
use strict;
use warnings;

# If run outside the normal test suite on an ASCII platform, you can
# just create a latin1_to_native() function that just returns its
# inputs, because that's the only function used from test.pl
require "test.pl";

# Test qr/\X/ and the \p{} regular expression constructs.  This file is
# constructed by mktables from the tables it generates, so if mktables is
# buggy, this won't necessarily catch those bugs.  Tests are generated for all
# feasible properties; a few aren't currently feasible; see
# is_code_point_usable() in mktables for details.

# Standard test packages are not used because this manipulates SIG_WARN.  It
# exits 0 if every non-skipped test succeeded; -1 if any failed.

my $Tests = 0;
my $Fails = 0;

sub Expect($$$$) {
    my $expected = shift;
    my $ord = shift;
    my $regex  = shift;
    my $warning_type = shift;   # Type of warning message, like 'deprecated'
                                # or empty if none
    my $line   = (caller)[2];
    $ord = ord(latin1_to_native(chr($ord)));

    # Convert the code point to hex form
    my $string = sprintf "\"\\x{%04X}\"", $ord;

    my @tests = "";

    # The first time through, use all warnings.  If the input should generate
    # a warning, add another time through with them turned off
    push @tests, "no warnings '$warning_type';" if $warning_type;

    foreach my $no_warnings (@tests) {

        # Store any warning messages instead of outputting them
        local $SIG{__WARN__} = $SIG{__WARN__};
        my $warning_message;
        $SIG{__WARN__} = sub { $warning_message = $_[0] };

        $Tests++;

        # A string eval is needed because of the 'no warnings'.
        # Assumes no parens in the regular expression
        my $result = eval "$no_warnings
                            my \$RegObj = qr($regex);
                            $string =~ \$RegObj ? 1 : 0";
        if (not defined $result) {
            print "not ok $Tests - couldn't compile /$regex/; line $line: $@\n";
            $Fails++;
        }
        elsif ($result ^ $expected) {
            print "not ok $Tests - expected $expected but got $result for $string =~ qr/$regex/; line $line\n";
            $Fails++;
        }
        elsif ($warning_message) {
            if (! $warning_type || ($warning_type && $no_warnings)) {
                print "not ok $Tests - for qr/$regex/ did not expect warning message '$warning_message'; line $line\n";
                $Fails++;
            }
            else {
                print "ok $Tests - expected and got a warning message for qr/$regex/; line $line\n";
            }
        }
        elsif ($warning_type && ! $no_warnings) {
            print "not ok $Tests - for qr/$regex/ expected a $warning_type warning message, but got none; line $line\n";
            $Fails++;
        }
        else {
            print "ok $Tests - got $result for $string =~ qr/$regex/; line $line\n";
        }
    }
    return;
}

sub Error($) {
    my $regex  = shift;
    $Tests++;
    if (eval { 'x' =~ qr/$regex/; 1 }) {
        $Fails++;
        my $line = (caller)[2];
        print "not ok $Tests - re compiled ok, but expected error for qr/$regex/; line $line: $@\n";
    }
    else {
        my $line = (caller)[2];
        print "ok $Tests - got and expected error for qr/$regex/; line $line\n";
    }
    return;
}

# GCBTest.txt character that separates grapheme clusters
my $breakable_utf8 = my $breakable = chr(0xF7);
utf8::upgrade($breakable_utf8);

# GCBTest.txt character that indicates that the adjoining code points are part
# of the same grapheme cluster
my $nobreak_utf8 = my $nobreak = chr(0xD7);
utf8::upgrade($nobreak_utf8);

sub Test_X($) {
    # Test qr/\X/ matches.  The input is a line from auxiliary/GCBTest.txt
    # Each such line is a sequence of code points given by their hex numbers,
    # separated by the two characters defined just before this subroutine that
    # indicate that either there can or cannot be a break between the adjacent
    # code points.  If there isn't a break, that means the sequence forms an
    # extended grapheme cluster, which means that \X should match the whole
    # thing.  If there is a break, \X should stop there.  This is all
    # converted by this routine into a match:
    #   $string =~ /(\X)/,
    # Each \X should match the next cluster; and that is what is checked.

    my $template = shift;

    my $line   = (caller)[2];

    # The line contains characters above the ASCII range, but in Latin1.  It
    # may or may not be in utf8, and if it is, it may or may not know it.  So,
    # convert these characters to 8 bits.  If knows is in utf8, simply
    # downgrade.
    if (utf8::is_utf8($template)) {
        utf8::downgrade($template);
    } else {

        # Otherwise, if it is in utf8, but doesn't know it, the next lines
        # convert the two problematic characters to their 8-bit equivalents.
        # If it isn't in utf8, they don't harm anything.
        use bytes;
        $template =~ s/$nobreak_utf8/$nobreak/g;
        $template =~ s/$breakable_utf8/$breakable/g;
    }

    # Get rid of the leading and trailing breakables
    $template =~ s/^ \s* $breakable \s* //x;
    $template =~ s/ \s* $breakable \s* $ //x;

    # And no-breaks become just a space.
    $template =~ s/ \s* $nobreak \s* / /xg;

    # Split the input into segments that are breakable between them.
    my @segments = split /\s*$breakable\s*/, $template;

    my $string = "";
    my $display_string = "";
    my @should_match;
    my @should_display;

    # Convert the code point sequence in each segment into a Perl string of
    # characters
    foreach my $segment (@segments) {
        my @code_points = split /\s+/, $segment;
        my $this_string = "";
        my $this_display = "";
        foreach my $code_point (@code_points) {
            $this_string .= latin1_to_native(chr(hex $code_point));
            $this_display .= "\\x{$code_point}";
        }

        # The next cluster should match the string in this segment.
        push @should_match, $this_string;
        push @should_display, $this_display;
        $string .= $this_string;
        $display_string .= $this_display;
    }

    # If a string can be represented in both non-ut8 and utf8, test both cases
    UPGRADE:
    for my $to_upgrade (0 .. 1) {

        if ($to_upgrade) {

            # If already in utf8, would just be a repeat
            next UPGRADE if utf8::is_utf8($string);

            utf8::upgrade($string);
        }

        # Finally, do the \X match.
        my @matches = $string =~ /(\X)/g;

        # Look through each matched cluster to verify that it matches what we
        # expect.
        my $min = (@matches < @should_match) ? @matches : @should_match;
        for my $i (0 .. $min - 1) {
            $Tests++;
            if ($matches[$i] eq $should_match[$i]) {
                print "ok $Tests - ";
                if ($i == 0) {
                    print "In \"$display_string\" =~ /(\\X)/g, \\X #1";
                } else {
                    print "And \\X #", $i + 1,
                }
                print " correctly matched $should_display[$i]; line $line\n";
            } else {
                $matches[$i] = join("", map { sprintf "\\x{%04X}", $_ }
                                                    unpack("U*", $matches[$i]));
                print "not ok $Tests - In \"$display_string\" =~ /(\\X)/g, \\X #",
                    $i + 1,
                    " should have matched $should_display[$i]",
                    " but instead matched $matches[$i]",
                    ".  Abandoning rest of line $line\n";
                next UPGRADE;
            }
        }

        # And the number of matches should equal the number of expected matches.
        $Tests++;
        if (@matches == @should_match) {
            print "ok $Tests - Nothing was left over; line $line\n";
        } else {
            print "not ok $Tests - There were ", scalar @should_match, " \\X matches expected, but got ", scalar @matches, " instead; line $line\n";
        }
    }

    return;
}

sub Finished() {
    print "1..$Tests\n";
    exit($Fails ? -1 : 0);
}

Error('\p{Script=InGreek}');    # Bug #69018
Test_X("1100 $nobreak 1161");  # Bug #70940
Expect(0, 0x2028, '\p{Print}', ""); # Bug # 71722
Expect(0, 0x2029, '\p{Print}', ""); # Bug # 71722
Expect(1, 0xFF10, '\p{XDigit}', ""); # Bug # 71726
Error('\p{Line_Break=:=	Space}');
Error('\P{Line_Break=:=	Space}');
Expect(1, 32, '\p{Line_Break=space}', "");
Expect(0, 32, '\p{^Line_Break=space}', "");
Expect(0, 32, '\P{Line_Break=space}', "");
Expect(1, 32, '\P{^Line_Break=space}', "");
Expect(0, 33, '\p{Line_Break=space}', "");
Expect(1, 33, '\p{^Line_Break=space}', "");
Expect(1, 33, '\P{Line_Break=space}', "");
Expect(0, 33, '\P{^Line_Break=space}', "");
Expect(1, 32, '\p{Line_Break=--Space}', "");
Expect(0, 32, '\p{^Line_Break=--Space}', "");
Expect(0, 32, '\P{Line_Break=--Space}', "");
Expect(1, 32, '\P{^Line_Break=--Space}', "");
Expect(0, 33, '\p{Line_Break=--Space}', "");
Expect(1, 33, '\p{^Line_Break=--Space}', "");
Expect(1, 33, '\P{Line_Break=--Space}', "");
Expect(0, 33, '\P{^Line_Break=--Space}', "");
Error('\p{Lb=		SP:=}');
Error('\P{Lb=		SP:=}');
Expect(1, 32, '\p{Lb=sp}', "");
Expect(0, 32, '\p{^Lb=sp}', "");
Expect(0, 32, '\P{Lb=sp}', "");
Expect(1, 32, '\P{^Lb=sp}', "");
Expect(0, 33, '\p{Lb=sp}', "");
Expect(1, 33, '\p{^Lb=sp}', "");
Expect(1, 33, '\P{Lb=sp}', "");
Expect(0, 33, '\P{^Lb=sp}', "");
Expect(1, 32, '\p{Lb:	SP}', "");
Expect(0, 32, '\p{^Lb:	SP}', "");
Expect(0, 32, '\P{Lb:	SP}', "");
Expect(1, 32, '\P{^Lb:	SP}', "");
Expect(0, 33, '\p{Lb:	SP}', "");
Expect(1, 33, '\p{^Lb:	SP}', "");
Expect(1, 33, '\P{Lb:	SP}', "");
Expect(0, 33, '\P{^Lb:	SP}', "");
Error('\p{Is_Line_Break=/a/		SPACE}');
Error('\P{Is_Line_Break=/a/		SPACE}');
Expect(1, 32, '\p{Is_Line_Break=space}', "");
Expect(0, 32, '\p{^Is_Line_Break=space}', "");
Expect(0, 32, '\P{Is_Line_Break=space}', "");
Expect(1, 32, '\P{^Is_Line_Break=space}', "");
Expect(0, 33, '\p{Is_Line_Break=space}', "");
Expect(1, 33, '\p{^Is_Line_Break=space}', "");
Expect(1, 33, '\P{Is_Line_Break=space}', "");
Expect(0, 33, '\P{^Is_Line_Break=space}', "");
Expect(1, 32, '\p{Is_Line_Break=-Space}', "");
Expect(0, 32, '\p{^Is_Line_Break=-Space}', "");
Expect(0, 32, '\P{Is_Line_Break=-Space}', "");
Expect(1, 32, '\P{^Is_Line_Break=-Space}', "");
Expect(0, 33, '\p{Is_Line_Break=-Space}', "");
Expect(1, 33, '\p{^Is_Line_Break=-Space}', "");
Expect(1, 33, '\P{Is_Line_Break=-Space}', "");
Expect(0, 33, '\P{^Is_Line_Break=-Space}', "");
Error('\p{Is_Lb=		SP/a/}');
Error('\P{Is_Lb=		SP/a/}');
Expect(1, 32, '\p{Is_Lb=sp}', "");
Expect(0, 32, '\p{^Is_Lb=sp}', "");
Expect(0, 32, '\P{Is_Lb=sp}', "");
Expect(1, 32, '\P{^Is_Lb=sp}', "");
Expect(0, 33, '\p{Is_Lb=sp}', "");
Expect(1, 33, '\p{^Is_Lb=sp}', "");
Expect(1, 33, '\P{Is_Lb=sp}', "");
Expect(0, 33, '\P{^Is_Lb=sp}', "");
Expect(1, 32, '\p{Is_Lb=_-sp}', "");
Expect(0, 32, '\p{^Is_Lb=_-sp}', "");
Expect(0, 32, '\P{Is_Lb=_-sp}', "");
Expect(1, 32, '\P{^Is_Lb=_-sp}', "");
Expect(0, 33, '\p{Is_Lb=_-sp}', "");
Expect(1, 33, '\p{^Is_Lb=_-sp}', "");
Expect(1, 33, '\P{Is_Lb=_-sp}', "");
Expect(0, 33, '\P{^Is_Lb=_-sp}', "");
Error('\p{Line_Break=_ GLUE:=}');
Error('\P{Line_Break=_ GLUE:=}');
Expect(1, 8239, '\p{Line_Break=glue}', "");
Expect(0, 8239, '\p{^Line_Break=glue}', "");
Expect(0, 8239, '\P{Line_Break=glue}', "");
Expect(1, 8239, '\P{^Line_Break=glue}', "");
Expect(0, 8240, '\p{Line_Break=glue}', "");
Expect(1, 8240, '\p{^Line_Break=glue}', "");
Expect(1, 8240, '\P{Line_Break=glue}', "");
Expect(0, 8240, '\P{^Line_Break=glue}', "");
Expect(1, 8239, '\p{Line_Break=_	glue}', "");
Expect(0, 8239, '\p{^Line_Break=_	glue}', "");
Expect(0, 8239, '\P{Line_Break=_	glue}', "");
Expect(1, 8239, '\P{^Line_Break=_	glue}', "");
Expect(0, 8240, '\p{Line_Break=_	glue}', "");
Expect(1, 8240, '\p{^Line_Break=_	glue}', "");
Expect(1, 8240, '\P{Line_Break=_	glue}', "");
Expect(0, 8240, '\P{^Line_Break=_	glue}', "");
Error('\p{Lb:   /a/	GL}');
Error('\P{Lb:   /a/	GL}');
Expect(1, 8239, '\p{Lb=gl}', "");
Expect(0, 8239, '\p{^Lb=gl}', "");
Expect(0, 8239, '\P{Lb=gl}', "");
Expect(1, 8239, '\P{^Lb=gl}', "");
Expect(0, 8240, '\p{Lb=gl}', "");
Expect(1, 8240, '\p{^Lb=gl}', "");
Expect(1, 8240, '\P{Lb=gl}', "");
Expect(0, 8240, '\P{^Lb=gl}', "");
Expect(1, 8239, '\p{Lb= gl}', "");
Expect(0, 8239, '\p{^Lb= gl}', "");
Expect(0, 8239, '\P{Lb= gl}', "");
Expect(1, 8239, '\P{^Lb= gl}', "");
Expect(0, 8240, '\p{Lb= gl}', "");
Expect(1, 8240, '\p{^Lb= gl}', "");
Expect(1, 8240, '\P{Lb= gl}', "");
Expect(0, 8240, '\P{^Lb= gl}', "");
Error('\p{Is_Line_Break::=-	Glue}');
Error('\P{Is_Line_Break::=-	Glue}');
Expect(1, 8239, '\p{Is_Line_Break=glue}', "");
Expect(0, 8239, '\p{^Is_Line_Break=glue}', "");
Expect(0, 8239, '\P{Is_Line_Break=glue}', "");
Expect(1, 8239, '\P{^Is_Line_Break=glue}', "");
Expect(0, 8240, '\p{Is_Line_Break=glue}', "");
Expect(1, 8240, '\p{^Is_Line_Break=glue}', "");
Expect(1, 8240, '\P{Is_Line_Break=glue}', "");
Expect(0, 8240, '\P{^Is_Line_Break=glue}', "");
Expect(1, 8239, '\p{Is_Line_Break:  Glue}', "");
Expect(0, 8239, '\p{^Is_Line_Break:  Glue}', "");
Expect(0, 8239, '\P{Is_Line_Break:  Glue}', "");
Expect(1, 8239, '\P{^Is_Line_Break:  Glue}', "");
Expect(0, 8240, '\p{Is_Line_Break:  Glue}', "");
Expect(1, 8240, '\p{^Is_Line_Break:  Glue}', "");
Expect(1, 8240, '\P{Is_Line_Break:  Glue}', "");
Expect(0, 8240, '\P{^Is_Line_Break:  Glue}', "");
Error('\p{Is_Lb=	:=GL}');
Error('\P{Is_Lb=	:=GL}');
Expect(1, 8239, '\p{Is_Lb=gl}', "");
Expect(0, 8239, '\p{^Is_Lb=gl}', "");
Expect(0, 8239, '\P{Is_Lb=gl}', "");
Expect(1, 8239, '\P{^Is_Lb=gl}', "");
Expect(0, 8240, '\p{Is_Lb=gl}', "");
Expect(1, 8240, '\p{^Is_Lb=gl}', "");
Expect(1, 8240, '\P{Is_Lb=gl}', "");
Expect(0, 8240, '\P{^Is_Lb=gl}', "");
Expect(1, 8239, '\p{Is_Lb=_	GL}', "");
Expect(0, 8239, '\p{^Is_Lb=_	GL}', "");
Expect(0, 8239, '\P{Is_Lb=_	GL}', "");
Expect(1, 8239, '\P{^Is_Lb=_	GL}', "");
Expect(0, 8240, '\p{Is_Lb=_	GL}', "");
Expect(1, 8240, '\p{^Is_Lb=_	GL}', "");
Expect(1, 8240, '\P{Is_Lb=_	GL}', "");
Expect(0, 8240, '\P{^Is_Lb=_	GL}', "");
Error('\p{Line_Break= Surrogate:=}');
Error('\P{Line_Break= Surrogate:=}');
Expect(1, 57343, '\p{Line_Break=surrogate}', 'deprecated');
Expect(0, 57343, '\p{^Line_Break=surrogate}', 'deprecated');
Expect(0, 57343, '\P{Line_Break=surrogate}', 'deprecated');
Expect(1, 57343, '\P{^Line_Break=surrogate}', 'deprecated');
Expect(0, 57344, '\p{Line_Break=surrogate}', 'deprecated');
Expect(1, 57344, '\p{^Line_Break=surrogate}', 'deprecated');
Expect(1, 57344, '\P{Line_Break=surrogate}', 'deprecated');
Expect(0, 57344, '\P{^Line_Break=surrogate}', 'deprecated');
Expect(1, 57343, '\p{Line_Break:   -_Surrogate}', 'deprecated');
Expect(0, 57343, '\p{^Line_Break:   -_Surrogate}', 'deprecated');
Expect(0, 57343, '\P{Line_Break:   -_Surrogate}', 'deprecated');
Expect(1, 57343, '\P{^Line_Break:   -_Surrogate}', 'deprecated');
Expect(0, 57344, '\p{Line_Break:   -_Surrogate}', 'deprecated');
Expect(1, 57344, '\p{^Line_Break:   -_Surrogate}', 'deprecated');
Expect(1, 57344, '\P{Line_Break:   -_Surrogate}', 'deprecated');
Expect(0, 57344, '\P{^Line_Break:   -_Surrogate}', 'deprecated');
Error('\p{Lb=-_sg/a/}');
Error('\P{Lb=-_sg/a/}');
Expect(1, 57343, '\p{Lb=sg}', 'deprecated');
Expect(0, 57343, '\p{^Lb=sg}', 'deprecated');
Expect(0, 57343, '\P{Lb=sg}', 'deprecated');
Expect(1, 57343, '\P{^Lb=sg}', 'deprecated');
Expect(0, 57344, '\p{Lb=sg}', 'deprecated');
Expect(1, 57344, '\p{^Lb=sg}', 'deprecated');
Expect(1, 57344, '\P{Lb=sg}', 'deprecated');
Expect(0, 57344, '\P{^Lb=sg}', 'deprecated');
Expect(1, 57343, '\p{Lb=	 SG}', 'deprecated');
Expect(0, 57343, '\p{^Lb=	 SG}', 'deprecated');
Expect(0, 57343, '\P{Lb=	 SG}', 'deprecated');
Expect(1, 57343, '\P{^Lb=	 SG}', 'deprecated');
Expect(0, 57344, '\p{Lb=	 SG}', 'deprecated');
Expect(1, 57344, '\p{^Lb=	 SG}', 'deprecated');
Expect(1, 57344, '\P{Lb=	 SG}', 'deprecated');
Expect(0, 57344, '\P{^Lb=	 SG}', 'deprecated');
Error('\p{Is_Line_Break:	:=_SURROGATE}');
Error('\P{Is_Line_Break:	:=_SURROGATE}');
Expect(1, 57343, '\p{Is_Line_Break=surrogate}', 'deprecated');
Expect(0, 57343, '\p{^Is_Line_Break=surrogate}', 'deprecated');
Expect(0, 57343, '\P{Is_Line_Break=surrogate}', 'deprecated');
Expect(1, 57343, '\P{^Is_Line_Break=surrogate}', 'deprecated');
Expect(0, 57344, '\p{Is_Line_Break=surrogate}', 'deprecated');
Expect(1, 57344, '\p{^Is_Line_Break=surrogate}', 'deprecated');
Expect(1, 57344, '\P{Is_Line_Break=surrogate}', 'deprecated');
Expect(0, 57344, '\P{^Is_Line_Break=surrogate}', 'deprecated');
Expect(1, 57343, '\p{Is_Line_Break= SURROGATE}', 'deprecated');
Expect(0, 57343, '\p{^Is_Line_Break= SURROGATE}', 'deprecated');
Expect(0, 57343, '\P{Is_Line_Break= SURROGATE}', 'deprecated');
Expect(1, 57343, '\P{^Is_Line_Break= SURROGATE}', 'deprecated');
Expect(0, 57344, '\p{Is_Line_Break= SURROGATE}', 'deprecated');
Expect(1, 57344, '\p{^Is_Line_Break= SURROGATE}', 'deprecated');
Expect(1, 57344, '\P{Is_Line_Break= SURROGATE}', 'deprecated');
Expect(0, 57344, '\P{^Is_Line_Break= SURROGATE}', 'deprecated');
Error('\p{Is_Lb=/a/--sg}');
Error('\P{Is_Lb=/a/--sg}');
Expect(1, 57343, '\p{Is_Lb=sg}', 'deprecated');
Expect(0, 57343, '\p{^Is_Lb=sg}', 'deprecated');
Expect(0, 57343, '\P{Is_Lb=sg}', 'deprecated');
Expect(1, 57343, '\P{^Is_Lb=sg}', 'deprecated');
Expect(0, 57344, '\p{Is_Lb=sg}', 'deprecated');
Expect(1, 57344, '\p{^Is_Lb=sg}', 'deprecated');
Expect(1, 57344, '\P{Is_Lb=sg}', 'deprecated');
Expect(0, 57344, '\P{^Is_Lb=sg}', 'deprecated');
Expect(1, 57343, '\p{Is_Lb: sg}', 'deprecated');
Expect(0, 57343, '\p{^Is_Lb: sg}', 'deprecated');
Expect(0, 57343, '\P{Is_Lb: sg}', 'deprecated');
Expect(1, 57343, '\P{^Is_Lb: sg}', 'deprecated');
Expect(0, 57344, '\p{Is_Lb: sg}', 'deprecated');
Expect(1, 57344, '\p{^Is_Lb: sg}', 'deprecated');
Expect(1, 57344, '\P{Is_Lb: sg}', 'deprecated');
Expect(0, 57344, '\P{^Is_Lb: sg}', 'deprecated');
Error('\p{Line_Break= Unknown/a/}');
Error('\P{Line_Break= Unknown/a/}');
Expect(1, 918000, '\p{Line_Break=unknown}', "");
Expect(0, 918000, '\p{^Line_Break=unknown}', "");
Expect(0, 918000, '\P{Line_Break=unknown}', "");
Expect(1, 918000, '\P{^Line_Break=unknown}', "");
Expect(0, 917999, '\p{Line_Break=unknown}', "");
Expect(1, 917999, '\p{^Line_Break=unknown}', "");
Expect(1, 917999, '\P{Line_Break=unknown}', "");
Expect(0, 917999, '\P{^Line_Break=unknown}', "");
Expect(1, 918000, '\p{Line_Break=-	unknown}', "");
Expect(0, 918000, '\p{^Line_Break=-	unknown}', "");
Expect(0, 918000, '\P{Line_Break=-	unknown}', "");
Expect(1, 918000, '\P{^Line_Break=-	unknown}', "");
Expect(0, 917999, '\p{Line_Break=-	unknown}', "");
Expect(1, 917999, '\p{^Line_Break=-	unknown}', "");
Expect(1, 917999, '\P{Line_Break=-	unknown}', "");
Expect(0, 917999, '\P{^Line_Break=-	unknown}', "");
Error('\p{Lb=:=XX}');
Error('\P{Lb=:=XX}');
Expect(1, 918000, '\p{Lb=xx}', "");
Expect(0, 918000, '\p{^Lb=xx}', "");
Expect(0, 918000, '\P{Lb=xx}', "");
Expect(1, 918000, '\P{^Lb=xx}', "");
Expect(0, 917999, '\p{Lb=xx}', "");
Expect(1, 917999, '\p{^Lb=xx}', "");
Expect(1, 917999, '\P{Lb=xx}', "");
Expect(0, 917999, '\P{^Lb=xx}', "");
Expect(1, 918000, '\p{Lb=__XX}', "");
Expect(0, 918000, '\p{^Lb=__XX}', "");
Expect(0, 918000, '\P{Lb=__XX}', "");
Expect(1, 918000, '\P{^Lb=__XX}', "");
Expect(0, 917999, '\p{Lb=__XX}', "");
Expect(1, 917999, '\p{^Lb=__XX}', "");
Expect(1, 917999, '\P{Lb=__XX}', "");
Expect(0, 917999, '\P{^Lb=__XX}', "");
Error('\p{Is_Line_Break=	UNKNOWN/a/}');
Error('\P{Is_Line_Break=	UNKNOWN/a/}');
Expect(1, 918000, '\p{Is_Line_Break:unknown}', "");
Expect(0, 918000, '\p{^Is_Line_Break:unknown}', "");
Expect(0, 918000, '\P{Is_Line_Break:unknown}', "");
Expect(1, 918000, '\P{^Is_Line_Break:unknown}', "");
Expect(0, 917999, '\p{Is_Line_Break:unknown}', "");
Expect(1, 917999, '\p{^Is_Line_Break:unknown}', "");
Expect(1, 917999, '\P{Is_Line_Break:unknown}', "");
Expect(0, 917999, '\P{^Is_Line_Break:unknown}', "");
Expect(1, 918000, '\p{Is_Line_Break=_-Unknown}', "");
Expect(0, 918000, '\p{^Is_Line_Break=_-Unknown}', "");
Expect(0, 918000, '\P{Is_Line_Break=_-Unknown}', "");
Expect(1, 918000, '\P{^Is_Line_Break=_-Unknown}', "");
Expect(0, 917999, '\p{Is_Line_Break=_-Unknown}', "");
Expect(1, 917999, '\p{^Is_Line_Break=_-Unknown}', "");
Expect(1, 917999, '\P{Is_Line_Break=_-Unknown}', "");
Expect(0, 917999, '\P{^Is_Line_Break=_-Unknown}', "");
Error('\p{Is_Lb=/a/		XX}');
Error('\P{Is_Lb=/a/		XX}');
Expect(1, 918000, '\p{Is_Lb=xx}', "");
Expect(0, 918000, '\p{^Is_Lb=xx}', "");
Expect(0, 918000, '\P{Is_Lb=xx}', "");
Expect(1, 918000, '\P{^Is_Lb=xx}', "");
Expect(0, 917999, '\p{Is_Lb=xx}', "");
Expect(1, 917999, '\p{^Is_Lb=xx}', "");
Expect(1, 917999, '\P{Is_Lb=xx}', "");
Expect(0, 917999, '\P{^Is_Lb=xx}', "");
Expect(1, 918000, '\p{Is_Lb=	-XX}', "");
Expect(0, 918000, '\p{^Is_Lb=	-XX}', "");
Expect(0, 918000, '\P{Is_Lb=	-XX}', "");
Expect(1, 918000, '\P{^Is_Lb=	-XX}', "");
Expect(0, 917999, '\p{Is_Lb=	-XX}', "");
Expect(1, 917999, '\p{^Is_Lb=	-XX}', "");
Expect(1, 917999, '\P{Is_Lb=	-XX}', "");
Expect(0, 917999, '\P{^Is_Lb=	-XX}', "");
Error('\p{Line_Break=-numeric:=}');
Error('\P{Line_Break=-numeric:=}');
Expect(1, 120831, '\p{Line_Break=numeric}', "");
Expect(0, 120831, '\p{^Line_Break=numeric}', "");
Expect(0, 120831, '\P{Line_Break=numeric}', "");
Expect(1, 120831, '\P{^Line_Break=numeric}', "");
Expect(0, 120832, '\p{Line_Break=numeric}', "");
Expect(1, 120832, '\p{^Line_Break=numeric}', "");
Expect(1, 120832, '\P{Line_Break=numeric}', "");
Expect(0, 120832, '\P{^Line_Break=numeric}', "");
Expect(1, 120831, '\p{Line_Break= NUMERIC}', "");
Expect(0, 120831, '\p{^Line_Break= NUMERIC}', "");
Expect(0, 120831, '\P{Line_Break= NUMERIC}', "");
Expect(1, 120831, '\P{^Line_Break= NUMERIC}', "");
Expect(0, 120832, '\p{Line_Break= NUMERIC}', "");
Expect(1, 120832, '\p{^Line_Break= NUMERIC}', "");
Expect(1, 120832, '\P{Line_Break= NUMERIC}', "");
Expect(0, 120832, '\P{^Line_Break= NUMERIC}', "");
Error('\p{Lb=/a/	-NU}');
Error('\P{Lb=/a/	-NU}');
Expect(1, 120831, '\p{Lb=nu}', "");
Expect(0, 120831, '\p{^Lb=nu}', "");
Expect(0, 120831, '\P{Lb=nu}', "");
Expect(1, 120831, '\P{^Lb=nu}', "");
Expect(0, 120832, '\p{Lb=nu}', "");
Expect(1, 120832, '\p{^Lb=nu}', "");
Expect(1, 120832, '\P{Lb=nu}', "");
Expect(0, 120832, '\P{^Lb=nu}', "");
Expect(1, 120831, '\p{Lb=-nu}', "");
Expect(0, 120831, '\p{^Lb=-nu}', "");
Expect(0, 120831, '\P{Lb=-nu}', "");
Expect(1, 120831, '\P{^Lb=-nu}', "");
Expect(0, 120832, '\p{Lb=-nu}', "");
Expect(1, 120832, '\p{^Lb=-nu}', "");
Expect(1, 120832, '\P{Lb=-nu}', "");
Expect(0, 120832, '\P{^Lb=-nu}', "");
Error('\p{Is_Line_Break=	 NUMERIC:=}');
Error('\P{Is_Line_Break=	 NUMERIC:=}');
Expect(1, 120831, '\p{Is_Line_Break=numeric}', "");
Expect(0, 120831, '\p{^Is_Line_Break=numeric}', "");
Expect(0, 120831, '\P{Is_Line_Break=numeric}', "");
Expect(1, 120831, '\P{^Is_Line_Break=numeric}', "");
Expect(0, 120832, '\p{Is_Line_Break=numeric}', "");
Expect(1, 120832, '\p{^Is_Line_Break=numeric}', "");
Expect(1, 120832, '\P{Is_Line_Break=numeric}', "");
Expect(0, 120832, '\P{^Is_Line_Break=numeric}', "");
Expect(1, 120831, '\p{Is_Line_Break=  Numeric}', "");
Expect(0, 120831, '\p{^Is_Line_Break=  Numeric}', "");
Expect(0, 120831, '\P{Is_Line_Break=  Numeric}', "");
Expect(1, 120831, '\P{^Is_Line_Break=  Numeric}', "");
Expect(0, 120832, '\p{Is_Line_Break=  Numeric}', "");
Expect(1, 120832, '\p{^Is_Line_Break=  Numeric}', "");
Expect(1, 120832, '\P{Is_Line_Break=  Numeric}', "");
Expect(0, 120832, '\P{^Is_Line_Break=  Numeric}', "");
Error('\p{Is_Lb=-NU/a/}');
Error('\P{Is_Lb=-NU/a/}');
Expect(1, 120831, '\p{Is_Lb=nu}', "");
Expect(0, 120831, '\p{^Is_Lb=nu}', "");
Expect(0, 120831, '\P{Is_Lb=nu}', "");
Expect(1, 120831, '\P{^Is_Lb=nu}', "");
Expect(0, 120832, '\p{Is_Lb=nu}', "");
Expect(1, 120832, '\p{^Is_Lb=nu}', "");
Expect(1, 120832, '\P{Is_Lb=nu}', "");
Expect(0, 120832, '\P{^Is_Lb=nu}', "");
Expect(1, 120831, '\p{Is_Lb=NU}', "");
Expect(0, 120831, '\p{^Is_Lb=NU}', "");
Expect(0, 120831, '\P{Is_Lb=NU}', "");
Expect(1, 120831, '\P{^Is_Lb=NU}', "");
Expect(0, 120832, '\p{Is_Lb=NU}', "");
Expect(1, 120832, '\p{^Is_Lb=NU}', "");
Expect(1, 120832, '\P{Is_Lb=NU}', "");
Expect(0, 120832, '\P{^Is_Lb=NU}', "");
Error('\p{Line_Break=:=- Break_before}');
Error('\P{Line_Break=:=- Break_before}');
Expect(1, 43125, '\p{Line_Break=breakbefore}', "");
Expect(0, 43125, '\p{^Line_Break=breakbefore}', "");
Expect(0, 43125, '\P{Line_Break=breakbefore}', "");
Expect(1, 43125, '\P{^Line_Break=breakbefore}', "");
Expect(0, 43126, '\p{Line_Break=breakbefore}', "");
Expect(1, 43126, '\p{^Line_Break=breakbefore}', "");
Expect(1, 43126, '\P{Line_Break=breakbefore}', "");
Expect(0, 43126, '\P{^Line_Break=breakbefore}', "");
Expect(1, 43125, '\p{Line_Break=-_Break_BEFORE}', "");
Expect(0, 43125, '\p{^Line_Break=-_Break_BEFORE}', "");
Expect(0, 43125, '\P{Line_Break=-_Break_BEFORE}', "");
Expect(1, 43125, '\P{^Line_Break=-_Break_BEFORE}', "");
Expect(0, 43126, '\p{Line_Break=-_Break_BEFORE}', "");
Expect(1, 43126, '\p{^Line_Break=-_Break_BEFORE}', "");
Expect(1, 43126, '\P{Line_Break=-_Break_BEFORE}', "");
Expect(0, 43126, '\P{^Line_Break=-_Break_BEFORE}', "");
Error('\p{Lb: /a/ bb}');
Error('\P{Lb: /a/ bb}');
Expect(1, 43125, '\p{Lb=bb}', "");
Expect(0, 43125, '\p{^Lb=bb}', "");
Expect(0, 43125, '\P{Lb=bb}', "");
Expect(1, 43125, '\P{^Lb=bb}', "");
Expect(0, 43126, '\p{Lb=bb}', "");
Expect(1, 43126, '\p{^Lb=bb}', "");
Expect(1, 43126, '\P{Lb=bb}', "");
Expect(0, 43126, '\P{^Lb=bb}', "");
Expect(1, 43125, '\p{Lb=		BB}', "");
Expect(0, 43125, '\p{^Lb=		BB}', "");
Expect(0, 43125, '\P{Lb=		BB}', "");
Expect(1, 43125, '\P{^Lb=		BB}', "");
Expect(0, 43126, '\p{Lb=		BB}', "");
Expect(1, 43126, '\p{^Lb=		BB}', "");
Expect(1, 43126, '\P{Lb=		BB}', "");
Expect(0, 43126, '\P{^Lb=		BB}', "");
Error('\p{Is_Line_Break=	:=Break_BEFORE}');
Error('\P{Is_Line_Break=	:=Break_BEFORE}');
Expect(1, 43125, '\p{Is_Line_Break:breakbefore}', "");
Expect(0, 43125, '\p{^Is_Line_Break:breakbefore}', "");
Expect(0, 43125, '\P{Is_Line_Break:breakbefore}', "");
Expect(1, 43125, '\P{^Is_Line_Break:breakbefore}', "");
Expect(0, 43126, '\p{Is_Line_Break:breakbefore}', "");
Expect(1, 43126, '\p{^Is_Line_Break:breakbefore}', "");
Expect(1, 43126, '\P{Is_Line_Break:breakbefore}', "");
Expect(0, 43126, '\P{^Is_Line_Break:breakbefore}', "");
Expect(1, 43125, '\p{Is_Line_Break:   BREAK_BEFORE}', "");
Expect(0, 43125, '\p{^Is_Line_Break:   BREAK_BEFORE}', "");
Expect(0, 43125, '\P{Is_Line_Break:   BREAK_BEFORE}', "");
Expect(1, 43125, '\P{^Is_Line_Break:   BREAK_BEFORE}', "");
Expect(0, 43126, '\p{Is_Line_Break:   BREAK_BEFORE}', "");
Expect(1, 43126, '\p{^Is_Line_Break:   BREAK_BEFORE}', "");
Expect(1, 43126, '\P{Is_Line_Break:   BREAK_BEFORE}', "");
Expect(0, 43126, '\P{^Is_Line_Break:   BREAK_BEFORE}', "");
Error('\p{Is_Lb=-/a/BB}');
Error('\P{Is_Lb=-/a/BB}');
Expect(1, 43125, '\p{Is_Lb=bb}', "");
Expect(0, 43125, '\p{^Is_Lb=bb}', "");
Expect(0, 43125, '\P{Is_Lb=bb}', "");
Expect(1, 43125, '\P{^Is_Lb=bb}', "");
Expect(0, 43126, '\p{Is_Lb=bb}', "");
Expect(1, 43126, '\p{^Is_Lb=bb}', "");
Expect(1, 43126, '\P{Is_Lb=bb}', "");
Expect(0, 43126, '\P{^Is_Lb=bb}', "");
Expect(1, 43125, '\p{Is_Lb= _BB}', "");
Expect(0, 43125, '\p{^Is_Lb= _BB}', "");
Expect(0, 43125, '\P{Is_Lb= _BB}', "");
Expect(1, 43125, '\P{^Is_Lb= _BB}', "");
Expect(0, 43126, '\p{Is_Lb= _BB}', "");
Expect(1, 43126, '\p{^Is_Lb= _BB}', "");
Expect(1, 43126, '\P{Is_Lb= _BB}', "");
Expect(0, 43126, '\P{^Is_Lb= _BB}', "");
Error('\p{Line_Break=_ZWSpace:=}');
Error('\P{Line_Break=_ZWSpace:=}');
Expect(1, 8203, '\p{Line_Break=zwspace}', "");
Expect(0, 8203, '\p{^Line_Break=zwspace}', "");
Expect(0, 8203, '\P{Line_Break=zwspace}', "");
Expect(1, 8203, '\P{^Line_Break=zwspace}', "");
Expect(0, 8204, '\p{Line_Break=zwspace}', "");
Expect(1, 8204, '\p{^Line_Break=zwspace}', "");
Expect(1, 8204, '\P{Line_Break=zwspace}', "");
Expect(0, 8204, '\P{^Line_Break=zwspace}', "");
Expect(1, 8203, '\p{Line_Break=-	ZWSpace}', "");
Expect(0, 8203, '\p{^Line_Break=-	ZWSpace}', "");
Expect(0, 8203, '\P{Line_Break=-	ZWSpace}', "");
Expect(1, 8203, '\P{^Line_Break=-	ZWSpace}', "");
Expect(0, 8204, '\p{Line_Break=-	ZWSpace}', "");
Expect(1, 8204, '\p{^Line_Break=-	ZWSpace}', "");
Expect(1, 8204, '\P{Line_Break=-	ZWSpace}', "");
Expect(0, 8204, '\P{^Line_Break=-	ZWSpace}', "");
Error('\p{Lb:/a/ _ZW}');
Error('\P{Lb:/a/ _ZW}');
Expect(1, 8203, '\p{Lb=zw}', "");
Expect(0, 8203, '\p{^Lb=zw}', "");
Expect(0, 8203, '\P{Lb=zw}', "");
Expect(1, 8203, '\P{^Lb=zw}', "");
Expect(0, 8204, '\p{Lb=zw}', "");
Expect(1, 8204, '\p{^Lb=zw}', "");
Expect(1, 8204, '\P{Lb=zw}', "");
Expect(0, 8204, '\P{^Lb=zw}', "");
Expect(1, 8203, '\p{Lb=_ZW}', "");
Expect(0, 8203, '\p{^Lb=_ZW}', "");
Expect(0, 8203, '\P{Lb=_ZW}', "");
Expect(1, 8203, '\P{^Lb=_ZW}', "");
Expect(0, 8204, '\p{Lb=_ZW}', "");
Expect(1, 8204, '\p{^Lb=_ZW}', "");
Expect(1, 8204, '\P{Lb=_ZW}', "");
Expect(0, 8204, '\P{^Lb=_ZW}', "");
Error('\p{Is_Line_Break=/a/  ZWSpace}');
Error('\P{Is_Line_Break=/a/  ZWSpace}');
Expect(1, 8203, '\p{Is_Line_Break=zwspace}', "");
Expect(0, 8203, '\p{^Is_Line_Break=zwspace}', "");
Expect(0, 8203, '\P{Is_Line_Break=zwspace}', "");
Expect(1, 8203, '\P{^Is_Line_Break=zwspace}', "");
Expect(0, 8204, '\p{Is_Line_Break=zwspace}', "");
Expect(1, 8204, '\p{^Is_Line_Break=zwspace}', "");
Expect(1, 8204, '\P{Is_Line_Break=zwspace}', "");
Expect(0, 8204, '\P{^Is_Line_Break=zwspace}', "");
Expect(1, 8203, '\p{Is_Line_Break=		ZWSPACE}', "");
Expect(0, 8203, '\p{^Is_Line_Break=		ZWSPACE}', "");
Expect(0, 8203, '\P{Is_Line_Break=		ZWSPACE}', "");
Expect(1, 8203, '\P{^Is_Line_Break=		ZWSPACE}', "");
Expect(0, 8204, '\p{Is_Line_Break=		ZWSPACE}', "");
Expect(1, 8204, '\p{^Is_Line_Break=		ZWSPACE}', "");
Expect(1, 8204, '\P{Is_Line_Break=		ZWSPACE}', "");
Expect(0, 8204, '\P{^Is_Line_Break=		ZWSPACE}', "");
Error('\p{Is_Lb=/a/_ZW}');
Error('\P{Is_Lb=/a/_ZW}');
Expect(1, 8203, '\p{Is_Lb=zw}', "");
Expect(0, 8203, '\p{^Is_Lb=zw}', "");
Expect(0, 8203, '\P{Is_Lb=zw}', "");
Expect(1, 8203, '\P{^Is_Lb=zw}', "");
Expect(0, 8204, '\p{Is_Lb=zw}', "");
Expect(1, 8204, '\p{^Is_Lb=zw}', "");
Expect(1, 8204, '\P{Is_Lb=zw}', "");
Expect(0, 8204, '\P{^Is_Lb=zw}', "");
Expect(1, 8203, '\p{Is_Lb: ZW}', "");
Expect(0, 8203, '\p{^Is_Lb: ZW}', "");
Expect(0, 8203, '\P{Is_Lb: ZW}', "");
Expect(1, 8203, '\P{^Is_Lb: ZW}', "");
Expect(0, 8204, '\p{Is_Lb: ZW}', "");
Expect(1, 8204, '\p{^Is_Lb: ZW}', "");
Expect(1, 8204, '\P{Is_Lb: ZW}', "");
Expect(0, 8204, '\P{^Is_Lb: ZW}', "");
Error('\p{Line_Break=/a/__jl}');
Error('\P{Line_Break=/a/__jl}');
Expect(1, 43388, '\p{Line_Break=jl}', "");
Expect(0, 43388, '\p{^Line_Break=jl}', "");
Expect(0, 43388, '\P{Line_Break=jl}', "");
Expect(1, 43388, '\P{^Line_Break=jl}', "");
Expect(0, 43389, '\p{Line_Break=jl}', "");
Expect(1, 43389, '\p{^Line_Break=jl}', "");
Expect(1, 43389, '\P{Line_Break=jl}', "");
Expect(0, 43389, '\P{^Line_Break=jl}', "");
Expect(1, 43388, '\p{Line_Break= 	jl}', "");
Expect(0, 43388, '\p{^Line_Break= 	jl}', "");
Expect(0, 43388, '\P{Line_Break= 	jl}', "");
Expect(1, 43388, '\P{^Line_Break= 	jl}', "");
Expect(0, 43389, '\p{Line_Break= 	jl}', "");
Expect(1, 43389, '\p{^Line_Break= 	jl}', "");
Expect(1, 43389, '\P{Line_Break= 	jl}', "");
Expect(0, 43389, '\P{^Line_Break= 	jl}', "");
Error('\p{Lb=:=	-jl}');
Error('\P{Lb=:=	-jl}');
Expect(1, 43388, '\p{Lb=jl}', "");
Expect(0, 43388, '\p{^Lb=jl}', "");
Expect(0, 43388, '\P{Lb=jl}', "");
Expect(1, 43388, '\P{^Lb=jl}', "");
Expect(0, 43389, '\p{Lb=jl}', "");
Expect(1, 43389, '\p{^Lb=jl}', "");
Expect(1, 43389, '\P{Lb=jl}', "");
Expect(0, 43389, '\P{^Lb=jl}', "");
Expect(1, 43388, '\p{Lb=_JL}', "");
Expect(0, 43388, '\p{^Lb=_JL}', "");
Expect(0, 43388, '\P{Lb=_JL}', "");
Expect(1, 43388, '\P{^Lb=_JL}', "");
Expect(0, 43389, '\p{Lb=_JL}', "");
Expect(1, 43389, '\p{^Lb=_JL}', "");
Expect(1, 43389, '\P{Lb=_JL}', "");
Expect(0, 43389, '\P{^Lb=_JL}', "");
Error('\p{Is_Line_Break= -JL/a/}');
Error('\P{Is_Line_Break= -JL/a/}');
Expect(1, 43388, '\p{Is_Line_Break=jl}', "");
Expect(0, 43388, '\p{^Is_Line_Break=jl}', "");
Expect(0, 43388, '\P{Is_Line_Break=jl}', "");
Expect(1, 43388, '\P{^Is_Line_Break=jl}', "");
Expect(0, 43389, '\p{Is_Line_Break=jl}', "");
Expect(1, 43389, '\p{^Is_Line_Break=jl}', "");
Expect(1, 43389, '\P{Is_Line_Break=jl}', "");
Expect(0, 43389, '\P{^Is_Line_Break=jl}', "");
Expect(1, 43388, '\p{Is_Line_Break= _JL}', "");
Expect(0, 43388, '\p{^Is_Line_Break= _JL}', "");
Expect(0, 43388, '\P{Is_Line_Break= _JL}', "");
Expect(1, 43388, '\P{^Is_Line_Break= _JL}', "");
Expect(0, 43389, '\p{Is_Line_Break= _JL}', "");
Expect(1, 43389, '\p{^Is_Line_Break= _JL}', "");
Expect(1, 43389, '\P{Is_Line_Break= _JL}', "");
Expect(0, 43389, '\P{^Is_Line_Break= _JL}', "");
Error('\p{Is_Lb=:=jl}');
Error('\P{Is_Lb=:=jl}');
Expect(1, 43388, '\p{Is_Lb:	jl}', "");
Expect(0, 43388, '\p{^Is_Lb:	jl}', "");
Expect(0, 43388, '\P{Is_Lb:	jl}', "");
Expect(1, 43388, '\P{^Is_Lb:	jl}', "");
Expect(0, 43389, '\p{Is_Lb:	jl}', "");
Expect(1, 43389, '\p{^Is_Lb:	jl}', "");
Expect(1, 43389, '\P{Is_Lb:	jl}', "");
Expect(0, 43389, '\P{^Is_Lb:	jl}', "");
Expect(1, 43388, '\p{Is_Lb: jl}', "");
Expect(0, 43388, '\p{^Is_Lb: jl}', "");
Expect(0, 43388, '\P{Is_Lb: jl}', "");
Expect(1, 43388, '\P{^Is_Lb: jl}', "");
Expect(0, 43389, '\p{Is_Lb: jl}', "");
Expect(1, 43389, '\p{^Is_Lb: jl}', "");
Expect(1, 43389, '\P{Is_Lb: jl}', "");
Expect(0, 43389, '\P{^Is_Lb: jl}', "");
Error('\p{Line_Break=/a/ 	BREAK_Symbols}');
Error('\P{Line_Break=/a/ 	BREAK_Symbols}');
Expect(1, 47, '\p{Line_Break=breaksymbols}', "");
Expect(0, 47, '\p{^Line_Break=breaksymbols}', "");
Expect(0, 47, '\P{Line_Break=breaksymbols}', "");
Expect(1, 47, '\P{^Line_Break=breaksymbols}', "");
Expect(0, 48, '\p{Line_Break=breaksymbols}', "");
Expect(1, 48, '\p{^Line_Break=breaksymbols}', "");
Expect(1, 48, '\P{Line_Break=breaksymbols}', "");
Expect(0, 48, '\P{^Line_Break=breaksymbols}', "");
Expect(1, 47, '\p{Line_Break= 	Break_Symbols}', "");
Expect(0, 47, '\p{^Line_Break= 	Break_Symbols}', "");
Expect(0, 47, '\P{Line_Break= 	Break_Symbols}', "");
Expect(1, 47, '\P{^Line_Break= 	Break_Symbols}', "");
Expect(0, 48, '\p{Line_Break= 	Break_Symbols}', "");
Expect(1, 48, '\p{^Line_Break= 	Break_Symbols}', "");
Expect(1, 48, '\P{Line_Break= 	Break_Symbols}', "");
Expect(0, 48, '\P{^Line_Break= 	Break_Symbols}', "");
Error('\p{Lb:	/a/	-sy}');
Error('\P{Lb:	/a/	-sy}');
Expect(1, 47, '\p{Lb:   sy}', "");
Expect(0, 47, '\p{^Lb:   sy}', "");
Expect(0, 47, '\P{Lb:   sy}', "");
Expect(1, 47, '\P{^Lb:   sy}', "");
Expect(0, 48, '\p{Lb:   sy}', "");
Expect(1, 48, '\p{^Lb:   sy}', "");
Expect(1, 48, '\P{Lb:   sy}', "");
Expect(0, 48, '\P{^Lb:   sy}', "");
Expect(1, 47, '\p{Lb=SY}', "");
Expect(0, 47, '\p{^Lb=SY}', "");
Expect(0, 47, '\P{Lb=SY}', "");
Expect(1, 47, '\P{^Lb=SY}', "");
Expect(0, 48, '\p{Lb=SY}', "");
Expect(1, 48, '\p{^Lb=SY}', "");
Expect(1, 48, '\P{Lb=SY}', "");
Expect(0, 48, '\P{^Lb=SY}', "");
Error('\p{Is_Line_Break=	/a/break_Symbols}');
Error('\P{Is_Line_Break=	/a/break_Symbols}');
Expect(1, 47, '\p{Is_Line_Break=breaksymbols}', "");
Expect(0, 47, '\p{^Is_Line_Break=breaksymbols}', "");
Expect(0, 47, '\P{Is_Line_Break=breaksymbols}', "");
Expect(1, 47, '\P{^Is_Line_Break=breaksymbols}', "");
Expect(0, 48, '\p{Is_Line_Break=breaksymbols}', "");
Expect(1, 48, '\p{^Is_Line_Break=breaksymbols}', "");
Expect(1, 48, '\P{Is_Line_Break=breaksymbols}', "");
Expect(0, 48, '\P{^Is_Line_Break=breaksymbols}', "");
Expect(1, 47, '\p{Is_Line_Break=_	BREAK_symbols}', "");
Expect(0, 47, '\p{^Is_Line_Break=_	BREAK_symbols}', "");
Expect(0, 47, '\P{Is_Line_Break=_	BREAK_symbols}', "");
Expect(1, 47, '\P{^Is_Line_Break=_	BREAK_symbols}', "");
Expect(0, 48, '\p{Is_Line_Break=_	BREAK_symbols}', "");
Expect(1, 48, '\p{^Is_Line_Break=_	BREAK_symbols}', "");
Expect(1, 48, '\P{Is_Line_Break=_	BREAK_symbols}', "");
Expect(0, 48, '\P{^Is_Line_Break=_	BREAK_symbols}', "");
Error('\p{Is_Lb=	:=SY}');
Error('\P{Is_Lb=	:=SY}');
Expect(1, 47, '\p{Is_Lb=sy}', "");
Expect(0, 47, '\p{^Is_Lb=sy}', "");
Expect(0, 47, '\P{Is_Lb=sy}', "");
Expect(1, 47, '\P{^Is_Lb=sy}', "");
Expect(0, 48, '\p{Is_Lb=sy}', "");
Expect(1, 48, '\p{^Is_Lb=sy}', "");
Expect(1, 48, '\P{Is_Lb=sy}', "");
Expect(0, 48, '\P{^Is_Lb=sy}', "");
Expect(1, 47, '\p{Is_Lb:	_SY}', "");
Expect(0, 47, '\p{^Is_Lb:	_SY}', "");
Expect(0, 47, '\P{Is_Lb:	_SY}', "");
Expect(1, 47, '\P{^Is_Lb:	_SY}', "");
Expect(0, 48, '\p{Is_Lb:	_SY}', "");
Expect(1, 48, '\p{^Is_Lb:	_SY}', "");
Expect(1, 48, '\P{Is_Lb:	_SY}', "");
Expect(0, 48, '\P{^Is_Lb:	_SY}', "");
Error('\p{Line_Break=/a/- Prefix_Numeric}');
Error('\P{Line_Break=/a/- Prefix_Numeric}');
Expect(1, 65510, '\p{Line_Break=prefixnumeric}', "");
Expect(0, 65510, '\p{^Line_Break=prefixnumeric}', "");
Expect(0, 65510, '\P{Line_Break=prefixnumeric}', "");
Expect(1, 65510, '\P{^Line_Break=prefixnumeric}', "");
Expect(0, 65511, '\p{Line_Break=prefixnumeric}', "");
Expect(1, 65511, '\p{^Line_Break=prefixnumeric}', "");
Expect(1, 65511, '\P{Line_Break=prefixnumeric}', "");
Expect(0, 65511, '\P{^Line_Break=prefixnumeric}', "");
Expect(1, 65510, '\p{Line_Break=_-Prefix_Numeric}', "");
Expect(0, 65510, '\p{^Line_Break=_-Prefix_Numeric}', "");
Expect(0, 65510, '\P{Line_Break=_-Prefix_Numeric}', "");
Expect(1, 65510, '\P{^Line_Break=_-Prefix_Numeric}', "");
Expect(0, 65511, '\p{Line_Break=_-Prefix_Numeric}', "");
Expect(1, 65511, '\p{^Line_Break=_-Prefix_Numeric}', "");
Expect(1, 65511, '\P{Line_Break=_-Prefix_Numeric}', "");
Expect(0, 65511, '\P{^Line_Break=_-Prefix_Numeric}', "");
Error('\p{Lb:/a/_	pr}');
Error('\P{Lb:/a/_	pr}');
Expect(1, 65510, '\p{Lb:   pr}', "");
Expect(0, 65510, '\p{^Lb:   pr}', "");
Expect(0, 65510, '\P{Lb:   pr}', "");
Expect(1, 65510, '\P{^Lb:   pr}', "");
Expect(0, 65511, '\p{Lb:   pr}', "");
Expect(1, 65511, '\p{^Lb:   pr}', "");
Expect(1, 65511, '\P{Lb:   pr}', "");
Expect(0, 65511, '\P{^Lb:   pr}', "");
Expect(1, 65510, '\p{Lb= -PR}', "");
Expect(0, 65510, '\p{^Lb= -PR}', "");
Expect(0, 65510, '\P{Lb= -PR}', "");
Expect(1, 65510, '\P{^Lb= -PR}', "");
Expect(0, 65511, '\p{Lb= -PR}', "");
Expect(1, 65511, '\p{^Lb= -PR}', "");
Expect(1, 65511, '\P{Lb= -PR}', "");
Expect(0, 65511, '\P{^Lb= -PR}', "");
Error('\p{Is_Line_Break=_:=Prefix_numeric}');
Error('\P{Is_Line_Break=_:=Prefix_numeric}');
Expect(1, 65510, '\p{Is_Line_Break=prefixnumeric}', "");
Expect(0, 65510, '\p{^Is_Line_Break=prefixnumeric}', "");
Expect(0, 65510, '\P{Is_Line_Break=prefixnumeric}', "");
Expect(1, 65510, '\P{^Is_Line_Break=prefixnumeric}', "");
Expect(0, 65511, '\p{Is_Line_Break=prefixnumeric}', "");
Expect(1, 65511, '\p{^Is_Line_Break=prefixnumeric}', "");
Expect(1, 65511, '\P{Is_Line_Break=prefixnumeric}', "");
Expect(0, 65511, '\P{^Is_Line_Break=prefixnumeric}', "");
Expect(1, 65510, '\p{Is_Line_Break=	prefix_Numeric}', "");
Expect(0, 65510, '\p{^Is_Line_Break=	prefix_Numeric}', "");
Expect(0, 65510, '\P{Is_Line_Break=	prefix_Numeric}', "");
Expect(1, 65510, '\P{^Is_Line_Break=	prefix_Numeric}', "");
Expect(0, 65511, '\p{Is_Line_Break=	prefix_Numeric}', "");
Expect(1, 65511, '\p{^Is_Line_Break=	prefix_Numeric}', "");
Expect(1, 65511, '\P{Is_Line_Break=	prefix_Numeric}', "");
Expect(0, 65511, '\P{^Is_Line_Break=	prefix_Numeric}', "");
Error('\p{Is_Lb=-	PR:=}');
Error('\P{Is_Lb=-	PR:=}');
Expect(1, 65510, '\p{Is_Lb=pr}', "");
Expect(0, 65510, '\p{^Is_Lb=pr}', "");
Expect(0, 65510, '\P{Is_Lb=pr}', "");
Expect(1, 65510, '\P{^Is_Lb=pr}', "");
Expect(0, 65511, '\p{Is_Lb=pr}', "");
Expect(1, 65511, '\p{^Is_Lb=pr}', "");
Expect(1, 65511, '\P{Is_Lb=pr}', "");
Expect(0, 65511, '\P{^Is_Lb=pr}', "");
Expect(1, 65510, '\p{Is_Lb=_ pr}', "");
Expect(0, 65510, '\p{^Is_Lb=_ pr}', "");
Expect(0, 65510, '\P{Is_Lb=_ pr}', "");
Expect(1, 65510, '\P{^Is_Lb=_ pr}', "");
Expect(0, 65511, '\p{Is_Lb=_ pr}', "");
Expect(1, 65511, '\p{^Is_Lb=_ pr}', "");
Expect(1, 65511, '\P{Is_Lb=_ pr}', "");
Expect(0, 65511, '\P{^Is_Lb=_ pr}', "");
Error('\p{Line_Break=	close_PARENTHESIS:=}');
Error('\P{Line_Break=	close_PARENTHESIS:=}');
Expect(1, 93, '\p{Line_Break:   closeparenthesis}', "");
Expect(0, 93, '\p{^Line_Break:   closeparenthesis}', "");
Expect(0, 93, '\P{Line_Break:   closeparenthesis}', "");
Expect(1, 93, '\P{^Line_Break:   closeparenthesis}', "");
Expect(0, 94, '\p{Line_Break:   closeparenthesis}', "");
Expect(1, 94, '\p{^Line_Break:   closeparenthesis}', "");
Expect(1, 94, '\P{Line_Break:   closeparenthesis}', "");
Expect(0, 94, '\P{^Line_Break:   closeparenthesis}', "");
Expect(1, 93, '\p{Line_Break:  -close_PARENTHESIS}', "");
Expect(0, 93, '\p{^Line_Break:  -close_PARENTHESIS}', "");
Expect(0, 93, '\P{Line_Break:  -close_PARENTHESIS}', "");
Expect(1, 93, '\P{^Line_Break:  -close_PARENTHESIS}', "");
Expect(0, 94, '\p{Line_Break:  -close_PARENTHESIS}', "");
Expect(1, 94, '\p{^Line_Break:  -close_PARENTHESIS}', "");
Expect(1, 94, '\P{Line_Break:  -close_PARENTHESIS}', "");
Expect(0, 94, '\P{^Line_Break:  -close_PARENTHESIS}', "");
Error('\p{Lb=:=- CP}');
Error('\P{Lb=:=- CP}');
Expect(1, 93, '\p{Lb:   cp}', "");
Expect(0, 93, '\p{^Lb:   cp}', "");
Expect(0, 93, '\P{Lb:   cp}', "");
Expect(1, 93, '\P{^Lb:   cp}', "");
Expect(0, 94, '\p{Lb:   cp}', "");
Expect(1, 94, '\p{^Lb:   cp}', "");
Expect(1, 94, '\P{Lb:   cp}', "");
Expect(0, 94, '\P{^Lb:   cp}', "");
Expect(1, 93, '\p{Lb=- CP}', "");
Expect(0, 93, '\p{^Lb=- CP}', "");
Expect(0, 93, '\P{Lb=- CP}', "");
Expect(1, 93, '\P{^Lb=- CP}', "");
Expect(0, 94, '\p{Lb=- CP}', "");
Expect(1, 94, '\p{^Lb=- CP}', "");
Expect(1, 94, '\P{Lb=- CP}', "");
Expect(0, 94, '\P{^Lb=- CP}', "");
Error('\p{Is_Line_Break=	-close_Parenthesis:=}');
Error('\P{Is_Line_Break=	-close_Parenthesis:=}');
Expect(1, 93, '\p{Is_Line_Break=closeparenthesis}', "");
Expect(0, 93, '\p{^Is_Line_Break=closeparenthesis}', "");
Expect(0, 93, '\P{Is_Line_Break=closeparenthesis}', "");
Expect(1, 93, '\P{^Is_Line_Break=closeparenthesis}', "");
Expect(0, 94, '\p{Is_Line_Break=closeparenthesis}', "");
Expect(1, 94, '\p{^Is_Line_Break=closeparenthesis}', "");
Expect(1, 94, '\P{Is_Line_Break=closeparenthesis}', "");
Expect(0, 94, '\P{^Is_Line_Break=closeparenthesis}', "");
Expect(1, 93, '\p{Is_Line_Break=  Close_Parenthesis}', "");
Expect(0, 93, '\p{^Is_Line_Break=  Close_Parenthesis}', "");
Expect(0, 93, '\P{Is_Line_Break=  Close_Parenthesis}', "");
Expect(1, 93, '\P{^Is_Line_Break=  Close_Parenthesis}', "");
Expect(0, 94, '\p{Is_Line_Break=  Close_Parenthesis}', "");
Expect(1, 94, '\p{^Is_Line_Break=  Close_Parenthesis}', "");
Expect(1, 94, '\P{Is_Line_Break=  Close_Parenthesis}', "");
Expect(0, 94, '\P{^Is_Line_Break=  Close_Parenthesis}', "");
Error('\p{Is_Lb=/a/ CP}');
Error('\P{Is_Lb=/a/ CP}');
Expect(1, 93, '\p{Is_Lb:   cp}', "");
Expect(0, 93, '\p{^Is_Lb:   cp}', "");
Expect(0, 93, '\P{Is_Lb:   cp}', "");
Expect(1, 93, '\P{^Is_Lb:   cp}', "");
Expect(0, 94, '\p{Is_Lb:   cp}', "");
Expect(1, 94, '\p{^Is_Lb:   cp}', "");
Expect(1, 94, '\P{Is_Lb:   cp}', "");
Expect(0, 94, '\P{^Is_Lb:   cp}', "");
Expect(1, 93, '\p{Is_Lb=- cp}', "");
Expect(0, 93, '\p{^Is_Lb=- cp}', "");
Expect(0, 93, '\P{Is_Lb=- cp}', "");
Expect(1, 93, '\P{^Is_Lb=- cp}', "");
Expect(0, 94, '\p{Is_Lb=- cp}', "");
Expect(1, 94, '\p{^Is_Lb=- cp}', "");
Expect(1, 94, '\P{Is_Lb=- cp}', "");
Expect(0, 94, '\P{^Is_Lb=- cp}', "");
Error('\p{Line_Break=:=-H3}');
Error('\P{Line_Break=:=-H3}');
Expect(1, 55203, '\p{Line_Break=h3}', "");
Expect(0, 55203, '\p{^Line_Break=h3}', "");
Expect(0, 55203, '\P{Line_Break=h3}', "");
Expect(1, 55203, '\P{^Line_Break=h3}', "");
Expect(0, 55204, '\p{Line_Break=h3}', "");
Expect(1, 55204, '\p{^Line_Break=h3}', "");
Expect(1, 55204, '\P{Line_Break=h3}', "");
Expect(0, 55204, '\P{^Line_Break=h3}', "");
Expect(1, 55203, '\p{Line_Break= H3}', "");
Expect(0, 55203, '\p{^Line_Break= H3}', "");
Expect(0, 55203, '\P{Line_Break= H3}', "");
Expect(1, 55203, '\P{^Line_Break= H3}', "");
Expect(0, 55204, '\p{Line_Break= H3}', "");
Expect(1, 55204, '\p{^Line_Break= H3}', "");
Expect(1, 55204, '\P{Line_Break= H3}', "");
Expect(0, 55204, '\P{^Line_Break= H3}', "");
Error('\p{Lb=/a/H3}');
Error('\P{Lb=/a/H3}');
Expect(1, 55203, '\p{Lb:	h3}', "");
Expect(0, 55203, '\p{^Lb:	h3}', "");
Expect(0, 55203, '\P{Lb:	h3}', "");
Expect(1, 55203, '\P{^Lb:	h3}', "");
Expect(0, 55204, '\p{Lb:	h3}', "");
Expect(1, 55204, '\p{^Lb:	h3}', "");
Expect(1, 55204, '\P{Lb:	h3}', "");
Expect(0, 55204, '\P{^Lb:	h3}', "");
Expect(1, 55203, '\p{Lb= -H3}', "");
Expect(0, 55203, '\p{^Lb= -H3}', "");
Expect(0, 55203, '\P{Lb= -H3}', "");
Expect(1, 55203, '\P{^Lb= -H3}', "");
Expect(0, 55204, '\p{Lb= -H3}', "");
Expect(1, 55204, '\p{^Lb= -H3}', "");
Expect(1, 55204, '\P{Lb= -H3}', "");
Expect(0, 55204, '\P{^Lb= -H3}', "");
Error('\p{Is_Line_Break=:=_	h3}');
Error('\P{Is_Line_Break=:=_	h3}');
Expect(1, 55203, '\p{Is_Line_Break=h3}', "");
Expect(0, 55203, '\p{^Is_Line_Break=h3}', "");
Expect(0, 55203, '\P{Is_Line_Break=h3}', "");
Expect(1, 55203, '\P{^Is_Line_Break=h3}', "");
Expect(0, 55204, '\p{Is_Line_Break=h3}', "");
Expect(1, 55204, '\p{^Is_Line_Break=h3}', "");
Expect(1, 55204, '\P{Is_Line_Break=h3}', "");
Expect(0, 55204, '\P{^Is_Line_Break=h3}', "");
Expect(1, 55203, '\p{Is_Line_Break:--H3}', "");
Expect(0, 55203, '\p{^Is_Line_Break:--H3}', "");
Expect(0, 55203, '\P{Is_Line_Break:--H3}', "");
Expect(1, 55203, '\P{^Is_Line_Break:--H3}', "");
Expect(0, 55204, '\p{Is_Line_Break:--H3}', "");
Expect(1, 55204, '\p{^Is_Line_Break:--H3}', "");
Expect(1, 55204, '\P{Is_Line_Break:--H3}', "");
Expect(0, 55204, '\P{^Is_Line_Break:--H3}', "");
Error('\p{Is_Lb=_/a/H3}');
Error('\P{Is_Lb=_/a/H3}');
Expect(1, 55203, '\p{Is_Lb=h3}', "");
Expect(0, 55203, '\p{^Is_Lb=h3}', "");
Expect(0, 55203, '\P{Is_Lb=h3}', "");
Expect(1, 55203, '\P{^Is_Lb=h3}', "");
Expect(0, 55204, '\p{Is_Lb=h3}', "");
Expect(1, 55204, '\p{^Is_Lb=h3}', "");
Expect(1, 55204, '\P{Is_Lb=h3}', "");
Expect(0, 55204, '\P{^Is_Lb=h3}', "");
Expect(1, 55203, '\p{Is_Lb:   H3}', "");
Expect(0, 55203, '\p{^Is_Lb:   H3}', "");
Expect(0, 55203, '\P{Is_Lb:   H3}', "");
Expect(1, 55203, '\P{^Is_Lb:   H3}', "");
Expect(0, 55204, '\p{Is_Lb:   H3}', "");
Expect(1, 55204, '\p{^Is_Lb:   H3}', "");
Expect(1, 55204, '\P{Is_Lb:   H3}', "");
Expect(0, 55204, '\P{^Is_Lb:   H3}', "");
Error('\p{Line_Break=/a/ Hebrew_LETTER}');
Error('\P{Line_Break=/a/ Hebrew_LETTER}');
Expect(1, 64335, '\p{Line_Break: hebrewletter}', "");
Expect(0, 64335, '\p{^Line_Break: hebrewletter}', "");
Expect(0, 64335, '\P{Line_Break: hebrewletter}', "");
Expect(1, 64335, '\P{^Line_Break: hebrewletter}', "");
Expect(0, 64336, '\p{Line_Break: hebrewletter}', "");
Expect(1, 64336, '\p{^Line_Break: hebrewletter}', "");
Expect(1, 64336, '\P{Line_Break: hebrewletter}', "");
Expect(0, 64336, '\P{^Line_Break: hebrewletter}', "");
Expect(1, 64335, '\p{Line_Break= hebrew_letter}', "");
Expect(0, 64335, '\p{^Line_Break= hebrew_letter}', "");
Expect(0, 64335, '\P{Line_Break= hebrew_letter}', "");
Expect(1, 64335, '\P{^Line_Break= hebrew_letter}', "");
Expect(0, 64336, '\p{Line_Break= hebrew_letter}', "");
Expect(1, 64336, '\p{^Line_Break= hebrew_letter}', "");
Expect(1, 64336, '\P{Line_Break= hebrew_letter}', "");
Expect(0, 64336, '\P{^Line_Break= hebrew_letter}', "");
Error('\p{Lb=	_hl/a/}');
Error('\P{Lb=	_hl/a/}');
Expect(1, 64335, '\p{Lb=hl}', "");
Expect(0, 64335, '\p{^Lb=hl}', "");
Expect(0, 64335, '\P{Lb=hl}', "");
Expect(1, 64335, '\P{^Lb=hl}', "");
Expect(0, 64336, '\p{Lb=hl}', "");
Expect(1, 64336, '\p{^Lb=hl}', "");
Expect(1, 64336, '\P{Lb=hl}', "");
Expect(0, 64336, '\P{^Lb=hl}', "");
Expect(1, 64335, '\p{Lb= HL}', "");
Expect(0, 64335, '\p{^Lb= HL}', "");
Expect(0, 64335, '\P{Lb= HL}', "");
Expect(1, 64335, '\P{^Lb= HL}', "");
Expect(0, 64336, '\p{Lb= HL}', "");
Expect(1, 64336, '\p{^Lb= HL}', "");
Expect(1, 64336, '\P{Lb= HL}', "");
Expect(0, 64336, '\P{^Lb= HL}', "");
Error('\p{Is_Line_Break=/a/	hebrew_Letter}');
Error('\P{Is_Line_Break=/a/	hebrew_Letter}');
Expect(1, 64335, '\p{Is_Line_Break=hebrewletter}', "");
Expect(0, 64335, '\p{^Is_Line_Break=hebrewletter}', "");
Expect(0, 64335, '\P{Is_Line_Break=hebrewletter}', "");
Expect(1, 64335, '\P{^Is_Line_Break=hebrewletter}', "");
Expect(0, 64336, '\p{Is_Line_Break=hebrewletter}', "");
Expect(1, 64336, '\p{^Is_Line_Break=hebrewletter}', "");
Expect(1, 64336, '\P{Is_Line_Break=hebrewletter}', "");
Expect(0, 64336, '\P{^Is_Line_Break=hebrewletter}', "");
Expect(1, 64335, '\p{Is_Line_Break:- HEBREW_letter}', "");
Expect(0, 64335, '\p{^Is_Line_Break:- HEBREW_letter}', "");
Expect(0, 64335, '\P{Is_Line_Break:- HEBREW_letter}', "");
Expect(1, 64335, '\P{^Is_Line_Break:- HEBREW_letter}', "");
Expect(0, 64336, '\p{Is_Line_Break:- HEBREW_letter}', "");
Expect(1, 64336, '\p{^Is_Line_Break:- HEBREW_letter}', "");
Expect(1, 64336, '\P{Is_Line_Break:- HEBREW_letter}', "");
Expect(0, 64336, '\P{^Is_Line_Break:- HEBREW_letter}', "");
Error('\p{Is_Lb=/a/_HL}');
Error('\P{Is_Lb=/a/_HL}');
Expect(1, 64335, '\p{Is_Lb=hl}', "");
Expect(0, 64335, '\p{^Is_Lb=hl}', "");
Expect(0, 64335, '\P{Is_Lb=hl}', "");
Expect(1, 64335, '\P{^Is_Lb=hl}', "");
Expect(0, 64336, '\p{Is_Lb=hl}', "");
Expect(1, 64336, '\p{^Is_Lb=hl}', "");
Expect(1, 64336, '\P{Is_Lb=hl}', "");
Expect(0, 64336, '\P{^Is_Lb=hl}', "");
Expect(1, 64335, '\p{Is_Lb= _HL}', "");
Expect(0, 64335, '\p{^Is_Lb= _HL}', "");
Expect(0, 64335, '\P{Is_Lb= _HL}', "");
Expect(1, 64335, '\P{^Is_Lb= _HL}', "");
Expect(0, 64336, '\p{Is_Lb= _HL}', "");
Expect(1, 64336, '\p{^Is_Lb= _HL}', "");
Expect(1, 64336, '\P{Is_Lb= _HL}', "");
Expect(0, 64336, '\P{^Is_Lb= _HL}', "");
Error('\p{Line_Break= :=CARRIAGE_Return}');
Error('\P{Line_Break= :=CARRIAGE_Return}');
Expect(1, 13, '\p{Line_Break=carriagereturn}', "");
Expect(0, 13, '\p{^Line_Break=carriagereturn}', "");
Expect(0, 13, '\P{Line_Break=carriagereturn}', "");
Expect(1, 13, '\P{^Line_Break=carriagereturn}', "");
Expect(0, 14, '\p{Line_Break=carriagereturn}', "");
Expect(1, 14, '\p{^Line_Break=carriagereturn}', "");
Expect(1, 14, '\P{Line_Break=carriagereturn}', "");
Expect(0, 14, '\P{^Line_Break=carriagereturn}', "");
Expect(1, 13, '\p{Line_Break=_CARRIAGE_return}', "");
Expect(0, 13, '\p{^Line_Break=_CARRIAGE_return}', "");
Expect(0, 13, '\P{Line_Break=_CARRIAGE_return}', "");
Expect(1, 13, '\P{^Line_Break=_CARRIAGE_return}', "");
Expect(0, 14, '\p{Line_Break=_CARRIAGE_return}', "");
Expect(1, 14, '\p{^Line_Break=_CARRIAGE_return}', "");
Expect(1, 14, '\P{Line_Break=_CARRIAGE_return}', "");
Expect(0, 14, '\P{^Line_Break=_CARRIAGE_return}', "");
Error('\p{Lb=/a/- CR}');
Error('\P{Lb=/a/- CR}');
Expect(1, 13, '\p{Lb=cr}', "");
Expect(0, 13, '\p{^Lb=cr}', "");
Expect(0, 13, '\P{Lb=cr}', "");
Expect(1, 13, '\P{^Lb=cr}', "");
Expect(0, 14, '\p{Lb=cr}', "");
Expect(1, 14, '\p{^Lb=cr}', "");
Expect(1, 14, '\P{Lb=cr}', "");
Expect(0, 14, '\P{^Lb=cr}', "");
Expect(1, 13, '\p{Lb=	-cr}', "");
Expect(0, 13, '\p{^Lb=	-cr}', "");
Expect(0, 13, '\P{Lb=	-cr}', "");
Expect(1, 13, '\P{^Lb=	-cr}', "");
Expect(0, 14, '\p{Lb=	-cr}', "");
Expect(1, 14, '\p{^Lb=	-cr}', "");
Expect(1, 14, '\P{Lb=	-cr}', "");
Expect(0, 14, '\P{^Lb=	-cr}', "");
Error('\p{Is_Line_Break:-CARRIAGE_return:=}');
Error('\P{Is_Line_Break:-CARRIAGE_return:=}');
Expect(1, 13, '\p{Is_Line_Break: carriagereturn}', "");
Expect(0, 13, '\p{^Is_Line_Break: carriagereturn}', "");
Expect(0, 13, '\P{Is_Line_Break: carriagereturn}', "");
Expect(1, 13, '\P{^Is_Line_Break: carriagereturn}', "");
Expect(0, 14, '\p{Is_Line_Break: carriagereturn}', "");
Expect(1, 14, '\p{^Is_Line_Break: carriagereturn}', "");
Expect(1, 14, '\P{Is_Line_Break: carriagereturn}', "");
Expect(0, 14, '\P{^Is_Line_Break: carriagereturn}', "");
Expect(1, 13, '\p{Is_Line_Break:_ carriage_RETURN}', "");
Expect(0, 13, '\p{^Is_Line_Break:_ carriage_RETURN}', "");
Expect(0, 13, '\P{Is_Line_Break:_ carriage_RETURN}', "");
Expect(1, 13, '\P{^Is_Line_Break:_ carriage_RETURN}', "");
Expect(0, 14, '\p{Is_Line_Break:_ carriage_RETURN}', "");
Expect(1, 14, '\p{^Is_Line_Break:_ carriage_RETURN}', "");
Expect(1, 14, '\P{Is_Line_Break:_ carriage_RETURN}', "");
Expect(0, 14, '\P{^Is_Line_Break:_ carriage_RETURN}', "");
Error('\p{Is_Lb:	 :=cr}');
Error('\P{Is_Lb:	 :=cr}');
Expect(1, 13, '\p{Is_Lb=cr}', "");
Expect(0, 13, '\p{^Is_Lb=cr}', "");
Expect(0, 13, '\P{Is_Lb=cr}', "");
Expect(1, 13, '\P{^Is_Lb=cr}', "");
Expect(0, 14, '\p{Is_Lb=cr}', "");
Expect(1, 14, '\p{^Is_Lb=cr}', "");
Expect(1, 14, '\P{Is_Lb=cr}', "");
Expect(0, 14, '\P{^Is_Lb=cr}', "");
Expect(1, 13, '\p{Is_Lb: - cr}', "");
Expect(0, 13, '\p{^Is_Lb: - cr}', "");
Expect(0, 13, '\P{Is_Lb: - cr}', "");
Expect(1, 13, '\P{^Is_Lb: - cr}', "");
Expect(0, 14, '\p{Is_Lb: - cr}', "");
Expect(1, 14, '\p{^Is_Lb: - cr}', "");
Expect(1, 14, '\P{Is_Lb: - cr}', "");
Expect(0, 14, '\P{^Is_Lb: - cr}', "");
Error('\p{Line_Break= /a/CLOSE_punctuation}');
Error('\P{Line_Break= /a/CLOSE_punctuation}');
Expect(1, 78715, '\p{Line_Break=closepunctuation}', "");
Expect(0, 78715, '\p{^Line_Break=closepunctuation}', "");
Expect(0, 78715, '\P{Line_Break=closepunctuation}', "");
Expect(1, 78715, '\P{^Line_Break=closepunctuation}', "");
Expect(0, 78716, '\p{Line_Break=closepunctuation}', "");
Expect(1, 78716, '\p{^Line_Break=closepunctuation}', "");
Expect(1, 78716, '\P{Line_Break=closepunctuation}', "");
Expect(0, 78716, '\P{^Line_Break=closepunctuation}', "");
Expect(1, 78715, '\p{Line_Break=- Close_punctuation}', "");
Expect(0, 78715, '\p{^Line_Break=- Close_punctuation}', "");
Expect(0, 78715, '\P{Line_Break=- Close_punctuation}', "");
Expect(1, 78715, '\P{^Line_Break=- Close_punctuation}', "");
Expect(0, 78716, '\p{Line_Break=- Close_punctuation}', "");
Expect(1, 78716, '\p{^Line_Break=- Close_punctuation}', "");
Expect(1, 78716, '\P{Line_Break=- Close_punctuation}', "");
Expect(0, 78716, '\P{^Line_Break=- Close_punctuation}', "");
Error('\p{Lb:	/a/-	CL}');
Error('\P{Lb:	/a/-	CL}');
Expect(1, 78715, '\p{Lb=cl}', "");
Expect(0, 78715, '\p{^Lb=cl}', "");
Expect(0, 78715, '\P{Lb=cl}', "");
Expect(1, 78715, '\P{^Lb=cl}', "");
Expect(0, 78716, '\p{Lb=cl}', "");
Expect(1, 78716, '\p{^Lb=cl}', "");
Expect(1, 78716, '\P{Lb=cl}', "");
Expect(0, 78716, '\P{^Lb=cl}', "");
Expect(1, 78715, '\p{Lb:	 -CL}', "");
Expect(0, 78715, '\p{^Lb:	 -CL}', "");
Expect(0, 78715, '\P{Lb:	 -CL}', "");
Expect(1, 78715, '\P{^Lb:	 -CL}', "");
Expect(0, 78716, '\p{Lb:	 -CL}', "");
Expect(1, 78716, '\p{^Lb:	 -CL}', "");
Expect(1, 78716, '\P{Lb:	 -CL}', "");
Expect(0, 78716, '\P{^Lb:	 -CL}', "");
Error('\p{Is_Line_Break: 	Close_punctuation:=}');
Error('\P{Is_Line_Break: 	Close_punctuation:=}');
Expect(1, 78715, '\p{Is_Line_Break:	closepunctuation}', "");
Expect(0, 78715, '\p{^Is_Line_Break:	closepunctuation}', "");
Expect(0, 78715, '\P{Is_Line_Break:	closepunctuation}', "");
Expect(1, 78715, '\P{^Is_Line_Break:	closepunctuation}', "");
Expect(0, 78716, '\p{Is_Line_Break:	closepunctuation}', "");
Expect(1, 78716, '\p{^Is_Line_Break:	closepunctuation}', "");
Expect(1, 78716, '\P{Is_Line_Break:	closepunctuation}', "");
Expect(0, 78716, '\P{^Is_Line_Break:	closepunctuation}', "");
Expect(1, 78715, '\p{Is_Line_Break=-_CLOSE_PUNCTUATION}', "");
Expect(0, 78715, '\p{^Is_Line_Break=-_CLOSE_PUNCTUATION}', "");
Expect(0, 78715, '\P{Is_Line_Break=-_CLOSE_PUNCTUATION}', "");
Expect(1, 78715, '\P{^Is_Line_Break=-_CLOSE_PUNCTUATION}', "");
Expect(0, 78716, '\p{Is_Line_Break=-_CLOSE_PUNCTUATION}', "");
Expect(1, 78716, '\p{^Is_Line_Break=-_CLOSE_PUNCTUATION}', "");
Expect(1, 78716, '\P{Is_Line_Break=-_CLOSE_PUNCTUATION}', "");
Expect(0, 78716, '\P{^Is_Line_Break=-_CLOSE_PUNCTUATION}', "");
Error('\p{Is_Lb=_	CL:=}');
Error('\P{Is_Lb=_	CL:=}');
Expect(1, 78715, '\p{Is_Lb=cl}', "");
Expect(0, 78715, '\p{^Is_Lb=cl}', "");
Expect(0, 78715, '\P{Is_Lb=cl}', "");
Expect(1, 78715, '\P{^Is_Lb=cl}', "");
Expect(0, 78716, '\p{Is_Lb=cl}', "");
Expect(1, 78716, '\p{^Is_Lb=cl}', "");
Expect(1, 78716, '\P{Is_Lb=cl}', "");
Expect(0, 78716, '\P{^Is_Lb=cl}', "");
Expect(1, 78715, '\p{Is_Lb=-CL}', "");
Expect(0, 78715, '\p{^Is_Lb=-CL}', "");
Expect(0, 78715, '\P{Is_Lb=-CL}', "");
Expect(1, 78715, '\P{^Is_Lb=-CL}', "");
Expect(0, 78716, '\p{Is_Lb=-CL}', "");
Expect(1, 78716, '\p{^Is_Lb=-CL}', "");
Expect(1, 78716, '\P{Is_Lb=-CL}', "");
Expect(0, 78716, '\P{^Is_Lb=-CL}', "");
Error('\p{Line_Break: -_Combining_MARK/a/}');
Error('\P{Line_Break: -_Combining_MARK/a/}');
Expect(1, 917999, '\p{Line_Break:   combiningmark}', "");
Expect(0, 917999, '\p{^Line_Break:   combiningmark}', "");
Expect(0, 917999, '\P{Line_Break:   combiningmark}', "");
Expect(1, 917999, '\P{^Line_Break:   combiningmark}', "");
Expect(0, 918000, '\p{Line_Break:   combiningmark}', "");
Expect(1, 918000, '\p{^Line_Break:   combiningmark}', "");
Expect(1, 918000, '\P{Line_Break:   combiningmark}', "");
Expect(0, 918000, '\P{^Line_Break:   combiningmark}', "");
Expect(1, 917999, '\p{Line_Break=-combining_mark}', "");
Expect(0, 917999, '\p{^Line_Break=-combining_mark}', "");
Expect(0, 917999, '\P{Line_Break=-combining_mark}', "");
Expect(1, 917999, '\P{^Line_Break=-combining_mark}', "");
Expect(0, 918000, '\p{Line_Break=-combining_mark}', "");
Expect(1, 918000, '\p{^Line_Break=-combining_mark}', "");
Expect(1, 918000, '\P{Line_Break=-combining_mark}', "");
Expect(0, 918000, '\P{^Line_Break=-combining_mark}', "");
Error('\p{Lb=/a/CM}');
Error('\P{Lb=/a/CM}');
Expect(1, 917999, '\p{Lb=cm}', "");
Expect(0, 917999, '\p{^Lb=cm}', "");
Expect(0, 917999, '\P{Lb=cm}', "");
Expect(1, 917999, '\P{^Lb=cm}', "");
Expect(0, 918000, '\p{Lb=cm}', "");
Expect(1, 918000, '\p{^Lb=cm}', "");
Expect(1, 918000, '\P{Lb=cm}', "");
Expect(0, 918000, '\P{^Lb=cm}', "");
Expect(1, 917999, '\p{Lb:   -CM}', "");
Expect(0, 917999, '\p{^Lb:   -CM}', "");
Expect(0, 917999, '\P{Lb:   -CM}', "");
Expect(1, 917999, '\P{^Lb:   -CM}', "");
Expect(0, 918000, '\p{Lb:   -CM}', "");
Expect(1, 918000, '\p{^Lb:   -CM}', "");
Expect(1, 918000, '\P{Lb:   -CM}', "");
Expect(0, 918000, '\P{^Lb:   -CM}', "");
Error('\p{Is_Line_Break= /a/Combining_mark}');
Error('\P{Is_Line_Break= /a/Combining_mark}');
Expect(1, 917999, '\p{Is_Line_Break=combiningmark}', "");
Expect(0, 917999, '\p{^Is_Line_Break=combiningmark}', "");
Expect(0, 917999, '\P{Is_Line_Break=combiningmark}', "");
Expect(1, 917999, '\P{^Is_Line_Break=combiningmark}', "");
Expect(0, 918000, '\p{Is_Line_Break=combiningmark}', "");
Expect(1, 918000, '\p{^Is_Line_Break=combiningmark}', "");
Expect(1, 918000, '\P{Is_Line_Break=combiningmark}', "");
Expect(0, 918000, '\P{^Is_Line_Break=combiningmark}', "");
Expect(1, 917999, '\p{Is_Line_Break= combining_Mark}', "");
Expect(0, 917999, '\p{^Is_Line_Break= combining_Mark}', "");
Expect(0, 917999, '\P{Is_Line_Break= combining_Mark}', "");
Expect(1, 917999, '\P{^Is_Line_Break= combining_Mark}', "");
Expect(0, 918000, '\p{Is_Line_Break= combining_Mark}', "");
Expect(1, 918000, '\p{^Is_Line_Break= combining_Mark}', "");
Expect(1, 918000, '\P{Is_Line_Break= combining_Mark}', "");
Expect(0, 918000, '\P{^Is_Line_Break= combining_Mark}', "");
Error('\p{Is_Lb=/a/ CM}');
Error('\P{Is_Lb=/a/ CM}');
Expect(1, 917999, '\p{Is_Lb=cm}', "");
Expect(0, 917999, '\p{^Is_Lb=cm}', "");
Expect(0, 917999, '\P{Is_Lb=cm}', "");
Expect(1, 917999, '\P{^Is_Lb=cm}', "");
Expect(0, 918000, '\p{Is_Lb=cm}', "");
Expect(1, 918000, '\p{^Is_Lb=cm}', "");
Expect(1, 918000, '\P{Is_Lb=cm}', "");
Expect(0, 918000, '\P{^Is_Lb=cm}', "");
Error('\p{Line_Break=/a/	-Mandatory_Break}');
Error('\P{Line_Break=/a/	-Mandatory_Break}');
Expect(1, 8233, '\p{Line_Break=mandatorybreak}', "");
Expect(0, 8233, '\p{^Line_Break=mandatorybreak}', "");
Expect(0, 8233, '\P{Line_Break=mandatorybreak}', "");
Expect(1, 8233, '\P{^Line_Break=mandatorybreak}', "");
Expect(0, 8234, '\p{Line_Break=mandatorybreak}', "");
Expect(1, 8234, '\p{^Line_Break=mandatorybreak}', "");
Expect(1, 8234, '\P{Line_Break=mandatorybreak}', "");
Expect(0, 8234, '\P{^Line_Break=mandatorybreak}', "");
Expect(1, 8233, '\p{Line_Break=		Mandatory_break}', "");
Expect(0, 8233, '\p{^Line_Break=		Mandatory_break}', "");
Expect(0, 8233, '\P{Line_Break=		Mandatory_break}', "");
Expect(1, 8233, '\P{^Line_Break=		Mandatory_break}', "");
Expect(0, 8234, '\p{Line_Break=		Mandatory_break}', "");
Expect(1, 8234, '\p{^Line_Break=		Mandatory_break}', "");
Expect(1, 8234, '\P{Line_Break=		Mandatory_break}', "");
Expect(0, 8234, '\P{^Line_Break=		Mandatory_break}', "");
Error('\p{Lb=:=_BK}');
Error('\P{Lb=:=_BK}');
Expect(1, 8233, '\p{Lb=bk}', "");
Expect(0, 8233, '\p{^Lb=bk}', "");
Expect(0, 8233, '\P{Lb=bk}', "");
Expect(1, 8233, '\P{^Lb=bk}', "");
Expect(0, 8234, '\p{Lb=bk}', "");
Expect(1, 8234, '\p{^Lb=bk}', "");
Expect(1, 8234, '\P{Lb=bk}', "");
Expect(0, 8234, '\P{^Lb=bk}', "");
Expect(1, 8233, '\p{Lb=- BK}', "");
Expect(0, 8233, '\p{^Lb=- BK}', "");
Expect(0, 8233, '\P{Lb=- BK}', "");
Expect(1, 8233, '\P{^Lb=- BK}', "");
Expect(0, 8234, '\p{Lb=- BK}', "");
Expect(1, 8234, '\p{^Lb=- BK}', "");
Expect(1, 8234, '\P{Lb=- BK}', "");
Expect(0, 8234, '\P{^Lb=- BK}', "");
Error('\p{Is_Line_Break=-:=Mandatory_Break}');
Error('\P{Is_Line_Break=-:=Mandatory_Break}');
Expect(1, 8233, '\p{Is_Line_Break=mandatorybreak}', "");
Expect(0, 8233, '\p{^Is_Line_Break=mandatorybreak}', "");
Expect(0, 8233, '\P{Is_Line_Break=mandatorybreak}', "");
Expect(1, 8233, '\P{^Is_Line_Break=mandatorybreak}', "");
Expect(0, 8234, '\p{Is_Line_Break=mandatorybreak}', "");
Expect(1, 8234, '\p{^Is_Line_Break=mandatorybreak}', "");
Expect(1, 8234, '\P{Is_Line_Break=mandatorybreak}', "");
Expect(0, 8234, '\P{^Is_Line_Break=mandatorybreak}', "");
Expect(1, 8233, '\p{Is_Line_Break=_MANDATORY_BREAK}', "");
Expect(0, 8233, '\p{^Is_Line_Break=_MANDATORY_BREAK}', "");
Expect(0, 8233, '\P{Is_Line_Break=_MANDATORY_BREAK}', "");
Expect(1, 8233, '\P{^Is_Line_Break=_MANDATORY_BREAK}', "");
Expect(0, 8234, '\p{Is_Line_Break=_MANDATORY_BREAK}', "");
Expect(1, 8234, '\p{^Is_Line_Break=_MANDATORY_BREAK}', "");
Expect(1, 8234, '\P{Is_Line_Break=_MANDATORY_BREAK}', "");
Expect(0, 8234, '\P{^Is_Line_Break=_MANDATORY_BREAK}', "");
Error('\p{Is_Lb=	 BK/a/}');
Error('\P{Is_Lb=	 BK/a/}');
Expect(1, 8233, '\p{Is_Lb: bk}', "");
Expect(0, 8233, '\p{^Is_Lb: bk}', "");
Expect(0, 8233, '\P{Is_Lb: bk}', "");
Expect(1, 8233, '\P{^Is_Lb: bk}', "");
Expect(0, 8234, '\p{Is_Lb: bk}', "");
Expect(1, 8234, '\p{^Is_Lb: bk}', "");
Expect(1, 8234, '\P{Is_Lb: bk}', "");
Expect(0, 8234, '\P{^Is_Lb: bk}', "");
Expect(1, 8233, '\p{Is_Lb=		BK}', "");
Expect(0, 8233, '\p{^Is_Lb=		BK}', "");
Expect(0, 8233, '\P{Is_Lb=		BK}', "");
Expect(1, 8233, '\P{^Is_Lb=		BK}', "");
Expect(0, 8234, '\p{Is_Lb=		BK}', "");
Expect(1, 8234, '\p{^Is_Lb=		BK}', "");
Expect(1, 8234, '\P{Is_Lb=		BK}', "");
Expect(0, 8234, '\P{^Is_Lb=		BK}', "");
Error('\p{Line_Break=-nonstarter/a/}');
Error('\P{Line_Break=-nonstarter/a/}');
Expect(1, 65439, '\p{Line_Break=nonstarter}', "");
Expect(0, 65439, '\p{^Line_Break=nonstarter}', "");
Expect(0, 65439, '\P{Line_Break=nonstarter}', "");
Expect(1, 65439, '\P{^Line_Break=nonstarter}', "");
Expect(0, 65440, '\p{Line_Break=nonstarter}', "");
Expect(1, 65440, '\p{^Line_Break=nonstarter}', "");
Expect(1, 65440, '\P{Line_Break=nonstarter}', "");
Expect(0, 65440, '\P{^Line_Break=nonstarter}', "");
Expect(1, 65439, '\p{Line_Break=		nonstarter}', "");
Expect(0, 65439, '\p{^Line_Break=		nonstarter}', "");
Expect(0, 65439, '\P{Line_Break=		nonstarter}', "");
Expect(1, 65439, '\P{^Line_Break=		nonstarter}', "");
Expect(0, 65440, '\p{Line_Break=		nonstarter}', "");
Expect(1, 65440, '\p{^Line_Break=		nonstarter}', "");
Expect(1, 65440, '\P{Line_Break=		nonstarter}', "");
Expect(0, 65440, '\P{^Line_Break=		nonstarter}', "");
Error('\p{Lb=	NS/a/}');
Error('\P{Lb=	NS/a/}');
Expect(1, 65439, '\p{Lb=ns}', "");
Expect(0, 65439, '\p{^Lb=ns}', "");
Expect(0, 65439, '\P{Lb=ns}', "");
Expect(1, 65439, '\P{^Lb=ns}', "");
Expect(0, 65440, '\p{Lb=ns}', "");
Expect(1, 65440, '\p{^Lb=ns}', "");
Expect(1, 65440, '\P{Lb=ns}', "");
Expect(0, 65440, '\P{^Lb=ns}', "");
Expect(1, 65439, '\p{Lb=--NS}', "");
Expect(0, 65439, '\p{^Lb=--NS}', "");
Expect(0, 65439, '\P{Lb=--NS}', "");
Expect(1, 65439, '\P{^Lb=--NS}', "");
Expect(0, 65440, '\p{Lb=--NS}', "");
Expect(1, 65440, '\p{^Lb=--NS}', "");
Expect(1, 65440, '\P{Lb=--NS}', "");
Expect(0, 65440, '\P{^Lb=--NS}', "");
Error('\p{Is_Line_Break:	_:=NONSTARTER}');
Error('\P{Is_Line_Break:	_:=NONSTARTER}');
Expect(1, 65439, '\p{Is_Line_Break=nonstarter}', "");
Expect(0, 65439, '\p{^Is_Line_Break=nonstarter}', "");
Expect(0, 65439, '\P{Is_Line_Break=nonstarter}', "");
Expect(1, 65439, '\P{^Is_Line_Break=nonstarter}', "");
Expect(0, 65440, '\p{Is_Line_Break=nonstarter}', "");
Expect(1, 65440, '\p{^Is_Line_Break=nonstarter}', "");
Expect(1, 65440, '\P{Is_Line_Break=nonstarter}', "");
Expect(0, 65440, '\P{^Is_Line_Break=nonstarter}', "");
Expect(1, 65439, '\p{Is_Line_Break=Nonstarter}', "");
Expect(0, 65439, '\p{^Is_Line_Break=Nonstarter}', "");
Expect(0, 65439, '\P{Is_Line_Break=Nonstarter}', "");
Expect(1, 65439, '\P{^Is_Line_Break=Nonstarter}', "");
Expect(0, 65440, '\p{Is_Line_Break=Nonstarter}', "");
Expect(1, 65440, '\p{^Is_Line_Break=Nonstarter}', "");
Expect(1, 65440, '\P{Is_Line_Break=Nonstarter}', "");
Expect(0, 65440, '\P{^Is_Line_Break=Nonstarter}', "");
Error('\p{Is_Lb=:= ns}');
Error('\P{Is_Lb=:= ns}');
Expect(1, 65439, '\p{Is_Lb:ns}', "");
Expect(0, 65439, '\p{^Is_Lb:ns}', "");
Expect(0, 65439, '\P{Is_Lb:ns}', "");
Expect(1, 65439, '\P{^Is_Lb:ns}', "");
Expect(0, 65440, '\p{Is_Lb:ns}', "");
Expect(1, 65440, '\p{^Is_Lb:ns}', "");
Expect(1, 65440, '\P{Is_Lb:ns}', "");
Expect(0, 65440, '\P{^Is_Lb:ns}', "");
Expect(1, 65439, '\p{Is_Lb=__NS}', "");
Expect(0, 65439, '\p{^Is_Lb=__NS}', "");
Expect(0, 65439, '\P{Is_Lb=__NS}', "");
Expect(1, 65439, '\P{^Is_Lb=__NS}', "");
Expect(0, 65440, '\p{Is_Lb=__NS}', "");
Expect(1, 65440, '\p{^Is_Lb=__NS}', "");
Expect(1, 65440, '\P{Is_Lb=__NS}', "");
Expect(0, 65440, '\P{^Is_Lb=__NS}', "");
Error('\p{Line_Break=:=JV}');
Error('\P{Line_Break=:=JV}');
Expect(1, 55238, '\p{Line_Break=jv}', "");
Expect(0, 55238, '\p{^Line_Break=jv}', "");
Expect(0, 55238, '\P{Line_Break=jv}', "");
Expect(1, 55238, '\P{^Line_Break=jv}', "");
Expect(0, 55239, '\p{Line_Break=jv}', "");
Expect(1, 55239, '\p{^Line_Break=jv}', "");
Expect(1, 55239, '\P{Line_Break=jv}', "");
Expect(0, 55239, '\P{^Line_Break=jv}', "");
Expect(1, 55238, '\p{Line_Break= JV}', "");
Expect(0, 55238, '\p{^Line_Break= JV}', "");
Expect(0, 55238, '\P{Line_Break= JV}', "");
Expect(1, 55238, '\P{^Line_Break= JV}', "");
Expect(0, 55239, '\p{Line_Break= JV}', "");
Expect(1, 55239, '\p{^Line_Break= JV}', "");
Expect(1, 55239, '\P{Line_Break= JV}', "");
Expect(0, 55239, '\P{^Line_Break= JV}', "");
Error('\p{Lb=:=_jv}');
Error('\P{Lb=:=_jv}');
Expect(1, 55238, '\p{Lb=jv}', "");
Expect(0, 55238, '\p{^Lb=jv}', "");
Expect(0, 55238, '\P{Lb=jv}', "");
Expect(1, 55238, '\P{^Lb=jv}', "");
Expect(0, 55239, '\p{Lb=jv}', "");
Expect(1, 55239, '\p{^Lb=jv}', "");
Expect(1, 55239, '\P{Lb=jv}', "");
Expect(0, 55239, '\P{^Lb=jv}', "");
Expect(1, 55238, '\p{Lb=		JV}', "");
Expect(0, 55238, '\p{^Lb=		JV}', "");
Expect(0, 55238, '\P{Lb=		JV}', "");
Expect(1, 55238, '\P{^Lb=		JV}', "");
Expect(0, 55239, '\p{Lb=		JV}', "");
Expect(1, 55239, '\p{^Lb=		JV}', "");
Expect(1, 55239, '\P{Lb=		JV}', "");
Expect(0, 55239, '\P{^Lb=		JV}', "");
Error('\p{Is_Line_Break=-:=JV}');
Error('\P{Is_Line_Break=-:=JV}');
Expect(1, 55238, '\p{Is_Line_Break=jv}', "");
Expect(0, 55238, '\p{^Is_Line_Break=jv}', "");
Expect(0, 55238, '\P{Is_Line_Break=jv}', "");
Expect(1, 55238, '\P{^Is_Line_Break=jv}', "");
Expect(0, 55239, '\p{Is_Line_Break=jv}', "");
Expect(1, 55239, '\p{^Is_Line_Break=jv}', "");
Expect(1, 55239, '\P{Is_Line_Break=jv}', "");
Expect(0, 55239, '\P{^Is_Line_Break=jv}', "");
Expect(1, 55238, '\p{Is_Line_Break=_JV}', "");
Expect(0, 55238, '\p{^Is_Line_Break=_JV}', "");
Expect(0, 55238, '\P{Is_Line_Break=_JV}', "");
Expect(1, 55238, '\P{^Is_Line_Break=_JV}', "");
Expect(0, 55239, '\p{Is_Line_Break=_JV}', "");
Expect(1, 55239, '\p{^Is_Line_Break=_JV}', "");
Expect(1, 55239, '\P{Is_Line_Break=_JV}', "");
Expect(0, 55239, '\P{^Is_Line_Break=_JV}', "");
Error('\p{Is_Lb=_-JV/a/}');
Error('\P{Is_Lb=_-JV/a/}');
Expect(1, 55238, '\p{Is_Lb=jv}', "");
Expect(0, 55238, '\p{^Is_Lb=jv}', "");
Expect(0, 55238, '\P{Is_Lb=jv}', "");
Expect(1, 55238, '\P{^Is_Lb=jv}', "");
Expect(0, 55239, '\p{Is_Lb=jv}', "");
Expect(1, 55239, '\p{^Is_Lb=jv}', "");
Expect(1, 55239, '\P{Is_Lb=jv}', "");
Expect(0, 55239, '\P{^Is_Lb=jv}', "");
Error('\p{Line_Break=:=_quotation}');
Error('\P{Line_Break=:=_quotation}');
Expect(1, 11809, '\p{Line_Break=quotation}', "");
Expect(0, 11809, '\p{^Line_Break=quotation}', "");
Expect(0, 11809, '\P{Line_Break=quotation}', "");
Expect(1, 11809, '\P{^Line_Break=quotation}', "");
Expect(0, 11810, '\p{Line_Break=quotation}', "");
Expect(1, 11810, '\p{^Line_Break=quotation}', "");
Expect(1, 11810, '\P{Line_Break=quotation}', "");
Expect(0, 11810, '\P{^Line_Break=quotation}', "");
Expect(1, 11809, '\p{Line_Break: --Quotation}', "");
Expect(0, 11809, '\p{^Line_Break: --Quotation}', "");
Expect(0, 11809, '\P{Line_Break: --Quotation}', "");
Expect(1, 11809, '\P{^Line_Break: --Quotation}', "");
Expect(0, 11810, '\p{Line_Break: --Quotation}', "");
Expect(1, 11810, '\p{^Line_Break: --Quotation}', "");
Expect(1, 11810, '\P{Line_Break: --Quotation}', "");
Expect(0, 11810, '\P{^Line_Break: --Quotation}', "");
Error('\p{Lb=--QU:=}');
Error('\P{Lb=--QU:=}');
Expect(1, 11809, '\p{Lb:qu}', "");
Expect(0, 11809, '\p{^Lb:qu}', "");
Expect(0, 11809, '\P{Lb:qu}', "");
Expect(1, 11809, '\P{^Lb:qu}', "");
Expect(0, 11810, '\p{Lb:qu}', "");
Expect(1, 11810, '\p{^Lb:qu}', "");
Expect(1, 11810, '\P{Lb:qu}', "");
Expect(0, 11810, '\P{^Lb:qu}', "");
Expect(1, 11809, '\p{Lb:	_QU}', "");
Expect(0, 11809, '\p{^Lb:	_QU}', "");
Expect(0, 11809, '\P{Lb:	_QU}', "");
Expect(1, 11809, '\P{^Lb:	_QU}', "");
Expect(0, 11810, '\p{Lb:	_QU}', "");
Expect(1, 11810, '\p{^Lb:	_QU}', "");
Expect(1, 11810, '\P{Lb:	_QU}', "");
Expect(0, 11810, '\P{^Lb:	_QU}', "");
Error('\p{Is_Line_Break=-Quotation:=}');
Error('\P{Is_Line_Break=-Quotation:=}');
Expect(1, 11809, '\p{Is_Line_Break=quotation}', "");
Expect(0, 11809, '\p{^Is_Line_Break=quotation}', "");
Expect(0, 11809, '\P{Is_Line_Break=quotation}', "");
Expect(1, 11809, '\P{^Is_Line_Break=quotation}', "");
Expect(0, 11810, '\p{Is_Line_Break=quotation}', "");
Expect(1, 11810, '\p{^Is_Line_Break=quotation}', "");
Expect(1, 11810, '\P{Is_Line_Break=quotation}', "");
Expect(0, 11810, '\P{^Is_Line_Break=quotation}', "");
Expect(1, 11809, '\p{Is_Line_Break=_ QUOTATION}', "");
Expect(0, 11809, '\p{^Is_Line_Break=_ QUOTATION}', "");
Expect(0, 11809, '\P{Is_Line_Break=_ QUOTATION}', "");
Expect(1, 11809, '\P{^Is_Line_Break=_ QUOTATION}', "");
Expect(0, 11810, '\p{Is_Line_Break=_ QUOTATION}', "");
Expect(1, 11810, '\p{^Is_Line_Break=_ QUOTATION}', "");
Expect(1, 11810, '\P{Is_Line_Break=_ QUOTATION}', "");
Expect(0, 11810, '\P{^Is_Line_Break=_ QUOTATION}', "");
Error('\p{Is_Lb=_/a/qu}');
Error('\P{Is_Lb=_/a/qu}');
Expect(1, 11809, '\p{Is_Lb=qu}', "");
Expect(0, 11809, '\p{^Is_Lb=qu}', "");
Expect(0, 11809, '\P{Is_Lb=qu}', "");
Expect(1, 11809, '\P{^Is_Lb=qu}', "");
Expect(0, 11810, '\p{Is_Lb=qu}', "");
Expect(1, 11810, '\p{^Is_Lb=qu}', "");
Expect(1, 11810, '\P{Is_Lb=qu}', "");
Expect(0, 11810, '\P{^Is_Lb=qu}', "");
Expect(1, 11809, '\p{Is_Lb=_-QU}', "");
Expect(0, 11809, '\p{^Is_Lb=_-QU}', "");
Expect(0, 11809, '\P{Is_Lb=_-QU}', "");
Expect(1, 11809, '\P{^Is_Lb=_-QU}', "");
Expect(0, 11810, '\p{Is_Lb=_-QU}', "");
Expect(1, 11810, '\p{^Is_Lb=_-QU}', "");
Expect(1, 11810, '\P{Is_Lb=_-QU}', "");
Expect(0, 11810, '\P{^Is_Lb=_-QU}', "");
Error('\p{Line_Break=:=-infix_Numeric}');
Error('\P{Line_Break=:=-infix_Numeric}');
Expect(1, 65044, '\p{Line_Break=infixnumeric}', "");
Expect(0, 65044, '\p{^Line_Break=infixnumeric}', "");
Expect(0, 65044, '\P{Line_Break=infixnumeric}', "");
Expect(1, 65044, '\P{^Line_Break=infixnumeric}', "");
Expect(0, 65045, '\p{Line_Break=infixnumeric}', "");
Expect(1, 65045, '\p{^Line_Break=infixnumeric}', "");
Expect(1, 65045, '\P{Line_Break=infixnumeric}', "");
Expect(0, 65045, '\P{^Line_Break=infixnumeric}', "");
Expect(1, 65044, '\p{Line_Break=	Infix_Numeric}', "");
Expect(0, 65044, '\p{^Line_Break=	Infix_Numeric}', "");
Expect(0, 65044, '\P{Line_Break=	Infix_Numeric}', "");
Expect(1, 65044, '\P{^Line_Break=	Infix_Numeric}', "");
Expect(0, 65045, '\p{Line_Break=	Infix_Numeric}', "");
Expect(1, 65045, '\p{^Line_Break=	Infix_Numeric}', "");
Expect(1, 65045, '\P{Line_Break=	Infix_Numeric}', "");
Expect(0, 65045, '\P{^Line_Break=	Infix_Numeric}', "");
Error('\p{Lb: := 	IS}');
Error('\P{Lb: := 	IS}');
Expect(1, 65044, '\p{Lb=is}', "");
Expect(0, 65044, '\p{^Lb=is}', "");
Expect(0, 65044, '\P{Lb=is}', "");
Expect(1, 65044, '\P{^Lb=is}', "");
Expect(0, 65045, '\p{Lb=is}', "");
Expect(1, 65045, '\p{^Lb=is}', "");
Expect(1, 65045, '\P{Lb=is}', "");
Expect(0, 65045, '\P{^Lb=is}', "");
Expect(1, 65044, '\p{Lb=_IS}', "");
Expect(0, 65044, '\p{^Lb=_IS}', "");
Expect(0, 65044, '\P{Lb=_IS}', "");
Expect(1, 65044, '\P{^Lb=_IS}', "");
Expect(0, 65045, '\p{Lb=_IS}', "");
Expect(1, 65045, '\p{^Lb=_IS}', "");
Expect(1, 65045, '\P{Lb=_IS}', "");
Expect(0, 65045, '\P{^Lb=_IS}', "");
Error('\p{Is_Line_Break=- infix_Numeric:=}');
Error('\P{Is_Line_Break=- infix_Numeric:=}');
Expect(1, 65044, '\p{Is_Line_Break=infixnumeric}', "");
Expect(0, 65044, '\p{^Is_Line_Break=infixnumeric}', "");
Expect(0, 65044, '\P{Is_Line_Break=infixnumeric}', "");
Expect(1, 65044, '\P{^Is_Line_Break=infixnumeric}', "");
Expect(0, 65045, '\p{Is_Line_Break=infixnumeric}', "");
Expect(1, 65045, '\p{^Is_Line_Break=infixnumeric}', "");
Expect(1, 65045, '\P{Is_Line_Break=infixnumeric}', "");
Expect(0, 65045, '\P{^Is_Line_Break=infixnumeric}', "");
Expect(1, 65044, '\p{Is_Line_Break=-_infix_NUMERIC}', "");
Expect(0, 65044, '\p{^Is_Line_Break=-_infix_NUMERIC}', "");
Expect(0, 65044, '\P{Is_Line_Break=-_infix_NUMERIC}', "");
Expect(1, 65044, '\P{^Is_Line_Break=-_infix_NUMERIC}', "");
Expect(0, 65045, '\p{Is_Line_Break=-_infix_NUMERIC}', "");
Expect(1, 65045, '\p{^Is_Line_Break=-_infix_NUMERIC}', "");
Expect(1, 65045, '\P{Is_Line_Break=-_infix_NUMERIC}', "");
Expect(0, 65045, '\P{^Is_Line_Break=-_infix_NUMERIC}', "");
Error('\p{Is_Lb=:=__is}');
Error('\P{Is_Lb=:=__is}');
Expect(1, 65044, '\p{Is_Lb=is}', "");
Expect(0, 65044, '\p{^Is_Lb=is}', "");
Expect(0, 65044, '\P{Is_Lb=is}', "");
Expect(1, 65044, '\P{^Is_Lb=is}', "");
Expect(0, 65045, '\p{Is_Lb=is}', "");
Expect(1, 65045, '\p{^Is_Lb=is}', "");
Expect(1, 65045, '\P{Is_Lb=is}', "");
Expect(0, 65045, '\P{^Is_Lb=is}', "");
Expect(1, 65044, '\p{Is_Lb=  is}', "");
Expect(0, 65044, '\p{^Is_Lb=  is}', "");
Expect(0, 65044, '\P{Is_Lb=  is}', "");
Expect(1, 65044, '\P{^Is_Lb=  is}', "");
Expect(0, 65045, '\p{Is_Lb=  is}', "");
Expect(1, 65045, '\p{^Is_Lb=  is}', "");
Expect(1, 65045, '\P{Is_Lb=  is}', "");
Expect(0, 65045, '\P{^Is_Lb=  is}', "");
Error('\p{Line_Break=	 inseparable:=}');
Error('\P{Line_Break=	 inseparable:=}');
Expect(1, 65049, '\p{Line_Break=inseparable}', "");
Expect(0, 65049, '\p{^Line_Break=inseparable}', "");
Expect(0, 65049, '\P{Line_Break=inseparable}', "");
Expect(1, 65049, '\P{^Line_Break=inseparable}', "");
Expect(0, 65050, '\p{Line_Break=inseparable}', "");
Expect(1, 65050, '\p{^Line_Break=inseparable}', "");
Expect(1, 65050, '\P{Line_Break=inseparable}', "");
Expect(0, 65050, '\P{^Line_Break=inseparable}', "");
Expect(1, 65049, '\p{Line_Break:  INSEPARABLE}', "");
Expect(0, 65049, '\p{^Line_Break:  INSEPARABLE}', "");
Expect(0, 65049, '\P{Line_Break:  INSEPARABLE}', "");
Expect(1, 65049, '\P{^Line_Break:  INSEPARABLE}', "");
Expect(0, 65050, '\p{Line_Break:  INSEPARABLE}', "");
Expect(1, 65050, '\p{^Line_Break:  INSEPARABLE}', "");
Expect(1, 65050, '\P{Line_Break:  INSEPARABLE}', "");
Expect(0, 65050, '\P{^Line_Break:  INSEPARABLE}', "");
Error('\p{Lb=	_in/a/}');
Error('\P{Lb=	_in/a/}');
Expect(1, 65049, '\p{Lb=in}', "");
Expect(0, 65049, '\p{^Lb=in}', "");
Expect(0, 65049, '\P{Lb=in}', "");
Expect(1, 65049, '\P{^Lb=in}', "");
Expect(0, 65050, '\p{Lb=in}', "");
Expect(1, 65050, '\p{^Lb=in}', "");
Expect(1, 65050, '\P{Lb=in}', "");
Expect(0, 65050, '\P{^Lb=in}', "");
Expect(1, 65049, '\p{Lb=_-IN}', "");
Expect(0, 65049, '\p{^Lb=_-IN}', "");
Expect(0, 65049, '\P{Lb=_-IN}', "");
Expect(1, 65049, '\P{^Lb=_-IN}', "");
Expect(0, 65050, '\p{Lb=_-IN}', "");
Expect(1, 65050, '\p{^Lb=_-IN}', "");
Expect(1, 65050, '\P{Lb=_-IN}', "");
Expect(0, 65050, '\P{^Lb=_-IN}', "");
Error('\p{Is_Line_Break: _INSEPERABLE/a/}');
Error('\P{Is_Line_Break: _INSEPERABLE/a/}');
Expect(1, 65049, '\p{Is_Line_Break=inseperable}', "");
Expect(0, 65049, '\p{^Is_Line_Break=inseperable}', "");
Expect(0, 65049, '\P{Is_Line_Break=inseperable}', "");
Expect(1, 65049, '\P{^Is_Line_Break=inseperable}', "");
Expect(0, 65050, '\p{Is_Line_Break=inseperable}', "");
Expect(1, 65050, '\p{^Is_Line_Break=inseperable}', "");
Expect(1, 65050, '\P{Is_Line_Break=inseperable}', "");
Expect(0, 65050, '\P{^Is_Line_Break=inseperable}', "");
Expect(1, 65049, '\p{Is_Line_Break=	Inseperable}', "");
Expect(0, 65049, '\p{^Is_Line_Break=	Inseperable}', "");
Expect(0, 65049, '\P{Is_Line_Break=	Inseperable}', "");
Expect(1, 65049, '\P{^Is_Line_Break=	Inseperable}', "");
Expect(0, 65050, '\p{Is_Line_Break=	Inseperable}', "");
Expect(1, 65050, '\p{^Is_Line_Break=	Inseperable}', "");
Expect(1, 65050, '\P{Is_Line_Break=	Inseperable}', "");
Expect(0, 65050, '\P{^Is_Line_Break=	Inseperable}', "");
Error('\p{Is_Lb:   _INSEPARABLE:=}');
Error('\P{Is_Lb:   _INSEPARABLE:=}');
Expect(1, 65049, '\p{Is_Lb=inseparable}', "");
Expect(0, 65049, '\p{^Is_Lb=inseparable}', "");
Expect(0, 65049, '\P{Is_Lb=inseparable}', "");
Expect(1, 65049, '\P{^Is_Lb=inseparable}', "");
Expect(0, 65050, '\p{Is_Lb=inseparable}', "");
Expect(1, 65050, '\p{^Is_Lb=inseparable}', "");
Expect(1, 65050, '\P{Is_Lb=inseparable}', "");
Expect(0, 65050, '\P{^Is_Lb=inseparable}', "");
Expect(1, 65049, '\p{Is_Lb:-	Inseparable}', "");
Expect(0, 65049, '\p{^Is_Lb:-	Inseparable}', "");
Expect(0, 65049, '\P{Is_Lb:-	Inseparable}', "");
Expect(1, 65049, '\P{^Is_Lb:-	Inseparable}', "");
Expect(0, 65050, '\p{Is_Lb:-	Inseparable}', "");
Expect(1, 65050, '\p{^Is_Lb:-	Inseparable}', "");
Expect(1, 65050, '\P{Is_Lb:-	Inseparable}', "");
Expect(0, 65050, '\P{^Is_Lb:-	Inseparable}', "");
Error('\p{Line_Break=/a/Complex_Context}');
Error('\P{Line_Break=/a/Complex_Context}');
Expect(1, 43743, '\p{Line_Break=complexcontext}', "");
Expect(0, 43743, '\p{^Line_Break=complexcontext}', "");
Expect(0, 43743, '\P{Line_Break=complexcontext}', "");
Expect(1, 43743, '\P{^Line_Break=complexcontext}', "");
Expect(0, 43744, '\p{Line_Break=complexcontext}', "");
Expect(1, 43744, '\p{^Line_Break=complexcontext}', "");
Expect(1, 43744, '\P{Line_Break=complexcontext}', "");
Expect(0, 43744, '\P{^Line_Break=complexcontext}', "");
Expect(1, 43743, '\p{Line_Break=_Complex_Context}', "");
Expect(0, 43743, '\p{^Line_Break=_Complex_Context}', "");
Expect(0, 43743, '\P{Line_Break=_Complex_Context}', "");
Expect(1, 43743, '\P{^Line_Break=_Complex_Context}', "");
Expect(0, 43744, '\p{Line_Break=_Complex_Context}', "");
Expect(1, 43744, '\p{^Line_Break=_Complex_Context}', "");
Expect(1, 43744, '\P{Line_Break=_Complex_Context}', "");
Expect(0, 43744, '\P{^Line_Break=_Complex_Context}', "");
Error('\p{Lb=-:=SA}');
Error('\P{Lb=-:=SA}');
Expect(1, 43743, '\p{Lb=sa}', "");
Expect(0, 43743, '\p{^Lb=sa}', "");
Expect(0, 43743, '\P{Lb=sa}', "");
Expect(1, 43743, '\P{^Lb=sa}', "");
Expect(0, 43744, '\p{Lb=sa}', "");
Expect(1, 43744, '\p{^Lb=sa}', "");
Expect(1, 43744, '\P{Lb=sa}', "");
Expect(0, 43744, '\P{^Lb=sa}', "");
Expect(1, 43743, '\p{Lb=-	sa}', "");
Expect(0, 43743, '\p{^Lb=-	sa}', "");
Expect(0, 43743, '\P{Lb=-	sa}', "");
Expect(1, 43743, '\P{^Lb=-	sa}', "");
Expect(0, 43744, '\p{Lb=-	sa}', "");
Expect(1, 43744, '\p{^Lb=-	sa}', "");
Expect(1, 43744, '\P{Lb=-	sa}', "");
Expect(0, 43744, '\P{^Lb=-	sa}', "");
Error('\p{Is_Line_Break=	:=Complex_Context}');
Error('\P{Is_Line_Break=	:=Complex_Context}');
Expect(1, 43743, '\p{Is_Line_Break=complexcontext}', "");
Expect(0, 43743, '\p{^Is_Line_Break=complexcontext}', "");
Expect(0, 43743, '\P{Is_Line_Break=complexcontext}', "");
Expect(1, 43743, '\P{^Is_Line_Break=complexcontext}', "");
Expect(0, 43744, '\p{Is_Line_Break=complexcontext}', "");
Expect(1, 43744, '\p{^Is_Line_Break=complexcontext}', "");
Expect(1, 43744, '\P{Is_Line_Break=complexcontext}', "");
Expect(0, 43744, '\P{^Is_Line_Break=complexcontext}', "");
Expect(1, 43743, '\p{Is_Line_Break=_ Complex_Context}', "");
Expect(0, 43743, '\p{^Is_Line_Break=_ Complex_Context}', "");
Expect(0, 43743, '\P{Is_Line_Break=_ Complex_Context}', "");
Expect(1, 43743, '\P{^Is_Line_Break=_ Complex_Context}', "");
Expect(0, 43744, '\p{Is_Line_Break=_ Complex_Context}', "");
Expect(1, 43744, '\p{^Is_Line_Break=_ Complex_Context}', "");
Expect(1, 43744, '\P{Is_Line_Break=_ Complex_Context}', "");
Expect(0, 43744, '\P{^Is_Line_Break=_ Complex_Context}', "");
Error('\p{Is_Lb=_:=SA}');
Error('\P{Is_Lb=_:=SA}');
Expect(1, 43743, '\p{Is_Lb=sa}', "");
Expect(0, 43743, '\p{^Is_Lb=sa}', "");
Expect(0, 43743, '\P{Is_Lb=sa}', "");
Expect(1, 43743, '\P{^Is_Lb=sa}', "");
Expect(0, 43744, '\p{Is_Lb=sa}', "");
Expect(1, 43744, '\p{^Is_Lb=sa}', "");
Expect(1, 43744, '\P{Is_Lb=sa}', "");
Expect(0, 43744, '\P{^Is_Lb=sa}', "");
Expect(1, 43743, '\p{Is_Lb=SA}', "");
Expect(0, 43743, '\p{^Is_Lb=SA}', "");
Expect(0, 43743, '\P{Is_Lb=SA}', "");
Expect(1, 43743, '\P{^Is_Lb=SA}', "");
Expect(0, 43744, '\p{Is_Lb=SA}', "");
Expect(1, 43744, '\p{^Is_Lb=SA}', "");
Expect(1, 43744, '\P{Is_Lb=SA}', "");
Expect(0, 43744, '\P{^Is_Lb=SA}', "");
Error('\p{Line_Break=:= -hyphen}');
Error('\P{Line_Break=:= -hyphen}');
Expect(1, 45, '\p{Line_Break=hyphen}', "");
Expect(0, 45, '\p{^Line_Break=hyphen}', "");
Expect(0, 45, '\P{Line_Break=hyphen}', "");
Expect(1, 45, '\P{^Line_Break=hyphen}', "");
Expect(0, 46, '\p{Line_Break=hyphen}', "");
Expect(1, 46, '\p{^Line_Break=hyphen}', "");
Expect(1, 46, '\P{Line_Break=hyphen}', "");
Expect(0, 46, '\P{^Line_Break=hyphen}', "");
Expect(1, 45, '\p{Line_Break=- hyphen}', "");
Expect(0, 45, '\p{^Line_Break=- hyphen}', "");
Expect(0, 45, '\P{Line_Break=- hyphen}', "");
Expect(1, 45, '\P{^Line_Break=- hyphen}', "");
Expect(0, 46, '\p{Line_Break=- hyphen}', "");
Expect(1, 46, '\p{^Line_Break=- hyphen}', "");
Expect(1, 46, '\P{Line_Break=- hyphen}', "");
Expect(0, 46, '\P{^Line_Break=- hyphen}', "");
Error('\p{Lb=__hy/a/}');
Error('\P{Lb=__hy/a/}');
Expect(1, 45, '\p{Lb=hy}', "");
Expect(0, 45, '\p{^Lb=hy}', "");
Expect(0, 45, '\P{Lb=hy}', "");
Expect(1, 45, '\P{^Lb=hy}', "");
Expect(0, 46, '\p{Lb=hy}', "");
Expect(1, 46, '\p{^Lb=hy}', "");
Expect(1, 46, '\P{Lb=hy}', "");
Expect(0, 46, '\P{^Lb=hy}', "");
Expect(1, 45, '\p{Lb=HY}', "");
Expect(0, 45, '\p{^Lb=HY}', "");
Expect(0, 45, '\P{Lb=HY}', "");
Expect(1, 45, '\P{^Lb=HY}', "");
Expect(0, 46, '\p{Lb=HY}', "");
Expect(1, 46, '\p{^Lb=HY}', "");
Expect(1, 46, '\P{Lb=HY}', "");
Expect(0, 46, '\P{^Lb=HY}', "");
Error('\p{Is_Line_Break=_ Hyphen:=}');
Error('\P{Is_Line_Break=_ Hyphen:=}');
Expect(1, 45, '\p{Is_Line_Break=hyphen}', "");
Expect(0, 45, '\p{^Is_Line_Break=hyphen}', "");
Expect(0, 45, '\P{Is_Line_Break=hyphen}', "");
Expect(1, 45, '\P{^Is_Line_Break=hyphen}', "");
Expect(0, 46, '\p{Is_Line_Break=hyphen}', "");
Expect(1, 46, '\p{^Is_Line_Break=hyphen}', "");
Expect(1, 46, '\P{Is_Line_Break=hyphen}', "");
Expect(0, 46, '\P{^Is_Line_Break=hyphen}', "");
Expect(1, 45, '\p{Is_Line_Break=- hyphen}', "");
Expect(0, 45, '\p{^Is_Line_Break=- hyphen}', "");
Expect(0, 45, '\P{Is_Line_Break=- hyphen}', "");
Expect(1, 45, '\P{^Is_Line_Break=- hyphen}', "");
Expect(0, 46, '\p{Is_Line_Break=- hyphen}', "");
Expect(1, 46, '\p{^Is_Line_Break=- hyphen}', "");
Expect(1, 46, '\P{Is_Line_Break=- hyphen}', "");
Expect(0, 46, '\P{^Is_Line_Break=- hyphen}', "");
Error('\p{Is_Lb=_HY:=}');
Error('\P{Is_Lb=_HY:=}');
Expect(1, 45, '\p{Is_Lb=hy}', "");
Expect(0, 45, '\p{^Is_Lb=hy}', "");
Expect(0, 45, '\P{Is_Lb=hy}', "");
Expect(1, 45, '\P{^Is_Lb=hy}', "");
Expect(0, 46, '\p{Is_Lb=hy}', "");
Expect(1, 46, '\p{^Is_Lb=hy}', "");
Expect(1, 46, '\P{Is_Lb=hy}', "");
Expect(0, 46, '\P{^Is_Lb=hy}', "");
Expect(1, 45, '\p{Is_Lb=_hy}', "");
Expect(0, 45, '\p{^Is_Lb=_hy}', "");
Expect(0, 45, '\P{Is_Lb=_hy}', "");
Expect(1, 45, '\P{^Is_Lb=_hy}', "");
Expect(0, 46, '\p{Is_Lb=_hy}', "");
Expect(1, 46, '\p{^Is_Lb=_hy}', "");
Expect(1, 46, '\P{Is_Lb=_hy}', "");
Expect(0, 46, '\P{^Is_Lb=_hy}', "");
Error('\p{Line_Break=-/a/Break_AFTER}');
Error('\P{Line_Break=-/a/Break_AFTER}');
Expect(1, 74867, '\p{Line_Break=breakafter}', "");
Expect(0, 74867, '\p{^Line_Break=breakafter}', "");
Expect(0, 74867, '\P{Line_Break=breakafter}', "");
Expect(1, 74867, '\P{^Line_Break=breakafter}', "");
Expect(0, 74868, '\p{Line_Break=breakafter}', "");
Expect(1, 74868, '\p{^Line_Break=breakafter}', "");
Expect(1, 74868, '\P{Line_Break=breakafter}', "");
Expect(0, 74868, '\P{^Line_Break=breakafter}', "");
Expect(1, 74867, '\p{Line_Break=_-BREAK_After}', "");
Expect(0, 74867, '\p{^Line_Break=_-BREAK_After}', "");
Expect(0, 74867, '\P{Line_Break=_-BREAK_After}', "");
Expect(1, 74867, '\P{^Line_Break=_-BREAK_After}', "");
Expect(0, 74868, '\p{Line_Break=_-BREAK_After}', "");
Expect(1, 74868, '\p{^Line_Break=_-BREAK_After}', "");
Expect(1, 74868, '\P{Line_Break=_-BREAK_After}', "");
Expect(0, 74868, '\P{^Line_Break=_-BREAK_After}', "");
Error('\p{Lb=--ba/a/}');
Error('\P{Lb=--ba/a/}');
Expect(1, 74867, '\p{Lb=ba}', "");
Expect(0, 74867, '\p{^Lb=ba}', "");
Expect(0, 74867, '\P{Lb=ba}', "");
Expect(1, 74867, '\P{^Lb=ba}', "");
Expect(0, 74868, '\p{Lb=ba}', "");
Expect(1, 74868, '\p{^Lb=ba}', "");
Expect(1, 74868, '\P{Lb=ba}', "");
Expect(0, 74868, '\P{^Lb=ba}', "");
Expect(1, 74867, '\p{Lb=	BA}', "");
Expect(0, 74867, '\p{^Lb=	BA}', "");
Expect(0, 74867, '\P{Lb=	BA}', "");
Expect(1, 74867, '\P{^Lb=	BA}', "");
Expect(0, 74868, '\p{Lb=	BA}', "");
Expect(1, 74868, '\p{^Lb=	BA}', "");
Expect(1, 74868, '\P{Lb=	BA}', "");
Expect(0, 74868, '\P{^Lb=	BA}', "");
Error('\p{Is_Line_Break=	-Break_after:=}');
Error('\P{Is_Line_Break=	-Break_after:=}');
Expect(1, 74867, '\p{Is_Line_Break:breakafter}', "");
Expect(0, 74867, '\p{^Is_Line_Break:breakafter}', "");
Expect(0, 74867, '\P{Is_Line_Break:breakafter}', "");
Expect(1, 74867, '\P{^Is_Line_Break:breakafter}', "");
Expect(0, 74868, '\p{Is_Line_Break:breakafter}', "");
Expect(1, 74868, '\p{^Is_Line_Break:breakafter}', "");
Expect(1, 74868, '\P{Is_Line_Break:breakafter}', "");
Expect(0, 74868, '\P{^Is_Line_Break:breakafter}', "");
Expect(1, 74867, '\p{Is_Line_Break=- Break_After}', "");
Expect(0, 74867, '\p{^Is_Line_Break=- Break_After}', "");
Expect(0, 74867, '\P{Is_Line_Break=- Break_After}', "");
Expect(1, 74867, '\P{^Is_Line_Break=- Break_After}', "");
Expect(0, 74868, '\p{Is_Line_Break=- Break_After}', "");
Expect(1, 74868, '\p{^Is_Line_Break=- Break_After}', "");
Expect(1, 74868, '\P{Is_Line_Break=- Break_After}', "");
Expect(0, 74868, '\P{^Is_Line_Break=- Break_After}', "");
Error('\p{Is_Lb=	/a/BA}');
Error('\P{Is_Lb=	/a/BA}');
Expect(1, 74867, '\p{Is_Lb=ba}', "");
Expect(0, 74867, '\p{^Is_Lb=ba}', "");
Expect(0, 74867, '\P{Is_Lb=ba}', "");
Expect(1, 74867, '\P{^Is_Lb=ba}', "");
Expect(0, 74868, '\p{Is_Lb=ba}', "");
Expect(1, 74868, '\p{^Is_Lb=ba}', "");
Expect(1, 74868, '\P{Is_Lb=ba}', "");
Expect(0, 74868, '\P{^Is_Lb=ba}', "");
Expect(1, 74867, '\p{Is_Lb=_-BA}', "");
Expect(0, 74867, '\p{^Is_Lb=_-BA}', "");
Expect(0, 74867, '\P{Is_Lb=_-BA}', "");
Expect(1, 74867, '\P{^Is_Lb=_-BA}', "");
Expect(0, 74868, '\p{Is_Lb=_-BA}', "");
Expect(1, 74868, '\p{^Is_Lb=_-BA}', "");
Expect(1, 74868, '\P{Is_Lb=_-BA}', "");
Expect(0, 74868, '\P{^Is_Lb=_-BA}', "");
Error('\p{Line_Break=/a/_-Line_Feed}');
Error('\P{Line_Break=/a/_-Line_Feed}');
Expect(1, 10, '\p{Line_Break=linefeed}', "");
Expect(0, 10, '\p{^Line_Break=linefeed}', "");
Expect(0, 10, '\P{Line_Break=linefeed}', "");
Expect(1, 10, '\P{^Line_Break=linefeed}', "");
Expect(0, 11, '\p{Line_Break=linefeed}', "");
Expect(1, 11, '\p{^Line_Break=linefeed}', "");
Expect(1, 11, '\P{Line_Break=linefeed}', "");
Expect(0, 11, '\P{^Line_Break=linefeed}', "");
Expect(1, 10, '\p{Line_Break=	-line_feed}', "");
Expect(0, 10, '\p{^Line_Break=	-line_feed}', "");
Expect(0, 10, '\P{Line_Break=	-line_feed}', "");
Expect(1, 10, '\P{^Line_Break=	-line_feed}', "");
Expect(0, 11, '\p{Line_Break=	-line_feed}', "");
Expect(1, 11, '\p{^Line_Break=	-line_feed}', "");
Expect(1, 11, '\P{Line_Break=	-line_feed}', "");
Expect(0, 11, '\P{^Line_Break=	-line_feed}', "");
Error('\p{Lb:-:=LF}');
Error('\P{Lb:-:=LF}');
Expect(1, 10, '\p{Lb=lf}', "");
Expect(0, 10, '\p{^Lb=lf}', "");
Expect(0, 10, '\P{Lb=lf}', "");
Expect(1, 10, '\P{^Lb=lf}', "");
Expect(0, 11, '\p{Lb=lf}', "");
Expect(1, 11, '\p{^Lb=lf}', "");
Expect(1, 11, '\P{Lb=lf}', "");
Expect(0, 11, '\P{^Lb=lf}', "");
Expect(1, 10, '\p{Lb=	_LF}', "");
Expect(0, 10, '\p{^Lb=	_LF}', "");
Expect(0, 10, '\P{Lb=	_LF}', "");
Expect(1, 10, '\P{^Lb=	_LF}', "");
Expect(0, 11, '\p{Lb=	_LF}', "");
Expect(1, 11, '\p{^Lb=	_LF}', "");
Expect(1, 11, '\P{Lb=	_LF}', "");
Expect(0, 11, '\P{^Lb=	_LF}', "");
Error('\p{Is_Line_Break=/a/__LINE_feed}');
Error('\P{Is_Line_Break=/a/__LINE_feed}');
Expect(1, 10, '\p{Is_Line_Break=linefeed}', "");
Expect(0, 10, '\p{^Is_Line_Break=linefeed}', "");
Expect(0, 10, '\P{Is_Line_Break=linefeed}', "");
Expect(1, 10, '\P{^Is_Line_Break=linefeed}', "");
Expect(0, 11, '\p{Is_Line_Break=linefeed}', "");
Expect(1, 11, '\p{^Is_Line_Break=linefeed}', "");
Expect(1, 11, '\P{Is_Line_Break=linefeed}', "");
Expect(0, 11, '\P{^Is_Line_Break=linefeed}', "");
Expect(1, 10, '\p{Is_Line_Break=	 line_feed}', "");
Expect(0, 10, '\p{^Is_Line_Break=	 line_feed}', "");
Expect(0, 10, '\P{Is_Line_Break=	 line_feed}', "");
Expect(1, 10, '\P{^Is_Line_Break=	 line_feed}', "");
Expect(0, 11, '\p{Is_Line_Break=	 line_feed}', "");
Expect(1, 11, '\p{^Is_Line_Break=	 line_feed}', "");
Expect(1, 11, '\P{Is_Line_Break=	 line_feed}', "");
Expect(0, 11, '\P{^Is_Line_Break=	 line_feed}', "");
Error('\p{Is_Lb=_:=LF}');
Error('\P{Is_Lb=_:=LF}');
Expect(1, 10, '\p{Is_Lb:   lf}', "");
Expect(0, 10, '\p{^Is_Lb:   lf}', "");
Expect(0, 10, '\P{Is_Lb:   lf}', "");
Expect(1, 10, '\P{^Is_Lb:   lf}', "");
Expect(0, 11, '\p{Is_Lb:   lf}', "");
Expect(1, 11, '\p{^Is_Lb:   lf}', "");
Expect(1, 11, '\P{Is_Lb:   lf}', "");
Expect(0, 11, '\P{^Is_Lb:   lf}', "");
Expect(1, 10, '\p{Is_Lb= _LF}', "");
Expect(0, 10, '\p{^Is_Lb= _LF}', "");
Expect(0, 10, '\P{Is_Lb= _LF}', "");
Expect(1, 10, '\P{^Is_Lb= _LF}', "");
Expect(0, 11, '\p{Is_Lb= _LF}', "");
Expect(1, 11, '\p{^Is_Lb= _LF}', "");
Expect(1, 11, '\P{Is_Lb= _LF}', "");
Expect(0, 11, '\P{^Is_Lb= _LF}', "");
Error('\p{Line_Break=:= Ambiguous}');
Error('\P{Line_Break=:= Ambiguous}');
Expect(1, 127386, '\p{Line_Break=ambiguous}', "");
Expect(0, 127386, '\p{^Line_Break=ambiguous}', "");
Expect(0, 127386, '\P{Line_Break=ambiguous}', "");
Expect(1, 127386, '\P{^Line_Break=ambiguous}', "");
Expect(0, 127387, '\p{Line_Break=ambiguous}', "");
Expect(1, 127387, '\p{^Line_Break=ambiguous}', "");
Expect(1, 127387, '\P{Line_Break=ambiguous}', "");
Expect(0, 127387, '\P{^Line_Break=ambiguous}', "");
Expect(1, 127386, '\p{Line_Break= _Ambiguous}', "");
Expect(0, 127386, '\p{^Line_Break= _Ambiguous}', "");
Expect(0, 127386, '\P{Line_Break= _Ambiguous}', "");
Expect(1, 127386, '\P{^Line_Break= _Ambiguous}', "");
Expect(0, 127387, '\p{Line_Break= _Ambiguous}', "");
Expect(1, 127387, '\p{^Line_Break= _Ambiguous}', "");
Expect(1, 127387, '\P{Line_Break= _Ambiguous}', "");
Expect(0, 127387, '\P{^Line_Break= _Ambiguous}', "");
Error('\p{Lb:	AI:=}');
Error('\P{Lb:	AI:=}');
Expect(1, 127386, '\p{Lb=ai}', "");
Expect(0, 127386, '\p{^Lb=ai}', "");
Expect(0, 127386, '\P{Lb=ai}', "");
Expect(1, 127386, '\P{^Lb=ai}', "");
Expect(0, 127387, '\p{Lb=ai}', "");
Expect(1, 127387, '\p{^Lb=ai}', "");
Expect(1, 127387, '\P{Lb=ai}', "");
Expect(0, 127387, '\P{^Lb=ai}', "");
Expect(1, 127386, '\p{Lb:	AI}', "");
Expect(0, 127386, '\p{^Lb:	AI}', "");
Expect(0, 127386, '\P{Lb:	AI}', "");
Expect(1, 127386, '\P{^Lb:	AI}', "");
Expect(0, 127387, '\p{Lb:	AI}', "");
Expect(1, 127387, '\p{^Lb:	AI}', "");
Expect(1, 127387, '\P{Lb:	AI}', "");
Expect(0, 127387, '\P{^Lb:	AI}', "");
Error('\p{Is_Line_Break=	-AMBIGUOUS:=}');
Error('\P{Is_Line_Break=	-AMBIGUOUS:=}');
Expect(1, 127386, '\p{Is_Line_Break=ambiguous}', "");
Expect(0, 127386, '\p{^Is_Line_Break=ambiguous}', "");
Expect(0, 127386, '\P{Is_Line_Break=ambiguous}', "");
Expect(1, 127386, '\P{^Is_Line_Break=ambiguous}', "");
Expect(0, 127387, '\p{Is_Line_Break=ambiguous}', "");
Expect(1, 127387, '\p{^Is_Line_Break=ambiguous}', "");
Expect(1, 127387, '\P{Is_Line_Break=ambiguous}', "");
Expect(0, 127387, '\P{^Is_Line_Break=ambiguous}', "");
Expect(1, 127386, '\p{Is_Line_Break=_Ambiguous}', "");
Expect(0, 127386, '\p{^Is_Line_Break=_Ambiguous}', "");
Expect(0, 127386, '\P{Is_Line_Break=_Ambiguous}', "");
Expect(1, 127386, '\P{^Is_Line_Break=_Ambiguous}', "");
Expect(0, 127387, '\p{Is_Line_Break=_Ambiguous}', "");
Expect(1, 127387, '\p{^Is_Line_Break=_Ambiguous}', "");
Expect(1, 127387, '\P{Is_Line_Break=_Ambiguous}', "");
Expect(0, 127387, '\P{^Is_Line_Break=_Ambiguous}', "");
Error('\p{Is_Lb=	:=AI}');
Error('\P{Is_Lb=	:=AI}');
Expect(1, 127386, '\p{Is_Lb=ai}', "");
Expect(0, 127386, '\p{^Is_Lb=ai}', "");
Expect(0, 127386, '\P{Is_Lb=ai}', "");
Expect(1, 127386, '\P{^Is_Lb=ai}', "");
Expect(0, 127387, '\p{Is_Lb=ai}', "");
Expect(1, 127387, '\p{^Is_Lb=ai}', "");
Expect(1, 127387, '\P{Is_Lb=ai}', "");
Expect(0, 127387, '\P{^Is_Lb=ai}', "");
Expect(1, 127386, '\p{Is_Lb:_	AI}', "");
Expect(0, 127386, '\p{^Is_Lb:_	AI}', "");
Expect(0, 127386, '\P{Is_Lb:_	AI}', "");
Expect(1, 127386, '\P{^Is_Lb:_	AI}', "");
Expect(0, 127387, '\p{Is_Lb:_	AI}', "");
Expect(1, 127387, '\p{^Is_Lb:_	AI}', "");
Expect(1, 127387, '\P{Is_Lb:_	AI}', "");
Expect(0, 127387, '\P{^Is_Lb:_	AI}', "");
Error('\p{Line_Break=__conditional_japanese_STARTER/a/}');
Error('\P{Line_Break=__conditional_japanese_STARTER/a/}');
Expect(1, 65392, '\p{Line_Break:	conditionaljapanesestarter}', "");
Expect(0, 65392, '\p{^Line_Break:	conditionaljapanesestarter}', "");
Expect(0, 65392, '\P{Line_Break:	conditionaljapanesestarter}', "");
Expect(1, 65392, '\P{^Line_Break:	conditionaljapanesestarter}', "");
Expect(0, 65393, '\p{Line_Break:	conditionaljapanesestarter}', "");
Expect(1, 65393, '\p{^Line_Break:	conditionaljapanesestarter}', "");
Expect(1, 65393, '\P{Line_Break:	conditionaljapanesestarter}', "");
Expect(0, 65393, '\P{^Line_Break:	conditionaljapanesestarter}', "");
Expect(1, 65392, '\p{Line_Break=_ conditional_Japanese_Starter}', "");
Expect(0, 65392, '\p{^Line_Break=_ conditional_Japanese_Starter}', "");
Expect(0, 65392, '\P{Line_Break=_ conditional_Japanese_Starter}', "");
Expect(1, 65392, '\P{^Line_Break=_ conditional_Japanese_Starter}', "");
Expect(0, 65393, '\p{Line_Break=_ conditional_Japanese_Starter}', "");
Expect(1, 65393, '\p{^Line_Break=_ conditional_Japanese_Starter}', "");
Expect(1, 65393, '\P{Line_Break=_ conditional_Japanese_Starter}', "");
Expect(0, 65393, '\P{^Line_Break=_ conditional_Japanese_Starter}', "");
Error('\p{Lb=-CJ:=}');
Error('\P{Lb=-CJ:=}');
Expect(1, 65392, '\p{Lb=cj}', "");
Expect(0, 65392, '\p{^Lb=cj}', "");
Expect(0, 65392, '\P{Lb=cj}', "");
Expect(1, 65392, '\P{^Lb=cj}', "");
Expect(0, 65393, '\p{Lb=cj}', "");
Expect(1, 65393, '\p{^Lb=cj}', "");
Expect(1, 65393, '\P{Lb=cj}', "");
Expect(0, 65393, '\P{^Lb=cj}', "");
Expect(1, 65392, '\p{Lb=-CJ}', "");
Expect(0, 65392, '\p{^Lb=-CJ}', "");
Expect(0, 65392, '\P{Lb=-CJ}', "");
Expect(1, 65392, '\P{^Lb=-CJ}', "");
Expect(0, 65393, '\p{Lb=-CJ}', "");
Expect(1, 65393, '\p{^Lb=-CJ}', "");
Expect(1, 65393, '\P{Lb=-CJ}', "");
Expect(0, 65393, '\P{^Lb=-CJ}', "");
Error('\p{Is_Line_Break=:=Conditional_Japanese_STARTER}');
Error('\P{Is_Line_Break=:=Conditional_Japanese_STARTER}');
Expect(1, 65392, '\p{Is_Line_Break=conditionaljapanesestarter}', "");
Expect(0, 65392, '\p{^Is_Line_Break=conditionaljapanesestarter}', "");
Expect(0, 65392, '\P{Is_Line_Break=conditionaljapanesestarter}', "");
Expect(1, 65392, '\P{^Is_Line_Break=conditionaljapanesestarter}', "");
Expect(0, 65393, '\p{Is_Line_Break=conditionaljapanesestarter}', "");
Expect(1, 65393, '\p{^Is_Line_Break=conditionaljapanesestarter}', "");
Expect(1, 65393, '\P{Is_Line_Break=conditionaljapanesestarter}', "");
Expect(0, 65393, '\P{^Is_Line_Break=conditionaljapanesestarter}', "");
Expect(1, 65392, '\p{Is_Line_Break=_-CONDITIONAL_Japanese_starter}', "");
Expect(0, 65392, '\p{^Is_Line_Break=_-CONDITIONAL_Japanese_starter}', "");
Expect(0, 65392, '\P{Is_Line_Break=_-CONDITIONAL_Japanese_starter}', "");
Expect(1, 65392, '\P{^Is_Line_Break=_-CONDITIONAL_Japanese_starter}', "");
Expect(0, 65393, '\p{Is_Line_Break=_-CONDITIONAL_Japanese_starter}', "");
Expect(1, 65393, '\p{^Is_Line_Break=_-CONDITIONAL_Japanese_starter}', "");
Expect(1, 65393, '\P{Is_Line_Break=_-CONDITIONAL_Japanese_starter}', "");
Expect(0, 65393, '\P{^Is_Line_Break=_-CONDITIONAL_Japanese_starter}', "");
Error('\p{Is_Lb= -cj:=}');
Error('\P{Is_Lb= -cj:=}');
Expect(1, 65392, '\p{Is_Lb=cj}', "");
Expect(0, 65392, '\p{^Is_Lb=cj}', "");
Expect(0, 65392, '\P{Is_Lb=cj}', "");
Expect(1, 65392, '\P{^Is_Lb=cj}', "");
Expect(0, 65393, '\p{Is_Lb=cj}', "");
Expect(1, 65393, '\p{^Is_Lb=cj}', "");
Expect(1, 65393, '\P{Is_Lb=cj}', "");
Expect(0, 65393, '\P{^Is_Lb=cj}', "");
Expect(1, 65392, '\p{Is_Lb= CJ}', "");
Expect(0, 65392, '\p{^Is_Lb= CJ}', "");
Expect(0, 65392, '\P{Is_Lb= CJ}', "");
Expect(1, 65392, '\P{^Is_Lb= CJ}', "");
Expect(0, 65393, '\p{Is_Lb= CJ}', "");
Expect(1, 65393, '\p{^Is_Lb= CJ}', "");
Expect(1, 65393, '\P{Is_Lb= CJ}', "");
Expect(0, 65393, '\P{^Is_Lb= CJ}', "");
Error('\p{Line_Break=_/a/Contingent_break}');
Error('\P{Line_Break=_/a/Contingent_break}');
Expect(1, 65532, '\p{Line_Break=contingentbreak}', "");
Expect(0, 65532, '\p{^Line_Break=contingentbreak}', "");
Expect(0, 65532, '\P{Line_Break=contingentbreak}', "");
Expect(1, 65532, '\P{^Line_Break=contingentbreak}', "");
Expect(0, 65533, '\p{Line_Break=contingentbreak}', "");
Expect(1, 65533, '\p{^Line_Break=contingentbreak}', "");
Expect(1, 65533, '\P{Line_Break=contingentbreak}', "");
Expect(0, 65533, '\P{^Line_Break=contingentbreak}', "");
Expect(1, 65532, '\p{Line_Break=_-Contingent_Break}', "");
Expect(0, 65532, '\p{^Line_Break=_-Contingent_Break}', "");
Expect(0, 65532, '\P{Line_Break=_-Contingent_Break}', "");
Expect(1, 65532, '\P{^Line_Break=_-Contingent_Break}', "");
Expect(0, 65533, '\p{Line_Break=_-Contingent_Break}', "");
Expect(1, 65533, '\p{^Line_Break=_-Contingent_Break}', "");
Expect(1, 65533, '\P{Line_Break=_-Contingent_Break}', "");
Expect(0, 65533, '\P{^Line_Break=_-Contingent_Break}', "");
Error('\p{Lb= _cb/a/}');
Error('\P{Lb= _cb/a/}');
Expect(1, 65532, '\p{Lb=cb}', "");
Expect(0, 65532, '\p{^Lb=cb}', "");
Expect(0, 65532, '\P{Lb=cb}', "");
Expect(1, 65532, '\P{^Lb=cb}', "");
Expect(0, 65533, '\p{Lb=cb}', "");
Expect(1, 65533, '\p{^Lb=cb}', "");
Expect(1, 65533, '\P{Lb=cb}', "");
Expect(0, 65533, '\P{^Lb=cb}', "");
Expect(1, 65532, '\p{Lb=	 cb}', "");
Expect(0, 65532, '\p{^Lb=	 cb}', "");
Expect(0, 65532, '\P{Lb=	 cb}', "");
Expect(1, 65532, '\P{^Lb=	 cb}', "");
Expect(0, 65533, '\p{Lb=	 cb}', "");
Expect(1, 65533, '\p{^Lb=	 cb}', "");
Expect(1, 65533, '\P{Lb=	 cb}', "");
Expect(0, 65533, '\P{^Lb=	 cb}', "");
Error('\p{Is_Line_Break=	:=Contingent_Break}');
Error('\P{Is_Line_Break=	:=Contingent_Break}');
Expect(1, 65532, '\p{Is_Line_Break=contingentbreak}', "");
Expect(0, 65532, '\p{^Is_Line_Break=contingentbreak}', "");
Expect(0, 65532, '\P{Is_Line_Break=contingentbreak}', "");
Expect(1, 65532, '\P{^Is_Line_Break=contingentbreak}', "");
Expect(0, 65533, '\p{Is_Line_Break=contingentbreak}', "");
Expect(1, 65533, '\p{^Is_Line_Break=contingentbreak}', "");
Expect(1, 65533, '\P{Is_Line_Break=contingentbreak}', "");
Expect(0, 65533, '\P{^Is_Line_Break=contingentbreak}', "");
Expect(1, 65532, '\p{Is_Line_Break=		Contingent_break}', "");
Expect(0, 65532, '\p{^Is_Line_Break=		Contingent_break}', "");
Expect(0, 65532, '\P{Is_Line_Break=		Contingent_break}', "");
Expect(1, 65532, '\P{^Is_Line_Break=		Contingent_break}', "");
Expect(0, 65533, '\p{Is_Line_Break=		Contingent_break}', "");
Expect(1, 65533, '\p{^Is_Line_Break=		Contingent_break}', "");
Expect(1, 65533, '\P{Is_Line_Break=		Contingent_break}', "");
Expect(0, 65533, '\P{^Is_Line_Break=		Contingent_break}', "");
Error('\p{Is_Lb=/a/- CB}');
Error('\P{Is_Lb=/a/- CB}');
Expect(1, 65532, '\p{Is_Lb=cb}', "");
Expect(0, 65532, '\p{^Is_Lb=cb}', "");
Expect(0, 65532, '\P{Is_Lb=cb}', "");
Expect(1, 65532, '\P{^Is_Lb=cb}', "");
Expect(0, 65533, '\p{Is_Lb=cb}', "");
Expect(1, 65533, '\p{^Is_Lb=cb}', "");
Expect(1, 65533, '\P{Is_Lb=cb}', "");
Expect(0, 65533, '\P{^Is_Lb=cb}', "");
Expect(1, 65532, '\p{Is_Lb=--CB}', "");
Expect(0, 65532, '\p{^Is_Lb=--CB}', "");
Expect(0, 65532, '\P{Is_Lb=--CB}', "");
Expect(1, 65532, '\P{^Is_Lb=--CB}', "");
Expect(0, 65533, '\p{Is_Lb=--CB}', "");
Expect(1, 65533, '\p{^Is_Lb=--CB}', "");
Expect(1, 65533, '\P{Is_Lb=--CB}', "");
Expect(0, 65533, '\P{^Is_Lb=--CB}', "");
Error('\p{Line_Break=	/a/IDEOGRAPHIC}');
Error('\P{Line_Break=	/a/IDEOGRAPHIC}');
Expect(1, 262141, '\p{Line_Break=ideographic}', "");
Expect(0, 262141, '\p{^Line_Break=ideographic}', "");
Expect(0, 262141, '\P{Line_Break=ideographic}', "");
Expect(1, 262141, '\P{^Line_Break=ideographic}', "");
Expect(0, 262144, '\p{Line_Break=ideographic}', "");
Expect(1, 262144, '\p{^Line_Break=ideographic}', "");
Expect(1, 262144, '\P{Line_Break=ideographic}', "");
Expect(0, 262144, '\P{^Line_Break=ideographic}', "");
Expect(1, 262141, '\p{Line_Break=	_ideographic}', "");
Expect(0, 262141, '\p{^Line_Break=	_ideographic}', "");
Expect(0, 262141, '\P{Line_Break=	_ideographic}', "");
Expect(1, 262141, '\P{^Line_Break=	_ideographic}', "");
Expect(0, 262144, '\p{Line_Break=	_ideographic}', "");
Expect(1, 262144, '\p{^Line_Break=	_ideographic}', "");
Expect(1, 262144, '\P{Line_Break=	_ideographic}', "");
Expect(0, 262144, '\P{^Line_Break=	_ideographic}', "");
Error('\p{Lb=:=_ ID}');
Error('\P{Lb=:=_ ID}');
Expect(1, 262141, '\p{Lb=id}', "");
Expect(0, 262141, '\p{^Lb=id}', "");
Expect(0, 262141, '\P{Lb=id}', "");
Expect(1, 262141, '\P{^Lb=id}', "");
Expect(0, 262144, '\p{Lb=id}', "");
Expect(1, 262144, '\p{^Lb=id}', "");
Expect(1, 262144, '\P{Lb=id}', "");
Expect(0, 262144, '\P{^Lb=id}', "");
Expect(1, 262141, '\p{Lb= 	ID}', "");
Expect(0, 262141, '\p{^Lb= 	ID}', "");
Expect(0, 262141, '\P{Lb= 	ID}', "");
Expect(1, 262141, '\P{^Lb= 	ID}', "");
Expect(0, 262144, '\p{Lb= 	ID}', "");
Expect(1, 262144, '\p{^Lb= 	ID}', "");
Expect(1, 262144, '\P{Lb= 	ID}', "");
Expect(0, 262144, '\P{^Lb= 	ID}', "");
Error('\p{Is_Line_Break=-/a/IDEOGRAPHIC}');
Error('\P{Is_Line_Break=-/a/IDEOGRAPHIC}');
Expect(1, 262141, '\p{Is_Line_Break=ideographic}', "");
Expect(0, 262141, '\p{^Is_Line_Break=ideographic}', "");
Expect(0, 262141, '\P{Is_Line_Break=ideographic}', "");
Expect(1, 262141, '\P{^Is_Line_Break=ideographic}', "");
Expect(0, 262144, '\p{Is_Line_Break=ideographic}', "");
Expect(1, 262144, '\p{^Is_Line_Break=ideographic}', "");
Expect(1, 262144, '\P{Is_Line_Break=ideographic}', "");
Expect(0, 262144, '\P{^Is_Line_Break=ideographic}', "");
Expect(1, 262141, '\p{Is_Line_Break=- ideographic}', "");
Expect(0, 262141, '\p{^Is_Line_Break=- ideographic}', "");
Expect(0, 262141, '\P{Is_Line_Break=- ideographic}', "");
Expect(1, 262141, '\P{^Is_Line_Break=- ideographic}', "");
Expect(0, 262144, '\p{Is_Line_Break=- ideographic}', "");
Expect(1, 262144, '\p{^Is_Line_Break=- ideographic}', "");
Expect(1, 262144, '\P{Is_Line_Break=- ideographic}', "");
Expect(0, 262144, '\P{^Is_Line_Break=- ideographic}', "");
Error('\p{Is_Lb:   		ID:=}');
Error('\P{Is_Lb:   		ID:=}');
Expect(1, 262141, '\p{Is_Lb=id}', "");
Expect(0, 262141, '\p{^Is_Lb=id}', "");
Expect(0, 262141, '\P{Is_Lb=id}', "");
Expect(1, 262141, '\P{^Is_Lb=id}', "");
Expect(0, 262144, '\p{Is_Lb=id}', "");
Expect(1, 262144, '\p{^Is_Lb=id}', "");
Expect(1, 262144, '\P{Is_Lb=id}', "");
Expect(0, 262144, '\P{^Is_Lb=id}', "");
Expect(1, 262141, '\p{Is_Lb=-id}', "");
Expect(0, 262141, '\p{^Is_Lb=-id}', "");
Expect(0, 262141, '\P{Is_Lb=-id}', "");
Expect(1, 262141, '\P{^Is_Lb=-id}', "");
Expect(0, 262144, '\p{Is_Lb=-id}', "");
Expect(1, 262144, '\p{^Is_Lb=-id}', "");
Expect(1, 262144, '\P{Is_Lb=-id}', "");
Expect(0, 262144, '\P{^Is_Lb=-id}', "");
Error('\p{Line_Break=:=_Exclamation}');
Error('\P{Line_Break=:=_Exclamation}');
Expect(1, 65311, '\p{Line_Break=exclamation}', "");
Expect(0, 65311, '\p{^Line_Break=exclamation}', "");
Expect(0, 65311, '\P{Line_Break=exclamation}', "");
Expect(1, 65311, '\P{^Line_Break=exclamation}', "");
Expect(0, 65312, '\p{Line_Break=exclamation}', "");
Expect(1, 65312, '\p{^Line_Break=exclamation}', "");
Expect(1, 65312, '\P{Line_Break=exclamation}', "");
Expect(0, 65312, '\P{^Line_Break=exclamation}', "");
Expect(1, 65311, '\p{Line_Break=_ Exclamation}', "");
Expect(0, 65311, '\p{^Line_Break=_ Exclamation}', "");
Expect(0, 65311, '\P{Line_Break=_ Exclamation}', "");
Expect(1, 65311, '\P{^Line_Break=_ Exclamation}', "");
Expect(0, 65312, '\p{Line_Break=_ Exclamation}', "");
Expect(1, 65312, '\p{^Line_Break=_ Exclamation}', "");
Expect(1, 65312, '\P{Line_Break=_ Exclamation}', "");
Expect(0, 65312, '\P{^Line_Break=_ Exclamation}', "");
Error('\p{Lb=/a/EX}');
Error('\P{Lb=/a/EX}');
Expect(1, 65311, '\p{Lb:   ex}', "");
Expect(0, 65311, '\p{^Lb:   ex}', "");
Expect(0, 65311, '\P{Lb:   ex}', "");
Expect(1, 65311, '\P{^Lb:   ex}', "");
Expect(0, 65312, '\p{Lb:   ex}', "");
Expect(1, 65312, '\p{^Lb:   ex}', "");
Expect(1, 65312, '\P{Lb:   ex}', "");
Expect(0, 65312, '\P{^Lb:   ex}', "");
Expect(1, 65311, '\p{Lb=-ex}', "");
Expect(0, 65311, '\p{^Lb=-ex}', "");
Expect(0, 65311, '\P{Lb=-ex}', "");
Expect(1, 65311, '\P{^Lb=-ex}', "");
Expect(0, 65312, '\p{Lb=-ex}', "");
Expect(1, 65312, '\p{^Lb=-ex}', "");
Expect(1, 65312, '\P{Lb=-ex}', "");
Expect(0, 65312, '\P{^Lb=-ex}', "");
Error('\p{Is_Line_Break= :=Exclamation}');
Error('\P{Is_Line_Break= :=Exclamation}');
Expect(1, 65311, '\p{Is_Line_Break=exclamation}', "");
Expect(0, 65311, '\p{^Is_Line_Break=exclamation}', "");
Expect(0, 65311, '\P{Is_Line_Break=exclamation}', "");
Expect(1, 65311, '\P{^Is_Line_Break=exclamation}', "");
Expect(0, 65312, '\p{Is_Line_Break=exclamation}', "");
Expect(1, 65312, '\p{^Is_Line_Break=exclamation}', "");
Expect(1, 65312, '\P{Is_Line_Break=exclamation}', "");
Expect(0, 65312, '\P{^Is_Line_Break=exclamation}', "");
Expect(1, 65311, '\p{Is_Line_Break=_Exclamation}', "");
Expect(0, 65311, '\p{^Is_Line_Break=_Exclamation}', "");
Expect(0, 65311, '\P{Is_Line_Break=_Exclamation}', "");
Expect(1, 65311, '\P{^Is_Line_Break=_Exclamation}', "");
Expect(0, 65312, '\p{Is_Line_Break=_Exclamation}', "");
Expect(1, 65312, '\p{^Is_Line_Break=_Exclamation}', "");
Expect(1, 65312, '\P{Is_Line_Break=_Exclamation}', "");
Expect(0, 65312, '\P{^Is_Line_Break=_Exclamation}', "");
Error('\p{Is_Lb= :=EX}');
Error('\P{Is_Lb= :=EX}');
Expect(1, 65311, '\p{Is_Lb=ex}', "");
Expect(0, 65311, '\p{^Is_Lb=ex}', "");
Expect(0, 65311, '\P{Is_Lb=ex}', "");
Expect(1, 65311, '\P{^Is_Lb=ex}', "");
Expect(0, 65312, '\p{Is_Lb=ex}', "");
Expect(1, 65312, '\p{^Is_Lb=ex}', "");
Expect(1, 65312, '\P{Is_Lb=ex}', "");
Expect(0, 65312, '\P{^Is_Lb=ex}', "");
Expect(1, 65311, '\p{Is_Lb=	EX}', "");
Expect(0, 65311, '\p{^Is_Lb=	EX}', "");
Expect(0, 65311, '\P{Is_Lb=	EX}', "");
Expect(1, 65311, '\P{^Is_Lb=	EX}', "");
Expect(0, 65312, '\p{Is_Lb=	EX}', "");
Expect(1, 65312, '\p{^Is_Lb=	EX}', "");
Expect(1, 65312, '\P{Is_Lb=	EX}', "");
Expect(0, 65312, '\P{^Is_Lb=	EX}', "");
Error('\p{Line_Break=:=	 Word_joiner}');
Error('\P{Line_Break=:=	 Word_joiner}');
Expect(1, 65279, '\p{Line_Break=wordjoiner}', "");
Expect(0, 65279, '\p{^Line_Break=wordjoiner}', "");
Expect(0, 65279, '\P{Line_Break=wordjoiner}', "");
Expect(1, 65279, '\P{^Line_Break=wordjoiner}', "");
Expect(0, 65280, '\p{Line_Break=wordjoiner}', "");
Expect(1, 65280, '\p{^Line_Break=wordjoiner}', "");
Expect(1, 65280, '\P{Line_Break=wordjoiner}', "");
Expect(0, 65280, '\P{^Line_Break=wordjoiner}', "");
Expect(1, 65279, '\p{Line_Break=  word_JOINER}', "");
Expect(0, 65279, '\p{^Line_Break=  word_JOINER}', "");
Expect(0, 65279, '\P{Line_Break=  word_JOINER}', "");
Expect(1, 65279, '\P{^Line_Break=  word_JOINER}', "");
Expect(0, 65280, '\p{Line_Break=  word_JOINER}', "");
Expect(1, 65280, '\p{^Line_Break=  word_JOINER}', "");
Expect(1, 65280, '\P{Line_Break=  word_JOINER}', "");
Expect(0, 65280, '\P{^Line_Break=  word_JOINER}', "");
Error('\p{Lb=-WJ:=}');
Error('\P{Lb=-WJ:=}');
Expect(1, 65279, '\p{Lb=wj}', "");
Expect(0, 65279, '\p{^Lb=wj}', "");
Expect(0, 65279, '\P{Lb=wj}', "");
Expect(1, 65279, '\P{^Lb=wj}', "");
Expect(0, 65280, '\p{Lb=wj}', "");
Expect(1, 65280, '\p{^Lb=wj}', "");
Expect(1, 65280, '\P{Lb=wj}', "");
Expect(0, 65280, '\P{^Lb=wj}', "");
Error('\p{Is_Line_Break=-:=WORD_joiner}');
Error('\P{Is_Line_Break=-:=WORD_joiner}');
Expect(1, 65279, '\p{Is_Line_Break=wordjoiner}', "");
Expect(0, 65279, '\p{^Is_Line_Break=wordjoiner}', "");
Expect(0, 65279, '\P{Is_Line_Break=wordjoiner}', "");
Expect(1, 65279, '\P{^Is_Line_Break=wordjoiner}', "");
Expect(0, 65280, '\p{Is_Line_Break=wordjoiner}', "");
Expect(1, 65280, '\p{^Is_Line_Break=wordjoiner}', "");
Expect(1, 65280, '\P{Is_Line_Break=wordjoiner}', "");
Expect(0, 65280, '\P{^Is_Line_Break=wordjoiner}', "");
Expect(1, 65279, '\p{Is_Line_Break=--WORD_JOINER}', "");
Expect(0, 65279, '\p{^Is_Line_Break=--WORD_JOINER}', "");
Expect(0, 65279, '\P{Is_Line_Break=--WORD_JOINER}', "");
Expect(1, 65279, '\P{^Is_Line_Break=--WORD_JOINER}', "");
Expect(0, 65280, '\p{Is_Line_Break=--WORD_JOINER}', "");
Expect(1, 65280, '\p{^Is_Line_Break=--WORD_JOINER}', "");
Expect(1, 65280, '\P{Is_Line_Break=--WORD_JOINER}', "");
Expect(0, 65280, '\P{^Is_Line_Break=--WORD_JOINER}', "");
Error('\p{Is_Lb=	wj:=}');
Error('\P{Is_Lb=	wj:=}');
Expect(1, 65279, '\p{Is_Lb=wj}', "");
Expect(0, 65279, '\p{^Is_Lb=wj}', "");
Expect(0, 65279, '\P{Is_Lb=wj}', "");
Expect(1, 65279, '\P{^Is_Lb=wj}', "");
Expect(0, 65280, '\p{Is_Lb=wj}', "");
Expect(1, 65280, '\p{^Is_Lb=wj}', "");
Expect(1, 65280, '\P{Is_Lb=wj}', "");
Expect(0, 65280, '\P{^Is_Lb=wj}', "");
Expect(1, 65279, '\p{Is_Lb=- WJ}', "");
Expect(0, 65279, '\p{^Is_Lb=- WJ}', "");
Expect(0, 65279, '\P{Is_Lb=- WJ}', "");
Expect(1, 65279, '\P{^Is_Lb=- WJ}', "");
Expect(0, 65280, '\p{Is_Lb=- WJ}', "");
Expect(1, 65280, '\p{^Is_Lb=- WJ}', "");
Expect(1, 65280, '\P{Is_Lb=- WJ}', "");
Expect(0, 65280, '\P{^Is_Lb=- WJ}', "");
Error('\p{Line_Break= H2/a/}');
Error('\P{Line_Break= H2/a/}');
Expect(1, 55176, '\p{Line_Break=h2}', "");
Expect(0, 55176, '\p{^Line_Break=h2}', "");
Expect(0, 55176, '\P{Line_Break=h2}', "");
Expect(1, 55176, '\P{^Line_Break=h2}', "");
Expect(0, 55177, '\p{Line_Break=h2}', "");
Expect(1, 55177, '\p{^Line_Break=h2}', "");
Expect(1, 55177, '\P{Line_Break=h2}', "");
Expect(0, 55177, '\P{^Line_Break=h2}', "");
Expect(1, 55176, '\p{Line_Break:    -H2}', "");
Expect(0, 55176, '\p{^Line_Break:    -H2}', "");
Expect(0, 55176, '\P{Line_Break:    -H2}', "");
Expect(1, 55176, '\P{^Line_Break:    -H2}', "");
Expect(0, 55177, '\p{Line_Break:    -H2}', "");
Expect(1, 55177, '\p{^Line_Break:    -H2}', "");
Expect(1, 55177, '\P{Line_Break:    -H2}', "");
Expect(0, 55177, '\P{^Line_Break:    -H2}', "");
Error('\p{Lb=:=H2}');
Error('\P{Lb=:=H2}');
Expect(1, 55176, '\p{Lb=h2}', "");
Expect(0, 55176, '\p{^Lb=h2}', "");
Expect(0, 55176, '\P{Lb=h2}', "");
Expect(1, 55176, '\P{^Lb=h2}', "");
Expect(0, 55177, '\p{Lb=h2}', "");
Expect(1, 55177, '\p{^Lb=h2}', "");
Expect(1, 55177, '\P{Lb=h2}', "");
Expect(0, 55177, '\P{^Lb=h2}', "");
Expect(1, 55176, '\p{Lb:  -H2}', "");
Expect(0, 55176, '\p{^Lb:  -H2}', "");
Expect(0, 55176, '\P{Lb:  -H2}', "");
Expect(1, 55176, '\P{^Lb:  -H2}', "");
Expect(0, 55177, '\p{Lb:  -H2}', "");
Expect(1, 55177, '\p{^Lb:  -H2}', "");
Expect(1, 55177, '\P{Lb:  -H2}', "");
Expect(0, 55177, '\P{^Lb:  -H2}', "");
Error('\p{Is_Line_Break=-:=H2}');
Error('\P{Is_Line_Break=-:=H2}');
Expect(1, 55176, '\p{Is_Line_Break=h2}', "");
Expect(0, 55176, '\p{^Is_Line_Break=h2}', "");
Expect(0, 55176, '\P{Is_Line_Break=h2}', "");
Expect(1, 55176, '\P{^Is_Line_Break=h2}', "");
Expect(0, 55177, '\p{Is_Line_Break=h2}', "");
Expect(1, 55177, '\p{^Is_Line_Break=h2}', "");
Expect(1, 55177, '\P{Is_Line_Break=h2}', "");
Expect(0, 55177, '\P{^Is_Line_Break=h2}', "");
Expect(1, 55176, '\p{Is_Line_Break:	-H2}', "");
Expect(0, 55176, '\p{^Is_Line_Break:	-H2}', "");
Expect(0, 55176, '\P{Is_Line_Break:	-H2}', "");
Expect(1, 55176, '\P{^Is_Line_Break:	-H2}', "");
Expect(0, 55177, '\p{Is_Line_Break:	-H2}', "");
Expect(1, 55177, '\p{^Is_Line_Break:	-H2}', "");
Expect(1, 55177, '\P{Is_Line_Break:	-H2}', "");
Expect(0, 55177, '\P{^Is_Line_Break:	-H2}', "");
Error('\p{Is_Lb: __H2:=}');
Error('\P{Is_Lb: __H2:=}');
Expect(1, 55176, '\p{Is_Lb=h2}', "");
Expect(0, 55176, '\p{^Is_Lb=h2}', "");
Expect(0, 55176, '\P{Is_Lb=h2}', "");
Expect(1, 55176, '\P{^Is_Lb=h2}', "");
Expect(0, 55177, '\p{Is_Lb=h2}', "");
Expect(1, 55177, '\p{^Is_Lb=h2}', "");
Expect(1, 55177, '\P{Is_Lb=h2}', "");
Expect(0, 55177, '\P{^Is_Lb=h2}', "");
Expect(1, 55176, '\p{Is_Lb=	-H2}', "");
Expect(0, 55176, '\p{^Is_Lb=	-H2}', "");
Expect(0, 55176, '\P{Is_Lb=	-H2}', "");
Expect(1, 55176, '\P{^Is_Lb=	-H2}', "");
Expect(0, 55177, '\p{Is_Lb=	-H2}', "");
Expect(1, 55177, '\p{^Is_Lb=	-H2}', "");
Expect(1, 55177, '\P{Is_Lb=	-H2}', "");
Expect(0, 55177, '\P{^Is_Lb=	-H2}', "");
Error('\p{Line_Break=/a/__BREAK_Both}');
Error('\P{Line_Break=/a/__BREAK_Both}');
Expect(1, 11835, '\p{Line_Break=breakboth}', "");
Expect(0, 11835, '\p{^Line_Break=breakboth}', "");
Expect(0, 11835, '\P{Line_Break=breakboth}', "");
Expect(1, 11835, '\P{^Line_Break=breakboth}', "");
Expect(0, 11836, '\p{Line_Break=breakboth}', "");
Expect(1, 11836, '\p{^Line_Break=breakboth}', "");
Expect(1, 11836, '\P{Line_Break=breakboth}', "");
Expect(0, 11836, '\P{^Line_Break=breakboth}', "");
Expect(1, 11835, '\p{Line_Break=	Break_BOTH}', "");
Expect(0, 11835, '\p{^Line_Break=	Break_BOTH}', "");
Expect(0, 11835, '\P{Line_Break=	Break_BOTH}', "");
Expect(1, 11835, '\P{^Line_Break=	Break_BOTH}', "");
Expect(0, 11836, '\p{Line_Break=	Break_BOTH}', "");
Expect(1, 11836, '\p{^Line_Break=	Break_BOTH}', "");
Expect(1, 11836, '\P{Line_Break=	Break_BOTH}', "");
Expect(0, 11836, '\P{^Line_Break=	Break_BOTH}', "");
Error('\p{Lb= 	B2/a/}');
Error('\P{Lb= 	B2/a/}');
Expect(1, 11835, '\p{Lb: b2}', "");
Expect(0, 11835, '\p{^Lb: b2}', "");
Expect(0, 11835, '\P{Lb: b2}', "");
Expect(1, 11835, '\P{^Lb: b2}', "");
Expect(0, 11836, '\p{Lb: b2}', "");
Expect(1, 11836, '\p{^Lb: b2}', "");
Expect(1, 11836, '\P{Lb: b2}', "");
Expect(0, 11836, '\P{^Lb: b2}', "");
Expect(1, 11835, '\p{Lb: -B2}', "");
Expect(0, 11835, '\p{^Lb: -B2}', "");
Expect(0, 11835, '\P{Lb: -B2}', "");
Expect(1, 11835, '\P{^Lb: -B2}', "");
Expect(0, 11836, '\p{Lb: -B2}', "");
Expect(1, 11836, '\p{^Lb: -B2}', "");
Expect(1, 11836, '\P{Lb: -B2}', "");
Expect(0, 11836, '\P{^Lb: -B2}', "");
Error('\p{Is_Line_Break:   _Break_BOTH:=}');
Error('\P{Is_Line_Break:   _Break_BOTH:=}');
Expect(1, 11835, '\p{Is_Line_Break=breakboth}', "");
Expect(0, 11835, '\p{^Is_Line_Break=breakboth}', "");
Expect(0, 11835, '\P{Is_Line_Break=breakboth}', "");
Expect(1, 11835, '\P{^Is_Line_Break=breakboth}', "");
Expect(0, 11836, '\p{Is_Line_Break=breakboth}', "");
Expect(1, 11836, '\p{^Is_Line_Break=breakboth}', "");
Expect(1, 11836, '\P{Is_Line_Break=breakboth}', "");
Expect(0, 11836, '\P{^Is_Line_Break=breakboth}', "");
Expect(1, 11835, '\p{Is_Line_Break=_ break_both}', "");
Expect(0, 11835, '\p{^Is_Line_Break=_ break_both}', "");
Expect(0, 11835, '\P{Is_Line_Break=_ break_both}', "");
Expect(1, 11835, '\P{^Is_Line_Break=_ break_both}', "");
Expect(0, 11836, '\p{Is_Line_Break=_ break_both}', "");
Expect(1, 11836, '\p{^Is_Line_Break=_ break_both}', "");
Expect(1, 11836, '\P{Is_Line_Break=_ break_both}', "");
Expect(0, 11836, '\P{^Is_Line_Break=_ break_both}', "");
Error('\p{Is_Lb=/a/B2}');
Error('\P{Is_Lb=/a/B2}');
Expect(1, 11835, '\p{Is_Lb=b2}', "");
Expect(0, 11835, '\p{^Is_Lb=b2}', "");
Expect(0, 11835, '\P{Is_Lb=b2}', "");
Expect(1, 11835, '\P{^Is_Lb=b2}', "");
Expect(0, 11836, '\p{Is_Lb=b2}', "");
Expect(1, 11836, '\p{^Is_Lb=b2}', "");
Expect(1, 11836, '\P{Is_Lb=b2}', "");
Expect(0, 11836, '\P{^Is_Lb=b2}', "");
Expect(1, 11835, '\p{Is_Lb=-	B2}', "");
Expect(0, 11835, '\p{^Is_Lb=-	B2}', "");
Expect(0, 11835, '\P{Is_Lb=-	B2}', "");
Expect(1, 11835, '\P{^Is_Lb=-	B2}', "");
Expect(0, 11836, '\p{Is_Lb=-	B2}', "");
Expect(1, 11836, '\p{^Is_Lb=-	B2}', "");
Expect(1, 11836, '\P{Is_Lb=-	B2}', "");
Expect(0, 11836, '\P{^Is_Lb=-	B2}', "");
Error('\p{Line_Break=:=Alphabetic}');
Error('\P{Line_Break=:=Alphabetic}');
Expect(1, 128883, '\p{Line_Break=alphabetic}', "");
Expect(0, 128883, '\p{^Line_Break=alphabetic}', "");
Expect(0, 128883, '\P{Line_Break=alphabetic}', "");
Expect(1, 128883, '\P{^Line_Break=alphabetic}', "");
Expect(0, 128884, '\p{Line_Break=alphabetic}', "");
Expect(1, 128884, '\p{^Line_Break=alphabetic}', "");
Expect(1, 128884, '\P{Line_Break=alphabetic}', "");
Expect(0, 128884, '\P{^Line_Break=alphabetic}', "");
Expect(1, 128883, '\p{Line_Break=--Alphabetic}', "");
Expect(0, 128883, '\p{^Line_Break=--Alphabetic}', "");
Expect(0, 128883, '\P{Line_Break=--Alphabetic}', "");
Expect(1, 128883, '\P{^Line_Break=--Alphabetic}', "");
Expect(0, 128884, '\p{Line_Break=--Alphabetic}', "");
Expect(1, 128884, '\p{^Line_Break=--Alphabetic}', "");
Expect(1, 128884, '\P{Line_Break=--Alphabetic}', "");
Expect(0, 128884, '\P{^Line_Break=--Alphabetic}', "");
Error('\p{Lb=:=_	al}');
Error('\P{Lb=:=_	al}');
Expect(1, 128883, '\p{Lb: al}', "");
Expect(0, 128883, '\p{^Lb: al}', "");
Expect(0, 128883, '\P{Lb: al}', "");
Expect(1, 128883, '\P{^Lb: al}', "");
Expect(0, 128884, '\p{Lb: al}', "");
Expect(1, 128884, '\p{^Lb: al}', "");
Expect(1, 128884, '\P{Lb: al}', "");
Expect(0, 128884, '\P{^Lb: al}', "");
Expect(1, 128883, '\p{Lb=-AL}', "");
Expect(0, 128883, '\p{^Lb=-AL}', "");
Expect(0, 128883, '\P{Lb=-AL}', "");
Expect(1, 128883, '\P{^Lb=-AL}', "");
Expect(0, 128884, '\p{Lb=-AL}', "");
Expect(1, 128884, '\p{^Lb=-AL}', "");
Expect(1, 128884, '\P{Lb=-AL}', "");
Expect(0, 128884, '\P{^Lb=-AL}', "");
Error('\p{Is_Line_Break=	:=Alphabetic}');
Error('\P{Is_Line_Break=	:=Alphabetic}');
Expect(1, 128883, '\p{Is_Line_Break=alphabetic}', "");
Expect(0, 128883, '\p{^Is_Line_Break=alphabetic}', "");
Expect(0, 128883, '\P{Is_Line_Break=alphabetic}', "");
Expect(1, 128883, '\P{^Is_Line_Break=alphabetic}', "");
Expect(0, 128884, '\p{Is_Line_Break=alphabetic}', "");
Expect(1, 128884, '\p{^Is_Line_Break=alphabetic}', "");
Expect(1, 128884, '\P{Is_Line_Break=alphabetic}', "");
Expect(0, 128884, '\P{^Is_Line_Break=alphabetic}', "");
Expect(1, 128883, '\p{Is_Line_Break= 	alphabetic}', "");
Expect(0, 128883, '\p{^Is_Line_Break= 	alphabetic}', "");
Expect(0, 128883, '\P{Is_Line_Break= 	alphabetic}', "");
Expect(1, 128883, '\P{^Is_Line_Break= 	alphabetic}', "");
Expect(0, 128884, '\p{Is_Line_Break= 	alphabetic}', "");
Expect(1, 128884, '\p{^Is_Line_Break= 	alphabetic}', "");
Expect(1, 128884, '\P{Is_Line_Break= 	alphabetic}', "");
Expect(0, 128884, '\P{^Is_Line_Break= 	alphabetic}', "");
Error('\p{Is_Lb=/a/-	al}');
Error('\P{Is_Lb=/a/-	al}');
Expect(1, 128883, '\p{Is_Lb=al}', "");
Expect(0, 128883, '\p{^Is_Lb=al}', "");
Expect(0, 128883, '\P{Is_Lb=al}', "");
Expect(1, 128883, '\P{^Is_Lb=al}', "");
Expect(0, 128884, '\p{Is_Lb=al}', "");
Expect(1, 128884, '\p{^Is_Lb=al}', "");
Expect(1, 128884, '\P{Is_Lb=al}', "");
Expect(0, 128884, '\P{^Is_Lb=al}', "");
Expect(1, 128883, '\p{Is_Lb=-AL}', "");
Expect(0, 128883, '\p{^Is_Lb=-AL}', "");
Expect(0, 128883, '\P{Is_Lb=-AL}', "");
Expect(1, 128883, '\P{^Is_Lb=-AL}', "");
Expect(0, 128884, '\p{Is_Lb=-AL}', "");
Expect(1, 128884, '\p{^Is_Lb=-AL}', "");
Expect(1, 128884, '\P{Is_Lb=-AL}', "");
Expect(0, 128884, '\P{^Is_Lb=-AL}', "");
Error('\p{Line_Break=/a/_POSTFIX_Numeric}');
Error('\P{Line_Break=/a/_POSTFIX_Numeric}');
Expect(1, 65504, '\p{Line_Break=postfixnumeric}', "");
Expect(0, 65504, '\p{^Line_Break=postfixnumeric}', "");
Expect(0, 65504, '\P{Line_Break=postfixnumeric}', "");
Expect(1, 65504, '\P{^Line_Break=postfixnumeric}', "");
Expect(0, 65505, '\p{Line_Break=postfixnumeric}', "");
Expect(1, 65505, '\p{^Line_Break=postfixnumeric}', "");
Expect(1, 65505, '\P{Line_Break=postfixnumeric}', "");
Expect(0, 65505, '\P{^Line_Break=postfixnumeric}', "");
Expect(1, 65504, '\p{Line_Break:   _ Postfix_Numeric}', "");
Expect(0, 65504, '\p{^Line_Break:   _ Postfix_Numeric}', "");
Expect(0, 65504, '\P{Line_Break:   _ Postfix_Numeric}', "");
Expect(1, 65504, '\P{^Line_Break:   _ Postfix_Numeric}', "");
Expect(0, 65505, '\p{Line_Break:   _ Postfix_Numeric}', "");
Expect(1, 65505, '\p{^Line_Break:   _ Postfix_Numeric}', "");
Expect(1, 65505, '\P{Line_Break:   _ Postfix_Numeric}', "");
Expect(0, 65505, '\P{^Line_Break:   _ Postfix_Numeric}', "");
Error('\p{Lb=	PO:=}');
Error('\P{Lb=	PO:=}');
Expect(1, 65504, '\p{Lb=po}', "");
Expect(0, 65504, '\p{^Lb=po}', "");
Expect(0, 65504, '\P{Lb=po}', "");
Expect(1, 65504, '\P{^Lb=po}', "");
Expect(0, 65505, '\p{Lb=po}', "");
Expect(1, 65505, '\p{^Lb=po}', "");
Expect(1, 65505, '\P{Lb=po}', "");
Expect(0, 65505, '\P{^Lb=po}', "");
Expect(1, 65504, '\p{Lb=_-PO}', "");
Expect(0, 65504, '\p{^Lb=_-PO}', "");
Expect(0, 65504, '\P{Lb=_-PO}', "");
Expect(1, 65504, '\P{^Lb=_-PO}', "");
Expect(0, 65505, '\p{Lb=_-PO}', "");
Expect(1, 65505, '\p{^Lb=_-PO}', "");
Expect(1, 65505, '\P{Lb=_-PO}', "");
Expect(0, 65505, '\P{^Lb=_-PO}', "");
Error('\p{Is_Line_Break=:=  Postfix_NUMERIC}');
Error('\P{Is_Line_Break=:=  Postfix_NUMERIC}');
Expect(1, 65504, '\p{Is_Line_Break=postfixnumeric}', "");
Expect(0, 65504, '\p{^Is_Line_Break=postfixnumeric}', "");
Expect(0, 65504, '\P{Is_Line_Break=postfixnumeric}', "");
Expect(1, 65504, '\P{^Is_Line_Break=postfixnumeric}', "");
Expect(0, 65505, '\p{Is_Line_Break=postfixnumeric}', "");
Expect(1, 65505, '\p{^Is_Line_Break=postfixnumeric}', "");
Expect(1, 65505, '\P{Is_Line_Break=postfixnumeric}', "");
Expect(0, 65505, '\P{^Is_Line_Break=postfixnumeric}', "");
Expect(1, 65504, '\p{Is_Line_Break=- Postfix_NUMERIC}', "");
Expect(0, 65504, '\p{^Is_Line_Break=- Postfix_NUMERIC}', "");
Expect(0, 65504, '\P{Is_Line_Break=- Postfix_NUMERIC}', "");
Expect(1, 65504, '\P{^Is_Line_Break=- Postfix_NUMERIC}', "");
Expect(0, 65505, '\p{Is_Line_Break=- Postfix_NUMERIC}', "");
Expect(1, 65505, '\p{^Is_Line_Break=- Postfix_NUMERIC}', "");
Expect(1, 65505, '\P{Is_Line_Break=- Postfix_NUMERIC}', "");
Expect(0, 65505, '\P{^Is_Line_Break=- Postfix_NUMERIC}', "");
Error('\p{Is_Lb=--PO:=}');
Error('\P{Is_Lb=--PO:=}');
Expect(1, 65504, '\p{Is_Lb=po}', "");
Expect(0, 65504, '\p{^Is_Lb=po}', "");
Expect(0, 65504, '\P{Is_Lb=po}', "");
Expect(1, 65504, '\P{^Is_Lb=po}', "");
Expect(0, 65505, '\p{Is_Lb=po}', "");
Expect(1, 65505, '\p{^Is_Lb=po}', "");
Expect(1, 65505, '\P{Is_Lb=po}', "");
Expect(0, 65505, '\P{^Is_Lb=po}', "");
Expect(1, 65504, '\p{Is_Lb=--po}', "");
Expect(0, 65504, '\p{^Is_Lb=--po}', "");
Expect(0, 65504, '\P{Is_Lb=--po}', "");
Expect(1, 65504, '\P{^Is_Lb=--po}', "");
Expect(0, 65505, '\p{Is_Lb=--po}', "");
Expect(1, 65505, '\p{^Is_Lb=--po}', "");
Expect(1, 65505, '\P{Is_Lb=--po}', "");
Expect(0, 65505, '\P{^Is_Lb=--po}', "");
Error('\p{Line_Break=-:=OPEN_Punctuation}');
Error('\P{Line_Break=-:=OPEN_Punctuation}');
Expect(1, 78713, '\p{Line_Break=openpunctuation}', "");
Expect(0, 78713, '\p{^Line_Break=openpunctuation}', "");
Expect(0, 78713, '\P{Line_Break=openpunctuation}', "");
Expect(1, 78713, '\P{^Line_Break=openpunctuation}', "");
Expect(0, 78714, '\p{Line_Break=openpunctuation}', "");
Expect(1, 78714, '\p{^Line_Break=openpunctuation}', "");
Expect(1, 78714, '\P{Line_Break=openpunctuation}', "");
Expect(0, 78714, '\P{^Line_Break=openpunctuation}', "");
Expect(1, 78713, '\p{Line_Break=	OPEN_Punctuation}', "");
Expect(0, 78713, '\p{^Line_Break=	OPEN_Punctuation}', "");
Expect(0, 78713, '\P{Line_Break=	OPEN_Punctuation}', "");
Expect(1, 78713, '\P{^Line_Break=	OPEN_Punctuation}', "");
Expect(0, 78714, '\p{Line_Break=	OPEN_Punctuation}', "");
Expect(1, 78714, '\p{^Line_Break=	OPEN_Punctuation}', "");
Expect(1, 78714, '\P{Line_Break=	OPEN_Punctuation}', "");
Expect(0, 78714, '\P{^Line_Break=	OPEN_Punctuation}', "");
Error('\p{Lb=/a/ op}');
Error('\P{Lb=/a/ op}');
Expect(1, 78713, '\p{Lb=op}', "");
Expect(0, 78713, '\p{^Lb=op}', "");
Expect(0, 78713, '\P{Lb=op}', "");
Expect(1, 78713, '\P{^Lb=op}', "");
Expect(0, 78714, '\p{Lb=op}', "");
Expect(1, 78714, '\p{^Lb=op}', "");
Expect(1, 78714, '\P{Lb=op}', "");
Expect(0, 78714, '\P{^Lb=op}', "");
Expect(1, 78713, '\p{Lb=_OP}', "");
Expect(0, 78713, '\p{^Lb=_OP}', "");
Expect(0, 78713, '\P{Lb=_OP}', "");
Expect(1, 78713, '\P{^Lb=_OP}', "");
Expect(0, 78714, '\p{Lb=_OP}', "");
Expect(1, 78714, '\p{^Lb=_OP}', "");
Expect(1, 78714, '\P{Lb=_OP}', "");
Expect(0, 78714, '\P{^Lb=_OP}', "");
Error('\p{Is_Line_Break=open_PUNCTUATION:=}');
Error('\P{Is_Line_Break=open_PUNCTUATION:=}');
Expect(1, 78713, '\p{Is_Line_Break=openpunctuation}', "");
Expect(0, 78713, '\p{^Is_Line_Break=openpunctuation}', "");
Expect(0, 78713, '\P{Is_Line_Break=openpunctuation}', "");
Expect(1, 78713, '\P{^Is_Line_Break=openpunctuation}', "");
Expect(0, 78714, '\p{Is_Line_Break=openpunctuation}', "");
Expect(1, 78714, '\p{^Is_Line_Break=openpunctuation}', "");
Expect(1, 78714, '\P{Is_Line_Break=openpunctuation}', "");
Expect(0, 78714, '\P{^Is_Line_Break=openpunctuation}', "");
Expect(1, 78713, '\p{Is_Line_Break=	 Open_punctuation}', "");
Expect(0, 78713, '\p{^Is_Line_Break=	 Open_punctuation}', "");
Expect(0, 78713, '\P{Is_Line_Break=	 Open_punctuation}', "");
Expect(1, 78713, '\P{^Is_Line_Break=	 Open_punctuation}', "");
Expect(0, 78714, '\p{Is_Line_Break=	 Open_punctuation}', "");
Expect(1, 78714, '\p{^Is_Line_Break=	 Open_punctuation}', "");
Expect(1, 78714, '\P{Is_Line_Break=	 Open_punctuation}', "");
Expect(0, 78714, '\P{^Is_Line_Break=	 Open_punctuation}', "");
Error('\p{Is_Lb=/a/op}');
Error('\P{Is_Lb=/a/op}');
Expect(1, 78713, '\p{Is_Lb=op}', "");
Expect(0, 78713, '\p{^Is_Lb=op}', "");
Expect(0, 78713, '\P{Is_Lb=op}', "");
Expect(1, 78713, '\P{^Is_Lb=op}', "");
Expect(0, 78714, '\p{Is_Lb=op}', "");
Expect(1, 78714, '\p{^Is_Lb=op}', "");
Expect(1, 78714, '\P{Is_Lb=op}', "");
Expect(0, 78714, '\P{^Is_Lb=op}', "");
Expect(1, 78713, '\p{Is_Lb=_op}', "");
Expect(0, 78713, '\p{^Is_Lb=_op}', "");
Expect(0, 78713, '\P{Is_Lb=_op}', "");
Expect(1, 78713, '\P{^Is_Lb=_op}', "");
Expect(0, 78714, '\p{Is_Lb=_op}', "");
Expect(1, 78714, '\p{^Is_Lb=_op}', "");
Expect(1, 78714, '\P{Is_Lb=_op}', "");
Expect(0, 78714, '\P{^Is_Lb=_op}', "");
Error('\p{Line_Break= NEXT_Line/a/}');
Error('\P{Line_Break= NEXT_Line/a/}');
Expect(1, 133, '\p{Line_Break=nextline}', "");
Expect(0, 133, '\p{^Line_Break=nextline}', "");
Expect(0, 133, '\P{Line_Break=nextline}', "");
Expect(1, 133, '\P{^Line_Break=nextline}', "");
Expect(0, 134, '\p{Line_Break=nextline}', "");
Expect(1, 134, '\p{^Line_Break=nextline}', "");
Expect(1, 134, '\P{Line_Break=nextline}', "");
Expect(0, 134, '\P{^Line_Break=nextline}', "");
Expect(1, 133, '\p{Line_Break=	NEXT_Line}', "");
Expect(0, 133, '\p{^Line_Break=	NEXT_Line}', "");
Expect(0, 133, '\P{Line_Break=	NEXT_Line}', "");
Expect(1, 133, '\P{^Line_Break=	NEXT_Line}', "");
Expect(0, 134, '\p{Line_Break=	NEXT_Line}', "");
Expect(1, 134, '\p{^Line_Break=	NEXT_Line}', "");
Expect(1, 134, '\P{Line_Break=	NEXT_Line}', "");
Expect(0, 134, '\P{^Line_Break=	NEXT_Line}', "");
Error('\p{Lb=-NL:=}');
Error('\P{Lb=-NL:=}');
Expect(1, 133, '\p{Lb=nl}', "");
Expect(0, 133, '\p{^Lb=nl}', "");
Expect(0, 133, '\P{Lb=nl}', "");
Expect(1, 133, '\P{^Lb=nl}', "");
Expect(0, 134, '\p{Lb=nl}', "");
Expect(1, 134, '\p{^Lb=nl}', "");
Expect(1, 134, '\P{Lb=nl}', "");
Expect(0, 134, '\P{^Lb=nl}', "");
Expect(1, 133, '\p{Lb=-NL}', "");
Expect(0, 133, '\p{^Lb=-NL}', "");
Expect(0, 133, '\P{Lb=-NL}', "");
Expect(1, 133, '\P{^Lb=-NL}', "");
Expect(0, 134, '\p{Lb=-NL}', "");
Expect(1, 134, '\p{^Lb=-NL}', "");
Expect(1, 134, '\P{Lb=-NL}', "");
Expect(0, 134, '\P{^Lb=-NL}', "");
Error('\p{Is_Line_Break= -Next_line/a/}');
Error('\P{Is_Line_Break= -Next_line/a/}');
Expect(1, 133, '\p{Is_Line_Break=nextline}', "");
Expect(0, 133, '\p{^Is_Line_Break=nextline}', "");
Expect(0, 133, '\P{Is_Line_Break=nextline}', "");
Expect(1, 133, '\P{^Is_Line_Break=nextline}', "");
Expect(0, 134, '\p{Is_Line_Break=nextline}', "");
Expect(1, 134, '\p{^Is_Line_Break=nextline}', "");
Expect(1, 134, '\P{Is_Line_Break=nextline}', "");
Expect(0, 134, '\P{^Is_Line_Break=nextline}', "");
Expect(1, 133, '\p{Is_Line_Break=_	next_line}', "");
Expect(0, 133, '\p{^Is_Line_Break=_	next_line}', "");
Expect(0, 133, '\P{Is_Line_Break=_	next_line}', "");
Expect(1, 133, '\P{^Is_Line_Break=_	next_line}', "");
Expect(0, 134, '\p{Is_Line_Break=_	next_line}', "");
Expect(1, 134, '\p{^Is_Line_Break=_	next_line}', "");
Expect(1, 134, '\P{Is_Line_Break=_	next_line}', "");
Expect(0, 134, '\P{^Is_Line_Break=_	next_line}', "");
Error('\p{Is_Lb=_ NL/a/}');
Error('\P{Is_Lb=_ NL/a/}');
Expect(1, 133, '\p{Is_Lb=nl}', "");
Expect(0, 133, '\p{^Is_Lb=nl}', "");
Expect(0, 133, '\P{Is_Lb=nl}', "");
Expect(1, 133, '\P{^Is_Lb=nl}', "");
Expect(0, 134, '\p{Is_Lb=nl}', "");
Expect(1, 134, '\p{^Is_Lb=nl}', "");
Expect(1, 134, '\P{Is_Lb=nl}', "");
Expect(0, 134, '\P{^Is_Lb=nl}', "");
Expect(1, 133, '\p{Is_Lb= 	NL}', "");
Expect(0, 133, '\p{^Is_Lb= 	NL}', "");
Expect(0, 133, '\P{Is_Lb= 	NL}', "");
Expect(1, 133, '\P{^Is_Lb= 	NL}', "");
Expect(0, 134, '\p{Is_Lb= 	NL}', "");
Expect(1, 134, '\p{^Is_Lb= 	NL}', "");
Expect(1, 134, '\P{Is_Lb= 	NL}', "");
Expect(0, 134, '\P{^Is_Lb= 	NL}', "");
Error('\p{Line_Break= :=JT}');
Error('\P{Line_Break= :=JT}');
Expect(1, 55291, '\p{Line_Break=jt}', "");
Expect(0, 55291, '\p{^Line_Break=jt}', "");
Expect(0, 55291, '\P{Line_Break=jt}', "");
Expect(1, 55291, '\P{^Line_Break=jt}', "");
Expect(0, 55292, '\p{Line_Break=jt}', "");
Expect(1, 55292, '\p{^Line_Break=jt}', "");
Expect(1, 55292, '\P{Line_Break=jt}', "");
Expect(0, 55292, '\P{^Line_Break=jt}', "");
Expect(1, 55291, '\p{Line_Break=		jt}', "");
Expect(0, 55291, '\p{^Line_Break=		jt}', "");
Expect(0, 55291, '\P{Line_Break=		jt}', "");
Expect(1, 55291, '\P{^Line_Break=		jt}', "");
Expect(0, 55292, '\p{Line_Break=		jt}', "");
Expect(1, 55292, '\p{^Line_Break=		jt}', "");
Expect(1, 55292, '\P{Line_Break=		jt}', "");
Expect(0, 55292, '\P{^Line_Break=		jt}', "");
Error('\p{Lb:		/a/JT}');
Error('\P{Lb:		/a/JT}');
Expect(1, 55291, '\p{Lb=jt}', "");
Expect(0, 55291, '\p{^Lb=jt}', "");
Expect(0, 55291, '\P{Lb=jt}', "");
Expect(1, 55291, '\P{^Lb=jt}', "");
Expect(0, 55292, '\p{Lb=jt}', "");
Expect(1, 55292, '\p{^Lb=jt}', "");
Expect(1, 55292, '\P{Lb=jt}', "");
Expect(0, 55292, '\P{^Lb=jt}', "");
Expect(1, 55291, '\p{Lb=	 JT}', "");
Expect(0, 55291, '\p{^Lb=	 JT}', "");
Expect(0, 55291, '\P{Lb=	 JT}', "");
Expect(1, 55291, '\P{^Lb=	 JT}', "");
Expect(0, 55292, '\p{Lb=	 JT}', "");
Expect(1, 55292, '\p{^Lb=	 JT}', "");
Expect(1, 55292, '\P{Lb=	 JT}', "");
Expect(0, 55292, '\P{^Lb=	 JT}', "");
Error('\p{Is_Line_Break:   := -JT}');
Error('\P{Is_Line_Break:   := -JT}');
Expect(1, 55291, '\p{Is_Line_Break:   jt}', "");
Expect(0, 55291, '\p{^Is_Line_Break:   jt}', "");
Expect(0, 55291, '\P{Is_Line_Break:   jt}', "");
Expect(1, 55291, '\P{^Is_Line_Break:   jt}', "");
Expect(0, 55292, '\p{Is_Line_Break:   jt}', "");
Expect(1, 55292, '\p{^Is_Line_Break:   jt}', "");
Expect(1, 55292, '\P{Is_Line_Break:   jt}', "");
Expect(0, 55292, '\P{^Is_Line_Break:   jt}', "");
Expect(1, 55291, '\p{Is_Line_Break= -JT}', "");
Expect(0, 55291, '\p{^Is_Line_Break= -JT}', "");
Expect(0, 55291, '\P{Is_Line_Break= -JT}', "");
Expect(1, 55291, '\P{^Is_Line_Break= -JT}', "");
Expect(0, 55292, '\p{Is_Line_Break= -JT}', "");
Expect(1, 55292, '\p{^Is_Line_Break= -JT}', "");
Expect(1, 55292, '\P{Is_Line_Break= -JT}', "");
Expect(0, 55292, '\P{^Is_Line_Break= -JT}', "");
Error('\p{Is_Lb= _JT:=}');
Error('\P{Is_Lb= _JT:=}');
Expect(1, 55291, '\p{Is_Lb=jt}', "");
Expect(0, 55291, '\p{^Is_Lb=jt}', "");
Expect(0, 55291, '\P{Is_Lb=jt}', "");
Expect(1, 55291, '\P{^Is_Lb=jt}', "");
Expect(0, 55292, '\p{Is_Lb=jt}', "");
Expect(1, 55292, '\p{^Is_Lb=jt}', "");
Expect(1, 55292, '\P{Is_Lb=jt}', "");
Expect(0, 55292, '\P{^Is_Lb=jt}', "");
Expect(1, 55291, '\p{Is_Lb=-jt}', "");
Expect(0, 55291, '\p{^Is_Lb=-jt}', "");
Expect(0, 55291, '\P{Is_Lb=-jt}', "");
Expect(1, 55291, '\P{^Is_Lb=-jt}', "");
Expect(0, 55292, '\p{Is_Lb=-jt}', "");
Expect(1, 55292, '\p{^Is_Lb=-jt}', "");
Expect(1, 55292, '\P{Is_Lb=-jt}', "");
Expect(0, 55292, '\P{^Is_Lb=-jt}', "");
Error('\p{Terminal_Punctuation=/a/- Yes}');
Error('\P{Terminal_Punctuation=/a/- Yes}');
Expect(1, 74867, '\p{Terminal_Punctuation=yes}', "");
Expect(0, 74867, '\p{^Terminal_Punctuation=yes}', "");
Expect(0, 74867, '\P{Terminal_Punctuation=yes}', "");
Expect(1, 74867, '\P{^Terminal_Punctuation=yes}', "");
Expect(0, 74868, '\p{Terminal_Punctuation=yes}', "");
Expect(1, 74868, '\p{^Terminal_Punctuation=yes}', "");
Expect(1, 74868, '\P{Terminal_Punctuation=yes}', "");
Expect(0, 74868, '\P{^Terminal_Punctuation=yes}', "");
Expect(1, 74867, '\p{Terminal_Punctuation=	-Yes}', "");
Expect(0, 74867, '\p{^Terminal_Punctuation=	-Yes}', "");
Expect(0, 74867, '\P{Terminal_Punctuation=	-Yes}', "");
Expect(1, 74867, '\P{^Terminal_Punctuation=	-Yes}', "");
Expect(0, 74868, '\p{Terminal_Punctuation=	-Yes}', "");
Expect(1, 74868, '\p{^Terminal_Punctuation=	-Yes}', "");
Expect(1, 74868, '\P{Terminal_Punctuation=	-Yes}', "");
Expect(0, 74868, '\P{^Terminal_Punctuation=	-Yes}', "");
Error('\p{Term=	Y:=}');
Error('\P{Term=	Y:=}');
Expect(1, 74867, '\p{Term=y}', "");
Expect(0, 74867, '\p{^Term=y}', "");
Expect(0, 74867, '\P{Term=y}', "");
Expect(1, 74867, '\P{^Term=y}', "");
Expect(0, 74868, '\p{Term=y}', "");
Expect(1, 74868, '\p{^Term=y}', "");
Expect(1, 74868, '\P{Term=y}', "");
Expect(0, 74868, '\P{^Term=y}', "");
Expect(1, 74867, '\p{Term=__Y}', "");
Expect(0, 74867, '\p{^Term=__Y}', "");
Expect(0, 74867, '\P{Term=__Y}', "");
Expect(1, 74867, '\P{^Term=__Y}', "");
Expect(0, 74868, '\p{Term=__Y}', "");
Expect(1, 74868, '\p{^Term=__Y}', "");
Expect(1, 74868, '\P{Term=__Y}', "");
Expect(0, 74868, '\P{^Term=__Y}', "");
Error('\p{Is_Terminal_Punctuation=	-t/a/}');
Error('\P{Is_Terminal_Punctuation=	-t/a/}');
Expect(1, 74867, '\p{Is_Terminal_Punctuation=t}', "");
Expect(0, 74867, '\p{^Is_Terminal_Punctuation=t}', "");
Expect(0, 74867, '\P{Is_Terminal_Punctuation=t}', "");
Expect(1, 74867, '\P{^Is_Terminal_Punctuation=t}', "");
Expect(0, 74868, '\p{Is_Terminal_Punctuation=t}', "");
Expect(1, 74868, '\p{^Is_Terminal_Punctuation=t}', "");
Expect(1, 74868, '\P{Is_Terminal_Punctuation=t}', "");
Expect(0, 74868, '\P{^Is_Terminal_Punctuation=t}', "");
Expect(1, 74867, '\p{Is_Terminal_Punctuation=_t}', "");
Expect(0, 74867, '\p{^Is_Terminal_Punctuation=_t}', "");
Expect(0, 74867, '\P{Is_Terminal_Punctuation=_t}', "");
Expect(1, 74867, '\P{^Is_Terminal_Punctuation=_t}', "");
Expect(0, 74868, '\p{Is_Terminal_Punctuation=_t}', "");
Expect(1, 74868, '\p{^Is_Terminal_Punctuation=_t}', "");
Expect(1, 74868, '\P{Is_Terminal_Punctuation=_t}', "");
Expect(0, 74868, '\P{^Is_Terminal_Punctuation=_t}', "");
Error('\p{Is_Term=	_true:=}');
Error('\P{Is_Term=	_true:=}');
Expect(1, 74867, '\p{Is_Term=true}', "");
Expect(0, 74867, '\p{^Is_Term=true}', "");
Expect(0, 74867, '\P{Is_Term=true}', "");
Expect(1, 74867, '\P{^Is_Term=true}', "");
Expect(0, 74868, '\p{Is_Term=true}', "");
Expect(1, 74868, '\p{^Is_Term=true}', "");
Expect(1, 74868, '\P{Is_Term=true}', "");
Expect(0, 74868, '\P{^Is_Term=true}', "");
Expect(1, 74867, '\p{Is_Term:_	true}', "");
Expect(0, 74867, '\p{^Is_Term:_	true}', "");
Expect(0, 74867, '\P{Is_Term:_	true}', "");
Expect(1, 74867, '\P{^Is_Term:_	true}', "");
Expect(0, 74868, '\p{Is_Term:_	true}', "");
Expect(1, 74868, '\p{^Is_Term:_	true}', "");
Expect(1, 74868, '\P{Is_Term:_	true}', "");
Expect(0, 74868, '\P{^Is_Term:_	true}', "");
Error('\p{Terminal_Punctuation:   :=	No}');
Error('\P{Terminal_Punctuation:   :=	No}');
Expect(1, 74868, '\p{Terminal_Punctuation=no}', "");
Expect(0, 74868, '\p{^Terminal_Punctuation=no}', "");
Expect(0, 74868, '\P{Terminal_Punctuation=no}', "");
Expect(1, 74868, '\P{^Terminal_Punctuation=no}', "");
Expect(0, 74867, '\p{Terminal_Punctuation=no}', "");
Expect(1, 74867, '\p{^Terminal_Punctuation=no}', "");
Expect(1, 74867, '\P{Terminal_Punctuation=no}', "");
Expect(0, 74867, '\P{^Terminal_Punctuation=no}', "");
Expect(1, 74868, '\p{Terminal_Punctuation=	No}', "");
Expect(0, 74868, '\p{^Terminal_Punctuation=	No}', "");
Expect(0, 74868, '\P{Terminal_Punctuation=	No}', "");
Expect(1, 74868, '\P{^Terminal_Punctuation=	No}', "");
Expect(0, 74867, '\p{Terminal_Punctuation=	No}', "");
Expect(1, 74867, '\p{^Terminal_Punctuation=	No}', "");
Expect(1, 74867, '\P{Terminal_Punctuation=	No}', "");
Expect(0, 74867, '\P{^Terminal_Punctuation=	No}', "");
Error('\p{Term=		N:=}');
Error('\P{Term=		N:=}');
Expect(1, 74868, '\p{Term:n}', "");
Expect(0, 74868, '\p{^Term:n}', "");
Expect(0, 74868, '\P{Term:n}', "");
Expect(1, 74868, '\P{^Term:n}', "");
Expect(0, 74867, '\p{Term:n}', "");
Expect(1, 74867, '\p{^Term:n}', "");
Expect(1, 74867, '\P{Term:n}', "");
Expect(0, 74867, '\P{^Term:n}', "");
Expect(1, 74868, '\p{Term=	n}', "");
Expect(0, 74868, '\p{^Term=	n}', "");
Expect(0, 74868, '\P{Term=	n}', "");
Expect(1, 74868, '\P{^Term=	n}', "");
Expect(0, 74867, '\p{Term=	n}', "");
Expect(1, 74867, '\p{^Term=	n}', "");
Expect(1, 74867, '\P{Term=	n}', "");
Expect(0, 74867, '\P{^Term=	n}', "");
Error('\p{Is_Terminal_Punctuation: /a/F}');
Error('\P{Is_Terminal_Punctuation: /a/F}');
Expect(1, 74868, '\p{Is_Terminal_Punctuation=f}', "");
Expect(0, 74868, '\p{^Is_Terminal_Punctuation=f}', "");
Expect(0, 74868, '\P{Is_Terminal_Punctuation=f}', "");
Expect(1, 74868, '\P{^Is_Terminal_Punctuation=f}', "");
Expect(0, 74867, '\p{Is_Terminal_Punctuation=f}', "");
Expect(1, 74867, '\p{^Is_Terminal_Punctuation=f}', "");
Expect(1, 74867, '\P{Is_Terminal_Punctuation=f}', "");
Expect(0, 74867, '\P{^Is_Terminal_Punctuation=f}', "");
Expect(1, 74868, '\p{Is_Terminal_Punctuation=	f}', "");
Expect(0, 74868, '\p{^Is_Terminal_Punctuation=	f}', "");
Expect(0, 74868, '\P{Is_Terminal_Punctuation=	f}', "");
Expect(1, 74868, '\P{^Is_Terminal_Punctuation=	f}', "");
Expect(0, 74867, '\p{Is_Terminal_Punctuation=	f}', "");
Expect(1, 74867, '\p{^Is_Terminal_Punctuation=	f}', "");
Expect(1, 74867, '\P{Is_Terminal_Punctuation=	f}', "");
Expect(0, 74867, '\P{^Is_Terminal_Punctuation=	f}', "");
Error('\p{Is_Term=/a/ false}');
Error('\P{Is_Term=/a/ false}');
Expect(1, 74868, '\p{Is_Term=false}', "");
Expect(0, 74868, '\p{^Is_Term=false}', "");
Expect(0, 74868, '\P{Is_Term=false}', "");
Expect(1, 74868, '\P{^Is_Term=false}', "");
Expect(0, 74867, '\p{Is_Term=false}', "");
Expect(1, 74867, '\p{^Is_Term=false}', "");
Expect(1, 74867, '\P{Is_Term=false}', "");
Expect(0, 74867, '\P{^Is_Term=false}', "");
Expect(1, 74868, '\p{Is_Term=_FALSE}', "");
Expect(0, 74868, '\p{^Is_Term=_FALSE}', "");
Expect(0, 74868, '\P{Is_Term=_FALSE}', "");
Expect(1, 74868, '\P{^Is_Term=_FALSE}', "");
Expect(0, 74867, '\p{Is_Term=_FALSE}', "");
Expect(1, 74867, '\p{^Is_Term=_FALSE}', "");
Expect(1, 74867, '\P{Is_Term=_FALSE}', "");
Expect(0, 74867, '\P{^Is_Term=_FALSE}', "");
Error('\p{Expands_On_NFKD=Yes}');
Error('\P{Expands_On_NFKD=Yes}');
Error('\p{XO_NFKD=Y}');
Error('\P{XO_NFKD=Y}');
Error('\p{Is_Expands_On_NFKD=T}');
Error('\P{Is_Expands_On_NFKD=T}');
Error('\p{Is_XO_NFKD=True}');
Error('\P{Is_XO_NFKD=True}');
Error('\p{Expands_On_NFKD=No}');
Error('\P{Expands_On_NFKD=No}');
Error('\p{XO_NFKD=N}');
Error('\P{XO_NFKD=N}');
Error('\p{Is_Expands_On_NFKD=F}');
Error('\P{Is_Expands_On_NFKD=F}');
Error('\p{Is_XO_NFKD: False}');
Error('\P{Is_XO_NFKD: False}');
Error('\p{Pattern_Syntax=:=_ YES}');
Error('\P{Pattern_Syntax=:=_ YES}');
Expect(1, 65094, '\p{Pattern_Syntax=yes}', "");
Expect(0, 65094, '\p{^Pattern_Syntax=yes}', "");
Expect(0, 65094, '\P{Pattern_Syntax=yes}', "");
Expect(1, 65094, '\P{^Pattern_Syntax=yes}', "");
Expect(0, 65095, '\p{Pattern_Syntax=yes}', "");
Expect(1, 65095, '\p{^Pattern_Syntax=yes}', "");
Expect(1, 65095, '\P{Pattern_Syntax=yes}', "");
Expect(0, 65095, '\P{^Pattern_Syntax=yes}', "");
Expect(1, 65094, '\p{Pattern_Syntax=_-Yes}', "");
Expect(0, 65094, '\p{^Pattern_Syntax=_-Yes}', "");
Expect(0, 65094, '\P{Pattern_Syntax=_-Yes}', "");
Expect(1, 65094, '\P{^Pattern_Syntax=_-Yes}', "");
Expect(0, 65095, '\p{Pattern_Syntax=_-Yes}', "");
Expect(1, 65095, '\p{^Pattern_Syntax=_-Yes}', "");
Expect(1, 65095, '\P{Pattern_Syntax=_-Yes}', "");
Expect(0, 65095, '\P{^Pattern_Syntax=_-Yes}', "");
Error('\p{Pat_Syn=_	Y/a/}');
Error('\P{Pat_Syn=_	Y/a/}');
Expect(1, 65094, '\p{Pat_Syn=y}', "");
Expect(0, 65094, '\p{^Pat_Syn=y}', "");
Expect(0, 65094, '\P{Pat_Syn=y}', "");
Expect(1, 65094, '\P{^Pat_Syn=y}', "");
Expect(0, 65095, '\p{Pat_Syn=y}', "");
Expect(1, 65095, '\p{^Pat_Syn=y}', "");
Expect(1, 65095, '\P{Pat_Syn=y}', "");
Expect(0, 65095, '\P{^Pat_Syn=y}', "");
Expect(1, 65094, '\p{Pat_Syn=-_Y}', "");
Expect(0, 65094, '\p{^Pat_Syn=-_Y}', "");
Expect(0, 65094, '\P{Pat_Syn=-_Y}', "");
Expect(1, 65094, '\P{^Pat_Syn=-_Y}', "");
Expect(0, 65095, '\p{Pat_Syn=-_Y}', "");
Expect(1, 65095, '\p{^Pat_Syn=-_Y}', "");
Expect(1, 65095, '\P{Pat_Syn=-_Y}', "");
Expect(0, 65095, '\P{^Pat_Syn=-_Y}', "");
Error('\p{Is_Pattern_Syntax=	_t:=}');
Error('\P{Is_Pattern_Syntax=	_t:=}');
Expect(1, 65094, '\p{Is_Pattern_Syntax:	t}', "");
Expect(0, 65094, '\p{^Is_Pattern_Syntax:	t}', "");
Expect(0, 65094, '\P{Is_Pattern_Syntax:	t}', "");
Expect(1, 65094, '\P{^Is_Pattern_Syntax:	t}', "");
Expect(0, 65095, '\p{Is_Pattern_Syntax:	t}', "");
Expect(1, 65095, '\p{^Is_Pattern_Syntax:	t}', "");
Expect(1, 65095, '\P{Is_Pattern_Syntax:	t}', "");
Expect(0, 65095, '\P{^Is_Pattern_Syntax:	t}', "");
Expect(1, 65094, '\p{Is_Pattern_Syntax=  T}', "");
Expect(0, 65094, '\p{^Is_Pattern_Syntax=  T}', "");
Expect(0, 65094, '\P{Is_Pattern_Syntax=  T}', "");
Expect(1, 65094, '\P{^Is_Pattern_Syntax=  T}', "");
Expect(0, 65095, '\p{Is_Pattern_Syntax=  T}', "");
Expect(1, 65095, '\p{^Is_Pattern_Syntax=  T}', "");
Expect(1, 65095, '\P{Is_Pattern_Syntax=  T}', "");
Expect(0, 65095, '\P{^Is_Pattern_Syntax=  T}', "");
Error('\p{Is_Pat_Syn=/a/	True}');
Error('\P{Is_Pat_Syn=/a/	True}');
Expect(1, 65094, '\p{Is_Pat_Syn=true}', "");
Expect(0, 65094, '\p{^Is_Pat_Syn=true}', "");
Expect(0, 65094, '\P{Is_Pat_Syn=true}', "");
Expect(1, 65094, '\P{^Is_Pat_Syn=true}', "");
Expect(0, 65095, '\p{Is_Pat_Syn=true}', "");
Expect(1, 65095, '\p{^Is_Pat_Syn=true}', "");
Expect(1, 65095, '\P{Is_Pat_Syn=true}', "");
Expect(0, 65095, '\P{^Is_Pat_Syn=true}', "");
Expect(1, 65094, '\p{Is_Pat_Syn=- True}', "");
Expect(0, 65094, '\p{^Is_Pat_Syn=- True}', "");
Expect(0, 65094, '\P{Is_Pat_Syn=- True}', "");
Expect(1, 65094, '\P{^Is_Pat_Syn=- True}', "");
Expect(0, 65095, '\p{Is_Pat_Syn=- True}', "");
Expect(1, 65095, '\p{^Is_Pat_Syn=- True}', "");
Expect(1, 65095, '\P{Is_Pat_Syn=- True}', "");
Expect(0, 65095, '\P{^Is_Pat_Syn=- True}', "");
Error('\p{Pattern_Syntax=		No/a/}');
Error('\P{Pattern_Syntax=		No/a/}');
Expect(1, 65095, '\p{Pattern_Syntax:no}', "");
Expect(0, 65095, '\p{^Pattern_Syntax:no}', "");
Expect(0, 65095, '\P{Pattern_Syntax:no}', "");
Expect(1, 65095, '\P{^Pattern_Syntax:no}', "");
Expect(0, 65094, '\p{Pattern_Syntax:no}', "");
Expect(1, 65094, '\p{^Pattern_Syntax:no}', "");
Expect(1, 65094, '\P{Pattern_Syntax:no}', "");
Expect(0, 65094, '\P{^Pattern_Syntax:no}', "");
Expect(1, 65095, '\p{Pattern_Syntax=	no}', "");
Expect(0, 65095, '\p{^Pattern_Syntax=	no}', "");
Expect(0, 65095, '\P{Pattern_Syntax=	no}', "");
Expect(1, 65095, '\P{^Pattern_Syntax=	no}', "");
Expect(0, 65094, '\p{Pattern_Syntax=	no}', "");
Expect(1, 65094, '\p{^Pattern_Syntax=	no}', "");
Expect(1, 65094, '\P{Pattern_Syntax=	no}', "");
Expect(0, 65094, '\P{^Pattern_Syntax=	no}', "");
Error('\p{Pat_Syn=_	n:=}');
Error('\P{Pat_Syn=_	n:=}');
Expect(1, 65095, '\p{Pat_Syn:n}', "");
Expect(0, 65095, '\p{^Pat_Syn:n}', "");
Expect(0, 65095, '\P{Pat_Syn:n}', "");
Expect(1, 65095, '\P{^Pat_Syn:n}', "");
Expect(0, 65094, '\p{Pat_Syn:n}', "");
Expect(1, 65094, '\p{^Pat_Syn:n}', "");
Expect(1, 65094, '\P{Pat_Syn:n}', "");
Expect(0, 65094, '\P{^Pat_Syn:n}', "");
Expect(1, 65095, '\p{Pat_Syn= -N}', "");
Expect(0, 65095, '\p{^Pat_Syn= -N}', "");
Expect(0, 65095, '\P{Pat_Syn= -N}', "");
Expect(1, 65095, '\P{^Pat_Syn= -N}', "");
Expect(0, 65094, '\p{Pat_Syn= -N}', "");
Expect(1, 65094, '\p{^Pat_Syn= -N}', "");
Expect(1, 65094, '\P{Pat_Syn= -N}', "");
Expect(0, 65094, '\P{^Pat_Syn= -N}', "");
Error('\p{Is_Pattern_Syntax=/a/  F}');
Error('\P{Is_Pattern_Syntax=/a/  F}');
Expect(1, 65095, '\p{Is_Pattern_Syntax=f}', "");
Expect(0, 65095, '\p{^Is_Pattern_Syntax=f}', "");
Expect(0, 65095, '\P{Is_Pattern_Syntax=f}', "");
Expect(1, 65095, '\P{^Is_Pattern_Syntax=f}', "");
Expect(0, 65094, '\p{Is_Pattern_Syntax=f}', "");
Expect(1, 65094, '\p{^Is_Pattern_Syntax=f}', "");
Expect(1, 65094, '\P{Is_Pattern_Syntax=f}', "");
Expect(0, 65094, '\P{^Is_Pattern_Syntax=f}', "");
Expect(1, 65095, '\p{Is_Pattern_Syntax=  F}', "");
Expect(0, 65095, '\p{^Is_Pattern_Syntax=  F}', "");
Expect(0, 65095, '\P{Is_Pattern_Syntax=  F}', "");
Expect(1, 65095, '\P{^Is_Pattern_Syntax=  F}', "");
Expect(0, 65094, '\p{Is_Pattern_Syntax=  F}', "");
Expect(1, 65094, '\p{^Is_Pattern_Syntax=  F}', "");
Expect(1, 65094, '\P{Is_Pattern_Syntax=  F}', "");
Expect(0, 65094, '\P{^Is_Pattern_Syntax=  F}', "");
Error('\p{Is_Pat_Syn=:=	-false}');
Error('\P{Is_Pat_Syn=:=	-false}');
Expect(1, 65095, '\p{Is_Pat_Syn=false}', "");
Expect(0, 65095, '\p{^Is_Pat_Syn=false}', "");
Expect(0, 65095, '\P{Is_Pat_Syn=false}', "");
Expect(1, 65095, '\P{^Is_Pat_Syn=false}', "");
Expect(0, 65094, '\p{Is_Pat_Syn=false}', "");
Expect(1, 65094, '\p{^Is_Pat_Syn=false}', "");
Expect(1, 65094, '\P{Is_Pat_Syn=false}', "");
Expect(0, 65094, '\P{^Is_Pat_Syn=false}', "");
Expect(1, 65095, '\p{Is_Pat_Syn=-	False}', "");
Expect(0, 65095, '\p{^Is_Pat_Syn=-	False}', "");
Expect(0, 65095, '\P{Is_Pat_Syn=-	False}', "");
Expect(1, 65095, '\P{^Is_Pat_Syn=-	False}', "");
Expect(0, 65094, '\p{Is_Pat_Syn=-	False}', "");
Expect(1, 65094, '\p{^Is_Pat_Syn=-	False}', "");
Expect(1, 65094, '\P{Is_Pat_Syn=-	False}', "");
Expect(0, 65094, '\P{^Is_Pat_Syn=-	False}', "");
Error('\p{ID_Start=_/a/Yes}');
Error('\P{ID_Start=_/a/Yes}');
Expect(1, 195101, '\p{ID_Start=yes}', "");
Expect(0, 195101, '\p{^ID_Start=yes}', "");
Expect(0, 195101, '\P{ID_Start=yes}', "");
Expect(1, 195101, '\P{^ID_Start=yes}', "");
Expect(0, 195102, '\p{ID_Start=yes}', "");
Expect(1, 195102, '\p{^ID_Start=yes}', "");
Expect(1, 195102, '\P{ID_Start=yes}', "");
Expect(0, 195102, '\P{^ID_Start=yes}', "");
Expect(1, 195101, '\p{ID_Start:-_YES}', "");
Expect(0, 195101, '\p{^ID_Start:-_YES}', "");
Expect(0, 195101, '\P{ID_Start:-_YES}', "");
Expect(1, 195101, '\P{^ID_Start:-_YES}', "");
Expect(0, 195102, '\p{ID_Start:-_YES}', "");
Expect(1, 195102, '\p{^ID_Start:-_YES}', "");
Expect(1, 195102, '\P{ID_Start:-_YES}', "");
Expect(0, 195102, '\P{^ID_Start:-_YES}', "");
Error('\p{IDS=_-Y/a/}');
Error('\P{IDS=_-Y/a/}');
Expect(1, 195101, '\p{IDS: y}', "");
Expect(0, 195101, '\p{^IDS: y}', "");
Expect(0, 195101, '\P{IDS: y}', "");
Expect(1, 195101, '\P{^IDS: y}', "");
Expect(0, 195102, '\p{IDS: y}', "");
Expect(1, 195102, '\p{^IDS: y}', "");
Expect(1, 195102, '\P{IDS: y}', "");
Expect(0, 195102, '\P{^IDS: y}', "");
Expect(1, 195101, '\p{IDS= -y}', "");
Expect(0, 195101, '\p{^IDS= -y}', "");
Expect(0, 195101, '\P{IDS= -y}', "");
Expect(1, 195101, '\P{^IDS= -y}', "");
Expect(0, 195102, '\p{IDS= -y}', "");
Expect(1, 195102, '\p{^IDS= -y}', "");
Expect(1, 195102, '\P{IDS= -y}', "");
Expect(0, 195102, '\P{^IDS= -y}', "");
Error('\p{Is_ID_Start= T/a/}');
Error('\P{Is_ID_Start= T/a/}');
Expect(1, 195101, '\p{Is_ID_Start=t}', "");
Expect(0, 195101, '\p{^Is_ID_Start=t}', "");
Expect(0, 195101, '\P{Is_ID_Start=t}', "");
Expect(1, 195101, '\P{^Is_ID_Start=t}', "");
Expect(0, 195102, '\p{Is_ID_Start=t}', "");
Expect(1, 195102, '\p{^Is_ID_Start=t}', "");
Expect(1, 195102, '\P{Is_ID_Start=t}', "");
Expect(0, 195102, '\P{^Is_ID_Start=t}', "");
Expect(1, 195101, '\p{Is_ID_Start=__T}', "");
Expect(0, 195101, '\p{^Is_ID_Start=__T}', "");
Expect(0, 195101, '\P{Is_ID_Start=__T}', "");
Expect(1, 195101, '\P{^Is_ID_Start=__T}', "");
Expect(0, 195102, '\p{Is_ID_Start=__T}', "");
Expect(1, 195102, '\p{^Is_ID_Start=__T}', "");
Expect(1, 195102, '\P{Is_ID_Start=__T}', "");
Expect(0, 195102, '\P{^Is_ID_Start=__T}', "");
Error('\p{Is_IDS=/a/true}');
Error('\P{Is_IDS=/a/true}');
Expect(1, 195101, '\p{Is_IDS=true}', "");
Expect(0, 195101, '\p{^Is_IDS=true}', "");
Expect(0, 195101, '\P{Is_IDS=true}', "");
Expect(1, 195101, '\P{^Is_IDS=true}', "");
Expect(0, 195102, '\p{Is_IDS=true}', "");
Expect(1, 195102, '\p{^Is_IDS=true}', "");
Expect(1, 195102, '\P{Is_IDS=true}', "");
Expect(0, 195102, '\P{^Is_IDS=true}', "");
Expect(1, 195101, '\p{Is_IDS=		TRUE}', "");
Expect(0, 195101, '\p{^Is_IDS=		TRUE}', "");
Expect(0, 195101, '\P{Is_IDS=		TRUE}', "");
Expect(1, 195101, '\P{^Is_IDS=		TRUE}', "");
Expect(0, 195102, '\p{Is_IDS=		TRUE}', "");
Expect(1, 195102, '\p{^Is_IDS=		TRUE}', "");
Expect(1, 195102, '\P{Is_IDS=		TRUE}', "");
Expect(0, 195102, '\P{^Is_IDS=		TRUE}', "");
Error('\p{ID_Start=_no/a/}');
Error('\P{ID_Start=_no/a/}');
Expect(1, 195102, '\p{ID_Start=no}', "");
Expect(0, 195102, '\p{^ID_Start=no}', "");
Expect(0, 195102, '\P{ID_Start=no}', "");
Expect(1, 195102, '\P{^ID_Start=no}', "");
Expect(0, 195101, '\p{ID_Start=no}', "");
Expect(1, 195101, '\p{^ID_Start=no}', "");
Expect(1, 195101, '\P{ID_Start=no}', "");
Expect(0, 195101, '\P{^ID_Start=no}', "");
Expect(1, 195102, '\p{ID_Start=_-NO}', "");
Expect(0, 195102, '\p{^ID_Start=_-NO}', "");
Expect(0, 195102, '\P{ID_Start=_-NO}', "");
Expect(1, 195102, '\P{^ID_Start=_-NO}', "");
Expect(0, 195101, '\p{ID_Start=_-NO}', "");
Expect(1, 195101, '\p{^ID_Start=_-NO}', "");
Expect(1, 195101, '\P{ID_Start=_-NO}', "");
Expect(0, 195101, '\P{^ID_Start=_-NO}', "");
Error('\p{IDS=:=_N}');
Error('\P{IDS=:=_N}');
Expect(1, 195102, '\p{IDS=n}', "");
Expect(0, 195102, '\p{^IDS=n}', "");
Expect(0, 195102, '\P{IDS=n}', "");
Expect(1, 195102, '\P{^IDS=n}', "");
Expect(0, 195101, '\p{IDS=n}', "");
Expect(1, 195101, '\p{^IDS=n}', "");
Expect(1, 195101, '\P{IDS=n}', "");
Expect(0, 195101, '\P{^IDS=n}', "");
Expect(1, 195102, '\p{IDS=-_N}', "");
Expect(0, 195102, '\p{^IDS=-_N}', "");
Expect(0, 195102, '\P{IDS=-_N}', "");
Expect(1, 195102, '\P{^IDS=-_N}', "");
Expect(0, 195101, '\p{IDS=-_N}', "");
Expect(1, 195101, '\p{^IDS=-_N}', "");
Expect(1, 195101, '\P{IDS=-_N}', "");
Expect(0, 195101, '\P{^IDS=-_N}', "");
Error('\p{Is_ID_Start=_:=F}');
Error('\P{Is_ID_Start=_:=F}');
Expect(1, 195102, '\p{Is_ID_Start:f}', "");
Expect(0, 195102, '\p{^Is_ID_Start:f}', "");
Expect(0, 195102, '\P{Is_ID_Start:f}', "");
Expect(1, 195102, '\P{^Is_ID_Start:f}', "");
Expect(0, 195101, '\p{Is_ID_Start:f}', "");
Expect(1, 195101, '\p{^Is_ID_Start:f}', "");
Expect(1, 195101, '\P{Is_ID_Start:f}', "");
Expect(0, 195101, '\P{^Is_ID_Start:f}', "");
Expect(1, 195102, '\p{Is_ID_Start=F}', "");
Expect(0, 195102, '\p{^Is_ID_Start=F}', "");
Expect(0, 195102, '\P{Is_ID_Start=F}', "");
Expect(1, 195102, '\P{^Is_ID_Start=F}', "");
Expect(0, 195101, '\p{Is_ID_Start=F}', "");
Expect(1, 195101, '\p{^Is_ID_Start=F}', "");
Expect(1, 195101, '\P{Is_ID_Start=F}', "");
Expect(0, 195101, '\P{^Is_ID_Start=F}', "");
Error('\p{Is_IDS=:= False}');
Error('\P{Is_IDS=:= False}');
Expect(1, 195102, '\p{Is_IDS=false}', "");
Expect(0, 195102, '\p{^Is_IDS=false}', "");
Expect(0, 195102, '\P{Is_IDS=false}', "");
Expect(1, 195102, '\P{^Is_IDS=false}', "");
Expect(0, 195101, '\p{Is_IDS=false}', "");
Expect(1, 195101, '\p{^Is_IDS=false}', "");
Expect(1, 195101, '\P{Is_IDS=false}', "");
Expect(0, 195101, '\P{^Is_IDS=false}', "");
Expect(1, 195102, '\p{Is_IDS=	false}', "");
Expect(0, 195102, '\p{^Is_IDS=	false}', "");
Expect(0, 195102, '\P{Is_IDS=	false}', "");
Expect(1, 195102, '\P{^Is_IDS=	false}', "");
Expect(0, 195101, '\p{Is_IDS=	false}', "");
Expect(1, 195101, '\p{^Is_IDS=	false}', "");
Expect(1, 195101, '\P{Is_IDS=	false}', "");
Expect(0, 195101, '\P{^Is_IDS=	false}', "");
Error('\p{Sentence_Break=/a/ sp}');
Error('\P{Sentence_Break=/a/ sp}');
Expect(1, 12288, '\p{Sentence_Break=sp}', "");
Expect(0, 12288, '\p{^Sentence_Break=sp}', "");
Expect(0, 12288, '\P{Sentence_Break=sp}', "");
Expect(1, 12288, '\P{^Sentence_Break=sp}', "");
Expect(0, 12289, '\p{Sentence_Break=sp}', "");
Expect(1, 12289, '\p{^Sentence_Break=sp}', "");
Expect(1, 12289, '\P{Sentence_Break=sp}', "");
Expect(0, 12289, '\P{^Sentence_Break=sp}', "");
Expect(1, 12288, '\p{Sentence_Break=	_Sp}', "");
Expect(0, 12288, '\p{^Sentence_Break=	_Sp}', "");
Expect(0, 12288, '\P{Sentence_Break=	_Sp}', "");
Expect(1, 12288, '\P{^Sentence_Break=	_Sp}', "");
Expect(0, 12289, '\p{Sentence_Break=	_Sp}', "");
Expect(1, 12289, '\p{^Sentence_Break=	_Sp}', "");
Expect(1, 12289, '\P{Sentence_Break=	_Sp}', "");
Expect(0, 12289, '\P{^Sentence_Break=	_Sp}', "");
Error('\p{SB=	/a/SP}');
Error('\P{SB=	/a/SP}');
Expect(1, 12288, '\p{SB=sp}', "");
Expect(0, 12288, '\p{^SB=sp}', "");
Expect(0, 12288, '\P{SB=sp}', "");
Expect(1, 12288, '\P{^SB=sp}', "");
Expect(0, 12289, '\p{SB=sp}', "");
Expect(1, 12289, '\p{^SB=sp}', "");
Expect(1, 12289, '\P{SB=sp}', "");
Expect(0, 12289, '\P{^SB=sp}', "");
Expect(1, 12288, '\p{SB:--SP}', "");
Expect(0, 12288, '\p{^SB:--SP}', "");
Expect(0, 12288, '\P{SB:--SP}', "");
Expect(1, 12288, '\P{^SB:--SP}', "");
Expect(0, 12289, '\p{SB:--SP}', "");
Expect(1, 12289, '\p{^SB:--SP}', "");
Expect(1, 12289, '\P{SB:--SP}', "");
Expect(0, 12289, '\P{^SB:--SP}', "");
Error('\p{Is_Sentence_Break=_/a/Sp}');
Error('\P{Is_Sentence_Break=_/a/Sp}');
Expect(1, 12288, '\p{Is_Sentence_Break: sp}', "");
Expect(0, 12288, '\p{^Is_Sentence_Break: sp}', "");
Expect(0, 12288, '\P{Is_Sentence_Break: sp}', "");
Expect(1, 12288, '\P{^Is_Sentence_Break: sp}', "");
Expect(0, 12289, '\p{Is_Sentence_Break: sp}', "");
Expect(1, 12289, '\p{^Is_Sentence_Break: sp}', "");
Expect(1, 12289, '\P{Is_Sentence_Break: sp}', "");
Expect(0, 12289, '\P{^Is_Sentence_Break: sp}', "");
Expect(1, 12288, '\p{Is_Sentence_Break=_ sp}', "");
Expect(0, 12288, '\p{^Is_Sentence_Break=_ sp}', "");
Expect(0, 12288, '\P{Is_Sentence_Break=_ sp}', "");
Expect(1, 12288, '\P{^Is_Sentence_Break=_ sp}', "");
Expect(0, 12289, '\p{Is_Sentence_Break=_ sp}', "");
Expect(1, 12289, '\p{^Is_Sentence_Break=_ sp}', "");
Expect(1, 12289, '\P{Is_Sentence_Break=_ sp}', "");
Expect(0, 12289, '\P{^Is_Sentence_Break=_ sp}', "");
Error('\p{Is_SB=_Sp/a/}');
Error('\P{Is_SB=_Sp/a/}');
Expect(1, 12288, '\p{Is_SB=sp}', "");
Expect(0, 12288, '\p{^Is_SB=sp}', "");
Expect(0, 12288, '\P{Is_SB=sp}', "");
Expect(1, 12288, '\P{^Is_SB=sp}', "");
Expect(0, 12289, '\p{Is_SB=sp}', "");
Expect(1, 12289, '\p{^Is_SB=sp}', "");
Expect(1, 12289, '\P{Is_SB=sp}', "");
Expect(0, 12289, '\P{^Is_SB=sp}', "");
Expect(1, 12288, '\p{Is_SB=-_Sp}', "");
Expect(0, 12288, '\p{^Is_SB=-_Sp}', "");
Expect(0, 12288, '\P{Is_SB=-_Sp}', "");
Expect(1, 12288, '\P{^Is_SB=-_Sp}', "");
Expect(0, 12289, '\p{Is_SB=-_Sp}', "");
Expect(1, 12289, '\p{^Is_SB=-_Sp}', "");
Expect(1, 12289, '\P{Is_SB=-_Sp}', "");
Expect(0, 12289, '\P{^Is_SB=-_Sp}', "");
Error('\p{Sentence_Break=:=  scontinue}');
Error('\P{Sentence_Break=:=  scontinue}');
Expect(1, 65380, '\p{Sentence_Break=scontinue}', "");
Expect(0, 65380, '\p{^Sentence_Break=scontinue}', "");
Expect(0, 65380, '\P{Sentence_Break=scontinue}', "");
Expect(1, 65380, '\P{^Sentence_Break=scontinue}', "");
Expect(0, 65381, '\p{Sentence_Break=scontinue}', "");
Expect(1, 65381, '\p{^Sentence_Break=scontinue}', "");
Expect(1, 65381, '\P{Sentence_Break=scontinue}', "");
Expect(0, 65381, '\P{^Sentence_Break=scontinue}', "");
Expect(1, 65380, '\p{Sentence_Break=_	SContinue}', "");
Expect(0, 65380, '\p{^Sentence_Break=_	SContinue}', "");
Expect(0, 65380, '\P{Sentence_Break=_	SContinue}', "");
Expect(1, 65380, '\P{^Sentence_Break=_	SContinue}', "");
Expect(0, 65381, '\p{Sentence_Break=_	SContinue}', "");
Expect(1, 65381, '\p{^Sentence_Break=_	SContinue}', "");
Expect(1, 65381, '\P{Sentence_Break=_	SContinue}', "");
Expect(0, 65381, '\P{^Sentence_Break=_	SContinue}', "");
Error('\p{SB=/a/-	SC}');
Error('\P{SB=/a/-	SC}');
Expect(1, 65380, '\p{SB=sc}', "");
Expect(0, 65380, '\p{^SB=sc}', "");
Expect(0, 65380, '\P{SB=sc}', "");
Expect(1, 65380, '\P{^SB=sc}', "");
Expect(0, 65381, '\p{SB=sc}', "");
Expect(1, 65381, '\p{^SB=sc}', "");
Expect(1, 65381, '\P{SB=sc}', "");
Expect(0, 65381, '\P{^SB=sc}', "");
Expect(1, 65380, '\p{SB=SC}', "");
Expect(0, 65380, '\p{^SB=SC}', "");
Expect(0, 65380, '\P{SB=SC}', "");
Expect(1, 65380, '\P{^SB=SC}', "");
Expect(0, 65381, '\p{SB=SC}', "");
Expect(1, 65381, '\p{^SB=SC}', "");
Expect(1, 65381, '\P{SB=SC}', "");
Expect(0, 65381, '\P{^SB=SC}', "");
Error('\p{Is_Sentence_Break=:=scontinue}');
Error('\P{Is_Sentence_Break=:=scontinue}');
Expect(1, 65380, '\p{Is_Sentence_Break=scontinue}', "");
Expect(0, 65380, '\p{^Is_Sentence_Break=scontinue}', "");
Expect(0, 65380, '\P{Is_Sentence_Break=scontinue}', "");
Expect(1, 65380, '\P{^Is_Sentence_Break=scontinue}', "");
Expect(0, 65381, '\p{Is_Sentence_Break=scontinue}', "");
Expect(1, 65381, '\p{^Is_Sentence_Break=scontinue}', "");
Expect(1, 65381, '\P{Is_Sentence_Break=scontinue}', "");
Expect(0, 65381, '\P{^Is_Sentence_Break=scontinue}', "");
Expect(1, 65380, '\p{Is_Sentence_Break:	 _SContinue}', "");
Expect(0, 65380, '\p{^Is_Sentence_Break:	 _SContinue}', "");
Expect(0, 65380, '\P{Is_Sentence_Break:	 _SContinue}', "");
Expect(1, 65380, '\P{^Is_Sentence_Break:	 _SContinue}', "");
Expect(0, 65381, '\p{Is_Sentence_Break:	 _SContinue}', "");
Expect(1, 65381, '\p{^Is_Sentence_Break:	 _SContinue}', "");
Expect(1, 65381, '\P{Is_Sentence_Break:	 _SContinue}', "");
Expect(0, 65381, '\P{^Is_Sentence_Break:	 _SContinue}', "");
Error('\p{Is_SB=:=SC}');
Error('\P{Is_SB=:=SC}');
Expect(1, 65380, '\p{Is_SB=sc}', "");
Expect(0, 65380, '\p{^Is_SB=sc}', "");
Expect(0, 65380, '\P{Is_SB=sc}', "");
Expect(1, 65380, '\P{^Is_SB=sc}', "");
Expect(0, 65381, '\p{Is_SB=sc}', "");
Expect(1, 65381, '\p{^Is_SB=sc}', "");
Expect(1, 65381, '\P{Is_SB=sc}', "");
Expect(0, 65381, '\P{^Is_SB=sc}', "");
Expect(1, 65380, '\p{Is_SB=	 SC}', "");
Expect(0, 65380, '\p{^Is_SB=	 SC}', "");
Expect(0, 65380, '\P{Is_SB=	 SC}', "");
Expect(1, 65380, '\P{^Is_SB=	 SC}', "");
Expect(0, 65381, '\p{Is_SB=	 SC}', "");
Expect(1, 65381, '\p{^Is_SB=	 SC}', "");
Expect(1, 65381, '\P{Is_SB=	 SC}', "");
Expect(0, 65381, '\P{^Is_SB=	 SC}', "");
Error('\p{Sentence_Break=ATerm/a/}');
Error('\P{Sentence_Break=ATerm/a/}');
Expect(1, 65294, '\p{Sentence_Break=aterm}', "");
Expect(0, 65294, '\p{^Sentence_Break=aterm}', "");
Expect(0, 65294, '\P{Sentence_Break=aterm}', "");
Expect(1, 65294, '\P{^Sentence_Break=aterm}', "");
Expect(0, 65295, '\p{Sentence_Break=aterm}', "");
Expect(1, 65295, '\p{^Sentence_Break=aterm}', "");
Expect(1, 65295, '\P{Sentence_Break=aterm}', "");
Expect(0, 65295, '\P{^Sentence_Break=aterm}', "");
Expect(1, 65294, '\p{Sentence_Break=__ATERM}', "");
Expect(0, 65294, '\p{^Sentence_Break=__ATERM}', "");
Expect(0, 65294, '\P{Sentence_Break=__ATERM}', "");
Expect(1, 65294, '\P{^Sentence_Break=__ATERM}', "");
Expect(0, 65295, '\p{Sentence_Break=__ATERM}', "");
Expect(1, 65295, '\p{^Sentence_Break=__ATERM}', "");
Expect(1, 65295, '\P{Sentence_Break=__ATERM}', "");
Expect(0, 65295, '\P{^Sentence_Break=__ATERM}', "");
Error('\p{SB=/a/_at}');
Error('\P{SB=/a/_at}');
Expect(1, 65294, '\p{SB=at}', "");
Expect(0, 65294, '\p{^SB=at}', "");
Expect(0, 65294, '\P{SB=at}', "");
Expect(1, 65294, '\P{^SB=at}', "");
Expect(0, 65295, '\p{SB=at}', "");
Expect(1, 65295, '\p{^SB=at}', "");
Expect(1, 65295, '\P{SB=at}', "");
Expect(0, 65295, '\P{^SB=at}', "");
Expect(1, 65294, '\p{SB=-_AT}', "");
Expect(0, 65294, '\p{^SB=-_AT}', "");
Expect(0, 65294, '\P{SB=-_AT}', "");
Expect(1, 65294, '\P{^SB=-_AT}', "");
Expect(0, 65295, '\p{SB=-_AT}', "");
Expect(1, 65295, '\p{^SB=-_AT}', "");
Expect(1, 65295, '\P{SB=-_AT}', "");
Expect(0, 65295, '\P{^SB=-_AT}', "");
Error('\p{Is_Sentence_Break=:=ATERM}');
Error('\P{Is_Sentence_Break=:=ATERM}');
Expect(1, 65294, '\p{Is_Sentence_Break=aterm}', "");
Expect(0, 65294, '\p{^Is_Sentence_Break=aterm}', "");
Expect(0, 65294, '\P{Is_Sentence_Break=aterm}', "");
Expect(1, 65294, '\P{^Is_Sentence_Break=aterm}', "");
Expect(0, 65295, '\p{Is_Sentence_Break=aterm}', "");
Expect(1, 65295, '\p{^Is_Sentence_Break=aterm}', "");
Expect(1, 65295, '\P{Is_Sentence_Break=aterm}', "");
Expect(0, 65295, '\P{^Is_Sentence_Break=aterm}', "");
Expect(1, 65294, '\p{Is_Sentence_Break=-_ATerm}', "");
Expect(0, 65294, '\p{^Is_Sentence_Break=-_ATerm}', "");
Expect(0, 65294, '\P{Is_Sentence_Break=-_ATerm}', "");
Expect(1, 65294, '\P{^Is_Sentence_Break=-_ATerm}', "");
Expect(0, 65295, '\p{Is_Sentence_Break=-_ATerm}', "");
Expect(1, 65295, '\p{^Is_Sentence_Break=-_ATerm}', "");
Expect(1, 65295, '\P{Is_Sentence_Break=-_ATerm}', "");
Expect(0, 65295, '\P{^Is_Sentence_Break=-_ATerm}', "");
Error('\p{Is_SB=/a/	AT}');
Error('\P{Is_SB=/a/	AT}');
Expect(1, 65294, '\p{Is_SB:	at}', "");
Expect(0, 65294, '\p{^Is_SB:	at}', "");
Expect(0, 65294, '\P{Is_SB:	at}', "");
Expect(1, 65294, '\P{^Is_SB:	at}', "");
Expect(0, 65295, '\p{Is_SB:	at}', "");
Expect(1, 65295, '\p{^Is_SB:	at}', "");
Expect(1, 65295, '\P{Is_SB:	at}', "");
Expect(0, 65295, '\P{^Is_SB:	at}', "");
Expect(1, 65294, '\p{Is_SB= at}', "");
Expect(0, 65294, '\p{^Is_SB= at}', "");
Expect(0, 65294, '\P{Is_SB= at}', "");
Expect(1, 65294, '\P{^Is_SB= at}', "");
Expect(0, 65295, '\p{Is_SB= at}', "");
Expect(1, 65295, '\p{^Is_SB= at}', "");
Expect(1, 65295, '\P{Is_SB= at}', "");
Expect(0, 65295, '\P{^Is_SB= at}', "");
Error('\p{Sentence_Break=_OLetter:=}');
Error('\P{Sentence_Break=_OLetter:=}');
Expect(1, 195101, '\p{Sentence_Break=oletter}', "");
Expect(0, 195101, '\p{^Sentence_Break=oletter}', "");
Expect(0, 195101, '\P{Sentence_Break=oletter}', "");
Expect(1, 195101, '\P{^Sentence_Break=oletter}', "");
Expect(0, 195102, '\p{Sentence_Break=oletter}', "");
Expect(1, 195102, '\p{^Sentence_Break=oletter}', "");
Expect(1, 195102, '\P{Sentence_Break=oletter}', "");
Expect(0, 195102, '\P{^Sentence_Break=oletter}', "");
Expect(1, 195101, '\p{Sentence_Break=_OLetter}', "");
Expect(0, 195101, '\p{^Sentence_Break=_OLetter}', "");
Expect(0, 195101, '\P{Sentence_Break=_OLetter}', "");
Expect(1, 195101, '\P{^Sentence_Break=_OLetter}', "");
Expect(0, 195102, '\p{Sentence_Break=_OLetter}', "");
Expect(1, 195102, '\p{^Sentence_Break=_OLetter}', "");
Expect(1, 195102, '\P{Sentence_Break=_OLetter}', "");
Expect(0, 195102, '\P{^Sentence_Break=_OLetter}', "");
Error('\p{SB=/a/	LE}');
Error('\P{SB=/a/	LE}');
Expect(1, 195101, '\p{SB=le}', "");
Expect(0, 195101, '\p{^SB=le}', "");
Expect(0, 195101, '\P{SB=le}', "");
Expect(1, 195101, '\P{^SB=le}', "");
Expect(0, 195102, '\p{SB=le}', "");
Expect(1, 195102, '\p{^SB=le}', "");
Expect(1, 195102, '\P{SB=le}', "");
Expect(0, 195102, '\P{^SB=le}', "");
Expect(1, 195101, '\p{SB=_LE}', "");
Expect(0, 195101, '\p{^SB=_LE}', "");
Expect(0, 195101, '\P{SB=_LE}', "");
Expect(1, 195101, '\P{^SB=_LE}', "");
Expect(0, 195102, '\p{SB=_LE}', "");
Expect(1, 195102, '\p{^SB=_LE}', "");
Expect(1, 195102, '\P{SB=_LE}', "");
Expect(0, 195102, '\P{^SB=_LE}', "");
Error('\p{Is_Sentence_Break=-	OLetter/a/}');
Error('\P{Is_Sentence_Break=-	OLetter/a/}');
Expect(1, 195101, '\p{Is_Sentence_Break=oletter}', "");
Expect(0, 195101, '\p{^Is_Sentence_Break=oletter}', "");
Expect(0, 195101, '\P{Is_Sentence_Break=oletter}', "");
Expect(1, 195101, '\P{^Is_Sentence_Break=oletter}', "");
Expect(0, 195102, '\p{Is_Sentence_Break=oletter}', "");
Expect(1, 195102, '\p{^Is_Sentence_Break=oletter}', "");
Expect(1, 195102, '\P{Is_Sentence_Break=oletter}', "");
Expect(0, 195102, '\P{^Is_Sentence_Break=oletter}', "");
Expect(1, 195101, '\p{Is_Sentence_Break=_OLetter}', "");
Expect(0, 195101, '\p{^Is_Sentence_Break=_OLetter}', "");
Expect(0, 195101, '\P{Is_Sentence_Break=_OLetter}', "");
Expect(1, 195101, '\P{^Is_Sentence_Break=_OLetter}', "");
Expect(0, 195102, '\p{Is_Sentence_Break=_OLetter}', "");
Expect(1, 195102, '\p{^Is_Sentence_Break=_OLetter}', "");
Expect(1, 195102, '\P{Is_Sentence_Break=_OLetter}', "");
Expect(0, 195102, '\P{^Is_Sentence_Break=_OLetter}', "");
Error('\p{Is_SB:	/a/--LE}');
Error('\P{Is_SB:	/a/--LE}');
Expect(1, 195101, '\p{Is_SB=le}', "");
Expect(0, 195101, '\p{^Is_SB=le}', "");
Expect(0, 195101, '\P{Is_SB=le}', "");
Expect(1, 195101, '\P{^Is_SB=le}', "");
Expect(0, 195102, '\p{Is_SB=le}', "");
Expect(1, 195102, '\p{^Is_SB=le}', "");
Expect(1, 195102, '\P{Is_SB=le}', "");
Expect(0, 195102, '\P{^Is_SB=le}', "");
Expect(1, 195101, '\p{Is_SB= 	LE}', "");
Expect(0, 195101, '\p{^Is_SB= 	LE}', "");
Expect(0, 195101, '\P{Is_SB= 	LE}', "");
Expect(1, 195101, '\P{^Is_SB= 	LE}', "");
Expect(0, 195102, '\p{Is_SB= 	LE}', "");
Expect(1, 195102, '\p{^Is_SB= 	LE}', "");
Expect(1, 195102, '\P{Is_SB= 	LE}', "");
Expect(0, 195102, '\P{^Is_SB= 	LE}', "");
Error('\p{Sentence_Break=	:=CR}');
Error('\P{Sentence_Break=	:=CR}');
Expect(1, 13, '\p{Sentence_Break=cr}', "");
Expect(0, 13, '\p{^Sentence_Break=cr}', "");
Expect(0, 13, '\P{Sentence_Break=cr}', "");
Expect(1, 13, '\P{^Sentence_Break=cr}', "");
Expect(0, 14, '\p{Sentence_Break=cr}', "");
Expect(1, 14, '\p{^Sentence_Break=cr}', "");
Expect(1, 14, '\P{Sentence_Break=cr}', "");
Expect(0, 14, '\P{^Sentence_Break=cr}', "");
Expect(1, 13, '\p{Sentence_Break=-CR}', "");
Expect(0, 13, '\p{^Sentence_Break=-CR}', "");
Expect(0, 13, '\P{Sentence_Break=-CR}', "");
Expect(1, 13, '\P{^Sentence_Break=-CR}', "");
Expect(0, 14, '\p{Sentence_Break=-CR}', "");
Expect(1, 14, '\p{^Sentence_Break=-CR}', "");
Expect(1, 14, '\P{Sentence_Break=-CR}', "");
Expect(0, 14, '\P{^Sentence_Break=-CR}', "");
Error('\p{SB=:=-_CR}');
Error('\P{SB=:=-_CR}');
Expect(1, 13, '\p{SB=cr}', "");
Expect(0, 13, '\p{^SB=cr}', "");
Expect(0, 13, '\P{SB=cr}', "");
Expect(1, 13, '\P{^SB=cr}', "");
Expect(0, 14, '\p{SB=cr}', "");
Expect(1, 14, '\p{^SB=cr}', "");
Expect(1, 14, '\P{SB=cr}', "");
Expect(0, 14, '\P{^SB=cr}', "");
Expect(1, 13, '\p{SB=	-CR}', "");
Expect(0, 13, '\p{^SB=	-CR}', "");
Expect(0, 13, '\P{SB=	-CR}', "");
Expect(1, 13, '\P{^SB=	-CR}', "");
Expect(0, 14, '\p{SB=	-CR}', "");
Expect(1, 14, '\p{^SB=	-CR}', "");
Expect(1, 14, '\P{SB=	-CR}', "");
Expect(0, 14, '\P{^SB=	-CR}', "");
Error('\p{Is_Sentence_Break=_/a/CR}');
Error('\P{Is_Sentence_Break=_/a/CR}');
Expect(1, 13, '\p{Is_Sentence_Break=cr}', "");
Expect(0, 13, '\p{^Is_Sentence_Break=cr}', "");
Expect(0, 13, '\P{Is_Sentence_Break=cr}', "");
Expect(1, 13, '\P{^Is_Sentence_Break=cr}', "");
Expect(0, 14, '\p{Is_Sentence_Break=cr}', "");
Expect(1, 14, '\p{^Is_Sentence_Break=cr}', "");
Expect(1, 14, '\P{Is_Sentence_Break=cr}', "");
Expect(0, 14, '\P{^Is_Sentence_Break=cr}', "");
Expect(1, 13, '\p{Is_Sentence_Break:   --CR}', "");
Expect(0, 13, '\p{^Is_Sentence_Break:   --CR}', "");
Expect(0, 13, '\P{Is_Sentence_Break:   --CR}', "");
Expect(1, 13, '\P{^Is_Sentence_Break:   --CR}', "");
Expect(0, 14, '\p{Is_Sentence_Break:   --CR}', "");
Expect(1, 14, '\p{^Is_Sentence_Break:   --CR}', "");
Expect(1, 14, '\P{Is_Sentence_Break:   --CR}', "");
Expect(0, 14, '\P{^Is_Sentence_Break:   --CR}', "");
Error('\p{Is_SB=/a/	-CR}');
Error('\P{Is_SB=/a/	-CR}');
Expect(1, 13, '\p{Is_SB=cr}', "");
Expect(0, 13, '\p{^Is_SB=cr}', "");
Expect(0, 13, '\P{Is_SB=cr}', "");
Expect(1, 13, '\P{^Is_SB=cr}', "");
Expect(0, 14, '\p{Is_SB=cr}', "");
Expect(1, 14, '\p{^Is_SB=cr}', "");
Expect(1, 14, '\P{Is_SB=cr}', "");
Expect(0, 14, '\P{^Is_SB=cr}', "");
Expect(1, 13, '\p{Is_SB=_	cr}', "");
Expect(0, 13, '\p{^Is_SB=_	cr}', "");
Expect(0, 13, '\P{Is_SB=_	cr}', "");
Expect(1, 13, '\P{^Is_SB=_	cr}', "");
Expect(0, 14, '\p{Is_SB=_	cr}', "");
Expect(1, 14, '\p{^Is_SB=_	cr}', "");
Expect(1, 14, '\P{Is_SB=_	cr}', "");
Expect(0, 14, '\P{^Is_SB=_	cr}', "");
Error('\p{Sentence_Break=/a/Other}');
Error('\P{Sentence_Break=/a/Other}');
Expect(1, 918000, '\p{Sentence_Break=other}', "");
Expect(0, 918000, '\p{^Sentence_Break=other}', "");
Expect(0, 918000, '\P{Sentence_Break=other}', "");
Expect(1, 918000, '\P{^Sentence_Break=other}', "");
Expect(0, 917999, '\p{Sentence_Break=other}', "");
Expect(1, 917999, '\p{^Sentence_Break=other}', "");
Expect(1, 917999, '\P{Sentence_Break=other}', "");
Expect(0, 917999, '\P{^Sentence_Break=other}', "");
Expect(1, 918000, '\p{Sentence_Break:	 -Other}', "");
Expect(0, 918000, '\p{^Sentence_Break:	 -Other}', "");
Expect(0, 918000, '\P{Sentence_Break:	 -Other}', "");
Expect(1, 918000, '\P{^Sentence_Break:	 -Other}', "");
Expect(0, 917999, '\p{Sentence_Break:	 -Other}', "");
Expect(1, 917999, '\p{^Sentence_Break:	 -Other}', "");
Expect(1, 917999, '\P{Sentence_Break:	 -Other}', "");
Expect(0, 917999, '\P{^Sentence_Break:	 -Other}', "");
Error('\p{SB:   /a/-xx}');
Error('\P{SB:   /a/-xx}');
Expect(1, 918000, '\p{SB=xx}', "");
Expect(0, 918000, '\p{^SB=xx}', "");
Expect(0, 918000, '\P{SB=xx}', "");
Expect(1, 918000, '\P{^SB=xx}', "");
Expect(0, 917999, '\p{SB=xx}', "");
Expect(1, 917999, '\p{^SB=xx}', "");
Expect(1, 917999, '\P{SB=xx}', "");
Expect(0, 917999, '\P{^SB=xx}', "");
Expect(1, 918000, '\p{SB= XX}', "");
Expect(0, 918000, '\p{^SB= XX}', "");
Expect(0, 918000, '\P{SB= XX}', "");
Expect(1, 918000, '\P{^SB= XX}', "");
Expect(0, 917999, '\p{SB= XX}', "");
Expect(1, 917999, '\p{^SB= XX}', "");
Expect(1, 917999, '\P{SB= XX}', "");
Expect(0, 917999, '\P{^SB= XX}', "");
Error('\p{Is_Sentence_Break=OTHER:=}');
Error('\P{Is_Sentence_Break=OTHER:=}');
Expect(1, 918000, '\p{Is_Sentence_Break: other}', "");
Expect(0, 918000, '\p{^Is_Sentence_Break: other}', "");
Expect(0, 918000, '\P{Is_Sentence_Break: other}', "");
Expect(1, 918000, '\P{^Is_Sentence_Break: other}', "");
Expect(0, 917999, '\p{Is_Sentence_Break: other}', "");
Expect(1, 917999, '\p{^Is_Sentence_Break: other}', "");
Expect(1, 917999, '\P{Is_Sentence_Break: other}', "");
Expect(0, 917999, '\P{^Is_Sentence_Break: other}', "");
Expect(1, 918000, '\p{Is_Sentence_Break=	Other}', "");
Expect(0, 918000, '\p{^Is_Sentence_Break=	Other}', "");
Expect(0, 918000, '\P{Is_Sentence_Break=	Other}', "");
Expect(1, 918000, '\P{^Is_Sentence_Break=	Other}', "");
Expect(0, 917999, '\p{Is_Sentence_Break=	Other}', "");
Expect(1, 917999, '\p{^Is_Sentence_Break=	Other}', "");
Expect(1, 917999, '\P{Is_Sentence_Break=	Other}', "");
Expect(0, 917999, '\P{^Is_Sentence_Break=	Other}', "");
Error('\p{Is_SB=/a/  XX}');
Error('\P{Is_SB=/a/  XX}');
Expect(1, 918000, '\p{Is_SB=xx}', "");
Expect(0, 918000, '\p{^Is_SB=xx}', "");
Expect(0, 918000, '\P{Is_SB=xx}', "");
Expect(1, 918000, '\P{^Is_SB=xx}', "");
Expect(0, 917999, '\p{Is_SB=xx}', "");
Expect(1, 917999, '\p{^Is_SB=xx}', "");
Expect(1, 917999, '\P{Is_SB=xx}', "");
Expect(0, 917999, '\P{^Is_SB=xx}', "");
Expect(1, 918000, '\p{Is_SB=--XX}', "");
Expect(0, 918000, '\p{^Is_SB=--XX}', "");
Expect(0, 918000, '\P{Is_SB=--XX}', "");
Expect(1, 918000, '\P{^Is_SB=--XX}', "");
Expect(0, 917999, '\p{Is_SB=--XX}', "");
Expect(1, 917999, '\p{^Is_SB=--XX}', "");
Expect(1, 917999, '\P{Is_SB=--XX}', "");
Expect(0, 917999, '\P{^Is_SB=--XX}', "");
Error('\p{Sentence_Break=-	NUMERIC/a/}');
Error('\P{Sentence_Break=-	NUMERIC/a/}');
Expect(1, 120831, '\p{Sentence_Break=numeric}', "");
Expect(0, 120831, '\p{^Sentence_Break=numeric}', "");
Expect(0, 120831, '\P{Sentence_Break=numeric}', "");
Expect(1, 120831, '\P{^Sentence_Break=numeric}', "");
Expect(0, 120832, '\p{Sentence_Break=numeric}', "");
Expect(1, 120832, '\p{^Sentence_Break=numeric}', "");
Expect(1, 120832, '\P{Sentence_Break=numeric}', "");
Expect(0, 120832, '\P{^Sentence_Break=numeric}', "");
Expect(1, 120831, '\p{Sentence_Break= 	Numeric}', "");
Expect(0, 120831, '\p{^Sentence_Break= 	Numeric}', "");
Expect(0, 120831, '\P{Sentence_Break= 	Numeric}', "");
Expect(1, 120831, '\P{^Sentence_Break= 	Numeric}', "");
Expect(0, 120832, '\p{Sentence_Break= 	Numeric}', "");
Expect(1, 120832, '\p{^Sentence_Break= 	Numeric}', "");
Expect(1, 120832, '\P{Sentence_Break= 	Numeric}', "");
Expect(0, 120832, '\P{^Sentence_Break= 	Numeric}', "");
Error('\p{SB=_	NU/a/}');
Error('\P{SB=_	NU/a/}');
Expect(1, 120831, '\p{SB=nu}', "");
Expect(0, 120831, '\p{^SB=nu}', "");
Expect(0, 120831, '\P{SB=nu}', "");
Expect(1, 120831, '\P{^SB=nu}', "");
Expect(0, 120832, '\p{SB=nu}', "");
Expect(1, 120832, '\p{^SB=nu}', "");
Expect(1, 120832, '\P{SB=nu}', "");
Expect(0, 120832, '\P{^SB=nu}', "");
Expect(1, 120831, '\p{SB= -nu}', "");
Expect(0, 120831, '\p{^SB= -nu}', "");
Expect(0, 120831, '\P{SB= -nu}', "");
Expect(1, 120831, '\P{^SB= -nu}', "");
Expect(0, 120832, '\p{SB= -nu}', "");
Expect(1, 120832, '\p{^SB= -nu}', "");
Expect(1, 120832, '\P{SB= -nu}', "");
Expect(0, 120832, '\P{^SB= -nu}', "");
Error('\p{Is_Sentence_Break=	/a/NUMERIC}');
Error('\P{Is_Sentence_Break=	/a/NUMERIC}');
Expect(1, 120831, '\p{Is_Sentence_Break=numeric}', "");
Expect(0, 120831, '\p{^Is_Sentence_Break=numeric}', "");
Expect(0, 120831, '\P{Is_Sentence_Break=numeric}', "");
Expect(1, 120831, '\P{^Is_Sentence_Break=numeric}', "");
Expect(0, 120832, '\p{Is_Sentence_Break=numeric}', "");
Expect(1, 120832, '\p{^Is_Sentence_Break=numeric}', "");
Expect(1, 120832, '\P{Is_Sentence_Break=numeric}', "");
Expect(0, 120832, '\P{^Is_Sentence_Break=numeric}', "");
Expect(1, 120831, '\p{Is_Sentence_Break=-	numeric}', "");
Expect(0, 120831, '\p{^Is_Sentence_Break=-	numeric}', "");
Expect(0, 120831, '\P{Is_Sentence_Break=-	numeric}', "");
Expect(1, 120831, '\P{^Is_Sentence_Break=-	numeric}', "");
Expect(0, 120832, '\p{Is_Sentence_Break=-	numeric}', "");
Expect(1, 120832, '\p{^Is_Sentence_Break=-	numeric}', "");
Expect(1, 120832, '\P{Is_Sentence_Break=-	numeric}', "");
Expect(0, 120832, '\P{^Is_Sentence_Break=-	numeric}', "");
Error('\p{Is_SB=/a/ 	nu}');
Error('\P{Is_SB=/a/ 	nu}');
Expect(1, 120831, '\p{Is_SB=nu}', "");
Expect(0, 120831, '\p{^Is_SB=nu}', "");
Expect(0, 120831, '\P{Is_SB=nu}', "");
Expect(1, 120831, '\P{^Is_SB=nu}', "");
Expect(0, 120832, '\p{Is_SB=nu}', "");
Expect(1, 120832, '\p{^Is_SB=nu}', "");
Expect(1, 120832, '\P{Is_SB=nu}', "");
Expect(0, 120832, '\P{^Is_SB=nu}', "");
Expect(1, 120831, '\p{Is_SB:  NU}', "");
Expect(0, 120831, '\p{^Is_SB:  NU}', "");
Expect(0, 120831, '\P{Is_SB:  NU}', "");
Expect(1, 120831, '\P{^Is_SB:  NU}', "");
Expect(0, 120832, '\p{Is_SB:  NU}', "");
Expect(1, 120832, '\p{^Is_SB:  NU}', "");
Expect(1, 120832, '\P{Is_SB:  NU}', "");
Expect(0, 120832, '\P{^Is_SB:  NU}', "");
Error('\p{Sentence_Break: FORMAT:=}');
Error('\P{Sentence_Break: FORMAT:=}');
Expect(1, 917631, '\p{Sentence_Break=format}', "");
Expect(0, 917631, '\p{^Sentence_Break=format}', "");
Expect(0, 917631, '\P{Sentence_Break=format}', "");
Expect(1, 917631, '\P{^Sentence_Break=format}', "");
Expect(0, 917632, '\p{Sentence_Break=format}', "");
Expect(1, 917632, '\p{^Sentence_Break=format}', "");
Expect(1, 917632, '\P{Sentence_Break=format}', "");
Expect(0, 917632, '\P{^Sentence_Break=format}', "");
Expect(1, 917631, '\p{Sentence_Break=_-Format}', "");
Expect(0, 917631, '\p{^Sentence_Break=_-Format}', "");
Expect(0, 917631, '\P{Sentence_Break=_-Format}', "");
Expect(1, 917631, '\P{^Sentence_Break=_-Format}', "");
Expect(0, 917632, '\p{Sentence_Break=_-Format}', "");
Expect(1, 917632, '\p{^Sentence_Break=_-Format}', "");
Expect(1, 917632, '\P{Sentence_Break=_-Format}', "");
Expect(0, 917632, '\P{^Sentence_Break=_-Format}', "");
Error('\p{SB= :=fo}');
Error('\P{SB= :=fo}');
Expect(1, 917631, '\p{SB=fo}', "");
Expect(0, 917631, '\p{^SB=fo}', "");
Expect(0, 917631, '\P{SB=fo}', "");
Expect(1, 917631, '\P{^SB=fo}', "");
Expect(0, 917632, '\p{SB=fo}', "");
Expect(1, 917632, '\p{^SB=fo}', "");
Expect(1, 917632, '\P{SB=fo}', "");
Expect(0, 917632, '\P{^SB=fo}', "");
Expect(1, 917631, '\p{SB=-FO}', "");
Expect(0, 917631, '\p{^SB=-FO}', "");
Expect(0, 917631, '\P{SB=-FO}', "");
Expect(1, 917631, '\P{^SB=-FO}', "");
Expect(0, 917632, '\p{SB=-FO}', "");
Expect(1, 917632, '\p{^SB=-FO}', "");
Expect(1, 917632, '\P{SB=-FO}', "");
Expect(0, 917632, '\P{^SB=-FO}', "");
Error('\p{Is_Sentence_Break: format/a/}');
Error('\P{Is_Sentence_Break: format/a/}');
Expect(1, 917631, '\p{Is_Sentence_Break=format}', "");
Expect(0, 917631, '\p{^Is_Sentence_Break=format}', "");
Expect(0, 917631, '\P{Is_Sentence_Break=format}', "");
Expect(1, 917631, '\P{^Is_Sentence_Break=format}', "");
Expect(0, 917632, '\p{Is_Sentence_Break=format}', "");
Expect(1, 917632, '\p{^Is_Sentence_Break=format}', "");
Expect(1, 917632, '\P{Is_Sentence_Break=format}', "");
Expect(0, 917632, '\P{^Is_Sentence_Break=format}', "");
Expect(1, 917631, '\p{Is_Sentence_Break=_ Format}', "");
Expect(0, 917631, '\p{^Is_Sentence_Break=_ Format}', "");
Expect(0, 917631, '\P{Is_Sentence_Break=_ Format}', "");
Expect(1, 917631, '\P{^Is_Sentence_Break=_ Format}', "");
Expect(0, 917632, '\p{Is_Sentence_Break=_ Format}', "");
Expect(1, 917632, '\p{^Is_Sentence_Break=_ Format}', "");
Expect(1, 917632, '\P{Is_Sentence_Break=_ Format}', "");
Expect(0, 917632, '\P{^Is_Sentence_Break=_ Format}', "");
Error('\p{Is_SB::=fo}');
Error('\P{Is_SB::=fo}');
Expect(1, 917631, '\p{Is_SB=fo}', "");
Expect(0, 917631, '\p{^Is_SB=fo}', "");
Expect(0, 917631, '\P{Is_SB=fo}', "");
Expect(1, 917631, '\P{^Is_SB=fo}', "");
Expect(0, 917632, '\p{Is_SB=fo}', "");
Expect(1, 917632, '\p{^Is_SB=fo}', "");
Expect(1, 917632, '\P{Is_SB=fo}', "");
Expect(0, 917632, '\P{^Is_SB=fo}', "");
Expect(1, 917631, '\p{Is_SB=-FO}', "");
Expect(0, 917631, '\p{^Is_SB=-FO}', "");
Expect(0, 917631, '\P{Is_SB=-FO}', "");
Expect(1, 917631, '\P{^Is_SB=-FO}', "");
Expect(0, 917632, '\p{Is_SB=-FO}', "");
Expect(1, 917632, '\p{^Is_SB=-FO}', "");
Expect(1, 917632, '\P{Is_SB=-FO}', "");
Expect(0, 917632, '\P{^Is_SB=-FO}', "");
Error('\p{Sentence_Break= /a/lower}');
Error('\P{Sentence_Break= /a/lower}');
Expect(1, 120779, '\p{Sentence_Break=lower}', "");
Expect(0, 120779, '\p{^Sentence_Break=lower}', "");
Expect(0, 120779, '\P{Sentence_Break=lower}', "");
Expect(1, 120779, '\P{^Sentence_Break=lower}', "");
Expect(0, 120780, '\p{Sentence_Break=lower}', "");
Expect(1, 120780, '\p{^Sentence_Break=lower}', "");
Expect(1, 120780, '\P{Sentence_Break=lower}', "");
Expect(0, 120780, '\P{^Sentence_Break=lower}', "");
Expect(1, 120779, '\p{Sentence_Break=-	Lower}', "");
Expect(0, 120779, '\p{^Sentence_Break=-	Lower}', "");
Expect(0, 120779, '\P{Sentence_Break=-	Lower}', "");
Expect(1, 120779, '\P{^Sentence_Break=-	Lower}', "");
Expect(0, 120780, '\p{Sentence_Break=-	Lower}', "");
Expect(1, 120780, '\p{^Sentence_Break=-	Lower}', "");
Expect(1, 120780, '\P{Sentence_Break=-	Lower}', "");
Expect(0, 120780, '\P{^Sentence_Break=-	Lower}', "");
Error('\p{SB= 	LO/a/}');
Error('\P{SB= 	LO/a/}');
Expect(1, 120779, '\p{SB=lo}', "");
Expect(0, 120779, '\p{^SB=lo}', "");
Expect(0, 120779, '\P{SB=lo}', "");
Expect(1, 120779, '\P{^SB=lo}', "");
Expect(0, 120780, '\p{SB=lo}', "");
Expect(1, 120780, '\p{^SB=lo}', "");
Expect(1, 120780, '\P{SB=lo}', "");
Expect(0, 120780, '\P{^SB=lo}', "");
Expect(1, 120779, '\p{SB=	LO}', "");
Expect(0, 120779, '\p{^SB=	LO}', "");
Expect(0, 120779, '\P{SB=	LO}', "");
Expect(1, 120779, '\P{^SB=	LO}', "");
Expect(0, 120780, '\p{SB=	LO}', "");
Expect(1, 120780, '\p{^SB=	LO}', "");
Expect(1, 120780, '\P{SB=	LO}', "");
Expect(0, 120780, '\P{^SB=	LO}', "");
Error('\p{Is_Sentence_Break=	:=lower}');
Error('\P{Is_Sentence_Break=	:=lower}');
Expect(1, 120779, '\p{Is_Sentence_Break:   lower}', "");
Expect(0, 120779, '\p{^Is_Sentence_Break:   lower}', "");
Expect(0, 120779, '\P{Is_Sentence_Break:   lower}', "");
Expect(1, 120779, '\P{^Is_Sentence_Break:   lower}', "");
Expect(0, 120780, '\p{Is_Sentence_Break:   lower}', "");
Expect(1, 120780, '\p{^Is_Sentence_Break:   lower}', "");
Expect(1, 120780, '\P{Is_Sentence_Break:   lower}', "");
Expect(0, 120780, '\P{^Is_Sentence_Break:   lower}', "");
Expect(1, 120779, '\p{Is_Sentence_Break=__lower}', "");
Expect(0, 120779, '\p{^Is_Sentence_Break=__lower}', "");
Expect(0, 120779, '\P{Is_Sentence_Break=__lower}', "");
Expect(1, 120779, '\P{^Is_Sentence_Break=__lower}', "");
Expect(0, 120780, '\p{Is_Sentence_Break=__lower}', "");
Expect(1, 120780, '\p{^Is_Sentence_Break=__lower}', "");
Expect(1, 120780, '\P{Is_Sentence_Break=__lower}', "");
Expect(0, 120780, '\P{^Is_Sentence_Break=__lower}', "");
Error('\p{Is_SB=:=LO}');
Error('\P{Is_SB=:=LO}');
Expect(1, 120779, '\p{Is_SB=lo}', "");
Expect(0, 120779, '\p{^Is_SB=lo}', "");
Expect(0, 120779, '\P{Is_SB=lo}', "");
Expect(1, 120779, '\P{^Is_SB=lo}', "");
Expect(0, 120780, '\p{Is_SB=lo}', "");
Expect(1, 120780, '\p{^Is_SB=lo}', "");
Expect(1, 120780, '\P{Is_SB=lo}', "");
Expect(0, 120780, '\P{^Is_SB=lo}', "");
Expect(1, 120779, '\p{Is_SB= LO}', "");
Expect(0, 120779, '\p{^Is_SB= LO}', "");
Expect(0, 120779, '\P{Is_SB= LO}', "");
Expect(1, 120779, '\P{^Is_SB= LO}', "");
Expect(0, 120780, '\p{Is_SB= LO}', "");
Expect(1, 120780, '\p{^Is_SB= LO}', "");
Expect(1, 120780, '\P{Is_SB= LO}', "");
Expect(0, 120780, '\P{^Is_SB= LO}', "");
Error('\p{Sentence_Break=-Upper/a/}');
Error('\P{Sentence_Break=-Upper/a/}');
Expect(1, 120778, '\p{Sentence_Break=upper}', "");
Expect(0, 120778, '\p{^Sentence_Break=upper}', "");
Expect(0, 120778, '\P{Sentence_Break=upper}', "");
Expect(1, 120778, '\P{^Sentence_Break=upper}', "");
Expect(0, 120779, '\p{Sentence_Break=upper}', "");
Expect(1, 120779, '\p{^Sentence_Break=upper}', "");
Expect(1, 120779, '\P{Sentence_Break=upper}', "");
Expect(0, 120779, '\P{^Sentence_Break=upper}', "");
Expect(1, 120778, '\p{Sentence_Break=		Upper}', "");
Expect(0, 120778, '\p{^Sentence_Break=		Upper}', "");
Expect(0, 120778, '\P{Sentence_Break=		Upper}', "");
Expect(1, 120778, '\P{^Sentence_Break=		Upper}', "");
Expect(0, 120779, '\p{Sentence_Break=		Upper}', "");
Expect(1, 120779, '\p{^Sentence_Break=		Upper}', "");
Expect(1, 120779, '\P{Sentence_Break=		Upper}', "");
Expect(0, 120779, '\P{^Sentence_Break=		Upper}', "");
Error('\p{SB=-UP/a/}');
Error('\P{SB=-UP/a/}');
Expect(1, 120778, '\p{SB=up}', "");
Expect(0, 120778, '\p{^SB=up}', "");
Expect(0, 120778, '\P{SB=up}', "");
Expect(1, 120778, '\P{^SB=up}', "");
Expect(0, 120779, '\p{SB=up}', "");
Expect(1, 120779, '\p{^SB=up}', "");
Expect(1, 120779, '\P{SB=up}', "");
Expect(0, 120779, '\P{^SB=up}', "");
Expect(1, 120778, '\p{SB=-	UP}', "");
Expect(0, 120778, '\p{^SB=-	UP}', "");
Expect(0, 120778, '\P{SB=-	UP}', "");
Expect(1, 120778, '\P{^SB=-	UP}', "");
Expect(0, 120779, '\p{SB=-	UP}', "");
Expect(1, 120779, '\p{^SB=-	UP}', "");
Expect(1, 120779, '\P{SB=-	UP}', "");
Expect(0, 120779, '\P{^SB=-	UP}', "");
Error('\p{Is_Sentence_Break=-_UPPER:=}');
Error('\P{Is_Sentence_Break=-_UPPER:=}');
Expect(1, 120778, '\p{Is_Sentence_Break=upper}', "");
Expect(0, 120778, '\p{^Is_Sentence_Break=upper}', "");
Expect(0, 120778, '\P{Is_Sentence_Break=upper}', "");
Expect(1, 120778, '\P{^Is_Sentence_Break=upper}', "");
Expect(0, 120779, '\p{Is_Sentence_Break=upper}', "");
Expect(1, 120779, '\p{^Is_Sentence_Break=upper}', "");
Expect(1, 120779, '\P{Is_Sentence_Break=upper}', "");
Expect(0, 120779, '\P{^Is_Sentence_Break=upper}', "");
Expect(1, 120778, '\p{Is_Sentence_Break=  upper}', "");
Expect(0, 120778, '\p{^Is_Sentence_Break=  upper}', "");
Expect(0, 120778, '\P{Is_Sentence_Break=  upper}', "");
Expect(1, 120778, '\P{^Is_Sentence_Break=  upper}', "");
Expect(0, 120779, '\p{Is_Sentence_Break=  upper}', "");
Expect(1, 120779, '\p{^Is_Sentence_Break=  upper}', "");
Expect(1, 120779, '\P{Is_Sentence_Break=  upper}', "");
Expect(0, 120779, '\P{^Is_Sentence_Break=  upper}', "");
Error('\p{Is_SB=_ UP:=}');
Error('\P{Is_SB=_ UP:=}');
Expect(1, 120778, '\p{Is_SB=up}', "");
Expect(0, 120778, '\p{^Is_SB=up}', "");
Expect(0, 120778, '\P{Is_SB=up}', "");
Expect(1, 120778, '\P{^Is_SB=up}', "");
Expect(0, 120779, '\p{Is_SB=up}', "");
Expect(1, 120779, '\p{^Is_SB=up}', "");
Expect(1, 120779, '\P{Is_SB=up}', "");
Expect(0, 120779, '\P{^Is_SB=up}', "");
Expect(1, 120778, '\p{Is_SB=-UP}', "");
Expect(0, 120778, '\p{^Is_SB=-UP}', "");
Expect(0, 120778, '\P{Is_SB=-UP}', "");
Expect(1, 120778, '\P{^Is_SB=-UP}', "");
Expect(0, 120779, '\p{Is_SB=-UP}', "");
Expect(1, 120779, '\p{^Is_SB=-UP}', "");
Expect(1, 120779, '\P{Is_SB=-UP}', "");
Expect(0, 120779, '\P{^Is_SB=-UP}', "");
Error('\p{Sentence_Break=  SEP:=}');
Error('\P{Sentence_Break=  SEP:=}');
Expect(1, 8233, '\p{Sentence_Break=sep}', "");
Expect(0, 8233, '\p{^Sentence_Break=sep}', "");
Expect(0, 8233, '\P{Sentence_Break=sep}', "");
Expect(1, 8233, '\P{^Sentence_Break=sep}', "");
Expect(0, 8234, '\p{Sentence_Break=sep}', "");
Expect(1, 8234, '\p{^Sentence_Break=sep}', "");
Expect(1, 8234, '\P{Sentence_Break=sep}', "");
Expect(0, 8234, '\P{^Sentence_Break=sep}', "");
Expect(1, 8233, '\p{Sentence_Break=	_Sep}', "");
Expect(0, 8233, '\p{^Sentence_Break=	_Sep}', "");
Expect(0, 8233, '\P{Sentence_Break=	_Sep}', "");
Expect(1, 8233, '\P{^Sentence_Break=	_Sep}', "");
Expect(0, 8234, '\p{Sentence_Break=	_Sep}', "");
Expect(1, 8234, '\p{^Sentence_Break=	_Sep}', "");
Expect(1, 8234, '\P{Sentence_Break=	_Sep}', "");
Expect(0, 8234, '\P{^Sentence_Break=	_Sep}', "");
Error('\p{SB=	/a/SE}');
Error('\P{SB=	/a/SE}');
Expect(1, 8233, '\p{SB=se}', "");
Expect(0, 8233, '\p{^SB=se}', "");
Expect(0, 8233, '\P{SB=se}', "");
Expect(1, 8233, '\P{^SB=se}', "");
Expect(0, 8234, '\p{SB=se}', "");
Expect(1, 8234, '\p{^SB=se}', "");
Expect(1, 8234, '\P{SB=se}', "");
Expect(0, 8234, '\P{^SB=se}', "");
Expect(1, 8233, '\p{SB=_SE}', "");
Expect(0, 8233, '\p{^SB=_SE}', "");
Expect(0, 8233, '\P{SB=_SE}', "");
Expect(1, 8233, '\P{^SB=_SE}', "");
Expect(0, 8234, '\p{SB=_SE}', "");
Expect(1, 8234, '\p{^SB=_SE}', "");
Expect(1, 8234, '\P{SB=_SE}', "");
Expect(0, 8234, '\P{^SB=_SE}', "");
Error('\p{Is_Sentence_Break=_	Sep:=}');
Error('\P{Is_Sentence_Break=_	Sep:=}');
Expect(1, 8233, '\p{Is_Sentence_Break:	sep}', "");
Expect(0, 8233, '\p{^Is_Sentence_Break:	sep}', "");
Expect(0, 8233, '\P{Is_Sentence_Break:	sep}', "");
Expect(1, 8233, '\P{^Is_Sentence_Break:	sep}', "");
Expect(0, 8234, '\p{Is_Sentence_Break:	sep}', "");
Expect(1, 8234, '\p{^Is_Sentence_Break:	sep}', "");
Expect(1, 8234, '\P{Is_Sentence_Break:	sep}', "");
Expect(0, 8234, '\P{^Is_Sentence_Break:	sep}', "");
Expect(1, 8233, '\p{Is_Sentence_Break=	 Sep}', "");
Expect(0, 8233, '\p{^Is_Sentence_Break=	 Sep}', "");
Expect(0, 8233, '\P{Is_Sentence_Break=	 Sep}', "");
Expect(1, 8233, '\P{^Is_Sentence_Break=	 Sep}', "");
Expect(0, 8234, '\p{Is_Sentence_Break=	 Sep}', "");
Expect(1, 8234, '\p{^Is_Sentence_Break=	 Sep}', "");
Expect(1, 8234, '\P{Is_Sentence_Break=	 Sep}', "");
Expect(0, 8234, '\P{^Is_Sentence_Break=	 Sep}', "");
Error('\p{Is_SB=/a/-SE}');
Error('\P{Is_SB=/a/-SE}');
Expect(1, 8233, '\p{Is_SB=se}', "");
Expect(0, 8233, '\p{^Is_SB=se}', "");
Expect(0, 8233, '\P{Is_SB=se}', "");
Expect(1, 8233, '\P{^Is_SB=se}', "");
Expect(0, 8234, '\p{Is_SB=se}', "");
Expect(1, 8234, '\p{^Is_SB=se}', "");
Expect(1, 8234, '\P{Is_SB=se}', "");
Expect(0, 8234, '\P{^Is_SB=se}', "");
Expect(1, 8233, '\p{Is_SB= se}', "");
Expect(0, 8233, '\p{^Is_SB= se}', "");
Expect(0, 8233, '\P{Is_SB= se}', "");
Expect(1, 8233, '\P{^Is_SB= se}', "");
Expect(0, 8234, '\p{Is_SB= se}', "");
Expect(1, 8234, '\p{^Is_SB= se}', "");
Expect(1, 8234, '\P{Is_SB= se}', "");
Expect(0, 8234, '\P{^Is_SB= se}', "");
Error('\p{Sentence_Break=/a/	extend}');
Error('\P{Sentence_Break=/a/	extend}');
Expect(1, 917999, '\p{Sentence_Break=extend}', "");
Expect(0, 917999, '\p{^Sentence_Break=extend}', "");
Expect(0, 917999, '\P{Sentence_Break=extend}', "");
Expect(1, 917999, '\P{^Sentence_Break=extend}', "");
Expect(0, 918000, '\p{Sentence_Break=extend}', "");
Expect(1, 918000, '\p{^Sentence_Break=extend}', "");
Expect(1, 918000, '\P{Sentence_Break=extend}', "");
Expect(0, 918000, '\P{^Sentence_Break=extend}', "");
Expect(1, 917999, '\p{Sentence_Break=	_EXTEND}', "");
Expect(0, 917999, '\p{^Sentence_Break=	_EXTEND}', "");
Expect(0, 917999, '\P{Sentence_Break=	_EXTEND}', "");
Expect(1, 917999, '\P{^Sentence_Break=	_EXTEND}', "");
Expect(0, 918000, '\p{Sentence_Break=	_EXTEND}', "");
Expect(1, 918000, '\p{^Sentence_Break=	_EXTEND}', "");
Expect(1, 918000, '\P{Sentence_Break=	_EXTEND}', "");
Expect(0, 918000, '\P{^Sentence_Break=	_EXTEND}', "");
Error('\p{SB=/a/ex}');
Error('\P{SB=/a/ex}');
Expect(1, 917999, '\p{SB=ex}', "");
Expect(0, 917999, '\p{^SB=ex}', "");
Expect(0, 917999, '\P{SB=ex}', "");
Expect(1, 917999, '\P{^SB=ex}', "");
Expect(0, 918000, '\p{SB=ex}', "");
Expect(1, 918000, '\p{^SB=ex}', "");
Expect(1, 918000, '\P{SB=ex}', "");
Expect(0, 918000, '\P{^SB=ex}', "");
Expect(1, 917999, '\p{SB= _ex}', "");
Expect(0, 917999, '\p{^SB= _ex}', "");
Expect(0, 917999, '\P{SB= _ex}', "");
Expect(1, 917999, '\P{^SB= _ex}', "");
Expect(0, 918000, '\p{SB= _ex}', "");
Expect(1, 918000, '\p{^SB= _ex}', "");
Expect(1, 918000, '\P{SB= _ex}', "");
Expect(0, 918000, '\P{^SB= _ex}', "");
Error('\p{Is_Sentence_Break= _EXTEND:=}');
Error('\P{Is_Sentence_Break= _EXTEND:=}');
Expect(1, 917999, '\p{Is_Sentence_Break:   extend}', "");
Expect(0, 917999, '\p{^Is_Sentence_Break:   extend}', "");
Expect(0, 917999, '\P{Is_Sentence_Break:   extend}', "");
Expect(1, 917999, '\P{^Is_Sentence_Break:   extend}', "");
Expect(0, 918000, '\p{Is_Sentence_Break:   extend}', "");
Expect(1, 918000, '\p{^Is_Sentence_Break:   extend}', "");
Expect(1, 918000, '\P{Is_Sentence_Break:   extend}', "");
Expect(0, 918000, '\P{^Is_Sentence_Break:   extend}', "");
Expect(1, 917999, '\p{Is_Sentence_Break= -EXTEND}', "");
Expect(0, 917999, '\p{^Is_Sentence_Break= -EXTEND}', "");
Expect(0, 917999, '\P{Is_Sentence_Break= -EXTEND}', "");
Expect(1, 917999, '\P{^Is_Sentence_Break= -EXTEND}', "");
Expect(0, 918000, '\p{Is_Sentence_Break= -EXTEND}', "");
Expect(1, 918000, '\p{^Is_Sentence_Break= -EXTEND}', "");
Expect(1, 918000, '\P{Is_Sentence_Break= -EXTEND}', "");
Expect(0, 918000, '\P{^Is_Sentence_Break= -EXTEND}', "");
Error('\p{Is_SB=:=_EX}');
Error('\P{Is_SB=:=_EX}');
Expect(1, 917999, '\p{Is_SB=ex}', "");
Expect(0, 917999, '\p{^Is_SB=ex}', "");
Expect(0, 917999, '\P{Is_SB=ex}', "");
Expect(1, 917999, '\P{^Is_SB=ex}', "");
Expect(0, 918000, '\p{Is_SB=ex}', "");
Expect(1, 918000, '\p{^Is_SB=ex}', "");
Expect(1, 918000, '\P{Is_SB=ex}', "");
Expect(0, 918000, '\P{^Is_SB=ex}', "");
Expect(1, 917999, '\p{Is_SB=  ex}', "");
Expect(0, 917999, '\p{^Is_SB=  ex}', "");
Expect(0, 917999, '\P{Is_SB=  ex}', "");
Expect(1, 917999, '\P{^Is_SB=  ex}', "");
Expect(0, 918000, '\p{Is_SB=  ex}', "");
Expect(1, 918000, '\p{^Is_SB=  ex}', "");
Expect(1, 918000, '\P{Is_SB=  ex}', "");
Expect(0, 918000, '\P{^Is_SB=  ex}', "");
Error('\p{Sentence_Break=/a/-sterm}');
Error('\P{Sentence_Break=/a/-sterm}');
Expect(1, 70086, '\p{Sentence_Break=sterm}', "");
Expect(0, 70086, '\p{^Sentence_Break=sterm}', "");
Expect(0, 70086, '\P{Sentence_Break=sterm}', "");
Expect(1, 70086, '\P{^Sentence_Break=sterm}', "");
Expect(0, 70087, '\p{Sentence_Break=sterm}', "");
Expect(1, 70087, '\p{^Sentence_Break=sterm}', "");
Expect(1, 70087, '\P{Sentence_Break=sterm}', "");
Expect(0, 70087, '\P{^Sentence_Break=sterm}', "");
Expect(1, 70086, '\p{Sentence_Break=STerm}', "");
Expect(0, 70086, '\p{^Sentence_Break=STerm}', "");
Expect(0, 70086, '\P{Sentence_Break=STerm}', "");
Expect(1, 70086, '\P{^Sentence_Break=STerm}', "");
Expect(0, 70087, '\p{Sentence_Break=STerm}', "");
Expect(1, 70087, '\p{^Sentence_Break=STerm}', "");
Expect(1, 70087, '\P{Sentence_Break=STerm}', "");
Expect(0, 70087, '\P{^Sentence_Break=STerm}', "");
Error('\p{SB=-	ST/a/}');
Error('\P{SB=-	ST/a/}');
Expect(1, 70086, '\p{SB=st}', "");
Expect(0, 70086, '\p{^SB=st}', "");
Expect(0, 70086, '\P{SB=st}', "");
Expect(1, 70086, '\P{^SB=st}', "");
Expect(0, 70087, '\p{SB=st}', "");
Expect(1, 70087, '\p{^SB=st}', "");
Expect(1, 70087, '\P{SB=st}', "");
Expect(0, 70087, '\P{^SB=st}', "");
Expect(1, 70086, '\p{SB=_ST}', "");
Expect(0, 70086, '\p{^SB=_ST}', "");
Expect(0, 70086, '\P{SB=_ST}', "");
Expect(1, 70086, '\P{^SB=_ST}', "");
Expect(0, 70087, '\p{SB=_ST}', "");
Expect(1, 70087, '\p{^SB=_ST}', "");
Expect(1, 70087, '\P{SB=_ST}', "");
Expect(0, 70087, '\P{^SB=_ST}', "");
Error('\p{Is_Sentence_Break=		sterm/a/}');
Error('\P{Is_Sentence_Break=		sterm/a/}');
Expect(1, 70086, '\p{Is_Sentence_Break=sterm}', "");
Expect(0, 70086, '\p{^Is_Sentence_Break=sterm}', "");
Expect(0, 70086, '\P{Is_Sentence_Break=sterm}', "");
Expect(1, 70086, '\P{^Is_Sentence_Break=sterm}', "");
Expect(0, 70087, '\p{Is_Sentence_Break=sterm}', "");
Expect(1, 70087, '\p{^Is_Sentence_Break=sterm}', "");
Expect(1, 70087, '\P{Is_Sentence_Break=sterm}', "");
Expect(0, 70087, '\P{^Is_Sentence_Break=sterm}', "");
Expect(1, 70086, '\p{Is_Sentence_Break=_STERM}', "");
Expect(0, 70086, '\p{^Is_Sentence_Break=_STERM}', "");
Expect(0, 70086, '\P{Is_Sentence_Break=_STERM}', "");
Expect(1, 70086, '\P{^Is_Sentence_Break=_STERM}', "");
Expect(0, 70087, '\p{Is_Sentence_Break=_STERM}', "");
Expect(1, 70087, '\p{^Is_Sentence_Break=_STERM}', "");
Expect(1, 70087, '\P{Is_Sentence_Break=_STERM}', "");
Expect(0, 70087, '\P{^Is_Sentence_Break=_STERM}', "");
Error('\p{Is_SB:   _ST:=}');
Error('\P{Is_SB:   _ST:=}');
Expect(1, 70086, '\p{Is_SB=st}', "");
Expect(0, 70086, '\p{^Is_SB=st}', "");
Expect(0, 70086, '\P{Is_SB=st}', "");
Expect(1, 70086, '\P{^Is_SB=st}', "");
Expect(0, 70087, '\p{Is_SB=st}', "");
Expect(1, 70087, '\p{^Is_SB=st}', "");
Expect(1, 70087, '\P{Is_SB=st}', "");
Expect(0, 70087, '\P{^Is_SB=st}', "");
Expect(1, 70086, '\p{Is_SB=	st}', "");
Expect(0, 70086, '\p{^Is_SB=	st}', "");
Expect(0, 70086, '\P{Is_SB=	st}', "");
Expect(1, 70086, '\P{^Is_SB=	st}', "");
Expect(0, 70087, '\p{Is_SB=	st}', "");
Expect(1, 70087, '\p{^Is_SB=	st}', "");
Expect(1, 70087, '\P{Is_SB=	st}', "");
Expect(0, 70087, '\P{^Is_SB=	st}', "");
Error('\p{Sentence_Break=-_LF:=}');
Error('\P{Sentence_Break=-_LF:=}');
Expect(1, 10, '\p{Sentence_Break=lf}', "");
Expect(0, 10, '\p{^Sentence_Break=lf}', "");
Expect(0, 10, '\P{Sentence_Break=lf}', "");
Expect(1, 10, '\P{^Sentence_Break=lf}', "");
Expect(0, 11, '\p{Sentence_Break=lf}', "");
Expect(1, 11, '\p{^Sentence_Break=lf}', "");
Expect(1, 11, '\P{Sentence_Break=lf}', "");
Expect(0, 11, '\P{^Sentence_Break=lf}', "");
Expect(1, 10, '\p{Sentence_Break: - lf}', "");
Expect(0, 10, '\p{^Sentence_Break: - lf}', "");
Expect(0, 10, '\P{Sentence_Break: - lf}', "");
Expect(1, 10, '\P{^Sentence_Break: - lf}', "");
Expect(0, 11, '\p{Sentence_Break: - lf}', "");
Expect(1, 11, '\p{^Sentence_Break: - lf}', "");
Expect(1, 11, '\P{Sentence_Break: - lf}', "");
Expect(0, 11, '\P{^Sentence_Break: - lf}', "");
Error('\p{SB=	:=LF}');
Error('\P{SB=	:=LF}');
Expect(1, 10, '\p{SB=lf}', "");
Expect(0, 10, '\p{^SB=lf}', "");
Expect(0, 10, '\P{SB=lf}', "");
Expect(1, 10, '\P{^SB=lf}', "");
Expect(0, 11, '\p{SB=lf}', "");
Expect(1, 11, '\p{^SB=lf}', "");
Expect(1, 11, '\P{SB=lf}', "");
Expect(0, 11, '\P{^SB=lf}', "");
Expect(1, 10, '\p{SB= LF}', "");
Expect(0, 10, '\p{^SB= LF}', "");
Expect(0, 10, '\P{SB= LF}', "");
Expect(1, 10, '\P{^SB= LF}', "");
Expect(0, 11, '\p{SB= LF}', "");
Expect(1, 11, '\p{^SB= LF}', "");
Expect(1, 11, '\P{SB= LF}', "");
Expect(0, 11, '\P{^SB= LF}', "");
Error('\p{Is_Sentence_Break=	 LF/a/}');
Error('\P{Is_Sentence_Break=	 LF/a/}');
Expect(1, 10, '\p{Is_Sentence_Break=lf}', "");
Expect(0, 10, '\p{^Is_Sentence_Break=lf}', "");
Expect(0, 10, '\P{Is_Sentence_Break=lf}', "");
Expect(1, 10, '\P{^Is_Sentence_Break=lf}', "");
Expect(0, 11, '\p{Is_Sentence_Break=lf}', "");
Expect(1, 11, '\p{^Is_Sentence_Break=lf}', "");
Expect(1, 11, '\P{Is_Sentence_Break=lf}', "");
Expect(0, 11, '\P{^Is_Sentence_Break=lf}', "");
Expect(1, 10, '\p{Is_Sentence_Break=-_lf}', "");
Expect(0, 10, '\p{^Is_Sentence_Break=-_lf}', "");
Expect(0, 10, '\P{Is_Sentence_Break=-_lf}', "");
Expect(1, 10, '\P{^Is_Sentence_Break=-_lf}', "");
Expect(0, 11, '\p{Is_Sentence_Break=-_lf}', "");
Expect(1, 11, '\p{^Is_Sentence_Break=-_lf}', "");
Expect(1, 11, '\P{Is_Sentence_Break=-_lf}', "");
Expect(0, 11, '\P{^Is_Sentence_Break=-_lf}', "");
Error('\p{Is_SB=/a/_	LF}');
Error('\P{Is_SB=/a/_	LF}');
Expect(1, 10, '\p{Is_SB=lf}', "");
Expect(0, 10, '\p{^Is_SB=lf}', "");
Expect(0, 10, '\P{Is_SB=lf}', "");
Expect(1, 10, '\P{^Is_SB=lf}', "");
Expect(0, 11, '\p{Is_SB=lf}', "");
Expect(1, 11, '\p{^Is_SB=lf}', "");
Expect(1, 11, '\P{Is_SB=lf}', "");
Expect(0, 11, '\P{^Is_SB=lf}', "");
Expect(1, 10, '\p{Is_SB=-LF}', "");
Expect(0, 10, '\p{^Is_SB=-LF}', "");
Expect(0, 10, '\P{Is_SB=-LF}', "");
Expect(1, 10, '\P{^Is_SB=-LF}', "");
Expect(0, 11, '\p{Is_SB=-LF}', "");
Expect(1, 11, '\p{^Is_SB=-LF}', "");
Expect(1, 11, '\P{Is_SB=-LF}', "");
Expect(0, 11, '\P{^Is_SB=-LF}', "");
Error('\p{Sentence_Break=:=Close}');
Error('\P{Sentence_Break=:=Close}');
Expect(1, 65379, '\p{Sentence_Break=close}', "");
Expect(0, 65379, '\p{^Sentence_Break=close}', "");
Expect(0, 65379, '\P{Sentence_Break=close}', "");
Expect(1, 65379, '\P{^Sentence_Break=close}', "");
Expect(0, 65380, '\p{Sentence_Break=close}', "");
Expect(1, 65380, '\p{^Sentence_Break=close}', "");
Expect(1, 65380, '\P{Sentence_Break=close}', "");
Expect(0, 65380, '\P{^Sentence_Break=close}', "");
Expect(1, 65379, '\p{Sentence_Break=	-Close}', "");
Expect(0, 65379, '\p{^Sentence_Break=	-Close}', "");
Expect(0, 65379, '\P{Sentence_Break=	-Close}', "");
Expect(1, 65379, '\P{^Sentence_Break=	-Close}', "");
Expect(0, 65380, '\p{Sentence_Break=	-Close}', "");
Expect(1, 65380, '\p{^Sentence_Break=	-Close}', "");
Expect(1, 65380, '\P{Sentence_Break=	-Close}', "");
Expect(0, 65380, '\P{^Sentence_Break=	-Close}', "");
Error('\p{SB=:=CL}');
Error('\P{SB=:=CL}');
Expect(1, 65379, '\p{SB=cl}', "");
Expect(0, 65379, '\p{^SB=cl}', "");
Expect(0, 65379, '\P{SB=cl}', "");
Expect(1, 65379, '\P{^SB=cl}', "");
Expect(0, 65380, '\p{SB=cl}', "");
Expect(1, 65380, '\p{^SB=cl}', "");
Expect(1, 65380, '\P{SB=cl}', "");
Expect(0, 65380, '\P{^SB=cl}', "");
Expect(1, 65379, '\p{SB: 	-CL}', "");
Expect(0, 65379, '\p{^SB: 	-CL}', "");
Expect(0, 65379, '\P{SB: 	-CL}', "");
Expect(1, 65379, '\P{^SB: 	-CL}', "");
Expect(0, 65380, '\p{SB: 	-CL}', "");
Expect(1, 65380, '\p{^SB: 	-CL}', "");
Expect(1, 65380, '\P{SB: 	-CL}', "");
Expect(0, 65380, '\P{^SB: 	-CL}', "");
Error('\p{Is_Sentence_Break=:=-close}');
Error('\P{Is_Sentence_Break=:=-close}');
Expect(1, 65379, '\p{Is_Sentence_Break=close}', "");
Expect(0, 65379, '\p{^Is_Sentence_Break=close}', "");
Expect(0, 65379, '\P{Is_Sentence_Break=close}', "");
Expect(1, 65379, '\P{^Is_Sentence_Break=close}', "");
Expect(0, 65380, '\p{Is_Sentence_Break=close}', "");
Expect(1, 65380, '\p{^Is_Sentence_Break=close}', "");
Expect(1, 65380, '\P{Is_Sentence_Break=close}', "");
Expect(0, 65380, '\P{^Is_Sentence_Break=close}', "");
Expect(1, 65379, '\p{Is_Sentence_Break:_ close}', "");
Expect(0, 65379, '\p{^Is_Sentence_Break:_ close}', "");
Expect(0, 65379, '\P{Is_Sentence_Break:_ close}', "");
Expect(1, 65379, '\P{^Is_Sentence_Break:_ close}', "");
Expect(0, 65380, '\p{Is_Sentence_Break:_ close}', "");
Expect(1, 65380, '\p{^Is_Sentence_Break:_ close}', "");
Expect(1, 65380, '\P{Is_Sentence_Break:_ close}', "");
Expect(0, 65380, '\P{^Is_Sentence_Break:_ close}', "");
Error('\p{Is_SB=  cl/a/}');
Error('\P{Is_SB=  cl/a/}');
Expect(1, 65379, '\p{Is_SB=cl}', "");
Expect(0, 65379, '\p{^Is_SB=cl}', "");
Expect(0, 65379, '\P{Is_SB=cl}', "");
Expect(1, 65379, '\P{^Is_SB=cl}', "");
Expect(0, 65380, '\p{Is_SB=cl}', "");
Expect(1, 65380, '\p{^Is_SB=cl}', "");
Expect(1, 65380, '\P{Is_SB=cl}', "");
Expect(0, 65380, '\P{^Is_SB=cl}', "");
Expect(1, 65379, '\p{Is_SB:_CL}', "");
Expect(0, 65379, '\p{^Is_SB:_CL}', "");
Expect(0, 65379, '\P{Is_SB:_CL}', "");
Expect(1, 65379, '\P{^Is_SB:_CL}', "");
Expect(0, 65380, '\p{Is_SB:_CL}', "");
Expect(1, 65380, '\p{^Is_SB:_CL}', "");
Expect(1, 65380, '\P{Is_SB:_CL}', "");
Expect(0, 65380, '\P{^Is_SB:_CL}', "");
Error('\p{Script_Extensions=/a/	Gothic}');
Error('\P{Script_Extensions=/a/	Gothic}');
Expect(1, 66378, '\p{Script_Extensions=gothic}', "");
Expect(0, 66378, '\p{^Script_Extensions=gothic}', "");
Expect(0, 66378, '\P{Script_Extensions=gothic}', "");
Expect(1, 66378, '\P{^Script_Extensions=gothic}', "");
Expect(0, 66379, '\p{Script_Extensions=gothic}', "");
Expect(1, 66379, '\p{^Script_Extensions=gothic}', "");
Expect(1, 66379, '\P{Script_Extensions=gothic}', "");
Expect(0, 66379, '\P{^Script_Extensions=gothic}', "");
Expect(1, 66378, '\p{Script_Extensions=-Gothic}', "");
Expect(0, 66378, '\p{^Script_Extensions=-Gothic}', "");
Expect(0, 66378, '\P{Script_Extensions=-Gothic}', "");
Expect(1, 66378, '\P{^Script_Extensions=-Gothic}', "");
Expect(0, 66379, '\p{Script_Extensions=-Gothic}', "");
Expect(1, 66379, '\p{^Script_Extensions=-Gothic}', "");
Expect(1, 66379, '\P{Script_Extensions=-Gothic}', "");
Expect(0, 66379, '\P{^Script_Extensions=-Gothic}', "");
Error('\p{Scx=/a/-	Goth}');
Error('\P{Scx=/a/-	Goth}');
Expect(1, 66378, '\p{Scx:goth}', "");
Expect(0, 66378, '\p{^Scx:goth}', "");
Expect(0, 66378, '\P{Scx:goth}', "");
Expect(1, 66378, '\P{^Scx:goth}', "");
Expect(0, 66379, '\p{Scx:goth}', "");
Expect(1, 66379, '\p{^Scx:goth}', "");
Expect(1, 66379, '\P{Scx:goth}', "");
Expect(0, 66379, '\P{^Scx:goth}', "");
Expect(1, 66378, '\p{Scx=_Goth}', "");
Expect(0, 66378, '\p{^Scx=_Goth}', "");
Expect(0, 66378, '\P{Scx=_Goth}', "");
Expect(1, 66378, '\P{^Scx=_Goth}', "");
Expect(0, 66379, '\p{Scx=_Goth}', "");
Expect(1, 66379, '\p{^Scx=_Goth}', "");
Expect(1, 66379, '\P{Scx=_Goth}', "");
Expect(0, 66379, '\P{^Scx=_Goth}', "");
Error('\p{Is_Script_Extensions=_Gothic:=}');
Error('\P{Is_Script_Extensions=_Gothic:=}');
Expect(1, 66378, '\p{Is_Script_Extensions=gothic}', "");
Expect(0, 66378, '\p{^Is_Script_Extensions=gothic}', "");
Expect(0, 66378, '\P{Is_Script_Extensions=gothic}', "");
Expect(1, 66378, '\P{^Is_Script_Extensions=gothic}', "");
Expect(0, 66379, '\p{Is_Script_Extensions=gothic}', "");
Expect(1, 66379, '\p{^Is_Script_Extensions=gothic}', "");
Expect(1, 66379, '\P{Is_Script_Extensions=gothic}', "");
Expect(0, 66379, '\P{^Is_Script_Extensions=gothic}', "");
Expect(1, 66378, '\p{Is_Script_Extensions=Gothic}', "");
Expect(0, 66378, '\p{^Is_Script_Extensions=Gothic}', "");
Expect(0, 66378, '\P{Is_Script_Extensions=Gothic}', "");
Expect(1, 66378, '\P{^Is_Script_Extensions=Gothic}', "");
Expect(0, 66379, '\p{Is_Script_Extensions=Gothic}', "");
Expect(1, 66379, '\p{^Is_Script_Extensions=Gothic}', "");
Expect(1, 66379, '\P{Is_Script_Extensions=Gothic}', "");
Expect(0, 66379, '\P{^Is_Script_Extensions=Gothic}', "");
Error('\p{Is_Scx=--Goth:=}');
Error('\P{Is_Scx=--Goth:=}');
Expect(1, 66378, '\p{Is_Scx=goth}', "");
Expect(0, 66378, '\p{^Is_Scx=goth}', "");
Expect(0, 66378, '\P{Is_Scx=goth}', "");
Expect(1, 66378, '\P{^Is_Scx=goth}', "");
Expect(0, 66379, '\p{Is_Scx=goth}', "");
Expect(1, 66379, '\p{^Is_Scx=goth}', "");
Expect(1, 66379, '\P{Is_Scx=goth}', "");
Expect(0, 66379, '\P{^Is_Scx=goth}', "");
Expect(1, 66378, '\p{Is_Scx=--GOTH}', "");
Expect(0, 66378, '\p{^Is_Scx=--GOTH}', "");
Expect(0, 66378, '\P{Is_Scx=--GOTH}', "");
Expect(1, 66378, '\P{^Is_Scx=--GOTH}', "");
Expect(0, 66379, '\p{Is_Scx=--GOTH}', "");
Expect(1, 66379, '\p{^Is_Scx=--GOTH}', "");
Expect(1, 66379, '\P{Is_Scx=--GOTH}', "");
Expect(0, 66379, '\P{^Is_Scx=--GOTH}', "");
Error('\p{Script_Extensions=:=_-New_tai_lue}');
Error('\P{Script_Extensions=:=_-New_tai_lue}');
Expect(1, 6623, '\p{Script_Extensions=newtailue}', "");
Expect(0, 6623, '\p{^Script_Extensions=newtailue}', "");
Expect(0, 6623, '\P{Script_Extensions=newtailue}', "");
Expect(1, 6623, '\P{^Script_Extensions=newtailue}', "");
Expect(0, 6624, '\p{Script_Extensions=newtailue}', "");
Expect(1, 6624, '\p{^Script_Extensions=newtailue}', "");
Expect(1, 6624, '\P{Script_Extensions=newtailue}', "");
Expect(0, 6624, '\P{^Script_Extensions=newtailue}', "");
Expect(1, 6623, '\p{Script_Extensions:   - New_Tai_Lue}', "");
Expect(0, 6623, '\p{^Script_Extensions:   - New_Tai_Lue}', "");
Expect(0, 6623, '\P{Script_Extensions:   - New_Tai_Lue}', "");
Expect(1, 6623, '\P{^Script_Extensions:   - New_Tai_Lue}', "");
Expect(0, 6624, '\p{Script_Extensions:   - New_Tai_Lue}', "");
Expect(1, 6624, '\p{^Script_Extensions:   - New_Tai_Lue}', "");
Expect(1, 6624, '\P{Script_Extensions:   - New_Tai_Lue}', "");
Expect(0, 6624, '\P{^Script_Extensions:   - New_Tai_Lue}', "");
Error('\p{Scx=	/a/Talu}');
Error('\P{Scx=	/a/Talu}');
Expect(1, 6623, '\p{Scx=talu}', "");
Expect(0, 6623, '\p{^Scx=talu}', "");
Expect(0, 6623, '\P{Scx=talu}', "");
Expect(1, 6623, '\P{^Scx=talu}', "");
Expect(0, 6624, '\p{Scx=talu}', "");
Expect(1, 6624, '\p{^Scx=talu}', "");
Expect(1, 6624, '\P{Scx=talu}', "");
Expect(0, 6624, '\P{^Scx=talu}', "");
Expect(1, 6623, '\p{Scx=-Talu}', "");
Expect(0, 6623, '\p{^Scx=-Talu}', "");
Expect(0, 6623, '\P{Scx=-Talu}', "");
Expect(1, 6623, '\P{^Scx=-Talu}', "");
Expect(0, 6624, '\p{Scx=-Talu}', "");
Expect(1, 6624, '\p{^Scx=-Talu}', "");
Expect(1, 6624, '\P{Scx=-Talu}', "");
Expect(0, 6624, '\P{^Scx=-Talu}', "");
Error('\p{Is_Script_Extensions:--New_TAI_Lue/a/}');
Error('\P{Is_Script_Extensions:--New_TAI_Lue/a/}');
Expect(1, 6623, '\p{Is_Script_Extensions=newtailue}', "");
Expect(0, 6623, '\p{^Is_Script_Extensions=newtailue}', "");
Expect(0, 6623, '\P{Is_Script_Extensions=newtailue}', "");
Expect(1, 6623, '\P{^Is_Script_Extensions=newtailue}', "");
Expect(0, 6624, '\p{Is_Script_Extensions=newtailue}', "");
Expect(1, 6624, '\p{^Is_Script_Extensions=newtailue}', "");
Expect(1, 6624, '\P{Is_Script_Extensions=newtailue}', "");
Expect(0, 6624, '\P{^Is_Script_Extensions=newtailue}', "");
Expect(1, 6623, '\p{Is_Script_Extensions=--new_Tai_Lue}', "");
Expect(0, 6623, '\p{^Is_Script_Extensions=--new_Tai_Lue}', "");
Expect(0, 6623, '\P{Is_Script_Extensions=--new_Tai_Lue}', "");
Expect(1, 6623, '\P{^Is_Script_Extensions=--new_Tai_Lue}', "");
Expect(0, 6624, '\p{Is_Script_Extensions=--new_Tai_Lue}', "");
Expect(1, 6624, '\p{^Is_Script_Extensions=--new_Tai_Lue}', "");
Expect(1, 6624, '\P{Is_Script_Extensions=--new_Tai_Lue}', "");
Expect(0, 6624, '\P{^Is_Script_Extensions=--new_Tai_Lue}', "");
Error('\p{Is_Scx=:=-Talu}');
Error('\P{Is_Scx=:=-Talu}');
Expect(1, 6623, '\p{Is_Scx:   talu}', "");
Expect(0, 6623, '\p{^Is_Scx:   talu}', "");
Expect(0, 6623, '\P{Is_Scx:   talu}', "");
Expect(1, 6623, '\P{^Is_Scx:   talu}', "");
Expect(0, 6624, '\p{Is_Scx:   talu}', "");
Expect(1, 6624, '\p{^Is_Scx:   talu}', "");
Expect(1, 6624, '\P{Is_Scx:   talu}', "");
Expect(0, 6624, '\P{^Is_Scx:   talu}', "");
Expect(1, 6623, '\p{Is_Scx=-TALU}', "");
Expect(0, 6623, '\p{^Is_Scx=-TALU}', "");
Expect(0, 6623, '\P{Is_Scx=-TALU}', "");
Expect(1, 6623, '\P{^Is_Scx=-TALU}', "");
Expect(0, 6624, '\p{Is_Scx=-TALU}', "");
Expect(1, 6624, '\p{^Is_Scx=-TALU}', "");
Expect(1, 6624, '\P{Is_Scx=-TALU}', "");
Expect(0, 6624, '\P{^Is_Scx=-TALU}', "");
Error('\p{Script_Extensions=_/a/old_Italic}');
Error('\P{Script_Extensions=_/a/old_Italic}');
Expect(1, 66339, '\p{Script_Extensions=olditalic}', "");
Expect(0, 66339, '\p{^Script_Extensions=olditalic}', "");
Expect(0, 66339, '\P{Script_Extensions=olditalic}', "");
Expect(1, 66339, '\P{^Script_Extensions=olditalic}', "");
Expect(0, 66340, '\p{Script_Extensions=olditalic}', "");
Expect(1, 66340, '\p{^Script_Extensions=olditalic}', "");
Expect(1, 66340, '\P{Script_Extensions=olditalic}', "");
Expect(0, 66340, '\P{^Script_Extensions=olditalic}', "");
Expect(1, 66339, '\p{Script_Extensions=_-Old_Italic}', "");
Expect(0, 66339, '\p{^Script_Extensions=_-Old_Italic}', "");
Expect(0, 66339, '\P{Script_Extensions=_-Old_Italic}', "");
Expect(1, 66339, '\P{^Script_Extensions=_-Old_Italic}', "");
Expect(0, 66340, '\p{Script_Extensions=_-Old_Italic}', "");
Expect(1, 66340, '\p{^Script_Extensions=_-Old_Italic}', "");
Expect(1, 66340, '\P{Script_Extensions=_-Old_Italic}', "");
Expect(0, 66340, '\P{^Script_Extensions=_-Old_Italic}', "");
Error('\p{Scx=:=ITAL}');
Error('\P{Scx=:=ITAL}');
Expect(1, 66339, '\p{Scx=ital}', "");
Expect(0, 66339, '\p{^Scx=ital}', "");
Expect(0, 66339, '\P{Scx=ital}', "");
Expect(1, 66339, '\P{^Scx=ital}', "");
Expect(0, 66340, '\p{Scx=ital}', "");
Expect(1, 66340, '\p{^Scx=ital}', "");
Expect(1, 66340, '\P{Scx=ital}', "");
Expect(0, 66340, '\P{^Scx=ital}', "");
Expect(1, 66339, '\p{Scx=  Ital}', "");
Expect(0, 66339, '\p{^Scx=  Ital}', "");
Expect(0, 66339, '\P{Scx=  Ital}', "");
Expect(1, 66339, '\P{^Scx=  Ital}', "");
Expect(0, 66340, '\p{Scx=  Ital}', "");
Expect(1, 66340, '\p{^Scx=  Ital}', "");
Expect(1, 66340, '\P{Scx=  Ital}', "");
Expect(0, 66340, '\P{^Scx=  Ital}', "");
Error('\p{Is_Script_Extensions=/a/-_old_Italic}');
Error('\P{Is_Script_Extensions=/a/-_old_Italic}');
Expect(1, 66339, '\p{Is_Script_Extensions=olditalic}', "");
Expect(0, 66339, '\p{^Is_Script_Extensions=olditalic}', "");
Expect(0, 66339, '\P{Is_Script_Extensions=olditalic}', "");
Expect(1, 66339, '\P{^Is_Script_Extensions=olditalic}', "");
Expect(0, 66340, '\p{Is_Script_Extensions=olditalic}', "");
Expect(1, 66340, '\p{^Is_Script_Extensions=olditalic}', "");
Expect(1, 66340, '\P{Is_Script_Extensions=olditalic}', "");
Expect(0, 66340, '\P{^Is_Script_Extensions=olditalic}', "");
Expect(1, 66339, '\p{Is_Script_Extensions=	_Old_ITALIC}', "");
Expect(0, 66339, '\p{^Is_Script_Extensions=	_Old_ITALIC}', "");
Expect(0, 66339, '\P{Is_Script_Extensions=	_Old_ITALIC}', "");
Expect(1, 66339, '\P{^Is_Script_Extensions=	_Old_ITALIC}', "");
Expect(0, 66340, '\p{Is_Script_Extensions=	_Old_ITALIC}', "");
Expect(1, 66340, '\p{^Is_Script_Extensions=	_Old_ITALIC}', "");
Expect(1, 66340, '\P{Is_Script_Extensions=	_Old_ITALIC}', "");
Expect(0, 66340, '\P{^Is_Script_Extensions=	_Old_ITALIC}', "");
Error('\p{Is_Scx=-:=ITAL}');
Error('\P{Is_Scx=-:=ITAL}');
Expect(1, 66339, '\p{Is_Scx:   ital}', "");
Expect(0, 66339, '\p{^Is_Scx:   ital}', "");
Expect(0, 66339, '\P{Is_Scx:   ital}', "");
Expect(1, 66339, '\P{^Is_Scx:   ital}', "");
Expect(0, 66340, '\p{Is_Scx:   ital}', "");
Expect(1, 66340, '\p{^Is_Scx:   ital}', "");
Expect(1, 66340, '\P{Is_Scx:   ital}', "");
Expect(0, 66340, '\P{^Is_Scx:   ital}', "");
Expect(1, 66339, '\p{Is_Scx=	 Ital}', "");
Expect(0, 66339, '\p{^Is_Scx=	 Ital}', "");
Expect(0, 66339, '\P{Is_Scx=	 Ital}', "");
Expect(1, 66339, '\P{^Is_Scx=	 Ital}', "");
Expect(0, 66340, '\p{Is_Scx=	 Ital}', "");
Expect(1, 66340, '\p{^Is_Scx=	 Ital}', "");
Expect(1, 66340, '\P{Is_Scx=	 Ital}', "");
Expect(0, 66340, '\P{^Is_Scx=	 Ital}', "");
Error('\p{Script_Extensions=:=	 Bamum}');
Error('\P{Script_Extensions=:=	 Bamum}');
Expect(1, 92728, '\p{Script_Extensions=bamum}', "");
Expect(0, 92728, '\p{^Script_Extensions=bamum}', "");
Expect(0, 92728, '\P{Script_Extensions=bamum}', "");
Expect(1, 92728, '\P{^Script_Extensions=bamum}', "");
Expect(0, 92729, '\p{Script_Extensions=bamum}', "");
Expect(1, 92729, '\p{^Script_Extensions=bamum}', "");
Expect(1, 92729, '\P{Script_Extensions=bamum}', "");
Expect(0, 92729, '\P{^Script_Extensions=bamum}', "");
Expect(1, 92728, '\p{Script_Extensions=		Bamum}', "");
Expect(0, 92728, '\p{^Script_Extensions=		Bamum}', "");
Expect(0, 92728, '\P{Script_Extensions=		Bamum}', "");
Expect(1, 92728, '\P{^Script_Extensions=		Bamum}', "");
Expect(0, 92729, '\p{Script_Extensions=		Bamum}', "");
Expect(1, 92729, '\p{^Script_Extensions=		Bamum}', "");
Expect(1, 92729, '\P{Script_Extensions=		Bamum}', "");
Expect(0, 92729, '\P{^Script_Extensions=		Bamum}', "");
Error('\p{Scx=	/a/Bamu}');
Error('\P{Scx=	/a/Bamu}');
Expect(1, 92728, '\p{Scx=bamu}', "");
Expect(0, 92728, '\p{^Scx=bamu}', "");
Expect(0, 92728, '\P{Scx=bamu}', "");
Expect(1, 92728, '\P{^Scx=bamu}', "");
Expect(0, 92729, '\p{Scx=bamu}', "");
Expect(1, 92729, '\p{^Scx=bamu}', "");
Expect(1, 92729, '\P{Scx=bamu}', "");
Expect(0, 92729, '\P{^Scx=bamu}', "");
Expect(1, 92728, '\p{Scx=_Bamu}', "");
Expect(0, 92728, '\p{^Scx=_Bamu}', "");
Expect(0, 92728, '\P{Scx=_Bamu}', "");
Expect(1, 92728, '\P{^Scx=_Bamu}', "");
Expect(0, 92729, '\p{Scx=_Bamu}', "");
Expect(1, 92729, '\p{^Scx=_Bamu}', "");
Expect(1, 92729, '\P{Scx=_Bamu}', "");
Expect(0, 92729, '\P{^Scx=_Bamu}', "");
Error('\p{Is_Script_Extensions=- bamum:=}');
Error('\P{Is_Script_Extensions=- bamum:=}');
Expect(1, 92728, '\p{Is_Script_Extensions=bamum}', "");
Expect(0, 92728, '\p{^Is_Script_Extensions=bamum}', "");
Expect(0, 92728, '\P{Is_Script_Extensions=bamum}', "");
Expect(1, 92728, '\P{^Is_Script_Extensions=bamum}', "");
Expect(0, 92729, '\p{Is_Script_Extensions=bamum}', "");
Expect(1, 92729, '\p{^Is_Script_Extensions=bamum}', "");
Expect(1, 92729, '\P{Is_Script_Extensions=bamum}', "");
Expect(0, 92729, '\P{^Is_Script_Extensions=bamum}', "");
Expect(1, 92728, '\p{Is_Script_Extensions=_ bamum}', "");
Expect(0, 92728, '\p{^Is_Script_Extensions=_ bamum}', "");
Expect(0, 92728, '\P{Is_Script_Extensions=_ bamum}', "");
Expect(1, 92728, '\P{^Is_Script_Extensions=_ bamum}', "");
Expect(0, 92729, '\p{Is_Script_Extensions=_ bamum}', "");
Expect(1, 92729, '\p{^Is_Script_Extensions=_ bamum}', "");
Expect(1, 92729, '\P{Is_Script_Extensions=_ bamum}', "");
Expect(0, 92729, '\P{^Is_Script_Extensions=_ bamum}', "");
Error('\p{Is_Scx=-:=bamu}');
Error('\P{Is_Scx=-:=bamu}');
Expect(1, 92728, '\p{Is_Scx=bamu}', "");
Expect(0, 92728, '\p{^Is_Scx=bamu}', "");
Expect(0, 92728, '\P{Is_Scx=bamu}', "");
Expect(1, 92728, '\P{^Is_Scx=bamu}', "");
Expect(0, 92729, '\p{Is_Scx=bamu}', "");
Expect(1, 92729, '\p{^Is_Scx=bamu}', "");
Expect(1, 92729, '\P{Is_Scx=bamu}', "");
Expect(0, 92729, '\P{^Is_Scx=bamu}', "");
Expect(1, 92728, '\p{Is_Scx: --Bamu}', "");
Expect(0, 92728, '\p{^Is_Scx: --Bamu}', "");
Expect(0, 92728, '\P{Is_Scx: --Bamu}', "");
Expect(1, 92728, '\P{^Is_Scx: --Bamu}', "");
Expect(0, 92729, '\p{Is_Scx: --Bamu}', "");
Expect(1, 92729, '\p{^Is_Scx: --Bamu}', "");
Expect(1, 92729, '\P{Is_Scx: --Bamu}', "");
Expect(0, 92729, '\P{^Is_Scx: --Bamu}', "");
Error('\p{Script_Extensions=-_BRAHMI/a/}');
Error('\P{Script_Extensions=-_BRAHMI/a/}');
Expect(1, 69743, '\p{Script_Extensions:	brahmi}', "");
Expect(0, 69743, '\p{^Script_Extensions:	brahmi}', "");
Expect(0, 69743, '\P{Script_Extensions:	brahmi}', "");
Expect(1, 69743, '\P{^Script_Extensions:	brahmi}', "");
Expect(0, 69744, '\p{Script_Extensions:	brahmi}', "");
Expect(1, 69744, '\p{^Script_Extensions:	brahmi}', "");
Expect(1, 69744, '\P{Script_Extensions:	brahmi}', "");
Expect(0, 69744, '\P{^Script_Extensions:	brahmi}', "");
Expect(1, 69743, '\p{Script_Extensions= -BRAHMI}', "");
Expect(0, 69743, '\p{^Script_Extensions= -BRAHMI}', "");
Expect(0, 69743, '\P{Script_Extensions= -BRAHMI}', "");
Expect(1, 69743, '\P{^Script_Extensions= -BRAHMI}', "");
Expect(0, 69744, '\p{Script_Extensions= -BRAHMI}', "");
Expect(1, 69744, '\p{^Script_Extensions= -BRAHMI}', "");
Expect(1, 69744, '\P{Script_Extensions= -BRAHMI}', "");
Expect(0, 69744, '\P{^Script_Extensions= -BRAHMI}', "");
Error('\p{Scx=/a/ 	Brah}');
Error('\P{Scx=/a/ 	Brah}');
Expect(1, 69743, '\p{Scx=brah}', "");
Expect(0, 69743, '\p{^Scx=brah}', "");
Expect(0, 69743, '\P{Scx=brah}', "");
Expect(1, 69743, '\P{^Scx=brah}', "");
Expect(0, 69744, '\p{Scx=brah}', "");
Expect(1, 69744, '\p{^Scx=brah}', "");
Expect(1, 69744, '\P{Scx=brah}', "");
Expect(0, 69744, '\P{^Scx=brah}', "");
Error('\p{Is_Script_Extensions=/a/_-Brahmi}');
Error('\P{Is_Script_Extensions=/a/_-Brahmi}');
Expect(1, 69743, '\p{Is_Script_Extensions=brahmi}', "");
Expect(0, 69743, '\p{^Is_Script_Extensions=brahmi}', "");
Expect(0, 69743, '\P{Is_Script_Extensions=brahmi}', "");
Expect(1, 69743, '\P{^Is_Script_Extensions=brahmi}', "");
Expect(0, 69744, '\p{Is_Script_Extensions=brahmi}', "");
Expect(1, 69744, '\p{^Is_Script_Extensions=brahmi}', "");
Expect(1, 69744, '\P{Is_Script_Extensions=brahmi}', "");
Expect(0, 69744, '\P{^Is_Script_Extensions=brahmi}', "");
Expect(1, 69743, '\p{Is_Script_Extensions=	Brahmi}', "");
Expect(0, 69743, '\p{^Is_Script_Extensions=	Brahmi}', "");
Expect(0, 69743, '\P{Is_Script_Extensions=	Brahmi}', "");
Expect(1, 69743, '\P{^Is_Script_Extensions=	Brahmi}', "");
Expect(0, 69744, '\p{Is_Script_Extensions=	Brahmi}', "");
Expect(1, 69744, '\p{^Is_Script_Extensions=	Brahmi}', "");
Expect(1, 69744, '\P{Is_Script_Extensions=	Brahmi}', "");
Expect(0, 69744, '\P{^Is_Script_Extensions=	Brahmi}', "");
Error('\p{Is_Scx=/a/ -Brah}');
Error('\P{Is_Scx=/a/ -Brah}');
Expect(1, 69743, '\p{Is_Scx=brah}', "");
Expect(0, 69743, '\p{^Is_Scx=brah}', "");
Expect(0, 69743, '\P{Is_Scx=brah}', "");
Expect(1, 69743, '\P{^Is_Scx=brah}', "");
Expect(0, 69744, '\p{Is_Scx=brah}', "");
Expect(1, 69744, '\p{^Is_Scx=brah}', "");
Expect(1, 69744, '\P{Is_Scx=brah}', "");
Expect(0, 69744, '\P{^Is_Scx=brah}', "");
Expect(1, 69743, '\p{Is_Scx=_Brah}', "");
Expect(0, 69743, '\p{^Is_Scx=_Brah}', "");
Expect(0, 69743, '\P{Is_Scx=_Brah}', "");
Expect(1, 69743, '\P{^Is_Scx=_Brah}', "");
Expect(0, 69744, '\p{Is_Scx=_Brah}', "");
Expect(1, 69744, '\p{^Is_Scx=_Brah}', "");
Expect(1, 69744, '\P{Is_Scx=_Brah}', "");
Expect(0, 69744, '\P{^Is_Scx=_Brah}', "");
Error('\p{Script_Extensions= _Inscriptional_parthian:=}');
Error('\P{Script_Extensions= _Inscriptional_parthian:=}');
Expect(1, 68447, '\p{Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Script_Extensions=	 inscriptional_parthian}', "");
Expect(0, 68447, '\p{^Script_Extensions=	 inscriptional_parthian}', "");
Expect(0, 68447, '\P{Script_Extensions=	 inscriptional_parthian}', "");
Expect(1, 68447, '\P{^Script_Extensions=	 inscriptional_parthian}', "");
Expect(0, 68448, '\p{Script_Extensions=	 inscriptional_parthian}', "");
Expect(1, 68448, '\p{^Script_Extensions=	 inscriptional_parthian}', "");
Expect(1, 68448, '\P{Script_Extensions=	 inscriptional_parthian}', "");
Expect(0, 68448, '\P{^Script_Extensions=	 inscriptional_parthian}', "");
Error('\p{Scx=__Prti:=}');
Error('\P{Scx=__Prti:=}');
Expect(1, 68447, '\p{Scx=prti}', "");
Expect(0, 68447, '\p{^Scx=prti}', "");
Expect(0, 68447, '\P{Scx=prti}', "");
Expect(1, 68447, '\P{^Scx=prti}', "");
Expect(0, 68448, '\p{Scx=prti}', "");
Expect(1, 68448, '\p{^Scx=prti}', "");
Expect(1, 68448, '\P{Scx=prti}', "");
Expect(0, 68448, '\P{^Scx=prti}', "");
Expect(1, 68447, '\p{Scx:	-PRTI}', "");
Expect(0, 68447, '\p{^Scx:	-PRTI}', "");
Expect(0, 68447, '\P{Scx:	-PRTI}', "");
Expect(1, 68447, '\P{^Scx:	-PRTI}', "");
Expect(0, 68448, '\p{Scx:	-PRTI}', "");
Expect(1, 68448, '\p{^Scx:	-PRTI}', "");
Expect(1, 68448, '\P{Scx:	-PRTI}', "");
Expect(0, 68448, '\P{^Scx:	-PRTI}', "");
Error('\p{Is_Script_Extensions=-/a/inscriptional_Parthian}');
Error('\P{Is_Script_Extensions=-/a/inscriptional_Parthian}');
Expect(1, 68447, '\p{Is_Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Is_Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Is_Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Is_Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Is_Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Is_Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Is_Script_Extensions=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Is_Script_Extensions=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Is_Script_Extensions=__Inscriptional_Parthian}', "");
Expect(0, 68447, '\p{^Is_Script_Extensions=__Inscriptional_Parthian}', "");
Expect(0, 68447, '\P{Is_Script_Extensions=__Inscriptional_Parthian}', "");
Expect(1, 68447, '\P{^Is_Script_Extensions=__Inscriptional_Parthian}', "");
Expect(0, 68448, '\p{Is_Script_Extensions=__Inscriptional_Parthian}', "");
Expect(1, 68448, '\p{^Is_Script_Extensions=__Inscriptional_Parthian}', "");
Expect(1, 68448, '\P{Is_Script_Extensions=__Inscriptional_Parthian}', "");
Expect(0, 68448, '\P{^Is_Script_Extensions=__Inscriptional_Parthian}', "");
Error('\p{Is_Scx::=_Prti}');
Error('\P{Is_Scx::=_Prti}');
Expect(1, 68447, '\p{Is_Scx=prti}', "");
Expect(0, 68447, '\p{^Is_Scx=prti}', "");
Expect(0, 68447, '\P{Is_Scx=prti}', "");
Expect(1, 68447, '\P{^Is_Scx=prti}', "");
Expect(0, 68448, '\p{Is_Scx=prti}', "");
Expect(1, 68448, '\p{^Is_Scx=prti}', "");
Expect(1, 68448, '\P{Is_Scx=prti}', "");
Expect(0, 68448, '\P{^Is_Scx=prti}', "");
Expect(1, 68447, '\p{Is_Scx= _PRTI}', "");
Expect(0, 68447, '\p{^Is_Scx= _PRTI}', "");
Expect(0, 68447, '\P{Is_Scx= _PRTI}', "");
Expect(1, 68447, '\P{^Is_Scx= _PRTI}', "");
Expect(0, 68448, '\p{Is_Scx= _PRTI}', "");
Expect(1, 68448, '\p{^Is_Scx= _PRTI}', "");
Expect(1, 68448, '\P{Is_Scx= _PRTI}', "");
Expect(0, 68448, '\P{^Is_Scx= _PRTI}', "");
Error('\p{Script_Extensions=/a/ -vai}');
Error('\P{Script_Extensions=/a/ -vai}');
Expect(1, 42539, '\p{Script_Extensions=vai}', "");
Expect(0, 42539, '\p{^Script_Extensions=vai}', "");
Expect(0, 42539, '\P{Script_Extensions=vai}', "");
Expect(1, 42539, '\P{^Script_Extensions=vai}', "");
Expect(0, 42540, '\p{Script_Extensions=vai}', "");
Expect(1, 42540, '\p{^Script_Extensions=vai}', "");
Expect(1, 42540, '\P{Script_Extensions=vai}', "");
Expect(0, 42540, '\P{^Script_Extensions=vai}', "");
Expect(1, 42539, '\p{Script_Extensions:   - Vai}', "");
Expect(0, 42539, '\p{^Script_Extensions:   - Vai}', "");
Expect(0, 42539, '\P{Script_Extensions:   - Vai}', "");
Expect(1, 42539, '\P{^Script_Extensions:   - Vai}', "");
Expect(0, 42540, '\p{Script_Extensions:   - Vai}', "");
Expect(1, 42540, '\p{^Script_Extensions:   - Vai}', "");
Expect(1, 42540, '\P{Script_Extensions:   - Vai}', "");
Expect(0, 42540, '\P{^Script_Extensions:   - Vai}', "");
Error('\p{Scx=/a/VAII}');
Error('\P{Scx=/a/VAII}');
Expect(1, 42539, '\p{Scx=vaii}', "");
Expect(0, 42539, '\p{^Scx=vaii}', "");
Expect(0, 42539, '\P{Scx=vaii}', "");
Expect(1, 42539, '\P{^Scx=vaii}', "");
Expect(0, 42540, '\p{Scx=vaii}', "");
Expect(1, 42540, '\p{^Scx=vaii}', "");
Expect(1, 42540, '\P{Scx=vaii}', "");
Expect(0, 42540, '\P{^Scx=vaii}', "");
Expect(1, 42539, '\p{Scx=		Vaii}', "");
Expect(0, 42539, '\p{^Scx=		Vaii}', "");
Expect(0, 42539, '\P{Scx=		Vaii}', "");
Expect(1, 42539, '\P{^Scx=		Vaii}', "");
Expect(0, 42540, '\p{Scx=		Vaii}', "");
Expect(1, 42540, '\p{^Scx=		Vaii}', "");
Expect(1, 42540, '\P{Scx=		Vaii}', "");
Expect(0, 42540, '\P{^Scx=		Vaii}', "");
Error('\p{Is_Script_Extensions=:=  Vai}');
Error('\P{Is_Script_Extensions=:=  Vai}');
Expect(1, 42539, '\p{Is_Script_Extensions=vai}', "");
Expect(0, 42539, '\p{^Is_Script_Extensions=vai}', "");
Expect(0, 42539, '\P{Is_Script_Extensions=vai}', "");
Expect(1, 42539, '\P{^Is_Script_Extensions=vai}', "");
Expect(0, 42540, '\p{Is_Script_Extensions=vai}', "");
Expect(1, 42540, '\p{^Is_Script_Extensions=vai}', "");
Expect(1, 42540, '\P{Is_Script_Extensions=vai}', "");
Expect(0, 42540, '\P{^Is_Script_Extensions=vai}', "");
Expect(1, 42539, '\p{Is_Script_Extensions= _Vai}', "");
Expect(0, 42539, '\p{^Is_Script_Extensions= _Vai}', "");
Expect(0, 42539, '\P{Is_Script_Extensions= _Vai}', "");
Expect(1, 42539, '\P{^Is_Script_Extensions= _Vai}', "");
Expect(0, 42540, '\p{Is_Script_Extensions= _Vai}', "");
Expect(1, 42540, '\p{^Is_Script_Extensions= _Vai}', "");
Expect(1, 42540, '\P{Is_Script_Extensions= _Vai}', "");
Expect(0, 42540, '\P{^Is_Script_Extensions= _Vai}', "");
Error('\p{Is_Scx=_ Vaii:=}');
Error('\P{Is_Scx=_ Vaii:=}');
Expect(1, 42539, '\p{Is_Scx=vaii}', "");
Expect(0, 42539, '\p{^Is_Scx=vaii}', "");
Expect(0, 42539, '\P{Is_Scx=vaii}', "");
Expect(1, 42539, '\P{^Is_Scx=vaii}', "");
Expect(0, 42540, '\p{Is_Scx=vaii}', "");
Expect(1, 42540, '\p{^Is_Scx=vaii}', "");
Expect(1, 42540, '\P{Is_Scx=vaii}', "");
Expect(0, 42540, '\P{^Is_Scx=vaii}', "");
Expect(1, 42539, '\p{Is_Scx:   -	Vaii}', "");
Expect(0, 42539, '\p{^Is_Scx:   -	Vaii}', "");
Expect(0, 42539, '\P{Is_Scx:   -	Vaii}', "");
Expect(1, 42539, '\P{^Is_Scx:   -	Vaii}', "");
Expect(0, 42540, '\p{Is_Scx:   -	Vaii}', "");
Expect(1, 42540, '\p{^Is_Scx:   -	Vaii}', "");
Expect(1, 42540, '\P{Is_Scx:   -	Vaii}', "");
Expect(0, 42540, '\P{^Is_Scx:   -	Vaii}', "");
Error('\p{Script_Extensions=/a/ Cherokee}');
Error('\P{Script_Extensions=/a/ Cherokee}');
Expect(1, 5108, '\p{Script_Extensions=cherokee}', "");
Expect(0, 5108, '\p{^Script_Extensions=cherokee}', "");
Expect(0, 5108, '\P{Script_Extensions=cherokee}', "");
Expect(1, 5108, '\P{^Script_Extensions=cherokee}', "");
Expect(0, 5109, '\p{Script_Extensions=cherokee}', "");
Expect(1, 5109, '\p{^Script_Extensions=cherokee}', "");
Expect(1, 5109, '\P{Script_Extensions=cherokee}', "");
Expect(0, 5109, '\P{^Script_Extensions=cherokee}', "");
Expect(1, 5108, '\p{Script_Extensions=  CHEROKEE}', "");
Expect(0, 5108, '\p{^Script_Extensions=  CHEROKEE}', "");
Expect(0, 5108, '\P{Script_Extensions=  CHEROKEE}', "");
Expect(1, 5108, '\P{^Script_Extensions=  CHEROKEE}', "");
Expect(0, 5109, '\p{Script_Extensions=  CHEROKEE}', "");
Expect(1, 5109, '\p{^Script_Extensions=  CHEROKEE}', "");
Expect(1, 5109, '\P{Script_Extensions=  CHEROKEE}', "");
Expect(0, 5109, '\P{^Script_Extensions=  CHEROKEE}', "");
Error('\p{Scx=	/a/Cher}');
Error('\P{Scx=	/a/Cher}');
Expect(1, 5108, '\p{Scx:cher}', "");
Expect(0, 5108, '\p{^Scx:cher}', "");
Expect(0, 5108, '\P{Scx:cher}', "");
Expect(1, 5108, '\P{^Scx:cher}', "");
Expect(0, 5109, '\p{Scx:cher}', "");
Expect(1, 5109, '\p{^Scx:cher}', "");
Expect(1, 5109, '\P{Scx:cher}', "");
Expect(0, 5109, '\P{^Scx:cher}', "");
Expect(1, 5108, '\p{Scx= -Cher}', "");
Expect(0, 5108, '\p{^Scx= -Cher}', "");
Expect(0, 5108, '\P{Scx= -Cher}', "");
Expect(1, 5108, '\P{^Scx= -Cher}', "");
Expect(0, 5109, '\p{Scx= -Cher}', "");
Expect(1, 5109, '\p{^Scx= -Cher}', "");
Expect(1, 5109, '\P{Scx= -Cher}', "");
Expect(0, 5109, '\P{^Scx= -Cher}', "");
Error('\p{Is_Script_Extensions:		_cherokee/a/}');
Error('\P{Is_Script_Extensions:		_cherokee/a/}');
Expect(1, 5108, '\p{Is_Script_Extensions=cherokee}', "");
Expect(0, 5108, '\p{^Is_Script_Extensions=cherokee}', "");
Expect(0, 5108, '\P{Is_Script_Extensions=cherokee}', "");
Expect(1, 5108, '\P{^Is_Script_Extensions=cherokee}', "");
Expect(0, 5109, '\p{Is_Script_Extensions=cherokee}', "");
Expect(1, 5109, '\p{^Is_Script_Extensions=cherokee}', "");
Expect(1, 5109, '\P{Is_Script_Extensions=cherokee}', "");
Expect(0, 5109, '\P{^Is_Script_Extensions=cherokee}', "");
Expect(1, 5108, '\p{Is_Script_Extensions:		CHEROKEE}', "");
Expect(0, 5108, '\p{^Is_Script_Extensions:		CHEROKEE}', "");
Expect(0, 5108, '\P{Is_Script_Extensions:		CHEROKEE}', "");
Expect(1, 5108, '\P{^Is_Script_Extensions:		CHEROKEE}', "");
Expect(0, 5109, '\p{Is_Script_Extensions:		CHEROKEE}', "");
Expect(1, 5109, '\p{^Is_Script_Extensions:		CHEROKEE}', "");
Expect(1, 5109, '\P{Is_Script_Extensions:		CHEROKEE}', "");
Expect(0, 5109, '\P{^Is_Script_Extensions:		CHEROKEE}', "");
Error('\p{Is_Scx=:=		Cher}');
Error('\P{Is_Scx=:=		Cher}');
Expect(1, 5108, '\p{Is_Scx=cher}', "");
Expect(0, 5108, '\p{^Is_Scx=cher}', "");
Expect(0, 5108, '\P{Is_Scx=cher}', "");
Expect(1, 5108, '\P{^Is_Scx=cher}', "");
Expect(0, 5109, '\p{Is_Scx=cher}', "");
Expect(1, 5109, '\p{^Is_Scx=cher}', "");
Expect(1, 5109, '\P{Is_Scx=cher}', "");
Expect(0, 5109, '\P{^Is_Scx=cher}', "");
Expect(1, 5108, '\p{Is_Scx=-CHER}', "");
Expect(0, 5108, '\p{^Is_Scx=-CHER}', "");
Expect(0, 5108, '\P{Is_Scx=-CHER}', "");
Expect(1, 5108, '\P{^Is_Scx=-CHER}', "");
Expect(0, 5109, '\p{Is_Scx=-CHER}', "");
Expect(1, 5109, '\p{^Is_Scx=-CHER}', "");
Expect(1, 5109, '\P{Is_Scx=-CHER}', "");
Expect(0, 5109, '\P{^Is_Scx=-CHER}', "");
Error('\p{Script_Extensions=-:=Runic}');
Error('\P{Script_Extensions=-:=Runic}');
Expect(1, 5872, '\p{Script_Extensions:runic}', "");
Expect(0, 5872, '\p{^Script_Extensions:runic}', "");
Expect(0, 5872, '\P{Script_Extensions:runic}', "");
Expect(1, 5872, '\P{^Script_Extensions:runic}', "");
Expect(0, 5873, '\p{Script_Extensions:runic}', "");
Expect(1, 5873, '\p{^Script_Extensions:runic}', "");
Expect(1, 5873, '\P{Script_Extensions:runic}', "");
Expect(0, 5873, '\P{^Script_Extensions:runic}', "");
Expect(1, 5872, '\p{Script_Extensions=	_RUNIC}', "");
Expect(0, 5872, '\p{^Script_Extensions=	_RUNIC}', "");
Expect(0, 5872, '\P{Script_Extensions=	_RUNIC}', "");
Expect(1, 5872, '\P{^Script_Extensions=	_RUNIC}', "");
Expect(0, 5873, '\p{Script_Extensions=	_RUNIC}', "");
Expect(1, 5873, '\p{^Script_Extensions=	_RUNIC}', "");
Expect(1, 5873, '\P{Script_Extensions=	_RUNIC}', "");
Expect(0, 5873, '\P{^Script_Extensions=	_RUNIC}', "");
Error('\p{Scx:_:=Runr}');
Error('\P{Scx:_:=Runr}');
Expect(1, 5872, '\p{Scx=runr}', "");
Expect(0, 5872, '\p{^Scx=runr}', "");
Expect(0, 5872, '\P{Scx=runr}', "");
Expect(1, 5872, '\P{^Scx=runr}', "");
Expect(0, 5873, '\p{Scx=runr}', "");
Expect(1, 5873, '\p{^Scx=runr}', "");
Expect(1, 5873, '\P{Scx=runr}', "");
Expect(0, 5873, '\P{^Scx=runr}', "");
Expect(1, 5872, '\p{Scx:    runr}', "");
Expect(0, 5872, '\p{^Scx:    runr}', "");
Expect(0, 5872, '\P{Scx:    runr}', "");
Expect(1, 5872, '\P{^Scx:    runr}', "");
Expect(0, 5873, '\p{Scx:    runr}', "");
Expect(1, 5873, '\p{^Scx:    runr}', "");
Expect(1, 5873, '\P{Scx:    runr}', "");
Expect(0, 5873, '\P{^Scx:    runr}', "");
Error('\p{Is_Script_Extensions=	_Runic/a/}');
Error('\P{Is_Script_Extensions=	_Runic/a/}');
Expect(1, 5872, '\p{Is_Script_Extensions=runic}', "");
Expect(0, 5872, '\p{^Is_Script_Extensions=runic}', "");
Expect(0, 5872, '\P{Is_Script_Extensions=runic}', "");
Expect(1, 5872, '\P{^Is_Script_Extensions=runic}', "");
Expect(0, 5873, '\p{Is_Script_Extensions=runic}', "");
Expect(1, 5873, '\p{^Is_Script_Extensions=runic}', "");
Expect(1, 5873, '\P{Is_Script_Extensions=runic}', "");
Expect(0, 5873, '\P{^Is_Script_Extensions=runic}', "");
Expect(1, 5872, '\p{Is_Script_Extensions=--RUNIC}', "");
Expect(0, 5872, '\p{^Is_Script_Extensions=--RUNIC}', "");
Expect(0, 5872, '\P{Is_Script_Extensions=--RUNIC}', "");
Expect(1, 5872, '\P{^Is_Script_Extensions=--RUNIC}', "");
Expect(0, 5873, '\p{Is_Script_Extensions=--RUNIC}', "");
Expect(1, 5873, '\p{^Is_Script_Extensions=--RUNIC}', "");
Expect(1, 5873, '\P{Is_Script_Extensions=--RUNIC}', "");
Expect(0, 5873, '\P{^Is_Script_Extensions=--RUNIC}', "");
Error('\p{Is_Scx= Runr:=}');
Error('\P{Is_Scx= Runr:=}');
Expect(1, 5872, '\p{Is_Scx=runr}', "");
Expect(0, 5872, '\p{^Is_Scx=runr}', "");
Expect(0, 5872, '\P{Is_Scx=runr}', "");
Expect(1, 5872, '\P{^Is_Scx=runr}', "");
Expect(0, 5873, '\p{Is_Scx=runr}', "");
Expect(1, 5873, '\p{^Is_Scx=runr}', "");
Expect(1, 5873, '\P{Is_Scx=runr}', "");
Expect(0, 5873, '\P{^Is_Scx=runr}', "");
Expect(1, 5872, '\p{Is_Scx:	-_runr}', "");
Expect(0, 5872, '\p{^Is_Scx:	-_runr}', "");
Expect(0, 5872, '\P{Is_Scx:	-_runr}', "");
Expect(1, 5872, '\P{^Is_Scx:	-_runr}', "");
Expect(0, 5873, '\p{Is_Scx:	-_runr}', "");
Expect(1, 5873, '\p{^Is_Scx:	-_runr}', "");
Expect(1, 5873, '\P{Is_Scx:	-_runr}', "");
Expect(0, 5873, '\P{^Is_Scx:	-_runr}', "");
Error('\p{Script_Extensions= 	MEROITIC_Hieroglyphs:=}');
Error('\P{Script_Extensions= 	MEROITIC_Hieroglyphs:=}');
Expect(1, 67999, '\p{Script_Extensions: meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Script_Extensions: meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Script_Extensions: meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Script_Extensions: meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Script_Extensions: meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Script_Extensions: meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Script_Extensions: meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Script_Extensions: meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Script_Extensions=	meroitic_HIEROGLYPHS}', "");
Expect(0, 67999, '\p{^Script_Extensions=	meroitic_HIEROGLYPHS}', "");
Expect(0, 67999, '\P{Script_Extensions=	meroitic_HIEROGLYPHS}', "");
Expect(1, 67999, '\P{^Script_Extensions=	meroitic_HIEROGLYPHS}', "");
Expect(0, 68000, '\p{Script_Extensions=	meroitic_HIEROGLYPHS}', "");
Expect(1, 68000, '\p{^Script_Extensions=	meroitic_HIEROGLYPHS}', "");
Expect(1, 68000, '\P{Script_Extensions=	meroitic_HIEROGLYPHS}', "");
Expect(0, 68000, '\P{^Script_Extensions=	meroitic_HIEROGLYPHS}', "");
Error('\p{Scx=:= -Mero}');
Error('\P{Scx=:= -Mero}');
Expect(1, 67999, '\p{Scx=mero}', "");
Expect(0, 67999, '\p{^Scx=mero}', "");
Expect(0, 67999, '\P{Scx=mero}', "");
Expect(1, 67999, '\P{^Scx=mero}', "");
Expect(0, 68000, '\p{Scx=mero}', "");
Expect(1, 68000, '\p{^Scx=mero}', "");
Expect(1, 68000, '\P{Scx=mero}', "");
Expect(0, 68000, '\P{^Scx=mero}', "");
Expect(1, 67999, '\p{Scx: _	Mero}', "");
Expect(0, 67999, '\p{^Scx: _	Mero}', "");
Expect(0, 67999, '\P{Scx: _	Mero}', "");
Expect(1, 67999, '\P{^Scx: _	Mero}', "");
Expect(0, 68000, '\p{Scx: _	Mero}', "");
Expect(1, 68000, '\p{^Scx: _	Mero}', "");
Expect(1, 68000, '\P{Scx: _	Mero}', "");
Expect(0, 68000, '\P{^Scx: _	Mero}', "");
Error('\p{Is_Script_Extensions=:=		Meroitic_hieroglyphs}');
Error('\P{Is_Script_Extensions=:=		Meroitic_hieroglyphs}');
Expect(1, 67999, '\p{Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Script_Extensions=meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Is_Script_Extensions=_	MEROITIC_Hieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Script_Extensions=_	MEROITIC_Hieroglyphs}', "");
Expect(0, 67999, '\P{Is_Script_Extensions=_	MEROITIC_Hieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Script_Extensions=_	MEROITIC_Hieroglyphs}', "");
Expect(0, 68000, '\p{Is_Script_Extensions=_	MEROITIC_Hieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Script_Extensions=_	MEROITIC_Hieroglyphs}', "");
Expect(1, 68000, '\P{Is_Script_Extensions=_	MEROITIC_Hieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Script_Extensions=_	MEROITIC_Hieroglyphs}', "");
Error('\p{Is_Scx=_ Mero:=}');
Error('\P{Is_Scx=_ Mero:=}');
Expect(1, 67999, '\p{Is_Scx=mero}', "");
Expect(0, 67999, '\p{^Is_Scx=mero}', "");
Expect(0, 67999, '\P{Is_Scx=mero}', "");
Expect(1, 67999, '\P{^Is_Scx=mero}', "");
Expect(0, 68000, '\p{Is_Scx=mero}', "");
Expect(1, 68000, '\p{^Is_Scx=mero}', "");
Expect(1, 68000, '\P{Is_Scx=mero}', "");
Expect(0, 68000, '\P{^Is_Scx=mero}', "");
Expect(1, 67999, '\p{Is_Scx= -mero}', "");
Expect(0, 67999, '\p{^Is_Scx= -mero}', "");
Expect(0, 67999, '\P{Is_Scx= -mero}', "");
Expect(1, 67999, '\P{^Is_Scx= -mero}', "");
Expect(0, 68000, '\p{Is_Scx= -mero}', "");
Expect(1, 68000, '\p{^Is_Scx= -mero}', "");
Expect(1, 68000, '\P{Is_Scx= -mero}', "");
Expect(0, 68000, '\P{^Is_Scx= -mero}', "");
Error('\p{Script_Extensions=	OL_Chiki:=}');
Error('\P{Script_Extensions=	OL_Chiki:=}');
Expect(1, 7295, '\p{Script_Extensions=olchiki}', "");
Expect(0, 7295, '\p{^Script_Extensions=olchiki}', "");
Expect(0, 7295, '\P{Script_Extensions=olchiki}', "");
Expect(1, 7295, '\P{^Script_Extensions=olchiki}', "");
Expect(0, 7296, '\p{Script_Extensions=olchiki}', "");
Expect(1, 7296, '\p{^Script_Extensions=olchiki}', "");
Expect(1, 7296, '\P{Script_Extensions=olchiki}', "");
Expect(0, 7296, '\P{^Script_Extensions=olchiki}', "");
Expect(1, 7295, '\p{Script_Extensions=	_Ol_CHIKI}', "");
Expect(0, 7295, '\p{^Script_Extensions=	_Ol_CHIKI}', "");
Expect(0, 7295, '\P{Script_Extensions=	_Ol_CHIKI}', "");
Expect(1, 7295, '\P{^Script_Extensions=	_Ol_CHIKI}', "");
Expect(0, 7296, '\p{Script_Extensions=	_Ol_CHIKI}', "");
Expect(1, 7296, '\p{^Script_Extensions=	_Ol_CHIKI}', "");
Expect(1, 7296, '\P{Script_Extensions=	_Ol_CHIKI}', "");
Expect(0, 7296, '\P{^Script_Extensions=	_Ol_CHIKI}', "");
Error('\p{Scx=:=Olck}');
Error('\P{Scx=:=Olck}');
Expect(1, 7295, '\p{Scx=olck}', "");
Expect(0, 7295, '\p{^Scx=olck}', "");
Expect(0, 7295, '\P{Scx=olck}', "");
Expect(1, 7295, '\P{^Scx=olck}', "");
Expect(0, 7296, '\p{Scx=olck}', "");
Expect(1, 7296, '\p{^Scx=olck}', "");
Expect(1, 7296, '\P{Scx=olck}', "");
Expect(0, 7296, '\P{^Scx=olck}', "");
Expect(1, 7295, '\p{Scx=_-Olck}', "");
Expect(0, 7295, '\p{^Scx=_-Olck}', "");
Expect(0, 7295, '\P{Scx=_-Olck}', "");
Expect(1, 7295, '\P{^Scx=_-Olck}', "");
Expect(0, 7296, '\p{Scx=_-Olck}', "");
Expect(1, 7296, '\p{^Scx=_-Olck}', "");
Expect(1, 7296, '\P{Scx=_-Olck}', "");
Expect(0, 7296, '\P{^Scx=_-Olck}', "");
Error('\p{Is_Script_Extensions=-/a/Ol_CHIKI}');
Error('\P{Is_Script_Extensions=-/a/Ol_CHIKI}');
Expect(1, 7295, '\p{Is_Script_Extensions=olchiki}', "");
Expect(0, 7295, '\p{^Is_Script_Extensions=olchiki}', "");
Expect(0, 7295, '\P{Is_Script_Extensions=olchiki}', "");
Expect(1, 7295, '\P{^Is_Script_Extensions=olchiki}', "");
Expect(0, 7296, '\p{Is_Script_Extensions=olchiki}', "");
Expect(1, 7296, '\p{^Is_Script_Extensions=olchiki}', "");
Expect(1, 7296, '\P{Is_Script_Extensions=olchiki}', "");
Expect(0, 7296, '\P{^Is_Script_Extensions=olchiki}', "");
Expect(1, 7295, '\p{Is_Script_Extensions=_-Ol_Chiki}', "");
Expect(0, 7295, '\p{^Is_Script_Extensions=_-Ol_Chiki}', "");
Expect(0, 7295, '\P{Is_Script_Extensions=_-Ol_Chiki}', "");
Expect(1, 7295, '\P{^Is_Script_Extensions=_-Ol_Chiki}', "");
Expect(0, 7296, '\p{Is_Script_Extensions=_-Ol_Chiki}', "");
Expect(1, 7296, '\p{^Is_Script_Extensions=_-Ol_Chiki}', "");
Expect(1, 7296, '\P{Is_Script_Extensions=_-Ol_Chiki}', "");
Expect(0, 7296, '\P{^Is_Script_Extensions=_-Ol_Chiki}', "");
Error('\p{Is_Scx=_:=Olck}');
Error('\P{Is_Scx=_:=Olck}');
Expect(1, 7295, '\p{Is_Scx=olck}', "");
Expect(0, 7295, '\p{^Is_Scx=olck}', "");
Expect(0, 7295, '\P{Is_Scx=olck}', "");
Expect(1, 7295, '\P{^Is_Scx=olck}', "");
Expect(0, 7296, '\p{Is_Scx=olck}', "");
Expect(1, 7296, '\p{^Is_Scx=olck}', "");
Expect(1, 7296, '\P{Is_Scx=olck}', "");
Expect(0, 7296, '\P{^Is_Scx=olck}', "");
Expect(1, 7295, '\p{Is_Scx=		Olck}', "");
Expect(0, 7295, '\p{^Is_Scx=		Olck}', "");
Expect(0, 7295, '\P{Is_Scx=		Olck}', "");
Expect(1, 7295, '\P{^Is_Scx=		Olck}', "");
Expect(0, 7296, '\p{Is_Scx=		Olck}', "");
Expect(1, 7296, '\p{^Is_Scx=		Olck}', "");
Expect(1, 7296, '\P{Is_Scx=		Olck}', "");
Expect(0, 7296, '\P{^Is_Scx=		Olck}', "");
Error('\p{Script_Extensions=_Phags_PA/a/}');
Error('\P{Script_Extensions=_Phags_PA/a/}');
Expect(1, 43127, '\p{Script_Extensions=phagspa}', "");
Expect(0, 43127, '\p{^Script_Extensions=phagspa}', "");
Expect(0, 43127, '\P{Script_Extensions=phagspa}', "");
Expect(1, 43127, '\P{^Script_Extensions=phagspa}', "");
Expect(0, 43128, '\p{Script_Extensions=phagspa}', "");
Expect(1, 43128, '\p{^Script_Extensions=phagspa}', "");
Expect(1, 43128, '\P{Script_Extensions=phagspa}', "");
Expect(0, 43128, '\P{^Script_Extensions=phagspa}', "");
Expect(1, 43127, '\p{Script_Extensions=-_PHAGS_Pa}', "");
Expect(0, 43127, '\p{^Script_Extensions=-_PHAGS_Pa}', "");
Expect(0, 43127, '\P{Script_Extensions=-_PHAGS_Pa}', "");
Expect(1, 43127, '\P{^Script_Extensions=-_PHAGS_Pa}', "");
Expect(0, 43128, '\p{Script_Extensions=-_PHAGS_Pa}', "");
Expect(1, 43128, '\p{^Script_Extensions=-_PHAGS_Pa}', "");
Expect(1, 43128, '\P{Script_Extensions=-_PHAGS_Pa}', "");
Expect(0, 43128, '\P{^Script_Extensions=-_PHAGS_Pa}', "");
Error('\p{Scx=/a/-_PHAG}');
Error('\P{Scx=/a/-_PHAG}');
Expect(1, 43127, '\p{Scx=phag}', "");
Expect(0, 43127, '\p{^Scx=phag}', "");
Expect(0, 43127, '\P{Scx=phag}', "");
Expect(1, 43127, '\P{^Scx=phag}', "");
Expect(0, 43128, '\p{Scx=phag}', "");
Expect(1, 43128, '\p{^Scx=phag}', "");
Expect(1, 43128, '\P{Scx=phag}', "");
Expect(0, 43128, '\P{^Scx=phag}', "");
Expect(1, 43127, '\p{Scx=_PHAG}', "");
Expect(0, 43127, '\p{^Scx=_PHAG}', "");
Expect(0, 43127, '\P{Scx=_PHAG}', "");
Expect(1, 43127, '\P{^Scx=_PHAG}', "");
Expect(0, 43128, '\p{Scx=_PHAG}', "");
Expect(1, 43128, '\p{^Scx=_PHAG}', "");
Expect(1, 43128, '\P{Scx=_PHAG}', "");
Expect(0, 43128, '\P{^Scx=_PHAG}', "");
Error('\p{Is_Script_Extensions=_ Phags_Pa/a/}');
Error('\P{Is_Script_Extensions=_ Phags_Pa/a/}');
Expect(1, 43127, '\p{Is_Script_Extensions:	phagspa}', "");
Expect(0, 43127, '\p{^Is_Script_Extensions:	phagspa}', "");
Expect(0, 43127, '\P{Is_Script_Extensions:	phagspa}', "");
Expect(1, 43127, '\P{^Is_Script_Extensions:	phagspa}', "");
Expect(0, 43128, '\p{Is_Script_Extensions:	phagspa}', "");
Expect(1, 43128, '\p{^Is_Script_Extensions:	phagspa}', "");
Expect(1, 43128, '\P{Is_Script_Extensions:	phagspa}', "");
Expect(0, 43128, '\P{^Is_Script_Extensions:	phagspa}', "");
Expect(1, 43127, '\p{Is_Script_Extensions= _phags_pa}', "");
Expect(0, 43127, '\p{^Is_Script_Extensions= _phags_pa}', "");
Expect(0, 43127, '\P{Is_Script_Extensions= _phags_pa}', "");
Expect(1, 43127, '\P{^Is_Script_Extensions= _phags_pa}', "");
Expect(0, 43128, '\p{Is_Script_Extensions= _phags_pa}', "");
Expect(1, 43128, '\p{^Is_Script_Extensions= _phags_pa}', "");
Expect(1, 43128, '\P{Is_Script_Extensions= _phags_pa}', "");
Expect(0, 43128, '\P{^Is_Script_Extensions= _phags_pa}', "");
Error('\p{Is_Scx=_/a/Phag}');
Error('\P{Is_Scx=_/a/Phag}');
Expect(1, 43127, '\p{Is_Scx: phag}', "");
Expect(0, 43127, '\p{^Is_Scx: phag}', "");
Expect(0, 43127, '\P{Is_Scx: phag}', "");
Expect(1, 43127, '\P{^Is_Scx: phag}', "");
Expect(0, 43128, '\p{Is_Scx: phag}', "");
Expect(1, 43128, '\p{^Is_Scx: phag}', "");
Expect(1, 43128, '\P{Is_Scx: phag}', "");
Expect(0, 43128, '\P{^Is_Scx: phag}', "");
Expect(1, 43127, '\p{Is_Scx: 	_Phag}', "");
Expect(0, 43127, '\p{^Is_Scx: 	_Phag}', "");
Expect(0, 43127, '\P{Is_Scx: 	_Phag}', "");
Expect(1, 43127, '\P{^Is_Scx: 	_Phag}', "");
Expect(0, 43128, '\p{Is_Scx: 	_Phag}', "");
Expect(1, 43128, '\p{^Is_Scx: 	_Phag}', "");
Expect(1, 43128, '\P{Is_Scx: 	_Phag}', "");
Expect(0, 43128, '\P{^Is_Scx: 	_Phag}', "");
Error('\p{Script_Extensions=_:=Ugaritic}');
Error('\P{Script_Extensions=_:=Ugaritic}');
Expect(1, 66463, '\p{Script_Extensions:   ugaritic}', "");
Expect(0, 66463, '\p{^Script_Extensions:   ugaritic}', "");
Expect(0, 66463, '\P{Script_Extensions:   ugaritic}', "");
Expect(1, 66463, '\P{^Script_Extensions:   ugaritic}', "");
Expect(0, 66464, '\p{Script_Extensions:   ugaritic}', "");
Expect(1, 66464, '\p{^Script_Extensions:   ugaritic}', "");
Expect(1, 66464, '\P{Script_Extensions:   ugaritic}', "");
Expect(0, 66464, '\P{^Script_Extensions:   ugaritic}', "");
Expect(1, 66463, '\p{Script_Extensions=-	UGARITIC}', "");
Expect(0, 66463, '\p{^Script_Extensions=-	UGARITIC}', "");
Expect(0, 66463, '\P{Script_Extensions=-	UGARITIC}', "");
Expect(1, 66463, '\P{^Script_Extensions=-	UGARITIC}', "");
Expect(0, 66464, '\p{Script_Extensions=-	UGARITIC}', "");
Expect(1, 66464, '\p{^Script_Extensions=-	UGARITIC}', "");
Expect(1, 66464, '\P{Script_Extensions=-	UGARITIC}', "");
Expect(0, 66464, '\P{^Script_Extensions=-	UGARITIC}', "");
Error('\p{Scx=_ Ugar/a/}');
Error('\P{Scx=_ Ugar/a/}');
Expect(1, 66463, '\p{Scx:   ugar}', "");
Expect(0, 66463, '\p{^Scx:   ugar}', "");
Expect(0, 66463, '\P{Scx:   ugar}', "");
Expect(1, 66463, '\P{^Scx:   ugar}', "");
Expect(0, 66464, '\p{Scx:   ugar}', "");
Expect(1, 66464, '\p{^Scx:   ugar}', "");
Expect(1, 66464, '\P{Scx:   ugar}', "");
Expect(0, 66464, '\P{^Scx:   ugar}', "");
Expect(1, 66463, '\p{Scx=	UGAR}', "");
Expect(0, 66463, '\p{^Scx=	UGAR}', "");
Expect(0, 66463, '\P{Scx=	UGAR}', "");
Expect(1, 66463, '\P{^Scx=	UGAR}', "");
Expect(0, 66464, '\p{Scx=	UGAR}', "");
Expect(1, 66464, '\p{^Scx=	UGAR}', "");
Expect(1, 66464, '\P{Scx=	UGAR}', "");
Expect(0, 66464, '\P{^Scx=	UGAR}', "");
Error('\p{Is_Script_Extensions: _Ugaritic/a/}');
Error('\P{Is_Script_Extensions: _Ugaritic/a/}');
Expect(1, 66463, '\p{Is_Script_Extensions=ugaritic}', "");
Expect(0, 66463, '\p{^Is_Script_Extensions=ugaritic}', "");
Expect(0, 66463, '\P{Is_Script_Extensions=ugaritic}', "");
Expect(1, 66463, '\P{^Is_Script_Extensions=ugaritic}', "");
Expect(0, 66464, '\p{Is_Script_Extensions=ugaritic}', "");
Expect(1, 66464, '\p{^Is_Script_Extensions=ugaritic}', "");
Expect(1, 66464, '\P{Is_Script_Extensions=ugaritic}', "");
Expect(0, 66464, '\P{^Is_Script_Extensions=ugaritic}', "");
Expect(1, 66463, '\p{Is_Script_Extensions=_ugaritic}', "");
Expect(0, 66463, '\p{^Is_Script_Extensions=_ugaritic}', "");
Expect(0, 66463, '\P{Is_Script_Extensions=_ugaritic}', "");
Expect(1, 66463, '\P{^Is_Script_Extensions=_ugaritic}', "");
Expect(0, 66464, '\p{Is_Script_Extensions=_ugaritic}', "");
Expect(1, 66464, '\p{^Is_Script_Extensions=_ugaritic}', "");
Expect(1, 66464, '\P{Is_Script_Extensions=_ugaritic}', "");
Expect(0, 66464, '\P{^Is_Script_Extensions=_ugaritic}', "");
Error('\p{Is_Scx=-	Ugar/a/}');
Error('\P{Is_Scx=-	Ugar/a/}');
Expect(1, 66463, '\p{Is_Scx: ugar}', "");
Expect(0, 66463, '\p{^Is_Scx: ugar}', "");
Expect(0, 66463, '\P{Is_Scx: ugar}', "");
Expect(1, 66463, '\P{^Is_Scx: ugar}', "");
Expect(0, 66464, '\p{Is_Scx: ugar}', "");
Expect(1, 66464, '\p{^Is_Scx: ugar}', "");
Expect(1, 66464, '\P{Is_Scx: ugar}', "");
Expect(0, 66464, '\P{^Is_Scx: ugar}', "");
Expect(1, 66463, '\p{Is_Scx=	_UGAR}', "");
Expect(0, 66463, '\p{^Is_Scx=	_UGAR}', "");
Expect(0, 66463, '\P{Is_Scx=	_UGAR}', "");
Expect(1, 66463, '\P{^Is_Scx=	_UGAR}', "");
Expect(0, 66464, '\p{Is_Scx=	_UGAR}', "");
Expect(1, 66464, '\p{^Is_Scx=	_UGAR}', "");
Expect(1, 66464, '\P{Is_Scx=	_UGAR}', "");
Expect(0, 66464, '\P{^Is_Scx=	_UGAR}', "");
Error('\p{Script_Extensions=-_Armenian:=}');
Error('\P{Script_Extensions=-_Armenian:=}');
Expect(1, 64279, '\p{Script_Extensions=armenian}', "");
Expect(0, 64279, '\p{^Script_Extensions=armenian}', "");
Expect(0, 64279, '\P{Script_Extensions=armenian}', "");
Expect(1, 64279, '\P{^Script_Extensions=armenian}', "");
Expect(0, 64280, '\p{Script_Extensions=armenian}', "");
Expect(1, 64280, '\p{^Script_Extensions=armenian}', "");
Expect(1, 64280, '\P{Script_Extensions=armenian}', "");
Expect(0, 64280, '\P{^Script_Extensions=armenian}', "");
Expect(1, 64279, '\p{Script_Extensions=-_ARMENIAN}', "");
Expect(0, 64279, '\p{^Script_Extensions=-_ARMENIAN}', "");
Expect(0, 64279, '\P{Script_Extensions=-_ARMENIAN}', "");
Expect(1, 64279, '\P{^Script_Extensions=-_ARMENIAN}', "");
Expect(0, 64280, '\p{Script_Extensions=-_ARMENIAN}', "");
Expect(1, 64280, '\p{^Script_Extensions=-_ARMENIAN}', "");
Expect(1, 64280, '\P{Script_Extensions=-_ARMENIAN}', "");
Expect(0, 64280, '\P{^Script_Extensions=-_ARMENIAN}', "");
Error('\p{Scx::=Armn}');
Error('\P{Scx::=Armn}');
Expect(1, 64279, '\p{Scx=armn}', "");
Expect(0, 64279, '\p{^Scx=armn}', "");
Expect(0, 64279, '\P{Scx=armn}', "");
Expect(1, 64279, '\P{^Scx=armn}', "");
Expect(0, 64280, '\p{Scx=armn}', "");
Expect(1, 64280, '\p{^Scx=armn}', "");
Expect(1, 64280, '\P{Scx=armn}', "");
Expect(0, 64280, '\P{^Scx=armn}', "");
Expect(1, 64279, '\p{Scx=-Armn}', "");
Expect(0, 64279, '\p{^Scx=-Armn}', "");
Expect(0, 64279, '\P{Scx=-Armn}', "");
Expect(1, 64279, '\P{^Scx=-Armn}', "");
Expect(0, 64280, '\p{Scx=-Armn}', "");
Expect(1, 64280, '\p{^Scx=-Armn}', "");
Expect(1, 64280, '\P{Scx=-Armn}', "");
Expect(0, 64280, '\P{^Scx=-Armn}', "");
Error('\p{Is_Script_Extensions=:=_ARMENIAN}');
Error('\P{Is_Script_Extensions=:=_ARMENIAN}');
Expect(1, 64279, '\p{Is_Script_Extensions=armenian}', "");
Expect(0, 64279, '\p{^Is_Script_Extensions=armenian}', "");
Expect(0, 64279, '\P{Is_Script_Extensions=armenian}', "");
Expect(1, 64279, '\P{^Is_Script_Extensions=armenian}', "");
Expect(0, 64280, '\p{Is_Script_Extensions=armenian}', "");
Expect(1, 64280, '\p{^Is_Script_Extensions=armenian}', "");
Expect(1, 64280, '\P{Is_Script_Extensions=armenian}', "");
Expect(0, 64280, '\P{^Is_Script_Extensions=armenian}', "");
Expect(1, 64279, '\p{Is_Script_Extensions=	_Armenian}', "");
Expect(0, 64279, '\p{^Is_Script_Extensions=	_Armenian}', "");
Expect(0, 64279, '\P{Is_Script_Extensions=	_Armenian}', "");
Expect(1, 64279, '\P{^Is_Script_Extensions=	_Armenian}', "");
Expect(0, 64280, '\p{Is_Script_Extensions=	_Armenian}', "");
Expect(1, 64280, '\p{^Is_Script_Extensions=	_Armenian}', "");
Expect(1, 64280, '\P{Is_Script_Extensions=	_Armenian}', "");
Expect(0, 64280, '\P{^Is_Script_Extensions=	_Armenian}', "");
Error('\p{Is_Scx=	 armn/a/}');
Error('\P{Is_Scx=	 armn/a/}');
Expect(1, 64279, '\p{Is_Scx=armn}', "");
Expect(0, 64279, '\p{^Is_Scx=armn}', "");
Expect(0, 64279, '\P{Is_Scx=armn}', "");
Expect(1, 64279, '\P{^Is_Scx=armn}', "");
Expect(0, 64280, '\p{Is_Scx=armn}', "");
Expect(1, 64280, '\p{^Is_Scx=armn}', "");
Expect(1, 64280, '\P{Is_Scx=armn}', "");
Expect(0, 64280, '\P{^Is_Scx=armn}', "");
Expect(1, 64279, '\p{Is_Scx=  ARMN}', "");
Expect(0, 64279, '\p{^Is_Scx=  ARMN}', "");
Expect(0, 64279, '\P{Is_Scx=  ARMN}', "");
Expect(1, 64279, '\P{^Is_Scx=  ARMN}', "");
Expect(0, 64280, '\p{Is_Scx=  ARMN}', "");
Expect(1, 64280, '\p{^Is_Scx=  ARMN}', "");
Expect(1, 64280, '\P{Is_Scx=  ARMN}', "");
Expect(0, 64280, '\P{^Is_Scx=  ARMN}', "");
Error('\p{Script_Extensions= -SYRIAC/a/}');
Error('\P{Script_Extensions= -SYRIAC/a/}');
Expect(1, 1871, '\p{Script_Extensions=syriac}', "");
Expect(0, 1871, '\p{^Script_Extensions=syriac}', "");
Expect(0, 1871, '\P{Script_Extensions=syriac}', "");
Expect(1, 1871, '\P{^Script_Extensions=syriac}', "");
Expect(0, 1872, '\p{Script_Extensions=syriac}', "");
Expect(1, 1872, '\p{^Script_Extensions=syriac}', "");
Expect(1, 1872, '\P{Script_Extensions=syriac}', "");
Expect(0, 1872, '\P{^Script_Extensions=syriac}', "");
Expect(1, 1871, '\p{Script_Extensions:     syriac}', "");
Expect(0, 1871, '\p{^Script_Extensions:     syriac}', "");
Expect(0, 1871, '\P{Script_Extensions:     syriac}', "");
Expect(1, 1871, '\P{^Script_Extensions:     syriac}', "");
Expect(0, 1872, '\p{Script_Extensions:     syriac}', "");
Expect(1, 1872, '\p{^Script_Extensions:     syriac}', "");
Expect(1, 1872, '\P{Script_Extensions:     syriac}', "");
Expect(0, 1872, '\P{^Script_Extensions:     syriac}', "");
Error('\p{Scx:_/a/SYRC}');
Error('\P{Scx:_/a/SYRC}');
Expect(1, 1871, '\p{Scx=syrc}', "");
Expect(0, 1871, '\p{^Scx=syrc}', "");
Expect(0, 1871, '\P{Scx=syrc}', "");
Expect(1, 1871, '\P{^Scx=syrc}', "");
Expect(0, 1872, '\p{Scx=syrc}', "");
Expect(1, 1872, '\p{^Scx=syrc}', "");
Expect(1, 1872, '\P{Scx=syrc}', "");
Expect(0, 1872, '\P{^Scx=syrc}', "");
Expect(1, 1871, '\p{Scx=		syrc}', "");
Expect(0, 1871, '\p{^Scx=		syrc}', "");
Expect(0, 1871, '\P{Scx=		syrc}', "");
Expect(1, 1871, '\P{^Scx=		syrc}', "");
Expect(0, 1872, '\p{Scx=		syrc}', "");
Expect(1, 1872, '\p{^Scx=		syrc}', "");
Expect(1, 1872, '\P{Scx=		syrc}', "");
Expect(0, 1872, '\P{^Scx=		syrc}', "");
Error('\p{Is_Script_Extensions:    -Syriac:=}');
Error('\P{Is_Script_Extensions:    -Syriac:=}');
Expect(1, 1871, '\p{Is_Script_Extensions:	syriac}', "");
Expect(0, 1871, '\p{^Is_Script_Extensions:	syriac}', "");
Expect(0, 1871, '\P{Is_Script_Extensions:	syriac}', "");
Expect(1, 1871, '\P{^Is_Script_Extensions:	syriac}', "");
Expect(0, 1872, '\p{Is_Script_Extensions:	syriac}', "");
Expect(1, 1872, '\p{^Is_Script_Extensions:	syriac}', "");
Expect(1, 1872, '\P{Is_Script_Extensions:	syriac}', "");
Expect(0, 1872, '\P{^Is_Script_Extensions:	syriac}', "");
Expect(1, 1871, '\p{Is_Script_Extensions=- Syriac}', "");
Expect(0, 1871, '\p{^Is_Script_Extensions=- Syriac}', "");
Expect(0, 1871, '\P{Is_Script_Extensions=- Syriac}', "");
Expect(1, 1871, '\P{^Is_Script_Extensions=- Syriac}', "");
Expect(0, 1872, '\p{Is_Script_Extensions=- Syriac}', "");
Expect(1, 1872, '\p{^Is_Script_Extensions=- Syriac}', "");
Expect(1, 1872, '\P{Is_Script_Extensions=- Syriac}', "");
Expect(0, 1872, '\P{^Is_Script_Extensions=- Syriac}', "");
Error('\p{Is_Scx=	/a/Syrc}');
Error('\P{Is_Scx=	/a/Syrc}');
Expect(1, 1871, '\p{Is_Scx=syrc}', "");
Expect(0, 1871, '\p{^Is_Scx=syrc}', "");
Expect(0, 1871, '\P{Is_Scx=syrc}', "");
Expect(1, 1871, '\P{^Is_Scx=syrc}', "");
Expect(0, 1872, '\p{Is_Scx=syrc}', "");
Expect(1, 1872, '\p{^Is_Scx=syrc}', "");
Expect(1, 1872, '\P{Is_Scx=syrc}', "");
Expect(0, 1872, '\P{^Is_Scx=syrc}', "");
Expect(1, 1871, '\p{Is_Scx=_-Syrc}', "");
Expect(0, 1871, '\p{^Is_Scx=_-Syrc}', "");
Expect(0, 1871, '\P{Is_Scx=_-Syrc}', "");
Expect(1, 1871, '\P{^Is_Scx=_-Syrc}', "");
Expect(0, 1872, '\p{Is_Scx=_-Syrc}', "");
Expect(1, 1872, '\p{^Is_Scx=_-Syrc}', "");
Expect(1, 1872, '\P{Is_Scx=_-Syrc}', "");
Expect(0, 1872, '\P{^Is_Scx=_-Syrc}', "");
Error('\p{Script_Extensions=:= -katakana}');
Error('\P{Script_Extensions=:= -katakana}');
Expect(1, 110592, '\p{Script_Extensions=katakana}', "");
Expect(0, 110592, '\p{^Script_Extensions=katakana}', "");
Expect(0, 110592, '\P{Script_Extensions=katakana}', "");
Expect(1, 110592, '\P{^Script_Extensions=katakana}', "");
Expect(0, 110593, '\p{Script_Extensions=katakana}', "");
Expect(1, 110593, '\p{^Script_Extensions=katakana}', "");
Expect(1, 110593, '\P{Script_Extensions=katakana}', "");
Expect(0, 110593, '\P{^Script_Extensions=katakana}', "");
Expect(1, 110592, '\p{Script_Extensions=-_KATAKANA}', "");
Expect(0, 110592, '\p{^Script_Extensions=-_KATAKANA}', "");
Expect(0, 110592, '\P{Script_Extensions=-_KATAKANA}', "");
Expect(1, 110592, '\P{^Script_Extensions=-_KATAKANA}', "");
Expect(0, 110593, '\p{Script_Extensions=-_KATAKANA}', "");
Expect(1, 110593, '\p{^Script_Extensions=-_KATAKANA}', "");
Expect(1, 110593, '\P{Script_Extensions=-_KATAKANA}', "");
Expect(0, 110593, '\P{^Script_Extensions=-_KATAKANA}', "");
Error('\p{Scx=KANA/a/}');
Error('\P{Scx=KANA/a/}');
Expect(1, 110592, '\p{Scx=kana}', "");
Expect(0, 110592, '\p{^Scx=kana}', "");
Expect(0, 110592, '\P{Scx=kana}', "");
Expect(1, 110592, '\P{^Scx=kana}', "");
Expect(0, 110593, '\p{Scx=kana}', "");
Expect(1, 110593, '\p{^Scx=kana}', "");
Expect(1, 110593, '\P{Scx=kana}', "");
Expect(0, 110593, '\P{^Scx=kana}', "");
Expect(1, 110592, '\p{Scx= -Kana}', "");
Expect(0, 110592, '\p{^Scx= -Kana}', "");
Expect(0, 110592, '\P{Scx= -Kana}', "");
Expect(1, 110592, '\P{^Scx= -Kana}', "");
Expect(0, 110593, '\p{Scx= -Kana}', "");
Expect(1, 110593, '\p{^Scx= -Kana}', "");
Expect(1, 110593, '\P{Scx= -Kana}', "");
Expect(0, 110593, '\P{^Scx= -Kana}', "");
Error('\p{Is_Script_Extensions= 	Katakana/a/}');
Error('\P{Is_Script_Extensions= 	Katakana/a/}');
Expect(1, 110592, '\p{Is_Script_Extensions=katakana}', "");
Expect(0, 110592, '\p{^Is_Script_Extensions=katakana}', "");
Expect(0, 110592, '\P{Is_Script_Extensions=katakana}', "");
Expect(1, 110592, '\P{^Is_Script_Extensions=katakana}', "");
Expect(0, 110593, '\p{Is_Script_Extensions=katakana}', "");
Expect(1, 110593, '\p{^Is_Script_Extensions=katakana}', "");
Expect(1, 110593, '\P{Is_Script_Extensions=katakana}', "");
Expect(0, 110593, '\P{^Is_Script_Extensions=katakana}', "");
Expect(1, 110592, '\p{Is_Script_Extensions=	Katakana}', "");
Expect(0, 110592, '\p{^Is_Script_Extensions=	Katakana}', "");
Expect(0, 110592, '\P{Is_Script_Extensions=	Katakana}', "");
Expect(1, 110592, '\P{^Is_Script_Extensions=	Katakana}', "");
Expect(0, 110593, '\p{Is_Script_Extensions=	Katakana}', "");
Expect(1, 110593, '\p{^Is_Script_Extensions=	Katakana}', "");
Expect(1, 110593, '\P{Is_Script_Extensions=	Katakana}', "");
Expect(0, 110593, '\P{^Is_Script_Extensions=	Katakana}', "");
Error('\p{Is_Scx=:=	-kana}');
Error('\P{Is_Scx=:=	-kana}');
Expect(1, 110592, '\p{Is_Scx=kana}', "");
Expect(0, 110592, '\p{^Is_Scx=kana}', "");
Expect(0, 110592, '\P{Is_Scx=kana}', "");
Expect(1, 110592, '\P{^Is_Scx=kana}', "");
Expect(0, 110593, '\p{Is_Scx=kana}', "");
Expect(1, 110593, '\p{^Is_Scx=kana}', "");
Expect(1, 110593, '\P{Is_Scx=kana}', "");
Expect(0, 110593, '\P{^Is_Scx=kana}', "");
Expect(1, 110592, '\p{Is_Scx=	-Kana}', "");
Expect(0, 110592, '\p{^Is_Scx=	-Kana}', "");
Expect(0, 110592, '\P{Is_Scx=	-Kana}', "");
Expect(1, 110592, '\P{^Is_Scx=	-Kana}', "");
Expect(0, 110593, '\p{Is_Scx=	-Kana}', "");
Expect(1, 110593, '\p{^Is_Scx=	-Kana}', "");
Expect(1, 110593, '\P{Is_Scx=	-Kana}', "");
Expect(0, 110593, '\P{^Is_Scx=	-Kana}', "");
Error('\p{Script_Extensions=	/a/Yi}');
Error('\P{Script_Extensions=	/a/Yi}');
Expect(1, 65381, '\p{Script_Extensions=yi}', "");
Expect(0, 65381, '\p{^Script_Extensions=yi}', "");
Expect(0, 65381, '\P{Script_Extensions=yi}', "");
Expect(1, 65381, '\P{^Script_Extensions=yi}', "");
Expect(0, 65382, '\p{Script_Extensions=yi}', "");
Expect(1, 65382, '\p{^Script_Extensions=yi}', "");
Expect(1, 65382, '\P{Script_Extensions=yi}', "");
Expect(0, 65382, '\P{^Script_Extensions=yi}', "");
Expect(1, 65381, '\p{Script_Extensions=	 Yi}', "");
Expect(0, 65381, '\p{^Script_Extensions=	 Yi}', "");
Expect(0, 65381, '\P{Script_Extensions=	 Yi}', "");
Expect(1, 65381, '\P{^Script_Extensions=	 Yi}', "");
Expect(0, 65382, '\p{Script_Extensions=	 Yi}', "");
Expect(1, 65382, '\p{^Script_Extensions=	 Yi}', "");
Expect(1, 65382, '\P{Script_Extensions=	 Yi}', "");
Expect(0, 65382, '\P{^Script_Extensions=	 Yi}', "");
Error('\p{Scx=:= YIII}');
Error('\P{Scx=:= YIII}');
Expect(1, 65381, '\p{Scx:   yiii}', "");
Expect(0, 65381, '\p{^Scx:   yiii}', "");
Expect(0, 65381, '\P{Scx:   yiii}', "");
Expect(1, 65381, '\P{^Scx:   yiii}', "");
Expect(0, 65382, '\p{Scx:   yiii}', "");
Expect(1, 65382, '\p{^Scx:   yiii}', "");
Expect(1, 65382, '\P{Scx:   yiii}', "");
Expect(0, 65382, '\P{^Scx:   yiii}', "");
Expect(1, 65381, '\p{Scx=-yiii}', "");
Expect(0, 65381, '\p{^Scx=-yiii}', "");
Expect(0, 65381, '\P{Scx=-yiii}', "");
Expect(1, 65381, '\P{^Scx=-yiii}', "");
Expect(0, 65382, '\p{Scx=-yiii}', "");
Expect(1, 65382, '\p{^Scx=-yiii}', "");
Expect(1, 65382, '\P{Scx=-yiii}', "");
Expect(0, 65382, '\P{^Scx=-yiii}', "");
Error('\p{Is_Script_Extensions=:=Yi}');
Error('\P{Is_Script_Extensions=:=Yi}');
Expect(1, 65381, '\p{Is_Script_Extensions=yi}', "");
Expect(0, 65381, '\p{^Is_Script_Extensions=yi}', "");
Expect(0, 65381, '\P{Is_Script_Extensions=yi}', "");
Expect(1, 65381, '\P{^Is_Script_Extensions=yi}', "");
Expect(0, 65382, '\p{Is_Script_Extensions=yi}', "");
Expect(1, 65382, '\p{^Is_Script_Extensions=yi}', "");
Expect(1, 65382, '\P{Is_Script_Extensions=yi}', "");
Expect(0, 65382, '\P{^Is_Script_Extensions=yi}', "");
Expect(1, 65381, '\p{Is_Script_Extensions= 	YI}', "");
Expect(0, 65381, '\p{^Is_Script_Extensions= 	YI}', "");
Expect(0, 65381, '\P{Is_Script_Extensions= 	YI}', "");
Expect(1, 65381, '\P{^Is_Script_Extensions= 	YI}', "");
Expect(0, 65382, '\p{Is_Script_Extensions= 	YI}', "");
Expect(1, 65382, '\p{^Is_Script_Extensions= 	YI}', "");
Expect(1, 65382, '\P{Is_Script_Extensions= 	YI}', "");
Expect(0, 65382, '\P{^Is_Script_Extensions= 	YI}', "");
Error('\p{Is_Scx=	:=yiii}');
Error('\P{Is_Scx=	:=yiii}');
Expect(1, 65381, '\p{Is_Scx=yiii}', "");
Expect(0, 65381, '\p{^Is_Scx=yiii}', "");
Expect(0, 65381, '\P{Is_Scx=yiii}', "");
Expect(1, 65381, '\P{^Is_Scx=yiii}', "");
Expect(0, 65382, '\p{Is_Scx=yiii}', "");
Expect(1, 65382, '\p{^Is_Scx=yiii}', "");
Expect(1, 65382, '\P{Is_Scx=yiii}', "");
Expect(0, 65382, '\P{^Is_Scx=yiii}', "");
Expect(1, 65381, '\p{Is_Scx=-yiii}', "");
Expect(0, 65381, '\p{^Is_Scx=-yiii}', "");
Expect(0, 65381, '\P{Is_Scx=-yiii}', "");
Expect(1, 65381, '\P{^Is_Scx=-yiii}', "");
Expect(0, 65382, '\p{Is_Scx=-yiii}', "");
Expect(1, 65382, '\p{^Is_Scx=-yiii}', "");
Expect(1, 65382, '\P{Is_Scx=-yiii}', "");
Expect(0, 65382, '\P{^Is_Scx=-yiii}', "");
Error('\p{Script_Extensions=/a/_GLAGOLITIC}');
Error('\P{Script_Extensions=/a/_GLAGOLITIC}');
Expect(1, 11358, '\p{Script_Extensions=glagolitic}', "");
Expect(0, 11358, '\p{^Script_Extensions=glagolitic}', "");
Expect(0, 11358, '\P{Script_Extensions=glagolitic}', "");
Expect(1, 11358, '\P{^Script_Extensions=glagolitic}', "");
Expect(0, 11359, '\p{Script_Extensions=glagolitic}', "");
Expect(1, 11359, '\p{^Script_Extensions=glagolitic}', "");
Expect(1, 11359, '\P{Script_Extensions=glagolitic}', "");
Expect(0, 11359, '\P{^Script_Extensions=glagolitic}', "");
Expect(1, 11358, '\p{Script_Extensions=- GLAGOLITIC}', "");
Expect(0, 11358, '\p{^Script_Extensions=- GLAGOLITIC}', "");
Expect(0, 11358, '\P{Script_Extensions=- GLAGOLITIC}', "");
Expect(1, 11358, '\P{^Script_Extensions=- GLAGOLITIC}', "");
Expect(0, 11359, '\p{Script_Extensions=- GLAGOLITIC}', "");
Expect(1, 11359, '\p{^Script_Extensions=- GLAGOLITIC}', "");
Expect(1, 11359, '\P{Script_Extensions=- GLAGOLITIC}', "");
Expect(0, 11359, '\P{^Script_Extensions=- GLAGOLITIC}', "");
Error('\p{Scx=  Glag/a/}');
Error('\P{Scx=  Glag/a/}');
Expect(1, 11358, '\p{Scx=glag}', "");
Expect(0, 11358, '\p{^Scx=glag}', "");
Expect(0, 11358, '\P{Scx=glag}', "");
Expect(1, 11358, '\P{^Scx=glag}', "");
Expect(0, 11359, '\p{Scx=glag}', "");
Expect(1, 11359, '\p{^Scx=glag}', "");
Expect(1, 11359, '\P{Scx=glag}', "");
Expect(0, 11359, '\P{^Scx=glag}', "");
Expect(1, 11358, '\p{Scx=-	Glag}', "");
Expect(0, 11358, '\p{^Scx=-	Glag}', "");
Expect(0, 11358, '\P{Scx=-	Glag}', "");
Expect(1, 11358, '\P{^Scx=-	Glag}', "");
Expect(0, 11359, '\p{Scx=-	Glag}', "");
Expect(1, 11359, '\p{^Scx=-	Glag}', "");
Expect(1, 11359, '\P{Scx=-	Glag}', "");
Expect(0, 11359, '\P{^Scx=-	Glag}', "");
Error('\p{Is_Script_Extensions= _GLAGOLITIC:=}');
Error('\P{Is_Script_Extensions= _GLAGOLITIC:=}');
Expect(1, 11358, '\p{Is_Script_Extensions=glagolitic}', "");
Expect(0, 11358, '\p{^Is_Script_Extensions=glagolitic}', "");
Expect(0, 11358, '\P{Is_Script_Extensions=glagolitic}', "");
Expect(1, 11358, '\P{^Is_Script_Extensions=glagolitic}', "");
Expect(0, 11359, '\p{Is_Script_Extensions=glagolitic}', "");
Expect(1, 11359, '\p{^Is_Script_Extensions=glagolitic}', "");
Expect(1, 11359, '\P{Is_Script_Extensions=glagolitic}', "");
Expect(0, 11359, '\P{^Is_Script_Extensions=glagolitic}', "");
Expect(1, 11358, '\p{Is_Script_Extensions=-	Glagolitic}', "");
Expect(0, 11358, '\p{^Is_Script_Extensions=-	Glagolitic}', "");
Expect(0, 11358, '\P{Is_Script_Extensions=-	Glagolitic}', "");
Expect(1, 11358, '\P{^Is_Script_Extensions=-	Glagolitic}', "");
Expect(0, 11359, '\p{Is_Script_Extensions=-	Glagolitic}', "");
Expect(1, 11359, '\p{^Is_Script_Extensions=-	Glagolitic}', "");
Expect(1, 11359, '\P{Is_Script_Extensions=-	Glagolitic}', "");
Expect(0, 11359, '\P{^Is_Script_Extensions=-	Glagolitic}', "");
Error('\p{Is_Scx= 	GLAG/a/}');
Error('\P{Is_Scx= 	GLAG/a/}');
Expect(1, 11358, '\p{Is_Scx=glag}', "");
Expect(0, 11358, '\p{^Is_Scx=glag}', "");
Expect(0, 11358, '\P{Is_Scx=glag}', "");
Expect(1, 11358, '\P{^Is_Scx=glag}', "");
Expect(0, 11359, '\p{Is_Scx=glag}', "");
Expect(1, 11359, '\p{^Is_Scx=glag}', "");
Expect(1, 11359, '\P{Is_Scx=glag}', "");
Expect(0, 11359, '\P{^Is_Scx=glag}', "");
Expect(1, 11358, '\p{Is_Scx=	 Glag}', "");
Expect(0, 11358, '\p{^Is_Scx=	 Glag}', "");
Expect(0, 11358, '\P{Is_Scx=	 Glag}', "");
Expect(1, 11358, '\P{^Is_Scx=	 Glag}', "");
Expect(0, 11359, '\p{Is_Scx=	 Glag}', "");
Expect(1, 11359, '\p{^Is_Scx=	 Glag}', "");
Expect(1, 11359, '\P{Is_Scx=	 Glag}', "");
Expect(0, 11359, '\P{^Is_Scx=	 Glag}', "");
Error('\p{Script_Extensions=	-Lepcha/a/}');
Error('\P{Script_Extensions=	-Lepcha/a/}');
Expect(1, 7247, '\p{Script_Extensions:	lepcha}', "");
Expect(0, 7247, '\p{^Script_Extensions:	lepcha}', "");
Expect(0, 7247, '\P{Script_Extensions:	lepcha}', "");
Expect(1, 7247, '\P{^Script_Extensions:	lepcha}', "");
Expect(0, 7248, '\p{Script_Extensions:	lepcha}', "");
Expect(1, 7248, '\p{^Script_Extensions:	lepcha}', "");
Expect(1, 7248, '\P{Script_Extensions:	lepcha}', "");
Expect(0, 7248, '\P{^Script_Extensions:	lepcha}', "");
Expect(1, 7247, '\p{Script_Extensions=-lepcha}', "");
Expect(0, 7247, '\p{^Script_Extensions=-lepcha}', "");
Expect(0, 7247, '\P{Script_Extensions=-lepcha}', "");
Expect(1, 7247, '\P{^Script_Extensions=-lepcha}', "");
Expect(0, 7248, '\p{Script_Extensions=-lepcha}', "");
Expect(1, 7248, '\p{^Script_Extensions=-lepcha}', "");
Expect(1, 7248, '\P{Script_Extensions=-lepcha}', "");
Expect(0, 7248, '\P{^Script_Extensions=-lepcha}', "");
Error('\p{Scx=/a/-LEPC}');
Error('\P{Scx=/a/-LEPC}');
Expect(1, 7247, '\p{Scx=lepc}', "");
Expect(0, 7247, '\p{^Scx=lepc}', "");
Expect(0, 7247, '\P{Scx=lepc}', "");
Expect(1, 7247, '\P{^Scx=lepc}', "");
Expect(0, 7248, '\p{Scx=lepc}', "");
Expect(1, 7248, '\p{^Scx=lepc}', "");
Expect(1, 7248, '\P{Scx=lepc}', "");
Expect(0, 7248, '\P{^Scx=lepc}', "");
Expect(1, 7247, '\p{Scx=_lepc}', "");
Expect(0, 7247, '\p{^Scx=_lepc}', "");
Expect(0, 7247, '\P{Scx=_lepc}', "");
Expect(1, 7247, '\P{^Scx=_lepc}', "");
Expect(0, 7248, '\p{Scx=_lepc}', "");
Expect(1, 7248, '\p{^Scx=_lepc}', "");
Expect(1, 7248, '\P{Scx=_lepc}', "");
Expect(0, 7248, '\P{^Scx=_lepc}', "");
Error('\p{Is_Script_Extensions=-:=Lepcha}');
Error('\P{Is_Script_Extensions=-:=Lepcha}');
Expect(1, 7247, '\p{Is_Script_Extensions=lepcha}', "");
Expect(0, 7247, '\p{^Is_Script_Extensions=lepcha}', "");
Expect(0, 7247, '\P{Is_Script_Extensions=lepcha}', "");
Expect(1, 7247, '\P{^Is_Script_Extensions=lepcha}', "");
Expect(0, 7248, '\p{Is_Script_Extensions=lepcha}', "");
Expect(1, 7248, '\p{^Is_Script_Extensions=lepcha}', "");
Expect(1, 7248, '\P{Is_Script_Extensions=lepcha}', "");
Expect(0, 7248, '\P{^Is_Script_Extensions=lepcha}', "");
Expect(1, 7247, '\p{Is_Script_Extensions=		LEPCHA}', "");
Expect(0, 7247, '\p{^Is_Script_Extensions=		LEPCHA}', "");
Expect(0, 7247, '\P{Is_Script_Extensions=		LEPCHA}', "");
Expect(1, 7247, '\P{^Is_Script_Extensions=		LEPCHA}', "");
Expect(0, 7248, '\p{Is_Script_Extensions=		LEPCHA}', "");
Expect(1, 7248, '\p{^Is_Script_Extensions=		LEPCHA}', "");
Expect(1, 7248, '\P{Is_Script_Extensions=		LEPCHA}', "");
Expect(0, 7248, '\P{^Is_Script_Extensions=		LEPCHA}', "");
Error('\p{Is_Scx: :=  LEPC}');
Error('\P{Is_Scx: :=  LEPC}');
Expect(1, 7247, '\p{Is_Scx=lepc}', "");
Expect(0, 7247, '\p{^Is_Scx=lepc}', "");
Expect(0, 7247, '\P{Is_Scx=lepc}', "");
Expect(1, 7247, '\P{^Is_Scx=lepc}', "");
Expect(0, 7248, '\p{Is_Scx=lepc}', "");
Expect(1, 7248, '\p{^Is_Scx=lepc}', "");
Expect(1, 7248, '\P{Is_Scx=lepc}', "");
Expect(0, 7248, '\P{^Is_Scx=lepc}', "");
Expect(1, 7247, '\p{Is_Scx= _LEPC}', "");
Expect(0, 7247, '\p{^Is_Scx= _LEPC}', "");
Expect(0, 7247, '\P{Is_Scx= _LEPC}', "");
Expect(1, 7247, '\P{^Is_Scx= _LEPC}', "");
Expect(0, 7248, '\p{Is_Scx= _LEPC}', "");
Expect(1, 7248, '\p{^Is_Scx= _LEPC}', "");
Expect(1, 7248, '\P{Is_Scx= _LEPC}', "");
Expect(0, 7248, '\P{^Is_Scx= _LEPC}', "");
Error('\p{Script_Extensions=_thaana:=}');
Error('\P{Script_Extensions=_thaana:=}');
Expect(1, 65021, '\p{Script_Extensions=thaana}', "");
Expect(0, 65021, '\p{^Script_Extensions=thaana}', "");
Expect(0, 65021, '\P{Script_Extensions=thaana}', "");
Expect(1, 65021, '\P{^Script_Extensions=thaana}', "");
Expect(0, 65022, '\p{Script_Extensions=thaana}', "");
Expect(1, 65022, '\p{^Script_Extensions=thaana}', "");
Expect(1, 65022, '\P{Script_Extensions=thaana}', "");
Expect(0, 65022, '\P{^Script_Extensions=thaana}', "");
Expect(1, 65021, '\p{Script_Extensions=-THAANA}', "");
Expect(0, 65021, '\p{^Script_Extensions=-THAANA}', "");
Expect(0, 65021, '\P{Script_Extensions=-THAANA}', "");
Expect(1, 65021, '\P{^Script_Extensions=-THAANA}', "");
Expect(0, 65022, '\p{Script_Extensions=-THAANA}', "");
Expect(1, 65022, '\p{^Script_Extensions=-THAANA}', "");
Expect(1, 65022, '\P{Script_Extensions=-THAANA}', "");
Expect(0, 65022, '\P{^Script_Extensions=-THAANA}', "");
Error('\p{Scx=--Thaa:=}');
Error('\P{Scx=--Thaa:=}');
Expect(1, 65021, '\p{Scx=thaa}', "");
Expect(0, 65021, '\p{^Scx=thaa}', "");
Expect(0, 65021, '\P{Scx=thaa}', "");
Expect(1, 65021, '\P{^Scx=thaa}', "");
Expect(0, 65022, '\p{Scx=thaa}', "");
Expect(1, 65022, '\p{^Scx=thaa}', "");
Expect(1, 65022, '\P{Scx=thaa}', "");
Expect(0, 65022, '\P{^Scx=thaa}', "");
Expect(1, 65021, '\p{Scx=	 Thaa}', "");
Expect(0, 65021, '\p{^Scx=	 Thaa}', "");
Expect(0, 65021, '\P{Scx=	 Thaa}', "");
Expect(1, 65021, '\P{^Scx=	 Thaa}', "");
Expect(0, 65022, '\p{Scx=	 Thaa}', "");
Expect(1, 65022, '\p{^Scx=	 Thaa}', "");
Expect(1, 65022, '\P{Scx=	 Thaa}', "");
Expect(0, 65022, '\P{^Scx=	 Thaa}', "");
Error('\p{Is_Script_Extensions=:= 	Thaana}');
Error('\P{Is_Script_Extensions=:= 	Thaana}');
Expect(1, 65021, '\p{Is_Script_Extensions=thaana}', "");
Expect(0, 65021, '\p{^Is_Script_Extensions=thaana}', "");
Expect(0, 65021, '\P{Is_Script_Extensions=thaana}', "");
Expect(1, 65021, '\P{^Is_Script_Extensions=thaana}', "");
Expect(0, 65022, '\p{Is_Script_Extensions=thaana}', "");
Expect(1, 65022, '\p{^Is_Script_Extensions=thaana}', "");
Expect(1, 65022, '\P{Is_Script_Extensions=thaana}', "");
Expect(0, 65022, '\P{^Is_Script_Extensions=thaana}', "");
Expect(1, 65021, '\p{Is_Script_Extensions=-thaana}', "");
Expect(0, 65021, '\p{^Is_Script_Extensions=-thaana}', "");
Expect(0, 65021, '\P{Is_Script_Extensions=-thaana}', "");
Expect(1, 65021, '\P{^Is_Script_Extensions=-thaana}', "");
Expect(0, 65022, '\p{Is_Script_Extensions=-thaana}', "");
Expect(1, 65022, '\p{^Is_Script_Extensions=-thaana}', "");
Expect(1, 65022, '\P{Is_Script_Extensions=-thaana}', "");
Expect(0, 65022, '\P{^Is_Script_Extensions=-thaana}', "");
Error('\p{Is_Scx=/a/--thaa}');
Error('\P{Is_Scx=/a/--thaa}');
Expect(1, 65021, '\p{Is_Scx=thaa}', "");
Expect(0, 65021, '\p{^Is_Scx=thaa}', "");
Expect(0, 65021, '\P{Is_Scx=thaa}', "");
Expect(1, 65021, '\P{^Is_Scx=thaa}', "");
Expect(0, 65022, '\p{Is_Scx=thaa}', "");
Expect(1, 65022, '\p{^Is_Scx=thaa}', "");
Expect(1, 65022, '\P{Is_Scx=thaa}', "");
Expect(0, 65022, '\P{^Is_Scx=thaa}', "");
Expect(1, 65021, '\p{Is_Scx=THAA}', "");
Expect(0, 65021, '\p{^Is_Scx=THAA}', "");
Expect(0, 65021, '\P{Is_Scx=THAA}', "");
Expect(1, 65021, '\P{^Is_Scx=THAA}', "");
Expect(0, 65022, '\p{Is_Scx=THAA}', "");
Expect(1, 65022, '\p{^Is_Scx=THAA}', "");
Expect(1, 65022, '\P{Is_Scx=THAA}', "");
Expect(0, 65022, '\P{^Is_Scx=THAA}', "");
Error('\p{Script_Extensions=__Coptic:=}');
Error('\P{Script_Extensions=__Coptic:=}');
Expect(1, 11519, '\p{Script_Extensions=coptic}', "");
Expect(0, 11519, '\p{^Script_Extensions=coptic}', "");
Expect(0, 11519, '\P{Script_Extensions=coptic}', "");
Expect(1, 11519, '\P{^Script_Extensions=coptic}', "");
Expect(0, 11520, '\p{Script_Extensions=coptic}', "");
Expect(1, 11520, '\p{^Script_Extensions=coptic}', "");
Expect(1, 11520, '\P{Script_Extensions=coptic}', "");
Expect(0, 11520, '\P{^Script_Extensions=coptic}', "");
Expect(1, 11519, '\p{Script_Extensions=	-Coptic}', "");
Expect(0, 11519, '\p{^Script_Extensions=	-Coptic}', "");
Expect(0, 11519, '\P{Script_Extensions=	-Coptic}', "");
Expect(1, 11519, '\P{^Script_Extensions=	-Coptic}', "");
Expect(0, 11520, '\p{Script_Extensions=	-Coptic}', "");
Expect(1, 11520, '\p{^Script_Extensions=	-Coptic}', "");
Expect(1, 11520, '\P{Script_Extensions=	-Coptic}', "");
Expect(0, 11520, '\P{^Script_Extensions=	-Coptic}', "");
Error('\p{Scx=/a/ 	Copt}');
Error('\P{Scx=/a/ 	Copt}');
Expect(1, 11519, '\p{Scx=copt}', "");
Expect(0, 11519, '\p{^Scx=copt}', "");
Expect(0, 11519, '\P{Scx=copt}', "");
Expect(1, 11519, '\P{^Scx=copt}', "");
Expect(0, 11520, '\p{Scx=copt}', "");
Expect(1, 11520, '\p{^Scx=copt}', "");
Expect(1, 11520, '\P{Scx=copt}', "");
Expect(0, 11520, '\P{^Scx=copt}', "");
Expect(1, 11519, '\p{Scx=	Copt}', "");
Expect(0, 11519, '\p{^Scx=	Copt}', "");
Expect(0, 11519, '\P{Scx=	Copt}', "");
Expect(1, 11519, '\P{^Scx=	Copt}', "");
Expect(0, 11520, '\p{Scx=	Copt}', "");
Expect(1, 11520, '\p{^Scx=	Copt}', "");
Expect(1, 11520, '\P{Scx=	Copt}', "");
Expect(0, 11520, '\P{^Scx=	Copt}', "");
Error('\p{Is_Script_Extensions= :=Qaac}');
Error('\P{Is_Script_Extensions= :=Qaac}');
Expect(1, 11519, '\p{Is_Script_Extensions=qaac}', "");
Expect(0, 11519, '\p{^Is_Script_Extensions=qaac}', "");
Expect(0, 11519, '\P{Is_Script_Extensions=qaac}', "");
Expect(1, 11519, '\P{^Is_Script_Extensions=qaac}', "");
Expect(0, 11520, '\p{Is_Script_Extensions=qaac}', "");
Expect(1, 11520, '\p{^Is_Script_Extensions=qaac}', "");
Expect(1, 11520, '\P{Is_Script_Extensions=qaac}', "");
Expect(0, 11520, '\P{^Is_Script_Extensions=qaac}', "");
Expect(1, 11519, '\p{Is_Script_Extensions:	_-qaac}', "");
Expect(0, 11519, '\p{^Is_Script_Extensions:	_-qaac}', "");
Expect(0, 11519, '\P{Is_Script_Extensions:	_-qaac}', "");
Expect(1, 11519, '\P{^Is_Script_Extensions:	_-qaac}', "");
Expect(0, 11520, '\p{Is_Script_Extensions:	_-qaac}', "");
Expect(1, 11520, '\p{^Is_Script_Extensions:	_-qaac}', "");
Expect(1, 11520, '\P{Is_Script_Extensions:	_-qaac}', "");
Expect(0, 11520, '\P{^Is_Script_Extensions:	_-qaac}', "");
Error('\p{Is_Scx= :=Coptic}');
Error('\P{Is_Scx= :=Coptic}');
Expect(1, 11519, '\p{Is_Scx=coptic}', "");
Expect(0, 11519, '\p{^Is_Scx=coptic}', "");
Expect(0, 11519, '\P{Is_Scx=coptic}', "");
Expect(1, 11519, '\P{^Is_Scx=coptic}', "");
Expect(0, 11520, '\p{Is_Scx=coptic}', "");
Expect(1, 11520, '\p{^Is_Scx=coptic}', "");
Expect(1, 11520, '\P{Is_Scx=coptic}', "");
Expect(0, 11520, '\P{^Is_Scx=coptic}', "");
Expect(1, 11519, '\p{Is_Scx= _COPTIC}', "");
Expect(0, 11519, '\p{^Is_Scx= _COPTIC}', "");
Expect(0, 11519, '\P{Is_Scx= _COPTIC}', "");
Expect(1, 11519, '\P{^Is_Scx= _COPTIC}', "");
Expect(0, 11520, '\p{Is_Scx= _COPTIC}', "");
Expect(1, 11520, '\p{^Is_Scx= _COPTIC}', "");
Expect(1, 11520, '\P{Is_Scx= _COPTIC}', "");
Expect(0, 11520, '\P{^Is_Scx= _COPTIC}', "");
Error('\p{Script_Extensions=	:=BALINESE}');
Error('\P{Script_Extensions=	:=BALINESE}');
Expect(1, 7036, '\p{Script_Extensions=balinese}', "");
Expect(0, 7036, '\p{^Script_Extensions=balinese}', "");
Expect(0, 7036, '\P{Script_Extensions=balinese}', "");
Expect(1, 7036, '\P{^Script_Extensions=balinese}', "");
Expect(0, 7037, '\p{Script_Extensions=balinese}', "");
Expect(1, 7037, '\p{^Script_Extensions=balinese}', "");
Expect(1, 7037, '\P{Script_Extensions=balinese}', "");
Expect(0, 7037, '\P{^Script_Extensions=balinese}', "");
Expect(1, 7036, '\p{Script_Extensions=--Balinese}', "");
Expect(0, 7036, '\p{^Script_Extensions=--Balinese}', "");
Expect(0, 7036, '\P{Script_Extensions=--Balinese}', "");
Expect(1, 7036, '\P{^Script_Extensions=--Balinese}', "");
Expect(0, 7037, '\p{Script_Extensions=--Balinese}', "");
Expect(1, 7037, '\p{^Script_Extensions=--Balinese}', "");
Expect(1, 7037, '\P{Script_Extensions=--Balinese}', "");
Expect(0, 7037, '\P{^Script_Extensions=--Balinese}', "");
Error('\p{Scx=	:=bali}');
Error('\P{Scx=	:=bali}');
Expect(1, 7036, '\p{Scx=bali}', "");
Expect(0, 7036, '\p{^Scx=bali}', "");
Expect(0, 7036, '\P{Scx=bali}', "");
Expect(1, 7036, '\P{^Scx=bali}', "");
Expect(0, 7037, '\p{Scx=bali}', "");
Expect(1, 7037, '\p{^Scx=bali}', "");
Expect(1, 7037, '\P{Scx=bali}', "");
Expect(0, 7037, '\P{^Scx=bali}', "");
Expect(1, 7036, '\p{Scx=-_bali}', "");
Expect(0, 7036, '\p{^Scx=-_bali}', "");
Expect(0, 7036, '\P{Scx=-_bali}', "");
Expect(1, 7036, '\P{^Scx=-_bali}', "");
Expect(0, 7037, '\p{Scx=-_bali}', "");
Expect(1, 7037, '\p{^Scx=-_bali}', "");
Expect(1, 7037, '\P{Scx=-_bali}', "");
Expect(0, 7037, '\P{^Scx=-_bali}', "");
Error('\p{Is_Script_Extensions=:=_	BALINESE}');
Error('\P{Is_Script_Extensions=:=_	BALINESE}');
Expect(1, 7036, '\p{Is_Script_Extensions=balinese}', "");
Expect(0, 7036, '\p{^Is_Script_Extensions=balinese}', "");
Expect(0, 7036, '\P{Is_Script_Extensions=balinese}', "");
Expect(1, 7036, '\P{^Is_Script_Extensions=balinese}', "");
Expect(0, 7037, '\p{Is_Script_Extensions=balinese}', "");
Expect(1, 7037, '\p{^Is_Script_Extensions=balinese}', "");
Expect(1, 7037, '\P{Is_Script_Extensions=balinese}', "");
Expect(0, 7037, '\P{^Is_Script_Extensions=balinese}', "");
Expect(1, 7036, '\p{Is_Script_Extensions:	 _BALINESE}', "");
Expect(0, 7036, '\p{^Is_Script_Extensions:	 _BALINESE}', "");
Expect(0, 7036, '\P{Is_Script_Extensions:	 _BALINESE}', "");
Expect(1, 7036, '\P{^Is_Script_Extensions:	 _BALINESE}', "");
Expect(0, 7037, '\p{Is_Script_Extensions:	 _BALINESE}', "");
Expect(1, 7037, '\p{^Is_Script_Extensions:	 _BALINESE}', "");
Expect(1, 7037, '\P{Is_Script_Extensions:	 _BALINESE}', "");
Expect(0, 7037, '\P{^Is_Script_Extensions:	 _BALINESE}', "");
Error('\p{Is_Scx=/a/_bali}');
Error('\P{Is_Scx=/a/_bali}');
Expect(1, 7036, '\p{Is_Scx:	bali}', "");
Expect(0, 7036, '\p{^Is_Scx:	bali}', "");
Expect(0, 7036, '\P{Is_Scx:	bali}', "");
Expect(1, 7036, '\P{^Is_Scx:	bali}', "");
Expect(0, 7037, '\p{Is_Scx:	bali}', "");
Expect(1, 7037, '\p{^Is_Scx:	bali}', "");
Expect(1, 7037, '\P{Is_Scx:	bali}', "");
Expect(0, 7037, '\P{^Is_Scx:	bali}', "");
Expect(1, 7036, '\p{Is_Scx= -BALI}', "");
Expect(0, 7036, '\p{^Is_Scx= -BALI}', "");
Expect(0, 7036, '\P{Is_Scx= -BALI}', "");
Expect(1, 7036, '\P{^Is_Scx= -BALI}', "");
Expect(0, 7037, '\p{Is_Scx= -BALI}', "");
Expect(1, 7037, '\p{^Is_Scx= -BALI}', "");
Expect(1, 7037, '\P{Is_Scx= -BALI}', "");
Expect(0, 7037, '\P{^Is_Scx= -BALI}', "");
Error('\p{Script_Extensions=Katakana_Or_Hiragana}');
Error('\P{Script_Extensions=Katakana_Or_Hiragana}');
Error('\p{Scx:	Hrkt}');
Error('\P{Scx:	Hrkt}');
Error('\p{Is_Script_Extensions=Katakana_Or_Hiragana}');
Error('\P{Is_Script_Extensions=Katakana_Or_Hiragana}');
Error('\p{Is_Scx=Hrkt}');
Error('\P{Is_Scx=Hrkt}');
Error('\p{Script_Extensions=/a/_-Meetei_mayek}');
Error('\P{Script_Extensions=/a/_-Meetei_mayek}');
Expect(1, 44025, '\p{Script_Extensions=meeteimayek}', "");
Expect(0, 44025, '\p{^Script_Extensions=meeteimayek}', "");
Expect(0, 44025, '\P{Script_Extensions=meeteimayek}', "");
Expect(1, 44025, '\P{^Script_Extensions=meeteimayek}', "");
Expect(0, 44026, '\p{Script_Extensions=meeteimayek}', "");
Expect(1, 44026, '\p{^Script_Extensions=meeteimayek}', "");
Expect(1, 44026, '\P{Script_Extensions=meeteimayek}', "");
Expect(0, 44026, '\P{^Script_Extensions=meeteimayek}', "");
Expect(1, 44025, '\p{Script_Extensions= _meetei_Mayek}', "");
Expect(0, 44025, '\p{^Script_Extensions= _meetei_Mayek}', "");
Expect(0, 44025, '\P{Script_Extensions= _meetei_Mayek}', "");
Expect(1, 44025, '\P{^Script_Extensions= _meetei_Mayek}', "");
Expect(0, 44026, '\p{Script_Extensions= _meetei_Mayek}', "");
Expect(1, 44026, '\p{^Script_Extensions= _meetei_Mayek}', "");
Expect(1, 44026, '\P{Script_Extensions= _meetei_Mayek}', "");
Expect(0, 44026, '\P{^Script_Extensions= _meetei_Mayek}', "");
Error('\p{Scx=/a/Mtei}');
Error('\P{Scx=/a/Mtei}');
Expect(1, 44025, '\p{Scx:   mtei}', "");
Expect(0, 44025, '\p{^Scx:   mtei}', "");
Expect(0, 44025, '\P{Scx:   mtei}', "");
Expect(1, 44025, '\P{^Scx:   mtei}', "");
Expect(0, 44026, '\p{Scx:   mtei}', "");
Expect(1, 44026, '\p{^Scx:   mtei}', "");
Expect(1, 44026, '\P{Scx:   mtei}', "");
Expect(0, 44026, '\P{^Scx:   mtei}', "");
Expect(1, 44025, '\p{Scx=_ MTEI}', "");
Expect(0, 44025, '\p{^Scx=_ MTEI}', "");
Expect(0, 44025, '\P{Scx=_ MTEI}', "");
Expect(1, 44025, '\P{^Scx=_ MTEI}', "");
Expect(0, 44026, '\p{Scx=_ MTEI}', "");
Expect(1, 44026, '\p{^Scx=_ MTEI}', "");
Expect(1, 44026, '\P{Scx=_ MTEI}', "");
Expect(0, 44026, '\P{^Scx=_ MTEI}', "");
Error('\p{Is_Script_Extensions=-/a/meetei_Mayek}');
Error('\P{Is_Script_Extensions=-/a/meetei_Mayek}');
Expect(1, 44025, '\p{Is_Script_Extensions:	meeteimayek}', "");
Expect(0, 44025, '\p{^Is_Script_Extensions:	meeteimayek}', "");
Expect(0, 44025, '\P{Is_Script_Extensions:	meeteimayek}', "");
Expect(1, 44025, '\P{^Is_Script_Extensions:	meeteimayek}', "");
Expect(0, 44026, '\p{Is_Script_Extensions:	meeteimayek}', "");
Expect(1, 44026, '\p{^Is_Script_Extensions:	meeteimayek}', "");
Expect(1, 44026, '\P{Is_Script_Extensions:	meeteimayek}', "");
Expect(0, 44026, '\P{^Is_Script_Extensions:	meeteimayek}', "");
Expect(1, 44025, '\p{Is_Script_Extensions=		Meetei_Mayek}', "");
Expect(0, 44025, '\p{^Is_Script_Extensions=		Meetei_Mayek}', "");
Expect(0, 44025, '\P{Is_Script_Extensions=		Meetei_Mayek}', "");
Expect(1, 44025, '\P{^Is_Script_Extensions=		Meetei_Mayek}', "");
Expect(0, 44026, '\p{Is_Script_Extensions=		Meetei_Mayek}', "");
Expect(1, 44026, '\p{^Is_Script_Extensions=		Meetei_Mayek}', "");
Expect(1, 44026, '\P{Is_Script_Extensions=		Meetei_Mayek}', "");
Expect(0, 44026, '\P{^Is_Script_Extensions=		Meetei_Mayek}', "");
Error('\p{Is_Scx=/a/- Mtei}');
Error('\P{Is_Scx=/a/- Mtei}');
Expect(1, 44025, '\p{Is_Scx=mtei}', "");
Expect(0, 44025, '\p{^Is_Scx=mtei}', "");
Expect(0, 44025, '\P{Is_Scx=mtei}', "");
Expect(1, 44025, '\P{^Is_Scx=mtei}', "");
Expect(0, 44026, '\p{Is_Scx=mtei}', "");
Expect(1, 44026, '\p{^Is_Scx=mtei}', "");
Expect(1, 44026, '\P{Is_Scx=mtei}', "");
Expect(0, 44026, '\P{^Is_Scx=mtei}', "");
Expect(1, 44025, '\p{Is_Scx=-Mtei}', "");
Expect(0, 44025, '\p{^Is_Scx=-Mtei}', "");
Expect(0, 44025, '\P{Is_Scx=-Mtei}', "");
Expect(1, 44025, '\P{^Is_Scx=-Mtei}', "");
Expect(0, 44026, '\p{Is_Scx=-Mtei}', "");
Expect(1, 44026, '\p{^Is_Scx=-Mtei}', "");
Expect(1, 44026, '\P{Is_Scx=-Mtei}', "");
Expect(0, 44026, '\P{^Is_Scx=-Mtei}', "");
Error('\p{Script_Extensions=/a/	_Syloti_nagri}');
Error('\P{Script_Extensions=/a/	_Syloti_nagri}');
Expect(1, 43051, '\p{Script_Extensions=sylotinagri}', "");
Expect(0, 43051, '\p{^Script_Extensions=sylotinagri}', "");
Expect(0, 43051, '\P{Script_Extensions=sylotinagri}', "");
Expect(1, 43051, '\P{^Script_Extensions=sylotinagri}', "");
Expect(0, 43052, '\p{Script_Extensions=sylotinagri}', "");
Expect(1, 43052, '\p{^Script_Extensions=sylotinagri}', "");
Expect(1, 43052, '\P{Script_Extensions=sylotinagri}', "");
Expect(0, 43052, '\P{^Script_Extensions=sylotinagri}', "");
Expect(1, 43051, '\p{Script_Extensions=	SYLOTI_Nagri}', "");
Expect(0, 43051, '\p{^Script_Extensions=	SYLOTI_Nagri}', "");
Expect(0, 43051, '\P{Script_Extensions=	SYLOTI_Nagri}', "");
Expect(1, 43051, '\P{^Script_Extensions=	SYLOTI_Nagri}', "");
Expect(0, 43052, '\p{Script_Extensions=	SYLOTI_Nagri}', "");
Expect(1, 43052, '\p{^Script_Extensions=	SYLOTI_Nagri}', "");
Expect(1, 43052, '\P{Script_Extensions=	SYLOTI_Nagri}', "");
Expect(0, 43052, '\P{^Script_Extensions=	SYLOTI_Nagri}', "");
Error('\p{Scx=-	sylo:=}');
Error('\P{Scx=-	sylo:=}');
Expect(1, 43051, '\p{Scx=sylo}', "");
Expect(0, 43051, '\p{^Scx=sylo}', "");
Expect(0, 43051, '\P{Scx=sylo}', "");
Expect(1, 43051, '\P{^Scx=sylo}', "");
Expect(0, 43052, '\p{Scx=sylo}', "");
Expect(1, 43052, '\p{^Scx=sylo}', "");
Expect(1, 43052, '\P{Scx=sylo}', "");
Expect(0, 43052, '\P{^Scx=sylo}', "");
Expect(1, 43051, '\p{Scx= 	Sylo}', "");
Expect(0, 43051, '\p{^Scx= 	Sylo}', "");
Expect(0, 43051, '\P{Scx= 	Sylo}', "");
Expect(1, 43051, '\P{^Scx= 	Sylo}', "");
Expect(0, 43052, '\p{Scx= 	Sylo}', "");
Expect(1, 43052, '\p{^Scx= 	Sylo}', "");
Expect(1, 43052, '\P{Scx= 	Sylo}', "");
Expect(0, 43052, '\P{^Scx= 	Sylo}', "");
Error('\p{Is_Script_Extensions=/a/	Syloti_nagri}');
Error('\P{Is_Script_Extensions=/a/	Syloti_nagri}');
Expect(1, 43051, '\p{Is_Script_Extensions=sylotinagri}', "");
Expect(0, 43051, '\p{^Is_Script_Extensions=sylotinagri}', "");
Expect(0, 43051, '\P{Is_Script_Extensions=sylotinagri}', "");
Expect(1, 43051, '\P{^Is_Script_Extensions=sylotinagri}', "");
Expect(0, 43052, '\p{Is_Script_Extensions=sylotinagri}', "");
Expect(1, 43052, '\p{^Is_Script_Extensions=sylotinagri}', "");
Expect(1, 43052, '\P{Is_Script_Extensions=sylotinagri}', "");
Expect(0, 43052, '\P{^Is_Script_Extensions=sylotinagri}', "");
Expect(1, 43051, '\p{Is_Script_Extensions= Syloti_Nagri}', "");
Expect(0, 43051, '\p{^Is_Script_Extensions= Syloti_Nagri}', "");
Expect(0, 43051, '\P{Is_Script_Extensions= Syloti_Nagri}', "");
Expect(1, 43051, '\P{^Is_Script_Extensions= Syloti_Nagri}', "");
Expect(0, 43052, '\p{Is_Script_Extensions= Syloti_Nagri}', "");
Expect(1, 43052, '\p{^Is_Script_Extensions= Syloti_Nagri}', "");
Expect(1, 43052, '\P{Is_Script_Extensions= Syloti_Nagri}', "");
Expect(0, 43052, '\P{^Is_Script_Extensions= Syloti_Nagri}', "");
Error('\p{Is_Scx= 	Sylo/a/}');
Error('\P{Is_Scx= 	Sylo/a/}');
Expect(1, 43051, '\p{Is_Scx=sylo}', "");
Expect(0, 43051, '\p{^Is_Scx=sylo}', "");
Expect(0, 43051, '\P{Is_Scx=sylo}', "");
Expect(1, 43051, '\P{^Is_Scx=sylo}', "");
Expect(0, 43052, '\p{Is_Scx=sylo}', "");
Expect(1, 43052, '\p{^Is_Scx=sylo}', "");
Expect(1, 43052, '\P{Is_Scx=sylo}', "");
Expect(0, 43052, '\P{^Is_Scx=sylo}', "");
Expect(1, 43051, '\p{Is_Scx=-	SYLO}', "");
Expect(0, 43051, '\p{^Is_Scx=-	SYLO}', "");
Expect(0, 43051, '\P{Is_Scx=-	SYLO}', "");
Expect(1, 43051, '\P{^Is_Scx=-	SYLO}', "");
Expect(0, 43052, '\p{Is_Scx=-	SYLO}', "");
Expect(1, 43052, '\p{^Is_Scx=-	SYLO}', "");
Expect(1, 43052, '\P{Is_Scx=-	SYLO}', "");
Expect(0, 43052, '\P{^Is_Scx=-	SYLO}', "");
Error('\p{Script_Extensions=-	tagbanwa/a/}');
Error('\P{Script_Extensions=-	tagbanwa/a/}');
Expect(1, 6003, '\p{Script_Extensions=tagbanwa}', "");
Expect(0, 6003, '\p{^Script_Extensions=tagbanwa}', "");
Expect(0, 6003, '\P{Script_Extensions=tagbanwa}', "");
Expect(1, 6003, '\P{^Script_Extensions=tagbanwa}', "");
Expect(0, 6004, '\p{Script_Extensions=tagbanwa}', "");
Expect(1, 6004, '\p{^Script_Extensions=tagbanwa}', "");
Expect(1, 6004, '\P{Script_Extensions=tagbanwa}', "");
Expect(0, 6004, '\P{^Script_Extensions=tagbanwa}', "");
Expect(1, 6003, '\p{Script_Extensions: 		Tagbanwa}', "");
Expect(0, 6003, '\p{^Script_Extensions: 		Tagbanwa}', "");
Expect(0, 6003, '\P{Script_Extensions: 		Tagbanwa}', "");
Expect(1, 6003, '\P{^Script_Extensions: 		Tagbanwa}', "");
Expect(0, 6004, '\p{Script_Extensions: 		Tagbanwa}', "");
Expect(1, 6004, '\p{^Script_Extensions: 		Tagbanwa}', "");
Expect(1, 6004, '\P{Script_Extensions: 		Tagbanwa}', "");
Expect(0, 6004, '\P{^Script_Extensions: 		Tagbanwa}', "");
Error('\p{Scx=_:=TAGB}');
Error('\P{Scx=_:=TAGB}');
Expect(1, 6003, '\p{Scx=tagb}', "");
Expect(0, 6003, '\p{^Scx=tagb}', "");
Expect(0, 6003, '\P{Scx=tagb}', "");
Expect(1, 6003, '\P{^Scx=tagb}', "");
Expect(0, 6004, '\p{Scx=tagb}', "");
Expect(1, 6004, '\p{^Scx=tagb}', "");
Expect(1, 6004, '\P{Scx=tagb}', "");
Expect(0, 6004, '\P{^Scx=tagb}', "");
Expect(1, 6003, '\p{Scx=	_tagb}', "");
Expect(0, 6003, '\p{^Scx=	_tagb}', "");
Expect(0, 6003, '\P{Scx=	_tagb}', "");
Expect(1, 6003, '\P{^Scx=	_tagb}', "");
Expect(0, 6004, '\p{Scx=	_tagb}', "");
Expect(1, 6004, '\p{^Scx=	_tagb}', "");
Expect(1, 6004, '\P{Scx=	_tagb}', "");
Expect(0, 6004, '\P{^Scx=	_tagb}', "");
Error('\p{Is_Script_Extensions=:=	_tagbanwa}');
Error('\P{Is_Script_Extensions=:=	_tagbanwa}');
Expect(1, 6003, '\p{Is_Script_Extensions=tagbanwa}', "");
Expect(0, 6003, '\p{^Is_Script_Extensions=tagbanwa}', "");
Expect(0, 6003, '\P{Is_Script_Extensions=tagbanwa}', "");
Expect(1, 6003, '\P{^Is_Script_Extensions=tagbanwa}', "");
Expect(0, 6004, '\p{Is_Script_Extensions=tagbanwa}', "");
Expect(1, 6004, '\p{^Is_Script_Extensions=tagbanwa}', "");
Expect(1, 6004, '\P{Is_Script_Extensions=tagbanwa}', "");
Expect(0, 6004, '\P{^Is_Script_Extensions=tagbanwa}', "");
Expect(1, 6003, '\p{Is_Script_Extensions=	tagbanwa}', "");
Expect(0, 6003, '\p{^Is_Script_Extensions=	tagbanwa}', "");
Expect(0, 6003, '\P{Is_Script_Extensions=	tagbanwa}', "");
Expect(1, 6003, '\P{^Is_Script_Extensions=	tagbanwa}', "");
Expect(0, 6004, '\p{Is_Script_Extensions=	tagbanwa}', "");
Expect(1, 6004, '\p{^Is_Script_Extensions=	tagbanwa}', "");
Expect(1, 6004, '\P{Is_Script_Extensions=	tagbanwa}', "");
Expect(0, 6004, '\P{^Is_Script_Extensions=	tagbanwa}', "");
Error('\p{Is_Scx=_Tagb/a/}');
Error('\P{Is_Scx=_Tagb/a/}');
Expect(1, 6003, '\p{Is_Scx=tagb}', "");
Expect(0, 6003, '\p{^Is_Scx=tagb}', "");
Expect(0, 6003, '\P{Is_Scx=tagb}', "");
Expect(1, 6003, '\P{^Is_Scx=tagb}', "");
Expect(0, 6004, '\p{Is_Scx=tagb}', "");
Expect(1, 6004, '\p{^Is_Scx=tagb}', "");
Expect(1, 6004, '\P{Is_Scx=tagb}', "");
Expect(0, 6004, '\P{^Is_Scx=tagb}', "");
Expect(1, 6003, '\p{Is_Scx=TAGB}', "");
Expect(0, 6003, '\p{^Is_Scx=TAGB}', "");
Expect(0, 6003, '\P{Is_Scx=TAGB}', "");
Expect(1, 6003, '\P{^Is_Scx=TAGB}', "");
Expect(0, 6004, '\p{Is_Scx=TAGB}', "");
Expect(1, 6004, '\p{^Is_Scx=TAGB}', "");
Expect(1, 6004, '\P{Is_Scx=TAGB}', "");
Expect(0, 6004, '\P{^Is_Scx=TAGB}', "");
Error('\p{Script_Extensions=:=_Phoenician}');
Error('\P{Script_Extensions=:=_Phoenician}');
Expect(1, 67871, '\p{Script_Extensions=phoenician}', "");
Expect(0, 67871, '\p{^Script_Extensions=phoenician}', "");
Expect(0, 67871, '\P{Script_Extensions=phoenician}', "");
Expect(1, 67871, '\P{^Script_Extensions=phoenician}', "");
Expect(0, 67872, '\p{Script_Extensions=phoenician}', "");
Expect(1, 67872, '\p{^Script_Extensions=phoenician}', "");
Expect(1, 67872, '\P{Script_Extensions=phoenician}', "");
Expect(0, 67872, '\P{^Script_Extensions=phoenician}', "");
Expect(1, 67871, '\p{Script_Extensions=- Phoenician}', "");
Expect(0, 67871, '\p{^Script_Extensions=- Phoenician}', "");
Expect(0, 67871, '\P{Script_Extensions=- Phoenician}', "");
Expect(1, 67871, '\P{^Script_Extensions=- Phoenician}', "");
Expect(0, 67872, '\p{Script_Extensions=- Phoenician}', "");
Expect(1, 67872, '\p{^Script_Extensions=- Phoenician}', "");
Expect(1, 67872, '\P{Script_Extensions=- Phoenician}', "");
Expect(0, 67872, '\P{^Script_Extensions=- Phoenician}', "");
Error('\p{Scx=/a/_ PHNX}');
Error('\P{Scx=/a/_ PHNX}');
Expect(1, 67871, '\p{Scx=phnx}', "");
Expect(0, 67871, '\p{^Scx=phnx}', "");
Expect(0, 67871, '\P{Scx=phnx}', "");
Expect(1, 67871, '\P{^Scx=phnx}', "");
Expect(0, 67872, '\p{Scx=phnx}', "");
Expect(1, 67872, '\p{^Scx=phnx}', "");
Expect(1, 67872, '\P{Scx=phnx}', "");
Expect(0, 67872, '\P{^Scx=phnx}', "");
Expect(1, 67871, '\p{Scx= Phnx}', "");
Expect(0, 67871, '\p{^Scx= Phnx}', "");
Expect(0, 67871, '\P{Scx= Phnx}', "");
Expect(1, 67871, '\P{^Scx= Phnx}', "");
Expect(0, 67872, '\p{Scx= Phnx}', "");
Expect(1, 67872, '\p{^Scx= Phnx}', "");
Expect(1, 67872, '\P{Scx= Phnx}', "");
Expect(0, 67872, '\P{^Scx= Phnx}', "");
Error('\p{Is_Script_Extensions=/a/-PHOENICIAN}');
Error('\P{Is_Script_Extensions=/a/-PHOENICIAN}');
Expect(1, 67871, '\p{Is_Script_Extensions=phoenician}', "");
Expect(0, 67871, '\p{^Is_Script_Extensions=phoenician}', "");
Expect(0, 67871, '\P{Is_Script_Extensions=phoenician}', "");
Expect(1, 67871, '\P{^Is_Script_Extensions=phoenician}', "");
Expect(0, 67872, '\p{Is_Script_Extensions=phoenician}', "");
Expect(1, 67872, '\p{^Is_Script_Extensions=phoenician}', "");
Expect(1, 67872, '\P{Is_Script_Extensions=phoenician}', "");
Expect(0, 67872, '\P{^Is_Script_Extensions=phoenician}', "");
Expect(1, 67871, '\p{Is_Script_Extensions=_Phoenician}', "");
Expect(0, 67871, '\p{^Is_Script_Extensions=_Phoenician}', "");
Expect(0, 67871, '\P{Is_Script_Extensions=_Phoenician}', "");
Expect(1, 67871, '\P{^Is_Script_Extensions=_Phoenician}', "");
Expect(0, 67872, '\p{Is_Script_Extensions=_Phoenician}', "");
Expect(1, 67872, '\p{^Is_Script_Extensions=_Phoenician}', "");
Expect(1, 67872, '\P{Is_Script_Extensions=_Phoenician}', "");
Expect(0, 67872, '\P{^Is_Script_Extensions=_Phoenician}', "");
Error('\p{Is_Scx=/a/	-phnx}');
Error('\P{Is_Scx=/a/	-phnx}');
Expect(1, 67871, '\p{Is_Scx=phnx}', "");
Expect(0, 67871, '\p{^Is_Scx=phnx}', "");
Expect(0, 67871, '\P{Is_Scx=phnx}', "");
Expect(1, 67871, '\P{^Is_Scx=phnx}', "");
Expect(0, 67872, '\p{Is_Scx=phnx}', "");
Expect(1, 67872, '\p{^Is_Scx=phnx}', "");
Expect(1, 67872, '\P{Is_Scx=phnx}', "");
Expect(0, 67872, '\P{^Is_Scx=phnx}', "");
Expect(1, 67871, '\p{Is_Scx=		Phnx}', "");
Expect(0, 67871, '\p{^Is_Scx=		Phnx}', "");
Expect(0, 67871, '\P{Is_Scx=		Phnx}', "");
Expect(1, 67871, '\P{^Is_Scx=		Phnx}', "");
Expect(0, 67872, '\p{Is_Scx=		Phnx}', "");
Expect(1, 67872, '\p{^Is_Scx=		Phnx}', "");
Expect(1, 67872, '\P{Is_Scx=		Phnx}', "");
Expect(0, 67872, '\P{^Is_Scx=		Phnx}', "");
Error('\p{Script_Extensions=--Devanagari:=}');
Error('\P{Script_Extensions=--Devanagari:=}');
Expect(1, 43259, '\p{Script_Extensions=devanagari}', "");
Expect(0, 43259, '\p{^Script_Extensions=devanagari}', "");
Expect(0, 43259, '\P{Script_Extensions=devanagari}', "");
Expect(1, 43259, '\P{^Script_Extensions=devanagari}', "");
Expect(0, 43260, '\p{Script_Extensions=devanagari}', "");
Expect(1, 43260, '\p{^Script_Extensions=devanagari}', "");
Expect(1, 43260, '\P{Script_Extensions=devanagari}', "");
Expect(0, 43260, '\P{^Script_Extensions=devanagari}', "");
Expect(1, 43259, '\p{Script_Extensions=_	Devanagari}', "");
Expect(0, 43259, '\p{^Script_Extensions=_	Devanagari}', "");
Expect(0, 43259, '\P{Script_Extensions=_	Devanagari}', "");
Expect(1, 43259, '\P{^Script_Extensions=_	Devanagari}', "");
Expect(0, 43260, '\p{Script_Extensions=_	Devanagari}', "");
Expect(1, 43260, '\p{^Script_Extensions=_	Devanagari}', "");
Expect(1, 43260, '\P{Script_Extensions=_	Devanagari}', "");
Expect(0, 43260, '\P{^Script_Extensions=_	Devanagari}', "");
Error('\p{Scx= :=Deva}');
Error('\P{Scx= :=Deva}');
Expect(1, 43259, '\p{Scx=deva}', "");
Expect(0, 43259, '\p{^Scx=deva}', "");
Expect(0, 43259, '\P{Scx=deva}', "");
Expect(1, 43259, '\P{^Scx=deva}', "");
Expect(0, 43260, '\p{Scx=deva}', "");
Expect(1, 43260, '\p{^Scx=deva}', "");
Expect(1, 43260, '\P{Scx=deva}', "");
Expect(0, 43260, '\P{^Scx=deva}', "");
Expect(1, 43259, '\p{Scx= 	Deva}', "");
Expect(0, 43259, '\p{^Scx= 	Deva}', "");
Expect(0, 43259, '\P{Scx= 	Deva}', "");
Expect(1, 43259, '\P{^Scx= 	Deva}', "");
Expect(0, 43260, '\p{Scx= 	Deva}', "");
Expect(1, 43260, '\p{^Scx= 	Deva}', "");
Expect(1, 43260, '\P{Scx= 	Deva}', "");
Expect(0, 43260, '\P{^Scx= 	Deva}', "");
Error('\p{Is_Script_Extensions=:=	DEVANAGARI}');
Error('\P{Is_Script_Extensions=:=	DEVANAGARI}');
Expect(1, 43259, '\p{Is_Script_Extensions=devanagari}', "");
Expect(0, 43259, '\p{^Is_Script_Extensions=devanagari}', "");
Expect(0, 43259, '\P{Is_Script_Extensions=devanagari}', "");
Expect(1, 43259, '\P{^Is_Script_Extensions=devanagari}', "");
Expect(0, 43260, '\p{Is_Script_Extensions=devanagari}', "");
Expect(1, 43260, '\p{^Is_Script_Extensions=devanagari}', "");
Expect(1, 43260, '\P{Is_Script_Extensions=devanagari}', "");
Expect(0, 43260, '\P{^Is_Script_Extensions=devanagari}', "");
Expect(1, 43259, '\p{Is_Script_Extensions=	_devanagari}', "");
Expect(0, 43259, '\p{^Is_Script_Extensions=	_devanagari}', "");
Expect(0, 43259, '\P{Is_Script_Extensions=	_devanagari}', "");
Expect(1, 43259, '\P{^Is_Script_Extensions=	_devanagari}', "");
Expect(0, 43260, '\p{Is_Script_Extensions=	_devanagari}', "");
Expect(1, 43260, '\p{^Is_Script_Extensions=	_devanagari}', "");
Expect(1, 43260, '\P{Is_Script_Extensions=	_devanagari}', "");
Expect(0, 43260, '\P{^Is_Script_Extensions=	_devanagari}', "");
Error('\p{Is_Scx= _Deva:=}');
Error('\P{Is_Scx= _Deva:=}');
Expect(1, 43259, '\p{Is_Scx=deva}', "");
Expect(0, 43259, '\p{^Is_Scx=deva}', "");
Expect(0, 43259, '\P{Is_Scx=deva}', "");
Expect(1, 43259, '\P{^Is_Scx=deva}', "");
Expect(0, 43260, '\p{Is_Scx=deva}', "");
Expect(1, 43260, '\p{^Is_Scx=deva}', "");
Expect(1, 43260, '\P{Is_Scx=deva}', "");
Expect(0, 43260, '\P{^Is_Scx=deva}', "");
Expect(1, 43259, '\p{Is_Scx=-_DEVA}', "");
Expect(0, 43259, '\p{^Is_Scx=-_DEVA}', "");
Expect(0, 43259, '\P{Is_Scx=-_DEVA}', "");
Expect(1, 43259, '\P{^Is_Scx=-_DEVA}', "");
Expect(0, 43260, '\p{Is_Scx=-_DEVA}', "");
Expect(1, 43260, '\p{^Is_Scx=-_DEVA}', "");
Expect(1, 43260, '\P{Is_Scx=-_DEVA}', "");
Expect(0, 43260, '\P{^Is_Scx=-_DEVA}', "");
Error('\p{Script_Extensions=_:=CARIAN}');
Error('\P{Script_Extensions=_:=CARIAN}');
Expect(1, 66256, '\p{Script_Extensions=carian}', "");
Expect(0, 66256, '\p{^Script_Extensions=carian}', "");
Expect(0, 66256, '\P{Script_Extensions=carian}', "");
Expect(1, 66256, '\P{^Script_Extensions=carian}', "");
Expect(0, 66257, '\p{Script_Extensions=carian}', "");
Expect(1, 66257, '\p{^Script_Extensions=carian}', "");
Expect(1, 66257, '\P{Script_Extensions=carian}', "");
Expect(0, 66257, '\P{^Script_Extensions=carian}', "");
Expect(1, 66256, '\p{Script_Extensions=_	carian}', "");
Expect(0, 66256, '\p{^Script_Extensions=_	carian}', "");
Expect(0, 66256, '\P{Script_Extensions=_	carian}', "");
Expect(1, 66256, '\P{^Script_Extensions=_	carian}', "");
Expect(0, 66257, '\p{Script_Extensions=_	carian}', "");
Expect(1, 66257, '\p{^Script_Extensions=_	carian}', "");
Expect(1, 66257, '\P{Script_Extensions=_	carian}', "");
Expect(0, 66257, '\P{^Script_Extensions=_	carian}', "");
Error('\p{Scx=-:=Cari}');
Error('\P{Scx=-:=Cari}');
Expect(1, 66256, '\p{Scx:cari}', "");
Expect(0, 66256, '\p{^Scx:cari}', "");
Expect(0, 66256, '\P{Scx:cari}', "");
Expect(1, 66256, '\P{^Scx:cari}', "");
Expect(0, 66257, '\p{Scx:cari}', "");
Expect(1, 66257, '\p{^Scx:cari}', "");
Expect(1, 66257, '\P{Scx:cari}', "");
Expect(0, 66257, '\P{^Scx:cari}', "");
Expect(1, 66256, '\p{Scx=--Cari}', "");
Expect(0, 66256, '\p{^Scx=--Cari}', "");
Expect(0, 66256, '\P{Scx=--Cari}', "");
Expect(1, 66256, '\P{^Scx=--Cari}', "");
Expect(0, 66257, '\p{Scx=--Cari}', "");
Expect(1, 66257, '\p{^Scx=--Cari}', "");
Expect(1, 66257, '\P{Scx=--Cari}', "");
Expect(0, 66257, '\P{^Scx=--Cari}', "");
Error('\p{Is_Script_Extensions=_:=CARIAN}');
Error('\P{Is_Script_Extensions=_:=CARIAN}');
Expect(1, 66256, '\p{Is_Script_Extensions=carian}', "");
Expect(0, 66256, '\p{^Is_Script_Extensions=carian}', "");
Expect(0, 66256, '\P{Is_Script_Extensions=carian}', "");
Expect(1, 66256, '\P{^Is_Script_Extensions=carian}', "");
Expect(0, 66257, '\p{Is_Script_Extensions=carian}', "");
Expect(1, 66257, '\p{^Is_Script_Extensions=carian}', "");
Expect(1, 66257, '\P{Is_Script_Extensions=carian}', "");
Expect(0, 66257, '\P{^Is_Script_Extensions=carian}', "");
Expect(1, 66256, '\p{Is_Script_Extensions=_Carian}', "");
Expect(0, 66256, '\p{^Is_Script_Extensions=_Carian}', "");
Expect(0, 66256, '\P{Is_Script_Extensions=_Carian}', "");
Expect(1, 66256, '\P{^Is_Script_Extensions=_Carian}', "");
Expect(0, 66257, '\p{Is_Script_Extensions=_Carian}', "");
Expect(1, 66257, '\p{^Is_Script_Extensions=_Carian}', "");
Expect(1, 66257, '\P{Is_Script_Extensions=_Carian}', "");
Expect(0, 66257, '\P{^Is_Script_Extensions=_Carian}', "");
Error('\p{Is_Scx=:=Cari}');
Error('\P{Is_Scx=:=Cari}');
Expect(1, 66256, '\p{Is_Scx=cari}', "");
Expect(0, 66256, '\p{^Is_Scx=cari}', "");
Expect(0, 66256, '\P{Is_Scx=cari}', "");
Expect(1, 66256, '\P{^Is_Scx=cari}', "");
Expect(0, 66257, '\p{Is_Scx=cari}', "");
Expect(1, 66257, '\p{^Is_Scx=cari}', "");
Expect(1, 66257, '\P{Is_Scx=cari}', "");
Expect(0, 66257, '\P{^Is_Scx=cari}', "");
Error('\p{Script_Extensions=:=--sharada}');
Error('\P{Script_Extensions=:=--sharada}');
Expect(1, 70105, '\p{Script_Extensions=sharada}', "");
Expect(0, 70105, '\p{^Script_Extensions=sharada}', "");
Expect(0, 70105, '\P{Script_Extensions=sharada}', "");
Expect(1, 70105, '\P{^Script_Extensions=sharada}', "");
Expect(0, 70106, '\p{Script_Extensions=sharada}', "");
Expect(1, 70106, '\p{^Script_Extensions=sharada}', "");
Expect(1, 70106, '\P{Script_Extensions=sharada}', "");
Expect(0, 70106, '\P{^Script_Extensions=sharada}', "");
Expect(1, 70105, '\p{Script_Extensions=	 Sharada}', "");
Expect(0, 70105, '\p{^Script_Extensions=	 Sharada}', "");
Expect(0, 70105, '\P{Script_Extensions=	 Sharada}', "");
Expect(1, 70105, '\P{^Script_Extensions=	 Sharada}', "");
Expect(0, 70106, '\p{Script_Extensions=	 Sharada}', "");
Expect(1, 70106, '\p{^Script_Extensions=	 Sharada}', "");
Expect(1, 70106, '\P{Script_Extensions=	 Sharada}', "");
Expect(0, 70106, '\P{^Script_Extensions=	 Sharada}', "");
Error('\p{Scx=-	shrd:=}');
Error('\P{Scx=-	shrd:=}');
Expect(1, 70105, '\p{Scx:   shrd}', "");
Expect(0, 70105, '\p{^Scx:   shrd}', "");
Expect(0, 70105, '\P{Scx:   shrd}', "");
Expect(1, 70105, '\P{^Scx:   shrd}', "");
Expect(0, 70106, '\p{Scx:   shrd}', "");
Expect(1, 70106, '\p{^Scx:   shrd}', "");
Expect(1, 70106, '\P{Scx:   shrd}', "");
Expect(0, 70106, '\P{^Scx:   shrd}', "");
Expect(1, 70105, '\p{Scx=	 Shrd}', "");
Expect(0, 70105, '\p{^Scx=	 Shrd}', "");
Expect(0, 70105, '\P{Scx=	 Shrd}', "");
Expect(1, 70105, '\P{^Scx=	 Shrd}', "");
Expect(0, 70106, '\p{Scx=	 Shrd}', "");
Expect(1, 70106, '\p{^Scx=	 Shrd}', "");
Expect(1, 70106, '\P{Scx=	 Shrd}', "");
Expect(0, 70106, '\P{^Scx=	 Shrd}', "");
Error('\p{Is_Script_Extensions=-:=Sharada}');
Error('\P{Is_Script_Extensions=-:=Sharada}');
Expect(1, 70105, '\p{Is_Script_Extensions=sharada}', "");
Expect(0, 70105, '\p{^Is_Script_Extensions=sharada}', "");
Expect(0, 70105, '\P{Is_Script_Extensions=sharada}', "");
Expect(1, 70105, '\P{^Is_Script_Extensions=sharada}', "");
Expect(0, 70106, '\p{Is_Script_Extensions=sharada}', "");
Expect(1, 70106, '\p{^Is_Script_Extensions=sharada}', "");
Expect(1, 70106, '\P{Is_Script_Extensions=sharada}', "");
Expect(0, 70106, '\P{^Is_Script_Extensions=sharada}', "");
Expect(1, 70105, '\p{Is_Script_Extensions=	sharada}', "");
Expect(0, 70105, '\p{^Is_Script_Extensions=	sharada}', "");
Expect(0, 70105, '\P{Is_Script_Extensions=	sharada}', "");
Expect(1, 70105, '\P{^Is_Script_Extensions=	sharada}', "");
Expect(0, 70106, '\p{Is_Script_Extensions=	sharada}', "");
Expect(1, 70106, '\p{^Is_Script_Extensions=	sharada}', "");
Expect(1, 70106, '\P{Is_Script_Extensions=	sharada}', "");
Expect(0, 70106, '\P{^Is_Script_Extensions=	sharada}', "");
Error('\p{Is_Scx=/a/-Shrd}');
Error('\P{Is_Scx=/a/-Shrd}');
Expect(1, 70105, '\p{Is_Scx=shrd}', "");
Expect(0, 70105, '\p{^Is_Scx=shrd}', "");
Expect(0, 70105, '\P{Is_Scx=shrd}', "");
Expect(1, 70105, '\P{^Is_Scx=shrd}', "");
Expect(0, 70106, '\p{Is_Scx=shrd}', "");
Expect(1, 70106, '\p{^Is_Scx=shrd}', "");
Expect(1, 70106, '\P{Is_Scx=shrd}', "");
Expect(0, 70106, '\P{^Is_Scx=shrd}', "");
Expect(1, 70105, '\p{Is_Scx= 	Shrd}', "");
Expect(0, 70105, '\p{^Is_Scx= 	Shrd}', "");
Expect(0, 70105, '\P{Is_Scx= 	Shrd}', "");
Expect(1, 70105, '\P{^Is_Scx= 	Shrd}', "");
Expect(0, 70106, '\p{Is_Scx= 	Shrd}', "");
Expect(1, 70106, '\p{^Is_Scx= 	Shrd}', "");
Expect(1, 70106, '\P{Is_Scx= 	Shrd}', "");
Expect(0, 70106, '\P{^Is_Scx= 	Shrd}', "");
Error('\p{Script_Extensions= :=Buhid}');
Error('\P{Script_Extensions= :=Buhid}');
Expect(1, 5971, '\p{Script_Extensions=buhid}', "");
Expect(0, 5971, '\p{^Script_Extensions=buhid}', "");
Expect(0, 5971, '\P{Script_Extensions=buhid}', "");
Expect(1, 5971, '\P{^Script_Extensions=buhid}', "");
Expect(0, 5972, '\p{Script_Extensions=buhid}', "");
Expect(1, 5972, '\p{^Script_Extensions=buhid}', "");
Expect(1, 5972, '\P{Script_Extensions=buhid}', "");
Expect(0, 5972, '\P{^Script_Extensions=buhid}', "");
Expect(1, 5971, '\p{Script_Extensions=-Buhid}', "");
Expect(0, 5971, '\p{^Script_Extensions=-Buhid}', "");
Expect(0, 5971, '\P{Script_Extensions=-Buhid}', "");
Expect(1, 5971, '\P{^Script_Extensions=-Buhid}', "");
Expect(0, 5972, '\p{Script_Extensions=-Buhid}', "");
Expect(1, 5972, '\p{^Script_Extensions=-Buhid}', "");
Expect(1, 5972, '\P{Script_Extensions=-Buhid}', "");
Expect(0, 5972, '\P{^Script_Extensions=-Buhid}', "");
Error('\p{Scx=-	BUHD/a/}');
Error('\P{Scx=-	BUHD/a/}');
Expect(1, 5971, '\p{Scx=buhd}', "");
Expect(0, 5971, '\p{^Scx=buhd}', "");
Expect(0, 5971, '\P{Scx=buhd}', "");
Expect(1, 5971, '\P{^Scx=buhd}', "");
Expect(0, 5972, '\p{Scx=buhd}', "");
Expect(1, 5972, '\p{^Scx=buhd}', "");
Expect(1, 5972, '\P{Scx=buhd}', "");
Expect(0, 5972, '\P{^Scx=buhd}', "");
Expect(1, 5971, '\p{Scx=	-buhd}', "");
Expect(0, 5971, '\p{^Scx=	-buhd}', "");
Expect(0, 5971, '\P{Scx=	-buhd}', "");
Expect(1, 5971, '\P{^Scx=	-buhd}', "");
Expect(0, 5972, '\p{Scx=	-buhd}', "");
Expect(1, 5972, '\p{^Scx=	-buhd}', "");
Expect(1, 5972, '\P{Scx=	-buhd}', "");
Expect(0, 5972, '\P{^Scx=	-buhd}', "");
Error('\p{Is_Script_Extensions=-:=Buhid}');
Error('\P{Is_Script_Extensions=-:=Buhid}');
Expect(1, 5971, '\p{Is_Script_Extensions=buhid}', "");
Expect(0, 5971, '\p{^Is_Script_Extensions=buhid}', "");
Expect(0, 5971, '\P{Is_Script_Extensions=buhid}', "");
Expect(1, 5971, '\P{^Is_Script_Extensions=buhid}', "");
Expect(0, 5972, '\p{Is_Script_Extensions=buhid}', "");
Expect(1, 5972, '\p{^Is_Script_Extensions=buhid}', "");
Expect(1, 5972, '\P{Is_Script_Extensions=buhid}', "");
Expect(0, 5972, '\P{^Is_Script_Extensions=buhid}', "");
Expect(1, 5971, '\p{Is_Script_Extensions=		BUHID}', "");
Expect(0, 5971, '\p{^Is_Script_Extensions=		BUHID}', "");
Expect(0, 5971, '\P{Is_Script_Extensions=		BUHID}', "");
Expect(1, 5971, '\P{^Is_Script_Extensions=		BUHID}', "");
Expect(0, 5972, '\p{Is_Script_Extensions=		BUHID}', "");
Expect(1, 5972, '\p{^Is_Script_Extensions=		BUHID}', "");
Expect(1, 5972, '\P{Is_Script_Extensions=		BUHID}', "");
Expect(0, 5972, '\P{^Is_Script_Extensions=		BUHID}', "");
Error('\p{Is_Scx=:= BUHD}');
Error('\P{Is_Scx=:= BUHD}');
Expect(1, 5971, '\p{Is_Scx=buhd}', "");
Expect(0, 5971, '\p{^Is_Scx=buhd}', "");
Expect(0, 5971, '\P{Is_Scx=buhd}', "");
Expect(1, 5971, '\P{^Is_Scx=buhd}', "");
Expect(0, 5972, '\p{Is_Scx=buhd}', "");
Expect(1, 5972, '\p{^Is_Scx=buhd}', "");
Expect(1, 5972, '\P{Is_Scx=buhd}', "");
Expect(0, 5972, '\P{^Is_Scx=buhd}', "");
Expect(1, 5971, '\p{Is_Scx=- buhd}', "");
Expect(0, 5971, '\p{^Is_Scx=- buhd}', "");
Expect(0, 5971, '\P{Is_Scx=- buhd}', "");
Expect(1, 5971, '\P{^Is_Scx=- buhd}', "");
Expect(0, 5972, '\p{Is_Scx=- buhd}', "");
Expect(1, 5972, '\p{^Is_Scx=- buhd}', "");
Expect(1, 5972, '\P{Is_Scx=- buhd}', "");
Expect(0, 5972, '\P{^Is_Scx=- buhd}', "");
Error('\p{Script_Extensions=/a/-Limbu}');
Error('\P{Script_Extensions=/a/-Limbu}');
Expect(1, 6479, '\p{Script_Extensions=limbu}', "");
Expect(0, 6479, '\p{^Script_Extensions=limbu}', "");
Expect(0, 6479, '\P{Script_Extensions=limbu}', "");
Expect(1, 6479, '\P{^Script_Extensions=limbu}', "");
Expect(0, 6480, '\p{Script_Extensions=limbu}', "");
Expect(1, 6480, '\p{^Script_Extensions=limbu}', "");
Expect(1, 6480, '\P{Script_Extensions=limbu}', "");
Expect(0, 6480, '\P{^Script_Extensions=limbu}', "");
Expect(1, 6479, '\p{Script_Extensions=	_Limbu}', "");
Expect(0, 6479, '\p{^Script_Extensions=	_Limbu}', "");
Expect(0, 6479, '\P{Script_Extensions=	_Limbu}', "");
Expect(1, 6479, '\P{^Script_Extensions=	_Limbu}', "");
Expect(0, 6480, '\p{Script_Extensions=	_Limbu}', "");
Expect(1, 6480, '\p{^Script_Extensions=	_Limbu}', "");
Expect(1, 6480, '\P{Script_Extensions=	_Limbu}', "");
Expect(0, 6480, '\P{^Script_Extensions=	_Limbu}', "");
Error('\p{Scx:	/a/_	Limb}');
Error('\P{Scx:	/a/_	Limb}');
Expect(1, 6479, '\p{Scx=limb}', "");
Expect(0, 6479, '\p{^Scx=limb}', "");
Expect(0, 6479, '\P{Scx=limb}', "");
Expect(1, 6479, '\P{^Scx=limb}', "");
Expect(0, 6480, '\p{Scx=limb}', "");
Expect(1, 6480, '\p{^Scx=limb}', "");
Expect(1, 6480, '\P{Scx=limb}', "");
Expect(0, 6480, '\P{^Scx=limb}', "");
Expect(1, 6479, '\p{Scx=__Limb}', "");
Expect(0, 6479, '\p{^Scx=__Limb}', "");
Expect(0, 6479, '\P{Scx=__Limb}', "");
Expect(1, 6479, '\P{^Scx=__Limb}', "");
Expect(0, 6480, '\p{Scx=__Limb}', "");
Expect(1, 6480, '\p{^Scx=__Limb}', "");
Expect(1, 6480, '\P{Scx=__Limb}', "");
Expect(0, 6480, '\P{^Scx=__Limb}', "");
Error('\p{Is_Script_Extensions=:=-Limbu}');
Error('\P{Is_Script_Extensions=:=-Limbu}');
Expect(1, 6479, '\p{Is_Script_Extensions=limbu}', "");
Expect(0, 6479, '\p{^Is_Script_Extensions=limbu}', "");
Expect(0, 6479, '\P{Is_Script_Extensions=limbu}', "");
Expect(1, 6479, '\P{^Is_Script_Extensions=limbu}', "");
Expect(0, 6480, '\p{Is_Script_Extensions=limbu}', "");
Expect(1, 6480, '\p{^Is_Script_Extensions=limbu}', "");
Expect(1, 6480, '\P{Is_Script_Extensions=limbu}', "");
Expect(0, 6480, '\P{^Is_Script_Extensions=limbu}', "");
Expect(1, 6479, '\p{Is_Script_Extensions:    Limbu}', "");
Expect(0, 6479, '\p{^Is_Script_Extensions:    Limbu}', "");
Expect(0, 6479, '\P{Is_Script_Extensions:    Limbu}', "");
Expect(1, 6479, '\P{^Is_Script_Extensions:    Limbu}', "");
Expect(0, 6480, '\p{Is_Script_Extensions:    Limbu}', "");
Expect(1, 6480, '\p{^Is_Script_Extensions:    Limbu}', "");
Expect(1, 6480, '\P{Is_Script_Extensions:    Limbu}', "");
Expect(0, 6480, '\P{^Is_Script_Extensions:    Limbu}', "");
Error('\p{Is_Scx=:=- limb}');
Error('\P{Is_Scx=:=- limb}');
Expect(1, 6479, '\p{Is_Scx=limb}', "");
Expect(0, 6479, '\p{^Is_Scx=limb}', "");
Expect(0, 6479, '\P{Is_Scx=limb}', "");
Expect(1, 6479, '\P{^Is_Scx=limb}', "");
Expect(0, 6480, '\p{Is_Scx=limb}', "");
Expect(1, 6480, '\p{^Is_Scx=limb}', "");
Expect(1, 6480, '\P{Is_Scx=limb}', "");
Expect(0, 6480, '\P{^Is_Scx=limb}', "");
Expect(1, 6479, '\p{Is_Scx:-_LIMB}', "");
Expect(0, 6479, '\p{^Is_Scx:-_LIMB}', "");
Expect(0, 6479, '\P{Is_Scx:-_LIMB}', "");
Expect(1, 6479, '\P{^Is_Scx:-_LIMB}', "");
Expect(0, 6480, '\p{Is_Scx:-_LIMB}', "");
Expect(1, 6480, '\p{^Is_Scx:-_LIMB}', "");
Expect(1, 6480, '\P{Is_Scx:-_LIMB}', "");
Expect(0, 6480, '\P{^Is_Scx:-_LIMB}', "");
Error('\p{Script_Extensions=_/a/egyptian_hieroglyphs}');
Error('\P{Script_Extensions=_/a/egyptian_hieroglyphs}');
Expect(1, 78894, '\p{Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78894, '\p{^Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78894, '\P{Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78894, '\P{^Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{^Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{^Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78894, '\p{Script_Extensions=--Egyptian_HIEROGLYPHS}', "");
Expect(0, 78894, '\p{^Script_Extensions=--Egyptian_HIEROGLYPHS}', "");
Expect(0, 78894, '\P{Script_Extensions=--Egyptian_HIEROGLYPHS}', "");
Expect(1, 78894, '\P{^Script_Extensions=--Egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\p{Script_Extensions=--Egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\p{^Script_Extensions=--Egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\P{Script_Extensions=--Egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\P{^Script_Extensions=--Egyptian_HIEROGLYPHS}', "");
Error('\p{Scx= Egyp:=}');
Error('\P{Scx= Egyp:=}');
Expect(1, 78894, '\p{Scx=egyp}', "");
Expect(0, 78894, '\p{^Scx=egyp}', "");
Expect(0, 78894, '\P{Scx=egyp}', "");
Expect(1, 78894, '\P{^Scx=egyp}', "");
Expect(0, 78895, '\p{Scx=egyp}', "");
Expect(1, 78895, '\p{^Scx=egyp}', "");
Expect(1, 78895, '\P{Scx=egyp}', "");
Expect(0, 78895, '\P{^Scx=egyp}', "");
Expect(1, 78894, '\p{Scx:   --EGYP}', "");
Expect(0, 78894, '\p{^Scx:   --EGYP}', "");
Expect(0, 78894, '\P{Scx:   --EGYP}', "");
Expect(1, 78894, '\P{^Scx:   --EGYP}', "");
Expect(0, 78895, '\p{Scx:   --EGYP}', "");
Expect(1, 78895, '\p{^Scx:   --EGYP}', "");
Expect(1, 78895, '\P{Scx:   --EGYP}', "");
Expect(0, 78895, '\P{^Scx:   --EGYP}', "");
Error('\p{Is_Script_Extensions=  Egyptian_Hieroglyphs/a/}');
Error('\P{Is_Script_Extensions=  Egyptian_Hieroglyphs/a/}');
Expect(1, 78894, '\p{Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78894, '\p{^Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78894, '\P{Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78894, '\P{^Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{^Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{^Is_Script_Extensions=egyptianhieroglyphs}', "");
Expect(1, 78894, '\p{Is_Script_Extensions= egyptian_HIEROGLYPHS}', "");
Expect(0, 78894, '\p{^Is_Script_Extensions= egyptian_HIEROGLYPHS}', "");
Expect(0, 78894, '\P{Is_Script_Extensions= egyptian_HIEROGLYPHS}', "");
Expect(1, 78894, '\P{^Is_Script_Extensions= egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\p{Is_Script_Extensions= egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\p{^Is_Script_Extensions= egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\P{Is_Script_Extensions= egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\P{^Is_Script_Extensions= egyptian_HIEROGLYPHS}', "");
Error('\p{Is_Scx=:=	 Egyp}');
Error('\P{Is_Scx=:=	 Egyp}');
Expect(1, 78894, '\p{Is_Scx=egyp}', "");
Expect(0, 78894, '\p{^Is_Scx=egyp}', "");
Expect(0, 78894, '\P{Is_Scx=egyp}', "");
Expect(1, 78894, '\P{^Is_Scx=egyp}', "");
Expect(0, 78895, '\p{Is_Scx=egyp}', "");
Expect(1, 78895, '\p{^Is_Scx=egyp}', "");
Expect(1, 78895, '\P{Is_Scx=egyp}', "");
Expect(0, 78895, '\P{^Is_Scx=egyp}', "");
Expect(1, 78894, '\p{Is_Scx=	-egyp}', "");
Expect(0, 78894, '\p{^Is_Scx=	-egyp}', "");
Expect(0, 78894, '\P{Is_Scx=	-egyp}', "");
Expect(1, 78894, '\P{^Is_Scx=	-egyp}', "");
Expect(0, 78895, '\p{Is_Scx=	-egyp}', "");
Expect(1, 78895, '\p{^Is_Scx=	-egyp}', "");
Expect(1, 78895, '\P{Is_Scx=	-egyp}', "");
Expect(0, 78895, '\P{^Is_Scx=	-egyp}', "");
Error('\p{Script_Extensions=:=	-javanese}');
Error('\P{Script_Extensions=:=	-javanese}');
Expect(1, 43487, '\p{Script_Extensions=javanese}', "");
Expect(0, 43487, '\p{^Script_Extensions=javanese}', "");
Expect(0, 43487, '\P{Script_Extensions=javanese}', "");
Expect(1, 43487, '\P{^Script_Extensions=javanese}', "");
Expect(0, 43488, '\p{Script_Extensions=javanese}', "");
Expect(1, 43488, '\p{^Script_Extensions=javanese}', "");
Expect(1, 43488, '\P{Script_Extensions=javanese}', "");
Expect(0, 43488, '\P{^Script_Extensions=javanese}', "");
Expect(1, 43487, '\p{Script_Extensions=JAVANESE}', "");
Expect(0, 43487, '\p{^Script_Extensions=JAVANESE}', "");
Expect(0, 43487, '\P{Script_Extensions=JAVANESE}', "");
Expect(1, 43487, '\P{^Script_Extensions=JAVANESE}', "");
Expect(0, 43488, '\p{Script_Extensions=JAVANESE}', "");
Expect(1, 43488, '\p{^Script_Extensions=JAVANESE}', "");
Expect(1, 43488, '\P{Script_Extensions=JAVANESE}', "");
Expect(0, 43488, '\P{^Script_Extensions=JAVANESE}', "");
Error('\p{Scx=:= JAVA}');
Error('\P{Scx=:= JAVA}');
Expect(1, 43487, '\p{Scx=java}', "");
Expect(0, 43487, '\p{^Scx=java}', "");
Expect(0, 43487, '\P{Scx=java}', "");
Expect(1, 43487, '\P{^Scx=java}', "");
Expect(0, 43488, '\p{Scx=java}', "");
Expect(1, 43488, '\p{^Scx=java}', "");
Expect(1, 43488, '\P{Scx=java}', "");
Expect(0, 43488, '\P{^Scx=java}', "");
Expect(1, 43487, '\p{Scx:	 	JAVA}', "");
Expect(0, 43487, '\p{^Scx:	 	JAVA}', "");
Expect(0, 43487, '\P{Scx:	 	JAVA}', "");
Expect(1, 43487, '\P{^Scx:	 	JAVA}', "");
Expect(0, 43488, '\p{Scx:	 	JAVA}', "");
Expect(1, 43488, '\p{^Scx:	 	JAVA}', "");
Expect(1, 43488, '\P{Scx:	 	JAVA}', "");
Expect(0, 43488, '\P{^Scx:	 	JAVA}', "");
Error('\p{Is_Script_Extensions=	:=Javanese}');
Error('\P{Is_Script_Extensions=	:=Javanese}');
Expect(1, 43487, '\p{Is_Script_Extensions=javanese}', "");
Expect(0, 43487, '\p{^Is_Script_Extensions=javanese}', "");
Expect(0, 43487, '\P{Is_Script_Extensions=javanese}', "");
Expect(1, 43487, '\P{^Is_Script_Extensions=javanese}', "");
Expect(0, 43488, '\p{Is_Script_Extensions=javanese}', "");
Expect(1, 43488, '\p{^Is_Script_Extensions=javanese}', "");
Expect(1, 43488, '\P{Is_Script_Extensions=javanese}', "");
Expect(0, 43488, '\P{^Is_Script_Extensions=javanese}', "");
Expect(1, 43487, '\p{Is_Script_Extensions=--Javanese}', "");
Expect(0, 43487, '\p{^Is_Script_Extensions=--Javanese}', "");
Expect(0, 43487, '\P{Is_Script_Extensions=--Javanese}', "");
Expect(1, 43487, '\P{^Is_Script_Extensions=--Javanese}', "");
Expect(0, 43488, '\p{Is_Script_Extensions=--Javanese}', "");
Expect(1, 43488, '\p{^Is_Script_Extensions=--Javanese}', "");
Expect(1, 43488, '\P{Is_Script_Extensions=--Javanese}', "");
Expect(0, 43488, '\P{^Is_Script_Extensions=--Javanese}', "");
Error('\p{Is_Scx=-_Java:=}');
Error('\P{Is_Scx=-_Java:=}');
Expect(1, 43487, '\p{Is_Scx=java}', "");
Expect(0, 43487, '\p{^Is_Scx=java}', "");
Expect(0, 43487, '\P{Is_Scx=java}', "");
Expect(1, 43487, '\P{^Is_Scx=java}', "");
Expect(0, 43488, '\p{Is_Scx=java}', "");
Expect(1, 43488, '\p{^Is_Scx=java}', "");
Expect(1, 43488, '\P{Is_Scx=java}', "");
Expect(0, 43488, '\P{^Is_Scx=java}', "");
Expect(1, 43487, '\p{Is_Scx:  JAVA}', "");
Expect(0, 43487, '\p{^Is_Scx:  JAVA}', "");
Expect(0, 43487, '\P{Is_Scx:  JAVA}', "");
Expect(1, 43487, '\P{^Is_Scx:  JAVA}', "");
Expect(0, 43488, '\p{Is_Scx:  JAVA}', "");
Expect(1, 43488, '\p{^Is_Scx:  JAVA}', "");
Expect(1, 43488, '\P{Is_Scx:  JAVA}', "");
Expect(0, 43488, '\P{^Is_Scx:  JAVA}', "");
Error('\p{Script_Extensions=:=_-Chakma}');
Error('\P{Script_Extensions=:=_-Chakma}');
Expect(1, 69955, '\p{Script_Extensions=chakma}', "");
Expect(0, 69955, '\p{^Script_Extensions=chakma}', "");
Expect(0, 69955, '\P{Script_Extensions=chakma}', "");
Expect(1, 69955, '\P{^Script_Extensions=chakma}', "");
Expect(0, 69956, '\p{Script_Extensions=chakma}', "");
Expect(1, 69956, '\p{^Script_Extensions=chakma}', "");
Expect(1, 69956, '\P{Script_Extensions=chakma}', "");
Expect(0, 69956, '\P{^Script_Extensions=chakma}', "");
Expect(1, 69955, '\p{Script_Extensions=-	CHAKMA}', "");
Expect(0, 69955, '\p{^Script_Extensions=-	CHAKMA}', "");
Expect(0, 69955, '\P{Script_Extensions=-	CHAKMA}', "");
Expect(1, 69955, '\P{^Script_Extensions=-	CHAKMA}', "");
Expect(0, 69956, '\p{Script_Extensions=-	CHAKMA}', "");
Expect(1, 69956, '\p{^Script_Extensions=-	CHAKMA}', "");
Expect(1, 69956, '\P{Script_Extensions=-	CHAKMA}', "");
Expect(0, 69956, '\P{^Script_Extensions=-	CHAKMA}', "");
Error('\p{Scx=Cakm:=}');
Error('\P{Scx=Cakm:=}');
Expect(1, 69955, '\p{Scx=cakm}', "");
Expect(0, 69955, '\p{^Scx=cakm}', "");
Expect(0, 69955, '\P{Scx=cakm}', "");
Expect(1, 69955, '\P{^Scx=cakm}', "");
Expect(0, 69956, '\p{Scx=cakm}', "");
Expect(1, 69956, '\p{^Scx=cakm}', "");
Expect(1, 69956, '\P{Scx=cakm}', "");
Expect(0, 69956, '\P{^Scx=cakm}', "");
Expect(1, 69955, '\p{Scx=	cakm}', "");
Expect(0, 69955, '\p{^Scx=	cakm}', "");
Expect(0, 69955, '\P{Scx=	cakm}', "");
Expect(1, 69955, '\P{^Scx=	cakm}', "");
Expect(0, 69956, '\p{Scx=	cakm}', "");
Expect(1, 69956, '\p{^Scx=	cakm}', "");
Expect(1, 69956, '\P{Scx=	cakm}', "");
Expect(0, 69956, '\P{^Scx=	cakm}', "");
Error('\p{Is_Script_Extensions=:= _Chakma}');
Error('\P{Is_Script_Extensions=:= _Chakma}');
Expect(1, 69955, '\p{Is_Script_Extensions=chakma}', "");
Expect(0, 69955, '\p{^Is_Script_Extensions=chakma}', "");
Expect(0, 69955, '\P{Is_Script_Extensions=chakma}', "");
Expect(1, 69955, '\P{^Is_Script_Extensions=chakma}', "");
Expect(0, 69956, '\p{Is_Script_Extensions=chakma}', "");
Expect(1, 69956, '\p{^Is_Script_Extensions=chakma}', "");
Expect(1, 69956, '\P{Is_Script_Extensions=chakma}', "");
Expect(0, 69956, '\P{^Is_Script_Extensions=chakma}', "");
Expect(1, 69955, '\p{Is_Script_Extensions=	_CHAKMA}', "");
Expect(0, 69955, '\p{^Is_Script_Extensions=	_CHAKMA}', "");
Expect(0, 69955, '\P{Is_Script_Extensions=	_CHAKMA}', "");
Expect(1, 69955, '\P{^Is_Script_Extensions=	_CHAKMA}', "");
Expect(0, 69956, '\p{Is_Script_Extensions=	_CHAKMA}', "");
Expect(1, 69956, '\p{^Is_Script_Extensions=	_CHAKMA}', "");
Expect(1, 69956, '\P{Is_Script_Extensions=	_CHAKMA}', "");
Expect(0, 69956, '\P{^Is_Script_Extensions=	_CHAKMA}', "");
Error('\p{Is_Scx=-:=CAKM}');
Error('\P{Is_Scx=-:=CAKM}');
Expect(1, 69955, '\p{Is_Scx=cakm}', "");
Expect(0, 69955, '\p{^Is_Scx=cakm}', "");
Expect(0, 69955, '\P{Is_Scx=cakm}', "");
Expect(1, 69955, '\P{^Is_Scx=cakm}', "");
Expect(0, 69956, '\p{Is_Scx=cakm}', "");
Expect(1, 69956, '\p{^Is_Scx=cakm}', "");
Expect(1, 69956, '\P{Is_Scx=cakm}', "");
Expect(0, 69956, '\P{^Is_Scx=cakm}', "");
Expect(1, 69955, '\p{Is_Scx=	-cakm}', "");
Expect(0, 69955, '\p{^Is_Scx=	-cakm}', "");
Expect(0, 69955, '\P{Is_Scx=	-cakm}', "");
Expect(1, 69955, '\P{^Is_Scx=	-cakm}', "");
Expect(0, 69956, '\p{Is_Scx=	-cakm}', "");
Expect(1, 69956, '\p{^Is_Scx=	-cakm}', "");
Expect(1, 69956, '\P{Is_Scx=	-cakm}', "");
Expect(0, 69956, '\P{^Is_Scx=	-cakm}', "");
Error('\p{Script_Extensions=/a/_ Common}');
Error('\P{Script_Extensions=/a/_ Common}');
Expect(1, 917631, '\p{Script_Extensions=common}', "");
Expect(0, 917631, '\p{^Script_Extensions=common}', "");
Expect(0, 917631, '\P{Script_Extensions=common}', "");
Expect(1, 917631, '\P{^Script_Extensions=common}', "");
Expect(0, 917632, '\p{Script_Extensions=common}', "");
Expect(1, 917632, '\p{^Script_Extensions=common}', "");
Expect(1, 917632, '\P{Script_Extensions=common}', "");
Expect(0, 917632, '\P{^Script_Extensions=common}', "");
Error('\p{Scx=/a/_ zyyy}');
Error('\P{Scx=/a/_ zyyy}');
Expect(1, 917631, '\p{Scx=zyyy}', "");
Expect(0, 917631, '\p{^Scx=zyyy}', "");
Expect(0, 917631, '\P{Scx=zyyy}', "");
Expect(1, 917631, '\P{^Scx=zyyy}', "");
Expect(0, 917632, '\p{Scx=zyyy}', "");
Expect(1, 917632, '\p{^Scx=zyyy}', "");
Expect(1, 917632, '\P{Scx=zyyy}', "");
Expect(0, 917632, '\P{^Scx=zyyy}', "");
Expect(1, 917631, '\p{Scx= zyyy}', "");
Expect(0, 917631, '\p{^Scx= zyyy}', "");
Expect(0, 917631, '\P{Scx= zyyy}', "");
Expect(1, 917631, '\P{^Scx= zyyy}', "");
Expect(0, 917632, '\p{Scx= zyyy}', "");
Expect(1, 917632, '\p{^Scx= zyyy}', "");
Expect(1, 917632, '\P{Scx= zyyy}', "");
Expect(0, 917632, '\P{^Scx= zyyy}', "");
Error('\p{Is_Script_Extensions=/a/_ Common}');
Error('\P{Is_Script_Extensions=/a/_ Common}');
Expect(1, 917631, '\p{Is_Script_Extensions=common}', "");
Expect(0, 917631, '\p{^Is_Script_Extensions=common}', "");
Expect(0, 917631, '\P{Is_Script_Extensions=common}', "");
Expect(1, 917631, '\P{^Is_Script_Extensions=common}', "");
Expect(0, 917632, '\p{Is_Script_Extensions=common}', "");
Expect(1, 917632, '\p{^Is_Script_Extensions=common}', "");
Expect(1, 917632, '\P{Is_Script_Extensions=common}', "");
Expect(0, 917632, '\P{^Is_Script_Extensions=common}', "");
Expect(1, 917631, '\p{Is_Script_Extensions= _common}', "");
Expect(0, 917631, '\p{^Is_Script_Extensions= _common}', "");
Expect(0, 917631, '\P{Is_Script_Extensions= _common}', "");
Expect(1, 917631, '\P{^Is_Script_Extensions= _common}', "");
Expect(0, 917632, '\p{Is_Script_Extensions= _common}', "");
Expect(1, 917632, '\p{^Is_Script_Extensions= _common}', "");
Expect(1, 917632, '\P{Is_Script_Extensions= _common}', "");
Expect(0, 917632, '\P{^Is_Script_Extensions= _common}', "");
Error('\p{Is_Scx=:=Zyyy}');
Error('\P{Is_Scx=:=Zyyy}');
Expect(1, 917631, '\p{Is_Scx=zyyy}', "");
Expect(0, 917631, '\p{^Is_Scx=zyyy}', "");
Expect(0, 917631, '\P{Is_Scx=zyyy}', "");
Expect(1, 917631, '\P{^Is_Scx=zyyy}', "");
Expect(0, 917632, '\p{Is_Scx=zyyy}', "");
Expect(1, 917632, '\p{^Is_Scx=zyyy}', "");
Expect(1, 917632, '\P{Is_Scx=zyyy}', "");
Expect(0, 917632, '\P{^Is_Scx=zyyy}', "");
Expect(1, 917631, '\p{Is_Scx=_Zyyy}', "");
Expect(0, 917631, '\p{^Is_Scx=_Zyyy}', "");
Expect(0, 917631, '\P{Is_Scx=_Zyyy}', "");
Expect(1, 917631, '\P{^Is_Scx=_Zyyy}', "");
Expect(0, 917632, '\p{Is_Scx=_Zyyy}', "");
Expect(1, 917632, '\p{^Is_Scx=_Zyyy}', "");
Expect(1, 917632, '\P{Is_Scx=_Zyyy}', "");
Expect(0, 917632, '\P{^Is_Scx=_Zyyy}', "");
Error('\p{Script_Extensions:	/a/Latin}');
Error('\P{Script_Extensions:	/a/Latin}');
Expect(1, 65370, '\p{Script_Extensions=latin}', "");
Expect(0, 65370, '\p{^Script_Extensions=latin}', "");
Expect(0, 65370, '\P{Script_Extensions=latin}', "");
Expect(1, 65370, '\P{^Script_Extensions=latin}', "");
Expect(0, 65371, '\p{Script_Extensions=latin}', "");
Expect(1, 65371, '\p{^Script_Extensions=latin}', "");
Expect(1, 65371, '\P{Script_Extensions=latin}', "");
Expect(0, 65371, '\P{^Script_Extensions=latin}', "");
Expect(1, 65370, '\p{Script_Extensions=- Latin}', "");
Expect(0, 65370, '\p{^Script_Extensions=- Latin}', "");
Expect(0, 65370, '\P{Script_Extensions=- Latin}', "");
Expect(1, 65370, '\P{^Script_Extensions=- Latin}', "");
Expect(0, 65371, '\p{Script_Extensions=- Latin}', "");
Expect(1, 65371, '\p{^Script_Extensions=- Latin}', "");
Expect(1, 65371, '\P{Script_Extensions=- Latin}', "");
Expect(0, 65371, '\P{^Script_Extensions=- Latin}', "");
Error('\p{Scx=_:=Latn}');
Error('\P{Scx=_:=Latn}');
Expect(1, 65370, '\p{Scx:latn}', "");
Expect(0, 65370, '\p{^Scx:latn}', "");
Expect(0, 65370, '\P{Scx:latn}', "");
Expect(1, 65370, '\P{^Scx:latn}', "");
Expect(0, 65371, '\p{Scx:latn}', "");
Expect(1, 65371, '\p{^Scx:latn}', "");
Expect(1, 65371, '\P{Scx:latn}', "");
Expect(0, 65371, '\P{^Scx:latn}', "");
Expect(1, 65370, '\p{Scx=_-LATN}', "");
Expect(0, 65370, '\p{^Scx=_-LATN}', "");
Expect(0, 65370, '\P{Scx=_-LATN}', "");
Expect(1, 65370, '\P{^Scx=_-LATN}', "");
Expect(0, 65371, '\p{Scx=_-LATN}', "");
Expect(1, 65371, '\p{^Scx=_-LATN}', "");
Expect(1, 65371, '\P{Scx=_-LATN}', "");
Expect(0, 65371, '\P{^Scx=_-LATN}', "");
Error('\p{Is_Script_Extensions=_ Latin:=}');
Error('\P{Is_Script_Extensions=_ Latin:=}');
Expect(1, 65370, '\p{Is_Script_Extensions=latin}', "");
Expect(0, 65370, '\p{^Is_Script_Extensions=latin}', "");
Expect(0, 65370, '\P{Is_Script_Extensions=latin}', "");
Expect(1, 65370, '\P{^Is_Script_Extensions=latin}', "");
Expect(0, 65371, '\p{Is_Script_Extensions=latin}', "");
Expect(1, 65371, '\p{^Is_Script_Extensions=latin}', "");
Expect(1, 65371, '\P{Is_Script_Extensions=latin}', "");
Expect(0, 65371, '\P{^Is_Script_Extensions=latin}', "");
Expect(1, 65370, '\p{Is_Script_Extensions=__Latin}', "");
Expect(0, 65370, '\p{^Is_Script_Extensions=__Latin}', "");
Expect(0, 65370, '\P{Is_Script_Extensions=__Latin}', "");
Expect(1, 65370, '\P{^Is_Script_Extensions=__Latin}', "");
Expect(0, 65371, '\p{Is_Script_Extensions=__Latin}', "");
Expect(1, 65371, '\p{^Is_Script_Extensions=__Latin}', "");
Expect(1, 65371, '\P{Is_Script_Extensions=__Latin}', "");
Expect(0, 65371, '\P{^Is_Script_Extensions=__Latin}', "");
Error('\p{Is_Scx=_-Latn/a/}');
Error('\P{Is_Scx=_-Latn/a/}');
Expect(1, 65370, '\p{Is_Scx=latn}', "");
Expect(0, 65370, '\p{^Is_Scx=latn}', "");
Expect(0, 65370, '\P{Is_Scx=latn}', "");
Expect(1, 65370, '\P{^Is_Scx=latn}', "");
Expect(0, 65371, '\p{Is_Scx=latn}', "");
Expect(1, 65371, '\p{^Is_Scx=latn}', "");
Expect(1, 65371, '\P{Is_Scx=latn}', "");
Expect(0, 65371, '\P{^Is_Scx=latn}', "");
Expect(1, 65370, '\p{Is_Scx=Latn}', "");
Expect(0, 65370, '\p{^Is_Scx=Latn}', "");
Expect(0, 65370, '\P{Is_Scx=Latn}', "");
Expect(1, 65370, '\P{^Is_Scx=Latn}', "");
Expect(0, 65371, '\p{Is_Scx=Latn}', "");
Expect(1, 65371, '\p{^Is_Scx=Latn}', "");
Expect(1, 65371, '\P{Is_Scx=Latn}', "");
Expect(0, 65371, '\P{^Is_Scx=Latn}', "");
Error('\p{Script_Extensions=/a/Ethiopic}');
Error('\P{Script_Extensions=/a/Ethiopic}');
Expect(1, 43822, '\p{Script_Extensions=ethiopic}', "");
Expect(0, 43822, '\p{^Script_Extensions=ethiopic}', "");
Expect(0, 43822, '\P{Script_Extensions=ethiopic}', "");
Expect(1, 43822, '\P{^Script_Extensions=ethiopic}', "");
Expect(0, 43823, '\p{Script_Extensions=ethiopic}', "");
Expect(1, 43823, '\p{^Script_Extensions=ethiopic}', "");
Expect(1, 43823, '\P{Script_Extensions=ethiopic}', "");
Expect(0, 43823, '\P{^Script_Extensions=ethiopic}', "");
Expect(1, 43822, '\p{Script_Extensions=	 ethiopic}', "");
Expect(0, 43822, '\p{^Script_Extensions=	 ethiopic}', "");
Expect(0, 43822, '\P{Script_Extensions=	 ethiopic}', "");
Expect(1, 43822, '\P{^Script_Extensions=	 ethiopic}', "");
Expect(0, 43823, '\p{Script_Extensions=	 ethiopic}', "");
Expect(1, 43823, '\p{^Script_Extensions=	 ethiopic}', "");
Expect(1, 43823, '\P{Script_Extensions=	 ethiopic}', "");
Expect(0, 43823, '\P{^Script_Extensions=	 ethiopic}', "");
Error('\p{Scx:   __ethi/a/}');
Error('\P{Scx:   __ethi/a/}');
Expect(1, 43822, '\p{Scx=ethi}', "");
Expect(0, 43822, '\p{^Scx=ethi}', "");
Expect(0, 43822, '\P{Scx=ethi}', "");
Expect(1, 43822, '\P{^Scx=ethi}', "");
Expect(0, 43823, '\p{Scx=ethi}', "");
Expect(1, 43823, '\p{^Scx=ethi}', "");
Expect(1, 43823, '\P{Scx=ethi}', "");
Expect(0, 43823, '\P{^Scx=ethi}', "");
Expect(1, 43822, '\p{Scx=-	Ethi}', "");
Expect(0, 43822, '\p{^Scx=-	Ethi}', "");
Expect(0, 43822, '\P{Scx=-	Ethi}', "");
Expect(1, 43822, '\P{^Scx=-	Ethi}', "");
Expect(0, 43823, '\p{Scx=-	Ethi}', "");
Expect(1, 43823, '\p{^Scx=-	Ethi}', "");
Expect(1, 43823, '\P{Scx=-	Ethi}', "");
Expect(0, 43823, '\P{^Scx=-	Ethi}', "");
Error('\p{Is_Script_Extensions=:=_ Ethiopic}');
Error('\P{Is_Script_Extensions=:=_ Ethiopic}');
Expect(1, 43822, '\p{Is_Script_Extensions=ethiopic}', "");
Expect(0, 43822, '\p{^Is_Script_Extensions=ethiopic}', "");
Expect(0, 43822, '\P{Is_Script_Extensions=ethiopic}', "");
Expect(1, 43822, '\P{^Is_Script_Extensions=ethiopic}', "");
Expect(0, 43823, '\p{Is_Script_Extensions=ethiopic}', "");
Expect(1, 43823, '\p{^Is_Script_Extensions=ethiopic}', "");
Expect(1, 43823, '\P{Is_Script_Extensions=ethiopic}', "");
Expect(0, 43823, '\P{^Is_Script_Extensions=ethiopic}', "");
Expect(1, 43822, '\p{Is_Script_Extensions=		Ethiopic}', "");
Expect(0, 43822, '\p{^Is_Script_Extensions=		Ethiopic}', "");
Expect(0, 43822, '\P{Is_Script_Extensions=		Ethiopic}', "");
Expect(1, 43822, '\P{^Is_Script_Extensions=		Ethiopic}', "");
Expect(0, 43823, '\p{Is_Script_Extensions=		Ethiopic}', "");
Expect(1, 43823, '\p{^Is_Script_Extensions=		Ethiopic}', "");
Expect(1, 43823, '\P{Is_Script_Extensions=		Ethiopic}', "");
Expect(0, 43823, '\P{^Is_Script_Extensions=		Ethiopic}', "");
Error('\p{Is_Scx=_	ethi:=}');
Error('\P{Is_Scx=_	ethi:=}');
Expect(1, 43822, '\p{Is_Scx=ethi}', "");
Expect(0, 43822, '\p{^Is_Scx=ethi}', "");
Expect(0, 43822, '\P{Is_Scx=ethi}', "");
Expect(1, 43822, '\P{^Is_Scx=ethi}', "");
Expect(0, 43823, '\p{Is_Scx=ethi}', "");
Expect(1, 43823, '\p{^Is_Scx=ethi}', "");
Expect(1, 43823, '\P{Is_Scx=ethi}', "");
Expect(0, 43823, '\P{^Is_Scx=ethi}', "");
Expect(1, 43822, '\p{Is_Scx=_-ethi}', "");
Expect(0, 43822, '\p{^Is_Scx=_-ethi}', "");
Expect(0, 43822, '\P{Is_Scx=_-ethi}', "");
Expect(1, 43822, '\P{^Is_Scx=_-ethi}', "");
Expect(0, 43823, '\p{Is_Scx=_-ethi}', "");
Expect(1, 43823, '\p{^Is_Scx=_-ethi}', "");
Expect(1, 43823, '\P{Is_Scx=_-ethi}', "");
Expect(0, 43823, '\P{^Is_Scx=_-ethi}', "");
Error('\p{Script_Extensions=Buginese/a/}');
Error('\P{Script_Extensions=Buginese/a/}');
Expect(1, 6687, '\p{Script_Extensions:buginese}', "");
Expect(0, 6687, '\p{^Script_Extensions:buginese}', "");
Expect(0, 6687, '\P{Script_Extensions:buginese}', "");
Expect(1, 6687, '\P{^Script_Extensions:buginese}', "");
Expect(0, 6688, '\p{Script_Extensions:buginese}', "");
Expect(1, 6688, '\p{^Script_Extensions:buginese}', "");
Expect(1, 6688, '\P{Script_Extensions:buginese}', "");
Expect(0, 6688, '\P{^Script_Extensions:buginese}', "");
Expect(1, 6687, '\p{Script_Extensions= -BUGINESE}', "");
Expect(0, 6687, '\p{^Script_Extensions= -BUGINESE}', "");
Expect(0, 6687, '\P{Script_Extensions= -BUGINESE}', "");
Expect(1, 6687, '\P{^Script_Extensions= -BUGINESE}', "");
Expect(0, 6688, '\p{Script_Extensions= -BUGINESE}', "");
Expect(1, 6688, '\p{^Script_Extensions= -BUGINESE}', "");
Expect(1, 6688, '\P{Script_Extensions= -BUGINESE}', "");
Expect(0, 6688, '\P{^Script_Extensions= -BUGINESE}', "");
Error('\p{Scx=/a/	Bugi}');
Error('\P{Scx=/a/	Bugi}');
Expect(1, 6687, '\p{Scx=bugi}', "");
Expect(0, 6687, '\p{^Scx=bugi}', "");
Expect(0, 6687, '\P{Scx=bugi}', "");
Expect(1, 6687, '\P{^Scx=bugi}', "");
Expect(0, 6688, '\p{Scx=bugi}', "");
Expect(1, 6688, '\p{^Scx=bugi}', "");
Expect(1, 6688, '\P{Scx=bugi}', "");
Expect(0, 6688, '\P{^Scx=bugi}', "");
Expect(1, 6687, '\p{Scx=-	Bugi}', "");
Expect(0, 6687, '\p{^Scx=-	Bugi}', "");
Expect(0, 6687, '\P{Scx=-	Bugi}', "");
Expect(1, 6687, '\P{^Scx=-	Bugi}', "");
Expect(0, 6688, '\p{Scx=-	Bugi}', "");
Expect(1, 6688, '\p{^Scx=-	Bugi}', "");
Expect(1, 6688, '\P{Scx=-	Bugi}', "");
Expect(0, 6688, '\P{^Scx=-	Bugi}', "");
Error('\p{Is_Script_Extensions= Buginese/a/}');
Error('\P{Is_Script_Extensions= Buginese/a/}');
Expect(1, 6687, '\p{Is_Script_Extensions=buginese}', "");
Expect(0, 6687, '\p{^Is_Script_Extensions=buginese}', "");
Expect(0, 6687, '\P{Is_Script_Extensions=buginese}', "");
Expect(1, 6687, '\P{^Is_Script_Extensions=buginese}', "");
Expect(0, 6688, '\p{Is_Script_Extensions=buginese}', "");
Expect(1, 6688, '\p{^Is_Script_Extensions=buginese}', "");
Expect(1, 6688, '\P{Is_Script_Extensions=buginese}', "");
Expect(0, 6688, '\P{^Is_Script_Extensions=buginese}', "");
Expect(1, 6687, '\p{Is_Script_Extensions:Buginese}', "");
Expect(0, 6687, '\p{^Is_Script_Extensions:Buginese}', "");
Expect(0, 6687, '\P{Is_Script_Extensions:Buginese}', "");
Expect(1, 6687, '\P{^Is_Script_Extensions:Buginese}', "");
Expect(0, 6688, '\p{Is_Script_Extensions:Buginese}', "");
Expect(1, 6688, '\p{^Is_Script_Extensions:Buginese}', "");
Expect(1, 6688, '\P{Is_Script_Extensions:Buginese}', "");
Expect(0, 6688, '\P{^Is_Script_Extensions:Buginese}', "");
Error('\p{Is_Scx:   :=-_BUGI}');
Error('\P{Is_Scx:   :=-_BUGI}');
Expect(1, 6687, '\p{Is_Scx:	bugi}', "");
Expect(0, 6687, '\p{^Is_Scx:	bugi}', "");
Expect(0, 6687, '\P{Is_Scx:	bugi}', "");
Expect(1, 6687, '\P{^Is_Scx:	bugi}', "");
Expect(0, 6688, '\p{Is_Scx:	bugi}', "");
Expect(1, 6688, '\p{^Is_Scx:	bugi}', "");
Expect(1, 6688, '\P{Is_Scx:	bugi}', "");
Expect(0, 6688, '\P{^Is_Scx:	bugi}', "");
Expect(1, 6687, '\p{Is_Scx=	BUGI}', "");
Expect(0, 6687, '\p{^Is_Scx=	BUGI}', "");
Expect(0, 6687, '\P{Is_Scx=	BUGI}', "");
Expect(1, 6687, '\P{^Is_Scx=	BUGI}', "");
Expect(0, 6688, '\p{Is_Scx=	BUGI}', "");
Expect(1, 6688, '\p{^Is_Scx=	BUGI}', "");
Expect(1, 6688, '\P{Is_Scx=	BUGI}', "");
Expect(0, 6688, '\P{^Is_Scx=	BUGI}', "");
Error('\p{Script_Extensions: -:=Hebrew}');
Error('\P{Script_Extensions: -:=Hebrew}');
Expect(1, 64335, '\p{Script_Extensions=hebrew}', "");
Expect(0, 64335, '\p{^Script_Extensions=hebrew}', "");
Expect(0, 64335, '\P{Script_Extensions=hebrew}', "");
Expect(1, 64335, '\P{^Script_Extensions=hebrew}', "");
Expect(0, 64336, '\p{Script_Extensions=hebrew}', "");
Expect(1, 64336, '\p{^Script_Extensions=hebrew}', "");
Expect(1, 64336, '\P{Script_Extensions=hebrew}', "");
Expect(0, 64336, '\P{^Script_Extensions=hebrew}', "");
Expect(1, 64335, '\p{Script_Extensions= -Hebrew}', "");
Expect(0, 64335, '\p{^Script_Extensions= -Hebrew}', "");
Expect(0, 64335, '\P{Script_Extensions= -Hebrew}', "");
Expect(1, 64335, '\P{^Script_Extensions= -Hebrew}', "");
Expect(0, 64336, '\p{Script_Extensions= -Hebrew}', "");
Expect(1, 64336, '\p{^Script_Extensions= -Hebrew}', "");
Expect(1, 64336, '\P{Script_Extensions= -Hebrew}', "");
Expect(0, 64336, '\P{^Script_Extensions= -Hebrew}', "");
Error('\p{Scx= _Hebr/a/}');
Error('\P{Scx= _Hebr/a/}');
Expect(1, 64335, '\p{Scx=hebr}', "");
Expect(0, 64335, '\p{^Scx=hebr}', "");
Expect(0, 64335, '\P{Scx=hebr}', "");
Expect(1, 64335, '\P{^Scx=hebr}', "");
Expect(0, 64336, '\p{Scx=hebr}', "");
Expect(1, 64336, '\p{^Scx=hebr}', "");
Expect(1, 64336, '\P{Scx=hebr}', "");
Expect(0, 64336, '\P{^Scx=hebr}', "");
Expect(1, 64335, '\p{Scx=  hebr}', "");
Expect(0, 64335, '\p{^Scx=  hebr}', "");
Expect(0, 64335, '\P{Scx=  hebr}', "");
Expect(1, 64335, '\P{^Scx=  hebr}', "");
Expect(0, 64336, '\p{Scx=  hebr}', "");
Expect(1, 64336, '\p{^Scx=  hebr}', "");
Expect(1, 64336, '\P{Scx=  hebr}', "");
Expect(0, 64336, '\P{^Scx=  hebr}', "");
Error('\p{Is_Script_Extensions=_:=HEBREW}');
Error('\P{Is_Script_Extensions=_:=HEBREW}');
Expect(1, 64335, '\p{Is_Script_Extensions=hebrew}', "");
Expect(0, 64335, '\p{^Is_Script_Extensions=hebrew}', "");
Expect(0, 64335, '\P{Is_Script_Extensions=hebrew}', "");
Expect(1, 64335, '\P{^Is_Script_Extensions=hebrew}', "");
Expect(0, 64336, '\p{Is_Script_Extensions=hebrew}', "");
Expect(1, 64336, '\p{^Is_Script_Extensions=hebrew}', "");
Expect(1, 64336, '\P{Is_Script_Extensions=hebrew}', "");
Expect(0, 64336, '\P{^Is_Script_Extensions=hebrew}', "");
Expect(1, 64335, '\p{Is_Script_Extensions: - Hebrew}', "");
Expect(0, 64335, '\p{^Is_Script_Extensions: - Hebrew}', "");
Expect(0, 64335, '\P{Is_Script_Extensions: - Hebrew}', "");
Expect(1, 64335, '\P{^Is_Script_Extensions: - Hebrew}', "");
Expect(0, 64336, '\p{Is_Script_Extensions: - Hebrew}', "");
Expect(1, 64336, '\p{^Is_Script_Extensions: - Hebrew}', "");
Expect(1, 64336, '\P{Is_Script_Extensions: - Hebrew}', "");
Expect(0, 64336, '\P{^Is_Script_Extensions: - Hebrew}', "");
Error('\p{Is_Scx:   Hebr:=}');
Error('\P{Is_Scx:   Hebr:=}');
Expect(1, 64335, '\p{Is_Scx:hebr}', "");
Expect(0, 64335, '\p{^Is_Scx:hebr}', "");
Expect(0, 64335, '\P{Is_Scx:hebr}', "");
Expect(1, 64335, '\P{^Is_Scx:hebr}', "");
Expect(0, 64336, '\p{Is_Scx:hebr}', "");
Expect(1, 64336, '\p{^Is_Scx:hebr}', "");
Expect(1, 64336, '\P{Is_Scx:hebr}', "");
Expect(0, 64336, '\P{^Is_Scx:hebr}', "");
Expect(1, 64335, '\p{Is_Scx=--HEBR}', "");
Expect(0, 64335, '\p{^Is_Scx=--HEBR}', "");
Expect(0, 64335, '\P{Is_Scx=--HEBR}', "");
Expect(1, 64335, '\P{^Is_Scx=--HEBR}', "");
Expect(0, 64336, '\p{Is_Scx=--HEBR}', "");
Expect(1, 64336, '\p{^Is_Scx=--HEBR}', "");
Expect(1, 64336, '\P{Is_Scx=--HEBR}', "");
Expect(0, 64336, '\P{^Is_Scx=--HEBR}', "");
Error('\p{Script_Extensions=:= _SORA_Sompeng}');
Error('\P{Script_Extensions=:= _SORA_Sompeng}');
Expect(1, 69881, '\p{Script_Extensions=sorasompeng}', "");
Expect(0, 69881, '\p{^Script_Extensions=sorasompeng}', "");
Expect(0, 69881, '\P{Script_Extensions=sorasompeng}', "");
Expect(1, 69881, '\P{^Script_Extensions=sorasompeng}', "");
Expect(0, 69882, '\p{Script_Extensions=sorasompeng}', "");
Expect(1, 69882, '\p{^Script_Extensions=sorasompeng}', "");
Expect(1, 69882, '\P{Script_Extensions=sorasompeng}', "");
Expect(0, 69882, '\P{^Script_Extensions=sorasompeng}', "");
Expect(1, 69881, '\p{Script_Extensions= 	sora_Sompeng}', "");
Expect(0, 69881, '\p{^Script_Extensions= 	sora_Sompeng}', "");
Expect(0, 69881, '\P{Script_Extensions= 	sora_Sompeng}', "");
Expect(1, 69881, '\P{^Script_Extensions= 	sora_Sompeng}', "");
Expect(0, 69882, '\p{Script_Extensions= 	sora_Sompeng}', "");
Expect(1, 69882, '\p{^Script_Extensions= 	sora_Sompeng}', "");
Expect(1, 69882, '\P{Script_Extensions= 	sora_Sompeng}', "");
Expect(0, 69882, '\P{^Script_Extensions= 	sora_Sompeng}', "");
Error('\p{Scx=:=		Sora}');
Error('\P{Scx=:=		Sora}');
Expect(1, 69881, '\p{Scx=sora}', "");
Expect(0, 69881, '\p{^Scx=sora}', "");
Expect(0, 69881, '\P{Scx=sora}', "");
Expect(1, 69881, '\P{^Scx=sora}', "");
Expect(0, 69882, '\p{Scx=sora}', "");
Expect(1, 69882, '\p{^Scx=sora}', "");
Expect(1, 69882, '\P{Scx=sora}', "");
Expect(0, 69882, '\P{^Scx=sora}', "");
Expect(1, 69881, '\p{Scx=_ Sora}', "");
Expect(0, 69881, '\p{^Scx=_ Sora}', "");
Expect(0, 69881, '\P{Scx=_ Sora}', "");
Expect(1, 69881, '\P{^Scx=_ Sora}', "");
Expect(0, 69882, '\p{Scx=_ Sora}', "");
Expect(1, 69882, '\p{^Scx=_ Sora}', "");
Expect(1, 69882, '\P{Scx=_ Sora}', "");
Expect(0, 69882, '\P{^Scx=_ Sora}', "");
Error('\p{Is_Script_Extensions=:= -SORA_Sompeng}');
Error('\P{Is_Script_Extensions=:= -SORA_Sompeng}');
Expect(1, 69881, '\p{Is_Script_Extensions=sorasompeng}', "");
Expect(0, 69881, '\p{^Is_Script_Extensions=sorasompeng}', "");
Expect(0, 69881, '\P{Is_Script_Extensions=sorasompeng}', "");
Expect(1, 69881, '\P{^Is_Script_Extensions=sorasompeng}', "");
Expect(0, 69882, '\p{Is_Script_Extensions=sorasompeng}', "");
Expect(1, 69882, '\p{^Is_Script_Extensions=sorasompeng}', "");
Expect(1, 69882, '\P{Is_Script_Extensions=sorasompeng}', "");
Expect(0, 69882, '\P{^Is_Script_Extensions=sorasompeng}', "");
Expect(1, 69881, '\p{Is_Script_Extensions=		sora_SOMPENG}', "");
Expect(0, 69881, '\p{^Is_Script_Extensions=		sora_SOMPENG}', "");
Expect(0, 69881, '\P{Is_Script_Extensions=		sora_SOMPENG}', "");
Expect(1, 69881, '\P{^Is_Script_Extensions=		sora_SOMPENG}', "");
Expect(0, 69882, '\p{Is_Script_Extensions=		sora_SOMPENG}', "");
Expect(1, 69882, '\p{^Is_Script_Extensions=		sora_SOMPENG}', "");
Expect(1, 69882, '\P{Is_Script_Extensions=		sora_SOMPENG}', "");
Expect(0, 69882, '\P{^Is_Script_Extensions=		sora_SOMPENG}', "");
Error('\p{Is_Scx=_:=Sora}');
Error('\P{Is_Scx=_:=Sora}');
Expect(1, 69881, '\p{Is_Scx=sora}', "");
Expect(0, 69881, '\p{^Is_Scx=sora}', "");
Expect(0, 69881, '\P{Is_Scx=sora}', "");
Expect(1, 69881, '\P{^Is_Scx=sora}', "");
Expect(0, 69882, '\p{Is_Scx=sora}', "");
Expect(1, 69882, '\p{^Is_Scx=sora}', "");
Expect(1, 69882, '\P{Is_Scx=sora}', "");
Expect(0, 69882, '\P{^Is_Scx=sora}', "");
Expect(1, 69881, '\p{Is_Scx=_	SORA}', "");
Expect(0, 69881, '\p{^Is_Scx=_	SORA}', "");
Expect(0, 69881, '\P{Is_Scx=_	SORA}', "");
Expect(1, 69881, '\P{^Is_Scx=_	SORA}', "");
Expect(0, 69882, '\p{Is_Scx=_	SORA}', "");
Expect(1, 69882, '\p{^Is_Scx=_	SORA}', "");
Expect(1, 69882, '\P{Is_Scx=_	SORA}', "");
Expect(0, 69882, '\P{^Is_Scx=_	SORA}', "");
Error('\p{Script_Extensions=	_Inscriptional_PAHLAVI:=}');
Error('\P{Script_Extensions=	_Inscriptional_PAHLAVI:=}');
Expect(1, 68479, '\p{Script_Extensions:   inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Script_Extensions:   inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Script_Extensions:   inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Script_Extensions:   inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Script_Extensions:   inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Script_Extensions:   inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Script_Extensions:   inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Script_Extensions:   inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Script_Extensions=--inscriptional_Pahlavi}', "");
Expect(0, 68479, '\p{^Script_Extensions=--inscriptional_Pahlavi}', "");
Expect(0, 68479, '\P{Script_Extensions=--inscriptional_Pahlavi}', "");
Expect(1, 68479, '\P{^Script_Extensions=--inscriptional_Pahlavi}', "");
Expect(0, 68480, '\p{Script_Extensions=--inscriptional_Pahlavi}', "");
Expect(1, 68480, '\p{^Script_Extensions=--inscriptional_Pahlavi}', "");
Expect(1, 68480, '\P{Script_Extensions=--inscriptional_Pahlavi}', "");
Expect(0, 68480, '\P{^Script_Extensions=--inscriptional_Pahlavi}', "");
Error('\p{Scx:-	PHLI:=}');
Error('\P{Scx:-	PHLI:=}');
Expect(1, 68479, '\p{Scx=phli}', "");
Expect(0, 68479, '\p{^Scx=phli}', "");
Expect(0, 68479, '\P{Scx=phli}', "");
Expect(1, 68479, '\P{^Scx=phli}', "");
Expect(0, 68480, '\p{Scx=phli}', "");
Expect(1, 68480, '\p{^Scx=phli}', "");
Expect(1, 68480, '\P{Scx=phli}', "");
Expect(0, 68480, '\P{^Scx=phli}', "");
Expect(1, 68479, '\p{Scx= Phli}', "");
Expect(0, 68479, '\p{^Scx= Phli}', "");
Expect(0, 68479, '\P{Scx= Phli}', "");
Expect(1, 68479, '\P{^Scx= Phli}', "");
Expect(0, 68480, '\p{Scx= Phli}', "");
Expect(1, 68480, '\p{^Scx= Phli}', "");
Expect(1, 68480, '\P{Scx= Phli}', "");
Expect(0, 68480, '\P{^Scx= Phli}', "");
Error('\p{Is_Script_Extensions=/a/ 	inscriptional_PAHLAVI}');
Error('\P{Is_Script_Extensions=/a/ 	inscriptional_PAHLAVI}');
Expect(1, 68479, '\p{Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Is_Script_Extensions=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Is_Script_Extensions= 	INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68479, '\p{^Is_Script_Extensions= 	INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68479, '\P{Is_Script_Extensions= 	INSCRIPTIONAL_Pahlavi}', "");
Expect(1, 68479, '\P{^Is_Script_Extensions= 	INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68480, '\p{Is_Script_Extensions= 	INSCRIPTIONAL_Pahlavi}', "");
Expect(1, 68480, '\p{^Is_Script_Extensions= 	INSCRIPTIONAL_Pahlavi}', "");
Expect(1, 68480, '\P{Is_Script_Extensions= 	INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68480, '\P{^Is_Script_Extensions= 	INSCRIPTIONAL_Pahlavi}', "");
Error('\p{Is_Scx=	/a/phli}');
Error('\P{Is_Scx=	/a/phli}');
Expect(1, 68479, '\p{Is_Scx=phli}', "");
Expect(0, 68479, '\p{^Is_Scx=phli}', "");
Expect(0, 68479, '\P{Is_Scx=phli}', "");
Expect(1, 68479, '\P{^Is_Scx=phli}', "");
Expect(0, 68480, '\p{Is_Scx=phli}', "");
Expect(1, 68480, '\p{^Is_Scx=phli}', "");
Expect(1, 68480, '\P{Is_Scx=phli}', "");
Expect(0, 68480, '\P{^Is_Scx=phli}', "");
Expect(1, 68479, '\p{Is_Scx=		Phli}', "");
Expect(0, 68479, '\p{^Is_Scx=		Phli}', "");
Expect(0, 68479, '\P{Is_Scx=		Phli}', "");
Expect(1, 68479, '\P{^Is_Scx=		Phli}', "");
Expect(0, 68480, '\p{Is_Scx=		Phli}', "");
Expect(1, 68480, '\p{^Is_Scx=		Phli}', "");
Expect(1, 68480, '\P{Is_Scx=		Phli}', "");
Expect(0, 68480, '\P{^Is_Scx=		Phli}', "");
Error('\p{Script_Extensions=/a/ tai_Tham}');
Error('\P{Script_Extensions=/a/ tai_Tham}');
Expect(1, 6829, '\p{Script_Extensions=taitham}', "");
Expect(0, 6829, '\p{^Script_Extensions=taitham}', "");
Expect(0, 6829, '\P{Script_Extensions=taitham}', "");
Expect(1, 6829, '\P{^Script_Extensions=taitham}', "");
Expect(0, 6830, '\p{Script_Extensions=taitham}', "");
Expect(1, 6830, '\p{^Script_Extensions=taitham}', "");
Expect(1, 6830, '\P{Script_Extensions=taitham}', "");
Expect(0, 6830, '\P{^Script_Extensions=taitham}', "");
Expect(1, 6829, '\p{Script_Extensions=_ TAI_tham}', "");
Expect(0, 6829, '\p{^Script_Extensions=_ TAI_tham}', "");
Expect(0, 6829, '\P{Script_Extensions=_ TAI_tham}', "");
Expect(1, 6829, '\P{^Script_Extensions=_ TAI_tham}', "");
Expect(0, 6830, '\p{Script_Extensions=_ TAI_tham}', "");
Expect(1, 6830, '\p{^Script_Extensions=_ TAI_tham}', "");
Expect(1, 6830, '\P{Script_Extensions=_ TAI_tham}', "");
Expect(0, 6830, '\P{^Script_Extensions=_ TAI_tham}', "");
Error('\p{Scx=_/a/LANA}');
Error('\P{Scx=_/a/LANA}');
Expect(1, 6829, '\p{Scx: lana}', "");
Expect(0, 6829, '\p{^Scx: lana}', "");
Expect(0, 6829, '\P{Scx: lana}', "");
Expect(1, 6829, '\P{^Scx: lana}', "");
Expect(0, 6830, '\p{Scx: lana}', "");
Expect(1, 6830, '\p{^Scx: lana}', "");
Expect(1, 6830, '\P{Scx: lana}', "");
Expect(0, 6830, '\P{^Scx: lana}', "");
Expect(1, 6829, '\p{Scx=_lana}', "");
Expect(0, 6829, '\p{^Scx=_lana}', "");
Expect(0, 6829, '\P{Scx=_lana}', "");
Expect(1, 6829, '\P{^Scx=_lana}', "");
Expect(0, 6830, '\p{Scx=_lana}', "");
Expect(1, 6830, '\p{^Scx=_lana}', "");
Expect(1, 6830, '\P{Scx=_lana}', "");
Expect(0, 6830, '\P{^Scx=_lana}', "");
Error('\p{Is_Script_Extensions=_/a/Tai_THAM}');
Error('\P{Is_Script_Extensions=_/a/Tai_THAM}');
Expect(1, 6829, '\p{Is_Script_Extensions=taitham}', "");
Expect(0, 6829, '\p{^Is_Script_Extensions=taitham}', "");
Expect(0, 6829, '\P{Is_Script_Extensions=taitham}', "");
Expect(1, 6829, '\P{^Is_Script_Extensions=taitham}', "");
Expect(0, 6830, '\p{Is_Script_Extensions=taitham}', "");
Expect(1, 6830, '\p{^Is_Script_Extensions=taitham}', "");
Expect(1, 6830, '\P{Is_Script_Extensions=taitham}', "");
Expect(0, 6830, '\P{^Is_Script_Extensions=taitham}', "");
Expect(1, 6829, '\p{Is_Script_Extensions=-tai_tham}', "");
Expect(0, 6829, '\p{^Is_Script_Extensions=-tai_tham}', "");
Expect(0, 6829, '\P{Is_Script_Extensions=-tai_tham}', "");
Expect(1, 6829, '\P{^Is_Script_Extensions=-tai_tham}', "");
Expect(0, 6830, '\p{Is_Script_Extensions=-tai_tham}', "");
Expect(1, 6830, '\p{^Is_Script_Extensions=-tai_tham}', "");
Expect(1, 6830, '\P{Is_Script_Extensions=-tai_tham}', "");
Expect(0, 6830, '\P{^Is_Script_Extensions=-tai_tham}', "");
Error('\p{Is_Scx=:=	Lana}');
Error('\P{Is_Scx=:=	Lana}');
Expect(1, 6829, '\p{Is_Scx=lana}', "");
Expect(0, 6829, '\p{^Is_Scx=lana}', "");
Expect(0, 6829, '\P{Is_Scx=lana}', "");
Expect(1, 6829, '\P{^Is_Scx=lana}', "");
Expect(0, 6830, '\p{Is_Scx=lana}', "");
Expect(1, 6830, '\p{^Is_Scx=lana}', "");
Expect(1, 6830, '\P{Is_Scx=lana}', "");
Expect(0, 6830, '\P{^Is_Scx=lana}', "");
Expect(1, 6829, '\p{Is_Scx=	_Lana}', "");
Expect(0, 6829, '\p{^Is_Scx=	_Lana}', "");
Expect(0, 6829, '\P{Is_Scx=	_Lana}', "");
Expect(1, 6829, '\P{^Is_Scx=	_Lana}', "");
Expect(0, 6830, '\p{Is_Scx=	_Lana}', "");
Expect(1, 6830, '\p{^Is_Scx=	_Lana}', "");
Expect(1, 6830, '\P{Is_Scx=	_Lana}', "");
Expect(0, 6830, '\P{^Is_Scx=	_Lana}', "");
Error('\p{Script_Extensions:	/a/Mandaic}');
Error('\P{Script_Extensions:	/a/Mandaic}');
Expect(1, 2142, '\p{Script_Extensions=mandaic}', "");
Expect(0, 2142, '\p{^Script_Extensions=mandaic}', "");
Expect(0, 2142, '\P{Script_Extensions=mandaic}', "");
Expect(1, 2142, '\P{^Script_Extensions=mandaic}', "");
Expect(0, 2143, '\p{Script_Extensions=mandaic}', "");
Expect(1, 2143, '\p{^Script_Extensions=mandaic}', "");
Expect(1, 2143, '\P{Script_Extensions=mandaic}', "");
Expect(0, 2143, '\P{^Script_Extensions=mandaic}', "");
Expect(1, 2142, '\p{Script_Extensions=	-Mandaic}', "");
Expect(0, 2142, '\p{^Script_Extensions=	-Mandaic}', "");
Expect(0, 2142, '\P{Script_Extensions=	-Mandaic}', "");
Expect(1, 2142, '\P{^Script_Extensions=	-Mandaic}', "");
Expect(0, 2143, '\p{Script_Extensions=	-Mandaic}', "");
Expect(1, 2143, '\p{^Script_Extensions=	-Mandaic}', "");
Expect(1, 2143, '\P{Script_Extensions=	-Mandaic}', "");
Expect(0, 2143, '\P{^Script_Extensions=	-Mandaic}', "");
Error('\p{Scx=  Mand:=}');
Error('\P{Scx=  Mand:=}');
Expect(1, 2142, '\p{Scx=mand}', "");
Expect(0, 2142, '\p{^Scx=mand}', "");
Expect(0, 2142, '\P{Scx=mand}', "");
Expect(1, 2142, '\P{^Scx=mand}', "");
Expect(0, 2143, '\p{Scx=mand}', "");
Expect(1, 2143, '\p{^Scx=mand}', "");
Expect(1, 2143, '\P{Scx=mand}', "");
Expect(0, 2143, '\P{^Scx=mand}', "");
Expect(1, 2142, '\p{Scx=  mand}', "");
Expect(0, 2142, '\p{^Scx=  mand}', "");
Expect(0, 2142, '\P{Scx=  mand}', "");
Expect(1, 2142, '\P{^Scx=  mand}', "");
Expect(0, 2143, '\p{Scx=  mand}', "");
Expect(1, 2143, '\p{^Scx=  mand}', "");
Expect(1, 2143, '\P{Scx=  mand}', "");
Expect(0, 2143, '\P{^Scx=  mand}', "");
Error('\p{Is_Script_Extensions=:= -Mandaic}');
Error('\P{Is_Script_Extensions=:= -Mandaic}');
Expect(1, 2142, '\p{Is_Script_Extensions=mandaic}', "");
Expect(0, 2142, '\p{^Is_Script_Extensions=mandaic}', "");
Expect(0, 2142, '\P{Is_Script_Extensions=mandaic}', "");
Expect(1, 2142, '\P{^Is_Script_Extensions=mandaic}', "");
Expect(0, 2143, '\p{Is_Script_Extensions=mandaic}', "");
Expect(1, 2143, '\p{^Is_Script_Extensions=mandaic}', "");
Expect(1, 2143, '\P{Is_Script_Extensions=mandaic}', "");
Expect(0, 2143, '\P{^Is_Script_Extensions=mandaic}', "");
Expect(1, 2142, '\p{Is_Script_Extensions=  Mandaic}', "");
Expect(0, 2142, '\p{^Is_Script_Extensions=  Mandaic}', "");
Expect(0, 2142, '\P{Is_Script_Extensions=  Mandaic}', "");
Expect(1, 2142, '\P{^Is_Script_Extensions=  Mandaic}', "");
Expect(0, 2143, '\p{Is_Script_Extensions=  Mandaic}', "");
Expect(1, 2143, '\p{^Is_Script_Extensions=  Mandaic}', "");
Expect(1, 2143, '\P{Is_Script_Extensions=  Mandaic}', "");
Expect(0, 2143, '\P{^Is_Script_Extensions=  Mandaic}', "");
Error('\p{Is_Scx=/a/Mand}');
Error('\P{Is_Scx=/a/Mand}');
Expect(1, 2142, '\p{Is_Scx=mand}', "");
Expect(0, 2142, '\p{^Is_Scx=mand}', "");
Expect(0, 2142, '\P{Is_Scx=mand}', "");
Expect(1, 2142, '\P{^Is_Scx=mand}', "");
Expect(0, 2143, '\p{Is_Scx=mand}', "");
Expect(1, 2143, '\p{^Is_Scx=mand}', "");
Expect(1, 2143, '\P{Is_Scx=mand}', "");
Expect(0, 2143, '\P{^Is_Scx=mand}', "");
Expect(1, 2142, '\p{Is_Scx=_	Mand}', "");
Expect(0, 2142, '\p{^Is_Scx=_	Mand}', "");
Expect(0, 2142, '\P{Is_Scx=_	Mand}', "");
Expect(1, 2142, '\P{^Is_Scx=_	Mand}', "");
Expect(0, 2143, '\p{Is_Scx=_	Mand}', "");
Expect(1, 2143, '\p{^Is_Scx=_	Mand}', "");
Expect(1, 2143, '\P{Is_Scx=_	Mand}', "");
Expect(0, 2143, '\P{^Is_Scx=_	Mand}', "");
Error('\p{Script_Extensions=_/a/Gurmukhi}');
Error('\P{Script_Extensions=_/a/Gurmukhi}');
Expect(1, 43065, '\p{Script_Extensions=gurmukhi}', "");
Expect(0, 43065, '\p{^Script_Extensions=gurmukhi}', "");
Expect(0, 43065, '\P{Script_Extensions=gurmukhi}', "");
Expect(1, 43065, '\P{^Script_Extensions=gurmukhi}', "");
Expect(0, 43066, '\p{Script_Extensions=gurmukhi}', "");
Expect(1, 43066, '\p{^Script_Extensions=gurmukhi}', "");
Expect(1, 43066, '\P{Script_Extensions=gurmukhi}', "");
Expect(0, 43066, '\P{^Script_Extensions=gurmukhi}', "");
Expect(1, 43065, '\p{Script_Extensions=-_GURMUKHI}', "");
Expect(0, 43065, '\p{^Script_Extensions=-_GURMUKHI}', "");
Expect(0, 43065, '\P{Script_Extensions=-_GURMUKHI}', "");
Expect(1, 43065, '\P{^Script_Extensions=-_GURMUKHI}', "");
Expect(0, 43066, '\p{Script_Extensions=-_GURMUKHI}', "");
Expect(1, 43066, '\p{^Script_Extensions=-_GURMUKHI}', "");
Expect(1, 43066, '\P{Script_Extensions=-_GURMUKHI}', "");
Expect(0, 43066, '\P{^Script_Extensions=-_GURMUKHI}', "");
Error('\p{Scx=		Guru/a/}');
Error('\P{Scx=		Guru/a/}');
Expect(1, 43065, '\p{Scx=guru}', "");
Expect(0, 43065, '\p{^Scx=guru}', "");
Expect(0, 43065, '\P{Scx=guru}', "");
Expect(1, 43065, '\P{^Scx=guru}', "");
Expect(0, 43066, '\p{Scx=guru}', "");
Expect(1, 43066, '\p{^Scx=guru}', "");
Expect(1, 43066, '\P{Scx=guru}', "");
Expect(0, 43066, '\P{^Scx=guru}', "");
Expect(1, 43065, '\p{Scx=--GURU}', "");
Expect(0, 43065, '\p{^Scx=--GURU}', "");
Expect(0, 43065, '\P{Scx=--GURU}', "");
Expect(1, 43065, '\P{^Scx=--GURU}', "");
Expect(0, 43066, '\p{Scx=--GURU}', "");
Expect(1, 43066, '\p{^Scx=--GURU}', "");
Expect(1, 43066, '\P{Scx=--GURU}', "");
Expect(0, 43066, '\P{^Scx=--GURU}', "");
Error('\p{Is_Script_Extensions=:=		Gurmukhi}');
Error('\P{Is_Script_Extensions=:=		Gurmukhi}');
Expect(1, 43065, '\p{Is_Script_Extensions=gurmukhi}', "");
Expect(0, 43065, '\p{^Is_Script_Extensions=gurmukhi}', "");
Expect(0, 43065, '\P{Is_Script_Extensions=gurmukhi}', "");
Expect(1, 43065, '\P{^Is_Script_Extensions=gurmukhi}', "");
Expect(0, 43066, '\p{Is_Script_Extensions=gurmukhi}', "");
Expect(1, 43066, '\p{^Is_Script_Extensions=gurmukhi}', "");
Expect(1, 43066, '\P{Is_Script_Extensions=gurmukhi}', "");
Expect(0, 43066, '\P{^Is_Script_Extensions=gurmukhi}', "");
Expect(1, 43065, '\p{Is_Script_Extensions=		GURMUKHI}', "");
Expect(0, 43065, '\p{^Is_Script_Extensions=		GURMUKHI}', "");
Expect(0, 43065, '\P{Is_Script_Extensions=		GURMUKHI}', "");
Expect(1, 43065, '\P{^Is_Script_Extensions=		GURMUKHI}', "");
Expect(0, 43066, '\p{Is_Script_Extensions=		GURMUKHI}', "");
Expect(1, 43066, '\p{^Is_Script_Extensions=		GURMUKHI}', "");
Expect(1, 43066, '\P{Is_Script_Extensions=		GURMUKHI}', "");
Expect(0, 43066, '\P{^Is_Script_Extensions=		GURMUKHI}', "");
Error('\p{Is_Scx=:=-	GURU}');
Error('\P{Is_Scx=:=-	GURU}');
Expect(1, 43065, '\p{Is_Scx=guru}', "");
Expect(0, 43065, '\p{^Is_Scx=guru}', "");
Expect(0, 43065, '\P{Is_Scx=guru}', "");
Expect(1, 43065, '\P{^Is_Scx=guru}', "");
Expect(0, 43066, '\p{Is_Scx=guru}', "");
Expect(1, 43066, '\p{^Is_Scx=guru}', "");
Expect(1, 43066, '\P{Is_Scx=guru}', "");
Expect(0, 43066, '\P{^Is_Scx=guru}', "");
Expect(1, 43065, '\p{Is_Scx=-Guru}', "");
Expect(0, 43065, '\p{^Is_Scx=-Guru}', "");
Expect(0, 43065, '\P{Is_Scx=-Guru}', "");
Expect(1, 43065, '\P{^Is_Scx=-Guru}', "");
Expect(0, 43066, '\p{Is_Scx=-Guru}', "");
Expect(1, 43066, '\p{^Is_Scx=-Guru}', "");
Expect(1, 43066, '\P{Is_Scx=-Guru}', "");
Expect(0, 43066, '\P{^Is_Scx=-Guru}', "");
Error('\p{Script_Extensions=/a/Batak}');
Error('\P{Script_Extensions=/a/Batak}');
Expect(1, 7167, '\p{Script_Extensions=batak}', "");
Expect(0, 7167, '\p{^Script_Extensions=batak}', "");
Expect(0, 7167, '\P{Script_Extensions=batak}', "");
Expect(1, 7167, '\P{^Script_Extensions=batak}', "");
Expect(0, 7168, '\p{Script_Extensions=batak}', "");
Expect(1, 7168, '\p{^Script_Extensions=batak}', "");
Expect(1, 7168, '\P{Script_Extensions=batak}', "");
Expect(0, 7168, '\P{^Script_Extensions=batak}', "");
Expect(1, 7167, '\p{Script_Extensions=_-Batak}', "");
Expect(0, 7167, '\p{^Script_Extensions=_-Batak}', "");
Expect(0, 7167, '\P{Script_Extensions=_-Batak}', "");
Expect(1, 7167, '\P{^Script_Extensions=_-Batak}', "");
Expect(0, 7168, '\p{Script_Extensions=_-Batak}', "");
Expect(1, 7168, '\p{^Script_Extensions=_-Batak}', "");
Expect(1, 7168, '\P{Script_Extensions=_-Batak}', "");
Expect(0, 7168, '\P{^Script_Extensions=_-Batak}', "");
Error('\p{Scx=/a/  Batk}');
Error('\P{Scx=/a/  Batk}');
Expect(1, 7167, '\p{Scx=batk}', "");
Expect(0, 7167, '\p{^Scx=batk}', "");
Expect(0, 7167, '\P{Scx=batk}', "");
Expect(1, 7167, '\P{^Scx=batk}', "");
Expect(0, 7168, '\p{Scx=batk}', "");
Expect(1, 7168, '\p{^Scx=batk}', "");
Expect(1, 7168, '\P{Scx=batk}', "");
Expect(0, 7168, '\P{^Scx=batk}', "");
Expect(1, 7167, '\p{Scx=- batk}', "");
Expect(0, 7167, '\p{^Scx=- batk}', "");
Expect(0, 7167, '\P{Scx=- batk}', "");
Expect(1, 7167, '\P{^Scx=- batk}', "");
Expect(0, 7168, '\p{Scx=- batk}', "");
Expect(1, 7168, '\p{^Scx=- batk}', "");
Expect(1, 7168, '\P{Scx=- batk}', "");
Expect(0, 7168, '\P{^Scx=- batk}', "");
Error('\p{Is_Script_Extensions=/a/_ batak}');
Error('\P{Is_Script_Extensions=/a/_ batak}');
Expect(1, 7167, '\p{Is_Script_Extensions=batak}', "");
Expect(0, 7167, '\p{^Is_Script_Extensions=batak}', "");
Expect(0, 7167, '\P{Is_Script_Extensions=batak}', "");
Expect(1, 7167, '\P{^Is_Script_Extensions=batak}', "");
Expect(0, 7168, '\p{Is_Script_Extensions=batak}', "");
Expect(1, 7168, '\p{^Is_Script_Extensions=batak}', "");
Expect(1, 7168, '\P{Is_Script_Extensions=batak}', "");
Expect(0, 7168, '\P{^Is_Script_Extensions=batak}', "");
Expect(1, 7167, '\p{Is_Script_Extensions=-_batak}', "");
Expect(0, 7167, '\p{^Is_Script_Extensions=-_batak}', "");
Expect(0, 7167, '\P{Is_Script_Extensions=-_batak}', "");
Expect(1, 7167, '\P{^Is_Script_Extensions=-_batak}', "");
Expect(0, 7168, '\p{Is_Script_Extensions=-_batak}', "");
Expect(1, 7168, '\p{^Is_Script_Extensions=-_batak}', "");
Expect(1, 7168, '\P{Is_Script_Extensions=-_batak}', "");
Expect(0, 7168, '\P{^Is_Script_Extensions=-_batak}', "");
Error('\p{Is_Scx:	-BATK/a/}');
Error('\P{Is_Scx:	-BATK/a/}');
Expect(1, 7167, '\p{Is_Scx=batk}', "");
Expect(0, 7167, '\p{^Is_Scx=batk}', "");
Expect(0, 7167, '\P{Is_Scx=batk}', "");
Expect(1, 7167, '\P{^Is_Scx=batk}', "");
Expect(0, 7168, '\p{Is_Scx=batk}', "");
Expect(1, 7168, '\p{^Is_Scx=batk}', "");
Expect(1, 7168, '\P{Is_Scx=batk}', "");
Expect(0, 7168, '\P{^Is_Scx=batk}', "");
Expect(1, 7167, '\p{Is_Scx=	-BATK}', "");
Expect(0, 7167, '\p{^Is_Scx=	-BATK}', "");
Expect(0, 7167, '\P{Is_Scx=	-BATK}', "");
Expect(1, 7167, '\P{^Is_Scx=	-BATK}', "");
Expect(0, 7168, '\p{Is_Scx=	-BATK}', "");
Expect(1, 7168, '\p{^Is_Scx=	-BATK}', "");
Expect(1, 7168, '\P{Is_Scx=	-BATK}', "");
Expect(0, 7168, '\P{^Is_Scx=	-BATK}', "");
Error('\p{Script_Extensions:/a/cypriot}');
Error('\P{Script_Extensions:/a/cypriot}');
Expect(1, 67647, '\p{Script_Extensions:cypriot}', "");
Expect(0, 67647, '\p{^Script_Extensions:cypriot}', "");
Expect(0, 67647, '\P{Script_Extensions:cypriot}', "");
Expect(1, 67647, '\P{^Script_Extensions:cypriot}', "");
Expect(0, 67648, '\p{Script_Extensions:cypriot}', "");
Expect(1, 67648, '\p{^Script_Extensions:cypriot}', "");
Expect(1, 67648, '\P{Script_Extensions:cypriot}', "");
Expect(0, 67648, '\P{^Script_Extensions:cypriot}', "");
Expect(1, 67647, '\p{Script_Extensions=_-cypriot}', "");
Expect(0, 67647, '\p{^Script_Extensions=_-cypriot}', "");
Expect(0, 67647, '\P{Script_Extensions=_-cypriot}', "");
Expect(1, 67647, '\P{^Script_Extensions=_-cypriot}', "");
Expect(0, 67648, '\p{Script_Extensions=_-cypriot}', "");
Expect(1, 67648, '\p{^Script_Extensions=_-cypriot}', "");
Expect(1, 67648, '\P{Script_Extensions=_-cypriot}', "");
Expect(0, 67648, '\P{^Script_Extensions=_-cypriot}', "");
Error('\p{Scx=:= Cprt}');
Error('\P{Scx=:= Cprt}');
Expect(1, 67647, '\p{Scx=cprt}', "");
Expect(0, 67647, '\p{^Scx=cprt}', "");
Expect(0, 67647, '\P{Scx=cprt}', "");
Expect(1, 67647, '\P{^Scx=cprt}', "");
Expect(0, 67648, '\p{Scx=cprt}', "");
Expect(1, 67648, '\p{^Scx=cprt}', "");
Expect(1, 67648, '\P{Scx=cprt}', "");
Expect(0, 67648, '\P{^Scx=cprt}', "");
Expect(1, 67647, '\p{Scx:-cprt}', "");
Expect(0, 67647, '\p{^Scx:-cprt}', "");
Expect(0, 67647, '\P{Scx:-cprt}', "");
Expect(1, 67647, '\P{^Scx:-cprt}', "");
Expect(0, 67648, '\p{Scx:-cprt}', "");
Expect(1, 67648, '\p{^Scx:-cprt}', "");
Expect(1, 67648, '\P{Scx:-cprt}', "");
Expect(0, 67648, '\P{^Scx:-cprt}', "");
Error('\p{Is_Script_Extensions=/a/ -Cypriot}');
Error('\P{Is_Script_Extensions=/a/ -Cypriot}');
Expect(1, 67647, '\p{Is_Script_Extensions=cypriot}', "");
Expect(0, 67647, '\p{^Is_Script_Extensions=cypriot}', "");
Expect(0, 67647, '\P{Is_Script_Extensions=cypriot}', "");
Expect(1, 67647, '\P{^Is_Script_Extensions=cypriot}', "");
Expect(0, 67648, '\p{Is_Script_Extensions=cypriot}', "");
Expect(1, 67648, '\p{^Is_Script_Extensions=cypriot}', "");
Expect(1, 67648, '\P{Is_Script_Extensions=cypriot}', "");
Expect(0, 67648, '\P{^Is_Script_Extensions=cypriot}', "");
Expect(1, 67647, '\p{Is_Script_Extensions=--Cypriot}', "");
Expect(0, 67647, '\p{^Is_Script_Extensions=--Cypriot}', "");
Expect(0, 67647, '\P{Is_Script_Extensions=--Cypriot}', "");
Expect(1, 67647, '\P{^Is_Script_Extensions=--Cypriot}', "");
Expect(0, 67648, '\p{Is_Script_Extensions=--Cypriot}', "");
Expect(1, 67648, '\p{^Is_Script_Extensions=--Cypriot}', "");
Expect(1, 67648, '\P{Is_Script_Extensions=--Cypriot}', "");
Expect(0, 67648, '\P{^Is_Script_Extensions=--Cypriot}', "");
Error('\p{Is_Scx=/a/--CPRT}');
Error('\P{Is_Scx=/a/--CPRT}');
Expect(1, 67647, '\p{Is_Scx=cprt}', "");
Expect(0, 67647, '\p{^Is_Scx=cprt}', "");
Expect(0, 67647, '\P{Is_Scx=cprt}', "");
Expect(1, 67647, '\P{^Is_Scx=cprt}', "");
Expect(0, 67648, '\p{Is_Scx=cprt}', "");
Expect(1, 67648, '\p{^Is_Scx=cprt}', "");
Expect(1, 67648, '\P{Is_Scx=cprt}', "");
Expect(0, 67648, '\P{^Is_Scx=cprt}', "");
Expect(1, 67647, '\p{Is_Scx=_	Cprt}', "");
Expect(0, 67647, '\p{^Is_Scx=_	Cprt}', "");
Expect(0, 67647, '\P{Is_Scx=_	Cprt}', "");
Expect(1, 67647, '\P{^Is_Scx=_	Cprt}', "");
Expect(0, 67648, '\p{Is_Scx=_	Cprt}', "");
Expect(1, 67648, '\p{^Is_Scx=_	Cprt}', "");
Expect(1, 67648, '\P{Is_Scx=_	Cprt}', "");
Expect(0, 67648, '\P{^Is_Scx=_	Cprt}', "");
Error('\p{Script_Extensions= HANGUL:=}');
Error('\P{Script_Extensions= HANGUL:=}');
Expect(1, 65500, '\p{Script_Extensions=hangul}', "");
Expect(0, 65500, '\p{^Script_Extensions=hangul}', "");
Expect(0, 65500, '\P{Script_Extensions=hangul}', "");
Expect(1, 65500, '\P{^Script_Extensions=hangul}', "");
Expect(0, 65501, '\p{Script_Extensions=hangul}', "");
Expect(1, 65501, '\p{^Script_Extensions=hangul}', "");
Expect(1, 65501, '\P{Script_Extensions=hangul}', "");
Expect(0, 65501, '\P{^Script_Extensions=hangul}', "");
Expect(1, 65500, '\p{Script_Extensions=- hangul}', "");
Expect(0, 65500, '\p{^Script_Extensions=- hangul}', "");
Expect(0, 65500, '\P{Script_Extensions=- hangul}', "");
Expect(1, 65500, '\P{^Script_Extensions=- hangul}', "");
Expect(0, 65501, '\p{Script_Extensions=- hangul}', "");
Expect(1, 65501, '\p{^Script_Extensions=- hangul}', "");
Expect(1, 65501, '\P{Script_Extensions=- hangul}', "");
Expect(0, 65501, '\P{^Script_Extensions=- hangul}', "");
Error('\p{Scx=	/a/HANG}');
Error('\P{Scx=	/a/HANG}');
Expect(1, 65500, '\p{Scx=hang}', "");
Expect(0, 65500, '\p{^Scx=hang}', "");
Expect(0, 65500, '\P{Scx=hang}', "");
Expect(1, 65500, '\P{^Scx=hang}', "");
Expect(0, 65501, '\p{Scx=hang}', "");
Expect(1, 65501, '\p{^Scx=hang}', "");
Expect(1, 65501, '\P{Scx=hang}', "");
Expect(0, 65501, '\P{^Scx=hang}', "");
Expect(1, 65500, '\p{Scx:   - Hang}', "");
Expect(0, 65500, '\p{^Scx:   - Hang}', "");
Expect(0, 65500, '\P{Scx:   - Hang}', "");
Expect(1, 65500, '\P{^Scx:   - Hang}', "");
Expect(0, 65501, '\p{Scx:   - Hang}', "");
Expect(1, 65501, '\p{^Scx:   - Hang}', "");
Expect(1, 65501, '\P{Scx:   - Hang}', "");
Expect(0, 65501, '\P{^Scx:   - Hang}', "");
Error('\p{Is_Script_Extensions=:=	 Hangul}');
Error('\P{Is_Script_Extensions=:=	 Hangul}');
Expect(1, 65500, '\p{Is_Script_Extensions: hangul}', "");
Expect(0, 65500, '\p{^Is_Script_Extensions: hangul}', "");
Expect(0, 65500, '\P{Is_Script_Extensions: hangul}', "");
Expect(1, 65500, '\P{^Is_Script_Extensions: hangul}', "");
Expect(0, 65501, '\p{Is_Script_Extensions: hangul}', "");
Expect(1, 65501, '\p{^Is_Script_Extensions: hangul}', "");
Expect(1, 65501, '\P{Is_Script_Extensions: hangul}', "");
Expect(0, 65501, '\P{^Is_Script_Extensions: hangul}', "");
Expect(1, 65500, '\p{Is_Script_Extensions:		hangul}', "");
Expect(0, 65500, '\p{^Is_Script_Extensions:		hangul}', "");
Expect(0, 65500, '\P{Is_Script_Extensions:		hangul}', "");
Expect(1, 65500, '\P{^Is_Script_Extensions:		hangul}', "");
Expect(0, 65501, '\p{Is_Script_Extensions:		hangul}', "");
Expect(1, 65501, '\p{^Is_Script_Extensions:		hangul}', "");
Expect(1, 65501, '\P{Is_Script_Extensions:		hangul}', "");
Expect(0, 65501, '\P{^Is_Script_Extensions:		hangul}', "");
Error('\p{Is_Scx=__Hang/a/}');
Error('\P{Is_Scx=__Hang/a/}');
Expect(1, 65500, '\p{Is_Scx=hang}', "");
Expect(0, 65500, '\p{^Is_Scx=hang}', "");
Expect(0, 65500, '\P{Is_Scx=hang}', "");
Expect(1, 65500, '\P{^Is_Scx=hang}', "");
Expect(0, 65501, '\p{Is_Scx=hang}', "");
Expect(1, 65501, '\p{^Is_Scx=hang}', "");
Expect(1, 65501, '\P{Is_Scx=hang}', "");
Expect(0, 65501, '\P{^Is_Scx=hang}', "");
Expect(1, 65500, '\p{Is_Scx= 	HANG}', "");
Expect(0, 65500, '\p{^Is_Scx= 	HANG}', "");
Expect(0, 65500, '\P{Is_Scx= 	HANG}', "");
Expect(1, 65500, '\P{^Is_Scx= 	HANG}', "");
Expect(0, 65501, '\p{Is_Scx= 	HANG}', "");
Expect(1, 65501, '\p{^Is_Scx= 	HANG}', "");
Expect(1, 65501, '\P{Is_Scx= 	HANG}', "");
Expect(0, 65501, '\P{^Is_Scx= 	HANG}', "");
Error('\p{Script_Extensions=-/a/Deseret}');
Error('\P{Script_Extensions=-/a/Deseret}');
Expect(1, 66639, '\p{Script_Extensions=deseret}', "");
Expect(0, 66639, '\p{^Script_Extensions=deseret}', "");
Expect(0, 66639, '\P{Script_Extensions=deseret}', "");
Expect(1, 66639, '\P{^Script_Extensions=deseret}', "");
Expect(0, 66640, '\p{Script_Extensions=deseret}', "");
Expect(1, 66640, '\p{^Script_Extensions=deseret}', "");
Expect(1, 66640, '\P{Script_Extensions=deseret}', "");
Expect(0, 66640, '\P{^Script_Extensions=deseret}', "");
Expect(1, 66639, '\p{Script_Extensions= DESERET}', "");
Expect(0, 66639, '\p{^Script_Extensions= DESERET}', "");
Expect(0, 66639, '\P{Script_Extensions= DESERET}', "");
Expect(1, 66639, '\P{^Script_Extensions= DESERET}', "");
Expect(0, 66640, '\p{Script_Extensions= DESERET}', "");
Expect(1, 66640, '\p{^Script_Extensions= DESERET}', "");
Expect(1, 66640, '\P{Script_Extensions= DESERET}', "");
Expect(0, 66640, '\P{^Script_Extensions= DESERET}', "");
Error('\p{Scx= 	Dsrt/a/}');
Error('\P{Scx= 	Dsrt/a/}');
Expect(1, 66639, '\p{Scx=dsrt}', "");
Expect(0, 66639, '\p{^Scx=dsrt}', "");
Expect(0, 66639, '\P{Scx=dsrt}', "");
Expect(1, 66639, '\P{^Scx=dsrt}', "");
Expect(0, 66640, '\p{Scx=dsrt}', "");
Expect(1, 66640, '\p{^Scx=dsrt}', "");
Expect(1, 66640, '\P{Scx=dsrt}', "");
Expect(0, 66640, '\P{^Scx=dsrt}', "");
Expect(1, 66639, '\p{Scx=_DSRT}', "");
Expect(0, 66639, '\p{^Scx=_DSRT}', "");
Expect(0, 66639, '\P{Scx=_DSRT}', "");
Expect(1, 66639, '\P{^Scx=_DSRT}', "");
Expect(0, 66640, '\p{Scx=_DSRT}', "");
Expect(1, 66640, '\p{^Scx=_DSRT}', "");
Expect(1, 66640, '\P{Scx=_DSRT}', "");
Expect(0, 66640, '\P{^Scx=_DSRT}', "");
Error('\p{Is_Script_Extensions=-/a/Deseret}');
Error('\P{Is_Script_Extensions=-/a/Deseret}');
Expect(1, 66639, '\p{Is_Script_Extensions:deseret}', "");
Expect(0, 66639, '\p{^Is_Script_Extensions:deseret}', "");
Expect(0, 66639, '\P{Is_Script_Extensions:deseret}', "");
Expect(1, 66639, '\P{^Is_Script_Extensions:deseret}', "");
Expect(0, 66640, '\p{Is_Script_Extensions:deseret}', "");
Expect(1, 66640, '\p{^Is_Script_Extensions:deseret}', "");
Expect(1, 66640, '\P{Is_Script_Extensions:deseret}', "");
Expect(0, 66640, '\P{^Is_Script_Extensions:deseret}', "");
Expect(1, 66639, '\p{Is_Script_Extensions=	_deseret}', "");
Expect(0, 66639, '\p{^Is_Script_Extensions=	_deseret}', "");
Expect(0, 66639, '\P{Is_Script_Extensions=	_deseret}', "");
Expect(1, 66639, '\P{^Is_Script_Extensions=	_deseret}', "");
Expect(0, 66640, '\p{Is_Script_Extensions=	_deseret}', "");
Expect(1, 66640, '\p{^Is_Script_Extensions=	_deseret}', "");
Expect(1, 66640, '\P{Is_Script_Extensions=	_deseret}', "");
Expect(0, 66640, '\P{^Is_Script_Extensions=	_deseret}', "");
Error('\p{Is_Scx=_DSRT:=}');
Error('\P{Is_Scx=_DSRT:=}');
Expect(1, 66639, '\p{Is_Scx=dsrt}', "");
Expect(0, 66639, '\p{^Is_Scx=dsrt}', "");
Expect(0, 66639, '\P{Is_Scx=dsrt}', "");
Expect(1, 66639, '\P{^Is_Scx=dsrt}', "");
Expect(0, 66640, '\p{Is_Scx=dsrt}', "");
Expect(1, 66640, '\p{^Is_Scx=dsrt}', "");
Expect(1, 66640, '\P{Is_Scx=dsrt}', "");
Expect(0, 66640, '\P{^Is_Scx=dsrt}', "");
Expect(1, 66639, '\p{Is_Scx=-_Dsrt}', "");
Expect(0, 66639, '\p{^Is_Scx=-_Dsrt}', "");
Expect(0, 66639, '\P{Is_Scx=-_Dsrt}', "");
Expect(1, 66639, '\P{^Is_Scx=-_Dsrt}', "");
Expect(0, 66640, '\p{Is_Scx=-_Dsrt}', "");
Expect(1, 66640, '\p{^Is_Scx=-_Dsrt}', "");
Expect(1, 66640, '\P{Is_Scx=-_Dsrt}', "");
Expect(0, 66640, '\P{^Is_Scx=-_Dsrt}', "");
Error('\p{Script_Extensions=	UNKNOWN/a/}');
Error('\P{Script_Extensions=	UNKNOWN/a/}');
Expect(1, 918000, '\p{Script_Extensions=unknown}', "");
Expect(0, 918000, '\p{^Script_Extensions=unknown}', "");
Expect(0, 918000, '\P{Script_Extensions=unknown}', "");
Expect(1, 918000, '\P{^Script_Extensions=unknown}', "");
Expect(0, 917999, '\p{Script_Extensions=unknown}', "");
Expect(1, 917999, '\p{^Script_Extensions=unknown}', "");
Expect(1, 917999, '\P{Script_Extensions=unknown}', "");
Expect(0, 917999, '\P{^Script_Extensions=unknown}', "");
Expect(1, 918000, '\p{Script_Extensions=-	Unknown}', "");
Expect(0, 918000, '\p{^Script_Extensions=-	Unknown}', "");
Expect(0, 918000, '\P{Script_Extensions=-	Unknown}', "");
Expect(1, 918000, '\P{^Script_Extensions=-	Unknown}', "");
Expect(0, 917999, '\p{Script_Extensions=-	Unknown}', "");
Expect(1, 917999, '\p{^Script_Extensions=-	Unknown}', "");
Expect(1, 917999, '\P{Script_Extensions=-	Unknown}', "");
Expect(0, 917999, '\P{^Script_Extensions=-	Unknown}', "");
Error('\p{Scx=:=--ZZZZ}');
Error('\P{Scx=:=--ZZZZ}');
Expect(1, 918000, '\p{Scx=zzzz}', "");
Expect(0, 918000, '\p{^Scx=zzzz}', "");
Expect(0, 918000, '\P{Scx=zzzz}', "");
Expect(1, 918000, '\P{^Scx=zzzz}', "");
Expect(0, 917999, '\p{Scx=zzzz}', "");
Expect(1, 917999, '\p{^Scx=zzzz}', "");
Expect(1, 917999, '\P{Scx=zzzz}', "");
Expect(0, 917999, '\P{^Scx=zzzz}', "");
Expect(1, 918000, '\p{Scx=_	Zzzz}', "");
Expect(0, 918000, '\p{^Scx=_	Zzzz}', "");
Expect(0, 918000, '\P{Scx=_	Zzzz}', "");
Expect(1, 918000, '\P{^Scx=_	Zzzz}', "");
Expect(0, 917999, '\p{Scx=_	Zzzz}', "");
Expect(1, 917999, '\p{^Scx=_	Zzzz}', "");
Expect(1, 917999, '\P{Scx=_	Zzzz}', "");
Expect(0, 917999, '\P{^Scx=_	Zzzz}', "");
Error('\p{Is_Script_Extensions=:=	unknown}');
Error('\P{Is_Script_Extensions=:=	unknown}');
Expect(1, 918000, '\p{Is_Script_Extensions=unknown}', "");
Expect(0, 918000, '\p{^Is_Script_Extensions=unknown}', "");
Expect(0, 918000, '\P{Is_Script_Extensions=unknown}', "");
Expect(1, 918000, '\P{^Is_Script_Extensions=unknown}', "");
Expect(0, 917999, '\p{Is_Script_Extensions=unknown}', "");
Expect(1, 917999, '\p{^Is_Script_Extensions=unknown}', "");
Expect(1, 917999, '\P{Is_Script_Extensions=unknown}', "");
Expect(0, 917999, '\P{^Is_Script_Extensions=unknown}', "");
Expect(1, 918000, '\p{Is_Script_Extensions=_	Unknown}', "");
Expect(0, 918000, '\p{^Is_Script_Extensions=_	Unknown}', "");
Expect(0, 918000, '\P{Is_Script_Extensions=_	Unknown}', "");
Expect(1, 918000, '\P{^Is_Script_Extensions=_	Unknown}', "");
Expect(0, 917999, '\p{Is_Script_Extensions=_	Unknown}', "");
Expect(1, 917999, '\p{^Is_Script_Extensions=_	Unknown}', "");
Expect(1, 917999, '\P{Is_Script_Extensions=_	Unknown}', "");
Expect(0, 917999, '\P{^Is_Script_Extensions=_	Unknown}', "");
Error('\p{Is_Scx:   :=  Zzzz}');
Error('\P{Is_Scx:   :=  Zzzz}');
Expect(1, 918000, '\p{Is_Scx=zzzz}', "");
Expect(0, 918000, '\p{^Is_Scx=zzzz}', "");
Expect(0, 918000, '\P{Is_Scx=zzzz}', "");
Expect(1, 918000, '\P{^Is_Scx=zzzz}', "");
Expect(0, 917999, '\p{Is_Scx=zzzz}', "");
Expect(1, 917999, '\p{^Is_Scx=zzzz}', "");
Expect(1, 917999, '\P{Is_Scx=zzzz}', "");
Expect(0, 917999, '\P{^Is_Scx=zzzz}', "");
Expect(1, 918000, '\p{Is_Scx=	zzzz}', "");
Expect(0, 918000, '\p{^Is_Scx=	zzzz}', "");
Expect(0, 918000, '\P{Is_Scx=	zzzz}', "");
Expect(1, 918000, '\P{^Is_Scx=	zzzz}', "");
Expect(0, 917999, '\p{Is_Scx=	zzzz}', "");
Expect(1, 917999, '\p{^Is_Scx=	zzzz}', "");
Expect(1, 917999, '\P{Is_Scx=	zzzz}', "");
Expect(0, 917999, '\P{^Is_Scx=	zzzz}', "");
Error('\p{Script_Extensions:   := cuneiform}');
Error('\P{Script_Extensions:   := cuneiform}');
Expect(1, 74867, '\p{Script_Extensions=cuneiform}', "");
Expect(0, 74867, '\p{^Script_Extensions=cuneiform}', "");
Expect(0, 74867, '\P{Script_Extensions=cuneiform}', "");
Expect(1, 74867, '\P{^Script_Extensions=cuneiform}', "");
Expect(0, 74868, '\p{Script_Extensions=cuneiform}', "");
Expect(1, 74868, '\p{^Script_Extensions=cuneiform}', "");
Expect(1, 74868, '\P{Script_Extensions=cuneiform}', "");
Expect(0, 74868, '\P{^Script_Extensions=cuneiform}', "");
Expect(1, 74867, '\p{Script_Extensions= -Cuneiform}', "");
Expect(0, 74867, '\p{^Script_Extensions= -Cuneiform}', "");
Expect(0, 74867, '\P{Script_Extensions= -Cuneiform}', "");
Expect(1, 74867, '\P{^Script_Extensions= -Cuneiform}', "");
Expect(0, 74868, '\p{Script_Extensions= -Cuneiform}', "");
Expect(1, 74868, '\p{^Script_Extensions= -Cuneiform}', "");
Expect(1, 74868, '\P{Script_Extensions= -Cuneiform}', "");
Expect(0, 74868, '\P{^Script_Extensions= -Cuneiform}', "");
Error('\p{Scx=	/a/xsux}');
Error('\P{Scx=	/a/xsux}');
Expect(1, 74867, '\p{Scx=xsux}', "");
Expect(0, 74867, '\p{^Scx=xsux}', "");
Expect(0, 74867, '\P{Scx=xsux}', "");
Expect(1, 74867, '\P{^Scx=xsux}', "");
Expect(0, 74868, '\p{Scx=xsux}', "");
Expect(1, 74868, '\p{^Scx=xsux}', "");
Expect(1, 74868, '\P{Scx=xsux}', "");
Expect(0, 74868, '\P{^Scx=xsux}', "");
Expect(1, 74867, '\p{Scx=	 Xsux}', "");
Expect(0, 74867, '\p{^Scx=	 Xsux}', "");
Expect(0, 74867, '\P{Scx=	 Xsux}', "");
Expect(1, 74867, '\P{^Scx=	 Xsux}', "");
Expect(0, 74868, '\p{Scx=	 Xsux}', "");
Expect(1, 74868, '\p{^Scx=	 Xsux}', "");
Expect(1, 74868, '\P{Scx=	 Xsux}', "");
Expect(0, 74868, '\P{^Scx=	 Xsux}', "");
Error('\p{Is_Script_Extensions: _cuneiform:=}');
Error('\P{Is_Script_Extensions: _cuneiform:=}');
Expect(1, 74867, '\p{Is_Script_Extensions=cuneiform}', "");
Expect(0, 74867, '\p{^Is_Script_Extensions=cuneiform}', "");
Expect(0, 74867, '\P{Is_Script_Extensions=cuneiform}', "");
Expect(1, 74867, '\P{^Is_Script_Extensions=cuneiform}', "");
Expect(0, 74868, '\p{Is_Script_Extensions=cuneiform}', "");
Expect(1, 74868, '\p{^Is_Script_Extensions=cuneiform}', "");
Expect(1, 74868, '\P{Is_Script_Extensions=cuneiform}', "");
Expect(0, 74868, '\P{^Is_Script_Extensions=cuneiform}', "");
Expect(1, 74867, '\p{Is_Script_Extensions=	_CUNEIFORM}', "");
Expect(0, 74867, '\p{^Is_Script_Extensions=	_CUNEIFORM}', "");
Expect(0, 74867, '\P{Is_Script_Extensions=	_CUNEIFORM}', "");
Expect(1, 74867, '\P{^Is_Script_Extensions=	_CUNEIFORM}', "");
Expect(0, 74868, '\p{Is_Script_Extensions=	_CUNEIFORM}', "");
Expect(1, 74868, '\p{^Is_Script_Extensions=	_CUNEIFORM}', "");
Expect(1, 74868, '\P{Is_Script_Extensions=	_CUNEIFORM}', "");
Expect(0, 74868, '\P{^Is_Script_Extensions=	_CUNEIFORM}', "");
Error('\p{Is_Scx=Xsux:=}');
Error('\P{Is_Scx=Xsux:=}');
Expect(1, 74867, '\p{Is_Scx:xsux}', "");
Expect(0, 74867, '\p{^Is_Scx:xsux}', "");
Expect(0, 74867, '\P{Is_Scx:xsux}', "");
Expect(1, 74867, '\P{^Is_Scx:xsux}', "");
Expect(0, 74868, '\p{Is_Scx:xsux}', "");
Expect(1, 74868, '\p{^Is_Scx:xsux}', "");
Expect(1, 74868, '\P{Is_Scx:xsux}', "");
Expect(0, 74868, '\P{^Is_Scx:xsux}', "");
Expect(1, 74867, '\p{Is_Scx=-xsux}', "");
Expect(0, 74867, '\p{^Is_Scx=-xsux}', "");
Expect(0, 74867, '\P{Is_Scx=-xsux}', "");
Expect(1, 74867, '\P{^Is_Scx=-xsux}', "");
Expect(0, 74868, '\p{Is_Scx=-xsux}', "");
Expect(1, 74868, '\p{^Is_Scx=-xsux}', "");
Expect(1, 74868, '\P{Is_Scx=-xsux}', "");
Expect(0, 74868, '\P{^Is_Scx=-xsux}', "");
Error('\p{Script_Extensions=/a/ -inherited}');
Error('\P{Script_Extensions=/a/ -inherited}');
Expect(1, 917999, '\p{Script_Extensions=inherited}', "");
Expect(0, 917999, '\p{^Script_Extensions=inherited}', "");
Expect(0, 917999, '\P{Script_Extensions=inherited}', "");
Expect(1, 917999, '\P{^Script_Extensions=inherited}', "");
Expect(0, 918000, '\p{Script_Extensions=inherited}', "");
Expect(1, 918000, '\p{^Script_Extensions=inherited}', "");
Expect(1, 918000, '\P{Script_Extensions=inherited}', "");
Expect(0, 918000, '\P{^Script_Extensions=inherited}', "");
Expect(1, 917999, '\p{Script_Extensions=		INHERITED}', "");
Expect(0, 917999, '\p{^Script_Extensions=		INHERITED}', "");
Expect(0, 917999, '\P{Script_Extensions=		INHERITED}', "");
Expect(1, 917999, '\P{^Script_Extensions=		INHERITED}', "");
Expect(0, 918000, '\p{Script_Extensions=		INHERITED}', "");
Expect(1, 918000, '\p{^Script_Extensions=		INHERITED}', "");
Expect(1, 918000, '\P{Script_Extensions=		INHERITED}', "");
Expect(0, 918000, '\P{^Script_Extensions=		INHERITED}', "");
Error('\p{Scx=:=  ZINH}');
Error('\P{Scx=:=  ZINH}');
Expect(1, 917999, '\p{Scx: zinh}', "");
Expect(0, 917999, '\p{^Scx: zinh}', "");
Expect(0, 917999, '\P{Scx: zinh}', "");
Expect(1, 917999, '\P{^Scx: zinh}', "");
Expect(0, 918000, '\p{Scx: zinh}', "");
Expect(1, 918000, '\p{^Scx: zinh}', "");
Expect(1, 918000, '\P{Scx: zinh}', "");
Expect(0, 918000, '\P{^Scx: zinh}', "");
Expect(1, 917999, '\p{Scx=	_Zinh}', "");
Expect(0, 917999, '\p{^Scx=	_Zinh}', "");
Expect(0, 917999, '\P{Scx=	_Zinh}', "");
Expect(1, 917999, '\P{^Scx=	_Zinh}', "");
Expect(0, 918000, '\p{Scx=	_Zinh}', "");
Expect(1, 918000, '\p{^Scx=	_Zinh}', "");
Expect(1, 918000, '\P{Scx=	_Zinh}', "");
Expect(0, 918000, '\P{^Scx=	_Zinh}', "");
Error('\p{Is_Script_Extensions= :=QAAI}');
Error('\P{Is_Script_Extensions= :=QAAI}');
Expect(1, 917999, '\p{Is_Script_Extensions=qaai}', "");
Expect(0, 917999, '\p{^Is_Script_Extensions=qaai}', "");
Expect(0, 917999, '\P{Is_Script_Extensions=qaai}', "");
Expect(1, 917999, '\P{^Is_Script_Extensions=qaai}', "");
Expect(0, 918000, '\p{Is_Script_Extensions=qaai}', "");
Expect(1, 918000, '\p{^Is_Script_Extensions=qaai}', "");
Expect(1, 918000, '\P{Is_Script_Extensions=qaai}', "");
Expect(0, 918000, '\P{^Is_Script_Extensions=qaai}', "");
Expect(1, 917999, '\p{Is_Script_Extensions=_-QAAI}', "");
Expect(0, 917999, '\p{^Is_Script_Extensions=_-QAAI}', "");
Expect(0, 917999, '\P{Is_Script_Extensions=_-QAAI}', "");
Expect(1, 917999, '\P{^Is_Script_Extensions=_-QAAI}', "");
Expect(0, 918000, '\p{Is_Script_Extensions=_-QAAI}', "");
Expect(1, 918000, '\p{^Is_Script_Extensions=_-QAAI}', "");
Expect(1, 918000, '\P{Is_Script_Extensions=_-QAAI}', "");
Expect(0, 918000, '\P{^Is_Script_Extensions=_-QAAI}', "");
Error('\p{Is_Scx=:=	Inherited}');
Error('\P{Is_Scx=:=	Inherited}');
Expect(1, 917999, '\p{Is_Scx:   inherited}', "");
Expect(0, 917999, '\p{^Is_Scx:   inherited}', "");
Expect(0, 917999, '\P{Is_Scx:   inherited}', "");
Expect(1, 917999, '\P{^Is_Scx:   inherited}', "");
Expect(0, 918000, '\p{Is_Scx:   inherited}', "");
Expect(1, 918000, '\p{^Is_Scx:   inherited}', "");
Expect(1, 918000, '\P{Is_Scx:   inherited}', "");
Expect(0, 918000, '\P{^Is_Scx:   inherited}', "");
Expect(1, 917999, '\p{Is_Scx:	__Inherited}', "");
Expect(0, 917999, '\p{^Is_Scx:	__Inherited}', "");
Expect(0, 917999, '\P{Is_Scx:	__Inherited}', "");
Expect(1, 917999, '\P{^Is_Scx:	__Inherited}', "");
Expect(0, 918000, '\p{Is_Scx:	__Inherited}', "");
Expect(1, 918000, '\p{^Is_Scx:	__Inherited}', "");
Expect(1, 918000, '\P{Is_Scx:	__Inherited}', "");
Expect(0, 918000, '\P{^Is_Scx:	__Inherited}', "");
Error('\p{Script_Extensions= :=Kaithi}');
Error('\P{Script_Extensions= :=Kaithi}');
Expect(1, 69825, '\p{Script_Extensions=kaithi}', "");
Expect(0, 69825, '\p{^Script_Extensions=kaithi}', "");
Expect(0, 69825, '\P{Script_Extensions=kaithi}', "");
Expect(1, 69825, '\P{^Script_Extensions=kaithi}', "");
Expect(0, 69826, '\p{Script_Extensions=kaithi}', "");
Expect(1, 69826, '\p{^Script_Extensions=kaithi}', "");
Expect(1, 69826, '\P{Script_Extensions=kaithi}', "");
Expect(0, 69826, '\P{^Script_Extensions=kaithi}', "");
Expect(1, 69825, '\p{Script_Extensions=  Kaithi}', "");
Expect(0, 69825, '\p{^Script_Extensions=  Kaithi}', "");
Expect(0, 69825, '\P{Script_Extensions=  Kaithi}', "");
Expect(1, 69825, '\P{^Script_Extensions=  Kaithi}', "");
Expect(0, 69826, '\p{Script_Extensions=  Kaithi}', "");
Expect(1, 69826, '\p{^Script_Extensions=  Kaithi}', "");
Expect(1, 69826, '\P{Script_Extensions=  Kaithi}', "");
Expect(0, 69826, '\P{^Script_Extensions=  Kaithi}', "");
Error('\p{Scx=__KTHI:=}');
Error('\P{Scx=__KTHI:=}');
Expect(1, 69825, '\p{Scx:	kthi}', "");
Expect(0, 69825, '\p{^Scx:	kthi}', "");
Expect(0, 69825, '\P{Scx:	kthi}', "");
Expect(1, 69825, '\P{^Scx:	kthi}', "");
Expect(0, 69826, '\p{Scx:	kthi}', "");
Expect(1, 69826, '\p{^Scx:	kthi}', "");
Expect(1, 69826, '\P{Scx:	kthi}', "");
Expect(0, 69826, '\P{^Scx:	kthi}', "");
Expect(1, 69825, '\p{Scx: -	kthi}', "");
Expect(0, 69825, '\p{^Scx: -	kthi}', "");
Expect(0, 69825, '\P{Scx: -	kthi}', "");
Expect(1, 69825, '\P{^Scx: -	kthi}', "");
Expect(0, 69826, '\p{Scx: -	kthi}', "");
Expect(1, 69826, '\p{^Scx: -	kthi}', "");
Expect(1, 69826, '\P{Scx: -	kthi}', "");
Expect(0, 69826, '\P{^Scx: -	kthi}', "");
Error('\p{Is_Script_Extensions=/a/_-Kaithi}');
Error('\P{Is_Script_Extensions=/a/_-Kaithi}');
Expect(1, 69825, '\p{Is_Script_Extensions=kaithi}', "");
Expect(0, 69825, '\p{^Is_Script_Extensions=kaithi}', "");
Expect(0, 69825, '\P{Is_Script_Extensions=kaithi}', "");
Expect(1, 69825, '\P{^Is_Script_Extensions=kaithi}', "");
Expect(0, 69826, '\p{Is_Script_Extensions=kaithi}', "");
Expect(1, 69826, '\p{^Is_Script_Extensions=kaithi}', "");
Expect(1, 69826, '\P{Is_Script_Extensions=kaithi}', "");
Expect(0, 69826, '\P{^Is_Script_Extensions=kaithi}', "");
Expect(1, 69825, '\p{Is_Script_Extensions=__Kaithi}', "");
Expect(0, 69825, '\p{^Is_Script_Extensions=__Kaithi}', "");
Expect(0, 69825, '\P{Is_Script_Extensions=__Kaithi}', "");
Expect(1, 69825, '\P{^Is_Script_Extensions=__Kaithi}', "");
Expect(0, 69826, '\p{Is_Script_Extensions=__Kaithi}', "");
Expect(1, 69826, '\p{^Is_Script_Extensions=__Kaithi}', "");
Expect(1, 69826, '\P{Is_Script_Extensions=__Kaithi}', "");
Expect(0, 69826, '\P{^Is_Script_Extensions=__Kaithi}', "");
Error('\p{Is_Scx=  KTHI:=}');
Error('\P{Is_Scx=  KTHI:=}');
Expect(1, 69825, '\p{Is_Scx=kthi}', "");
Expect(0, 69825, '\p{^Is_Scx=kthi}', "");
Expect(0, 69825, '\P{Is_Scx=kthi}', "");
Expect(1, 69825, '\P{^Is_Scx=kthi}', "");
Expect(0, 69826, '\p{Is_Scx=kthi}', "");
Expect(1, 69826, '\p{^Is_Scx=kthi}', "");
Expect(1, 69826, '\P{Is_Scx=kthi}', "");
Expect(0, 69826, '\P{^Is_Scx=kthi}', "");
Expect(1, 69825, '\p{Is_Scx=	kthi}', "");
Expect(0, 69825, '\p{^Is_Scx=	kthi}', "");
Expect(0, 69825, '\P{Is_Scx=	kthi}', "");
Expect(1, 69825, '\P{^Is_Scx=	kthi}', "");
Expect(0, 69826, '\p{Is_Scx=	kthi}', "");
Expect(1, 69826, '\p{^Is_Scx=	kthi}', "");
Expect(1, 69826, '\P{Is_Scx=	kthi}', "");
Expect(0, 69826, '\P{^Is_Scx=	kthi}', "");
Error('\p{Script_Extensions= /a/ARABIC}');
Error('\P{Script_Extensions= /a/ARABIC}');
Expect(1, 126705, '\p{Script_Extensions=arabic}', "");
Expect(0, 126705, '\p{^Script_Extensions=arabic}', "");
Expect(0, 126705, '\P{Script_Extensions=arabic}', "");
Expect(1, 126705, '\P{^Script_Extensions=arabic}', "");
Expect(0, 126706, '\p{Script_Extensions=arabic}', "");
Expect(1, 126706, '\p{^Script_Extensions=arabic}', "");
Expect(1, 126706, '\P{Script_Extensions=arabic}', "");
Expect(0, 126706, '\P{^Script_Extensions=arabic}', "");
Expect(1, 126705, '\p{Script_Extensions=-_Arabic}', "");
Expect(0, 126705, '\p{^Script_Extensions=-_Arabic}', "");
Expect(0, 126705, '\P{Script_Extensions=-_Arabic}', "");
Expect(1, 126705, '\P{^Script_Extensions=-_Arabic}', "");
Expect(0, 126706, '\p{Script_Extensions=-_Arabic}', "");
Expect(1, 126706, '\p{^Script_Extensions=-_Arabic}', "");
Expect(1, 126706, '\P{Script_Extensions=-_Arabic}', "");
Expect(0, 126706, '\P{^Script_Extensions=-_Arabic}', "");
Error('\p{Scx=:= 	ARAB}');
Error('\P{Scx=:= 	ARAB}');
Expect(1, 126705, '\p{Scx=arab}', "");
Expect(0, 126705, '\p{^Scx=arab}', "");
Expect(0, 126705, '\P{Scx=arab}', "");
Expect(1, 126705, '\P{^Scx=arab}', "");
Expect(0, 126706, '\p{Scx=arab}', "");
Expect(1, 126706, '\p{^Scx=arab}', "");
Expect(1, 126706, '\P{Scx=arab}', "");
Expect(0, 126706, '\P{^Scx=arab}', "");
Expect(1, 126705, '\p{Scx=--Arab}', "");
Expect(0, 126705, '\p{^Scx=--Arab}', "");
Expect(0, 126705, '\P{Scx=--Arab}', "");
Expect(1, 126705, '\P{^Scx=--Arab}', "");
Expect(0, 126706, '\p{Scx=--Arab}', "");
Expect(1, 126706, '\p{^Scx=--Arab}', "");
Expect(1, 126706, '\P{Scx=--Arab}', "");
Expect(0, 126706, '\P{^Scx=--Arab}', "");
Error('\p{Is_Script_Extensions=_ arabic:=}');
Error('\P{Is_Script_Extensions=_ arabic:=}');
Expect(1, 126705, '\p{Is_Script_Extensions=arabic}', "");
Expect(0, 126705, '\p{^Is_Script_Extensions=arabic}', "");
Expect(0, 126705, '\P{Is_Script_Extensions=arabic}', "");
Expect(1, 126705, '\P{^Is_Script_Extensions=arabic}', "");
Expect(0, 126706, '\p{Is_Script_Extensions=arabic}', "");
Expect(1, 126706, '\p{^Is_Script_Extensions=arabic}', "");
Expect(1, 126706, '\P{Is_Script_Extensions=arabic}', "");
Expect(0, 126706, '\P{^Is_Script_Extensions=arabic}', "");
Expect(1, 126705, '\p{Is_Script_Extensions: 	 Arabic}', "");
Expect(0, 126705, '\p{^Is_Script_Extensions: 	 Arabic}', "");
Expect(0, 126705, '\P{Is_Script_Extensions: 	 Arabic}', "");
Expect(1, 126705, '\P{^Is_Script_Extensions: 	 Arabic}', "");
Expect(0, 126706, '\p{Is_Script_Extensions: 	 Arabic}', "");
Expect(1, 126706, '\p{^Is_Script_Extensions: 	 Arabic}', "");
Expect(1, 126706, '\P{Is_Script_Extensions: 	 Arabic}', "");
Expect(0, 126706, '\P{^Is_Script_Extensions: 	 Arabic}', "");
Error('\p{Is_Scx= 	ARAB:=}');
Error('\P{Is_Scx= 	ARAB:=}');
Expect(1, 126705, '\p{Is_Scx: arab}', "");
Expect(0, 126705, '\p{^Is_Scx: arab}', "");
Expect(0, 126705, '\P{Is_Scx: arab}', "");
Expect(1, 126705, '\P{^Is_Scx: arab}', "");
Expect(0, 126706, '\p{Is_Scx: arab}', "");
Expect(1, 126706, '\p{^Is_Scx: arab}', "");
Expect(1, 126706, '\P{Is_Scx: arab}', "");
Expect(0, 126706, '\P{^Is_Scx: arab}', "");
Expect(1, 126705, '\p{Is_Scx: _ARAB}', "");
Expect(0, 126705, '\p{^Is_Scx: _ARAB}', "");
Expect(0, 126705, '\P{Is_Scx: _ARAB}', "");
Expect(1, 126705, '\P{^Is_Scx: _ARAB}', "");
Expect(0, 126706, '\p{Is_Scx: _ARAB}', "");
Expect(1, 126706, '\p{^Is_Scx: _ARAB}', "");
Expect(1, 126706, '\P{Is_Scx: _ARAB}', "");
Expect(0, 126706, '\P{^Is_Scx: _ARAB}', "");
Error('\p{Script_Extensions=/a/__Canadian_Aboriginal}');
Error('\P{Script_Extensions=/a/__Canadian_Aboriginal}');
Expect(1, 6389, '\p{Script_Extensions=canadianaboriginal}', "");
Expect(0, 6389, '\p{^Script_Extensions=canadianaboriginal}', "");
Expect(0, 6389, '\P{Script_Extensions=canadianaboriginal}', "");
Expect(1, 6389, '\P{^Script_Extensions=canadianaboriginal}', "");
Expect(0, 6390, '\p{Script_Extensions=canadianaboriginal}', "");
Expect(1, 6390, '\p{^Script_Extensions=canadianaboriginal}', "");
Expect(1, 6390, '\P{Script_Extensions=canadianaboriginal}', "");
Expect(0, 6390, '\P{^Script_Extensions=canadianaboriginal}', "");
Expect(1, 6389, '\p{Script_Extensions=	Canadian_ABORIGINAL}', "");
Expect(0, 6389, '\p{^Script_Extensions=	Canadian_ABORIGINAL}', "");
Expect(0, 6389, '\P{Script_Extensions=	Canadian_ABORIGINAL}', "");
Expect(1, 6389, '\P{^Script_Extensions=	Canadian_ABORIGINAL}', "");
Expect(0, 6390, '\p{Script_Extensions=	Canadian_ABORIGINAL}', "");
Expect(1, 6390, '\p{^Script_Extensions=	Canadian_ABORIGINAL}', "");
Expect(1, 6390, '\P{Script_Extensions=	Canadian_ABORIGINAL}', "");
Expect(0, 6390, '\P{^Script_Extensions=	Canadian_ABORIGINAL}', "");
Error('\p{Scx=	/a/Cans}');
Error('\P{Scx=	/a/Cans}');
Expect(1, 6389, '\p{Scx=cans}', "");
Expect(0, 6389, '\p{^Scx=cans}', "");
Expect(0, 6389, '\P{Scx=cans}', "");
Expect(1, 6389, '\P{^Scx=cans}', "");
Expect(0, 6390, '\p{Scx=cans}', "");
Expect(1, 6390, '\p{^Scx=cans}', "");
Expect(1, 6390, '\P{Scx=cans}', "");
Expect(0, 6390, '\P{^Scx=cans}', "");
Expect(1, 6389, '\p{Scx=	-cans}', "");
Expect(0, 6389, '\p{^Scx=	-cans}', "");
Expect(0, 6389, '\P{Scx=	-cans}', "");
Expect(1, 6389, '\P{^Scx=	-cans}', "");
Expect(0, 6390, '\p{Scx=	-cans}', "");
Expect(1, 6390, '\p{^Scx=	-cans}', "");
Expect(1, 6390, '\P{Scx=	-cans}', "");
Expect(0, 6390, '\P{^Scx=	-cans}', "");
Error('\p{Is_Script_Extensions=_Canadian_ABORIGINAL:=}');
Error('\P{Is_Script_Extensions=_Canadian_ABORIGINAL:=}');
Expect(1, 6389, '\p{Is_Script_Extensions=canadianaboriginal}', "");
Expect(0, 6389, '\p{^Is_Script_Extensions=canadianaboriginal}', "");
Expect(0, 6389, '\P{Is_Script_Extensions=canadianaboriginal}', "");
Expect(1, 6389, '\P{^Is_Script_Extensions=canadianaboriginal}', "");
Expect(0, 6390, '\p{Is_Script_Extensions=canadianaboriginal}', "");
Expect(1, 6390, '\p{^Is_Script_Extensions=canadianaboriginal}', "");
Expect(1, 6390, '\P{Is_Script_Extensions=canadianaboriginal}', "");
Expect(0, 6390, '\P{^Is_Script_Extensions=canadianaboriginal}', "");
Expect(1, 6389, '\p{Is_Script_Extensions=	_Canadian_ABORIGINAL}', "");
Expect(0, 6389, '\p{^Is_Script_Extensions=	_Canadian_ABORIGINAL}', "");
Expect(0, 6389, '\P{Is_Script_Extensions=	_Canadian_ABORIGINAL}', "");
Expect(1, 6389, '\P{^Is_Script_Extensions=	_Canadian_ABORIGINAL}', "");
Expect(0, 6390, '\p{Is_Script_Extensions=	_Canadian_ABORIGINAL}', "");
Expect(1, 6390, '\p{^Is_Script_Extensions=	_Canadian_ABORIGINAL}', "");
Expect(1, 6390, '\P{Is_Script_Extensions=	_Canadian_ABORIGINAL}', "");
Expect(0, 6390, '\P{^Is_Script_Extensions=	_Canadian_ABORIGINAL}', "");
Error('\p{Is_Scx=/a/  Cans}');
Error('\P{Is_Scx=/a/  Cans}');
Expect(1, 6389, '\p{Is_Scx:cans}', "");
Expect(0, 6389, '\p{^Is_Scx:cans}', "");
Expect(0, 6389, '\P{Is_Scx:cans}', "");
Expect(1, 6389, '\P{^Is_Scx:cans}', "");
Expect(0, 6390, '\p{Is_Scx:cans}', "");
Expect(1, 6390, '\p{^Is_Scx:cans}', "");
Expect(1, 6390, '\P{Is_Scx:cans}', "");
Expect(0, 6390, '\P{^Is_Scx:cans}', "");
Expect(1, 6389, '\p{Is_Scx=_Cans}', "");
Expect(0, 6389, '\p{^Is_Scx=_Cans}', "");
Expect(0, 6389, '\P{Is_Scx=_Cans}', "");
Expect(1, 6389, '\P{^Is_Scx=_Cans}', "");
Expect(0, 6390, '\p{Is_Scx=_Cans}', "");
Expect(1, 6390, '\p{^Is_Scx=_Cans}', "");
Expect(1, 6390, '\P{Is_Scx=_Cans}', "");
Expect(0, 6390, '\P{^Is_Scx=_Cans}', "");
Error('\p{Script_Extensions= :=Shavian}');
Error('\P{Script_Extensions= :=Shavian}');
Expect(1, 66687, '\p{Script_Extensions=shavian}', "");
Expect(0, 66687, '\p{^Script_Extensions=shavian}', "");
Expect(0, 66687, '\P{Script_Extensions=shavian}', "");
Expect(1, 66687, '\P{^Script_Extensions=shavian}', "");
Expect(0, 66688, '\p{Script_Extensions=shavian}', "");
Expect(1, 66688, '\p{^Script_Extensions=shavian}', "");
Expect(1, 66688, '\P{Script_Extensions=shavian}', "");
Expect(0, 66688, '\P{^Script_Extensions=shavian}', "");
Expect(1, 66687, '\p{Script_Extensions: _ SHAVIAN}', "");
Expect(0, 66687, '\p{^Script_Extensions: _ SHAVIAN}', "");
Expect(0, 66687, '\P{Script_Extensions: _ SHAVIAN}', "");
Expect(1, 66687, '\P{^Script_Extensions: _ SHAVIAN}', "");
Expect(0, 66688, '\p{Script_Extensions: _ SHAVIAN}', "");
Expect(1, 66688, '\p{^Script_Extensions: _ SHAVIAN}', "");
Expect(1, 66688, '\P{Script_Extensions: _ SHAVIAN}', "");
Expect(0, 66688, '\P{^Script_Extensions: _ SHAVIAN}', "");
Error('\p{Scx= Shaw:=}');
Error('\P{Scx= Shaw:=}');
Expect(1, 66687, '\p{Scx=shaw}', "");
Expect(0, 66687, '\p{^Scx=shaw}', "");
Expect(0, 66687, '\P{Scx=shaw}', "");
Expect(1, 66687, '\P{^Scx=shaw}', "");
Expect(0, 66688, '\p{Scx=shaw}', "");
Expect(1, 66688, '\p{^Scx=shaw}', "");
Expect(1, 66688, '\P{Scx=shaw}', "");
Expect(0, 66688, '\P{^Scx=shaw}', "");
Expect(1, 66687, '\p{Scx=-Shaw}', "");
Expect(0, 66687, '\p{^Scx=-Shaw}', "");
Expect(0, 66687, '\P{Scx=-Shaw}', "");
Expect(1, 66687, '\P{^Scx=-Shaw}', "");
Expect(0, 66688, '\p{Scx=-Shaw}', "");
Expect(1, 66688, '\p{^Scx=-Shaw}', "");
Expect(1, 66688, '\P{Scx=-Shaw}', "");
Expect(0, 66688, '\P{^Scx=-Shaw}', "");
Error('\p{Is_Script_Extensions=/a/-	shavian}');
Error('\P{Is_Script_Extensions=/a/-	shavian}');
Expect(1, 66687, '\p{Is_Script_Extensions=shavian}', "");
Expect(0, 66687, '\p{^Is_Script_Extensions=shavian}', "");
Expect(0, 66687, '\P{Is_Script_Extensions=shavian}', "");
Expect(1, 66687, '\P{^Is_Script_Extensions=shavian}', "");
Expect(0, 66688, '\p{Is_Script_Extensions=shavian}', "");
Expect(1, 66688, '\p{^Is_Script_Extensions=shavian}', "");
Expect(1, 66688, '\P{Is_Script_Extensions=shavian}', "");
Expect(0, 66688, '\P{^Is_Script_Extensions=shavian}', "");
Expect(1, 66687, '\p{Is_Script_Extensions=--shavian}', "");
Expect(0, 66687, '\p{^Is_Script_Extensions=--shavian}', "");
Expect(0, 66687, '\P{Is_Script_Extensions=--shavian}', "");
Expect(1, 66687, '\P{^Is_Script_Extensions=--shavian}', "");
Expect(0, 66688, '\p{Is_Script_Extensions=--shavian}', "");
Expect(1, 66688, '\p{^Is_Script_Extensions=--shavian}', "");
Expect(1, 66688, '\P{Is_Script_Extensions=--shavian}', "");
Expect(0, 66688, '\P{^Is_Script_Extensions=--shavian}', "");
Error('\p{Is_Scx=- shaw/a/}');
Error('\P{Is_Scx=- shaw/a/}');
Expect(1, 66687, '\p{Is_Scx: shaw}', "");
Expect(0, 66687, '\p{^Is_Scx: shaw}', "");
Expect(0, 66687, '\P{Is_Scx: shaw}', "");
Expect(1, 66687, '\P{^Is_Scx: shaw}', "");
Expect(0, 66688, '\p{Is_Scx: shaw}', "");
Expect(1, 66688, '\p{^Is_Scx: shaw}', "");
Expect(1, 66688, '\P{Is_Scx: shaw}', "");
Expect(0, 66688, '\P{^Is_Scx: shaw}', "");
Expect(1, 66687, '\p{Is_Scx=_Shaw}', "");
Expect(0, 66687, '\p{^Is_Scx=_Shaw}', "");
Expect(0, 66687, '\P{Is_Scx=_Shaw}', "");
Expect(1, 66687, '\P{^Is_Scx=_Shaw}', "");
Expect(0, 66688, '\p{Is_Scx=_Shaw}', "");
Expect(1, 66688, '\p{^Is_Scx=_Shaw}', "");
Expect(1, 66688, '\P{Is_Scx=_Shaw}', "");
Expect(0, 66688, '\P{^Is_Scx=_Shaw}', "");
Error('\p{Script_Extensions=:= Old_South_Arabian}');
Error('\P{Script_Extensions=:= Old_South_Arabian}');
Expect(1, 68223, '\p{Script_Extensions: oldsoutharabian}', "");
Expect(0, 68223, '\p{^Script_Extensions: oldsoutharabian}', "");
Expect(0, 68223, '\P{Script_Extensions: oldsoutharabian}', "");
Expect(1, 68223, '\P{^Script_Extensions: oldsoutharabian}', "");
Expect(0, 68224, '\p{Script_Extensions: oldsoutharabian}', "");
Expect(1, 68224, '\p{^Script_Extensions: oldsoutharabian}', "");
Expect(1, 68224, '\P{Script_Extensions: oldsoutharabian}', "");
Expect(0, 68224, '\P{^Script_Extensions: oldsoutharabian}', "");
Expect(1, 68223, '\p{Script_Extensions=	old_South_Arabian}', "");
Expect(0, 68223, '\p{^Script_Extensions=	old_South_Arabian}', "");
Expect(0, 68223, '\P{Script_Extensions=	old_South_Arabian}', "");
Expect(1, 68223, '\P{^Script_Extensions=	old_South_Arabian}', "");
Expect(0, 68224, '\p{Script_Extensions=	old_South_Arabian}', "");
Expect(1, 68224, '\p{^Script_Extensions=	old_South_Arabian}', "");
Expect(1, 68224, '\P{Script_Extensions=	old_South_Arabian}', "");
Expect(0, 68224, '\P{^Script_Extensions=	old_South_Arabian}', "");
Error('\p{Scx=:=sarb}');
Error('\P{Scx=:=sarb}');
Expect(1, 68223, '\p{Scx=sarb}', "");
Expect(0, 68223, '\p{^Scx=sarb}', "");
Expect(0, 68223, '\P{Scx=sarb}', "");
Expect(1, 68223, '\P{^Scx=sarb}', "");
Expect(0, 68224, '\p{Scx=sarb}', "");
Expect(1, 68224, '\p{^Scx=sarb}', "");
Expect(1, 68224, '\P{Scx=sarb}', "");
Expect(0, 68224, '\P{^Scx=sarb}', "");
Expect(1, 68223, '\p{Scx=-_Sarb}', "");
Expect(0, 68223, '\p{^Scx=-_Sarb}', "");
Expect(0, 68223, '\P{Scx=-_Sarb}', "");
Expect(1, 68223, '\P{^Scx=-_Sarb}', "");
Expect(0, 68224, '\p{Scx=-_Sarb}', "");
Expect(1, 68224, '\p{^Scx=-_Sarb}', "");
Expect(1, 68224, '\P{Scx=-_Sarb}', "");
Expect(0, 68224, '\P{^Scx=-_Sarb}', "");
Error('\p{Is_Script_Extensions=_ OLD_South_arabian:=}');
Error('\P{Is_Script_Extensions=_ OLD_South_arabian:=}');
Expect(1, 68223, '\p{Is_Script_Extensions=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Is_Script_Extensions=oldsoutharabian}', "");
Expect(0, 68223, '\P{Is_Script_Extensions=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Is_Script_Extensions=oldsoutharabian}', "");
Expect(0, 68224, '\p{Is_Script_Extensions=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Is_Script_Extensions=oldsoutharabian}', "");
Expect(1, 68224, '\P{Is_Script_Extensions=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Is_Script_Extensions=oldsoutharabian}', "");
Expect(1, 68223, '\p{Is_Script_Extensions=-old_south_arabian}', "");
Expect(0, 68223, '\p{^Is_Script_Extensions=-old_south_arabian}', "");
Expect(0, 68223, '\P{Is_Script_Extensions=-old_south_arabian}', "");
Expect(1, 68223, '\P{^Is_Script_Extensions=-old_south_arabian}', "");
Expect(0, 68224, '\p{Is_Script_Extensions=-old_south_arabian}', "");
Expect(1, 68224, '\p{^Is_Script_Extensions=-old_south_arabian}', "");
Expect(1, 68224, '\P{Is_Script_Extensions=-old_south_arabian}', "");
Expect(0, 68224, '\P{^Is_Script_Extensions=-old_south_arabian}', "");
Error('\p{Is_Scx=_-Sarb/a/}');
Error('\P{Is_Scx=_-Sarb/a/}');
Expect(1, 68223, '\p{Is_Scx=sarb}', "");
Expect(0, 68223, '\p{^Is_Scx=sarb}', "");
Expect(0, 68223, '\P{Is_Scx=sarb}', "");
Expect(1, 68223, '\P{^Is_Scx=sarb}', "");
Expect(0, 68224, '\p{Is_Scx=sarb}', "");
Expect(1, 68224, '\p{^Is_Scx=sarb}', "");
Expect(1, 68224, '\P{Is_Scx=sarb}', "");
Expect(0, 68224, '\P{^Is_Scx=sarb}', "");
Expect(1, 68223, '\p{Is_Scx=-SARB}', "");
Expect(0, 68223, '\p{^Is_Scx=-SARB}', "");
Expect(0, 68223, '\P{Is_Scx=-SARB}', "");
Expect(1, 68223, '\P{^Is_Scx=-SARB}', "");
Expect(0, 68224, '\p{Is_Scx=-SARB}', "");
Expect(1, 68224, '\p{^Is_Scx=-SARB}', "");
Expect(1, 68224, '\P{Is_Scx=-SARB}', "");
Expect(0, 68224, '\P{^Is_Scx=-SARB}', "");
Error('\p{Script_Extensions=	/a/Lydian}');
Error('\P{Script_Extensions=	/a/Lydian}');
Expect(1, 67903, '\p{Script_Extensions=lydian}', "");
Expect(0, 67903, '\p{^Script_Extensions=lydian}', "");
Expect(0, 67903, '\P{Script_Extensions=lydian}', "");
Expect(1, 67903, '\P{^Script_Extensions=lydian}', "");
Expect(0, 67904, '\p{Script_Extensions=lydian}', "");
Expect(1, 67904, '\p{^Script_Extensions=lydian}', "");
Expect(1, 67904, '\P{Script_Extensions=lydian}', "");
Expect(0, 67904, '\P{^Script_Extensions=lydian}', "");
Expect(1, 67903, '\p{Script_Extensions=-	Lydian}', "");
Expect(0, 67903, '\p{^Script_Extensions=-	Lydian}', "");
Expect(0, 67903, '\P{Script_Extensions=-	Lydian}', "");
Expect(1, 67903, '\P{^Script_Extensions=-	Lydian}', "");
Expect(0, 67904, '\p{Script_Extensions=-	Lydian}', "");
Expect(1, 67904, '\p{^Script_Extensions=-	Lydian}', "");
Expect(1, 67904, '\P{Script_Extensions=-	Lydian}', "");
Expect(0, 67904, '\P{^Script_Extensions=-	Lydian}', "");
Error('\p{Scx: := -Lydi}');
Error('\P{Scx: := -Lydi}');
Expect(1, 67903, '\p{Scx=lydi}', "");
Expect(0, 67903, '\p{^Scx=lydi}', "");
Expect(0, 67903, '\P{Scx=lydi}', "");
Expect(1, 67903, '\P{^Scx=lydi}', "");
Expect(0, 67904, '\p{Scx=lydi}', "");
Expect(1, 67904, '\p{^Scx=lydi}', "");
Expect(1, 67904, '\P{Scx=lydi}', "");
Expect(0, 67904, '\P{^Scx=lydi}', "");
Expect(1, 67903, '\p{Scx= _Lydi}', "");
Expect(0, 67903, '\p{^Scx= _Lydi}', "");
Expect(0, 67903, '\P{Scx= _Lydi}', "");
Expect(1, 67903, '\P{^Scx= _Lydi}', "");
Expect(0, 67904, '\p{Scx= _Lydi}', "");
Expect(1, 67904, '\p{^Scx= _Lydi}', "");
Expect(1, 67904, '\P{Scx= _Lydi}', "");
Expect(0, 67904, '\P{^Scx= _Lydi}', "");
Error('\p{Is_Script_Extensions= /a/Lydian}');
Error('\P{Is_Script_Extensions= /a/Lydian}');
Expect(1, 67903, '\p{Is_Script_Extensions=lydian}', "");
Expect(0, 67903, '\p{^Is_Script_Extensions=lydian}', "");
Expect(0, 67903, '\P{Is_Script_Extensions=lydian}', "");
Expect(1, 67903, '\P{^Is_Script_Extensions=lydian}', "");
Expect(0, 67904, '\p{Is_Script_Extensions=lydian}', "");
Expect(1, 67904, '\p{^Is_Script_Extensions=lydian}', "");
Expect(1, 67904, '\P{Is_Script_Extensions=lydian}', "");
Expect(0, 67904, '\P{^Is_Script_Extensions=lydian}', "");
Expect(1, 67903, '\p{Is_Script_Extensions=		Lydian}', "");
Expect(0, 67903, '\p{^Is_Script_Extensions=		Lydian}', "");
Expect(0, 67903, '\P{Is_Script_Extensions=		Lydian}', "");
Expect(1, 67903, '\P{^Is_Script_Extensions=		Lydian}', "");
Expect(0, 67904, '\p{Is_Script_Extensions=		Lydian}', "");
Expect(1, 67904, '\p{^Is_Script_Extensions=		Lydian}', "");
Expect(1, 67904, '\P{Is_Script_Extensions=		Lydian}', "");
Expect(0, 67904, '\P{^Is_Script_Extensions=		Lydian}', "");
Error('\p{Is_Scx=_-lydi/a/}');
Error('\P{Is_Scx=_-lydi/a/}');
Expect(1, 67903, '\p{Is_Scx=lydi}', "");
Expect(0, 67903, '\p{^Is_Scx=lydi}', "");
Expect(0, 67903, '\P{Is_Scx=lydi}', "");
Expect(1, 67903, '\P{^Is_Scx=lydi}', "");
Expect(0, 67904, '\p{Is_Scx=lydi}', "");
Expect(1, 67904, '\p{^Is_Scx=lydi}', "");
Expect(1, 67904, '\P{Is_Scx=lydi}', "");
Expect(0, 67904, '\P{^Is_Scx=lydi}', "");
Expect(1, 67903, '\p{Is_Scx=_Lydi}', "");
Expect(0, 67903, '\p{^Is_Scx=_Lydi}', "");
Expect(0, 67903, '\P{Is_Scx=_Lydi}', "");
Expect(1, 67903, '\P{^Is_Scx=_Lydi}', "");
Expect(0, 67904, '\p{Is_Scx=_Lydi}', "");
Expect(1, 67904, '\p{^Is_Scx=_Lydi}', "");
Expect(1, 67904, '\P{Is_Scx=_Lydi}', "");
Expect(0, 67904, '\P{^Is_Scx=_Lydi}', "");
Error('\p{Script_Extensions=-Lao:=}');
Error('\P{Script_Extensions=-Lao:=}');
Expect(1, 3807, '\p{Script_Extensions=lao}', "");
Expect(0, 3807, '\p{^Script_Extensions=lao}', "");
Expect(0, 3807, '\P{Script_Extensions=lao}', "");
Expect(1, 3807, '\P{^Script_Extensions=lao}', "");
Expect(0, 3808, '\p{Script_Extensions=lao}', "");
Expect(1, 3808, '\p{^Script_Extensions=lao}', "");
Expect(1, 3808, '\P{Script_Extensions=lao}', "");
Expect(0, 3808, '\P{^Script_Extensions=lao}', "");
Expect(1, 3807, '\p{Script_Extensions:Lao}', "");
Expect(0, 3807, '\p{^Script_Extensions:Lao}', "");
Expect(0, 3807, '\P{Script_Extensions:Lao}', "");
Expect(1, 3807, '\P{^Script_Extensions:Lao}', "");
Expect(0, 3808, '\p{Script_Extensions:Lao}', "");
Expect(1, 3808, '\p{^Script_Extensions:Lao}', "");
Expect(1, 3808, '\P{Script_Extensions:Lao}', "");
Expect(0, 3808, '\P{^Script_Extensions:Lao}', "");
Error('\p{Scx=:=- Laoo}');
Error('\P{Scx=:=- Laoo}');
Expect(1, 3807, '\p{Scx=laoo}', "");
Expect(0, 3807, '\p{^Scx=laoo}', "");
Expect(0, 3807, '\P{Scx=laoo}', "");
Expect(1, 3807, '\P{^Scx=laoo}', "");
Expect(0, 3808, '\p{Scx=laoo}', "");
Expect(1, 3808, '\p{^Scx=laoo}', "");
Expect(1, 3808, '\P{Scx=laoo}', "");
Expect(0, 3808, '\P{^Scx=laoo}', "");
Expect(1, 3807, '\p{Scx=_ LAOO}', "");
Expect(0, 3807, '\p{^Scx=_ LAOO}', "");
Expect(0, 3807, '\P{Scx=_ LAOO}', "");
Expect(1, 3807, '\P{^Scx=_ LAOO}', "");
Expect(0, 3808, '\p{Scx=_ LAOO}', "");
Expect(1, 3808, '\p{^Scx=_ LAOO}', "");
Expect(1, 3808, '\P{Scx=_ LAOO}', "");
Expect(0, 3808, '\P{^Scx=_ LAOO}', "");
Error('\p{Is_Script_Extensions= 	Lao:=}');
Error('\P{Is_Script_Extensions= 	Lao:=}');
Expect(1, 3807, '\p{Is_Script_Extensions=lao}', "");
Expect(0, 3807, '\p{^Is_Script_Extensions=lao}', "");
Expect(0, 3807, '\P{Is_Script_Extensions=lao}', "");
Expect(1, 3807, '\P{^Is_Script_Extensions=lao}', "");
Expect(0, 3808, '\p{Is_Script_Extensions=lao}', "");
Expect(1, 3808, '\p{^Is_Script_Extensions=lao}', "");
Expect(1, 3808, '\P{Is_Script_Extensions=lao}', "");
Expect(0, 3808, '\P{^Is_Script_Extensions=lao}', "");
Expect(1, 3807, '\p{Is_Script_Extensions=	-lao}', "");
Expect(0, 3807, '\p{^Is_Script_Extensions=	-lao}', "");
Expect(0, 3807, '\P{Is_Script_Extensions=	-lao}', "");
Expect(1, 3807, '\P{^Is_Script_Extensions=	-lao}', "");
Expect(0, 3808, '\p{Is_Script_Extensions=	-lao}', "");
Expect(1, 3808, '\p{^Is_Script_Extensions=	-lao}', "");
Expect(1, 3808, '\P{Is_Script_Extensions=	-lao}', "");
Expect(0, 3808, '\P{^Is_Script_Extensions=	-lao}', "");
Error('\p{Is_Scx=/a/Laoo}');
Error('\P{Is_Scx=/a/Laoo}');
Expect(1, 3807, '\p{Is_Scx=laoo}', "");
Expect(0, 3807, '\p{^Is_Scx=laoo}', "");
Expect(0, 3807, '\P{Is_Scx=laoo}', "");
Expect(1, 3807, '\P{^Is_Scx=laoo}', "");
Expect(0, 3808, '\p{Is_Scx=laoo}', "");
Expect(1, 3808, '\p{^Is_Scx=laoo}', "");
Expect(1, 3808, '\P{Is_Scx=laoo}', "");
Expect(0, 3808, '\P{^Is_Scx=laoo}', "");
Expect(1, 3807, '\p{Is_Scx=- LAOO}', "");
Expect(0, 3807, '\p{^Is_Scx=- LAOO}', "");
Expect(0, 3807, '\P{Is_Scx=- LAOO}', "");
Expect(1, 3807, '\P{^Is_Scx=- LAOO}', "");
Expect(0, 3808, '\p{Is_Scx=- LAOO}', "");
Expect(1, 3808, '\p{^Is_Scx=- LAOO}', "");
Expect(1, 3808, '\P{Is_Scx=- LAOO}', "");
Expect(0, 3808, '\P{^Is_Scx=- LAOO}', "");
Error('\p{Script_Extensions=-:=MONGOLIAN}');
Error('\P{Script_Extensions=-:=MONGOLIAN}');
Expect(1, 6314, '\p{Script_Extensions=mongolian}', "");
Expect(0, 6314, '\p{^Script_Extensions=mongolian}', "");
Expect(0, 6314, '\P{Script_Extensions=mongolian}', "");
Expect(1, 6314, '\P{^Script_Extensions=mongolian}', "");
Expect(0, 6315, '\p{Script_Extensions=mongolian}', "");
Expect(1, 6315, '\p{^Script_Extensions=mongolian}', "");
Expect(1, 6315, '\P{Script_Extensions=mongolian}', "");
Expect(0, 6315, '\P{^Script_Extensions=mongolian}', "");
Expect(1, 6314, '\p{Script_Extensions=-	MONGOLIAN}', "");
Expect(0, 6314, '\p{^Script_Extensions=-	MONGOLIAN}', "");
Expect(0, 6314, '\P{Script_Extensions=-	MONGOLIAN}', "");
Expect(1, 6314, '\P{^Script_Extensions=-	MONGOLIAN}', "");
Expect(0, 6315, '\p{Script_Extensions=-	MONGOLIAN}', "");
Expect(1, 6315, '\p{^Script_Extensions=-	MONGOLIAN}', "");
Expect(1, 6315, '\P{Script_Extensions=-	MONGOLIAN}', "");
Expect(0, 6315, '\P{^Script_Extensions=-	MONGOLIAN}', "");
Error('\p{Scx=--mong:=}');
Error('\P{Scx=--mong:=}');
Expect(1, 6314, '\p{Scx=mong}', "");
Expect(0, 6314, '\p{^Scx=mong}', "");
Expect(0, 6314, '\P{Scx=mong}', "");
Expect(1, 6314, '\P{^Scx=mong}', "");
Expect(0, 6315, '\p{Scx=mong}', "");
Expect(1, 6315, '\p{^Scx=mong}', "");
Expect(1, 6315, '\P{Scx=mong}', "");
Expect(0, 6315, '\P{^Scx=mong}', "");
Expect(1, 6314, '\p{Scx=	MONG}', "");
Expect(0, 6314, '\p{^Scx=	MONG}', "");
Expect(0, 6314, '\P{Scx=	MONG}', "");
Expect(1, 6314, '\P{^Scx=	MONG}', "");
Expect(0, 6315, '\p{Scx=	MONG}', "");
Expect(1, 6315, '\p{^Scx=	MONG}', "");
Expect(1, 6315, '\P{Scx=	MONG}', "");
Expect(0, 6315, '\P{^Scx=	MONG}', "");
Error('\p{Is_Script_Extensions=_ MONGOLIAN:=}');
Error('\P{Is_Script_Extensions=_ MONGOLIAN:=}');
Expect(1, 6314, '\p{Is_Script_Extensions=mongolian}', "");
Expect(0, 6314, '\p{^Is_Script_Extensions=mongolian}', "");
Expect(0, 6314, '\P{Is_Script_Extensions=mongolian}', "");
Expect(1, 6314, '\P{^Is_Script_Extensions=mongolian}', "");
Expect(0, 6315, '\p{Is_Script_Extensions=mongolian}', "");
Expect(1, 6315, '\p{^Is_Script_Extensions=mongolian}', "");
Expect(1, 6315, '\P{Is_Script_Extensions=mongolian}', "");
Expect(0, 6315, '\P{^Is_Script_Extensions=mongolian}', "");
Expect(1, 6314, '\p{Is_Script_Extensions=_Mongolian}', "");
Expect(0, 6314, '\p{^Is_Script_Extensions=_Mongolian}', "");
Expect(0, 6314, '\P{Is_Script_Extensions=_Mongolian}', "");
Expect(1, 6314, '\P{^Is_Script_Extensions=_Mongolian}', "");
Expect(0, 6315, '\p{Is_Script_Extensions=_Mongolian}', "");
Expect(1, 6315, '\p{^Is_Script_Extensions=_Mongolian}', "");
Expect(1, 6315, '\P{Is_Script_Extensions=_Mongolian}', "");
Expect(0, 6315, '\P{^Is_Script_Extensions=_Mongolian}', "");
Error('\p{Is_Scx=/a/	Mong}');
Error('\P{Is_Scx=/a/	Mong}');
Expect(1, 6314, '\p{Is_Scx=mong}', "");
Expect(0, 6314, '\p{^Is_Scx=mong}', "");
Expect(0, 6314, '\P{Is_Scx=mong}', "");
Expect(1, 6314, '\P{^Is_Scx=mong}', "");
Expect(0, 6315, '\p{Is_Scx=mong}', "");
Expect(1, 6315, '\p{^Is_Scx=mong}', "");
Expect(1, 6315, '\P{Is_Scx=mong}', "");
Expect(0, 6315, '\P{^Is_Scx=mong}', "");
Expect(1, 6314, '\p{Is_Scx:	Mong}', "");
Expect(0, 6314, '\p{^Is_Scx:	Mong}', "");
Expect(0, 6314, '\P{Is_Scx:	Mong}', "");
Expect(1, 6314, '\P{^Is_Scx:	Mong}', "");
Expect(0, 6315, '\p{Is_Scx:	Mong}', "");
Expect(1, 6315, '\p{^Is_Scx:	Mong}', "");
Expect(1, 6315, '\P{Is_Scx:	Mong}', "");
Expect(0, 6315, '\P{^Is_Scx:	Mong}', "");
Error('\p{Script_Extensions=_cyrillic/a/}');
Error('\P{Script_Extensions=_cyrillic/a/}');
Expect(1, 42655, '\p{Script_Extensions=cyrillic}', "");
Expect(0, 42655, '\p{^Script_Extensions=cyrillic}', "");
Expect(0, 42655, '\P{Script_Extensions=cyrillic}', "");
Expect(1, 42655, '\P{^Script_Extensions=cyrillic}', "");
Expect(0, 42656, '\p{Script_Extensions=cyrillic}', "");
Expect(1, 42656, '\p{^Script_Extensions=cyrillic}', "");
Expect(1, 42656, '\P{Script_Extensions=cyrillic}', "");
Expect(0, 42656, '\P{^Script_Extensions=cyrillic}', "");
Expect(1, 42655, '\p{Script_Extensions:--CYRILLIC}', "");
Expect(0, 42655, '\p{^Script_Extensions:--CYRILLIC}', "");
Expect(0, 42655, '\P{Script_Extensions:--CYRILLIC}', "");
Expect(1, 42655, '\P{^Script_Extensions:--CYRILLIC}', "");
Expect(0, 42656, '\p{Script_Extensions:--CYRILLIC}', "");
Expect(1, 42656, '\p{^Script_Extensions:--CYRILLIC}', "");
Expect(1, 42656, '\P{Script_Extensions:--CYRILLIC}', "");
Expect(0, 42656, '\P{^Script_Extensions:--CYRILLIC}', "");
Error('\p{Scx=_cyrl:=}');
Error('\P{Scx=_cyrl:=}');
Expect(1, 42655, '\p{Scx=cyrl}', "");
Expect(0, 42655, '\p{^Scx=cyrl}', "");
Expect(0, 42655, '\P{Scx=cyrl}', "");
Expect(1, 42655, '\P{^Scx=cyrl}', "");
Expect(0, 42656, '\p{Scx=cyrl}', "");
Expect(1, 42656, '\p{^Scx=cyrl}', "");
Expect(1, 42656, '\P{Scx=cyrl}', "");
Expect(0, 42656, '\P{^Scx=cyrl}', "");
Expect(1, 42655, '\p{Scx=		Cyrl}', "");
Expect(0, 42655, '\p{^Scx=		Cyrl}', "");
Expect(0, 42655, '\P{Scx=		Cyrl}', "");
Expect(1, 42655, '\P{^Scx=		Cyrl}', "");
Expect(0, 42656, '\p{Scx=		Cyrl}', "");
Expect(1, 42656, '\p{^Scx=		Cyrl}', "");
Expect(1, 42656, '\P{Scx=		Cyrl}', "");
Expect(0, 42656, '\P{^Scx=		Cyrl}', "");
Error('\p{Is_Script_Extensions=	:=Cyrillic}');
Error('\P{Is_Script_Extensions=	:=Cyrillic}');
Expect(1, 42655, '\p{Is_Script_Extensions=cyrillic}', "");
Expect(0, 42655, '\p{^Is_Script_Extensions=cyrillic}', "");
Expect(0, 42655, '\P{Is_Script_Extensions=cyrillic}', "");
Expect(1, 42655, '\P{^Is_Script_Extensions=cyrillic}', "");
Expect(0, 42656, '\p{Is_Script_Extensions=cyrillic}', "");
Expect(1, 42656, '\p{^Is_Script_Extensions=cyrillic}', "");
Expect(1, 42656, '\P{Is_Script_Extensions=cyrillic}', "");
Expect(0, 42656, '\P{^Is_Script_Extensions=cyrillic}', "");
Expect(1, 42655, '\p{Is_Script_Extensions:   __cyrillic}', "");
Expect(0, 42655, '\p{^Is_Script_Extensions:   __cyrillic}', "");
Expect(0, 42655, '\P{Is_Script_Extensions:   __cyrillic}', "");
Expect(1, 42655, '\P{^Is_Script_Extensions:   __cyrillic}', "");
Expect(0, 42656, '\p{Is_Script_Extensions:   __cyrillic}', "");
Expect(1, 42656, '\p{^Is_Script_Extensions:   __cyrillic}', "");
Expect(1, 42656, '\P{Is_Script_Extensions:   __cyrillic}', "");
Expect(0, 42656, '\P{^Is_Script_Extensions:   __cyrillic}', "");
Error('\p{Is_Scx=__Cyrl:=}');
Error('\P{Is_Scx=__Cyrl:=}');
Expect(1, 42655, '\p{Is_Scx=cyrl}', "");
Expect(0, 42655, '\p{^Is_Scx=cyrl}', "");
Expect(0, 42655, '\P{Is_Scx=cyrl}', "");
Expect(1, 42655, '\P{^Is_Scx=cyrl}', "");
Expect(0, 42656, '\p{Is_Scx=cyrl}', "");
Expect(1, 42656, '\p{^Is_Scx=cyrl}', "");
Expect(1, 42656, '\P{Is_Scx=cyrl}', "");
Expect(0, 42656, '\P{^Is_Scx=cyrl}', "");
Expect(1, 42655, '\p{Is_Scx=--CYRL}', "");
Expect(0, 42655, '\p{^Is_Scx=--CYRL}', "");
Expect(0, 42655, '\P{Is_Scx=--CYRL}', "");
Expect(1, 42655, '\P{^Is_Scx=--CYRL}', "");
Expect(0, 42656, '\p{Is_Scx=--CYRL}', "");
Expect(1, 42656, '\p{^Is_Scx=--CYRL}', "");
Expect(1, 42656, '\P{Is_Scx=--CYRL}', "");
Expect(0, 42656, '\P{^Is_Scx=--CYRL}', "");
Error('\p{Script_Extensions=	:=Hiragana}');
Error('\P{Script_Extensions=	:=Hiragana}');
Expect(1, 127488, '\p{Script_Extensions=hiragana}', "");
Expect(0, 127488, '\p{^Script_Extensions=hiragana}', "");
Expect(0, 127488, '\P{Script_Extensions=hiragana}', "");
Expect(1, 127488, '\P{^Script_Extensions=hiragana}', "");
Expect(0, 127489, '\p{Script_Extensions=hiragana}', "");
Expect(1, 127489, '\p{^Script_Extensions=hiragana}', "");
Expect(1, 127489, '\P{Script_Extensions=hiragana}', "");
Expect(0, 127489, '\P{^Script_Extensions=hiragana}', "");
Expect(1, 127488, '\p{Script_Extensions=- HIRAGANA}', "");
Expect(0, 127488, '\p{^Script_Extensions=- HIRAGANA}', "");
Expect(0, 127488, '\P{Script_Extensions=- HIRAGANA}', "");
Expect(1, 127488, '\P{^Script_Extensions=- HIRAGANA}', "");
Expect(0, 127489, '\p{Script_Extensions=- HIRAGANA}', "");
Expect(1, 127489, '\p{^Script_Extensions=- HIRAGANA}', "");
Expect(1, 127489, '\P{Script_Extensions=- HIRAGANA}', "");
Expect(0, 127489, '\P{^Script_Extensions=- HIRAGANA}', "");
Error('\p{Scx= /a/HIRA}');
Error('\P{Scx= /a/HIRA}');
Expect(1, 127488, '\p{Scx=hira}', "");
Expect(0, 127488, '\p{^Scx=hira}', "");
Expect(0, 127488, '\P{Scx=hira}', "");
Expect(1, 127488, '\P{^Scx=hira}', "");
Expect(0, 127489, '\p{Scx=hira}', "");
Expect(1, 127489, '\p{^Scx=hira}', "");
Expect(1, 127489, '\P{Scx=hira}', "");
Expect(0, 127489, '\P{^Scx=hira}', "");
Expect(1, 127488, '\p{Scx=	-Hira}', "");
Expect(0, 127488, '\p{^Scx=	-Hira}', "");
Expect(0, 127488, '\P{Scx=	-Hira}', "");
Expect(1, 127488, '\P{^Scx=	-Hira}', "");
Expect(0, 127489, '\p{Scx=	-Hira}', "");
Expect(1, 127489, '\p{^Scx=	-Hira}', "");
Expect(1, 127489, '\P{Scx=	-Hira}', "");
Expect(0, 127489, '\P{^Scx=	-Hira}', "");
Error('\p{Is_Script_Extensions=hiragana/a/}');
Error('\P{Is_Script_Extensions=hiragana/a/}');
Expect(1, 127488, '\p{Is_Script_Extensions=hiragana}', "");
Expect(0, 127488, '\p{^Is_Script_Extensions=hiragana}', "");
Expect(0, 127488, '\P{Is_Script_Extensions=hiragana}', "");
Expect(1, 127488, '\P{^Is_Script_Extensions=hiragana}', "");
Expect(0, 127489, '\p{Is_Script_Extensions=hiragana}', "");
Expect(1, 127489, '\p{^Is_Script_Extensions=hiragana}', "");
Expect(1, 127489, '\P{Is_Script_Extensions=hiragana}', "");
Expect(0, 127489, '\P{^Is_Script_Extensions=hiragana}', "");
Expect(1, 127488, '\p{Is_Script_Extensions=-_hiragana}', "");
Expect(0, 127488, '\p{^Is_Script_Extensions=-_hiragana}', "");
Expect(0, 127488, '\P{Is_Script_Extensions=-_hiragana}', "");
Expect(1, 127488, '\P{^Is_Script_Extensions=-_hiragana}', "");
Expect(0, 127489, '\p{Is_Script_Extensions=-_hiragana}', "");
Expect(1, 127489, '\p{^Is_Script_Extensions=-_hiragana}', "");
Expect(1, 127489, '\P{Is_Script_Extensions=-_hiragana}', "");
Expect(0, 127489, '\P{^Is_Script_Extensions=-_hiragana}', "");
Error('\p{Is_Scx:__hira/a/}');
Error('\P{Is_Scx:__hira/a/}');
Expect(1, 127488, '\p{Is_Scx=hira}', "");
Expect(0, 127488, '\p{^Is_Scx=hira}', "");
Expect(0, 127488, '\P{Is_Scx=hira}', "");
Expect(1, 127488, '\P{^Is_Scx=hira}', "");
Expect(0, 127489, '\p{Is_Scx=hira}', "");
Expect(1, 127489, '\p{^Is_Scx=hira}', "");
Expect(1, 127489, '\P{Is_Scx=hira}', "");
Expect(0, 127489, '\P{^Is_Scx=hira}', "");
Expect(1, 127488, '\p{Is_Scx=--HIRA}', "");
Expect(0, 127488, '\p{^Is_Scx=--HIRA}', "");
Expect(0, 127488, '\P{Is_Scx=--HIRA}', "");
Expect(1, 127488, '\P{^Is_Scx=--HIRA}', "");
Expect(0, 127489, '\p{Is_Scx=--HIRA}', "");
Expect(1, 127489, '\p{^Is_Scx=--HIRA}', "");
Expect(1, 127489, '\P{Is_Scx=--HIRA}', "");
Expect(0, 127489, '\P{^Is_Scx=--HIRA}', "");
Error('\p{Script_Extensions=_:=Braille}');
Error('\P{Script_Extensions=_:=Braille}');
Expect(1, 10495, '\p{Script_Extensions=braille}', "");
Expect(0, 10495, '\p{^Script_Extensions=braille}', "");
Expect(0, 10495, '\P{Script_Extensions=braille}', "");
Expect(1, 10495, '\P{^Script_Extensions=braille}', "");
Expect(0, 10496, '\p{Script_Extensions=braille}', "");
Expect(1, 10496, '\p{^Script_Extensions=braille}', "");
Expect(1, 10496, '\P{Script_Extensions=braille}', "");
Expect(0, 10496, '\P{^Script_Extensions=braille}', "");
Expect(1, 10495, '\p{Script_Extensions=_-Braille}', "");
Expect(0, 10495, '\p{^Script_Extensions=_-Braille}', "");
Expect(0, 10495, '\P{Script_Extensions=_-Braille}', "");
Expect(1, 10495, '\P{^Script_Extensions=_-Braille}', "");
Expect(0, 10496, '\p{Script_Extensions=_-Braille}', "");
Expect(1, 10496, '\p{^Script_Extensions=_-Braille}', "");
Expect(1, 10496, '\P{Script_Extensions=_-Braille}', "");
Expect(0, 10496, '\P{^Script_Extensions=_-Braille}', "");
Error('\p{Scx=-_brai:=}');
Error('\P{Scx=-_brai:=}');
Expect(1, 10495, '\p{Scx=brai}', "");
Expect(0, 10495, '\p{^Scx=brai}', "");
Expect(0, 10495, '\P{Scx=brai}', "");
Expect(1, 10495, '\P{^Scx=brai}', "");
Expect(0, 10496, '\p{Scx=brai}', "");
Expect(1, 10496, '\p{^Scx=brai}', "");
Expect(1, 10496, '\P{Scx=brai}', "");
Expect(0, 10496, '\P{^Scx=brai}', "");
Expect(1, 10495, '\p{Scx=__Brai}', "");
Expect(0, 10495, '\p{^Scx=__Brai}', "");
Expect(0, 10495, '\P{Scx=__Brai}', "");
Expect(1, 10495, '\P{^Scx=__Brai}', "");
Expect(0, 10496, '\p{Scx=__Brai}', "");
Expect(1, 10496, '\p{^Scx=__Brai}', "");
Expect(1, 10496, '\P{Scx=__Brai}', "");
Expect(0, 10496, '\P{^Scx=__Brai}', "");
Error('\p{Is_Script_Extensions:   	/a/BRAILLE}');
Error('\P{Is_Script_Extensions:   	/a/BRAILLE}');
Expect(1, 10495, '\p{Is_Script_Extensions:	braille}', "");
Expect(0, 10495, '\p{^Is_Script_Extensions:	braille}', "");
Expect(0, 10495, '\P{Is_Script_Extensions:	braille}', "");
Expect(1, 10495, '\P{^Is_Script_Extensions:	braille}', "");
Expect(0, 10496, '\p{Is_Script_Extensions:	braille}', "");
Expect(1, 10496, '\p{^Is_Script_Extensions:	braille}', "");
Expect(1, 10496, '\P{Is_Script_Extensions:	braille}', "");
Expect(0, 10496, '\P{^Is_Script_Extensions:	braille}', "");
Expect(1, 10495, '\p{Is_Script_Extensions=_BRAILLE}', "");
Expect(0, 10495, '\p{^Is_Script_Extensions=_BRAILLE}', "");
Expect(0, 10495, '\P{Is_Script_Extensions=_BRAILLE}', "");
Expect(1, 10495, '\P{^Is_Script_Extensions=_BRAILLE}', "");
Expect(0, 10496, '\p{Is_Script_Extensions=_BRAILLE}', "");
Expect(1, 10496, '\p{^Is_Script_Extensions=_BRAILLE}', "");
Expect(1, 10496, '\P{Is_Script_Extensions=_BRAILLE}', "");
Expect(0, 10496, '\P{^Is_Script_Extensions=_BRAILLE}', "");
Error('\p{Is_Scx=/a/		brai}');
Error('\P{Is_Scx=/a/		brai}');
Expect(1, 10495, '\p{Is_Scx=brai}', "");
Expect(0, 10495, '\p{^Is_Scx=brai}', "");
Expect(0, 10495, '\P{Is_Scx=brai}', "");
Expect(1, 10495, '\P{^Is_Scx=brai}', "");
Expect(0, 10496, '\p{Is_Scx=brai}', "");
Expect(1, 10496, '\p{^Is_Scx=brai}', "");
Expect(1, 10496, '\P{Is_Scx=brai}', "");
Expect(0, 10496, '\P{^Is_Scx=brai}', "");
Expect(1, 10495, '\p{Is_Scx=_	Brai}', "");
Expect(0, 10495, '\p{^Is_Scx=_	Brai}', "");
Expect(0, 10495, '\P{Is_Scx=_	Brai}', "");
Expect(1, 10495, '\P{^Is_Scx=_	Brai}', "");
Expect(0, 10496, '\p{Is_Scx=_	Brai}', "");
Expect(1, 10496, '\p{^Is_Scx=_	Brai}', "");
Expect(1, 10496, '\P{Is_Scx=_	Brai}', "");
Expect(0, 10496, '\P{^Is_Scx=_	Brai}', "");
Error('\p{Script_Extensions=_ Bengali/a/}');
Error('\P{Script_Extensions=_ Bengali/a/}');
Expect(1, 2555, '\p{Script_Extensions=bengali}', "");
Expect(0, 2555, '\p{^Script_Extensions=bengali}', "");
Expect(0, 2555, '\P{Script_Extensions=bengali}', "");
Expect(1, 2555, '\P{^Script_Extensions=bengali}', "");
Expect(0, 2556, '\p{Script_Extensions=bengali}', "");
Expect(1, 2556, '\p{^Script_Extensions=bengali}', "");
Expect(1, 2556, '\P{Script_Extensions=bengali}', "");
Expect(0, 2556, '\P{^Script_Extensions=bengali}', "");
Expect(1, 2555, '\p{Script_Extensions=-	bengali}', "");
Expect(0, 2555, '\p{^Script_Extensions=-	bengali}', "");
Expect(0, 2555, '\P{Script_Extensions=-	bengali}', "");
Expect(1, 2555, '\P{^Script_Extensions=-	bengali}', "");
Expect(0, 2556, '\p{Script_Extensions=-	bengali}', "");
Expect(1, 2556, '\p{^Script_Extensions=-	bengali}', "");
Expect(1, 2556, '\P{Script_Extensions=-	bengali}', "");
Expect(0, 2556, '\P{^Script_Extensions=-	bengali}', "");
Error('\p{Scx=:=Beng}');
Error('\P{Scx=:=Beng}');
Expect(1, 2555, '\p{Scx=beng}', "");
Expect(0, 2555, '\p{^Scx=beng}', "");
Expect(0, 2555, '\P{Scx=beng}', "");
Expect(1, 2555, '\P{^Scx=beng}', "");
Expect(0, 2556, '\p{Scx=beng}', "");
Expect(1, 2556, '\p{^Scx=beng}', "");
Expect(1, 2556, '\P{Scx=beng}', "");
Expect(0, 2556, '\P{^Scx=beng}', "");
Expect(1, 2555, '\p{Scx=	beng}', "");
Expect(0, 2555, '\p{^Scx=	beng}', "");
Expect(0, 2555, '\P{Scx=	beng}', "");
Expect(1, 2555, '\P{^Scx=	beng}', "");
Expect(0, 2556, '\p{Scx=	beng}', "");
Expect(1, 2556, '\p{^Scx=	beng}', "");
Expect(1, 2556, '\P{Scx=	beng}', "");
Expect(0, 2556, '\P{^Scx=	beng}', "");
Error('\p{Is_Script_Extensions=	Bengali/a/}');
Error('\P{Is_Script_Extensions=	Bengali/a/}');
Expect(1, 2555, '\p{Is_Script_Extensions=bengali}', "");
Expect(0, 2555, '\p{^Is_Script_Extensions=bengali}', "");
Expect(0, 2555, '\P{Is_Script_Extensions=bengali}', "");
Expect(1, 2555, '\P{^Is_Script_Extensions=bengali}', "");
Expect(0, 2556, '\p{Is_Script_Extensions=bengali}', "");
Expect(1, 2556, '\p{^Is_Script_Extensions=bengali}', "");
Expect(1, 2556, '\P{Is_Script_Extensions=bengali}', "");
Expect(0, 2556, '\P{^Is_Script_Extensions=bengali}', "");
Expect(1, 2555, '\p{Is_Script_Extensions=	Bengali}', "");
Expect(0, 2555, '\p{^Is_Script_Extensions=	Bengali}', "");
Expect(0, 2555, '\P{Is_Script_Extensions=	Bengali}', "");
Expect(1, 2555, '\P{^Is_Script_Extensions=	Bengali}', "");
Expect(0, 2556, '\p{Is_Script_Extensions=	Bengali}', "");
Expect(1, 2556, '\p{^Is_Script_Extensions=	Bengali}', "");
Expect(1, 2556, '\P{Is_Script_Extensions=	Bengali}', "");
Expect(0, 2556, '\P{^Is_Script_Extensions=	Bengali}', "");
Error('\p{Is_Scx=	/a/Beng}');
Error('\P{Is_Scx=	/a/Beng}');
Expect(1, 2555, '\p{Is_Scx=beng}', "");
Expect(0, 2555, '\p{^Is_Scx=beng}', "");
Expect(0, 2555, '\P{Is_Scx=beng}', "");
Expect(1, 2555, '\P{^Is_Scx=beng}', "");
Expect(0, 2556, '\p{Is_Scx=beng}', "");
Expect(1, 2556, '\p{^Is_Scx=beng}', "");
Expect(1, 2556, '\P{Is_Scx=beng}', "");
Expect(0, 2556, '\P{^Is_Scx=beng}', "");
Expect(1, 2555, '\p{Is_Scx=-_BENG}', "");
Expect(0, 2555, '\p{^Is_Scx=-_BENG}', "");
Expect(0, 2555, '\P{Is_Scx=-_BENG}', "");
Expect(1, 2555, '\P{^Is_Scx=-_BENG}', "");
Expect(0, 2556, '\p{Is_Scx=-_BENG}', "");
Expect(1, 2556, '\p{^Is_Scx=-_BENG}', "");
Expect(1, 2556, '\P{Is_Scx=-_BENG}', "");
Expect(0, 2556, '\P{^Is_Scx=-_BENG}', "");
Error('\p{Script_Extensions=  Kharoshthi:=}');
Error('\P{Script_Extensions=  Kharoshthi:=}');
Expect(1, 68184, '\p{Script_Extensions=kharoshthi}', "");
Expect(0, 68184, '\p{^Script_Extensions=kharoshthi}', "");
Expect(0, 68184, '\P{Script_Extensions=kharoshthi}', "");
Expect(1, 68184, '\P{^Script_Extensions=kharoshthi}', "");
Expect(0, 68185, '\p{Script_Extensions=kharoshthi}', "");
Expect(1, 68185, '\p{^Script_Extensions=kharoshthi}', "");
Expect(1, 68185, '\P{Script_Extensions=kharoshthi}', "");
Expect(0, 68185, '\P{^Script_Extensions=kharoshthi}', "");
Expect(1, 68184, '\p{Script_Extensions=_	Kharoshthi}', "");
Expect(0, 68184, '\p{^Script_Extensions=_	Kharoshthi}', "");
Expect(0, 68184, '\P{Script_Extensions=_	Kharoshthi}', "");
Expect(1, 68184, '\P{^Script_Extensions=_	Kharoshthi}', "");
Expect(0, 68185, '\p{Script_Extensions=_	Kharoshthi}', "");
Expect(1, 68185, '\p{^Script_Extensions=_	Kharoshthi}', "");
Expect(1, 68185, '\P{Script_Extensions=_	Kharoshthi}', "");
Expect(0, 68185, '\P{^Script_Extensions=_	Kharoshthi}', "");
Error('\p{Scx=/a/Khar}');
Error('\P{Scx=/a/Khar}');
Expect(1, 68184, '\p{Scx=khar}', "");
Expect(0, 68184, '\p{^Scx=khar}', "");
Expect(0, 68184, '\P{Scx=khar}', "");
Expect(1, 68184, '\P{^Scx=khar}', "");
Expect(0, 68185, '\p{Scx=khar}', "");
Expect(1, 68185, '\p{^Scx=khar}', "");
Expect(1, 68185, '\P{Scx=khar}', "");
Expect(0, 68185, '\P{^Scx=khar}', "");
Expect(1, 68184, '\p{Scx=__KHAR}', "");
Expect(0, 68184, '\p{^Scx=__KHAR}', "");
Expect(0, 68184, '\P{Scx=__KHAR}', "");
Expect(1, 68184, '\P{^Scx=__KHAR}', "");
Expect(0, 68185, '\p{Scx=__KHAR}', "");
Expect(1, 68185, '\p{^Scx=__KHAR}', "");
Expect(1, 68185, '\P{Scx=__KHAR}', "");
Expect(0, 68185, '\P{^Scx=__KHAR}', "");
Error('\p{Is_Script_Extensions=_	Kharoshthi:=}');
Error('\P{Is_Script_Extensions=_	Kharoshthi:=}');
Expect(1, 68184, '\p{Is_Script_Extensions=kharoshthi}', "");
Expect(0, 68184, '\p{^Is_Script_Extensions=kharoshthi}', "");
Expect(0, 68184, '\P{Is_Script_Extensions=kharoshthi}', "");
Expect(1, 68184, '\P{^Is_Script_Extensions=kharoshthi}', "");
Expect(0, 68185, '\p{Is_Script_Extensions=kharoshthi}', "");
Expect(1, 68185, '\p{^Is_Script_Extensions=kharoshthi}', "");
Expect(1, 68185, '\P{Is_Script_Extensions=kharoshthi}', "");
Expect(0, 68185, '\P{^Is_Script_Extensions=kharoshthi}', "");
Expect(1, 68184, '\p{Is_Script_Extensions=_-Kharoshthi}', "");
Expect(0, 68184, '\p{^Is_Script_Extensions=_-Kharoshthi}', "");
Expect(0, 68184, '\P{Is_Script_Extensions=_-Kharoshthi}', "");
Expect(1, 68184, '\P{^Is_Script_Extensions=_-Kharoshthi}', "");
Expect(0, 68185, '\p{Is_Script_Extensions=_-Kharoshthi}', "");
Expect(1, 68185, '\p{^Is_Script_Extensions=_-Kharoshthi}', "");
Expect(1, 68185, '\P{Is_Script_Extensions=_-Kharoshthi}', "");
Expect(0, 68185, '\P{^Is_Script_Extensions=_-Kharoshthi}', "");
Error('\p{Is_Scx=/a/ KHAR}');
Error('\P{Is_Scx=/a/ KHAR}');
Expect(1, 68184, '\p{Is_Scx=khar}', "");
Expect(0, 68184, '\p{^Is_Scx=khar}', "");
Expect(0, 68184, '\P{Is_Scx=khar}', "");
Expect(1, 68184, '\P{^Is_Scx=khar}', "");
Expect(0, 68185, '\p{Is_Scx=khar}', "");
Expect(1, 68185, '\p{^Is_Scx=khar}', "");
Expect(1, 68185, '\P{Is_Scx=khar}', "");
Expect(0, 68185, '\P{^Is_Scx=khar}', "");
Expect(1, 68184, '\p{Is_Scx=--Khar}', "");
Expect(0, 68184, '\p{^Is_Scx=--Khar}', "");
Expect(0, 68184, '\P{Is_Scx=--Khar}', "");
Expect(1, 68184, '\P{^Is_Scx=--Khar}', "");
Expect(0, 68185, '\p{Is_Scx=--Khar}', "");
Expect(1, 68185, '\p{^Is_Scx=--Khar}', "");
Expect(1, 68185, '\P{Is_Scx=--Khar}', "");
Expect(0, 68185, '\P{^Is_Scx=--Khar}', "");
Error('\p{Script_Extensions:  lycian/a/}');
Error('\P{Script_Extensions:  lycian/a/}');
Expect(1, 66204, '\p{Script_Extensions=lycian}', "");
Expect(0, 66204, '\p{^Script_Extensions=lycian}', "");
Expect(0, 66204, '\P{Script_Extensions=lycian}', "");
Expect(1, 66204, '\P{^Script_Extensions=lycian}', "");
Expect(0, 66205, '\p{Script_Extensions=lycian}', "");
Expect(1, 66205, '\p{^Script_Extensions=lycian}', "");
Expect(1, 66205, '\P{Script_Extensions=lycian}', "");
Expect(0, 66205, '\P{^Script_Extensions=lycian}', "");
Expect(1, 66204, '\p{Script_Extensions:   _	Lycian}', "");
Expect(0, 66204, '\p{^Script_Extensions:   _	Lycian}', "");
Expect(0, 66204, '\P{Script_Extensions:   _	Lycian}', "");
Expect(1, 66204, '\P{^Script_Extensions:   _	Lycian}', "");
Expect(0, 66205, '\p{Script_Extensions:   _	Lycian}', "");
Expect(1, 66205, '\p{^Script_Extensions:   _	Lycian}', "");
Expect(1, 66205, '\P{Script_Extensions:   _	Lycian}', "");
Expect(0, 66205, '\P{^Script_Extensions:   _	Lycian}', "");
Error('\p{Scx=/a/	Lyci}');
Error('\P{Scx=/a/	Lyci}');
Expect(1, 66204, '\p{Scx=lyci}', "");
Expect(0, 66204, '\p{^Scx=lyci}', "");
Expect(0, 66204, '\P{Scx=lyci}', "");
Expect(1, 66204, '\P{^Scx=lyci}', "");
Expect(0, 66205, '\p{Scx=lyci}', "");
Expect(1, 66205, '\p{^Scx=lyci}', "");
Expect(1, 66205, '\P{Scx=lyci}', "");
Expect(0, 66205, '\P{^Scx=lyci}', "");
Expect(1, 66204, '\p{Scx=  LYCI}', "");
Expect(0, 66204, '\p{^Scx=  LYCI}', "");
Expect(0, 66204, '\P{Scx=  LYCI}', "");
Expect(1, 66204, '\P{^Scx=  LYCI}', "");
Expect(0, 66205, '\p{Scx=  LYCI}', "");
Expect(1, 66205, '\p{^Scx=  LYCI}', "");
Expect(1, 66205, '\P{Scx=  LYCI}', "");
Expect(0, 66205, '\P{^Scx=  LYCI}', "");
Error('\p{Is_Script_Extensions=/a/--lycian}');
Error('\P{Is_Script_Extensions=/a/--lycian}');
Expect(1, 66204, '\p{Is_Script_Extensions=lycian}', "");
Expect(0, 66204, '\p{^Is_Script_Extensions=lycian}', "");
Expect(0, 66204, '\P{Is_Script_Extensions=lycian}', "");
Expect(1, 66204, '\P{^Is_Script_Extensions=lycian}', "");
Expect(0, 66205, '\p{Is_Script_Extensions=lycian}', "");
Expect(1, 66205, '\p{^Is_Script_Extensions=lycian}', "");
Expect(1, 66205, '\P{Is_Script_Extensions=lycian}', "");
Expect(0, 66205, '\P{^Is_Script_Extensions=lycian}', "");
Expect(1, 66204, '\p{Is_Script_Extensions=- Lycian}', "");
Expect(0, 66204, '\p{^Is_Script_Extensions=- Lycian}', "");
Expect(0, 66204, '\P{Is_Script_Extensions=- Lycian}', "");
Expect(1, 66204, '\P{^Is_Script_Extensions=- Lycian}', "");
Expect(0, 66205, '\p{Is_Script_Extensions=- Lycian}', "");
Expect(1, 66205, '\p{^Is_Script_Extensions=- Lycian}', "");
Expect(1, 66205, '\P{Is_Script_Extensions=- Lycian}', "");
Expect(0, 66205, '\P{^Is_Script_Extensions=- Lycian}', "");
Error('\p{Is_Scx=:=_	Lyci}');
Error('\P{Is_Scx=:=_	Lyci}');
Expect(1, 66204, '\p{Is_Scx=lyci}', "");
Expect(0, 66204, '\p{^Is_Scx=lyci}', "");
Expect(0, 66204, '\P{Is_Scx=lyci}', "");
Expect(1, 66204, '\P{^Is_Scx=lyci}', "");
Expect(0, 66205, '\p{Is_Scx=lyci}', "");
Expect(1, 66205, '\p{^Is_Scx=lyci}', "");
Expect(1, 66205, '\P{Is_Scx=lyci}', "");
Expect(0, 66205, '\P{^Is_Scx=lyci}', "");
Expect(1, 66204, '\p{Is_Scx=_LYCI}', "");
Expect(0, 66204, '\p{^Is_Scx=_LYCI}', "");
Expect(0, 66204, '\P{Is_Scx=_LYCI}', "");
Expect(1, 66204, '\P{^Is_Scx=_LYCI}', "");
Expect(0, 66205, '\p{Is_Scx=_LYCI}', "");
Expect(1, 66205, '\p{^Is_Scx=_LYCI}', "");
Expect(1, 66205, '\P{Is_Scx=_LYCI}', "");
Expect(0, 66205, '\P{^Is_Scx=_LYCI}', "");
Error('\p{Script_Extensions=-/a/bopomofo}');
Error('\P{Script_Extensions=-/a/bopomofo}');
Expect(1, 65381, '\p{Script_Extensions:	bopomofo}', "");
Expect(0, 65381, '\p{^Script_Extensions:	bopomofo}', "");
Expect(0, 65381, '\P{Script_Extensions:	bopomofo}', "");
Expect(1, 65381, '\P{^Script_Extensions:	bopomofo}', "");
Expect(0, 65382, '\p{Script_Extensions:	bopomofo}', "");
Expect(1, 65382, '\p{^Script_Extensions:	bopomofo}', "");
Expect(1, 65382, '\P{Script_Extensions:	bopomofo}', "");
Expect(0, 65382, '\P{^Script_Extensions:	bopomofo}', "");
Expect(1, 65381, '\p{Script_Extensions= -Bopomofo}', "");
Expect(0, 65381, '\p{^Script_Extensions= -Bopomofo}', "");
Expect(0, 65381, '\P{Script_Extensions= -Bopomofo}', "");
Expect(1, 65381, '\P{^Script_Extensions= -Bopomofo}', "");
Expect(0, 65382, '\p{Script_Extensions= -Bopomofo}', "");
Expect(1, 65382, '\p{^Script_Extensions= -Bopomofo}', "");
Expect(1, 65382, '\P{Script_Extensions= -Bopomofo}', "");
Expect(0, 65382, '\P{^Script_Extensions= -Bopomofo}', "");
Error('\p{Scx= 	Bopo/a/}');
Error('\P{Scx= 	Bopo/a/}');
Expect(1, 65381, '\p{Scx=bopo}', "");
Expect(0, 65381, '\p{^Scx=bopo}', "");
Expect(0, 65381, '\P{Scx=bopo}', "");
Expect(1, 65381, '\P{^Scx=bopo}', "");
Expect(0, 65382, '\p{Scx=bopo}', "");
Expect(1, 65382, '\p{^Scx=bopo}', "");
Expect(1, 65382, '\P{Scx=bopo}', "");
Expect(0, 65382, '\P{^Scx=bopo}', "");
Expect(1, 65381, '\p{Scx=--bopo}', "");
Expect(0, 65381, '\p{^Scx=--bopo}', "");
Expect(0, 65381, '\P{Scx=--bopo}', "");
Expect(1, 65381, '\P{^Scx=--bopo}', "");
Expect(0, 65382, '\p{Scx=--bopo}', "");
Expect(1, 65382, '\p{^Scx=--bopo}', "");
Expect(1, 65382, '\P{Scx=--bopo}', "");
Expect(0, 65382, '\P{^Scx=--bopo}', "");
Error('\p{Is_Script_Extensions:   	:=bopomofo}');
Error('\P{Is_Script_Extensions:   	:=bopomofo}');
Expect(1, 65381, '\p{Is_Script_Extensions=bopomofo}', "");
Expect(0, 65381, '\p{^Is_Script_Extensions=bopomofo}', "");
Expect(0, 65381, '\P{Is_Script_Extensions=bopomofo}', "");
Expect(1, 65381, '\P{^Is_Script_Extensions=bopomofo}', "");
Expect(0, 65382, '\p{Is_Script_Extensions=bopomofo}', "");
Expect(1, 65382, '\p{^Is_Script_Extensions=bopomofo}', "");
Expect(1, 65382, '\P{Is_Script_Extensions=bopomofo}', "");
Expect(0, 65382, '\P{^Is_Script_Extensions=bopomofo}', "");
Expect(1, 65381, '\p{Is_Script_Extensions= _bopomofo}', "");
Expect(0, 65381, '\p{^Is_Script_Extensions= _bopomofo}', "");
Expect(0, 65381, '\P{Is_Script_Extensions= _bopomofo}', "");
Expect(1, 65381, '\P{^Is_Script_Extensions= _bopomofo}', "");
Expect(0, 65382, '\p{Is_Script_Extensions= _bopomofo}', "");
Expect(1, 65382, '\p{^Is_Script_Extensions= _bopomofo}', "");
Expect(1, 65382, '\P{Is_Script_Extensions= _bopomofo}', "");
Expect(0, 65382, '\P{^Is_Script_Extensions= _bopomofo}', "");
Error('\p{Is_Scx=  Bopo/a/}');
Error('\P{Is_Scx=  Bopo/a/}');
Expect(1, 65381, '\p{Is_Scx=bopo}', "");
Expect(0, 65381, '\p{^Is_Scx=bopo}', "");
Expect(0, 65381, '\P{Is_Scx=bopo}', "");
Expect(1, 65381, '\P{^Is_Scx=bopo}', "");
Expect(0, 65382, '\p{Is_Scx=bopo}', "");
Expect(1, 65382, '\p{^Is_Scx=bopo}', "");
Expect(1, 65382, '\P{Is_Scx=bopo}', "");
Expect(0, 65382, '\P{^Is_Scx=bopo}', "");
Expect(1, 65381, '\p{Is_Scx= _BOPO}', "");
Expect(0, 65381, '\p{^Is_Scx= _BOPO}', "");
Expect(0, 65381, '\P{Is_Scx= _BOPO}', "");
Expect(1, 65381, '\P{^Is_Scx= _BOPO}', "");
Expect(0, 65382, '\p{Is_Scx= _BOPO}', "");
Expect(1, 65382, '\p{^Is_Scx= _BOPO}', "");
Expect(1, 65382, '\P{Is_Scx= _BOPO}', "");
Expect(0, 65382, '\P{^Is_Scx= _BOPO}', "");
Error('\p{Script_Extensions= /a/Meroitic_Cursive}');
Error('\P{Script_Extensions= /a/Meroitic_Cursive}');
Expect(1, 68031, '\p{Script_Extensions=meroiticcursive}', "");
Expect(0, 68031, '\p{^Script_Extensions=meroiticcursive}', "");
Expect(0, 68031, '\P{Script_Extensions=meroiticcursive}', "");
Expect(1, 68031, '\P{^Script_Extensions=meroiticcursive}', "");
Expect(0, 68032, '\p{Script_Extensions=meroiticcursive}', "");
Expect(1, 68032, '\p{^Script_Extensions=meroiticcursive}', "");
Expect(1, 68032, '\P{Script_Extensions=meroiticcursive}', "");
Expect(0, 68032, '\P{^Script_Extensions=meroiticcursive}', "");
Expect(1, 68031, '\p{Script_Extensions:	meroitic_CURSIVE}', "");
Expect(0, 68031, '\p{^Script_Extensions:	meroitic_CURSIVE}', "");
Expect(0, 68031, '\P{Script_Extensions:	meroitic_CURSIVE}', "");
Expect(1, 68031, '\P{^Script_Extensions:	meroitic_CURSIVE}', "");
Expect(0, 68032, '\p{Script_Extensions:	meroitic_CURSIVE}', "");
Expect(1, 68032, '\p{^Script_Extensions:	meroitic_CURSIVE}', "");
Expect(1, 68032, '\P{Script_Extensions:	meroitic_CURSIVE}', "");
Expect(0, 68032, '\P{^Script_Extensions:	meroitic_CURSIVE}', "");
Error('\p{Scx=:=Merc}');
Error('\P{Scx=:=Merc}');
Expect(1, 68031, '\p{Scx=merc}', "");
Expect(0, 68031, '\p{^Scx=merc}', "");
Expect(0, 68031, '\P{Scx=merc}', "");
Expect(1, 68031, '\P{^Scx=merc}', "");
Expect(0, 68032, '\p{Scx=merc}', "");
Expect(1, 68032, '\p{^Scx=merc}', "");
Expect(1, 68032, '\P{Scx=merc}', "");
Expect(0, 68032, '\P{^Scx=merc}', "");
Expect(1, 68031, '\p{Scx=	Merc}', "");
Expect(0, 68031, '\p{^Scx=	Merc}', "");
Expect(0, 68031, '\P{Scx=	Merc}', "");
Expect(1, 68031, '\P{^Scx=	Merc}', "");
Expect(0, 68032, '\p{Scx=	Merc}', "");
Expect(1, 68032, '\p{^Scx=	Merc}', "");
Expect(1, 68032, '\P{Scx=	Merc}', "");
Expect(0, 68032, '\P{^Scx=	Merc}', "");
Error('\p{Is_Script_Extensions=/a/MEROITIC_Cursive}');
Error('\P{Is_Script_Extensions=/a/MEROITIC_Cursive}');
Expect(1, 68031, '\p{Is_Script_Extensions=meroiticcursive}', "");
Expect(0, 68031, '\p{^Is_Script_Extensions=meroiticcursive}', "");
Expect(0, 68031, '\P{Is_Script_Extensions=meroiticcursive}', "");
Expect(1, 68031, '\P{^Is_Script_Extensions=meroiticcursive}', "");
Expect(0, 68032, '\p{Is_Script_Extensions=meroiticcursive}', "");
Expect(1, 68032, '\p{^Is_Script_Extensions=meroiticcursive}', "");
Expect(1, 68032, '\P{Is_Script_Extensions=meroiticcursive}', "");
Expect(0, 68032, '\P{^Is_Script_Extensions=meroiticcursive}', "");
Expect(1, 68031, '\p{Is_Script_Extensions:		Meroitic_CURSIVE}', "");
Expect(0, 68031, '\p{^Is_Script_Extensions:		Meroitic_CURSIVE}', "");
Expect(0, 68031, '\P{Is_Script_Extensions:		Meroitic_CURSIVE}', "");
Expect(1, 68031, '\P{^Is_Script_Extensions:		Meroitic_CURSIVE}', "");
Expect(0, 68032, '\p{Is_Script_Extensions:		Meroitic_CURSIVE}', "");
Expect(1, 68032, '\p{^Is_Script_Extensions:		Meroitic_CURSIVE}', "");
Expect(1, 68032, '\P{Is_Script_Extensions:		Meroitic_CURSIVE}', "");
Expect(0, 68032, '\P{^Is_Script_Extensions:		Meroitic_CURSIVE}', "");
Error('\p{Is_Scx=	/a/merc}');
Error('\P{Is_Scx=	/a/merc}');
Expect(1, 68031, '\p{Is_Scx=merc}', "");
Expect(0, 68031, '\p{^Is_Scx=merc}', "");
Expect(0, 68031, '\P{Is_Scx=merc}', "");
Expect(1, 68031, '\P{^Is_Scx=merc}', "");
Expect(0, 68032, '\p{Is_Scx=merc}', "");
Expect(1, 68032, '\p{^Is_Scx=merc}', "");
Expect(1, 68032, '\P{Is_Scx=merc}', "");
Expect(0, 68032, '\P{^Is_Scx=merc}', "");
Expect(1, 68031, '\p{Is_Scx:		-Merc}', "");
Expect(0, 68031, '\p{^Is_Scx:		-Merc}', "");
Expect(0, 68031, '\P{Is_Scx:		-Merc}', "");
Expect(1, 68031, '\P{^Is_Scx:		-Merc}', "");
Expect(0, 68032, '\p{Is_Scx:		-Merc}', "");
Expect(1, 68032, '\p{^Is_Scx:		-Merc}', "");
Expect(1, 68032, '\P{Is_Scx:		-Merc}', "");
Expect(0, 68032, '\P{^Is_Scx:		-Merc}', "");
Error('\p{Script_Extensions=-_Rejang/a/}');
Error('\P{Script_Extensions=-_Rejang/a/}');
Expect(1, 43359, '\p{Script_Extensions=rejang}', "");
Expect(0, 43359, '\p{^Script_Extensions=rejang}', "");
Expect(0, 43359, '\P{Script_Extensions=rejang}', "");
Expect(1, 43359, '\P{^Script_Extensions=rejang}', "");
Expect(0, 43360, '\p{Script_Extensions=rejang}', "");
Expect(1, 43360, '\p{^Script_Extensions=rejang}', "");
Expect(1, 43360, '\P{Script_Extensions=rejang}', "");
Expect(0, 43360, '\P{^Script_Extensions=rejang}', "");
Expect(1, 43359, '\p{Script_Extensions=_Rejang}', "");
Expect(0, 43359, '\p{^Script_Extensions=_Rejang}', "");
Expect(0, 43359, '\P{Script_Extensions=_Rejang}', "");
Expect(1, 43359, '\P{^Script_Extensions=_Rejang}', "");
Expect(0, 43360, '\p{Script_Extensions=_Rejang}', "");
Expect(1, 43360, '\p{^Script_Extensions=_Rejang}', "");
Expect(1, 43360, '\P{Script_Extensions=_Rejang}', "");
Expect(0, 43360, '\P{^Script_Extensions=_Rejang}', "");
Error('\p{Scx=:=Rjng}');
Error('\P{Scx=:=Rjng}');
Expect(1, 43359, '\p{Scx:rjng}', "");
Expect(0, 43359, '\p{^Scx:rjng}', "");
Expect(0, 43359, '\P{Scx:rjng}', "");
Expect(1, 43359, '\P{^Scx:rjng}', "");
Expect(0, 43360, '\p{Scx:rjng}', "");
Expect(1, 43360, '\p{^Scx:rjng}', "");
Expect(1, 43360, '\P{Scx:rjng}', "");
Expect(0, 43360, '\P{^Scx:rjng}', "");
Expect(1, 43359, '\p{Scx=		RJNG}', "");
Expect(0, 43359, '\p{^Scx=		RJNG}', "");
Expect(0, 43359, '\P{Scx=		RJNG}', "");
Expect(1, 43359, '\P{^Scx=		RJNG}', "");
Expect(0, 43360, '\p{Scx=		RJNG}', "");
Expect(1, 43360, '\p{^Scx=		RJNG}', "");
Expect(1, 43360, '\P{Scx=		RJNG}', "");
Expect(0, 43360, '\P{^Scx=		RJNG}', "");
Error('\p{Is_Script_Extensions=-rejang/a/}');
Error('\P{Is_Script_Extensions=-rejang/a/}');
Expect(1, 43359, '\p{Is_Script_Extensions=rejang}', "");
Expect(0, 43359, '\p{^Is_Script_Extensions=rejang}', "");
Expect(0, 43359, '\P{Is_Script_Extensions=rejang}', "");
Expect(1, 43359, '\P{^Is_Script_Extensions=rejang}', "");
Expect(0, 43360, '\p{Is_Script_Extensions=rejang}', "");
Expect(1, 43360, '\p{^Is_Script_Extensions=rejang}', "");
Expect(1, 43360, '\P{Is_Script_Extensions=rejang}', "");
Expect(0, 43360, '\P{^Is_Script_Extensions=rejang}', "");
Expect(1, 43359, '\p{Is_Script_Extensions:		-REJANG}', "");
Expect(0, 43359, '\p{^Is_Script_Extensions:		-REJANG}', "");
Expect(0, 43359, '\P{Is_Script_Extensions:		-REJANG}', "");
Expect(1, 43359, '\P{^Is_Script_Extensions:		-REJANG}', "");
Expect(0, 43360, '\p{Is_Script_Extensions:		-REJANG}', "");
Expect(1, 43360, '\p{^Is_Script_Extensions:		-REJANG}', "");
Expect(1, 43360, '\P{Is_Script_Extensions:		-REJANG}', "");
Expect(0, 43360, '\P{^Is_Script_Extensions:		-REJANG}', "");
Error('\p{Is_Scx=	:=RJNG}');
Error('\P{Is_Scx=	:=RJNG}');
Expect(1, 43359, '\p{Is_Scx=rjng}', "");
Expect(0, 43359, '\p{^Is_Scx=rjng}', "");
Expect(0, 43359, '\P{Is_Scx=rjng}', "");
Expect(1, 43359, '\P{^Is_Scx=rjng}', "");
Expect(0, 43360, '\p{Is_Scx=rjng}', "");
Expect(1, 43360, '\p{^Is_Scx=rjng}', "");
Expect(1, 43360, '\P{Is_Scx=rjng}', "");
Expect(0, 43360, '\P{^Is_Scx=rjng}', "");
Expect(1, 43359, '\p{Is_Scx=-Rjng}', "");
Expect(0, 43359, '\p{^Is_Scx=-Rjng}', "");
Expect(0, 43359, '\P{Is_Scx=-Rjng}', "");
Expect(1, 43359, '\P{^Is_Scx=-Rjng}', "");
Expect(0, 43360, '\p{Is_Scx=-Rjng}', "");
Expect(1, 43360, '\p{^Is_Scx=-Rjng}', "");
Expect(1, 43360, '\P{Is_Scx=-Rjng}', "");
Expect(0, 43360, '\P{^Is_Scx=-Rjng}', "");
Error('\p{Script_Extensions=-	hanunoo/a/}');
Error('\P{Script_Extensions=-	hanunoo/a/}');
Expect(1, 5942, '\p{Script_Extensions=hanunoo}', "");
Expect(0, 5942, '\p{^Script_Extensions=hanunoo}', "");
Expect(0, 5942, '\P{Script_Extensions=hanunoo}', "");
Expect(1, 5942, '\P{^Script_Extensions=hanunoo}', "");
Expect(0, 5943, '\p{Script_Extensions=hanunoo}', "");
Expect(1, 5943, '\p{^Script_Extensions=hanunoo}', "");
Expect(1, 5943, '\P{Script_Extensions=hanunoo}', "");
Expect(0, 5943, '\P{^Script_Extensions=hanunoo}', "");
Expect(1, 5942, '\p{Script_Extensions=-_Hanunoo}', "");
Expect(0, 5942, '\p{^Script_Extensions=-_Hanunoo}', "");
Expect(0, 5942, '\P{Script_Extensions=-_Hanunoo}', "");
Expect(1, 5942, '\P{^Script_Extensions=-_Hanunoo}', "");
Expect(0, 5943, '\p{Script_Extensions=-_Hanunoo}', "");
Expect(1, 5943, '\p{^Script_Extensions=-_Hanunoo}', "");
Expect(1, 5943, '\P{Script_Extensions=-_Hanunoo}', "");
Expect(0, 5943, '\P{^Script_Extensions=-_Hanunoo}', "");
Error('\p{Scx=:=- Hano}');
Error('\P{Scx=:=- Hano}');
Expect(1, 5942, '\p{Scx=hano}', "");
Expect(0, 5942, '\p{^Scx=hano}', "");
Expect(0, 5942, '\P{Scx=hano}', "");
Expect(1, 5942, '\P{^Scx=hano}', "");
Expect(0, 5943, '\p{Scx=hano}', "");
Expect(1, 5943, '\p{^Scx=hano}', "");
Expect(1, 5943, '\P{Scx=hano}', "");
Expect(0, 5943, '\P{^Scx=hano}', "");
Expect(1, 5942, '\p{Scx: 	_Hano}', "");
Expect(0, 5942, '\p{^Scx: 	_Hano}', "");
Expect(0, 5942, '\P{Scx: 	_Hano}', "");
Expect(1, 5942, '\P{^Scx: 	_Hano}', "");
Expect(0, 5943, '\p{Scx: 	_Hano}', "");
Expect(1, 5943, '\p{^Scx: 	_Hano}', "");
Expect(1, 5943, '\P{Scx: 	_Hano}', "");
Expect(0, 5943, '\P{^Scx: 	_Hano}', "");
Error('\p{Is_Script_Extensions=-Hanunoo/a/}');
Error('\P{Is_Script_Extensions=-Hanunoo/a/}');
Expect(1, 5942, '\p{Is_Script_Extensions=hanunoo}', "");
Expect(0, 5942, '\p{^Is_Script_Extensions=hanunoo}', "");
Expect(0, 5942, '\P{Is_Script_Extensions=hanunoo}', "");
Expect(1, 5942, '\P{^Is_Script_Extensions=hanunoo}', "");
Expect(0, 5943, '\p{Is_Script_Extensions=hanunoo}', "");
Expect(1, 5943, '\p{^Is_Script_Extensions=hanunoo}', "");
Expect(1, 5943, '\P{Is_Script_Extensions=hanunoo}', "");
Expect(0, 5943, '\P{^Is_Script_Extensions=hanunoo}', "");
Expect(1, 5942, '\p{Is_Script_Extensions=--hanunoo}', "");
Expect(0, 5942, '\p{^Is_Script_Extensions=--hanunoo}', "");
Expect(0, 5942, '\P{Is_Script_Extensions=--hanunoo}', "");
Expect(1, 5942, '\P{^Is_Script_Extensions=--hanunoo}', "");
Expect(0, 5943, '\p{Is_Script_Extensions=--hanunoo}', "");
Expect(1, 5943, '\p{^Is_Script_Extensions=--hanunoo}', "");
Expect(1, 5943, '\P{Is_Script_Extensions=--hanunoo}', "");
Expect(0, 5943, '\P{^Is_Script_Extensions=--hanunoo}', "");
Error('\p{Is_Scx=-:=HANO}');
Error('\P{Is_Scx=-:=HANO}');
Expect(1, 5942, '\p{Is_Scx=hano}', "");
Expect(0, 5942, '\p{^Is_Scx=hano}', "");
Expect(0, 5942, '\P{Is_Scx=hano}', "");
Expect(1, 5942, '\P{^Is_Scx=hano}', "");
Expect(0, 5943, '\p{Is_Scx=hano}', "");
Expect(1, 5943, '\p{^Is_Scx=hano}', "");
Expect(1, 5943, '\P{Is_Scx=hano}', "");
Expect(0, 5943, '\P{^Is_Scx=hano}', "");
Expect(1, 5942, '\p{Is_Scx= Hano}', "");
Expect(0, 5942, '\p{^Is_Scx= Hano}', "");
Expect(0, 5942, '\P{Is_Scx= Hano}', "");
Expect(1, 5942, '\P{^Is_Scx= Hano}', "");
Expect(0, 5943, '\p{Is_Scx= Hano}', "");
Expect(1, 5943, '\p{^Is_Scx= Hano}', "");
Expect(1, 5943, '\P{Is_Scx= Hano}', "");
Expect(0, 5943, '\P{^Is_Scx= Hano}', "");
Error('\p{Script_Extensions=:=-	IMPERIAL_ARAMAIC}');
Error('\P{Script_Extensions=:=-	IMPERIAL_ARAMAIC}');
Expect(1, 67679, '\p{Script_Extensions=imperialaramaic}', "");
Expect(0, 67679, '\p{^Script_Extensions=imperialaramaic}', "");
Expect(0, 67679, '\P{Script_Extensions=imperialaramaic}', "");
Expect(1, 67679, '\P{^Script_Extensions=imperialaramaic}', "");
Expect(0, 67680, '\p{Script_Extensions=imperialaramaic}', "");
Expect(1, 67680, '\p{^Script_Extensions=imperialaramaic}', "");
Expect(1, 67680, '\P{Script_Extensions=imperialaramaic}', "");
Expect(0, 67680, '\P{^Script_Extensions=imperialaramaic}', "");
Expect(1, 67679, '\p{Script_Extensions=_imperial_ARAMAIC}', "");
Expect(0, 67679, '\p{^Script_Extensions=_imperial_ARAMAIC}', "");
Expect(0, 67679, '\P{Script_Extensions=_imperial_ARAMAIC}', "");
Expect(1, 67679, '\P{^Script_Extensions=_imperial_ARAMAIC}', "");
Expect(0, 67680, '\p{Script_Extensions=_imperial_ARAMAIC}', "");
Expect(1, 67680, '\p{^Script_Extensions=_imperial_ARAMAIC}', "");
Expect(1, 67680, '\P{Script_Extensions=_imperial_ARAMAIC}', "");
Expect(0, 67680, '\P{^Script_Extensions=_imperial_ARAMAIC}', "");
Error('\p{Scx= /a/Armi}');
Error('\P{Scx= /a/Armi}');
Expect(1, 67679, '\p{Scx=armi}', "");
Expect(0, 67679, '\p{^Scx=armi}', "");
Expect(0, 67679, '\P{Scx=armi}', "");
Expect(1, 67679, '\P{^Scx=armi}', "");
Expect(0, 67680, '\p{Scx=armi}', "");
Expect(1, 67680, '\p{^Scx=armi}', "");
Expect(1, 67680, '\P{Scx=armi}', "");
Expect(0, 67680, '\P{^Scx=armi}', "");
Expect(1, 67679, '\p{Scx= Armi}', "");
Expect(0, 67679, '\p{^Scx= Armi}', "");
Expect(0, 67679, '\P{Scx= Armi}', "");
Expect(1, 67679, '\P{^Scx= Armi}', "");
Expect(0, 67680, '\p{Scx= Armi}', "");
Expect(1, 67680, '\p{^Scx= Armi}', "");
Expect(1, 67680, '\P{Scx= Armi}', "");
Expect(0, 67680, '\P{^Scx= Armi}', "");
Error('\p{Is_Script_Extensions= _IMPERIAL_aramaic/a/}');
Error('\P{Is_Script_Extensions= _IMPERIAL_aramaic/a/}');
Expect(1, 67679, '\p{Is_Script_Extensions=imperialaramaic}', "");
Expect(0, 67679, '\p{^Is_Script_Extensions=imperialaramaic}', "");
Expect(0, 67679, '\P{Is_Script_Extensions=imperialaramaic}', "");
Expect(1, 67679, '\P{^Is_Script_Extensions=imperialaramaic}', "");
Expect(0, 67680, '\p{Is_Script_Extensions=imperialaramaic}', "");
Expect(1, 67680, '\p{^Is_Script_Extensions=imperialaramaic}', "");
Expect(1, 67680, '\P{Is_Script_Extensions=imperialaramaic}', "");
Expect(0, 67680, '\P{^Is_Script_Extensions=imperialaramaic}', "");
Expect(1, 67679, '\p{Is_Script_Extensions=- Imperial_Aramaic}', "");
Expect(0, 67679, '\p{^Is_Script_Extensions=- Imperial_Aramaic}', "");
Expect(0, 67679, '\P{Is_Script_Extensions=- Imperial_Aramaic}', "");
Expect(1, 67679, '\P{^Is_Script_Extensions=- Imperial_Aramaic}', "");
Expect(0, 67680, '\p{Is_Script_Extensions=- Imperial_Aramaic}', "");
Expect(1, 67680, '\p{^Is_Script_Extensions=- Imperial_Aramaic}', "");
Expect(1, 67680, '\P{Is_Script_Extensions=- Imperial_Aramaic}', "");
Expect(0, 67680, '\P{^Is_Script_Extensions=- Imperial_Aramaic}', "");
Error('\p{Is_Scx=/a/_Armi}');
Error('\P{Is_Scx=/a/_Armi}');
Expect(1, 67679, '\p{Is_Scx=armi}', "");
Expect(0, 67679, '\p{^Is_Scx=armi}', "");
Expect(0, 67679, '\P{Is_Scx=armi}', "");
Expect(1, 67679, '\P{^Is_Scx=armi}', "");
Expect(0, 67680, '\p{Is_Scx=armi}', "");
Expect(1, 67680, '\p{^Is_Scx=armi}', "");
Expect(1, 67680, '\P{Is_Scx=armi}', "");
Expect(0, 67680, '\P{^Is_Scx=armi}', "");
Expect(1, 67679, '\p{Is_Scx=--armi}', "");
Expect(0, 67679, '\p{^Is_Scx=--armi}', "");
Expect(0, 67679, '\P{Is_Scx=--armi}', "");
Expect(1, 67679, '\P{^Is_Scx=--armi}', "");
Expect(0, 67680, '\p{Is_Scx=--armi}', "");
Expect(1, 67680, '\p{^Is_Scx=--armi}', "");
Expect(1, 67680, '\P{Is_Scx=--armi}', "");
Expect(0, 67680, '\P{^Is_Scx=--armi}', "");
Error('\p{Script_Extensions=:=	 kannada}');
Error('\P{Script_Extensions=:=	 kannada}');
Expect(1, 3314, '\p{Script_Extensions=kannada}', "");
Expect(0, 3314, '\p{^Script_Extensions=kannada}', "");
Expect(0, 3314, '\P{Script_Extensions=kannada}', "");
Expect(1, 3314, '\P{^Script_Extensions=kannada}', "");
Expect(0, 3315, '\p{Script_Extensions=kannada}', "");
Expect(1, 3315, '\p{^Script_Extensions=kannada}', "");
Expect(1, 3315, '\P{Script_Extensions=kannada}', "");
Expect(0, 3315, '\P{^Script_Extensions=kannada}', "");
Expect(1, 3314, '\p{Script_Extensions:-	Kannada}', "");
Expect(0, 3314, '\p{^Script_Extensions:-	Kannada}', "");
Expect(0, 3314, '\P{Script_Extensions:-	Kannada}', "");
Expect(1, 3314, '\P{^Script_Extensions:-	Kannada}', "");
Expect(0, 3315, '\p{Script_Extensions:-	Kannada}', "");
Expect(1, 3315, '\p{^Script_Extensions:-	Kannada}', "");
Expect(1, 3315, '\P{Script_Extensions:-	Kannada}', "");
Expect(0, 3315, '\P{^Script_Extensions:-	Kannada}', "");
Error('\p{Scx=/a/__Knda}');
Error('\P{Scx=/a/__Knda}');
Expect(1, 3314, '\p{Scx=knda}', "");
Expect(0, 3314, '\p{^Scx=knda}', "");
Expect(0, 3314, '\P{Scx=knda}', "");
Expect(1, 3314, '\P{^Scx=knda}', "");
Expect(0, 3315, '\p{Scx=knda}', "");
Expect(1, 3315, '\p{^Scx=knda}', "");
Expect(1, 3315, '\P{Scx=knda}', "");
Expect(0, 3315, '\P{^Scx=knda}', "");
Expect(1, 3314, '\p{Scx=	knda}', "");
Expect(0, 3314, '\p{^Scx=	knda}', "");
Expect(0, 3314, '\P{Scx=	knda}', "");
Expect(1, 3314, '\P{^Scx=	knda}', "");
Expect(0, 3315, '\p{Scx=	knda}', "");
Expect(1, 3315, '\p{^Scx=	knda}', "");
Expect(1, 3315, '\P{Scx=	knda}', "");
Expect(0, 3315, '\P{^Scx=	knda}', "");
Error('\p{Is_Script_Extensions=/a/_-Kannada}');
Error('\P{Is_Script_Extensions=/a/_-Kannada}');
Expect(1, 3314, '\p{Is_Script_Extensions=kannada}', "");
Expect(0, 3314, '\p{^Is_Script_Extensions=kannada}', "");
Expect(0, 3314, '\P{Is_Script_Extensions=kannada}', "");
Expect(1, 3314, '\P{^Is_Script_Extensions=kannada}', "");
Expect(0, 3315, '\p{Is_Script_Extensions=kannada}', "");
Expect(1, 3315, '\p{^Is_Script_Extensions=kannada}', "");
Expect(1, 3315, '\P{Is_Script_Extensions=kannada}', "");
Expect(0, 3315, '\P{^Is_Script_Extensions=kannada}', "");
Expect(1, 3314, '\p{Is_Script_Extensions=_ kannada}', "");
Expect(0, 3314, '\p{^Is_Script_Extensions=_ kannada}', "");
Expect(0, 3314, '\P{Is_Script_Extensions=_ kannada}', "");
Expect(1, 3314, '\P{^Is_Script_Extensions=_ kannada}', "");
Expect(0, 3315, '\p{Is_Script_Extensions=_ kannada}', "");
Expect(1, 3315, '\p{^Is_Script_Extensions=_ kannada}', "");
Expect(1, 3315, '\P{Is_Script_Extensions=_ kannada}', "");
Expect(0, 3315, '\P{^Is_Script_Extensions=_ kannada}', "");
Error('\p{Is_Scx:	:=KNDA}');
Error('\P{Is_Scx:	:=KNDA}');
Expect(1, 3314, '\p{Is_Scx=knda}', "");
Expect(0, 3314, '\p{^Is_Scx=knda}', "");
Expect(0, 3314, '\P{Is_Scx=knda}', "");
Expect(1, 3314, '\P{^Is_Scx=knda}', "");
Expect(0, 3315, '\p{Is_Scx=knda}', "");
Expect(1, 3315, '\p{^Is_Scx=knda}', "");
Expect(1, 3315, '\P{Is_Scx=knda}', "");
Expect(0, 3315, '\P{^Is_Scx=knda}', "");
Expect(1, 3314, '\p{Is_Scx=-KNDA}', "");
Expect(0, 3314, '\p{^Is_Scx=-KNDA}', "");
Expect(0, 3314, '\P{Is_Scx=-KNDA}', "");
Expect(1, 3314, '\P{^Is_Scx=-KNDA}', "");
Expect(0, 3315, '\p{Is_Scx=-KNDA}', "");
Expect(1, 3315, '\p{^Is_Scx=-KNDA}', "");
Expect(1, 3315, '\P{Is_Scx=-KNDA}', "");
Expect(0, 3315, '\P{^Is_Scx=-KNDA}', "");
Error('\p{Script_Extensions=-:=Khmer}');
Error('\P{Script_Extensions=-:=Khmer}');
Expect(1, 6655, '\p{Script_Extensions=khmer}', "");
Expect(0, 6655, '\p{^Script_Extensions=khmer}', "");
Expect(0, 6655, '\P{Script_Extensions=khmer}', "");
Expect(1, 6655, '\P{^Script_Extensions=khmer}', "");
Expect(0, 6656, '\p{Script_Extensions=khmer}', "");
Expect(1, 6656, '\p{^Script_Extensions=khmer}', "");
Expect(1, 6656, '\P{Script_Extensions=khmer}', "");
Expect(0, 6656, '\P{^Script_Extensions=khmer}', "");
Expect(1, 6655, '\p{Script_Extensions=_-Khmer}', "");
Expect(0, 6655, '\p{^Script_Extensions=_-Khmer}', "");
Expect(0, 6655, '\P{Script_Extensions=_-Khmer}', "");
Expect(1, 6655, '\P{^Script_Extensions=_-Khmer}', "");
Expect(0, 6656, '\p{Script_Extensions=_-Khmer}', "");
Expect(1, 6656, '\p{^Script_Extensions=_-Khmer}', "");
Expect(1, 6656, '\P{Script_Extensions=_-Khmer}', "");
Expect(0, 6656, '\P{^Script_Extensions=_-Khmer}', "");
Error('\p{Scx=/a/ Khmr}');
Error('\P{Scx=/a/ Khmr}');
Expect(1, 6655, '\p{Scx=khmr}', "");
Expect(0, 6655, '\p{^Scx=khmr}', "");
Expect(0, 6655, '\P{Scx=khmr}', "");
Expect(1, 6655, '\P{^Scx=khmr}', "");
Expect(0, 6656, '\p{Scx=khmr}', "");
Expect(1, 6656, '\p{^Scx=khmr}', "");
Expect(1, 6656, '\P{Scx=khmr}', "");
Expect(0, 6656, '\P{^Scx=khmr}', "");
Expect(1, 6655, '\p{Scx=__KHMR}', "");
Expect(0, 6655, '\p{^Scx=__KHMR}', "");
Expect(0, 6655, '\P{Scx=__KHMR}', "");
Expect(1, 6655, '\P{^Scx=__KHMR}', "");
Expect(0, 6656, '\p{Scx=__KHMR}', "");
Expect(1, 6656, '\p{^Scx=__KHMR}', "");
Expect(1, 6656, '\P{Scx=__KHMR}', "");
Expect(0, 6656, '\P{^Scx=__KHMR}', "");
Error('\p{Is_Script_Extensions=_-khmer/a/}');
Error('\P{Is_Script_Extensions=_-khmer/a/}');
Expect(1, 6655, '\p{Is_Script_Extensions=khmer}', "");
Expect(0, 6655, '\p{^Is_Script_Extensions=khmer}', "");
Expect(0, 6655, '\P{Is_Script_Extensions=khmer}', "");
Expect(1, 6655, '\P{^Is_Script_Extensions=khmer}', "");
Expect(0, 6656, '\p{Is_Script_Extensions=khmer}', "");
Expect(1, 6656, '\p{^Is_Script_Extensions=khmer}', "");
Expect(1, 6656, '\P{Is_Script_Extensions=khmer}', "");
Expect(0, 6656, '\P{^Is_Script_Extensions=khmer}', "");
Expect(1, 6655, '\p{Is_Script_Extensions=-	khmer}', "");
Expect(0, 6655, '\p{^Is_Script_Extensions=-	khmer}', "");
Expect(0, 6655, '\P{Is_Script_Extensions=-	khmer}', "");
Expect(1, 6655, '\P{^Is_Script_Extensions=-	khmer}', "");
Expect(0, 6656, '\p{Is_Script_Extensions=-	khmer}', "");
Expect(1, 6656, '\p{^Is_Script_Extensions=-	khmer}', "");
Expect(1, 6656, '\P{Is_Script_Extensions=-	khmer}', "");
Expect(0, 6656, '\P{^Is_Script_Extensions=-	khmer}', "");
Error('\p{Is_Scx=-:=Khmr}');
Error('\P{Is_Scx=-:=Khmr}');
Expect(1, 6655, '\p{Is_Scx=khmr}', "");
Expect(0, 6655, '\p{^Is_Scx=khmr}', "");
Expect(0, 6655, '\P{Is_Scx=khmr}', "");
Expect(1, 6655, '\P{^Is_Scx=khmr}', "");
Expect(0, 6656, '\p{Is_Scx=khmr}', "");
Expect(1, 6656, '\p{^Is_Scx=khmr}', "");
Expect(1, 6656, '\P{Is_Scx=khmr}', "");
Expect(0, 6656, '\P{^Is_Scx=khmr}', "");
Expect(1, 6655, '\p{Is_Scx:   KHMR}', "");
Expect(0, 6655, '\p{^Is_Scx:   KHMR}', "");
Expect(0, 6655, '\P{Is_Scx:   KHMR}', "");
Expect(1, 6655, '\P{^Is_Scx:   KHMR}', "");
Expect(0, 6656, '\p{Is_Scx:   KHMR}', "");
Expect(1, 6656, '\p{^Is_Scx:   KHMR}', "");
Expect(1, 6656, '\P{Is_Scx:   KHMR}', "");
Expect(0, 6656, '\P{^Is_Scx:   KHMR}', "");
Error('\p{Script_Extensions=/a/_	CHAM}');
Error('\P{Script_Extensions=/a/_	CHAM}');
Expect(1, 43615, '\p{Script_Extensions=cham}', "");
Expect(0, 43615, '\p{^Script_Extensions=cham}', "");
Expect(0, 43615, '\P{Script_Extensions=cham}', "");
Expect(1, 43615, '\P{^Script_Extensions=cham}', "");
Expect(0, 43616, '\p{Script_Extensions=cham}', "");
Expect(1, 43616, '\p{^Script_Extensions=cham}', "");
Expect(1, 43616, '\P{Script_Extensions=cham}', "");
Expect(0, 43616, '\P{^Script_Extensions=cham}', "");
Expect(1, 43615, '\p{Script_Extensions:   - Cham}', "");
Expect(0, 43615, '\p{^Script_Extensions:   - Cham}', "");
Expect(0, 43615, '\P{Script_Extensions:   - Cham}', "");
Expect(1, 43615, '\P{^Script_Extensions:   - Cham}', "");
Expect(0, 43616, '\p{Script_Extensions:   - Cham}', "");
Expect(1, 43616, '\p{^Script_Extensions:   - Cham}', "");
Expect(1, 43616, '\P{Script_Extensions:   - Cham}', "");
Expect(0, 43616, '\P{^Script_Extensions:   - Cham}', "");
Error('\p{Scx=--CHAM:=}');
Error('\P{Scx=--CHAM:=}');
Expect(1, 43615, '\p{Scx=cham}', "");
Expect(0, 43615, '\p{^Scx=cham}', "");
Expect(0, 43615, '\P{Scx=cham}', "");
Expect(1, 43615, '\P{^Scx=cham}', "");
Expect(0, 43616, '\p{Scx=cham}', "");
Expect(1, 43616, '\p{^Scx=cham}', "");
Expect(1, 43616, '\P{Scx=cham}', "");
Expect(0, 43616, '\P{^Scx=cham}', "");
Expect(1, 43615, '\p{Scx=_	Cham}', "");
Expect(0, 43615, '\p{^Scx=_	Cham}', "");
Expect(0, 43615, '\P{Scx=_	Cham}', "");
Expect(1, 43615, '\P{^Scx=_	Cham}', "");
Expect(0, 43616, '\p{Scx=_	Cham}', "");
Expect(1, 43616, '\p{^Scx=_	Cham}', "");
Expect(1, 43616, '\P{Scx=_	Cham}', "");
Expect(0, 43616, '\P{^Scx=_	Cham}', "");
Error('\p{Is_Script_Extensions=-:=cham}');
Error('\P{Is_Script_Extensions=-:=cham}');
Expect(1, 43615, '\p{Is_Script_Extensions=cham}', "");
Expect(0, 43615, '\p{^Is_Script_Extensions=cham}', "");
Expect(0, 43615, '\P{Is_Script_Extensions=cham}', "");
Expect(1, 43615, '\P{^Is_Script_Extensions=cham}', "");
Expect(0, 43616, '\p{Is_Script_Extensions=cham}', "");
Expect(1, 43616, '\p{^Is_Script_Extensions=cham}', "");
Expect(1, 43616, '\P{Is_Script_Extensions=cham}', "");
Expect(0, 43616, '\P{^Is_Script_Extensions=cham}', "");
Expect(1, 43615, '\p{Is_Script_Extensions=_Cham}', "");
Expect(0, 43615, '\p{^Is_Script_Extensions=_Cham}', "");
Expect(0, 43615, '\P{Is_Script_Extensions=_Cham}', "");
Expect(1, 43615, '\P{^Is_Script_Extensions=_Cham}', "");
Expect(0, 43616, '\p{Is_Script_Extensions=_Cham}', "");
Expect(1, 43616, '\p{^Is_Script_Extensions=_Cham}', "");
Expect(1, 43616, '\P{Is_Script_Extensions=_Cham}', "");
Expect(0, 43616, '\P{^Is_Script_Extensions=_Cham}', "");
Error('\p{Is_Scx=/a/	-CHAM}');
Error('\P{Is_Scx=/a/	-CHAM}');
Expect(1, 43615, '\p{Is_Scx=cham}', "");
Expect(0, 43615, '\p{^Is_Scx=cham}', "");
Expect(0, 43615, '\P{Is_Scx=cham}', "");
Expect(1, 43615, '\P{^Is_Scx=cham}', "");
Expect(0, 43616, '\p{Is_Scx=cham}', "");
Expect(1, 43616, '\p{^Is_Scx=cham}', "");
Expect(1, 43616, '\P{Is_Scx=cham}', "");
Expect(0, 43616, '\P{^Is_Scx=cham}', "");
Expect(1, 43615, '\p{Is_Scx=-_Cham}', "");
Expect(0, 43615, '\p{^Is_Scx=-_Cham}', "");
Expect(0, 43615, '\P{Is_Scx=-_Cham}', "");
Expect(1, 43615, '\P{^Is_Scx=-_Cham}', "");
Expect(0, 43616, '\p{Is_Scx=-_Cham}', "");
Expect(1, 43616, '\p{^Is_Scx=-_Cham}', "");
Expect(1, 43616, '\P{Is_Scx=-_Cham}', "");
Expect(0, 43616, '\P{^Is_Scx=-_Cham}', "");
Error('\p{Script_Extensions=tibetan/a/}');
Error('\P{Script_Extensions=tibetan/a/}');
Expect(1, 4058, '\p{Script_Extensions=tibetan}', "");
Expect(0, 4058, '\p{^Script_Extensions=tibetan}', "");
Expect(0, 4058, '\P{Script_Extensions=tibetan}', "");
Expect(1, 4058, '\P{^Script_Extensions=tibetan}', "");
Expect(0, 4059, '\p{Script_Extensions=tibetan}', "");
Expect(1, 4059, '\p{^Script_Extensions=tibetan}', "");
Expect(1, 4059, '\P{Script_Extensions=tibetan}', "");
Expect(0, 4059, '\P{^Script_Extensions=tibetan}', "");
Expect(1, 4058, '\p{Script_Extensions:	TIBETAN}', "");
Expect(0, 4058, '\p{^Script_Extensions:	TIBETAN}', "");
Expect(0, 4058, '\P{Script_Extensions:	TIBETAN}', "");
Expect(1, 4058, '\P{^Script_Extensions:	TIBETAN}', "");
Expect(0, 4059, '\p{Script_Extensions:	TIBETAN}', "");
Expect(1, 4059, '\p{^Script_Extensions:	TIBETAN}', "");
Expect(1, 4059, '\P{Script_Extensions:	TIBETAN}', "");
Expect(0, 4059, '\P{^Script_Extensions:	TIBETAN}', "");
Error('\p{Scx=:=	_Tibt}');
Error('\P{Scx=:=	_Tibt}');
Expect(1, 4058, '\p{Scx=tibt}', "");
Expect(0, 4058, '\p{^Scx=tibt}', "");
Expect(0, 4058, '\P{Scx=tibt}', "");
Expect(1, 4058, '\P{^Scx=tibt}', "");
Expect(0, 4059, '\p{Scx=tibt}', "");
Expect(1, 4059, '\p{^Scx=tibt}', "");
Expect(1, 4059, '\P{Scx=tibt}', "");
Expect(0, 4059, '\P{^Scx=tibt}', "");
Expect(1, 4058, '\p{Scx= Tibt}', "");
Expect(0, 4058, '\p{^Scx= Tibt}', "");
Expect(0, 4058, '\P{Scx= Tibt}', "");
Expect(1, 4058, '\P{^Scx= Tibt}', "");
Expect(0, 4059, '\p{Scx= Tibt}', "");
Expect(1, 4059, '\p{^Scx= Tibt}', "");
Expect(1, 4059, '\P{Scx= Tibt}', "");
Expect(0, 4059, '\P{^Scx= Tibt}', "");
Error('\p{Is_Script_Extensions=_/a/Tibetan}');
Error('\P{Is_Script_Extensions=_/a/Tibetan}');
Expect(1, 4058, '\p{Is_Script_Extensions=tibetan}', "");
Expect(0, 4058, '\p{^Is_Script_Extensions=tibetan}', "");
Expect(0, 4058, '\P{Is_Script_Extensions=tibetan}', "");
Expect(1, 4058, '\P{^Is_Script_Extensions=tibetan}', "");
Expect(0, 4059, '\p{Is_Script_Extensions=tibetan}', "");
Expect(1, 4059, '\p{^Is_Script_Extensions=tibetan}', "");
Expect(1, 4059, '\P{Is_Script_Extensions=tibetan}', "");
Expect(0, 4059, '\P{^Is_Script_Extensions=tibetan}', "");
Expect(1, 4058, '\p{Is_Script_Extensions=	 Tibetan}', "");
Expect(0, 4058, '\p{^Is_Script_Extensions=	 Tibetan}', "");
Expect(0, 4058, '\P{Is_Script_Extensions=	 Tibetan}', "");
Expect(1, 4058, '\P{^Is_Script_Extensions=	 Tibetan}', "");
Expect(0, 4059, '\p{Is_Script_Extensions=	 Tibetan}', "");
Expect(1, 4059, '\p{^Is_Script_Extensions=	 Tibetan}', "");
Expect(1, 4059, '\P{Is_Script_Extensions=	 Tibetan}', "");
Expect(0, 4059, '\P{^Is_Script_Extensions=	 Tibetan}', "");
Error('\p{Is_Scx:   :=Tibt}');
Error('\P{Is_Scx:   :=Tibt}');
Expect(1, 4058, '\p{Is_Scx=tibt}', "");
Expect(0, 4058, '\p{^Is_Scx=tibt}', "");
Expect(0, 4058, '\P{Is_Scx=tibt}', "");
Expect(1, 4058, '\P{^Is_Scx=tibt}', "");
Expect(0, 4059, '\p{Is_Scx=tibt}', "");
Expect(1, 4059, '\p{^Is_Scx=tibt}', "");
Expect(1, 4059, '\P{Is_Scx=tibt}', "");
Expect(0, 4059, '\P{^Is_Scx=tibt}', "");
Expect(1, 4058, '\p{Is_Scx= TIBT}', "");
Expect(0, 4058, '\p{^Is_Scx= TIBT}', "");
Expect(0, 4058, '\P{Is_Scx= TIBT}', "");
Expect(1, 4058, '\P{^Is_Scx= TIBT}', "");
Expect(0, 4059, '\p{Is_Scx= TIBT}', "");
Expect(1, 4059, '\p{^Is_Scx= TIBT}', "");
Expect(1, 4059, '\P{Is_Scx= TIBT}', "");
Expect(0, 4059, '\P{^Is_Scx= TIBT}', "");
Error('\p{Script_Extensions=:=_	ogham}');
Error('\P{Script_Extensions=:=_	ogham}');
Expect(1, 5788, '\p{Script_Extensions=ogham}', "");
Expect(0, 5788, '\p{^Script_Extensions=ogham}', "");
Expect(0, 5788, '\P{Script_Extensions=ogham}', "");
Expect(1, 5788, '\P{^Script_Extensions=ogham}', "");
Expect(0, 5789, '\p{Script_Extensions=ogham}', "");
Expect(1, 5789, '\p{^Script_Extensions=ogham}', "");
Expect(1, 5789, '\P{Script_Extensions=ogham}', "");
Expect(0, 5789, '\P{^Script_Extensions=ogham}', "");
Expect(1, 5788, '\p{Script_Extensions=	-Ogham}', "");
Expect(0, 5788, '\p{^Script_Extensions=	-Ogham}', "");
Expect(0, 5788, '\P{Script_Extensions=	-Ogham}', "");
Expect(1, 5788, '\P{^Script_Extensions=	-Ogham}', "");
Expect(0, 5789, '\p{Script_Extensions=	-Ogham}', "");
Expect(1, 5789, '\p{^Script_Extensions=	-Ogham}', "");
Expect(1, 5789, '\P{Script_Extensions=	-Ogham}', "");
Expect(0, 5789, '\P{^Script_Extensions=	-Ogham}', "");
Error('\p{Scx= _OGAM/a/}');
Error('\P{Scx= _OGAM/a/}');
Expect(1, 5788, '\p{Scx=ogam}', "");
Expect(0, 5788, '\p{^Scx=ogam}', "");
Expect(0, 5788, '\P{Scx=ogam}', "");
Expect(1, 5788, '\P{^Scx=ogam}', "");
Expect(0, 5789, '\p{Scx=ogam}', "");
Expect(1, 5789, '\p{^Scx=ogam}', "");
Expect(1, 5789, '\P{Scx=ogam}', "");
Expect(0, 5789, '\P{^Scx=ogam}', "");
Expect(1, 5788, '\p{Scx=		ogam}', "");
Expect(0, 5788, '\p{^Scx=		ogam}', "");
Expect(0, 5788, '\P{Scx=		ogam}', "");
Expect(1, 5788, '\P{^Scx=		ogam}', "");
Expect(0, 5789, '\p{Scx=		ogam}', "");
Expect(1, 5789, '\p{^Scx=		ogam}', "");
Expect(1, 5789, '\P{Scx=		ogam}', "");
Expect(0, 5789, '\P{^Scx=		ogam}', "");
Error('\p{Is_Script_Extensions= /a/Ogham}');
Error('\P{Is_Script_Extensions= /a/Ogham}');
Expect(1, 5788, '\p{Is_Script_Extensions=ogham}', "");
Expect(0, 5788, '\p{^Is_Script_Extensions=ogham}', "");
Expect(0, 5788, '\P{Is_Script_Extensions=ogham}', "");
Expect(1, 5788, '\P{^Is_Script_Extensions=ogham}', "");
Expect(0, 5789, '\p{Is_Script_Extensions=ogham}', "");
Expect(1, 5789, '\p{^Is_Script_Extensions=ogham}', "");
Expect(1, 5789, '\P{Is_Script_Extensions=ogham}', "");
Expect(0, 5789, '\P{^Is_Script_Extensions=ogham}', "");
Expect(1, 5788, '\p{Is_Script_Extensions=_	Ogham}', "");
Expect(0, 5788, '\p{^Is_Script_Extensions=_	Ogham}', "");
Expect(0, 5788, '\P{Is_Script_Extensions=_	Ogham}', "");
Expect(1, 5788, '\P{^Is_Script_Extensions=_	Ogham}', "");
Expect(0, 5789, '\p{Is_Script_Extensions=_	Ogham}', "");
Expect(1, 5789, '\p{^Is_Script_Extensions=_	Ogham}', "");
Expect(1, 5789, '\P{Is_Script_Extensions=_	Ogham}', "");
Expect(0, 5789, '\P{^Is_Script_Extensions=_	Ogham}', "");
Error('\p{Is_Scx=:=OGAM}');
Error('\P{Is_Scx=:=OGAM}');
Expect(1, 5788, '\p{Is_Scx=ogam}', "");
Expect(0, 5788, '\p{^Is_Scx=ogam}', "");
Expect(0, 5788, '\P{Is_Scx=ogam}', "");
Expect(1, 5788, '\P{^Is_Scx=ogam}', "");
Expect(0, 5789, '\p{Is_Scx=ogam}', "");
Expect(1, 5789, '\p{^Is_Scx=ogam}', "");
Expect(1, 5789, '\P{Is_Scx=ogam}', "");
Expect(0, 5789, '\P{^Is_Scx=ogam}', "");
Expect(1, 5788, '\p{Is_Scx=_	Ogam}', "");
Expect(0, 5788, '\p{^Is_Scx=_	Ogam}', "");
Expect(0, 5788, '\P{Is_Scx=_	Ogam}', "");
Expect(1, 5788, '\P{^Is_Scx=_	Ogam}', "");
Expect(0, 5789, '\p{Is_Scx=_	Ogam}', "");
Expect(1, 5789, '\p{^Is_Scx=_	Ogam}', "");
Expect(1, 5789, '\P{Is_Scx=_	Ogam}', "");
Expect(0, 5789, '\P{^Is_Scx=_	Ogam}', "");
Error('\p{Script_Extensions= /a/Tai_Viet}');
Error('\P{Script_Extensions= /a/Tai_Viet}');
Expect(1, 43743, '\p{Script_Extensions=taiviet}', "");
Expect(0, 43743, '\p{^Script_Extensions=taiviet}', "");
Expect(0, 43743, '\P{Script_Extensions=taiviet}', "");
Expect(1, 43743, '\P{^Script_Extensions=taiviet}', "");
Expect(0, 43744, '\p{Script_Extensions=taiviet}', "");
Expect(1, 43744, '\p{^Script_Extensions=taiviet}', "");
Expect(1, 43744, '\P{Script_Extensions=taiviet}', "");
Expect(0, 43744, '\P{^Script_Extensions=taiviet}', "");
Expect(1, 43743, '\p{Script_Extensions=_Tai_Viet}', "");
Expect(0, 43743, '\p{^Script_Extensions=_Tai_Viet}', "");
Expect(0, 43743, '\P{Script_Extensions=_Tai_Viet}', "");
Expect(1, 43743, '\P{^Script_Extensions=_Tai_Viet}', "");
Expect(0, 43744, '\p{Script_Extensions=_Tai_Viet}', "");
Expect(1, 43744, '\p{^Script_Extensions=_Tai_Viet}', "");
Expect(1, 43744, '\P{Script_Extensions=_Tai_Viet}', "");
Expect(0, 43744, '\P{^Script_Extensions=_Tai_Viet}', "");
Error('\p{Scx=/a/Tavt}');
Error('\P{Scx=/a/Tavt}');
Expect(1, 43743, '\p{Scx:tavt}', "");
Expect(0, 43743, '\p{^Scx:tavt}', "");
Expect(0, 43743, '\P{Scx:tavt}', "");
Expect(1, 43743, '\P{^Scx:tavt}', "");
Expect(0, 43744, '\p{Scx:tavt}', "");
Expect(1, 43744, '\p{^Scx:tavt}', "");
Expect(1, 43744, '\P{Scx:tavt}', "");
Expect(0, 43744, '\P{^Scx:tavt}', "");
Expect(1, 43743, '\p{Scx=_TAVT}', "");
Expect(0, 43743, '\p{^Scx=_TAVT}', "");
Expect(0, 43743, '\P{Scx=_TAVT}', "");
Expect(1, 43743, '\P{^Scx=_TAVT}', "");
Expect(0, 43744, '\p{Scx=_TAVT}', "");
Expect(1, 43744, '\p{^Scx=_TAVT}', "");
Expect(1, 43744, '\P{Scx=_TAVT}', "");
Expect(0, 43744, '\P{^Scx=_TAVT}', "");
Error('\p{Is_Script_Extensions=:=tai_VIET}');
Error('\P{Is_Script_Extensions=:=tai_VIET}');
Expect(1, 43743, '\p{Is_Script_Extensions=taiviet}', "");
Expect(0, 43743, '\p{^Is_Script_Extensions=taiviet}', "");
Expect(0, 43743, '\P{Is_Script_Extensions=taiviet}', "");
Expect(1, 43743, '\P{^Is_Script_Extensions=taiviet}', "");
Expect(0, 43744, '\p{Is_Script_Extensions=taiviet}', "");
Expect(1, 43744, '\p{^Is_Script_Extensions=taiviet}', "");
Expect(1, 43744, '\P{Is_Script_Extensions=taiviet}', "");
Expect(0, 43744, '\P{^Is_Script_Extensions=taiviet}', "");
Expect(1, 43743, '\p{Is_Script_Extensions=- Tai_Viet}', "");
Expect(0, 43743, '\p{^Is_Script_Extensions=- Tai_Viet}', "");
Expect(0, 43743, '\P{Is_Script_Extensions=- Tai_Viet}', "");
Expect(1, 43743, '\P{^Is_Script_Extensions=- Tai_Viet}', "");
Expect(0, 43744, '\p{Is_Script_Extensions=- Tai_Viet}', "");
Expect(1, 43744, '\p{^Is_Script_Extensions=- Tai_Viet}', "");
Expect(1, 43744, '\P{Is_Script_Extensions=- Tai_Viet}', "");
Expect(0, 43744, '\P{^Is_Script_Extensions=- Tai_Viet}', "");
Error('\p{Is_Scx=/a/-Tavt}');
Error('\P{Is_Scx=/a/-Tavt}');
Expect(1, 43743, '\p{Is_Scx=tavt}', "");
Expect(0, 43743, '\p{^Is_Scx=tavt}', "");
Expect(0, 43743, '\P{Is_Scx=tavt}', "");
Expect(1, 43743, '\P{^Is_Scx=tavt}', "");
Expect(0, 43744, '\p{Is_Scx=tavt}', "");
Expect(1, 43744, '\p{^Is_Scx=tavt}', "");
Expect(1, 43744, '\P{Is_Scx=tavt}', "");
Expect(0, 43744, '\P{^Is_Scx=tavt}', "");
Expect(1, 43743, '\p{Is_Scx:	_ tavt}', "");
Expect(0, 43743, '\p{^Is_Scx:	_ tavt}', "");
Expect(0, 43743, '\P{Is_Scx:	_ tavt}', "");
Expect(1, 43743, '\P{^Is_Scx:	_ tavt}', "");
Expect(0, 43744, '\p{Is_Scx:	_ tavt}', "");
Expect(1, 43744, '\p{^Is_Scx:	_ tavt}', "");
Expect(1, 43744, '\P{Is_Scx:	_ tavt}', "");
Expect(0, 43744, '\P{^Is_Scx:	_ tavt}', "");
Error('\p{Script_Extensions: :=MALAYALAM}');
Error('\P{Script_Extensions: :=MALAYALAM}');
Expect(1, 3455, '\p{Script_Extensions=malayalam}', "");
Expect(0, 3455, '\p{^Script_Extensions=malayalam}', "");
Expect(0, 3455, '\P{Script_Extensions=malayalam}', "");
Expect(1, 3455, '\P{^Script_Extensions=malayalam}', "");
Expect(0, 3456, '\p{Script_Extensions=malayalam}', "");
Expect(1, 3456, '\p{^Script_Extensions=malayalam}', "");
Expect(1, 3456, '\P{Script_Extensions=malayalam}', "");
Expect(0, 3456, '\P{^Script_Extensions=malayalam}', "");
Expect(1, 3455, '\p{Script_Extensions=_	MALAYALAM}', "");
Expect(0, 3455, '\p{^Script_Extensions=_	MALAYALAM}', "");
Expect(0, 3455, '\P{Script_Extensions=_	MALAYALAM}', "");
Expect(1, 3455, '\P{^Script_Extensions=_	MALAYALAM}', "");
Expect(0, 3456, '\p{Script_Extensions=_	MALAYALAM}', "");
Expect(1, 3456, '\p{^Script_Extensions=_	MALAYALAM}', "");
Expect(1, 3456, '\P{Script_Extensions=_	MALAYALAM}', "");
Expect(0, 3456, '\P{^Script_Extensions=_	MALAYALAM}', "");
Error('\p{Scx=_:=MLYM}');
Error('\P{Scx=_:=MLYM}');
Expect(1, 3455, '\p{Scx=mlym}', "");
Expect(0, 3455, '\p{^Scx=mlym}', "");
Expect(0, 3455, '\P{Scx=mlym}', "");
Expect(1, 3455, '\P{^Scx=mlym}', "");
Expect(0, 3456, '\p{Scx=mlym}', "");
Expect(1, 3456, '\p{^Scx=mlym}', "");
Expect(1, 3456, '\P{Scx=mlym}', "");
Expect(0, 3456, '\P{^Scx=mlym}', "");
Expect(1, 3455, '\p{Scx=-Mlym}', "");
Expect(0, 3455, '\p{^Scx=-Mlym}', "");
Expect(0, 3455, '\P{Scx=-Mlym}', "");
Expect(1, 3455, '\P{^Scx=-Mlym}', "");
Expect(0, 3456, '\p{Scx=-Mlym}', "");
Expect(1, 3456, '\p{^Scx=-Mlym}', "");
Expect(1, 3456, '\P{Scx=-Mlym}', "");
Expect(0, 3456, '\P{^Scx=-Mlym}', "");
Error('\p{Is_Script_Extensions=-/a/Malayalam}');
Error('\P{Is_Script_Extensions=-/a/Malayalam}');
Expect(1, 3455, '\p{Is_Script_Extensions=malayalam}', "");
Expect(0, 3455, '\p{^Is_Script_Extensions=malayalam}', "");
Expect(0, 3455, '\P{Is_Script_Extensions=malayalam}', "");
Expect(1, 3455, '\P{^Is_Script_Extensions=malayalam}', "");
Expect(0, 3456, '\p{Is_Script_Extensions=malayalam}', "");
Expect(1, 3456, '\p{^Is_Script_Extensions=malayalam}', "");
Expect(1, 3456, '\P{Is_Script_Extensions=malayalam}', "");
Expect(0, 3456, '\P{^Is_Script_Extensions=malayalam}', "");
Expect(1, 3455, '\p{Is_Script_Extensions=-	Malayalam}', "");
Expect(0, 3455, '\p{^Is_Script_Extensions=-	Malayalam}', "");
Expect(0, 3455, '\P{Is_Script_Extensions=-	Malayalam}', "");
Expect(1, 3455, '\P{^Is_Script_Extensions=-	Malayalam}', "");
Expect(0, 3456, '\p{Is_Script_Extensions=-	Malayalam}', "");
Expect(1, 3456, '\p{^Is_Script_Extensions=-	Malayalam}', "");
Expect(1, 3456, '\P{Is_Script_Extensions=-	Malayalam}', "");
Expect(0, 3456, '\P{^Is_Script_Extensions=-	Malayalam}', "");
Error('\p{Is_Scx=-:=MLYM}');
Error('\P{Is_Scx=-:=MLYM}');
Expect(1, 3455, '\p{Is_Scx=mlym}', "");
Expect(0, 3455, '\p{^Is_Scx=mlym}', "");
Expect(0, 3455, '\P{Is_Scx=mlym}', "");
Expect(1, 3455, '\P{^Is_Scx=mlym}', "");
Expect(0, 3456, '\p{Is_Scx=mlym}', "");
Expect(1, 3456, '\p{^Is_Scx=mlym}', "");
Expect(1, 3456, '\P{Is_Scx=mlym}', "");
Expect(0, 3456, '\P{^Is_Scx=mlym}', "");
Expect(1, 3455, '\p{Is_Scx=_-MLYM}', "");
Expect(0, 3455, '\p{^Is_Scx=_-MLYM}', "");
Expect(0, 3455, '\P{Is_Scx=_-MLYM}', "");
Expect(1, 3455, '\P{^Is_Scx=_-MLYM}', "");
Expect(0, 3456, '\p{Is_Scx=_-MLYM}', "");
Expect(1, 3456, '\p{^Is_Scx=_-MLYM}', "");
Expect(1, 3456, '\P{Is_Scx=_-MLYM}', "");
Expect(0, 3456, '\P{^Is_Scx=_-MLYM}', "");
Error('\p{Script_Extensions= :=OLD_persian}');
Error('\P{Script_Extensions= :=OLD_persian}');
Expect(1, 66517, '\p{Script_Extensions=oldpersian}', "");
Expect(0, 66517, '\p{^Script_Extensions=oldpersian}', "");
Expect(0, 66517, '\P{Script_Extensions=oldpersian}', "");
Expect(1, 66517, '\P{^Script_Extensions=oldpersian}', "");
Expect(0, 66518, '\p{Script_Extensions=oldpersian}', "");
Expect(1, 66518, '\p{^Script_Extensions=oldpersian}', "");
Expect(1, 66518, '\P{Script_Extensions=oldpersian}', "");
Expect(0, 66518, '\P{^Script_Extensions=oldpersian}', "");
Expect(1, 66517, '\p{Script_Extensions=__old_Persian}', "");
Expect(0, 66517, '\p{^Script_Extensions=__old_Persian}', "");
Expect(0, 66517, '\P{Script_Extensions=__old_Persian}', "");
Expect(1, 66517, '\P{^Script_Extensions=__old_Persian}', "");
Expect(0, 66518, '\p{Script_Extensions=__old_Persian}', "");
Expect(1, 66518, '\p{^Script_Extensions=__old_Persian}', "");
Expect(1, 66518, '\P{Script_Extensions=__old_Persian}', "");
Expect(0, 66518, '\P{^Script_Extensions=__old_Persian}', "");
Error('\p{Scx= 	xpeo/a/}');
Error('\P{Scx= 	xpeo/a/}');
Expect(1, 66517, '\p{Scx=xpeo}', "");
Expect(0, 66517, '\p{^Scx=xpeo}', "");
Expect(0, 66517, '\P{Scx=xpeo}', "");
Expect(1, 66517, '\P{^Scx=xpeo}', "");
Expect(0, 66518, '\p{Scx=xpeo}', "");
Expect(1, 66518, '\p{^Scx=xpeo}', "");
Expect(1, 66518, '\P{Scx=xpeo}', "");
Expect(0, 66518, '\P{^Scx=xpeo}', "");
Expect(1, 66517, '\p{Scx=_	XPEO}', "");
Expect(0, 66517, '\p{^Scx=_	XPEO}', "");
Expect(0, 66517, '\P{Scx=_	XPEO}', "");
Expect(1, 66517, '\P{^Scx=_	XPEO}', "");
Expect(0, 66518, '\p{Scx=_	XPEO}', "");
Expect(1, 66518, '\p{^Scx=_	XPEO}', "");
Expect(1, 66518, '\P{Scx=_	XPEO}', "");
Expect(0, 66518, '\P{^Scx=_	XPEO}', "");
Error('\p{Is_Script_Extensions=:=Old_Persian}');
Error('\P{Is_Script_Extensions=:=Old_Persian}');
Expect(1, 66517, '\p{Is_Script_Extensions=oldpersian}', "");
Expect(0, 66517, '\p{^Is_Script_Extensions=oldpersian}', "");
Expect(0, 66517, '\P{Is_Script_Extensions=oldpersian}', "");
Expect(1, 66517, '\P{^Is_Script_Extensions=oldpersian}', "");
Expect(0, 66518, '\p{Is_Script_Extensions=oldpersian}', "");
Expect(1, 66518, '\p{^Is_Script_Extensions=oldpersian}', "");
Expect(1, 66518, '\P{Is_Script_Extensions=oldpersian}', "");
Expect(0, 66518, '\P{^Is_Script_Extensions=oldpersian}', "");
Expect(1, 66517, '\p{Is_Script_Extensions:   -	old_Persian}', "");
Expect(0, 66517, '\p{^Is_Script_Extensions:   -	old_Persian}', "");
Expect(0, 66517, '\P{Is_Script_Extensions:   -	old_Persian}', "");
Expect(1, 66517, '\P{^Is_Script_Extensions:   -	old_Persian}', "");
Expect(0, 66518, '\p{Is_Script_Extensions:   -	old_Persian}', "");
Expect(1, 66518, '\p{^Is_Script_Extensions:   -	old_Persian}', "");
Expect(1, 66518, '\P{Is_Script_Extensions:   -	old_Persian}', "");
Expect(0, 66518, '\P{^Is_Script_Extensions:   -	old_Persian}', "");
Error('\p{Is_Scx=/a/-_Xpeo}');
Error('\P{Is_Scx=/a/-_Xpeo}');
Expect(1, 66517, '\p{Is_Scx=xpeo}', "");
Expect(0, 66517, '\p{^Is_Scx=xpeo}', "");
Expect(0, 66517, '\P{Is_Scx=xpeo}', "");
Expect(1, 66517, '\P{^Is_Scx=xpeo}', "");
Expect(0, 66518, '\p{Is_Scx=xpeo}', "");
Expect(1, 66518, '\p{^Is_Scx=xpeo}', "");
Expect(1, 66518, '\P{Is_Scx=xpeo}', "");
Expect(0, 66518, '\P{^Is_Scx=xpeo}', "");
Expect(1, 66517, '\p{Is_Scx= XPEO}', "");
Expect(0, 66517, '\p{^Is_Scx= XPEO}', "");
Expect(0, 66517, '\P{Is_Scx= XPEO}', "");
Expect(1, 66517, '\P{^Is_Scx= XPEO}', "");
Expect(0, 66518, '\p{Is_Scx= XPEO}', "");
Expect(1, 66518, '\p{^Is_Scx= XPEO}', "");
Expect(1, 66518, '\P{Is_Scx= XPEO}', "");
Expect(0, 66518, '\P{^Is_Scx= XPEO}', "");
Error('\p{Script_Extensions=:=	 myanmar}');
Error('\P{Script_Extensions=:=	 myanmar}');
Expect(1, 43643, '\p{Script_Extensions=myanmar}', "");
Expect(0, 43643, '\p{^Script_Extensions=myanmar}', "");
Expect(0, 43643, '\P{Script_Extensions=myanmar}', "");
Expect(1, 43643, '\P{^Script_Extensions=myanmar}', "");
Expect(0, 43644, '\p{Script_Extensions=myanmar}', "");
Expect(1, 43644, '\p{^Script_Extensions=myanmar}', "");
Expect(1, 43644, '\P{Script_Extensions=myanmar}', "");
Expect(0, 43644, '\P{^Script_Extensions=myanmar}', "");
Expect(1, 43643, '\p{Script_Extensions=  Myanmar}', "");
Expect(0, 43643, '\p{^Script_Extensions=  Myanmar}', "");
Expect(0, 43643, '\P{Script_Extensions=  Myanmar}', "");
Expect(1, 43643, '\P{^Script_Extensions=  Myanmar}', "");
Expect(0, 43644, '\p{Script_Extensions=  Myanmar}', "");
Expect(1, 43644, '\p{^Script_Extensions=  Myanmar}', "");
Expect(1, 43644, '\P{Script_Extensions=  Myanmar}', "");
Expect(0, 43644, '\P{^Script_Extensions=  Myanmar}', "");
Error('\p{Scx:	- mymr/a/}');
Error('\P{Scx:	- mymr/a/}');
Expect(1, 43643, '\p{Scx=mymr}', "");
Expect(0, 43643, '\p{^Scx=mymr}', "");
Expect(0, 43643, '\P{Scx=mymr}', "");
Expect(1, 43643, '\P{^Scx=mymr}', "");
Expect(0, 43644, '\p{Scx=mymr}', "");
Expect(1, 43644, '\p{^Scx=mymr}', "");
Expect(1, 43644, '\P{Scx=mymr}', "");
Expect(0, 43644, '\P{^Scx=mymr}', "");
Expect(1, 43643, '\p{Scx:   	Mymr}', "");
Expect(0, 43643, '\p{^Scx:   	Mymr}', "");
Expect(0, 43643, '\P{Scx:   	Mymr}', "");
Expect(1, 43643, '\P{^Scx:   	Mymr}', "");
Expect(0, 43644, '\p{Scx:   	Mymr}', "");
Expect(1, 43644, '\p{^Scx:   	Mymr}', "");
Expect(1, 43644, '\P{Scx:   	Mymr}', "");
Expect(0, 43644, '\P{^Scx:   	Mymr}', "");
Error('\p{Is_Script_Extensions=  MYANMAR:=}');
Error('\P{Is_Script_Extensions=  MYANMAR:=}');
Expect(1, 43643, '\p{Is_Script_Extensions=myanmar}', "");
Expect(0, 43643, '\p{^Is_Script_Extensions=myanmar}', "");
Expect(0, 43643, '\P{Is_Script_Extensions=myanmar}', "");
Expect(1, 43643, '\P{^Is_Script_Extensions=myanmar}', "");
Expect(0, 43644, '\p{Is_Script_Extensions=myanmar}', "");
Expect(1, 43644, '\p{^Is_Script_Extensions=myanmar}', "");
Expect(1, 43644, '\P{Is_Script_Extensions=myanmar}', "");
Expect(0, 43644, '\P{^Is_Script_Extensions=myanmar}', "");
Expect(1, 43643, '\p{Is_Script_Extensions=-_Myanmar}', "");
Expect(0, 43643, '\p{^Is_Script_Extensions=-_Myanmar}', "");
Expect(0, 43643, '\P{Is_Script_Extensions=-_Myanmar}', "");
Expect(1, 43643, '\P{^Is_Script_Extensions=-_Myanmar}', "");
Expect(0, 43644, '\p{Is_Script_Extensions=-_Myanmar}', "");
Expect(1, 43644, '\p{^Is_Script_Extensions=-_Myanmar}', "");
Expect(1, 43644, '\P{Is_Script_Extensions=-_Myanmar}', "");
Expect(0, 43644, '\P{^Is_Script_Extensions=-_Myanmar}', "");
Error('\p{Is_Scx=-_Mymr/a/}');
Error('\P{Is_Scx=-_Mymr/a/}');
Expect(1, 43643, '\p{Is_Scx=mymr}', "");
Expect(0, 43643, '\p{^Is_Scx=mymr}', "");
Expect(0, 43643, '\P{Is_Scx=mymr}', "");
Expect(1, 43643, '\P{^Is_Scx=mymr}', "");
Expect(0, 43644, '\p{Is_Scx=mymr}', "");
Expect(1, 43644, '\p{^Is_Scx=mymr}', "");
Expect(1, 43644, '\P{Is_Scx=mymr}', "");
Expect(0, 43644, '\P{^Is_Scx=mymr}', "");
Expect(1, 43643, '\p{Is_Scx=  Mymr}', "");
Expect(0, 43643, '\p{^Is_Scx=  Mymr}', "");
Expect(0, 43643, '\P{Is_Scx=  Mymr}', "");
Expect(1, 43643, '\P{^Is_Scx=  Mymr}', "");
Expect(0, 43644, '\p{Is_Scx=  Mymr}', "");
Expect(1, 43644, '\p{^Is_Scx=  Mymr}', "");
Expect(1, 43644, '\P{Is_Scx=  Mymr}', "");
Expect(0, 43644, '\P{^Is_Scx=  Mymr}', "");
Error('\p{Script_Extensions=--OLD_Turkic:=}');
Error('\P{Script_Extensions=--OLD_Turkic:=}');
Expect(1, 68680, '\p{Script_Extensions=oldturkic}', "");
Expect(0, 68680, '\p{^Script_Extensions=oldturkic}', "");
Expect(0, 68680, '\P{Script_Extensions=oldturkic}', "");
Expect(1, 68680, '\P{^Script_Extensions=oldturkic}', "");
Expect(0, 68681, '\p{Script_Extensions=oldturkic}', "");
Expect(1, 68681, '\p{^Script_Extensions=oldturkic}', "");
Expect(1, 68681, '\P{Script_Extensions=oldturkic}', "");
Expect(0, 68681, '\P{^Script_Extensions=oldturkic}', "");
Expect(1, 68680, '\p{Script_Extensions=_Old_Turkic}', "");
Expect(0, 68680, '\p{^Script_Extensions=_Old_Turkic}', "");
Expect(0, 68680, '\P{Script_Extensions=_Old_Turkic}', "");
Expect(1, 68680, '\P{^Script_Extensions=_Old_Turkic}', "");
Expect(0, 68681, '\p{Script_Extensions=_Old_Turkic}', "");
Expect(1, 68681, '\p{^Script_Extensions=_Old_Turkic}', "");
Expect(1, 68681, '\P{Script_Extensions=_Old_Turkic}', "");
Expect(0, 68681, '\P{^Script_Extensions=_Old_Turkic}', "");
Error('\p{Scx= :=Orkh}');
Error('\P{Scx= :=Orkh}');
Expect(1, 68680, '\p{Scx=orkh}', "");
Expect(0, 68680, '\p{^Scx=orkh}', "");
Expect(0, 68680, '\P{Scx=orkh}', "");
Expect(1, 68680, '\P{^Scx=orkh}', "");
Expect(0, 68681, '\p{Scx=orkh}', "");
Expect(1, 68681, '\p{^Scx=orkh}', "");
Expect(1, 68681, '\P{Scx=orkh}', "");
Expect(0, 68681, '\P{^Scx=orkh}', "");
Expect(1, 68680, '\p{Scx=--Orkh}', "");
Expect(0, 68680, '\p{^Scx=--Orkh}', "");
Expect(0, 68680, '\P{Scx=--Orkh}', "");
Expect(1, 68680, '\P{^Scx=--Orkh}', "");
Expect(0, 68681, '\p{Scx=--Orkh}', "");
Expect(1, 68681, '\p{^Scx=--Orkh}', "");
Expect(1, 68681, '\P{Scx=--Orkh}', "");
Expect(0, 68681, '\P{^Scx=--Orkh}', "");
Error('\p{Is_Script_Extensions=/a/ 	Old_TURKIC}');
Error('\P{Is_Script_Extensions=/a/ 	Old_TURKIC}');
Expect(1, 68680, '\p{Is_Script_Extensions=oldturkic}', "");
Expect(0, 68680, '\p{^Is_Script_Extensions=oldturkic}', "");
Expect(0, 68680, '\P{Is_Script_Extensions=oldturkic}', "");
Expect(1, 68680, '\P{^Is_Script_Extensions=oldturkic}', "");
Expect(0, 68681, '\p{Is_Script_Extensions=oldturkic}', "");
Expect(1, 68681, '\p{^Is_Script_Extensions=oldturkic}', "");
Expect(1, 68681, '\P{Is_Script_Extensions=oldturkic}', "");
Expect(0, 68681, '\P{^Is_Script_Extensions=oldturkic}', "");
Expect(1, 68680, '\p{Is_Script_Extensions= Old_turkic}', "");
Expect(0, 68680, '\p{^Is_Script_Extensions= Old_turkic}', "");
Expect(0, 68680, '\P{Is_Script_Extensions= Old_turkic}', "");
Expect(1, 68680, '\P{^Is_Script_Extensions= Old_turkic}', "");
Expect(0, 68681, '\p{Is_Script_Extensions= Old_turkic}', "");
Expect(1, 68681, '\p{^Is_Script_Extensions= Old_turkic}', "");
Expect(1, 68681, '\P{Is_Script_Extensions= Old_turkic}', "");
Expect(0, 68681, '\P{^Is_Script_Extensions= Old_turkic}', "");
Error('\p{Is_Scx=_ orkh/a/}');
Error('\P{Is_Scx=_ orkh/a/}');
Expect(1, 68680, '\p{Is_Scx=orkh}', "");
Expect(0, 68680, '\p{^Is_Scx=orkh}', "");
Expect(0, 68680, '\P{Is_Scx=orkh}', "");
Expect(1, 68680, '\P{^Is_Scx=orkh}', "");
Expect(0, 68681, '\p{Is_Scx=orkh}', "");
Expect(1, 68681, '\p{^Is_Scx=orkh}', "");
Expect(1, 68681, '\P{Is_Scx=orkh}', "");
Expect(0, 68681, '\P{^Is_Scx=orkh}', "");
Expect(1, 68680, '\p{Is_Scx=- Orkh}', "");
Expect(0, 68680, '\p{^Is_Scx=- Orkh}', "");
Expect(0, 68680, '\P{Is_Scx=- Orkh}', "");
Expect(1, 68680, '\P{^Is_Scx=- Orkh}', "");
Expect(0, 68681, '\p{Is_Scx=- Orkh}', "");
Expect(1, 68681, '\p{^Is_Scx=- Orkh}', "");
Expect(1, 68681, '\P{Is_Scx=- Orkh}', "");
Expect(0, 68681, '\P{^Is_Scx=- Orkh}', "");
Error('\p{Script_Extensions=	/a/osmanya}');
Error('\P{Script_Extensions=	/a/osmanya}');
Expect(1, 66729, '\p{Script_Extensions=osmanya}', "");
Expect(0, 66729, '\p{^Script_Extensions=osmanya}', "");
Expect(0, 66729, '\P{Script_Extensions=osmanya}', "");
Expect(1, 66729, '\P{^Script_Extensions=osmanya}', "");
Expect(0, 66730, '\p{Script_Extensions=osmanya}', "");
Expect(1, 66730, '\p{^Script_Extensions=osmanya}', "");
Expect(1, 66730, '\P{Script_Extensions=osmanya}', "");
Expect(0, 66730, '\P{^Script_Extensions=osmanya}', "");
Expect(1, 66729, '\p{Script_Extensions= _Osmanya}', "");
Expect(0, 66729, '\p{^Script_Extensions= _Osmanya}', "");
Expect(0, 66729, '\P{Script_Extensions= _Osmanya}', "");
Expect(1, 66729, '\P{^Script_Extensions= _Osmanya}', "");
Expect(0, 66730, '\p{Script_Extensions= _Osmanya}', "");
Expect(1, 66730, '\p{^Script_Extensions= _Osmanya}', "");
Expect(1, 66730, '\P{Script_Extensions= _Osmanya}', "");
Expect(0, 66730, '\P{^Script_Extensions= _Osmanya}', "");
Error('\p{Scx=/a/Osma}');
Error('\P{Scx=/a/Osma}');
Expect(1, 66729, '\p{Scx=osma}', "");
Expect(0, 66729, '\p{^Scx=osma}', "");
Expect(0, 66729, '\P{Scx=osma}', "");
Expect(1, 66729, '\P{^Scx=osma}', "");
Expect(0, 66730, '\p{Scx=osma}', "");
Expect(1, 66730, '\p{^Scx=osma}', "");
Expect(1, 66730, '\P{Scx=osma}', "");
Expect(0, 66730, '\P{^Scx=osma}', "");
Expect(1, 66729, '\p{Scx=	 OSMA}', "");
Expect(0, 66729, '\p{^Scx=	 OSMA}', "");
Expect(0, 66729, '\P{Scx=	 OSMA}', "");
Expect(1, 66729, '\P{^Scx=	 OSMA}', "");
Expect(0, 66730, '\p{Scx=	 OSMA}', "");
Expect(1, 66730, '\p{^Scx=	 OSMA}', "");
Expect(1, 66730, '\P{Scx=	 OSMA}', "");
Expect(0, 66730, '\P{^Scx=	 OSMA}', "");
Error('\p{Is_Script_Extensions=--Osmanya/a/}');
Error('\P{Is_Script_Extensions=--Osmanya/a/}');
Expect(1, 66729, '\p{Is_Script_Extensions=osmanya}', "");
Expect(0, 66729, '\p{^Is_Script_Extensions=osmanya}', "");
Expect(0, 66729, '\P{Is_Script_Extensions=osmanya}', "");
Expect(1, 66729, '\P{^Is_Script_Extensions=osmanya}', "");
Expect(0, 66730, '\p{Is_Script_Extensions=osmanya}', "");
Expect(1, 66730, '\p{^Is_Script_Extensions=osmanya}', "");
Expect(1, 66730, '\P{Is_Script_Extensions=osmanya}', "");
Expect(0, 66730, '\P{^Is_Script_Extensions=osmanya}', "");
Expect(1, 66729, '\p{Is_Script_Extensions=_Osmanya}', "");
Expect(0, 66729, '\p{^Is_Script_Extensions=_Osmanya}', "");
Expect(0, 66729, '\P{Is_Script_Extensions=_Osmanya}', "");
Expect(1, 66729, '\P{^Is_Script_Extensions=_Osmanya}', "");
Expect(0, 66730, '\p{Is_Script_Extensions=_Osmanya}', "");
Expect(1, 66730, '\p{^Is_Script_Extensions=_Osmanya}', "");
Expect(1, 66730, '\P{Is_Script_Extensions=_Osmanya}', "");
Expect(0, 66730, '\P{^Is_Script_Extensions=_Osmanya}', "");
Error('\p{Is_Scx=:=osma}');
Error('\P{Is_Scx=:=osma}');
Expect(1, 66729, '\p{Is_Scx=osma}', "");
Expect(0, 66729, '\p{^Is_Scx=osma}', "");
Expect(0, 66729, '\P{Is_Scx=osma}', "");
Expect(1, 66729, '\P{^Is_Scx=osma}', "");
Expect(0, 66730, '\p{Is_Scx=osma}', "");
Expect(1, 66730, '\p{^Is_Scx=osma}', "");
Expect(1, 66730, '\P{Is_Scx=osma}', "");
Expect(0, 66730, '\P{^Is_Scx=osma}', "");
Expect(1, 66729, '\p{Is_Scx=- osma}', "");
Expect(0, 66729, '\p{^Is_Scx=- osma}', "");
Expect(0, 66729, '\P{Is_Scx=- osma}', "");
Expect(1, 66729, '\P{^Is_Scx=- osma}', "");
Expect(0, 66730, '\p{Is_Scx=- osma}', "");
Expect(1, 66730, '\p{^Is_Scx=- osma}', "");
Expect(1, 66730, '\P{Is_Scx=- osma}', "");
Expect(0, 66730, '\P{^Is_Scx=- osma}', "");
Error('\p{Script_Extensions=/a/ 	TAGALOG}');
Error('\P{Script_Extensions=/a/ 	TAGALOG}');
Expect(1, 5942, '\p{Script_Extensions=tagalog}', "");
Expect(0, 5942, '\p{^Script_Extensions=tagalog}', "");
Expect(0, 5942, '\P{Script_Extensions=tagalog}', "");
Expect(1, 5942, '\P{^Script_Extensions=tagalog}', "");
Expect(0, 5943, '\p{Script_Extensions=tagalog}', "");
Expect(1, 5943, '\p{^Script_Extensions=tagalog}', "");
Expect(1, 5943, '\P{Script_Extensions=tagalog}', "");
Expect(0, 5943, '\P{^Script_Extensions=tagalog}', "");
Expect(1, 5942, '\p{Script_Extensions=-tagalog}', "");
Expect(0, 5942, '\p{^Script_Extensions=-tagalog}', "");
Expect(0, 5942, '\P{Script_Extensions=-tagalog}', "");
Expect(1, 5942, '\P{^Script_Extensions=-tagalog}', "");
Expect(0, 5943, '\p{Script_Extensions=-tagalog}', "");
Expect(1, 5943, '\p{^Script_Extensions=-tagalog}', "");
Expect(1, 5943, '\P{Script_Extensions=-tagalog}', "");
Expect(0, 5943, '\P{^Script_Extensions=-tagalog}', "");
Error('\p{Scx=:=tglg}');
Error('\P{Scx=:=tglg}');
Expect(1, 5942, '\p{Scx=tglg}', "");
Expect(0, 5942, '\p{^Scx=tglg}', "");
Expect(0, 5942, '\P{Scx=tglg}', "");
Expect(1, 5942, '\P{^Scx=tglg}', "");
Expect(0, 5943, '\p{Scx=tglg}', "");
Expect(1, 5943, '\p{^Scx=tglg}', "");
Expect(1, 5943, '\P{Scx=tglg}', "");
Expect(0, 5943, '\P{^Scx=tglg}', "");
Expect(1, 5942, '\p{Scx=Tglg}', "");
Expect(0, 5942, '\p{^Scx=Tglg}', "");
Expect(0, 5942, '\P{Scx=Tglg}', "");
Expect(1, 5942, '\P{^Scx=Tglg}', "");
Expect(0, 5943, '\p{Scx=Tglg}', "");
Expect(1, 5943, '\p{^Scx=Tglg}', "");
Expect(1, 5943, '\P{Scx=Tglg}', "");
Expect(0, 5943, '\P{^Scx=Tglg}', "");
Error('\p{Is_Script_Extensions=-	tagalog/a/}');
Error('\P{Is_Script_Extensions=-	tagalog/a/}');
Expect(1, 5942, '\p{Is_Script_Extensions=tagalog}', "");
Expect(0, 5942, '\p{^Is_Script_Extensions=tagalog}', "");
Expect(0, 5942, '\P{Is_Script_Extensions=tagalog}', "");
Expect(1, 5942, '\P{^Is_Script_Extensions=tagalog}', "");
Expect(0, 5943, '\p{Is_Script_Extensions=tagalog}', "");
Expect(1, 5943, '\p{^Is_Script_Extensions=tagalog}', "");
Expect(1, 5943, '\P{Is_Script_Extensions=tagalog}', "");
Expect(0, 5943, '\P{^Is_Script_Extensions=tagalog}', "");
Expect(1, 5942, '\p{Is_Script_Extensions:			Tagalog}', "");
Expect(0, 5942, '\p{^Is_Script_Extensions:			Tagalog}', "");
Expect(0, 5942, '\P{Is_Script_Extensions:			Tagalog}', "");
Expect(1, 5942, '\P{^Is_Script_Extensions:			Tagalog}', "");
Expect(0, 5943, '\p{Is_Script_Extensions:			Tagalog}', "");
Expect(1, 5943, '\p{^Is_Script_Extensions:			Tagalog}', "");
Expect(1, 5943, '\P{Is_Script_Extensions:			Tagalog}', "");
Expect(0, 5943, '\P{^Is_Script_Extensions:			Tagalog}', "");
Error('\p{Is_Scx= :=Tglg}');
Error('\P{Is_Scx= :=Tglg}');
Expect(1, 5942, '\p{Is_Scx:   tglg}', "");
Expect(0, 5942, '\p{^Is_Scx:   tglg}', "");
Expect(0, 5942, '\P{Is_Scx:   tglg}', "");
Expect(1, 5942, '\P{^Is_Scx:   tglg}', "");
Expect(0, 5943, '\p{Is_Scx:   tglg}', "");
Expect(1, 5943, '\p{^Is_Scx:   tglg}', "");
Expect(1, 5943, '\P{Is_Scx:   tglg}', "");
Expect(0, 5943, '\P{^Is_Scx:   tglg}', "");
Expect(1, 5942, '\p{Is_Scx=	tglg}', "");
Expect(0, 5942, '\p{^Is_Scx=	tglg}', "");
Expect(0, 5942, '\P{Is_Scx=	tglg}', "");
Expect(1, 5942, '\P{^Is_Scx=	tglg}', "");
Expect(0, 5943, '\p{Is_Scx=	tglg}', "");
Expect(1, 5943, '\p{^Is_Scx=	tglg}', "");
Expect(1, 5943, '\P{Is_Scx=	tglg}', "");
Expect(0, 5943, '\P{^Is_Scx=	tglg}', "");
Error('\p{Script_Extensions=:=_-Gujarati}');
Error('\P{Script_Extensions=:=_-Gujarati}');
Expect(1, 43065, '\p{Script_Extensions=gujarati}', "");
Expect(0, 43065, '\p{^Script_Extensions=gujarati}', "");
Expect(0, 43065, '\P{Script_Extensions=gujarati}', "");
Expect(1, 43065, '\P{^Script_Extensions=gujarati}', "");
Expect(0, 43066, '\p{Script_Extensions=gujarati}', "");
Expect(1, 43066, '\p{^Script_Extensions=gujarati}', "");
Expect(1, 43066, '\P{Script_Extensions=gujarati}', "");
Expect(0, 43066, '\P{^Script_Extensions=gujarati}', "");
Expect(1, 43065, '\p{Script_Extensions=	 Gujarati}', "");
Expect(0, 43065, '\p{^Script_Extensions=	 Gujarati}', "");
Expect(0, 43065, '\P{Script_Extensions=	 Gujarati}', "");
Expect(1, 43065, '\P{^Script_Extensions=	 Gujarati}', "");
Expect(0, 43066, '\p{Script_Extensions=	 Gujarati}', "");
Expect(1, 43066, '\p{^Script_Extensions=	 Gujarati}', "");
Expect(1, 43066, '\P{Script_Extensions=	 Gujarati}', "");
Expect(0, 43066, '\P{^Script_Extensions=	 Gujarati}', "");
Error('\p{Scx= /a/Gujr}');
Error('\P{Scx= /a/Gujr}');
Expect(1, 43065, '\p{Scx=gujr}', "");
Expect(0, 43065, '\p{^Scx=gujr}', "");
Expect(0, 43065, '\P{Scx=gujr}', "");
Expect(1, 43065, '\P{^Scx=gujr}', "");
Expect(0, 43066, '\p{Scx=gujr}', "");
Expect(1, 43066, '\p{^Scx=gujr}', "");
Expect(1, 43066, '\P{Scx=gujr}', "");
Expect(0, 43066, '\P{^Scx=gujr}', "");
Expect(1, 43065, '\p{Scx= GUJR}', "");
Expect(0, 43065, '\p{^Scx= GUJR}', "");
Expect(0, 43065, '\P{Scx= GUJR}', "");
Expect(1, 43065, '\P{^Scx= GUJR}', "");
Expect(0, 43066, '\p{Scx= GUJR}', "");
Expect(1, 43066, '\p{^Scx= GUJR}', "");
Expect(1, 43066, '\P{Scx= GUJR}', "");
Expect(0, 43066, '\P{^Scx= GUJR}', "");
Error('\p{Is_Script_Extensions=:=_ GUJARATI}');
Error('\P{Is_Script_Extensions=:=_ GUJARATI}');
Expect(1, 43065, '\p{Is_Script_Extensions:gujarati}', "");
Expect(0, 43065, '\p{^Is_Script_Extensions:gujarati}', "");
Expect(0, 43065, '\P{Is_Script_Extensions:gujarati}', "");
Expect(1, 43065, '\P{^Is_Script_Extensions:gujarati}', "");
Expect(0, 43066, '\p{Is_Script_Extensions:gujarati}', "");
Expect(1, 43066, '\p{^Is_Script_Extensions:gujarati}', "");
Expect(1, 43066, '\P{Is_Script_Extensions:gujarati}', "");
Expect(0, 43066, '\P{^Is_Script_Extensions:gujarati}', "");
Expect(1, 43065, '\p{Is_Script_Extensions:-	GUJARATI}', "");
Expect(0, 43065, '\p{^Is_Script_Extensions:-	GUJARATI}', "");
Expect(0, 43065, '\P{Is_Script_Extensions:-	GUJARATI}', "");
Expect(1, 43065, '\P{^Is_Script_Extensions:-	GUJARATI}', "");
Expect(0, 43066, '\p{Is_Script_Extensions:-	GUJARATI}', "");
Expect(1, 43066, '\p{^Is_Script_Extensions:-	GUJARATI}', "");
Expect(1, 43066, '\P{Is_Script_Extensions:-	GUJARATI}', "");
Expect(0, 43066, '\P{^Is_Script_Extensions:-	GUJARATI}', "");
Error('\p{Is_Scx=/a/--Gujr}');
Error('\P{Is_Scx=/a/--Gujr}');
Expect(1, 43065, '\p{Is_Scx=gujr}', "");
Expect(0, 43065, '\p{^Is_Scx=gujr}', "");
Expect(0, 43065, '\P{Is_Scx=gujr}', "");
Expect(1, 43065, '\P{^Is_Scx=gujr}', "");
Expect(0, 43066, '\p{Is_Scx=gujr}', "");
Expect(1, 43066, '\p{^Is_Scx=gujr}', "");
Expect(1, 43066, '\P{Is_Scx=gujr}', "");
Expect(0, 43066, '\P{^Is_Scx=gujr}', "");
Expect(1, 43065, '\p{Is_Scx=GUJR}', "");
Expect(0, 43065, '\p{^Is_Scx=GUJR}', "");
Expect(0, 43065, '\P{Is_Scx=GUJR}', "");
Expect(1, 43065, '\P{^Is_Scx=GUJR}', "");
Expect(0, 43066, '\p{Is_Scx=GUJR}', "");
Expect(1, 43066, '\p{^Is_Scx=GUJR}', "");
Expect(1, 43066, '\P{Is_Scx=GUJR}', "");
Expect(0, 43066, '\P{^Is_Scx=GUJR}', "");
Error('\p{Script_Extensions=:= -oriya}');
Error('\P{Script_Extensions=:= -oriya}');
Expect(1, 2935, '\p{Script_Extensions=oriya}', "");
Expect(0, 2935, '\p{^Script_Extensions=oriya}', "");
Expect(0, 2935, '\P{Script_Extensions=oriya}', "");
Expect(1, 2935, '\P{^Script_Extensions=oriya}', "");
Expect(0, 2936, '\p{Script_Extensions=oriya}', "");
Expect(1, 2936, '\p{^Script_Extensions=oriya}', "");
Expect(1, 2936, '\P{Script_Extensions=oriya}', "");
Expect(0, 2936, '\P{^Script_Extensions=oriya}', "");
Expect(1, 2935, '\p{Script_Extensions=-	Oriya}', "");
Expect(0, 2935, '\p{^Script_Extensions=-	Oriya}', "");
Expect(0, 2935, '\P{Script_Extensions=-	Oriya}', "");
Expect(1, 2935, '\P{^Script_Extensions=-	Oriya}', "");
Expect(0, 2936, '\p{Script_Extensions=-	Oriya}', "");
Expect(1, 2936, '\p{^Script_Extensions=-	Oriya}', "");
Expect(1, 2936, '\P{Script_Extensions=-	Oriya}', "");
Expect(0, 2936, '\P{^Script_Extensions=-	Oriya}', "");
Error('\p{Scx=	ORYA/a/}');
Error('\P{Scx=	ORYA/a/}');
Expect(1, 2935, '\p{Scx=orya}', "");
Expect(0, 2935, '\p{^Scx=orya}', "");
Expect(0, 2935, '\P{Scx=orya}', "");
Expect(1, 2935, '\P{^Scx=orya}', "");
Expect(0, 2936, '\p{Scx=orya}', "");
Expect(1, 2936, '\p{^Scx=orya}', "");
Expect(1, 2936, '\P{Scx=orya}', "");
Expect(0, 2936, '\P{^Scx=orya}', "");
Expect(1, 2935, '\p{Scx=__Orya}', "");
Expect(0, 2935, '\p{^Scx=__Orya}', "");
Expect(0, 2935, '\P{Scx=__Orya}', "");
Expect(1, 2935, '\P{^Scx=__Orya}', "");
Expect(0, 2936, '\p{Scx=__Orya}', "");
Expect(1, 2936, '\p{^Scx=__Orya}', "");
Expect(1, 2936, '\P{Scx=__Orya}', "");
Expect(0, 2936, '\P{^Scx=__Orya}', "");
Error('\p{Is_Script_Extensions=-:=oriya}');
Error('\P{Is_Script_Extensions=-:=oriya}');
Expect(1, 2935, '\p{Is_Script_Extensions=oriya}', "");
Expect(0, 2935, '\p{^Is_Script_Extensions=oriya}', "");
Expect(0, 2935, '\P{Is_Script_Extensions=oriya}', "");
Expect(1, 2935, '\P{^Is_Script_Extensions=oriya}', "");
Expect(0, 2936, '\p{Is_Script_Extensions=oriya}', "");
Expect(1, 2936, '\p{^Is_Script_Extensions=oriya}', "");
Expect(1, 2936, '\P{Is_Script_Extensions=oriya}', "");
Expect(0, 2936, '\P{^Is_Script_Extensions=oriya}', "");
Expect(1, 2935, '\p{Is_Script_Extensions= 	ORIYA}', "");
Expect(0, 2935, '\p{^Is_Script_Extensions= 	ORIYA}', "");
Expect(0, 2935, '\P{Is_Script_Extensions= 	ORIYA}', "");
Expect(1, 2935, '\P{^Is_Script_Extensions= 	ORIYA}', "");
Expect(0, 2936, '\p{Is_Script_Extensions= 	ORIYA}', "");
Expect(1, 2936, '\p{^Is_Script_Extensions= 	ORIYA}', "");
Expect(1, 2936, '\P{Is_Script_Extensions= 	ORIYA}', "");
Expect(0, 2936, '\P{^Is_Script_Extensions= 	ORIYA}', "");
Error('\p{Is_Scx:-/a/orya}');
Error('\P{Is_Scx:-/a/orya}');
Expect(1, 2935, '\p{Is_Scx=orya}', "");
Expect(0, 2935, '\p{^Is_Scx=orya}', "");
Expect(0, 2935, '\P{Is_Scx=orya}', "");
Expect(1, 2935, '\P{^Is_Scx=orya}', "");
Expect(0, 2936, '\p{Is_Scx=orya}', "");
Expect(1, 2936, '\p{^Is_Scx=orya}', "");
Expect(1, 2936, '\P{Is_Scx=orya}', "");
Expect(0, 2936, '\P{^Is_Scx=orya}', "");
Expect(1, 2935, '\p{Is_Scx=	ORYA}', "");
Expect(0, 2935, '\p{^Is_Scx=	ORYA}', "");
Expect(0, 2935, '\P{Is_Scx=	ORYA}', "");
Expect(1, 2935, '\P{^Is_Scx=	ORYA}', "");
Expect(0, 2936, '\p{Is_Scx=	ORYA}', "");
Expect(1, 2936, '\p{^Is_Scx=	ORYA}', "");
Expect(1, 2936, '\P{Is_Scx=	ORYA}', "");
Expect(0, 2936, '\P{^Is_Scx=	ORYA}', "");
Error('\p{Script_Extensions=/a/__lisu}');
Error('\P{Script_Extensions=/a/__lisu}');
Expect(1, 42239, '\p{Script_Extensions=lisu}', "");
Expect(0, 42239, '\p{^Script_Extensions=lisu}', "");
Expect(0, 42239, '\P{Script_Extensions=lisu}', "");
Expect(1, 42239, '\P{^Script_Extensions=lisu}', "");
Expect(0, 42240, '\p{Script_Extensions=lisu}', "");
Expect(1, 42240, '\p{^Script_Extensions=lisu}', "");
Expect(1, 42240, '\P{Script_Extensions=lisu}', "");
Expect(0, 42240, '\P{^Script_Extensions=lisu}', "");
Expect(1, 42239, '\p{Script_Extensions= 	Lisu}', "");
Expect(0, 42239, '\p{^Script_Extensions= 	Lisu}', "");
Expect(0, 42239, '\P{Script_Extensions= 	Lisu}', "");
Expect(1, 42239, '\P{^Script_Extensions= 	Lisu}', "");
Expect(0, 42240, '\p{Script_Extensions= 	Lisu}', "");
Expect(1, 42240, '\p{^Script_Extensions= 	Lisu}', "");
Expect(1, 42240, '\P{Script_Extensions= 	Lisu}', "");
Expect(0, 42240, '\P{^Script_Extensions= 	Lisu}', "");
Error('\p{Scx=/a/Lisu}');
Error('\P{Scx=/a/Lisu}');
Expect(1, 42239, '\p{Scx=lisu}', "");
Expect(0, 42239, '\p{^Scx=lisu}', "");
Expect(0, 42239, '\P{Scx=lisu}', "");
Expect(1, 42239, '\P{^Scx=lisu}', "");
Expect(0, 42240, '\p{Scx=lisu}', "");
Expect(1, 42240, '\p{^Scx=lisu}', "");
Expect(1, 42240, '\P{Scx=lisu}', "");
Expect(0, 42240, '\P{^Scx=lisu}', "");
Expect(1, 42239, '\p{Scx=_-LISU}', "");
Expect(0, 42239, '\p{^Scx=_-LISU}', "");
Expect(0, 42239, '\P{Scx=_-LISU}', "");
Expect(1, 42239, '\P{^Scx=_-LISU}', "");
Expect(0, 42240, '\p{Scx=_-LISU}', "");
Expect(1, 42240, '\p{^Scx=_-LISU}', "");
Expect(1, 42240, '\P{Scx=_-LISU}', "");
Expect(0, 42240, '\P{^Scx=_-LISU}', "");
Error('\p{Is_Script_Extensions=:=	-Lisu}');
Error('\P{Is_Script_Extensions=:=	-Lisu}');
Expect(1, 42239, '\p{Is_Script_Extensions=lisu}', "");
Expect(0, 42239, '\p{^Is_Script_Extensions=lisu}', "");
Expect(0, 42239, '\P{Is_Script_Extensions=lisu}', "");
Expect(1, 42239, '\P{^Is_Script_Extensions=lisu}', "");
Expect(0, 42240, '\p{Is_Script_Extensions=lisu}', "");
Expect(1, 42240, '\p{^Is_Script_Extensions=lisu}', "");
Expect(1, 42240, '\P{Is_Script_Extensions=lisu}', "");
Expect(0, 42240, '\P{^Is_Script_Extensions=lisu}', "");
Expect(1, 42239, '\p{Is_Script_Extensions:	-_Lisu}', "");
Expect(0, 42239, '\p{^Is_Script_Extensions:	-_Lisu}', "");
Expect(0, 42239, '\P{Is_Script_Extensions:	-_Lisu}', "");
Expect(1, 42239, '\P{^Is_Script_Extensions:	-_Lisu}', "");
Expect(0, 42240, '\p{Is_Script_Extensions:	-_Lisu}', "");
Expect(1, 42240, '\p{^Is_Script_Extensions:	-_Lisu}', "");
Expect(1, 42240, '\P{Is_Script_Extensions:	-_Lisu}', "");
Expect(0, 42240, '\P{^Is_Script_Extensions:	-_Lisu}', "");
Error('\p{Is_Scx=-:=lisu}');
Error('\P{Is_Scx=-:=lisu}');
Expect(1, 42239, '\p{Is_Scx=lisu}', "");
Expect(0, 42239, '\p{^Is_Scx=lisu}', "");
Expect(0, 42239, '\P{Is_Scx=lisu}', "");
Expect(1, 42239, '\P{^Is_Scx=lisu}', "");
Expect(0, 42240, '\p{Is_Scx=lisu}', "");
Expect(1, 42240, '\p{^Is_Scx=lisu}', "");
Expect(1, 42240, '\P{Is_Scx=lisu}', "");
Expect(0, 42240, '\P{^Is_Scx=lisu}', "");
Expect(1, 42239, '\p{Is_Scx=	_Lisu}', "");
Expect(0, 42239, '\p{^Is_Scx=	_Lisu}', "");
Expect(0, 42239, '\P{Is_Scx=	_Lisu}', "");
Expect(1, 42239, '\P{^Is_Scx=	_Lisu}', "");
Expect(0, 42240, '\p{Is_Scx=	_Lisu}', "");
Expect(1, 42240, '\p{^Is_Scx=	_Lisu}', "");
Expect(1, 42240, '\P{Is_Scx=	_Lisu}', "");
Expect(0, 42240, '\P{^Is_Scx=	_Lisu}', "");
Error('\p{Script_Extensions:   /a/-	avestan}');
Error('\P{Script_Extensions:   /a/-	avestan}');
Expect(1, 68415, '\p{Script_Extensions=avestan}', "");
Expect(0, 68415, '\p{^Script_Extensions=avestan}', "");
Expect(0, 68415, '\P{Script_Extensions=avestan}', "");
Expect(1, 68415, '\P{^Script_Extensions=avestan}', "");
Expect(0, 68416, '\p{Script_Extensions=avestan}', "");
Expect(1, 68416, '\p{^Script_Extensions=avestan}', "");
Expect(1, 68416, '\P{Script_Extensions=avestan}', "");
Expect(0, 68416, '\P{^Script_Extensions=avestan}', "");
Expect(1, 68415, '\p{Script_Extensions=	 Avestan}', "");
Expect(0, 68415, '\p{^Script_Extensions=	 Avestan}', "");
Expect(0, 68415, '\P{Script_Extensions=	 Avestan}', "");
Expect(1, 68415, '\P{^Script_Extensions=	 Avestan}', "");
Expect(0, 68416, '\p{Script_Extensions=	 Avestan}', "");
Expect(1, 68416, '\p{^Script_Extensions=	 Avestan}', "");
Expect(1, 68416, '\P{Script_Extensions=	 Avestan}', "");
Expect(0, 68416, '\P{^Script_Extensions=	 Avestan}', "");
Error('\p{Scx=/a/-avst}');
Error('\P{Scx=/a/-avst}');
Expect(1, 68415, '\p{Scx:   avst}', "");
Expect(0, 68415, '\p{^Scx:   avst}', "");
Expect(0, 68415, '\P{Scx:   avst}', "");
Expect(1, 68415, '\P{^Scx:   avst}', "");
Expect(0, 68416, '\p{Scx:   avst}', "");
Expect(1, 68416, '\p{^Scx:   avst}', "");
Expect(1, 68416, '\P{Scx:   avst}', "");
Expect(0, 68416, '\P{^Scx:   avst}', "");
Expect(1, 68415, '\p{Scx=_ Avst}', "");
Expect(0, 68415, '\p{^Scx=_ Avst}', "");
Expect(0, 68415, '\P{Scx=_ Avst}', "");
Expect(1, 68415, '\P{^Scx=_ Avst}', "");
Expect(0, 68416, '\p{Scx=_ Avst}', "");
Expect(1, 68416, '\p{^Scx=_ Avst}', "");
Expect(1, 68416, '\P{Scx=_ Avst}', "");
Expect(0, 68416, '\P{^Scx=_ Avst}', "");
Error('\p{Is_Script_Extensions: := avestan}');
Error('\P{Is_Script_Extensions: := avestan}');
Expect(1, 68415, '\p{Is_Script_Extensions=avestan}', "");
Expect(0, 68415, '\p{^Is_Script_Extensions=avestan}', "");
Expect(0, 68415, '\P{Is_Script_Extensions=avestan}', "");
Expect(1, 68415, '\P{^Is_Script_Extensions=avestan}', "");
Expect(0, 68416, '\p{Is_Script_Extensions=avestan}', "");
Expect(1, 68416, '\p{^Is_Script_Extensions=avestan}', "");
Expect(1, 68416, '\P{Is_Script_Extensions=avestan}', "");
Expect(0, 68416, '\P{^Is_Script_Extensions=avestan}', "");
Expect(1, 68415, '\p{Is_Script_Extensions:	Avestan}', "");
Expect(0, 68415, '\p{^Is_Script_Extensions:	Avestan}', "");
Expect(0, 68415, '\P{Is_Script_Extensions:	Avestan}', "");
Expect(1, 68415, '\P{^Is_Script_Extensions:	Avestan}', "");
Expect(0, 68416, '\p{Is_Script_Extensions:	Avestan}', "");
Expect(1, 68416, '\p{^Is_Script_Extensions:	Avestan}', "");
Expect(1, 68416, '\P{Is_Script_Extensions:	Avestan}', "");
Expect(0, 68416, '\P{^Is_Script_Extensions:	Avestan}', "");
Error('\p{Is_Scx=/a/- avst}');
Error('\P{Is_Scx=/a/- avst}');
Expect(1, 68415, '\p{Is_Scx=avst}', "");
Expect(0, 68415, '\p{^Is_Scx=avst}', "");
Expect(0, 68415, '\P{Is_Scx=avst}', "");
Expect(1, 68415, '\P{^Is_Scx=avst}', "");
Expect(0, 68416, '\p{Is_Scx=avst}', "");
Expect(1, 68416, '\p{^Is_Scx=avst}', "");
Expect(1, 68416, '\P{Is_Scx=avst}', "");
Expect(0, 68416, '\P{^Is_Scx=avst}', "");
Expect(1, 68415, '\p{Is_Scx= 	avst}', "");
Expect(0, 68415, '\p{^Is_Scx= 	avst}', "");
Expect(0, 68415, '\P{Is_Scx= 	avst}', "");
Expect(1, 68415, '\P{^Is_Scx= 	avst}', "");
Expect(0, 68416, '\p{Is_Scx= 	avst}', "");
Expect(1, 68416, '\p{^Is_Scx= 	avst}', "");
Expect(1, 68416, '\P{Is_Scx= 	avst}', "");
Expect(0, 68416, '\P{^Is_Scx= 	avst}', "");
Error('\p{Script_Extensions=:=	Tai_le}');
Error('\P{Script_Extensions=:=	Tai_le}');
Expect(1, 6516, '\p{Script_Extensions=taile}', "");
Expect(0, 6516, '\p{^Script_Extensions=taile}', "");
Expect(0, 6516, '\P{Script_Extensions=taile}', "");
Expect(1, 6516, '\P{^Script_Extensions=taile}', "");
Expect(0, 6517, '\p{Script_Extensions=taile}', "");
Expect(1, 6517, '\p{^Script_Extensions=taile}', "");
Expect(1, 6517, '\P{Script_Extensions=taile}', "");
Expect(0, 6517, '\P{^Script_Extensions=taile}', "");
Expect(1, 6516, '\p{Script_Extensions:   	-Tai_LE}', "");
Expect(0, 6516, '\p{^Script_Extensions:   	-Tai_LE}', "");
Expect(0, 6516, '\P{Script_Extensions:   	-Tai_LE}', "");
Expect(1, 6516, '\P{^Script_Extensions:   	-Tai_LE}', "");
Expect(0, 6517, '\p{Script_Extensions:   	-Tai_LE}', "");
Expect(1, 6517, '\p{^Script_Extensions:   	-Tai_LE}', "");
Expect(1, 6517, '\P{Script_Extensions:   	-Tai_LE}', "");
Expect(0, 6517, '\P{^Script_Extensions:   	-Tai_LE}', "");
Error('\p{Scx= 	Tale/a/}');
Error('\P{Scx= 	Tale/a/}');
Expect(1, 6516, '\p{Scx=tale}', "");
Expect(0, 6516, '\p{^Scx=tale}', "");
Expect(0, 6516, '\P{Scx=tale}', "");
Expect(1, 6516, '\P{^Scx=tale}', "");
Expect(0, 6517, '\p{Scx=tale}', "");
Expect(1, 6517, '\p{^Scx=tale}', "");
Expect(1, 6517, '\P{Scx=tale}', "");
Expect(0, 6517, '\P{^Scx=tale}', "");
Expect(1, 6516, '\p{Scx= Tale}', "");
Expect(0, 6516, '\p{^Scx= Tale}', "");
Expect(0, 6516, '\P{Scx= Tale}', "");
Expect(1, 6516, '\P{^Scx= Tale}', "");
Expect(0, 6517, '\p{Scx= Tale}', "");
Expect(1, 6517, '\p{^Scx= Tale}', "");
Expect(1, 6517, '\P{Scx= Tale}', "");
Expect(0, 6517, '\P{^Scx= Tale}', "");
Error('\p{Is_Script_Extensions=	tai_le:=}');
Error('\P{Is_Script_Extensions=	tai_le:=}');
Expect(1, 6516, '\p{Is_Script_Extensions=taile}', "");
Expect(0, 6516, '\p{^Is_Script_Extensions=taile}', "");
Expect(0, 6516, '\P{Is_Script_Extensions=taile}', "");
Expect(1, 6516, '\P{^Is_Script_Extensions=taile}', "");
Expect(0, 6517, '\p{Is_Script_Extensions=taile}', "");
Expect(1, 6517, '\p{^Is_Script_Extensions=taile}', "");
Expect(1, 6517, '\P{Is_Script_Extensions=taile}', "");
Expect(0, 6517, '\P{^Is_Script_Extensions=taile}', "");
Expect(1, 6516, '\p{Is_Script_Extensions=_ tai_Le}', "");
Expect(0, 6516, '\p{^Is_Script_Extensions=_ tai_Le}', "");
Expect(0, 6516, '\P{Is_Script_Extensions=_ tai_Le}', "");
Expect(1, 6516, '\P{^Is_Script_Extensions=_ tai_Le}', "");
Expect(0, 6517, '\p{Is_Script_Extensions=_ tai_Le}', "");
Expect(1, 6517, '\p{^Is_Script_Extensions=_ tai_Le}', "");
Expect(1, 6517, '\P{Is_Script_Extensions=_ tai_Le}', "");
Expect(0, 6517, '\P{^Is_Script_Extensions=_ tai_Le}', "");
Error('\p{Is_Scx:   _/a/TALE}');
Error('\P{Is_Scx:   _/a/TALE}');
Expect(1, 6516, '\p{Is_Scx=tale}', "");
Expect(0, 6516, '\p{^Is_Scx=tale}', "");
Expect(0, 6516, '\P{Is_Scx=tale}', "");
Expect(1, 6516, '\P{^Is_Scx=tale}', "");
Expect(0, 6517, '\p{Is_Scx=tale}', "");
Expect(1, 6517, '\p{^Is_Scx=tale}', "");
Expect(1, 6517, '\P{Is_Scx=tale}', "");
Expect(0, 6517, '\P{^Is_Scx=tale}', "");
Expect(1, 6516, '\p{Is_Scx:   __Tale}', "");
Expect(0, 6516, '\p{^Is_Scx:   __Tale}', "");
Expect(0, 6516, '\P{Is_Scx:   __Tale}', "");
Expect(1, 6516, '\P{^Is_Scx:   __Tale}', "");
Expect(0, 6517, '\p{Is_Scx:   __Tale}', "");
Expect(1, 6517, '\p{^Is_Scx:   __Tale}', "");
Expect(1, 6517, '\P{Is_Scx:   __Tale}', "");
Expect(0, 6517, '\P{^Is_Scx:   __Tale}', "");
Error('\p{Script_Extensions=	:=Georgian}');
Error('\P{Script_Extensions=	:=Georgian}');
Expect(1, 11565, '\p{Script_Extensions=georgian}', "");
Expect(0, 11565, '\p{^Script_Extensions=georgian}', "");
Expect(0, 11565, '\P{Script_Extensions=georgian}', "");
Expect(1, 11565, '\P{^Script_Extensions=georgian}', "");
Expect(0, 11566, '\p{Script_Extensions=georgian}', "");
Expect(1, 11566, '\p{^Script_Extensions=georgian}', "");
Expect(1, 11566, '\P{Script_Extensions=georgian}', "");
Expect(0, 11566, '\P{^Script_Extensions=georgian}', "");
Expect(1, 11565, '\p{Script_Extensions=	 Georgian}', "");
Expect(0, 11565, '\p{^Script_Extensions=	 Georgian}', "");
Expect(0, 11565, '\P{Script_Extensions=	 Georgian}', "");
Expect(1, 11565, '\P{^Script_Extensions=	 Georgian}', "");
Expect(0, 11566, '\p{Script_Extensions=	 Georgian}', "");
Expect(1, 11566, '\p{^Script_Extensions=	 Georgian}', "");
Expect(1, 11566, '\P{Script_Extensions=	 Georgian}', "");
Expect(0, 11566, '\P{^Script_Extensions=	 Georgian}', "");
Error('\p{Scx=:=	 Geor}');
Error('\P{Scx=:=	 Geor}');
Expect(1, 11565, '\p{Scx=geor}', "");
Expect(0, 11565, '\p{^Scx=geor}', "");
Expect(0, 11565, '\P{Scx=geor}', "");
Expect(1, 11565, '\P{^Scx=geor}', "");
Expect(0, 11566, '\p{Scx=geor}', "");
Expect(1, 11566, '\p{^Scx=geor}', "");
Expect(1, 11566, '\P{Scx=geor}', "");
Expect(0, 11566, '\P{^Scx=geor}', "");
Expect(1, 11565, '\p{Scx=	Geor}', "");
Expect(0, 11565, '\p{^Scx=	Geor}', "");
Expect(0, 11565, '\P{Scx=	Geor}', "");
Expect(1, 11565, '\P{^Scx=	Geor}', "");
Expect(0, 11566, '\p{Scx=	Geor}', "");
Expect(1, 11566, '\p{^Scx=	Geor}', "");
Expect(1, 11566, '\P{Scx=	Geor}', "");
Expect(0, 11566, '\P{^Scx=	Geor}', "");
Error('\p{Is_Script_Extensions=	-GEORGIAN/a/}');
Error('\P{Is_Script_Extensions=	-GEORGIAN/a/}');
Expect(1, 11565, '\p{Is_Script_Extensions=georgian}', "");
Expect(0, 11565, '\p{^Is_Script_Extensions=georgian}', "");
Expect(0, 11565, '\P{Is_Script_Extensions=georgian}', "");
Expect(1, 11565, '\P{^Is_Script_Extensions=georgian}', "");
Expect(0, 11566, '\p{Is_Script_Extensions=georgian}', "");
Expect(1, 11566, '\p{^Is_Script_Extensions=georgian}', "");
Expect(1, 11566, '\P{Is_Script_Extensions=georgian}', "");
Expect(0, 11566, '\P{^Is_Script_Extensions=georgian}', "");
Expect(1, 11565, '\p{Is_Script_Extensions=_-Georgian}', "");
Expect(0, 11565, '\p{^Is_Script_Extensions=_-Georgian}', "");
Expect(0, 11565, '\P{Is_Script_Extensions=_-Georgian}', "");
Expect(1, 11565, '\P{^Is_Script_Extensions=_-Georgian}', "");
Expect(0, 11566, '\p{Is_Script_Extensions=_-Georgian}', "");
Expect(1, 11566, '\p{^Is_Script_Extensions=_-Georgian}', "");
Expect(1, 11566, '\P{Is_Script_Extensions=_-Georgian}', "");
Expect(0, 11566, '\P{^Is_Script_Extensions=_-Georgian}', "");
Error('\p{Is_Scx=:=_GEOR}');
Error('\P{Is_Scx=:=_GEOR}');
Expect(1, 11565, '\p{Is_Scx=geor}', "");
Expect(0, 11565, '\p{^Is_Scx=geor}', "");
Expect(0, 11565, '\P{Is_Scx=geor}', "");
Expect(1, 11565, '\P{^Is_Scx=geor}', "");
Expect(0, 11566, '\p{Is_Scx=geor}', "");
Expect(1, 11566, '\p{^Is_Scx=geor}', "");
Expect(1, 11566, '\P{Is_Scx=geor}', "");
Expect(0, 11566, '\P{^Is_Scx=geor}', "");
Expect(1, 11565, '\p{Is_Scx=	-Geor}', "");
Expect(0, 11565, '\p{^Is_Scx=	-Geor}', "");
Expect(0, 11565, '\P{Is_Scx=	-Geor}', "");
Expect(1, 11565, '\P{^Is_Scx=	-Geor}', "");
Expect(0, 11566, '\p{Is_Scx=	-Geor}', "");
Expect(1, 11566, '\p{^Is_Scx=	-Geor}', "");
Expect(1, 11566, '\P{Is_Scx=	-Geor}', "");
Expect(0, 11566, '\P{^Is_Scx=	-Geor}', "");
Error('\p{Script_Extensions= /a/Telugu}');
Error('\P{Script_Extensions= /a/Telugu}');
Expect(1, 3199, '\p{Script_Extensions=telugu}', "");
Expect(0, 3199, '\p{^Script_Extensions=telugu}', "");
Expect(0, 3199, '\P{Script_Extensions=telugu}', "");
Expect(1, 3199, '\P{^Script_Extensions=telugu}', "");
Expect(0, 3200, '\p{Script_Extensions=telugu}', "");
Expect(1, 3200, '\p{^Script_Extensions=telugu}', "");
Expect(1, 3200, '\P{Script_Extensions=telugu}', "");
Expect(0, 3200, '\P{^Script_Extensions=telugu}', "");
Expect(1, 3199, '\p{Script_Extensions=--Telugu}', "");
Expect(0, 3199, '\p{^Script_Extensions=--Telugu}', "");
Expect(0, 3199, '\P{Script_Extensions=--Telugu}', "");
Expect(1, 3199, '\P{^Script_Extensions=--Telugu}', "");
Expect(0, 3200, '\p{Script_Extensions=--Telugu}', "");
Expect(1, 3200, '\p{^Script_Extensions=--Telugu}', "");
Expect(1, 3200, '\P{Script_Extensions=--Telugu}', "");
Expect(0, 3200, '\P{^Script_Extensions=--Telugu}', "");
Error('\p{Scx=	 Telu:=}');
Error('\P{Scx=	 Telu:=}');
Expect(1, 3199, '\p{Scx=telu}', "");
Expect(0, 3199, '\p{^Scx=telu}', "");
Expect(0, 3199, '\P{Scx=telu}', "");
Expect(1, 3199, '\P{^Scx=telu}', "");
Expect(0, 3200, '\p{Scx=telu}', "");
Expect(1, 3200, '\p{^Scx=telu}', "");
Expect(1, 3200, '\P{Scx=telu}', "");
Expect(0, 3200, '\P{^Scx=telu}', "");
Expect(1, 3199, '\p{Scx=	TELU}', "");
Expect(0, 3199, '\p{^Scx=	TELU}', "");
Expect(0, 3199, '\P{Scx=	TELU}', "");
Expect(1, 3199, '\P{^Scx=	TELU}', "");
Expect(0, 3200, '\p{Scx=	TELU}', "");
Expect(1, 3200, '\p{^Scx=	TELU}', "");
Expect(1, 3200, '\P{Scx=	TELU}', "");
Expect(0, 3200, '\P{^Scx=	TELU}', "");
Error('\p{Is_Script_Extensions=	/a/Telugu}');
Error('\P{Is_Script_Extensions=	/a/Telugu}');
Expect(1, 3199, '\p{Is_Script_Extensions=telugu}', "");
Expect(0, 3199, '\p{^Is_Script_Extensions=telugu}', "");
Expect(0, 3199, '\P{Is_Script_Extensions=telugu}', "");
Expect(1, 3199, '\P{^Is_Script_Extensions=telugu}', "");
Expect(0, 3200, '\p{Is_Script_Extensions=telugu}', "");
Expect(1, 3200, '\p{^Is_Script_Extensions=telugu}', "");
Expect(1, 3200, '\P{Is_Script_Extensions=telugu}', "");
Expect(0, 3200, '\P{^Is_Script_Extensions=telugu}', "");
Expect(1, 3199, '\p{Is_Script_Extensions=_TELUGU}', "");
Expect(0, 3199, '\p{^Is_Script_Extensions=_TELUGU}', "");
Expect(0, 3199, '\P{Is_Script_Extensions=_TELUGU}', "");
Expect(1, 3199, '\P{^Is_Script_Extensions=_TELUGU}', "");
Expect(0, 3200, '\p{Is_Script_Extensions=_TELUGU}', "");
Expect(1, 3200, '\p{^Is_Script_Extensions=_TELUGU}', "");
Expect(1, 3200, '\P{Is_Script_Extensions=_TELUGU}', "");
Expect(0, 3200, '\P{^Is_Script_Extensions=_TELUGU}', "");
Error('\p{Is_Scx=/a/Telu}');
Error('\P{Is_Scx=/a/Telu}');
Expect(1, 3199, '\p{Is_Scx=telu}', "");
Expect(0, 3199, '\p{^Is_Scx=telu}', "");
Expect(0, 3199, '\P{Is_Scx=telu}', "");
Expect(1, 3199, '\P{^Is_Scx=telu}', "");
Expect(0, 3200, '\p{Is_Scx=telu}', "");
Expect(1, 3200, '\p{^Is_Scx=telu}', "");
Expect(1, 3200, '\P{Is_Scx=telu}', "");
Expect(0, 3200, '\P{^Is_Scx=telu}', "");
Expect(1, 3199, '\p{Is_Scx=	-TELU}', "");
Expect(0, 3199, '\p{^Is_Scx=	-TELU}', "");
Expect(0, 3199, '\P{Is_Scx=	-TELU}', "");
Expect(1, 3199, '\P{^Is_Scx=	-TELU}', "");
Expect(0, 3200, '\p{Is_Scx=	-TELU}', "");
Expect(1, 3200, '\p{^Is_Scx=	-TELU}', "");
Expect(1, 3200, '\P{Is_Scx=	-TELU}', "");
Expect(0, 3200, '\P{^Is_Scx=	-TELU}', "");
Error('\p{Script_Extensions=_/a/TIFINAGH}');
Error('\P{Script_Extensions=_/a/TIFINAGH}');
Expect(1, 11647, '\p{Script_Extensions=tifinagh}', "");
Expect(0, 11647, '\p{^Script_Extensions=tifinagh}', "");
Expect(0, 11647, '\P{Script_Extensions=tifinagh}', "");
Expect(1, 11647, '\P{^Script_Extensions=tifinagh}', "");
Expect(0, 11648, '\p{Script_Extensions=tifinagh}', "");
Expect(1, 11648, '\p{^Script_Extensions=tifinagh}', "");
Expect(1, 11648, '\P{Script_Extensions=tifinagh}', "");
Expect(0, 11648, '\P{^Script_Extensions=tifinagh}', "");
Expect(1, 11647, '\p{Script_Extensions=__Tifinagh}', "");
Expect(0, 11647, '\p{^Script_Extensions=__Tifinagh}', "");
Expect(0, 11647, '\P{Script_Extensions=__Tifinagh}', "");
Expect(1, 11647, '\P{^Script_Extensions=__Tifinagh}', "");
Expect(0, 11648, '\p{Script_Extensions=__Tifinagh}', "");
Expect(1, 11648, '\p{^Script_Extensions=__Tifinagh}', "");
Expect(1, 11648, '\P{Script_Extensions=__Tifinagh}', "");
Expect(0, 11648, '\P{^Script_Extensions=__Tifinagh}', "");
Error('\p{Scx:/a/--tfng}');
Error('\P{Scx:/a/--tfng}');
Expect(1, 11647, '\p{Scx=tfng}', "");
Expect(0, 11647, '\p{^Scx=tfng}', "");
Expect(0, 11647, '\P{Scx=tfng}', "");
Expect(1, 11647, '\P{^Scx=tfng}', "");
Expect(0, 11648, '\p{Scx=tfng}', "");
Expect(1, 11648, '\p{^Scx=tfng}', "");
Expect(1, 11648, '\P{Scx=tfng}', "");
Expect(0, 11648, '\P{^Scx=tfng}', "");
Expect(1, 11647, '\p{Scx=	-Tfng}', "");
Expect(0, 11647, '\p{^Scx=	-Tfng}', "");
Expect(0, 11647, '\P{Scx=	-Tfng}', "");
Expect(1, 11647, '\P{^Scx=	-Tfng}', "");
Expect(0, 11648, '\p{Scx=	-Tfng}', "");
Expect(1, 11648, '\p{^Scx=	-Tfng}', "");
Expect(1, 11648, '\P{Scx=	-Tfng}', "");
Expect(0, 11648, '\P{^Scx=	-Tfng}', "");
Error('\p{Is_Script_Extensions=/a/_-tifinagh}');
Error('\P{Is_Script_Extensions=/a/_-tifinagh}');
Expect(1, 11647, '\p{Is_Script_Extensions=tifinagh}', "");
Expect(0, 11647, '\p{^Is_Script_Extensions=tifinagh}', "");
Expect(0, 11647, '\P{Is_Script_Extensions=tifinagh}', "");
Expect(1, 11647, '\P{^Is_Script_Extensions=tifinagh}', "");
Expect(0, 11648, '\p{Is_Script_Extensions=tifinagh}', "");
Expect(1, 11648, '\p{^Is_Script_Extensions=tifinagh}', "");
Expect(1, 11648, '\P{Is_Script_Extensions=tifinagh}', "");
Expect(0, 11648, '\P{^Is_Script_Extensions=tifinagh}', "");
Expect(1, 11647, '\p{Is_Script_Extensions=_Tifinagh}', "");
Expect(0, 11647, '\p{^Is_Script_Extensions=_Tifinagh}', "");
Expect(0, 11647, '\P{Is_Script_Extensions=_Tifinagh}', "");
Expect(1, 11647, '\P{^Is_Script_Extensions=_Tifinagh}', "");
Expect(0, 11648, '\p{Is_Script_Extensions=_Tifinagh}', "");
Expect(1, 11648, '\p{^Is_Script_Extensions=_Tifinagh}', "");
Expect(1, 11648, '\P{Is_Script_Extensions=_Tifinagh}', "");
Expect(0, 11648, '\P{^Is_Script_Extensions=_Tifinagh}', "");
Error('\p{Is_Scx=	/a/TFNG}');
Error('\P{Is_Scx=	/a/TFNG}');
Expect(1, 11647, '\p{Is_Scx=tfng}', "");
Expect(0, 11647, '\p{^Is_Scx=tfng}', "");
Expect(0, 11647, '\P{Is_Scx=tfng}', "");
Expect(1, 11647, '\P{^Is_Scx=tfng}', "");
Expect(0, 11648, '\p{Is_Scx=tfng}', "");
Expect(1, 11648, '\p{^Is_Scx=tfng}', "");
Expect(1, 11648, '\P{Is_Scx=tfng}', "");
Expect(0, 11648, '\P{^Is_Scx=tfng}', "");
Expect(1, 11647, '\p{Is_Scx=	-Tfng}', "");
Expect(0, 11647, '\p{^Is_Scx=	-Tfng}', "");
Expect(0, 11647, '\P{Is_Scx=	-Tfng}', "");
Expect(1, 11647, '\P{^Is_Scx=	-Tfng}', "");
Expect(0, 11648, '\p{Is_Scx=	-Tfng}', "");
Expect(1, 11648, '\p{^Is_Scx=	-Tfng}', "");
Expect(1, 11648, '\P{Is_Scx=	-Tfng}', "");
Expect(0, 11648, '\P{^Is_Scx=	-Tfng}', "");
Error('\p{Script_Extensions= :=Sinhala}');
Error('\P{Script_Extensions= :=Sinhala}');
Expect(1, 3572, '\p{Script_Extensions=sinhala}', "");
Expect(0, 3572, '\p{^Script_Extensions=sinhala}', "");
Expect(0, 3572, '\P{Script_Extensions=sinhala}', "");
Expect(1, 3572, '\P{^Script_Extensions=sinhala}', "");
Expect(0, 3573, '\p{Script_Extensions=sinhala}', "");
Expect(1, 3573, '\p{^Script_Extensions=sinhala}', "");
Expect(1, 3573, '\P{Script_Extensions=sinhala}', "");
Expect(0, 3573, '\P{^Script_Extensions=sinhala}', "");
Expect(1, 3572, '\p{Script_Extensions=_SINHALA}', "");
Expect(0, 3572, '\p{^Script_Extensions=_SINHALA}', "");
Expect(0, 3572, '\P{Script_Extensions=_SINHALA}', "");
Expect(1, 3572, '\P{^Script_Extensions=_SINHALA}', "");
Expect(0, 3573, '\p{Script_Extensions=_SINHALA}', "");
Expect(1, 3573, '\p{^Script_Extensions=_SINHALA}', "");
Expect(1, 3573, '\P{Script_Extensions=_SINHALA}', "");
Expect(0, 3573, '\P{^Script_Extensions=_SINHALA}', "");
Error('\p{Scx=/a/_-SINH}');
Error('\P{Scx=/a/_-SINH}');
Expect(1, 3572, '\p{Scx=sinh}', "");
Expect(0, 3572, '\p{^Scx=sinh}', "");
Expect(0, 3572, '\P{Scx=sinh}', "");
Expect(1, 3572, '\P{^Scx=sinh}', "");
Expect(0, 3573, '\p{Scx=sinh}', "");
Expect(1, 3573, '\p{^Scx=sinh}', "");
Expect(1, 3573, '\P{Scx=sinh}', "");
Expect(0, 3573, '\P{^Scx=sinh}', "");
Expect(1, 3572, '\p{Scx= 	Sinh}', "");
Expect(0, 3572, '\p{^Scx= 	Sinh}', "");
Expect(0, 3572, '\P{Scx= 	Sinh}', "");
Expect(1, 3572, '\P{^Scx= 	Sinh}', "");
Expect(0, 3573, '\p{Scx= 	Sinh}', "");
Expect(1, 3573, '\p{^Scx= 	Sinh}', "");
Expect(1, 3573, '\P{Scx= 	Sinh}', "");
Expect(0, 3573, '\P{^Scx= 	Sinh}', "");
Error('\p{Is_Script_Extensions=-_SINHALA/a/}');
Error('\P{Is_Script_Extensions=-_SINHALA/a/}');
Expect(1, 3572, '\p{Is_Script_Extensions=sinhala}', "");
Expect(0, 3572, '\p{^Is_Script_Extensions=sinhala}', "");
Expect(0, 3572, '\P{Is_Script_Extensions=sinhala}', "");
Expect(1, 3572, '\P{^Is_Script_Extensions=sinhala}', "");
Expect(0, 3573, '\p{Is_Script_Extensions=sinhala}', "");
Expect(1, 3573, '\p{^Is_Script_Extensions=sinhala}', "");
Expect(1, 3573, '\P{Is_Script_Extensions=sinhala}', "");
Expect(0, 3573, '\P{^Is_Script_Extensions=sinhala}', "");
Expect(1, 3572, '\p{Is_Script_Extensions=-sinhala}', "");
Expect(0, 3572, '\p{^Is_Script_Extensions=-sinhala}', "");
Expect(0, 3572, '\P{Is_Script_Extensions=-sinhala}', "");
Expect(1, 3572, '\P{^Is_Script_Extensions=-sinhala}', "");
Expect(0, 3573, '\p{Is_Script_Extensions=-sinhala}', "");
Expect(1, 3573, '\p{^Is_Script_Extensions=-sinhala}', "");
Expect(1, 3573, '\P{Is_Script_Extensions=-sinhala}', "");
Expect(0, 3573, '\P{^Is_Script_Extensions=-sinhala}', "");
Error('\p{Is_Scx=_/a/sinh}');
Error('\P{Is_Scx=_/a/sinh}');
Expect(1, 3572, '\p{Is_Scx=sinh}', "");
Expect(0, 3572, '\p{^Is_Scx=sinh}', "");
Expect(0, 3572, '\P{Is_Scx=sinh}', "");
Expect(1, 3572, '\P{^Is_Scx=sinh}', "");
Expect(0, 3573, '\p{Is_Scx=sinh}', "");
Expect(1, 3573, '\p{^Is_Scx=sinh}', "");
Expect(1, 3573, '\P{Is_Scx=sinh}', "");
Expect(0, 3573, '\P{^Is_Scx=sinh}', "");
Expect(1, 3572, '\p{Is_Scx:	_Sinh}', "");
Expect(0, 3572, '\p{^Is_Scx:	_Sinh}', "");
Expect(0, 3572, '\P{Is_Scx:	_Sinh}', "");
Expect(1, 3572, '\P{^Is_Scx:	_Sinh}', "");
Expect(0, 3573, '\p{Is_Scx:	_Sinh}', "");
Expect(1, 3573, '\p{^Is_Scx:	_Sinh}', "");
Expect(1, 3573, '\P{Is_Scx:	_Sinh}', "");
Expect(0, 3573, '\P{^Is_Scx:	_Sinh}', "");
Error('\p{Script_Extensions=:=Miao}');
Error('\P{Script_Extensions=:=Miao}');
Expect(1, 94111, '\p{Script_Extensions=miao}', "");
Expect(0, 94111, '\p{^Script_Extensions=miao}', "");
Expect(0, 94111, '\P{Script_Extensions=miao}', "");
Expect(1, 94111, '\P{^Script_Extensions=miao}', "");
Expect(0, 94112, '\p{Script_Extensions=miao}', "");
Expect(1, 94112, '\p{^Script_Extensions=miao}', "");
Expect(1, 94112, '\P{Script_Extensions=miao}', "");
Expect(0, 94112, '\P{^Script_Extensions=miao}', "");
Expect(1, 94111, '\p{Script_Extensions=_	miao}', "");
Expect(0, 94111, '\p{^Script_Extensions=_	miao}', "");
Expect(0, 94111, '\P{Script_Extensions=_	miao}', "");
Expect(1, 94111, '\P{^Script_Extensions=_	miao}', "");
Expect(0, 94112, '\p{Script_Extensions=_	miao}', "");
Expect(1, 94112, '\p{^Script_Extensions=_	miao}', "");
Expect(1, 94112, '\P{Script_Extensions=_	miao}', "");
Expect(0, 94112, '\P{^Script_Extensions=_	miao}', "");
Error('\p{Scx= /a/PLRD}');
Error('\P{Scx= /a/PLRD}');
Expect(1, 94111, '\p{Scx=plrd}', "");
Expect(0, 94111, '\p{^Scx=plrd}', "");
Expect(0, 94111, '\P{Scx=plrd}', "");
Expect(1, 94111, '\P{^Scx=plrd}', "");
Expect(0, 94112, '\p{Scx=plrd}', "");
Expect(1, 94112, '\p{^Scx=plrd}', "");
Expect(1, 94112, '\P{Scx=plrd}', "");
Expect(0, 94112, '\P{^Scx=plrd}', "");
Expect(1, 94111, '\p{Scx=	_PLRD}', "");
Expect(0, 94111, '\p{^Scx=	_PLRD}', "");
Expect(0, 94111, '\P{Scx=	_PLRD}', "");
Expect(1, 94111, '\P{^Scx=	_PLRD}', "");
Expect(0, 94112, '\p{Scx=	_PLRD}', "");
Expect(1, 94112, '\p{^Scx=	_PLRD}', "");
Expect(1, 94112, '\P{Scx=	_PLRD}', "");
Expect(0, 94112, '\P{^Scx=	_PLRD}', "");
Error('\p{Is_Script_Extensions=:= miao}');
Error('\P{Is_Script_Extensions=:= miao}');
Expect(1, 94111, '\p{Is_Script_Extensions:miao}', "");
Expect(0, 94111, '\p{^Is_Script_Extensions:miao}', "");
Expect(0, 94111, '\P{Is_Script_Extensions:miao}', "");
Expect(1, 94111, '\P{^Is_Script_Extensions:miao}', "");
Expect(0, 94112, '\p{Is_Script_Extensions:miao}', "");
Expect(1, 94112, '\p{^Is_Script_Extensions:miao}', "");
Expect(1, 94112, '\P{Is_Script_Extensions:miao}', "");
Expect(0, 94112, '\P{^Is_Script_Extensions:miao}', "");
Expect(1, 94111, '\p{Is_Script_Extensions=		miao}', "");
Expect(0, 94111, '\p{^Is_Script_Extensions=		miao}', "");
Expect(0, 94111, '\P{Is_Script_Extensions=		miao}', "");
Expect(1, 94111, '\P{^Is_Script_Extensions=		miao}', "");
Expect(0, 94112, '\p{Is_Script_Extensions=		miao}', "");
Expect(1, 94112, '\p{^Is_Script_Extensions=		miao}', "");
Expect(1, 94112, '\P{Is_Script_Extensions=		miao}', "");
Expect(0, 94112, '\P{^Is_Script_Extensions=		miao}', "");
Error('\p{Is_Scx:_ plrd/a/}');
Error('\P{Is_Scx:_ plrd/a/}');
Expect(1, 94111, '\p{Is_Scx: plrd}', "");
Expect(0, 94111, '\p{^Is_Scx: plrd}', "");
Expect(0, 94111, '\P{Is_Scx: plrd}', "");
Expect(1, 94111, '\P{^Is_Scx: plrd}', "");
Expect(0, 94112, '\p{Is_Scx: plrd}', "");
Expect(1, 94112, '\p{^Is_Scx: plrd}', "");
Expect(1, 94112, '\P{Is_Scx: plrd}', "");
Expect(0, 94112, '\P{^Is_Scx: plrd}', "");
Expect(1, 94111, '\p{Is_Scx=-PLRD}', "");
Expect(0, 94111, '\p{^Is_Scx=-PLRD}', "");
Expect(0, 94111, '\P{Is_Scx=-PLRD}', "");
Expect(1, 94111, '\P{^Is_Scx=-PLRD}', "");
Expect(0, 94112, '\p{Is_Scx=-PLRD}', "");
Expect(1, 94112, '\p{^Is_Scx=-PLRD}', "");
Expect(1, 94112, '\P{Is_Scx=-PLRD}', "");
Expect(0, 94112, '\P{^Is_Scx=-PLRD}', "");
Error('\p{Script_Extensions=	:=linear_b}');
Error('\P{Script_Extensions=	:=linear_b}');
Expect(1, 65855, '\p{Script_Extensions=linearb}', "");
Expect(0, 65855, '\p{^Script_Extensions=linearb}', "");
Expect(0, 65855, '\P{Script_Extensions=linearb}', "");
Expect(1, 65855, '\P{^Script_Extensions=linearb}', "");
Expect(0, 65856, '\p{Script_Extensions=linearb}', "");
Expect(1, 65856, '\p{^Script_Extensions=linearb}', "");
Expect(1, 65856, '\P{Script_Extensions=linearb}', "");
Expect(0, 65856, '\P{^Script_Extensions=linearb}', "");
Expect(1, 65855, '\p{Script_Extensions= -LINEAR_B}', "");
Expect(0, 65855, '\p{^Script_Extensions= -LINEAR_B}', "");
Expect(0, 65855, '\P{Script_Extensions= -LINEAR_B}', "");
Expect(1, 65855, '\P{^Script_Extensions= -LINEAR_B}', "");
Expect(0, 65856, '\p{Script_Extensions= -LINEAR_B}', "");
Expect(1, 65856, '\p{^Script_Extensions= -LINEAR_B}', "");
Expect(1, 65856, '\P{Script_Extensions= -LINEAR_B}', "");
Expect(0, 65856, '\P{^Script_Extensions= -LINEAR_B}', "");
Error('\p{Scx=:=	_Linb}');
Error('\P{Scx=:=	_Linb}');
Expect(1, 65855, '\p{Scx=linb}', "");
Expect(0, 65855, '\p{^Scx=linb}', "");
Expect(0, 65855, '\P{Scx=linb}', "");
Expect(1, 65855, '\P{^Scx=linb}', "");
Expect(0, 65856, '\p{Scx=linb}', "");
Expect(1, 65856, '\p{^Scx=linb}', "");
Expect(1, 65856, '\P{Scx=linb}', "");
Expect(0, 65856, '\P{^Scx=linb}', "");
Expect(1, 65855, '\p{Scx=__linb}', "");
Expect(0, 65855, '\p{^Scx=__linb}', "");
Expect(0, 65855, '\P{Scx=__linb}', "");
Expect(1, 65855, '\P{^Scx=__linb}', "");
Expect(0, 65856, '\p{Scx=__linb}', "");
Expect(1, 65856, '\p{^Scx=__linb}', "");
Expect(1, 65856, '\P{Scx=__linb}', "");
Expect(0, 65856, '\P{^Scx=__linb}', "");
Error('\p{Is_Script_Extensions=_:=LINEAR_B}');
Error('\P{Is_Script_Extensions=_:=LINEAR_B}');
Expect(1, 65855, '\p{Is_Script_Extensions=linearb}', "");
Expect(0, 65855, '\p{^Is_Script_Extensions=linearb}', "");
Expect(0, 65855, '\P{Is_Script_Extensions=linearb}', "");
Expect(1, 65855, '\P{^Is_Script_Extensions=linearb}', "");
Expect(0, 65856, '\p{Is_Script_Extensions=linearb}', "");
Expect(1, 65856, '\p{^Is_Script_Extensions=linearb}', "");
Expect(1, 65856, '\P{Is_Script_Extensions=linearb}', "");
Expect(0, 65856, '\P{^Is_Script_Extensions=linearb}', "");
Expect(1, 65855, '\p{Is_Script_Extensions=		Linear_B}', "");
Expect(0, 65855, '\p{^Is_Script_Extensions=		Linear_B}', "");
Expect(0, 65855, '\P{Is_Script_Extensions=		Linear_B}', "");
Expect(1, 65855, '\P{^Is_Script_Extensions=		Linear_B}', "");
Expect(0, 65856, '\p{Is_Script_Extensions=		Linear_B}', "");
Expect(1, 65856, '\p{^Is_Script_Extensions=		Linear_B}', "");
Expect(1, 65856, '\P{Is_Script_Extensions=		Linear_B}', "");
Expect(0, 65856, '\P{^Is_Script_Extensions=		Linear_B}', "");
Error('\p{Is_Scx=	:=linb}');
Error('\P{Is_Scx=	:=linb}');
Expect(1, 65855, '\p{Is_Scx=linb}', "");
Expect(0, 65855, '\p{^Is_Scx=linb}', "");
Expect(0, 65855, '\P{Is_Scx=linb}', "");
Expect(1, 65855, '\P{^Is_Scx=linb}', "");
Expect(0, 65856, '\p{Is_Scx=linb}', "");
Expect(1, 65856, '\p{^Is_Scx=linb}', "");
Expect(1, 65856, '\P{Is_Scx=linb}', "");
Expect(0, 65856, '\P{^Is_Scx=linb}', "");
Expect(1, 65855, '\p{Is_Scx=_Linb}', "");
Expect(0, 65855, '\p{^Is_Scx=_Linb}', "");
Expect(0, 65855, '\P{Is_Scx=_Linb}', "");
Expect(1, 65855, '\P{^Is_Scx=_Linb}', "");
Expect(0, 65856, '\p{Is_Scx=_Linb}', "");
Expect(1, 65856, '\p{^Is_Scx=_Linb}', "");
Expect(1, 65856, '\P{Is_Scx=_Linb}', "");
Expect(0, 65856, '\P{^Is_Scx=_Linb}', "");
Error('\p{Script_Extensions=/a/_SAMARITAN}');
Error('\P{Script_Extensions=/a/_SAMARITAN}');
Expect(1, 2110, '\p{Script_Extensions=samaritan}', "");
Expect(0, 2110, '\p{^Script_Extensions=samaritan}', "");
Expect(0, 2110, '\P{Script_Extensions=samaritan}', "");
Expect(1, 2110, '\P{^Script_Extensions=samaritan}', "");
Expect(0, 2111, '\p{Script_Extensions=samaritan}', "");
Expect(1, 2111, '\p{^Script_Extensions=samaritan}', "");
Expect(1, 2111, '\P{Script_Extensions=samaritan}', "");
Expect(0, 2111, '\P{^Script_Extensions=samaritan}', "");
Expect(1, 2110, '\p{Script_Extensions=_ Samaritan}', "");
Expect(0, 2110, '\p{^Script_Extensions=_ Samaritan}', "");
Expect(0, 2110, '\P{Script_Extensions=_ Samaritan}', "");
Expect(1, 2110, '\P{^Script_Extensions=_ Samaritan}', "");
Expect(0, 2111, '\p{Script_Extensions=_ Samaritan}', "");
Expect(1, 2111, '\p{^Script_Extensions=_ Samaritan}', "");
Expect(1, 2111, '\P{Script_Extensions=_ Samaritan}', "");
Expect(0, 2111, '\P{^Script_Extensions=_ Samaritan}', "");
Error('\p{Scx=:=_	Samr}');
Error('\P{Scx=:=_	Samr}');
Expect(1, 2110, '\p{Scx=samr}', "");
Expect(0, 2110, '\p{^Scx=samr}', "");
Expect(0, 2110, '\P{Scx=samr}', "");
Expect(1, 2110, '\P{^Scx=samr}', "");
Expect(0, 2111, '\p{Scx=samr}', "");
Expect(1, 2111, '\p{^Scx=samr}', "");
Expect(1, 2111, '\P{Scx=samr}', "");
Expect(0, 2111, '\P{^Scx=samr}', "");
Expect(1, 2110, '\p{Scx:  Samr}', "");
Expect(0, 2110, '\p{^Scx:  Samr}', "");
Expect(0, 2110, '\P{Scx:  Samr}', "");
Expect(1, 2110, '\P{^Scx:  Samr}', "");
Expect(0, 2111, '\p{Scx:  Samr}', "");
Expect(1, 2111, '\p{^Scx:  Samr}', "");
Expect(1, 2111, '\P{Scx:  Samr}', "");
Expect(0, 2111, '\P{^Scx:  Samr}', "");
Error('\p{Is_Script_Extensions=:= samaritan}');
Error('\P{Is_Script_Extensions=:= samaritan}');
Expect(1, 2110, '\p{Is_Script_Extensions=samaritan}', "");
Expect(0, 2110, '\p{^Is_Script_Extensions=samaritan}', "");
Expect(0, 2110, '\P{Is_Script_Extensions=samaritan}', "");
Expect(1, 2110, '\P{^Is_Script_Extensions=samaritan}', "");
Expect(0, 2111, '\p{Is_Script_Extensions=samaritan}', "");
Expect(1, 2111, '\p{^Is_Script_Extensions=samaritan}', "");
Expect(1, 2111, '\P{Is_Script_Extensions=samaritan}', "");
Expect(0, 2111, '\P{^Is_Script_Extensions=samaritan}', "");
Expect(1, 2110, '\p{Is_Script_Extensions=--Samaritan}', "");
Expect(0, 2110, '\p{^Is_Script_Extensions=--Samaritan}', "");
Expect(0, 2110, '\P{Is_Script_Extensions=--Samaritan}', "");
Expect(1, 2110, '\P{^Is_Script_Extensions=--Samaritan}', "");
Expect(0, 2111, '\p{Is_Script_Extensions=--Samaritan}', "");
Expect(1, 2111, '\p{^Is_Script_Extensions=--Samaritan}', "");
Expect(1, 2111, '\P{Is_Script_Extensions=--Samaritan}', "");
Expect(0, 2111, '\P{^Is_Script_Extensions=--Samaritan}', "");
Error('\p{Is_Scx=/a/ -samr}');
Error('\P{Is_Scx=/a/ -samr}');
Expect(1, 2110, '\p{Is_Scx=samr}', "");
Expect(0, 2110, '\p{^Is_Scx=samr}', "");
Expect(0, 2110, '\P{Is_Scx=samr}', "");
Expect(1, 2110, '\P{^Is_Scx=samr}', "");
Expect(0, 2111, '\p{Is_Scx=samr}', "");
Expect(1, 2111, '\p{^Is_Scx=samr}', "");
Expect(1, 2111, '\P{Is_Scx=samr}', "");
Expect(0, 2111, '\P{^Is_Scx=samr}', "");
Expect(1, 2110, '\p{Is_Scx:-	samr}', "");
Expect(0, 2110, '\p{^Is_Scx:-	samr}', "");
Expect(0, 2110, '\P{Is_Scx:-	samr}', "");
Expect(1, 2110, '\P{^Is_Scx:-	samr}', "");
Expect(0, 2111, '\p{Is_Scx:-	samr}', "");
Expect(1, 2111, '\p{^Is_Scx:-	samr}', "");
Expect(1, 2111, '\P{Is_Scx:-	samr}', "");
Expect(0, 2111, '\P{^Is_Scx:-	samr}', "");
Error('\p{Script_Extensions=saurashtra:=}');
Error('\P{Script_Extensions=saurashtra:=}');
Expect(1, 43225, '\p{Script_Extensions=saurashtra}', "");
Expect(0, 43225, '\p{^Script_Extensions=saurashtra}', "");
Expect(0, 43225, '\P{Script_Extensions=saurashtra}', "");
Expect(1, 43225, '\P{^Script_Extensions=saurashtra}', "");
Expect(0, 43226, '\p{Script_Extensions=saurashtra}', "");
Expect(1, 43226, '\p{^Script_Extensions=saurashtra}', "");
Expect(1, 43226, '\P{Script_Extensions=saurashtra}', "");
Expect(0, 43226, '\P{^Script_Extensions=saurashtra}', "");
Expect(1, 43225, '\p{Script_Extensions=-	Saurashtra}', "");
Expect(0, 43225, '\p{^Script_Extensions=-	Saurashtra}', "");
Expect(0, 43225, '\P{Script_Extensions=-	Saurashtra}', "");
Expect(1, 43225, '\P{^Script_Extensions=-	Saurashtra}', "");
Expect(0, 43226, '\p{Script_Extensions=-	Saurashtra}', "");
Expect(1, 43226, '\p{^Script_Extensions=-	Saurashtra}', "");
Expect(1, 43226, '\P{Script_Extensions=-	Saurashtra}', "");
Expect(0, 43226, '\P{^Script_Extensions=-	Saurashtra}', "");
Error('\p{Scx=/a/Saur}');
Error('\P{Scx=/a/Saur}');
Expect(1, 43225, '\p{Scx=saur}', "");
Expect(0, 43225, '\p{^Scx=saur}', "");
Expect(0, 43225, '\P{Scx=saur}', "");
Expect(1, 43225, '\P{^Scx=saur}', "");
Expect(0, 43226, '\p{Scx=saur}', "");
Expect(1, 43226, '\p{^Scx=saur}', "");
Expect(1, 43226, '\P{Scx=saur}', "");
Expect(0, 43226, '\P{^Scx=saur}', "");
Expect(1, 43225, '\p{Scx=--SAUR}', "");
Expect(0, 43225, '\p{^Scx=--SAUR}', "");
Expect(0, 43225, '\P{Scx=--SAUR}', "");
Expect(1, 43225, '\P{^Scx=--SAUR}', "");
Expect(0, 43226, '\p{Scx=--SAUR}', "");
Expect(1, 43226, '\p{^Scx=--SAUR}', "");
Expect(1, 43226, '\P{Scx=--SAUR}', "");
Expect(0, 43226, '\P{^Scx=--SAUR}', "");
Error('\p{Is_Script_Extensions=:=_-SAURASHTRA}');
Error('\P{Is_Script_Extensions=:=_-SAURASHTRA}');
Expect(1, 43225, '\p{Is_Script_Extensions: saurashtra}', "");
Expect(0, 43225, '\p{^Is_Script_Extensions: saurashtra}', "");
Expect(0, 43225, '\P{Is_Script_Extensions: saurashtra}', "");
Expect(1, 43225, '\P{^Is_Script_Extensions: saurashtra}', "");
Expect(0, 43226, '\p{Is_Script_Extensions: saurashtra}', "");
Expect(1, 43226, '\p{^Is_Script_Extensions: saurashtra}', "");
Expect(1, 43226, '\P{Is_Script_Extensions: saurashtra}', "");
Expect(0, 43226, '\P{^Is_Script_Extensions: saurashtra}', "");
Expect(1, 43225, '\p{Is_Script_Extensions=-saurashtra}', "");
Expect(0, 43225, '\p{^Is_Script_Extensions=-saurashtra}', "");
Expect(0, 43225, '\P{Is_Script_Extensions=-saurashtra}', "");
Expect(1, 43225, '\P{^Is_Script_Extensions=-saurashtra}', "");
Expect(0, 43226, '\p{Is_Script_Extensions=-saurashtra}', "");
Expect(1, 43226, '\p{^Is_Script_Extensions=-saurashtra}', "");
Expect(1, 43226, '\P{Is_Script_Extensions=-saurashtra}', "");
Expect(0, 43226, '\P{^Is_Script_Extensions=-saurashtra}', "");
Error('\p{Is_Scx:	:=	 Saur}');
Error('\P{Is_Scx:	:=	 Saur}');
Expect(1, 43225, '\p{Is_Scx=saur}', "");
Expect(0, 43225, '\p{^Is_Scx=saur}', "");
Expect(0, 43225, '\P{Is_Scx=saur}', "");
Expect(1, 43225, '\P{^Is_Scx=saur}', "");
Expect(0, 43226, '\p{Is_Scx=saur}', "");
Expect(1, 43226, '\p{^Is_Scx=saur}', "");
Expect(1, 43226, '\P{Is_Scx=saur}', "");
Expect(0, 43226, '\P{^Is_Scx=saur}', "");
Expect(1, 43225, '\p{Is_Scx:  Saur}', "");
Expect(0, 43225, '\p{^Is_Scx:  Saur}', "");
Expect(0, 43225, '\P{Is_Scx:  Saur}', "");
Expect(1, 43225, '\P{^Is_Scx:  Saur}', "");
Expect(0, 43226, '\p{Is_Scx:  Saur}', "");
Expect(1, 43226, '\p{^Is_Scx:  Saur}', "");
Expect(1, 43226, '\P{Is_Scx:  Saur}', "");
Expect(0, 43226, '\P{^Is_Scx:  Saur}', "");
Error('\p{Script_Extensions=	sundanese:=}');
Error('\P{Script_Extensions=	sundanese:=}');
Expect(1, 7367, '\p{Script_Extensions:	sundanese}', "");
Expect(0, 7367, '\p{^Script_Extensions:	sundanese}', "");
Expect(0, 7367, '\P{Script_Extensions:	sundanese}', "");
Expect(1, 7367, '\P{^Script_Extensions:	sundanese}', "");
Expect(0, 7368, '\p{Script_Extensions:	sundanese}', "");
Expect(1, 7368, '\p{^Script_Extensions:	sundanese}', "");
Expect(1, 7368, '\P{Script_Extensions:	sundanese}', "");
Expect(0, 7368, '\P{^Script_Extensions:	sundanese}', "");
Expect(1, 7367, '\p{Script_Extensions=_	sundanese}', "");
Expect(0, 7367, '\p{^Script_Extensions=_	sundanese}', "");
Expect(0, 7367, '\P{Script_Extensions=_	sundanese}', "");
Expect(1, 7367, '\P{^Script_Extensions=_	sundanese}', "");
Expect(0, 7368, '\p{Script_Extensions=_	sundanese}', "");
Expect(1, 7368, '\p{^Script_Extensions=_	sundanese}', "");
Expect(1, 7368, '\P{Script_Extensions=_	sundanese}', "");
Expect(0, 7368, '\P{^Script_Extensions=_	sundanese}', "");
Error('\p{Scx=:=-_Sund}');
Error('\P{Scx=:=-_Sund}');
Expect(1, 7367, '\p{Scx=sund}', "");
Expect(0, 7367, '\p{^Scx=sund}', "");
Expect(0, 7367, '\P{Scx=sund}', "");
Expect(1, 7367, '\P{^Scx=sund}', "");
Expect(0, 7368, '\p{Scx=sund}', "");
Expect(1, 7368, '\p{^Scx=sund}', "");
Expect(1, 7368, '\P{Scx=sund}', "");
Expect(0, 7368, '\P{^Scx=sund}', "");
Expect(1, 7367, '\p{Scx=__Sund}', "");
Expect(0, 7367, '\p{^Scx=__Sund}', "");
Expect(0, 7367, '\P{Scx=__Sund}', "");
Expect(1, 7367, '\P{^Scx=__Sund}', "");
Expect(0, 7368, '\p{Scx=__Sund}', "");
Expect(1, 7368, '\p{^Scx=__Sund}', "");
Expect(1, 7368, '\P{Scx=__Sund}', "");
Expect(0, 7368, '\P{^Scx=__Sund}', "");
Error('\p{Is_Script_Extensions=--sundanese:=}');
Error('\P{Is_Script_Extensions=--sundanese:=}');
Expect(1, 7367, '\p{Is_Script_Extensions=sundanese}', "");
Expect(0, 7367, '\p{^Is_Script_Extensions=sundanese}', "");
Expect(0, 7367, '\P{Is_Script_Extensions=sundanese}', "");
Expect(1, 7367, '\P{^Is_Script_Extensions=sundanese}', "");
Expect(0, 7368, '\p{Is_Script_Extensions=sundanese}', "");
Expect(1, 7368, '\p{^Is_Script_Extensions=sundanese}', "");
Expect(1, 7368, '\P{Is_Script_Extensions=sundanese}', "");
Expect(0, 7368, '\P{^Is_Script_Extensions=sundanese}', "");
Expect(1, 7367, '\p{Is_Script_Extensions=	 Sundanese}', "");
Expect(0, 7367, '\p{^Is_Script_Extensions=	 Sundanese}', "");
Expect(0, 7367, '\P{Is_Script_Extensions=	 Sundanese}', "");
Expect(1, 7367, '\P{^Is_Script_Extensions=	 Sundanese}', "");
Expect(0, 7368, '\p{Is_Script_Extensions=	 Sundanese}', "");
Expect(1, 7368, '\p{^Is_Script_Extensions=	 Sundanese}', "");
Expect(1, 7368, '\P{Is_Script_Extensions=	 Sundanese}', "");
Expect(0, 7368, '\P{^Is_Script_Extensions=	 Sundanese}', "");
Error('\p{Is_Scx=/a/--Sund}');
Error('\P{Is_Scx=/a/--Sund}');
Expect(1, 7367, '\p{Is_Scx:   sund}', "");
Expect(0, 7367, '\p{^Is_Scx:   sund}', "");
Expect(0, 7367, '\P{Is_Scx:   sund}', "");
Expect(1, 7367, '\P{^Is_Scx:   sund}', "");
Expect(0, 7368, '\p{Is_Scx:   sund}', "");
Expect(1, 7368, '\p{^Is_Scx:   sund}', "");
Expect(1, 7368, '\P{Is_Scx:   sund}', "");
Expect(0, 7368, '\P{^Is_Scx:   sund}', "");
Expect(1, 7367, '\p{Is_Scx= Sund}', "");
Expect(0, 7367, '\p{^Is_Scx= Sund}', "");
Expect(0, 7367, '\P{Is_Scx= Sund}', "");
Expect(1, 7367, '\P{^Is_Scx= Sund}', "");
Expect(0, 7368, '\p{Is_Scx= Sund}', "");
Expect(1, 7368, '\p{^Is_Scx= Sund}', "");
Expect(1, 7368, '\P{Is_Scx= Sund}', "");
Expect(0, 7368, '\P{^Is_Scx= Sund}', "");
Error('\p{Script_Extensions=:=- nko}');
Error('\P{Script_Extensions=:=- nko}');
Expect(1, 2042, '\p{Script_Extensions=nko}', "");
Expect(0, 2042, '\p{^Script_Extensions=nko}', "");
Expect(0, 2042, '\P{Script_Extensions=nko}', "");
Expect(1, 2042, '\P{^Script_Extensions=nko}', "");
Expect(0, 2043, '\p{Script_Extensions=nko}', "");
Expect(1, 2043, '\p{^Script_Extensions=nko}', "");
Expect(1, 2043, '\P{Script_Extensions=nko}', "");
Expect(0, 2043, '\P{^Script_Extensions=nko}', "");
Expect(1, 2042, '\p{Script_Extensions=- NKO}', "");
Expect(0, 2042, '\p{^Script_Extensions=- NKO}', "");
Expect(0, 2042, '\P{Script_Extensions=- NKO}', "");
Expect(1, 2042, '\P{^Script_Extensions=- NKO}', "");
Expect(0, 2043, '\p{Script_Extensions=- NKO}', "");
Expect(1, 2043, '\p{^Script_Extensions=- NKO}', "");
Expect(1, 2043, '\P{Script_Extensions=- NKO}', "");
Expect(0, 2043, '\P{^Script_Extensions=- NKO}', "");
Error('\p{Scx=_	NKOO:=}');
Error('\P{Scx=_	NKOO:=}');
Expect(1, 2042, '\p{Scx=nkoo}', "");
Expect(0, 2042, '\p{^Scx=nkoo}', "");
Expect(0, 2042, '\P{Scx=nkoo}', "");
Expect(1, 2042, '\P{^Scx=nkoo}', "");
Expect(0, 2043, '\p{Scx=nkoo}', "");
Expect(1, 2043, '\p{^Scx=nkoo}', "");
Expect(1, 2043, '\P{Scx=nkoo}', "");
Expect(0, 2043, '\P{^Scx=nkoo}', "");
Expect(1, 2042, '\p{Scx= -Nkoo}', "");
Expect(0, 2042, '\p{^Scx= -Nkoo}', "");
Expect(0, 2042, '\P{Scx= -Nkoo}', "");
Expect(1, 2042, '\P{^Scx= -Nkoo}', "");
Expect(0, 2043, '\p{Scx= -Nkoo}', "");
Expect(1, 2043, '\p{^Scx= -Nkoo}', "");
Expect(1, 2043, '\P{Scx= -Nkoo}', "");
Expect(0, 2043, '\P{^Scx= -Nkoo}', "");
Error('\p{Is_Script_Extensions=_	nko/a/}');
Error('\P{Is_Script_Extensions=_	nko/a/}');
Expect(1, 2042, '\p{Is_Script_Extensions=nko}', "");
Expect(0, 2042, '\p{^Is_Script_Extensions=nko}', "");
Expect(0, 2042, '\P{Is_Script_Extensions=nko}', "");
Expect(1, 2042, '\P{^Is_Script_Extensions=nko}', "");
Expect(0, 2043, '\p{Is_Script_Extensions=nko}', "");
Expect(1, 2043, '\p{^Is_Script_Extensions=nko}', "");
Expect(1, 2043, '\P{Is_Script_Extensions=nko}', "");
Expect(0, 2043, '\P{^Is_Script_Extensions=nko}', "");
Expect(1, 2042, '\p{Is_Script_Extensions=  nko}', "");
Expect(0, 2042, '\p{^Is_Script_Extensions=  nko}', "");
Expect(0, 2042, '\P{Is_Script_Extensions=  nko}', "");
Expect(1, 2042, '\P{^Is_Script_Extensions=  nko}', "");
Expect(0, 2043, '\p{Is_Script_Extensions=  nko}', "");
Expect(1, 2043, '\p{^Is_Script_Extensions=  nko}', "");
Expect(1, 2043, '\P{Is_Script_Extensions=  nko}', "");
Expect(0, 2043, '\P{^Is_Script_Extensions=  nko}', "");
Error('\p{Is_Scx:   NKOO/a/}');
Error('\P{Is_Scx:   NKOO/a/}');
Expect(1, 2042, '\p{Is_Scx=nkoo}', "");
Expect(0, 2042, '\p{^Is_Scx=nkoo}', "");
Expect(0, 2042, '\P{Is_Scx=nkoo}', "");
Expect(1, 2042, '\P{^Is_Scx=nkoo}', "");
Expect(0, 2043, '\p{Is_Scx=nkoo}', "");
Expect(1, 2043, '\p{^Is_Scx=nkoo}', "");
Expect(1, 2043, '\P{Is_Scx=nkoo}', "");
Expect(0, 2043, '\P{^Is_Scx=nkoo}', "");
Expect(1, 2042, '\p{Is_Scx=--Nkoo}', "");
Expect(0, 2042, '\p{^Is_Scx=--Nkoo}', "");
Expect(0, 2042, '\P{Is_Scx=--Nkoo}', "");
Expect(1, 2042, '\P{^Is_Scx=--Nkoo}', "");
Expect(0, 2043, '\p{Is_Scx=--Nkoo}', "");
Expect(1, 2043, '\p{^Is_Scx=--Nkoo}', "");
Expect(1, 2043, '\P{Is_Scx=--Nkoo}', "");
Expect(0, 2043, '\P{^Is_Scx=--Nkoo}', "");
Error('\p{Script_Extensions: -	han/a/}');
Error('\P{Script_Extensions: -	han/a/}');
Expect(1, 195101, '\p{Script_Extensions:han}', "");
Expect(0, 195101, '\p{^Script_Extensions:han}', "");
Expect(0, 195101, '\P{Script_Extensions:han}', "");
Expect(1, 195101, '\P{^Script_Extensions:han}', "");
Expect(0, 195102, '\p{Script_Extensions:han}', "");
Expect(1, 195102, '\p{^Script_Extensions:han}', "");
Expect(1, 195102, '\P{Script_Extensions:han}', "");
Expect(0, 195102, '\P{^Script_Extensions:han}', "");
Expect(1, 195101, '\p{Script_Extensions=Han}', "");
Expect(0, 195101, '\p{^Script_Extensions=Han}', "");
Expect(0, 195101, '\P{Script_Extensions=Han}', "");
Expect(1, 195101, '\P{^Script_Extensions=Han}', "");
Expect(0, 195102, '\p{Script_Extensions=Han}', "");
Expect(1, 195102, '\p{^Script_Extensions=Han}', "");
Expect(1, 195102, '\P{Script_Extensions=Han}', "");
Expect(0, 195102, '\P{^Script_Extensions=Han}', "");
Error('\p{Scx=:=hani}');
Error('\P{Scx=:=hani}');
Expect(1, 195101, '\p{Scx=hani}', "");
Expect(0, 195101, '\p{^Scx=hani}', "");
Expect(0, 195101, '\P{Scx=hani}', "");
Expect(1, 195101, '\P{^Scx=hani}', "");
Expect(0, 195102, '\p{Scx=hani}', "");
Expect(1, 195102, '\p{^Scx=hani}', "");
Expect(1, 195102, '\P{Scx=hani}', "");
Expect(0, 195102, '\P{^Scx=hani}', "");
Expect(1, 195101, '\p{Scx=__HANI}', "");
Expect(0, 195101, '\p{^Scx=__HANI}', "");
Expect(0, 195101, '\P{Scx=__HANI}', "");
Expect(1, 195101, '\P{^Scx=__HANI}', "");
Expect(0, 195102, '\p{Scx=__HANI}', "");
Expect(1, 195102, '\p{^Scx=__HANI}', "");
Expect(1, 195102, '\P{Scx=__HANI}', "");
Expect(0, 195102, '\P{^Scx=__HANI}', "");
Error('\p{Is_Script_Extensions=_-HAN/a/}');
Error('\P{Is_Script_Extensions=_-HAN/a/}');
Expect(1, 195101, '\p{Is_Script_Extensions=han}', "");
Expect(0, 195101, '\p{^Is_Script_Extensions=han}', "");
Expect(0, 195101, '\P{Is_Script_Extensions=han}', "");
Expect(1, 195101, '\P{^Is_Script_Extensions=han}', "");
Expect(0, 195102, '\p{Is_Script_Extensions=han}', "");
Expect(1, 195102, '\p{^Is_Script_Extensions=han}', "");
Expect(1, 195102, '\P{Is_Script_Extensions=han}', "");
Expect(0, 195102, '\P{^Is_Script_Extensions=han}', "");
Expect(1, 195101, '\p{Is_Script_Extensions=__Han}', "");
Expect(0, 195101, '\p{^Is_Script_Extensions=__Han}', "");
Expect(0, 195101, '\P{Is_Script_Extensions=__Han}', "");
Expect(1, 195101, '\P{^Is_Script_Extensions=__Han}', "");
Expect(0, 195102, '\p{Is_Script_Extensions=__Han}', "");
Expect(1, 195102, '\p{^Is_Script_Extensions=__Han}', "");
Expect(1, 195102, '\P{Is_Script_Extensions=__Han}', "");
Expect(0, 195102, '\P{^Is_Script_Extensions=__Han}', "");
Error('\p{Is_Scx:   	:=Hani}');
Error('\P{Is_Scx:   	:=Hani}');
Expect(1, 195101, '\p{Is_Scx=hani}', "");
Expect(0, 195101, '\p{^Is_Scx=hani}', "");
Expect(0, 195101, '\P{Is_Scx=hani}', "");
Expect(1, 195101, '\P{^Is_Scx=hani}', "");
Expect(0, 195102, '\p{Is_Scx=hani}', "");
Expect(1, 195102, '\p{^Is_Scx=hani}', "");
Expect(1, 195102, '\P{Is_Scx=hani}', "");
Expect(0, 195102, '\P{^Is_Scx=hani}', "");
Expect(1, 195101, '\p{Is_Scx=-Hani}', "");
Expect(0, 195101, '\p{^Is_Scx=-Hani}', "");
Expect(0, 195101, '\P{Is_Scx=-Hani}', "");
Expect(1, 195101, '\P{^Is_Scx=-Hani}', "");
Expect(0, 195102, '\p{Is_Scx=-Hani}', "");
Expect(1, 195102, '\p{^Is_Scx=-Hani}', "");
Expect(1, 195102, '\P{Is_Scx=-Hani}', "");
Expect(0, 195102, '\P{^Is_Scx=-Hani}', "");
Error('\p{Script_Extensions:	:=Takri}');
Error('\P{Script_Extensions:	:=Takri}');
Expect(1, 71369, '\p{Script_Extensions=takri}', "");
Expect(0, 71369, '\p{^Script_Extensions=takri}', "");
Expect(0, 71369, '\P{Script_Extensions=takri}', "");
Expect(1, 71369, '\P{^Script_Extensions=takri}', "");
Expect(0, 71370, '\p{Script_Extensions=takri}', "");
Expect(1, 71370, '\p{^Script_Extensions=takri}', "");
Expect(1, 71370, '\P{Script_Extensions=takri}', "");
Expect(0, 71370, '\P{^Script_Extensions=takri}', "");
Expect(1, 71369, '\p{Script_Extensions=	Takri}', "");
Expect(0, 71369, '\p{^Script_Extensions=	Takri}', "");
Expect(0, 71369, '\P{Script_Extensions=	Takri}', "");
Expect(1, 71369, '\P{^Script_Extensions=	Takri}', "");
Expect(0, 71370, '\p{Script_Extensions=	Takri}', "");
Expect(1, 71370, '\p{^Script_Extensions=	Takri}', "");
Expect(1, 71370, '\P{Script_Extensions=	Takri}', "");
Expect(0, 71370, '\P{^Script_Extensions=	Takri}', "");
Error('\p{Scx=	_Takr:=}');
Error('\P{Scx=	_Takr:=}');
Expect(1, 71369, '\p{Scx=takr}', "");
Expect(0, 71369, '\p{^Scx=takr}', "");
Expect(0, 71369, '\P{Scx=takr}', "");
Expect(1, 71369, '\P{^Scx=takr}', "");
Expect(0, 71370, '\p{Scx=takr}', "");
Expect(1, 71370, '\p{^Scx=takr}', "");
Expect(1, 71370, '\P{Scx=takr}', "");
Expect(0, 71370, '\P{^Scx=takr}', "");
Expect(1, 71369, '\p{Scx=	_TAKR}', "");
Expect(0, 71369, '\p{^Scx=	_TAKR}', "");
Expect(0, 71369, '\P{Scx=	_TAKR}', "");
Expect(1, 71369, '\P{^Scx=	_TAKR}', "");
Expect(0, 71370, '\p{Scx=	_TAKR}', "");
Expect(1, 71370, '\p{^Scx=	_TAKR}', "");
Expect(1, 71370, '\P{Scx=	_TAKR}', "");
Expect(0, 71370, '\P{^Scx=	_TAKR}', "");
Error('\p{Is_Script_Extensions=-/a/takri}');
Error('\P{Is_Script_Extensions=-/a/takri}');
Expect(1, 71369, '\p{Is_Script_Extensions=takri}', "");
Expect(0, 71369, '\p{^Is_Script_Extensions=takri}', "");
Expect(0, 71369, '\P{Is_Script_Extensions=takri}', "");
Expect(1, 71369, '\P{^Is_Script_Extensions=takri}', "");
Expect(0, 71370, '\p{Is_Script_Extensions=takri}', "");
Expect(1, 71370, '\p{^Is_Script_Extensions=takri}', "");
Expect(1, 71370, '\P{Is_Script_Extensions=takri}', "");
Expect(0, 71370, '\P{^Is_Script_Extensions=takri}', "");
Expect(1, 71369, '\p{Is_Script_Extensions=		TAKRI}', "");
Expect(0, 71369, '\p{^Is_Script_Extensions=		TAKRI}', "");
Expect(0, 71369, '\P{Is_Script_Extensions=		TAKRI}', "");
Expect(1, 71369, '\P{^Is_Script_Extensions=		TAKRI}', "");
Expect(0, 71370, '\p{Is_Script_Extensions=		TAKRI}', "");
Expect(1, 71370, '\p{^Is_Script_Extensions=		TAKRI}', "");
Expect(1, 71370, '\P{Is_Script_Extensions=		TAKRI}', "");
Expect(0, 71370, '\P{^Is_Script_Extensions=		TAKRI}', "");
Error('\p{Is_Scx=:= -TAKR}');
Error('\P{Is_Scx=:= -TAKR}');
Expect(1, 71369, '\p{Is_Scx:   takr}', "");
Expect(0, 71369, '\p{^Is_Scx:   takr}', "");
Expect(0, 71369, '\P{Is_Scx:   takr}', "");
Expect(1, 71369, '\P{^Is_Scx:   takr}', "");
Expect(0, 71370, '\p{Is_Scx:   takr}', "");
Expect(1, 71370, '\p{^Is_Scx:   takr}', "");
Expect(1, 71370, '\P{Is_Scx:   takr}', "");
Expect(0, 71370, '\P{^Is_Scx:   takr}', "");
Expect(1, 71369, '\p{Is_Scx:	_TAKR}', "");
Expect(0, 71369, '\p{^Is_Scx:	_TAKR}', "");
Expect(0, 71369, '\P{Is_Scx:	_TAKR}', "");
Expect(1, 71369, '\P{^Is_Scx:	_TAKR}', "");
Expect(0, 71370, '\p{Is_Scx:	_TAKR}', "");
Expect(1, 71370, '\p{^Is_Scx:	_TAKR}', "");
Expect(1, 71370, '\P{Is_Scx:	_TAKR}', "");
Expect(0, 71370, '\P{^Is_Scx:	_TAKR}', "");
Error('\p{Script_Extensions=	_greek/a/}');
Error('\P{Script_Extensions=	_greek/a/}');
Expect(1, 119365, '\p{Script_Extensions=greek}', "");
Expect(0, 119365, '\p{^Script_Extensions=greek}', "");
Expect(0, 119365, '\P{Script_Extensions=greek}', "");
Expect(1, 119365, '\P{^Script_Extensions=greek}', "");
Expect(0, 119366, '\p{Script_Extensions=greek}', "");
Expect(1, 119366, '\p{^Script_Extensions=greek}', "");
Expect(1, 119366, '\P{Script_Extensions=greek}', "");
Expect(0, 119366, '\P{^Script_Extensions=greek}', "");
Expect(1, 119365, '\p{Script_Extensions:   Greek}', "");
Expect(0, 119365, '\p{^Script_Extensions:   Greek}', "");
Expect(0, 119365, '\P{Script_Extensions:   Greek}', "");
Expect(1, 119365, '\P{^Script_Extensions:   Greek}', "");
Expect(0, 119366, '\p{Script_Extensions:   Greek}', "");
Expect(1, 119366, '\p{^Script_Extensions:   Greek}', "");
Expect(1, 119366, '\P{Script_Extensions:   Greek}', "");
Expect(0, 119366, '\P{^Script_Extensions:   Greek}', "");
Error('\p{Scx=:=_grek}');
Error('\P{Scx=:=_grek}');
Expect(1, 119365, '\p{Scx=grek}', "");
Expect(0, 119365, '\p{^Scx=grek}', "");
Expect(0, 119365, '\P{Scx=grek}', "");
Expect(1, 119365, '\P{^Scx=grek}', "");
Expect(0, 119366, '\p{Scx=grek}', "");
Expect(1, 119366, '\p{^Scx=grek}', "");
Expect(1, 119366, '\P{Scx=grek}', "");
Expect(0, 119366, '\P{^Scx=grek}', "");
Expect(1, 119365, '\p{Scx=__grek}', "");
Expect(0, 119365, '\p{^Scx=__grek}', "");
Expect(0, 119365, '\P{Scx=__grek}', "");
Expect(1, 119365, '\P{^Scx=__grek}', "");
Expect(0, 119366, '\p{Scx=__grek}', "");
Expect(1, 119366, '\p{^Scx=__grek}', "");
Expect(1, 119366, '\P{Scx=__grek}', "");
Expect(0, 119366, '\P{^Scx=__grek}', "");
Error('\p{Is_Script_Extensions=  Greek/a/}');
Error('\P{Is_Script_Extensions=  Greek/a/}');
Expect(1, 119365, '\p{Is_Script_Extensions=greek}', "");
Expect(0, 119365, '\p{^Is_Script_Extensions=greek}', "");
Expect(0, 119365, '\P{Is_Script_Extensions=greek}', "");
Expect(1, 119365, '\P{^Is_Script_Extensions=greek}', "");
Expect(0, 119366, '\p{Is_Script_Extensions=greek}', "");
Expect(1, 119366, '\p{^Is_Script_Extensions=greek}', "");
Expect(1, 119366, '\P{Is_Script_Extensions=greek}', "");
Expect(0, 119366, '\P{^Is_Script_Extensions=greek}', "");
Expect(1, 119365, '\p{Is_Script_Extensions= 	GREEK}', "");
Expect(0, 119365, '\p{^Is_Script_Extensions= 	GREEK}', "");
Expect(0, 119365, '\P{Is_Script_Extensions= 	GREEK}', "");
Expect(1, 119365, '\P{^Is_Script_Extensions= 	GREEK}', "");
Expect(0, 119366, '\p{Is_Script_Extensions= 	GREEK}', "");
Expect(1, 119366, '\p{^Is_Script_Extensions= 	GREEK}', "");
Expect(1, 119366, '\P{Is_Script_Extensions= 	GREEK}', "");
Expect(0, 119366, '\P{^Is_Script_Extensions= 	GREEK}', "");
Error('\p{Is_Scx: -/a/Grek}');
Error('\P{Is_Scx: -/a/Grek}');
Expect(1, 119365, '\p{Is_Scx=grek}', "");
Expect(0, 119365, '\p{^Is_Scx=grek}', "");
Expect(0, 119365, '\P{Is_Scx=grek}', "");
Expect(1, 119365, '\P{^Is_Scx=grek}', "");
Expect(0, 119366, '\p{Is_Scx=grek}', "");
Expect(1, 119366, '\p{^Is_Scx=grek}', "");
Expect(1, 119366, '\P{Is_Scx=grek}', "");
Expect(0, 119366, '\P{^Is_Scx=grek}', "");
Expect(1, 119365, '\p{Is_Scx=	-Grek}', "");
Expect(0, 119365, '\p{^Is_Scx=	-Grek}', "");
Expect(0, 119365, '\P{Is_Scx=	-Grek}', "");
Expect(1, 119365, '\P{^Is_Scx=	-Grek}', "");
Expect(0, 119366, '\p{Is_Scx=	-Grek}', "");
Expect(1, 119366, '\p{^Is_Scx=	-Grek}', "");
Expect(1, 119366, '\P{Is_Scx=	-Grek}', "");
Expect(0, 119366, '\P{^Is_Scx=	-Grek}', "");
Error('\p{Script_Extensions= :=Tamil}');
Error('\P{Script_Extensions= :=Tamil}');
Expect(1, 3066, '\p{Script_Extensions=tamil}', "");
Expect(0, 3066, '\p{^Script_Extensions=tamil}', "");
Expect(0, 3066, '\P{Script_Extensions=tamil}', "");
Expect(1, 3066, '\P{^Script_Extensions=tamil}', "");
Expect(0, 3067, '\p{Script_Extensions=tamil}', "");
Expect(1, 3067, '\p{^Script_Extensions=tamil}', "");
Expect(1, 3067, '\P{Script_Extensions=tamil}', "");
Expect(0, 3067, '\P{^Script_Extensions=tamil}', "");
Expect(1, 3066, '\p{Script_Extensions= Tamil}', "");
Expect(0, 3066, '\p{^Script_Extensions= Tamil}', "");
Expect(0, 3066, '\P{Script_Extensions= Tamil}', "");
Expect(1, 3066, '\P{^Script_Extensions= Tamil}', "");
Expect(0, 3067, '\p{Script_Extensions= Tamil}', "");
Expect(1, 3067, '\p{^Script_Extensions= Tamil}', "");
Expect(1, 3067, '\P{Script_Extensions= Tamil}', "");
Expect(0, 3067, '\P{^Script_Extensions= Tamil}', "");
Error('\p{Scx:		Taml:=}');
Error('\P{Scx:		Taml:=}');
Expect(1, 3066, '\p{Scx=taml}', "");
Expect(0, 3066, '\p{^Scx=taml}', "");
Expect(0, 3066, '\P{Scx=taml}', "");
Expect(1, 3066, '\P{^Scx=taml}', "");
Expect(0, 3067, '\p{Scx=taml}', "");
Expect(1, 3067, '\p{^Scx=taml}', "");
Expect(1, 3067, '\P{Scx=taml}', "");
Expect(0, 3067, '\P{^Scx=taml}', "");
Expect(1, 3066, '\p{Scx=		TAML}', "");
Expect(0, 3066, '\p{^Scx=		TAML}', "");
Expect(0, 3066, '\P{Scx=		TAML}', "");
Expect(1, 3066, '\P{^Scx=		TAML}', "");
Expect(0, 3067, '\p{Scx=		TAML}', "");
Expect(1, 3067, '\p{^Scx=		TAML}', "");
Expect(1, 3067, '\P{Scx=		TAML}', "");
Expect(0, 3067, '\P{^Scx=		TAML}', "");
Error('\p{Is_Script_Extensions=__Tamil:=}');
Error('\P{Is_Script_Extensions=__Tamil:=}');
Expect(1, 3066, '\p{Is_Script_Extensions=tamil}', "");
Expect(0, 3066, '\p{^Is_Script_Extensions=tamil}', "");
Expect(0, 3066, '\P{Is_Script_Extensions=tamil}', "");
Expect(1, 3066, '\P{^Is_Script_Extensions=tamil}', "");
Expect(0, 3067, '\p{Is_Script_Extensions=tamil}', "");
Expect(1, 3067, '\p{^Is_Script_Extensions=tamil}', "");
Expect(1, 3067, '\P{Is_Script_Extensions=tamil}', "");
Expect(0, 3067, '\P{^Is_Script_Extensions=tamil}', "");
Expect(1, 3066, '\p{Is_Script_Extensions= tamil}', "");
Expect(0, 3066, '\p{^Is_Script_Extensions= tamil}', "");
Expect(0, 3066, '\P{Is_Script_Extensions= tamil}', "");
Expect(1, 3066, '\P{^Is_Script_Extensions= tamil}', "");
Expect(0, 3067, '\p{Is_Script_Extensions= tamil}', "");
Expect(1, 3067, '\p{^Is_Script_Extensions= tamil}', "");
Expect(1, 3067, '\P{Is_Script_Extensions= tamil}', "");
Expect(0, 3067, '\P{^Is_Script_Extensions= tamil}', "");
Error('\p{Is_Scx=Taml:=}');
Error('\P{Is_Scx=Taml:=}');
Expect(1, 3066, '\p{Is_Scx=taml}', "");
Expect(0, 3066, '\p{^Is_Scx=taml}', "");
Expect(0, 3066, '\P{Is_Scx=taml}', "");
Expect(1, 3066, '\P{^Is_Scx=taml}', "");
Expect(0, 3067, '\p{Is_Scx=taml}', "");
Expect(1, 3067, '\p{^Is_Scx=taml}', "");
Expect(1, 3067, '\P{Is_Scx=taml}', "");
Expect(0, 3067, '\P{^Is_Scx=taml}', "");
Expect(1, 3066, '\p{Is_Scx=__TAML}', "");
Expect(0, 3066, '\p{^Is_Scx=__TAML}', "");
Expect(0, 3066, '\P{Is_Scx=__TAML}', "");
Expect(1, 3066, '\P{^Is_Scx=__TAML}', "");
Expect(0, 3067, '\p{Is_Scx=__TAML}', "");
Expect(1, 3067, '\p{^Is_Scx=__TAML}', "");
Expect(1, 3067, '\P{Is_Scx=__TAML}', "");
Expect(0, 3067, '\P{^Is_Scx=__TAML}', "");
Error('\p{Script_Extensions= _kayah_li/a/}');
Error('\P{Script_Extensions= _kayah_li/a/}');
Expect(1, 43311, '\p{Script_Extensions=kayahli}', "");
Expect(0, 43311, '\p{^Script_Extensions=kayahli}', "");
Expect(0, 43311, '\P{Script_Extensions=kayahli}', "");
Expect(1, 43311, '\P{^Script_Extensions=kayahli}', "");
Expect(0, 43312, '\p{Script_Extensions=kayahli}', "");
Expect(1, 43312, '\p{^Script_Extensions=kayahli}', "");
Expect(1, 43312, '\P{Script_Extensions=kayahli}', "");
Expect(0, 43312, '\P{^Script_Extensions=kayahli}', "");
Expect(1, 43311, '\p{Script_Extensions= -KAYAH_LI}', "");
Expect(0, 43311, '\p{^Script_Extensions= -KAYAH_LI}', "");
Expect(0, 43311, '\P{Script_Extensions= -KAYAH_LI}', "");
Expect(1, 43311, '\P{^Script_Extensions= -KAYAH_LI}', "");
Expect(0, 43312, '\p{Script_Extensions= -KAYAH_LI}', "");
Expect(1, 43312, '\p{^Script_Extensions= -KAYAH_LI}', "");
Expect(1, 43312, '\P{Script_Extensions= -KAYAH_LI}', "");
Expect(0, 43312, '\P{^Script_Extensions= -KAYAH_LI}', "");
Error('\p{Scx= _KALI/a/}');
Error('\P{Scx= _KALI/a/}');
Expect(1, 43311, '\p{Scx=kali}', "");
Expect(0, 43311, '\p{^Scx=kali}', "");
Expect(0, 43311, '\P{Scx=kali}', "");
Expect(1, 43311, '\P{^Scx=kali}', "");
Expect(0, 43312, '\p{Scx=kali}', "");
Expect(1, 43312, '\p{^Scx=kali}', "");
Expect(1, 43312, '\P{Scx=kali}', "");
Expect(0, 43312, '\P{^Scx=kali}', "");
Expect(1, 43311, '\p{Scx=	_Kali}', "");
Expect(0, 43311, '\p{^Scx=	_Kali}', "");
Expect(0, 43311, '\P{Scx=	_Kali}', "");
Expect(1, 43311, '\P{^Scx=	_Kali}', "");
Expect(0, 43312, '\p{Scx=	_Kali}', "");
Expect(1, 43312, '\p{^Scx=	_Kali}', "");
Expect(1, 43312, '\P{Scx=	_Kali}', "");
Expect(0, 43312, '\P{^Scx=	_Kali}', "");
Error('\p{Is_Script_Extensions=_ kayah_li:=}');
Error('\P{Is_Script_Extensions=_ kayah_li:=}');
Expect(1, 43311, '\p{Is_Script_Extensions=kayahli}', "");
Expect(0, 43311, '\p{^Is_Script_Extensions=kayahli}', "");
Expect(0, 43311, '\P{Is_Script_Extensions=kayahli}', "");
Expect(1, 43311, '\P{^Is_Script_Extensions=kayahli}', "");
Expect(0, 43312, '\p{Is_Script_Extensions=kayahli}', "");
Expect(1, 43312, '\p{^Is_Script_Extensions=kayahli}', "");
Expect(1, 43312, '\P{Is_Script_Extensions=kayahli}', "");
Expect(0, 43312, '\P{^Is_Script_Extensions=kayahli}', "");
Expect(1, 43311, '\p{Is_Script_Extensions=-Kayah_li}', "");
Expect(0, 43311, '\p{^Is_Script_Extensions=-Kayah_li}', "");
Expect(0, 43311, '\P{Is_Script_Extensions=-Kayah_li}', "");
Expect(1, 43311, '\P{^Is_Script_Extensions=-Kayah_li}', "");
Expect(0, 43312, '\p{Is_Script_Extensions=-Kayah_li}', "");
Expect(1, 43312, '\p{^Is_Script_Extensions=-Kayah_li}', "");
Expect(1, 43312, '\P{Is_Script_Extensions=-Kayah_li}', "");
Expect(0, 43312, '\P{^Is_Script_Extensions=-Kayah_li}', "");
Error('\p{Is_Scx= /a/KALI}');
Error('\P{Is_Scx= /a/KALI}');
Expect(1, 43311, '\p{Is_Scx=kali}', "");
Expect(0, 43311, '\p{^Is_Scx=kali}', "");
Expect(0, 43311, '\P{Is_Scx=kali}', "");
Expect(1, 43311, '\P{^Is_Scx=kali}', "");
Expect(0, 43312, '\p{Is_Scx=kali}', "");
Expect(1, 43312, '\p{^Is_Scx=kali}', "");
Expect(1, 43312, '\P{Is_Scx=kali}', "");
Expect(0, 43312, '\P{^Is_Scx=kali}', "");
Expect(1, 43311, '\p{Is_Scx=	Kali}', "");
Expect(0, 43311, '\p{^Is_Scx=	Kali}', "");
Expect(0, 43311, '\P{Is_Scx=	Kali}', "");
Expect(1, 43311, '\P{^Is_Scx=	Kali}', "");
Expect(0, 43312, '\p{Is_Scx=	Kali}', "");
Expect(1, 43312, '\p{^Is_Scx=	Kali}', "");
Expect(1, 43312, '\P{Is_Scx=	Kali}', "");
Expect(0, 43312, '\P{^Is_Scx=	Kali}', "");
Error('\p{Script_Extensions=	/a/Thai}');
Error('\P{Script_Extensions=	/a/Thai}');
Expect(1, 3675, '\p{Script_Extensions=thai}', "");
Expect(0, 3675, '\p{^Script_Extensions=thai}', "");
Expect(0, 3675, '\P{Script_Extensions=thai}', "");
Expect(1, 3675, '\P{^Script_Extensions=thai}', "");
Expect(0, 3676, '\p{Script_Extensions=thai}', "");
Expect(1, 3676, '\p{^Script_Extensions=thai}', "");
Expect(1, 3676, '\P{Script_Extensions=thai}', "");
Expect(0, 3676, '\P{^Script_Extensions=thai}', "");
Expect(1, 3675, '\p{Script_Extensions= 	Thai}', "");
Expect(0, 3675, '\p{^Script_Extensions= 	Thai}', "");
Expect(0, 3675, '\P{Script_Extensions= 	Thai}', "");
Expect(1, 3675, '\P{^Script_Extensions= 	Thai}', "");
Expect(0, 3676, '\p{Script_Extensions= 	Thai}', "");
Expect(1, 3676, '\p{^Script_Extensions= 	Thai}', "");
Expect(1, 3676, '\P{Script_Extensions= 	Thai}', "");
Expect(0, 3676, '\P{^Script_Extensions= 	Thai}', "");
Error('\p{Scx=-/a/thai}');
Error('\P{Scx=-/a/thai}');
Expect(1, 3675, '\p{Scx=thai}', "");
Expect(0, 3675, '\p{^Scx=thai}', "");
Expect(0, 3675, '\P{Scx=thai}', "");
Expect(1, 3675, '\P{^Scx=thai}', "");
Expect(0, 3676, '\p{Scx=thai}', "");
Expect(1, 3676, '\p{^Scx=thai}', "");
Expect(1, 3676, '\P{Scx=thai}', "");
Expect(0, 3676, '\P{^Scx=thai}', "");
Expect(1, 3675, '\p{Scx=_Thai}', "");
Expect(0, 3675, '\p{^Scx=_Thai}', "");
Expect(0, 3675, '\P{Scx=_Thai}', "");
Expect(1, 3675, '\P{^Scx=_Thai}', "");
Expect(0, 3676, '\p{Scx=_Thai}', "");
Expect(1, 3676, '\p{^Scx=_Thai}', "");
Expect(1, 3676, '\P{Scx=_Thai}', "");
Expect(0, 3676, '\P{^Scx=_Thai}', "");
Error('\p{Is_Script_Extensions=:=-	Thai}');
Error('\P{Is_Script_Extensions=:=-	Thai}');
Expect(1, 3675, '\p{Is_Script_Extensions=thai}', "");
Expect(0, 3675, '\p{^Is_Script_Extensions=thai}', "");
Expect(0, 3675, '\P{Is_Script_Extensions=thai}', "");
Expect(1, 3675, '\P{^Is_Script_Extensions=thai}', "");
Expect(0, 3676, '\p{Is_Script_Extensions=thai}', "");
Expect(1, 3676, '\p{^Is_Script_Extensions=thai}', "");
Expect(1, 3676, '\P{Is_Script_Extensions=thai}', "");
Expect(0, 3676, '\P{^Is_Script_Extensions=thai}', "");
Expect(1, 3675, '\p{Is_Script_Extensions= 	THAI}', "");
Expect(0, 3675, '\p{^Is_Script_Extensions= 	THAI}', "");
Expect(0, 3675, '\P{Is_Script_Extensions= 	THAI}', "");
Expect(1, 3675, '\P{^Is_Script_Extensions= 	THAI}', "");
Expect(0, 3676, '\p{Is_Script_Extensions= 	THAI}', "");
Expect(1, 3676, '\p{^Is_Script_Extensions= 	THAI}', "");
Expect(1, 3676, '\P{Is_Script_Extensions= 	THAI}', "");
Expect(0, 3676, '\P{^Is_Script_Extensions= 	THAI}', "");
Error('\p{Is_Scx=_:=thai}');
Error('\P{Is_Scx=_:=thai}');
Expect(1, 3675, '\p{Is_Scx=thai}', "");
Expect(0, 3675, '\p{^Is_Scx=thai}', "");
Expect(0, 3675, '\P{Is_Scx=thai}', "");
Expect(1, 3675, '\P{^Is_Scx=thai}', "");
Expect(0, 3676, '\p{Is_Scx=thai}', "");
Expect(1, 3676, '\p{^Is_Scx=thai}', "");
Expect(1, 3676, '\P{Is_Scx=thai}', "");
Expect(0, 3676, '\P{^Is_Scx=thai}', "");
Expect(1, 3675, '\p{Is_Scx:_	THAI}', "");
Expect(0, 3675, '\p{^Is_Scx:_	THAI}', "");
Expect(0, 3675, '\P{Is_Scx:_	THAI}', "");
Expect(1, 3675, '\P{^Is_Scx:_	THAI}', "");
Expect(0, 3676, '\p{Is_Scx:_	THAI}', "");
Expect(1, 3676, '\p{^Is_Scx:_	THAI}', "");
Expect(1, 3676, '\P{Is_Scx:_	THAI}', "");
Expect(0, 3676, '\P{^Is_Scx:_	THAI}', "");
Error('\p{NFKC_Quick_Check:		:=Yes}');
Error('\P{NFKC_Quick_Check:		:=Yes}');
Expect(1, 195102, '\p{NFKC_Quick_Check:   yes}', "");
Expect(0, 195102, '\p{^NFKC_Quick_Check:   yes}', "");
Expect(0, 195102, '\P{NFKC_Quick_Check:   yes}', "");
Expect(1, 195102, '\P{^NFKC_Quick_Check:   yes}', "");
Expect(0, 195101, '\p{NFKC_Quick_Check:   yes}', "");
Expect(1, 195101, '\p{^NFKC_Quick_Check:   yes}', "");
Expect(1, 195101, '\P{NFKC_Quick_Check:   yes}', "");
Expect(0, 195101, '\P{^NFKC_Quick_Check:   yes}', "");
Expect(1, 195102, '\p{NFKC_Quick_Check:    -Yes}', "");
Expect(0, 195102, '\p{^NFKC_Quick_Check:    -Yes}', "");
Expect(0, 195102, '\P{NFKC_Quick_Check:    -Yes}', "");
Expect(1, 195102, '\P{^NFKC_Quick_Check:    -Yes}', "");
Expect(0, 195101, '\p{NFKC_Quick_Check:    -Yes}', "");
Expect(1, 195101, '\p{^NFKC_Quick_Check:    -Yes}', "");
Expect(1, 195101, '\P{NFKC_Quick_Check:    -Yes}', "");
Expect(0, 195101, '\P{^NFKC_Quick_Check:    -Yes}', "");
Error('\p{NFKC_QC: :=	 Y}');
Error('\P{NFKC_QC: :=	 Y}');
Expect(1, 195102, '\p{NFKC_QC: y}', "");
Expect(0, 195102, '\p{^NFKC_QC: y}', "");
Expect(0, 195102, '\P{NFKC_QC: y}', "");
Expect(1, 195102, '\P{^NFKC_QC: y}', "");
Expect(0, 195101, '\p{NFKC_QC: y}', "");
Expect(1, 195101, '\p{^NFKC_QC: y}', "");
Expect(1, 195101, '\P{NFKC_QC: y}', "");
Expect(0, 195101, '\P{^NFKC_QC: y}', "");
Expect(1, 195102, '\p{NFKC_QC=_	Y}', "");
Expect(0, 195102, '\p{^NFKC_QC=_	Y}', "");
Expect(0, 195102, '\P{NFKC_QC=_	Y}', "");
Expect(1, 195102, '\P{^NFKC_QC=_	Y}', "");
Expect(0, 195101, '\p{NFKC_QC=_	Y}', "");
Expect(1, 195101, '\p{^NFKC_QC=_	Y}', "");
Expect(1, 195101, '\P{NFKC_QC=_	Y}', "");
Expect(0, 195101, '\P{^NFKC_QC=_	Y}', "");
Error('\p{Is_NFKC_Quick_Check=:=--yes}');
Error('\P{Is_NFKC_Quick_Check=:=--yes}');
Expect(1, 195102, '\p{Is_NFKC_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^Is_NFKC_Quick_Check=yes}', "");
Expect(0, 195102, '\P{Is_NFKC_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^Is_NFKC_Quick_Check=yes}', "");
Expect(0, 195101, '\p{Is_NFKC_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^Is_NFKC_Quick_Check=yes}', "");
Expect(1, 195101, '\P{Is_NFKC_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^Is_NFKC_Quick_Check=yes}', "");
Expect(1, 195102, '\p{Is_NFKC_Quick_Check:   	Yes}', "");
Expect(0, 195102, '\p{^Is_NFKC_Quick_Check:   	Yes}', "");
Expect(0, 195102, '\P{Is_NFKC_Quick_Check:   	Yes}', "");
Expect(1, 195102, '\P{^Is_NFKC_Quick_Check:   	Yes}', "");
Expect(0, 195101, '\p{Is_NFKC_Quick_Check:   	Yes}', "");
Expect(1, 195101, '\p{^Is_NFKC_Quick_Check:   	Yes}', "");
Expect(1, 195101, '\P{Is_NFKC_Quick_Check:   	Yes}', "");
Expect(0, 195101, '\P{^Is_NFKC_Quick_Check:   	Yes}', "");
Error('\p{Is_NFKC_QC=	_Y:=}');
Error('\P{Is_NFKC_QC=	_Y:=}');
Expect(1, 195102, '\p{Is_NFKC_QC=y}', "");
Expect(0, 195102, '\p{^Is_NFKC_QC=y}', "");
Expect(0, 195102, '\P{Is_NFKC_QC=y}', "");
Expect(1, 195102, '\P{^Is_NFKC_QC=y}', "");
Expect(0, 195101, '\p{Is_NFKC_QC=y}', "");
Expect(1, 195101, '\p{^Is_NFKC_QC=y}', "");
Expect(1, 195101, '\P{Is_NFKC_QC=y}', "");
Expect(0, 195101, '\P{^Is_NFKC_QC=y}', "");
Expect(1, 195102, '\p{Is_NFKC_QC=		y}', "");
Expect(0, 195102, '\p{^Is_NFKC_QC=		y}', "");
Expect(0, 195102, '\P{Is_NFKC_QC=		y}', "");
Expect(1, 195102, '\P{^Is_NFKC_QC=		y}', "");
Expect(0, 195101, '\p{Is_NFKC_QC=		y}', "");
Expect(1, 195101, '\p{^Is_NFKC_QC=		y}', "");
Expect(1, 195101, '\P{Is_NFKC_QC=		y}', "");
Expect(0, 195101, '\P{^Is_NFKC_QC=		y}', "");
Error('\p{NFKC_Quick_Check=	:=No}');
Error('\P{NFKC_Quick_Check=	:=No}');
Expect(1, 195101, '\p{NFKC_Quick_Check=no}', "");
Expect(0, 195101, '\p{^NFKC_Quick_Check=no}', "");
Expect(0, 195101, '\P{NFKC_Quick_Check=no}', "");
Expect(1, 195101, '\P{^NFKC_Quick_Check=no}', "");
Expect(0, 195102, '\p{NFKC_Quick_Check=no}', "");
Expect(1, 195102, '\p{^NFKC_Quick_Check=no}', "");
Expect(1, 195102, '\P{NFKC_Quick_Check=no}', "");
Expect(0, 195102, '\P{^NFKC_Quick_Check=no}', "");
Expect(1, 195101, '\p{NFKC_Quick_Check=-no}', "");
Expect(0, 195101, '\p{^NFKC_Quick_Check=-no}', "");
Expect(0, 195101, '\P{NFKC_Quick_Check=-no}', "");
Expect(1, 195101, '\P{^NFKC_Quick_Check=-no}', "");
Expect(0, 195102, '\p{NFKC_Quick_Check=-no}', "");
Expect(1, 195102, '\p{^NFKC_Quick_Check=-no}', "");
Expect(1, 195102, '\P{NFKC_Quick_Check=-no}', "");
Expect(0, 195102, '\P{^NFKC_Quick_Check=-no}', "");
Error('\p{NFKC_QC=	:=N}');
Error('\P{NFKC_QC=	:=N}');
Expect(1, 195101, '\p{NFKC_QC:n}', "");
Expect(0, 195101, '\p{^NFKC_QC:n}', "");
Expect(0, 195101, '\P{NFKC_QC:n}', "");
Expect(1, 195101, '\P{^NFKC_QC:n}', "");
Expect(0, 195102, '\p{NFKC_QC:n}', "");
Expect(1, 195102, '\p{^NFKC_QC:n}', "");
Expect(1, 195102, '\P{NFKC_QC:n}', "");
Expect(0, 195102, '\P{^NFKC_QC:n}', "");
Expect(1, 195101, '\p{NFKC_QC= _N}', "");
Expect(0, 195101, '\p{^NFKC_QC= _N}', "");
Expect(0, 195101, '\P{NFKC_QC= _N}', "");
Expect(1, 195101, '\P{^NFKC_QC= _N}', "");
Expect(0, 195102, '\p{NFKC_QC= _N}', "");
Expect(1, 195102, '\p{^NFKC_QC= _N}', "");
Expect(1, 195102, '\P{NFKC_QC= _N}', "");
Expect(0, 195102, '\P{^NFKC_QC= _N}', "");
Error('\p{Is_NFKC_Quick_Check:  /a/no}');
Error('\P{Is_NFKC_Quick_Check:  /a/no}');
Expect(1, 195101, '\p{Is_NFKC_Quick_Check=no}', "");
Expect(0, 195101, '\p{^Is_NFKC_Quick_Check=no}', "");
Expect(0, 195101, '\P{Is_NFKC_Quick_Check=no}', "");
Expect(1, 195101, '\P{^Is_NFKC_Quick_Check=no}', "");
Expect(0, 195102, '\p{Is_NFKC_Quick_Check=no}', "");
Expect(1, 195102, '\p{^Is_NFKC_Quick_Check=no}', "");
Expect(1, 195102, '\P{Is_NFKC_Quick_Check=no}', "");
Expect(0, 195102, '\P{^Is_NFKC_Quick_Check=no}', "");
Expect(1, 195101, '\p{Is_NFKC_Quick_Check=-	no}', "");
Expect(0, 195101, '\p{^Is_NFKC_Quick_Check=-	no}', "");
Expect(0, 195101, '\P{Is_NFKC_Quick_Check=-	no}', "");
Expect(1, 195101, '\P{^Is_NFKC_Quick_Check=-	no}', "");
Expect(0, 195102, '\p{Is_NFKC_Quick_Check=-	no}', "");
Expect(1, 195102, '\p{^Is_NFKC_Quick_Check=-	no}', "");
Expect(1, 195102, '\P{Is_NFKC_Quick_Check=-	no}', "");
Expect(0, 195102, '\P{^Is_NFKC_Quick_Check=-	no}', "");
Error('\p{Is_NFKC_QC=:=__N}');
Error('\P{Is_NFKC_QC=:=__N}');
Expect(1, 195101, '\p{Is_NFKC_QC=n}', "");
Expect(0, 195101, '\p{^Is_NFKC_QC=n}', "");
Expect(0, 195101, '\P{Is_NFKC_QC=n}', "");
Expect(1, 195101, '\P{^Is_NFKC_QC=n}', "");
Expect(0, 195102, '\p{Is_NFKC_QC=n}', "");
Expect(1, 195102, '\p{^Is_NFKC_QC=n}', "");
Expect(1, 195102, '\P{Is_NFKC_QC=n}', "");
Expect(0, 195102, '\P{^Is_NFKC_QC=n}', "");
Expect(1, 195101, '\p{Is_NFKC_QC=-	N}', "");
Expect(0, 195101, '\p{^Is_NFKC_QC=-	N}', "");
Expect(0, 195101, '\P{Is_NFKC_QC=-	N}', "");
Expect(1, 195101, '\P{^Is_NFKC_QC=-	N}', "");
Expect(0, 195102, '\p{Is_NFKC_QC=-	N}', "");
Expect(1, 195102, '\p{^Is_NFKC_QC=-	N}', "");
Expect(1, 195102, '\P{Is_NFKC_QC=-	N}', "");
Expect(0, 195102, '\P{^Is_NFKC_QC=-	N}', "");
Error('\p{NFKC_Quick_Check= _MAYBE/a/}');
Error('\P{NFKC_Quick_Check= _MAYBE/a/}');
Expect(1, 69927, '\p{NFKC_Quick_Check:	maybe}', "");
Expect(0, 69927, '\p{^NFKC_Quick_Check:	maybe}', "");
Expect(0, 69927, '\P{NFKC_Quick_Check:	maybe}', "");
Expect(1, 69927, '\P{^NFKC_Quick_Check:	maybe}', "");
Expect(0, 69928, '\p{NFKC_Quick_Check:	maybe}', "");
Expect(1, 69928, '\p{^NFKC_Quick_Check:	maybe}', "");
Expect(1, 69928, '\P{NFKC_Quick_Check:	maybe}', "");
Expect(0, 69928, '\P{^NFKC_Quick_Check:	maybe}', "");
Expect(1, 69927, '\p{NFKC_Quick_Check=	-maybe}', "");
Expect(0, 69927, '\p{^NFKC_Quick_Check=	-maybe}', "");
Expect(0, 69927, '\P{NFKC_Quick_Check=	-maybe}', "");
Expect(1, 69927, '\P{^NFKC_Quick_Check=	-maybe}', "");
Expect(0, 69928, '\p{NFKC_Quick_Check=	-maybe}', "");
Expect(1, 69928, '\p{^NFKC_Quick_Check=	-maybe}', "");
Expect(1, 69928, '\P{NFKC_Quick_Check=	-maybe}', "");
Expect(0, 69928, '\P{^NFKC_Quick_Check=	-maybe}', "");
Error('\p{NFKC_QC:  _M/a/}');
Error('\P{NFKC_QC:  _M/a/}');
Expect(1, 69927, '\p{NFKC_QC=m}', "");
Expect(0, 69927, '\p{^NFKC_QC=m}', "");
Expect(0, 69927, '\P{NFKC_QC=m}', "");
Expect(1, 69927, '\P{^NFKC_QC=m}', "");
Expect(0, 69928, '\p{NFKC_QC=m}', "");
Expect(1, 69928, '\p{^NFKC_QC=m}', "");
Expect(1, 69928, '\P{NFKC_QC=m}', "");
Expect(0, 69928, '\P{^NFKC_QC=m}', "");
Expect(1, 69927, '\p{NFKC_QC=- M}', "");
Expect(0, 69927, '\p{^NFKC_QC=- M}', "");
Expect(0, 69927, '\P{NFKC_QC=- M}', "");
Expect(1, 69927, '\P{^NFKC_QC=- M}', "");
Expect(0, 69928, '\p{NFKC_QC=- M}', "");
Expect(1, 69928, '\p{^NFKC_QC=- M}', "");
Expect(1, 69928, '\P{NFKC_QC=- M}', "");
Expect(0, 69928, '\P{^NFKC_QC=- M}', "");
Error('\p{Is_NFKC_Quick_Check=--Maybe/a/}');
Error('\P{Is_NFKC_Quick_Check=--Maybe/a/}');
Expect(1, 69927, '\p{Is_NFKC_Quick_Check=maybe}', "");
Expect(0, 69927, '\p{^Is_NFKC_Quick_Check=maybe}', "");
Expect(0, 69927, '\P{Is_NFKC_Quick_Check=maybe}', "");
Expect(1, 69927, '\P{^Is_NFKC_Quick_Check=maybe}', "");
Expect(0, 69928, '\p{Is_NFKC_Quick_Check=maybe}', "");
Expect(1, 69928, '\p{^Is_NFKC_Quick_Check=maybe}', "");
Expect(1, 69928, '\P{Is_NFKC_Quick_Check=maybe}', "");
Expect(0, 69928, '\P{^Is_NFKC_Quick_Check=maybe}', "");
Expect(1, 69927, '\p{Is_NFKC_Quick_Check:   	MAYBE}', "");
Expect(0, 69927, '\p{^Is_NFKC_Quick_Check:   	MAYBE}', "");
Expect(0, 69927, '\P{Is_NFKC_Quick_Check:   	MAYBE}', "");
Expect(1, 69927, '\P{^Is_NFKC_Quick_Check:   	MAYBE}', "");
Expect(0, 69928, '\p{Is_NFKC_Quick_Check:   	MAYBE}', "");
Expect(1, 69928, '\p{^Is_NFKC_Quick_Check:   	MAYBE}', "");
Expect(1, 69928, '\P{Is_NFKC_Quick_Check:   	MAYBE}', "");
Expect(0, 69928, '\P{^Is_NFKC_Quick_Check:   	MAYBE}', "");
Error('\p{Is_NFKC_QC=-:=m}');
Error('\P{Is_NFKC_QC=-:=m}');
Expect(1, 69927, '\p{Is_NFKC_QC=m}', "");
Expect(0, 69927, '\p{^Is_NFKC_QC=m}', "");
Expect(0, 69927, '\P{Is_NFKC_QC=m}', "");
Expect(1, 69927, '\P{^Is_NFKC_QC=m}', "");
Expect(0, 69928, '\p{Is_NFKC_QC=m}', "");
Expect(1, 69928, '\p{^Is_NFKC_QC=m}', "");
Expect(1, 69928, '\P{Is_NFKC_QC=m}', "");
Expect(0, 69928, '\P{^Is_NFKC_QC=m}', "");
Expect(1, 69927, '\p{Is_NFKC_QC=- m}', "");
Expect(0, 69927, '\p{^Is_NFKC_QC=- m}', "");
Expect(0, 69927, '\P{Is_NFKC_QC=- m}', "");
Expect(1, 69927, '\P{^Is_NFKC_QC=- m}', "");
Expect(0, 69928, '\p{Is_NFKC_QC=- m}', "");
Expect(1, 69928, '\p{^Is_NFKC_QC=- m}', "");
Expect(1, 69928, '\P{Is_NFKC_QC=- m}', "");
Expect(0, 69928, '\P{^Is_NFKC_QC=- m}', "");
Error('\p{Other_Alphabetic=Yes}');
Error('\P{Other_Alphabetic=Yes}');
Error('\p{OAlpha=Y}');
Error('\P{OAlpha=Y}');
Error('\p{Is_Other_Alphabetic=T}');
Error('\P{Is_Other_Alphabetic=T}');
Error('\p{Is_OAlpha=True}');
Error('\P{Is_OAlpha=True}');
Error('\p{Other_Alphabetic:   No}');
Error('\P{Other_Alphabetic:   No}');
Error('\p{OAlpha=N}');
Error('\P{OAlpha=N}');
Error('\p{Is_Other_Alphabetic=F}');
Error('\P{Is_Other_Alphabetic=F}');
Error('\p{Is_OAlpha:False}');
Error('\P{Is_OAlpha:False}');
Error('\p{Indic_Matra_Category=Left}');
Error('\P{Indic_Matra_Category=Left}');
Error('\p{InMC=Left}');
Error('\P{InMC=Left}');
Error('\p{Is_Indic_Matra_Category:   Left}');
Error('\P{Is_Indic_Matra_Category:   Left}');
Error('\p{Is_InMC=Left}');
Error('\P{Is_InMC=Left}');
Error('\p{Indic_Matra_Category:Right}');
Error('\P{Indic_Matra_Category:Right}');
Error('\p{InMC=Right}');
Error('\P{InMC=Right}');
Error('\p{Is_Indic_Matra_Category=Right}');
Error('\P{Is_Indic_Matra_Category=Right}');
Error('\p{Is_InMC=Right}');
Error('\P{Is_InMC=Right}');
Error('\p{Indic_Matra_Category: Top_And_Left}');
Error('\P{Indic_Matra_Category: Top_And_Left}');
Error('\p{InMC=Top_And_Left}');
Error('\P{InMC=Top_And_Left}');
Error('\p{Is_Indic_Matra_Category=Top_And_Left}');
Error('\P{Is_Indic_Matra_Category=Top_And_Left}');
Error('\p{Is_InMC=Top_And_Left}');
Error('\P{Is_InMC=Top_And_Left}');
Error('\p{Indic_Matra_Category=Top_And_Bottom_And_Right}');
Error('\P{Indic_Matra_Category=Top_And_Bottom_And_Right}');
Error('\p{InMC=Top_And_Bottom_And_Right}');
Error('\P{InMC=Top_And_Bottom_And_Right}');
Error('\p{Is_Indic_Matra_Category:Top_And_Bottom_And_Right}');
Error('\P{Is_Indic_Matra_Category:Top_And_Bottom_And_Right}');
Error('\p{Is_InMC:   Top_And_Bottom_And_Right}');
Error('\P{Is_InMC:   Top_And_Bottom_And_Right}');
Error('\p{Indic_Matra_Category=Overstruck}');
Error('\P{Indic_Matra_Category=Overstruck}');
Error('\p{InMC=Overstruck}');
Error('\P{InMC=Overstruck}');
Error('\p{Is_Indic_Matra_Category=Overstruck}');
Error('\P{Is_Indic_Matra_Category=Overstruck}');
Error('\p{Is_InMC=Overstruck}');
Error('\P{Is_InMC=Overstruck}');
Error('\p{Indic_Matra_Category=Top_And_Bottom}');
Error('\P{Indic_Matra_Category=Top_And_Bottom}');
Error('\p{InMC=Top_And_Bottom}');
Error('\P{InMC=Top_And_Bottom}');
Error('\p{Is_Indic_Matra_Category=Top_And_Bottom}');
Error('\P{Is_Indic_Matra_Category=Top_And_Bottom}');
Error('\p{Is_InMC=Top_And_Bottom}');
Error('\P{Is_InMC=Top_And_Bottom}');
Error('\p{Indic_Matra_Category=Bottom_And_Right}');
Error('\P{Indic_Matra_Category=Bottom_And_Right}');
Error('\p{InMC=Bottom_And_Right}');
Error('\P{InMC=Bottom_And_Right}');
Error('\p{Is_Indic_Matra_Category=Bottom_And_Right}');
Error('\P{Is_Indic_Matra_Category=Bottom_And_Right}');
Error('\p{Is_InMC=Bottom_And_Right}');
Error('\P{Is_InMC=Bottom_And_Right}');
Error('\p{Indic_Matra_Category:   Visual_Order_Left}');
Error('\P{Indic_Matra_Category:   Visual_Order_Left}');
Error('\p{InMC=Visual_Order_Left}');
Error('\P{InMC=Visual_Order_Left}');
Error('\p{Is_Indic_Matra_Category=Visual_Order_Left}');
Error('\P{Is_Indic_Matra_Category=Visual_Order_Left}');
Error('\p{Is_InMC:	Visual_Order_Left}');
Error('\P{Is_InMC:	Visual_Order_Left}');
Error('\p{Indic_Matra_Category=Top_And_Right}');
Error('\P{Indic_Matra_Category=Top_And_Right}');
Error('\p{InMC=Top_And_Right}');
Error('\P{InMC=Top_And_Right}');
Error('\p{Is_Indic_Matra_Category=Top_And_Right}');
Error('\P{Is_Indic_Matra_Category=Top_And_Right}');
Error('\p{Is_InMC=Top_And_Right}');
Error('\P{Is_InMC=Top_And_Right}');
Error('\p{Indic_Matra_Category=Top}');
Error('\P{Indic_Matra_Category=Top}');
Error('\p{InMC=Top}');
Error('\P{InMC=Top}');
Error('\p{Is_Indic_Matra_Category=Top}');
Error('\P{Is_Indic_Matra_Category=Top}');
Error('\p{Is_InMC=Top}');
Error('\P{Is_InMC=Top}');
Error('\p{Indic_Matra_Category=NA}');
Error('\P{Indic_Matra_Category=NA}');
Error('\p{InMC=NA}');
Error('\P{InMC=NA}');
Error('\p{Is_Indic_Matra_Category=NA}');
Error('\P{Is_Indic_Matra_Category=NA}');
Error('\p{Is_InMC=NA}');
Error('\P{Is_InMC=NA}');
Error('\p{Indic_Matra_Category=Bottom}');
Error('\P{Indic_Matra_Category=Bottom}');
Error('\p{InMC=Bottom}');
Error('\P{InMC=Bottom}');
Error('\p{Is_Indic_Matra_Category=Bottom}');
Error('\P{Is_Indic_Matra_Category=Bottom}');
Error('\p{Is_InMC=Bottom}');
Error('\P{Is_InMC=Bottom}');
Error('\p{Indic_Matra_Category=Invisible}');
Error('\P{Indic_Matra_Category=Invisible}');
Error('\p{InMC=Invisible}');
Error('\P{InMC=Invisible}');
Error('\p{Is_Indic_Matra_Category=Invisible}');
Error('\P{Is_Indic_Matra_Category=Invisible}');
Error('\p{Is_InMC=Invisible}');
Error('\P{Is_InMC=Invisible}');
Error('\p{Indic_Matra_Category=Left_And_Right}');
Error('\P{Indic_Matra_Category=Left_And_Right}');
Error('\p{InMC=Left_And_Right}');
Error('\P{InMC=Left_And_Right}');
Error('\p{Is_Indic_Matra_Category=Left_And_Right}');
Error('\P{Is_Indic_Matra_Category=Left_And_Right}');
Error('\p{Is_InMC=Left_And_Right}');
Error('\P{Is_InMC=Left_And_Right}');
Error('\p{Indic_Matra_Category=Top_And_Left_And_Right}');
Error('\P{Indic_Matra_Category=Top_And_Left_And_Right}');
Error('\p{InMC:Top_And_Left_And_Right}');
Error('\P{InMC:Top_And_Left_And_Right}');
Error('\p{Is_Indic_Matra_Category=Top_And_Left_And_Right}');
Error('\P{Is_Indic_Matra_Category=Top_And_Left_And_Right}');
Error('\p{Is_InMC=Top_And_Left_And_Right}');
Error('\P{Is_InMC=Top_And_Left_And_Right}');
Error('\p{General_Category=/a/_-Symbol}');
Error('\P{General_Category=/a/_-Symbol}');
Expect(1, 128883, '\p{General_Category=symbol}', "");
Expect(0, 128883, '\p{^General_Category=symbol}', "");
Expect(0, 128883, '\P{General_Category=symbol}', "");
Expect(1, 128883, '\P{^General_Category=symbol}', "");
Expect(0, 128884, '\p{General_Category=symbol}', "");
Expect(1, 128884, '\p{^General_Category=symbol}', "");
Expect(1, 128884, '\P{General_Category=symbol}', "");
Expect(0, 128884, '\P{^General_Category=symbol}', "");
Expect(1, 128883, '\p{General_Category= Symbol}', "");
Expect(0, 128883, '\p{^General_Category= Symbol}', "");
Expect(0, 128883, '\P{General_Category= Symbol}', "");
Expect(1, 128883, '\P{^General_Category= Symbol}', "");
Expect(0, 128884, '\p{General_Category= Symbol}', "");
Expect(1, 128884, '\p{^General_Category= Symbol}', "");
Expect(1, 128884, '\P{General_Category= Symbol}', "");
Expect(0, 128884, '\P{^General_Category= Symbol}', "");
Error('\p{Gc:    S:=}');
Error('\P{Gc:    S:=}');
Expect(1, 128883, '\p{Gc=s}', "");
Expect(0, 128883, '\p{^Gc=s}', "");
Expect(0, 128883, '\P{Gc=s}', "");
Expect(1, 128883, '\P{^Gc=s}', "");
Expect(0, 128884, '\p{Gc=s}', "");
Expect(1, 128884, '\p{^Gc=s}', "");
Expect(1, 128884, '\P{Gc=s}', "");
Expect(0, 128884, '\P{^Gc=s}', "");
Expect(1, 128883, '\p{Gc= -s}', "");
Expect(0, 128883, '\p{^Gc= -s}', "");
Expect(0, 128883, '\P{Gc= -s}', "");
Expect(1, 128883, '\P{^Gc= -s}', "");
Expect(0, 128884, '\p{Gc= -s}', "");
Expect(1, 128884, '\p{^Gc= -s}', "");
Expect(1, 128884, '\P{Gc= -s}', "");
Expect(0, 128884, '\P{^Gc= -s}', "");
Error('\p{Category=_	symbol:=}');
Error('\P{Category=_	symbol:=}');
Expect(1, 128883, '\p{Category:symbol}', "");
Expect(0, 128883, '\p{^Category:symbol}', "");
Expect(0, 128883, '\P{Category:symbol}', "");
Expect(1, 128883, '\P{^Category:symbol}', "");
Expect(0, 128884, '\p{Category:symbol}', "");
Expect(1, 128884, '\p{^Category:symbol}', "");
Expect(1, 128884, '\P{Category:symbol}', "");
Expect(0, 128884, '\P{^Category:symbol}', "");
Expect(1, 128883, '\p{Category=_-symbol}', "");
Expect(0, 128883, '\p{^Category=_-symbol}', "");
Expect(0, 128883, '\P{Category=_-symbol}', "");
Expect(1, 128883, '\P{^Category=_-symbol}', "");
Expect(0, 128884, '\p{Category=_-symbol}', "");
Expect(1, 128884, '\p{^Category=_-symbol}', "");
Expect(1, 128884, '\P{Category=_-symbol}', "");
Expect(0, 128884, '\P{^Category=_-symbol}', "");
Error('\p{Is_General_Category=_	S/a/}');
Error('\P{Is_General_Category=_	S/a/}');
Expect(1, 128883, '\p{Is_General_Category=s}', "");
Expect(0, 128883, '\p{^Is_General_Category=s}', "");
Expect(0, 128883, '\P{Is_General_Category=s}', "");
Expect(1, 128883, '\P{^Is_General_Category=s}', "");
Expect(0, 128884, '\p{Is_General_Category=s}', "");
Expect(1, 128884, '\p{^Is_General_Category=s}', "");
Expect(1, 128884, '\P{Is_General_Category=s}', "");
Expect(0, 128884, '\P{^Is_General_Category=s}', "");
Expect(1, 128883, '\p{Is_General_Category= S}', "");
Expect(0, 128883, '\p{^Is_General_Category= S}', "");
Expect(0, 128883, '\P{Is_General_Category= S}', "");
Expect(1, 128883, '\P{^Is_General_Category= S}', "");
Expect(0, 128884, '\p{Is_General_Category= S}', "");
Expect(1, 128884, '\p{^Is_General_Category= S}', "");
Expect(1, 128884, '\P{Is_General_Category= S}', "");
Expect(0, 128884, '\P{^Is_General_Category= S}', "");
Error('\p{Is_Gc=/a/Symbol}');
Error('\P{Is_Gc=/a/Symbol}');
Expect(1, 128883, '\p{Is_Gc=symbol}', "");
Expect(0, 128883, '\p{^Is_Gc=symbol}', "");
Expect(0, 128883, '\P{Is_Gc=symbol}', "");
Expect(1, 128883, '\P{^Is_Gc=symbol}', "");
Expect(0, 128884, '\p{Is_Gc=symbol}', "");
Expect(1, 128884, '\p{^Is_Gc=symbol}', "");
Expect(1, 128884, '\P{Is_Gc=symbol}', "");
Expect(0, 128884, '\P{^Is_Gc=symbol}', "");
Expect(1, 128883, '\p{Is_Gc=_-Symbol}', "");
Expect(0, 128883, '\p{^Is_Gc=_-Symbol}', "");
Expect(0, 128883, '\P{Is_Gc=_-Symbol}', "");
Expect(1, 128883, '\P{^Is_Gc=_-Symbol}', "");
Expect(0, 128884, '\p{Is_Gc=_-Symbol}', "");
Expect(1, 128884, '\p{^Is_Gc=_-Symbol}', "");
Expect(1, 128884, '\P{Is_Gc=_-Symbol}', "");
Expect(0, 128884, '\P{^Is_Gc=_-Symbol}', "");
Error('\p{Is_Category= -s/a/}');
Error('\P{Is_Category= -s/a/}');
Expect(1, 128883, '\p{Is_Category=s}', "");
Expect(0, 128883, '\p{^Is_Category=s}', "");
Expect(0, 128883, '\P{Is_Category=s}', "");
Expect(1, 128883, '\P{^Is_Category=s}', "");
Expect(0, 128884, '\p{Is_Category=s}', "");
Expect(1, 128884, '\p{^Is_Category=s}', "");
Expect(1, 128884, '\P{Is_Category=s}', "");
Expect(0, 128884, '\P{^Is_Category=s}', "");
Expect(1, 128883, '\p{Is_Category= S}', "");
Expect(0, 128883, '\p{^Is_Category= S}', "");
Expect(0, 128883, '\P{Is_Category= S}', "");
Expect(1, 128883, '\P{^Is_Category= S}', "");
Expect(0, 128884, '\p{Is_Category= S}', "");
Expect(1, 128884, '\p{^Is_Category= S}', "");
Expect(1, 128884, '\P{Is_Category= S}', "");
Expect(0, 128884, '\P{^Is_Category= S}', "");
Error('\p{General_Category=-/a/spacing_Mark}');
Error('\P{General_Category=-/a/spacing_Mark}');
Expect(1, 119154, '\p{General_Category:	spacingmark}', "");
Expect(0, 119154, '\p{^General_Category:	spacingmark}', "");
Expect(0, 119154, '\P{General_Category:	spacingmark}', "");
Expect(1, 119154, '\P{^General_Category:	spacingmark}', "");
Expect(0, 119155, '\p{General_Category:	spacingmark}', "");
Expect(1, 119155, '\p{^General_Category:	spacingmark}', "");
Expect(1, 119155, '\P{General_Category:	spacingmark}', "");
Expect(0, 119155, '\P{^General_Category:	spacingmark}', "");
Expect(1, 119154, '\p{General_Category= Spacing_Mark}', "");
Expect(0, 119154, '\p{^General_Category= Spacing_Mark}', "");
Expect(0, 119154, '\P{General_Category= Spacing_Mark}', "");
Expect(1, 119154, '\P{^General_Category= Spacing_Mark}', "");
Expect(0, 119155, '\p{General_Category= Spacing_Mark}', "");
Expect(1, 119155, '\p{^General_Category= Spacing_Mark}', "");
Expect(1, 119155, '\P{General_Category= Spacing_Mark}', "");
Expect(0, 119155, '\P{^General_Category= Spacing_Mark}', "");
Error('\p{Gc=/a/ _Mc}');
Error('\P{Gc=/a/ _Mc}');
Expect(1, 119154, '\p{Gc=mc}', "");
Expect(0, 119154, '\p{^Gc=mc}', "");
Expect(0, 119154, '\P{Gc=mc}', "");
Expect(1, 119154, '\P{^Gc=mc}', "");
Expect(0, 119155, '\p{Gc=mc}', "");
Expect(1, 119155, '\p{^Gc=mc}', "");
Expect(1, 119155, '\P{Gc=mc}', "");
Expect(0, 119155, '\P{^Gc=mc}', "");
Expect(1, 119154, '\p{Gc=		MC}', "");
Expect(0, 119154, '\p{^Gc=		MC}', "");
Expect(0, 119154, '\P{Gc=		MC}', "");
Expect(1, 119154, '\P{^Gc=		MC}', "");
Expect(0, 119155, '\p{Gc=		MC}', "");
Expect(1, 119155, '\p{^Gc=		MC}', "");
Expect(1, 119155, '\P{Gc=		MC}', "");
Expect(0, 119155, '\P{^Gc=		MC}', "");
Error('\p{Category=_:=Spacing_MARK}');
Error('\P{Category=_:=Spacing_MARK}');
Expect(1, 119154, '\p{Category=spacingmark}', "");
Expect(0, 119154, '\p{^Category=spacingmark}', "");
Expect(0, 119154, '\P{Category=spacingmark}', "");
Expect(1, 119154, '\P{^Category=spacingmark}', "");
Expect(0, 119155, '\p{Category=spacingmark}', "");
Expect(1, 119155, '\p{^Category=spacingmark}', "");
Expect(1, 119155, '\P{Category=spacingmark}', "");
Expect(0, 119155, '\P{^Category=spacingmark}', "");
Expect(1, 119154, '\p{Category=_	Spacing_Mark}', "");
Expect(0, 119154, '\p{^Category=_	Spacing_Mark}', "");
Expect(0, 119154, '\P{Category=_	Spacing_Mark}', "");
Expect(1, 119154, '\P{^Category=_	Spacing_Mark}', "");
Expect(0, 119155, '\p{Category=_	Spacing_Mark}', "");
Expect(1, 119155, '\p{^Category=_	Spacing_Mark}', "");
Expect(1, 119155, '\P{Category=_	Spacing_Mark}', "");
Expect(0, 119155, '\P{^Category=_	Spacing_Mark}', "");
Error('\p{Is_General_Category=:=_	Mc}');
Error('\P{Is_General_Category=:=_	Mc}');
Expect(1, 119154, '\p{Is_General_Category=mc}', "");
Expect(0, 119154, '\p{^Is_General_Category=mc}', "");
Expect(0, 119154, '\P{Is_General_Category=mc}', "");
Expect(1, 119154, '\P{^Is_General_Category=mc}', "");
Expect(0, 119155, '\p{Is_General_Category=mc}', "");
Expect(1, 119155, '\p{^Is_General_Category=mc}', "");
Expect(1, 119155, '\P{Is_General_Category=mc}', "");
Expect(0, 119155, '\P{^Is_General_Category=mc}', "");
Expect(1, 119154, '\p{Is_General_Category= Mc}', "");
Expect(0, 119154, '\p{^Is_General_Category= Mc}', "");
Expect(0, 119154, '\P{Is_General_Category= Mc}', "");
Expect(1, 119154, '\P{^Is_General_Category= Mc}', "");
Expect(0, 119155, '\p{Is_General_Category= Mc}', "");
Expect(1, 119155, '\p{^Is_General_Category= Mc}', "");
Expect(1, 119155, '\P{Is_General_Category= Mc}', "");
Expect(0, 119155, '\P{^Is_General_Category= Mc}', "");
Error('\p{Is_Gc= 	spacing_Mark:=}');
Error('\P{Is_Gc= 	spacing_Mark:=}');
Expect(1, 119154, '\p{Is_Gc=spacingmark}', "");
Expect(0, 119154, '\p{^Is_Gc=spacingmark}', "");
Expect(0, 119154, '\P{Is_Gc=spacingmark}', "");
Expect(1, 119154, '\P{^Is_Gc=spacingmark}', "");
Expect(0, 119155, '\p{Is_Gc=spacingmark}', "");
Expect(1, 119155, '\p{^Is_Gc=spacingmark}', "");
Expect(1, 119155, '\P{Is_Gc=spacingmark}', "");
Expect(0, 119155, '\P{^Is_Gc=spacingmark}', "");
Expect(1, 119154, '\p{Is_Gc=- Spacing_mark}', "");
Expect(0, 119154, '\p{^Is_Gc=- Spacing_mark}', "");
Expect(0, 119154, '\P{Is_Gc=- Spacing_mark}', "");
Expect(1, 119154, '\P{^Is_Gc=- Spacing_mark}', "");
Expect(0, 119155, '\p{Is_Gc=- Spacing_mark}', "");
Expect(1, 119155, '\p{^Is_Gc=- Spacing_mark}', "");
Expect(1, 119155, '\P{Is_Gc=- Spacing_mark}', "");
Expect(0, 119155, '\P{^Is_Gc=- Spacing_mark}', "");
Error('\p{Is_Category::= Mc}');
Error('\P{Is_Category::= Mc}');
Expect(1, 119154, '\p{Is_Category: mc}', "");
Expect(0, 119154, '\p{^Is_Category: mc}', "");
Expect(0, 119154, '\P{Is_Category: mc}', "");
Expect(1, 119154, '\P{^Is_Category: mc}', "");
Expect(0, 119155, '\p{Is_Category: mc}', "");
Expect(1, 119155, '\p{^Is_Category: mc}', "");
Expect(1, 119155, '\P{Is_Category: mc}', "");
Expect(0, 119155, '\P{^Is_Category: mc}', "");
Expect(1, 119154, '\p{Is_Category= mc}', "");
Expect(0, 119154, '\p{^Is_Category= mc}', "");
Expect(0, 119154, '\P{Is_Category= mc}', "");
Expect(1, 119154, '\P{^Is_Category= mc}', "");
Expect(0, 119155, '\p{Is_Category= mc}', "");
Expect(1, 119155, '\p{^Is_Category= mc}', "");
Expect(1, 119155, '\P{Is_Category= mc}', "");
Expect(0, 119155, '\P{^Is_Category= mc}', "");
Error('\p{General_Category=:= 	Cased_Letter}');
Error('\P{General_Category=:= 	Cased_Letter}');
Expect(1, 120779, '\p{General_Category:	casedletter}', "");
Expect(0, 120779, '\p{^General_Category:	casedletter}', "");
Expect(0, 120779, '\P{General_Category:	casedletter}', "");
Expect(1, 120779, '\P{^General_Category:	casedletter}', "");
Expect(0, 120780, '\p{General_Category:	casedletter}', "");
Expect(1, 120780, '\p{^General_Category:	casedletter}', "");
Expect(1, 120780, '\P{General_Category:	casedletter}', "");
Expect(0, 120780, '\P{^General_Category:	casedletter}', "");
Expect(1, 120779, '\p{General_Category=-Cased_Letter}', "");
Expect(0, 120779, '\p{^General_Category=-Cased_Letter}', "");
Expect(0, 120779, '\P{General_Category=-Cased_Letter}', "");
Expect(1, 120779, '\P{^General_Category=-Cased_Letter}', "");
Expect(0, 120780, '\p{General_Category=-Cased_Letter}', "");
Expect(1, 120780, '\p{^General_Category=-Cased_Letter}', "");
Expect(1, 120780, '\P{General_Category=-Cased_Letter}', "");
Expect(0, 120780, '\P{^General_Category=-Cased_Letter}', "");
Error('\p{Gc=__LC:=}');
Error('\P{Gc=__LC:=}');
Expect(1, 120779, '\p{Gc: lc}', "");
Expect(0, 120779, '\p{^Gc: lc}', "");
Expect(0, 120779, '\P{Gc: lc}', "");
Expect(1, 120779, '\P{^Gc: lc}', "");
Expect(0, 120780, '\p{Gc: lc}', "");
Expect(1, 120780, '\p{^Gc: lc}', "");
Expect(1, 120780, '\P{Gc: lc}', "");
Expect(0, 120780, '\P{^Gc: lc}', "");
Expect(1, 120779, '\p{Gc=-LC}', "");
Expect(0, 120779, '\p{^Gc=-LC}', "");
Expect(0, 120779, '\P{Gc=-LC}', "");
Expect(1, 120779, '\P{^Gc=-LC}', "");
Expect(0, 120780, '\p{Gc=-LC}', "");
Expect(1, 120780, '\p{^Gc=-LC}', "");
Expect(1, 120780, '\P{Gc=-LC}', "");
Expect(0, 120780, '\P{^Gc=-LC}', "");
Error('\p{Category=L_:=}');
Error('\P{Category=L_:=}');
Expect(1, 120779, '\p{Category=l_}', "");
Expect(0, 120779, '\p{^Category=l_}', "");
Expect(0, 120779, '\P{Category=l_}', "");
Expect(1, 120779, '\P{^Category=l_}', "");
Expect(0, 120780, '\p{Category=l_}', "");
Expect(1, 120780, '\p{^Category=l_}', "");
Expect(1, 120780, '\P{Category=l_}', "");
Expect(0, 120780, '\P{^Category=l_}', "");
Expect(1, 120779, '\p{Category=-l_}', "");
Expect(0, 120779, '\p{^Category=-l_}', "");
Expect(0, 120779, '\P{Category=-l_}', "");
Expect(1, 120779, '\P{^Category=-l_}', "");
Expect(0, 120780, '\p{Category=-l_}', "");
Expect(1, 120780, '\p{^Category=-l_}', "");
Expect(1, 120780, '\P{Category=-l_}', "");
Expect(0, 120780, '\P{^Category=-l_}', "");
Error('\p{Is_General_Category=/a/_	L&}');
Error('\P{Is_General_Category=/a/_	L&}');
Expect(1, 120779, '\p{Is_General_Category=l&}', "");
Expect(0, 120779, '\p{^Is_General_Category=l&}', "");
Expect(0, 120779, '\P{Is_General_Category=l&}', "");
Expect(1, 120779, '\P{^Is_General_Category=l&}', "");
Expect(0, 120780, '\p{Is_General_Category=l&}', "");
Expect(1, 120780, '\p{^Is_General_Category=l&}', "");
Expect(1, 120780, '\P{Is_General_Category=l&}', "");
Expect(0, 120780, '\P{^Is_General_Category=l&}', "");
Expect(1, 120779, '\p{Is_General_Category= l&}', "");
Expect(0, 120779, '\p{^Is_General_Category= l&}', "");
Expect(0, 120779, '\P{Is_General_Category= l&}', "");
Expect(1, 120779, '\P{^Is_General_Category= l&}', "");
Expect(0, 120780, '\p{Is_General_Category= l&}', "");
Expect(1, 120780, '\p{^Is_General_Category= l&}', "");
Expect(1, 120780, '\P{Is_General_Category= l&}', "");
Expect(0, 120780, '\P{^Is_General_Category= l&}', "");
Error('\p{Is_Gc=/a/ CASED_Letter}');
Error('\P{Is_Gc=/a/ CASED_Letter}');
Expect(1, 120779, '\p{Is_Gc=casedletter}', "");
Expect(0, 120779, '\p{^Is_Gc=casedletter}', "");
Expect(0, 120779, '\P{Is_Gc=casedletter}', "");
Expect(1, 120779, '\P{^Is_Gc=casedletter}', "");
Expect(0, 120780, '\p{Is_Gc=casedletter}', "");
Expect(1, 120780, '\p{^Is_Gc=casedletter}', "");
Expect(1, 120780, '\P{Is_Gc=casedletter}', "");
Expect(0, 120780, '\P{^Is_Gc=casedletter}', "");
Expect(1, 120779, '\p{Is_Gc=_cased_letter}', "");
Expect(0, 120779, '\p{^Is_Gc=_cased_letter}', "");
Expect(0, 120779, '\P{Is_Gc=_cased_letter}', "");
Expect(1, 120779, '\P{^Is_Gc=_cased_letter}', "");
Expect(0, 120780, '\p{Is_Gc=_cased_letter}', "");
Expect(1, 120780, '\p{^Is_Gc=_cased_letter}', "");
Expect(1, 120780, '\P{Is_Gc=_cased_letter}', "");
Expect(0, 120780, '\P{^Is_Gc=_cased_letter}', "");
Error('\p{Is_Category=	LC/a/}');
Error('\P{Is_Category=	LC/a/}');
Expect(1, 120779, '\p{Is_Category:   lc}', "");
Expect(0, 120779, '\p{^Is_Category:   lc}', "");
Expect(0, 120779, '\P{Is_Category:   lc}', "");
Expect(1, 120779, '\P{^Is_Category:   lc}', "");
Expect(0, 120780, '\p{Is_Category:   lc}', "");
Expect(1, 120780, '\p{^Is_Category:   lc}', "");
Expect(1, 120780, '\P{Is_Category:   lc}', "");
Expect(0, 120780, '\P{^Is_Category:   lc}', "");
Expect(1, 120779, '\p{Is_Category=-LC}', "");
Expect(0, 120779, '\p{^Is_Category=-LC}', "");
Expect(0, 120779, '\P{Is_Category=-LC}', "");
Expect(1, 120779, '\P{^Is_Category=-LC}', "");
Expect(0, 120780, '\p{Is_Category=-LC}', "");
Expect(1, 120780, '\p{^Is_Category=-LC}', "");
Expect(1, 120780, '\P{Is_Category=-LC}', "");
Expect(0, 120780, '\P{^Is_Category=-LC}', "");
Error('\p{General_Category=:=_	Connector_PUNCTUATION}');
Error('\P{General_Category=:=_	Connector_PUNCTUATION}');
Expect(1, 65343, '\p{General_Category=connectorpunctuation}', "");
Expect(0, 65343, '\p{^General_Category=connectorpunctuation}', "");
Expect(0, 65343, '\P{General_Category=connectorpunctuation}', "");
Expect(1, 65343, '\P{^General_Category=connectorpunctuation}', "");
Expect(0, 65344, '\p{General_Category=connectorpunctuation}', "");
Expect(1, 65344, '\p{^General_Category=connectorpunctuation}', "");
Expect(1, 65344, '\P{General_Category=connectorpunctuation}', "");
Expect(0, 65344, '\P{^General_Category=connectorpunctuation}', "");
Expect(1, 65343, '\p{General_Category=_ Connector_punctuation}', "");
Expect(0, 65343, '\p{^General_Category=_ Connector_punctuation}', "");
Expect(0, 65343, '\P{General_Category=_ Connector_punctuation}', "");
Expect(1, 65343, '\P{^General_Category=_ Connector_punctuation}', "");
Expect(0, 65344, '\p{General_Category=_ Connector_punctuation}', "");
Expect(1, 65344, '\p{^General_Category=_ Connector_punctuation}', "");
Expect(1, 65344, '\P{General_Category=_ Connector_punctuation}', "");
Expect(0, 65344, '\P{^General_Category=_ Connector_punctuation}', "");
Error('\p{Gc=-	PC/a/}');
Error('\P{Gc=-	PC/a/}');
Expect(1, 65343, '\p{Gc=pc}', "");
Expect(0, 65343, '\p{^Gc=pc}', "");
Expect(0, 65343, '\P{Gc=pc}', "");
Expect(1, 65343, '\P{^Gc=pc}', "");
Expect(0, 65344, '\p{Gc=pc}', "");
Expect(1, 65344, '\p{^Gc=pc}', "");
Expect(1, 65344, '\P{Gc=pc}', "");
Expect(0, 65344, '\P{^Gc=pc}', "");
Expect(1, 65343, '\p{Gc=-_Pc}', "");
Expect(0, 65343, '\p{^Gc=-_Pc}', "");
Expect(0, 65343, '\P{Gc=-_Pc}', "");
Expect(1, 65343, '\P{^Gc=-_Pc}', "");
Expect(0, 65344, '\p{Gc=-_Pc}', "");
Expect(1, 65344, '\p{^Gc=-_Pc}', "");
Expect(1, 65344, '\P{Gc=-_Pc}', "");
Expect(0, 65344, '\P{^Gc=-_Pc}', "");
Error('\p{Category=:=_	CONNECTOR_PUNCTUATION}');
Error('\P{Category=:=_	CONNECTOR_PUNCTUATION}');
Expect(1, 65343, '\p{Category=connectorpunctuation}', "");
Expect(0, 65343, '\p{^Category=connectorpunctuation}', "");
Expect(0, 65343, '\P{Category=connectorpunctuation}', "");
Expect(1, 65343, '\P{^Category=connectorpunctuation}', "");
Expect(0, 65344, '\p{Category=connectorpunctuation}', "");
Expect(1, 65344, '\p{^Category=connectorpunctuation}', "");
Expect(1, 65344, '\P{Category=connectorpunctuation}', "");
Expect(0, 65344, '\P{^Category=connectorpunctuation}', "");
Expect(1, 65343, '\p{Category: _Connector_Punctuation}', "");
Expect(0, 65343, '\p{^Category: _Connector_Punctuation}', "");
Expect(0, 65343, '\P{Category: _Connector_Punctuation}', "");
Expect(1, 65343, '\P{^Category: _Connector_Punctuation}', "");
Expect(0, 65344, '\p{Category: _Connector_Punctuation}', "");
Expect(1, 65344, '\p{^Category: _Connector_Punctuation}', "");
Expect(1, 65344, '\P{Category: _Connector_Punctuation}', "");
Expect(0, 65344, '\P{^Category: _Connector_Punctuation}', "");
Error('\p{Is_General_Category=	_Pc:=}');
Error('\P{Is_General_Category=	_Pc:=}');
Expect(1, 65343, '\p{Is_General_Category:   pc}', "");
Expect(0, 65343, '\p{^Is_General_Category:   pc}', "");
Expect(0, 65343, '\P{Is_General_Category:   pc}', "");
Expect(1, 65343, '\P{^Is_General_Category:   pc}', "");
Expect(0, 65344, '\p{Is_General_Category:   pc}', "");
Expect(1, 65344, '\p{^Is_General_Category:   pc}', "");
Expect(1, 65344, '\P{Is_General_Category:   pc}', "");
Expect(0, 65344, '\P{^Is_General_Category:   pc}', "");
Expect(1, 65343, '\p{Is_General_Category=-pc}', "");
Expect(0, 65343, '\p{^Is_General_Category=-pc}', "");
Expect(0, 65343, '\P{Is_General_Category=-pc}', "");
Expect(1, 65343, '\P{^Is_General_Category=-pc}', "");
Expect(0, 65344, '\p{Is_General_Category=-pc}', "");
Expect(1, 65344, '\p{^Is_General_Category=-pc}', "");
Expect(1, 65344, '\P{Is_General_Category=-pc}', "");
Expect(0, 65344, '\P{^Is_General_Category=-pc}', "");
Error('\p{Is_Gc=-/a/CONNECTOR_punctuation}');
Error('\P{Is_Gc=-/a/CONNECTOR_punctuation}');
Expect(1, 65343, '\p{Is_Gc=connectorpunctuation}', "");
Expect(0, 65343, '\p{^Is_Gc=connectorpunctuation}', "");
Expect(0, 65343, '\P{Is_Gc=connectorpunctuation}', "");
Expect(1, 65343, '\P{^Is_Gc=connectorpunctuation}', "");
Expect(0, 65344, '\p{Is_Gc=connectorpunctuation}', "");
Expect(1, 65344, '\p{^Is_Gc=connectorpunctuation}', "");
Expect(1, 65344, '\P{Is_Gc=connectorpunctuation}', "");
Expect(0, 65344, '\P{^Is_Gc=connectorpunctuation}', "");
Expect(1, 65343, '\p{Is_Gc=-	CONNECTOR_Punctuation}', "");
Expect(0, 65343, '\p{^Is_Gc=-	CONNECTOR_Punctuation}', "");
Expect(0, 65343, '\P{Is_Gc=-	CONNECTOR_Punctuation}', "");
Expect(1, 65343, '\P{^Is_Gc=-	CONNECTOR_Punctuation}', "");
Expect(0, 65344, '\p{Is_Gc=-	CONNECTOR_Punctuation}', "");
Expect(1, 65344, '\p{^Is_Gc=-	CONNECTOR_Punctuation}', "");
Expect(1, 65344, '\P{Is_Gc=-	CONNECTOR_Punctuation}', "");
Expect(0, 65344, '\P{^Is_Gc=-	CONNECTOR_Punctuation}', "");
Error('\p{Is_Category=:=PC}');
Error('\P{Is_Category=:=PC}');
Expect(1, 65343, '\p{Is_Category=pc}', "");
Expect(0, 65343, '\p{^Is_Category=pc}', "");
Expect(0, 65343, '\P{Is_Category=pc}', "");
Expect(1, 65343, '\P{^Is_Category=pc}', "");
Expect(0, 65344, '\p{Is_Category=pc}', "");
Expect(1, 65344, '\p{^Is_Category=pc}', "");
Expect(1, 65344, '\P{Is_Category=pc}', "");
Expect(0, 65344, '\P{^Is_Category=pc}', "");
Expect(1, 65343, '\p{Is_Category=- PC}', "");
Expect(0, 65343, '\p{^Is_Category=- PC}', "");
Expect(0, 65343, '\P{Is_Category=- PC}', "");
Expect(1, 65343, '\P{^Is_Category=- PC}', "");
Expect(0, 65344, '\p{Is_Category=- PC}', "");
Expect(1, 65344, '\p{^Is_Category=- PC}', "");
Expect(1, 65344, '\P{Is_Category=- PC}', "");
Expect(0, 65344, '\P{^Is_Category=- PC}', "");
Error('\p{General_Category= -Surrogate:=}');
Error('\P{General_Category= -Surrogate:=}');
Expect(1, 57343, '\p{General_Category=surrogate}', "");
Expect(0, 57343, '\p{^General_Category=surrogate}', "");
Expect(0, 57343, '\P{General_Category=surrogate}', "");
Expect(1, 57343, '\P{^General_Category=surrogate}', "");
Expect(0, 57344, '\p{General_Category=surrogate}', "");
Expect(1, 57344, '\p{^General_Category=surrogate}', "");
Expect(1, 57344, '\P{General_Category=surrogate}', "");
Expect(0, 57344, '\P{^General_Category=surrogate}', "");
Expect(1, 57343, '\p{General_Category=_Surrogate}', "");
Expect(0, 57343, '\p{^General_Category=_Surrogate}', "");
Expect(0, 57343, '\P{General_Category=_Surrogate}', "");
Expect(1, 57343, '\P{^General_Category=_Surrogate}', "");
Expect(0, 57344, '\p{General_Category=_Surrogate}', "");
Expect(1, 57344, '\p{^General_Category=_Surrogate}', "");
Expect(1, 57344, '\P{General_Category=_Surrogate}', "");
Expect(0, 57344, '\P{^General_Category=_Surrogate}', "");
Error('\p{Gc=:=Cs}');
Error('\P{Gc=:=Cs}');
Expect(1, 57343, '\p{Gc=cs}', "");
Expect(0, 57343, '\p{^Gc=cs}', "");
Expect(0, 57343, '\P{Gc=cs}', "");
Expect(1, 57343, '\P{^Gc=cs}', "");
Expect(0, 57344, '\p{Gc=cs}', "");
Expect(1, 57344, '\p{^Gc=cs}', "");
Expect(1, 57344, '\P{Gc=cs}', "");
Expect(0, 57344, '\P{^Gc=cs}', "");
Expect(1, 57343, '\p{Gc= cs}', "");
Expect(0, 57343, '\p{^Gc= cs}', "");
Expect(0, 57343, '\P{Gc= cs}', "");
Expect(1, 57343, '\P{^Gc= cs}', "");
Expect(0, 57344, '\p{Gc= cs}', "");
Expect(1, 57344, '\p{^Gc= cs}', "");
Expect(1, 57344, '\P{Gc= cs}', "");
Expect(0, 57344, '\P{^Gc= cs}', "");
Error('\p{Category=:=- Surrogate}');
Error('\P{Category=:=- Surrogate}');
Expect(1, 57343, '\p{Category=surrogate}', "");
Expect(0, 57343, '\p{^Category=surrogate}', "");
Expect(0, 57343, '\P{Category=surrogate}', "");
Expect(1, 57343, '\P{^Category=surrogate}', "");
Expect(0, 57344, '\p{Category=surrogate}', "");
Expect(1, 57344, '\p{^Category=surrogate}', "");
Expect(1, 57344, '\P{Category=surrogate}', "");
Expect(0, 57344, '\P{^Category=surrogate}', "");
Expect(1, 57343, '\p{Category= _SURROGATE}', "");
Expect(0, 57343, '\p{^Category= _SURROGATE}', "");
Expect(0, 57343, '\P{Category= _SURROGATE}', "");
Expect(1, 57343, '\P{^Category= _SURROGATE}', "");
Expect(0, 57344, '\p{Category= _SURROGATE}', "");
Expect(1, 57344, '\p{^Category= _SURROGATE}', "");
Expect(1, 57344, '\P{Category= _SURROGATE}', "");
Expect(0, 57344, '\P{^Category= _SURROGATE}', "");
Error('\p{Is_General_Category:/a/_-cs}');
Error('\P{Is_General_Category:/a/_-cs}');
Expect(1, 57343, '\p{Is_General_Category=cs}', "");
Expect(0, 57343, '\p{^Is_General_Category=cs}', "");
Expect(0, 57343, '\P{Is_General_Category=cs}', "");
Expect(1, 57343, '\P{^Is_General_Category=cs}', "");
Expect(0, 57344, '\p{Is_General_Category=cs}', "");
Expect(1, 57344, '\p{^Is_General_Category=cs}', "");
Expect(1, 57344, '\P{Is_General_Category=cs}', "");
Expect(0, 57344, '\P{^Is_General_Category=cs}', "");
Expect(1, 57343, '\p{Is_General_Category= -Cs}', "");
Expect(0, 57343, '\p{^Is_General_Category= -Cs}', "");
Expect(0, 57343, '\P{Is_General_Category= -Cs}', "");
Expect(1, 57343, '\P{^Is_General_Category= -Cs}', "");
Expect(0, 57344, '\p{Is_General_Category= -Cs}', "");
Expect(1, 57344, '\p{^Is_General_Category= -Cs}', "");
Expect(1, 57344, '\P{Is_General_Category= -Cs}', "");
Expect(0, 57344, '\P{^Is_General_Category= -Cs}', "");
Error('\p{Is_Gc=--Surrogate:=}');
Error('\P{Is_Gc=--Surrogate:=}');
Expect(1, 57343, '\p{Is_Gc=surrogate}', "");
Expect(0, 57343, '\p{^Is_Gc=surrogate}', "");
Expect(0, 57343, '\P{Is_Gc=surrogate}', "");
Expect(1, 57343, '\P{^Is_Gc=surrogate}', "");
Expect(0, 57344, '\p{Is_Gc=surrogate}', "");
Expect(1, 57344, '\p{^Is_Gc=surrogate}', "");
Expect(1, 57344, '\P{Is_Gc=surrogate}', "");
Expect(0, 57344, '\P{^Is_Gc=surrogate}', "");
Expect(1, 57343, '\p{Is_Gc: _	surrogate}', "");
Expect(0, 57343, '\p{^Is_Gc: _	surrogate}', "");
Expect(0, 57343, '\P{Is_Gc: _	surrogate}', "");
Expect(1, 57343, '\P{^Is_Gc: _	surrogate}', "");
Expect(0, 57344, '\p{Is_Gc: _	surrogate}', "");
Expect(1, 57344, '\p{^Is_Gc: _	surrogate}', "");
Expect(1, 57344, '\P{Is_Gc: _	surrogate}', "");
Expect(0, 57344, '\P{^Is_Gc: _	surrogate}', "");
Error('\p{Is_Category= /a/Cs}');
Error('\P{Is_Category= /a/Cs}');
Expect(1, 57343, '\p{Is_Category=cs}', "");
Expect(0, 57343, '\p{^Is_Category=cs}', "");
Expect(0, 57343, '\P{Is_Category=cs}', "");
Expect(1, 57343, '\P{^Is_Category=cs}', "");
Expect(0, 57344, '\p{Is_Category=cs}', "");
Expect(1, 57344, '\p{^Is_Category=cs}', "");
Expect(1, 57344, '\P{Is_Category=cs}', "");
Expect(0, 57344, '\P{^Is_Category=cs}', "");
Expect(1, 57343, '\p{Is_Category:	  Cs}', "");
Expect(0, 57343, '\p{^Is_Category:	  Cs}', "");
Expect(0, 57343, '\P{Is_Category:	  Cs}', "");
Expect(1, 57343, '\P{^Is_Category:	  Cs}', "");
Expect(0, 57344, '\p{Is_Category:	  Cs}', "");
Expect(1, 57344, '\p{^Is_Category:	  Cs}', "");
Expect(1, 57344, '\P{Is_Category:	  Cs}', "");
Expect(0, 57344, '\P{^Is_Category:	  Cs}', "");
Error('\p{General_Category=  PRIVATE_use:=}');
Error('\P{General_Category=  PRIVATE_use:=}');
Expect(1, 1114109, '\p{General_Category:privateuse}', "");
Expect(0, 1114109, '\p{^General_Category:privateuse}', "");
Expect(0, 1114109, '\P{General_Category:privateuse}', "");
Expect(1, 1114109, '\P{^General_Category:privateuse}', "");
Expect(0, 63744, '\p{General_Category:privateuse}', "");
Expect(1, 63744, '\p{^General_Category:privateuse}', "");
Expect(1, 63744, '\P{General_Category:privateuse}', "");
Expect(0, 63744, '\P{^General_Category:privateuse}', "");
Expect(1, 1114109, '\p{General_Category=-PRIVATE_USE}', "");
Expect(0, 1114109, '\p{^General_Category=-PRIVATE_USE}', "");
Expect(0, 1114109, '\P{General_Category=-PRIVATE_USE}', "");
Expect(1, 1114109, '\P{^General_Category=-PRIVATE_USE}', "");
Expect(0, 63744, '\p{General_Category=-PRIVATE_USE}', "");
Expect(1, 63744, '\p{^General_Category=-PRIVATE_USE}', "");
Expect(1, 63744, '\P{General_Category=-PRIVATE_USE}', "");
Expect(0, 63744, '\P{^General_Category=-PRIVATE_USE}', "");
Error('\p{Gc=-:=co}');
Error('\P{Gc=-:=co}');
Expect(1, 1114109, '\p{Gc=co}', "");
Expect(0, 1114109, '\p{^Gc=co}', "");
Expect(0, 1114109, '\P{Gc=co}', "");
Expect(1, 1114109, '\P{^Gc=co}', "");
Expect(0, 63744, '\p{Gc=co}', "");
Expect(1, 63744, '\p{^Gc=co}', "");
Expect(1, 63744, '\P{Gc=co}', "");
Expect(0, 63744, '\P{^Gc=co}', "");
Expect(1, 1114109, '\p{Gc=_ Co}', "");
Expect(0, 1114109, '\p{^Gc=_ Co}', "");
Expect(0, 1114109, '\P{Gc=_ Co}', "");
Expect(1, 1114109, '\P{^Gc=_ Co}', "");
Expect(0, 63744, '\p{Gc=_ Co}', "");
Expect(1, 63744, '\p{^Gc=_ Co}', "");
Expect(1, 63744, '\P{Gc=_ Co}', "");
Expect(0, 63744, '\P{^Gc=_ Co}', "");
Error('\p{Category: /a/	PRIVATE_use}');
Error('\P{Category: /a/	PRIVATE_use}');
Expect(1, 1114109, '\p{Category=privateuse}', "");
Expect(0, 1114109, '\p{^Category=privateuse}', "");
Expect(0, 1114109, '\P{Category=privateuse}', "");
Expect(1, 1114109, '\P{^Category=privateuse}', "");
Expect(0, 63744, '\p{Category=privateuse}', "");
Expect(1, 63744, '\p{^Category=privateuse}', "");
Expect(1, 63744, '\P{Category=privateuse}', "");
Expect(0, 63744, '\P{^Category=privateuse}', "");
Expect(1, 1114109, '\p{Category:  Private_Use}', "");
Expect(0, 1114109, '\p{^Category:  Private_Use}', "");
Expect(0, 1114109, '\P{Category:  Private_Use}', "");
Expect(1, 1114109, '\P{^Category:  Private_Use}', "");
Expect(0, 63744, '\p{Category:  Private_Use}', "");
Expect(1, 63744, '\p{^Category:  Private_Use}', "");
Expect(1, 63744, '\P{Category:  Private_Use}', "");
Expect(0, 63744, '\P{^Category:  Private_Use}', "");
Error('\p{Is_General_Category= co/a/}');
Error('\P{Is_General_Category= co/a/}');
Expect(1, 1114109, '\p{Is_General_Category=co}', "");
Expect(0, 1114109, '\p{^Is_General_Category=co}', "");
Expect(0, 1114109, '\P{Is_General_Category=co}', "");
Expect(1, 1114109, '\P{^Is_General_Category=co}', "");
Expect(0, 63744, '\p{Is_General_Category=co}', "");
Expect(1, 63744, '\p{^Is_General_Category=co}', "");
Expect(1, 63744, '\P{Is_General_Category=co}', "");
Expect(0, 63744, '\P{^Is_General_Category=co}', "");
Expect(1, 1114109, '\p{Is_General_Category=	_Co}', "");
Expect(0, 1114109, '\p{^Is_General_Category=	_Co}', "");
Expect(0, 1114109, '\P{Is_General_Category=	_Co}', "");
Expect(1, 1114109, '\P{^Is_General_Category=	_Co}', "");
Expect(0, 63744, '\p{Is_General_Category=	_Co}', "");
Expect(1, 63744, '\p{^Is_General_Category=	_Co}', "");
Expect(1, 63744, '\P{Is_General_Category=	_Co}', "");
Expect(0, 63744, '\P{^Is_General_Category=	_Co}', "");
Error('\p{Is_Gc= /a/private_Use}');
Error('\P{Is_Gc= /a/private_Use}');
Expect(1, 1114109, '\p{Is_Gc=privateuse}', "");
Expect(0, 1114109, '\p{^Is_Gc=privateuse}', "");
Expect(0, 1114109, '\P{Is_Gc=privateuse}', "");
Expect(1, 1114109, '\P{^Is_Gc=privateuse}', "");
Expect(0, 63744, '\p{Is_Gc=privateuse}', "");
Expect(1, 63744, '\p{^Is_Gc=privateuse}', "");
Expect(1, 63744, '\P{Is_Gc=privateuse}', "");
Expect(0, 63744, '\P{^Is_Gc=privateuse}', "");
Expect(1, 1114109, '\p{Is_Gc=-_Private_Use}', "");
Expect(0, 1114109, '\p{^Is_Gc=-_Private_Use}', "");
Expect(0, 1114109, '\P{Is_Gc=-_Private_Use}', "");
Expect(1, 1114109, '\P{^Is_Gc=-_Private_Use}', "");
Expect(0, 63744, '\p{Is_Gc=-_Private_Use}', "");
Expect(1, 63744, '\p{^Is_Gc=-_Private_Use}', "");
Expect(1, 63744, '\P{Is_Gc=-_Private_Use}', "");
Expect(0, 63744, '\P{^Is_Gc=-_Private_Use}', "");
Error('\p{Is_Category=/a/_ co}');
Error('\P{Is_Category=/a/_ co}');
Expect(1, 1114109, '\p{Is_Category=co}', "");
Expect(0, 1114109, '\p{^Is_Category=co}', "");
Expect(0, 1114109, '\P{Is_Category=co}', "");
Expect(1, 1114109, '\P{^Is_Category=co}', "");
Expect(0, 63744, '\p{Is_Category=co}', "");
Expect(1, 63744, '\p{^Is_Category=co}', "");
Expect(1, 63744, '\P{Is_Category=co}', "");
Expect(0, 63744, '\P{^Is_Category=co}', "");
Expect(1, 1114109, '\p{Is_Category=-	co}', "");
Expect(0, 1114109, '\p{^Is_Category=-	co}', "");
Expect(0, 1114109, '\P{Is_Category=-	co}', "");
Expect(1, 1114109, '\P{^Is_Category=-	co}', "");
Expect(0, 63744, '\p{Is_Category=-	co}', "");
Expect(1, 63744, '\p{^Is_Category=-	co}', "");
Expect(1, 63744, '\P{Is_Category=-	co}', "");
Expect(0, 63744, '\P{^Is_Category=-	co}', "");
Error('\p{General_Category=:=	lowercase_letter}');
Error('\P{General_Category=:=	lowercase_letter}');
Expect(1, 120779, '\p{General_Category:lowercaseletter}', "");
Expect(0, 120779, '\p{^General_Category:lowercaseletter}', "");
Expect(0, 120779, '\P{General_Category:lowercaseletter}', "");
Expect(1, 120779, '\P{^General_Category:lowercaseletter}', "");
Expect(0, 120780, '\p{General_Category:lowercaseletter}', "");
Expect(1, 120780, '\p{^General_Category:lowercaseletter}', "");
Expect(1, 120780, '\P{General_Category:lowercaseletter}', "");
Expect(0, 120780, '\P{^General_Category:lowercaseletter}', "");
Expect(1, 120779, '\p{General_Category=	Lowercase_letter}', "");
Expect(0, 120779, '\p{^General_Category=	Lowercase_letter}', "");
Expect(0, 120779, '\P{General_Category=	Lowercase_letter}', "");
Expect(1, 120779, '\P{^General_Category=	Lowercase_letter}', "");
Expect(0, 120780, '\p{General_Category=	Lowercase_letter}', "");
Expect(1, 120780, '\p{^General_Category=	Lowercase_letter}', "");
Expect(1, 120780, '\P{General_Category=	Lowercase_letter}', "");
Expect(0, 120780, '\P{^General_Category=	Lowercase_letter}', "");
Error('\p{Gc=:=	Ll}');
Error('\P{Gc=:=	Ll}');
Expect(1, 120779, '\p{Gc=ll}', "");
Expect(0, 120779, '\p{^Gc=ll}', "");
Expect(0, 120779, '\P{Gc=ll}', "");
Expect(1, 120779, '\P{^Gc=ll}', "");
Expect(0, 120780, '\p{Gc=ll}', "");
Expect(1, 120780, '\p{^Gc=ll}', "");
Expect(1, 120780, '\P{Gc=ll}', "");
Expect(0, 120780, '\P{^Gc=ll}', "");
Expect(1, 120779, '\p{Gc= -ll}', "");
Expect(0, 120779, '\p{^Gc= -ll}', "");
Expect(0, 120779, '\P{Gc= -ll}', "");
Expect(1, 120779, '\P{^Gc= -ll}', "");
Expect(0, 120780, '\p{Gc= -ll}', "");
Expect(1, 120780, '\p{^Gc= -ll}', "");
Expect(1, 120780, '\P{Gc= -ll}', "");
Expect(0, 120780, '\P{^Gc= -ll}', "");
Error('\p{Category=/a/ _Lowercase_LETTER}');
Error('\P{Category=/a/ _Lowercase_LETTER}');
Expect(1, 120779, '\p{Category=lowercaseletter}', "");
Expect(0, 120779, '\p{^Category=lowercaseletter}', "");
Expect(0, 120779, '\P{Category=lowercaseletter}', "");
Expect(1, 120779, '\P{^Category=lowercaseletter}', "");
Expect(0, 120780, '\p{Category=lowercaseletter}', "");
Expect(1, 120780, '\p{^Category=lowercaseletter}', "");
Expect(1, 120780, '\P{Category=lowercaseletter}', "");
Expect(0, 120780, '\P{^Category=lowercaseletter}', "");
Expect(1, 120779, '\p{Category=	_LOWERCASE_letter}', "");
Expect(0, 120779, '\p{^Category=	_LOWERCASE_letter}', "");
Expect(0, 120779, '\P{Category=	_LOWERCASE_letter}', "");
Expect(1, 120779, '\P{^Category=	_LOWERCASE_letter}', "");
Expect(0, 120780, '\p{Category=	_LOWERCASE_letter}', "");
Expect(1, 120780, '\p{^Category=	_LOWERCASE_letter}', "");
Expect(1, 120780, '\P{Category=	_LOWERCASE_letter}', "");
Expect(0, 120780, '\P{^Category=	_LOWERCASE_letter}', "");
Error('\p{Is_General_Category=:=_-Ll}');
Error('\P{Is_General_Category=:=_-Ll}');
Expect(1, 120779, '\p{Is_General_Category=ll}', "");
Expect(0, 120779, '\p{^Is_General_Category=ll}', "");
Expect(0, 120779, '\P{Is_General_Category=ll}', "");
Expect(1, 120779, '\P{^Is_General_Category=ll}', "");
Expect(0, 120780, '\p{Is_General_Category=ll}', "");
Expect(1, 120780, '\p{^Is_General_Category=ll}', "");
Expect(1, 120780, '\P{Is_General_Category=ll}', "");
Expect(0, 120780, '\P{^Is_General_Category=ll}', "");
Expect(1, 120779, '\p{Is_General_Category=_ll}', "");
Expect(0, 120779, '\p{^Is_General_Category=_ll}', "");
Expect(0, 120779, '\P{Is_General_Category=_ll}', "");
Expect(1, 120779, '\P{^Is_General_Category=_ll}', "");
Expect(0, 120780, '\p{Is_General_Category=_ll}', "");
Expect(1, 120780, '\p{^Is_General_Category=_ll}', "");
Expect(1, 120780, '\P{Is_General_Category=_ll}', "");
Expect(0, 120780, '\P{^Is_General_Category=_ll}', "");
Error('\p{Is_Gc=/a/Lowercase_letter}');
Error('\P{Is_Gc=/a/Lowercase_letter}');
Expect(1, 120779, '\p{Is_Gc=lowercaseletter}', "");
Expect(0, 120779, '\p{^Is_Gc=lowercaseletter}', "");
Expect(0, 120779, '\P{Is_Gc=lowercaseletter}', "");
Expect(1, 120779, '\P{^Is_Gc=lowercaseletter}', "");
Expect(0, 120780, '\p{Is_Gc=lowercaseletter}', "");
Expect(1, 120780, '\p{^Is_Gc=lowercaseletter}', "");
Expect(1, 120780, '\P{Is_Gc=lowercaseletter}', "");
Expect(0, 120780, '\P{^Is_Gc=lowercaseletter}', "");
Expect(1, 120779, '\p{Is_Gc=__LOWERCASE_Letter}', "");
Expect(0, 120779, '\p{^Is_Gc=__LOWERCASE_Letter}', "");
Expect(0, 120779, '\P{Is_Gc=__LOWERCASE_Letter}', "");
Expect(1, 120779, '\P{^Is_Gc=__LOWERCASE_Letter}', "");
Expect(0, 120780, '\p{Is_Gc=__LOWERCASE_Letter}', "");
Expect(1, 120780, '\p{^Is_Gc=__LOWERCASE_Letter}', "");
Expect(1, 120780, '\P{Is_Gc=__LOWERCASE_Letter}', "");
Expect(0, 120780, '\P{^Is_Gc=__LOWERCASE_Letter}', "");
Error('\p{Is_Category=-/a/ll}');
Error('\P{Is_Category=-/a/ll}');
Expect(1, 120779, '\p{Is_Category=ll}', "");
Expect(0, 120779, '\p{^Is_Category=ll}', "");
Expect(0, 120779, '\P{Is_Category=ll}', "");
Expect(1, 120779, '\P{^Is_Category=ll}', "");
Expect(0, 120780, '\p{Is_Category=ll}', "");
Expect(1, 120780, '\p{^Is_Category=ll}', "");
Expect(1, 120780, '\P{Is_Category=ll}', "");
Expect(0, 120780, '\P{^Is_Category=ll}', "");
Expect(1, 120779, '\p{Is_Category= _LL}', "");
Expect(0, 120779, '\p{^Is_Category= _LL}', "");
Expect(0, 120779, '\P{Is_Category= _LL}', "");
Expect(1, 120779, '\P{^Is_Category= _LL}', "");
Expect(0, 120780, '\p{Is_Category= _LL}', "");
Expect(1, 120780, '\p{^Is_Category= _LL}', "");
Expect(1, 120780, '\P{Is_Category= _LL}', "");
Expect(0, 120780, '\P{^Is_Category= _LL}', "");
Error('\p{General_Category=:=-	Close_PUNCTUATION}');
Error('\P{General_Category=:=-	Close_PUNCTUATION}');
Expect(1, 65379, '\p{General_Category=closepunctuation}', "");
Expect(0, 65379, '\p{^General_Category=closepunctuation}', "");
Expect(0, 65379, '\P{General_Category=closepunctuation}', "");
Expect(1, 65379, '\P{^General_Category=closepunctuation}', "");
Expect(0, 65380, '\p{General_Category=closepunctuation}', "");
Expect(1, 65380, '\p{^General_Category=closepunctuation}', "");
Expect(1, 65380, '\P{General_Category=closepunctuation}', "");
Expect(0, 65380, '\P{^General_Category=closepunctuation}', "");
Expect(1, 65379, '\p{General_Category= close_Punctuation}', "");
Expect(0, 65379, '\p{^General_Category= close_Punctuation}', "");
Expect(0, 65379, '\P{General_Category= close_Punctuation}', "");
Expect(1, 65379, '\P{^General_Category= close_Punctuation}', "");
Expect(0, 65380, '\p{General_Category= close_Punctuation}', "");
Expect(1, 65380, '\p{^General_Category= close_Punctuation}', "");
Expect(1, 65380, '\P{General_Category= close_Punctuation}', "");
Expect(0, 65380, '\P{^General_Category= close_Punctuation}', "");
Error('\p{Gc=Pe/a/}');
Error('\P{Gc=Pe/a/}');
Expect(1, 65379, '\p{Gc=pe}', "");
Expect(0, 65379, '\p{^Gc=pe}', "");
Expect(0, 65379, '\P{Gc=pe}', "");
Expect(1, 65379, '\P{^Gc=pe}', "");
Expect(0, 65380, '\p{Gc=pe}', "");
Expect(1, 65380, '\p{^Gc=pe}', "");
Expect(1, 65380, '\P{Gc=pe}', "");
Expect(0, 65380, '\P{^Gc=pe}', "");
Expect(1, 65379, '\p{Gc: PE}', "");
Expect(0, 65379, '\p{^Gc: PE}', "");
Expect(0, 65379, '\P{Gc: PE}', "");
Expect(1, 65379, '\P{^Gc: PE}', "");
Expect(0, 65380, '\p{Gc: PE}', "");
Expect(1, 65380, '\p{^Gc: PE}', "");
Expect(1, 65380, '\P{Gc: PE}', "");
Expect(0, 65380, '\P{^Gc: PE}', "");
Error('\p{Category=:=_-Close_Punctuation}');
Error('\P{Category=:=_-Close_Punctuation}');
Expect(1, 65379, '\p{Category=closepunctuation}', "");
Expect(0, 65379, '\p{^Category=closepunctuation}', "");
Expect(0, 65379, '\P{Category=closepunctuation}', "");
Expect(1, 65379, '\P{^Category=closepunctuation}', "");
Expect(0, 65380, '\p{Category=closepunctuation}', "");
Expect(1, 65380, '\p{^Category=closepunctuation}', "");
Expect(1, 65380, '\P{Category=closepunctuation}', "");
Expect(0, 65380, '\P{^Category=closepunctuation}', "");
Expect(1, 65379, '\p{Category=_ Close_punctuation}', "");
Expect(0, 65379, '\p{^Category=_ Close_punctuation}', "");
Expect(0, 65379, '\P{Category=_ Close_punctuation}', "");
Expect(1, 65379, '\P{^Category=_ Close_punctuation}', "");
Expect(0, 65380, '\p{Category=_ Close_punctuation}', "");
Expect(1, 65380, '\p{^Category=_ Close_punctuation}', "");
Expect(1, 65380, '\P{Category=_ Close_punctuation}', "");
Expect(0, 65380, '\P{^Category=_ Close_punctuation}', "");
Error('\p{Is_General_Category=:=Pe}');
Error('\P{Is_General_Category=:=Pe}');
Expect(1, 65379, '\p{Is_General_Category:	pe}', "");
Expect(0, 65379, '\p{^Is_General_Category:	pe}', "");
Expect(0, 65379, '\P{Is_General_Category:	pe}', "");
Expect(1, 65379, '\P{^Is_General_Category:	pe}', "");
Expect(0, 65380, '\p{Is_General_Category:	pe}', "");
Expect(1, 65380, '\p{^Is_General_Category:	pe}', "");
Expect(1, 65380, '\P{Is_General_Category:	pe}', "");
Expect(0, 65380, '\P{^Is_General_Category:	pe}', "");
Expect(1, 65379, '\p{Is_General_Category= PE}', "");
Expect(0, 65379, '\p{^Is_General_Category= PE}', "");
Expect(0, 65379, '\P{Is_General_Category= PE}', "");
Expect(1, 65379, '\P{^Is_General_Category= PE}', "");
Expect(0, 65380, '\p{Is_General_Category= PE}', "");
Expect(1, 65380, '\p{^Is_General_Category= PE}', "");
Expect(1, 65380, '\P{Is_General_Category= PE}', "");
Expect(0, 65380, '\P{^Is_General_Category= PE}', "");
Error('\p{Is_Gc=	close_punctuation/a/}');
Error('\P{Is_Gc=	close_punctuation/a/}');
Expect(1, 65379, '\p{Is_Gc=closepunctuation}', "");
Expect(0, 65379, '\p{^Is_Gc=closepunctuation}', "");
Expect(0, 65379, '\P{Is_Gc=closepunctuation}', "");
Expect(1, 65379, '\P{^Is_Gc=closepunctuation}', "");
Expect(0, 65380, '\p{Is_Gc=closepunctuation}', "");
Expect(1, 65380, '\p{^Is_Gc=closepunctuation}', "");
Expect(1, 65380, '\P{Is_Gc=closepunctuation}', "");
Expect(0, 65380, '\P{^Is_Gc=closepunctuation}', "");
Expect(1, 65379, '\p{Is_Gc=	Close_Punctuation}', "");
Expect(0, 65379, '\p{^Is_Gc=	Close_Punctuation}', "");
Expect(0, 65379, '\P{Is_Gc=	Close_Punctuation}', "");
Expect(1, 65379, '\P{^Is_Gc=	Close_Punctuation}', "");
Expect(0, 65380, '\p{Is_Gc=	Close_Punctuation}', "");
Expect(1, 65380, '\p{^Is_Gc=	Close_Punctuation}', "");
Expect(1, 65380, '\P{Is_Gc=	Close_Punctuation}', "");
Expect(0, 65380, '\P{^Is_Gc=	Close_Punctuation}', "");
Error('\p{Is_Category=_/a/pe}');
Error('\P{Is_Category=_/a/pe}');
Expect(1, 65379, '\p{Is_Category=pe}', "");
Expect(0, 65379, '\p{^Is_Category=pe}', "");
Expect(0, 65379, '\P{Is_Category=pe}', "");
Expect(1, 65379, '\P{^Is_Category=pe}', "");
Expect(0, 65380, '\p{Is_Category=pe}', "");
Expect(1, 65380, '\p{^Is_Category=pe}', "");
Expect(1, 65380, '\P{Is_Category=pe}', "");
Expect(0, 65380, '\P{^Is_Category=pe}', "");
Expect(1, 65379, '\p{Is_Category=__Pe}', "");
Expect(0, 65379, '\p{^Is_Category=__Pe}', "");
Expect(0, 65379, '\P{Is_Category=__Pe}', "");
Expect(1, 65379, '\P{^Is_Category=__Pe}', "");
Expect(0, 65380, '\p{Is_Category=__Pe}', "");
Expect(1, 65380, '\p{^Is_Category=__Pe}', "");
Expect(1, 65380, '\P{Is_Category=__Pe}', "");
Expect(0, 65380, '\P{^Is_Category=__Pe}', "");
Error('\p{General_Category= Letter:=}');
Error('\P{General_Category= Letter:=}');
Expect(1, 195101, '\p{General_Category:letter}', "");
Expect(0, 195101, '\p{^General_Category:letter}', "");
Expect(0, 195101, '\P{General_Category:letter}', "");
Expect(1, 195101, '\P{^General_Category:letter}', "");
Expect(0, 195102, '\p{General_Category:letter}', "");
Expect(1, 195102, '\p{^General_Category:letter}', "");
Expect(1, 195102, '\P{General_Category:letter}', "");
Expect(0, 195102, '\P{^General_Category:letter}', "");
Expect(1, 195101, '\p{General_Category=	LETTER}', "");
Expect(0, 195101, '\p{^General_Category=	LETTER}', "");
Expect(0, 195101, '\P{General_Category=	LETTER}', "");
Expect(1, 195101, '\P{^General_Category=	LETTER}', "");
Expect(0, 195102, '\p{General_Category=	LETTER}', "");
Expect(1, 195102, '\p{^General_Category=	LETTER}', "");
Expect(1, 195102, '\P{General_Category=	LETTER}', "");
Expect(0, 195102, '\P{^General_Category=	LETTER}', "");
Error('\p{Gc=:=	-l}');
Error('\P{Gc=:=	-l}');
Expect(1, 195101, '\p{Gc=l}', "");
Expect(0, 195101, '\p{^Gc=l}', "");
Expect(0, 195101, '\P{Gc=l}', "");
Expect(1, 195101, '\P{^Gc=l}', "");
Expect(0, 195102, '\p{Gc=l}', "");
Expect(1, 195102, '\p{^Gc=l}', "");
Expect(1, 195102, '\P{Gc=l}', "");
Expect(0, 195102, '\P{^Gc=l}', "");
Expect(1, 195101, '\p{Gc=	-L}', "");
Expect(0, 195101, '\p{^Gc=	-L}', "");
Expect(0, 195101, '\P{Gc=	-L}', "");
Expect(1, 195101, '\P{^Gc=	-L}', "");
Expect(0, 195102, '\p{Gc=	-L}', "");
Expect(1, 195102, '\p{^Gc=	-L}', "");
Expect(1, 195102, '\P{Gc=	-L}', "");
Expect(0, 195102, '\P{^Gc=	-L}', "");
Error('\p{Category=--LETTER:=}');
Error('\P{Category=--LETTER:=}');
Expect(1, 195101, '\p{Category=letter}', "");
Expect(0, 195101, '\p{^Category=letter}', "");
Expect(0, 195101, '\P{Category=letter}', "");
Expect(1, 195101, '\P{^Category=letter}', "");
Expect(0, 195102, '\p{Category=letter}', "");
Expect(1, 195102, '\p{^Category=letter}', "");
Expect(1, 195102, '\P{Category=letter}', "");
Expect(0, 195102, '\P{^Category=letter}', "");
Expect(1, 195101, '\p{Category= letter}', "");
Expect(0, 195101, '\p{^Category= letter}', "");
Expect(0, 195101, '\P{Category= letter}', "");
Expect(1, 195101, '\P{^Category= letter}', "");
Expect(0, 195102, '\p{Category= letter}', "");
Expect(1, 195102, '\p{^Category= letter}', "");
Expect(1, 195102, '\P{Category= letter}', "");
Expect(0, 195102, '\P{^Category= letter}', "");
Error('\p{Is_General_Category=:=	-L}');
Error('\P{Is_General_Category=:=	-L}');
Expect(1, 195101, '\p{Is_General_Category=l}', "");
Expect(0, 195101, '\p{^Is_General_Category=l}', "");
Expect(0, 195101, '\P{Is_General_Category=l}', "");
Expect(1, 195101, '\P{^Is_General_Category=l}', "");
Expect(0, 195102, '\p{Is_General_Category=l}', "");
Expect(1, 195102, '\p{^Is_General_Category=l}', "");
Expect(1, 195102, '\P{Is_General_Category=l}', "");
Expect(0, 195102, '\P{^Is_General_Category=l}', "");
Expect(1, 195101, '\p{Is_General_Category=--l}', "");
Expect(0, 195101, '\p{^Is_General_Category=--l}', "");
Expect(0, 195101, '\P{Is_General_Category=--l}', "");
Expect(1, 195101, '\P{^Is_General_Category=--l}', "");
Expect(0, 195102, '\p{Is_General_Category=--l}', "");
Expect(1, 195102, '\p{^Is_General_Category=--l}', "");
Expect(1, 195102, '\P{Is_General_Category=--l}', "");
Expect(0, 195102, '\P{^Is_General_Category=--l}', "");
Error('\p{Is_Gc= _Letter/a/}');
Error('\P{Is_Gc= _Letter/a/}');
Expect(1, 195101, '\p{Is_Gc=letter}', "");
Expect(0, 195101, '\p{^Is_Gc=letter}', "");
Expect(0, 195101, '\P{Is_Gc=letter}', "");
Expect(1, 195101, '\P{^Is_Gc=letter}', "");
Expect(0, 195102, '\p{Is_Gc=letter}', "");
Expect(1, 195102, '\p{^Is_Gc=letter}', "");
Expect(1, 195102, '\P{Is_Gc=letter}', "");
Expect(0, 195102, '\P{^Is_Gc=letter}', "");
Expect(1, 195101, '\p{Is_Gc:		LETTER}', "");
Expect(0, 195101, '\p{^Is_Gc:		LETTER}', "");
Expect(0, 195101, '\P{Is_Gc:		LETTER}', "");
Expect(1, 195101, '\P{^Is_Gc:		LETTER}', "");
Expect(0, 195102, '\p{Is_Gc:		LETTER}', "");
Expect(1, 195102, '\p{^Is_Gc:		LETTER}', "");
Expect(1, 195102, '\P{Is_Gc:		LETTER}', "");
Expect(0, 195102, '\P{^Is_Gc:		LETTER}', "");
Error('\p{Is_Category=/a/_L}');
Error('\P{Is_Category=/a/_L}');
Expect(1, 195101, '\p{Is_Category=l}', "");
Expect(0, 195101, '\p{^Is_Category=l}', "");
Expect(0, 195101, '\P{Is_Category=l}', "");
Expect(1, 195101, '\P{^Is_Category=l}', "");
Expect(0, 195102, '\p{Is_Category=l}', "");
Expect(1, 195102, '\p{^Is_Category=l}', "");
Expect(1, 195102, '\P{Is_Category=l}', "");
Expect(0, 195102, '\P{^Is_Category=l}', "");
Expect(1, 195101, '\p{Is_Category=	L}', "");
Expect(0, 195101, '\p{^Is_Category=	L}', "");
Expect(0, 195101, '\P{Is_Category=	L}', "");
Expect(1, 195101, '\P{^Is_Category=	L}', "");
Expect(0, 195102, '\p{Is_Category=	L}', "");
Expect(1, 195102, '\p{^Is_Category=	L}', "");
Expect(1, 195102, '\P{Is_Category=	L}', "");
Expect(0, 195102, '\P{^Is_Category=	L}', "");
Error('\p{General_Category=/a/-	MODIFIER_Symbol}');
Error('\P{General_Category=/a/-	MODIFIER_Symbol}');
Expect(1, 65507, '\p{General_Category=modifiersymbol}', "");
Expect(0, 65507, '\p{^General_Category=modifiersymbol}', "");
Expect(0, 65507, '\P{General_Category=modifiersymbol}', "");
Expect(1, 65507, '\P{^General_Category=modifiersymbol}', "");
Expect(0, 65508, '\p{General_Category=modifiersymbol}', "");
Expect(1, 65508, '\p{^General_Category=modifiersymbol}', "");
Expect(1, 65508, '\P{General_Category=modifiersymbol}', "");
Expect(0, 65508, '\P{^General_Category=modifiersymbol}', "");
Expect(1, 65507, '\p{General_Category:     Modifier_SYMBOL}', "");
Expect(0, 65507, '\p{^General_Category:     Modifier_SYMBOL}', "");
Expect(0, 65507, '\P{General_Category:     Modifier_SYMBOL}', "");
Expect(1, 65507, '\P{^General_Category:     Modifier_SYMBOL}', "");
Expect(0, 65508, '\p{General_Category:     Modifier_SYMBOL}', "");
Expect(1, 65508, '\p{^General_Category:     Modifier_SYMBOL}', "");
Expect(1, 65508, '\P{General_Category:     Modifier_SYMBOL}', "");
Expect(0, 65508, '\P{^General_Category:     Modifier_SYMBOL}', "");
Error('\p{Gc=:=Sk}');
Error('\P{Gc=:=Sk}');
Expect(1, 65507, '\p{Gc=sk}', "");
Expect(0, 65507, '\p{^Gc=sk}', "");
Expect(0, 65507, '\P{Gc=sk}', "");
Expect(1, 65507, '\P{^Gc=sk}', "");
Expect(0, 65508, '\p{Gc=sk}', "");
Expect(1, 65508, '\p{^Gc=sk}', "");
Expect(1, 65508, '\P{Gc=sk}', "");
Expect(0, 65508, '\P{^Gc=sk}', "");
Expect(1, 65507, '\p{Gc=-_Sk}', "");
Expect(0, 65507, '\p{^Gc=-_Sk}', "");
Expect(0, 65507, '\P{Gc=-_Sk}', "");
Expect(1, 65507, '\P{^Gc=-_Sk}', "");
Expect(0, 65508, '\p{Gc=-_Sk}', "");
Expect(1, 65508, '\p{^Gc=-_Sk}', "");
Expect(1, 65508, '\P{Gc=-_Sk}', "");
Expect(0, 65508, '\P{^Gc=-_Sk}', "");
Error('\p{Category=-/a/modifier_Symbol}');
Error('\P{Category=-/a/modifier_Symbol}');
Expect(1, 65507, '\p{Category=modifiersymbol}', "");
Expect(0, 65507, '\p{^Category=modifiersymbol}', "");
Expect(0, 65507, '\P{Category=modifiersymbol}', "");
Expect(1, 65507, '\P{^Category=modifiersymbol}', "");
Expect(0, 65508, '\p{Category=modifiersymbol}', "");
Expect(1, 65508, '\p{^Category=modifiersymbol}', "");
Expect(1, 65508, '\P{Category=modifiersymbol}', "");
Expect(0, 65508, '\P{^Category=modifiersymbol}', "");
Expect(1, 65507, '\p{Category=-_Modifier_SYMBOL}', "");
Expect(0, 65507, '\p{^Category=-_Modifier_SYMBOL}', "");
Expect(0, 65507, '\P{Category=-_Modifier_SYMBOL}', "");
Expect(1, 65507, '\P{^Category=-_Modifier_SYMBOL}', "");
Expect(0, 65508, '\p{Category=-_Modifier_SYMBOL}', "");
Expect(1, 65508, '\p{^Category=-_Modifier_SYMBOL}', "");
Expect(1, 65508, '\P{Category=-_Modifier_SYMBOL}', "");
Expect(0, 65508, '\P{^Category=-_Modifier_SYMBOL}', "");
Error('\p{Is_General_Category=/a/ Sk}');
Error('\P{Is_General_Category=/a/ Sk}');
Expect(1, 65507, '\p{Is_General_Category:sk}', "");
Expect(0, 65507, '\p{^Is_General_Category:sk}', "");
Expect(0, 65507, '\P{Is_General_Category:sk}', "");
Expect(1, 65507, '\P{^Is_General_Category:sk}', "");
Expect(0, 65508, '\p{Is_General_Category:sk}', "");
Expect(1, 65508, '\p{^Is_General_Category:sk}', "");
Expect(1, 65508, '\P{Is_General_Category:sk}', "");
Expect(0, 65508, '\P{^Is_General_Category:sk}', "");
Expect(1, 65507, '\p{Is_General_Category=__SK}', "");
Expect(0, 65507, '\p{^Is_General_Category=__SK}', "");
Expect(0, 65507, '\P{Is_General_Category=__SK}', "");
Expect(1, 65507, '\P{^Is_General_Category=__SK}', "");
Expect(0, 65508, '\p{Is_General_Category=__SK}', "");
Expect(1, 65508, '\p{^Is_General_Category=__SK}', "");
Expect(1, 65508, '\P{Is_General_Category=__SK}', "");
Expect(0, 65508, '\P{^Is_General_Category=__SK}', "");
Error('\p{Is_Gc=_-Modifier_Symbol:=}');
Error('\P{Is_Gc=_-Modifier_Symbol:=}');
Expect(1, 65507, '\p{Is_Gc=modifiersymbol}', "");
Expect(0, 65507, '\p{^Is_Gc=modifiersymbol}', "");
Expect(0, 65507, '\P{Is_Gc=modifiersymbol}', "");
Expect(1, 65507, '\P{^Is_Gc=modifiersymbol}', "");
Expect(0, 65508, '\p{Is_Gc=modifiersymbol}', "");
Expect(1, 65508, '\p{^Is_Gc=modifiersymbol}', "");
Expect(1, 65508, '\P{Is_Gc=modifiersymbol}', "");
Expect(0, 65508, '\P{^Is_Gc=modifiersymbol}', "");
Expect(1, 65507, '\p{Is_Gc=__Modifier_Symbol}', "");
Expect(0, 65507, '\p{^Is_Gc=__Modifier_Symbol}', "");
Expect(0, 65507, '\P{Is_Gc=__Modifier_Symbol}', "");
Expect(1, 65507, '\P{^Is_Gc=__Modifier_Symbol}', "");
Expect(0, 65508, '\p{Is_Gc=__Modifier_Symbol}', "");
Expect(1, 65508, '\p{^Is_Gc=__Modifier_Symbol}', "");
Expect(1, 65508, '\P{Is_Gc=__Modifier_Symbol}', "");
Expect(0, 65508, '\P{^Is_Gc=__Modifier_Symbol}', "");
Error('\p{Is_Category=-_SK:=}');
Error('\P{Is_Category=-_SK:=}');
Expect(1, 65507, '\p{Is_Category=sk}', "");
Expect(0, 65507, '\p{^Is_Category=sk}', "");
Expect(0, 65507, '\P{Is_Category=sk}', "");
Expect(1, 65507, '\P{^Is_Category=sk}', "");
Expect(0, 65508, '\p{Is_Category=sk}', "");
Expect(1, 65508, '\p{^Is_Category=sk}', "");
Expect(1, 65508, '\P{Is_Category=sk}', "");
Expect(0, 65508, '\P{^Is_Category=sk}', "");
Expect(1, 65507, '\p{Is_Category=-	Sk}', "");
Expect(0, 65507, '\p{^Is_Category=-	Sk}', "");
Expect(0, 65507, '\P{Is_Category=-	Sk}', "");
Expect(1, 65507, '\P{^Is_Category=-	Sk}', "");
Expect(0, 65508, '\p{Is_Category=-	Sk}', "");
Expect(1, 65508, '\p{^Is_Category=-	Sk}', "");
Expect(1, 65508, '\P{Is_Category=-	Sk}', "");
Expect(0, 65508, '\P{^Is_Category=-	Sk}', "");
Error('\p{General_Category=:=-Mark}');
Error('\P{General_Category=:=-Mark}');
Expect(1, 917999, '\p{General_Category=mark}', "");
Expect(0, 917999, '\p{^General_Category=mark}', "");
Expect(0, 917999, '\P{General_Category=mark}', "");
Expect(1, 917999, '\P{^General_Category=mark}', "");
Expect(0, 918000, '\p{General_Category=mark}', "");
Expect(1, 918000, '\p{^General_Category=mark}', "");
Expect(1, 918000, '\P{General_Category=mark}', "");
Expect(0, 918000, '\P{^General_Category=mark}', "");
Expect(1, 917999, '\p{General_Category:  _MARK}', "");
Expect(0, 917999, '\p{^General_Category:  _MARK}', "");
Expect(0, 917999, '\P{General_Category:  _MARK}', "");
Expect(1, 917999, '\P{^General_Category:  _MARK}', "");
Expect(0, 918000, '\p{General_Category:  _MARK}', "");
Expect(1, 918000, '\p{^General_Category:  _MARK}', "");
Expect(1, 918000, '\P{General_Category:  _MARK}', "");
Expect(0, 918000, '\P{^General_Category:  _MARK}', "");
Error('\p{Gc=:=	-M}');
Error('\P{Gc=:=	-M}');
Expect(1, 917999, '\p{Gc=m}', "");
Expect(0, 917999, '\p{^Gc=m}', "");
Expect(0, 917999, '\P{Gc=m}', "");
Expect(1, 917999, '\P{^Gc=m}', "");
Expect(0, 918000, '\p{Gc=m}', "");
Expect(1, 918000, '\p{^Gc=m}', "");
Expect(1, 918000, '\P{Gc=m}', "");
Expect(0, 918000, '\P{^Gc=m}', "");
Expect(1, 917999, '\p{Gc=--m}', "");
Expect(0, 917999, '\p{^Gc=--m}', "");
Expect(0, 917999, '\P{Gc=--m}', "");
Expect(1, 917999, '\P{^Gc=--m}', "");
Expect(0, 918000, '\p{Gc=--m}', "");
Expect(1, 918000, '\p{^Gc=--m}', "");
Expect(1, 918000, '\P{Gc=--m}', "");
Expect(0, 918000, '\P{^Gc=--m}', "");
Error('\p{Category=_/a/Combining_mark}');
Error('\P{Category=_/a/Combining_mark}');
Expect(1, 917999, '\p{Category=combiningmark}', "");
Expect(0, 917999, '\p{^Category=combiningmark}', "");
Expect(0, 917999, '\P{Category=combiningmark}', "");
Expect(1, 917999, '\P{^Category=combiningmark}', "");
Expect(0, 918000, '\p{Category=combiningmark}', "");
Expect(1, 918000, '\p{^Category=combiningmark}', "");
Expect(1, 918000, '\P{Category=combiningmark}', "");
Expect(0, 918000, '\P{^Category=combiningmark}', "");
Expect(1, 917999, '\p{Category=- COMBINING_MARK}', "");
Expect(0, 917999, '\p{^Category=- COMBINING_MARK}', "");
Expect(0, 917999, '\P{Category=- COMBINING_MARK}', "");
Expect(1, 917999, '\P{^Category=- COMBINING_MARK}', "");
Expect(0, 918000, '\p{Category=- COMBINING_MARK}', "");
Expect(1, 918000, '\p{^Category=- COMBINING_MARK}', "");
Expect(1, 918000, '\P{Category=- COMBINING_MARK}', "");
Expect(0, 918000, '\P{^Category=- COMBINING_MARK}', "");
Error('\p{Is_General_Category=	mark:=}');
Error('\P{Is_General_Category=	mark:=}');
Expect(1, 917999, '\p{Is_General_Category=mark}', "");
Expect(0, 917999, '\p{^Is_General_Category=mark}', "");
Expect(0, 917999, '\P{Is_General_Category=mark}', "");
Expect(1, 917999, '\P{^Is_General_Category=mark}', "");
Expect(0, 918000, '\p{Is_General_Category=mark}', "");
Expect(1, 918000, '\p{^Is_General_Category=mark}', "");
Expect(1, 918000, '\P{Is_General_Category=mark}', "");
Expect(0, 918000, '\P{^Is_General_Category=mark}', "");
Expect(1, 917999, '\p{Is_General_Category=	mark}', "");
Expect(0, 917999, '\p{^Is_General_Category=	mark}', "");
Expect(0, 917999, '\P{Is_General_Category=	mark}', "");
Expect(1, 917999, '\P{^Is_General_Category=	mark}', "");
Expect(0, 918000, '\p{Is_General_Category=	mark}', "");
Expect(1, 918000, '\p{^Is_General_Category=	mark}', "");
Expect(1, 918000, '\P{Is_General_Category=	mark}', "");
Expect(0, 918000, '\P{^Is_General_Category=	mark}', "");
Error('\p{Is_Gc=	 M/a/}');
Error('\P{Is_Gc=	 M/a/}');
Expect(1, 917999, '\p{Is_Gc=m}', "");
Expect(0, 917999, '\p{^Is_Gc=m}', "");
Expect(0, 917999, '\P{Is_Gc=m}', "");
Expect(1, 917999, '\P{^Is_Gc=m}', "");
Expect(0, 918000, '\p{Is_Gc=m}', "");
Expect(1, 918000, '\p{^Is_Gc=m}', "");
Expect(1, 918000, '\P{Is_Gc=m}', "");
Expect(0, 918000, '\P{^Is_Gc=m}', "");
Expect(1, 917999, '\p{Is_Gc=M}', "");
Expect(0, 917999, '\p{^Is_Gc=M}', "");
Expect(0, 917999, '\P{Is_Gc=M}', "");
Expect(1, 917999, '\P{^Is_Gc=M}', "");
Expect(0, 918000, '\p{Is_Gc=M}', "");
Expect(1, 918000, '\p{^Is_Gc=M}', "");
Expect(1, 918000, '\P{Is_Gc=M}', "");
Expect(0, 918000, '\P{^Is_Gc=M}', "");
Error('\p{Is_Category:    COMBINING_MARK:=}');
Error('\P{Is_Category:    COMBINING_MARK:=}');
Expect(1, 917999, '\p{Is_Category=combiningmark}', "");
Expect(0, 917999, '\p{^Is_Category=combiningmark}', "");
Expect(0, 917999, '\P{Is_Category=combiningmark}', "");
Expect(1, 917999, '\P{^Is_Category=combiningmark}', "");
Expect(0, 918000, '\p{Is_Category=combiningmark}', "");
Expect(1, 918000, '\p{^Is_Category=combiningmark}', "");
Expect(1, 918000, '\P{Is_Category=combiningmark}', "");
Expect(0, 918000, '\P{^Is_Category=combiningmark}', "");
Expect(1, 917999, '\p{Is_Category=__combining_Mark}', "");
Expect(0, 917999, '\p{^Is_Category=__combining_Mark}', "");
Expect(0, 917999, '\P{Is_Category=__combining_Mark}', "");
Expect(1, 917999, '\P{^Is_Category=__combining_Mark}', "");
Expect(0, 918000, '\p{Is_Category=__combining_Mark}', "");
Expect(1, 918000, '\p{^Is_Category=__combining_Mark}', "");
Expect(1, 918000, '\P{Is_Category=__combining_Mark}', "");
Expect(0, 918000, '\P{^Is_Category=__combining_Mark}', "");
Error('\p{General_Category=	/a/control}');
Error('\P{General_Category=	/a/control}');
Expect(1, 159, '\p{General_Category=control}', "");
Expect(0, 159, '\p{^General_Category=control}', "");
Expect(0, 159, '\P{General_Category=control}', "");
Expect(1, 159, '\P{^General_Category=control}', "");
Expect(0, 160, '\p{General_Category=control}', "");
Expect(1, 160, '\p{^General_Category=control}', "");
Expect(1, 160, '\P{General_Category=control}', "");
Expect(0, 160, '\P{^General_Category=control}', "");
Expect(1, 159, '\p{General_Category=-Control}', "");
Expect(0, 159, '\p{^General_Category=-Control}', "");
Expect(0, 159, '\P{General_Category=-Control}', "");
Expect(1, 159, '\P{^General_Category=-Control}', "");
Expect(0, 160, '\p{General_Category=-Control}', "");
Expect(1, 160, '\p{^General_Category=-Control}', "");
Expect(1, 160, '\P{General_Category=-Control}', "");
Expect(0, 160, '\P{^General_Category=-Control}', "");
Error('\p{Gc= 	Cc:=}');
Error('\P{Gc= 	Cc:=}');
Expect(1, 159, '\p{Gc=cc}', "");
Expect(0, 159, '\p{^Gc=cc}', "");
Expect(0, 159, '\P{Gc=cc}', "");
Expect(1, 159, '\P{^Gc=cc}', "");
Expect(0, 160, '\p{Gc=cc}', "");
Expect(1, 160, '\p{^Gc=cc}', "");
Expect(1, 160, '\P{Gc=cc}', "");
Expect(0, 160, '\P{^Gc=cc}', "");
Expect(1, 159, '\p{Gc=- Cc}', "");
Expect(0, 159, '\p{^Gc=- Cc}', "");
Expect(0, 159, '\P{Gc=- Cc}', "");
Expect(1, 159, '\P{^Gc=- Cc}', "");
Expect(0, 160, '\p{Gc=- Cc}', "");
Expect(1, 160, '\p{^Gc=- Cc}', "");
Expect(1, 160, '\P{Gc=- Cc}', "");
Expect(0, 160, '\P{^Gc=- Cc}', "");
Error('\p{Category=_/a/Cntrl}');
Error('\P{Category=_/a/Cntrl}');
Expect(1, 159, '\p{Category=cntrl}', "");
Expect(0, 159, '\p{^Category=cntrl}', "");
Expect(0, 159, '\P{Category=cntrl}', "");
Expect(1, 159, '\P{^Category=cntrl}', "");
Expect(0, 160, '\p{Category=cntrl}', "");
Expect(1, 160, '\p{^Category=cntrl}', "");
Expect(1, 160, '\P{Category=cntrl}', "");
Expect(0, 160, '\P{^Category=cntrl}', "");
Expect(1, 159, '\p{Category=_CNTRL}', "");
Expect(0, 159, '\p{^Category=_CNTRL}', "");
Expect(0, 159, '\P{Category=_CNTRL}', "");
Expect(1, 159, '\P{^Category=_CNTRL}', "");
Expect(0, 160, '\p{Category=_CNTRL}', "");
Expect(1, 160, '\p{^Category=_CNTRL}', "");
Expect(1, 160, '\P{Category=_CNTRL}', "");
Expect(0, 160, '\P{^Category=_CNTRL}', "");
Error('\p{Is_General_Category= /a/Control}');
Error('\P{Is_General_Category= /a/Control}');
Expect(1, 159, '\p{Is_General_Category: control}', "");
Expect(0, 159, '\p{^Is_General_Category: control}', "");
Expect(0, 159, '\P{Is_General_Category: control}', "");
Expect(1, 159, '\P{^Is_General_Category: control}', "");
Expect(0, 160, '\p{Is_General_Category: control}', "");
Expect(1, 160, '\p{^Is_General_Category: control}', "");
Expect(1, 160, '\P{Is_General_Category: control}', "");
Expect(0, 160, '\P{^Is_General_Category: control}', "");
Expect(1, 159, '\p{Is_General_Category:     Control}', "");
Expect(0, 159, '\p{^Is_General_Category:     Control}', "");
Expect(0, 159, '\P{Is_General_Category:     Control}', "");
Expect(1, 159, '\P{^Is_General_Category:     Control}', "");
Expect(0, 160, '\p{Is_General_Category:     Control}', "");
Expect(1, 160, '\p{^Is_General_Category:     Control}', "");
Expect(1, 160, '\P{Is_General_Category:     Control}', "");
Expect(0, 160, '\P{^Is_General_Category:     Control}', "");
Error('\p{Is_Gc= Cc/a/}');
Error('\P{Is_Gc= Cc/a/}');
Expect(1, 159, '\p{Is_Gc=cc}', "");
Expect(0, 159, '\p{^Is_Gc=cc}', "");
Expect(0, 159, '\P{Is_Gc=cc}', "");
Expect(1, 159, '\P{^Is_Gc=cc}', "");
Expect(0, 160, '\p{Is_Gc=cc}', "");
Expect(1, 160, '\p{^Is_Gc=cc}', "");
Expect(1, 160, '\P{Is_Gc=cc}', "");
Expect(0, 160, '\P{^Is_Gc=cc}', "");
Expect(1, 159, '\p{Is_Gc=-CC}', "");
Expect(0, 159, '\p{^Is_Gc=-CC}', "");
Expect(0, 159, '\P{Is_Gc=-CC}', "");
Expect(1, 159, '\P{^Is_Gc=-CC}', "");
Expect(0, 160, '\p{Is_Gc=-CC}', "");
Expect(1, 160, '\p{^Is_Gc=-CC}', "");
Expect(1, 160, '\P{Is_Gc=-CC}', "");
Expect(0, 160, '\P{^Is_Gc=-CC}', "");
Error('\p{Is_Category=/a/ 	Cntrl}');
Error('\P{Is_Category=/a/ 	Cntrl}');
Expect(1, 159, '\p{Is_Category=cntrl}', "");
Expect(0, 159, '\p{^Is_Category=cntrl}', "");
Expect(0, 159, '\P{Is_Category=cntrl}', "");
Expect(1, 159, '\P{^Is_Category=cntrl}', "");
Expect(0, 160, '\p{Is_Category=cntrl}', "");
Expect(1, 160, '\p{^Is_Category=cntrl}', "");
Expect(1, 160, '\P{Is_Category=cntrl}', "");
Expect(0, 160, '\P{^Is_Category=cntrl}', "");
Expect(1, 159, '\p{Is_Category=_Cntrl}', "");
Expect(0, 159, '\p{^Is_Category=_Cntrl}', "");
Expect(0, 159, '\P{Is_Category=_Cntrl}', "");
Expect(1, 159, '\P{^Is_Category=_Cntrl}', "");
Expect(0, 160, '\p{Is_Category=_Cntrl}', "");
Expect(1, 160, '\p{^Is_Category=_Cntrl}', "");
Expect(1, 160, '\P{Is_Category=_Cntrl}', "");
Expect(0, 160, '\P{^Is_Category=_Cntrl}', "");
Error('\p{General_Category=		LINE_Separator:=}');
Error('\P{General_Category=		LINE_Separator:=}');
Expect(1, 8232, '\p{General_Category=lineseparator}', "");
Expect(0, 8232, '\p{^General_Category=lineseparator}', "");
Expect(0, 8232, '\P{General_Category=lineseparator}', "");
Expect(1, 8232, '\P{^General_Category=lineseparator}', "");
Expect(0, 8233, '\p{General_Category=lineseparator}', "");
Expect(1, 8233, '\p{^General_Category=lineseparator}', "");
Expect(1, 8233, '\P{General_Category=lineseparator}', "");
Expect(0, 8233, '\P{^General_Category=lineseparator}', "");
Expect(1, 8232, '\p{General_Category=_	Line_Separator}', "");
Expect(0, 8232, '\p{^General_Category=_	Line_Separator}', "");
Expect(0, 8232, '\P{General_Category=_	Line_Separator}', "");
Expect(1, 8232, '\P{^General_Category=_	Line_Separator}', "");
Expect(0, 8233, '\p{General_Category=_	Line_Separator}', "");
Expect(1, 8233, '\p{^General_Category=_	Line_Separator}', "");
Expect(1, 8233, '\P{General_Category=_	Line_Separator}', "");
Expect(0, 8233, '\P{^General_Category=_	Line_Separator}', "");
Error('\p{Gc=/a/		zl}');
Error('\P{Gc=/a/		zl}');
Expect(1, 8232, '\p{Gc=zl}', "");
Expect(0, 8232, '\p{^Gc=zl}', "");
Expect(0, 8232, '\P{Gc=zl}', "");
Expect(1, 8232, '\P{^Gc=zl}', "");
Expect(0, 8233, '\p{Gc=zl}', "");
Expect(1, 8233, '\p{^Gc=zl}', "");
Expect(1, 8233, '\P{Gc=zl}', "");
Expect(0, 8233, '\P{^Gc=zl}', "");
Expect(1, 8232, '\p{Gc: -ZL}', "");
Expect(0, 8232, '\p{^Gc: -ZL}', "");
Expect(0, 8232, '\P{Gc: -ZL}', "");
Expect(1, 8232, '\P{^Gc: -ZL}', "");
Expect(0, 8233, '\p{Gc: -ZL}', "");
Expect(1, 8233, '\p{^Gc: -ZL}', "");
Expect(1, 8233, '\P{Gc: -ZL}', "");
Expect(0, 8233, '\P{^Gc: -ZL}', "");
Error('\p{Category=/a/__LINE_separator}');
Error('\P{Category=/a/__LINE_separator}');
Expect(1, 8232, '\p{Category=lineseparator}', "");
Expect(0, 8232, '\p{^Category=lineseparator}', "");
Expect(0, 8232, '\P{Category=lineseparator}', "");
Expect(1, 8232, '\P{^Category=lineseparator}', "");
Expect(0, 8233, '\p{Category=lineseparator}', "");
Expect(1, 8233, '\p{^Category=lineseparator}', "");
Expect(1, 8233, '\P{Category=lineseparator}', "");
Expect(0, 8233, '\P{^Category=lineseparator}', "");
Expect(1, 8232, '\p{Category=- Line_Separator}', "");
Expect(0, 8232, '\p{^Category=- Line_Separator}', "");
Expect(0, 8232, '\P{Category=- Line_Separator}', "");
Expect(1, 8232, '\P{^Category=- Line_Separator}', "");
Expect(0, 8233, '\p{Category=- Line_Separator}', "");
Expect(1, 8233, '\p{^Category=- Line_Separator}', "");
Expect(1, 8233, '\P{Category=- Line_Separator}', "");
Expect(0, 8233, '\P{^Category=- Line_Separator}', "");
Error('\p{Is_General_Category=:=_-Zl}');
Error('\P{Is_General_Category=:=_-Zl}');
Expect(1, 8232, '\p{Is_General_Category=zl}', "");
Expect(0, 8232, '\p{^Is_General_Category=zl}', "");
Expect(0, 8232, '\P{Is_General_Category=zl}', "");
Expect(1, 8232, '\P{^Is_General_Category=zl}', "");
Expect(0, 8233, '\p{Is_General_Category=zl}', "");
Expect(1, 8233, '\p{^Is_General_Category=zl}', "");
Expect(1, 8233, '\P{Is_General_Category=zl}', "");
Expect(0, 8233, '\P{^Is_General_Category=zl}', "");
Expect(1, 8232, '\p{Is_General_Category=Zl}', "");
Expect(0, 8232, '\p{^Is_General_Category=Zl}', "");
Expect(0, 8232, '\P{Is_General_Category=Zl}', "");
Expect(1, 8232, '\P{^Is_General_Category=Zl}', "");
Expect(0, 8233, '\p{Is_General_Category=Zl}', "");
Expect(1, 8233, '\p{^Is_General_Category=Zl}', "");
Expect(1, 8233, '\P{Is_General_Category=Zl}', "");
Expect(0, 8233, '\P{^Is_General_Category=Zl}', "");
Error('\p{Is_Gc=/a/-line_Separator}');
Error('\P{Is_Gc=/a/-line_Separator}');
Expect(1, 8232, '\p{Is_Gc=lineseparator}', "");
Expect(0, 8232, '\p{^Is_Gc=lineseparator}', "");
Expect(0, 8232, '\P{Is_Gc=lineseparator}', "");
Expect(1, 8232, '\P{^Is_Gc=lineseparator}', "");
Expect(0, 8233, '\p{Is_Gc=lineseparator}', "");
Expect(1, 8233, '\p{^Is_Gc=lineseparator}', "");
Expect(1, 8233, '\P{Is_Gc=lineseparator}', "");
Expect(0, 8233, '\P{^Is_Gc=lineseparator}', "");
Expect(1, 8232, '\p{Is_Gc:	-Line_SEPARATOR}', "");
Expect(0, 8232, '\p{^Is_Gc:	-Line_SEPARATOR}', "");
Expect(0, 8232, '\P{Is_Gc:	-Line_SEPARATOR}', "");
Expect(1, 8232, '\P{^Is_Gc:	-Line_SEPARATOR}', "");
Expect(0, 8233, '\p{Is_Gc:	-Line_SEPARATOR}', "");
Expect(1, 8233, '\p{^Is_Gc:	-Line_SEPARATOR}', "");
Expect(1, 8233, '\P{Is_Gc:	-Line_SEPARATOR}', "");
Expect(0, 8233, '\P{^Is_Gc:	-Line_SEPARATOR}', "");
Error('\p{Is_Category=:=_zl}');
Error('\P{Is_Category=:=_zl}');
Expect(1, 8232, '\p{Is_Category: zl}', "");
Expect(0, 8232, '\p{^Is_Category: zl}', "");
Expect(0, 8232, '\P{Is_Category: zl}', "");
Expect(1, 8232, '\P{^Is_Category: zl}', "");
Expect(0, 8233, '\p{Is_Category: zl}', "");
Expect(1, 8233, '\p{^Is_Category: zl}', "");
Expect(1, 8233, '\P{Is_Category: zl}', "");
Expect(0, 8233, '\P{^Is_Category: zl}', "");
Expect(1, 8232, '\p{Is_Category=-Zl}', "");
Expect(0, 8232, '\p{^Is_Category=-Zl}', "");
Expect(0, 8232, '\P{Is_Category=-Zl}', "");
Expect(1, 8232, '\P{^Is_Category=-Zl}', "");
Expect(0, 8233, '\p{Is_Category=-Zl}', "");
Expect(1, 8233, '\p{^Is_Category=-Zl}', "");
Expect(1, 8233, '\P{Is_Category=-Zl}', "");
Expect(0, 8233, '\P{^Is_Category=-Zl}', "");
Error('\p{General_Category=:=- Open_Punctuation}');
Error('\P{General_Category=:=- Open_Punctuation}');
Expect(1, 65378, '\p{General_Category=openpunctuation}', "");
Expect(0, 65378, '\p{^General_Category=openpunctuation}', "");
Expect(0, 65378, '\P{General_Category=openpunctuation}', "");
Expect(1, 65378, '\P{^General_Category=openpunctuation}', "");
Expect(0, 65379, '\p{General_Category=openpunctuation}', "");
Expect(1, 65379, '\p{^General_Category=openpunctuation}', "");
Expect(1, 65379, '\P{General_Category=openpunctuation}', "");
Expect(0, 65379, '\P{^General_Category=openpunctuation}', "");
Expect(1, 65378, '\p{General_Category:  -Open_Punctuation}', "");
Expect(0, 65378, '\p{^General_Category:  -Open_Punctuation}', "");
Expect(0, 65378, '\P{General_Category:  -Open_Punctuation}', "");
Expect(1, 65378, '\P{^General_Category:  -Open_Punctuation}', "");
Expect(0, 65379, '\p{General_Category:  -Open_Punctuation}', "");
Expect(1, 65379, '\p{^General_Category:  -Open_Punctuation}', "");
Expect(1, 65379, '\P{General_Category:  -Open_Punctuation}', "");
Expect(0, 65379, '\P{^General_Category:  -Open_Punctuation}', "");
Error('\p{Gc=/a/PS}');
Error('\P{Gc=/a/PS}');
Expect(1, 65378, '\p{Gc=ps}', "");
Expect(0, 65378, '\p{^Gc=ps}', "");
Expect(0, 65378, '\P{Gc=ps}', "");
Expect(1, 65378, '\P{^Gc=ps}', "");
Expect(0, 65379, '\p{Gc=ps}', "");
Expect(1, 65379, '\p{^Gc=ps}', "");
Expect(1, 65379, '\P{Gc=ps}', "");
Expect(0, 65379, '\P{^Gc=ps}', "");
Expect(1, 65378, '\p{Gc=--Ps}', "");
Expect(0, 65378, '\p{^Gc=--Ps}', "");
Expect(0, 65378, '\P{Gc=--Ps}', "");
Expect(1, 65378, '\P{^Gc=--Ps}', "");
Expect(0, 65379, '\p{Gc=--Ps}', "");
Expect(1, 65379, '\p{^Gc=--Ps}', "");
Expect(1, 65379, '\P{Gc=--Ps}', "");
Expect(0, 65379, '\P{^Gc=--Ps}', "");
Error('\p{Category= _open_Punctuation:=}');
Error('\P{Category= _open_Punctuation:=}');
Expect(1, 65378, '\p{Category=openpunctuation}', "");
Expect(0, 65378, '\p{^Category=openpunctuation}', "");
Expect(0, 65378, '\P{Category=openpunctuation}', "");
Expect(1, 65378, '\P{^Category=openpunctuation}', "");
Expect(0, 65379, '\p{Category=openpunctuation}', "");
Expect(1, 65379, '\p{^Category=openpunctuation}', "");
Expect(1, 65379, '\P{Category=openpunctuation}', "");
Expect(0, 65379, '\P{^Category=openpunctuation}', "");
Expect(1, 65378, '\p{Category= 	Open_Punctuation}', "");
Expect(0, 65378, '\p{^Category= 	Open_Punctuation}', "");
Expect(0, 65378, '\P{Category= 	Open_Punctuation}', "");
Expect(1, 65378, '\P{^Category= 	Open_Punctuation}', "");
Expect(0, 65379, '\p{Category= 	Open_Punctuation}', "");
Expect(1, 65379, '\p{^Category= 	Open_Punctuation}', "");
Expect(1, 65379, '\P{Category= 	Open_Punctuation}', "");
Expect(0, 65379, '\P{^Category= 	Open_Punctuation}', "");
Error('\p{Is_General_Category=	ps:=}');
Error('\P{Is_General_Category=	ps:=}');
Expect(1, 65378, '\p{Is_General_Category=ps}', "");
Expect(0, 65378, '\p{^Is_General_Category=ps}', "");
Expect(0, 65378, '\P{Is_General_Category=ps}', "");
Expect(1, 65378, '\P{^Is_General_Category=ps}', "");
Expect(0, 65379, '\p{Is_General_Category=ps}', "");
Expect(1, 65379, '\p{^Is_General_Category=ps}', "");
Expect(1, 65379, '\P{Is_General_Category=ps}', "");
Expect(0, 65379, '\P{^Is_General_Category=ps}', "");
Expect(1, 65378, '\p{Is_General_Category= _PS}', "");
Expect(0, 65378, '\p{^Is_General_Category= _PS}', "");
Expect(0, 65378, '\P{Is_General_Category= _PS}', "");
Expect(1, 65378, '\P{^Is_General_Category= _PS}', "");
Expect(0, 65379, '\p{Is_General_Category= _PS}', "");
Expect(1, 65379, '\p{^Is_General_Category= _PS}', "");
Expect(1, 65379, '\P{Is_General_Category= _PS}', "");
Expect(0, 65379, '\P{^Is_General_Category= _PS}', "");
Error('\p{Is_Gc:		:=Open_PUNCTUATION}');
Error('\P{Is_Gc:		:=Open_PUNCTUATION}');
Expect(1, 65378, '\p{Is_Gc=openpunctuation}', "");
Expect(0, 65378, '\p{^Is_Gc=openpunctuation}', "");
Expect(0, 65378, '\P{Is_Gc=openpunctuation}', "");
Expect(1, 65378, '\P{^Is_Gc=openpunctuation}', "");
Expect(0, 65379, '\p{Is_Gc=openpunctuation}', "");
Expect(1, 65379, '\p{^Is_Gc=openpunctuation}', "");
Expect(1, 65379, '\P{Is_Gc=openpunctuation}', "");
Expect(0, 65379, '\P{^Is_Gc=openpunctuation}', "");
Expect(1, 65378, '\p{Is_Gc: -Open_PUNCTUATION}', "");
Expect(0, 65378, '\p{^Is_Gc: -Open_PUNCTUATION}', "");
Expect(0, 65378, '\P{Is_Gc: -Open_PUNCTUATION}', "");
Expect(1, 65378, '\P{^Is_Gc: -Open_PUNCTUATION}', "");
Expect(0, 65379, '\p{Is_Gc: -Open_PUNCTUATION}', "");
Expect(1, 65379, '\p{^Is_Gc: -Open_PUNCTUATION}', "");
Expect(1, 65379, '\P{Is_Gc: -Open_PUNCTUATION}', "");
Expect(0, 65379, '\P{^Is_Gc: -Open_PUNCTUATION}', "");
Error('\p{Is_Category= :=PS}');
Error('\P{Is_Category= :=PS}');
Expect(1, 65378, '\p{Is_Category=ps}', "");
Expect(0, 65378, '\p{^Is_Category=ps}', "");
Expect(0, 65378, '\P{Is_Category=ps}', "");
Expect(1, 65378, '\P{^Is_Category=ps}', "");
Expect(0, 65379, '\p{Is_Category=ps}', "");
Expect(1, 65379, '\p{^Is_Category=ps}', "");
Expect(1, 65379, '\P{Is_Category=ps}', "");
Expect(0, 65379, '\P{^Is_Category=ps}', "");
Expect(1, 65378, '\p{Is_Category=	 ps}', "");
Expect(0, 65378, '\p{^Is_Category=	 ps}', "");
Expect(0, 65378, '\P{Is_Category=	 ps}', "");
Expect(1, 65378, '\P{^Is_Category=	 ps}', "");
Expect(0, 65379, '\p{Is_Category=	 ps}', "");
Expect(1, 65379, '\p{^Is_Category=	 ps}', "");
Expect(1, 65379, '\P{Is_Category=	 ps}', "");
Expect(0, 65379, '\P{^Is_Category=	 ps}', "");
Error('\p{General_Category=_-MATH_SYMBOL/a/}');
Error('\P{General_Category=_-MATH_SYMBOL/a/}');
Expect(1, 126705, '\p{General_Category=mathsymbol}', "");
Expect(0, 126705, '\p{^General_Category=mathsymbol}', "");
Expect(0, 126705, '\P{General_Category=mathsymbol}', "");
Expect(1, 126705, '\P{^General_Category=mathsymbol}', "");
Expect(0, 126706, '\p{General_Category=mathsymbol}', "");
Expect(1, 126706, '\p{^General_Category=mathsymbol}', "");
Expect(1, 126706, '\P{General_Category=mathsymbol}', "");
Expect(0, 126706, '\P{^General_Category=mathsymbol}', "");
Expect(1, 126705, '\p{General_Category= _Math_Symbol}', "");
Expect(0, 126705, '\p{^General_Category= _Math_Symbol}', "");
Expect(0, 126705, '\P{General_Category= _Math_Symbol}', "");
Expect(1, 126705, '\P{^General_Category= _Math_Symbol}', "");
Expect(0, 126706, '\p{General_Category= _Math_Symbol}', "");
Expect(1, 126706, '\p{^General_Category= _Math_Symbol}', "");
Expect(1, 126706, '\P{General_Category= _Math_Symbol}', "");
Expect(0, 126706, '\P{^General_Category= _Math_Symbol}', "");
Error('\p{Gc=_	sm:=}');
Error('\P{Gc=_	sm:=}');
Expect(1, 126705, '\p{Gc=sm}', "");
Expect(0, 126705, '\p{^Gc=sm}', "");
Expect(0, 126705, '\P{Gc=sm}', "");
Expect(1, 126705, '\P{^Gc=sm}', "");
Expect(0, 126706, '\p{Gc=sm}', "");
Expect(1, 126706, '\p{^Gc=sm}', "");
Expect(1, 126706, '\P{Gc=sm}', "");
Expect(0, 126706, '\P{^Gc=sm}', "");
Expect(1, 126705, '\p{Gc=	-SM}', "");
Expect(0, 126705, '\p{^Gc=	-SM}', "");
Expect(0, 126705, '\P{Gc=	-SM}', "");
Expect(1, 126705, '\P{^Gc=	-SM}', "");
Expect(0, 126706, '\p{Gc=	-SM}', "");
Expect(1, 126706, '\p{^Gc=	-SM}', "");
Expect(1, 126706, '\P{Gc=	-SM}', "");
Expect(0, 126706, '\P{^Gc=	-SM}', "");
Error('\p{Category= Math_Symbol:=}');
Error('\P{Category= Math_Symbol:=}');
Expect(1, 126705, '\p{Category=mathsymbol}', "");
Expect(0, 126705, '\p{^Category=mathsymbol}', "");
Expect(0, 126705, '\P{Category=mathsymbol}', "");
Expect(1, 126705, '\P{^Category=mathsymbol}', "");
Expect(0, 126706, '\p{Category=mathsymbol}', "");
Expect(1, 126706, '\p{^Category=mathsymbol}', "");
Expect(1, 126706, '\P{Category=mathsymbol}', "");
Expect(0, 126706, '\P{^Category=mathsymbol}', "");
Expect(1, 126705, '\p{Category:	_Math_symbol}', "");
Expect(0, 126705, '\p{^Category:	_Math_symbol}', "");
Expect(0, 126705, '\P{Category:	_Math_symbol}', "");
Expect(1, 126705, '\P{^Category:	_Math_symbol}', "");
Expect(0, 126706, '\p{Category:	_Math_symbol}', "");
Expect(1, 126706, '\p{^Category:	_Math_symbol}', "");
Expect(1, 126706, '\P{Category:	_Math_symbol}', "");
Expect(0, 126706, '\P{^Category:	_Math_symbol}', "");
Error('\p{Is_General_Category=/a/--SM}');
Error('\P{Is_General_Category=/a/--SM}');
Expect(1, 126705, '\p{Is_General_Category=sm}', "");
Expect(0, 126705, '\p{^Is_General_Category=sm}', "");
Expect(0, 126705, '\P{Is_General_Category=sm}', "");
Expect(1, 126705, '\P{^Is_General_Category=sm}', "");
Expect(0, 126706, '\p{Is_General_Category=sm}', "");
Expect(1, 126706, '\p{^Is_General_Category=sm}', "");
Expect(1, 126706, '\P{Is_General_Category=sm}', "");
Expect(0, 126706, '\P{^Is_General_Category=sm}', "");
Expect(1, 126705, '\p{Is_General_Category=	sm}', "");
Expect(0, 126705, '\p{^Is_General_Category=	sm}', "");
Expect(0, 126705, '\P{Is_General_Category=	sm}', "");
Expect(1, 126705, '\P{^Is_General_Category=	sm}', "");
Expect(0, 126706, '\p{Is_General_Category=	sm}', "");
Expect(1, 126706, '\p{^Is_General_Category=	sm}', "");
Expect(1, 126706, '\P{Is_General_Category=	sm}', "");
Expect(0, 126706, '\P{^Is_General_Category=	sm}', "");
Error('\p{Is_Gc=--Math_SYMBOL:=}');
Error('\P{Is_Gc=--Math_SYMBOL:=}');
Expect(1, 126705, '\p{Is_Gc=mathsymbol}', "");
Expect(0, 126705, '\p{^Is_Gc=mathsymbol}', "");
Expect(0, 126705, '\P{Is_Gc=mathsymbol}', "");
Expect(1, 126705, '\P{^Is_Gc=mathsymbol}', "");
Expect(0, 126706, '\p{Is_Gc=mathsymbol}', "");
Expect(1, 126706, '\p{^Is_Gc=mathsymbol}', "");
Expect(1, 126706, '\P{Is_Gc=mathsymbol}', "");
Expect(0, 126706, '\P{^Is_Gc=mathsymbol}', "");
Expect(1, 126705, '\p{Is_Gc:	-	Math_Symbol}', "");
Expect(0, 126705, '\p{^Is_Gc:	-	Math_Symbol}', "");
Expect(0, 126705, '\P{Is_Gc:	-	Math_Symbol}', "");
Expect(1, 126705, '\P{^Is_Gc:	-	Math_Symbol}', "");
Expect(0, 126706, '\p{Is_Gc:	-	Math_Symbol}', "");
Expect(1, 126706, '\p{^Is_Gc:	-	Math_Symbol}', "");
Expect(1, 126706, '\P{Is_Gc:	-	Math_Symbol}', "");
Expect(0, 126706, '\P{^Is_Gc:	-	Math_Symbol}', "");
Error('\p{Is_Category=--Sm:=}');
Error('\P{Is_Category=--Sm:=}');
Expect(1, 126705, '\p{Is_Category=sm}', "");
Expect(0, 126705, '\p{^Is_Category=sm}', "");
Expect(0, 126705, '\P{Is_Category=sm}', "");
Expect(1, 126705, '\P{^Is_Category=sm}', "");
Expect(0, 126706, '\p{Is_Category=sm}', "");
Expect(1, 126706, '\p{^Is_Category=sm}', "");
Expect(1, 126706, '\P{Is_Category=sm}', "");
Expect(0, 126706, '\P{^Is_Category=sm}', "");
Expect(1, 126705, '\p{Is_Category:   - sm}', "");
Expect(0, 126705, '\p{^Is_Category:   - sm}', "");
Expect(0, 126705, '\P{Is_Category:   - sm}', "");
Expect(1, 126705, '\P{^Is_Category:   - sm}', "");
Expect(0, 126706, '\p{Is_Category:   - sm}', "");
Expect(1, 126706, '\p{^Is_Category:   - sm}', "");
Expect(1, 126706, '\P{Is_Category:   - sm}', "");
Expect(0, 126706, '\P{^Is_Category:   - sm}', "");
Error('\p{General_Category=/a/ 	Other_letter}');
Error('\P{General_Category=/a/ 	Other_letter}');
Expect(1, 195101, '\p{General_Category=otherletter}', "");
Expect(0, 195101, '\p{^General_Category=otherletter}', "");
Expect(0, 195101, '\P{General_Category=otherletter}', "");
Expect(1, 195101, '\P{^General_Category=otherletter}', "");
Expect(0, 195102, '\p{General_Category=otherletter}', "");
Expect(1, 195102, '\p{^General_Category=otherletter}', "");
Expect(1, 195102, '\P{General_Category=otherletter}', "");
Expect(0, 195102, '\P{^General_Category=otherletter}', "");
Expect(1, 195101, '\p{General_Category=- other_Letter}', "");
Expect(0, 195101, '\p{^General_Category=- other_Letter}', "");
Expect(0, 195101, '\P{General_Category=- other_Letter}', "");
Expect(1, 195101, '\P{^General_Category=- other_Letter}', "");
Expect(0, 195102, '\p{General_Category=- other_Letter}', "");
Expect(1, 195102, '\p{^General_Category=- other_Letter}', "");
Expect(1, 195102, '\P{General_Category=- other_Letter}', "");
Expect(0, 195102, '\P{^General_Category=- other_Letter}', "");
Error('\p{Gc=/a/_Lo}');
Error('\P{Gc=/a/_Lo}');
Expect(1, 195101, '\p{Gc=lo}', "");
Expect(0, 195101, '\p{^Gc=lo}', "");
Expect(0, 195101, '\P{Gc=lo}', "");
Expect(1, 195101, '\P{^Gc=lo}', "");
Expect(0, 195102, '\p{Gc=lo}', "");
Expect(1, 195102, '\p{^Gc=lo}', "");
Expect(1, 195102, '\P{Gc=lo}', "");
Expect(0, 195102, '\P{^Gc=lo}', "");
Expect(1, 195101, '\p{Gc: -LO}', "");
Expect(0, 195101, '\p{^Gc: -LO}', "");
Expect(0, 195101, '\P{Gc: -LO}', "");
Expect(1, 195101, '\P{^Gc: -LO}', "");
Expect(0, 195102, '\p{Gc: -LO}', "");
Expect(1, 195102, '\p{^Gc: -LO}', "");
Expect(1, 195102, '\P{Gc: -LO}', "");
Expect(0, 195102, '\P{^Gc: -LO}', "");
Error('\p{Category=-OTHER_LETTER:=}');
Error('\P{Category=-OTHER_LETTER:=}');
Expect(1, 195101, '\p{Category=otherletter}', "");
Expect(0, 195101, '\p{^Category=otherletter}', "");
Expect(0, 195101, '\P{Category=otherletter}', "");
Expect(1, 195101, '\P{^Category=otherletter}', "");
Expect(0, 195102, '\p{Category=otherletter}', "");
Expect(1, 195102, '\p{^Category=otherletter}', "");
Expect(1, 195102, '\P{Category=otherletter}', "");
Expect(0, 195102, '\P{^Category=otherletter}', "");
Expect(1, 195101, '\p{Category= _other_LETTER}', "");
Expect(0, 195101, '\p{^Category= _other_LETTER}', "");
Expect(0, 195101, '\P{Category= _other_LETTER}', "");
Expect(1, 195101, '\P{^Category= _other_LETTER}', "");
Expect(0, 195102, '\p{Category= _other_LETTER}', "");
Expect(1, 195102, '\p{^Category= _other_LETTER}', "");
Expect(1, 195102, '\P{Category= _other_LETTER}', "");
Expect(0, 195102, '\P{^Category= _other_LETTER}', "");
Error('\p{Is_General_Category=:=	_lo}');
Error('\P{Is_General_Category=:=	_lo}');
Expect(1, 195101, '\p{Is_General_Category=lo}', "");
Expect(0, 195101, '\p{^Is_General_Category=lo}', "");
Expect(0, 195101, '\P{Is_General_Category=lo}', "");
Expect(1, 195101, '\P{^Is_General_Category=lo}', "");
Expect(0, 195102, '\p{Is_General_Category=lo}', "");
Expect(1, 195102, '\p{^Is_General_Category=lo}', "");
Expect(1, 195102, '\P{Is_General_Category=lo}', "");
Expect(0, 195102, '\P{^Is_General_Category=lo}', "");
Expect(1, 195101, '\p{Is_General_Category= _lo}', "");
Expect(0, 195101, '\p{^Is_General_Category= _lo}', "");
Expect(0, 195101, '\P{Is_General_Category= _lo}', "");
Expect(1, 195101, '\P{^Is_General_Category= _lo}', "");
Expect(0, 195102, '\p{Is_General_Category= _lo}', "");
Expect(1, 195102, '\p{^Is_General_Category= _lo}', "");
Expect(1, 195102, '\P{Is_General_Category= _lo}', "");
Expect(0, 195102, '\P{^Is_General_Category= _lo}', "");
Error('\p{Is_Gc= :=Other_Letter}');
Error('\P{Is_Gc= :=Other_Letter}');
Expect(1, 195101, '\p{Is_Gc=otherletter}', "");
Expect(0, 195101, '\p{^Is_Gc=otherletter}', "");
Expect(0, 195101, '\P{Is_Gc=otherletter}', "");
Expect(1, 195101, '\P{^Is_Gc=otherletter}', "");
Expect(0, 195102, '\p{Is_Gc=otherletter}', "");
Expect(1, 195102, '\p{^Is_Gc=otherletter}', "");
Expect(1, 195102, '\P{Is_Gc=otherletter}', "");
Expect(0, 195102, '\P{^Is_Gc=otherletter}', "");
Expect(1, 195101, '\p{Is_Gc=	-other_Letter}', "");
Expect(0, 195101, '\p{^Is_Gc=	-other_Letter}', "");
Expect(0, 195101, '\P{Is_Gc=	-other_Letter}', "");
Expect(1, 195101, '\P{^Is_Gc=	-other_Letter}', "");
Expect(0, 195102, '\p{Is_Gc=	-other_Letter}', "");
Expect(1, 195102, '\p{^Is_Gc=	-other_Letter}', "");
Expect(1, 195102, '\P{Is_Gc=	-other_Letter}', "");
Expect(0, 195102, '\P{^Is_Gc=	-other_Letter}', "");
Error('\p{Is_Category:   _	lo:=}');
Error('\P{Is_Category:   _	lo:=}');
Expect(1, 195101, '\p{Is_Category=lo}', "");
Expect(0, 195101, '\p{^Is_Category=lo}', "");
Expect(0, 195101, '\P{Is_Category=lo}', "");
Expect(1, 195101, '\P{^Is_Category=lo}', "");
Expect(0, 195102, '\p{Is_Category=lo}', "");
Expect(1, 195102, '\p{^Is_Category=lo}', "");
Expect(1, 195102, '\P{Is_Category=lo}', "");
Expect(0, 195102, '\P{^Is_Category=lo}', "");
Expect(1, 195101, '\p{Is_Category=-Lo}', "");
Expect(0, 195101, '\p{^Is_Category=-Lo}', "");
Expect(0, 195101, '\P{Is_Category=-Lo}', "");
Expect(1, 195101, '\P{^Is_Category=-Lo}', "");
Expect(0, 195102, '\p{Is_Category=-Lo}', "");
Expect(1, 195102, '\p{^Is_Category=-Lo}', "");
Expect(1, 195102, '\P{Is_Category=-Lo}', "");
Expect(0, 195102, '\P{^Is_Category=-Lo}', "");
Error('\p{General_Category=	:=SEPARATOR}');
Error('\P{General_Category=	:=SEPARATOR}');
Expect(1, 12288, '\p{General_Category=separator}', "");
Expect(0, 12288, '\p{^General_Category=separator}', "");
Expect(0, 12288, '\P{General_Category=separator}', "");
Expect(1, 12288, '\P{^General_Category=separator}', "");
Expect(0, 12289, '\p{General_Category=separator}', "");
Expect(1, 12289, '\p{^General_Category=separator}', "");
Expect(1, 12289, '\P{General_Category=separator}', "");
Expect(0, 12289, '\P{^General_Category=separator}', "");
Expect(1, 12288, '\p{General_Category=	-separator}', "");
Expect(0, 12288, '\p{^General_Category=	-separator}', "");
Expect(0, 12288, '\P{General_Category=	-separator}', "");
Expect(1, 12288, '\P{^General_Category=	-separator}', "");
Expect(0, 12289, '\p{General_Category=	-separator}', "");
Expect(1, 12289, '\p{^General_Category=	-separator}', "");
Expect(1, 12289, '\P{General_Category=	-separator}', "");
Expect(0, 12289, '\P{^General_Category=	-separator}', "");
Error('\p{Gc=/a/_ Z}');
Error('\P{Gc=/a/_ Z}');
Expect(1, 12288, '\p{Gc=z}', "");
Expect(0, 12288, '\p{^Gc=z}', "");
Expect(0, 12288, '\P{Gc=z}', "");
Expect(1, 12288, '\P{^Gc=z}', "");
Expect(0, 12289, '\p{Gc=z}', "");
Expect(1, 12289, '\p{^Gc=z}', "");
Expect(1, 12289, '\P{Gc=z}', "");
Expect(0, 12289, '\P{^Gc=z}', "");
Expect(1, 12288, '\p{Gc=-z}', "");
Expect(0, 12288, '\p{^Gc=-z}', "");
Expect(0, 12288, '\P{Gc=-z}', "");
Expect(1, 12288, '\P{^Gc=-z}', "");
Expect(0, 12289, '\p{Gc=-z}', "");
Expect(1, 12289, '\p{^Gc=-z}', "");
Expect(1, 12289, '\P{Gc=-z}', "");
Expect(0, 12289, '\P{^Gc=-z}', "");
Error('\p{Category:   separator:=}');
Error('\P{Category:   separator:=}');
Expect(1, 12288, '\p{Category: separator}', "");
Expect(0, 12288, '\p{^Category: separator}', "");
Expect(0, 12288, '\P{Category: separator}', "");
Expect(1, 12288, '\P{^Category: separator}', "");
Expect(0, 12289, '\p{Category: separator}', "");
Expect(1, 12289, '\p{^Category: separator}', "");
Expect(1, 12289, '\P{Category: separator}', "");
Expect(0, 12289, '\P{^Category: separator}', "");
Expect(1, 12288, '\p{Category=_ Separator}', "");
Expect(0, 12288, '\p{^Category=_ Separator}', "");
Expect(0, 12288, '\P{Category=_ Separator}', "");
Expect(1, 12288, '\P{^Category=_ Separator}', "");
Expect(0, 12289, '\p{Category=_ Separator}', "");
Expect(1, 12289, '\p{^Category=_ Separator}', "");
Expect(1, 12289, '\P{Category=_ Separator}', "");
Expect(0, 12289, '\P{^Category=_ Separator}', "");
Error('\p{Is_General_Category: --z:=}');
Error('\P{Is_General_Category: --z:=}');
Expect(1, 12288, '\p{Is_General_Category:	z}', "");
Expect(0, 12288, '\p{^Is_General_Category:	z}', "");
Expect(0, 12288, '\P{Is_General_Category:	z}', "");
Expect(1, 12288, '\P{^Is_General_Category:	z}', "");
Expect(0, 12289, '\p{Is_General_Category:	z}', "");
Expect(1, 12289, '\p{^Is_General_Category:	z}', "");
Expect(1, 12289, '\P{Is_General_Category:	z}', "");
Expect(0, 12289, '\P{^Is_General_Category:	z}', "");
Expect(1, 12288, '\p{Is_General_Category=-Z}', "");
Expect(0, 12288, '\p{^Is_General_Category=-Z}', "");
Expect(0, 12288, '\P{Is_General_Category=-Z}', "");
Expect(1, 12288, '\P{^Is_General_Category=-Z}', "");
Expect(0, 12289, '\p{Is_General_Category=-Z}', "");
Expect(1, 12289, '\p{^Is_General_Category=-Z}', "");
Expect(1, 12289, '\P{Is_General_Category=-Z}', "");
Expect(0, 12289, '\P{^Is_General_Category=-Z}', "");
Error('\p{Is_Gc=:=_	Separator}');
Error('\P{Is_Gc=:=_	Separator}');
Expect(1, 12288, '\p{Is_Gc=separator}', "");
Expect(0, 12288, '\p{^Is_Gc=separator}', "");
Expect(0, 12288, '\P{Is_Gc=separator}', "");
Expect(1, 12288, '\P{^Is_Gc=separator}', "");
Expect(0, 12289, '\p{Is_Gc=separator}', "");
Expect(1, 12289, '\p{^Is_Gc=separator}', "");
Expect(1, 12289, '\P{Is_Gc=separator}', "");
Expect(0, 12289, '\P{^Is_Gc=separator}', "");
Expect(1, 12288, '\p{Is_Gc:_-SEPARATOR}', "");
Expect(0, 12288, '\p{^Is_Gc:_-SEPARATOR}', "");
Expect(0, 12288, '\P{Is_Gc:_-SEPARATOR}', "");
Expect(1, 12288, '\P{^Is_Gc:_-SEPARATOR}', "");
Expect(0, 12289, '\p{Is_Gc:_-SEPARATOR}', "");
Expect(1, 12289, '\p{^Is_Gc:_-SEPARATOR}', "");
Expect(1, 12289, '\P{Is_Gc:_-SEPARATOR}', "");
Expect(0, 12289, '\P{^Is_Gc:_-SEPARATOR}', "");
Error('\p{Is_Category=_	Z/a/}');
Error('\P{Is_Category=_	Z/a/}');
Expect(1, 12288, '\p{Is_Category=z}', "");
Expect(0, 12288, '\p{^Is_Category=z}', "");
Expect(0, 12288, '\P{Is_Category=z}', "");
Expect(1, 12288, '\P{^Is_Category=z}', "");
Expect(0, 12289, '\p{Is_Category=z}', "");
Expect(1, 12289, '\p{^Is_Category=z}', "");
Expect(1, 12289, '\P{Is_Category=z}', "");
Expect(0, 12289, '\P{^Is_Category=z}', "");
Expect(1, 12288, '\p{Is_Category:	__Z}', "");
Expect(0, 12288, '\p{^Is_Category:	__Z}', "");
Expect(0, 12288, '\P{Is_Category:	__Z}', "");
Expect(1, 12288, '\P{^Is_Category:	__Z}', "");
Expect(0, 12289, '\p{Is_Category:	__Z}', "");
Expect(1, 12289, '\p{^Is_Category:	__Z}', "");
Expect(1, 12289, '\P{Is_Category:	__Z}', "");
Expect(0, 12289, '\P{^Is_Category:	__Z}', "");
Error('\p{General_Category=_/a/other_PUNCTUATION}');
Error('\P{General_Category=_/a/other_PUNCTUATION}');
Expect(1, 74867, '\p{General_Category=otherpunctuation}', "");
Expect(0, 74867, '\p{^General_Category=otherpunctuation}', "");
Expect(0, 74867, '\P{General_Category=otherpunctuation}', "");
Expect(1, 74867, '\P{^General_Category=otherpunctuation}', "");
Expect(0, 74868, '\p{General_Category=otherpunctuation}', "");
Expect(1, 74868, '\p{^General_Category=otherpunctuation}', "");
Expect(1, 74868, '\P{General_Category=otherpunctuation}', "");
Expect(0, 74868, '\P{^General_Category=otherpunctuation}', "");
Expect(1, 74867, '\p{General_Category=		Other_PUNCTUATION}', "");
Expect(0, 74867, '\p{^General_Category=		Other_PUNCTUATION}', "");
Expect(0, 74867, '\P{General_Category=		Other_PUNCTUATION}', "");
Expect(1, 74867, '\P{^General_Category=		Other_PUNCTUATION}', "");
Expect(0, 74868, '\p{General_Category=		Other_PUNCTUATION}', "");
Expect(1, 74868, '\p{^General_Category=		Other_PUNCTUATION}', "");
Expect(1, 74868, '\P{General_Category=		Other_PUNCTUATION}', "");
Expect(0, 74868, '\P{^General_Category=		Other_PUNCTUATION}', "");
Error('\p{Gc=-:=PO}');
Error('\P{Gc=-:=PO}');
Expect(1, 74867, '\p{Gc=po}', "");
Expect(0, 74867, '\p{^Gc=po}', "");
Expect(0, 74867, '\P{Gc=po}', "");
Expect(1, 74867, '\P{^Gc=po}', "");
Expect(0, 74868, '\p{Gc=po}', "");
Expect(1, 74868, '\p{^Gc=po}', "");
Expect(1, 74868, '\P{Gc=po}', "");
Expect(0, 74868, '\P{^Gc=po}', "");
Expect(1, 74867, '\p{Gc:	-Po}', "");
Expect(0, 74867, '\p{^Gc:	-Po}', "");
Expect(0, 74867, '\P{Gc:	-Po}', "");
Expect(1, 74867, '\P{^Gc:	-Po}', "");
Expect(0, 74868, '\p{Gc:	-Po}', "");
Expect(1, 74868, '\p{^Gc:	-Po}', "");
Expect(1, 74868, '\P{Gc:	-Po}', "");
Expect(0, 74868, '\P{^Gc:	-Po}', "");
Error('\p{Category=/a/ Other_Punctuation}');
Error('\P{Category=/a/ Other_Punctuation}');
Expect(1, 74867, '\p{Category: otherpunctuation}', "");
Expect(0, 74867, '\p{^Category: otherpunctuation}', "");
Expect(0, 74867, '\P{Category: otherpunctuation}', "");
Expect(1, 74867, '\P{^Category: otherpunctuation}', "");
Expect(0, 74868, '\p{Category: otherpunctuation}', "");
Expect(1, 74868, '\p{^Category: otherpunctuation}', "");
Expect(1, 74868, '\P{Category: otherpunctuation}', "");
Expect(0, 74868, '\P{^Category: otherpunctuation}', "");
Expect(1, 74867, '\p{Category=		Other_Punctuation}', "");
Expect(0, 74867, '\p{^Category=		Other_Punctuation}', "");
Expect(0, 74867, '\P{Category=		Other_Punctuation}', "");
Expect(1, 74867, '\P{^Category=		Other_Punctuation}', "");
Expect(0, 74868, '\p{Category=		Other_Punctuation}', "");
Expect(1, 74868, '\p{^Category=		Other_Punctuation}', "");
Expect(1, 74868, '\P{Category=		Other_Punctuation}', "");
Expect(0, 74868, '\P{^Category=		Other_Punctuation}', "");
Error('\p{Is_General_Category= -PO:=}');
Error('\P{Is_General_Category= -PO:=}');
Expect(1, 74867, '\p{Is_General_Category:po}', "");
Expect(0, 74867, '\p{^Is_General_Category:po}', "");
Expect(0, 74867, '\P{Is_General_Category:po}', "");
Expect(1, 74867, '\P{^Is_General_Category:po}', "");
Expect(0, 74868, '\p{Is_General_Category:po}', "");
Expect(1, 74868, '\p{^Is_General_Category:po}', "");
Expect(1, 74868, '\P{Is_General_Category:po}', "");
Expect(0, 74868, '\P{^Is_General_Category:po}', "");
Expect(1, 74867, '\p{Is_General_Category= 	po}', "");
Expect(0, 74867, '\p{^Is_General_Category= 	po}', "");
Expect(0, 74867, '\P{Is_General_Category= 	po}', "");
Expect(1, 74867, '\P{^Is_General_Category= 	po}', "");
Expect(0, 74868, '\p{Is_General_Category= 	po}', "");
Expect(1, 74868, '\p{^Is_General_Category= 	po}', "");
Expect(1, 74868, '\P{Is_General_Category= 	po}', "");
Expect(0, 74868, '\P{^Is_General_Category= 	po}', "");
Error('\p{Is_Gc=/a/_	OTHER_Punctuation}');
Error('\P{Is_Gc=/a/_	OTHER_Punctuation}');
Expect(1, 74867, '\p{Is_Gc=otherpunctuation}', "");
Expect(0, 74867, '\p{^Is_Gc=otherpunctuation}', "");
Expect(0, 74867, '\P{Is_Gc=otherpunctuation}', "");
Expect(1, 74867, '\P{^Is_Gc=otherpunctuation}', "");
Expect(0, 74868, '\p{Is_Gc=otherpunctuation}', "");
Expect(1, 74868, '\p{^Is_Gc=otherpunctuation}', "");
Expect(1, 74868, '\P{Is_Gc=otherpunctuation}', "");
Expect(0, 74868, '\P{^Is_Gc=otherpunctuation}', "");
Expect(1, 74867, '\p{Is_Gc=_OTHER_PUNCTUATION}', "");
Expect(0, 74867, '\p{^Is_Gc=_OTHER_PUNCTUATION}', "");
Expect(0, 74867, '\P{Is_Gc=_OTHER_PUNCTUATION}', "");
Expect(1, 74867, '\P{^Is_Gc=_OTHER_PUNCTUATION}', "");
Expect(0, 74868, '\p{Is_Gc=_OTHER_PUNCTUATION}', "");
Expect(1, 74868, '\p{^Is_Gc=_OTHER_PUNCTUATION}', "");
Expect(1, 74868, '\P{Is_Gc=_OTHER_PUNCTUATION}', "");
Expect(0, 74868, '\P{^Is_Gc=_OTHER_PUNCTUATION}', "");
Error('\p{Is_Category=:=_	PO}');
Error('\P{Is_Category=:=_	PO}');
Expect(1, 74867, '\p{Is_Category=po}', "");
Expect(0, 74867, '\p{^Is_Category=po}', "");
Expect(0, 74867, '\P{Is_Category=po}', "");
Expect(1, 74867, '\P{^Is_Category=po}', "");
Expect(0, 74868, '\p{Is_Category=po}', "");
Expect(1, 74868, '\p{^Is_Category=po}', "");
Expect(1, 74868, '\P{Is_Category=po}', "");
Expect(0, 74868, '\P{^Is_Category=po}', "");
Expect(1, 74867, '\p{Is_Category=_Po}', "");
Expect(0, 74867, '\p{^Is_Category=_Po}', "");
Expect(0, 74867, '\P{Is_Category=_Po}', "");
Expect(1, 74867, '\P{^Is_Category=_Po}', "");
Expect(0, 74868, '\p{Is_Category=_Po}', "");
Expect(1, 74868, '\p{^Is_Category=_Po}', "");
Expect(1, 74868, '\P{Is_Category=_Po}', "");
Expect(0, 74868, '\P{^Is_Category=_Po}', "");
Error('\p{General_Category=:=		Space_Separator}');
Error('\P{General_Category=:=		Space_Separator}');
Expect(1, 12288, '\p{General_Category=spaceseparator}', "");
Expect(0, 12288, '\p{^General_Category=spaceseparator}', "");
Expect(0, 12288, '\P{General_Category=spaceseparator}', "");
Expect(1, 12288, '\P{^General_Category=spaceseparator}', "");
Expect(0, 12289, '\p{General_Category=spaceseparator}', "");
Expect(1, 12289, '\p{^General_Category=spaceseparator}', "");
Expect(1, 12289, '\P{General_Category=spaceseparator}', "");
Expect(0, 12289, '\P{^General_Category=spaceseparator}', "");
Expect(1, 12288, '\p{General_Category=	 SPACE_separator}', "");
Expect(0, 12288, '\p{^General_Category=	 SPACE_separator}', "");
Expect(0, 12288, '\P{General_Category=	 SPACE_separator}', "");
Expect(1, 12288, '\P{^General_Category=	 SPACE_separator}', "");
Expect(0, 12289, '\p{General_Category=	 SPACE_separator}', "");
Expect(1, 12289, '\p{^General_Category=	 SPACE_separator}', "");
Expect(1, 12289, '\P{General_Category=	 SPACE_separator}', "");
Expect(0, 12289, '\P{^General_Category=	 SPACE_separator}', "");
Error('\p{Gc=/a/-Zs}');
Error('\P{Gc=/a/-Zs}');
Expect(1, 12288, '\p{Gc=zs}', "");
Expect(0, 12288, '\p{^Gc=zs}', "");
Expect(0, 12288, '\P{Gc=zs}', "");
Expect(1, 12288, '\P{^Gc=zs}', "");
Expect(0, 12289, '\p{Gc=zs}', "");
Expect(1, 12289, '\p{^Gc=zs}', "");
Expect(1, 12289, '\P{Gc=zs}', "");
Expect(0, 12289, '\P{^Gc=zs}', "");
Expect(1, 12288, '\p{Gc:		Zs}', "");
Expect(0, 12288, '\p{^Gc:		Zs}', "");
Expect(0, 12288, '\P{Gc:		Zs}', "");
Expect(1, 12288, '\P{^Gc:		Zs}', "");
Expect(0, 12289, '\p{Gc:		Zs}', "");
Expect(1, 12289, '\p{^Gc:		Zs}', "");
Expect(1, 12289, '\P{Gc:		Zs}', "");
Expect(0, 12289, '\P{^Gc:		Zs}', "");
Error('\p{Category= /a/Space_separator}');
Error('\P{Category= /a/Space_separator}');
Expect(1, 12288, '\p{Category=spaceseparator}', "");
Expect(0, 12288, '\p{^Category=spaceseparator}', "");
Expect(0, 12288, '\P{Category=spaceseparator}', "");
Expect(1, 12288, '\P{^Category=spaceseparator}', "");
Expect(0, 12289, '\p{Category=spaceseparator}', "");
Expect(1, 12289, '\p{^Category=spaceseparator}', "");
Expect(1, 12289, '\P{Category=spaceseparator}', "");
Expect(0, 12289, '\P{^Category=spaceseparator}', "");
Expect(1, 12288, '\p{Category=	_Space_Separator}', "");
Expect(0, 12288, '\p{^Category=	_Space_Separator}', "");
Expect(0, 12288, '\P{Category=	_Space_Separator}', "");
Expect(1, 12288, '\P{^Category=	_Space_Separator}', "");
Expect(0, 12289, '\p{Category=	_Space_Separator}', "");
Expect(1, 12289, '\p{^Category=	_Space_Separator}', "");
Expect(1, 12289, '\P{Category=	_Space_Separator}', "");
Expect(0, 12289, '\P{^Category=	_Space_Separator}', "");
Error('\p{Is_General_Category=:=	 zs}');
Error('\P{Is_General_Category=:=	 zs}');
Expect(1, 12288, '\p{Is_General_Category=zs}', "");
Expect(0, 12288, '\p{^Is_General_Category=zs}', "");
Expect(0, 12288, '\P{Is_General_Category=zs}', "");
Expect(1, 12288, '\P{^Is_General_Category=zs}', "");
Expect(0, 12289, '\p{Is_General_Category=zs}', "");
Expect(1, 12289, '\p{^Is_General_Category=zs}', "");
Expect(1, 12289, '\P{Is_General_Category=zs}', "");
Expect(0, 12289, '\P{^Is_General_Category=zs}', "");
Expect(1, 12288, '\p{Is_General_Category=_ Zs}', "");
Expect(0, 12288, '\p{^Is_General_Category=_ Zs}', "");
Expect(0, 12288, '\P{Is_General_Category=_ Zs}', "");
Expect(1, 12288, '\P{^Is_General_Category=_ Zs}', "");
Expect(0, 12289, '\p{Is_General_Category=_ Zs}', "");
Expect(1, 12289, '\p{^Is_General_Category=_ Zs}', "");
Expect(1, 12289, '\P{Is_General_Category=_ Zs}', "");
Expect(0, 12289, '\P{^Is_General_Category=_ Zs}', "");
Error('\p{Is_Gc=:=	Space_SEPARATOR}');
Error('\P{Is_Gc=:=	Space_SEPARATOR}');
Expect(1, 12288, '\p{Is_Gc=spaceseparator}', "");
Expect(0, 12288, '\p{^Is_Gc=spaceseparator}', "");
Expect(0, 12288, '\P{Is_Gc=spaceseparator}', "");
Expect(1, 12288, '\P{^Is_Gc=spaceseparator}', "");
Expect(0, 12289, '\p{Is_Gc=spaceseparator}', "");
Expect(1, 12289, '\p{^Is_Gc=spaceseparator}', "");
Expect(1, 12289, '\P{Is_Gc=spaceseparator}', "");
Expect(0, 12289, '\P{^Is_Gc=spaceseparator}', "");
Expect(1, 12288, '\p{Is_Gc= SPACE_Separator}', "");
Expect(0, 12288, '\p{^Is_Gc= SPACE_Separator}', "");
Expect(0, 12288, '\P{Is_Gc= SPACE_Separator}', "");
Expect(1, 12288, '\P{^Is_Gc= SPACE_Separator}', "");
Expect(0, 12289, '\p{Is_Gc= SPACE_Separator}', "");
Expect(1, 12289, '\p{^Is_Gc= SPACE_Separator}', "");
Expect(1, 12289, '\P{Is_Gc= SPACE_Separator}', "");
Expect(0, 12289, '\P{^Is_Gc= SPACE_Separator}', "");
Error('\p{Is_Category: _/a/Zs}');
Error('\P{Is_Category: _/a/Zs}');
Expect(1, 12288, '\p{Is_Category=zs}', "");
Expect(0, 12288, '\p{^Is_Category=zs}', "");
Expect(0, 12288, '\P{Is_Category=zs}', "");
Expect(1, 12288, '\P{^Is_Category=zs}', "");
Expect(0, 12289, '\p{Is_Category=zs}', "");
Expect(1, 12289, '\p{^Is_Category=zs}', "");
Expect(1, 12289, '\P{Is_Category=zs}', "");
Expect(0, 12289, '\P{^Is_Category=zs}', "");
Expect(1, 12288, '\p{Is_Category=	 zs}', "");
Expect(0, 12288, '\p{^Is_Category=	 zs}', "");
Expect(0, 12288, '\P{Is_Category=	 zs}', "");
Expect(1, 12288, '\P{^Is_Category=	 zs}', "");
Expect(0, 12289, '\p{Is_Category=	 zs}', "");
Expect(1, 12289, '\p{^Is_Category=	 zs}', "");
Expect(1, 12289, '\P{Is_Category=	 zs}', "");
Expect(0, 12289, '\P{^Is_Category=	 zs}', "");
Error('\p{General_Category:   	 OTHER_NUMBER/a/}');
Error('\P{General_Category:   	 OTHER_NUMBER/a/}');
Expect(1, 127242, '\p{General_Category=othernumber}', "");
Expect(0, 127242, '\p{^General_Category=othernumber}', "");
Expect(0, 127242, '\P{General_Category=othernumber}', "");
Expect(1, 127242, '\P{^General_Category=othernumber}', "");
Expect(0, 127243, '\p{General_Category=othernumber}', "");
Expect(1, 127243, '\p{^General_Category=othernumber}', "");
Expect(1, 127243, '\P{General_Category=othernumber}', "");
Expect(0, 127243, '\P{^General_Category=othernumber}', "");
Expect(1, 127242, '\p{General_Category=_Other_Number}', "");
Expect(0, 127242, '\p{^General_Category=_Other_Number}', "");
Expect(0, 127242, '\P{General_Category=_Other_Number}', "");
Expect(1, 127242, '\P{^General_Category=_Other_Number}', "");
Expect(0, 127243, '\p{General_Category=_Other_Number}', "");
Expect(1, 127243, '\p{^General_Category=_Other_Number}', "");
Expect(1, 127243, '\P{General_Category=_Other_Number}', "");
Expect(0, 127243, '\P{^General_Category=_Other_Number}', "");
Error('\p{Gc=:=-NO}');
Error('\P{Gc=:=-NO}');
Expect(1, 127242, '\p{Gc=no}', "");
Expect(0, 127242, '\p{^Gc=no}', "");
Expect(0, 127242, '\P{Gc=no}', "");
Expect(1, 127242, '\P{^Gc=no}', "");
Expect(0, 127243, '\p{Gc=no}', "");
Expect(1, 127243, '\p{^Gc=no}', "");
Expect(1, 127243, '\P{Gc=no}', "");
Expect(0, 127243, '\P{^Gc=no}', "");
Expect(1, 127242, '\p{Gc:  No}', "");
Expect(0, 127242, '\p{^Gc:  No}', "");
Expect(0, 127242, '\P{Gc:  No}', "");
Expect(1, 127242, '\P{^Gc:  No}', "");
Expect(0, 127243, '\p{Gc:  No}', "");
Expect(1, 127243, '\p{^Gc:  No}', "");
Expect(1, 127243, '\P{Gc:  No}', "");
Expect(0, 127243, '\P{^Gc:  No}', "");
Error('\p{Category=-/a/other_Number}');
Error('\P{Category=-/a/other_Number}');
Expect(1, 127242, '\p{Category: othernumber}', "");
Expect(0, 127242, '\p{^Category: othernumber}', "");
Expect(0, 127242, '\P{Category: othernumber}', "");
Expect(1, 127242, '\P{^Category: othernumber}', "");
Expect(0, 127243, '\p{Category: othernumber}', "");
Expect(1, 127243, '\p{^Category: othernumber}', "");
Expect(1, 127243, '\P{Category: othernumber}', "");
Expect(0, 127243, '\P{^Category: othernumber}', "");
Expect(1, 127242, '\p{Category= _Other_number}', "");
Expect(0, 127242, '\p{^Category= _Other_number}', "");
Expect(0, 127242, '\P{Category= _Other_number}', "");
Expect(1, 127242, '\P{^Category= _Other_number}', "");
Expect(0, 127243, '\p{Category= _Other_number}', "");
Expect(1, 127243, '\p{^Category= _Other_number}', "");
Expect(1, 127243, '\P{Category= _Other_number}', "");
Expect(0, 127243, '\P{^Category= _Other_number}', "");
Error('\p{Is_General_Category=_/a/NO}');
Error('\P{Is_General_Category=_/a/NO}');
Expect(1, 127242, '\p{Is_General_Category=no}', "");
Expect(0, 127242, '\p{^Is_General_Category=no}', "");
Expect(0, 127242, '\P{Is_General_Category=no}', "");
Expect(1, 127242, '\P{^Is_General_Category=no}', "");
Expect(0, 127243, '\p{Is_General_Category=no}', "");
Expect(1, 127243, '\p{^Is_General_Category=no}', "");
Expect(1, 127243, '\P{Is_General_Category=no}', "");
Expect(0, 127243, '\P{^Is_General_Category=no}', "");
Expect(1, 127242, '\p{Is_General_Category=- No}', "");
Expect(0, 127242, '\p{^Is_General_Category=- No}', "");
Expect(0, 127242, '\P{Is_General_Category=- No}', "");
Expect(1, 127242, '\P{^Is_General_Category=- No}', "");
Expect(0, 127243, '\p{Is_General_Category=- No}', "");
Expect(1, 127243, '\p{^Is_General_Category=- No}', "");
Expect(1, 127243, '\P{Is_General_Category=- No}', "");
Expect(0, 127243, '\P{^Is_General_Category=- No}', "");
Error('\p{Is_Gc=__Other_number/a/}');
Error('\P{Is_Gc=__Other_number/a/}');
Expect(1, 127242, '\p{Is_Gc=othernumber}', "");
Expect(0, 127242, '\p{^Is_Gc=othernumber}', "");
Expect(0, 127242, '\P{Is_Gc=othernumber}', "");
Expect(1, 127242, '\P{^Is_Gc=othernumber}', "");
Expect(0, 127243, '\p{Is_Gc=othernumber}', "");
Expect(1, 127243, '\p{^Is_Gc=othernumber}', "");
Expect(1, 127243, '\P{Is_Gc=othernumber}', "");
Expect(0, 127243, '\P{^Is_Gc=othernumber}', "");
Expect(1, 127242, '\p{Is_Gc= -OTHER_Number}', "");
Expect(0, 127242, '\p{^Is_Gc= -OTHER_Number}', "");
Expect(0, 127242, '\P{Is_Gc= -OTHER_Number}', "");
Expect(1, 127242, '\P{^Is_Gc= -OTHER_Number}', "");
Expect(0, 127243, '\p{Is_Gc= -OTHER_Number}', "");
Expect(1, 127243, '\p{^Is_Gc= -OTHER_Number}', "");
Expect(1, 127243, '\P{Is_Gc= -OTHER_Number}', "");
Expect(0, 127243, '\P{^Is_Gc= -OTHER_Number}', "");
Error('\p{Is_Category=-NO/a/}');
Error('\P{Is_Category=-NO/a/}');
Expect(1, 127242, '\p{Is_Category=no}', "");
Expect(0, 127242, '\p{^Is_Category=no}', "");
Expect(0, 127242, '\P{Is_Category=no}', "");
Expect(1, 127242, '\P{^Is_Category=no}', "");
Expect(0, 127243, '\p{Is_Category=no}', "");
Expect(1, 127243, '\p{^Is_Category=no}', "");
Expect(1, 127243, '\P{Is_Category=no}', "");
Expect(0, 127243, '\P{^Is_Category=no}', "");
Expect(1, 127242, '\p{Is_Category=_	no}', "");
Expect(0, 127242, '\p{^Is_Category=_	no}', "");
Expect(0, 127242, '\P{Is_Category=_	no}', "");
Expect(1, 127242, '\P{^Is_Category=_	no}', "");
Expect(0, 127243, '\p{Is_Category=_	no}', "");
Expect(1, 127243, '\p{^Is_Category=_	no}', "");
Expect(1, 127243, '\P{Is_Category=_	no}', "");
Expect(0, 127243, '\P{^Is_Category=_	no}', "");
Error('\p{General_Category=:=- other}');
Error('\P{General_Category=:=- other}');
Expect(1, 918000, '\p{General_Category:   other}', "");
Expect(0, 918000, '\p{^General_Category:   other}', "");
Expect(0, 918000, '\P{General_Category:   other}', "");
Expect(1, 918000, '\P{^General_Category:   other}', "");
Expect(0, 917999, '\p{General_Category:   other}', "");
Expect(1, 917999, '\p{^General_Category:   other}', "");
Expect(1, 917999, '\P{General_Category:   other}', "");
Expect(0, 917999, '\P{^General_Category:   other}', "");
Expect(1, 918000, '\p{General_Category=	 Other}', "");
Expect(0, 918000, '\p{^General_Category=	 Other}', "");
Expect(0, 918000, '\P{General_Category=	 Other}', "");
Expect(1, 918000, '\P{^General_Category=	 Other}', "");
Expect(0, 917999, '\p{General_Category=	 Other}', "");
Expect(1, 917999, '\p{^General_Category=	 Other}', "");
Expect(1, 917999, '\P{General_Category=	 Other}', "");
Expect(0, 917999, '\P{^General_Category=	 Other}', "");
Error('\p{Gc:		-C/a/}');
Error('\P{Gc:		-C/a/}');
Expect(1, 918000, '\p{Gc=c}', "");
Expect(0, 918000, '\p{^Gc=c}', "");
Expect(0, 918000, '\P{Gc=c}', "");
Expect(1, 918000, '\P{^Gc=c}', "");
Expect(0, 917999, '\p{Gc=c}', "");
Expect(1, 917999, '\p{^Gc=c}', "");
Expect(1, 917999, '\P{Gc=c}', "");
Expect(0, 917999, '\P{^Gc=c}', "");
Expect(1, 918000, '\p{Gc:   _	C}', "");
Expect(0, 918000, '\p{^Gc:   _	C}', "");
Expect(0, 918000, '\P{Gc:   _	C}', "");
Expect(1, 918000, '\P{^Gc:   _	C}', "");
Expect(0, 917999, '\p{Gc:   _	C}', "");
Expect(1, 917999, '\p{^Gc:   _	C}', "");
Expect(1, 917999, '\P{Gc:   _	C}', "");
Expect(0, 917999, '\P{^Gc:   _	C}', "");
Error('\p{Category=/a/_Other}');
Error('\P{Category=/a/_Other}');
Expect(1, 918000, '\p{Category=other}', "");
Expect(0, 918000, '\p{^Category=other}', "");
Expect(0, 918000, '\P{Category=other}', "");
Expect(1, 918000, '\P{^Category=other}', "");
Expect(0, 917999, '\p{Category=other}', "");
Expect(1, 917999, '\p{^Category=other}', "");
Expect(1, 917999, '\P{Category=other}', "");
Expect(0, 917999, '\P{^Category=other}', "");
Expect(1, 918000, '\p{Category=_-Other}', "");
Expect(0, 918000, '\p{^Category=_-Other}', "");
Expect(0, 918000, '\P{Category=_-Other}', "");
Expect(1, 918000, '\P{^Category=_-Other}', "");
Expect(0, 917999, '\p{Category=_-Other}', "");
Expect(1, 917999, '\p{^Category=_-Other}', "");
Expect(1, 917999, '\P{Category=_-Other}', "");
Expect(0, 917999, '\P{^Category=_-Other}', "");
Error('\p{Is_General_Category=:=-	C}');
Error('\P{Is_General_Category=:=-	C}');
Expect(1, 918000, '\p{Is_General_Category=c}', "");
Expect(0, 918000, '\p{^Is_General_Category=c}', "");
Expect(0, 918000, '\P{Is_General_Category=c}', "");
Expect(1, 918000, '\P{^Is_General_Category=c}', "");
Expect(0, 917999, '\p{Is_General_Category=c}', "");
Expect(1, 917999, '\p{^Is_General_Category=c}', "");
Expect(1, 917999, '\P{Is_General_Category=c}', "");
Expect(0, 917999, '\P{^Is_General_Category=c}', "");
Expect(1, 918000, '\p{Is_General_Category=-_C}', "");
Expect(0, 918000, '\p{^Is_General_Category=-_C}', "");
Expect(0, 918000, '\P{Is_General_Category=-_C}', "");
Expect(1, 918000, '\P{^Is_General_Category=-_C}', "");
Expect(0, 917999, '\p{Is_General_Category=-_C}', "");
Expect(1, 917999, '\p{^Is_General_Category=-_C}', "");
Expect(1, 917999, '\P{Is_General_Category=-_C}', "");
Expect(0, 917999, '\P{^Is_General_Category=-_C}', "");
Error('\p{Is_Gc= :=OTHER}');
Error('\P{Is_Gc= :=OTHER}');
Expect(1, 918000, '\p{Is_Gc=other}', "");
Expect(0, 918000, '\p{^Is_Gc=other}', "");
Expect(0, 918000, '\P{Is_Gc=other}', "");
Expect(1, 918000, '\P{^Is_Gc=other}', "");
Expect(0, 917999, '\p{Is_Gc=other}', "");
Expect(1, 917999, '\p{^Is_Gc=other}', "");
Expect(1, 917999, '\P{Is_Gc=other}', "");
Expect(0, 917999, '\P{^Is_Gc=other}', "");
Expect(1, 918000, '\p{Is_Gc=	-Other}', "");
Expect(0, 918000, '\p{^Is_Gc=	-Other}', "");
Expect(0, 918000, '\P{Is_Gc=	-Other}', "");
Expect(1, 918000, '\P{^Is_Gc=	-Other}', "");
Expect(0, 917999, '\p{Is_Gc=	-Other}', "");
Expect(1, 917999, '\p{^Is_Gc=	-Other}', "");
Expect(1, 917999, '\P{Is_Gc=	-Other}', "");
Expect(0, 917999, '\P{^Is_Gc=	-Other}', "");
Error('\p{Is_Category=_:=C}');
Error('\P{Is_Category=_:=C}');
Expect(1, 918000, '\p{Is_Category=c}', "");
Expect(0, 918000, '\p{^Is_Category=c}', "");
Expect(0, 918000, '\P{Is_Category=c}', "");
Expect(1, 918000, '\P{^Is_Category=c}', "");
Expect(0, 917999, '\p{Is_Category=c}', "");
Expect(1, 917999, '\p{^Is_Category=c}', "");
Expect(1, 917999, '\P{Is_Category=c}', "");
Expect(0, 917999, '\P{^Is_Category=c}', "");
Expect(1, 918000, '\p{Is_Category= _C}', "");
Expect(0, 918000, '\p{^Is_Category= _C}', "");
Expect(0, 918000, '\P{Is_Category= _C}', "");
Expect(1, 918000, '\P{^Is_Category= _C}', "");
Expect(0, 917999, '\p{Is_Category= _C}', "");
Expect(1, 917999, '\p{^Is_Category= _C}', "");
Expect(1, 917999, '\P{Is_Category= _C}', "");
Expect(0, 917999, '\P{^Is_Category= _C}', "");
Error('\p{General_Category:	/a/ -Enclosing_Mark}');
Error('\P{General_Category:	/a/ -Enclosing_Mark}');
Expect(1, 42610, '\p{General_Category=enclosingmark}', "");
Expect(0, 42610, '\p{^General_Category=enclosingmark}', "");
Expect(0, 42610, '\P{General_Category=enclosingmark}', "");
Expect(1, 42610, '\P{^General_Category=enclosingmark}', "");
Expect(0, 42611, '\p{General_Category=enclosingmark}', "");
Expect(1, 42611, '\p{^General_Category=enclosingmark}', "");
Expect(1, 42611, '\P{General_Category=enclosingmark}', "");
Expect(0, 42611, '\P{^General_Category=enclosingmark}', "");
Expect(1, 42610, '\p{General_Category= Enclosing_MARK}', "");
Expect(0, 42610, '\p{^General_Category= Enclosing_MARK}', "");
Expect(0, 42610, '\P{General_Category= Enclosing_MARK}', "");
Expect(1, 42610, '\P{^General_Category= Enclosing_MARK}', "");
Expect(0, 42611, '\p{General_Category= Enclosing_MARK}', "");
Expect(1, 42611, '\p{^General_Category= Enclosing_MARK}', "");
Expect(1, 42611, '\P{General_Category= Enclosing_MARK}', "");
Expect(0, 42611, '\P{^General_Category= Enclosing_MARK}', "");
Error('\p{Gc=/a/-_me}');
Error('\P{Gc=/a/-_me}');
Expect(1, 42610, '\p{Gc=me}', "");
Expect(0, 42610, '\p{^Gc=me}', "");
Expect(0, 42610, '\P{Gc=me}', "");
Expect(1, 42610, '\P{^Gc=me}', "");
Expect(0, 42611, '\p{Gc=me}', "");
Expect(1, 42611, '\p{^Gc=me}', "");
Expect(1, 42611, '\P{Gc=me}', "");
Expect(0, 42611, '\P{^Gc=me}', "");
Expect(1, 42610, '\p{Gc=_ Me}', "");
Expect(0, 42610, '\p{^Gc=_ Me}', "");
Expect(0, 42610, '\P{Gc=_ Me}', "");
Expect(1, 42610, '\P{^Gc=_ Me}', "");
Expect(0, 42611, '\p{Gc=_ Me}', "");
Expect(1, 42611, '\p{^Gc=_ Me}', "");
Expect(1, 42611, '\P{Gc=_ Me}', "");
Expect(0, 42611, '\P{^Gc=_ Me}', "");
Error('\p{Category=:=Enclosing_Mark}');
Error('\P{Category=:=Enclosing_Mark}');
Expect(1, 42610, '\p{Category: enclosingmark}', "");
Expect(0, 42610, '\p{^Category: enclosingmark}', "");
Expect(0, 42610, '\P{Category: enclosingmark}', "");
Expect(1, 42610, '\P{^Category: enclosingmark}', "");
Expect(0, 42611, '\p{Category: enclosingmark}', "");
Expect(1, 42611, '\p{^Category: enclosingmark}', "");
Expect(1, 42611, '\P{Category: enclosingmark}', "");
Expect(0, 42611, '\P{^Category: enclosingmark}', "");
Expect(1, 42610, '\p{Category= _Enclosing_MARK}', "");
Expect(0, 42610, '\p{^Category= _Enclosing_MARK}', "");
Expect(0, 42610, '\P{Category= _Enclosing_MARK}', "");
Expect(1, 42610, '\P{^Category= _Enclosing_MARK}', "");
Expect(0, 42611, '\p{Category= _Enclosing_MARK}', "");
Expect(1, 42611, '\p{^Category= _Enclosing_MARK}', "");
Expect(1, 42611, '\P{Category= _Enclosing_MARK}', "");
Expect(0, 42611, '\P{^Category= _Enclosing_MARK}', "");
Error('\p{Is_General_Category= _me/a/}');
Error('\P{Is_General_Category= _me/a/}');
Expect(1, 42610, '\p{Is_General_Category=me}', "");
Expect(0, 42610, '\p{^Is_General_Category=me}', "");
Expect(0, 42610, '\P{Is_General_Category=me}', "");
Expect(1, 42610, '\P{^Is_General_Category=me}', "");
Expect(0, 42611, '\p{Is_General_Category=me}', "");
Expect(1, 42611, '\p{^Is_General_Category=me}', "");
Expect(1, 42611, '\P{Is_General_Category=me}', "");
Expect(0, 42611, '\P{^Is_General_Category=me}', "");
Expect(1, 42610, '\p{Is_General_Category= Me}', "");
Expect(0, 42610, '\p{^Is_General_Category= Me}', "");
Expect(0, 42610, '\P{Is_General_Category= Me}', "");
Expect(1, 42610, '\P{^Is_General_Category= Me}', "");
Expect(0, 42611, '\p{Is_General_Category= Me}', "");
Expect(1, 42611, '\p{^Is_General_Category= Me}', "");
Expect(1, 42611, '\P{Is_General_Category= Me}', "");
Expect(0, 42611, '\P{^Is_General_Category= Me}', "");
Error('\p{Is_Gc=:=ENCLOSING_MARK}');
Error('\P{Is_Gc=:=ENCLOSING_MARK}');
Expect(1, 42610, '\p{Is_Gc=enclosingmark}', "");
Expect(0, 42610, '\p{^Is_Gc=enclosingmark}', "");
Expect(0, 42610, '\P{Is_Gc=enclosingmark}', "");
Expect(1, 42610, '\P{^Is_Gc=enclosingmark}', "");
Expect(0, 42611, '\p{Is_Gc=enclosingmark}', "");
Expect(1, 42611, '\p{^Is_Gc=enclosingmark}', "");
Expect(1, 42611, '\P{Is_Gc=enclosingmark}', "");
Expect(0, 42611, '\P{^Is_Gc=enclosingmark}', "");
Expect(1, 42610, '\p{Is_Gc:_ENCLOSING_Mark}', "");
Expect(0, 42610, '\p{^Is_Gc:_ENCLOSING_Mark}', "");
Expect(0, 42610, '\P{Is_Gc:_ENCLOSING_Mark}', "");
Expect(1, 42610, '\P{^Is_Gc:_ENCLOSING_Mark}', "");
Expect(0, 42611, '\p{Is_Gc:_ENCLOSING_Mark}', "");
Expect(1, 42611, '\p{^Is_Gc:_ENCLOSING_Mark}', "");
Expect(1, 42611, '\P{Is_Gc:_ENCLOSING_Mark}', "");
Expect(0, 42611, '\P{^Is_Gc:_ENCLOSING_Mark}', "");
Error('\p{Is_Category=-Me:=}');
Error('\P{Is_Category=-Me:=}');
Expect(1, 42610, '\p{Is_Category: me}', "");
Expect(0, 42610, '\p{^Is_Category: me}', "");
Expect(0, 42610, '\P{Is_Category: me}', "");
Expect(1, 42610, '\P{^Is_Category: me}', "");
Expect(0, 42611, '\p{Is_Category: me}', "");
Expect(1, 42611, '\p{^Is_Category: me}', "");
Expect(1, 42611, '\P{Is_Category: me}', "");
Expect(0, 42611, '\P{^Is_Category: me}', "");
Expect(1, 42610, '\p{Is_Category= _me}', "");
Expect(0, 42610, '\p{^Is_Category= _me}', "");
Expect(0, 42610, '\P{Is_Category= _me}', "");
Expect(1, 42610, '\P{^Is_Category= _me}', "");
Expect(0, 42611, '\p{Is_Category= _me}', "");
Expect(1, 42611, '\p{^Is_Category= _me}', "");
Expect(1, 42611, '\P{Is_Category= _me}', "");
Expect(0, 42611, '\P{^Is_Category= _me}', "");
Error('\p{General_Category=:=	 Titlecase_Letter}');
Error('\P{General_Category=:=	 Titlecase_Letter}');
Expect(1, 8188, '\p{General_Category=titlecaseletter}', "");
Expect(0, 8188, '\p{^General_Category=titlecaseletter}', "");
Expect(0, 8188, '\P{General_Category=titlecaseletter}', "");
Expect(1, 8188, '\P{^General_Category=titlecaseletter}', "");
Expect(0, 8189, '\p{General_Category=titlecaseletter}', "");
Expect(1, 8189, '\p{^General_Category=titlecaseletter}', "");
Expect(1, 8189, '\P{General_Category=titlecaseletter}', "");
Expect(0, 8189, '\P{^General_Category=titlecaseletter}', "");
Expect(1, 8188, '\p{General_Category= 	titlecase_Letter}', "");
Expect(0, 8188, '\p{^General_Category= 	titlecase_Letter}', "");
Expect(0, 8188, '\P{General_Category= 	titlecase_Letter}', "");
Expect(1, 8188, '\P{^General_Category= 	titlecase_Letter}', "");
Expect(0, 8189, '\p{General_Category= 	titlecase_Letter}', "");
Expect(1, 8189, '\p{^General_Category= 	titlecase_Letter}', "");
Expect(1, 8189, '\P{General_Category= 	titlecase_Letter}', "");
Expect(0, 8189, '\P{^General_Category= 	titlecase_Letter}', "");
Error('\p{Gc=:=Lt}');
Error('\P{Gc=:=Lt}');
Expect(1, 8188, '\p{Gc=lt}', "");
Expect(0, 8188, '\p{^Gc=lt}', "");
Expect(0, 8188, '\P{Gc=lt}', "");
Expect(1, 8188, '\P{^Gc=lt}', "");
Expect(0, 8189, '\p{Gc=lt}', "");
Expect(1, 8189, '\p{^Gc=lt}', "");
Expect(1, 8189, '\P{Gc=lt}', "");
Expect(0, 8189, '\P{^Gc=lt}', "");
Expect(1, 8188, '\p{Gc= LT}', "");
Expect(0, 8188, '\p{^Gc= LT}', "");
Expect(0, 8188, '\P{Gc= LT}', "");
Expect(1, 8188, '\P{^Gc= LT}', "");
Expect(0, 8189, '\p{Gc= LT}', "");
Expect(1, 8189, '\p{^Gc= LT}', "");
Expect(1, 8189, '\P{Gc= LT}', "");
Expect(0, 8189, '\P{^Gc= LT}', "");
Error('\p{Category= _titlecase_letter/a/}');
Error('\P{Category= _titlecase_letter/a/}');
Expect(1, 8188, '\p{Category=titlecaseletter}', "");
Expect(0, 8188, '\p{^Category=titlecaseletter}', "");
Expect(0, 8188, '\P{Category=titlecaseletter}', "");
Expect(1, 8188, '\P{^Category=titlecaseletter}', "");
Expect(0, 8189, '\p{Category=titlecaseletter}', "");
Expect(1, 8189, '\p{^Category=titlecaseletter}', "");
Expect(1, 8189, '\P{Category=titlecaseletter}', "");
Expect(0, 8189, '\P{^Category=titlecaseletter}', "");
Expect(1, 8188, '\p{Category= _TITLECASE_letter}', "");
Expect(0, 8188, '\p{^Category= _TITLECASE_letter}', "");
Expect(0, 8188, '\P{Category= _TITLECASE_letter}', "");
Expect(1, 8188, '\P{^Category= _TITLECASE_letter}', "");
Expect(0, 8189, '\p{Category= _TITLECASE_letter}', "");
Expect(1, 8189, '\p{^Category= _TITLECASE_letter}', "");
Expect(1, 8189, '\P{Category= _TITLECASE_letter}', "");
Expect(0, 8189, '\P{^Category= _TITLECASE_letter}', "");
Error('\p{Is_General_Category=-:=lt}');
Error('\P{Is_General_Category=-:=lt}');
Expect(1, 8188, '\p{Is_General_Category=lt}', "");
Expect(0, 8188, '\p{^Is_General_Category=lt}', "");
Expect(0, 8188, '\P{Is_General_Category=lt}', "");
Expect(1, 8188, '\P{^Is_General_Category=lt}', "");
Expect(0, 8189, '\p{Is_General_Category=lt}', "");
Expect(1, 8189, '\p{^Is_General_Category=lt}', "");
Expect(1, 8189, '\P{Is_General_Category=lt}', "");
Expect(0, 8189, '\P{^Is_General_Category=lt}', "");
Expect(1, 8188, '\p{Is_General_Category=Lt}', "");
Expect(0, 8188, '\p{^Is_General_Category=Lt}', "");
Expect(0, 8188, '\P{Is_General_Category=Lt}', "");
Expect(1, 8188, '\P{^Is_General_Category=Lt}', "");
Expect(0, 8189, '\p{Is_General_Category=Lt}', "");
Expect(1, 8189, '\p{^Is_General_Category=Lt}', "");
Expect(1, 8189, '\P{Is_General_Category=Lt}', "");
Expect(0, 8189, '\P{^Is_General_Category=Lt}', "");
Error('\p{Is_Gc: /a/_ Titlecase_letter}');
Error('\P{Is_Gc: /a/_ Titlecase_letter}');
Expect(1, 8188, '\p{Is_Gc=titlecaseletter}', "");
Expect(0, 8188, '\p{^Is_Gc=titlecaseletter}', "");
Expect(0, 8188, '\P{Is_Gc=titlecaseletter}', "");
Expect(1, 8188, '\P{^Is_Gc=titlecaseletter}', "");
Expect(0, 8189, '\p{Is_Gc=titlecaseletter}', "");
Expect(1, 8189, '\p{^Is_Gc=titlecaseletter}', "");
Expect(1, 8189, '\P{Is_Gc=titlecaseletter}', "");
Expect(0, 8189, '\P{^Is_Gc=titlecaseletter}', "");
Expect(1, 8188, '\p{Is_Gc=_titlecase_LETTER}', "");
Expect(0, 8188, '\p{^Is_Gc=_titlecase_LETTER}', "");
Expect(0, 8188, '\P{Is_Gc=_titlecase_LETTER}', "");
Expect(1, 8188, '\P{^Is_Gc=_titlecase_LETTER}', "");
Expect(0, 8189, '\p{Is_Gc=_titlecase_LETTER}', "");
Expect(1, 8189, '\p{^Is_Gc=_titlecase_LETTER}', "");
Expect(1, 8189, '\P{Is_Gc=_titlecase_LETTER}', "");
Expect(0, 8189, '\P{^Is_Gc=_titlecase_LETTER}', "");
Error('\p{Is_Category:		/a/LT}');
Error('\P{Is_Category:		/a/LT}');
Expect(1, 8188, '\p{Is_Category=lt}', "");
Expect(0, 8188, '\p{^Is_Category=lt}', "");
Expect(0, 8188, '\P{Is_Category=lt}', "");
Expect(1, 8188, '\P{^Is_Category=lt}', "");
Expect(0, 8189, '\p{Is_Category=lt}', "");
Expect(1, 8189, '\p{^Is_Category=lt}', "");
Expect(1, 8189, '\P{Is_Category=lt}', "");
Expect(0, 8189, '\P{^Is_Category=lt}', "");
Expect(1, 8188, '\p{Is_Category=  Lt}', "");
Expect(0, 8188, '\p{^Is_Category=  Lt}', "");
Expect(0, 8188, '\P{Is_Category=  Lt}', "");
Expect(1, 8188, '\P{^Is_Category=  Lt}', "");
Expect(0, 8189, '\p{Is_Category=  Lt}', "");
Expect(1, 8189, '\p{^Is_Category=  Lt}', "");
Expect(1, 8189, '\P{Is_Category=  Lt}', "");
Expect(0, 8189, '\P{^Is_Category=  Lt}', "");
Error('\p{General_Category=-/a/Punctuation}');
Error('\P{General_Category=-/a/Punctuation}');
Expect(1, 74867, '\p{General_Category=punctuation}', "");
Expect(0, 74867, '\p{^General_Category=punctuation}', "");
Expect(0, 74867, '\P{General_Category=punctuation}', "");
Expect(1, 74867, '\P{^General_Category=punctuation}', "");
Expect(0, 74868, '\p{General_Category=punctuation}', "");
Expect(1, 74868, '\p{^General_Category=punctuation}', "");
Expect(1, 74868, '\P{General_Category=punctuation}', "");
Expect(0, 74868, '\P{^General_Category=punctuation}', "");
Expect(1, 74867, '\p{General_Category=- Punctuation}', "");
Expect(0, 74867, '\p{^General_Category=- Punctuation}', "");
Expect(0, 74867, '\P{General_Category=- Punctuation}', "");
Expect(1, 74867, '\P{^General_Category=- Punctuation}', "");
Expect(0, 74868, '\p{General_Category=- Punctuation}', "");
Expect(1, 74868, '\p{^General_Category=- Punctuation}', "");
Expect(1, 74868, '\P{General_Category=- Punctuation}', "");
Expect(0, 74868, '\P{^General_Category=- Punctuation}', "");
Error('\p{Gc=-:=p}');
Error('\P{Gc=-:=p}');
Expect(1, 74867, '\p{Gc=p}', "");
Expect(0, 74867, '\p{^Gc=p}', "");
Expect(0, 74867, '\P{Gc=p}', "");
Expect(1, 74867, '\P{^Gc=p}', "");
Expect(0, 74868, '\p{Gc=p}', "");
Expect(1, 74868, '\p{^Gc=p}', "");
Expect(1, 74868, '\P{Gc=p}', "");
Expect(0, 74868, '\P{^Gc=p}', "");
Expect(1, 74867, '\p{Gc= _P}', "");
Expect(0, 74867, '\p{^Gc= _P}', "");
Expect(0, 74867, '\P{Gc= _P}', "");
Expect(1, 74867, '\P{^Gc= _P}', "");
Expect(0, 74868, '\p{Gc= _P}', "");
Expect(1, 74868, '\p{^Gc= _P}', "");
Expect(1, 74868, '\P{Gc= _P}', "");
Expect(0, 74868, '\P{^Gc= _P}', "");
Error('\p{Category=/a/		punct}');
Error('\P{Category=/a/		punct}');
Expect(1, 74867, '\p{Category:   punct}', "");
Expect(0, 74867, '\p{^Category:   punct}', "");
Expect(0, 74867, '\P{Category:   punct}', "");
Expect(1, 74867, '\P{^Category:   punct}', "");
Expect(0, 74868, '\p{Category:   punct}', "");
Expect(1, 74868, '\p{^Category:   punct}', "");
Expect(1, 74868, '\P{Category:   punct}', "");
Expect(0, 74868, '\P{^Category:   punct}', "");
Expect(1, 74867, '\p{Category= _Punct}', "");
Expect(0, 74867, '\p{^Category= _Punct}', "");
Expect(0, 74867, '\P{Category= _Punct}', "");
Expect(1, 74867, '\P{^Category= _Punct}', "");
Expect(0, 74868, '\p{Category= _Punct}', "");
Expect(1, 74868, '\p{^Category= _Punct}', "");
Expect(1, 74868, '\P{Category= _Punct}', "");
Expect(0, 74868, '\P{^Category= _Punct}', "");
Error('\p{Is_General_Category=	/a/PUNCTUATION}');
Error('\P{Is_General_Category=	/a/PUNCTUATION}');
Expect(1, 74867, '\p{Is_General_Category=punctuation}', "");
Expect(0, 74867, '\p{^Is_General_Category=punctuation}', "");
Expect(0, 74867, '\P{Is_General_Category=punctuation}', "");
Expect(1, 74867, '\P{^Is_General_Category=punctuation}', "");
Expect(0, 74868, '\p{Is_General_Category=punctuation}', "");
Expect(1, 74868, '\p{^Is_General_Category=punctuation}', "");
Expect(1, 74868, '\P{Is_General_Category=punctuation}', "");
Expect(0, 74868, '\P{^Is_General_Category=punctuation}', "");
Expect(1, 74867, '\p{Is_General_Category= Punctuation}', "");
Expect(0, 74867, '\p{^Is_General_Category= Punctuation}', "");
Expect(0, 74867, '\P{Is_General_Category= Punctuation}', "");
Expect(1, 74867, '\P{^Is_General_Category= Punctuation}', "");
Expect(0, 74868, '\p{Is_General_Category= Punctuation}', "");
Expect(1, 74868, '\p{^Is_General_Category= Punctuation}', "");
Expect(1, 74868, '\P{Is_General_Category= Punctuation}', "");
Expect(0, 74868, '\P{^Is_General_Category= Punctuation}', "");
Error('\p{Is_Gc=-:=P}');
Error('\P{Is_Gc=-:=P}');
Expect(1, 74867, '\p{Is_Gc=p}', "");
Expect(0, 74867, '\p{^Is_Gc=p}', "");
Expect(0, 74867, '\P{Is_Gc=p}', "");
Expect(1, 74867, '\P{^Is_Gc=p}', "");
Expect(0, 74868, '\p{Is_Gc=p}', "");
Expect(1, 74868, '\p{^Is_Gc=p}', "");
Expect(1, 74868, '\P{Is_Gc=p}', "");
Expect(0, 74868, '\P{^Is_Gc=p}', "");
Expect(1, 74867, '\p{Is_Gc= 	p}', "");
Expect(0, 74867, '\p{^Is_Gc= 	p}', "");
Expect(0, 74867, '\P{Is_Gc= 	p}', "");
Expect(1, 74867, '\P{^Is_Gc= 	p}', "");
Expect(0, 74868, '\p{Is_Gc= 	p}', "");
Expect(1, 74868, '\p{^Is_Gc= 	p}', "");
Expect(1, 74868, '\P{Is_Gc= 	p}', "");
Expect(0, 74868, '\P{^Is_Gc= 	p}', "");
Error('\p{Is_Category=_Punct:=}');
Error('\P{Is_Category=_Punct:=}');
Expect(1, 74867, '\p{Is_Category=punct}', "");
Expect(0, 74867, '\p{^Is_Category=punct}', "");
Expect(0, 74867, '\P{Is_Category=punct}', "");
Expect(1, 74867, '\P{^Is_Category=punct}', "");
Expect(0, 74868, '\p{Is_Category=punct}', "");
Expect(1, 74868, '\p{^Is_Category=punct}', "");
Expect(1, 74868, '\P{Is_Category=punct}', "");
Expect(0, 74868, '\P{^Is_Category=punct}', "");
Expect(1, 74867, '\p{Is_Category=_	Punct}', "");
Expect(0, 74867, '\p{^Is_Category=_	Punct}', "");
Expect(0, 74867, '\P{Is_Category=_	Punct}', "");
Expect(1, 74867, '\P{^Is_Category=_	Punct}', "");
Expect(0, 74868, '\p{Is_Category=_	Punct}', "");
Expect(1, 74868, '\p{^Is_Category=_	Punct}', "");
Expect(1, 74868, '\P{Is_Category=_	Punct}', "");
Expect(0, 74868, '\P{^Is_Category=_	Punct}', "");
Error('\p{General_Category=:=	_Uppercase_LETTER}');
Error('\P{General_Category=:=	_Uppercase_LETTER}');
Expect(1, 120778, '\p{General_Category=uppercaseletter}', "");
Expect(0, 120778, '\p{^General_Category=uppercaseletter}', "");
Expect(0, 120778, '\P{General_Category=uppercaseletter}', "");
Expect(1, 120778, '\P{^General_Category=uppercaseletter}', "");
Expect(0, 120779, '\p{General_Category=uppercaseletter}', "");
Expect(1, 120779, '\p{^General_Category=uppercaseletter}', "");
Expect(1, 120779, '\P{General_Category=uppercaseletter}', "");
Expect(0, 120779, '\P{^General_Category=uppercaseletter}', "");
Expect(1, 120778, '\p{General_Category=_ uppercase_Letter}', "");
Expect(0, 120778, '\p{^General_Category=_ uppercase_Letter}', "");
Expect(0, 120778, '\P{General_Category=_ uppercase_Letter}', "");
Expect(1, 120778, '\P{^General_Category=_ uppercase_Letter}', "");
Expect(0, 120779, '\p{General_Category=_ uppercase_Letter}', "");
Expect(1, 120779, '\p{^General_Category=_ uppercase_Letter}', "");
Expect(1, 120779, '\P{General_Category=_ uppercase_Letter}', "");
Expect(0, 120779, '\P{^General_Category=_ uppercase_Letter}', "");
Error('\p{Gc=:= 	LU}');
Error('\P{Gc=:= 	LU}');
Expect(1, 120778, '\p{Gc=lu}', "");
Expect(0, 120778, '\p{^Gc=lu}', "");
Expect(0, 120778, '\P{Gc=lu}', "");
Expect(1, 120778, '\P{^Gc=lu}', "");
Expect(0, 120779, '\p{Gc=lu}', "");
Expect(1, 120779, '\p{^Gc=lu}', "");
Expect(1, 120779, '\P{Gc=lu}', "");
Expect(0, 120779, '\P{^Gc=lu}', "");
Expect(1, 120778, '\p{Gc=_lu}', "");
Expect(0, 120778, '\p{^Gc=_lu}', "");
Expect(0, 120778, '\P{Gc=_lu}', "");
Expect(1, 120778, '\P{^Gc=_lu}', "");
Expect(0, 120779, '\p{Gc=_lu}', "");
Expect(1, 120779, '\p{^Gc=_lu}', "");
Expect(1, 120779, '\P{Gc=_lu}', "");
Expect(0, 120779, '\P{^Gc=_lu}', "");
Error('\p{Category=-/a/Uppercase_Letter}');
Error('\P{Category=-/a/Uppercase_Letter}');
Expect(1, 120778, '\p{Category=uppercaseletter}', "");
Expect(0, 120778, '\p{^Category=uppercaseletter}', "");
Expect(0, 120778, '\P{Category=uppercaseletter}', "");
Expect(1, 120778, '\P{^Category=uppercaseletter}', "");
Expect(0, 120779, '\p{Category=uppercaseletter}', "");
Expect(1, 120779, '\p{^Category=uppercaseletter}', "");
Expect(1, 120779, '\P{Category=uppercaseletter}', "");
Expect(0, 120779, '\P{^Category=uppercaseletter}', "");
Expect(1, 120778, '\p{Category:     Uppercase_letter}', "");
Expect(0, 120778, '\p{^Category:     Uppercase_letter}', "");
Expect(0, 120778, '\P{Category:     Uppercase_letter}', "");
Expect(1, 120778, '\P{^Category:     Uppercase_letter}', "");
Expect(0, 120779, '\p{Category:     Uppercase_letter}', "");
Expect(1, 120779, '\p{^Category:     Uppercase_letter}', "");
Expect(1, 120779, '\P{Category:     Uppercase_letter}', "");
Expect(0, 120779, '\P{^Category:     Uppercase_letter}', "");
Error('\p{Is_General_Category= :=Lu}');
Error('\P{Is_General_Category= :=Lu}');
Expect(1, 120778, '\p{Is_General_Category: lu}', "");
Expect(0, 120778, '\p{^Is_General_Category: lu}', "");
Expect(0, 120778, '\P{Is_General_Category: lu}', "");
Expect(1, 120778, '\P{^Is_General_Category: lu}', "");
Expect(0, 120779, '\p{Is_General_Category: lu}', "");
Expect(1, 120779, '\p{^Is_General_Category: lu}', "");
Expect(1, 120779, '\P{Is_General_Category: lu}', "");
Expect(0, 120779, '\P{^Is_General_Category: lu}', "");
Expect(1, 120778, '\p{Is_General_Category=-_lu}', "");
Expect(0, 120778, '\p{^Is_General_Category=-_lu}', "");
Expect(0, 120778, '\P{Is_General_Category=-_lu}', "");
Expect(1, 120778, '\P{^Is_General_Category=-_lu}', "");
Expect(0, 120779, '\p{Is_General_Category=-_lu}', "");
Expect(1, 120779, '\p{^Is_General_Category=-_lu}', "");
Expect(1, 120779, '\P{Is_General_Category=-_lu}', "");
Expect(0, 120779, '\P{^Is_General_Category=-_lu}', "");
Error('\p{Is_Gc= :=Uppercase_Letter}');
Error('\P{Is_Gc= :=Uppercase_Letter}');
Expect(1, 120778, '\p{Is_Gc=uppercaseletter}', "");
Expect(0, 120778, '\p{^Is_Gc=uppercaseletter}', "");
Expect(0, 120778, '\P{Is_Gc=uppercaseletter}', "");
Expect(1, 120778, '\P{^Is_Gc=uppercaseletter}', "");
Expect(0, 120779, '\p{Is_Gc=uppercaseletter}', "");
Expect(1, 120779, '\p{^Is_Gc=uppercaseletter}', "");
Expect(1, 120779, '\P{Is_Gc=uppercaseletter}', "");
Expect(0, 120779, '\P{^Is_Gc=uppercaseletter}', "");
Expect(1, 120778, '\p{Is_Gc=	-UPPERCASE_Letter}', "");
Expect(0, 120778, '\p{^Is_Gc=	-UPPERCASE_Letter}', "");
Expect(0, 120778, '\P{Is_Gc=	-UPPERCASE_Letter}', "");
Expect(1, 120778, '\P{^Is_Gc=	-UPPERCASE_Letter}', "");
Expect(0, 120779, '\p{Is_Gc=	-UPPERCASE_Letter}', "");
Expect(1, 120779, '\p{^Is_Gc=	-UPPERCASE_Letter}', "");
Expect(1, 120779, '\P{Is_Gc=	-UPPERCASE_Letter}', "");
Expect(0, 120779, '\P{^Is_Gc=	-UPPERCASE_Letter}', "");
Error('\p{Is_Category=/a/_lu}');
Error('\P{Is_Category=/a/_lu}');
Expect(1, 120778, '\p{Is_Category=lu}', "");
Expect(0, 120778, '\p{^Is_Category=lu}', "");
Expect(0, 120778, '\P{Is_Category=lu}', "");
Expect(1, 120778, '\P{^Is_Category=lu}', "");
Expect(0, 120779, '\p{Is_Category=lu}', "");
Expect(1, 120779, '\p{^Is_Category=lu}', "");
Expect(1, 120779, '\P{Is_Category=lu}', "");
Expect(0, 120779, '\P{^Is_Category=lu}', "");
Expect(1, 120778, '\p{Is_Category=_-LU}', "");
Expect(0, 120778, '\p{^Is_Category=_-LU}', "");
Expect(0, 120778, '\P{Is_Category=_-LU}', "");
Expect(1, 120778, '\P{^Is_Category=_-LU}', "");
Expect(0, 120779, '\p{Is_Category=_-LU}', "");
Expect(1, 120779, '\p{^Is_Category=_-LU}', "");
Expect(1, 120779, '\P{Is_Category=_-LU}', "");
Expect(0, 120779, '\P{^Is_Category=_-LU}', "");
Error('\p{General_Category=:=  Number}');
Error('\P{General_Category=:=  Number}');
Expect(1, 127242, '\p{General_Category=number}', "");
Expect(0, 127242, '\p{^General_Category=number}', "");
Expect(0, 127242, '\P{General_Category=number}', "");
Expect(1, 127242, '\P{^General_Category=number}', "");
Expect(0, 127243, '\p{General_Category=number}', "");
Expect(1, 127243, '\p{^General_Category=number}', "");
Expect(1, 127243, '\P{General_Category=number}', "");
Expect(0, 127243, '\P{^General_Category=number}', "");
Expect(1, 127242, '\p{General_Category=_	NUMBER}', "");
Expect(0, 127242, '\p{^General_Category=_	NUMBER}', "");
Expect(0, 127242, '\P{General_Category=_	NUMBER}', "");
Expect(1, 127242, '\P{^General_Category=_	NUMBER}', "");
Expect(0, 127243, '\p{General_Category=_	NUMBER}', "");
Expect(1, 127243, '\p{^General_Category=_	NUMBER}', "");
Expect(1, 127243, '\P{General_Category=_	NUMBER}', "");
Expect(0, 127243, '\P{^General_Category=_	NUMBER}', "");
Error('\p{Gc: 	-N:=}');
Error('\P{Gc: 	-N:=}');
Expect(1, 127242, '\p{Gc:n}', "");
Expect(0, 127242, '\p{^Gc:n}', "");
Expect(0, 127242, '\P{Gc:n}', "");
Expect(1, 127242, '\P{^Gc:n}', "");
Expect(0, 127243, '\p{Gc:n}', "");
Expect(1, 127243, '\p{^Gc:n}', "");
Expect(1, 127243, '\P{Gc:n}', "");
Expect(0, 127243, '\P{^Gc:n}', "");
Expect(1, 127242, '\p{Gc=_ N}', "");
Expect(0, 127242, '\p{^Gc=_ N}', "");
Expect(0, 127242, '\P{Gc=_ N}', "");
Expect(1, 127242, '\P{^Gc=_ N}', "");
Expect(0, 127243, '\p{Gc=_ N}', "");
Expect(1, 127243, '\p{^Gc=_ N}', "");
Expect(1, 127243, '\P{Gc=_ N}', "");
Expect(0, 127243, '\P{^Gc=_ N}', "");
Error('\p{Category=:= -number}');
Error('\P{Category=:= -number}');
Expect(1, 127242, '\p{Category=number}', "");
Expect(0, 127242, '\p{^Category=number}', "");
Expect(0, 127242, '\P{Category=number}', "");
Expect(1, 127242, '\P{^Category=number}', "");
Expect(0, 127243, '\p{Category=number}', "");
Expect(1, 127243, '\p{^Category=number}', "");
Expect(1, 127243, '\P{Category=number}', "");
Expect(0, 127243, '\P{^Category=number}', "");
Expect(1, 127242, '\p{Category= Number}', "");
Expect(0, 127242, '\p{^Category= Number}', "");
Expect(0, 127242, '\P{Category= Number}', "");
Expect(1, 127242, '\P{^Category= Number}', "");
Expect(0, 127243, '\p{Category= Number}', "");
Expect(1, 127243, '\p{^Category= Number}', "");
Expect(1, 127243, '\P{Category= Number}', "");
Expect(0, 127243, '\P{^Category= Number}', "");
Error('\p{Is_General_Category:/a/ -N}');
Error('\P{Is_General_Category:/a/ -N}');
Expect(1, 127242, '\p{Is_General_Category=n}', "");
Expect(0, 127242, '\p{^Is_General_Category=n}', "");
Expect(0, 127242, '\P{Is_General_Category=n}', "");
Expect(1, 127242, '\P{^Is_General_Category=n}', "");
Expect(0, 127243, '\p{Is_General_Category=n}', "");
Expect(1, 127243, '\p{^Is_General_Category=n}', "");
Expect(1, 127243, '\P{Is_General_Category=n}', "");
Expect(0, 127243, '\P{^Is_General_Category=n}', "");
Expect(1, 127242, '\p{Is_General_Category=-N}', "");
Expect(0, 127242, '\p{^Is_General_Category=-N}', "");
Expect(0, 127242, '\P{Is_General_Category=-N}', "");
Expect(1, 127242, '\P{^Is_General_Category=-N}', "");
Expect(0, 127243, '\p{Is_General_Category=-N}', "");
Expect(1, 127243, '\p{^Is_General_Category=-N}', "");
Expect(1, 127243, '\P{Is_General_Category=-N}', "");
Expect(0, 127243, '\P{^Is_General_Category=-N}', "");
Error('\p{Is_Gc=:= _number}');
Error('\P{Is_Gc=:= _number}');
Expect(1, 127242, '\p{Is_Gc=number}', "");
Expect(0, 127242, '\p{^Is_Gc=number}', "");
Expect(0, 127242, '\P{Is_Gc=number}', "");
Expect(1, 127242, '\P{^Is_Gc=number}', "");
Expect(0, 127243, '\p{Is_Gc=number}', "");
Expect(1, 127243, '\p{^Is_Gc=number}', "");
Expect(1, 127243, '\P{Is_Gc=number}', "");
Expect(0, 127243, '\P{^Is_Gc=number}', "");
Expect(1, 127242, '\p{Is_Gc=-NUMBER}', "");
Expect(0, 127242, '\p{^Is_Gc=-NUMBER}', "");
Expect(0, 127242, '\P{Is_Gc=-NUMBER}', "");
Expect(1, 127242, '\P{^Is_Gc=-NUMBER}', "");
Expect(0, 127243, '\p{Is_Gc=-NUMBER}', "");
Expect(1, 127243, '\p{^Is_Gc=-NUMBER}', "");
Expect(1, 127243, '\P{Is_Gc=-NUMBER}', "");
Expect(0, 127243, '\P{^Is_Gc=-NUMBER}', "");
Error('\p{Is_Category=:=_N}');
Error('\P{Is_Category=:=_N}');
Expect(1, 127242, '\p{Is_Category=n}', "");
Expect(0, 127242, '\p{^Is_Category=n}', "");
Expect(0, 127242, '\P{Is_Category=n}', "");
Expect(1, 127242, '\P{^Is_Category=n}', "");
Expect(0, 127243, '\p{Is_Category=n}', "");
Expect(1, 127243, '\p{^Is_Category=n}', "");
Expect(1, 127243, '\P{Is_Category=n}', "");
Expect(0, 127243, '\P{^Is_Category=n}', "");
Expect(1, 127242, '\p{Is_Category=	-n}', "");
Expect(0, 127242, '\p{^Is_Category=	-n}', "");
Expect(0, 127242, '\P{Is_Category=	-n}', "");
Expect(1, 127242, '\P{^Is_Category=	-n}', "");
Expect(0, 127243, '\p{Is_Category=	-n}', "");
Expect(1, 127243, '\p{^Is_Category=	-n}', "");
Expect(1, 127243, '\P{Is_Category=	-n}', "");
Expect(0, 127243, '\P{^Is_Category=	-n}', "");
Error('\p{General_Category=_/a/FINAL_Punctuation}');
Error('\P{General_Category=_/a/FINAL_Punctuation}');
Expect(1, 11809, '\p{General_Category=finalpunctuation}', "");
Expect(0, 11809, '\p{^General_Category=finalpunctuation}', "");
Expect(0, 11809, '\P{General_Category=finalpunctuation}', "");
Expect(1, 11809, '\P{^General_Category=finalpunctuation}', "");
Expect(0, 11810, '\p{General_Category=finalpunctuation}', "");
Expect(1, 11810, '\p{^General_Category=finalpunctuation}', "");
Expect(1, 11810, '\P{General_Category=finalpunctuation}', "");
Expect(0, 11810, '\P{^General_Category=finalpunctuation}', "");
Expect(1, 11809, '\p{General_Category=  Final_Punctuation}', "");
Expect(0, 11809, '\p{^General_Category=  Final_Punctuation}', "");
Expect(0, 11809, '\P{General_Category=  Final_Punctuation}', "");
Expect(1, 11809, '\P{^General_Category=  Final_Punctuation}', "");
Expect(0, 11810, '\p{General_Category=  Final_Punctuation}', "");
Expect(1, 11810, '\p{^General_Category=  Final_Punctuation}', "");
Expect(1, 11810, '\P{General_Category=  Final_Punctuation}', "");
Expect(0, 11810, '\P{^General_Category=  Final_Punctuation}', "");
Error('\p{Gc=_/a/Pf}');
Error('\P{Gc=_/a/Pf}');
Expect(1, 11809, '\p{Gc=pf}', "");
Expect(0, 11809, '\p{^Gc=pf}', "");
Expect(0, 11809, '\P{Gc=pf}', "");
Expect(1, 11809, '\P{^Gc=pf}', "");
Expect(0, 11810, '\p{Gc=pf}', "");
Expect(1, 11810, '\p{^Gc=pf}', "");
Expect(1, 11810, '\P{Gc=pf}', "");
Expect(0, 11810, '\P{^Gc=pf}', "");
Expect(1, 11809, '\p{Gc=		PF}', "");
Expect(0, 11809, '\p{^Gc=		PF}', "");
Expect(0, 11809, '\P{Gc=		PF}', "");
Expect(1, 11809, '\P{^Gc=		PF}', "");
Expect(0, 11810, '\p{Gc=		PF}', "");
Expect(1, 11810, '\p{^Gc=		PF}', "");
Expect(1, 11810, '\P{Gc=		PF}', "");
Expect(0, 11810, '\P{^Gc=		PF}', "");
Error('\p{Category=	:=FINAL_PUNCTUATION}');
Error('\P{Category=	:=FINAL_PUNCTUATION}');
Expect(1, 11809, '\p{Category:	finalpunctuation}', "");
Expect(0, 11809, '\p{^Category:	finalpunctuation}', "");
Expect(0, 11809, '\P{Category:	finalpunctuation}', "");
Expect(1, 11809, '\P{^Category:	finalpunctuation}', "");
Expect(0, 11810, '\p{Category:	finalpunctuation}', "");
Expect(1, 11810, '\p{^Category:	finalpunctuation}', "");
Expect(1, 11810, '\P{Category:	finalpunctuation}', "");
Expect(0, 11810, '\P{^Category:	finalpunctuation}', "");
Expect(1, 11809, '\p{Category=_-FINAL_Punctuation}', "");
Expect(0, 11809, '\p{^Category=_-FINAL_Punctuation}', "");
Expect(0, 11809, '\P{Category=_-FINAL_Punctuation}', "");
Expect(1, 11809, '\P{^Category=_-FINAL_Punctuation}', "");
Expect(0, 11810, '\p{Category=_-FINAL_Punctuation}', "");
Expect(1, 11810, '\p{^Category=_-FINAL_Punctuation}', "");
Expect(1, 11810, '\P{Category=_-FINAL_Punctuation}', "");
Expect(0, 11810, '\P{^Category=_-FINAL_Punctuation}', "");
Error('\p{Is_General_Category=pf/a/}');
Error('\P{Is_General_Category=pf/a/}');
Expect(1, 11809, '\p{Is_General_Category=pf}', "");
Expect(0, 11809, '\p{^Is_General_Category=pf}', "");
Expect(0, 11809, '\P{Is_General_Category=pf}', "");
Expect(1, 11809, '\P{^Is_General_Category=pf}', "");
Expect(0, 11810, '\p{Is_General_Category=pf}', "");
Expect(1, 11810, '\p{^Is_General_Category=pf}', "");
Expect(1, 11810, '\P{Is_General_Category=pf}', "");
Expect(0, 11810, '\P{^Is_General_Category=pf}', "");
Expect(1, 11809, '\p{Is_General_Category: PF}', "");
Expect(0, 11809, '\p{^Is_General_Category: PF}', "");
Expect(0, 11809, '\P{Is_General_Category: PF}', "");
Expect(1, 11809, '\P{^Is_General_Category: PF}', "");
Expect(0, 11810, '\p{Is_General_Category: PF}', "");
Expect(1, 11810, '\p{^Is_General_Category: PF}', "");
Expect(1, 11810, '\P{Is_General_Category: PF}', "");
Expect(0, 11810, '\P{^Is_General_Category: PF}', "");
Error('\p{Is_Gc=	 Final_Punctuation/a/}');
Error('\P{Is_Gc=	 Final_Punctuation/a/}');
Expect(1, 11809, '\p{Is_Gc=finalpunctuation}', "");
Expect(0, 11809, '\p{^Is_Gc=finalpunctuation}', "");
Expect(0, 11809, '\P{Is_Gc=finalpunctuation}', "");
Expect(1, 11809, '\P{^Is_Gc=finalpunctuation}', "");
Expect(0, 11810, '\p{Is_Gc=finalpunctuation}', "");
Expect(1, 11810, '\p{^Is_Gc=finalpunctuation}', "");
Expect(1, 11810, '\P{Is_Gc=finalpunctuation}', "");
Expect(0, 11810, '\P{^Is_Gc=finalpunctuation}', "");
Expect(1, 11809, '\p{Is_Gc=_Final_Punctuation}', "");
Expect(0, 11809, '\p{^Is_Gc=_Final_Punctuation}', "");
Expect(0, 11809, '\P{Is_Gc=_Final_Punctuation}', "");
Expect(1, 11809, '\P{^Is_Gc=_Final_Punctuation}', "");
Expect(0, 11810, '\p{Is_Gc=_Final_Punctuation}', "");
Expect(1, 11810, '\p{^Is_Gc=_Final_Punctuation}', "");
Expect(1, 11810, '\P{Is_Gc=_Final_Punctuation}', "");
Expect(0, 11810, '\P{^Is_Gc=_Final_Punctuation}', "");
Error('\p{Is_Category: /a/__PF}');
Error('\P{Is_Category: /a/__PF}');
Expect(1, 11809, '\p{Is_Category=pf}', "");
Expect(0, 11809, '\p{^Is_Category=pf}', "");
Expect(0, 11809, '\P{Is_Category=pf}', "");
Expect(1, 11809, '\P{^Is_Category=pf}', "");
Expect(0, 11810, '\p{Is_Category=pf}', "");
Expect(1, 11810, '\p{^Is_Category=pf}', "");
Expect(1, 11810, '\P{Is_Category=pf}', "");
Expect(0, 11810, '\P{^Is_Category=pf}', "");
Expect(1, 11809, '\p{Is_Category=-	Pf}', "");
Expect(0, 11809, '\p{^Is_Category=-	Pf}', "");
Expect(0, 11809, '\P{Is_Category=-	Pf}', "");
Expect(1, 11809, '\P{^Is_Category=-	Pf}', "");
Expect(0, 11810, '\p{Is_Category=-	Pf}', "");
Expect(1, 11810, '\p{^Is_Category=-	Pf}', "");
Expect(1, 11810, '\P{Is_Category=-	Pf}', "");
Expect(0, 11810, '\P{^Is_Category=-	Pf}', "");
Error('\p{General_Category=_/a/Unassigned}');
Error('\P{General_Category=_/a/Unassigned}');
Expect(1, 918000, '\p{General_Category=unassigned}', "");
Expect(0, 918000, '\p{^General_Category=unassigned}', "");
Expect(0, 918000, '\P{General_Category=unassigned}', "");
Expect(1, 918000, '\P{^General_Category=unassigned}', "");
Expect(0, 1114109, '\p{General_Category=unassigned}', "");
Expect(1, 1114109, '\p{^General_Category=unassigned}', "");
Expect(1, 1114109, '\P{General_Category=unassigned}', "");
Expect(0, 1114109, '\P{^General_Category=unassigned}', "");
Expect(1, 918000, '\p{General_Category:   Unassigned}', "");
Expect(0, 918000, '\p{^General_Category:   Unassigned}', "");
Expect(0, 918000, '\P{General_Category:   Unassigned}', "");
Expect(1, 918000, '\P{^General_Category:   Unassigned}', "");
Expect(0, 1114109, '\p{General_Category:   Unassigned}', "");
Expect(1, 1114109, '\p{^General_Category:   Unassigned}', "");
Expect(1, 1114109, '\P{General_Category:   Unassigned}', "");
Expect(0, 1114109, '\P{^General_Category:   Unassigned}', "");
Error('\p{Gc:    	cn/a/}');
Error('\P{Gc:    	cn/a/}');
Expect(1, 918000, '\p{Gc=cn}', "");
Expect(0, 918000, '\p{^Gc=cn}', "");
Expect(0, 918000, '\P{Gc=cn}', "");
Expect(1, 918000, '\P{^Gc=cn}', "");
Expect(0, 1114109, '\p{Gc=cn}', "");
Expect(1, 1114109, '\p{^Gc=cn}', "");
Expect(1, 1114109, '\P{Gc=cn}', "");
Expect(0, 1114109, '\P{^Gc=cn}', "");
Expect(1, 918000, '\p{Gc= Cn}', "");
Expect(0, 918000, '\p{^Gc= Cn}', "");
Expect(0, 918000, '\P{Gc= Cn}', "");
Expect(1, 918000, '\P{^Gc= Cn}', "");
Expect(0, 1114109, '\p{Gc= Cn}', "");
Expect(1, 1114109, '\p{^Gc= Cn}', "");
Expect(1, 1114109, '\P{Gc= Cn}', "");
Expect(0, 1114109, '\P{^Gc= Cn}', "");
Error('\p{Category=-:=Unassigned}');
Error('\P{Category=-:=Unassigned}');
Expect(1, 918000, '\p{Category=unassigned}', "");
Expect(0, 918000, '\p{^Category=unassigned}', "");
Expect(0, 918000, '\P{Category=unassigned}', "");
Expect(1, 918000, '\P{^Category=unassigned}', "");
Expect(0, 1114109, '\p{Category=unassigned}', "");
Expect(1, 1114109, '\p{^Category=unassigned}', "");
Expect(1, 1114109, '\P{Category=unassigned}', "");
Expect(0, 1114109, '\P{^Category=unassigned}', "");
Expect(1, 918000, '\p{Category= -UNASSIGNED}', "");
Expect(0, 918000, '\p{^Category= -UNASSIGNED}', "");
Expect(0, 918000, '\P{Category= -UNASSIGNED}', "");
Expect(1, 918000, '\P{^Category= -UNASSIGNED}', "");
Expect(0, 1114109, '\p{Category= -UNASSIGNED}', "");
Expect(1, 1114109, '\p{^Category= -UNASSIGNED}', "");
Expect(1, 1114109, '\P{Category= -UNASSIGNED}', "");
Expect(0, 1114109, '\P{^Category= -UNASSIGNED}', "");
Error('\p{Is_General_Category= Cn:=}');
Error('\P{Is_General_Category= Cn:=}');
Expect(1, 918000, '\p{Is_General_Category=cn}', "");
Expect(0, 918000, '\p{^Is_General_Category=cn}', "");
Expect(0, 918000, '\P{Is_General_Category=cn}', "");
Expect(1, 918000, '\P{^Is_General_Category=cn}', "");
Expect(0, 1114109, '\p{Is_General_Category=cn}', "");
Expect(1, 1114109, '\p{^Is_General_Category=cn}', "");
Expect(1, 1114109, '\P{Is_General_Category=cn}', "");
Expect(0, 1114109, '\P{^Is_General_Category=cn}', "");
Expect(1, 918000, '\p{Is_General_Category=-	CN}', "");
Expect(0, 918000, '\p{^Is_General_Category=-	CN}', "");
Expect(0, 918000, '\P{Is_General_Category=-	CN}', "");
Expect(1, 918000, '\P{^Is_General_Category=-	CN}', "");
Expect(0, 1114109, '\p{Is_General_Category=-	CN}', "");
Expect(1, 1114109, '\p{^Is_General_Category=-	CN}', "");
Expect(1, 1114109, '\P{Is_General_Category=-	CN}', "");
Expect(0, 1114109, '\P{^Is_General_Category=-	CN}', "");
Error('\p{Is_Gc=_:=unassigned}');
Error('\P{Is_Gc=_:=unassigned}');
Expect(1, 918000, '\p{Is_Gc: unassigned}', "");
Expect(0, 918000, '\p{^Is_Gc: unassigned}', "");
Expect(0, 918000, '\P{Is_Gc: unassigned}', "");
Expect(1, 918000, '\P{^Is_Gc: unassigned}', "");
Expect(0, 1114109, '\p{Is_Gc: unassigned}', "");
Expect(1, 1114109, '\p{^Is_Gc: unassigned}', "");
Expect(1, 1114109, '\P{Is_Gc: unassigned}', "");
Expect(0, 1114109, '\P{^Is_Gc: unassigned}', "");
Expect(1, 918000, '\p{Is_Gc:		Unassigned}', "");
Expect(0, 918000, '\p{^Is_Gc:		Unassigned}', "");
Expect(0, 918000, '\P{Is_Gc:		Unassigned}', "");
Expect(1, 918000, '\P{^Is_Gc:		Unassigned}', "");
Expect(0, 1114109, '\p{Is_Gc:		Unassigned}', "");
Expect(1, 1114109, '\p{^Is_Gc:		Unassigned}', "");
Expect(1, 1114109, '\P{Is_Gc:		Unassigned}', "");
Expect(0, 1114109, '\P{^Is_Gc:		Unassigned}', "");
Error('\p{Is_Category=-/a/Cn}');
Error('\P{Is_Category=-/a/Cn}');
Expect(1, 918000, '\p{Is_Category=cn}', "");
Expect(0, 918000, '\p{^Is_Category=cn}', "");
Expect(0, 918000, '\P{Is_Category=cn}', "");
Expect(1, 918000, '\P{^Is_Category=cn}', "");
Expect(0, 1114109, '\p{Is_Category=cn}', "");
Expect(1, 1114109, '\p{^Is_Category=cn}', "");
Expect(1, 1114109, '\P{Is_Category=cn}', "");
Expect(0, 1114109, '\P{^Is_Category=cn}', "");
Expect(1, 918000, '\p{Is_Category=_	Cn}', "");
Expect(0, 918000, '\p{^Is_Category=_	Cn}', "");
Expect(0, 918000, '\P{Is_Category=_	Cn}', "");
Expect(1, 918000, '\P{^Is_Category=_	Cn}', "");
Expect(0, 1114109, '\p{Is_Category=_	Cn}', "");
Expect(1, 1114109, '\p{^Is_Category=_	Cn}', "");
Expect(1, 1114109, '\P{Is_Category=_	Cn}', "");
Expect(0, 1114109, '\P{^Is_Category=_	Cn}', "");
Error('\p{General_Category=_:=Decimal_Number}');
Error('\P{General_Category=_:=Decimal_Number}');
Expect(1, 120831, '\p{General_Category=decimalnumber}', "");
Expect(0, 120831, '\p{^General_Category=decimalnumber}', "");
Expect(0, 120831, '\P{General_Category=decimalnumber}', "");
Expect(1, 120831, '\P{^General_Category=decimalnumber}', "");
Expect(0, 120832, '\p{General_Category=decimalnumber}', "");
Expect(1, 120832, '\p{^General_Category=decimalnumber}', "");
Expect(1, 120832, '\P{General_Category=decimalnumber}', "");
Expect(0, 120832, '\P{^General_Category=decimalnumber}', "");
Expect(1, 120831, '\p{General_Category= DECIMAL_number}', "");
Expect(0, 120831, '\p{^General_Category= DECIMAL_number}', "");
Expect(0, 120831, '\P{General_Category= DECIMAL_number}', "");
Expect(1, 120831, '\P{^General_Category= DECIMAL_number}', "");
Expect(0, 120832, '\p{General_Category= DECIMAL_number}', "");
Expect(1, 120832, '\p{^General_Category= DECIMAL_number}', "");
Expect(1, 120832, '\P{General_Category= DECIMAL_number}', "");
Expect(0, 120832, '\P{^General_Category= DECIMAL_number}', "");
Error('\p{Gc=:=ND}');
Error('\P{Gc=:=ND}');
Expect(1, 120831, '\p{Gc=nd}', "");
Expect(0, 120831, '\p{^Gc=nd}', "");
Expect(0, 120831, '\P{Gc=nd}', "");
Expect(1, 120831, '\P{^Gc=nd}', "");
Expect(0, 120832, '\p{Gc=nd}', "");
Expect(1, 120832, '\p{^Gc=nd}', "");
Expect(1, 120832, '\P{Gc=nd}', "");
Expect(0, 120832, '\P{^Gc=nd}', "");
Error('\p{Category:	 _Digit:=}');
Error('\P{Category:	 _Digit:=}');
Expect(1, 120831, '\p{Category=digit}', "");
Expect(0, 120831, '\p{^Category=digit}', "");
Expect(0, 120831, '\P{Category=digit}', "");
Expect(1, 120831, '\P{^Category=digit}', "");
Expect(0, 120832, '\p{Category=digit}', "");
Expect(1, 120832, '\p{^Category=digit}', "");
Expect(1, 120832, '\P{Category=digit}', "");
Expect(0, 120832, '\P{^Category=digit}', "");
Expect(1, 120831, '\p{Category=  Digit}', "");
Expect(0, 120831, '\p{^Category=  Digit}', "");
Expect(0, 120831, '\P{Category=  Digit}', "");
Expect(1, 120831, '\P{^Category=  Digit}', "");
Expect(0, 120832, '\p{Category=  Digit}', "");
Expect(1, 120832, '\p{^Category=  Digit}', "");
Expect(1, 120832, '\P{Category=  Digit}', "");
Expect(0, 120832, '\P{^Category=  Digit}', "");
Error('\p{Is_General_Category=/a/- Decimal_NUMBER}');
Error('\P{Is_General_Category=/a/- Decimal_NUMBER}');
Expect(1, 120831, '\p{Is_General_Category=decimalnumber}', "");
Expect(0, 120831, '\p{^Is_General_Category=decimalnumber}', "");
Expect(0, 120831, '\P{Is_General_Category=decimalnumber}', "");
Expect(1, 120831, '\P{^Is_General_Category=decimalnumber}', "");
Expect(0, 120832, '\p{Is_General_Category=decimalnumber}', "");
Expect(1, 120832, '\p{^Is_General_Category=decimalnumber}', "");
Expect(1, 120832, '\P{Is_General_Category=decimalnumber}', "");
Expect(0, 120832, '\P{^Is_General_Category=decimalnumber}', "");
Expect(1, 120831, '\p{Is_General_Category:   __Decimal_NUMBER}', "");
Expect(0, 120831, '\p{^Is_General_Category:   __Decimal_NUMBER}', "");
Expect(0, 120831, '\P{Is_General_Category:   __Decimal_NUMBER}', "");
Expect(1, 120831, '\P{^Is_General_Category:   __Decimal_NUMBER}', "");
Expect(0, 120832, '\p{Is_General_Category:   __Decimal_NUMBER}', "");
Expect(1, 120832, '\p{^Is_General_Category:   __Decimal_NUMBER}', "");
Expect(1, 120832, '\P{Is_General_Category:   __Decimal_NUMBER}', "");
Expect(0, 120832, '\P{^Is_General_Category:   __Decimal_NUMBER}', "");
Error('\p{Is_Gc= ND/a/}');
Error('\P{Is_Gc= ND/a/}');
Expect(1, 120831, '\p{Is_Gc=nd}', "");
Expect(0, 120831, '\p{^Is_Gc=nd}', "");
Expect(0, 120831, '\P{Is_Gc=nd}', "");
Expect(1, 120831, '\P{^Is_Gc=nd}', "");
Expect(0, 120832, '\p{Is_Gc=nd}', "");
Expect(1, 120832, '\p{^Is_Gc=nd}', "");
Expect(1, 120832, '\P{Is_Gc=nd}', "");
Expect(0, 120832, '\P{^Is_Gc=nd}', "");
Expect(1, 120831, '\p{Is_Gc=ND}', "");
Expect(0, 120831, '\p{^Is_Gc=ND}', "");
Expect(0, 120831, '\P{Is_Gc=ND}', "");
Expect(1, 120831, '\P{^Is_Gc=ND}', "");
Expect(0, 120832, '\p{Is_Gc=ND}', "");
Expect(1, 120832, '\p{^Is_Gc=ND}', "");
Expect(1, 120832, '\P{Is_Gc=ND}', "");
Expect(0, 120832, '\P{^Is_Gc=ND}', "");
Error('\p{Is_Category=:=	_Digit}');
Error('\P{Is_Category=:=	_Digit}');
Expect(1, 120831, '\p{Is_Category=digit}', "");
Expect(0, 120831, '\p{^Is_Category=digit}', "");
Expect(0, 120831, '\P{Is_Category=digit}', "");
Expect(1, 120831, '\P{^Is_Category=digit}', "");
Expect(0, 120832, '\p{Is_Category=digit}', "");
Expect(1, 120832, '\p{^Is_Category=digit}', "");
Expect(1, 120832, '\P{Is_Category=digit}', "");
Expect(0, 120832, '\P{^Is_Category=digit}', "");
Expect(1, 120831, '\p{Is_Category=-_DIGIT}', "");
Expect(0, 120831, '\p{^Is_Category=-_DIGIT}', "");
Expect(0, 120831, '\P{Is_Category=-_DIGIT}', "");
Expect(1, 120831, '\P{^Is_Category=-_DIGIT}', "");
Expect(0, 120832, '\p{Is_Category=-_DIGIT}', "");
Expect(1, 120832, '\p{^Is_Category=-_DIGIT}', "");
Expect(1, 120832, '\P{Is_Category=-_DIGIT}', "");
Expect(0, 120832, '\P{^Is_Category=-_DIGIT}', "");
Error('\p{General_Category=:=_paragraph_separator}');
Error('\P{General_Category=:=_paragraph_separator}');
Expect(1, 8233, '\p{General_Category=paragraphseparator}', "");
Expect(0, 8233, '\p{^General_Category=paragraphseparator}', "");
Expect(0, 8233, '\P{General_Category=paragraphseparator}', "");
Expect(1, 8233, '\P{^General_Category=paragraphseparator}', "");
Expect(0, 8234, '\p{General_Category=paragraphseparator}', "");
Expect(1, 8234, '\p{^General_Category=paragraphseparator}', "");
Expect(1, 8234, '\P{General_Category=paragraphseparator}', "");
Expect(0, 8234, '\P{^General_Category=paragraphseparator}', "");
Expect(1, 8233, '\p{General_Category= 	Paragraph_Separator}', "");
Expect(0, 8233, '\p{^General_Category= 	Paragraph_Separator}', "");
Expect(0, 8233, '\P{General_Category= 	Paragraph_Separator}', "");
Expect(1, 8233, '\P{^General_Category= 	Paragraph_Separator}', "");
Expect(0, 8234, '\p{General_Category= 	Paragraph_Separator}', "");
Expect(1, 8234, '\p{^General_Category= 	Paragraph_Separator}', "");
Expect(1, 8234, '\P{General_Category= 	Paragraph_Separator}', "");
Expect(0, 8234, '\P{^General_Category= 	Paragraph_Separator}', "");
Error('\p{Gc=	:=Zp}');
Error('\P{Gc=	:=Zp}');
Expect(1, 8233, '\p{Gc=zp}', "");
Expect(0, 8233, '\p{^Gc=zp}', "");
Expect(0, 8233, '\P{Gc=zp}', "");
Expect(1, 8233, '\P{^Gc=zp}', "");
Expect(0, 8234, '\p{Gc=zp}', "");
Expect(1, 8234, '\p{^Gc=zp}', "");
Expect(1, 8234, '\P{Gc=zp}', "");
Expect(0, 8234, '\P{^Gc=zp}', "");
Expect(1, 8233, '\p{Gc=--Zp}', "");
Expect(0, 8233, '\p{^Gc=--Zp}', "");
Expect(0, 8233, '\P{Gc=--Zp}', "");
Expect(1, 8233, '\P{^Gc=--Zp}', "");
Expect(0, 8234, '\p{Gc=--Zp}', "");
Expect(1, 8234, '\p{^Gc=--Zp}', "");
Expect(1, 8234, '\P{Gc=--Zp}', "");
Expect(0, 8234, '\P{^Gc=--Zp}', "");
Error('\p{Category=:=		Paragraph_Separator}');
Error('\P{Category=:=		Paragraph_Separator}');
Expect(1, 8233, '\p{Category=paragraphseparator}', "");
Expect(0, 8233, '\p{^Category=paragraphseparator}', "");
Expect(0, 8233, '\P{Category=paragraphseparator}', "");
Expect(1, 8233, '\P{^Category=paragraphseparator}', "");
Expect(0, 8234, '\p{Category=paragraphseparator}', "");
Expect(1, 8234, '\p{^Category=paragraphseparator}', "");
Expect(1, 8234, '\P{Category=paragraphseparator}', "");
Expect(0, 8234, '\P{^Category=paragraphseparator}', "");
Expect(1, 8233, '\p{Category=_	PARAGRAPH_Separator}', "");
Expect(0, 8233, '\p{^Category=_	PARAGRAPH_Separator}', "");
Expect(0, 8233, '\P{Category=_	PARAGRAPH_Separator}', "");
Expect(1, 8233, '\P{^Category=_	PARAGRAPH_Separator}', "");
Expect(0, 8234, '\p{Category=_	PARAGRAPH_Separator}', "");
Expect(1, 8234, '\p{^Category=_	PARAGRAPH_Separator}', "");
Expect(1, 8234, '\P{Category=_	PARAGRAPH_Separator}', "");
Expect(0, 8234, '\P{^Category=_	PARAGRAPH_Separator}', "");
Error('\p{Is_General_Category:-_zp/a/}');
Error('\P{Is_General_Category:-_zp/a/}');
Expect(1, 8233, '\p{Is_General_Category:   zp}', "");
Expect(0, 8233, '\p{^Is_General_Category:   zp}', "");
Expect(0, 8233, '\P{Is_General_Category:   zp}', "");
Expect(1, 8233, '\P{^Is_General_Category:   zp}', "");
Expect(0, 8234, '\p{Is_General_Category:   zp}', "");
Expect(1, 8234, '\p{^Is_General_Category:   zp}', "");
Expect(1, 8234, '\P{Is_General_Category:   zp}', "");
Expect(0, 8234, '\P{^Is_General_Category:   zp}', "");
Expect(1, 8233, '\p{Is_General_Category= Zp}', "");
Expect(0, 8233, '\p{^Is_General_Category= Zp}', "");
Expect(0, 8233, '\P{Is_General_Category= Zp}', "");
Expect(1, 8233, '\P{^Is_General_Category= Zp}', "");
Expect(0, 8234, '\p{Is_General_Category= Zp}', "");
Expect(1, 8234, '\p{^Is_General_Category= Zp}', "");
Expect(1, 8234, '\P{Is_General_Category= Zp}', "");
Expect(0, 8234, '\P{^Is_General_Category= Zp}', "");
Error('\p{Is_Gc=/a/PARAGRAPH_separator}');
Error('\P{Is_Gc=/a/PARAGRAPH_separator}');
Expect(1, 8233, '\p{Is_Gc=paragraphseparator}', "");
Expect(0, 8233, '\p{^Is_Gc=paragraphseparator}', "");
Expect(0, 8233, '\P{Is_Gc=paragraphseparator}', "");
Expect(1, 8233, '\P{^Is_Gc=paragraphseparator}', "");
Expect(0, 8234, '\p{Is_Gc=paragraphseparator}', "");
Expect(1, 8234, '\p{^Is_Gc=paragraphseparator}', "");
Expect(1, 8234, '\P{Is_Gc=paragraphseparator}', "");
Expect(0, 8234, '\P{^Is_Gc=paragraphseparator}', "");
Expect(1, 8233, '\p{Is_Gc=-	Paragraph_separator}', "");
Expect(0, 8233, '\p{^Is_Gc=-	Paragraph_separator}', "");
Expect(0, 8233, '\P{Is_Gc=-	Paragraph_separator}', "");
Expect(1, 8233, '\P{^Is_Gc=-	Paragraph_separator}', "");
Expect(0, 8234, '\p{Is_Gc=-	Paragraph_separator}', "");
Expect(1, 8234, '\p{^Is_Gc=-	Paragraph_separator}', "");
Expect(1, 8234, '\P{Is_Gc=-	Paragraph_separator}', "");
Expect(0, 8234, '\P{^Is_Gc=-	Paragraph_separator}', "");
Error('\p{Is_Category=_Zp:=}');
Error('\P{Is_Category=_Zp:=}');
Expect(1, 8233, '\p{Is_Category: zp}', "");
Expect(0, 8233, '\p{^Is_Category: zp}', "");
Expect(0, 8233, '\P{Is_Category: zp}', "");
Expect(1, 8233, '\P{^Is_Category: zp}', "");
Expect(0, 8234, '\p{Is_Category: zp}', "");
Expect(1, 8234, '\p{^Is_Category: zp}', "");
Expect(1, 8234, '\P{Is_Category: zp}', "");
Expect(0, 8234, '\P{^Is_Category: zp}', "");
Expect(1, 8233, '\p{Is_Category= Zp}', "");
Expect(0, 8233, '\p{^Is_Category= Zp}', "");
Expect(0, 8233, '\P{Is_Category= Zp}', "");
Expect(1, 8233, '\P{^Is_Category= Zp}', "");
Expect(0, 8234, '\p{Is_Category= Zp}', "");
Expect(1, 8234, '\p{^Is_Category= Zp}', "");
Expect(1, 8234, '\P{Is_Category= Zp}', "");
Expect(0, 8234, '\P{^Is_Category= Zp}', "");
Error('\p{General_Category=	 MODIFIER_Letter:=}');
Error('\P{General_Category=	 MODIFIER_Letter:=}');
Expect(1, 94111, '\p{General_Category=modifierletter}', "");
Expect(0, 94111, '\p{^General_Category=modifierletter}', "");
Expect(0, 94111, '\P{General_Category=modifierletter}', "");
Expect(1, 94111, '\P{^General_Category=modifierletter}', "");
Expect(0, 94112, '\p{General_Category=modifierletter}', "");
Expect(1, 94112, '\p{^General_Category=modifierletter}', "");
Expect(1, 94112, '\P{General_Category=modifierletter}', "");
Expect(0, 94112, '\P{^General_Category=modifierletter}', "");
Expect(1, 94111, '\p{General_Category=- modifier_Letter}', "");
Expect(0, 94111, '\p{^General_Category=- modifier_Letter}', "");
Expect(0, 94111, '\P{General_Category=- modifier_Letter}', "");
Expect(1, 94111, '\P{^General_Category=- modifier_Letter}', "");
Expect(0, 94112, '\p{General_Category=- modifier_Letter}', "");
Expect(1, 94112, '\p{^General_Category=- modifier_Letter}', "");
Expect(1, 94112, '\P{General_Category=- modifier_Letter}', "");
Expect(0, 94112, '\P{^General_Category=- modifier_Letter}', "");
Error('\p{Gc=__Lm:=}');
Error('\P{Gc=__Lm:=}');
Expect(1, 94111, '\p{Gc=lm}', "");
Expect(0, 94111, '\p{^Gc=lm}', "");
Expect(0, 94111, '\P{Gc=lm}', "");
Expect(1, 94111, '\P{^Gc=lm}', "");
Expect(0, 94112, '\p{Gc=lm}', "");
Expect(1, 94112, '\p{^Gc=lm}', "");
Expect(1, 94112, '\P{Gc=lm}', "");
Expect(0, 94112, '\P{^Gc=lm}', "");
Expect(1, 94111, '\p{Gc:   _	lm}', "");
Expect(0, 94111, '\p{^Gc:   _	lm}', "");
Expect(0, 94111, '\P{Gc:   _	lm}', "");
Expect(1, 94111, '\P{^Gc:   _	lm}', "");
Expect(0, 94112, '\p{Gc:   _	lm}', "");
Expect(1, 94112, '\p{^Gc:   _	lm}', "");
Expect(1, 94112, '\P{Gc:   _	lm}', "");
Expect(0, 94112, '\P{^Gc:   _	lm}', "");
Error('\p{Category= modifier_Letter:=}');
Error('\P{Category= modifier_Letter:=}');
Expect(1, 94111, '\p{Category=modifierletter}', "");
Expect(0, 94111, '\p{^Category=modifierletter}', "");
Expect(0, 94111, '\P{Category=modifierletter}', "");
Expect(1, 94111, '\P{^Category=modifierletter}', "");
Expect(0, 94112, '\p{Category=modifierletter}', "");
Expect(1, 94112, '\p{^Category=modifierletter}', "");
Expect(1, 94112, '\P{Category=modifierletter}', "");
Expect(0, 94112, '\P{^Category=modifierletter}', "");
Expect(1, 94111, '\p{Category= Modifier_LETTER}', "");
Expect(0, 94111, '\p{^Category= Modifier_LETTER}', "");
Expect(0, 94111, '\P{Category= Modifier_LETTER}', "");
Expect(1, 94111, '\P{^Category= Modifier_LETTER}', "");
Expect(0, 94112, '\p{Category= Modifier_LETTER}', "");
Expect(1, 94112, '\p{^Category= Modifier_LETTER}', "");
Expect(1, 94112, '\P{Category= Modifier_LETTER}', "");
Expect(0, 94112, '\P{^Category= Modifier_LETTER}', "");
Error('\p{Is_General_Category=/a/- LM}');
Error('\P{Is_General_Category=/a/- LM}');
Expect(1, 94111, '\p{Is_General_Category=lm}', "");
Expect(0, 94111, '\p{^Is_General_Category=lm}', "");
Expect(0, 94111, '\P{Is_General_Category=lm}', "");
Expect(1, 94111, '\P{^Is_General_Category=lm}', "");
Expect(0, 94112, '\p{Is_General_Category=lm}', "");
Expect(1, 94112, '\p{^Is_General_Category=lm}', "");
Expect(1, 94112, '\P{Is_General_Category=lm}', "");
Expect(0, 94112, '\P{^Is_General_Category=lm}', "");
Expect(1, 94111, '\p{Is_General_Category=_ LM}', "");
Expect(0, 94111, '\p{^Is_General_Category=_ LM}', "");
Expect(0, 94111, '\P{Is_General_Category=_ LM}', "");
Expect(1, 94111, '\P{^Is_General_Category=_ LM}', "");
Expect(0, 94112, '\p{Is_General_Category=_ LM}', "");
Expect(1, 94112, '\p{^Is_General_Category=_ LM}', "");
Expect(1, 94112, '\P{Is_General_Category=_ LM}', "");
Expect(0, 94112, '\P{^Is_General_Category=_ LM}', "");
Error('\p{Is_Gc= :=Modifier_letter}');
Error('\P{Is_Gc= :=Modifier_letter}');
Expect(1, 94111, '\p{Is_Gc=modifierletter}', "");
Expect(0, 94111, '\p{^Is_Gc=modifierletter}', "");
Expect(0, 94111, '\P{Is_Gc=modifierletter}', "");
Expect(1, 94111, '\P{^Is_Gc=modifierletter}', "");
Expect(0, 94112, '\p{Is_Gc=modifierletter}', "");
Expect(1, 94112, '\p{^Is_Gc=modifierletter}', "");
Expect(1, 94112, '\P{Is_Gc=modifierletter}', "");
Expect(0, 94112, '\P{^Is_Gc=modifierletter}', "");
Expect(1, 94111, '\p{Is_Gc=-_modifier_Letter}', "");
Expect(0, 94111, '\p{^Is_Gc=-_modifier_Letter}', "");
Expect(0, 94111, '\P{Is_Gc=-_modifier_Letter}', "");
Expect(1, 94111, '\P{^Is_Gc=-_modifier_Letter}', "");
Expect(0, 94112, '\p{Is_Gc=-_modifier_Letter}', "");
Expect(1, 94112, '\p{^Is_Gc=-_modifier_Letter}', "");
Expect(1, 94112, '\P{Is_Gc=-_modifier_Letter}', "");
Expect(0, 94112, '\P{^Is_Gc=-_modifier_Letter}', "");
Error('\p{Is_Category=-/a/Lm}');
Error('\P{Is_Category=-/a/Lm}');
Expect(1, 94111, '\p{Is_Category=lm}', "");
Expect(0, 94111, '\p{^Is_Category=lm}', "");
Expect(0, 94111, '\P{Is_Category=lm}', "");
Expect(1, 94111, '\P{^Is_Category=lm}', "");
Expect(0, 94112, '\p{Is_Category=lm}', "");
Expect(1, 94112, '\p{^Is_Category=lm}', "");
Expect(1, 94112, '\P{Is_Category=lm}', "");
Expect(0, 94112, '\P{^Is_Category=lm}', "");
Expect(1, 94111, '\p{Is_Category=__Lm}', "");
Expect(0, 94111, '\p{^Is_Category=__Lm}', "");
Expect(0, 94111, '\P{Is_Category=__Lm}', "");
Expect(1, 94111, '\P{^Is_Category=__Lm}', "");
Expect(0, 94112, '\p{Is_Category=__Lm}', "");
Expect(1, 94112, '\p{^Is_Category=__Lm}', "");
Expect(1, 94112, '\P{Is_Category=__Lm}', "");
Expect(0, 94112, '\P{^Is_Category=__Lm}', "");
Error('\p{General_Category=/a/	Other_SYMBOL}');
Error('\P{General_Category=/a/	Other_SYMBOL}');
Expect(1, 128883, '\p{General_Category=othersymbol}', "");
Expect(0, 128883, '\p{^General_Category=othersymbol}', "");
Expect(0, 128883, '\P{General_Category=othersymbol}', "");
Expect(1, 128883, '\P{^General_Category=othersymbol}', "");
Expect(0, 128884, '\p{General_Category=othersymbol}', "");
Expect(1, 128884, '\p{^General_Category=othersymbol}', "");
Expect(1, 128884, '\P{General_Category=othersymbol}', "");
Expect(0, 128884, '\P{^General_Category=othersymbol}', "");
Expect(1, 128883, '\p{General_Category=	Other_SYMBOL}', "");
Expect(0, 128883, '\p{^General_Category=	Other_SYMBOL}', "");
Expect(0, 128883, '\P{General_Category=	Other_SYMBOL}', "");
Expect(1, 128883, '\P{^General_Category=	Other_SYMBOL}', "");
Expect(0, 128884, '\p{General_Category=	Other_SYMBOL}', "");
Expect(1, 128884, '\p{^General_Category=	Other_SYMBOL}', "");
Expect(1, 128884, '\P{General_Category=	Other_SYMBOL}', "");
Expect(0, 128884, '\P{^General_Category=	Other_SYMBOL}', "");
Error('\p{Gc=--So:=}');
Error('\P{Gc=--So:=}');
Expect(1, 128883, '\p{Gc=so}', "");
Expect(0, 128883, '\p{^Gc=so}', "");
Expect(0, 128883, '\P{Gc=so}', "");
Expect(1, 128883, '\P{^Gc=so}', "");
Expect(0, 128884, '\p{Gc=so}', "");
Expect(1, 128884, '\p{^Gc=so}', "");
Expect(1, 128884, '\P{Gc=so}', "");
Expect(0, 128884, '\P{^Gc=so}', "");
Expect(1, 128883, '\p{Gc=_SO}', "");
Expect(0, 128883, '\p{^Gc=_SO}', "");
Expect(0, 128883, '\P{Gc=_SO}', "");
Expect(1, 128883, '\P{^Gc=_SO}', "");
Expect(0, 128884, '\p{Gc=_SO}', "");
Expect(1, 128884, '\p{^Gc=_SO}', "");
Expect(1, 128884, '\P{Gc=_SO}', "");
Expect(0, 128884, '\P{^Gc=_SO}', "");
Error('\p{Category=-	Other_Symbol:=}');
Error('\P{Category=-	Other_Symbol:=}');
Expect(1, 128883, '\p{Category=othersymbol}', "");
Expect(0, 128883, '\p{^Category=othersymbol}', "");
Expect(0, 128883, '\P{Category=othersymbol}', "");
Expect(1, 128883, '\P{^Category=othersymbol}', "");
Expect(0, 128884, '\p{Category=othersymbol}', "");
Expect(1, 128884, '\p{^Category=othersymbol}', "");
Expect(1, 128884, '\P{Category=othersymbol}', "");
Expect(0, 128884, '\P{^Category=othersymbol}', "");
Expect(1, 128883, '\p{Category=--OTHER_SYMBOL}', "");
Expect(0, 128883, '\p{^Category=--OTHER_SYMBOL}', "");
Expect(0, 128883, '\P{Category=--OTHER_SYMBOL}', "");
Expect(1, 128883, '\P{^Category=--OTHER_SYMBOL}', "");
Expect(0, 128884, '\p{Category=--OTHER_SYMBOL}', "");
Expect(1, 128884, '\p{^Category=--OTHER_SYMBOL}', "");
Expect(1, 128884, '\P{Category=--OTHER_SYMBOL}', "");
Expect(0, 128884, '\P{^Category=--OTHER_SYMBOL}', "");
Error('\p{Is_General_Category=:=-SO}');
Error('\P{Is_General_Category=:=-SO}');
Expect(1, 128883, '\p{Is_General_Category=so}', "");
Expect(0, 128883, '\p{^Is_General_Category=so}', "");
Expect(0, 128883, '\P{Is_General_Category=so}', "");
Expect(1, 128883, '\P{^Is_General_Category=so}', "");
Expect(0, 128884, '\p{Is_General_Category=so}', "");
Expect(1, 128884, '\p{^Is_General_Category=so}', "");
Expect(1, 128884, '\P{Is_General_Category=so}', "");
Expect(0, 128884, '\P{^Is_General_Category=so}', "");
Expect(1, 128883, '\p{Is_General_Category=_ So}', "");
Expect(0, 128883, '\p{^Is_General_Category=_ So}', "");
Expect(0, 128883, '\P{Is_General_Category=_ So}', "");
Expect(1, 128883, '\P{^Is_General_Category=_ So}', "");
Expect(0, 128884, '\p{Is_General_Category=_ So}', "");
Expect(1, 128884, '\p{^Is_General_Category=_ So}', "");
Expect(1, 128884, '\P{Is_General_Category=_ So}', "");
Expect(0, 128884, '\P{^Is_General_Category=_ So}', "");
Error('\p{Is_Gc=_/a/Other_symbol}');
Error('\P{Is_Gc=_/a/Other_symbol}');
Expect(1, 128883, '\p{Is_Gc:   othersymbol}', "");
Expect(0, 128883, '\p{^Is_Gc:   othersymbol}', "");
Expect(0, 128883, '\P{Is_Gc:   othersymbol}', "");
Expect(1, 128883, '\P{^Is_Gc:   othersymbol}', "");
Expect(0, 128884, '\p{Is_Gc:   othersymbol}', "");
Expect(1, 128884, '\p{^Is_Gc:   othersymbol}', "");
Expect(1, 128884, '\P{Is_Gc:   othersymbol}', "");
Expect(0, 128884, '\P{^Is_Gc:   othersymbol}', "");
Expect(1, 128883, '\p{Is_Gc:	_	OTHER_symbol}', "");
Expect(0, 128883, '\p{^Is_Gc:	_	OTHER_symbol}', "");
Expect(0, 128883, '\P{Is_Gc:	_	OTHER_symbol}', "");
Expect(1, 128883, '\P{^Is_Gc:	_	OTHER_symbol}', "");
Expect(0, 128884, '\p{Is_Gc:	_	OTHER_symbol}', "");
Expect(1, 128884, '\p{^Is_Gc:	_	OTHER_symbol}', "");
Expect(1, 128884, '\P{Is_Gc:	_	OTHER_symbol}', "");
Expect(0, 128884, '\P{^Is_Gc:	_	OTHER_symbol}', "");
Error('\p{Is_Category=/a/SO}');
Error('\P{Is_Category=/a/SO}');
Expect(1, 128883, '\p{Is_Category=so}', "");
Expect(0, 128883, '\p{^Is_Category=so}', "");
Expect(0, 128883, '\P{Is_Category=so}', "");
Expect(1, 128883, '\P{^Is_Category=so}', "");
Expect(0, 128884, '\p{Is_Category=so}', "");
Expect(1, 128884, '\p{^Is_Category=so}', "");
Expect(1, 128884, '\P{Is_Category=so}', "");
Expect(0, 128884, '\P{^Is_Category=so}', "");
Expect(1, 128883, '\p{Is_Category= So}', "");
Expect(0, 128883, '\p{^Is_Category= So}', "");
Expect(0, 128883, '\P{Is_Category= So}', "");
Expect(1, 128883, '\P{^Is_Category= So}', "");
Expect(0, 128884, '\p{Is_Category= So}', "");
Expect(1, 128884, '\p{^Is_Category= So}', "");
Expect(1, 128884, '\P{Is_Category= So}', "");
Expect(0, 128884, '\P{^Is_Category= So}', "");
Error('\p{General_Category=/a/_-nonspacing_MARK}');
Error('\P{General_Category=/a/_-nonspacing_MARK}');
Expect(1, 917999, '\p{General_Category=nonspacingmark}', "");
Expect(0, 917999, '\p{^General_Category=nonspacingmark}', "");
Expect(0, 917999, '\P{General_Category=nonspacingmark}', "");
Expect(1, 917999, '\P{^General_Category=nonspacingmark}', "");
Expect(0, 918000, '\p{General_Category=nonspacingmark}', "");
Expect(1, 918000, '\p{^General_Category=nonspacingmark}', "");
Expect(1, 918000, '\P{General_Category=nonspacingmark}', "");
Expect(0, 918000, '\P{^General_Category=nonspacingmark}', "");
Expect(1, 917999, '\p{General_Category=	 Nonspacing_MARK}', "");
Expect(0, 917999, '\p{^General_Category=	 Nonspacing_MARK}', "");
Expect(0, 917999, '\P{General_Category=	 Nonspacing_MARK}', "");
Expect(1, 917999, '\P{^General_Category=	 Nonspacing_MARK}', "");
Expect(0, 918000, '\p{General_Category=	 Nonspacing_MARK}', "");
Expect(1, 918000, '\p{^General_Category=	 Nonspacing_MARK}', "");
Expect(1, 918000, '\P{General_Category=	 Nonspacing_MARK}', "");
Expect(0, 918000, '\P{^General_Category=	 Nonspacing_MARK}', "");
Error('\p{Gc= _MN:=}');
Error('\P{Gc= _MN:=}');
Expect(1, 917999, '\p{Gc=mn}', "");
Expect(0, 917999, '\p{^Gc=mn}', "");
Expect(0, 917999, '\P{Gc=mn}', "");
Expect(1, 917999, '\P{^Gc=mn}', "");
Expect(0, 918000, '\p{Gc=mn}', "");
Expect(1, 918000, '\p{^Gc=mn}', "");
Expect(1, 918000, '\P{Gc=mn}', "");
Expect(0, 918000, '\P{^Gc=mn}', "");
Expect(1, 917999, '\p{Gc=_-Mn}', "");
Expect(0, 917999, '\p{^Gc=_-Mn}', "");
Expect(0, 917999, '\P{Gc=_-Mn}', "");
Expect(1, 917999, '\P{^Gc=_-Mn}', "");
Expect(0, 918000, '\p{Gc=_-Mn}', "");
Expect(1, 918000, '\p{^Gc=_-Mn}', "");
Expect(1, 918000, '\P{Gc=_-Mn}', "");
Expect(0, 918000, '\P{^Gc=_-Mn}', "");
Error('\p{Category=:= 	Nonspacing_mark}');
Error('\P{Category=:= 	Nonspacing_mark}');
Expect(1, 917999, '\p{Category=nonspacingmark}', "");
Expect(0, 917999, '\p{^Category=nonspacingmark}', "");
Expect(0, 917999, '\P{Category=nonspacingmark}', "");
Expect(1, 917999, '\P{^Category=nonspacingmark}', "");
Expect(0, 918000, '\p{Category=nonspacingmark}', "");
Expect(1, 918000, '\p{^Category=nonspacingmark}', "");
Expect(1, 918000, '\P{Category=nonspacingmark}', "");
Expect(0, 918000, '\P{^Category=nonspacingmark}', "");
Expect(1, 917999, '\p{Category:  	NONSPACING_Mark}', "");
Expect(0, 917999, '\p{^Category:  	NONSPACING_Mark}', "");
Expect(0, 917999, '\P{Category:  	NONSPACING_Mark}', "");
Expect(1, 917999, '\P{^Category:  	NONSPACING_Mark}', "");
Expect(0, 918000, '\p{Category:  	NONSPACING_Mark}', "");
Expect(1, 918000, '\p{^Category:  	NONSPACING_Mark}', "");
Expect(1, 918000, '\P{Category:  	NONSPACING_Mark}', "");
Expect(0, 918000, '\P{^Category:  	NONSPACING_Mark}', "");
Error('\p{Is_General_Category=/a/	MN}');
Error('\P{Is_General_Category=/a/	MN}');
Expect(1, 917999, '\p{Is_General_Category=mn}', "");
Expect(0, 917999, '\p{^Is_General_Category=mn}', "");
Expect(0, 917999, '\P{Is_General_Category=mn}', "");
Expect(1, 917999, '\P{^Is_General_Category=mn}', "");
Expect(0, 918000, '\p{Is_General_Category=mn}', "");
Expect(1, 918000, '\p{^Is_General_Category=mn}', "");
Expect(1, 918000, '\P{Is_General_Category=mn}', "");
Expect(0, 918000, '\P{^Is_General_Category=mn}', "");
Expect(1, 917999, '\p{Is_General_Category:	_Mn}', "");
Expect(0, 917999, '\p{^Is_General_Category:	_Mn}', "");
Expect(0, 917999, '\P{Is_General_Category:	_Mn}', "");
Expect(1, 917999, '\P{^Is_General_Category:	_Mn}', "");
Expect(0, 918000, '\p{Is_General_Category:	_Mn}', "");
Expect(1, 918000, '\p{^Is_General_Category:	_Mn}', "");
Expect(1, 918000, '\P{Is_General_Category:	_Mn}', "");
Expect(0, 918000, '\P{^Is_General_Category:	_Mn}', "");
Error('\p{Is_Gc=-:=Nonspacing_MARK}');
Error('\P{Is_Gc=-:=Nonspacing_MARK}');
Expect(1, 917999, '\p{Is_Gc=nonspacingmark}', "");
Expect(0, 917999, '\p{^Is_Gc=nonspacingmark}', "");
Expect(0, 917999, '\P{Is_Gc=nonspacingmark}', "");
Expect(1, 917999, '\P{^Is_Gc=nonspacingmark}', "");
Expect(0, 918000, '\p{Is_Gc=nonspacingmark}', "");
Expect(1, 918000, '\p{^Is_Gc=nonspacingmark}', "");
Expect(1, 918000, '\P{Is_Gc=nonspacingmark}', "");
Expect(0, 918000, '\P{^Is_Gc=nonspacingmark}', "");
Expect(1, 917999, '\p{Is_Gc=-_NONSPACING_Mark}', "");
Expect(0, 917999, '\p{^Is_Gc=-_NONSPACING_Mark}', "");
Expect(0, 917999, '\P{Is_Gc=-_NONSPACING_Mark}', "");
Expect(1, 917999, '\P{^Is_Gc=-_NONSPACING_Mark}', "");
Expect(0, 918000, '\p{Is_Gc=-_NONSPACING_Mark}', "");
Expect(1, 918000, '\p{^Is_Gc=-_NONSPACING_Mark}', "");
Expect(1, 918000, '\P{Is_Gc=-_NONSPACING_Mark}', "");
Expect(0, 918000, '\P{^Is_Gc=-_NONSPACING_Mark}', "");
Error('\p{Is_Category=	 Mn/a/}');
Error('\P{Is_Category=	 Mn/a/}');
Expect(1, 917999, '\p{Is_Category=mn}', "");
Expect(0, 917999, '\p{^Is_Category=mn}', "");
Expect(0, 917999, '\P{Is_Category=mn}', "");
Expect(1, 917999, '\P{^Is_Category=mn}', "");
Expect(0, 918000, '\p{Is_Category=mn}', "");
Expect(1, 918000, '\p{^Is_Category=mn}', "");
Expect(1, 918000, '\P{Is_Category=mn}', "");
Expect(0, 918000, '\P{^Is_Category=mn}', "");
Expect(1, 917999, '\p{Is_Category= -Mn}', "");
Expect(0, 917999, '\p{^Is_Category= -Mn}', "");
Expect(0, 917999, '\P{Is_Category= -Mn}', "");
Expect(1, 917999, '\P{^Is_Category= -Mn}', "");
Expect(0, 918000, '\p{Is_Category= -Mn}', "");
Expect(1, 918000, '\p{^Is_Category= -Mn}', "");
Expect(1, 918000, '\P{Is_Category= -Mn}', "");
Expect(0, 918000, '\P{^Is_Category= -Mn}', "");
Error('\p{General_Category=-/a/Dash_Punctuation}');
Error('\P{General_Category=-/a/Dash_Punctuation}');
Expect(1, 65293, '\p{General_Category:dashpunctuation}', "");
Expect(0, 65293, '\p{^General_Category:dashpunctuation}', "");
Expect(0, 65293, '\P{General_Category:dashpunctuation}', "");
Expect(1, 65293, '\P{^General_Category:dashpunctuation}', "");
Expect(0, 65294, '\p{General_Category:dashpunctuation}', "");
Expect(1, 65294, '\p{^General_Category:dashpunctuation}', "");
Expect(1, 65294, '\P{General_Category:dashpunctuation}', "");
Expect(0, 65294, '\P{^General_Category:dashpunctuation}', "");
Expect(1, 65293, '\p{General_Category=--Dash_PUNCTUATION}', "");
Expect(0, 65293, '\p{^General_Category=--Dash_PUNCTUATION}', "");
Expect(0, 65293, '\P{General_Category=--Dash_PUNCTUATION}', "");
Expect(1, 65293, '\P{^General_Category=--Dash_PUNCTUATION}', "");
Expect(0, 65294, '\p{General_Category=--Dash_PUNCTUATION}', "");
Expect(1, 65294, '\p{^General_Category=--Dash_PUNCTUATION}', "");
Expect(1, 65294, '\P{General_Category=--Dash_PUNCTUATION}', "");
Expect(0, 65294, '\P{^General_Category=--Dash_PUNCTUATION}', "");
Error('\p{Gc: :=_pd}');
Error('\P{Gc: :=_pd}');
Expect(1, 65293, '\p{Gc=pd}', "");
Expect(0, 65293, '\p{^Gc=pd}', "");
Expect(0, 65293, '\P{Gc=pd}', "");
Expect(1, 65293, '\P{^Gc=pd}', "");
Expect(0, 65294, '\p{Gc=pd}', "");
Expect(1, 65294, '\p{^Gc=pd}', "");
Expect(1, 65294, '\P{Gc=pd}', "");
Expect(0, 65294, '\P{^Gc=pd}', "");
Expect(1, 65293, '\p{Gc=-	Pd}', "");
Expect(0, 65293, '\p{^Gc=-	Pd}', "");
Expect(0, 65293, '\P{Gc=-	Pd}', "");
Expect(1, 65293, '\P{^Gc=-	Pd}', "");
Expect(0, 65294, '\p{Gc=-	Pd}', "");
Expect(1, 65294, '\p{^Gc=-	Pd}', "");
Expect(1, 65294, '\P{Gc=-	Pd}', "");
Expect(0, 65294, '\P{^Gc=-	Pd}', "");
Error('\p{Category=-	Dash_PUNCTUATION:=}');
Error('\P{Category=-	Dash_PUNCTUATION:=}');
Expect(1, 65293, '\p{Category=dashpunctuation}', "");
Expect(0, 65293, '\p{^Category=dashpunctuation}', "");
Expect(0, 65293, '\P{Category=dashpunctuation}', "");
Expect(1, 65293, '\P{^Category=dashpunctuation}', "");
Expect(0, 65294, '\p{Category=dashpunctuation}', "");
Expect(1, 65294, '\p{^Category=dashpunctuation}', "");
Expect(1, 65294, '\P{Category=dashpunctuation}', "");
Expect(0, 65294, '\P{^Category=dashpunctuation}', "");
Expect(1, 65293, '\p{Category: _Dash_Punctuation}', "");
Expect(0, 65293, '\p{^Category: _Dash_Punctuation}', "");
Expect(0, 65293, '\P{Category: _Dash_Punctuation}', "");
Expect(1, 65293, '\P{^Category: _Dash_Punctuation}', "");
Expect(0, 65294, '\p{Category: _Dash_Punctuation}', "");
Expect(1, 65294, '\p{^Category: _Dash_Punctuation}', "");
Expect(1, 65294, '\P{Category: _Dash_Punctuation}', "");
Expect(0, 65294, '\P{^Category: _Dash_Punctuation}', "");
Error('\p{Is_General_Category= _Pd:=}');
Error('\P{Is_General_Category= _Pd:=}');
Expect(1, 65293, '\p{Is_General_Category=pd}', "");
Expect(0, 65293, '\p{^Is_General_Category=pd}', "");
Expect(0, 65293, '\P{Is_General_Category=pd}', "");
Expect(1, 65293, '\P{^Is_General_Category=pd}', "");
Expect(0, 65294, '\p{Is_General_Category=pd}', "");
Expect(1, 65294, '\p{^Is_General_Category=pd}', "");
Expect(1, 65294, '\P{Is_General_Category=pd}', "");
Expect(0, 65294, '\P{^Is_General_Category=pd}', "");
Expect(1, 65293, '\p{Is_General_Category=	 PD}', "");
Expect(0, 65293, '\p{^Is_General_Category=	 PD}', "");
Expect(0, 65293, '\P{Is_General_Category=	 PD}', "");
Expect(1, 65293, '\P{^Is_General_Category=	 PD}', "");
Expect(0, 65294, '\p{Is_General_Category=	 PD}', "");
Expect(1, 65294, '\p{^Is_General_Category=	 PD}', "");
Expect(1, 65294, '\P{Is_General_Category=	 PD}', "");
Expect(0, 65294, '\P{^Is_General_Category=	 PD}', "");
Error('\p{Is_Gc=dash_PUNCTUATION:=}');
Error('\P{Is_Gc=dash_PUNCTUATION:=}');
Expect(1, 65293, '\p{Is_Gc=dashpunctuation}', "");
Expect(0, 65293, '\p{^Is_Gc=dashpunctuation}', "");
Expect(0, 65293, '\P{Is_Gc=dashpunctuation}', "");
Expect(1, 65293, '\P{^Is_Gc=dashpunctuation}', "");
Expect(0, 65294, '\p{Is_Gc=dashpunctuation}', "");
Expect(1, 65294, '\p{^Is_Gc=dashpunctuation}', "");
Expect(1, 65294, '\P{Is_Gc=dashpunctuation}', "");
Expect(0, 65294, '\P{^Is_Gc=dashpunctuation}', "");
Expect(1, 65293, '\p{Is_Gc=-	dash_PUNCTUATION}', "");
Expect(0, 65293, '\p{^Is_Gc=-	dash_PUNCTUATION}', "");
Expect(0, 65293, '\P{Is_Gc=-	dash_PUNCTUATION}', "");
Expect(1, 65293, '\P{^Is_Gc=-	dash_PUNCTUATION}', "");
Expect(0, 65294, '\p{Is_Gc=-	dash_PUNCTUATION}', "");
Expect(1, 65294, '\p{^Is_Gc=-	dash_PUNCTUATION}', "");
Expect(1, 65294, '\P{Is_Gc=-	dash_PUNCTUATION}', "");
Expect(0, 65294, '\P{^Is_Gc=-	dash_PUNCTUATION}', "");
Error('\p{Is_Category=/a/_Pd}');
Error('\P{Is_Category=/a/_Pd}');
Expect(1, 65293, '\p{Is_Category=pd}', "");
Expect(0, 65293, '\p{^Is_Category=pd}', "");
Expect(0, 65293, '\P{Is_Category=pd}', "");
Expect(1, 65293, '\P{^Is_Category=pd}', "");
Expect(0, 65294, '\p{Is_Category=pd}', "");
Expect(1, 65294, '\p{^Is_Category=pd}', "");
Expect(1, 65294, '\P{Is_Category=pd}', "");
Expect(0, 65294, '\P{^Is_Category=pd}', "");
Expect(1, 65293, '\p{Is_Category=	Pd}', "");
Expect(0, 65293, '\p{^Is_Category=	Pd}', "");
Expect(0, 65293, '\P{Is_Category=	Pd}', "");
Expect(1, 65293, '\P{^Is_Category=	Pd}', "");
Expect(0, 65294, '\p{Is_Category=	Pd}', "");
Expect(1, 65294, '\p{^Is_Category=	Pd}', "");
Expect(1, 65294, '\P{Is_Category=	Pd}', "");
Expect(0, 65294, '\P{^Is_Category=	Pd}', "");
Error('\p{General_Category=/a/	-Letter_Number}');
Error('\P{General_Category=/a/	-Letter_Number}');
Expect(1, 74850, '\p{General_Category=letternumber}', "");
Expect(0, 74850, '\p{^General_Category=letternumber}', "");
Expect(0, 74850, '\P{General_Category=letternumber}', "");
Expect(1, 74850, '\P{^General_Category=letternumber}', "");
Expect(0, 74851, '\p{General_Category=letternumber}', "");
Expect(1, 74851, '\p{^General_Category=letternumber}', "");
Expect(1, 74851, '\P{General_Category=letternumber}', "");
Expect(0, 74851, '\P{^General_Category=letternumber}', "");
Expect(1, 74850, '\p{General_Category=__Letter_NUMBER}', "");
Expect(0, 74850, '\p{^General_Category=__Letter_NUMBER}', "");
Expect(0, 74850, '\P{General_Category=__Letter_NUMBER}', "");
Expect(1, 74850, '\P{^General_Category=__Letter_NUMBER}', "");
Expect(0, 74851, '\p{General_Category=__Letter_NUMBER}', "");
Expect(1, 74851, '\p{^General_Category=__Letter_NUMBER}', "");
Expect(1, 74851, '\P{General_Category=__Letter_NUMBER}', "");
Expect(0, 74851, '\P{^General_Category=__Letter_NUMBER}', "");
Error('\p{Gc:   /a/ Nl}');
Error('\P{Gc:   /a/ Nl}');
Expect(1, 74850, '\p{Gc:nl}', "");
Expect(0, 74850, '\p{^Gc:nl}', "");
Expect(0, 74850, '\P{Gc:nl}', "");
Expect(1, 74850, '\P{^Gc:nl}', "");
Expect(0, 74851, '\p{Gc:nl}', "");
Expect(1, 74851, '\p{^Gc:nl}', "");
Expect(1, 74851, '\P{Gc:nl}', "");
Expect(0, 74851, '\P{^Gc:nl}', "");
Expect(1, 74850, '\p{Gc:_-Nl}', "");
Expect(0, 74850, '\p{^Gc:_-Nl}', "");
Expect(0, 74850, '\P{Gc:_-Nl}', "");
Expect(1, 74850, '\P{^Gc:_-Nl}', "");
Expect(0, 74851, '\p{Gc:_-Nl}', "");
Expect(1, 74851, '\p{^Gc:_-Nl}', "");
Expect(1, 74851, '\P{Gc:_-Nl}', "");
Expect(0, 74851, '\P{^Gc:_-Nl}', "");
Error('\p{Category=letter_Number/a/}');
Error('\P{Category=letter_Number/a/}');
Expect(1, 74850, '\p{Category=letternumber}', "");
Expect(0, 74850, '\p{^Category=letternumber}', "");
Expect(0, 74850, '\P{Category=letternumber}', "");
Expect(1, 74850, '\P{^Category=letternumber}', "");
Expect(0, 74851, '\p{Category=letternumber}', "");
Expect(1, 74851, '\p{^Category=letternumber}', "");
Expect(1, 74851, '\P{Category=letternumber}', "");
Expect(0, 74851, '\P{^Category=letternumber}', "");
Expect(1, 74850, '\p{Category=	_letter_Number}', "");
Expect(0, 74850, '\p{^Category=	_letter_Number}', "");
Expect(0, 74850, '\P{Category=	_letter_Number}', "");
Expect(1, 74850, '\P{^Category=	_letter_Number}', "");
Expect(0, 74851, '\p{Category=	_letter_Number}', "");
Expect(1, 74851, '\p{^Category=	_letter_Number}', "");
Expect(1, 74851, '\P{Category=	_letter_Number}', "");
Expect(0, 74851, '\P{^Category=	_letter_Number}', "");
Error('\p{Is_General_Category=--nl/a/}');
Error('\P{Is_General_Category=--nl/a/}');
Expect(1, 74850, '\p{Is_General_Category=nl}', "");
Expect(0, 74850, '\p{^Is_General_Category=nl}', "");
Expect(0, 74850, '\P{Is_General_Category=nl}', "");
Expect(1, 74850, '\P{^Is_General_Category=nl}', "");
Expect(0, 74851, '\p{Is_General_Category=nl}', "");
Expect(1, 74851, '\p{^Is_General_Category=nl}', "");
Expect(1, 74851, '\P{Is_General_Category=nl}', "");
Expect(0, 74851, '\P{^Is_General_Category=nl}', "");
Expect(1, 74850, '\p{Is_General_Category= NL}', "");
Expect(0, 74850, '\p{^Is_General_Category= NL}', "");
Expect(0, 74850, '\P{Is_General_Category= NL}', "");
Expect(1, 74850, '\P{^Is_General_Category= NL}', "");
Expect(0, 74851, '\p{Is_General_Category= NL}', "");
Expect(1, 74851, '\p{^Is_General_Category= NL}', "");
Expect(1, 74851, '\P{Is_General_Category= NL}', "");
Expect(0, 74851, '\P{^Is_General_Category= NL}', "");
Error('\p{Is_Gc=-:=LETTER_number}');
Error('\P{Is_Gc=-:=LETTER_number}');
Expect(1, 74850, '\p{Is_Gc=letternumber}', "");
Expect(0, 74850, '\p{^Is_Gc=letternumber}', "");
Expect(0, 74850, '\P{Is_Gc=letternumber}', "");
Expect(1, 74850, '\P{^Is_Gc=letternumber}', "");
Expect(0, 74851, '\p{Is_Gc=letternumber}', "");
Expect(1, 74851, '\p{^Is_Gc=letternumber}', "");
Expect(1, 74851, '\P{Is_Gc=letternumber}', "");
Expect(0, 74851, '\P{^Is_Gc=letternumber}', "");
Expect(1, 74850, '\p{Is_Gc=-letter_number}', "");
Expect(0, 74850, '\p{^Is_Gc=-letter_number}', "");
Expect(0, 74850, '\P{Is_Gc=-letter_number}', "");
Expect(1, 74850, '\P{^Is_Gc=-letter_number}', "");
Expect(0, 74851, '\p{Is_Gc=-letter_number}', "");
Expect(1, 74851, '\p{^Is_Gc=-letter_number}', "");
Expect(1, 74851, '\P{Is_Gc=-letter_number}', "");
Expect(0, 74851, '\P{^Is_Gc=-letter_number}', "");
Error('\p{Is_Category= /a/NL}');
Error('\P{Is_Category= /a/NL}');
Expect(1, 74850, '\p{Is_Category=nl}', "");
Expect(0, 74850, '\p{^Is_Category=nl}', "");
Expect(0, 74850, '\P{Is_Category=nl}', "");
Expect(1, 74850, '\P{^Is_Category=nl}', "");
Expect(0, 74851, '\p{Is_Category=nl}', "");
Expect(1, 74851, '\p{^Is_Category=nl}', "");
Expect(1, 74851, '\P{Is_Category=nl}', "");
Expect(0, 74851, '\P{^Is_Category=nl}', "");
Expect(1, 74850, '\p{Is_Category=_	Nl}', "");
Expect(0, 74850, '\p{^Is_Category=_	Nl}', "");
Expect(0, 74850, '\P{Is_Category=_	Nl}', "");
Expect(1, 74850, '\P{^Is_Category=_	Nl}', "");
Expect(0, 74851, '\p{Is_Category=_	Nl}', "");
Expect(1, 74851, '\p{^Is_Category=_	Nl}', "");
Expect(1, 74851, '\P{Is_Category=_	Nl}', "");
Expect(0, 74851, '\P{^Is_Category=_	Nl}', "");
Error('\p{General_Category=	initial_punctuation/a/}');
Error('\P{General_Category=	initial_punctuation/a/}');
Expect(1, 11808, '\p{General_Category=initialpunctuation}', "");
Expect(0, 11808, '\p{^General_Category=initialpunctuation}', "");
Expect(0, 11808, '\P{General_Category=initialpunctuation}', "");
Expect(1, 11808, '\P{^General_Category=initialpunctuation}', "");
Expect(0, 11809, '\p{General_Category=initialpunctuation}', "");
Expect(1, 11809, '\p{^General_Category=initialpunctuation}', "");
Expect(1, 11809, '\P{General_Category=initialpunctuation}', "");
Expect(0, 11809, '\P{^General_Category=initialpunctuation}', "");
Expect(1, 11808, '\p{General_Category= _Initial_punctuation}', "");
Expect(0, 11808, '\p{^General_Category= _Initial_punctuation}', "");
Expect(0, 11808, '\P{General_Category= _Initial_punctuation}', "");
Expect(1, 11808, '\P{^General_Category= _Initial_punctuation}', "");
Expect(0, 11809, '\p{General_Category= _Initial_punctuation}', "");
Expect(1, 11809, '\p{^General_Category= _Initial_punctuation}', "");
Expect(1, 11809, '\P{General_Category= _Initial_punctuation}', "");
Expect(0, 11809, '\P{^General_Category= _Initial_punctuation}', "");
Error('\p{Gc=		PI/a/}');
Error('\P{Gc=		PI/a/}');
Expect(1, 11808, '\p{Gc=pi}', "");
Expect(0, 11808, '\p{^Gc=pi}', "");
Expect(0, 11808, '\P{Gc=pi}', "");
Expect(1, 11808, '\P{^Gc=pi}', "");
Expect(0, 11809, '\p{Gc=pi}', "");
Expect(1, 11809, '\p{^Gc=pi}', "");
Expect(1, 11809, '\P{Gc=pi}', "");
Expect(0, 11809, '\P{^Gc=pi}', "");
Expect(1, 11808, '\p{Gc=	Pi}', "");
Expect(0, 11808, '\p{^Gc=	Pi}', "");
Expect(0, 11808, '\P{Gc=	Pi}', "");
Expect(1, 11808, '\P{^Gc=	Pi}', "");
Expect(0, 11809, '\p{Gc=	Pi}', "");
Expect(1, 11809, '\p{^Gc=	Pi}', "");
Expect(1, 11809, '\P{Gc=	Pi}', "");
Expect(0, 11809, '\P{^Gc=	Pi}', "");
Error('\p{Category=_/a/Initial_Punctuation}');
Error('\P{Category=_/a/Initial_Punctuation}');
Expect(1, 11808, '\p{Category=initialpunctuation}', "");
Expect(0, 11808, '\p{^Category=initialpunctuation}', "");
Expect(0, 11808, '\P{Category=initialpunctuation}', "");
Expect(1, 11808, '\P{^Category=initialpunctuation}', "");
Expect(0, 11809, '\p{Category=initialpunctuation}', "");
Expect(1, 11809, '\p{^Category=initialpunctuation}', "");
Expect(1, 11809, '\P{Category=initialpunctuation}', "");
Expect(0, 11809, '\P{^Category=initialpunctuation}', "");
Expect(1, 11808, '\p{Category=	initial_PUNCTUATION}', "");
Expect(0, 11808, '\p{^Category=	initial_PUNCTUATION}', "");
Expect(0, 11808, '\P{Category=	initial_PUNCTUATION}', "");
Expect(1, 11808, '\P{^Category=	initial_PUNCTUATION}', "");
Expect(0, 11809, '\p{Category=	initial_PUNCTUATION}', "");
Expect(1, 11809, '\p{^Category=	initial_PUNCTUATION}', "");
Expect(1, 11809, '\P{Category=	initial_PUNCTUATION}', "");
Expect(0, 11809, '\P{^Category=	initial_PUNCTUATION}', "");
Error('\p{Is_General_Category=-_pi:=}');
Error('\P{Is_General_Category=-_pi:=}');
Expect(1, 11808, '\p{Is_General_Category=pi}', "");
Expect(0, 11808, '\p{^Is_General_Category=pi}', "");
Expect(0, 11808, '\P{Is_General_Category=pi}', "");
Expect(1, 11808, '\P{^Is_General_Category=pi}', "");
Expect(0, 11809, '\p{Is_General_Category=pi}', "");
Expect(1, 11809, '\p{^Is_General_Category=pi}', "");
Expect(1, 11809, '\P{Is_General_Category=pi}', "");
Expect(0, 11809, '\P{^Is_General_Category=pi}', "");
Expect(1, 11808, '\p{Is_General_Category=__Pi}', "");
Expect(0, 11808, '\p{^Is_General_Category=__Pi}', "");
Expect(0, 11808, '\P{Is_General_Category=__Pi}', "");
Expect(1, 11808, '\P{^Is_General_Category=__Pi}', "");
Expect(0, 11809, '\p{Is_General_Category=__Pi}', "");
Expect(1, 11809, '\p{^Is_General_Category=__Pi}', "");
Expect(1, 11809, '\P{Is_General_Category=__Pi}', "");
Expect(0, 11809, '\P{^Is_General_Category=__Pi}', "");
Error('\p{Is_Gc=/a/initial_PUNCTUATION}');
Error('\P{Is_Gc=/a/initial_PUNCTUATION}');
Expect(1, 11808, '\p{Is_Gc=initialpunctuation}', "");
Expect(0, 11808, '\p{^Is_Gc=initialpunctuation}', "");
Expect(0, 11808, '\P{Is_Gc=initialpunctuation}', "");
Expect(1, 11808, '\P{^Is_Gc=initialpunctuation}', "");
Expect(0, 11809, '\p{Is_Gc=initialpunctuation}', "");
Expect(1, 11809, '\p{^Is_Gc=initialpunctuation}', "");
Expect(1, 11809, '\P{Is_Gc=initialpunctuation}', "");
Expect(0, 11809, '\P{^Is_Gc=initialpunctuation}', "");
Expect(1, 11808, '\p{Is_Gc= Initial_PUNCTUATION}', "");
Expect(0, 11808, '\p{^Is_Gc= Initial_PUNCTUATION}', "");
Expect(0, 11808, '\P{Is_Gc= Initial_PUNCTUATION}', "");
Expect(1, 11808, '\P{^Is_Gc= Initial_PUNCTUATION}', "");
Expect(0, 11809, '\p{Is_Gc= Initial_PUNCTUATION}', "");
Expect(1, 11809, '\p{^Is_Gc= Initial_PUNCTUATION}', "");
Expect(1, 11809, '\P{Is_Gc= Initial_PUNCTUATION}', "");
Expect(0, 11809, '\P{^Is_Gc= Initial_PUNCTUATION}', "");
Error('\p{Is_Category=_:=pi}');
Error('\P{Is_Category=_:=pi}');
Expect(1, 11808, '\p{Is_Category=pi}', "");
Expect(0, 11808, '\p{^Is_Category=pi}', "");
Expect(0, 11808, '\P{Is_Category=pi}', "");
Expect(1, 11808, '\P{^Is_Category=pi}', "");
Expect(0, 11809, '\p{Is_Category=pi}', "");
Expect(1, 11809, '\p{^Is_Category=pi}', "");
Expect(1, 11809, '\P{Is_Category=pi}', "");
Expect(0, 11809, '\P{^Is_Category=pi}', "");
Expect(1, 11808, '\p{Is_Category=		pi}', "");
Expect(0, 11808, '\p{^Is_Category=		pi}', "");
Expect(0, 11808, '\P{Is_Category=		pi}', "");
Expect(1, 11808, '\P{^Is_Category=		pi}', "");
Expect(0, 11809, '\p{Is_Category=		pi}', "");
Expect(1, 11809, '\p{^Is_Category=		pi}', "");
Expect(1, 11809, '\P{Is_Category=		pi}', "");
Expect(0, 11809, '\P{^Is_Category=		pi}', "");
Error('\p{General_Category= Currency_SYMBOL:=}');
Error('\P{General_Category= Currency_SYMBOL:=}');
Expect(1, 65510, '\p{General_Category: currencysymbol}', "");
Expect(0, 65510, '\p{^General_Category: currencysymbol}', "");
Expect(0, 65510, '\P{General_Category: currencysymbol}', "");
Expect(1, 65510, '\P{^General_Category: currencysymbol}', "");
Expect(0, 65511, '\p{General_Category: currencysymbol}', "");
Expect(1, 65511, '\p{^General_Category: currencysymbol}', "");
Expect(1, 65511, '\P{General_Category: currencysymbol}', "");
Expect(0, 65511, '\P{^General_Category: currencysymbol}', "");
Expect(1, 65510, '\p{General_Category= 	Currency_Symbol}', "");
Expect(0, 65510, '\p{^General_Category= 	Currency_Symbol}', "");
Expect(0, 65510, '\P{General_Category= 	Currency_Symbol}', "");
Expect(1, 65510, '\P{^General_Category= 	Currency_Symbol}', "");
Expect(0, 65511, '\p{General_Category= 	Currency_Symbol}', "");
Expect(1, 65511, '\p{^General_Category= 	Currency_Symbol}', "");
Expect(1, 65511, '\P{General_Category= 	Currency_Symbol}', "");
Expect(0, 65511, '\P{^General_Category= 	Currency_Symbol}', "");
Error('\p{Gc=-Sc/a/}');
Error('\P{Gc=-Sc/a/}');
Expect(1, 65510, '\p{Gc=sc}', "");
Expect(0, 65510, '\p{^Gc=sc}', "");
Expect(0, 65510, '\P{Gc=sc}', "");
Expect(1, 65510, '\P{^Gc=sc}', "");
Expect(0, 65511, '\p{Gc=sc}', "");
Expect(1, 65511, '\p{^Gc=sc}', "");
Expect(1, 65511, '\P{Gc=sc}', "");
Expect(0, 65511, '\P{^Gc=sc}', "");
Expect(1, 65510, '\p{Gc=_ sc}', "");
Expect(0, 65510, '\p{^Gc=_ sc}', "");
Expect(0, 65510, '\P{Gc=_ sc}', "");
Expect(1, 65510, '\P{^Gc=_ sc}', "");
Expect(0, 65511, '\p{Gc=_ sc}', "");
Expect(1, 65511, '\p{^Gc=_ sc}', "");
Expect(1, 65511, '\P{Gc=_ sc}', "");
Expect(0, 65511, '\P{^Gc=_ sc}', "");
Error('\p{Category=/a/Currency_symbol}');
Error('\P{Category=/a/Currency_symbol}');
Expect(1, 65510, '\p{Category=currencysymbol}', "");
Expect(0, 65510, '\p{^Category=currencysymbol}', "");
Expect(0, 65510, '\P{Category=currencysymbol}', "");
Expect(1, 65510, '\P{^Category=currencysymbol}', "");
Expect(0, 65511, '\p{Category=currencysymbol}', "");
Expect(1, 65511, '\p{^Category=currencysymbol}', "");
Expect(1, 65511, '\P{Category=currencysymbol}', "");
Expect(0, 65511, '\P{^Category=currencysymbol}', "");
Expect(1, 65510, '\p{Category: -currency_Symbol}', "");
Expect(0, 65510, '\p{^Category: -currency_Symbol}', "");
Expect(0, 65510, '\P{Category: -currency_Symbol}', "");
Expect(1, 65510, '\P{^Category: -currency_Symbol}', "");
Expect(0, 65511, '\p{Category: -currency_Symbol}', "");
Expect(1, 65511, '\p{^Category: -currency_Symbol}', "");
Expect(1, 65511, '\P{Category: -currency_Symbol}', "");
Expect(0, 65511, '\P{^Category: -currency_Symbol}', "");
Error('\p{Is_General_Category=	Sc/a/}');
Error('\P{Is_General_Category=	Sc/a/}');
Expect(1, 65510, '\p{Is_General_Category=sc}', "");
Expect(0, 65510, '\p{^Is_General_Category=sc}', "");
Expect(0, 65510, '\P{Is_General_Category=sc}', "");
Expect(1, 65510, '\P{^Is_General_Category=sc}', "");
Expect(0, 65511, '\p{Is_General_Category=sc}', "");
Expect(1, 65511, '\p{^Is_General_Category=sc}', "");
Expect(1, 65511, '\P{Is_General_Category=sc}', "");
Expect(0, 65511, '\P{^Is_General_Category=sc}', "");
Expect(1, 65510, '\p{Is_General_Category:	_SC}', "");
Expect(0, 65510, '\p{^Is_General_Category:	_SC}', "");
Expect(0, 65510, '\P{Is_General_Category:	_SC}', "");
Expect(1, 65510, '\P{^Is_General_Category:	_SC}', "");
Expect(0, 65511, '\p{Is_General_Category:	_SC}', "");
Expect(1, 65511, '\p{^Is_General_Category:	_SC}', "");
Expect(1, 65511, '\P{Is_General_Category:	_SC}', "");
Expect(0, 65511, '\P{^Is_General_Category:	_SC}', "");
Error('\p{Is_Gc: _:=currency_Symbol}');
Error('\P{Is_Gc: _:=currency_Symbol}');
Expect(1, 65510, '\p{Is_Gc:currencysymbol}', "");
Expect(0, 65510, '\p{^Is_Gc:currencysymbol}', "");
Expect(0, 65510, '\P{Is_Gc:currencysymbol}', "");
Expect(1, 65510, '\P{^Is_Gc:currencysymbol}', "");
Expect(0, 65511, '\p{Is_Gc:currencysymbol}', "");
Expect(1, 65511, '\p{^Is_Gc:currencysymbol}', "");
Expect(1, 65511, '\P{Is_Gc:currencysymbol}', "");
Expect(0, 65511, '\P{^Is_Gc:currencysymbol}', "");
Expect(1, 65510, '\p{Is_Gc= CURRENCY_Symbol}', "");
Expect(0, 65510, '\p{^Is_Gc= CURRENCY_Symbol}', "");
Expect(0, 65510, '\P{Is_Gc= CURRENCY_Symbol}', "");
Expect(1, 65510, '\P{^Is_Gc= CURRENCY_Symbol}', "");
Expect(0, 65511, '\p{Is_Gc= CURRENCY_Symbol}', "");
Expect(1, 65511, '\p{^Is_Gc= CURRENCY_Symbol}', "");
Expect(1, 65511, '\P{Is_Gc= CURRENCY_Symbol}', "");
Expect(0, 65511, '\P{^Is_Gc= CURRENCY_Symbol}', "");
Error('\p{Is_Category= -SC/a/}');
Error('\P{Is_Category= -SC/a/}');
Expect(1, 65510, '\p{Is_Category=sc}', "");
Expect(0, 65510, '\p{^Is_Category=sc}', "");
Expect(0, 65510, '\P{Is_Category=sc}', "");
Expect(1, 65510, '\P{^Is_Category=sc}', "");
Expect(0, 65511, '\p{Is_Category=sc}', "");
Expect(1, 65511, '\p{^Is_Category=sc}', "");
Expect(1, 65511, '\P{Is_Category=sc}', "");
Expect(0, 65511, '\P{^Is_Category=sc}', "");
Expect(1, 65510, '\p{Is_Category=_	sc}', "");
Expect(0, 65510, '\p{^Is_Category=_	sc}', "");
Expect(0, 65510, '\P{Is_Category=_	sc}', "");
Expect(1, 65510, '\P{^Is_Category=_	sc}', "");
Expect(0, 65511, '\p{Is_Category=_	sc}', "");
Expect(1, 65511, '\p{^Is_Category=_	sc}', "");
Expect(1, 65511, '\P{Is_Category=_	sc}', "");
Expect(0, 65511, '\P{^Is_Category=_	sc}', "");
Error('\p{General_Category=:=		Format}');
Error('\P{General_Category=:=		Format}');
Expect(1, 917631, '\p{General_Category:format}', "");
Expect(0, 917631, '\p{^General_Category:format}', "");
Expect(0, 917631, '\P{General_Category:format}', "");
Expect(1, 917631, '\P{^General_Category:format}', "");
Expect(0, 917632, '\p{General_Category:format}', "");
Expect(1, 917632, '\p{^General_Category:format}', "");
Expect(1, 917632, '\P{General_Category:format}', "");
Expect(0, 917632, '\P{^General_Category:format}', "");
Expect(1, 917631, '\p{General_Category=- format}', "");
Expect(0, 917631, '\p{^General_Category=- format}', "");
Expect(0, 917631, '\P{General_Category=- format}', "");
Expect(1, 917631, '\P{^General_Category=- format}', "");
Expect(0, 917632, '\p{General_Category=- format}', "");
Expect(1, 917632, '\p{^General_Category=- format}', "");
Expect(1, 917632, '\P{General_Category=- format}', "");
Expect(0, 917632, '\P{^General_Category=- format}', "");
Error('\p{Gc=/a/Cf}');
Error('\P{Gc=/a/Cf}');
Expect(1, 917631, '\p{Gc=cf}', "");
Expect(0, 917631, '\p{^Gc=cf}', "");
Expect(0, 917631, '\P{Gc=cf}', "");
Expect(1, 917631, '\P{^Gc=cf}', "");
Expect(0, 917632, '\p{Gc=cf}', "");
Expect(1, 917632, '\p{^Gc=cf}', "");
Expect(1, 917632, '\P{Gc=cf}', "");
Expect(0, 917632, '\P{^Gc=cf}', "");
Expect(1, 917631, '\p{Gc=_	cf}', "");
Expect(0, 917631, '\p{^Gc=_	cf}', "");
Expect(0, 917631, '\P{Gc=_	cf}', "");
Expect(1, 917631, '\P{^Gc=_	cf}', "");
Expect(0, 917632, '\p{Gc=_	cf}', "");
Expect(1, 917632, '\p{^Gc=_	cf}', "");
Expect(1, 917632, '\P{Gc=_	cf}', "");
Expect(0, 917632, '\P{^Gc=_	cf}', "");
Error('\p{Category::=-FORMAT}');
Error('\P{Category::=-FORMAT}');
Expect(1, 917631, '\p{Category=format}', "");
Expect(0, 917631, '\p{^Category=format}', "");
Expect(0, 917631, '\P{Category=format}', "");
Expect(1, 917631, '\P{^Category=format}', "");
Expect(0, 917632, '\p{Category=format}', "");
Expect(1, 917632, '\p{^Category=format}', "");
Expect(1, 917632, '\P{Category=format}', "");
Expect(0, 917632, '\P{^Category=format}', "");
Expect(1, 917631, '\p{Category= -FORMAT}', "");
Expect(0, 917631, '\p{^Category= -FORMAT}', "");
Expect(0, 917631, '\P{Category= -FORMAT}', "");
Expect(1, 917631, '\P{^Category= -FORMAT}', "");
Expect(0, 917632, '\p{Category= -FORMAT}', "");
Expect(1, 917632, '\p{^Category= -FORMAT}', "");
Expect(1, 917632, '\P{Category= -FORMAT}', "");
Expect(0, 917632, '\P{^Category= -FORMAT}', "");
Error('\p{Is_General_Category=-Cf/a/}');
Error('\P{Is_General_Category=-Cf/a/}');
Expect(1, 917631, '\p{Is_General_Category=cf}', "");
Expect(0, 917631, '\p{^Is_General_Category=cf}', "");
Expect(0, 917631, '\P{Is_General_Category=cf}', "");
Expect(1, 917631, '\P{^Is_General_Category=cf}', "");
Expect(0, 917632, '\p{Is_General_Category=cf}', "");
Expect(1, 917632, '\p{^Is_General_Category=cf}', "");
Expect(1, 917632, '\P{Is_General_Category=cf}', "");
Expect(0, 917632, '\P{^Is_General_Category=cf}', "");
Expect(1, 917631, '\p{Is_General_Category:	_Cf}', "");
Expect(0, 917631, '\p{^Is_General_Category:	_Cf}', "");
Expect(0, 917631, '\P{Is_General_Category:	_Cf}', "");
Expect(1, 917631, '\P{^Is_General_Category:	_Cf}', "");
Expect(0, 917632, '\p{Is_General_Category:	_Cf}', "");
Expect(1, 917632, '\p{^Is_General_Category:	_Cf}', "");
Expect(1, 917632, '\P{Is_General_Category:	_Cf}', "");
Expect(0, 917632, '\P{^Is_General_Category:	_Cf}', "");
Error('\p{Is_Gc=/a/-FORMAT}');
Error('\P{Is_Gc=/a/-FORMAT}');
Expect(1, 917631, '\p{Is_Gc=format}', "");
Expect(0, 917631, '\p{^Is_Gc=format}', "");
Expect(0, 917631, '\P{Is_Gc=format}', "");
Expect(1, 917631, '\P{^Is_Gc=format}', "");
Expect(0, 917632, '\p{Is_Gc=format}', "");
Expect(1, 917632, '\p{^Is_Gc=format}', "");
Expect(1, 917632, '\P{Is_Gc=format}', "");
Expect(0, 917632, '\P{^Is_Gc=format}', "");
Expect(1, 917631, '\p{Is_Gc= -Format}', "");
Expect(0, 917631, '\p{^Is_Gc= -Format}', "");
Expect(0, 917631, '\P{Is_Gc= -Format}', "");
Expect(1, 917631, '\P{^Is_Gc= -Format}', "");
Expect(0, 917632, '\p{Is_Gc= -Format}', "");
Expect(1, 917632, '\p{^Is_Gc= -Format}', "");
Expect(1, 917632, '\P{Is_Gc= -Format}', "");
Expect(0, 917632, '\P{^Is_Gc= -Format}', "");
Error('\p{Is_Category=-_Cf:=}');
Error('\P{Is_Category=-_Cf:=}');
Expect(1, 917631, '\p{Is_Category=cf}', "");
Expect(0, 917631, '\p{^Is_Category=cf}', "");
Expect(0, 917631, '\P{Is_Category=cf}', "");
Expect(1, 917631, '\P{^Is_Category=cf}', "");
Expect(0, 917632, '\p{Is_Category=cf}', "");
Expect(1, 917632, '\p{^Is_Category=cf}', "");
Expect(1, 917632, '\P{Is_Category=cf}', "");
Expect(0, 917632, '\P{^Is_Category=cf}', "");
Expect(1, 917631, '\p{Is_Category= -cf}', "");
Expect(0, 917631, '\p{^Is_Category= -cf}', "");
Expect(0, 917631, '\P{Is_Category= -cf}', "");
Expect(1, 917631, '\P{^Is_Category= -cf}', "");
Expect(0, 917632, '\p{Is_Category= -cf}', "");
Expect(1, 917632, '\p{^Is_Category= -cf}', "");
Expect(1, 917632, '\P{Is_Category= -cf}', "");
Expect(0, 917632, '\P{^Is_Category= -cf}', "");
Error('\p{IDS_Trinary_Operator=/a/Yes}');
Error('\P{IDS_Trinary_Operator=/a/Yes}');
Expect(1, 12275, '\p{IDS_Trinary_Operator:	yes}', "");
Expect(0, 12275, '\p{^IDS_Trinary_Operator:	yes}', "");
Expect(0, 12275, '\P{IDS_Trinary_Operator:	yes}', "");
Expect(1, 12275, '\P{^IDS_Trinary_Operator:	yes}', "");
Expect(0, 12276, '\p{IDS_Trinary_Operator:	yes}', "");
Expect(1, 12276, '\p{^IDS_Trinary_Operator:	yes}', "");
Expect(1, 12276, '\P{IDS_Trinary_Operator:	yes}', "");
Expect(0, 12276, '\P{^IDS_Trinary_Operator:	yes}', "");
Expect(1, 12275, '\p{IDS_Trinary_Operator=	YES}', "");
Expect(0, 12275, '\p{^IDS_Trinary_Operator=	YES}', "");
Expect(0, 12275, '\P{IDS_Trinary_Operator=	YES}', "");
Expect(1, 12275, '\P{^IDS_Trinary_Operator=	YES}', "");
Expect(0, 12276, '\p{IDS_Trinary_Operator=	YES}', "");
Expect(1, 12276, '\p{^IDS_Trinary_Operator=	YES}', "");
Expect(1, 12276, '\P{IDS_Trinary_Operator=	YES}', "");
Expect(0, 12276, '\P{^IDS_Trinary_Operator=	YES}', "");
Error('\p{IDST=/a/_	y}');
Error('\P{IDST=/a/_	y}');
Expect(1, 12275, '\p{IDST=y}', "");
Expect(0, 12275, '\p{^IDST=y}', "");
Expect(0, 12275, '\P{IDST=y}', "");
Expect(1, 12275, '\P{^IDST=y}', "");
Expect(0, 12276, '\p{IDST=y}', "");
Expect(1, 12276, '\p{^IDST=y}', "");
Expect(1, 12276, '\P{IDST=y}', "");
Expect(0, 12276, '\P{^IDST=y}', "");
Expect(1, 12275, '\p{IDST=	Y}', "");
Expect(0, 12275, '\p{^IDST=	Y}', "");
Expect(0, 12275, '\P{IDST=	Y}', "");
Expect(1, 12275, '\P{^IDST=	Y}', "");
Expect(0, 12276, '\p{IDST=	Y}', "");
Expect(1, 12276, '\p{^IDST=	Y}', "");
Expect(1, 12276, '\P{IDST=	Y}', "");
Expect(0, 12276, '\P{^IDST=	Y}', "");
Error('\p{Is_IDS_Trinary_Operator=	:=T}');
Error('\P{Is_IDS_Trinary_Operator=	:=T}');
Expect(1, 12275, '\p{Is_IDS_Trinary_Operator=t}', "");
Expect(0, 12275, '\p{^Is_IDS_Trinary_Operator=t}', "");
Expect(0, 12275, '\P{Is_IDS_Trinary_Operator=t}', "");
Expect(1, 12275, '\P{^Is_IDS_Trinary_Operator=t}', "");
Expect(0, 12276, '\p{Is_IDS_Trinary_Operator=t}', "");
Expect(1, 12276, '\p{^Is_IDS_Trinary_Operator=t}', "");
Expect(1, 12276, '\P{Is_IDS_Trinary_Operator=t}', "");
Expect(0, 12276, '\P{^Is_IDS_Trinary_Operator=t}', "");
Expect(1, 12275, '\p{Is_IDS_Trinary_Operator= T}', "");
Expect(0, 12275, '\p{^Is_IDS_Trinary_Operator= T}', "");
Expect(0, 12275, '\P{Is_IDS_Trinary_Operator= T}', "");
Expect(1, 12275, '\P{^Is_IDS_Trinary_Operator= T}', "");
Expect(0, 12276, '\p{Is_IDS_Trinary_Operator= T}', "");
Expect(1, 12276, '\p{^Is_IDS_Trinary_Operator= T}', "");
Expect(1, 12276, '\P{Is_IDS_Trinary_Operator= T}', "");
Expect(0, 12276, '\P{^Is_IDS_Trinary_Operator= T}', "");
Error('\p{Is_IDST=/a/ 	TRUE}');
Error('\P{Is_IDST=/a/ 	TRUE}');
Expect(1, 12275, '\p{Is_IDST=true}', "");
Expect(0, 12275, '\p{^Is_IDST=true}', "");
Expect(0, 12275, '\P{Is_IDST=true}', "");
Expect(1, 12275, '\P{^Is_IDST=true}', "");
Expect(0, 12276, '\p{Is_IDST=true}', "");
Expect(1, 12276, '\p{^Is_IDST=true}', "");
Expect(1, 12276, '\P{Is_IDST=true}', "");
Expect(0, 12276, '\P{^Is_IDST=true}', "");
Expect(1, 12275, '\p{Is_IDST= -True}', "");
Expect(0, 12275, '\p{^Is_IDST= -True}', "");
Expect(0, 12275, '\P{Is_IDST= -True}', "");
Expect(1, 12275, '\P{^Is_IDST= -True}', "");
Expect(0, 12276, '\p{Is_IDST= -True}', "");
Expect(1, 12276, '\p{^Is_IDST= -True}', "");
Expect(1, 12276, '\P{Is_IDST= -True}', "");
Expect(0, 12276, '\P{^Is_IDST= -True}', "");
Error('\p{IDS_Trinary_Operator=:=-NO}');
Error('\P{IDS_Trinary_Operator=:=-NO}');
Expect(1, 12276, '\p{IDS_Trinary_Operator=no}', "");
Expect(0, 12276, '\p{^IDS_Trinary_Operator=no}', "");
Expect(0, 12276, '\P{IDS_Trinary_Operator=no}', "");
Expect(1, 12276, '\P{^IDS_Trinary_Operator=no}', "");
Expect(0, 12275, '\p{IDS_Trinary_Operator=no}', "");
Expect(1, 12275, '\p{^IDS_Trinary_Operator=no}', "");
Expect(1, 12275, '\P{IDS_Trinary_Operator=no}', "");
Expect(0, 12275, '\P{^IDS_Trinary_Operator=no}', "");
Expect(1, 12276, '\p{IDS_Trinary_Operator=		NO}', "");
Expect(0, 12276, '\p{^IDS_Trinary_Operator=		NO}', "");
Expect(0, 12276, '\P{IDS_Trinary_Operator=		NO}', "");
Expect(1, 12276, '\P{^IDS_Trinary_Operator=		NO}', "");
Expect(0, 12275, '\p{IDS_Trinary_Operator=		NO}', "");
Expect(1, 12275, '\p{^IDS_Trinary_Operator=		NO}', "");
Expect(1, 12275, '\P{IDS_Trinary_Operator=		NO}', "");
Expect(0, 12275, '\P{^IDS_Trinary_Operator=		NO}', "");
Error('\p{IDST=/a/  N}');
Error('\P{IDST=/a/  N}');
Expect(1, 12276, '\p{IDST=n}', "");
Expect(0, 12276, '\p{^IDST=n}', "");
Expect(0, 12276, '\P{IDST=n}', "");
Expect(1, 12276, '\P{^IDST=n}', "");
Expect(0, 12275, '\p{IDST=n}', "");
Expect(1, 12275, '\p{^IDST=n}', "");
Expect(1, 12275, '\P{IDST=n}', "");
Expect(0, 12275, '\P{^IDST=n}', "");
Expect(1, 12276, '\p{IDST=-	N}', "");
Expect(0, 12276, '\p{^IDST=-	N}', "");
Expect(0, 12276, '\P{IDST=-	N}', "");
Expect(1, 12276, '\P{^IDST=-	N}', "");
Expect(0, 12275, '\p{IDST=-	N}', "");
Expect(1, 12275, '\p{^IDST=-	N}', "");
Expect(1, 12275, '\P{IDST=-	N}', "");
Expect(0, 12275, '\P{^IDST=-	N}', "");
Error('\p{Is_IDS_Trinary_Operator= _f/a/}');
Error('\P{Is_IDS_Trinary_Operator= _f/a/}');
Expect(1, 12276, '\p{Is_IDS_Trinary_Operator: f}', "");
Expect(0, 12276, '\p{^Is_IDS_Trinary_Operator: f}', "");
Expect(0, 12276, '\P{Is_IDS_Trinary_Operator: f}', "");
Expect(1, 12276, '\P{^Is_IDS_Trinary_Operator: f}', "");
Expect(0, 12275, '\p{Is_IDS_Trinary_Operator: f}', "");
Expect(1, 12275, '\p{^Is_IDS_Trinary_Operator: f}', "");
Expect(1, 12275, '\P{Is_IDS_Trinary_Operator: f}', "");
Expect(0, 12275, '\P{^Is_IDS_Trinary_Operator: f}', "");
Expect(1, 12276, '\p{Is_IDS_Trinary_Operator=	_F}', "");
Expect(0, 12276, '\p{^Is_IDS_Trinary_Operator=	_F}', "");
Expect(0, 12276, '\P{Is_IDS_Trinary_Operator=	_F}', "");
Expect(1, 12276, '\P{^Is_IDS_Trinary_Operator=	_F}', "");
Expect(0, 12275, '\p{Is_IDS_Trinary_Operator=	_F}', "");
Expect(1, 12275, '\p{^Is_IDS_Trinary_Operator=	_F}', "");
Expect(1, 12275, '\P{Is_IDS_Trinary_Operator=	_F}', "");
Expect(0, 12275, '\P{^Is_IDS_Trinary_Operator=	_F}', "");
Error('\p{Is_IDST=  FALSE:=}');
Error('\P{Is_IDST=  FALSE:=}');
Expect(1, 12276, '\p{Is_IDST=false}', "");
Expect(0, 12276, '\p{^Is_IDST=false}', "");
Expect(0, 12276, '\P{Is_IDST=false}', "");
Expect(1, 12276, '\P{^Is_IDST=false}', "");
Expect(0, 12275, '\p{Is_IDST=false}', "");
Expect(1, 12275, '\p{^Is_IDST=false}', "");
Expect(1, 12275, '\P{Is_IDST=false}', "");
Expect(0, 12275, '\P{^Is_IDST=false}', "");
Expect(1, 12276, '\p{Is_IDST=_False}', "");
Expect(0, 12276, '\p{^Is_IDST=_False}', "");
Expect(0, 12276, '\P{Is_IDST=_False}', "");
Expect(1, 12276, '\P{^Is_IDST=_False}', "");
Expect(0, 12275, '\p{Is_IDST=_False}', "");
Expect(1, 12275, '\p{^Is_IDST=_False}', "");
Expect(1, 12275, '\P{Is_IDST=_False}', "");
Expect(0, 12275, '\P{^Is_IDST=_False}', "");
Error('\p{Grapheme_Cluster_Break:	__spacingmark/a/}');
Error('\P{Grapheme_Cluster_Break:	__spacingmark/a/}');
Expect(1, 119149, '\p{Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119149, '\p{^Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119149, '\P{Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119149, '\P{^Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119150, '\p{Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119150, '\p{^Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119150, '\P{Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119150, '\P{^Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119149, '\p{Grapheme_Cluster_Break=-	SPACINGMARK}', "");
Expect(0, 119149, '\p{^Grapheme_Cluster_Break=-	SPACINGMARK}', "");
Expect(0, 119149, '\P{Grapheme_Cluster_Break=-	SPACINGMARK}', "");
Expect(1, 119149, '\P{^Grapheme_Cluster_Break=-	SPACINGMARK}', "");
Expect(0, 119150, '\p{Grapheme_Cluster_Break=-	SPACINGMARK}', "");
Expect(1, 119150, '\p{^Grapheme_Cluster_Break=-	SPACINGMARK}', "");
Expect(1, 119150, '\P{Grapheme_Cluster_Break=-	SPACINGMARK}', "");
Expect(0, 119150, '\P{^Grapheme_Cluster_Break=-	SPACINGMARK}', "");
Error('\p{GCB= :=sm}');
Error('\P{GCB= :=sm}');
Expect(1, 119149, '\p{GCB=sm}', "");
Expect(0, 119149, '\p{^GCB=sm}', "");
Expect(0, 119149, '\P{GCB=sm}', "");
Expect(1, 119149, '\P{^GCB=sm}', "");
Expect(0, 119150, '\p{GCB=sm}', "");
Expect(1, 119150, '\p{^GCB=sm}', "");
Expect(1, 119150, '\P{GCB=sm}', "");
Expect(0, 119150, '\P{^GCB=sm}', "");
Expect(1, 119149, '\p{GCB= 	SM}', "");
Expect(0, 119149, '\p{^GCB= 	SM}', "");
Expect(0, 119149, '\P{GCB= 	SM}', "");
Expect(1, 119149, '\P{^GCB= 	SM}', "");
Expect(0, 119150, '\p{GCB= 	SM}', "");
Expect(1, 119150, '\p{^GCB= 	SM}', "");
Expect(1, 119150, '\P{GCB= 	SM}', "");
Expect(0, 119150, '\P{^GCB= 	SM}', "");
Error('\p{Is_Grapheme_Cluster_Break=/a/ spacingmark}');
Error('\P{Is_Grapheme_Cluster_Break=/a/ spacingmark}');
Expect(1, 119149, '\p{Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119149, '\p{^Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119149, '\P{Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119149, '\P{^Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119150, '\p{Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119150, '\p{^Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119150, '\P{Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(0, 119150, '\P{^Is_Grapheme_Cluster_Break=spacingmark}', "");
Expect(1, 119149, '\p{Is_Grapheme_Cluster_Break=  SPACINGMARK}', "");
Expect(0, 119149, '\p{^Is_Grapheme_Cluster_Break=  SPACINGMARK}', "");
Expect(0, 119149, '\P{Is_Grapheme_Cluster_Break=  SPACINGMARK}', "");
Expect(1, 119149, '\P{^Is_Grapheme_Cluster_Break=  SPACINGMARK}', "");
Expect(0, 119150, '\p{Is_Grapheme_Cluster_Break=  SPACINGMARK}', "");
Expect(1, 119150, '\p{^Is_Grapheme_Cluster_Break=  SPACINGMARK}', "");
Expect(1, 119150, '\P{Is_Grapheme_Cluster_Break=  SPACINGMARK}', "");
Expect(0, 119150, '\P{^Is_Grapheme_Cluster_Break=  SPACINGMARK}', "");
Error('\p{Is_GCB:	 :=SM}');
Error('\P{Is_GCB:	 :=SM}');
Expect(1, 119149, '\p{Is_GCB=sm}', "");
Expect(0, 119149, '\p{^Is_GCB=sm}', "");
Expect(0, 119149, '\P{Is_GCB=sm}', "");
Expect(1, 119149, '\P{^Is_GCB=sm}', "");
Expect(0, 119150, '\p{Is_GCB=sm}', "");
Expect(1, 119150, '\p{^Is_GCB=sm}', "");
Expect(1, 119150, '\P{Is_GCB=sm}', "");
Expect(0, 119150, '\P{^Is_GCB=sm}', "");
Expect(1, 119149, '\p{Is_GCB=--SM}', "");
Expect(0, 119149, '\p{^Is_GCB=--SM}', "");
Expect(0, 119149, '\P{Is_GCB=--SM}', "");
Expect(1, 119149, '\P{^Is_GCB=--SM}', "");
Expect(0, 119150, '\p{Is_GCB=--SM}', "");
Expect(1, 119150, '\p{^Is_GCB=--SM}', "");
Expect(1, 119150, '\P{Is_GCB=--SM}', "");
Expect(0, 119150, '\P{^Is_GCB=--SM}', "");
Error('\p{Grapheme_Cluster_Break=_/a/T}');
Error('\P{Grapheme_Cluster_Break=_/a/T}');
Expect(1, 55291, '\p{Grapheme_Cluster_Break=t}', "");
Expect(0, 55291, '\p{^Grapheme_Cluster_Break=t}', "");
Expect(0, 55291, '\P{Grapheme_Cluster_Break=t}', "");
Expect(1, 55291, '\P{^Grapheme_Cluster_Break=t}', "");
Expect(0, 55292, '\p{Grapheme_Cluster_Break=t}', "");
Expect(1, 55292, '\p{^Grapheme_Cluster_Break=t}', "");
Expect(1, 55292, '\P{Grapheme_Cluster_Break=t}', "");
Expect(0, 55292, '\P{^Grapheme_Cluster_Break=t}', "");
Expect(1, 55291, '\p{Grapheme_Cluster_Break:	-_T}', "");
Expect(0, 55291, '\p{^Grapheme_Cluster_Break:	-_T}', "");
Expect(0, 55291, '\P{Grapheme_Cluster_Break:	-_T}', "");
Expect(1, 55291, '\P{^Grapheme_Cluster_Break:	-_T}', "");
Expect(0, 55292, '\p{Grapheme_Cluster_Break:	-_T}', "");
Expect(1, 55292, '\p{^Grapheme_Cluster_Break:	-_T}', "");
Expect(1, 55292, '\P{Grapheme_Cluster_Break:	-_T}', "");
Expect(0, 55292, '\P{^Grapheme_Cluster_Break:	-_T}', "");
Error('\p{GCB= -T:=}');
Error('\P{GCB= -T:=}');
Expect(1, 55291, '\p{GCB=t}', "");
Expect(0, 55291, '\p{^GCB=t}', "");
Expect(0, 55291, '\P{GCB=t}', "");
Expect(1, 55291, '\P{^GCB=t}', "");
Expect(0, 55292, '\p{GCB=t}', "");
Expect(1, 55292, '\p{^GCB=t}', "");
Expect(1, 55292, '\P{GCB=t}', "");
Expect(0, 55292, '\P{^GCB=t}', "");
Expect(1, 55291, '\p{GCB=__T}', "");
Expect(0, 55291, '\p{^GCB=__T}', "");
Expect(0, 55291, '\P{GCB=__T}', "");
Expect(1, 55291, '\P{^GCB=__T}', "");
Expect(0, 55292, '\p{GCB=__T}', "");
Expect(1, 55292, '\p{^GCB=__T}', "");
Expect(1, 55292, '\P{GCB=__T}', "");
Expect(0, 55292, '\P{^GCB=__T}', "");
Error('\p{Is_Grapheme_Cluster_Break= 	T:=}');
Error('\P{Is_Grapheme_Cluster_Break= 	T:=}');
Expect(1, 55291, '\p{Is_Grapheme_Cluster_Break=t}', "");
Expect(0, 55291, '\p{^Is_Grapheme_Cluster_Break=t}', "");
Expect(0, 55291, '\P{Is_Grapheme_Cluster_Break=t}', "");
Expect(1, 55291, '\P{^Is_Grapheme_Cluster_Break=t}', "");
Expect(0, 55292, '\p{Is_Grapheme_Cluster_Break=t}', "");
Expect(1, 55292, '\p{^Is_Grapheme_Cluster_Break=t}', "");
Expect(1, 55292, '\P{Is_Grapheme_Cluster_Break=t}', "");
Expect(0, 55292, '\P{^Is_Grapheme_Cluster_Break=t}', "");
Expect(1, 55291, '\p{Is_Grapheme_Cluster_Break=	T}', "");
Expect(0, 55291, '\p{^Is_Grapheme_Cluster_Break=	T}', "");
Expect(0, 55291, '\P{Is_Grapheme_Cluster_Break=	T}', "");
Expect(1, 55291, '\P{^Is_Grapheme_Cluster_Break=	T}', "");
Expect(0, 55292, '\p{Is_Grapheme_Cluster_Break=	T}', "");
Expect(1, 55292, '\p{^Is_Grapheme_Cluster_Break=	T}', "");
Expect(1, 55292, '\P{Is_Grapheme_Cluster_Break=	T}', "");
Expect(0, 55292, '\P{^Is_Grapheme_Cluster_Break=	T}', "");
Error('\p{Is_GCB:-/a/T}');
Error('\P{Is_GCB:-/a/T}');
Expect(1, 55291, '\p{Is_GCB=t}', "");
Expect(0, 55291, '\p{^Is_GCB=t}', "");
Expect(0, 55291, '\P{Is_GCB=t}', "");
Expect(1, 55291, '\P{^Is_GCB=t}', "");
Expect(0, 55292, '\p{Is_GCB=t}', "");
Expect(1, 55292, '\p{^Is_GCB=t}', "");
Expect(1, 55292, '\P{Is_GCB=t}', "");
Expect(0, 55292, '\P{^Is_GCB=t}', "");
Expect(1, 55291, '\p{Is_GCB= T}', "");
Expect(0, 55291, '\p{^Is_GCB= T}', "");
Expect(0, 55291, '\P{Is_GCB= T}', "");
Expect(1, 55291, '\P{^Is_GCB= T}', "");
Expect(0, 55292, '\p{Is_GCB= T}', "");
Expect(1, 55292, '\p{^Is_GCB= T}', "");
Expect(1, 55292, '\P{Is_GCB= T}', "");
Expect(0, 55292, '\P{^Is_GCB= T}', "");
Error('\p{Grapheme_Cluster_Break= _LVT/a/}');
Error('\P{Grapheme_Cluster_Break= _LVT/a/}');
Expect(1, 55203, '\p{Grapheme_Cluster_Break=lvt}', "");
Expect(0, 55203, '\p{^Grapheme_Cluster_Break=lvt}', "");
Expect(0, 55203, '\P{Grapheme_Cluster_Break=lvt}', "");
Expect(1, 55203, '\P{^Grapheme_Cluster_Break=lvt}', "");
Expect(0, 55204, '\p{Grapheme_Cluster_Break=lvt}', "");
Expect(1, 55204, '\p{^Grapheme_Cluster_Break=lvt}', "");
Expect(1, 55204, '\P{Grapheme_Cluster_Break=lvt}', "");
Expect(0, 55204, '\P{^Grapheme_Cluster_Break=lvt}', "");
Expect(1, 55203, '\p{Grapheme_Cluster_Break: _ LVT}', "");
Expect(0, 55203, '\p{^Grapheme_Cluster_Break: _ LVT}', "");
Expect(0, 55203, '\P{Grapheme_Cluster_Break: _ LVT}', "");
Expect(1, 55203, '\P{^Grapheme_Cluster_Break: _ LVT}', "");
Expect(0, 55204, '\p{Grapheme_Cluster_Break: _ LVT}', "");
Expect(1, 55204, '\p{^Grapheme_Cluster_Break: _ LVT}', "");
Expect(1, 55204, '\P{Grapheme_Cluster_Break: _ LVT}', "");
Expect(0, 55204, '\P{^Grapheme_Cluster_Break: _ LVT}', "");
Error('\p{GCB=:=LVT}');
Error('\P{GCB=:=LVT}');
Expect(1, 55203, '\p{GCB=lvt}', "");
Expect(0, 55203, '\p{^GCB=lvt}', "");
Expect(0, 55203, '\P{GCB=lvt}', "");
Expect(1, 55203, '\P{^GCB=lvt}', "");
Expect(0, 55204, '\p{GCB=lvt}', "");
Expect(1, 55204, '\p{^GCB=lvt}', "");
Expect(1, 55204, '\P{GCB=lvt}', "");
Expect(0, 55204, '\P{^GCB=lvt}', "");
Expect(1, 55203, '\p{GCB=- LVT}', "");
Expect(0, 55203, '\p{^GCB=- LVT}', "");
Expect(0, 55203, '\P{GCB=- LVT}', "");
Expect(1, 55203, '\P{^GCB=- LVT}', "");
Expect(0, 55204, '\p{GCB=- LVT}', "");
Expect(1, 55204, '\p{^GCB=- LVT}', "");
Expect(1, 55204, '\P{GCB=- LVT}', "");
Expect(0, 55204, '\P{^GCB=- LVT}', "");
Error('\p{Is_Grapheme_Cluster_Break=:=_	LVT}');
Error('\P{Is_Grapheme_Cluster_Break=:=_	LVT}');
Expect(1, 55203, '\p{Is_Grapheme_Cluster_Break=lvt}', "");
Expect(0, 55203, '\p{^Is_Grapheme_Cluster_Break=lvt}', "");
Expect(0, 55203, '\P{Is_Grapheme_Cluster_Break=lvt}', "");
Expect(1, 55203, '\P{^Is_Grapheme_Cluster_Break=lvt}', "");
Expect(0, 55204, '\p{Is_Grapheme_Cluster_Break=lvt}', "");
Expect(1, 55204, '\p{^Is_Grapheme_Cluster_Break=lvt}', "");
Expect(1, 55204, '\P{Is_Grapheme_Cluster_Break=lvt}', "");
Expect(0, 55204, '\P{^Is_Grapheme_Cluster_Break=lvt}', "");
Expect(1, 55203, '\p{Is_Grapheme_Cluster_Break=_lvt}', "");
Expect(0, 55203, '\p{^Is_Grapheme_Cluster_Break=_lvt}', "");
Expect(0, 55203, '\P{Is_Grapheme_Cluster_Break=_lvt}', "");
Expect(1, 55203, '\P{^Is_Grapheme_Cluster_Break=_lvt}', "");
Expect(0, 55204, '\p{Is_Grapheme_Cluster_Break=_lvt}', "");
Expect(1, 55204, '\p{^Is_Grapheme_Cluster_Break=_lvt}', "");
Expect(1, 55204, '\P{Is_Grapheme_Cluster_Break=_lvt}', "");
Expect(0, 55204, '\P{^Is_Grapheme_Cluster_Break=_lvt}', "");
Error('\p{Is_GCB=_	LVT:=}');
Error('\P{Is_GCB=_	LVT:=}');
Expect(1, 55203, '\p{Is_GCB=lvt}', "");
Expect(0, 55203, '\p{^Is_GCB=lvt}', "");
Expect(0, 55203, '\P{Is_GCB=lvt}', "");
Expect(1, 55203, '\P{^Is_GCB=lvt}', "");
Expect(0, 55204, '\p{Is_GCB=lvt}', "");
Expect(1, 55204, '\p{^Is_GCB=lvt}', "");
Expect(1, 55204, '\P{Is_GCB=lvt}', "");
Expect(0, 55204, '\P{^Is_GCB=lvt}', "");
Expect(1, 55203, '\p{Is_GCB=		LVT}', "");
Expect(0, 55203, '\p{^Is_GCB=		LVT}', "");
Expect(0, 55203, '\P{Is_GCB=		LVT}', "");
Expect(1, 55203, '\P{^Is_GCB=		LVT}', "");
Expect(0, 55204, '\p{Is_GCB=		LVT}', "");
Expect(1, 55204, '\p{^Is_GCB=		LVT}', "");
Expect(1, 55204, '\P{Is_GCB=		LVT}', "");
Expect(0, 55204, '\P{^Is_GCB=		LVT}', "");
Error('\p{Grapheme_Cluster_Break=/a/	 control}');
Error('\P{Grapheme_Cluster_Break=/a/	 control}');
Expect(1, 921599, '\p{Grapheme_Cluster_Break=control}', "");
Expect(0, 921599, '\p{^Grapheme_Cluster_Break=control}', "");
Expect(0, 921599, '\P{Grapheme_Cluster_Break=control}', "");
Expect(1, 921599, '\P{^Grapheme_Cluster_Break=control}', "");
Expect(0, 921600, '\p{Grapheme_Cluster_Break=control}', "");
Expect(1, 921600, '\p{^Grapheme_Cluster_Break=control}', "");
Expect(1, 921600, '\P{Grapheme_Cluster_Break=control}', "");
Expect(0, 921600, '\P{^Grapheme_Cluster_Break=control}', "");
Expect(1, 921599, '\p{Grapheme_Cluster_Break=_control}', "");
Expect(0, 921599, '\p{^Grapheme_Cluster_Break=_control}', "");
Expect(0, 921599, '\P{Grapheme_Cluster_Break=_control}', "");
Expect(1, 921599, '\P{^Grapheme_Cluster_Break=_control}', "");
Expect(0, 921600, '\p{Grapheme_Cluster_Break=_control}', "");
Expect(1, 921600, '\p{^Grapheme_Cluster_Break=_control}', "");
Expect(1, 921600, '\P{Grapheme_Cluster_Break=_control}', "");
Expect(0, 921600, '\P{^Grapheme_Cluster_Break=_control}', "");
Error('\p{GCB=-/a/CN}');
Error('\P{GCB=-/a/CN}');
Expect(1, 921599, '\p{GCB:cn}', "");
Expect(0, 921599, '\p{^GCB:cn}', "");
Expect(0, 921599, '\P{GCB:cn}', "");
Expect(1, 921599, '\P{^GCB:cn}', "");
Expect(0, 921600, '\p{GCB:cn}', "");
Expect(1, 921600, '\p{^GCB:cn}', "");
Expect(1, 921600, '\P{GCB:cn}', "");
Expect(0, 921600, '\P{^GCB:cn}', "");
Expect(1, 921599, '\p{GCB:  _CN}', "");
Expect(0, 921599, '\p{^GCB:  _CN}', "");
Expect(0, 921599, '\P{GCB:  _CN}', "");
Expect(1, 921599, '\P{^GCB:  _CN}', "");
Expect(0, 921600, '\p{GCB:  _CN}', "");
Expect(1, 921600, '\p{^GCB:  _CN}', "");
Expect(1, 921600, '\P{GCB:  _CN}', "");
Expect(0, 921600, '\P{^GCB:  _CN}', "");
Error('\p{Is_Grapheme_Cluster_Break=	:=Control}');
Error('\P{Is_Grapheme_Cluster_Break=	:=Control}');
Expect(1, 921599, '\p{Is_Grapheme_Cluster_Break=control}', "");
Expect(0, 921599, '\p{^Is_Grapheme_Cluster_Break=control}', "");
Expect(0, 921599, '\P{Is_Grapheme_Cluster_Break=control}', "");
Expect(1, 921599, '\P{^Is_Grapheme_Cluster_Break=control}', "");
Expect(0, 921600, '\p{Is_Grapheme_Cluster_Break=control}', "");
Expect(1, 921600, '\p{^Is_Grapheme_Cluster_Break=control}', "");
Expect(1, 921600, '\P{Is_Grapheme_Cluster_Break=control}', "");
Expect(0, 921600, '\P{^Is_Grapheme_Cluster_Break=control}', "");
Expect(1, 921599, '\p{Is_Grapheme_Cluster_Break= Control}', "");
Expect(0, 921599, '\p{^Is_Grapheme_Cluster_Break= Control}', "");
Expect(0, 921599, '\P{Is_Grapheme_Cluster_Break= Control}', "");
Expect(1, 921599, '\P{^Is_Grapheme_Cluster_Break= Control}', "");
Expect(0, 921600, '\p{Is_Grapheme_Cluster_Break= Control}', "");
Expect(1, 921600, '\p{^Is_Grapheme_Cluster_Break= Control}', "");
Expect(1, 921600, '\P{Is_Grapheme_Cluster_Break= Control}', "");
Expect(0, 921600, '\P{^Is_Grapheme_Cluster_Break= Control}', "");
Error('\p{Is_GCB=_CN/a/}');
Error('\P{Is_GCB=_CN/a/}');
Expect(1, 921599, '\p{Is_GCB=cn}', "");
Expect(0, 921599, '\p{^Is_GCB=cn}', "");
Expect(0, 921599, '\P{Is_GCB=cn}', "");
Expect(1, 921599, '\P{^Is_GCB=cn}', "");
Expect(0, 921600, '\p{Is_GCB=cn}', "");
Expect(1, 921600, '\p{^Is_GCB=cn}', "");
Expect(1, 921600, '\P{Is_GCB=cn}', "");
Expect(0, 921600, '\P{^Is_GCB=cn}', "");
Expect(1, 921599, '\p{Is_GCB=	CN}', "");
Expect(0, 921599, '\p{^Is_GCB=	CN}', "");
Expect(0, 921599, '\P{Is_GCB=	CN}', "");
Expect(1, 921599, '\P{^Is_GCB=	CN}', "");
Expect(0, 921600, '\p{Is_GCB=	CN}', "");
Expect(1, 921600, '\p{^Is_GCB=	CN}', "");
Expect(1, 921600, '\P{Is_GCB=	CN}', "");
Expect(0, 921600, '\P{^Is_GCB=	CN}', "");
Error('\p{Grapheme_Cluster_Break=-:=Prepend}');
Error('\P{Grapheme_Cluster_Break=-:=Prepend}');
Expect(0, 1, '\p{Grapheme_Cluster_Break=prepend}', "");
Expect(1, 1, '\p{^Grapheme_Cluster_Break=prepend}', "");
Expect(1, 1, '\P{Grapheme_Cluster_Break=prepend}', "");
Expect(0, 1, '\P{^Grapheme_Cluster_Break=prepend}', "");
Expect(0, 1, '\p{Grapheme_Cluster_Break=-PREPEND}', "");
Expect(1, 1, '\p{^Grapheme_Cluster_Break=-PREPEND}', "");
Expect(1, 1, '\P{Grapheme_Cluster_Break=-PREPEND}', "");
Expect(0, 1, '\P{^Grapheme_Cluster_Break=-PREPEND}', "");
Error('\p{GCB=_-PP:=}');
Error('\P{GCB=_-PP:=}');
Expect(0, 1, '\p{GCB=pp}', "");
Expect(1, 1, '\p{^GCB=pp}', "");
Expect(1, 1, '\P{GCB=pp}', "");
Expect(0, 1, '\P{^GCB=pp}', "");
Expect(0, 1, '\p{GCB=	 pp}', "");
Expect(1, 1, '\p{^GCB=	 pp}', "");
Expect(1, 1, '\P{GCB=	 pp}', "");
Expect(0, 1, '\P{^GCB=	 pp}', "");
Error('\p{Is_Grapheme_Cluster_Break=/a/_	Prepend}');
Error('\P{Is_Grapheme_Cluster_Break=/a/_	Prepend}');
Expect(0, 1, '\p{Is_Grapheme_Cluster_Break=prepend}', "");
Expect(1, 1, '\p{^Is_Grapheme_Cluster_Break=prepend}', "");
Expect(1, 1, '\P{Is_Grapheme_Cluster_Break=prepend}', "");
Expect(0, 1, '\P{^Is_Grapheme_Cluster_Break=prepend}', "");
Expect(0, 1, '\p{Is_Grapheme_Cluster_Break=		Prepend}', "");
Expect(1, 1, '\p{^Is_Grapheme_Cluster_Break=		Prepend}', "");
Expect(1, 1, '\P{Is_Grapheme_Cluster_Break=		Prepend}', "");
Expect(0, 1, '\P{^Is_Grapheme_Cluster_Break=		Prepend}', "");
Error('\p{Is_GCB=_:=PP}');
Error('\P{Is_GCB=_:=PP}');
Expect(0, 1, '\p{Is_GCB:	pp}', "");
Expect(1, 1, '\p{^Is_GCB:	pp}', "");
Expect(1, 1, '\P{Is_GCB:	pp}', "");
Expect(0, 1, '\P{^Is_GCB:	pp}', "");
Expect(0, 1, '\p{Is_GCB=_ PP}', "");
Expect(1, 1, '\p{^Is_GCB=_ PP}', "");
Expect(1, 1, '\P{Is_GCB=_ PP}', "");
Expect(0, 1, '\P{^Is_GCB=_ PP}', "");
Error('\p{Grapheme_Cluster_Break=-:=LV}');
Error('\P{Grapheme_Cluster_Break=-:=LV}');
Expect(1, 55176, '\p{Grapheme_Cluster_Break=lv}', "");
Expect(0, 55176, '\p{^Grapheme_Cluster_Break=lv}', "");
Expect(0, 55176, '\P{Grapheme_Cluster_Break=lv}', "");
Expect(1, 55176, '\P{^Grapheme_Cluster_Break=lv}', "");
Expect(0, 55177, '\p{Grapheme_Cluster_Break=lv}', "");
Expect(1, 55177, '\p{^Grapheme_Cluster_Break=lv}', "");
Expect(1, 55177, '\P{Grapheme_Cluster_Break=lv}', "");
Expect(0, 55177, '\P{^Grapheme_Cluster_Break=lv}', "");
Expect(1, 55176, '\p{Grapheme_Cluster_Break=LV}', "");
Expect(0, 55176, '\p{^Grapheme_Cluster_Break=LV}', "");
Expect(0, 55176, '\P{Grapheme_Cluster_Break=LV}', "");
Expect(1, 55176, '\P{^Grapheme_Cluster_Break=LV}', "");
Expect(0, 55177, '\p{Grapheme_Cluster_Break=LV}', "");
Expect(1, 55177, '\p{^Grapheme_Cluster_Break=LV}', "");
Expect(1, 55177, '\P{Grapheme_Cluster_Break=LV}', "");
Expect(0, 55177, '\P{^Grapheme_Cluster_Break=LV}', "");
Error('\p{GCB=	/a/LV}');
Error('\P{GCB=	/a/LV}');
Expect(1, 55176, '\p{GCB=lv}', "");
Expect(0, 55176, '\p{^GCB=lv}', "");
Expect(0, 55176, '\P{GCB=lv}', "");
Expect(1, 55176, '\P{^GCB=lv}', "");
Expect(0, 55177, '\p{GCB=lv}', "");
Expect(1, 55177, '\p{^GCB=lv}', "");
Expect(1, 55177, '\P{GCB=lv}', "");
Expect(0, 55177, '\P{^GCB=lv}', "");
Expect(1, 55176, '\p{GCB=_LV}', "");
Expect(0, 55176, '\p{^GCB=_LV}', "");
Expect(0, 55176, '\P{GCB=_LV}', "");
Expect(1, 55176, '\P{^GCB=_LV}', "");
Expect(0, 55177, '\p{GCB=_LV}', "");
Expect(1, 55177, '\p{^GCB=_LV}', "");
Expect(1, 55177, '\P{GCB=_LV}', "");
Expect(0, 55177, '\P{^GCB=_LV}', "");
Error('\p{Is_Grapheme_Cluster_Break=:=_LV}');
Error('\P{Is_Grapheme_Cluster_Break=:=_LV}');
Expect(1, 55176, '\p{Is_Grapheme_Cluster_Break: lv}', "");
Expect(0, 55176, '\p{^Is_Grapheme_Cluster_Break: lv}', "");
Expect(0, 55176, '\P{Is_Grapheme_Cluster_Break: lv}', "");
Expect(1, 55176, '\P{^Is_Grapheme_Cluster_Break: lv}', "");
Expect(0, 55177, '\p{Is_Grapheme_Cluster_Break: lv}', "");
Expect(1, 55177, '\p{^Is_Grapheme_Cluster_Break: lv}', "");
Expect(1, 55177, '\P{Is_Grapheme_Cluster_Break: lv}', "");
Expect(0, 55177, '\P{^Is_Grapheme_Cluster_Break: lv}', "");
Expect(1, 55176, '\p{Is_Grapheme_Cluster_Break=	LV}', "");
Expect(0, 55176, '\p{^Is_Grapheme_Cluster_Break=	LV}', "");
Expect(0, 55176, '\P{Is_Grapheme_Cluster_Break=	LV}', "");
Expect(1, 55176, '\P{^Is_Grapheme_Cluster_Break=	LV}', "");
Expect(0, 55177, '\p{Is_Grapheme_Cluster_Break=	LV}', "");
Expect(1, 55177, '\p{^Is_Grapheme_Cluster_Break=	LV}', "");
Expect(1, 55177, '\P{Is_Grapheme_Cluster_Break=	LV}', "");
Expect(0, 55177, '\P{^Is_Grapheme_Cluster_Break=	LV}', "");
Error('\p{Is_GCB= LV:=}');
Error('\P{Is_GCB= LV:=}');
Expect(1, 55176, '\p{Is_GCB=lv}', "");
Expect(0, 55176, '\p{^Is_GCB=lv}', "");
Expect(0, 55176, '\P{Is_GCB=lv}', "");
Expect(1, 55176, '\P{^Is_GCB=lv}', "");
Expect(0, 55177, '\p{Is_GCB=lv}', "");
Expect(1, 55177, '\p{^Is_GCB=lv}', "");
Expect(1, 55177, '\P{Is_GCB=lv}', "");
Expect(0, 55177, '\P{^Is_GCB=lv}', "");
Expect(1, 55176, '\p{Is_GCB=_LV}', "");
Expect(0, 55176, '\p{^Is_GCB=_LV}', "");
Expect(0, 55176, '\P{Is_GCB=_LV}', "");
Expect(1, 55176, '\P{^Is_GCB=_LV}', "");
Expect(0, 55177, '\p{Is_GCB=_LV}', "");
Expect(1, 55177, '\p{^Is_GCB=_LV}', "");
Expect(1, 55177, '\P{Is_GCB=_LV}', "");
Expect(0, 55177, '\P{^Is_GCB=_LV}', "");
Error('\p{Grapheme_Cluster_Break=-:=CR}');
Error('\P{Grapheme_Cluster_Break=-:=CR}');
Expect(1, 13, '\p{Grapheme_Cluster_Break=cr}', "");
Expect(0, 13, '\p{^Grapheme_Cluster_Break=cr}', "");
Expect(0, 13, '\P{Grapheme_Cluster_Break=cr}', "");
Expect(1, 13, '\P{^Grapheme_Cluster_Break=cr}', "");
Expect(0, 14, '\p{Grapheme_Cluster_Break=cr}', "");
Expect(1, 14, '\p{^Grapheme_Cluster_Break=cr}', "");
Expect(1, 14, '\P{Grapheme_Cluster_Break=cr}', "");
Expect(0, 14, '\P{^Grapheme_Cluster_Break=cr}', "");
Expect(1, 13, '\p{Grapheme_Cluster_Break=- CR}', "");
Expect(0, 13, '\p{^Grapheme_Cluster_Break=- CR}', "");
Expect(0, 13, '\P{Grapheme_Cluster_Break=- CR}', "");
Expect(1, 13, '\P{^Grapheme_Cluster_Break=- CR}', "");
Expect(0, 14, '\p{Grapheme_Cluster_Break=- CR}', "");
Expect(1, 14, '\p{^Grapheme_Cluster_Break=- CR}', "");
Expect(1, 14, '\P{Grapheme_Cluster_Break=- CR}', "");
Expect(0, 14, '\P{^Grapheme_Cluster_Break=- CR}', "");
Error('\p{GCB=	cr:=}');
Error('\P{GCB=	cr:=}');
Expect(1, 13, '\p{GCB=cr}', "");
Expect(0, 13, '\p{^GCB=cr}', "");
Expect(0, 13, '\P{GCB=cr}', "");
Expect(1, 13, '\P{^GCB=cr}', "");
Expect(0, 14, '\p{GCB=cr}', "");
Expect(1, 14, '\p{^GCB=cr}', "");
Expect(1, 14, '\P{GCB=cr}', "");
Expect(0, 14, '\P{^GCB=cr}', "");
Expect(1, 13, '\p{GCB=CR}', "");
Expect(0, 13, '\p{^GCB=CR}', "");
Expect(0, 13, '\P{GCB=CR}', "");
Expect(1, 13, '\P{^GCB=CR}', "");
Expect(0, 14, '\p{GCB=CR}', "");
Expect(1, 14, '\p{^GCB=CR}', "");
Expect(1, 14, '\P{GCB=CR}', "");
Expect(0, 14, '\P{^GCB=CR}', "");
Error('\p{Is_Grapheme_Cluster_Break=-/a/CR}');
Error('\P{Is_Grapheme_Cluster_Break=-/a/CR}');
Expect(1, 13, '\p{Is_Grapheme_Cluster_Break:	cr}', "");
Expect(0, 13, '\p{^Is_Grapheme_Cluster_Break:	cr}', "");
Expect(0, 13, '\P{Is_Grapheme_Cluster_Break:	cr}', "");
Expect(1, 13, '\P{^Is_Grapheme_Cluster_Break:	cr}', "");
Expect(0, 14, '\p{Is_Grapheme_Cluster_Break:	cr}', "");
Expect(1, 14, '\p{^Is_Grapheme_Cluster_Break:	cr}', "");
Expect(1, 14, '\P{Is_Grapheme_Cluster_Break:	cr}', "");
Expect(0, 14, '\P{^Is_Grapheme_Cluster_Break:	cr}', "");
Expect(1, 13, '\p{Is_Grapheme_Cluster_Break= 	cr}', "");
Expect(0, 13, '\p{^Is_Grapheme_Cluster_Break= 	cr}', "");
Expect(0, 13, '\P{Is_Grapheme_Cluster_Break= 	cr}', "");
Expect(1, 13, '\P{^Is_Grapheme_Cluster_Break= 	cr}', "");
Expect(0, 14, '\p{Is_Grapheme_Cluster_Break= 	cr}', "");
Expect(1, 14, '\p{^Is_Grapheme_Cluster_Break= 	cr}', "");
Expect(1, 14, '\P{Is_Grapheme_Cluster_Break= 	cr}', "");
Expect(0, 14, '\P{^Is_Grapheme_Cluster_Break= 	cr}', "");
Error('\p{Is_GCB= 	CR:=}');
Error('\P{Is_GCB= 	CR:=}');
Expect(1, 13, '\p{Is_GCB=cr}', "");
Expect(0, 13, '\p{^Is_GCB=cr}', "");
Expect(0, 13, '\P{Is_GCB=cr}', "");
Expect(1, 13, '\P{^Is_GCB=cr}', "");
Expect(0, 14, '\p{Is_GCB=cr}', "");
Expect(1, 14, '\p{^Is_GCB=cr}', "");
Expect(1, 14, '\P{Is_GCB=cr}', "");
Expect(0, 14, '\P{^Is_GCB=cr}', "");
Expect(1, 13, '\p{Is_GCB=_cr}', "");
Expect(0, 13, '\p{^Is_GCB=_cr}', "");
Expect(0, 13, '\P{Is_GCB=_cr}', "");
Expect(1, 13, '\P{^Is_GCB=_cr}', "");
Expect(0, 14, '\p{Is_GCB=_cr}', "");
Expect(1, 14, '\p{^Is_GCB=_cr}', "");
Expect(1, 14, '\P{Is_GCB=_cr}', "");
Expect(0, 14, '\P{^Is_GCB=_cr}', "");
Error('\p{Grapheme_Cluster_Break=  Other:=}');
Error('\P{Grapheme_Cluster_Break=  Other:=}');
Expect(1, 921600, '\p{Grapheme_Cluster_Break:	other}', "");
Expect(0, 921600, '\p{^Grapheme_Cluster_Break:	other}', "");
Expect(0, 921600, '\P{Grapheme_Cluster_Break:	other}', "");
Expect(1, 921600, '\P{^Grapheme_Cluster_Break:	other}', "");
Expect(0, 921599, '\p{Grapheme_Cluster_Break:	other}', "");
Expect(1, 921599, '\p{^Grapheme_Cluster_Break:	other}', "");
Expect(1, 921599, '\P{Grapheme_Cluster_Break:	other}', "");
Expect(0, 921599, '\P{^Grapheme_Cluster_Break:	other}', "");
Expect(1, 921600, '\p{Grapheme_Cluster_Break= Other}', "");
Expect(0, 921600, '\p{^Grapheme_Cluster_Break= Other}', "");
Expect(0, 921600, '\P{Grapheme_Cluster_Break= Other}', "");
Expect(1, 921600, '\P{^Grapheme_Cluster_Break= Other}', "");
Expect(0, 921599, '\p{Grapheme_Cluster_Break= Other}', "");
Expect(1, 921599, '\p{^Grapheme_Cluster_Break= Other}', "");
Expect(1, 921599, '\P{Grapheme_Cluster_Break= Other}', "");
Expect(0, 921599, '\P{^Grapheme_Cluster_Break= Other}', "");
Error('\p{GCB= :=XX}');
Error('\P{GCB= :=XX}');
Expect(1, 921600, '\p{GCB=xx}', "");
Expect(0, 921600, '\p{^GCB=xx}', "");
Expect(0, 921600, '\P{GCB=xx}', "");
Expect(1, 921600, '\P{^GCB=xx}', "");
Expect(0, 921599, '\p{GCB=xx}', "");
Expect(1, 921599, '\p{^GCB=xx}', "");
Expect(1, 921599, '\P{GCB=xx}', "");
Expect(0, 921599, '\P{^GCB=xx}', "");
Expect(1, 921600, '\p{GCB= xx}', "");
Expect(0, 921600, '\p{^GCB= xx}', "");
Expect(0, 921600, '\P{GCB= xx}', "");
Expect(1, 921600, '\P{^GCB= xx}', "");
Expect(0, 921599, '\p{GCB= xx}', "");
Expect(1, 921599, '\p{^GCB= xx}', "");
Expect(1, 921599, '\P{GCB= xx}', "");
Expect(0, 921599, '\P{^GCB= xx}', "");
Error('\p{Is_Grapheme_Cluster_Break=_Other:=}');
Error('\P{Is_Grapheme_Cluster_Break=_Other:=}');
Expect(1, 921600, '\p{Is_Grapheme_Cluster_Break=other}', "");
Expect(0, 921600, '\p{^Is_Grapheme_Cluster_Break=other}', "");
Expect(0, 921600, '\P{Is_Grapheme_Cluster_Break=other}', "");
Expect(1, 921600, '\P{^Is_Grapheme_Cluster_Break=other}', "");
Expect(0, 921599, '\p{Is_Grapheme_Cluster_Break=other}', "");
Expect(1, 921599, '\p{^Is_Grapheme_Cluster_Break=other}', "");
Expect(1, 921599, '\P{Is_Grapheme_Cluster_Break=other}', "");
Expect(0, 921599, '\P{^Is_Grapheme_Cluster_Break=other}', "");
Expect(1, 921600, '\p{Is_Grapheme_Cluster_Break=-other}', "");
Expect(0, 921600, '\p{^Is_Grapheme_Cluster_Break=-other}', "");
Expect(0, 921600, '\P{Is_Grapheme_Cluster_Break=-other}', "");
Expect(1, 921600, '\P{^Is_Grapheme_Cluster_Break=-other}', "");
Expect(0, 921599, '\p{Is_Grapheme_Cluster_Break=-other}', "");
Expect(1, 921599, '\p{^Is_Grapheme_Cluster_Break=-other}', "");
Expect(1, 921599, '\P{Is_Grapheme_Cluster_Break=-other}', "");
Expect(0, 921599, '\P{^Is_Grapheme_Cluster_Break=-other}', "");
Error('\p{Is_GCB=	:=XX}');
Error('\P{Is_GCB=	:=XX}');
Expect(1, 921600, '\p{Is_GCB=xx}', "");
Expect(0, 921600, '\p{^Is_GCB=xx}', "");
Expect(0, 921600, '\P{Is_GCB=xx}', "");
Expect(1, 921600, '\P{^Is_GCB=xx}', "");
Expect(0, 921599, '\p{Is_GCB=xx}', "");
Expect(1, 921599, '\p{^Is_GCB=xx}', "");
Expect(1, 921599, '\P{Is_GCB=xx}', "");
Expect(0, 921599, '\P{^Is_GCB=xx}', "");
Expect(1, 921600, '\p{Is_GCB= XX}', "");
Expect(0, 921600, '\p{^Is_GCB= XX}', "");
Expect(0, 921600, '\P{Is_GCB= XX}', "");
Expect(1, 921600, '\P{^Is_GCB= XX}', "");
Expect(0, 921599, '\p{Is_GCB= XX}', "");
Expect(1, 921599, '\p{^Is_GCB= XX}', "");
Expect(1, 921599, '\P{Is_GCB= XX}', "");
Expect(0, 921599, '\P{^Is_GCB= XX}', "");
Error('\p{Grapheme_Cluster_Break=/a/EXTEND}');
Error('\P{Grapheme_Cluster_Break=/a/EXTEND}');
Expect(1, 917999, '\p{Grapheme_Cluster_Break=extend}', "");
Expect(0, 917999, '\p{^Grapheme_Cluster_Break=extend}', "");
Expect(0, 917999, '\P{Grapheme_Cluster_Break=extend}', "");
Expect(1, 917999, '\P{^Grapheme_Cluster_Break=extend}', "");
Expect(0, 918000, '\p{Grapheme_Cluster_Break=extend}', "");
Expect(1, 918000, '\p{^Grapheme_Cluster_Break=extend}', "");
Expect(1, 918000, '\P{Grapheme_Cluster_Break=extend}', "");
Expect(0, 918000, '\P{^Grapheme_Cluster_Break=extend}', "");
Expect(1, 917999, '\p{Grapheme_Cluster_Break= Extend}', "");
Expect(0, 917999, '\p{^Grapheme_Cluster_Break= Extend}', "");
Expect(0, 917999, '\P{Grapheme_Cluster_Break= Extend}', "");
Expect(1, 917999, '\P{^Grapheme_Cluster_Break= Extend}', "");
Expect(0, 918000, '\p{Grapheme_Cluster_Break= Extend}', "");
Expect(1, 918000, '\p{^Grapheme_Cluster_Break= Extend}', "");
Expect(1, 918000, '\P{Grapheme_Cluster_Break= Extend}', "");
Expect(0, 918000, '\P{^Grapheme_Cluster_Break= Extend}', "");
Error('\p{GCB=:=-EX}');
Error('\P{GCB=:=-EX}');
Expect(1, 917999, '\p{GCB=ex}', "");
Expect(0, 917999, '\p{^GCB=ex}', "");
Expect(0, 917999, '\P{GCB=ex}', "");
Expect(1, 917999, '\P{^GCB=ex}', "");
Expect(0, 918000, '\p{GCB=ex}', "");
Expect(1, 918000, '\p{^GCB=ex}', "");
Expect(1, 918000, '\P{GCB=ex}', "");
Expect(0, 918000, '\P{^GCB=ex}', "");
Expect(1, 917999, '\p{GCB=	_EX}', "");
Expect(0, 917999, '\p{^GCB=	_EX}', "");
Expect(0, 917999, '\P{GCB=	_EX}', "");
Expect(1, 917999, '\P{^GCB=	_EX}', "");
Expect(0, 918000, '\p{GCB=	_EX}', "");
Expect(1, 918000, '\p{^GCB=	_EX}', "");
Expect(1, 918000, '\P{GCB=	_EX}', "");
Expect(0, 918000, '\P{^GCB=	_EX}', "");
Error('\p{Is_Grapheme_Cluster_Break=_ extend/a/}');
Error('\P{Is_Grapheme_Cluster_Break=_ extend/a/}');
Expect(1, 917999, '\p{Is_Grapheme_Cluster_Break:   extend}', "");
Expect(0, 917999, '\p{^Is_Grapheme_Cluster_Break:   extend}', "");
Expect(0, 917999, '\P{Is_Grapheme_Cluster_Break:   extend}', "");
Expect(1, 917999, '\P{^Is_Grapheme_Cluster_Break:   extend}', "");
Expect(0, 918000, '\p{Is_Grapheme_Cluster_Break:   extend}', "");
Expect(1, 918000, '\p{^Is_Grapheme_Cluster_Break:   extend}', "");
Expect(1, 918000, '\P{Is_Grapheme_Cluster_Break:   extend}', "");
Expect(0, 918000, '\P{^Is_Grapheme_Cluster_Break:   extend}', "");
Expect(1, 917999, '\p{Is_Grapheme_Cluster_Break=	_extend}', "");
Expect(0, 917999, '\p{^Is_Grapheme_Cluster_Break=	_extend}', "");
Expect(0, 917999, '\P{Is_Grapheme_Cluster_Break=	_extend}', "");
Expect(1, 917999, '\P{^Is_Grapheme_Cluster_Break=	_extend}', "");
Expect(0, 918000, '\p{Is_Grapheme_Cluster_Break=	_extend}', "");
Expect(1, 918000, '\p{^Is_Grapheme_Cluster_Break=	_extend}', "");
Expect(1, 918000, '\P{Is_Grapheme_Cluster_Break=	_extend}', "");
Expect(0, 918000, '\P{^Is_Grapheme_Cluster_Break=	_extend}', "");
Error('\p{Is_GCB=_:=ex}');
Error('\P{Is_GCB=_:=ex}');
Expect(1, 917999, '\p{Is_GCB=ex}', "");
Expect(0, 917999, '\p{^Is_GCB=ex}', "");
Expect(0, 917999, '\P{Is_GCB=ex}', "");
Expect(1, 917999, '\P{^Is_GCB=ex}', "");
Expect(0, 918000, '\p{Is_GCB=ex}', "");
Expect(1, 918000, '\p{^Is_GCB=ex}', "");
Expect(1, 918000, '\P{Is_GCB=ex}', "");
Expect(0, 918000, '\P{^Is_GCB=ex}', "");
Expect(1, 917999, '\p{Is_GCB=- EX}', "");
Expect(0, 917999, '\p{^Is_GCB=- EX}', "");
Expect(0, 917999, '\P{Is_GCB=- EX}', "");
Expect(1, 917999, '\P{^Is_GCB=- EX}', "");
Expect(0, 918000, '\p{Is_GCB=- EX}', "");
Expect(1, 918000, '\p{^Is_GCB=- EX}', "");
Expect(1, 918000, '\P{Is_GCB=- EX}', "");
Expect(0, 918000, '\P{^Is_GCB=- EX}', "");
Error('\p{Grapheme_Cluster_Break= /a/v}');
Error('\P{Grapheme_Cluster_Break= /a/v}');
Expect(1, 55238, '\p{Grapheme_Cluster_Break=v}', "");
Expect(0, 55238, '\p{^Grapheme_Cluster_Break=v}', "");
Expect(0, 55238, '\P{Grapheme_Cluster_Break=v}', "");
Expect(1, 55238, '\P{^Grapheme_Cluster_Break=v}', "");
Expect(0, 55239, '\p{Grapheme_Cluster_Break=v}', "");
Expect(1, 55239, '\p{^Grapheme_Cluster_Break=v}', "");
Expect(1, 55239, '\P{Grapheme_Cluster_Break=v}', "");
Expect(0, 55239, '\P{^Grapheme_Cluster_Break=v}', "");
Expect(1, 55238, '\p{Grapheme_Cluster_Break=-v}', "");
Expect(0, 55238, '\p{^Grapheme_Cluster_Break=-v}', "");
Expect(0, 55238, '\P{Grapheme_Cluster_Break=-v}', "");
Expect(1, 55238, '\P{^Grapheme_Cluster_Break=-v}', "");
Expect(0, 55239, '\p{Grapheme_Cluster_Break=-v}', "");
Expect(1, 55239, '\p{^Grapheme_Cluster_Break=-v}', "");
Expect(1, 55239, '\P{Grapheme_Cluster_Break=-v}', "");
Expect(0, 55239, '\P{^Grapheme_Cluster_Break=-v}', "");
Error('\p{GCB=	/a/v}');
Error('\P{GCB=	/a/v}');
Expect(1, 55238, '\p{GCB=v}', "");
Expect(0, 55238, '\p{^GCB=v}', "");
Expect(0, 55238, '\P{GCB=v}', "");
Expect(1, 55238, '\P{^GCB=v}', "");
Expect(0, 55239, '\p{GCB=v}', "");
Expect(1, 55239, '\p{^GCB=v}', "");
Expect(1, 55239, '\P{GCB=v}', "");
Expect(0, 55239, '\P{^GCB=v}', "");
Expect(1, 55238, '\p{GCB=--V}', "");
Expect(0, 55238, '\p{^GCB=--V}', "");
Expect(0, 55238, '\P{GCB=--V}', "");
Expect(1, 55238, '\P{^GCB=--V}', "");
Expect(0, 55239, '\p{GCB=--V}', "");
Expect(1, 55239, '\p{^GCB=--V}', "");
Expect(1, 55239, '\P{GCB=--V}', "");
Expect(0, 55239, '\P{^GCB=--V}', "");
Error('\p{Is_Grapheme_Cluster_Break=--V/a/}');
Error('\P{Is_Grapheme_Cluster_Break=--V/a/}');
Expect(1, 55238, '\p{Is_Grapheme_Cluster_Break=v}', "");
Expect(0, 55238, '\p{^Is_Grapheme_Cluster_Break=v}', "");
Expect(0, 55238, '\P{Is_Grapheme_Cluster_Break=v}', "");
Expect(1, 55238, '\P{^Is_Grapheme_Cluster_Break=v}', "");
Expect(0, 55239, '\p{Is_Grapheme_Cluster_Break=v}', "");
Expect(1, 55239, '\p{^Is_Grapheme_Cluster_Break=v}', "");
Expect(1, 55239, '\P{Is_Grapheme_Cluster_Break=v}', "");
Expect(0, 55239, '\P{^Is_Grapheme_Cluster_Break=v}', "");
Expect(1, 55238, '\p{Is_Grapheme_Cluster_Break=-V}', "");
Expect(0, 55238, '\p{^Is_Grapheme_Cluster_Break=-V}', "");
Expect(0, 55238, '\P{Is_Grapheme_Cluster_Break=-V}', "");
Expect(1, 55238, '\P{^Is_Grapheme_Cluster_Break=-V}', "");
Expect(0, 55239, '\p{Is_Grapheme_Cluster_Break=-V}', "");
Expect(1, 55239, '\p{^Is_Grapheme_Cluster_Break=-V}', "");
Expect(1, 55239, '\P{Is_Grapheme_Cluster_Break=-V}', "");
Expect(0, 55239, '\P{^Is_Grapheme_Cluster_Break=-V}', "");
Error('\p{Is_GCB=:=__V}');
Error('\P{Is_GCB=:=__V}');
Expect(1, 55238, '\p{Is_GCB=v}', "");
Expect(0, 55238, '\p{^Is_GCB=v}', "");
Expect(0, 55238, '\P{Is_GCB=v}', "");
Expect(1, 55238, '\P{^Is_GCB=v}', "");
Expect(0, 55239, '\p{Is_GCB=v}', "");
Expect(1, 55239, '\p{^Is_GCB=v}', "");
Expect(1, 55239, '\P{Is_GCB=v}', "");
Expect(0, 55239, '\P{^Is_GCB=v}', "");
Expect(1, 55238, '\p{Is_GCB=- V}', "");
Expect(0, 55238, '\p{^Is_GCB=- V}', "");
Expect(0, 55238, '\P{Is_GCB=- V}', "");
Expect(1, 55238, '\P{^Is_GCB=- V}', "");
Expect(0, 55239, '\p{Is_GCB=- V}', "");
Expect(1, 55239, '\p{^Is_GCB=- V}', "");
Expect(1, 55239, '\P{Is_GCB=- V}', "");
Expect(0, 55239, '\P{^Is_GCB=- V}', "");
Error('\p{Grapheme_Cluster_Break=-:=LF}');
Error('\P{Grapheme_Cluster_Break=-:=LF}');
Expect(1, 10, '\p{Grapheme_Cluster_Break:   lf}', "");
Expect(0, 10, '\p{^Grapheme_Cluster_Break:   lf}', "");
Expect(0, 10, '\P{Grapheme_Cluster_Break:   lf}', "");
Expect(1, 10, '\P{^Grapheme_Cluster_Break:   lf}', "");
Expect(0, 11, '\p{Grapheme_Cluster_Break:   lf}', "");
Expect(1, 11, '\p{^Grapheme_Cluster_Break:   lf}', "");
Expect(1, 11, '\P{Grapheme_Cluster_Break:   lf}', "");
Expect(0, 11, '\P{^Grapheme_Cluster_Break:   lf}', "");
Expect(1, 10, '\p{Grapheme_Cluster_Break=-LF}', "");
Expect(0, 10, '\p{^Grapheme_Cluster_Break=-LF}', "");
Expect(0, 10, '\P{Grapheme_Cluster_Break=-LF}', "");
Expect(1, 10, '\P{^Grapheme_Cluster_Break=-LF}', "");
Expect(0, 11, '\p{Grapheme_Cluster_Break=-LF}', "");
Expect(1, 11, '\p{^Grapheme_Cluster_Break=-LF}', "");
Expect(1, 11, '\P{Grapheme_Cluster_Break=-LF}', "");
Expect(0, 11, '\P{^Grapheme_Cluster_Break=-LF}', "");
Error('\p{GCB=/a/LF}');
Error('\P{GCB=/a/LF}');
Expect(1, 10, '\p{GCB=lf}', "");
Expect(0, 10, '\p{^GCB=lf}', "");
Expect(0, 10, '\P{GCB=lf}', "");
Expect(1, 10, '\P{^GCB=lf}', "");
Expect(0, 11, '\p{GCB=lf}', "");
Expect(1, 11, '\p{^GCB=lf}', "");
Expect(1, 11, '\P{GCB=lf}', "");
Expect(0, 11, '\P{^GCB=lf}', "");
Expect(1, 10, '\p{GCB=	lf}', "");
Expect(0, 10, '\p{^GCB=	lf}', "");
Expect(0, 10, '\P{GCB=	lf}', "");
Expect(1, 10, '\P{^GCB=	lf}', "");
Expect(0, 11, '\p{GCB=	lf}', "");
Expect(1, 11, '\p{^GCB=	lf}', "");
Expect(1, 11, '\P{GCB=	lf}', "");
Expect(0, 11, '\P{^GCB=	lf}', "");
Error('\p{Is_Grapheme_Cluster_Break=--LF:=}');
Error('\P{Is_Grapheme_Cluster_Break=--LF:=}');
Expect(1, 10, '\p{Is_Grapheme_Cluster_Break=lf}', "");
Expect(0, 10, '\p{^Is_Grapheme_Cluster_Break=lf}', "");
Expect(0, 10, '\P{Is_Grapheme_Cluster_Break=lf}', "");
Expect(1, 10, '\P{^Is_Grapheme_Cluster_Break=lf}', "");
Expect(0, 11, '\p{Is_Grapheme_Cluster_Break=lf}', "");
Expect(1, 11, '\p{^Is_Grapheme_Cluster_Break=lf}', "");
Expect(1, 11, '\P{Is_Grapheme_Cluster_Break=lf}', "");
Expect(0, 11, '\P{^Is_Grapheme_Cluster_Break=lf}', "");
Expect(1, 10, '\p{Is_Grapheme_Cluster_Break=__LF}', "");
Expect(0, 10, '\p{^Is_Grapheme_Cluster_Break=__LF}', "");
Expect(0, 10, '\P{Is_Grapheme_Cluster_Break=__LF}', "");
Expect(1, 10, '\P{^Is_Grapheme_Cluster_Break=__LF}', "");
Expect(0, 11, '\p{Is_Grapheme_Cluster_Break=__LF}', "");
Expect(1, 11, '\p{^Is_Grapheme_Cluster_Break=__LF}', "");
Expect(1, 11, '\P{Is_Grapheme_Cluster_Break=__LF}', "");
Expect(0, 11, '\P{^Is_Grapheme_Cluster_Break=__LF}', "");
Error('\p{Is_GCB=	/a/lf}');
Error('\P{Is_GCB=	/a/lf}');
Expect(1, 10, '\p{Is_GCB=lf}', "");
Expect(0, 10, '\p{^Is_GCB=lf}', "");
Expect(0, 10, '\P{Is_GCB=lf}', "");
Expect(1, 10, '\P{^Is_GCB=lf}', "");
Expect(0, 11, '\p{Is_GCB=lf}', "");
Expect(1, 11, '\p{^Is_GCB=lf}', "");
Expect(1, 11, '\P{Is_GCB=lf}', "");
Expect(0, 11, '\P{^Is_GCB=lf}', "");
Expect(1, 10, '\p{Is_GCB:	 LF}', "");
Expect(0, 10, '\p{^Is_GCB:	 LF}', "");
Expect(0, 10, '\P{Is_GCB:	 LF}', "");
Expect(1, 10, '\P{^Is_GCB:	 LF}', "");
Expect(0, 11, '\p{Is_GCB:	 LF}', "");
Expect(1, 11, '\p{^Is_GCB:	 LF}', "");
Expect(1, 11, '\P{Is_GCB:	 LF}', "");
Expect(0, 11, '\P{^Is_GCB:	 LF}', "");
Error('\p{Grapheme_Cluster_Break=-_l:=}');
Error('\P{Grapheme_Cluster_Break=-_l:=}');
Expect(1, 43388, '\p{Grapheme_Cluster_Break=l}', "");
Expect(0, 43388, '\p{^Grapheme_Cluster_Break=l}', "");
Expect(0, 43388, '\P{Grapheme_Cluster_Break=l}', "");
Expect(1, 43388, '\P{^Grapheme_Cluster_Break=l}', "");
Expect(0, 43389, '\p{Grapheme_Cluster_Break=l}', "");
Expect(1, 43389, '\p{^Grapheme_Cluster_Break=l}', "");
Expect(1, 43389, '\P{Grapheme_Cluster_Break=l}', "");
Expect(0, 43389, '\P{^Grapheme_Cluster_Break=l}', "");
Expect(1, 43388, '\p{Grapheme_Cluster_Break=- L}', "");
Expect(0, 43388, '\p{^Grapheme_Cluster_Break=- L}', "");
Expect(0, 43388, '\P{Grapheme_Cluster_Break=- L}', "");
Expect(1, 43388, '\P{^Grapheme_Cluster_Break=- L}', "");
Expect(0, 43389, '\p{Grapheme_Cluster_Break=- L}', "");
Expect(1, 43389, '\p{^Grapheme_Cluster_Break=- L}', "");
Expect(1, 43389, '\P{Grapheme_Cluster_Break=- L}', "");
Expect(0, 43389, '\P{^Grapheme_Cluster_Break=- L}', "");
Error('\p{GCB=	 l:=}');
Error('\P{GCB=	 l:=}');
Expect(1, 43388, '\p{GCB=l}', "");
Expect(0, 43388, '\p{^GCB=l}', "");
Expect(0, 43388, '\P{GCB=l}', "");
Expect(1, 43388, '\P{^GCB=l}', "");
Expect(0, 43389, '\p{GCB=l}', "");
Expect(1, 43389, '\p{^GCB=l}', "");
Expect(1, 43389, '\P{GCB=l}', "");
Expect(0, 43389, '\P{^GCB=l}', "");
Expect(1, 43388, '\p{GCB=-L}', "");
Expect(0, 43388, '\p{^GCB=-L}', "");
Expect(0, 43388, '\P{GCB=-L}', "");
Expect(1, 43388, '\P{^GCB=-L}', "");
Expect(0, 43389, '\p{GCB=-L}', "");
Expect(1, 43389, '\p{^GCB=-L}', "");
Expect(1, 43389, '\P{GCB=-L}', "");
Expect(0, 43389, '\P{^GCB=-L}', "");
Error('\p{Is_Grapheme_Cluster_Break=:=- L}');
Error('\P{Is_Grapheme_Cluster_Break=:=- L}');
Expect(1, 43388, '\p{Is_Grapheme_Cluster_Break=l}', "");
Expect(0, 43388, '\p{^Is_Grapheme_Cluster_Break=l}', "");
Expect(0, 43388, '\P{Is_Grapheme_Cluster_Break=l}', "");
Expect(1, 43388, '\P{^Is_Grapheme_Cluster_Break=l}', "");
Expect(0, 43389, '\p{Is_Grapheme_Cluster_Break=l}', "");
Expect(1, 43389, '\p{^Is_Grapheme_Cluster_Break=l}', "");
Expect(1, 43389, '\P{Is_Grapheme_Cluster_Break=l}', "");
Expect(0, 43389, '\P{^Is_Grapheme_Cluster_Break=l}', "");
Expect(1, 43388, '\p{Is_Grapheme_Cluster_Break=  L}', "");
Expect(0, 43388, '\p{^Is_Grapheme_Cluster_Break=  L}', "");
Expect(0, 43388, '\P{Is_Grapheme_Cluster_Break=  L}', "");
Expect(1, 43388, '\P{^Is_Grapheme_Cluster_Break=  L}', "");
Expect(0, 43389, '\p{Is_Grapheme_Cluster_Break=  L}', "");
Expect(1, 43389, '\p{^Is_Grapheme_Cluster_Break=  L}', "");
Expect(1, 43389, '\P{Is_Grapheme_Cluster_Break=  L}', "");
Expect(0, 43389, '\P{^Is_Grapheme_Cluster_Break=  L}', "");
Error('\p{Is_GCB:	-_L:=}');
Error('\P{Is_GCB:	-_L:=}');
Expect(1, 43388, '\p{Is_GCB=l}', "");
Expect(0, 43388, '\p{^Is_GCB=l}', "");
Expect(0, 43388, '\P{Is_GCB=l}', "");
Expect(1, 43388, '\P{^Is_GCB=l}', "");
Expect(0, 43389, '\p{Is_GCB=l}', "");
Expect(1, 43389, '\p{^Is_GCB=l}', "");
Expect(1, 43389, '\P{Is_GCB=l}', "");
Expect(0, 43389, '\P{^Is_GCB=l}', "");
Expect(1, 43388, '\p{Is_GCB=	L}', "");
Expect(0, 43388, '\p{^Is_GCB=	L}', "");
Expect(0, 43388, '\P{Is_GCB=	L}', "");
Expect(1, 43388, '\P{^Is_GCB=	L}', "");
Expect(0, 43389, '\p{Is_GCB=	L}', "");
Expect(1, 43389, '\p{^Is_GCB=	L}', "");
Expect(1, 43389, '\P{Is_GCB=	L}', "");
Expect(0, 43389, '\P{^Is_GCB=	L}', "");
Error('\p{Indic_Syllabic_Category=Consonant_Dead}');
Error('\P{Indic_Syllabic_Category=Consonant_Dead}');
Error('\p{InSC=Consonant_Dead}');
Error('\P{InSC=Consonant_Dead}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Dead}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Dead}');
Error('\p{Is_InSC: Consonant_Dead}');
Error('\P{Is_InSC: Consonant_Dead}');
Error('\p{Indic_Syllabic_Category=Avagraha}');
Error('\P{Indic_Syllabic_Category=Avagraha}');
Error('\p{InSC=Avagraha}');
Error('\P{InSC=Avagraha}');
Error('\p{Is_Indic_Syllabic_Category:   Avagraha}');
Error('\P{Is_Indic_Syllabic_Category:   Avagraha}');
Error('\p{Is_InSC=Avagraha}');
Error('\P{Is_InSC=Avagraha}');
Error('\p{Indic_Syllabic_Category=Vowel_Independent}');
Error('\P{Indic_Syllabic_Category=Vowel_Independent}');
Error('\p{InSC=Vowel_Independent}');
Error('\P{InSC=Vowel_Independent}');
Error('\p{Is_Indic_Syllabic_Category=Vowel_Independent}');
Error('\P{Is_Indic_Syllabic_Category=Vowel_Independent}');
Error('\p{Is_InSC=Vowel_Independent}');
Error('\P{Is_InSC=Vowel_Independent}');
Error('\p{Indic_Syllabic_Category=Bindu}');
Error('\P{Indic_Syllabic_Category=Bindu}');
Error('\p{InSC=Bindu}');
Error('\P{InSC=Bindu}');
Error('\p{Is_Indic_Syllabic_Category=Bindu}');
Error('\P{Is_Indic_Syllabic_Category=Bindu}');
Error('\p{Is_InSC=Bindu}');
Error('\P{Is_InSC=Bindu}');
Error('\p{Indic_Syllabic_Category:   Tone_Letter}');
Error('\P{Indic_Syllabic_Category:   Tone_Letter}');
Error('\p{InSC=Tone_Letter}');
Error('\P{InSC=Tone_Letter}');
Error('\p{Is_Indic_Syllabic_Category=Tone_Letter}');
Error('\P{Is_Indic_Syllabic_Category=Tone_Letter}');
Error('\p{Is_InSC:   Tone_Letter}');
Error('\P{Is_InSC:   Tone_Letter}');
Error('\p{Indic_Syllabic_Category=Consonant_Final}');
Error('\P{Indic_Syllabic_Category=Consonant_Final}');
Error('\p{InSC=Consonant_Final}');
Error('\P{InSC=Consonant_Final}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Final}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Final}');
Error('\p{Is_InSC=Consonant_Final}');
Error('\P{Is_InSC=Consonant_Final}');
Error('\p{Indic_Syllabic_Category=Consonant_Medial}');
Error('\P{Indic_Syllabic_Category=Consonant_Medial}');
Error('\p{InSC=Consonant_Medial}');
Error('\P{InSC=Consonant_Medial}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Medial}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Medial}');
Error('\p{Is_InSC=Consonant_Medial}');
Error('\P{Is_InSC=Consonant_Medial}');
Error('\p{Indic_Syllabic_Category=Register_Shifter}');
Error('\P{Indic_Syllabic_Category=Register_Shifter}');
Error('\p{InSC=Register_Shifter}');
Error('\P{InSC=Register_Shifter}');
Error('\p{Is_Indic_Syllabic_Category=Register_Shifter}');
Error('\P{Is_Indic_Syllabic_Category=Register_Shifter}');
Error('\p{Is_InSC=Register_Shifter}');
Error('\P{Is_InSC=Register_Shifter}');
Error('\p{Indic_Syllabic_Category=Vowel_Dependent}');
Error('\P{Indic_Syllabic_Category=Vowel_Dependent}');
Error('\p{InSC=Vowel_Dependent}');
Error('\P{InSC=Vowel_Dependent}');
Error('\p{Is_Indic_Syllabic_Category=Vowel_Dependent}');
Error('\P{Is_Indic_Syllabic_Category=Vowel_Dependent}');
Error('\p{Is_InSC=Vowel_Dependent}');
Error('\P{Is_InSC=Vowel_Dependent}');
Error('\p{Indic_Syllabic_Category=Tone_Mark}');
Error('\P{Indic_Syllabic_Category=Tone_Mark}');
Error('\p{InSC=Tone_Mark}');
Error('\P{InSC=Tone_Mark}');
Error('\p{Is_Indic_Syllabic_Category=Tone_Mark}');
Error('\P{Is_Indic_Syllabic_Category=Tone_Mark}');
Error('\p{Is_InSC=Tone_Mark}');
Error('\P{Is_InSC=Tone_Mark}');
Error('\p{Indic_Syllabic_Category=Nukta}');
Error('\P{Indic_Syllabic_Category=Nukta}');
Error('\p{InSC: Nukta}');
Error('\P{InSC: Nukta}');
Error('\p{Is_Indic_Syllabic_Category=Nukta}');
Error('\P{Is_Indic_Syllabic_Category=Nukta}');
Error('\p{Is_InSC=Nukta}');
Error('\P{Is_InSC=Nukta}');
Error('\p{Indic_Syllabic_Category=Consonant_Repha}');
Error('\P{Indic_Syllabic_Category=Consonant_Repha}');
Error('\p{InSC=Consonant_Repha}');
Error('\P{InSC=Consonant_Repha}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Repha}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Repha}');
Error('\p{Is_InSC=Consonant_Repha}');
Error('\P{Is_InSC=Consonant_Repha}');
Error('\p{Indic_Syllabic_Category: Consonant_Subjoined}');
Error('\P{Indic_Syllabic_Category: Consonant_Subjoined}');
Error('\p{InSC=Consonant_Subjoined}');
Error('\P{InSC=Consonant_Subjoined}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Subjoined}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Subjoined}');
Error('\p{Is_InSC=Consonant_Subjoined}');
Error('\P{Is_InSC=Consonant_Subjoined}');
Error('\p{Indic_Syllabic_Category:Virama}');
Error('\P{Indic_Syllabic_Category:Virama}');
Error('\p{InSC=Virama}');
Error('\P{InSC=Virama}');
Error('\p{Is_Indic_Syllabic_Category:Virama}');
Error('\P{Is_Indic_Syllabic_Category:Virama}');
Error('\p{Is_InSC=Virama}');
Error('\P{Is_InSC=Virama}');
Error('\p{Indic_Syllabic_Category=Consonant}');
Error('\P{Indic_Syllabic_Category=Consonant}');
Error('\p{InSC=Consonant}');
Error('\P{InSC=Consonant}');
Error('\p{Is_Indic_Syllabic_Category=Consonant}');
Error('\P{Is_Indic_Syllabic_Category=Consonant}');
Error('\p{Is_InSC=Consonant}');
Error('\P{Is_InSC=Consonant}');
Error('\p{Indic_Syllabic_Category=Consonant_Head_Letter}');
Error('\P{Indic_Syllabic_Category=Consonant_Head_Letter}');
Error('\p{InSC=Consonant_Head_Letter}');
Error('\P{InSC=Consonant_Head_Letter}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Head_Letter}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Head_Letter}');
Error('\p{Is_InSC=Consonant_Head_Letter}');
Error('\P{Is_InSC=Consonant_Head_Letter}');
Error('\p{Indic_Syllabic_Category=Visarga}');
Error('\P{Indic_Syllabic_Category=Visarga}');
Error('\p{InSC=Visarga}');
Error('\P{InSC=Visarga}');
Error('\p{Is_Indic_Syllabic_Category=Visarga}');
Error('\P{Is_Indic_Syllabic_Category=Visarga}');
Error('\p{Is_InSC=Visarga}');
Error('\P{Is_InSC=Visarga}');
Error('\p{Indic_Syllabic_Category=Modifying_Letter}');
Error('\P{Indic_Syllabic_Category=Modifying_Letter}');
Error('\p{InSC=Modifying_Letter}');
Error('\P{InSC=Modifying_Letter}');
Error('\p{Is_Indic_Syllabic_Category=Modifying_Letter}');
Error('\P{Is_Indic_Syllabic_Category=Modifying_Letter}');
Error('\p{Is_InSC=Modifying_Letter}');
Error('\P{Is_InSC=Modifying_Letter}');
Error('\p{Indic_Syllabic_Category=Other}');
Error('\P{Indic_Syllabic_Category=Other}');
Error('\p{InSC=Other}');
Error('\P{InSC=Other}');
Error('\p{Is_Indic_Syllabic_Category=Other}');
Error('\P{Is_Indic_Syllabic_Category=Other}');
Error('\p{Is_InSC=Other}');
Error('\P{Is_InSC=Other}');
Error('\p{Indic_Syllabic_Category=Consonant_Placeholder}');
Error('\P{Indic_Syllabic_Category=Consonant_Placeholder}');
Error('\p{InSC=Consonant_Placeholder}');
Error('\P{InSC=Consonant_Placeholder}');
Error('\p{Is_Indic_Syllabic_Category=Consonant_Placeholder}');
Error('\P{Is_Indic_Syllabic_Category=Consonant_Placeholder}');
Error('\p{Is_InSC=Consonant_Placeholder}');
Error('\P{Is_InSC=Consonant_Placeholder}');
Error('\p{Indic_Syllabic_Category=Vowel}');
Error('\P{Indic_Syllabic_Category=Vowel}');
Error('\p{InSC=Vowel}');
Error('\P{InSC=Vowel}');
Error('\p{Is_Indic_Syllabic_Category=Vowel}');
Error('\P{Is_Indic_Syllabic_Category=Vowel}');
Error('\p{Is_InSC:	Vowel}');
Error('\P{Is_InSC:	Vowel}');
Error('\p{Default_Ignorable_Code_Point=/a/Yes}');
Error('\P{Default_Ignorable_Code_Point=/a/Yes}');
Expect(1, 921599, '\p{Default_Ignorable_Code_Point=yes}', "");
Expect(0, 921599, '\p{^Default_Ignorable_Code_Point=yes}', "");
Expect(0, 921599, '\P{Default_Ignorable_Code_Point=yes}', "");
Expect(1, 921599, '\P{^Default_Ignorable_Code_Point=yes}', "");
Expect(0, 921600, '\p{Default_Ignorable_Code_Point=yes}', "");
Expect(1, 921600, '\p{^Default_Ignorable_Code_Point=yes}', "");
Expect(1, 921600, '\P{Default_Ignorable_Code_Point=yes}', "");
Expect(0, 921600, '\P{^Default_Ignorable_Code_Point=yes}', "");
Expect(1, 921599, '\p{Default_Ignorable_Code_Point= -Yes}', "");
Expect(0, 921599, '\p{^Default_Ignorable_Code_Point= -Yes}', "");
Expect(0, 921599, '\P{Default_Ignorable_Code_Point= -Yes}', "");
Expect(1, 921599, '\P{^Default_Ignorable_Code_Point= -Yes}', "");
Expect(0, 921600, '\p{Default_Ignorable_Code_Point= -Yes}', "");
Expect(1, 921600, '\p{^Default_Ignorable_Code_Point= -Yes}', "");
Expect(1, 921600, '\P{Default_Ignorable_Code_Point= -Yes}', "");
Expect(0, 921600, '\P{^Default_Ignorable_Code_Point= -Yes}', "");
Error('\p{DI=/a/_	Y}');
Error('\P{DI=/a/_	Y}');
Expect(1, 921599, '\p{DI=y}', "");
Expect(0, 921599, '\p{^DI=y}', "");
Expect(0, 921599, '\P{DI=y}', "");
Expect(1, 921599, '\P{^DI=y}', "");
Expect(0, 921600, '\p{DI=y}', "");
Expect(1, 921600, '\p{^DI=y}', "");
Expect(1, 921600, '\P{DI=y}', "");
Expect(0, 921600, '\P{^DI=y}', "");
Expect(1, 921599, '\p{DI=- Y}', "");
Expect(0, 921599, '\p{^DI=- Y}', "");
Expect(0, 921599, '\P{DI=- Y}', "");
Expect(1, 921599, '\P{^DI=- Y}', "");
Expect(0, 921600, '\p{DI=- Y}', "");
Expect(1, 921600, '\p{^DI=- Y}', "");
Expect(1, 921600, '\P{DI=- Y}', "");
Expect(0, 921600, '\P{^DI=- Y}', "");
Error('\p{Is_Default_Ignorable_Code_Point=:=-T}');
Error('\P{Is_Default_Ignorable_Code_Point=:=-T}');
Expect(1, 921599, '\p{Is_Default_Ignorable_Code_Point=t}', "");
Expect(0, 921599, '\p{^Is_Default_Ignorable_Code_Point=t}', "");
Expect(0, 921599, '\P{Is_Default_Ignorable_Code_Point=t}', "");
Expect(1, 921599, '\P{^Is_Default_Ignorable_Code_Point=t}', "");
Expect(0, 921600, '\p{Is_Default_Ignorable_Code_Point=t}', "");
Expect(1, 921600, '\p{^Is_Default_Ignorable_Code_Point=t}', "");
Expect(1, 921600, '\P{Is_Default_Ignorable_Code_Point=t}', "");
Expect(0, 921600, '\P{^Is_Default_Ignorable_Code_Point=t}', "");
Expect(1, 921599, '\p{Is_Default_Ignorable_Code_Point= _T}', "");
Expect(0, 921599, '\p{^Is_Default_Ignorable_Code_Point= _T}', "");
Expect(0, 921599, '\P{Is_Default_Ignorable_Code_Point= _T}', "");
Expect(1, 921599, '\P{^Is_Default_Ignorable_Code_Point= _T}', "");
Expect(0, 921600, '\p{Is_Default_Ignorable_Code_Point= _T}', "");
Expect(1, 921600, '\p{^Is_Default_Ignorable_Code_Point= _T}', "");
Expect(1, 921600, '\P{Is_Default_Ignorable_Code_Point= _T}', "");
Expect(0, 921600, '\P{^Is_Default_Ignorable_Code_Point= _T}', "");
Error('\p{Is_DI=_	True/a/}');
Error('\P{Is_DI=_	True/a/}');
Expect(1, 921599, '\p{Is_DI=true}', "");
Expect(0, 921599, '\p{^Is_DI=true}', "");
Expect(0, 921599, '\P{Is_DI=true}', "");
Expect(1, 921599, '\P{^Is_DI=true}', "");
Expect(0, 921600, '\p{Is_DI=true}', "");
Expect(1, 921600, '\p{^Is_DI=true}', "");
Expect(1, 921600, '\P{Is_DI=true}', "");
Expect(0, 921600, '\P{^Is_DI=true}', "");
Expect(1, 921599, '\p{Is_DI=--True}', "");
Expect(0, 921599, '\p{^Is_DI=--True}', "");
Expect(0, 921599, '\P{Is_DI=--True}', "");
Expect(1, 921599, '\P{^Is_DI=--True}', "");
Expect(0, 921600, '\p{Is_DI=--True}', "");
Expect(1, 921600, '\p{^Is_DI=--True}', "");
Expect(1, 921600, '\P{Is_DI=--True}', "");
Expect(0, 921600, '\P{^Is_DI=--True}', "");
Error('\p{Default_Ignorable_Code_Point=-_No:=}');
Error('\P{Default_Ignorable_Code_Point=-_No:=}');
Expect(1, 921600, '\p{Default_Ignorable_Code_Point=no}', "");
Expect(0, 921600, '\p{^Default_Ignorable_Code_Point=no}', "");
Expect(0, 921600, '\P{Default_Ignorable_Code_Point=no}', "");
Expect(1, 921600, '\P{^Default_Ignorable_Code_Point=no}', "");
Expect(0, 921599, '\p{Default_Ignorable_Code_Point=no}', "");
Expect(1, 921599, '\p{^Default_Ignorable_Code_Point=no}', "");
Expect(1, 921599, '\P{Default_Ignorable_Code_Point=no}', "");
Expect(0, 921599, '\P{^Default_Ignorable_Code_Point=no}', "");
Expect(1, 921600, '\p{Default_Ignorable_Code_Point=-_No}', "");
Expect(0, 921600, '\p{^Default_Ignorable_Code_Point=-_No}', "");
Expect(0, 921600, '\P{Default_Ignorable_Code_Point=-_No}', "");
Expect(1, 921600, '\P{^Default_Ignorable_Code_Point=-_No}', "");
Expect(0, 921599, '\p{Default_Ignorable_Code_Point=-_No}', "");
Expect(1, 921599, '\p{^Default_Ignorable_Code_Point=-_No}', "");
Expect(1, 921599, '\P{Default_Ignorable_Code_Point=-_No}', "");
Expect(0, 921599, '\P{^Default_Ignorable_Code_Point=-_No}', "");
Error('\p{DI=/a/-	N}');
Error('\P{DI=/a/-	N}');
Expect(1, 921600, '\p{DI=n}', "");
Expect(0, 921600, '\p{^DI=n}', "");
Expect(0, 921600, '\P{DI=n}', "");
Expect(1, 921600, '\P{^DI=n}', "");
Expect(0, 921599, '\p{DI=n}', "");
Expect(1, 921599, '\p{^DI=n}', "");
Expect(1, 921599, '\P{DI=n}', "");
Expect(0, 921599, '\P{^DI=n}', "");
Expect(1, 921600, '\p{DI:	 n}', "");
Expect(0, 921600, '\p{^DI:	 n}', "");
Expect(0, 921600, '\P{DI:	 n}', "");
Expect(1, 921600, '\P{^DI:	 n}', "");
Expect(0, 921599, '\p{DI:	 n}', "");
Expect(1, 921599, '\p{^DI:	 n}', "");
Expect(1, 921599, '\P{DI:	 n}', "");
Expect(0, 921599, '\P{^DI:	 n}', "");
Error('\p{Is_Default_Ignorable_Code_Point= :=F}');
Error('\P{Is_Default_Ignorable_Code_Point= :=F}');
Expect(1, 921600, '\p{Is_Default_Ignorable_Code_Point=f}', "");
Expect(0, 921600, '\p{^Is_Default_Ignorable_Code_Point=f}', "");
Expect(0, 921600, '\P{Is_Default_Ignorable_Code_Point=f}', "");
Expect(1, 921600, '\P{^Is_Default_Ignorable_Code_Point=f}', "");
Expect(0, 921599, '\p{Is_Default_Ignorable_Code_Point=f}', "");
Expect(1, 921599, '\p{^Is_Default_Ignorable_Code_Point=f}', "");
Expect(1, 921599, '\P{Is_Default_Ignorable_Code_Point=f}', "");
Expect(0, 921599, '\P{^Is_Default_Ignorable_Code_Point=f}', "");
Expect(1, 921600, '\p{Is_Default_Ignorable_Code_Point=--f}', "");
Expect(0, 921600, '\p{^Is_Default_Ignorable_Code_Point=--f}', "");
Expect(0, 921600, '\P{Is_Default_Ignorable_Code_Point=--f}', "");
Expect(1, 921600, '\P{^Is_Default_Ignorable_Code_Point=--f}', "");
Expect(0, 921599, '\p{Is_Default_Ignorable_Code_Point=--f}', "");
Expect(1, 921599, '\p{^Is_Default_Ignorable_Code_Point=--f}', "");
Expect(1, 921599, '\P{Is_Default_Ignorable_Code_Point=--f}', "");
Expect(0, 921599, '\P{^Is_Default_Ignorable_Code_Point=--f}', "");
Error('\p{Is_DI=:=	False}');
Error('\P{Is_DI=:=	False}');
Expect(1, 921600, '\p{Is_DI=false}', "");
Expect(0, 921600, '\p{^Is_DI=false}', "");
Expect(0, 921600, '\P{Is_DI=false}', "");
Expect(1, 921600, '\P{^Is_DI=false}', "");
Expect(0, 921599, '\p{Is_DI=false}', "");
Expect(1, 921599, '\p{^Is_DI=false}', "");
Expect(1, 921599, '\P{Is_DI=false}', "");
Expect(0, 921599, '\P{^Is_DI=false}', "");
Expect(1, 921600, '\p{Is_DI= 	FALSE}', "");
Expect(0, 921600, '\p{^Is_DI= 	FALSE}', "");
Expect(0, 921600, '\P{Is_DI= 	FALSE}', "");
Expect(1, 921600, '\P{^Is_DI= 	FALSE}', "");
Expect(0, 921599, '\p{Is_DI= 	FALSE}', "");
Expect(1, 921599, '\p{^Is_DI= 	FALSE}', "");
Expect(1, 921599, '\P{Is_DI= 	FALSE}', "");
Expect(0, 921599, '\P{^Is_DI= 	FALSE}', "");
Error('\p{ASCII_Hex_Digit:_:=yes}');
Error('\P{ASCII_Hex_Digit:_:=yes}');
Expect(1, 102, '\p{ASCII_Hex_Digit=yes}', "");
Expect(0, 102, '\p{^ASCII_Hex_Digit=yes}', "");
Expect(0, 102, '\P{ASCII_Hex_Digit=yes}', "");
Expect(1, 102, '\P{^ASCII_Hex_Digit=yes}', "");
Expect(0, 103, '\p{ASCII_Hex_Digit=yes}', "");
Expect(1, 103, '\p{^ASCII_Hex_Digit=yes}', "");
Expect(1, 103, '\P{ASCII_Hex_Digit=yes}', "");
Expect(0, 103, '\P{^ASCII_Hex_Digit=yes}', "");
Expect(1, 102, '\p{ASCII_Hex_Digit= YES}', "");
Expect(0, 102, '\p{^ASCII_Hex_Digit= YES}', "");
Expect(0, 102, '\P{ASCII_Hex_Digit= YES}', "");
Expect(1, 102, '\P{^ASCII_Hex_Digit= YES}', "");
Expect(0, 103, '\p{ASCII_Hex_Digit= YES}', "");
Expect(1, 103, '\p{^ASCII_Hex_Digit= YES}', "");
Expect(1, 103, '\P{ASCII_Hex_Digit= YES}', "");
Expect(0, 103, '\P{^ASCII_Hex_Digit= YES}', "");
Error('\p{AHex: /a/ _Y}');
Error('\P{AHex: /a/ _Y}');
Expect(1, 102, '\p{AHex=y}', "");
Expect(0, 102, '\p{^AHex=y}', "");
Expect(0, 102, '\P{AHex=y}', "");
Expect(1, 102, '\P{^AHex=y}', "");
Expect(0, 103, '\p{AHex=y}', "");
Expect(1, 103, '\p{^AHex=y}', "");
Expect(1, 103, '\P{AHex=y}', "");
Expect(0, 103, '\P{^AHex=y}', "");
Expect(1, 102, '\p{AHex= -y}', "");
Expect(0, 102, '\p{^AHex= -y}', "");
Expect(0, 102, '\P{AHex= -y}', "");
Expect(1, 102, '\P{^AHex= -y}', "");
Expect(0, 103, '\p{AHex= -y}', "");
Expect(1, 103, '\p{^AHex= -y}', "");
Expect(1, 103, '\P{AHex= -y}', "");
Expect(0, 103, '\P{^AHex= -y}', "");
Error('\p{Is_ASCII_Hex_Digit= /a/T}');
Error('\P{Is_ASCII_Hex_Digit= /a/T}');
Expect(1, 102, '\p{Is_ASCII_Hex_Digit=t}', "");
Expect(0, 102, '\p{^Is_ASCII_Hex_Digit=t}', "");
Expect(0, 102, '\P{Is_ASCII_Hex_Digit=t}', "");
Expect(1, 102, '\P{^Is_ASCII_Hex_Digit=t}', "");
Expect(0, 103, '\p{Is_ASCII_Hex_Digit=t}', "");
Expect(1, 103, '\p{^Is_ASCII_Hex_Digit=t}', "");
Expect(1, 103, '\P{Is_ASCII_Hex_Digit=t}', "");
Expect(0, 103, '\P{^Is_ASCII_Hex_Digit=t}', "");
Expect(1, 102, '\p{Is_ASCII_Hex_Digit=	T}', "");
Expect(0, 102, '\p{^Is_ASCII_Hex_Digit=	T}', "");
Expect(0, 102, '\P{Is_ASCII_Hex_Digit=	T}', "");
Expect(1, 102, '\P{^Is_ASCII_Hex_Digit=	T}', "");
Expect(0, 103, '\p{Is_ASCII_Hex_Digit=	T}', "");
Expect(1, 103, '\p{^Is_ASCII_Hex_Digit=	T}', "");
Expect(1, 103, '\P{Is_ASCII_Hex_Digit=	T}', "");
Expect(0, 103, '\P{^Is_ASCII_Hex_Digit=	T}', "");
Error('\p{Is_AHex=/a/_true}');
Error('\P{Is_AHex=/a/_true}');
Expect(1, 102, '\p{Is_AHex: true}', "");
Expect(0, 102, '\p{^Is_AHex: true}', "");
Expect(0, 102, '\P{Is_AHex: true}', "");
Expect(1, 102, '\P{^Is_AHex: true}', "");
Expect(0, 103, '\p{Is_AHex: true}', "");
Expect(1, 103, '\p{^Is_AHex: true}', "");
Expect(1, 103, '\P{Is_AHex: true}', "");
Expect(0, 103, '\P{^Is_AHex: true}', "");
Expect(1, 102, '\p{Is_AHex=_	True}', "");
Expect(0, 102, '\p{^Is_AHex=_	True}', "");
Expect(0, 102, '\P{Is_AHex=_	True}', "");
Expect(1, 102, '\P{^Is_AHex=_	True}', "");
Expect(0, 103, '\p{Is_AHex=_	True}', "");
Expect(1, 103, '\p{^Is_AHex=_	True}', "");
Expect(1, 103, '\P{Is_AHex=_	True}', "");
Expect(0, 103, '\P{^Is_AHex=_	True}', "");
Error('\p{ASCII_Hex_Digit=/a/	-NO}');
Error('\P{ASCII_Hex_Digit=/a/	-NO}');
Expect(1, 103, '\p{ASCII_Hex_Digit=no}', "");
Expect(0, 103, '\p{^ASCII_Hex_Digit=no}', "");
Expect(0, 103, '\P{ASCII_Hex_Digit=no}', "");
Expect(1, 103, '\P{^ASCII_Hex_Digit=no}', "");
Expect(0, 102, '\p{ASCII_Hex_Digit=no}', "");
Expect(1, 102, '\p{^ASCII_Hex_Digit=no}', "");
Expect(1, 102, '\P{ASCII_Hex_Digit=no}', "");
Expect(0, 102, '\P{^ASCII_Hex_Digit=no}', "");
Expect(1, 103, '\p{ASCII_Hex_Digit=__no}', "");
Expect(0, 103, '\p{^ASCII_Hex_Digit=__no}', "");
Expect(0, 103, '\P{ASCII_Hex_Digit=__no}', "");
Expect(1, 103, '\P{^ASCII_Hex_Digit=__no}', "");
Expect(0, 102, '\p{ASCII_Hex_Digit=__no}', "");
Expect(1, 102, '\p{^ASCII_Hex_Digit=__no}', "");
Expect(1, 102, '\P{ASCII_Hex_Digit=__no}', "");
Expect(0, 102, '\P{^ASCII_Hex_Digit=__no}', "");
Error('\p{AHex= _N:=}');
Error('\P{AHex= _N:=}');
Expect(1, 103, '\p{AHex=n}', "");
Expect(0, 103, '\p{^AHex=n}', "");
Expect(0, 103, '\P{AHex=n}', "");
Expect(1, 103, '\P{^AHex=n}', "");
Expect(0, 102, '\p{AHex=n}', "");
Expect(1, 102, '\p{^AHex=n}', "");
Expect(1, 102, '\P{AHex=n}', "");
Expect(0, 102, '\P{^AHex=n}', "");
Expect(1, 103, '\p{AHex=_	N}', "");
Expect(0, 103, '\p{^AHex=_	N}', "");
Expect(0, 103, '\P{AHex=_	N}', "");
Expect(1, 103, '\P{^AHex=_	N}', "");
Expect(0, 102, '\p{AHex=_	N}', "");
Expect(1, 102, '\p{^AHex=_	N}', "");
Expect(1, 102, '\P{AHex=_	N}', "");
Expect(0, 102, '\P{^AHex=_	N}', "");
Error('\p{Is_ASCII_Hex_Digit:	-f/a/}');
Error('\P{Is_ASCII_Hex_Digit:	-f/a/}');
Expect(1, 103, '\p{Is_ASCII_Hex_Digit=f}', "");
Expect(0, 103, '\p{^Is_ASCII_Hex_Digit=f}', "");
Expect(0, 103, '\P{Is_ASCII_Hex_Digit=f}', "");
Expect(1, 103, '\P{^Is_ASCII_Hex_Digit=f}', "");
Expect(0, 102, '\p{Is_ASCII_Hex_Digit=f}', "");
Expect(1, 102, '\p{^Is_ASCII_Hex_Digit=f}', "");
Expect(1, 102, '\P{Is_ASCII_Hex_Digit=f}', "");
Expect(0, 102, '\P{^Is_ASCII_Hex_Digit=f}', "");
Expect(1, 103, '\p{Is_ASCII_Hex_Digit=-_F}', "");
Expect(0, 103, '\p{^Is_ASCII_Hex_Digit=-_F}', "");
Expect(0, 103, '\P{Is_ASCII_Hex_Digit=-_F}', "");
Expect(1, 103, '\P{^Is_ASCII_Hex_Digit=-_F}', "");
Expect(0, 102, '\p{Is_ASCII_Hex_Digit=-_F}', "");
Expect(1, 102, '\p{^Is_ASCII_Hex_Digit=-_F}', "");
Expect(1, 102, '\P{Is_ASCII_Hex_Digit=-_F}', "");
Expect(0, 102, '\P{^Is_ASCII_Hex_Digit=-_F}', "");
Error('\p{Is_AHex:	_-false:=}');
Error('\P{Is_AHex:	_-false:=}');
Expect(1, 103, '\p{Is_AHex:   false}', "");
Expect(0, 103, '\p{^Is_AHex:   false}', "");
Expect(0, 103, '\P{Is_AHex:   false}', "");
Expect(1, 103, '\P{^Is_AHex:   false}', "");
Expect(0, 102, '\p{Is_AHex:   false}', "");
Expect(1, 102, '\p{^Is_AHex:   false}', "");
Expect(1, 102, '\P{Is_AHex:   false}', "");
Expect(0, 102, '\P{^Is_AHex:   false}', "");
Expect(1, 103, '\p{Is_AHex=__FALSE}', "");
Expect(0, 103, '\p{^Is_AHex=__FALSE}', "");
Expect(0, 103, '\P{Is_AHex=__FALSE}', "");
Expect(1, 103, '\P{^Is_AHex=__FALSE}', "");
Expect(0, 102, '\p{Is_AHex=__FALSE}', "");
Expect(1, 102, '\p{^Is_AHex=__FALSE}', "");
Expect(1, 102, '\P{Is_AHex=__FALSE}', "");
Expect(0, 102, '\P{^Is_AHex=__FALSE}', "");
Error('\p{NFD_Quick_Check=-:=Yes}');
Error('\P{NFD_Quick_Check=-:=Yes}');
Expect(1, 195102, '\p{NFD_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^NFD_Quick_Check=yes}', "");
Expect(0, 195102, '\P{NFD_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^NFD_Quick_Check=yes}', "");
Expect(0, 195101, '\p{NFD_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^NFD_Quick_Check=yes}', "");
Expect(1, 195101, '\P{NFD_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^NFD_Quick_Check=yes}', "");
Expect(1, 195102, '\p{NFD_Quick_Check=--yes}', "");
Expect(0, 195102, '\p{^NFD_Quick_Check=--yes}', "");
Expect(0, 195102, '\P{NFD_Quick_Check=--yes}', "");
Expect(1, 195102, '\P{^NFD_Quick_Check=--yes}', "");
Expect(0, 195101, '\p{NFD_Quick_Check=--yes}', "");
Expect(1, 195101, '\p{^NFD_Quick_Check=--yes}', "");
Expect(1, 195101, '\P{NFD_Quick_Check=--yes}', "");
Expect(0, 195101, '\P{^NFD_Quick_Check=--yes}', "");
Error('\p{NFD_QC:/a/ Y}');
Error('\P{NFD_QC:/a/ Y}');
Expect(1, 195102, '\p{NFD_QC=y}', "");
Expect(0, 195102, '\p{^NFD_QC=y}', "");
Expect(0, 195102, '\P{NFD_QC=y}', "");
Expect(1, 195102, '\P{^NFD_QC=y}', "");
Expect(0, 195101, '\p{NFD_QC=y}', "");
Expect(1, 195101, '\p{^NFD_QC=y}', "");
Expect(1, 195101, '\P{NFD_QC=y}', "");
Expect(0, 195101, '\P{^NFD_QC=y}', "");
Expect(1, 195102, '\p{NFD_QC=__y}', "");
Expect(0, 195102, '\p{^NFD_QC=__y}', "");
Expect(0, 195102, '\P{NFD_QC=__y}', "");
Expect(1, 195102, '\P{^NFD_QC=__y}', "");
Expect(0, 195101, '\p{NFD_QC=__y}', "");
Expect(1, 195101, '\p{^NFD_QC=__y}', "");
Expect(1, 195101, '\P{NFD_QC=__y}', "");
Expect(0, 195101, '\P{^NFD_QC=__y}', "");
Error('\p{Is_NFD_Quick_Check=_ Yes:=}');
Error('\P{Is_NFD_Quick_Check=_ Yes:=}');
Expect(1, 195102, '\p{Is_NFD_Quick_Check:   yes}', "");
Expect(0, 195102, '\p{^Is_NFD_Quick_Check:   yes}', "");
Expect(0, 195102, '\P{Is_NFD_Quick_Check:   yes}', "");
Expect(1, 195102, '\P{^Is_NFD_Quick_Check:   yes}', "");
Expect(0, 195101, '\p{Is_NFD_Quick_Check:   yes}', "");
Expect(1, 195101, '\p{^Is_NFD_Quick_Check:   yes}', "");
Expect(1, 195101, '\P{Is_NFD_Quick_Check:   yes}', "");
Expect(0, 195101, '\P{^Is_NFD_Quick_Check:   yes}', "");
Expect(1, 195102, '\p{Is_NFD_Quick_Check=--Yes}', "");
Expect(0, 195102, '\p{^Is_NFD_Quick_Check=--Yes}', "");
Expect(0, 195102, '\P{Is_NFD_Quick_Check=--Yes}', "");
Expect(1, 195102, '\P{^Is_NFD_Quick_Check=--Yes}', "");
Expect(0, 195101, '\p{Is_NFD_Quick_Check=--Yes}', "");
Expect(1, 195101, '\p{^Is_NFD_Quick_Check=--Yes}', "");
Expect(1, 195101, '\P{Is_NFD_Quick_Check=--Yes}', "");
Expect(0, 195101, '\P{^Is_NFD_Quick_Check=--Yes}', "");
Error('\p{Is_NFD_QC= _Y/a/}');
Error('\P{Is_NFD_QC= _Y/a/}');
Expect(1, 195102, '\p{Is_NFD_QC=y}', "");
Expect(0, 195102, '\p{^Is_NFD_QC=y}', "");
Expect(0, 195102, '\P{Is_NFD_QC=y}', "");
Expect(1, 195102, '\P{^Is_NFD_QC=y}', "");
Expect(0, 195101, '\p{Is_NFD_QC=y}', "");
Expect(1, 195101, '\p{^Is_NFD_QC=y}', "");
Expect(1, 195101, '\P{Is_NFD_QC=y}', "");
Expect(0, 195101, '\P{^Is_NFD_QC=y}', "");
Expect(1, 195102, '\p{Is_NFD_QC= _Y}', "");
Expect(0, 195102, '\p{^Is_NFD_QC= _Y}', "");
Expect(0, 195102, '\P{Is_NFD_QC= _Y}', "");
Expect(1, 195102, '\P{^Is_NFD_QC= _Y}', "");
Expect(0, 195101, '\p{Is_NFD_QC= _Y}', "");
Expect(1, 195101, '\p{^Is_NFD_QC= _Y}', "");
Expect(1, 195101, '\P{Is_NFD_QC= _Y}', "");
Expect(0, 195101, '\P{^Is_NFD_QC= _Y}', "");
Error('\p{NFD_Quick_Check=-no/a/}');
Error('\P{NFD_Quick_Check=-no/a/}');
Expect(1, 195101, '\p{NFD_Quick_Check=no}', "");
Expect(0, 195101, '\p{^NFD_Quick_Check=no}', "");
Expect(0, 195101, '\P{NFD_Quick_Check=no}', "");
Expect(1, 195101, '\P{^NFD_Quick_Check=no}', "");
Expect(0, 195102, '\p{NFD_Quick_Check=no}', "");
Expect(1, 195102, '\p{^NFD_Quick_Check=no}', "");
Expect(1, 195102, '\P{NFD_Quick_Check=no}', "");
Expect(0, 195102, '\P{^NFD_Quick_Check=no}', "");
Expect(1, 195101, '\p{NFD_Quick_Check=	-No}', "");
Expect(0, 195101, '\p{^NFD_Quick_Check=	-No}', "");
Expect(0, 195101, '\P{NFD_Quick_Check=	-No}', "");
Expect(1, 195101, '\P{^NFD_Quick_Check=	-No}', "");
Expect(0, 195102, '\p{NFD_Quick_Check=	-No}', "");
Expect(1, 195102, '\p{^NFD_Quick_Check=	-No}', "");
Expect(1, 195102, '\P{NFD_Quick_Check=	-No}', "");
Expect(0, 195102, '\P{^NFD_Quick_Check=	-No}', "");
Error('\p{NFD_QC=	_n/a/}');
Error('\P{NFD_QC=	_n/a/}');
Expect(1, 195101, '\p{NFD_QC=n}', "");
Expect(0, 195101, '\p{^NFD_QC=n}', "");
Expect(0, 195101, '\P{NFD_QC=n}', "");
Expect(1, 195101, '\P{^NFD_QC=n}', "");
Expect(0, 195102, '\p{NFD_QC=n}', "");
Expect(1, 195102, '\p{^NFD_QC=n}', "");
Expect(1, 195102, '\P{NFD_QC=n}', "");
Expect(0, 195102, '\P{^NFD_QC=n}', "");
Expect(1, 195101, '\p{NFD_QC=	N}', "");
Expect(0, 195101, '\p{^NFD_QC=	N}', "");
Expect(0, 195101, '\P{NFD_QC=	N}', "");
Expect(1, 195101, '\P{^NFD_QC=	N}', "");
Expect(0, 195102, '\p{NFD_QC=	N}', "");
Expect(1, 195102, '\p{^NFD_QC=	N}', "");
Expect(1, 195102, '\P{NFD_QC=	N}', "");
Expect(0, 195102, '\P{^NFD_QC=	N}', "");
Error('\p{Is_NFD_Quick_Check:   		No/a/}');
Error('\P{Is_NFD_Quick_Check:   		No/a/}');
Expect(1, 195101, '\p{Is_NFD_Quick_Check=no}', "");
Expect(0, 195101, '\p{^Is_NFD_Quick_Check=no}', "");
Expect(0, 195101, '\P{Is_NFD_Quick_Check=no}', "");
Expect(1, 195101, '\P{^Is_NFD_Quick_Check=no}', "");
Expect(0, 195102, '\p{Is_NFD_Quick_Check=no}', "");
Expect(1, 195102, '\p{^Is_NFD_Quick_Check=no}', "");
Expect(1, 195102, '\P{Is_NFD_Quick_Check=no}', "");
Expect(0, 195102, '\P{^Is_NFD_Quick_Check=no}', "");
Expect(1, 195101, '\p{Is_NFD_Quick_Check=-	no}', "");
Expect(0, 195101, '\p{^Is_NFD_Quick_Check=-	no}', "");
Expect(0, 195101, '\P{Is_NFD_Quick_Check=-	no}', "");
Expect(1, 195101, '\P{^Is_NFD_Quick_Check=-	no}', "");
Expect(0, 195102, '\p{Is_NFD_Quick_Check=-	no}', "");
Expect(1, 195102, '\p{^Is_NFD_Quick_Check=-	no}', "");
Expect(1, 195102, '\P{Is_NFD_Quick_Check=-	no}', "");
Expect(0, 195102, '\P{^Is_NFD_Quick_Check=-	no}', "");
Error('\p{Is_NFD_QC=:=  n}');
Error('\P{Is_NFD_QC=:=  n}');
Expect(1, 195101, '\p{Is_NFD_QC=n}', "");
Expect(0, 195101, '\p{^Is_NFD_QC=n}', "");
Expect(0, 195101, '\P{Is_NFD_QC=n}', "");
Expect(1, 195101, '\P{^Is_NFD_QC=n}', "");
Expect(0, 195102, '\p{Is_NFD_QC=n}', "");
Expect(1, 195102, '\p{^Is_NFD_QC=n}', "");
Expect(1, 195102, '\P{Is_NFD_QC=n}', "");
Expect(0, 195102, '\P{^Is_NFD_QC=n}', "");
Expect(1, 195101, '\p{Is_NFD_QC=_ n}', "");
Expect(0, 195101, '\p{^Is_NFD_QC=_ n}', "");
Expect(0, 195101, '\P{Is_NFD_QC=_ n}', "");
Expect(1, 195101, '\P{^Is_NFD_QC=_ n}', "");
Expect(0, 195102, '\p{Is_NFD_QC=_ n}', "");
Expect(1, 195102, '\p{^Is_NFD_QC=_ n}', "");
Expect(1, 195102, '\P{Is_NFD_QC=_ n}', "");
Expect(0, 195102, '\P{^Is_NFD_QC=_ n}', "");
Error('\p{Expands_On_NFKC=Yes}');
Error('\P{Expands_On_NFKC=Yes}');
Error('\p{XO_NFKC=Y}');
Error('\P{XO_NFKC=Y}');
Error('\p{Is_Expands_On_NFKC=T}');
Error('\P{Is_Expands_On_NFKC=T}');
Error('\p{Is_XO_NFKC=True}');
Error('\P{Is_XO_NFKC=True}');
Error('\p{Expands_On_NFKC=No}');
Error('\P{Expands_On_NFKC=No}');
Error('\p{XO_NFKC=N}');
Error('\P{XO_NFKC=N}');
Error('\p{Is_Expands_On_NFKC=F}');
Error('\P{Is_Expands_On_NFKC=F}');
Error('\p{Is_XO_NFKC=False}');
Error('\P{Is_XO_NFKC=False}');
Error('\p{Other_Lowercase=Yes}');
Error('\P{Other_Lowercase=Yes}');
Error('\p{OLower=Y}');
Error('\P{OLower=Y}');
Error('\p{Is_Other_Lowercase=T}');
Error('\P{Is_Other_Lowercase=T}');
Error('\p{Is_OLower:   True}');
Error('\P{Is_OLower:   True}');
Error('\p{Other_Lowercase=No}');
Error('\P{Other_Lowercase=No}');
Error('\p{OLower=N}');
Error('\P{OLower=N}');
Error('\p{Is_Other_Lowercase:	F}');
Error('\P{Is_Other_Lowercase:	F}');
Error('\p{Is_OLower=False}');
Error('\P{Is_OLower=False}');
Error('\p{Pattern_White_Space=	/a/Yes}');
Error('\P{Pattern_White_Space=	/a/Yes}');
Expect(1, 8233, '\p{Pattern_White_Space=yes}', "");
Expect(0, 8233, '\p{^Pattern_White_Space=yes}', "");
Expect(0, 8233, '\P{Pattern_White_Space=yes}', "");
Expect(1, 8233, '\P{^Pattern_White_Space=yes}', "");
Expect(0, 8234, '\p{Pattern_White_Space=yes}', "");
Expect(1, 8234, '\p{^Pattern_White_Space=yes}', "");
Expect(1, 8234, '\P{Pattern_White_Space=yes}', "");
Expect(0, 8234, '\P{^Pattern_White_Space=yes}', "");
Expect(1, 8233, '\p{Pattern_White_Space=	yes}', "");
Expect(0, 8233, '\p{^Pattern_White_Space=	yes}', "");
Expect(0, 8233, '\P{Pattern_White_Space=	yes}', "");
Expect(1, 8233, '\P{^Pattern_White_Space=	yes}', "");
Expect(0, 8234, '\p{Pattern_White_Space=	yes}', "");
Expect(1, 8234, '\p{^Pattern_White_Space=	yes}', "");
Expect(1, 8234, '\P{Pattern_White_Space=	yes}', "");
Expect(0, 8234, '\P{^Pattern_White_Space=	yes}', "");
Error('\p{Pat_WS= /a/Y}');
Error('\P{Pat_WS= /a/Y}');
Expect(1, 8233, '\p{Pat_WS:   y}', "");
Expect(0, 8233, '\p{^Pat_WS:   y}', "");
Expect(0, 8233, '\P{Pat_WS:   y}', "");
Expect(1, 8233, '\P{^Pat_WS:   y}', "");
Expect(0, 8234, '\p{Pat_WS:   y}', "");
Expect(1, 8234, '\p{^Pat_WS:   y}', "");
Expect(1, 8234, '\P{Pat_WS:   y}', "");
Expect(0, 8234, '\P{^Pat_WS:   y}', "");
Expect(1, 8233, '\p{Pat_WS:		Y}', "");
Expect(0, 8233, '\p{^Pat_WS:		Y}', "");
Expect(0, 8233, '\P{Pat_WS:		Y}', "");
Expect(1, 8233, '\P{^Pat_WS:		Y}', "");
Expect(0, 8234, '\p{Pat_WS:		Y}', "");
Expect(1, 8234, '\p{^Pat_WS:		Y}', "");
Expect(1, 8234, '\P{Pat_WS:		Y}', "");
Expect(0, 8234, '\P{^Pat_WS:		Y}', "");
Error('\p{Is_Pattern_White_Space=_/a/T}');
Error('\P{Is_Pattern_White_Space=_/a/T}');
Expect(1, 8233, '\p{Is_Pattern_White_Space=t}', "");
Expect(0, 8233, '\p{^Is_Pattern_White_Space=t}', "");
Expect(0, 8233, '\P{Is_Pattern_White_Space=t}', "");
Expect(1, 8233, '\P{^Is_Pattern_White_Space=t}', "");
Expect(0, 8234, '\p{Is_Pattern_White_Space=t}', "");
Expect(1, 8234, '\p{^Is_Pattern_White_Space=t}', "");
Expect(1, 8234, '\P{Is_Pattern_White_Space=t}', "");
Expect(0, 8234, '\P{^Is_Pattern_White_Space=t}', "");
Expect(1, 8233, '\p{Is_Pattern_White_Space= -T}', "");
Expect(0, 8233, '\p{^Is_Pattern_White_Space= -T}', "");
Expect(0, 8233, '\P{Is_Pattern_White_Space= -T}', "");
Expect(1, 8233, '\P{^Is_Pattern_White_Space= -T}', "");
Expect(0, 8234, '\p{Is_Pattern_White_Space= -T}', "");
Expect(1, 8234, '\p{^Is_Pattern_White_Space= -T}', "");
Expect(1, 8234, '\P{Is_Pattern_White_Space= -T}', "");
Expect(0, 8234, '\P{^Is_Pattern_White_Space= -T}', "");
Error('\p{Is_Pat_WS=/a/_true}');
Error('\P{Is_Pat_WS=/a/_true}');
Expect(1, 8233, '\p{Is_Pat_WS=true}', "");
Expect(0, 8233, '\p{^Is_Pat_WS=true}', "");
Expect(0, 8233, '\P{Is_Pat_WS=true}', "");
Expect(1, 8233, '\P{^Is_Pat_WS=true}', "");
Expect(0, 8234, '\p{Is_Pat_WS=true}', "");
Expect(1, 8234, '\p{^Is_Pat_WS=true}', "");
Expect(1, 8234, '\P{Is_Pat_WS=true}', "");
Expect(0, 8234, '\P{^Is_Pat_WS=true}', "");
Expect(1, 8233, '\p{Is_Pat_WS=	-TRUE}', "");
Expect(0, 8233, '\p{^Is_Pat_WS=	-TRUE}', "");
Expect(0, 8233, '\P{Is_Pat_WS=	-TRUE}', "");
Expect(1, 8233, '\P{^Is_Pat_WS=	-TRUE}', "");
Expect(0, 8234, '\p{Is_Pat_WS=	-TRUE}', "");
Expect(1, 8234, '\p{^Is_Pat_WS=	-TRUE}', "");
Expect(1, 8234, '\P{Is_Pat_WS=	-TRUE}', "");
Expect(0, 8234, '\P{^Is_Pat_WS=	-TRUE}', "");
Error('\p{Pattern_White_Space=:=  No}');
Error('\P{Pattern_White_Space=:=  No}');
Expect(1, 8234, '\p{Pattern_White_Space=no}', "");
Expect(0, 8234, '\p{^Pattern_White_Space=no}', "");
Expect(0, 8234, '\P{Pattern_White_Space=no}', "");
Expect(1, 8234, '\P{^Pattern_White_Space=no}', "");
Expect(0, 8233, '\p{Pattern_White_Space=no}', "");
Expect(1, 8233, '\p{^Pattern_White_Space=no}', "");
Expect(1, 8233, '\P{Pattern_White_Space=no}', "");
Expect(0, 8233, '\P{^Pattern_White_Space=no}', "");
Expect(1, 8234, '\p{Pattern_White_Space:	no}', "");
Expect(0, 8234, '\p{^Pattern_White_Space:	no}', "");
Expect(0, 8234, '\P{Pattern_White_Space:	no}', "");
Expect(1, 8234, '\P{^Pattern_White_Space:	no}', "");
Expect(0, 8233, '\p{Pattern_White_Space:	no}', "");
Expect(1, 8233, '\p{^Pattern_White_Space:	no}', "");
Expect(1, 8233, '\P{Pattern_White_Space:	no}', "");
Expect(0, 8233, '\P{^Pattern_White_Space:	no}', "");
Error('\p{Pat_WS=		n/a/}');
Error('\P{Pat_WS=		n/a/}');
Expect(1, 8234, '\p{Pat_WS=n}', "");
Expect(0, 8234, '\p{^Pat_WS=n}', "");
Expect(0, 8234, '\P{Pat_WS=n}', "");
Expect(1, 8234, '\P{^Pat_WS=n}', "");
Expect(0, 8233, '\p{Pat_WS=n}', "");
Expect(1, 8233, '\p{^Pat_WS=n}', "");
Expect(1, 8233, '\P{Pat_WS=n}', "");
Expect(0, 8233, '\P{^Pat_WS=n}', "");
Expect(1, 8234, '\p{Pat_WS=_N}', "");
Expect(0, 8234, '\p{^Pat_WS=_N}', "");
Expect(0, 8234, '\P{Pat_WS=_N}', "");
Expect(1, 8234, '\P{^Pat_WS=_N}', "");
Expect(0, 8233, '\p{Pat_WS=_N}', "");
Expect(1, 8233, '\p{^Pat_WS=_N}', "");
Expect(1, 8233, '\P{Pat_WS=_N}', "");
Expect(0, 8233, '\P{^Pat_WS=_N}', "");
Error('\p{Is_Pattern_White_Space=-/a/f}');
Error('\P{Is_Pattern_White_Space=-/a/f}');
Expect(1, 8234, '\p{Is_Pattern_White_Space=f}', "");
Expect(0, 8234, '\p{^Is_Pattern_White_Space=f}', "");
Expect(0, 8234, '\P{Is_Pattern_White_Space=f}', "");
Expect(1, 8234, '\P{^Is_Pattern_White_Space=f}', "");
Expect(0, 8233, '\p{Is_Pattern_White_Space=f}', "");
Expect(1, 8233, '\p{^Is_Pattern_White_Space=f}', "");
Expect(1, 8233, '\P{Is_Pattern_White_Space=f}', "");
Expect(0, 8233, '\P{^Is_Pattern_White_Space=f}', "");
Expect(1, 8234, '\p{Is_Pattern_White_Space:- F}', "");
Expect(0, 8234, '\p{^Is_Pattern_White_Space:- F}', "");
Expect(0, 8234, '\P{Is_Pattern_White_Space:- F}', "");
Expect(1, 8234, '\P{^Is_Pattern_White_Space:- F}', "");
Expect(0, 8233, '\p{Is_Pattern_White_Space:- F}', "");
Expect(1, 8233, '\p{^Is_Pattern_White_Space:- F}', "");
Expect(1, 8233, '\P{Is_Pattern_White_Space:- F}', "");
Expect(0, 8233, '\P{^Is_Pattern_White_Space:- F}', "");
Error('\p{Is_Pat_WS=	/a/False}');
Error('\P{Is_Pat_WS=	/a/False}');
Expect(1, 8234, '\p{Is_Pat_WS=false}', "");
Expect(0, 8234, '\p{^Is_Pat_WS=false}', "");
Expect(0, 8234, '\P{Is_Pat_WS=false}', "");
Expect(1, 8234, '\P{^Is_Pat_WS=false}', "");
Expect(0, 8233, '\p{Is_Pat_WS=false}', "");
Expect(1, 8233, '\p{^Is_Pat_WS=false}', "");
Expect(1, 8233, '\P{Is_Pat_WS=false}', "");
Expect(0, 8233, '\P{^Is_Pat_WS=false}', "");
Expect(1, 8234, '\p{Is_Pat_WS= -false}', "");
Expect(0, 8234, '\p{^Is_Pat_WS= -false}', "");
Expect(0, 8234, '\P{Is_Pat_WS= -false}', "");
Expect(1, 8234, '\P{^Is_Pat_WS= -false}', "");
Expect(0, 8233, '\p{Is_Pat_WS= -false}', "");
Expect(1, 8233, '\p{^Is_Pat_WS= -false}', "");
Expect(1, 8233, '\P{Is_Pat_WS= -false}', "");
Expect(0, 8233, '\P{^Is_Pat_WS= -false}', "");
Error('\p{Joining_Group:	/a/rohingya_YEH}');
Error('\P{Joining_Group:	/a/rohingya_YEH}');
Expect(1, 2220, '\p{Joining_Group=rohingyayeh}', "");
Expect(0, 2220, '\p{^Joining_Group=rohingyayeh}', "");
Expect(0, 2220, '\P{Joining_Group=rohingyayeh}', "");
Expect(1, 2220, '\P{^Joining_Group=rohingyayeh}', "");
Expect(0, 2221, '\p{Joining_Group=rohingyayeh}', "");
Expect(1, 2221, '\p{^Joining_Group=rohingyayeh}', "");
Expect(1, 2221, '\P{Joining_Group=rohingyayeh}', "");
Expect(0, 2221, '\P{^Joining_Group=rohingyayeh}', "");
Expect(1, 2220, '\p{Joining_Group=_ Rohingya_Yeh}', "");
Expect(0, 2220, '\p{^Joining_Group=_ Rohingya_Yeh}', "");
Expect(0, 2220, '\P{Joining_Group=_ Rohingya_Yeh}', "");
Expect(1, 2220, '\P{^Joining_Group=_ Rohingya_Yeh}', "");
Expect(0, 2221, '\p{Joining_Group=_ Rohingya_Yeh}', "");
Expect(1, 2221, '\p{^Joining_Group=_ Rohingya_Yeh}', "");
Expect(1, 2221, '\P{Joining_Group=_ Rohingya_Yeh}', "");
Expect(0, 2221, '\P{^Joining_Group=_ Rohingya_Yeh}', "");
Error('\p{Jg=_:=Rohingya_Yeh}');
Error('\P{Jg=_:=Rohingya_Yeh}');
Expect(1, 2220, '\p{Jg=rohingyayeh}', "");
Expect(0, 2220, '\p{^Jg=rohingyayeh}', "");
Expect(0, 2220, '\P{Jg=rohingyayeh}', "");
Expect(1, 2220, '\P{^Jg=rohingyayeh}', "");
Expect(0, 2221, '\p{Jg=rohingyayeh}', "");
Expect(1, 2221, '\p{^Jg=rohingyayeh}', "");
Expect(1, 2221, '\P{Jg=rohingyayeh}', "");
Expect(0, 2221, '\P{^Jg=rohingyayeh}', "");
Expect(1, 2220, '\p{Jg= 	Rohingya_Yeh}', "");
Expect(0, 2220, '\p{^Jg= 	Rohingya_Yeh}', "");
Expect(0, 2220, '\P{Jg= 	Rohingya_Yeh}', "");
Expect(1, 2220, '\P{^Jg= 	Rohingya_Yeh}', "");
Expect(0, 2221, '\p{Jg= 	Rohingya_Yeh}', "");
Expect(1, 2221, '\p{^Jg= 	Rohingya_Yeh}', "");
Expect(1, 2221, '\P{Jg= 	Rohingya_Yeh}', "");
Expect(0, 2221, '\P{^Jg= 	Rohingya_Yeh}', "");
Error('\p{Is_Joining_Group: _-ROHINGYA_Yeh/a/}');
Error('\P{Is_Joining_Group: _-ROHINGYA_Yeh/a/}');
Expect(1, 2220, '\p{Is_Joining_Group=rohingyayeh}', "");
Expect(0, 2220, '\p{^Is_Joining_Group=rohingyayeh}', "");
Expect(0, 2220, '\P{Is_Joining_Group=rohingyayeh}', "");
Expect(1, 2220, '\P{^Is_Joining_Group=rohingyayeh}', "");
Expect(0, 2221, '\p{Is_Joining_Group=rohingyayeh}', "");
Expect(1, 2221, '\p{^Is_Joining_Group=rohingyayeh}', "");
Expect(1, 2221, '\P{Is_Joining_Group=rohingyayeh}', "");
Expect(0, 2221, '\P{^Is_Joining_Group=rohingyayeh}', "");
Expect(1, 2220, '\p{Is_Joining_Group=-Rohingya_Yeh}', "");
Expect(0, 2220, '\p{^Is_Joining_Group=-Rohingya_Yeh}', "");
Expect(0, 2220, '\P{Is_Joining_Group=-Rohingya_Yeh}', "");
Expect(1, 2220, '\P{^Is_Joining_Group=-Rohingya_Yeh}', "");
Expect(0, 2221, '\p{Is_Joining_Group=-Rohingya_Yeh}', "");
Expect(1, 2221, '\p{^Is_Joining_Group=-Rohingya_Yeh}', "");
Expect(1, 2221, '\P{Is_Joining_Group=-Rohingya_Yeh}', "");
Expect(0, 2221, '\P{^Is_Joining_Group=-Rohingya_Yeh}', "");
Error('\p{Is_Jg=_-rohingya_Yeh/a/}');
Error('\P{Is_Jg=_-rohingya_Yeh/a/}');
Expect(1, 2220, '\p{Is_Jg=rohingyayeh}', "");
Expect(0, 2220, '\p{^Is_Jg=rohingyayeh}', "");
Expect(0, 2220, '\P{Is_Jg=rohingyayeh}', "");
Expect(1, 2220, '\P{^Is_Jg=rohingyayeh}', "");
Expect(0, 2221, '\p{Is_Jg=rohingyayeh}', "");
Expect(1, 2221, '\p{^Is_Jg=rohingyayeh}', "");
Expect(1, 2221, '\P{Is_Jg=rohingyayeh}', "");
Expect(0, 2221, '\P{^Is_Jg=rohingyayeh}', "");
Expect(1, 2220, '\p{Is_Jg:   Rohingya_Yeh}', "");
Expect(0, 2220, '\p{^Is_Jg:   Rohingya_Yeh}', "");
Expect(0, 2220, '\P{Is_Jg:   Rohingya_Yeh}', "");
Expect(1, 2220, '\P{^Is_Jg:   Rohingya_Yeh}', "");
Expect(0, 2221, '\p{Is_Jg:   Rohingya_Yeh}', "");
Expect(1, 2221, '\p{^Is_Jg:   Rohingya_Yeh}', "");
Expect(1, 2221, '\P{Is_Jg:   Rohingya_Yeh}', "");
Expect(0, 2221, '\P{^Is_Jg:   Rohingya_Yeh}', "");
Error('\p{Joining_Group:_Swash_kaf/a/}');
Error('\P{Joining_Group:_Swash_kaf/a/}');
Expect(1, 1706, '\p{Joining_Group=swashkaf}', "");
Expect(0, 1706, '\p{^Joining_Group=swashkaf}', "");
Expect(0, 1706, '\P{Joining_Group=swashkaf}', "");
Expect(1, 1706, '\P{^Joining_Group=swashkaf}', "");
Expect(0, 1707, '\p{Joining_Group=swashkaf}', "");
Expect(1, 1707, '\p{^Joining_Group=swashkaf}', "");
Expect(1, 1707, '\P{Joining_Group=swashkaf}', "");
Expect(0, 1707, '\P{^Joining_Group=swashkaf}', "");
Expect(1, 1706, '\p{Joining_Group=	-Swash_KAF}', "");
Expect(0, 1706, '\p{^Joining_Group=	-Swash_KAF}', "");
Expect(0, 1706, '\P{Joining_Group=	-Swash_KAF}', "");
Expect(1, 1706, '\P{^Joining_Group=	-Swash_KAF}', "");
Expect(0, 1707, '\p{Joining_Group=	-Swash_KAF}', "");
Expect(1, 1707, '\p{^Joining_Group=	-Swash_KAF}', "");
Expect(1, 1707, '\P{Joining_Group=	-Swash_KAF}', "");
Expect(0, 1707, '\P{^Joining_Group=	-Swash_KAF}', "");
Error('\p{Jg= SWASH_KAF:=}');
Error('\P{Jg= SWASH_KAF:=}');
Expect(1, 1706, '\p{Jg=swashkaf}', "");
Expect(0, 1706, '\p{^Jg=swashkaf}', "");
Expect(0, 1706, '\P{Jg=swashkaf}', "");
Expect(1, 1706, '\P{^Jg=swashkaf}', "");
Expect(0, 1707, '\p{Jg=swashkaf}', "");
Expect(1, 1707, '\p{^Jg=swashkaf}', "");
Expect(1, 1707, '\P{Jg=swashkaf}', "");
Expect(0, 1707, '\P{^Jg=swashkaf}', "");
Expect(1, 1706, '\p{Jg:_-Swash_KAF}', "");
Expect(0, 1706, '\p{^Jg:_-Swash_KAF}', "");
Expect(0, 1706, '\P{Jg:_-Swash_KAF}', "");
Expect(1, 1706, '\P{^Jg:_-Swash_KAF}', "");
Expect(0, 1707, '\p{Jg:_-Swash_KAF}', "");
Expect(1, 1707, '\p{^Jg:_-Swash_KAF}', "");
Expect(1, 1707, '\P{Jg:_-Swash_KAF}', "");
Expect(0, 1707, '\P{^Jg:_-Swash_KAF}', "");
Error('\p{Is_Joining_Group= _swash_Kaf:=}');
Error('\P{Is_Joining_Group= _swash_Kaf:=}');
Expect(1, 1706, '\p{Is_Joining_Group:swashkaf}', "");
Expect(0, 1706, '\p{^Is_Joining_Group:swashkaf}', "");
Expect(0, 1706, '\P{Is_Joining_Group:swashkaf}', "");
Expect(1, 1706, '\P{^Is_Joining_Group:swashkaf}', "");
Expect(0, 1707, '\p{Is_Joining_Group:swashkaf}', "");
Expect(1, 1707, '\p{^Is_Joining_Group:swashkaf}', "");
Expect(1, 1707, '\P{Is_Joining_Group:swashkaf}', "");
Expect(0, 1707, '\P{^Is_Joining_Group:swashkaf}', "");
Expect(1, 1706, '\p{Is_Joining_Group=		swash_kaf}', "");
Expect(0, 1706, '\p{^Is_Joining_Group=		swash_kaf}', "");
Expect(0, 1706, '\P{Is_Joining_Group=		swash_kaf}', "");
Expect(1, 1706, '\P{^Is_Joining_Group=		swash_kaf}', "");
Expect(0, 1707, '\p{Is_Joining_Group=		swash_kaf}', "");
Expect(1, 1707, '\p{^Is_Joining_Group=		swash_kaf}', "");
Expect(1, 1707, '\P{Is_Joining_Group=		swash_kaf}', "");
Expect(0, 1707, '\P{^Is_Joining_Group=		swash_kaf}', "");
Error('\p{Is_Jg=_	Swash_Kaf:=}');
Error('\P{Is_Jg=_	Swash_Kaf:=}');
Expect(1, 1706, '\p{Is_Jg=swashkaf}', "");
Expect(0, 1706, '\p{^Is_Jg=swashkaf}', "");
Expect(0, 1706, '\P{Is_Jg=swashkaf}', "");
Expect(1, 1706, '\P{^Is_Jg=swashkaf}', "");
Expect(0, 1707, '\p{Is_Jg=swashkaf}', "");
Expect(1, 1707, '\p{^Is_Jg=swashkaf}', "");
Expect(1, 1707, '\P{Is_Jg=swashkaf}', "");
Expect(0, 1707, '\P{^Is_Jg=swashkaf}', "");
Expect(1, 1706, '\p{Is_Jg=- SWASH_Kaf}', "");
Expect(0, 1706, '\p{^Is_Jg=- SWASH_Kaf}', "");
Expect(0, 1706, '\P{Is_Jg=- SWASH_Kaf}', "");
Expect(1, 1706, '\P{^Is_Jg=- SWASH_Kaf}', "");
Expect(0, 1707, '\p{Is_Jg=- SWASH_Kaf}', "");
Expect(1, 1707, '\p{^Is_Jg=- SWASH_Kaf}', "");
Expect(1, 1707, '\P{Is_Jg=- SWASH_Kaf}', "");
Expect(0, 1707, '\P{^Is_Jg=- SWASH_Kaf}', "");
Error('\p{Joining_Group=_	seen:=}');
Error('\P{Joining_Group=_	seen:=}');
Expect(1, 1918, '\p{Joining_Group=seen}', "");
Expect(0, 1918, '\p{^Joining_Group=seen}', "");
Expect(0, 1918, '\P{Joining_Group=seen}', "");
Expect(1, 1918, '\P{^Joining_Group=seen}', "");
Expect(0, 1919, '\p{Joining_Group=seen}', "");
Expect(1, 1919, '\p{^Joining_Group=seen}', "");
Expect(1, 1919, '\P{Joining_Group=seen}', "");
Expect(0, 1919, '\P{^Joining_Group=seen}', "");
Expect(1, 1918, '\p{Joining_Group=	seen}', "");
Expect(0, 1918, '\p{^Joining_Group=	seen}', "");
Expect(0, 1918, '\P{Joining_Group=	seen}', "");
Expect(1, 1918, '\P{^Joining_Group=	seen}', "");
Expect(0, 1919, '\p{Joining_Group=	seen}', "");
Expect(1, 1919, '\p{^Joining_Group=	seen}', "");
Expect(1, 1919, '\P{Joining_Group=	seen}', "");
Expect(0, 1919, '\P{^Joining_Group=	seen}', "");
Error('\p{Jg=seen:=}');
Error('\P{Jg=seen:=}');
Expect(1, 1918, '\p{Jg=seen}', "");
Expect(0, 1918, '\p{^Jg=seen}', "");
Expect(0, 1918, '\P{Jg=seen}', "");
Expect(1, 1918, '\P{^Jg=seen}', "");
Expect(0, 1919, '\p{Jg=seen}', "");
Expect(1, 1919, '\p{^Jg=seen}', "");
Expect(1, 1919, '\P{Jg=seen}', "");
Expect(0, 1919, '\P{^Jg=seen}', "");
Expect(1, 1918, '\p{Jg=_	SEEN}', "");
Expect(0, 1918, '\p{^Jg=_	SEEN}', "");
Expect(0, 1918, '\P{Jg=_	SEEN}', "");
Expect(1, 1918, '\P{^Jg=_	SEEN}', "");
Expect(0, 1919, '\p{Jg=_	SEEN}', "");
Expect(1, 1919, '\p{^Jg=_	SEEN}', "");
Expect(1, 1919, '\P{Jg=_	SEEN}', "");
Expect(0, 1919, '\P{^Jg=_	SEEN}', "");
Error('\p{Is_Joining_Group:   _/a/SEEN}');
Error('\P{Is_Joining_Group:   _/a/SEEN}');
Expect(1, 1918, '\p{Is_Joining_Group=seen}', "");
Expect(0, 1918, '\p{^Is_Joining_Group=seen}', "");
Expect(0, 1918, '\P{Is_Joining_Group=seen}', "");
Expect(1, 1918, '\P{^Is_Joining_Group=seen}', "");
Expect(0, 1919, '\p{Is_Joining_Group=seen}', "");
Expect(1, 1919, '\p{^Is_Joining_Group=seen}', "");
Expect(1, 1919, '\P{Is_Joining_Group=seen}', "");
Expect(0, 1919, '\P{^Is_Joining_Group=seen}', "");
Expect(1, 1918, '\p{Is_Joining_Group:   -	SEEN}', "");
Expect(0, 1918, '\p{^Is_Joining_Group:   -	SEEN}', "");
Expect(0, 1918, '\P{Is_Joining_Group:   -	SEEN}', "");
Expect(1, 1918, '\P{^Is_Joining_Group:   -	SEEN}', "");
Expect(0, 1919, '\p{Is_Joining_Group:   -	SEEN}', "");
Expect(1, 1919, '\p{^Is_Joining_Group:   -	SEEN}', "");
Expect(1, 1919, '\P{Is_Joining_Group:   -	SEEN}', "");
Expect(0, 1919, '\P{^Is_Joining_Group:   -	SEEN}', "");
Error('\p{Is_Jg=_-Seen:=}');
Error('\P{Is_Jg=_-Seen:=}');
Expect(1, 1918, '\p{Is_Jg:seen}', "");
Expect(0, 1918, '\p{^Is_Jg:seen}', "");
Expect(0, 1918, '\P{Is_Jg:seen}', "");
Expect(1, 1918, '\P{^Is_Jg:seen}', "");
Expect(0, 1919, '\p{Is_Jg:seen}', "");
Expect(1, 1919, '\p{^Is_Jg:seen}', "");
Expect(1, 1919, '\P{Is_Jg:seen}', "");
Expect(0, 1919, '\P{^Is_Jg:seen}', "");
Expect(1, 1918, '\p{Is_Jg=Seen}', "");
Expect(0, 1918, '\p{^Is_Jg=Seen}', "");
Expect(0, 1918, '\P{Is_Jg=Seen}', "");
Expect(1, 1918, '\P{^Is_Jg=Seen}', "");
Expect(0, 1919, '\p{Is_Jg=Seen}', "");
Expect(1, 1919, '\p{^Is_Jg=Seen}', "");
Expect(1, 1919, '\P{Is_Jg=Seen}', "");
Expect(0, 1919, '\P{^Is_Jg=Seen}', "");
Error('\p{Joining_Group:	/a/		Beth}');
Error('\P{Joining_Group:	/a/		Beth}');
Expect(1, 1837, '\p{Joining_Group=beth}', "");
Expect(0, 1837, '\p{^Joining_Group=beth}', "");
Expect(0, 1837, '\P{Joining_Group=beth}', "");
Expect(1, 1837, '\P{^Joining_Group=beth}', "");
Expect(0, 1838, '\p{Joining_Group=beth}', "");
Expect(1, 1838, '\p{^Joining_Group=beth}', "");
Expect(1, 1838, '\P{Joining_Group=beth}', "");
Expect(0, 1838, '\P{^Joining_Group=beth}', "");
Expect(1, 1837, '\p{Joining_Group=__BETH}', "");
Expect(0, 1837, '\p{^Joining_Group=__BETH}', "");
Expect(0, 1837, '\P{Joining_Group=__BETH}', "");
Expect(1, 1837, '\P{^Joining_Group=__BETH}', "");
Expect(0, 1838, '\p{Joining_Group=__BETH}', "");
Expect(1, 1838, '\p{^Joining_Group=__BETH}', "");
Expect(1, 1838, '\P{Joining_Group=__BETH}', "");
Expect(0, 1838, '\P{^Joining_Group=__BETH}', "");
Error('\p{Jg=-BETH/a/}');
Error('\P{Jg=-BETH/a/}');
Expect(1, 1837, '\p{Jg=beth}', "");
Expect(0, 1837, '\p{^Jg=beth}', "");
Expect(0, 1837, '\P{Jg=beth}', "");
Expect(1, 1837, '\P{^Jg=beth}', "");
Expect(0, 1838, '\p{Jg=beth}', "");
Expect(1, 1838, '\p{^Jg=beth}', "");
Expect(1, 1838, '\P{Jg=beth}', "");
Expect(0, 1838, '\P{^Jg=beth}', "");
Expect(1, 1837, '\p{Jg=-BETH}', "");
Expect(0, 1837, '\p{^Jg=-BETH}', "");
Expect(0, 1837, '\P{Jg=-BETH}', "");
Expect(1, 1837, '\P{^Jg=-BETH}', "");
Expect(0, 1838, '\p{Jg=-BETH}', "");
Expect(1, 1838, '\p{^Jg=-BETH}', "");
Expect(1, 1838, '\P{Jg=-BETH}', "");
Expect(0, 1838, '\P{^Jg=-BETH}', "");
Error('\p{Is_Joining_Group:	_beth/a/}');
Error('\P{Is_Joining_Group:	_beth/a/}');
Expect(1, 1837, '\p{Is_Joining_Group=beth}', "");
Expect(0, 1837, '\p{^Is_Joining_Group=beth}', "");
Expect(0, 1837, '\P{Is_Joining_Group=beth}', "");
Expect(1, 1837, '\P{^Is_Joining_Group=beth}', "");
Expect(0, 1838, '\p{Is_Joining_Group=beth}', "");
Expect(1, 1838, '\p{^Is_Joining_Group=beth}', "");
Expect(1, 1838, '\P{Is_Joining_Group=beth}', "");
Expect(0, 1838, '\P{^Is_Joining_Group=beth}', "");
Expect(1, 1837, '\p{Is_Joining_Group=-_Beth}', "");
Expect(0, 1837, '\p{^Is_Joining_Group=-_Beth}', "");
Expect(0, 1837, '\P{Is_Joining_Group=-_Beth}', "");
Expect(1, 1837, '\P{^Is_Joining_Group=-_Beth}', "");
Expect(0, 1838, '\p{Is_Joining_Group=-_Beth}', "");
Expect(1, 1838, '\p{^Is_Joining_Group=-_Beth}', "");
Expect(1, 1838, '\P{Is_Joining_Group=-_Beth}', "");
Expect(0, 1838, '\P{^Is_Joining_Group=-_Beth}', "");
Error('\p{Is_Jg=:=	_Beth}');
Error('\P{Is_Jg=:=	_Beth}');
Expect(1, 1837, '\p{Is_Jg=beth}', "");
Expect(0, 1837, '\p{^Is_Jg=beth}', "");
Expect(0, 1837, '\P{Is_Jg=beth}', "");
Expect(1, 1837, '\P{^Is_Jg=beth}', "");
Expect(0, 1838, '\p{Is_Jg=beth}', "");
Expect(1, 1838, '\p{^Is_Jg=beth}', "");
Expect(1, 1838, '\P{Is_Jg=beth}', "");
Expect(0, 1838, '\P{^Is_Jg=beth}', "");
Expect(1, 1837, '\p{Is_Jg=-	BETH}', "");
Expect(0, 1837, '\p{^Is_Jg=-	BETH}', "");
Expect(0, 1837, '\P{Is_Jg=-	BETH}', "");
Expect(1, 1837, '\P{^Is_Jg=-	BETH}', "");
Expect(0, 1838, '\p{Is_Jg=-	BETH}', "");
Expect(1, 1838, '\p{^Is_Jg=-	BETH}', "");
Expect(1, 1838, '\P{Is_Jg=-	BETH}', "");
Expect(0, 1838, '\P{^Is_Jg=-	BETH}', "");
Error('\p{Joining_Group= -Semkath:=}');
Error('\P{Joining_Group= -Semkath:=}');
Expect(1, 1827, '\p{Joining_Group=semkath}', "");
Expect(0, 1827, '\p{^Joining_Group=semkath}', "");
Expect(0, 1827, '\P{Joining_Group=semkath}', "");
Expect(1, 1827, '\P{^Joining_Group=semkath}', "");
Expect(0, 1828, '\p{Joining_Group=semkath}', "");
Expect(1, 1828, '\p{^Joining_Group=semkath}', "");
Expect(1, 1828, '\P{Joining_Group=semkath}', "");
Expect(0, 1828, '\P{^Joining_Group=semkath}', "");
Expect(1, 1827, '\p{Joining_Group:   _	Semkath}', "");
Expect(0, 1827, '\p{^Joining_Group:   _	Semkath}', "");
Expect(0, 1827, '\P{Joining_Group:   _	Semkath}', "");
Expect(1, 1827, '\P{^Joining_Group:   _	Semkath}', "");
Expect(0, 1828, '\p{Joining_Group:   _	Semkath}', "");
Expect(1, 1828, '\p{^Joining_Group:   _	Semkath}', "");
Expect(1, 1828, '\P{Joining_Group:   _	Semkath}', "");
Expect(0, 1828, '\P{^Joining_Group:   _	Semkath}', "");
Error('\p{Jg:/a/-_semkath}');
Error('\P{Jg:/a/-_semkath}');
Expect(1, 1827, '\p{Jg=semkath}', "");
Expect(0, 1827, '\p{^Jg=semkath}', "");
Expect(0, 1827, '\P{Jg=semkath}', "");
Expect(1, 1827, '\P{^Jg=semkath}', "");
Expect(0, 1828, '\p{Jg=semkath}', "");
Expect(1, 1828, '\p{^Jg=semkath}', "");
Expect(1, 1828, '\P{Jg=semkath}', "");
Expect(0, 1828, '\P{^Jg=semkath}', "");
Expect(1, 1827, '\p{Jg= -Semkath}', "");
Expect(0, 1827, '\p{^Jg= -Semkath}', "");
Expect(0, 1827, '\P{Jg= -Semkath}', "");
Expect(1, 1827, '\P{^Jg= -Semkath}', "");
Expect(0, 1828, '\p{Jg= -Semkath}', "");
Expect(1, 1828, '\p{^Jg= -Semkath}', "");
Expect(1, 1828, '\P{Jg= -Semkath}', "");
Expect(0, 1828, '\P{^Jg= -Semkath}', "");
Error('\p{Is_Joining_Group: :=Semkath}');
Error('\P{Is_Joining_Group: :=Semkath}');
Expect(1, 1827, '\p{Is_Joining_Group=semkath}', "");
Expect(0, 1827, '\p{^Is_Joining_Group=semkath}', "");
Expect(0, 1827, '\P{Is_Joining_Group=semkath}', "");
Expect(1, 1827, '\P{^Is_Joining_Group=semkath}', "");
Expect(0, 1828, '\p{Is_Joining_Group=semkath}', "");
Expect(1, 1828, '\p{^Is_Joining_Group=semkath}', "");
Expect(1, 1828, '\P{Is_Joining_Group=semkath}', "");
Expect(0, 1828, '\P{^Is_Joining_Group=semkath}', "");
Expect(1, 1827, '\p{Is_Joining_Group= Semkath}', "");
Expect(0, 1827, '\p{^Is_Joining_Group= Semkath}', "");
Expect(0, 1827, '\P{Is_Joining_Group= Semkath}', "");
Expect(1, 1827, '\P{^Is_Joining_Group= Semkath}', "");
Expect(0, 1828, '\p{Is_Joining_Group= Semkath}', "");
Expect(1, 1828, '\p{^Is_Joining_Group= Semkath}', "");
Expect(1, 1828, '\P{Is_Joining_Group= Semkath}', "");
Expect(0, 1828, '\P{^Is_Joining_Group= Semkath}', "");
Error('\p{Is_Jg=	_Semkath/a/}');
Error('\P{Is_Jg=	_Semkath/a/}');
Expect(1, 1827, '\p{Is_Jg:   semkath}', "");
Expect(0, 1827, '\p{^Is_Jg:   semkath}', "");
Expect(0, 1827, '\P{Is_Jg:   semkath}', "");
Expect(1, 1827, '\P{^Is_Jg:   semkath}', "");
Expect(0, 1828, '\p{Is_Jg:   semkath}', "");
Expect(1, 1828, '\p{^Is_Jg:   semkath}', "");
Expect(1, 1828, '\P{Is_Jg:   semkath}', "");
Expect(0, 1828, '\P{^Is_Jg:   semkath}', "");
Expect(1, 1827, '\p{Is_Jg=_-Semkath}', "");
Expect(0, 1827, '\p{^Is_Jg=_-Semkath}', "");
Expect(0, 1827, '\P{Is_Jg=_-Semkath}', "");
Expect(1, 1827, '\P{^Is_Jg=_-Semkath}', "");
Expect(0, 1828, '\p{Is_Jg=_-Semkath}', "");
Expect(1, 1828, '\p{^Is_Jg=_-Semkath}', "");
Expect(1, 1828, '\P{Is_Jg=_-Semkath}', "");
Expect(0, 1828, '\P{^Is_Jg=_-Semkath}', "");
Error('\p{Joining_Group=/a/	hah}');
Error('\P{Joining_Group=/a/	hah}');
Expect(1, 2210, '\p{Joining_Group=hah}', "");
Expect(0, 2210, '\p{^Joining_Group=hah}', "");
Expect(0, 2210, '\P{Joining_Group=hah}', "");
Expect(1, 2210, '\P{^Joining_Group=hah}', "");
Expect(0, 2211, '\p{Joining_Group=hah}', "");
Expect(1, 2211, '\p{^Joining_Group=hah}', "");
Expect(1, 2211, '\P{Joining_Group=hah}', "");
Expect(0, 2211, '\P{^Joining_Group=hah}', "");
Expect(1, 2210, '\p{Joining_Group=-Hah}', "");
Expect(0, 2210, '\p{^Joining_Group=-Hah}', "");
Expect(0, 2210, '\P{Joining_Group=-Hah}', "");
Expect(1, 2210, '\P{^Joining_Group=-Hah}', "");
Expect(0, 2211, '\p{Joining_Group=-Hah}', "");
Expect(1, 2211, '\p{^Joining_Group=-Hah}', "");
Expect(1, 2211, '\P{Joining_Group=-Hah}', "");
Expect(0, 2211, '\P{^Joining_Group=-Hah}', "");
Error('\p{Jg=	HAH/a/}');
Error('\P{Jg=	HAH/a/}');
Expect(1, 2210, '\p{Jg=hah}', "");
Expect(0, 2210, '\p{^Jg=hah}', "");
Expect(0, 2210, '\P{Jg=hah}', "");
Expect(1, 2210, '\P{^Jg=hah}', "");
Expect(0, 2211, '\p{Jg=hah}', "");
Expect(1, 2211, '\p{^Jg=hah}', "");
Expect(1, 2211, '\P{Jg=hah}', "");
Expect(0, 2211, '\P{^Jg=hah}', "");
Expect(1, 2210, '\p{Jg=	 Hah}', "");
Expect(0, 2210, '\p{^Jg=	 Hah}', "");
Expect(0, 2210, '\P{Jg=	 Hah}', "");
Expect(1, 2210, '\P{^Jg=	 Hah}', "");
Expect(0, 2211, '\p{Jg=	 Hah}', "");
Expect(1, 2211, '\p{^Jg=	 Hah}', "");
Expect(1, 2211, '\P{Jg=	 Hah}', "");
Expect(0, 2211, '\P{^Jg=	 Hah}', "");
Error('\p{Is_Joining_Group=:=--Hah}');
Error('\P{Is_Joining_Group=:=--Hah}');
Expect(1, 2210, '\p{Is_Joining_Group=hah}', "");
Expect(0, 2210, '\p{^Is_Joining_Group=hah}', "");
Expect(0, 2210, '\P{Is_Joining_Group=hah}', "");
Expect(1, 2210, '\P{^Is_Joining_Group=hah}', "");
Expect(0, 2211, '\p{Is_Joining_Group=hah}', "");
Expect(1, 2211, '\p{^Is_Joining_Group=hah}', "");
Expect(1, 2211, '\P{Is_Joining_Group=hah}', "");
Expect(0, 2211, '\P{^Is_Joining_Group=hah}', "");
Expect(1, 2210, '\p{Is_Joining_Group= hah}', "");
Expect(0, 2210, '\p{^Is_Joining_Group= hah}', "");
Expect(0, 2210, '\P{Is_Joining_Group= hah}', "");
Expect(1, 2210, '\P{^Is_Joining_Group= hah}', "");
Expect(0, 2211, '\p{Is_Joining_Group= hah}', "");
Expect(1, 2211, '\p{^Is_Joining_Group= hah}', "");
Expect(1, 2211, '\P{Is_Joining_Group= hah}', "");
Expect(0, 2211, '\P{^Is_Joining_Group= hah}', "");
Error('\p{Is_Jg=/a/	hah}');
Error('\P{Is_Jg=/a/	hah}');
Expect(1, 2210, '\p{Is_Jg=hah}', "");
Expect(0, 2210, '\p{^Is_Jg=hah}', "");
Expect(0, 2210, '\P{Is_Jg=hah}', "");
Expect(1, 2210, '\P{^Is_Jg=hah}', "");
Expect(0, 2211, '\p{Is_Jg=hah}', "");
Expect(1, 2211, '\p{^Is_Jg=hah}', "");
Expect(1, 2211, '\P{Is_Jg=hah}', "");
Expect(0, 2211, '\P{^Is_Jg=hah}', "");
Expect(1, 2210, '\p{Is_Jg=--Hah}', "");
Expect(0, 2210, '\p{^Is_Jg=--Hah}', "");
Expect(0, 2210, '\P{Is_Jg=--Hah}', "");
Expect(1, 2210, '\P{^Is_Jg=--Hah}', "");
Expect(0, 2211, '\p{Is_Jg=--Hah}', "");
Expect(1, 2211, '\p{^Is_Jg=--Hah}', "");
Expect(1, 2211, '\P{Is_Jg=--Hah}', "");
Expect(0, 2211, '\P{^Is_Jg=--Hah}', "");
Error('\p{Joining_Group= Syriac_Waw:=}');
Error('\P{Joining_Group= Syriac_Waw:=}');
Expect(1, 1816, '\p{Joining_Group=syriacwaw}', "");
Expect(0, 1816, '\p{^Joining_Group=syriacwaw}', "");
Expect(0, 1816, '\P{Joining_Group=syriacwaw}', "");
Expect(1, 1816, '\P{^Joining_Group=syriacwaw}', "");
Expect(0, 1817, '\p{Joining_Group=syriacwaw}', "");
Expect(1, 1817, '\p{^Joining_Group=syriacwaw}', "");
Expect(1, 1817, '\P{Joining_Group=syriacwaw}', "");
Expect(0, 1817, '\P{^Joining_Group=syriacwaw}', "");
Expect(1, 1816, '\p{Joining_Group=	Syriac_Waw}', "");
Expect(0, 1816, '\p{^Joining_Group=	Syriac_Waw}', "");
Expect(0, 1816, '\P{Joining_Group=	Syriac_Waw}', "");
Expect(1, 1816, '\P{^Joining_Group=	Syriac_Waw}', "");
Expect(0, 1817, '\p{Joining_Group=	Syriac_Waw}', "");
Expect(1, 1817, '\p{^Joining_Group=	Syriac_Waw}', "");
Expect(1, 1817, '\P{Joining_Group=	Syriac_Waw}', "");
Expect(0, 1817, '\P{^Joining_Group=	Syriac_Waw}', "");
Error('\p{Jg=/a/-	syriac_Waw}');
Error('\P{Jg=/a/-	syriac_Waw}');
Expect(1, 1816, '\p{Jg=syriacwaw}', "");
Expect(0, 1816, '\p{^Jg=syriacwaw}', "");
Expect(0, 1816, '\P{Jg=syriacwaw}', "");
Expect(1, 1816, '\P{^Jg=syriacwaw}', "");
Expect(0, 1817, '\p{Jg=syriacwaw}', "");
Expect(1, 1817, '\p{^Jg=syriacwaw}', "");
Expect(1, 1817, '\P{Jg=syriacwaw}', "");
Expect(0, 1817, '\P{^Jg=syriacwaw}', "");
Expect(1, 1816, '\p{Jg=__Syriac_Waw}', "");
Expect(0, 1816, '\p{^Jg=__Syriac_Waw}', "");
Expect(0, 1816, '\P{Jg=__Syriac_Waw}', "");
Expect(1, 1816, '\P{^Jg=__Syriac_Waw}', "");
Expect(0, 1817, '\p{Jg=__Syriac_Waw}', "");
Expect(1, 1817, '\p{^Jg=__Syriac_Waw}', "");
Expect(1, 1817, '\P{Jg=__Syriac_Waw}', "");
Expect(0, 1817, '\P{^Jg=__Syriac_Waw}', "");
Error('\p{Is_Joining_Group=:=_Syriac_waw}');
Error('\P{Is_Joining_Group=:=_Syriac_waw}');
Expect(1, 1816, '\p{Is_Joining_Group=syriacwaw}', "");
Expect(0, 1816, '\p{^Is_Joining_Group=syriacwaw}', "");
Expect(0, 1816, '\P{Is_Joining_Group=syriacwaw}', "");
Expect(1, 1816, '\P{^Is_Joining_Group=syriacwaw}', "");
Expect(0, 1817, '\p{Is_Joining_Group=syriacwaw}', "");
Expect(1, 1817, '\p{^Is_Joining_Group=syriacwaw}', "");
Expect(1, 1817, '\P{Is_Joining_Group=syriacwaw}', "");
Expect(0, 1817, '\P{^Is_Joining_Group=syriacwaw}', "");
Expect(1, 1816, '\p{Is_Joining_Group= Syriac_WAW}', "");
Expect(0, 1816, '\p{^Is_Joining_Group= Syriac_WAW}', "");
Expect(0, 1816, '\P{Is_Joining_Group= Syriac_WAW}', "");
Expect(1, 1816, '\P{^Is_Joining_Group= Syriac_WAW}', "");
Expect(0, 1817, '\p{Is_Joining_Group= Syriac_WAW}', "");
Expect(1, 1817, '\p{^Is_Joining_Group= Syriac_WAW}', "");
Expect(1, 1817, '\P{Is_Joining_Group= Syriac_WAW}', "");
Expect(0, 1817, '\P{^Is_Joining_Group= Syriac_WAW}', "");
Error('\p{Is_Jg=/a/SYRIAC_WAW}');
Error('\P{Is_Jg=/a/SYRIAC_WAW}');
Expect(1, 1816, '\p{Is_Jg=syriacwaw}', "");
Expect(0, 1816, '\p{^Is_Jg=syriacwaw}', "");
Expect(0, 1816, '\P{Is_Jg=syriacwaw}', "");
Expect(1, 1816, '\P{^Is_Jg=syriacwaw}', "");
Expect(0, 1817, '\p{Is_Jg=syriacwaw}', "");
Expect(1, 1817, '\p{^Is_Jg=syriacwaw}', "");
Expect(1, 1817, '\P{Is_Jg=syriacwaw}', "");
Expect(0, 1817, '\P{^Is_Jg=syriacwaw}', "");
Expect(1, 1816, '\p{Is_Jg= _syriac_Waw}', "");
Expect(0, 1816, '\p{^Is_Jg= _syriac_Waw}', "");
Expect(0, 1816, '\P{Is_Jg= _syriac_Waw}', "");
Expect(1, 1816, '\P{^Is_Jg= _syriac_Waw}', "");
Expect(0, 1817, '\p{Is_Jg= _syriac_Waw}', "");
Expect(1, 1817, '\p{^Is_Jg= _syriac_Waw}', "");
Expect(1, 1817, '\P{Is_Jg= _syriac_Waw}', "");
Expect(0, 1817, '\P{^Is_Jg= _syriac_Waw}', "");
Error('\p{Joining_Group=/a/	_Taw}');
Error('\P{Joining_Group=/a/	_Taw}');
Expect(1, 1836, '\p{Joining_Group=taw}', "");
Expect(0, 1836, '\p{^Joining_Group=taw}', "");
Expect(0, 1836, '\P{Joining_Group=taw}', "");
Expect(1, 1836, '\P{^Joining_Group=taw}', "");
Expect(0, 1837, '\p{Joining_Group=taw}', "");
Expect(1, 1837, '\p{^Joining_Group=taw}', "");
Expect(1, 1837, '\P{Joining_Group=taw}', "");
Expect(0, 1837, '\P{^Joining_Group=taw}', "");
Expect(1, 1836, '\p{Joining_Group=		taw}', "");
Expect(0, 1836, '\p{^Joining_Group=		taw}', "");
Expect(0, 1836, '\P{Joining_Group=		taw}', "");
Expect(1, 1836, '\P{^Joining_Group=		taw}', "");
Expect(0, 1837, '\p{Joining_Group=		taw}', "");
Expect(1, 1837, '\p{^Joining_Group=		taw}', "");
Expect(1, 1837, '\P{Joining_Group=		taw}', "");
Expect(0, 1837, '\P{^Joining_Group=		taw}', "");
Error('\p{Jg= 	Taw:=}');
Error('\P{Jg= 	Taw:=}');
Expect(1, 1836, '\p{Jg=taw}', "");
Expect(0, 1836, '\p{^Jg=taw}', "");
Expect(0, 1836, '\P{Jg=taw}', "");
Expect(1, 1836, '\P{^Jg=taw}', "");
Expect(0, 1837, '\p{Jg=taw}', "");
Expect(1, 1837, '\p{^Jg=taw}', "");
Expect(1, 1837, '\P{Jg=taw}', "");
Expect(0, 1837, '\P{^Jg=taw}', "");
Expect(1, 1836, '\p{Jg=-_taw}', "");
Expect(0, 1836, '\p{^Jg=-_taw}', "");
Expect(0, 1836, '\P{Jg=-_taw}', "");
Expect(1, 1836, '\P{^Jg=-_taw}', "");
Expect(0, 1837, '\p{Jg=-_taw}', "");
Expect(1, 1837, '\p{^Jg=-_taw}', "");
Expect(1, 1837, '\P{Jg=-_taw}', "");
Expect(0, 1837, '\P{^Jg=-_taw}', "");
Error('\p{Is_Joining_Group=:=	_TAW}');
Error('\P{Is_Joining_Group=:=	_TAW}');
Expect(1, 1836, '\p{Is_Joining_Group=taw}', "");
Expect(0, 1836, '\p{^Is_Joining_Group=taw}', "");
Expect(0, 1836, '\P{Is_Joining_Group=taw}', "");
Expect(1, 1836, '\P{^Is_Joining_Group=taw}', "");
Expect(0, 1837, '\p{Is_Joining_Group=taw}', "");
Expect(1, 1837, '\p{^Is_Joining_Group=taw}', "");
Expect(1, 1837, '\P{Is_Joining_Group=taw}', "");
Expect(0, 1837, '\P{^Is_Joining_Group=taw}', "");
Expect(1, 1836, '\p{Is_Joining_Group=	_TAW}', "");
Expect(0, 1836, '\p{^Is_Joining_Group=	_TAW}', "");
Expect(0, 1836, '\P{Is_Joining_Group=	_TAW}', "");
Expect(1, 1836, '\P{^Is_Joining_Group=	_TAW}', "");
Expect(0, 1837, '\p{Is_Joining_Group=	_TAW}', "");
Expect(1, 1837, '\p{^Is_Joining_Group=	_TAW}', "");
Expect(1, 1837, '\P{Is_Joining_Group=	_TAW}', "");
Expect(0, 1837, '\P{^Is_Joining_Group=	_TAW}', "");
Error('\p{Is_Jg= -taw/a/}');
Error('\P{Is_Jg= -taw/a/}');
Expect(1, 1836, '\p{Is_Jg:	taw}', "");
Expect(0, 1836, '\p{^Is_Jg:	taw}', "");
Expect(0, 1836, '\P{Is_Jg:	taw}', "");
Expect(1, 1836, '\P{^Is_Jg:	taw}', "");
Expect(0, 1837, '\p{Is_Jg:	taw}', "");
Expect(1, 1837, '\p{^Is_Jg:	taw}', "");
Expect(1, 1837, '\P{Is_Jg:	taw}', "");
Expect(0, 1837, '\P{^Is_Jg:	taw}', "");
Expect(1, 1836, '\p{Is_Jg=	 Taw}', "");
Expect(0, 1836, '\p{^Is_Jg=	 Taw}', "");
Expect(0, 1836, '\P{Is_Jg=	 Taw}', "");
Expect(1, 1836, '\P{^Is_Jg=	 Taw}', "");
Expect(0, 1837, '\p{Is_Jg=	 Taw}', "");
Expect(1, 1837, '\p{^Is_Jg=	 Taw}', "");
Expect(1, 1837, '\P{Is_Jg=	 Taw}', "");
Expect(0, 1837, '\P{^Is_Jg=	 Taw}', "");
Error('\p{Joining_Group=	_Final_Semkath/a/}');
Error('\P{Joining_Group=	_Final_Semkath/a/}');
Expect(1, 1828, '\p{Joining_Group=finalsemkath}', "");
Expect(0, 1828, '\p{^Joining_Group=finalsemkath}', "");
Expect(0, 1828, '\P{Joining_Group=finalsemkath}', "");
Expect(1, 1828, '\P{^Joining_Group=finalsemkath}', "");
Expect(0, 1829, '\p{Joining_Group=finalsemkath}', "");
Expect(1, 1829, '\p{^Joining_Group=finalsemkath}', "");
Expect(1, 1829, '\P{Joining_Group=finalsemkath}', "");
Expect(0, 1829, '\P{^Joining_Group=finalsemkath}', "");
Expect(1, 1828, '\p{Joining_Group=_-final_semkath}', "");
Expect(0, 1828, '\p{^Joining_Group=_-final_semkath}', "");
Expect(0, 1828, '\P{Joining_Group=_-final_semkath}', "");
Expect(1, 1828, '\P{^Joining_Group=_-final_semkath}', "");
Expect(0, 1829, '\p{Joining_Group=_-final_semkath}', "");
Expect(1, 1829, '\p{^Joining_Group=_-final_semkath}', "");
Expect(1, 1829, '\P{Joining_Group=_-final_semkath}', "");
Expect(0, 1829, '\P{^Joining_Group=_-final_semkath}', "");
Error('\p{Jg:   -/a/Final_semkath}');
Error('\P{Jg:   -/a/Final_semkath}');
Expect(1, 1828, '\p{Jg=finalsemkath}', "");
Expect(0, 1828, '\p{^Jg=finalsemkath}', "");
Expect(0, 1828, '\P{Jg=finalsemkath}', "");
Expect(1, 1828, '\P{^Jg=finalsemkath}', "");
Expect(0, 1829, '\p{Jg=finalsemkath}', "");
Expect(1, 1829, '\p{^Jg=finalsemkath}', "");
Expect(1, 1829, '\P{Jg=finalsemkath}', "");
Expect(0, 1829, '\P{^Jg=finalsemkath}', "");
Expect(1, 1828, '\p{Jg=  Final_Semkath}', "");
Expect(0, 1828, '\p{^Jg=  Final_Semkath}', "");
Expect(0, 1828, '\P{Jg=  Final_Semkath}', "");
Expect(1, 1828, '\P{^Jg=  Final_Semkath}', "");
Expect(0, 1829, '\p{Jg=  Final_Semkath}', "");
Expect(1, 1829, '\p{^Jg=  Final_Semkath}', "");
Expect(1, 1829, '\P{Jg=  Final_Semkath}', "");
Expect(0, 1829, '\P{^Jg=  Final_Semkath}', "");
Error('\p{Is_Joining_Group=/a/final_SEMKATH}');
Error('\P{Is_Joining_Group=/a/final_SEMKATH}');
Expect(1, 1828, '\p{Is_Joining_Group=finalsemkath}', "");
Expect(0, 1828, '\p{^Is_Joining_Group=finalsemkath}', "");
Expect(0, 1828, '\P{Is_Joining_Group=finalsemkath}', "");
Expect(1, 1828, '\P{^Is_Joining_Group=finalsemkath}', "");
Expect(0, 1829, '\p{Is_Joining_Group=finalsemkath}', "");
Expect(1, 1829, '\p{^Is_Joining_Group=finalsemkath}', "");
Expect(1, 1829, '\P{Is_Joining_Group=finalsemkath}', "");
Expect(0, 1829, '\P{^Is_Joining_Group=finalsemkath}', "");
Expect(1, 1828, '\p{Is_Joining_Group=FINAL_Semkath}', "");
Expect(0, 1828, '\p{^Is_Joining_Group=FINAL_Semkath}', "");
Expect(0, 1828, '\P{Is_Joining_Group=FINAL_Semkath}', "");
Expect(1, 1828, '\P{^Is_Joining_Group=FINAL_Semkath}', "");
Expect(0, 1829, '\p{Is_Joining_Group=FINAL_Semkath}', "");
Expect(1, 1829, '\p{^Is_Joining_Group=FINAL_Semkath}', "");
Expect(1, 1829, '\P{Is_Joining_Group=FINAL_Semkath}', "");
Expect(0, 1829, '\P{^Is_Joining_Group=FINAL_Semkath}', "");
Error('\p{Is_Jg:- Final_Semkath:=}');
Error('\P{Is_Jg:- Final_Semkath:=}');
Expect(1, 1828, '\p{Is_Jg=finalsemkath}', "");
Expect(0, 1828, '\p{^Is_Jg=finalsemkath}', "");
Expect(0, 1828, '\P{Is_Jg=finalsemkath}', "");
Expect(1, 1828, '\P{^Is_Jg=finalsemkath}', "");
Expect(0, 1829, '\p{Is_Jg=finalsemkath}', "");
Expect(1, 1829, '\p{^Is_Jg=finalsemkath}', "");
Expect(1, 1829, '\P{Is_Jg=finalsemkath}', "");
Expect(0, 1829, '\P{^Is_Jg=finalsemkath}', "");
Expect(1, 1828, '\p{Is_Jg=-Final_Semkath}', "");
Expect(0, 1828, '\p{^Is_Jg=-Final_Semkath}', "");
Expect(0, 1828, '\P{Is_Jg=-Final_Semkath}', "");
Expect(1, 1828, '\P{^Is_Jg=-Final_Semkath}', "");
Expect(0, 1829, '\p{Is_Jg=-Final_Semkath}', "");
Expect(1, 1829, '\p{^Is_Jg=-Final_Semkath}', "");
Expect(1, 1829, '\P{Is_Jg=-Final_Semkath}', "");
Expect(0, 1829, '\P{^Is_Jg=-Final_Semkath}', "");
Error('\p{Joining_Group=_ Knotted_Heh:=}');
Error('\P{Joining_Group=_ Knotted_Heh:=}');
Expect(1, 1791, '\p{Joining_Group=knottedheh}', "");
Expect(0, 1791, '\p{^Joining_Group=knottedheh}', "");
Expect(0, 1791, '\P{Joining_Group=knottedheh}', "");
Expect(1, 1791, '\P{^Joining_Group=knottedheh}', "");
Expect(0, 1792, '\p{Joining_Group=knottedheh}', "");
Expect(1, 1792, '\p{^Joining_Group=knottedheh}', "");
Expect(1, 1792, '\P{Joining_Group=knottedheh}', "");
Expect(0, 1792, '\P{^Joining_Group=knottedheh}', "");
Expect(1, 1791, '\p{Joining_Group= _Knotted_Heh}', "");
Expect(0, 1791, '\p{^Joining_Group= _Knotted_Heh}', "");
Expect(0, 1791, '\P{Joining_Group= _Knotted_Heh}', "");
Expect(1, 1791, '\P{^Joining_Group= _Knotted_Heh}', "");
Expect(0, 1792, '\p{Joining_Group= _Knotted_Heh}', "");
Expect(1, 1792, '\p{^Joining_Group= _Knotted_Heh}', "");
Expect(1, 1792, '\P{Joining_Group= _Knotted_Heh}', "");
Expect(0, 1792, '\P{^Joining_Group= _Knotted_Heh}', "");
Error('\p{Jg=--Knotted_Heh/a/}');
Error('\P{Jg=--Knotted_Heh/a/}');
Expect(1, 1791, '\p{Jg=knottedheh}', "");
Expect(0, 1791, '\p{^Jg=knottedheh}', "");
Expect(0, 1791, '\P{Jg=knottedheh}', "");
Expect(1, 1791, '\P{^Jg=knottedheh}', "");
Expect(0, 1792, '\p{Jg=knottedheh}', "");
Expect(1, 1792, '\p{^Jg=knottedheh}', "");
Expect(1, 1792, '\P{Jg=knottedheh}', "");
Expect(0, 1792, '\P{^Jg=knottedheh}', "");
Expect(1, 1791, '\p{Jg=--knotted_Heh}', "");
Expect(0, 1791, '\p{^Jg=--knotted_Heh}', "");
Expect(0, 1791, '\P{Jg=--knotted_Heh}', "");
Expect(1, 1791, '\P{^Jg=--knotted_Heh}', "");
Expect(0, 1792, '\p{Jg=--knotted_Heh}', "");
Expect(1, 1792, '\p{^Jg=--knotted_Heh}', "");
Expect(1, 1792, '\P{Jg=--knotted_Heh}', "");
Expect(0, 1792, '\P{^Jg=--knotted_Heh}', "");
Error('\p{Is_Joining_Group=/a/	_Knotted_Heh}');
Error('\P{Is_Joining_Group=/a/	_Knotted_Heh}');
Expect(1, 1791, '\p{Is_Joining_Group:   knottedheh}', "");
Expect(0, 1791, '\p{^Is_Joining_Group:   knottedheh}', "");
Expect(0, 1791, '\P{Is_Joining_Group:   knottedheh}', "");
Expect(1, 1791, '\P{^Is_Joining_Group:   knottedheh}', "");
Expect(0, 1792, '\p{Is_Joining_Group:   knottedheh}', "");
Expect(1, 1792, '\p{^Is_Joining_Group:   knottedheh}', "");
Expect(1, 1792, '\P{Is_Joining_Group:   knottedheh}', "");
Expect(0, 1792, '\P{^Is_Joining_Group:   knottedheh}', "");
Expect(1, 1791, '\p{Is_Joining_Group:   	_Knotted_heh}', "");
Expect(0, 1791, '\p{^Is_Joining_Group:   	_Knotted_heh}', "");
Expect(0, 1791, '\P{Is_Joining_Group:   	_Knotted_heh}', "");
Expect(1, 1791, '\P{^Is_Joining_Group:   	_Knotted_heh}', "");
Expect(0, 1792, '\p{Is_Joining_Group:   	_Knotted_heh}', "");
Expect(1, 1792, '\p{^Is_Joining_Group:   	_Knotted_heh}', "");
Expect(1, 1792, '\P{Is_Joining_Group:   	_Knotted_heh}', "");
Expect(0, 1792, '\P{^Is_Joining_Group:   	_Knotted_heh}', "");
Error('\p{Is_Jg=/a/knotted_Heh}');
Error('\P{Is_Jg=/a/knotted_Heh}');
Expect(1, 1791, '\p{Is_Jg=knottedheh}', "");
Expect(0, 1791, '\p{^Is_Jg=knottedheh}', "");
Expect(0, 1791, '\P{Is_Jg=knottedheh}', "");
Expect(1, 1791, '\P{^Is_Jg=knottedheh}', "");
Expect(0, 1792, '\p{Is_Jg=knottedheh}', "");
Expect(1, 1792, '\p{^Is_Jg=knottedheh}', "");
Expect(1, 1792, '\P{Is_Jg=knottedheh}', "");
Expect(0, 1792, '\P{^Is_Jg=knottedheh}', "");
Expect(1, 1791, '\p{Is_Jg: 	Knotted_Heh}', "");
Expect(0, 1791, '\p{^Is_Jg: 	Knotted_Heh}', "");
Expect(0, 1791, '\P{Is_Jg: 	Knotted_Heh}', "");
Expect(1, 1791, '\P{^Is_Jg: 	Knotted_Heh}', "");
Expect(0, 1792, '\p{Is_Jg: 	Knotted_Heh}', "");
Expect(1, 1792, '\p{^Is_Jg: 	Knotted_Heh}', "");
Expect(1, 1792, '\P{Is_Jg: 	Knotted_Heh}', "");
Expect(0, 1792, '\P{^Is_Jg: 	Knotted_Heh}', "");
Error('\p{Joining_Group=/a/ kaf}');
Error('\P{Joining_Group=/a/ kaf}');
Expect(1, 1919, '\p{Joining_Group:kaf}', "");
Expect(0, 1919, '\p{^Joining_Group:kaf}', "");
Expect(0, 1919, '\P{Joining_Group:kaf}', "");
Expect(1, 1919, '\P{^Joining_Group:kaf}', "");
Expect(0, 1920, '\p{Joining_Group:kaf}', "");
Expect(1, 1920, '\p{^Joining_Group:kaf}', "");
Expect(1, 1920, '\P{Joining_Group:kaf}', "");
Expect(0, 1920, '\P{^Joining_Group:kaf}', "");
Expect(1, 1919, '\p{Joining_Group=_	KAF}', "");
Expect(0, 1919, '\p{^Joining_Group=_	KAF}', "");
Expect(0, 1919, '\P{Joining_Group=_	KAF}', "");
Expect(1, 1919, '\P{^Joining_Group=_	KAF}', "");
Expect(0, 1920, '\p{Joining_Group=_	KAF}', "");
Expect(1, 1920, '\p{^Joining_Group=_	KAF}', "");
Expect(1, 1920, '\P{Joining_Group=_	KAF}', "");
Expect(0, 1920, '\P{^Joining_Group=_	KAF}', "");
Error('\p{Jg=/a/Kaf}');
Error('\P{Jg=/a/Kaf}');
Expect(1, 1919, '\p{Jg=kaf}', "");
Expect(0, 1919, '\p{^Jg=kaf}', "");
Expect(0, 1919, '\P{Jg=kaf}', "");
Expect(1, 1919, '\P{^Jg=kaf}', "");
Expect(0, 1920, '\p{Jg=kaf}', "");
Expect(1, 1920, '\p{^Jg=kaf}', "");
Expect(1, 1920, '\P{Jg=kaf}', "");
Expect(0, 1920, '\P{^Jg=kaf}', "");
Expect(1, 1919, '\p{Jg=__Kaf}', "");
Expect(0, 1919, '\p{^Jg=__Kaf}', "");
Expect(0, 1919, '\P{Jg=__Kaf}', "");
Expect(1, 1919, '\P{^Jg=__Kaf}', "");
Expect(0, 1920, '\p{Jg=__Kaf}', "");
Expect(1, 1920, '\p{^Jg=__Kaf}', "");
Expect(1, 1920, '\P{Jg=__Kaf}', "");
Expect(0, 1920, '\P{^Jg=__Kaf}', "");
Error('\p{Is_Joining_Group=_Kaf/a/}');
Error('\P{Is_Joining_Group=_Kaf/a/}');
Expect(1, 1919, '\p{Is_Joining_Group=kaf}', "");
Expect(0, 1919, '\p{^Is_Joining_Group=kaf}', "");
Expect(0, 1919, '\P{Is_Joining_Group=kaf}', "");
Expect(1, 1919, '\P{^Is_Joining_Group=kaf}', "");
Expect(0, 1920, '\p{Is_Joining_Group=kaf}', "");
Expect(1, 1920, '\p{^Is_Joining_Group=kaf}', "");
Expect(1, 1920, '\P{Is_Joining_Group=kaf}', "");
Expect(0, 1920, '\P{^Is_Joining_Group=kaf}', "");
Expect(1, 1919, '\p{Is_Joining_Group=- KAF}', "");
Expect(0, 1919, '\p{^Is_Joining_Group=- KAF}', "");
Expect(0, 1919, '\P{Is_Joining_Group=- KAF}', "");
Expect(1, 1919, '\P{^Is_Joining_Group=- KAF}', "");
Expect(0, 1920, '\p{Is_Joining_Group=- KAF}', "");
Expect(1, 1920, '\p{^Is_Joining_Group=- KAF}', "");
Expect(1, 1920, '\P{Is_Joining_Group=- KAF}', "");
Expect(0, 1920, '\P{^Is_Joining_Group=- KAF}', "");
Error('\p{Is_Jg= _Kaf/a/}');
Error('\P{Is_Jg= _Kaf/a/}');
Expect(1, 1919, '\p{Is_Jg=kaf}', "");
Expect(0, 1919, '\p{^Is_Jg=kaf}', "");
Expect(0, 1919, '\P{Is_Jg=kaf}', "");
Expect(1, 1919, '\P{^Is_Jg=kaf}', "");
Expect(0, 1920, '\p{Is_Jg=kaf}', "");
Expect(1, 1920, '\p{^Is_Jg=kaf}', "");
Expect(1, 1920, '\P{Is_Jg=kaf}', "");
Expect(0, 1920, '\P{^Is_Jg=kaf}', "");
Expect(1, 1919, '\p{Is_Jg=-Kaf}', "");
Expect(0, 1919, '\p{^Is_Jg=-Kaf}', "");
Expect(0, 1919, '\P{Is_Jg=-Kaf}', "");
Expect(1, 1919, '\P{^Is_Jg=-Kaf}', "");
Expect(0, 1920, '\p{Is_Jg=-Kaf}', "");
Expect(1, 1920, '\p{^Is_Jg=-Kaf}', "");
Expect(1, 1920, '\P{Is_Jg=-Kaf}', "");
Expect(0, 1920, '\P{^Is_Jg=-Kaf}', "");
Error('\p{Joining_Group:		FE:=}');
Error('\P{Joining_Group:		FE:=}');
Expect(1, 1871, '\p{Joining_Group=fe}', "");
Expect(0, 1871, '\p{^Joining_Group=fe}', "");
Expect(0, 1871, '\P{Joining_Group=fe}', "");
Expect(1, 1871, '\P{^Joining_Group=fe}', "");
Expect(0, 1872, '\p{Joining_Group=fe}', "");
Expect(1, 1872, '\p{^Joining_Group=fe}', "");
Expect(1, 1872, '\P{Joining_Group=fe}', "");
Expect(0, 1872, '\P{^Joining_Group=fe}', "");
Expect(1, 1871, '\p{Joining_Group=	 Fe}', "");
Expect(0, 1871, '\p{^Joining_Group=	 Fe}', "");
Expect(0, 1871, '\P{Joining_Group=	 Fe}', "");
Expect(1, 1871, '\P{^Joining_Group=	 Fe}', "");
Expect(0, 1872, '\p{Joining_Group=	 Fe}', "");
Expect(1, 1872, '\p{^Joining_Group=	 Fe}', "");
Expect(1, 1872, '\P{Joining_Group=	 Fe}', "");
Expect(0, 1872, '\P{^Joining_Group=	 Fe}', "");
Error('\p{Jg:/a/	_fe}');
Error('\P{Jg:/a/	_fe}');
Expect(1, 1871, '\p{Jg=fe}', "");
Expect(0, 1871, '\p{^Jg=fe}', "");
Expect(0, 1871, '\P{Jg=fe}', "");
Expect(1, 1871, '\P{^Jg=fe}', "");
Expect(0, 1872, '\p{Jg=fe}', "");
Expect(1, 1872, '\p{^Jg=fe}', "");
Expect(1, 1872, '\P{Jg=fe}', "");
Expect(0, 1872, '\P{^Jg=fe}', "");
Expect(1, 1871, '\p{Jg=-	fe}', "");
Expect(0, 1871, '\p{^Jg=-	fe}', "");
Expect(0, 1871, '\P{Jg=-	fe}', "");
Expect(1, 1871, '\P{^Jg=-	fe}', "");
Expect(0, 1872, '\p{Jg=-	fe}', "");
Expect(1, 1872, '\p{^Jg=-	fe}', "");
Expect(1, 1872, '\P{Jg=-	fe}', "");
Expect(0, 1872, '\P{^Jg=-	fe}', "");
Error('\p{Is_Joining_Group:   :=-_Fe}');
Error('\P{Is_Joining_Group:   :=-_Fe}');
Expect(1, 1871, '\p{Is_Joining_Group: fe}', "");
Expect(0, 1871, '\p{^Is_Joining_Group: fe}', "");
Expect(0, 1871, '\P{Is_Joining_Group: fe}', "");
Expect(1, 1871, '\P{^Is_Joining_Group: fe}', "");
Expect(0, 1872, '\p{Is_Joining_Group: fe}', "");
Expect(1, 1872, '\p{^Is_Joining_Group: fe}', "");
Expect(1, 1872, '\P{Is_Joining_Group: fe}', "");
Expect(0, 1872, '\P{^Is_Joining_Group: fe}', "");
Expect(1, 1871, '\p{Is_Joining_Group=- fe}', "");
Expect(0, 1871, '\p{^Is_Joining_Group=- fe}', "");
Expect(0, 1871, '\P{Is_Joining_Group=- fe}', "");
Expect(1, 1871, '\P{^Is_Joining_Group=- fe}', "");
Expect(0, 1872, '\p{Is_Joining_Group=- fe}', "");
Expect(1, 1872, '\p{^Is_Joining_Group=- fe}', "");
Expect(1, 1872, '\P{Is_Joining_Group=- fe}', "");
Expect(0, 1872, '\P{^Is_Joining_Group=- fe}', "");
Error('\p{Is_Jg=:=	Fe}');
Error('\P{Is_Jg=:=	Fe}');
Expect(1, 1871, '\p{Is_Jg=fe}', "");
Expect(0, 1871, '\p{^Is_Jg=fe}', "");
Expect(0, 1871, '\P{Is_Jg=fe}', "");
Expect(1, 1871, '\P{^Is_Jg=fe}', "");
Expect(0, 1872, '\p{Is_Jg=fe}', "");
Expect(1, 1872, '\p{^Is_Jg=fe}', "");
Expect(1, 1872, '\P{Is_Jg=fe}', "");
Expect(0, 1872, '\P{^Is_Jg=fe}', "");
Expect(1, 1871, '\p{Is_Jg= _FE}', "");
Expect(0, 1871, '\p{^Is_Jg= _FE}', "");
Expect(0, 1871, '\P{Is_Jg= _FE}', "");
Expect(1, 1871, '\P{^Is_Jg= _FE}', "");
Expect(0, 1872, '\p{Is_Jg= _FE}', "");
Expect(1, 1872, '\p{^Is_Jg= _FE}', "");
Expect(1, 1872, '\P{Is_Jg= _FE}', "");
Expect(0, 1872, '\P{^Is_Jg= _FE}', "");
Error('\p{Joining_Group=-Hamza_On_Heh_Goal:=}');
Error('\P{Joining_Group=-Hamza_On_Heh_Goal:=}');
Expect(1, 1731, '\p{Joining_Group=hamzaonhehgoal}', "");
Expect(0, 1731, '\p{^Joining_Group=hamzaonhehgoal}', "");
Expect(0, 1731, '\P{Joining_Group=hamzaonhehgoal}', "");
Expect(1, 1731, '\P{^Joining_Group=hamzaonhehgoal}', "");
Expect(0, 1732, '\p{Joining_Group=hamzaonhehgoal}', "");
Expect(1, 1732, '\p{^Joining_Group=hamzaonhehgoal}', "");
Expect(1, 1732, '\P{Joining_Group=hamzaonhehgoal}', "");
Expect(0, 1732, '\P{^Joining_Group=hamzaonhehgoal}', "");
Expect(1, 1731, '\p{Joining_Group= Hamza_On_Heh_Goal}', "");
Expect(0, 1731, '\p{^Joining_Group= Hamza_On_Heh_Goal}', "");
Expect(0, 1731, '\P{Joining_Group= Hamza_On_Heh_Goal}', "");
Expect(1, 1731, '\P{^Joining_Group= Hamza_On_Heh_Goal}', "");
Expect(0, 1732, '\p{Joining_Group= Hamza_On_Heh_Goal}', "");
Expect(1, 1732, '\p{^Joining_Group= Hamza_On_Heh_Goal}', "");
Expect(1, 1732, '\P{Joining_Group= Hamza_On_Heh_Goal}', "");
Expect(0, 1732, '\P{^Joining_Group= Hamza_On_Heh_Goal}', "");
Error('\p{Jg:   	Teh_Marbuta_goal:=}');
Error('\P{Jg:   	Teh_Marbuta_goal:=}');
Expect(1, 1731, '\p{Jg=tehmarbutagoal}', "");
Expect(0, 1731, '\p{^Jg=tehmarbutagoal}', "");
Expect(0, 1731, '\P{Jg=tehmarbutagoal}', "");
Expect(1, 1731, '\P{^Jg=tehmarbutagoal}', "");
Expect(0, 1732, '\p{Jg=tehmarbutagoal}', "");
Expect(1, 1732, '\p{^Jg=tehmarbutagoal}', "");
Expect(1, 1732, '\P{Jg=tehmarbutagoal}', "");
Expect(0, 1732, '\P{^Jg=tehmarbutagoal}', "");
Expect(1, 1731, '\p{Jg=__teh_Marbuta_goal}', "");
Expect(0, 1731, '\p{^Jg=__teh_Marbuta_goal}', "");
Expect(0, 1731, '\P{Jg=__teh_Marbuta_goal}', "");
Expect(1, 1731, '\P{^Jg=__teh_Marbuta_goal}', "");
Expect(0, 1732, '\p{Jg=__teh_Marbuta_goal}', "");
Expect(1, 1732, '\p{^Jg=__teh_Marbuta_goal}', "");
Expect(1, 1732, '\P{Jg=__teh_Marbuta_goal}', "");
Expect(0, 1732, '\P{^Jg=__teh_Marbuta_goal}', "");
Error('\p{Is_Joining_Group=  HAMZA_on_HEH_Goal:=}');
Error('\P{Is_Joining_Group=  HAMZA_on_HEH_Goal:=}');
Expect(1, 1731, '\p{Is_Joining_Group=hamzaonhehgoal}', "");
Expect(0, 1731, '\p{^Is_Joining_Group=hamzaonhehgoal}', "");
Expect(0, 1731, '\P{Is_Joining_Group=hamzaonhehgoal}', "");
Expect(1, 1731, '\P{^Is_Joining_Group=hamzaonhehgoal}', "");
Expect(0, 1732, '\p{Is_Joining_Group=hamzaonhehgoal}', "");
Expect(1, 1732, '\p{^Is_Joining_Group=hamzaonhehgoal}', "");
Expect(1, 1732, '\P{Is_Joining_Group=hamzaonhehgoal}', "");
Expect(0, 1732, '\P{^Is_Joining_Group=hamzaonhehgoal}', "");
Expect(1, 1731, '\p{Is_Joining_Group:HAMZA_On_HEH_GOAL}', "");
Expect(0, 1731, '\p{^Is_Joining_Group:HAMZA_On_HEH_GOAL}', "");
Expect(0, 1731, '\P{Is_Joining_Group:HAMZA_On_HEH_GOAL}', "");
Expect(1, 1731, '\P{^Is_Joining_Group:HAMZA_On_HEH_GOAL}', "");
Expect(0, 1732, '\p{Is_Joining_Group:HAMZA_On_HEH_GOAL}', "");
Expect(1, 1732, '\p{^Is_Joining_Group:HAMZA_On_HEH_GOAL}', "");
Expect(1, 1732, '\P{Is_Joining_Group:HAMZA_On_HEH_GOAL}', "");
Expect(0, 1732, '\P{^Is_Joining_Group:HAMZA_On_HEH_GOAL}', "");
Error('\p{Is_Jg=/a/Teh_Marbuta_Goal}');
Error('\P{Is_Jg=/a/Teh_Marbuta_Goal}');
Expect(1, 1731, '\p{Is_Jg=tehmarbutagoal}', "");
Expect(0, 1731, '\p{^Is_Jg=tehmarbutagoal}', "");
Expect(0, 1731, '\P{Is_Jg=tehmarbutagoal}', "");
Expect(1, 1731, '\P{^Is_Jg=tehmarbutagoal}', "");
Expect(0, 1732, '\p{Is_Jg=tehmarbutagoal}', "");
Expect(1, 1732, '\p{^Is_Jg=tehmarbutagoal}', "");
Expect(1, 1732, '\P{Is_Jg=tehmarbutagoal}', "");
Expect(0, 1732, '\P{^Is_Jg=tehmarbutagoal}', "");
Expect(1, 1731, '\p{Is_Jg: _-Teh_marbuta_Goal}', "");
Expect(0, 1731, '\p{^Is_Jg: _-Teh_marbuta_Goal}', "");
Expect(0, 1731, '\P{Is_Jg: _-Teh_marbuta_Goal}', "");
Expect(1, 1731, '\P{^Is_Jg: _-Teh_marbuta_Goal}', "");
Expect(0, 1732, '\p{Is_Jg: _-Teh_marbuta_Goal}', "");
Expect(1, 1732, '\p{^Is_Jg: _-Teh_marbuta_Goal}', "");
Expect(1, 1732, '\P{Is_Jg: _-Teh_marbuta_Goal}', "");
Expect(0, 1732, '\P{^Is_Jg: _-Teh_marbuta_Goal}', "");
Error('\p{Joining_Group= 	ALAPH:=}');
Error('\P{Joining_Group= 	ALAPH:=}');
Expect(1, 1808, '\p{Joining_Group=alaph}', "");
Expect(0, 1808, '\p{^Joining_Group=alaph}', "");
Expect(0, 1808, '\P{Joining_Group=alaph}', "");
Expect(1, 1808, '\P{^Joining_Group=alaph}', "");
Expect(0, 1809, '\p{Joining_Group=alaph}', "");
Expect(1, 1809, '\p{^Joining_Group=alaph}', "");
Expect(1, 1809, '\P{Joining_Group=alaph}', "");
Expect(0, 1809, '\P{^Joining_Group=alaph}', "");
Expect(1, 1808, '\p{Joining_Group=_ Alaph}', "");
Expect(0, 1808, '\p{^Joining_Group=_ Alaph}', "");
Expect(0, 1808, '\P{Joining_Group=_ Alaph}', "");
Expect(1, 1808, '\P{^Joining_Group=_ Alaph}', "");
Expect(0, 1809, '\p{Joining_Group=_ Alaph}', "");
Expect(1, 1809, '\p{^Joining_Group=_ Alaph}', "");
Expect(1, 1809, '\P{Joining_Group=_ Alaph}', "");
Expect(0, 1809, '\P{^Joining_Group=_ Alaph}', "");
Error('\p{Jg=_:=Alaph}');
Error('\P{Jg=_:=Alaph}');
Expect(1, 1808, '\p{Jg=alaph}', "");
Expect(0, 1808, '\p{^Jg=alaph}', "");
Expect(0, 1808, '\P{Jg=alaph}', "");
Expect(1, 1808, '\P{^Jg=alaph}', "");
Expect(0, 1809, '\p{Jg=alaph}', "");
Expect(1, 1809, '\p{^Jg=alaph}', "");
Expect(1, 1809, '\P{Jg=alaph}', "");
Expect(0, 1809, '\P{^Jg=alaph}', "");
Expect(1, 1808, '\p{Jg=	 alaph}', "");
Expect(0, 1808, '\p{^Jg=	 alaph}', "");
Expect(0, 1808, '\P{Jg=	 alaph}', "");
Expect(1, 1808, '\P{^Jg=	 alaph}', "");
Expect(0, 1809, '\p{Jg=	 alaph}', "");
Expect(1, 1809, '\p{^Jg=	 alaph}', "");
Expect(1, 1809, '\P{Jg=	 alaph}', "");
Expect(0, 1809, '\P{^Jg=	 alaph}', "");
Error('\p{Is_Joining_Group=:= _alaph}');
Error('\P{Is_Joining_Group=:= _alaph}');
Expect(1, 1808, '\p{Is_Joining_Group=alaph}', "");
Expect(0, 1808, '\p{^Is_Joining_Group=alaph}', "");
Expect(0, 1808, '\P{Is_Joining_Group=alaph}', "");
Expect(1, 1808, '\P{^Is_Joining_Group=alaph}', "");
Expect(0, 1809, '\p{Is_Joining_Group=alaph}', "");
Expect(1, 1809, '\p{^Is_Joining_Group=alaph}', "");
Expect(1, 1809, '\P{Is_Joining_Group=alaph}', "");
Expect(0, 1809, '\P{^Is_Joining_Group=alaph}', "");
Expect(1, 1808, '\p{Is_Joining_Group:_Alaph}', "");
Expect(0, 1808, '\p{^Is_Joining_Group:_Alaph}', "");
Expect(0, 1808, '\P{Is_Joining_Group:_Alaph}', "");
Expect(1, 1808, '\P{^Is_Joining_Group:_Alaph}', "");
Expect(0, 1809, '\p{Is_Joining_Group:_Alaph}', "");
Expect(1, 1809, '\p{^Is_Joining_Group:_Alaph}', "");
Expect(1, 1809, '\P{Is_Joining_Group:_Alaph}', "");
Expect(0, 1809, '\P{^Is_Joining_Group:_Alaph}', "");
Error('\p{Is_Jg=/a/ALAPH}');
Error('\P{Is_Jg=/a/ALAPH}');
Expect(1, 1808, '\p{Is_Jg=alaph}', "");
Expect(0, 1808, '\p{^Is_Jg=alaph}', "");
Expect(0, 1808, '\P{Is_Jg=alaph}', "");
Expect(1, 1808, '\P{^Is_Jg=alaph}', "");
Expect(0, 1809, '\p{Is_Jg=alaph}', "");
Expect(1, 1809, '\p{^Is_Jg=alaph}', "");
Expect(1, 1809, '\P{Is_Jg=alaph}', "");
Expect(0, 1809, '\P{^Is_Jg=alaph}', "");
Expect(1, 1808, '\p{Is_Jg:   _-ALAPH}', "");
Expect(0, 1808, '\p{^Is_Jg:   _-ALAPH}', "");
Expect(0, 1808, '\P{Is_Jg:   _-ALAPH}', "");
Expect(1, 1808, '\P{^Is_Jg:   _-ALAPH}', "");
Expect(0, 1809, '\p{Is_Jg:   _-ALAPH}', "");
Expect(1, 1809, '\p{^Is_Jg:   _-ALAPH}', "");
Expect(1, 1809, '\P{Is_Jg:   _-ALAPH}', "");
Expect(0, 1809, '\P{^Is_Jg:   _-ALAPH}', "");
Error('\p{Joining_Group=/a/ 	lamadh}');
Error('\P{Joining_Group=/a/ 	lamadh}');
Expect(1, 1824, '\p{Joining_Group=lamadh}', "");
Expect(0, 1824, '\p{^Joining_Group=lamadh}', "");
Expect(0, 1824, '\P{Joining_Group=lamadh}', "");
Expect(1, 1824, '\P{^Joining_Group=lamadh}', "");
Expect(0, 1825, '\p{Joining_Group=lamadh}', "");
Expect(1, 1825, '\p{^Joining_Group=lamadh}', "");
Expect(1, 1825, '\P{Joining_Group=lamadh}', "");
Expect(0, 1825, '\P{^Joining_Group=lamadh}', "");
Expect(1, 1824, '\p{Joining_Group=  lamadh}', "");
Expect(0, 1824, '\p{^Joining_Group=  lamadh}', "");
Expect(0, 1824, '\P{Joining_Group=  lamadh}', "");
Expect(1, 1824, '\P{^Joining_Group=  lamadh}', "");
Expect(0, 1825, '\p{Joining_Group=  lamadh}', "");
Expect(1, 1825, '\p{^Joining_Group=  lamadh}', "");
Expect(1, 1825, '\P{Joining_Group=  lamadh}', "");
Expect(0, 1825, '\P{^Joining_Group=  lamadh}', "");
Error('\p{Jg= 	Lamadh:=}');
Error('\P{Jg= 	Lamadh:=}');
Expect(1, 1824, '\p{Jg=lamadh}', "");
Expect(0, 1824, '\p{^Jg=lamadh}', "");
Expect(0, 1824, '\P{Jg=lamadh}', "");
Expect(1, 1824, '\P{^Jg=lamadh}', "");
Expect(0, 1825, '\p{Jg=lamadh}', "");
Expect(1, 1825, '\p{^Jg=lamadh}', "");
Expect(1, 1825, '\P{Jg=lamadh}', "");
Expect(0, 1825, '\P{^Jg=lamadh}', "");
Expect(1, 1824, '\p{Jg=_Lamadh}', "");
Expect(0, 1824, '\p{^Jg=_Lamadh}', "");
Expect(0, 1824, '\P{Jg=_Lamadh}', "");
Expect(1, 1824, '\P{^Jg=_Lamadh}', "");
Expect(0, 1825, '\p{Jg=_Lamadh}', "");
Expect(1, 1825, '\p{^Jg=_Lamadh}', "");
Expect(1, 1825, '\P{Jg=_Lamadh}', "");
Expect(0, 1825, '\P{^Jg=_Lamadh}', "");
Error('\p{Is_Joining_Group=/a/	Lamadh}');
Error('\P{Is_Joining_Group=/a/	Lamadh}');
Expect(1, 1824, '\p{Is_Joining_Group=lamadh}', "");
Expect(0, 1824, '\p{^Is_Joining_Group=lamadh}', "");
Expect(0, 1824, '\P{Is_Joining_Group=lamadh}', "");
Expect(1, 1824, '\P{^Is_Joining_Group=lamadh}', "");
Expect(0, 1825, '\p{Is_Joining_Group=lamadh}', "");
Expect(1, 1825, '\p{^Is_Joining_Group=lamadh}', "");
Expect(1, 1825, '\P{Is_Joining_Group=lamadh}', "");
Expect(0, 1825, '\P{^Is_Joining_Group=lamadh}', "");
Expect(1, 1824, '\p{Is_Joining_Group:   	-Lamadh}', "");
Expect(0, 1824, '\p{^Is_Joining_Group:   	-Lamadh}', "");
Expect(0, 1824, '\P{Is_Joining_Group:   	-Lamadh}', "");
Expect(1, 1824, '\P{^Is_Joining_Group:   	-Lamadh}', "");
Expect(0, 1825, '\p{Is_Joining_Group:   	-Lamadh}', "");
Expect(1, 1825, '\p{^Is_Joining_Group:   	-Lamadh}', "");
Expect(1, 1825, '\P{Is_Joining_Group:   	-Lamadh}', "");
Expect(0, 1825, '\P{^Is_Joining_Group:   	-Lamadh}', "");
Error('\p{Is_Jg=	/a/Lamadh}');
Error('\P{Is_Jg=	/a/Lamadh}');
Expect(1, 1824, '\p{Is_Jg=lamadh}', "");
Expect(0, 1824, '\p{^Is_Jg=lamadh}', "");
Expect(0, 1824, '\P{Is_Jg=lamadh}', "");
Expect(1, 1824, '\P{^Is_Jg=lamadh}', "");
Expect(0, 1825, '\p{Is_Jg=lamadh}', "");
Expect(1, 1825, '\p{^Is_Jg=lamadh}', "");
Expect(1, 1825, '\P{Is_Jg=lamadh}', "");
Expect(0, 1825, '\P{^Is_Jg=lamadh}', "");
Expect(1, 1824, '\p{Is_Jg:   --lamadh}', "");
Expect(0, 1824, '\p{^Is_Jg:   --lamadh}', "");
Expect(0, 1824, '\P{Is_Jg:   --lamadh}', "");
Expect(1, 1824, '\P{^Is_Jg:   --lamadh}', "");
Expect(0, 1825, '\p{Is_Jg:   --lamadh}', "");
Expect(1, 1825, '\p{^Is_Jg:   --lamadh}', "");
Expect(1, 1825, '\P{Is_Jg:   --lamadh}', "");
Expect(0, 1825, '\P{^Is_Jg:   --lamadh}', "");
Error('\p{Joining_Group=_/a/e}');
Error('\P{Joining_Group=_/a/e}');
Expect(1, 1829, '\p{Joining_Group=e}', "");
Expect(0, 1829, '\p{^Joining_Group=e}', "");
Expect(0, 1829, '\P{Joining_Group=e}', "");
Expect(1, 1829, '\P{^Joining_Group=e}', "");
Expect(0, 1830, '\p{Joining_Group=e}', "");
Expect(1, 1830, '\p{^Joining_Group=e}', "");
Expect(1, 1830, '\P{Joining_Group=e}', "");
Expect(0, 1830, '\P{^Joining_Group=e}', "");
Expect(1, 1829, '\p{Joining_Group=	-E}', "");
Expect(0, 1829, '\p{^Joining_Group=	-E}', "");
Expect(0, 1829, '\P{Joining_Group=	-E}', "");
Expect(1, 1829, '\P{^Joining_Group=	-E}', "");
Expect(0, 1830, '\p{Joining_Group=	-E}', "");
Expect(1, 1830, '\p{^Joining_Group=	-E}', "");
Expect(1, 1830, '\P{Joining_Group=	-E}', "");
Expect(0, 1830, '\P{^Joining_Group=	-E}', "");
Error('\p{Jg=/a/_ E}');
Error('\P{Jg=/a/_ E}');
Expect(1, 1829, '\p{Jg=e}', "");
Expect(0, 1829, '\p{^Jg=e}', "");
Expect(0, 1829, '\P{Jg=e}', "");
Expect(1, 1829, '\P{^Jg=e}', "");
Expect(0, 1830, '\p{Jg=e}', "");
Expect(1, 1830, '\p{^Jg=e}', "");
Expect(1, 1830, '\P{Jg=e}', "");
Expect(0, 1830, '\P{^Jg=e}', "");
Expect(1, 1829, '\p{Jg=	_E}', "");
Expect(0, 1829, '\p{^Jg=	_E}', "");
Expect(0, 1829, '\P{Jg=	_E}', "");
Expect(1, 1829, '\P{^Jg=	_E}', "");
Expect(0, 1830, '\p{Jg=	_E}', "");
Expect(1, 1830, '\p{^Jg=	_E}', "");
Expect(1, 1830, '\P{Jg=	_E}', "");
Expect(0, 1830, '\P{^Jg=	_E}', "");
Error('\p{Is_Joining_Group=	e/a/}');
Error('\P{Is_Joining_Group=	e/a/}');
Expect(1, 1829, '\p{Is_Joining_Group=e}', "");
Expect(0, 1829, '\p{^Is_Joining_Group=e}', "");
Expect(0, 1829, '\P{Is_Joining_Group=e}', "");
Expect(1, 1829, '\P{^Is_Joining_Group=e}', "");
Expect(0, 1830, '\p{Is_Joining_Group=e}', "");
Expect(1, 1830, '\p{^Is_Joining_Group=e}', "");
Expect(1, 1830, '\P{Is_Joining_Group=e}', "");
Expect(0, 1830, '\P{^Is_Joining_Group=e}', "");
Expect(1, 1829, '\p{Is_Joining_Group= E}', "");
Expect(0, 1829, '\p{^Is_Joining_Group= E}', "");
Expect(0, 1829, '\P{Is_Joining_Group= E}', "");
Expect(1, 1829, '\P{^Is_Joining_Group= E}', "");
Expect(0, 1830, '\p{Is_Joining_Group= E}', "");
Expect(1, 1830, '\p{^Is_Joining_Group= E}', "");
Expect(1, 1830, '\P{Is_Joining_Group= E}', "");
Expect(0, 1830, '\P{^Is_Joining_Group= E}', "");
Error('\p{Is_Jg=:=__E}');
Error('\P{Is_Jg=:=__E}');
Expect(1, 1829, '\p{Is_Jg=e}', "");
Expect(0, 1829, '\p{^Is_Jg=e}', "");
Expect(0, 1829, '\P{Is_Jg=e}', "");
Expect(1, 1829, '\P{^Is_Jg=e}', "");
Expect(0, 1830, '\p{Is_Jg=e}', "");
Expect(1, 1830, '\p{^Is_Jg=e}', "");
Expect(1, 1830, '\P{Is_Jg=e}', "");
Expect(0, 1830, '\P{^Is_Jg=e}', "");
Expect(1, 1829, '\p{Is_Jg=-e}', "");
Expect(0, 1829, '\p{^Is_Jg=-e}', "");
Expect(0, 1829, '\P{Is_Jg=-e}', "");
Expect(1, 1829, '\P{^Is_Jg=-e}', "");
Expect(0, 1830, '\p{Is_Jg=-e}', "");
Expect(1, 1830, '\p{^Is_Jg=-e}', "");
Expect(1, 1830, '\P{Is_Jg=-e}', "");
Expect(0, 1830, '\P{^Is_Jg=-e}', "");
Error('\p{Joining_Group=_/a/Yeh_BARREE}');
Error('\P{Joining_Group=_/a/Yeh_BARREE}');
Expect(1, 1747, '\p{Joining_Group=yehbarree}', "");
Expect(0, 1747, '\p{^Joining_Group=yehbarree}', "");
Expect(0, 1747, '\P{Joining_Group=yehbarree}', "");
Expect(1, 1747, '\P{^Joining_Group=yehbarree}', "");
Expect(0, 1748, '\p{Joining_Group=yehbarree}', "");
Expect(1, 1748, '\p{^Joining_Group=yehbarree}', "");
Expect(1, 1748, '\P{Joining_Group=yehbarree}', "");
Expect(0, 1748, '\P{^Joining_Group=yehbarree}', "");
Expect(1, 1747, '\p{Joining_Group=-	YEH_barree}', "");
Expect(0, 1747, '\p{^Joining_Group=-	YEH_barree}', "");
Expect(0, 1747, '\P{Joining_Group=-	YEH_barree}', "");
Expect(1, 1747, '\P{^Joining_Group=-	YEH_barree}', "");
Expect(0, 1748, '\p{Joining_Group=-	YEH_barree}', "");
Expect(1, 1748, '\p{^Joining_Group=-	YEH_barree}', "");
Expect(1, 1748, '\P{Joining_Group=-	YEH_barree}', "");
Expect(0, 1748, '\P{^Joining_Group=-	YEH_barree}', "");
Error('\p{Jg=/a/YEH_Barree}');
Error('\P{Jg=/a/YEH_Barree}');
Expect(1, 1747, '\p{Jg=yehbarree}', "");
Expect(0, 1747, '\p{^Jg=yehbarree}', "");
Expect(0, 1747, '\P{Jg=yehbarree}', "");
Expect(1, 1747, '\P{^Jg=yehbarree}', "");
Expect(0, 1748, '\p{Jg=yehbarree}', "");
Expect(1, 1748, '\p{^Jg=yehbarree}', "");
Expect(1, 1748, '\P{Jg=yehbarree}', "");
Expect(0, 1748, '\P{^Jg=yehbarree}', "");
Expect(1, 1747, '\p{Jg=-YEH_Barree}', "");
Expect(0, 1747, '\p{^Jg=-YEH_Barree}', "");
Expect(0, 1747, '\P{Jg=-YEH_Barree}', "");
Expect(1, 1747, '\P{^Jg=-YEH_Barree}', "");
Expect(0, 1748, '\p{Jg=-YEH_Barree}', "");
Expect(1, 1748, '\p{^Jg=-YEH_Barree}', "");
Expect(1, 1748, '\P{Jg=-YEH_Barree}', "");
Expect(0, 1748, '\P{^Jg=-YEH_Barree}', "");
Error('\p{Is_Joining_Group=		yeh_BARREE:=}');
Error('\P{Is_Joining_Group=		yeh_BARREE:=}');
Expect(1, 1747, '\p{Is_Joining_Group=yehbarree}', "");
Expect(0, 1747, '\p{^Is_Joining_Group=yehbarree}', "");
Expect(0, 1747, '\P{Is_Joining_Group=yehbarree}', "");
Expect(1, 1747, '\P{^Is_Joining_Group=yehbarree}', "");
Expect(0, 1748, '\p{Is_Joining_Group=yehbarree}', "");
Expect(1, 1748, '\p{^Is_Joining_Group=yehbarree}', "");
Expect(1, 1748, '\P{Is_Joining_Group=yehbarree}', "");
Expect(0, 1748, '\P{^Is_Joining_Group=yehbarree}', "");
Expect(1, 1747, '\p{Is_Joining_Group=Yeh_BARREE}', "");
Expect(0, 1747, '\p{^Is_Joining_Group=Yeh_BARREE}', "");
Expect(0, 1747, '\P{Is_Joining_Group=Yeh_BARREE}', "");
Expect(1, 1747, '\P{^Is_Joining_Group=Yeh_BARREE}', "");
Expect(0, 1748, '\p{Is_Joining_Group=Yeh_BARREE}', "");
Expect(1, 1748, '\p{^Is_Joining_Group=Yeh_BARREE}', "");
Expect(1, 1748, '\P{Is_Joining_Group=Yeh_BARREE}', "");
Expect(0, 1748, '\P{^Is_Joining_Group=Yeh_BARREE}', "");
Error('\p{Is_Jg=/a/_ Yeh_BARREE}');
Error('\P{Is_Jg=/a/_ Yeh_BARREE}');
Expect(1, 1747, '\p{Is_Jg=yehbarree}', "");
Expect(0, 1747, '\p{^Is_Jg=yehbarree}', "");
Expect(0, 1747, '\P{Is_Jg=yehbarree}', "");
Expect(1, 1747, '\P{^Is_Jg=yehbarree}', "");
Expect(0, 1748, '\p{Is_Jg=yehbarree}', "");
Expect(1, 1748, '\p{^Is_Jg=yehbarree}', "");
Expect(1, 1748, '\P{Is_Jg=yehbarree}', "");
Expect(0, 1748, '\P{^Is_Jg=yehbarree}', "");
Expect(1, 1747, '\p{Is_Jg=_-Yeh_Barree}', "");
Expect(0, 1747, '\p{^Is_Jg=_-Yeh_Barree}', "");
Expect(0, 1747, '\P{Is_Jg=_-Yeh_Barree}', "");
Expect(1, 1747, '\P{^Is_Jg=_-Yeh_Barree}', "");
Expect(0, 1748, '\p{Is_Jg=_-Yeh_Barree}', "");
Expect(1, 1748, '\p{^Is_Jg=_-Yeh_Barree}', "");
Expect(1, 1748, '\P{Is_Jg=_-Yeh_Barree}', "");
Expect(0, 1748, '\P{^Is_Jg=_-Yeh_Barree}', "");
Error('\p{Joining_Group=/a/	nya}');
Error('\P{Joining_Group=/a/	nya}');
Expect(1, 1725, '\p{Joining_Group=nya}', "");
Expect(0, 1725, '\p{^Joining_Group=nya}', "");
Expect(0, 1725, '\P{Joining_Group=nya}', "");
Expect(1, 1725, '\P{^Joining_Group=nya}', "");
Expect(0, 1726, '\p{Joining_Group=nya}', "");
Expect(1, 1726, '\p{^Joining_Group=nya}', "");
Expect(1, 1726, '\P{Joining_Group=nya}', "");
Expect(0, 1726, '\P{^Joining_Group=nya}', "");
Expect(1, 1725, '\p{Joining_Group:     NYA}', "");
Expect(0, 1725, '\p{^Joining_Group:     NYA}', "");
Expect(0, 1725, '\P{Joining_Group:     NYA}', "");
Expect(1, 1725, '\P{^Joining_Group:     NYA}', "");
Expect(0, 1726, '\p{Joining_Group:     NYA}', "");
Expect(1, 1726, '\p{^Joining_Group:     NYA}', "");
Expect(1, 1726, '\P{Joining_Group:     NYA}', "");
Expect(0, 1726, '\P{^Joining_Group:     NYA}', "");
Error('\p{Jg=__nya/a/}');
Error('\P{Jg=__nya/a/}');
Expect(1, 1725, '\p{Jg=nya}', "");
Expect(0, 1725, '\p{^Jg=nya}', "");
Expect(0, 1725, '\P{Jg=nya}', "");
Expect(1, 1725, '\P{^Jg=nya}', "");
Expect(0, 1726, '\p{Jg=nya}', "");
Expect(1, 1726, '\p{^Jg=nya}', "");
Expect(1, 1726, '\P{Jg=nya}', "");
Expect(0, 1726, '\P{^Jg=nya}', "");
Expect(1, 1725, '\p{Jg=	-Nya}', "");
Expect(0, 1725, '\p{^Jg=	-Nya}', "");
Expect(0, 1725, '\P{Jg=	-Nya}', "");
Expect(1, 1725, '\P{^Jg=	-Nya}', "");
Expect(0, 1726, '\p{Jg=	-Nya}', "");
Expect(1, 1726, '\p{^Jg=	-Nya}', "");
Expect(1, 1726, '\P{Jg=	-Nya}', "");
Expect(0, 1726, '\P{^Jg=	-Nya}', "");
Error('\p{Is_Joining_Group=-:=Nya}');
Error('\P{Is_Joining_Group=-:=Nya}');
Expect(1, 1725, '\p{Is_Joining_Group=nya}', "");
Expect(0, 1725, '\p{^Is_Joining_Group=nya}', "");
Expect(0, 1725, '\P{Is_Joining_Group=nya}', "");
Expect(1, 1725, '\P{^Is_Joining_Group=nya}', "");
Expect(0, 1726, '\p{Is_Joining_Group=nya}', "");
Expect(1, 1726, '\p{^Is_Joining_Group=nya}', "");
Expect(1, 1726, '\P{Is_Joining_Group=nya}', "");
Expect(0, 1726, '\P{^Is_Joining_Group=nya}', "");
Expect(1, 1725, '\p{Is_Joining_Group=_NYA}', "");
Expect(0, 1725, '\p{^Is_Joining_Group=_NYA}', "");
Expect(0, 1725, '\P{Is_Joining_Group=_NYA}', "");
Expect(1, 1725, '\P{^Is_Joining_Group=_NYA}', "");
Expect(0, 1726, '\p{Is_Joining_Group=_NYA}', "");
Expect(1, 1726, '\p{^Is_Joining_Group=_NYA}', "");
Expect(1, 1726, '\P{Is_Joining_Group=_NYA}', "");
Expect(0, 1726, '\P{^Is_Joining_Group=_NYA}', "");
Error('\p{Is_Jg=/a/_NYA}');
Error('\P{Is_Jg=/a/_NYA}');
Expect(1, 1725, '\p{Is_Jg=nya}', "");
Expect(0, 1725, '\p{^Is_Jg=nya}', "");
Expect(0, 1725, '\P{Is_Jg=nya}', "");
Expect(1, 1725, '\P{^Is_Jg=nya}', "");
Expect(0, 1726, '\p{Is_Jg=nya}', "");
Expect(1, 1726, '\p{^Is_Jg=nya}', "");
Expect(1, 1726, '\P{Is_Jg=nya}', "");
Expect(0, 1726, '\P{^Is_Jg=nya}', "");
Expect(1, 1725, '\p{Is_Jg=	Nya}', "");
Expect(0, 1725, '\p{^Is_Jg=	Nya}', "");
Expect(0, 1725, '\P{Is_Jg=	Nya}', "");
Expect(1, 1725, '\P{^Is_Jg=	Nya}', "");
Expect(0, 1726, '\p{Is_Jg=	Nya}', "");
Expect(1, 1726, '\p{^Is_Jg=	Nya}', "");
Expect(1, 1726, '\P{Is_Jg=	Nya}', "");
Expect(0, 1726, '\P{^Is_Jg=	Nya}', "");
Error('\p{Joining_Group=/a/ NUN}');
Error('\P{Joining_Group=/a/ NUN}');
Expect(1, 1826, '\p{Joining_Group=nun}', "");
Expect(0, 1826, '\p{^Joining_Group=nun}', "");
Expect(0, 1826, '\P{Joining_Group=nun}', "");
Expect(1, 1826, '\P{^Joining_Group=nun}', "");
Expect(0, 1827, '\p{Joining_Group=nun}', "");
Expect(1, 1827, '\p{^Joining_Group=nun}', "");
Expect(1, 1827, '\P{Joining_Group=nun}', "");
Expect(0, 1827, '\P{^Joining_Group=nun}', "");
Expect(1, 1826, '\p{Joining_Group:_Nun}', "");
Expect(0, 1826, '\p{^Joining_Group:_Nun}', "");
Expect(0, 1826, '\P{Joining_Group:_Nun}', "");
Expect(1, 1826, '\P{^Joining_Group:_Nun}', "");
Expect(0, 1827, '\p{Joining_Group:_Nun}', "");
Expect(1, 1827, '\p{^Joining_Group:_Nun}', "");
Expect(1, 1827, '\P{Joining_Group:_Nun}', "");
Expect(0, 1827, '\P{^Joining_Group:_Nun}', "");
Error('\p{Jg=/a/	_Nun}');
Error('\P{Jg=/a/	_Nun}');
Expect(1, 1826, '\p{Jg=nun}', "");
Expect(0, 1826, '\p{^Jg=nun}', "");
Expect(0, 1826, '\P{Jg=nun}', "");
Expect(1, 1826, '\P{^Jg=nun}', "");
Expect(0, 1827, '\p{Jg=nun}', "");
Expect(1, 1827, '\p{^Jg=nun}', "");
Expect(1, 1827, '\P{Jg=nun}', "");
Expect(0, 1827, '\P{^Jg=nun}', "");
Expect(1, 1826, '\p{Jg=-Nun}', "");
Expect(0, 1826, '\p{^Jg=-Nun}', "");
Expect(0, 1826, '\P{Jg=-Nun}', "");
Expect(1, 1826, '\P{^Jg=-Nun}', "");
Expect(0, 1827, '\p{Jg=-Nun}', "");
Expect(1, 1827, '\p{^Jg=-Nun}', "");
Expect(1, 1827, '\P{Jg=-Nun}', "");
Expect(0, 1827, '\P{^Jg=-Nun}', "");
Error('\p{Is_Joining_Group=-:=Nun}');
Error('\P{Is_Joining_Group=-:=Nun}');
Expect(1, 1826, '\p{Is_Joining_Group=nun}', "");
Expect(0, 1826, '\p{^Is_Joining_Group=nun}', "");
Expect(0, 1826, '\P{Is_Joining_Group=nun}', "");
Expect(1, 1826, '\P{^Is_Joining_Group=nun}', "");
Expect(0, 1827, '\p{Is_Joining_Group=nun}', "");
Expect(1, 1827, '\p{^Is_Joining_Group=nun}', "");
Expect(1, 1827, '\P{Is_Joining_Group=nun}', "");
Expect(0, 1827, '\P{^Is_Joining_Group=nun}', "");
Expect(1, 1826, '\p{Is_Joining_Group=_	Nun}', "");
Expect(0, 1826, '\p{^Is_Joining_Group=_	Nun}', "");
Expect(0, 1826, '\P{Is_Joining_Group=_	Nun}', "");
Expect(1, 1826, '\P{^Is_Joining_Group=_	Nun}', "");
Expect(0, 1827, '\p{Is_Joining_Group=_	Nun}', "");
Expect(1, 1827, '\p{^Is_Joining_Group=_	Nun}', "");
Expect(1, 1827, '\P{Is_Joining_Group=_	Nun}', "");
Expect(0, 1827, '\P{^Is_Joining_Group=_	Nun}', "");
Error('\p{Is_Jg=/a/	-NUN}');
Error('\P{Is_Jg=/a/	-NUN}');
Expect(1, 1826, '\p{Is_Jg=nun}', "");
Expect(0, 1826, '\p{^Is_Jg=nun}', "");
Expect(0, 1826, '\P{Is_Jg=nun}', "");
Expect(1, 1826, '\P{^Is_Jg=nun}', "");
Expect(0, 1827, '\p{Is_Jg=nun}', "");
Expect(1, 1827, '\p{^Is_Jg=nun}', "");
Expect(1, 1827, '\P{Is_Jg=nun}', "");
Expect(0, 1827, '\P{^Is_Jg=nun}', "");
Expect(1, 1826, '\p{Is_Jg:	--nun}', "");
Expect(0, 1826, '\p{^Is_Jg:	--nun}', "");
Expect(0, 1826, '\P{Is_Jg:	--nun}', "");
Expect(1, 1826, '\P{^Is_Jg:	--nun}', "");
Expect(0, 1827, '\p{Is_Jg:	--nun}', "");
Expect(1, 1827, '\p{^Is_Jg:	--nun}', "");
Expect(1, 1827, '\P{Is_Jg:	--nun}', "");
Expect(0, 1827, '\P{^Is_Jg:	--nun}', "");
Error('\p{Joining_Group=  Lam/a/}');
Error('\P{Joining_Group=  Lam/a/}');
Expect(1, 2214, '\p{Joining_Group=lam}', "");
Expect(0, 2214, '\p{^Joining_Group=lam}', "");
Expect(0, 2214, '\P{Joining_Group=lam}', "");
Expect(1, 2214, '\P{^Joining_Group=lam}', "");
Expect(0, 2215, '\p{Joining_Group=lam}', "");
Expect(1, 2215, '\p{^Joining_Group=lam}', "");
Expect(1, 2215, '\P{Joining_Group=lam}', "");
Expect(0, 2215, '\P{^Joining_Group=lam}', "");
Expect(1, 2214, '\p{Joining_Group=_Lam}', "");
Expect(0, 2214, '\p{^Joining_Group=_Lam}', "");
Expect(0, 2214, '\P{Joining_Group=_Lam}', "");
Expect(1, 2214, '\P{^Joining_Group=_Lam}', "");
Expect(0, 2215, '\p{Joining_Group=_Lam}', "");
Expect(1, 2215, '\p{^Joining_Group=_Lam}', "");
Expect(1, 2215, '\P{Joining_Group=_Lam}', "");
Expect(0, 2215, '\P{^Joining_Group=_Lam}', "");
Error('\p{Jg=-:=LAM}');
Error('\P{Jg=-:=LAM}');
Expect(1, 2214, '\p{Jg=lam}', "");
Expect(0, 2214, '\p{^Jg=lam}', "");
Expect(0, 2214, '\P{Jg=lam}', "");
Expect(1, 2214, '\P{^Jg=lam}', "");
Expect(0, 2215, '\p{Jg=lam}', "");
Expect(1, 2215, '\p{^Jg=lam}', "");
Expect(1, 2215, '\P{Jg=lam}', "");
Expect(0, 2215, '\P{^Jg=lam}', "");
Expect(1, 2214, '\p{Jg=	Lam}', "");
Expect(0, 2214, '\p{^Jg=	Lam}', "");
Expect(0, 2214, '\P{Jg=	Lam}', "");
Expect(1, 2214, '\P{^Jg=	Lam}', "");
Expect(0, 2215, '\p{Jg=	Lam}', "");
Expect(1, 2215, '\p{^Jg=	Lam}', "");
Expect(1, 2215, '\P{Jg=	Lam}', "");
Expect(0, 2215, '\P{^Jg=	Lam}', "");
Error('\p{Is_Joining_Group=/a/Lam}');
Error('\P{Is_Joining_Group=/a/Lam}');
Expect(1, 2214, '\p{Is_Joining_Group=lam}', "");
Expect(0, 2214, '\p{^Is_Joining_Group=lam}', "");
Expect(0, 2214, '\P{Is_Joining_Group=lam}', "");
Expect(1, 2214, '\P{^Is_Joining_Group=lam}', "");
Expect(0, 2215, '\p{Is_Joining_Group=lam}', "");
Expect(1, 2215, '\p{^Is_Joining_Group=lam}', "");
Expect(1, 2215, '\P{Is_Joining_Group=lam}', "");
Expect(0, 2215, '\P{^Is_Joining_Group=lam}', "");
Expect(1, 2214, '\p{Is_Joining_Group=_ lam}', "");
Expect(0, 2214, '\p{^Is_Joining_Group=_ lam}', "");
Expect(0, 2214, '\P{Is_Joining_Group=_ lam}', "");
Expect(1, 2214, '\P{^Is_Joining_Group=_ lam}', "");
Expect(0, 2215, '\p{Is_Joining_Group=_ lam}', "");
Expect(1, 2215, '\p{^Is_Joining_Group=_ lam}', "");
Expect(1, 2215, '\P{Is_Joining_Group=_ lam}', "");
Expect(0, 2215, '\P{^Is_Joining_Group=_ lam}', "");
Error('\p{Is_Jg=	:=Lam}');
Error('\P{Is_Jg=	:=Lam}');
Expect(1, 2214, '\p{Is_Jg=lam}', "");
Expect(0, 2214, '\p{^Is_Jg=lam}', "");
Expect(0, 2214, '\P{Is_Jg=lam}', "");
Expect(1, 2214, '\P{^Is_Jg=lam}', "");
Expect(0, 2215, '\p{Is_Jg=lam}', "");
Expect(1, 2215, '\p{^Is_Jg=lam}', "");
Expect(1, 2215, '\P{Is_Jg=lam}', "");
Expect(0, 2215, '\P{^Is_Jg=lam}', "");
Expect(1, 2214, '\p{Is_Jg:   --Lam}', "");
Expect(0, 2214, '\p{^Is_Jg:   --Lam}', "");
Expect(0, 2214, '\P{Is_Jg:   --Lam}', "");
Expect(1, 2214, '\P{^Is_Jg:   --Lam}', "");
Expect(0, 2215, '\p{Is_Jg:   --Lam}', "");
Expect(1, 2215, '\p{^Is_Jg:   --Lam}', "");
Expect(1, 2215, '\P{Is_Jg:   --Lam}', "");
Expect(0, 2215, '\P{^Is_Jg:   --Lam}', "");
Error('\p{Joining_Group=:=- ain}');
Error('\P{Joining_Group=:=- ain}');
Expect(1, 1887, '\p{Joining_Group:   ain}', "");
Expect(0, 1887, '\p{^Joining_Group:   ain}', "");
Expect(0, 1887, '\P{Joining_Group:   ain}', "");
Expect(1, 1887, '\P{^Joining_Group:   ain}', "");
Expect(0, 1888, '\p{Joining_Group:   ain}', "");
Expect(1, 1888, '\p{^Joining_Group:   ain}', "");
Expect(1, 1888, '\P{Joining_Group:   ain}', "");
Expect(0, 1888, '\P{^Joining_Group:   ain}', "");
Expect(1, 1887, '\p{Joining_Group:		-ain}', "");
Expect(0, 1887, '\p{^Joining_Group:		-ain}', "");
Expect(0, 1887, '\P{Joining_Group:		-ain}', "");
Expect(1, 1887, '\P{^Joining_Group:		-ain}', "");
Expect(0, 1888, '\p{Joining_Group:		-ain}', "");
Expect(1, 1888, '\p{^Joining_Group:		-ain}', "");
Expect(1, 1888, '\P{Joining_Group:		-ain}', "");
Expect(0, 1888, '\P{^Joining_Group:		-ain}', "");
Error('\p{Jg=_:=Ain}');
Error('\P{Jg=_:=Ain}');
Expect(1, 1887, '\p{Jg=ain}', "");
Expect(0, 1887, '\p{^Jg=ain}', "");
Expect(0, 1887, '\P{Jg=ain}', "");
Expect(1, 1887, '\P{^Jg=ain}', "");
Expect(0, 1888, '\p{Jg=ain}', "");
Expect(1, 1888, '\p{^Jg=ain}', "");
Expect(1, 1888, '\P{Jg=ain}', "");
Expect(0, 1888, '\P{^Jg=ain}', "");
Expect(1, 1887, '\p{Jg=_-Ain}', "");
Expect(0, 1887, '\p{^Jg=_-Ain}', "");
Expect(0, 1887, '\P{Jg=_-Ain}', "");
Expect(1, 1887, '\P{^Jg=_-Ain}', "");
Expect(0, 1888, '\p{Jg=_-Ain}', "");
Expect(1, 1888, '\p{^Jg=_-Ain}', "");
Expect(1, 1888, '\P{Jg=_-Ain}', "");
Expect(0, 1888, '\P{^Jg=_-Ain}', "");
Error('\p{Is_Joining_Group= _ain:=}');
Error('\P{Is_Joining_Group= _ain:=}');
Expect(1, 1887, '\p{Is_Joining_Group=ain}', "");
Expect(0, 1887, '\p{^Is_Joining_Group=ain}', "");
Expect(0, 1887, '\P{Is_Joining_Group=ain}', "");
Expect(1, 1887, '\P{^Is_Joining_Group=ain}', "");
Expect(0, 1888, '\p{Is_Joining_Group=ain}', "");
Expect(1, 1888, '\p{^Is_Joining_Group=ain}', "");
Expect(1, 1888, '\P{Is_Joining_Group=ain}', "");
Expect(0, 1888, '\P{^Is_Joining_Group=ain}', "");
Expect(1, 1887, '\p{Is_Joining_Group=	-AIN}', "");
Expect(0, 1887, '\p{^Is_Joining_Group=	-AIN}', "");
Expect(0, 1887, '\P{Is_Joining_Group=	-AIN}', "");
Expect(1, 1887, '\P{^Is_Joining_Group=	-AIN}', "");
Expect(0, 1888, '\p{Is_Joining_Group=	-AIN}', "");
Expect(1, 1888, '\p{^Is_Joining_Group=	-AIN}', "");
Expect(1, 1888, '\P{Is_Joining_Group=	-AIN}', "");
Expect(0, 1888, '\P{^Is_Joining_Group=	-AIN}', "");
Error('\p{Is_Jg=/a/-ain}');
Error('\P{Is_Jg=/a/-ain}');
Expect(1, 1887, '\p{Is_Jg:ain}', "");
Expect(0, 1887, '\p{^Is_Jg:ain}', "");
Expect(0, 1887, '\P{Is_Jg:ain}', "");
Expect(1, 1887, '\P{^Is_Jg:ain}', "");
Expect(0, 1888, '\p{Is_Jg:ain}', "");
Expect(1, 1888, '\p{^Is_Jg:ain}', "");
Expect(1, 1888, '\P{Is_Jg:ain}', "");
Expect(0, 1888, '\P{^Is_Jg:ain}', "");
Expect(1, 1887, '\p{Is_Jg=	 AIN}', "");
Expect(0, 1887, '\p{^Is_Jg=	 AIN}', "");
Expect(0, 1887, '\P{Is_Jg=	 AIN}', "");
Expect(1, 1887, '\P{^Is_Jg=	 AIN}', "");
Expect(0, 1888, '\p{Is_Jg=	 AIN}', "");
Expect(1, 1888, '\p{^Is_Jg=	 AIN}', "");
Expect(1, 1888, '\P{Is_Jg=	 AIN}', "");
Expect(0, 1888, '\P{^Is_Jg=	 AIN}', "");
Error('\p{Joining_Group=_/a/Heh}');
Error('\P{Joining_Group=_/a/Heh}');
Expect(1, 1607, '\p{Joining_Group: heh}', "");
Expect(0, 1607, '\p{^Joining_Group: heh}', "");
Expect(0, 1607, '\P{Joining_Group: heh}', "");
Expect(1, 1607, '\P{^Joining_Group: heh}', "");
Expect(0, 1608, '\p{Joining_Group: heh}', "");
Expect(1, 1608, '\p{^Joining_Group: heh}', "");
Expect(1, 1608, '\P{Joining_Group: heh}', "");
Expect(0, 1608, '\P{^Joining_Group: heh}', "");
Expect(1, 1607, '\p{Joining_Group=_heh}', "");
Expect(0, 1607, '\p{^Joining_Group=_heh}', "");
Expect(0, 1607, '\P{Joining_Group=_heh}', "");
Expect(1, 1607, '\P{^Joining_Group=_heh}', "");
Expect(0, 1608, '\p{Joining_Group=_heh}', "");
Expect(1, 1608, '\p{^Joining_Group=_heh}', "");
Expect(1, 1608, '\P{Joining_Group=_heh}', "");
Expect(0, 1608, '\P{^Joining_Group=_heh}', "");
Error('\p{Jg:    _HEH:=}');
Error('\P{Jg:    _HEH:=}');
Expect(1, 1607, '\p{Jg=heh}', "");
Expect(0, 1607, '\p{^Jg=heh}', "");
Expect(0, 1607, '\P{Jg=heh}', "");
Expect(1, 1607, '\P{^Jg=heh}', "");
Expect(0, 1608, '\p{Jg=heh}', "");
Expect(1, 1608, '\p{^Jg=heh}', "");
Expect(1, 1608, '\P{Jg=heh}', "");
Expect(0, 1608, '\P{^Jg=heh}', "");
Expect(1, 1607, '\p{Jg=		Heh}', "");
Expect(0, 1607, '\p{^Jg=		Heh}', "");
Expect(0, 1607, '\P{Jg=		Heh}', "");
Expect(1, 1607, '\P{^Jg=		Heh}', "");
Expect(0, 1608, '\p{Jg=		Heh}', "");
Expect(1, 1608, '\p{^Jg=		Heh}', "");
Expect(1, 1608, '\P{Jg=		Heh}', "");
Expect(0, 1608, '\P{^Jg=		Heh}', "");
Error('\p{Is_Joining_Group::=-_Heh}');
Error('\P{Is_Joining_Group::=-_Heh}');
Expect(1, 1607, '\p{Is_Joining_Group=heh}', "");
Expect(0, 1607, '\p{^Is_Joining_Group=heh}', "");
Expect(0, 1607, '\P{Is_Joining_Group=heh}', "");
Expect(1, 1607, '\P{^Is_Joining_Group=heh}', "");
Expect(0, 1608, '\p{Is_Joining_Group=heh}', "");
Expect(1, 1608, '\p{^Is_Joining_Group=heh}', "");
Expect(1, 1608, '\P{Is_Joining_Group=heh}', "");
Expect(0, 1608, '\P{^Is_Joining_Group=heh}', "");
Expect(1, 1607, '\p{Is_Joining_Group= _Heh}', "");
Expect(0, 1607, '\p{^Is_Joining_Group= _Heh}', "");
Expect(0, 1607, '\P{Is_Joining_Group= _Heh}', "");
Expect(1, 1607, '\P{^Is_Joining_Group= _Heh}', "");
Expect(0, 1608, '\p{Is_Joining_Group= _Heh}', "");
Expect(1, 1608, '\p{^Is_Joining_Group= _Heh}', "");
Expect(1, 1608, '\P{Is_Joining_Group= _Heh}', "");
Expect(0, 1608, '\P{^Is_Joining_Group= _Heh}', "");
Error('\p{Is_Jg=_Heh/a/}');
Error('\P{Is_Jg=_Heh/a/}');
Expect(1, 1607, '\p{Is_Jg=heh}', "");
Expect(0, 1607, '\p{^Is_Jg=heh}', "");
Expect(0, 1607, '\P{Is_Jg=heh}', "");
Expect(1, 1607, '\P{^Is_Jg=heh}', "");
Expect(0, 1608, '\p{Is_Jg=heh}', "");
Expect(1, 1608, '\p{^Is_Jg=heh}', "");
Expect(1, 1608, '\P{Is_Jg=heh}', "");
Expect(0, 1608, '\P{^Is_Jg=heh}', "");
Expect(1, 1607, '\p{Is_Jg:_ heh}', "");
Expect(0, 1607, '\p{^Is_Jg:_ heh}', "");
Expect(0, 1607, '\P{Is_Jg:_ heh}', "");
Expect(1, 1607, '\P{^Is_Jg:_ heh}', "");
Expect(0, 1608, '\p{Is_Jg:_ heh}', "");
Expect(1, 1608, '\p{^Is_Jg:_ heh}', "");
Expect(1, 1608, '\P{Is_Jg:_ heh}', "");
Expect(0, 1608, '\P{^Is_Jg:_ heh}', "");
Error('\p{Joining_Group:   _/a/Qaph}');
Error('\P{Joining_Group:   _/a/Qaph}');
Expect(1, 1833, '\p{Joining_Group=qaph}', "");
Expect(0, 1833, '\p{^Joining_Group=qaph}', "");
Expect(0, 1833, '\P{Joining_Group=qaph}', "");
Expect(1, 1833, '\P{^Joining_Group=qaph}', "");
Expect(0, 1834, '\p{Joining_Group=qaph}', "");
Expect(1, 1834, '\p{^Joining_Group=qaph}', "");
Expect(1, 1834, '\P{Joining_Group=qaph}', "");
Expect(0, 1834, '\P{^Joining_Group=qaph}', "");
Expect(1, 1833, '\p{Joining_Group=	_qaph}', "");
Expect(0, 1833, '\p{^Joining_Group=	_qaph}', "");
Expect(0, 1833, '\P{Joining_Group=	_qaph}', "");
Expect(1, 1833, '\P{^Joining_Group=	_qaph}', "");
Expect(0, 1834, '\p{Joining_Group=	_qaph}', "");
Expect(1, 1834, '\p{^Joining_Group=	_qaph}', "");
Expect(1, 1834, '\P{Joining_Group=	_qaph}', "");
Expect(0, 1834, '\P{^Joining_Group=	_qaph}', "");
Error('\p{Jg=:=-QAPH}');
Error('\P{Jg=:=-QAPH}');
Expect(1, 1833, '\p{Jg=qaph}', "");
Expect(0, 1833, '\p{^Jg=qaph}', "");
Expect(0, 1833, '\P{Jg=qaph}', "");
Expect(1, 1833, '\P{^Jg=qaph}', "");
Expect(0, 1834, '\p{Jg=qaph}', "");
Expect(1, 1834, '\p{^Jg=qaph}', "");
Expect(1, 1834, '\P{Jg=qaph}', "");
Expect(0, 1834, '\P{^Jg=qaph}', "");
Expect(1, 1833, '\p{Jg=_Qaph}', "");
Expect(0, 1833, '\p{^Jg=_Qaph}', "");
Expect(0, 1833, '\P{Jg=_Qaph}', "");
Expect(1, 1833, '\P{^Jg=_Qaph}', "");
Expect(0, 1834, '\p{Jg=_Qaph}', "");
Expect(1, 1834, '\p{^Jg=_Qaph}', "");
Expect(1, 1834, '\P{Jg=_Qaph}', "");
Expect(0, 1834, '\P{^Jg=_Qaph}', "");
Error('\p{Is_Joining_Group=__Qaph:=}');
Error('\P{Is_Joining_Group=__Qaph:=}');
Expect(1, 1833, '\p{Is_Joining_Group=qaph}', "");
Expect(0, 1833, '\p{^Is_Joining_Group=qaph}', "");
Expect(0, 1833, '\P{Is_Joining_Group=qaph}', "");
Expect(1, 1833, '\P{^Is_Joining_Group=qaph}', "");
Expect(0, 1834, '\p{Is_Joining_Group=qaph}', "");
Expect(1, 1834, '\p{^Is_Joining_Group=qaph}', "");
Expect(1, 1834, '\P{Is_Joining_Group=qaph}', "");
Expect(0, 1834, '\P{^Is_Joining_Group=qaph}', "");
Expect(1, 1833, '\p{Is_Joining_Group= QAPH}', "");
Expect(0, 1833, '\p{^Is_Joining_Group= QAPH}', "");
Expect(0, 1833, '\P{Is_Joining_Group= QAPH}', "");
Expect(1, 1833, '\P{^Is_Joining_Group= QAPH}', "");
Expect(0, 1834, '\p{Is_Joining_Group= QAPH}', "");
Expect(1, 1834, '\p{^Is_Joining_Group= QAPH}', "");
Expect(1, 1834, '\P{Is_Joining_Group= QAPH}', "");
Expect(0, 1834, '\P{^Is_Joining_Group= QAPH}', "");
Error('\p{Is_Jg=	-Qaph:=}');
Error('\P{Is_Jg=	-Qaph:=}');
Expect(1, 1833, '\p{Is_Jg=qaph}', "");
Expect(0, 1833, '\p{^Is_Jg=qaph}', "");
Expect(0, 1833, '\P{Is_Jg=qaph}', "");
Expect(1, 1833, '\P{^Is_Jg=qaph}', "");
Expect(0, 1834, '\p{Is_Jg=qaph}', "");
Expect(1, 1834, '\p{^Is_Jg=qaph}', "");
Expect(1, 1834, '\P{Is_Jg=qaph}', "");
Expect(0, 1834, '\P{^Is_Jg=qaph}', "");
Expect(1, 1833, '\p{Is_Jg=QAPH}', "");
Expect(0, 1833, '\p{^Is_Jg=QAPH}', "");
Expect(0, 1833, '\P{Is_Jg=QAPH}', "");
Expect(1, 1833, '\P{^Is_Jg=QAPH}', "");
Expect(0, 1834, '\p{Is_Jg=QAPH}', "");
Expect(1, 1834, '\p{^Is_Jg=QAPH}', "");
Expect(1, 1834, '\P{Is_Jg=QAPH}', "");
Expect(0, 1834, '\P{^Is_Jg=QAPH}', "");
Error('\p{Joining_Group=:=_-HETH}');
Error('\P{Joining_Group=:=_-HETH}');
Expect(1, 1818, '\p{Joining_Group:	heth}', "");
Expect(0, 1818, '\p{^Joining_Group:	heth}', "");
Expect(0, 1818, '\P{Joining_Group:	heth}', "");
Expect(1, 1818, '\P{^Joining_Group:	heth}', "");
Expect(0, 1819, '\p{Joining_Group:	heth}', "");
Expect(1, 1819, '\p{^Joining_Group:	heth}', "");
Expect(1, 1819, '\P{Joining_Group:	heth}', "");
Expect(0, 1819, '\P{^Joining_Group:	heth}', "");
Error('\p{Jg=/a/	Heth}');
Error('\P{Jg=/a/	Heth}');
Expect(1, 1818, '\p{Jg=heth}', "");
Expect(0, 1818, '\p{^Jg=heth}', "");
Expect(0, 1818, '\P{Jg=heth}', "");
Expect(1, 1818, '\P{^Jg=heth}', "");
Expect(0, 1819, '\p{Jg=heth}', "");
Expect(1, 1819, '\p{^Jg=heth}', "");
Expect(1, 1819, '\P{Jg=heth}', "");
Expect(0, 1819, '\P{^Jg=heth}', "");
Expect(1, 1818, '\p{Jg=-HETH}', "");
Expect(0, 1818, '\p{^Jg=-HETH}', "");
Expect(0, 1818, '\P{Jg=-HETH}', "");
Expect(1, 1818, '\P{^Jg=-HETH}', "");
Expect(0, 1819, '\p{Jg=-HETH}', "");
Expect(1, 1819, '\p{^Jg=-HETH}', "");
Expect(1, 1819, '\P{Jg=-HETH}', "");
Expect(0, 1819, '\P{^Jg=-HETH}', "");
Error('\p{Is_Joining_Group=:=Heth}');
Error('\P{Is_Joining_Group=:=Heth}');
Expect(1, 1818, '\p{Is_Joining_Group:	heth}', "");
Expect(0, 1818, '\p{^Is_Joining_Group:	heth}', "");
Expect(0, 1818, '\P{Is_Joining_Group:	heth}', "");
Expect(1, 1818, '\P{^Is_Joining_Group:	heth}', "");
Expect(0, 1819, '\p{Is_Joining_Group:	heth}', "");
Expect(1, 1819, '\p{^Is_Joining_Group:	heth}', "");
Expect(1, 1819, '\P{Is_Joining_Group:	heth}', "");
Expect(0, 1819, '\P{^Is_Joining_Group:	heth}', "");
Expect(1, 1818, '\p{Is_Joining_Group=Heth}', "");
Expect(0, 1818, '\p{^Is_Joining_Group=Heth}', "");
Expect(0, 1818, '\P{Is_Joining_Group=Heth}', "");
Expect(1, 1818, '\P{^Is_Joining_Group=Heth}', "");
Expect(0, 1819, '\p{Is_Joining_Group=Heth}', "");
Expect(1, 1819, '\p{^Is_Joining_Group=Heth}', "");
Expect(1, 1819, '\P{Is_Joining_Group=Heth}', "");
Expect(0, 1819, '\P{^Is_Joining_Group=Heth}', "");
Error('\p{Is_Jg=_:=Heth}');
Error('\P{Is_Jg=_:=Heth}');
Expect(1, 1818, '\p{Is_Jg=heth}', "");
Expect(0, 1818, '\p{^Is_Jg=heth}', "");
Expect(0, 1818, '\P{Is_Jg=heth}', "");
Expect(1, 1818, '\P{^Is_Jg=heth}', "");
Expect(0, 1819, '\p{Is_Jg=heth}', "");
Expect(1, 1819, '\p{^Is_Jg=heth}', "");
Expect(1, 1819, '\P{Is_Jg=heth}', "");
Expect(0, 1819, '\P{^Is_Jg=heth}', "");
Expect(1, 1818, '\p{Is_Jg=-HETH}', "");
Expect(0, 1818, '\p{^Is_Jg=-HETH}', "");
Expect(0, 1818, '\P{Is_Jg=-HETH}', "");
Expect(1, 1818, '\P{^Is_Jg=-HETH}', "");
Expect(0, 1819, '\p{Is_Jg=-HETH}', "");
Expect(1, 1819, '\p{^Is_Jg=-HETH}', "");
Expect(1, 1819, '\P{Is_Jg=-HETH}', "");
Expect(0, 1819, '\P{^Is_Jg=-HETH}', "");
Error('\p{Joining_Group=/a/Teth}');
Error('\P{Joining_Group=/a/Teth}');
Expect(1, 1820, '\p{Joining_Group=teth}', "");
Expect(0, 1820, '\p{^Joining_Group=teth}', "");
Expect(0, 1820, '\P{Joining_Group=teth}', "");
Expect(1, 1820, '\P{^Joining_Group=teth}', "");
Expect(0, 1821, '\p{Joining_Group=teth}', "");
Expect(1, 1821, '\p{^Joining_Group=teth}', "");
Expect(1, 1821, '\P{Joining_Group=teth}', "");
Expect(0, 1821, '\P{^Joining_Group=teth}', "");
Expect(1, 1820, '\p{Joining_Group: __Teth}', "");
Expect(0, 1820, '\p{^Joining_Group: __Teth}', "");
Expect(0, 1820, '\P{Joining_Group: __Teth}', "");
Expect(1, 1820, '\P{^Joining_Group: __Teth}', "");
Expect(0, 1821, '\p{Joining_Group: __Teth}', "");
Expect(1, 1821, '\p{^Joining_Group: __Teth}', "");
Expect(1, 1821, '\P{Joining_Group: __Teth}', "");
Expect(0, 1821, '\P{^Joining_Group: __Teth}', "");
Error('\p{Jg=	/a/teth}');
Error('\P{Jg=	/a/teth}');
Expect(1, 1820, '\p{Jg=teth}', "");
Expect(0, 1820, '\p{^Jg=teth}', "");
Expect(0, 1820, '\P{Jg=teth}', "");
Expect(1, 1820, '\P{^Jg=teth}', "");
Expect(0, 1821, '\p{Jg=teth}', "");
Expect(1, 1821, '\p{^Jg=teth}', "");
Expect(1, 1821, '\P{Jg=teth}', "");
Expect(0, 1821, '\P{^Jg=teth}', "");
Expect(1, 1820, '\p{Jg=__teth}', "");
Expect(0, 1820, '\p{^Jg=__teth}', "");
Expect(0, 1820, '\P{Jg=__teth}', "");
Expect(1, 1820, '\P{^Jg=__teth}', "");
Expect(0, 1821, '\p{Jg=__teth}', "");
Expect(1, 1821, '\p{^Jg=__teth}', "");
Expect(1, 1821, '\P{Jg=__teth}', "");
Expect(0, 1821, '\P{^Jg=__teth}', "");
Error('\p{Is_Joining_Group=/a/	teth}');
Error('\P{Is_Joining_Group=/a/	teth}');
Expect(1, 1820, '\p{Is_Joining_Group=teth}', "");
Expect(0, 1820, '\p{^Is_Joining_Group=teth}', "");
Expect(0, 1820, '\P{Is_Joining_Group=teth}', "");
Expect(1, 1820, '\P{^Is_Joining_Group=teth}', "");
Expect(0, 1821, '\p{Is_Joining_Group=teth}', "");
Expect(1, 1821, '\p{^Is_Joining_Group=teth}', "");
Expect(1, 1821, '\P{Is_Joining_Group=teth}', "");
Expect(0, 1821, '\P{^Is_Joining_Group=teth}', "");
Expect(1, 1820, '\p{Is_Joining_Group=	_Teth}', "");
Expect(0, 1820, '\p{^Is_Joining_Group=	_Teth}', "");
Expect(0, 1820, '\P{Is_Joining_Group=	_Teth}', "");
Expect(1, 1820, '\P{^Is_Joining_Group=	_Teth}', "");
Expect(0, 1821, '\p{Is_Joining_Group=	_Teth}', "");
Expect(1, 1821, '\p{^Is_Joining_Group=	_Teth}', "");
Expect(1, 1821, '\P{Is_Joining_Group=	_Teth}', "");
Expect(0, 1821, '\P{^Is_Joining_Group=	_Teth}', "");
Error('\p{Is_Jg=/a/-TETH}');
Error('\P{Is_Jg=/a/-TETH}');
Expect(1, 1820, '\p{Is_Jg=teth}', "");
Expect(0, 1820, '\p{^Is_Jg=teth}', "");
Expect(0, 1820, '\P{Is_Jg=teth}', "");
Expect(1, 1820, '\P{^Is_Jg=teth}', "");
Expect(0, 1821, '\p{Is_Jg=teth}', "");
Expect(1, 1821, '\p{^Is_Jg=teth}', "");
Expect(1, 1821, '\P{Is_Jg=teth}', "");
Expect(0, 1821, '\P{^Is_Jg=teth}', "");
Expect(1, 1820, '\p{Is_Jg=_Teth}', "");
Expect(0, 1820, '\p{^Is_Jg=_Teth}', "");
Expect(0, 1820, '\P{Is_Jg=_Teth}', "");
Expect(1, 1820, '\P{^Is_Jg=_Teth}', "");
Expect(0, 1821, '\p{Is_Jg=_Teth}', "");
Expect(1, 1821, '\p{^Is_Jg=_Teth}', "");
Expect(1, 1821, '\P{Is_Jg=_Teth}', "");
Expect(0, 1821, '\P{^Is_Jg=_Teth}', "");
Error('\p{Joining_Group=_:=dalath_Rish}');
Error('\P{Joining_Group=_:=dalath_Rish}');
Expect(1, 1839, '\p{Joining_Group=dalathrish}', "");
Expect(0, 1839, '\p{^Joining_Group=dalathrish}', "");
Expect(0, 1839, '\P{Joining_Group=dalathrish}', "");
Expect(1, 1839, '\P{^Joining_Group=dalathrish}', "");
Expect(0, 1840, '\p{Joining_Group=dalathrish}', "");
Expect(1, 1840, '\p{^Joining_Group=dalathrish}', "");
Expect(1, 1840, '\P{Joining_Group=dalathrish}', "");
Expect(0, 1840, '\P{^Joining_Group=dalathrish}', "");
Expect(1, 1839, '\p{Joining_Group=_dalath_RISH}', "");
Expect(0, 1839, '\p{^Joining_Group=_dalath_RISH}', "");
Expect(0, 1839, '\P{Joining_Group=_dalath_RISH}', "");
Expect(1, 1839, '\P{^Joining_Group=_dalath_RISH}', "");
Expect(0, 1840, '\p{Joining_Group=_dalath_RISH}', "");
Expect(1, 1840, '\p{^Joining_Group=_dalath_RISH}', "");
Expect(1, 1840, '\P{Joining_Group=_dalath_RISH}', "");
Expect(0, 1840, '\P{^Joining_Group=_dalath_RISH}', "");
Error('\p{Jg=	/a/dalath_Rish}');
Error('\P{Jg=	/a/dalath_Rish}');
Expect(1, 1839, '\p{Jg: dalathrish}', "");
Expect(0, 1839, '\p{^Jg: dalathrish}', "");
Expect(0, 1839, '\P{Jg: dalathrish}', "");
Expect(1, 1839, '\P{^Jg: dalathrish}', "");
Expect(0, 1840, '\p{Jg: dalathrish}', "");
Expect(1, 1840, '\p{^Jg: dalathrish}', "");
Expect(1, 1840, '\P{Jg: dalathrish}', "");
Expect(0, 1840, '\P{^Jg: dalathrish}', "");
Expect(1, 1839, '\p{Jg=	Dalath_RISH}', "");
Expect(0, 1839, '\p{^Jg=	Dalath_RISH}', "");
Expect(0, 1839, '\P{Jg=	Dalath_RISH}', "");
Expect(1, 1839, '\P{^Jg=	Dalath_RISH}', "");
Expect(0, 1840, '\p{Jg=	Dalath_RISH}', "");
Expect(1, 1840, '\p{^Jg=	Dalath_RISH}', "");
Expect(1, 1840, '\P{Jg=	Dalath_RISH}', "");
Expect(0, 1840, '\P{^Jg=	Dalath_RISH}', "");
Error('\p{Is_Joining_Group=-:=dalath_Rish}');
Error('\P{Is_Joining_Group=-:=dalath_Rish}');
Expect(1, 1839, '\p{Is_Joining_Group=dalathrish}', "");
Expect(0, 1839, '\p{^Is_Joining_Group=dalathrish}', "");
Expect(0, 1839, '\P{Is_Joining_Group=dalathrish}', "");
Expect(1, 1839, '\P{^Is_Joining_Group=dalathrish}', "");
Expect(0, 1840, '\p{Is_Joining_Group=dalathrish}', "");
Expect(1, 1840, '\p{^Is_Joining_Group=dalathrish}', "");
Expect(1, 1840, '\P{Is_Joining_Group=dalathrish}', "");
Expect(0, 1840, '\P{^Is_Joining_Group=dalathrish}', "");
Expect(1, 1839, '\p{Is_Joining_Group=-dalath_Rish}', "");
Expect(0, 1839, '\p{^Is_Joining_Group=-dalath_Rish}', "");
Expect(0, 1839, '\P{Is_Joining_Group=-dalath_Rish}', "");
Expect(1, 1839, '\P{^Is_Joining_Group=-dalath_Rish}', "");
Expect(0, 1840, '\p{Is_Joining_Group=-dalath_Rish}', "");
Expect(1, 1840, '\p{^Is_Joining_Group=-dalath_Rish}', "");
Expect(1, 1840, '\P{Is_Joining_Group=-dalath_Rish}', "");
Expect(0, 1840, '\P{^Is_Joining_Group=-dalath_Rish}', "");
Error('\p{Is_Jg:	:=Dalath_rish}');
Error('\P{Is_Jg:	:=Dalath_rish}');
Expect(1, 1839, '\p{Is_Jg:	dalathrish}', "");
Expect(0, 1839, '\p{^Is_Jg:	dalathrish}', "");
Expect(0, 1839, '\P{Is_Jg:	dalathrish}', "");
Expect(1, 1839, '\P{^Is_Jg:	dalathrish}', "");
Expect(0, 1840, '\p{Is_Jg:	dalathrish}', "");
Expect(1, 1840, '\p{^Is_Jg:	dalathrish}', "");
Expect(1, 1840, '\P{Is_Jg:	dalathrish}', "");
Expect(0, 1840, '\P{^Is_Jg:	dalathrish}', "");
Expect(1, 1839, '\p{Is_Jg=__DALATH_rish}', "");
Expect(0, 1839, '\p{^Is_Jg=__DALATH_rish}', "");
Expect(0, 1839, '\P{Is_Jg=__DALATH_rish}', "");
Expect(1, 1839, '\P{^Is_Jg=__DALATH_rish}', "");
Expect(0, 1840, '\p{Is_Jg=__DALATH_rish}', "");
Expect(1, 1840, '\p{^Is_Jg=__DALATH_rish}', "");
Expect(1, 1840, '\P{Is_Jg=__DALATH_rish}', "");
Expect(0, 1840, '\P{^Is_Jg=__DALATH_rish}', "");
Error('\p{Joining_Group=_:=Yudh_He}');
Error('\P{Joining_Group=_:=Yudh_He}');
Expect(1, 1822, '\p{Joining_Group:	yudhhe}', "");
Expect(0, 1822, '\p{^Joining_Group:	yudhhe}', "");
Expect(0, 1822, '\P{Joining_Group:	yudhhe}', "");
Expect(1, 1822, '\P{^Joining_Group:	yudhhe}', "");
Expect(0, 1823, '\p{Joining_Group:	yudhhe}', "");
Expect(1, 1823, '\p{^Joining_Group:	yudhhe}', "");
Expect(1, 1823, '\P{Joining_Group:	yudhhe}', "");
Expect(0, 1823, '\P{^Joining_Group:	yudhhe}', "");
Expect(1, 1822, '\p{Joining_Group=_Yudh_He}', "");
Expect(0, 1822, '\p{^Joining_Group=_Yudh_He}', "");
Expect(0, 1822, '\P{Joining_Group=_Yudh_He}', "");
Expect(1, 1822, '\P{^Joining_Group=_Yudh_He}', "");
Expect(0, 1823, '\p{Joining_Group=_Yudh_He}', "");
Expect(1, 1823, '\p{^Joining_Group=_Yudh_He}', "");
Expect(1, 1823, '\P{Joining_Group=_Yudh_He}', "");
Expect(0, 1823, '\P{^Joining_Group=_Yudh_He}', "");
Error('\p{Jg=/a/Yudh_he}');
Error('\P{Jg=/a/Yudh_he}');
Expect(1, 1822, '\p{Jg=yudhhe}', "");
Expect(0, 1822, '\p{^Jg=yudhhe}', "");
Expect(0, 1822, '\P{Jg=yudhhe}', "");
Expect(1, 1822, '\P{^Jg=yudhhe}', "");
Expect(0, 1823, '\p{Jg=yudhhe}', "");
Expect(1, 1823, '\p{^Jg=yudhhe}', "");
Expect(1, 1823, '\P{Jg=yudhhe}', "");
Expect(0, 1823, '\P{^Jg=yudhhe}', "");
Expect(1, 1822, '\p{Jg=  Yudh_He}', "");
Expect(0, 1822, '\p{^Jg=  Yudh_He}', "");
Expect(0, 1822, '\P{Jg=  Yudh_He}', "");
Expect(1, 1822, '\P{^Jg=  Yudh_He}', "");
Expect(0, 1823, '\p{Jg=  Yudh_He}', "");
Expect(1, 1823, '\p{^Jg=  Yudh_He}', "");
Expect(1, 1823, '\P{Jg=  Yudh_He}', "");
Expect(0, 1823, '\P{^Jg=  Yudh_He}', "");
Error('\p{Is_Joining_Group= -YUDH_he/a/}');
Error('\P{Is_Joining_Group= -YUDH_he/a/}');
Expect(1, 1822, '\p{Is_Joining_Group: yudhhe}', "");
Expect(0, 1822, '\p{^Is_Joining_Group: yudhhe}', "");
Expect(0, 1822, '\P{Is_Joining_Group: yudhhe}', "");
Expect(1, 1822, '\P{^Is_Joining_Group: yudhhe}', "");
Expect(0, 1823, '\p{Is_Joining_Group: yudhhe}', "");
Expect(1, 1823, '\p{^Is_Joining_Group: yudhhe}', "");
Expect(1, 1823, '\P{Is_Joining_Group: yudhhe}', "");
Expect(0, 1823, '\P{^Is_Joining_Group: yudhhe}', "");
Expect(1, 1822, '\p{Is_Joining_Group=  yudh_He}', "");
Expect(0, 1822, '\p{^Is_Joining_Group=  yudh_He}', "");
Expect(0, 1822, '\P{Is_Joining_Group=  yudh_He}', "");
Expect(1, 1822, '\P{^Is_Joining_Group=  yudh_He}', "");
Expect(0, 1823, '\p{Is_Joining_Group=  yudh_He}', "");
Expect(1, 1823, '\p{^Is_Joining_Group=  yudh_He}', "");
Expect(1, 1823, '\P{Is_Joining_Group=  yudh_He}', "");
Expect(0, 1823, '\P{^Is_Joining_Group=  yudh_He}', "");
Error('\p{Is_Jg=:=--Yudh_HE}');
Error('\P{Is_Jg=:=--Yudh_HE}');
Expect(1, 1822, '\p{Is_Jg=yudhhe}', "");
Expect(0, 1822, '\p{^Is_Jg=yudhhe}', "");
Expect(0, 1822, '\P{Is_Jg=yudhhe}', "");
Expect(1, 1822, '\P{^Is_Jg=yudhhe}', "");
Expect(0, 1823, '\p{Is_Jg=yudhhe}', "");
Expect(1, 1823, '\p{^Is_Jg=yudhhe}', "");
Expect(1, 1823, '\P{Is_Jg=yudhhe}', "");
Expect(0, 1823, '\P{^Is_Jg=yudhhe}', "");
Expect(1, 1822, '\p{Is_Jg=- YUDH_He}', "");
Expect(0, 1822, '\p{^Is_Jg=- YUDH_He}', "");
Expect(0, 1822, '\P{Is_Jg=- YUDH_He}', "");
Expect(1, 1822, '\P{^Is_Jg=- YUDH_He}', "");
Expect(0, 1823, '\p{Is_Jg=- YUDH_He}', "");
Expect(1, 1823, '\p{^Is_Jg=- YUDH_He}', "");
Expect(1, 1823, '\P{Is_Jg=- YUDH_He}', "");
Expect(0, 1823, '\P{^Is_Jg=- YUDH_He}', "");
Error('\p{Joining_Group=/a/Noon}');
Error('\P{Joining_Group=/a/Noon}');
Expect(1, 1897, '\p{Joining_Group=noon}', "");
Expect(0, 1897, '\p{^Joining_Group=noon}', "");
Expect(0, 1897, '\P{Joining_Group=noon}', "");
Expect(1, 1897, '\P{^Joining_Group=noon}', "");
Expect(0, 1898, '\p{Joining_Group=noon}', "");
Expect(1, 1898, '\p{^Joining_Group=noon}', "");
Expect(1, 1898, '\P{Joining_Group=noon}', "");
Expect(0, 1898, '\P{^Joining_Group=noon}', "");
Expect(1, 1897, '\p{Joining_Group=--NOON}', "");
Expect(0, 1897, '\p{^Joining_Group=--NOON}', "");
Expect(0, 1897, '\P{Joining_Group=--NOON}', "");
Expect(1, 1897, '\P{^Joining_Group=--NOON}', "");
Expect(0, 1898, '\p{Joining_Group=--NOON}', "");
Expect(1, 1898, '\p{^Joining_Group=--NOON}', "");
Expect(1, 1898, '\P{Joining_Group=--NOON}', "");
Expect(0, 1898, '\P{^Joining_Group=--NOON}', "");
Error('\p{Jg=:=-Noon}');
Error('\P{Jg=:=-Noon}');
Expect(1, 1897, '\p{Jg:   noon}', "");
Expect(0, 1897, '\p{^Jg:   noon}', "");
Expect(0, 1897, '\P{Jg:   noon}', "");
Expect(1, 1897, '\P{^Jg:   noon}', "");
Expect(0, 1898, '\p{Jg:   noon}', "");
Expect(1, 1898, '\p{^Jg:   noon}', "");
Expect(1, 1898, '\P{Jg:   noon}', "");
Expect(0, 1898, '\P{^Jg:   noon}', "");
Expect(1, 1897, '\p{Jg=-_Noon}', "");
Expect(0, 1897, '\p{^Jg=-_Noon}', "");
Expect(0, 1897, '\P{Jg=-_Noon}', "");
Expect(1, 1897, '\P{^Jg=-_Noon}', "");
Expect(0, 1898, '\p{Jg=-_Noon}', "");
Expect(1, 1898, '\p{^Jg=-_Noon}', "");
Expect(1, 1898, '\P{Jg=-_Noon}', "");
Expect(0, 1898, '\P{^Jg=-_Noon}', "");
Error('\p{Is_Joining_Group:   	NOON/a/}');
Error('\P{Is_Joining_Group:   	NOON/a/}');
Expect(1, 1897, '\p{Is_Joining_Group=noon}', "");
Expect(0, 1897, '\p{^Is_Joining_Group=noon}', "");
Expect(0, 1897, '\P{Is_Joining_Group=noon}', "");
Expect(1, 1897, '\P{^Is_Joining_Group=noon}', "");
Expect(0, 1898, '\p{Is_Joining_Group=noon}', "");
Expect(1, 1898, '\p{^Is_Joining_Group=noon}', "");
Expect(1, 1898, '\P{Is_Joining_Group=noon}', "");
Expect(0, 1898, '\P{^Is_Joining_Group=noon}', "");
Expect(1, 1897, '\p{Is_Joining_Group=	-noon}', "");
Expect(0, 1897, '\p{^Is_Joining_Group=	-noon}', "");
Expect(0, 1897, '\P{Is_Joining_Group=	-noon}', "");
Expect(1, 1897, '\P{^Is_Joining_Group=	-noon}', "");
Expect(0, 1898, '\p{Is_Joining_Group=	-noon}', "");
Expect(1, 1898, '\p{^Is_Joining_Group=	-noon}', "");
Expect(1, 1898, '\P{Is_Joining_Group=	-noon}', "");
Expect(0, 1898, '\P{^Is_Joining_Group=	-noon}', "");
Error('\p{Is_Jg=	/a/Noon}');
Error('\P{Is_Jg=	/a/Noon}');
Expect(1, 1897, '\p{Is_Jg=noon}', "");
Expect(0, 1897, '\p{^Is_Jg=noon}', "");
Expect(0, 1897, '\P{Is_Jg=noon}', "");
Expect(1, 1897, '\P{^Is_Jg=noon}', "");
Expect(0, 1898, '\p{Is_Jg=noon}', "");
Expect(1, 1898, '\p{^Is_Jg=noon}', "");
Expect(1, 1898, '\P{Is_Jg=noon}', "");
Expect(0, 1898, '\P{^Is_Jg=noon}', "");
Expect(1, 1897, '\p{Is_Jg=Noon}', "");
Expect(0, 1897, '\p{^Is_Jg=Noon}', "");
Expect(0, 1897, '\P{Is_Jg=Noon}', "");
Expect(1, 1897, '\P{^Is_Jg=Noon}', "");
Expect(0, 1898, '\p{Is_Jg=Noon}', "");
Expect(1, 1898, '\p{^Is_Jg=Noon}', "");
Expect(1, 1898, '\P{Is_Jg=Noon}', "");
Expect(0, 1898, '\P{^Is_Jg=Noon}', "");
Error('\p{Joining_Group=	:=Mim}');
Error('\P{Joining_Group=	:=Mim}');
Expect(1, 1825, '\p{Joining_Group=mim}', "");
Expect(0, 1825, '\p{^Joining_Group=mim}', "");
Expect(0, 1825, '\P{Joining_Group=mim}', "");
Expect(1, 1825, '\P{^Joining_Group=mim}', "");
Expect(0, 1826, '\p{Joining_Group=mim}', "");
Expect(1, 1826, '\p{^Joining_Group=mim}', "");
Expect(1, 1826, '\P{Joining_Group=mim}', "");
Expect(0, 1826, '\P{^Joining_Group=mim}', "");
Expect(1, 1825, '\p{Joining_Group= mim}', "");
Expect(0, 1825, '\p{^Joining_Group= mim}', "");
Expect(0, 1825, '\P{Joining_Group= mim}', "");
Expect(1, 1825, '\P{^Joining_Group= mim}', "");
Expect(0, 1826, '\p{Joining_Group= mim}', "");
Expect(1, 1826, '\p{^Joining_Group= mim}', "");
Expect(1, 1826, '\P{Joining_Group= mim}', "");
Expect(0, 1826, '\P{^Joining_Group= mim}', "");
Error('\p{Jg=:=MIM}');
Error('\P{Jg=:=MIM}');
Expect(1, 1825, '\p{Jg=mim}', "");
Expect(0, 1825, '\p{^Jg=mim}', "");
Expect(0, 1825, '\P{Jg=mim}', "");
Expect(1, 1825, '\P{^Jg=mim}', "");
Expect(0, 1826, '\p{Jg=mim}', "");
Expect(1, 1826, '\p{^Jg=mim}', "");
Expect(1, 1826, '\P{Jg=mim}', "");
Expect(0, 1826, '\P{^Jg=mim}', "");
Expect(1, 1825, '\p{Jg=_-MIM}', "");
Expect(0, 1825, '\p{^Jg=_-MIM}', "");
Expect(0, 1825, '\P{Jg=_-MIM}', "");
Expect(1, 1825, '\P{^Jg=_-MIM}', "");
Expect(0, 1826, '\p{Jg=_-MIM}', "");
Expect(1, 1826, '\p{^Jg=_-MIM}', "");
Expect(1, 1826, '\P{Jg=_-MIM}', "");
Expect(0, 1826, '\P{^Jg=_-MIM}', "");
Error('\p{Is_Joining_Group=- Mim:=}');
Error('\P{Is_Joining_Group=- Mim:=}');
Expect(1, 1825, '\p{Is_Joining_Group=mim}', "");
Expect(0, 1825, '\p{^Is_Joining_Group=mim}', "");
Expect(0, 1825, '\P{Is_Joining_Group=mim}', "");
Expect(1, 1825, '\P{^Is_Joining_Group=mim}', "");
Expect(0, 1826, '\p{Is_Joining_Group=mim}', "");
Expect(1, 1826, '\p{^Is_Joining_Group=mim}', "");
Expect(1, 1826, '\P{Is_Joining_Group=mim}', "");
Expect(0, 1826, '\P{^Is_Joining_Group=mim}', "");
Expect(1, 1825, '\p{Is_Joining_Group=_ MIM}', "");
Expect(0, 1825, '\p{^Is_Joining_Group=_ MIM}', "");
Expect(0, 1825, '\P{Is_Joining_Group=_ MIM}', "");
Expect(1, 1825, '\P{^Is_Joining_Group=_ MIM}', "");
Expect(0, 1826, '\p{Is_Joining_Group=_ MIM}', "");
Expect(1, 1826, '\p{^Is_Joining_Group=_ MIM}', "");
Expect(1, 1826, '\P{Is_Joining_Group=_ MIM}', "");
Expect(0, 1826, '\P{^Is_Joining_Group=_ MIM}', "");
Error('\p{Is_Jg=__Mim:=}');
Error('\P{Is_Jg=__Mim:=}');
Expect(1, 1825, '\p{Is_Jg=mim}', "");
Expect(0, 1825, '\p{^Is_Jg=mim}', "");
Expect(0, 1825, '\P{Is_Jg=mim}', "");
Expect(1, 1825, '\P{^Is_Jg=mim}', "");
Expect(0, 1826, '\p{Is_Jg=mim}', "");
Expect(1, 1826, '\p{^Is_Jg=mim}', "");
Expect(1, 1826, '\P{Is_Jg=mim}', "");
Expect(0, 1826, '\P{^Is_Jg=mim}', "");
Expect(1, 1825, '\p{Is_Jg=-_MIM}', "");
Expect(0, 1825, '\p{^Is_Jg=-_MIM}', "");
Expect(0, 1825, '\P{Is_Jg=-_MIM}', "");
Expect(1, 1825, '\P{^Is_Jg=-_MIM}', "");
Expect(0, 1826, '\p{Is_Jg=-_MIM}', "");
Expect(1, 1826, '\p{^Is_Jg=-_MIM}', "");
Expect(1, 1826, '\P{Is_Jg=-_MIM}', "");
Expect(0, 1826, '\P{^Is_Jg=-_MIM}', "");
Error('\p{Joining_Group=_ HEH_Goal/a/}');
Error('\P{Joining_Group=_ HEH_Goal/a/}');
Expect(1, 1730, '\p{Joining_Group=hehgoal}', "");
Expect(0, 1730, '\p{^Joining_Group=hehgoal}', "");
Expect(0, 1730, '\P{Joining_Group=hehgoal}', "");
Expect(1, 1730, '\P{^Joining_Group=hehgoal}', "");
Expect(0, 1731, '\p{Joining_Group=hehgoal}', "");
Expect(1, 1731, '\p{^Joining_Group=hehgoal}', "");
Expect(1, 1731, '\P{Joining_Group=hehgoal}', "");
Expect(0, 1731, '\P{^Joining_Group=hehgoal}', "");
Expect(1, 1730, '\p{Joining_Group=_	Heh_goal}', "");
Expect(0, 1730, '\p{^Joining_Group=_	Heh_goal}', "");
Expect(0, 1730, '\P{Joining_Group=_	Heh_goal}', "");
Expect(1, 1730, '\P{^Joining_Group=_	Heh_goal}', "");
Expect(0, 1731, '\p{Joining_Group=_	Heh_goal}', "");
Expect(1, 1731, '\p{^Joining_Group=_	Heh_goal}', "");
Expect(1, 1731, '\P{Joining_Group=_	Heh_goal}', "");
Expect(0, 1731, '\P{^Joining_Group=_	Heh_goal}', "");
Error('\p{Jg:    _Heh_Goal:=}');
Error('\P{Jg:    _Heh_Goal:=}');
Expect(1, 1730, '\p{Jg=hehgoal}', "");
Expect(0, 1730, '\p{^Jg=hehgoal}', "");
Expect(0, 1730, '\P{Jg=hehgoal}', "");
Expect(1, 1730, '\P{^Jg=hehgoal}', "");
Expect(0, 1731, '\p{Jg=hehgoal}', "");
Expect(1, 1731, '\p{^Jg=hehgoal}', "");
Expect(1, 1731, '\P{Jg=hehgoal}', "");
Expect(0, 1731, '\P{^Jg=hehgoal}', "");
Expect(1, 1730, '\p{Jg=_-Heh_GOAL}', "");
Expect(0, 1730, '\p{^Jg=_-Heh_GOAL}', "");
Expect(0, 1730, '\P{Jg=_-Heh_GOAL}', "");
Expect(1, 1730, '\P{^Jg=_-Heh_GOAL}', "");
Expect(0, 1731, '\p{Jg=_-Heh_GOAL}', "");
Expect(1, 1731, '\p{^Jg=_-Heh_GOAL}', "");
Expect(1, 1731, '\P{Jg=_-Heh_GOAL}', "");
Expect(0, 1731, '\P{^Jg=_-Heh_GOAL}', "");
Error('\p{Is_Joining_Group=/a/_	Heh_Goal}');
Error('\P{Is_Joining_Group=/a/_	Heh_Goal}');
Expect(1, 1730, '\p{Is_Joining_Group=hehgoal}', "");
Expect(0, 1730, '\p{^Is_Joining_Group=hehgoal}', "");
Expect(0, 1730, '\P{Is_Joining_Group=hehgoal}', "");
Expect(1, 1730, '\P{^Is_Joining_Group=hehgoal}', "");
Expect(0, 1731, '\p{Is_Joining_Group=hehgoal}', "");
Expect(1, 1731, '\p{^Is_Joining_Group=hehgoal}', "");
Expect(1, 1731, '\P{Is_Joining_Group=hehgoal}', "");
Expect(0, 1731, '\P{^Is_Joining_Group=hehgoal}', "");
Expect(1, 1730, '\p{Is_Joining_Group=--heh_GOAL}', "");
Expect(0, 1730, '\p{^Is_Joining_Group=--heh_GOAL}', "");
Expect(0, 1730, '\P{Is_Joining_Group=--heh_GOAL}', "");
Expect(1, 1730, '\P{^Is_Joining_Group=--heh_GOAL}', "");
Expect(0, 1731, '\p{Is_Joining_Group=--heh_GOAL}', "");
Expect(1, 1731, '\p{^Is_Joining_Group=--heh_GOAL}', "");
Expect(1, 1731, '\P{Is_Joining_Group=--heh_GOAL}', "");
Expect(0, 1731, '\P{^Is_Joining_Group=--heh_GOAL}', "");
Error('\p{Is_Jg:/a/HEH_goal}');
Error('\P{Is_Jg:/a/HEH_goal}');
Expect(1, 1730, '\p{Is_Jg=hehgoal}', "");
Expect(0, 1730, '\p{^Is_Jg=hehgoal}', "");
Expect(0, 1730, '\P{Is_Jg=hehgoal}', "");
Expect(1, 1730, '\P{^Is_Jg=hehgoal}', "");
Expect(0, 1731, '\p{Is_Jg=hehgoal}', "");
Expect(1, 1731, '\p{^Is_Jg=hehgoal}', "");
Expect(1, 1731, '\P{Is_Jg=hehgoal}', "");
Expect(0, 1731, '\P{^Is_Jg=hehgoal}', "");
Expect(1, 1730, '\p{Is_Jg=_HEH_Goal}', "");
Expect(0, 1730, '\p{^Is_Jg=_HEH_Goal}', "");
Expect(0, 1730, '\P{Is_Jg=_HEH_Goal}', "");
Expect(1, 1730, '\P{^Is_Jg=_HEH_Goal}', "");
Expect(0, 1731, '\p{Is_Jg=_HEH_Goal}', "");
Expect(1, 1731, '\p{^Is_Jg=_HEH_Goal}', "");
Expect(1, 1731, '\P{Is_Jg=_HEH_Goal}', "");
Expect(0, 1731, '\P{^Is_Jg=_HEH_Goal}', "");
Error('\p{Joining_Group= 	khaph/a/}');
Error('\P{Joining_Group= 	khaph/a/}');
Expect(1, 1870, '\p{Joining_Group=khaph}', "");
Expect(0, 1870, '\p{^Joining_Group=khaph}', "");
Expect(0, 1870, '\P{Joining_Group=khaph}', "");
Expect(1, 1870, '\P{^Joining_Group=khaph}', "");
Expect(0, 1871, '\p{Joining_Group=khaph}', "");
Expect(1, 1871, '\p{^Joining_Group=khaph}', "");
Expect(1, 1871, '\P{Joining_Group=khaph}', "");
Expect(0, 1871, '\P{^Joining_Group=khaph}', "");
Expect(1, 1870, '\p{Joining_Group: -KHAPH}', "");
Expect(0, 1870, '\p{^Joining_Group: -KHAPH}', "");
Expect(0, 1870, '\P{Joining_Group: -KHAPH}', "");
Expect(1, 1870, '\P{^Joining_Group: -KHAPH}', "");
Expect(0, 1871, '\p{Joining_Group: -KHAPH}', "");
Expect(1, 1871, '\p{^Joining_Group: -KHAPH}', "");
Expect(1, 1871, '\P{Joining_Group: -KHAPH}', "");
Expect(0, 1871, '\P{^Joining_Group: -KHAPH}', "");
Error('\p{Jg=:=		Khaph}');
Error('\P{Jg=:=		Khaph}');
Expect(1, 1870, '\p{Jg=khaph}', "");
Expect(0, 1870, '\p{^Jg=khaph}', "");
Expect(0, 1870, '\P{Jg=khaph}', "");
Expect(1, 1870, '\P{^Jg=khaph}', "");
Expect(0, 1871, '\p{Jg=khaph}', "");
Expect(1, 1871, '\p{^Jg=khaph}', "");
Expect(1, 1871, '\P{Jg=khaph}', "");
Expect(0, 1871, '\P{^Jg=khaph}', "");
Expect(1, 1870, '\p{Jg=KHAPH}', "");
Expect(0, 1870, '\p{^Jg=KHAPH}', "");
Expect(0, 1870, '\P{Jg=KHAPH}', "");
Expect(1, 1870, '\P{^Jg=KHAPH}', "");
Expect(0, 1871, '\p{Jg=KHAPH}', "");
Expect(1, 1871, '\p{^Jg=KHAPH}', "");
Expect(1, 1871, '\P{Jg=KHAPH}', "");
Expect(0, 1871, '\P{^Jg=KHAPH}', "");
Error('\p{Is_Joining_Group= :=Khaph}');
Error('\P{Is_Joining_Group= :=Khaph}');
Expect(1, 1870, '\p{Is_Joining_Group=khaph}', "");
Expect(0, 1870, '\p{^Is_Joining_Group=khaph}', "");
Expect(0, 1870, '\P{Is_Joining_Group=khaph}', "");
Expect(1, 1870, '\P{^Is_Joining_Group=khaph}', "");
Expect(0, 1871, '\p{Is_Joining_Group=khaph}', "");
Expect(1, 1871, '\p{^Is_Joining_Group=khaph}', "");
Expect(1, 1871, '\P{Is_Joining_Group=khaph}', "");
Expect(0, 1871, '\P{^Is_Joining_Group=khaph}', "");
Expect(1, 1870, '\p{Is_Joining_Group=_-Khaph}', "");
Expect(0, 1870, '\p{^Is_Joining_Group=_-Khaph}', "");
Expect(0, 1870, '\P{Is_Joining_Group=_-Khaph}', "");
Expect(1, 1870, '\P{^Is_Joining_Group=_-Khaph}', "");
Expect(0, 1871, '\p{Is_Joining_Group=_-Khaph}', "");
Expect(1, 1871, '\p{^Is_Joining_Group=_-Khaph}', "");
Expect(1, 1871, '\P{Is_Joining_Group=_-Khaph}', "");
Expect(0, 1871, '\P{^Is_Joining_Group=_-Khaph}', "");
Error('\p{Is_Jg=:=	KHAPH}');
Error('\P{Is_Jg=:=	KHAPH}');
Expect(1, 1870, '\p{Is_Jg=khaph}', "");
Expect(0, 1870, '\p{^Is_Jg=khaph}', "");
Expect(0, 1870, '\P{Is_Jg=khaph}', "");
Expect(1, 1870, '\P{^Is_Jg=khaph}', "");
Expect(0, 1871, '\p{Is_Jg=khaph}', "");
Expect(1, 1871, '\p{^Is_Jg=khaph}', "");
Expect(1, 1871, '\P{Is_Jg=khaph}', "");
Expect(0, 1871, '\P{^Is_Jg=khaph}', "");
Expect(1, 1870, '\p{Is_Jg=	_KHAPH}', "");
Expect(0, 1870, '\p{^Is_Jg=	_KHAPH}', "");
Expect(0, 1870, '\P{Is_Jg=	_KHAPH}', "");
Expect(1, 1870, '\P{^Is_Jg=	_KHAPH}', "");
Expect(0, 1871, '\p{Is_Jg=	_KHAPH}', "");
Expect(1, 1871, '\p{^Is_Jg=	_KHAPH}', "");
Expect(1, 1871, '\P{Is_Jg=	_KHAPH}', "");
Expect(0, 1871, '\P{^Is_Jg=	_KHAPH}', "");
Error('\p{Joining_Group=/a/_SAD}');
Error('\P{Joining_Group=/a/_SAD}');
Expect(1, 1787, '\p{Joining_Group=sad}', "");
Expect(0, 1787, '\p{^Joining_Group=sad}', "");
Expect(0, 1787, '\P{Joining_Group=sad}', "");
Expect(1, 1787, '\P{^Joining_Group=sad}', "");
Expect(0, 1788, '\p{Joining_Group=sad}', "");
Expect(1, 1788, '\p{^Joining_Group=sad}', "");
Expect(1, 1788, '\P{Joining_Group=sad}', "");
Expect(0, 1788, '\P{^Joining_Group=sad}', "");
Expect(1, 1787, '\p{Joining_Group=_Sad}', "");
Expect(0, 1787, '\p{^Joining_Group=_Sad}', "");
Expect(0, 1787, '\P{Joining_Group=_Sad}', "");
Expect(1, 1787, '\P{^Joining_Group=_Sad}', "");
Expect(0, 1788, '\p{Joining_Group=_Sad}', "");
Expect(1, 1788, '\p{^Joining_Group=_Sad}', "");
Expect(1, 1788, '\P{Joining_Group=_Sad}', "");
Expect(0, 1788, '\P{^Joining_Group=_Sad}', "");
Error('\p{Jg=_sad/a/}');
Error('\P{Jg=_sad/a/}');
Expect(1, 1787, '\p{Jg=sad}', "");
Expect(0, 1787, '\p{^Jg=sad}', "");
Expect(0, 1787, '\P{Jg=sad}', "");
Expect(1, 1787, '\P{^Jg=sad}', "");
Expect(0, 1788, '\p{Jg=sad}', "");
Expect(1, 1788, '\p{^Jg=sad}', "");
Expect(1, 1788, '\P{Jg=sad}', "");
Expect(0, 1788, '\P{^Jg=sad}', "");
Expect(1, 1787, '\p{Jg=_ Sad}', "");
Expect(0, 1787, '\p{^Jg=_ Sad}', "");
Expect(0, 1787, '\P{Jg=_ Sad}', "");
Expect(1, 1787, '\P{^Jg=_ Sad}', "");
Expect(0, 1788, '\p{Jg=_ Sad}', "");
Expect(1, 1788, '\p{^Jg=_ Sad}', "");
Expect(1, 1788, '\P{Jg=_ Sad}', "");
Expect(0, 1788, '\P{^Jg=_ Sad}', "");
Error('\p{Is_Joining_Group=  SAD/a/}');
Error('\P{Is_Joining_Group=  SAD/a/}');
Expect(1, 1787, '\p{Is_Joining_Group=sad}', "");
Expect(0, 1787, '\p{^Is_Joining_Group=sad}', "");
Expect(0, 1787, '\P{Is_Joining_Group=sad}', "");
Expect(1, 1787, '\P{^Is_Joining_Group=sad}', "");
Expect(0, 1788, '\p{Is_Joining_Group=sad}', "");
Expect(1, 1788, '\p{^Is_Joining_Group=sad}', "");
Expect(1, 1788, '\P{Is_Joining_Group=sad}', "");
Expect(0, 1788, '\P{^Is_Joining_Group=sad}', "");
Expect(1, 1787, '\p{Is_Joining_Group: - Sad}', "");
Expect(0, 1787, '\p{^Is_Joining_Group: - Sad}', "");
Expect(0, 1787, '\P{Is_Joining_Group: - Sad}', "");
Expect(1, 1787, '\P{^Is_Joining_Group: - Sad}', "");
Expect(0, 1788, '\p{Is_Joining_Group: - Sad}', "");
Expect(1, 1788, '\p{^Is_Joining_Group: - Sad}', "");
Expect(1, 1788, '\P{Is_Joining_Group: - Sad}', "");
Expect(0, 1788, '\P{^Is_Joining_Group: - Sad}', "");
Error('\p{Is_Jg=:=SAD}');
Error('\P{Is_Jg=:=SAD}');
Expect(1, 1787, '\p{Is_Jg=sad}', "");
Expect(0, 1787, '\p{^Is_Jg=sad}', "");
Expect(0, 1787, '\P{Is_Jg=sad}', "");
Expect(1, 1787, '\P{^Is_Jg=sad}', "");
Expect(0, 1788, '\p{Is_Jg=sad}', "");
Expect(1, 1788, '\p{^Is_Jg=sad}', "");
Expect(1, 1788, '\P{Is_Jg=sad}', "");
Expect(0, 1788, '\P{^Is_Jg=sad}', "");
Expect(1, 1787, '\p{Is_Jg=_ Sad}', "");
Expect(0, 1787, '\p{^Is_Jg=_ Sad}', "");
Expect(0, 1787, '\P{Is_Jg=_ Sad}', "");
Expect(1, 1787, '\P{^Is_Jg=_ Sad}', "");
Expect(0, 1788, '\p{Is_Jg=_ Sad}', "");
Expect(1, 1788, '\p{^Is_Jg=_ Sad}', "");
Expect(1, 1788, '\P{Is_Jg=_ Sad}', "");
Expect(0, 1788, '\P{^Is_Jg=_ Sad}', "");
Error('\p{Joining_Group: /a/ -QAF}');
Error('\P{Joining_Group: /a/ -QAF}');
Expect(1, 2213, '\p{Joining_Group:qaf}', "");
Expect(0, 2213, '\p{^Joining_Group:qaf}', "");
Expect(0, 2213, '\P{Joining_Group:qaf}', "");
Expect(1, 2213, '\P{^Joining_Group:qaf}', "");
Expect(0, 2214, '\p{Joining_Group:qaf}', "");
Expect(1, 2214, '\p{^Joining_Group:qaf}', "");
Expect(1, 2214, '\P{Joining_Group:qaf}', "");
Expect(0, 2214, '\P{^Joining_Group:qaf}', "");
Expect(1, 2213, '\p{Joining_Group=-Qaf}', "");
Expect(0, 2213, '\p{^Joining_Group=-Qaf}', "");
Expect(0, 2213, '\P{Joining_Group=-Qaf}', "");
Expect(1, 2213, '\P{^Joining_Group=-Qaf}', "");
Expect(0, 2214, '\p{Joining_Group=-Qaf}', "");
Expect(1, 2214, '\p{^Joining_Group=-Qaf}', "");
Expect(1, 2214, '\P{Joining_Group=-Qaf}', "");
Expect(0, 2214, '\P{^Joining_Group=-Qaf}', "");
Error('\p{Jg=-QAF:=}');
Error('\P{Jg=-QAF:=}');
Expect(1, 2213, '\p{Jg=qaf}', "");
Expect(0, 2213, '\p{^Jg=qaf}', "");
Expect(0, 2213, '\P{Jg=qaf}', "");
Expect(1, 2213, '\P{^Jg=qaf}', "");
Expect(0, 2214, '\p{Jg=qaf}', "");
Expect(1, 2214, '\p{^Jg=qaf}', "");
Expect(1, 2214, '\P{Jg=qaf}', "");
Expect(0, 2214, '\P{^Jg=qaf}', "");
Expect(1, 2213, '\p{Jg=	QAF}', "");
Expect(0, 2213, '\p{^Jg=	QAF}', "");
Expect(0, 2213, '\P{Jg=	QAF}', "");
Expect(1, 2213, '\P{^Jg=	QAF}', "");
Expect(0, 2214, '\p{Jg=	QAF}', "");
Expect(1, 2214, '\p{^Jg=	QAF}', "");
Expect(1, 2214, '\P{Jg=	QAF}', "");
Expect(0, 2214, '\P{^Jg=	QAF}', "");
Error('\p{Is_Joining_Group=__QAF:=}');
Error('\P{Is_Joining_Group=__QAF:=}');
Expect(1, 2213, '\p{Is_Joining_Group:   qaf}', "");
Expect(0, 2213, '\p{^Is_Joining_Group:   qaf}', "");
Expect(0, 2213, '\P{Is_Joining_Group:   qaf}', "");
Expect(1, 2213, '\P{^Is_Joining_Group:   qaf}', "");
Expect(0, 2214, '\p{Is_Joining_Group:   qaf}', "");
Expect(1, 2214, '\p{^Is_Joining_Group:   qaf}', "");
Expect(1, 2214, '\P{Is_Joining_Group:   qaf}', "");
Expect(0, 2214, '\P{^Is_Joining_Group:   qaf}', "");
Expect(1, 2213, '\p{Is_Joining_Group:	__Qaf}', "");
Expect(0, 2213, '\p{^Is_Joining_Group:	__Qaf}', "");
Expect(0, 2213, '\P{Is_Joining_Group:	__Qaf}', "");
Expect(1, 2213, '\P{^Is_Joining_Group:	__Qaf}', "");
Expect(0, 2214, '\p{Is_Joining_Group:	__Qaf}', "");
Expect(1, 2214, '\p{^Is_Joining_Group:	__Qaf}', "");
Expect(1, 2214, '\P{Is_Joining_Group:	__Qaf}', "");
Expect(0, 2214, '\P{^Is_Joining_Group:	__Qaf}', "");
Error('\p{Is_Jg=/a/qaf}');
Error('\P{Is_Jg=/a/qaf}');
Expect(1, 2213, '\p{Is_Jg=qaf}', "");
Expect(0, 2213, '\p{^Is_Jg=qaf}', "");
Expect(0, 2213, '\P{Is_Jg=qaf}', "");
Expect(1, 2213, '\P{^Is_Jg=qaf}', "");
Expect(0, 2214, '\p{Is_Jg=qaf}', "");
Expect(1, 2214, '\p{^Is_Jg=qaf}', "");
Expect(1, 2214, '\P{Is_Jg=qaf}', "");
Expect(0, 2214, '\P{^Is_Jg=qaf}', "");
Expect(1, 2213, '\p{Is_Jg=  Qaf}', "");
Expect(0, 2213, '\p{^Is_Jg=  Qaf}', "");
Expect(0, 2213, '\P{Is_Jg=  Qaf}', "");
Expect(1, 2213, '\P{^Is_Jg=  Qaf}', "");
Expect(0, 2214, '\p{Is_Jg=  Qaf}', "");
Expect(1, 2214, '\p{^Is_Jg=  Qaf}', "");
Expect(1, 2214, '\P{Is_Jg=  Qaf}', "");
Expect(0, 2214, '\P{^Is_Jg=  Qaf}', "");
Error('\p{Joining_Group= _kaph:=}');
Error('\P{Joining_Group= _kaph:=}');
Expect(1, 1823, '\p{Joining_Group=kaph}', "");
Expect(0, 1823, '\p{^Joining_Group=kaph}', "");
Expect(0, 1823, '\P{Joining_Group=kaph}', "");
Expect(1, 1823, '\P{^Joining_Group=kaph}', "");
Expect(0, 1824, '\p{Joining_Group=kaph}', "");
Expect(1, 1824, '\p{^Joining_Group=kaph}', "");
Expect(1, 1824, '\P{Joining_Group=kaph}', "");
Expect(0, 1824, '\P{^Joining_Group=kaph}', "");
Expect(1, 1823, '\p{Joining_Group=  Kaph}', "");
Expect(0, 1823, '\p{^Joining_Group=  Kaph}', "");
Expect(0, 1823, '\P{Joining_Group=  Kaph}', "");
Expect(1, 1823, '\P{^Joining_Group=  Kaph}', "");
Expect(0, 1824, '\p{Joining_Group=  Kaph}', "");
Expect(1, 1824, '\p{^Joining_Group=  Kaph}', "");
Expect(1, 1824, '\P{Joining_Group=  Kaph}', "");
Expect(0, 1824, '\P{^Joining_Group=  Kaph}', "");
Error('\p{Jg=KAPH/a/}');
Error('\P{Jg=KAPH/a/}');
Expect(1, 1823, '\p{Jg=kaph}', "");
Expect(0, 1823, '\p{^Jg=kaph}', "");
Expect(0, 1823, '\P{Jg=kaph}', "");
Expect(1, 1823, '\P{^Jg=kaph}', "");
Expect(0, 1824, '\p{Jg=kaph}', "");
Expect(1, 1824, '\p{^Jg=kaph}', "");
Expect(1, 1824, '\P{Jg=kaph}', "");
Expect(0, 1824, '\P{^Jg=kaph}', "");
Expect(1, 1823, '\p{Jg=-KAPH}', "");
Expect(0, 1823, '\p{^Jg=-KAPH}', "");
Expect(0, 1823, '\P{Jg=-KAPH}', "");
Expect(1, 1823, '\P{^Jg=-KAPH}', "");
Expect(0, 1824, '\p{Jg=-KAPH}', "");
Expect(1, 1824, '\p{^Jg=-KAPH}', "");
Expect(1, 1824, '\P{Jg=-KAPH}', "");
Expect(0, 1824, '\P{^Jg=-KAPH}', "");
Error('\p{Is_Joining_Group= :=Kaph}');
Error('\P{Is_Joining_Group= :=Kaph}');
Expect(1, 1823, '\p{Is_Joining_Group=kaph}', "");
Expect(0, 1823, '\p{^Is_Joining_Group=kaph}', "");
Expect(0, 1823, '\P{Is_Joining_Group=kaph}', "");
Expect(1, 1823, '\P{^Is_Joining_Group=kaph}', "");
Expect(0, 1824, '\p{Is_Joining_Group=kaph}', "");
Expect(1, 1824, '\p{^Is_Joining_Group=kaph}', "");
Expect(1, 1824, '\P{Is_Joining_Group=kaph}', "");
Expect(0, 1824, '\P{^Is_Joining_Group=kaph}', "");
Expect(1, 1823, '\p{Is_Joining_Group=	-KAPH}', "");
Expect(0, 1823, '\p{^Is_Joining_Group=	-KAPH}', "");
Expect(0, 1823, '\P{Is_Joining_Group=	-KAPH}', "");
Expect(1, 1823, '\P{^Is_Joining_Group=	-KAPH}', "");
Expect(0, 1824, '\p{Is_Joining_Group=	-KAPH}', "");
Expect(1, 1824, '\p{^Is_Joining_Group=	-KAPH}', "");
Expect(1, 1824, '\P{Is_Joining_Group=	-KAPH}', "");
Expect(0, 1824, '\P{^Is_Joining_Group=	-KAPH}', "");
Error('\p{Is_Jg=	_KAPH:=}');
Error('\P{Is_Jg=	_KAPH:=}');
Expect(1, 1823, '\p{Is_Jg=kaph}', "");
Expect(0, 1823, '\p{^Is_Jg=kaph}', "");
Expect(0, 1823, '\P{Is_Jg=kaph}', "");
Expect(1, 1823, '\P{^Is_Jg=kaph}', "");
Expect(0, 1824, '\p{Is_Jg=kaph}', "");
Expect(1, 1824, '\p{^Is_Jg=kaph}', "");
Expect(1, 1824, '\P{Is_Jg=kaph}', "");
Expect(0, 1824, '\P{^Is_Jg=kaph}', "");
Expect(1, 1823, '\p{Is_Jg=	_KAPH}', "");
Expect(0, 1823, '\p{^Is_Jg=	_KAPH}', "");
Expect(0, 1823, '\P{Is_Jg=	_KAPH}', "");
Expect(1, 1823, '\P{^Is_Jg=	_KAPH}', "");
Expect(0, 1824, '\p{Is_Jg=	_KAPH}', "");
Expect(1, 1824, '\p{^Is_Jg=	_KAPH}', "");
Expect(1, 1824, '\P{Is_Jg=	_KAPH}', "");
Expect(0, 1824, '\P{^Is_Jg=	_KAPH}', "");
Error('\p{Joining_Group= -Feh:=}');
Error('\P{Joining_Group= -Feh:=}');
Expect(1, 2212, '\p{Joining_Group=feh}', "");
Expect(0, 2212, '\p{^Joining_Group=feh}', "");
Expect(0, 2212, '\P{Joining_Group=feh}', "");
Expect(1, 2212, '\P{^Joining_Group=feh}', "");
Expect(0, 2213, '\p{Joining_Group=feh}', "");
Expect(1, 2213, '\p{^Joining_Group=feh}', "");
Expect(1, 2213, '\P{Joining_Group=feh}', "");
Expect(0, 2213, '\P{^Joining_Group=feh}', "");
Expect(1, 2212, '\p{Joining_Group= FEH}', "");
Expect(0, 2212, '\p{^Joining_Group= FEH}', "");
Expect(0, 2212, '\P{Joining_Group= FEH}', "");
Expect(1, 2212, '\P{^Joining_Group= FEH}', "");
Expect(0, 2213, '\p{Joining_Group= FEH}', "");
Expect(1, 2213, '\p{^Joining_Group= FEH}', "");
Expect(1, 2213, '\P{Joining_Group= FEH}', "");
Expect(0, 2213, '\P{^Joining_Group= FEH}', "");
Error('\p{Jg=_ FEH/a/}');
Error('\P{Jg=_ FEH/a/}');
Expect(1, 2212, '\p{Jg=feh}', "");
Expect(0, 2212, '\p{^Jg=feh}', "");
Expect(0, 2212, '\P{Jg=feh}', "");
Expect(1, 2212, '\P{^Jg=feh}', "");
Expect(0, 2213, '\p{Jg=feh}', "");
Expect(1, 2213, '\p{^Jg=feh}', "");
Expect(1, 2213, '\P{Jg=feh}', "");
Expect(0, 2213, '\P{^Jg=feh}', "");
Expect(1, 2212, '\p{Jg=-_Feh}', "");
Expect(0, 2212, '\p{^Jg=-_Feh}', "");
Expect(0, 2212, '\P{Jg=-_Feh}', "");
Expect(1, 2212, '\P{^Jg=-_Feh}', "");
Expect(0, 2213, '\p{Jg=-_Feh}', "");
Expect(1, 2213, '\p{^Jg=-_Feh}', "");
Expect(1, 2213, '\P{Jg=-_Feh}', "");
Expect(0, 2213, '\P{^Jg=-_Feh}', "");
Error('\p{Is_Joining_Group=/a/	feh}');
Error('\P{Is_Joining_Group=/a/	feh}');
Expect(1, 2212, '\p{Is_Joining_Group=feh}', "");
Expect(0, 2212, '\p{^Is_Joining_Group=feh}', "");
Expect(0, 2212, '\P{Is_Joining_Group=feh}', "");
Expect(1, 2212, '\P{^Is_Joining_Group=feh}', "");
Expect(0, 2213, '\p{Is_Joining_Group=feh}', "");
Expect(1, 2213, '\p{^Is_Joining_Group=feh}', "");
Expect(1, 2213, '\P{Is_Joining_Group=feh}', "");
Expect(0, 2213, '\P{^Is_Joining_Group=feh}', "");
Expect(1, 2212, '\p{Is_Joining_Group=_ FEH}', "");
Expect(0, 2212, '\p{^Is_Joining_Group=_ FEH}', "");
Expect(0, 2212, '\P{Is_Joining_Group=_ FEH}', "");
Expect(1, 2212, '\P{^Is_Joining_Group=_ FEH}', "");
Expect(0, 2213, '\p{Is_Joining_Group=_ FEH}', "");
Expect(1, 2213, '\p{^Is_Joining_Group=_ FEH}', "");
Expect(1, 2213, '\P{Is_Joining_Group=_ FEH}', "");
Expect(0, 2213, '\P{^Is_Joining_Group=_ FEH}', "");
Error('\p{Is_Jg=	/a/FEH}');
Error('\P{Is_Jg=	/a/FEH}');
Expect(1, 2212, '\p{Is_Jg=feh}', "");
Expect(0, 2212, '\p{^Is_Jg=feh}', "");
Expect(0, 2212, '\P{Is_Jg=feh}', "");
Expect(1, 2212, '\P{^Is_Jg=feh}', "");
Expect(0, 2213, '\p{Is_Jg=feh}', "");
Expect(1, 2213, '\p{^Is_Jg=feh}', "");
Expect(1, 2213, '\P{Is_Jg=feh}', "");
Expect(0, 2213, '\P{^Is_Jg=feh}', "");
Error('\p{Joining_Group=/a/-FARSI_Yeh}');
Error('\P{Joining_Group=/a/-FARSI_Yeh}');
Expect(1, 1910, '\p{Joining_Group=farsiyeh}', "");
Expect(0, 1910, '\p{^Joining_Group=farsiyeh}', "");
Expect(0, 1910, '\P{Joining_Group=farsiyeh}', "");
Expect(1, 1910, '\P{^Joining_Group=farsiyeh}', "");
Expect(0, 1911, '\p{Joining_Group=farsiyeh}', "");
Expect(1, 1911, '\p{^Joining_Group=farsiyeh}', "");
Expect(1, 1911, '\P{Joining_Group=farsiyeh}', "");
Expect(0, 1911, '\P{^Joining_Group=farsiyeh}', "");
Expect(1, 1910, '\p{Joining_Group=--FARSI_YEH}', "");
Expect(0, 1910, '\p{^Joining_Group=--FARSI_YEH}', "");
Expect(0, 1910, '\P{Joining_Group=--FARSI_YEH}', "");
Expect(1, 1910, '\P{^Joining_Group=--FARSI_YEH}', "");
Expect(0, 1911, '\p{Joining_Group=--FARSI_YEH}', "");
Expect(1, 1911, '\p{^Joining_Group=--FARSI_YEH}', "");
Expect(1, 1911, '\P{Joining_Group=--FARSI_YEH}', "");
Expect(0, 1911, '\P{^Joining_Group=--FARSI_YEH}', "");
Error('\p{Jg:   	/a/Farsi_yeh}');
Error('\P{Jg:   	/a/Farsi_yeh}');
Expect(1, 1910, '\p{Jg=farsiyeh}', "");
Expect(0, 1910, '\p{^Jg=farsiyeh}', "");
Expect(0, 1910, '\P{Jg=farsiyeh}', "");
Expect(1, 1910, '\P{^Jg=farsiyeh}', "");
Expect(0, 1911, '\p{Jg=farsiyeh}', "");
Expect(1, 1911, '\p{^Jg=farsiyeh}', "");
Expect(1, 1911, '\P{Jg=farsiyeh}', "");
Expect(0, 1911, '\P{^Jg=farsiyeh}', "");
Expect(1, 1910, '\p{Jg=--Farsi_yeh}', "");
Expect(0, 1910, '\p{^Jg=--Farsi_yeh}', "");
Expect(0, 1910, '\P{Jg=--Farsi_yeh}', "");
Expect(1, 1910, '\P{^Jg=--Farsi_yeh}', "");
Expect(0, 1911, '\p{Jg=--Farsi_yeh}', "");
Expect(1, 1911, '\p{^Jg=--Farsi_yeh}', "");
Expect(1, 1911, '\P{Jg=--Farsi_yeh}', "");
Expect(0, 1911, '\P{^Jg=--Farsi_yeh}', "");
Error('\p{Is_Joining_Group= -Farsi_Yeh/a/}');
Error('\P{Is_Joining_Group= -Farsi_Yeh/a/}');
Expect(1, 1910, '\p{Is_Joining_Group=farsiyeh}', "");
Expect(0, 1910, '\p{^Is_Joining_Group=farsiyeh}', "");
Expect(0, 1910, '\P{Is_Joining_Group=farsiyeh}', "");
Expect(1, 1910, '\P{^Is_Joining_Group=farsiyeh}', "");
Expect(0, 1911, '\p{Is_Joining_Group=farsiyeh}', "");
Expect(1, 1911, '\p{^Is_Joining_Group=farsiyeh}', "");
Expect(1, 1911, '\P{Is_Joining_Group=farsiyeh}', "");
Expect(0, 1911, '\P{^Is_Joining_Group=farsiyeh}', "");
Expect(1, 1910, '\p{Is_Joining_Group=		farsi_Yeh}', "");
Expect(0, 1910, '\p{^Is_Joining_Group=		farsi_Yeh}', "");
Expect(0, 1910, '\P{Is_Joining_Group=		farsi_Yeh}', "");
Expect(1, 1910, '\P{^Is_Joining_Group=		farsi_Yeh}', "");
Expect(0, 1911, '\p{Is_Joining_Group=		farsi_Yeh}', "");
Expect(1, 1911, '\p{^Is_Joining_Group=		farsi_Yeh}', "");
Expect(1, 1911, '\P{Is_Joining_Group=		farsi_Yeh}', "");
Expect(0, 1911, '\P{^Is_Joining_Group=		farsi_Yeh}', "");
Error('\p{Is_Jg=_:=FARSI_YEH}');
Error('\P{Is_Jg=_:=FARSI_YEH}');
Expect(1, 1910, '\p{Is_Jg=farsiyeh}', "");
Expect(0, 1910, '\p{^Is_Jg=farsiyeh}', "");
Expect(0, 1910, '\P{Is_Jg=farsiyeh}', "");
Expect(1, 1910, '\P{^Is_Jg=farsiyeh}', "");
Expect(0, 1911, '\p{Is_Jg=farsiyeh}', "");
Expect(1, 1911, '\p{^Is_Jg=farsiyeh}', "");
Expect(1, 1911, '\P{Is_Jg=farsiyeh}', "");
Expect(0, 1911, '\P{^Is_Jg=farsiyeh}', "");
Expect(1, 1910, '\p{Is_Jg=	farsi_Yeh}', "");
Expect(0, 1910, '\p{^Is_Jg=	farsi_Yeh}', "");
Expect(0, 1910, '\P{Is_Jg=	farsi_Yeh}', "");
Expect(1, 1910, '\P{^Is_Jg=	farsi_Yeh}', "");
Expect(0, 1911, '\p{Is_Jg=	farsi_Yeh}', "");
Expect(1, 1911, '\p{^Is_Jg=	farsi_Yeh}', "");
Expect(1, 1911, '\P{Is_Jg=	farsi_Yeh}', "");
Expect(0, 1911, '\P{^Is_Jg=	farsi_Yeh}', "");
Error('\p{Joining_Group=-/a/Reversed_Pe}');
Error('\P{Joining_Group=-/a/Reversed_Pe}');
Expect(1, 1831, '\p{Joining_Group=reversedpe}', "");
Expect(0, 1831, '\p{^Joining_Group=reversedpe}', "");
Expect(0, 1831, '\P{Joining_Group=reversedpe}', "");
Expect(1, 1831, '\P{^Joining_Group=reversedpe}', "");
Expect(0, 1832, '\p{Joining_Group=reversedpe}', "");
Expect(1, 1832, '\p{^Joining_Group=reversedpe}', "");
Expect(1, 1832, '\P{Joining_Group=reversedpe}', "");
Expect(0, 1832, '\P{^Joining_Group=reversedpe}', "");
Expect(1, 1831, '\p{Joining_Group:	--reversed_pe}', "");
Expect(0, 1831, '\p{^Joining_Group:	--reversed_pe}', "");
Expect(0, 1831, '\P{Joining_Group:	--reversed_pe}', "");
Expect(1, 1831, '\P{^Joining_Group:	--reversed_pe}', "");
Expect(0, 1832, '\p{Joining_Group:	--reversed_pe}', "");
Expect(1, 1832, '\p{^Joining_Group:	--reversed_pe}', "");
Expect(1, 1832, '\P{Joining_Group:	--reversed_pe}', "");
Expect(0, 1832, '\P{^Joining_Group:	--reversed_pe}', "");
Error('\p{Jg=		reversed_Pe/a/}');
Error('\P{Jg=		reversed_Pe/a/}');
Expect(1, 1831, '\p{Jg:   reversedpe}', "");
Expect(0, 1831, '\p{^Jg:   reversedpe}', "");
Expect(0, 1831, '\P{Jg:   reversedpe}', "");
Expect(1, 1831, '\P{^Jg:   reversedpe}', "");
Expect(0, 1832, '\p{Jg:   reversedpe}', "");
Expect(1, 1832, '\p{^Jg:   reversedpe}', "");
Expect(1, 1832, '\P{Jg:   reversedpe}', "");
Expect(0, 1832, '\P{^Jg:   reversedpe}', "");
Expect(1, 1831, '\p{Jg=-Reversed_Pe}', "");
Expect(0, 1831, '\p{^Jg=-Reversed_Pe}', "");
Expect(0, 1831, '\P{Jg=-Reversed_Pe}', "");
Expect(1, 1831, '\P{^Jg=-Reversed_Pe}', "");
Expect(0, 1832, '\p{Jg=-Reversed_Pe}', "");
Expect(1, 1832, '\p{^Jg=-Reversed_Pe}', "");
Expect(1, 1832, '\P{Jg=-Reversed_Pe}', "");
Expect(0, 1832, '\P{^Jg=-Reversed_Pe}', "");
Error('\p{Is_Joining_Group=	:=Reversed_PE}');
Error('\P{Is_Joining_Group=	:=Reversed_PE}');
Expect(1, 1831, '\p{Is_Joining_Group=reversedpe}', "");
Expect(0, 1831, '\p{^Is_Joining_Group=reversedpe}', "");
Expect(0, 1831, '\P{Is_Joining_Group=reversedpe}', "");
Expect(1, 1831, '\P{^Is_Joining_Group=reversedpe}', "");
Expect(0, 1832, '\p{Is_Joining_Group=reversedpe}', "");
Expect(1, 1832, '\p{^Is_Joining_Group=reversedpe}', "");
Expect(1, 1832, '\P{Is_Joining_Group=reversedpe}', "");
Expect(0, 1832, '\P{^Is_Joining_Group=reversedpe}', "");
Expect(1, 1831, '\p{Is_Joining_Group=-	REVERSED_Pe}', "");
Expect(0, 1831, '\p{^Is_Joining_Group=-	REVERSED_Pe}', "");
Expect(0, 1831, '\P{Is_Joining_Group=-	REVERSED_Pe}', "");
Expect(1, 1831, '\P{^Is_Joining_Group=-	REVERSED_Pe}', "");
Expect(0, 1832, '\p{Is_Joining_Group=-	REVERSED_Pe}', "");
Expect(1, 1832, '\p{^Is_Joining_Group=-	REVERSED_Pe}', "");
Expect(1, 1832, '\P{Is_Joining_Group=-	REVERSED_Pe}', "");
Expect(0, 1832, '\P{^Is_Joining_Group=-	REVERSED_Pe}', "");
Error('\p{Is_Jg=-	reversed_pe/a/}');
Error('\P{Is_Jg=-	reversed_pe/a/}');
Expect(1, 1831, '\p{Is_Jg=reversedpe}', "");
Expect(0, 1831, '\p{^Is_Jg=reversedpe}', "");
Expect(0, 1831, '\P{Is_Jg=reversedpe}', "");
Expect(1, 1831, '\P{^Is_Jg=reversedpe}', "");
Expect(0, 1832, '\p{Is_Jg=reversedpe}', "");
Expect(1, 1832, '\p{^Is_Jg=reversedpe}', "");
Expect(1, 1832, '\P{Is_Jg=reversedpe}', "");
Expect(0, 1832, '\P{^Is_Jg=reversedpe}', "");
Expect(1, 1831, '\p{Is_Jg=Reversed_PE}', "");
Expect(0, 1831, '\p{^Is_Jg=Reversed_PE}', "");
Expect(0, 1831, '\P{Is_Jg=Reversed_PE}', "");
Expect(1, 1831, '\P{^Is_Jg=Reversed_PE}', "");
Expect(0, 1832, '\p{Is_Jg=Reversed_PE}', "");
Expect(1, 1832, '\p{^Is_Jg=Reversed_PE}', "");
Expect(1, 1832, '\P{Is_Jg=Reversed_PE}', "");
Expect(0, 1832, '\P{^Is_Jg=Reversed_PE}', "");
Error('\p{Joining_Group=:=Shin}');
Error('\P{Joining_Group=:=Shin}');
Expect(1, 1835, '\p{Joining_Group=shin}', "");
Expect(0, 1835, '\p{^Joining_Group=shin}', "");
Expect(0, 1835, '\P{Joining_Group=shin}', "");
Expect(1, 1835, '\P{^Joining_Group=shin}', "");
Expect(0, 1836, '\p{Joining_Group=shin}', "");
Expect(1, 1836, '\p{^Joining_Group=shin}', "");
Expect(1, 1836, '\P{Joining_Group=shin}', "");
Expect(0, 1836, '\P{^Joining_Group=shin}', "");
Expect(1, 1835, '\p{Joining_Group=-	shin}', "");
Expect(0, 1835, '\p{^Joining_Group=-	shin}', "");
Expect(0, 1835, '\P{Joining_Group=-	shin}', "");
Expect(1, 1835, '\P{^Joining_Group=-	shin}', "");
Expect(0, 1836, '\p{Joining_Group=-	shin}', "");
Expect(1, 1836, '\p{^Joining_Group=-	shin}', "");
Expect(1, 1836, '\P{Joining_Group=-	shin}', "");
Expect(0, 1836, '\P{^Joining_Group=-	shin}', "");
Error('\p{Jg=/a/	Shin}');
Error('\P{Jg=/a/	Shin}');
Expect(1, 1835, '\p{Jg=shin}', "");
Expect(0, 1835, '\p{^Jg=shin}', "");
Expect(0, 1835, '\P{Jg=shin}', "");
Expect(1, 1835, '\P{^Jg=shin}', "");
Expect(0, 1836, '\p{Jg=shin}', "");
Expect(1, 1836, '\p{^Jg=shin}', "");
Expect(1, 1836, '\P{Jg=shin}', "");
Expect(0, 1836, '\P{^Jg=shin}', "");
Expect(1, 1835, '\p{Jg= Shin}', "");
Expect(0, 1835, '\p{^Jg= Shin}', "");
Expect(0, 1835, '\P{Jg= Shin}', "");
Expect(1, 1835, '\P{^Jg= Shin}', "");
Expect(0, 1836, '\p{Jg= Shin}', "");
Expect(1, 1836, '\p{^Jg= Shin}', "");
Expect(1, 1836, '\P{Jg= Shin}', "");
Expect(0, 1836, '\P{^Jg= Shin}', "");
Error('\p{Is_Joining_Group:   :=-SHIN}');
Error('\P{Is_Joining_Group:   :=-SHIN}');
Expect(1, 1835, '\p{Is_Joining_Group=shin}', "");
Expect(0, 1835, '\p{^Is_Joining_Group=shin}', "");
Expect(0, 1835, '\P{Is_Joining_Group=shin}', "");
Expect(1, 1835, '\P{^Is_Joining_Group=shin}', "");
Expect(0, 1836, '\p{Is_Joining_Group=shin}', "");
Expect(1, 1836, '\p{^Is_Joining_Group=shin}', "");
Expect(1, 1836, '\P{Is_Joining_Group=shin}', "");
Expect(0, 1836, '\P{^Is_Joining_Group=shin}', "");
Expect(1, 1835, '\p{Is_Joining_Group= _Shin}', "");
Expect(0, 1835, '\p{^Is_Joining_Group= _Shin}', "");
Expect(0, 1835, '\P{Is_Joining_Group= _Shin}', "");
Expect(1, 1835, '\P{^Is_Joining_Group= _Shin}', "");
Expect(0, 1836, '\p{Is_Joining_Group= _Shin}', "");
Expect(1, 1836, '\p{^Is_Joining_Group= _Shin}', "");
Expect(1, 1836, '\P{Is_Joining_Group= _Shin}', "");
Expect(0, 1836, '\P{^Is_Joining_Group= _Shin}', "");
Error('\p{Is_Jg:   /a/Shin}');
Error('\P{Is_Jg:   /a/Shin}');
Expect(1, 1835, '\p{Is_Jg=shin}', "");
Expect(0, 1835, '\p{^Is_Jg=shin}', "");
Expect(0, 1835, '\P{Is_Jg=shin}', "");
Expect(1, 1835, '\P{^Is_Jg=shin}', "");
Expect(0, 1836, '\p{Is_Jg=shin}', "");
Expect(1, 1836, '\p{^Is_Jg=shin}', "");
Expect(1, 1836, '\P{Is_Jg=shin}', "");
Expect(0, 1836, '\P{^Is_Jg=shin}', "");
Expect(1, 1835, '\p{Is_Jg= Shin}', "");
Expect(0, 1835, '\p{^Is_Jg= Shin}', "");
Expect(0, 1835, '\P{Is_Jg= Shin}', "");
Expect(1, 1835, '\P{^Is_Jg= Shin}', "");
Expect(0, 1836, '\p{Is_Jg= Shin}', "");
Expect(1, 1836, '\p{^Is_Jg= Shin}', "");
Expect(1, 1836, '\P{Is_Jg= Shin}', "");
Expect(0, 1836, '\P{^Is_Jg= Shin}', "");
Error('\p{Joining_Group:/a/YEH}');
Error('\P{Joining_Group:/a/YEH}');
Expect(1, 2217, '\p{Joining_Group=yeh}', "");
Expect(0, 2217, '\p{^Joining_Group=yeh}', "");
Expect(0, 2217, '\P{Joining_Group=yeh}', "");
Expect(1, 2217, '\P{^Joining_Group=yeh}', "");
Expect(0, 2218, '\p{Joining_Group=yeh}', "");
Expect(1, 2218, '\p{^Joining_Group=yeh}', "");
Expect(1, 2218, '\P{Joining_Group=yeh}', "");
Expect(0, 2218, '\P{^Joining_Group=yeh}', "");
Expect(1, 2217, '\p{Joining_Group= YEH}', "");
Expect(0, 2217, '\p{^Joining_Group= YEH}', "");
Expect(0, 2217, '\P{Joining_Group= YEH}', "");
Expect(1, 2217, '\P{^Joining_Group= YEH}', "");
Expect(0, 2218, '\p{Joining_Group= YEH}', "");
Expect(1, 2218, '\p{^Joining_Group= YEH}', "");
Expect(1, 2218, '\P{Joining_Group= YEH}', "");
Expect(0, 2218, '\P{^Joining_Group= YEH}', "");
Error('\p{Jg=		yeh/a/}');
Error('\P{Jg=		yeh/a/}');
Expect(1, 2217, '\p{Jg=yeh}', "");
Expect(0, 2217, '\p{^Jg=yeh}', "");
Expect(0, 2217, '\P{Jg=yeh}', "");
Expect(1, 2217, '\P{^Jg=yeh}', "");
Expect(0, 2218, '\p{Jg=yeh}', "");
Expect(1, 2218, '\p{^Jg=yeh}', "");
Expect(1, 2218, '\P{Jg=yeh}', "");
Expect(0, 2218, '\P{^Jg=yeh}', "");
Expect(1, 2217, '\p{Jg=  Yeh}', "");
Expect(0, 2217, '\p{^Jg=  Yeh}', "");
Expect(0, 2217, '\P{Jg=  Yeh}', "");
Expect(1, 2217, '\P{^Jg=  Yeh}', "");
Expect(0, 2218, '\p{Jg=  Yeh}', "");
Expect(1, 2218, '\p{^Jg=  Yeh}', "");
Expect(1, 2218, '\P{Jg=  Yeh}', "");
Expect(0, 2218, '\P{^Jg=  Yeh}', "");
Error('\p{Is_Joining_Group=:=	 Yeh}');
Error('\P{Is_Joining_Group=:=	 Yeh}');
Expect(1, 2217, '\p{Is_Joining_Group=yeh}', "");
Expect(0, 2217, '\p{^Is_Joining_Group=yeh}', "");
Expect(0, 2217, '\P{Is_Joining_Group=yeh}', "");
Expect(1, 2217, '\P{^Is_Joining_Group=yeh}', "");
Expect(0, 2218, '\p{Is_Joining_Group=yeh}', "");
Expect(1, 2218, '\p{^Is_Joining_Group=yeh}', "");
Expect(1, 2218, '\P{Is_Joining_Group=yeh}', "");
Expect(0, 2218, '\P{^Is_Joining_Group=yeh}', "");
Expect(1, 2217, '\p{Is_Joining_Group:    	Yeh}', "");
Expect(0, 2217, '\p{^Is_Joining_Group:    	Yeh}', "");
Expect(0, 2217, '\P{Is_Joining_Group:    	Yeh}', "");
Expect(1, 2217, '\P{^Is_Joining_Group:    	Yeh}', "");
Expect(0, 2218, '\p{Is_Joining_Group:    	Yeh}', "");
Expect(1, 2218, '\p{^Is_Joining_Group:    	Yeh}', "");
Expect(1, 2218, '\P{Is_Joining_Group:    	Yeh}', "");
Expect(0, 2218, '\P{^Is_Joining_Group:    	Yeh}', "");
Error('\p{Is_Jg=_/a/Yeh}');
Error('\P{Is_Jg=_/a/Yeh}');
Expect(1, 2217, '\p{Is_Jg=yeh}', "");
Expect(0, 2217, '\p{^Is_Jg=yeh}', "");
Expect(0, 2217, '\P{Is_Jg=yeh}', "");
Expect(1, 2217, '\P{^Is_Jg=yeh}', "");
Expect(0, 2218, '\p{Is_Jg=yeh}', "");
Expect(1, 2218, '\p{^Is_Jg=yeh}', "");
Expect(1, 2218, '\P{Is_Jg=yeh}', "");
Expect(0, 2218, '\P{^Is_Jg=yeh}', "");
Expect(1, 2217, '\p{Is_Jg=__yeh}', "");
Expect(0, 2217, '\p{^Is_Jg=__yeh}', "");
Expect(0, 2217, '\P{Is_Jg=__yeh}', "");
Expect(1, 2217, '\P{^Is_Jg=__yeh}', "");
Expect(0, 2218, '\p{Is_Jg=__yeh}', "");
Expect(1, 2218, '\p{^Is_Jg=__yeh}', "");
Expect(1, 2218, '\P{Is_Jg=__yeh}', "");
Expect(0, 2218, '\P{^Is_Jg=__yeh}', "");
Error('\p{Joining_Group=-_Dal:=}');
Error('\P{Joining_Group=-_Dal:=}');
Expect(1, 1882, '\p{Joining_Group=dal}', "");
Expect(0, 1882, '\p{^Joining_Group=dal}', "");
Expect(0, 1882, '\P{Joining_Group=dal}', "");
Expect(1, 1882, '\P{^Joining_Group=dal}', "");
Expect(0, 1883, '\p{Joining_Group=dal}', "");
Expect(1, 1883, '\p{^Joining_Group=dal}', "");
Expect(1, 1883, '\P{Joining_Group=dal}', "");
Expect(0, 1883, '\P{^Joining_Group=dal}', "");
Expect(1, 1882, '\p{Joining_Group:	-DAL}', "");
Expect(0, 1882, '\p{^Joining_Group:	-DAL}', "");
Expect(0, 1882, '\P{Joining_Group:	-DAL}', "");
Expect(1, 1882, '\P{^Joining_Group:	-DAL}', "");
Expect(0, 1883, '\p{Joining_Group:	-DAL}', "");
Expect(1, 1883, '\p{^Joining_Group:	-DAL}', "");
Expect(1, 1883, '\P{Joining_Group:	-DAL}', "");
Expect(0, 1883, '\P{^Joining_Group:	-DAL}', "");
Error('\p{Jg:   :=	-DAL}');
Error('\P{Jg:   :=	-DAL}');
Expect(1, 1882, '\p{Jg=dal}', "");
Expect(0, 1882, '\p{^Jg=dal}', "");
Expect(0, 1882, '\P{Jg=dal}', "");
Expect(1, 1882, '\P{^Jg=dal}', "");
Expect(0, 1883, '\p{Jg=dal}', "");
Expect(1, 1883, '\p{^Jg=dal}', "");
Expect(1, 1883, '\P{Jg=dal}', "");
Expect(0, 1883, '\P{^Jg=dal}', "");
Expect(1, 1882, '\p{Jg=	Dal}', "");
Expect(0, 1882, '\p{^Jg=	Dal}', "");
Expect(0, 1882, '\P{Jg=	Dal}', "");
Expect(1, 1882, '\P{^Jg=	Dal}', "");
Expect(0, 1883, '\p{Jg=	Dal}', "");
Expect(1, 1883, '\p{^Jg=	Dal}', "");
Expect(1, 1883, '\P{Jg=	Dal}', "");
Expect(0, 1883, '\P{^Jg=	Dal}', "");
Error('\p{Is_Joining_Group=_/a/Dal}');
Error('\P{Is_Joining_Group=_/a/Dal}');
Expect(1, 1882, '\p{Is_Joining_Group=dal}', "");
Expect(0, 1882, '\p{^Is_Joining_Group=dal}', "");
Expect(0, 1882, '\P{Is_Joining_Group=dal}', "");
Expect(1, 1882, '\P{^Is_Joining_Group=dal}', "");
Expect(0, 1883, '\p{Is_Joining_Group=dal}', "");
Expect(1, 1883, '\p{^Is_Joining_Group=dal}', "");
Expect(1, 1883, '\P{Is_Joining_Group=dal}', "");
Expect(0, 1883, '\P{^Is_Joining_Group=dal}', "");
Error('\p{Is_Jg= :=DAL}');
Error('\P{Is_Jg= :=DAL}');
Expect(1, 1882, '\p{Is_Jg=dal}', "");
Expect(0, 1882, '\p{^Is_Jg=dal}', "");
Expect(0, 1882, '\P{Is_Jg=dal}', "");
Expect(1, 1882, '\P{^Is_Jg=dal}', "");
Expect(0, 1883, '\p{Is_Jg=dal}', "");
Expect(1, 1883, '\p{^Is_Jg=dal}', "");
Expect(1, 1883, '\P{Is_Jg=dal}', "");
Expect(0, 1883, '\P{^Is_Jg=dal}', "");
Expect(1, 1882, '\p{Is_Jg=	-Dal}', "");
Expect(0, 1882, '\p{^Is_Jg=	-Dal}', "");
Expect(0, 1882, '\P{Is_Jg=	-Dal}', "");
Expect(1, 1882, '\P{^Is_Jg=	-Dal}', "");
Expect(0, 1883, '\p{Is_Jg=	-Dal}', "");
Expect(1, 1883, '\p{^Is_Jg=	-Dal}', "");
Expect(1, 1883, '\P{Is_Jg=	-Dal}', "");
Expect(0, 1883, '\P{^Is_Jg=	-Dal}', "");
Error('\p{Joining_Group=He:=}');
Error('\P{Joining_Group=He:=}');
Expect(1, 1815, '\p{Joining_Group:he}', "");
Expect(0, 1815, '\p{^Joining_Group:he}', "");
Expect(0, 1815, '\P{Joining_Group:he}', "");
Expect(1, 1815, '\P{^Joining_Group:he}', "");
Expect(0, 1816, '\p{Joining_Group:he}', "");
Expect(1, 1816, '\p{^Joining_Group:he}', "");
Expect(1, 1816, '\P{Joining_Group:he}', "");
Expect(0, 1816, '\P{^Joining_Group:he}', "");
Expect(1, 1815, '\p{Joining_Group= 	He}', "");
Expect(0, 1815, '\p{^Joining_Group= 	He}', "");
Expect(0, 1815, '\P{Joining_Group= 	He}', "");
Expect(1, 1815, '\P{^Joining_Group= 	He}', "");
Expect(0, 1816, '\p{Joining_Group= 	He}', "");
Expect(1, 1816, '\p{^Joining_Group= 	He}', "");
Expect(1, 1816, '\P{Joining_Group= 	He}', "");
Expect(0, 1816, '\P{^Joining_Group= 	He}', "");
Error('\p{Jg= :=he}');
Error('\P{Jg= :=he}');
Expect(1, 1815, '\p{Jg=he}', "");
Expect(0, 1815, '\p{^Jg=he}', "");
Expect(0, 1815, '\P{Jg=he}', "");
Expect(1, 1815, '\P{^Jg=he}', "");
Expect(0, 1816, '\p{Jg=he}', "");
Expect(1, 1816, '\p{^Jg=he}', "");
Expect(1, 1816, '\P{Jg=he}', "");
Expect(0, 1816, '\P{^Jg=he}', "");
Expect(1, 1815, '\p{Jg=	He}', "");
Expect(0, 1815, '\p{^Jg=	He}', "");
Expect(0, 1815, '\P{Jg=	He}', "");
Expect(1, 1815, '\P{^Jg=	He}', "");
Expect(0, 1816, '\p{Jg=	He}', "");
Expect(1, 1816, '\p{^Jg=	He}', "");
Expect(1, 1816, '\P{Jg=	He}', "");
Expect(0, 1816, '\P{^Jg=	He}', "");
Error('\p{Is_Joining_Group=	/a/he}');
Error('\P{Is_Joining_Group=	/a/he}');
Expect(1, 1815, '\p{Is_Joining_Group=he}', "");
Expect(0, 1815, '\p{^Is_Joining_Group=he}', "");
Expect(0, 1815, '\P{Is_Joining_Group=he}', "");
Expect(1, 1815, '\P{^Is_Joining_Group=he}', "");
Expect(0, 1816, '\p{Is_Joining_Group=he}', "");
Expect(1, 1816, '\p{^Is_Joining_Group=he}', "");
Expect(1, 1816, '\P{Is_Joining_Group=he}', "");
Expect(0, 1816, '\P{^Is_Joining_Group=he}', "");
Expect(1, 1815, '\p{Is_Joining_Group:		_he}', "");
Expect(0, 1815, '\p{^Is_Joining_Group:		_he}', "");
Expect(0, 1815, '\P{Is_Joining_Group:		_he}', "");
Expect(1, 1815, '\P{^Is_Joining_Group:		_he}', "");
Expect(0, 1816, '\p{Is_Joining_Group:		_he}', "");
Expect(1, 1816, '\p{^Is_Joining_Group:		_he}', "");
Expect(1, 1816, '\P{Is_Joining_Group:		_he}', "");
Expect(0, 1816, '\P{^Is_Joining_Group:		_he}', "");
Error('\p{Is_Jg=-	HE:=}');
Error('\P{Is_Jg=-	HE:=}');
Expect(1, 1815, '\p{Is_Jg=he}', "");
Expect(0, 1815, '\p{^Is_Jg=he}', "");
Expect(0, 1815, '\P{Is_Jg=he}', "");
Expect(1, 1815, '\P{^Is_Jg=he}', "");
Expect(0, 1816, '\p{Is_Jg=he}', "");
Expect(1, 1816, '\p{^Is_Jg=he}', "");
Expect(1, 1816, '\P{Is_Jg=he}', "");
Expect(0, 1816, '\P{^Is_Jg=he}', "");
Expect(1, 1815, '\p{Is_Jg=_He}', "");
Expect(0, 1815, '\p{^Is_Jg=_He}', "");
Expect(0, 1815, '\P{Is_Jg=_He}', "");
Expect(1, 1815, '\P{^Is_Jg=_He}', "");
Expect(0, 1816, '\p{Is_Jg=_He}', "");
Expect(1, 1816, '\p{^Is_Jg=_He}', "");
Expect(1, 1816, '\P{Is_Jg=_He}', "");
Expect(0, 1816, '\P{^Is_Jg=_He}', "");
Error('\p{Joining_Group=-/a/no_Joining_Group}');
Error('\P{Joining_Group=-/a/no_Joining_Group}');
Expect(1, 2221, '\p{Joining_Group=nojoininggroup}', "");
Expect(0, 2221, '\p{^Joining_Group=nojoininggroup}', "");
Expect(0, 2221, '\P{Joining_Group=nojoininggroup}', "");
Expect(1, 2221, '\P{^Joining_Group=nojoininggroup}', "");
Expect(0, 2220, '\p{Joining_Group=nojoininggroup}', "");
Expect(1, 2220, '\p{^Joining_Group=nojoininggroup}', "");
Expect(1, 2220, '\P{Joining_Group=nojoininggroup}', "");
Expect(0, 2220, '\P{^Joining_Group=nojoininggroup}', "");
Expect(1, 2221, '\p{Joining_Group: -_No_joining_Group}', "");
Expect(0, 2221, '\p{^Joining_Group: -_No_joining_Group}', "");
Expect(0, 2221, '\P{Joining_Group: -_No_joining_Group}', "");
Expect(1, 2221, '\P{^Joining_Group: -_No_joining_Group}', "");
Expect(0, 2220, '\p{Joining_Group: -_No_joining_Group}', "");
Expect(1, 2220, '\p{^Joining_Group: -_No_joining_Group}', "");
Expect(1, 2220, '\P{Joining_Group: -_No_joining_Group}', "");
Expect(0, 2220, '\P{^Joining_Group: -_No_joining_Group}', "");
Error('\p{Jg= No_Joining_group:=}');
Error('\P{Jg= No_Joining_group:=}');
Expect(1, 2221, '\p{Jg=nojoininggroup}', "");
Expect(0, 2221, '\p{^Jg=nojoininggroup}', "");
Expect(0, 2221, '\P{Jg=nojoininggroup}', "");
Expect(1, 2221, '\P{^Jg=nojoininggroup}', "");
Expect(0, 2220, '\p{Jg=nojoininggroup}', "");
Expect(1, 2220, '\p{^Jg=nojoininggroup}', "");
Expect(1, 2220, '\P{Jg=nojoininggroup}', "");
Expect(0, 2220, '\P{^Jg=nojoininggroup}', "");
Expect(1, 2221, '\p{Jg:  _No_JOINING_Group}', "");
Expect(0, 2221, '\p{^Jg:  _No_JOINING_Group}', "");
Expect(0, 2221, '\P{Jg:  _No_JOINING_Group}', "");
Expect(1, 2221, '\P{^Jg:  _No_JOINING_Group}', "");
Expect(0, 2220, '\p{Jg:  _No_JOINING_Group}', "");
Expect(1, 2220, '\p{^Jg:  _No_JOINING_Group}', "");
Expect(1, 2220, '\P{Jg:  _No_JOINING_Group}', "");
Expect(0, 2220, '\P{^Jg:  _No_JOINING_Group}', "");
Error('\p{Is_Joining_Group=	:=NO_joining_GROUP}');
Error('\P{Is_Joining_Group=	:=NO_joining_GROUP}');
Expect(1, 2221, '\p{Is_Joining_Group=nojoininggroup}', "");
Expect(0, 2221, '\p{^Is_Joining_Group=nojoininggroup}', "");
Expect(0, 2221, '\P{Is_Joining_Group=nojoininggroup}', "");
Expect(1, 2221, '\P{^Is_Joining_Group=nojoininggroup}', "");
Expect(0, 2220, '\p{Is_Joining_Group=nojoininggroup}', "");
Expect(1, 2220, '\p{^Is_Joining_Group=nojoininggroup}', "");
Expect(1, 2220, '\P{Is_Joining_Group=nojoininggroup}', "");
Expect(0, 2220, '\P{^Is_Joining_Group=nojoininggroup}', "");
Expect(1, 2221, '\p{Is_Joining_Group=_NO_Joining_Group}', "");
Expect(0, 2221, '\p{^Is_Joining_Group=_NO_Joining_Group}', "");
Expect(0, 2221, '\P{Is_Joining_Group=_NO_Joining_Group}', "");
Expect(1, 2221, '\P{^Is_Joining_Group=_NO_Joining_Group}', "");
Expect(0, 2220, '\p{Is_Joining_Group=_NO_Joining_Group}', "");
Expect(1, 2220, '\p{^Is_Joining_Group=_NO_Joining_Group}', "");
Expect(1, 2220, '\P{Is_Joining_Group=_NO_Joining_Group}', "");
Expect(0, 2220, '\P{^Is_Joining_Group=_NO_Joining_Group}', "");
Error('\p{Is_Jg= /a/No_JOINING_Group}');
Error('\P{Is_Jg= /a/No_JOINING_Group}');
Expect(1, 2221, '\p{Is_Jg=nojoininggroup}', "");
Expect(0, 2221, '\p{^Is_Jg=nojoininggroup}', "");
Expect(0, 2221, '\P{Is_Jg=nojoininggroup}', "");
Expect(1, 2221, '\P{^Is_Jg=nojoininggroup}', "");
Expect(0, 2220, '\p{Is_Jg=nojoininggroup}', "");
Expect(1, 2220, '\p{^Is_Jg=nojoininggroup}', "");
Expect(1, 2220, '\P{Is_Jg=nojoininggroup}', "");
Expect(0, 2220, '\P{^Is_Jg=nojoininggroup}', "");
Expect(1, 2221, '\p{Is_Jg=	NO_joining_Group}', "");
Expect(0, 2221, '\p{^Is_Jg=	NO_joining_Group}', "");
Expect(0, 2221, '\P{Is_Jg=	NO_joining_Group}', "");
Expect(1, 2221, '\P{^Is_Jg=	NO_joining_Group}', "");
Expect(0, 2220, '\p{Is_Jg=	NO_joining_Group}', "");
Expect(1, 2220, '\p{^Is_Jg=	NO_joining_Group}', "");
Expect(1, 2220, '\P{Is_Jg=	NO_joining_Group}', "");
Expect(0, 2220, '\P{^Is_Jg=	NO_joining_Group}', "");
Error('\p{Joining_Group=/a/ Burushaski_Yeh_Barree}');
Error('\P{Joining_Group=/a/ Burushaski_Yeh_Barree}');
Expect(1, 1915, '\p{Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1915, '\p{^Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1915, '\P{Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1915, '\P{^Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1916, '\p{Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1916, '\p{^Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1916, '\P{Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1916, '\P{^Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1915, '\p{Joining_Group=		burushaski_Yeh_BARREE}', "");
Expect(0, 1915, '\p{^Joining_Group=		burushaski_Yeh_BARREE}', "");
Expect(0, 1915, '\P{Joining_Group=		burushaski_Yeh_BARREE}', "");
Expect(1, 1915, '\P{^Joining_Group=		burushaski_Yeh_BARREE}', "");
Expect(0, 1916, '\p{Joining_Group=		burushaski_Yeh_BARREE}', "");
Expect(1, 1916, '\p{^Joining_Group=		burushaski_Yeh_BARREE}', "");
Expect(1, 1916, '\P{Joining_Group=		burushaski_Yeh_BARREE}', "");
Expect(0, 1916, '\P{^Joining_Group=		burushaski_Yeh_BARREE}', "");
Error('\p{Jg=-_Burushaski_yeh_Barree:=}');
Error('\P{Jg=-_Burushaski_yeh_Barree:=}');
Expect(1, 1915, '\p{Jg=burushaskiyehbarree}', "");
Expect(0, 1915, '\p{^Jg=burushaskiyehbarree}', "");
Expect(0, 1915, '\P{Jg=burushaskiyehbarree}', "");
Expect(1, 1915, '\P{^Jg=burushaskiyehbarree}', "");
Expect(0, 1916, '\p{Jg=burushaskiyehbarree}', "");
Expect(1, 1916, '\p{^Jg=burushaskiyehbarree}', "");
Expect(1, 1916, '\P{Jg=burushaskiyehbarree}', "");
Expect(0, 1916, '\P{^Jg=burushaskiyehbarree}', "");
Expect(1, 1915, '\p{Jg= _burushaski_Yeh_Barree}', "");
Expect(0, 1915, '\p{^Jg= _burushaski_Yeh_Barree}', "");
Expect(0, 1915, '\P{Jg= _burushaski_Yeh_Barree}', "");
Expect(1, 1915, '\P{^Jg= _burushaski_Yeh_Barree}', "");
Expect(0, 1916, '\p{Jg= _burushaski_Yeh_Barree}', "");
Expect(1, 1916, '\p{^Jg= _burushaski_Yeh_Barree}', "");
Expect(1, 1916, '\P{Jg= _burushaski_Yeh_Barree}', "");
Expect(0, 1916, '\P{^Jg= _burushaski_Yeh_Barree}', "");
Error('\p{Is_Joining_Group=/a/	burushaski_yeh_BARREE}');
Error('\P{Is_Joining_Group=/a/	burushaski_yeh_BARREE}');
Expect(1, 1915, '\p{Is_Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1915, '\p{^Is_Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1915, '\P{Is_Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1915, '\P{^Is_Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1916, '\p{Is_Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1916, '\p{^Is_Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1916, '\P{Is_Joining_Group=burushaskiyehbarree}', "");
Expect(0, 1916, '\P{^Is_Joining_Group=burushaskiyehbarree}', "");
Expect(1, 1915, '\p{Is_Joining_Group=-_Burushaski_YEH_BARREE}', "");
Expect(0, 1915, '\p{^Is_Joining_Group=-_Burushaski_YEH_BARREE}', "");
Expect(0, 1915, '\P{Is_Joining_Group=-_Burushaski_YEH_BARREE}', "");
Expect(1, 1915, '\P{^Is_Joining_Group=-_Burushaski_YEH_BARREE}', "");
Expect(0, 1916, '\p{Is_Joining_Group=-_Burushaski_YEH_BARREE}', "");
Expect(1, 1916, '\p{^Is_Joining_Group=-_Burushaski_YEH_BARREE}', "");
Expect(1, 1916, '\P{Is_Joining_Group=-_Burushaski_YEH_BARREE}', "");
Expect(0, 1916, '\P{^Is_Joining_Group=-_Burushaski_YEH_BARREE}', "");
Error('\p{Is_Jg: /a/-BURUSHASKI_Yeh_BARREE}');
Error('\P{Is_Jg: /a/-BURUSHASKI_Yeh_BARREE}');
Expect(1, 1915, '\p{Is_Jg:burushaskiyehbarree}', "");
Expect(0, 1915, '\p{^Is_Jg:burushaskiyehbarree}', "");
Expect(0, 1915, '\P{Is_Jg:burushaskiyehbarree}', "");
Expect(1, 1915, '\P{^Is_Jg:burushaskiyehbarree}', "");
Expect(0, 1916, '\p{Is_Jg:burushaskiyehbarree}', "");
Expect(1, 1916, '\p{^Is_Jg:burushaskiyehbarree}', "");
Expect(1, 1916, '\P{Is_Jg:burushaskiyehbarree}', "");
Expect(0, 1916, '\P{^Is_Jg:burushaskiyehbarree}', "");
Expect(1, 1915, '\p{Is_Jg=- BURUSHASKI_Yeh_Barree}', "");
Expect(0, 1915, '\p{^Is_Jg=- BURUSHASKI_Yeh_Barree}', "");
Expect(0, 1915, '\P{Is_Jg=- BURUSHASKI_Yeh_Barree}', "");
Expect(1, 1915, '\P{^Is_Jg=- BURUSHASKI_Yeh_Barree}', "");
Expect(0, 1916, '\p{Is_Jg=- BURUSHASKI_Yeh_Barree}', "");
Expect(1, 1916, '\p{^Is_Jg=- BURUSHASKI_Yeh_Barree}', "");
Expect(1, 1916, '\P{Is_Jg=- BURUSHASKI_Yeh_Barree}', "");
Expect(0, 1916, '\P{^Is_Jg=- BURUSHASKI_Yeh_Barree}', "");
Error('\p{Joining_Group=/a/ 	TEH_Marbuta}');
Error('\P{Joining_Group=/a/ 	TEH_Marbuta}');
Expect(1, 1749, '\p{Joining_Group:   tehmarbuta}', "");
Expect(0, 1749, '\p{^Joining_Group:   tehmarbuta}', "");
Expect(0, 1749, '\P{Joining_Group:   tehmarbuta}', "");
Expect(1, 1749, '\P{^Joining_Group:   tehmarbuta}', "");
Expect(0, 1750, '\p{Joining_Group:   tehmarbuta}', "");
Expect(1, 1750, '\p{^Joining_Group:   tehmarbuta}', "");
Expect(1, 1750, '\P{Joining_Group:   tehmarbuta}', "");
Expect(0, 1750, '\P{^Joining_Group:   tehmarbuta}', "");
Expect(1, 1749, '\p{Joining_Group=	 teh_Marbuta}', "");
Expect(0, 1749, '\p{^Joining_Group=	 teh_Marbuta}', "");
Expect(0, 1749, '\P{Joining_Group=	 teh_Marbuta}', "");
Expect(1, 1749, '\P{^Joining_Group=	 teh_Marbuta}', "");
Expect(0, 1750, '\p{Joining_Group=	 teh_Marbuta}', "");
Expect(1, 1750, '\p{^Joining_Group=	 teh_Marbuta}', "");
Expect(1, 1750, '\P{Joining_Group=	 teh_Marbuta}', "");
Expect(0, 1750, '\P{^Joining_Group=	 teh_Marbuta}', "");
Error('\p{Jg=:=-Teh_marbuta}');
Error('\P{Jg=:=-Teh_marbuta}');
Expect(1, 1749, '\p{Jg=tehmarbuta}', "");
Expect(0, 1749, '\p{^Jg=tehmarbuta}', "");
Expect(0, 1749, '\P{Jg=tehmarbuta}', "");
Expect(1, 1749, '\P{^Jg=tehmarbuta}', "");
Expect(0, 1750, '\p{Jg=tehmarbuta}', "");
Expect(1, 1750, '\p{^Jg=tehmarbuta}', "");
Expect(1, 1750, '\P{Jg=tehmarbuta}', "");
Expect(0, 1750, '\P{^Jg=tehmarbuta}', "");
Expect(1, 1749, '\p{Jg= 	Teh_Marbuta}', "");
Expect(0, 1749, '\p{^Jg= 	Teh_Marbuta}', "");
Expect(0, 1749, '\P{Jg= 	Teh_Marbuta}', "");
Expect(1, 1749, '\P{^Jg= 	Teh_Marbuta}', "");
Expect(0, 1750, '\p{Jg= 	Teh_Marbuta}', "");
Expect(1, 1750, '\p{^Jg= 	Teh_Marbuta}', "");
Expect(1, 1750, '\P{Jg= 	Teh_Marbuta}', "");
Expect(0, 1750, '\P{^Jg= 	Teh_Marbuta}', "");
Error('\p{Is_Joining_Group:  -Teh_marbuta/a/}');
Error('\P{Is_Joining_Group:  -Teh_marbuta/a/}');
Expect(1, 1749, '\p{Is_Joining_Group=tehmarbuta}', "");
Expect(0, 1749, '\p{^Is_Joining_Group=tehmarbuta}', "");
Expect(0, 1749, '\P{Is_Joining_Group=tehmarbuta}', "");
Expect(1, 1749, '\P{^Is_Joining_Group=tehmarbuta}', "");
Expect(0, 1750, '\p{Is_Joining_Group=tehmarbuta}', "");
Expect(1, 1750, '\p{^Is_Joining_Group=tehmarbuta}', "");
Expect(1, 1750, '\P{Is_Joining_Group=tehmarbuta}', "");
Expect(0, 1750, '\P{^Is_Joining_Group=tehmarbuta}', "");
Expect(1, 1749, '\p{Is_Joining_Group=_-teh_Marbuta}', "");
Expect(0, 1749, '\p{^Is_Joining_Group=_-teh_Marbuta}', "");
Expect(0, 1749, '\P{Is_Joining_Group=_-teh_Marbuta}', "");
Expect(1, 1749, '\P{^Is_Joining_Group=_-teh_Marbuta}', "");
Expect(0, 1750, '\p{Is_Joining_Group=_-teh_Marbuta}', "");
Expect(1, 1750, '\p{^Is_Joining_Group=_-teh_Marbuta}', "");
Expect(1, 1750, '\P{Is_Joining_Group=_-teh_Marbuta}', "");
Expect(0, 1750, '\P{^Is_Joining_Group=_-teh_Marbuta}', "");
Error('\p{Is_Jg=:=Teh_Marbuta}');
Error('\P{Is_Jg=:=Teh_Marbuta}');
Expect(1, 1749, '\p{Is_Jg:	tehmarbuta}', "");
Expect(0, 1749, '\p{^Is_Jg:	tehmarbuta}', "");
Expect(0, 1749, '\P{Is_Jg:	tehmarbuta}', "");
Expect(1, 1749, '\P{^Is_Jg:	tehmarbuta}', "");
Expect(0, 1750, '\p{Is_Jg:	tehmarbuta}', "");
Expect(1, 1750, '\p{^Is_Jg:	tehmarbuta}', "");
Expect(1, 1750, '\P{Is_Jg:	tehmarbuta}', "");
Expect(0, 1750, '\P{^Is_Jg:	tehmarbuta}', "");
Expect(1, 1749, '\p{Is_Jg=		TEH_Marbuta}', "");
Expect(0, 1749, '\p{^Is_Jg=		TEH_Marbuta}', "");
Expect(0, 1749, '\P{Is_Jg=		TEH_Marbuta}', "");
Expect(1, 1749, '\P{^Is_Jg=		TEH_Marbuta}', "");
Expect(0, 1750, '\p{Is_Jg=		TEH_Marbuta}', "");
Expect(1, 1750, '\p{^Is_Jg=		TEH_Marbuta}', "");
Expect(1, 1750, '\P{Is_Jg=		TEH_Marbuta}', "");
Expect(0, 1750, '\P{^Is_Jg=		TEH_Marbuta}', "");
Error('\p{Joining_Group=	:=reh}');
Error('\P{Joining_Group=	:=reh}');
Expect(1, 2218, '\p{Joining_Group=reh}', "");
Expect(0, 2218, '\p{^Joining_Group=reh}', "");
Expect(0, 2218, '\P{Joining_Group=reh}', "");
Expect(1, 2218, '\P{^Joining_Group=reh}', "");
Expect(0, 2219, '\p{Joining_Group=reh}', "");
Expect(1, 2219, '\p{^Joining_Group=reh}', "");
Expect(1, 2219, '\P{Joining_Group=reh}', "");
Expect(0, 2219, '\P{^Joining_Group=reh}', "");
Expect(1, 2218, '\p{Joining_Group=  Reh}', "");
Expect(0, 2218, '\p{^Joining_Group=  Reh}', "");
Expect(0, 2218, '\P{Joining_Group=  Reh}', "");
Expect(1, 2218, '\P{^Joining_Group=  Reh}', "");
Expect(0, 2219, '\p{Joining_Group=  Reh}', "");
Expect(1, 2219, '\p{^Joining_Group=  Reh}', "");
Expect(1, 2219, '\P{Joining_Group=  Reh}', "");
Expect(0, 2219, '\P{^Joining_Group=  Reh}', "");
Error('\p{Jg= Reh:=}');
Error('\P{Jg= Reh:=}');
Expect(1, 2218, '\p{Jg=reh}', "");
Expect(0, 2218, '\p{^Jg=reh}', "");
Expect(0, 2218, '\P{Jg=reh}', "");
Expect(1, 2218, '\P{^Jg=reh}', "");
Expect(0, 2219, '\p{Jg=reh}', "");
Expect(1, 2219, '\p{^Jg=reh}', "");
Expect(1, 2219, '\P{Jg=reh}', "");
Expect(0, 2219, '\P{^Jg=reh}', "");
Expect(1, 2218, '\p{Jg= REH}', "");
Expect(0, 2218, '\p{^Jg= REH}', "");
Expect(0, 2218, '\P{Jg= REH}', "");
Expect(1, 2218, '\P{^Jg= REH}', "");
Expect(0, 2219, '\p{Jg= REH}', "");
Expect(1, 2219, '\p{^Jg= REH}', "");
Expect(1, 2219, '\P{Jg= REH}', "");
Expect(0, 2219, '\P{^Jg= REH}', "");
Error('\p{Is_Joining_Group=_ Reh/a/}');
Error('\P{Is_Joining_Group=_ Reh/a/}');
Expect(1, 2218, '\p{Is_Joining_Group=reh}', "");
Expect(0, 2218, '\p{^Is_Joining_Group=reh}', "");
Expect(0, 2218, '\P{Is_Joining_Group=reh}', "");
Expect(1, 2218, '\P{^Is_Joining_Group=reh}', "");
Expect(0, 2219, '\p{Is_Joining_Group=reh}', "");
Expect(1, 2219, '\p{^Is_Joining_Group=reh}', "");
Expect(1, 2219, '\P{Is_Joining_Group=reh}', "");
Expect(0, 2219, '\P{^Is_Joining_Group=reh}', "");
Expect(1, 2218, '\p{Is_Joining_Group:   -reh}', "");
Expect(0, 2218, '\p{^Is_Joining_Group:   -reh}', "");
Expect(0, 2218, '\P{Is_Joining_Group:   -reh}', "");
Expect(1, 2218, '\P{^Is_Joining_Group:   -reh}', "");
Expect(0, 2219, '\p{Is_Joining_Group:   -reh}', "");
Expect(1, 2219, '\p{^Is_Joining_Group:   -reh}', "");
Expect(1, 2219, '\P{Is_Joining_Group:   -reh}', "");
Expect(0, 2219, '\P{^Is_Joining_Group:   -reh}', "");
Error('\p{Is_Jg= /a/reh}');
Error('\P{Is_Jg= /a/reh}');
Expect(1, 2218, '\p{Is_Jg=reh}', "");
Expect(0, 2218, '\p{^Is_Jg=reh}', "");
Expect(0, 2218, '\P{Is_Jg=reh}', "");
Expect(1, 2218, '\P{^Is_Jg=reh}', "");
Expect(0, 2219, '\p{Is_Jg=reh}', "");
Expect(1, 2219, '\p{^Is_Jg=reh}', "");
Expect(1, 2219, '\P{Is_Jg=reh}', "");
Expect(0, 2219, '\P{^Is_Jg=reh}', "");
Expect(1, 2218, '\p{Is_Jg=	_Reh}', "");
Expect(0, 2218, '\p{^Is_Jg=	_Reh}', "");
Expect(0, 2218, '\P{Is_Jg=	_Reh}', "");
Expect(1, 2218, '\P{^Is_Jg=	_Reh}', "");
Expect(0, 2219, '\p{Is_Jg=	_Reh}', "");
Expect(1, 2219, '\p{^Is_Jg=	_Reh}', "");
Expect(1, 2219, '\P{Is_Jg=	_Reh}', "");
Expect(0, 2219, '\P{^Is_Jg=	_Reh}', "");
Error('\p{Joining_Group=:=__Zain}');
Error('\P{Joining_Group=:=__Zain}');
Expect(1, 1817, '\p{Joining_Group=zain}', "");
Expect(0, 1817, '\p{^Joining_Group=zain}', "");
Expect(0, 1817, '\P{Joining_Group=zain}', "");
Expect(1, 1817, '\P{^Joining_Group=zain}', "");
Expect(0, 1818, '\p{Joining_Group=zain}', "");
Expect(1, 1818, '\p{^Joining_Group=zain}', "");
Expect(1, 1818, '\P{Joining_Group=zain}', "");
Expect(0, 1818, '\P{^Joining_Group=zain}', "");
Expect(1, 1817, '\p{Joining_Group=	Zain}', "");
Expect(0, 1817, '\p{^Joining_Group=	Zain}', "");
Expect(0, 1817, '\P{Joining_Group=	Zain}', "");
Expect(1, 1817, '\P{^Joining_Group=	Zain}', "");
Expect(0, 1818, '\p{Joining_Group=	Zain}', "");
Expect(1, 1818, '\p{^Joining_Group=	Zain}', "");
Expect(1, 1818, '\P{Joining_Group=	Zain}', "");
Expect(0, 1818, '\P{^Joining_Group=	Zain}', "");
Error('\p{Jg=_:=Zain}');
Error('\P{Jg=_:=Zain}');
Expect(1, 1817, '\p{Jg=zain}', "");
Expect(0, 1817, '\p{^Jg=zain}', "");
Expect(0, 1817, '\P{Jg=zain}', "");
Expect(1, 1817, '\P{^Jg=zain}', "");
Expect(0, 1818, '\p{Jg=zain}', "");
Expect(1, 1818, '\p{^Jg=zain}', "");
Expect(1, 1818, '\P{Jg=zain}', "");
Expect(0, 1818, '\P{^Jg=zain}', "");
Expect(1, 1817, '\p{Jg= zain}', "");
Expect(0, 1817, '\p{^Jg= zain}', "");
Expect(0, 1817, '\P{Jg= zain}', "");
Expect(1, 1817, '\P{^Jg= zain}', "");
Expect(0, 1818, '\p{Jg= zain}', "");
Expect(1, 1818, '\p{^Jg= zain}', "");
Expect(1, 1818, '\P{Jg= zain}', "");
Expect(0, 1818, '\P{^Jg= zain}', "");
Error('\p{Is_Joining_Group=:= ZAIN}');
Error('\P{Is_Joining_Group=:= ZAIN}');
Expect(1, 1817, '\p{Is_Joining_Group=zain}', "");
Expect(0, 1817, '\p{^Is_Joining_Group=zain}', "");
Expect(0, 1817, '\P{Is_Joining_Group=zain}', "");
Expect(1, 1817, '\P{^Is_Joining_Group=zain}', "");
Expect(0, 1818, '\p{Is_Joining_Group=zain}', "");
Expect(1, 1818, '\p{^Is_Joining_Group=zain}', "");
Expect(1, 1818, '\P{Is_Joining_Group=zain}', "");
Expect(0, 1818, '\P{^Is_Joining_Group=zain}', "");
Expect(1, 1817, '\p{Is_Joining_Group=		zain}', "");
Expect(0, 1817, '\p{^Is_Joining_Group=		zain}', "");
Expect(0, 1817, '\P{Is_Joining_Group=		zain}', "");
Expect(1, 1817, '\P{^Is_Joining_Group=		zain}', "");
Expect(0, 1818, '\p{Is_Joining_Group=		zain}', "");
Expect(1, 1818, '\p{^Is_Joining_Group=		zain}', "");
Expect(1, 1818, '\P{Is_Joining_Group=		zain}', "");
Expect(0, 1818, '\P{^Is_Joining_Group=		zain}', "");
Error('\p{Is_Jg=_/a/Zain}');
Error('\P{Is_Jg=_/a/Zain}');
Expect(1, 1817, '\p{Is_Jg=zain}', "");
Expect(0, 1817, '\p{^Is_Jg=zain}', "");
Expect(0, 1817, '\P{Is_Jg=zain}', "");
Expect(1, 1817, '\P{^Is_Jg=zain}', "");
Expect(0, 1818, '\p{Is_Jg=zain}', "");
Expect(1, 1818, '\p{^Is_Jg=zain}', "");
Expect(1, 1818, '\P{Is_Jg=zain}', "");
Expect(0, 1818, '\P{^Is_Jg=zain}', "");
Expect(1, 1817, '\p{Is_Jg= Zain}', "");
Expect(0, 1817, '\p{^Is_Jg= Zain}', "");
Expect(0, 1817, '\P{Is_Jg= Zain}', "");
Expect(1, 1817, '\P{^Is_Jg= Zain}', "");
Expect(0, 1818, '\p{Is_Jg= Zain}', "");
Expect(1, 1818, '\p{^Is_Jg= Zain}', "");
Expect(1, 1818, '\P{Is_Jg= Zain}', "");
Expect(0, 1818, '\P{^Is_Jg= Zain}', "");
Error('\p{Joining_Group= 	ALEF/a/}');
Error('\P{Joining_Group= 	ALEF/a/}');
Expect(1, 1908, '\p{Joining_Group=alef}', "");
Expect(0, 1908, '\p{^Joining_Group=alef}', "");
Expect(0, 1908, '\P{Joining_Group=alef}', "");
Expect(1, 1908, '\P{^Joining_Group=alef}', "");
Expect(0, 1909, '\p{Joining_Group=alef}', "");
Expect(1, 1909, '\p{^Joining_Group=alef}', "");
Expect(1, 1909, '\P{Joining_Group=alef}', "");
Expect(0, 1909, '\P{^Joining_Group=alef}', "");
Expect(1, 1908, '\p{Joining_Group=  ALEF}', "");
Expect(0, 1908, '\p{^Joining_Group=  ALEF}', "");
Expect(0, 1908, '\P{Joining_Group=  ALEF}', "");
Expect(1, 1908, '\P{^Joining_Group=  ALEF}', "");
Expect(0, 1909, '\p{Joining_Group=  ALEF}', "");
Expect(1, 1909, '\p{^Joining_Group=  ALEF}', "");
Expect(1, 1909, '\P{Joining_Group=  ALEF}', "");
Expect(0, 1909, '\P{^Joining_Group=  ALEF}', "");
Error('\p{Jg=_:=Alef}');
Error('\P{Jg=_:=Alef}');
Expect(1, 1908, '\p{Jg=alef}', "");
Expect(0, 1908, '\p{^Jg=alef}', "");
Expect(0, 1908, '\P{Jg=alef}', "");
Expect(1, 1908, '\P{^Jg=alef}', "");
Expect(0, 1909, '\p{Jg=alef}', "");
Expect(1, 1909, '\p{^Jg=alef}', "");
Expect(1, 1909, '\P{Jg=alef}', "");
Expect(0, 1909, '\P{^Jg=alef}', "");
Expect(1, 1908, '\p{Jg=- Alef}', "");
Expect(0, 1908, '\p{^Jg=- Alef}', "");
Expect(0, 1908, '\P{Jg=- Alef}', "");
Expect(1, 1908, '\P{^Jg=- Alef}', "");
Expect(0, 1909, '\p{Jg=- Alef}', "");
Expect(1, 1909, '\p{^Jg=- Alef}', "");
Expect(1, 1909, '\P{Jg=- Alef}', "");
Expect(0, 1909, '\P{^Jg=- Alef}', "");
Error('\p{Is_Joining_Group:	:=Alef}');
Error('\P{Is_Joining_Group:	:=Alef}');
Expect(1, 1908, '\p{Is_Joining_Group=alef}', "");
Expect(0, 1908, '\p{^Is_Joining_Group=alef}', "");
Expect(0, 1908, '\P{Is_Joining_Group=alef}', "");
Expect(1, 1908, '\P{^Is_Joining_Group=alef}', "");
Expect(0, 1909, '\p{Is_Joining_Group=alef}', "");
Expect(1, 1909, '\p{^Is_Joining_Group=alef}', "");
Expect(1, 1909, '\P{Is_Joining_Group=alef}', "");
Expect(0, 1909, '\P{^Is_Joining_Group=alef}', "");
Expect(1, 1908, '\p{Is_Joining_Group=	Alef}', "");
Expect(0, 1908, '\p{^Is_Joining_Group=	Alef}', "");
Expect(0, 1908, '\P{Is_Joining_Group=	Alef}', "");
Expect(1, 1908, '\P{^Is_Joining_Group=	Alef}', "");
Expect(0, 1909, '\p{Is_Joining_Group=	Alef}', "");
Expect(1, 1909, '\p{^Is_Joining_Group=	Alef}', "");
Expect(1, 1909, '\P{Is_Joining_Group=	Alef}', "");
Expect(0, 1909, '\P{^Is_Joining_Group=	Alef}', "");
Error('\p{Is_Jg=-/a/alef}');
Error('\P{Is_Jg=-/a/alef}');
Expect(1, 1908, '\p{Is_Jg=alef}', "");
Expect(0, 1908, '\p{^Is_Jg=alef}', "");
Expect(0, 1908, '\P{Is_Jg=alef}', "");
Expect(1, 1908, '\P{^Is_Jg=alef}', "");
Expect(0, 1909, '\p{Is_Jg=alef}', "");
Expect(1, 1909, '\p{^Is_Jg=alef}', "");
Expect(1, 1909, '\P{Is_Jg=alef}', "");
Expect(0, 1909, '\P{^Is_Jg=alef}', "");
Expect(1, 1908, '\p{Is_Jg=	alef}', "");
Expect(0, 1908, '\p{^Is_Jg=	alef}', "");
Expect(0, 1908, '\P{Is_Jg=	alef}', "");
Expect(1, 1908, '\P{^Is_Jg=	alef}', "");
Expect(0, 1909, '\p{Is_Jg=	alef}', "");
Expect(1, 1909, '\p{^Is_Jg=	alef}', "");
Expect(1, 1909, '\P{Is_Jg=	alef}', "");
Expect(0, 1909, '\P{^Is_Jg=	alef}', "");
Error('\p{Joining_Group=	:=gamal}');
Error('\P{Joining_Group=	:=gamal}');
Expect(1, 1838, '\p{Joining_Group=gamal}', "");
Expect(0, 1838, '\p{^Joining_Group=gamal}', "");
Expect(0, 1838, '\P{Joining_Group=gamal}', "");
Expect(1, 1838, '\P{^Joining_Group=gamal}', "");
Expect(0, 1839, '\p{Joining_Group=gamal}', "");
Expect(1, 1839, '\p{^Joining_Group=gamal}', "");
Expect(1, 1839, '\P{Joining_Group=gamal}', "");
Expect(0, 1839, '\P{^Joining_Group=gamal}', "");
Expect(1, 1838, '\p{Joining_Group=	 gamal}', "");
Expect(0, 1838, '\p{^Joining_Group=	 gamal}', "");
Expect(0, 1838, '\P{Joining_Group=	 gamal}', "");
Expect(1, 1838, '\P{^Joining_Group=	 gamal}', "");
Expect(0, 1839, '\p{Joining_Group=	 gamal}', "");
Expect(1, 1839, '\p{^Joining_Group=	 gamal}', "");
Expect(1, 1839, '\P{Joining_Group=	 gamal}', "");
Expect(0, 1839, '\P{^Joining_Group=	 gamal}', "");
Error('\p{Jg=	gamal:=}');
Error('\P{Jg=	gamal:=}');
Expect(1, 1838, '\p{Jg: gamal}', "");
Expect(0, 1838, '\p{^Jg: gamal}', "");
Expect(0, 1838, '\P{Jg: gamal}', "");
Expect(1, 1838, '\P{^Jg: gamal}', "");
Expect(0, 1839, '\p{Jg: gamal}', "");
Expect(1, 1839, '\p{^Jg: gamal}', "");
Expect(1, 1839, '\P{Jg: gamal}', "");
Expect(0, 1839, '\P{^Jg: gamal}', "");
Expect(1, 1838, '\p{Jg=-Gamal}', "");
Expect(0, 1838, '\p{^Jg=-Gamal}', "");
Expect(0, 1838, '\P{Jg=-Gamal}', "");
Expect(1, 1838, '\P{^Jg=-Gamal}', "");
Expect(0, 1839, '\p{Jg=-Gamal}', "");
Expect(1, 1839, '\p{^Jg=-Gamal}', "");
Expect(1, 1839, '\P{Jg=-Gamal}', "");
Expect(0, 1839, '\P{^Jg=-Gamal}', "");
Error('\p{Is_Joining_Group:   :=_Gamal}');
Error('\P{Is_Joining_Group:   :=_Gamal}');
Expect(1, 1838, '\p{Is_Joining_Group=gamal}', "");
Expect(0, 1838, '\p{^Is_Joining_Group=gamal}', "");
Expect(0, 1838, '\P{Is_Joining_Group=gamal}', "");
Expect(1, 1838, '\P{^Is_Joining_Group=gamal}', "");
Expect(0, 1839, '\p{Is_Joining_Group=gamal}', "");
Expect(1, 1839, '\p{^Is_Joining_Group=gamal}', "");
Expect(1, 1839, '\P{Is_Joining_Group=gamal}', "");
Expect(0, 1839, '\P{^Is_Joining_Group=gamal}', "");
Expect(1, 1838, '\p{Is_Joining_Group=		Gamal}', "");
Expect(0, 1838, '\p{^Is_Joining_Group=		Gamal}', "");
Expect(0, 1838, '\P{Is_Joining_Group=		Gamal}', "");
Expect(1, 1838, '\P{^Is_Joining_Group=		Gamal}', "");
Expect(0, 1839, '\p{Is_Joining_Group=		Gamal}', "");
Expect(1, 1839, '\p{^Is_Joining_Group=		Gamal}', "");
Expect(1, 1839, '\P{Is_Joining_Group=		Gamal}', "");
Expect(0, 1839, '\P{^Is_Joining_Group=		Gamal}', "");
Error('\p{Is_Jg:	-Gamal/a/}');
Error('\P{Is_Jg:	-Gamal/a/}');
Expect(1, 1838, '\p{Is_Jg=gamal}', "");
Expect(0, 1838, '\p{^Is_Jg=gamal}', "");
Expect(0, 1838, '\P{Is_Jg=gamal}', "");
Expect(1, 1838, '\P{^Is_Jg=gamal}', "");
Expect(0, 1839, '\p{Is_Jg=gamal}', "");
Expect(1, 1839, '\p{^Is_Jg=gamal}', "");
Expect(1, 1839, '\P{Is_Jg=gamal}', "");
Expect(0, 1839, '\P{^Is_Jg=gamal}', "");
Expect(1, 1838, '\p{Is_Jg=__Gamal}', "");
Expect(0, 1838, '\p{^Is_Jg=__Gamal}', "");
Expect(0, 1838, '\P{Is_Jg=__Gamal}', "");
Expect(1, 1838, '\P{^Is_Jg=__Gamal}', "");
Expect(0, 1839, '\p{Is_Jg=__Gamal}', "");
Expect(1, 1839, '\p{^Is_Jg=__Gamal}', "");
Expect(1, 1839, '\P{Is_Jg=__Gamal}', "");
Expect(0, 1839, '\P{^Is_Jg=__Gamal}', "");
Error('\p{Joining_Group=:=_	Meem}');
Error('\P{Joining_Group=:=_	Meem}');
Expect(1, 2215, '\p{Joining_Group=meem}', "");
Expect(0, 2215, '\p{^Joining_Group=meem}', "");
Expect(0, 2215, '\P{Joining_Group=meem}', "");
Expect(1, 2215, '\P{^Joining_Group=meem}', "");
Expect(0, 2216, '\p{Joining_Group=meem}', "");
Expect(1, 2216, '\p{^Joining_Group=meem}', "");
Expect(1, 2216, '\P{Joining_Group=meem}', "");
Expect(0, 2216, '\P{^Joining_Group=meem}', "");
Expect(1, 2215, '\p{Joining_Group=	 Meem}', "");
Expect(0, 2215, '\p{^Joining_Group=	 Meem}', "");
Expect(0, 2215, '\P{Joining_Group=	 Meem}', "");
Expect(1, 2215, '\P{^Joining_Group=	 Meem}', "");
Expect(0, 2216, '\p{Joining_Group=	 Meem}', "");
Expect(1, 2216, '\p{^Joining_Group=	 Meem}', "");
Expect(1, 2216, '\P{Joining_Group=	 Meem}', "");
Expect(0, 2216, '\P{^Joining_Group=	 Meem}', "");
Error('\p{Jg=/a/_ meem}');
Error('\P{Jg=/a/_ meem}');
Expect(1, 2215, '\p{Jg:meem}', "");
Expect(0, 2215, '\p{^Jg:meem}', "");
Expect(0, 2215, '\P{Jg:meem}', "");
Expect(1, 2215, '\P{^Jg:meem}', "");
Expect(0, 2216, '\p{Jg:meem}', "");
Expect(1, 2216, '\p{^Jg:meem}', "");
Expect(1, 2216, '\P{Jg:meem}', "");
Expect(0, 2216, '\P{^Jg:meem}', "");
Expect(1, 2215, '\p{Jg=_-MEEM}', "");
Expect(0, 2215, '\p{^Jg=_-MEEM}', "");
Expect(0, 2215, '\P{Jg=_-MEEM}', "");
Expect(1, 2215, '\P{^Jg=_-MEEM}', "");
Expect(0, 2216, '\p{Jg=_-MEEM}', "");
Expect(1, 2216, '\p{^Jg=_-MEEM}', "");
Expect(1, 2216, '\P{Jg=_-MEEM}', "");
Expect(0, 2216, '\P{^Jg=_-MEEM}', "");
Error('\p{Is_Joining_Group=_:=Meem}');
Error('\P{Is_Joining_Group=_:=Meem}');
Expect(1, 2215, '\p{Is_Joining_Group=meem}', "");
Expect(0, 2215, '\p{^Is_Joining_Group=meem}', "");
Expect(0, 2215, '\P{Is_Joining_Group=meem}', "");
Expect(1, 2215, '\P{^Is_Joining_Group=meem}', "");
Expect(0, 2216, '\p{Is_Joining_Group=meem}', "");
Expect(1, 2216, '\p{^Is_Joining_Group=meem}', "");
Expect(1, 2216, '\P{Is_Joining_Group=meem}', "");
Expect(0, 2216, '\P{^Is_Joining_Group=meem}', "");
Expect(1, 2215, '\p{Is_Joining_Group=_Meem}', "");
Expect(0, 2215, '\p{^Is_Joining_Group=_Meem}', "");
Expect(0, 2215, '\P{Is_Joining_Group=_Meem}', "");
Expect(1, 2215, '\P{^Is_Joining_Group=_Meem}', "");
Expect(0, 2216, '\p{Is_Joining_Group=_Meem}', "");
Expect(1, 2216, '\p{^Is_Joining_Group=_Meem}', "");
Expect(1, 2216, '\P{Is_Joining_Group=_Meem}', "");
Expect(0, 2216, '\P{^Is_Joining_Group=_Meem}', "");
Error('\p{Is_Jg=/a/Meem}');
Error('\P{Is_Jg=/a/Meem}');
Expect(1, 2215, '\p{Is_Jg=meem}', "");
Expect(0, 2215, '\p{^Is_Jg=meem}', "");
Expect(0, 2215, '\P{Is_Jg=meem}', "");
Expect(1, 2215, '\P{^Is_Jg=meem}', "");
Expect(0, 2216, '\p{Is_Jg=meem}', "");
Expect(1, 2216, '\p{^Is_Jg=meem}', "");
Expect(1, 2216, '\P{Is_Jg=meem}', "");
Expect(0, 2216, '\P{^Is_Jg=meem}', "");
Expect(1, 2215, '\p{Is_Jg=_-Meem}', "");
Expect(0, 2215, '\p{^Is_Jg=_-Meem}', "");
Expect(0, 2215, '\P{Is_Jg=_-Meem}', "");
Expect(1, 2215, '\P{^Is_Jg=_-Meem}', "");
Expect(0, 2216, '\p{Is_Jg=_-Meem}', "");
Expect(1, 2216, '\p{^Is_Jg=_-Meem}', "");
Expect(1, 2216, '\P{Is_Jg=_-Meem}', "");
Expect(0, 2216, '\P{^Is_Jg=_-Meem}', "");
Error('\p{Joining_Group=	 Sadhe/a/}');
Error('\P{Joining_Group=	 Sadhe/a/}');
Expect(1, 1832, '\p{Joining_Group=sadhe}', "");
Expect(0, 1832, '\p{^Joining_Group=sadhe}', "");
Expect(0, 1832, '\P{Joining_Group=sadhe}', "");
Expect(1, 1832, '\P{^Joining_Group=sadhe}', "");
Expect(0, 1833, '\p{Joining_Group=sadhe}', "");
Expect(1, 1833, '\p{^Joining_Group=sadhe}', "");
Expect(1, 1833, '\P{Joining_Group=sadhe}', "");
Expect(0, 1833, '\P{^Joining_Group=sadhe}', "");
Expect(1, 1832, '\p{Joining_Group= Sadhe}', "");
Expect(0, 1832, '\p{^Joining_Group= Sadhe}', "");
Expect(0, 1832, '\P{Joining_Group= Sadhe}', "");
Expect(1, 1832, '\P{^Joining_Group= Sadhe}', "");
Expect(0, 1833, '\p{Joining_Group= Sadhe}', "");
Expect(1, 1833, '\p{^Joining_Group= Sadhe}', "");
Expect(1, 1833, '\P{Joining_Group= Sadhe}', "");
Expect(0, 1833, '\P{^Joining_Group= Sadhe}', "");
Error('\p{Jg=:= 	SADHE}');
Error('\P{Jg=:= 	SADHE}');
Expect(1, 1832, '\p{Jg=sadhe}', "");
Expect(0, 1832, '\p{^Jg=sadhe}', "");
Expect(0, 1832, '\P{Jg=sadhe}', "");
Expect(1, 1832, '\P{^Jg=sadhe}', "");
Expect(0, 1833, '\p{Jg=sadhe}', "");
Expect(1, 1833, '\p{^Jg=sadhe}', "");
Expect(1, 1833, '\P{Jg=sadhe}', "");
Expect(0, 1833, '\P{^Jg=sadhe}', "");
Expect(1, 1832, '\p{Jg=_ Sadhe}', "");
Expect(0, 1832, '\p{^Jg=_ Sadhe}', "");
Expect(0, 1832, '\P{Jg=_ Sadhe}', "");
Expect(1, 1832, '\P{^Jg=_ Sadhe}', "");
Expect(0, 1833, '\p{Jg=_ Sadhe}', "");
Expect(1, 1833, '\p{^Jg=_ Sadhe}', "");
Expect(1, 1833, '\P{Jg=_ Sadhe}', "");
Expect(0, 1833, '\P{^Jg=_ Sadhe}', "");
Error('\p{Is_Joining_Group=	-Sadhe:=}');
Error('\P{Is_Joining_Group=	-Sadhe:=}');
Expect(1, 1832, '\p{Is_Joining_Group=sadhe}', "");
Expect(0, 1832, '\p{^Is_Joining_Group=sadhe}', "");
Expect(0, 1832, '\P{Is_Joining_Group=sadhe}', "");
Expect(1, 1832, '\P{^Is_Joining_Group=sadhe}', "");
Expect(0, 1833, '\p{Is_Joining_Group=sadhe}', "");
Expect(1, 1833, '\p{^Is_Joining_Group=sadhe}', "");
Expect(1, 1833, '\P{Is_Joining_Group=sadhe}', "");
Expect(0, 1833, '\P{^Is_Joining_Group=sadhe}', "");
Expect(1, 1832, '\p{Is_Joining_Group=-sadhe}', "");
Expect(0, 1832, '\p{^Is_Joining_Group=-sadhe}', "");
Expect(0, 1832, '\P{Is_Joining_Group=-sadhe}', "");
Expect(1, 1832, '\P{^Is_Joining_Group=-sadhe}', "");
Expect(0, 1833, '\p{Is_Joining_Group=-sadhe}', "");
Expect(1, 1833, '\p{^Is_Joining_Group=-sadhe}', "");
Expect(1, 1833, '\P{Is_Joining_Group=-sadhe}', "");
Expect(0, 1833, '\P{^Is_Joining_Group=-sadhe}', "");
Error('\p{Is_Jg=_ SADHE/a/}');
Error('\P{Is_Jg=_ SADHE/a/}');
Expect(1, 1832, '\p{Is_Jg=sadhe}', "");
Expect(0, 1832, '\p{^Is_Jg=sadhe}', "");
Expect(0, 1832, '\P{Is_Jg=sadhe}', "");
Expect(1, 1832, '\P{^Is_Jg=sadhe}', "");
Expect(0, 1833, '\p{Is_Jg=sadhe}', "");
Expect(1, 1833, '\p{^Is_Jg=sadhe}', "");
Expect(1, 1833, '\P{Is_Jg=sadhe}', "");
Expect(0, 1833, '\P{^Is_Jg=sadhe}', "");
Expect(1, 1832, '\p{Is_Jg= SADHE}', "");
Expect(0, 1832, '\p{^Is_Jg= SADHE}', "");
Expect(0, 1832, '\P{Is_Jg= SADHE}', "");
Expect(1, 1832, '\P{^Is_Jg= SADHE}', "");
Expect(0, 1833, '\p{Is_Jg= SADHE}', "");
Expect(1, 1833, '\p{^Is_Jg= SADHE}', "");
Expect(1, 1833, '\P{Is_Jg= SADHE}', "");
Expect(0, 1833, '\P{^Is_Jg= SADHE}', "");
Error('\p{Joining_Group:  /a/beh}');
Error('\P{Joining_Group:  /a/beh}');
Expect(1, 2208, '\p{Joining_Group=beh}', "");
Expect(0, 2208, '\p{^Joining_Group=beh}', "");
Expect(0, 2208, '\P{Joining_Group=beh}', "");
Expect(1, 2208, '\P{^Joining_Group=beh}', "");
Expect(0, 2209, '\p{Joining_Group=beh}', "");
Expect(1, 2209, '\p{^Joining_Group=beh}', "");
Expect(1, 2209, '\P{Joining_Group=beh}', "");
Expect(0, 2209, '\P{^Joining_Group=beh}', "");
Expect(1, 2208, '\p{Joining_Group=Beh}', "");
Expect(0, 2208, '\p{^Joining_Group=Beh}', "");
Expect(0, 2208, '\P{Joining_Group=Beh}', "");
Expect(1, 2208, '\P{^Joining_Group=Beh}', "");
Expect(0, 2209, '\p{Joining_Group=Beh}', "");
Expect(1, 2209, '\p{^Joining_Group=Beh}', "");
Expect(1, 2209, '\P{Joining_Group=Beh}', "");
Expect(0, 2209, '\P{^Joining_Group=Beh}', "");
Error('\p{Jg=-:=Beh}');
Error('\P{Jg=-:=Beh}');
Expect(1, 2208, '\p{Jg=beh}', "");
Expect(0, 2208, '\p{^Jg=beh}', "");
Expect(0, 2208, '\P{Jg=beh}', "");
Expect(1, 2208, '\P{^Jg=beh}', "");
Expect(0, 2209, '\p{Jg=beh}', "");
Expect(1, 2209, '\p{^Jg=beh}', "");
Expect(1, 2209, '\P{Jg=beh}', "");
Expect(0, 2209, '\P{^Jg=beh}', "");
Expect(1, 2208, '\p{Jg:	_Beh}', "");
Expect(0, 2208, '\p{^Jg:	_Beh}', "");
Expect(0, 2208, '\P{Jg:	_Beh}', "");
Expect(1, 2208, '\P{^Jg:	_Beh}', "");
Expect(0, 2209, '\p{Jg:	_Beh}', "");
Expect(1, 2209, '\p{^Jg:	_Beh}', "");
Expect(1, 2209, '\P{Jg:	_Beh}', "");
Expect(0, 2209, '\P{^Jg:	_Beh}', "");
Error('\p{Is_Joining_Group=_ Beh/a/}');
Error('\P{Is_Joining_Group=_ Beh/a/}');
Expect(1, 2208, '\p{Is_Joining_Group=beh}', "");
Expect(0, 2208, '\p{^Is_Joining_Group=beh}', "");
Expect(0, 2208, '\P{Is_Joining_Group=beh}', "");
Expect(1, 2208, '\P{^Is_Joining_Group=beh}', "");
Expect(0, 2209, '\p{Is_Joining_Group=beh}', "");
Expect(1, 2209, '\p{^Is_Joining_Group=beh}', "");
Expect(1, 2209, '\P{Is_Joining_Group=beh}', "");
Expect(0, 2209, '\P{^Is_Joining_Group=beh}', "");
Expect(1, 2208, '\p{Is_Joining_Group=-	BEH}', "");
Expect(0, 2208, '\p{^Is_Joining_Group=-	BEH}', "");
Expect(0, 2208, '\P{Is_Joining_Group=-	BEH}', "");
Expect(1, 2208, '\P{^Is_Joining_Group=-	BEH}', "");
Expect(0, 2209, '\p{Is_Joining_Group=-	BEH}', "");
Expect(1, 2209, '\p{^Is_Joining_Group=-	BEH}', "");
Expect(1, 2209, '\P{Is_Joining_Group=-	BEH}', "");
Expect(0, 2209, '\P{^Is_Joining_Group=-	BEH}', "");
Error('\p{Is_Jg= /a/Beh}');
Error('\P{Is_Jg= /a/Beh}');
Expect(1, 2208, '\p{Is_Jg=beh}', "");
Expect(0, 2208, '\p{^Is_Jg=beh}', "");
Expect(0, 2208, '\P{Is_Jg=beh}', "");
Expect(1, 2208, '\P{^Is_Jg=beh}', "");
Expect(0, 2209, '\p{Is_Jg=beh}', "");
Expect(1, 2209, '\p{^Is_Jg=beh}', "");
Expect(1, 2209, '\P{Is_Jg=beh}', "");
Expect(0, 2209, '\P{^Is_Jg=beh}', "");
Expect(1, 2208, '\p{Is_Jg=-Beh}', "");
Expect(0, 2208, '\p{^Is_Jg=-Beh}', "");
Expect(0, 2208, '\P{Is_Jg=-Beh}', "");
Expect(1, 2208, '\P{^Is_Jg=-Beh}', "");
Expect(0, 2209, '\p{Is_Jg=-Beh}', "");
Expect(1, 2209, '\p{^Is_Jg=-Beh}', "");
Expect(1, 2209, '\P{Is_Jg=-Beh}', "");
Expect(0, 2209, '\P{^Is_Jg=-Beh}', "");
Error('\p{Joining_Group=		YEH_with_Tail:=}');
Error('\P{Joining_Group=		YEH_with_Tail:=}');
Expect(1, 1741, '\p{Joining_Group:   yehwithtail}', "");
Expect(0, 1741, '\p{^Joining_Group:   yehwithtail}', "");
Expect(0, 1741, '\P{Joining_Group:   yehwithtail}', "");
Expect(1, 1741, '\P{^Joining_Group:   yehwithtail}', "");
Expect(0, 1742, '\p{Joining_Group:   yehwithtail}', "");
Expect(1, 1742, '\p{^Joining_Group:   yehwithtail}', "");
Expect(1, 1742, '\P{Joining_Group:   yehwithtail}', "");
Expect(0, 1742, '\P{^Joining_Group:   yehwithtail}', "");
Expect(1, 1741, '\p{Joining_Group=_	yeh_With_TAIL}', "");
Expect(0, 1741, '\p{^Joining_Group=_	yeh_With_TAIL}', "");
Expect(0, 1741, '\P{Joining_Group=_	yeh_With_TAIL}', "");
Expect(1, 1741, '\P{^Joining_Group=_	yeh_With_TAIL}', "");
Expect(0, 1742, '\p{Joining_Group=_	yeh_With_TAIL}', "");
Expect(1, 1742, '\p{^Joining_Group=_	yeh_With_TAIL}', "");
Expect(1, 1742, '\P{Joining_Group=_	yeh_With_TAIL}', "");
Expect(0, 1742, '\P{^Joining_Group=_	yeh_With_TAIL}', "");
Error('\p{Jg=:=Yeh_with_Tail}');
Error('\P{Jg=:=Yeh_with_Tail}');
Expect(1, 1741, '\p{Jg=yehwithtail}', "");
Expect(0, 1741, '\p{^Jg=yehwithtail}', "");
Expect(0, 1741, '\P{Jg=yehwithtail}', "");
Expect(1, 1741, '\P{^Jg=yehwithtail}', "");
Expect(0, 1742, '\p{Jg=yehwithtail}', "");
Expect(1, 1742, '\p{^Jg=yehwithtail}', "");
Expect(1, 1742, '\P{Jg=yehwithtail}', "");
Expect(0, 1742, '\P{^Jg=yehwithtail}', "");
Expect(1, 1741, '\p{Jg: _yeh_With_TAIL}', "");
Expect(0, 1741, '\p{^Jg: _yeh_With_TAIL}', "");
Expect(0, 1741, '\P{Jg: _yeh_With_TAIL}', "");
Expect(1, 1741, '\P{^Jg: _yeh_With_TAIL}', "");
Expect(0, 1742, '\p{Jg: _yeh_With_TAIL}', "");
Expect(1, 1742, '\p{^Jg: _yeh_With_TAIL}', "");
Expect(1, 1742, '\P{Jg: _yeh_With_TAIL}', "");
Expect(0, 1742, '\P{^Jg: _yeh_With_TAIL}', "");
Error('\p{Is_Joining_Group=	/a/YEH_with_Tail}');
Error('\P{Is_Joining_Group=	/a/YEH_with_Tail}');
Expect(1, 1741, '\p{Is_Joining_Group=yehwithtail}', "");
Expect(0, 1741, '\p{^Is_Joining_Group=yehwithtail}', "");
Expect(0, 1741, '\P{Is_Joining_Group=yehwithtail}', "");
Expect(1, 1741, '\P{^Is_Joining_Group=yehwithtail}', "");
Expect(0, 1742, '\p{Is_Joining_Group=yehwithtail}', "");
Expect(1, 1742, '\p{^Is_Joining_Group=yehwithtail}', "");
Expect(1, 1742, '\P{Is_Joining_Group=yehwithtail}', "");
Expect(0, 1742, '\P{^Is_Joining_Group=yehwithtail}', "");
Expect(1, 1741, '\p{Is_Joining_Group= 	yeh_with_tail}', "");
Expect(0, 1741, '\p{^Is_Joining_Group= 	yeh_with_tail}', "");
Expect(0, 1741, '\P{Is_Joining_Group= 	yeh_with_tail}', "");
Expect(1, 1741, '\P{^Is_Joining_Group= 	yeh_with_tail}', "");
Expect(0, 1742, '\p{Is_Joining_Group= 	yeh_with_tail}', "");
Expect(1, 1742, '\p{^Is_Joining_Group= 	yeh_with_tail}', "");
Expect(1, 1742, '\P{Is_Joining_Group= 	yeh_with_tail}', "");
Expect(0, 1742, '\P{^Is_Joining_Group= 	yeh_with_tail}', "");
Error('\p{Is_Jg=/a/yeh_With_tail}');
Error('\P{Is_Jg=/a/yeh_With_tail}');
Expect(1, 1741, '\p{Is_Jg=yehwithtail}', "");
Expect(0, 1741, '\p{^Is_Jg=yehwithtail}', "");
Expect(0, 1741, '\P{Is_Jg=yehwithtail}', "");
Expect(1, 1741, '\P{^Is_Jg=yehwithtail}', "");
Expect(0, 1742, '\p{Is_Jg=yehwithtail}', "");
Expect(1, 1742, '\p{^Is_Jg=yehwithtail}', "");
Expect(1, 1742, '\P{Is_Jg=yehwithtail}', "");
Expect(0, 1742, '\P{^Is_Jg=yehwithtail}', "");
Expect(1, 1741, '\p{Is_Jg:   -_YEH_With_TAIL}', "");
Expect(0, 1741, '\p{^Is_Jg:   -_YEH_With_TAIL}', "");
Expect(0, 1741, '\P{Is_Jg:   -_YEH_With_TAIL}', "");
Expect(1, 1741, '\P{^Is_Jg:   -_YEH_With_TAIL}', "");
Expect(0, 1742, '\p{Is_Jg:   -_YEH_With_TAIL}', "");
Expect(1, 1742, '\p{^Is_Jg:   -_YEH_With_TAIL}', "");
Expect(1, 1742, '\P{Is_Jg:   -_YEH_With_TAIL}', "");
Expect(0, 1742, '\P{^Is_Jg:   -_YEH_With_TAIL}', "");
Error('\p{Joining_Group=:=- tah}');
Error('\P{Joining_Group=:=- tah}');
Expect(1, 2211, '\p{Joining_Group:tah}', "");
Expect(0, 2211, '\p{^Joining_Group:tah}', "");
Expect(0, 2211, '\P{Joining_Group:tah}', "");
Expect(1, 2211, '\P{^Joining_Group:tah}', "");
Expect(0, 2212, '\p{Joining_Group:tah}', "");
Expect(1, 2212, '\p{^Joining_Group:tah}', "");
Expect(1, 2212, '\P{Joining_Group:tah}', "");
Expect(0, 2212, '\P{^Joining_Group:tah}', "");
Expect(1, 2211, '\p{Joining_Group:  _Tah}', "");
Expect(0, 2211, '\p{^Joining_Group:  _Tah}', "");
Expect(0, 2211, '\P{Joining_Group:  _Tah}', "");
Expect(1, 2211, '\P{^Joining_Group:  _Tah}', "");
Expect(0, 2212, '\p{Joining_Group:  _Tah}', "");
Expect(1, 2212, '\p{^Joining_Group:  _Tah}', "");
Expect(1, 2212, '\P{Joining_Group:  _Tah}', "");
Expect(0, 2212, '\P{^Joining_Group:  _Tah}', "");
Error('\p{Jg=-_Tah/a/}');
Error('\P{Jg=-_Tah/a/}');
Expect(1, 2211, '\p{Jg=tah}', "");
Expect(0, 2211, '\p{^Jg=tah}', "");
Expect(0, 2211, '\P{Jg=tah}', "");
Expect(1, 2211, '\P{^Jg=tah}', "");
Expect(0, 2212, '\p{Jg=tah}', "");
Expect(1, 2212, '\p{^Jg=tah}', "");
Expect(1, 2212, '\P{Jg=tah}', "");
Expect(0, 2212, '\P{^Jg=tah}', "");
Expect(1, 2211, '\p{Jg=_	Tah}', "");
Expect(0, 2211, '\p{^Jg=_	Tah}', "");
Expect(0, 2211, '\P{Jg=_	Tah}', "");
Expect(1, 2211, '\P{^Jg=_	Tah}', "");
Expect(0, 2212, '\p{Jg=_	Tah}', "");
Expect(1, 2212, '\p{^Jg=_	Tah}', "");
Expect(1, 2212, '\P{Jg=_	Tah}', "");
Expect(0, 2212, '\P{^Jg=_	Tah}', "");
Error('\p{Is_Joining_Group=/a/TAH}');
Error('\P{Is_Joining_Group=/a/TAH}');
Expect(1, 2211, '\p{Is_Joining_Group=tah}', "");
Expect(0, 2211, '\p{^Is_Joining_Group=tah}', "");
Expect(0, 2211, '\P{Is_Joining_Group=tah}', "");
Expect(1, 2211, '\P{^Is_Joining_Group=tah}', "");
Expect(0, 2212, '\p{Is_Joining_Group=tah}', "");
Expect(1, 2212, '\p{^Is_Joining_Group=tah}', "");
Expect(1, 2212, '\P{Is_Joining_Group=tah}', "");
Expect(0, 2212, '\P{^Is_Joining_Group=tah}', "");
Expect(1, 2211, '\p{Is_Joining_Group=Tah}', "");
Expect(0, 2211, '\p{^Is_Joining_Group=Tah}', "");
Expect(0, 2211, '\P{Is_Joining_Group=Tah}', "");
Expect(1, 2211, '\P{^Is_Joining_Group=Tah}', "");
Expect(0, 2212, '\p{Is_Joining_Group=Tah}', "");
Expect(1, 2212, '\p{^Is_Joining_Group=Tah}', "");
Expect(1, 2212, '\P{Is_Joining_Group=Tah}', "");
Expect(0, 2212, '\P{^Is_Joining_Group=Tah}', "");
Error('\p{Is_Jg=:=TAH}');
Error('\P{Is_Jg=:=TAH}');
Expect(1, 2211, '\p{Is_Jg=tah}', "");
Expect(0, 2211, '\p{^Is_Jg=tah}', "");
Expect(0, 2211, '\P{Is_Jg=tah}', "");
Expect(1, 2211, '\P{^Is_Jg=tah}', "");
Expect(0, 2212, '\p{Is_Jg=tah}', "");
Expect(1, 2212, '\p{^Is_Jg=tah}', "");
Expect(1, 2212, '\P{Is_Jg=tah}', "");
Expect(0, 2212, '\P{^Is_Jg=tah}', "");
Expect(1, 2211, '\p{Is_Jg=_	TAH}', "");
Expect(0, 2211, '\p{^Is_Jg=_	TAH}', "");
Expect(0, 2211, '\P{Is_Jg=_	TAH}', "");
Expect(1, 2211, '\P{^Is_Jg=_	TAH}', "");
Expect(0, 2212, '\p{Is_Jg=_	TAH}', "");
Expect(1, 2212, '\p{^Is_Jg=_	TAH}', "");
Expect(1, 2212, '\P{Is_Jg=_	TAH}', "");
Expect(0, 2212, '\P{^Is_Jg=_	TAH}', "");
Error('\p{Joining_Group=/a/  Zhain}');
Error('\P{Joining_Group=/a/  Zhain}');
Expect(1, 1869, '\p{Joining_Group:   zhain}', "");
Expect(0, 1869, '\p{^Joining_Group:   zhain}', "");
Expect(0, 1869, '\P{Joining_Group:   zhain}', "");
Expect(1, 1869, '\P{^Joining_Group:   zhain}', "");
Expect(0, 1870, '\p{Joining_Group:   zhain}', "");
Expect(1, 1870, '\p{^Joining_Group:   zhain}', "");
Expect(1, 1870, '\P{Joining_Group:   zhain}', "");
Expect(0, 1870, '\P{^Joining_Group:   zhain}', "");
Expect(1, 1869, '\p{Joining_Group=-	ZHAIN}', "");
Expect(0, 1869, '\p{^Joining_Group=-	ZHAIN}', "");
Expect(0, 1869, '\P{Joining_Group=-	ZHAIN}', "");
Expect(1, 1869, '\P{^Joining_Group=-	ZHAIN}', "");
Expect(0, 1870, '\p{Joining_Group=-	ZHAIN}', "");
Expect(1, 1870, '\p{^Joining_Group=-	ZHAIN}', "");
Expect(1, 1870, '\P{Joining_Group=-	ZHAIN}', "");
Expect(0, 1870, '\P{^Joining_Group=-	ZHAIN}', "");
Error('\p{Jg:-_Zhain:=}');
Error('\P{Jg:-_Zhain:=}');
Expect(1, 1869, '\p{Jg=zhain}', "");
Expect(0, 1869, '\p{^Jg=zhain}', "");
Expect(0, 1869, '\P{Jg=zhain}', "");
Expect(1, 1869, '\P{^Jg=zhain}', "");
Expect(0, 1870, '\p{Jg=zhain}', "");
Expect(1, 1870, '\p{^Jg=zhain}', "");
Expect(1, 1870, '\P{Jg=zhain}', "");
Expect(0, 1870, '\P{^Jg=zhain}', "");
Expect(1, 1869, '\p{Jg=	Zhain}', "");
Expect(0, 1869, '\p{^Jg=	Zhain}', "");
Expect(0, 1869, '\P{Jg=	Zhain}', "");
Expect(1, 1869, '\P{^Jg=	Zhain}', "");
Expect(0, 1870, '\p{Jg=	Zhain}', "");
Expect(1, 1870, '\p{^Jg=	Zhain}', "");
Expect(1, 1870, '\P{Jg=	Zhain}', "");
Expect(0, 1870, '\P{^Jg=	Zhain}', "");
Error('\p{Is_Joining_Group:	/a/- Zhain}');
Error('\P{Is_Joining_Group:	/a/- Zhain}');
Expect(1, 1869, '\p{Is_Joining_Group:   zhain}', "");
Expect(0, 1869, '\p{^Is_Joining_Group:   zhain}', "");
Expect(0, 1869, '\P{Is_Joining_Group:   zhain}', "");
Expect(1, 1869, '\P{^Is_Joining_Group:   zhain}', "");
Expect(0, 1870, '\p{Is_Joining_Group:   zhain}', "");
Expect(1, 1870, '\p{^Is_Joining_Group:   zhain}', "");
Expect(1, 1870, '\P{Is_Joining_Group:   zhain}', "");
Expect(0, 1870, '\P{^Is_Joining_Group:   zhain}', "");
Expect(1, 1869, '\p{Is_Joining_Group=	-Zhain}', "");
Expect(0, 1869, '\p{^Is_Joining_Group=	-Zhain}', "");
Expect(0, 1869, '\P{Is_Joining_Group=	-Zhain}', "");
Expect(1, 1869, '\P{^Is_Joining_Group=	-Zhain}', "");
Expect(0, 1870, '\p{Is_Joining_Group=	-Zhain}', "");
Expect(1, 1870, '\p{^Is_Joining_Group=	-Zhain}', "");
Expect(1, 1870, '\P{Is_Joining_Group=	-Zhain}', "");
Expect(0, 1870, '\P{^Is_Joining_Group=	-Zhain}', "");
Error('\p{Is_Jg= _zhain:=}');
Error('\P{Is_Jg= _zhain:=}');
Expect(1, 1869, '\p{Is_Jg=zhain}', "");
Expect(0, 1869, '\p{^Is_Jg=zhain}', "");
Expect(0, 1869, '\P{Is_Jg=zhain}', "");
Expect(1, 1869, '\P{^Is_Jg=zhain}', "");
Expect(0, 1870, '\p{Is_Jg=zhain}', "");
Expect(1, 1870, '\p{^Is_Jg=zhain}', "");
Expect(1, 1870, '\P{Is_Jg=zhain}', "");
Expect(0, 1870, '\P{^Is_Jg=zhain}', "");
Expect(1, 1869, '\p{Is_Jg:	 _Zhain}', "");
Expect(0, 1869, '\p{^Is_Jg:	 _Zhain}', "");
Expect(0, 1869, '\P{Is_Jg:	 _Zhain}', "");
Expect(1, 1869, '\P{^Is_Jg:	 _Zhain}', "");
Expect(0, 1870, '\p{Is_Jg:	 _Zhain}', "");
Expect(1, 1870, '\p{^Is_Jg:	 _Zhain}', "");
Expect(1, 1870, '\P{Is_Jg:	 _Zhain}', "");
Expect(0, 1870, '\P{^Is_Jg:	 _Zhain}', "");
Error('\p{Joining_Group=/a/__Pe}');
Error('\P{Joining_Group=/a/__Pe}');
Expect(1, 1830, '\p{Joining_Group=pe}', "");
Expect(0, 1830, '\p{^Joining_Group=pe}', "");
Expect(0, 1830, '\P{Joining_Group=pe}', "");
Expect(1, 1830, '\P{^Joining_Group=pe}', "");
Expect(0, 1831, '\p{Joining_Group=pe}', "");
Expect(1, 1831, '\p{^Joining_Group=pe}', "");
Expect(1, 1831, '\P{Joining_Group=pe}', "");
Expect(0, 1831, '\P{^Joining_Group=pe}', "");
Expect(1, 1830, '\p{Joining_Group:		_Pe}', "");
Expect(0, 1830, '\p{^Joining_Group:		_Pe}', "");
Expect(0, 1830, '\P{Joining_Group:		_Pe}', "");
Expect(1, 1830, '\P{^Joining_Group:		_Pe}', "");
Expect(0, 1831, '\p{Joining_Group:		_Pe}', "");
Expect(1, 1831, '\p{^Joining_Group:		_Pe}', "");
Expect(1, 1831, '\P{Joining_Group:		_Pe}', "");
Expect(0, 1831, '\P{^Joining_Group:		_Pe}', "");
Error('\p{Jg:   /a/ PE}');
Error('\P{Jg:   /a/ PE}');
Expect(1, 1830, '\p{Jg=pe}', "");
Expect(0, 1830, '\p{^Jg=pe}', "");
Expect(0, 1830, '\P{Jg=pe}', "");
Expect(1, 1830, '\P{^Jg=pe}', "");
Expect(0, 1831, '\p{Jg=pe}', "");
Expect(1, 1831, '\p{^Jg=pe}', "");
Expect(1, 1831, '\P{Jg=pe}', "");
Expect(0, 1831, '\P{^Jg=pe}', "");
Expect(1, 1830, '\p{Jg= Pe}', "");
Expect(0, 1830, '\p{^Jg= Pe}', "");
Expect(0, 1830, '\P{Jg= Pe}', "");
Expect(1, 1830, '\P{^Jg= Pe}', "");
Expect(0, 1831, '\p{Jg= Pe}', "");
Expect(1, 1831, '\p{^Jg= Pe}', "");
Expect(1, 1831, '\P{Jg= Pe}', "");
Expect(0, 1831, '\P{^Jg= Pe}', "");
Error('\p{Is_Joining_Group=_/a/Pe}');
Error('\P{Is_Joining_Group=_/a/Pe}');
Expect(1, 1830, '\p{Is_Joining_Group=pe}', "");
Expect(0, 1830, '\p{^Is_Joining_Group=pe}', "");
Expect(0, 1830, '\P{Is_Joining_Group=pe}', "");
Expect(1, 1830, '\P{^Is_Joining_Group=pe}', "");
Expect(0, 1831, '\p{Is_Joining_Group=pe}', "");
Expect(1, 1831, '\p{^Is_Joining_Group=pe}', "");
Expect(1, 1831, '\P{Is_Joining_Group=pe}', "");
Expect(0, 1831, '\P{^Is_Joining_Group=pe}', "");
Expect(1, 1830, '\p{Is_Joining_Group=-PE}', "");
Expect(0, 1830, '\p{^Is_Joining_Group=-PE}', "");
Expect(0, 1830, '\P{Is_Joining_Group=-PE}', "");
Expect(1, 1830, '\P{^Is_Joining_Group=-PE}', "");
Expect(0, 1831, '\p{Is_Joining_Group=-PE}', "");
Expect(1, 1831, '\p{^Is_Joining_Group=-PE}', "");
Expect(1, 1831, '\P{Is_Joining_Group=-PE}', "");
Expect(0, 1831, '\P{^Is_Joining_Group=-PE}', "");
Error('\p{Is_Jg=_/a/Pe}');
Error('\P{Is_Jg=_/a/Pe}');
Expect(1, 1830, '\p{Is_Jg=pe}', "");
Expect(0, 1830, '\p{^Is_Jg=pe}', "");
Expect(0, 1830, '\P{Is_Jg=pe}', "");
Expect(1, 1830, '\P{^Is_Jg=pe}', "");
Expect(0, 1831, '\p{Is_Jg=pe}', "");
Expect(1, 1831, '\p{^Is_Jg=pe}', "");
Expect(1, 1831, '\P{Is_Jg=pe}', "");
Expect(0, 1831, '\P{^Is_Jg=pe}', "");
Expect(1, 1830, '\p{Is_Jg:	-_pe}', "");
Expect(0, 1830, '\p{^Is_Jg:	-_pe}', "");
Expect(0, 1830, '\P{Is_Jg:	-_pe}', "");
Expect(1, 1830, '\P{^Is_Jg:	-_pe}', "");
Expect(0, 1831, '\p{Is_Jg:	-_pe}', "");
Expect(1, 1831, '\p{^Is_Jg:	-_pe}', "");
Expect(1, 1831, '\P{Is_Jg:	-_pe}', "");
Expect(0, 1831, '\P{^Is_Jg:	-_pe}', "");
Error('\p{Joining_Group= /a/GAF}');
Error('\P{Joining_Group= /a/GAF}');
Expect(1, 1892, '\p{Joining_Group:   gaf}', "");
Expect(0, 1892, '\p{^Joining_Group:   gaf}', "");
Expect(0, 1892, '\P{Joining_Group:   gaf}', "");
Expect(1, 1892, '\P{^Joining_Group:   gaf}', "");
Expect(0, 1893, '\p{Joining_Group:   gaf}', "");
Expect(1, 1893, '\p{^Joining_Group:   gaf}', "");
Expect(1, 1893, '\P{Joining_Group:   gaf}', "");
Expect(0, 1893, '\P{^Joining_Group:   gaf}', "");
Expect(1, 1892, '\p{Joining_Group=	Gaf}', "");
Expect(0, 1892, '\p{^Joining_Group=	Gaf}', "");
Expect(0, 1892, '\P{Joining_Group=	Gaf}', "");
Expect(1, 1892, '\P{^Joining_Group=	Gaf}', "");
Expect(0, 1893, '\p{Joining_Group=	Gaf}', "");
Expect(1, 1893, '\p{^Joining_Group=	Gaf}', "");
Expect(1, 1893, '\P{Joining_Group=	Gaf}', "");
Expect(0, 1893, '\P{^Joining_Group=	Gaf}', "");
Error('\p{Jg:   -/a/gaf}');
Error('\P{Jg:   -/a/gaf}');
Expect(1, 1892, '\p{Jg=gaf}', "");
Expect(0, 1892, '\p{^Jg=gaf}', "");
Expect(0, 1892, '\P{Jg=gaf}', "");
Expect(1, 1892, '\P{^Jg=gaf}', "");
Expect(0, 1893, '\p{Jg=gaf}', "");
Expect(1, 1893, '\p{^Jg=gaf}', "");
Expect(1, 1893, '\P{Jg=gaf}', "");
Expect(0, 1893, '\P{^Jg=gaf}', "");
Expect(1, 1892, '\p{Jg=	Gaf}', "");
Expect(0, 1892, '\p{^Jg=	Gaf}', "");
Expect(0, 1892, '\P{Jg=	Gaf}', "");
Expect(1, 1892, '\P{^Jg=	Gaf}', "");
Expect(0, 1893, '\p{Jg=	Gaf}', "");
Expect(1, 1893, '\p{^Jg=	Gaf}', "");
Expect(1, 1893, '\P{Jg=	Gaf}', "");
Expect(0, 1893, '\P{^Jg=	Gaf}', "");
Error('\p{Is_Joining_Group=-	Gaf:=}');
Error('\P{Is_Joining_Group=-	Gaf:=}');
Expect(1, 1892, '\p{Is_Joining_Group=gaf}', "");
Expect(0, 1892, '\p{^Is_Joining_Group=gaf}', "");
Expect(0, 1892, '\P{Is_Joining_Group=gaf}', "");
Expect(1, 1892, '\P{^Is_Joining_Group=gaf}', "");
Expect(0, 1893, '\p{Is_Joining_Group=gaf}', "");
Expect(1, 1893, '\p{^Is_Joining_Group=gaf}', "");
Expect(1, 1893, '\P{Is_Joining_Group=gaf}', "");
Expect(0, 1893, '\P{^Is_Joining_Group=gaf}', "");
Expect(1, 1892, '\p{Is_Joining_Group=  Gaf}', "");
Expect(0, 1892, '\p{^Is_Joining_Group=  Gaf}', "");
Expect(0, 1892, '\P{Is_Joining_Group=  Gaf}', "");
Expect(1, 1892, '\P{^Is_Joining_Group=  Gaf}', "");
Expect(0, 1893, '\p{Is_Joining_Group=  Gaf}', "");
Expect(1, 1893, '\p{^Is_Joining_Group=  Gaf}', "");
Expect(1, 1893, '\P{Is_Joining_Group=  Gaf}', "");
Expect(0, 1893, '\P{^Is_Joining_Group=  Gaf}', "");
Error('\p{Is_Jg=	:=Gaf}');
Error('\P{Is_Jg=	:=Gaf}');
Expect(1, 1892, '\p{Is_Jg=gaf}', "");
Expect(0, 1892, '\p{^Is_Jg=gaf}', "");
Expect(0, 1892, '\P{Is_Jg=gaf}', "");
Expect(1, 1892, '\P{^Is_Jg=gaf}', "");
Expect(0, 1893, '\p{Is_Jg=gaf}', "");
Expect(1, 1893, '\p{^Is_Jg=gaf}', "");
Expect(1, 1893, '\P{Is_Jg=gaf}', "");
Expect(0, 1893, '\P{^Is_Jg=gaf}', "");
Expect(1, 1892, '\p{Is_Jg= gaf}', "");
Expect(0, 1892, '\p{^Is_Jg= gaf}', "");
Expect(0, 1892, '\P{Is_Jg= gaf}', "");
Expect(1, 1892, '\P{^Is_Jg= gaf}', "");
Expect(0, 1893, '\p{Is_Jg= gaf}', "");
Expect(1, 1893, '\p{^Is_Jg= gaf}', "");
Expect(1, 1893, '\P{Is_Jg= gaf}', "");
Expect(0, 1893, '\P{^Is_Jg= gaf}', "");
Error('\p{Joining_Group=_	WAW/a/}');
Error('\P{Joining_Group=_	WAW/a/}');
Expect(1, 2219, '\p{Joining_Group:	waw}', "");
Expect(0, 2219, '\p{^Joining_Group:	waw}', "");
Expect(0, 2219, '\P{Joining_Group:	waw}', "");
Expect(1, 2219, '\P{^Joining_Group:	waw}', "");
Expect(0, 2220, '\p{Joining_Group:	waw}', "");
Expect(1, 2220, '\p{^Joining_Group:	waw}', "");
Expect(1, 2220, '\P{Joining_Group:	waw}', "");
Expect(0, 2220, '\P{^Joining_Group:	waw}', "");
Expect(1, 2219, '\p{Joining_Group=	Waw}', "");
Expect(0, 2219, '\p{^Joining_Group=	Waw}', "");
Expect(0, 2219, '\P{Joining_Group=	Waw}', "");
Expect(1, 2219, '\P{^Joining_Group=	Waw}', "");
Expect(0, 2220, '\p{Joining_Group=	Waw}', "");
Expect(1, 2220, '\p{^Joining_Group=	Waw}', "");
Expect(1, 2220, '\P{Joining_Group=	Waw}', "");
Expect(0, 2220, '\P{^Joining_Group=	Waw}', "");
Error('\p{Jg=:=__Waw}');
Error('\P{Jg=:=__Waw}');
Expect(1, 2219, '\p{Jg=waw}', "");
Expect(0, 2219, '\p{^Jg=waw}', "");
Expect(0, 2219, '\P{Jg=waw}', "");
Expect(1, 2219, '\P{^Jg=waw}', "");
Expect(0, 2220, '\p{Jg=waw}', "");
Expect(1, 2220, '\p{^Jg=waw}', "");
Expect(1, 2220, '\P{Jg=waw}', "");
Expect(0, 2220, '\P{^Jg=waw}', "");
Expect(1, 2219, '\p{Jg=	_waw}', "");
Expect(0, 2219, '\p{^Jg=	_waw}', "");
Expect(0, 2219, '\P{Jg=	_waw}', "");
Expect(1, 2219, '\P{^Jg=	_waw}', "");
Expect(0, 2220, '\p{Jg=	_waw}', "");
Expect(1, 2220, '\p{^Jg=	_waw}', "");
Expect(1, 2220, '\P{Jg=	_waw}', "");
Expect(0, 2220, '\P{^Jg=	_waw}', "");
Error('\p{Is_Joining_Group=/a/ 	waw}');
Error('\P{Is_Joining_Group=/a/ 	waw}');
Expect(1, 2219, '\p{Is_Joining_Group:   waw}', "");
Expect(0, 2219, '\p{^Is_Joining_Group:   waw}', "");
Expect(0, 2219, '\P{Is_Joining_Group:   waw}', "");
Expect(1, 2219, '\P{^Is_Joining_Group:   waw}', "");
Expect(0, 2220, '\p{Is_Joining_Group:   waw}', "");
Expect(1, 2220, '\p{^Is_Joining_Group:   waw}', "");
Expect(1, 2220, '\P{Is_Joining_Group:   waw}', "");
Expect(0, 2220, '\P{^Is_Joining_Group:   waw}', "");
Expect(1, 2219, '\p{Is_Joining_Group=__Waw}', "");
Expect(0, 2219, '\p{^Is_Joining_Group=__Waw}', "");
Expect(0, 2219, '\P{Is_Joining_Group=__Waw}', "");
Expect(1, 2219, '\P{^Is_Joining_Group=__Waw}', "");
Expect(0, 2220, '\p{Is_Joining_Group=__Waw}', "");
Expect(1, 2220, '\p{^Is_Joining_Group=__Waw}', "");
Expect(1, 2220, '\P{Is_Joining_Group=__Waw}', "");
Expect(0, 2220, '\P{^Is_Joining_Group=__Waw}', "");
Error('\p{Is_Jg=	/a/Waw}');
Error('\P{Is_Jg=	/a/Waw}');
Expect(1, 2219, '\p{Is_Jg=waw}', "");
Expect(0, 2219, '\p{^Is_Jg=waw}', "");
Expect(0, 2219, '\P{Is_Jg=waw}', "");
Expect(1, 2219, '\P{^Is_Jg=waw}', "");
Expect(0, 2220, '\p{Is_Jg=waw}', "");
Expect(1, 2220, '\p{^Is_Jg=waw}', "");
Expect(1, 2220, '\P{Is_Jg=waw}', "");
Expect(0, 2220, '\P{^Is_Jg=waw}', "");
Expect(1, 2219, '\p{Is_Jg: -	waw}', "");
Expect(0, 2219, '\p{^Is_Jg: -	waw}', "");
Expect(0, 2219, '\P{Is_Jg: -	waw}', "");
Expect(1, 2219, '\P{^Is_Jg: -	waw}', "");
Expect(0, 2220, '\p{Is_Jg: -	waw}', "");
Expect(1, 2220, '\p{^Is_Jg: -	waw}', "");
Expect(1, 2220, '\P{Is_Jg: -	waw}', "");
Expect(0, 2220, '\P{^Is_Jg: -	waw}', "");
Error('\p{Joining_Group=:=Yudh}');
Error('\P{Joining_Group=:=Yudh}');
Expect(1, 1821, '\p{Joining_Group=yudh}', "");
Expect(0, 1821, '\p{^Joining_Group=yudh}', "");
Expect(0, 1821, '\P{Joining_Group=yudh}', "");
Expect(1, 1821, '\P{^Joining_Group=yudh}', "");
Expect(0, 1822, '\p{Joining_Group=yudh}', "");
Expect(1, 1822, '\p{^Joining_Group=yudh}', "");
Expect(1, 1822, '\P{Joining_Group=yudh}', "");
Expect(0, 1822, '\P{^Joining_Group=yudh}', "");
Expect(1, 1821, '\p{Joining_Group=_	Yudh}', "");
Expect(0, 1821, '\p{^Joining_Group=_	Yudh}', "");
Expect(0, 1821, '\P{Joining_Group=_	Yudh}', "");
Expect(1, 1821, '\P{^Joining_Group=_	Yudh}', "");
Expect(0, 1822, '\p{Joining_Group=_	Yudh}', "");
Expect(1, 1822, '\p{^Joining_Group=_	Yudh}', "");
Expect(1, 1822, '\P{Joining_Group=_	Yudh}', "");
Expect(0, 1822, '\P{^Joining_Group=_	Yudh}', "");
Error('\p{Jg=_Yudh/a/}');
Error('\P{Jg=_Yudh/a/}');
Expect(1, 1821, '\p{Jg=yudh}', "");
Expect(0, 1821, '\p{^Jg=yudh}', "");
Expect(0, 1821, '\P{Jg=yudh}', "");
Expect(1, 1821, '\P{^Jg=yudh}', "");
Expect(0, 1822, '\p{Jg=yudh}', "");
Expect(1, 1822, '\p{^Jg=yudh}', "");
Expect(1, 1822, '\P{Jg=yudh}', "");
Expect(0, 1822, '\P{^Jg=yudh}', "");
Expect(1, 1821, '\p{Jg=		Yudh}', "");
Expect(0, 1821, '\p{^Jg=		Yudh}', "");
Expect(0, 1821, '\P{Jg=		Yudh}', "");
Expect(1, 1821, '\P{^Jg=		Yudh}', "");
Expect(0, 1822, '\p{Jg=		Yudh}', "");
Expect(1, 1822, '\p{^Jg=		Yudh}', "");
Expect(1, 1822, '\P{Jg=		Yudh}', "");
Expect(0, 1822, '\P{^Jg=		Yudh}', "");
Error('\p{Is_Joining_Group=-/a/Yudh}');
Error('\P{Is_Joining_Group=-/a/Yudh}');
Expect(1, 1821, '\p{Is_Joining_Group=yudh}', "");
Expect(0, 1821, '\p{^Is_Joining_Group=yudh}', "");
Expect(0, 1821, '\P{Is_Joining_Group=yudh}', "");
Expect(1, 1821, '\P{^Is_Joining_Group=yudh}', "");
Expect(0, 1822, '\p{Is_Joining_Group=yudh}', "");
Expect(1, 1822, '\p{^Is_Joining_Group=yudh}', "");
Expect(1, 1822, '\P{Is_Joining_Group=yudh}', "");
Expect(0, 1822, '\P{^Is_Joining_Group=yudh}', "");
Expect(1, 1821, '\p{Is_Joining_Group=	 YUDH}', "");
Expect(0, 1821, '\p{^Is_Joining_Group=	 YUDH}', "");
Expect(0, 1821, '\P{Is_Joining_Group=	 YUDH}', "");
Expect(1, 1821, '\P{^Is_Joining_Group=	 YUDH}', "");
Expect(0, 1822, '\p{Is_Joining_Group=	 YUDH}', "");
Expect(1, 1822, '\p{^Is_Joining_Group=	 YUDH}', "");
Expect(1, 1822, '\P{Is_Joining_Group=	 YUDH}', "");
Expect(0, 1822, '\P{^Is_Joining_Group=	 YUDH}', "");
Error('\p{Is_Jg=-_yudh/a/}');
Error('\P{Is_Jg=-_yudh/a/}');
Expect(1, 1821, '\p{Is_Jg=yudh}', "");
Expect(0, 1821, '\p{^Is_Jg=yudh}', "");
Expect(0, 1821, '\P{Is_Jg=yudh}', "");
Expect(1, 1821, '\P{^Is_Jg=yudh}', "");
Expect(0, 1822, '\p{Is_Jg=yudh}', "");
Expect(1, 1822, '\p{^Is_Jg=yudh}', "");
Expect(1, 1822, '\P{Is_Jg=yudh}', "");
Expect(0, 1822, '\P{^Is_Jg=yudh}', "");
Expect(1, 1821, '\p{Is_Jg:	-Yudh}', "");
Expect(0, 1821, '\p{^Is_Jg:	-Yudh}', "");
Expect(0, 1821, '\P{Is_Jg:	-Yudh}', "");
Expect(1, 1821, '\P{^Is_Jg:	-Yudh}', "");
Expect(0, 1822, '\p{Is_Jg:	-Yudh}', "");
Expect(1, 1822, '\p{^Is_Jg:	-Yudh}', "");
Expect(1, 1822, '\P{Is_Jg:	-Yudh}', "");
Expect(0, 1822, '\P{^Is_Jg:	-Yudh}', "");
Error('\p{Bidi_Mirrored=_/a/YES}');
Error('\P{Bidi_Mirrored=_/a/YES}');
Expect(1, 120771, '\p{Bidi_Mirrored=yes}', "");
Expect(0, 120771, '\p{^Bidi_Mirrored=yes}', "");
Expect(0, 120771, '\P{Bidi_Mirrored=yes}', "");
Expect(1, 120771, '\P{^Bidi_Mirrored=yes}', "");
Expect(0, 120772, '\p{Bidi_Mirrored=yes}', "");
Expect(1, 120772, '\p{^Bidi_Mirrored=yes}', "");
Expect(1, 120772, '\P{Bidi_Mirrored=yes}', "");
Expect(0, 120772, '\P{^Bidi_Mirrored=yes}', "");
Expect(1, 120771, '\p{Bidi_Mirrored=-yes}', "");
Expect(0, 120771, '\p{^Bidi_Mirrored=-yes}', "");
Expect(0, 120771, '\P{Bidi_Mirrored=-yes}', "");
Expect(1, 120771, '\P{^Bidi_Mirrored=-yes}', "");
Expect(0, 120772, '\p{Bidi_Mirrored=-yes}', "");
Expect(1, 120772, '\p{^Bidi_Mirrored=-yes}', "");
Expect(1, 120772, '\P{Bidi_Mirrored=-yes}', "");
Expect(0, 120772, '\P{^Bidi_Mirrored=-yes}', "");
Error('\p{Bidi_M=_ y:=}');
Error('\P{Bidi_M=_ y:=}');
Expect(1, 120771, '\p{Bidi_M:   y}', "");
Expect(0, 120771, '\p{^Bidi_M:   y}', "");
Expect(0, 120771, '\P{Bidi_M:   y}', "");
Expect(1, 120771, '\P{^Bidi_M:   y}', "");
Expect(0, 120772, '\p{Bidi_M:   y}', "");
Expect(1, 120772, '\p{^Bidi_M:   y}', "");
Expect(1, 120772, '\P{Bidi_M:   y}', "");
Expect(0, 120772, '\P{^Bidi_M:   y}', "");
Expect(1, 120771, '\p{Bidi_M=	_Y}', "");
Expect(0, 120771, '\p{^Bidi_M=	_Y}', "");
Expect(0, 120771, '\P{Bidi_M=	_Y}', "");
Expect(1, 120771, '\P{^Bidi_M=	_Y}', "");
Expect(0, 120772, '\p{Bidi_M=	_Y}', "");
Expect(1, 120772, '\p{^Bidi_M=	_Y}', "");
Expect(1, 120772, '\P{Bidi_M=	_Y}', "");
Expect(0, 120772, '\P{^Bidi_M=	_Y}', "");
Error('\p{Is_Bidi_Mirrored=:= T}');
Error('\P{Is_Bidi_Mirrored=:= T}');
Expect(1, 120771, '\p{Is_Bidi_Mirrored=t}', "");
Expect(0, 120771, '\p{^Is_Bidi_Mirrored=t}', "");
Expect(0, 120771, '\P{Is_Bidi_Mirrored=t}', "");
Expect(1, 120771, '\P{^Is_Bidi_Mirrored=t}', "");
Expect(0, 120772, '\p{Is_Bidi_Mirrored=t}', "");
Expect(1, 120772, '\p{^Is_Bidi_Mirrored=t}', "");
Expect(1, 120772, '\P{Is_Bidi_Mirrored=t}', "");
Expect(0, 120772, '\P{^Is_Bidi_Mirrored=t}', "");
Expect(1, 120771, '\p{Is_Bidi_Mirrored=  T}', "");
Expect(0, 120771, '\p{^Is_Bidi_Mirrored=  T}', "");
Expect(0, 120771, '\P{Is_Bidi_Mirrored=  T}', "");
Expect(1, 120771, '\P{^Is_Bidi_Mirrored=  T}', "");
Expect(0, 120772, '\p{Is_Bidi_Mirrored=  T}', "");
Expect(1, 120772, '\p{^Is_Bidi_Mirrored=  T}', "");
Expect(1, 120772, '\P{Is_Bidi_Mirrored=  T}', "");
Expect(0, 120772, '\P{^Is_Bidi_Mirrored=  T}', "");
Error('\p{Is_Bidi_M=/a/	 true}');
Error('\P{Is_Bidi_M=/a/	 true}');
Expect(1, 120771, '\p{Is_Bidi_M=true}', "");
Expect(0, 120771, '\p{^Is_Bidi_M=true}', "");
Expect(0, 120771, '\P{Is_Bidi_M=true}', "");
Expect(1, 120771, '\P{^Is_Bidi_M=true}', "");
Expect(0, 120772, '\p{Is_Bidi_M=true}', "");
Expect(1, 120772, '\p{^Is_Bidi_M=true}', "");
Expect(1, 120772, '\P{Is_Bidi_M=true}', "");
Expect(0, 120772, '\P{^Is_Bidi_M=true}', "");
Expect(1, 120771, '\p{Is_Bidi_M=	 true}', "");
Expect(0, 120771, '\p{^Is_Bidi_M=	 true}', "");
Expect(0, 120771, '\P{Is_Bidi_M=	 true}', "");
Expect(1, 120771, '\P{^Is_Bidi_M=	 true}', "");
Expect(0, 120772, '\p{Is_Bidi_M=	 true}', "");
Expect(1, 120772, '\p{^Is_Bidi_M=	 true}', "");
Expect(1, 120772, '\P{Is_Bidi_M=	 true}', "");
Expect(0, 120772, '\P{^Is_Bidi_M=	 true}', "");
Error('\p{Bidi_Mirrored=/a/NO}');
Error('\P{Bidi_Mirrored=/a/NO}');
Expect(1, 120772, '\p{Bidi_Mirrored=no}', "");
Expect(0, 120772, '\p{^Bidi_Mirrored=no}', "");
Expect(0, 120772, '\P{Bidi_Mirrored=no}', "");
Expect(1, 120772, '\P{^Bidi_Mirrored=no}', "");
Expect(0, 120771, '\p{Bidi_Mirrored=no}', "");
Expect(1, 120771, '\p{^Bidi_Mirrored=no}', "");
Expect(1, 120771, '\P{Bidi_Mirrored=no}', "");
Expect(0, 120771, '\P{^Bidi_Mirrored=no}', "");
Expect(1, 120772, '\p{Bidi_Mirrored=_ No}', "");
Expect(0, 120772, '\p{^Bidi_Mirrored=_ No}', "");
Expect(0, 120772, '\P{Bidi_Mirrored=_ No}', "");
Expect(1, 120772, '\P{^Bidi_Mirrored=_ No}', "");
Expect(0, 120771, '\p{Bidi_Mirrored=_ No}', "");
Expect(1, 120771, '\p{^Bidi_Mirrored=_ No}', "");
Expect(1, 120771, '\P{Bidi_Mirrored=_ No}', "");
Expect(0, 120771, '\P{^Bidi_Mirrored=_ No}', "");
Error('\p{Bidi_M:    N/a/}');
Error('\P{Bidi_M:    N/a/}');
Expect(1, 120772, '\p{Bidi_M=n}', "");
Expect(0, 120772, '\p{^Bidi_M=n}', "");
Expect(0, 120772, '\P{Bidi_M=n}', "");
Expect(1, 120772, '\P{^Bidi_M=n}', "");
Expect(0, 120771, '\p{Bidi_M=n}', "");
Expect(1, 120771, '\p{^Bidi_M=n}', "");
Expect(1, 120771, '\P{Bidi_M=n}', "");
Expect(0, 120771, '\P{^Bidi_M=n}', "");
Expect(1, 120772, '\p{Bidi_M=	N}', "");
Expect(0, 120772, '\p{^Bidi_M=	N}', "");
Expect(0, 120772, '\P{Bidi_M=	N}', "");
Expect(1, 120772, '\P{^Bidi_M=	N}', "");
Expect(0, 120771, '\p{Bidi_M=	N}', "");
Expect(1, 120771, '\p{^Bidi_M=	N}', "");
Expect(1, 120771, '\P{Bidi_M=	N}', "");
Expect(0, 120771, '\P{^Bidi_M=	N}', "");
Error('\p{Is_Bidi_Mirrored=	 F/a/}');
Error('\P{Is_Bidi_Mirrored=	 F/a/}');
Expect(1, 120772, '\p{Is_Bidi_Mirrored:f}', "");
Expect(0, 120772, '\p{^Is_Bidi_Mirrored:f}', "");
Expect(0, 120772, '\P{Is_Bidi_Mirrored:f}', "");
Expect(1, 120772, '\P{^Is_Bidi_Mirrored:f}', "");
Expect(0, 120771, '\p{Is_Bidi_Mirrored:f}', "");
Expect(1, 120771, '\p{^Is_Bidi_Mirrored:f}', "");
Expect(1, 120771, '\P{Is_Bidi_Mirrored:f}', "");
Expect(0, 120771, '\P{^Is_Bidi_Mirrored:f}', "");
Expect(1, 120772, '\p{Is_Bidi_Mirrored=	F}', "");
Expect(0, 120772, '\p{^Is_Bidi_Mirrored=	F}', "");
Expect(0, 120772, '\P{Is_Bidi_Mirrored=	F}', "");
Expect(1, 120772, '\P{^Is_Bidi_Mirrored=	F}', "");
Expect(0, 120771, '\p{Is_Bidi_Mirrored=	F}', "");
Expect(1, 120771, '\p{^Is_Bidi_Mirrored=	F}', "");
Expect(1, 120771, '\P{Is_Bidi_Mirrored=	F}', "");
Expect(0, 120771, '\P{^Is_Bidi_Mirrored=	F}', "");
Error('\p{Is_Bidi_M=:=	False}');
Error('\P{Is_Bidi_M=:=	False}');
Expect(1, 120772, '\p{Is_Bidi_M:   false}', "");
Expect(0, 120772, '\p{^Is_Bidi_M:   false}', "");
Expect(0, 120772, '\P{Is_Bidi_M:   false}', "");
Expect(1, 120772, '\P{^Is_Bidi_M:   false}', "");
Expect(0, 120771, '\p{Is_Bidi_M:   false}', "");
Expect(1, 120771, '\p{^Is_Bidi_M:   false}', "");
Expect(1, 120771, '\P{Is_Bidi_M:   false}', "");
Expect(0, 120771, '\P{^Is_Bidi_M:   false}', "");
Expect(1, 120772, '\p{Is_Bidi_M=--False}', "");
Expect(0, 120772, '\p{^Is_Bidi_M=--False}', "");
Expect(0, 120772, '\P{Is_Bidi_M=--False}', "");
Expect(1, 120772, '\P{^Is_Bidi_M=--False}', "");
Expect(0, 120771, '\p{Is_Bidi_M=--False}', "");
Expect(1, 120771, '\p{^Is_Bidi_M=--False}', "");
Expect(1, 120771, '\P{Is_Bidi_M=--False}', "");
Expect(0, 120771, '\P{^Is_Bidi_M=--False}', "");
Error('\p{Radical=:=-Yes}');
Error('\P{Radical=:=-Yes}');
Expect(1, 12245, '\p{Radical=yes}', "");
Expect(0, 12245, '\p{^Radical=yes}', "");
Expect(0, 12245, '\P{Radical=yes}', "");
Expect(1, 12245, '\P{^Radical=yes}', "");
Expect(0, 12246, '\p{Radical=yes}', "");
Expect(1, 12246, '\p{^Radical=yes}', "");
Expect(1, 12246, '\P{Radical=yes}', "");
Expect(0, 12246, '\P{^Radical=yes}', "");
Expect(1, 12245, '\p{Radical=_-YES}', "");
Expect(0, 12245, '\p{^Radical=_-YES}', "");
Expect(0, 12245, '\P{Radical=_-YES}', "");
Expect(1, 12245, '\P{^Radical=_-YES}', "");
Expect(0, 12246, '\p{Radical=_-YES}', "");
Expect(1, 12246, '\p{^Radical=_-YES}', "");
Expect(1, 12246, '\P{Radical=_-YES}', "");
Expect(0, 12246, '\P{^Radical=_-YES}', "");
Error('\p{Is_Radical=:=y}');
Error('\P{Is_Radical=:=y}');
Expect(1, 12245, '\p{Is_Radical=y}', "");
Expect(0, 12245, '\p{^Is_Radical=y}', "");
Expect(0, 12245, '\P{Is_Radical=y}', "");
Expect(1, 12245, '\P{^Is_Radical=y}', "");
Expect(0, 12246, '\p{Is_Radical=y}', "");
Expect(1, 12246, '\p{^Is_Radical=y}', "");
Expect(1, 12246, '\P{Is_Radical=y}', "");
Expect(0, 12246, '\P{^Is_Radical=y}', "");
Expect(1, 12245, '\p{Is_Radical=-Y}', "");
Expect(0, 12245, '\p{^Is_Radical=-Y}', "");
Expect(0, 12245, '\P{Is_Radical=-Y}', "");
Expect(1, 12245, '\P{^Is_Radical=-Y}', "");
Expect(0, 12246, '\p{Is_Radical=-Y}', "");
Expect(1, 12246, '\p{^Is_Radical=-Y}', "");
Expect(1, 12246, '\P{Is_Radical=-Y}', "");
Expect(0, 12246, '\P{^Is_Radical=-Y}', "");
Error('\p{Radical=-	T:=}');
Error('\P{Radical=-	T:=}');
Expect(1, 12245, '\p{Radical=t}', "");
Expect(0, 12245, '\p{^Radical=t}', "");
Expect(0, 12245, '\P{Radical=t}', "");
Expect(1, 12245, '\P{^Radical=t}', "");
Expect(0, 12246, '\p{Radical=t}', "");
Expect(1, 12246, '\p{^Radical=t}', "");
Expect(1, 12246, '\P{Radical=t}', "");
Expect(0, 12246, '\P{^Radical=t}', "");
Expect(1, 12245, '\p{Radical=	-T}', "");
Expect(0, 12245, '\p{^Radical=	-T}', "");
Expect(0, 12245, '\P{Radical=	-T}', "");
Expect(1, 12245, '\P{^Radical=	-T}', "");
Expect(0, 12246, '\p{Radical=	-T}', "");
Expect(1, 12246, '\p{^Radical=	-T}', "");
Expect(1, 12246, '\P{Radical=	-T}', "");
Expect(0, 12246, '\P{^Radical=	-T}', "");
Error('\p{Is_Radical=/a/	TRUE}');
Error('\P{Is_Radical=/a/	TRUE}');
Expect(1, 12245, '\p{Is_Radical=true}', "");
Expect(0, 12245, '\p{^Is_Radical=true}', "");
Expect(0, 12245, '\P{Is_Radical=true}', "");
Expect(1, 12245, '\P{^Is_Radical=true}', "");
Expect(0, 12246, '\p{Is_Radical=true}', "");
Expect(1, 12246, '\p{^Is_Radical=true}', "");
Expect(1, 12246, '\P{Is_Radical=true}', "");
Expect(0, 12246, '\P{^Is_Radical=true}', "");
Expect(1, 12245, '\p{Is_Radical:	 True}', "");
Expect(0, 12245, '\p{^Is_Radical:	 True}', "");
Expect(0, 12245, '\P{Is_Radical:	 True}', "");
Expect(1, 12245, '\P{^Is_Radical:	 True}', "");
Expect(0, 12246, '\p{Is_Radical:	 True}', "");
Expect(1, 12246, '\p{^Is_Radical:	 True}', "");
Expect(1, 12246, '\P{Is_Radical:	 True}', "");
Expect(0, 12246, '\P{^Is_Radical:	 True}', "");
Error('\p{Radical:	 :=no}');
Error('\P{Radical:	 :=no}');
Expect(1, 12246, '\p{Radical=no}', "");
Expect(0, 12246, '\p{^Radical=no}', "");
Expect(0, 12246, '\P{Radical=no}', "");
Expect(1, 12246, '\P{^Radical=no}', "");
Expect(0, 12245, '\p{Radical=no}', "");
Expect(1, 12245, '\p{^Radical=no}', "");
Expect(1, 12245, '\P{Radical=no}', "");
Expect(0, 12245, '\P{^Radical=no}', "");
Expect(1, 12246, '\p{Radical=- no}', "");
Expect(0, 12246, '\p{^Radical=- no}', "");
Expect(0, 12246, '\P{Radical=- no}', "");
Expect(1, 12246, '\P{^Radical=- no}', "");
Expect(0, 12245, '\p{Radical=- no}', "");
Expect(1, 12245, '\p{^Radical=- no}', "");
Expect(1, 12245, '\P{Radical=- no}', "");
Expect(0, 12245, '\P{^Radical=- no}', "");
Error('\p{Is_Radical=_-N/a/}');
Error('\P{Is_Radical=_-N/a/}');
Expect(1, 12246, '\p{Is_Radical=n}', "");
Expect(0, 12246, '\p{^Is_Radical=n}', "");
Expect(0, 12246, '\P{Is_Radical=n}', "");
Expect(1, 12246, '\P{^Is_Radical=n}', "");
Expect(0, 12245, '\p{Is_Radical=n}', "");
Expect(1, 12245, '\p{^Is_Radical=n}', "");
Expect(1, 12245, '\P{Is_Radical=n}', "");
Expect(0, 12245, '\P{^Is_Radical=n}', "");
Expect(1, 12246, '\p{Is_Radical=	-N}', "");
Expect(0, 12246, '\p{^Is_Radical=	-N}', "");
Expect(0, 12246, '\P{Is_Radical=	-N}', "");
Expect(1, 12246, '\P{^Is_Radical=	-N}', "");
Expect(0, 12245, '\p{Is_Radical=	-N}', "");
Expect(1, 12245, '\p{^Is_Radical=	-N}', "");
Expect(1, 12245, '\P{Is_Radical=	-N}', "");
Expect(0, 12245, '\P{^Is_Radical=	-N}', "");
Error('\p{Radical=		f/a/}');
Error('\P{Radical=		f/a/}');
Expect(1, 12246, '\p{Radical=f}', "");
Expect(0, 12246, '\p{^Radical=f}', "");
Expect(0, 12246, '\P{Radical=f}', "");
Expect(1, 12246, '\P{^Radical=f}', "");
Expect(0, 12245, '\p{Radical=f}', "");
Expect(1, 12245, '\p{^Radical=f}', "");
Expect(1, 12245, '\P{Radical=f}', "");
Expect(0, 12245, '\P{^Radical=f}', "");
Expect(1, 12246, '\p{Radical:	 -F}', "");
Expect(0, 12246, '\p{^Radical:	 -F}', "");
Expect(0, 12246, '\P{Radical:	 -F}', "");
Expect(1, 12246, '\P{^Radical:	 -F}', "");
Expect(0, 12245, '\p{Radical:	 -F}', "");
Expect(1, 12245, '\p{^Radical:	 -F}', "");
Expect(1, 12245, '\P{Radical:	 -F}', "");
Expect(0, 12245, '\P{^Radical:	 -F}', "");
Error('\p{Is_Radical:/a/ False}');
Error('\P{Is_Radical:/a/ False}');
Expect(1, 12246, '\p{Is_Radical:   false}', "");
Expect(0, 12246, '\p{^Is_Radical:   false}', "");
Expect(0, 12246, '\P{Is_Radical:   false}', "");
Expect(1, 12246, '\P{^Is_Radical:   false}', "");
Expect(0, 12245, '\p{Is_Radical:   false}', "");
Expect(1, 12245, '\p{^Is_Radical:   false}', "");
Expect(1, 12245, '\P{Is_Radical:   false}', "");
Expect(0, 12245, '\P{^Is_Radical:   false}', "");
Expect(1, 12246, '\p{Is_Radical=	_FALSE}', "");
Expect(0, 12246, '\p{^Is_Radical=	_FALSE}', "");
Expect(0, 12246, '\P{Is_Radical=	_FALSE}', "");
Expect(1, 12246, '\P{^Is_Radical=	_FALSE}', "");
Expect(0, 12245, '\p{Is_Radical=	_FALSE}', "");
Expect(1, 12245, '\p{^Is_Radical=	_FALSE}', "");
Expect(1, 12245, '\P{Is_Radical=	_FALSE}', "");
Expect(0, 12245, '\P{^Is_Radical=	_FALSE}', "");
Error('\p{Present_In=003.2/a/}');
Error('\P{Present_In=003.2/a/}');
Expect(1, 983040, '\p{Present_In=0_0_0_0_0_003.2}', "");
Expect(0, 983040, '\p{^Present_In=0_0_0_0_0_003.2}', "");
Expect(0, 983040, '\P{Present_In=0_0_0_0_0_003.2}', "");
Expect(1, 983040, '\P{^Present_In=0_0_0_0_0_003.2}', "");
Expect(0, 983037, '\p{Present_In=0_0_0_0_0_003.2}', "");
Expect(1, 983037, '\p{^Present_In=0_0_0_0_0_003.2}', "");
Expect(1, 983037, '\P{Present_In=0_0_0_0_0_003.2}', "");
Expect(0, 983037, '\P{^Present_In=0_0_0_0_0_003.2}', "");
Error('\p{In=:=_+00_00_00_003.2}');
Error('\P{In=:=_+00_00_00_003.2}');
Expect(1, 983040, '\p{In=+000000000_3.2}', "");
Expect(0, 983040, '\p{^In=+000000000_3.2}', "");
Expect(0, 983040, '\P{In=+000000000_3.2}', "");
Expect(1, 983040, '\P{^In=+000000000_3.2}', "");
Expect(0, 983037, '\p{In=+000000000_3.2}', "");
Expect(1, 983037, '\p{^In=+000000000_3.2}', "");
Expect(1, 983037, '\P{In=+000000000_3.2}', "");
Expect(0, 983037, '\P{^In=+000000000_3.2}', "");
Error('\p{Is_Present_In=	000003.2:=}');
Error('\P{Is_Present_In=	000003.2:=}');
Expect(1, 983040, '\p{Is_Present_In=00_3.2}', "");
Expect(0, 983040, '\p{^Is_Present_In=00_3.2}', "");
Expect(0, 983040, '\P{Is_Present_In=00_3.2}', "");
Expect(1, 983040, '\P{^Is_Present_In=00_3.2}', "");
Expect(0, 983037, '\p{Is_Present_In=00_3.2}', "");
Expect(1, 983037, '\p{^Is_Present_In=00_3.2}', "");
Expect(1, 983037, '\P{Is_Present_In=00_3.2}', "");
Expect(0, 983037, '\P{^Is_Present_In=00_3.2}', "");
Error('\p{Is_In=/a/	_003.2}');
Error('\P{Is_In=/a/	_003.2}');
Expect(1, 983040, '\p{Is_In=+0000003.2}', "");
Expect(0, 983040, '\p{^Is_In=+0000003.2}', "");
Expect(0, 983040, '\P{Is_In=+0000003.2}', "");
Expect(1, 983040, '\P{^Is_In=+0000003.2}', "");
Expect(0, 983037, '\p{Is_In=+0000003.2}', "");
Expect(1, 983037, '\p{^Is_In=+0000003.2}', "");
Expect(1, 983037, '\P{Is_In=+0000003.2}', "");
Expect(0, 983037, '\P{^Is_In=+0000003.2}', "");
Error('\p{Present_In=_:=000_2.0}');
Error('\P{Present_In=_:=000_2.0}');
Expect(1, 983040, '\p{Present_In=+2.0}', "");
Expect(0, 983040, '\p{^Present_In=+2.0}', "");
Expect(0, 983040, '\P{Present_In=+2.0}', "");
Expect(1, 983040, '\P{^Present_In=+2.0}', "");
Expect(0, 983037, '\p{Present_In=+2.0}', "");
Expect(1, 983037, '\p{^Present_In=+2.0}', "");
Expect(1, 983037, '\P{Present_In=+2.0}', "");
Expect(0, 983037, '\P{^Present_In=+2.0}', "");
Error('\p{In=	/a/2.0}');
Error('\P{In=	/a/2.0}');
Expect(1, 983040, '\p{In=0002.0}', "");
Expect(0, 983040, '\p{^In=0002.0}', "");
Expect(0, 983040, '\P{In=0002.0}', "");
Expect(1, 983040, '\P{^In=0002.0}', "");
Expect(0, 983037, '\p{In=0002.0}', "");
Expect(1, 983037, '\p{^In=0002.0}', "");
Expect(1, 983037, '\P{In=0002.0}', "");
Expect(0, 983037, '\P{^In=0002.0}', "");
Error('\p{Is_Present_In=--+00_2.0/a/}');
Error('\P{Is_Present_In=--+00_2.0/a/}');
Expect(1, 983040, '\p{Is_Present_In:	2.0}', "");
Expect(0, 983040, '\p{^Is_Present_In:	2.0}', "");
Expect(0, 983040, '\P{Is_Present_In:	2.0}', "");
Expect(1, 983040, '\P{^Is_Present_In:	2.0}', "");
Expect(0, 983037, '\p{Is_Present_In:	2.0}', "");
Expect(1, 983037, '\p{^Is_Present_In:	2.0}', "");
Expect(1, 983037, '\P{Is_Present_In:	2.0}', "");
Expect(0, 983037, '\P{^Is_Present_In:	2.0}', "");
Error('\p{Is_In=-/a/0_2.0}');
Error('\P{Is_In=-/a/0_2.0}');
Expect(1, 983040, '\p{Is_In:0_0_0_0_0_0_0_02.0}', "");
Expect(0, 983040, '\p{^Is_In:0_0_0_0_0_0_0_02.0}', "");
Expect(0, 983040, '\P{Is_In:0_0_0_0_0_0_0_02.0}', "");
Expect(1, 983040, '\P{^Is_In:0_0_0_0_0_0_0_02.0}', "");
Expect(0, 983037, '\p{Is_In:0_0_0_0_0_0_0_02.0}', "");
Expect(1, 983037, '\p{^Is_In:0_0_0_0_0_0_0_02.0}', "");
Expect(1, 983037, '\P{Is_In:0_0_0_0_0_0_0_02.0}', "");
Expect(0, 983037, '\P{^Is_In:0_0_0_0_0_0_0_02.0}', "");
Error('\p{Present_In=/a/_1.1}');
Error('\P{Present_In=/a/_1.1}');
Expect(1, 65533, '\p{Present_In=+00000001.1}', "");
Expect(0, 65533, '\p{^Present_In=+00000001.1}', "");
Expect(0, 65533, '\P{Present_In=+00000001.1}', "");
Expect(1, 65533, '\P{^Present_In=+00000001.1}', "");
Expect(0, 65536, '\p{Present_In=+00000001.1}', "");
Expect(1, 65536, '\p{^Present_In=+00000001.1}', "");
Expect(1, 65536, '\P{Present_In=+00000001.1}', "");
Expect(0, 65536, '\P{^Present_In=+00000001.1}', "");
Error('\p{In=/a/		000_000_1.1}');
Error('\P{In=/a/		000_000_1.1}');
Expect(1, 65533, '\p{In=00000000_1.1}', "");
Expect(0, 65533, '\p{^In=00000000_1.1}', "");
Expect(0, 65533, '\P{In=00000000_1.1}', "");
Expect(1, 65533, '\P{^In=00000000_1.1}', "");
Expect(0, 65536, '\p{In=00000000_1.1}', "");
Expect(1, 65536, '\p{^In=00000000_1.1}', "");
Expect(1, 65536, '\P{In=00000000_1.1}', "");
Expect(0, 65536, '\P{^In=00000000_1.1}', "");
Error('\p{Is_Present_In=:=		000000_1.1}');
Error('\P{Is_Present_In=:=		000000_1.1}');
Expect(1, 65533, '\p{Is_Present_In: 01.1}', "");
Expect(0, 65533, '\p{^Is_Present_In: 01.1}', "");
Expect(0, 65533, '\P{Is_Present_In: 01.1}', "");
Expect(1, 65533, '\P{^Is_Present_In: 01.1}', "");
Expect(0, 65536, '\p{Is_Present_In: 01.1}', "");
Expect(1, 65536, '\p{^Is_Present_In: 01.1}', "");
Expect(1, 65536, '\P{Is_Present_In: 01.1}', "");
Expect(0, 65536, '\P{^Is_Present_In: 01.1}', "");
Error('\p{Is_In=/a/-0001.1}');
Error('\P{Is_In=/a/-0001.1}');
Expect(1, 65533, '\p{Is_In=0000000001.1}', "");
Expect(0, 65533, '\p{^Is_In=0000000001.1}', "");
Expect(0, 65533, '\P{Is_In=0000000001.1}', "");
Expect(1, 65533, '\P{^Is_In=0000000001.1}', "");
Expect(0, 65536, '\p{Is_In=0000000001.1}', "");
Expect(1, 65536, '\p{^Is_In=0000000001.1}', "");
Expect(1, 65536, '\P{Is_In=0000000001.1}', "");
Expect(0, 65536, '\P{^Is_In=0000000001.1}', "");
Error('\p{Present_In=_+00005.0/a/}');
Error('\P{Present_In=_+00005.0/a/}');
Expect(1, 983040, '\p{Present_In=+000_000_05.0}', "");
Expect(0, 983040, '\p{^Present_In=+000_000_05.0}', "");
Expect(0, 983040, '\P{Present_In=+000_000_05.0}', "");
Expect(1, 983040, '\P{^Present_In=+000_000_05.0}', "");
Expect(0, 983037, '\p{Present_In=+000_000_05.0}', "");
Expect(1, 983037, '\p{^Present_In=+000_000_05.0}', "");
Expect(1, 983037, '\P{Present_In=+000_000_05.0}', "");
Expect(0, 983037, '\P{^Present_In=+000_000_05.0}', "");
Error('\p{In=	 00_00_5.0:=}');
Error('\P{In=	 00_00_5.0:=}');
Expect(1, 983040, '\p{In=00005.0}', "");
Expect(0, 983040, '\p{^In=00005.0}', "");
Expect(0, 983040, '\P{In=00005.0}', "");
Expect(1, 983040, '\P{^In=00005.0}', "");
Expect(0, 983037, '\p{In=00005.0}', "");
Expect(1, 983037, '\p{^In=00005.0}', "");
Expect(1, 983037, '\P{In=00005.0}', "");
Expect(0, 983037, '\P{^In=00005.0}', "");
Error('\p{Is_Present_In=_ 000005.0/a/}');
Error('\P{Is_Present_In=_ 000005.0/a/}');
Expect(1, 983040, '\p{Is_Present_In:	0000005.0}', "");
Expect(0, 983040, '\p{^Is_Present_In:	0000005.0}', "");
Expect(0, 983040, '\P{Is_Present_In:	0000005.0}', "");
Expect(1, 983040, '\P{^Is_Present_In:	0000005.0}', "");
Expect(0, 983037, '\p{Is_Present_In:	0000005.0}', "");
Expect(1, 983037, '\p{^Is_Present_In:	0000005.0}', "");
Expect(1, 983037, '\P{Is_Present_In:	0000005.0}', "");
Expect(0, 983037, '\P{^Is_Present_In:	0000005.0}', "");
Error('\p{Is_In=:=_ 0_0_05.0}');
Error('\P{Is_In=:=_ 0_0_05.0}');
Expect(1, 983040, '\p{Is_In=0000000005.0}', "");
Expect(0, 983040, '\p{^Is_In=0000000005.0}', "");
Expect(0, 983040, '\P{Is_In=0000000005.0}', "");
Expect(1, 983040, '\P{^Is_In=0000000005.0}', "");
Expect(0, 983037, '\p{Is_In=0000000005.0}', "");
Expect(1, 983037, '\p{^Is_In=0000000005.0}', "");
Expect(1, 983037, '\P{Is_In=0000000005.0}', "");
Expect(0, 983037, '\P{^Is_In=0000000005.0}', "");
Error('\p{Present_In=	/a/000003.0}');
Error('\P{Present_In=	/a/000003.0}');
Expect(1, 983040, '\p{Present_In=003.0}', "");
Expect(0, 983040, '\p{^Present_In=003.0}', "");
Expect(0, 983040, '\P{Present_In=003.0}', "");
Expect(1, 983040, '\P{^Present_In=003.0}', "");
Expect(0, 983037, '\p{Present_In=003.0}', "");
Expect(1, 983037, '\p{^Present_In=003.0}', "");
Expect(1, 983037, '\P{Present_In=003.0}', "");
Expect(0, 983037, '\P{^Present_In=003.0}', "");
Error('\p{In=:=	0003.0}');
Error('\P{In=:=	0003.0}');
Expect(1, 983040, '\p{In=00000_3.0}', "");
Expect(0, 983040, '\p{^In=00000_3.0}', "");
Expect(0, 983040, '\P{In=00000_3.0}', "");
Expect(1, 983040, '\P{^In=00000_3.0}', "");
Expect(0, 983037, '\p{In=00000_3.0}', "");
Expect(1, 983037, '\p{^In=00000_3.0}', "");
Expect(1, 983037, '\P{In=00000_3.0}', "");
Expect(0, 983037, '\P{^In=00000_3.0}', "");
Error('\p{Is_Present_In=:=_+0_0_0_03.0}');
Error('\P{Is_Present_In=:=_+0_0_0_03.0}');
Expect(1, 983040, '\p{Is_Present_In=3.0}', "");
Expect(0, 983040, '\p{^Is_Present_In=3.0}', "");
Expect(0, 983040, '\P{Is_Present_In=3.0}', "");
Expect(1, 983040, '\P{^Is_Present_In=3.0}', "");
Expect(0, 983037, '\p{Is_Present_In=3.0}', "");
Expect(1, 983037, '\p{^Is_Present_In=3.0}', "");
Expect(1, 983037, '\P{Is_Present_In=3.0}', "");
Expect(0, 983037, '\P{^Is_Present_In=3.0}', "");
Error('\p{Is_In:   :=		+000_000_3.0}');
Error('\P{Is_In:   :=		+000_000_3.0}');
Expect(1, 983040, '\p{Is_In=00_00_03.0}', "");
Expect(0, 983040, '\p{^Is_In=00_00_03.0}', "");
Expect(0, 983040, '\P{Is_In=00_00_03.0}', "");
Expect(1, 983040, '\P{^Is_In=00_00_03.0}', "");
Expect(0, 983037, '\p{Is_In=00_00_03.0}', "");
Expect(1, 983037, '\p{^Is_In=00_00_03.0}', "");
Expect(1, 983037, '\P{Is_In=00_00_03.0}', "");
Expect(0, 983037, '\P{^Is_In=00_00_03.0}', "");
Error('\p{Present_In: _/a/06.0}');
Error('\P{Present_In: _/a/06.0}');
Expect(1, 983040, '\p{Present_In=006.0}', "");
Expect(0, 983040, '\p{^Present_In=006.0}', "");
Expect(0, 983040, '\P{Present_In=006.0}', "");
Expect(1, 983040, '\P{^Present_In=006.0}', "");
Expect(0, 983037, '\p{Present_In=006.0}', "");
Expect(1, 983037, '\p{^Present_In=006.0}', "");
Expect(1, 983037, '\P{Present_In=006.0}', "");
Expect(0, 983037, '\P{^Present_In=006.0}', "");
Error('\p{In=_:=+0000_0000_06.0}');
Error('\P{In=_:=+0000_0000_06.0}');
Expect(1, 983040, '\p{In=00_00_6.0}', "");
Expect(0, 983040, '\p{^In=00_00_6.0}', "");
Expect(0, 983040, '\P{In=00_00_6.0}', "");
Expect(1, 983040, '\P{^In=00_00_6.0}', "");
Expect(0, 983037, '\p{In=00_00_6.0}', "");
Expect(1, 983037, '\p{^In=00_00_6.0}', "");
Expect(1, 983037, '\P{In=00_00_6.0}', "");
Expect(0, 983037, '\P{^In=00_00_6.0}', "");
Error('\p{Is_Present_In= :=+00000006.0}');
Error('\P{Is_Present_In= :=+00000006.0}');
Expect(1, 983040, '\p{Is_Present_In:   00000_6.0}', "");
Expect(0, 983040, '\p{^Is_Present_In:   00000_6.0}', "");
Expect(0, 983040, '\P{Is_Present_In:   00000_6.0}', "");
Expect(1, 983040, '\P{^Is_Present_In:   00000_6.0}', "");
Expect(0, 983037, '\p{Is_Present_In:   00000_6.0}', "");
Expect(1, 983037, '\p{^Is_Present_In:   00000_6.0}', "");
Expect(1, 983037, '\P{Is_Present_In:   00000_6.0}', "");
Expect(0, 983037, '\P{^Is_Present_In:   00000_6.0}', "");
Error('\p{Is_In=:=--00000006.0}');
Error('\P{Is_In=:=--00000006.0}');
Expect(1, 983040, '\p{Is_In=0000006.0}', "");
Expect(0, 983040, '\p{^Is_In=0000006.0}', "");
Expect(0, 983040, '\P{Is_In=0000006.0}', "");
Expect(1, 983040, '\P{^Is_In=0000006.0}', "");
Expect(0, 983037, '\p{Is_In=0000006.0}', "");
Expect(1, 983037, '\p{^Is_In=0000006.0}', "");
Expect(1, 983037, '\P{Is_In=0000006.0}', "");
Expect(0, 983037, '\P{^Is_In=0000006.0}', "");
Error('\p{Present_In=--Unassigned/a/}');
Error('\P{Present_In=--Unassigned/a/}');
Expect(1, 983037, '\p{Present_In=unassigned}', "");
Expect(0, 983037, '\p{^Present_In=unassigned}', "");
Expect(0, 983037, '\P{Present_In=unassigned}', "");
Expect(1, 983037, '\P{^Present_In=unassigned}', "");
Expect(0, 983040, '\p{Present_In=unassigned}', "");
Expect(1, 983040, '\p{^Present_In=unassigned}', "");
Expect(1, 983040, '\P{Present_In=unassigned}', "");
Expect(0, 983040, '\P{^Present_In=unassigned}', "");
Expect(1, 983037, '\p{Present_In= _Unassigned}', "");
Expect(0, 983037, '\p{^Present_In= _Unassigned}', "");
Expect(0, 983037, '\P{Present_In= _Unassigned}', "");
Expect(1, 983037, '\P{^Present_In= _Unassigned}', "");
Expect(0, 983040, '\p{Present_In= _Unassigned}', "");
Expect(1, 983040, '\p{^Present_In= _Unassigned}', "");
Expect(1, 983040, '\P{Present_In= _Unassigned}', "");
Expect(0, 983040, '\P{^Present_In= _Unassigned}', "");
Error('\p{In= :=unassigned}');
Error('\P{In= :=unassigned}');
Expect(1, 983037, '\p{In=unassigned}', "");
Expect(0, 983037, '\p{^In=unassigned}', "");
Expect(0, 983037, '\P{In=unassigned}', "");
Expect(1, 983037, '\P{^In=unassigned}', "");
Expect(0, 983040, '\p{In=unassigned}', "");
Expect(1, 983040, '\p{^In=unassigned}', "");
Expect(1, 983040, '\P{In=unassigned}', "");
Expect(0, 983040, '\P{^In=unassigned}', "");
Expect(1, 983037, '\p{In=-	Unassigned}', "");
Expect(0, 983037, '\p{^In=-	Unassigned}', "");
Expect(0, 983037, '\P{In=-	Unassigned}', "");
Expect(1, 983037, '\P{^In=-	Unassigned}', "");
Expect(0, 983040, '\p{In=-	Unassigned}', "");
Expect(1, 983040, '\p{^In=-	Unassigned}', "");
Expect(1, 983040, '\P{In=-	Unassigned}', "");
Expect(0, 983040, '\P{^In=-	Unassigned}', "");
Error('\p{Is_Present_In=_:=Unassigned}');
Error('\P{Is_Present_In=_:=Unassigned}');
Expect(1, 983037, '\p{Is_Present_In=unassigned}', "");
Expect(0, 983037, '\p{^Is_Present_In=unassigned}', "");
Expect(0, 983037, '\P{Is_Present_In=unassigned}', "");
Expect(1, 983037, '\P{^Is_Present_In=unassigned}', "");
Expect(0, 983040, '\p{Is_Present_In=unassigned}', "");
Expect(1, 983040, '\p{^Is_Present_In=unassigned}', "");
Expect(1, 983040, '\P{Is_Present_In=unassigned}', "");
Expect(0, 983040, '\P{^Is_Present_In=unassigned}', "");
Expect(1, 983037, '\p{Is_Present_In=_Unassigned}', "");
Expect(0, 983037, '\p{^Is_Present_In=_Unassigned}', "");
Expect(0, 983037, '\P{Is_Present_In=_Unassigned}', "");
Expect(1, 983037, '\P{^Is_Present_In=_Unassigned}', "");
Expect(0, 983040, '\p{Is_Present_In=_Unassigned}', "");
Expect(1, 983040, '\p{^Is_Present_In=_Unassigned}', "");
Expect(1, 983040, '\P{Is_Present_In=_Unassigned}', "");
Expect(0, 983040, '\P{^Is_Present_In=_Unassigned}', "");
Error('\p{Is_In=_Unassigned/a/}');
Error('\P{Is_In=_Unassigned/a/}');
Expect(1, 983037, '\p{Is_In=unassigned}', "");
Expect(0, 983037, '\p{^Is_In=unassigned}', "");
Expect(0, 983037, '\P{Is_In=unassigned}', "");
Expect(1, 983037, '\P{^Is_In=unassigned}', "");
Expect(0, 983040, '\p{Is_In=unassigned}', "");
Expect(1, 983040, '\p{^Is_In=unassigned}', "");
Expect(1, 983040, '\P{Is_In=unassigned}', "");
Expect(0, 983040, '\P{^Is_In=unassigned}', "");
Expect(1, 983037, '\p{Is_In=	 Unassigned}', "");
Expect(0, 983037, '\p{^Is_In=	 Unassigned}', "");
Expect(0, 983037, '\P{Is_In=	 Unassigned}', "");
Expect(1, 983037, '\P{^Is_In=	 Unassigned}', "");
Expect(0, 983040, '\p{Is_In=	 Unassigned}', "");
Expect(1, 983040, '\p{^Is_In=	 Unassigned}', "");
Expect(1, 983040, '\P{Is_In=	 Unassigned}', "");
Expect(0, 983040, '\P{^Is_In=	 Unassigned}', "");
Error('\p{Present_In=_0000005.1:=}');
Error('\P{Present_In=_0000005.1:=}');
Expect(1, 983040, '\p{Present_In=+0_0_0_0_0_0_0_05.1}', "");
Expect(0, 983040, '\p{^Present_In=+0_0_0_0_0_0_0_05.1}', "");
Expect(0, 983040, '\P{Present_In=+0_0_0_0_0_0_0_05.1}', "");
Expect(1, 983040, '\P{^Present_In=+0_0_0_0_0_0_0_05.1}', "");
Expect(0, 983037, '\p{Present_In=+0_0_0_0_0_0_0_05.1}', "");
Expect(1, 983037, '\p{^Present_In=+0_0_0_0_0_0_0_05.1}', "");
Expect(1, 983037, '\P{Present_In=+0_0_0_0_0_0_0_05.1}', "");
Expect(0, 983037, '\P{^Present_In=+0_0_0_0_0_0_0_05.1}', "");
Error('\p{In=	_05.1:=}');
Error('\P{In=	_05.1:=}');
Expect(1, 983040, '\p{In=+0_0_0_05.1}', "");
Expect(0, 983040, '\p{^In=+0_0_0_05.1}', "");
Expect(0, 983040, '\P{In=+0_0_0_05.1}', "");
Expect(1, 983040, '\P{^In=+0_0_0_05.1}', "");
Expect(0, 983037, '\p{In=+0_0_0_05.1}', "");
Expect(1, 983037, '\p{^In=+0_0_0_05.1}', "");
Expect(1, 983037, '\P{In=+0_0_0_05.1}', "");
Expect(0, 983037, '\P{^In=+0_0_0_05.1}', "");
Error('\p{Is_Present_In=:=		005.1}');
Error('\P{Is_Present_In=:=		005.1}');
Expect(1, 983040, '\p{Is_Present_In=5.1}', "");
Expect(0, 983040, '\p{^Is_Present_In=5.1}', "");
Expect(0, 983040, '\P{Is_Present_In=5.1}', "");
Expect(1, 983040, '\P{^Is_Present_In=5.1}', "");
Expect(0, 983037, '\p{Is_Present_In=5.1}', "");
Expect(1, 983037, '\p{^Is_Present_In=5.1}', "");
Expect(1, 983037, '\P{Is_Present_In=5.1}', "");
Expect(0, 983037, '\P{^Is_Present_In=5.1}', "");
Error('\p{Is_In=/a/	-00_00_5.1}');
Error('\P{Is_In=/a/	-00_00_5.1}');
Expect(1, 983040, '\p{Is_In=0_0_05.1}', "");
Expect(0, 983040, '\p{^Is_In=0_0_05.1}', "");
Expect(0, 983040, '\P{Is_In=0_0_05.1}', "");
Expect(1, 983040, '\P{^Is_In=0_0_05.1}', "");
Expect(0, 983037, '\p{Is_In=0_0_05.1}', "");
Expect(1, 983037, '\p{^Is_In=0_0_05.1}', "");
Expect(1, 983037, '\P{Is_In=0_0_05.1}', "");
Expect(0, 983037, '\P{^Is_In=0_0_05.1}', "");
Error('\p{Present_In= +00004.0/a/}');
Error('\P{Present_In= +00004.0/a/}');
Expect(1, 983040, '\p{Present_In=+0004.0}', "");
Expect(0, 983040, '\p{^Present_In=+0004.0}', "");
Expect(0, 983040, '\P{Present_In=+0004.0}', "");
Expect(1, 983040, '\P{^Present_In=+0004.0}', "");
Expect(0, 983037, '\p{Present_In=+0004.0}', "");
Expect(1, 983037, '\p{^Present_In=+0004.0}', "");
Expect(1, 983037, '\P{Present_In=+0004.0}', "");
Expect(0, 983037, '\P{^Present_In=+0004.0}', "");
Error('\p{In= 000_4.0:=}');
Error('\P{In= 000_4.0:=}');
Expect(1, 983040, '\p{In=+0_0_0_0_0004.0}', "");
Expect(0, 983040, '\p{^In=+0_0_0_0_0004.0}', "");
Expect(0, 983040, '\P{In=+0_0_0_0_0004.0}', "");
Expect(1, 983040, '\P{^In=+0_0_0_0_0004.0}', "");
Expect(0, 983037, '\p{In=+0_0_0_0_0004.0}', "");
Expect(1, 983037, '\p{^In=+0_0_0_0_0004.0}', "");
Expect(1, 983037, '\P{In=+0_0_0_0_0004.0}', "");
Expect(0, 983037, '\P{^In=+0_0_0_0_0004.0}', "");
Error('\p{Is_Present_In=_:=000000004.0}');
Error('\P{Is_Present_In=_:=000000004.0}');
Expect(1, 983040, '\p{Is_Present_In=+000004.0}', "");
Expect(0, 983040, '\p{^Is_Present_In=+000004.0}', "");
Expect(0, 983040, '\P{Is_Present_In=+000004.0}', "");
Expect(1, 983040, '\P{^Is_Present_In=+000004.0}', "");
Expect(0, 983037, '\p{Is_Present_In=+000004.0}', "");
Expect(1, 983037, '\p{^Is_Present_In=+000004.0}', "");
Expect(1, 983037, '\P{Is_Present_In=+000004.0}', "");
Expect(0, 983037, '\P{^Is_Present_In=+000004.0}', "");
Error('\p{Is_In=:= -000004.0}');
Error('\P{Is_In=:= -000004.0}');
Expect(1, 983040, '\p{Is_In=+000000004.0}', "");
Expect(0, 983040, '\p{^Is_In=+000000004.0}', "");
Expect(0, 983040, '\P{Is_In=+000000004.0}', "");
Expect(1, 983040, '\P{^Is_In=+000000004.0}', "");
Expect(0, 983037, '\p{Is_In=+000000004.0}', "");
Expect(1, 983037, '\p{^Is_In=+000000004.0}', "");
Expect(1, 983037, '\P{Is_In=+000000004.0}', "");
Expect(0, 983037, '\P{^Is_In=+000000004.0}', "");
Error('\p{Present_In=_ 000000005.2/a/}');
Error('\P{Present_In=_ 000000005.2/a/}');
Expect(1, 983040, '\p{Present_In=5.2}', "");
Expect(0, 983040, '\p{^Present_In=5.2}', "");
Expect(0, 983040, '\P{Present_In=5.2}', "");
Expect(1, 983040, '\P{^Present_In=5.2}', "");
Expect(0, 983037, '\p{Present_In=5.2}', "");
Expect(1, 983037, '\p{^Present_In=5.2}', "");
Expect(1, 983037, '\P{Present_In=5.2}', "");
Expect(0, 983037, '\P{^Present_In=5.2}', "");
Error('\p{In=_/a/0005.2}');
Error('\P{In=_/a/0005.2}');
Expect(1, 983040, '\p{In=+0000000005.2}', "");
Expect(0, 983040, '\p{^In=+0000000005.2}', "");
Expect(0, 983040, '\P{In=+0000000005.2}', "");
Expect(1, 983040, '\P{^In=+0000000005.2}', "");
Expect(0, 983037, '\p{In=+0000000005.2}', "");
Expect(1, 983037, '\p{^In=+0000000005.2}', "");
Expect(1, 983037, '\P{In=+0000000005.2}', "");
Expect(0, 983037, '\P{^In=+0000000005.2}', "");
Error('\p{Is_Present_In=/a/	-000000_5.2}');
Error('\P{Is_Present_In=/a/	-000000_5.2}');
Expect(1, 983040, '\p{Is_Present_In:	00000005.2}', "");
Expect(0, 983040, '\p{^Is_Present_In:	00000005.2}', "");
Expect(0, 983040, '\P{Is_Present_In:	00000005.2}', "");
Expect(1, 983040, '\P{^Is_Present_In:	00000005.2}', "");
Expect(0, 983037, '\p{Is_Present_In:	00000005.2}', "");
Expect(1, 983037, '\p{^Is_Present_In:	00000005.2}', "");
Expect(1, 983037, '\P{Is_Present_In:	00000005.2}', "");
Expect(0, 983037, '\P{^Is_Present_In:	00000005.2}', "");
Error('\p{Is_In=_+00_00_00_00_05.2/a/}');
Error('\P{Is_In=_+00_00_00_00_05.2/a/}');
Expect(1, 983040, '\p{Is_In=+00_00_00_05.2}', "");
Expect(0, 983040, '\p{^Is_In=+00_00_00_05.2}', "");
Expect(0, 983040, '\P{Is_In=+00_00_00_05.2}', "");
Expect(1, 983040, '\P{^Is_In=+00_00_00_05.2}', "");
Expect(0, 983037, '\p{Is_In=+00_00_00_05.2}', "");
Expect(1, 983037, '\p{^Is_In=+00_00_00_05.2}', "");
Expect(1, 983037, '\P{Is_In=+00_00_00_05.2}', "");
Expect(0, 983037, '\P{^Is_In=+00_00_00_05.2}', "");
Error('\p{Present_In:	 	+0003.1/a/}');
Error('\P{Present_In:	 	+0003.1/a/}');
Expect(1, 983040, '\p{Present_In=0000003.1}', "");
Expect(0, 983040, '\p{^Present_In=0000003.1}', "");
Expect(0, 983040, '\P{Present_In=0000003.1}', "");
Expect(1, 983040, '\P{^Present_In=0000003.1}', "");
Expect(0, 983037, '\p{Present_In=0000003.1}', "");
Expect(1, 983037, '\p{^Present_In=0000003.1}', "");
Expect(1, 983037, '\P{Present_In=0000003.1}', "");
Expect(0, 983037, '\P{^Present_In=0000003.1}', "");
Error('\p{In=/a/	0000000003.1}');
Error('\P{In=/a/	0000000003.1}');
Expect(1, 983040, '\p{In=+0000000003.1}', "");
Expect(0, 983040, '\p{^In=+0000000003.1}', "");
Expect(0, 983040, '\P{In=+0000000003.1}', "");
Expect(1, 983040, '\P{^In=+0000000003.1}', "");
Expect(0, 983037, '\p{In=+0000000003.1}', "");
Expect(1, 983037, '\p{^In=+0000000003.1}', "");
Expect(1, 983037, '\P{In=+0000000003.1}', "");
Expect(0, 983037, '\P{^In=+0000000003.1}', "");
Error('\p{Is_Present_In=:=_	0_0_0_0_003.1}');
Error('\P{Is_Present_In=:=_	0_0_0_0_003.1}');
Expect(1, 983040, '\p{Is_Present_In=0_3.1}', "");
Expect(0, 983040, '\p{^Is_Present_In=0_3.1}', "");
Expect(0, 983040, '\P{Is_Present_In=0_3.1}', "");
Expect(1, 983040, '\P{^Is_Present_In=0_3.1}', "");
Expect(0, 983037, '\p{Is_Present_In=0_3.1}', "");
Expect(1, 983037, '\p{^Is_Present_In=0_3.1}', "");
Expect(1, 983037, '\P{Is_Present_In=0_3.1}', "");
Expect(0, 983037, '\P{^Is_Present_In=0_3.1}', "");
Error('\p{Is_In=_	000003.1/a/}');
Error('\P{Is_In=_	000003.1/a/}');
Expect(1, 983040, '\p{Is_In: 0000000003.1}', "");
Expect(0, 983040, '\p{^Is_In: 0000000003.1}', "");
Expect(0, 983040, '\P{Is_In: 0000000003.1}', "");
Expect(1, 983040, '\P{^Is_In: 0000000003.1}', "");
Expect(0, 983037, '\p{Is_In: 0000000003.1}', "");
Expect(1, 983037, '\p{^Is_In: 0000000003.1}', "");
Expect(1, 983037, '\P{Is_In: 0000000003.1}', "");
Expect(0, 983037, '\P{^Is_In: 0000000003.1}', "");
Error('\p{Present_In=:=_+00_6.1}');
Error('\P{Present_In=:=_+00_6.1}');
Expect(1, 983040, '\p{Present_In=0000006.1}', "");
Expect(0, 983040, '\p{^Present_In=0000006.1}', "");
Expect(0, 983040, '\P{Present_In=0000006.1}', "");
Expect(1, 983040, '\P{^Present_In=0000006.1}', "");
Expect(0, 983037, '\p{Present_In=0000006.1}', "");
Expect(1, 983037, '\p{^Present_In=0000006.1}', "");
Expect(1, 983037, '\P{Present_In=0000006.1}', "");
Expect(0, 983037, '\P{^Present_In=0000006.1}', "");
Error('\p{In=/a/-+0_0_0_0_006.1}');
Error('\P{In=/a/-+0_0_0_0_006.1}');
Expect(1, 983040, '\p{In=+006.1}', "");
Expect(0, 983040, '\p{^In=+006.1}', "");
Expect(0, 983040, '\P{In=+006.1}', "");
Expect(1, 983040, '\P{^In=+006.1}', "");
Expect(0, 983037, '\p{In=+006.1}', "");
Expect(1, 983037, '\p{^In=+006.1}', "");
Expect(1, 983037, '\P{In=+006.1}', "");
Expect(0, 983037, '\P{^In=+006.1}', "");
Error('\p{Is_Present_In=/a/ +06.1}');
Error('\P{Is_Present_In=/a/ +06.1}');
Expect(1, 983040, '\p{Is_Present_In=0_6.1}', "");
Expect(0, 983040, '\p{^Is_Present_In=0_6.1}', "");
Expect(0, 983040, '\P{Is_Present_In=0_6.1}', "");
Expect(1, 983040, '\P{^Is_Present_In=0_6.1}', "");
Expect(0, 983037, '\p{Is_Present_In=0_6.1}', "");
Expect(1, 983037, '\p{^Is_Present_In=0_6.1}', "");
Expect(1, 983037, '\P{Is_Present_In=0_6.1}', "");
Expect(0, 983037, '\P{^Is_Present_In=0_6.1}', "");
Error('\p{Is_In=-+0_0_0_0_0_0_006.1:=}');
Error('\P{Is_In=-+0_0_0_0_0_0_006.1:=}');
Expect(1, 983040, '\p{Is_In=6.1}', "");
Expect(0, 983040, '\p{^Is_In=6.1}', "");
Expect(0, 983040, '\P{Is_In=6.1}', "");
Expect(1, 983040, '\P{^Is_In=6.1}', "");
Expect(0, 983037, '\p{Is_In=6.1}', "");
Expect(1, 983037, '\p{^Is_In=6.1}', "");
Expect(1, 983037, '\P{Is_In=6.1}', "");
Expect(0, 983037, '\P{^Is_In=6.1}', "");
Error('\p{Present_In:   :=-2.1}');
Error('\P{Present_In:   :=-2.1}');
Expect(1, 983040, '\p{Present_In:   002.1}', "");
Expect(0, 983040, '\p{^Present_In:   002.1}', "");
Expect(0, 983040, '\P{Present_In:   002.1}', "");
Expect(1, 983040, '\P{^Present_In:   002.1}', "");
Expect(0, 983037, '\p{Present_In:   002.1}', "");
Expect(1, 983037, '\p{^Present_In:   002.1}', "");
Expect(1, 983037, '\P{Present_In:   002.1}', "");
Expect(0, 983037, '\P{^Present_In:   002.1}', "");
Error('\p{In:	- 0002.1:=}');
Error('\P{In:	- 0002.1:=}');
Expect(1, 983040, '\p{In=0_0_0_0_0_0_0_0_02.1}', "");
Expect(0, 983040, '\p{^In=0_0_0_0_0_0_0_0_02.1}', "");
Expect(0, 983040, '\P{In=0_0_0_0_0_0_0_0_02.1}', "");
Expect(1, 983040, '\P{^In=0_0_0_0_0_0_0_0_02.1}', "");
Expect(0, 983037, '\p{In=0_0_0_0_0_0_0_0_02.1}', "");
Expect(1, 983037, '\p{^In=0_0_0_0_0_0_0_0_02.1}', "");
Expect(1, 983037, '\P{In=0_0_0_0_0_0_0_0_02.1}', "");
Expect(0, 983037, '\P{^In=0_0_0_0_0_0_0_0_02.1}', "");
Error('\p{Is_Present_In=:=_	00002.1}');
Error('\P{Is_Present_In=:=_	00002.1}');
Expect(1, 983040, '\p{Is_Present_In=+0002.1}', "");
Expect(0, 983040, '\p{^Is_Present_In=+0002.1}', "");
Expect(0, 983040, '\P{Is_Present_In=+0002.1}', "");
Expect(1, 983040, '\P{^Is_Present_In=+0002.1}', "");
Expect(0, 983037, '\p{Is_Present_In=+0002.1}', "");
Expect(1, 983037, '\p{^Is_Present_In=+0002.1}', "");
Expect(1, 983037, '\P{Is_Present_In=+0002.1}', "");
Expect(0, 983037, '\P{^Is_Present_In=+0002.1}', "");
Error('\p{Is_In=	_0_0_02.1/a/}');
Error('\P{Is_In=	_0_0_02.1/a/}');
Expect(1, 983040, '\p{Is_In=00_00_00_02.1}', "");
Expect(0, 983040, '\p{^Is_In=00_00_00_02.1}', "");
Expect(0, 983040, '\P{Is_In=00_00_00_02.1}', "");
Expect(1, 983040, '\P{^Is_In=00_00_00_02.1}', "");
Expect(0, 983037, '\p{Is_In=00_00_00_02.1}', "");
Expect(1, 983037, '\p{^Is_In=00_00_00_02.1}', "");
Expect(1, 983037, '\P{Is_In=00_00_00_02.1}', "");
Expect(0, 983037, '\P{^Is_In=00_00_00_02.1}', "");
Error('\p{Present_In=/a/		00000004.1}');
Error('\P{Present_In=/a/		00000004.1}');
Expect(1, 983040, '\p{Present_In=000000004.1}', "");
Expect(0, 983040, '\p{^Present_In=000000004.1}', "");
Expect(0, 983040, '\P{Present_In=000000004.1}', "");
Expect(1, 983040, '\P{^Present_In=000000004.1}', "");
Expect(0, 983037, '\p{Present_In=000000004.1}', "");
Expect(1, 983037, '\p{^Present_In=000000004.1}', "");
Expect(1, 983037, '\P{Present_In=000000004.1}', "");
Expect(0, 983037, '\P{^Present_In=000000004.1}', "");
Error('\p{In=:= 004.1}');
Error('\P{In=:= 004.1}');
Expect(1, 983040, '\p{In=000004.1}', "");
Expect(0, 983040, '\p{^In=000004.1}', "");
Expect(0, 983040, '\P{In=000004.1}', "");
Expect(1, 983040, '\P{^In=000004.1}', "");
Expect(0, 983037, '\p{In=000004.1}', "");
Expect(1, 983037, '\p{^In=000004.1}', "");
Expect(1, 983037, '\P{In=000004.1}', "");
Expect(0, 983037, '\P{^In=000004.1}', "");
Error('\p{Is_Present_In=_-0000000004.1:=}');
Error('\P{Is_Present_In=_-0000000004.1:=}');
Expect(1, 983040, '\p{Is_Present_In=0_4.1}', "");
Expect(0, 983040, '\p{^Is_Present_In=0_4.1}', "");
Expect(0, 983040, '\P{Is_Present_In=0_4.1}', "");
Expect(1, 983040, '\P{^Is_Present_In=0_4.1}', "");
Expect(0, 983037, '\p{Is_Present_In=0_4.1}', "");
Expect(1, 983037, '\p{^Is_Present_In=0_4.1}', "");
Expect(1, 983037, '\P{Is_Present_In=0_4.1}', "");
Expect(0, 983037, '\P{^Is_Present_In=0_4.1}', "");
Error('\p{Is_In=:=_	04.1}');
Error('\P{Is_In=:=_	04.1}');
Expect(1, 983040, '\p{Is_In=0_0_0_0_0_004.1}', "");
Expect(0, 983040, '\p{^Is_In=0_0_0_0_0_004.1}', "");
Expect(0, 983040, '\P{Is_In=0_0_0_0_0_004.1}', "");
Expect(1, 983040, '\P{^Is_In=0_0_0_0_0_004.1}', "");
Expect(0, 983037, '\p{Is_In=0_0_0_0_0_004.1}', "");
Expect(1, 983037, '\p{^Is_In=0_0_0_0_0_004.1}', "");
Expect(1, 983037, '\P{Is_In=0_0_0_0_0_004.1}', "");
Expect(0, 983037, '\P{^Is_In=0_0_0_0_0_004.1}', "");
Error('\p{STerm=-/a/Yes}');
Error('\P{STerm=-/a/Yes}');
Expect(1, 70086, '\p{STerm=yes}', "");
Expect(0, 70086, '\p{^STerm=yes}', "");
Expect(0, 70086, '\P{STerm=yes}', "");
Expect(1, 70086, '\P{^STerm=yes}', "");
Expect(0, 70087, '\p{STerm=yes}', "");
Expect(1, 70087, '\p{^STerm=yes}', "");
Expect(1, 70087, '\P{STerm=yes}', "");
Expect(0, 70087, '\P{^STerm=yes}', "");
Expect(1, 70086, '\p{STerm=-YES}', "");
Expect(0, 70086, '\p{^STerm=-YES}', "");
Expect(0, 70086, '\P{STerm=-YES}', "");
Expect(1, 70086, '\P{^STerm=-YES}', "");
Expect(0, 70087, '\p{STerm=-YES}', "");
Expect(1, 70087, '\p{^STerm=-YES}', "");
Expect(1, 70087, '\P{STerm=-YES}', "");
Expect(0, 70087, '\P{^STerm=-YES}', "");
Error('\p{Is_STerm=_/a/Y}');
Error('\P{Is_STerm=_/a/Y}');
Expect(1, 70086, '\p{Is_STerm=y}', "");
Expect(0, 70086, '\p{^Is_STerm=y}', "");
Expect(0, 70086, '\P{Is_STerm=y}', "");
Expect(1, 70086, '\P{^Is_STerm=y}', "");
Expect(0, 70087, '\p{Is_STerm=y}', "");
Expect(1, 70087, '\p{^Is_STerm=y}', "");
Expect(1, 70087, '\P{Is_STerm=y}', "");
Expect(0, 70087, '\P{^Is_STerm=y}', "");
Expect(1, 70086, '\p{Is_STerm= -Y}', "");
Expect(0, 70086, '\p{^Is_STerm= -Y}', "");
Expect(0, 70086, '\P{Is_STerm= -Y}', "");
Expect(1, 70086, '\P{^Is_STerm= -Y}', "");
Expect(0, 70087, '\p{Is_STerm= -Y}', "");
Expect(1, 70087, '\p{^Is_STerm= -Y}', "");
Expect(1, 70087, '\P{Is_STerm= -Y}', "");
Expect(0, 70087, '\P{^Is_STerm= -Y}', "");
Error('\p{STerm=- t:=}');
Error('\P{STerm=- t:=}');
Expect(1, 70086, '\p{STerm=t}', "");
Expect(0, 70086, '\p{^STerm=t}', "");
Expect(0, 70086, '\P{STerm=t}', "");
Expect(1, 70086, '\P{^STerm=t}', "");
Expect(0, 70087, '\p{STerm=t}', "");
Expect(1, 70087, '\p{^STerm=t}', "");
Expect(1, 70087, '\P{STerm=t}', "");
Expect(0, 70087, '\P{^STerm=t}', "");
Expect(1, 70086, '\p{STerm=__t}', "");
Expect(0, 70086, '\p{^STerm=__t}', "");
Expect(0, 70086, '\P{STerm=__t}', "");
Expect(1, 70086, '\P{^STerm=__t}', "");
Expect(0, 70087, '\p{STerm=__t}', "");
Expect(1, 70087, '\p{^STerm=__t}', "");
Expect(1, 70087, '\P{STerm=__t}', "");
Expect(0, 70087, '\P{^STerm=__t}', "");
Error('\p{Is_STerm=_:=true}');
Error('\P{Is_STerm=_:=true}');
Expect(1, 70086, '\p{Is_STerm=true}', "");
Expect(0, 70086, '\p{^Is_STerm=true}', "");
Expect(0, 70086, '\P{Is_STerm=true}', "");
Expect(1, 70086, '\P{^Is_STerm=true}', "");
Expect(0, 70087, '\p{Is_STerm=true}', "");
Expect(1, 70087, '\p{^Is_STerm=true}', "");
Expect(1, 70087, '\P{Is_STerm=true}', "");
Expect(0, 70087, '\P{^Is_STerm=true}', "");
Expect(1, 70086, '\p{Is_STerm=	true}', "");
Expect(0, 70086, '\p{^Is_STerm=	true}', "");
Expect(0, 70086, '\P{Is_STerm=	true}', "");
Expect(1, 70086, '\P{^Is_STerm=	true}', "");
Expect(0, 70087, '\p{Is_STerm=	true}', "");
Expect(1, 70087, '\p{^Is_STerm=	true}', "");
Expect(1, 70087, '\P{Is_STerm=	true}', "");
Expect(0, 70087, '\P{^Is_STerm=	true}', "");
Error('\p{STerm=/a/No}');
Error('\P{STerm=/a/No}');
Expect(1, 70087, '\p{STerm:no}', "");
Expect(0, 70087, '\p{^STerm:no}', "");
Expect(0, 70087, '\P{STerm:no}', "");
Expect(1, 70087, '\P{^STerm:no}', "");
Expect(0, 70086, '\p{STerm:no}', "");
Expect(1, 70086, '\p{^STerm:no}', "");
Expect(1, 70086, '\P{STerm:no}', "");
Expect(0, 70086, '\P{^STerm:no}', "");
Expect(1, 70087, '\p{STerm=_NO}', "");
Expect(0, 70087, '\p{^STerm=_NO}', "");
Expect(0, 70087, '\P{STerm=_NO}', "");
Expect(1, 70087, '\P{^STerm=_NO}', "");
Expect(0, 70086, '\p{STerm=_NO}', "");
Expect(1, 70086, '\p{^STerm=_NO}', "");
Expect(1, 70086, '\P{STerm=_NO}', "");
Expect(0, 70086, '\P{^STerm=_NO}', "");
Error('\p{Is_STerm:	 /a/n}');
Error('\P{Is_STerm:	 /a/n}');
Expect(1, 70087, '\p{Is_STerm=n}', "");
Expect(0, 70087, '\p{^Is_STerm=n}', "");
Expect(0, 70087, '\P{Is_STerm=n}', "");
Expect(1, 70087, '\P{^Is_STerm=n}', "");
Expect(0, 70086, '\p{Is_STerm=n}', "");
Expect(1, 70086, '\p{^Is_STerm=n}', "");
Expect(1, 70086, '\P{Is_STerm=n}', "");
Expect(0, 70086, '\P{^Is_STerm=n}', "");
Expect(1, 70087, '\p{Is_STerm=- n}', "");
Expect(0, 70087, '\p{^Is_STerm=- n}', "");
Expect(0, 70087, '\P{Is_STerm=- n}', "");
Expect(1, 70087, '\P{^Is_STerm=- n}', "");
Expect(0, 70086, '\p{Is_STerm=- n}', "");
Expect(1, 70086, '\p{^Is_STerm=- n}', "");
Expect(1, 70086, '\P{Is_STerm=- n}', "");
Expect(0, 70086, '\P{^Is_STerm=- n}', "");
Error('\p{STerm=/a/ -f}');
Error('\P{STerm=/a/ -f}');
Expect(1, 70087, '\p{STerm=f}', "");
Expect(0, 70087, '\p{^STerm=f}', "");
Expect(0, 70087, '\P{STerm=f}', "");
Expect(1, 70087, '\P{^STerm=f}', "");
Expect(0, 70086, '\p{STerm=f}', "");
Expect(1, 70086, '\p{^STerm=f}', "");
Expect(1, 70086, '\P{STerm=f}', "");
Expect(0, 70086, '\P{^STerm=f}', "");
Expect(1, 70087, '\p{STerm=  F}', "");
Expect(0, 70087, '\p{^STerm=  F}', "");
Expect(0, 70087, '\P{STerm=  F}', "");
Expect(1, 70087, '\P{^STerm=  F}', "");
Expect(0, 70086, '\p{STerm=  F}', "");
Expect(1, 70086, '\p{^STerm=  F}', "");
Expect(1, 70086, '\P{STerm=  F}', "");
Expect(0, 70086, '\P{^STerm=  F}', "");
Error('\p{Is_STerm=_:=FALSE}');
Error('\P{Is_STerm=_:=FALSE}');
Expect(1, 70087, '\p{Is_STerm=false}', "");
Expect(0, 70087, '\p{^Is_STerm=false}', "");
Expect(0, 70087, '\P{Is_STerm=false}', "");
Expect(1, 70087, '\P{^Is_STerm=false}', "");
Expect(0, 70086, '\p{Is_STerm=false}', "");
Expect(1, 70086, '\p{^Is_STerm=false}', "");
Expect(1, 70086, '\P{Is_STerm=false}', "");
Expect(0, 70086, '\P{^Is_STerm=false}', "");
Expect(1, 70087, '\p{Is_STerm=	 FALSE}', "");
Expect(0, 70087, '\p{^Is_STerm=	 FALSE}', "");
Expect(0, 70087, '\P{Is_STerm=	 FALSE}', "");
Expect(1, 70087, '\P{^Is_STerm=	 FALSE}', "");
Expect(0, 70086, '\p{Is_STerm=	 FALSE}', "");
Expect(1, 70086, '\p{^Is_STerm=	 FALSE}', "");
Expect(1, 70086, '\P{Is_STerm=	 FALSE}', "");
Expect(0, 70086, '\P{^Is_STerm=	 FALSE}', "");
Error('\p{Extender=--yes:=}');
Error('\P{Extender=--yes:=}');
Expect(1, 65392, '\p{Extender=yes}', "");
Expect(0, 65392, '\p{^Extender=yes}', "");
Expect(0, 65392, '\P{Extender=yes}', "");
Expect(1, 65392, '\P{^Extender=yes}', "");
Expect(0, 65393, '\p{Extender=yes}', "");
Expect(1, 65393, '\p{^Extender=yes}', "");
Expect(1, 65393, '\P{Extender=yes}', "");
Expect(0, 65393, '\P{^Extender=yes}', "");
Expect(1, 65392, '\p{Extender= 	YES}', "");
Expect(0, 65392, '\p{^Extender= 	YES}', "");
Expect(0, 65392, '\P{Extender= 	YES}', "");
Expect(1, 65392, '\P{^Extender= 	YES}', "");
Expect(0, 65393, '\p{Extender= 	YES}', "");
Expect(1, 65393, '\p{^Extender= 	YES}', "");
Expect(1, 65393, '\P{Extender= 	YES}', "");
Expect(0, 65393, '\P{^Extender= 	YES}', "");
Error('\p{Ext=-	Y:=}');
Error('\P{Ext=-	Y:=}');
Expect(1, 65392, '\p{Ext=y}', "");
Expect(0, 65392, '\p{^Ext=y}', "");
Expect(0, 65392, '\P{Ext=y}', "");
Expect(1, 65392, '\P{^Ext=y}', "");
Expect(0, 65393, '\p{Ext=y}', "");
Expect(1, 65393, '\p{^Ext=y}', "");
Expect(1, 65393, '\P{Ext=y}', "");
Expect(0, 65393, '\P{^Ext=y}', "");
Expect(1, 65392, '\p{Ext=-Y}', "");
Expect(0, 65392, '\p{^Ext=-Y}', "");
Expect(0, 65392, '\P{Ext=-Y}', "");
Expect(1, 65392, '\P{^Ext=-Y}', "");
Expect(0, 65393, '\p{Ext=-Y}', "");
Expect(1, 65393, '\p{^Ext=-Y}', "");
Expect(1, 65393, '\P{Ext=-Y}', "");
Expect(0, 65393, '\P{^Ext=-Y}', "");
Error('\p{Is_Extender=T:=}');
Error('\P{Is_Extender=T:=}');
Expect(1, 65392, '\p{Is_Extender=t}', "");
Expect(0, 65392, '\p{^Is_Extender=t}', "");
Expect(0, 65392, '\P{Is_Extender=t}', "");
Expect(1, 65392, '\P{^Is_Extender=t}', "");
Expect(0, 65393, '\p{Is_Extender=t}', "");
Expect(1, 65393, '\p{^Is_Extender=t}', "");
Expect(1, 65393, '\P{Is_Extender=t}', "");
Expect(0, 65393, '\P{^Is_Extender=t}', "");
Expect(1, 65392, '\p{Is_Extender:   --T}', "");
Expect(0, 65392, '\p{^Is_Extender:   --T}', "");
Expect(0, 65392, '\P{Is_Extender:   --T}', "");
Expect(1, 65392, '\P{^Is_Extender:   --T}', "");
Expect(0, 65393, '\p{Is_Extender:   --T}', "");
Expect(1, 65393, '\p{^Is_Extender:   --T}', "");
Expect(1, 65393, '\P{Is_Extender:   --T}', "");
Expect(0, 65393, '\P{^Is_Extender:   --T}', "");
Error('\p{Is_Ext=/a/	True}');
Error('\P{Is_Ext=/a/	True}');
Expect(1, 65392, '\p{Is_Ext: true}', "");
Expect(0, 65392, '\p{^Is_Ext: true}', "");
Expect(0, 65392, '\P{Is_Ext: true}', "");
Expect(1, 65392, '\P{^Is_Ext: true}', "");
Expect(0, 65393, '\p{Is_Ext: true}', "");
Expect(1, 65393, '\p{^Is_Ext: true}', "");
Expect(1, 65393, '\P{Is_Ext: true}', "");
Expect(0, 65393, '\P{^Is_Ext: true}', "");
Expect(1, 65392, '\p{Is_Ext=	TRUE}', "");
Expect(0, 65392, '\p{^Is_Ext=	TRUE}', "");
Expect(0, 65392, '\P{Is_Ext=	TRUE}', "");
Expect(1, 65392, '\P{^Is_Ext=	TRUE}', "");
Expect(0, 65393, '\p{Is_Ext=	TRUE}', "");
Expect(1, 65393, '\p{^Is_Ext=	TRUE}', "");
Expect(1, 65393, '\P{Is_Ext=	TRUE}', "");
Expect(0, 65393, '\P{^Is_Ext=	TRUE}', "");
Error('\p{Extender=--No/a/}');
Error('\P{Extender=--No/a/}');
Expect(1, 65393, '\p{Extender=no}', "");
Expect(0, 65393, '\p{^Extender=no}', "");
Expect(0, 65393, '\P{Extender=no}', "");
Expect(1, 65393, '\P{^Extender=no}', "");
Expect(0, 65392, '\p{Extender=no}', "");
Expect(1, 65392, '\p{^Extender=no}', "");
Expect(1, 65392, '\P{Extender=no}', "");
Expect(0, 65392, '\P{^Extender=no}', "");
Expect(1, 65393, '\p{Extender=	_No}', "");
Expect(0, 65393, '\p{^Extender=	_No}', "");
Expect(0, 65393, '\P{Extender=	_No}', "");
Expect(1, 65393, '\P{^Extender=	_No}', "");
Expect(0, 65392, '\p{Extender=	_No}', "");
Expect(1, 65392, '\p{^Extender=	_No}', "");
Expect(1, 65392, '\P{Extender=	_No}', "");
Expect(0, 65392, '\P{^Extender=	_No}', "");
Error('\p{Ext=:=_ N}');
Error('\P{Ext=:=_ N}');
Expect(1, 65393, '\p{Ext=n}', "");
Expect(0, 65393, '\p{^Ext=n}', "");
Expect(0, 65393, '\P{Ext=n}', "");
Expect(1, 65393, '\P{^Ext=n}', "");
Expect(0, 65392, '\p{Ext=n}', "");
Expect(1, 65392, '\p{^Ext=n}', "");
Expect(1, 65392, '\P{Ext=n}', "");
Expect(0, 65392, '\P{^Ext=n}', "");
Expect(1, 65393, '\p{Ext=-_n}', "");
Expect(0, 65393, '\p{^Ext=-_n}', "");
Expect(0, 65393, '\P{Ext=-_n}', "");
Expect(1, 65393, '\P{^Ext=-_n}', "");
Expect(0, 65392, '\p{Ext=-_n}', "");
Expect(1, 65392, '\p{^Ext=-_n}', "");
Expect(1, 65392, '\P{Ext=-_n}', "");
Expect(0, 65392, '\P{^Ext=-_n}', "");
Error('\p{Is_Extender=	:=F}');
Error('\P{Is_Extender=	:=F}');
Expect(1, 65393, '\p{Is_Extender=f}', "");
Expect(0, 65393, '\p{^Is_Extender=f}', "");
Expect(0, 65393, '\P{Is_Extender=f}', "");
Expect(1, 65393, '\P{^Is_Extender=f}', "");
Expect(0, 65392, '\p{Is_Extender=f}', "");
Expect(1, 65392, '\p{^Is_Extender=f}', "");
Expect(1, 65392, '\P{Is_Extender=f}', "");
Expect(0, 65392, '\P{^Is_Extender=f}', "");
Expect(1, 65393, '\p{Is_Extender=	F}', "");
Expect(0, 65393, '\p{^Is_Extender=	F}', "");
Expect(0, 65393, '\P{Is_Extender=	F}', "");
Expect(1, 65393, '\P{^Is_Extender=	F}', "");
Expect(0, 65392, '\p{Is_Extender=	F}', "");
Expect(1, 65392, '\p{^Is_Extender=	F}', "");
Expect(1, 65392, '\P{Is_Extender=	F}', "");
Expect(0, 65392, '\P{^Is_Extender=	F}', "");
Error('\p{Is_Ext= -FALSE:=}');
Error('\P{Is_Ext= -FALSE:=}');
Expect(1, 65393, '\p{Is_Ext=false}', "");
Expect(0, 65393, '\p{^Is_Ext=false}', "");
Expect(0, 65393, '\P{Is_Ext=false}', "");
Expect(1, 65393, '\P{^Is_Ext=false}', "");
Expect(0, 65392, '\p{Is_Ext=false}', "");
Expect(1, 65392, '\p{^Is_Ext=false}', "");
Expect(1, 65392, '\P{Is_Ext=false}', "");
Expect(0, 65392, '\P{^Is_Ext=false}', "");
Expect(1, 65393, '\p{Is_Ext=-False}', "");
Expect(0, 65393, '\p{^Is_Ext=-False}', "");
Expect(0, 65393, '\P{Is_Ext=-False}', "");
Expect(1, 65393, '\P{^Is_Ext=-False}', "");
Expect(0, 65392, '\p{Is_Ext=-False}', "");
Expect(1, 65392, '\p{^Is_Ext=-False}', "");
Expect(1, 65392, '\P{Is_Ext=-False}', "");
Expect(0, 65392, '\P{^Is_Ext=-False}', "");
Error('\p{Other_Math=Yes}');
Error('\P{Other_Math=Yes}');
Error('\p{OMath=Y}');
Error('\P{OMath=Y}');
Error('\p{Is_Other_Math=T}');
Error('\P{Is_Other_Math=T}');
Error('\p{Is_OMath=True}');
Error('\P{Is_OMath=True}');
Error('\p{Other_Math=No}');
Error('\P{Other_Math=No}');
Error('\p{OMath:   N}');
Error('\P{OMath:   N}');
Error('\p{Is_Other_Math=F}');
Error('\P{Is_Other_Math=F}');
Error('\p{Is_OMath=False}');
Error('\P{Is_OMath=False}');
Error('\p{Other_Default_Ignorable_Code_Point=Yes}');
Error('\P{Other_Default_Ignorable_Code_Point=Yes}');
Error('\p{ODI=Y}');
Error('\P{ODI=Y}');
Error('\p{Is_Other_Default_Ignorable_Code_Point=T}');
Error('\P{Is_Other_Default_Ignorable_Code_Point=T}');
Error('\p{Is_ODI=True}');
Error('\P{Is_ODI=True}');
Error('\p{Other_Default_Ignorable_Code_Point=No}');
Error('\P{Other_Default_Ignorable_Code_Point=No}');
Error('\p{ODI=N}');
Error('\P{ODI=N}');
Error('\p{Is_Other_Default_Ignorable_Code_Point=F}');
Error('\P{Is_Other_Default_Ignorable_Code_Point=F}');
Error('\p{Is_ODI=False}');
Error('\P{Is_ODI=False}');
Error('\p{XID_Start=/a/Yes}');
Error('\P{XID_Start=/a/Yes}');
Expect(1, 195101, '\p{XID_Start=yes}', "");
Expect(0, 195101, '\p{^XID_Start=yes}', "");
Expect(0, 195101, '\P{XID_Start=yes}', "");
Expect(1, 195101, '\P{^XID_Start=yes}', "");
Expect(0, 195102, '\p{XID_Start=yes}', "");
Expect(1, 195102, '\p{^XID_Start=yes}', "");
Expect(1, 195102, '\P{XID_Start=yes}', "");
Expect(0, 195102, '\P{^XID_Start=yes}', "");
Expect(1, 195101, '\p{XID_Start=-_YES}', "");
Expect(0, 195101, '\p{^XID_Start=-_YES}', "");
Expect(0, 195101, '\P{XID_Start=-_YES}', "");
Expect(1, 195101, '\P{^XID_Start=-_YES}', "");
Expect(0, 195102, '\p{XID_Start=-_YES}', "");
Expect(1, 195102, '\p{^XID_Start=-_YES}', "");
Expect(1, 195102, '\P{XID_Start=-_YES}', "");
Expect(0, 195102, '\P{^XID_Start=-_YES}', "");
Error('\p{XIDS=-/a/Y}');
Error('\P{XIDS=-/a/Y}');
Expect(1, 195101, '\p{XIDS=y}', "");
Expect(0, 195101, '\p{^XIDS=y}', "");
Expect(0, 195101, '\P{XIDS=y}', "");
Expect(1, 195101, '\P{^XIDS=y}', "");
Expect(0, 195102, '\p{XIDS=y}', "");
Expect(1, 195102, '\p{^XIDS=y}', "");
Expect(1, 195102, '\P{XIDS=y}', "");
Expect(0, 195102, '\P{^XIDS=y}', "");
Expect(1, 195101, '\p{XIDS: _	Y}', "");
Expect(0, 195101, '\p{^XIDS: _	Y}', "");
Expect(0, 195101, '\P{XIDS: _	Y}', "");
Expect(1, 195101, '\P{^XIDS: _	Y}', "");
Expect(0, 195102, '\p{XIDS: _	Y}', "");
Expect(1, 195102, '\p{^XIDS: _	Y}', "");
Expect(1, 195102, '\P{XIDS: _	Y}', "");
Expect(0, 195102, '\P{^XIDS: _	Y}', "");
Error('\p{Is_XID_Start=:=-	T}');
Error('\P{Is_XID_Start=:=-	T}');
Expect(1, 195101, '\p{Is_XID_Start=t}', "");
Expect(0, 195101, '\p{^Is_XID_Start=t}', "");
Expect(0, 195101, '\P{Is_XID_Start=t}', "");
Expect(1, 195101, '\P{^Is_XID_Start=t}', "");
Expect(0, 195102, '\p{Is_XID_Start=t}', "");
Expect(1, 195102, '\p{^Is_XID_Start=t}', "");
Expect(1, 195102, '\P{Is_XID_Start=t}', "");
Expect(0, 195102, '\P{^Is_XID_Start=t}', "");
Expect(1, 195101, '\p{Is_XID_Start= -T}', "");
Expect(0, 195101, '\p{^Is_XID_Start= -T}', "");
Expect(0, 195101, '\P{Is_XID_Start= -T}', "");
Expect(1, 195101, '\P{^Is_XID_Start= -T}', "");
Expect(0, 195102, '\p{Is_XID_Start= -T}', "");
Expect(1, 195102, '\p{^Is_XID_Start= -T}', "");
Expect(1, 195102, '\P{Is_XID_Start= -T}', "");
Expect(0, 195102, '\P{^Is_XID_Start= -T}', "");
Error('\p{Is_XIDS=:=	 True}');
Error('\P{Is_XIDS=:=	 True}');
Expect(1, 195101, '\p{Is_XIDS=true}', "");
Expect(0, 195101, '\p{^Is_XIDS=true}', "");
Expect(0, 195101, '\P{Is_XIDS=true}', "");
Expect(1, 195101, '\P{^Is_XIDS=true}', "");
Expect(0, 195102, '\p{Is_XIDS=true}', "");
Expect(1, 195102, '\p{^Is_XIDS=true}', "");
Expect(1, 195102, '\P{Is_XIDS=true}', "");
Expect(0, 195102, '\P{^Is_XIDS=true}', "");
Expect(1, 195101, '\p{Is_XIDS= -true}', "");
Expect(0, 195101, '\p{^Is_XIDS= -true}', "");
Expect(0, 195101, '\P{Is_XIDS= -true}', "");
Expect(1, 195101, '\P{^Is_XIDS= -true}', "");
Expect(0, 195102, '\p{Is_XIDS= -true}', "");
Expect(1, 195102, '\p{^Is_XIDS= -true}', "");
Expect(1, 195102, '\P{Is_XIDS= -true}', "");
Expect(0, 195102, '\P{^Is_XIDS= -true}', "");
Error('\p{XID_Start=:=_	no}');
Error('\P{XID_Start=:=_	no}');
Expect(1, 195102, '\p{XID_Start=no}', "");
Expect(0, 195102, '\p{^XID_Start=no}', "");
Expect(0, 195102, '\P{XID_Start=no}', "");
Expect(1, 195102, '\P{^XID_Start=no}', "");
Expect(0, 195101, '\p{XID_Start=no}', "");
Expect(1, 195101, '\p{^XID_Start=no}', "");
Expect(1, 195101, '\P{XID_Start=no}', "");
Expect(0, 195101, '\P{^XID_Start=no}', "");
Expect(1, 195102, '\p{XID_Start= no}', "");
Expect(0, 195102, '\p{^XID_Start= no}', "");
Expect(0, 195102, '\P{XID_Start= no}', "");
Expect(1, 195102, '\P{^XID_Start= no}', "");
Expect(0, 195101, '\p{XID_Start= no}', "");
Expect(1, 195101, '\p{^XID_Start= no}', "");
Expect(1, 195101, '\P{XID_Start= no}', "");
Expect(0, 195101, '\P{^XID_Start= no}', "");
Error('\p{XIDS=:= _N}');
Error('\P{XIDS=:= _N}');
Expect(1, 195102, '\p{XIDS=n}', "");
Expect(0, 195102, '\p{^XIDS=n}', "");
Expect(0, 195102, '\P{XIDS=n}', "");
Expect(1, 195102, '\P{^XIDS=n}', "");
Expect(0, 195101, '\p{XIDS=n}', "");
Expect(1, 195101, '\p{^XIDS=n}', "");
Expect(1, 195101, '\P{XIDS=n}', "");
Expect(0, 195101, '\P{^XIDS=n}', "");
Expect(1, 195102, '\p{XIDS=-n}', "");
Expect(0, 195102, '\p{^XIDS=-n}', "");
Expect(0, 195102, '\P{XIDS=-n}', "");
Expect(1, 195102, '\P{^XIDS=-n}', "");
Expect(0, 195101, '\p{XIDS=-n}', "");
Expect(1, 195101, '\p{^XIDS=-n}', "");
Expect(1, 195101, '\P{XIDS=-n}', "");
Expect(0, 195101, '\P{^XIDS=-n}', "");
Error('\p{Is_XID_Start=/a/ -F}');
Error('\P{Is_XID_Start=/a/ -F}');
Expect(1, 195102, '\p{Is_XID_Start=f}', "");
Expect(0, 195102, '\p{^Is_XID_Start=f}', "");
Expect(0, 195102, '\P{Is_XID_Start=f}', "");
Expect(1, 195102, '\P{^Is_XID_Start=f}', "");
Expect(0, 195101, '\p{Is_XID_Start=f}', "");
Expect(1, 195101, '\p{^Is_XID_Start=f}', "");
Expect(1, 195101, '\P{Is_XID_Start=f}', "");
Expect(0, 195101, '\P{^Is_XID_Start=f}', "");
Expect(1, 195102, '\p{Is_XID_Start=F}', "");
Expect(0, 195102, '\p{^Is_XID_Start=F}', "");
Expect(0, 195102, '\P{Is_XID_Start=F}', "");
Expect(1, 195102, '\P{^Is_XID_Start=F}', "");
Expect(0, 195101, '\p{Is_XID_Start=F}', "");
Expect(1, 195101, '\p{^Is_XID_Start=F}', "");
Expect(1, 195101, '\P{Is_XID_Start=F}', "");
Expect(0, 195101, '\P{^Is_XID_Start=F}', "");
Error('\p{Is_XIDS= /a/False}');
Error('\P{Is_XIDS= /a/False}');
Expect(1, 195102, '\p{Is_XIDS=false}', "");
Expect(0, 195102, '\p{^Is_XIDS=false}', "");
Expect(0, 195102, '\P{Is_XIDS=false}', "");
Expect(1, 195102, '\P{^Is_XIDS=false}', "");
Expect(0, 195101, '\p{Is_XIDS=false}', "");
Expect(1, 195101, '\p{^Is_XIDS=false}', "");
Expect(1, 195101, '\P{Is_XIDS=false}', "");
Expect(0, 195101, '\P{^Is_XIDS=false}', "");
Expect(1, 195102, '\p{Is_XIDS=-	False}', "");
Expect(0, 195102, '\p{^Is_XIDS=-	False}', "");
Expect(0, 195102, '\P{Is_XIDS=-	False}', "");
Expect(1, 195102, '\P{^Is_XIDS=-	False}', "");
Expect(0, 195101, '\p{Is_XIDS=-	False}', "");
Expect(1, 195101, '\p{^Is_XIDS=-	False}', "");
Expect(1, 195101, '\P{Is_XIDS=-	False}', "");
Expect(0, 195101, '\P{^Is_XIDS=-	False}', "");
Error('\p{Changes_When_NFKC_Casefolded=	yes/a/}');
Error('\P{Changes_When_NFKC_Casefolded=	yes/a/}');
Expect(1, 921599, '\p{Changes_When_NFKC_Casefolded=yes}', "");
Expect(0, 921599, '\p{^Changes_When_NFKC_Casefolded=yes}', "");
Expect(0, 921599, '\P{Changes_When_NFKC_Casefolded=yes}', "");
Expect(1, 921599, '\P{^Changes_When_NFKC_Casefolded=yes}', "");
Expect(0, 921600, '\p{Changes_When_NFKC_Casefolded=yes}', "");
Expect(1, 921600, '\p{^Changes_When_NFKC_Casefolded=yes}', "");
Expect(1, 921600, '\P{Changes_When_NFKC_Casefolded=yes}', "");
Expect(0, 921600, '\P{^Changes_When_NFKC_Casefolded=yes}', "");
Expect(1, 921599, '\p{Changes_When_NFKC_Casefolded=	YES}', "");
Expect(0, 921599, '\p{^Changes_When_NFKC_Casefolded=	YES}', "");
Expect(0, 921599, '\P{Changes_When_NFKC_Casefolded=	YES}', "");
Expect(1, 921599, '\P{^Changes_When_NFKC_Casefolded=	YES}', "");
Expect(0, 921600, '\p{Changes_When_NFKC_Casefolded=	YES}', "");
Expect(1, 921600, '\p{^Changes_When_NFKC_Casefolded=	YES}', "");
Expect(1, 921600, '\P{Changes_When_NFKC_Casefolded=	YES}', "");
Expect(0, 921600, '\P{^Changes_When_NFKC_Casefolded=	YES}', "");
Error('\p{CWKCF=/a/	Y}');
Error('\P{CWKCF=/a/	Y}');
Expect(1, 921599, '\p{CWKCF:y}', "");
Expect(0, 921599, '\p{^CWKCF:y}', "");
Expect(0, 921599, '\P{CWKCF:y}', "");
Expect(1, 921599, '\P{^CWKCF:y}', "");
Expect(0, 921600, '\p{CWKCF:y}', "");
Expect(1, 921600, '\p{^CWKCF:y}', "");
Expect(1, 921600, '\P{CWKCF:y}', "");
Expect(0, 921600, '\P{^CWKCF:y}', "");
Expect(1, 921599, '\p{CWKCF= -Y}', "");
Expect(0, 921599, '\p{^CWKCF= -Y}', "");
Expect(0, 921599, '\P{CWKCF= -Y}', "");
Expect(1, 921599, '\P{^CWKCF= -Y}', "");
Expect(0, 921600, '\p{CWKCF= -Y}', "");
Expect(1, 921600, '\p{^CWKCF= -Y}', "");
Expect(1, 921600, '\P{CWKCF= -Y}', "");
Expect(0, 921600, '\P{^CWKCF= -Y}', "");
Error('\p{Is_Changes_When_NFKC_Casefolded=/a/-t}');
Error('\P{Is_Changes_When_NFKC_Casefolded=/a/-t}');
Expect(1, 921599, '\p{Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(0, 921599, '\p{^Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(0, 921599, '\P{Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(1, 921599, '\P{^Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(0, 921600, '\p{Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(1, 921600, '\p{^Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(1, 921600, '\P{Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(0, 921600, '\P{^Is_Changes_When_NFKC_Casefolded=t}', "");
Expect(1, 921599, '\p{Is_Changes_When_NFKC_Casefolded: 	 T}', "");
Expect(0, 921599, '\p{^Is_Changes_When_NFKC_Casefolded: 	 T}', "");
Expect(0, 921599, '\P{Is_Changes_When_NFKC_Casefolded: 	 T}', "");
Expect(1, 921599, '\P{^Is_Changes_When_NFKC_Casefolded: 	 T}', "");
Expect(0, 921600, '\p{Is_Changes_When_NFKC_Casefolded: 	 T}', "");
Expect(1, 921600, '\p{^Is_Changes_When_NFKC_Casefolded: 	 T}', "");
Expect(1, 921600, '\P{Is_Changes_When_NFKC_Casefolded: 	 T}', "");
Expect(0, 921600, '\P{^Is_Changes_When_NFKC_Casefolded: 	 T}', "");
Error('\p{Is_CWKCF=_/a/TRUE}');
Error('\P{Is_CWKCF=_/a/TRUE}');
Expect(1, 921599, '\p{Is_CWKCF=true}', "");
Expect(0, 921599, '\p{^Is_CWKCF=true}', "");
Expect(0, 921599, '\P{Is_CWKCF=true}', "");
Expect(1, 921599, '\P{^Is_CWKCF=true}', "");
Expect(0, 921600, '\p{Is_CWKCF=true}', "");
Expect(1, 921600, '\p{^Is_CWKCF=true}', "");
Expect(1, 921600, '\P{Is_CWKCF=true}', "");
Expect(0, 921600, '\P{^Is_CWKCF=true}', "");
Expect(1, 921599, '\p{Is_CWKCF=_TRUE}', "");
Expect(0, 921599, '\p{^Is_CWKCF=_TRUE}', "");
Expect(0, 921599, '\P{Is_CWKCF=_TRUE}', "");
Expect(1, 921599, '\P{^Is_CWKCF=_TRUE}', "");
Expect(0, 921600, '\p{Is_CWKCF=_TRUE}', "");
Expect(1, 921600, '\p{^Is_CWKCF=_TRUE}', "");
Expect(1, 921600, '\P{Is_CWKCF=_TRUE}', "");
Expect(0, 921600, '\P{^Is_CWKCF=_TRUE}', "");
Error('\p{Changes_When_NFKC_Casefolded=	/a/No}');
Error('\P{Changes_When_NFKC_Casefolded=	/a/No}');
Expect(1, 921600, '\p{Changes_When_NFKC_Casefolded=no}', "");
Expect(0, 921600, '\p{^Changes_When_NFKC_Casefolded=no}', "");
Expect(0, 921600, '\P{Changes_When_NFKC_Casefolded=no}', "");
Expect(1, 921600, '\P{^Changes_When_NFKC_Casefolded=no}', "");
Expect(0, 921599, '\p{Changes_When_NFKC_Casefolded=no}', "");
Expect(1, 921599, '\p{^Changes_When_NFKC_Casefolded=no}', "");
Expect(1, 921599, '\P{Changes_When_NFKC_Casefolded=no}', "");
Expect(0, 921599, '\P{^Changes_When_NFKC_Casefolded=no}', "");
Expect(1, 921600, '\p{Changes_When_NFKC_Casefolded=_-NO}', "");
Expect(0, 921600, '\p{^Changes_When_NFKC_Casefolded=_-NO}', "");
Expect(0, 921600, '\P{Changes_When_NFKC_Casefolded=_-NO}', "");
Expect(1, 921600, '\P{^Changes_When_NFKC_Casefolded=_-NO}', "");
Expect(0, 921599, '\p{Changes_When_NFKC_Casefolded=_-NO}', "");
Expect(1, 921599, '\p{^Changes_When_NFKC_Casefolded=_-NO}', "");
Expect(1, 921599, '\P{Changes_When_NFKC_Casefolded=_-NO}', "");
Expect(0, 921599, '\P{^Changes_When_NFKC_Casefolded=_-NO}', "");
Error('\p{CWKCF= -n:=}');
Error('\P{CWKCF= -n:=}');
Expect(1, 921600, '\p{CWKCF=n}', "");
Expect(0, 921600, '\p{^CWKCF=n}', "");
Expect(0, 921600, '\P{CWKCF=n}', "");
Expect(1, 921600, '\P{^CWKCF=n}', "");
Expect(0, 921599, '\p{CWKCF=n}', "");
Expect(1, 921599, '\p{^CWKCF=n}', "");
Expect(1, 921599, '\P{CWKCF=n}', "");
Expect(0, 921599, '\P{^CWKCF=n}', "");
Expect(1, 921600, '\p{CWKCF=__N}', "");
Expect(0, 921600, '\p{^CWKCF=__N}', "");
Expect(0, 921600, '\P{CWKCF=__N}', "");
Expect(1, 921600, '\P{^CWKCF=__N}', "");
Expect(0, 921599, '\p{CWKCF=__N}', "");
Expect(1, 921599, '\p{^CWKCF=__N}', "");
Expect(1, 921599, '\P{CWKCF=__N}', "");
Expect(0, 921599, '\P{^CWKCF=__N}', "");
Error('\p{Is_Changes_When_NFKC_Casefolded=_:=F}');
Error('\P{Is_Changes_When_NFKC_Casefolded=_:=F}');
Expect(1, 921600, '\p{Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(0, 921600, '\p{^Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(0, 921600, '\P{Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(1, 921600, '\P{^Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(0, 921599, '\p{Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(1, 921599, '\p{^Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(1, 921599, '\P{Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(0, 921599, '\P{^Is_Changes_When_NFKC_Casefolded:	f}', "");
Expect(1, 921600, '\p{Is_Changes_When_NFKC_Casefolded=-_F}', "");
Expect(0, 921600, '\p{^Is_Changes_When_NFKC_Casefolded=-_F}', "");
Expect(0, 921600, '\P{Is_Changes_When_NFKC_Casefolded=-_F}', "");
Expect(1, 921600, '\P{^Is_Changes_When_NFKC_Casefolded=-_F}', "");
Expect(0, 921599, '\p{Is_Changes_When_NFKC_Casefolded=-_F}', "");
Expect(1, 921599, '\p{^Is_Changes_When_NFKC_Casefolded=-_F}', "");
Expect(1, 921599, '\P{Is_Changes_When_NFKC_Casefolded=-_F}', "");
Expect(0, 921599, '\P{^Is_Changes_When_NFKC_Casefolded=-_F}', "");
Error('\p{Is_CWKCF=:=-_False}');
Error('\P{Is_CWKCF=:=-_False}');
Expect(1, 921600, '\p{Is_CWKCF=false}', "");
Expect(0, 921600, '\p{^Is_CWKCF=false}', "");
Expect(0, 921600, '\P{Is_CWKCF=false}', "");
Expect(1, 921600, '\P{^Is_CWKCF=false}', "");
Expect(0, 921599, '\p{Is_CWKCF=false}', "");
Expect(1, 921599, '\p{^Is_CWKCF=false}', "");
Expect(1, 921599, '\P{Is_CWKCF=false}', "");
Expect(0, 921599, '\P{^Is_CWKCF=false}', "");
Expect(1, 921600, '\p{Is_CWKCF= FALSE}', "");
Expect(0, 921600, '\p{^Is_CWKCF= FALSE}', "");
Expect(0, 921600, '\P{Is_CWKCF= FALSE}', "");
Expect(1, 921600, '\P{^Is_CWKCF= FALSE}', "");
Expect(0, 921599, '\p{Is_CWKCF= FALSE}', "");
Expect(1, 921599, '\p{^Is_CWKCF= FALSE}', "");
Expect(1, 921599, '\P{Is_CWKCF= FALSE}', "");
Expect(0, 921599, '\P{^Is_CWKCF= FALSE}', "");
Error('\p{East_Asian_Width=	 Wide/a/}');
Error('\P{East_Asian_Width=	 Wide/a/}');
Expect(1, 262141, '\p{East_Asian_Width=wide}', "");
Expect(0, 262141, '\p{^East_Asian_Width=wide}', "");
Expect(0, 262141, '\P{East_Asian_Width=wide}', "");
Expect(1, 262141, '\P{^East_Asian_Width=wide}', "");
Expect(0, 262144, '\p{East_Asian_Width=wide}', "");
Expect(1, 262144, '\p{^East_Asian_Width=wide}', "");
Expect(1, 262144, '\P{East_Asian_Width=wide}', "");
Expect(0, 262144, '\P{^East_Asian_Width=wide}', "");
Expect(1, 262141, '\p{East_Asian_Width=-Wide}', "");
Expect(0, 262141, '\p{^East_Asian_Width=-Wide}', "");
Expect(0, 262141, '\P{East_Asian_Width=-Wide}', "");
Expect(1, 262141, '\P{^East_Asian_Width=-Wide}', "");
Expect(0, 262144, '\p{East_Asian_Width=-Wide}', "");
Expect(1, 262144, '\p{^East_Asian_Width=-Wide}', "");
Expect(1, 262144, '\P{East_Asian_Width=-Wide}', "");
Expect(0, 262144, '\P{^East_Asian_Width=-Wide}', "");
Error('\p{Ea=	-W:=}');
Error('\P{Ea=	-W:=}');
Expect(1, 262141, '\p{Ea=w}', "");
Expect(0, 262141, '\p{^Ea=w}', "");
Expect(0, 262141, '\P{Ea=w}', "");
Expect(1, 262141, '\P{^Ea=w}', "");
Expect(0, 262144, '\p{Ea=w}', "");
Expect(1, 262144, '\p{^Ea=w}', "");
Expect(1, 262144, '\P{Ea=w}', "");
Expect(0, 262144, '\P{^Ea=w}', "");
Expect(1, 262141, '\p{Ea=-_W}', "");
Expect(0, 262141, '\p{^Ea=-_W}', "");
Expect(0, 262141, '\P{Ea=-_W}', "");
Expect(1, 262141, '\P{^Ea=-_W}', "");
Expect(0, 262144, '\p{Ea=-_W}', "");
Expect(1, 262144, '\p{^Ea=-_W}', "");
Expect(1, 262144, '\P{Ea=-_W}', "");
Expect(0, 262144, '\P{^Ea=-_W}', "");
Error('\p{Is_East_Asian_Width:   :=  Wide}');
Error('\P{Is_East_Asian_Width:   :=  Wide}');
Expect(1, 262141, '\p{Is_East_Asian_Width=wide}', "");
Expect(0, 262141, '\p{^Is_East_Asian_Width=wide}', "");
Expect(0, 262141, '\P{Is_East_Asian_Width=wide}', "");
Expect(1, 262141, '\P{^Is_East_Asian_Width=wide}', "");
Expect(0, 262144, '\p{Is_East_Asian_Width=wide}', "");
Expect(1, 262144, '\p{^Is_East_Asian_Width=wide}', "");
Expect(1, 262144, '\P{Is_East_Asian_Width=wide}', "");
Expect(0, 262144, '\P{^Is_East_Asian_Width=wide}', "");
Expect(1, 262141, '\p{Is_East_Asian_Width=__wide}', "");
Expect(0, 262141, '\p{^Is_East_Asian_Width=__wide}', "");
Expect(0, 262141, '\P{Is_East_Asian_Width=__wide}', "");
Expect(1, 262141, '\P{^Is_East_Asian_Width=__wide}', "");
Expect(0, 262144, '\p{Is_East_Asian_Width=__wide}', "");
Expect(1, 262144, '\p{^Is_East_Asian_Width=__wide}', "");
Expect(1, 262144, '\P{Is_East_Asian_Width=__wide}', "");
Expect(0, 262144, '\P{^Is_East_Asian_Width=__wide}', "");
Error('\p{Is_Ea=	_W/a/}');
Error('\P{Is_Ea=	_W/a/}');
Expect(1, 262141, '\p{Is_Ea=w}', "");
Expect(0, 262141, '\p{^Is_Ea=w}', "");
Expect(0, 262141, '\P{Is_Ea=w}', "");
Expect(1, 262141, '\P{^Is_Ea=w}', "");
Expect(0, 262144, '\p{Is_Ea=w}', "");
Expect(1, 262144, '\p{^Is_Ea=w}', "");
Expect(1, 262144, '\P{Is_Ea=w}', "");
Expect(0, 262144, '\P{^Is_Ea=w}', "");
Expect(1, 262141, '\p{Is_Ea=	W}', "");
Expect(0, 262141, '\p{^Is_Ea=	W}', "");
Expect(0, 262141, '\P{Is_Ea=	W}', "");
Expect(1, 262141, '\P{^Is_Ea=	W}', "");
Expect(0, 262144, '\p{Is_Ea=	W}', "");
Expect(1, 262144, '\p{^Is_Ea=	W}', "");
Expect(1, 262144, '\P{Is_Ea=	W}', "");
Expect(0, 262144, '\P{^Is_Ea=	W}', "");
Error('\p{East_Asian_Width=-	fullwidth/a/}');
Error('\P{East_Asian_Width=-	fullwidth/a/}');
Expect(1, 65510, '\p{East_Asian_Width=fullwidth}', "");
Expect(0, 65510, '\p{^East_Asian_Width=fullwidth}', "");
Expect(0, 65510, '\P{East_Asian_Width=fullwidth}', "");
Expect(1, 65510, '\P{^East_Asian_Width=fullwidth}', "");
Expect(0, 65511, '\p{East_Asian_Width=fullwidth}', "");
Expect(1, 65511, '\p{^East_Asian_Width=fullwidth}', "");
Expect(1, 65511, '\P{East_Asian_Width=fullwidth}', "");
Expect(0, 65511, '\P{^East_Asian_Width=fullwidth}', "");
Expect(1, 65510, '\p{East_Asian_Width:   - FULLWIDTH}', "");
Expect(0, 65510, '\p{^East_Asian_Width:   - FULLWIDTH}', "");
Expect(0, 65510, '\P{East_Asian_Width:   - FULLWIDTH}', "");
Expect(1, 65510, '\P{^East_Asian_Width:   - FULLWIDTH}', "");
Expect(0, 65511, '\p{East_Asian_Width:   - FULLWIDTH}', "");
Expect(1, 65511, '\p{^East_Asian_Width:   - FULLWIDTH}', "");
Expect(1, 65511, '\P{East_Asian_Width:   - FULLWIDTH}', "");
Expect(0, 65511, '\P{^East_Asian_Width:   - FULLWIDTH}', "");
Error('\p{Ea=/a/F}');
Error('\P{Ea=/a/F}');
Expect(1, 65510, '\p{Ea=f}', "");
Expect(0, 65510, '\p{^Ea=f}', "");
Expect(0, 65510, '\P{Ea=f}', "");
Expect(1, 65510, '\P{^Ea=f}', "");
Expect(0, 65511, '\p{Ea=f}', "");
Expect(1, 65511, '\p{^Ea=f}', "");
Expect(1, 65511, '\P{Ea=f}', "");
Expect(0, 65511, '\P{^Ea=f}', "");
Expect(1, 65510, '\p{Ea=	_f}', "");
Expect(0, 65510, '\p{^Ea=	_f}', "");
Expect(0, 65510, '\P{Ea=	_f}', "");
Expect(1, 65510, '\P{^Ea=	_f}', "");
Expect(0, 65511, '\p{Ea=	_f}', "");
Expect(1, 65511, '\p{^Ea=	_f}', "");
Expect(1, 65511, '\P{Ea=	_f}', "");
Expect(0, 65511, '\P{^Ea=	_f}', "");
Error('\p{Is_East_Asian_Width=- fullwidth:=}');
Error('\P{Is_East_Asian_Width=- fullwidth:=}');
Expect(1, 65510, '\p{Is_East_Asian_Width=fullwidth}', "");
Expect(0, 65510, '\p{^Is_East_Asian_Width=fullwidth}', "");
Expect(0, 65510, '\P{Is_East_Asian_Width=fullwidth}', "");
Expect(1, 65510, '\P{^Is_East_Asian_Width=fullwidth}', "");
Expect(0, 65511, '\p{Is_East_Asian_Width=fullwidth}', "");
Expect(1, 65511, '\p{^Is_East_Asian_Width=fullwidth}', "");
Expect(1, 65511, '\P{Is_East_Asian_Width=fullwidth}', "");
Expect(0, 65511, '\P{^Is_East_Asian_Width=fullwidth}', "");
Expect(1, 65510, '\p{Is_East_Asian_Width=- Fullwidth}', "");
Expect(0, 65510, '\p{^Is_East_Asian_Width=- Fullwidth}', "");
Expect(0, 65510, '\P{Is_East_Asian_Width=- Fullwidth}', "");
Expect(1, 65510, '\P{^Is_East_Asian_Width=- Fullwidth}', "");
Expect(0, 65511, '\p{Is_East_Asian_Width=- Fullwidth}', "");
Expect(1, 65511, '\p{^Is_East_Asian_Width=- Fullwidth}', "");
Expect(1, 65511, '\P{Is_East_Asian_Width=- Fullwidth}', "");
Expect(0, 65511, '\P{^Is_East_Asian_Width=- Fullwidth}', "");
Error('\p{Is_Ea=	 F:=}');
Error('\P{Is_Ea=	 F:=}');
Expect(1, 65510, '\p{Is_Ea=f}', "");
Expect(0, 65510, '\p{^Is_Ea=f}', "");
Expect(0, 65510, '\P{Is_Ea=f}', "");
Expect(1, 65510, '\P{^Is_Ea=f}', "");
Expect(0, 65511, '\p{Is_Ea=f}', "");
Expect(1, 65511, '\p{^Is_Ea=f}', "");
Expect(1, 65511, '\P{Is_Ea=f}', "");
Expect(0, 65511, '\P{^Is_Ea=f}', "");
Expect(1, 65510, '\p{Is_Ea=_ F}', "");
Expect(0, 65510, '\p{^Is_Ea=_ F}', "");
Expect(0, 65510, '\P{Is_Ea=_ F}', "");
Expect(1, 65510, '\P{^Is_Ea=_ F}', "");
Expect(0, 65511, '\p{Is_Ea=_ F}', "");
Expect(1, 65511, '\p{^Is_Ea=_ F}', "");
Expect(1, 65511, '\P{Is_Ea=_ F}', "");
Expect(0, 65511, '\P{^Is_Ea=_ F}', "");
Error('\p{East_Asian_Width=-_Ambiguous:=}');
Error('\P{East_Asian_Width=-_Ambiguous:=}');
Expect(1, 1114109, '\p{East_Asian_Width=ambiguous}', "");
Expect(0, 1114109, '\p{^East_Asian_Width=ambiguous}', "");
Expect(0, 1114109, '\P{East_Asian_Width=ambiguous}', "");
Expect(1, 1114109, '\P{^East_Asian_Width=ambiguous}', "");
Expect(0, 918000, '\p{East_Asian_Width=ambiguous}', "");
Expect(1, 918000, '\p{^East_Asian_Width=ambiguous}', "");
Expect(1, 918000, '\P{East_Asian_Width=ambiguous}', "");
Expect(0, 918000, '\P{^East_Asian_Width=ambiguous}', "");
Expect(1, 1114109, '\p{East_Asian_Width=	_Ambiguous}', "");
Expect(0, 1114109, '\p{^East_Asian_Width=	_Ambiguous}', "");
Expect(0, 1114109, '\P{East_Asian_Width=	_Ambiguous}', "");
Expect(1, 1114109, '\P{^East_Asian_Width=	_Ambiguous}', "");
Expect(0, 918000, '\p{East_Asian_Width=	_Ambiguous}', "");
Expect(1, 918000, '\p{^East_Asian_Width=	_Ambiguous}', "");
Expect(1, 918000, '\P{East_Asian_Width=	_Ambiguous}', "");
Expect(0, 918000, '\P{^East_Asian_Width=	_Ambiguous}', "");
Error('\p{Ea=	/a/A}');
Error('\P{Ea=	/a/A}');
Expect(1, 1114109, '\p{Ea=a}', "");
Expect(0, 1114109, '\p{^Ea=a}', "");
Expect(0, 1114109, '\P{Ea=a}', "");
Expect(1, 1114109, '\P{^Ea=a}', "");
Expect(0, 918000, '\p{Ea=a}', "");
Expect(1, 918000, '\p{^Ea=a}', "");
Expect(1, 918000, '\P{Ea=a}', "");
Expect(0, 918000, '\P{^Ea=a}', "");
Expect(1, 1114109, '\p{Ea= a}', "");
Expect(0, 1114109, '\p{^Ea= a}', "");
Expect(0, 1114109, '\P{Ea= a}', "");
Expect(1, 1114109, '\P{^Ea= a}', "");
Expect(0, 918000, '\p{Ea= a}', "");
Expect(1, 918000, '\p{^Ea= a}', "");
Expect(1, 918000, '\P{Ea= a}', "");
Expect(0, 918000, '\P{^Ea= a}', "");
Error('\p{Is_East_Asian_Width= Ambiguous/a/}');
Error('\P{Is_East_Asian_Width= Ambiguous/a/}');
Expect(1, 1114109, '\p{Is_East_Asian_Width=ambiguous}', "");
Expect(0, 1114109, '\p{^Is_East_Asian_Width=ambiguous}', "");
Expect(0, 1114109, '\P{Is_East_Asian_Width=ambiguous}', "");
Expect(1, 1114109, '\P{^Is_East_Asian_Width=ambiguous}', "");
Expect(0, 918000, '\p{Is_East_Asian_Width=ambiguous}', "");
Expect(1, 918000, '\p{^Is_East_Asian_Width=ambiguous}', "");
Expect(1, 918000, '\P{Is_East_Asian_Width=ambiguous}', "");
Expect(0, 918000, '\P{^Is_East_Asian_Width=ambiguous}', "");
Expect(1, 1114109, '\p{Is_East_Asian_Width= 	AMBIGUOUS}', "");
Expect(0, 1114109, '\p{^Is_East_Asian_Width= 	AMBIGUOUS}', "");
Expect(0, 1114109, '\P{Is_East_Asian_Width= 	AMBIGUOUS}', "");
Expect(1, 1114109, '\P{^Is_East_Asian_Width= 	AMBIGUOUS}', "");
Expect(0, 918000, '\p{Is_East_Asian_Width= 	AMBIGUOUS}', "");
Expect(1, 918000, '\p{^Is_East_Asian_Width= 	AMBIGUOUS}', "");
Expect(1, 918000, '\P{Is_East_Asian_Width= 	AMBIGUOUS}', "");
Expect(0, 918000, '\P{^Is_East_Asian_Width= 	AMBIGUOUS}', "");
Error('\p{Is_Ea=:= _A}');
Error('\P{Is_Ea=:= _A}');
Expect(1, 1114109, '\p{Is_Ea=a}', "");
Expect(0, 1114109, '\p{^Is_Ea=a}', "");
Expect(0, 1114109, '\P{Is_Ea=a}', "");
Expect(1, 1114109, '\P{^Is_Ea=a}', "");
Expect(0, 918000, '\p{Is_Ea=a}', "");
Expect(1, 918000, '\p{^Is_Ea=a}', "");
Expect(1, 918000, '\P{Is_Ea=a}', "");
Expect(0, 918000, '\P{^Is_Ea=a}', "");
Expect(1, 1114109, '\p{Is_Ea:	__A}', "");
Expect(0, 1114109, '\p{^Is_Ea:	__A}', "");
Expect(0, 1114109, '\P{Is_Ea:	__A}', "");
Expect(1, 1114109, '\P{^Is_Ea:	__A}', "");
Expect(0, 918000, '\p{Is_Ea:	__A}', "");
Expect(1, 918000, '\p{^Is_Ea:	__A}', "");
Expect(1, 918000, '\P{Is_Ea:	__A}', "");
Expect(0, 918000, '\P{^Is_Ea:	__A}', "");
Error('\p{East_Asian_Width= 	NARROW:=}');
Error('\P{East_Asian_Width= 	NARROW:=}');
Expect(1, 10630, '\p{East_Asian_Width=narrow}', "");
Expect(0, 10630, '\p{^East_Asian_Width=narrow}', "");
Expect(0, 10630, '\P{East_Asian_Width=narrow}', "");
Expect(1, 10630, '\P{^East_Asian_Width=narrow}', "");
Expect(0, 10631, '\p{East_Asian_Width=narrow}', "");
Expect(1, 10631, '\p{^East_Asian_Width=narrow}', "");
Expect(1, 10631, '\P{East_Asian_Width=narrow}', "");
Expect(0, 10631, '\P{^East_Asian_Width=narrow}', "");
Expect(1, 10630, '\p{East_Asian_Width=_narrow}', "");
Expect(0, 10630, '\p{^East_Asian_Width=_narrow}', "");
Expect(0, 10630, '\P{East_Asian_Width=_narrow}', "");
Expect(1, 10630, '\P{^East_Asian_Width=_narrow}', "");
Expect(0, 10631, '\p{East_Asian_Width=_narrow}', "");
Expect(1, 10631, '\p{^East_Asian_Width=_narrow}', "");
Expect(1, 10631, '\P{East_Asian_Width=_narrow}', "");
Expect(0, 10631, '\P{^East_Asian_Width=_narrow}', "");
Error('\p{Ea= /a/NA}');
Error('\P{Ea= /a/NA}');
Expect(1, 10630, '\p{Ea=na}', "");
Expect(0, 10630, '\p{^Ea=na}', "");
Expect(0, 10630, '\P{Ea=na}', "");
Expect(1, 10630, '\P{^Ea=na}', "");
Expect(0, 10631, '\p{Ea=na}', "");
Expect(1, 10631, '\p{^Ea=na}', "");
Expect(1, 10631, '\P{Ea=na}', "");
Expect(0, 10631, '\P{^Ea=na}', "");
Expect(1, 10630, '\p{Ea=	 NA}', "");
Expect(0, 10630, '\p{^Ea=	 NA}', "");
Expect(0, 10630, '\P{Ea=	 NA}', "");
Expect(1, 10630, '\P{^Ea=	 NA}', "");
Expect(0, 10631, '\p{Ea=	 NA}', "");
Expect(1, 10631, '\p{^Ea=	 NA}', "");
Expect(1, 10631, '\P{Ea=	 NA}', "");
Expect(0, 10631, '\P{^Ea=	 NA}', "");
Error('\p{Is_East_Asian_Width=-/a/narrow}');
Error('\P{Is_East_Asian_Width=-/a/narrow}');
Expect(1, 10630, '\p{Is_East_Asian_Width=narrow}', "");
Expect(0, 10630, '\p{^Is_East_Asian_Width=narrow}', "");
Expect(0, 10630, '\P{Is_East_Asian_Width=narrow}', "");
Expect(1, 10630, '\P{^Is_East_Asian_Width=narrow}', "");
Expect(0, 10631, '\p{Is_East_Asian_Width=narrow}', "");
Expect(1, 10631, '\p{^Is_East_Asian_Width=narrow}', "");
Expect(1, 10631, '\P{Is_East_Asian_Width=narrow}', "");
Expect(0, 10631, '\P{^Is_East_Asian_Width=narrow}', "");
Expect(1, 10630, '\p{Is_East_Asian_Width=-Narrow}', "");
Expect(0, 10630, '\p{^Is_East_Asian_Width=-Narrow}', "");
Expect(0, 10630, '\P{Is_East_Asian_Width=-Narrow}', "");
Expect(1, 10630, '\P{^Is_East_Asian_Width=-Narrow}', "");
Expect(0, 10631, '\p{Is_East_Asian_Width=-Narrow}', "");
Expect(1, 10631, '\p{^Is_East_Asian_Width=-Narrow}', "");
Expect(1, 10631, '\P{Is_East_Asian_Width=-Narrow}', "");
Expect(0, 10631, '\P{^Is_East_Asian_Width=-Narrow}', "");
Error('\p{Is_Ea=		NA:=}');
Error('\P{Is_Ea=		NA:=}');
Expect(1, 10630, '\p{Is_Ea=na}', "");
Expect(0, 10630, '\p{^Is_Ea=na}', "");
Expect(0, 10630, '\P{Is_Ea=na}', "");
Expect(1, 10630, '\P{^Is_Ea=na}', "");
Expect(0, 10631, '\p{Is_Ea=na}', "");
Expect(1, 10631, '\p{^Is_Ea=na}', "");
Expect(1, 10631, '\P{Is_Ea=na}', "");
Expect(0, 10631, '\P{^Is_Ea=na}', "");
Expect(1, 10630, '\p{Is_Ea= NA}', "");
Expect(0, 10630, '\p{^Is_Ea= NA}', "");
Expect(0, 10630, '\P{Is_Ea= NA}', "");
Expect(1, 10630, '\P{^Is_Ea= NA}', "");
Expect(0, 10631, '\p{Is_Ea= NA}', "");
Expect(1, 10631, '\p{^Is_Ea= NA}', "");
Expect(1, 10631, '\P{Is_Ea= NA}', "");
Expect(0, 10631, '\P{^Is_Ea= NA}', "");
Error('\p{East_Asian_Width=	-NEUTRAL:=}');
Error('\P{East_Asian_Width=	-NEUTRAL:=}');
Expect(1, 918000, '\p{East_Asian_Width=neutral}', "");
Expect(0, 918000, '\p{^East_Asian_Width=neutral}', "");
Expect(0, 918000, '\P{East_Asian_Width=neutral}', "");
Expect(1, 918000, '\P{^East_Asian_Width=neutral}', "");
Expect(0, 1114109, '\p{East_Asian_Width=neutral}', "");
Expect(1, 1114109, '\p{^East_Asian_Width=neutral}', "");
Expect(1, 1114109, '\P{East_Asian_Width=neutral}', "");
Expect(0, 1114109, '\P{^East_Asian_Width=neutral}', "");
Expect(1, 918000, '\p{East_Asian_Width=-NEUTRAL}', "");
Expect(0, 918000, '\p{^East_Asian_Width=-NEUTRAL}', "");
Expect(0, 918000, '\P{East_Asian_Width=-NEUTRAL}', "");
Expect(1, 918000, '\P{^East_Asian_Width=-NEUTRAL}', "");
Expect(0, 1114109, '\p{East_Asian_Width=-NEUTRAL}', "");
Expect(1, 1114109, '\p{^East_Asian_Width=-NEUTRAL}', "");
Expect(1, 1114109, '\P{East_Asian_Width=-NEUTRAL}', "");
Expect(0, 1114109, '\P{^East_Asian_Width=-NEUTRAL}', "");
Error('\p{Ea=_-N/a/}');
Error('\P{Ea=_-N/a/}');
Expect(1, 918000, '\p{Ea=n}', "");
Expect(0, 918000, '\p{^Ea=n}', "");
Expect(0, 918000, '\P{Ea=n}', "");
Expect(1, 918000, '\P{^Ea=n}', "");
Expect(0, 1114109, '\p{Ea=n}', "");
Expect(1, 1114109, '\p{^Ea=n}', "");
Expect(1, 1114109, '\P{Ea=n}', "");
Expect(0, 1114109, '\P{^Ea=n}', "");
Expect(1, 918000, '\p{Ea= N}', "");
Expect(0, 918000, '\p{^Ea= N}', "");
Expect(0, 918000, '\P{Ea= N}', "");
Expect(1, 918000, '\P{^Ea= N}', "");
Expect(0, 1114109, '\p{Ea= N}', "");
Expect(1, 1114109, '\p{^Ea= N}', "");
Expect(1, 1114109, '\P{Ea= N}', "");
Expect(0, 1114109, '\P{^Ea= N}', "");
Error('\p{Is_East_Asian_Width=:=_neutral}');
Error('\P{Is_East_Asian_Width=:=_neutral}');
Expect(1, 918000, '\p{Is_East_Asian_Width=neutral}', "");
Expect(0, 918000, '\p{^Is_East_Asian_Width=neutral}', "");
Expect(0, 918000, '\P{Is_East_Asian_Width=neutral}', "");
Expect(1, 918000, '\P{^Is_East_Asian_Width=neutral}', "");
Expect(0, 1114109, '\p{Is_East_Asian_Width=neutral}', "");
Expect(1, 1114109, '\p{^Is_East_Asian_Width=neutral}', "");
Expect(1, 1114109, '\P{Is_East_Asian_Width=neutral}', "");
Expect(0, 1114109, '\P{^Is_East_Asian_Width=neutral}', "");
Expect(1, 918000, '\p{Is_East_Asian_Width= Neutral}', "");
Expect(0, 918000, '\p{^Is_East_Asian_Width= Neutral}', "");
Expect(0, 918000, '\P{Is_East_Asian_Width= Neutral}', "");
Expect(1, 918000, '\P{^Is_East_Asian_Width= Neutral}', "");
Expect(0, 1114109, '\p{Is_East_Asian_Width= Neutral}', "");
Expect(1, 1114109, '\p{^Is_East_Asian_Width= Neutral}', "");
Expect(1, 1114109, '\P{Is_East_Asian_Width= Neutral}', "");
Expect(0, 1114109, '\P{^Is_East_Asian_Width= Neutral}', "");
Error('\p{Is_Ea=:=-	N}');
Error('\P{Is_Ea=:=-	N}');
Expect(1, 918000, '\p{Is_Ea=n}', "");
Expect(0, 918000, '\p{^Is_Ea=n}', "");
Expect(0, 918000, '\P{Is_Ea=n}', "");
Expect(1, 918000, '\P{^Is_Ea=n}', "");
Expect(0, 1114109, '\p{Is_Ea=n}', "");
Expect(1, 1114109, '\p{^Is_Ea=n}', "");
Expect(1, 1114109, '\P{Is_Ea=n}', "");
Expect(0, 1114109, '\P{^Is_Ea=n}', "");
Expect(1, 918000, '\p{Is_Ea=	N}', "");
Expect(0, 918000, '\p{^Is_Ea=	N}', "");
Expect(0, 918000, '\P{Is_Ea=	N}', "");
Expect(1, 918000, '\P{^Is_Ea=	N}', "");
Expect(0, 1114109, '\p{Is_Ea=	N}', "");
Expect(1, 1114109, '\p{^Is_Ea=	N}', "");
Expect(1, 1114109, '\P{Is_Ea=	N}', "");
Expect(0, 1114109, '\P{^Is_Ea=	N}', "");
Error('\p{East_Asian_Width=/a/ halfwidth}');
Error('\P{East_Asian_Width=/a/ halfwidth}');
Expect(1, 65518, '\p{East_Asian_Width=halfwidth}', "");
Expect(0, 65518, '\p{^East_Asian_Width=halfwidth}', "");
Expect(0, 65518, '\P{East_Asian_Width=halfwidth}', "");
Expect(1, 65518, '\P{^East_Asian_Width=halfwidth}', "");
Expect(0, 65519, '\p{East_Asian_Width=halfwidth}', "");
Expect(1, 65519, '\p{^East_Asian_Width=halfwidth}', "");
Expect(1, 65519, '\P{East_Asian_Width=halfwidth}', "");
Expect(0, 65519, '\P{^East_Asian_Width=halfwidth}', "");
Expect(1, 65518, '\p{East_Asian_Width=--HALFWIDTH}', "");
Expect(0, 65518, '\p{^East_Asian_Width=--HALFWIDTH}', "");
Expect(0, 65518, '\P{East_Asian_Width=--HALFWIDTH}', "");
Expect(1, 65518, '\P{^East_Asian_Width=--HALFWIDTH}', "");
Expect(0, 65519, '\p{East_Asian_Width=--HALFWIDTH}', "");
Expect(1, 65519, '\p{^East_Asian_Width=--HALFWIDTH}', "");
Expect(1, 65519, '\P{East_Asian_Width=--HALFWIDTH}', "");
Expect(0, 65519, '\P{^East_Asian_Width=--HALFWIDTH}', "");
Error('\p{Ea=_H/a/}');
Error('\P{Ea=_H/a/}');
Expect(1, 65518, '\p{Ea=h}', "");
Expect(0, 65518, '\p{^Ea=h}', "");
Expect(0, 65518, '\P{Ea=h}', "");
Expect(1, 65518, '\P{^Ea=h}', "");
Expect(0, 65519, '\p{Ea=h}', "");
Expect(1, 65519, '\p{^Ea=h}', "");
Expect(1, 65519, '\P{Ea=h}', "");
Expect(0, 65519, '\P{^Ea=h}', "");
Expect(1, 65518, '\p{Ea=_H}', "");
Expect(0, 65518, '\p{^Ea=_H}', "");
Expect(0, 65518, '\P{Ea=_H}', "");
Expect(1, 65518, '\P{^Ea=_H}', "");
Expect(0, 65519, '\p{Ea=_H}', "");
Expect(1, 65519, '\p{^Ea=_H}', "");
Expect(1, 65519, '\P{Ea=_H}', "");
Expect(0, 65519, '\P{^Ea=_H}', "");
Error('\p{Is_East_Asian_Width=/a/-halfwidth}');
Error('\P{Is_East_Asian_Width=/a/-halfwidth}');
Expect(1, 65518, '\p{Is_East_Asian_Width=halfwidth}', "");
Expect(0, 65518, '\p{^Is_East_Asian_Width=halfwidth}', "");
Expect(0, 65518, '\P{Is_East_Asian_Width=halfwidth}', "");
Expect(1, 65518, '\P{^Is_East_Asian_Width=halfwidth}', "");
Expect(0, 65519, '\p{Is_East_Asian_Width=halfwidth}', "");
Expect(1, 65519, '\p{^Is_East_Asian_Width=halfwidth}', "");
Expect(1, 65519, '\P{Is_East_Asian_Width=halfwidth}', "");
Expect(0, 65519, '\P{^Is_East_Asian_Width=halfwidth}', "");
Expect(1, 65518, '\p{Is_East_Asian_Width=_halfwidth}', "");
Expect(0, 65518, '\p{^Is_East_Asian_Width=_halfwidth}', "");
Expect(0, 65518, '\P{Is_East_Asian_Width=_halfwidth}', "");
Expect(1, 65518, '\P{^Is_East_Asian_Width=_halfwidth}', "");
Expect(0, 65519, '\p{Is_East_Asian_Width=_halfwidth}', "");
Expect(1, 65519, '\p{^Is_East_Asian_Width=_halfwidth}', "");
Expect(1, 65519, '\P{Is_East_Asian_Width=_halfwidth}', "");
Expect(0, 65519, '\P{^Is_East_Asian_Width=_halfwidth}', "");
Error('\p{Is_Ea= :=h}');
Error('\P{Is_Ea= :=h}');
Expect(1, 65518, '\p{Is_Ea=h}', "");
Expect(0, 65518, '\p{^Is_Ea=h}', "");
Expect(0, 65518, '\P{Is_Ea=h}', "");
Expect(1, 65518, '\P{^Is_Ea=h}', "");
Expect(0, 65519, '\p{Is_Ea=h}', "");
Expect(1, 65519, '\p{^Is_Ea=h}', "");
Expect(1, 65519, '\P{Is_Ea=h}', "");
Expect(0, 65519, '\P{^Is_Ea=h}', "");
Expect(1, 65518, '\p{Is_Ea:  	h}', "");
Expect(0, 65518, '\p{^Is_Ea:  	h}', "");
Expect(0, 65518, '\P{Is_Ea:  	h}', "");
Expect(1, 65518, '\P{^Is_Ea:  	h}', "");
Expect(0, 65519, '\p{Is_Ea:  	h}', "");
Expect(1, 65519, '\p{^Is_Ea:  	h}', "");
Expect(1, 65519, '\P{Is_Ea:  	h}', "");
Expect(0, 65519, '\P{^Is_Ea:  	h}', "");
Error('\p{Quotation_Mark=/a/yes}');
Error('\P{Quotation_Mark=/a/yes}');
Expect(1, 65379, '\p{Quotation_Mark=yes}', "");
Expect(0, 65379, '\p{^Quotation_Mark=yes}', "");
Expect(0, 65379, '\P{Quotation_Mark=yes}', "");
Expect(1, 65379, '\P{^Quotation_Mark=yes}', "");
Expect(0, 65380, '\p{Quotation_Mark=yes}', "");
Expect(1, 65380, '\p{^Quotation_Mark=yes}', "");
Expect(1, 65380, '\P{Quotation_Mark=yes}', "");
Expect(0, 65380, '\P{^Quotation_Mark=yes}', "");
Expect(1, 65379, '\p{Quotation_Mark=		Yes}', "");
Expect(0, 65379, '\p{^Quotation_Mark=		Yes}', "");
Expect(0, 65379, '\P{Quotation_Mark=		Yes}', "");
Expect(1, 65379, '\P{^Quotation_Mark=		Yes}', "");
Expect(0, 65380, '\p{Quotation_Mark=		Yes}', "");
Expect(1, 65380, '\p{^Quotation_Mark=		Yes}', "");
Expect(1, 65380, '\P{Quotation_Mark=		Yes}', "");
Expect(0, 65380, '\P{^Quotation_Mark=		Yes}', "");
Error('\p{QMark= Y/a/}');
Error('\P{QMark= Y/a/}');
Expect(1, 65379, '\p{QMark=y}', "");
Expect(0, 65379, '\p{^QMark=y}', "");
Expect(0, 65379, '\P{QMark=y}', "");
Expect(1, 65379, '\P{^QMark=y}', "");
Expect(0, 65380, '\p{QMark=y}', "");
Expect(1, 65380, '\p{^QMark=y}', "");
Expect(1, 65380, '\P{QMark=y}', "");
Expect(0, 65380, '\P{^QMark=y}', "");
Expect(1, 65379, '\p{QMark=		Y}', "");
Expect(0, 65379, '\p{^QMark=		Y}', "");
Expect(0, 65379, '\P{QMark=		Y}', "");
Expect(1, 65379, '\P{^QMark=		Y}', "");
Expect(0, 65380, '\p{QMark=		Y}', "");
Expect(1, 65380, '\p{^QMark=		Y}', "");
Expect(1, 65380, '\P{QMark=		Y}', "");
Expect(0, 65380, '\P{^QMark=		Y}', "");
Error('\p{Is_Quotation_Mark=:=	-T}');
Error('\P{Is_Quotation_Mark=:=	-T}');
Expect(1, 65379, '\p{Is_Quotation_Mark:t}', "");
Expect(0, 65379, '\p{^Is_Quotation_Mark:t}', "");
Expect(0, 65379, '\P{Is_Quotation_Mark:t}', "");
Expect(1, 65379, '\P{^Is_Quotation_Mark:t}', "");
Expect(0, 65380, '\p{Is_Quotation_Mark:t}', "");
Expect(1, 65380, '\p{^Is_Quotation_Mark:t}', "");
Expect(1, 65380, '\P{Is_Quotation_Mark:t}', "");
Expect(0, 65380, '\P{^Is_Quotation_Mark:t}', "");
Expect(1, 65379, '\p{Is_Quotation_Mark=_ t}', "");
Expect(0, 65379, '\p{^Is_Quotation_Mark=_ t}', "");
Expect(0, 65379, '\P{Is_Quotation_Mark=_ t}', "");
Expect(1, 65379, '\P{^Is_Quotation_Mark=_ t}', "");
Expect(0, 65380, '\p{Is_Quotation_Mark=_ t}', "");
Expect(1, 65380, '\p{^Is_Quotation_Mark=_ t}', "");
Expect(1, 65380, '\P{Is_Quotation_Mark=_ t}', "");
Expect(0, 65380, '\P{^Is_Quotation_Mark=_ t}', "");
Error('\p{Is_QMark=-true:=}');
Error('\P{Is_QMark=-true:=}');
Expect(1, 65379, '\p{Is_QMark=true}', "");
Expect(0, 65379, '\p{^Is_QMark=true}', "");
Expect(0, 65379, '\P{Is_QMark=true}', "");
Expect(1, 65379, '\P{^Is_QMark=true}', "");
Expect(0, 65380, '\p{Is_QMark=true}', "");
Expect(1, 65380, '\p{^Is_QMark=true}', "");
Expect(1, 65380, '\P{Is_QMark=true}', "");
Expect(0, 65380, '\P{^Is_QMark=true}', "");
Expect(1, 65379, '\p{Is_QMark=		TRUE}', "");
Expect(0, 65379, '\p{^Is_QMark=		TRUE}', "");
Expect(0, 65379, '\P{Is_QMark=		TRUE}', "");
Expect(1, 65379, '\P{^Is_QMark=		TRUE}', "");
Expect(0, 65380, '\p{Is_QMark=		TRUE}', "");
Expect(1, 65380, '\p{^Is_QMark=		TRUE}', "");
Expect(1, 65380, '\P{Is_QMark=		TRUE}', "");
Expect(0, 65380, '\P{^Is_QMark=		TRUE}', "");
Error('\p{Quotation_Mark=_NO/a/}');
Error('\P{Quotation_Mark=_NO/a/}');
Expect(1, 65380, '\p{Quotation_Mark=no}', "");
Expect(0, 65380, '\p{^Quotation_Mark=no}', "");
Expect(0, 65380, '\P{Quotation_Mark=no}', "");
Expect(1, 65380, '\P{^Quotation_Mark=no}', "");
Expect(0, 65379, '\p{Quotation_Mark=no}', "");
Expect(1, 65379, '\p{^Quotation_Mark=no}', "");
Expect(1, 65379, '\P{Quotation_Mark=no}', "");
Expect(0, 65379, '\P{^Quotation_Mark=no}', "");
Expect(1, 65380, '\p{Quotation_Mark=	_No}', "");
Expect(0, 65380, '\p{^Quotation_Mark=	_No}', "");
Expect(0, 65380, '\P{Quotation_Mark=	_No}', "");
Expect(1, 65380, '\P{^Quotation_Mark=	_No}', "");
Expect(0, 65379, '\p{Quotation_Mark=	_No}', "");
Expect(1, 65379, '\p{^Quotation_Mark=	_No}', "");
Expect(1, 65379, '\P{Quotation_Mark=	_No}', "");
Expect(0, 65379, '\P{^Quotation_Mark=	_No}', "");
Error('\p{QMark=:=	-n}');
Error('\P{QMark=:=	-n}');
Expect(1, 65380, '\p{QMark=n}', "");
Expect(0, 65380, '\p{^QMark=n}', "");
Expect(0, 65380, '\P{QMark=n}', "");
Expect(1, 65380, '\P{^QMark=n}', "");
Expect(0, 65379, '\p{QMark=n}', "");
Expect(1, 65379, '\p{^QMark=n}', "");
Expect(1, 65379, '\P{QMark=n}', "");
Expect(0, 65379, '\P{^QMark=n}', "");
Expect(1, 65380, '\p{QMark=	n}', "");
Expect(0, 65380, '\p{^QMark=	n}', "");
Expect(0, 65380, '\P{QMark=	n}', "");
Expect(1, 65380, '\P{^QMark=	n}', "");
Expect(0, 65379, '\p{QMark=	n}', "");
Expect(1, 65379, '\p{^QMark=	n}', "");
Expect(1, 65379, '\P{QMark=	n}', "");
Expect(0, 65379, '\P{^QMark=	n}', "");
Error('\p{Is_Quotation_Mark=_F/a/}');
Error('\P{Is_Quotation_Mark=_F/a/}');
Expect(1, 65380, '\p{Is_Quotation_Mark=f}', "");
Expect(0, 65380, '\p{^Is_Quotation_Mark=f}', "");
Expect(0, 65380, '\P{Is_Quotation_Mark=f}', "");
Expect(1, 65380, '\P{^Is_Quotation_Mark=f}', "");
Expect(0, 65379, '\p{Is_Quotation_Mark=f}', "");
Expect(1, 65379, '\p{^Is_Quotation_Mark=f}', "");
Expect(1, 65379, '\P{Is_Quotation_Mark=f}', "");
Expect(0, 65379, '\P{^Is_Quotation_Mark=f}', "");
Expect(1, 65380, '\p{Is_Quotation_Mark:-F}', "");
Expect(0, 65380, '\p{^Is_Quotation_Mark:-F}', "");
Expect(0, 65380, '\P{Is_Quotation_Mark:-F}', "");
Expect(1, 65380, '\P{^Is_Quotation_Mark:-F}', "");
Expect(0, 65379, '\p{Is_Quotation_Mark:-F}', "");
Expect(1, 65379, '\p{^Is_Quotation_Mark:-F}', "");
Expect(1, 65379, '\P{Is_Quotation_Mark:-F}', "");
Expect(0, 65379, '\P{^Is_Quotation_Mark:-F}', "");
Error('\p{Is_QMark=-:=False}');
Error('\P{Is_QMark=-:=False}');
Expect(1, 65380, '\p{Is_QMark=false}', "");
Expect(0, 65380, '\p{^Is_QMark=false}', "");
Expect(0, 65380, '\P{Is_QMark=false}', "");
Expect(1, 65380, '\P{^Is_QMark=false}', "");
Expect(0, 65379, '\p{Is_QMark=false}', "");
Expect(1, 65379, '\p{^Is_QMark=false}', "");
Expect(1, 65379, '\P{Is_QMark=false}', "");
Expect(0, 65379, '\P{^Is_QMark=false}', "");
Expect(1, 65380, '\p{Is_QMark=_False}', "");
Expect(0, 65380, '\p{^Is_QMark=_False}', "");
Expect(0, 65380, '\P{Is_QMark=_False}', "");
Expect(1, 65380, '\P{^Is_QMark=_False}', "");
Expect(0, 65379, '\p{Is_QMark=_False}', "");
Expect(1, 65379, '\p{^Is_QMark=_False}', "");
Expect(1, 65379, '\P{Is_QMark=_False}', "");
Expect(0, 65379, '\P{^Is_QMark=_False}', "");
Error('\p{Grapheme_Link=Yes}');
Error('\P{Grapheme_Link=Yes}');
Error('\p{Gr_Link=Y}');
Error('\P{Gr_Link=Y}');
Error('\p{Is_Grapheme_Link=T}');
Error('\P{Is_Grapheme_Link=T}');
Error('\p{Is_Gr_Link=True}');
Error('\P{Is_Gr_Link=True}');
Error('\p{Grapheme_Link=No}');
Error('\P{Grapheme_Link=No}');
Error('\p{Gr_Link=N}');
Error('\P{Gr_Link=N}');
Error('\p{Is_Grapheme_Link=F}');
Error('\P{Is_Grapheme_Link=F}');
Error('\p{Is_Gr_Link=False}');
Error('\P{Is_Gr_Link=False}');
Error('\p{Bidi_Control= Yes/a/}');
Error('\P{Bidi_Control= Yes/a/}');
Expect(1, 8238, '\p{Bidi_Control=yes}', "");
Expect(0, 8238, '\p{^Bidi_Control=yes}', "");
Expect(0, 8238, '\P{Bidi_Control=yes}', "");
Expect(1, 8238, '\P{^Bidi_Control=yes}', "");
Expect(0, 8239, '\p{Bidi_Control=yes}', "");
Expect(1, 8239, '\p{^Bidi_Control=yes}', "");
Expect(1, 8239, '\P{Bidi_Control=yes}', "");
Expect(0, 8239, '\P{^Bidi_Control=yes}', "");
Expect(1, 8238, '\p{Bidi_Control=  yes}', "");
Expect(0, 8238, '\p{^Bidi_Control=  yes}', "");
Expect(0, 8238, '\P{Bidi_Control=  yes}', "");
Expect(1, 8238, '\P{^Bidi_Control=  yes}', "");
Expect(0, 8239, '\p{Bidi_Control=  yes}', "");
Expect(1, 8239, '\p{^Bidi_Control=  yes}', "");
Expect(1, 8239, '\P{Bidi_Control=  yes}', "");
Expect(0, 8239, '\P{^Bidi_Control=  yes}', "");
Error('\p{Bidi_C=/a/	y}');
Error('\P{Bidi_C=/a/	y}');
Expect(1, 8238, '\p{Bidi_C=y}', "");
Expect(0, 8238, '\p{^Bidi_C=y}', "");
Expect(0, 8238, '\P{Bidi_C=y}', "");
Expect(1, 8238, '\P{^Bidi_C=y}', "");
Expect(0, 8239, '\p{Bidi_C=y}', "");
Expect(1, 8239, '\p{^Bidi_C=y}', "");
Expect(1, 8239, '\P{Bidi_C=y}', "");
Expect(0, 8239, '\P{^Bidi_C=y}', "");
Expect(1, 8238, '\p{Bidi_C:   		Y}', "");
Expect(0, 8238, '\p{^Bidi_C:   		Y}', "");
Expect(0, 8238, '\P{Bidi_C:   		Y}', "");
Expect(1, 8238, '\P{^Bidi_C:   		Y}', "");
Expect(0, 8239, '\p{Bidi_C:   		Y}', "");
Expect(1, 8239, '\p{^Bidi_C:   		Y}', "");
Expect(1, 8239, '\P{Bidi_C:   		Y}', "");
Expect(0, 8239, '\P{^Bidi_C:   		Y}', "");
Error('\p{Is_Bidi_Control=/a/-	t}');
Error('\P{Is_Bidi_Control=/a/-	t}');
Expect(1, 8238, '\p{Is_Bidi_Control=t}', "");
Expect(0, 8238, '\p{^Is_Bidi_Control=t}', "");
Expect(0, 8238, '\P{Is_Bidi_Control=t}', "");
Expect(1, 8238, '\P{^Is_Bidi_Control=t}', "");
Expect(0, 8239, '\p{Is_Bidi_Control=t}', "");
Expect(1, 8239, '\p{^Is_Bidi_Control=t}', "");
Expect(1, 8239, '\P{Is_Bidi_Control=t}', "");
Expect(0, 8239, '\P{^Is_Bidi_Control=t}', "");
Expect(1, 8238, '\p{Is_Bidi_Control=-T}', "");
Expect(0, 8238, '\p{^Is_Bidi_Control=-T}', "");
Expect(0, 8238, '\P{Is_Bidi_Control=-T}', "");
Expect(1, 8238, '\P{^Is_Bidi_Control=-T}', "");
Expect(0, 8239, '\p{Is_Bidi_Control=-T}', "");
Expect(1, 8239, '\p{^Is_Bidi_Control=-T}', "");
Expect(1, 8239, '\P{Is_Bidi_Control=-T}', "");
Expect(0, 8239, '\P{^Is_Bidi_Control=-T}', "");
Error('\p{Is_Bidi_C=:=	true}');
Error('\P{Is_Bidi_C=:=	true}');
Expect(1, 8238, '\p{Is_Bidi_C=true}', "");
Expect(0, 8238, '\p{^Is_Bidi_C=true}', "");
Expect(0, 8238, '\P{Is_Bidi_C=true}', "");
Expect(1, 8238, '\P{^Is_Bidi_C=true}', "");
Expect(0, 8239, '\p{Is_Bidi_C=true}', "");
Expect(1, 8239, '\p{^Is_Bidi_C=true}', "");
Expect(1, 8239, '\P{Is_Bidi_C=true}', "");
Expect(0, 8239, '\P{^Is_Bidi_C=true}', "");
Expect(1, 8238, '\p{Is_Bidi_C=_-True}', "");
Expect(0, 8238, '\p{^Is_Bidi_C=_-True}', "");
Expect(0, 8238, '\P{Is_Bidi_C=_-True}', "");
Expect(1, 8238, '\P{^Is_Bidi_C=_-True}', "");
Expect(0, 8239, '\p{Is_Bidi_C=_-True}', "");
Expect(1, 8239, '\p{^Is_Bidi_C=_-True}', "");
Expect(1, 8239, '\P{Is_Bidi_C=_-True}', "");
Expect(0, 8239, '\P{^Is_Bidi_C=_-True}', "");
Error('\p{Bidi_Control:	:=	no}');
Error('\P{Bidi_Control:	:=	no}');
Expect(1, 8239, '\p{Bidi_Control=no}', "");
Expect(0, 8239, '\p{^Bidi_Control=no}', "");
Expect(0, 8239, '\P{Bidi_Control=no}', "");
Expect(1, 8239, '\P{^Bidi_Control=no}', "");
Expect(0, 8238, '\p{Bidi_Control=no}', "");
Expect(1, 8238, '\p{^Bidi_Control=no}', "");
Expect(1, 8238, '\P{Bidi_Control=no}', "");
Expect(0, 8238, '\P{^Bidi_Control=no}', "");
Expect(1, 8239, '\p{Bidi_Control=_no}', "");
Expect(0, 8239, '\p{^Bidi_Control=_no}', "");
Expect(0, 8239, '\P{Bidi_Control=_no}', "");
Expect(1, 8239, '\P{^Bidi_Control=_no}', "");
Expect(0, 8238, '\p{Bidi_Control=_no}', "");
Expect(1, 8238, '\p{^Bidi_Control=_no}', "");
Expect(1, 8238, '\P{Bidi_Control=_no}', "");
Expect(0, 8238, '\P{^Bidi_Control=_no}', "");
Error('\p{Bidi_C= -n/a/}');
Error('\P{Bidi_C= -n/a/}');
Expect(1, 8239, '\p{Bidi_C=n}', "");
Expect(0, 8239, '\p{^Bidi_C=n}', "");
Expect(0, 8239, '\P{Bidi_C=n}', "");
Expect(1, 8239, '\P{^Bidi_C=n}', "");
Expect(0, 8238, '\p{Bidi_C=n}', "");
Expect(1, 8238, '\p{^Bidi_C=n}', "");
Expect(1, 8238, '\P{Bidi_C=n}', "");
Expect(0, 8238, '\P{^Bidi_C=n}', "");
Expect(1, 8239, '\p{Bidi_C= _N}', "");
Expect(0, 8239, '\p{^Bidi_C= _N}', "");
Expect(0, 8239, '\P{Bidi_C= _N}', "");
Expect(1, 8239, '\P{^Bidi_C= _N}', "");
Expect(0, 8238, '\p{Bidi_C= _N}', "");
Expect(1, 8238, '\p{^Bidi_C= _N}', "");
Expect(1, 8238, '\P{Bidi_C= _N}', "");
Expect(0, 8238, '\P{^Bidi_C= _N}', "");
Error('\p{Is_Bidi_Control=/a/ 	F}');
Error('\P{Is_Bidi_Control=/a/ 	F}');
Expect(1, 8239, '\p{Is_Bidi_Control=f}', "");
Expect(0, 8239, '\p{^Is_Bidi_Control=f}', "");
Expect(0, 8239, '\P{Is_Bidi_Control=f}', "");
Expect(1, 8239, '\P{^Is_Bidi_Control=f}', "");
Expect(0, 8238, '\p{Is_Bidi_Control=f}', "");
Expect(1, 8238, '\p{^Is_Bidi_Control=f}', "");
Expect(1, 8238, '\P{Is_Bidi_Control=f}', "");
Expect(0, 8238, '\P{^Is_Bidi_Control=f}', "");
Expect(1, 8239, '\p{Is_Bidi_Control=_f}', "");
Expect(0, 8239, '\p{^Is_Bidi_Control=_f}', "");
Expect(0, 8239, '\P{Is_Bidi_Control=_f}', "");
Expect(1, 8239, '\P{^Is_Bidi_Control=_f}', "");
Expect(0, 8238, '\p{Is_Bidi_Control=_f}', "");
Expect(1, 8238, '\p{^Is_Bidi_Control=_f}', "");
Expect(1, 8238, '\P{Is_Bidi_Control=_f}', "");
Expect(0, 8238, '\P{^Is_Bidi_Control=_f}', "");
Error('\p{Is_Bidi_C: _FALSE:=}');
Error('\P{Is_Bidi_C: _FALSE:=}');
Expect(1, 8239, '\p{Is_Bidi_C=false}', "");
Expect(0, 8239, '\p{^Is_Bidi_C=false}', "");
Expect(0, 8239, '\P{Is_Bidi_C=false}', "");
Expect(1, 8239, '\P{^Is_Bidi_C=false}', "");
Expect(0, 8238, '\p{Is_Bidi_C=false}', "");
Expect(1, 8238, '\p{^Is_Bidi_C=false}', "");
Expect(1, 8238, '\P{Is_Bidi_C=false}', "");
Expect(0, 8238, '\P{^Is_Bidi_C=false}', "");
Expect(1, 8239, '\p{Is_Bidi_C= false}', "");
Expect(0, 8239, '\p{^Is_Bidi_C= false}', "");
Expect(0, 8239, '\P{Is_Bidi_C= false}', "");
Expect(1, 8239, '\P{^Is_Bidi_C= false}', "");
Expect(0, 8238, '\p{Is_Bidi_C= false}', "");
Expect(1, 8238, '\p{^Is_Bidi_C= false}', "");
Expect(1, 8238, '\P{Is_Bidi_C= false}', "");
Expect(0, 8238, '\P{^Is_Bidi_C= false}', "");
Error('\p{Script=  new_TAI_Lue:=}');
Error('\P{Script=  new_TAI_Lue:=}');
Expect(1, 6623, '\p{Script=newtailue}', "");
Expect(0, 6623, '\p{^Script=newtailue}', "");
Expect(0, 6623, '\P{Script=newtailue}', "");
Expect(1, 6623, '\P{^Script=newtailue}', "");
Expect(0, 6624, '\p{Script=newtailue}', "");
Expect(1, 6624, '\p{^Script=newtailue}', "");
Expect(1, 6624, '\P{Script=newtailue}', "");
Expect(0, 6624, '\P{^Script=newtailue}', "");
Expect(1, 6623, '\p{Script:	new_tai_LUE}', "");
Expect(0, 6623, '\p{^Script:	new_tai_LUE}', "");
Expect(0, 6623, '\P{Script:	new_tai_LUE}', "");
Expect(1, 6623, '\P{^Script:	new_tai_LUE}', "");
Expect(0, 6624, '\p{Script:	new_tai_LUE}', "");
Expect(1, 6624, '\p{^Script:	new_tai_LUE}', "");
Expect(1, 6624, '\P{Script:	new_tai_LUE}', "");
Expect(0, 6624, '\P{^Script:	new_tai_LUE}', "");
Error('\p{Sc=_:=TALU}');
Error('\P{Sc=_:=TALU}');
Expect(1, 6623, '\p{Sc=talu}', "");
Expect(0, 6623, '\p{^Sc=talu}', "");
Expect(0, 6623, '\P{Sc=talu}', "");
Expect(1, 6623, '\P{^Sc=talu}', "");
Expect(0, 6624, '\p{Sc=talu}', "");
Expect(1, 6624, '\p{^Sc=talu}', "");
Expect(1, 6624, '\P{Sc=talu}', "");
Expect(0, 6624, '\P{^Sc=talu}', "");
Expect(1, 6623, '\p{Sc=--TALU}', "");
Expect(0, 6623, '\p{^Sc=--TALU}', "");
Expect(0, 6623, '\P{Sc=--TALU}', "");
Expect(1, 6623, '\P{^Sc=--TALU}', "");
Expect(0, 6624, '\p{Sc=--TALU}', "");
Expect(1, 6624, '\p{^Sc=--TALU}', "");
Expect(1, 6624, '\P{Sc=--TALU}', "");
Expect(0, 6624, '\P{^Sc=--TALU}', "");
Error('\p{Is_Script=/a/	 new_Tai_lue}');
Error('\P{Is_Script=/a/	 new_Tai_lue}');
Expect(1, 6623, '\p{Is_Script:newtailue}', "");
Expect(0, 6623, '\p{^Is_Script:newtailue}', "");
Expect(0, 6623, '\P{Is_Script:newtailue}', "");
Expect(1, 6623, '\P{^Is_Script:newtailue}', "");
Expect(0, 6624, '\p{Is_Script:newtailue}', "");
Expect(1, 6624, '\p{^Is_Script:newtailue}', "");
Expect(1, 6624, '\P{Is_Script:newtailue}', "");
Expect(0, 6624, '\P{^Is_Script:newtailue}', "");
Expect(1, 6623, '\p{Is_Script=_ New_tai_LUE}', "");
Expect(0, 6623, '\p{^Is_Script=_ New_tai_LUE}', "");
Expect(0, 6623, '\P{Is_Script=_ New_tai_LUE}', "");
Expect(1, 6623, '\P{^Is_Script=_ New_tai_LUE}', "");
Expect(0, 6624, '\p{Is_Script=_ New_tai_LUE}', "");
Expect(1, 6624, '\p{^Is_Script=_ New_tai_LUE}', "");
Expect(1, 6624, '\P{Is_Script=_ New_tai_LUE}', "");
Expect(0, 6624, '\P{^Is_Script=_ New_tai_LUE}', "");
Error('\p{Is_Sc=-:=talu}');
Error('\P{Is_Sc=-:=talu}');
Expect(1, 6623, '\p{Is_Sc=talu}', "");
Expect(0, 6623, '\p{^Is_Sc=talu}', "");
Expect(0, 6623, '\P{Is_Sc=talu}', "");
Expect(1, 6623, '\P{^Is_Sc=talu}', "");
Expect(0, 6624, '\p{Is_Sc=talu}', "");
Expect(1, 6624, '\p{^Is_Sc=talu}', "");
Expect(1, 6624, '\P{Is_Sc=talu}', "");
Expect(0, 6624, '\P{^Is_Sc=talu}', "");
Expect(1, 6623, '\p{Is_Sc=	Talu}', "");
Expect(0, 6623, '\p{^Is_Sc=	Talu}', "");
Expect(0, 6623, '\P{Is_Sc=	Talu}', "");
Expect(1, 6623, '\P{^Is_Sc=	Talu}', "");
Expect(0, 6624, '\p{Is_Sc=	Talu}', "");
Expect(1, 6624, '\p{^Is_Sc=	Talu}', "");
Expect(1, 6624, '\P{Is_Sc=	Talu}', "");
Expect(0, 6624, '\P{^Is_Sc=	Talu}', "");
Error('\p{Script=-/a/gothic}');
Error('\P{Script=-/a/gothic}');
Expect(1, 66378, '\p{Script=gothic}', "");
Expect(0, 66378, '\p{^Script=gothic}', "");
Expect(0, 66378, '\P{Script=gothic}', "");
Expect(1, 66378, '\P{^Script=gothic}', "");
Expect(0, 66379, '\p{Script=gothic}', "");
Expect(1, 66379, '\p{^Script=gothic}', "");
Expect(1, 66379, '\P{Script=gothic}', "");
Expect(0, 66379, '\P{^Script=gothic}', "");
Expect(1, 66378, '\p{Script= 	gothic}', "");
Expect(0, 66378, '\p{^Script= 	gothic}', "");
Expect(0, 66378, '\P{Script= 	gothic}', "");
Expect(1, 66378, '\P{^Script= 	gothic}', "");
Expect(0, 66379, '\p{Script= 	gothic}', "");
Expect(1, 66379, '\p{^Script= 	gothic}', "");
Expect(1, 66379, '\P{Script= 	gothic}', "");
Expect(0, 66379, '\P{^Script= 	gothic}', "");
Error('\p{Sc=:= 	GOTH}');
Error('\P{Sc=:= 	GOTH}');
Expect(1, 66378, '\p{Sc:   goth}', "");
Expect(0, 66378, '\p{^Sc:   goth}', "");
Expect(0, 66378, '\P{Sc:   goth}', "");
Expect(1, 66378, '\P{^Sc:   goth}', "");
Expect(0, 66379, '\p{Sc:   goth}', "");
Expect(1, 66379, '\p{^Sc:   goth}', "");
Expect(1, 66379, '\P{Sc:   goth}', "");
Expect(0, 66379, '\P{^Sc:   goth}', "");
Expect(1, 66378, '\p{Sc=-Goth}', "");
Expect(0, 66378, '\p{^Sc=-Goth}', "");
Expect(0, 66378, '\P{Sc=-Goth}', "");
Expect(1, 66378, '\P{^Sc=-Goth}', "");
Expect(0, 66379, '\p{Sc=-Goth}', "");
Expect(1, 66379, '\p{^Sc=-Goth}', "");
Expect(1, 66379, '\P{Sc=-Goth}', "");
Expect(0, 66379, '\P{^Sc=-Goth}', "");
Error('\p{Is_Script=_ gothic:=}');
Error('\P{Is_Script=_ gothic:=}');
Expect(1, 66378, '\p{Is_Script=gothic}', "");
Expect(0, 66378, '\p{^Is_Script=gothic}', "");
Expect(0, 66378, '\P{Is_Script=gothic}', "");
Expect(1, 66378, '\P{^Is_Script=gothic}', "");
Expect(0, 66379, '\p{Is_Script=gothic}', "");
Expect(1, 66379, '\p{^Is_Script=gothic}', "");
Expect(1, 66379, '\P{Is_Script=gothic}', "");
Expect(0, 66379, '\P{^Is_Script=gothic}', "");
Expect(1, 66378, '\p{Is_Script=	Gothic}', "");
Expect(0, 66378, '\p{^Is_Script=	Gothic}', "");
Expect(0, 66378, '\P{Is_Script=	Gothic}', "");
Expect(1, 66378, '\P{^Is_Script=	Gothic}', "");
Expect(0, 66379, '\p{Is_Script=	Gothic}', "");
Expect(1, 66379, '\p{^Is_Script=	Gothic}', "");
Expect(1, 66379, '\P{Is_Script=	Gothic}', "");
Expect(0, 66379, '\P{^Is_Script=	Gothic}', "");
Error('\p{Is_Sc=:=-_GOTH}');
Error('\P{Is_Sc=:=-_GOTH}');
Expect(1, 66378, '\p{Is_Sc=goth}', "");
Expect(0, 66378, '\p{^Is_Sc=goth}', "");
Expect(0, 66378, '\P{Is_Sc=goth}', "");
Expect(1, 66378, '\P{^Is_Sc=goth}', "");
Expect(0, 66379, '\p{Is_Sc=goth}', "");
Expect(1, 66379, '\p{^Is_Sc=goth}', "");
Expect(1, 66379, '\P{Is_Sc=goth}', "");
Expect(0, 66379, '\P{^Is_Sc=goth}', "");
Expect(1, 66378, '\p{Is_Sc=	goth}', "");
Expect(0, 66378, '\p{^Is_Sc=	goth}', "");
Expect(0, 66378, '\P{Is_Sc=	goth}', "");
Expect(1, 66378, '\P{^Is_Sc=	goth}', "");
Expect(0, 66379, '\p{Is_Sc=	goth}', "");
Expect(1, 66379, '\p{^Is_Sc=	goth}', "");
Expect(1, 66379, '\P{Is_Sc=	goth}', "");
Expect(0, 66379, '\P{^Is_Sc=	goth}', "");
Error('\p{Script=/a/	_OLD_ITALIC}');
Error('\P{Script=/a/	_OLD_ITALIC}');
Expect(1, 66339, '\p{Script=olditalic}', "");
Expect(0, 66339, '\p{^Script=olditalic}', "");
Expect(0, 66339, '\P{Script=olditalic}', "");
Expect(1, 66339, '\P{^Script=olditalic}', "");
Expect(0, 66340, '\p{Script=olditalic}', "");
Expect(1, 66340, '\p{^Script=olditalic}', "");
Expect(1, 66340, '\P{Script=olditalic}', "");
Expect(0, 66340, '\P{^Script=olditalic}', "");
Expect(1, 66339, '\p{Script= -OLD_ITALIC}', "");
Expect(0, 66339, '\p{^Script= -OLD_ITALIC}', "");
Expect(0, 66339, '\P{Script= -OLD_ITALIC}', "");
Expect(1, 66339, '\P{^Script= -OLD_ITALIC}', "");
Expect(0, 66340, '\p{Script= -OLD_ITALIC}', "");
Expect(1, 66340, '\p{^Script= -OLD_ITALIC}', "");
Expect(1, 66340, '\P{Script= -OLD_ITALIC}', "");
Expect(0, 66340, '\P{^Script= -OLD_ITALIC}', "");
Error('\p{Sc=Ital:=}');
Error('\P{Sc=Ital:=}');
Expect(1, 66339, '\p{Sc=ital}', "");
Expect(0, 66339, '\p{^Sc=ital}', "");
Expect(0, 66339, '\P{Sc=ital}', "");
Expect(1, 66339, '\P{^Sc=ital}', "");
Expect(0, 66340, '\p{Sc=ital}', "");
Expect(1, 66340, '\p{^Sc=ital}', "");
Expect(1, 66340, '\P{Sc=ital}', "");
Expect(0, 66340, '\P{^Sc=ital}', "");
Expect(1, 66339, '\p{Sc=--ital}', "");
Expect(0, 66339, '\p{^Sc=--ital}', "");
Expect(0, 66339, '\P{Sc=--ital}', "");
Expect(1, 66339, '\P{^Sc=--ital}', "");
Expect(0, 66340, '\p{Sc=--ital}', "");
Expect(1, 66340, '\p{^Sc=--ital}', "");
Expect(1, 66340, '\P{Sc=--ital}', "");
Expect(0, 66340, '\P{^Sc=--ital}', "");
Error('\p{Is_Script=	OLD_Italic/a/}');
Error('\P{Is_Script=	OLD_Italic/a/}');
Expect(1, 66339, '\p{Is_Script=olditalic}', "");
Expect(0, 66339, '\p{^Is_Script=olditalic}', "");
Expect(0, 66339, '\P{Is_Script=olditalic}', "");
Expect(1, 66339, '\P{^Is_Script=olditalic}', "");
Expect(0, 66340, '\p{Is_Script=olditalic}', "");
Expect(1, 66340, '\p{^Is_Script=olditalic}', "");
Expect(1, 66340, '\P{Is_Script=olditalic}', "");
Expect(0, 66340, '\P{^Is_Script=olditalic}', "");
Expect(1, 66339, '\p{Is_Script=	 OLD_Italic}', "");
Expect(0, 66339, '\p{^Is_Script=	 OLD_Italic}', "");
Expect(0, 66339, '\P{Is_Script=	 OLD_Italic}', "");
Expect(1, 66339, '\P{^Is_Script=	 OLD_Italic}', "");
Expect(0, 66340, '\p{Is_Script=	 OLD_Italic}', "");
Expect(1, 66340, '\p{^Is_Script=	 OLD_Italic}', "");
Expect(1, 66340, '\P{Is_Script=	 OLD_Italic}', "");
Expect(0, 66340, '\P{^Is_Script=	 OLD_Italic}', "");
Error('\p{Is_Sc=-	ITAL/a/}');
Error('\P{Is_Sc=-	ITAL/a/}');
Expect(1, 66339, '\p{Is_Sc=ital}', "");
Expect(0, 66339, '\p{^Is_Sc=ital}', "");
Expect(0, 66339, '\P{Is_Sc=ital}', "");
Expect(1, 66339, '\P{^Is_Sc=ital}', "");
Expect(0, 66340, '\p{Is_Sc=ital}', "");
Expect(1, 66340, '\p{^Is_Sc=ital}', "");
Expect(1, 66340, '\P{Is_Sc=ital}', "");
Expect(0, 66340, '\P{^Is_Sc=ital}', "");
Expect(1, 66339, '\p{Is_Sc=	 Ital}', "");
Expect(0, 66339, '\p{^Is_Sc=	 Ital}', "");
Expect(0, 66339, '\P{Is_Sc=	 Ital}', "");
Expect(1, 66339, '\P{^Is_Sc=	 Ital}', "");
Expect(0, 66340, '\p{Is_Sc=	 Ital}', "");
Expect(1, 66340, '\p{^Is_Sc=	 Ital}', "");
Expect(1, 66340, '\P{Is_Sc=	 Ital}', "");
Expect(0, 66340, '\P{^Is_Sc=	 Ital}', "");
Error('\p{Script=/a/-_Bamum}');
Error('\P{Script=/a/-_Bamum}');
Expect(1, 92728, '\p{Script=bamum}', "");
Expect(0, 92728, '\p{^Script=bamum}', "");
Expect(0, 92728, '\P{Script=bamum}', "");
Expect(1, 92728, '\P{^Script=bamum}', "");
Expect(0, 92729, '\p{Script=bamum}', "");
Expect(1, 92729, '\p{^Script=bamum}', "");
Expect(1, 92729, '\P{Script=bamum}', "");
Expect(0, 92729, '\P{^Script=bamum}', "");
Expect(1, 92728, '\p{Script=- bamum}', "");
Expect(0, 92728, '\p{^Script=- bamum}', "");
Expect(0, 92728, '\P{Script=- bamum}', "");
Expect(1, 92728, '\P{^Script=- bamum}', "");
Expect(0, 92729, '\p{Script=- bamum}', "");
Expect(1, 92729, '\p{^Script=- bamum}', "");
Expect(1, 92729, '\P{Script=- bamum}', "");
Expect(0, 92729, '\P{^Script=- bamum}', "");
Error('\p{Sc=:=-Bamu}');
Error('\P{Sc=:=-Bamu}');
Expect(1, 92728, '\p{Sc:bamu}', "");
Expect(0, 92728, '\p{^Sc:bamu}', "");
Expect(0, 92728, '\P{Sc:bamu}', "");
Expect(1, 92728, '\P{^Sc:bamu}', "");
Expect(0, 92729, '\p{Sc:bamu}', "");
Expect(1, 92729, '\p{^Sc:bamu}', "");
Expect(1, 92729, '\P{Sc:bamu}', "");
Expect(0, 92729, '\P{^Sc:bamu}', "");
Expect(1, 92728, '\p{Sc: -_BAMU}', "");
Expect(0, 92728, '\p{^Sc: -_BAMU}', "");
Expect(0, 92728, '\P{Sc: -_BAMU}', "");
Expect(1, 92728, '\P{^Sc: -_BAMU}', "");
Expect(0, 92729, '\p{Sc: -_BAMU}', "");
Expect(1, 92729, '\p{^Sc: -_BAMU}', "");
Expect(1, 92729, '\P{Sc: -_BAMU}', "");
Expect(0, 92729, '\P{^Sc: -_BAMU}', "");
Error('\p{Is_Script=:=	 BAMUM}');
Error('\P{Is_Script=:=	 BAMUM}');
Expect(1, 92728, '\p{Is_Script=bamum}', "");
Expect(0, 92728, '\p{^Is_Script=bamum}', "");
Expect(0, 92728, '\P{Is_Script=bamum}', "");
Expect(1, 92728, '\P{^Is_Script=bamum}', "");
Expect(0, 92729, '\p{Is_Script=bamum}', "");
Expect(1, 92729, '\p{^Is_Script=bamum}', "");
Expect(1, 92729, '\P{Is_Script=bamum}', "");
Expect(0, 92729, '\P{^Is_Script=bamum}', "");
Expect(1, 92728, '\p{Is_Script=-	BAMUM}', "");
Expect(0, 92728, '\p{^Is_Script=-	BAMUM}', "");
Expect(0, 92728, '\P{Is_Script=-	BAMUM}', "");
Expect(1, 92728, '\P{^Is_Script=-	BAMUM}', "");
Expect(0, 92729, '\p{Is_Script=-	BAMUM}', "");
Expect(1, 92729, '\p{^Is_Script=-	BAMUM}', "");
Expect(1, 92729, '\P{Is_Script=-	BAMUM}', "");
Expect(0, 92729, '\P{^Is_Script=-	BAMUM}', "");
Error('\p{Is_Sc= /a/bamu}');
Error('\P{Is_Sc= /a/bamu}');
Expect(1, 92728, '\p{Is_Sc=bamu}', "");
Expect(0, 92728, '\p{^Is_Sc=bamu}', "");
Expect(0, 92728, '\P{Is_Sc=bamu}', "");
Expect(1, 92728, '\P{^Is_Sc=bamu}', "");
Expect(0, 92729, '\p{Is_Sc=bamu}', "");
Expect(1, 92729, '\p{^Is_Sc=bamu}', "");
Expect(1, 92729, '\P{Is_Sc=bamu}', "");
Expect(0, 92729, '\P{^Is_Sc=bamu}', "");
Expect(1, 92728, '\p{Is_Sc=		Bamu}', "");
Expect(0, 92728, '\p{^Is_Sc=		Bamu}', "");
Expect(0, 92728, '\P{Is_Sc=		Bamu}', "");
Expect(1, 92728, '\P{^Is_Sc=		Bamu}', "");
Expect(0, 92729, '\p{Is_Sc=		Bamu}', "");
Expect(1, 92729, '\p{^Is_Sc=		Bamu}', "");
Expect(1, 92729, '\P{Is_Sc=		Bamu}', "");
Expect(0, 92729, '\P{^Is_Sc=		Bamu}', "");
Error('\p{Script= :=Brahmi}');
Error('\P{Script= :=Brahmi}');
Expect(1, 69743, '\p{Script=brahmi}', "");
Expect(0, 69743, '\p{^Script=brahmi}', "");
Expect(0, 69743, '\P{Script=brahmi}', "");
Expect(1, 69743, '\P{^Script=brahmi}', "");
Expect(0, 69744, '\p{Script=brahmi}', "");
Expect(1, 69744, '\p{^Script=brahmi}', "");
Expect(1, 69744, '\P{Script=brahmi}', "");
Expect(0, 69744, '\P{^Script=brahmi}', "");
Expect(1, 69743, '\p{Script=- Brahmi}', "");
Expect(0, 69743, '\p{^Script=- Brahmi}', "");
Expect(0, 69743, '\P{Script=- Brahmi}', "");
Expect(1, 69743, '\P{^Script=- Brahmi}', "");
Expect(0, 69744, '\p{Script=- Brahmi}', "");
Expect(1, 69744, '\p{^Script=- Brahmi}', "");
Expect(1, 69744, '\P{Script=- Brahmi}', "");
Expect(0, 69744, '\P{^Script=- Brahmi}', "");
Error('\p{Sc= /a/brah}');
Error('\P{Sc= /a/brah}');
Expect(1, 69743, '\p{Sc=brah}', "");
Expect(0, 69743, '\p{^Sc=brah}', "");
Expect(0, 69743, '\P{Sc=brah}', "");
Expect(1, 69743, '\P{^Sc=brah}', "");
Expect(0, 69744, '\p{Sc=brah}', "");
Expect(1, 69744, '\p{^Sc=brah}', "");
Expect(1, 69744, '\P{Sc=brah}', "");
Expect(0, 69744, '\P{^Sc=brah}', "");
Expect(1, 69743, '\p{Sc=-	Brah}', "");
Expect(0, 69743, '\p{^Sc=-	Brah}', "");
Expect(0, 69743, '\P{Sc=-	Brah}', "");
Expect(1, 69743, '\P{^Sc=-	Brah}', "");
Expect(0, 69744, '\p{Sc=-	Brah}', "");
Expect(1, 69744, '\p{^Sc=-	Brah}', "");
Expect(1, 69744, '\P{Sc=-	Brah}', "");
Expect(0, 69744, '\P{^Sc=-	Brah}', "");
Error('\p{Is_Script=-	Brahmi/a/}');
Error('\P{Is_Script=-	Brahmi/a/}');
Expect(1, 69743, '\p{Is_Script:brahmi}', "");
Expect(0, 69743, '\p{^Is_Script:brahmi}', "");
Expect(0, 69743, '\P{Is_Script:brahmi}', "");
Expect(1, 69743, '\P{^Is_Script:brahmi}', "");
Expect(0, 69744, '\p{Is_Script:brahmi}', "");
Expect(1, 69744, '\p{^Is_Script:brahmi}', "");
Expect(1, 69744, '\P{Is_Script:brahmi}', "");
Expect(0, 69744, '\P{^Is_Script:brahmi}', "");
Expect(1, 69743, '\p{Is_Script=	_brahmi}', "");
Expect(0, 69743, '\p{^Is_Script=	_brahmi}', "");
Expect(0, 69743, '\P{Is_Script=	_brahmi}', "");
Expect(1, 69743, '\P{^Is_Script=	_brahmi}', "");
Expect(0, 69744, '\p{Is_Script=	_brahmi}', "");
Expect(1, 69744, '\p{^Is_Script=	_brahmi}', "");
Expect(1, 69744, '\P{Is_Script=	_brahmi}', "");
Expect(0, 69744, '\P{^Is_Script=	_brahmi}', "");
Error('\p{Is_Sc=--Brah:=}');
Error('\P{Is_Sc=--Brah:=}');
Expect(1, 69743, '\p{Is_Sc=brah}', "");
Expect(0, 69743, '\p{^Is_Sc=brah}', "");
Expect(0, 69743, '\P{Is_Sc=brah}', "");
Expect(1, 69743, '\P{^Is_Sc=brah}', "");
Expect(0, 69744, '\p{Is_Sc=brah}', "");
Expect(1, 69744, '\p{^Is_Sc=brah}', "");
Expect(1, 69744, '\P{Is_Sc=brah}', "");
Expect(0, 69744, '\P{^Is_Sc=brah}', "");
Expect(1, 69743, '\p{Is_Sc= Brah}', "");
Expect(0, 69743, '\p{^Is_Sc= Brah}', "");
Expect(0, 69743, '\P{Is_Sc= Brah}', "");
Expect(1, 69743, '\P{^Is_Sc= Brah}', "");
Expect(0, 69744, '\p{Is_Sc= Brah}', "");
Expect(1, 69744, '\p{^Is_Sc= Brah}', "");
Expect(1, 69744, '\P{Is_Sc= Brah}', "");
Expect(0, 69744, '\P{^Is_Sc= Brah}', "");
Error('\p{Script=/a/ Vai}');
Error('\P{Script=/a/ Vai}');
Expect(1, 42539, '\p{Script=vai}', "");
Expect(0, 42539, '\p{^Script=vai}', "");
Expect(0, 42539, '\P{Script=vai}', "");
Expect(1, 42539, '\P{^Script=vai}', "");
Expect(0, 42540, '\p{Script=vai}', "");
Expect(1, 42540, '\p{^Script=vai}', "");
Expect(1, 42540, '\P{Script=vai}', "");
Expect(0, 42540, '\P{^Script=vai}', "");
Expect(1, 42539, '\p{Script= Vai}', "");
Expect(0, 42539, '\p{^Script= Vai}', "");
Expect(0, 42539, '\P{Script= Vai}', "");
Expect(1, 42539, '\P{^Script= Vai}', "");
Expect(0, 42540, '\p{Script= Vai}', "");
Expect(1, 42540, '\p{^Script= Vai}', "");
Expect(1, 42540, '\P{Script= Vai}', "");
Expect(0, 42540, '\P{^Script= Vai}', "");
Error('\p{Sc=_Vaii:=}');
Error('\P{Sc=_Vaii:=}');
Expect(1, 42539, '\p{Sc=vaii}', "");
Expect(0, 42539, '\p{^Sc=vaii}', "");
Expect(0, 42539, '\P{Sc=vaii}', "");
Expect(1, 42539, '\P{^Sc=vaii}', "");
Expect(0, 42540, '\p{Sc=vaii}', "");
Expect(1, 42540, '\p{^Sc=vaii}', "");
Expect(1, 42540, '\P{Sc=vaii}', "");
Expect(0, 42540, '\P{^Sc=vaii}', "");
Expect(1, 42539, '\p{Sc=	 VAII}', "");
Expect(0, 42539, '\p{^Sc=	 VAII}', "");
Expect(0, 42539, '\P{Sc=	 VAII}', "");
Expect(1, 42539, '\P{^Sc=	 VAII}', "");
Expect(0, 42540, '\p{Sc=	 VAII}', "");
Expect(1, 42540, '\p{^Sc=	 VAII}', "");
Expect(1, 42540, '\P{Sc=	 VAII}', "");
Expect(0, 42540, '\P{^Sc=	 VAII}', "");
Error('\p{Is_Script=:=		Vai}');
Error('\P{Is_Script=:=		Vai}');
Expect(1, 42539, '\p{Is_Script=vai}', "");
Expect(0, 42539, '\p{^Is_Script=vai}', "");
Expect(0, 42539, '\P{Is_Script=vai}', "");
Expect(1, 42539, '\P{^Is_Script=vai}', "");
Expect(0, 42540, '\p{Is_Script=vai}', "");
Expect(1, 42540, '\p{^Is_Script=vai}', "");
Expect(1, 42540, '\P{Is_Script=vai}', "");
Expect(0, 42540, '\P{^Is_Script=vai}', "");
Expect(1, 42539, '\p{Is_Script=	 VAI}', "");
Expect(0, 42539, '\p{^Is_Script=	 VAI}', "");
Expect(0, 42539, '\P{Is_Script=	 VAI}', "");
Expect(1, 42539, '\P{^Is_Script=	 VAI}', "");
Expect(0, 42540, '\p{Is_Script=	 VAI}', "");
Expect(1, 42540, '\p{^Is_Script=	 VAI}', "");
Expect(1, 42540, '\P{Is_Script=	 VAI}', "");
Expect(0, 42540, '\P{^Is_Script=	 VAI}', "");
Error('\p{Is_Sc=:=--VAII}');
Error('\P{Is_Sc=:=--VAII}');
Expect(1, 42539, '\p{Is_Sc=vaii}', "");
Expect(0, 42539, '\p{^Is_Sc=vaii}', "");
Expect(0, 42539, '\P{Is_Sc=vaii}', "");
Expect(1, 42539, '\P{^Is_Sc=vaii}', "");
Expect(0, 42540, '\p{Is_Sc=vaii}', "");
Expect(1, 42540, '\p{^Is_Sc=vaii}', "");
Expect(1, 42540, '\P{Is_Sc=vaii}', "");
Expect(0, 42540, '\P{^Is_Sc=vaii}', "");
Expect(1, 42539, '\p{Is_Sc=		Vaii}', "");
Expect(0, 42539, '\p{^Is_Sc=		Vaii}', "");
Expect(0, 42539, '\P{Is_Sc=		Vaii}', "");
Expect(1, 42539, '\P{^Is_Sc=		Vaii}', "");
Expect(0, 42540, '\p{Is_Sc=		Vaii}', "");
Expect(1, 42540, '\p{^Is_Sc=		Vaii}', "");
Expect(1, 42540, '\P{Is_Sc=		Vaii}', "");
Expect(0, 42540, '\P{^Is_Sc=		Vaii}', "");
Error('\p{Script=:=	_inscriptional_Parthian}');
Error('\P{Script=:=	_inscriptional_Parthian}');
Expect(1, 68447, '\p{Script=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Script=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Script=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Script=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Script=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Script=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Script=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Script=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Script=_	Inscriptional_parthian}', "");
Expect(0, 68447, '\p{^Script=_	Inscriptional_parthian}', "");
Expect(0, 68447, '\P{Script=_	Inscriptional_parthian}', "");
Expect(1, 68447, '\P{^Script=_	Inscriptional_parthian}', "");
Expect(0, 68448, '\p{Script=_	Inscriptional_parthian}', "");
Expect(1, 68448, '\p{^Script=_	Inscriptional_parthian}', "");
Expect(1, 68448, '\P{Script=_	Inscriptional_parthian}', "");
Expect(0, 68448, '\P{^Script=_	Inscriptional_parthian}', "");
Error('\p{Sc=/a/ _Prti}');
Error('\P{Sc=/a/ _Prti}');
Expect(1, 68447, '\p{Sc=prti}', "");
Expect(0, 68447, '\p{^Sc=prti}', "");
Expect(0, 68447, '\P{Sc=prti}', "");
Expect(1, 68447, '\P{^Sc=prti}', "");
Expect(0, 68448, '\p{Sc=prti}', "");
Expect(1, 68448, '\p{^Sc=prti}', "");
Expect(1, 68448, '\P{Sc=prti}', "");
Expect(0, 68448, '\P{^Sc=prti}', "");
Expect(1, 68447, '\p{Sc=	Prti}', "");
Expect(0, 68447, '\p{^Sc=	Prti}', "");
Expect(0, 68447, '\P{Sc=	Prti}', "");
Expect(1, 68447, '\P{^Sc=	Prti}', "");
Expect(0, 68448, '\p{Sc=	Prti}', "");
Expect(1, 68448, '\p{^Sc=	Prti}', "");
Expect(1, 68448, '\P{Sc=	Prti}', "");
Expect(0, 68448, '\P{^Sc=	Prti}', "");
Error('\p{Is_Script=:=-_Inscriptional_PARTHIAN}');
Error('\P{Is_Script=:=-_Inscriptional_PARTHIAN}');
Expect(1, 68447, '\p{Is_Script=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Is_Script=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Is_Script=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Is_Script=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Is_Script=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Is_Script=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Is_Script=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Is_Script=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Is_Script= inscriptional_parthian}', "");
Expect(0, 68447, '\p{^Is_Script= inscriptional_parthian}', "");
Expect(0, 68447, '\P{Is_Script= inscriptional_parthian}', "");
Expect(1, 68447, '\P{^Is_Script= inscriptional_parthian}', "");
Expect(0, 68448, '\p{Is_Script= inscriptional_parthian}', "");
Expect(1, 68448, '\p{^Is_Script= inscriptional_parthian}', "");
Expect(1, 68448, '\P{Is_Script= inscriptional_parthian}', "");
Expect(0, 68448, '\P{^Is_Script= inscriptional_parthian}', "");
Error('\p{Is_Sc=/a/_ PRTI}');
Error('\P{Is_Sc=/a/_ PRTI}');
Expect(1, 68447, '\p{Is_Sc=prti}', "");
Expect(0, 68447, '\p{^Is_Sc=prti}', "");
Expect(0, 68447, '\P{Is_Sc=prti}', "");
Expect(1, 68447, '\P{^Is_Sc=prti}', "");
Expect(0, 68448, '\p{Is_Sc=prti}', "");
Expect(1, 68448, '\p{^Is_Sc=prti}', "");
Expect(1, 68448, '\P{Is_Sc=prti}', "");
Expect(0, 68448, '\P{^Is_Sc=prti}', "");
Expect(1, 68447, '\p{Is_Sc=-	PRTI}', "");
Expect(0, 68447, '\p{^Is_Sc=-	PRTI}', "");
Expect(0, 68447, '\P{Is_Sc=-	PRTI}', "");
Expect(1, 68447, '\P{^Is_Sc=-	PRTI}', "");
Expect(0, 68448, '\p{Is_Sc=-	PRTI}', "");
Expect(1, 68448, '\p{^Is_Sc=-	PRTI}', "");
Expect(1, 68448, '\P{Is_Sc=-	PRTI}', "");
Expect(0, 68448, '\P{^Is_Sc=-	PRTI}', "");
Error('\p{Script=--runic:=}');
Error('\P{Script=--runic:=}');
Expect(1, 5872, '\p{Script:runic}', "");
Expect(0, 5872, '\p{^Script:runic}', "");
Expect(0, 5872, '\P{Script:runic}', "");
Expect(1, 5872, '\P{^Script:runic}', "");
Expect(0, 5873, '\p{Script:runic}', "");
Expect(1, 5873, '\p{^Script:runic}', "");
Expect(1, 5873, '\P{Script:runic}', "");
Expect(0, 5873, '\P{^Script:runic}', "");
Expect(1, 5872, '\p{Script=_Runic}', "");
Expect(0, 5872, '\p{^Script=_Runic}', "");
Expect(0, 5872, '\P{Script=_Runic}', "");
Expect(1, 5872, '\P{^Script=_Runic}', "");
Expect(0, 5873, '\p{Script=_Runic}', "");
Expect(1, 5873, '\p{^Script=_Runic}', "");
Expect(1, 5873, '\P{Script=_Runic}', "");
Expect(0, 5873, '\P{^Script=_Runic}', "");
Error('\p{Sc=:= 	Runr}');
Error('\P{Sc=:= 	Runr}');
Expect(1, 5872, '\p{Sc=runr}', "");
Expect(0, 5872, '\p{^Sc=runr}', "");
Expect(0, 5872, '\P{Sc=runr}', "");
Expect(1, 5872, '\P{^Sc=runr}', "");
Expect(0, 5873, '\p{Sc=runr}', "");
Expect(1, 5873, '\p{^Sc=runr}', "");
Expect(1, 5873, '\P{Sc=runr}', "");
Expect(0, 5873, '\P{^Sc=runr}', "");
Expect(1, 5872, '\p{Sc=	 runr}', "");
Expect(0, 5872, '\p{^Sc=	 runr}', "");
Expect(0, 5872, '\P{Sc=	 runr}', "");
Expect(1, 5872, '\P{^Sc=	 runr}', "");
Expect(0, 5873, '\p{Sc=	 runr}', "");
Expect(1, 5873, '\p{^Sc=	 runr}', "");
Expect(1, 5873, '\P{Sc=	 runr}', "");
Expect(0, 5873, '\P{^Sc=	 runr}', "");
Error('\p{Is_Script=	/a/RUNIC}');
Error('\P{Is_Script=	/a/RUNIC}');
Expect(1, 5872, '\p{Is_Script=runic}', "");
Expect(0, 5872, '\p{^Is_Script=runic}', "");
Expect(0, 5872, '\P{Is_Script=runic}', "");
Expect(1, 5872, '\P{^Is_Script=runic}', "");
Expect(0, 5873, '\p{Is_Script=runic}', "");
Expect(1, 5873, '\p{^Is_Script=runic}', "");
Expect(1, 5873, '\P{Is_Script=runic}', "");
Expect(0, 5873, '\P{^Is_Script=runic}', "");
Expect(1, 5872, '\p{Is_Script=__Runic}', "");
Expect(0, 5872, '\p{^Is_Script=__Runic}', "");
Expect(0, 5872, '\P{Is_Script=__Runic}', "");
Expect(1, 5872, '\P{^Is_Script=__Runic}', "");
Expect(0, 5873, '\p{Is_Script=__Runic}', "");
Expect(1, 5873, '\p{^Is_Script=__Runic}', "");
Expect(1, 5873, '\P{Is_Script=__Runic}', "");
Expect(0, 5873, '\P{^Is_Script=__Runic}', "");
Error('\p{Is_Sc: :=RUNR}');
Error('\P{Is_Sc: :=RUNR}');
Expect(1, 5872, '\p{Is_Sc=runr}', "");
Expect(0, 5872, '\p{^Is_Sc=runr}', "");
Expect(0, 5872, '\P{Is_Sc=runr}', "");
Expect(1, 5872, '\P{^Is_Sc=runr}', "");
Expect(0, 5873, '\p{Is_Sc=runr}', "");
Expect(1, 5873, '\p{^Is_Sc=runr}', "");
Expect(1, 5873, '\P{Is_Sc=runr}', "");
Expect(0, 5873, '\P{^Is_Sc=runr}', "");
Expect(1, 5872, '\p{Is_Sc=-runr}', "");
Expect(0, 5872, '\p{^Is_Sc=-runr}', "");
Expect(0, 5872, '\P{Is_Sc=-runr}', "");
Expect(1, 5872, '\P{^Is_Sc=-runr}', "");
Expect(0, 5873, '\p{Is_Sc=-runr}', "");
Expect(1, 5873, '\p{^Is_Sc=-runr}', "");
Expect(1, 5873, '\P{Is_Sc=-runr}', "");
Expect(0, 5873, '\P{^Is_Sc=-runr}', "");
Error('\p{Script=	_Cherokee:=}');
Error('\P{Script=	_Cherokee:=}');
Expect(1, 5108, '\p{Script=cherokee}', "");
Expect(0, 5108, '\p{^Script=cherokee}', "");
Expect(0, 5108, '\P{Script=cherokee}', "");
Expect(1, 5108, '\P{^Script=cherokee}', "");
Expect(0, 5109, '\p{Script=cherokee}', "");
Expect(1, 5109, '\p{^Script=cherokee}', "");
Expect(1, 5109, '\P{Script=cherokee}', "");
Expect(0, 5109, '\P{^Script=cherokee}', "");
Expect(1, 5108, '\p{Script= Cherokee}', "");
Expect(0, 5108, '\p{^Script= Cherokee}', "");
Expect(0, 5108, '\P{Script= Cherokee}', "");
Expect(1, 5108, '\P{^Script= Cherokee}', "");
Expect(0, 5109, '\p{Script= Cherokee}', "");
Expect(1, 5109, '\p{^Script= Cherokee}', "");
Expect(1, 5109, '\P{Script= Cherokee}', "");
Expect(0, 5109, '\P{^Script= Cherokee}', "");
Error('\p{Sc=:=	Cher}');
Error('\P{Sc=:=	Cher}');
Expect(1, 5108, '\p{Sc=cher}', "");
Expect(0, 5108, '\p{^Sc=cher}', "");
Expect(0, 5108, '\P{Sc=cher}', "");
Expect(1, 5108, '\P{^Sc=cher}', "");
Expect(0, 5109, '\p{Sc=cher}', "");
Expect(1, 5109, '\p{^Sc=cher}', "");
Expect(1, 5109, '\P{Sc=cher}', "");
Expect(0, 5109, '\P{^Sc=cher}', "");
Expect(1, 5108, '\p{Sc=	-CHER}', "");
Expect(0, 5108, '\p{^Sc=	-CHER}', "");
Expect(0, 5108, '\P{Sc=	-CHER}', "");
Expect(1, 5108, '\P{^Sc=	-CHER}', "");
Expect(0, 5109, '\p{Sc=	-CHER}', "");
Expect(1, 5109, '\p{^Sc=	-CHER}', "");
Expect(1, 5109, '\P{Sc=	-CHER}', "");
Expect(0, 5109, '\P{^Sc=	-CHER}', "");
Error('\p{Is_Script: -_Cherokee/a/}');
Error('\P{Is_Script: -_Cherokee/a/}');
Expect(1, 5108, '\p{Is_Script=cherokee}', "");
Expect(0, 5108, '\p{^Is_Script=cherokee}', "");
Expect(0, 5108, '\P{Is_Script=cherokee}', "");
Expect(1, 5108, '\P{^Is_Script=cherokee}', "");
Expect(0, 5109, '\p{Is_Script=cherokee}', "");
Expect(1, 5109, '\p{^Is_Script=cherokee}', "");
Expect(1, 5109, '\P{Is_Script=cherokee}', "");
Expect(0, 5109, '\P{^Is_Script=cherokee}', "");
Expect(1, 5108, '\p{Is_Script=__cherokee}', "");
Expect(0, 5108, '\p{^Is_Script=__cherokee}', "");
Expect(0, 5108, '\P{Is_Script=__cherokee}', "");
Expect(1, 5108, '\P{^Is_Script=__cherokee}', "");
Expect(0, 5109, '\p{Is_Script=__cherokee}', "");
Expect(1, 5109, '\p{^Is_Script=__cherokee}', "");
Expect(1, 5109, '\P{Is_Script=__cherokee}', "");
Expect(0, 5109, '\P{^Is_Script=__cherokee}', "");
Error('\p{Is_Sc=_ Cher/a/}');
Error('\P{Is_Sc=_ Cher/a/}');
Expect(1, 5108, '\p{Is_Sc=cher}', "");
Expect(0, 5108, '\p{^Is_Sc=cher}', "");
Expect(0, 5108, '\P{Is_Sc=cher}', "");
Expect(1, 5108, '\P{^Is_Sc=cher}', "");
Expect(0, 5109, '\p{Is_Sc=cher}', "");
Expect(1, 5109, '\p{^Is_Sc=cher}', "");
Expect(1, 5109, '\P{Is_Sc=cher}', "");
Expect(0, 5109, '\P{^Is_Sc=cher}', "");
Expect(1, 5108, '\p{Is_Sc: _CHER}', "");
Expect(0, 5108, '\p{^Is_Sc: _CHER}', "");
Expect(0, 5108, '\P{Is_Sc: _CHER}', "");
Expect(1, 5108, '\P{^Is_Sc: _CHER}', "");
Expect(0, 5109, '\p{Is_Sc: _CHER}', "");
Expect(1, 5109, '\p{^Is_Sc: _CHER}', "");
Expect(1, 5109, '\P{Is_Sc: _CHER}', "");
Expect(0, 5109, '\P{^Is_Sc: _CHER}', "");
Error('\p{Script=	 Meroitic_hieroglyphs:=}');
Error('\P{Script=	 Meroitic_hieroglyphs:=}');
Expect(1, 67999, '\p{Script=meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Script=meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Script=meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Script=meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Script=meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Script=meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Script=meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Script=meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Script=_ Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\p{^Script=_ Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\P{Script=_ Meroitic_Hieroglyphs}', "");
Expect(1, 67999, '\P{^Script=_ Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\p{Script=_ Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\p{^Script=_ Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\P{Script=_ Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\P{^Script=_ Meroitic_Hieroglyphs}', "");
Error('\p{Sc= :=Mero}');
Error('\P{Sc= :=Mero}');
Expect(1, 67999, '\p{Sc=mero}', "");
Expect(0, 67999, '\p{^Sc=mero}', "");
Expect(0, 67999, '\P{Sc=mero}', "");
Expect(1, 67999, '\P{^Sc=mero}', "");
Expect(0, 68000, '\p{Sc=mero}', "");
Expect(1, 68000, '\p{^Sc=mero}', "");
Expect(1, 68000, '\P{Sc=mero}', "");
Expect(0, 68000, '\P{^Sc=mero}', "");
Expect(1, 67999, '\p{Sc=_MERO}', "");
Expect(0, 67999, '\p{^Sc=_MERO}', "");
Expect(0, 67999, '\P{Sc=_MERO}', "");
Expect(1, 67999, '\P{^Sc=_MERO}', "");
Expect(0, 68000, '\p{Sc=_MERO}', "");
Expect(1, 68000, '\p{^Sc=_MERO}', "");
Expect(1, 68000, '\P{Sc=_MERO}', "");
Expect(0, 68000, '\P{^Sc=_MERO}', "");
Error('\p{Is_Script=/a/MEROITIC_Hieroglyphs}');
Error('\P{Is_Script=/a/MEROITIC_Hieroglyphs}');
Expect(1, 67999, '\p{Is_Script=meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Script=meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Is_Script=meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Script=meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Is_Script=meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Script=meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Is_Script=meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Script=meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Is_Script: 		Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Script: 		Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\P{Is_Script: 		Meroitic_Hieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Script: 		Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\p{Is_Script: 		Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Script: 		Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\P{Is_Script: 		Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Script: 		Meroitic_Hieroglyphs}', "");
Error('\p{Is_Sc=	:=mero}');
Error('\P{Is_Sc=	:=mero}');
Expect(1, 67999, '\p{Is_Sc=mero}', "");
Expect(0, 67999, '\p{^Is_Sc=mero}', "");
Expect(0, 67999, '\P{Is_Sc=mero}', "");
Expect(1, 67999, '\P{^Is_Sc=mero}', "");
Expect(0, 68000, '\p{Is_Sc=mero}', "");
Expect(1, 68000, '\p{^Is_Sc=mero}', "");
Expect(1, 68000, '\P{Is_Sc=mero}', "");
Expect(0, 68000, '\P{^Is_Sc=mero}', "");
Expect(1, 67999, '\p{Is_Sc=_	mero}', "");
Expect(0, 67999, '\p{^Is_Sc=_	mero}', "");
Expect(0, 67999, '\P{Is_Sc=_	mero}', "");
Expect(1, 67999, '\P{^Is_Sc=_	mero}', "");
Expect(0, 68000, '\p{Is_Sc=_	mero}', "");
Expect(1, 68000, '\p{^Is_Sc=_	mero}', "");
Expect(1, 68000, '\P{Is_Sc=_	mero}', "");
Expect(0, 68000, '\P{^Is_Sc=_	mero}', "");
Error('\p{Script=/a/__Ol_Chiki}');
Error('\P{Script=/a/__Ol_Chiki}');
Expect(1, 7295, '\p{Script=olchiki}', "");
Expect(0, 7295, '\p{^Script=olchiki}', "");
Expect(0, 7295, '\P{Script=olchiki}', "");
Expect(1, 7295, '\P{^Script=olchiki}', "");
Expect(0, 7296, '\p{Script=olchiki}', "");
Expect(1, 7296, '\p{^Script=olchiki}', "");
Expect(1, 7296, '\P{Script=olchiki}', "");
Expect(0, 7296, '\P{^Script=olchiki}', "");
Expect(1, 7295, '\p{Script=_ ol_chiki}', "");
Expect(0, 7295, '\p{^Script=_ ol_chiki}', "");
Expect(0, 7295, '\P{Script=_ ol_chiki}', "");
Expect(1, 7295, '\P{^Script=_ ol_chiki}', "");
Expect(0, 7296, '\p{Script=_ ol_chiki}', "");
Expect(1, 7296, '\p{^Script=_ ol_chiki}', "");
Expect(1, 7296, '\P{Script=_ ol_chiki}', "");
Expect(0, 7296, '\P{^Script=_ ol_chiki}', "");
Error('\p{Sc=	/a/olck}');
Error('\P{Sc=	/a/olck}');
Expect(1, 7295, '\p{Sc=olck}', "");
Expect(0, 7295, '\p{^Sc=olck}', "");
Expect(0, 7295, '\P{Sc=olck}', "");
Expect(1, 7295, '\P{^Sc=olck}', "");
Expect(0, 7296, '\p{Sc=olck}', "");
Expect(1, 7296, '\p{^Sc=olck}', "");
Expect(1, 7296, '\P{Sc=olck}', "");
Expect(0, 7296, '\P{^Sc=olck}', "");
Expect(1, 7295, '\p{Sc=__OLCK}', "");
Expect(0, 7295, '\p{^Sc=__OLCK}', "");
Expect(0, 7295, '\P{Sc=__OLCK}', "");
Expect(1, 7295, '\P{^Sc=__OLCK}', "");
Expect(0, 7296, '\p{Sc=__OLCK}', "");
Expect(1, 7296, '\p{^Sc=__OLCK}', "");
Expect(1, 7296, '\P{Sc=__OLCK}', "");
Expect(0, 7296, '\P{^Sc=__OLCK}', "");
Error('\p{Is_Script=:=	OL_chiki}');
Error('\P{Is_Script=:=	OL_chiki}');
Expect(1, 7295, '\p{Is_Script=olchiki}', "");
Expect(0, 7295, '\p{^Is_Script=olchiki}', "");
Expect(0, 7295, '\P{Is_Script=olchiki}', "");
Expect(1, 7295, '\P{^Is_Script=olchiki}', "");
Expect(0, 7296, '\p{Is_Script=olchiki}', "");
Expect(1, 7296, '\p{^Is_Script=olchiki}', "");
Expect(1, 7296, '\P{Is_Script=olchiki}', "");
Expect(0, 7296, '\P{^Is_Script=olchiki}', "");
Expect(1, 7295, '\p{Is_Script=-ol_Chiki}', "");
Expect(0, 7295, '\p{^Is_Script=-ol_Chiki}', "");
Expect(0, 7295, '\P{Is_Script=-ol_Chiki}', "");
Expect(1, 7295, '\P{^Is_Script=-ol_Chiki}', "");
Expect(0, 7296, '\p{Is_Script=-ol_Chiki}', "");
Expect(1, 7296, '\p{^Is_Script=-ol_Chiki}', "");
Expect(1, 7296, '\P{Is_Script=-ol_Chiki}', "");
Expect(0, 7296, '\P{^Is_Script=-ol_Chiki}', "");
Error('\p{Is_Sc=/a/--Olck}');
Error('\P{Is_Sc=/a/--Olck}');
Expect(1, 7295, '\p{Is_Sc=olck}', "");
Expect(0, 7295, '\p{^Is_Sc=olck}', "");
Expect(0, 7295, '\P{Is_Sc=olck}', "");
Expect(1, 7295, '\P{^Is_Sc=olck}', "");
Expect(0, 7296, '\p{Is_Sc=olck}', "");
Expect(1, 7296, '\p{^Is_Sc=olck}', "");
Expect(1, 7296, '\P{Is_Sc=olck}', "");
Expect(0, 7296, '\P{^Is_Sc=olck}', "");
Expect(1, 7295, '\p{Is_Sc=	 olck}', "");
Expect(0, 7295, '\p{^Is_Sc=	 olck}', "");
Expect(0, 7295, '\P{Is_Sc=	 olck}', "");
Expect(1, 7295, '\P{^Is_Sc=	 olck}', "");
Expect(0, 7296, '\p{Is_Sc=	 olck}', "");
Expect(1, 7296, '\p{^Is_Sc=	 olck}', "");
Expect(1, 7296, '\P{Is_Sc=	 olck}', "");
Expect(0, 7296, '\P{^Is_Sc=	 olck}', "");
Error('\p{Script=_Ugaritic:=}');
Error('\P{Script=_Ugaritic:=}');
Expect(1, 66463, '\p{Script=ugaritic}', "");
Expect(0, 66463, '\p{^Script=ugaritic}', "");
Expect(0, 66463, '\P{Script=ugaritic}', "");
Expect(1, 66463, '\P{^Script=ugaritic}', "");
Expect(0, 66464, '\p{Script=ugaritic}', "");
Expect(1, 66464, '\p{^Script=ugaritic}', "");
Expect(1, 66464, '\P{Script=ugaritic}', "");
Expect(0, 66464, '\P{^Script=ugaritic}', "");
Expect(1, 66463, '\p{Script=	_UGARITIC}', "");
Expect(0, 66463, '\p{^Script=	_UGARITIC}', "");
Expect(0, 66463, '\P{Script=	_UGARITIC}', "");
Expect(1, 66463, '\P{^Script=	_UGARITIC}', "");
Expect(0, 66464, '\p{Script=	_UGARITIC}', "");
Expect(1, 66464, '\p{^Script=	_UGARITIC}', "");
Expect(1, 66464, '\P{Script=	_UGARITIC}', "");
Expect(0, 66464, '\P{^Script=	_UGARITIC}', "");
Error('\p{Sc=UGAR/a/}');
Error('\P{Sc=UGAR/a/}');
Expect(1, 66463, '\p{Sc=ugar}', "");
Expect(0, 66463, '\p{^Sc=ugar}', "");
Expect(0, 66463, '\P{Sc=ugar}', "");
Expect(1, 66463, '\P{^Sc=ugar}', "");
Expect(0, 66464, '\p{Sc=ugar}', "");
Expect(1, 66464, '\p{^Sc=ugar}', "");
Expect(1, 66464, '\P{Sc=ugar}', "");
Expect(0, 66464, '\P{^Sc=ugar}', "");
Expect(1, 66463, '\p{Sc= -Ugar}', "");
Expect(0, 66463, '\p{^Sc= -Ugar}', "");
Expect(0, 66463, '\P{Sc= -Ugar}', "");
Expect(1, 66463, '\P{^Sc= -Ugar}', "");
Expect(0, 66464, '\p{Sc= -Ugar}', "");
Expect(1, 66464, '\p{^Sc= -Ugar}', "");
Expect(1, 66464, '\P{Sc= -Ugar}', "");
Expect(0, 66464, '\P{^Sc= -Ugar}', "");
Error('\p{Is_Script=/a/- UGARITIC}');
Error('\P{Is_Script=/a/- UGARITIC}');
Expect(1, 66463, '\p{Is_Script=ugaritic}', "");
Expect(0, 66463, '\p{^Is_Script=ugaritic}', "");
Expect(0, 66463, '\P{Is_Script=ugaritic}', "");
Expect(1, 66463, '\P{^Is_Script=ugaritic}', "");
Expect(0, 66464, '\p{Is_Script=ugaritic}', "");
Expect(1, 66464, '\p{^Is_Script=ugaritic}', "");
Expect(1, 66464, '\P{Is_Script=ugaritic}', "");
Expect(0, 66464, '\P{^Is_Script=ugaritic}', "");
Expect(1, 66463, '\p{Is_Script=  ugaritic}', "");
Expect(0, 66463, '\p{^Is_Script=  ugaritic}', "");
Expect(0, 66463, '\P{Is_Script=  ugaritic}', "");
Expect(1, 66463, '\P{^Is_Script=  ugaritic}', "");
Expect(0, 66464, '\p{Is_Script=  ugaritic}', "");
Expect(1, 66464, '\p{^Is_Script=  ugaritic}', "");
Expect(1, 66464, '\P{Is_Script=  ugaritic}', "");
Expect(0, 66464, '\P{^Is_Script=  ugaritic}', "");
Error('\p{Is_Sc=:=	 Ugar}');
Error('\P{Is_Sc=:=	 Ugar}');
Expect(1, 66463, '\p{Is_Sc=ugar}', "");
Expect(0, 66463, '\p{^Is_Sc=ugar}', "");
Expect(0, 66463, '\P{Is_Sc=ugar}', "");
Expect(1, 66463, '\P{^Is_Sc=ugar}', "");
Expect(0, 66464, '\p{Is_Sc=ugar}', "");
Expect(1, 66464, '\p{^Is_Sc=ugar}', "");
Expect(1, 66464, '\P{Is_Sc=ugar}', "");
Expect(0, 66464, '\P{^Is_Sc=ugar}', "");
Expect(1, 66463, '\p{Is_Sc:Ugar}', "");
Expect(0, 66463, '\p{^Is_Sc:Ugar}', "");
Expect(0, 66463, '\P{Is_Sc:Ugar}', "");
Expect(1, 66463, '\P{^Is_Sc:Ugar}', "");
Expect(0, 66464, '\p{Is_Sc:Ugar}', "");
Expect(1, 66464, '\p{^Is_Sc:Ugar}', "");
Expect(1, 66464, '\P{Is_Sc:Ugar}', "");
Expect(0, 66464, '\P{^Is_Sc:Ugar}', "");
Error('\p{Script=_:=Phags_pa}');
Error('\P{Script=_:=Phags_pa}');
Expect(1, 43127, '\p{Script=phagspa}', "");
Expect(0, 43127, '\p{^Script=phagspa}', "");
Expect(0, 43127, '\P{Script=phagspa}', "");
Expect(1, 43127, '\P{^Script=phagspa}', "");
Expect(0, 43128, '\p{Script=phagspa}', "");
Expect(1, 43128, '\p{^Script=phagspa}', "");
Expect(1, 43128, '\P{Script=phagspa}', "");
Expect(0, 43128, '\P{^Script=phagspa}', "");
Expect(1, 43127, '\p{Script=-phags_Pa}', "");
Expect(0, 43127, '\p{^Script=-phags_Pa}', "");
Expect(0, 43127, '\P{Script=-phags_Pa}', "");
Expect(1, 43127, '\P{^Script=-phags_Pa}', "");
Expect(0, 43128, '\p{Script=-phags_Pa}', "");
Expect(1, 43128, '\p{^Script=-phags_Pa}', "");
Expect(1, 43128, '\P{Script=-phags_Pa}', "");
Expect(0, 43128, '\P{^Script=-phags_Pa}', "");
Error('\p{Sc=-phag/a/}');
Error('\P{Sc=-phag/a/}');
Expect(1, 43127, '\p{Sc=phag}', "");
Expect(0, 43127, '\p{^Sc=phag}', "");
Expect(0, 43127, '\P{Sc=phag}', "");
Expect(1, 43127, '\P{^Sc=phag}', "");
Expect(0, 43128, '\p{Sc=phag}', "");
Expect(1, 43128, '\p{^Sc=phag}', "");
Expect(1, 43128, '\P{Sc=phag}', "");
Expect(0, 43128, '\P{^Sc=phag}', "");
Expect(1, 43127, '\p{Sc=__PHAG}', "");
Expect(0, 43127, '\p{^Sc=__PHAG}', "");
Expect(0, 43127, '\P{Sc=__PHAG}', "");
Expect(1, 43127, '\P{^Sc=__PHAG}', "");
Expect(0, 43128, '\p{Sc=__PHAG}', "");
Expect(1, 43128, '\p{^Sc=__PHAG}', "");
Expect(1, 43128, '\P{Sc=__PHAG}', "");
Expect(0, 43128, '\P{^Sc=__PHAG}', "");
Error('\p{Is_Script=/a/phags_Pa}');
Error('\P{Is_Script=/a/phags_Pa}');
Expect(1, 43127, '\p{Is_Script=phagspa}', "");
Expect(0, 43127, '\p{^Is_Script=phagspa}', "");
Expect(0, 43127, '\P{Is_Script=phagspa}', "");
Expect(1, 43127, '\P{^Is_Script=phagspa}', "");
Expect(0, 43128, '\p{Is_Script=phagspa}', "");
Expect(1, 43128, '\p{^Is_Script=phagspa}', "");
Expect(1, 43128, '\P{Is_Script=phagspa}', "");
Expect(0, 43128, '\P{^Is_Script=phagspa}', "");
Expect(1, 43127, '\p{Is_Script: 	-PHAGS_Pa}', "");
Expect(0, 43127, '\p{^Is_Script: 	-PHAGS_Pa}', "");
Expect(0, 43127, '\P{Is_Script: 	-PHAGS_Pa}', "");
Expect(1, 43127, '\P{^Is_Script: 	-PHAGS_Pa}', "");
Expect(0, 43128, '\p{Is_Script: 	-PHAGS_Pa}', "");
Expect(1, 43128, '\p{^Is_Script: 	-PHAGS_Pa}', "");
Expect(1, 43128, '\P{Is_Script: 	-PHAGS_Pa}', "");
Expect(0, 43128, '\P{^Is_Script: 	-PHAGS_Pa}', "");
Error('\p{Is_Sc=-:=PHAG}');
Error('\P{Is_Sc=-:=PHAG}');
Expect(1, 43127, '\p{Is_Sc=phag}', "");
Expect(0, 43127, '\p{^Is_Sc=phag}', "");
Expect(0, 43127, '\P{Is_Sc=phag}', "");
Expect(1, 43127, '\P{^Is_Sc=phag}', "");
Expect(0, 43128, '\p{Is_Sc=phag}', "");
Expect(1, 43128, '\p{^Is_Sc=phag}', "");
Expect(1, 43128, '\P{Is_Sc=phag}', "");
Expect(0, 43128, '\P{^Is_Sc=phag}', "");
Expect(1, 43127, '\p{Is_Sc= 	phag}', "");
Expect(0, 43127, '\p{^Is_Sc= 	phag}', "");
Expect(0, 43127, '\P{Is_Sc= 	phag}', "");
Expect(1, 43127, '\P{^Is_Sc= 	phag}', "");
Expect(0, 43128, '\p{Is_Sc= 	phag}', "");
Expect(1, 43128, '\p{^Is_Sc= 	phag}', "");
Expect(1, 43128, '\P{Is_Sc= 	phag}', "");
Expect(0, 43128, '\P{^Is_Sc= 	phag}', "");
Error('\p{Script=_ARMENIAN/a/}');
Error('\P{Script=_ARMENIAN/a/}');
Expect(1, 64279, '\p{Script=armenian}', "");
Expect(0, 64279, '\p{^Script=armenian}', "");
Expect(0, 64279, '\P{Script=armenian}', "");
Expect(1, 64279, '\P{^Script=armenian}', "");
Expect(0, 64280, '\p{Script=armenian}', "");
Expect(1, 64280, '\p{^Script=armenian}', "");
Expect(1, 64280, '\P{Script=armenian}', "");
Expect(0, 64280, '\P{^Script=armenian}', "");
Expect(1, 64279, '\p{Script=	 Armenian}', "");
Expect(0, 64279, '\p{^Script=	 Armenian}', "");
Expect(0, 64279, '\P{Script=	 Armenian}', "");
Expect(1, 64279, '\P{^Script=	 Armenian}', "");
Expect(0, 64280, '\p{Script=	 Armenian}', "");
Expect(1, 64280, '\p{^Script=	 Armenian}', "");
Expect(1, 64280, '\P{Script=	 Armenian}', "");
Expect(0, 64280, '\P{^Script=	 Armenian}', "");
Error('\p{Sc=/a/_armn}');
Error('\P{Sc=/a/_armn}');
Expect(1, 64279, '\p{Sc=armn}', "");
Expect(0, 64279, '\p{^Sc=armn}', "");
Expect(0, 64279, '\P{Sc=armn}', "");
Expect(1, 64279, '\P{^Sc=armn}', "");
Expect(0, 64280, '\p{Sc=armn}', "");
Expect(1, 64280, '\p{^Sc=armn}', "");
Expect(1, 64280, '\P{Sc=armn}', "");
Expect(0, 64280, '\P{^Sc=armn}', "");
Expect(1, 64279, '\p{Sc=	armn}', "");
Expect(0, 64279, '\p{^Sc=	armn}', "");
Expect(0, 64279, '\P{Sc=	armn}', "");
Expect(1, 64279, '\P{^Sc=	armn}', "");
Expect(0, 64280, '\p{Sc=	armn}', "");
Expect(1, 64280, '\p{^Sc=	armn}', "");
Expect(1, 64280, '\P{Sc=	armn}', "");
Expect(0, 64280, '\P{^Sc=	armn}', "");
Error('\p{Is_Script:   /a/--ARMENIAN}');
Error('\P{Is_Script:   /a/--ARMENIAN}');
Expect(1, 64279, '\p{Is_Script=armenian}', "");
Expect(0, 64279, '\p{^Is_Script=armenian}', "");
Expect(0, 64279, '\P{Is_Script=armenian}', "");
Expect(1, 64279, '\P{^Is_Script=armenian}', "");
Expect(0, 64280, '\p{Is_Script=armenian}', "");
Expect(1, 64280, '\p{^Is_Script=armenian}', "");
Expect(1, 64280, '\P{Is_Script=armenian}', "");
Expect(0, 64280, '\P{^Is_Script=armenian}', "");
Expect(1, 64279, '\p{Is_Script=	_armenian}', "");
Expect(0, 64279, '\p{^Is_Script=	_armenian}', "");
Expect(0, 64279, '\P{Is_Script=	_armenian}', "");
Expect(1, 64279, '\P{^Is_Script=	_armenian}', "");
Expect(0, 64280, '\p{Is_Script=	_armenian}', "");
Expect(1, 64280, '\p{^Is_Script=	_armenian}', "");
Expect(1, 64280, '\P{Is_Script=	_armenian}', "");
Expect(0, 64280, '\P{^Is_Script=	_armenian}', "");
Error('\p{Is_Sc=:=  Armn}');
Error('\P{Is_Sc=:=  Armn}');
Expect(1, 64279, '\p{Is_Sc=armn}', "");
Expect(0, 64279, '\p{^Is_Sc=armn}', "");
Expect(0, 64279, '\P{Is_Sc=armn}', "");
Expect(1, 64279, '\P{^Is_Sc=armn}', "");
Expect(0, 64280, '\p{Is_Sc=armn}', "");
Expect(1, 64280, '\p{^Is_Sc=armn}', "");
Expect(1, 64280, '\P{Is_Sc=armn}', "");
Expect(0, 64280, '\P{^Is_Sc=armn}', "");
Expect(1, 64279, '\p{Is_Sc=		Armn}', "");
Expect(0, 64279, '\p{^Is_Sc=		Armn}', "");
Expect(0, 64279, '\P{Is_Sc=		Armn}', "");
Expect(1, 64279, '\P{^Is_Sc=		Armn}', "");
Expect(0, 64280, '\p{Is_Sc=		Armn}', "");
Expect(1, 64280, '\p{^Is_Sc=		Armn}', "");
Expect(1, 64280, '\P{Is_Sc=		Armn}', "");
Expect(0, 64280, '\P{^Is_Sc=		Armn}', "");
Error('\p{Script= 	syriac:=}');
Error('\P{Script= 	syriac:=}');
Expect(1, 1871, '\p{Script=syriac}', "");
Expect(0, 1871, '\p{^Script=syriac}', "");
Expect(0, 1871, '\P{Script=syriac}', "");
Expect(1, 1871, '\P{^Script=syriac}', "");
Expect(0, 1872, '\p{Script=syriac}', "");
Expect(1, 1872, '\p{^Script=syriac}', "");
Expect(1, 1872, '\P{Script=syriac}', "");
Expect(0, 1872, '\P{^Script=syriac}', "");
Expect(1, 1871, '\p{Script=_	syriac}', "");
Expect(0, 1871, '\p{^Script=_	syriac}', "");
Expect(0, 1871, '\P{Script=_	syriac}', "");
Expect(1, 1871, '\P{^Script=_	syriac}', "");
Expect(0, 1872, '\p{Script=_	syriac}', "");
Expect(1, 1872, '\p{^Script=_	syriac}', "");
Expect(1, 1872, '\P{Script=_	syriac}', "");
Expect(0, 1872, '\P{^Script=_	syriac}', "");
Error('\p{Sc=-_SYRC:=}');
Error('\P{Sc=-_SYRC:=}');
Expect(1, 1871, '\p{Sc=syrc}', "");
Expect(0, 1871, '\p{^Sc=syrc}', "");
Expect(0, 1871, '\P{Sc=syrc}', "");
Expect(1, 1871, '\P{^Sc=syrc}', "");
Expect(0, 1872, '\p{Sc=syrc}', "");
Expect(1, 1872, '\p{^Sc=syrc}', "");
Expect(1, 1872, '\P{Sc=syrc}', "");
Expect(0, 1872, '\P{^Sc=syrc}', "");
Expect(1, 1871, '\p{Sc=_-Syrc}', "");
Expect(0, 1871, '\p{^Sc=_-Syrc}', "");
Expect(0, 1871, '\P{Sc=_-Syrc}', "");
Expect(1, 1871, '\P{^Sc=_-Syrc}', "");
Expect(0, 1872, '\p{Sc=_-Syrc}', "");
Expect(1, 1872, '\p{^Sc=_-Syrc}', "");
Expect(1, 1872, '\P{Sc=_-Syrc}', "");
Expect(0, 1872, '\P{^Sc=_-Syrc}', "");
Error('\p{Is_Script=-:=Syriac}');
Error('\P{Is_Script=-:=Syriac}');
Expect(1, 1871, '\p{Is_Script=syriac}', "");
Expect(0, 1871, '\p{^Is_Script=syriac}', "");
Expect(0, 1871, '\P{Is_Script=syriac}', "");
Expect(1, 1871, '\P{^Is_Script=syriac}', "");
Expect(0, 1872, '\p{Is_Script=syriac}', "");
Expect(1, 1872, '\p{^Is_Script=syriac}', "");
Expect(1, 1872, '\P{Is_Script=syriac}', "");
Expect(0, 1872, '\P{^Is_Script=syriac}', "");
Expect(1, 1871, '\p{Is_Script=	 Syriac}', "");
Expect(0, 1871, '\p{^Is_Script=	 Syriac}', "");
Expect(0, 1871, '\P{Is_Script=	 Syriac}', "");
Expect(1, 1871, '\P{^Is_Script=	 Syriac}', "");
Expect(0, 1872, '\p{Is_Script=	 Syriac}', "");
Expect(1, 1872, '\p{^Is_Script=	 Syriac}', "");
Expect(1, 1872, '\P{Is_Script=	 Syriac}', "");
Expect(0, 1872, '\P{^Is_Script=	 Syriac}', "");
Error('\p{Is_Sc=-/a/Syrc}');
Error('\P{Is_Sc=-/a/Syrc}');
Expect(1, 1871, '\p{Is_Sc=syrc}', "");
Expect(0, 1871, '\p{^Is_Sc=syrc}', "");
Expect(0, 1871, '\P{Is_Sc=syrc}', "");
Expect(1, 1871, '\P{^Is_Sc=syrc}', "");
Expect(0, 1872, '\p{Is_Sc=syrc}', "");
Expect(1, 1872, '\p{^Is_Sc=syrc}', "");
Expect(1, 1872, '\P{Is_Sc=syrc}', "");
Expect(0, 1872, '\P{^Is_Sc=syrc}', "");
Expect(1, 1871, '\p{Is_Sc=_ Syrc}', "");
Expect(0, 1871, '\p{^Is_Sc=_ Syrc}', "");
Expect(0, 1871, '\P{Is_Sc=_ Syrc}', "");
Expect(1, 1871, '\P{^Is_Sc=_ Syrc}', "");
Expect(0, 1872, '\p{Is_Sc=_ Syrc}', "");
Expect(1, 1872, '\p{^Is_Sc=_ Syrc}', "");
Expect(1, 1872, '\P{Is_Sc=_ Syrc}', "");
Expect(0, 1872, '\P{^Is_Sc=_ Syrc}', "");
Error('\p{Script=	:=Yi}');
Error('\P{Script=	:=Yi}');
Expect(1, 42182, '\p{Script=yi}', "");
Expect(0, 42182, '\p{^Script=yi}', "");
Expect(0, 42182, '\P{Script=yi}', "");
Expect(1, 42182, '\P{^Script=yi}', "");
Expect(0, 42183, '\p{Script=yi}', "");
Expect(1, 42183, '\p{^Script=yi}', "");
Expect(1, 42183, '\P{Script=yi}', "");
Expect(0, 42183, '\P{^Script=yi}', "");
Expect(1, 42182, '\p{Script= _Yi}', "");
Expect(0, 42182, '\p{^Script= _Yi}', "");
Expect(0, 42182, '\P{Script= _Yi}', "");
Expect(1, 42182, '\P{^Script= _Yi}', "");
Expect(0, 42183, '\p{Script= _Yi}', "");
Expect(1, 42183, '\p{^Script= _Yi}', "");
Expect(1, 42183, '\P{Script= _Yi}', "");
Expect(0, 42183, '\P{^Script= _Yi}', "");
Error('\p{Sc:	:=__Yiii}');
Error('\P{Sc:	:=__Yiii}');
Expect(1, 42182, '\p{Sc=yiii}', "");
Expect(0, 42182, '\p{^Sc=yiii}', "");
Expect(0, 42182, '\P{Sc=yiii}', "");
Expect(1, 42182, '\P{^Sc=yiii}', "");
Expect(0, 42183, '\p{Sc=yiii}', "");
Expect(1, 42183, '\p{^Sc=yiii}', "");
Expect(1, 42183, '\P{Sc=yiii}', "");
Expect(0, 42183, '\P{^Sc=yiii}', "");
Expect(1, 42182, '\p{Sc=  Yiii}', "");
Expect(0, 42182, '\p{^Sc=  Yiii}', "");
Expect(0, 42182, '\P{Sc=  Yiii}', "");
Expect(1, 42182, '\P{^Sc=  Yiii}', "");
Expect(0, 42183, '\p{Sc=  Yiii}', "");
Expect(1, 42183, '\p{^Sc=  Yiii}', "");
Expect(1, 42183, '\P{Sc=  Yiii}', "");
Expect(0, 42183, '\P{^Sc=  Yiii}', "");
Error('\p{Is_Script= _yi:=}');
Error('\P{Is_Script= _yi:=}');
Expect(1, 42182, '\p{Is_Script=yi}', "");
Expect(0, 42182, '\p{^Is_Script=yi}', "");
Expect(0, 42182, '\P{Is_Script=yi}', "");
Expect(1, 42182, '\P{^Is_Script=yi}', "");
Expect(0, 42183, '\p{Is_Script=yi}', "");
Expect(1, 42183, '\p{^Is_Script=yi}', "");
Expect(1, 42183, '\P{Is_Script=yi}', "");
Expect(0, 42183, '\P{^Is_Script=yi}', "");
Expect(1, 42182, '\p{Is_Script=- yi}', "");
Expect(0, 42182, '\p{^Is_Script=- yi}', "");
Expect(0, 42182, '\P{Is_Script=- yi}', "");
Expect(1, 42182, '\P{^Is_Script=- yi}', "");
Expect(0, 42183, '\p{Is_Script=- yi}', "");
Expect(1, 42183, '\p{^Is_Script=- yi}', "");
Expect(1, 42183, '\P{Is_Script=- yi}', "");
Expect(0, 42183, '\P{^Is_Script=- yi}', "");
Error('\p{Is_Sc=_/a/Yiii}');
Error('\P{Is_Sc=_/a/Yiii}');
Expect(1, 42182, '\p{Is_Sc=yiii}', "");
Expect(0, 42182, '\p{^Is_Sc=yiii}', "");
Expect(0, 42182, '\P{Is_Sc=yiii}', "");
Expect(1, 42182, '\P{^Is_Sc=yiii}', "");
Expect(0, 42183, '\p{Is_Sc=yiii}', "");
Expect(1, 42183, '\p{^Is_Sc=yiii}', "");
Expect(1, 42183, '\P{Is_Sc=yiii}', "");
Expect(0, 42183, '\P{^Is_Sc=yiii}', "");
Expect(1, 42182, '\p{Is_Sc:	-Yiii}', "");
Expect(0, 42182, '\p{^Is_Sc:	-Yiii}', "");
Expect(0, 42182, '\P{Is_Sc:	-Yiii}', "");
Expect(1, 42182, '\P{^Is_Sc:	-Yiii}', "");
Expect(0, 42183, '\p{Is_Sc:	-Yiii}', "");
Expect(1, 42183, '\p{^Is_Sc:	-Yiii}', "");
Expect(1, 42183, '\P{Is_Sc:	-Yiii}', "");
Expect(0, 42183, '\P{^Is_Sc:	-Yiii}', "");
Error('\p{Script=	:=Katakana}');
Error('\P{Script=	:=Katakana}');
Expect(1, 110592, '\p{Script:	katakana}', "");
Expect(0, 110592, '\p{^Script:	katakana}', "");
Expect(0, 110592, '\P{Script:	katakana}', "");
Expect(1, 110592, '\P{^Script:	katakana}', "");
Expect(0, 110593, '\p{Script:	katakana}', "");
Expect(1, 110593, '\p{^Script:	katakana}', "");
Expect(1, 110593, '\P{Script:	katakana}', "");
Expect(0, 110593, '\P{^Script:	katakana}', "");
Expect(1, 110592, '\p{Script=_katakana}', "");
Expect(0, 110592, '\p{^Script=_katakana}', "");
Expect(0, 110592, '\P{Script=_katakana}', "");
Expect(1, 110592, '\P{^Script=_katakana}', "");
Expect(0, 110593, '\p{Script=_katakana}', "");
Expect(1, 110593, '\p{^Script=_katakana}', "");
Expect(1, 110593, '\P{Script=_katakana}', "");
Expect(0, 110593, '\P{^Script=_katakana}', "");
Error('\p{Sc=:=_	Kana}');
Error('\P{Sc=:=_	Kana}');
Expect(1, 110592, '\p{Sc=kana}', "");
Expect(0, 110592, '\p{^Sc=kana}', "");
Expect(0, 110592, '\P{Sc=kana}', "");
Expect(1, 110592, '\P{^Sc=kana}', "");
Expect(0, 110593, '\p{Sc=kana}', "");
Expect(1, 110593, '\p{^Sc=kana}', "");
Expect(1, 110593, '\P{Sc=kana}', "");
Expect(0, 110593, '\P{^Sc=kana}', "");
Expect(1, 110592, '\p{Sc=		KANA}', "");
Expect(0, 110592, '\p{^Sc=		KANA}', "");
Expect(0, 110592, '\P{Sc=		KANA}', "");
Expect(1, 110592, '\P{^Sc=		KANA}', "");
Expect(0, 110593, '\p{Sc=		KANA}', "");
Expect(1, 110593, '\p{^Sc=		KANA}', "");
Expect(1, 110593, '\P{Sc=		KANA}', "");
Expect(0, 110593, '\P{^Sc=		KANA}', "");
Error('\p{Is_Script= :=katakana}');
Error('\P{Is_Script= :=katakana}');
Expect(1, 110592, '\p{Is_Script=katakana}', "");
Expect(0, 110592, '\p{^Is_Script=katakana}', "");
Expect(0, 110592, '\P{Is_Script=katakana}', "");
Expect(1, 110592, '\P{^Is_Script=katakana}', "");
Expect(0, 110593, '\p{Is_Script=katakana}', "");
Expect(1, 110593, '\p{^Is_Script=katakana}', "");
Expect(1, 110593, '\P{Is_Script=katakana}', "");
Expect(0, 110593, '\P{^Is_Script=katakana}', "");
Expect(1, 110592, '\p{Is_Script=	katakana}', "");
Expect(0, 110592, '\p{^Is_Script=	katakana}', "");
Expect(0, 110592, '\P{Is_Script=	katakana}', "");
Expect(1, 110592, '\P{^Is_Script=	katakana}', "");
Expect(0, 110593, '\p{Is_Script=	katakana}', "");
Expect(1, 110593, '\p{^Is_Script=	katakana}', "");
Expect(1, 110593, '\P{Is_Script=	katakana}', "");
Expect(0, 110593, '\P{^Is_Script=	katakana}', "");
Error('\p{Is_Sc=--Kana:=}');
Error('\P{Is_Sc=--Kana:=}');
Expect(1, 110592, '\p{Is_Sc=kana}', "");
Expect(0, 110592, '\p{^Is_Sc=kana}', "");
Expect(0, 110592, '\P{Is_Sc=kana}', "");
Expect(1, 110592, '\P{^Is_Sc=kana}', "");
Expect(0, 110593, '\p{Is_Sc=kana}', "");
Expect(1, 110593, '\p{^Is_Sc=kana}', "");
Expect(1, 110593, '\P{Is_Sc=kana}', "");
Expect(0, 110593, '\P{^Is_Sc=kana}', "");
Expect(1, 110592, '\p{Is_Sc=_-Kana}', "");
Expect(0, 110592, '\p{^Is_Sc=_-Kana}', "");
Expect(0, 110592, '\P{Is_Sc=_-Kana}', "");
Expect(1, 110592, '\P{^Is_Sc=_-Kana}', "");
Expect(0, 110593, '\p{Is_Sc=_-Kana}', "");
Expect(1, 110593, '\p{^Is_Sc=_-Kana}', "");
Expect(1, 110593, '\P{Is_Sc=_-Kana}', "");
Expect(0, 110593, '\P{^Is_Sc=_-Kana}', "");
Error('\p{Script=-/a/Glagolitic}');
Error('\P{Script=-/a/Glagolitic}');
Expect(1, 11358, '\p{Script=glagolitic}', "");
Expect(0, 11358, '\p{^Script=glagolitic}', "");
Expect(0, 11358, '\P{Script=glagolitic}', "");
Expect(1, 11358, '\P{^Script=glagolitic}', "");
Expect(0, 11359, '\p{Script=glagolitic}', "");
Expect(1, 11359, '\p{^Script=glagolitic}', "");
Expect(1, 11359, '\P{Script=glagolitic}', "");
Expect(0, 11359, '\P{^Script=glagolitic}', "");
Expect(1, 11358, '\p{Script=__GLAGOLITIC}', "");
Expect(0, 11358, '\p{^Script=__GLAGOLITIC}', "");
Expect(0, 11358, '\P{Script=__GLAGOLITIC}', "");
Expect(1, 11358, '\P{^Script=__GLAGOLITIC}', "");
Expect(0, 11359, '\p{Script=__GLAGOLITIC}', "");
Expect(1, 11359, '\p{^Script=__GLAGOLITIC}', "");
Expect(1, 11359, '\P{Script=__GLAGOLITIC}', "");
Expect(0, 11359, '\P{^Script=__GLAGOLITIC}', "");
Error('\p{Sc=--glag/a/}');
Error('\P{Sc=--glag/a/}');
Expect(1, 11358, '\p{Sc=glag}', "");
Expect(0, 11358, '\p{^Sc=glag}', "");
Expect(0, 11358, '\P{Sc=glag}', "");
Expect(1, 11358, '\P{^Sc=glag}', "");
Expect(0, 11359, '\p{Sc=glag}', "");
Expect(1, 11359, '\p{^Sc=glag}', "");
Expect(1, 11359, '\P{Sc=glag}', "");
Expect(0, 11359, '\P{^Sc=glag}', "");
Expect(1, 11358, '\p{Sc=-GLAG}', "");
Expect(0, 11358, '\p{^Sc=-GLAG}', "");
Expect(0, 11358, '\P{Sc=-GLAG}', "");
Expect(1, 11358, '\P{^Sc=-GLAG}', "");
Expect(0, 11359, '\p{Sc=-GLAG}', "");
Expect(1, 11359, '\p{^Sc=-GLAG}', "");
Expect(1, 11359, '\P{Sc=-GLAG}', "");
Expect(0, 11359, '\P{^Sc=-GLAG}', "");
Error('\p{Is_Script=/a/-	glagolitic}');
Error('\P{Is_Script=/a/-	glagolitic}');
Expect(1, 11358, '\p{Is_Script=glagolitic}', "");
Expect(0, 11358, '\p{^Is_Script=glagolitic}', "");
Expect(0, 11358, '\P{Is_Script=glagolitic}', "");
Expect(1, 11358, '\P{^Is_Script=glagolitic}', "");
Expect(0, 11359, '\p{Is_Script=glagolitic}', "");
Expect(1, 11359, '\p{^Is_Script=glagolitic}', "");
Expect(1, 11359, '\P{Is_Script=glagolitic}', "");
Expect(0, 11359, '\P{^Is_Script=glagolitic}', "");
Expect(1, 11358, '\p{Is_Script=	GLAGOLITIC}', "");
Expect(0, 11358, '\p{^Is_Script=	GLAGOLITIC}', "");
Expect(0, 11358, '\P{Is_Script=	GLAGOLITIC}', "");
Expect(1, 11358, '\P{^Is_Script=	GLAGOLITIC}', "");
Expect(0, 11359, '\p{Is_Script=	GLAGOLITIC}', "");
Expect(1, 11359, '\p{^Is_Script=	GLAGOLITIC}', "");
Expect(1, 11359, '\P{Is_Script=	GLAGOLITIC}', "");
Expect(0, 11359, '\P{^Is_Script=	GLAGOLITIC}', "");
Error('\p{Is_Sc=/a/- GLAG}');
Error('\P{Is_Sc=/a/- GLAG}');
Expect(1, 11358, '\p{Is_Sc=glag}', "");
Expect(0, 11358, '\p{^Is_Sc=glag}', "");
Expect(0, 11358, '\P{Is_Sc=glag}', "");
Expect(1, 11358, '\P{^Is_Sc=glag}', "");
Expect(0, 11359, '\p{Is_Sc=glag}', "");
Expect(1, 11359, '\p{^Is_Sc=glag}', "");
Expect(1, 11359, '\P{Is_Sc=glag}', "");
Expect(0, 11359, '\P{^Is_Sc=glag}', "");
Expect(1, 11358, '\p{Is_Sc=	 glag}', "");
Expect(0, 11358, '\p{^Is_Sc=	 glag}', "");
Expect(0, 11358, '\P{Is_Sc=	 glag}', "");
Expect(1, 11358, '\P{^Is_Sc=	 glag}', "");
Expect(0, 11359, '\p{Is_Sc=	 glag}', "");
Expect(1, 11359, '\p{^Is_Sc=	 glag}', "");
Expect(1, 11359, '\P{Is_Sc=	 glag}', "");
Expect(0, 11359, '\P{^Is_Sc=	 glag}', "");
Error('\p{Script=:= Thaana}');
Error('\P{Script=:= Thaana}');
Expect(1, 1969, '\p{Script=thaana}', "");
Expect(0, 1969, '\p{^Script=thaana}', "");
Expect(0, 1969, '\P{Script=thaana}', "");
Expect(1, 1969, '\P{^Script=thaana}', "");
Expect(0, 1970, '\p{Script=thaana}', "");
Expect(1, 1970, '\p{^Script=thaana}', "");
Expect(1, 1970, '\P{Script=thaana}', "");
Expect(0, 1970, '\P{^Script=thaana}', "");
Expect(1, 1969, '\p{Script=_thaana}', "");
Expect(0, 1969, '\p{^Script=_thaana}', "");
Expect(0, 1969, '\P{Script=_thaana}', "");
Expect(1, 1969, '\P{^Script=_thaana}', "");
Expect(0, 1970, '\p{Script=_thaana}', "");
Expect(1, 1970, '\p{^Script=_thaana}', "");
Expect(1, 1970, '\P{Script=_thaana}', "");
Expect(0, 1970, '\P{^Script=_thaana}', "");
Error('\p{Sc=_/a/Thaa}');
Error('\P{Sc=_/a/Thaa}');
Expect(1, 1969, '\p{Sc=thaa}', "");
Expect(0, 1969, '\p{^Sc=thaa}', "");
Expect(0, 1969, '\P{Sc=thaa}', "");
Expect(1, 1969, '\P{^Sc=thaa}', "");
Expect(0, 1970, '\p{Sc=thaa}', "");
Expect(1, 1970, '\p{^Sc=thaa}', "");
Expect(1, 1970, '\P{Sc=thaa}', "");
Expect(0, 1970, '\P{^Sc=thaa}', "");
Expect(1, 1969, '\p{Sc=_Thaa}', "");
Expect(0, 1969, '\p{^Sc=_Thaa}', "");
Expect(0, 1969, '\P{Sc=_Thaa}', "");
Expect(1, 1969, '\P{^Sc=_Thaa}', "");
Expect(0, 1970, '\p{Sc=_Thaa}', "");
Expect(1, 1970, '\p{^Sc=_Thaa}', "");
Expect(1, 1970, '\P{Sc=_Thaa}', "");
Expect(0, 1970, '\P{^Sc=_Thaa}', "");
Error('\p{Is_Script=	THAANA/a/}');
Error('\P{Is_Script=	THAANA/a/}');
Expect(1, 1969, '\p{Is_Script=thaana}', "");
Expect(0, 1969, '\p{^Is_Script=thaana}', "");
Expect(0, 1969, '\P{Is_Script=thaana}', "");
Expect(1, 1969, '\P{^Is_Script=thaana}', "");
Expect(0, 1970, '\p{Is_Script=thaana}', "");
Expect(1, 1970, '\p{^Is_Script=thaana}', "");
Expect(1, 1970, '\P{Is_Script=thaana}', "");
Expect(0, 1970, '\P{^Is_Script=thaana}', "");
Expect(1, 1969, '\p{Is_Script=-_Thaana}', "");
Expect(0, 1969, '\p{^Is_Script=-_Thaana}', "");
Expect(0, 1969, '\P{Is_Script=-_Thaana}', "");
Expect(1, 1969, '\P{^Is_Script=-_Thaana}', "");
Expect(0, 1970, '\p{Is_Script=-_Thaana}', "");
Expect(1, 1970, '\p{^Is_Script=-_Thaana}', "");
Expect(1, 1970, '\P{Is_Script=-_Thaana}', "");
Expect(0, 1970, '\P{^Is_Script=-_Thaana}', "");
Error('\p{Is_Sc=_	Thaa:=}');
Error('\P{Is_Sc=_	Thaa:=}');
Expect(1, 1969, '\p{Is_Sc=thaa}', "");
Expect(0, 1969, '\p{^Is_Sc=thaa}', "");
Expect(0, 1969, '\P{Is_Sc=thaa}', "");
Expect(1, 1969, '\P{^Is_Sc=thaa}', "");
Expect(0, 1970, '\p{Is_Sc=thaa}', "");
Expect(1, 1970, '\p{^Is_Sc=thaa}', "");
Expect(1, 1970, '\P{Is_Sc=thaa}', "");
Expect(0, 1970, '\P{^Is_Sc=thaa}', "");
Expect(1, 1969, '\p{Is_Sc=	_Thaa}', "");
Expect(0, 1969, '\p{^Is_Sc=	_Thaa}', "");
Expect(0, 1969, '\P{Is_Sc=	_Thaa}', "");
Expect(1, 1969, '\P{^Is_Sc=	_Thaa}', "");
Expect(0, 1970, '\p{Is_Sc=	_Thaa}', "");
Expect(1, 1970, '\p{^Is_Sc=	_Thaa}', "");
Expect(1, 1970, '\P{Is_Sc=	_Thaa}', "");
Expect(0, 1970, '\P{^Is_Sc=	_Thaa}', "");
Error('\p{Script=:=	LEPCHA}');
Error('\P{Script=:=	LEPCHA}');
Expect(1, 7247, '\p{Script=lepcha}', "");
Expect(0, 7247, '\p{^Script=lepcha}', "");
Expect(0, 7247, '\P{Script=lepcha}', "");
Expect(1, 7247, '\P{^Script=lepcha}', "");
Expect(0, 7248, '\p{Script=lepcha}', "");
Expect(1, 7248, '\p{^Script=lepcha}', "");
Expect(1, 7248, '\P{Script=lepcha}', "");
Expect(0, 7248, '\P{^Script=lepcha}', "");
Expect(1, 7247, '\p{Script= 	lepcha}', "");
Expect(0, 7247, '\p{^Script= 	lepcha}', "");
Expect(0, 7247, '\P{Script= 	lepcha}', "");
Expect(1, 7247, '\P{^Script= 	lepcha}', "");
Expect(0, 7248, '\p{Script= 	lepcha}', "");
Expect(1, 7248, '\p{^Script= 	lepcha}', "");
Expect(1, 7248, '\P{Script= 	lepcha}', "");
Expect(0, 7248, '\P{^Script= 	lepcha}', "");
Error('\p{Sc= LEPC:=}');
Error('\P{Sc= LEPC:=}');
Expect(1, 7247, '\p{Sc=lepc}', "");
Expect(0, 7247, '\p{^Sc=lepc}', "");
Expect(0, 7247, '\P{Sc=lepc}', "");
Expect(1, 7247, '\P{^Sc=lepc}', "");
Expect(0, 7248, '\p{Sc=lepc}', "");
Expect(1, 7248, '\p{^Sc=lepc}', "");
Expect(1, 7248, '\P{Sc=lepc}', "");
Expect(0, 7248, '\P{^Sc=lepc}', "");
Expect(1, 7247, '\p{Sc= LEPC}', "");
Expect(0, 7247, '\p{^Sc= LEPC}', "");
Expect(0, 7247, '\P{Sc= LEPC}', "");
Expect(1, 7247, '\P{^Sc= LEPC}', "");
Expect(0, 7248, '\p{Sc= LEPC}', "");
Expect(1, 7248, '\p{^Sc= LEPC}', "");
Expect(1, 7248, '\P{Sc= LEPC}', "");
Expect(0, 7248, '\P{^Sc= LEPC}', "");
Error('\p{Is_Script=:=_Lepcha}');
Error('\P{Is_Script=:=_Lepcha}');
Expect(1, 7247, '\p{Is_Script=lepcha}', "");
Expect(0, 7247, '\p{^Is_Script=lepcha}', "");
Expect(0, 7247, '\P{Is_Script=lepcha}', "");
Expect(1, 7247, '\P{^Is_Script=lepcha}', "");
Expect(0, 7248, '\p{Is_Script=lepcha}', "");
Expect(1, 7248, '\p{^Is_Script=lepcha}', "");
Expect(1, 7248, '\P{Is_Script=lepcha}', "");
Expect(0, 7248, '\P{^Is_Script=lepcha}', "");
Expect(1, 7247, '\p{Is_Script= _lepcha}', "");
Expect(0, 7247, '\p{^Is_Script= _lepcha}', "");
Expect(0, 7247, '\P{Is_Script= _lepcha}', "");
Expect(1, 7247, '\P{^Is_Script= _lepcha}', "");
Expect(0, 7248, '\p{Is_Script= _lepcha}', "");
Expect(1, 7248, '\p{^Is_Script= _lepcha}', "");
Expect(1, 7248, '\P{Is_Script= _lepcha}', "");
Expect(0, 7248, '\P{^Is_Script= _lepcha}', "");
Error('\p{Is_Sc:   	/a/Lepc}');
Error('\P{Is_Sc:   	/a/Lepc}');
Expect(1, 7247, '\p{Is_Sc=lepc}', "");
Expect(0, 7247, '\p{^Is_Sc=lepc}', "");
Expect(0, 7247, '\P{Is_Sc=lepc}', "");
Expect(1, 7247, '\P{^Is_Sc=lepc}', "");
Expect(0, 7248, '\p{Is_Sc=lepc}', "");
Expect(1, 7248, '\p{^Is_Sc=lepc}', "");
Expect(1, 7248, '\P{Is_Sc=lepc}', "");
Expect(0, 7248, '\P{^Is_Sc=lepc}', "");
Expect(1, 7247, '\p{Is_Sc=	Lepc}', "");
Expect(0, 7247, '\p{^Is_Sc=	Lepc}', "");
Expect(0, 7247, '\P{Is_Sc=	Lepc}', "");
Expect(1, 7247, '\P{^Is_Sc=	Lepc}', "");
Expect(0, 7248, '\p{Is_Sc=	Lepc}', "");
Expect(1, 7248, '\p{^Is_Sc=	Lepc}', "");
Expect(1, 7248, '\P{Is_Sc=	Lepc}', "");
Expect(0, 7248, '\P{^Is_Sc=	Lepc}', "");
Error('\p{Script=/a/		coptic}');
Error('\P{Script=/a/		coptic}');
Expect(1, 11519, '\p{Script=coptic}', "");
Expect(0, 11519, '\p{^Script=coptic}', "");
Expect(0, 11519, '\P{Script=coptic}', "");
Expect(1, 11519, '\P{^Script=coptic}', "");
Expect(0, 11520, '\p{Script=coptic}', "");
Expect(1, 11520, '\p{^Script=coptic}', "");
Expect(1, 11520, '\P{Script=coptic}', "");
Expect(0, 11520, '\P{^Script=coptic}', "");
Expect(1, 11519, '\p{Script=- Coptic}', "");
Expect(0, 11519, '\p{^Script=- Coptic}', "");
Expect(0, 11519, '\P{Script=- Coptic}', "");
Expect(1, 11519, '\P{^Script=- Coptic}', "");
Expect(0, 11520, '\p{Script=- Coptic}', "");
Expect(1, 11520, '\p{^Script=- Coptic}', "");
Expect(1, 11520, '\P{Script=- Coptic}', "");
Expect(0, 11520, '\P{^Script=- Coptic}', "");
Error('\p{Sc=/a/ 	Copt}');
Error('\P{Sc=/a/ 	Copt}');
Expect(1, 11519, '\p{Sc=copt}', "");
Expect(0, 11519, '\p{^Sc=copt}', "");
Expect(0, 11519, '\P{Sc=copt}', "");
Expect(1, 11519, '\P{^Sc=copt}', "");
Expect(0, 11520, '\p{Sc=copt}', "");
Expect(1, 11520, '\p{^Sc=copt}', "");
Expect(1, 11520, '\P{Sc=copt}', "");
Expect(0, 11520, '\P{^Sc=copt}', "");
Expect(1, 11519, '\p{Sc=--COPT}', "");
Expect(0, 11519, '\p{^Sc=--COPT}', "");
Expect(0, 11519, '\P{Sc=--COPT}', "");
Expect(1, 11519, '\P{^Sc=--COPT}', "");
Expect(0, 11520, '\p{Sc=--COPT}', "");
Expect(1, 11520, '\p{^Sc=--COPT}', "");
Expect(1, 11520, '\P{Sc=--COPT}', "");
Expect(0, 11520, '\P{^Sc=--COPT}', "");
Error('\p{Is_Script= _Qaac/a/}');
Error('\P{Is_Script= _Qaac/a/}');
Expect(1, 11519, '\p{Is_Script=qaac}', "");
Expect(0, 11519, '\p{^Is_Script=qaac}', "");
Expect(0, 11519, '\P{Is_Script=qaac}', "");
Expect(1, 11519, '\P{^Is_Script=qaac}', "");
Expect(0, 11520, '\p{Is_Script=qaac}', "");
Expect(1, 11520, '\p{^Is_Script=qaac}', "");
Expect(1, 11520, '\P{Is_Script=qaac}', "");
Expect(0, 11520, '\P{^Is_Script=qaac}', "");
Expect(1, 11519, '\p{Is_Script=		qaac}', "");
Expect(0, 11519, '\p{^Is_Script=		qaac}', "");
Expect(0, 11519, '\P{Is_Script=		qaac}', "");
Expect(1, 11519, '\P{^Is_Script=		qaac}', "");
Expect(0, 11520, '\p{Is_Script=		qaac}', "");
Expect(1, 11520, '\p{^Is_Script=		qaac}', "");
Expect(1, 11520, '\P{Is_Script=		qaac}', "");
Expect(0, 11520, '\P{^Is_Script=		qaac}', "");
Error('\p{Is_Sc=:= 	COPTIC}');
Error('\P{Is_Sc=:= 	COPTIC}');
Expect(1, 11519, '\p{Is_Sc=coptic}', "");
Expect(0, 11519, '\p{^Is_Sc=coptic}', "");
Expect(0, 11519, '\P{Is_Sc=coptic}', "");
Expect(1, 11519, '\P{^Is_Sc=coptic}', "");
Expect(0, 11520, '\p{Is_Sc=coptic}', "");
Expect(1, 11520, '\p{^Is_Sc=coptic}', "");
Expect(1, 11520, '\P{Is_Sc=coptic}', "");
Expect(0, 11520, '\P{^Is_Sc=coptic}', "");
Expect(1, 11519, '\p{Is_Sc=- Coptic}', "");
Expect(0, 11519, '\p{^Is_Sc=- Coptic}', "");
Expect(0, 11519, '\P{Is_Sc=- Coptic}', "");
Expect(1, 11519, '\P{^Is_Sc=- Coptic}', "");
Expect(0, 11520, '\p{Is_Sc=- Coptic}', "");
Expect(1, 11520, '\p{^Is_Sc=- Coptic}', "");
Expect(1, 11520, '\P{Is_Sc=- Coptic}', "");
Expect(0, 11520, '\P{^Is_Sc=- Coptic}', "");
Error('\p{Script=Katakana_Or_Hiragana}');
Error('\P{Script=Katakana_Or_Hiragana}');
Error('\p{Sc=Hrkt}');
Error('\P{Sc=Hrkt}');
Error('\p{Is_Script=Katakana_Or_Hiragana}');
Error('\P{Is_Script=Katakana_Or_Hiragana}');
Error('\p{Is_Sc=Hrkt}');
Error('\P{Is_Sc=Hrkt}');
Error('\p{Script=_/a/BALINESE}');
Error('\P{Script=_/a/BALINESE}');
Expect(1, 7036, '\p{Script=balinese}', "");
Expect(0, 7036, '\p{^Script=balinese}', "");
Expect(0, 7036, '\P{Script=balinese}', "");
Expect(1, 7036, '\P{^Script=balinese}', "");
Expect(0, 7037, '\p{Script=balinese}', "");
Expect(1, 7037, '\p{^Script=balinese}', "");
Expect(1, 7037, '\P{Script=balinese}', "");
Expect(0, 7037, '\P{^Script=balinese}', "");
Expect(1, 7036, '\p{Script:  BALINESE}', "");
Expect(0, 7036, '\p{^Script:  BALINESE}', "");
Expect(0, 7036, '\P{Script:  BALINESE}', "");
Expect(1, 7036, '\P{^Script:  BALINESE}', "");
Expect(0, 7037, '\p{Script:  BALINESE}', "");
Expect(1, 7037, '\p{^Script:  BALINESE}', "");
Expect(1, 7037, '\P{Script:  BALINESE}', "");
Expect(0, 7037, '\P{^Script:  BALINESE}', "");
Error('\p{Sc=	/a/Bali}');
Error('\P{Sc=	/a/Bali}');
Expect(1, 7036, '\p{Sc=bali}', "");
Expect(0, 7036, '\p{^Sc=bali}', "");
Expect(0, 7036, '\P{Sc=bali}', "");
Expect(1, 7036, '\P{^Sc=bali}', "");
Expect(0, 7037, '\p{Sc=bali}', "");
Expect(1, 7037, '\p{^Sc=bali}', "");
Expect(1, 7037, '\P{Sc=bali}', "");
Expect(0, 7037, '\P{^Sc=bali}', "");
Expect(1, 7036, '\p{Sc= BALI}', "");
Expect(0, 7036, '\p{^Sc= BALI}', "");
Expect(0, 7036, '\P{Sc= BALI}', "");
Expect(1, 7036, '\P{^Sc= BALI}', "");
Expect(0, 7037, '\p{Sc= BALI}', "");
Expect(1, 7037, '\p{^Sc= BALI}', "");
Expect(1, 7037, '\P{Sc= BALI}', "");
Expect(0, 7037, '\P{^Sc= BALI}', "");
Error('\p{Is_Script=-Balinese:=}');
Error('\P{Is_Script=-Balinese:=}');
Expect(1, 7036, '\p{Is_Script=balinese}', "");
Expect(0, 7036, '\p{^Is_Script=balinese}', "");
Expect(0, 7036, '\P{Is_Script=balinese}', "");
Expect(1, 7036, '\P{^Is_Script=balinese}', "");
Expect(0, 7037, '\p{Is_Script=balinese}', "");
Expect(1, 7037, '\p{^Is_Script=balinese}', "");
Expect(1, 7037, '\P{Is_Script=balinese}', "");
Expect(0, 7037, '\P{^Is_Script=balinese}', "");
Expect(1, 7036, '\p{Is_Script=	BALINESE}', "");
Expect(0, 7036, '\p{^Is_Script=	BALINESE}', "");
Expect(0, 7036, '\P{Is_Script=	BALINESE}', "");
Expect(1, 7036, '\P{^Is_Script=	BALINESE}', "");
Expect(0, 7037, '\p{Is_Script=	BALINESE}', "");
Expect(1, 7037, '\p{^Is_Script=	BALINESE}', "");
Expect(1, 7037, '\P{Is_Script=	BALINESE}', "");
Expect(0, 7037, '\P{^Is_Script=	BALINESE}', "");
Error('\p{Is_Sc: Bali:=}');
Error('\P{Is_Sc: Bali:=}');
Expect(1, 7036, '\p{Is_Sc=bali}', "");
Expect(0, 7036, '\p{^Is_Sc=bali}', "");
Expect(0, 7036, '\P{Is_Sc=bali}', "");
Expect(1, 7036, '\P{^Is_Sc=bali}', "");
Expect(0, 7037, '\p{Is_Sc=bali}', "");
Expect(1, 7037, '\p{^Is_Sc=bali}', "");
Expect(1, 7037, '\P{Is_Sc=bali}', "");
Expect(0, 7037, '\P{^Is_Sc=bali}', "");
Expect(1, 7036, '\p{Is_Sc=-_Bali}', "");
Expect(0, 7036, '\p{^Is_Sc=-_Bali}', "");
Expect(0, 7036, '\P{Is_Sc=-_Bali}', "");
Expect(1, 7036, '\P{^Is_Sc=-_Bali}', "");
Expect(0, 7037, '\p{Is_Sc=-_Bali}', "");
Expect(1, 7037, '\p{^Is_Sc=-_Bali}', "");
Expect(1, 7037, '\P{Is_Sc=-_Bali}', "");
Expect(0, 7037, '\P{^Is_Sc=-_Bali}', "");
Error('\p{Script=	-Tagbanwa:=}');
Error('\P{Script=	-Tagbanwa:=}');
Expect(1, 6003, '\p{Script:   tagbanwa}', "");
Expect(0, 6003, '\p{^Script:   tagbanwa}', "");
Expect(0, 6003, '\P{Script:   tagbanwa}', "");
Expect(1, 6003, '\P{^Script:   tagbanwa}', "");
Expect(0, 6004, '\p{Script:   tagbanwa}', "");
Expect(1, 6004, '\p{^Script:   tagbanwa}', "");
Expect(1, 6004, '\P{Script:   tagbanwa}', "");
Expect(0, 6004, '\P{^Script:   tagbanwa}', "");
Expect(1, 6003, '\p{Script= -Tagbanwa}', "");
Expect(0, 6003, '\p{^Script= -Tagbanwa}', "");
Expect(0, 6003, '\P{Script= -Tagbanwa}', "");
Expect(1, 6003, '\P{^Script= -Tagbanwa}', "");
Expect(0, 6004, '\p{Script= -Tagbanwa}', "");
Expect(1, 6004, '\p{^Script= -Tagbanwa}', "");
Expect(1, 6004, '\P{Script= -Tagbanwa}', "");
Expect(0, 6004, '\P{^Script= -Tagbanwa}', "");
Error('\p{Sc=/a/ Tagb}');
Error('\P{Sc=/a/ Tagb}');
Expect(1, 6003, '\p{Sc:   tagb}', "");
Expect(0, 6003, '\p{^Sc:   tagb}', "");
Expect(0, 6003, '\P{Sc:   tagb}', "");
Expect(1, 6003, '\P{^Sc:   tagb}', "");
Expect(0, 6004, '\p{Sc:   tagb}', "");
Expect(1, 6004, '\p{^Sc:   tagb}', "");
Expect(1, 6004, '\P{Sc:   tagb}', "");
Expect(0, 6004, '\P{^Sc:   tagb}', "");
Expect(1, 6003, '\p{Sc=_-TAGB}', "");
Expect(0, 6003, '\p{^Sc=_-TAGB}', "");
Expect(0, 6003, '\P{Sc=_-TAGB}', "");
Expect(1, 6003, '\P{^Sc=_-TAGB}', "");
Expect(0, 6004, '\p{Sc=_-TAGB}', "");
Expect(1, 6004, '\p{^Sc=_-TAGB}', "");
Expect(1, 6004, '\P{Sc=_-TAGB}', "");
Expect(0, 6004, '\P{^Sc=_-TAGB}', "");
Error('\p{Is_Script=/a/ -Tagbanwa}');
Error('\P{Is_Script=/a/ -Tagbanwa}');
Expect(1, 6003, '\p{Is_Script=tagbanwa}', "");
Expect(0, 6003, '\p{^Is_Script=tagbanwa}', "");
Expect(0, 6003, '\P{Is_Script=tagbanwa}', "");
Expect(1, 6003, '\P{^Is_Script=tagbanwa}', "");
Expect(0, 6004, '\p{Is_Script=tagbanwa}', "");
Expect(1, 6004, '\p{^Is_Script=tagbanwa}', "");
Expect(1, 6004, '\P{Is_Script=tagbanwa}', "");
Expect(0, 6004, '\P{^Is_Script=tagbanwa}', "");
Expect(1, 6003, '\p{Is_Script= Tagbanwa}', "");
Expect(0, 6003, '\p{^Is_Script= Tagbanwa}', "");
Expect(0, 6003, '\P{Is_Script= Tagbanwa}', "");
Expect(1, 6003, '\P{^Is_Script= Tagbanwa}', "");
Expect(0, 6004, '\p{Is_Script= Tagbanwa}', "");
Expect(1, 6004, '\p{^Is_Script= Tagbanwa}', "");
Expect(1, 6004, '\P{Is_Script= Tagbanwa}', "");
Expect(0, 6004, '\P{^Is_Script= Tagbanwa}', "");
Error('\p{Is_Sc= _Tagb:=}');
Error('\P{Is_Sc= _Tagb:=}');
Expect(1, 6003, '\p{Is_Sc=tagb}', "");
Expect(0, 6003, '\p{^Is_Sc=tagb}', "");
Expect(0, 6003, '\P{Is_Sc=tagb}', "");
Expect(1, 6003, '\P{^Is_Sc=tagb}', "");
Expect(0, 6004, '\p{Is_Sc=tagb}', "");
Expect(1, 6004, '\p{^Is_Sc=tagb}', "");
Expect(1, 6004, '\P{Is_Sc=tagb}', "");
Expect(0, 6004, '\P{^Is_Sc=tagb}', "");
Expect(1, 6003, '\p{Is_Sc=	-TAGB}', "");
Expect(0, 6003, '\p{^Is_Sc=	-TAGB}', "");
Expect(0, 6003, '\P{Is_Sc=	-TAGB}', "");
Expect(1, 6003, '\P{^Is_Sc=	-TAGB}', "");
Expect(0, 6004, '\p{Is_Sc=	-TAGB}', "");
Expect(1, 6004, '\p{^Is_Sc=	-TAGB}', "");
Expect(1, 6004, '\P{Is_Sc=	-TAGB}', "");
Expect(0, 6004, '\P{^Is_Sc=	-TAGB}', "");
Error('\p{Script=/a/ 	SYLOTI_NAGRI}');
Error('\P{Script=/a/ 	SYLOTI_NAGRI}');
Expect(1, 43051, '\p{Script=sylotinagri}', "");
Expect(0, 43051, '\p{^Script=sylotinagri}', "");
Expect(0, 43051, '\P{Script=sylotinagri}', "");
Expect(1, 43051, '\P{^Script=sylotinagri}', "");
Expect(0, 43052, '\p{Script=sylotinagri}', "");
Expect(1, 43052, '\p{^Script=sylotinagri}', "");
Expect(1, 43052, '\P{Script=sylotinagri}', "");
Expect(0, 43052, '\P{^Script=sylotinagri}', "");
Expect(1, 43051, '\p{Script=_ SYLOTI_nagri}', "");
Expect(0, 43051, '\p{^Script=_ SYLOTI_nagri}', "");
Expect(0, 43051, '\P{Script=_ SYLOTI_nagri}', "");
Expect(1, 43051, '\P{^Script=_ SYLOTI_nagri}', "");
Expect(0, 43052, '\p{Script=_ SYLOTI_nagri}', "");
Expect(1, 43052, '\p{^Script=_ SYLOTI_nagri}', "");
Expect(1, 43052, '\P{Script=_ SYLOTI_nagri}', "");
Expect(0, 43052, '\P{^Script=_ SYLOTI_nagri}', "");
Error('\p{Sc=:=- Sylo}');
Error('\P{Sc=:=- Sylo}');
Expect(1, 43051, '\p{Sc=sylo}', "");
Expect(0, 43051, '\p{^Sc=sylo}', "");
Expect(0, 43051, '\P{Sc=sylo}', "");
Expect(1, 43051, '\P{^Sc=sylo}', "");
Expect(0, 43052, '\p{Sc=sylo}', "");
Expect(1, 43052, '\p{^Sc=sylo}', "");
Expect(1, 43052, '\P{Sc=sylo}', "");
Expect(0, 43052, '\P{^Sc=sylo}', "");
Expect(1, 43051, '\p{Sc=-	sylo}', "");
Expect(0, 43051, '\p{^Sc=-	sylo}', "");
Expect(0, 43051, '\P{Sc=-	sylo}', "");
Expect(1, 43051, '\P{^Sc=-	sylo}', "");
Expect(0, 43052, '\p{Sc=-	sylo}', "");
Expect(1, 43052, '\p{^Sc=-	sylo}', "");
Expect(1, 43052, '\P{Sc=-	sylo}', "");
Expect(0, 43052, '\P{^Sc=-	sylo}', "");
Error('\p{Is_Script=/a/		Syloti_Nagri}');
Error('\P{Is_Script=/a/		Syloti_Nagri}');
Expect(1, 43051, '\p{Is_Script=sylotinagri}', "");
Expect(0, 43051, '\p{^Is_Script=sylotinagri}', "");
Expect(0, 43051, '\P{Is_Script=sylotinagri}', "");
Expect(1, 43051, '\P{^Is_Script=sylotinagri}', "");
Expect(0, 43052, '\p{Is_Script=sylotinagri}', "");
Expect(1, 43052, '\p{^Is_Script=sylotinagri}', "");
Expect(1, 43052, '\P{Is_Script=sylotinagri}', "");
Expect(0, 43052, '\P{^Is_Script=sylotinagri}', "");
Expect(1, 43051, '\p{Is_Script=	 Syloti_Nagri}', "");
Expect(0, 43051, '\p{^Is_Script=	 Syloti_Nagri}', "");
Expect(0, 43051, '\P{Is_Script=	 Syloti_Nagri}', "");
Expect(1, 43051, '\P{^Is_Script=	 Syloti_Nagri}', "");
Expect(0, 43052, '\p{Is_Script=	 Syloti_Nagri}', "");
Expect(1, 43052, '\p{^Is_Script=	 Syloti_Nagri}', "");
Expect(1, 43052, '\P{Is_Script=	 Syloti_Nagri}', "");
Expect(0, 43052, '\P{^Is_Script=	 Syloti_Nagri}', "");
Error('\p{Is_Sc=-SYLO:=}');
Error('\P{Is_Sc=-SYLO:=}');
Expect(1, 43051, '\p{Is_Sc=sylo}', "");
Expect(0, 43051, '\p{^Is_Sc=sylo}', "");
Expect(0, 43051, '\P{Is_Sc=sylo}', "");
Expect(1, 43051, '\P{^Is_Sc=sylo}', "");
Expect(0, 43052, '\p{Is_Sc=sylo}', "");
Expect(1, 43052, '\p{^Is_Sc=sylo}', "");
Expect(1, 43052, '\P{Is_Sc=sylo}', "");
Expect(0, 43052, '\P{^Is_Sc=sylo}', "");
Expect(1, 43051, '\p{Is_Sc=	-Sylo}', "");
Expect(0, 43051, '\p{^Is_Sc=	-Sylo}', "");
Expect(0, 43051, '\P{Is_Sc=	-Sylo}', "");
Expect(1, 43051, '\P{^Is_Sc=	-Sylo}', "");
Expect(0, 43052, '\p{Is_Sc=	-Sylo}', "");
Expect(1, 43052, '\p{^Is_Sc=	-Sylo}', "");
Expect(1, 43052, '\P{Is_Sc=	-Sylo}', "");
Expect(0, 43052, '\P{^Is_Sc=	-Sylo}', "");
Error('\p{Script= 	MEETEI_MAYEK/a/}');
Error('\P{Script= 	MEETEI_MAYEK/a/}');
Expect(1, 44025, '\p{Script=meeteimayek}', "");
Expect(0, 44025, '\p{^Script=meeteimayek}', "");
Expect(0, 44025, '\P{Script=meeteimayek}', "");
Expect(1, 44025, '\P{^Script=meeteimayek}', "");
Expect(0, 44026, '\p{Script=meeteimayek}', "");
Expect(1, 44026, '\p{^Script=meeteimayek}', "");
Expect(1, 44026, '\P{Script=meeteimayek}', "");
Expect(0, 44026, '\P{^Script=meeteimayek}', "");
Expect(1, 44025, '\p{Script=  Meetei_mayek}', "");
Expect(0, 44025, '\p{^Script=  Meetei_mayek}', "");
Expect(0, 44025, '\P{Script=  Meetei_mayek}', "");
Expect(1, 44025, '\P{^Script=  Meetei_mayek}', "");
Expect(0, 44026, '\p{Script=  Meetei_mayek}', "");
Expect(1, 44026, '\p{^Script=  Meetei_mayek}', "");
Expect(1, 44026, '\P{Script=  Meetei_mayek}', "");
Expect(0, 44026, '\P{^Script=  Meetei_mayek}', "");
Error('\p{Sc=-/a/mtei}');
Error('\P{Sc=-/a/mtei}');
Expect(1, 44025, '\p{Sc=mtei}', "");
Expect(0, 44025, '\p{^Sc=mtei}', "");
Expect(0, 44025, '\P{Sc=mtei}', "");
Expect(1, 44025, '\P{^Sc=mtei}', "");
Expect(0, 44026, '\p{Sc=mtei}', "");
Expect(1, 44026, '\p{^Sc=mtei}', "");
Expect(1, 44026, '\P{Sc=mtei}', "");
Expect(0, 44026, '\P{^Sc=mtei}', "");
Expect(1, 44025, '\p{Sc= 	MTEI}', "");
Expect(0, 44025, '\p{^Sc= 	MTEI}', "");
Expect(0, 44025, '\P{Sc= 	MTEI}', "");
Expect(1, 44025, '\P{^Sc= 	MTEI}', "");
Expect(0, 44026, '\p{Sc= 	MTEI}', "");
Expect(1, 44026, '\p{^Sc= 	MTEI}', "");
Expect(1, 44026, '\P{Sc= 	MTEI}', "");
Expect(0, 44026, '\P{^Sc= 	MTEI}', "");
Error('\p{Is_Script=:=- Meetei_mayek}');
Error('\P{Is_Script=:=- Meetei_mayek}');
Expect(1, 44025, '\p{Is_Script=meeteimayek}', "");
Expect(0, 44025, '\p{^Is_Script=meeteimayek}', "");
Expect(0, 44025, '\P{Is_Script=meeteimayek}', "");
Expect(1, 44025, '\P{^Is_Script=meeteimayek}', "");
Expect(0, 44026, '\p{Is_Script=meeteimayek}', "");
Expect(1, 44026, '\p{^Is_Script=meeteimayek}', "");
Expect(1, 44026, '\P{Is_Script=meeteimayek}', "");
Expect(0, 44026, '\P{^Is_Script=meeteimayek}', "");
Expect(1, 44025, '\p{Is_Script:	- meetei_MAYEK}', "");
Expect(0, 44025, '\p{^Is_Script:	- meetei_MAYEK}', "");
Expect(0, 44025, '\P{Is_Script:	- meetei_MAYEK}', "");
Expect(1, 44025, '\P{^Is_Script:	- meetei_MAYEK}', "");
Expect(0, 44026, '\p{Is_Script:	- meetei_MAYEK}', "");
Expect(1, 44026, '\p{^Is_Script:	- meetei_MAYEK}', "");
Expect(1, 44026, '\P{Is_Script:	- meetei_MAYEK}', "");
Expect(0, 44026, '\P{^Is_Script:	- meetei_MAYEK}', "");
Error('\p{Is_Sc= Mtei/a/}');
Error('\P{Is_Sc= Mtei/a/}');
Expect(1, 44025, '\p{Is_Sc: mtei}', "");
Expect(0, 44025, '\p{^Is_Sc: mtei}', "");
Expect(0, 44025, '\P{Is_Sc: mtei}', "");
Expect(1, 44025, '\P{^Is_Sc: mtei}', "");
Expect(0, 44026, '\p{Is_Sc: mtei}', "");
Expect(1, 44026, '\p{^Is_Sc: mtei}', "");
Expect(1, 44026, '\P{Is_Sc: mtei}', "");
Expect(0, 44026, '\P{^Is_Sc: mtei}', "");
Expect(1, 44025, '\p{Is_Sc= mtei}', "");
Expect(0, 44025, '\p{^Is_Sc= mtei}', "");
Expect(0, 44025, '\P{Is_Sc= mtei}', "");
Expect(1, 44025, '\P{^Is_Sc= mtei}', "");
Expect(0, 44026, '\p{Is_Sc= mtei}', "");
Expect(1, 44026, '\p{^Is_Sc= mtei}', "");
Expect(1, 44026, '\P{Is_Sc= mtei}', "");
Expect(0, 44026, '\P{^Is_Sc= mtei}', "");
Error('\p{Script=_:=PHOENICIAN}');
Error('\P{Script=_:=PHOENICIAN}');
Expect(1, 67871, '\p{Script=phoenician}', "");
Expect(0, 67871, '\p{^Script=phoenician}', "");
Expect(0, 67871, '\P{Script=phoenician}', "");
Expect(1, 67871, '\P{^Script=phoenician}', "");
Expect(0, 67872, '\p{Script=phoenician}', "");
Expect(1, 67872, '\p{^Script=phoenician}', "");
Expect(1, 67872, '\P{Script=phoenician}', "");
Expect(0, 67872, '\P{^Script=phoenician}', "");
Expect(1, 67871, '\p{Script= 	Phoenician}', "");
Expect(0, 67871, '\p{^Script= 	Phoenician}', "");
Expect(0, 67871, '\P{Script= 	Phoenician}', "");
Expect(1, 67871, '\P{^Script= 	Phoenician}', "");
Expect(0, 67872, '\p{Script= 	Phoenician}', "");
Expect(1, 67872, '\p{^Script= 	Phoenician}', "");
Expect(1, 67872, '\P{Script= 	Phoenician}', "");
Expect(0, 67872, '\P{^Script= 	Phoenician}', "");
Error('\p{Sc= _Phnx/a/}');
Error('\P{Sc= _Phnx/a/}');
Expect(1, 67871, '\p{Sc=phnx}', "");
Expect(0, 67871, '\p{^Sc=phnx}', "");
Expect(0, 67871, '\P{Sc=phnx}', "");
Expect(1, 67871, '\P{^Sc=phnx}', "");
Expect(0, 67872, '\p{Sc=phnx}', "");
Expect(1, 67872, '\p{^Sc=phnx}', "");
Expect(1, 67872, '\P{Sc=phnx}', "");
Expect(0, 67872, '\P{^Sc=phnx}', "");
Expect(1, 67871, '\p{Sc:			Phnx}', "");
Expect(0, 67871, '\p{^Sc:			Phnx}', "");
Expect(0, 67871, '\P{Sc:			Phnx}', "");
Expect(1, 67871, '\P{^Sc:			Phnx}', "");
Expect(0, 67872, '\p{Sc:			Phnx}', "");
Expect(1, 67872, '\p{^Sc:			Phnx}', "");
Expect(1, 67872, '\P{Sc:			Phnx}', "");
Expect(0, 67872, '\P{^Sc:			Phnx}', "");
Error('\p{Is_Script= Phoenician/a/}');
Error('\P{Is_Script= Phoenician/a/}');
Expect(1, 67871, '\p{Is_Script=phoenician}', "");
Expect(0, 67871, '\p{^Is_Script=phoenician}', "");
Expect(0, 67871, '\P{Is_Script=phoenician}', "");
Expect(1, 67871, '\P{^Is_Script=phoenician}', "");
Expect(0, 67872, '\p{Is_Script=phoenician}', "");
Expect(1, 67872, '\p{^Is_Script=phoenician}', "");
Expect(1, 67872, '\P{Is_Script=phoenician}', "");
Expect(0, 67872, '\P{^Is_Script=phoenician}', "");
Error('\p{Is_Sc=_Phnx:=}');
Error('\P{Is_Sc=_Phnx:=}');
Expect(1, 67871, '\p{Is_Sc=phnx}', "");
Expect(0, 67871, '\p{^Is_Sc=phnx}', "");
Expect(0, 67871, '\P{Is_Sc=phnx}', "");
Expect(1, 67871, '\P{^Is_Sc=phnx}', "");
Expect(0, 67872, '\p{Is_Sc=phnx}', "");
Expect(1, 67872, '\p{^Is_Sc=phnx}', "");
Expect(1, 67872, '\P{Is_Sc=phnx}', "");
Expect(0, 67872, '\P{^Is_Sc=phnx}', "");
Expect(1, 67871, '\p{Is_Sc=--PHNX}', "");
Expect(0, 67871, '\p{^Is_Sc=--PHNX}', "");
Expect(0, 67871, '\P{Is_Sc=--PHNX}', "");
Expect(1, 67871, '\P{^Is_Sc=--PHNX}', "");
Expect(0, 67872, '\p{Is_Sc=--PHNX}', "");
Expect(1, 67872, '\p{^Is_Sc=--PHNX}', "");
Expect(1, 67872, '\P{Is_Sc=--PHNX}', "");
Expect(0, 67872, '\P{^Is_Sc=--PHNX}', "");
Error('\p{Script:	:=DEVANAGARI}');
Error('\P{Script:	:=DEVANAGARI}');
Expect(1, 43259, '\p{Script=devanagari}', "");
Expect(0, 43259, '\p{^Script=devanagari}', "");
Expect(0, 43259, '\P{Script=devanagari}', "");
Expect(1, 43259, '\P{^Script=devanagari}', "");
Expect(0, 43260, '\p{Script=devanagari}', "");
Expect(1, 43260, '\p{^Script=devanagari}', "");
Expect(1, 43260, '\P{Script=devanagari}', "");
Expect(0, 43260, '\P{^Script=devanagari}', "");
Expect(1, 43259, '\p{Script=_ Devanagari}', "");
Expect(0, 43259, '\p{^Script=_ Devanagari}', "");
Expect(0, 43259, '\P{Script=_ Devanagari}', "");
Expect(1, 43259, '\P{^Script=_ Devanagari}', "");
Expect(0, 43260, '\p{Script=_ Devanagari}', "");
Expect(1, 43260, '\p{^Script=_ Devanagari}', "");
Expect(1, 43260, '\P{Script=_ Devanagari}', "");
Expect(0, 43260, '\P{^Script=_ Devanagari}', "");
Error('\p{Sc= DEVA:=}');
Error('\P{Sc= DEVA:=}');
Expect(1, 43259, '\p{Sc=deva}', "");
Expect(0, 43259, '\p{^Sc=deva}', "");
Expect(0, 43259, '\P{Sc=deva}', "");
Expect(1, 43259, '\P{^Sc=deva}', "");
Expect(0, 43260, '\p{Sc=deva}', "");
Expect(1, 43260, '\p{^Sc=deva}', "");
Expect(1, 43260, '\P{Sc=deva}', "");
Expect(0, 43260, '\P{^Sc=deva}', "");
Expect(1, 43259, '\p{Sc: --Deva}', "");
Expect(0, 43259, '\p{^Sc: --Deva}', "");
Expect(0, 43259, '\P{Sc: --Deva}', "");
Expect(1, 43259, '\P{^Sc: --Deva}', "");
Expect(0, 43260, '\p{Sc: --Deva}', "");
Expect(1, 43260, '\p{^Sc: --Deva}', "");
Expect(1, 43260, '\P{Sc: --Deva}', "");
Expect(0, 43260, '\P{^Sc: --Deva}', "");
Error('\p{Is_Script=-_Devanagari:=}');
Error('\P{Is_Script=-_Devanagari:=}');
Expect(1, 43259, '\p{Is_Script=devanagari}', "");
Expect(0, 43259, '\p{^Is_Script=devanagari}', "");
Expect(0, 43259, '\P{Is_Script=devanagari}', "");
Expect(1, 43259, '\P{^Is_Script=devanagari}', "");
Expect(0, 43260, '\p{Is_Script=devanagari}', "");
Expect(1, 43260, '\p{^Is_Script=devanagari}', "");
Expect(1, 43260, '\P{Is_Script=devanagari}', "");
Expect(0, 43260, '\P{^Is_Script=devanagari}', "");
Expect(1, 43259, '\p{Is_Script= -devanagari}', "");
Expect(0, 43259, '\p{^Is_Script= -devanagari}', "");
Expect(0, 43259, '\P{Is_Script= -devanagari}', "");
Expect(1, 43259, '\P{^Is_Script= -devanagari}', "");
Expect(0, 43260, '\p{Is_Script= -devanagari}', "");
Expect(1, 43260, '\p{^Is_Script= -devanagari}', "");
Expect(1, 43260, '\P{Is_Script= -devanagari}', "");
Expect(0, 43260, '\P{^Is_Script= -devanagari}', "");
Error('\p{Is_Sc=/a/Deva}');
Error('\P{Is_Sc=/a/Deva}');
Expect(1, 43259, '\p{Is_Sc=deva}', "");
Expect(0, 43259, '\p{^Is_Sc=deva}', "");
Expect(0, 43259, '\P{Is_Sc=deva}', "");
Expect(1, 43259, '\P{^Is_Sc=deva}', "");
Expect(0, 43260, '\p{Is_Sc=deva}', "");
Expect(1, 43260, '\p{^Is_Sc=deva}', "");
Expect(1, 43260, '\P{Is_Sc=deva}', "");
Expect(0, 43260, '\P{^Is_Sc=deva}', "");
Expect(1, 43259, '\p{Is_Sc=-Deva}', "");
Expect(0, 43259, '\p{^Is_Sc=-Deva}', "");
Expect(0, 43259, '\P{Is_Sc=-Deva}', "");
Expect(1, 43259, '\P{^Is_Sc=-Deva}', "");
Expect(0, 43260, '\p{Is_Sc=-Deva}', "");
Expect(1, 43260, '\p{^Is_Sc=-Deva}', "");
Expect(1, 43260, '\P{Is_Sc=-Deva}', "");
Expect(0, 43260, '\P{^Is_Sc=-Deva}', "");
Error('\p{Script=:= _SHARADA}');
Error('\P{Script=:= _SHARADA}');
Expect(1, 70105, '\p{Script=sharada}', "");
Expect(0, 70105, '\p{^Script=sharada}', "");
Expect(0, 70105, '\P{Script=sharada}', "");
Expect(1, 70105, '\P{^Script=sharada}', "");
Expect(0, 70106, '\p{Script=sharada}', "");
Expect(1, 70106, '\p{^Script=sharada}', "");
Expect(1, 70106, '\P{Script=sharada}', "");
Expect(0, 70106, '\P{^Script=sharada}', "");
Expect(1, 70105, '\p{Script=Sharada}', "");
Expect(0, 70105, '\p{^Script=Sharada}', "");
Expect(0, 70105, '\P{Script=Sharada}', "");
Expect(1, 70105, '\P{^Script=Sharada}', "");
Expect(0, 70106, '\p{Script=Sharada}', "");
Expect(1, 70106, '\p{^Script=Sharada}', "");
Expect(1, 70106, '\P{Script=Sharada}', "");
Expect(0, 70106, '\P{^Script=Sharada}', "");
Error('\p{Sc=	_Shrd:=}');
Error('\P{Sc=	_Shrd:=}');
Expect(1, 70105, '\p{Sc=shrd}', "");
Expect(0, 70105, '\p{^Sc=shrd}', "");
Expect(0, 70105, '\P{Sc=shrd}', "");
Expect(1, 70105, '\P{^Sc=shrd}', "");
Expect(0, 70106, '\p{Sc=shrd}', "");
Expect(1, 70106, '\p{^Sc=shrd}', "");
Expect(1, 70106, '\P{Sc=shrd}', "");
Expect(0, 70106, '\P{^Sc=shrd}', "");
Expect(1, 70105, '\p{Sc=  shrd}', "");
Expect(0, 70105, '\p{^Sc=  shrd}', "");
Expect(0, 70105, '\P{Sc=  shrd}', "");
Expect(1, 70105, '\P{^Sc=  shrd}', "");
Expect(0, 70106, '\p{Sc=  shrd}', "");
Expect(1, 70106, '\p{^Sc=  shrd}', "");
Expect(1, 70106, '\P{Sc=  shrd}', "");
Expect(0, 70106, '\P{^Sc=  shrd}', "");
Error('\p{Is_Script=	Sharada:=}');
Error('\P{Is_Script=	Sharada:=}');
Expect(1, 70105, '\p{Is_Script=sharada}', "");
Expect(0, 70105, '\p{^Is_Script=sharada}', "");
Expect(0, 70105, '\P{Is_Script=sharada}', "");
Expect(1, 70105, '\P{^Is_Script=sharada}', "");
Expect(0, 70106, '\p{Is_Script=sharada}', "");
Expect(1, 70106, '\p{^Is_Script=sharada}', "");
Expect(1, 70106, '\P{Is_Script=sharada}', "");
Expect(0, 70106, '\P{^Is_Script=sharada}', "");
Expect(1, 70105, '\p{Is_Script=_ SHARADA}', "");
Expect(0, 70105, '\p{^Is_Script=_ SHARADA}', "");
Expect(0, 70105, '\P{Is_Script=_ SHARADA}', "");
Expect(1, 70105, '\P{^Is_Script=_ SHARADA}', "");
Expect(0, 70106, '\p{Is_Script=_ SHARADA}', "");
Expect(1, 70106, '\p{^Is_Script=_ SHARADA}', "");
Expect(1, 70106, '\P{Is_Script=_ SHARADA}', "");
Expect(0, 70106, '\P{^Is_Script=_ SHARADA}', "");
Error('\p{Is_Sc=:=-	shrd}');
Error('\P{Is_Sc=:=-	shrd}');
Expect(1, 70105, '\p{Is_Sc=shrd}', "");
Expect(0, 70105, '\p{^Is_Sc=shrd}', "");
Expect(0, 70105, '\P{Is_Sc=shrd}', "");
Expect(1, 70105, '\P{^Is_Sc=shrd}', "");
Expect(0, 70106, '\p{Is_Sc=shrd}', "");
Expect(1, 70106, '\p{^Is_Sc=shrd}', "");
Expect(1, 70106, '\P{Is_Sc=shrd}', "");
Expect(0, 70106, '\P{^Is_Sc=shrd}', "");
Expect(1, 70105, '\p{Is_Sc=-Shrd}', "");
Expect(0, 70105, '\p{^Is_Sc=-Shrd}', "");
Expect(0, 70105, '\P{Is_Sc=-Shrd}', "");
Expect(1, 70105, '\P{^Is_Sc=-Shrd}', "");
Expect(0, 70106, '\p{Is_Sc=-Shrd}', "");
Expect(1, 70106, '\p{^Is_Sc=-Shrd}', "");
Expect(1, 70106, '\P{Is_Sc=-Shrd}', "");
Expect(0, 70106, '\P{^Is_Sc=-Shrd}', "");
Error('\p{Script=_ Carian:=}');
Error('\P{Script=_ Carian:=}');
Expect(1, 66256, '\p{Script=carian}', "");
Expect(0, 66256, '\p{^Script=carian}', "");
Expect(0, 66256, '\P{Script=carian}', "");
Expect(1, 66256, '\P{^Script=carian}', "");
Expect(0, 66257, '\p{Script=carian}', "");
Expect(1, 66257, '\p{^Script=carian}', "");
Expect(1, 66257, '\P{Script=carian}', "");
Expect(0, 66257, '\P{^Script=carian}', "");
Expect(1, 66256, '\p{Script=  Carian}', "");
Expect(0, 66256, '\p{^Script=  Carian}', "");
Expect(0, 66256, '\P{Script=  Carian}', "");
Expect(1, 66256, '\P{^Script=  Carian}', "");
Expect(0, 66257, '\p{Script=  Carian}', "");
Expect(1, 66257, '\p{^Script=  Carian}', "");
Expect(1, 66257, '\P{Script=  Carian}', "");
Expect(0, 66257, '\P{^Script=  Carian}', "");
Error('\p{Sc:_	Cari/a/}');
Error('\P{Sc:_	Cari/a/}');
Expect(1, 66256, '\p{Sc=cari}', "");
Expect(0, 66256, '\p{^Sc=cari}', "");
Expect(0, 66256, '\P{Sc=cari}', "");
Expect(1, 66256, '\P{^Sc=cari}', "");
Expect(0, 66257, '\p{Sc=cari}', "");
Expect(1, 66257, '\p{^Sc=cari}', "");
Expect(1, 66257, '\P{Sc=cari}', "");
Expect(0, 66257, '\P{^Sc=cari}', "");
Expect(1, 66256, '\p{Sc=		Cari}', "");
Expect(0, 66256, '\p{^Sc=		Cari}', "");
Expect(0, 66256, '\P{Sc=		Cari}', "");
Expect(1, 66256, '\P{^Sc=		Cari}', "");
Expect(0, 66257, '\p{Sc=		Cari}', "");
Expect(1, 66257, '\p{^Sc=		Cari}', "");
Expect(1, 66257, '\P{Sc=		Cari}', "");
Expect(0, 66257, '\P{^Sc=		Cari}', "");
Error('\p{Is_Script=/a/	_CARIAN}');
Error('\P{Is_Script=/a/	_CARIAN}');
Expect(1, 66256, '\p{Is_Script:	carian}', "");
Expect(0, 66256, '\p{^Is_Script:	carian}', "");
Expect(0, 66256, '\P{Is_Script:	carian}', "");
Expect(1, 66256, '\P{^Is_Script:	carian}', "");
Expect(0, 66257, '\p{Is_Script:	carian}', "");
Expect(1, 66257, '\p{^Is_Script:	carian}', "");
Expect(1, 66257, '\P{Is_Script:	carian}', "");
Expect(0, 66257, '\P{^Is_Script:	carian}', "");
Expect(1, 66256, '\p{Is_Script= 	carian}', "");
Expect(0, 66256, '\p{^Is_Script= 	carian}', "");
Expect(0, 66256, '\P{Is_Script= 	carian}', "");
Expect(1, 66256, '\P{^Is_Script= 	carian}', "");
Expect(0, 66257, '\p{Is_Script= 	carian}', "");
Expect(1, 66257, '\p{^Is_Script= 	carian}', "");
Expect(1, 66257, '\P{Is_Script= 	carian}', "");
Expect(0, 66257, '\P{^Is_Script= 	carian}', "");
Error('\p{Is_Sc= CARI/a/}');
Error('\P{Is_Sc= CARI/a/}');
Expect(1, 66256, '\p{Is_Sc=cari}', "");
Expect(0, 66256, '\p{^Is_Sc=cari}', "");
Expect(0, 66256, '\P{Is_Sc=cari}', "");
Expect(1, 66256, '\P{^Is_Sc=cari}', "");
Expect(0, 66257, '\p{Is_Sc=cari}', "");
Expect(1, 66257, '\p{^Is_Sc=cari}', "");
Expect(1, 66257, '\P{Is_Sc=cari}', "");
Expect(0, 66257, '\P{^Is_Sc=cari}', "");
Expect(1, 66256, '\p{Is_Sc=  Cari}', "");
Expect(0, 66256, '\p{^Is_Sc=  Cari}', "");
Expect(0, 66256, '\P{Is_Sc=  Cari}', "");
Expect(1, 66256, '\P{^Is_Sc=  Cari}', "");
Expect(0, 66257, '\p{Is_Sc=  Cari}', "");
Expect(1, 66257, '\p{^Is_Sc=  Cari}', "");
Expect(1, 66257, '\P{Is_Sc=  Cari}', "");
Expect(0, 66257, '\P{^Is_Sc=  Cari}', "");
Error('\p{Script=buhid/a/}');
Error('\P{Script=buhid/a/}');
Expect(1, 5971, '\p{Script=buhid}', "");
Expect(0, 5971, '\p{^Script=buhid}', "");
Expect(0, 5971, '\P{Script=buhid}', "");
Expect(1, 5971, '\P{^Script=buhid}', "");
Expect(0, 5972, '\p{Script=buhid}', "");
Expect(1, 5972, '\p{^Script=buhid}', "");
Expect(1, 5972, '\P{Script=buhid}', "");
Expect(0, 5972, '\P{^Script=buhid}', "");
Expect(1, 5971, '\p{Script= Buhid}', "");
Expect(0, 5971, '\p{^Script= Buhid}', "");
Expect(0, 5971, '\P{Script= Buhid}', "");
Expect(1, 5971, '\P{^Script= Buhid}', "");
Expect(0, 5972, '\p{Script= Buhid}', "");
Expect(1, 5972, '\p{^Script= Buhid}', "");
Expect(1, 5972, '\P{Script= Buhid}', "");
Expect(0, 5972, '\P{^Script= Buhid}', "");
Error('\p{Sc= 	BUHD/a/}');
Error('\P{Sc= 	BUHD/a/}');
Expect(1, 5971, '\p{Sc:buhd}', "");
Expect(0, 5971, '\p{^Sc:buhd}', "");
Expect(0, 5971, '\P{Sc:buhd}', "");
Expect(1, 5971, '\P{^Sc:buhd}', "");
Expect(0, 5972, '\p{Sc:buhd}', "");
Expect(1, 5972, '\p{^Sc:buhd}', "");
Expect(1, 5972, '\P{Sc:buhd}', "");
Expect(0, 5972, '\P{^Sc:buhd}', "");
Expect(1, 5971, '\p{Sc=-_Buhd}', "");
Expect(0, 5971, '\p{^Sc=-_Buhd}', "");
Expect(0, 5971, '\P{Sc=-_Buhd}', "");
Expect(1, 5971, '\P{^Sc=-_Buhd}', "");
Expect(0, 5972, '\p{Sc=-_Buhd}', "");
Expect(1, 5972, '\p{^Sc=-_Buhd}', "");
Expect(1, 5972, '\P{Sc=-_Buhd}', "");
Expect(0, 5972, '\P{^Sc=-_Buhd}', "");
Error('\p{Is_Script=	/a/Buhid}');
Error('\P{Is_Script=	/a/Buhid}');
Expect(1, 5971, '\p{Is_Script=buhid}', "");
Expect(0, 5971, '\p{^Is_Script=buhid}', "");
Expect(0, 5971, '\P{Is_Script=buhid}', "");
Expect(1, 5971, '\P{^Is_Script=buhid}', "");
Expect(0, 5972, '\p{Is_Script=buhid}', "");
Expect(1, 5972, '\p{^Is_Script=buhid}', "");
Expect(1, 5972, '\P{Is_Script=buhid}', "");
Expect(0, 5972, '\P{^Is_Script=buhid}', "");
Expect(1, 5971, '\p{Is_Script=-BUHID}', "");
Expect(0, 5971, '\p{^Is_Script=-BUHID}', "");
Expect(0, 5971, '\P{Is_Script=-BUHID}', "");
Expect(1, 5971, '\P{^Is_Script=-BUHID}', "");
Expect(0, 5972, '\p{Is_Script=-BUHID}', "");
Expect(1, 5972, '\p{^Is_Script=-BUHID}', "");
Expect(1, 5972, '\P{Is_Script=-BUHID}', "");
Expect(0, 5972, '\P{^Is_Script=-BUHID}', "");
Error('\p{Is_Sc=/a/_-Buhd}');
Error('\P{Is_Sc=/a/_-Buhd}');
Expect(1, 5971, '\p{Is_Sc=buhd}', "");
Expect(0, 5971, '\p{^Is_Sc=buhd}', "");
Expect(0, 5971, '\P{Is_Sc=buhd}', "");
Expect(1, 5971, '\P{^Is_Sc=buhd}', "");
Expect(0, 5972, '\p{Is_Sc=buhd}', "");
Expect(1, 5972, '\p{^Is_Sc=buhd}', "");
Expect(1, 5972, '\P{Is_Sc=buhd}', "");
Expect(0, 5972, '\P{^Is_Sc=buhd}', "");
Expect(1, 5971, '\p{Is_Sc:    -Buhd}', "");
Expect(0, 5971, '\p{^Is_Sc:    -Buhd}', "");
Expect(0, 5971, '\P{Is_Sc:    -Buhd}', "");
Expect(1, 5971, '\P{^Is_Sc:    -Buhd}', "");
Expect(0, 5972, '\p{Is_Sc:    -Buhd}', "");
Expect(1, 5972, '\p{^Is_Sc:    -Buhd}', "");
Expect(1, 5972, '\P{Is_Sc:    -Buhd}', "");
Expect(0, 5972, '\P{^Is_Sc:    -Buhd}', "");
Error('\p{Script=:=-_limbu}');
Error('\P{Script=:=-_limbu}');
Expect(1, 6479, '\p{Script=limbu}', "");
Expect(0, 6479, '\p{^Script=limbu}', "");
Expect(0, 6479, '\P{Script=limbu}', "");
Expect(1, 6479, '\P{^Script=limbu}', "");
Expect(0, 6480, '\p{Script=limbu}', "");
Expect(1, 6480, '\p{^Script=limbu}', "");
Expect(1, 6480, '\P{Script=limbu}', "");
Expect(0, 6480, '\P{^Script=limbu}', "");
Error('\p{Sc=:=_ LIMB}');
Error('\P{Sc=:=_ LIMB}');
Expect(1, 6479, '\p{Sc=limb}', "");
Expect(0, 6479, '\p{^Sc=limb}', "");
Expect(0, 6479, '\P{Sc=limb}', "");
Expect(1, 6479, '\P{^Sc=limb}', "");
Expect(0, 6480, '\p{Sc=limb}', "");
Expect(1, 6480, '\p{^Sc=limb}', "");
Expect(1, 6480, '\P{Sc=limb}', "");
Expect(0, 6480, '\P{^Sc=limb}', "");
Expect(1, 6479, '\p{Sc=_-limb}', "");
Expect(0, 6479, '\p{^Sc=_-limb}', "");
Expect(0, 6479, '\P{Sc=_-limb}', "");
Expect(1, 6479, '\P{^Sc=_-limb}', "");
Expect(0, 6480, '\p{Sc=_-limb}', "");
Expect(1, 6480, '\p{^Sc=_-limb}', "");
Expect(1, 6480, '\P{Sc=_-limb}', "");
Expect(0, 6480, '\P{^Sc=_-limb}', "");
Error('\p{Is_Script=/a/ Limbu}');
Error('\P{Is_Script=/a/ Limbu}');
Expect(1, 6479, '\p{Is_Script=limbu}', "");
Expect(0, 6479, '\p{^Is_Script=limbu}', "");
Expect(0, 6479, '\P{Is_Script=limbu}', "");
Expect(1, 6479, '\P{^Is_Script=limbu}', "");
Expect(0, 6480, '\p{Is_Script=limbu}', "");
Expect(1, 6480, '\p{^Is_Script=limbu}', "");
Expect(1, 6480, '\P{Is_Script=limbu}', "");
Expect(0, 6480, '\P{^Is_Script=limbu}', "");
Expect(1, 6479, '\p{Is_Script= _limbu}', "");
Expect(0, 6479, '\p{^Is_Script= _limbu}', "");
Expect(0, 6479, '\P{Is_Script= _limbu}', "");
Expect(1, 6479, '\P{^Is_Script= _limbu}', "");
Expect(0, 6480, '\p{Is_Script= _limbu}', "");
Expect(1, 6480, '\p{^Is_Script= _limbu}', "");
Expect(1, 6480, '\P{Is_Script= _limbu}', "");
Expect(0, 6480, '\P{^Is_Script= _limbu}', "");
Error('\p{Is_Sc=-limb/a/}');
Error('\P{Is_Sc=-limb/a/}');
Expect(1, 6479, '\p{Is_Sc=limb}', "");
Expect(0, 6479, '\p{^Is_Sc=limb}', "");
Expect(0, 6479, '\P{Is_Sc=limb}', "");
Expect(1, 6479, '\P{^Is_Sc=limb}', "");
Expect(0, 6480, '\p{Is_Sc=limb}', "");
Expect(1, 6480, '\p{^Is_Sc=limb}', "");
Expect(1, 6480, '\P{Is_Sc=limb}', "");
Expect(0, 6480, '\P{^Is_Sc=limb}', "");
Expect(1, 6479, '\p{Is_Sc= 	Limb}', "");
Expect(0, 6479, '\p{^Is_Sc= 	Limb}', "");
Expect(0, 6479, '\P{Is_Sc= 	Limb}', "");
Expect(1, 6479, '\P{^Is_Sc= 	Limb}', "");
Expect(0, 6480, '\p{Is_Sc= 	Limb}', "");
Expect(1, 6480, '\p{^Is_Sc= 	Limb}', "");
Expect(1, 6480, '\P{Is_Sc= 	Limb}', "");
Expect(0, 6480, '\P{^Is_Sc= 	Limb}', "");
Error('\p{Script:_:=javanese}');
Error('\P{Script:_:=javanese}');
Expect(1, 43487, '\p{Script:	javanese}', "");
Expect(0, 43487, '\p{^Script:	javanese}', "");
Expect(0, 43487, '\P{Script:	javanese}', "");
Expect(1, 43487, '\P{^Script:	javanese}', "");
Expect(0, 43488, '\p{Script:	javanese}', "");
Expect(1, 43488, '\p{^Script:	javanese}', "");
Expect(1, 43488, '\P{Script:	javanese}', "");
Expect(0, 43488, '\P{^Script:	javanese}', "");
Expect(1, 43487, '\p{Script=	-Javanese}', "");
Expect(0, 43487, '\p{^Script=	-Javanese}', "");
Expect(0, 43487, '\P{Script=	-Javanese}', "");
Expect(1, 43487, '\P{^Script=	-Javanese}', "");
Expect(0, 43488, '\p{Script=	-Javanese}', "");
Expect(1, 43488, '\p{^Script=	-Javanese}', "");
Expect(1, 43488, '\P{Script=	-Javanese}', "");
Expect(0, 43488, '\P{^Script=	-Javanese}', "");
Error('\p{Sc=	/a/Java}');
Error('\P{Sc=	/a/Java}');
Expect(1, 43487, '\p{Sc=java}', "");
Expect(0, 43487, '\p{^Sc=java}', "");
Expect(0, 43487, '\P{Sc=java}', "");
Expect(1, 43487, '\P{^Sc=java}', "");
Expect(0, 43488, '\p{Sc=java}', "");
Expect(1, 43488, '\p{^Sc=java}', "");
Expect(1, 43488, '\P{Sc=java}', "");
Expect(0, 43488, '\P{^Sc=java}', "");
Expect(1, 43487, '\p{Sc=-	JAVA}', "");
Expect(0, 43487, '\p{^Sc=-	JAVA}', "");
Expect(0, 43487, '\P{Sc=-	JAVA}', "");
Expect(1, 43487, '\P{^Sc=-	JAVA}', "");
Expect(0, 43488, '\p{Sc=-	JAVA}', "");
Expect(1, 43488, '\p{^Sc=-	JAVA}', "");
Expect(1, 43488, '\P{Sc=-	JAVA}', "");
Expect(0, 43488, '\P{^Sc=-	JAVA}', "");
Error('\p{Is_Script=	 JAVANESE:=}');
Error('\P{Is_Script=	 JAVANESE:=}');
Expect(1, 43487, '\p{Is_Script=javanese}', "");
Expect(0, 43487, '\p{^Is_Script=javanese}', "");
Expect(0, 43487, '\P{Is_Script=javanese}', "");
Expect(1, 43487, '\P{^Is_Script=javanese}', "");
Expect(0, 43488, '\p{Is_Script=javanese}', "");
Expect(1, 43488, '\p{^Is_Script=javanese}', "");
Expect(1, 43488, '\P{Is_Script=javanese}', "");
Expect(0, 43488, '\P{^Is_Script=javanese}', "");
Expect(1, 43487, '\p{Is_Script=_	JAVANESE}', "");
Expect(0, 43487, '\p{^Is_Script=_	JAVANESE}', "");
Expect(0, 43487, '\P{Is_Script=_	JAVANESE}', "");
Expect(1, 43487, '\P{^Is_Script=_	JAVANESE}', "");
Expect(0, 43488, '\p{Is_Script=_	JAVANESE}', "");
Expect(1, 43488, '\p{^Is_Script=_	JAVANESE}', "");
Expect(1, 43488, '\P{Is_Script=_	JAVANESE}', "");
Expect(0, 43488, '\P{^Is_Script=_	JAVANESE}', "");
Error('\p{Is_Sc=:=JAVA}');
Error('\P{Is_Sc=:=JAVA}');
Expect(1, 43487, '\p{Is_Sc=java}', "");
Expect(0, 43487, '\p{^Is_Sc=java}', "");
Expect(0, 43487, '\P{Is_Sc=java}', "");
Expect(1, 43487, '\P{^Is_Sc=java}', "");
Expect(0, 43488, '\p{Is_Sc=java}', "");
Expect(1, 43488, '\p{^Is_Sc=java}', "");
Expect(1, 43488, '\P{Is_Sc=java}', "");
Expect(0, 43488, '\P{^Is_Sc=java}', "");
Expect(1, 43487, '\p{Is_Sc= 	java}', "");
Expect(0, 43487, '\p{^Is_Sc= 	java}', "");
Expect(0, 43487, '\P{Is_Sc= 	java}', "");
Expect(1, 43487, '\P{^Is_Sc= 	java}', "");
Expect(0, 43488, '\p{Is_Sc= 	java}', "");
Expect(1, 43488, '\p{^Is_Sc= 	java}', "");
Expect(1, 43488, '\P{Is_Sc= 	java}', "");
Expect(0, 43488, '\P{^Is_Sc= 	java}', "");
Error('\p{Script=:=Egyptian_HIEROGLYPHS}');
Error('\P{Script=:=Egyptian_HIEROGLYPHS}');
Expect(1, 78894, '\p{Script=egyptianhieroglyphs}', "");
Expect(0, 78894, '\p{^Script=egyptianhieroglyphs}', "");
Expect(0, 78894, '\P{Script=egyptianhieroglyphs}', "");
Expect(1, 78894, '\P{^Script=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{Script=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{^Script=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{Script=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{^Script=egyptianhieroglyphs}', "");
Expect(1, 78894, '\p{Script= -egyptian_HIEROGLYPHS}', "");
Expect(0, 78894, '\p{^Script= -egyptian_HIEROGLYPHS}', "");
Expect(0, 78894, '\P{Script= -egyptian_HIEROGLYPHS}', "");
Expect(1, 78894, '\P{^Script= -egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\p{Script= -egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\p{^Script= -egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\P{Script= -egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\P{^Script= -egyptian_HIEROGLYPHS}', "");
Error('\p{Sc=/a/	_Egyp}');
Error('\P{Sc=/a/	_Egyp}');
Expect(1, 78894, '\p{Sc=egyp}', "");
Expect(0, 78894, '\p{^Sc=egyp}', "");
Expect(0, 78894, '\P{Sc=egyp}', "");
Expect(1, 78894, '\P{^Sc=egyp}', "");
Expect(0, 78895, '\p{Sc=egyp}', "");
Expect(1, 78895, '\p{^Sc=egyp}', "");
Expect(1, 78895, '\P{Sc=egyp}', "");
Expect(0, 78895, '\P{^Sc=egyp}', "");
Expect(1, 78894, '\p{Sc= Egyp}', "");
Expect(0, 78894, '\p{^Sc= Egyp}', "");
Expect(0, 78894, '\P{Sc= Egyp}', "");
Expect(1, 78894, '\P{^Sc= Egyp}', "");
Expect(0, 78895, '\p{Sc= Egyp}', "");
Expect(1, 78895, '\p{^Sc= Egyp}', "");
Expect(1, 78895, '\P{Sc= Egyp}', "");
Expect(0, 78895, '\P{^Sc= Egyp}', "");
Error('\p{Is_Script=-:=egyptian_Hieroglyphs}');
Error('\P{Is_Script=-:=egyptian_Hieroglyphs}');
Expect(1, 78894, '\p{Is_Script=egyptianhieroglyphs}', "");
Expect(0, 78894, '\p{^Is_Script=egyptianhieroglyphs}', "");
Expect(0, 78894, '\P{Is_Script=egyptianhieroglyphs}', "");
Expect(1, 78894, '\P{^Is_Script=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{Is_Script=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{^Is_Script=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{Is_Script=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{^Is_Script=egyptianhieroglyphs}', "");
Expect(1, 78894, '\p{Is_Script= 	Egyptian_HIEROGLYPHS}', "");
Expect(0, 78894, '\p{^Is_Script= 	Egyptian_HIEROGLYPHS}', "");
Expect(0, 78894, '\P{Is_Script= 	Egyptian_HIEROGLYPHS}', "");
Expect(1, 78894, '\P{^Is_Script= 	Egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\p{Is_Script= 	Egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\p{^Is_Script= 	Egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\P{Is_Script= 	Egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\P{^Is_Script= 	Egyptian_HIEROGLYPHS}', "");
Error('\p{Is_Sc=__Egyp:=}');
Error('\P{Is_Sc=__Egyp:=}');
Expect(1, 78894, '\p{Is_Sc:	egyp}', "");
Expect(0, 78894, '\p{^Is_Sc:	egyp}', "");
Expect(0, 78894, '\P{Is_Sc:	egyp}', "");
Expect(1, 78894, '\P{^Is_Sc:	egyp}', "");
Expect(0, 78895, '\p{Is_Sc:	egyp}', "");
Expect(1, 78895, '\p{^Is_Sc:	egyp}', "");
Expect(1, 78895, '\P{Is_Sc:	egyp}', "");
Expect(0, 78895, '\P{^Is_Sc:	egyp}', "");
Expect(1, 78894, '\p{Is_Sc=_Egyp}', "");
Expect(0, 78894, '\p{^Is_Sc=_Egyp}', "");
Expect(0, 78894, '\P{Is_Sc=_Egyp}', "");
Expect(1, 78894, '\P{^Is_Sc=_Egyp}', "");
Expect(0, 78895, '\p{Is_Sc=_Egyp}', "");
Expect(1, 78895, '\p{^Is_Sc=_Egyp}', "");
Expect(1, 78895, '\P{Is_Sc=_Egyp}', "");
Expect(0, 78895, '\P{^Is_Sc=_Egyp}', "");
Error('\p{Script= :=common}');
Error('\P{Script= :=common}');
Expect(1, 917631, '\p{Script=common}', "");
Expect(0, 917631, '\p{^Script=common}', "");
Expect(0, 917631, '\P{Script=common}', "");
Expect(1, 917631, '\P{^Script=common}', "");
Expect(0, 917632, '\p{Script=common}', "");
Expect(1, 917632, '\p{^Script=common}', "");
Expect(1, 917632, '\P{Script=common}', "");
Expect(0, 917632, '\P{^Script=common}', "");
Expect(1, 917631, '\p{Script= Common}', "");
Expect(0, 917631, '\p{^Script= Common}', "");
Expect(0, 917631, '\P{Script= Common}', "");
Expect(1, 917631, '\P{^Script= Common}', "");
Expect(0, 917632, '\p{Script= Common}', "");
Expect(1, 917632, '\p{^Script= Common}', "");
Expect(1, 917632, '\P{Script= Common}', "");
Expect(0, 917632, '\P{^Script= Common}', "");
Error('\p{Sc:   :=-Zyyy}');
Error('\P{Sc:   :=-Zyyy}');
Expect(1, 917631, '\p{Sc=zyyy}', "");
Expect(0, 917631, '\p{^Sc=zyyy}', "");
Expect(0, 917631, '\P{Sc=zyyy}', "");
Expect(1, 917631, '\P{^Sc=zyyy}', "");
Expect(0, 917632, '\p{Sc=zyyy}', "");
Expect(1, 917632, '\p{^Sc=zyyy}', "");
Expect(1, 917632, '\P{Sc=zyyy}', "");
Expect(0, 917632, '\P{^Sc=zyyy}', "");
Expect(1, 917631, '\p{Sc= 	Zyyy}', "");
Expect(0, 917631, '\p{^Sc= 	Zyyy}', "");
Expect(0, 917631, '\P{Sc= 	Zyyy}', "");
Expect(1, 917631, '\P{^Sc= 	Zyyy}', "");
Expect(0, 917632, '\p{Sc= 	Zyyy}', "");
Expect(1, 917632, '\p{^Sc= 	Zyyy}', "");
Expect(1, 917632, '\P{Sc= 	Zyyy}', "");
Expect(0, 917632, '\P{^Sc= 	Zyyy}', "");
Error('\p{Is_Script:/a/-	Common}');
Error('\P{Is_Script:/a/-	Common}');
Expect(1, 917631, '\p{Is_Script:common}', "");
Expect(0, 917631, '\p{^Is_Script:common}', "");
Expect(0, 917631, '\P{Is_Script:common}', "");
Expect(1, 917631, '\P{^Is_Script:common}', "");
Expect(0, 917632, '\p{Is_Script:common}', "");
Expect(1, 917632, '\p{^Is_Script:common}', "");
Expect(1, 917632, '\P{Is_Script:common}', "");
Expect(0, 917632, '\P{^Is_Script:common}', "");
Expect(1, 917631, '\p{Is_Script=-common}', "");
Expect(0, 917631, '\p{^Is_Script=-common}', "");
Expect(0, 917631, '\P{Is_Script=-common}', "");
Expect(1, 917631, '\P{^Is_Script=-common}', "");
Expect(0, 917632, '\p{Is_Script=-common}', "");
Expect(1, 917632, '\p{^Is_Script=-common}', "");
Expect(1, 917632, '\P{Is_Script=-common}', "");
Expect(0, 917632, '\P{^Is_Script=-common}', "");
Error('\p{Is_Sc=	_ZYYY:=}');
Error('\P{Is_Sc=	_ZYYY:=}');
Expect(1, 917631, '\p{Is_Sc=zyyy}', "");
Expect(0, 917631, '\p{^Is_Sc=zyyy}', "");
Expect(0, 917631, '\P{Is_Sc=zyyy}', "");
Expect(1, 917631, '\P{^Is_Sc=zyyy}', "");
Expect(0, 917632, '\p{Is_Sc=zyyy}', "");
Expect(1, 917632, '\p{^Is_Sc=zyyy}', "");
Expect(1, 917632, '\P{Is_Sc=zyyy}', "");
Expect(0, 917632, '\P{^Is_Sc=zyyy}', "");
Expect(1, 917631, '\p{Is_Sc=	_ZYYY}', "");
Expect(0, 917631, '\p{^Is_Sc=	_ZYYY}', "");
Expect(0, 917631, '\P{Is_Sc=	_ZYYY}', "");
Expect(1, 917631, '\P{^Is_Sc=	_ZYYY}', "");
Expect(0, 917632, '\p{Is_Sc=	_ZYYY}', "");
Expect(1, 917632, '\p{^Is_Sc=	_ZYYY}', "");
Expect(1, 917632, '\P{Is_Sc=	_ZYYY}', "");
Expect(0, 917632, '\P{^Is_Sc=	_ZYYY}', "");
Error('\p{Script=-/a/CHAKMA}');
Error('\P{Script=-/a/CHAKMA}');
Expect(1, 69955, '\p{Script=chakma}', "");
Expect(0, 69955, '\p{^Script=chakma}', "");
Expect(0, 69955, '\P{Script=chakma}', "");
Expect(1, 69955, '\P{^Script=chakma}', "");
Expect(0, 69956, '\p{Script=chakma}', "");
Expect(1, 69956, '\p{^Script=chakma}', "");
Expect(1, 69956, '\P{Script=chakma}', "");
Expect(0, 69956, '\P{^Script=chakma}', "");
Expect(1, 69955, '\p{Script:	-CHAKMA}', "");
Expect(0, 69955, '\p{^Script:	-CHAKMA}', "");
Expect(0, 69955, '\P{Script:	-CHAKMA}', "");
Expect(1, 69955, '\P{^Script:	-CHAKMA}', "");
Expect(0, 69956, '\p{Script:	-CHAKMA}', "");
Expect(1, 69956, '\p{^Script:	-CHAKMA}', "");
Expect(1, 69956, '\P{Script:	-CHAKMA}', "");
Expect(0, 69956, '\P{^Script:	-CHAKMA}', "");
Error('\p{Sc=_Cakm/a/}');
Error('\P{Sc=_Cakm/a/}');
Expect(1, 69955, '\p{Sc=cakm}', "");
Expect(0, 69955, '\p{^Sc=cakm}', "");
Expect(0, 69955, '\P{Sc=cakm}', "");
Expect(1, 69955, '\P{^Sc=cakm}', "");
Expect(0, 69956, '\p{Sc=cakm}', "");
Expect(1, 69956, '\p{^Sc=cakm}', "");
Expect(1, 69956, '\P{Sc=cakm}', "");
Expect(0, 69956, '\P{^Sc=cakm}', "");
Expect(1, 69955, '\p{Sc= _Cakm}', "");
Expect(0, 69955, '\p{^Sc= _Cakm}', "");
Expect(0, 69955, '\P{Sc= _Cakm}', "");
Expect(1, 69955, '\P{^Sc= _Cakm}', "");
Expect(0, 69956, '\p{Sc= _Cakm}', "");
Expect(1, 69956, '\p{^Sc= _Cakm}', "");
Expect(1, 69956, '\P{Sc= _Cakm}', "");
Expect(0, 69956, '\P{^Sc= _Cakm}', "");
Error('\p{Is_Script=/a/ _CHAKMA}');
Error('\P{Is_Script=/a/ _CHAKMA}');
Expect(1, 69955, '\p{Is_Script=chakma}', "");
Expect(0, 69955, '\p{^Is_Script=chakma}', "");
Expect(0, 69955, '\P{Is_Script=chakma}', "");
Expect(1, 69955, '\P{^Is_Script=chakma}', "");
Expect(0, 69956, '\p{Is_Script=chakma}', "");
Expect(1, 69956, '\p{^Is_Script=chakma}', "");
Expect(1, 69956, '\P{Is_Script=chakma}', "");
Expect(0, 69956, '\P{^Is_Script=chakma}', "");
Expect(1, 69955, '\p{Is_Script=__Chakma}', "");
Expect(0, 69955, '\p{^Is_Script=__Chakma}', "");
Expect(0, 69955, '\P{Is_Script=__Chakma}', "");
Expect(1, 69955, '\P{^Is_Script=__Chakma}', "");
Expect(0, 69956, '\p{Is_Script=__Chakma}', "");
Expect(1, 69956, '\p{^Is_Script=__Chakma}', "");
Expect(1, 69956, '\P{Is_Script=__Chakma}', "");
Expect(0, 69956, '\P{^Is_Script=__Chakma}', "");
Error('\p{Is_Sc=--Cakm/a/}');
Error('\P{Is_Sc=--Cakm/a/}');
Expect(1, 69955, '\p{Is_Sc=cakm}', "");
Expect(0, 69955, '\p{^Is_Sc=cakm}', "");
Expect(0, 69955, '\P{Is_Sc=cakm}', "");
Expect(1, 69955, '\P{^Is_Sc=cakm}', "");
Expect(0, 69956, '\p{Is_Sc=cakm}', "");
Expect(1, 69956, '\p{^Is_Sc=cakm}', "");
Expect(1, 69956, '\P{Is_Sc=cakm}', "");
Expect(0, 69956, '\P{^Is_Sc=cakm}', "");
Expect(1, 69955, '\p{Is_Sc=--Cakm}', "");
Expect(0, 69955, '\p{^Is_Sc=--Cakm}', "");
Expect(0, 69955, '\P{Is_Sc=--Cakm}', "");
Expect(1, 69955, '\P{^Is_Sc=--Cakm}', "");
Expect(0, 69956, '\p{Is_Sc=--Cakm}', "");
Expect(1, 69956, '\p{^Is_Sc=--Cakm}', "");
Expect(1, 69956, '\P{Is_Sc=--Cakm}', "");
Expect(0, 69956, '\P{^Is_Sc=--Cakm}', "");
Error('\p{Script=:=	 latin}');
Error('\P{Script=:=	 latin}');
Expect(1, 65370, '\p{Script=latin}', "");
Expect(0, 65370, '\p{^Script=latin}', "");
Expect(0, 65370, '\P{Script=latin}', "");
Expect(1, 65370, '\P{^Script=latin}', "");
Expect(0, 65371, '\p{Script=latin}', "");
Expect(1, 65371, '\p{^Script=latin}', "");
Expect(1, 65371, '\P{Script=latin}', "");
Expect(0, 65371, '\P{^Script=latin}', "");
Expect(1, 65370, '\p{Script=		Latin}', "");
Expect(0, 65370, '\p{^Script=		Latin}', "");
Expect(0, 65370, '\P{Script=		Latin}', "");
Expect(1, 65370, '\P{^Script=		Latin}', "");
Expect(0, 65371, '\p{Script=		Latin}', "");
Expect(1, 65371, '\p{^Script=		Latin}', "");
Expect(1, 65371, '\P{Script=		Latin}', "");
Expect(0, 65371, '\P{^Script=		Latin}', "");
Error('\p{Sc:	_/a/Latn}');
Error('\P{Sc:	_/a/Latn}');
Expect(1, 65370, '\p{Sc: latn}', "");
Expect(0, 65370, '\p{^Sc: latn}', "");
Expect(0, 65370, '\P{Sc: latn}', "");
Expect(1, 65370, '\P{^Sc: latn}', "");
Expect(0, 65371, '\p{Sc: latn}', "");
Expect(1, 65371, '\p{^Sc: latn}', "");
Expect(1, 65371, '\P{Sc: latn}', "");
Expect(0, 65371, '\P{^Sc: latn}', "");
Expect(1, 65370, '\p{Sc=_-LATN}', "");
Expect(0, 65370, '\p{^Sc=_-LATN}', "");
Expect(0, 65370, '\P{Sc=_-LATN}', "");
Expect(1, 65370, '\P{^Sc=_-LATN}', "");
Expect(0, 65371, '\p{Sc=_-LATN}', "");
Expect(1, 65371, '\p{^Sc=_-LATN}', "");
Expect(1, 65371, '\P{Sc=_-LATN}', "");
Expect(0, 65371, '\P{^Sc=_-LATN}', "");
Error('\p{Is_Script=:=- latin}');
Error('\P{Is_Script=:=- latin}');
Expect(1, 65370, '\p{Is_Script=latin}', "");
Expect(0, 65370, '\p{^Is_Script=latin}', "");
Expect(0, 65370, '\P{Is_Script=latin}', "");
Expect(1, 65370, '\P{^Is_Script=latin}', "");
Expect(0, 65371, '\p{Is_Script=latin}', "");
Expect(1, 65371, '\p{^Is_Script=latin}', "");
Expect(1, 65371, '\P{Is_Script=latin}', "");
Expect(0, 65371, '\P{^Is_Script=latin}', "");
Expect(1, 65370, '\p{Is_Script=	 Latin}', "");
Expect(0, 65370, '\p{^Is_Script=	 Latin}', "");
Expect(0, 65370, '\P{Is_Script=	 Latin}', "");
Expect(1, 65370, '\P{^Is_Script=	 Latin}', "");
Expect(0, 65371, '\p{Is_Script=	 Latin}', "");
Expect(1, 65371, '\p{^Is_Script=	 Latin}', "");
Expect(1, 65371, '\P{Is_Script=	 Latin}', "");
Expect(0, 65371, '\P{^Is_Script=	 Latin}', "");
Error('\p{Is_Sc= _LATN:=}');
Error('\P{Is_Sc= _LATN:=}');
Expect(1, 65370, '\p{Is_Sc=latn}', "");
Expect(0, 65370, '\p{^Is_Sc=latn}', "");
Expect(0, 65370, '\P{Is_Sc=latn}', "");
Expect(1, 65370, '\P{^Is_Sc=latn}', "");
Expect(0, 65371, '\p{Is_Sc=latn}', "");
Expect(1, 65371, '\p{^Is_Sc=latn}', "");
Expect(1, 65371, '\P{Is_Sc=latn}', "");
Expect(0, 65371, '\P{^Is_Sc=latn}', "");
Expect(1, 65370, '\p{Is_Sc=_	latn}', "");
Expect(0, 65370, '\p{^Is_Sc=_	latn}', "");
Expect(0, 65370, '\P{Is_Sc=_	latn}', "");
Expect(1, 65370, '\P{^Is_Sc=_	latn}', "");
Expect(0, 65371, '\p{Is_Sc=_	latn}', "");
Expect(1, 65371, '\p{^Is_Sc=_	latn}', "");
Expect(1, 65371, '\P{Is_Sc=_	latn}', "");
Expect(0, 65371, '\P{^Is_Sc=_	latn}', "");
Error('\p{Script=_-ethiopic/a/}');
Error('\P{Script=_-ethiopic/a/}');
Expect(1, 43822, '\p{Script=ethiopic}', "");
Expect(0, 43822, '\p{^Script=ethiopic}', "");
Expect(0, 43822, '\P{Script=ethiopic}', "");
Expect(1, 43822, '\P{^Script=ethiopic}', "");
Expect(0, 43823, '\p{Script=ethiopic}', "");
Expect(1, 43823, '\p{^Script=ethiopic}', "");
Expect(1, 43823, '\P{Script=ethiopic}', "");
Expect(0, 43823, '\P{^Script=ethiopic}', "");
Expect(1, 43822, '\p{Script= _ETHIOPIC}', "");
Expect(0, 43822, '\p{^Script= _ETHIOPIC}', "");
Expect(0, 43822, '\P{Script= _ETHIOPIC}', "");
Expect(1, 43822, '\P{^Script= _ETHIOPIC}', "");
Expect(0, 43823, '\p{Script= _ETHIOPIC}', "");
Expect(1, 43823, '\p{^Script= _ETHIOPIC}', "");
Expect(1, 43823, '\P{Script= _ETHIOPIC}', "");
Expect(0, 43823, '\P{^Script= _ETHIOPIC}', "");
Error('\p{Sc= -Ethi/a/}');
Error('\P{Sc= -Ethi/a/}');
Expect(1, 43822, '\p{Sc=ethi}', "");
Expect(0, 43822, '\p{^Sc=ethi}', "");
Expect(0, 43822, '\P{Sc=ethi}', "");
Expect(1, 43822, '\P{^Sc=ethi}', "");
Expect(0, 43823, '\p{Sc=ethi}', "");
Expect(1, 43823, '\p{^Sc=ethi}', "");
Expect(1, 43823, '\P{Sc=ethi}', "");
Expect(0, 43823, '\P{^Sc=ethi}', "");
Expect(1, 43822, '\p{Sc= ETHI}', "");
Expect(0, 43822, '\p{^Sc= ETHI}', "");
Expect(0, 43822, '\P{Sc= ETHI}', "");
Expect(1, 43822, '\P{^Sc= ETHI}', "");
Expect(0, 43823, '\p{Sc= ETHI}', "");
Expect(1, 43823, '\p{^Sc= ETHI}', "");
Expect(1, 43823, '\P{Sc= ETHI}', "");
Expect(0, 43823, '\P{^Sc= ETHI}', "");
Error('\p{Is_Script=:=--Ethiopic}');
Error('\P{Is_Script=:=--Ethiopic}');
Expect(1, 43822, '\p{Is_Script:ethiopic}', "");
Expect(0, 43822, '\p{^Is_Script:ethiopic}', "");
Expect(0, 43822, '\P{Is_Script:ethiopic}', "");
Expect(1, 43822, '\P{^Is_Script:ethiopic}', "");
Expect(0, 43823, '\p{Is_Script:ethiopic}', "");
Expect(1, 43823, '\p{^Is_Script:ethiopic}', "");
Expect(1, 43823, '\P{Is_Script:ethiopic}', "");
Expect(0, 43823, '\P{^Is_Script:ethiopic}', "");
Expect(1, 43822, '\p{Is_Script=	ETHIOPIC}', "");
Expect(0, 43822, '\p{^Is_Script=	ETHIOPIC}', "");
Expect(0, 43822, '\P{Is_Script=	ETHIOPIC}', "");
Expect(1, 43822, '\P{^Is_Script=	ETHIOPIC}', "");
Expect(0, 43823, '\p{Is_Script=	ETHIOPIC}', "");
Expect(1, 43823, '\p{^Is_Script=	ETHIOPIC}', "");
Expect(1, 43823, '\P{Is_Script=	ETHIOPIC}', "");
Expect(0, 43823, '\P{^Is_Script=	ETHIOPIC}', "");
Error('\p{Is_Sc=	Ethi:=}');
Error('\P{Is_Sc=	Ethi:=}');
Expect(1, 43822, '\p{Is_Sc=ethi}', "");
Expect(0, 43822, '\p{^Is_Sc=ethi}', "");
Expect(0, 43822, '\P{Is_Sc=ethi}', "");
Expect(1, 43822, '\P{^Is_Sc=ethi}', "");
Expect(0, 43823, '\p{Is_Sc=ethi}', "");
Expect(1, 43823, '\p{^Is_Sc=ethi}', "");
Expect(1, 43823, '\P{Is_Sc=ethi}', "");
Expect(0, 43823, '\P{^Is_Sc=ethi}', "");
Expect(1, 43822, '\p{Is_Sc=_-ethi}', "");
Expect(0, 43822, '\p{^Is_Sc=_-ethi}', "");
Expect(0, 43822, '\P{Is_Sc=_-ethi}', "");
Expect(1, 43822, '\P{^Is_Sc=_-ethi}', "");
Expect(0, 43823, '\p{Is_Sc=_-ethi}', "");
Expect(1, 43823, '\p{^Is_Sc=_-ethi}', "");
Expect(1, 43823, '\P{Is_Sc=_-ethi}', "");
Expect(0, 43823, '\P{^Is_Sc=_-ethi}', "");
Error('\p{Script=/a/	-SORA_sompeng}');
Error('\P{Script=/a/	-SORA_sompeng}');
Expect(1, 69881, '\p{Script: sorasompeng}', "");
Expect(0, 69881, '\p{^Script: sorasompeng}', "");
Expect(0, 69881, '\P{Script: sorasompeng}', "");
Expect(1, 69881, '\P{^Script: sorasompeng}', "");
Expect(0, 69882, '\p{Script: sorasompeng}', "");
Expect(1, 69882, '\p{^Script: sorasompeng}', "");
Expect(1, 69882, '\P{Script: sorasompeng}', "");
Expect(0, 69882, '\P{^Script: sorasompeng}', "");
Expect(1, 69881, '\p{Script= SORA_Sompeng}', "");
Expect(0, 69881, '\p{^Script= SORA_Sompeng}', "");
Expect(0, 69881, '\P{Script= SORA_Sompeng}', "");
Expect(1, 69881, '\P{^Script= SORA_Sompeng}', "");
Expect(0, 69882, '\p{Script= SORA_Sompeng}', "");
Expect(1, 69882, '\p{^Script= SORA_Sompeng}', "");
Expect(1, 69882, '\P{Script= SORA_Sompeng}', "");
Expect(0, 69882, '\P{^Script= SORA_Sompeng}', "");
Error('\p{Sc= Sora/a/}');
Error('\P{Sc= Sora/a/}');
Expect(1, 69881, '\p{Sc=sora}', "");
Expect(0, 69881, '\p{^Sc=sora}', "");
Expect(0, 69881, '\P{Sc=sora}', "");
Expect(1, 69881, '\P{^Sc=sora}', "");
Expect(0, 69882, '\p{Sc=sora}', "");
Expect(1, 69882, '\p{^Sc=sora}', "");
Expect(1, 69882, '\P{Sc=sora}', "");
Expect(0, 69882, '\P{^Sc=sora}', "");
Expect(1, 69881, '\p{Sc=_-Sora}', "");
Expect(0, 69881, '\p{^Sc=_-Sora}', "");
Expect(0, 69881, '\P{Sc=_-Sora}', "");
Expect(1, 69881, '\P{^Sc=_-Sora}', "");
Expect(0, 69882, '\p{Sc=_-Sora}', "");
Expect(1, 69882, '\p{^Sc=_-Sora}', "");
Expect(1, 69882, '\P{Sc=_-Sora}', "");
Expect(0, 69882, '\P{^Sc=_-Sora}', "");
Error('\p{Is_Script=	_sora_Sompeng/a/}');
Error('\P{Is_Script=	_sora_Sompeng/a/}');
Expect(1, 69881, '\p{Is_Script=sorasompeng}', "");
Expect(0, 69881, '\p{^Is_Script=sorasompeng}', "");
Expect(0, 69881, '\P{Is_Script=sorasompeng}', "");
Expect(1, 69881, '\P{^Is_Script=sorasompeng}', "");
Expect(0, 69882, '\p{Is_Script=sorasompeng}', "");
Expect(1, 69882, '\p{^Is_Script=sorasompeng}', "");
Expect(1, 69882, '\P{Is_Script=sorasompeng}', "");
Expect(0, 69882, '\P{^Is_Script=sorasompeng}', "");
Expect(1, 69881, '\p{Is_Script=	-Sora_Sompeng}', "");
Expect(0, 69881, '\p{^Is_Script=	-Sora_Sompeng}', "");
Expect(0, 69881, '\P{Is_Script=	-Sora_Sompeng}', "");
Expect(1, 69881, '\P{^Is_Script=	-Sora_Sompeng}', "");
Expect(0, 69882, '\p{Is_Script=	-Sora_Sompeng}', "");
Expect(1, 69882, '\p{^Is_Script=	-Sora_Sompeng}', "");
Expect(1, 69882, '\P{Is_Script=	-Sora_Sompeng}', "");
Expect(0, 69882, '\P{^Is_Script=	-Sora_Sompeng}', "");
Error('\p{Is_Sc=:=_SORA}');
Error('\P{Is_Sc=:=_SORA}');
Expect(1, 69881, '\p{Is_Sc=sora}', "");
Expect(0, 69881, '\p{^Is_Sc=sora}', "");
Expect(0, 69881, '\P{Is_Sc=sora}', "");
Expect(1, 69881, '\P{^Is_Sc=sora}', "");
Expect(0, 69882, '\p{Is_Sc=sora}', "");
Expect(1, 69882, '\p{^Is_Sc=sora}', "");
Expect(1, 69882, '\P{Is_Sc=sora}', "");
Expect(0, 69882, '\P{^Is_Sc=sora}', "");
Expect(1, 69881, '\p{Is_Sc=	sora}', "");
Expect(0, 69881, '\p{^Is_Sc=	sora}', "");
Expect(0, 69881, '\P{Is_Sc=	sora}', "");
Expect(1, 69881, '\P{^Is_Sc=	sora}', "");
Expect(0, 69882, '\p{Is_Sc=	sora}', "");
Expect(1, 69882, '\p{^Is_Sc=	sora}', "");
Expect(1, 69882, '\P{Is_Sc=	sora}', "");
Expect(0, 69882, '\P{^Is_Sc=	sora}', "");
Error('\p{Script=-Hebrew:=}');
Error('\P{Script=-Hebrew:=}');
Expect(1, 64335, '\p{Script=hebrew}', "");
Expect(0, 64335, '\p{^Script=hebrew}', "");
Expect(0, 64335, '\P{Script=hebrew}', "");
Expect(1, 64335, '\P{^Script=hebrew}', "");
Expect(0, 64336, '\p{Script=hebrew}', "");
Expect(1, 64336, '\p{^Script=hebrew}', "");
Expect(1, 64336, '\P{Script=hebrew}', "");
Expect(0, 64336, '\P{^Script=hebrew}', "");
Expect(1, 64335, '\p{Script=	-Hebrew}', "");
Expect(0, 64335, '\p{^Script=	-Hebrew}', "");
Expect(0, 64335, '\P{Script=	-Hebrew}', "");
Expect(1, 64335, '\P{^Script=	-Hebrew}', "");
Expect(0, 64336, '\p{Script=	-Hebrew}', "");
Expect(1, 64336, '\p{^Script=	-Hebrew}', "");
Expect(1, 64336, '\P{Script=	-Hebrew}', "");
Expect(0, 64336, '\P{^Script=	-Hebrew}', "");
Error('\p{Sc=/a/- hebr}');
Error('\P{Sc=/a/- hebr}');
Expect(1, 64335, '\p{Sc=hebr}', "");
Expect(0, 64335, '\p{^Sc=hebr}', "");
Expect(0, 64335, '\P{Sc=hebr}', "");
Expect(1, 64335, '\P{^Sc=hebr}', "");
Expect(0, 64336, '\p{Sc=hebr}', "");
Expect(1, 64336, '\p{^Sc=hebr}', "");
Expect(1, 64336, '\P{Sc=hebr}', "");
Expect(0, 64336, '\P{^Sc=hebr}', "");
Error('\p{Is_Script:	/a/Hebrew}');
Error('\P{Is_Script:	/a/Hebrew}');
Expect(1, 64335, '\p{Is_Script=hebrew}', "");
Expect(0, 64335, '\p{^Is_Script=hebrew}', "");
Expect(0, 64335, '\P{Is_Script=hebrew}', "");
Expect(1, 64335, '\P{^Is_Script=hebrew}', "");
Expect(0, 64336, '\p{Is_Script=hebrew}', "");
Expect(1, 64336, '\p{^Is_Script=hebrew}', "");
Expect(1, 64336, '\P{Is_Script=hebrew}', "");
Expect(0, 64336, '\P{^Is_Script=hebrew}', "");
Expect(1, 64335, '\p{Is_Script:__hebrew}', "");
Expect(0, 64335, '\p{^Is_Script:__hebrew}', "");
Expect(0, 64335, '\P{Is_Script:__hebrew}', "");
Expect(1, 64335, '\P{^Is_Script:__hebrew}', "");
Expect(0, 64336, '\p{Is_Script:__hebrew}', "");
Expect(1, 64336, '\p{^Is_Script:__hebrew}', "");
Expect(1, 64336, '\P{Is_Script:__hebrew}', "");
Expect(0, 64336, '\P{^Is_Script:__hebrew}', "");
Error('\p{Is_Sc= 	hebr/a/}');
Error('\P{Is_Sc= 	hebr/a/}');
Expect(1, 64335, '\p{Is_Sc=hebr}', "");
Expect(0, 64335, '\p{^Is_Sc=hebr}', "");
Expect(0, 64335, '\P{Is_Sc=hebr}', "");
Expect(1, 64335, '\P{^Is_Sc=hebr}', "");
Expect(0, 64336, '\p{Is_Sc=hebr}', "");
Expect(1, 64336, '\p{^Is_Sc=hebr}', "");
Expect(1, 64336, '\P{Is_Sc=hebr}', "");
Expect(0, 64336, '\P{^Is_Sc=hebr}', "");
Expect(1, 64335, '\p{Is_Sc=	 Hebr}', "");
Expect(0, 64335, '\p{^Is_Sc=	 Hebr}', "");
Expect(0, 64335, '\P{Is_Sc=	 Hebr}', "");
Expect(1, 64335, '\P{^Is_Sc=	 Hebr}', "");
Expect(0, 64336, '\p{Is_Sc=	 Hebr}', "");
Expect(1, 64336, '\p{^Is_Sc=	 Hebr}', "");
Expect(1, 64336, '\P{Is_Sc=	 Hebr}', "");
Expect(0, 64336, '\P{^Is_Sc=	 Hebr}', "");
Error('\p{Script=:=	Buginese}');
Error('\P{Script=:=	Buginese}');
Expect(1, 6687, '\p{Script=buginese}', "");
Expect(0, 6687, '\p{^Script=buginese}', "");
Expect(0, 6687, '\P{Script=buginese}', "");
Expect(1, 6687, '\P{^Script=buginese}', "");
Expect(0, 6688, '\p{Script=buginese}', "");
Expect(1, 6688, '\p{^Script=buginese}', "");
Expect(1, 6688, '\P{Script=buginese}', "");
Expect(0, 6688, '\P{^Script=buginese}', "");
Expect(1, 6687, '\p{Script=_-buginese}', "");
Expect(0, 6687, '\p{^Script=_-buginese}', "");
Expect(0, 6687, '\P{Script=_-buginese}', "");
Expect(1, 6687, '\P{^Script=_-buginese}', "");
Expect(0, 6688, '\p{Script=_-buginese}', "");
Expect(1, 6688, '\p{^Script=_-buginese}', "");
Expect(1, 6688, '\P{Script=_-buginese}', "");
Expect(0, 6688, '\P{^Script=_-buginese}', "");
Error('\p{Sc=_/a/bugi}');
Error('\P{Sc=_/a/bugi}');
Expect(1, 6687, '\p{Sc=bugi}', "");
Expect(0, 6687, '\p{^Sc=bugi}', "");
Expect(0, 6687, '\P{Sc=bugi}', "");
Expect(1, 6687, '\P{^Sc=bugi}', "");
Expect(0, 6688, '\p{Sc=bugi}', "");
Expect(1, 6688, '\p{^Sc=bugi}', "");
Expect(1, 6688, '\P{Sc=bugi}', "");
Expect(0, 6688, '\P{^Sc=bugi}', "");
Expect(1, 6687, '\p{Sc=-BUGI}', "");
Expect(0, 6687, '\p{^Sc=-BUGI}', "");
Expect(0, 6687, '\P{Sc=-BUGI}', "");
Expect(1, 6687, '\P{^Sc=-BUGI}', "");
Expect(0, 6688, '\p{Sc=-BUGI}', "");
Expect(1, 6688, '\p{^Sc=-BUGI}', "");
Expect(1, 6688, '\P{Sc=-BUGI}', "");
Expect(0, 6688, '\P{^Sc=-BUGI}', "");
Error('\p{Is_Script=:=	Buginese}');
Error('\P{Is_Script=:=	Buginese}');
Expect(1, 6687, '\p{Is_Script=buginese}', "");
Expect(0, 6687, '\p{^Is_Script=buginese}', "");
Expect(0, 6687, '\P{Is_Script=buginese}', "");
Expect(1, 6687, '\P{^Is_Script=buginese}', "");
Expect(0, 6688, '\p{Is_Script=buginese}', "");
Expect(1, 6688, '\p{^Is_Script=buginese}', "");
Expect(1, 6688, '\P{Is_Script=buginese}', "");
Expect(0, 6688, '\P{^Is_Script=buginese}', "");
Expect(1, 6687, '\p{Is_Script=	_Buginese}', "");
Expect(0, 6687, '\p{^Is_Script=	_Buginese}', "");
Expect(0, 6687, '\P{Is_Script=	_Buginese}', "");
Expect(1, 6687, '\P{^Is_Script=	_Buginese}', "");
Expect(0, 6688, '\p{Is_Script=	_Buginese}', "");
Expect(1, 6688, '\p{^Is_Script=	_Buginese}', "");
Expect(1, 6688, '\P{Is_Script=	_Buginese}', "");
Expect(0, 6688, '\P{^Is_Script=	_Buginese}', "");
Error('\p{Is_Sc=_/a/Bugi}');
Error('\P{Is_Sc=_/a/Bugi}');
Expect(1, 6687, '\p{Is_Sc=bugi}', "");
Expect(0, 6687, '\p{^Is_Sc=bugi}', "");
Expect(0, 6687, '\P{Is_Sc=bugi}', "");
Expect(1, 6687, '\P{^Is_Sc=bugi}', "");
Expect(0, 6688, '\p{Is_Sc=bugi}', "");
Expect(1, 6688, '\p{^Is_Sc=bugi}', "");
Expect(1, 6688, '\P{Is_Sc=bugi}', "");
Expect(0, 6688, '\P{^Is_Sc=bugi}', "");
Expect(1, 6687, '\p{Is_Sc=		bugi}', "");
Expect(0, 6687, '\p{^Is_Sc=		bugi}', "");
Expect(0, 6687, '\P{Is_Sc=		bugi}', "");
Expect(1, 6687, '\P{^Is_Sc=		bugi}', "");
Expect(0, 6688, '\p{Is_Sc=		bugi}', "");
Expect(1, 6688, '\p{^Is_Sc=		bugi}', "");
Expect(1, 6688, '\P{Is_Sc=		bugi}', "");
Expect(0, 6688, '\P{^Is_Sc=		bugi}', "");
Error('\p{Script=/a/INSCRIPTIONAL_Pahlavi}');
Error('\P{Script=/a/INSCRIPTIONAL_Pahlavi}');
Expect(1, 68479, '\p{Script=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Script=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Script=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Script=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Script=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Script=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Script=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Script=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Script:   --Inscriptional_Pahlavi}', "");
Expect(0, 68479, '\p{^Script:   --Inscriptional_Pahlavi}', "");
Expect(0, 68479, '\P{Script:   --Inscriptional_Pahlavi}', "");
Expect(1, 68479, '\P{^Script:   --Inscriptional_Pahlavi}', "");
Expect(0, 68480, '\p{Script:   --Inscriptional_Pahlavi}', "");
Expect(1, 68480, '\p{^Script:   --Inscriptional_Pahlavi}', "");
Expect(1, 68480, '\P{Script:   --Inscriptional_Pahlavi}', "");
Expect(0, 68480, '\P{^Script:   --Inscriptional_Pahlavi}', "");
Error('\p{Sc: Phli/a/}');
Error('\P{Sc: Phli/a/}');
Expect(1, 68479, '\p{Sc=phli}', "");
Expect(0, 68479, '\p{^Sc=phli}', "");
Expect(0, 68479, '\P{Sc=phli}', "");
Expect(1, 68479, '\P{^Sc=phli}', "");
Expect(0, 68480, '\p{Sc=phli}', "");
Expect(1, 68480, '\p{^Sc=phli}', "");
Expect(1, 68480, '\P{Sc=phli}', "");
Expect(0, 68480, '\P{^Sc=phli}', "");
Expect(1, 68479, '\p{Sc=_	Phli}', "");
Expect(0, 68479, '\p{^Sc=_	Phli}', "");
Expect(0, 68479, '\P{Sc=_	Phli}', "");
Expect(1, 68479, '\P{^Sc=_	Phli}', "");
Expect(0, 68480, '\p{Sc=_	Phli}', "");
Expect(1, 68480, '\p{^Sc=_	Phli}', "");
Expect(1, 68480, '\P{Sc=_	Phli}', "");
Expect(0, 68480, '\P{^Sc=_	Phli}', "");
Error('\p{Is_Script= 	Inscriptional_Pahlavi/a/}');
Error('\P{Is_Script= 	Inscriptional_Pahlavi/a/}');
Expect(1, 68479, '\p{Is_Script=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Is_Script=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Is_Script=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Is_Script=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Is_Script=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Is_Script=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Is_Script=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Is_Script=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Is_Script=	_Inscriptional_Pahlavi}', "");
Expect(0, 68479, '\p{^Is_Script=	_Inscriptional_Pahlavi}', "");
Expect(0, 68479, '\P{Is_Script=	_Inscriptional_Pahlavi}', "");
Expect(1, 68479, '\P{^Is_Script=	_Inscriptional_Pahlavi}', "");
Expect(0, 68480, '\p{Is_Script=	_Inscriptional_Pahlavi}', "");
Expect(1, 68480, '\p{^Is_Script=	_Inscriptional_Pahlavi}', "");
Expect(1, 68480, '\P{Is_Script=	_Inscriptional_Pahlavi}', "");
Expect(0, 68480, '\P{^Is_Script=	_Inscriptional_Pahlavi}', "");
Error('\p{Is_Sc=:=	_phli}');
Error('\P{Is_Sc=:=	_phli}');
Expect(1, 68479, '\p{Is_Sc:phli}', "");
Expect(0, 68479, '\p{^Is_Sc:phli}', "");
Expect(0, 68479, '\P{Is_Sc:phli}', "");
Expect(1, 68479, '\P{^Is_Sc:phli}', "");
Expect(0, 68480, '\p{Is_Sc:phli}', "");
Expect(1, 68480, '\p{^Is_Sc:phli}', "");
Expect(1, 68480, '\P{Is_Sc:phli}', "");
Expect(0, 68480, '\P{^Is_Sc:phli}', "");
Expect(1, 68479, '\p{Is_Sc=		phli}', "");
Expect(0, 68479, '\p{^Is_Sc=		phli}', "");
Expect(0, 68479, '\P{Is_Sc=		phli}', "");
Expect(1, 68479, '\P{^Is_Sc=		phli}', "");
Expect(0, 68480, '\p{Is_Sc=		phli}', "");
Expect(1, 68480, '\p{^Is_Sc=		phli}', "");
Expect(1, 68480, '\P{Is_Sc=		phli}', "");
Expect(0, 68480, '\P{^Is_Sc=		phli}', "");
Error('\p{Script=		Mandaic:=}');
Error('\P{Script=		Mandaic:=}');
Expect(1, 2142, '\p{Script=mandaic}', "");
Expect(0, 2142, '\p{^Script=mandaic}', "");
Expect(0, 2142, '\P{Script=mandaic}', "");
Expect(1, 2142, '\P{^Script=mandaic}', "");
Expect(0, 2143, '\p{Script=mandaic}', "");
Expect(1, 2143, '\p{^Script=mandaic}', "");
Expect(1, 2143, '\P{Script=mandaic}', "");
Expect(0, 2143, '\P{^Script=mandaic}', "");
Expect(1, 2142, '\p{Script: -_mandaic}', "");
Expect(0, 2142, '\p{^Script: -_mandaic}', "");
Expect(0, 2142, '\P{Script: -_mandaic}', "");
Expect(1, 2142, '\P{^Script: -_mandaic}', "");
Expect(0, 2143, '\p{Script: -_mandaic}', "");
Expect(1, 2143, '\p{^Script: -_mandaic}', "");
Expect(1, 2143, '\P{Script: -_mandaic}', "");
Expect(0, 2143, '\P{^Script: -_mandaic}', "");
Error('\p{Sc:	:=	_Mand}');
Error('\P{Sc:	:=	_Mand}');
Expect(1, 2142, '\p{Sc=mand}', "");
Expect(0, 2142, '\p{^Sc=mand}', "");
Expect(0, 2142, '\P{Sc=mand}', "");
Expect(1, 2142, '\P{^Sc=mand}', "");
Expect(0, 2143, '\p{Sc=mand}', "");
Expect(1, 2143, '\p{^Sc=mand}', "");
Expect(1, 2143, '\P{Sc=mand}', "");
Expect(0, 2143, '\P{^Sc=mand}', "");
Expect(1, 2142, '\p{Sc=_-Mand}', "");
Expect(0, 2142, '\p{^Sc=_-Mand}', "");
Expect(0, 2142, '\P{Sc=_-Mand}', "");
Expect(1, 2142, '\P{^Sc=_-Mand}', "");
Expect(0, 2143, '\p{Sc=_-Mand}', "");
Expect(1, 2143, '\p{^Sc=_-Mand}', "");
Expect(1, 2143, '\P{Sc=_-Mand}', "");
Expect(0, 2143, '\P{^Sc=_-Mand}', "");
Error('\p{Is_Script=:=-	MANDAIC}');
Error('\P{Is_Script=:=-	MANDAIC}');
Expect(1, 2142, '\p{Is_Script=mandaic}', "");
Expect(0, 2142, '\p{^Is_Script=mandaic}', "");
Expect(0, 2142, '\P{Is_Script=mandaic}', "");
Expect(1, 2142, '\P{^Is_Script=mandaic}', "");
Expect(0, 2143, '\p{Is_Script=mandaic}', "");
Expect(1, 2143, '\p{^Is_Script=mandaic}', "");
Expect(1, 2143, '\P{Is_Script=mandaic}', "");
Expect(0, 2143, '\P{^Is_Script=mandaic}', "");
Expect(1, 2142, '\p{Is_Script=_-Mandaic}', "");
Expect(0, 2142, '\p{^Is_Script=_-Mandaic}', "");
Expect(0, 2142, '\P{Is_Script=_-Mandaic}', "");
Expect(1, 2142, '\P{^Is_Script=_-Mandaic}', "");
Expect(0, 2143, '\p{Is_Script=_-Mandaic}', "");
Expect(1, 2143, '\p{^Is_Script=_-Mandaic}', "");
Expect(1, 2143, '\P{Is_Script=_-Mandaic}', "");
Expect(0, 2143, '\P{^Is_Script=_-Mandaic}', "");
Error('\p{Is_Sc= :=MAND}');
Error('\P{Is_Sc= :=MAND}');
Expect(1, 2142, '\p{Is_Sc=mand}', "");
Expect(0, 2142, '\p{^Is_Sc=mand}', "");
Expect(0, 2142, '\P{Is_Sc=mand}', "");
Expect(1, 2142, '\P{^Is_Sc=mand}', "");
Expect(0, 2143, '\p{Is_Sc=mand}', "");
Expect(1, 2143, '\p{^Is_Sc=mand}', "");
Expect(1, 2143, '\P{Is_Sc=mand}', "");
Expect(0, 2143, '\P{^Is_Sc=mand}', "");
Expect(1, 2142, '\p{Is_Sc=	_mand}', "");
Expect(0, 2142, '\p{^Is_Sc=	_mand}', "");
Expect(0, 2142, '\P{Is_Sc=	_mand}', "");
Expect(1, 2142, '\P{^Is_Sc=	_mand}', "");
Expect(0, 2143, '\p{Is_Sc=	_mand}', "");
Expect(1, 2143, '\p{^Is_Sc=	_mand}', "");
Expect(1, 2143, '\P{Is_Sc=	_mand}', "");
Expect(0, 2143, '\P{^Is_Sc=	_mand}', "");
Error('\p{Script=_:=TAI_Tham}');
Error('\P{Script=_:=TAI_Tham}');
Expect(1, 6829, '\p{Script: taitham}', "");
Expect(0, 6829, '\p{^Script: taitham}', "");
Expect(0, 6829, '\P{Script: taitham}', "");
Expect(1, 6829, '\P{^Script: taitham}', "");
Expect(0, 6830, '\p{Script: taitham}', "");
Expect(1, 6830, '\p{^Script: taitham}', "");
Expect(1, 6830, '\P{Script: taitham}', "");
Expect(0, 6830, '\P{^Script: taitham}', "");
Expect(1, 6829, '\p{Script=__Tai_Tham}', "");
Expect(0, 6829, '\p{^Script=__Tai_Tham}', "");
Expect(0, 6829, '\P{Script=__Tai_Tham}', "");
Expect(1, 6829, '\P{^Script=__Tai_Tham}', "");
Expect(0, 6830, '\p{Script=__Tai_Tham}', "");
Expect(1, 6830, '\p{^Script=__Tai_Tham}', "");
Expect(1, 6830, '\P{Script=__Tai_Tham}', "");
Expect(0, 6830, '\P{^Script=__Tai_Tham}', "");
Error('\p{Sc= /a/lana}');
Error('\P{Sc= /a/lana}');
Expect(1, 6829, '\p{Sc=lana}', "");
Expect(0, 6829, '\p{^Sc=lana}', "");
Expect(0, 6829, '\P{Sc=lana}', "");
Expect(1, 6829, '\P{^Sc=lana}', "");
Expect(0, 6830, '\p{Sc=lana}', "");
Expect(1, 6830, '\p{^Sc=lana}', "");
Expect(1, 6830, '\P{Sc=lana}', "");
Expect(0, 6830, '\P{^Sc=lana}', "");
Expect(1, 6829, '\p{Sc:    -Lana}', "");
Expect(0, 6829, '\p{^Sc:    -Lana}', "");
Expect(0, 6829, '\P{Sc:    -Lana}', "");
Expect(1, 6829, '\P{^Sc:    -Lana}', "");
Expect(0, 6830, '\p{Sc:    -Lana}', "");
Expect(1, 6830, '\p{^Sc:    -Lana}', "");
Expect(1, 6830, '\P{Sc:    -Lana}', "");
Expect(0, 6830, '\P{^Sc:    -Lana}', "");
Error('\p{Is_Script= Tai_tham/a/}');
Error('\P{Is_Script= Tai_tham/a/}');
Expect(1, 6829, '\p{Is_Script=taitham}', "");
Expect(0, 6829, '\p{^Is_Script=taitham}', "");
Expect(0, 6829, '\P{Is_Script=taitham}', "");
Expect(1, 6829, '\P{^Is_Script=taitham}', "");
Expect(0, 6830, '\p{Is_Script=taitham}', "");
Expect(1, 6830, '\p{^Is_Script=taitham}', "");
Expect(1, 6830, '\P{Is_Script=taitham}', "");
Expect(0, 6830, '\P{^Is_Script=taitham}', "");
Expect(1, 6829, '\p{Is_Script=	-tai_tham}', "");
Expect(0, 6829, '\p{^Is_Script=	-tai_tham}', "");
Expect(0, 6829, '\P{Is_Script=	-tai_tham}', "");
Expect(1, 6829, '\P{^Is_Script=	-tai_tham}', "");
Expect(0, 6830, '\p{Is_Script=	-tai_tham}', "");
Expect(1, 6830, '\p{^Is_Script=	-tai_tham}', "");
Expect(1, 6830, '\P{Is_Script=	-tai_tham}', "");
Expect(0, 6830, '\P{^Is_Script=	-tai_tham}', "");
Error('\p{Is_Sc:-/a/Lana}');
Error('\P{Is_Sc:-/a/Lana}');
Expect(1, 6829, '\p{Is_Sc=lana}', "");
Expect(0, 6829, '\p{^Is_Sc=lana}', "");
Expect(0, 6829, '\P{Is_Sc=lana}', "");
Expect(1, 6829, '\P{^Is_Sc=lana}', "");
Expect(0, 6830, '\p{Is_Sc=lana}', "");
Expect(1, 6830, '\p{^Is_Sc=lana}', "");
Expect(1, 6830, '\P{Is_Sc=lana}', "");
Expect(0, 6830, '\P{^Is_Sc=lana}', "");
Expect(1, 6829, '\p{Is_Sc=-Lana}', "");
Expect(0, 6829, '\p{^Is_Sc=-Lana}', "");
Expect(0, 6829, '\P{Is_Sc=-Lana}', "");
Expect(1, 6829, '\P{^Is_Sc=-Lana}', "");
Expect(0, 6830, '\p{Is_Sc=-Lana}', "");
Expect(1, 6830, '\p{^Is_Sc=-Lana}', "");
Expect(1, 6830, '\P{Is_Sc=-Lana}', "");
Expect(0, 6830, '\P{^Is_Sc=-Lana}', "");
Error('\p{Script=_/a/gurmukhi}');
Error('\P{Script=_/a/gurmukhi}');
Expect(1, 2677, '\p{Script=gurmukhi}', "");
Expect(0, 2677, '\p{^Script=gurmukhi}', "");
Expect(0, 2677, '\P{Script=gurmukhi}', "");
Expect(1, 2677, '\P{^Script=gurmukhi}', "");
Expect(0, 2678, '\p{Script=gurmukhi}', "");
Expect(1, 2678, '\p{^Script=gurmukhi}', "");
Expect(1, 2678, '\P{Script=gurmukhi}', "");
Expect(0, 2678, '\P{^Script=gurmukhi}', "");
Expect(1, 2677, '\p{Script=_-GURMUKHI}', "");
Expect(0, 2677, '\p{^Script=_-GURMUKHI}', "");
Expect(0, 2677, '\P{Script=_-GURMUKHI}', "");
Expect(1, 2677, '\P{^Script=_-GURMUKHI}', "");
Expect(0, 2678, '\p{Script=_-GURMUKHI}', "");
Expect(1, 2678, '\p{^Script=_-GURMUKHI}', "");
Expect(1, 2678, '\P{Script=_-GURMUKHI}', "");
Expect(0, 2678, '\P{^Script=_-GURMUKHI}', "");
Error('\p{Sc=/a/	guru}');
Error('\P{Sc=/a/	guru}');
Expect(1, 2677, '\p{Sc=guru}', "");
Expect(0, 2677, '\p{^Sc=guru}', "");
Expect(0, 2677, '\P{Sc=guru}', "");
Expect(1, 2677, '\P{^Sc=guru}', "");
Expect(0, 2678, '\p{Sc=guru}', "");
Expect(1, 2678, '\p{^Sc=guru}', "");
Expect(1, 2678, '\P{Sc=guru}', "");
Expect(0, 2678, '\P{^Sc=guru}', "");
Expect(1, 2677, '\p{Sc=	_Guru}', "");
Expect(0, 2677, '\p{^Sc=	_Guru}', "");
Expect(0, 2677, '\P{Sc=	_Guru}', "");
Expect(1, 2677, '\P{^Sc=	_Guru}', "");
Expect(0, 2678, '\p{Sc=	_Guru}', "");
Expect(1, 2678, '\p{^Sc=	_Guru}', "");
Expect(1, 2678, '\P{Sc=	_Guru}', "");
Expect(0, 2678, '\P{^Sc=	_Guru}', "");
Error('\p{Is_Script= gurmukhi/a/}');
Error('\P{Is_Script= gurmukhi/a/}');
Expect(1, 2677, '\p{Is_Script=gurmukhi}', "");
Expect(0, 2677, '\p{^Is_Script=gurmukhi}', "");
Expect(0, 2677, '\P{Is_Script=gurmukhi}', "");
Expect(1, 2677, '\P{^Is_Script=gurmukhi}', "");
Expect(0, 2678, '\p{Is_Script=gurmukhi}', "");
Expect(1, 2678, '\p{^Is_Script=gurmukhi}', "");
Expect(1, 2678, '\P{Is_Script=gurmukhi}', "");
Expect(0, 2678, '\P{^Is_Script=gurmukhi}', "");
Expect(1, 2677, '\p{Is_Script=  Gurmukhi}', "");
Expect(0, 2677, '\p{^Is_Script=  Gurmukhi}', "");
Expect(0, 2677, '\P{Is_Script=  Gurmukhi}', "");
Expect(1, 2677, '\P{^Is_Script=  Gurmukhi}', "");
Expect(0, 2678, '\p{Is_Script=  Gurmukhi}', "");
Expect(1, 2678, '\p{^Is_Script=  Gurmukhi}', "");
Expect(1, 2678, '\P{Is_Script=  Gurmukhi}', "");
Expect(0, 2678, '\P{^Is_Script=  Gurmukhi}', "");
Error('\p{Is_Sc=	:=Guru}');
Error('\P{Is_Sc=	:=Guru}');
Expect(1, 2677, '\p{Is_Sc=guru}', "");
Expect(0, 2677, '\p{^Is_Sc=guru}', "");
Expect(0, 2677, '\P{Is_Sc=guru}', "");
Expect(1, 2677, '\P{^Is_Sc=guru}', "");
Expect(0, 2678, '\p{Is_Sc=guru}', "");
Expect(1, 2678, '\p{^Is_Sc=guru}', "");
Expect(1, 2678, '\P{Is_Sc=guru}', "");
Expect(0, 2678, '\P{^Is_Sc=guru}', "");
Expect(1, 2677, '\p{Is_Sc=	-Guru}', "");
Expect(0, 2677, '\p{^Is_Sc=	-Guru}', "");
Expect(0, 2677, '\P{Is_Sc=	-Guru}', "");
Expect(1, 2677, '\P{^Is_Sc=	-Guru}', "");
Expect(0, 2678, '\p{Is_Sc=	-Guru}', "");
Expect(1, 2678, '\p{^Is_Sc=	-Guru}', "");
Expect(1, 2678, '\P{Is_Sc=	-Guru}', "");
Expect(0, 2678, '\P{^Is_Sc=	-Guru}', "");
Error('\p{Script=-:=Cypriot}');
Error('\P{Script=-:=Cypriot}');
Expect(1, 67647, '\p{Script: cypriot}', "");
Expect(0, 67647, '\p{^Script: cypriot}', "");
Expect(0, 67647, '\P{Script: cypriot}', "");
Expect(1, 67647, '\P{^Script: cypriot}', "");
Expect(0, 67648, '\p{Script: cypriot}', "");
Expect(1, 67648, '\p{^Script: cypriot}', "");
Expect(1, 67648, '\P{Script: cypriot}', "");
Expect(0, 67648, '\P{^Script: cypriot}', "");
Expect(1, 67647, '\p{Script= CYPRIOT}', "");
Expect(0, 67647, '\p{^Script= CYPRIOT}', "");
Expect(0, 67647, '\P{Script= CYPRIOT}', "");
Expect(1, 67647, '\P{^Script= CYPRIOT}', "");
Expect(0, 67648, '\p{Script= CYPRIOT}', "");
Expect(1, 67648, '\p{^Script= CYPRIOT}', "");
Expect(1, 67648, '\P{Script= CYPRIOT}', "");
Expect(0, 67648, '\P{^Script= CYPRIOT}', "");
Error('\p{Sc=/a/cprt}');
Error('\P{Sc=/a/cprt}');
Expect(1, 67647, '\p{Sc=cprt}', "");
Expect(0, 67647, '\p{^Sc=cprt}', "");
Expect(0, 67647, '\P{Sc=cprt}', "");
Expect(1, 67647, '\P{^Sc=cprt}', "");
Expect(0, 67648, '\p{Sc=cprt}', "");
Expect(1, 67648, '\p{^Sc=cprt}', "");
Expect(1, 67648, '\P{Sc=cprt}', "");
Expect(0, 67648, '\P{^Sc=cprt}', "");
Expect(1, 67647, '\p{Sc:   -	cprt}', "");
Expect(0, 67647, '\p{^Sc:   -	cprt}', "");
Expect(0, 67647, '\P{Sc:   -	cprt}', "");
Expect(1, 67647, '\P{^Sc:   -	cprt}', "");
Expect(0, 67648, '\p{Sc:   -	cprt}', "");
Expect(1, 67648, '\p{^Sc:   -	cprt}', "");
Expect(1, 67648, '\P{Sc:   -	cprt}', "");
Expect(0, 67648, '\P{^Sc:   -	cprt}', "");
Error('\p{Is_Script=:= 	Cypriot}');
Error('\P{Is_Script=:= 	Cypriot}');
Expect(1, 67647, '\p{Is_Script=cypriot}', "");
Expect(0, 67647, '\p{^Is_Script=cypriot}', "");
Expect(0, 67647, '\P{Is_Script=cypriot}', "");
Expect(1, 67647, '\P{^Is_Script=cypriot}', "");
Expect(0, 67648, '\p{Is_Script=cypriot}', "");
Expect(1, 67648, '\p{^Is_Script=cypriot}', "");
Expect(1, 67648, '\P{Is_Script=cypriot}', "");
Expect(0, 67648, '\P{^Is_Script=cypriot}', "");
Expect(1, 67647, '\p{Is_Script=_ CYPRIOT}', "");
Expect(0, 67647, '\p{^Is_Script=_ CYPRIOT}', "");
Expect(0, 67647, '\P{Is_Script=_ CYPRIOT}', "");
Expect(1, 67647, '\P{^Is_Script=_ CYPRIOT}', "");
Expect(0, 67648, '\p{Is_Script=_ CYPRIOT}', "");
Expect(1, 67648, '\p{^Is_Script=_ CYPRIOT}', "");
Expect(1, 67648, '\P{Is_Script=_ CYPRIOT}', "");
Expect(0, 67648, '\P{^Is_Script=_ CYPRIOT}', "");
Error('\p{Is_Sc: _-Cprt/a/}');
Error('\P{Is_Sc: _-Cprt/a/}');
Expect(1, 67647, '\p{Is_Sc=cprt}', "");
Expect(0, 67647, '\p{^Is_Sc=cprt}', "");
Expect(0, 67647, '\P{Is_Sc=cprt}', "");
Expect(1, 67647, '\P{^Is_Sc=cprt}', "");
Expect(0, 67648, '\p{Is_Sc=cprt}', "");
Expect(1, 67648, '\p{^Is_Sc=cprt}', "");
Expect(1, 67648, '\P{Is_Sc=cprt}', "");
Expect(0, 67648, '\P{^Is_Sc=cprt}', "");
Expect(1, 67647, '\p{Is_Sc=	 Cprt}', "");
Expect(0, 67647, '\p{^Is_Sc=	 Cprt}', "");
Expect(0, 67647, '\P{Is_Sc=	 Cprt}', "");
Expect(1, 67647, '\P{^Is_Sc=	 Cprt}', "");
Expect(0, 67648, '\p{Is_Sc=	 Cprt}', "");
Expect(1, 67648, '\p{^Is_Sc=	 Cprt}', "");
Expect(1, 67648, '\P{Is_Sc=	 Cprt}', "");
Expect(0, 67648, '\P{^Is_Sc=	 Cprt}', "");
Error('\p{Script=:=BATAK}');
Error('\P{Script=:=BATAK}');
Expect(1, 7167, '\p{Script=batak}', "");
Expect(0, 7167, '\p{^Script=batak}', "");
Expect(0, 7167, '\P{Script=batak}', "");
Expect(1, 7167, '\P{^Script=batak}', "");
Expect(0, 7168, '\p{Script=batak}', "");
Expect(1, 7168, '\p{^Script=batak}', "");
Expect(1, 7168, '\P{Script=batak}', "");
Expect(0, 7168, '\P{^Script=batak}', "");
Expect(1, 7167, '\p{Script=_	BATAK}', "");
Expect(0, 7167, '\p{^Script=_	BATAK}', "");
Expect(0, 7167, '\P{Script=_	BATAK}', "");
Expect(1, 7167, '\P{^Script=_	BATAK}', "");
Expect(0, 7168, '\p{Script=_	BATAK}', "");
Expect(1, 7168, '\p{^Script=_	BATAK}', "");
Expect(1, 7168, '\P{Script=_	BATAK}', "");
Expect(0, 7168, '\P{^Script=_	BATAK}', "");
Error('\p{Sc=:= Batk}');
Error('\P{Sc=:= Batk}');
Expect(1, 7167, '\p{Sc=batk}', "");
Expect(0, 7167, '\p{^Sc=batk}', "");
Expect(0, 7167, '\P{Sc=batk}', "");
Expect(1, 7167, '\P{^Sc=batk}', "");
Expect(0, 7168, '\p{Sc=batk}', "");
Expect(1, 7168, '\p{^Sc=batk}', "");
Expect(1, 7168, '\P{Sc=batk}', "");
Expect(0, 7168, '\P{^Sc=batk}', "");
Expect(1, 7167, '\p{Sc=_-batk}', "");
Expect(0, 7167, '\p{^Sc=_-batk}', "");
Expect(0, 7167, '\P{Sc=_-batk}', "");
Expect(1, 7167, '\P{^Sc=_-batk}', "");
Expect(0, 7168, '\p{Sc=_-batk}', "");
Expect(1, 7168, '\p{^Sc=_-batk}', "");
Expect(1, 7168, '\P{Sc=_-batk}', "");
Expect(0, 7168, '\P{^Sc=_-batk}', "");
Error('\p{Is_Script=_batak/a/}');
Error('\P{Is_Script=_batak/a/}');
Expect(1, 7167, '\p{Is_Script=batak}', "");
Expect(0, 7167, '\p{^Is_Script=batak}', "");
Expect(0, 7167, '\P{Is_Script=batak}', "");
Expect(1, 7167, '\P{^Is_Script=batak}', "");
Expect(0, 7168, '\p{Is_Script=batak}', "");
Expect(1, 7168, '\p{^Is_Script=batak}', "");
Expect(1, 7168, '\P{Is_Script=batak}', "");
Expect(0, 7168, '\P{^Is_Script=batak}', "");
Expect(1, 7167, '\p{Is_Script:--Batak}', "");
Expect(0, 7167, '\p{^Is_Script:--Batak}', "");
Expect(0, 7167, '\P{Is_Script:--Batak}', "");
Expect(1, 7167, '\P{^Is_Script:--Batak}', "");
Expect(0, 7168, '\p{Is_Script:--Batak}', "");
Expect(1, 7168, '\p{^Is_Script:--Batak}', "");
Expect(1, 7168, '\P{Is_Script:--Batak}', "");
Expect(0, 7168, '\P{^Is_Script:--Batak}', "");
Error('\p{Is_Sc:    _BATK/a/}');
Error('\P{Is_Sc:    _BATK/a/}');
Expect(1, 7167, '\p{Is_Sc:	batk}', "");
Expect(0, 7167, '\p{^Is_Sc:	batk}', "");
Expect(0, 7167, '\P{Is_Sc:	batk}', "");
Expect(1, 7167, '\P{^Is_Sc:	batk}', "");
Expect(0, 7168, '\p{Is_Sc:	batk}', "");
Expect(1, 7168, '\p{^Is_Sc:	batk}', "");
Expect(1, 7168, '\P{Is_Sc:	batk}', "");
Expect(0, 7168, '\P{^Is_Sc:	batk}', "");
Expect(1, 7167, '\p{Is_Sc=		Batk}', "");
Expect(0, 7167, '\p{^Is_Sc=		Batk}', "");
Expect(0, 7167, '\P{Is_Sc=		Batk}', "");
Expect(1, 7167, '\P{^Is_Sc=		Batk}', "");
Expect(0, 7168, '\p{Is_Sc=		Batk}', "");
Expect(1, 7168, '\p{^Is_Sc=		Batk}', "");
Expect(1, 7168, '\P{Is_Sc=		Batk}', "");
Expect(0, 7168, '\P{^Is_Sc=		Batk}', "");
Error('\p{Script=/a/_Hangul}');
Error('\P{Script=/a/_Hangul}');
Expect(1, 65500, '\p{Script=hangul}', "");
Expect(0, 65500, '\p{^Script=hangul}', "");
Expect(0, 65500, '\P{Script=hangul}', "");
Expect(1, 65500, '\P{^Script=hangul}', "");
Expect(0, 65501, '\p{Script=hangul}', "");
Expect(1, 65501, '\p{^Script=hangul}', "");
Expect(1, 65501, '\P{Script=hangul}', "");
Expect(0, 65501, '\P{^Script=hangul}', "");
Expect(1, 65500, '\p{Script=  Hangul}', "");
Expect(0, 65500, '\p{^Script=  Hangul}', "");
Expect(0, 65500, '\P{Script=  Hangul}', "");
Expect(1, 65500, '\P{^Script=  Hangul}', "");
Expect(0, 65501, '\p{Script=  Hangul}', "");
Expect(1, 65501, '\p{^Script=  Hangul}', "");
Expect(1, 65501, '\P{Script=  Hangul}', "");
Expect(0, 65501, '\P{^Script=  Hangul}', "");
Error('\p{Sc=:=-_HANG}');
Error('\P{Sc=:=-_HANG}');
Expect(1, 65500, '\p{Sc: hang}', "");
Expect(0, 65500, '\p{^Sc: hang}', "");
Expect(0, 65500, '\P{Sc: hang}', "");
Expect(1, 65500, '\P{^Sc: hang}', "");
Expect(0, 65501, '\p{Sc: hang}', "");
Expect(1, 65501, '\p{^Sc: hang}', "");
Expect(1, 65501, '\P{Sc: hang}', "");
Expect(0, 65501, '\P{^Sc: hang}', "");
Expect(1, 65500, '\p{Sc=-_hang}', "");
Expect(0, 65500, '\p{^Sc=-_hang}', "");
Expect(0, 65500, '\P{Sc=-_hang}', "");
Expect(1, 65500, '\P{^Sc=-_hang}', "");
Expect(0, 65501, '\p{Sc=-_hang}', "");
Expect(1, 65501, '\p{^Sc=-_hang}', "");
Expect(1, 65501, '\P{Sc=-_hang}', "");
Expect(0, 65501, '\P{^Sc=-_hang}', "");
Error('\p{Is_Script=:=	_HANGUL}');
Error('\P{Is_Script=:=	_HANGUL}');
Expect(1, 65500, '\p{Is_Script=hangul}', "");
Expect(0, 65500, '\p{^Is_Script=hangul}', "");
Expect(0, 65500, '\P{Is_Script=hangul}', "");
Expect(1, 65500, '\P{^Is_Script=hangul}', "");
Expect(0, 65501, '\p{Is_Script=hangul}', "");
Expect(1, 65501, '\p{^Is_Script=hangul}', "");
Expect(1, 65501, '\P{Is_Script=hangul}', "");
Expect(0, 65501, '\P{^Is_Script=hangul}', "");
Expect(1, 65500, '\p{Is_Script=-	hangul}', "");
Expect(0, 65500, '\p{^Is_Script=-	hangul}', "");
Expect(0, 65500, '\P{Is_Script=-	hangul}', "");
Expect(1, 65500, '\P{^Is_Script=-	hangul}', "");
Expect(0, 65501, '\p{Is_Script=-	hangul}', "");
Expect(1, 65501, '\p{^Is_Script=-	hangul}', "");
Expect(1, 65501, '\P{Is_Script=-	hangul}', "");
Expect(0, 65501, '\P{^Is_Script=-	hangul}', "");
Error('\p{Is_Sc=/a/-hang}');
Error('\P{Is_Sc=/a/-hang}');
Expect(1, 65500, '\p{Is_Sc=hang}', "");
Expect(0, 65500, '\p{^Is_Sc=hang}', "");
Expect(0, 65500, '\P{Is_Sc=hang}', "");
Expect(1, 65500, '\P{^Is_Sc=hang}', "");
Expect(0, 65501, '\p{Is_Sc=hang}', "");
Expect(1, 65501, '\p{^Is_Sc=hang}', "");
Expect(1, 65501, '\P{Is_Sc=hang}', "");
Expect(0, 65501, '\P{^Is_Sc=hang}', "");
Expect(1, 65500, '\p{Is_Sc=	Hang}', "");
Expect(0, 65500, '\p{^Is_Sc=	Hang}', "");
Expect(0, 65500, '\P{Is_Sc=	Hang}', "");
Expect(1, 65500, '\P{^Is_Sc=	Hang}', "");
Expect(0, 65501, '\p{Is_Sc=	Hang}', "");
Expect(1, 65501, '\p{^Is_Sc=	Hang}', "");
Expect(1, 65501, '\P{Is_Sc=	Hang}', "");
Expect(0, 65501, '\P{^Is_Sc=	Hang}', "");
Error('\p{Script=/a/UNKNOWN}');
Error('\P{Script=/a/UNKNOWN}');
Expect(1, 918000, '\p{Script=unknown}', "");
Expect(0, 918000, '\p{^Script=unknown}', "");
Expect(0, 918000, '\P{Script=unknown}', "");
Expect(1, 918000, '\P{^Script=unknown}', "");
Expect(0, 917999, '\p{Script=unknown}', "");
Expect(1, 917999, '\p{^Script=unknown}', "");
Expect(1, 917999, '\P{Script=unknown}', "");
Expect(0, 917999, '\P{^Script=unknown}', "");
Expect(1, 918000, '\p{Script=	 UNKNOWN}', "");
Expect(0, 918000, '\p{^Script=	 UNKNOWN}', "");
Expect(0, 918000, '\P{Script=	 UNKNOWN}', "");
Expect(1, 918000, '\P{^Script=	 UNKNOWN}', "");
Expect(0, 917999, '\p{Script=	 UNKNOWN}', "");
Expect(1, 917999, '\p{^Script=	 UNKNOWN}', "");
Expect(1, 917999, '\P{Script=	 UNKNOWN}', "");
Expect(0, 917999, '\P{^Script=	 UNKNOWN}', "");
Error('\p{Sc=/a/-ZZZZ}');
Error('\P{Sc=/a/-ZZZZ}');
Expect(1, 918000, '\p{Sc=zzzz}', "");
Expect(0, 918000, '\p{^Sc=zzzz}', "");
Expect(0, 918000, '\P{Sc=zzzz}', "");
Expect(1, 918000, '\P{^Sc=zzzz}', "");
Expect(0, 917999, '\p{Sc=zzzz}', "");
Expect(1, 917999, '\p{^Sc=zzzz}', "");
Expect(1, 917999, '\P{Sc=zzzz}', "");
Expect(0, 917999, '\P{^Sc=zzzz}', "");
Expect(1, 918000, '\p{Sc=_zzzz}', "");
Expect(0, 918000, '\p{^Sc=_zzzz}', "");
Expect(0, 918000, '\P{Sc=_zzzz}', "");
Expect(1, 918000, '\P{^Sc=_zzzz}', "");
Expect(0, 917999, '\p{Sc=_zzzz}', "");
Expect(1, 917999, '\p{^Sc=_zzzz}', "");
Expect(1, 917999, '\P{Sc=_zzzz}', "");
Expect(0, 917999, '\P{^Sc=_zzzz}', "");
Error('\p{Is_Script=_-UNKNOWN/a/}');
Error('\P{Is_Script=_-UNKNOWN/a/}');
Expect(1, 918000, '\p{Is_Script=unknown}', "");
Expect(0, 918000, '\p{^Is_Script=unknown}', "");
Expect(0, 918000, '\P{Is_Script=unknown}', "");
Expect(1, 918000, '\P{^Is_Script=unknown}', "");
Expect(0, 917999, '\p{Is_Script=unknown}', "");
Expect(1, 917999, '\p{^Is_Script=unknown}', "");
Expect(1, 917999, '\P{Is_Script=unknown}', "");
Expect(0, 917999, '\P{^Is_Script=unknown}', "");
Expect(1, 918000, '\p{Is_Script=-_Unknown}', "");
Expect(0, 918000, '\p{^Is_Script=-_Unknown}', "");
Expect(0, 918000, '\P{Is_Script=-_Unknown}', "");
Expect(1, 918000, '\P{^Is_Script=-_Unknown}', "");
Expect(0, 917999, '\p{Is_Script=-_Unknown}', "");
Expect(1, 917999, '\p{^Is_Script=-_Unknown}', "");
Expect(1, 917999, '\P{Is_Script=-_Unknown}', "");
Expect(0, 917999, '\P{^Is_Script=-_Unknown}', "");
Error('\p{Is_Sc=_/a/zzzz}');
Error('\P{Is_Sc=_/a/zzzz}');
Expect(1, 918000, '\p{Is_Sc=zzzz}', "");
Expect(0, 918000, '\p{^Is_Sc=zzzz}', "");
Expect(0, 918000, '\P{Is_Sc=zzzz}', "");
Expect(1, 918000, '\P{^Is_Sc=zzzz}', "");
Expect(0, 917999, '\p{Is_Sc=zzzz}', "");
Expect(1, 917999, '\p{^Is_Sc=zzzz}', "");
Expect(1, 917999, '\P{Is_Sc=zzzz}', "");
Expect(0, 917999, '\P{^Is_Sc=zzzz}', "");
Expect(1, 918000, '\p{Is_Sc=- Zzzz}', "");
Expect(0, 918000, '\p{^Is_Sc=- Zzzz}', "");
Expect(0, 918000, '\P{Is_Sc=- Zzzz}', "");
Expect(1, 918000, '\P{^Is_Sc=- Zzzz}', "");
Expect(0, 917999, '\p{Is_Sc=- Zzzz}', "");
Expect(1, 917999, '\p{^Is_Sc=- Zzzz}', "");
Expect(1, 917999, '\P{Is_Sc=- Zzzz}', "");
Expect(0, 917999, '\P{^Is_Sc=- Zzzz}', "");
Error('\p{Script=_ DESERET:=}');
Error('\P{Script=_ DESERET:=}');
Expect(1, 66639, '\p{Script=deseret}', "");
Expect(0, 66639, '\p{^Script=deseret}', "");
Expect(0, 66639, '\P{Script=deseret}', "");
Expect(1, 66639, '\P{^Script=deseret}', "");
Expect(0, 66640, '\p{Script=deseret}', "");
Expect(1, 66640, '\p{^Script=deseret}', "");
Expect(1, 66640, '\P{Script=deseret}', "");
Expect(0, 66640, '\P{^Script=deseret}', "");
Expect(1, 66639, '\p{Script= _Deseret}', "");
Expect(0, 66639, '\p{^Script= _Deseret}', "");
Expect(0, 66639, '\P{Script= _Deseret}', "");
Expect(1, 66639, '\P{^Script= _Deseret}', "");
Expect(0, 66640, '\p{Script= _Deseret}', "");
Expect(1, 66640, '\p{^Script= _Deseret}', "");
Expect(1, 66640, '\P{Script= _Deseret}', "");
Expect(0, 66640, '\P{^Script= _Deseret}', "");
Error('\p{Sc= :=dsrt}');
Error('\P{Sc= :=dsrt}');
Expect(1, 66639, '\p{Sc=dsrt}', "");
Expect(0, 66639, '\p{^Sc=dsrt}', "");
Expect(0, 66639, '\P{Sc=dsrt}', "");
Expect(1, 66639, '\P{^Sc=dsrt}', "");
Expect(0, 66640, '\p{Sc=dsrt}', "");
Expect(1, 66640, '\p{^Sc=dsrt}', "");
Expect(1, 66640, '\P{Sc=dsrt}', "");
Expect(0, 66640, '\P{^Sc=dsrt}', "");
Expect(1, 66639, '\p{Sc=- Dsrt}', "");
Expect(0, 66639, '\p{^Sc=- Dsrt}', "");
Expect(0, 66639, '\P{Sc=- Dsrt}', "");
Expect(1, 66639, '\P{^Sc=- Dsrt}', "");
Expect(0, 66640, '\p{Sc=- Dsrt}', "");
Expect(1, 66640, '\p{^Sc=- Dsrt}', "");
Expect(1, 66640, '\P{Sc=- Dsrt}', "");
Expect(0, 66640, '\P{^Sc=- Dsrt}', "");
Error('\p{Is_Script=/a/ Deseret}');
Error('\P{Is_Script=/a/ Deseret}');
Expect(1, 66639, '\p{Is_Script=deseret}', "");
Expect(0, 66639, '\p{^Is_Script=deseret}', "");
Expect(0, 66639, '\P{Is_Script=deseret}', "");
Expect(1, 66639, '\P{^Is_Script=deseret}', "");
Expect(0, 66640, '\p{Is_Script=deseret}', "");
Expect(1, 66640, '\p{^Is_Script=deseret}', "");
Expect(1, 66640, '\P{Is_Script=deseret}', "");
Expect(0, 66640, '\P{^Is_Script=deseret}', "");
Expect(1, 66639, '\p{Is_Script= deseret}', "");
Expect(0, 66639, '\p{^Is_Script= deseret}', "");
Expect(0, 66639, '\P{Is_Script= deseret}', "");
Expect(1, 66639, '\P{^Is_Script= deseret}', "");
Expect(0, 66640, '\p{Is_Script= deseret}', "");
Expect(1, 66640, '\p{^Is_Script= deseret}', "");
Expect(1, 66640, '\P{Is_Script= deseret}', "");
Expect(0, 66640, '\P{^Is_Script= deseret}', "");
Error('\p{Is_Sc=:=_dsrt}');
Error('\P{Is_Sc=:=_dsrt}');
Expect(1, 66639, '\p{Is_Sc=dsrt}', "");
Expect(0, 66639, '\p{^Is_Sc=dsrt}', "");
Expect(0, 66639, '\P{Is_Sc=dsrt}', "");
Expect(1, 66639, '\P{^Is_Sc=dsrt}', "");
Expect(0, 66640, '\p{Is_Sc=dsrt}', "");
Expect(1, 66640, '\p{^Is_Sc=dsrt}', "");
Expect(1, 66640, '\P{Is_Sc=dsrt}', "");
Expect(0, 66640, '\P{^Is_Sc=dsrt}', "");
Expect(1, 66639, '\p{Is_Sc=  Dsrt}', "");
Expect(0, 66639, '\p{^Is_Sc=  Dsrt}', "");
Expect(0, 66639, '\P{Is_Sc=  Dsrt}', "");
Expect(1, 66639, '\P{^Is_Sc=  Dsrt}', "");
Expect(0, 66640, '\p{Is_Sc=  Dsrt}', "");
Expect(1, 66640, '\p{^Is_Sc=  Dsrt}', "");
Expect(1, 66640, '\P{Is_Sc=  Dsrt}', "");
Expect(0, 66640, '\P{^Is_Sc=  Dsrt}', "");
Error('\p{Script=_	CUNEIFORM/a/}');
Error('\P{Script=_	CUNEIFORM/a/}');
Expect(1, 74867, '\p{Script=cuneiform}', "");
Expect(0, 74867, '\p{^Script=cuneiform}', "");
Expect(0, 74867, '\P{Script=cuneiform}', "");
Expect(1, 74867, '\P{^Script=cuneiform}', "");
Expect(0, 74868, '\p{Script=cuneiform}', "");
Expect(1, 74868, '\p{^Script=cuneiform}', "");
Expect(1, 74868, '\P{Script=cuneiform}', "");
Expect(0, 74868, '\P{^Script=cuneiform}', "");
Expect(1, 74867, '\p{Script: CUNEIFORM}', "");
Expect(0, 74867, '\p{^Script: CUNEIFORM}', "");
Expect(0, 74867, '\P{Script: CUNEIFORM}', "");
Expect(1, 74867, '\P{^Script: CUNEIFORM}', "");
Expect(0, 74868, '\p{Script: CUNEIFORM}', "");
Expect(1, 74868, '\p{^Script: CUNEIFORM}', "");
Expect(1, 74868, '\P{Script: CUNEIFORM}', "");
Expect(0, 74868, '\P{^Script: CUNEIFORM}', "");
Error('\p{Sc=/a/__XSUX}');
Error('\P{Sc=/a/__XSUX}');
Expect(1, 74867, '\p{Sc=xsux}', "");
Expect(0, 74867, '\p{^Sc=xsux}', "");
Expect(0, 74867, '\P{Sc=xsux}', "");
Expect(1, 74867, '\P{^Sc=xsux}', "");
Expect(0, 74868, '\p{Sc=xsux}', "");
Expect(1, 74868, '\p{^Sc=xsux}', "");
Expect(1, 74868, '\P{Sc=xsux}', "");
Expect(0, 74868, '\P{^Sc=xsux}', "");
Expect(1, 74867, '\p{Sc:_-Xsux}', "");
Expect(0, 74867, '\p{^Sc:_-Xsux}', "");
Expect(0, 74867, '\P{Sc:_-Xsux}', "");
Expect(1, 74867, '\P{^Sc:_-Xsux}', "");
Expect(0, 74868, '\p{Sc:_-Xsux}', "");
Expect(1, 74868, '\p{^Sc:_-Xsux}', "");
Expect(1, 74868, '\P{Sc:_-Xsux}', "");
Expect(0, 74868, '\P{^Sc:_-Xsux}', "");
Error('\p{Is_Script=-Cuneiform:=}');
Error('\P{Is_Script=-Cuneiform:=}');
Expect(1, 74867, '\p{Is_Script=cuneiform}', "");
Expect(0, 74867, '\p{^Is_Script=cuneiform}', "");
Expect(0, 74867, '\P{Is_Script=cuneiform}', "");
Expect(1, 74867, '\P{^Is_Script=cuneiform}', "");
Expect(0, 74868, '\p{Is_Script=cuneiform}', "");
Expect(1, 74868, '\p{^Is_Script=cuneiform}', "");
Expect(1, 74868, '\P{Is_Script=cuneiform}', "");
Expect(0, 74868, '\P{^Is_Script=cuneiform}', "");
Expect(1, 74867, '\p{Is_Script=_CUNEIFORM}', "");
Expect(0, 74867, '\p{^Is_Script=_CUNEIFORM}', "");
Expect(0, 74867, '\P{Is_Script=_CUNEIFORM}', "");
Expect(1, 74867, '\P{^Is_Script=_CUNEIFORM}', "");
Expect(0, 74868, '\p{Is_Script=_CUNEIFORM}', "");
Expect(1, 74868, '\p{^Is_Script=_CUNEIFORM}', "");
Expect(1, 74868, '\P{Is_Script=_CUNEIFORM}', "");
Expect(0, 74868, '\P{^Is_Script=_CUNEIFORM}', "");
Error('\p{Is_Sc: :=Xsux}');
Error('\P{Is_Sc: :=Xsux}');
Expect(1, 74867, '\p{Is_Sc: xsux}', "");
Expect(0, 74867, '\p{^Is_Sc: xsux}', "");
Expect(0, 74867, '\P{Is_Sc: xsux}', "");
Expect(1, 74867, '\P{^Is_Sc: xsux}', "");
Expect(0, 74868, '\p{Is_Sc: xsux}', "");
Expect(1, 74868, '\p{^Is_Sc: xsux}', "");
Expect(1, 74868, '\P{Is_Sc: xsux}', "");
Expect(0, 74868, '\P{^Is_Sc: xsux}', "");
Expect(1, 74867, '\p{Is_Sc=-_XSUX}', "");
Expect(0, 74867, '\p{^Is_Sc=-_XSUX}', "");
Expect(0, 74867, '\P{Is_Sc=-_XSUX}', "");
Expect(1, 74867, '\P{^Is_Sc=-_XSUX}', "");
Expect(0, 74868, '\p{Is_Sc=-_XSUX}', "");
Expect(1, 74868, '\p{^Is_Sc=-_XSUX}', "");
Expect(1, 74868, '\P{Is_Sc=-_XSUX}', "");
Expect(0, 74868, '\P{^Is_Sc=-_XSUX}', "");
Error('\p{Script=:=	INHERITED}');
Error('\P{Script=:=	INHERITED}');
Expect(1, 917999, '\p{Script=inherited}', "");
Expect(0, 917999, '\p{^Script=inherited}', "");
Expect(0, 917999, '\P{Script=inherited}', "");
Expect(1, 917999, '\P{^Script=inherited}', "");
Expect(0, 918000, '\p{Script=inherited}', "");
Expect(1, 918000, '\p{^Script=inherited}', "");
Expect(1, 918000, '\P{Script=inherited}', "");
Expect(0, 918000, '\P{^Script=inherited}', "");
Expect(1, 917999, '\p{Script=-_Inherited}', "");
Expect(0, 917999, '\p{^Script=-_Inherited}', "");
Expect(0, 917999, '\P{Script=-_Inherited}', "");
Expect(1, 917999, '\P{^Script=-_Inherited}', "");
Expect(0, 918000, '\p{Script=-_Inherited}', "");
Expect(1, 918000, '\p{^Script=-_Inherited}', "");
Expect(1, 918000, '\P{Script=-_Inherited}', "");
Expect(0, 918000, '\P{^Script=-_Inherited}', "");
Error('\p{Sc=_-zinh/a/}');
Error('\P{Sc=_-zinh/a/}');
Expect(1, 917999, '\p{Sc=zinh}', "");
Expect(0, 917999, '\p{^Sc=zinh}', "");
Expect(0, 917999, '\P{Sc=zinh}', "");
Expect(1, 917999, '\P{^Sc=zinh}', "");
Expect(0, 918000, '\p{Sc=zinh}', "");
Expect(1, 918000, '\p{^Sc=zinh}', "");
Expect(1, 918000, '\P{Sc=zinh}', "");
Expect(0, 918000, '\P{^Sc=zinh}', "");
Expect(1, 917999, '\p{Sc:   --Zinh}', "");
Expect(0, 917999, '\p{^Sc:   --Zinh}', "");
Expect(0, 917999, '\P{Sc:   --Zinh}', "");
Expect(1, 917999, '\P{^Sc:   --Zinh}', "");
Expect(0, 918000, '\p{Sc:   --Zinh}', "");
Expect(1, 918000, '\p{^Sc:   --Zinh}', "");
Expect(1, 918000, '\P{Sc:   --Zinh}', "");
Expect(0, 918000, '\P{^Sc:   --Zinh}', "");
Error('\p{Is_Script=:=QAAI}');
Error('\P{Is_Script=:=QAAI}');
Expect(1, 917999, '\p{Is_Script=qaai}', "");
Expect(0, 917999, '\p{^Is_Script=qaai}', "");
Expect(0, 917999, '\P{Is_Script=qaai}', "");
Expect(1, 917999, '\P{^Is_Script=qaai}', "");
Expect(0, 918000, '\p{Is_Script=qaai}', "");
Expect(1, 918000, '\p{^Is_Script=qaai}', "");
Expect(1, 918000, '\P{Is_Script=qaai}', "");
Expect(0, 918000, '\P{^Is_Script=qaai}', "");
Expect(1, 917999, '\p{Is_Script=-	Qaai}', "");
Expect(0, 917999, '\p{^Is_Script=-	Qaai}', "");
Expect(0, 917999, '\P{Is_Script=-	Qaai}', "");
Expect(1, 917999, '\P{^Is_Script=-	Qaai}', "");
Expect(0, 918000, '\p{Is_Script=-	Qaai}', "");
Expect(1, 918000, '\p{^Is_Script=-	Qaai}', "");
Expect(1, 918000, '\P{Is_Script=-	Qaai}', "");
Expect(0, 918000, '\P{^Is_Script=-	Qaai}', "");
Error('\p{Is_Sc=_:=INHERITED}');
Error('\P{Is_Sc=_:=INHERITED}');
Expect(1, 917999, '\p{Is_Sc=inherited}', "");
Expect(0, 917999, '\p{^Is_Sc=inherited}', "");
Expect(0, 917999, '\P{Is_Sc=inherited}', "");
Expect(1, 917999, '\P{^Is_Sc=inherited}', "");
Expect(0, 918000, '\p{Is_Sc=inherited}', "");
Expect(1, 918000, '\p{^Is_Sc=inherited}', "");
Expect(1, 918000, '\P{Is_Sc=inherited}', "");
Expect(0, 918000, '\P{^Is_Sc=inherited}', "");
Expect(1, 917999, '\p{Is_Sc=	inherited}', "");
Expect(0, 917999, '\p{^Is_Sc=	inherited}', "");
Expect(0, 917999, '\P{Is_Sc=	inherited}', "");
Expect(1, 917999, '\P{^Is_Sc=	inherited}', "");
Expect(0, 918000, '\p{Is_Sc=	inherited}', "");
Expect(1, 918000, '\p{^Is_Sc=	inherited}', "");
Expect(1, 918000, '\P{Is_Sc=	inherited}', "");
Expect(0, 918000, '\P{^Is_Sc=	inherited}', "");
Error('\p{Script= kaithi:=}');
Error('\P{Script= kaithi:=}');
Expect(1, 69825, '\p{Script=kaithi}', "");
Expect(0, 69825, '\p{^Script=kaithi}', "");
Expect(0, 69825, '\P{Script=kaithi}', "");
Expect(1, 69825, '\P{^Script=kaithi}', "");
Expect(0, 69826, '\p{Script=kaithi}', "");
Expect(1, 69826, '\p{^Script=kaithi}', "");
Expect(1, 69826, '\P{Script=kaithi}', "");
Expect(0, 69826, '\P{^Script=kaithi}', "");
Expect(1, 69825, '\p{Script=-kaithi}', "");
Expect(0, 69825, '\p{^Script=-kaithi}', "");
Expect(0, 69825, '\P{Script=-kaithi}', "");
Expect(1, 69825, '\P{^Script=-kaithi}', "");
Expect(0, 69826, '\p{Script=-kaithi}', "");
Expect(1, 69826, '\p{^Script=-kaithi}', "");
Expect(1, 69826, '\P{Script=-kaithi}', "");
Expect(0, 69826, '\P{^Script=-kaithi}', "");
Error('\p{Sc=-	Kthi:=}');
Error('\P{Sc=-	Kthi:=}');
Expect(1, 69825, '\p{Sc=kthi}', "");
Expect(0, 69825, '\p{^Sc=kthi}', "");
Expect(0, 69825, '\P{Sc=kthi}', "");
Expect(1, 69825, '\P{^Sc=kthi}', "");
Expect(0, 69826, '\p{Sc=kthi}', "");
Expect(1, 69826, '\p{^Sc=kthi}', "");
Expect(1, 69826, '\P{Sc=kthi}', "");
Expect(0, 69826, '\P{^Sc=kthi}', "");
Expect(1, 69825, '\p{Sc: _	Kthi}', "");
Expect(0, 69825, '\p{^Sc: _	Kthi}', "");
Expect(0, 69825, '\P{Sc: _	Kthi}', "");
Expect(1, 69825, '\P{^Sc: _	Kthi}', "");
Expect(0, 69826, '\p{Sc: _	Kthi}', "");
Expect(1, 69826, '\p{^Sc: _	Kthi}', "");
Expect(1, 69826, '\P{Sc: _	Kthi}', "");
Expect(0, 69826, '\P{^Sc: _	Kthi}', "");
Error('\p{Is_Script=Kaithi/a/}');
Error('\P{Is_Script=Kaithi/a/}');
Expect(1, 69825, '\p{Is_Script=kaithi}', "");
Expect(0, 69825, '\p{^Is_Script=kaithi}', "");
Expect(0, 69825, '\P{Is_Script=kaithi}', "");
Expect(1, 69825, '\P{^Is_Script=kaithi}', "");
Expect(0, 69826, '\p{Is_Script=kaithi}', "");
Expect(1, 69826, '\p{^Is_Script=kaithi}', "");
Expect(1, 69826, '\P{Is_Script=kaithi}', "");
Expect(0, 69826, '\P{^Is_Script=kaithi}', "");
Expect(1, 69825, '\p{Is_Script=- kaithi}', "");
Expect(0, 69825, '\p{^Is_Script=- kaithi}', "");
Expect(0, 69825, '\P{Is_Script=- kaithi}', "");
Expect(1, 69825, '\P{^Is_Script=- kaithi}', "");
Expect(0, 69826, '\p{Is_Script=- kaithi}', "");
Expect(1, 69826, '\p{^Is_Script=- kaithi}', "");
Expect(1, 69826, '\P{Is_Script=- kaithi}', "");
Expect(0, 69826, '\P{^Is_Script=- kaithi}', "");
Error('\p{Is_Sc=:=Kthi}');
Error('\P{Is_Sc=:=Kthi}');
Expect(1, 69825, '\p{Is_Sc=kthi}', "");
Expect(0, 69825, '\p{^Is_Sc=kthi}', "");
Expect(0, 69825, '\P{Is_Sc=kthi}', "");
Expect(1, 69825, '\P{^Is_Sc=kthi}', "");
Expect(0, 69826, '\p{Is_Sc=kthi}', "");
Expect(1, 69826, '\p{^Is_Sc=kthi}', "");
Expect(1, 69826, '\P{Is_Sc=kthi}', "");
Expect(0, 69826, '\P{^Is_Sc=kthi}', "");
Expect(1, 69825, '\p{Is_Sc=_	kthi}', "");
Expect(0, 69825, '\p{^Is_Sc=_	kthi}', "");
Expect(0, 69825, '\P{Is_Sc=_	kthi}', "");
Expect(1, 69825, '\P{^Is_Sc=_	kthi}', "");
Expect(0, 69826, '\p{Is_Sc=_	kthi}', "");
Expect(1, 69826, '\p{^Is_Sc=_	kthi}', "");
Expect(1, 69826, '\P{Is_Sc=_	kthi}', "");
Expect(0, 69826, '\P{^Is_Sc=_	kthi}', "");
Error('\p{Script=	/a/ARABIC}');
Error('\P{Script=	/a/ARABIC}');
Expect(1, 126705, '\p{Script=arabic}', "");
Expect(0, 126705, '\p{^Script=arabic}', "");
Expect(0, 126705, '\P{Script=arabic}', "");
Expect(1, 126705, '\P{^Script=arabic}', "");
Expect(0, 126706, '\p{Script=arabic}', "");
Expect(1, 126706, '\p{^Script=arabic}', "");
Expect(1, 126706, '\P{Script=arabic}', "");
Expect(0, 126706, '\P{^Script=arabic}', "");
Expect(1, 126705, '\p{Script=-_arabic}', "");
Expect(0, 126705, '\p{^Script=-_arabic}', "");
Expect(0, 126705, '\P{Script=-_arabic}', "");
Expect(1, 126705, '\P{^Script=-_arabic}', "");
Expect(0, 126706, '\p{Script=-_arabic}', "");
Expect(1, 126706, '\p{^Script=-_arabic}', "");
Expect(1, 126706, '\P{Script=-_arabic}', "");
Expect(0, 126706, '\P{^Script=-_arabic}', "");
Error('\p{Sc=-Arab/a/}');
Error('\P{Sc=-Arab/a/}');
Expect(1, 126705, '\p{Sc=arab}', "");
Expect(0, 126705, '\p{^Sc=arab}', "");
Expect(0, 126705, '\P{Sc=arab}', "");
Expect(1, 126705, '\P{^Sc=arab}', "");
Expect(0, 126706, '\p{Sc=arab}', "");
Expect(1, 126706, '\p{^Sc=arab}', "");
Expect(1, 126706, '\P{Sc=arab}', "");
Expect(0, 126706, '\P{^Sc=arab}', "");
Expect(1, 126705, '\p{Sc:    	Arab}', "");
Expect(0, 126705, '\p{^Sc:    	Arab}', "");
Expect(0, 126705, '\P{Sc:    	Arab}', "");
Expect(1, 126705, '\P{^Sc:    	Arab}', "");
Expect(0, 126706, '\p{Sc:    	Arab}', "");
Expect(1, 126706, '\p{^Sc:    	Arab}', "");
Expect(1, 126706, '\P{Sc:    	Arab}', "");
Expect(0, 126706, '\P{^Sc:    	Arab}', "");
Error('\p{Is_Script=/a/_ Arabic}');
Error('\P{Is_Script=/a/_ Arabic}');
Expect(1, 126705, '\p{Is_Script=arabic}', "");
Expect(0, 126705, '\p{^Is_Script=arabic}', "");
Expect(0, 126705, '\P{Is_Script=arabic}', "");
Expect(1, 126705, '\P{^Is_Script=arabic}', "");
Expect(0, 126706, '\p{Is_Script=arabic}', "");
Expect(1, 126706, '\p{^Is_Script=arabic}', "");
Expect(1, 126706, '\P{Is_Script=arabic}', "");
Expect(0, 126706, '\P{^Is_Script=arabic}', "");
Expect(1, 126705, '\p{Is_Script=	Arabic}', "");
Expect(0, 126705, '\p{^Is_Script=	Arabic}', "");
Expect(0, 126705, '\P{Is_Script=	Arabic}', "");
Expect(1, 126705, '\P{^Is_Script=	Arabic}', "");
Expect(0, 126706, '\p{Is_Script=	Arabic}', "");
Expect(1, 126706, '\p{^Is_Script=	Arabic}', "");
Expect(1, 126706, '\P{Is_Script=	Arabic}', "");
Expect(0, 126706, '\P{^Is_Script=	Arabic}', "");
Error('\p{Is_Sc:/a/	-Arab}');
Error('\P{Is_Sc:/a/	-Arab}');
Expect(1, 126705, '\p{Is_Sc=arab}', "");
Expect(0, 126705, '\p{^Is_Sc=arab}', "");
Expect(0, 126705, '\P{Is_Sc=arab}', "");
Expect(1, 126705, '\P{^Is_Sc=arab}', "");
Expect(0, 126706, '\p{Is_Sc=arab}', "");
Expect(1, 126706, '\p{^Is_Sc=arab}', "");
Expect(1, 126706, '\P{Is_Sc=arab}', "");
Expect(0, 126706, '\P{^Is_Sc=arab}', "");
Expect(1, 126705, '\p{Is_Sc=	Arab}', "");
Expect(0, 126705, '\p{^Is_Sc=	Arab}', "");
Expect(0, 126705, '\P{Is_Sc=	Arab}', "");
Expect(1, 126705, '\P{^Is_Sc=	Arab}', "");
Expect(0, 126706, '\p{Is_Sc=	Arab}', "");
Expect(1, 126706, '\p{^Is_Sc=	Arab}', "");
Expect(1, 126706, '\P{Is_Sc=	Arab}', "");
Expect(0, 126706, '\P{^Is_Sc=	Arab}', "");
Error('\p{Script= Shavian/a/}');
Error('\P{Script= Shavian/a/}');
Expect(1, 66687, '\p{Script=shavian}', "");
Expect(0, 66687, '\p{^Script=shavian}', "");
Expect(0, 66687, '\P{Script=shavian}', "");
Expect(1, 66687, '\P{^Script=shavian}', "");
Expect(0, 66688, '\p{Script=shavian}', "");
Expect(1, 66688, '\p{^Script=shavian}', "");
Expect(1, 66688, '\P{Script=shavian}', "");
Expect(0, 66688, '\P{^Script=shavian}', "");
Expect(1, 66687, '\p{Script=_shavian}', "");
Expect(0, 66687, '\p{^Script=_shavian}', "");
Expect(0, 66687, '\P{Script=_shavian}', "");
Expect(1, 66687, '\P{^Script=_shavian}', "");
Expect(0, 66688, '\p{Script=_shavian}', "");
Expect(1, 66688, '\p{^Script=_shavian}', "");
Expect(1, 66688, '\P{Script=_shavian}', "");
Expect(0, 66688, '\P{^Script=_shavian}', "");
Error('\p{Sc=:=	-Shaw}');
Error('\P{Sc=:=	-Shaw}');
Expect(1, 66687, '\p{Sc=shaw}', "");
Expect(0, 66687, '\p{^Sc=shaw}', "");
Expect(0, 66687, '\P{Sc=shaw}', "");
Expect(1, 66687, '\P{^Sc=shaw}', "");
Expect(0, 66688, '\p{Sc=shaw}', "");
Expect(1, 66688, '\p{^Sc=shaw}', "");
Expect(1, 66688, '\P{Sc=shaw}', "");
Expect(0, 66688, '\P{^Sc=shaw}', "");
Expect(1, 66687, '\p{Sc= -SHAW}', "");
Expect(0, 66687, '\p{^Sc= -SHAW}', "");
Expect(0, 66687, '\P{Sc= -SHAW}', "");
Expect(1, 66687, '\P{^Sc= -SHAW}', "");
Expect(0, 66688, '\p{Sc= -SHAW}', "");
Expect(1, 66688, '\p{^Sc= -SHAW}', "");
Expect(1, 66688, '\P{Sc= -SHAW}', "");
Expect(0, 66688, '\P{^Sc= -SHAW}', "");
Error('\p{Is_Script=/a/  SHAVIAN}');
Error('\P{Is_Script=/a/  SHAVIAN}');
Expect(1, 66687, '\p{Is_Script:shavian}', "");
Expect(0, 66687, '\p{^Is_Script:shavian}', "");
Expect(0, 66687, '\P{Is_Script:shavian}', "");
Expect(1, 66687, '\P{^Is_Script:shavian}', "");
Expect(0, 66688, '\p{Is_Script:shavian}', "");
Expect(1, 66688, '\p{^Is_Script:shavian}', "");
Expect(1, 66688, '\P{Is_Script:shavian}', "");
Expect(0, 66688, '\P{^Is_Script:shavian}', "");
Expect(1, 66687, '\p{Is_Script=	-Shavian}', "");
Expect(0, 66687, '\p{^Is_Script=	-Shavian}', "");
Expect(0, 66687, '\P{Is_Script=	-Shavian}', "");
Expect(1, 66687, '\P{^Is_Script=	-Shavian}', "");
Expect(0, 66688, '\p{Is_Script=	-Shavian}', "");
Expect(1, 66688, '\p{^Is_Script=	-Shavian}', "");
Expect(1, 66688, '\P{Is_Script=	-Shavian}', "");
Expect(0, 66688, '\P{^Is_Script=	-Shavian}', "");
Error('\p{Is_Sc=/a/Shaw}');
Error('\P{Is_Sc=/a/Shaw}');
Expect(1, 66687, '\p{Is_Sc:   shaw}', "");
Expect(0, 66687, '\p{^Is_Sc:   shaw}', "");
Expect(0, 66687, '\P{Is_Sc:   shaw}', "");
Expect(1, 66687, '\P{^Is_Sc:   shaw}', "");
Expect(0, 66688, '\p{Is_Sc:   shaw}', "");
Expect(1, 66688, '\p{^Is_Sc:   shaw}', "");
Expect(1, 66688, '\P{Is_Sc:   shaw}', "");
Expect(0, 66688, '\P{^Is_Sc:   shaw}', "");
Expect(1, 66687, '\p{Is_Sc:   	Shaw}', "");
Expect(0, 66687, '\p{^Is_Sc:   	Shaw}', "");
Expect(0, 66687, '\P{Is_Sc:   	Shaw}', "");
Expect(1, 66687, '\P{^Is_Sc:   	Shaw}', "");
Expect(0, 66688, '\p{Is_Sc:   	Shaw}', "");
Expect(1, 66688, '\p{^Is_Sc:   	Shaw}', "");
Expect(1, 66688, '\P{Is_Sc:   	Shaw}', "");
Expect(0, 66688, '\P{^Is_Sc:   	Shaw}', "");
Error('\p{Script=	_Canadian_aboriginal/a/}');
Error('\P{Script=	_Canadian_aboriginal/a/}');
Expect(1, 6389, '\p{Script=canadianaboriginal}', "");
Expect(0, 6389, '\p{^Script=canadianaboriginal}', "");
Expect(0, 6389, '\P{Script=canadianaboriginal}', "");
Expect(1, 6389, '\P{^Script=canadianaboriginal}', "");
Expect(0, 6390, '\p{Script=canadianaboriginal}', "");
Expect(1, 6390, '\p{^Script=canadianaboriginal}', "");
Expect(1, 6390, '\P{Script=canadianaboriginal}', "");
Expect(0, 6390, '\P{^Script=canadianaboriginal}', "");
Expect(1, 6389, '\p{Script=  Canadian_Aboriginal}', "");
Expect(0, 6389, '\p{^Script=  Canadian_Aboriginal}', "");
Expect(0, 6389, '\P{Script=  Canadian_Aboriginal}', "");
Expect(1, 6389, '\P{^Script=  Canadian_Aboriginal}', "");
Expect(0, 6390, '\p{Script=  Canadian_Aboriginal}', "");
Expect(1, 6390, '\p{^Script=  Canadian_Aboriginal}', "");
Expect(1, 6390, '\P{Script=  Canadian_Aboriginal}', "");
Expect(0, 6390, '\P{^Script=  Canadian_Aboriginal}', "");
Error('\p{Sc=	:=Cans}');
Error('\P{Sc=	:=Cans}');
Expect(1, 6389, '\p{Sc=cans}', "");
Expect(0, 6389, '\p{^Sc=cans}', "");
Expect(0, 6389, '\P{Sc=cans}', "");
Expect(1, 6389, '\P{^Sc=cans}', "");
Expect(0, 6390, '\p{Sc=cans}', "");
Expect(1, 6390, '\p{^Sc=cans}', "");
Expect(1, 6390, '\P{Sc=cans}', "");
Expect(0, 6390, '\P{^Sc=cans}', "");
Error('\p{Is_Script=-:=CANADIAN_Aboriginal}');
Error('\P{Is_Script=-:=CANADIAN_Aboriginal}');
Expect(1, 6389, '\p{Is_Script=canadianaboriginal}', "");
Expect(0, 6389, '\p{^Is_Script=canadianaboriginal}', "");
Expect(0, 6389, '\P{Is_Script=canadianaboriginal}', "");
Expect(1, 6389, '\P{^Is_Script=canadianaboriginal}', "");
Expect(0, 6390, '\p{Is_Script=canadianaboriginal}', "");
Expect(1, 6390, '\p{^Is_Script=canadianaboriginal}', "");
Expect(1, 6390, '\P{Is_Script=canadianaboriginal}', "");
Expect(0, 6390, '\P{^Is_Script=canadianaboriginal}', "");
Expect(1, 6389, '\p{Is_Script=		Canadian_aboriginal}', "");
Expect(0, 6389, '\p{^Is_Script=		Canadian_aboriginal}', "");
Expect(0, 6389, '\P{Is_Script=		Canadian_aboriginal}', "");
Expect(1, 6389, '\P{^Is_Script=		Canadian_aboriginal}', "");
Expect(0, 6390, '\p{Is_Script=		Canadian_aboriginal}', "");
Expect(1, 6390, '\p{^Is_Script=		Canadian_aboriginal}', "");
Expect(1, 6390, '\P{Is_Script=		Canadian_aboriginal}', "");
Expect(0, 6390, '\P{^Is_Script=		Canadian_aboriginal}', "");
Error('\p{Is_Sc=_-cans/a/}');
Error('\P{Is_Sc=_-cans/a/}');
Expect(1, 6389, '\p{Is_Sc=cans}', "");
Expect(0, 6389, '\p{^Is_Sc=cans}', "");
Expect(0, 6389, '\P{Is_Sc=cans}', "");
Expect(1, 6389, '\P{^Is_Sc=cans}', "");
Expect(0, 6390, '\p{Is_Sc=cans}', "");
Expect(1, 6390, '\p{^Is_Sc=cans}', "");
Expect(1, 6390, '\P{Is_Sc=cans}', "");
Expect(0, 6390, '\P{^Is_Sc=cans}', "");
Expect(1, 6389, '\p{Is_Sc=--cans}', "");
Expect(0, 6389, '\p{^Is_Sc=--cans}', "");
Expect(0, 6389, '\P{Is_Sc=--cans}', "");
Expect(1, 6389, '\P{^Is_Sc=--cans}', "");
Expect(0, 6390, '\p{Is_Sc=--cans}', "");
Expect(1, 6390, '\p{^Is_Sc=--cans}', "");
Expect(1, 6390, '\P{Is_Sc=--cans}', "");
Expect(0, 6390, '\P{^Is_Sc=--cans}', "");
Error('\p{Script=:= _old_SOUTH_Arabian}');
Error('\P{Script=:= _old_SOUTH_Arabian}');
Expect(1, 68223, '\p{Script=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Script=oldsoutharabian}', "");
Expect(0, 68223, '\P{Script=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Script=oldsoutharabian}', "");
Expect(0, 68224, '\p{Script=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Script=oldsoutharabian}', "");
Expect(1, 68224, '\P{Script=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Script=oldsoutharabian}', "");
Expect(1, 68223, '\p{Script= -Old_south_Arabian}', "");
Expect(0, 68223, '\p{^Script= -Old_south_Arabian}', "");
Expect(0, 68223, '\P{Script= -Old_south_Arabian}', "");
Expect(1, 68223, '\P{^Script= -Old_south_Arabian}', "");
Expect(0, 68224, '\p{Script= -Old_south_Arabian}', "");
Expect(1, 68224, '\p{^Script= -Old_south_Arabian}', "");
Expect(1, 68224, '\P{Script= -Old_south_Arabian}', "");
Expect(0, 68224, '\P{^Script= -Old_south_Arabian}', "");
Error('\p{Sc=/a/	 Sarb}');
Error('\P{Sc=/a/	 Sarb}');
Expect(1, 68223, '\p{Sc=sarb}', "");
Expect(0, 68223, '\p{^Sc=sarb}', "");
Expect(0, 68223, '\P{Sc=sarb}', "");
Expect(1, 68223, '\P{^Sc=sarb}', "");
Expect(0, 68224, '\p{Sc=sarb}', "");
Expect(1, 68224, '\p{^Sc=sarb}', "");
Expect(1, 68224, '\P{Sc=sarb}', "");
Expect(0, 68224, '\P{^Sc=sarb}', "");
Expect(1, 68223, '\p{Sc=	_sarb}', "");
Expect(0, 68223, '\p{^Sc=	_sarb}', "");
Expect(0, 68223, '\P{Sc=	_sarb}', "");
Expect(1, 68223, '\P{^Sc=	_sarb}', "");
Expect(0, 68224, '\p{Sc=	_sarb}', "");
Expect(1, 68224, '\p{^Sc=	_sarb}', "");
Expect(1, 68224, '\P{Sc=	_sarb}', "");
Expect(0, 68224, '\P{^Sc=	_sarb}', "");
Error('\p{Is_Script=--Old_South_ARABIAN/a/}');
Error('\P{Is_Script=--Old_South_ARABIAN/a/}');
Expect(1, 68223, '\p{Is_Script=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Is_Script=oldsoutharabian}', "");
Expect(0, 68223, '\P{Is_Script=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Is_Script=oldsoutharabian}', "");
Expect(0, 68224, '\p{Is_Script=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Is_Script=oldsoutharabian}', "");
Expect(1, 68224, '\P{Is_Script=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Is_Script=oldsoutharabian}', "");
Expect(1, 68223, '\p{Is_Script= OLD_south_arabian}', "");
Expect(0, 68223, '\p{^Is_Script= OLD_south_arabian}', "");
Expect(0, 68223, '\P{Is_Script= OLD_south_arabian}', "");
Expect(1, 68223, '\P{^Is_Script= OLD_south_arabian}', "");
Expect(0, 68224, '\p{Is_Script= OLD_south_arabian}', "");
Expect(1, 68224, '\p{^Is_Script= OLD_south_arabian}', "");
Expect(1, 68224, '\P{Is_Script= OLD_south_arabian}', "");
Expect(0, 68224, '\P{^Is_Script= OLD_south_arabian}', "");
Error('\p{Is_Sc=:=-_sarb}');
Error('\P{Is_Sc=:=-_sarb}');
Expect(1, 68223, '\p{Is_Sc=sarb}', "");
Expect(0, 68223, '\p{^Is_Sc=sarb}', "");
Expect(0, 68223, '\P{Is_Sc=sarb}', "");
Expect(1, 68223, '\P{^Is_Sc=sarb}', "");
Expect(0, 68224, '\p{Is_Sc=sarb}', "");
Expect(1, 68224, '\p{^Is_Sc=sarb}', "");
Expect(1, 68224, '\P{Is_Sc=sarb}', "");
Expect(0, 68224, '\P{^Is_Sc=sarb}', "");
Expect(1, 68223, '\p{Is_Sc=_sarb}', "");
Expect(0, 68223, '\p{^Is_Sc=_sarb}', "");
Expect(0, 68223, '\P{Is_Sc=_sarb}', "");
Expect(1, 68223, '\P{^Is_Sc=_sarb}', "");
Expect(0, 68224, '\p{Is_Sc=_sarb}', "");
Expect(1, 68224, '\p{^Is_Sc=_sarb}', "");
Expect(1, 68224, '\P{Is_Sc=_sarb}', "");
Expect(0, 68224, '\P{^Is_Sc=_sarb}', "");
Error('\p{Script=	-Lydian/a/}');
Error('\P{Script=	-Lydian/a/}');
Expect(1, 67903, '\p{Script=lydian}', "");
Expect(0, 67903, '\p{^Script=lydian}', "");
Expect(0, 67903, '\P{Script=lydian}', "");
Expect(1, 67903, '\P{^Script=lydian}', "");
Expect(0, 67904, '\p{Script=lydian}', "");
Expect(1, 67904, '\p{^Script=lydian}', "");
Expect(1, 67904, '\P{Script=lydian}', "");
Expect(0, 67904, '\P{^Script=lydian}', "");
Expect(1, 67903, '\p{Script=	 lydian}', "");
Expect(0, 67903, '\p{^Script=	 lydian}', "");
Expect(0, 67903, '\P{Script=	 lydian}', "");
Expect(1, 67903, '\P{^Script=	 lydian}', "");
Expect(0, 67904, '\p{Script=	 lydian}', "");
Expect(1, 67904, '\p{^Script=	 lydian}', "");
Expect(1, 67904, '\P{Script=	 lydian}', "");
Expect(0, 67904, '\P{^Script=	 lydian}', "");
Error('\p{Sc=:=  LYDI}');
Error('\P{Sc=:=  LYDI}');
Expect(1, 67903, '\p{Sc=lydi}', "");
Expect(0, 67903, '\p{^Sc=lydi}', "");
Expect(0, 67903, '\P{Sc=lydi}', "");
Expect(1, 67903, '\P{^Sc=lydi}', "");
Expect(0, 67904, '\p{Sc=lydi}', "");
Expect(1, 67904, '\p{^Sc=lydi}', "");
Expect(1, 67904, '\P{Sc=lydi}', "");
Expect(0, 67904, '\P{^Sc=lydi}', "");
Expect(1, 67903, '\p{Sc=-lydi}', "");
Expect(0, 67903, '\p{^Sc=-lydi}', "");
Expect(0, 67903, '\P{Sc=-lydi}', "");
Expect(1, 67903, '\P{^Sc=-lydi}', "");
Expect(0, 67904, '\p{Sc=-lydi}', "");
Expect(1, 67904, '\p{^Sc=-lydi}', "");
Expect(1, 67904, '\P{Sc=-lydi}', "");
Expect(0, 67904, '\P{^Sc=-lydi}', "");
Error('\p{Is_Script= -LYDIAN/a/}');
Error('\P{Is_Script= -LYDIAN/a/}');
Expect(1, 67903, '\p{Is_Script=lydian}', "");
Expect(0, 67903, '\p{^Is_Script=lydian}', "");
Expect(0, 67903, '\P{Is_Script=lydian}', "");
Expect(1, 67903, '\P{^Is_Script=lydian}', "");
Expect(0, 67904, '\p{Is_Script=lydian}', "");
Expect(1, 67904, '\p{^Is_Script=lydian}', "");
Expect(1, 67904, '\P{Is_Script=lydian}', "");
Expect(0, 67904, '\P{^Is_Script=lydian}', "");
Expect(1, 67903, '\p{Is_Script= 	Lydian}', "");
Expect(0, 67903, '\p{^Is_Script= 	Lydian}', "");
Expect(0, 67903, '\P{Is_Script= 	Lydian}', "");
Expect(1, 67903, '\P{^Is_Script= 	Lydian}', "");
Expect(0, 67904, '\p{Is_Script= 	Lydian}', "");
Expect(1, 67904, '\p{^Is_Script= 	Lydian}', "");
Expect(1, 67904, '\P{Is_Script= 	Lydian}', "");
Expect(0, 67904, '\P{^Is_Script= 	Lydian}', "");
Error('\p{Is_Sc=	/a/Lydi}');
Error('\P{Is_Sc=	/a/Lydi}');
Expect(1, 67903, '\p{Is_Sc=lydi}', "");
Expect(0, 67903, '\p{^Is_Sc=lydi}', "");
Expect(0, 67903, '\P{Is_Sc=lydi}', "");
Expect(1, 67903, '\P{^Is_Sc=lydi}', "");
Expect(0, 67904, '\p{Is_Sc=lydi}', "");
Expect(1, 67904, '\p{^Is_Sc=lydi}', "");
Expect(1, 67904, '\P{Is_Sc=lydi}', "");
Expect(0, 67904, '\P{^Is_Sc=lydi}', "");
Error('\p{Script=-	Mongolian/a/}');
Error('\P{Script=-	Mongolian/a/}');
Expect(1, 6314, '\p{Script=mongolian}', "");
Expect(0, 6314, '\p{^Script=mongolian}', "");
Expect(0, 6314, '\P{Script=mongolian}', "");
Expect(1, 6314, '\P{^Script=mongolian}', "");
Expect(0, 6315, '\p{Script=mongolian}', "");
Expect(1, 6315, '\p{^Script=mongolian}', "");
Expect(1, 6315, '\P{Script=mongolian}', "");
Expect(0, 6315, '\P{^Script=mongolian}', "");
Expect(1, 6314, '\p{Script=_mongolian}', "");
Expect(0, 6314, '\p{^Script=_mongolian}', "");
Expect(0, 6314, '\P{Script=_mongolian}', "");
Expect(1, 6314, '\P{^Script=_mongolian}', "");
Expect(0, 6315, '\p{Script=_mongolian}', "");
Expect(1, 6315, '\p{^Script=_mongolian}', "");
Expect(1, 6315, '\P{Script=_mongolian}', "");
Expect(0, 6315, '\P{^Script=_mongolian}', "");
Error('\p{Sc=		Mong:=}');
Error('\P{Sc=		Mong:=}');
Expect(1, 6314, '\p{Sc=mong}', "");
Expect(0, 6314, '\p{^Sc=mong}', "");
Expect(0, 6314, '\P{Sc=mong}', "");
Expect(1, 6314, '\P{^Sc=mong}', "");
Expect(0, 6315, '\p{Sc=mong}', "");
Expect(1, 6315, '\p{^Sc=mong}', "");
Expect(1, 6315, '\P{Sc=mong}', "");
Expect(0, 6315, '\P{^Sc=mong}', "");
Expect(1, 6314, '\p{Sc=_Mong}', "");
Expect(0, 6314, '\p{^Sc=_Mong}', "");
Expect(0, 6314, '\P{Sc=_Mong}', "");
Expect(1, 6314, '\P{^Sc=_Mong}', "");
Expect(0, 6315, '\p{Sc=_Mong}', "");
Expect(1, 6315, '\p{^Sc=_Mong}', "");
Expect(1, 6315, '\P{Sc=_Mong}', "");
Expect(0, 6315, '\P{^Sc=_Mong}', "");
Error('\p{Is_Script=/a/		Mongolian}');
Error('\P{Is_Script=/a/		Mongolian}');
Expect(1, 6314, '\p{Is_Script=mongolian}', "");
Expect(0, 6314, '\p{^Is_Script=mongolian}', "");
Expect(0, 6314, '\P{Is_Script=mongolian}', "");
Expect(1, 6314, '\P{^Is_Script=mongolian}', "");
Expect(0, 6315, '\p{Is_Script=mongolian}', "");
Expect(1, 6315, '\p{^Is_Script=mongolian}', "");
Expect(1, 6315, '\P{Is_Script=mongolian}', "");
Expect(0, 6315, '\P{^Is_Script=mongolian}', "");
Expect(1, 6314, '\p{Is_Script=-Mongolian}', "");
Expect(0, 6314, '\p{^Is_Script=-Mongolian}', "");
Expect(0, 6314, '\P{Is_Script=-Mongolian}', "");
Expect(1, 6314, '\P{^Is_Script=-Mongolian}', "");
Expect(0, 6315, '\p{Is_Script=-Mongolian}', "");
Expect(1, 6315, '\p{^Is_Script=-Mongolian}', "");
Expect(1, 6315, '\P{Is_Script=-Mongolian}', "");
Expect(0, 6315, '\P{^Is_Script=-Mongolian}', "");
Error('\p{Is_Sc=-:=Mong}');
Error('\P{Is_Sc=-:=Mong}');
Expect(1, 6314, '\p{Is_Sc=mong}', "");
Expect(0, 6314, '\p{^Is_Sc=mong}', "");
Expect(0, 6314, '\P{Is_Sc=mong}', "");
Expect(1, 6314, '\P{^Is_Sc=mong}', "");
Expect(0, 6315, '\p{Is_Sc=mong}', "");
Expect(1, 6315, '\p{^Is_Sc=mong}', "");
Expect(1, 6315, '\P{Is_Sc=mong}', "");
Expect(0, 6315, '\P{^Is_Sc=mong}', "");
Expect(1, 6314, '\p{Is_Sc=_ mong}', "");
Expect(0, 6314, '\p{^Is_Sc=_ mong}', "");
Expect(0, 6314, '\P{Is_Sc=_ mong}', "");
Expect(1, 6314, '\P{^Is_Sc=_ mong}', "");
Expect(0, 6315, '\p{Is_Sc=_ mong}', "");
Expect(1, 6315, '\p{^Is_Sc=_ mong}', "");
Expect(1, 6315, '\P{Is_Sc=_ mong}', "");
Expect(0, 6315, '\P{^Is_Sc=_ mong}', "");
Error('\p{Script=	_lao/a/}');
Error('\P{Script=	_lao/a/}');
Expect(1, 3807, '\p{Script=lao}', "");
Expect(0, 3807, '\p{^Script=lao}', "");
Expect(0, 3807, '\P{Script=lao}', "");
Expect(1, 3807, '\P{^Script=lao}', "");
Expect(0, 3808, '\p{Script=lao}', "");
Expect(1, 3808, '\p{^Script=lao}', "");
Expect(1, 3808, '\P{Script=lao}', "");
Expect(0, 3808, '\P{^Script=lao}', "");
Expect(1, 3807, '\p{Script= LAO}', "");
Expect(0, 3807, '\p{^Script= LAO}', "");
Expect(0, 3807, '\P{Script= LAO}', "");
Expect(1, 3807, '\P{^Script= LAO}', "");
Expect(0, 3808, '\p{Script= LAO}', "");
Expect(1, 3808, '\p{^Script= LAO}', "");
Expect(1, 3808, '\P{Script= LAO}', "");
Expect(0, 3808, '\P{^Script= LAO}', "");
Error('\p{Sc=	_laoo/a/}');
Error('\P{Sc=	_laoo/a/}');
Expect(1, 3807, '\p{Sc=laoo}', "");
Expect(0, 3807, '\p{^Sc=laoo}', "");
Expect(0, 3807, '\P{Sc=laoo}', "");
Expect(1, 3807, '\P{^Sc=laoo}', "");
Expect(0, 3808, '\p{Sc=laoo}', "");
Expect(1, 3808, '\p{^Sc=laoo}', "");
Expect(1, 3808, '\P{Sc=laoo}', "");
Expect(0, 3808, '\P{^Sc=laoo}', "");
Expect(1, 3807, '\p{Sc=	Laoo}', "");
Expect(0, 3807, '\p{^Sc=	Laoo}', "");
Expect(0, 3807, '\P{Sc=	Laoo}', "");
Expect(1, 3807, '\P{^Sc=	Laoo}', "");
Expect(0, 3808, '\p{Sc=	Laoo}', "");
Expect(1, 3808, '\p{^Sc=	Laoo}', "");
Expect(1, 3808, '\P{Sc=	Laoo}', "");
Expect(0, 3808, '\P{^Sc=	Laoo}', "");
Error('\p{Is_Script=:=		lao}');
Error('\P{Is_Script=:=		lao}');
Expect(1, 3807, '\p{Is_Script=lao}', "");
Expect(0, 3807, '\p{^Is_Script=lao}', "");
Expect(0, 3807, '\P{Is_Script=lao}', "");
Expect(1, 3807, '\P{^Is_Script=lao}', "");
Expect(0, 3808, '\p{Is_Script=lao}', "");
Expect(1, 3808, '\p{^Is_Script=lao}', "");
Expect(1, 3808, '\P{Is_Script=lao}', "");
Expect(0, 3808, '\P{^Is_Script=lao}', "");
Expect(1, 3807, '\p{Is_Script= -Lao}', "");
Expect(0, 3807, '\p{^Is_Script= -Lao}', "");
Expect(0, 3807, '\P{Is_Script= -Lao}', "");
Expect(1, 3807, '\P{^Is_Script= -Lao}', "");
Expect(0, 3808, '\p{Is_Script= -Lao}', "");
Expect(1, 3808, '\p{^Is_Script= -Lao}', "");
Expect(1, 3808, '\P{Is_Script= -Lao}', "");
Expect(0, 3808, '\P{^Is_Script= -Lao}', "");
Error('\p{Is_Sc:/a/__LAOO}');
Error('\P{Is_Sc:/a/__LAOO}');
Expect(1, 3807, '\p{Is_Sc=laoo}', "");
Expect(0, 3807, '\p{^Is_Sc=laoo}', "");
Expect(0, 3807, '\P{Is_Sc=laoo}', "");
Expect(1, 3807, '\P{^Is_Sc=laoo}', "");
Expect(0, 3808, '\p{Is_Sc=laoo}', "");
Expect(1, 3808, '\p{^Is_Sc=laoo}', "");
Expect(1, 3808, '\P{Is_Sc=laoo}', "");
Expect(0, 3808, '\P{^Is_Sc=laoo}', "");
Expect(1, 3807, '\p{Is_Sc=-_laoo}', "");
Expect(0, 3807, '\p{^Is_Sc=-_laoo}', "");
Expect(0, 3807, '\P{Is_Sc=-_laoo}', "");
Expect(1, 3807, '\P{^Is_Sc=-_laoo}', "");
Expect(0, 3808, '\p{Is_Sc=-_laoo}', "");
Expect(1, 3808, '\p{^Is_Sc=-_laoo}', "");
Expect(1, 3808, '\P{Is_Sc=-_laoo}', "");
Expect(0, 3808, '\P{^Is_Sc=-_laoo}', "");
Error('\p{Script=:=-	Cyrillic}');
Error('\P{Script=:=-	Cyrillic}');
Expect(1, 42655, '\p{Script=cyrillic}', "");
Expect(0, 42655, '\p{^Script=cyrillic}', "");
Expect(0, 42655, '\P{Script=cyrillic}', "");
Expect(1, 42655, '\P{^Script=cyrillic}', "");
Expect(0, 42656, '\p{Script=cyrillic}', "");
Expect(1, 42656, '\p{^Script=cyrillic}', "");
Expect(1, 42656, '\P{Script=cyrillic}', "");
Expect(0, 42656, '\P{^Script=cyrillic}', "");
Expect(1, 42655, '\p{Script=	-CYRILLIC}', "");
Expect(0, 42655, '\p{^Script=	-CYRILLIC}', "");
Expect(0, 42655, '\P{Script=	-CYRILLIC}', "");
Expect(1, 42655, '\P{^Script=	-CYRILLIC}', "");
Expect(0, 42656, '\p{Script=	-CYRILLIC}', "");
Expect(1, 42656, '\p{^Script=	-CYRILLIC}', "");
Expect(1, 42656, '\P{Script=	-CYRILLIC}', "");
Expect(0, 42656, '\P{^Script=	-CYRILLIC}', "");
Error('\p{Sc=	-cyrl/a/}');
Error('\P{Sc=	-cyrl/a/}');
Expect(1, 42655, '\p{Sc=cyrl}', "");
Expect(0, 42655, '\p{^Sc=cyrl}', "");
Expect(0, 42655, '\P{Sc=cyrl}', "");
Expect(1, 42655, '\P{^Sc=cyrl}', "");
Expect(0, 42656, '\p{Sc=cyrl}', "");
Expect(1, 42656, '\p{^Sc=cyrl}', "");
Expect(1, 42656, '\P{Sc=cyrl}', "");
Expect(0, 42656, '\P{^Sc=cyrl}', "");
Expect(1, 42655, '\p{Sc: - Cyrl}', "");
Expect(0, 42655, '\p{^Sc: - Cyrl}', "");
Expect(0, 42655, '\P{Sc: - Cyrl}', "");
Expect(1, 42655, '\P{^Sc: - Cyrl}', "");
Expect(0, 42656, '\p{Sc: - Cyrl}', "");
Expect(1, 42656, '\p{^Sc: - Cyrl}', "");
Expect(1, 42656, '\P{Sc: - Cyrl}', "");
Expect(0, 42656, '\P{^Sc: - Cyrl}', "");
Error('\p{Is_Script=:=- Cyrillic}');
Error('\P{Is_Script=:=- Cyrillic}');
Expect(1, 42655, '\p{Is_Script=cyrillic}', "");
Expect(0, 42655, '\p{^Is_Script=cyrillic}', "");
Expect(0, 42655, '\P{Is_Script=cyrillic}', "");
Expect(1, 42655, '\P{^Is_Script=cyrillic}', "");
Expect(0, 42656, '\p{Is_Script=cyrillic}', "");
Expect(1, 42656, '\p{^Is_Script=cyrillic}', "");
Expect(1, 42656, '\P{Is_Script=cyrillic}', "");
Expect(0, 42656, '\P{^Is_Script=cyrillic}', "");
Expect(1, 42655, '\p{Is_Script=_cyrillic}', "");
Expect(0, 42655, '\p{^Is_Script=_cyrillic}', "");
Expect(0, 42655, '\P{Is_Script=_cyrillic}', "");
Expect(1, 42655, '\P{^Is_Script=_cyrillic}', "");
Expect(0, 42656, '\p{Is_Script=_cyrillic}', "");
Expect(1, 42656, '\p{^Is_Script=_cyrillic}', "");
Expect(1, 42656, '\P{Is_Script=_cyrillic}', "");
Expect(0, 42656, '\P{^Is_Script=_cyrillic}', "");
Error('\p{Is_Sc=	 Cyrl/a/}');
Error('\P{Is_Sc=	 Cyrl/a/}');
Expect(1, 42655, '\p{Is_Sc=cyrl}', "");
Expect(0, 42655, '\p{^Is_Sc=cyrl}', "");
Expect(0, 42655, '\P{Is_Sc=cyrl}', "");
Expect(1, 42655, '\P{^Is_Sc=cyrl}', "");
Expect(0, 42656, '\p{Is_Sc=cyrl}', "");
Expect(1, 42656, '\p{^Is_Sc=cyrl}', "");
Expect(1, 42656, '\P{Is_Sc=cyrl}', "");
Expect(0, 42656, '\P{^Is_Sc=cyrl}', "");
Expect(1, 42655, '\p{Is_Sc=-Cyrl}', "");
Expect(0, 42655, '\p{^Is_Sc=-Cyrl}', "");
Expect(0, 42655, '\P{Is_Sc=-Cyrl}', "");
Expect(1, 42655, '\P{^Is_Sc=-Cyrl}', "");
Expect(0, 42656, '\p{Is_Sc=-Cyrl}', "");
Expect(1, 42656, '\p{^Is_Sc=-Cyrl}', "");
Expect(1, 42656, '\P{Is_Sc=-Cyrl}', "");
Expect(0, 42656, '\P{^Is_Sc=-Cyrl}', "");
Error('\p{Script=HIRAGANA:=}');
Error('\P{Script=HIRAGANA:=}');
Expect(1, 127488, '\p{Script=hiragana}', "");
Expect(0, 127488, '\p{^Script=hiragana}', "");
Expect(0, 127488, '\P{Script=hiragana}', "");
Expect(1, 127488, '\P{^Script=hiragana}', "");
Expect(0, 127489, '\p{Script=hiragana}', "");
Expect(1, 127489, '\p{^Script=hiragana}', "");
Expect(1, 127489, '\P{Script=hiragana}', "");
Expect(0, 127489, '\P{^Script=hiragana}', "");
Expect(1, 127488, '\p{Script= HIRAGANA}', "");
Expect(0, 127488, '\p{^Script= HIRAGANA}', "");
Expect(0, 127488, '\P{Script= HIRAGANA}', "");
Expect(1, 127488, '\P{^Script= HIRAGANA}', "");
Expect(0, 127489, '\p{Script= HIRAGANA}', "");
Expect(1, 127489, '\p{^Script= HIRAGANA}', "");
Expect(1, 127489, '\P{Script= HIRAGANA}', "");
Expect(0, 127489, '\P{^Script= HIRAGANA}', "");
Error('\p{Sc=:=	-Hira}');
Error('\P{Sc=:=	-Hira}');
Expect(1, 127488, '\p{Sc=hira}', "");
Expect(0, 127488, '\p{^Sc=hira}', "");
Expect(0, 127488, '\P{Sc=hira}', "");
Expect(1, 127488, '\P{^Sc=hira}', "");
Expect(0, 127489, '\p{Sc=hira}', "");
Expect(1, 127489, '\p{^Sc=hira}', "");
Expect(1, 127489, '\P{Sc=hira}', "");
Expect(0, 127489, '\P{^Sc=hira}', "");
Expect(1, 127488, '\p{Sc=-	Hira}', "");
Expect(0, 127488, '\p{^Sc=-	Hira}', "");
Expect(0, 127488, '\P{Sc=-	Hira}', "");
Expect(1, 127488, '\P{^Sc=-	Hira}', "");
Expect(0, 127489, '\p{Sc=-	Hira}', "");
Expect(1, 127489, '\p{^Sc=-	Hira}', "");
Expect(1, 127489, '\P{Sc=-	Hira}', "");
Expect(0, 127489, '\P{^Sc=-	Hira}', "");
Error('\p{Is_Script=/a/--HIRAGANA}');
Error('\P{Is_Script=/a/--HIRAGANA}');
Expect(1, 127488, '\p{Is_Script=hiragana}', "");
Expect(0, 127488, '\p{^Is_Script=hiragana}', "");
Expect(0, 127488, '\P{Is_Script=hiragana}', "");
Expect(1, 127488, '\P{^Is_Script=hiragana}', "");
Expect(0, 127489, '\p{Is_Script=hiragana}', "");
Expect(1, 127489, '\p{^Is_Script=hiragana}', "");
Expect(1, 127489, '\P{Is_Script=hiragana}', "");
Expect(0, 127489, '\P{^Is_Script=hiragana}', "");
Expect(1, 127488, '\p{Is_Script=Hiragana}', "");
Expect(0, 127488, '\p{^Is_Script=Hiragana}', "");
Expect(0, 127488, '\P{Is_Script=Hiragana}', "");
Expect(1, 127488, '\P{^Is_Script=Hiragana}', "");
Expect(0, 127489, '\p{Is_Script=Hiragana}', "");
Expect(1, 127489, '\p{^Is_Script=Hiragana}', "");
Expect(1, 127489, '\P{Is_Script=Hiragana}', "");
Expect(0, 127489, '\P{^Is_Script=Hiragana}', "");
Error('\p{Is_Sc=/a/hira}');
Error('\P{Is_Sc=/a/hira}');
Expect(1, 127488, '\p{Is_Sc=hira}', "");
Expect(0, 127488, '\p{^Is_Sc=hira}', "");
Expect(0, 127488, '\P{Is_Sc=hira}', "");
Expect(1, 127488, '\P{^Is_Sc=hira}', "");
Expect(0, 127489, '\p{Is_Sc=hira}', "");
Expect(1, 127489, '\p{^Is_Sc=hira}', "");
Expect(1, 127489, '\P{Is_Sc=hira}', "");
Expect(0, 127489, '\P{^Is_Sc=hira}', "");
Expect(1, 127488, '\p{Is_Sc= Hira}', "");
Expect(0, 127488, '\p{^Is_Sc= Hira}', "");
Expect(0, 127488, '\P{Is_Sc= Hira}', "");
Expect(1, 127488, '\P{^Is_Sc= Hira}', "");
Expect(0, 127489, '\p{Is_Sc= Hira}', "");
Expect(1, 127489, '\p{^Is_Sc= Hira}', "");
Expect(1, 127489, '\P{Is_Sc= Hira}', "");
Expect(0, 127489, '\P{^Is_Sc= Hira}', "");
Error('\p{Script=	BENGALI:=}');
Error('\P{Script=	BENGALI:=}');
Expect(1, 2555, '\p{Script=bengali}', "");
Expect(0, 2555, '\p{^Script=bengali}', "");
Expect(0, 2555, '\P{Script=bengali}', "");
Expect(1, 2555, '\P{^Script=bengali}', "");
Expect(0, 2556, '\p{Script=bengali}', "");
Expect(1, 2556, '\p{^Script=bengali}', "");
Expect(1, 2556, '\P{Script=bengali}', "");
Expect(0, 2556, '\P{^Script=bengali}', "");
Expect(1, 2555, '\p{Script= _BENGALI}', "");
Expect(0, 2555, '\p{^Script= _BENGALI}', "");
Expect(0, 2555, '\P{Script= _BENGALI}', "");
Expect(1, 2555, '\P{^Script= _BENGALI}', "");
Expect(0, 2556, '\p{Script= _BENGALI}', "");
Expect(1, 2556, '\p{^Script= _BENGALI}', "");
Expect(1, 2556, '\P{Script= _BENGALI}', "");
Expect(0, 2556, '\P{^Script= _BENGALI}', "");
Error('\p{Sc= Beng:=}');
Error('\P{Sc= Beng:=}');
Expect(1, 2555, '\p{Sc=beng}', "");
Expect(0, 2555, '\p{^Sc=beng}', "");
Expect(0, 2555, '\P{Sc=beng}', "");
Expect(1, 2555, '\P{^Sc=beng}', "");
Expect(0, 2556, '\p{Sc=beng}', "");
Expect(1, 2556, '\p{^Sc=beng}', "");
Expect(1, 2556, '\P{Sc=beng}', "");
Expect(0, 2556, '\P{^Sc=beng}', "");
Expect(1, 2555, '\p{Sc=_-beng}', "");
Expect(0, 2555, '\p{^Sc=_-beng}', "");
Expect(0, 2555, '\P{Sc=_-beng}', "");
Expect(1, 2555, '\P{^Sc=_-beng}', "");
Expect(0, 2556, '\p{Sc=_-beng}', "");
Expect(1, 2556, '\p{^Sc=_-beng}', "");
Expect(1, 2556, '\P{Sc=_-beng}', "");
Expect(0, 2556, '\P{^Sc=_-beng}', "");
Error('\p{Is_Script=-Bengali:=}');
Error('\P{Is_Script=-Bengali:=}');
Expect(1, 2555, '\p{Is_Script: bengali}', "");
Expect(0, 2555, '\p{^Is_Script: bengali}', "");
Expect(0, 2555, '\P{Is_Script: bengali}', "");
Expect(1, 2555, '\P{^Is_Script: bengali}', "");
Expect(0, 2556, '\p{Is_Script: bengali}', "");
Expect(1, 2556, '\p{^Is_Script: bengali}', "");
Expect(1, 2556, '\P{Is_Script: bengali}', "");
Expect(0, 2556, '\P{^Is_Script: bengali}', "");
Expect(1, 2555, '\p{Is_Script=	_Bengali}', "");
Expect(0, 2555, '\p{^Is_Script=	_Bengali}', "");
Expect(0, 2555, '\P{Is_Script=	_Bengali}', "");
Expect(1, 2555, '\P{^Is_Script=	_Bengali}', "");
Expect(0, 2556, '\p{Is_Script=	_Bengali}', "");
Expect(1, 2556, '\p{^Is_Script=	_Bengali}', "");
Expect(1, 2556, '\P{Is_Script=	_Bengali}', "");
Expect(0, 2556, '\P{^Is_Script=	_Bengali}', "");
Error('\p{Is_Sc=_ BENG/a/}');
Error('\P{Is_Sc=_ BENG/a/}');
Expect(1, 2555, '\p{Is_Sc=beng}', "");
Expect(0, 2555, '\p{^Is_Sc=beng}', "");
Expect(0, 2555, '\P{Is_Sc=beng}', "");
Expect(1, 2555, '\P{^Is_Sc=beng}', "");
Expect(0, 2556, '\p{Is_Sc=beng}', "");
Expect(1, 2556, '\p{^Is_Sc=beng}', "");
Expect(1, 2556, '\P{Is_Sc=beng}', "");
Expect(0, 2556, '\P{^Is_Sc=beng}', "");
Expect(1, 2555, '\p{Is_Sc=_ BENG}', "");
Expect(0, 2555, '\p{^Is_Sc=_ BENG}', "");
Expect(0, 2555, '\P{Is_Sc=_ BENG}', "");
Expect(1, 2555, '\P{^Is_Sc=_ BENG}', "");
Expect(0, 2556, '\p{Is_Sc=_ BENG}', "");
Expect(1, 2556, '\p{^Is_Sc=_ BENG}', "");
Expect(1, 2556, '\P{Is_Sc=_ BENG}', "");
Expect(0, 2556, '\P{^Is_Sc=_ BENG}', "");
Error('\p{Script=/a/_	Braille}');
Error('\P{Script=/a/_	Braille}');
Expect(1, 10495, '\p{Script=braille}', "");
Expect(0, 10495, '\p{^Script=braille}', "");
Expect(0, 10495, '\P{Script=braille}', "");
Expect(1, 10495, '\P{^Script=braille}', "");
Expect(0, 10496, '\p{Script=braille}', "");
Expect(1, 10496, '\p{^Script=braille}', "");
Expect(1, 10496, '\P{Script=braille}', "");
Expect(0, 10496, '\P{^Script=braille}', "");
Expect(1, 10495, '\p{Script=-Braille}', "");
Expect(0, 10495, '\p{^Script=-Braille}', "");
Expect(0, 10495, '\P{Script=-Braille}', "");
Expect(1, 10495, '\P{^Script=-Braille}', "");
Expect(0, 10496, '\p{Script=-Braille}', "");
Expect(1, 10496, '\p{^Script=-Braille}', "");
Expect(1, 10496, '\P{Script=-Braille}', "");
Expect(0, 10496, '\P{^Script=-Braille}', "");
Error('\p{Sc= /a/brai}');
Error('\P{Sc= /a/brai}');
Expect(1, 10495, '\p{Sc=brai}', "");
Expect(0, 10495, '\p{^Sc=brai}', "");
Expect(0, 10495, '\P{Sc=brai}', "");
Expect(1, 10495, '\P{^Sc=brai}', "");
Expect(0, 10496, '\p{Sc=brai}', "");
Expect(1, 10496, '\p{^Sc=brai}', "");
Expect(1, 10496, '\P{Sc=brai}', "");
Expect(0, 10496, '\P{^Sc=brai}', "");
Expect(1, 10495, '\p{Sc=	 Brai}', "");
Expect(0, 10495, '\p{^Sc=	 Brai}', "");
Expect(0, 10495, '\P{Sc=	 Brai}', "");
Expect(1, 10495, '\P{^Sc=	 Brai}', "");
Expect(0, 10496, '\p{Sc=	 Brai}', "");
Expect(1, 10496, '\p{^Sc=	 Brai}', "");
Expect(1, 10496, '\P{Sc=	 Brai}', "");
Expect(0, 10496, '\P{^Sc=	 Brai}', "");
Error('\p{Is_Script:   -_BRAILLE/a/}');
Error('\P{Is_Script:   -_BRAILLE/a/}');
Expect(1, 10495, '\p{Is_Script=braille}', "");
Expect(0, 10495, '\p{^Is_Script=braille}', "");
Expect(0, 10495, '\P{Is_Script=braille}', "");
Expect(1, 10495, '\P{^Is_Script=braille}', "");
Expect(0, 10496, '\p{Is_Script=braille}', "");
Expect(1, 10496, '\p{^Is_Script=braille}', "");
Expect(1, 10496, '\P{Is_Script=braille}', "");
Expect(0, 10496, '\P{^Is_Script=braille}', "");
Expect(1, 10495, '\p{Is_Script:	-Braille}', "");
Expect(0, 10495, '\p{^Is_Script:	-Braille}', "");
Expect(0, 10495, '\P{Is_Script:	-Braille}', "");
Expect(1, 10495, '\P{^Is_Script:	-Braille}', "");
Expect(0, 10496, '\p{Is_Script:	-Braille}', "");
Expect(1, 10496, '\p{^Is_Script:	-Braille}', "");
Expect(1, 10496, '\P{Is_Script:	-Braille}', "");
Expect(0, 10496, '\P{^Is_Script:	-Braille}', "");
Error('\p{Is_Sc=_/a/Brai}');
Error('\P{Is_Sc=_/a/Brai}');
Expect(1, 10495, '\p{Is_Sc=brai}', "");
Expect(0, 10495, '\p{^Is_Sc=brai}', "");
Expect(0, 10495, '\P{Is_Sc=brai}', "");
Expect(1, 10495, '\P{^Is_Sc=brai}', "");
Expect(0, 10496, '\p{Is_Sc=brai}', "");
Expect(1, 10496, '\p{^Is_Sc=brai}', "");
Expect(1, 10496, '\P{Is_Sc=brai}', "");
Expect(0, 10496, '\P{^Is_Sc=brai}', "");
Expect(1, 10495, '\p{Is_Sc:_-brai}', "");
Expect(0, 10495, '\p{^Is_Sc:_-brai}', "");
Expect(0, 10495, '\P{Is_Sc:_-brai}', "");
Expect(1, 10495, '\P{^Is_Sc:_-brai}', "");
Expect(0, 10496, '\p{Is_Sc:_-brai}', "");
Expect(1, 10496, '\p{^Is_Sc:_-brai}', "");
Expect(1, 10496, '\P{Is_Sc:_-brai}', "");
Expect(0, 10496, '\P{^Is_Sc:_-brai}', "");
Error('\p{Script=	 KHAROSHTHI/a/}');
Error('\P{Script=	 KHAROSHTHI/a/}');
Expect(1, 68184, '\p{Script=kharoshthi}', "");
Expect(0, 68184, '\p{^Script=kharoshthi}', "");
Expect(0, 68184, '\P{Script=kharoshthi}', "");
Expect(1, 68184, '\P{^Script=kharoshthi}', "");
Expect(0, 68185, '\p{Script=kharoshthi}', "");
Expect(1, 68185, '\p{^Script=kharoshthi}', "");
Expect(1, 68185, '\P{Script=kharoshthi}', "");
Expect(0, 68185, '\P{^Script=kharoshthi}', "");
Expect(1, 68184, '\p{Script=_Kharoshthi}', "");
Expect(0, 68184, '\p{^Script=_Kharoshthi}', "");
Expect(0, 68184, '\P{Script=_Kharoshthi}', "");
Expect(1, 68184, '\P{^Script=_Kharoshthi}', "");
Expect(0, 68185, '\p{Script=_Kharoshthi}', "");
Expect(1, 68185, '\p{^Script=_Kharoshthi}', "");
Expect(1, 68185, '\P{Script=_Kharoshthi}', "");
Expect(0, 68185, '\P{^Script=_Kharoshthi}', "");
Error('\p{Sc=/a/_-khar}');
Error('\P{Sc=/a/_-khar}');
Expect(1, 68184, '\p{Sc=khar}', "");
Expect(0, 68184, '\p{^Sc=khar}', "");
Expect(0, 68184, '\P{Sc=khar}', "");
Expect(1, 68184, '\P{^Sc=khar}', "");
Expect(0, 68185, '\p{Sc=khar}', "");
Expect(1, 68185, '\p{^Sc=khar}', "");
Expect(1, 68185, '\P{Sc=khar}', "");
Expect(0, 68185, '\P{^Sc=khar}', "");
Expect(1, 68184, '\p{Sc=_khar}', "");
Expect(0, 68184, '\p{^Sc=_khar}', "");
Expect(0, 68184, '\P{Sc=_khar}', "");
Expect(1, 68184, '\P{^Sc=_khar}', "");
Expect(0, 68185, '\p{Sc=_khar}', "");
Expect(1, 68185, '\p{^Sc=_khar}', "");
Expect(1, 68185, '\P{Sc=_khar}', "");
Expect(0, 68185, '\P{^Sc=_khar}', "");
Error('\p{Is_Script=:=_ Kharoshthi}');
Error('\P{Is_Script=:=_ Kharoshthi}');
Expect(1, 68184, '\p{Is_Script=kharoshthi}', "");
Expect(0, 68184, '\p{^Is_Script=kharoshthi}', "");
Expect(0, 68184, '\P{Is_Script=kharoshthi}', "");
Expect(1, 68184, '\P{^Is_Script=kharoshthi}', "");
Expect(0, 68185, '\p{Is_Script=kharoshthi}', "");
Expect(1, 68185, '\p{^Is_Script=kharoshthi}', "");
Expect(1, 68185, '\P{Is_Script=kharoshthi}', "");
Expect(0, 68185, '\P{^Is_Script=kharoshthi}', "");
Expect(1, 68184, '\p{Is_Script:   _Kharoshthi}', "");
Expect(0, 68184, '\p{^Is_Script:   _Kharoshthi}', "");
Expect(0, 68184, '\P{Is_Script:   _Kharoshthi}', "");
Expect(1, 68184, '\P{^Is_Script:   _Kharoshthi}', "");
Expect(0, 68185, '\p{Is_Script:   _Kharoshthi}', "");
Expect(1, 68185, '\p{^Is_Script:   _Kharoshthi}', "");
Expect(1, 68185, '\P{Is_Script:   _Kharoshthi}', "");
Expect(0, 68185, '\P{^Is_Script:   _Kharoshthi}', "");
Error('\p{Is_Sc=-:=khar}');
Error('\P{Is_Sc=-:=khar}');
Expect(1, 68184, '\p{Is_Sc=khar}', "");
Expect(0, 68184, '\p{^Is_Sc=khar}', "");
Expect(0, 68184, '\P{Is_Sc=khar}', "");
Expect(1, 68184, '\P{^Is_Sc=khar}', "");
Expect(0, 68185, '\p{Is_Sc=khar}', "");
Expect(1, 68185, '\p{^Is_Sc=khar}', "");
Expect(1, 68185, '\P{Is_Sc=khar}', "");
Expect(0, 68185, '\P{^Is_Sc=khar}', "");
Expect(1, 68184, '\p{Is_Sc= _Khar}', "");
Expect(0, 68184, '\p{^Is_Sc= _Khar}', "");
Expect(0, 68184, '\P{Is_Sc= _Khar}', "");
Expect(1, 68184, '\P{^Is_Sc= _Khar}', "");
Expect(0, 68185, '\p{Is_Sc= _Khar}', "");
Expect(1, 68185, '\p{^Is_Sc= _Khar}', "");
Expect(1, 68185, '\P{Is_Sc= _Khar}', "");
Expect(0, 68185, '\P{^Is_Sc= _Khar}', "");
Error('\p{Script=-:=Lycian}');
Error('\P{Script=-:=Lycian}');
Expect(1, 66204, '\p{Script=lycian}', "");
Expect(0, 66204, '\p{^Script=lycian}', "");
Expect(0, 66204, '\P{Script=lycian}', "");
Expect(1, 66204, '\P{^Script=lycian}', "");
Expect(0, 66205, '\p{Script=lycian}', "");
Expect(1, 66205, '\p{^Script=lycian}', "");
Expect(1, 66205, '\P{Script=lycian}', "");
Expect(0, 66205, '\P{^Script=lycian}', "");
Expect(1, 66204, '\p{Script:	- Lycian}', "");
Expect(0, 66204, '\p{^Script:	- Lycian}', "");
Expect(0, 66204, '\P{Script:	- Lycian}', "");
Expect(1, 66204, '\P{^Script:	- Lycian}', "");
Expect(0, 66205, '\p{Script:	- Lycian}', "");
Expect(1, 66205, '\p{^Script:	- Lycian}', "");
Expect(1, 66205, '\P{Script:	- Lycian}', "");
Expect(0, 66205, '\P{^Script:	- Lycian}', "");
Error('\p{Sc=_	lyci:=}');
Error('\P{Sc=_	lyci:=}');
Expect(1, 66204, '\p{Sc=lyci}', "");
Expect(0, 66204, '\p{^Sc=lyci}', "");
Expect(0, 66204, '\P{Sc=lyci}', "");
Expect(1, 66204, '\P{^Sc=lyci}', "");
Expect(0, 66205, '\p{Sc=lyci}', "");
Expect(1, 66205, '\p{^Sc=lyci}', "");
Expect(1, 66205, '\P{Sc=lyci}', "");
Expect(0, 66205, '\P{^Sc=lyci}', "");
Expect(1, 66204, '\p{Sc=- lyci}', "");
Expect(0, 66204, '\p{^Sc=- lyci}', "");
Expect(0, 66204, '\P{Sc=- lyci}', "");
Expect(1, 66204, '\P{^Sc=- lyci}', "");
Expect(0, 66205, '\p{Sc=- lyci}', "");
Expect(1, 66205, '\p{^Sc=- lyci}', "");
Expect(1, 66205, '\P{Sc=- lyci}', "");
Expect(0, 66205, '\P{^Sc=- lyci}', "");
Error('\p{Is_Script=:=	 lycian}');
Error('\P{Is_Script=:=	 lycian}');
Expect(1, 66204, '\p{Is_Script=lycian}', "");
Expect(0, 66204, '\p{^Is_Script=lycian}', "");
Expect(0, 66204, '\P{Is_Script=lycian}', "");
Expect(1, 66204, '\P{^Is_Script=lycian}', "");
Expect(0, 66205, '\p{Is_Script=lycian}', "");
Expect(1, 66205, '\p{^Is_Script=lycian}', "");
Expect(1, 66205, '\P{Is_Script=lycian}', "");
Expect(0, 66205, '\P{^Is_Script=lycian}', "");
Expect(1, 66204, '\p{Is_Script=_	LYCIAN}', "");
Expect(0, 66204, '\p{^Is_Script=_	LYCIAN}', "");
Expect(0, 66204, '\P{Is_Script=_	LYCIAN}', "");
Expect(1, 66204, '\P{^Is_Script=_	LYCIAN}', "");
Expect(0, 66205, '\p{Is_Script=_	LYCIAN}', "");
Expect(1, 66205, '\p{^Is_Script=_	LYCIAN}', "");
Expect(1, 66205, '\P{Is_Script=_	LYCIAN}', "");
Expect(0, 66205, '\P{^Is_Script=_	LYCIAN}', "");
Error('\p{Is_Sc=_Lyci:=}');
Error('\P{Is_Sc=_Lyci:=}');
Expect(1, 66204, '\p{Is_Sc=lyci}', "");
Expect(0, 66204, '\p{^Is_Sc=lyci}', "");
Expect(0, 66204, '\P{Is_Sc=lyci}', "");
Expect(1, 66204, '\P{^Is_Sc=lyci}', "");
Expect(0, 66205, '\p{Is_Sc=lyci}', "");
Expect(1, 66205, '\p{^Is_Sc=lyci}', "");
Expect(1, 66205, '\P{Is_Sc=lyci}', "");
Expect(0, 66205, '\P{^Is_Sc=lyci}', "");
Expect(1, 66204, '\p{Is_Sc=	Lyci}', "");
Expect(0, 66204, '\p{^Is_Sc=	Lyci}', "");
Expect(0, 66204, '\P{Is_Sc=	Lyci}', "");
Expect(1, 66204, '\P{^Is_Sc=	Lyci}', "");
Expect(0, 66205, '\p{Is_Sc=	Lyci}', "");
Expect(1, 66205, '\p{^Is_Sc=	Lyci}', "");
Expect(1, 66205, '\P{Is_Sc=	Lyci}', "");
Expect(0, 66205, '\P{^Is_Sc=	Lyci}', "");
Error('\p{Script:   /a/_	Bopomofo}');
Error('\P{Script:   /a/_	Bopomofo}');
Expect(1, 12730, '\p{Script=bopomofo}', "");
Expect(0, 12730, '\p{^Script=bopomofo}', "");
Expect(0, 12730, '\P{Script=bopomofo}', "");
Expect(1, 12730, '\P{^Script=bopomofo}', "");
Expect(0, 12731, '\p{Script=bopomofo}', "");
Expect(1, 12731, '\p{^Script=bopomofo}', "");
Expect(1, 12731, '\P{Script=bopomofo}', "");
Expect(0, 12731, '\P{^Script=bopomofo}', "");
Expect(1, 12730, '\p{Script=BOPOMOFO}', "");
Expect(0, 12730, '\p{^Script=BOPOMOFO}', "");
Expect(0, 12730, '\P{Script=BOPOMOFO}', "");
Expect(1, 12730, '\P{^Script=BOPOMOFO}', "");
Expect(0, 12731, '\p{Script=BOPOMOFO}', "");
Expect(1, 12731, '\p{^Script=BOPOMOFO}', "");
Expect(1, 12731, '\P{Script=BOPOMOFO}', "");
Expect(0, 12731, '\P{^Script=BOPOMOFO}', "");
Error('\p{Sc= /a/BOPO}');
Error('\P{Sc= /a/BOPO}');
Expect(1, 12730, '\p{Sc=bopo}', "");
Expect(0, 12730, '\p{^Sc=bopo}', "");
Expect(0, 12730, '\P{Sc=bopo}', "");
Expect(1, 12730, '\P{^Sc=bopo}', "");
Expect(0, 12731, '\p{Sc=bopo}', "");
Expect(1, 12731, '\p{^Sc=bopo}', "");
Expect(1, 12731, '\P{Sc=bopo}', "");
Expect(0, 12731, '\P{^Sc=bopo}', "");
Expect(1, 12730, '\p{Sc=  Bopo}', "");
Expect(0, 12730, '\p{^Sc=  Bopo}', "");
Expect(0, 12730, '\P{Sc=  Bopo}', "");
Expect(1, 12730, '\P{^Sc=  Bopo}', "");
Expect(0, 12731, '\p{Sc=  Bopo}', "");
Expect(1, 12731, '\p{^Sc=  Bopo}', "");
Expect(1, 12731, '\P{Sc=  Bopo}', "");
Expect(0, 12731, '\P{^Sc=  Bopo}', "");
Error('\p{Is_Script=:= bopomofo}');
Error('\P{Is_Script=:= bopomofo}');
Expect(1, 12730, '\p{Is_Script=bopomofo}', "");
Expect(0, 12730, '\p{^Is_Script=bopomofo}', "");
Expect(0, 12730, '\P{Is_Script=bopomofo}', "");
Expect(1, 12730, '\P{^Is_Script=bopomofo}', "");
Expect(0, 12731, '\p{Is_Script=bopomofo}', "");
Expect(1, 12731, '\p{^Is_Script=bopomofo}', "");
Expect(1, 12731, '\P{Is_Script=bopomofo}', "");
Expect(0, 12731, '\P{^Is_Script=bopomofo}', "");
Expect(1, 12730, '\p{Is_Script=	_bopomofo}', "");
Expect(0, 12730, '\p{^Is_Script=	_bopomofo}', "");
Expect(0, 12730, '\P{Is_Script=	_bopomofo}', "");
Expect(1, 12730, '\P{^Is_Script=	_bopomofo}', "");
Expect(0, 12731, '\p{Is_Script=	_bopomofo}', "");
Expect(1, 12731, '\p{^Is_Script=	_bopomofo}', "");
Expect(1, 12731, '\P{Is_Script=	_bopomofo}', "");
Expect(0, 12731, '\P{^Is_Script=	_bopomofo}', "");
Error('\p{Is_Sc:   :=--Bopo}');
Error('\P{Is_Sc:   :=--Bopo}');
Expect(1, 12730, '\p{Is_Sc=bopo}', "");
Expect(0, 12730, '\p{^Is_Sc=bopo}', "");
Expect(0, 12730, '\P{Is_Sc=bopo}', "");
Expect(1, 12730, '\P{^Is_Sc=bopo}', "");
Expect(0, 12731, '\p{Is_Sc=bopo}', "");
Expect(1, 12731, '\p{^Is_Sc=bopo}', "");
Expect(1, 12731, '\P{Is_Sc=bopo}', "");
Expect(0, 12731, '\P{^Is_Sc=bopo}', "");
Expect(1, 12730, '\p{Is_Sc=_BOPO}', "");
Expect(0, 12730, '\p{^Is_Sc=_BOPO}', "");
Expect(0, 12730, '\P{Is_Sc=_BOPO}', "");
Expect(1, 12730, '\P{^Is_Sc=_BOPO}', "");
Expect(0, 12731, '\p{Is_Sc=_BOPO}', "");
Expect(1, 12731, '\p{^Is_Sc=_BOPO}', "");
Expect(1, 12731, '\P{Is_Sc=_BOPO}', "");
Expect(0, 12731, '\P{^Is_Sc=_BOPO}', "");
Error('\p{Script=:=REJANG}');
Error('\P{Script=:=REJANG}');
Expect(1, 43359, '\p{Script=rejang}', "");
Expect(0, 43359, '\p{^Script=rejang}', "");
Expect(0, 43359, '\P{Script=rejang}', "");
Expect(1, 43359, '\P{^Script=rejang}', "");
Expect(0, 43360, '\p{Script=rejang}', "");
Expect(1, 43360, '\p{^Script=rejang}', "");
Expect(1, 43360, '\P{Script=rejang}', "");
Expect(0, 43360, '\P{^Script=rejang}', "");
Expect(1, 43359, '\p{Script=_-REJANG}', "");
Expect(0, 43359, '\p{^Script=_-REJANG}', "");
Expect(0, 43359, '\P{Script=_-REJANG}', "");
Expect(1, 43359, '\P{^Script=_-REJANG}', "");
Expect(0, 43360, '\p{Script=_-REJANG}', "");
Expect(1, 43360, '\p{^Script=_-REJANG}', "");
Expect(1, 43360, '\P{Script=_-REJANG}', "");
Expect(0, 43360, '\P{^Script=_-REJANG}', "");
Error('\p{Sc=		Rjng/a/}');
Error('\P{Sc=		Rjng/a/}');
Expect(1, 43359, '\p{Sc=rjng}', "");
Expect(0, 43359, '\p{^Sc=rjng}', "");
Expect(0, 43359, '\P{Sc=rjng}', "");
Expect(1, 43359, '\P{^Sc=rjng}', "");
Expect(0, 43360, '\p{Sc=rjng}', "");
Expect(1, 43360, '\p{^Sc=rjng}', "");
Expect(1, 43360, '\P{Sc=rjng}', "");
Expect(0, 43360, '\P{^Sc=rjng}', "");
Expect(1, 43359, '\p{Sc=_ Rjng}', "");
Expect(0, 43359, '\p{^Sc=_ Rjng}', "");
Expect(0, 43359, '\P{Sc=_ Rjng}', "");
Expect(1, 43359, '\P{^Sc=_ Rjng}', "");
Expect(0, 43360, '\p{Sc=_ Rjng}', "");
Expect(1, 43360, '\p{^Sc=_ Rjng}', "");
Expect(1, 43360, '\P{Sc=_ Rjng}', "");
Expect(0, 43360, '\P{^Sc=_ Rjng}', "");
Error('\p{Is_Script: :=	 Rejang}');
Error('\P{Is_Script: :=	 Rejang}');
Expect(1, 43359, '\p{Is_Script: rejang}', "");
Expect(0, 43359, '\p{^Is_Script: rejang}', "");
Expect(0, 43359, '\P{Is_Script: rejang}', "");
Expect(1, 43359, '\P{^Is_Script: rejang}', "");
Expect(0, 43360, '\p{Is_Script: rejang}', "");
Expect(1, 43360, '\p{^Is_Script: rejang}', "");
Expect(1, 43360, '\P{Is_Script: rejang}', "");
Expect(0, 43360, '\P{^Is_Script: rejang}', "");
Expect(1, 43359, '\p{Is_Script:_ Rejang}', "");
Expect(0, 43359, '\p{^Is_Script:_ Rejang}', "");
Expect(0, 43359, '\P{Is_Script:_ Rejang}', "");
Expect(1, 43359, '\P{^Is_Script:_ Rejang}', "");
Expect(0, 43360, '\p{Is_Script:_ Rejang}', "");
Expect(1, 43360, '\p{^Is_Script:_ Rejang}', "");
Expect(1, 43360, '\P{Is_Script:_ Rejang}', "");
Expect(0, 43360, '\P{^Is_Script:_ Rejang}', "");
Error('\p{Is_Sc=	rjng/a/}');
Error('\P{Is_Sc=	rjng/a/}');
Expect(1, 43359, '\p{Is_Sc=rjng}', "");
Expect(0, 43359, '\p{^Is_Sc=rjng}', "");
Expect(0, 43359, '\P{Is_Sc=rjng}', "");
Expect(1, 43359, '\P{^Is_Sc=rjng}', "");
Expect(0, 43360, '\p{Is_Sc=rjng}', "");
Expect(1, 43360, '\p{^Is_Sc=rjng}', "");
Expect(1, 43360, '\P{Is_Sc=rjng}', "");
Expect(0, 43360, '\P{^Is_Sc=rjng}', "");
Expect(1, 43359, '\p{Is_Sc=-	RJNG}', "");
Expect(0, 43359, '\p{^Is_Sc=-	RJNG}', "");
Expect(0, 43359, '\P{Is_Sc=-	RJNG}', "");
Expect(1, 43359, '\P{^Is_Sc=-	RJNG}', "");
Expect(0, 43360, '\p{Is_Sc=-	RJNG}', "");
Expect(1, 43360, '\p{^Is_Sc=-	RJNG}', "");
Expect(1, 43360, '\P{Is_Sc=-	RJNG}', "");
Expect(0, 43360, '\P{^Is_Sc=-	RJNG}', "");
Error('\p{Script=-/a/MEROITIC_CURSIVE}');
Error('\P{Script=-/a/MEROITIC_CURSIVE}');
Expect(1, 68031, '\p{Script=meroiticcursive}', "");
Expect(0, 68031, '\p{^Script=meroiticcursive}', "");
Expect(0, 68031, '\P{Script=meroiticcursive}', "");
Expect(1, 68031, '\P{^Script=meroiticcursive}', "");
Expect(0, 68032, '\p{Script=meroiticcursive}', "");
Expect(1, 68032, '\p{^Script=meroiticcursive}', "");
Expect(1, 68032, '\P{Script=meroiticcursive}', "");
Expect(0, 68032, '\P{^Script=meroiticcursive}', "");
Expect(1, 68031, '\p{Script=-MEROITIC_Cursive}', "");
Expect(0, 68031, '\p{^Script=-MEROITIC_Cursive}', "");
Expect(0, 68031, '\P{Script=-MEROITIC_Cursive}', "");
Expect(1, 68031, '\P{^Script=-MEROITIC_Cursive}', "");
Expect(0, 68032, '\p{Script=-MEROITIC_Cursive}', "");
Expect(1, 68032, '\p{^Script=-MEROITIC_Cursive}', "");
Expect(1, 68032, '\P{Script=-MEROITIC_Cursive}', "");
Expect(0, 68032, '\P{^Script=-MEROITIC_Cursive}', "");
Error('\p{Sc=	 MERC/a/}');
Error('\P{Sc=	 MERC/a/}');
Expect(1, 68031, '\p{Sc=merc}', "");
Expect(0, 68031, '\p{^Sc=merc}', "");
Expect(0, 68031, '\P{Sc=merc}', "");
Expect(1, 68031, '\P{^Sc=merc}', "");
Expect(0, 68032, '\p{Sc=merc}', "");
Expect(1, 68032, '\p{^Sc=merc}', "");
Expect(1, 68032, '\P{Sc=merc}', "");
Expect(0, 68032, '\P{^Sc=merc}', "");
Expect(1, 68031, '\p{Sc: Merc}', "");
Expect(0, 68031, '\p{^Sc: Merc}', "");
Expect(0, 68031, '\P{Sc: Merc}', "");
Expect(1, 68031, '\P{^Sc: Merc}', "");
Expect(0, 68032, '\p{Sc: Merc}', "");
Expect(1, 68032, '\p{^Sc: Merc}', "");
Expect(1, 68032, '\P{Sc: Merc}', "");
Expect(0, 68032, '\P{^Sc: Merc}', "");
Error('\p{Is_Script:	:=	_Meroitic_cursive}');
Error('\P{Is_Script:	:=	_Meroitic_cursive}');
Expect(1, 68031, '\p{Is_Script=meroiticcursive}', "");
Expect(0, 68031, '\p{^Is_Script=meroiticcursive}', "");
Expect(0, 68031, '\P{Is_Script=meroiticcursive}', "");
Expect(1, 68031, '\P{^Is_Script=meroiticcursive}', "");
Expect(0, 68032, '\p{Is_Script=meroiticcursive}', "");
Expect(1, 68032, '\p{^Is_Script=meroiticcursive}', "");
Expect(1, 68032, '\P{Is_Script=meroiticcursive}', "");
Expect(0, 68032, '\P{^Is_Script=meroiticcursive}', "");
Expect(1, 68031, '\p{Is_Script=  Meroitic_Cursive}', "");
Expect(0, 68031, '\p{^Is_Script=  Meroitic_Cursive}', "");
Expect(0, 68031, '\P{Is_Script=  Meroitic_Cursive}', "");
Expect(1, 68031, '\P{^Is_Script=  Meroitic_Cursive}', "");
Expect(0, 68032, '\p{Is_Script=  Meroitic_Cursive}', "");
Expect(1, 68032, '\p{^Is_Script=  Meroitic_Cursive}', "");
Expect(1, 68032, '\P{Is_Script=  Meroitic_Cursive}', "");
Expect(0, 68032, '\P{^Is_Script=  Meroitic_Cursive}', "");
Error('\p{Is_Sc= /a/merc}');
Error('\P{Is_Sc= /a/merc}');
Expect(1, 68031, '\p{Is_Sc=merc}', "");
Expect(0, 68031, '\p{^Is_Sc=merc}', "");
Expect(0, 68031, '\P{Is_Sc=merc}', "");
Expect(1, 68031, '\P{^Is_Sc=merc}', "");
Expect(0, 68032, '\p{Is_Sc=merc}', "");
Expect(1, 68032, '\p{^Is_Sc=merc}', "");
Expect(1, 68032, '\P{Is_Sc=merc}', "");
Expect(0, 68032, '\P{^Is_Sc=merc}', "");
Expect(1, 68031, '\p{Is_Sc=_	Merc}', "");
Expect(0, 68031, '\p{^Is_Sc=_	Merc}', "");
Expect(0, 68031, '\P{Is_Sc=_	Merc}', "");
Expect(1, 68031, '\P{^Is_Sc=_	Merc}', "");
Expect(0, 68032, '\p{Is_Sc=_	Merc}', "");
Expect(1, 68032, '\p{^Is_Sc=_	Merc}', "");
Expect(1, 68032, '\P{Is_Sc=_	Merc}', "");
Expect(0, 68032, '\P{^Is_Sc=_	Merc}', "");
Error('\p{Script=	 HANUNOO/a/}');
Error('\P{Script=	 HANUNOO/a/}');
Expect(1, 5940, '\p{Script=hanunoo}', "");
Expect(0, 5940, '\p{^Script=hanunoo}', "");
Expect(0, 5940, '\P{Script=hanunoo}', "");
Expect(1, 5940, '\P{^Script=hanunoo}', "");
Expect(0, 5941, '\p{Script=hanunoo}', "");
Expect(1, 5941, '\p{^Script=hanunoo}', "");
Expect(1, 5941, '\P{Script=hanunoo}', "");
Expect(0, 5941, '\P{^Script=hanunoo}', "");
Expect(1, 5940, '\p{Script= -Hanunoo}', "");
Expect(0, 5940, '\p{^Script= -Hanunoo}', "");
Expect(0, 5940, '\P{Script= -Hanunoo}', "");
Expect(1, 5940, '\P{^Script= -Hanunoo}', "");
Expect(0, 5941, '\p{Script= -Hanunoo}', "");
Expect(1, 5941, '\p{^Script= -Hanunoo}', "");
Expect(1, 5941, '\P{Script= -Hanunoo}', "");
Expect(0, 5941, '\P{^Script= -Hanunoo}', "");
Error('\p{Sc=-HANO/a/}');
Error('\P{Sc=-HANO/a/}');
Expect(1, 5940, '\p{Sc=hano}', "");
Expect(0, 5940, '\p{^Sc=hano}', "");
Expect(0, 5940, '\P{Sc=hano}', "");
Expect(1, 5940, '\P{^Sc=hano}', "");
Expect(0, 5941, '\p{Sc=hano}', "");
Expect(1, 5941, '\p{^Sc=hano}', "");
Expect(1, 5941, '\P{Sc=hano}', "");
Expect(0, 5941, '\P{^Sc=hano}', "");
Expect(1, 5940, '\p{Sc=_ HANO}', "");
Expect(0, 5940, '\p{^Sc=_ HANO}', "");
Expect(0, 5940, '\P{Sc=_ HANO}', "");
Expect(1, 5940, '\P{^Sc=_ HANO}', "");
Expect(0, 5941, '\p{Sc=_ HANO}', "");
Expect(1, 5941, '\p{^Sc=_ HANO}', "");
Expect(1, 5941, '\P{Sc=_ HANO}', "");
Expect(0, 5941, '\P{^Sc=_ HANO}', "");
Error('\p{Is_Script=-HANUNOO/a/}');
Error('\P{Is_Script=-HANUNOO/a/}');
Expect(1, 5940, '\p{Is_Script=hanunoo}', "");
Expect(0, 5940, '\p{^Is_Script=hanunoo}', "");
Expect(0, 5940, '\P{Is_Script=hanunoo}', "");
Expect(1, 5940, '\P{^Is_Script=hanunoo}', "");
Expect(0, 5941, '\p{Is_Script=hanunoo}', "");
Expect(1, 5941, '\p{^Is_Script=hanunoo}', "");
Expect(1, 5941, '\P{Is_Script=hanunoo}', "");
Expect(0, 5941, '\P{^Is_Script=hanunoo}', "");
Expect(1, 5940, '\p{Is_Script=	Hanunoo}', "");
Expect(0, 5940, '\p{^Is_Script=	Hanunoo}', "");
Expect(0, 5940, '\P{Is_Script=	Hanunoo}', "");
Expect(1, 5940, '\P{^Is_Script=	Hanunoo}', "");
Expect(0, 5941, '\p{Is_Script=	Hanunoo}', "");
Expect(1, 5941, '\p{^Is_Script=	Hanunoo}', "");
Expect(1, 5941, '\P{Is_Script=	Hanunoo}', "");
Expect(0, 5941, '\P{^Is_Script=	Hanunoo}', "");
Error('\p{Is_Sc:	:=-HANO}');
Error('\P{Is_Sc:	:=-HANO}');
Expect(1, 5940, '\p{Is_Sc=hano}', "");
Expect(0, 5940, '\p{^Is_Sc=hano}', "");
Expect(0, 5940, '\P{Is_Sc=hano}', "");
Expect(1, 5940, '\P{^Is_Sc=hano}', "");
Expect(0, 5941, '\p{Is_Sc=hano}', "");
Expect(1, 5941, '\p{^Is_Sc=hano}', "");
Expect(1, 5941, '\P{Is_Sc=hano}', "");
Expect(0, 5941, '\P{^Is_Sc=hano}', "");
Expect(1, 5940, '\p{Is_Sc= _Hano}', "");
Expect(0, 5940, '\p{^Is_Sc= _Hano}', "");
Expect(0, 5940, '\P{Is_Sc= _Hano}', "");
Expect(1, 5940, '\P{^Is_Sc= _Hano}', "");
Expect(0, 5941, '\p{Is_Sc= _Hano}', "");
Expect(1, 5941, '\p{^Is_Sc= _Hano}', "");
Expect(1, 5941, '\P{Is_Sc= _Hano}', "");
Expect(0, 5941, '\P{^Is_Sc= _Hano}', "");
Error('\p{Script=	-Imperial_aramaic/a/}');
Error('\P{Script=	-Imperial_aramaic/a/}');
Expect(1, 67679, '\p{Script=imperialaramaic}', "");
Expect(0, 67679, '\p{^Script=imperialaramaic}', "");
Expect(0, 67679, '\P{Script=imperialaramaic}', "");
Expect(1, 67679, '\P{^Script=imperialaramaic}', "");
Expect(0, 67680, '\p{Script=imperialaramaic}', "");
Expect(1, 67680, '\p{^Script=imperialaramaic}', "");
Expect(1, 67680, '\P{Script=imperialaramaic}', "");
Expect(0, 67680, '\P{^Script=imperialaramaic}', "");
Expect(1, 67679, '\p{Script=-	Imperial_ARAMAIC}', "");
Expect(0, 67679, '\p{^Script=-	Imperial_ARAMAIC}', "");
Expect(0, 67679, '\P{Script=-	Imperial_ARAMAIC}', "");
Expect(1, 67679, '\P{^Script=-	Imperial_ARAMAIC}', "");
Expect(0, 67680, '\p{Script=-	Imperial_ARAMAIC}', "");
Expect(1, 67680, '\p{^Script=-	Imperial_ARAMAIC}', "");
Expect(1, 67680, '\P{Script=-	Imperial_ARAMAIC}', "");
Expect(0, 67680, '\P{^Script=-	Imperial_ARAMAIC}', "");
Error('\p{Sc=/a/armi}');
Error('\P{Sc=/a/armi}');
Expect(1, 67679, '\p{Sc=armi}', "");
Expect(0, 67679, '\p{^Sc=armi}', "");
Expect(0, 67679, '\P{Sc=armi}', "");
Expect(1, 67679, '\P{^Sc=armi}', "");
Expect(0, 67680, '\p{Sc=armi}', "");
Expect(1, 67680, '\p{^Sc=armi}', "");
Expect(1, 67680, '\P{Sc=armi}', "");
Expect(0, 67680, '\P{^Sc=armi}', "");
Expect(1, 67679, '\p{Sc:   _	Armi}', "");
Expect(0, 67679, '\p{^Sc:   _	Armi}', "");
Expect(0, 67679, '\P{Sc:   _	Armi}', "");
Expect(1, 67679, '\P{^Sc:   _	Armi}', "");
Expect(0, 67680, '\p{Sc:   _	Armi}', "");
Expect(1, 67680, '\p{^Sc:   _	Armi}', "");
Expect(1, 67680, '\P{Sc:   _	Armi}', "");
Expect(0, 67680, '\P{^Sc:   _	Armi}', "");
Error('\p{Is_Script=-:=Imperial_aramaic}');
Error('\P{Is_Script=-:=Imperial_aramaic}');
Expect(1, 67679, '\p{Is_Script=imperialaramaic}', "");
Expect(0, 67679, '\p{^Is_Script=imperialaramaic}', "");
Expect(0, 67679, '\P{Is_Script=imperialaramaic}', "");
Expect(1, 67679, '\P{^Is_Script=imperialaramaic}', "");
Expect(0, 67680, '\p{Is_Script=imperialaramaic}', "");
Expect(1, 67680, '\p{^Is_Script=imperialaramaic}', "");
Expect(1, 67680, '\P{Is_Script=imperialaramaic}', "");
Expect(0, 67680, '\P{^Is_Script=imperialaramaic}', "");
Expect(1, 67679, '\p{Is_Script=IMPERIAL_aramaic}', "");
Expect(0, 67679, '\p{^Is_Script=IMPERIAL_aramaic}', "");
Expect(0, 67679, '\P{Is_Script=IMPERIAL_aramaic}', "");
Expect(1, 67679, '\P{^Is_Script=IMPERIAL_aramaic}', "");
Expect(0, 67680, '\p{Is_Script=IMPERIAL_aramaic}', "");
Expect(1, 67680, '\p{^Is_Script=IMPERIAL_aramaic}', "");
Expect(1, 67680, '\P{Is_Script=IMPERIAL_aramaic}', "");
Expect(0, 67680, '\P{^Is_Script=IMPERIAL_aramaic}', "");
Error('\p{Is_Sc= :=Armi}');
Error('\P{Is_Sc= :=Armi}');
Expect(1, 67679, '\p{Is_Sc=armi}', "");
Expect(0, 67679, '\p{^Is_Sc=armi}', "");
Expect(0, 67679, '\P{Is_Sc=armi}', "");
Expect(1, 67679, '\P{^Is_Sc=armi}', "");
Expect(0, 67680, '\p{Is_Sc=armi}', "");
Expect(1, 67680, '\p{^Is_Sc=armi}', "");
Expect(1, 67680, '\P{Is_Sc=armi}', "");
Expect(0, 67680, '\P{^Is_Sc=armi}', "");
Expect(1, 67679, '\p{Is_Sc=	ARMI}', "");
Expect(0, 67679, '\p{^Is_Sc=	ARMI}', "");
Expect(0, 67679, '\P{Is_Sc=	ARMI}', "");
Expect(1, 67679, '\P{^Is_Sc=	ARMI}', "");
Expect(0, 67680, '\p{Is_Sc=	ARMI}', "");
Expect(1, 67680, '\p{^Is_Sc=	ARMI}', "");
Expect(1, 67680, '\P{Is_Sc=	ARMI}', "");
Expect(0, 67680, '\P{^Is_Sc=	ARMI}', "");
Error('\p{Script=KANNADA:=}');
Error('\P{Script=KANNADA:=}');
Expect(1, 3314, '\p{Script=kannada}', "");
Expect(0, 3314, '\p{^Script=kannada}', "");
Expect(0, 3314, '\P{Script=kannada}', "");
Expect(1, 3314, '\P{^Script=kannada}', "");
Expect(0, 3315, '\p{Script=kannada}', "");
Expect(1, 3315, '\p{^Script=kannada}', "");
Expect(1, 3315, '\P{Script=kannada}', "");
Expect(0, 3315, '\P{^Script=kannada}', "");
Expect(1, 3314, '\p{Script= _KANNADA}', "");
Expect(0, 3314, '\p{^Script= _KANNADA}', "");
Expect(0, 3314, '\P{Script= _KANNADA}', "");
Expect(1, 3314, '\P{^Script= _KANNADA}', "");
Expect(0, 3315, '\p{Script= _KANNADA}', "");
Expect(1, 3315, '\p{^Script= _KANNADA}', "");
Expect(1, 3315, '\P{Script= _KANNADA}', "");
Expect(0, 3315, '\P{^Script= _KANNADA}', "");
Error('\p{Sc=__knda:=}');
Error('\P{Sc=__knda:=}');
Expect(1, 3314, '\p{Sc=knda}', "");
Expect(0, 3314, '\p{^Sc=knda}', "");
Expect(0, 3314, '\P{Sc=knda}', "");
Expect(1, 3314, '\P{^Sc=knda}', "");
Expect(0, 3315, '\p{Sc=knda}', "");
Expect(1, 3315, '\p{^Sc=knda}', "");
Expect(1, 3315, '\P{Sc=knda}', "");
Expect(0, 3315, '\P{^Sc=knda}', "");
Expect(1, 3314, '\p{Sc=-_knda}', "");
Expect(0, 3314, '\p{^Sc=-_knda}', "");
Expect(0, 3314, '\P{Sc=-_knda}', "");
Expect(1, 3314, '\P{^Sc=-_knda}', "");
Expect(0, 3315, '\p{Sc=-_knda}', "");
Expect(1, 3315, '\p{^Sc=-_knda}', "");
Expect(1, 3315, '\P{Sc=-_knda}', "");
Expect(0, 3315, '\P{^Sc=-_knda}', "");
Error('\p{Is_Script=	/a/KANNADA}');
Error('\P{Is_Script=	/a/KANNADA}');
Expect(1, 3314, '\p{Is_Script=kannada}', "");
Expect(0, 3314, '\p{^Is_Script=kannada}', "");
Expect(0, 3314, '\P{Is_Script=kannada}', "");
Expect(1, 3314, '\P{^Is_Script=kannada}', "");
Expect(0, 3315, '\p{Is_Script=kannada}', "");
Expect(1, 3315, '\p{^Is_Script=kannada}', "");
Expect(1, 3315, '\P{Is_Script=kannada}', "");
Expect(0, 3315, '\P{^Is_Script=kannada}', "");
Expect(1, 3314, '\p{Is_Script= Kannada}', "");
Expect(0, 3314, '\p{^Is_Script= Kannada}', "");
Expect(0, 3314, '\P{Is_Script= Kannada}', "");
Expect(1, 3314, '\P{^Is_Script= Kannada}', "");
Expect(0, 3315, '\p{Is_Script= Kannada}', "");
Expect(1, 3315, '\p{^Is_Script= Kannada}', "");
Expect(1, 3315, '\P{Is_Script= Kannada}', "");
Expect(0, 3315, '\P{^Is_Script= Kannada}', "");
Error('\p{Is_Sc:  Knda:=}');
Error('\P{Is_Sc:  Knda:=}');
Expect(1, 3314, '\p{Is_Sc=knda}', "");
Expect(0, 3314, '\p{^Is_Sc=knda}', "");
Expect(0, 3314, '\P{Is_Sc=knda}', "");
Expect(1, 3314, '\P{^Is_Sc=knda}', "");
Expect(0, 3315, '\p{Is_Sc=knda}', "");
Expect(1, 3315, '\p{^Is_Sc=knda}', "");
Expect(1, 3315, '\P{Is_Sc=knda}', "");
Expect(0, 3315, '\P{^Is_Sc=knda}', "");
Expect(1, 3314, '\p{Is_Sc: 	 Knda}', "");
Expect(0, 3314, '\p{^Is_Sc: 	 Knda}', "");
Expect(0, 3314, '\P{Is_Sc: 	 Knda}', "");
Expect(1, 3314, '\P{^Is_Sc: 	 Knda}', "");
Expect(0, 3315, '\p{Is_Sc: 	 Knda}', "");
Expect(1, 3315, '\p{^Is_Sc: 	 Knda}', "");
Expect(1, 3315, '\P{Is_Sc: 	 Knda}', "");
Expect(0, 3315, '\P{^Is_Sc: 	 Knda}', "");
Error('\p{Script=/a/ khmer}');
Error('\P{Script=/a/ khmer}');
Expect(1, 6655, '\p{Script=khmer}', "");
Expect(0, 6655, '\p{^Script=khmer}', "");
Expect(0, 6655, '\P{Script=khmer}', "");
Expect(1, 6655, '\P{^Script=khmer}', "");
Expect(0, 6656, '\p{Script=khmer}', "");
Expect(1, 6656, '\p{^Script=khmer}', "");
Expect(1, 6656, '\P{Script=khmer}', "");
Expect(0, 6656, '\P{^Script=khmer}', "");
Expect(1, 6655, '\p{Script=-	KHMER}', "");
Expect(0, 6655, '\p{^Script=-	KHMER}', "");
Expect(0, 6655, '\P{Script=-	KHMER}', "");
Expect(1, 6655, '\P{^Script=-	KHMER}', "");
Expect(0, 6656, '\p{Script=-	KHMER}', "");
Expect(1, 6656, '\p{^Script=-	KHMER}', "");
Expect(1, 6656, '\P{Script=-	KHMER}', "");
Expect(0, 6656, '\P{^Script=-	KHMER}', "");
Error('\p{Sc=_/a/KHMR}');
Error('\P{Sc=_/a/KHMR}');
Expect(1, 6655, '\p{Sc=khmr}', "");
Expect(0, 6655, '\p{^Sc=khmr}', "");
Expect(0, 6655, '\P{Sc=khmr}', "");
Expect(1, 6655, '\P{^Sc=khmr}', "");
Expect(0, 6656, '\p{Sc=khmr}', "");
Expect(1, 6656, '\p{^Sc=khmr}', "");
Expect(1, 6656, '\P{Sc=khmr}', "");
Expect(0, 6656, '\P{^Sc=khmr}', "");
Expect(1, 6655, '\p{Sc=-	Khmr}', "");
Expect(0, 6655, '\p{^Sc=-	Khmr}', "");
Expect(0, 6655, '\P{Sc=-	Khmr}', "");
Expect(1, 6655, '\P{^Sc=-	Khmr}', "");
Expect(0, 6656, '\p{Sc=-	Khmr}', "");
Expect(1, 6656, '\p{^Sc=-	Khmr}', "");
Expect(1, 6656, '\P{Sc=-	Khmr}', "");
Expect(0, 6656, '\P{^Sc=-	Khmr}', "");
Error('\p{Is_Script=	/a/khmer}');
Error('\P{Is_Script=	/a/khmer}');
Expect(1, 6655, '\p{Is_Script=khmer}', "");
Expect(0, 6655, '\p{^Is_Script=khmer}', "");
Expect(0, 6655, '\P{Is_Script=khmer}', "");
Expect(1, 6655, '\P{^Is_Script=khmer}', "");
Expect(0, 6656, '\p{Is_Script=khmer}', "");
Expect(1, 6656, '\p{^Is_Script=khmer}', "");
Expect(1, 6656, '\P{Is_Script=khmer}', "");
Expect(0, 6656, '\P{^Is_Script=khmer}', "");
Expect(1, 6655, '\p{Is_Script=-khmer}', "");
Expect(0, 6655, '\p{^Is_Script=-khmer}', "");
Expect(0, 6655, '\P{Is_Script=-khmer}', "");
Expect(1, 6655, '\P{^Is_Script=-khmer}', "");
Expect(0, 6656, '\p{Is_Script=-khmer}', "");
Expect(1, 6656, '\p{^Is_Script=-khmer}', "");
Expect(1, 6656, '\P{Is_Script=-khmer}', "");
Expect(0, 6656, '\P{^Is_Script=-khmer}', "");
Error('\p{Is_Sc=:=-khmr}');
Error('\P{Is_Sc=:=-khmr}');
Expect(1, 6655, '\p{Is_Sc=khmr}', "");
Expect(0, 6655, '\p{^Is_Sc=khmr}', "");
Expect(0, 6655, '\P{Is_Sc=khmr}', "");
Expect(1, 6655, '\P{^Is_Sc=khmr}', "");
Expect(0, 6656, '\p{Is_Sc=khmr}', "");
Expect(1, 6656, '\p{^Is_Sc=khmr}', "");
Expect(1, 6656, '\P{Is_Sc=khmr}', "");
Expect(0, 6656, '\P{^Is_Sc=khmr}', "");
Expect(1, 6655, '\p{Is_Sc=-khmr}', "");
Expect(0, 6655, '\p{^Is_Sc=-khmr}', "");
Expect(0, 6655, '\P{Is_Sc=-khmr}', "");
Expect(1, 6655, '\P{^Is_Sc=-khmr}', "");
Expect(0, 6656, '\p{Is_Sc=-khmr}', "");
Expect(1, 6656, '\p{^Is_Sc=-khmr}', "");
Expect(1, 6656, '\P{Is_Sc=-khmr}', "");
Expect(0, 6656, '\P{^Is_Sc=-khmr}', "");
Error('\p{Script:	_/a/CHAM}');
Error('\P{Script:	_/a/CHAM}');
Expect(1, 43615, '\p{Script=cham}', "");
Expect(0, 43615, '\p{^Script=cham}', "");
Expect(0, 43615, '\P{Script=cham}', "");
Expect(1, 43615, '\P{^Script=cham}', "");
Expect(0, 43616, '\p{Script=cham}', "");
Expect(1, 43616, '\p{^Script=cham}', "");
Expect(1, 43616, '\P{Script=cham}', "");
Expect(0, 43616, '\P{^Script=cham}', "");
Expect(1, 43615, '\p{Script=	cham}', "");
Expect(0, 43615, '\p{^Script=	cham}', "");
Expect(0, 43615, '\P{Script=	cham}', "");
Expect(1, 43615, '\P{^Script=	cham}', "");
Expect(0, 43616, '\p{Script=	cham}', "");
Expect(1, 43616, '\p{^Script=	cham}', "");
Expect(1, 43616, '\P{Script=	cham}', "");
Expect(0, 43616, '\P{^Script=	cham}', "");
Error('\p{Sc=__CHAM/a/}');
Error('\P{Sc=__CHAM/a/}');
Expect(1, 43615, '\p{Sc:cham}', "");
Expect(0, 43615, '\p{^Sc:cham}', "");
Expect(0, 43615, '\P{Sc:cham}', "");
Expect(1, 43615, '\P{^Sc:cham}', "");
Expect(0, 43616, '\p{Sc:cham}', "");
Expect(1, 43616, '\p{^Sc:cham}', "");
Expect(1, 43616, '\P{Sc:cham}', "");
Expect(0, 43616, '\P{^Sc:cham}', "");
Expect(1, 43615, '\p{Sc=-CHAM}', "");
Expect(0, 43615, '\p{^Sc=-CHAM}', "");
Expect(0, 43615, '\P{Sc=-CHAM}', "");
Expect(1, 43615, '\P{^Sc=-CHAM}', "");
Expect(0, 43616, '\p{Sc=-CHAM}', "");
Expect(1, 43616, '\p{^Sc=-CHAM}', "");
Expect(1, 43616, '\P{Sc=-CHAM}', "");
Expect(0, 43616, '\P{^Sc=-CHAM}', "");
Error('\p{Is_Script=	CHAM:=}');
Error('\P{Is_Script=	CHAM:=}');
Expect(1, 43615, '\p{Is_Script=cham}', "");
Expect(0, 43615, '\p{^Is_Script=cham}', "");
Expect(0, 43615, '\P{Is_Script=cham}', "");
Expect(1, 43615, '\P{^Is_Script=cham}', "");
Expect(0, 43616, '\p{Is_Script=cham}', "");
Expect(1, 43616, '\p{^Is_Script=cham}', "");
Expect(1, 43616, '\P{Is_Script=cham}', "");
Expect(0, 43616, '\P{^Is_Script=cham}', "");
Expect(1, 43615, '\p{Is_Script= 	Cham}', "");
Expect(0, 43615, '\p{^Is_Script= 	Cham}', "");
Expect(0, 43615, '\P{Is_Script= 	Cham}', "");
Expect(1, 43615, '\P{^Is_Script= 	Cham}', "");
Expect(0, 43616, '\p{Is_Script= 	Cham}', "");
Expect(1, 43616, '\p{^Is_Script= 	Cham}', "");
Expect(1, 43616, '\P{Is_Script= 	Cham}', "");
Expect(0, 43616, '\P{^Is_Script= 	Cham}', "");
Error('\p{Is_Sc=:= 	cham}');
Error('\P{Is_Sc=:= 	cham}');
Expect(1, 43615, '\p{Is_Sc=cham}', "");
Expect(0, 43615, '\p{^Is_Sc=cham}', "");
Expect(0, 43615, '\P{Is_Sc=cham}', "");
Expect(1, 43615, '\P{^Is_Sc=cham}', "");
Expect(0, 43616, '\p{Is_Sc=cham}', "");
Expect(1, 43616, '\p{^Is_Sc=cham}', "");
Expect(1, 43616, '\P{Is_Sc=cham}', "");
Expect(0, 43616, '\P{^Is_Sc=cham}', "");
Error('\p{Script=:=_Tibetan}');
Error('\P{Script=:=_Tibetan}');
Expect(1, 4058, '\p{Script=tibetan}', "");
Expect(0, 4058, '\p{^Script=tibetan}', "");
Expect(0, 4058, '\P{Script=tibetan}', "");
Expect(1, 4058, '\P{^Script=tibetan}', "");
Expect(0, 4059, '\p{Script=tibetan}', "");
Expect(1, 4059, '\p{^Script=tibetan}', "");
Expect(1, 4059, '\P{Script=tibetan}', "");
Expect(0, 4059, '\P{^Script=tibetan}', "");
Expect(1, 4058, '\p{Script=-Tibetan}', "");
Expect(0, 4058, '\p{^Script=-Tibetan}', "");
Expect(0, 4058, '\P{Script=-Tibetan}', "");
Expect(1, 4058, '\P{^Script=-Tibetan}', "");
Expect(0, 4059, '\p{Script=-Tibetan}', "");
Expect(1, 4059, '\p{^Script=-Tibetan}', "");
Expect(1, 4059, '\P{Script=-Tibetan}', "");
Expect(0, 4059, '\P{^Script=-Tibetan}', "");
Error('\p{Sc=_ tibt:=}');
Error('\P{Sc=_ tibt:=}');
Expect(1, 4058, '\p{Sc=tibt}', "");
Expect(0, 4058, '\p{^Sc=tibt}', "");
Expect(0, 4058, '\P{Sc=tibt}', "");
Expect(1, 4058, '\P{^Sc=tibt}', "");
Expect(0, 4059, '\p{Sc=tibt}', "");
Expect(1, 4059, '\p{^Sc=tibt}', "");
Expect(1, 4059, '\P{Sc=tibt}', "");
Expect(0, 4059, '\P{^Sc=tibt}', "");
Expect(1, 4058, '\p{Sc=	_Tibt}', "");
Expect(0, 4058, '\p{^Sc=	_Tibt}', "");
Expect(0, 4058, '\P{Sc=	_Tibt}', "");
Expect(1, 4058, '\P{^Sc=	_Tibt}', "");
Expect(0, 4059, '\p{Sc=	_Tibt}', "");
Expect(1, 4059, '\p{^Sc=	_Tibt}', "");
Expect(1, 4059, '\P{Sc=	_Tibt}', "");
Expect(0, 4059, '\P{^Sc=	_Tibt}', "");
Error('\p{Is_Script=:=Tibetan}');
Error('\P{Is_Script=:=Tibetan}');
Expect(1, 4058, '\p{Is_Script=tibetan}', "");
Expect(0, 4058, '\p{^Is_Script=tibetan}', "");
Expect(0, 4058, '\P{Is_Script=tibetan}', "");
Expect(1, 4058, '\P{^Is_Script=tibetan}', "");
Expect(0, 4059, '\p{Is_Script=tibetan}', "");
Expect(1, 4059, '\p{^Is_Script=tibetan}', "");
Expect(1, 4059, '\P{Is_Script=tibetan}', "");
Expect(0, 4059, '\P{^Is_Script=tibetan}', "");
Expect(1, 4058, '\p{Is_Script:    	Tibetan}', "");
Expect(0, 4058, '\p{^Is_Script:    	Tibetan}', "");
Expect(0, 4058, '\P{Is_Script:    	Tibetan}', "");
Expect(1, 4058, '\P{^Is_Script:    	Tibetan}', "");
Expect(0, 4059, '\p{Is_Script:    	Tibetan}', "");
Expect(1, 4059, '\p{^Is_Script:    	Tibetan}', "");
Expect(1, 4059, '\P{Is_Script:    	Tibetan}', "");
Expect(0, 4059, '\P{^Is_Script:    	Tibetan}', "");
Error('\p{Is_Sc=/a/Tibt}');
Error('\P{Is_Sc=/a/Tibt}');
Expect(1, 4058, '\p{Is_Sc=tibt}', "");
Expect(0, 4058, '\p{^Is_Sc=tibt}', "");
Expect(0, 4058, '\P{Is_Sc=tibt}', "");
Expect(1, 4058, '\P{^Is_Sc=tibt}', "");
Expect(0, 4059, '\p{Is_Sc=tibt}', "");
Expect(1, 4059, '\p{^Is_Sc=tibt}', "");
Expect(1, 4059, '\P{Is_Sc=tibt}', "");
Expect(0, 4059, '\P{^Is_Sc=tibt}', "");
Expect(1, 4058, '\p{Is_Sc= Tibt}', "");
Expect(0, 4058, '\p{^Is_Sc= Tibt}', "");
Expect(0, 4058, '\P{Is_Sc= Tibt}', "");
Expect(1, 4058, '\P{^Is_Sc= Tibt}', "");
Expect(0, 4059, '\p{Is_Sc= Tibt}', "");
Expect(1, 4059, '\p{^Is_Sc= Tibt}', "");
Expect(1, 4059, '\P{Is_Sc= Tibt}', "");
Expect(0, 4059, '\P{^Is_Sc= Tibt}', "");
Error('\p{Script: 	 Ogham:=}');
Error('\P{Script: 	 Ogham:=}');
Expect(1, 5788, '\p{Script=ogham}', "");
Expect(0, 5788, '\p{^Script=ogham}', "");
Expect(0, 5788, '\P{Script=ogham}', "");
Expect(1, 5788, '\P{^Script=ogham}', "");
Expect(0, 5789, '\p{Script=ogham}', "");
Expect(1, 5789, '\p{^Script=ogham}', "");
Expect(1, 5789, '\P{Script=ogham}', "");
Expect(0, 5789, '\P{^Script=ogham}', "");
Expect(1, 5788, '\p{Script:	_OGHAM}', "");
Expect(0, 5788, '\p{^Script:	_OGHAM}', "");
Expect(0, 5788, '\P{Script:	_OGHAM}', "");
Expect(1, 5788, '\P{^Script:	_OGHAM}', "");
Expect(0, 5789, '\p{Script:	_OGHAM}', "");
Expect(1, 5789, '\p{^Script:	_OGHAM}', "");
Expect(1, 5789, '\P{Script:	_OGHAM}', "");
Expect(0, 5789, '\P{^Script:	_OGHAM}', "");
Error('\p{Sc=:=- Ogam}');
Error('\P{Sc=:=- Ogam}');
Expect(1, 5788, '\p{Sc=ogam}', "");
Expect(0, 5788, '\p{^Sc=ogam}', "");
Expect(0, 5788, '\P{Sc=ogam}', "");
Expect(1, 5788, '\P{^Sc=ogam}', "");
Expect(0, 5789, '\p{Sc=ogam}', "");
Expect(1, 5789, '\p{^Sc=ogam}', "");
Expect(1, 5789, '\P{Sc=ogam}', "");
Expect(0, 5789, '\P{^Sc=ogam}', "");
Expect(1, 5788, '\p{Sc= -Ogam}', "");
Expect(0, 5788, '\p{^Sc= -Ogam}', "");
Expect(0, 5788, '\P{Sc= -Ogam}', "");
Expect(1, 5788, '\P{^Sc= -Ogam}', "");
Expect(0, 5789, '\p{Sc= -Ogam}', "");
Expect(1, 5789, '\p{^Sc= -Ogam}', "");
Expect(1, 5789, '\P{Sc= -Ogam}', "");
Expect(0, 5789, '\P{^Sc= -Ogam}', "");
Error('\p{Is_Script=ogham/a/}');
Error('\P{Is_Script=ogham/a/}');
Expect(1, 5788, '\p{Is_Script=ogham}', "");
Expect(0, 5788, '\p{^Is_Script=ogham}', "");
Expect(0, 5788, '\P{Is_Script=ogham}', "");
Expect(1, 5788, '\P{^Is_Script=ogham}', "");
Expect(0, 5789, '\p{Is_Script=ogham}', "");
Expect(1, 5789, '\p{^Is_Script=ogham}', "");
Expect(1, 5789, '\P{Is_Script=ogham}', "");
Expect(0, 5789, '\P{^Is_Script=ogham}', "");
Expect(1, 5788, '\p{Is_Script=--ogham}', "");
Expect(0, 5788, '\p{^Is_Script=--ogham}', "");
Expect(0, 5788, '\P{Is_Script=--ogham}', "");
Expect(1, 5788, '\P{^Is_Script=--ogham}', "");
Expect(0, 5789, '\p{Is_Script=--ogham}', "");
Expect(1, 5789, '\p{^Is_Script=--ogham}', "");
Expect(1, 5789, '\P{Is_Script=--ogham}', "");
Expect(0, 5789, '\P{^Is_Script=--ogham}', "");
Error('\p{Is_Sc=/a/_-ogam}');
Error('\P{Is_Sc=/a/_-ogam}');
Expect(1, 5788, '\p{Is_Sc=ogam}', "");
Expect(0, 5788, '\p{^Is_Sc=ogam}', "");
Expect(0, 5788, '\P{Is_Sc=ogam}', "");
Expect(1, 5788, '\P{^Is_Sc=ogam}', "");
Expect(0, 5789, '\p{Is_Sc=ogam}', "");
Expect(1, 5789, '\p{^Is_Sc=ogam}', "");
Expect(1, 5789, '\P{Is_Sc=ogam}', "");
Expect(0, 5789, '\P{^Is_Sc=ogam}', "");
Expect(1, 5788, '\p{Is_Sc=-_Ogam}', "");
Expect(0, 5788, '\p{^Is_Sc=-_Ogam}', "");
Expect(0, 5788, '\P{Is_Sc=-_Ogam}', "");
Expect(1, 5788, '\P{^Is_Sc=-_Ogam}', "");
Expect(0, 5789, '\p{Is_Sc=-_Ogam}', "");
Expect(1, 5789, '\p{^Is_Sc=-_Ogam}', "");
Expect(1, 5789, '\P{Is_Sc=-_Ogam}', "");
Expect(0, 5789, '\P{^Is_Sc=-_Ogam}', "");
Error('\p{Script=_/a/TAI_VIET}');
Error('\P{Script=_/a/TAI_VIET}');
Expect(1, 43743, '\p{Script=taiviet}', "");
Expect(0, 43743, '\p{^Script=taiviet}', "");
Expect(0, 43743, '\P{Script=taiviet}', "");
Expect(1, 43743, '\P{^Script=taiviet}', "");
Expect(0, 43744, '\p{Script=taiviet}', "");
Expect(1, 43744, '\p{^Script=taiviet}', "");
Expect(1, 43744, '\P{Script=taiviet}', "");
Expect(0, 43744, '\P{^Script=taiviet}', "");
Expect(1, 43743, '\p{Script=		Tai_VIET}', "");
Expect(0, 43743, '\p{^Script=		Tai_VIET}', "");
Expect(0, 43743, '\P{Script=		Tai_VIET}', "");
Expect(1, 43743, '\P{^Script=		Tai_VIET}', "");
Expect(0, 43744, '\p{Script=		Tai_VIET}', "");
Expect(1, 43744, '\p{^Script=		Tai_VIET}', "");
Expect(1, 43744, '\P{Script=		Tai_VIET}', "");
Expect(0, 43744, '\P{^Script=		Tai_VIET}', "");
Error('\p{Sc=/a/ tavt}');
Error('\P{Sc=/a/ tavt}');
Expect(1, 43743, '\p{Sc=tavt}', "");
Expect(0, 43743, '\p{^Sc=tavt}', "");
Expect(0, 43743, '\P{Sc=tavt}', "");
Expect(1, 43743, '\P{^Sc=tavt}', "");
Expect(0, 43744, '\p{Sc=tavt}', "");
Expect(1, 43744, '\p{^Sc=tavt}', "");
Expect(1, 43744, '\P{Sc=tavt}', "");
Expect(0, 43744, '\P{^Sc=tavt}', "");
Expect(1, 43743, '\p{Sc=-Tavt}', "");
Expect(0, 43743, '\p{^Sc=-Tavt}', "");
Expect(0, 43743, '\P{Sc=-Tavt}', "");
Expect(1, 43743, '\P{^Sc=-Tavt}', "");
Expect(0, 43744, '\p{Sc=-Tavt}', "");
Expect(1, 43744, '\p{^Sc=-Tavt}', "");
Expect(1, 43744, '\P{Sc=-Tavt}', "");
Expect(0, 43744, '\P{^Sc=-Tavt}', "");
Error('\p{Is_Script=/a/ 	Tai_viet}');
Error('\P{Is_Script=/a/ 	Tai_viet}');
Expect(1, 43743, '\p{Is_Script=taiviet}', "");
Expect(0, 43743, '\p{^Is_Script=taiviet}', "");
Expect(0, 43743, '\P{Is_Script=taiviet}', "");
Expect(1, 43743, '\P{^Is_Script=taiviet}', "");
Expect(0, 43744, '\p{Is_Script=taiviet}', "");
Expect(1, 43744, '\p{^Is_Script=taiviet}', "");
Expect(1, 43744, '\P{Is_Script=taiviet}', "");
Expect(0, 43744, '\P{^Is_Script=taiviet}', "");
Expect(1, 43743, '\p{Is_Script=	Tai_Viet}', "");
Expect(0, 43743, '\p{^Is_Script=	Tai_Viet}', "");
Expect(0, 43743, '\P{Is_Script=	Tai_Viet}', "");
Expect(1, 43743, '\P{^Is_Script=	Tai_Viet}', "");
Expect(0, 43744, '\p{Is_Script=	Tai_Viet}', "");
Expect(1, 43744, '\p{^Is_Script=	Tai_Viet}', "");
Expect(1, 43744, '\P{Is_Script=	Tai_Viet}', "");
Expect(0, 43744, '\P{^Is_Script=	Tai_Viet}', "");
Error('\p{Is_Sc=/a/-	Tavt}');
Error('\P{Is_Sc=/a/-	Tavt}');
Expect(1, 43743, '\p{Is_Sc=tavt}', "");
Expect(0, 43743, '\p{^Is_Sc=tavt}', "");
Expect(0, 43743, '\P{Is_Sc=tavt}', "");
Expect(1, 43743, '\P{^Is_Sc=tavt}', "");
Expect(0, 43744, '\p{Is_Sc=tavt}', "");
Expect(1, 43744, '\p{^Is_Sc=tavt}', "");
Expect(1, 43744, '\P{Is_Sc=tavt}', "");
Expect(0, 43744, '\P{^Is_Sc=tavt}', "");
Expect(1, 43743, '\p{Is_Sc=-tavt}', "");
Expect(0, 43743, '\p{^Is_Sc=-tavt}', "");
Expect(0, 43743, '\P{Is_Sc=-tavt}', "");
Expect(1, 43743, '\P{^Is_Sc=-tavt}', "");
Expect(0, 43744, '\p{Is_Sc=-tavt}', "");
Expect(1, 43744, '\p{^Is_Sc=-tavt}', "");
Expect(1, 43744, '\P{Is_Sc=-tavt}', "");
Expect(0, 43744, '\P{^Is_Sc=-tavt}', "");
Error('\p{Script=:=MALAYALAM}');
Error('\P{Script=:=MALAYALAM}');
Expect(1, 3455, '\p{Script:   malayalam}', "");
Expect(0, 3455, '\p{^Script:   malayalam}', "");
Expect(0, 3455, '\P{Script:   malayalam}', "");
Expect(1, 3455, '\P{^Script:   malayalam}', "");
Expect(0, 3456, '\p{Script:   malayalam}', "");
Expect(1, 3456, '\p{^Script:   malayalam}', "");
Expect(1, 3456, '\P{Script:   malayalam}', "");
Expect(0, 3456, '\P{^Script:   malayalam}', "");
Expect(1, 3455, '\p{Script=	malayalam}', "");
Expect(0, 3455, '\p{^Script=	malayalam}', "");
Expect(0, 3455, '\P{Script=	malayalam}', "");
Expect(1, 3455, '\P{^Script=	malayalam}', "");
Expect(0, 3456, '\p{Script=	malayalam}', "");
Expect(1, 3456, '\p{^Script=	malayalam}', "");
Expect(1, 3456, '\P{Script=	malayalam}', "");
Expect(0, 3456, '\P{^Script=	malayalam}', "");
Error('\p{Sc=:= -Mlym}');
Error('\P{Sc=:= -Mlym}');
Expect(1, 3455, '\p{Sc:mlym}', "");
Expect(0, 3455, '\p{^Sc:mlym}', "");
Expect(0, 3455, '\P{Sc:mlym}', "");
Expect(1, 3455, '\P{^Sc:mlym}', "");
Expect(0, 3456, '\p{Sc:mlym}', "");
Expect(1, 3456, '\p{^Sc:mlym}', "");
Expect(1, 3456, '\P{Sc:mlym}', "");
Expect(0, 3456, '\P{^Sc:mlym}', "");
Expect(1, 3455, '\p{Sc:   		mlym}', "");
Expect(0, 3455, '\p{^Sc:   		mlym}', "");
Expect(0, 3455, '\P{Sc:   		mlym}', "");
Expect(1, 3455, '\P{^Sc:   		mlym}', "");
Expect(0, 3456, '\p{Sc:   		mlym}', "");
Expect(1, 3456, '\p{^Sc:   		mlym}', "");
Expect(1, 3456, '\P{Sc:   		mlym}', "");
Expect(0, 3456, '\P{^Sc:   		mlym}', "");
Error('\p{Is_Script=_:=Malayalam}');
Error('\P{Is_Script=_:=Malayalam}');
Expect(1, 3455, '\p{Is_Script=malayalam}', "");
Expect(0, 3455, '\p{^Is_Script=malayalam}', "");
Expect(0, 3455, '\P{Is_Script=malayalam}', "");
Expect(1, 3455, '\P{^Is_Script=malayalam}', "");
Expect(0, 3456, '\p{Is_Script=malayalam}', "");
Expect(1, 3456, '\p{^Is_Script=malayalam}', "");
Expect(1, 3456, '\P{Is_Script=malayalam}', "");
Expect(0, 3456, '\P{^Is_Script=malayalam}', "");
Expect(1, 3455, '\p{Is_Script=_malayalam}', "");
Expect(0, 3455, '\p{^Is_Script=_malayalam}', "");
Expect(0, 3455, '\P{Is_Script=_malayalam}', "");
Expect(1, 3455, '\P{^Is_Script=_malayalam}', "");
Expect(0, 3456, '\p{Is_Script=_malayalam}', "");
Expect(1, 3456, '\p{^Is_Script=_malayalam}', "");
Expect(1, 3456, '\P{Is_Script=_malayalam}', "");
Expect(0, 3456, '\P{^Is_Script=_malayalam}', "");
Error('\p{Is_Sc=--Mlym/a/}');
Error('\P{Is_Sc=--Mlym/a/}');
Expect(1, 3455, '\p{Is_Sc=mlym}', "");
Expect(0, 3455, '\p{^Is_Sc=mlym}', "");
Expect(0, 3455, '\P{Is_Sc=mlym}', "");
Expect(1, 3455, '\P{^Is_Sc=mlym}', "");
Expect(0, 3456, '\p{Is_Sc=mlym}', "");
Expect(1, 3456, '\p{^Is_Sc=mlym}', "");
Expect(1, 3456, '\P{Is_Sc=mlym}', "");
Expect(0, 3456, '\P{^Is_Sc=mlym}', "");
Expect(1, 3455, '\p{Is_Sc= _Mlym}', "");
Expect(0, 3455, '\p{^Is_Sc= _Mlym}', "");
Expect(0, 3455, '\P{Is_Sc= _Mlym}', "");
Expect(1, 3455, '\P{^Is_Sc= _Mlym}', "");
Expect(0, 3456, '\p{Is_Sc= _Mlym}', "");
Expect(1, 3456, '\p{^Is_Sc= _Mlym}', "");
Expect(1, 3456, '\P{Is_Sc= _Mlym}', "");
Expect(0, 3456, '\P{^Is_Sc= _Mlym}', "");
Error('\p{Script=-:=Old_Persian}');
Error('\P{Script=-:=Old_Persian}');
Expect(1, 66517, '\p{Script=oldpersian}', "");
Expect(0, 66517, '\p{^Script=oldpersian}', "");
Expect(0, 66517, '\P{Script=oldpersian}', "");
Expect(1, 66517, '\P{^Script=oldpersian}', "");
Expect(0, 66518, '\p{Script=oldpersian}', "");
Expect(1, 66518, '\p{^Script=oldpersian}', "");
Expect(1, 66518, '\P{Script=oldpersian}', "");
Expect(0, 66518, '\P{^Script=oldpersian}', "");
Expect(1, 66517, '\p{Script=-_Old_Persian}', "");
Expect(0, 66517, '\p{^Script=-_Old_Persian}', "");
Expect(0, 66517, '\P{Script=-_Old_Persian}', "");
Expect(1, 66517, '\P{^Script=-_Old_Persian}', "");
Expect(0, 66518, '\p{Script=-_Old_Persian}', "");
Expect(1, 66518, '\p{^Script=-_Old_Persian}', "");
Expect(1, 66518, '\P{Script=-_Old_Persian}', "");
Expect(0, 66518, '\P{^Script=-_Old_Persian}', "");
Error('\p{Sc=/a/XPEO}');
Error('\P{Sc=/a/XPEO}');
Expect(1, 66517, '\p{Sc=xpeo}', "");
Expect(0, 66517, '\p{^Sc=xpeo}', "");
Expect(0, 66517, '\P{Sc=xpeo}', "");
Expect(1, 66517, '\P{^Sc=xpeo}', "");
Expect(0, 66518, '\p{Sc=xpeo}', "");
Expect(1, 66518, '\p{^Sc=xpeo}', "");
Expect(1, 66518, '\P{Sc=xpeo}', "");
Expect(0, 66518, '\P{^Sc=xpeo}', "");
Expect(1, 66517, '\p{Sc=		XPEO}', "");
Expect(0, 66517, '\p{^Sc=		XPEO}', "");
Expect(0, 66517, '\P{Sc=		XPEO}', "");
Expect(1, 66517, '\P{^Sc=		XPEO}', "");
Expect(0, 66518, '\p{Sc=		XPEO}', "");
Expect(1, 66518, '\p{^Sc=		XPEO}', "");
Expect(1, 66518, '\P{Sc=		XPEO}', "");
Expect(0, 66518, '\P{^Sc=		XPEO}', "");
Error('\p{Is_Script=:= -old_PERSIAN}');
Error('\P{Is_Script=:= -old_PERSIAN}');
Expect(1, 66517, '\p{Is_Script=oldpersian}', "");
Expect(0, 66517, '\p{^Is_Script=oldpersian}', "");
Expect(0, 66517, '\P{Is_Script=oldpersian}', "");
Expect(1, 66517, '\P{^Is_Script=oldpersian}', "");
Expect(0, 66518, '\p{Is_Script=oldpersian}', "");
Expect(1, 66518, '\p{^Is_Script=oldpersian}', "");
Expect(1, 66518, '\P{Is_Script=oldpersian}', "");
Expect(0, 66518, '\P{^Is_Script=oldpersian}', "");
Expect(1, 66517, '\p{Is_Script=	Old_PERSIAN}', "");
Expect(0, 66517, '\p{^Is_Script=	Old_PERSIAN}', "");
Expect(0, 66517, '\P{Is_Script=	Old_PERSIAN}', "");
Expect(1, 66517, '\P{^Is_Script=	Old_PERSIAN}', "");
Expect(0, 66518, '\p{Is_Script=	Old_PERSIAN}', "");
Expect(1, 66518, '\p{^Is_Script=	Old_PERSIAN}', "");
Expect(1, 66518, '\P{Is_Script=	Old_PERSIAN}', "");
Expect(0, 66518, '\P{^Is_Script=	Old_PERSIAN}', "");
Error('\p{Is_Sc=/a/_Xpeo}');
Error('\P{Is_Sc=/a/_Xpeo}');
Expect(1, 66517, '\p{Is_Sc=xpeo}', "");
Expect(0, 66517, '\p{^Is_Sc=xpeo}', "");
Expect(0, 66517, '\P{Is_Sc=xpeo}', "");
Expect(1, 66517, '\P{^Is_Sc=xpeo}', "");
Expect(0, 66518, '\p{Is_Sc=xpeo}', "");
Expect(1, 66518, '\p{^Is_Sc=xpeo}', "");
Expect(1, 66518, '\P{Is_Sc=xpeo}', "");
Expect(0, 66518, '\P{^Is_Sc=xpeo}', "");
Expect(1, 66517, '\p{Is_Sc= Xpeo}', "");
Expect(0, 66517, '\p{^Is_Sc= Xpeo}', "");
Expect(0, 66517, '\P{Is_Sc= Xpeo}', "");
Expect(1, 66517, '\P{^Is_Sc= Xpeo}', "");
Expect(0, 66518, '\p{Is_Sc= Xpeo}', "");
Expect(1, 66518, '\p{^Is_Sc= Xpeo}', "");
Expect(1, 66518, '\P{Is_Sc= Xpeo}', "");
Expect(0, 66518, '\P{^Is_Sc= Xpeo}', "");
Error('\p{Script=_ MYANMAR/a/}');
Error('\P{Script=_ MYANMAR/a/}');
Expect(1, 43643, '\p{Script=myanmar}', "");
Expect(0, 43643, '\p{^Script=myanmar}', "");
Expect(0, 43643, '\P{Script=myanmar}', "");
Expect(1, 43643, '\P{^Script=myanmar}', "");
Expect(0, 43644, '\p{Script=myanmar}', "");
Expect(1, 43644, '\p{^Script=myanmar}', "");
Expect(1, 43644, '\P{Script=myanmar}', "");
Expect(0, 43644, '\P{^Script=myanmar}', "");
Expect(1, 43643, '\p{Script=__Myanmar}', "");
Expect(0, 43643, '\p{^Script=__Myanmar}', "");
Expect(0, 43643, '\P{Script=__Myanmar}', "");
Expect(1, 43643, '\P{^Script=__Myanmar}', "");
Expect(0, 43644, '\p{Script=__Myanmar}', "");
Expect(1, 43644, '\p{^Script=__Myanmar}', "");
Expect(1, 43644, '\P{Script=__Myanmar}', "");
Expect(0, 43644, '\P{^Script=__Myanmar}', "");
Error('\p{Sc=	/a/MYMR}');
Error('\P{Sc=	/a/MYMR}');
Expect(1, 43643, '\p{Sc=mymr}', "");
Expect(0, 43643, '\p{^Sc=mymr}', "");
Expect(0, 43643, '\P{Sc=mymr}', "");
Expect(1, 43643, '\P{^Sc=mymr}', "");
Expect(0, 43644, '\p{Sc=mymr}', "");
Expect(1, 43644, '\p{^Sc=mymr}', "");
Expect(1, 43644, '\P{Sc=mymr}', "");
Expect(0, 43644, '\P{^Sc=mymr}', "");
Expect(1, 43643, '\p{Sc= 	Mymr}', "");
Expect(0, 43643, '\p{^Sc= 	Mymr}', "");
Expect(0, 43643, '\P{Sc= 	Mymr}', "");
Expect(1, 43643, '\P{^Sc= 	Mymr}', "");
Expect(0, 43644, '\p{Sc= 	Mymr}', "");
Expect(1, 43644, '\p{^Sc= 	Mymr}', "");
Expect(1, 43644, '\P{Sc= 	Mymr}', "");
Expect(0, 43644, '\P{^Sc= 	Mymr}', "");
Error('\p{Is_Script= Myanmar/a/}');
Error('\P{Is_Script= Myanmar/a/}');
Expect(1, 43643, '\p{Is_Script=myanmar}', "");
Expect(0, 43643, '\p{^Is_Script=myanmar}', "");
Expect(0, 43643, '\P{Is_Script=myanmar}', "");
Expect(1, 43643, '\P{^Is_Script=myanmar}', "");
Expect(0, 43644, '\p{Is_Script=myanmar}', "");
Expect(1, 43644, '\p{^Is_Script=myanmar}', "");
Expect(1, 43644, '\P{Is_Script=myanmar}', "");
Expect(0, 43644, '\P{^Is_Script=myanmar}', "");
Error('\p{Is_Sc=	_Mymr/a/}');
Error('\P{Is_Sc=	_Mymr/a/}');
Expect(1, 43643, '\p{Is_Sc=mymr}', "");
Expect(0, 43643, '\p{^Is_Sc=mymr}', "");
Expect(0, 43643, '\P{Is_Sc=mymr}', "");
Expect(1, 43643, '\P{^Is_Sc=mymr}', "");
Expect(0, 43644, '\p{Is_Sc=mymr}', "");
Expect(1, 43644, '\p{^Is_Sc=mymr}', "");
Expect(1, 43644, '\P{Is_Sc=mymr}', "");
Expect(0, 43644, '\P{^Is_Sc=mymr}', "");
Expect(1, 43643, '\p{Is_Sc= mymr}', "");
Expect(0, 43643, '\p{^Is_Sc= mymr}', "");
Expect(0, 43643, '\P{Is_Sc= mymr}', "");
Expect(1, 43643, '\P{^Is_Sc= mymr}', "");
Expect(0, 43644, '\p{Is_Sc= mymr}', "");
Expect(1, 43644, '\p{^Is_Sc= mymr}', "");
Expect(1, 43644, '\P{Is_Sc= mymr}', "");
Expect(0, 43644, '\P{^Is_Sc= mymr}', "");
Error('\p{Script=/a/	-Old_turkic}');
Error('\P{Script=/a/	-Old_turkic}');
Expect(1, 68680, '\p{Script=oldturkic}', "");
Expect(0, 68680, '\p{^Script=oldturkic}', "");
Expect(0, 68680, '\P{Script=oldturkic}', "");
Expect(1, 68680, '\P{^Script=oldturkic}', "");
Expect(0, 68681, '\p{Script=oldturkic}', "");
Expect(1, 68681, '\p{^Script=oldturkic}', "");
Expect(1, 68681, '\P{Script=oldturkic}', "");
Expect(0, 68681, '\P{^Script=oldturkic}', "");
Expect(1, 68680, '\p{Script=- Old_turkic}', "");
Expect(0, 68680, '\p{^Script=- Old_turkic}', "");
Expect(0, 68680, '\P{Script=- Old_turkic}', "");
Expect(1, 68680, '\P{^Script=- Old_turkic}', "");
Expect(0, 68681, '\p{Script=- Old_turkic}', "");
Expect(1, 68681, '\p{^Script=- Old_turkic}', "");
Expect(1, 68681, '\P{Script=- Old_turkic}', "");
Expect(0, 68681, '\P{^Script=- Old_turkic}', "");
Error('\p{Sc=_ORKH:=}');
Error('\P{Sc=_ORKH:=}');
Expect(1, 68680, '\p{Sc=orkh}', "");
Expect(0, 68680, '\p{^Sc=orkh}', "");
Expect(0, 68680, '\P{Sc=orkh}', "");
Expect(1, 68680, '\P{^Sc=orkh}', "");
Expect(0, 68681, '\p{Sc=orkh}', "");
Expect(1, 68681, '\p{^Sc=orkh}', "");
Expect(1, 68681, '\P{Sc=orkh}', "");
Expect(0, 68681, '\P{^Sc=orkh}', "");
Expect(1, 68680, '\p{Sc=__orkh}', "");
Expect(0, 68680, '\p{^Sc=__orkh}', "");
Expect(0, 68680, '\P{Sc=__orkh}', "");
Expect(1, 68680, '\P{^Sc=__orkh}', "");
Expect(0, 68681, '\p{Sc=__orkh}', "");
Expect(1, 68681, '\p{^Sc=__orkh}', "");
Expect(1, 68681, '\P{Sc=__orkh}', "");
Expect(0, 68681, '\P{^Sc=__orkh}', "");
Error('\p{Is_Script=- OLD_TURKIC:=}');
Error('\P{Is_Script=- OLD_TURKIC:=}');
Expect(1, 68680, '\p{Is_Script=oldturkic}', "");
Expect(0, 68680, '\p{^Is_Script=oldturkic}', "");
Expect(0, 68680, '\P{Is_Script=oldturkic}', "");
Expect(1, 68680, '\P{^Is_Script=oldturkic}', "");
Expect(0, 68681, '\p{Is_Script=oldturkic}', "");
Expect(1, 68681, '\p{^Is_Script=oldturkic}', "");
Expect(1, 68681, '\P{Is_Script=oldturkic}', "");
Expect(0, 68681, '\P{^Is_Script=oldturkic}', "");
Expect(1, 68680, '\p{Is_Script=-old_Turkic}', "");
Expect(0, 68680, '\p{^Is_Script=-old_Turkic}', "");
Expect(0, 68680, '\P{Is_Script=-old_Turkic}', "");
Expect(1, 68680, '\P{^Is_Script=-old_Turkic}', "");
Expect(0, 68681, '\p{Is_Script=-old_Turkic}', "");
Expect(1, 68681, '\p{^Is_Script=-old_Turkic}', "");
Expect(1, 68681, '\P{Is_Script=-old_Turkic}', "");
Expect(0, 68681, '\P{^Is_Script=-old_Turkic}', "");
Error('\p{Is_Sc=_/a/orkh}');
Error('\P{Is_Sc=_/a/orkh}');
Expect(1, 68680, '\p{Is_Sc=orkh}', "");
Expect(0, 68680, '\p{^Is_Sc=orkh}', "");
Expect(0, 68680, '\P{Is_Sc=orkh}', "");
Expect(1, 68680, '\P{^Is_Sc=orkh}', "");
Expect(0, 68681, '\p{Is_Sc=orkh}', "");
Expect(1, 68681, '\p{^Is_Sc=orkh}', "");
Expect(1, 68681, '\P{Is_Sc=orkh}', "");
Expect(0, 68681, '\P{^Is_Sc=orkh}', "");
Expect(1, 68680, '\p{Is_Sc:	 _Orkh}', "");
Expect(0, 68680, '\p{^Is_Sc:	 _Orkh}', "");
Expect(0, 68680, '\P{Is_Sc:	 _Orkh}', "");
Expect(1, 68680, '\P{^Is_Sc:	 _Orkh}', "");
Expect(0, 68681, '\p{Is_Sc:	 _Orkh}', "");
Expect(1, 68681, '\p{^Is_Sc:	 _Orkh}', "");
Expect(1, 68681, '\P{Is_Sc:	 _Orkh}', "");
Expect(0, 68681, '\P{^Is_Sc:	 _Orkh}', "");
Error('\p{Script=OSMANYA/a/}');
Error('\P{Script=OSMANYA/a/}');
Expect(1, 66729, '\p{Script=osmanya}', "");
Expect(0, 66729, '\p{^Script=osmanya}', "");
Expect(0, 66729, '\P{Script=osmanya}', "");
Expect(1, 66729, '\P{^Script=osmanya}', "");
Expect(0, 66730, '\p{Script=osmanya}', "");
Expect(1, 66730, '\p{^Script=osmanya}', "");
Expect(1, 66730, '\P{Script=osmanya}', "");
Expect(0, 66730, '\P{^Script=osmanya}', "");
Expect(1, 66729, '\p{Script= osmanya}', "");
Expect(0, 66729, '\p{^Script= osmanya}', "");
Expect(0, 66729, '\P{Script= osmanya}', "");
Expect(1, 66729, '\P{^Script= osmanya}', "");
Expect(0, 66730, '\p{Script= osmanya}', "");
Expect(1, 66730, '\p{^Script= osmanya}', "");
Expect(1, 66730, '\P{Script= osmanya}', "");
Expect(0, 66730, '\P{^Script= osmanya}', "");
Error('\p{Sc=	_Osma:=}');
Error('\P{Sc=	_Osma:=}');
Expect(1, 66729, '\p{Sc=osma}', "");
Expect(0, 66729, '\p{^Sc=osma}', "");
Expect(0, 66729, '\P{Sc=osma}', "");
Expect(1, 66729, '\P{^Sc=osma}', "");
Expect(0, 66730, '\p{Sc=osma}', "");
Expect(1, 66730, '\p{^Sc=osma}', "");
Expect(1, 66730, '\P{Sc=osma}', "");
Expect(0, 66730, '\P{^Sc=osma}', "");
Expect(1, 66729, '\p{Sc=--OSMA}', "");
Expect(0, 66729, '\p{^Sc=--OSMA}', "");
Expect(0, 66729, '\P{Sc=--OSMA}', "");
Expect(1, 66729, '\P{^Sc=--OSMA}', "");
Expect(0, 66730, '\p{Sc=--OSMA}', "");
Expect(1, 66730, '\p{^Sc=--OSMA}', "");
Expect(1, 66730, '\P{Sc=--OSMA}', "");
Expect(0, 66730, '\P{^Sc=--OSMA}', "");
Error('\p{Is_Script=__Osmanya/a/}');
Error('\P{Is_Script=__Osmanya/a/}');
Expect(1, 66729, '\p{Is_Script=osmanya}', "");
Expect(0, 66729, '\p{^Is_Script=osmanya}', "");
Expect(0, 66729, '\P{Is_Script=osmanya}', "");
Expect(1, 66729, '\P{^Is_Script=osmanya}', "");
Expect(0, 66730, '\p{Is_Script=osmanya}', "");
Expect(1, 66730, '\p{^Is_Script=osmanya}', "");
Expect(1, 66730, '\P{Is_Script=osmanya}', "");
Expect(0, 66730, '\P{^Is_Script=osmanya}', "");
Expect(1, 66729, '\p{Is_Script= Osmanya}', "");
Expect(0, 66729, '\p{^Is_Script= Osmanya}', "");
Expect(0, 66729, '\P{Is_Script= Osmanya}', "");
Expect(1, 66729, '\P{^Is_Script= Osmanya}', "");
Expect(0, 66730, '\p{Is_Script= Osmanya}', "");
Expect(1, 66730, '\p{^Is_Script= Osmanya}', "");
Expect(1, 66730, '\P{Is_Script= Osmanya}', "");
Expect(0, 66730, '\P{^Is_Script= Osmanya}', "");
Error('\p{Is_Sc=/a/	_osma}');
Error('\P{Is_Sc=/a/	_osma}');
Expect(1, 66729, '\p{Is_Sc=osma}', "");
Expect(0, 66729, '\p{^Is_Sc=osma}', "");
Expect(0, 66729, '\P{Is_Sc=osma}', "");
Expect(1, 66729, '\P{^Is_Sc=osma}', "");
Expect(0, 66730, '\p{Is_Sc=osma}', "");
Expect(1, 66730, '\p{^Is_Sc=osma}', "");
Expect(1, 66730, '\P{Is_Sc=osma}', "");
Expect(0, 66730, '\P{^Is_Sc=osma}', "");
Expect(1, 66729, '\p{Is_Sc: OSMA}', "");
Expect(0, 66729, '\p{^Is_Sc: OSMA}', "");
Expect(0, 66729, '\P{Is_Sc: OSMA}', "");
Expect(1, 66729, '\P{^Is_Sc: OSMA}', "");
Expect(0, 66730, '\p{Is_Sc: OSMA}', "");
Expect(1, 66730, '\p{^Is_Sc: OSMA}', "");
Expect(1, 66730, '\P{Is_Sc: OSMA}', "");
Expect(0, 66730, '\P{^Is_Sc: OSMA}', "");
Error('\p{Script:/a/_TAGALOG}');
Error('\P{Script:/a/_TAGALOG}');
Expect(1, 5908, '\p{Script: tagalog}', "");
Expect(0, 5908, '\p{^Script: tagalog}', "");
Expect(0, 5908, '\P{Script: tagalog}', "");
Expect(1, 5908, '\P{^Script: tagalog}', "");
Expect(0, 5909, '\p{Script: tagalog}', "");
Expect(1, 5909, '\p{^Script: tagalog}', "");
Expect(1, 5909, '\P{Script: tagalog}', "");
Expect(0, 5909, '\P{^Script: tagalog}', "");
Expect(1, 5908, '\p{Script= -TAGALOG}', "");
Expect(0, 5908, '\p{^Script= -TAGALOG}', "");
Expect(0, 5908, '\P{Script= -TAGALOG}', "");
Expect(1, 5908, '\P{^Script= -TAGALOG}', "");
Expect(0, 5909, '\p{Script= -TAGALOG}', "");
Expect(1, 5909, '\p{^Script= -TAGALOG}', "");
Expect(1, 5909, '\P{Script= -TAGALOG}', "");
Expect(0, 5909, '\P{^Script= -TAGALOG}', "");
Error('\p{Sc=:= tglg}');
Error('\P{Sc=:= tglg}');
Expect(1, 5908, '\p{Sc=tglg}', "");
Expect(0, 5908, '\p{^Sc=tglg}', "");
Expect(0, 5908, '\P{Sc=tglg}', "");
Expect(1, 5908, '\P{^Sc=tglg}', "");
Expect(0, 5909, '\p{Sc=tglg}', "");
Expect(1, 5909, '\p{^Sc=tglg}', "");
Expect(1, 5909, '\P{Sc=tglg}', "");
Expect(0, 5909, '\P{^Sc=tglg}', "");
Expect(1, 5908, '\p{Sc=_-Tglg}', "");
Expect(0, 5908, '\p{^Sc=_-Tglg}', "");
Expect(0, 5908, '\P{Sc=_-Tglg}', "");
Expect(1, 5908, '\P{^Sc=_-Tglg}', "");
Expect(0, 5909, '\p{Sc=_-Tglg}', "");
Expect(1, 5909, '\p{^Sc=_-Tglg}', "");
Expect(1, 5909, '\P{Sc=_-Tglg}', "");
Expect(0, 5909, '\P{^Sc=_-Tglg}', "");
Error('\p{Is_Script: 	/a/tagalog}');
Error('\P{Is_Script: 	/a/tagalog}');
Expect(1, 5908, '\p{Is_Script=tagalog}', "");
Expect(0, 5908, '\p{^Is_Script=tagalog}', "");
Expect(0, 5908, '\P{Is_Script=tagalog}', "");
Expect(1, 5908, '\P{^Is_Script=tagalog}', "");
Expect(0, 5909, '\p{Is_Script=tagalog}', "");
Expect(1, 5909, '\p{^Is_Script=tagalog}', "");
Expect(1, 5909, '\P{Is_Script=tagalog}', "");
Expect(0, 5909, '\P{^Is_Script=tagalog}', "");
Expect(1, 5908, '\p{Is_Script=-	TAGALOG}', "");
Expect(0, 5908, '\p{^Is_Script=-	TAGALOG}', "");
Expect(0, 5908, '\P{Is_Script=-	TAGALOG}', "");
Expect(1, 5908, '\P{^Is_Script=-	TAGALOG}', "");
Expect(0, 5909, '\p{Is_Script=-	TAGALOG}', "");
Expect(1, 5909, '\p{^Is_Script=-	TAGALOG}', "");
Expect(1, 5909, '\P{Is_Script=-	TAGALOG}', "");
Expect(0, 5909, '\P{^Is_Script=-	TAGALOG}', "");
Error('\p{Is_Sc=:=	TGLG}');
Error('\P{Is_Sc=:=	TGLG}');
Expect(1, 5908, '\p{Is_Sc=tglg}', "");
Expect(0, 5908, '\p{^Is_Sc=tglg}', "");
Expect(0, 5908, '\P{Is_Sc=tglg}', "");
Expect(1, 5908, '\P{^Is_Sc=tglg}', "");
Expect(0, 5909, '\p{Is_Sc=tglg}', "");
Expect(1, 5909, '\p{^Is_Sc=tglg}', "");
Expect(1, 5909, '\P{Is_Sc=tglg}', "");
Expect(0, 5909, '\P{^Is_Sc=tglg}', "");
Expect(1, 5908, '\p{Is_Sc=-_tglg}', "");
Expect(0, 5908, '\p{^Is_Sc=-_tglg}', "");
Expect(0, 5908, '\P{Is_Sc=-_tglg}', "");
Expect(1, 5908, '\P{^Is_Sc=-_tglg}', "");
Expect(0, 5909, '\p{Is_Sc=-_tglg}', "");
Expect(1, 5909, '\p{^Is_Sc=-_tglg}', "");
Expect(1, 5909, '\P{Is_Sc=-_tglg}', "");
Expect(0, 5909, '\P{^Is_Sc=-_tglg}', "");
Error('\p{Script=-:=gujarati}');
Error('\P{Script=-:=gujarati}');
Expect(1, 2801, '\p{Script=gujarati}', "");
Expect(0, 2801, '\p{^Script=gujarati}', "");
Expect(0, 2801, '\P{Script=gujarati}', "");
Expect(1, 2801, '\P{^Script=gujarati}', "");
Expect(0, 2802, '\p{Script=gujarati}', "");
Expect(1, 2802, '\p{^Script=gujarati}', "");
Expect(1, 2802, '\P{Script=gujarati}', "");
Expect(0, 2802, '\P{^Script=gujarati}', "");
Expect(1, 2801, '\p{Script=-gujarati}', "");
Expect(0, 2801, '\p{^Script=-gujarati}', "");
Expect(0, 2801, '\P{Script=-gujarati}', "");
Expect(1, 2801, '\P{^Script=-gujarati}', "");
Expect(0, 2802, '\p{Script=-gujarati}', "");
Expect(1, 2802, '\p{^Script=-gujarati}', "");
Expect(1, 2802, '\P{Script=-gujarati}', "");
Expect(0, 2802, '\P{^Script=-gujarati}', "");
Error('\p{Sc=:=Gujr}');
Error('\P{Sc=:=Gujr}');
Expect(1, 2801, '\p{Sc=gujr}', "");
Expect(0, 2801, '\p{^Sc=gujr}', "");
Expect(0, 2801, '\P{Sc=gujr}', "");
Expect(1, 2801, '\P{^Sc=gujr}', "");
Expect(0, 2802, '\p{Sc=gujr}', "");
Expect(1, 2802, '\p{^Sc=gujr}', "");
Expect(1, 2802, '\P{Sc=gujr}', "");
Expect(0, 2802, '\P{^Sc=gujr}', "");
Expect(1, 2801, '\p{Sc=  Gujr}', "");
Expect(0, 2801, '\p{^Sc=  Gujr}', "");
Expect(0, 2801, '\P{Sc=  Gujr}', "");
Expect(1, 2801, '\P{^Sc=  Gujr}', "");
Expect(0, 2802, '\p{Sc=  Gujr}', "");
Expect(1, 2802, '\p{^Sc=  Gujr}', "");
Expect(1, 2802, '\P{Sc=  Gujr}', "");
Expect(0, 2802, '\P{^Sc=  Gujr}', "");
Error('\p{Is_Script=	/a/Gujarati}');
Error('\P{Is_Script=	/a/Gujarati}');
Expect(1, 2801, '\p{Is_Script:gujarati}', "");
Expect(0, 2801, '\p{^Is_Script:gujarati}', "");
Expect(0, 2801, '\P{Is_Script:gujarati}', "");
Expect(1, 2801, '\P{^Is_Script:gujarati}', "");
Expect(0, 2802, '\p{Is_Script:gujarati}', "");
Expect(1, 2802, '\p{^Is_Script:gujarati}', "");
Expect(1, 2802, '\P{Is_Script:gujarati}', "");
Expect(0, 2802, '\P{^Is_Script:gujarati}', "");
Expect(1, 2801, '\p{Is_Script=-	Gujarati}', "");
Expect(0, 2801, '\p{^Is_Script=-	Gujarati}', "");
Expect(0, 2801, '\P{Is_Script=-	Gujarati}', "");
Expect(1, 2801, '\P{^Is_Script=-	Gujarati}', "");
Expect(0, 2802, '\p{Is_Script=-	Gujarati}', "");
Expect(1, 2802, '\p{^Is_Script=-	Gujarati}', "");
Expect(1, 2802, '\P{Is_Script=-	Gujarati}', "");
Expect(0, 2802, '\P{^Is_Script=-	Gujarati}', "");
Error('\p{Is_Sc=--Gujr:=}');
Error('\P{Is_Sc=--Gujr:=}');
Expect(1, 2801, '\p{Is_Sc=gujr}', "");
Expect(0, 2801, '\p{^Is_Sc=gujr}', "");
Expect(0, 2801, '\P{Is_Sc=gujr}', "");
Expect(1, 2801, '\P{^Is_Sc=gujr}', "");
Expect(0, 2802, '\p{Is_Sc=gujr}', "");
Expect(1, 2802, '\p{^Is_Sc=gujr}', "");
Expect(1, 2802, '\P{Is_Sc=gujr}', "");
Expect(0, 2802, '\P{^Is_Sc=gujr}', "");
Expect(1, 2801, '\p{Is_Sc=	-gujr}', "");
Expect(0, 2801, '\p{^Is_Sc=	-gujr}', "");
Expect(0, 2801, '\P{Is_Sc=	-gujr}', "");
Expect(1, 2801, '\P{^Is_Sc=	-gujr}', "");
Expect(0, 2802, '\p{Is_Sc=	-gujr}', "");
Expect(1, 2802, '\p{^Is_Sc=	-gujr}', "");
Expect(1, 2802, '\P{Is_Sc=	-gujr}', "");
Expect(0, 2802, '\P{^Is_Sc=	-gujr}', "");
Error('\p{Script=__ORIYA:=}');
Error('\P{Script=__ORIYA:=}');
Expect(1, 2935, '\p{Script=oriya}', "");
Expect(0, 2935, '\p{^Script=oriya}', "");
Expect(0, 2935, '\P{Script=oriya}', "");
Expect(1, 2935, '\P{^Script=oriya}', "");
Expect(0, 2936, '\p{Script=oriya}', "");
Expect(1, 2936, '\p{^Script=oriya}', "");
Expect(1, 2936, '\P{Script=oriya}', "");
Expect(0, 2936, '\P{^Script=oriya}', "");
Expect(1, 2935, '\p{Script= _ORIYA}', "");
Expect(0, 2935, '\p{^Script= _ORIYA}', "");
Expect(0, 2935, '\P{Script= _ORIYA}', "");
Expect(1, 2935, '\P{^Script= _ORIYA}', "");
Expect(0, 2936, '\p{Script= _ORIYA}', "");
Expect(1, 2936, '\p{^Script= _ORIYA}', "");
Expect(1, 2936, '\P{Script= _ORIYA}', "");
Expect(0, 2936, '\P{^Script= _ORIYA}', "");
Error('\p{Sc= _orya:=}');
Error('\P{Sc= _orya:=}');
Expect(1, 2935, '\p{Sc=orya}', "");
Expect(0, 2935, '\p{^Sc=orya}', "");
Expect(0, 2935, '\P{Sc=orya}', "");
Expect(1, 2935, '\P{^Sc=orya}', "");
Expect(0, 2936, '\p{Sc=orya}', "");
Expect(1, 2936, '\p{^Sc=orya}', "");
Expect(1, 2936, '\P{Sc=orya}', "");
Expect(0, 2936, '\P{^Sc=orya}', "");
Expect(1, 2935, '\p{Sc=_orya}', "");
Expect(0, 2935, '\p{^Sc=_orya}', "");
Expect(0, 2935, '\P{Sc=_orya}', "");
Expect(1, 2935, '\P{^Sc=_orya}', "");
Expect(0, 2936, '\p{Sc=_orya}', "");
Expect(1, 2936, '\p{^Sc=_orya}', "");
Expect(1, 2936, '\P{Sc=_orya}', "");
Expect(0, 2936, '\P{^Sc=_orya}', "");
Error('\p{Is_Script=_:=oriya}');
Error('\P{Is_Script=_:=oriya}');
Expect(1, 2935, '\p{Is_Script=oriya}', "");
Expect(0, 2935, '\p{^Is_Script=oriya}', "");
Expect(0, 2935, '\P{Is_Script=oriya}', "");
Expect(1, 2935, '\P{^Is_Script=oriya}', "");
Expect(0, 2936, '\p{Is_Script=oriya}', "");
Expect(1, 2936, '\p{^Is_Script=oriya}', "");
Expect(1, 2936, '\P{Is_Script=oriya}', "");
Expect(0, 2936, '\P{^Is_Script=oriya}', "");
Expect(1, 2935, '\p{Is_Script=	_ORIYA}', "");
Expect(0, 2935, '\p{^Is_Script=	_ORIYA}', "");
Expect(0, 2935, '\P{Is_Script=	_ORIYA}', "");
Expect(1, 2935, '\P{^Is_Script=	_ORIYA}', "");
Expect(0, 2936, '\p{Is_Script=	_ORIYA}', "");
Expect(1, 2936, '\p{^Is_Script=	_ORIYA}', "");
Expect(1, 2936, '\P{Is_Script=	_ORIYA}', "");
Expect(0, 2936, '\P{^Is_Script=	_ORIYA}', "");
Error('\p{Is_Sc=/a/_Orya}');
Error('\P{Is_Sc=/a/_Orya}');
Expect(1, 2935, '\p{Is_Sc=orya}', "");
Expect(0, 2935, '\p{^Is_Sc=orya}', "");
Expect(0, 2935, '\P{Is_Sc=orya}', "");
Expect(1, 2935, '\P{^Is_Sc=orya}', "");
Expect(0, 2936, '\p{Is_Sc=orya}', "");
Expect(1, 2936, '\p{^Is_Sc=orya}', "");
Expect(1, 2936, '\P{Is_Sc=orya}', "");
Expect(0, 2936, '\P{^Is_Sc=orya}', "");
Expect(1, 2935, '\p{Is_Sc=-orya}', "");
Expect(0, 2935, '\p{^Is_Sc=-orya}', "");
Expect(0, 2935, '\P{Is_Sc=-orya}', "");
Expect(1, 2935, '\P{^Is_Sc=-orya}', "");
Expect(0, 2936, '\p{Is_Sc=-orya}', "");
Expect(1, 2936, '\p{^Is_Sc=-orya}', "");
Expect(1, 2936, '\P{Is_Sc=-orya}', "");
Expect(0, 2936, '\P{^Is_Sc=-orya}', "");
Error('\p{Script=:=  LISU}');
Error('\P{Script=:=  LISU}');
Expect(1, 42239, '\p{Script=lisu}', "");
Expect(0, 42239, '\p{^Script=lisu}', "");
Expect(0, 42239, '\P{Script=lisu}', "");
Expect(1, 42239, '\P{^Script=lisu}', "");
Expect(0, 42240, '\p{Script=lisu}', "");
Expect(1, 42240, '\p{^Script=lisu}', "");
Expect(1, 42240, '\P{Script=lisu}', "");
Expect(0, 42240, '\P{^Script=lisu}', "");
Expect(1, 42239, '\p{Script= Lisu}', "");
Expect(0, 42239, '\p{^Script= Lisu}', "");
Expect(0, 42239, '\P{Script= Lisu}', "");
Expect(1, 42239, '\P{^Script= Lisu}', "");
Expect(0, 42240, '\p{Script= Lisu}', "");
Expect(1, 42240, '\p{^Script= Lisu}', "");
Expect(1, 42240, '\P{Script= Lisu}', "");
Expect(0, 42240, '\P{^Script= Lisu}', "");
Error('\p{Sc=  LISU/a/}');
Error('\P{Sc=  LISU/a/}');
Expect(1, 42239, '\p{Sc=lisu}', "");
Expect(0, 42239, '\p{^Sc=lisu}', "");
Expect(0, 42239, '\P{Sc=lisu}', "");
Expect(1, 42239, '\P{^Sc=lisu}', "");
Expect(0, 42240, '\p{Sc=lisu}', "");
Expect(1, 42240, '\p{^Sc=lisu}', "");
Expect(1, 42240, '\P{Sc=lisu}', "");
Expect(0, 42240, '\P{^Sc=lisu}', "");
Expect(1, 42239, '\p{Sc= 	LISU}', "");
Expect(0, 42239, '\p{^Sc= 	LISU}', "");
Expect(0, 42239, '\P{Sc= 	LISU}', "");
Expect(1, 42239, '\P{^Sc= 	LISU}', "");
Expect(0, 42240, '\p{Sc= 	LISU}', "");
Expect(1, 42240, '\p{^Sc= 	LISU}', "");
Expect(1, 42240, '\P{Sc= 	LISU}', "");
Expect(0, 42240, '\P{^Sc= 	LISU}', "");
Error('\p{Is_Script:	_lisu:=}');
Error('\P{Is_Script:	_lisu:=}');
Expect(1, 42239, '\p{Is_Script=lisu}', "");
Expect(0, 42239, '\p{^Is_Script=lisu}', "");
Expect(0, 42239, '\P{Is_Script=lisu}', "");
Expect(1, 42239, '\P{^Is_Script=lisu}', "");
Expect(0, 42240, '\p{Is_Script=lisu}', "");
Expect(1, 42240, '\p{^Is_Script=lisu}', "");
Expect(1, 42240, '\P{Is_Script=lisu}', "");
Expect(0, 42240, '\P{^Is_Script=lisu}', "");
Expect(1, 42239, '\p{Is_Script=--Lisu}', "");
Expect(0, 42239, '\p{^Is_Script=--Lisu}', "");
Expect(0, 42239, '\P{Is_Script=--Lisu}', "");
Expect(1, 42239, '\P{^Is_Script=--Lisu}', "");
Expect(0, 42240, '\p{Is_Script=--Lisu}', "");
Expect(1, 42240, '\p{^Is_Script=--Lisu}', "");
Expect(1, 42240, '\P{Is_Script=--Lisu}', "");
Expect(0, 42240, '\P{^Is_Script=--Lisu}', "");
Error('\p{Is_Sc=/a/_	lisu}');
Error('\P{Is_Sc=/a/_	lisu}');
Expect(1, 42239, '\p{Is_Sc=lisu}', "");
Expect(0, 42239, '\p{^Is_Sc=lisu}', "");
Expect(0, 42239, '\P{Is_Sc=lisu}', "");
Expect(1, 42239, '\P{^Is_Sc=lisu}', "");
Expect(0, 42240, '\p{Is_Sc=lisu}', "");
Expect(1, 42240, '\p{^Is_Sc=lisu}', "");
Expect(1, 42240, '\P{Is_Sc=lisu}', "");
Expect(0, 42240, '\P{^Is_Sc=lisu}', "");
Expect(1, 42239, '\p{Is_Sc=_ LISU}', "");
Expect(0, 42239, '\p{^Is_Sc=_ LISU}', "");
Expect(0, 42239, '\P{Is_Sc=_ LISU}', "");
Expect(1, 42239, '\P{^Is_Sc=_ LISU}', "");
Expect(0, 42240, '\p{Is_Sc=_ LISU}', "");
Expect(1, 42240, '\p{^Is_Sc=_ LISU}', "");
Expect(1, 42240, '\P{Is_Sc=_ LISU}', "");
Expect(0, 42240, '\P{^Is_Sc=_ LISU}', "");
Error('\p{Script=:=__tai_Le}');
Error('\P{Script=:=__tai_Le}');
Expect(1, 6516, '\p{Script=taile}', "");
Expect(0, 6516, '\p{^Script=taile}', "");
Expect(0, 6516, '\P{Script=taile}', "");
Expect(1, 6516, '\P{^Script=taile}', "");
Expect(0, 6517, '\p{Script=taile}', "");
Expect(1, 6517, '\p{^Script=taile}', "");
Expect(1, 6517, '\P{Script=taile}', "");
Expect(0, 6517, '\P{^Script=taile}', "");
Expect(1, 6516, '\p{Script=	Tai_LE}', "");
Expect(0, 6516, '\p{^Script=	Tai_LE}', "");
Expect(0, 6516, '\P{Script=	Tai_LE}', "");
Expect(1, 6516, '\P{^Script=	Tai_LE}', "");
Expect(0, 6517, '\p{Script=	Tai_LE}', "");
Expect(1, 6517, '\p{^Script=	Tai_LE}', "");
Expect(1, 6517, '\P{Script=	Tai_LE}', "");
Expect(0, 6517, '\P{^Script=	Tai_LE}', "");
Error('\p{Sc=/a/-	tale}');
Error('\P{Sc=/a/-	tale}');
Expect(1, 6516, '\p{Sc=tale}', "");
Expect(0, 6516, '\p{^Sc=tale}', "");
Expect(0, 6516, '\P{Sc=tale}', "");
Expect(1, 6516, '\P{^Sc=tale}', "");
Expect(0, 6517, '\p{Sc=tale}', "");
Expect(1, 6517, '\p{^Sc=tale}', "");
Expect(1, 6517, '\P{Sc=tale}', "");
Expect(0, 6517, '\P{^Sc=tale}', "");
Expect(1, 6516, '\p{Sc:- tale}', "");
Expect(0, 6516, '\p{^Sc:- tale}', "");
Expect(0, 6516, '\P{Sc:- tale}', "");
Expect(1, 6516, '\P{^Sc:- tale}', "");
Expect(0, 6517, '\p{Sc:- tale}', "");
Expect(1, 6517, '\p{^Sc:- tale}', "");
Expect(1, 6517, '\P{Sc:- tale}', "");
Expect(0, 6517, '\P{^Sc:- tale}', "");
Error('\p{Is_Script=/a/Tai_LE}');
Error('\P{Is_Script=/a/Tai_LE}');
Expect(1, 6516, '\p{Is_Script:	taile}', "");
Expect(0, 6516, '\p{^Is_Script:	taile}', "");
Expect(0, 6516, '\P{Is_Script:	taile}', "");
Expect(1, 6516, '\P{^Is_Script:	taile}', "");
Expect(0, 6517, '\p{Is_Script:	taile}', "");
Expect(1, 6517, '\p{^Is_Script:	taile}', "");
Expect(1, 6517, '\P{Is_Script:	taile}', "");
Expect(0, 6517, '\P{^Is_Script:	taile}', "");
Expect(1, 6516, '\p{Is_Script=		tai_Le}', "");
Expect(0, 6516, '\p{^Is_Script=		tai_Le}', "");
Expect(0, 6516, '\P{Is_Script=		tai_Le}', "");
Expect(1, 6516, '\P{^Is_Script=		tai_Le}', "");
Expect(0, 6517, '\p{Is_Script=		tai_Le}', "");
Expect(1, 6517, '\p{^Is_Script=		tai_Le}', "");
Expect(1, 6517, '\P{Is_Script=		tai_Le}', "");
Expect(0, 6517, '\P{^Is_Script=		tai_Le}', "");
Error('\p{Is_Sc=:=_Tale}');
Error('\P{Is_Sc=:=_Tale}');
Expect(1, 6516, '\p{Is_Sc=tale}', "");
Expect(0, 6516, '\p{^Is_Sc=tale}', "");
Expect(0, 6516, '\P{Is_Sc=tale}', "");
Expect(1, 6516, '\P{^Is_Sc=tale}', "");
Expect(0, 6517, '\p{Is_Sc=tale}', "");
Expect(1, 6517, '\p{^Is_Sc=tale}', "");
Expect(1, 6517, '\P{Is_Sc=tale}', "");
Expect(0, 6517, '\P{^Is_Sc=tale}', "");
Expect(1, 6516, '\p{Is_Sc=	TALE}', "");
Expect(0, 6516, '\p{^Is_Sc=	TALE}', "");
Expect(0, 6516, '\P{Is_Sc=	TALE}', "");
Expect(1, 6516, '\P{^Is_Sc=	TALE}', "");
Expect(0, 6517, '\p{Is_Sc=	TALE}', "");
Expect(1, 6517, '\p{^Is_Sc=	TALE}', "");
Expect(1, 6517, '\P{Is_Sc=	TALE}', "");
Expect(0, 6517, '\P{^Is_Sc=	TALE}', "");
Error('\p{Script=	_AVESTAN:=}');
Error('\P{Script=	_AVESTAN:=}');
Expect(1, 68415, '\p{Script=avestan}', "");
Expect(0, 68415, '\p{^Script=avestan}', "");
Expect(0, 68415, '\P{Script=avestan}', "");
Expect(1, 68415, '\P{^Script=avestan}', "");
Expect(0, 68416, '\p{Script=avestan}', "");
Expect(1, 68416, '\p{^Script=avestan}', "");
Expect(1, 68416, '\P{Script=avestan}', "");
Expect(0, 68416, '\P{^Script=avestan}', "");
Expect(1, 68415, '\p{Script=--avestan}', "");
Expect(0, 68415, '\p{^Script=--avestan}', "");
Expect(0, 68415, '\P{Script=--avestan}', "");
Expect(1, 68415, '\P{^Script=--avestan}', "");
Expect(0, 68416, '\p{Script=--avestan}', "");
Expect(1, 68416, '\p{^Script=--avestan}', "");
Expect(1, 68416, '\P{Script=--avestan}', "");
Expect(0, 68416, '\P{^Script=--avestan}', "");
Error('\p{Sc: __Avst/a/}');
Error('\P{Sc: __Avst/a/}');
Expect(1, 68415, '\p{Sc=avst}', "");
Expect(0, 68415, '\p{^Sc=avst}', "");
Expect(0, 68415, '\P{Sc=avst}', "");
Expect(1, 68415, '\P{^Sc=avst}', "");
Expect(0, 68416, '\p{Sc=avst}', "");
Expect(1, 68416, '\p{^Sc=avst}', "");
Expect(1, 68416, '\P{Sc=avst}', "");
Expect(0, 68416, '\P{^Sc=avst}', "");
Expect(1, 68415, '\p{Sc=-AVST}', "");
Expect(0, 68415, '\p{^Sc=-AVST}', "");
Expect(0, 68415, '\P{Sc=-AVST}', "");
Expect(1, 68415, '\P{^Sc=-AVST}', "");
Expect(0, 68416, '\p{Sc=-AVST}', "");
Expect(1, 68416, '\p{^Sc=-AVST}', "");
Expect(1, 68416, '\P{Sc=-AVST}', "");
Expect(0, 68416, '\P{^Sc=-AVST}', "");
Error('\p{Is_Script=-_avestan:=}');
Error('\P{Is_Script=-_avestan:=}');
Expect(1, 68415, '\p{Is_Script=avestan}', "");
Expect(0, 68415, '\p{^Is_Script=avestan}', "");
Expect(0, 68415, '\P{Is_Script=avestan}', "");
Expect(1, 68415, '\P{^Is_Script=avestan}', "");
Expect(0, 68416, '\p{Is_Script=avestan}', "");
Expect(1, 68416, '\p{^Is_Script=avestan}', "");
Expect(1, 68416, '\P{Is_Script=avestan}', "");
Expect(0, 68416, '\P{^Is_Script=avestan}', "");
Expect(1, 68415, '\p{Is_Script:	 	avestan}', "");
Expect(0, 68415, '\p{^Is_Script:	 	avestan}', "");
Expect(0, 68415, '\P{Is_Script:	 	avestan}', "");
Expect(1, 68415, '\P{^Is_Script:	 	avestan}', "");
Expect(0, 68416, '\p{Is_Script:	 	avestan}', "");
Expect(1, 68416, '\p{^Is_Script:	 	avestan}', "");
Expect(1, 68416, '\P{Is_Script:	 	avestan}', "");
Expect(0, 68416, '\P{^Is_Script:	 	avestan}', "");
Error('\p{Is_Sc=:=_AVST}');
Error('\P{Is_Sc=:=_AVST}');
Expect(1, 68415, '\p{Is_Sc=avst}', "");
Expect(0, 68415, '\p{^Is_Sc=avst}', "");
Expect(0, 68415, '\P{Is_Sc=avst}', "");
Expect(1, 68415, '\P{^Is_Sc=avst}', "");
Expect(0, 68416, '\p{Is_Sc=avst}', "");
Expect(1, 68416, '\p{^Is_Sc=avst}', "");
Expect(1, 68416, '\P{Is_Sc=avst}', "");
Expect(0, 68416, '\P{^Is_Sc=avst}', "");
Expect(1, 68415, '\p{Is_Sc= Avst}', "");
Expect(0, 68415, '\p{^Is_Sc= Avst}', "");
Expect(0, 68415, '\P{Is_Sc= Avst}', "");
Expect(1, 68415, '\P{^Is_Sc= Avst}', "");
Expect(0, 68416, '\p{Is_Sc= Avst}', "");
Expect(1, 68416, '\p{^Is_Sc= Avst}', "");
Expect(1, 68416, '\P{Is_Sc= Avst}', "");
Expect(0, 68416, '\P{^Is_Sc= Avst}', "");
Error('\p{Script= -Georgian:=}');
Error('\P{Script= -Georgian:=}');
Expect(1, 11565, '\p{Script=georgian}', "");
Expect(0, 11565, '\p{^Script=georgian}', "");
Expect(0, 11565, '\P{Script=georgian}', "");
Expect(1, 11565, '\P{^Script=georgian}', "");
Expect(0, 11566, '\p{Script=georgian}', "");
Expect(1, 11566, '\p{^Script=georgian}', "");
Expect(1, 11566, '\P{Script=georgian}', "");
Expect(0, 11566, '\P{^Script=georgian}', "");
Expect(1, 11565, '\p{Script=-Georgian}', "");
Expect(0, 11565, '\p{^Script=-Georgian}', "");
Expect(0, 11565, '\P{Script=-Georgian}', "");
Expect(1, 11565, '\P{^Script=-Georgian}', "");
Expect(0, 11566, '\p{Script=-Georgian}', "");
Expect(1, 11566, '\p{^Script=-Georgian}', "");
Expect(1, 11566, '\P{Script=-Georgian}', "");
Expect(0, 11566, '\P{^Script=-Georgian}', "");
Error('\p{Sc= :=geor}');
Error('\P{Sc= :=geor}');
Expect(1, 11565, '\p{Sc=geor}', "");
Expect(0, 11565, '\p{^Sc=geor}', "");
Expect(0, 11565, '\P{Sc=geor}', "");
Expect(1, 11565, '\P{^Sc=geor}', "");
Expect(0, 11566, '\p{Sc=geor}', "");
Expect(1, 11566, '\p{^Sc=geor}', "");
Expect(1, 11566, '\P{Sc=geor}', "");
Expect(0, 11566, '\P{^Sc=geor}', "");
Expect(1, 11565, '\p{Sc=_	Geor}', "");
Expect(0, 11565, '\p{^Sc=_	Geor}', "");
Expect(0, 11565, '\P{Sc=_	Geor}', "");
Expect(1, 11565, '\P{^Sc=_	Geor}', "");
Expect(0, 11566, '\p{Sc=_	Geor}', "");
Expect(1, 11566, '\p{^Sc=_	Geor}', "");
Expect(1, 11566, '\P{Sc=_	Geor}', "");
Expect(0, 11566, '\P{^Sc=_	Geor}', "");
Error('\p{Is_Script= Georgian/a/}');
Error('\P{Is_Script= Georgian/a/}');
Expect(1, 11565, '\p{Is_Script=georgian}', "");
Expect(0, 11565, '\p{^Is_Script=georgian}', "");
Expect(0, 11565, '\P{Is_Script=georgian}', "");
Expect(1, 11565, '\P{^Is_Script=georgian}', "");
Expect(0, 11566, '\p{Is_Script=georgian}', "");
Expect(1, 11566, '\p{^Is_Script=georgian}', "");
Expect(1, 11566, '\P{Is_Script=georgian}', "");
Expect(0, 11566, '\P{^Is_Script=georgian}', "");
Error('\p{Is_Sc:	:=		Geor}');
Error('\P{Is_Sc:	:=		Geor}');
Expect(1, 11565, '\p{Is_Sc=geor}', "");
Expect(0, 11565, '\p{^Is_Sc=geor}', "");
Expect(0, 11565, '\P{Is_Sc=geor}', "");
Expect(1, 11565, '\P{^Is_Sc=geor}', "");
Expect(0, 11566, '\p{Is_Sc=geor}', "");
Expect(1, 11566, '\p{^Is_Sc=geor}', "");
Expect(1, 11566, '\P{Is_Sc=geor}', "");
Expect(0, 11566, '\P{^Is_Sc=geor}', "");
Expect(1, 11565, '\p{Is_Sc:	-	geor}', "");
Expect(0, 11565, '\p{^Is_Sc:	-	geor}', "");
Expect(0, 11565, '\P{Is_Sc:	-	geor}', "");
Expect(1, 11565, '\P{^Is_Sc:	-	geor}', "");
Expect(0, 11566, '\p{Is_Sc:	-	geor}', "");
Expect(1, 11566, '\p{^Is_Sc:	-	geor}', "");
Expect(1, 11566, '\P{Is_Sc:	-	geor}', "");
Expect(0, 11566, '\P{^Is_Sc:	-	geor}', "");
Error('\p{Script=--TELUGU/a/}');
Error('\P{Script=--TELUGU/a/}');
Expect(1, 3199, '\p{Script=telugu}', "");
Expect(0, 3199, '\p{^Script=telugu}', "");
Expect(0, 3199, '\P{Script=telugu}', "");
Expect(1, 3199, '\P{^Script=telugu}', "");
Expect(0, 3200, '\p{Script=telugu}', "");
Expect(1, 3200, '\p{^Script=telugu}', "");
Expect(1, 3200, '\P{Script=telugu}', "");
Expect(0, 3200, '\P{^Script=telugu}', "");
Expect(1, 3199, '\p{Script=  telugu}', "");
Expect(0, 3199, '\p{^Script=  telugu}', "");
Expect(0, 3199, '\P{Script=  telugu}', "");
Expect(1, 3199, '\P{^Script=  telugu}', "");
Expect(0, 3200, '\p{Script=  telugu}', "");
Expect(1, 3200, '\p{^Script=  telugu}', "");
Expect(1, 3200, '\P{Script=  telugu}', "");
Expect(0, 3200, '\P{^Script=  telugu}', "");
Error('\p{Sc= :=telu}');
Error('\P{Sc= :=telu}');
Expect(1, 3199, '\p{Sc=telu}', "");
Expect(0, 3199, '\p{^Sc=telu}', "");
Expect(0, 3199, '\P{Sc=telu}', "");
Expect(1, 3199, '\P{^Sc=telu}', "");
Expect(0, 3200, '\p{Sc=telu}', "");
Expect(1, 3200, '\p{^Sc=telu}', "");
Expect(1, 3200, '\P{Sc=telu}', "");
Expect(0, 3200, '\P{^Sc=telu}', "");
Expect(1, 3199, '\p{Sc=--telu}', "");
Expect(0, 3199, '\p{^Sc=--telu}', "");
Expect(0, 3199, '\P{Sc=--telu}', "");
Expect(1, 3199, '\P{^Sc=--telu}', "");
Expect(0, 3200, '\p{Sc=--telu}', "");
Expect(1, 3200, '\p{^Sc=--telu}', "");
Expect(1, 3200, '\P{Sc=--telu}', "");
Expect(0, 3200, '\P{^Sc=--telu}', "");
Error('\p{Is_Script=-:=telugu}');
Error('\P{Is_Script=-:=telugu}');
Expect(1, 3199, '\p{Is_Script=telugu}', "");
Expect(0, 3199, '\p{^Is_Script=telugu}', "");
Expect(0, 3199, '\P{Is_Script=telugu}', "");
Expect(1, 3199, '\P{^Is_Script=telugu}', "");
Expect(0, 3200, '\p{Is_Script=telugu}', "");
Expect(1, 3200, '\p{^Is_Script=telugu}', "");
Expect(1, 3200, '\P{Is_Script=telugu}', "");
Expect(0, 3200, '\P{^Is_Script=telugu}', "");
Expect(1, 3199, '\p{Is_Script:	--telugu}', "");
Expect(0, 3199, '\p{^Is_Script:	--telugu}', "");
Expect(0, 3199, '\P{Is_Script:	--telugu}', "");
Expect(1, 3199, '\P{^Is_Script:	--telugu}', "");
Expect(0, 3200, '\p{Is_Script:	--telugu}', "");
Expect(1, 3200, '\p{^Is_Script:	--telugu}', "");
Expect(1, 3200, '\P{Is_Script:	--telugu}', "");
Expect(0, 3200, '\P{^Is_Script:	--telugu}', "");
Error('\p{Is_Sc=:=_ Telu}');
Error('\P{Is_Sc=:=_ Telu}');
Expect(1, 3199, '\p{Is_Sc=telu}', "");
Expect(0, 3199, '\p{^Is_Sc=telu}', "");
Expect(0, 3199, '\P{Is_Sc=telu}', "");
Expect(1, 3199, '\P{^Is_Sc=telu}', "");
Expect(0, 3200, '\p{Is_Sc=telu}', "");
Expect(1, 3200, '\p{^Is_Sc=telu}', "");
Expect(1, 3200, '\P{Is_Sc=telu}', "");
Expect(0, 3200, '\P{^Is_Sc=telu}', "");
Expect(1, 3199, '\p{Is_Sc= _telu}', "");
Expect(0, 3199, '\p{^Is_Sc= _telu}', "");
Expect(0, 3199, '\P{Is_Sc= _telu}', "");
Expect(1, 3199, '\P{^Is_Sc= _telu}', "");
Expect(0, 3200, '\p{Is_Sc= _telu}', "");
Expect(1, 3200, '\p{^Is_Sc= _telu}', "");
Expect(1, 3200, '\P{Is_Sc= _telu}', "");
Expect(0, 3200, '\P{^Is_Sc= _telu}', "");
Error('\p{Script=	TIFINAGH/a/}');
Error('\P{Script=	TIFINAGH/a/}');
Expect(1, 11647, '\p{Script: tifinagh}', "");
Expect(0, 11647, '\p{^Script: tifinagh}', "");
Expect(0, 11647, '\P{Script: tifinagh}', "");
Expect(1, 11647, '\P{^Script: tifinagh}', "");
Expect(0, 11648, '\p{Script: tifinagh}', "");
Expect(1, 11648, '\p{^Script: tifinagh}', "");
Expect(1, 11648, '\P{Script: tifinagh}', "");
Expect(0, 11648, '\P{^Script: tifinagh}', "");
Expect(1, 11647, '\p{Script=-TIFINAGH}', "");
Expect(0, 11647, '\p{^Script=-TIFINAGH}', "");
Expect(0, 11647, '\P{Script=-TIFINAGH}', "");
Expect(1, 11647, '\P{^Script=-TIFINAGH}', "");
Expect(0, 11648, '\p{Script=-TIFINAGH}', "");
Expect(1, 11648, '\p{^Script=-TIFINAGH}', "");
Expect(1, 11648, '\P{Script=-TIFINAGH}', "");
Expect(0, 11648, '\P{^Script=-TIFINAGH}', "");
Error('\p{Sc= :=Tfng}');
Error('\P{Sc= :=Tfng}');
Expect(1, 11647, '\p{Sc=tfng}', "");
Expect(0, 11647, '\p{^Sc=tfng}', "");
Expect(0, 11647, '\P{Sc=tfng}', "");
Expect(1, 11647, '\P{^Sc=tfng}', "");
Expect(0, 11648, '\p{Sc=tfng}', "");
Expect(1, 11648, '\p{^Sc=tfng}', "");
Expect(1, 11648, '\P{Sc=tfng}', "");
Expect(0, 11648, '\P{^Sc=tfng}', "");
Expect(1, 11647, '\p{Sc=__TFNG}', "");
Expect(0, 11647, '\p{^Sc=__TFNG}', "");
Expect(0, 11647, '\P{Sc=__TFNG}', "");
Expect(1, 11647, '\P{^Sc=__TFNG}', "");
Expect(0, 11648, '\p{Sc=__TFNG}', "");
Expect(1, 11648, '\p{^Sc=__TFNG}', "");
Expect(1, 11648, '\P{Sc=__TFNG}', "");
Expect(0, 11648, '\P{^Sc=__TFNG}', "");
Error('\p{Is_Script=:=Tifinagh}');
Error('\P{Is_Script=:=Tifinagh}');
Expect(1, 11647, '\p{Is_Script=tifinagh}', "");
Expect(0, 11647, '\p{^Is_Script=tifinagh}', "");
Expect(0, 11647, '\P{Is_Script=tifinagh}', "");
Expect(1, 11647, '\P{^Is_Script=tifinagh}', "");
Expect(0, 11648, '\p{Is_Script=tifinagh}', "");
Expect(1, 11648, '\p{^Is_Script=tifinagh}', "");
Expect(1, 11648, '\P{Is_Script=tifinagh}', "");
Expect(0, 11648, '\P{^Is_Script=tifinagh}', "");
Expect(1, 11647, '\p{Is_Script=-_TIFINAGH}', "");
Expect(0, 11647, '\p{^Is_Script=-_TIFINAGH}', "");
Expect(0, 11647, '\P{Is_Script=-_TIFINAGH}', "");
Expect(1, 11647, '\P{^Is_Script=-_TIFINAGH}', "");
Expect(0, 11648, '\p{Is_Script=-_TIFINAGH}', "");
Expect(1, 11648, '\p{^Is_Script=-_TIFINAGH}', "");
Expect(1, 11648, '\P{Is_Script=-_TIFINAGH}', "");
Expect(0, 11648, '\P{^Is_Script=-_TIFINAGH}', "");
Error('\p{Is_Sc=/a/_tfng}');
Error('\P{Is_Sc=/a/_tfng}');
Expect(1, 11647, '\p{Is_Sc=tfng}', "");
Expect(0, 11647, '\p{^Is_Sc=tfng}', "");
Expect(0, 11647, '\P{Is_Sc=tfng}', "");
Expect(1, 11647, '\P{^Is_Sc=tfng}', "");
Expect(0, 11648, '\p{Is_Sc=tfng}', "");
Expect(1, 11648, '\p{^Is_Sc=tfng}', "");
Expect(1, 11648, '\P{Is_Sc=tfng}', "");
Expect(0, 11648, '\P{^Is_Sc=tfng}', "");
Expect(1, 11647, '\p{Is_Sc=_Tfng}', "");
Expect(0, 11647, '\p{^Is_Sc=_Tfng}', "");
Expect(0, 11647, '\P{Is_Sc=_Tfng}', "");
Expect(1, 11647, '\P{^Is_Sc=_Tfng}', "");
Expect(0, 11648, '\p{Is_Sc=_Tfng}', "");
Expect(1, 11648, '\p{^Is_Sc=_Tfng}', "");
Expect(1, 11648, '\P{Is_Sc=_Tfng}', "");
Expect(0, 11648, '\P{^Is_Sc=_Tfng}', "");
Error('\p{Script=	_SINHALA/a/}');
Error('\P{Script=	_SINHALA/a/}');
Expect(1, 3572, '\p{Script=sinhala}', "");
Expect(0, 3572, '\p{^Script=sinhala}', "");
Expect(0, 3572, '\P{Script=sinhala}', "");
Expect(1, 3572, '\P{^Script=sinhala}', "");
Expect(0, 3573, '\p{Script=sinhala}', "");
Expect(1, 3573, '\p{^Script=sinhala}', "");
Expect(1, 3573, '\P{Script=sinhala}', "");
Expect(0, 3573, '\P{^Script=sinhala}', "");
Expect(1, 3572, '\p{Script=__Sinhala}', "");
Expect(0, 3572, '\p{^Script=__Sinhala}', "");
Expect(0, 3572, '\P{Script=__Sinhala}', "");
Expect(1, 3572, '\P{^Script=__Sinhala}', "");
Expect(0, 3573, '\p{Script=__Sinhala}', "");
Expect(1, 3573, '\p{^Script=__Sinhala}', "");
Expect(1, 3573, '\P{Script=__Sinhala}', "");
Expect(0, 3573, '\P{^Script=__Sinhala}', "");
Error('\p{Sc=-/a/Sinh}');
Error('\P{Sc=-/a/Sinh}');
Expect(1, 3572, '\p{Sc=sinh}', "");
Expect(0, 3572, '\p{^Sc=sinh}', "");
Expect(0, 3572, '\P{Sc=sinh}', "");
Expect(1, 3572, '\P{^Sc=sinh}', "");
Expect(0, 3573, '\p{Sc=sinh}', "");
Expect(1, 3573, '\p{^Sc=sinh}', "");
Expect(1, 3573, '\P{Sc=sinh}', "");
Expect(0, 3573, '\P{^Sc=sinh}', "");
Expect(1, 3572, '\p{Sc= Sinh}', "");
Expect(0, 3572, '\p{^Sc= Sinh}', "");
Expect(0, 3572, '\P{Sc= Sinh}', "");
Expect(1, 3572, '\P{^Sc= Sinh}', "");
Expect(0, 3573, '\p{Sc= Sinh}', "");
Expect(1, 3573, '\p{^Sc= Sinh}', "");
Expect(1, 3573, '\P{Sc= Sinh}', "");
Expect(0, 3573, '\P{^Sc= Sinh}', "");
Error('\p{Is_Script=_-Sinhala/a/}');
Error('\P{Is_Script=_-Sinhala/a/}');
Expect(1, 3572, '\p{Is_Script=sinhala}', "");
Expect(0, 3572, '\p{^Is_Script=sinhala}', "");
Expect(0, 3572, '\P{Is_Script=sinhala}', "");
Expect(1, 3572, '\P{^Is_Script=sinhala}', "");
Expect(0, 3573, '\p{Is_Script=sinhala}', "");
Expect(1, 3573, '\p{^Is_Script=sinhala}', "");
Expect(1, 3573, '\P{Is_Script=sinhala}', "");
Expect(0, 3573, '\P{^Is_Script=sinhala}', "");
Expect(1, 3572, '\p{Is_Script:	_SINHALA}', "");
Expect(0, 3572, '\p{^Is_Script:	_SINHALA}', "");
Expect(0, 3572, '\P{Is_Script:	_SINHALA}', "");
Expect(1, 3572, '\P{^Is_Script:	_SINHALA}', "");
Expect(0, 3573, '\p{Is_Script:	_SINHALA}', "");
Expect(1, 3573, '\p{^Is_Script:	_SINHALA}', "");
Expect(1, 3573, '\P{Is_Script:	_SINHALA}', "");
Expect(0, 3573, '\P{^Is_Script:	_SINHALA}', "");
Error('\p{Is_Sc=:=-SINH}');
Error('\P{Is_Sc=:=-SINH}');
Expect(1, 3572, '\p{Is_Sc: sinh}', "");
Expect(0, 3572, '\p{^Is_Sc: sinh}', "");
Expect(0, 3572, '\P{Is_Sc: sinh}', "");
Expect(1, 3572, '\P{^Is_Sc: sinh}', "");
Expect(0, 3573, '\p{Is_Sc: sinh}', "");
Expect(1, 3573, '\p{^Is_Sc: sinh}', "");
Expect(1, 3573, '\P{Is_Sc: sinh}', "");
Expect(0, 3573, '\P{^Is_Sc: sinh}', "");
Expect(1, 3572, '\p{Is_Sc: _Sinh}', "");
Expect(0, 3572, '\p{^Is_Sc: _Sinh}', "");
Expect(0, 3572, '\P{Is_Sc: _Sinh}', "");
Expect(1, 3572, '\P{^Is_Sc: _Sinh}', "");
Expect(0, 3573, '\p{Is_Sc: _Sinh}', "");
Expect(1, 3573, '\p{^Is_Sc: _Sinh}', "");
Expect(1, 3573, '\P{Is_Sc: _Sinh}', "");
Expect(0, 3573, '\P{^Is_Sc: _Sinh}', "");
Error('\p{Script=	 miao/a/}');
Error('\P{Script=	 miao/a/}');
Expect(1, 94111, '\p{Script=miao}', "");
Expect(0, 94111, '\p{^Script=miao}', "");
Expect(0, 94111, '\P{Script=miao}', "");
Expect(1, 94111, '\P{^Script=miao}', "");
Expect(0, 94112, '\p{Script=miao}', "");
Expect(1, 94112, '\p{^Script=miao}', "");
Expect(1, 94112, '\P{Script=miao}', "");
Expect(0, 94112, '\P{^Script=miao}', "");
Expect(1, 94111, '\p{Script= Miao}', "");
Expect(0, 94111, '\p{^Script= Miao}', "");
Expect(0, 94111, '\P{Script= Miao}', "");
Expect(1, 94111, '\P{^Script= Miao}', "");
Expect(0, 94112, '\p{Script= Miao}', "");
Expect(1, 94112, '\p{^Script= Miao}', "");
Expect(1, 94112, '\P{Script= Miao}', "");
Expect(0, 94112, '\P{^Script= Miao}', "");
Error('\p{Sc=_-Plrd/a/}');
Error('\P{Sc=_-Plrd/a/}');
Expect(1, 94111, '\p{Sc=plrd}', "");
Expect(0, 94111, '\p{^Sc=plrd}', "");
Expect(0, 94111, '\P{Sc=plrd}', "");
Expect(1, 94111, '\P{^Sc=plrd}', "");
Expect(0, 94112, '\p{Sc=plrd}', "");
Expect(1, 94112, '\p{^Sc=plrd}', "");
Expect(1, 94112, '\P{Sc=plrd}', "");
Expect(0, 94112, '\P{^Sc=plrd}', "");
Expect(1, 94111, '\p{Sc=--Plrd}', "");
Expect(0, 94111, '\p{^Sc=--Plrd}', "");
Expect(0, 94111, '\P{Sc=--Plrd}', "");
Expect(1, 94111, '\P{^Sc=--Plrd}', "");
Expect(0, 94112, '\p{Sc=--Plrd}', "");
Expect(1, 94112, '\p{^Sc=--Plrd}', "");
Expect(1, 94112, '\P{Sc=--Plrd}', "");
Expect(0, 94112, '\P{^Sc=--Plrd}', "");
Error('\p{Is_Script=-/a/Miao}');
Error('\P{Is_Script=-/a/Miao}');
Expect(1, 94111, '\p{Is_Script=miao}', "");
Expect(0, 94111, '\p{^Is_Script=miao}', "");
Expect(0, 94111, '\P{Is_Script=miao}', "");
Expect(1, 94111, '\P{^Is_Script=miao}', "");
Expect(0, 94112, '\p{Is_Script=miao}', "");
Expect(1, 94112, '\p{^Is_Script=miao}', "");
Expect(1, 94112, '\P{Is_Script=miao}', "");
Expect(0, 94112, '\P{^Is_Script=miao}', "");
Expect(1, 94111, '\p{Is_Script=- Miao}', "");
Expect(0, 94111, '\p{^Is_Script=- Miao}', "");
Expect(0, 94111, '\P{Is_Script=- Miao}', "");
Expect(1, 94111, '\P{^Is_Script=- Miao}', "");
Expect(0, 94112, '\p{Is_Script=- Miao}', "");
Expect(1, 94112, '\p{^Is_Script=- Miao}', "");
Expect(1, 94112, '\P{Is_Script=- Miao}', "");
Expect(0, 94112, '\P{^Is_Script=- Miao}', "");
Error('\p{Is_Sc=:= Plrd}');
Error('\P{Is_Sc=:= Plrd}');
Expect(1, 94111, '\p{Is_Sc:plrd}', "");
Expect(0, 94111, '\p{^Is_Sc:plrd}', "");
Expect(0, 94111, '\P{Is_Sc:plrd}', "");
Expect(1, 94111, '\P{^Is_Sc:plrd}', "");
Expect(0, 94112, '\p{Is_Sc:plrd}', "");
Expect(1, 94112, '\p{^Is_Sc:plrd}', "");
Expect(1, 94112, '\P{Is_Sc:plrd}', "");
Expect(0, 94112, '\P{^Is_Sc:plrd}', "");
Expect(1, 94111, '\p{Is_Sc: Plrd}', "");
Expect(0, 94111, '\p{^Is_Sc: Plrd}', "");
Expect(0, 94111, '\P{Is_Sc: Plrd}', "");
Expect(1, 94111, '\P{^Is_Sc: Plrd}', "");
Expect(0, 94112, '\p{Is_Sc: Plrd}', "");
Expect(1, 94112, '\p{^Is_Sc: Plrd}', "");
Expect(1, 94112, '\P{Is_Sc: Plrd}', "");
Expect(0, 94112, '\P{^Is_Sc: Plrd}', "");
Error('\p{Script=/a/ _linear_b}');
Error('\P{Script=/a/ _linear_b}');
Expect(1, 65786, '\p{Script=linearb}', "");
Expect(0, 65786, '\p{^Script=linearb}', "");
Expect(0, 65786, '\P{Script=linearb}', "");
Expect(1, 65786, '\P{^Script=linearb}', "");
Expect(0, 65787, '\p{Script=linearb}', "");
Expect(1, 65787, '\p{^Script=linearb}', "");
Expect(1, 65787, '\P{Script=linearb}', "");
Expect(0, 65787, '\P{^Script=linearb}', "");
Expect(1, 65786, '\p{Script= Linear_B}', "");
Expect(0, 65786, '\p{^Script= Linear_B}', "");
Expect(0, 65786, '\P{Script= Linear_B}', "");
Expect(1, 65786, '\P{^Script= Linear_B}', "");
Expect(0, 65787, '\p{Script= Linear_B}', "");
Expect(1, 65787, '\p{^Script= Linear_B}', "");
Expect(1, 65787, '\P{Script= Linear_B}', "");
Expect(0, 65787, '\P{^Script= Linear_B}', "");
Error('\p{Sc=-	LINB/a/}');
Error('\P{Sc=-	LINB/a/}');
Expect(1, 65786, '\p{Sc=linb}', "");
Expect(0, 65786, '\p{^Sc=linb}', "");
Expect(0, 65786, '\P{Sc=linb}', "");
Expect(1, 65786, '\P{^Sc=linb}', "");
Expect(0, 65787, '\p{Sc=linb}', "");
Expect(1, 65787, '\p{^Sc=linb}', "");
Expect(1, 65787, '\P{Sc=linb}', "");
Expect(0, 65787, '\P{^Sc=linb}', "");
Expect(1, 65786, '\p{Sc:   -Linb}', "");
Expect(0, 65786, '\p{^Sc:   -Linb}', "");
Expect(0, 65786, '\P{Sc:   -Linb}', "");
Expect(1, 65786, '\P{^Sc:   -Linb}', "");
Expect(0, 65787, '\p{Sc:   -Linb}', "");
Expect(1, 65787, '\p{^Sc:   -Linb}', "");
Expect(1, 65787, '\P{Sc:   -Linb}', "");
Expect(0, 65787, '\P{^Sc:   -Linb}', "");
Error('\p{Is_Script:   /a/_	Linear_B}');
Error('\P{Is_Script:   /a/_	Linear_B}');
Expect(1, 65786, '\p{Is_Script=linearb}', "");
Expect(0, 65786, '\p{^Is_Script=linearb}', "");
Expect(0, 65786, '\P{Is_Script=linearb}', "");
Expect(1, 65786, '\P{^Is_Script=linearb}', "");
Expect(0, 65787, '\p{Is_Script=linearb}', "");
Expect(1, 65787, '\p{^Is_Script=linearb}', "");
Expect(1, 65787, '\P{Is_Script=linearb}', "");
Expect(0, 65787, '\P{^Is_Script=linearb}', "");
Expect(1, 65786, '\p{Is_Script=	 Linear_B}', "");
Expect(0, 65786, '\p{^Is_Script=	 Linear_B}', "");
Expect(0, 65786, '\P{Is_Script=	 Linear_B}', "");
Expect(1, 65786, '\P{^Is_Script=	 Linear_B}', "");
Expect(0, 65787, '\p{Is_Script=	 Linear_B}', "");
Expect(1, 65787, '\p{^Is_Script=	 Linear_B}', "");
Expect(1, 65787, '\P{Is_Script=	 Linear_B}', "");
Expect(0, 65787, '\P{^Is_Script=	 Linear_B}', "");
Error('\p{Is_Sc:/a/ Linb}');
Error('\P{Is_Sc:/a/ Linb}');
Expect(1, 65786, '\p{Is_Sc=linb}', "");
Expect(0, 65786, '\p{^Is_Sc=linb}', "");
Expect(0, 65786, '\P{Is_Sc=linb}', "");
Expect(1, 65786, '\P{^Is_Sc=linb}', "");
Expect(0, 65787, '\p{Is_Sc=linb}', "");
Expect(1, 65787, '\p{^Is_Sc=linb}', "");
Expect(1, 65787, '\P{Is_Sc=linb}', "");
Expect(0, 65787, '\P{^Is_Sc=linb}', "");
Expect(1, 65786, '\p{Is_Sc=-	LINB}', "");
Expect(0, 65786, '\p{^Is_Sc=-	LINB}', "");
Expect(0, 65786, '\P{Is_Sc=-	LINB}', "");
Expect(1, 65786, '\P{^Is_Sc=-	LINB}', "");
Expect(0, 65787, '\p{Is_Sc=-	LINB}', "");
Expect(1, 65787, '\p{^Is_Sc=-	LINB}', "");
Expect(1, 65787, '\P{Is_Sc=-	LINB}', "");
Expect(0, 65787, '\P{^Is_Sc=-	LINB}', "");
Error('\p{Script=-/a/Samaritan}');
Error('\P{Script=-/a/Samaritan}');
Expect(1, 2110, '\p{Script:   samaritan}', "");
Expect(0, 2110, '\p{^Script:   samaritan}', "");
Expect(0, 2110, '\P{Script:   samaritan}', "");
Expect(1, 2110, '\P{^Script:   samaritan}', "");
Expect(0, 2111, '\p{Script:   samaritan}', "");
Expect(1, 2111, '\p{^Script:   samaritan}', "");
Expect(1, 2111, '\P{Script:   samaritan}', "");
Expect(0, 2111, '\P{^Script:   samaritan}', "");
Expect(1, 2110, '\p{Script=--Samaritan}', "");
Expect(0, 2110, '\p{^Script=--Samaritan}', "");
Expect(0, 2110, '\P{Script=--Samaritan}', "");
Expect(1, 2110, '\P{^Script=--Samaritan}', "");
Expect(0, 2111, '\p{Script=--Samaritan}', "");
Expect(1, 2111, '\p{^Script=--Samaritan}', "");
Expect(1, 2111, '\P{Script=--Samaritan}', "");
Expect(0, 2111, '\P{^Script=--Samaritan}', "");
Error('\p{Sc=:=__Samr}');
Error('\P{Sc=:=__Samr}');
Expect(1, 2110, '\p{Sc=samr}', "");
Expect(0, 2110, '\p{^Sc=samr}', "");
Expect(0, 2110, '\P{Sc=samr}', "");
Expect(1, 2110, '\P{^Sc=samr}', "");
Expect(0, 2111, '\p{Sc=samr}', "");
Expect(1, 2111, '\p{^Sc=samr}', "");
Expect(1, 2111, '\P{Sc=samr}', "");
Expect(0, 2111, '\P{^Sc=samr}', "");
Expect(1, 2110, '\p{Sc=_ SAMR}', "");
Expect(0, 2110, '\p{^Sc=_ SAMR}', "");
Expect(0, 2110, '\P{Sc=_ SAMR}', "");
Expect(1, 2110, '\P{^Sc=_ SAMR}', "");
Expect(0, 2111, '\p{Sc=_ SAMR}', "");
Expect(1, 2111, '\p{^Sc=_ SAMR}', "");
Expect(1, 2111, '\P{Sc=_ SAMR}', "");
Expect(0, 2111, '\P{^Sc=_ SAMR}', "");
Error('\p{Is_Script=/a/	-SAMARITAN}');
Error('\P{Is_Script=/a/	-SAMARITAN}');
Expect(1, 2110, '\p{Is_Script=samaritan}', "");
Expect(0, 2110, '\p{^Is_Script=samaritan}', "");
Expect(0, 2110, '\P{Is_Script=samaritan}', "");
Expect(1, 2110, '\P{^Is_Script=samaritan}', "");
Expect(0, 2111, '\p{Is_Script=samaritan}', "");
Expect(1, 2111, '\p{^Is_Script=samaritan}', "");
Expect(1, 2111, '\P{Is_Script=samaritan}', "");
Expect(0, 2111, '\P{^Is_Script=samaritan}', "");
Expect(1, 2110, '\p{Is_Script=  samaritan}', "");
Expect(0, 2110, '\p{^Is_Script=  samaritan}', "");
Expect(0, 2110, '\P{Is_Script=  samaritan}', "");
Expect(1, 2110, '\P{^Is_Script=  samaritan}', "");
Expect(0, 2111, '\p{Is_Script=  samaritan}', "");
Expect(1, 2111, '\p{^Is_Script=  samaritan}', "");
Expect(1, 2111, '\P{Is_Script=  samaritan}', "");
Expect(0, 2111, '\P{^Is_Script=  samaritan}', "");
Error('\p{Is_Sc:   := _Samr}');
Error('\P{Is_Sc:   := _Samr}');
Expect(1, 2110, '\p{Is_Sc:samr}', "");
Expect(0, 2110, '\p{^Is_Sc:samr}', "");
Expect(0, 2110, '\P{Is_Sc:samr}', "");
Expect(1, 2110, '\P{^Is_Sc:samr}', "");
Expect(0, 2111, '\p{Is_Sc:samr}', "");
Expect(1, 2111, '\p{^Is_Sc:samr}', "");
Expect(1, 2111, '\P{Is_Sc:samr}', "");
Expect(0, 2111, '\P{^Is_Sc:samr}', "");
Expect(1, 2110, '\p{Is_Sc=-	Samr}', "");
Expect(0, 2110, '\p{^Is_Sc=-	Samr}', "");
Expect(0, 2110, '\P{Is_Sc=-	Samr}', "");
Expect(1, 2110, '\P{^Is_Sc=-	Samr}', "");
Expect(0, 2111, '\p{Is_Sc=-	Samr}', "");
Expect(1, 2111, '\p{^Is_Sc=-	Samr}', "");
Expect(1, 2111, '\P{Is_Sc=-	Samr}', "");
Expect(0, 2111, '\P{^Is_Sc=-	Samr}', "");
Error('\p{Script:	:=SAURASHTRA}');
Error('\P{Script:	:=SAURASHTRA}');
Expect(1, 43225, '\p{Script=saurashtra}', "");
Expect(0, 43225, '\p{^Script=saurashtra}', "");
Expect(0, 43225, '\P{Script=saurashtra}', "");
Expect(1, 43225, '\P{^Script=saurashtra}', "");
Expect(0, 43226, '\p{Script=saurashtra}', "");
Expect(1, 43226, '\p{^Script=saurashtra}', "");
Expect(1, 43226, '\P{Script=saurashtra}', "");
Expect(0, 43226, '\P{^Script=saurashtra}', "");
Expect(1, 43225, '\p{Script=	_SAURASHTRA}', "");
Expect(0, 43225, '\p{^Script=	_SAURASHTRA}', "");
Expect(0, 43225, '\P{Script=	_SAURASHTRA}', "");
Expect(1, 43225, '\P{^Script=	_SAURASHTRA}', "");
Expect(0, 43226, '\p{Script=	_SAURASHTRA}', "");
Expect(1, 43226, '\p{^Script=	_SAURASHTRA}', "");
Expect(1, 43226, '\P{Script=	_SAURASHTRA}', "");
Expect(0, 43226, '\P{^Script=	_SAURASHTRA}', "");
Error('\p{Sc: _-SAUR:=}');
Error('\P{Sc: _-SAUR:=}');
Expect(1, 43225, '\p{Sc=saur}', "");
Expect(0, 43225, '\p{^Sc=saur}', "");
Expect(0, 43225, '\P{Sc=saur}', "");
Expect(1, 43225, '\P{^Sc=saur}', "");
Expect(0, 43226, '\p{Sc=saur}', "");
Expect(1, 43226, '\p{^Sc=saur}', "");
Expect(1, 43226, '\P{Sc=saur}', "");
Expect(0, 43226, '\P{^Sc=saur}', "");
Expect(1, 43225, '\p{Sc=_	Saur}', "");
Expect(0, 43225, '\p{^Sc=_	Saur}', "");
Expect(0, 43225, '\P{Sc=_	Saur}', "");
Expect(1, 43225, '\P{^Sc=_	Saur}', "");
Expect(0, 43226, '\p{Sc=_	Saur}', "");
Expect(1, 43226, '\p{^Sc=_	Saur}', "");
Expect(1, 43226, '\P{Sc=_	Saur}', "");
Expect(0, 43226, '\P{^Sc=_	Saur}', "");
Error('\p{Is_Script=_	Saurashtra/a/}');
Error('\P{Is_Script=_	Saurashtra/a/}');
Expect(1, 43225, '\p{Is_Script=saurashtra}', "");
Expect(0, 43225, '\p{^Is_Script=saurashtra}', "");
Expect(0, 43225, '\P{Is_Script=saurashtra}', "");
Expect(1, 43225, '\P{^Is_Script=saurashtra}', "");
Expect(0, 43226, '\p{Is_Script=saurashtra}', "");
Expect(1, 43226, '\p{^Is_Script=saurashtra}', "");
Expect(1, 43226, '\P{Is_Script=saurashtra}', "");
Expect(0, 43226, '\P{^Is_Script=saurashtra}', "");
Expect(1, 43225, '\p{Is_Script=_saurashtra}', "");
Expect(0, 43225, '\p{^Is_Script=_saurashtra}', "");
Expect(0, 43225, '\P{Is_Script=_saurashtra}', "");
Expect(1, 43225, '\P{^Is_Script=_saurashtra}', "");
Expect(0, 43226, '\p{Is_Script=_saurashtra}', "");
Expect(1, 43226, '\p{^Is_Script=_saurashtra}', "");
Expect(1, 43226, '\P{Is_Script=_saurashtra}', "");
Expect(0, 43226, '\P{^Is_Script=_saurashtra}', "");
Error('\p{Is_Sc:   Saur:=}');
Error('\P{Is_Sc:   Saur:=}');
Expect(1, 43225, '\p{Is_Sc=saur}', "");
Expect(0, 43225, '\p{^Is_Sc=saur}', "");
Expect(0, 43225, '\P{Is_Sc=saur}', "");
Expect(1, 43225, '\P{^Is_Sc=saur}', "");
Expect(0, 43226, '\p{Is_Sc=saur}', "");
Expect(1, 43226, '\p{^Is_Sc=saur}', "");
Expect(1, 43226, '\P{Is_Sc=saur}', "");
Expect(0, 43226, '\P{^Is_Sc=saur}', "");
Expect(1, 43225, '\p{Is_Sc:    SAUR}', "");
Expect(0, 43225, '\p{^Is_Sc:    SAUR}', "");
Expect(0, 43225, '\P{Is_Sc:    SAUR}', "");
Expect(1, 43225, '\P{^Is_Sc:    SAUR}', "");
Expect(0, 43226, '\p{Is_Sc:    SAUR}', "");
Expect(1, 43226, '\p{^Is_Sc:    SAUR}', "");
Expect(1, 43226, '\P{Is_Sc:    SAUR}', "");
Expect(0, 43226, '\P{^Is_Sc:    SAUR}', "");
Error('\p{Script=_/a/sundanese}');
Error('\P{Script=_/a/sundanese}');
Expect(1, 7367, '\p{Script=sundanese}', "");
Expect(0, 7367, '\p{^Script=sundanese}', "");
Expect(0, 7367, '\P{Script=sundanese}', "");
Expect(1, 7367, '\P{^Script=sundanese}', "");
Expect(0, 7368, '\p{Script=sundanese}', "");
Expect(1, 7368, '\p{^Script=sundanese}', "");
Expect(1, 7368, '\P{Script=sundanese}', "");
Expect(0, 7368, '\P{^Script=sundanese}', "");
Expect(1, 7367, '\p{Script=- SUNDANESE}', "");
Expect(0, 7367, '\p{^Script=- SUNDANESE}', "");
Expect(0, 7367, '\P{Script=- SUNDANESE}', "");
Expect(1, 7367, '\P{^Script=- SUNDANESE}', "");
Expect(0, 7368, '\p{Script=- SUNDANESE}', "");
Expect(1, 7368, '\p{^Script=- SUNDANESE}', "");
Expect(1, 7368, '\P{Script=- SUNDANESE}', "");
Expect(0, 7368, '\P{^Script=- SUNDANESE}', "");
Error('\p{Sc= _sund/a/}');
Error('\P{Sc= _sund/a/}');
Expect(1, 7367, '\p{Sc=sund}', "");
Expect(0, 7367, '\p{^Sc=sund}', "");
Expect(0, 7367, '\P{Sc=sund}', "");
Expect(1, 7367, '\P{^Sc=sund}', "");
Expect(0, 7368, '\p{Sc=sund}', "");
Expect(1, 7368, '\p{^Sc=sund}', "");
Expect(1, 7368, '\P{Sc=sund}', "");
Expect(0, 7368, '\P{^Sc=sund}', "");
Expect(1, 7367, '\p{Sc=	-Sund}', "");
Expect(0, 7367, '\p{^Sc=	-Sund}', "");
Expect(0, 7367, '\P{Sc=	-Sund}', "");
Expect(1, 7367, '\P{^Sc=	-Sund}', "");
Expect(0, 7368, '\p{Sc=	-Sund}', "");
Expect(1, 7368, '\p{^Sc=	-Sund}', "");
Expect(1, 7368, '\P{Sc=	-Sund}', "");
Expect(0, 7368, '\P{^Sc=	-Sund}', "");
Error('\p{Is_Script=  sundanese/a/}');
Error('\P{Is_Script=  sundanese/a/}');
Expect(1, 7367, '\p{Is_Script=sundanese}', "");
Expect(0, 7367, '\p{^Is_Script=sundanese}', "");
Expect(0, 7367, '\P{Is_Script=sundanese}', "");
Expect(1, 7367, '\P{^Is_Script=sundanese}', "");
Expect(0, 7368, '\p{Is_Script=sundanese}', "");
Expect(1, 7368, '\p{^Is_Script=sundanese}', "");
Expect(1, 7368, '\P{Is_Script=sundanese}', "");
Expect(0, 7368, '\P{^Is_Script=sundanese}', "");
Expect(1, 7367, '\p{Is_Script=-Sundanese}', "");
Expect(0, 7367, '\p{^Is_Script=-Sundanese}', "");
Expect(0, 7367, '\P{Is_Script=-Sundanese}', "");
Expect(1, 7367, '\P{^Is_Script=-Sundanese}', "");
Expect(0, 7368, '\p{Is_Script=-Sundanese}', "");
Expect(1, 7368, '\p{^Is_Script=-Sundanese}', "");
Expect(1, 7368, '\P{Is_Script=-Sundanese}', "");
Expect(0, 7368, '\P{^Is_Script=-Sundanese}', "");
Error('\p{Is_Sc= 	Sund:=}');
Error('\P{Is_Sc= 	Sund:=}');
Expect(1, 7367, '\p{Is_Sc=sund}', "");
Expect(0, 7367, '\p{^Is_Sc=sund}', "");
Expect(0, 7367, '\P{Is_Sc=sund}', "");
Expect(1, 7367, '\P{^Is_Sc=sund}', "");
Expect(0, 7368, '\p{Is_Sc=sund}', "");
Expect(1, 7368, '\p{^Is_Sc=sund}', "");
Expect(1, 7368, '\P{Is_Sc=sund}', "");
Expect(0, 7368, '\P{^Is_Sc=sund}', "");
Expect(1, 7367, '\p{Is_Sc: _-Sund}', "");
Expect(0, 7367, '\p{^Is_Sc: _-Sund}', "");
Expect(0, 7367, '\P{Is_Sc: _-Sund}', "");
Expect(1, 7367, '\P{^Is_Sc: _-Sund}', "");
Expect(0, 7368, '\p{Is_Sc: _-Sund}', "");
Expect(1, 7368, '\p{^Is_Sc: _-Sund}', "");
Expect(1, 7368, '\P{Is_Sc: _-Sund}', "");
Expect(0, 7368, '\P{^Is_Sc: _-Sund}', "");
Error('\p{Script=/a/nko}');
Error('\P{Script=/a/nko}');
Expect(1, 2042, '\p{Script=nko}', "");
Expect(0, 2042, '\p{^Script=nko}', "");
Expect(0, 2042, '\P{Script=nko}', "");
Expect(1, 2042, '\P{^Script=nko}', "");
Expect(0, 2043, '\p{Script=nko}', "");
Expect(1, 2043, '\p{^Script=nko}', "");
Expect(1, 2043, '\P{Script=nko}', "");
Expect(0, 2043, '\P{^Script=nko}', "");
Expect(1, 2042, '\p{Script= -NKO}', "");
Expect(0, 2042, '\p{^Script= -NKO}', "");
Expect(0, 2042, '\P{Script= -NKO}', "");
Expect(1, 2042, '\P{^Script= -NKO}', "");
Expect(0, 2043, '\p{Script= -NKO}', "");
Expect(1, 2043, '\p{^Script= -NKO}', "");
Expect(1, 2043, '\P{Script= -NKO}', "");
Expect(0, 2043, '\P{^Script= -NKO}', "");
Error('\p{Sc=_:=Nkoo}');
Error('\P{Sc=_:=Nkoo}');
Expect(1, 2042, '\p{Sc=nkoo}', "");
Expect(0, 2042, '\p{^Sc=nkoo}', "");
Expect(0, 2042, '\P{Sc=nkoo}', "");
Expect(1, 2042, '\P{^Sc=nkoo}', "");
Expect(0, 2043, '\p{Sc=nkoo}', "");
Expect(1, 2043, '\p{^Sc=nkoo}', "");
Expect(1, 2043, '\P{Sc=nkoo}', "");
Expect(0, 2043, '\P{^Sc=nkoo}', "");
Expect(1, 2042, '\p{Sc=  nkoo}', "");
Expect(0, 2042, '\p{^Sc=  nkoo}', "");
Expect(0, 2042, '\P{Sc=  nkoo}', "");
Expect(1, 2042, '\P{^Sc=  nkoo}', "");
Expect(0, 2043, '\p{Sc=  nkoo}', "");
Expect(1, 2043, '\p{^Sc=  nkoo}', "");
Expect(1, 2043, '\P{Sc=  nkoo}', "");
Expect(0, 2043, '\P{^Sc=  nkoo}', "");
Error('\p{Is_Script=-_Nko:=}');
Error('\P{Is_Script=-_Nko:=}');
Expect(1, 2042, '\p{Is_Script=nko}', "");
Expect(0, 2042, '\p{^Is_Script=nko}', "");
Expect(0, 2042, '\P{Is_Script=nko}', "");
Expect(1, 2042, '\P{^Is_Script=nko}', "");
Expect(0, 2043, '\p{Is_Script=nko}', "");
Expect(1, 2043, '\p{^Is_Script=nko}', "");
Expect(1, 2043, '\P{Is_Script=nko}', "");
Expect(0, 2043, '\P{^Is_Script=nko}', "");
Expect(1, 2042, '\p{Is_Script= Nko}', "");
Expect(0, 2042, '\p{^Is_Script= Nko}', "");
Expect(0, 2042, '\P{Is_Script= Nko}', "");
Expect(1, 2042, '\P{^Is_Script= Nko}', "");
Expect(0, 2043, '\p{Is_Script= Nko}', "");
Expect(1, 2043, '\p{^Is_Script= Nko}', "");
Expect(1, 2043, '\P{Is_Script= Nko}', "");
Expect(0, 2043, '\P{^Is_Script= Nko}', "");
Error('\p{Is_Sc=_nkoo/a/}');
Error('\P{Is_Sc=_nkoo/a/}');
Expect(1, 2042, '\p{Is_Sc=nkoo}', "");
Expect(0, 2042, '\p{^Is_Sc=nkoo}', "");
Expect(0, 2042, '\P{Is_Sc=nkoo}', "");
Expect(1, 2042, '\P{^Is_Sc=nkoo}', "");
Expect(0, 2043, '\p{Is_Sc=nkoo}', "");
Expect(1, 2043, '\p{^Is_Sc=nkoo}', "");
Expect(1, 2043, '\P{Is_Sc=nkoo}', "");
Expect(0, 2043, '\P{^Is_Sc=nkoo}', "");
Expect(1, 2042, '\p{Is_Sc=	Nkoo}', "");
Expect(0, 2042, '\p{^Is_Sc=	Nkoo}', "");
Expect(0, 2042, '\P{Is_Sc=	Nkoo}', "");
Expect(1, 2042, '\P{^Is_Sc=	Nkoo}', "");
Expect(0, 2043, '\p{Is_Sc=	Nkoo}', "");
Expect(1, 2043, '\p{^Is_Sc=	Nkoo}', "");
Expect(1, 2043, '\P{Is_Sc=	Nkoo}', "");
Expect(0, 2043, '\P{^Is_Sc=	Nkoo}', "");
Error('\p{Script=	/a/han}');
Error('\P{Script=	/a/han}');
Expect(1, 195101, '\p{Script=han}', "");
Expect(0, 195101, '\p{^Script=han}', "");
Expect(0, 195101, '\P{Script=han}', "");
Expect(1, 195101, '\P{^Script=han}', "");
Expect(0, 195102, '\p{Script=han}', "");
Expect(1, 195102, '\p{^Script=han}', "");
Expect(1, 195102, '\P{Script=han}', "");
Expect(0, 195102, '\P{^Script=han}', "");
Expect(1, 195101, '\p{Script=	Han}', "");
Expect(0, 195101, '\p{^Script=	Han}', "");
Expect(0, 195101, '\P{Script=	Han}', "");
Expect(1, 195101, '\P{^Script=	Han}', "");
Expect(0, 195102, '\p{Script=	Han}', "");
Expect(1, 195102, '\p{^Script=	Han}', "");
Expect(1, 195102, '\P{Script=	Han}', "");
Expect(0, 195102, '\P{^Script=	Han}', "");
Error('\p{Sc:/a/_hani}');
Error('\P{Sc:/a/_hani}');
Expect(1, 195101, '\p{Sc=hani}', "");
Expect(0, 195101, '\p{^Sc=hani}', "");
Expect(0, 195101, '\P{Sc=hani}', "");
Expect(1, 195101, '\P{^Sc=hani}', "");
Expect(0, 195102, '\p{Sc=hani}', "");
Expect(1, 195102, '\p{^Sc=hani}', "");
Expect(1, 195102, '\P{Sc=hani}', "");
Expect(0, 195102, '\P{^Sc=hani}', "");
Expect(1, 195101, '\p{Sc= HANI}', "");
Expect(0, 195101, '\p{^Sc= HANI}', "");
Expect(0, 195101, '\P{Sc= HANI}', "");
Expect(1, 195101, '\P{^Sc= HANI}', "");
Expect(0, 195102, '\p{Sc= HANI}', "");
Expect(1, 195102, '\p{^Sc= HANI}', "");
Expect(1, 195102, '\P{Sc= HANI}', "");
Expect(0, 195102, '\P{^Sc= HANI}', "");
Error('\p{Is_Script= _han:=}');
Error('\P{Is_Script= _han:=}');
Expect(1, 195101, '\p{Is_Script=han}', "");
Expect(0, 195101, '\p{^Is_Script=han}', "");
Expect(0, 195101, '\P{Is_Script=han}', "");
Expect(1, 195101, '\P{^Is_Script=han}', "");
Expect(0, 195102, '\p{Is_Script=han}', "");
Expect(1, 195102, '\p{^Is_Script=han}', "");
Expect(1, 195102, '\P{Is_Script=han}', "");
Expect(0, 195102, '\P{^Is_Script=han}', "");
Expect(1, 195101, '\p{Is_Script=_Han}', "");
Expect(0, 195101, '\p{^Is_Script=_Han}', "");
Expect(0, 195101, '\P{Is_Script=_Han}', "");
Expect(1, 195101, '\P{^Is_Script=_Han}', "");
Expect(0, 195102, '\p{Is_Script=_Han}', "");
Expect(1, 195102, '\p{^Is_Script=_Han}', "");
Expect(1, 195102, '\P{Is_Script=_Han}', "");
Expect(0, 195102, '\P{^Is_Script=_Han}', "");
Error('\p{Is_Sc=-HANI/a/}');
Error('\P{Is_Sc=-HANI/a/}');
Expect(1, 195101, '\p{Is_Sc=hani}', "");
Expect(0, 195101, '\p{^Is_Sc=hani}', "");
Expect(0, 195101, '\P{Is_Sc=hani}', "");
Expect(1, 195101, '\P{^Is_Sc=hani}', "");
Expect(0, 195102, '\p{Is_Sc=hani}', "");
Expect(1, 195102, '\p{^Is_Sc=hani}', "");
Expect(1, 195102, '\P{Is_Sc=hani}', "");
Expect(0, 195102, '\P{^Is_Sc=hani}', "");
Expect(1, 195101, '\p{Is_Sc=-	hani}', "");
Expect(0, 195101, '\p{^Is_Sc=-	hani}', "");
Expect(0, 195101, '\P{Is_Sc=-	hani}', "");
Expect(1, 195101, '\P{^Is_Sc=-	hani}', "");
Expect(0, 195102, '\p{Is_Sc=-	hani}', "");
Expect(1, 195102, '\p{^Is_Sc=-	hani}', "");
Expect(1, 195102, '\P{Is_Sc=-	hani}', "");
Expect(0, 195102, '\P{^Is_Sc=-	hani}', "");
Error('\p{Script=-takri/a/}');
Error('\P{Script=-takri/a/}');
Expect(1, 71369, '\p{Script=takri}', "");
Expect(0, 71369, '\p{^Script=takri}', "");
Expect(0, 71369, '\P{Script=takri}', "");
Expect(1, 71369, '\P{^Script=takri}', "");
Expect(0, 71370, '\p{Script=takri}', "");
Expect(1, 71370, '\p{^Script=takri}', "");
Expect(1, 71370, '\P{Script=takri}', "");
Expect(0, 71370, '\P{^Script=takri}', "");
Expect(1, 71369, '\p{Script=--Takri}', "");
Expect(0, 71369, '\p{^Script=--Takri}', "");
Expect(0, 71369, '\P{Script=--Takri}', "");
Expect(1, 71369, '\P{^Script=--Takri}', "");
Expect(0, 71370, '\p{Script=--Takri}', "");
Expect(1, 71370, '\p{^Script=--Takri}', "");
Expect(1, 71370, '\P{Script=--Takri}', "");
Expect(0, 71370, '\P{^Script=--Takri}', "");
Error('\p{Sc= 	Takr:=}');
Error('\P{Sc= 	Takr:=}');
Expect(1, 71369, '\p{Sc=takr}', "");
Expect(0, 71369, '\p{^Sc=takr}', "");
Expect(0, 71369, '\P{Sc=takr}', "");
Expect(1, 71369, '\P{^Sc=takr}', "");
Expect(0, 71370, '\p{Sc=takr}', "");
Expect(1, 71370, '\p{^Sc=takr}', "");
Expect(1, 71370, '\P{Sc=takr}', "");
Expect(0, 71370, '\P{^Sc=takr}', "");
Expect(1, 71369, '\p{Sc= _Takr}', "");
Expect(0, 71369, '\p{^Sc= _Takr}', "");
Expect(0, 71369, '\P{Sc= _Takr}', "");
Expect(1, 71369, '\P{^Sc= _Takr}', "");
Expect(0, 71370, '\p{Sc= _Takr}', "");
Expect(1, 71370, '\p{^Sc= _Takr}', "");
Expect(1, 71370, '\P{Sc= _Takr}', "");
Expect(0, 71370, '\P{^Sc= _Takr}', "");
Error('\p{Is_Script:   _	Takri:=}');
Error('\P{Is_Script:   _	Takri:=}');
Expect(1, 71369, '\p{Is_Script:	takri}', "");
Expect(0, 71369, '\p{^Is_Script:	takri}', "");
Expect(0, 71369, '\P{Is_Script:	takri}', "");
Expect(1, 71369, '\P{^Is_Script:	takri}', "");
Expect(0, 71370, '\p{Is_Script:	takri}', "");
Expect(1, 71370, '\p{^Is_Script:	takri}', "");
Expect(1, 71370, '\P{Is_Script:	takri}', "");
Expect(0, 71370, '\P{^Is_Script:	takri}', "");
Expect(1, 71369, '\p{Is_Script=_ Takri}', "");
Expect(0, 71369, '\p{^Is_Script=_ Takri}', "");
Expect(0, 71369, '\P{Is_Script=_ Takri}', "");
Expect(1, 71369, '\P{^Is_Script=_ Takri}', "");
Expect(0, 71370, '\p{Is_Script=_ Takri}', "");
Expect(1, 71370, '\p{^Is_Script=_ Takri}', "");
Expect(1, 71370, '\P{Is_Script=_ Takri}', "");
Expect(0, 71370, '\P{^Is_Script=_ Takri}', "");
Error('\p{Is_Sc=_/a/Takr}');
Error('\P{Is_Sc=_/a/Takr}');
Expect(1, 71369, '\p{Is_Sc=takr}', "");
Expect(0, 71369, '\p{^Is_Sc=takr}', "");
Expect(0, 71369, '\P{Is_Sc=takr}', "");
Expect(1, 71369, '\P{^Is_Sc=takr}', "");
Expect(0, 71370, '\p{Is_Sc=takr}', "");
Expect(1, 71370, '\p{^Is_Sc=takr}', "");
Expect(1, 71370, '\P{Is_Sc=takr}', "");
Expect(0, 71370, '\P{^Is_Sc=takr}', "");
Expect(1, 71369, '\p{Is_Sc= 	takr}', "");
Expect(0, 71369, '\p{^Is_Sc= 	takr}', "");
Expect(0, 71369, '\P{Is_Sc= 	takr}', "");
Expect(1, 71369, '\P{^Is_Sc= 	takr}', "");
Expect(0, 71370, '\p{Is_Sc= 	takr}', "");
Expect(1, 71370, '\p{^Is_Sc= 	takr}', "");
Expect(1, 71370, '\P{Is_Sc= 	takr}', "");
Expect(0, 71370, '\P{^Is_Sc= 	takr}', "");
Error('\p{Script=	Greek/a/}');
Error('\P{Script=	Greek/a/}');
Expect(1, 119365, '\p{Script=greek}', "");
Expect(0, 119365, '\p{^Script=greek}', "");
Expect(0, 119365, '\P{Script=greek}', "");
Expect(1, 119365, '\P{^Script=greek}', "");
Expect(0, 119366, '\p{Script=greek}', "");
Expect(1, 119366, '\p{^Script=greek}', "");
Expect(1, 119366, '\P{Script=greek}', "");
Expect(0, 119366, '\P{^Script=greek}', "");
Expect(1, 119365, '\p{Script=_ greek}', "");
Expect(0, 119365, '\p{^Script=_ greek}', "");
Expect(0, 119365, '\P{Script=_ greek}', "");
Expect(1, 119365, '\P{^Script=_ greek}', "");
Expect(0, 119366, '\p{Script=_ greek}', "");
Expect(1, 119366, '\p{^Script=_ greek}', "");
Expect(1, 119366, '\P{Script=_ greek}', "");
Expect(0, 119366, '\P{^Script=_ greek}', "");
Error('\p{Sc= _grek/a/}');
Error('\P{Sc= _grek/a/}');
Expect(1, 119365, '\p{Sc=grek}', "");
Expect(0, 119365, '\p{^Sc=grek}', "");
Expect(0, 119365, '\P{Sc=grek}', "");
Expect(1, 119365, '\P{^Sc=grek}', "");
Expect(0, 119366, '\p{Sc=grek}', "");
Expect(1, 119366, '\p{^Sc=grek}', "");
Expect(1, 119366, '\P{Sc=grek}', "");
Expect(0, 119366, '\P{^Sc=grek}', "");
Expect(1, 119365, '\p{Sc= Grek}', "");
Expect(0, 119365, '\p{^Sc= Grek}', "");
Expect(0, 119365, '\P{Sc= Grek}', "");
Expect(1, 119365, '\P{^Sc= Grek}', "");
Expect(0, 119366, '\p{Sc= Grek}', "");
Expect(1, 119366, '\p{^Sc= Grek}', "");
Expect(1, 119366, '\P{Sc= Grek}', "");
Expect(0, 119366, '\P{^Sc= Grek}', "");
Error('\p{Is_Script:	-_Greek:=}');
Error('\P{Is_Script:	-_Greek:=}');
Expect(1, 119365, '\p{Is_Script=greek}', "");
Expect(0, 119365, '\p{^Is_Script=greek}', "");
Expect(0, 119365, '\P{Is_Script=greek}', "");
Expect(1, 119365, '\P{^Is_Script=greek}', "");
Expect(0, 119366, '\p{Is_Script=greek}', "");
Expect(1, 119366, '\p{^Is_Script=greek}', "");
Expect(1, 119366, '\P{Is_Script=greek}', "");
Expect(0, 119366, '\P{^Is_Script=greek}', "");
Expect(1, 119365, '\p{Is_Script=		Greek}', "");
Expect(0, 119365, '\p{^Is_Script=		Greek}', "");
Expect(0, 119365, '\P{Is_Script=		Greek}', "");
Expect(1, 119365, '\P{^Is_Script=		Greek}', "");
Expect(0, 119366, '\p{Is_Script=		Greek}', "");
Expect(1, 119366, '\p{^Is_Script=		Greek}', "");
Expect(1, 119366, '\P{Is_Script=		Greek}', "");
Expect(0, 119366, '\P{^Is_Script=		Greek}', "");
Error('\p{Is_Sc=/a/ _Grek}');
Error('\P{Is_Sc=/a/ _Grek}');
Expect(1, 119365, '\p{Is_Sc:grek}', "");
Expect(0, 119365, '\p{^Is_Sc:grek}', "");
Expect(0, 119365, '\P{Is_Sc:grek}', "");
Expect(1, 119365, '\P{^Is_Sc:grek}', "");
Expect(0, 119366, '\p{Is_Sc:grek}', "");
Expect(1, 119366, '\p{^Is_Sc:grek}', "");
Expect(1, 119366, '\P{Is_Sc:grek}', "");
Expect(0, 119366, '\P{^Is_Sc:grek}', "");
Expect(1, 119365, '\p{Is_Sc= -GREK}', "");
Expect(0, 119365, '\p{^Is_Sc= -GREK}', "");
Expect(0, 119365, '\P{Is_Sc= -GREK}', "");
Expect(1, 119365, '\P{^Is_Sc= -GREK}', "");
Expect(0, 119366, '\p{Is_Sc= -GREK}', "");
Expect(1, 119366, '\p{^Is_Sc= -GREK}', "");
Expect(1, 119366, '\P{Is_Sc= -GREK}', "");
Expect(0, 119366, '\P{^Is_Sc= -GREK}', "");
Error('\p{Script=/a/ _Tamil}');
Error('\P{Script=/a/ _Tamil}');
Expect(1, 3066, '\p{Script:	tamil}', "");
Expect(0, 3066, '\p{^Script:	tamil}', "");
Expect(0, 3066, '\P{Script:	tamil}', "");
Expect(1, 3066, '\P{^Script:	tamil}', "");
Expect(0, 3067, '\p{Script:	tamil}', "");
Expect(1, 3067, '\p{^Script:	tamil}', "");
Expect(1, 3067, '\P{Script:	tamil}', "");
Expect(0, 3067, '\P{^Script:	tamil}', "");
Expect(1, 3066, '\p{Script=-_TAMIL}', "");
Expect(0, 3066, '\p{^Script=-_TAMIL}', "");
Expect(0, 3066, '\P{Script=-_TAMIL}', "");
Expect(1, 3066, '\P{^Script=-_TAMIL}', "");
Expect(0, 3067, '\p{Script=-_TAMIL}', "");
Expect(1, 3067, '\p{^Script=-_TAMIL}', "");
Expect(1, 3067, '\P{Script=-_TAMIL}', "");
Expect(0, 3067, '\P{^Script=-_TAMIL}', "");
Error('\p{Sc:  	taml/a/}');
Error('\P{Sc:  	taml/a/}');
Expect(1, 3066, '\p{Sc=taml}', "");
Expect(0, 3066, '\p{^Sc=taml}', "");
Expect(0, 3066, '\P{Sc=taml}', "");
Expect(1, 3066, '\P{^Sc=taml}', "");
Expect(0, 3067, '\p{Sc=taml}', "");
Expect(1, 3067, '\p{^Sc=taml}', "");
Expect(1, 3067, '\P{Sc=taml}', "");
Expect(0, 3067, '\P{^Sc=taml}', "");
Expect(1, 3066, '\p{Sc=_Taml}', "");
Expect(0, 3066, '\p{^Sc=_Taml}', "");
Expect(0, 3066, '\P{Sc=_Taml}', "");
Expect(1, 3066, '\P{^Sc=_Taml}', "");
Expect(0, 3067, '\p{Sc=_Taml}', "");
Expect(1, 3067, '\p{^Sc=_Taml}', "");
Expect(1, 3067, '\P{Sc=_Taml}', "");
Expect(0, 3067, '\P{^Sc=_Taml}', "");
Error('\p{Is_Script=:=		tamil}');
Error('\P{Is_Script=:=		tamil}');
Expect(1, 3066, '\p{Is_Script=tamil}', "");
Expect(0, 3066, '\p{^Is_Script=tamil}', "");
Expect(0, 3066, '\P{Is_Script=tamil}', "");
Expect(1, 3066, '\P{^Is_Script=tamil}', "");
Expect(0, 3067, '\p{Is_Script=tamil}', "");
Expect(1, 3067, '\p{^Is_Script=tamil}', "");
Expect(1, 3067, '\P{Is_Script=tamil}', "");
Expect(0, 3067, '\P{^Is_Script=tamil}', "");
Expect(1, 3066, '\p{Is_Script=_	tamil}', "");
Expect(0, 3066, '\p{^Is_Script=_	tamil}', "");
Expect(0, 3066, '\P{Is_Script=_	tamil}', "");
Expect(1, 3066, '\P{^Is_Script=_	tamil}', "");
Expect(0, 3067, '\p{Is_Script=_	tamil}', "");
Expect(1, 3067, '\p{^Is_Script=_	tamil}', "");
Expect(1, 3067, '\P{Is_Script=_	tamil}', "");
Expect(0, 3067, '\P{^Is_Script=_	tamil}', "");
Error('\p{Is_Sc: /a/Taml}');
Error('\P{Is_Sc: /a/Taml}');
Expect(1, 3066, '\p{Is_Sc=taml}', "");
Expect(0, 3066, '\p{^Is_Sc=taml}', "");
Expect(0, 3066, '\P{Is_Sc=taml}', "");
Expect(1, 3066, '\P{^Is_Sc=taml}', "");
Expect(0, 3067, '\p{Is_Sc=taml}', "");
Expect(1, 3067, '\p{^Is_Sc=taml}', "");
Expect(1, 3067, '\P{Is_Sc=taml}', "");
Expect(0, 3067, '\P{^Is_Sc=taml}', "");
Expect(1, 3066, '\p{Is_Sc=-	Taml}', "");
Expect(0, 3066, '\p{^Is_Sc=-	Taml}', "");
Expect(0, 3066, '\P{Is_Sc=-	Taml}', "");
Expect(1, 3066, '\P{^Is_Sc=-	Taml}', "");
Expect(0, 3067, '\p{Is_Sc=-	Taml}', "");
Expect(1, 3067, '\p{^Is_Sc=-	Taml}', "");
Expect(1, 3067, '\P{Is_Sc=-	Taml}', "");
Expect(0, 3067, '\P{^Is_Sc=-	Taml}', "");
Error('\p{Script=/a/-Kayah_Li}');
Error('\P{Script=/a/-Kayah_Li}');
Expect(1, 43311, '\p{Script=kayahli}', "");
Expect(0, 43311, '\p{^Script=kayahli}', "");
Expect(0, 43311, '\P{Script=kayahli}', "");
Expect(1, 43311, '\P{^Script=kayahli}', "");
Expect(0, 43312, '\p{Script=kayahli}', "");
Expect(1, 43312, '\p{^Script=kayahli}', "");
Expect(1, 43312, '\P{Script=kayahli}', "");
Expect(0, 43312, '\P{^Script=kayahli}', "");
Expect(1, 43311, '\p{Script=_ Kayah_LI}', "");
Expect(0, 43311, '\p{^Script=_ Kayah_LI}', "");
Expect(0, 43311, '\P{Script=_ Kayah_LI}', "");
Expect(1, 43311, '\P{^Script=_ Kayah_LI}', "");
Expect(0, 43312, '\p{Script=_ Kayah_LI}', "");
Expect(1, 43312, '\p{^Script=_ Kayah_LI}', "");
Expect(1, 43312, '\P{Script=_ Kayah_LI}', "");
Expect(0, 43312, '\P{^Script=_ Kayah_LI}', "");
Error('\p{Sc=/a/_KALI}');
Error('\P{Sc=/a/_KALI}');
Expect(1, 43311, '\p{Sc=kali}', "");
Expect(0, 43311, '\p{^Sc=kali}', "");
Expect(0, 43311, '\P{Sc=kali}', "");
Expect(1, 43311, '\P{^Sc=kali}', "");
Expect(0, 43312, '\p{Sc=kali}', "");
Expect(1, 43312, '\p{^Sc=kali}', "");
Expect(1, 43312, '\P{Sc=kali}', "");
Expect(0, 43312, '\P{^Sc=kali}', "");
Expect(1, 43311, '\p{Sc=	Kali}', "");
Expect(0, 43311, '\p{^Sc=	Kali}', "");
Expect(0, 43311, '\P{Sc=	Kali}', "");
Expect(1, 43311, '\P{^Sc=	Kali}', "");
Expect(0, 43312, '\p{Sc=	Kali}', "");
Expect(1, 43312, '\p{^Sc=	Kali}', "");
Expect(1, 43312, '\P{Sc=	Kali}', "");
Expect(0, 43312, '\P{^Sc=	Kali}', "");
Error('\p{Is_Script=-/a/KAYAH_Li}');
Error('\P{Is_Script=-/a/KAYAH_Li}');
Expect(1, 43311, '\p{Is_Script=kayahli}', "");
Expect(0, 43311, '\p{^Is_Script=kayahli}', "");
Expect(0, 43311, '\P{Is_Script=kayahli}', "");
Expect(1, 43311, '\P{^Is_Script=kayahli}', "");
Expect(0, 43312, '\p{Is_Script=kayahli}', "");
Expect(1, 43312, '\p{^Is_Script=kayahli}', "");
Expect(1, 43312, '\P{Is_Script=kayahli}', "");
Expect(0, 43312, '\P{^Is_Script=kayahli}', "");
Expect(1, 43311, '\p{Is_Script=_ Kayah_li}', "");
Expect(0, 43311, '\p{^Is_Script=_ Kayah_li}', "");
Expect(0, 43311, '\P{Is_Script=_ Kayah_li}', "");
Expect(1, 43311, '\P{^Is_Script=_ Kayah_li}', "");
Expect(0, 43312, '\p{Is_Script=_ Kayah_li}', "");
Expect(1, 43312, '\p{^Is_Script=_ Kayah_li}', "");
Expect(1, 43312, '\P{Is_Script=_ Kayah_li}', "");
Expect(0, 43312, '\P{^Is_Script=_ Kayah_li}', "");
Error('\p{Is_Sc=/a/	-Kali}');
Error('\P{Is_Sc=/a/	-Kali}');
Expect(1, 43311, '\p{Is_Sc=kali}', "");
Expect(0, 43311, '\p{^Is_Sc=kali}', "");
Expect(0, 43311, '\P{Is_Sc=kali}', "");
Expect(1, 43311, '\P{^Is_Sc=kali}', "");
Expect(0, 43312, '\p{Is_Sc=kali}', "");
Expect(1, 43312, '\p{^Is_Sc=kali}', "");
Expect(1, 43312, '\P{Is_Sc=kali}', "");
Expect(0, 43312, '\P{^Is_Sc=kali}', "");
Expect(1, 43311, '\p{Is_Sc: 	-KALI}', "");
Expect(0, 43311, '\p{^Is_Sc: 	-KALI}', "");
Expect(0, 43311, '\P{Is_Sc: 	-KALI}', "");
Expect(1, 43311, '\P{^Is_Sc: 	-KALI}', "");
Expect(0, 43312, '\p{Is_Sc: 	-KALI}', "");
Expect(1, 43312, '\p{^Is_Sc: 	-KALI}', "");
Expect(1, 43312, '\P{Is_Sc: 	-KALI}', "");
Expect(0, 43312, '\P{^Is_Sc: 	-KALI}', "");
Error('\p{Script= Thai:=}');
Error('\P{Script= Thai:=}');
Expect(1, 3675, '\p{Script:thai}', "");
Expect(0, 3675, '\p{^Script:thai}', "");
Expect(0, 3675, '\P{Script:thai}', "");
Expect(1, 3675, '\P{^Script:thai}', "");
Expect(0, 3676, '\p{Script:thai}', "");
Expect(1, 3676, '\p{^Script:thai}', "");
Expect(1, 3676, '\P{Script:thai}', "");
Expect(0, 3676, '\P{^Script:thai}', "");
Expect(1, 3675, '\p{Script=-	Thai}', "");
Expect(0, 3675, '\p{^Script=-	Thai}', "");
Expect(0, 3675, '\P{Script=-	Thai}', "");
Expect(1, 3675, '\P{^Script=-	Thai}', "");
Expect(0, 3676, '\p{Script=-	Thai}', "");
Expect(1, 3676, '\p{^Script=-	Thai}', "");
Expect(1, 3676, '\P{Script=-	Thai}', "");
Expect(0, 3676, '\P{^Script=-	Thai}', "");
Error('\p{Sc=_-Thai:=}');
Error('\P{Sc=_-Thai:=}');
Expect(1, 3675, '\p{Sc=thai}', "");
Expect(0, 3675, '\p{^Sc=thai}', "");
Expect(0, 3675, '\P{Sc=thai}', "");
Expect(1, 3675, '\P{^Sc=thai}', "");
Expect(0, 3676, '\p{Sc=thai}', "");
Expect(1, 3676, '\p{^Sc=thai}', "");
Expect(1, 3676, '\P{Sc=thai}', "");
Expect(0, 3676, '\P{^Sc=thai}', "");
Expect(1, 3675, '\p{Sc= Thai}', "");
Expect(0, 3675, '\p{^Sc= Thai}', "");
Expect(0, 3675, '\P{Sc= Thai}', "");
Expect(1, 3675, '\P{^Sc= Thai}', "");
Expect(0, 3676, '\p{Sc= Thai}', "");
Expect(1, 3676, '\p{^Sc= Thai}', "");
Expect(1, 3676, '\P{Sc= Thai}', "");
Expect(0, 3676, '\P{^Sc= Thai}', "");
Error('\p{Is_Script=_	Thai:=}');
Error('\P{Is_Script=_	Thai:=}');
Expect(1, 3675, '\p{Is_Script=thai}', "");
Expect(0, 3675, '\p{^Is_Script=thai}', "");
Expect(0, 3675, '\P{Is_Script=thai}', "");
Expect(1, 3675, '\P{^Is_Script=thai}', "");
Expect(0, 3676, '\p{Is_Script=thai}', "");
Expect(1, 3676, '\p{^Is_Script=thai}', "");
Expect(1, 3676, '\P{Is_Script=thai}', "");
Expect(0, 3676, '\P{^Is_Script=thai}', "");
Expect(1, 3675, '\p{Is_Script=_Thai}', "");
Expect(0, 3675, '\p{^Is_Script=_Thai}', "");
Expect(0, 3675, '\P{Is_Script=_Thai}', "");
Expect(1, 3675, '\P{^Is_Script=_Thai}', "");
Expect(0, 3676, '\p{Is_Script=_Thai}', "");
Expect(1, 3676, '\p{^Is_Script=_Thai}', "");
Expect(1, 3676, '\P{Is_Script=_Thai}', "");
Expect(0, 3676, '\P{^Is_Script=_Thai}', "");
Error('\p{Is_Sc=_/a/THAI}');
Error('\P{Is_Sc=_/a/THAI}');
Expect(1, 3675, '\p{Is_Sc=thai}', "");
Expect(0, 3675, '\p{^Is_Sc=thai}', "");
Expect(0, 3675, '\P{Is_Sc=thai}', "");
Expect(1, 3675, '\P{^Is_Sc=thai}', "");
Expect(0, 3676, '\p{Is_Sc=thai}', "");
Expect(1, 3676, '\p{^Is_Sc=thai}', "");
Expect(1, 3676, '\P{Is_Sc=thai}', "");
Expect(0, 3676, '\P{^Is_Sc=thai}', "");
Expect(1, 3675, '\p{Is_Sc=_-Thai}', "");
Expect(0, 3675, '\p{^Is_Sc=_-Thai}', "");
Expect(0, 3675, '\P{Is_Sc=_-Thai}', "");
Expect(1, 3675, '\P{^Is_Sc=_-Thai}', "");
Expect(0, 3676, '\p{Is_Sc=_-Thai}', "");
Expect(1, 3676, '\p{^Is_Sc=_-Thai}', "");
Expect(1, 3676, '\P{Is_Sc=_-Thai}', "");
Expect(0, 3676, '\P{^Is_Sc=_-Thai}', "");
Error('\p{Composition_Exclusion=	:=Yes}');
Error('\P{Composition_Exclusion=	:=Yes}');
Expect(1, 119232, '\p{Composition_Exclusion:yes}', "");
Expect(0, 119232, '\p{^Composition_Exclusion:yes}', "");
Expect(0, 119232, '\P{Composition_Exclusion:yes}', "");
Expect(1, 119232, '\P{^Composition_Exclusion:yes}', "");
Expect(0, 119233, '\p{Composition_Exclusion:yes}', "");
Expect(1, 119233, '\p{^Composition_Exclusion:yes}', "");
Expect(1, 119233, '\P{Composition_Exclusion:yes}', "");
Expect(0, 119233, '\P{^Composition_Exclusion:yes}', "");
Expect(1, 119232, '\p{Composition_Exclusion= 	yes}', "");
Expect(0, 119232, '\p{^Composition_Exclusion= 	yes}', "");
Expect(0, 119232, '\P{Composition_Exclusion= 	yes}', "");
Expect(1, 119232, '\P{^Composition_Exclusion= 	yes}', "");
Expect(0, 119233, '\p{Composition_Exclusion= 	yes}', "");
Expect(1, 119233, '\p{^Composition_Exclusion= 	yes}', "");
Expect(1, 119233, '\P{Composition_Exclusion= 	yes}', "");
Expect(0, 119233, '\P{^Composition_Exclusion= 	yes}', "");
Error('\p{CE= :=y}');
Error('\P{CE= :=y}');
Expect(1, 119232, '\p{CE=y}', "");
Expect(0, 119232, '\p{^CE=y}', "");
Expect(0, 119232, '\P{CE=y}', "");
Expect(1, 119232, '\P{^CE=y}', "");
Expect(0, 119233, '\p{CE=y}', "");
Expect(1, 119233, '\p{^CE=y}', "");
Expect(1, 119233, '\P{CE=y}', "");
Expect(0, 119233, '\P{^CE=y}', "");
Expect(1, 119232, '\p{CE=-Y}', "");
Expect(0, 119232, '\p{^CE=-Y}', "");
Expect(0, 119232, '\P{CE=-Y}', "");
Expect(1, 119232, '\P{^CE=-Y}', "");
Expect(0, 119233, '\p{CE=-Y}', "");
Expect(1, 119233, '\p{^CE=-Y}', "");
Expect(1, 119233, '\P{CE=-Y}', "");
Expect(0, 119233, '\P{^CE=-Y}', "");
Error('\p{Is_Composition_Exclusion: /a/ _T}');
Error('\P{Is_Composition_Exclusion: /a/ _T}');
Expect(1, 119232, '\p{Is_Composition_Exclusion=t}', "");
Expect(0, 119232, '\p{^Is_Composition_Exclusion=t}', "");
Expect(0, 119232, '\P{Is_Composition_Exclusion=t}', "");
Expect(1, 119232, '\P{^Is_Composition_Exclusion=t}', "");
Expect(0, 119233, '\p{Is_Composition_Exclusion=t}', "");
Expect(1, 119233, '\p{^Is_Composition_Exclusion=t}', "");
Expect(1, 119233, '\P{Is_Composition_Exclusion=t}', "");
Expect(0, 119233, '\P{^Is_Composition_Exclusion=t}', "");
Expect(1, 119232, '\p{Is_Composition_Exclusion=-T}', "");
Expect(0, 119232, '\p{^Is_Composition_Exclusion=-T}', "");
Expect(0, 119232, '\P{Is_Composition_Exclusion=-T}', "");
Expect(1, 119232, '\P{^Is_Composition_Exclusion=-T}', "");
Expect(0, 119233, '\p{Is_Composition_Exclusion=-T}', "");
Expect(1, 119233, '\p{^Is_Composition_Exclusion=-T}', "");
Expect(1, 119233, '\P{Is_Composition_Exclusion=-T}', "");
Expect(0, 119233, '\P{^Is_Composition_Exclusion=-T}', "");
Error('\p{Is_CE=:=true}');
Error('\P{Is_CE=:=true}');
Expect(1, 119232, '\p{Is_CE:   true}', "");
Expect(0, 119232, '\p{^Is_CE:   true}', "");
Expect(0, 119232, '\P{Is_CE:   true}', "");
Expect(1, 119232, '\P{^Is_CE:   true}', "");
Expect(0, 119233, '\p{Is_CE:   true}', "");
Expect(1, 119233, '\p{^Is_CE:   true}', "");
Expect(1, 119233, '\P{Is_CE:   true}', "");
Expect(0, 119233, '\P{^Is_CE:   true}', "");
Expect(1, 119232, '\p{Is_CE=__TRUE}', "");
Expect(0, 119232, '\p{^Is_CE=__TRUE}', "");
Expect(0, 119232, '\P{Is_CE=__TRUE}', "");
Expect(1, 119232, '\P{^Is_CE=__TRUE}', "");
Expect(0, 119233, '\p{Is_CE=__TRUE}', "");
Expect(1, 119233, '\p{^Is_CE=__TRUE}', "");
Expect(1, 119233, '\P{Is_CE=__TRUE}', "");
Expect(0, 119233, '\P{^Is_CE=__TRUE}', "");
Error('\p{Composition_Exclusion=- No:=}');
Error('\P{Composition_Exclusion=- No:=}');
Expect(1, 119233, '\p{Composition_Exclusion=no}', "");
Expect(0, 119233, '\p{^Composition_Exclusion=no}', "");
Expect(0, 119233, '\P{Composition_Exclusion=no}', "");
Expect(1, 119233, '\P{^Composition_Exclusion=no}', "");
Expect(0, 119232, '\p{Composition_Exclusion=no}', "");
Expect(1, 119232, '\p{^Composition_Exclusion=no}', "");
Expect(1, 119232, '\P{Composition_Exclusion=no}', "");
Expect(0, 119232, '\P{^Composition_Exclusion=no}', "");
Expect(1, 119233, '\p{Composition_Exclusion= no}', "");
Expect(0, 119233, '\p{^Composition_Exclusion= no}', "");
Expect(0, 119233, '\P{Composition_Exclusion= no}', "");
Expect(1, 119233, '\P{^Composition_Exclusion= no}', "");
Expect(0, 119232, '\p{Composition_Exclusion= no}', "");
Expect(1, 119232, '\p{^Composition_Exclusion= no}', "");
Expect(1, 119232, '\P{Composition_Exclusion= no}', "");
Expect(0, 119232, '\P{^Composition_Exclusion= no}', "");
Error('\p{CE=	/a/N}');
Error('\P{CE=	/a/N}');
Expect(1, 119233, '\p{CE=n}', "");
Expect(0, 119233, '\p{^CE=n}', "");
Expect(0, 119233, '\P{CE=n}', "");
Expect(1, 119233, '\P{^CE=n}', "");
Expect(0, 119232, '\p{CE=n}', "");
Expect(1, 119232, '\p{^CE=n}', "");
Expect(1, 119232, '\P{CE=n}', "");
Expect(0, 119232, '\P{^CE=n}', "");
Expect(1, 119233, '\p{CE= -N}', "");
Expect(0, 119233, '\p{^CE= -N}', "");
Expect(0, 119233, '\P{CE= -N}', "");
Expect(1, 119233, '\P{^CE= -N}', "");
Expect(0, 119232, '\p{CE= -N}', "");
Expect(1, 119232, '\p{^CE= -N}', "");
Expect(1, 119232, '\P{CE= -N}', "");
Expect(0, 119232, '\P{^CE= -N}', "");
Error('\p{Is_Composition_Exclusion=:=--F}');
Error('\P{Is_Composition_Exclusion=:=--F}');
Expect(1, 119233, '\p{Is_Composition_Exclusion=f}', "");
Expect(0, 119233, '\p{^Is_Composition_Exclusion=f}', "");
Expect(0, 119233, '\P{Is_Composition_Exclusion=f}', "");
Expect(1, 119233, '\P{^Is_Composition_Exclusion=f}', "");
Expect(0, 119232, '\p{Is_Composition_Exclusion=f}', "");
Expect(1, 119232, '\p{^Is_Composition_Exclusion=f}', "");
Expect(1, 119232, '\P{Is_Composition_Exclusion=f}', "");
Expect(0, 119232, '\P{^Is_Composition_Exclusion=f}', "");
Expect(1, 119233, '\p{Is_Composition_Exclusion=_-F}', "");
Expect(0, 119233, '\p{^Is_Composition_Exclusion=_-F}', "");
Expect(0, 119233, '\P{Is_Composition_Exclusion=_-F}', "");
Expect(1, 119233, '\P{^Is_Composition_Exclusion=_-F}', "");
Expect(0, 119232, '\p{Is_Composition_Exclusion=_-F}', "");
Expect(1, 119232, '\p{^Is_Composition_Exclusion=_-F}', "");
Expect(1, 119232, '\P{Is_Composition_Exclusion=_-F}', "");
Expect(0, 119232, '\P{^Is_Composition_Exclusion=_-F}', "");
Error('\p{Is_CE=_:=false}');
Error('\P{Is_CE=_:=false}');
Expect(1, 119233, '\p{Is_CE=false}', "");
Expect(0, 119233, '\p{^Is_CE=false}', "");
Expect(0, 119233, '\P{Is_CE=false}', "");
Expect(1, 119233, '\P{^Is_CE=false}', "");
Expect(0, 119232, '\p{Is_CE=false}', "");
Expect(1, 119232, '\p{^Is_CE=false}', "");
Expect(1, 119232, '\P{Is_CE=false}', "");
Expect(0, 119232, '\P{^Is_CE=false}', "");
Expect(1, 119233, '\p{Is_CE=	 False}', "");
Expect(0, 119233, '\p{^Is_CE=	 False}', "");
Expect(0, 119233, '\P{Is_CE=	 False}', "");
Expect(1, 119233, '\P{^Is_CE=	 False}', "");
Expect(0, 119232, '\p{Is_CE=	 False}', "");
Expect(1, 119232, '\p{^Is_CE=	 False}', "");
Expect(1, 119232, '\P{Is_CE=	 False}', "");
Expect(0, 119232, '\P{^Is_CE=	 False}', "");
Error('\p{XID_Continue=	yes:=}');
Error('\P{XID_Continue=	yes:=}');
Expect(1, 917999, '\p{XID_Continue=yes}', "");
Expect(0, 917999, '\p{^XID_Continue=yes}', "");
Expect(0, 917999, '\P{XID_Continue=yes}', "");
Expect(1, 917999, '\P{^XID_Continue=yes}', "");
Expect(0, 918000, '\p{XID_Continue=yes}', "");
Expect(1, 918000, '\p{^XID_Continue=yes}', "");
Expect(1, 918000, '\P{XID_Continue=yes}', "");
Expect(0, 918000, '\P{^XID_Continue=yes}', "");
Expect(1, 917999, '\p{XID_Continue= 	yes}', "");
Expect(0, 917999, '\p{^XID_Continue= 	yes}', "");
Expect(0, 917999, '\P{XID_Continue= 	yes}', "");
Expect(1, 917999, '\P{^XID_Continue= 	yes}', "");
Expect(0, 918000, '\p{XID_Continue= 	yes}', "");
Expect(1, 918000, '\p{^XID_Continue= 	yes}', "");
Expect(1, 918000, '\P{XID_Continue= 	yes}', "");
Expect(0, 918000, '\P{^XID_Continue= 	yes}', "");
Error('\p{XIDC=_:=Y}');
Error('\P{XIDC=_:=Y}');
Expect(1, 917999, '\p{XIDC=y}', "");
Expect(0, 917999, '\p{^XIDC=y}', "");
Expect(0, 917999, '\P{XIDC=y}', "");
Expect(1, 917999, '\P{^XIDC=y}', "");
Expect(0, 918000, '\p{XIDC=y}', "");
Expect(1, 918000, '\p{^XIDC=y}', "");
Expect(1, 918000, '\P{XIDC=y}', "");
Expect(0, 918000, '\P{^XIDC=y}', "");
Expect(1, 917999, '\p{XIDC=	-Y}', "");
Expect(0, 917999, '\p{^XIDC=	-Y}', "");
Expect(0, 917999, '\P{XIDC=	-Y}', "");
Expect(1, 917999, '\P{^XIDC=	-Y}', "");
Expect(0, 918000, '\p{XIDC=	-Y}', "");
Expect(1, 918000, '\p{^XIDC=	-Y}', "");
Expect(1, 918000, '\P{XIDC=	-Y}', "");
Expect(0, 918000, '\P{^XIDC=	-Y}', "");
Error('\p{Is_XID_Continue=/a/ 	T}');
Error('\P{Is_XID_Continue=/a/ 	T}');
Expect(1, 917999, '\p{Is_XID_Continue=t}', "");
Expect(0, 917999, '\p{^Is_XID_Continue=t}', "");
Expect(0, 917999, '\P{Is_XID_Continue=t}', "");
Expect(1, 917999, '\P{^Is_XID_Continue=t}', "");
Expect(0, 918000, '\p{Is_XID_Continue=t}', "");
Expect(1, 918000, '\p{^Is_XID_Continue=t}', "");
Expect(1, 918000, '\P{Is_XID_Continue=t}', "");
Expect(0, 918000, '\P{^Is_XID_Continue=t}', "");
Expect(1, 917999, '\p{Is_XID_Continue=  T}', "");
Expect(0, 917999, '\p{^Is_XID_Continue=  T}', "");
Expect(0, 917999, '\P{Is_XID_Continue=  T}', "");
Expect(1, 917999, '\P{^Is_XID_Continue=  T}', "");
Expect(0, 918000, '\p{Is_XID_Continue=  T}', "");
Expect(1, 918000, '\p{^Is_XID_Continue=  T}', "");
Expect(1, 918000, '\P{Is_XID_Continue=  T}', "");
Expect(0, 918000, '\P{^Is_XID_Continue=  T}', "");
Error('\p{Is_XIDC=	_TRUE:=}');
Error('\P{Is_XIDC=	_TRUE:=}');
Expect(1, 917999, '\p{Is_XIDC=true}', "");
Expect(0, 917999, '\p{^Is_XIDC=true}', "");
Expect(0, 917999, '\P{Is_XIDC=true}', "");
Expect(1, 917999, '\P{^Is_XIDC=true}', "");
Expect(0, 918000, '\p{Is_XIDC=true}', "");
Expect(1, 918000, '\p{^Is_XIDC=true}', "");
Expect(1, 918000, '\P{Is_XIDC=true}', "");
Expect(0, 918000, '\P{^Is_XIDC=true}', "");
Expect(1, 917999, '\p{Is_XIDC=-	TRUE}', "");
Expect(0, 917999, '\p{^Is_XIDC=-	TRUE}', "");
Expect(0, 917999, '\P{Is_XIDC=-	TRUE}', "");
Expect(1, 917999, '\P{^Is_XIDC=-	TRUE}', "");
Expect(0, 918000, '\p{Is_XIDC=-	TRUE}', "");
Expect(1, 918000, '\p{^Is_XIDC=-	TRUE}', "");
Expect(1, 918000, '\P{Is_XIDC=-	TRUE}', "");
Expect(0, 918000, '\P{^Is_XIDC=-	TRUE}', "");
Error('\p{XID_Continue=:= No}');
Error('\P{XID_Continue=:= No}');
Expect(1, 918000, '\p{XID_Continue=no}', "");
Expect(0, 918000, '\p{^XID_Continue=no}', "");
Expect(0, 918000, '\P{XID_Continue=no}', "");
Expect(1, 918000, '\P{^XID_Continue=no}', "");
Expect(0, 917999, '\p{XID_Continue=no}', "");
Expect(1, 917999, '\p{^XID_Continue=no}', "");
Expect(1, 917999, '\P{XID_Continue=no}', "");
Expect(0, 917999, '\P{^XID_Continue=no}', "");
Expect(1, 918000, '\p{XID_Continue=-no}', "");
Expect(0, 918000, '\p{^XID_Continue=-no}', "");
Expect(0, 918000, '\P{XID_Continue=-no}', "");
Expect(1, 918000, '\P{^XID_Continue=-no}', "");
Expect(0, 917999, '\p{XID_Continue=-no}', "");
Expect(1, 917999, '\p{^XID_Continue=-no}', "");
Expect(1, 917999, '\P{XID_Continue=-no}', "");
Expect(0, 917999, '\P{^XID_Continue=-no}', "");
Error('\p{XIDC= /a/N}');
Error('\P{XIDC= /a/N}');
Expect(1, 918000, '\p{XIDC=n}', "");
Expect(0, 918000, '\p{^XIDC=n}', "");
Expect(0, 918000, '\P{XIDC=n}', "");
Expect(1, 918000, '\P{^XIDC=n}', "");
Expect(0, 917999, '\p{XIDC=n}', "");
Expect(1, 917999, '\p{^XIDC=n}', "");
Expect(1, 917999, '\P{XIDC=n}', "");
Expect(0, 917999, '\P{^XIDC=n}', "");
Expect(1, 918000, '\p{XIDC=-_n}', "");
Expect(0, 918000, '\p{^XIDC=-_n}', "");
Expect(0, 918000, '\P{XIDC=-_n}', "");
Expect(1, 918000, '\P{^XIDC=-_n}', "");
Expect(0, 917999, '\p{XIDC=-_n}', "");
Expect(1, 917999, '\p{^XIDC=-_n}', "");
Expect(1, 917999, '\P{XIDC=-_n}', "");
Expect(0, 917999, '\P{^XIDC=-_n}', "");
Error('\p{Is_XID_Continue=_-F/a/}');
Error('\P{Is_XID_Continue=_-F/a/}');
Expect(1, 918000, '\p{Is_XID_Continue=f}', "");
Expect(0, 918000, '\p{^Is_XID_Continue=f}', "");
Expect(0, 918000, '\P{Is_XID_Continue=f}', "");
Expect(1, 918000, '\P{^Is_XID_Continue=f}', "");
Expect(0, 917999, '\p{Is_XID_Continue=f}', "");
Expect(1, 917999, '\p{^Is_XID_Continue=f}', "");
Expect(1, 917999, '\P{Is_XID_Continue=f}', "");
Expect(0, 917999, '\P{^Is_XID_Continue=f}', "");
Expect(1, 918000, '\p{Is_XID_Continue=-F}', "");
Expect(0, 918000, '\p{^Is_XID_Continue=-F}', "");
Expect(0, 918000, '\P{Is_XID_Continue=-F}', "");
Expect(1, 918000, '\P{^Is_XID_Continue=-F}', "");
Expect(0, 917999, '\p{Is_XID_Continue=-F}', "");
Expect(1, 917999, '\p{^Is_XID_Continue=-F}', "");
Expect(1, 917999, '\P{Is_XID_Continue=-F}', "");
Expect(0, 917999, '\P{^Is_XID_Continue=-F}', "");
Error('\p{Is_XIDC=- False/a/}');
Error('\P{Is_XIDC=- False/a/}');
Expect(1, 918000, '\p{Is_XIDC=false}', "");
Expect(0, 918000, '\p{^Is_XIDC=false}', "");
Expect(0, 918000, '\P{Is_XIDC=false}', "");
Expect(1, 918000, '\P{^Is_XIDC=false}', "");
Expect(0, 917999, '\p{Is_XIDC=false}', "");
Expect(1, 917999, '\p{^Is_XIDC=false}', "");
Expect(1, 917999, '\P{Is_XIDC=false}', "");
Expect(0, 917999, '\P{^Is_XIDC=false}', "");
Expect(1, 918000, '\p{Is_XIDC=	False}', "");
Expect(0, 918000, '\p{^Is_XIDC=	False}', "");
Expect(0, 918000, '\P{Is_XIDC=	False}', "");
Expect(1, 918000, '\P{^Is_XIDC=	False}', "");
Expect(0, 917999, '\p{Is_XIDC=	False}', "");
Expect(1, 917999, '\p{^Is_XIDC=	False}', "");
Expect(1, 917999, '\P{Is_XIDC=	False}', "");
Expect(0, 917999, '\P{^Is_XIDC=	False}', "");
Error('\p{White_Space:  :=YES}');
Error('\P{White_Space:  :=YES}');
Expect(1, 12288, '\p{White_Space=yes}', "");
Expect(0, 12288, '\p{^White_Space=yes}', "");
Expect(0, 12288, '\P{White_Space=yes}', "");
Expect(1, 12288, '\P{^White_Space=yes}', "");
Expect(0, 12289, '\p{White_Space=yes}', "");
Expect(1, 12289, '\p{^White_Space=yes}', "");
Expect(1, 12289, '\P{White_Space=yes}', "");
Expect(0, 12289, '\P{^White_Space=yes}', "");
Expect(1, 12288, '\p{White_Space=__YES}', "");
Expect(0, 12288, '\p{^White_Space=__YES}', "");
Expect(0, 12288, '\P{White_Space=__YES}', "");
Expect(1, 12288, '\P{^White_Space=__YES}', "");
Expect(0, 12289, '\p{White_Space=__YES}', "");
Expect(1, 12289, '\p{^White_Space=__YES}', "");
Expect(1, 12289, '\P{White_Space=__YES}', "");
Expect(0, 12289, '\P{^White_Space=__YES}', "");
Error('\p{WSpace= :=Y}');
Error('\P{WSpace= :=Y}');
Expect(1, 12288, '\p{WSpace=y}', "");
Expect(0, 12288, '\p{^WSpace=y}', "");
Expect(0, 12288, '\P{WSpace=y}', "");
Expect(1, 12288, '\P{^WSpace=y}', "");
Expect(0, 12289, '\p{WSpace=y}', "");
Expect(1, 12289, '\p{^WSpace=y}', "");
Expect(1, 12289, '\P{WSpace=y}', "");
Expect(0, 12289, '\P{^WSpace=y}', "");
Expect(1, 12288, '\p{WSpace= -Y}', "");
Expect(0, 12288, '\p{^WSpace= -Y}', "");
Expect(0, 12288, '\P{WSpace= -Y}', "");
Expect(1, 12288, '\P{^WSpace= -Y}', "");
Expect(0, 12289, '\p{WSpace= -Y}', "");
Expect(1, 12289, '\p{^WSpace= -Y}', "");
Expect(1, 12289, '\P{WSpace= -Y}', "");
Expect(0, 12289, '\P{^WSpace= -Y}', "");
Error('\p{Space=	T:=}');
Error('\P{Space=	T:=}');
Expect(1, 12288, '\p{Space=t}', "");
Expect(0, 12288, '\p{^Space=t}', "");
Expect(0, 12288, '\P{Space=t}', "");
Expect(1, 12288, '\P{^Space=t}', "");
Expect(0, 12289, '\p{Space=t}', "");
Expect(1, 12289, '\p{^Space=t}', "");
Expect(1, 12289, '\P{Space=t}', "");
Expect(0, 12289, '\P{^Space=t}', "");
Expect(1, 12288, '\p{Space=-_T}', "");
Expect(0, 12288, '\p{^Space=-_T}', "");
Expect(0, 12288, '\P{Space=-_T}', "");
Expect(1, 12288, '\P{^Space=-_T}', "");
Expect(0, 12289, '\p{Space=-_T}', "");
Expect(1, 12289, '\p{^Space=-_T}', "");
Expect(1, 12289, '\P{Space=-_T}', "");
Expect(0, 12289, '\P{^Space=-_T}', "");
Error('\p{Is_White_Space:	_/a/true}');
Error('\P{Is_White_Space:	_/a/true}');
Expect(1, 12288, '\p{Is_White_Space=true}', "");
Expect(0, 12288, '\p{^Is_White_Space=true}', "");
Expect(0, 12288, '\P{Is_White_Space=true}', "");
Expect(1, 12288, '\P{^Is_White_Space=true}', "");
Expect(0, 12289, '\p{Is_White_Space=true}', "");
Expect(1, 12289, '\p{^Is_White_Space=true}', "");
Expect(1, 12289, '\P{Is_White_Space=true}', "");
Expect(0, 12289, '\P{^Is_White_Space=true}', "");
Expect(1, 12288, '\p{Is_White_Space= 	true}', "");
Expect(0, 12288, '\p{^Is_White_Space= 	true}', "");
Expect(0, 12288, '\P{Is_White_Space= 	true}', "");
Expect(1, 12288, '\P{^Is_White_Space= 	true}', "");
Expect(0, 12289, '\p{Is_White_Space= 	true}', "");
Expect(1, 12289, '\p{^Is_White_Space= 	true}', "");
Expect(1, 12289, '\P{Is_White_Space= 	true}', "");
Expect(0, 12289, '\P{^Is_White_Space= 	true}', "");
Error('\p{Is_WSpace=/a/	 Yes}');
Error('\P{Is_WSpace=/a/	 Yes}');
Expect(1, 12288, '\p{Is_WSpace=yes}', "");
Expect(0, 12288, '\p{^Is_WSpace=yes}', "");
Expect(0, 12288, '\P{Is_WSpace=yes}', "");
Expect(1, 12288, '\P{^Is_WSpace=yes}', "");
Expect(0, 12289, '\p{Is_WSpace=yes}', "");
Expect(1, 12289, '\p{^Is_WSpace=yes}', "");
Expect(1, 12289, '\P{Is_WSpace=yes}', "");
Expect(0, 12289, '\P{^Is_WSpace=yes}', "");
Expect(1, 12288, '\p{Is_WSpace:    -YES}', "");
Expect(0, 12288, '\p{^Is_WSpace:    -YES}', "");
Expect(0, 12288, '\P{Is_WSpace:    -YES}', "");
Expect(1, 12288, '\P{^Is_WSpace:    -YES}', "");
Expect(0, 12289, '\p{Is_WSpace:    -YES}', "");
Expect(1, 12289, '\p{^Is_WSpace:    -YES}', "");
Expect(1, 12289, '\P{Is_WSpace:    -YES}', "");
Expect(0, 12289, '\P{^Is_WSpace:    -YES}', "");
Error('\p{Is_Space: :=- y}');
Error('\P{Is_Space: :=- y}');
Expect(1, 12288, '\p{Is_Space=y}', "");
Expect(0, 12288, '\p{^Is_Space=y}', "");
Expect(0, 12288, '\P{Is_Space=y}', "");
Expect(1, 12288, '\P{^Is_Space=y}', "");
Expect(0, 12289, '\p{Is_Space=y}', "");
Expect(1, 12289, '\p{^Is_Space=y}', "");
Expect(1, 12289, '\P{Is_Space=y}', "");
Expect(0, 12289, '\P{^Is_Space=y}', "");
Expect(1, 12288, '\p{Is_Space=	 Y}', "");
Expect(0, 12288, '\p{^Is_Space=	 Y}', "");
Expect(0, 12288, '\P{Is_Space=	 Y}', "");
Expect(1, 12288, '\P{^Is_Space=	 Y}', "");
Expect(0, 12289, '\p{Is_Space=	 Y}', "");
Expect(1, 12289, '\p{^Is_Space=	 Y}', "");
Expect(1, 12289, '\P{Is_Space=	 Y}', "");
Expect(0, 12289, '\P{^Is_Space=	 Y}', "");
Error('\p{White_Space=:=NO}');
Error('\P{White_Space=:=NO}');
Expect(1, 12289, '\p{White_Space=no}', "");
Expect(0, 12289, '\p{^White_Space=no}', "");
Expect(0, 12289, '\P{White_Space=no}', "");
Expect(1, 12289, '\P{^White_Space=no}', "");
Expect(0, 12288, '\p{White_Space=no}', "");
Expect(1, 12288, '\p{^White_Space=no}', "");
Expect(1, 12288, '\P{White_Space=no}', "");
Expect(0, 12288, '\P{^White_Space=no}', "");
Expect(1, 12289, '\p{White_Space=_	No}', "");
Expect(0, 12289, '\p{^White_Space=_	No}', "");
Expect(0, 12289, '\P{White_Space=_	No}', "");
Expect(1, 12289, '\P{^White_Space=_	No}', "");
Expect(0, 12288, '\p{White_Space=_	No}', "");
Expect(1, 12288, '\p{^White_Space=_	No}', "");
Expect(1, 12288, '\P{White_Space=_	No}', "");
Expect(0, 12288, '\P{^White_Space=_	No}', "");
Error('\p{WSpace= N/a/}');
Error('\P{WSpace= N/a/}');
Expect(1, 12289, '\p{WSpace=n}', "");
Expect(0, 12289, '\p{^WSpace=n}', "");
Expect(0, 12289, '\P{WSpace=n}', "");
Expect(1, 12289, '\P{^WSpace=n}', "");
Expect(0, 12288, '\p{WSpace=n}', "");
Expect(1, 12288, '\p{^WSpace=n}', "");
Expect(1, 12288, '\P{WSpace=n}', "");
Expect(0, 12288, '\P{^WSpace=n}', "");
Expect(1, 12289, '\p{WSpace=_N}', "");
Expect(0, 12289, '\p{^WSpace=_N}', "");
Expect(0, 12289, '\P{WSpace=_N}', "");
Expect(1, 12289, '\P{^WSpace=_N}', "");
Expect(0, 12288, '\p{WSpace=_N}', "");
Expect(1, 12288, '\p{^WSpace=_N}', "");
Expect(1, 12288, '\P{WSpace=_N}', "");
Expect(0, 12288, '\P{^WSpace=_N}', "");
Error('\p{Space: _/a/F}');
Error('\P{Space: _/a/F}');
Expect(1, 12289, '\p{Space:   f}', "");
Expect(0, 12289, '\p{^Space:   f}', "");
Expect(0, 12289, '\P{Space:   f}', "");
Expect(1, 12289, '\P{^Space:   f}', "");
Expect(0, 12288, '\p{Space:   f}', "");
Expect(1, 12288, '\p{^Space:   f}', "");
Expect(1, 12288, '\P{Space:   f}', "");
Expect(0, 12288, '\P{^Space:   f}', "");
Expect(1, 12289, '\p{Space=-_F}', "");
Expect(0, 12289, '\p{^Space=-_F}', "");
Expect(0, 12289, '\P{Space=-_F}', "");
Expect(1, 12289, '\P{^Space=-_F}', "");
Expect(0, 12288, '\p{Space=-_F}', "");
Expect(1, 12288, '\p{^Space=-_F}', "");
Expect(1, 12288, '\P{Space=-_F}', "");
Expect(0, 12288, '\P{^Space=-_F}', "");
Error('\p{Is_White_Space=	FALSE:=}');
Error('\P{Is_White_Space=	FALSE:=}');
Expect(1, 12289, '\p{Is_White_Space=false}', "");
Expect(0, 12289, '\p{^Is_White_Space=false}', "");
Expect(0, 12289, '\P{Is_White_Space=false}', "");
Expect(1, 12289, '\P{^Is_White_Space=false}', "");
Expect(0, 12288, '\p{Is_White_Space=false}', "");
Expect(1, 12288, '\p{^Is_White_Space=false}', "");
Expect(1, 12288, '\P{Is_White_Space=false}', "");
Expect(0, 12288, '\P{^Is_White_Space=false}', "");
Expect(1, 12289, '\p{Is_White_Space=_	FALSE}', "");
Expect(0, 12289, '\p{^Is_White_Space=_	FALSE}', "");
Expect(0, 12289, '\P{Is_White_Space=_	FALSE}', "");
Expect(1, 12289, '\P{^Is_White_Space=_	FALSE}', "");
Expect(0, 12288, '\p{Is_White_Space=_	FALSE}', "");
Expect(1, 12288, '\p{^Is_White_Space=_	FALSE}', "");
Expect(1, 12288, '\P{Is_White_Space=_	FALSE}', "");
Expect(0, 12288, '\P{^Is_White_Space=_	FALSE}', "");
Error('\p{Is_WSpace=/a/-No}');
Error('\P{Is_WSpace=/a/-No}');
Expect(1, 12289, '\p{Is_WSpace=no}', "");
Expect(0, 12289, '\p{^Is_WSpace=no}', "");
Expect(0, 12289, '\P{Is_WSpace=no}', "");
Expect(1, 12289, '\P{^Is_WSpace=no}', "");
Expect(0, 12288, '\p{Is_WSpace=no}', "");
Expect(1, 12288, '\p{^Is_WSpace=no}', "");
Expect(1, 12288, '\P{Is_WSpace=no}', "");
Expect(0, 12288, '\P{^Is_WSpace=no}', "");
Expect(1, 12289, '\p{Is_WSpace=_-no}', "");
Expect(0, 12289, '\p{^Is_WSpace=_-no}', "");
Expect(0, 12289, '\P{Is_WSpace=_-no}', "");
Expect(1, 12289, '\P{^Is_WSpace=_-no}', "");
Expect(0, 12288, '\p{Is_WSpace=_-no}', "");
Expect(1, 12288, '\p{^Is_WSpace=_-no}', "");
Expect(1, 12288, '\P{Is_WSpace=_-no}', "");
Expect(0, 12288, '\P{^Is_WSpace=_-no}', "");
Error('\p{Is_Space= -n/a/}');
Error('\P{Is_Space= -n/a/}');
Expect(1, 12289, '\p{Is_Space=n}', "");
Expect(0, 12289, '\p{^Is_Space=n}', "");
Expect(0, 12289, '\P{Is_Space=n}', "");
Expect(1, 12289, '\P{^Is_Space=n}', "");
Expect(0, 12288, '\p{Is_Space=n}', "");
Expect(1, 12288, '\p{^Is_Space=n}', "");
Expect(1, 12288, '\P{Is_Space=n}', "");
Expect(0, 12288, '\P{^Is_Space=n}', "");
Expect(1, 12289, '\p{Is_Space=_N}', "");
Expect(0, 12289, '\p{^Is_Space=_N}', "");
Expect(0, 12289, '\P{Is_Space=_N}', "");
Expect(1, 12289, '\P{^Is_Space=_N}', "");
Expect(0, 12288, '\p{Is_Space=_N}', "");
Expect(1, 12288, '\p{^Is_Space=_N}', "");
Expect(1, 12288, '\P{Is_Space=_N}', "");
Expect(0, 12288, '\P{^Is_Space=_N}', "");
Error('\p{Math= :=yes}');
Error('\P{Math= :=yes}');
Expect(1, 126705, '\p{Math=yes}', "");
Expect(0, 126705, '\p{^Math=yes}', "");
Expect(0, 126705, '\P{Math=yes}', "");
Expect(1, 126705, '\P{^Math=yes}', "");
Expect(0, 126706, '\p{Math=yes}', "");
Expect(1, 126706, '\p{^Math=yes}', "");
Expect(1, 126706, '\P{Math=yes}', "");
Expect(0, 126706, '\P{^Math=yes}', "");
Expect(1, 126705, '\p{Math= _Yes}', "");
Expect(0, 126705, '\p{^Math= _Yes}', "");
Expect(0, 126705, '\P{Math= _Yes}', "");
Expect(1, 126705, '\P{^Math= _Yes}', "");
Expect(0, 126706, '\p{Math= _Yes}', "");
Expect(1, 126706, '\p{^Math= _Yes}', "");
Expect(1, 126706, '\P{Math= _Yes}', "");
Expect(0, 126706, '\P{^Math= _Yes}', "");
Error('\p{Is_Math=:=	_Y}');
Error('\P{Is_Math=:=	_Y}');
Expect(1, 126705, '\p{Is_Math=y}', "");
Expect(0, 126705, '\p{^Is_Math=y}', "");
Expect(0, 126705, '\P{Is_Math=y}', "");
Expect(1, 126705, '\P{^Is_Math=y}', "");
Expect(0, 126706, '\p{Is_Math=y}', "");
Expect(1, 126706, '\p{^Is_Math=y}', "");
Expect(1, 126706, '\P{Is_Math=y}', "");
Expect(0, 126706, '\P{^Is_Math=y}', "");
Expect(1, 126705, '\p{Is_Math=_	y}', "");
Expect(0, 126705, '\p{^Is_Math=_	y}', "");
Expect(0, 126705, '\P{Is_Math=_	y}', "");
Expect(1, 126705, '\P{^Is_Math=_	y}', "");
Expect(0, 126706, '\p{Is_Math=_	y}', "");
Expect(1, 126706, '\p{^Is_Math=_	y}', "");
Expect(1, 126706, '\P{Is_Math=_	y}', "");
Expect(0, 126706, '\P{^Is_Math=_	y}', "");
Error('\p{Math= t:=}');
Error('\P{Math= t:=}');
Expect(1, 126705, '\p{Math=t}', "");
Expect(0, 126705, '\p{^Math=t}', "");
Expect(0, 126705, '\P{Math=t}', "");
Expect(1, 126705, '\P{^Math=t}', "");
Expect(0, 126706, '\p{Math=t}', "");
Expect(1, 126706, '\p{^Math=t}', "");
Expect(1, 126706, '\P{Math=t}', "");
Expect(0, 126706, '\P{^Math=t}', "");
Expect(1, 126705, '\p{Math:  	T}', "");
Expect(0, 126705, '\p{^Math:  	T}', "");
Expect(0, 126705, '\P{Math:  	T}', "");
Expect(1, 126705, '\P{^Math:  	T}', "");
Expect(0, 126706, '\p{Math:  	T}', "");
Expect(1, 126706, '\p{^Math:  	T}', "");
Expect(1, 126706, '\P{Math:  	T}', "");
Expect(0, 126706, '\P{^Math:  	T}', "");
Error('\p{Is_Math=-_True/a/}');
Error('\P{Is_Math=-_True/a/}');
Expect(1, 126705, '\p{Is_Math=true}', "");
Expect(0, 126705, '\p{^Is_Math=true}', "");
Expect(0, 126705, '\P{Is_Math=true}', "");
Expect(1, 126705, '\P{^Is_Math=true}', "");
Expect(0, 126706, '\p{Is_Math=true}', "");
Expect(1, 126706, '\p{^Is_Math=true}', "");
Expect(1, 126706, '\P{Is_Math=true}', "");
Expect(0, 126706, '\P{^Is_Math=true}', "");
Expect(1, 126705, '\p{Is_Math= 	true}', "");
Expect(0, 126705, '\p{^Is_Math= 	true}', "");
Expect(0, 126705, '\P{Is_Math= 	true}', "");
Expect(1, 126705, '\P{^Is_Math= 	true}', "");
Expect(0, 126706, '\p{Is_Math= 	true}', "");
Expect(1, 126706, '\p{^Is_Math= 	true}', "");
Expect(1, 126706, '\P{Is_Math= 	true}', "");
Expect(0, 126706, '\P{^Is_Math= 	true}', "");
Error('\p{Math=_:=No}');
Error('\P{Math=_:=No}');
Expect(1, 126706, '\p{Math=no}', "");
Expect(0, 126706, '\p{^Math=no}', "");
Expect(0, 126706, '\P{Math=no}', "");
Expect(1, 126706, '\P{^Math=no}', "");
Expect(0, 126705, '\p{Math=no}', "");
Expect(1, 126705, '\p{^Math=no}', "");
Expect(1, 126705, '\P{Math=no}', "");
Expect(0, 126705, '\P{^Math=no}', "");
Expect(1, 126706, '\p{Math:   	no}', "");
Expect(0, 126706, '\p{^Math:   	no}', "");
Expect(0, 126706, '\P{Math:   	no}', "");
Expect(1, 126706, '\P{^Math:   	no}', "");
Expect(0, 126705, '\p{Math:   	no}', "");
Expect(1, 126705, '\p{^Math:   	no}', "");
Expect(1, 126705, '\P{Math:   	no}', "");
Expect(0, 126705, '\P{^Math:   	no}', "");
Error('\p{Is_Math=-/a/N}');
Error('\P{Is_Math=-/a/N}');
Expect(1, 126706, '\p{Is_Math=n}', "");
Expect(0, 126706, '\p{^Is_Math=n}', "");
Expect(0, 126706, '\P{Is_Math=n}', "");
Expect(1, 126706, '\P{^Is_Math=n}', "");
Expect(0, 126705, '\p{Is_Math=n}', "");
Expect(1, 126705, '\p{^Is_Math=n}', "");
Expect(1, 126705, '\P{Is_Math=n}', "");
Expect(0, 126705, '\P{^Is_Math=n}', "");
Expect(1, 126706, '\p{Is_Math=_N}', "");
Expect(0, 126706, '\p{^Is_Math=_N}', "");
Expect(0, 126706, '\P{Is_Math=_N}', "");
Expect(1, 126706, '\P{^Is_Math=_N}', "");
Expect(0, 126705, '\p{Is_Math=_N}', "");
Expect(1, 126705, '\p{^Is_Math=_N}', "");
Expect(1, 126705, '\P{Is_Math=_N}', "");
Expect(0, 126705, '\P{^Is_Math=_N}', "");
Error('\p{Math::=__F}');
Error('\P{Math::=__F}');
Expect(1, 126706, '\p{Math=f}', "");
Expect(0, 126706, '\p{^Math=f}', "");
Expect(0, 126706, '\P{Math=f}', "");
Expect(1, 126706, '\P{^Math=f}', "");
Expect(0, 126705, '\p{Math=f}', "");
Expect(1, 126705, '\p{^Math=f}', "");
Expect(1, 126705, '\P{Math=f}', "");
Expect(0, 126705, '\P{^Math=f}', "");
Expect(1, 126706, '\p{Math:     f}', "");
Expect(0, 126706, '\p{^Math:     f}', "");
Expect(0, 126706, '\P{Math:     f}', "");
Expect(1, 126706, '\P{^Math:     f}', "");
Expect(0, 126705, '\p{Math:     f}', "");
Expect(1, 126705, '\p{^Math:     f}', "");
Expect(1, 126705, '\P{Math:     f}', "");
Expect(0, 126705, '\P{^Math:     f}', "");
Error('\p{Is_Math=		False:=}');
Error('\P{Is_Math=		False:=}');
Expect(1, 126706, '\p{Is_Math=false}', "");
Expect(0, 126706, '\p{^Is_Math=false}', "");
Expect(0, 126706, '\P{Is_Math=false}', "");
Expect(1, 126706, '\P{^Is_Math=false}', "");
Expect(0, 126705, '\p{Is_Math=false}', "");
Expect(1, 126705, '\p{^Is_Math=false}', "");
Expect(1, 126705, '\P{Is_Math=false}', "");
Expect(0, 126705, '\P{^Is_Math=false}', "");
Expect(1, 126706, '\p{Is_Math: _False}', "");
Expect(0, 126706, '\p{^Is_Math: _False}', "");
Expect(0, 126706, '\P{Is_Math: _False}', "");
Expect(1, 126706, '\P{^Is_Math: _False}', "");
Expect(0, 126705, '\p{Is_Math: _False}', "");
Expect(1, 126705, '\p{^Is_Math: _False}', "");
Expect(1, 126705, '\P{Is_Math: _False}', "");
Expect(0, 126705, '\P{^Is_Math: _False}', "");
Error('\p{ID_Continue=YES:=}');
Error('\P{ID_Continue=YES:=}');
Expect(1, 917999, '\p{ID_Continue=yes}', "");
Expect(0, 917999, '\p{^ID_Continue=yes}', "");
Expect(0, 917999, '\P{ID_Continue=yes}', "");
Expect(1, 917999, '\P{^ID_Continue=yes}', "");
Expect(0, 918000, '\p{ID_Continue=yes}', "");
Expect(1, 918000, '\p{^ID_Continue=yes}', "");
Expect(1, 918000, '\P{ID_Continue=yes}', "");
Expect(0, 918000, '\P{^ID_Continue=yes}', "");
Expect(1, 917999, '\p{ID_Continue=	 Yes}', "");
Expect(0, 917999, '\p{^ID_Continue=	 Yes}', "");
Expect(0, 917999, '\P{ID_Continue=	 Yes}', "");
Expect(1, 917999, '\P{^ID_Continue=	 Yes}', "");
Expect(0, 918000, '\p{ID_Continue=	 Yes}', "");
Expect(1, 918000, '\p{^ID_Continue=	 Yes}', "");
Expect(1, 918000, '\P{ID_Continue=	 Yes}', "");
Expect(0, 918000, '\P{^ID_Continue=	 Yes}', "");
Error('\p{IDC=	-Y:=}');
Error('\P{IDC=	-Y:=}');
Expect(1, 917999, '\p{IDC=y}', "");
Expect(0, 917999, '\p{^IDC=y}', "");
Expect(0, 917999, '\P{IDC=y}', "");
Expect(1, 917999, '\P{^IDC=y}', "");
Expect(0, 918000, '\p{IDC=y}', "");
Expect(1, 918000, '\p{^IDC=y}', "");
Expect(1, 918000, '\P{IDC=y}', "");
Expect(0, 918000, '\P{^IDC=y}', "");
Expect(1, 917999, '\p{IDC=-Y}', "");
Expect(0, 917999, '\p{^IDC=-Y}', "");
Expect(0, 917999, '\P{IDC=-Y}', "");
Expect(1, 917999, '\P{^IDC=-Y}', "");
Expect(0, 918000, '\p{IDC=-Y}', "");
Expect(1, 918000, '\p{^IDC=-Y}', "");
Expect(1, 918000, '\P{IDC=-Y}', "");
Expect(0, 918000, '\P{^IDC=-Y}', "");
Error('\p{Is_ID_Continue=_	T/a/}');
Error('\P{Is_ID_Continue=_	T/a/}');
Expect(1, 917999, '\p{Is_ID_Continue=t}', "");
Expect(0, 917999, '\p{^Is_ID_Continue=t}', "");
Expect(0, 917999, '\P{Is_ID_Continue=t}', "");
Expect(1, 917999, '\P{^Is_ID_Continue=t}', "");
Expect(0, 918000, '\p{Is_ID_Continue=t}', "");
Expect(1, 918000, '\p{^Is_ID_Continue=t}', "");
Expect(1, 918000, '\P{Is_ID_Continue=t}', "");
Expect(0, 918000, '\P{^Is_ID_Continue=t}', "");
Expect(1, 917999, '\p{Is_ID_Continue=-T}', "");
Expect(0, 917999, '\p{^Is_ID_Continue=-T}', "");
Expect(0, 917999, '\P{Is_ID_Continue=-T}', "");
Expect(1, 917999, '\P{^Is_ID_Continue=-T}', "");
Expect(0, 918000, '\p{Is_ID_Continue=-T}', "");
Expect(1, 918000, '\p{^Is_ID_Continue=-T}', "");
Expect(1, 918000, '\P{Is_ID_Continue=-T}', "");
Expect(0, 918000, '\P{^Is_ID_Continue=-T}', "");
Error('\p{Is_IDC=/a/-	true}');
Error('\P{Is_IDC=/a/-	true}');
Expect(1, 917999, '\p{Is_IDC=true}', "");
Expect(0, 917999, '\p{^Is_IDC=true}', "");
Expect(0, 917999, '\P{Is_IDC=true}', "");
Expect(1, 917999, '\P{^Is_IDC=true}', "");
Expect(0, 918000, '\p{Is_IDC=true}', "");
Expect(1, 918000, '\p{^Is_IDC=true}', "");
Expect(1, 918000, '\P{Is_IDC=true}', "");
Expect(0, 918000, '\P{^Is_IDC=true}', "");
Expect(1, 917999, '\p{Is_IDC=	 True}', "");
Expect(0, 917999, '\p{^Is_IDC=	 True}', "");
Expect(0, 917999, '\P{Is_IDC=	 True}', "");
Expect(1, 917999, '\P{^Is_IDC=	 True}', "");
Expect(0, 918000, '\p{Is_IDC=	 True}', "");
Expect(1, 918000, '\p{^Is_IDC=	 True}', "");
Expect(1, 918000, '\P{Is_IDC=	 True}', "");
Expect(0, 918000, '\P{^Is_IDC=	 True}', "");
Error('\p{ID_Continue=/a/	 No}');
Error('\P{ID_Continue=/a/	 No}');
Expect(1, 918000, '\p{ID_Continue=no}', "");
Expect(0, 918000, '\p{^ID_Continue=no}', "");
Expect(0, 918000, '\P{ID_Continue=no}', "");
Expect(1, 918000, '\P{^ID_Continue=no}', "");
Expect(0, 917999, '\p{ID_Continue=no}', "");
Expect(1, 917999, '\p{^ID_Continue=no}', "");
Expect(1, 917999, '\P{ID_Continue=no}', "");
Expect(0, 917999, '\P{^ID_Continue=no}', "");
Expect(1, 918000, '\p{ID_Continue:   _	No}', "");
Expect(0, 918000, '\p{^ID_Continue:   _	No}', "");
Expect(0, 918000, '\P{ID_Continue:   _	No}', "");
Expect(1, 918000, '\P{^ID_Continue:   _	No}', "");
Expect(0, 917999, '\p{ID_Continue:   _	No}', "");
Expect(1, 917999, '\p{^ID_Continue:   _	No}', "");
Expect(1, 917999, '\P{ID_Continue:   _	No}', "");
Expect(0, 917999, '\P{^ID_Continue:   _	No}', "");
Error('\p{IDC:-	N:=}');
Error('\P{IDC:-	N:=}');
Expect(1, 918000, '\p{IDC=n}', "");
Expect(0, 918000, '\p{^IDC=n}', "");
Expect(0, 918000, '\P{IDC=n}', "");
Expect(1, 918000, '\P{^IDC=n}', "");
Expect(0, 917999, '\p{IDC=n}', "");
Expect(1, 917999, '\p{^IDC=n}', "");
Expect(1, 917999, '\P{IDC=n}', "");
Expect(0, 917999, '\P{^IDC=n}', "");
Expect(1, 918000, '\p{IDC=_N}', "");
Expect(0, 918000, '\p{^IDC=_N}', "");
Expect(0, 918000, '\P{IDC=_N}', "");
Expect(1, 918000, '\P{^IDC=_N}', "");
Expect(0, 917999, '\p{IDC=_N}', "");
Expect(1, 917999, '\p{^IDC=_N}', "");
Expect(1, 917999, '\P{IDC=_N}', "");
Expect(0, 917999, '\P{^IDC=_N}', "");
Error('\p{Is_ID_Continue=:=		f}');
Error('\P{Is_ID_Continue=:=		f}');
Expect(1, 918000, '\p{Is_ID_Continue:f}', "");
Expect(0, 918000, '\p{^Is_ID_Continue:f}', "");
Expect(0, 918000, '\P{Is_ID_Continue:f}', "");
Expect(1, 918000, '\P{^Is_ID_Continue:f}', "");
Expect(0, 917999, '\p{Is_ID_Continue:f}', "");
Expect(1, 917999, '\p{^Is_ID_Continue:f}', "");
Expect(1, 917999, '\P{Is_ID_Continue:f}', "");
Expect(0, 917999, '\P{^Is_ID_Continue:f}', "");
Expect(1, 918000, '\p{Is_ID_Continue=	 F}', "");
Expect(0, 918000, '\p{^Is_ID_Continue=	 F}', "");
Expect(0, 918000, '\P{Is_ID_Continue=	 F}', "");
Expect(1, 918000, '\P{^Is_ID_Continue=	 F}', "");
Expect(0, 917999, '\p{Is_ID_Continue=	 F}', "");
Expect(1, 917999, '\p{^Is_ID_Continue=	 F}', "");
Expect(1, 917999, '\P{Is_ID_Continue=	 F}', "");
Expect(0, 917999, '\P{^Is_ID_Continue=	 F}', "");
Error('\p{Is_IDC=_ FALSE/a/}');
Error('\P{Is_IDC=_ FALSE/a/}');
Expect(1, 918000, '\p{Is_IDC: false}', "");
Expect(0, 918000, '\p{^Is_IDC: false}', "");
Expect(0, 918000, '\P{Is_IDC: false}', "");
Expect(1, 918000, '\P{^Is_IDC: false}', "");
Expect(0, 917999, '\p{Is_IDC: false}', "");
Expect(1, 917999, '\p{^Is_IDC: false}', "");
Expect(1, 917999, '\P{Is_IDC: false}', "");
Expect(0, 917999, '\P{^Is_IDC: false}', "");
Expect(1, 918000, '\p{Is_IDC=_	False}', "");
Expect(0, 918000, '\p{^Is_IDC=_	False}', "");
Expect(0, 918000, '\P{Is_IDC=_	False}', "");
Expect(1, 918000, '\P{^Is_IDC=_	False}', "");
Expect(0, 917999, '\p{Is_IDC=_	False}', "");
Expect(1, 917999, '\p{^Is_IDC=_	False}', "");
Expect(1, 917999, '\P{Is_IDC=_	False}', "");
Expect(0, 917999, '\P{^Is_IDC=_	False}', "");
Error('\p{Variation_Selector:   /a/	-YES}');
Error('\P{Variation_Selector:   /a/	-YES}');
Expect(1, 917999, '\p{Variation_Selector=yes}', "");
Expect(0, 917999, '\p{^Variation_Selector=yes}', "");
Expect(0, 917999, '\P{Variation_Selector=yes}', "");
Expect(1, 917999, '\P{^Variation_Selector=yes}', "");
Expect(0, 918000, '\p{Variation_Selector=yes}', "");
Expect(1, 918000, '\p{^Variation_Selector=yes}', "");
Expect(1, 918000, '\P{Variation_Selector=yes}', "");
Expect(0, 918000, '\P{^Variation_Selector=yes}', "");
Expect(1, 917999, '\p{Variation_Selector=-_Yes}', "");
Expect(0, 917999, '\p{^Variation_Selector=-_Yes}', "");
Expect(0, 917999, '\P{Variation_Selector=-_Yes}', "");
Expect(1, 917999, '\P{^Variation_Selector=-_Yes}', "");
Expect(0, 918000, '\p{Variation_Selector=-_Yes}', "");
Expect(1, 918000, '\p{^Variation_Selector=-_Yes}', "");
Expect(1, 918000, '\P{Variation_Selector=-_Yes}', "");
Expect(0, 918000, '\P{^Variation_Selector=-_Yes}', "");
Error('\p{VS=/a/	-y}');
Error('\P{VS=/a/	-y}');
Expect(1, 917999, '\p{VS=y}', "");
Expect(0, 917999, '\p{^VS=y}', "");
Expect(0, 917999, '\P{VS=y}', "");
Expect(1, 917999, '\P{^VS=y}', "");
Expect(0, 918000, '\p{VS=y}', "");
Expect(1, 918000, '\p{^VS=y}', "");
Expect(1, 918000, '\P{VS=y}', "");
Expect(0, 918000, '\P{^VS=y}', "");
Expect(1, 917999, '\p{VS=	 Y}', "");
Expect(0, 917999, '\p{^VS=	 Y}', "");
Expect(0, 917999, '\P{VS=	 Y}', "");
Expect(1, 917999, '\P{^VS=	 Y}', "");
Expect(0, 918000, '\p{VS=	 Y}', "");
Expect(1, 918000, '\p{^VS=	 Y}', "");
Expect(1, 918000, '\P{VS=	 Y}', "");
Expect(0, 918000, '\P{^VS=	 Y}', "");
Error('\p{Is_Variation_Selector=-:=T}');
Error('\P{Is_Variation_Selector=-:=T}');
Expect(1, 917999, '\p{Is_Variation_Selector=t}', "");
Expect(0, 917999, '\p{^Is_Variation_Selector=t}', "");
Expect(0, 917999, '\P{Is_Variation_Selector=t}', "");
Expect(1, 917999, '\P{^Is_Variation_Selector=t}', "");
Expect(0, 918000, '\p{Is_Variation_Selector=t}', "");
Expect(1, 918000, '\p{^Is_Variation_Selector=t}', "");
Expect(1, 918000, '\P{Is_Variation_Selector=t}', "");
Expect(0, 918000, '\P{^Is_Variation_Selector=t}', "");
Expect(1, 917999, '\p{Is_Variation_Selector=_ T}', "");
Expect(0, 917999, '\p{^Is_Variation_Selector=_ T}', "");
Expect(0, 917999, '\P{Is_Variation_Selector=_ T}', "");
Expect(1, 917999, '\P{^Is_Variation_Selector=_ T}', "");
Expect(0, 918000, '\p{Is_Variation_Selector=_ T}', "");
Expect(1, 918000, '\p{^Is_Variation_Selector=_ T}', "");
Expect(1, 918000, '\P{Is_Variation_Selector=_ T}', "");
Expect(0, 918000, '\P{^Is_Variation_Selector=_ T}', "");
Error('\p{Is_VS=-/a/True}');
Error('\P{Is_VS=-/a/True}');
Expect(1, 917999, '\p{Is_VS=true}', "");
Expect(0, 917999, '\p{^Is_VS=true}', "");
Expect(0, 917999, '\P{Is_VS=true}', "");
Expect(1, 917999, '\P{^Is_VS=true}', "");
Expect(0, 918000, '\p{Is_VS=true}', "");
Expect(1, 918000, '\p{^Is_VS=true}', "");
Expect(1, 918000, '\P{Is_VS=true}', "");
Expect(0, 918000, '\P{^Is_VS=true}', "");
Expect(1, 917999, '\p{Is_VS=	True}', "");
Expect(0, 917999, '\p{^Is_VS=	True}', "");
Expect(0, 917999, '\P{Is_VS=	True}', "");
Expect(1, 917999, '\P{^Is_VS=	True}', "");
Expect(0, 918000, '\p{Is_VS=	True}', "");
Expect(1, 918000, '\p{^Is_VS=	True}', "");
Expect(1, 918000, '\P{Is_VS=	True}', "");
Expect(0, 918000, '\P{^Is_VS=	True}', "");
Error('\p{Variation_Selector=:=-No}');
Error('\P{Variation_Selector=:=-No}');
Expect(1, 918000, '\p{Variation_Selector:   no}', "");
Expect(0, 918000, '\p{^Variation_Selector:   no}', "");
Expect(0, 918000, '\P{Variation_Selector:   no}', "");
Expect(1, 918000, '\P{^Variation_Selector:   no}', "");
Expect(0, 917999, '\p{Variation_Selector:   no}', "");
Expect(1, 917999, '\p{^Variation_Selector:   no}', "");
Expect(1, 917999, '\P{Variation_Selector:   no}', "");
Expect(0, 917999, '\P{^Variation_Selector:   no}', "");
Expect(1, 918000, '\p{Variation_Selector=	_NO}', "");
Expect(0, 918000, '\p{^Variation_Selector=	_NO}', "");
Expect(0, 918000, '\P{Variation_Selector=	_NO}', "");
Expect(1, 918000, '\P{^Variation_Selector=	_NO}', "");
Expect(0, 917999, '\p{Variation_Selector=	_NO}', "");
Expect(1, 917999, '\p{^Variation_Selector=	_NO}', "");
Expect(1, 917999, '\P{Variation_Selector=	_NO}', "");
Expect(0, 917999, '\P{^Variation_Selector=	_NO}', "");
Error('\p{VS=-_N:=}');
Error('\P{VS=-_N:=}');
Expect(1, 918000, '\p{VS: n}', "");
Expect(0, 918000, '\p{^VS: n}', "");
Expect(0, 918000, '\P{VS: n}', "");
Expect(1, 918000, '\P{^VS: n}', "");
Expect(0, 917999, '\p{VS: n}', "");
Expect(1, 917999, '\p{^VS: n}', "");
Expect(1, 917999, '\P{VS: n}', "");
Expect(0, 917999, '\P{^VS: n}', "");
Expect(1, 918000, '\p{VS=_	n}', "");
Expect(0, 918000, '\p{^VS=_	n}', "");
Expect(0, 918000, '\P{VS=_	n}', "");
Expect(1, 918000, '\P{^VS=_	n}', "");
Expect(0, 917999, '\p{VS=_	n}', "");
Expect(1, 917999, '\p{^VS=_	n}', "");
Expect(1, 917999, '\P{VS=_	n}', "");
Expect(0, 917999, '\P{^VS=_	n}', "");
Error('\p{Is_Variation_Selector=/a/F}');
Error('\P{Is_Variation_Selector=/a/F}');
Expect(1, 918000, '\p{Is_Variation_Selector=f}', "");
Expect(0, 918000, '\p{^Is_Variation_Selector=f}', "");
Expect(0, 918000, '\P{Is_Variation_Selector=f}', "");
Expect(1, 918000, '\P{^Is_Variation_Selector=f}', "");
Expect(0, 917999, '\p{Is_Variation_Selector=f}', "");
Expect(1, 917999, '\p{^Is_Variation_Selector=f}', "");
Expect(1, 917999, '\P{Is_Variation_Selector=f}', "");
Expect(0, 917999, '\P{^Is_Variation_Selector=f}', "");
Expect(1, 918000, '\p{Is_Variation_Selector=	 F}', "");
Expect(0, 918000, '\p{^Is_Variation_Selector=	 F}', "");
Expect(0, 918000, '\P{Is_Variation_Selector=	 F}', "");
Expect(1, 918000, '\P{^Is_Variation_Selector=	 F}', "");
Expect(0, 917999, '\p{Is_Variation_Selector=	 F}', "");
Expect(1, 917999, '\p{^Is_Variation_Selector=	 F}', "");
Expect(1, 917999, '\P{Is_Variation_Selector=	 F}', "");
Expect(0, 917999, '\P{^Is_Variation_Selector=	 F}', "");
Error('\p{Is_VS=	_FALSE:=}');
Error('\P{Is_VS=	_FALSE:=}');
Expect(1, 918000, '\p{Is_VS=false}', "");
Expect(0, 918000, '\p{^Is_VS=false}', "");
Expect(0, 918000, '\P{Is_VS=false}', "");
Expect(1, 918000, '\P{^Is_VS=false}', "");
Expect(0, 917999, '\p{Is_VS=false}', "");
Expect(1, 917999, '\p{^Is_VS=false}', "");
Expect(1, 917999, '\P{Is_VS=false}', "");
Expect(0, 917999, '\P{^Is_VS=false}', "");
Expect(1, 918000, '\p{Is_VS= -False}', "");
Expect(0, 918000, '\p{^Is_VS= -False}', "");
Expect(0, 918000, '\P{Is_VS= -False}', "");
Expect(1, 918000, '\P{^Is_VS= -False}', "");
Expect(0, 917999, '\p{Is_VS= -False}', "");
Expect(1, 917999, '\p{^Is_VS= -False}', "");
Expect(1, 917999, '\P{Is_VS= -False}', "");
Expect(0, 917999, '\P{^Is_VS= -False}', "");
Error('\p{Changes_When_Casemapped=:=	YES}');
Error('\P{Changes_When_Casemapped=:=	YES}');
Expect(1, 66639, '\p{Changes_When_Casemapped=yes}', "");
Expect(0, 66639, '\p{^Changes_When_Casemapped=yes}', "");
Expect(0, 66639, '\P{Changes_When_Casemapped=yes}', "");
Expect(1, 66639, '\P{^Changes_When_Casemapped=yes}', "");
Expect(0, 66640, '\p{Changes_When_Casemapped=yes}', "");
Expect(1, 66640, '\p{^Changes_When_Casemapped=yes}', "");
Expect(1, 66640, '\P{Changes_When_Casemapped=yes}', "");
Expect(0, 66640, '\P{^Changes_When_Casemapped=yes}', "");
Expect(1, 66639, '\p{Changes_When_Casemapped=-_YES}', "");
Expect(0, 66639, '\p{^Changes_When_Casemapped=-_YES}', "");
Expect(0, 66639, '\P{Changes_When_Casemapped=-_YES}', "");
Expect(1, 66639, '\P{^Changes_When_Casemapped=-_YES}', "");
Expect(0, 66640, '\p{Changes_When_Casemapped=-_YES}', "");
Expect(1, 66640, '\p{^Changes_When_Casemapped=-_YES}', "");
Expect(1, 66640, '\P{Changes_When_Casemapped=-_YES}', "");
Expect(0, 66640, '\P{^Changes_When_Casemapped=-_YES}', "");
Error('\p{CWCM= -y/a/}');
Error('\P{CWCM= -y/a/}');
Expect(1, 66639, '\p{CWCM=y}', "");
Expect(0, 66639, '\p{^CWCM=y}', "");
Expect(0, 66639, '\P{CWCM=y}', "");
Expect(1, 66639, '\P{^CWCM=y}', "");
Expect(0, 66640, '\p{CWCM=y}', "");
Expect(1, 66640, '\p{^CWCM=y}', "");
Expect(1, 66640, '\P{CWCM=y}', "");
Expect(0, 66640, '\P{^CWCM=y}', "");
Expect(1, 66639, '\p{CWCM= Y}', "");
Expect(0, 66639, '\p{^CWCM= Y}', "");
Expect(0, 66639, '\P{CWCM= Y}', "");
Expect(1, 66639, '\P{^CWCM= Y}', "");
Expect(0, 66640, '\p{CWCM= Y}', "");
Expect(1, 66640, '\p{^CWCM= Y}', "");
Expect(1, 66640, '\P{CWCM= Y}', "");
Expect(0, 66640, '\P{^CWCM= Y}', "");
Error('\p{Is_Changes_When_Casemapped=_:=t}');
Error('\P{Is_Changes_When_Casemapped=_:=t}');
Expect(1, 66639, '\p{Is_Changes_When_Casemapped=t}', "");
Expect(0, 66639, '\p{^Is_Changes_When_Casemapped=t}', "");
Expect(0, 66639, '\P{Is_Changes_When_Casemapped=t}', "");
Expect(1, 66639, '\P{^Is_Changes_When_Casemapped=t}', "");
Expect(0, 66640, '\p{Is_Changes_When_Casemapped=t}', "");
Expect(1, 66640, '\p{^Is_Changes_When_Casemapped=t}', "");
Expect(1, 66640, '\P{Is_Changes_When_Casemapped=t}', "");
Expect(0, 66640, '\P{^Is_Changes_When_Casemapped=t}', "");
Expect(1, 66639, '\p{Is_Changes_When_Casemapped=-	T}', "");
Expect(0, 66639, '\p{^Is_Changes_When_Casemapped=-	T}', "");
Expect(0, 66639, '\P{Is_Changes_When_Casemapped=-	T}', "");
Expect(1, 66639, '\P{^Is_Changes_When_Casemapped=-	T}', "");
Expect(0, 66640, '\p{Is_Changes_When_Casemapped=-	T}', "");
Expect(1, 66640, '\p{^Is_Changes_When_Casemapped=-	T}', "");
Expect(1, 66640, '\P{Is_Changes_When_Casemapped=-	T}', "");
Expect(0, 66640, '\P{^Is_Changes_When_Casemapped=-	T}', "");
Error('\p{Is_CWCM=		True:=}');
Error('\P{Is_CWCM=		True:=}');
Expect(1, 66639, '\p{Is_CWCM=true}', "");
Expect(0, 66639, '\p{^Is_CWCM=true}', "");
Expect(0, 66639, '\P{Is_CWCM=true}', "");
Expect(1, 66639, '\P{^Is_CWCM=true}', "");
Expect(0, 66640, '\p{Is_CWCM=true}', "");
Expect(1, 66640, '\p{^Is_CWCM=true}', "");
Expect(1, 66640, '\P{Is_CWCM=true}', "");
Expect(0, 66640, '\P{^Is_CWCM=true}', "");
Expect(1, 66639, '\p{Is_CWCM=-True}', "");
Expect(0, 66639, '\p{^Is_CWCM=-True}', "");
Expect(0, 66639, '\P{Is_CWCM=-True}', "");
Expect(1, 66639, '\P{^Is_CWCM=-True}', "");
Expect(0, 66640, '\p{Is_CWCM=-True}', "");
Expect(1, 66640, '\p{^Is_CWCM=-True}', "");
Expect(1, 66640, '\P{Is_CWCM=-True}', "");
Expect(0, 66640, '\P{^Is_CWCM=-True}', "");
Error('\p{Changes_When_Casemapped= NO:=}');
Error('\P{Changes_When_Casemapped= NO:=}');
Expect(1, 66640, '\p{Changes_When_Casemapped=no}', "");
Expect(0, 66640, '\p{^Changes_When_Casemapped=no}', "");
Expect(0, 66640, '\P{Changes_When_Casemapped=no}', "");
Expect(1, 66640, '\P{^Changes_When_Casemapped=no}', "");
Expect(0, 66639, '\p{Changes_When_Casemapped=no}', "");
Expect(1, 66639, '\p{^Changes_When_Casemapped=no}', "");
Expect(1, 66639, '\P{Changes_When_Casemapped=no}', "");
Expect(0, 66639, '\P{^Changes_When_Casemapped=no}', "");
Error('\p{CWCM=__N:=}');
Error('\P{CWCM=__N:=}');
Expect(1, 66640, '\p{CWCM=n}', "");
Expect(0, 66640, '\p{^CWCM=n}', "");
Expect(0, 66640, '\P{CWCM=n}', "");
Expect(1, 66640, '\P{^CWCM=n}', "");
Expect(0, 66639, '\p{CWCM=n}', "");
Expect(1, 66639, '\p{^CWCM=n}', "");
Expect(1, 66639, '\P{CWCM=n}', "");
Expect(0, 66639, '\P{^CWCM=n}', "");
Expect(1, 66640, '\p{CWCM=	 N}', "");
Expect(0, 66640, '\p{^CWCM=	 N}', "");
Expect(0, 66640, '\P{CWCM=	 N}', "");
Expect(1, 66640, '\P{^CWCM=	 N}', "");
Expect(0, 66639, '\p{CWCM=	 N}', "");
Expect(1, 66639, '\p{^CWCM=	 N}', "");
Expect(1, 66639, '\P{CWCM=	 N}', "");
Expect(0, 66639, '\P{^CWCM=	 N}', "");
Error('\p{Is_Changes_When_Casemapped=:=f}');
Error('\P{Is_Changes_When_Casemapped=:=f}');
Expect(1, 66640, '\p{Is_Changes_When_Casemapped: f}', "");
Expect(0, 66640, '\p{^Is_Changes_When_Casemapped: f}', "");
Expect(0, 66640, '\P{Is_Changes_When_Casemapped: f}', "");
Expect(1, 66640, '\P{^Is_Changes_When_Casemapped: f}', "");
Expect(0, 66639, '\p{Is_Changes_When_Casemapped: f}', "");
Expect(1, 66639, '\p{^Is_Changes_When_Casemapped: f}', "");
Expect(1, 66639, '\P{Is_Changes_When_Casemapped: f}', "");
Expect(0, 66639, '\P{^Is_Changes_When_Casemapped: f}', "");
Expect(1, 66640, '\p{Is_Changes_When_Casemapped=	 F}', "");
Expect(0, 66640, '\p{^Is_Changes_When_Casemapped=	 F}', "");
Expect(0, 66640, '\P{Is_Changes_When_Casemapped=	 F}', "");
Expect(1, 66640, '\P{^Is_Changes_When_Casemapped=	 F}', "");
Expect(0, 66639, '\p{Is_Changes_When_Casemapped=	 F}', "");
Expect(1, 66639, '\p{^Is_Changes_When_Casemapped=	 F}', "");
Expect(1, 66639, '\P{Is_Changes_When_Casemapped=	 F}', "");
Expect(0, 66639, '\P{^Is_Changes_When_Casemapped=	 F}', "");
Error('\p{Is_CWCM=_ False/a/}');
Error('\P{Is_CWCM=_ False/a/}');
Expect(1, 66640, '\p{Is_CWCM=false}', "");
Expect(0, 66640, '\p{^Is_CWCM=false}', "");
Expect(0, 66640, '\P{Is_CWCM=false}', "");
Expect(1, 66640, '\P{^Is_CWCM=false}', "");
Expect(0, 66639, '\p{Is_CWCM=false}', "");
Expect(1, 66639, '\p{^Is_CWCM=false}', "");
Expect(1, 66639, '\P{Is_CWCM=false}', "");
Expect(0, 66639, '\P{^Is_CWCM=false}', "");
Expect(1, 66640, '\p{Is_CWCM=	False}', "");
Expect(0, 66640, '\p{^Is_CWCM=	False}', "");
Expect(0, 66640, '\P{Is_CWCM=	False}', "");
Expect(1, 66640, '\P{^Is_CWCM=	False}', "");
Expect(0, 66639, '\p{Is_CWCM=	False}', "");
Expect(1, 66639, '\p{^Is_CWCM=	False}', "");
Expect(1, 66639, '\P{Is_CWCM=	False}', "");
Expect(0, 66639, '\P{^Is_CWCM=	False}', "");
Error('\p{Full_Composition_Exclusion=_yes:=}');
Error('\P{Full_Composition_Exclusion=_yes:=}');
Expect(1, 195101, '\p{Full_Composition_Exclusion=yes}', "");
Expect(0, 195101, '\p{^Full_Composition_Exclusion=yes}', "");
Expect(0, 195101, '\P{Full_Composition_Exclusion=yes}', "");
Expect(1, 195101, '\P{^Full_Composition_Exclusion=yes}', "");
Expect(0, 195102, '\p{Full_Composition_Exclusion=yes}', "");
Expect(1, 195102, '\p{^Full_Composition_Exclusion=yes}', "");
Expect(1, 195102, '\P{Full_Composition_Exclusion=yes}', "");
Expect(0, 195102, '\P{^Full_Composition_Exclusion=yes}', "");
Expect(1, 195101, '\p{Full_Composition_Exclusion=-_yes}', "");
Expect(0, 195101, '\p{^Full_Composition_Exclusion=-_yes}', "");
Expect(0, 195101, '\P{Full_Composition_Exclusion=-_yes}', "");
Expect(1, 195101, '\P{^Full_Composition_Exclusion=-_yes}', "");
Expect(0, 195102, '\p{Full_Composition_Exclusion=-_yes}', "");
Expect(1, 195102, '\p{^Full_Composition_Exclusion=-_yes}', "");
Expect(1, 195102, '\P{Full_Composition_Exclusion=-_yes}', "");
Expect(0, 195102, '\P{^Full_Composition_Exclusion=-_yes}', "");
Error('\p{Comp_Ex=  Y:=}');
Error('\P{Comp_Ex=  Y:=}');
Expect(1, 195101, '\p{Comp_Ex=y}', "");
Expect(0, 195101, '\p{^Comp_Ex=y}', "");
Expect(0, 195101, '\P{Comp_Ex=y}', "");
Expect(1, 195101, '\P{^Comp_Ex=y}', "");
Expect(0, 195102, '\p{Comp_Ex=y}', "");
Expect(1, 195102, '\p{^Comp_Ex=y}', "");
Expect(1, 195102, '\P{Comp_Ex=y}', "");
Expect(0, 195102, '\P{^Comp_Ex=y}', "");
Expect(1, 195101, '\p{Comp_Ex=	Y}', "");
Expect(0, 195101, '\p{^Comp_Ex=	Y}', "");
Expect(0, 195101, '\P{Comp_Ex=	Y}', "");
Expect(1, 195101, '\P{^Comp_Ex=	Y}', "");
Expect(0, 195102, '\p{Comp_Ex=	Y}', "");
Expect(1, 195102, '\p{^Comp_Ex=	Y}', "");
Expect(1, 195102, '\P{Comp_Ex=	Y}', "");
Expect(0, 195102, '\P{^Comp_Ex=	Y}', "");
Error('\p{Is_Full_Composition_Exclusion=/a/--T}');
Error('\P{Is_Full_Composition_Exclusion=/a/--T}');
Expect(1, 195101, '\p{Is_Full_Composition_Exclusion=t}', "");
Expect(0, 195101, '\p{^Is_Full_Composition_Exclusion=t}', "");
Expect(0, 195101, '\P{Is_Full_Composition_Exclusion=t}', "");
Expect(1, 195101, '\P{^Is_Full_Composition_Exclusion=t}', "");
Expect(0, 195102, '\p{Is_Full_Composition_Exclusion=t}', "");
Expect(1, 195102, '\p{^Is_Full_Composition_Exclusion=t}', "");
Expect(1, 195102, '\P{Is_Full_Composition_Exclusion=t}', "");
Expect(0, 195102, '\P{^Is_Full_Composition_Exclusion=t}', "");
Expect(1, 195101, '\p{Is_Full_Composition_Exclusion=_-T}', "");
Expect(0, 195101, '\p{^Is_Full_Composition_Exclusion=_-T}', "");
Expect(0, 195101, '\P{Is_Full_Composition_Exclusion=_-T}', "");
Expect(1, 195101, '\P{^Is_Full_Composition_Exclusion=_-T}', "");
Expect(0, 195102, '\p{Is_Full_Composition_Exclusion=_-T}', "");
Expect(1, 195102, '\p{^Is_Full_Composition_Exclusion=_-T}', "");
Expect(1, 195102, '\P{Is_Full_Composition_Exclusion=_-T}', "");
Expect(0, 195102, '\P{^Is_Full_Composition_Exclusion=_-T}', "");
Error('\p{Is_Comp_Ex=	_True/a/}');
Error('\P{Is_Comp_Ex=	_True/a/}');
Expect(1, 195101, '\p{Is_Comp_Ex=true}', "");
Expect(0, 195101, '\p{^Is_Comp_Ex=true}', "");
Expect(0, 195101, '\P{Is_Comp_Ex=true}', "");
Expect(1, 195101, '\P{^Is_Comp_Ex=true}', "");
Expect(0, 195102, '\p{Is_Comp_Ex=true}', "");
Expect(1, 195102, '\p{^Is_Comp_Ex=true}', "");
Expect(1, 195102, '\P{Is_Comp_Ex=true}', "");
Expect(0, 195102, '\P{^Is_Comp_Ex=true}', "");
Expect(1, 195101, '\p{Is_Comp_Ex=	True}', "");
Expect(0, 195101, '\p{^Is_Comp_Ex=	True}', "");
Expect(0, 195101, '\P{Is_Comp_Ex=	True}', "");
Expect(1, 195101, '\P{^Is_Comp_Ex=	True}', "");
Expect(0, 195102, '\p{Is_Comp_Ex=	True}', "");
Expect(1, 195102, '\p{^Is_Comp_Ex=	True}', "");
Expect(1, 195102, '\P{Is_Comp_Ex=	True}', "");
Expect(0, 195102, '\P{^Is_Comp_Ex=	True}', "");
Error('\p{Full_Composition_Exclusion=	/a/No}');
Error('\P{Full_Composition_Exclusion=	/a/No}');
Expect(1, 195102, '\p{Full_Composition_Exclusion=no}', "");
Expect(0, 195102, '\p{^Full_Composition_Exclusion=no}', "");
Expect(0, 195102, '\P{Full_Composition_Exclusion=no}', "");
Expect(1, 195102, '\P{^Full_Composition_Exclusion=no}', "");
Expect(0, 195101, '\p{Full_Composition_Exclusion=no}', "");
Expect(1, 195101, '\p{^Full_Composition_Exclusion=no}', "");
Expect(1, 195101, '\P{Full_Composition_Exclusion=no}', "");
Expect(0, 195101, '\P{^Full_Composition_Exclusion=no}', "");
Expect(1, 195102, '\p{Full_Composition_Exclusion= no}', "");
Expect(0, 195102, '\p{^Full_Composition_Exclusion= no}', "");
Expect(0, 195102, '\P{Full_Composition_Exclusion= no}', "");
Expect(1, 195102, '\P{^Full_Composition_Exclusion= no}', "");
Expect(0, 195101, '\p{Full_Composition_Exclusion= no}', "");
Expect(1, 195101, '\p{^Full_Composition_Exclusion= no}', "");
Expect(1, 195101, '\P{Full_Composition_Exclusion= no}', "");
Expect(0, 195101, '\P{^Full_Composition_Exclusion= no}', "");
Error('\p{Comp_Ex=:= n}');
Error('\P{Comp_Ex=:= n}');
Expect(1, 195102, '\p{Comp_Ex=n}', "");
Expect(0, 195102, '\p{^Comp_Ex=n}', "");
Expect(0, 195102, '\P{Comp_Ex=n}', "");
Expect(1, 195102, '\P{^Comp_Ex=n}', "");
Expect(0, 195101, '\p{Comp_Ex=n}', "");
Expect(1, 195101, '\p{^Comp_Ex=n}', "");
Expect(1, 195101, '\P{Comp_Ex=n}', "");
Expect(0, 195101, '\P{^Comp_Ex=n}', "");
Expect(1, 195102, '\p{Comp_Ex=_-N}', "");
Expect(0, 195102, '\p{^Comp_Ex=_-N}', "");
Expect(0, 195102, '\P{Comp_Ex=_-N}', "");
Expect(1, 195102, '\P{^Comp_Ex=_-N}', "");
Expect(0, 195101, '\p{Comp_Ex=_-N}', "");
Expect(1, 195101, '\p{^Comp_Ex=_-N}', "");
Expect(1, 195101, '\P{Comp_Ex=_-N}', "");
Expect(0, 195101, '\P{^Comp_Ex=_-N}', "");
Error('\p{Is_Full_Composition_Exclusion=  F/a/}');
Error('\P{Is_Full_Composition_Exclusion=  F/a/}');
Expect(1, 195102, '\p{Is_Full_Composition_Exclusion=f}', "");
Expect(0, 195102, '\p{^Is_Full_Composition_Exclusion=f}', "");
Expect(0, 195102, '\P{Is_Full_Composition_Exclusion=f}', "");
Expect(1, 195102, '\P{^Is_Full_Composition_Exclusion=f}', "");
Expect(0, 195101, '\p{Is_Full_Composition_Exclusion=f}', "");
Expect(1, 195101, '\p{^Is_Full_Composition_Exclusion=f}', "");
Expect(1, 195101, '\P{Is_Full_Composition_Exclusion=f}', "");
Expect(0, 195101, '\P{^Is_Full_Composition_Exclusion=f}', "");
Expect(1, 195102, '\p{Is_Full_Composition_Exclusion=__f}', "");
Expect(0, 195102, '\p{^Is_Full_Composition_Exclusion=__f}', "");
Expect(0, 195102, '\P{Is_Full_Composition_Exclusion=__f}', "");
Expect(1, 195102, '\P{^Is_Full_Composition_Exclusion=__f}', "");
Expect(0, 195101, '\p{Is_Full_Composition_Exclusion=__f}', "");
Expect(1, 195101, '\p{^Is_Full_Composition_Exclusion=__f}', "");
Expect(1, 195101, '\P{Is_Full_Composition_Exclusion=__f}', "");
Expect(0, 195101, '\P{^Is_Full_Composition_Exclusion=__f}', "");
Error('\p{Is_Comp_Ex=:=_-False}');
Error('\P{Is_Comp_Ex=:=_-False}');
Expect(1, 195102, '\p{Is_Comp_Ex=false}', "");
Expect(0, 195102, '\p{^Is_Comp_Ex=false}', "");
Expect(0, 195102, '\P{Is_Comp_Ex=false}', "");
Expect(1, 195102, '\P{^Is_Comp_Ex=false}', "");
Expect(0, 195101, '\p{Is_Comp_Ex=false}', "");
Expect(1, 195101, '\p{^Is_Comp_Ex=false}', "");
Expect(1, 195101, '\P{Is_Comp_Ex=false}', "");
Expect(0, 195101, '\P{^Is_Comp_Ex=false}', "");
Expect(1, 195102, '\p{Is_Comp_Ex=__False}', "");
Expect(0, 195102, '\p{^Is_Comp_Ex=__False}', "");
Expect(0, 195102, '\P{Is_Comp_Ex=__False}', "");
Expect(1, 195102, '\P{^Is_Comp_Ex=__False}', "");
Expect(0, 195101, '\p{Is_Comp_Ex=__False}', "");
Expect(1, 195101, '\p{^Is_Comp_Ex=__False}', "");
Expect(1, 195101, '\P{Is_Comp_Ex=__False}', "");
Expect(0, 195101, '\P{^Is_Comp_Ex=__False}', "");
Error('\p{Expands_On_NFC=Yes}');
Error('\P{Expands_On_NFC=Yes}');
Error('\p{XO_NFC=Y}');
Error('\P{XO_NFC=Y}');
Error('\p{Is_Expands_On_NFC=T}');
Error('\P{Is_Expands_On_NFC=T}');
Error('\p{Is_XO_NFC=True}');
Error('\P{Is_XO_NFC=True}');
Error('\p{Expands_On_NFC: No}');
Error('\P{Expands_On_NFC: No}');
Error('\p{XO_NFC=N}');
Error('\P{XO_NFC=N}');
Error('\p{Is_Expands_On_NFC=F}');
Error('\P{Is_Expands_On_NFC=F}');
Error('\p{Is_XO_NFC=False}');
Error('\P{Is_XO_NFC=False}');
Error('\p{Changes_When_Uppercased=	 yes/a/}');
Error('\P{Changes_When_Uppercased=	 yes/a/}');
Expect(1, 66639, '\p{Changes_When_Uppercased=yes}', "");
Expect(0, 66639, '\p{^Changes_When_Uppercased=yes}', "");
Expect(0, 66639, '\P{Changes_When_Uppercased=yes}', "");
Expect(1, 66639, '\P{^Changes_When_Uppercased=yes}', "");
Expect(0, 66640, '\p{Changes_When_Uppercased=yes}', "");
Expect(1, 66640, '\p{^Changes_When_Uppercased=yes}', "");
Expect(1, 66640, '\P{Changes_When_Uppercased=yes}', "");
Expect(0, 66640, '\P{^Changes_When_Uppercased=yes}', "");
Expect(1, 66639, '\p{Changes_When_Uppercased=		Yes}', "");
Expect(0, 66639, '\p{^Changes_When_Uppercased=		Yes}', "");
Expect(0, 66639, '\P{Changes_When_Uppercased=		Yes}', "");
Expect(1, 66639, '\P{^Changes_When_Uppercased=		Yes}', "");
Expect(0, 66640, '\p{Changes_When_Uppercased=		Yes}', "");
Expect(1, 66640, '\p{^Changes_When_Uppercased=		Yes}', "");
Expect(1, 66640, '\P{Changes_When_Uppercased=		Yes}', "");
Expect(0, 66640, '\P{^Changes_When_Uppercased=		Yes}', "");
Error('\p{CWU=/a/-Y}');
Error('\P{CWU=/a/-Y}');
Expect(1, 66639, '\p{CWU=y}', "");
Expect(0, 66639, '\p{^CWU=y}', "");
Expect(0, 66639, '\P{CWU=y}', "");
Expect(1, 66639, '\P{^CWU=y}', "");
Expect(0, 66640, '\p{CWU=y}', "");
Expect(1, 66640, '\p{^CWU=y}', "");
Expect(1, 66640, '\P{CWU=y}', "");
Expect(0, 66640, '\P{^CWU=y}', "");
Expect(1, 66639, '\p{CWU=	_Y}', "");
Expect(0, 66639, '\p{^CWU=	_Y}', "");
Expect(0, 66639, '\P{CWU=	_Y}', "");
Expect(1, 66639, '\P{^CWU=	_Y}', "");
Expect(0, 66640, '\p{CWU=	_Y}', "");
Expect(1, 66640, '\p{^CWU=	_Y}', "");
Expect(1, 66640, '\P{CWU=	_Y}', "");
Expect(0, 66640, '\P{^CWU=	_Y}', "");
Error('\p{Is_Changes_When_Uppercased=/a/-_t}');
Error('\P{Is_Changes_When_Uppercased=/a/-_t}');
Expect(1, 66639, '\p{Is_Changes_When_Uppercased=t}', "");
Expect(0, 66639, '\p{^Is_Changes_When_Uppercased=t}', "");
Expect(0, 66639, '\P{Is_Changes_When_Uppercased=t}', "");
Expect(1, 66639, '\P{^Is_Changes_When_Uppercased=t}', "");
Expect(0, 66640, '\p{Is_Changes_When_Uppercased=t}', "");
Expect(1, 66640, '\p{^Is_Changes_When_Uppercased=t}', "");
Expect(1, 66640, '\P{Is_Changes_When_Uppercased=t}', "");
Expect(0, 66640, '\P{^Is_Changes_When_Uppercased=t}', "");
Expect(1, 66639, '\p{Is_Changes_When_Uppercased= T}', "");
Expect(0, 66639, '\p{^Is_Changes_When_Uppercased= T}', "");
Expect(0, 66639, '\P{Is_Changes_When_Uppercased= T}', "");
Expect(1, 66639, '\P{^Is_Changes_When_Uppercased= T}', "");
Expect(0, 66640, '\p{Is_Changes_When_Uppercased= T}', "");
Expect(1, 66640, '\p{^Is_Changes_When_Uppercased= T}', "");
Expect(1, 66640, '\P{Is_Changes_When_Uppercased= T}', "");
Expect(0, 66640, '\P{^Is_Changes_When_Uppercased= T}', "");
Error('\p{Is_CWU:   /a/ TRUE}');
Error('\P{Is_CWU:   /a/ TRUE}');
Expect(1, 66639, '\p{Is_CWU=true}', "");
Expect(0, 66639, '\p{^Is_CWU=true}', "");
Expect(0, 66639, '\P{Is_CWU=true}', "");
Expect(1, 66639, '\P{^Is_CWU=true}', "");
Expect(0, 66640, '\p{Is_CWU=true}', "");
Expect(1, 66640, '\p{^Is_CWU=true}', "");
Expect(1, 66640, '\P{Is_CWU=true}', "");
Expect(0, 66640, '\P{^Is_CWU=true}', "");
Expect(1, 66639, '\p{Is_CWU=_	True}', "");
Expect(0, 66639, '\p{^Is_CWU=_	True}', "");
Expect(0, 66639, '\P{Is_CWU=_	True}', "");
Expect(1, 66639, '\P{^Is_CWU=_	True}', "");
Expect(0, 66640, '\p{Is_CWU=_	True}', "");
Expect(1, 66640, '\p{^Is_CWU=_	True}', "");
Expect(1, 66640, '\P{Is_CWU=_	True}', "");
Expect(0, 66640, '\P{^Is_CWU=_	True}', "");
Error('\p{Changes_When_Uppercased=_/a/no}');
Error('\P{Changes_When_Uppercased=_/a/no}');
Expect(1, 66640, '\p{Changes_When_Uppercased=no}', "");
Expect(0, 66640, '\p{^Changes_When_Uppercased=no}', "");
Expect(0, 66640, '\P{Changes_When_Uppercased=no}', "");
Expect(1, 66640, '\P{^Changes_When_Uppercased=no}', "");
Expect(0, 66639, '\p{Changes_When_Uppercased=no}', "");
Expect(1, 66639, '\p{^Changes_When_Uppercased=no}', "");
Expect(1, 66639, '\P{Changes_When_Uppercased=no}', "");
Expect(0, 66639, '\P{^Changes_When_Uppercased=no}', "");
Expect(1, 66640, '\p{Changes_When_Uppercased=__NO}', "");
Expect(0, 66640, '\p{^Changes_When_Uppercased=__NO}', "");
Expect(0, 66640, '\P{Changes_When_Uppercased=__NO}', "");
Expect(1, 66640, '\P{^Changes_When_Uppercased=__NO}', "");
Expect(0, 66639, '\p{Changes_When_Uppercased=__NO}', "");
Expect(1, 66639, '\p{^Changes_When_Uppercased=__NO}', "");
Expect(1, 66639, '\P{Changes_When_Uppercased=__NO}', "");
Expect(0, 66639, '\P{^Changes_When_Uppercased=__NO}', "");
Error('\p{CWU=/a/-_N}');
Error('\P{CWU=/a/-_N}');
Expect(1, 66640, '\p{CWU=n}', "");
Expect(0, 66640, '\p{^CWU=n}', "");
Expect(0, 66640, '\P{CWU=n}', "");
Expect(1, 66640, '\P{^CWU=n}', "");
Expect(0, 66639, '\p{CWU=n}', "");
Expect(1, 66639, '\p{^CWU=n}', "");
Expect(1, 66639, '\P{CWU=n}', "");
Expect(0, 66639, '\P{^CWU=n}', "");
Expect(1, 66640, '\p{CWU: _N}', "");
Expect(0, 66640, '\p{^CWU: _N}', "");
Expect(0, 66640, '\P{CWU: _N}', "");
Expect(1, 66640, '\P{^CWU: _N}', "");
Expect(0, 66639, '\p{CWU: _N}', "");
Expect(1, 66639, '\p{^CWU: _N}', "");
Expect(1, 66639, '\P{CWU: _N}', "");
Expect(0, 66639, '\P{^CWU: _N}', "");
Error('\p{Is_Changes_When_Uppercased=-F/a/}');
Error('\P{Is_Changes_When_Uppercased=-F/a/}');
Expect(1, 66640, '\p{Is_Changes_When_Uppercased=f}', "");
Expect(0, 66640, '\p{^Is_Changes_When_Uppercased=f}', "");
Expect(0, 66640, '\P{Is_Changes_When_Uppercased=f}', "");
Expect(1, 66640, '\P{^Is_Changes_When_Uppercased=f}', "");
Expect(0, 66639, '\p{Is_Changes_When_Uppercased=f}', "");
Expect(1, 66639, '\p{^Is_Changes_When_Uppercased=f}', "");
Expect(1, 66639, '\P{Is_Changes_When_Uppercased=f}', "");
Expect(0, 66639, '\P{^Is_Changes_When_Uppercased=f}', "");
Expect(1, 66640, '\p{Is_Changes_When_Uppercased= -F}', "");
Expect(0, 66640, '\p{^Is_Changes_When_Uppercased= -F}', "");
Expect(0, 66640, '\P{Is_Changes_When_Uppercased= -F}', "");
Expect(1, 66640, '\P{^Is_Changes_When_Uppercased= -F}', "");
Expect(0, 66639, '\p{Is_Changes_When_Uppercased= -F}', "");
Expect(1, 66639, '\p{^Is_Changes_When_Uppercased= -F}', "");
Expect(1, 66639, '\P{Is_Changes_When_Uppercased= -F}', "");
Expect(0, 66639, '\P{^Is_Changes_When_Uppercased= -F}', "");
Error('\p{Is_CWU=-/a/False}');
Error('\P{Is_CWU=-/a/False}');
Expect(1, 66640, '\p{Is_CWU=false}', "");
Expect(0, 66640, '\p{^Is_CWU=false}', "");
Expect(0, 66640, '\P{Is_CWU=false}', "");
Expect(1, 66640, '\P{^Is_CWU=false}', "");
Expect(0, 66639, '\p{Is_CWU=false}', "");
Expect(1, 66639, '\p{^Is_CWU=false}', "");
Expect(1, 66639, '\P{Is_CWU=false}', "");
Expect(0, 66639, '\P{^Is_CWU=false}', "");
Expect(1, 66640, '\p{Is_CWU= FALSE}', "");
Expect(0, 66640, '\p{^Is_CWU= FALSE}', "");
Expect(0, 66640, '\P{Is_CWU= FALSE}', "");
Expect(1, 66640, '\P{^Is_CWU= FALSE}', "");
Expect(0, 66639, '\p{Is_CWU= FALSE}', "");
Expect(1, 66639, '\p{^Is_CWU= FALSE}', "");
Expect(1, 66639, '\P{Is_CWU= FALSE}', "");
Expect(0, 66639, '\P{^Is_CWU= FALSE}', "");
Error('\p{Alphabetic=:=yes}');
Error('\P{Alphabetic=:=yes}');
Expect(1, 195101, '\p{Alphabetic=yes}', "");
Expect(0, 195101, '\p{^Alphabetic=yes}', "");
Expect(0, 195101, '\P{Alphabetic=yes}', "");
Expect(1, 195101, '\P{^Alphabetic=yes}', "");
Expect(0, 195102, '\p{Alphabetic=yes}', "");
Expect(1, 195102, '\p{^Alphabetic=yes}', "");
Expect(1, 195102, '\P{Alphabetic=yes}', "");
Expect(0, 195102, '\P{^Alphabetic=yes}', "");
Expect(1, 195101, '\p{Alphabetic=-Yes}', "");
Expect(0, 195101, '\p{^Alphabetic=-Yes}', "");
Expect(0, 195101, '\P{Alphabetic=-Yes}', "");
Expect(1, 195101, '\P{^Alphabetic=-Yes}', "");
Expect(0, 195102, '\p{Alphabetic=-Yes}', "");
Expect(1, 195102, '\p{^Alphabetic=-Yes}', "");
Expect(1, 195102, '\P{Alphabetic=-Yes}', "");
Expect(0, 195102, '\P{^Alphabetic=-Yes}', "");
Error('\p{Alpha=	:=Y}');
Error('\P{Alpha=	:=Y}');
Expect(1, 195101, '\p{Alpha=y}', "");
Expect(0, 195101, '\p{^Alpha=y}', "");
Expect(0, 195101, '\P{Alpha=y}', "");
Expect(1, 195101, '\P{^Alpha=y}', "");
Expect(0, 195102, '\p{Alpha=y}', "");
Expect(1, 195102, '\p{^Alpha=y}', "");
Expect(1, 195102, '\P{Alpha=y}', "");
Expect(0, 195102, '\P{^Alpha=y}', "");
Expect(1, 195101, '\p{Alpha=_Y}', "");
Expect(0, 195101, '\p{^Alpha=_Y}', "");
Expect(0, 195101, '\P{Alpha=_Y}', "");
Expect(1, 195101, '\P{^Alpha=_Y}', "");
Expect(0, 195102, '\p{Alpha=_Y}', "");
Expect(1, 195102, '\p{^Alpha=_Y}', "");
Expect(1, 195102, '\P{Alpha=_Y}', "");
Expect(0, 195102, '\P{^Alpha=_Y}', "");
Error('\p{Is_Alphabetic=/a/-_t}');
Error('\P{Is_Alphabetic=/a/-_t}');
Expect(1, 195101, '\p{Is_Alphabetic=t}', "");
Expect(0, 195101, '\p{^Is_Alphabetic=t}', "");
Expect(0, 195101, '\P{Is_Alphabetic=t}', "");
Expect(1, 195101, '\P{^Is_Alphabetic=t}', "");
Expect(0, 195102, '\p{Is_Alphabetic=t}', "");
Expect(1, 195102, '\p{^Is_Alphabetic=t}', "");
Expect(1, 195102, '\P{Is_Alphabetic=t}', "");
Expect(0, 195102, '\P{^Is_Alphabetic=t}', "");
Expect(1, 195101, '\p{Is_Alphabetic:   	 T}', "");
Expect(0, 195101, '\p{^Is_Alphabetic:   	 T}', "");
Expect(0, 195101, '\P{Is_Alphabetic:   	 T}', "");
Expect(1, 195101, '\P{^Is_Alphabetic:   	 T}', "");
Expect(0, 195102, '\p{Is_Alphabetic:   	 T}', "");
Expect(1, 195102, '\p{^Is_Alphabetic:   	 T}', "");
Expect(1, 195102, '\P{Is_Alphabetic:   	 T}', "");
Expect(0, 195102, '\P{^Is_Alphabetic:   	 T}', "");
Error('\p{Is_Alpha= 	true/a/}');
Error('\P{Is_Alpha= 	true/a/}');
Expect(1, 195101, '\p{Is_Alpha=true}', "");
Expect(0, 195101, '\p{^Is_Alpha=true}', "");
Expect(0, 195101, '\P{Is_Alpha=true}', "");
Expect(1, 195101, '\P{^Is_Alpha=true}', "");
Expect(0, 195102, '\p{Is_Alpha=true}', "");
Expect(1, 195102, '\p{^Is_Alpha=true}', "");
Expect(1, 195102, '\P{Is_Alpha=true}', "");
Expect(0, 195102, '\P{^Is_Alpha=true}', "");
Expect(1, 195101, '\p{Is_Alpha=-True}', "");
Expect(0, 195101, '\p{^Is_Alpha=-True}', "");
Expect(0, 195101, '\P{Is_Alpha=-True}', "");
Expect(1, 195101, '\P{^Is_Alpha=-True}', "");
Expect(0, 195102, '\p{Is_Alpha=-True}', "");
Expect(1, 195102, '\p{^Is_Alpha=-True}', "");
Expect(1, 195102, '\P{Is_Alpha=-True}', "");
Expect(0, 195102, '\P{^Is_Alpha=-True}', "");
Error('\p{Alphabetic=_	no:=}');
Error('\P{Alphabetic=_	no:=}');
Expect(1, 195102, '\p{Alphabetic=no}', "");
Expect(0, 195102, '\p{^Alphabetic=no}', "");
Expect(0, 195102, '\P{Alphabetic=no}', "");
Expect(1, 195102, '\P{^Alphabetic=no}', "");
Expect(0, 195101, '\p{Alphabetic=no}', "");
Expect(1, 195101, '\p{^Alphabetic=no}', "");
Expect(1, 195101, '\P{Alphabetic=no}', "");
Expect(0, 195101, '\P{^Alphabetic=no}', "");
Expect(1, 195102, '\p{Alphabetic=	-no}', "");
Expect(0, 195102, '\p{^Alphabetic=	-no}', "");
Expect(0, 195102, '\P{Alphabetic=	-no}', "");
Expect(1, 195102, '\P{^Alphabetic=	-no}', "");
Expect(0, 195101, '\p{Alphabetic=	-no}', "");
Expect(1, 195101, '\p{^Alphabetic=	-no}', "");
Expect(1, 195101, '\P{Alphabetic=	-no}', "");
Expect(0, 195101, '\P{^Alphabetic=	-no}', "");
Error('\p{Alpha=:=	 N}');
Error('\P{Alpha=:=	 N}');
Expect(1, 195102, '\p{Alpha=n}', "");
Expect(0, 195102, '\p{^Alpha=n}', "");
Expect(0, 195102, '\P{Alpha=n}', "");
Expect(1, 195102, '\P{^Alpha=n}', "");
Expect(0, 195101, '\p{Alpha=n}', "");
Expect(1, 195101, '\p{^Alpha=n}', "");
Expect(1, 195101, '\P{Alpha=n}', "");
Expect(0, 195101, '\P{^Alpha=n}', "");
Expect(1, 195102, '\p{Alpha=- N}', "");
Expect(0, 195102, '\p{^Alpha=- N}', "");
Expect(0, 195102, '\P{Alpha=- N}', "");
Expect(1, 195102, '\P{^Alpha=- N}', "");
Expect(0, 195101, '\p{Alpha=- N}', "");
Expect(1, 195101, '\p{^Alpha=- N}', "");
Expect(1, 195101, '\P{Alpha=- N}', "");
Expect(0, 195101, '\P{^Alpha=- N}', "");
Error('\p{Is_Alphabetic: -:=F}');
Error('\P{Is_Alphabetic: -:=F}');
Expect(1, 195102, '\p{Is_Alphabetic=f}', "");
Expect(0, 195102, '\p{^Is_Alphabetic=f}', "");
Expect(0, 195102, '\P{Is_Alphabetic=f}', "");
Expect(1, 195102, '\P{^Is_Alphabetic=f}', "");
Expect(0, 195101, '\p{Is_Alphabetic=f}', "");
Expect(1, 195101, '\p{^Is_Alphabetic=f}', "");
Expect(1, 195101, '\P{Is_Alphabetic=f}', "");
Expect(0, 195101, '\P{^Is_Alphabetic=f}', "");
Expect(1, 195102, '\p{Is_Alphabetic=-F}', "");
Expect(0, 195102, '\p{^Is_Alphabetic=-F}', "");
Expect(0, 195102, '\P{Is_Alphabetic=-F}', "");
Expect(1, 195102, '\P{^Is_Alphabetic=-F}', "");
Expect(0, 195101, '\p{Is_Alphabetic=-F}', "");
Expect(1, 195101, '\p{^Is_Alphabetic=-F}', "");
Expect(1, 195101, '\P{Is_Alphabetic=-F}', "");
Expect(0, 195101, '\P{^Is_Alphabetic=-F}', "");
Error('\p{Is_Alpha=_:=FALSE}');
Error('\P{Is_Alpha=_:=FALSE}');
Expect(1, 195102, '\p{Is_Alpha=false}', "");
Expect(0, 195102, '\p{^Is_Alpha=false}', "");
Expect(0, 195102, '\P{Is_Alpha=false}', "");
Expect(1, 195102, '\P{^Is_Alpha=false}', "");
Expect(0, 195101, '\p{Is_Alpha=false}', "");
Expect(1, 195101, '\p{^Is_Alpha=false}', "");
Expect(1, 195101, '\P{Is_Alpha=false}', "");
Expect(0, 195101, '\P{^Is_Alpha=false}', "");
Expect(1, 195102, '\p{Is_Alpha=_False}', "");
Expect(0, 195102, '\p{^Is_Alpha=_False}', "");
Expect(0, 195102, '\P{Is_Alpha=_False}', "");
Expect(1, 195102, '\P{^Is_Alpha=_False}', "");
Expect(0, 195101, '\p{Is_Alpha=_False}', "");
Expect(1, 195101, '\p{^Is_Alpha=_False}', "");
Expect(1, 195101, '\P{Is_Alpha=_False}', "");
Expect(0, 195101, '\P{^Is_Alpha=_False}', "");
Error('\p{Hex_Digit=-:=Yes}');
Error('\P{Hex_Digit=-:=Yes}');
Expect(1, 65350, '\p{Hex_Digit=yes}', "");
Expect(0, 65350, '\p{^Hex_Digit=yes}', "");
Expect(0, 65350, '\P{Hex_Digit=yes}', "");
Expect(1, 65350, '\P{^Hex_Digit=yes}', "");
Expect(0, 65351, '\p{Hex_Digit=yes}', "");
Expect(1, 65351, '\p{^Hex_Digit=yes}', "");
Expect(1, 65351, '\P{Hex_Digit=yes}', "");
Expect(0, 65351, '\P{^Hex_Digit=yes}', "");
Expect(1, 65350, '\p{Hex_Digit=_-yes}', "");
Expect(0, 65350, '\p{^Hex_Digit=_-yes}', "");
Expect(0, 65350, '\P{Hex_Digit=_-yes}', "");
Expect(1, 65350, '\P{^Hex_Digit=_-yes}', "");
Expect(0, 65351, '\p{Hex_Digit=_-yes}', "");
Expect(1, 65351, '\p{^Hex_Digit=_-yes}', "");
Expect(1, 65351, '\P{Hex_Digit=_-yes}', "");
Expect(0, 65351, '\P{^Hex_Digit=_-yes}', "");
Error('\p{Hex=/a/Y}');
Error('\P{Hex=/a/Y}');
Expect(1, 65350, '\p{Hex=y}', "");
Expect(0, 65350, '\p{^Hex=y}', "");
Expect(0, 65350, '\P{Hex=y}', "");
Expect(1, 65350, '\P{^Hex=y}', "");
Expect(0, 65351, '\p{Hex=y}', "");
Expect(1, 65351, '\p{^Hex=y}', "");
Expect(1, 65351, '\P{Hex=y}', "");
Expect(0, 65351, '\P{^Hex=y}', "");
Expect(1, 65350, '\p{Hex=	-Y}', "");
Expect(0, 65350, '\p{^Hex=	-Y}', "");
Expect(0, 65350, '\P{Hex=	-Y}', "");
Expect(1, 65350, '\P{^Hex=	-Y}', "");
Expect(0, 65351, '\p{Hex=	-Y}', "");
Expect(1, 65351, '\p{^Hex=	-Y}', "");
Expect(1, 65351, '\P{Hex=	-Y}', "");
Expect(0, 65351, '\P{^Hex=	-Y}', "");
Error('\p{Is_Hex_Digit=:=_ T}');
Error('\P{Is_Hex_Digit=:=_ T}');
Expect(1, 65350, '\p{Is_Hex_Digit=t}', "");
Expect(0, 65350, '\p{^Is_Hex_Digit=t}', "");
Expect(0, 65350, '\P{Is_Hex_Digit=t}', "");
Expect(1, 65350, '\P{^Is_Hex_Digit=t}', "");
Expect(0, 65351, '\p{Is_Hex_Digit=t}', "");
Expect(1, 65351, '\p{^Is_Hex_Digit=t}', "");
Expect(1, 65351, '\P{Is_Hex_Digit=t}', "");
Expect(0, 65351, '\P{^Is_Hex_Digit=t}', "");
Expect(1, 65350, '\p{Is_Hex_Digit:	 T}', "");
Expect(0, 65350, '\p{^Is_Hex_Digit:	 T}', "");
Expect(0, 65350, '\P{Is_Hex_Digit:	 T}', "");
Expect(1, 65350, '\P{^Is_Hex_Digit:	 T}', "");
Expect(0, 65351, '\p{Is_Hex_Digit:	 T}', "");
Expect(1, 65351, '\p{^Is_Hex_Digit:	 T}', "");
Expect(1, 65351, '\P{Is_Hex_Digit:	 T}', "");
Expect(0, 65351, '\P{^Is_Hex_Digit:	 T}', "");
Error('\p{Is_Hex=-	True:=}');
Error('\P{Is_Hex=-	True:=}');
Expect(1, 65350, '\p{Is_Hex=true}', "");
Expect(0, 65350, '\p{^Is_Hex=true}', "");
Expect(0, 65350, '\P{Is_Hex=true}', "");
Expect(1, 65350, '\P{^Is_Hex=true}', "");
Expect(0, 65351, '\p{Is_Hex=true}', "");
Expect(1, 65351, '\p{^Is_Hex=true}', "");
Expect(1, 65351, '\P{Is_Hex=true}', "");
Expect(0, 65351, '\P{^Is_Hex=true}', "");
Expect(1, 65350, '\p{Is_Hex=-_TRUE}', "");
Expect(0, 65350, '\p{^Is_Hex=-_TRUE}', "");
Expect(0, 65350, '\P{Is_Hex=-_TRUE}', "");
Expect(1, 65350, '\P{^Is_Hex=-_TRUE}', "");
Expect(0, 65351, '\p{Is_Hex=-_TRUE}', "");
Expect(1, 65351, '\p{^Is_Hex=-_TRUE}', "");
Expect(1, 65351, '\P{Is_Hex=-_TRUE}', "");
Expect(0, 65351, '\P{^Is_Hex=-_TRUE}', "");
Error('\p{Hex_Digit=:=  NO}');
Error('\P{Hex_Digit=:=  NO}');
Expect(1, 65351, '\p{Hex_Digit=no}', "");
Expect(0, 65351, '\p{^Hex_Digit=no}', "");
Expect(0, 65351, '\P{Hex_Digit=no}', "");
Expect(1, 65351, '\P{^Hex_Digit=no}', "");
Expect(0, 65350, '\p{Hex_Digit=no}', "");
Expect(1, 65350, '\p{^Hex_Digit=no}', "");
Expect(1, 65350, '\P{Hex_Digit=no}', "");
Expect(0, 65350, '\P{^Hex_Digit=no}', "");
Expect(1, 65351, '\p{Hex_Digit=	NO}', "");
Expect(0, 65351, '\p{^Hex_Digit=	NO}', "");
Expect(0, 65351, '\P{Hex_Digit=	NO}', "");
Expect(1, 65351, '\P{^Hex_Digit=	NO}', "");
Expect(0, 65350, '\p{Hex_Digit=	NO}', "");
Expect(1, 65350, '\p{^Hex_Digit=	NO}', "");
Expect(1, 65350, '\P{Hex_Digit=	NO}', "");
Expect(0, 65350, '\P{^Hex_Digit=	NO}', "");
Error('\p{Hex= n:=}');
Error('\P{Hex= n:=}');
Expect(1, 65351, '\p{Hex=n}', "");
Expect(0, 65351, '\p{^Hex=n}', "");
Expect(0, 65351, '\P{Hex=n}', "");
Expect(1, 65351, '\P{^Hex=n}', "");
Expect(0, 65350, '\p{Hex=n}', "");
Expect(1, 65350, '\p{^Hex=n}', "");
Expect(1, 65350, '\P{Hex=n}', "");
Expect(0, 65350, '\P{^Hex=n}', "");
Expect(1, 65351, '\p{Hex=	N}', "");
Expect(0, 65351, '\p{^Hex=	N}', "");
Expect(0, 65351, '\P{Hex=	N}', "");
Expect(1, 65351, '\P{^Hex=	N}', "");
Expect(0, 65350, '\p{Hex=	N}', "");
Expect(1, 65350, '\p{^Hex=	N}', "");
Expect(1, 65350, '\P{Hex=	N}', "");
Expect(0, 65350, '\P{^Hex=	N}', "");
Error('\p{Is_Hex_Digit=--F:=}');
Error('\P{Is_Hex_Digit=--F:=}');
Expect(1, 65351, '\p{Is_Hex_Digit=f}', "");
Expect(0, 65351, '\p{^Is_Hex_Digit=f}', "");
Expect(0, 65351, '\P{Is_Hex_Digit=f}', "");
Expect(1, 65351, '\P{^Is_Hex_Digit=f}', "");
Expect(0, 65350, '\p{Is_Hex_Digit=f}', "");
Expect(1, 65350, '\p{^Is_Hex_Digit=f}', "");
Expect(1, 65350, '\P{Is_Hex_Digit=f}', "");
Expect(0, 65350, '\P{^Is_Hex_Digit=f}', "");
Expect(1, 65351, '\p{Is_Hex_Digit= f}', "");
Expect(0, 65351, '\p{^Is_Hex_Digit= f}', "");
Expect(0, 65351, '\P{Is_Hex_Digit= f}', "");
Expect(1, 65351, '\P{^Is_Hex_Digit= f}', "");
Expect(0, 65350, '\p{Is_Hex_Digit= f}', "");
Expect(1, 65350, '\p{^Is_Hex_Digit= f}', "");
Expect(1, 65350, '\P{Is_Hex_Digit= f}', "");
Expect(0, 65350, '\P{^Is_Hex_Digit= f}', "");
Error('\p{Is_Hex=	_False:=}');
Error('\P{Is_Hex=	_False:=}');
Expect(1, 65351, '\p{Is_Hex=false}', "");
Expect(0, 65351, '\p{^Is_Hex=false}', "");
Expect(0, 65351, '\P{Is_Hex=false}', "");
Expect(1, 65351, '\P{^Is_Hex=false}', "");
Expect(0, 65350, '\p{Is_Hex=false}', "");
Expect(1, 65350, '\p{^Is_Hex=false}', "");
Expect(1, 65350, '\P{Is_Hex=false}', "");
Expect(0, 65350, '\P{^Is_Hex=false}', "");
Expect(1, 65351, '\p{Is_Hex= false}', "");
Expect(0, 65351, '\p{^Is_Hex= false}', "");
Expect(0, 65351, '\P{Is_Hex= false}', "");
Expect(1, 65351, '\P{^Is_Hex= false}', "");
Expect(0, 65350, '\p{Is_Hex= false}', "");
Expect(1, 65350, '\p{^Is_Hex= false}', "");
Expect(1, 65350, '\P{Is_Hex= false}', "");
Expect(0, 65350, '\P{^Is_Hex= false}', "");
Error('\p{Dash=	_Yes/a/}');
Error('\P{Dash=	_Yes/a/}');
Expect(1, 65293, '\p{Dash:	yes}', "");
Expect(0, 65293, '\p{^Dash:	yes}', "");
Expect(0, 65293, '\P{Dash:	yes}', "");
Expect(1, 65293, '\P{^Dash:	yes}', "");
Expect(0, 65294, '\p{Dash:	yes}', "");
Expect(1, 65294, '\p{^Dash:	yes}', "");
Expect(1, 65294, '\P{Dash:	yes}', "");
Expect(0, 65294, '\P{^Dash:	yes}', "");
Expect(1, 65293, '\p{Dash=		Yes}', "");
Expect(0, 65293, '\p{^Dash=		Yes}', "");
Expect(0, 65293, '\P{Dash=		Yes}', "");
Expect(1, 65293, '\P{^Dash=		Yes}', "");
Expect(0, 65294, '\p{Dash=		Yes}', "");
Expect(1, 65294, '\p{^Dash=		Yes}', "");
Expect(1, 65294, '\P{Dash=		Yes}', "");
Expect(0, 65294, '\P{^Dash=		Yes}', "");
Error('\p{Is_Dash: :=	Y}');
Error('\P{Is_Dash: :=	Y}');
Expect(1, 65293, '\p{Is_Dash=y}', "");
Expect(0, 65293, '\p{^Is_Dash=y}', "");
Expect(0, 65293, '\P{Is_Dash=y}', "");
Expect(1, 65293, '\P{^Is_Dash=y}', "");
Expect(0, 65294, '\p{Is_Dash=y}', "");
Expect(1, 65294, '\p{^Is_Dash=y}', "");
Expect(1, 65294, '\P{Is_Dash=y}', "");
Expect(0, 65294, '\P{^Is_Dash=y}', "");
Expect(1, 65293, '\p{Is_Dash=-Y}', "");
Expect(0, 65293, '\p{^Is_Dash=-Y}', "");
Expect(0, 65293, '\P{Is_Dash=-Y}', "");
Expect(1, 65293, '\P{^Is_Dash=-Y}', "");
Expect(0, 65294, '\p{Is_Dash=-Y}', "");
Expect(1, 65294, '\p{^Is_Dash=-Y}', "");
Expect(1, 65294, '\P{Is_Dash=-Y}', "");
Expect(0, 65294, '\P{^Is_Dash=-Y}', "");
Error('\p{Dash= :=t}');
Error('\P{Dash= :=t}');
Expect(1, 65293, '\p{Dash=t}', "");
Expect(0, 65293, '\p{^Dash=t}', "");
Expect(0, 65293, '\P{Dash=t}', "");
Expect(1, 65293, '\P{^Dash=t}', "");
Expect(0, 65294, '\p{Dash=t}', "");
Expect(1, 65294, '\p{^Dash=t}', "");
Expect(1, 65294, '\P{Dash=t}', "");
Expect(0, 65294, '\P{^Dash=t}', "");
Expect(1, 65293, '\p{Dash=	 T}', "");
Expect(0, 65293, '\p{^Dash=	 T}', "");
Expect(0, 65293, '\P{Dash=	 T}', "");
Expect(1, 65293, '\P{^Dash=	 T}', "");
Expect(0, 65294, '\p{Dash=	 T}', "");
Expect(1, 65294, '\p{^Dash=	 T}', "");
Expect(1, 65294, '\P{Dash=	 T}', "");
Expect(0, 65294, '\P{^Dash=	 T}', "");
Error('\p{Is_Dash=/a/	True}');
Error('\P{Is_Dash=/a/	True}');
Expect(1, 65293, '\p{Is_Dash=true}', "");
Expect(0, 65293, '\p{^Is_Dash=true}', "");
Expect(0, 65293, '\P{Is_Dash=true}', "");
Expect(1, 65293, '\P{^Is_Dash=true}', "");
Expect(0, 65294, '\p{Is_Dash=true}', "");
Expect(1, 65294, '\p{^Is_Dash=true}', "");
Expect(1, 65294, '\P{Is_Dash=true}', "");
Expect(0, 65294, '\P{^Is_Dash=true}', "");
Expect(1, 65293, '\p{Is_Dash=-TRUE}', "");
Expect(0, 65293, '\p{^Is_Dash=-TRUE}', "");
Expect(0, 65293, '\P{Is_Dash=-TRUE}', "");
Expect(1, 65293, '\P{^Is_Dash=-TRUE}', "");
Expect(0, 65294, '\p{Is_Dash=-TRUE}', "");
Expect(1, 65294, '\p{^Is_Dash=-TRUE}', "");
Expect(1, 65294, '\P{Is_Dash=-TRUE}', "");
Expect(0, 65294, '\P{^Is_Dash=-TRUE}', "");
Error('\p{Dash=_ No/a/}');
Error('\P{Dash=_ No/a/}');
Expect(1, 65294, '\p{Dash: no}', "");
Expect(0, 65294, '\p{^Dash: no}', "");
Expect(0, 65294, '\P{Dash: no}', "");
Expect(1, 65294, '\P{^Dash: no}', "");
Expect(0, 65293, '\p{Dash: no}', "");
Expect(1, 65293, '\p{^Dash: no}', "");
Expect(1, 65293, '\P{Dash: no}', "");
Expect(0, 65293, '\P{^Dash: no}', "");
Expect(1, 65294, '\p{Dash=_ No}', "");
Expect(0, 65294, '\p{^Dash=_ No}', "");
Expect(0, 65294, '\P{Dash=_ No}', "");
Expect(1, 65294, '\P{^Dash=_ No}', "");
Expect(0, 65293, '\p{Dash=_ No}', "");
Expect(1, 65293, '\p{^Dash=_ No}', "");
Expect(1, 65293, '\P{Dash=_ No}', "");
Expect(0, 65293, '\P{^Dash=_ No}', "");
Error('\p{Is_Dash=	n:=}');
Error('\P{Is_Dash=	n:=}');
Expect(1, 65294, '\p{Is_Dash=n}', "");
Expect(0, 65294, '\p{^Is_Dash=n}', "");
Expect(0, 65294, '\P{Is_Dash=n}', "");
Expect(1, 65294, '\P{^Is_Dash=n}', "");
Expect(0, 65293, '\p{Is_Dash=n}', "");
Expect(1, 65293, '\p{^Is_Dash=n}', "");
Expect(1, 65293, '\P{Is_Dash=n}', "");
Expect(0, 65293, '\P{^Is_Dash=n}', "");
Expect(1, 65294, '\p{Is_Dash=--N}', "");
Expect(0, 65294, '\p{^Is_Dash=--N}', "");
Expect(0, 65294, '\P{Is_Dash=--N}', "");
Expect(1, 65294, '\P{^Is_Dash=--N}', "");
Expect(0, 65293, '\p{Is_Dash=--N}', "");
Expect(1, 65293, '\p{^Is_Dash=--N}', "");
Expect(1, 65293, '\P{Is_Dash=--N}', "");
Expect(0, 65293, '\P{^Is_Dash=--N}', "");
Error('\p{Dash=	/a/F}');
Error('\P{Dash=	/a/F}');
Expect(1, 65294, '\p{Dash=f}', "");
Expect(0, 65294, '\p{^Dash=f}', "");
Expect(0, 65294, '\P{Dash=f}', "");
Expect(1, 65294, '\P{^Dash=f}', "");
Expect(0, 65293, '\p{Dash=f}', "");
Expect(1, 65293, '\p{^Dash=f}', "");
Expect(1, 65293, '\P{Dash=f}', "");
Expect(0, 65293, '\P{^Dash=f}', "");
Expect(1, 65294, '\p{Dash= -F}', "");
Expect(0, 65294, '\p{^Dash= -F}', "");
Expect(0, 65294, '\P{Dash= -F}', "");
Expect(1, 65294, '\P{^Dash= -F}', "");
Expect(0, 65293, '\p{Dash= -F}', "");
Expect(1, 65293, '\p{^Dash= -F}', "");
Expect(1, 65293, '\P{Dash= -F}', "");
Expect(0, 65293, '\P{^Dash= -F}', "");
Error('\p{Is_Dash= FALSE/a/}');
Error('\P{Is_Dash= FALSE/a/}');
Expect(1, 65294, '\p{Is_Dash=false}', "");
Expect(0, 65294, '\p{^Is_Dash=false}', "");
Expect(0, 65294, '\P{Is_Dash=false}', "");
Expect(1, 65294, '\P{^Is_Dash=false}', "");
Expect(0, 65293, '\p{Is_Dash=false}', "");
Expect(1, 65293, '\p{^Is_Dash=false}', "");
Expect(1, 65293, '\P{Is_Dash=false}', "");
Expect(0, 65293, '\P{^Is_Dash=false}', "");
Expect(1, 65294, '\p{Is_Dash= 	false}', "");
Expect(0, 65294, '\p{^Is_Dash= 	false}', "");
Expect(0, 65294, '\P{Is_Dash= 	false}', "");
Expect(1, 65294, '\P{^Is_Dash= 	false}', "");
Expect(0, 65293, '\p{Is_Dash= 	false}', "");
Expect(1, 65293, '\p{^Is_Dash= 	false}', "");
Expect(1, 65293, '\P{Is_Dash= 	false}', "");
Expect(0, 65293, '\P{^Is_Dash= 	false}', "");
Error('\p{Changes_When_Lowercased=:=Yes}');
Error('\P{Changes_When_Lowercased=:=Yes}');
Expect(1, 66599, '\p{Changes_When_Lowercased=yes}', "");
Expect(0, 66599, '\p{^Changes_When_Lowercased=yes}', "");
Expect(0, 66599, '\P{Changes_When_Lowercased=yes}', "");
Expect(1, 66599, '\P{^Changes_When_Lowercased=yes}', "");
Expect(0, 66600, '\p{Changes_When_Lowercased=yes}', "");
Expect(1, 66600, '\p{^Changes_When_Lowercased=yes}', "");
Expect(1, 66600, '\P{Changes_When_Lowercased=yes}', "");
Expect(0, 66600, '\P{^Changes_When_Lowercased=yes}', "");
Expect(1, 66599, '\p{Changes_When_Lowercased=-Yes}', "");
Expect(0, 66599, '\p{^Changes_When_Lowercased=-Yes}', "");
Expect(0, 66599, '\P{Changes_When_Lowercased=-Yes}', "");
Expect(1, 66599, '\P{^Changes_When_Lowercased=-Yes}', "");
Expect(0, 66600, '\p{Changes_When_Lowercased=-Yes}', "");
Expect(1, 66600, '\p{^Changes_When_Lowercased=-Yes}', "");
Expect(1, 66600, '\P{Changes_When_Lowercased=-Yes}', "");
Expect(0, 66600, '\P{^Changes_When_Lowercased=-Yes}', "");
Error('\p{CWL=/a/  Y}');
Error('\P{CWL=/a/  Y}');
Expect(1, 66599, '\p{CWL=y}', "");
Expect(0, 66599, '\p{^CWL=y}', "");
Expect(0, 66599, '\P{CWL=y}', "");
Expect(1, 66599, '\P{^CWL=y}', "");
Expect(0, 66600, '\p{CWL=y}', "");
Expect(1, 66600, '\p{^CWL=y}', "");
Expect(1, 66600, '\P{CWL=y}', "");
Expect(0, 66600, '\P{^CWL=y}', "");
Expect(1, 66599, '\p{CWL=_-Y}', "");
Expect(0, 66599, '\p{^CWL=_-Y}', "");
Expect(0, 66599, '\P{CWL=_-Y}', "");
Expect(1, 66599, '\P{^CWL=_-Y}', "");
Expect(0, 66600, '\p{CWL=_-Y}', "");
Expect(1, 66600, '\p{^CWL=_-Y}', "");
Expect(1, 66600, '\P{CWL=_-Y}', "");
Expect(0, 66600, '\P{^CWL=_-Y}', "");
Error('\p{Is_Changes_When_Lowercased=:=-t}');
Error('\P{Is_Changes_When_Lowercased=:=-t}');
Expect(1, 66599, '\p{Is_Changes_When_Lowercased=t}', "");
Expect(0, 66599, '\p{^Is_Changes_When_Lowercased=t}', "");
Expect(0, 66599, '\P{Is_Changes_When_Lowercased=t}', "");
Expect(1, 66599, '\P{^Is_Changes_When_Lowercased=t}', "");
Expect(0, 66600, '\p{Is_Changes_When_Lowercased=t}', "");
Expect(1, 66600, '\p{^Is_Changes_When_Lowercased=t}', "");
Expect(1, 66600, '\P{Is_Changes_When_Lowercased=t}', "");
Expect(0, 66600, '\P{^Is_Changes_When_Lowercased=t}', "");
Expect(1, 66599, '\p{Is_Changes_When_Lowercased= 	T}', "");
Expect(0, 66599, '\p{^Is_Changes_When_Lowercased= 	T}', "");
Expect(0, 66599, '\P{Is_Changes_When_Lowercased= 	T}', "");
Expect(1, 66599, '\P{^Is_Changes_When_Lowercased= 	T}', "");
Expect(0, 66600, '\p{Is_Changes_When_Lowercased= 	T}', "");
Expect(1, 66600, '\p{^Is_Changes_When_Lowercased= 	T}', "");
Expect(1, 66600, '\P{Is_Changes_When_Lowercased= 	T}', "");
Expect(0, 66600, '\P{^Is_Changes_When_Lowercased= 	T}', "");
Error('\p{Is_CWL=:= -True}');
Error('\P{Is_CWL=:= -True}');
Expect(1, 66599, '\p{Is_CWL=true}', "");
Expect(0, 66599, '\p{^Is_CWL=true}', "");
Expect(0, 66599, '\P{Is_CWL=true}', "");
Expect(1, 66599, '\P{^Is_CWL=true}', "");
Expect(0, 66600, '\p{Is_CWL=true}', "");
Expect(1, 66600, '\p{^Is_CWL=true}', "");
Expect(1, 66600, '\P{Is_CWL=true}', "");
Expect(0, 66600, '\P{^Is_CWL=true}', "");
Expect(1, 66599, '\p{Is_CWL= -TRUE}', "");
Expect(0, 66599, '\p{^Is_CWL= -TRUE}', "");
Expect(0, 66599, '\P{Is_CWL= -TRUE}', "");
Expect(1, 66599, '\P{^Is_CWL= -TRUE}', "");
Expect(0, 66600, '\p{Is_CWL= -TRUE}', "");
Expect(1, 66600, '\p{^Is_CWL= -TRUE}', "");
Expect(1, 66600, '\P{Is_CWL= -TRUE}', "");
Expect(0, 66600, '\P{^Is_CWL= -TRUE}', "");
Error('\p{Changes_When_Lowercased=:= no}');
Error('\P{Changes_When_Lowercased=:= no}');
Expect(1, 66600, '\p{Changes_When_Lowercased=no}', "");
Expect(0, 66600, '\p{^Changes_When_Lowercased=no}', "");
Expect(0, 66600, '\P{Changes_When_Lowercased=no}', "");
Expect(1, 66600, '\P{^Changes_When_Lowercased=no}', "");
Expect(0, 66599, '\p{Changes_When_Lowercased=no}', "");
Expect(1, 66599, '\p{^Changes_When_Lowercased=no}', "");
Expect(1, 66599, '\P{Changes_When_Lowercased=no}', "");
Expect(0, 66599, '\P{^Changes_When_Lowercased=no}', "");
Expect(1, 66600, '\p{Changes_When_Lowercased=_no}', "");
Expect(0, 66600, '\p{^Changes_When_Lowercased=_no}', "");
Expect(0, 66600, '\P{Changes_When_Lowercased=_no}', "");
Expect(1, 66600, '\P{^Changes_When_Lowercased=_no}', "");
Expect(0, 66599, '\p{Changes_When_Lowercased=_no}', "");
Expect(1, 66599, '\p{^Changes_When_Lowercased=_no}', "");
Expect(1, 66599, '\P{Changes_When_Lowercased=_no}', "");
Expect(0, 66599, '\P{^Changes_When_Lowercased=_no}', "");
Error('\p{CWL:	/a/n}');
Error('\P{CWL:	/a/n}');
Expect(1, 66600, '\p{CWL=n}', "");
Expect(0, 66600, '\p{^CWL=n}', "");
Expect(0, 66600, '\P{CWL=n}', "");
Expect(1, 66600, '\P{^CWL=n}', "");
Expect(0, 66599, '\p{CWL=n}', "");
Expect(1, 66599, '\p{^CWL=n}', "");
Expect(1, 66599, '\P{CWL=n}', "");
Expect(0, 66599, '\P{^CWL=n}', "");
Expect(1, 66600, '\p{CWL=	N}', "");
Expect(0, 66600, '\p{^CWL=	N}', "");
Expect(0, 66600, '\P{CWL=	N}', "");
Expect(1, 66600, '\P{^CWL=	N}', "");
Expect(0, 66599, '\p{CWL=	N}', "");
Expect(1, 66599, '\p{^CWL=	N}', "");
Expect(1, 66599, '\P{CWL=	N}', "");
Expect(0, 66599, '\P{^CWL=	N}', "");
Error('\p{Is_Changes_When_Lowercased=--F/a/}');
Error('\P{Is_Changes_When_Lowercased=--F/a/}');
Expect(1, 66600, '\p{Is_Changes_When_Lowercased:	f}', "");
Expect(0, 66600, '\p{^Is_Changes_When_Lowercased:	f}', "");
Expect(0, 66600, '\P{Is_Changes_When_Lowercased:	f}', "");
Expect(1, 66600, '\P{^Is_Changes_When_Lowercased:	f}', "");
Expect(0, 66599, '\p{Is_Changes_When_Lowercased:	f}', "");
Expect(1, 66599, '\p{^Is_Changes_When_Lowercased:	f}', "");
Expect(1, 66599, '\P{Is_Changes_When_Lowercased:	f}', "");
Expect(0, 66599, '\P{^Is_Changes_When_Lowercased:	f}', "");
Expect(1, 66600, '\p{Is_Changes_When_Lowercased= 	F}', "");
Expect(0, 66600, '\p{^Is_Changes_When_Lowercased= 	F}', "");
Expect(0, 66600, '\P{Is_Changes_When_Lowercased= 	F}', "");
Expect(1, 66600, '\P{^Is_Changes_When_Lowercased= 	F}', "");
Expect(0, 66599, '\p{Is_Changes_When_Lowercased= 	F}', "");
Expect(1, 66599, '\p{^Is_Changes_When_Lowercased= 	F}', "");
Expect(1, 66599, '\P{Is_Changes_When_Lowercased= 	F}', "");
Expect(0, 66599, '\P{^Is_Changes_When_Lowercased= 	F}', "");
Error('\p{Is_CWL=_-False:=}');
Error('\P{Is_CWL=_-False:=}');
Expect(1, 66600, '\p{Is_CWL=false}', "");
Expect(0, 66600, '\p{^Is_CWL=false}', "");
Expect(0, 66600, '\P{Is_CWL=false}', "");
Expect(1, 66600, '\P{^Is_CWL=false}', "");
Expect(0, 66599, '\p{Is_CWL=false}', "");
Expect(1, 66599, '\p{^Is_CWL=false}', "");
Expect(1, 66599, '\P{Is_CWL=false}', "");
Expect(0, 66599, '\P{^Is_CWL=false}', "");
Expect(1, 66600, '\p{Is_CWL= False}', "");
Expect(0, 66600, '\p{^Is_CWL= False}', "");
Expect(0, 66600, '\P{Is_CWL= False}', "");
Expect(1, 66600, '\P{^Is_CWL= False}', "");
Expect(0, 66599, '\p{Is_CWL= False}', "");
Expect(1, 66599, '\p{^Is_CWL= False}', "");
Expect(1, 66599, '\P{Is_CWL= False}', "");
Expect(0, 66599, '\P{^Is_CWL= False}', "");
Error('\p{Grapheme_Base= /a/Yes}');
Error('\P{Grapheme_Base= /a/Yes}');
Expect(1, 195101, '\p{Grapheme_Base=yes}', "");
Expect(0, 195101, '\p{^Grapheme_Base=yes}', "");
Expect(0, 195101, '\P{Grapheme_Base=yes}', "");
Expect(1, 195101, '\P{^Grapheme_Base=yes}', "");
Expect(0, 195102, '\p{Grapheme_Base=yes}', "");
Expect(1, 195102, '\p{^Grapheme_Base=yes}', "");
Expect(1, 195102, '\P{Grapheme_Base=yes}', "");
Expect(0, 195102, '\P{^Grapheme_Base=yes}', "");
Expect(1, 195101, '\p{Grapheme_Base=	Yes}', "");
Expect(0, 195101, '\p{^Grapheme_Base=	Yes}', "");
Expect(0, 195101, '\P{Grapheme_Base=	Yes}', "");
Expect(1, 195101, '\P{^Grapheme_Base=	Yes}', "");
Expect(0, 195102, '\p{Grapheme_Base=	Yes}', "");
Expect(1, 195102, '\p{^Grapheme_Base=	Yes}', "");
Expect(1, 195102, '\P{Grapheme_Base=	Yes}', "");
Expect(0, 195102, '\P{^Grapheme_Base=	Yes}', "");
Error('\p{Gr_Base=-_Y/a/}');
Error('\P{Gr_Base=-_Y/a/}');
Expect(1, 195101, '\p{Gr_Base=y}', "");
Expect(0, 195101, '\p{^Gr_Base=y}', "");
Expect(0, 195101, '\P{Gr_Base=y}', "");
Expect(1, 195101, '\P{^Gr_Base=y}', "");
Expect(0, 195102, '\p{Gr_Base=y}', "");
Expect(1, 195102, '\p{^Gr_Base=y}', "");
Expect(1, 195102, '\P{Gr_Base=y}', "");
Expect(0, 195102, '\P{^Gr_Base=y}', "");
Expect(1, 195101, '\p{Gr_Base:   	 y}', "");
Expect(0, 195101, '\p{^Gr_Base:   	 y}', "");
Expect(0, 195101, '\P{Gr_Base:   	 y}', "");
Expect(1, 195101, '\P{^Gr_Base:   	 y}', "");
Expect(0, 195102, '\p{Gr_Base:   	 y}', "");
Expect(1, 195102, '\p{^Gr_Base:   	 y}', "");
Expect(1, 195102, '\P{Gr_Base:   	 y}', "");
Expect(0, 195102, '\P{^Gr_Base:   	 y}', "");
Error('\p{Is_Grapheme_Base:   :=t}');
Error('\P{Is_Grapheme_Base:   :=t}');
Expect(1, 195101, '\p{Is_Grapheme_Base=t}', "");
Expect(0, 195101, '\p{^Is_Grapheme_Base=t}', "");
Expect(0, 195101, '\P{Is_Grapheme_Base=t}', "");
Expect(1, 195101, '\P{^Is_Grapheme_Base=t}', "");
Expect(0, 195102, '\p{Is_Grapheme_Base=t}', "");
Expect(1, 195102, '\p{^Is_Grapheme_Base=t}', "");
Expect(1, 195102, '\P{Is_Grapheme_Base=t}', "");
Expect(0, 195102, '\P{^Is_Grapheme_Base=t}', "");
Expect(1, 195101, '\p{Is_Grapheme_Base:   _t}', "");
Expect(0, 195101, '\p{^Is_Grapheme_Base:   _t}', "");
Expect(0, 195101, '\P{Is_Grapheme_Base:   _t}', "");
Expect(1, 195101, '\P{^Is_Grapheme_Base:   _t}', "");
Expect(0, 195102, '\p{Is_Grapheme_Base:   _t}', "");
Expect(1, 195102, '\p{^Is_Grapheme_Base:   _t}', "");
Expect(1, 195102, '\P{Is_Grapheme_Base:   _t}', "");
Expect(0, 195102, '\P{^Is_Grapheme_Base:   _t}', "");
Error('\p{Is_Gr_Base= _TRUE:=}');
Error('\P{Is_Gr_Base= _TRUE:=}');
Expect(1, 195101, '\p{Is_Gr_Base=true}', "");
Expect(0, 195101, '\p{^Is_Gr_Base=true}', "");
Expect(0, 195101, '\P{Is_Gr_Base=true}', "");
Expect(1, 195101, '\P{^Is_Gr_Base=true}', "");
Expect(0, 195102, '\p{Is_Gr_Base=true}', "");
Expect(1, 195102, '\p{^Is_Gr_Base=true}', "");
Expect(1, 195102, '\P{Is_Gr_Base=true}', "");
Expect(0, 195102, '\P{^Is_Gr_Base=true}', "");
Expect(1, 195101, '\p{Is_Gr_Base=-	True}', "");
Expect(0, 195101, '\p{^Is_Gr_Base=-	True}', "");
Expect(0, 195101, '\P{Is_Gr_Base=-	True}', "");
Expect(1, 195101, '\P{^Is_Gr_Base=-	True}', "");
Expect(0, 195102, '\p{Is_Gr_Base=-	True}', "");
Expect(1, 195102, '\p{^Is_Gr_Base=-	True}', "");
Expect(1, 195102, '\P{Is_Gr_Base=-	True}', "");
Expect(0, 195102, '\P{^Is_Gr_Base=-	True}', "");
Error('\p{Grapheme_Base=/a/_ No}');
Error('\P{Grapheme_Base=/a/_ No}');
Expect(1, 195102, '\p{Grapheme_Base=no}', "");
Expect(0, 195102, '\p{^Grapheme_Base=no}', "");
Expect(0, 195102, '\P{Grapheme_Base=no}', "");
Expect(1, 195102, '\P{^Grapheme_Base=no}', "");
Expect(0, 195101, '\p{Grapheme_Base=no}', "");
Expect(1, 195101, '\p{^Grapheme_Base=no}', "");
Expect(1, 195101, '\P{Grapheme_Base=no}', "");
Expect(0, 195101, '\P{^Grapheme_Base=no}', "");
Expect(1, 195102, '\p{Grapheme_Base= _No}', "");
Expect(0, 195102, '\p{^Grapheme_Base= _No}', "");
Expect(0, 195102, '\P{Grapheme_Base= _No}', "");
Expect(1, 195102, '\P{^Grapheme_Base= _No}', "");
Expect(0, 195101, '\p{Grapheme_Base= _No}', "");
Expect(1, 195101, '\p{^Grapheme_Base= _No}', "");
Expect(1, 195101, '\P{Grapheme_Base= _No}', "");
Expect(0, 195101, '\P{^Grapheme_Base= _No}', "");
Error('\p{Gr_Base=	/a/N}');
Error('\P{Gr_Base=	/a/N}');
Expect(1, 195102, '\p{Gr_Base=n}', "");
Expect(0, 195102, '\p{^Gr_Base=n}', "");
Expect(0, 195102, '\P{Gr_Base=n}', "");
Expect(1, 195102, '\P{^Gr_Base=n}', "");
Expect(0, 195101, '\p{Gr_Base=n}', "");
Expect(1, 195101, '\p{^Gr_Base=n}', "");
Expect(1, 195101, '\P{Gr_Base=n}', "");
Expect(0, 195101, '\P{^Gr_Base=n}', "");
Expect(1, 195102, '\p{Gr_Base=_-N}', "");
Expect(0, 195102, '\p{^Gr_Base=_-N}', "");
Expect(0, 195102, '\P{Gr_Base=_-N}', "");
Expect(1, 195102, '\P{^Gr_Base=_-N}', "");
Expect(0, 195101, '\p{Gr_Base=_-N}', "");
Expect(1, 195101, '\p{^Gr_Base=_-N}', "");
Expect(1, 195101, '\P{Gr_Base=_-N}', "");
Expect(0, 195101, '\P{^Gr_Base=_-N}', "");
Error('\p{Is_Grapheme_Base=	-F/a/}');
Error('\P{Is_Grapheme_Base=	-F/a/}');
Expect(1, 195102, '\p{Is_Grapheme_Base=f}', "");
Expect(0, 195102, '\p{^Is_Grapheme_Base=f}', "");
Expect(0, 195102, '\P{Is_Grapheme_Base=f}', "");
Expect(1, 195102, '\P{^Is_Grapheme_Base=f}', "");
Expect(0, 195101, '\p{Is_Grapheme_Base=f}', "");
Expect(1, 195101, '\p{^Is_Grapheme_Base=f}', "");
Expect(1, 195101, '\P{Is_Grapheme_Base=f}', "");
Expect(0, 195101, '\P{^Is_Grapheme_Base=f}', "");
Expect(1, 195102, '\p{Is_Grapheme_Base=	-F}', "");
Expect(0, 195102, '\p{^Is_Grapheme_Base=	-F}', "");
Expect(0, 195102, '\P{Is_Grapheme_Base=	-F}', "");
Expect(1, 195102, '\P{^Is_Grapheme_Base=	-F}', "");
Expect(0, 195101, '\p{Is_Grapheme_Base=	-F}', "");
Expect(1, 195101, '\p{^Is_Grapheme_Base=	-F}', "");
Expect(1, 195101, '\P{Is_Grapheme_Base=	-F}', "");
Expect(0, 195101, '\P{^Is_Grapheme_Base=	-F}', "");
Error('\p{Is_Gr_Base=/a/	false}');
Error('\P{Is_Gr_Base=/a/	false}');
Expect(1, 195102, '\p{Is_Gr_Base: false}', "");
Expect(0, 195102, '\p{^Is_Gr_Base: false}', "");
Expect(0, 195102, '\P{Is_Gr_Base: false}', "");
Expect(1, 195102, '\P{^Is_Gr_Base: false}', "");
Expect(0, 195101, '\p{Is_Gr_Base: false}', "");
Expect(1, 195101, '\p{^Is_Gr_Base: false}', "");
Expect(1, 195101, '\P{Is_Gr_Base: false}', "");
Expect(0, 195101, '\P{^Is_Gr_Base: false}', "");
Expect(1, 195102, '\p{Is_Gr_Base=_ FALSE}', "");
Expect(0, 195102, '\p{^Is_Gr_Base=_ FALSE}', "");
Expect(0, 195102, '\P{Is_Gr_Base=_ FALSE}', "");
Expect(1, 195102, '\P{^Is_Gr_Base=_ FALSE}', "");
Expect(0, 195101, '\p{Is_Gr_Base=_ FALSE}', "");
Expect(1, 195101, '\p{^Is_Gr_Base=_ FALSE}', "");
Expect(1, 195101, '\P{Is_Gr_Base=_ FALSE}', "");
Expect(0, 195101, '\P{^Is_Gr_Base=_ FALSE}', "");
Error('\p{Bidi_Class=		Segment_Separator/a/}');
Error('\P{Bidi_Class=		Segment_Separator/a/}');
Expect(1, 31, '\p{Bidi_Class=segmentseparator}', "");
Expect(0, 31, '\p{^Bidi_Class=segmentseparator}', "");
Expect(0, 31, '\P{Bidi_Class=segmentseparator}', "");
Expect(1, 31, '\P{^Bidi_Class=segmentseparator}', "");
Expect(0, 32, '\p{Bidi_Class=segmentseparator}', "");
Expect(1, 32, '\p{^Bidi_Class=segmentseparator}', "");
Expect(1, 32, '\P{Bidi_Class=segmentseparator}', "");
Expect(0, 32, '\P{^Bidi_Class=segmentseparator}', "");
Expect(1, 31, '\p{Bidi_Class=- segment_Separator}', "");
Expect(0, 31, '\p{^Bidi_Class=- segment_Separator}', "");
Expect(0, 31, '\P{Bidi_Class=- segment_Separator}', "");
Expect(1, 31, '\P{^Bidi_Class=- segment_Separator}', "");
Expect(0, 32, '\p{Bidi_Class=- segment_Separator}', "");
Expect(1, 32, '\p{^Bidi_Class=- segment_Separator}', "");
Expect(1, 32, '\P{Bidi_Class=- segment_Separator}', "");
Expect(0, 32, '\P{^Bidi_Class=- segment_Separator}', "");
Error('\p{Bc=/a/_S}');
Error('\P{Bc=/a/_S}');
Expect(1, 31, '\p{Bc=s}', "");
Expect(0, 31, '\p{^Bc=s}', "");
Expect(0, 31, '\P{Bc=s}', "");
Expect(1, 31, '\P{^Bc=s}', "");
Expect(0, 32, '\p{Bc=s}', "");
Expect(1, 32, '\p{^Bc=s}', "");
Expect(1, 32, '\P{Bc=s}', "");
Expect(0, 32, '\P{^Bc=s}', "");
Expect(1, 31, '\p{Bc=	S}', "");
Expect(0, 31, '\p{^Bc=	S}', "");
Expect(0, 31, '\P{Bc=	S}', "");
Expect(1, 31, '\P{^Bc=	S}', "");
Expect(0, 32, '\p{Bc=	S}', "");
Expect(1, 32, '\p{^Bc=	S}', "");
Expect(1, 32, '\P{Bc=	S}', "");
Expect(0, 32, '\P{^Bc=	S}', "");
Error('\p{Is_Bidi_Class=--Segment_Separator/a/}');
Error('\P{Is_Bidi_Class=--Segment_Separator/a/}');
Expect(1, 31, '\p{Is_Bidi_Class=segmentseparator}', "");
Expect(0, 31, '\p{^Is_Bidi_Class=segmentseparator}', "");
Expect(0, 31, '\P{Is_Bidi_Class=segmentseparator}', "");
Expect(1, 31, '\P{^Is_Bidi_Class=segmentseparator}', "");
Expect(0, 32, '\p{Is_Bidi_Class=segmentseparator}', "");
Expect(1, 32, '\p{^Is_Bidi_Class=segmentseparator}', "");
Expect(1, 32, '\P{Is_Bidi_Class=segmentseparator}', "");
Expect(0, 32, '\P{^Is_Bidi_Class=segmentseparator}', "");
Expect(1, 31, '\p{Is_Bidi_Class:- Segment_SEPARATOR}', "");
Expect(0, 31, '\p{^Is_Bidi_Class:- Segment_SEPARATOR}', "");
Expect(0, 31, '\P{Is_Bidi_Class:- Segment_SEPARATOR}', "");
Expect(1, 31, '\P{^Is_Bidi_Class:- Segment_SEPARATOR}', "");
Expect(0, 32, '\p{Is_Bidi_Class:- Segment_SEPARATOR}', "");
Expect(1, 32, '\p{^Is_Bidi_Class:- Segment_SEPARATOR}', "");
Expect(1, 32, '\P{Is_Bidi_Class:- Segment_SEPARATOR}', "");
Expect(0, 32, '\P{^Is_Bidi_Class:- Segment_SEPARATOR}', "");
Error('\p{Is_Bc:	:=	 S}');
Error('\P{Is_Bc:	:=	 S}');
Expect(1, 31, '\p{Is_Bc:   s}', "");
Expect(0, 31, '\p{^Is_Bc:   s}', "");
Expect(0, 31, '\P{Is_Bc:   s}', "");
Expect(1, 31, '\P{^Is_Bc:   s}', "");
Expect(0, 32, '\p{Is_Bc:   s}', "");
Expect(1, 32, '\p{^Is_Bc:   s}', "");
Expect(1, 32, '\P{Is_Bc:   s}', "");
Expect(0, 32, '\P{^Is_Bc:   s}', "");
Expect(1, 31, '\p{Is_Bc=_ S}', "");
Expect(0, 31, '\p{^Is_Bc=_ S}', "");
Expect(0, 31, '\P{Is_Bc=_ S}', "");
Expect(1, 31, '\P{^Is_Bc=_ S}', "");
Expect(0, 32, '\p{Is_Bc=_ S}', "");
Expect(1, 32, '\p{^Is_Bc=_ S}', "");
Expect(1, 32, '\P{Is_Bc=_ S}', "");
Expect(0, 32, '\P{^Is_Bc=_ S}', "");
Error('\p{Bidi_Class=:=	right_to_LEFT_override}');
Error('\P{Bidi_Class=:=	right_to_LEFT_override}');
Expect(1, 8238, '\p{Bidi_Class=righttoleftoverride}', "");
Expect(0, 8238, '\p{^Bidi_Class=righttoleftoverride}', "");
Expect(0, 8238, '\P{Bidi_Class=righttoleftoverride}', "");
Expect(1, 8238, '\P{^Bidi_Class=righttoleftoverride}', "");
Expect(0, 8239, '\p{Bidi_Class=righttoleftoverride}', "");
Expect(1, 8239, '\p{^Bidi_Class=righttoleftoverride}', "");
Expect(1, 8239, '\P{Bidi_Class=righttoleftoverride}', "");
Expect(0, 8239, '\P{^Bidi_Class=righttoleftoverride}', "");
Expect(1, 8238, '\p{Bidi_Class=_ Right_To_Left_Override}', "");
Expect(0, 8238, '\p{^Bidi_Class=_ Right_To_Left_Override}', "");
Expect(0, 8238, '\P{Bidi_Class=_ Right_To_Left_Override}', "");
Expect(1, 8238, '\P{^Bidi_Class=_ Right_To_Left_Override}', "");
Expect(0, 8239, '\p{Bidi_Class=_ Right_To_Left_Override}', "");
Expect(1, 8239, '\p{^Bidi_Class=_ Right_To_Left_Override}', "");
Expect(1, 8239, '\P{Bidi_Class=_ Right_To_Left_Override}', "");
Expect(0, 8239, '\P{^Bidi_Class=_ Right_To_Left_Override}', "");
Error('\p{Bc:/a/RLO}');
Error('\P{Bc:/a/RLO}');
Expect(1, 8238, '\p{Bc: rlo}', "");
Expect(0, 8238, '\p{^Bc: rlo}', "");
Expect(0, 8238, '\P{Bc: rlo}', "");
Expect(1, 8238, '\P{^Bc: rlo}', "");
Expect(0, 8239, '\p{Bc: rlo}', "");
Expect(1, 8239, '\p{^Bc: rlo}', "");
Expect(1, 8239, '\P{Bc: rlo}', "");
Expect(0, 8239, '\P{^Bc: rlo}', "");
Expect(1, 8238, '\p{Bc: _ RLO}', "");
Expect(0, 8238, '\p{^Bc: _ RLO}', "");
Expect(0, 8238, '\P{Bc: _ RLO}', "");
Expect(1, 8238, '\P{^Bc: _ RLO}', "");
Expect(0, 8239, '\p{Bc: _ RLO}', "");
Expect(1, 8239, '\p{^Bc: _ RLO}', "");
Expect(1, 8239, '\P{Bc: _ RLO}', "");
Expect(0, 8239, '\P{^Bc: _ RLO}', "");
Error('\p{Is_Bidi_Class=_-Right_to_LEFT_override:=}');
Error('\P{Is_Bidi_Class=_-Right_to_LEFT_override:=}');
Expect(1, 8238, '\p{Is_Bidi_Class=righttoleftoverride}', "");
Expect(0, 8238, '\p{^Is_Bidi_Class=righttoleftoverride}', "");
Expect(0, 8238, '\P{Is_Bidi_Class=righttoleftoverride}', "");
Expect(1, 8238, '\P{^Is_Bidi_Class=righttoleftoverride}', "");
Expect(0, 8239, '\p{Is_Bidi_Class=righttoleftoverride}', "");
Expect(1, 8239, '\p{^Is_Bidi_Class=righttoleftoverride}', "");
Expect(1, 8239, '\P{Is_Bidi_Class=righttoleftoverride}', "");
Expect(0, 8239, '\P{^Is_Bidi_Class=righttoleftoverride}', "");
Expect(1, 8238, '\p{Is_Bidi_Class=-Right_To_LEFT_Override}', "");
Expect(0, 8238, '\p{^Is_Bidi_Class=-Right_To_LEFT_Override}', "");
Expect(0, 8238, '\P{Is_Bidi_Class=-Right_To_LEFT_Override}', "");
Expect(1, 8238, '\P{^Is_Bidi_Class=-Right_To_LEFT_Override}', "");
Expect(0, 8239, '\p{Is_Bidi_Class=-Right_To_LEFT_Override}', "");
Expect(1, 8239, '\p{^Is_Bidi_Class=-Right_To_LEFT_Override}', "");
Expect(1, 8239, '\P{Is_Bidi_Class=-Right_To_LEFT_Override}', "");
Expect(0, 8239, '\P{^Is_Bidi_Class=-Right_To_LEFT_Override}', "");
Error('\p{Is_Bc=_:=RLO}');
Error('\P{Is_Bc=_:=RLO}');
Expect(1, 8238, '\p{Is_Bc=rlo}', "");
Expect(0, 8238, '\p{^Is_Bc=rlo}', "");
Expect(0, 8238, '\P{Is_Bc=rlo}', "");
Expect(1, 8238, '\P{^Is_Bc=rlo}', "");
Expect(0, 8239, '\p{Is_Bc=rlo}', "");
Expect(1, 8239, '\p{^Is_Bc=rlo}', "");
Expect(1, 8239, '\P{Is_Bc=rlo}', "");
Expect(0, 8239, '\P{^Is_Bc=rlo}', "");
Expect(1, 8238, '\p{Is_Bc=	rlo}', "");
Expect(0, 8238, '\p{^Is_Bc=	rlo}', "");
Expect(0, 8238, '\P{Is_Bc=	rlo}', "");
Expect(1, 8238, '\P{^Is_Bc=	rlo}', "");
Expect(0, 8239, '\p{Is_Bc=	rlo}', "");
Expect(1, 8239, '\p{^Is_Bc=	rlo}', "");
Expect(1, 8239, '\P{Is_Bc=	rlo}', "");
Expect(0, 8239, '\P{^Is_Bc=	rlo}', "");
Error('\p{Bidi_Class=	/a/left_to_Right_Override}');
Error('\P{Bidi_Class=	/a/left_to_Right_Override}');
Expect(1, 8237, '\p{Bidi_Class=lefttorightoverride}', "");
Expect(0, 8237, '\p{^Bidi_Class=lefttorightoverride}', "");
Expect(0, 8237, '\P{Bidi_Class=lefttorightoverride}', "");
Expect(1, 8237, '\P{^Bidi_Class=lefttorightoverride}', "");
Expect(0, 8238, '\p{Bidi_Class=lefttorightoverride}', "");
Expect(1, 8238, '\p{^Bidi_Class=lefttorightoverride}', "");
Expect(1, 8238, '\P{Bidi_Class=lefttorightoverride}', "");
Expect(0, 8238, '\P{^Bidi_Class=lefttorightoverride}', "");
Expect(1, 8237, '\p{Bidi_Class=_left_TO_right_Override}', "");
Expect(0, 8237, '\p{^Bidi_Class=_left_TO_right_Override}', "");
Expect(0, 8237, '\P{Bidi_Class=_left_TO_right_Override}', "");
Expect(1, 8237, '\P{^Bidi_Class=_left_TO_right_Override}', "");
Expect(0, 8238, '\p{Bidi_Class=_left_TO_right_Override}', "");
Expect(1, 8238, '\p{^Bidi_Class=_left_TO_right_Override}', "");
Expect(1, 8238, '\P{Bidi_Class=_left_TO_right_Override}', "");
Expect(0, 8238, '\P{^Bidi_Class=_left_TO_right_Override}', "");
Error('\p{Bc=		LRO/a/}');
Error('\P{Bc=		LRO/a/}');
Expect(1, 8237, '\p{Bc=lro}', "");
Expect(0, 8237, '\p{^Bc=lro}', "");
Expect(0, 8237, '\P{Bc=lro}', "");
Expect(1, 8237, '\P{^Bc=lro}', "");
Expect(0, 8238, '\p{Bc=lro}', "");
Expect(1, 8238, '\p{^Bc=lro}', "");
Expect(1, 8238, '\P{Bc=lro}', "");
Expect(0, 8238, '\P{^Bc=lro}', "");
Expect(1, 8237, '\p{Bc= _LRO}', "");
Expect(0, 8237, '\p{^Bc= _LRO}', "");
Expect(0, 8237, '\P{Bc= _LRO}', "");
Expect(1, 8237, '\P{^Bc= _LRO}', "");
Expect(0, 8238, '\p{Bc= _LRO}', "");
Expect(1, 8238, '\p{^Bc= _LRO}', "");
Expect(1, 8238, '\P{Bc= _LRO}', "");
Expect(0, 8238, '\P{^Bc= _LRO}', "");
Error('\p{Is_Bidi_Class=	/a/Left_To_Right_OVERRIDE}');
Error('\P{Is_Bidi_Class=	/a/Left_To_Right_OVERRIDE}');
Expect(1, 8237, '\p{Is_Bidi_Class=lefttorightoverride}', "");
Expect(0, 8237, '\p{^Is_Bidi_Class=lefttorightoverride}', "");
Expect(0, 8237, '\P{Is_Bidi_Class=lefttorightoverride}', "");
Expect(1, 8237, '\P{^Is_Bidi_Class=lefttorightoverride}', "");
Expect(0, 8238, '\p{Is_Bidi_Class=lefttorightoverride}', "");
Expect(1, 8238, '\p{^Is_Bidi_Class=lefttorightoverride}', "");
Expect(1, 8238, '\P{Is_Bidi_Class=lefttorightoverride}', "");
Expect(0, 8238, '\P{^Is_Bidi_Class=lefttorightoverride}', "");
Expect(1, 8237, '\p{Is_Bidi_Class=		LEFT_to_Right_Override}', "");
Expect(0, 8237, '\p{^Is_Bidi_Class=		LEFT_to_Right_Override}', "");
Expect(0, 8237, '\P{Is_Bidi_Class=		LEFT_to_Right_Override}', "");
Expect(1, 8237, '\P{^Is_Bidi_Class=		LEFT_to_Right_Override}', "");
Expect(0, 8238, '\p{Is_Bidi_Class=		LEFT_to_Right_Override}', "");
Expect(1, 8238, '\p{^Is_Bidi_Class=		LEFT_to_Right_Override}', "");
Expect(1, 8238, '\P{Is_Bidi_Class=		LEFT_to_Right_Override}', "");
Expect(0, 8238, '\P{^Is_Bidi_Class=		LEFT_to_Right_Override}', "");
Error('\p{Is_Bc=:=_ LRO}');
Error('\P{Is_Bc=:=_ LRO}');
Expect(1, 8237, '\p{Is_Bc=lro}', "");
Expect(0, 8237, '\p{^Is_Bc=lro}', "");
Expect(0, 8237, '\P{Is_Bc=lro}', "");
Expect(1, 8237, '\P{^Is_Bc=lro}', "");
Expect(0, 8238, '\p{Is_Bc=lro}', "");
Expect(1, 8238, '\p{^Is_Bc=lro}', "");
Expect(1, 8238, '\P{Is_Bc=lro}', "");
Expect(0, 8238, '\P{^Is_Bc=lro}', "");
Expect(1, 8237, '\p{Is_Bc=-_LRO}', "");
Expect(0, 8237, '\p{^Is_Bc=-_LRO}', "");
Expect(0, 8237, '\P{Is_Bc=-_LRO}', "");
Expect(1, 8237, '\P{^Is_Bc=-_LRO}', "");
Expect(0, 8238, '\p{Is_Bc=-_LRO}', "");
Expect(1, 8238, '\p{^Is_Bc=-_LRO}', "");
Expect(1, 8238, '\P{Is_Bc=-_LRO}', "");
Expect(0, 8238, '\P{^Is_Bc=-_LRO}', "");
Error('\p{Bidi_Class= /a/LEFT_to_RIGHT}');
Error('\P{Bidi_Class= /a/LEFT_to_RIGHT}');
Expect(1, 1114109, '\p{Bidi_Class=lefttoright}', "");
Expect(0, 1114109, '\p{^Bidi_Class=lefttoright}', "");
Expect(0, 1114109, '\P{Bidi_Class=lefttoright}', "");
Expect(1, 1114109, '\P{^Bidi_Class=lefttoright}', "");
Expect(0, 921599, '\p{Bidi_Class=lefttoright}', "");
Expect(1, 921599, '\p{^Bidi_Class=lefttoright}', "");
Expect(1, 921599, '\P{Bidi_Class=lefttoright}', "");
Expect(0, 921599, '\P{^Bidi_Class=lefttoright}', "");
Expect(1, 1114109, '\p{Bidi_Class=_LEFT_to_Right}', "");
Expect(0, 1114109, '\p{^Bidi_Class=_LEFT_to_Right}', "");
Expect(0, 1114109, '\P{Bidi_Class=_LEFT_to_Right}', "");
Expect(1, 1114109, '\P{^Bidi_Class=_LEFT_to_Right}', "");
Expect(0, 921599, '\p{Bidi_Class=_LEFT_to_Right}', "");
Expect(1, 921599, '\p{^Bidi_Class=_LEFT_to_Right}', "");
Expect(1, 921599, '\P{Bidi_Class=_LEFT_to_Right}', "");
Expect(0, 921599, '\P{^Bidi_Class=_LEFT_to_Right}', "");
Error('\p{Bc=:=  L}');
Error('\P{Bc=:=  L}');
Expect(1, 1114109, '\p{Bc=l}', "");
Expect(0, 1114109, '\p{^Bc=l}', "");
Expect(0, 1114109, '\P{Bc=l}', "");
Expect(1, 1114109, '\P{^Bc=l}', "");
Expect(0, 921599, '\p{Bc=l}', "");
Expect(1, 921599, '\p{^Bc=l}', "");
Expect(1, 921599, '\P{Bc=l}', "");
Expect(0, 921599, '\P{^Bc=l}', "");
Expect(1, 1114109, '\p{Bc=-_L}', "");
Expect(0, 1114109, '\p{^Bc=-_L}', "");
Expect(0, 1114109, '\P{Bc=-_L}', "");
Expect(1, 1114109, '\P{^Bc=-_L}', "");
Expect(0, 921599, '\p{Bc=-_L}', "");
Expect(1, 921599, '\p{^Bc=-_L}', "");
Expect(1, 921599, '\P{Bc=-_L}', "");
Expect(0, 921599, '\P{^Bc=-_L}', "");
Error('\p{Is_Bidi_Class=:=	left_TO_right}');
Error('\P{Is_Bidi_Class=:=	left_TO_right}');
Expect(1, 1114109, '\p{Is_Bidi_Class=lefttoright}', "");
Expect(0, 1114109, '\p{^Is_Bidi_Class=lefttoright}', "");
Expect(0, 1114109, '\P{Is_Bidi_Class=lefttoright}', "");
Expect(1, 1114109, '\P{^Is_Bidi_Class=lefttoright}', "");
Expect(0, 921599, '\p{Is_Bidi_Class=lefttoright}', "");
Expect(1, 921599, '\p{^Is_Bidi_Class=lefttoright}', "");
Expect(1, 921599, '\P{Is_Bidi_Class=lefttoright}', "");
Expect(0, 921599, '\P{^Is_Bidi_Class=lefttoright}', "");
Expect(1, 1114109, '\p{Is_Bidi_Class=	Left_TO_Right}', "");
Expect(0, 1114109, '\p{^Is_Bidi_Class=	Left_TO_Right}', "");
Expect(0, 1114109, '\P{Is_Bidi_Class=	Left_TO_Right}', "");
Expect(1, 1114109, '\P{^Is_Bidi_Class=	Left_TO_Right}', "");
Expect(0, 921599, '\p{Is_Bidi_Class=	Left_TO_Right}', "");
Expect(1, 921599, '\p{^Is_Bidi_Class=	Left_TO_Right}', "");
Expect(1, 921599, '\P{Is_Bidi_Class=	Left_TO_Right}', "");
Expect(0, 921599, '\P{^Is_Bidi_Class=	Left_TO_Right}', "");
Error('\p{Is_Bc=-:=L}');
Error('\P{Is_Bc=-:=L}');
Expect(1, 1114109, '\p{Is_Bc=l}', "");
Expect(0, 1114109, '\p{^Is_Bc=l}', "");
Expect(0, 1114109, '\P{Is_Bc=l}', "");
Expect(1, 1114109, '\P{^Is_Bc=l}', "");
Expect(0, 921599, '\p{Is_Bc=l}', "");
Expect(1, 921599, '\p{^Is_Bc=l}', "");
Expect(1, 921599, '\P{Is_Bc=l}', "");
Expect(0, 921599, '\P{^Is_Bc=l}', "");
Expect(1, 1114109, '\p{Is_Bc= 	L}', "");
Expect(0, 1114109, '\p{^Is_Bc= 	L}', "");
Expect(0, 1114109, '\P{Is_Bc= 	L}', "");
Expect(1, 1114109, '\P{^Is_Bc= 	L}', "");
Expect(0, 921599, '\p{Is_Bc= 	L}', "");
Expect(1, 921599, '\p{^Is_Bc= 	L}', "");
Expect(1, 921599, '\P{Is_Bc= 	L}', "");
Expect(0, 921599, '\P{^Is_Bc= 	L}', "");
Error('\p{Bidi_Class:   -right_TO_LEFT_embedding:=}');
Error('\P{Bidi_Class:   -right_TO_LEFT_embedding:=}');
Expect(1, 8235, '\p{Bidi_Class=righttoleftembedding}', "");
Expect(0, 8235, '\p{^Bidi_Class=righttoleftembedding}', "");
Expect(0, 8235, '\P{Bidi_Class=righttoleftembedding}', "");
Expect(1, 8235, '\P{^Bidi_Class=righttoleftembedding}', "");
Expect(0, 8236, '\p{Bidi_Class=righttoleftembedding}', "");
Expect(1, 8236, '\p{^Bidi_Class=righttoleftembedding}', "");
Expect(1, 8236, '\P{Bidi_Class=righttoleftembedding}', "");
Expect(0, 8236, '\P{^Bidi_Class=righttoleftembedding}', "");
Expect(1, 8235, '\p{Bidi_Class=	Right_To_LEFT_Embedding}', "");
Expect(0, 8235, '\p{^Bidi_Class=	Right_To_LEFT_Embedding}', "");
Expect(0, 8235, '\P{Bidi_Class=	Right_To_LEFT_Embedding}', "");
Expect(1, 8235, '\P{^Bidi_Class=	Right_To_LEFT_Embedding}', "");
Expect(0, 8236, '\p{Bidi_Class=	Right_To_LEFT_Embedding}', "");
Expect(1, 8236, '\p{^Bidi_Class=	Right_To_LEFT_Embedding}', "");
Expect(1, 8236, '\P{Bidi_Class=	Right_To_LEFT_Embedding}', "");
Expect(0, 8236, '\P{^Bidi_Class=	Right_To_LEFT_Embedding}', "");
Error('\p{Bc= :=RLE}');
Error('\P{Bc= :=RLE}');
Expect(1, 8235, '\p{Bc=rle}', "");
Expect(0, 8235, '\p{^Bc=rle}', "");
Expect(0, 8235, '\P{Bc=rle}', "");
Expect(1, 8235, '\P{^Bc=rle}', "");
Expect(0, 8236, '\p{Bc=rle}', "");
Expect(1, 8236, '\p{^Bc=rle}', "");
Expect(1, 8236, '\P{Bc=rle}', "");
Expect(0, 8236, '\P{^Bc=rle}', "");
Expect(1, 8235, '\p{Bc= 	RLE}', "");
Expect(0, 8235, '\p{^Bc= 	RLE}', "");
Expect(0, 8235, '\P{Bc= 	RLE}', "");
Expect(1, 8235, '\P{^Bc= 	RLE}', "");
Expect(0, 8236, '\p{Bc= 	RLE}', "");
Expect(1, 8236, '\p{^Bc= 	RLE}', "");
Expect(1, 8236, '\P{Bc= 	RLE}', "");
Expect(0, 8236, '\P{^Bc= 	RLE}', "");
Error('\p{Is_Bidi_Class=	:=Right_To_Left_Embedding}');
Error('\P{Is_Bidi_Class=	:=Right_To_Left_Embedding}');
Expect(1, 8235, '\p{Is_Bidi_Class=righttoleftembedding}', "");
Expect(0, 8235, '\p{^Is_Bidi_Class=righttoleftembedding}', "");
Expect(0, 8235, '\P{Is_Bidi_Class=righttoleftembedding}', "");
Expect(1, 8235, '\P{^Is_Bidi_Class=righttoleftembedding}', "");
Expect(0, 8236, '\p{Is_Bidi_Class=righttoleftembedding}', "");
Expect(1, 8236, '\p{^Is_Bidi_Class=righttoleftembedding}', "");
Expect(1, 8236, '\P{Is_Bidi_Class=righttoleftembedding}', "");
Expect(0, 8236, '\P{^Is_Bidi_Class=righttoleftembedding}', "");
Expect(1, 8235, '\p{Is_Bidi_Class=-	RIGHT_TO_LEFT_Embedding}', "");
Expect(0, 8235, '\p{^Is_Bidi_Class=-	RIGHT_TO_LEFT_Embedding}', "");
Expect(0, 8235, '\P{Is_Bidi_Class=-	RIGHT_TO_LEFT_Embedding}', "");
Expect(1, 8235, '\P{^Is_Bidi_Class=-	RIGHT_TO_LEFT_Embedding}', "");
Expect(0, 8236, '\p{Is_Bidi_Class=-	RIGHT_TO_LEFT_Embedding}', "");
Expect(1, 8236, '\p{^Is_Bidi_Class=-	RIGHT_TO_LEFT_Embedding}', "");
Expect(1, 8236, '\P{Is_Bidi_Class=-	RIGHT_TO_LEFT_Embedding}', "");
Expect(0, 8236, '\P{^Is_Bidi_Class=-	RIGHT_TO_LEFT_Embedding}', "");
Error('\p{Is_Bc=:=_	RLE}');
Error('\P{Is_Bc=:=_	RLE}');
Expect(1, 8235, '\p{Is_Bc=rle}', "");
Expect(0, 8235, '\p{^Is_Bc=rle}', "");
Expect(0, 8235, '\P{Is_Bc=rle}', "");
Expect(1, 8235, '\P{^Is_Bc=rle}', "");
Expect(0, 8236, '\p{Is_Bc=rle}', "");
Expect(1, 8236, '\p{^Is_Bc=rle}', "");
Expect(1, 8236, '\P{Is_Bc=rle}', "");
Expect(0, 8236, '\P{^Is_Bc=rle}', "");
Expect(1, 8235, '\p{Is_Bc=_-RLE}', "");
Expect(0, 8235, '\p{^Is_Bc=_-RLE}', "");
Expect(0, 8235, '\P{Is_Bc=_-RLE}', "");
Expect(1, 8235, '\P{^Is_Bc=_-RLE}', "");
Expect(0, 8236, '\p{Is_Bc=_-RLE}', "");
Expect(1, 8236, '\p{^Is_Bc=_-RLE}', "");
Expect(1, 8236, '\P{Is_Bc=_-RLE}', "");
Expect(0, 8236, '\P{^Is_Bc=_-RLE}', "");
Error('\p{Bidi_Class= -other_neutral/a/}');
Error('\P{Bidi_Class= -other_neutral/a/}');
Expect(1, 128883, '\p{Bidi_Class:otherneutral}', "");
Expect(0, 128883, '\p{^Bidi_Class:otherneutral}', "");
Expect(0, 128883, '\P{Bidi_Class:otherneutral}', "");
Expect(1, 128883, '\P{^Bidi_Class:otherneutral}', "");
Expect(0, 128884, '\p{Bidi_Class:otherneutral}', "");
Expect(1, 128884, '\p{^Bidi_Class:otherneutral}', "");
Expect(1, 128884, '\P{Bidi_Class:otherneutral}', "");
Expect(0, 128884, '\P{^Bidi_Class:otherneutral}', "");
Expect(1, 128883, '\p{Bidi_Class=	Other_Neutral}', "");
Expect(0, 128883, '\p{^Bidi_Class=	Other_Neutral}', "");
Expect(0, 128883, '\P{Bidi_Class=	Other_Neutral}', "");
Expect(1, 128883, '\P{^Bidi_Class=	Other_Neutral}', "");
Expect(0, 128884, '\p{Bidi_Class=	Other_Neutral}', "");
Expect(1, 128884, '\p{^Bidi_Class=	Other_Neutral}', "");
Expect(1, 128884, '\P{Bidi_Class=	Other_Neutral}', "");
Expect(0, 128884, '\P{^Bidi_Class=	Other_Neutral}', "");
Error('\p{Bc=/a/ ON}');
Error('\P{Bc=/a/ ON}');
Expect(1, 128883, '\p{Bc=on}', "");
Expect(0, 128883, '\p{^Bc=on}', "");
Expect(0, 128883, '\P{Bc=on}', "");
Expect(1, 128883, '\P{^Bc=on}', "");
Expect(0, 128884, '\p{Bc=on}', "");
Expect(1, 128884, '\p{^Bc=on}', "");
Expect(1, 128884, '\P{Bc=on}', "");
Expect(0, 128884, '\P{^Bc=on}', "");
Expect(1, 128883, '\p{Bc= ON}', "");
Expect(0, 128883, '\p{^Bc= ON}', "");
Expect(0, 128883, '\P{Bc= ON}', "");
Expect(1, 128883, '\P{^Bc= ON}', "");
Expect(0, 128884, '\p{Bc= ON}', "");
Expect(1, 128884, '\p{^Bc= ON}', "");
Expect(1, 128884, '\P{Bc= ON}', "");
Expect(0, 128884, '\P{^Bc= ON}', "");
Error('\p{Is_Bidi_Class=	other_NEUTRAL:=}');
Error('\P{Is_Bidi_Class=	other_NEUTRAL:=}');
Expect(1, 128883, '\p{Is_Bidi_Class=otherneutral}', "");
Expect(0, 128883, '\p{^Is_Bidi_Class=otherneutral}', "");
Expect(0, 128883, '\P{Is_Bidi_Class=otherneutral}', "");
Expect(1, 128883, '\P{^Is_Bidi_Class=otherneutral}', "");
Expect(0, 128884, '\p{Is_Bidi_Class=otherneutral}', "");
Expect(1, 128884, '\p{^Is_Bidi_Class=otherneutral}', "");
Expect(1, 128884, '\P{Is_Bidi_Class=otherneutral}', "");
Expect(0, 128884, '\P{^Is_Bidi_Class=otherneutral}', "");
Expect(1, 128883, '\p{Is_Bidi_Class=- OTHER_neutral}', "");
Expect(0, 128883, '\p{^Is_Bidi_Class=- OTHER_neutral}', "");
Expect(0, 128883, '\P{Is_Bidi_Class=- OTHER_neutral}', "");
Expect(1, 128883, '\P{^Is_Bidi_Class=- OTHER_neutral}', "");
Expect(0, 128884, '\p{Is_Bidi_Class=- OTHER_neutral}', "");
Expect(1, 128884, '\p{^Is_Bidi_Class=- OTHER_neutral}', "");
Expect(1, 128884, '\P{Is_Bidi_Class=- OTHER_neutral}', "");
Expect(0, 128884, '\P{^Is_Bidi_Class=- OTHER_neutral}', "");
Error('\p{Is_Bc= /a/ON}');
Error('\P{Is_Bc= /a/ON}');
Expect(1, 128883, '\p{Is_Bc=on}', "");
Expect(0, 128883, '\p{^Is_Bc=on}', "");
Expect(0, 128883, '\P{Is_Bc=on}', "");
Expect(1, 128883, '\P{^Is_Bc=on}', "");
Expect(0, 128884, '\p{Is_Bc=on}', "");
Expect(1, 128884, '\p{^Is_Bc=on}', "");
Expect(1, 128884, '\P{Is_Bc=on}', "");
Expect(0, 128884, '\P{^Is_Bc=on}', "");
Expect(1, 128883, '\p{Is_Bc= ON}', "");
Expect(0, 128883, '\p{^Is_Bc= ON}', "");
Expect(0, 128883, '\P{Is_Bc= ON}', "");
Expect(1, 128883, '\P{^Is_Bc= ON}', "");
Expect(0, 128884, '\p{Is_Bc= ON}', "");
Expect(1, 128884, '\p{^Is_Bc= ON}', "");
Expect(1, 128884, '\P{Is_Bc= ON}', "");
Expect(0, 128884, '\P{^Is_Bc= ON}', "");
Error('\p{Bidi_Class= _Boundary_Neutral/a/}');
Error('\P{Bidi_Class= _Boundary_Neutral/a/}');
Expect(1, 921599, '\p{Bidi_Class=boundaryneutral}', "");
Expect(0, 921599, '\p{^Bidi_Class=boundaryneutral}', "");
Expect(0, 921599, '\P{Bidi_Class=boundaryneutral}', "");
Expect(1, 921599, '\P{^Bidi_Class=boundaryneutral}', "");
Expect(0, 1114109, '\p{Bidi_Class=boundaryneutral}', "");
Expect(1, 1114109, '\p{^Bidi_Class=boundaryneutral}', "");
Expect(1, 1114109, '\P{Bidi_Class=boundaryneutral}', "");
Expect(0, 1114109, '\P{^Bidi_Class=boundaryneutral}', "");
Expect(1, 921599, '\p{Bidi_Class= boundary_Neutral}', "");
Expect(0, 921599, '\p{^Bidi_Class= boundary_Neutral}', "");
Expect(0, 921599, '\P{Bidi_Class= boundary_Neutral}', "");
Expect(1, 921599, '\P{^Bidi_Class= boundary_Neutral}', "");
Expect(0, 1114109, '\p{Bidi_Class= boundary_Neutral}', "");
Expect(1, 1114109, '\p{^Bidi_Class= boundary_Neutral}', "");
Expect(1, 1114109, '\P{Bidi_Class= boundary_Neutral}', "");
Expect(0, 1114109, '\P{^Bidi_Class= boundary_Neutral}', "");
Error('\p{Bc=	 BN/a/}');
Error('\P{Bc=	 BN/a/}');
Expect(1, 921599, '\p{Bc=bn}', "");
Expect(0, 921599, '\p{^Bc=bn}', "");
Expect(0, 921599, '\P{Bc=bn}', "");
Expect(1, 921599, '\P{^Bc=bn}', "");
Expect(0, 1114109, '\p{Bc=bn}', "");
Expect(1, 1114109, '\p{^Bc=bn}', "");
Expect(1, 1114109, '\P{Bc=bn}', "");
Expect(0, 1114109, '\P{^Bc=bn}', "");
Expect(1, 921599, '\p{Bc=_BN}', "");
Expect(0, 921599, '\p{^Bc=_BN}', "");
Expect(0, 921599, '\P{Bc=_BN}', "");
Expect(1, 921599, '\P{^Bc=_BN}', "");
Expect(0, 1114109, '\p{Bc=_BN}', "");
Expect(1, 1114109, '\p{^Bc=_BN}', "");
Expect(1, 1114109, '\P{Bc=_BN}', "");
Expect(0, 1114109, '\P{^Bc=_BN}', "");
Error('\p{Is_Bidi_Class=/a/	boundary_NEUTRAL}');
Error('\P{Is_Bidi_Class=/a/	boundary_NEUTRAL}');
Expect(1, 921599, '\p{Is_Bidi_Class=boundaryneutral}', "");
Expect(0, 921599, '\p{^Is_Bidi_Class=boundaryneutral}', "");
Expect(0, 921599, '\P{Is_Bidi_Class=boundaryneutral}', "");
Expect(1, 921599, '\P{^Is_Bidi_Class=boundaryneutral}', "");
Expect(0, 1114109, '\p{Is_Bidi_Class=boundaryneutral}', "");
Expect(1, 1114109, '\p{^Is_Bidi_Class=boundaryneutral}', "");
Expect(1, 1114109, '\P{Is_Bidi_Class=boundaryneutral}', "");
Expect(0, 1114109, '\P{^Is_Bidi_Class=boundaryneutral}', "");
Expect(1, 921599, '\p{Is_Bidi_Class=Boundary_Neutral}', "");
Expect(0, 921599, '\p{^Is_Bidi_Class=Boundary_Neutral}', "");
Expect(0, 921599, '\P{Is_Bidi_Class=Boundary_Neutral}', "");
Expect(1, 921599, '\P{^Is_Bidi_Class=Boundary_Neutral}', "");
Expect(0, 1114109, '\p{Is_Bidi_Class=Boundary_Neutral}', "");
Expect(1, 1114109, '\p{^Is_Bidi_Class=Boundary_Neutral}', "");
Expect(1, 1114109, '\P{Is_Bidi_Class=Boundary_Neutral}', "");
Expect(0, 1114109, '\P{^Is_Bidi_Class=Boundary_Neutral}', "");
Error('\p{Is_Bc=-	BN:=}');
Error('\P{Is_Bc=-	BN:=}');
Expect(1, 921599, '\p{Is_Bc=bn}', "");
Expect(0, 921599, '\p{^Is_Bc=bn}', "");
Expect(0, 921599, '\P{Is_Bc=bn}', "");
Expect(1, 921599, '\P{^Is_Bc=bn}', "");
Expect(0, 1114109, '\p{Is_Bc=bn}', "");
Expect(1, 1114109, '\p{^Is_Bc=bn}', "");
Expect(1, 1114109, '\P{Is_Bc=bn}', "");
Expect(0, 1114109, '\P{^Is_Bc=bn}', "");
Expect(1, 921599, '\p{Is_Bc:	_BN}', "");
Expect(0, 921599, '\p{^Is_Bc:	_BN}', "");
Expect(0, 921599, '\P{Is_Bc:	_BN}', "");
Expect(1, 921599, '\P{^Is_Bc:	_BN}', "");
Expect(0, 1114109, '\p{Is_Bc:	_BN}', "");
Expect(1, 1114109, '\p{^Is_Bc:	_BN}', "");
Expect(1, 1114109, '\P{Is_Bc:	_BN}', "");
Expect(0, 1114109, '\P{^Is_Bc:	_BN}', "");
Error('\p{Bidi_Class=:=		pop_Directional_Format}');
Error('\P{Bidi_Class=:=		pop_Directional_Format}');
Expect(1, 8236, '\p{Bidi_Class=popdirectionalformat}', "");
Expect(0, 8236, '\p{^Bidi_Class=popdirectionalformat}', "");
Expect(0, 8236, '\P{Bidi_Class=popdirectionalformat}', "");
Expect(1, 8236, '\P{^Bidi_Class=popdirectionalformat}', "");
Expect(0, 8237, '\p{Bidi_Class=popdirectionalformat}', "");
Expect(1, 8237, '\p{^Bidi_Class=popdirectionalformat}', "");
Expect(1, 8237, '\P{Bidi_Class=popdirectionalformat}', "");
Expect(0, 8237, '\P{^Bidi_Class=popdirectionalformat}', "");
Expect(1, 8236, '\p{Bidi_Class=_-Pop_Directional_FORMAT}', "");
Expect(0, 8236, '\p{^Bidi_Class=_-Pop_Directional_FORMAT}', "");
Expect(0, 8236, '\P{Bidi_Class=_-Pop_Directional_FORMAT}', "");
Expect(1, 8236, '\P{^Bidi_Class=_-Pop_Directional_FORMAT}', "");
Expect(0, 8237, '\p{Bidi_Class=_-Pop_Directional_FORMAT}', "");
Expect(1, 8237, '\p{^Bidi_Class=_-Pop_Directional_FORMAT}', "");
Expect(1, 8237, '\P{Bidi_Class=_-Pop_Directional_FORMAT}', "");
Expect(0, 8237, '\P{^Bidi_Class=_-Pop_Directional_FORMAT}', "");
Error('\p{Bc=:=-_PDF}');
Error('\P{Bc=:=-_PDF}');
Expect(1, 8236, '\p{Bc=pdf}', "");
Expect(0, 8236, '\p{^Bc=pdf}', "");
Expect(0, 8236, '\P{Bc=pdf}', "");
Expect(1, 8236, '\P{^Bc=pdf}', "");
Expect(0, 8237, '\p{Bc=pdf}', "");
Expect(1, 8237, '\p{^Bc=pdf}', "");
Expect(1, 8237, '\P{Bc=pdf}', "");
Expect(0, 8237, '\P{^Bc=pdf}', "");
Expect(1, 8236, '\p{Bc: -	PDF}', "");
Expect(0, 8236, '\p{^Bc: -	PDF}', "");
Expect(0, 8236, '\P{Bc: -	PDF}', "");
Expect(1, 8236, '\P{^Bc: -	PDF}', "");
Expect(0, 8237, '\p{Bc: -	PDF}', "");
Expect(1, 8237, '\p{^Bc: -	PDF}', "");
Expect(1, 8237, '\P{Bc: -	PDF}', "");
Expect(0, 8237, '\P{^Bc: -	PDF}', "");
Error('\p{Is_Bidi_Class=:= Pop_Directional_Format}');
Error('\P{Is_Bidi_Class=:= Pop_Directional_Format}');
Expect(1, 8236, '\p{Is_Bidi_Class=popdirectionalformat}', "");
Expect(0, 8236, '\p{^Is_Bidi_Class=popdirectionalformat}', "");
Expect(0, 8236, '\P{Is_Bidi_Class=popdirectionalformat}', "");
Expect(1, 8236, '\P{^Is_Bidi_Class=popdirectionalformat}', "");
Expect(0, 8237, '\p{Is_Bidi_Class=popdirectionalformat}', "");
Expect(1, 8237, '\p{^Is_Bidi_Class=popdirectionalformat}', "");
Expect(1, 8237, '\P{Is_Bidi_Class=popdirectionalformat}', "");
Expect(0, 8237, '\P{^Is_Bidi_Class=popdirectionalformat}', "");
Expect(1, 8236, '\p{Is_Bidi_Class:   --Pop_directional_format}', "");
Expect(0, 8236, '\p{^Is_Bidi_Class:   --Pop_directional_format}', "");
Expect(0, 8236, '\P{Is_Bidi_Class:   --Pop_directional_format}', "");
Expect(1, 8236, '\P{^Is_Bidi_Class:   --Pop_directional_format}', "");
Expect(0, 8237, '\p{Is_Bidi_Class:   --Pop_directional_format}', "");
Expect(1, 8237, '\p{^Is_Bidi_Class:   --Pop_directional_format}', "");
Expect(1, 8237, '\P{Is_Bidi_Class:   --Pop_directional_format}', "");
Expect(0, 8237, '\P{^Is_Bidi_Class:   --Pop_directional_format}', "");
Error('\p{Is_Bc=  PDF/a/}');
Error('\P{Is_Bc=  PDF/a/}');
Expect(1, 8236, '\p{Is_Bc=pdf}', "");
Expect(0, 8236, '\p{^Is_Bc=pdf}', "");
Expect(0, 8236, '\P{Is_Bc=pdf}', "");
Expect(1, 8236, '\P{^Is_Bc=pdf}', "");
Expect(0, 8237, '\p{Is_Bc=pdf}', "");
Expect(1, 8237, '\p{^Is_Bc=pdf}', "");
Expect(1, 8237, '\P{Is_Bc=pdf}', "");
Expect(0, 8237, '\P{^Is_Bc=pdf}', "");
Expect(1, 8236, '\p{Is_Bc=-	PDF}', "");
Expect(0, 8236, '\p{^Is_Bc=-	PDF}', "");
Expect(0, 8236, '\P{Is_Bc=-	PDF}', "");
Expect(1, 8236, '\P{^Is_Bc=-	PDF}', "");
Expect(0, 8237, '\p{Is_Bc=-	PDF}', "");
Expect(1, 8237, '\p{^Is_Bc=-	PDF}', "");
Expect(1, 8237, '\P{Is_Bc=-	PDF}', "");
Expect(0, 8237, '\P{^Is_Bc=-	PDF}', "");
Error('\p{Bidi_Class:	_/a/Left_to_right_Embedding}');
Error('\P{Bidi_Class:	_/a/Left_to_right_Embedding}');
Expect(1, 8234, '\p{Bidi_Class=lefttorightembedding}', "");
Expect(0, 8234, '\p{^Bidi_Class=lefttorightembedding}', "");
Expect(0, 8234, '\P{Bidi_Class=lefttorightembedding}', "");
Expect(1, 8234, '\P{^Bidi_Class=lefttorightembedding}', "");
Expect(0, 8235, '\p{Bidi_Class=lefttorightembedding}', "");
Expect(1, 8235, '\p{^Bidi_Class=lefttorightembedding}', "");
Expect(1, 8235, '\P{Bidi_Class=lefttorightembedding}', "");
Expect(0, 8235, '\P{^Bidi_Class=lefttorightembedding}', "");
Expect(1, 8234, '\p{Bidi_Class:   	 LEFT_TO_Right_Embedding}', "");
Expect(0, 8234, '\p{^Bidi_Class:   	 LEFT_TO_Right_Embedding}', "");
Expect(0, 8234, '\P{Bidi_Class:   	 LEFT_TO_Right_Embedding}', "");
Expect(1, 8234, '\P{^Bidi_Class:   	 LEFT_TO_Right_Embedding}', "");
Expect(0, 8235, '\p{Bidi_Class:   	 LEFT_TO_Right_Embedding}', "");
Expect(1, 8235, '\p{^Bidi_Class:   	 LEFT_TO_Right_Embedding}', "");
Expect(1, 8235, '\P{Bidi_Class:   	 LEFT_TO_Right_Embedding}', "");
Expect(0, 8235, '\P{^Bidi_Class:   	 LEFT_TO_Right_Embedding}', "");
Error('\p{Bc=/a/	_lre}');
Error('\P{Bc=/a/	_lre}');
Expect(1, 8234, '\p{Bc=lre}', "");
Expect(0, 8234, '\p{^Bc=lre}', "");
Expect(0, 8234, '\P{Bc=lre}', "");
Expect(1, 8234, '\P{^Bc=lre}', "");
Expect(0, 8235, '\p{Bc=lre}', "");
Expect(1, 8235, '\p{^Bc=lre}', "");
Expect(1, 8235, '\P{Bc=lre}', "");
Expect(0, 8235, '\P{^Bc=lre}', "");
Expect(1, 8234, '\p{Bc=--lre}', "");
Expect(0, 8234, '\p{^Bc=--lre}', "");
Expect(0, 8234, '\P{Bc=--lre}', "");
Expect(1, 8234, '\P{^Bc=--lre}', "");
Expect(0, 8235, '\p{Bc=--lre}', "");
Expect(1, 8235, '\p{^Bc=--lre}', "");
Expect(1, 8235, '\P{Bc=--lre}', "");
Expect(0, 8235, '\P{^Bc=--lre}', "");
Error('\p{Is_Bidi_Class=-_LEFT_To_RIGHT_Embedding/a/}');
Error('\P{Is_Bidi_Class=-_LEFT_To_RIGHT_Embedding/a/}');
Expect(1, 8234, '\p{Is_Bidi_Class:	lefttorightembedding}', "");
Expect(0, 8234, '\p{^Is_Bidi_Class:	lefttorightembedding}', "");
Expect(0, 8234, '\P{Is_Bidi_Class:	lefttorightembedding}', "");
Expect(1, 8234, '\P{^Is_Bidi_Class:	lefttorightembedding}', "");
Expect(0, 8235, '\p{Is_Bidi_Class:	lefttorightembedding}', "");
Expect(1, 8235, '\p{^Is_Bidi_Class:	lefttorightembedding}', "");
Expect(1, 8235, '\P{Is_Bidi_Class:	lefttorightembedding}', "");
Expect(0, 8235, '\P{^Is_Bidi_Class:	lefttorightembedding}', "");
Expect(1, 8234, '\p{Is_Bidi_Class= Left_TO_Right_embedding}', "");
Expect(0, 8234, '\p{^Is_Bidi_Class= Left_TO_Right_embedding}', "");
Expect(0, 8234, '\P{Is_Bidi_Class= Left_TO_Right_embedding}', "");
Expect(1, 8234, '\P{^Is_Bidi_Class= Left_TO_Right_embedding}', "");
Expect(0, 8235, '\p{Is_Bidi_Class= Left_TO_Right_embedding}', "");
Expect(1, 8235, '\p{^Is_Bidi_Class= Left_TO_Right_embedding}', "");
Expect(1, 8235, '\P{Is_Bidi_Class= Left_TO_Right_embedding}', "");
Expect(0, 8235, '\P{^Is_Bidi_Class= Left_TO_Right_embedding}', "");
Error('\p{Is_Bc: 	:=lre}');
Error('\P{Is_Bc: 	:=lre}');
Expect(1, 8234, '\p{Is_Bc: lre}', "");
Expect(0, 8234, '\p{^Is_Bc: lre}', "");
Expect(0, 8234, '\P{Is_Bc: lre}', "");
Expect(1, 8234, '\P{^Is_Bc: lre}', "");
Expect(0, 8235, '\p{Is_Bc: lre}', "");
Expect(1, 8235, '\p{^Is_Bc: lre}', "");
Expect(1, 8235, '\P{Is_Bc: lre}', "");
Expect(0, 8235, '\P{^Is_Bc: lre}', "");
Expect(1, 8234, '\p{Is_Bc: -lre}', "");
Expect(0, 8234, '\p{^Is_Bc: -lre}', "");
Expect(0, 8234, '\P{Is_Bc: -lre}', "");
Expect(1, 8234, '\P{^Is_Bc: -lre}', "");
Expect(0, 8235, '\p{Is_Bc: -lre}', "");
Expect(1, 8235, '\p{^Is_Bc: -lre}', "");
Expect(1, 8235, '\P{Is_Bc: -lre}', "");
Expect(0, 8235, '\P{^Is_Bc: -lre}', "");
Error('\p{Bidi_Class=	_Arabic_letter/a/}');
Error('\P{Bidi_Class=	_Arabic_letter/a/}');
Expect(1, 126719, '\p{Bidi_Class=arabicletter}', "");
Expect(0, 126719, '\p{^Bidi_Class=arabicletter}', "");
Expect(0, 126719, '\P{Bidi_Class=arabicletter}', "");
Expect(1, 126719, '\P{^Bidi_Class=arabicletter}', "");
Expect(0, 126720, '\p{Bidi_Class=arabicletter}', "");
Expect(1, 126720, '\p{^Bidi_Class=arabicletter}', "");
Expect(1, 126720, '\P{Bidi_Class=arabicletter}', "");
Expect(0, 126720, '\P{^Bidi_Class=arabicletter}', "");
Expect(1, 126719, '\p{Bidi_Class=ARABIC_Letter}', "");
Expect(0, 126719, '\p{^Bidi_Class=ARABIC_Letter}', "");
Expect(0, 126719, '\P{Bidi_Class=ARABIC_Letter}', "");
Expect(1, 126719, '\P{^Bidi_Class=ARABIC_Letter}', "");
Expect(0, 126720, '\p{Bidi_Class=ARABIC_Letter}', "");
Expect(1, 126720, '\p{^Bidi_Class=ARABIC_Letter}', "");
Expect(1, 126720, '\P{Bidi_Class=ARABIC_Letter}', "");
Expect(0, 126720, '\P{^Bidi_Class=ARABIC_Letter}', "");
Error('\p{Bc=:=_-AL}');
Error('\P{Bc=:=_-AL}');
Expect(1, 126719, '\p{Bc=al}', "");
Expect(0, 126719, '\p{^Bc=al}', "");
Expect(0, 126719, '\P{Bc=al}', "");
Expect(1, 126719, '\P{^Bc=al}', "");
Expect(0, 126720, '\p{Bc=al}', "");
Expect(1, 126720, '\p{^Bc=al}', "");
Expect(1, 126720, '\P{Bc=al}', "");
Expect(0, 126720, '\P{^Bc=al}', "");
Expect(1, 126719, '\p{Bc=- AL}', "");
Expect(0, 126719, '\p{^Bc=- AL}', "");
Expect(0, 126719, '\P{Bc=- AL}', "");
Expect(1, 126719, '\P{^Bc=- AL}', "");
Expect(0, 126720, '\p{Bc=- AL}', "");
Expect(1, 126720, '\p{^Bc=- AL}', "");
Expect(1, 126720, '\P{Bc=- AL}', "");
Expect(0, 126720, '\P{^Bc=- AL}', "");
Error('\p{Is_Bidi_Class=/a/  Arabic_Letter}');
Error('\P{Is_Bidi_Class=/a/  Arabic_Letter}');
Expect(1, 126719, '\p{Is_Bidi_Class=arabicletter}', "");
Expect(0, 126719, '\p{^Is_Bidi_Class=arabicletter}', "");
Expect(0, 126719, '\P{Is_Bidi_Class=arabicletter}', "");
Expect(1, 126719, '\P{^Is_Bidi_Class=arabicletter}', "");
Expect(0, 126720, '\p{Is_Bidi_Class=arabicletter}', "");
Expect(1, 126720, '\p{^Is_Bidi_Class=arabicletter}', "");
Expect(1, 126720, '\P{Is_Bidi_Class=arabicletter}', "");
Expect(0, 126720, '\P{^Is_Bidi_Class=arabicletter}', "");
Expect(1, 126719, '\p{Is_Bidi_Class: 	 ARABIC_Letter}', "");
Expect(0, 126719, '\p{^Is_Bidi_Class: 	 ARABIC_Letter}', "");
Expect(0, 126719, '\P{Is_Bidi_Class: 	 ARABIC_Letter}', "");
Expect(1, 126719, '\P{^Is_Bidi_Class: 	 ARABIC_Letter}', "");
Expect(0, 126720, '\p{Is_Bidi_Class: 	 ARABIC_Letter}', "");
Expect(1, 126720, '\p{^Is_Bidi_Class: 	 ARABIC_Letter}', "");
Expect(1, 126720, '\P{Is_Bidi_Class: 	 ARABIC_Letter}', "");
Expect(0, 126720, '\P{^Is_Bidi_Class: 	 ARABIC_Letter}', "");
Error('\p{Is_Bc:	:=AL}');
Error('\P{Is_Bc:	:=AL}');
Expect(1, 126719, '\p{Is_Bc=al}', "");
Expect(0, 126719, '\p{^Is_Bc=al}', "");
Expect(0, 126719, '\P{Is_Bc=al}', "");
Expect(1, 126719, '\P{^Is_Bc=al}', "");
Expect(0, 126720, '\p{Is_Bc=al}', "");
Expect(1, 126720, '\p{^Is_Bc=al}', "");
Expect(1, 126720, '\P{Is_Bc=al}', "");
Expect(0, 126720, '\P{^Is_Bc=al}', "");
Expect(1, 126719, '\p{Is_Bc=__al}', "");
Expect(0, 126719, '\p{^Is_Bc=__al}', "");
Expect(0, 126719, '\P{Is_Bc=__al}', "");
Expect(1, 126719, '\P{^Is_Bc=__al}', "");
Expect(0, 126720, '\p{Is_Bc=__al}', "");
Expect(1, 126720, '\p{^Is_Bc=__al}', "");
Expect(1, 126720, '\P{Is_Bc=__al}', "");
Expect(0, 126720, '\P{^Is_Bc=__al}', "");
Error('\p{Bidi_Class=:=	_Common_separator}');
Error('\P{Bidi_Class=:=	_Common_separator}');
Expect(1, 65306, '\p{Bidi_Class=commonseparator}', "");
Expect(0, 65306, '\p{^Bidi_Class=commonseparator}', "");
Expect(0, 65306, '\P{Bidi_Class=commonseparator}', "");
Expect(1, 65306, '\P{^Bidi_Class=commonseparator}', "");
Expect(0, 65307, '\p{Bidi_Class=commonseparator}', "");
Expect(1, 65307, '\p{^Bidi_Class=commonseparator}', "");
Expect(1, 65307, '\P{Bidi_Class=commonseparator}', "");
Expect(0, 65307, '\P{^Bidi_Class=commonseparator}', "");
Expect(1, 65306, '\p{Bidi_Class=	-common_SEPARATOR}', "");
Expect(0, 65306, '\p{^Bidi_Class=	-common_SEPARATOR}', "");
Expect(0, 65306, '\P{Bidi_Class=	-common_SEPARATOR}', "");
Expect(1, 65306, '\P{^Bidi_Class=	-common_SEPARATOR}', "");
Expect(0, 65307, '\p{Bidi_Class=	-common_SEPARATOR}', "");
Expect(1, 65307, '\p{^Bidi_Class=	-common_SEPARATOR}', "");
Expect(1, 65307, '\P{Bidi_Class=	-common_SEPARATOR}', "");
Expect(0, 65307, '\P{^Bidi_Class=	-common_SEPARATOR}', "");
Error('\p{Bc=-:=CS}');
Error('\P{Bc=-:=CS}');
Expect(1, 65306, '\p{Bc=cs}', "");
Expect(0, 65306, '\p{^Bc=cs}', "");
Expect(0, 65306, '\P{Bc=cs}', "");
Expect(1, 65306, '\P{^Bc=cs}', "");
Expect(0, 65307, '\p{Bc=cs}', "");
Expect(1, 65307, '\p{^Bc=cs}', "");
Expect(1, 65307, '\P{Bc=cs}', "");
Expect(0, 65307, '\P{^Bc=cs}', "");
Expect(1, 65306, '\p{Bc= -cs}', "");
Expect(0, 65306, '\p{^Bc= -cs}', "");
Expect(0, 65306, '\P{Bc= -cs}', "");
Expect(1, 65306, '\P{^Bc= -cs}', "");
Expect(0, 65307, '\p{Bc= -cs}', "");
Expect(1, 65307, '\p{^Bc= -cs}', "");
Expect(1, 65307, '\P{Bc= -cs}', "");
Expect(0, 65307, '\P{^Bc= -cs}', "");
Error('\p{Is_Bidi_Class=_/a/Common_Separator}');
Error('\P{Is_Bidi_Class=_/a/Common_Separator}');
Expect(1, 65306, '\p{Is_Bidi_Class=commonseparator}', "");
Expect(0, 65306, '\p{^Is_Bidi_Class=commonseparator}', "");
Expect(0, 65306, '\P{Is_Bidi_Class=commonseparator}', "");
Expect(1, 65306, '\P{^Is_Bidi_Class=commonseparator}', "");
Expect(0, 65307, '\p{Is_Bidi_Class=commonseparator}', "");
Expect(1, 65307, '\p{^Is_Bidi_Class=commonseparator}', "");
Expect(1, 65307, '\P{Is_Bidi_Class=commonseparator}', "");
Expect(0, 65307, '\P{^Is_Bidi_Class=commonseparator}', "");
Expect(1, 65306, '\p{Is_Bidi_Class= common_SEPARATOR}', "");
Expect(0, 65306, '\p{^Is_Bidi_Class= common_SEPARATOR}', "");
Expect(0, 65306, '\P{Is_Bidi_Class= common_SEPARATOR}', "");
Expect(1, 65306, '\P{^Is_Bidi_Class= common_SEPARATOR}', "");
Expect(0, 65307, '\p{Is_Bidi_Class= common_SEPARATOR}', "");
Expect(1, 65307, '\p{^Is_Bidi_Class= common_SEPARATOR}', "");
Expect(1, 65307, '\P{Is_Bidi_Class= common_SEPARATOR}', "");
Expect(0, 65307, '\P{^Is_Bidi_Class= common_SEPARATOR}', "");
Error('\p{Is_Bc=/a/	-CS}');
Error('\P{Is_Bc=/a/	-CS}');
Expect(1, 65306, '\p{Is_Bc=cs}', "");
Expect(0, 65306, '\p{^Is_Bc=cs}', "");
Expect(0, 65306, '\P{Is_Bc=cs}', "");
Expect(1, 65306, '\P{^Is_Bc=cs}', "");
Expect(0, 65307, '\p{Is_Bc=cs}', "");
Expect(1, 65307, '\p{^Is_Bc=cs}', "");
Expect(1, 65307, '\P{Is_Bc=cs}', "");
Expect(0, 65307, '\P{^Is_Bc=cs}', "");
Expect(1, 65306, '\p{Is_Bc=CS}', "");
Expect(0, 65306, '\p{^Is_Bc=CS}', "");
Expect(0, 65306, '\P{Is_Bc=CS}', "");
Expect(1, 65306, '\P{^Is_Bc=CS}', "");
Expect(0, 65307, '\p{Is_Bc=CS}', "");
Expect(1, 65307, '\p{^Is_Bc=CS}', "");
Expect(1, 65307, '\P{Is_Bc=CS}', "");
Expect(0, 65307, '\P{^Is_Bc=CS}', "");
Error('\p{Bidi_Class:	:= 	WHITE_space}');
Error('\P{Bidi_Class:	:= 	WHITE_space}');
Expect(1, 12288, '\p{Bidi_Class: whitespace}', "");
Expect(0, 12288, '\p{^Bidi_Class: whitespace}', "");
Expect(0, 12288, '\P{Bidi_Class: whitespace}', "");
Expect(1, 12288, '\P{^Bidi_Class: whitespace}', "");
Expect(0, 12289, '\p{Bidi_Class: whitespace}', "");
Expect(1, 12289, '\p{^Bidi_Class: whitespace}', "");
Expect(1, 12289, '\P{Bidi_Class: whitespace}', "");
Expect(0, 12289, '\P{^Bidi_Class: whitespace}', "");
Expect(1, 12288, '\p{Bidi_Class=_WHITE_Space}', "");
Expect(0, 12288, '\p{^Bidi_Class=_WHITE_Space}', "");
Expect(0, 12288, '\P{Bidi_Class=_WHITE_Space}', "");
Expect(1, 12288, '\P{^Bidi_Class=_WHITE_Space}', "");
Expect(0, 12289, '\p{Bidi_Class=_WHITE_Space}', "");
Expect(1, 12289, '\p{^Bidi_Class=_WHITE_Space}', "");
Expect(1, 12289, '\P{Bidi_Class=_WHITE_Space}', "");
Expect(0, 12289, '\P{^Bidi_Class=_WHITE_Space}', "");
Error('\p{Bc=:=WS}');
Error('\P{Bc=:=WS}');
Expect(1, 12288, '\p{Bc:   ws}', "");
Expect(0, 12288, '\p{^Bc:   ws}', "");
Expect(0, 12288, '\P{Bc:   ws}', "");
Expect(1, 12288, '\P{^Bc:   ws}', "");
Expect(0, 12289, '\p{Bc:   ws}', "");
Expect(1, 12289, '\p{^Bc:   ws}', "");
Expect(1, 12289, '\P{Bc:   ws}', "");
Expect(0, 12289, '\P{^Bc:   ws}', "");
Expect(1, 12288, '\p{Bc: WS}', "");
Expect(0, 12288, '\p{^Bc: WS}', "");
Expect(0, 12288, '\P{Bc: WS}', "");
Expect(1, 12288, '\P{^Bc: WS}', "");
Expect(0, 12289, '\p{Bc: WS}', "");
Expect(1, 12289, '\p{^Bc: WS}', "");
Expect(1, 12289, '\P{Bc: WS}', "");
Expect(0, 12289, '\P{^Bc: WS}', "");
Error('\p{Is_Bidi_Class=:=  WHITE_SPACE}');
Error('\P{Is_Bidi_Class=:=  WHITE_SPACE}');
Expect(1, 12288, '\p{Is_Bidi_Class=whitespace}', "");
Expect(0, 12288, '\p{^Is_Bidi_Class=whitespace}', "");
Expect(0, 12288, '\P{Is_Bidi_Class=whitespace}', "");
Expect(1, 12288, '\P{^Is_Bidi_Class=whitespace}', "");
Expect(0, 12289, '\p{Is_Bidi_Class=whitespace}', "");
Expect(1, 12289, '\p{^Is_Bidi_Class=whitespace}', "");
Expect(1, 12289, '\P{Is_Bidi_Class=whitespace}', "");
Expect(0, 12289, '\P{^Is_Bidi_Class=whitespace}', "");
Expect(1, 12288, '\p{Is_Bidi_Class: -WHITE_SPACE}', "");
Expect(0, 12288, '\p{^Is_Bidi_Class: -WHITE_SPACE}', "");
Expect(0, 12288, '\P{Is_Bidi_Class: -WHITE_SPACE}', "");
Expect(1, 12288, '\P{^Is_Bidi_Class: -WHITE_SPACE}', "");
Expect(0, 12289, '\p{Is_Bidi_Class: -WHITE_SPACE}', "");
Expect(1, 12289, '\p{^Is_Bidi_Class: -WHITE_SPACE}', "");
Expect(1, 12289, '\P{Is_Bidi_Class: -WHITE_SPACE}', "");
Expect(0, 12289, '\P{^Is_Bidi_Class: -WHITE_SPACE}', "");
Error('\p{Is_Bc=	WS:=}');
Error('\P{Is_Bc=	WS:=}');
Expect(1, 12288, '\p{Is_Bc=ws}', "");
Expect(0, 12288, '\p{^Is_Bc=ws}', "");
Expect(0, 12288, '\P{Is_Bc=ws}', "");
Expect(1, 12288, '\P{^Is_Bc=ws}', "");
Expect(0, 12289, '\p{Is_Bc=ws}', "");
Expect(1, 12289, '\p{^Is_Bc=ws}', "");
Expect(1, 12289, '\P{Is_Bc=ws}', "");
Expect(0, 12289, '\P{^Is_Bc=ws}', "");
Expect(1, 12288, '\p{Is_Bc=_WS}', "");
Expect(0, 12288, '\p{^Is_Bc=_WS}', "");
Expect(0, 12288, '\P{Is_Bc=_WS}', "");
Expect(1, 12288, '\P{^Is_Bc=_WS}', "");
Expect(0, 12289, '\p{Is_Bc=_WS}', "");
Expect(1, 12289, '\p{^Is_Bc=_WS}', "");
Expect(1, 12289, '\P{Is_Bc=_WS}', "");
Expect(0, 12289, '\P{^Is_Bc=_WS}', "");
Error('\p{Bidi_Class=_-paragraph_SEPARATOR/a/}');
Error('\P{Bidi_Class=_-paragraph_SEPARATOR/a/}');
Expect(1, 8233, '\p{Bidi_Class=paragraphseparator}', "");
Expect(0, 8233, '\p{^Bidi_Class=paragraphseparator}', "");
Expect(0, 8233, '\P{Bidi_Class=paragraphseparator}', "");
Expect(1, 8233, '\P{^Bidi_Class=paragraphseparator}', "");
Expect(0, 8234, '\p{Bidi_Class=paragraphseparator}', "");
Expect(1, 8234, '\p{^Bidi_Class=paragraphseparator}', "");
Expect(1, 8234, '\P{Bidi_Class=paragraphseparator}', "");
Expect(0, 8234, '\P{^Bidi_Class=paragraphseparator}', "");
Expect(1, 8233, '\p{Bidi_Class:	 _paragraph_Separator}', "");
Expect(0, 8233, '\p{^Bidi_Class:	 _paragraph_Separator}', "");
Expect(0, 8233, '\P{Bidi_Class:	 _paragraph_Separator}', "");
Expect(1, 8233, '\P{^Bidi_Class:	 _paragraph_Separator}', "");
Expect(0, 8234, '\p{Bidi_Class:	 _paragraph_Separator}', "");
Expect(1, 8234, '\p{^Bidi_Class:	 _paragraph_Separator}', "");
Expect(1, 8234, '\P{Bidi_Class:	 _paragraph_Separator}', "");
Expect(0, 8234, '\P{^Bidi_Class:	 _paragraph_Separator}', "");
Error('\p{Bc=/a/_ B}');
Error('\P{Bc=/a/_ B}');
Expect(1, 8233, '\p{Bc=b}', "");
Expect(0, 8233, '\p{^Bc=b}', "");
Expect(0, 8233, '\P{Bc=b}', "");
Expect(1, 8233, '\P{^Bc=b}', "");
Expect(0, 8234, '\p{Bc=b}', "");
Expect(1, 8234, '\p{^Bc=b}', "");
Expect(1, 8234, '\P{Bc=b}', "");
Expect(0, 8234, '\P{^Bc=b}', "");
Expect(1, 8233, '\p{Bc=	-B}', "");
Expect(0, 8233, '\p{^Bc=	-B}', "");
Expect(0, 8233, '\P{Bc=	-B}', "");
Expect(1, 8233, '\P{^Bc=	-B}', "");
Expect(0, 8234, '\p{Bc=	-B}', "");
Expect(1, 8234, '\p{^Bc=	-B}', "");
Expect(1, 8234, '\P{Bc=	-B}', "");
Expect(0, 8234, '\P{^Bc=	-B}', "");
Error('\p{Is_Bidi_Class=	/a/Paragraph_Separator}');
Error('\P{Is_Bidi_Class=	/a/Paragraph_Separator}');
Expect(1, 8233, '\p{Is_Bidi_Class=paragraphseparator}', "");
Expect(0, 8233, '\p{^Is_Bidi_Class=paragraphseparator}', "");
Expect(0, 8233, '\P{Is_Bidi_Class=paragraphseparator}', "");
Expect(1, 8233, '\P{^Is_Bidi_Class=paragraphseparator}', "");
Expect(0, 8234, '\p{Is_Bidi_Class=paragraphseparator}', "");
Expect(1, 8234, '\p{^Is_Bidi_Class=paragraphseparator}', "");
Expect(1, 8234, '\P{Is_Bidi_Class=paragraphseparator}', "");
Expect(0, 8234, '\P{^Is_Bidi_Class=paragraphseparator}', "");
Expect(1, 8233, '\p{Is_Bidi_Class= _Paragraph_SEPARATOR}', "");
Expect(0, 8233, '\p{^Is_Bidi_Class= _Paragraph_SEPARATOR}', "");
Expect(0, 8233, '\P{Is_Bidi_Class= _Paragraph_SEPARATOR}', "");
Expect(1, 8233, '\P{^Is_Bidi_Class= _Paragraph_SEPARATOR}', "");
Expect(0, 8234, '\p{Is_Bidi_Class= _Paragraph_SEPARATOR}', "");
Expect(1, 8234, '\p{^Is_Bidi_Class= _Paragraph_SEPARATOR}', "");
Expect(1, 8234, '\P{Is_Bidi_Class= _Paragraph_SEPARATOR}', "");
Expect(0, 8234, '\P{^Is_Bidi_Class= _Paragraph_SEPARATOR}', "");
Error('\p{Is_Bc=/a/_B}');
Error('\P{Is_Bc=/a/_B}');
Expect(1, 8233, '\p{Is_Bc=b}', "");
Expect(0, 8233, '\p{^Is_Bc=b}', "");
Expect(0, 8233, '\P{Is_Bc=b}', "");
Expect(1, 8233, '\P{^Is_Bc=b}', "");
Expect(0, 8234, '\p{Is_Bc=b}', "");
Expect(1, 8234, '\p{^Is_Bc=b}', "");
Expect(1, 8234, '\P{Is_Bc=b}', "");
Expect(0, 8234, '\P{^Is_Bc=b}', "");
Expect(1, 8233, '\p{Is_Bc=-_B}', "");
Expect(0, 8233, '\p{^Is_Bc=-_B}', "");
Expect(0, 8233, '\P{Is_Bc=-_B}', "");
Expect(1, 8233, '\P{^Is_Bc=-_B}', "");
Expect(0, 8234, '\p{Is_Bc=-_B}', "");
Expect(1, 8234, '\p{^Is_Bc=-_B}', "");
Expect(1, 8234, '\P{Is_Bc=-_B}', "");
Expect(0, 8234, '\P{^Is_Bc=-_B}', "");
Error('\p{Bidi_Class=	-Nonspacing_Mark:=}');
Error('\P{Bidi_Class=	-Nonspacing_Mark:=}');
Expect(1, 917999, '\p{Bidi_Class:	nonspacingmark}', "");
Expect(0, 917999, '\p{^Bidi_Class:	nonspacingmark}', "");
Expect(0, 917999, '\P{Bidi_Class:	nonspacingmark}', "");
Expect(1, 917999, '\P{^Bidi_Class:	nonspacingmark}', "");
Expect(0, 918000, '\p{Bidi_Class:	nonspacingmark}', "");
Expect(1, 918000, '\p{^Bidi_Class:	nonspacingmark}', "");
Expect(1, 918000, '\P{Bidi_Class:	nonspacingmark}', "");
Expect(0, 918000, '\P{^Bidi_Class:	nonspacingmark}', "");
Expect(1, 917999, '\p{Bidi_Class=-	NONSPACING_mark}', "");
Expect(0, 917999, '\p{^Bidi_Class=-	NONSPACING_mark}', "");
Expect(0, 917999, '\P{Bidi_Class=-	NONSPACING_mark}', "");
Expect(1, 917999, '\P{^Bidi_Class=-	NONSPACING_mark}', "");
Expect(0, 918000, '\p{Bidi_Class=-	NONSPACING_mark}', "");
Expect(1, 918000, '\p{^Bidi_Class=-	NONSPACING_mark}', "");
Expect(1, 918000, '\P{Bidi_Class=-	NONSPACING_mark}', "");
Expect(0, 918000, '\P{^Bidi_Class=-	NONSPACING_mark}', "");
Error('\p{Bc=_-NSM:=}');
Error('\P{Bc=_-NSM:=}');
Expect(1, 917999, '\p{Bc=nsm}', "");
Expect(0, 917999, '\p{^Bc=nsm}', "");
Expect(0, 917999, '\P{Bc=nsm}', "");
Expect(1, 917999, '\P{^Bc=nsm}', "");
Expect(0, 918000, '\p{Bc=nsm}', "");
Expect(1, 918000, '\p{^Bc=nsm}', "");
Expect(1, 918000, '\P{Bc=nsm}', "");
Expect(0, 918000, '\P{^Bc=nsm}', "");
Expect(1, 917999, '\p{Bc=-_NSM}', "");
Expect(0, 917999, '\p{^Bc=-_NSM}', "");
Expect(0, 917999, '\P{Bc=-_NSM}', "");
Expect(1, 917999, '\P{^Bc=-_NSM}', "");
Expect(0, 918000, '\p{Bc=-_NSM}', "");
Expect(1, 918000, '\p{^Bc=-_NSM}', "");
Expect(1, 918000, '\P{Bc=-_NSM}', "");
Expect(0, 918000, '\P{^Bc=-_NSM}', "");
Error('\p{Is_Bidi_Class= /a/NONSPACING_Mark}');
Error('\P{Is_Bidi_Class= /a/NONSPACING_Mark}');
Expect(1, 917999, '\p{Is_Bidi_Class=nonspacingmark}', "");
Expect(0, 917999, '\p{^Is_Bidi_Class=nonspacingmark}', "");
Expect(0, 917999, '\P{Is_Bidi_Class=nonspacingmark}', "");
Expect(1, 917999, '\P{^Is_Bidi_Class=nonspacingmark}', "");
Expect(0, 918000, '\p{Is_Bidi_Class=nonspacingmark}', "");
Expect(1, 918000, '\p{^Is_Bidi_Class=nonspacingmark}', "");
Expect(1, 918000, '\P{Is_Bidi_Class=nonspacingmark}', "");
Expect(0, 918000, '\P{^Is_Bidi_Class=nonspacingmark}', "");
Expect(1, 917999, '\p{Is_Bidi_Class= _Nonspacing_Mark}', "");
Expect(0, 917999, '\p{^Is_Bidi_Class= _Nonspacing_Mark}', "");
Expect(0, 917999, '\P{Is_Bidi_Class= _Nonspacing_Mark}', "");
Expect(1, 917999, '\P{^Is_Bidi_Class= _Nonspacing_Mark}', "");
Expect(0, 918000, '\p{Is_Bidi_Class= _Nonspacing_Mark}', "");
Expect(1, 918000, '\p{^Is_Bidi_Class= _Nonspacing_Mark}', "");
Expect(1, 918000, '\P{Is_Bidi_Class= _Nonspacing_Mark}', "");
Expect(0, 918000, '\P{^Is_Bidi_Class= _Nonspacing_Mark}', "");
Error('\p{Is_Bc=_/a/NSM}');
Error('\P{Is_Bc=_/a/NSM}');
Expect(1, 917999, '\p{Is_Bc=nsm}', "");
Expect(0, 917999, '\p{^Is_Bc=nsm}', "");
Expect(0, 917999, '\P{Is_Bc=nsm}', "");
Expect(1, 917999, '\P{^Is_Bc=nsm}', "");
Expect(0, 918000, '\p{Is_Bc=nsm}', "");
Expect(1, 918000, '\p{^Is_Bc=nsm}', "");
Expect(1, 918000, '\P{Is_Bc=nsm}', "");
Expect(0, 918000, '\P{^Is_Bc=nsm}', "");
Expect(1, 917999, '\p{Is_Bc=		nsm}', "");
Expect(0, 917999, '\p{^Is_Bc=		nsm}', "");
Expect(0, 917999, '\P{Is_Bc=		nsm}', "");
Expect(1, 917999, '\P{^Is_Bc=		nsm}', "");
Expect(0, 918000, '\p{Is_Bc=		nsm}', "");
Expect(1, 918000, '\p{^Is_Bc=		nsm}', "");
Expect(1, 918000, '\P{Is_Bc=		nsm}', "");
Expect(0, 918000, '\P{^Is_Bc=		nsm}', "");
Error('\p{Bidi_Class=/a/ 	European_Separator}');
Error('\P{Bidi_Class=/a/ 	European_Separator}');
Expect(1, 65293, '\p{Bidi_Class=europeanseparator}', "");
Expect(0, 65293, '\p{^Bidi_Class=europeanseparator}', "");
Expect(0, 65293, '\P{Bidi_Class=europeanseparator}', "");
Expect(1, 65293, '\P{^Bidi_Class=europeanseparator}', "");
Expect(0, 65294, '\p{Bidi_Class=europeanseparator}', "");
Expect(1, 65294, '\p{^Bidi_Class=europeanseparator}', "");
Expect(1, 65294, '\P{Bidi_Class=europeanseparator}', "");
Expect(0, 65294, '\P{^Bidi_Class=europeanseparator}', "");
Expect(1, 65293, '\p{Bidi_Class=	 European_Separator}', "");
Expect(0, 65293, '\p{^Bidi_Class=	 European_Separator}', "");
Expect(0, 65293, '\P{Bidi_Class=	 European_Separator}', "");
Expect(1, 65293, '\P{^Bidi_Class=	 European_Separator}', "");
Expect(0, 65294, '\p{Bidi_Class=	 European_Separator}', "");
Expect(1, 65294, '\p{^Bidi_Class=	 European_Separator}', "");
Expect(1, 65294, '\P{Bidi_Class=	 European_Separator}', "");
Expect(0, 65294, '\P{^Bidi_Class=	 European_Separator}', "");
Error('\p{Bc=/a/-ES}');
Error('\P{Bc=/a/-ES}');
Expect(1, 65293, '\p{Bc=es}', "");
Expect(0, 65293, '\p{^Bc=es}', "");
Expect(0, 65293, '\P{Bc=es}', "");
Expect(1, 65293, '\P{^Bc=es}', "");
Expect(0, 65294, '\p{Bc=es}', "");
Expect(1, 65294, '\p{^Bc=es}', "");
Expect(1, 65294, '\P{Bc=es}', "");
Expect(0, 65294, '\P{^Bc=es}', "");
Expect(1, 65293, '\p{Bc=-	ES}', "");
Expect(0, 65293, '\p{^Bc=-	ES}', "");
Expect(0, 65293, '\P{Bc=-	ES}', "");
Expect(1, 65293, '\P{^Bc=-	ES}', "");
Expect(0, 65294, '\p{Bc=-	ES}', "");
Expect(1, 65294, '\p{^Bc=-	ES}', "");
Expect(1, 65294, '\P{Bc=-	ES}', "");
Expect(0, 65294, '\P{^Bc=-	ES}', "");
Error('\p{Is_Bidi_Class=-:=european_Separator}');
Error('\P{Is_Bidi_Class=-:=european_Separator}');
Expect(1, 65293, '\p{Is_Bidi_Class=europeanseparator}', "");
Expect(0, 65293, '\p{^Is_Bidi_Class=europeanseparator}', "");
Expect(0, 65293, '\P{Is_Bidi_Class=europeanseparator}', "");
Expect(1, 65293, '\P{^Is_Bidi_Class=europeanseparator}', "");
Expect(0, 65294, '\p{Is_Bidi_Class=europeanseparator}', "");
Expect(1, 65294, '\p{^Is_Bidi_Class=europeanseparator}', "");
Expect(1, 65294, '\P{Is_Bidi_Class=europeanseparator}', "");
Expect(0, 65294, '\P{^Is_Bidi_Class=europeanseparator}', "");
Expect(1, 65293, '\p{Is_Bidi_Class=-european_Separator}', "");
Expect(0, 65293, '\p{^Is_Bidi_Class=-european_Separator}', "");
Expect(0, 65293, '\P{Is_Bidi_Class=-european_Separator}', "");
Expect(1, 65293, '\P{^Is_Bidi_Class=-european_Separator}', "");
Expect(0, 65294, '\p{Is_Bidi_Class=-european_Separator}', "");
Expect(1, 65294, '\p{^Is_Bidi_Class=-european_Separator}', "");
Expect(1, 65294, '\P{Is_Bidi_Class=-european_Separator}', "");
Expect(0, 65294, '\P{^Is_Bidi_Class=-european_Separator}', "");
Error('\p{Is_Bc=ES/a/}');
Error('\P{Is_Bc=ES/a/}');
Expect(1, 65293, '\p{Is_Bc=es}', "");
Expect(0, 65293, '\p{^Is_Bc=es}', "");
Expect(0, 65293, '\P{Is_Bc=es}', "");
Expect(1, 65293, '\P{^Is_Bc=es}', "");
Expect(0, 65294, '\p{Is_Bc=es}', "");
Expect(1, 65294, '\p{^Is_Bc=es}', "");
Expect(1, 65294, '\P{Is_Bc=es}', "");
Expect(0, 65294, '\P{^Is_Bc=es}', "");
Expect(1, 65293, '\p{Is_Bc= es}', "");
Expect(0, 65293, '\p{^Is_Bc= es}', "");
Expect(0, 65293, '\P{Is_Bc= es}', "");
Expect(1, 65293, '\P{^Is_Bc= es}', "");
Expect(0, 65294, '\p{Is_Bc= es}', "");
Expect(1, 65294, '\p{^Is_Bc= es}', "");
Expect(1, 65294, '\P{Is_Bc= es}', "");
Expect(0, 65294, '\P{^Is_Bc= es}', "");
Error('\p{Bidi_Class=:=__European_Number}');
Error('\P{Bidi_Class=:=__European_Number}');
Expect(1, 127242, '\p{Bidi_Class=europeannumber}', "");
Expect(0, 127242, '\p{^Bidi_Class=europeannumber}', "");
Expect(0, 127242, '\P{Bidi_Class=europeannumber}', "");
Expect(1, 127242, '\P{^Bidi_Class=europeannumber}', "");
Expect(0, 127243, '\p{Bidi_Class=europeannumber}', "");
Expect(1, 127243, '\p{^Bidi_Class=europeannumber}', "");
Expect(1, 127243, '\P{Bidi_Class=europeannumber}', "");
Expect(0, 127243, '\P{^Bidi_Class=europeannumber}', "");
Expect(1, 127242, '\p{Bidi_Class=_ European_Number}', "");
Expect(0, 127242, '\p{^Bidi_Class=_ European_Number}', "");
Expect(0, 127242, '\P{Bidi_Class=_ European_Number}', "");
Expect(1, 127242, '\P{^Bidi_Class=_ European_Number}', "");
Expect(0, 127243, '\p{Bidi_Class=_ European_Number}', "");
Expect(1, 127243, '\p{^Bidi_Class=_ European_Number}', "");
Expect(1, 127243, '\P{Bidi_Class=_ European_Number}', "");
Expect(0, 127243, '\P{^Bidi_Class=_ European_Number}', "");
Error('\p{Bc=_ EN:=}');
Error('\P{Bc=_ EN:=}');
Expect(1, 127242, '\p{Bc=en}', "");
Expect(0, 127242, '\p{^Bc=en}', "");
Expect(0, 127242, '\P{Bc=en}', "");
Expect(1, 127242, '\P{^Bc=en}', "");
Expect(0, 127243, '\p{Bc=en}', "");
Expect(1, 127243, '\p{^Bc=en}', "");
Expect(1, 127243, '\P{Bc=en}', "");
Expect(0, 127243, '\P{^Bc=en}', "");
Expect(1, 127242, '\p{Bc=_-EN}', "");
Expect(0, 127242, '\p{^Bc=_-EN}', "");
Expect(0, 127242, '\P{Bc=_-EN}', "");
Expect(1, 127242, '\P{^Bc=_-EN}', "");
Expect(0, 127243, '\p{Bc=_-EN}', "");
Expect(1, 127243, '\p{^Bc=_-EN}', "");
Expect(1, 127243, '\P{Bc=_-EN}', "");
Expect(0, 127243, '\P{^Bc=_-EN}', "");
Error('\p{Is_Bidi_Class=/a/	 EUROPEAN_Number}');
Error('\P{Is_Bidi_Class=/a/	 EUROPEAN_Number}');
Expect(1, 127242, '\p{Is_Bidi_Class=europeannumber}', "");
Expect(0, 127242, '\p{^Is_Bidi_Class=europeannumber}', "");
Expect(0, 127242, '\P{Is_Bidi_Class=europeannumber}', "");
Expect(1, 127242, '\P{^Is_Bidi_Class=europeannumber}', "");
Expect(0, 127243, '\p{Is_Bidi_Class=europeannumber}', "");
Expect(1, 127243, '\p{^Is_Bidi_Class=europeannumber}', "");
Expect(1, 127243, '\P{Is_Bidi_Class=europeannumber}', "");
Expect(0, 127243, '\P{^Is_Bidi_Class=europeannumber}', "");
Expect(1, 127242, '\p{Is_Bidi_Class=	-european_Number}', "");
Expect(0, 127242, '\p{^Is_Bidi_Class=	-european_Number}', "");
Expect(0, 127242, '\P{Is_Bidi_Class=	-european_Number}', "");
Expect(1, 127242, '\P{^Is_Bidi_Class=	-european_Number}', "");
Expect(0, 127243, '\p{Is_Bidi_Class=	-european_Number}', "");
Expect(1, 127243, '\p{^Is_Bidi_Class=	-european_Number}', "");
Expect(1, 127243, '\P{Is_Bidi_Class=	-european_Number}', "");
Expect(0, 127243, '\P{^Is_Bidi_Class=	-european_Number}', "");
Error('\p{Is_Bc= /a/EN}');
Error('\P{Is_Bc= /a/EN}');
Expect(1, 127242, '\p{Is_Bc:en}', "");
Expect(0, 127242, '\p{^Is_Bc:en}', "");
Expect(0, 127242, '\P{Is_Bc:en}', "");
Expect(1, 127242, '\P{^Is_Bc:en}', "");
Expect(0, 127243, '\p{Is_Bc:en}', "");
Expect(1, 127243, '\p{^Is_Bc:en}', "");
Expect(1, 127243, '\P{Is_Bc:en}', "");
Expect(0, 127243, '\P{^Is_Bc:en}', "");
Expect(1, 127242, '\p{Is_Bc=	_EN}', "");
Expect(0, 127242, '\p{^Is_Bc=	_EN}', "");
Expect(0, 127242, '\P{Is_Bc=	_EN}', "");
Expect(1, 127242, '\P{^Is_Bc=	_EN}', "");
Expect(0, 127243, '\p{Is_Bc=	_EN}', "");
Expect(1, 127243, '\p{^Is_Bc=	_EN}', "");
Expect(1, 127243, '\P{Is_Bc=	_EN}', "");
Expect(0, 127243, '\P{^Is_Bc=	_EN}', "");
Error('\p{Bidi_Class=/a/_-european_Terminator}');
Error('\P{Bidi_Class=/a/_-european_Terminator}');
Expect(1, 65510, '\p{Bidi_Class=europeanterminator}', "");
Expect(0, 65510, '\p{^Bidi_Class=europeanterminator}', "");
Expect(0, 65510, '\P{Bidi_Class=europeanterminator}', "");
Expect(1, 65510, '\P{^Bidi_Class=europeanterminator}', "");
Expect(0, 65511, '\p{Bidi_Class=europeanterminator}', "");
Expect(1, 65511, '\p{^Bidi_Class=europeanterminator}', "");
Expect(1, 65511, '\P{Bidi_Class=europeanterminator}', "");
Expect(0, 65511, '\P{^Bidi_Class=europeanterminator}', "");
Expect(1, 65510, '\p{Bidi_Class:	 _european_Terminator}', "");
Expect(0, 65510, '\p{^Bidi_Class:	 _european_Terminator}', "");
Expect(0, 65510, '\P{Bidi_Class:	 _european_Terminator}', "");
Expect(1, 65510, '\P{^Bidi_Class:	 _european_Terminator}', "");
Expect(0, 65511, '\p{Bidi_Class:	 _european_Terminator}', "");
Expect(1, 65511, '\p{^Bidi_Class:	 _european_Terminator}', "");
Expect(1, 65511, '\P{Bidi_Class:	 _european_Terminator}', "");
Expect(0, 65511, '\P{^Bidi_Class:	 _european_Terminator}', "");
Error('\p{Bc=:=		ET}');
Error('\P{Bc=:=		ET}');
Expect(1, 65510, '\p{Bc: et}', "");
Expect(0, 65510, '\p{^Bc: et}', "");
Expect(0, 65510, '\P{Bc: et}', "");
Expect(1, 65510, '\P{^Bc: et}', "");
Expect(0, 65511, '\p{Bc: et}', "");
Expect(1, 65511, '\p{^Bc: et}', "");
Expect(1, 65511, '\P{Bc: et}', "");
Expect(0, 65511, '\P{^Bc: et}', "");
Expect(1, 65510, '\p{Bc=	ET}', "");
Expect(0, 65510, '\p{^Bc=	ET}', "");
Expect(0, 65510, '\P{Bc=	ET}', "");
Expect(1, 65510, '\P{^Bc=	ET}', "");
Expect(0, 65511, '\p{Bc=	ET}', "");
Expect(1, 65511, '\p{^Bc=	ET}', "");
Expect(1, 65511, '\P{Bc=	ET}', "");
Expect(0, 65511, '\P{^Bc=	ET}', "");
Error('\p{Is_Bidi_Class= _european_Terminator:=}');
Error('\P{Is_Bidi_Class= _european_Terminator:=}');
Expect(1, 65510, '\p{Is_Bidi_Class=europeanterminator}', "");
Expect(0, 65510, '\p{^Is_Bidi_Class=europeanterminator}', "");
Expect(0, 65510, '\P{Is_Bidi_Class=europeanterminator}', "");
Expect(1, 65510, '\P{^Is_Bidi_Class=europeanterminator}', "");
Expect(0, 65511, '\p{Is_Bidi_Class=europeanterminator}', "");
Expect(1, 65511, '\p{^Is_Bidi_Class=europeanterminator}', "");
Expect(1, 65511, '\P{Is_Bidi_Class=europeanterminator}', "");
Expect(0, 65511, '\P{^Is_Bidi_Class=europeanterminator}', "");
Expect(1, 65510, '\p{Is_Bidi_Class= _European_TERMINATOR}', "");
Expect(0, 65510, '\p{^Is_Bidi_Class= _European_TERMINATOR}', "");
Expect(0, 65510, '\P{Is_Bidi_Class= _European_TERMINATOR}', "");
Expect(1, 65510, '\P{^Is_Bidi_Class= _European_TERMINATOR}', "");
Expect(0, 65511, '\p{Is_Bidi_Class= _European_TERMINATOR}', "");
Expect(1, 65511, '\p{^Is_Bidi_Class= _European_TERMINATOR}', "");
Expect(1, 65511, '\P{Is_Bidi_Class= _European_TERMINATOR}', "");
Expect(0, 65511, '\P{^Is_Bidi_Class= _European_TERMINATOR}', "");
Error('\p{Is_Bc=- ET/a/}');
Error('\P{Is_Bc=- ET/a/}');
Expect(1, 65510, '\p{Is_Bc=et}', "");
Expect(0, 65510, '\p{^Is_Bc=et}', "");
Expect(0, 65510, '\P{Is_Bc=et}', "");
Expect(1, 65510, '\P{^Is_Bc=et}', "");
Expect(0, 65511, '\p{Is_Bc=et}', "");
Expect(1, 65511, '\p{^Is_Bc=et}', "");
Expect(1, 65511, '\P{Is_Bc=et}', "");
Expect(0, 65511, '\P{^Is_Bc=et}', "");
Expect(1, 65510, '\p{Is_Bc=_ET}', "");
Expect(0, 65510, '\p{^Is_Bc=_ET}', "");
Expect(0, 65510, '\P{Is_Bc=_ET}', "");
Expect(1, 65510, '\P{^Is_Bc=_ET}', "");
Expect(0, 65511, '\p{Is_Bc=_ET}', "");
Expect(1, 65511, '\p{^Is_Bc=_ET}', "");
Expect(1, 65511, '\P{Is_Bc=_ET}', "");
Expect(0, 65511, '\P{^Is_Bc=_ET}', "");
Error('\p{Bidi_Class=-	Arabic_NUMBER/a/}');
Error('\P{Bidi_Class=-	Arabic_NUMBER/a/}');
Expect(1, 69246, '\p{Bidi_Class=arabicnumber}', "");
Expect(0, 69246, '\p{^Bidi_Class=arabicnumber}', "");
Expect(0, 69246, '\P{Bidi_Class=arabicnumber}', "");
Expect(1, 69246, '\P{^Bidi_Class=arabicnumber}', "");
Expect(0, 69247, '\p{Bidi_Class=arabicnumber}', "");
Expect(1, 69247, '\p{^Bidi_Class=arabicnumber}', "");
Expect(1, 69247, '\P{Bidi_Class=arabicnumber}', "");
Expect(0, 69247, '\P{^Bidi_Class=arabicnumber}', "");
Expect(1, 69246, '\p{Bidi_Class= -Arabic_Number}', "");
Expect(0, 69246, '\p{^Bidi_Class= -Arabic_Number}', "");
Expect(0, 69246, '\P{Bidi_Class= -Arabic_Number}', "");
Expect(1, 69246, '\P{^Bidi_Class= -Arabic_Number}', "");
Expect(0, 69247, '\p{Bidi_Class= -Arabic_Number}', "");
Expect(1, 69247, '\p{^Bidi_Class= -Arabic_Number}', "");
Expect(1, 69247, '\P{Bidi_Class= -Arabic_Number}', "");
Expect(0, 69247, '\P{^Bidi_Class= -Arabic_Number}', "");
Error('\p{Bc=_	AN:=}');
Error('\P{Bc=_	AN:=}');
Expect(1, 69246, '\p{Bc=an}', "");
Expect(0, 69246, '\p{^Bc=an}', "");
Expect(0, 69246, '\P{Bc=an}', "");
Expect(1, 69246, '\P{^Bc=an}', "");
Expect(0, 69247, '\p{Bc=an}', "");
Expect(1, 69247, '\p{^Bc=an}', "");
Expect(1, 69247, '\P{Bc=an}', "");
Expect(0, 69247, '\P{^Bc=an}', "");
Expect(1, 69246, '\p{Bc=-AN}', "");
Expect(0, 69246, '\p{^Bc=-AN}', "");
Expect(0, 69246, '\P{Bc=-AN}', "");
Expect(1, 69246, '\P{^Bc=-AN}', "");
Expect(0, 69247, '\p{Bc=-AN}', "");
Expect(1, 69247, '\p{^Bc=-AN}', "");
Expect(1, 69247, '\P{Bc=-AN}', "");
Expect(0, 69247, '\P{^Bc=-AN}', "");
Error('\p{Is_Bidi_Class=Arabic_NUMBER:=}');
Error('\P{Is_Bidi_Class=Arabic_NUMBER:=}');
Expect(1, 69246, '\p{Is_Bidi_Class=arabicnumber}', "");
Expect(0, 69246, '\p{^Is_Bidi_Class=arabicnumber}', "");
Expect(0, 69246, '\P{Is_Bidi_Class=arabicnumber}', "");
Expect(1, 69246, '\P{^Is_Bidi_Class=arabicnumber}', "");
Expect(0, 69247, '\p{Is_Bidi_Class=arabicnumber}', "");
Expect(1, 69247, '\p{^Is_Bidi_Class=arabicnumber}', "");
Expect(1, 69247, '\P{Is_Bidi_Class=arabicnumber}', "");
Expect(0, 69247, '\P{^Is_Bidi_Class=arabicnumber}', "");
Expect(1, 69246, '\p{Is_Bidi_Class=_-ARABIC_number}', "");
Expect(0, 69246, '\p{^Is_Bidi_Class=_-ARABIC_number}', "");
Expect(0, 69246, '\P{Is_Bidi_Class=_-ARABIC_number}', "");
Expect(1, 69246, '\P{^Is_Bidi_Class=_-ARABIC_number}', "");
Expect(0, 69247, '\p{Is_Bidi_Class=_-ARABIC_number}', "");
Expect(1, 69247, '\p{^Is_Bidi_Class=_-ARABIC_number}', "");
Expect(1, 69247, '\P{Is_Bidi_Class=_-ARABIC_number}', "");
Expect(0, 69247, '\P{^Is_Bidi_Class=_-ARABIC_number}', "");
Error('\p{Is_Bc=_:=AN}');
Error('\P{Is_Bc=_:=AN}');
Expect(1, 69246, '\p{Is_Bc=an}', "");
Expect(0, 69246, '\p{^Is_Bc=an}', "");
Expect(0, 69246, '\P{Is_Bc=an}', "");
Expect(1, 69246, '\P{^Is_Bc=an}', "");
Expect(0, 69247, '\p{Is_Bc=an}', "");
Expect(1, 69247, '\p{^Is_Bc=an}', "");
Expect(1, 69247, '\P{Is_Bc=an}', "");
Expect(0, 69247, '\P{^Is_Bc=an}', "");
Expect(1, 69246, '\p{Is_Bc=-an}', "");
Expect(0, 69246, '\p{^Is_Bc=-an}', "");
Expect(0, 69246, '\P{Is_Bc=-an}', "");
Expect(1, 69246, '\P{^Is_Bc=-an}', "");
Expect(0, 69247, '\p{Is_Bc=-an}', "");
Expect(1, 69247, '\p{^Is_Bc=-an}', "");
Expect(1, 69247, '\P{Is_Bc=-an}', "");
Expect(0, 69247, '\P{^Is_Bc=-an}', "");
Error('\p{Bidi_Class=:=_RIGHT_TO_Left}');
Error('\P{Bidi_Class=:=_RIGHT_TO_Left}');
Expect(1, 126975, '\p{Bidi_Class=righttoleft}', "");
Expect(0, 126975, '\p{^Bidi_Class=righttoleft}', "");
Expect(0, 126975, '\P{Bidi_Class=righttoleft}', "");
Expect(1, 126975, '\P{^Bidi_Class=righttoleft}', "");
Expect(0, 126976, '\p{Bidi_Class=righttoleft}', "");
Expect(1, 126976, '\p{^Bidi_Class=righttoleft}', "");
Expect(1, 126976, '\P{Bidi_Class=righttoleft}', "");
Expect(0, 126976, '\P{^Bidi_Class=righttoleft}', "");
Expect(1, 126975, '\p{Bidi_Class=__Right_To_Left}', "");
Expect(0, 126975, '\p{^Bidi_Class=__Right_To_Left}', "");
Expect(0, 126975, '\P{Bidi_Class=__Right_To_Left}', "");
Expect(1, 126975, '\P{^Bidi_Class=__Right_To_Left}', "");
Expect(0, 126976, '\p{Bidi_Class=__Right_To_Left}', "");
Expect(1, 126976, '\p{^Bidi_Class=__Right_To_Left}', "");
Expect(1, 126976, '\P{Bidi_Class=__Right_To_Left}', "");
Expect(0, 126976, '\P{^Bidi_Class=__Right_To_Left}', "");
Error('\p{Bc=_ R/a/}');
Error('\P{Bc=_ R/a/}');
Expect(1, 126975, '\p{Bc=r}', "");
Expect(0, 126975, '\p{^Bc=r}', "");
Expect(0, 126975, '\P{Bc=r}', "");
Expect(1, 126975, '\P{^Bc=r}', "");
Expect(0, 126976, '\p{Bc=r}', "");
Expect(1, 126976, '\p{^Bc=r}', "");
Expect(1, 126976, '\P{Bc=r}', "");
Expect(0, 126976, '\P{^Bc=r}', "");
Expect(1, 126975, '\p{Bc:   	R}', "");
Expect(0, 126975, '\p{^Bc:   	R}', "");
Expect(0, 126975, '\P{Bc:   	R}', "");
Expect(1, 126975, '\P{^Bc:   	R}', "");
Expect(0, 126976, '\p{Bc:   	R}', "");
Expect(1, 126976, '\p{^Bc:   	R}', "");
Expect(1, 126976, '\P{Bc:   	R}', "");
Expect(0, 126976, '\P{^Bc:   	R}', "");
Error('\p{Is_Bidi_Class=:= -Right_To_LEFT}');
Error('\P{Is_Bidi_Class=:= -Right_To_LEFT}');
Expect(1, 126975, '\p{Is_Bidi_Class=righttoleft}', "");
Expect(0, 126975, '\p{^Is_Bidi_Class=righttoleft}', "");
Expect(0, 126975, '\P{Is_Bidi_Class=righttoleft}', "");
Expect(1, 126975, '\P{^Is_Bidi_Class=righttoleft}', "");
Expect(0, 126976, '\p{Is_Bidi_Class=righttoleft}', "");
Expect(1, 126976, '\p{^Is_Bidi_Class=righttoleft}', "");
Expect(1, 126976, '\P{Is_Bidi_Class=righttoleft}', "");
Expect(0, 126976, '\P{^Is_Bidi_Class=righttoleft}', "");
Expect(1, 126975, '\p{Is_Bidi_Class=-_Right_to_Left}', "");
Expect(0, 126975, '\p{^Is_Bidi_Class=-_Right_to_Left}', "");
Expect(0, 126975, '\P{Is_Bidi_Class=-_Right_to_Left}', "");
Expect(1, 126975, '\P{^Is_Bidi_Class=-_Right_to_Left}', "");
Expect(0, 126976, '\p{Is_Bidi_Class=-_Right_to_Left}', "");
Expect(1, 126976, '\p{^Is_Bidi_Class=-_Right_to_Left}', "");
Expect(1, 126976, '\P{Is_Bidi_Class=-_Right_to_Left}', "");
Expect(0, 126976, '\P{^Is_Bidi_Class=-_Right_to_Left}', "");
Error('\p{Is_Bc=:=_ R}');
Error('\P{Is_Bc=:=_ R}');
Expect(1, 126975, '\p{Is_Bc=r}', "");
Expect(0, 126975, '\p{^Is_Bc=r}', "");
Expect(0, 126975, '\P{Is_Bc=r}', "");
Expect(1, 126975, '\P{^Is_Bc=r}', "");
Expect(0, 126976, '\p{Is_Bc=r}', "");
Expect(1, 126976, '\p{^Is_Bc=r}', "");
Expect(1, 126976, '\P{Is_Bc=r}', "");
Expect(0, 126976, '\P{^Is_Bc=r}', "");
Expect(1, 126975, '\p{Is_Bc=-_R}', "");
Expect(0, 126975, '\p{^Is_Bc=-_R}', "");
Expect(0, 126975, '\P{Is_Bc=-_R}', "");
Expect(1, 126975, '\P{^Is_Bc=-_R}', "");
Expect(0, 126976, '\p{Is_Bc=-_R}', "");
Expect(1, 126976, '\p{^Is_Bc=-_R}', "");
Expect(1, 126976, '\P{Is_Bc=-_R}', "");
Expect(0, 126976, '\P{^Is_Bc=-_R}', "");
Error('\p{Expands_On_NFD=Yes}');
Error('\P{Expands_On_NFD=Yes}');
Error('\p{XO_NFD: Y}');
Error('\P{XO_NFD: Y}');
Error('\p{Is_Expands_On_NFD=T}');
Error('\P{Is_Expands_On_NFD=T}');
Error('\p{Is_XO_NFD=True}');
Error('\P{Is_XO_NFD=True}');
Error('\p{Expands_On_NFD=No}');
Error('\P{Expands_On_NFD=No}');
Error('\p{XO_NFD=N}');
Error('\P{XO_NFD=N}');
Error('\p{Is_Expands_On_NFD=F}');
Error('\P{Is_Expands_On_NFD=F}');
Error('\p{Is_XO_NFD=False}');
Error('\P{Is_XO_NFD=False}');
Error('\p{Other_ID_Start=Yes}');
Error('\P{Other_ID_Start=Yes}');
Error('\p{OIDS=Y}');
Error('\P{OIDS=Y}');
Error('\p{Is_Other_ID_Start=T}');
Error('\P{Is_Other_ID_Start=T}');
Error('\p{Is_OIDS=True}');
Error('\P{Is_OIDS=True}');
Error('\p{Other_ID_Start=No}');
Error('\P{Other_ID_Start=No}');
Error('\p{OIDS=N}');
Error('\P{OIDS=N}');
Error('\p{Is_Other_ID_Start=F}');
Error('\P{Is_Other_ID_Start=F}');
Error('\p{Is_OIDS=False}');
Error('\P{Is_OIDS=False}');
Error('\p{Decomposition_Type=/a/ _FRACTION}');
Error('\P{Decomposition_Type=/a/ _FRACTION}');
Expect(1, 8585, '\p{Decomposition_Type=fraction}', "");
Expect(0, 8585, '\p{^Decomposition_Type=fraction}', "");
Expect(0, 8585, '\P{Decomposition_Type=fraction}', "");
Expect(1, 8585, '\P{^Decomposition_Type=fraction}', "");
Expect(0, 8586, '\p{Decomposition_Type=fraction}', "");
Expect(1, 8586, '\p{^Decomposition_Type=fraction}', "");
Expect(1, 8586, '\P{Decomposition_Type=fraction}', "");
Expect(0, 8586, '\P{^Decomposition_Type=fraction}', "");
Error('\p{Dt=:=	Fra}');
Error('\P{Dt=:=	Fra}');
Expect(1, 8585, '\p{Dt: fra}', "");
Expect(0, 8585, '\p{^Dt: fra}', "");
Expect(0, 8585, '\P{Dt: fra}', "");
Expect(1, 8585, '\P{^Dt: fra}', "");
Expect(0, 8586, '\p{Dt: fra}', "");
Expect(1, 8586, '\p{^Dt: fra}', "");
Expect(1, 8586, '\P{Dt: fra}', "");
Expect(0, 8586, '\P{^Dt: fra}', "");
Expect(1, 8585, '\p{Dt=	Fra}', "");
Expect(0, 8585, '\p{^Dt=	Fra}', "");
Expect(0, 8585, '\P{Dt=	Fra}', "");
Expect(1, 8585, '\P{^Dt=	Fra}', "");
Expect(0, 8586, '\p{Dt=	Fra}', "");
Expect(1, 8586, '\p{^Dt=	Fra}', "");
Expect(1, 8586, '\P{Dt=	Fra}', "");
Expect(0, 8586, '\P{^Dt=	Fra}', "");
Error('\p{Is_Decomposition_Type=:= -Fraction}');
Error('\P{Is_Decomposition_Type=:= -Fraction}');
Expect(1, 8585, '\p{Is_Decomposition_Type=fraction}', "");
Expect(0, 8585, '\p{^Is_Decomposition_Type=fraction}', "");
Expect(0, 8585, '\P{Is_Decomposition_Type=fraction}', "");
Expect(1, 8585, '\P{^Is_Decomposition_Type=fraction}', "");
Expect(0, 8586, '\p{Is_Decomposition_Type=fraction}', "");
Expect(1, 8586, '\p{^Is_Decomposition_Type=fraction}', "");
Expect(1, 8586, '\P{Is_Decomposition_Type=fraction}', "");
Expect(0, 8586, '\P{^Is_Decomposition_Type=fraction}', "");
Expect(1, 8585, '\p{Is_Decomposition_Type= 	FRACTION}', "");
Expect(0, 8585, '\p{^Is_Decomposition_Type= 	FRACTION}', "");
Expect(0, 8585, '\P{Is_Decomposition_Type= 	FRACTION}', "");
Expect(1, 8585, '\P{^Is_Decomposition_Type= 	FRACTION}', "");
Expect(0, 8586, '\p{Is_Decomposition_Type= 	FRACTION}', "");
Expect(1, 8586, '\p{^Is_Decomposition_Type= 	FRACTION}', "");
Expect(1, 8586, '\P{Is_Decomposition_Type= 	FRACTION}', "");
Expect(0, 8586, '\P{^Is_Decomposition_Type= 	FRACTION}', "");
Error('\p{Is_Dt=/a/  fra}');
Error('\P{Is_Dt=/a/  fra}');
Expect(1, 8585, '\p{Is_Dt=fra}', "");
Expect(0, 8585, '\p{^Is_Dt=fra}', "");
Expect(0, 8585, '\P{Is_Dt=fra}', "");
Expect(1, 8585, '\P{^Is_Dt=fra}', "");
Expect(0, 8586, '\p{Is_Dt=fra}', "");
Expect(1, 8586, '\p{^Is_Dt=fra}', "");
Expect(1, 8586, '\P{Is_Dt=fra}', "");
Expect(0, 8586, '\P{^Is_Dt=fra}', "");
Expect(1, 8585, '\p{Is_Dt=	 fra}', "");
Expect(0, 8585, '\p{^Is_Dt=	 fra}', "");
Expect(0, 8585, '\P{Is_Dt=	 fra}', "");
Expect(1, 8585, '\P{^Is_Dt=	 fra}', "");
Expect(0, 8586, '\p{Is_Dt=	 fra}', "");
Expect(1, 8586, '\p{^Is_Dt=	 fra}', "");
Expect(1, 8586, '\P{Is_Dt=	 fra}', "");
Expect(0, 8586, '\P{^Is_Dt=	 fra}', "");
Error('\p{Decomposition_Type=--Circle/a/}');
Error('\P{Decomposition_Type=--Circle/a/}');
Expect(1, 127569, '\p{Decomposition_Type=circle}', "");
Expect(0, 127569, '\p{^Decomposition_Type=circle}', "");
Expect(0, 127569, '\P{Decomposition_Type=circle}', "");
Expect(1, 127569, '\P{^Decomposition_Type=circle}', "");
Expect(0, 127570, '\p{Decomposition_Type=circle}', "");
Expect(1, 127570, '\p{^Decomposition_Type=circle}', "");
Expect(1, 127570, '\P{Decomposition_Type=circle}', "");
Expect(0, 127570, '\P{^Decomposition_Type=circle}', "");
Expect(1, 127569, '\p{Decomposition_Type=		circle}', "");
Expect(0, 127569, '\p{^Decomposition_Type=		circle}', "");
Expect(0, 127569, '\P{Decomposition_Type=		circle}', "");
Expect(1, 127569, '\P{^Decomposition_Type=		circle}', "");
Expect(0, 127570, '\p{Decomposition_Type=		circle}', "");
Expect(1, 127570, '\p{^Decomposition_Type=		circle}', "");
Expect(1, 127570, '\P{Decomposition_Type=		circle}', "");
Expect(0, 127570, '\P{^Decomposition_Type=		circle}', "");
Error('\p{Dt:	/a/_	enc}');
Error('\P{Dt:	/a/_	enc}');
Expect(1, 127569, '\p{Dt:   enc}', "");
Expect(0, 127569, '\p{^Dt:   enc}', "");
Expect(0, 127569, '\P{Dt:   enc}', "");
Expect(1, 127569, '\P{^Dt:   enc}', "");
Expect(0, 127570, '\p{Dt:   enc}', "");
Expect(1, 127570, '\p{^Dt:   enc}', "");
Expect(1, 127570, '\P{Dt:   enc}', "");
Expect(0, 127570, '\P{^Dt:   enc}', "");
Expect(1, 127569, '\p{Dt=- Enc}', "");
Expect(0, 127569, '\p{^Dt=- Enc}', "");
Expect(0, 127569, '\P{Dt=- Enc}', "");
Expect(1, 127569, '\P{^Dt=- Enc}', "");
Expect(0, 127570, '\p{Dt=- Enc}', "");
Expect(1, 127570, '\p{^Dt=- Enc}', "");
Expect(1, 127570, '\P{Dt=- Enc}', "");
Expect(0, 127570, '\P{^Dt=- Enc}', "");
Error('\p{Is_Decomposition_Type=:=		Circle}');
Error('\P{Is_Decomposition_Type=:=		Circle}');
Expect(1, 127569, '\p{Is_Decomposition_Type=circle}', "");
Expect(0, 127569, '\p{^Is_Decomposition_Type=circle}', "");
Expect(0, 127569, '\P{Is_Decomposition_Type=circle}', "");
Expect(1, 127569, '\P{^Is_Decomposition_Type=circle}', "");
Expect(0, 127570, '\p{Is_Decomposition_Type=circle}', "");
Expect(1, 127570, '\p{^Is_Decomposition_Type=circle}', "");
Expect(1, 127570, '\P{Is_Decomposition_Type=circle}', "");
Expect(0, 127570, '\P{^Is_Decomposition_Type=circle}', "");
Expect(1, 127569, '\p{Is_Decomposition_Type= Circle}', "");
Expect(0, 127569, '\p{^Is_Decomposition_Type= Circle}', "");
Expect(0, 127569, '\P{Is_Decomposition_Type= Circle}', "");
Expect(1, 127569, '\P{^Is_Decomposition_Type= Circle}', "");
Expect(0, 127570, '\p{Is_Decomposition_Type= Circle}', "");
Expect(1, 127570, '\p{^Is_Decomposition_Type= Circle}', "");
Expect(1, 127570, '\P{Is_Decomposition_Type= Circle}', "");
Expect(0, 127570, '\P{^Is_Decomposition_Type= Circle}', "");
Error('\p{Is_Dt=_/a/Enc}');
Error('\P{Is_Dt=_/a/Enc}');
Expect(1, 127569, '\p{Is_Dt: enc}', "");
Expect(0, 127569, '\p{^Is_Dt: enc}', "");
Expect(0, 127569, '\P{Is_Dt: enc}', "");
Expect(1, 127569, '\P{^Is_Dt: enc}', "");
Expect(0, 127570, '\p{Is_Dt: enc}', "");
Expect(1, 127570, '\p{^Is_Dt: enc}', "");
Expect(1, 127570, '\P{Is_Dt: enc}', "");
Expect(0, 127570, '\P{^Is_Dt: enc}', "");
Expect(1, 127569, '\p{Is_Dt=	Enc}', "");
Expect(0, 127569, '\p{^Is_Dt=	Enc}', "");
Expect(0, 127569, '\P{Is_Dt=	Enc}', "");
Expect(1, 127569, '\P{^Is_Dt=	Enc}', "");
Expect(0, 127570, '\p{Is_Dt=	Enc}', "");
Expect(1, 127570, '\p{^Is_Dt=	Enc}', "");
Expect(1, 127570, '\P{Is_Dt=	Enc}', "");
Expect(0, 127570, '\P{^Is_Dt=	Enc}', "");
Error('\p{Decomposition_Type= Initial/a/}');
Error('\P{Decomposition_Type= Initial/a/}');
Expect(1, 65267, '\p{Decomposition_Type=initial}', "");
Expect(0, 65267, '\p{^Decomposition_Type=initial}', "");
Expect(0, 65267, '\P{Decomposition_Type=initial}', "");
Expect(1, 65267, '\P{^Decomposition_Type=initial}', "");
Expect(0, 65268, '\p{Decomposition_Type=initial}', "");
Expect(1, 65268, '\p{^Decomposition_Type=initial}', "");
Expect(1, 65268, '\P{Decomposition_Type=initial}', "");
Expect(0, 65268, '\P{^Decomposition_Type=initial}', "");
Expect(1, 65267, '\p{Decomposition_Type:		 Initial}', "");
Expect(0, 65267, '\p{^Decomposition_Type:		 Initial}', "");
Expect(0, 65267, '\P{Decomposition_Type:		 Initial}', "");
Expect(1, 65267, '\P{^Decomposition_Type:		 Initial}', "");
Expect(0, 65268, '\p{Decomposition_Type:		 Initial}', "");
Expect(1, 65268, '\p{^Decomposition_Type:		 Initial}', "");
Expect(1, 65268, '\P{Decomposition_Type:		 Initial}', "");
Expect(0, 65268, '\P{^Decomposition_Type:		 Initial}', "");
Error('\p{Dt::= 	Init}');
Error('\P{Dt::= 	Init}');
Expect(1, 65267, '\p{Dt=init}', "");
Expect(0, 65267, '\p{^Dt=init}', "");
Expect(0, 65267, '\P{Dt=init}', "");
Expect(1, 65267, '\P{^Dt=init}', "");
Expect(0, 65268, '\p{Dt=init}', "");
Expect(1, 65268, '\p{^Dt=init}', "");
Expect(1, 65268, '\P{Dt=init}', "");
Expect(0, 65268, '\P{^Dt=init}', "");
Expect(1, 65267, '\p{Dt=-	init}', "");
Expect(0, 65267, '\p{^Dt=-	init}', "");
Expect(0, 65267, '\P{Dt=-	init}', "");
Expect(1, 65267, '\P{^Dt=-	init}', "");
Expect(0, 65268, '\p{Dt=-	init}', "");
Expect(1, 65268, '\p{^Dt=-	init}', "");
Expect(1, 65268, '\P{Dt=-	init}', "");
Expect(0, 65268, '\P{^Dt=-	init}', "");
Error('\p{Is_Decomposition_Type=:=  INITIAL}');
Error('\P{Is_Decomposition_Type=:=  INITIAL}');
Expect(1, 65267, '\p{Is_Decomposition_Type=initial}', "");
Expect(0, 65267, '\p{^Is_Decomposition_Type=initial}', "");
Expect(0, 65267, '\P{Is_Decomposition_Type=initial}', "");
Expect(1, 65267, '\P{^Is_Decomposition_Type=initial}', "");
Expect(0, 65268, '\p{Is_Decomposition_Type=initial}', "");
Expect(1, 65268, '\p{^Is_Decomposition_Type=initial}', "");
Expect(1, 65268, '\P{Is_Decomposition_Type=initial}', "");
Expect(0, 65268, '\P{^Is_Decomposition_Type=initial}', "");
Expect(1, 65267, '\p{Is_Decomposition_Type=-_initial}', "");
Expect(0, 65267, '\p{^Is_Decomposition_Type=-_initial}', "");
Expect(0, 65267, '\P{Is_Decomposition_Type=-_initial}', "");
Expect(1, 65267, '\P{^Is_Decomposition_Type=-_initial}', "");
Expect(0, 65268, '\p{Is_Decomposition_Type=-_initial}', "");
Expect(1, 65268, '\p{^Is_Decomposition_Type=-_initial}', "");
Expect(1, 65268, '\P{Is_Decomposition_Type=-_initial}', "");
Expect(0, 65268, '\P{^Is_Decomposition_Type=-_initial}', "");
Error('\p{Is_Dt= /a/Init}');
Error('\P{Is_Dt= /a/Init}');
Expect(1, 65267, '\p{Is_Dt=init}', "");
Expect(0, 65267, '\p{^Is_Dt=init}', "");
Expect(0, 65267, '\P{Is_Dt=init}', "");
Expect(1, 65267, '\P{^Is_Dt=init}', "");
Expect(0, 65268, '\p{Is_Dt=init}', "");
Expect(1, 65268, '\p{^Is_Dt=init}', "");
Expect(1, 65268, '\P{Is_Dt=init}', "");
Expect(0, 65268, '\P{^Is_Dt=init}', "");
Expect(1, 65267, '\p{Is_Dt=_Init}', "");
Expect(0, 65267, '\p{^Is_Dt=_Init}', "");
Expect(0, 65267, '\P{Is_Dt=_Init}', "");
Expect(1, 65267, '\P{^Is_Dt=_Init}', "");
Expect(0, 65268, '\p{Is_Dt=_Init}', "");
Expect(1, 65268, '\p{^Is_Dt=_Init}', "");
Expect(1, 65268, '\P{Is_Dt=_Init}', "");
Expect(0, 65268, '\P{^Is_Dt=_Init}', "");
Error('\p{Decomposition_Type=	-Font/a/}');
Error('\P{Decomposition_Type=	-Font/a/}');
Expect(1, 126651, '\p{Decomposition_Type=font}', "");
Expect(0, 126651, '\p{^Decomposition_Type=font}', "");
Expect(0, 126651, '\P{Decomposition_Type=font}', "");
Expect(1, 126651, '\P{^Decomposition_Type=font}', "");
Expect(0, 126652, '\p{Decomposition_Type=font}', "");
Expect(1, 126652, '\p{^Decomposition_Type=font}', "");
Expect(1, 126652, '\P{Decomposition_Type=font}', "");
Expect(0, 126652, '\P{^Decomposition_Type=font}', "");
Expect(1, 126651, '\p{Decomposition_Type=	_FONT}', "");
Expect(0, 126651, '\p{^Decomposition_Type=	_FONT}', "");
Expect(0, 126651, '\P{Decomposition_Type=	_FONT}', "");
Expect(1, 126651, '\P{^Decomposition_Type=	_FONT}', "");
Expect(0, 126652, '\p{Decomposition_Type=	_FONT}', "");
Expect(1, 126652, '\p{^Decomposition_Type=	_FONT}', "");
Expect(1, 126652, '\P{Decomposition_Type=	_FONT}', "");
Expect(0, 126652, '\P{^Decomposition_Type=	_FONT}', "");
Error('\p{Dt:	 Font:=}');
Error('\P{Dt:	 Font:=}');
Expect(1, 126651, '\p{Dt=font}', "");
Expect(0, 126651, '\p{^Dt=font}', "");
Expect(0, 126651, '\P{Dt=font}', "");
Expect(1, 126651, '\P{^Dt=font}', "");
Expect(0, 126652, '\p{Dt=font}', "");
Expect(1, 126652, '\p{^Dt=font}', "");
Expect(1, 126652, '\P{Dt=font}', "");
Expect(0, 126652, '\P{^Dt=font}', "");
Expect(1, 126651, '\p{Dt=	-Font}', "");
Expect(0, 126651, '\p{^Dt=	-Font}', "");
Expect(0, 126651, '\P{Dt=	-Font}', "");
Expect(1, 126651, '\P{^Dt=	-Font}', "");
Expect(0, 126652, '\p{Dt=	-Font}', "");
Expect(1, 126652, '\p{^Dt=	-Font}', "");
Expect(1, 126652, '\P{Dt=	-Font}', "");
Expect(0, 126652, '\P{^Dt=	-Font}', "");
Error('\p{Is_Decomposition_Type= -FONT/a/}');
Error('\P{Is_Decomposition_Type= -FONT/a/}');
Expect(1, 126651, '\p{Is_Decomposition_Type=font}', "");
Expect(0, 126651, '\p{^Is_Decomposition_Type=font}', "");
Expect(0, 126651, '\P{Is_Decomposition_Type=font}', "");
Expect(1, 126651, '\P{^Is_Decomposition_Type=font}', "");
Expect(0, 126652, '\p{Is_Decomposition_Type=font}', "");
Expect(1, 126652, '\p{^Is_Decomposition_Type=font}', "");
Expect(1, 126652, '\P{Is_Decomposition_Type=font}', "");
Expect(0, 126652, '\P{^Is_Decomposition_Type=font}', "");
Expect(1, 126651, '\p{Is_Decomposition_Type=_font}', "");
Expect(0, 126651, '\p{^Is_Decomposition_Type=_font}', "");
Expect(0, 126651, '\P{Is_Decomposition_Type=_font}', "");
Expect(1, 126651, '\P{^Is_Decomposition_Type=_font}', "");
Expect(0, 126652, '\p{Is_Decomposition_Type=_font}', "");
Expect(1, 126652, '\p{^Is_Decomposition_Type=_font}', "");
Expect(1, 126652, '\P{Is_Decomposition_Type=_font}', "");
Expect(0, 126652, '\P{^Is_Decomposition_Type=_font}', "");
Error('\p{Is_Dt=	font:=}');
Error('\P{Is_Dt=	font:=}');
Expect(1, 126651, '\p{Is_Dt=font}', "");
Expect(0, 126651, '\p{^Is_Dt=font}', "");
Expect(0, 126651, '\P{Is_Dt=font}', "");
Expect(1, 126651, '\P{^Is_Dt=font}', "");
Expect(0, 126652, '\p{Is_Dt=font}', "");
Expect(1, 126652, '\p{^Is_Dt=font}', "");
Expect(1, 126652, '\P{Is_Dt=font}', "");
Expect(0, 126652, '\P{^Is_Dt=font}', "");
Expect(1, 126651, '\p{Is_Dt= _font}', "");
Expect(0, 126651, '\p{^Is_Dt= _font}', "");
Expect(0, 126651, '\P{Is_Dt= _font}', "");
Expect(1, 126651, '\P{^Is_Dt= _font}', "");
Expect(0, 126652, '\p{Is_Dt= _font}', "");
Expect(1, 126652, '\p{^Is_Dt= _font}', "");
Expect(1, 126652, '\P{Is_Dt= _font}', "");
Expect(0, 126652, '\P{^Is_Dt= _font}', "");
Error('\p{Decomposition_Type= NOBREAK/a/}');
Error('\P{Decomposition_Type= NOBREAK/a/}');
Expect(1, 8239, '\p{Decomposition_Type=nobreak}', "");
Expect(0, 8239, '\p{^Decomposition_Type=nobreak}', "");
Expect(0, 8239, '\P{Decomposition_Type=nobreak}', "");
Expect(1, 8239, '\P{^Decomposition_Type=nobreak}', "");
Expect(0, 8240, '\p{Decomposition_Type=nobreak}', "");
Expect(1, 8240, '\p{^Decomposition_Type=nobreak}', "");
Expect(1, 8240, '\P{Decomposition_Type=nobreak}', "");
Expect(0, 8240, '\P{^Decomposition_Type=nobreak}', "");
Expect(1, 8239, '\p{Decomposition_Type=Nobreak}', "");
Expect(0, 8239, '\p{^Decomposition_Type=Nobreak}', "");
Expect(0, 8239, '\P{Decomposition_Type=Nobreak}', "");
Expect(1, 8239, '\P{^Decomposition_Type=Nobreak}', "");
Expect(0, 8240, '\p{Decomposition_Type=Nobreak}', "");
Expect(1, 8240, '\p{^Decomposition_Type=Nobreak}', "");
Expect(1, 8240, '\P{Decomposition_Type=Nobreak}', "");
Expect(0, 8240, '\P{^Decomposition_Type=Nobreak}', "");
Error('\p{Dt=_NB/a/}');
Error('\P{Dt=_NB/a/}');
Expect(1, 8239, '\p{Dt=nb}', "");
Expect(0, 8239, '\p{^Dt=nb}', "");
Expect(0, 8239, '\P{Dt=nb}', "");
Expect(1, 8239, '\P{^Dt=nb}', "");
Expect(0, 8240, '\p{Dt=nb}', "");
Expect(1, 8240, '\p{^Dt=nb}', "");
Expect(1, 8240, '\P{Dt=nb}', "");
Expect(0, 8240, '\P{^Dt=nb}', "");
Expect(1, 8239, '\p{Dt=_NB}', "");
Expect(0, 8239, '\p{^Dt=_NB}', "");
Expect(0, 8239, '\P{Dt=_NB}', "");
Expect(1, 8239, '\P{^Dt=_NB}', "");
Expect(0, 8240, '\p{Dt=_NB}', "");
Expect(1, 8240, '\p{^Dt=_NB}', "");
Expect(1, 8240, '\P{Dt=_NB}', "");
Expect(0, 8240, '\P{^Dt=_NB}', "");
Error('\p{Is_Decomposition_Type= _NOBREAK/a/}');
Error('\P{Is_Decomposition_Type= _NOBREAK/a/}');
Expect(1, 8239, '\p{Is_Decomposition_Type=nobreak}', "");
Expect(0, 8239, '\p{^Is_Decomposition_Type=nobreak}', "");
Expect(0, 8239, '\P{Is_Decomposition_Type=nobreak}', "");
Expect(1, 8239, '\P{^Is_Decomposition_Type=nobreak}', "");
Expect(0, 8240, '\p{Is_Decomposition_Type=nobreak}', "");
Expect(1, 8240, '\p{^Is_Decomposition_Type=nobreak}', "");
Expect(1, 8240, '\P{Is_Decomposition_Type=nobreak}', "");
Expect(0, 8240, '\P{^Is_Decomposition_Type=nobreak}', "");
Expect(1, 8239, '\p{Is_Decomposition_Type: Nobreak}', "");
Expect(0, 8239, '\p{^Is_Decomposition_Type: Nobreak}', "");
Expect(0, 8239, '\P{Is_Decomposition_Type: Nobreak}', "");
Expect(1, 8239, '\P{^Is_Decomposition_Type: Nobreak}', "");
Expect(0, 8240, '\p{Is_Decomposition_Type: Nobreak}', "");
Expect(1, 8240, '\p{^Is_Decomposition_Type: Nobreak}', "");
Expect(1, 8240, '\P{Is_Decomposition_Type: Nobreak}', "");
Expect(0, 8240, '\P{^Is_Decomposition_Type: Nobreak}', "");
Error('\p{Is_Dt=:= -Nb}');
Error('\P{Is_Dt=:= -Nb}');
Expect(1, 8239, '\p{Is_Dt=nb}', "");
Expect(0, 8239, '\p{^Is_Dt=nb}', "");
Expect(0, 8239, '\P{Is_Dt=nb}', "");
Expect(1, 8239, '\P{^Is_Dt=nb}', "");
Expect(0, 8240, '\p{Is_Dt=nb}', "");
Expect(1, 8240, '\p{^Is_Dt=nb}', "");
Expect(1, 8240, '\P{Is_Dt=nb}', "");
Expect(0, 8240, '\P{^Is_Dt=nb}', "");
Expect(1, 8239, '\p{Is_Dt=-nb}', "");
Expect(0, 8239, '\p{^Is_Dt=-nb}', "");
Expect(0, 8239, '\P{Is_Dt=-nb}', "");
Expect(1, 8239, '\P{^Is_Dt=-nb}', "");
Expect(0, 8240, '\p{Is_Dt=-nb}', "");
Expect(1, 8240, '\p{^Is_Dt=-nb}', "");
Expect(1, 8240, '\P{Is_Dt=-nb}', "");
Expect(0, 8240, '\P{^Is_Dt=-nb}', "");
Error('\p{Decomposition_Type=	_Non_canonical:=}');
Error('\P{Decomposition_Type=	_Non_canonical:=}');
Expect(1, 127569, '\p{Decomposition_Type=noncanonical}', "");
Expect(0, 127569, '\p{^Decomposition_Type=noncanonical}', "");
Expect(0, 127569, '\P{Decomposition_Type=noncanonical}', "");
Expect(1, 127569, '\P{^Decomposition_Type=noncanonical}', "");
Expect(0, 127570, '\p{Decomposition_Type=noncanonical}', "");
Expect(1, 127570, '\p{^Decomposition_Type=noncanonical}', "");
Expect(1, 127570, '\P{Decomposition_Type=noncanonical}', "");
Expect(0, 127570, '\P{^Decomposition_Type=noncanonical}', "");
Expect(1, 127569, '\p{Decomposition_Type= 	non_canonical}', "");
Expect(0, 127569, '\p{^Decomposition_Type= 	non_canonical}', "");
Expect(0, 127569, '\P{Decomposition_Type= 	non_canonical}', "");
Expect(1, 127569, '\P{^Decomposition_Type= 	non_canonical}', "");
Expect(0, 127570, '\p{Decomposition_Type= 	non_canonical}', "");
Expect(1, 127570, '\p{^Decomposition_Type= 	non_canonical}', "");
Expect(1, 127570, '\P{Decomposition_Type= 	non_canonical}', "");
Expect(0, 127570, '\P{^Decomposition_Type= 	non_canonical}', "");
Error('\p{Dt= NON_canon:=}');
Error('\P{Dt= NON_canon:=}');
Expect(1, 127569, '\p{Dt=noncanon}', "");
Expect(0, 127569, '\p{^Dt=noncanon}', "");
Expect(0, 127569, '\P{Dt=noncanon}', "");
Expect(1, 127569, '\P{^Dt=noncanon}', "");
Expect(0, 127570, '\p{Dt=noncanon}', "");
Expect(1, 127570, '\p{^Dt=noncanon}', "");
Expect(1, 127570, '\P{Dt=noncanon}', "");
Expect(0, 127570, '\P{^Dt=noncanon}', "");
Expect(1, 127569, '\p{Dt=	_NON_canon}', "");
Expect(0, 127569, '\p{^Dt=	_NON_canon}', "");
Expect(0, 127569, '\P{Dt=	_NON_canon}', "");
Expect(1, 127569, '\P{^Dt=	_NON_canon}', "");
Expect(0, 127570, '\p{Dt=	_NON_canon}', "");
Expect(1, 127570, '\p{^Dt=	_NON_canon}', "");
Expect(1, 127570, '\P{Dt=	_NON_canon}', "");
Expect(0, 127570, '\P{^Dt=	_NON_canon}', "");
Error('\p{Is_Decomposition_Type=-non_Canonical:=}');
Error('\P{Is_Decomposition_Type=-non_Canonical:=}');
Expect(1, 127569, '\p{Is_Decomposition_Type:	noncanonical}', "");
Expect(0, 127569, '\p{^Is_Decomposition_Type:	noncanonical}', "");
Expect(0, 127569, '\P{Is_Decomposition_Type:	noncanonical}', "");
Expect(1, 127569, '\P{^Is_Decomposition_Type:	noncanonical}', "");
Expect(0, 127570, '\p{Is_Decomposition_Type:	noncanonical}', "");
Expect(1, 127570, '\p{^Is_Decomposition_Type:	noncanonical}', "");
Expect(1, 127570, '\P{Is_Decomposition_Type:	noncanonical}', "");
Expect(0, 127570, '\P{^Is_Decomposition_Type:	noncanonical}', "");
Expect(1, 127569, '\p{Is_Decomposition_Type:   NON_CANONICAL}', "");
Expect(0, 127569, '\p{^Is_Decomposition_Type:   NON_CANONICAL}', "");
Expect(0, 127569, '\P{Is_Decomposition_Type:   NON_CANONICAL}', "");
Expect(1, 127569, '\P{^Is_Decomposition_Type:   NON_CANONICAL}', "");
Expect(0, 127570, '\p{Is_Decomposition_Type:   NON_CANONICAL}', "");
Expect(1, 127570, '\p{^Is_Decomposition_Type:   NON_CANONICAL}', "");
Expect(1, 127570, '\P{Is_Decomposition_Type:   NON_CANONICAL}', "");
Expect(0, 127570, '\P{^Is_Decomposition_Type:   NON_CANONICAL}', "");
Error('\p{Is_Dt=	Non_canon/a/}');
Error('\P{Is_Dt=	Non_canon/a/}');
Expect(1, 127569, '\p{Is_Dt=noncanon}', "");
Expect(0, 127569, '\p{^Is_Dt=noncanon}', "");
Expect(0, 127569, '\P{Is_Dt=noncanon}', "");
Expect(1, 127569, '\P{^Is_Dt=noncanon}', "");
Expect(0, 127570, '\p{Is_Dt=noncanon}', "");
Expect(1, 127570, '\p{^Is_Dt=noncanon}', "");
Expect(1, 127570, '\P{Is_Dt=noncanon}', "");
Expect(0, 127570, '\P{^Is_Dt=noncanon}', "");
Expect(1, 127569, '\p{Is_Dt=__NON_CANON}', "");
Expect(0, 127569, '\p{^Is_Dt=__NON_CANON}', "");
Expect(0, 127569, '\P{Is_Dt=__NON_CANON}', "");
Expect(1, 127569, '\P{^Is_Dt=__NON_CANON}', "");
Expect(0, 127570, '\p{Is_Dt=__NON_CANON}', "");
Expect(1, 127570, '\p{^Is_Dt=__NON_CANON}', "");
Expect(1, 127570, '\P{Is_Dt=__NON_CANON}', "");
Expect(0, 127570, '\P{^Is_Dt=__NON_CANON}', "");
Error('\p{Decomposition_Type=/a/Final}');
Error('\P{Decomposition_Type=/a/Final}');
Expect(1, 65276, '\p{Decomposition_Type=final}', "");
Expect(0, 65276, '\p{^Decomposition_Type=final}', "");
Expect(0, 65276, '\P{Decomposition_Type=final}', "");
Expect(1, 65276, '\P{^Decomposition_Type=final}', "");
Expect(0, 65277, '\p{Decomposition_Type=final}', "");
Expect(1, 65277, '\p{^Decomposition_Type=final}', "");
Expect(1, 65277, '\P{Decomposition_Type=final}', "");
Expect(0, 65277, '\P{^Decomposition_Type=final}', "");
Expect(1, 65276, '\p{Decomposition_Type=  Final}', "");
Expect(0, 65276, '\p{^Decomposition_Type=  Final}', "");
Expect(0, 65276, '\P{Decomposition_Type=  Final}', "");
Expect(1, 65276, '\P{^Decomposition_Type=  Final}', "");
Expect(0, 65277, '\p{Decomposition_Type=  Final}', "");
Expect(1, 65277, '\p{^Decomposition_Type=  Final}', "");
Expect(1, 65277, '\P{Decomposition_Type=  Final}', "");
Expect(0, 65277, '\P{^Decomposition_Type=  Final}', "");
Error('\p{Dt=-	Fin/a/}');
Error('\P{Dt=-	Fin/a/}');
Expect(1, 65276, '\p{Dt=fin}', "");
Expect(0, 65276, '\p{^Dt=fin}', "");
Expect(0, 65276, '\P{Dt=fin}', "");
Expect(1, 65276, '\P{^Dt=fin}', "");
Expect(0, 65277, '\p{Dt=fin}', "");
Expect(1, 65277, '\p{^Dt=fin}', "");
Expect(1, 65277, '\P{Dt=fin}', "");
Expect(0, 65277, '\P{^Dt=fin}', "");
Expect(1, 65276, '\p{Dt=	Fin}', "");
Expect(0, 65276, '\p{^Dt=	Fin}', "");
Expect(0, 65276, '\P{Dt=	Fin}', "");
Expect(1, 65276, '\P{^Dt=	Fin}', "");
Expect(0, 65277, '\p{Dt=	Fin}', "");
Expect(1, 65277, '\p{^Dt=	Fin}', "");
Expect(1, 65277, '\P{Dt=	Fin}', "");
Expect(0, 65277, '\P{^Dt=	Fin}', "");
Error('\p{Is_Decomposition_Type=/a/_	final}');
Error('\P{Is_Decomposition_Type=/a/_	final}');
Expect(1, 65276, '\p{Is_Decomposition_Type=final}', "");
Expect(0, 65276, '\p{^Is_Decomposition_Type=final}', "");
Expect(0, 65276, '\P{Is_Decomposition_Type=final}', "");
Expect(1, 65276, '\P{^Is_Decomposition_Type=final}', "");
Expect(0, 65277, '\p{Is_Decomposition_Type=final}', "");
Expect(1, 65277, '\p{^Is_Decomposition_Type=final}', "");
Expect(1, 65277, '\P{Is_Decomposition_Type=final}', "");
Expect(0, 65277, '\P{^Is_Decomposition_Type=final}', "");
Expect(1, 65276, '\p{Is_Decomposition_Type=_	final}', "");
Expect(0, 65276, '\p{^Is_Decomposition_Type=_	final}', "");
Expect(0, 65276, '\P{Is_Decomposition_Type=_	final}', "");
Expect(1, 65276, '\P{^Is_Decomposition_Type=_	final}', "");
Expect(0, 65277, '\p{Is_Decomposition_Type=_	final}', "");
Expect(1, 65277, '\p{^Is_Decomposition_Type=_	final}', "");
Expect(1, 65277, '\P{Is_Decomposition_Type=_	final}', "");
Expect(0, 65277, '\P{^Is_Decomposition_Type=_	final}', "");
Error('\p{Is_Dt=:=	 Fin}');
Error('\P{Is_Dt=:=	 Fin}');
Expect(1, 65276, '\p{Is_Dt=fin}', "");
Expect(0, 65276, '\p{^Is_Dt=fin}', "");
Expect(0, 65276, '\P{Is_Dt=fin}', "");
Expect(1, 65276, '\P{^Is_Dt=fin}', "");
Expect(0, 65277, '\p{Is_Dt=fin}', "");
Expect(1, 65277, '\p{^Is_Dt=fin}', "");
Expect(1, 65277, '\P{Is_Dt=fin}', "");
Expect(0, 65277, '\P{^Is_Dt=fin}', "");
Expect(1, 65276, '\p{Is_Dt= _fin}', "");
Expect(0, 65276, '\p{^Is_Dt= _fin}', "");
Expect(0, 65276, '\P{Is_Dt= _fin}', "");
Expect(1, 65276, '\P{^Is_Dt= _fin}', "");
Expect(0, 65277, '\p{Is_Dt= _fin}', "");
Expect(1, 65277, '\p{^Is_Dt= _fin}', "");
Expect(1, 65277, '\P{Is_Dt= _fin}', "");
Expect(0, 65277, '\P{^Is_Dt= _fin}', "");
Error('\p{Decomposition_Type: 	wide:=}');
Error('\P{Decomposition_Type: 	wide:=}');
Expect(1, 65510, '\p{Decomposition_Type=wide}', "");
Expect(0, 65510, '\p{^Decomposition_Type=wide}', "");
Expect(0, 65510, '\P{Decomposition_Type=wide}', "");
Expect(1, 65510, '\P{^Decomposition_Type=wide}', "");
Expect(0, 65511, '\p{Decomposition_Type=wide}', "");
Expect(1, 65511, '\p{^Decomposition_Type=wide}', "");
Expect(1, 65511, '\P{Decomposition_Type=wide}', "");
Expect(0, 65511, '\P{^Decomposition_Type=wide}', "");
Expect(1, 65510, '\p{Decomposition_Type=-Wide}', "");
Expect(0, 65510, '\p{^Decomposition_Type=-Wide}', "");
Expect(0, 65510, '\P{Decomposition_Type=-Wide}', "");
Expect(1, 65510, '\P{^Decomposition_Type=-Wide}', "");
Expect(0, 65511, '\p{Decomposition_Type=-Wide}', "");
Expect(1, 65511, '\p{^Decomposition_Type=-Wide}', "");
Expect(1, 65511, '\P{Decomposition_Type=-Wide}', "");
Expect(0, 65511, '\P{^Decomposition_Type=-Wide}', "");
Error('\p{Dt=:=_ WIDE}');
Error('\P{Dt=:=_ WIDE}');
Expect(1, 65510, '\p{Dt=wide}', "");
Expect(0, 65510, '\p{^Dt=wide}', "");
Expect(0, 65510, '\P{Dt=wide}', "");
Expect(1, 65510, '\P{^Dt=wide}', "");
Expect(0, 65511, '\p{Dt=wide}', "");
Expect(1, 65511, '\p{^Dt=wide}', "");
Expect(1, 65511, '\P{Dt=wide}', "");
Expect(0, 65511, '\P{^Dt=wide}', "");
Expect(1, 65510, '\p{Dt=		wide}', "");
Expect(0, 65510, '\p{^Dt=		wide}', "");
Expect(0, 65510, '\P{Dt=		wide}', "");
Expect(1, 65510, '\P{^Dt=		wide}', "");
Expect(0, 65511, '\p{Dt=		wide}', "");
Expect(1, 65511, '\p{^Dt=		wide}', "");
Expect(1, 65511, '\P{Dt=		wide}', "");
Expect(0, 65511, '\P{^Dt=		wide}', "");
Error('\p{Is_Decomposition_Type=:= _Wide}');
Error('\P{Is_Decomposition_Type=:= _Wide}');
Expect(1, 65510, '\p{Is_Decomposition_Type:wide}', "");
Expect(0, 65510, '\p{^Is_Decomposition_Type:wide}', "");
Expect(0, 65510, '\P{Is_Decomposition_Type:wide}', "");
Expect(1, 65510, '\P{^Is_Decomposition_Type:wide}', "");
Expect(0, 65511, '\p{Is_Decomposition_Type:wide}', "");
Expect(1, 65511, '\p{^Is_Decomposition_Type:wide}', "");
Expect(1, 65511, '\P{Is_Decomposition_Type:wide}', "");
Expect(0, 65511, '\P{^Is_Decomposition_Type:wide}', "");
Expect(1, 65510, '\p{Is_Decomposition_Type=-Wide}', "");
Expect(0, 65510, '\p{^Is_Decomposition_Type=-Wide}', "");
Expect(0, 65510, '\P{Is_Decomposition_Type=-Wide}', "");
Expect(1, 65510, '\P{^Is_Decomposition_Type=-Wide}', "");
Expect(0, 65511, '\p{Is_Decomposition_Type=-Wide}', "");
Expect(1, 65511, '\p{^Is_Decomposition_Type=-Wide}', "");
Expect(1, 65511, '\P{Is_Decomposition_Type=-Wide}', "");
Expect(0, 65511, '\P{^Is_Decomposition_Type=-Wide}', "");
Error('\p{Is_Dt=:=	-WIDE}');
Error('\P{Is_Dt=:=	-WIDE}');
Expect(1, 65510, '\p{Is_Dt:wide}', "");
Expect(0, 65510, '\p{^Is_Dt:wide}', "");
Expect(0, 65510, '\P{Is_Dt:wide}', "");
Expect(1, 65510, '\P{^Is_Dt:wide}', "");
Expect(0, 65511, '\p{Is_Dt:wide}', "");
Expect(1, 65511, '\p{^Is_Dt:wide}', "");
Expect(1, 65511, '\P{Is_Dt:wide}', "");
Expect(0, 65511, '\P{^Is_Dt:wide}', "");
Expect(1, 65510, '\p{Is_Dt=_WIDE}', "");
Expect(0, 65510, '\p{^Is_Dt=_WIDE}', "");
Expect(0, 65510, '\P{Is_Dt=_WIDE}', "");
Expect(1, 65510, '\P{^Is_Dt=_WIDE}', "");
Expect(0, 65511, '\p{Is_Dt=_WIDE}', "");
Expect(1, 65511, '\p{^Is_Dt=_WIDE}', "");
Expect(1, 65511, '\P{Is_Dt=_WIDE}', "");
Expect(0, 65511, '\P{^Is_Dt=_WIDE}', "");
Error('\p{Decomposition_Type:		canonical/a/}');
Error('\P{Decomposition_Type:		canonical/a/}');
Expect(1, 195101, '\p{Decomposition_Type:   canonical}', "");
Expect(0, 195101, '\p{^Decomposition_Type:   canonical}', "");
Expect(0, 195101, '\P{Decomposition_Type:   canonical}', "");
Expect(1, 195101, '\P{^Decomposition_Type:   canonical}', "");
Expect(0, 195102, '\p{Decomposition_Type:   canonical}', "");
Expect(1, 195102, '\p{^Decomposition_Type:   canonical}', "");
Expect(1, 195102, '\P{Decomposition_Type:   canonical}', "");
Expect(0, 195102, '\P{^Decomposition_Type:   canonical}', "");
Expect(1, 195101, '\p{Decomposition_Type=-Canonical}', "");
Expect(0, 195101, '\p{^Decomposition_Type=-Canonical}', "");
Expect(0, 195101, '\P{Decomposition_Type=-Canonical}', "");
Expect(1, 195101, '\P{^Decomposition_Type=-Canonical}', "");
Expect(0, 195102, '\p{Decomposition_Type=-Canonical}', "");
Expect(1, 195102, '\p{^Decomposition_Type=-Canonical}', "");
Expect(1, 195102, '\P{Decomposition_Type=-Canonical}', "");
Expect(0, 195102, '\P{^Decomposition_Type=-Canonical}', "");
Error('\p{Dt=_Can/a/}');
Error('\P{Dt=_Can/a/}');
Expect(1, 195101, '\p{Dt=can}', "");
Expect(0, 195101, '\p{^Dt=can}', "");
Expect(0, 195101, '\P{Dt=can}', "");
Expect(1, 195101, '\P{^Dt=can}', "");
Expect(0, 195102, '\p{Dt=can}', "");
Expect(1, 195102, '\p{^Dt=can}', "");
Expect(1, 195102, '\P{Dt=can}', "");
Expect(0, 195102, '\P{^Dt=can}', "");
Expect(1, 195101, '\p{Dt=_ Can}', "");
Expect(0, 195101, '\p{^Dt=_ Can}', "");
Expect(0, 195101, '\P{Dt=_ Can}', "");
Expect(1, 195101, '\P{^Dt=_ Can}', "");
Expect(0, 195102, '\p{Dt=_ Can}', "");
Expect(1, 195102, '\p{^Dt=_ Can}', "");
Expect(1, 195102, '\P{Dt=_ Can}', "");
Expect(0, 195102, '\P{^Dt=_ Can}', "");
Error('\p{Is_Decomposition_Type=/a/canonical}');
Error('\P{Is_Decomposition_Type=/a/canonical}');
Expect(1, 195101, '\p{Is_Decomposition_Type=canonical}', "");
Expect(0, 195101, '\p{^Is_Decomposition_Type=canonical}', "");
Expect(0, 195101, '\P{Is_Decomposition_Type=canonical}', "");
Expect(1, 195101, '\P{^Is_Decomposition_Type=canonical}', "");
Expect(0, 195102, '\p{Is_Decomposition_Type=canonical}', "");
Expect(1, 195102, '\p{^Is_Decomposition_Type=canonical}', "");
Expect(1, 195102, '\P{Is_Decomposition_Type=canonical}', "");
Expect(0, 195102, '\P{^Is_Decomposition_Type=canonical}', "");
Expect(1, 195101, '\p{Is_Decomposition_Type=-Canonical}', "");
Expect(0, 195101, '\p{^Is_Decomposition_Type=-Canonical}', "");
Expect(0, 195101, '\P{Is_Decomposition_Type=-Canonical}', "");
Expect(1, 195101, '\P{^Is_Decomposition_Type=-Canonical}', "");
Expect(0, 195102, '\p{Is_Decomposition_Type=-Canonical}', "");
Expect(1, 195102, '\p{^Is_Decomposition_Type=-Canonical}', "");
Expect(1, 195102, '\P{Is_Decomposition_Type=-Canonical}', "");
Expect(0, 195102, '\P{^Is_Decomposition_Type=-Canonical}', "");
Error('\p{Is_Dt=_CAN/a/}');
Error('\P{Is_Dt=_CAN/a/}');
Expect(1, 195101, '\p{Is_Dt=can}', "");
Expect(0, 195101, '\p{^Is_Dt=can}', "");
Expect(0, 195101, '\P{Is_Dt=can}', "");
Expect(1, 195101, '\P{^Is_Dt=can}', "");
Expect(0, 195102, '\p{Is_Dt=can}', "");
Expect(1, 195102, '\p{^Is_Dt=can}', "");
Expect(1, 195102, '\P{Is_Dt=can}', "");
Expect(0, 195102, '\P{^Is_Dt=can}', "");
Expect(1, 195101, '\p{Is_Dt= Can}', "");
Expect(0, 195101, '\p{^Is_Dt= Can}', "");
Expect(0, 195101, '\P{Is_Dt= Can}', "");
Expect(1, 195101, '\P{^Is_Dt= Can}', "");
Expect(0, 195102, '\p{Is_Dt= Can}', "");
Expect(1, 195102, '\p{^Is_Dt= Can}', "");
Expect(1, 195102, '\P{Is_Dt= Can}', "");
Expect(0, 195102, '\P{^Is_Dt= Can}', "");
Error('\p{Decomposition_Type:	/a/ Vertical}');
Error('\P{Decomposition_Type:	/a/ Vertical}');
Expect(1, 65096, '\p{Decomposition_Type=vertical}', "");
Expect(0, 65096, '\p{^Decomposition_Type=vertical}', "");
Expect(0, 65096, '\P{Decomposition_Type=vertical}', "");
Expect(1, 65096, '\P{^Decomposition_Type=vertical}', "");
Expect(0, 65097, '\p{Decomposition_Type=vertical}', "");
Expect(1, 65097, '\p{^Decomposition_Type=vertical}', "");
Expect(1, 65097, '\P{Decomposition_Type=vertical}', "");
Expect(0, 65097, '\P{^Decomposition_Type=vertical}', "");
Expect(1, 65096, '\p{Decomposition_Type= _VERTICAL}', "");
Expect(0, 65096, '\p{^Decomposition_Type= _VERTICAL}', "");
Expect(0, 65096, '\P{Decomposition_Type= _VERTICAL}', "");
Expect(1, 65096, '\P{^Decomposition_Type= _VERTICAL}', "");
Expect(0, 65097, '\p{Decomposition_Type= _VERTICAL}', "");
Expect(1, 65097, '\p{^Decomposition_Type= _VERTICAL}', "");
Expect(1, 65097, '\P{Decomposition_Type= _VERTICAL}', "");
Expect(0, 65097, '\P{^Decomposition_Type= _VERTICAL}', "");
Error('\p{Dt=	:=VERT}');
Error('\P{Dt=	:=VERT}');
Expect(1, 65096, '\p{Dt=vert}', "");
Expect(0, 65096, '\p{^Dt=vert}', "");
Expect(0, 65096, '\P{Dt=vert}', "");
Expect(1, 65096, '\P{^Dt=vert}', "");
Expect(0, 65097, '\p{Dt=vert}', "");
Expect(1, 65097, '\p{^Dt=vert}', "");
Expect(1, 65097, '\P{Dt=vert}', "");
Expect(0, 65097, '\P{^Dt=vert}', "");
Expect(1, 65096, '\p{Dt=--VERT}', "");
Expect(0, 65096, '\p{^Dt=--VERT}', "");
Expect(0, 65096, '\P{Dt=--VERT}', "");
Expect(1, 65096, '\P{^Dt=--VERT}', "");
Expect(0, 65097, '\p{Dt=--VERT}', "");
Expect(1, 65097, '\p{^Dt=--VERT}', "");
Expect(1, 65097, '\P{Dt=--VERT}', "");
Expect(0, 65097, '\P{^Dt=--VERT}', "");
Error('\p{Is_Decomposition_Type=/a/Vertical}');
Error('\P{Is_Decomposition_Type=/a/Vertical}');
Expect(1, 65096, '\p{Is_Decomposition_Type=vertical}', "");
Expect(0, 65096, '\p{^Is_Decomposition_Type=vertical}', "");
Expect(0, 65096, '\P{Is_Decomposition_Type=vertical}', "");
Expect(1, 65096, '\P{^Is_Decomposition_Type=vertical}', "");
Expect(0, 65097, '\p{Is_Decomposition_Type=vertical}', "");
Expect(1, 65097, '\p{^Is_Decomposition_Type=vertical}', "");
Expect(1, 65097, '\P{Is_Decomposition_Type=vertical}', "");
Expect(0, 65097, '\P{^Is_Decomposition_Type=vertical}', "");
Expect(1, 65096, '\p{Is_Decomposition_Type=-	Vertical}', "");
Expect(0, 65096, '\p{^Is_Decomposition_Type=-	Vertical}', "");
Expect(0, 65096, '\P{Is_Decomposition_Type=-	Vertical}', "");
Expect(1, 65096, '\P{^Is_Decomposition_Type=-	Vertical}', "");
Expect(0, 65097, '\p{Is_Decomposition_Type=-	Vertical}', "");
Expect(1, 65097, '\p{^Is_Decomposition_Type=-	Vertical}', "");
Expect(1, 65097, '\P{Is_Decomposition_Type=-	Vertical}', "");
Expect(0, 65097, '\P{^Is_Decomposition_Type=-	Vertical}', "");
Error('\p{Is_Dt=_Vert:=}');
Error('\P{Is_Dt=_Vert:=}');
Expect(1, 65096, '\p{Is_Dt=vert}', "");
Expect(0, 65096, '\p{^Is_Dt=vert}', "");
Expect(0, 65096, '\P{Is_Dt=vert}', "");
Expect(1, 65096, '\P{^Is_Dt=vert}', "");
Expect(0, 65097, '\p{Is_Dt=vert}', "");
Expect(1, 65097, '\p{^Is_Dt=vert}', "");
Expect(1, 65097, '\P{Is_Dt=vert}', "");
Expect(0, 65097, '\P{^Is_Dt=vert}', "");
Expect(1, 65096, '\p{Is_Dt:	_ vert}', "");
Expect(0, 65096, '\p{^Is_Dt:	_ vert}', "");
Expect(0, 65096, '\P{Is_Dt:	_ vert}', "");
Expect(1, 65096, '\P{^Is_Dt:	_ vert}', "");
Expect(0, 65097, '\p{Is_Dt:	_ vert}', "");
Expect(1, 65097, '\p{^Is_Dt:	_ vert}', "");
Expect(1, 65097, '\P{Is_Dt:	_ vert}', "");
Expect(0, 65097, '\P{^Is_Dt:	_ vert}', "");
Error('\p{Decomposition_Type:   :=	_medial}');
Error('\P{Decomposition_Type:   :=	_medial}');
Expect(1, 65268, '\p{Decomposition_Type=medial}', "");
Expect(0, 65268, '\p{^Decomposition_Type=medial}', "");
Expect(0, 65268, '\P{Decomposition_Type=medial}', "");
Expect(1, 65268, '\P{^Decomposition_Type=medial}', "");
Expect(0, 65269, '\p{Decomposition_Type=medial}', "");
Expect(1, 65269, '\p{^Decomposition_Type=medial}', "");
Expect(1, 65269, '\P{Decomposition_Type=medial}', "");
Expect(0, 65269, '\P{^Decomposition_Type=medial}', "");
Expect(1, 65268, '\p{Decomposition_Type= Medial}', "");
Expect(0, 65268, '\p{^Decomposition_Type= Medial}', "");
Expect(0, 65268, '\P{Decomposition_Type= Medial}', "");
Expect(1, 65268, '\P{^Decomposition_Type= Medial}', "");
Expect(0, 65269, '\p{Decomposition_Type= Medial}', "");
Expect(1, 65269, '\p{^Decomposition_Type= Medial}', "");
Expect(1, 65269, '\P{Decomposition_Type= Medial}', "");
Expect(0, 65269, '\P{^Decomposition_Type= Medial}', "");
Error('\p{Dt=_	MED:=}');
Error('\P{Dt=_	MED:=}');
Expect(1, 65268, '\p{Dt=med}', "");
Expect(0, 65268, '\p{^Dt=med}', "");
Expect(0, 65268, '\P{Dt=med}', "");
Expect(1, 65268, '\P{^Dt=med}', "");
Expect(0, 65269, '\p{Dt=med}', "");
Expect(1, 65269, '\p{^Dt=med}', "");
Expect(1, 65269, '\P{Dt=med}', "");
Expect(0, 65269, '\P{^Dt=med}', "");
Expect(1, 65268, '\p{Dt:	 -Med}', "");
Expect(0, 65268, '\p{^Dt:	 -Med}', "");
Expect(0, 65268, '\P{Dt:	 -Med}', "");
Expect(1, 65268, '\P{^Dt:	 -Med}', "");
Expect(0, 65269, '\p{Dt:	 -Med}', "");
Expect(1, 65269, '\p{^Dt:	 -Med}', "");
Expect(1, 65269, '\P{Dt:	 -Med}', "");
Expect(0, 65269, '\P{^Dt:	 -Med}', "");
Error('\p{Is_Decomposition_Type=_ Medial/a/}');
Error('\P{Is_Decomposition_Type=_ Medial/a/}');
Expect(1, 65268, '\p{Is_Decomposition_Type:medial}', "");
Expect(0, 65268, '\p{^Is_Decomposition_Type:medial}', "");
Expect(0, 65268, '\P{Is_Decomposition_Type:medial}', "");
Expect(1, 65268, '\P{^Is_Decomposition_Type:medial}', "");
Expect(0, 65269, '\p{Is_Decomposition_Type:medial}', "");
Expect(1, 65269, '\p{^Is_Decomposition_Type:medial}', "");
Expect(1, 65269, '\P{Is_Decomposition_Type:medial}', "");
Expect(0, 65269, '\P{^Is_Decomposition_Type:medial}', "");
Expect(1, 65268, '\p{Is_Decomposition_Type=-Medial}', "");
Expect(0, 65268, '\p{^Is_Decomposition_Type=-Medial}', "");
Expect(0, 65268, '\P{Is_Decomposition_Type=-Medial}', "");
Expect(1, 65268, '\P{^Is_Decomposition_Type=-Medial}', "");
Expect(0, 65269, '\p{Is_Decomposition_Type=-Medial}', "");
Expect(1, 65269, '\p{^Is_Decomposition_Type=-Medial}', "");
Expect(1, 65269, '\P{Is_Decomposition_Type=-Medial}', "");
Expect(0, 65269, '\P{^Is_Decomposition_Type=-Medial}', "");
Error('\p{Is_Dt=-/a/med}');
Error('\P{Is_Dt=-/a/med}');
Expect(1, 65268, '\p{Is_Dt=med}', "");
Expect(0, 65268, '\p{^Is_Dt=med}', "");
Expect(0, 65268, '\P{Is_Dt=med}', "");
Expect(1, 65268, '\P{^Is_Dt=med}', "");
Expect(0, 65269, '\p{Is_Dt=med}', "");
Expect(1, 65269, '\p{^Is_Dt=med}', "");
Expect(1, 65269, '\P{Is_Dt=med}', "");
Expect(0, 65269, '\P{^Is_Dt=med}', "");
Expect(1, 65268, '\p{Is_Dt=	Med}', "");
Expect(0, 65268, '\p{^Is_Dt=	Med}', "");
Expect(0, 65268, '\P{Is_Dt=	Med}', "");
Expect(1, 65268, '\P{^Is_Dt=	Med}', "");
Expect(0, 65269, '\p{Is_Dt=	Med}', "");
Expect(1, 65269, '\p{^Is_Dt=	Med}', "");
Expect(1, 65269, '\P{Is_Dt=	Med}', "");
Expect(0, 65269, '\P{^Is_Dt=	Med}', "");
Error('\p{Decomposition_Type= 	NARROW/a/}');
Error('\P{Decomposition_Type= 	NARROW/a/}');
Expect(1, 65518, '\p{Decomposition_Type: narrow}', "");
Expect(0, 65518, '\p{^Decomposition_Type: narrow}', "");
Expect(0, 65518, '\P{Decomposition_Type: narrow}', "");
Expect(1, 65518, '\P{^Decomposition_Type: narrow}', "");
Expect(0, 65519, '\p{Decomposition_Type: narrow}', "");
Expect(1, 65519, '\p{^Decomposition_Type: narrow}', "");
Expect(1, 65519, '\P{Decomposition_Type: narrow}', "");
Expect(0, 65519, '\P{^Decomposition_Type: narrow}', "");
Expect(1, 65518, '\p{Decomposition_Type=- Narrow}', "");
Expect(0, 65518, '\p{^Decomposition_Type=- Narrow}', "");
Expect(0, 65518, '\P{Decomposition_Type=- Narrow}', "");
Expect(1, 65518, '\P{^Decomposition_Type=- Narrow}', "");
Expect(0, 65519, '\p{Decomposition_Type=- Narrow}', "");
Expect(1, 65519, '\p{^Decomposition_Type=- Narrow}', "");
Expect(1, 65519, '\P{Decomposition_Type=- Narrow}', "");
Expect(0, 65519, '\P{^Decomposition_Type=- Narrow}', "");
Error('\p{Dt=	_nar/a/}');
Error('\P{Dt=	_nar/a/}');
Expect(1, 65518, '\p{Dt=nar}', "");
Expect(0, 65518, '\p{^Dt=nar}', "");
Expect(0, 65518, '\P{Dt=nar}', "");
Expect(1, 65518, '\P{^Dt=nar}', "");
Expect(0, 65519, '\p{Dt=nar}', "");
Expect(1, 65519, '\p{^Dt=nar}', "");
Expect(1, 65519, '\P{Dt=nar}', "");
Expect(0, 65519, '\P{^Dt=nar}', "");
Expect(1, 65518, '\p{Dt=__Nar}', "");
Expect(0, 65518, '\p{^Dt=__Nar}', "");
Expect(0, 65518, '\P{Dt=__Nar}', "");
Expect(1, 65518, '\P{^Dt=__Nar}', "");
Expect(0, 65519, '\p{Dt=__Nar}', "");
Expect(1, 65519, '\p{^Dt=__Nar}', "");
Expect(1, 65519, '\P{Dt=__Nar}', "");
Expect(0, 65519, '\P{^Dt=__Nar}', "");
Error('\p{Is_Decomposition_Type:   	/a/NARROW}');
Error('\P{Is_Decomposition_Type:   	/a/NARROW}');
Expect(1, 65518, '\p{Is_Decomposition_Type=narrow}', "");
Expect(0, 65518, '\p{^Is_Decomposition_Type=narrow}', "");
Expect(0, 65518, '\P{Is_Decomposition_Type=narrow}', "");
Expect(1, 65518, '\P{^Is_Decomposition_Type=narrow}', "");
Expect(0, 65519, '\p{Is_Decomposition_Type=narrow}', "");
Expect(1, 65519, '\p{^Is_Decomposition_Type=narrow}', "");
Expect(1, 65519, '\P{Is_Decomposition_Type=narrow}', "");
Expect(0, 65519, '\P{^Is_Decomposition_Type=narrow}', "");
Expect(1, 65518, '\p{Is_Decomposition_Type:    Narrow}', "");
Expect(0, 65518, '\p{^Is_Decomposition_Type:    Narrow}', "");
Expect(0, 65518, '\P{Is_Decomposition_Type:    Narrow}', "");
Expect(1, 65518, '\P{^Is_Decomposition_Type:    Narrow}', "");
Expect(0, 65519, '\p{Is_Decomposition_Type:    Narrow}', "");
Expect(1, 65519, '\p{^Is_Decomposition_Type:    Narrow}', "");
Expect(1, 65519, '\P{Is_Decomposition_Type:    Narrow}', "");
Expect(0, 65519, '\P{^Is_Decomposition_Type:    Narrow}', "");
Error('\p{Is_Dt=:=NAR}');
Error('\P{Is_Dt=:=NAR}');
Expect(1, 65518, '\p{Is_Dt=nar}', "");
Expect(0, 65518, '\p{^Is_Dt=nar}', "");
Expect(0, 65518, '\P{Is_Dt=nar}', "");
Expect(1, 65518, '\P{^Is_Dt=nar}', "");
Expect(0, 65519, '\p{Is_Dt=nar}', "");
Expect(1, 65519, '\p{^Is_Dt=nar}', "");
Expect(1, 65519, '\P{Is_Dt=nar}', "");
Expect(0, 65519, '\P{^Is_Dt=nar}', "");
Expect(1, 65518, '\p{Is_Dt:   -Nar}', "");
Expect(0, 65518, '\p{^Is_Dt:   -Nar}', "");
Expect(0, 65518, '\P{Is_Dt:   -Nar}', "");
Expect(1, 65518, '\P{^Is_Dt:   -Nar}', "");
Expect(0, 65519, '\p{Is_Dt:   -Nar}', "");
Expect(1, 65519, '\p{^Is_Dt:   -Nar}', "");
Expect(1, 65519, '\P{Is_Dt:   -Nar}', "");
Expect(0, 65519, '\P{^Is_Dt:   -Nar}', "");
Error('\p{Decomposition_Type=	-sub/a/}');
Error('\P{Decomposition_Type=	-sub/a/}');
Expect(1, 11388, '\p{Decomposition_Type=sub}', "");
Expect(0, 11388, '\p{^Decomposition_Type=sub}', "");
Expect(0, 11388, '\P{Decomposition_Type=sub}', "");
Expect(1, 11388, '\P{^Decomposition_Type=sub}', "");
Expect(0, 11389, '\p{Decomposition_Type=sub}', "");
Expect(1, 11389, '\p{^Decomposition_Type=sub}', "");
Expect(1, 11389, '\P{Decomposition_Type=sub}', "");
Expect(0, 11389, '\P{^Decomposition_Type=sub}', "");
Expect(1, 11388, '\p{Decomposition_Type=	sub}', "");
Expect(0, 11388, '\p{^Decomposition_Type=	sub}', "");
Expect(0, 11388, '\P{Decomposition_Type=	sub}', "");
Expect(1, 11388, '\P{^Decomposition_Type=	sub}', "");
Expect(0, 11389, '\p{Decomposition_Type=	sub}', "");
Expect(1, 11389, '\p{^Decomposition_Type=	sub}', "");
Expect(1, 11389, '\P{Decomposition_Type=	sub}', "");
Expect(0, 11389, '\P{^Decomposition_Type=	sub}', "");
Error('\p{Dt: 	:=SUB}');
Error('\P{Dt: 	:=SUB}');
Expect(1, 11388, '\p{Dt=sub}', "");
Expect(0, 11388, '\p{^Dt=sub}', "");
Expect(0, 11388, '\P{Dt=sub}', "");
Expect(1, 11388, '\P{^Dt=sub}', "");
Expect(0, 11389, '\p{Dt=sub}', "");
Expect(1, 11389, '\p{^Dt=sub}', "");
Expect(1, 11389, '\P{Dt=sub}', "");
Expect(0, 11389, '\P{^Dt=sub}', "");
Expect(1, 11388, '\p{Dt=- SUB}', "");
Expect(0, 11388, '\p{^Dt=- SUB}', "");
Expect(0, 11388, '\P{Dt=- SUB}', "");
Expect(1, 11388, '\P{^Dt=- SUB}', "");
Expect(0, 11389, '\p{Dt=- SUB}', "");
Expect(1, 11389, '\p{^Dt=- SUB}', "");
Expect(1, 11389, '\P{Dt=- SUB}', "");
Expect(0, 11389, '\P{^Dt=- SUB}', "");
Error('\p{Is_Decomposition_Type=_ sub:=}');
Error('\P{Is_Decomposition_Type=_ sub:=}');
Expect(1, 11388, '\p{Is_Decomposition_Type=sub}', "");
Expect(0, 11388, '\p{^Is_Decomposition_Type=sub}', "");
Expect(0, 11388, '\P{Is_Decomposition_Type=sub}', "");
Expect(1, 11388, '\P{^Is_Decomposition_Type=sub}', "");
Expect(0, 11389, '\p{Is_Decomposition_Type=sub}', "");
Expect(1, 11389, '\p{^Is_Decomposition_Type=sub}', "");
Expect(1, 11389, '\P{Is_Decomposition_Type=sub}', "");
Expect(0, 11389, '\P{^Is_Decomposition_Type=sub}', "");
Expect(1, 11388, '\p{Is_Decomposition_Type: 	sub}', "");
Expect(0, 11388, '\p{^Is_Decomposition_Type: 	sub}', "");
Expect(0, 11388, '\P{Is_Decomposition_Type: 	sub}', "");
Expect(1, 11388, '\P{^Is_Decomposition_Type: 	sub}', "");
Expect(0, 11389, '\p{Is_Decomposition_Type: 	sub}', "");
Expect(1, 11389, '\p{^Is_Decomposition_Type: 	sub}', "");
Expect(1, 11389, '\P{Is_Decomposition_Type: 	sub}', "");
Expect(0, 11389, '\P{^Is_Decomposition_Type: 	sub}', "");
Error('\p{Is_Dt=--Sub/a/}');
Error('\P{Is_Dt=--Sub/a/}');
Expect(1, 11388, '\p{Is_Dt=sub}', "");
Expect(0, 11388, '\p{^Is_Dt=sub}', "");
Expect(0, 11388, '\P{Is_Dt=sub}', "");
Expect(1, 11388, '\P{^Is_Dt=sub}', "");
Expect(0, 11389, '\p{Is_Dt=sub}', "");
Expect(1, 11389, '\p{^Is_Dt=sub}', "");
Expect(1, 11389, '\P{Is_Dt=sub}', "");
Expect(0, 11389, '\P{^Is_Dt=sub}', "");
Expect(1, 11388, '\p{Is_Dt=_sub}', "");
Expect(0, 11388, '\p{^Is_Dt=_sub}', "");
Expect(0, 11388, '\P{Is_Dt=_sub}', "");
Expect(1, 11388, '\P{^Is_Dt=_sub}', "");
Expect(0, 11389, '\p{Is_Dt=_sub}', "");
Expect(1, 11389, '\p{^Is_Dt=_sub}', "");
Expect(1, 11389, '\P{Is_Dt=_sub}', "");
Expect(0, 11389, '\P{^Is_Dt=_sub}', "");
Error('\p{Decomposition_Type=_Square:=}');
Error('\P{Decomposition_Type=_Square:=}');
Expect(1, 127546, '\p{Decomposition_Type=square}', "");
Expect(0, 127546, '\p{^Decomposition_Type=square}', "");
Expect(0, 127546, '\P{Decomposition_Type=square}', "");
Expect(1, 127546, '\P{^Decomposition_Type=square}', "");
Expect(0, 127547, '\p{Decomposition_Type=square}', "");
Expect(1, 127547, '\p{^Decomposition_Type=square}', "");
Expect(1, 127547, '\P{Decomposition_Type=square}', "");
Expect(0, 127547, '\P{^Decomposition_Type=square}', "");
Expect(1, 127546, '\p{Decomposition_Type=--Square}', "");
Expect(0, 127546, '\p{^Decomposition_Type=--Square}', "");
Expect(0, 127546, '\P{Decomposition_Type=--Square}', "");
Expect(1, 127546, '\P{^Decomposition_Type=--Square}', "");
Expect(0, 127547, '\p{Decomposition_Type=--Square}', "");
Expect(1, 127547, '\p{^Decomposition_Type=--Square}', "");
Expect(1, 127547, '\P{Decomposition_Type=--Square}', "");
Expect(0, 127547, '\P{^Decomposition_Type=--Square}', "");
Error('\p{Dt=-_Sqr/a/}');
Error('\P{Dt=-_Sqr/a/}');
Expect(1, 127546, '\p{Dt=sqr}', "");
Expect(0, 127546, '\p{^Dt=sqr}', "");
Expect(0, 127546, '\P{Dt=sqr}', "");
Expect(1, 127546, '\P{^Dt=sqr}', "");
Expect(0, 127547, '\p{Dt=sqr}', "");
Expect(1, 127547, '\p{^Dt=sqr}', "");
Expect(1, 127547, '\P{Dt=sqr}', "");
Expect(0, 127547, '\P{^Dt=sqr}', "");
Expect(1, 127546, '\p{Dt=	_SQR}', "");
Expect(0, 127546, '\p{^Dt=	_SQR}', "");
Expect(0, 127546, '\P{Dt=	_SQR}', "");
Expect(1, 127546, '\P{^Dt=	_SQR}', "");
Expect(0, 127547, '\p{Dt=	_SQR}', "");
Expect(1, 127547, '\p{^Dt=	_SQR}', "");
Expect(1, 127547, '\P{Dt=	_SQR}', "");
Expect(0, 127547, '\P{^Dt=	_SQR}', "");
Error('\p{Is_Decomposition_Type=_ Square:=}');
Error('\P{Is_Decomposition_Type=_ Square:=}');
Expect(1, 127546, '\p{Is_Decomposition_Type=square}', "");
Expect(0, 127546, '\p{^Is_Decomposition_Type=square}', "");
Expect(0, 127546, '\P{Is_Decomposition_Type=square}', "");
Expect(1, 127546, '\P{^Is_Decomposition_Type=square}', "");
Expect(0, 127547, '\p{Is_Decomposition_Type=square}', "");
Expect(1, 127547, '\p{^Is_Decomposition_Type=square}', "");
Expect(1, 127547, '\P{Is_Decomposition_Type=square}', "");
Expect(0, 127547, '\P{^Is_Decomposition_Type=square}', "");
Expect(1, 127546, '\p{Is_Decomposition_Type=-	Square}', "");
Expect(0, 127546, '\p{^Is_Decomposition_Type=-	Square}', "");
Expect(0, 127546, '\P{Is_Decomposition_Type=-	Square}', "");
Expect(1, 127546, '\P{^Is_Decomposition_Type=-	Square}', "");
Expect(0, 127547, '\p{Is_Decomposition_Type=-	Square}', "");
Expect(1, 127547, '\p{^Is_Decomposition_Type=-	Square}', "");
Expect(1, 127547, '\P{Is_Decomposition_Type=-	Square}', "");
Expect(0, 127547, '\P{^Is_Decomposition_Type=-	Square}', "");
Error('\p{Is_Dt=SQR/a/}');
Error('\P{Is_Dt=SQR/a/}');
Expect(1, 127546, '\p{Is_Dt=sqr}', "");
Expect(0, 127546, '\p{^Is_Dt=sqr}', "");
Expect(0, 127546, '\P{Is_Dt=sqr}', "");
Expect(1, 127546, '\P{^Is_Dt=sqr}', "");
Expect(0, 127547, '\p{Is_Dt=sqr}', "");
Expect(1, 127547, '\p{^Is_Dt=sqr}', "");
Expect(1, 127547, '\P{Is_Dt=sqr}', "");
Expect(0, 127547, '\P{^Is_Dt=sqr}', "");
Expect(1, 127546, '\p{Is_Dt=  Sqr}', "");
Expect(0, 127546, '\p{^Is_Dt=  Sqr}', "");
Expect(0, 127546, '\P{Is_Dt=  Sqr}', "");
Expect(1, 127546, '\P{^Is_Dt=  Sqr}', "");
Expect(0, 127547, '\p{Is_Dt=  Sqr}', "");
Expect(1, 127547, '\p{^Is_Dt=  Sqr}', "");
Expect(1, 127547, '\P{Is_Dt=  Sqr}', "");
Expect(0, 127547, '\P{^Is_Dt=  Sqr}', "");
Error('\p{Decomposition_Type=/a/_Small}');
Error('\P{Decomposition_Type=/a/_Small}');
Expect(1, 65131, '\p{Decomposition_Type: small}', "");
Expect(0, 65131, '\p{^Decomposition_Type: small}', "");
Expect(0, 65131, '\P{Decomposition_Type: small}', "");
Expect(1, 65131, '\P{^Decomposition_Type: small}', "");
Expect(0, 65132, '\p{Decomposition_Type: small}', "");
Expect(1, 65132, '\p{^Decomposition_Type: small}', "");
Expect(1, 65132, '\P{Decomposition_Type: small}', "");
Expect(0, 65132, '\P{^Decomposition_Type: small}', "");
Expect(1, 65131, '\p{Decomposition_Type= _small}', "");
Expect(0, 65131, '\p{^Decomposition_Type= _small}', "");
Expect(0, 65131, '\P{Decomposition_Type= _small}', "");
Expect(1, 65131, '\P{^Decomposition_Type= _small}', "");
Expect(0, 65132, '\p{Decomposition_Type= _small}', "");
Expect(1, 65132, '\p{^Decomposition_Type= _small}', "");
Expect(1, 65132, '\P{Decomposition_Type= _small}', "");
Expect(0, 65132, '\P{^Decomposition_Type= _small}', "");
Error('\p{Dt=/a/ _SML}');
Error('\P{Dt=/a/ _SML}');
Expect(1, 65131, '\p{Dt=sml}', "");
Expect(0, 65131, '\p{^Dt=sml}', "");
Expect(0, 65131, '\P{Dt=sml}', "");
Expect(1, 65131, '\P{^Dt=sml}', "");
Expect(0, 65132, '\p{Dt=sml}', "");
Expect(1, 65132, '\p{^Dt=sml}', "");
Expect(1, 65132, '\P{Dt=sml}', "");
Expect(0, 65132, '\P{^Dt=sml}', "");
Expect(1, 65131, '\p{Dt= Sml}', "");
Expect(0, 65131, '\p{^Dt= Sml}', "");
Expect(0, 65131, '\P{Dt= Sml}', "");
Expect(1, 65131, '\P{^Dt= Sml}', "");
Expect(0, 65132, '\p{Dt= Sml}', "");
Expect(1, 65132, '\p{^Dt= Sml}', "");
Expect(1, 65132, '\P{Dt= Sml}', "");
Expect(0, 65132, '\P{^Dt= Sml}', "");
Error('\p{Is_Decomposition_Type:   /a/small}');
Error('\P{Is_Decomposition_Type:   /a/small}');
Expect(1, 65131, '\p{Is_Decomposition_Type=small}', "");
Expect(0, 65131, '\p{^Is_Decomposition_Type=small}', "");
Expect(0, 65131, '\P{Is_Decomposition_Type=small}', "");
Expect(1, 65131, '\P{^Is_Decomposition_Type=small}', "");
Expect(0, 65132, '\p{Is_Decomposition_Type=small}', "");
Expect(1, 65132, '\p{^Is_Decomposition_Type=small}', "");
Expect(1, 65132, '\P{Is_Decomposition_Type=small}', "");
Expect(0, 65132, '\P{^Is_Decomposition_Type=small}', "");
Expect(1, 65131, '\p{Is_Decomposition_Type= _small}', "");
Expect(0, 65131, '\p{^Is_Decomposition_Type= _small}', "");
Expect(0, 65131, '\P{Is_Decomposition_Type= _small}', "");
Expect(1, 65131, '\P{^Is_Decomposition_Type= _small}', "");
Expect(0, 65132, '\p{Is_Decomposition_Type= _small}', "");
Expect(1, 65132, '\p{^Is_Decomposition_Type= _small}', "");
Expect(1, 65132, '\P{Is_Decomposition_Type= _small}', "");
Expect(0, 65132, '\P{^Is_Decomposition_Type= _small}', "");
Error('\p{Is_Dt=_sml:=}');
Error('\P{Is_Dt=_sml:=}');
Expect(1, 65131, '\p{Is_Dt=sml}', "");
Expect(0, 65131, '\p{^Is_Dt=sml}', "");
Expect(0, 65131, '\P{Is_Dt=sml}', "");
Expect(1, 65131, '\P{^Is_Dt=sml}', "");
Expect(0, 65132, '\p{Is_Dt=sml}', "");
Expect(1, 65132, '\p{^Is_Dt=sml}', "");
Expect(1, 65132, '\P{Is_Dt=sml}', "");
Expect(0, 65132, '\P{^Is_Dt=sml}', "");
Expect(1, 65131, '\p{Is_Dt= Sml}', "");
Expect(0, 65131, '\p{^Is_Dt= Sml}', "");
Expect(0, 65131, '\P{Is_Dt= Sml}', "");
Expect(1, 65131, '\P{^Is_Dt= Sml}', "");
Expect(0, 65132, '\p{Is_Dt= Sml}', "");
Expect(1, 65132, '\p{^Is_Dt= Sml}', "");
Expect(1, 65132, '\P{Is_Dt= Sml}', "");
Expect(0, 65132, '\P{^Is_Dt= Sml}', "");
Error('\p{Decomposition_Type=:=-_none}');
Error('\P{Decomposition_Type=:=-_none}');
Expect(1, 195102, '\p{Decomposition_Type=none}', "");
Expect(0, 195102, '\p{^Decomposition_Type=none}', "");
Expect(0, 195102, '\P{Decomposition_Type=none}', "");
Expect(1, 195102, '\P{^Decomposition_Type=none}', "");
Expect(0, 195101, '\p{Decomposition_Type=none}', "");
Expect(1, 195101, '\p{^Decomposition_Type=none}', "");
Expect(1, 195101, '\P{Decomposition_Type=none}', "");
Expect(0, 195101, '\P{^Decomposition_Type=none}', "");
Expect(1, 195102, '\p{Decomposition_Type= 	None}', "");
Expect(0, 195102, '\p{^Decomposition_Type= 	None}', "");
Expect(0, 195102, '\P{Decomposition_Type= 	None}', "");
Expect(1, 195102, '\P{^Decomposition_Type= 	None}', "");
Expect(0, 195101, '\p{Decomposition_Type= 	None}', "");
Expect(1, 195101, '\p{^Decomposition_Type= 	None}', "");
Expect(1, 195101, '\P{Decomposition_Type= 	None}', "");
Expect(0, 195101, '\P{^Decomposition_Type= 	None}', "");
Error('\p{Dt=-_None:=}');
Error('\P{Dt=-_None:=}');
Expect(1, 195102, '\p{Dt=none}', "");
Expect(0, 195102, '\p{^Dt=none}', "");
Expect(0, 195102, '\P{Dt=none}', "");
Expect(1, 195102, '\P{^Dt=none}', "");
Expect(0, 195101, '\p{Dt=none}', "");
Expect(1, 195101, '\p{^Dt=none}', "");
Expect(1, 195101, '\P{Dt=none}', "");
Expect(0, 195101, '\P{^Dt=none}', "");
Expect(1, 195102, '\p{Dt=_	None}', "");
Expect(0, 195102, '\p{^Dt=_	None}', "");
Expect(0, 195102, '\P{Dt=_	None}', "");
Expect(1, 195102, '\P{^Dt=_	None}', "");
Expect(0, 195101, '\p{Dt=_	None}', "");
Expect(1, 195101, '\p{^Dt=_	None}', "");
Expect(1, 195101, '\P{Dt=_	None}', "");
Expect(0, 195101, '\P{^Dt=_	None}', "");
Error('\p{Is_Decomposition_Type=	/a/None}');
Error('\P{Is_Decomposition_Type=	/a/None}');
Expect(1, 195102, '\p{Is_Decomposition_Type=none}', "");
Expect(0, 195102, '\p{^Is_Decomposition_Type=none}', "");
Expect(0, 195102, '\P{Is_Decomposition_Type=none}', "");
Expect(1, 195102, '\P{^Is_Decomposition_Type=none}', "");
Expect(0, 195101, '\p{Is_Decomposition_Type=none}', "");
Expect(1, 195101, '\p{^Is_Decomposition_Type=none}', "");
Expect(1, 195101, '\P{Is_Decomposition_Type=none}', "");
Expect(0, 195101, '\P{^Is_Decomposition_Type=none}', "");
Expect(1, 195102, '\p{Is_Decomposition_Type: 	 None}', "");
Expect(0, 195102, '\p{^Is_Decomposition_Type: 	 None}', "");
Expect(0, 195102, '\P{Is_Decomposition_Type: 	 None}', "");
Expect(1, 195102, '\P{^Is_Decomposition_Type: 	 None}', "");
Expect(0, 195101, '\p{Is_Decomposition_Type: 	 None}', "");
Expect(1, 195101, '\p{^Is_Decomposition_Type: 	 None}', "");
Expect(1, 195101, '\P{Is_Decomposition_Type: 	 None}', "");
Expect(0, 195101, '\P{^Is_Decomposition_Type: 	 None}', "");
Error('\p{Is_Dt=:=-_none}');
Error('\P{Is_Dt=:=-_none}');
Expect(1, 195102, '\p{Is_Dt:	none}', "");
Expect(0, 195102, '\p{^Is_Dt:	none}', "");
Expect(0, 195102, '\P{Is_Dt:	none}', "");
Expect(1, 195102, '\P{^Is_Dt:	none}', "");
Expect(0, 195101, '\p{Is_Dt:	none}', "");
Expect(1, 195101, '\p{^Is_Dt:	none}', "");
Expect(1, 195101, '\P{Is_Dt:	none}', "");
Expect(0, 195101, '\P{^Is_Dt:	none}', "");
Expect(1, 195102, '\p{Is_Dt= 	none}', "");
Expect(0, 195102, '\p{^Is_Dt= 	none}', "");
Expect(0, 195102, '\P{Is_Dt= 	none}', "");
Expect(1, 195102, '\P{^Is_Dt= 	none}', "");
Expect(0, 195101, '\p{Is_Dt= 	none}', "");
Expect(1, 195101, '\p{^Is_Dt= 	none}', "");
Expect(1, 195101, '\P{Is_Dt= 	none}', "");
Expect(0, 195101, '\P{^Is_Dt= 	none}', "");
Error('\p{Decomposition_Type:	 Isolated/a/}');
Error('\P{Decomposition_Type:	 Isolated/a/}');
Expect(1, 65275, '\p{Decomposition_Type=isolated}', "");
Expect(0, 65275, '\p{^Decomposition_Type=isolated}', "");
Expect(0, 65275, '\P{Decomposition_Type=isolated}', "");
Expect(1, 65275, '\P{^Decomposition_Type=isolated}', "");
Expect(0, 65276, '\p{Decomposition_Type=isolated}', "");
Expect(1, 65276, '\p{^Decomposition_Type=isolated}', "");
Expect(1, 65276, '\P{Decomposition_Type=isolated}', "");
Expect(0, 65276, '\P{^Decomposition_Type=isolated}', "");
Expect(1, 65275, '\p{Decomposition_Type=_ISOLATED}', "");
Expect(0, 65275, '\p{^Decomposition_Type=_ISOLATED}', "");
Expect(0, 65275, '\P{Decomposition_Type=_ISOLATED}', "");
Expect(1, 65275, '\P{^Decomposition_Type=_ISOLATED}', "");
Expect(0, 65276, '\p{Decomposition_Type=_ISOLATED}', "");
Expect(1, 65276, '\p{^Decomposition_Type=_ISOLATED}', "");
Expect(1, 65276, '\P{Decomposition_Type=_ISOLATED}', "");
Expect(0, 65276, '\P{^Decomposition_Type=_ISOLATED}', "");
Error('\p{Dt=/a/Iso}');
Error('\P{Dt=/a/Iso}');
Expect(1, 65275, '\p{Dt=iso}', "");
Expect(0, 65275, '\p{^Dt=iso}', "");
Expect(0, 65275, '\P{Dt=iso}', "");
Expect(1, 65275, '\P{^Dt=iso}', "");
Expect(0, 65276, '\p{Dt=iso}', "");
Expect(1, 65276, '\p{^Dt=iso}', "");
Expect(1, 65276, '\P{Dt=iso}', "");
Expect(0, 65276, '\P{^Dt=iso}', "");
Expect(1, 65275, '\p{Dt=	 iso}', "");
Expect(0, 65275, '\p{^Dt=	 iso}', "");
Expect(0, 65275, '\P{Dt=	 iso}', "");
Expect(1, 65275, '\P{^Dt=	 iso}', "");
Expect(0, 65276, '\p{Dt=	 iso}', "");
Expect(1, 65276, '\p{^Dt=	 iso}', "");
Expect(1, 65276, '\P{Dt=	 iso}', "");
Expect(0, 65276, '\P{^Dt=	 iso}', "");
Error('\p{Is_Decomposition_Type=_/a/Isolated}');
Error('\P{Is_Decomposition_Type=_/a/Isolated}');
Expect(1, 65275, '\p{Is_Decomposition_Type:   isolated}', "");
Expect(0, 65275, '\p{^Is_Decomposition_Type:   isolated}', "");
Expect(0, 65275, '\P{Is_Decomposition_Type:   isolated}', "");
Expect(1, 65275, '\P{^Is_Decomposition_Type:   isolated}', "");
Expect(0, 65276, '\p{Is_Decomposition_Type:   isolated}', "");
Expect(1, 65276, '\p{^Is_Decomposition_Type:   isolated}', "");
Expect(1, 65276, '\P{Is_Decomposition_Type:   isolated}', "");
Expect(0, 65276, '\P{^Is_Decomposition_Type:   isolated}', "");
Expect(1, 65275, '\p{Is_Decomposition_Type=-_ISOLATED}', "");
Expect(0, 65275, '\p{^Is_Decomposition_Type=-_ISOLATED}', "");
Expect(0, 65275, '\P{Is_Decomposition_Type=-_ISOLATED}', "");
Expect(1, 65275, '\P{^Is_Decomposition_Type=-_ISOLATED}', "");
Expect(0, 65276, '\p{Is_Decomposition_Type=-_ISOLATED}', "");
Expect(1, 65276, '\p{^Is_Decomposition_Type=-_ISOLATED}', "");
Expect(1, 65276, '\P{Is_Decomposition_Type=-_ISOLATED}', "");
Expect(0, 65276, '\P{^Is_Decomposition_Type=-_ISOLATED}', "");
Error('\p{Is_Dt=/a/__ISO}');
Error('\P{Is_Dt=/a/__ISO}');
Expect(1, 65275, '\p{Is_Dt:	iso}', "");
Expect(0, 65275, '\p{^Is_Dt:	iso}', "");
Expect(0, 65275, '\P{Is_Dt:	iso}', "");
Expect(1, 65275, '\P{^Is_Dt:	iso}', "");
Expect(0, 65276, '\p{Is_Dt:	iso}', "");
Expect(1, 65276, '\p{^Is_Dt:	iso}', "");
Expect(1, 65276, '\P{Is_Dt:	iso}', "");
Expect(0, 65276, '\P{^Is_Dt:	iso}', "");
Expect(1, 65275, '\p{Is_Dt=  iso}', "");
Expect(0, 65275, '\p{^Is_Dt=  iso}', "");
Expect(0, 65275, '\P{Is_Dt=  iso}', "");
Expect(1, 65275, '\P{^Is_Dt=  iso}', "");
Expect(0, 65276, '\p{Is_Dt=  iso}', "");
Expect(1, 65276, '\p{^Is_Dt=  iso}', "");
Expect(1, 65276, '\P{Is_Dt=  iso}', "");
Expect(0, 65276, '\P{^Is_Dt=  iso}', "");
Error('\p{Decomposition_Type=:=-Compat}');
Error('\P{Decomposition_Type=:=-Compat}');
Expect(1, 127560, '\p{Decomposition_Type=compat}', "");
Expect(0, 127560, '\p{^Decomposition_Type=compat}', "");
Expect(0, 127560, '\P{Decomposition_Type=compat}', "");
Expect(1, 127560, '\P{^Decomposition_Type=compat}', "");
Expect(0, 127561, '\p{Decomposition_Type=compat}', "");
Expect(1, 127561, '\p{^Decomposition_Type=compat}', "");
Expect(1, 127561, '\P{Decomposition_Type=compat}', "");
Expect(0, 127561, '\P{^Decomposition_Type=compat}', "");
Expect(1, 127560, '\p{Decomposition_Type:	  COMPAT}', "");
Expect(0, 127560, '\p{^Decomposition_Type:	  COMPAT}', "");
Expect(0, 127560, '\P{Decomposition_Type:	  COMPAT}', "");
Expect(1, 127560, '\P{^Decomposition_Type:	  COMPAT}', "");
Expect(0, 127561, '\p{Decomposition_Type:	  COMPAT}', "");
Expect(1, 127561, '\p{^Decomposition_Type:	  COMPAT}', "");
Expect(1, 127561, '\P{Decomposition_Type:	  COMPAT}', "");
Expect(0, 127561, '\P{^Decomposition_Type:	  COMPAT}', "");
Error('\p{Dt=--com:=}');
Error('\P{Dt=--com:=}');
Expect(1, 127560, '\p{Dt=com}', "");
Expect(0, 127560, '\p{^Dt=com}', "");
Expect(0, 127560, '\P{Dt=com}', "");
Expect(1, 127560, '\P{^Dt=com}', "");
Expect(0, 127561, '\p{Dt=com}', "");
Expect(1, 127561, '\p{^Dt=com}', "");
Expect(1, 127561, '\P{Dt=com}', "");
Expect(0, 127561, '\P{^Dt=com}', "");
Expect(1, 127560, '\p{Dt=--com}', "");
Expect(0, 127560, '\p{^Dt=--com}', "");
Expect(0, 127560, '\P{Dt=--com}', "");
Expect(1, 127560, '\P{^Dt=--com}', "");
Expect(0, 127561, '\p{Dt=--com}', "");
Expect(1, 127561, '\p{^Dt=--com}', "");
Expect(1, 127561, '\P{Dt=--com}', "");
Expect(0, 127561, '\P{^Dt=--com}', "");
Error('\p{Is_Decomposition_Type=	Compat/a/}');
Error('\P{Is_Decomposition_Type=	Compat/a/}');
Expect(1, 127560, '\p{Is_Decomposition_Type=compat}', "");
Expect(0, 127560, '\p{^Is_Decomposition_Type=compat}', "");
Expect(0, 127560, '\P{Is_Decomposition_Type=compat}', "");
Expect(1, 127560, '\P{^Is_Decomposition_Type=compat}', "");
Expect(0, 127561, '\p{Is_Decomposition_Type=compat}', "");
Expect(1, 127561, '\p{^Is_Decomposition_Type=compat}', "");
Expect(1, 127561, '\P{Is_Decomposition_Type=compat}', "");
Expect(0, 127561, '\P{^Is_Decomposition_Type=compat}', "");
Expect(1, 127560, '\p{Is_Decomposition_Type=	Compat}', "");
Expect(0, 127560, '\p{^Is_Decomposition_Type=	Compat}', "");
Expect(0, 127560, '\P{Is_Decomposition_Type=	Compat}', "");
Expect(1, 127560, '\P{^Is_Decomposition_Type=	Compat}', "");
Expect(0, 127561, '\p{Is_Decomposition_Type=	Compat}', "");
Expect(1, 127561, '\p{^Is_Decomposition_Type=	Compat}', "");
Expect(1, 127561, '\P{Is_Decomposition_Type=	Compat}', "");
Expect(0, 127561, '\P{^Is_Decomposition_Type=	Compat}', "");
Error('\p{Is_Dt= -Com/a/}');
Error('\P{Is_Dt= -Com/a/}');
Expect(1, 127560, '\p{Is_Dt=com}', "");
Expect(0, 127560, '\p{^Is_Dt=com}', "");
Expect(0, 127560, '\P{Is_Dt=com}', "");
Expect(1, 127560, '\P{^Is_Dt=com}', "");
Expect(0, 127561, '\p{Is_Dt=com}', "");
Expect(1, 127561, '\p{^Is_Dt=com}', "");
Expect(1, 127561, '\P{Is_Dt=com}', "");
Expect(0, 127561, '\P{^Is_Dt=com}', "");
Expect(1, 127560, '\p{Is_Dt=	com}', "");
Expect(0, 127560, '\p{^Is_Dt=	com}', "");
Expect(0, 127560, '\P{Is_Dt=	com}', "");
Expect(1, 127560, '\P{^Is_Dt=	com}', "");
Expect(0, 127561, '\p{Is_Dt=	com}', "");
Expect(1, 127561, '\p{^Is_Dt=	com}', "");
Expect(1, 127561, '\P{Is_Dt=	com}', "");
Expect(0, 127561, '\P{^Is_Dt=	com}', "");
Error('\p{Decomposition_Type:	:=  super}');
Error('\P{Decomposition_Type:	:=  super}');
Expect(1, 127339, '\p{Decomposition_Type=super}', "");
Expect(0, 127339, '\p{^Decomposition_Type=super}', "");
Expect(0, 127339, '\P{Decomposition_Type=super}', "");
Expect(1, 127339, '\P{^Decomposition_Type=super}', "");
Expect(0, 127340, '\p{Decomposition_Type=super}', "");
Expect(1, 127340, '\p{^Decomposition_Type=super}', "");
Expect(1, 127340, '\P{Decomposition_Type=super}', "");
Expect(0, 127340, '\P{^Decomposition_Type=super}', "");
Expect(1, 127339, '\p{Decomposition_Type=-_super}', "");
Expect(0, 127339, '\p{^Decomposition_Type=-_super}', "");
Expect(0, 127339, '\P{Decomposition_Type=-_super}', "");
Expect(1, 127339, '\P{^Decomposition_Type=-_super}', "");
Expect(0, 127340, '\p{Decomposition_Type=-_super}', "");
Expect(1, 127340, '\p{^Decomposition_Type=-_super}', "");
Expect(1, 127340, '\P{Decomposition_Type=-_super}', "");
Expect(0, 127340, '\P{^Decomposition_Type=-_super}', "");
Error('\p{Dt=/a/_SUP}');
Error('\P{Dt=/a/_SUP}');
Expect(1, 127339, '\p{Dt=sup}', "");
Expect(0, 127339, '\p{^Dt=sup}', "");
Expect(0, 127339, '\P{Dt=sup}', "");
Expect(1, 127339, '\P{^Dt=sup}', "");
Expect(0, 127340, '\p{Dt=sup}', "");
Expect(1, 127340, '\p{^Dt=sup}', "");
Expect(1, 127340, '\P{Dt=sup}', "");
Expect(0, 127340, '\P{^Dt=sup}', "");
Expect(1, 127339, '\p{Dt=		SUP}', "");
Expect(0, 127339, '\p{^Dt=		SUP}', "");
Expect(0, 127339, '\P{Dt=		SUP}', "");
Expect(1, 127339, '\P{^Dt=		SUP}', "");
Expect(0, 127340, '\p{Dt=		SUP}', "");
Expect(1, 127340, '\p{^Dt=		SUP}', "");
Expect(1, 127340, '\P{Dt=		SUP}', "");
Expect(0, 127340, '\P{^Dt=		SUP}', "");
Error('\p{Is_Decomposition_Type=/a/_Super}');
Error('\P{Is_Decomposition_Type=/a/_Super}');
Expect(1, 127339, '\p{Is_Decomposition_Type=super}', "");
Expect(0, 127339, '\p{^Is_Decomposition_Type=super}', "");
Expect(0, 127339, '\P{Is_Decomposition_Type=super}', "");
Expect(1, 127339, '\P{^Is_Decomposition_Type=super}', "");
Expect(0, 127340, '\p{Is_Decomposition_Type=super}', "");
Expect(1, 127340, '\p{^Is_Decomposition_Type=super}', "");
Expect(1, 127340, '\P{Is_Decomposition_Type=super}', "");
Expect(0, 127340, '\P{^Is_Decomposition_Type=super}', "");
Expect(1, 127339, '\p{Is_Decomposition_Type: -Super}', "");
Expect(0, 127339, '\p{^Is_Decomposition_Type: -Super}', "");
Expect(0, 127339, '\P{Is_Decomposition_Type: -Super}', "");
Expect(1, 127339, '\P{^Is_Decomposition_Type: -Super}', "");
Expect(0, 127340, '\p{Is_Decomposition_Type: -Super}', "");
Expect(1, 127340, '\p{^Is_Decomposition_Type: -Super}', "");
Expect(1, 127340, '\P{Is_Decomposition_Type: -Super}', "");
Expect(0, 127340, '\P{^Is_Decomposition_Type: -Super}', "");
Error('\p{Is_Dt=_	sup:=}');
Error('\P{Is_Dt=_	sup:=}');
Expect(1, 127339, '\p{Is_Dt=sup}', "");
Expect(0, 127339, '\p{^Is_Dt=sup}', "");
Expect(0, 127339, '\P{Is_Dt=sup}', "");
Expect(1, 127339, '\P{^Is_Dt=sup}', "");
Expect(0, 127340, '\p{Is_Dt=sup}', "");
Expect(1, 127340, '\p{^Is_Dt=sup}', "");
Expect(1, 127340, '\P{Is_Dt=sup}', "");
Expect(0, 127340, '\P{^Is_Dt=sup}', "");
Expect(1, 127339, '\p{Is_Dt:   _sup}', "");
Expect(0, 127339, '\p{^Is_Dt:   _sup}', "");
Expect(0, 127339, '\P{Is_Dt:   _sup}', "");
Expect(1, 127339, '\P{^Is_Dt:   _sup}', "");
Expect(0, 127340, '\p{Is_Dt:   _sup}', "");
Expect(1, 127340, '\p{^Is_Dt:   _sup}', "");
Expect(1, 127340, '\P{Is_Dt:   _sup}', "");
Expect(0, 127340, '\P{^Is_Dt:   _sup}', "");
Error('\p{Jamo_Short_Name=S}');
Error('\P{Jamo_Short_Name=S}');
Error('\p{JSN=S}');
Error('\P{JSN=S}');
Error('\p{Is_Jamo_Short_Name=S}');
Error('\P{Is_Jamo_Short_Name=S}');
Error('\p{Is_JSN=S}');
Error('\P{Is_JSN=S}');
Error('\p{Jamo_Short_Name=GG}');
Error('\P{Jamo_Short_Name=GG}');
Error('\p{JSN:   GG}');
Error('\P{JSN:   GG}');
Error('\p{Is_Jamo_Short_Name=GG}');
Error('\P{Is_Jamo_Short_Name=GG}');
Error('\p{Is_JSN=GG}');
Error('\P{Is_JSN=GG}');
Error('\p{Jamo_Short_Name=YU}');
Error('\P{Jamo_Short_Name=YU}');
Error('\p{JSN=YU}');
Error('\P{JSN=YU}');
Error('\p{Is_Jamo_Short_Name=YU}');
Error('\P{Is_Jamo_Short_Name=YU}');
Error('\p{Is_JSN=YU}');
Error('\P{Is_JSN=YU}');
Error('\p{Jamo_Short_Name=WE}');
Error('\P{Jamo_Short_Name=WE}');
Error('\p{JSN=WE}');
Error('\P{JSN=WE}');
Error('\p{Is_Jamo_Short_Name=WE}');
Error('\P{Is_Jamo_Short_Name=WE}');
Error('\p{Is_JSN:   WE}');
Error('\P{Is_JSN:   WE}');
Error('\p{Jamo_Short_Name=LM}');
Error('\P{Jamo_Short_Name=LM}');
Error('\p{JSN=LM}');
Error('\P{JSN=LM}');
Error('\p{Is_Jamo_Short_Name=LM}');
Error('\P{Is_Jamo_Short_Name=LM}');
Error('\p{Is_JSN=LM}');
Error('\P{Is_JSN=LM}');
Error('\p{Jamo_Short_Name=K}');
Error('\P{Jamo_Short_Name=K}');
Error('\p{JSN=K}');
Error('\P{JSN=K}');
Error('\p{Is_Jamo_Short_Name=K}');
Error('\P{Is_Jamo_Short_Name=K}');
Error('\p{Is_JSN=K}');
Error('\P{Is_JSN=K}');
Error('\p{Jamo_Short_Name:   BB}');
Error('\P{Jamo_Short_Name:   BB}');
Error('\p{JSN=BB}');
Error('\P{JSN=BB}');
Error('\p{Is_Jamo_Short_Name=BB}');
Error('\P{Is_Jamo_Short_Name=BB}');
Error('\p{Is_JSN=BB}');
Error('\P{Is_JSN=BB}');
Error('\p{Jamo_Short_Name=D}');
Error('\P{Jamo_Short_Name=D}');
Error('\p{JSN=D}');
Error('\P{JSN=D}');
Error('\p{Is_Jamo_Short_Name=D}');
Error('\P{Is_Jamo_Short_Name=D}');
Error('\p{Is_JSN=D}');
Error('\P{Is_JSN=D}');
Error('\p{Jamo_Short_Name:   E}');
Error('\P{Jamo_Short_Name:   E}');
Error('\p{JSN=E}');
Error('\P{JSN=E}');
Error('\p{Is_Jamo_Short_Name:E}');
Error('\P{Is_Jamo_Short_Name:E}');
Error('\p{Is_JSN=E}');
Error('\P{Is_JSN=E}');
Error('\p{Jamo_Short_Name=JJ}');
Error('\P{Jamo_Short_Name=JJ}');
Error('\p{JSN=JJ}');
Error('\P{JSN=JJ}');
Error('\p{Is_Jamo_Short_Name=JJ}');
Error('\P{Is_Jamo_Short_Name=JJ}');
Error('\p{Is_JSN:   JJ}');
Error('\P{Is_JSN:   JJ}');
Error('\p{Jamo_Short_Name=G}');
Error('\P{Jamo_Short_Name=G}');
Error('\p{JSN=G}');
Error('\P{JSN=G}');
Error('\p{Is_Jamo_Short_Name=G}');
Error('\P{Is_Jamo_Short_Name=G}');
Error('\p{Is_JSN=G}');
Error('\P{Is_JSN=G}');
Error('\p{Jamo_Short_Name=YA}');
Error('\P{Jamo_Short_Name=YA}');
Error('\p{JSN=YA}');
Error('\P{JSN=YA}');
Error('\p{Is_Jamo_Short_Name=YA}');
Error('\P{Is_Jamo_Short_Name=YA}');
Error('\p{Is_JSN=YA}');
Error('\P{Is_JSN=YA}');
Error('\p{Jamo_Short_Name=SS}');
Error('\P{Jamo_Short_Name=SS}');
Error('\p{JSN=SS}');
Error('\P{JSN=SS}');
Error('\p{Is_Jamo_Short_Name=SS}');
Error('\P{Is_Jamo_Short_Name=SS}');
Error('\p{Is_JSN=SS}');
Error('\P{Is_JSN=SS}');
Error('\p{Jamo_Short_Name=LB}');
Error('\P{Jamo_Short_Name=LB}');
Error('\p{JSN=LB}');
Error('\P{JSN=LB}');
Error('\p{Is_Jamo_Short_Name=LB}');
Error('\P{Is_Jamo_Short_Name=LB}');
Error('\p{Is_JSN=LB}');
Error('\P{Is_JSN=LB}');
Error('\p{Jamo_Short_Name=J}');
Error('\P{Jamo_Short_Name=J}');
Error('\p{JSN=J}');
Error('\P{JSN=J}');
Error('\p{Is_Jamo_Short_Name:   J}');
Error('\P{Is_Jamo_Short_Name:   J}');
Error('\p{Is_JSN=J}');
Error('\P{Is_JSN=J}');
Error('\p{Jamo_Short_Name=NH}');
Error('\P{Jamo_Short_Name=NH}');
Error('\p{JSN=NH}');
Error('\P{JSN=NH}');
Error('\p{Is_Jamo_Short_Name=NH}');
Error('\P{Is_Jamo_Short_Name=NH}');
Error('\p{Is_JSN=NH}');
Error('\P{Is_JSN=NH}');
Error('\p{Jamo_Short_Name=YEO}');
Error('\P{Jamo_Short_Name=YEO}');
Error('\p{JSN=YEO}');
Error('\P{JSN=YEO}');
Error('\p{Is_Jamo_Short_Name=YEO}');
Error('\P{Is_Jamo_Short_Name=YEO}');
Error('\p{Is_JSN=YEO}');
Error('\P{Is_JSN=YEO}');
Error('\p{Jamo_Short_Name=LP}');
Error('\P{Jamo_Short_Name=LP}');
Error('\p{JSN=LP}');
Error('\P{JSN=LP}');
Error('\p{Is_Jamo_Short_Name=LP}');
Error('\P{Is_Jamo_Short_Name=LP}');
Error('\p{Is_JSN=LP}');
Error('\P{Is_JSN=LP}');
Error('\p{Jamo_Short_Name=NJ}');
Error('\P{Jamo_Short_Name=NJ}');
Error('\p{JSN:   NJ}');
Error('\P{JSN:   NJ}');
Error('\p{Is_Jamo_Short_Name:NJ}');
Error('\P{Is_Jamo_Short_Name:NJ}');
Error('\p{Is_JSN=NJ}');
Error('\P{Is_JSN=NJ}');
Error('\p{Jamo_Short_Name=B}');
Error('\P{Jamo_Short_Name=B}');
Error('\p{JSN=B}');
Error('\P{JSN=B}');
Error('\p{Is_Jamo_Short_Name=B}');
Error('\P{Is_Jamo_Short_Name=B}');
Error('\p{Is_JSN=B}');
Error('\P{Is_JSN=B}');
Error('\p{Jamo_Short_Name=WA}');
Error('\P{Jamo_Short_Name=WA}');
Error('\p{JSN=WA}');
Error('\P{JSN=WA}');
Error('\p{Is_Jamo_Short_Name=WA}');
Error('\P{Is_Jamo_Short_Name=WA}');
Error('\p{Is_JSN:WA}');
Error('\P{Is_JSN:WA}');
Error('\p{Jamo_Short_Name=LH}');
Error('\P{Jamo_Short_Name=LH}');
Error('\p{JSN=LH}');
Error('\P{JSN=LH}');
Error('\p{Is_Jamo_Short_Name=LH}');
Error('\P{Is_Jamo_Short_Name=LH}');
Error('\p{Is_JSN=LH}');
Error('\P{Is_JSN=LH}');
Error('\p{Jamo_Short_Name=YI}');
Error('\P{Jamo_Short_Name=YI}');
Error('\p{JSN=YI}');
Error('\P{JSN=YI}');
Error('\p{Is_Jamo_Short_Name=YI}');
Error('\P{Is_Jamo_Short_Name=YI}');
Error('\p{Is_JSN=YI}');
Error('\P{Is_JSN=YI}');
Error('\p{Jamo_Short_Name=WEO}');
Error('\P{Jamo_Short_Name=WEO}');
Error('\p{JSN=WEO}');
Error('\P{JSN=WEO}');
Error('\p{Is_Jamo_Short_Name=WEO}');
Error('\P{Is_Jamo_Short_Name=WEO}');
Error('\p{Is_JSN=WEO}');
Error('\P{Is_JSN=WEO}');
Error('\p{Jamo_Short_Name=LS}');
Error('\P{Jamo_Short_Name=LS}');
Error('\p{JSN:   LS}');
Error('\P{JSN:   LS}');
Error('\p{Is_Jamo_Short_Name=LS}');
Error('\P{Is_Jamo_Short_Name=LS}');
Error('\p{Is_JSN=LS}');
Error('\P{Is_JSN=LS}');
Error('\p{Jamo_Short_Name=BS}');
Error('\P{Jamo_Short_Name=BS}');
Error('\p{JSN=BS}');
Error('\P{JSN=BS}');
Error('\p{Is_Jamo_Short_Name=BS}');
Error('\P{Is_Jamo_Short_Name=BS}');
Error('\p{Is_JSN=BS}');
Error('\P{Is_JSN=BS}');
Error('\p{Jamo_Short_Name=YO}');
Error('\P{Jamo_Short_Name=YO}');
Error('\p{JSN=YO}');
Error('\P{JSN=YO}');
Error('\p{Is_Jamo_Short_Name=YO}');
Error('\P{Is_Jamo_Short_Name=YO}');
Error('\p{Is_JSN=YO}');
Error('\P{Is_JSN=YO}');
Error('\p{Jamo_Short_Name=OE}');
Error('\P{Jamo_Short_Name=OE}');
Error('\p{JSN=OE}');
Error('\P{JSN=OE}');
Error('\p{Is_Jamo_Short_Name=OE}');
Error('\P{Is_Jamo_Short_Name=OE}');
Error('\p{Is_JSN=OE}');
Error('\P{Is_JSN=OE}');
Error('\p{Jamo_Short_Name:   M}');
Error('\P{Jamo_Short_Name:   M}');
Error('\p{JSN=M}');
Error('\P{JSN=M}');
Error('\p{Is_Jamo_Short_Name=M}');
Error('\P{Is_Jamo_Short_Name=M}');
Error('\p{Is_JSN=M}');
Error('\P{Is_JSN=M}');
Error('\p{Jamo_Short_Name:EO}');
Error('\P{Jamo_Short_Name:EO}');
Error('\p{JSN=EO}');
Error('\P{JSN=EO}');
Error('\p{Is_Jamo_Short_Name=EO}');
Error('\P{Is_Jamo_Short_Name=EO}');
Error('\p{Is_JSN=EO}');
Error('\P{Is_JSN=EO}');
Error('\p{Jamo_Short_Name=C}');
Error('\P{Jamo_Short_Name=C}');
Error('\p{JSN:	C}');
Error('\P{JSN:	C}');
Error('\p{Is_Jamo_Short_Name:C}');
Error('\P{Is_Jamo_Short_Name:C}');
Error('\p{Is_JSN=C}');
Error('\P{Is_JSN=C}');
Error('\p{Jamo_Short_Name=WI}');
Error('\P{Jamo_Short_Name=WI}');
Error('\p{JSN=WI}');
Error('\P{JSN=WI}');
Error('\p{Is_Jamo_Short_Name=WI}');
Error('\P{Is_Jamo_Short_Name=WI}');
Error('\p{Is_JSN=WI}');
Error('\P{Is_JSN=WI}');
Error('\p{Jamo_Short_Name=EU}');
Error('\P{Jamo_Short_Name=EU}');
Error('\p{JSN: EU}');
Error('\P{JSN: EU}');
Error('\p{Is_Jamo_Short_Name=EU}');
Error('\P{Is_Jamo_Short_Name=EU}');
Error('\p{Is_JSN=EU}');
Error('\P{Is_JSN=EU}');
Error('\p{Jamo_Short_Name=LG}');
Error('\P{Jamo_Short_Name=LG}');
Error('\p{JSN=LG}');
Error('\P{JSN=LG}');
Error('\p{Is_Jamo_Short_Name: LG}');
Error('\P{Is_Jamo_Short_Name: LG}');
Error('\p{Is_JSN=LG}');
Error('\P{Is_JSN=LG}');
Error('\p{Jamo_Short_Name=L}');
Error('\P{Jamo_Short_Name=L}');
Error('\p{JSN=L}');
Error('\P{JSN=L}');
Error('\p{Is_Jamo_Short_Name=L}');
Error('\P{Is_Jamo_Short_Name=L}');
Error('\p{Is_JSN=L}');
Error('\P{Is_JSN=L}');
Error('\p{Jamo_Short_Name=P}');
Error('\P{Jamo_Short_Name=P}');
Error('\p{JSN=P}');
Error('\P{JSN=P}');
Error('\p{Is_Jamo_Short_Name=P}');
Error('\P{Is_Jamo_Short_Name=P}');
Error('\p{Is_JSN=P}');
Error('\P{Is_JSN=P}');
Error('\p{Jamo_Short_Name=DD}');
Error('\P{Jamo_Short_Name=DD}');
Error('\p{JSN=DD}');
Error('\P{JSN=DD}');
Error('\p{Is_Jamo_Short_Name=DD}');
Error('\P{Is_Jamo_Short_Name=DD}');
Error('\p{Is_JSN=DD}');
Error('\P{Is_JSN=DD}');
Error('\p{Jamo_Short_Name=NG}');
Error('\P{Jamo_Short_Name=NG}');
Error('\p{JSN=NG}');
Error('\P{JSN=NG}');
Error('\p{Is_Jamo_Short_Name=NG}');
Error('\P{Is_Jamo_Short_Name=NG}');
Error('\p{Is_JSN=NG}');
Error('\P{Is_JSN=NG}');
Error('\p{Jamo_Short_Name:   AE}');
Error('\P{Jamo_Short_Name:   AE}');
Error('\p{JSN:   AE}');
Error('\P{JSN:   AE}');
Error('\p{Is_Jamo_Short_Name=AE}');
Error('\P{Is_Jamo_Short_Name=AE}');
Error('\p{Is_JSN=AE}');
Error('\P{Is_JSN=AE}');
Error('\p{Jamo_Short_Name=T}');
Error('\P{Jamo_Short_Name=T}');
Error('\p{JSN=T}');
Error('\P{JSN=T}');
Error('\p{Is_Jamo_Short_Name=T}');
Error('\P{Is_Jamo_Short_Name=T}');
Error('\p{Is_JSN: T}');
Error('\P{Is_JSN: T}');
Error('\p{Jamo_Short_Name=A}');
Error('\P{Jamo_Short_Name=A}');
Error('\p{JSN=A}');
Error('\P{JSN=A}');
Error('\p{Is_Jamo_Short_Name=A}');
Error('\P{Is_Jamo_Short_Name=A}');
Error('\p{Is_JSN:	A}');
Error('\P{Is_JSN:	A}');
Error('\p{Jamo_Short_Name:N}');
Error('\P{Jamo_Short_Name:N}');
Error('\p{JSN=N}');
Error('\P{JSN=N}');
Error('\p{Is_Jamo_Short_Name=N}');
Error('\P{Is_Jamo_Short_Name=N}');
Error('\p{Is_JSN=N}');
Error('\P{Is_JSN=N}');
Error('\p{Jamo_Short_Name=YAE}');
Error('\P{Jamo_Short_Name=YAE}');
Error('\p{JSN=YAE}');
Error('\P{JSN=YAE}');
Error('\p{Is_Jamo_Short_Name=YAE}');
Error('\P{Is_Jamo_Short_Name=YAE}');
Error('\p{Is_JSN=YAE}');
Error('\P{Is_JSN=YAE}');
Error('\p{Jamo_Short_Name=U}');
Error('\P{Jamo_Short_Name=U}');
Error('\p{JSN=U}');
Error('\P{JSN=U}');
Error('\p{Is_Jamo_Short_Name=U}');
Error('\P{Is_Jamo_Short_Name=U}');
Error('\p{Is_JSN=U}');
Error('\P{Is_JSN=U}');
Error('\p{Jamo_Short_Name:   WAE}');
Error('\P{Jamo_Short_Name:   WAE}');
Error('\p{JSN: WAE}');
Error('\P{JSN: WAE}');
Error('\p{Is_Jamo_Short_Name: WAE}');
Error('\P{Is_Jamo_Short_Name: WAE}');
Error('\p{Is_JSN=WAE}');
Error('\P{Is_JSN=WAE}');
Error('\p{Jamo_Short_Name=GS}');
Error('\P{Jamo_Short_Name=GS}');
Error('\p{JSN=GS}');
Error('\P{JSN=GS}');
Error('\p{Is_Jamo_Short_Name=GS}');
Error('\P{Is_Jamo_Short_Name=GS}');
Error('\p{Is_JSN=GS}');
Error('\P{Is_JSN=GS}');
Error('\p{Jamo_Short_Name=H}');
Error('\P{Jamo_Short_Name=H}');
Error('\p{JSN=H}');
Error('\P{JSN=H}');
Error('\p{Is_Jamo_Short_Name=H}');
Error('\P{Is_Jamo_Short_Name=H}');
Error('\p{Is_JSN=H}');
Error('\P{Is_JSN=H}');
Error('\p{Jamo_Short_Name=LT}');
Error('\P{Jamo_Short_Name=LT}');
Error('\p{JSN=LT}');
Error('\P{JSN=LT}');
Error('\p{Is_Jamo_Short_Name=LT}');
Error('\P{Is_Jamo_Short_Name=LT}');
Error('\p{Is_JSN=LT}');
Error('\P{Is_JSN=LT}');
Error('\p{Jamo_Short_Name=YE}');
Error('\P{Jamo_Short_Name=YE}');
Error('\p{JSN:   YE}');
Error('\P{JSN:   YE}');
Error('\p{Is_Jamo_Short_Name=YE}');
Error('\P{Is_Jamo_Short_Name=YE}');
Error('\p{Is_JSN=YE}');
Error('\P{Is_JSN=YE}');
Error('\p{Jamo_Short_Name=I}');
Error('\P{Jamo_Short_Name=I}');
Error('\p{JSN=I}');
Error('\P{JSN=I}');
Error('\p{Is_Jamo_Short_Name=I}');
Error('\P{Is_Jamo_Short_Name=I}');
Error('\p{Is_JSN=I}');
Error('\P{Is_JSN=I}');
Error('\p{Jamo_Short_Name=R}');
Error('\P{Jamo_Short_Name=R}');
Error('\p{JSN=R}');
Error('\P{JSN=R}');
Error('\p{Is_Jamo_Short_Name=R}');
Error('\P{Is_Jamo_Short_Name=R}');
Error('\p{Is_JSN=R}');
Error('\P{Is_JSN=R}');
Error('\p{Jamo_Short_Name=O}');
Error('\P{Jamo_Short_Name=O}');
Error('\p{JSN=O}');
Error('\P{JSN=O}');
Error('\p{Is_Jamo_Short_Name=O}');
Error('\P{Is_Jamo_Short_Name=O}');
Error('\p{Is_JSN=O}');
Error('\P{Is_JSN=O}');
Error('\p{NFKD_Quick_Check=/a/ Yes}');
Error('\P{NFKD_Quick_Check=/a/ Yes}');
Expect(1, 195102, '\p{NFKD_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^NFKD_Quick_Check=yes}', "");
Expect(0, 195102, '\P{NFKD_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^NFKD_Quick_Check=yes}', "");
Expect(0, 195101, '\p{NFKD_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^NFKD_Quick_Check=yes}', "");
Expect(1, 195101, '\P{NFKD_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^NFKD_Quick_Check=yes}', "");
Expect(1, 195102, '\p{NFKD_Quick_Check=	_YES}', "");
Expect(0, 195102, '\p{^NFKD_Quick_Check=	_YES}', "");
Expect(0, 195102, '\P{NFKD_Quick_Check=	_YES}', "");
Expect(1, 195102, '\P{^NFKD_Quick_Check=	_YES}', "");
Expect(0, 195101, '\p{NFKD_Quick_Check=	_YES}', "");
Expect(1, 195101, '\p{^NFKD_Quick_Check=	_YES}', "");
Expect(1, 195101, '\P{NFKD_Quick_Check=	_YES}', "");
Expect(0, 195101, '\P{^NFKD_Quick_Check=	_YES}', "");
Error('\p{NFKD_QC=/a/		Y}');
Error('\P{NFKD_QC=/a/		Y}');
Expect(1, 195102, '\p{NFKD_QC=y}', "");
Expect(0, 195102, '\p{^NFKD_QC=y}', "");
Expect(0, 195102, '\P{NFKD_QC=y}', "");
Expect(1, 195102, '\P{^NFKD_QC=y}', "");
Expect(0, 195101, '\p{NFKD_QC=y}', "");
Expect(1, 195101, '\p{^NFKD_QC=y}', "");
Expect(1, 195101, '\P{NFKD_QC=y}', "");
Expect(0, 195101, '\P{^NFKD_QC=y}', "");
Expect(1, 195102, '\p{NFKD_QC=		Y}', "");
Expect(0, 195102, '\p{^NFKD_QC=		Y}', "");
Expect(0, 195102, '\P{NFKD_QC=		Y}', "");
Expect(1, 195102, '\P{^NFKD_QC=		Y}', "");
Expect(0, 195101, '\p{NFKD_QC=		Y}', "");
Expect(1, 195101, '\p{^NFKD_QC=		Y}', "");
Expect(1, 195101, '\P{NFKD_QC=		Y}', "");
Expect(0, 195101, '\P{^NFKD_QC=		Y}', "");
Error('\p{Is_NFKD_Quick_Check=:=- Yes}');
Error('\P{Is_NFKD_Quick_Check=:=- Yes}');
Expect(1, 195102, '\p{Is_NFKD_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^Is_NFKD_Quick_Check=yes}', "");
Expect(0, 195102, '\P{Is_NFKD_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^Is_NFKD_Quick_Check=yes}', "");
Expect(0, 195101, '\p{Is_NFKD_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^Is_NFKD_Quick_Check=yes}', "");
Expect(1, 195101, '\P{Is_NFKD_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^Is_NFKD_Quick_Check=yes}', "");
Expect(1, 195102, '\p{Is_NFKD_Quick_Check=-yes}', "");
Expect(0, 195102, '\p{^Is_NFKD_Quick_Check=-yes}', "");
Expect(0, 195102, '\P{Is_NFKD_Quick_Check=-yes}', "");
Expect(1, 195102, '\P{^Is_NFKD_Quick_Check=-yes}', "");
Expect(0, 195101, '\p{Is_NFKD_Quick_Check=-yes}', "");
Expect(1, 195101, '\p{^Is_NFKD_Quick_Check=-yes}', "");
Expect(1, 195101, '\P{Is_NFKD_Quick_Check=-yes}', "");
Expect(0, 195101, '\P{^Is_NFKD_Quick_Check=-yes}', "");
Error('\p{Is_NFKD_QC= y:=}');
Error('\P{Is_NFKD_QC= y:=}');
Expect(1, 195102, '\p{Is_NFKD_QC=y}', "");
Expect(0, 195102, '\p{^Is_NFKD_QC=y}', "");
Expect(0, 195102, '\P{Is_NFKD_QC=y}', "");
Expect(1, 195102, '\P{^Is_NFKD_QC=y}', "");
Expect(0, 195101, '\p{Is_NFKD_QC=y}', "");
Expect(1, 195101, '\p{^Is_NFKD_QC=y}', "");
Expect(1, 195101, '\P{Is_NFKD_QC=y}', "");
Expect(0, 195101, '\P{^Is_NFKD_QC=y}', "");
Expect(1, 195102, '\p{Is_NFKD_QC=- y}', "");
Expect(0, 195102, '\p{^Is_NFKD_QC=- y}', "");
Expect(0, 195102, '\P{Is_NFKD_QC=- y}', "");
Expect(1, 195102, '\P{^Is_NFKD_QC=- y}', "");
Expect(0, 195101, '\p{Is_NFKD_QC=- y}', "");
Expect(1, 195101, '\p{^Is_NFKD_QC=- y}', "");
Expect(1, 195101, '\P{Is_NFKD_QC=- y}', "");
Expect(0, 195101, '\P{^Is_NFKD_QC=- y}', "");
Error('\p{NFKD_Quick_Check=--No/a/}');
Error('\P{NFKD_Quick_Check=--No/a/}');
Expect(1, 195101, '\p{NFKD_Quick_Check=no}', "");
Expect(0, 195101, '\p{^NFKD_Quick_Check=no}', "");
Expect(0, 195101, '\P{NFKD_Quick_Check=no}', "");
Expect(1, 195101, '\P{^NFKD_Quick_Check=no}', "");
Expect(0, 195102, '\p{NFKD_Quick_Check=no}', "");
Expect(1, 195102, '\p{^NFKD_Quick_Check=no}', "");
Expect(1, 195102, '\P{NFKD_Quick_Check=no}', "");
Expect(0, 195102, '\P{^NFKD_Quick_Check=no}', "");
Expect(1, 195101, '\p{NFKD_Quick_Check=No}', "");
Expect(0, 195101, '\p{^NFKD_Quick_Check=No}', "");
Expect(0, 195101, '\P{NFKD_Quick_Check=No}', "");
Expect(1, 195101, '\P{^NFKD_Quick_Check=No}', "");
Expect(0, 195102, '\p{NFKD_Quick_Check=No}', "");
Expect(1, 195102, '\p{^NFKD_Quick_Check=No}', "");
Expect(1, 195102, '\P{NFKD_Quick_Check=No}', "");
Expect(0, 195102, '\P{^NFKD_Quick_Check=No}', "");
Error('\p{NFKD_QC=_/a/N}');
Error('\P{NFKD_QC=_/a/N}');
Expect(1, 195101, '\p{NFKD_QC=n}', "");
Expect(0, 195101, '\p{^NFKD_QC=n}', "");
Expect(0, 195101, '\P{NFKD_QC=n}', "");
Expect(1, 195101, '\P{^NFKD_QC=n}', "");
Expect(0, 195102, '\p{NFKD_QC=n}', "");
Expect(1, 195102, '\p{^NFKD_QC=n}', "");
Expect(1, 195102, '\P{NFKD_QC=n}', "");
Expect(0, 195102, '\P{^NFKD_QC=n}', "");
Expect(1, 195101, '\p{NFKD_QC= _n}', "");
Expect(0, 195101, '\p{^NFKD_QC= _n}', "");
Expect(0, 195101, '\P{NFKD_QC= _n}', "");
Expect(1, 195101, '\P{^NFKD_QC= _n}', "");
Expect(0, 195102, '\p{NFKD_QC= _n}', "");
Expect(1, 195102, '\p{^NFKD_QC= _n}', "");
Expect(1, 195102, '\P{NFKD_QC= _n}', "");
Expect(0, 195102, '\P{^NFKD_QC= _n}', "");
Error('\p{Is_NFKD_Quick_Check=/a/-No}');
Error('\P{Is_NFKD_Quick_Check=/a/-No}');
Expect(1, 195101, '\p{Is_NFKD_Quick_Check=no}', "");
Expect(0, 195101, '\p{^Is_NFKD_Quick_Check=no}', "");
Expect(0, 195101, '\P{Is_NFKD_Quick_Check=no}', "");
Expect(1, 195101, '\P{^Is_NFKD_Quick_Check=no}', "");
Expect(0, 195102, '\p{Is_NFKD_Quick_Check=no}', "");
Expect(1, 195102, '\p{^Is_NFKD_Quick_Check=no}', "");
Expect(1, 195102, '\P{Is_NFKD_Quick_Check=no}', "");
Expect(0, 195102, '\P{^Is_NFKD_Quick_Check=no}', "");
Expect(1, 195101, '\p{Is_NFKD_Quick_Check:No}', "");
Expect(0, 195101, '\p{^Is_NFKD_Quick_Check:No}', "");
Expect(0, 195101, '\P{Is_NFKD_Quick_Check:No}', "");
Expect(1, 195101, '\P{^Is_NFKD_Quick_Check:No}', "");
Expect(0, 195102, '\p{Is_NFKD_Quick_Check:No}', "");
Expect(1, 195102, '\p{^Is_NFKD_Quick_Check:No}', "");
Expect(1, 195102, '\P{Is_NFKD_Quick_Check:No}', "");
Expect(0, 195102, '\P{^Is_NFKD_Quick_Check:No}', "");
Error('\p{Is_NFKD_QC=_-n/a/}');
Error('\P{Is_NFKD_QC=_-n/a/}');
Expect(1, 195101, '\p{Is_NFKD_QC=n}', "");
Expect(0, 195101, '\p{^Is_NFKD_QC=n}', "");
Expect(0, 195101, '\P{Is_NFKD_QC=n}', "");
Expect(1, 195101, '\P{^Is_NFKD_QC=n}', "");
Expect(0, 195102, '\p{Is_NFKD_QC=n}', "");
Expect(1, 195102, '\p{^Is_NFKD_QC=n}', "");
Expect(1, 195102, '\P{Is_NFKD_QC=n}', "");
Expect(0, 195102, '\P{^Is_NFKD_QC=n}', "");
Expect(1, 195101, '\p{Is_NFKD_QC:	  n}', "");
Expect(0, 195101, '\p{^Is_NFKD_QC:	  n}', "");
Expect(0, 195101, '\P{Is_NFKD_QC:	  n}', "");
Expect(1, 195101, '\P{^Is_NFKD_QC:	  n}', "");
Expect(0, 195102, '\p{Is_NFKD_QC:	  n}', "");
Expect(1, 195102, '\p{^Is_NFKD_QC:	  n}', "");
Expect(1, 195102, '\P{Is_NFKD_QC:	  n}', "");
Expect(0, 195102, '\P{^Is_NFKD_QC:	  n}', "");
Error('\p{Join_Control=:=	YES}');
Error('\P{Join_Control=:=	YES}');
Expect(1, 8205, '\p{Join_Control=yes}', "");
Expect(0, 8205, '\p{^Join_Control=yes}', "");
Expect(0, 8205, '\P{Join_Control=yes}', "");
Expect(1, 8205, '\P{^Join_Control=yes}', "");
Expect(0, 8206, '\p{Join_Control=yes}', "");
Expect(1, 8206, '\p{^Join_Control=yes}', "");
Expect(1, 8206, '\P{Join_Control=yes}', "");
Expect(0, 8206, '\P{^Join_Control=yes}', "");
Expect(1, 8205, '\p{Join_Control= _yes}', "");
Expect(0, 8205, '\p{^Join_Control= _yes}', "");
Expect(0, 8205, '\P{Join_Control= _yes}', "");
Expect(1, 8205, '\P{^Join_Control= _yes}', "");
Expect(0, 8206, '\p{Join_Control= _yes}', "");
Expect(1, 8206, '\p{^Join_Control= _yes}', "");
Expect(1, 8206, '\P{Join_Control= _yes}', "");
Expect(0, 8206, '\P{^Join_Control= _yes}', "");
Error('\p{Join_C=/a/	Y}');
Error('\P{Join_C=/a/	Y}');
Expect(1, 8205, '\p{Join_C=y}', "");
Expect(0, 8205, '\p{^Join_C=y}', "");
Expect(0, 8205, '\P{Join_C=y}', "");
Expect(1, 8205, '\P{^Join_C=y}', "");
Expect(0, 8206, '\p{Join_C=y}', "");
Expect(1, 8206, '\p{^Join_C=y}', "");
Expect(1, 8206, '\P{Join_C=y}', "");
Expect(0, 8206, '\P{^Join_C=y}', "");
Expect(1, 8205, '\p{Join_C=_Y}', "");
Expect(0, 8205, '\p{^Join_C=_Y}', "");
Expect(0, 8205, '\P{Join_C=_Y}', "");
Expect(1, 8205, '\P{^Join_C=_Y}', "");
Expect(0, 8206, '\p{Join_C=_Y}', "");
Expect(1, 8206, '\p{^Join_C=_Y}', "");
Expect(1, 8206, '\P{Join_C=_Y}', "");
Expect(0, 8206, '\P{^Join_C=_Y}', "");
Error('\p{Is_Join_Control=-T/a/}');
Error('\P{Is_Join_Control=-T/a/}');
Expect(1, 8205, '\p{Is_Join_Control=t}', "");
Expect(0, 8205, '\p{^Is_Join_Control=t}', "");
Expect(0, 8205, '\P{Is_Join_Control=t}', "");
Expect(1, 8205, '\P{^Is_Join_Control=t}', "");
Expect(0, 8206, '\p{Is_Join_Control=t}', "");
Expect(1, 8206, '\p{^Is_Join_Control=t}', "");
Expect(1, 8206, '\P{Is_Join_Control=t}', "");
Expect(0, 8206, '\P{^Is_Join_Control=t}', "");
Expect(1, 8205, '\p{Is_Join_Control=_	T}', "");
Expect(0, 8205, '\p{^Is_Join_Control=_	T}', "");
Expect(0, 8205, '\P{Is_Join_Control=_	T}', "");
Expect(1, 8205, '\P{^Is_Join_Control=_	T}', "");
Expect(0, 8206, '\p{Is_Join_Control=_	T}', "");
Expect(1, 8206, '\p{^Is_Join_Control=_	T}', "");
Expect(1, 8206, '\P{Is_Join_Control=_	T}', "");
Expect(0, 8206, '\P{^Is_Join_Control=_	T}', "");
Error('\p{Is_Join_C=/a/ _True}');
Error('\P{Is_Join_C=/a/ _True}');
Expect(1, 8205, '\p{Is_Join_C=true}', "");
Expect(0, 8205, '\p{^Is_Join_C=true}', "");
Expect(0, 8205, '\P{Is_Join_C=true}', "");
Expect(1, 8205, '\P{^Is_Join_C=true}', "");
Expect(0, 8206, '\p{Is_Join_C=true}', "");
Expect(1, 8206, '\p{^Is_Join_C=true}', "");
Expect(1, 8206, '\P{Is_Join_C=true}', "");
Expect(0, 8206, '\P{^Is_Join_C=true}', "");
Expect(1, 8205, '\p{Is_Join_C:	 True}', "");
Expect(0, 8205, '\p{^Is_Join_C:	 True}', "");
Expect(0, 8205, '\P{Is_Join_C:	 True}', "");
Expect(1, 8205, '\P{^Is_Join_C:	 True}', "");
Expect(0, 8206, '\p{Is_Join_C:	 True}', "");
Expect(1, 8206, '\p{^Is_Join_C:	 True}', "");
Expect(1, 8206, '\P{Is_Join_C:	 True}', "");
Expect(0, 8206, '\P{^Is_Join_C:	 True}', "");
Error('\p{Join_Control=:=-no}');
Error('\P{Join_Control=:=-no}');
Expect(1, 8206, '\p{Join_Control=no}', "");
Expect(0, 8206, '\p{^Join_Control=no}', "");
Expect(0, 8206, '\P{Join_Control=no}', "");
Expect(1, 8206, '\P{^Join_Control=no}', "");
Expect(0, 8205, '\p{Join_Control=no}', "");
Expect(1, 8205, '\p{^Join_Control=no}', "");
Expect(1, 8205, '\P{Join_Control=no}', "");
Expect(0, 8205, '\P{^Join_Control=no}', "");
Expect(1, 8206, '\p{Join_Control=_ No}', "");
Expect(0, 8206, '\p{^Join_Control=_ No}', "");
Expect(0, 8206, '\P{Join_Control=_ No}', "");
Expect(1, 8206, '\P{^Join_Control=_ No}', "");
Expect(0, 8205, '\p{Join_Control=_ No}', "");
Expect(1, 8205, '\p{^Join_Control=_ No}', "");
Expect(1, 8205, '\P{Join_Control=_ No}', "");
Expect(0, 8205, '\P{^Join_Control=_ No}', "");
Error('\p{Join_C=_-N:=}');
Error('\P{Join_C=_-N:=}');
Expect(1, 8206, '\p{Join_C=n}', "");
Expect(0, 8206, '\p{^Join_C=n}', "");
Expect(0, 8206, '\P{Join_C=n}', "");
Expect(1, 8206, '\P{^Join_C=n}', "");
Expect(0, 8205, '\p{Join_C=n}', "");
Expect(1, 8205, '\p{^Join_C=n}', "");
Expect(1, 8205, '\P{Join_C=n}', "");
Expect(0, 8205, '\P{^Join_C=n}', "");
Expect(1, 8206, '\p{Join_C=- n}', "");
Expect(0, 8206, '\p{^Join_C=- n}', "");
Expect(0, 8206, '\P{Join_C=- n}', "");
Expect(1, 8206, '\P{^Join_C=- n}', "");
Expect(0, 8205, '\p{Join_C=- n}', "");
Expect(1, 8205, '\p{^Join_C=- n}', "");
Expect(1, 8205, '\P{Join_C=- n}', "");
Expect(0, 8205, '\P{^Join_C=- n}', "");
Error('\p{Is_Join_Control=/a/		f}');
Error('\P{Is_Join_Control=/a/		f}');
Expect(1, 8206, '\p{Is_Join_Control=f}', "");
Expect(0, 8206, '\p{^Is_Join_Control=f}', "");
Expect(0, 8206, '\P{Is_Join_Control=f}', "");
Expect(1, 8206, '\P{^Is_Join_Control=f}', "");
Expect(0, 8205, '\p{Is_Join_Control=f}', "");
Expect(1, 8205, '\p{^Is_Join_Control=f}', "");
Expect(1, 8205, '\P{Is_Join_Control=f}', "");
Expect(0, 8205, '\P{^Is_Join_Control=f}', "");
Expect(1, 8206, '\p{Is_Join_Control=-_F}', "");
Expect(0, 8206, '\p{^Is_Join_Control=-_F}', "");
Expect(0, 8206, '\P{Is_Join_Control=-_F}', "");
Expect(1, 8206, '\P{^Is_Join_Control=-_F}', "");
Expect(0, 8205, '\p{Is_Join_Control=-_F}', "");
Expect(1, 8205, '\p{^Is_Join_Control=-_F}', "");
Expect(1, 8205, '\P{Is_Join_Control=-_F}', "");
Expect(0, 8205, '\P{^Is_Join_Control=-_F}', "");
Error('\p{Is_Join_C=/a/	_FALSE}');
Error('\P{Is_Join_C=/a/	_FALSE}');
Expect(1, 8206, '\p{Is_Join_C=false}', "");
Expect(0, 8206, '\p{^Is_Join_C=false}', "");
Expect(0, 8206, '\P{Is_Join_C=false}', "");
Expect(1, 8206, '\P{^Is_Join_C=false}', "");
Expect(0, 8205, '\p{Is_Join_C=false}', "");
Expect(1, 8205, '\p{^Is_Join_C=false}', "");
Expect(1, 8205, '\P{Is_Join_C=false}', "");
Expect(0, 8205, '\P{^Is_Join_C=false}', "");
Expect(1, 8206, '\p{Is_Join_C=--false}', "");
Expect(0, 8206, '\p{^Is_Join_C=--false}', "");
Expect(0, 8206, '\P{Is_Join_C=--false}', "");
Expect(1, 8206, '\P{^Is_Join_C=--false}', "");
Expect(0, 8205, '\p{Is_Join_C=--false}', "");
Expect(1, 8205, '\p{^Is_Join_C=--false}', "");
Expect(1, 8205, '\P{Is_Join_C=--false}', "");
Expect(0, 8205, '\P{^Is_Join_C=--false}', "");
Error('\p{Numeric_Value=	00003_3/a/}');
Error('\P{Numeric_Value=	00003_3/a/}');
Expect(1, 12893, '\p{Numeric_Value=0033}', "");
Expect(0, 12893, '\p{^Numeric_Value=0033}', "");
Expect(0, 12893, '\P{Numeric_Value=0033}', "");
Expect(1, 12893, '\P{^Numeric_Value=0033}', "");
Expect(0, 12894, '\p{Numeric_Value=0033}', "");
Expect(1, 12894, '\p{^Numeric_Value=0033}', "");
Expect(1, 12894, '\P{Numeric_Value=0033}', "");
Expect(0, 12894, '\P{^Numeric_Value=0033}', "");
Error('\p{Nv:	__00000033:=}');
Error('\P{Nv:	__00000033:=}');
Expect(1, 12893, '\p{Nv=+0000033}', "");
Expect(0, 12893, '\p{^Nv=+0000033}', "");
Expect(0, 12893, '\P{Nv=+0000033}', "");
Expect(1, 12893, '\P{^Nv=+0000033}', "");
Expect(0, 12894, '\p{Nv=+0000033}', "");
Expect(1, 12894, '\p{^Nv=+0000033}', "");
Expect(1, 12894, '\P{Nv=+0000033}', "");
Expect(0, 12894, '\P{^Nv=+0000033}', "");
Error('\p{Is_Numeric_Value:	/a/00_00_33}');
Error('\P{Is_Numeric_Value:	/a/00_00_33}');
Expect(1, 12893, '\p{Is_Numeric_Value=+000000033}', "");
Expect(0, 12893, '\p{^Is_Numeric_Value=+000000033}', "");
Expect(0, 12893, '\P{Is_Numeric_Value=+000000033}', "");
Expect(1, 12893, '\P{^Is_Numeric_Value=+000000033}', "");
Expect(0, 12894, '\p{Is_Numeric_Value=+000000033}', "");
Expect(1, 12894, '\p{^Is_Numeric_Value=+000000033}', "");
Expect(1, 12894, '\P{Is_Numeric_Value=+000000033}', "");
Expect(0, 12894, '\P{^Is_Numeric_Value=+000000033}', "");
Error('\p{Is_Nv=-	00000033/a/}');
Error('\P{Is_Nv=-	00000033/a/}');
Expect(1, 12893, '\p{Is_Nv=+0000000003_3}', "");
Expect(0, 12893, '\p{^Is_Nv=+0000000003_3}', "");
Expect(0, 12893, '\P{Is_Nv=+0000000003_3}', "");
Expect(1, 12893, '\P{^Is_Nv=+0000000003_3}', "");
Expect(0, 12894, '\p{Is_Nv=+0000000003_3}', "");
Expect(1, 12894, '\p{^Is_Nv=+0000000003_3}', "");
Expect(1, 12894, '\P{Is_Nv=+0000000003_3}', "");
Expect(0, 12894, '\P{^Is_Nv=+0000000003_3}', "");
Error('\p{Numeric_Value=:=_	0000032}');
Error('\P{Numeric_Value=:=_	0000032}');
Expect(1, 12892, '\p{Numeric_Value=0_0_0_0_0_0_0_0_032}', "");
Expect(0, 12892, '\p{^Numeric_Value=0_0_0_0_0_0_0_0_032}', "");
Expect(0, 12892, '\P{Numeric_Value=0_0_0_0_0_0_0_0_032}', "");
Expect(1, 12892, '\P{^Numeric_Value=0_0_0_0_0_0_0_0_032}', "");
Expect(0, 12893, '\p{Numeric_Value=0_0_0_0_0_0_0_0_032}', "");
Expect(1, 12893, '\p{^Numeric_Value=0_0_0_0_0_0_0_0_032}', "");
Expect(1, 12893, '\P{Numeric_Value=0_0_0_0_0_0_0_0_032}', "");
Expect(0, 12893, '\P{^Numeric_Value=0_0_0_0_0_0_0_0_032}', "");
Error('\p{Nv=	_032:=}');
Error('\P{Nv=	_032:=}');
Expect(1, 12892, '\p{Nv=+032}', "");
Expect(0, 12892, '\p{^Nv=+032}', "");
Expect(0, 12892, '\P{Nv=+032}', "");
Expect(1, 12892, '\P{^Nv=+032}', "");
Expect(0, 12893, '\p{Nv=+032}', "");
Expect(1, 12893, '\p{^Nv=+032}', "");
Expect(1, 12893, '\P{Nv=+032}', "");
Expect(0, 12893, '\P{^Nv=+032}', "");
Error('\p{Is_Numeric_Value=_:=32}');
Error('\P{Is_Numeric_Value=_:=32}');
Expect(1, 12892, '\p{Is_Numeric_Value=+0_0_0_0_0032}', "");
Expect(0, 12892, '\p{^Is_Numeric_Value=+0_0_0_0_0032}', "");
Expect(0, 12892, '\P{Is_Numeric_Value=+0_0_0_0_0032}', "");
Expect(1, 12892, '\P{^Is_Numeric_Value=+0_0_0_0_0032}', "");
Expect(0, 12893, '\p{Is_Numeric_Value=+0_0_0_0_0032}', "");
Expect(1, 12893, '\p{^Is_Numeric_Value=+0_0_0_0_0032}', "");
Expect(1, 12893, '\P{Is_Numeric_Value=+0_0_0_0_0032}', "");
Expect(0, 12893, '\P{^Is_Numeric_Value=+0_0_0_0_0032}', "");
Error('\p{Is_Nv=:=-	0000000032}');
Error('\P{Is_Nv=:=-	0000000032}');
Expect(1, 12892, '\p{Is_Nv=0_0_32}', "");
Expect(0, 12892, '\p{^Is_Nv=0_0_32}', "");
Expect(0, 12892, '\P{Is_Nv=0_0_32}', "");
Expect(1, 12892, '\P{^Is_Nv=0_0_32}', "");
Expect(0, 12893, '\p{Is_Nv=0_0_32}', "");
Expect(1, 12893, '\p{^Is_Nv=0_0_32}', "");
Expect(1, 12893, '\P{Is_Nv=0_0_32}', "");
Expect(0, 12893, '\P{^Is_Nv=0_0_32}', "");
Error('\p{Numeric_Value=000000003/000000016:=}');
Error('\P{Numeric_Value=000000003/000000016:=}');
Expect(1, 43061, '\p{Numeric_Value=+03/00000016}', "");
Expect(0, 43061, '\p{^Numeric_Value=+03/00000016}', "");
Expect(0, 43061, '\P{Numeric_Value=+03/00000016}', "");
Expect(1, 43061, '\P{^Numeric_Value=+03/00000016}', "");
Expect(0, 43062, '\p{Numeric_Value=+03/00000016}', "");
Expect(1, 43062, '\p{^Numeric_Value=+03/00000016}', "");
Expect(1, 43062, '\P{Numeric_Value=+03/00000016}', "");
Expect(0, 43062, '\P{^Numeric_Value=+03/00000016}', "");
Error('\p{Numeric_Value=0.19}');
Error('\P{Numeric_Value=0.19}');
Expect(1, 43061, '\p{Numeric_Value=0.188}', "");
Expect(0, 43061, '\p{^Numeric_Value=0.188}', "");
Expect(0, 43061, '\P{Numeric_Value=0.188}', "");
Expect(1, 43061, '\P{^Numeric_Value=0.188}', "");
Expect(0, 43062, '\p{Numeric_Value=0.188}', "");
Expect(1, 43062, '\p{^Numeric_Value=0.188}', "");
Expect(1, 43062, '\P{Numeric_Value=0.188}', "");
Expect(0, 43062, '\P{^Numeric_Value=0.188}', "");
Error('\p{Nv=-000003/00000000016:=}');
Error('\P{Nv=-000003/00000000016:=}');
Expect(1, 43061, '\p{Nv=+0003/000016}', "");
Expect(0, 43061, '\p{^Nv=+0003/000016}', "");
Expect(0, 43061, '\P{Nv=+0003/000016}', "");
Expect(1, 43061, '\P{^Nv=+0003/000016}', "");
Expect(0, 43062, '\p{Nv=+0003/000016}', "");
Expect(1, 43062, '\p{^Nv=+0003/000016}', "");
Expect(1, 43062, '\P{Nv=+0003/000016}', "");
Expect(0, 43062, '\P{^Nv=+0003/000016}', "");
Error('\p{Nv=0.19}');
Error('\P{Nv=0.19}');
Expect(1, 43061, '\p{Nv=0.188}', "");
Expect(0, 43061, '\p{^Nv=0.188}', "");
Expect(0, 43061, '\P{Nv=0.188}', "");
Expect(1, 43061, '\P{^Nv=0.188}', "");
Expect(0, 43062, '\p{Nv=0.188}', "");
Expect(1, 43062, '\p{^Nv=0.188}', "");
Expect(1, 43062, '\P{Nv=0.188}', "");
Expect(0, 43062, '\P{^Nv=0.188}', "");
Error('\p{Is_Numeric_Value: -:=+3/0000000016}');
Error('\P{Is_Numeric_Value: -:=+3/0000000016}');
Expect(1, 43061, '\p{Is_Numeric_Value=0003/0000016}', "");
Expect(0, 43061, '\p{^Is_Numeric_Value=0003/0000016}', "");
Expect(0, 43061, '\P{Is_Numeric_Value=0003/0000016}', "");
Expect(1, 43061, '\P{^Is_Numeric_Value=0003/0000016}', "");
Expect(0, 43062, '\p{Is_Numeric_Value=0003/0000016}', "");
Expect(1, 43062, '\p{^Is_Numeric_Value=0003/0000016}', "");
Expect(1, 43062, '\P{Is_Numeric_Value=0003/0000016}', "");
Expect(0, 43062, '\P{^Is_Numeric_Value=0003/0000016}', "");
Error('\p{Is_Numeric_Value=0.19}');
Error('\P{Is_Numeric_Value=0.19}');
Expect(1, 43061, '\p{Is_Numeric_Value=0.188}', "");
Expect(0, 43061, '\p{^Is_Numeric_Value=0.188}', "");
Expect(0, 43061, '\P{Is_Numeric_Value=0.188}', "");
Expect(1, 43061, '\P{^Is_Numeric_Value=0.188}', "");
Expect(0, 43062, '\p{Is_Numeric_Value=0.188}', "");
Expect(1, 43062, '\p{^Is_Numeric_Value=0.188}', "");
Expect(1, 43062, '\P{Is_Numeric_Value=0.188}', "");
Expect(0, 43062, '\P{^Is_Numeric_Value=0.188}', "");
Error('\p{Is_Nv= 	0003/0000000016/a/}');
Error('\P{Is_Nv= 	0003/0000000016/a/}');
Expect(1, 43061, '\p{Is_Nv=0000000003/000016}', "");
Expect(0, 43061, '\p{^Is_Nv=0000000003/000016}', "");
Expect(0, 43061, '\P{Is_Nv=0000000003/000016}', "");
Expect(1, 43061, '\P{^Is_Nv=0000000003/000016}', "");
Expect(0, 43062, '\p{Is_Nv=0000000003/000016}', "");
Expect(1, 43062, '\p{^Is_Nv=0000000003/000016}', "");
Expect(1, 43062, '\P{Is_Nv=0000000003/000016}', "");
Expect(0, 43062, '\P{^Is_Nv=0000000003/000016}', "");
Error('\p{Is_Nv:0.19}');
Error('\P{Is_Nv:0.19}');
Expect(1, 43061, '\p{Is_Nv=0.188}', "");
Expect(0, 43061, '\p{^Is_Nv=0.188}', "");
Expect(0, 43061, '\P{Is_Nv=0.188}', "");
Expect(1, 43061, '\P{^Is_Nv=0.188}', "");
Expect(0, 43062, '\p{Is_Nv=0.188}', "");
Expect(1, 43062, '\p{^Is_Nv=0.188}', "");
Expect(1, 43062, '\P{Is_Nv=0.188}', "");
Expect(0, 43062, '\P{^Is_Nv=0.188}', "");
Error('\p{Numeric_Value=_:=21}');
Error('\P{Numeric_Value=_:=21}');
Expect(1, 12881, '\p{Numeric_Value=0_0_0_0_0_0_0_0021}', "");
Expect(0, 12881, '\p{^Numeric_Value=0_0_0_0_0_0_0_0021}', "");
Expect(0, 12881, '\P{Numeric_Value=0_0_0_0_0_0_0_0021}', "");
Expect(1, 12881, '\P{^Numeric_Value=0_0_0_0_0_0_0_0021}', "");
Expect(0, 12882, '\p{Numeric_Value=0_0_0_0_0_0_0_0021}', "");
Expect(1, 12882, '\p{^Numeric_Value=0_0_0_0_0_0_0_0021}', "");
Expect(1, 12882, '\P{Numeric_Value=0_0_0_0_0_0_0_0021}', "");
Expect(0, 12882, '\P{^Numeric_Value=0_0_0_0_0_0_0_0021}', "");
Error('\p{Nv=:=	_00_00_00_021}');
Error('\P{Nv=:=	_00_00_00_021}');
Expect(1, 12881, '\p{Nv=21}', "");
Expect(0, 12881, '\p{^Nv=21}', "");
Expect(0, 12881, '\P{Nv=21}', "");
Expect(1, 12881, '\P{^Nv=21}', "");
Expect(0, 12882, '\p{Nv=21}', "");
Expect(1, 12882, '\p{^Nv=21}', "");
Expect(1, 12882, '\P{Nv=21}', "");
Expect(0, 12882, '\P{^Nv=21}', "");
Error('\p{Is_Numeric_Value= -02_1/a/}');
Error('\P{Is_Numeric_Value= -02_1/a/}');
Expect(1, 12881, '\p{Is_Numeric_Value=00_00_00_00_21}', "");
Expect(0, 12881, '\p{^Is_Numeric_Value=00_00_00_00_21}', "");
Expect(0, 12881, '\P{Is_Numeric_Value=00_00_00_00_21}', "");
Expect(1, 12881, '\P{^Is_Numeric_Value=00_00_00_00_21}', "");
Expect(0, 12882, '\p{Is_Numeric_Value=00_00_00_00_21}', "");
Expect(1, 12882, '\p{^Is_Numeric_Value=00_00_00_00_21}', "");
Expect(1, 12882, '\P{Is_Numeric_Value=00_00_00_00_21}', "");
Expect(0, 12882, '\P{^Is_Numeric_Value=00_00_00_00_21}', "");
Error('\p{Is_Nv=/a/0_0_0_0_0_0_0_21}');
Error('\P{Is_Nv=/a/0_0_0_0_0_0_0_21}');
Expect(1, 12881, '\p{Is_Nv=00_00_02_1}', "");
Expect(0, 12881, '\p{^Is_Nv=00_00_02_1}', "");
Expect(0, 12881, '\P{Is_Nv=00_00_02_1}', "");
Expect(1, 12881, '\P{^Is_Nv=00_00_02_1}', "");
Expect(0, 12882, '\p{Is_Nv=00_00_02_1}', "");
Expect(1, 12882, '\p{^Is_Nv=00_00_02_1}', "");
Expect(1, 12882, '\P{Is_Nv=00_00_02_1}', "");
Expect(0, 12882, '\P{^Is_Nv=00_00_02_1}', "");
Error('\p{Numeric_Value=:=_ 0000_0000_90}');
Error('\P{Numeric_Value=:=_ 0000_0000_90}');
Expect(1, 119665, '\p{Numeric_Value=0090}', "");
Expect(0, 119665, '\p{^Numeric_Value=0090}', "");
Expect(0, 119665, '\P{Numeric_Value=0090}', "");
Expect(1, 119665, '\P{^Numeric_Value=0090}', "");
Expect(0, 119666, '\p{Numeric_Value=0090}', "");
Expect(1, 119666, '\p{^Numeric_Value=0090}', "");
Expect(1, 119666, '\P{Numeric_Value=0090}', "");
Expect(0, 119666, '\P{^Numeric_Value=0090}', "");
Error('\p{Nv=:= 	0090}');
Error('\P{Nv=:= 	0090}');
Expect(1, 119665, '\p{Nv=+000000090}', "");
Expect(0, 119665, '\p{^Nv=+000000090}', "");
Expect(0, 119665, '\P{Nv=+000000090}', "");
Expect(1, 119665, '\P{^Nv=+000000090}', "");
Expect(0, 119666, '\p{Nv=+000000090}', "");
Expect(1, 119666, '\p{^Nv=+000000090}', "");
Expect(1, 119666, '\P{Nv=+000000090}', "");
Expect(0, 119666, '\P{^Nv=+000000090}', "");
Error('\p{Is_Numeric_Value=--00000000090/a/}');
Error('\P{Is_Numeric_Value=--00000000090/a/}');
Expect(1, 119665, '\p{Is_Numeric_Value=00000090}', "");
Expect(0, 119665, '\p{^Is_Numeric_Value=00000090}', "");
Expect(0, 119665, '\P{Is_Numeric_Value=00000090}', "");
Expect(1, 119665, '\P{^Is_Numeric_Value=00000090}', "");
Expect(0, 119666, '\p{Is_Numeric_Value=00000090}', "");
Expect(1, 119666, '\p{^Is_Numeric_Value=00000090}', "");
Expect(1, 119666, '\P{Is_Numeric_Value=00000090}', "");
Expect(0, 119666, '\P{^Is_Numeric_Value=00000090}', "");
Error('\p{Is_Nv=/a/ 	0_0_0_0_0_0090}');
Error('\P{Is_Nv=/a/ 	0_0_0_0_0_0090}');
Expect(1, 119665, '\p{Is_Nv=0000090}', "");
Expect(0, 119665, '\p{^Is_Nv=0000090}', "");
Expect(0, 119665, '\P{Is_Nv=0000090}', "");
Expect(1, 119665, '\P{^Is_Nv=0000090}', "");
Expect(0, 119666, '\p{Is_Nv=0000090}', "");
Expect(1, 119666, '\p{^Is_Nv=0000090}', "");
Expect(1, 119666, '\P{Is_Nv=0000090}', "");
Expect(0, 119666, '\P{^Is_Nv=0000090}', "");
Error('\p{Numeric_Value: /a/+0007}');
Error('\P{Numeric_Value: /a/+0007}');
Expect(1, 131073, '\p{Numeric_Value=000000007}', "");
Expect(0, 131073, '\p{^Numeric_Value=000000007}', "");
Expect(0, 131073, '\P{Numeric_Value=000000007}', "");
Expect(1, 131073, '\P{^Numeric_Value=000000007}', "");
Expect(0, 131074, '\p{Numeric_Value=000000007}', "");
Expect(1, 131074, '\p{^Numeric_Value=000000007}', "");
Expect(1, 131074, '\P{Numeric_Value=000000007}', "");
Expect(0, 131074, '\P{^Numeric_Value=000000007}', "");
Error('\p{Nv=:=	0007}');
Error('\P{Nv=:=	0007}');
Expect(1, 131073, '\p{Nv=+00000007}', "");
Expect(0, 131073, '\p{^Nv=+00000007}', "");
Expect(0, 131073, '\P{Nv=+00000007}', "");
Expect(1, 131073, '\P{^Nv=+00000007}', "");
Expect(0, 131074, '\p{Nv=+00000007}', "");
Expect(1, 131074, '\p{^Nv=+00000007}', "");
Expect(1, 131074, '\P{Nv=+00000007}', "");
Expect(0, 131074, '\P{^Nv=+00000007}', "");
Error('\p{Is_Numeric_Value=	/a/0007}');
Error('\P{Is_Numeric_Value=	/a/0007}');
Expect(1, 131073, '\p{Is_Numeric_Value=0_0_0_0_0007}', "");
Expect(0, 131073, '\p{^Is_Numeric_Value=0_0_0_0_0007}', "");
Expect(0, 131073, '\P{Is_Numeric_Value=0_0_0_0_0007}', "");
Expect(1, 131073, '\P{^Is_Numeric_Value=0_0_0_0_0007}', "");
Expect(0, 131074, '\p{Is_Numeric_Value=0_0_0_0_0007}', "");
Expect(1, 131074, '\p{^Is_Numeric_Value=0_0_0_0_0007}', "");
Expect(1, 131074, '\P{Is_Numeric_Value=0_0_0_0_0007}', "");
Expect(0, 131074, '\P{^Is_Numeric_Value=0_0_0_0_0007}', "");
Error('\p{Is_Nv=/a/+00007}');
Error('\P{Is_Nv=/a/+00007}');
Expect(1, 131073, '\p{Is_Nv=7}', "");
Expect(0, 131073, '\p{^Is_Nv=7}', "");
Expect(0, 131073, '\P{Is_Nv=7}', "");
Expect(1, 131073, '\P{^Is_Nv=7}', "");
Expect(0, 131074, '\p{Is_Nv=7}', "");
Expect(1, 131074, '\p{^Is_Nv=7}', "");
Expect(1, 131074, '\P{Is_Nv=7}', "");
Expect(0, 131074, '\P{^Is_Nv=7}', "");
Error('\p{Numeric_Value=-:=0_0_0_0_0_0_6_00}');
Error('\P{Numeric_Value=-:=0_0_0_0_0_0_6_00}');
Expect(1, 69239, '\p{Numeric_Value=000060_0}', "");
Expect(0, 69239, '\p{^Numeric_Value=000060_0}', "");
Expect(0, 69239, '\P{Numeric_Value=000060_0}', "");
Expect(1, 69239, '\P{^Numeric_Value=000060_0}', "");
Expect(0, 69240, '\p{Numeric_Value=000060_0}', "");
Expect(1, 69240, '\p{^Numeric_Value=000060_0}', "");
Expect(1, 69240, '\P{Numeric_Value=000060_0}', "");
Expect(0, 69240, '\P{^Numeric_Value=000060_0}', "");
Error('\p{Nv=/a/	000_000_060_0}');
Error('\P{Nv=/a/	000_000_060_0}');
Expect(1, 69239, '\p{Nv=0_0_0_0_600}', "");
Expect(0, 69239, '\p{^Nv=0_0_0_0_600}', "");
Expect(0, 69239, '\P{Nv=0_0_0_0_600}', "");
Expect(1, 69239, '\P{^Nv=0_0_0_0_600}', "");
Expect(0, 69240, '\p{Nv=0_0_0_0_600}', "");
Expect(1, 69240, '\p{^Nv=0_0_0_0_600}', "");
Expect(1, 69240, '\P{Nv=0_0_0_0_600}', "");
Expect(0, 69240, '\P{^Nv=0_0_0_0_600}', "");
Error('\p{Is_Numeric_Value:   :=+000000600}');
Error('\P{Is_Numeric_Value:   :=+000000600}');
Expect(1, 69239, '\p{Is_Numeric_Value=0_0_0_0_600}', "");
Expect(0, 69239, '\p{^Is_Numeric_Value=0_0_0_0_600}', "");
Expect(0, 69239, '\P{Is_Numeric_Value=0_0_0_0_600}', "");
Expect(1, 69239, '\P{^Is_Numeric_Value=0_0_0_0_600}', "");
Expect(0, 69240, '\p{Is_Numeric_Value=0_0_0_0_600}', "");
Expect(1, 69240, '\p{^Is_Numeric_Value=0_0_0_0_600}', "");
Expect(1, 69240, '\P{Is_Numeric_Value=0_0_0_0_600}', "");
Expect(0, 69240, '\P{^Is_Numeric_Value=0_0_0_0_600}', "");
Error('\p{Is_Nv=	00000600/a/}');
Error('\P{Is_Nv=	00000600/a/}');
Expect(1, 69239, '\p{Is_Nv:	+0600}', "");
Expect(0, 69239, '\p{^Is_Nv:	+0600}', "");
Expect(0, 69239, '\P{Is_Nv:	+0600}', "");
Expect(1, 69239, '\P{^Is_Nv:	+0600}', "");
Expect(0, 69240, '\p{Is_Nv:	+0600}', "");
Expect(1, 69240, '\p{^Is_Nv:	+0600}', "");
Expect(1, 69240, '\P{Is_Nv:	+0600}', "");
Expect(0, 69240, '\P{^Is_Nv:	+0600}', "");
Error('\p{Numeric_Value=/a/__+0_0_0_26}');
Error('\P{Numeric_Value=/a/__+0_0_0_26}');
Expect(1, 12886, '\p{Numeric_Value=+00000000026}', "");
Expect(0, 12886, '\p{^Numeric_Value=+00000000026}', "");
Expect(0, 12886, '\P{Numeric_Value=+00000000026}', "");
Expect(1, 12886, '\P{^Numeric_Value=+00000000026}', "");
Expect(0, 12887, '\p{Numeric_Value=+00000000026}', "");
Expect(1, 12887, '\p{^Numeric_Value=+00000000026}', "");
Expect(1, 12887, '\P{Numeric_Value=+00000000026}', "");
Expect(0, 12887, '\P{^Numeric_Value=+00000000026}', "");
Error('\p{Nv= /a/02_6}');
Error('\P{Nv= /a/02_6}');
Expect(1, 12886, '\p{Nv:0000000026}', "");
Expect(0, 12886, '\p{^Nv:0000000026}', "");
Expect(0, 12886, '\P{Nv:0000000026}', "");
Expect(1, 12886, '\P{^Nv:0000000026}', "");
Expect(0, 12887, '\p{Nv:0000000026}', "");
Expect(1, 12887, '\p{^Nv:0000000026}', "");
Expect(1, 12887, '\P{Nv:0000000026}', "");
Expect(0, 12887, '\P{^Nv:0000000026}', "");
Error('\p{Is_Numeric_Value=:=026}');
Error('\P{Is_Numeric_Value=:=026}');
Expect(1, 12886, '\p{Is_Numeric_Value=0_0_0_0_26}', "");
Expect(0, 12886, '\p{^Is_Numeric_Value=0_0_0_0_26}', "");
Expect(0, 12886, '\P{Is_Numeric_Value=0_0_0_0_26}', "");
Expect(1, 12886, '\P{^Is_Numeric_Value=0_0_0_0_26}', "");
Expect(0, 12887, '\p{Is_Numeric_Value=0_0_0_0_26}', "");
Expect(1, 12887, '\p{^Is_Numeric_Value=0_0_0_0_26}', "");
Expect(1, 12887, '\P{Is_Numeric_Value=0_0_0_0_26}', "");
Expect(0, 12887, '\P{^Is_Numeric_Value=0_0_0_0_26}', "");
Error('\p{Is_Nv=/a/ 	0026}');
Error('\P{Is_Nv=/a/ 	0026}');
Expect(1, 12886, '\p{Is_Nv=0000_0000_26}', "");
Expect(0, 12886, '\p{^Is_Nv=0000_0000_26}', "");
Expect(0, 12886, '\P{Is_Nv=0000_0000_26}', "");
Expect(1, 12886, '\P{^Is_Nv=0000_0000_26}', "");
Expect(0, 12887, '\p{Is_Nv=0000_0000_26}', "");
Expect(1, 12887, '\p{^Is_Nv=0000_0000_26}', "");
Expect(1, 12887, '\P{Is_Nv=0000_0000_26}', "");
Expect(0, 12887, '\P{^Is_Nv=0000_0000_26}', "");
Error('\p{Numeric_Value:	_08_0:=}');
Error('\P{Numeric_Value:	_08_0:=}');
Expect(1, 119664, '\p{Numeric_Value=00080}', "");
Expect(0, 119664, '\p{^Numeric_Value=00080}', "");
Expect(0, 119664, '\P{Numeric_Value=00080}', "");
Expect(1, 119664, '\P{^Numeric_Value=00080}', "");
Expect(0, 119665, '\p{Numeric_Value=00080}', "");
Expect(1, 119665, '\p{^Numeric_Value=00080}', "");
Expect(1, 119665, '\P{Numeric_Value=00080}', "");
Expect(0, 119665, '\P{^Numeric_Value=00080}', "");
Error('\p{Nv: /a/+08_0}');
Error('\P{Nv: /a/+08_0}');
Expect(1, 119664, '\p{Nv=80}', "");
Expect(0, 119664, '\p{^Nv=80}', "");
Expect(0, 119664, '\P{Nv=80}', "");
Expect(1, 119664, '\P{^Nv=80}', "");
Expect(0, 119665, '\p{Nv=80}', "");
Expect(1, 119665, '\p{^Nv=80}', "");
Expect(1, 119665, '\P{Nv=80}', "");
Expect(0, 119665, '\P{^Nv=80}', "");
Error('\p{Is_Numeric_Value=:=- 0080}');
Error('\P{Is_Numeric_Value=:=- 0080}');
Expect(1, 119664, '\p{Is_Numeric_Value=+8_0}', "");
Expect(0, 119664, '\p{^Is_Numeric_Value=+8_0}', "");
Expect(0, 119664, '\P{Is_Numeric_Value=+8_0}', "");
Expect(1, 119664, '\P{^Is_Numeric_Value=+8_0}', "");
Expect(0, 119665, '\p{Is_Numeric_Value=+8_0}', "");
Expect(1, 119665, '\p{^Is_Numeric_Value=+8_0}', "");
Expect(1, 119665, '\P{Is_Numeric_Value=+8_0}', "");
Expect(0, 119665, '\P{^Is_Numeric_Value=+8_0}', "");
Error('\p{Is_Nv=_/a/0000080}');
Error('\P{Is_Nv=_/a/0000080}');
Expect(1, 119664, '\p{Is_Nv=00_00_08_0}', "");
Expect(0, 119664, '\p{^Is_Nv=00_00_08_0}', "");
Expect(0, 119664, '\P{Is_Nv=00_00_08_0}', "");
Expect(1, 119664, '\P{^Is_Nv=00_00_08_0}', "");
Expect(0, 119665, '\p{Is_Nv=00_00_08_0}', "");
Expect(1, 119665, '\p{^Is_Nv=00_00_08_0}', "");
Expect(1, 119665, '\P{Is_Nv=00_00_08_0}', "");
Expect(0, 119665, '\P{^Is_Nv=00_00_08_0}', "");
Error('\p{Numeric_Value=/a/__+0000200}');
Error('\P{Numeric_Value=/a/__+0000200}');
Expect(1, 69235, '\p{Numeric_Value=0_2_00}', "");
Expect(0, 69235, '\p{^Numeric_Value=0_2_00}', "");
Expect(0, 69235, '\P{Numeric_Value=0_2_00}', "");
Expect(1, 69235, '\P{^Numeric_Value=0_2_00}', "");
Expect(0, 69236, '\p{Numeric_Value=0_2_00}', "");
Expect(1, 69236, '\p{^Numeric_Value=0_2_00}', "");
Expect(1, 69236, '\P{Numeric_Value=0_2_00}', "");
Expect(0, 69236, '\P{^Numeric_Value=0_2_00}', "");
Error('\p{Nv=:=000000000200}');
Error('\P{Nv=:=000000000200}');
Expect(1, 69235, '\p{Nv=20_0}', "");
Expect(0, 69235, '\p{^Nv=20_0}', "");
Expect(0, 69235, '\P{Nv=20_0}', "");
Expect(1, 69235, '\P{^Nv=20_0}', "");
Expect(0, 69236, '\p{Nv=20_0}', "");
Expect(1, 69236, '\p{^Nv=20_0}', "");
Expect(1, 69236, '\P{Nv=20_0}', "");
Expect(0, 69236, '\P{^Nv=20_0}', "");
Error('\p{Is_Numeric_Value=:=  +00000200}');
Error('\P{Is_Numeric_Value=:=  +00000200}');
Expect(1, 69235, '\p{Is_Numeric_Value=0_0_0_0_200}', "");
Expect(0, 69235, '\p{^Is_Numeric_Value=0_0_0_0_200}', "");
Expect(0, 69235, '\P{Is_Numeric_Value=0_0_0_0_200}', "");
Expect(1, 69235, '\P{^Is_Numeric_Value=0_0_0_0_200}', "");
Expect(0, 69236, '\p{Is_Numeric_Value=0_0_0_0_200}', "");
Expect(1, 69236, '\p{^Is_Numeric_Value=0_0_0_0_200}', "");
Expect(1, 69236, '\P{Is_Numeric_Value=0_0_0_0_200}', "");
Expect(0, 69236, '\P{^Is_Numeric_Value=0_0_0_0_200}', "");
Error('\p{Is_Nv:-:=00_00_00_200}');
Error('\P{Is_Nv:-:=00_00_00_200}');
Expect(1, 69235, '\p{Is_Nv=000200}', "");
Expect(0, 69235, '\p{^Is_Nv=000200}', "");
Expect(0, 69235, '\P{Is_Nv=000200}', "");
Expect(1, 69235, '\P{^Is_Nv=000200}', "");
Expect(0, 69236, '\p{Is_Nv=000200}', "");
Expect(1, 69236, '\p{^Is_Nv=000200}', "");
Expect(1, 69236, '\P{Is_Nv=000200}', "");
Expect(0, 69236, '\P{^Is_Nv=000200}', "");
Error('\p{Numeric_Value=/a/	 0000001_8}');
Error('\P{Numeric_Value=/a/	 0000001_8}');
Expect(1, 9458, '\p{Numeric_Value=+0000000001_8}', "");
Expect(0, 9458, '\p{^Numeric_Value=+0000000001_8}', "");
Expect(0, 9458, '\P{Numeric_Value=+0000000001_8}', "");
Expect(1, 9458, '\P{^Numeric_Value=+0000000001_8}', "");
Expect(0, 9459, '\p{Numeric_Value=+0000000001_8}', "");
Expect(1, 9459, '\p{^Numeric_Value=+0000000001_8}', "");
Expect(1, 9459, '\P{Numeric_Value=+0000000001_8}', "");
Expect(0, 9459, '\P{^Numeric_Value=+0000000001_8}', "");
Error('\p{Nv= _+0_0_18:=}');
Error('\P{Nv= _+0_0_18:=}');
Expect(1, 9458, '\p{Nv=0000018}', "");
Expect(0, 9458, '\p{^Nv=0000018}', "");
Expect(0, 9458, '\P{Nv=0000018}', "");
Expect(1, 9458, '\P{^Nv=0000018}', "");
Expect(0, 9459, '\p{Nv=0000018}', "");
Expect(1, 9459, '\p{^Nv=0000018}', "");
Expect(1, 9459, '\P{Nv=0000018}', "");
Expect(0, 9459, '\P{^Nv=0000018}', "");
Error('\p{Is_Numeric_Value=/a/018}');
Error('\P{Is_Numeric_Value=/a/018}');
Expect(1, 9458, '\p{Is_Numeric_Value=+000_000_18}', "");
Expect(0, 9458, '\p{^Is_Numeric_Value=+000_000_18}', "");
Expect(0, 9458, '\P{Is_Numeric_Value=+000_000_18}', "");
Expect(1, 9458, '\P{^Is_Numeric_Value=+000_000_18}', "");
Expect(0, 9459, '\p{Is_Numeric_Value=+000_000_18}', "");
Expect(1, 9459, '\p{^Is_Numeric_Value=+000_000_18}', "");
Expect(1, 9459, '\P{Is_Numeric_Value=+000_000_18}', "");
Expect(0, 9459, '\P{^Is_Numeric_Value=+000_000_18}', "");
Error('\p{Is_Nv=_:=00000_00001_8}');
Error('\P{Is_Nv=_:=00000_00001_8}');
Expect(1, 9458, '\p{Is_Nv=000_000_18}', "");
Expect(0, 9458, '\p{^Is_Nv=000_000_18}', "");
Expect(0, 9458, '\P{Is_Nv=000_000_18}', "");
Expect(1, 9458, '\P{^Is_Nv=000_000_18}', "");
Expect(0, 9459, '\p{Is_Nv=000_000_18}', "");
Expect(1, 9459, '\p{^Is_Nv=000_000_18}', "");
Expect(1, 9459, '\P{Is_Nv=000_000_18}', "");
Expect(0, 9459, '\P{^Is_Nv=000_000_18}', "");
Error('\p{Numeric_Value=/a/_ 0003/000000008}');
Error('\P{Numeric_Value=/a/_ 0003/000000008}');
Expect(1, 8540, '\p{Numeric_Value=3/000000008}', "");
Expect(0, 8540, '\p{^Numeric_Value=3/000000008}', "");
Expect(0, 8540, '\P{Numeric_Value=3/000000008}', "");
Expect(1, 8540, '\P{^Numeric_Value=3/000000008}', "");
Expect(0, 8541, '\p{Numeric_Value=3/000000008}', "");
Expect(1, 8541, '\p{^Numeric_Value=3/000000008}', "");
Expect(1, 8541, '\P{Numeric_Value=3/000000008}', "");
Expect(0, 8541, '\P{^Numeric_Value=3/000000008}', "");
Error('\p{Numeric_Value:0.38}');
Error('\P{Numeric_Value:0.38}');
Expect(1, 8540, '\p{Numeric_Value=0.375}', "");
Expect(0, 8540, '\p{^Numeric_Value=0.375}', "");
Expect(0, 8540, '\P{Numeric_Value=0.375}', "");
Expect(1, 8540, '\P{^Numeric_Value=0.375}', "");
Expect(0, 8541, '\p{Numeric_Value=0.375}', "");
Expect(1, 8541, '\p{^Numeric_Value=0.375}', "");
Expect(1, 8541, '\P{Numeric_Value=0.375}', "");
Expect(0, 8541, '\P{^Numeric_Value=0.375}', "");
Error('\p{Nv:   -/a/00000003/00008}');
Error('\P{Nv:   -/a/00000003/00008}');
Expect(1, 8540, '\p{Nv=+000003/0000008}', "");
Expect(0, 8540, '\p{^Nv=+000003/0000008}', "");
Expect(0, 8540, '\P{Nv=+000003/0000008}', "");
Expect(1, 8540, '\P{^Nv=+000003/0000008}', "");
Expect(0, 8541, '\p{Nv=+000003/0000008}', "");
Expect(1, 8541, '\p{^Nv=+000003/0000008}', "");
Expect(1, 8541, '\P{Nv=+000003/0000008}', "");
Expect(0, 8541, '\P{^Nv=+000003/0000008}', "");
Error('\p{Nv:0.38}');
Error('\P{Nv:0.38}');
Expect(1, 8540, '\p{Nv:   0.375}', "");
Expect(0, 8540, '\p{^Nv:   0.375}', "");
Expect(0, 8540, '\P{Nv:   0.375}', "");
Expect(1, 8540, '\P{^Nv:   0.375}', "");
Expect(0, 8541, '\p{Nv:   0.375}', "");
Expect(1, 8541, '\p{^Nv:   0.375}', "");
Expect(1, 8541, '\P{Nv:   0.375}', "");
Expect(0, 8541, '\P{^Nv:   0.375}', "");
Error('\p{Is_Numeric_Value:    :=000003/008}');
Error('\P{Is_Numeric_Value:    :=000003/008}');
Expect(1, 8540, '\p{Is_Numeric_Value=+0000000003/00008}', "");
Expect(0, 8540, '\p{^Is_Numeric_Value=+0000000003/00008}', "");
Expect(0, 8540, '\P{Is_Numeric_Value=+0000000003/00008}', "");
Expect(1, 8540, '\P{^Is_Numeric_Value=+0000000003/00008}', "");
Expect(0, 8541, '\p{Is_Numeric_Value=+0000000003/00008}', "");
Expect(1, 8541, '\p{^Is_Numeric_Value=+0000000003/00008}', "");
Expect(1, 8541, '\P{Is_Numeric_Value=+0000000003/00008}', "");
Expect(0, 8541, '\P{^Is_Numeric_Value=+0000000003/00008}', "");
Error('\p{Is_Numeric_Value=0.38}');
Error('\P{Is_Numeric_Value=0.38}');
Expect(1, 8540, '\p{Is_Numeric_Value=0.375}', "");
Expect(0, 8540, '\p{^Is_Numeric_Value=0.375}', "");
Expect(0, 8540, '\P{Is_Numeric_Value=0.375}', "");
Expect(1, 8540, '\P{^Is_Numeric_Value=0.375}', "");
Expect(0, 8541, '\p{Is_Numeric_Value=0.375}', "");
Expect(1, 8541, '\p{^Is_Numeric_Value=0.375}', "");
Expect(1, 8541, '\P{Is_Numeric_Value=0.375}', "");
Expect(0, 8541, '\P{^Is_Numeric_Value=0.375}', "");
Error('\p{Is_Nv=-_00003/000000008:=}');
Error('\P{Is_Nv=-_00003/000000008:=}');
Expect(1, 8540, '\p{Is_Nv=+000003/08}', "");
Expect(0, 8540, '\p{^Is_Nv=+000003/08}', "");
Expect(0, 8540, '\P{Is_Nv=+000003/08}', "");
Expect(1, 8540, '\P{^Is_Nv=+000003/08}', "");
Expect(0, 8541, '\p{Is_Nv=+000003/08}', "");
Expect(1, 8541, '\p{^Is_Nv=+000003/08}', "");
Expect(1, 8541, '\P{Is_Nv=+000003/08}', "");
Expect(0, 8541, '\P{^Is_Nv=+000003/08}', "");
Error('\p{Is_Nv: 0.38}');
Error('\P{Is_Nv: 0.38}');
Expect(1, 8540, '\p{Is_Nv=0.375}', "");
Expect(0, 8540, '\p{^Is_Nv=0.375}', "");
Expect(0, 8540, '\P{Is_Nv=0.375}', "");
Expect(1, 8540, '\P{^Is_Nv=0.375}', "");
Expect(0, 8541, '\p{Is_Nv=0.375}', "");
Expect(1, 8541, '\p{^Is_Nv=0.375}', "");
Expect(1, 8541, '\P{Is_Nv=0.375}', "");
Expect(0, 8541, '\P{^Is_Nv=0.375}', "");
Error('\p{Numeric_Value=/a/	 +000000500}');
Error('\P{Numeric_Value=/a/	 +000000500}');
Expect(1, 69238, '\p{Numeric_Value=0000_0005_00}', "");
Expect(0, 69238, '\p{^Numeric_Value=0000_0005_00}', "");
Expect(0, 69238, '\P{Numeric_Value=0000_0005_00}', "");
Expect(1, 69238, '\P{^Numeric_Value=0000_0005_00}', "");
Expect(0, 69239, '\p{Numeric_Value=0000_0005_00}', "");
Expect(1, 69239, '\p{^Numeric_Value=0000_0005_00}', "");
Expect(1, 69239, '\P{Numeric_Value=0000_0005_00}', "");
Expect(0, 69239, '\P{^Numeric_Value=0000_0005_00}', "");
Error('\p{Nv:	 /a/0050_0}');
Error('\P{Nv:	 /a/0050_0}');
Expect(1, 69238, '\p{Nv=+00000000050_0}', "");
Expect(0, 69238, '\p{^Nv=+00000000050_0}', "");
Expect(0, 69238, '\P{Nv=+00000000050_0}', "");
Expect(1, 69238, '\P{^Nv=+00000000050_0}', "");
Expect(0, 69239, '\p{Nv=+00000000050_0}', "");
Expect(1, 69239, '\p{^Nv=+00000000050_0}', "");
Expect(1, 69239, '\P{Nv=+00000000050_0}', "");
Expect(0, 69239, '\P{^Nv=+00000000050_0}', "");
Error('\p{Is_Numeric_Value=	/a/0_0_0_0_500}');
Error('\P{Is_Numeric_Value=	/a/0_0_0_0_500}');
Expect(1, 69238, '\p{Is_Numeric_Value=50_0}', "");
Expect(0, 69238, '\p{^Is_Numeric_Value=50_0}', "");
Expect(0, 69238, '\P{Is_Numeric_Value=50_0}', "");
Expect(1, 69238, '\P{^Is_Numeric_Value=50_0}', "");
Expect(0, 69239, '\p{Is_Numeric_Value=50_0}', "");
Expect(1, 69239, '\p{^Is_Numeric_Value=50_0}', "");
Expect(1, 69239, '\P{Is_Numeric_Value=50_0}', "");
Expect(0, 69239, '\P{^Is_Numeric_Value=50_0}', "");
Error('\p{Is_Nv=-:=00_05_00}');
Error('\P{Is_Nv=-:=00_05_00}');
Expect(1, 69238, '\p{Is_Nv=+00000050_0}', "");
Expect(0, 69238, '\p{^Is_Nv=+00000050_0}', "");
Expect(0, 69238, '\P{Is_Nv=+00000050_0}', "");
Expect(1, 69238, '\P{^Is_Nv=+00000050_0}', "");
Expect(0, 69239, '\p{Is_Nv=+00000050_0}', "");
Expect(1, 69239, '\p{^Is_Nv=+00000050_0}', "");
Expect(1, 69239, '\P{Is_Nv=+00000050_0}', "");
Expect(0, 69239, '\P{^Is_Nv=+00000050_0}', "");
Error('\p{Numeric_Value=/a/00000000016}');
Error('\P{Numeric_Value=/a/00000000016}');
Expect(1, 9456, '\p{Numeric_Value=00016}', "");
Expect(0, 9456, '\p{^Numeric_Value=00016}', "");
Expect(0, 9456, '\P{Numeric_Value=00016}', "");
Expect(1, 9456, '\P{^Numeric_Value=00016}', "");
Expect(0, 9457, '\p{Numeric_Value=00016}', "");
Expect(1, 9457, '\p{^Numeric_Value=00016}', "");
Expect(1, 9457, '\P{Numeric_Value=00016}', "");
Expect(0, 9457, '\P{^Numeric_Value=00016}', "");
Error('\p{Nv=_:=00016}');
Error('\P{Nv=_:=00016}');
Expect(1, 9456, '\p{Nv=016}', "");
Expect(0, 9456, '\p{^Nv=016}', "");
Expect(0, 9456, '\P{Nv=016}', "");
Expect(1, 9456, '\P{^Nv=016}', "");
Expect(0, 9457, '\p{Nv=016}', "");
Expect(1, 9457, '\p{^Nv=016}', "");
Expect(1, 9457, '\P{Nv=016}', "");
Expect(0, 9457, '\P{^Nv=016}', "");
Error('\p{Is_Numeric_Value: /a/__1_6}');
Error('\P{Is_Numeric_Value: /a/__1_6}');
Expect(1, 9456, '\p{Is_Numeric_Value=000_000_001_6}', "");
Expect(0, 9456, '\p{^Is_Numeric_Value=000_000_001_6}', "");
Expect(0, 9456, '\P{Is_Numeric_Value=000_000_001_6}', "");
Expect(1, 9456, '\P{^Is_Numeric_Value=000_000_001_6}', "");
Expect(0, 9457, '\p{Is_Numeric_Value=000_000_001_6}', "");
Expect(1, 9457, '\p{^Is_Numeric_Value=000_000_001_6}', "");
Expect(1, 9457, '\P{Is_Numeric_Value=000_000_001_6}', "");
Expect(0, 9457, '\P{^Is_Numeric_Value=000_000_001_6}', "");
Error('\p{Is_Nv= :=+0_0_0_0_0_0_0_0_016}');
Error('\P{Is_Nv= :=+0_0_0_0_0_0_0_0_016}');
Expect(1, 9456, '\p{Is_Nv=016}', "");
Expect(0, 9456, '\p{^Is_Nv=016}', "");
Expect(0, 9456, '\P{Is_Nv=016}', "");
Expect(1, 9456, '\P{^Is_Nv=016}', "");
Expect(0, 9457, '\p{Is_Nv=016}', "");
Expect(1, 9457, '\p{^Is_Nv=016}', "");
Expect(1, 9457, '\P{Is_Nv=016}', "");
Expect(0, 9457, '\P{^Is_Nv=016}', "");
Error('\p{Numeric_Value=:=	-044}');
Error('\P{Numeric_Value=:=	-044}');
Expect(1, 12985, '\p{Numeric_Value=04_4}', "");
Expect(0, 12985, '\p{^Numeric_Value=04_4}', "");
Expect(0, 12985, '\P{Numeric_Value=04_4}', "");
Expect(1, 12985, '\P{^Numeric_Value=04_4}', "");
Expect(0, 12986, '\p{Numeric_Value=04_4}', "");
Expect(1, 12986, '\p{^Numeric_Value=04_4}', "");
Expect(1, 12986, '\P{Numeric_Value=04_4}', "");
Expect(0, 12986, '\P{^Numeric_Value=04_4}', "");
Error('\p{Nv=_:=00_00_00_44}');
Error('\P{Nv=_:=00_00_00_44}');
Expect(1, 12985, '\p{Nv=+0_0_44}', "");
Expect(0, 12985, '\p{^Nv=+0_0_44}', "");
Expect(0, 12985, '\P{Nv=+0_0_44}', "");
Expect(1, 12985, '\P{^Nv=+0_0_44}', "");
Expect(0, 12986, '\p{Nv=+0_0_44}', "");
Expect(1, 12986, '\p{^Nv=+0_0_44}', "");
Expect(1, 12986, '\P{Nv=+0_0_44}', "");
Expect(0, 12986, '\P{^Nv=+0_0_44}', "");
Error('\p{Is_Numeric_Value=_-+000044:=}');
Error('\P{Is_Numeric_Value=_-+000044:=}');
Expect(1, 12985, '\p{Is_Numeric_Value=0044}', "");
Expect(0, 12985, '\p{^Is_Numeric_Value=0044}', "");
Expect(0, 12985, '\P{Is_Numeric_Value=0044}', "");
Expect(1, 12985, '\P{^Is_Numeric_Value=0044}', "");
Expect(0, 12986, '\p{Is_Numeric_Value=0044}', "");
Expect(1, 12986, '\p{^Is_Numeric_Value=0044}', "");
Expect(1, 12986, '\P{Is_Numeric_Value=0044}', "");
Expect(0, 12986, '\P{^Is_Numeric_Value=0044}', "");
Error('\p{Is_Nv=:=+0000044}');
Error('\P{Is_Nv=:=+0000044}');
Expect(1, 12985, '\p{Is_Nv=+44}', "");
Expect(0, 12985, '\p{^Is_Nv=+44}', "");
Expect(0, 12985, '\P{Is_Nv=+44}', "");
Expect(1, 12985, '\P{^Is_Nv=+44}', "");
Expect(0, 12986, '\p{Is_Nv=+44}', "");
Expect(1, 12986, '\p{^Is_Nv=+44}', "");
Expect(1, 12986, '\P{Is_Nv=+44}', "");
Expect(0, 12986, '\P{^Is_Nv=+44}', "");
Error('\p{Numeric_Value= :=+00027}');
Error('\P{Numeric_Value= :=+00027}');
Expect(1, 12887, '\p{Numeric_Value=+000027}', "");
Expect(0, 12887, '\p{^Numeric_Value=+000027}', "");
Expect(0, 12887, '\P{Numeric_Value=+000027}', "");
Expect(1, 12887, '\P{^Numeric_Value=+000027}', "");
Expect(0, 12888, '\p{Numeric_Value=+000027}', "");
Expect(1, 12888, '\p{^Numeric_Value=+000027}', "");
Expect(1, 12888, '\P{Numeric_Value=+000027}', "");
Expect(0, 12888, '\P{^Numeric_Value=+000027}', "");
Error('\p{Nv:	_:=+00_02_7}');
Error('\P{Nv:	_:=+00_02_7}');
Expect(1, 12887, '\p{Nv=00027}', "");
Expect(0, 12887, '\p{^Nv=00027}', "");
Expect(0, 12887, '\P{Nv=00027}', "");
Expect(1, 12887, '\P{^Nv=00027}', "");
Expect(0, 12888, '\p{Nv=00027}', "");
Expect(1, 12888, '\p{^Nv=00027}', "");
Expect(1, 12888, '\P{Nv=00027}', "");
Expect(0, 12888, '\P{^Nv=00027}', "");
Error('\p{Is_Numeric_Value= :=000000002_7}');
Error('\P{Is_Numeric_Value= :=000000002_7}');
Expect(1, 12887, '\p{Is_Numeric_Value=00000000027}', "");
Expect(0, 12887, '\p{^Is_Numeric_Value=00000000027}', "");
Expect(0, 12887, '\P{Is_Numeric_Value=00000000027}', "");
Expect(1, 12887, '\P{^Is_Numeric_Value=00000000027}', "");
Expect(0, 12888, '\p{Is_Numeric_Value=00000000027}', "");
Expect(1, 12888, '\p{^Is_Numeric_Value=00000000027}', "");
Expect(1, 12888, '\P{Is_Numeric_Value=00000000027}', "");
Expect(0, 12888, '\P{^Is_Numeric_Value=00000000027}', "");
Error('\p{Is_Nv= 0_0_0_0_0_0_027:=}');
Error('\P{Is_Nv= 0_0_0_0_0_0_027:=}');
Expect(1, 12887, '\p{Is_Nv=+00027}', "");
Expect(0, 12887, '\p{^Is_Nv=+00027}', "");
Expect(0, 12887, '\P{Is_Nv=+00027}', "");
Expect(1, 12887, '\P{^Is_Nv=+00027}', "");
Expect(0, 12888, '\p{Is_Nv=+00027}', "");
Expect(1, 12888, '\p{^Is_Nv=+00027}', "");
Expect(1, 12888, '\P{Is_Nv=+00027}', "");
Expect(0, 12888, '\P{^Is_Nv=+00027}', "");
Error('\p{Numeric_Value=_:=01_00_00_00_00}');
Error('\P{Numeric_Value=_:=01_00_00_00_00}');
Expect(1, 20740, '\p{Numeric_Value=+00100000000}', "");
Expect(0, 20740, '\p{^Numeric_Value=+00100000000}', "");
Expect(0, 20740, '\P{Numeric_Value=+00100000000}', "");
Expect(1, 20740, '\P{^Numeric_Value=+00100000000}', "");
Expect(0, 20741, '\p{Numeric_Value=+00100000000}', "");
Expect(1, 20741, '\p{^Numeric_Value=+00100000000}', "");
Expect(1, 20741, '\P{Numeric_Value=+00100000000}', "");
Expect(0, 20741, '\P{^Numeric_Value=+00100000000}', "");
Error('\p{Nv:	/a/-000000100000000}');
Error('\P{Nv:	/a/-000000100000000}');
Expect(1, 20740, '\p{Nv=100000000}', "");
Expect(0, 20740, '\p{^Nv=100000000}', "");
Expect(0, 20740, '\P{Nv=100000000}', "");
Expect(1, 20740, '\P{^Nv=100000000}', "");
Expect(0, 20741, '\p{Nv=100000000}', "");
Expect(1, 20741, '\p{^Nv=100000000}', "");
Expect(1, 20741, '\P{Nv=100000000}', "");
Expect(0, 20741, '\P{^Nv=100000000}', "");
Error('\p{Is_Numeric_Value=--+000000100000000/a/}');
Error('\P{Is_Numeric_Value=--+000000100000000/a/}');
Expect(1, 20740, '\p{Is_Numeric_Value=+100000000}', "");
Expect(0, 20740, '\p{^Is_Numeric_Value=+100000000}', "");
Expect(0, 20740, '\P{Is_Numeric_Value=+100000000}', "");
Expect(1, 20740, '\P{^Is_Numeric_Value=+100000000}', "");
Expect(0, 20741, '\p{Is_Numeric_Value=+100000000}', "");
Expect(1, 20741, '\p{^Is_Numeric_Value=+100000000}', "");
Expect(1, 20741, '\P{Is_Numeric_Value=+100000000}', "");
Expect(0, 20741, '\P{^Is_Numeric_Value=+100000000}', "");
Error('\p{Is_Nv=/a/00_10_00_00_00_0}');
Error('\P{Is_Nv=/a/00_10_00_00_00_0}');
Expect(1, 20740, '\p{Is_Nv: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(0, 20740, '\p{^Is_Nv: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(0, 20740, '\P{Is_Nv: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(1, 20740, '\P{^Is_Nv: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(0, 20741, '\p{Is_Nv: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(1, 20741, '\p{^Is_Nv: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(1, 20741, '\P{Is_Nv: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Expect(0, 20741, '\P{^Is_Nv: 0_0_0_0_0_0_0_0_1_0_0_0_0_0_0_00}', "");
Error('\p{Numeric_Value=	/a/000001/0000010}');
Error('\P{Numeric_Value=	/a/000001/0000010}');
Expect(1, 8530, '\p{Numeric_Value=0001/00010}', "");
Expect(0, 8530, '\p{^Numeric_Value=0001/00010}', "");
Expect(0, 8530, '\P{Numeric_Value=0001/00010}', "");
Expect(1, 8530, '\P{^Numeric_Value=0001/00010}', "");
Expect(0, 8531, '\p{Numeric_Value=0001/00010}', "");
Expect(1, 8531, '\p{^Numeric_Value=0001/00010}', "");
Expect(1, 8531, '\P{Numeric_Value=0001/00010}', "");
Expect(0, 8531, '\P{^Numeric_Value=0001/00010}', "");
Expect(1, 8530, '\p{Numeric_Value=0.100}', "");
Expect(0, 8530, '\p{^Numeric_Value=0.100}', "");
Expect(0, 8530, '\P{Numeric_Value=0.100}', "");
Expect(1, 8530, '\P{^Numeric_Value=0.100}', "");
Expect(0, 8531, '\p{Numeric_Value=0.100}', "");
Expect(1, 8531, '\p{^Numeric_Value=0.100}', "");
Expect(1, 8531, '\P{Numeric_Value=0.100}', "");
Expect(0, 8531, '\P{^Numeric_Value=0.100}', "");
Error('\p{Nv=/a/ +0000001/0000010}');
Error('\P{Nv=/a/ +0000001/0000010}');
Expect(1, 8530, '\p{Nv=0000000001/00000000010}', "");
Expect(0, 8530, '\p{^Nv=0000000001/00000000010}', "");
Expect(0, 8530, '\P{Nv=0000000001/00000000010}', "");
Expect(1, 8530, '\P{^Nv=0000000001/00000000010}', "");
Expect(0, 8531, '\p{Nv=0000000001/00000000010}', "");
Expect(1, 8531, '\p{^Nv=0000000001/00000000010}', "");
Expect(1, 8531, '\P{Nv=0000000001/00000000010}', "");
Expect(0, 8531, '\P{^Nv=0000000001/00000000010}', "");
Expect(1, 8530, '\p{Nv:0.100}', "");
Expect(0, 8530, '\p{^Nv:0.100}', "");
Expect(0, 8530, '\P{Nv:0.100}', "");
Expect(1, 8530, '\P{^Nv:0.100}', "");
Expect(0, 8531, '\p{Nv:0.100}', "");
Expect(1, 8531, '\p{^Nv:0.100}', "");
Expect(1, 8531, '\P{Nv:0.100}', "");
Expect(0, 8531, '\P{^Nv:0.100}', "");
Error('\p{Is_Numeric_Value=/a/+0000001/0000000010}');
Error('\P{Is_Numeric_Value=/a/+0000001/0000000010}');
Expect(1, 8530, '\p{Is_Numeric_Value: 0001/10}', "");
Expect(0, 8530, '\p{^Is_Numeric_Value: 0001/10}', "");
Expect(0, 8530, '\P{Is_Numeric_Value: 0001/10}', "");
Expect(1, 8530, '\P{^Is_Numeric_Value: 0001/10}', "");
Expect(0, 8531, '\p{Is_Numeric_Value: 0001/10}', "");
Expect(1, 8531, '\p{^Is_Numeric_Value: 0001/10}', "");
Expect(1, 8531, '\P{Is_Numeric_Value: 0001/10}', "");
Expect(0, 8531, '\P{^Is_Numeric_Value: 0001/10}', "");
Expect(1, 8530, '\p{Is_Numeric_Value=0.100}', "");
Expect(0, 8530, '\p{^Is_Numeric_Value=0.100}', "");
Expect(0, 8530, '\P{Is_Numeric_Value=0.100}', "");
Expect(1, 8530, '\P{^Is_Numeric_Value=0.100}', "");
Expect(0, 8531, '\p{Is_Numeric_Value=0.100}', "");
Expect(1, 8531, '\p{^Is_Numeric_Value=0.100}', "");
Expect(1, 8531, '\P{Is_Numeric_Value=0.100}', "");
Expect(0, 8531, '\P{^Is_Numeric_Value=0.100}', "");
Error('\p{Is_Nv:	:=+0001/00010}');
Error('\P{Is_Nv:	:=+0001/00010}');
Expect(1, 8530, '\p{Is_Nv=000000001/00000000010}', "");
Expect(0, 8530, '\p{^Is_Nv=000000001/00000000010}', "");
Expect(0, 8530, '\P{Is_Nv=000000001/00000000010}', "");
Expect(1, 8530, '\P{^Is_Nv=000000001/00000000010}', "");
Expect(0, 8531, '\p{Is_Nv=000000001/00000000010}', "");
Expect(1, 8531, '\p{^Is_Nv=000000001/00000000010}', "");
Expect(1, 8531, '\P{Is_Nv=000000001/00000000010}', "");
Expect(0, 8531, '\P{^Is_Nv=000000001/00000000010}', "");
Expect(1, 8530, '\p{Is_Nv=0.100}', "");
Expect(0, 8530, '\p{^Is_Nv=0.100}', "");
Expect(0, 8530, '\P{Is_Nv=0.100}', "");
Expect(1, 8530, '\P{^Is_Nv=0.100}', "");
Expect(0, 8531, '\p{Is_Nv=0.100}', "");
Expect(1, 8531, '\p{^Is_Nv=0.100}', "");
Expect(1, 8531, '\P{Is_Nv=0.100}', "");
Expect(0, 8531, '\P{^Is_Nv=0.100}', "");
Error('\p{Numeric_Value=/a/ -000000040_0}');
Error('\P{Numeric_Value=/a/ -000000040_0}');
Expect(1, 69237, '\p{Numeric_Value=000000000400}', "");
Expect(0, 69237, '\p{^Numeric_Value=000000000400}', "");
Expect(0, 69237, '\P{Numeric_Value=000000000400}', "");
Expect(1, 69237, '\P{^Numeric_Value=000000000400}', "");
Expect(0, 69238, '\p{Numeric_Value=000000000400}', "");
Expect(1, 69238, '\p{^Numeric_Value=000000000400}', "");
Expect(1, 69238, '\P{Numeric_Value=000000000400}', "");
Expect(0, 69238, '\P{^Numeric_Value=000000000400}', "");
Error('\p{Nv=	_+0400:=}');
Error('\P{Nv=	_+0400:=}');
Expect(1, 69237, '\p{Nv=000000400}', "");
Expect(0, 69237, '\p{^Nv=000000400}', "");
Expect(0, 69237, '\P{Nv=000000400}', "");
Expect(1, 69237, '\P{^Nv=000000400}', "");
Expect(0, 69238, '\p{Nv=000000400}', "");
Expect(1, 69238, '\p{^Nv=000000400}', "");
Expect(1, 69238, '\P{Nv=000000400}', "");
Expect(0, 69238, '\P{^Nv=000000400}', "");
Error('\p{Is_Numeric_Value=	:=00000040_0}');
Error('\P{Is_Numeric_Value=	:=00000040_0}');
Expect(1, 69237, '\p{Is_Numeric_Value=0000400}', "");
Expect(0, 69237, '\p{^Is_Numeric_Value=0000400}', "");
Expect(0, 69237, '\P{Is_Numeric_Value=0000400}', "");
Expect(1, 69237, '\P{^Is_Numeric_Value=0000400}', "");
Expect(0, 69238, '\p{Is_Numeric_Value=0000400}', "");
Expect(1, 69238, '\p{^Is_Numeric_Value=0000400}', "");
Expect(1, 69238, '\P{Is_Numeric_Value=0000400}', "");
Expect(0, 69238, '\P{^Is_Numeric_Value=0000400}', "");
Error('\p{Is_Nv= 00_00_00_400:=}');
Error('\P{Is_Nv= 00_00_00_400:=}');
Expect(1, 69237, '\p{Is_Nv=+0000000400}', "");
Expect(0, 69237, '\p{^Is_Nv=+0000000400}', "");
Expect(0, 69237, '\P{Is_Nv=+0000000400}', "");
Expect(1, 69237, '\P{^Is_Nv=+0000000400}', "");
Expect(0, 69238, '\p{Is_Nv=+0000000400}', "");
Expect(1, 69238, '\p{^Is_Nv=+0000000400}', "");
Expect(1, 69238, '\P{Is_Nv=+0000000400}', "");
Expect(0, 69238, '\P{^Is_Nv=+0000000400}', "");
Error('\p{Numeric_Value=_:=00000005/0000000008}');
Error('\P{Numeric_Value=_:=00000005/0000000008}');
Expect(1, 8541, '\p{Numeric_Value=0005/008}', "");
Expect(0, 8541, '\p{^Numeric_Value=0005/008}', "");
Expect(0, 8541, '\P{Numeric_Value=0005/008}', "");
Expect(1, 8541, '\P{^Numeric_Value=0005/008}', "");
Expect(0, 8542, '\p{Numeric_Value=0005/008}', "");
Expect(1, 8542, '\p{^Numeric_Value=0005/008}', "");
Expect(1, 8542, '\P{Numeric_Value=0005/008}', "");
Expect(0, 8542, '\P{^Numeric_Value=0005/008}', "");
Error('\p{Numeric_Value=0.62}');
Error('\P{Numeric_Value=0.62}');
Expect(1, 8541, '\p{Numeric_Value=0.625}', "");
Expect(0, 8541, '\p{^Numeric_Value=0.625}', "");
Expect(0, 8541, '\P{Numeric_Value=0.625}', "");
Expect(1, 8541, '\P{^Numeric_Value=0.625}', "");
Expect(0, 8542, '\p{Numeric_Value=0.625}', "");
Expect(1, 8542, '\p{^Numeric_Value=0.625}', "");
Expect(1, 8542, '\P{Numeric_Value=0.625}', "");
Expect(0, 8542, '\P{^Numeric_Value=0.625}', "");
Error('\p{Nv:   -:=05/0000000008}');
Error('\P{Nv:   -:=05/0000000008}');
Expect(1, 8541, '\p{Nv=+0000005/008}', "");
Expect(0, 8541, '\p{^Nv=+0000005/008}', "");
Expect(0, 8541, '\P{Nv=+0000005/008}', "");
Expect(1, 8541, '\P{^Nv=+0000005/008}', "");
Expect(0, 8542, '\p{Nv=+0000005/008}', "");
Expect(1, 8542, '\p{^Nv=+0000005/008}', "");
Expect(1, 8542, '\P{Nv=+0000005/008}', "");
Expect(0, 8542, '\P{^Nv=+0000005/008}', "");
Error('\p{Nv=0.62}');
Error('\P{Nv=0.62}');
Expect(1, 8541, '\p{Nv=0.625}', "");
Expect(0, 8541, '\p{^Nv=0.625}', "");
Expect(0, 8541, '\P{Nv=0.625}', "");
Expect(1, 8541, '\P{^Nv=0.625}', "");
Expect(0, 8542, '\p{Nv=0.625}', "");
Expect(1, 8542, '\p{^Nv=0.625}', "");
Expect(1, 8542, '\P{Nv=0.625}', "");
Expect(0, 8542, '\P{^Nv=0.625}', "");
Error('\p{Is_Numeric_Value=/a/ 	00005/00008}');
Error('\P{Is_Numeric_Value=/a/ 	00005/00008}');
Expect(1, 8541, '\p{Is_Numeric_Value=000005/0000000008}', "");
Expect(0, 8541, '\p{^Is_Numeric_Value=000005/0000000008}', "");
Expect(0, 8541, '\P{Is_Numeric_Value=000005/0000000008}', "");
Expect(1, 8541, '\P{^Is_Numeric_Value=000005/0000000008}', "");
Expect(0, 8542, '\p{Is_Numeric_Value=000005/0000000008}', "");
Expect(1, 8542, '\p{^Is_Numeric_Value=000005/0000000008}', "");
Expect(1, 8542, '\P{Is_Numeric_Value=000005/0000000008}', "");
Expect(0, 8542, '\P{^Is_Numeric_Value=000005/0000000008}', "");
Error('\p{Is_Numeric_Value=0.62}');
Error('\P{Is_Numeric_Value=0.62}');
Expect(1, 8541, '\p{Is_Numeric_Value=0.625}', "");
Expect(0, 8541, '\p{^Is_Numeric_Value=0.625}', "");
Expect(0, 8541, '\P{Is_Numeric_Value=0.625}', "");
Expect(1, 8541, '\P{^Is_Numeric_Value=0.625}', "");
Expect(0, 8542, '\p{Is_Numeric_Value=0.625}', "");
Expect(1, 8542, '\p{^Is_Numeric_Value=0.625}', "");
Expect(1, 8542, '\P{Is_Numeric_Value=0.625}', "");
Expect(0, 8542, '\P{^Is_Numeric_Value=0.625}', "");
Error('\p{Is_Nv=/a/		+0000005/000008}');
Error('\P{Is_Nv=/a/		+0000005/000008}');
Expect(1, 8541, '\p{Is_Nv=000000005/00008}', "");
Expect(0, 8541, '\p{^Is_Nv=000000005/00008}', "");
Expect(0, 8541, '\P{Is_Nv=000000005/00008}', "");
Expect(1, 8541, '\P{^Is_Nv=000000005/00008}', "");
Expect(0, 8542, '\p{Is_Nv=000000005/00008}', "");
Expect(1, 8542, '\p{^Is_Nv=000000005/00008}', "");
Expect(1, 8542, '\P{Is_Nv=000000005/00008}', "");
Expect(0, 8542, '\P{^Is_Nv=000000005/00008}', "");
Error('\p{Is_Nv=0.62}');
Error('\P{Is_Nv=0.62}');
Expect(1, 8541, '\p{Is_Nv=0.625}', "");
Expect(0, 8541, '\p{^Is_Nv=0.625}', "");
Expect(0, 8541, '\P{Is_Nv=0.625}', "");
Expect(1, 8541, '\P{^Is_Nv=0.625}', "");
Expect(0, 8542, '\p{Is_Nv=0.625}', "");
Expect(1, 8542, '\p{^Is_Nv=0.625}', "");
Expect(1, 8542, '\P{Is_Nv=0.625}', "");
Expect(0, 8542, '\P{^Is_Nv=0.625}', "");
Error('\p{Numeric_Value: /a/+0000000020}');
Error('\P{Numeric_Value: /a/+0000000020}');
Expect(1, 119658, '\p{Numeric_Value=0000_0002_0}', "");
Expect(0, 119658, '\p{^Numeric_Value=0000_0002_0}', "");
Expect(0, 119658, '\P{Numeric_Value=0000_0002_0}', "");
Expect(1, 119658, '\P{^Numeric_Value=0000_0002_0}', "");
Expect(0, 119659, '\p{Numeric_Value=0000_0002_0}', "");
Expect(1, 119659, '\p{^Numeric_Value=0000_0002_0}', "");
Expect(1, 119659, '\P{Numeric_Value=0000_0002_0}', "");
Expect(0, 119659, '\P{^Numeric_Value=0000_0002_0}', "");
Error('\p{Nv= _00000020:=}');
Error('\P{Nv= _00000020:=}');
Expect(1, 119658, '\p{Nv=+000020}', "");
Expect(0, 119658, '\p{^Nv=+000020}', "");
Expect(0, 119658, '\P{Nv=+000020}', "");
Expect(1, 119658, '\P{^Nv=+000020}', "");
Expect(0, 119659, '\p{Nv=+000020}', "");
Expect(1, 119659, '\p{^Nv=+000020}', "");
Expect(1, 119659, '\P{Nv=+000020}', "");
Expect(0, 119659, '\P{^Nv=+000020}', "");
Error('\p{Is_Numeric_Value: :=-+00020}');
Error('\P{Is_Numeric_Value: :=-+00020}');
Expect(1, 119658, '\p{Is_Numeric_Value=+0_0_20}', "");
Expect(0, 119658, '\p{^Is_Numeric_Value=+0_0_20}', "");
Expect(0, 119658, '\P{Is_Numeric_Value=+0_0_20}', "");
Expect(1, 119658, '\P{^Is_Numeric_Value=+0_0_20}', "");
Expect(0, 119659, '\p{Is_Numeric_Value=+0_0_20}', "");
Expect(1, 119659, '\p{^Is_Numeric_Value=+0_0_20}', "");
Expect(1, 119659, '\P{Is_Numeric_Value=+0_0_20}', "");
Expect(0, 119659, '\P{^Is_Numeric_Value=+0_0_20}', "");
Error('\p{Is_Nv: 	0_0_0_0_0_0_0_020/a/}');
Error('\P{Is_Nv: 	0_0_0_0_0_0_0_020/a/}');
Expect(1, 119658, '\p{Is_Nv=0_0_0_0_0_0_020}', "");
Expect(0, 119658, '\p{^Is_Nv=0_0_0_0_0_0_020}', "");
Expect(0, 119658, '\P{Is_Nv=0_0_0_0_0_0_020}', "");
Expect(1, 119658, '\P{^Is_Nv=0_0_0_0_0_0_020}', "");
Expect(0, 119659, '\p{Is_Nv=0_0_0_0_0_0_020}', "");
Expect(1, 119659, '\p{^Is_Nv=0_0_0_0_0_0_020}', "");
Expect(1, 119659, '\P{Is_Nv=0_0_0_0_0_0_020}', "");
Expect(0, 119659, '\P{^Is_Nv=0_0_0_0_0_0_020}', "");
Error('\p{Numeric_Value=:=	-+0_0_0_3_0_000}');
Error('\P{Numeric_Value=:=	-+0_0_0_3_0_000}');
Expect(1, 65837, '\p{Numeric_Value:+000000030000}', "");
Expect(0, 65837, '\p{^Numeric_Value:+000000030000}', "");
Expect(0, 65837, '\P{Numeric_Value:+000000030000}', "");
Expect(1, 65837, '\P{^Numeric_Value:+000000030000}', "");
Expect(0, 65838, '\p{Numeric_Value:+000000030000}', "");
Expect(1, 65838, '\p{^Numeric_Value:+000000030000}', "");
Expect(1, 65838, '\P{Numeric_Value:+000000030000}', "");
Expect(0, 65838, '\P{^Numeric_Value:+000000030000}', "");
Error('\p{Nv=/a/0_3_0_0_00}');
Error('\P{Nv=/a/0_3_0_0_00}');
Expect(1, 65837, '\p{Nv=00030000}', "");
Expect(0, 65837, '\p{^Nv=00030000}', "");
Expect(0, 65837, '\P{Nv=00030000}', "");
Expect(1, 65837, '\P{^Nv=00030000}', "");
Expect(0, 65838, '\p{Nv=00030000}', "");
Expect(1, 65838, '\p{^Nv=00030000}', "");
Expect(1, 65838, '\P{Nv=00030000}', "");
Expect(0, 65838, '\P{^Nv=00030000}', "");
Error('\p{Is_Numeric_Value=/a/030000}');
Error('\P{Is_Numeric_Value=/a/030000}');
Expect(1, 65837, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_30000}', "");
Expect(0, 65837, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_30000}', "");
Expect(0, 65837, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_30000}', "");
Expect(1, 65837, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_30000}', "");
Expect(0, 65838, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_30000}', "");
Expect(1, 65838, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_30000}', "");
Expect(1, 65838, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_30000}', "");
Expect(0, 65838, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_30000}', "");
Error('\p{Is_Nv=:=	-0000_3000_0}');
Error('\P{Is_Nv=:=	-0000_3000_0}');
Expect(1, 65837, '\p{Is_Nv=00030000}', "");
Expect(0, 65837, '\p{^Is_Nv=00030000}', "");
Expect(0, 65837, '\P{Is_Nv=00030000}', "");
Expect(1, 65837, '\P{^Is_Nv=00030000}', "");
Expect(0, 65838, '\p{Is_Nv=00030000}', "");
Expect(1, 65838, '\p{^Is_Nv=00030000}', "");
Expect(1, 65838, '\P{Is_Nv=00030000}', "");
Expect(0, 65838, '\P{^Is_Nv=00030000}', "");
Error('\p{Numeric_Value= +000010/a/}');
Error('\P{Numeric_Value= +000010/a/}');
Expect(1, 119657, '\p{Numeric_Value=000_001_0}', "");
Expect(0, 119657, '\p{^Numeric_Value=000_001_0}', "");
Expect(0, 119657, '\P{Numeric_Value=000_001_0}', "");
Expect(1, 119657, '\P{^Numeric_Value=000_001_0}', "");
Expect(0, 119658, '\p{Numeric_Value=000_001_0}', "");
Expect(1, 119658, '\p{^Numeric_Value=000_001_0}', "");
Expect(1, 119658, '\P{Numeric_Value=000_001_0}', "");
Expect(0, 119658, '\P{^Numeric_Value=000_001_0}', "");
Error('\p{Nv=0_0_10:=}');
Error('\P{Nv=0_0_10:=}');
Expect(1, 119657, '\p{Nv=+0_0_0_0_0_0_010}', "");
Expect(0, 119657, '\p{^Nv=+0_0_0_0_0_0_010}', "");
Expect(0, 119657, '\P{Nv=+0_0_0_0_0_0_010}', "");
Expect(1, 119657, '\P{^Nv=+0_0_0_0_0_0_010}', "");
Expect(0, 119658, '\p{Nv=+0_0_0_0_0_0_010}', "");
Expect(1, 119658, '\p{^Nv=+0_0_0_0_0_0_010}', "");
Expect(1, 119658, '\P{Nv=+0_0_0_0_0_0_010}', "");
Expect(0, 119658, '\P{^Nv=+0_0_0_0_0_0_010}', "");
Error('\p{Is_Numeric_Value=/a/	-000000001_0}');
Error('\P{Is_Numeric_Value=/a/	-000000001_0}');
Expect(1, 119657, '\p{Is_Numeric_Value=+1_0}', "");
Expect(0, 119657, '\p{^Is_Numeric_Value=+1_0}', "");
Expect(0, 119657, '\P{Is_Numeric_Value=+1_0}', "");
Expect(1, 119657, '\P{^Is_Numeric_Value=+1_0}', "");
Expect(0, 119658, '\p{Is_Numeric_Value=+1_0}', "");
Expect(1, 119658, '\p{^Is_Numeric_Value=+1_0}', "");
Expect(1, 119658, '\P{Is_Numeric_Value=+1_0}', "");
Expect(0, 119658, '\P{^Is_Numeric_Value=+1_0}', "");
Error('\p{Is_Nv= /a/0_0_0_0_010}');
Error('\P{Is_Nv= /a/0_0_0_0_010}');
Expect(1, 119657, '\p{Is_Nv:+0_0_10}', "");
Expect(0, 119657, '\p{^Is_Nv:+0_0_10}', "");
Expect(0, 119657, '\P{Is_Nv:+0_0_10}', "");
Expect(1, 119657, '\P{^Is_Nv:+0_0_10}', "");
Expect(0, 119658, '\p{Is_Nv:+0_0_10}', "");
Expect(1, 119658, '\p{^Is_Nv:+0_0_10}', "");
Expect(1, 119658, '\P{Is_Nv:+0_0_10}', "");
Expect(0, 119658, '\P{^Is_Nv:+0_0_10}', "");
Error('\p{Numeric_Value=:=+0000_0000_31}');
Error('\P{Numeric_Value=:=+0000_0000_31}');
Expect(1, 12891, '\p{Numeric_Value: 0000000031}', "");
Expect(0, 12891, '\p{^Numeric_Value: 0000000031}', "");
Expect(0, 12891, '\P{Numeric_Value: 0000000031}', "");
Expect(1, 12891, '\P{^Numeric_Value: 0000000031}', "");
Expect(0, 12892, '\p{Numeric_Value: 0000000031}', "");
Expect(1, 12892, '\p{^Numeric_Value: 0000000031}', "");
Expect(1, 12892, '\P{Numeric_Value: 0000000031}', "");
Expect(0, 12892, '\P{^Numeric_Value: 0000000031}', "");
Error('\p{Nv: _+0_0_31/a/}');
Error('\P{Nv: _+0_0_31/a/}');
Expect(1, 12891, '\p{Nv=0_0_0_0_31}', "");
Expect(0, 12891, '\p{^Nv=0_0_0_0_31}', "");
Expect(0, 12891, '\P{Nv=0_0_0_0_31}', "");
Expect(1, 12891, '\P{^Nv=0_0_0_0_31}', "");
Expect(0, 12892, '\p{Nv=0_0_0_0_31}', "");
Expect(1, 12892, '\p{^Nv=0_0_0_0_31}', "");
Expect(1, 12892, '\P{Nv=0_0_0_0_31}', "");
Expect(0, 12892, '\P{^Nv=0_0_0_0_31}', "");
Error('\p{Is_Numeric_Value= :=0_0_0_31}');
Error('\P{Is_Numeric_Value= :=0_0_0_31}');
Expect(1, 12891, '\p{Is_Numeric_Value=000003_1}', "");
Expect(0, 12891, '\p{^Is_Numeric_Value=000003_1}', "");
Expect(0, 12891, '\P{Is_Numeric_Value=000003_1}', "");
Expect(1, 12891, '\P{^Is_Numeric_Value=000003_1}', "");
Expect(0, 12892, '\p{Is_Numeric_Value=000003_1}', "");
Expect(1, 12892, '\p{^Is_Numeric_Value=000003_1}', "");
Expect(1, 12892, '\P{Is_Numeric_Value=000003_1}', "");
Expect(0, 12892, '\P{^Is_Numeric_Value=000003_1}', "");
Error('\p{Is_Nv=	/a/+000000031}');
Error('\P{Is_Nv=	/a/+000000031}');
Expect(1, 12891, '\p{Is_Nv=000000031}', "");
Expect(0, 12891, '\p{^Is_Nv=000000031}', "");
Expect(0, 12891, '\P{Is_Nv=000000031}', "");
Expect(1, 12891, '\P{^Is_Nv=000000031}', "");
Expect(0, 12892, '\p{Is_Nv=000000031}', "");
Expect(1, 12892, '\p{^Is_Nv=000000031}', "");
Expect(1, 12892, '\P{Is_Nv=000000031}', "");
Expect(0, 12892, '\P{^Is_Nv=000000031}', "");
Error('\p{Numeric_Value:	_ 0000008000/a/}');
Error('\P{Numeric_Value:	_ 0000008000/a/}');
Expect(1, 65833, '\p{Numeric_Value:   000000008000}', "");
Expect(0, 65833, '\p{^Numeric_Value:   000000008000}', "");
Expect(0, 65833, '\P{Numeric_Value:   000000008000}', "");
Expect(1, 65833, '\P{^Numeric_Value:   000000008000}', "");
Expect(0, 65834, '\p{Numeric_Value:   000000008000}', "");
Expect(1, 65834, '\p{^Numeric_Value:   000000008000}', "");
Expect(1, 65834, '\P{Numeric_Value:   000000008000}', "");
Expect(0, 65834, '\P{^Numeric_Value:   000000008000}', "");
Error('\p{Nv= 0800_0:=}');
Error('\P{Nv= 0800_0:=}');
Expect(1, 65833, '\p{Nv=0000000008000}', "");
Expect(0, 65833, '\p{^Nv=0000000008000}', "");
Expect(0, 65833, '\P{Nv=0000000008000}', "");
Expect(1, 65833, '\P{^Nv=0000000008000}', "");
Expect(0, 65834, '\p{Nv=0000000008000}', "");
Expect(1, 65834, '\p{^Nv=0000000008000}', "");
Expect(1, 65834, '\P{Nv=0000000008000}', "");
Expect(0, 65834, '\P{^Nv=0000000008000}', "");
Error('\p{Is_Numeric_Value=:= _000000008000}');
Error('\P{Is_Numeric_Value=:= _000000008000}');
Expect(1, 65833, '\p{Is_Numeric_Value=+008000}', "");
Expect(0, 65833, '\p{^Is_Numeric_Value=+008000}', "");
Expect(0, 65833, '\P{Is_Numeric_Value=+008000}', "");
Expect(1, 65833, '\P{^Is_Numeric_Value=+008000}', "");
Expect(0, 65834, '\p{Is_Numeric_Value=+008000}', "");
Expect(1, 65834, '\p{^Is_Numeric_Value=+008000}', "");
Expect(1, 65834, '\P{Is_Numeric_Value=+008000}', "");
Expect(0, 65834, '\P{^Is_Numeric_Value=+008000}', "");
Error('\p{Is_Nv= /a/8000}');
Error('\P{Is_Nv= /a/8000}');
Expect(1, 65833, '\p{Is_Nv:8000}', "");
Expect(0, 65833, '\p{^Is_Nv:8000}', "");
Expect(0, 65833, '\P{Is_Nv:8000}', "");
Expect(1, 65833, '\P{^Is_Nv:8000}', "");
Expect(0, 65834, '\p{Is_Nv:8000}', "");
Expect(1, 65834, '\p{^Is_Nv:8000}', "");
Expect(1, 65834, '\P{Is_Nv:8000}', "");
Expect(0, 65834, '\P{^Is_Nv:8000}', "");
Error('\p{Numeric_Value=:=000000001000000000000}');
Error('\P{Numeric_Value=:=000000001000000000000}');
Expect(1, 20806, '\p{Numeric_Value=0_0_0_0_0_0_1_0_0_0_000000000}', "");
Expect(0, 20806, '\p{^Numeric_Value=0_0_0_0_0_0_1_0_0_0_000000000}', "");
Expect(0, 20806, '\P{Numeric_Value=0_0_0_0_0_0_1_0_0_0_000000000}', "");
Expect(1, 20806, '\P{^Numeric_Value=0_0_0_0_0_0_1_0_0_0_000000000}', "");
Expect(0, 20807, '\p{Numeric_Value=0_0_0_0_0_0_1_0_0_0_000000000}', "");
Expect(1, 20807, '\p{^Numeric_Value=0_0_0_0_0_0_1_0_0_0_000000000}', "");
Expect(1, 20807, '\P{Numeric_Value=0_0_0_0_0_0_1_0_0_0_000000000}', "");
Expect(0, 20807, '\P{^Numeric_Value=0_0_0_0_0_0_1_0_0_0_000000000}', "");
Error('\p{Nv=:=0_0_0_0_0_1_0_0_0_0_0_0_0_0_0_0_00}');
Error('\P{Nv=:=0_0_0_0_0_1_0_0_0_0_0_0_0_0_0_0_00}');
Expect(1, 20806, '\p{Nv=00001000000000000}', "");
Expect(0, 20806, '\p{^Nv=00001000000000000}', "");
Expect(0, 20806, '\P{Nv=00001000000000000}', "");
Expect(1, 20806, '\P{^Nv=00001000000000000}', "");
Expect(0, 20807, '\p{Nv=00001000000000000}', "");
Expect(1, 20807, '\p{^Nv=00001000000000000}', "");
Expect(1, 20807, '\P{Nv=00001000000000000}', "");
Expect(0, 20807, '\P{^Nv=00001000000000000}', "");
Error('\p{Is_Numeric_Value= 0000001000000000000:=}');
Error('\P{Is_Numeric_Value= 0000001000000000000:=}');
Expect(1, 20806, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(0, 20806, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(0, 20806, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(1, 20806, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(0, 20807, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(1, 20807, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(1, 20807, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(0, 20807, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Error('\p{Is_Nv=	 00_00_00_00_10_00_00_0000000/a/}');
Error('\P{Is_Nv=	 00_00_00_00_10_00_00_0000000/a/}');
Expect(1, 20806, '\p{Is_Nv=0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(0, 20806, '\p{^Is_Nv=0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(0, 20806, '\P{Is_Nv=0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(1, 20806, '\P{^Is_Nv=0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(0, 20807, '\p{Is_Nv=0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(1, 20807, '\p{^Is_Nv=0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(1, 20807, '\P{Is_Nv=0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Expect(0, 20807, '\P{^Is_Nv=0_0_0_0_0_0_0_0_1_0_0_0_0_00000000}', "");
Error('\p{Numeric_Value=/a/_ +0_0_0_0_0_035}');
Error('\P{Numeric_Value=/a/_ +0_0_0_0_0_035}');
Expect(1, 12895, '\p{Numeric_Value:   00000035}', "");
Expect(0, 12895, '\p{^Numeric_Value:   00000035}', "");
Expect(0, 12895, '\P{Numeric_Value:   00000035}', "");
Expect(1, 12895, '\P{^Numeric_Value:   00000035}', "");
Expect(0, 12896, '\p{Numeric_Value:   00000035}', "");
Expect(1, 12896, '\p{^Numeric_Value:   00000035}', "");
Expect(1, 12896, '\P{Numeric_Value:   00000035}', "");
Expect(0, 12896, '\P{^Numeric_Value:   00000035}', "");
Error('\p{Nv=/a/+0_0_0_035}');
Error('\P{Nv=/a/+0_0_0_035}');
Expect(1, 12895, '\p{Nv=+00000000035}', "");
Expect(0, 12895, '\p{^Nv=+00000000035}', "");
Expect(0, 12895, '\P{Nv=+00000000035}', "");
Expect(1, 12895, '\P{^Nv=+00000000035}', "");
Expect(0, 12896, '\p{Nv=+00000000035}', "");
Expect(1, 12896, '\p{^Nv=+00000000035}', "");
Expect(1, 12896, '\P{Nv=+00000000035}', "");
Expect(0, 12896, '\P{^Nv=+00000000035}', "");
Error('\p{Is_Numeric_Value=_00_00_00_00_035:=}');
Error('\P{Is_Numeric_Value=_00_00_00_00_035:=}');
Expect(1, 12895, '\p{Is_Numeric_Value=0_0_0_0_0_35}', "");
Expect(0, 12895, '\p{^Is_Numeric_Value=0_0_0_0_0_35}', "");
Expect(0, 12895, '\P{Is_Numeric_Value=0_0_0_0_0_35}', "");
Expect(1, 12895, '\P{^Is_Numeric_Value=0_0_0_0_0_35}', "");
Expect(0, 12896, '\p{Is_Numeric_Value=0_0_0_0_0_35}', "");
Expect(1, 12896, '\p{^Is_Numeric_Value=0_0_0_0_0_35}', "");
Expect(1, 12896, '\P{Is_Numeric_Value=0_0_0_0_0_35}', "");
Expect(0, 12896, '\P{^Is_Numeric_Value=0_0_0_0_0_35}', "");
Error('\p{Is_Nv=  +003_5:=}');
Error('\P{Is_Nv=  +003_5:=}');
Expect(1, 12895, '\p{Is_Nv:   0000035}', "");
Expect(0, 12895, '\p{^Is_Nv:   0000035}', "");
Expect(0, 12895, '\P{Is_Nv:   0000035}', "");
Expect(1, 12895, '\P{^Is_Nv:   0000035}', "");
Expect(0, 12896, '\p{Is_Nv:   0000035}', "");
Expect(1, 12896, '\p{^Is_Nv:   0000035}', "");
Expect(1, 12896, '\P{Is_Nv:   0000035}', "");
Expect(0, 12896, '\P{^Is_Nv:   0000035}', "");
Error('\p{Numeric_Value=_:=00000011}');
Error('\P{Numeric_Value=_:=00000011}');
Expect(1, 9451, '\p{Numeric_Value=11}', "");
Expect(0, 9451, '\p{^Numeric_Value=11}', "");
Expect(0, 9451, '\P{Numeric_Value=11}', "");
Expect(1, 9451, '\P{^Numeric_Value=11}', "");
Expect(0, 9452, '\p{Numeric_Value=11}', "");
Expect(1, 9452, '\p{^Numeric_Value=11}', "");
Expect(1, 9452, '\P{Numeric_Value=11}', "");
Expect(0, 9452, '\P{^Numeric_Value=11}', "");
Error('\p{Nv=-/a/+011}');
Error('\P{Nv=-/a/+011}');
Expect(1, 9451, '\p{Nv=00_00_00_00_011}', "");
Expect(0, 9451, '\p{^Nv=00_00_00_00_011}', "");
Expect(0, 9451, '\P{Nv=00_00_00_00_011}', "");
Expect(1, 9451, '\P{^Nv=00_00_00_00_011}', "");
Expect(0, 9452, '\p{Nv=00_00_00_00_011}', "");
Expect(1, 9452, '\p{^Nv=00_00_00_00_011}', "");
Expect(1, 9452, '\P{Nv=00_00_00_00_011}', "");
Expect(0, 9452, '\P{^Nv=00_00_00_00_011}', "");
Error('\p{Is_Numeric_Value=/a/ +00_00_01_1}');
Error('\P{Is_Numeric_Value=/a/ +00_00_01_1}');
Expect(1, 9451, '\p{Is_Numeric_Value=0000011}', "");
Expect(0, 9451, '\p{^Is_Numeric_Value=0000011}', "");
Expect(0, 9451, '\P{Is_Numeric_Value=0000011}', "");
Expect(1, 9451, '\P{^Is_Numeric_Value=0000011}', "");
Expect(0, 9452, '\p{Is_Numeric_Value=0000011}', "");
Expect(1, 9452, '\p{^Is_Numeric_Value=0000011}', "");
Expect(1, 9452, '\P{Is_Numeric_Value=0000011}', "");
Expect(0, 9452, '\P{^Is_Numeric_Value=0000011}', "");
Error('\p{Is_Nv=:=-0001_1}');
Error('\P{Is_Nv=:=-0001_1}');
Expect(1, 9451, '\p{Is_Nv=0001_1}', "");
Expect(0, 9451, '\p{^Is_Nv=0001_1}', "");
Expect(0, 9451, '\P{Is_Nv=0001_1}', "");
Expect(1, 9451, '\P{^Is_Nv=0001_1}', "");
Expect(0, 9452, '\p{Is_Nv=0001_1}', "");
Expect(1, 9452, '\p{^Is_Nv=0001_1}', "");
Expect(1, 9452, '\P{Is_Nv=0001_1}', "");
Expect(0, 9452, '\P{^Is_Nv=0001_1}', "");
Error('\p{Numeric_Value=-/a/1/00000005}');
Error('\P{Numeric_Value=-/a/1/00000005}');
Expect(1, 8533, '\p{Numeric_Value: +00000001/00005}', "");
Expect(0, 8533, '\p{^Numeric_Value: +00000001/00005}', "");
Expect(0, 8533, '\P{Numeric_Value: +00000001/00005}', "");
Expect(1, 8533, '\P{^Numeric_Value: +00000001/00005}', "");
Expect(0, 8534, '\p{Numeric_Value: +00000001/00005}', "");
Expect(1, 8534, '\p{^Numeric_Value: +00000001/00005}', "");
Expect(1, 8534, '\P{Numeric_Value: +00000001/00005}', "");
Expect(0, 8534, '\P{^Numeric_Value: +00000001/00005}', "");
Expect(1, 8533, '\p{Numeric_Value=0.200}', "");
Expect(0, 8533, '\p{^Numeric_Value=0.200}', "");
Expect(0, 8533, '\P{Numeric_Value=0.200}', "");
Expect(1, 8533, '\P{^Numeric_Value=0.200}', "");
Expect(0, 8534, '\p{Numeric_Value=0.200}', "");
Expect(1, 8534, '\p{^Numeric_Value=0.200}', "");
Expect(1, 8534, '\P{Numeric_Value=0.200}', "");
Expect(0, 8534, '\P{^Numeric_Value=0.200}', "");
Error('\p{Nv=	 +0000001/000005/a/}');
Error('\P{Nv=	 +0000001/000005/a/}');
Expect(1, 8533, '\p{Nv=+001/005}', "");
Expect(0, 8533, '\p{^Nv=+001/005}', "");
Expect(0, 8533, '\P{Nv=+001/005}', "");
Expect(1, 8533, '\P{^Nv=+001/005}', "");
Expect(0, 8534, '\p{Nv=+001/005}', "");
Expect(1, 8534, '\p{^Nv=+001/005}', "");
Expect(1, 8534, '\P{Nv=+001/005}', "");
Expect(0, 8534, '\P{^Nv=+001/005}', "");
Expect(1, 8533, '\p{Nv=0.200}', "");
Expect(0, 8533, '\p{^Nv=0.200}', "");
Expect(0, 8533, '\P{Nv=0.200}', "");
Expect(1, 8533, '\P{^Nv=0.200}', "");
Expect(0, 8534, '\p{Nv=0.200}', "");
Expect(1, 8534, '\p{^Nv=0.200}', "");
Expect(1, 8534, '\P{Nv=0.200}', "");
Expect(0, 8534, '\P{^Nv=0.200}', "");
Error('\p{Is_Numeric_Value=/a/-+0000000001/000005}');
Error('\P{Is_Numeric_Value=/a/-+0000000001/000005}');
Expect(1, 8533, '\p{Is_Numeric_Value=+01/00000005}', "");
Expect(0, 8533, '\p{^Is_Numeric_Value=+01/00000005}', "");
Expect(0, 8533, '\P{Is_Numeric_Value=+01/00000005}', "");
Expect(1, 8533, '\P{^Is_Numeric_Value=+01/00000005}', "");
Expect(0, 8534, '\p{Is_Numeric_Value=+01/00000005}', "");
Expect(1, 8534, '\p{^Is_Numeric_Value=+01/00000005}', "");
Expect(1, 8534, '\P{Is_Numeric_Value=+01/00000005}', "");
Expect(0, 8534, '\P{^Is_Numeric_Value=+01/00000005}', "");
Expect(1, 8533, '\p{Is_Numeric_Value: 0.200}', "");
Expect(0, 8533, '\p{^Is_Numeric_Value: 0.200}', "");
Expect(0, 8533, '\P{Is_Numeric_Value: 0.200}', "");
Expect(1, 8533, '\P{^Is_Numeric_Value: 0.200}', "");
Expect(0, 8534, '\p{Is_Numeric_Value: 0.200}', "");
Expect(1, 8534, '\p{^Is_Numeric_Value: 0.200}', "");
Expect(1, 8534, '\P{Is_Numeric_Value: 0.200}', "");
Expect(0, 8534, '\P{^Is_Numeric_Value: 0.200}', "");
Error('\p{Is_Nv=	/a/000001/000005}');
Error('\P{Is_Nv=	/a/000001/000005}');
Expect(1, 8533, '\p{Is_Nv=+000000001/00000005}', "");
Expect(0, 8533, '\p{^Is_Nv=+000000001/00000005}', "");
Expect(0, 8533, '\P{Is_Nv=+000000001/00000005}', "");
Expect(1, 8533, '\P{^Is_Nv=+000000001/00000005}', "");
Expect(0, 8534, '\p{Is_Nv=+000000001/00000005}', "");
Expect(1, 8534, '\p{^Is_Nv=+000000001/00000005}', "");
Expect(1, 8534, '\P{Is_Nv=+000000001/00000005}', "");
Expect(0, 8534, '\P{^Is_Nv=+000000001/00000005}', "");
Expect(1, 8533, '\p{Is_Nv=0.200}', "");
Expect(0, 8533, '\p{^Is_Nv=0.200}', "");
Expect(0, 8533, '\P{Is_Nv=0.200}', "");
Expect(1, 8533, '\P{^Is_Nv=0.200}', "");
Expect(0, 8534, '\p{Is_Nv=0.200}', "");
Expect(1, 8534, '\p{^Is_Nv=0.200}', "");
Expect(1, 8534, '\P{Is_Nv=0.200}', "");
Expect(0, 8534, '\P{^Is_Nv=0.200}', "");
Error('\p{Numeric_Value=__NaN:=}');
Error('\P{Numeric_Value=__NaN:=}');
Expect(1, 194705, '\p{Numeric_Value=nan}', "");
Expect(0, 194705, '\p{^Numeric_Value=nan}', "");
Expect(0, 194705, '\P{Numeric_Value=nan}', "");
Expect(1, 194705, '\P{^Numeric_Value=nan}', "");
Expect(0, 194704, '\p{Numeric_Value=nan}', "");
Expect(1, 194704, '\p{^Numeric_Value=nan}', "");
Expect(1, 194704, '\P{Numeric_Value=nan}', "");
Expect(0, 194704, '\P{^Numeric_Value=nan}', "");
Expect(1, 194705, '\p{Numeric_Value=_NaN}', "");
Expect(0, 194705, '\p{^Numeric_Value=_NaN}', "");
Expect(0, 194705, '\P{Numeric_Value=_NaN}', "");
Expect(1, 194705, '\P{^Numeric_Value=_NaN}', "");
Expect(0, 194704, '\p{Numeric_Value=_NaN}', "");
Expect(1, 194704, '\p{^Numeric_Value=_NaN}', "");
Expect(1, 194704, '\P{Numeric_Value=_NaN}', "");
Expect(0, 194704, '\P{^Numeric_Value=_NaN}', "");
Error('\p{Nv=-:=NAN}');
Error('\P{Nv=-:=NAN}');
Expect(1, 194705, '\p{Nv=nan}', "");
Expect(0, 194705, '\p{^Nv=nan}', "");
Expect(0, 194705, '\P{Nv=nan}', "");
Expect(1, 194705, '\P{^Nv=nan}', "");
Expect(0, 194704, '\p{Nv=nan}', "");
Expect(1, 194704, '\p{^Nv=nan}', "");
Expect(1, 194704, '\P{Nv=nan}', "");
Expect(0, 194704, '\P{^Nv=nan}', "");
Expect(1, 194705, '\p{Nv=		NAN}', "");
Expect(0, 194705, '\p{^Nv=		NAN}', "");
Expect(0, 194705, '\P{Nv=		NAN}', "");
Expect(1, 194705, '\P{^Nv=		NAN}', "");
Expect(0, 194704, '\p{Nv=		NAN}', "");
Expect(1, 194704, '\p{^Nv=		NAN}', "");
Expect(1, 194704, '\P{Nv=		NAN}', "");
Expect(0, 194704, '\P{^Nv=		NAN}', "");
Error('\p{Is_Numeric_Value=/a/NaN}');
Error('\P{Is_Numeric_Value=/a/NaN}');
Expect(1, 194705, '\p{Is_Numeric_Value=nan}', "");
Expect(0, 194705, '\p{^Is_Numeric_Value=nan}', "");
Expect(0, 194705, '\P{Is_Numeric_Value=nan}', "");
Expect(1, 194705, '\P{^Is_Numeric_Value=nan}', "");
Expect(0, 194704, '\p{Is_Numeric_Value=nan}', "");
Expect(1, 194704, '\p{^Is_Numeric_Value=nan}', "");
Expect(1, 194704, '\P{Is_Numeric_Value=nan}', "");
Expect(0, 194704, '\P{^Is_Numeric_Value=nan}', "");
Expect(1, 194705, '\p{Is_Numeric_Value=-NaN}', "");
Expect(0, 194705, '\p{^Is_Numeric_Value=-NaN}', "");
Expect(0, 194705, '\P{Is_Numeric_Value=-NaN}', "");
Expect(1, 194705, '\P{^Is_Numeric_Value=-NaN}', "");
Expect(0, 194704, '\p{Is_Numeric_Value=-NaN}', "");
Expect(1, 194704, '\p{^Is_Numeric_Value=-NaN}', "");
Expect(1, 194704, '\P{Is_Numeric_Value=-NaN}', "");
Expect(0, 194704, '\P{^Is_Numeric_Value=-NaN}', "");
Error('\p{Is_Nv=-	nan/a/}');
Error('\P{Is_Nv=-	nan/a/}');
Expect(1, 194705, '\p{Is_Nv=nan}', "");
Expect(0, 194705, '\p{^Is_Nv=nan}', "");
Expect(0, 194705, '\P{Is_Nv=nan}', "");
Expect(1, 194705, '\P{^Is_Nv=nan}', "");
Expect(0, 194704, '\p{Is_Nv=nan}', "");
Expect(1, 194704, '\p{^Is_Nv=nan}', "");
Expect(1, 194704, '\P{Is_Nv=nan}', "");
Expect(0, 194704, '\P{^Is_Nv=nan}', "");
Expect(1, 194705, '\p{Is_Nv=	 nan}', "");
Expect(0, 194705, '\p{^Is_Nv=	 nan}', "");
Expect(0, 194705, '\P{Is_Nv=	 nan}', "");
Expect(1, 194705, '\P{^Is_Nv=	 nan}', "");
Expect(0, 194704, '\p{Is_Nv=	 nan}', "");
Expect(1, 194704, '\p{^Is_Nv=	 nan}', "");
Expect(1, 194704, '\P{Is_Nv=	 nan}', "");
Expect(0, 194704, '\P{^Is_Nv=	 nan}', "");
Error('\p{Numeric_Value=	/a/+00015/002}');
Error('\P{Numeric_Value=	/a/+00015/002}');
Expect(1, 3889, '\p{Numeric_Value=+015/000000002}', "");
Expect(0, 3889, '\p{^Numeric_Value=+015/000000002}', "");
Expect(0, 3889, '\P{Numeric_Value=+015/000000002}', "");
Expect(1, 3889, '\P{^Numeric_Value=+015/000000002}', "");
Expect(0, 3890, '\p{Numeric_Value=+015/000000002}', "");
Expect(1, 3890, '\p{^Numeric_Value=+015/000000002}', "");
Expect(1, 3890, '\P{Numeric_Value=+015/000000002}', "");
Expect(0, 3890, '\P{^Numeric_Value=+015/000000002}', "");
Expect(1, 3889, '\p{Numeric_Value=7.500}', "");
Expect(0, 3889, '\p{^Numeric_Value=7.500}', "");
Expect(0, 3889, '\P{Numeric_Value=7.500}', "");
Expect(1, 3889, '\P{^Numeric_Value=7.500}', "");
Expect(0, 3890, '\p{Numeric_Value=7.500}', "");
Expect(1, 3890, '\p{^Numeric_Value=7.500}', "");
Expect(1, 3890, '\P{Numeric_Value=7.500}', "");
Expect(0, 3890, '\P{^Numeric_Value=7.500}', "");
Error('\p{Nv=:=-	000000015/00002}');
Error('\P{Nv=:=-	000000015/00002}');
Expect(1, 3889, '\p{Nv=000000015/0002}', "");
Expect(0, 3889, '\p{^Nv=000000015/0002}', "");
Expect(0, 3889, '\P{Nv=000000015/0002}', "");
Expect(1, 3889, '\P{^Nv=000000015/0002}', "");
Expect(0, 3890, '\p{Nv=000000015/0002}', "");
Expect(1, 3890, '\p{^Nv=000000015/0002}', "");
Expect(1, 3890, '\P{Nv=000000015/0002}', "");
Expect(0, 3890, '\P{^Nv=000000015/0002}', "");
Expect(1, 3889, '\p{Nv=7.500}', "");
Expect(0, 3889, '\p{^Nv=7.500}', "");
Expect(0, 3889, '\P{Nv=7.500}', "");
Expect(1, 3889, '\P{^Nv=7.500}', "");
Expect(0, 3890, '\p{Nv=7.500}', "");
Expect(1, 3890, '\p{^Nv=7.500}', "");
Expect(1, 3890, '\P{Nv=7.500}', "");
Expect(0, 3890, '\P{^Nv=7.500}', "");
Error('\p{Is_Numeric_Value=:=+000000015/02}');
Error('\P{Is_Numeric_Value=:=+000000015/02}');
Expect(1, 3889, '\p{Is_Numeric_Value=0015/0000000002}', "");
Expect(0, 3889, '\p{^Is_Numeric_Value=0015/0000000002}', "");
Expect(0, 3889, '\P{Is_Numeric_Value=0015/0000000002}', "");
Expect(1, 3889, '\P{^Is_Numeric_Value=0015/0000000002}', "");
Expect(0, 3890, '\p{Is_Numeric_Value=0015/0000000002}', "");
Expect(1, 3890, '\p{^Is_Numeric_Value=0015/0000000002}', "");
Expect(1, 3890, '\P{Is_Numeric_Value=0015/0000000002}', "");
Expect(0, 3890, '\P{^Is_Numeric_Value=0015/0000000002}', "");
Expect(1, 3889, '\p{Is_Numeric_Value=7.500}', "");
Expect(0, 3889, '\p{^Is_Numeric_Value=7.500}', "");
Expect(0, 3889, '\P{Is_Numeric_Value=7.500}', "");
Expect(1, 3889, '\P{^Is_Numeric_Value=7.500}', "");
Expect(0, 3890, '\p{Is_Numeric_Value=7.500}', "");
Expect(1, 3890, '\p{^Is_Numeric_Value=7.500}', "");
Expect(1, 3890, '\P{Is_Numeric_Value=7.500}', "");
Expect(0, 3890, '\P{^Is_Numeric_Value=7.500}', "");
Error('\p{Is_Nv=:=		000000015/0000002}');
Error('\P{Is_Nv=:=		000000015/0000002}');
Expect(1, 3889, '\p{Is_Nv=00015/0002}', "");
Expect(0, 3889, '\p{^Is_Nv=00015/0002}', "");
Expect(0, 3889, '\P{Is_Nv=00015/0002}', "");
Expect(1, 3889, '\P{^Is_Nv=00015/0002}', "");
Expect(0, 3890, '\p{Is_Nv=00015/0002}', "");
Expect(1, 3890, '\p{^Is_Nv=00015/0002}', "");
Expect(1, 3890, '\P{Is_Nv=00015/0002}', "");
Expect(0, 3890, '\P{^Is_Nv=00015/0002}', "");
Expect(1, 3889, '\p{Is_Nv=7.500}', "");
Expect(0, 3889, '\p{^Is_Nv=7.500}', "");
Expect(0, 3889, '\P{Is_Nv=7.500}', "");
Expect(1, 3889, '\P{^Is_Nv=7.500}', "");
Expect(0, 3890, '\p{Is_Nv=7.500}', "");
Expect(1, 3890, '\p{^Is_Nv=7.500}', "");
Expect(1, 3890, '\P{Is_Nv=7.500}', "");
Expect(0, 3890, '\P{^Is_Nv=7.500}', "");
Error('\p{Numeric_Value=:=0000000007/000008}');
Error('\P{Numeric_Value=:=0000000007/000008}');
Expect(1, 8542, '\p{Numeric_Value=07/08}', "");
Expect(0, 8542, '\p{^Numeric_Value=07/08}', "");
Expect(0, 8542, '\P{Numeric_Value=07/08}', "");
Expect(1, 8542, '\P{^Numeric_Value=07/08}', "");
Expect(0, 8543, '\p{Numeric_Value=07/08}', "");
Expect(1, 8543, '\p{^Numeric_Value=07/08}', "");
Expect(1, 8543, '\P{Numeric_Value=07/08}', "");
Expect(0, 8543, '\P{^Numeric_Value=07/08}', "");
Error('\p{Numeric_Value: 0.9}');
Error('\P{Numeric_Value: 0.9}');
Error('\p{Numeric_Value=0.88}');
Error('\P{Numeric_Value=0.88}');
Expect(1, 8542, '\p{Numeric_Value: 0.875}', "");
Expect(0, 8542, '\p{^Numeric_Value: 0.875}', "");
Expect(0, 8542, '\P{Numeric_Value: 0.875}', "");
Expect(1, 8542, '\P{^Numeric_Value: 0.875}', "");
Expect(0, 8543, '\p{Numeric_Value: 0.875}', "");
Expect(1, 8543, '\p{^Numeric_Value: 0.875}', "");
Expect(1, 8543, '\P{Numeric_Value: 0.875}', "");
Expect(0, 8543, '\P{^Numeric_Value: 0.875}', "");
Error('\p{Nv= :=7/000000008}');
Error('\P{Nv= :=7/000000008}');
Expect(1, 8542, '\p{Nv=0007/00008}', "");
Expect(0, 8542, '\p{^Nv=0007/00008}', "");
Expect(0, 8542, '\P{Nv=0007/00008}', "");
Expect(1, 8542, '\P{^Nv=0007/00008}', "");
Expect(0, 8543, '\p{Nv=0007/00008}', "");
Expect(1, 8543, '\p{^Nv=0007/00008}', "");
Expect(1, 8543, '\P{Nv=0007/00008}', "");
Expect(0, 8543, '\P{^Nv=0007/00008}', "");
Error('\p{Nv=0.9}');
Error('\P{Nv=0.9}');
Error('\p{Nv=0.88}');
Error('\P{Nv=0.88}');
Expect(1, 8542, '\p{Nv=0.875}', "");
Expect(0, 8542, '\p{^Nv=0.875}', "");
Expect(0, 8542, '\P{Nv=0.875}', "");
Expect(1, 8542, '\P{^Nv=0.875}', "");
Expect(0, 8543, '\p{Nv=0.875}', "");
Expect(1, 8543, '\p{^Nv=0.875}', "");
Expect(1, 8543, '\P{Nv=0.875}', "");
Expect(0, 8543, '\P{^Nv=0.875}', "");
Error('\p{Is_Numeric_Value=-:=7/0000008}');
Error('\P{Is_Numeric_Value=-:=7/0000008}');
Expect(1, 8542, '\p{Is_Numeric_Value=0000007/0000008}', "");
Expect(0, 8542, '\p{^Is_Numeric_Value=0000007/0000008}', "");
Expect(0, 8542, '\P{Is_Numeric_Value=0000007/0000008}', "");
Expect(1, 8542, '\P{^Is_Numeric_Value=0000007/0000008}', "");
Expect(0, 8543, '\p{Is_Numeric_Value=0000007/0000008}', "");
Expect(1, 8543, '\p{^Is_Numeric_Value=0000007/0000008}', "");
Expect(1, 8543, '\P{Is_Numeric_Value=0000007/0000008}', "");
Expect(0, 8543, '\P{^Is_Numeric_Value=0000007/0000008}', "");
Error('\p{Is_Numeric_Value=0.9}');
Error('\P{Is_Numeric_Value=0.9}');
Error('\p{Is_Numeric_Value=0.88}');
Error('\P{Is_Numeric_Value=0.88}');
Expect(1, 8542, '\p{Is_Numeric_Value=0.875}', "");
Expect(0, 8542, '\p{^Is_Numeric_Value=0.875}', "");
Expect(0, 8542, '\P{Is_Numeric_Value=0.875}', "");
Expect(1, 8542, '\P{^Is_Numeric_Value=0.875}', "");
Expect(0, 8543, '\p{Is_Numeric_Value=0.875}', "");
Expect(1, 8543, '\p{^Is_Numeric_Value=0.875}', "");
Expect(1, 8543, '\P{Is_Numeric_Value=0.875}', "");
Expect(0, 8543, '\P{^Is_Numeric_Value=0.875}', "");
Error('\p{Is_Nv:	__000007/00008:=}');
Error('\P{Is_Nv:	__000007/00008:=}');
Expect(1, 8542, '\p{Is_Nv=+0000007/08}', "");
Expect(0, 8542, '\p{^Is_Nv=+0000007/08}', "");
Expect(0, 8542, '\P{Is_Nv=+0000007/08}', "");
Expect(1, 8542, '\P{^Is_Nv=+0000007/08}', "");
Expect(0, 8543, '\p{Is_Nv=+0000007/08}', "");
Expect(1, 8543, '\p{^Is_Nv=+0000007/08}', "");
Expect(1, 8543, '\P{Is_Nv=+0000007/08}', "");
Expect(0, 8543, '\P{^Is_Nv=+0000007/08}', "");
Error('\p{Is_Nv=0.9}');
Error('\P{Is_Nv=0.9}');
Error('\p{Is_Nv=0.88}');
Error('\P{Is_Nv=0.88}');
Expect(1, 8542, '\p{Is_Nv=0.875}', "");
Expect(0, 8542, '\p{^Is_Nv=0.875}', "");
Expect(0, 8542, '\P{Is_Nv=0.875}', "");
Expect(1, 8542, '\P{^Is_Nv=0.875}', "");
Expect(0, 8543, '\p{Is_Nv=0.875}', "");
Expect(1, 8543, '\p{^Is_Nv=0.875}', "");
Expect(1, 8543, '\P{Is_Nv=0.875}', "");
Expect(0, 8543, '\P{^Is_Nv=0.875}', "");
Error('\p{Numeric_Value=_:=+000700}');
Error('\P{Numeric_Value=_:=+000700}');
Expect(1, 69240, '\p{Numeric_Value=00700}', "");
Expect(0, 69240, '\p{^Numeric_Value=00700}', "");
Expect(0, 69240, '\P{Numeric_Value=00700}', "");
Expect(1, 69240, '\P{^Numeric_Value=00700}', "");
Expect(0, 69241, '\p{Numeric_Value=00700}', "");
Expect(1, 69241, '\p{^Numeric_Value=00700}', "");
Expect(1, 69241, '\P{Numeric_Value=00700}', "");
Expect(0, 69241, '\P{^Numeric_Value=00700}', "");
Error('\p{Nv=	 000700:=}');
Error('\P{Nv=	 000700:=}');
Expect(1, 69240, '\p{Nv=+0_0_0_0_0_7_00}', "");
Expect(0, 69240, '\p{^Nv=+0_0_0_0_0_7_00}', "");
Expect(0, 69240, '\P{Nv=+0_0_0_0_0_7_00}', "");
Expect(1, 69240, '\P{^Nv=+0_0_0_0_0_7_00}', "");
Expect(0, 69241, '\p{Nv=+0_0_0_0_0_7_00}', "");
Expect(1, 69241, '\p{^Nv=+0_0_0_0_0_7_00}', "");
Expect(1, 69241, '\P{Nv=+0_0_0_0_0_7_00}', "");
Expect(0, 69241, '\P{^Nv=+0_0_0_0_0_7_00}', "");
Error('\p{Is_Numeric_Value=- 0000000700/a/}');
Error('\P{Is_Numeric_Value=- 0000000700/a/}');
Expect(1, 69240, '\p{Is_Numeric_Value=0700}', "");
Expect(0, 69240, '\p{^Is_Numeric_Value=0700}', "");
Expect(0, 69240, '\P{Is_Numeric_Value=0700}', "");
Expect(1, 69240, '\P{^Is_Numeric_Value=0700}', "");
Expect(0, 69241, '\p{Is_Numeric_Value=0700}', "");
Expect(1, 69241, '\p{^Is_Numeric_Value=0700}', "");
Expect(1, 69241, '\P{Is_Numeric_Value=0700}', "");
Expect(0, 69241, '\P{^Is_Numeric_Value=0700}', "");
Error('\p{Is_Nv=_-0000000070_0:=}');
Error('\P{Is_Nv=_-0000000070_0:=}');
Expect(1, 69240, '\p{Is_Nv=00000000700}', "");
Expect(0, 69240, '\p{^Is_Nv=00000000700}', "");
Expect(0, 69240, '\P{Is_Nv=00000000700}', "");
Expect(1, 69240, '\P{^Is_Nv=00000000700}', "");
Expect(0, 69241, '\p{Is_Nv=00000000700}', "");
Expect(1, 69241, '\p{^Is_Nv=00000000700}', "");
Expect(1, 69241, '\P{Is_Nv=00000000700}', "");
Expect(0, 69241, '\P{^Is_Nv=00000000700}', "");
Error('\p{Numeric_Value:-:=00000013/0000000002}');
Error('\P{Numeric_Value:-:=00000013/0000000002}');
Expect(1, 3888, '\p{Numeric_Value=+000013/02}', "");
Expect(0, 3888, '\p{^Numeric_Value=+000013/02}', "");
Expect(0, 3888, '\P{Numeric_Value=+000013/02}', "");
Expect(1, 3888, '\P{^Numeric_Value=+000013/02}', "");
Expect(0, 3889, '\p{Numeric_Value=+000013/02}', "");
Expect(1, 3889, '\p{^Numeric_Value=+000013/02}', "");
Expect(1, 3889, '\P{Numeric_Value=+000013/02}', "");
Expect(0, 3889, '\P{^Numeric_Value=+000013/02}', "");
Expect(1, 3888, '\p{Numeric_Value=6.500}', "");
Expect(0, 3888, '\p{^Numeric_Value=6.500}', "");
Expect(0, 3888, '\P{Numeric_Value=6.500}', "");
Expect(1, 3888, '\P{^Numeric_Value=6.500}', "");
Expect(0, 3889, '\p{Numeric_Value=6.500}', "");
Expect(1, 3889, '\p{^Numeric_Value=6.500}', "");
Expect(1, 3889, '\P{Numeric_Value=6.500}', "");
Expect(0, 3889, '\P{^Numeric_Value=6.500}', "");
Error('\p{Nv:	/a/	013/02}');
Error('\P{Nv:	/a/	013/02}');
Expect(1, 3888, '\p{Nv=00013/00000002}', "");
Expect(0, 3888, '\p{^Nv=00013/00000002}', "");
Expect(0, 3888, '\P{Nv=00013/00000002}', "");
Expect(1, 3888, '\P{^Nv=00013/00000002}', "");
Expect(0, 3889, '\p{Nv=00013/00000002}', "");
Expect(1, 3889, '\p{^Nv=00013/00000002}', "");
Expect(1, 3889, '\P{Nv=00013/00000002}', "");
Expect(0, 3889, '\P{^Nv=00013/00000002}', "");
Expect(1, 3888, '\p{Nv: 6.500}', "");
Expect(0, 3888, '\p{^Nv: 6.500}', "");
Expect(0, 3888, '\P{Nv: 6.500}', "");
Expect(1, 3888, '\P{^Nv: 6.500}', "");
Expect(0, 3889, '\p{Nv: 6.500}', "");
Expect(1, 3889, '\p{^Nv: 6.500}', "");
Expect(1, 3889, '\P{Nv: 6.500}', "");
Expect(0, 3889, '\P{^Nv: 6.500}', "");
Error('\p{Is_Numeric_Value=:= _0000000013/2}');
Error('\P{Is_Numeric_Value=:= _0000000013/2}');
Expect(1, 3888, '\p{Is_Numeric_Value=0000013/002}', "");
Expect(0, 3888, '\p{^Is_Numeric_Value=0000013/002}', "");
Expect(0, 3888, '\P{Is_Numeric_Value=0000013/002}', "");
Expect(1, 3888, '\P{^Is_Numeric_Value=0000013/002}', "");
Expect(0, 3889, '\p{Is_Numeric_Value=0000013/002}', "");
Expect(1, 3889, '\p{^Is_Numeric_Value=0000013/002}', "");
Expect(1, 3889, '\P{Is_Numeric_Value=0000013/002}', "");
Expect(0, 3889, '\P{^Is_Numeric_Value=0000013/002}', "");
Expect(1, 3888, '\p{Is_Numeric_Value=6.500}', "");
Expect(0, 3888, '\p{^Is_Numeric_Value=6.500}', "");
Expect(0, 3888, '\P{Is_Numeric_Value=6.500}', "");
Expect(1, 3888, '\P{^Is_Numeric_Value=6.500}', "");
Expect(0, 3889, '\p{Is_Numeric_Value=6.500}', "");
Expect(1, 3889, '\p{^Is_Numeric_Value=6.500}', "");
Expect(1, 3889, '\P{Is_Numeric_Value=6.500}', "");
Expect(0, 3889, '\P{^Is_Numeric_Value=6.500}', "");
Error('\p{Is_Nv=-	+013/00002/a/}');
Error('\P{Is_Nv=-	+013/00002/a/}');
Expect(1, 3888, '\p{Is_Nv=13/000000002}', "");
Expect(0, 3888, '\p{^Is_Nv=13/000000002}', "");
Expect(0, 3888, '\P{Is_Nv=13/000000002}', "");
Expect(1, 3888, '\P{^Is_Nv=13/000000002}', "");
Expect(0, 3889, '\p{Is_Nv=13/000000002}', "");
Expect(1, 3889, '\p{^Is_Nv=13/000000002}', "");
Expect(1, 3889, '\P{Is_Nv=13/000000002}', "");
Expect(0, 3889, '\P{^Is_Nv=13/000000002}', "");
Expect(1, 3888, '\p{Is_Nv=6.500}', "");
Expect(0, 3888, '\p{^Is_Nv=6.500}', "");
Expect(0, 3888, '\P{Is_Nv=6.500}', "");
Expect(1, 3888, '\P{^Is_Nv=6.500}', "");
Expect(0, 3889, '\p{Is_Nv=6.500}', "");
Expect(1, 3889, '\p{^Is_Nv=6.500}', "");
Expect(1, 3889, '\P{Is_Nv=6.500}', "");
Expect(0, 3889, '\P{^Is_Nv=6.500}', "");
Error('\p{Numeric_Value=/a/+0_0_0_0_0_048}');
Error('\P{Numeric_Value=/a/+0_0_0_0_0_048}');
Expect(1, 12989, '\p{Numeric_Value=0_0_48}', "");
Expect(0, 12989, '\p{^Numeric_Value=0_0_48}', "");
Expect(0, 12989, '\P{Numeric_Value=0_0_48}', "");
Expect(1, 12989, '\P{^Numeric_Value=0_0_48}', "");
Expect(0, 12990, '\p{Numeric_Value=0_0_48}', "");
Expect(1, 12990, '\p{^Numeric_Value=0_0_48}', "");
Expect(1, 12990, '\P{Numeric_Value=0_0_48}', "");
Expect(0, 12990, '\P{^Numeric_Value=0_0_48}', "");
Error('\p{Nv:   :=	00_00_48}');
Error('\P{Nv:   :=	00_00_48}');
Expect(1, 12989, '\p{Nv=00048}', "");
Expect(0, 12989, '\p{^Nv=00048}', "");
Expect(0, 12989, '\P{Nv=00048}', "");
Expect(1, 12989, '\P{^Nv=00048}', "");
Expect(0, 12990, '\p{Nv=00048}', "");
Expect(1, 12990, '\p{^Nv=00048}', "");
Expect(1, 12990, '\P{Nv=00048}', "");
Expect(0, 12990, '\P{^Nv=00048}', "");
Error('\p{Is_Numeric_Value=/a/ 	+4_8}');
Error('\P{Is_Numeric_Value=/a/ 	+4_8}');
Expect(1, 12989, '\p{Is_Numeric_Value=+0_0_0_0_0_0048}', "");
Expect(0, 12989, '\p{^Is_Numeric_Value=+0_0_0_0_0_0048}', "");
Expect(0, 12989, '\P{Is_Numeric_Value=+0_0_0_0_0_0048}', "");
Expect(1, 12989, '\P{^Is_Numeric_Value=+0_0_0_0_0_0048}', "");
Expect(0, 12990, '\p{Is_Numeric_Value=+0_0_0_0_0_0048}', "");
Expect(1, 12990, '\p{^Is_Numeric_Value=+0_0_0_0_0_0048}', "");
Expect(1, 12990, '\P{Is_Numeric_Value=+0_0_0_0_0_0048}', "");
Expect(0, 12990, '\P{^Is_Numeric_Value=+0_0_0_0_0_0048}', "");
Error('\p{Is_Nv=	:=00048}');
Error('\P{Is_Nv=	:=00048}');
Expect(1, 12989, '\p{Is_Nv=+00000004_8}', "");
Expect(0, 12989, '\p{^Is_Nv=+00000004_8}', "");
Expect(0, 12989, '\P{Is_Nv=+00000004_8}', "");
Expect(1, 12989, '\P{^Is_Nv=+00000004_8}', "");
Expect(0, 12990, '\p{Is_Nv=+00000004_8}', "");
Expect(1, 12990, '\p{^Is_Nv=+00000004_8}', "");
Expect(1, 12990, '\P{Is_Nv=+00000004_8}', "");
Expect(0, 12990, '\P{^Is_Nv=+00000004_8}', "");
Error('\p{Numeric_Value=- 000000001/00000003/a/}');
Error('\P{Numeric_Value=- 000000001/00000003/a/}');
Expect(1, 74845, '\p{Numeric_Value=000000001/000003}', "");
Expect(0, 74845, '\p{^Numeric_Value=000000001/000003}', "");
Expect(0, 74845, '\P{Numeric_Value=000000001/000003}', "");
Expect(1, 74845, '\P{^Numeric_Value=000000001/000003}', "");
Expect(0, 74846, '\p{Numeric_Value=000000001/000003}', "");
Expect(1, 74846, '\p{^Numeric_Value=000000001/000003}', "");
Expect(1, 74846, '\P{Numeric_Value=000000001/000003}', "");
Expect(0, 74846, '\P{^Numeric_Value=000000001/000003}', "");
Error('\p{Numeric_Value=0.3}');
Error('\P{Numeric_Value=0.3}');
Error('\p{Numeric_Value=0.33}');
Error('\P{Numeric_Value=0.33}');
Expect(1, 74845, '\p{Numeric_Value=0.333}', "");
Expect(0, 74845, '\p{^Numeric_Value=0.333}', "");
Expect(0, 74845, '\P{Numeric_Value=0.333}', "");
Expect(1, 74845, '\P{^Numeric_Value=0.333}', "");
Expect(0, 74846, '\p{Numeric_Value=0.333}', "");
Expect(1, 74846, '\p{^Numeric_Value=0.333}', "");
Expect(1, 74846, '\P{Numeric_Value=0.333}', "");
Expect(0, 74846, '\P{^Numeric_Value=0.333}', "");
Error('\p{Nv=	:=+000001/0000000003}');
Error('\P{Nv=	:=+000001/0000000003}');
Expect(1, 74845, '\p{Nv=0000000001/00000003}', "");
Expect(0, 74845, '\p{^Nv=0000000001/00000003}', "");
Expect(0, 74845, '\P{Nv=0000000001/00000003}', "");
Expect(1, 74845, '\P{^Nv=0000000001/00000003}', "");
Expect(0, 74846, '\p{Nv=0000000001/00000003}', "");
Expect(1, 74846, '\p{^Nv=0000000001/00000003}', "");
Expect(1, 74846, '\P{Nv=0000000001/00000003}', "");
Expect(0, 74846, '\P{^Nv=0000000001/00000003}', "");
Error('\p{Nv=0.3}');
Error('\P{Nv=0.3}');
Error('\p{Nv=0.33}');
Error('\P{Nv=0.33}');
Expect(1, 74845, '\p{Nv=0.333}', "");
Expect(0, 74845, '\p{^Nv=0.333}', "");
Expect(0, 74845, '\P{Nv=0.333}', "");
Expect(1, 74845, '\P{^Nv=0.333}', "");
Expect(0, 74846, '\p{Nv=0.333}', "");
Expect(1, 74846, '\p{^Nv=0.333}', "");
Expect(1, 74846, '\P{Nv=0.333}', "");
Expect(0, 74846, '\P{^Nv=0.333}', "");
Error('\p{Is_Numeric_Value=	-01/00003:=}');
Error('\P{Is_Numeric_Value=	-01/00003:=}');
Expect(1, 74845, '\p{Is_Numeric_Value=001/03}', "");
Expect(0, 74845, '\p{^Is_Numeric_Value=001/03}', "");
Expect(0, 74845, '\P{Is_Numeric_Value=001/03}', "");
Expect(1, 74845, '\P{^Is_Numeric_Value=001/03}', "");
Expect(0, 74846, '\p{Is_Numeric_Value=001/03}', "");
Expect(1, 74846, '\p{^Is_Numeric_Value=001/03}', "");
Expect(1, 74846, '\P{Is_Numeric_Value=001/03}', "");
Expect(0, 74846, '\P{^Is_Numeric_Value=001/03}', "");
Error('\p{Is_Numeric_Value=0.3}');
Error('\P{Is_Numeric_Value=0.3}');
Error('\p{Is_Numeric_Value=0.33}');
Error('\P{Is_Numeric_Value=0.33}');
Expect(1, 74845, '\p{Is_Numeric_Value=0.333}', "");
Expect(0, 74845, '\p{^Is_Numeric_Value=0.333}', "");
Expect(0, 74845, '\P{Is_Numeric_Value=0.333}', "");
Expect(1, 74845, '\P{^Is_Numeric_Value=0.333}', "");
Expect(0, 74846, '\p{Is_Numeric_Value=0.333}', "");
Expect(1, 74846, '\p{^Is_Numeric_Value=0.333}', "");
Expect(1, 74846, '\P{Is_Numeric_Value=0.333}', "");
Expect(0, 74846, '\P{^Is_Numeric_Value=0.333}', "");
Error('\p{Is_Nv::=000000001/00000003}');
Error('\P{Is_Nv::=000000001/00000003}');
Expect(1, 74845, '\p{Is_Nv=000000001/03}', "");
Expect(0, 74845, '\p{^Is_Nv=000000001/03}', "");
Expect(0, 74845, '\P{Is_Nv=000000001/03}', "");
Expect(1, 74845, '\P{^Is_Nv=000000001/03}', "");
Expect(0, 74846, '\p{Is_Nv=000000001/03}', "");
Expect(1, 74846, '\p{^Is_Nv=000000001/03}', "");
Expect(1, 74846, '\P{Is_Nv=000000001/03}', "");
Expect(0, 74846, '\P{^Is_Nv=000000001/03}', "");
Error('\p{Is_Nv=0.3}');
Error('\P{Is_Nv=0.3}');
Error('\p{Is_Nv=0.33}');
Error('\P{Is_Nv=0.33}');
Expect(1, 74845, '\p{Is_Nv=0.333}', "");
Expect(0, 74845, '\p{^Is_Nv=0.333}', "");
Expect(0, 74845, '\P{Is_Nv=0.333}', "");
Expect(1, 74845, '\P{^Is_Nv=0.333}', "");
Expect(0, 74846, '\p{Is_Nv=0.333}', "");
Expect(1, 74846, '\p{^Is_Nv=0.333}', "");
Expect(1, 74846, '\P{Is_Nv=0.333}', "");
Expect(0, 74846, '\P{^Is_Nv=0.333}', "");
Error('\p{Numeric_Value=/a/000000002000}');
Error('\P{Numeric_Value=/a/000000002000}');
Expect(1, 65827, '\p{Numeric_Value=200_0}', "");
Expect(0, 65827, '\p{^Numeric_Value=200_0}', "");
Expect(0, 65827, '\P{Numeric_Value=200_0}', "");
Expect(1, 65827, '\P{^Numeric_Value=200_0}', "");
Expect(0, 65828, '\p{Numeric_Value=200_0}', "");
Expect(1, 65828, '\p{^Numeric_Value=200_0}', "");
Expect(1, 65828, '\P{Numeric_Value=200_0}', "");
Expect(0, 65828, '\P{^Numeric_Value=200_0}', "");
Error('\p{Nv=/a/-_00_20_00}');
Error('\P{Nv=/a/-_00_20_00}');
Expect(1, 65827, '\p{Nv=+0_0_0_0_0_2_0_00}', "");
Expect(0, 65827, '\p{^Nv=+0_0_0_0_0_2_0_00}', "");
Expect(0, 65827, '\P{Nv=+0_0_0_0_0_2_0_00}', "");
Expect(1, 65827, '\P{^Nv=+0_0_0_0_0_2_0_00}', "");
Expect(0, 65828, '\p{Nv=+0_0_0_0_0_2_0_00}', "");
Expect(1, 65828, '\p{^Nv=+0_0_0_0_0_2_0_00}', "");
Expect(1, 65828, '\P{Nv=+0_0_0_0_0_2_0_00}', "");
Expect(0, 65828, '\P{^Nv=+0_0_0_0_0_2_0_00}', "");
Error('\p{Is_Numeric_Value:   /a/		0002000}');
Error('\P{Is_Numeric_Value:   /a/		0002000}');
Expect(1, 65827, '\p{Is_Numeric_Value:	+2000}', "");
Expect(0, 65827, '\p{^Is_Numeric_Value:	+2000}', "");
Expect(0, 65827, '\P{Is_Numeric_Value:	+2000}', "");
Expect(1, 65827, '\P{^Is_Numeric_Value:	+2000}', "");
Expect(0, 65828, '\p{Is_Numeric_Value:	+2000}', "");
Expect(1, 65828, '\p{^Is_Numeric_Value:	+2000}', "");
Expect(1, 65828, '\P{Is_Numeric_Value:	+2000}', "");
Expect(0, 65828, '\P{^Is_Numeric_Value:	+2000}', "");
Error('\p{Is_Nv=	:=+0000000002000}');
Error('\P{Is_Nv=	:=+0000000002000}');
Expect(1, 65827, '\p{Is_Nv=000000002000}', "");
Expect(0, 65827, '\p{^Is_Nv=000000002000}', "");
Expect(0, 65827, '\P{Is_Nv=000000002000}', "");
Expect(1, 65827, '\P{^Is_Nv=000000002000}', "");
Expect(0, 65828, '\p{Is_Nv=000000002000}', "");
Expect(1, 65828, '\p{^Is_Nv=000000002000}', "");
Expect(1, 65828, '\P{Is_Nv=000000002000}', "");
Expect(0, 65828, '\P{^Is_Nv=000000002000}', "");
Error('\p{Numeric_Value=-:=00_00_02_9}');
Error('\P{Numeric_Value=-:=00_00_02_9}');
Expect(1, 12889, '\p{Numeric_Value=0000029}', "");
Expect(0, 12889, '\p{^Numeric_Value=0000029}', "");
Expect(0, 12889, '\P{Numeric_Value=0000029}', "");
Expect(1, 12889, '\P{^Numeric_Value=0000029}', "");
Expect(0, 12890, '\p{Numeric_Value=0000029}', "");
Expect(1, 12890, '\p{^Numeric_Value=0000029}', "");
Expect(1, 12890, '\P{Numeric_Value=0000029}', "");
Expect(0, 12890, '\P{^Numeric_Value=0000029}', "");
Error('\p{Nv=	00000000029:=}');
Error('\P{Nv=	00000000029:=}');
Expect(1, 12889, '\p{Nv=029}', "");
Expect(0, 12889, '\p{^Nv=029}', "");
Expect(0, 12889, '\P{Nv=029}', "");
Expect(1, 12889, '\P{^Nv=029}', "");
Expect(0, 12890, '\p{Nv=029}', "");
Expect(1, 12890, '\p{^Nv=029}', "");
Expect(1, 12890, '\P{Nv=029}', "");
Expect(0, 12890, '\P{^Nv=029}', "");
Error('\p{Is_Numeric_Value= +02_9:=}');
Error('\P{Is_Numeric_Value= +02_9:=}');
Expect(1, 12889, '\p{Is_Numeric_Value=+2_9}', "");
Expect(0, 12889, '\p{^Is_Numeric_Value=+2_9}', "");
Expect(0, 12889, '\P{Is_Numeric_Value=+2_9}', "");
Expect(1, 12889, '\P{^Is_Numeric_Value=+2_9}', "");
Expect(0, 12890, '\p{Is_Numeric_Value=+2_9}', "");
Expect(1, 12890, '\p{^Is_Numeric_Value=+2_9}', "");
Expect(1, 12890, '\P{Is_Numeric_Value=+2_9}', "");
Expect(0, 12890, '\P{^Is_Numeric_Value=+2_9}', "");
Error('\p{Is_Nv=-	00000002_9/a/}');
Error('\P{Is_Nv=-	00000002_9/a/}');
Expect(1, 12889, '\p{Is_Nv=000000029}', "");
Expect(0, 12889, '\p{^Is_Nv=000000029}', "");
Expect(0, 12889, '\P{Is_Nv=000000029}', "");
Expect(1, 12889, '\P{^Is_Nv=000000029}', "");
Expect(0, 12890, '\p{Is_Nv=000000029}', "");
Expect(1, 12890, '\p{^Is_Nv=000000029}', "");
Expect(1, 12890, '\P{Is_Nv=000000029}', "");
Expect(0, 12890, '\P{^Is_Nv=000000029}', "");
Error('\p{Numeric_Value=:=0_1_0_0_0_00}');
Error('\P{Numeric_Value=:=0_1_0_0_0_00}');
Expect(1, 8584, '\p{Numeric_Value=0000100000}', "");
Expect(0, 8584, '\p{^Numeric_Value=0000100000}', "");
Expect(0, 8584, '\P{Numeric_Value=0000100000}', "");
Expect(1, 8584, '\P{^Numeric_Value=0000100000}', "");
Expect(0, 8585, '\p{Numeric_Value=0000100000}', "");
Expect(1, 8585, '\p{^Numeric_Value=0000100000}', "");
Expect(1, 8585, '\P{Numeric_Value=0000100000}', "");
Expect(0, 8585, '\P{^Numeric_Value=0000100000}', "");
Error('\p{Nv:	-:=+000_010_000_0}');
Error('\P{Nv:	-:=+000_010_000_0}');
Expect(1, 8584, '\p{Nv:   00_00_00_00_10_00_00}', "");
Expect(0, 8584, '\p{^Nv:   00_00_00_00_10_00_00}', "");
Expect(0, 8584, '\P{Nv:   00_00_00_00_10_00_00}', "");
Expect(1, 8584, '\P{^Nv:   00_00_00_00_10_00_00}', "");
Expect(0, 8585, '\p{Nv:   00_00_00_00_10_00_00}', "");
Expect(1, 8585, '\p{^Nv:   00_00_00_00_10_00_00}', "");
Expect(1, 8585, '\P{Nv:   00_00_00_00_10_00_00}', "");
Expect(0, 8585, '\P{^Nv:   00_00_00_00_10_00_00}', "");
Error('\p{Is_Numeric_Value: :=	0100000}');
Error('\P{Is_Numeric_Value: :=	0100000}');
Expect(1, 8584, '\p{Is_Numeric_Value=+000000_001000_00}', "");
Expect(0, 8584, '\p{^Is_Numeric_Value=+000000_001000_00}', "");
Expect(0, 8584, '\P{Is_Numeric_Value=+000000_001000_00}', "");
Expect(1, 8584, '\P{^Is_Numeric_Value=+000000_001000_00}', "");
Expect(0, 8585, '\p{Is_Numeric_Value=+000000_001000_00}', "");
Expect(1, 8585, '\p{^Is_Numeric_Value=+000000_001000_00}', "");
Expect(1, 8585, '\P{Is_Numeric_Value=+000000_001000_00}', "");
Expect(0, 8585, '\P{^Is_Numeric_Value=+000000_001000_00}', "");
Error('\p{Is_Nv=	:=00_10_00_00}');
Error('\P{Is_Nv=	:=00_10_00_00}');
Expect(1, 8584, '\p{Is_Nv=000000000100000}', "");
Expect(0, 8584, '\p{^Is_Nv=000000000100000}', "");
Expect(0, 8584, '\P{Is_Nv=000000000100000}', "");
Expect(1, 8584, '\P{^Is_Nv=000000000100000}', "");
Expect(0, 8585, '\p{Is_Nv=000000000100000}', "");
Expect(1, 8585, '\p{^Is_Nv=000000000100000}', "");
Expect(1, 8585, '\P{Is_Nv=000000000100000}', "");
Expect(0, 8585, '\P{^Is_Nv=000000000100000}', "");
Error('\p{Numeric_Value=-:=+0000000050}');
Error('\P{Numeric_Value=-:=+0000000050}');
Expect(1, 119661, '\p{Numeric_Value=0000000050}', "");
Expect(0, 119661, '\p{^Numeric_Value=0000000050}', "");
Expect(0, 119661, '\P{Numeric_Value=0000000050}', "");
Expect(1, 119661, '\P{^Numeric_Value=0000000050}', "");
Expect(0, 119662, '\p{Numeric_Value=0000000050}', "");
Expect(1, 119662, '\p{^Numeric_Value=0000000050}', "");
Expect(1, 119662, '\P{Numeric_Value=0000000050}', "");
Expect(0, 119662, '\P{^Numeric_Value=0000000050}', "");
Error('\p{Nv=:=50}');
Error('\P{Nv=:=50}');
Expect(1, 119661, '\p{Nv=0000_0005_0}', "");
Expect(0, 119661, '\p{^Nv=0000_0005_0}', "");
Expect(0, 119661, '\P{Nv=0000_0005_0}', "");
Expect(1, 119661, '\P{^Nv=0000_0005_0}', "");
Expect(0, 119662, '\p{Nv=0000_0005_0}', "");
Expect(1, 119662, '\p{^Nv=0000_0005_0}', "");
Expect(1, 119662, '\P{Nv=0000_0005_0}', "");
Expect(0, 119662, '\P{^Nv=0000_0005_0}', "");
Error('\p{Is_Numeric_Value=_:=+005_0}');
Error('\P{Is_Numeric_Value=_:=+005_0}');
Expect(1, 119661, '\p{Is_Numeric_Value=0000005_0}', "");
Expect(0, 119661, '\p{^Is_Numeric_Value=0000005_0}', "");
Expect(0, 119661, '\P{Is_Numeric_Value=0000005_0}', "");
Expect(1, 119661, '\P{^Is_Numeric_Value=0000005_0}', "");
Expect(0, 119662, '\p{Is_Numeric_Value=0000005_0}', "");
Expect(1, 119662, '\p{^Is_Numeric_Value=0000005_0}', "");
Expect(1, 119662, '\P{Is_Numeric_Value=0000005_0}', "");
Expect(0, 119662, '\P{^Is_Numeric_Value=0000005_0}', "");
Error('\p{Is_Nv:	/a/_	050}');
Error('\P{Is_Nv:	/a/_	050}');
Expect(1, 119661, '\p{Is_Nv=05_0}', "");
Expect(0, 119661, '\p{^Is_Nv=05_0}', "");
Expect(0, 119661, '\P{Is_Nv=05_0}', "");
Expect(1, 119661, '\P{^Is_Nv=05_0}', "");
Expect(0, 119662, '\p{Is_Nv=05_0}', "");
Expect(1, 119662, '\p{^Is_Nv=05_0}', "");
Expect(1, 119662, '\P{Is_Nv=05_0}', "");
Expect(0, 119662, '\P{^Is_Nv=05_0}', "");
Error('\p{Numeric_Value=- +00039/a/}');
Error('\P{Numeric_Value=- +00039/a/}');
Expect(1, 12980, '\p{Numeric_Value=0000000003_9}', "");
Expect(0, 12980, '\p{^Numeric_Value=0000000003_9}', "");
Expect(0, 12980, '\P{Numeric_Value=0000000003_9}', "");
Expect(1, 12980, '\P{^Numeric_Value=0000000003_9}', "");
Expect(0, 12981, '\p{Numeric_Value=0000000003_9}', "");
Expect(1, 12981, '\p{^Numeric_Value=0000000003_9}', "");
Expect(1, 12981, '\P{Numeric_Value=0000000003_9}', "");
Expect(0, 12981, '\P{^Numeric_Value=0000000003_9}', "");
Error('\p{Nv=	00000000039/a/}');
Error('\P{Nv=	00000000039/a/}');
Expect(1, 12980, '\p{Nv: 03_9}', "");
Expect(0, 12980, '\p{^Nv: 03_9}', "");
Expect(0, 12980, '\P{Nv: 03_9}', "");
Expect(1, 12980, '\P{^Nv: 03_9}', "");
Expect(0, 12981, '\p{Nv: 03_9}', "");
Expect(1, 12981, '\p{^Nv: 03_9}', "");
Expect(1, 12981, '\P{Nv: 03_9}', "");
Expect(0, 12981, '\P{^Nv: 03_9}', "");
Error('\p{Is_Numeric_Value=_ 000000039/a/}');
Error('\P{Is_Numeric_Value=_ 000000039/a/}');
Expect(1, 12980, '\p{Is_Numeric_Value: 00039}', "");
Expect(0, 12980, '\p{^Is_Numeric_Value: 00039}', "");
Expect(0, 12980, '\P{Is_Numeric_Value: 00039}', "");
Expect(1, 12980, '\P{^Is_Numeric_Value: 00039}', "");
Expect(0, 12981, '\p{Is_Numeric_Value: 00039}', "");
Expect(1, 12981, '\p{^Is_Numeric_Value: 00039}', "");
Expect(1, 12981, '\P{Is_Numeric_Value: 00039}', "");
Expect(0, 12981, '\P{^Is_Numeric_Value: 00039}', "");
Error('\p{Is_Nv=:= -000000039}');
Error('\P{Is_Nv=:= -000000039}');
Expect(1, 12980, '\p{Is_Nv=+000_000_003_9}', "");
Expect(0, 12980, '\p{^Is_Nv=+000_000_003_9}', "");
Expect(0, 12980, '\P{Is_Nv=+000_000_003_9}', "");
Expect(1, 12980, '\P{^Is_Nv=+000_000_003_9}', "");
Expect(0, 12981, '\p{Is_Nv=+000_000_003_9}', "");
Expect(1, 12981, '\p{^Is_Nv=+000_000_003_9}', "");
Expect(1, 12981, '\P{Is_Nv=+000_000_003_9}', "");
Expect(0, 12981, '\P{^Is_Nv=+000_000_003_9}', "");
Error('\p{Numeric_Value=/a/000001/00016}');
Error('\P{Numeric_Value=/a/000001/00016}');
Expect(1, 43059, '\p{Numeric_Value=+0000000001/0016}', "");
Expect(0, 43059, '\p{^Numeric_Value=+0000000001/0016}', "");
Expect(0, 43059, '\P{Numeric_Value=+0000000001/0016}', "");
Expect(1, 43059, '\P{^Numeric_Value=+0000000001/0016}', "");
Expect(0, 43060, '\p{Numeric_Value=+0000000001/0016}', "");
Expect(1, 43060, '\p{^Numeric_Value=+0000000001/0016}', "");
Expect(1, 43060, '\P{Numeric_Value=+0000000001/0016}', "");
Expect(0, 43060, '\P{^Numeric_Value=+0000000001/0016}', "");
Error('\p{Numeric_Value=0.06}');
Error('\P{Numeric_Value=0.06}');
Expect(1, 43059, '\p{Numeric_Value=0.062}', "");
Expect(0, 43059, '\p{^Numeric_Value=0.062}', "");
Expect(0, 43059, '\P{Numeric_Value=0.062}', "");
Expect(1, 43059, '\P{^Numeric_Value=0.062}', "");
Expect(0, 43060, '\p{Numeric_Value=0.062}', "");
Expect(1, 43060, '\p{^Numeric_Value=0.062}', "");
Expect(1, 43060, '\P{Numeric_Value=0.062}', "");
Expect(0, 43060, '\P{^Numeric_Value=0.062}', "");
Error('\p{Nv=	-0001/0000016/a/}');
Error('\P{Nv=	-0001/0000016/a/}');
Expect(1, 43059, '\p{Nv:	1/000000016}', "");
Expect(0, 43059, '\p{^Nv:	1/000000016}', "");
Expect(0, 43059, '\P{Nv:	1/000000016}', "");
Expect(1, 43059, '\P{^Nv:	1/000000016}', "");
Expect(0, 43060, '\p{Nv:	1/000000016}', "");
Expect(1, 43060, '\p{^Nv:	1/000000016}', "");
Expect(1, 43060, '\P{Nv:	1/000000016}', "");
Expect(0, 43060, '\P{^Nv:	1/000000016}', "");
Error('\p{Nv=0.06}');
Error('\P{Nv=0.06}');
Expect(1, 43059, '\p{Nv=0.062}', "");
Expect(0, 43059, '\p{^Nv=0.062}', "");
Expect(0, 43059, '\P{Nv=0.062}', "");
Expect(1, 43059, '\P{^Nv=0.062}', "");
Expect(0, 43060, '\p{Nv=0.062}', "");
Expect(1, 43060, '\p{^Nv=0.062}', "");
Expect(1, 43060, '\P{Nv=0.062}', "");
Expect(0, 43060, '\P{^Nv=0.062}', "");
Error('\p{Is_Numeric_Value=-_000000001/016:=}');
Error('\P{Is_Numeric_Value=-_000000001/016:=}');
Expect(1, 43059, '\p{Is_Numeric_Value=01/00016}', "");
Expect(0, 43059, '\p{^Is_Numeric_Value=01/00016}', "");
Expect(0, 43059, '\P{Is_Numeric_Value=01/00016}', "");
Expect(1, 43059, '\P{^Is_Numeric_Value=01/00016}', "");
Expect(0, 43060, '\p{Is_Numeric_Value=01/00016}', "");
Expect(1, 43060, '\p{^Is_Numeric_Value=01/00016}', "");
Expect(1, 43060, '\P{Is_Numeric_Value=01/00016}', "");
Expect(0, 43060, '\P{^Is_Numeric_Value=01/00016}', "");
Error('\p{Is_Numeric_Value=0.06}');
Error('\P{Is_Numeric_Value=0.06}');
Expect(1, 43059, '\p{Is_Numeric_Value=0.062}', "");
Expect(0, 43059, '\p{^Is_Numeric_Value=0.062}', "");
Expect(0, 43059, '\P{Is_Numeric_Value=0.062}', "");
Expect(1, 43059, '\P{^Is_Numeric_Value=0.062}', "");
Expect(0, 43060, '\p{Is_Numeric_Value=0.062}', "");
Expect(1, 43060, '\p{^Is_Numeric_Value=0.062}', "");
Expect(1, 43060, '\P{Is_Numeric_Value=0.062}', "");
Expect(0, 43060, '\P{^Is_Numeric_Value=0.062}', "");
Error('\p{Is_Nv=:= 01/0016}');
Error('\P{Is_Nv=:= 01/0016}');
Expect(1, 43059, '\p{Is_Nv: 1/0016}', "");
Expect(0, 43059, '\p{^Is_Nv: 1/0016}', "");
Expect(0, 43059, '\P{Is_Nv: 1/0016}', "");
Expect(1, 43059, '\P{^Is_Nv: 1/0016}', "");
Expect(0, 43060, '\p{Is_Nv: 1/0016}', "");
Expect(1, 43060, '\p{^Is_Nv: 1/0016}', "");
Expect(1, 43060, '\P{Is_Nv: 1/0016}', "");
Expect(0, 43060, '\P{^Is_Nv: 1/0016}', "");
Error('\p{Is_Nv=0.06}');
Error('\P{Is_Nv=0.06}');
Expect(1, 43059, '\p{Is_Nv=0.062}', "");
Expect(0, 43059, '\p{^Is_Nv=0.062}', "");
Expect(0, 43059, '\P{Is_Nv=0.062}', "");
Expect(1, 43059, '\P{^Is_Nv=0.062}', "");
Expect(0, 43060, '\p{Is_Nv=0.062}', "");
Expect(1, 43060, '\p{^Is_Nv=0.062}', "");
Expect(1, 43060, '\P{Is_Nv=0.062}', "");
Expect(0, 43060, '\P{^Is_Nv=0.062}', "");
Error('\p{Numeric_Value=__003/00000002:=}');
Error('\P{Numeric_Value=__003/00000002:=}');
Expect(1, 3883, '\p{Numeric_Value=+0000000003/0000002}', "");
Expect(0, 3883, '\p{^Numeric_Value=+0000000003/0000002}', "");
Expect(0, 3883, '\P{Numeric_Value=+0000000003/0000002}', "");
Expect(1, 3883, '\P{^Numeric_Value=+0000000003/0000002}', "");
Expect(0, 3884, '\p{Numeric_Value=+0000000003/0000002}', "");
Expect(1, 3884, '\p{^Numeric_Value=+0000000003/0000002}', "");
Expect(1, 3884, '\P{Numeric_Value=+0000000003/0000002}', "");
Expect(0, 3884, '\P{^Numeric_Value=+0000000003/0000002}', "");
Expect(1, 3883, '\p{Numeric_Value=1.500}', "");
Expect(0, 3883, '\p{^Numeric_Value=1.500}', "");
Expect(0, 3883, '\P{Numeric_Value=1.500}', "");
Expect(1, 3883, '\P{^Numeric_Value=1.500}', "");
Expect(0, 3884, '\p{Numeric_Value=1.500}', "");
Expect(1, 3884, '\p{^Numeric_Value=1.500}', "");
Expect(1, 3884, '\P{Numeric_Value=1.500}', "");
Expect(0, 3884, '\P{^Numeric_Value=1.500}', "");
Error('\p{Nv=_/a/3/000002}');
Error('\P{Nv=_/a/3/000002}');
Expect(1, 3883, '\p{Nv=003/00002}', "");
Expect(0, 3883, '\p{^Nv=003/00002}', "");
Expect(0, 3883, '\P{Nv=003/00002}', "");
Expect(1, 3883, '\P{^Nv=003/00002}', "");
Expect(0, 3884, '\p{Nv=003/00002}', "");
Expect(1, 3884, '\p{^Nv=003/00002}', "");
Expect(1, 3884, '\P{Nv=003/00002}', "");
Expect(0, 3884, '\P{^Nv=003/00002}', "");
Expect(1, 3883, '\p{Nv=1.500}', "");
Expect(0, 3883, '\p{^Nv=1.500}', "");
Expect(0, 3883, '\P{Nv=1.500}', "");
Expect(1, 3883, '\P{^Nv=1.500}', "");
Expect(0, 3884, '\p{Nv=1.500}', "");
Expect(1, 3884, '\p{^Nv=1.500}', "");
Expect(1, 3884, '\P{Nv=1.500}', "");
Expect(0, 3884, '\P{^Nv=1.500}', "");
Error('\p{Is_Numeric_Value= :=3/002}');
Error('\P{Is_Numeric_Value= :=3/002}');
Expect(1, 3883, '\p{Is_Numeric_Value=+003/00002}', "");
Expect(0, 3883, '\p{^Is_Numeric_Value=+003/00002}', "");
Expect(0, 3883, '\P{Is_Numeric_Value=+003/00002}', "");
Expect(1, 3883, '\P{^Is_Numeric_Value=+003/00002}', "");
Expect(0, 3884, '\p{Is_Numeric_Value=+003/00002}', "");
Expect(1, 3884, '\p{^Is_Numeric_Value=+003/00002}', "");
Expect(1, 3884, '\P{Is_Numeric_Value=+003/00002}', "");
Expect(0, 3884, '\P{^Is_Numeric_Value=+003/00002}', "");
Expect(1, 3883, '\p{Is_Numeric_Value=1.500}', "");
Expect(0, 3883, '\p{^Is_Numeric_Value=1.500}', "");
Expect(0, 3883, '\P{Is_Numeric_Value=1.500}', "");
Expect(1, 3883, '\P{^Is_Numeric_Value=1.500}', "");
Expect(0, 3884, '\p{Is_Numeric_Value=1.500}', "");
Expect(1, 3884, '\p{^Is_Numeric_Value=1.500}', "");
Expect(1, 3884, '\P{Is_Numeric_Value=1.500}', "");
Expect(0, 3884, '\P{^Is_Numeric_Value=1.500}', "");
Error('\p{Is_Nv=	/a/+0003/000002}');
Error('\P{Is_Nv=	/a/+0003/000002}');
Expect(1, 3883, '\p{Is_Nv=+000000003/2}', "");
Expect(0, 3883, '\p{^Is_Nv=+000000003/2}', "");
Expect(0, 3883, '\P{Is_Nv=+000000003/2}', "");
Expect(1, 3883, '\P{^Is_Nv=+000000003/2}', "");
Expect(0, 3884, '\p{Is_Nv=+000000003/2}', "");
Expect(1, 3884, '\p{^Is_Nv=+000000003/2}', "");
Expect(1, 3884, '\P{Is_Nv=+000000003/2}', "");
Expect(0, 3884, '\P{^Is_Nv=+000000003/2}', "");
Expect(1, 3883, '\p{Is_Nv=1.500}', "");
Expect(0, 3883, '\p{^Is_Nv=1.500}', "");
Expect(0, 3883, '\P{Is_Nv=1.500}', "");
Expect(1, 3883, '\P{^Is_Nv=1.500}', "");
Expect(0, 3884, '\p{Is_Nv=1.500}', "");
Expect(1, 3884, '\p{^Is_Nv=1.500}', "");
Expect(1, 3884, '\P{Is_Nv=1.500}', "");
Expect(0, 3884, '\P{^Is_Nv=1.500}', "");
Error('\p{Numeric_Value=/a/-_04_1}');
Error('\P{Numeric_Value=/a/-_04_1}');
Expect(1, 12982, '\p{Numeric_Value=00000000041}', "");
Expect(0, 12982, '\p{^Numeric_Value=00000000041}', "");
Expect(0, 12982, '\P{Numeric_Value=00000000041}', "");
Expect(1, 12982, '\P{^Numeric_Value=00000000041}', "");
Expect(0, 12983, '\p{Numeric_Value=00000000041}', "");
Expect(1, 12983, '\p{^Numeric_Value=00000000041}', "");
Expect(1, 12983, '\P{Numeric_Value=00000000041}', "");
Expect(0, 12983, '\P{^Numeric_Value=00000000041}', "");
Error('\p{Nv:    41/a/}');
Error('\P{Nv:    41/a/}');
Expect(1, 12982, '\p{Nv=0000000041}', "");
Expect(0, 12982, '\p{^Nv=0000000041}', "");
Expect(0, 12982, '\P{Nv=0000000041}', "");
Expect(1, 12982, '\P{^Nv=0000000041}', "");
Expect(0, 12983, '\p{Nv=0000000041}', "");
Expect(1, 12983, '\p{^Nv=0000000041}', "");
Expect(1, 12983, '\P{Nv=0000000041}', "");
Expect(0, 12983, '\P{^Nv=0000000041}', "");
Error('\p{Is_Numeric_Value=/a/_	041}');
Error('\P{Is_Numeric_Value=/a/_	041}');
Expect(1, 12982, '\p{Is_Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(0, 12982, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(0, 12982, '\P{Is_Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(1, 12982, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(0, 12983, '\p{Is_Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(1, 12983, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(1, 12983, '\P{Is_Numeric_Value=0_0_0_0_0_0_0041}', "");
Expect(0, 12983, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0041}', "");
Error('\p{Is_Nv=/a/  +000000041}');
Error('\P{Is_Nv=/a/  +000000041}');
Expect(1, 12982, '\p{Is_Nv=00_04_1}', "");
Expect(0, 12982, '\p{^Is_Nv=00_04_1}', "");
Expect(0, 12982, '\P{Is_Nv=00_04_1}', "");
Expect(1, 12982, '\P{^Is_Nv=00_04_1}', "");
Expect(0, 12983, '\p{Is_Nv=00_04_1}', "");
Expect(1, 12983, '\p{^Is_Nv=00_04_1}', "");
Expect(1, 12983, '\P{Is_Nv=00_04_1}', "");
Expect(0, 12983, '\P{^Is_Nv=00_04_1}', "");
Error('\p{Numeric_Value=_:=00_00_12}');
Error('\P{Numeric_Value=_:=00_00_12}');
Expect(1, 9452, '\p{Numeric_Value=+0000012}', "");
Expect(0, 9452, '\p{^Numeric_Value=+0000012}', "");
Expect(0, 9452, '\P{Numeric_Value=+0000012}', "");
Expect(1, 9452, '\P{^Numeric_Value=+0000012}', "");
Expect(0, 9453, '\p{Numeric_Value=+0000012}', "");
Expect(1, 9453, '\p{^Numeric_Value=+0000012}', "");
Expect(1, 9453, '\P{Numeric_Value=+0000012}', "");
Expect(0, 9453, '\P{^Numeric_Value=+0000012}', "");
Error('\p{Nv=:=	000000012}');
Error('\P{Nv=:=	000000012}');
Expect(1, 9452, '\p{Nv:01_2}', "");
Expect(0, 9452, '\p{^Nv:01_2}', "");
Expect(0, 9452, '\P{Nv:01_2}', "");
Expect(1, 9452, '\P{^Nv:01_2}', "");
Expect(0, 9453, '\p{Nv:01_2}', "");
Expect(1, 9453, '\p{^Nv:01_2}', "");
Expect(1, 9453, '\P{Nv:01_2}', "");
Expect(0, 9453, '\P{^Nv:01_2}', "");
Error('\p{Is_Numeric_Value=	/a/012}');
Error('\P{Is_Numeric_Value=	/a/012}');
Expect(1, 9452, '\p{Is_Numeric_Value=00_00_00_00_012}', "");
Expect(0, 9452, '\p{^Is_Numeric_Value=00_00_00_00_012}', "");
Expect(0, 9452, '\P{Is_Numeric_Value=00_00_00_00_012}', "");
Expect(1, 9452, '\P{^Is_Numeric_Value=00_00_00_00_012}', "");
Expect(0, 9453, '\p{Is_Numeric_Value=00_00_00_00_012}', "");
Expect(1, 9453, '\p{^Is_Numeric_Value=00_00_00_00_012}', "");
Expect(1, 9453, '\P{Is_Numeric_Value=00_00_00_00_012}', "");
Expect(0, 9453, '\P{^Is_Numeric_Value=00_00_00_00_012}', "");
Error('\p{Is_Nv=/a/-00012}');
Error('\P{Is_Nv=/a/-00012}');
Expect(1, 9452, '\p{Is_Nv=+00000001_2}', "");
Expect(0, 9452, '\p{^Is_Nv=+00000001_2}', "");
Expect(0, 9452, '\P{Is_Nv=+00000001_2}', "");
Expect(1, 9452, '\P{^Is_Nv=+00000001_2}', "");
Expect(0, 9453, '\p{Is_Nv=+00000001_2}', "");
Expect(1, 9453, '\p{^Is_Nv=+00000001_2}', "");
Expect(1, 9453, '\P{Is_Nv=+00000001_2}', "");
Expect(0, 9453, '\P{^Is_Nv=+00000001_2}', "");
Error('\p{Numeric_Value=_+015:=}');
Error('\P{Numeric_Value=_+015:=}');
Expect(1, 9455, '\p{Numeric_Value=+0_0_0_0_0_015}', "");
Expect(0, 9455, '\p{^Numeric_Value=+0_0_0_0_0_015}', "");
Expect(0, 9455, '\P{Numeric_Value=+0_0_0_0_0_015}', "");
Expect(1, 9455, '\P{^Numeric_Value=+0_0_0_0_0_015}', "");
Expect(0, 9456, '\p{Numeric_Value=+0_0_0_0_0_015}', "");
Expect(1, 9456, '\p{^Numeric_Value=+0_0_0_0_0_015}', "");
Expect(1, 9456, '\P{Numeric_Value=+0_0_0_0_0_015}', "");
Expect(0, 9456, '\P{^Numeric_Value=+0_0_0_0_0_015}', "");
Error('\p{Nv=:=	 000015}');
Error('\P{Nv=:=	 000015}');
Expect(1, 9455, '\p{Nv:015}', "");
Expect(0, 9455, '\p{^Nv:015}', "");
Expect(0, 9455, '\P{Nv:015}', "");
Expect(1, 9455, '\P{^Nv:015}', "");
Expect(0, 9456, '\p{Nv:015}', "");
Expect(1, 9456, '\p{^Nv:015}', "");
Expect(1, 9456, '\P{Nv:015}', "");
Expect(0, 9456, '\P{^Nv:015}', "");
Error('\p{Is_Numeric_Value=-/a/00000015}');
Error('\P{Is_Numeric_Value=-/a/00000015}');
Expect(1, 9455, '\p{Is_Numeric_Value=+00_00_00_01_5}', "");
Expect(0, 9455, '\p{^Is_Numeric_Value=+00_00_00_01_5}', "");
Expect(0, 9455, '\P{Is_Numeric_Value=+00_00_00_01_5}', "");
Expect(1, 9455, '\P{^Is_Numeric_Value=+00_00_00_01_5}', "");
Expect(0, 9456, '\p{Is_Numeric_Value=+00_00_00_01_5}', "");
Expect(1, 9456, '\p{^Is_Numeric_Value=+00_00_00_01_5}', "");
Expect(1, 9456, '\P{Is_Numeric_Value=+00_00_00_01_5}', "");
Expect(0, 9456, '\P{^Is_Numeric_Value=+00_00_00_01_5}', "");
Error('\p{Is_Nv=:= 	+0000015}');
Error('\P{Is_Nv=:= 	+0000015}');
Expect(1, 9455, '\p{Is_Nv=0_0_0_0_15}', "");
Expect(0, 9455, '\p{^Is_Nv=0_0_0_0_15}', "");
Expect(0, 9455, '\P{Is_Nv=0_0_0_0_15}', "");
Expect(1, 9455, '\P{^Is_Nv=0_0_0_0_15}', "");
Expect(0, 9456, '\p{Is_Nv=0_0_0_0_15}', "");
Expect(1, 9456, '\p{^Is_Nv=0_0_0_0_15}', "");
Expect(1, 9456, '\P{Is_Nv=0_0_0_0_15}', "");
Expect(0, 9456, '\P{^Is_Nv=0_0_0_0_15}', "");
Error('\p{Numeric_Value=-:=-000000001/00000002}');
Error('\P{Numeric_Value=-:=-000000001/00000002}');
Expect(1, 3891, '\p{Numeric_Value=-00000001/2}', "");
Expect(0, 3891, '\p{^Numeric_Value=-00000001/2}', "");
Expect(0, 3891, '\P{Numeric_Value=-00000001/2}', "");
Expect(1, 3891, '\P{^Numeric_Value=-00000001/2}', "");
Expect(0, 3892, '\p{Numeric_Value=-00000001/2}', "");
Expect(1, 3892, '\p{^Numeric_Value=-00000001/2}', "");
Expect(1, 3892, '\P{Numeric_Value=-00000001/2}', "");
Expect(0, 3892, '\P{^Numeric_Value=-00000001/2}', "");
Expect(1, 3891, '\p{Numeric_Value:-0.500}', "");
Expect(0, 3891, '\p{^Numeric_Value:-0.500}', "");
Expect(0, 3891, '\P{Numeric_Value:-0.500}', "");
Expect(1, 3891, '\P{^Numeric_Value:-0.500}', "");
Expect(0, 3892, '\p{Numeric_Value:-0.500}', "");
Expect(1, 3892, '\p{^Numeric_Value:-0.500}', "");
Expect(1, 3892, '\P{Numeric_Value:-0.500}', "");
Expect(0, 3892, '\P{^Numeric_Value:-0.500}', "");
Error('\p{Nv=/a/-_-0000000001/002}');
Error('\P{Nv=/a/-_-0000000001/002}');
Expect(1, 3891, '\p{Nv=-01/002}', "");
Expect(0, 3891, '\p{^Nv=-01/002}', "");
Expect(0, 3891, '\P{Nv=-01/002}', "");
Expect(1, 3891, '\P{^Nv=-01/002}', "");
Expect(0, 3892, '\p{Nv=-01/002}', "");
Expect(1, 3892, '\p{^Nv=-01/002}', "");
Expect(1, 3892, '\P{Nv=-01/002}', "");
Expect(0, 3892, '\P{^Nv=-01/002}', "");
Expect(1, 3891, '\p{Nv=-0.500}', "");
Expect(0, 3891, '\p{^Nv=-0.500}', "");
Expect(0, 3891, '\P{Nv=-0.500}', "");
Expect(1, 3891, '\P{^Nv=-0.500}', "");
Expect(0, 3892, '\p{Nv=-0.500}', "");
Expect(1, 3892, '\p{^Nv=-0.500}', "");
Expect(1, 3892, '\P{Nv=-0.500}', "");
Expect(0, 3892, '\P{^Nv=-0.500}', "");
Error('\p{Is_Numeric_Value=:=--000001/000002}');
Error('\P{Is_Numeric_Value=:=--000001/000002}');
Expect(1, 3891, '\p{Is_Numeric_Value=-00000001/0000000002}', "");
Expect(0, 3891, '\p{^Is_Numeric_Value=-00000001/0000000002}', "");
Expect(0, 3891, '\P{Is_Numeric_Value=-00000001/0000000002}', "");
Expect(1, 3891, '\P{^Is_Numeric_Value=-00000001/0000000002}', "");
Expect(0, 3892, '\p{Is_Numeric_Value=-00000001/0000000002}', "");
Expect(1, 3892, '\p{^Is_Numeric_Value=-00000001/0000000002}', "");
Expect(1, 3892, '\P{Is_Numeric_Value=-00000001/0000000002}', "");
Expect(0, 3892, '\P{^Is_Numeric_Value=-00000001/0000000002}', "");
Expect(1, 3891, '\p{Is_Numeric_Value=-0.500}', "");
Expect(0, 3891, '\p{^Is_Numeric_Value=-0.500}', "");
Expect(0, 3891, '\P{Is_Numeric_Value=-0.500}', "");
Expect(1, 3891, '\P{^Is_Numeric_Value=-0.500}', "");
Expect(0, 3892, '\p{Is_Numeric_Value=-0.500}', "");
Expect(1, 3892, '\p{^Is_Numeric_Value=-0.500}', "");
Expect(1, 3892, '\P{Is_Numeric_Value=-0.500}', "");
Expect(0, 3892, '\P{^Is_Numeric_Value=-0.500}', "");
Error('\p{Is_Nv=-	-0000000001/2:=}');
Error('\P{Is_Nv=-	-0000000001/2:=}');
Expect(1, 3891, '\p{Is_Nv=-0000000001/002}', "");
Expect(0, 3891, '\p{^Is_Nv=-0000000001/002}', "");
Expect(0, 3891, '\P{Is_Nv=-0000000001/002}', "");
Expect(1, 3891, '\P{^Is_Nv=-0000000001/002}', "");
Expect(0, 3892, '\p{Is_Nv=-0000000001/002}', "");
Expect(1, 3892, '\p{^Is_Nv=-0000000001/002}', "");
Expect(1, 3892, '\P{Is_Nv=-0000000001/002}', "");
Expect(0, 3892, '\P{^Is_Nv=-0000000001/002}', "");
Expect(1, 3891, '\p{Is_Nv=-0.500}', "");
Expect(0, 3891, '\p{^Is_Nv=-0.500}', "");
Expect(0, 3891, '\P{Is_Nv=-0.500}', "");
Expect(1, 3891, '\P{^Is_Nv=-0.500}', "");
Expect(0, 3892, '\p{Is_Nv=-0.500}', "");
Expect(1, 3892, '\p{^Is_Nv=-0.500}', "");
Expect(1, 3892, '\P{Is_Nv=-0.500}', "");
Expect(0, 3892, '\P{^Is_Nv=-0.500}', "");
Error('\p{Numeric_Value= :=+000_800_00}');
Error('\P{Numeric_Value= :=+000_800_00}');
Expect(1, 65842, '\p{Numeric_Value=00080000}', "");
Expect(0, 65842, '\p{^Numeric_Value=00080000}', "");
Expect(0, 65842, '\P{Numeric_Value=00080000}', "");
Expect(1, 65842, '\P{^Numeric_Value=00080000}', "");
Expect(0, 65843, '\p{Numeric_Value=00080000}', "");
Expect(1, 65843, '\p{^Numeric_Value=00080000}', "");
Expect(1, 65843, '\P{Numeric_Value=00080000}', "");
Expect(0, 65843, '\P{^Numeric_Value=00080000}', "");
Error('\p{Nv=  000_000_080_000/a/}');
Error('\P{Nv=  000_000_080_000/a/}');
Expect(1, 65842, '\p{Nv=80000}', "");
Expect(0, 65842, '\p{^Nv=80000}', "");
Expect(0, 65842, '\P{Nv=80000}', "");
Expect(1, 65842, '\P{^Nv=80000}', "");
Expect(0, 65843, '\p{Nv=80000}', "");
Expect(1, 65843, '\p{^Nv=80000}', "");
Expect(1, 65843, '\P{Nv=80000}', "");
Expect(0, 65843, '\P{^Nv=80000}', "");
Error('\p{Is_Numeric_Value=/a/-00000_00800_00}');
Error('\P{Is_Numeric_Value=/a/-00000_00800_00}');
Expect(1, 65842, '\p{Is_Numeric_Value=000000080000}', "");
Expect(0, 65842, '\p{^Is_Numeric_Value=000000080000}', "");
Expect(0, 65842, '\P{Is_Numeric_Value=000000080000}', "");
Expect(1, 65842, '\P{^Is_Numeric_Value=000000080000}', "");
Expect(0, 65843, '\p{Is_Numeric_Value=000000080000}', "");
Expect(1, 65843, '\p{^Is_Numeric_Value=000000080000}', "");
Expect(1, 65843, '\P{Is_Numeric_Value=000000080000}', "");
Expect(0, 65843, '\P{^Is_Numeric_Value=000000080000}', "");
Error('\p{Is_Nv=/a/-	000008000_0}');
Error('\P{Is_Nv=/a/-	000008000_0}');
Expect(1, 65842, '\p{Is_Nv: +0_0_8_0_000}', "");
Expect(0, 65842, '\p{^Is_Nv: +0_0_8_0_000}', "");
Expect(0, 65842, '\P{Is_Nv: +0_0_8_0_000}', "");
Expect(1, 65842, '\P{^Is_Nv: +0_0_8_0_000}', "");
Expect(0, 65843, '\p{Is_Nv: +0_0_8_0_000}', "");
Expect(1, 65843, '\p{^Is_Nv: +0_0_8_0_000}', "");
Expect(1, 65843, '\P{Is_Nv: +0_0_8_0_000}', "");
Expect(0, 65843, '\P{^Is_Nv: +0_0_8_0_000}', "");
Error('\p{Numeric_Value=/a/ 0_0_0_0_0_00060}');
Error('\P{Numeric_Value=/a/ 0_0_0_0_0_00060}');
Expect(1, 119662, '\p{Numeric_Value=0060}', "");
Expect(0, 119662, '\p{^Numeric_Value=0060}', "");
Expect(0, 119662, '\P{Numeric_Value=0060}', "");
Expect(1, 119662, '\P{^Numeric_Value=0060}', "");
Expect(0, 119663, '\p{Numeric_Value=0060}', "");
Expect(1, 119663, '\p{^Numeric_Value=0060}', "");
Expect(1, 119663, '\P{Numeric_Value=0060}', "");
Expect(0, 119663, '\P{^Numeric_Value=0060}', "");
Error('\p{Nv= /a/00060}');
Error('\P{Nv= /a/00060}');
Expect(1, 119662, '\p{Nv=+060}', "");
Expect(0, 119662, '\p{^Nv=+060}', "");
Expect(0, 119662, '\P{Nv=+060}', "");
Expect(1, 119662, '\P{^Nv=+060}', "");
Expect(0, 119663, '\p{Nv=+060}', "");
Expect(1, 119663, '\p{^Nv=+060}', "");
Expect(1, 119663, '\P{Nv=+060}', "");
Expect(0, 119663, '\P{^Nv=+060}', "");
Error('\p{Is_Numeric_Value= _+00000060:=}');
Error('\P{Is_Numeric_Value= _+00000060:=}');
Expect(1, 119662, '\p{Is_Numeric_Value=0000000060}', "");
Expect(0, 119662, '\p{^Is_Numeric_Value=0000000060}', "");
Expect(0, 119662, '\P{Is_Numeric_Value=0000000060}', "");
Expect(1, 119662, '\P{^Is_Numeric_Value=0000000060}', "");
Expect(0, 119663, '\p{Is_Numeric_Value=0000000060}', "");
Expect(1, 119663, '\p{^Is_Numeric_Value=0000000060}', "");
Expect(1, 119663, '\P{Is_Numeric_Value=0000000060}', "");
Expect(0, 119663, '\P{^Is_Numeric_Value=0000000060}', "");
Error('\p{Is_Nv: :=_	6_0}');
Error('\P{Is_Nv: :=_	6_0}');
Expect(1, 119662, '\p{Is_Nv:   060}', "");
Expect(0, 119662, '\p{^Is_Nv:   060}', "");
Expect(0, 119662, '\P{Is_Nv:   060}', "");
Expect(1, 119662, '\P{^Is_Nv:   060}', "");
Expect(0, 119663, '\p{Is_Nv:   060}', "");
Expect(1, 119663, '\p{^Is_Nv:   060}', "");
Expect(1, 119663, '\P{Is_Nv:   060}', "");
Expect(0, 119663, '\P{^Is_Nv:   060}', "");
Error('\p{Numeric_Value= 000_004_5/a/}');
Error('\P{Numeric_Value= 000_004_5/a/}');
Expect(1, 12986, '\p{Numeric_Value=0_0_0_0_0_0_00045}', "");
Expect(0, 12986, '\p{^Numeric_Value=0_0_0_0_0_0_00045}', "");
Expect(0, 12986, '\P{Numeric_Value=0_0_0_0_0_0_00045}', "");
Expect(1, 12986, '\P{^Numeric_Value=0_0_0_0_0_0_00045}', "");
Expect(0, 12987, '\p{Numeric_Value=0_0_0_0_0_0_00045}', "");
Expect(1, 12987, '\p{^Numeric_Value=0_0_0_0_0_0_00045}', "");
Expect(1, 12987, '\P{Numeric_Value=0_0_0_0_0_0_00045}', "");
Expect(0, 12987, '\P{^Numeric_Value=0_0_0_0_0_0_00045}', "");
Error('\p{Nv= 	4_5:=}');
Error('\P{Nv= 	4_5:=}');
Expect(1, 12986, '\p{Nv=0_0_0_0_0_0_45}', "");
Expect(0, 12986, '\p{^Nv=0_0_0_0_0_0_45}', "");
Expect(0, 12986, '\P{Nv=0_0_0_0_0_0_45}', "");
Expect(1, 12986, '\P{^Nv=0_0_0_0_0_0_45}', "");
Expect(0, 12987, '\p{Nv=0_0_0_0_0_0_45}', "");
Expect(1, 12987, '\p{^Nv=0_0_0_0_0_0_45}', "");
Expect(1, 12987, '\P{Nv=0_0_0_0_0_0_45}', "");
Expect(0, 12987, '\P{^Nv=0_0_0_0_0_0_45}', "");
Error('\p{Is_Numeric_Value= _45/a/}');
Error('\P{Is_Numeric_Value= _45/a/}');
Expect(1, 12986, '\p{Is_Numeric_Value=+00004_5}', "");
Expect(0, 12986, '\p{^Is_Numeric_Value=+00004_5}', "");
Expect(0, 12986, '\P{Is_Numeric_Value=+00004_5}', "");
Expect(1, 12986, '\P{^Is_Numeric_Value=+00004_5}', "");
Expect(0, 12987, '\p{Is_Numeric_Value=+00004_5}', "");
Expect(1, 12987, '\p{^Is_Numeric_Value=+00004_5}', "");
Expect(1, 12987, '\P{Is_Numeric_Value=+00004_5}', "");
Expect(0, 12987, '\P{^Is_Numeric_Value=+00004_5}', "");
Error('\p{Is_Nv=-45/a/}');
Error('\P{Is_Nv=-45/a/}');
Expect(1, 12986, '\p{Is_Nv=0000045}', "");
Expect(0, 12986, '\p{^Is_Nv=0000045}', "");
Expect(0, 12986, '\P{Is_Nv=0000045}', "");
Expect(1, 12986, '\P{^Is_Nv=0000045}', "");
Expect(0, 12987, '\p{Is_Nv=0000045}', "");
Expect(1, 12987, '\p{^Is_Nv=0000045}', "");
Expect(1, 12987, '\P{Is_Nv=0000045}', "");
Expect(0, 12987, '\P{^Is_Nv=0000045}', "");
Error('\p{Numeric_Value=_0000000019/a/}');
Error('\P{Numeric_Value=_0000000019/a/}');
Expect(1, 9459, '\p{Numeric_Value=0_0_0_0_0_0_19}', "");
Expect(0, 9459, '\p{^Numeric_Value=0_0_0_0_0_0_19}', "");
Expect(0, 9459, '\P{Numeric_Value=0_0_0_0_0_0_19}', "");
Expect(1, 9459, '\P{^Numeric_Value=0_0_0_0_0_0_19}', "");
Expect(0, 9460, '\p{Numeric_Value=0_0_0_0_0_0_19}', "");
Expect(1, 9460, '\p{^Numeric_Value=0_0_0_0_0_0_19}', "");
Expect(1, 9460, '\P{Numeric_Value=0_0_0_0_0_0_19}', "");
Expect(0, 9460, '\P{^Numeric_Value=0_0_0_0_0_0_19}', "");
Error('\p{Nv=	:=01_9}');
Error('\P{Nv=	:=01_9}');
Expect(1, 9459, '\p{Nv=000000019}', "");
Expect(0, 9459, '\p{^Nv=000000019}', "");
Expect(0, 9459, '\P{Nv=000000019}', "");
Expect(1, 9459, '\P{^Nv=000000019}', "");
Expect(0, 9460, '\p{Nv=000000019}', "");
Expect(1, 9460, '\p{^Nv=000000019}', "");
Expect(1, 9460, '\P{Nv=000000019}', "");
Expect(0, 9460, '\P{^Nv=000000019}', "");
Error('\p{Is_Numeric_Value=	 0000000019/a/}');
Error('\P{Is_Numeric_Value=	 0000000019/a/}');
Expect(1, 9459, '\p{Is_Numeric_Value=+000000001_9}', "");
Expect(0, 9459, '\p{^Is_Numeric_Value=+000000001_9}', "");
Expect(0, 9459, '\P{Is_Numeric_Value=+000000001_9}', "");
Expect(1, 9459, '\P{^Is_Numeric_Value=+000000001_9}', "");
Expect(0, 9460, '\p{Is_Numeric_Value=+000000001_9}', "");
Expect(1, 9460, '\p{^Is_Numeric_Value=+000000001_9}', "");
Expect(1, 9460, '\P{Is_Numeric_Value=+000000001_9}', "");
Expect(0, 9460, '\P{^Is_Numeric_Value=+000000001_9}', "");
Error('\p{Is_Nv=--000019/a/}');
Error('\P{Is_Nv=--000019/a/}');
Expect(1, 9459, '\p{Is_Nv=1_9}', "");
Expect(0, 9459, '\p{^Is_Nv=1_9}', "");
Expect(0, 9459, '\P{Is_Nv=1_9}', "");
Expect(1, 9459, '\P{^Is_Nv=1_9}', "");
Expect(0, 9460, '\p{Is_Nv=1_9}', "");
Expect(1, 9460, '\p{^Is_Nv=1_9}', "");
Expect(1, 9460, '\P{Is_Nv=1_9}', "");
Expect(0, 9460, '\P{^Is_Nv=1_9}', "");
Error('\p{Numeric_Value=-0000000011/02:=}');
Error('\P{Numeric_Value=-0000000011/02:=}');
Expect(1, 3887, '\p{Numeric_Value=011/2}', "");
Expect(0, 3887, '\p{^Numeric_Value=011/2}', "");
Expect(0, 3887, '\P{Numeric_Value=011/2}', "");
Expect(1, 3887, '\P{^Numeric_Value=011/2}', "");
Expect(0, 3888, '\p{Numeric_Value=011/2}', "");
Expect(1, 3888, '\p{^Numeric_Value=011/2}', "");
Expect(1, 3888, '\P{Numeric_Value=011/2}', "");
Expect(0, 3888, '\P{^Numeric_Value=011/2}', "");
Expect(1, 3887, '\p{Numeric_Value=5.500}', "");
Expect(0, 3887, '\p{^Numeric_Value=5.500}', "");
Expect(0, 3887, '\P{Numeric_Value=5.500}', "");
Expect(1, 3887, '\P{^Numeric_Value=5.500}', "");
Expect(0, 3888, '\p{Numeric_Value=5.500}', "");
Expect(1, 3888, '\p{^Numeric_Value=5.500}', "");
Expect(1, 3888, '\P{Numeric_Value=5.500}', "");
Expect(0, 3888, '\P{^Numeric_Value=5.500}', "");
Error('\p{Nv=/a/	+11/000002}');
Error('\P{Nv=/a/	+11/000002}');
Expect(1, 3887, '\p{Nv=00011/2}', "");
Expect(0, 3887, '\p{^Nv=00011/2}', "");
Expect(0, 3887, '\P{Nv=00011/2}', "");
Expect(1, 3887, '\P{^Nv=00011/2}', "");
Expect(0, 3888, '\p{Nv=00011/2}', "");
Expect(1, 3888, '\p{^Nv=00011/2}', "");
Expect(1, 3888, '\P{Nv=00011/2}', "");
Expect(0, 3888, '\P{^Nv=00011/2}', "");
Expect(1, 3887, '\p{Nv=5.500}', "");
Expect(0, 3887, '\p{^Nv=5.500}', "");
Expect(0, 3887, '\P{Nv=5.500}', "");
Expect(1, 3887, '\P{^Nv=5.500}', "");
Expect(0, 3888, '\p{Nv=5.500}', "");
Expect(1, 3888, '\p{^Nv=5.500}', "");
Expect(1, 3888, '\P{Nv=5.500}', "");
Expect(0, 3888, '\P{^Nv=5.500}', "");
Error('\p{Is_Numeric_Value=:=	-+0000011/0000000002}');
Error('\P{Is_Numeric_Value=:=	-+0000011/0000000002}');
Expect(1, 3887, '\p{Is_Numeric_Value=00011/02}', "");
Expect(0, 3887, '\p{^Is_Numeric_Value=00011/02}', "");
Expect(0, 3887, '\P{Is_Numeric_Value=00011/02}', "");
Expect(1, 3887, '\P{^Is_Numeric_Value=00011/02}', "");
Expect(0, 3888, '\p{Is_Numeric_Value=00011/02}', "");
Expect(1, 3888, '\p{^Is_Numeric_Value=00011/02}', "");
Expect(1, 3888, '\P{Is_Numeric_Value=00011/02}', "");
Expect(0, 3888, '\P{^Is_Numeric_Value=00011/02}', "");
Expect(1, 3887, '\p{Is_Numeric_Value=5.500}', "");
Expect(0, 3887, '\p{^Is_Numeric_Value=5.500}', "");
Expect(0, 3887, '\P{Is_Numeric_Value=5.500}', "");
Expect(1, 3887, '\P{^Is_Numeric_Value=5.500}', "");
Expect(0, 3888, '\p{Is_Numeric_Value=5.500}', "");
Expect(1, 3888, '\p{^Is_Numeric_Value=5.500}', "");
Expect(1, 3888, '\P{Is_Numeric_Value=5.500}', "");
Expect(0, 3888, '\P{^Is_Numeric_Value=5.500}', "");
Error('\p{Is_Nv=	_00011/0000002/a/}');
Error('\P{Is_Nv=	_00011/0000002/a/}');
Expect(1, 3887, '\p{Is_Nv=+00000000011/002}', "");
Expect(0, 3887, '\p{^Is_Nv=+00000000011/002}', "");
Expect(0, 3887, '\P{Is_Nv=+00000000011/002}', "");
Expect(1, 3887, '\P{^Is_Nv=+00000000011/002}', "");
Expect(0, 3888, '\p{Is_Nv=+00000000011/002}', "");
Expect(1, 3888, '\p{^Is_Nv=+00000000011/002}', "");
Expect(1, 3888, '\P{Is_Nv=+00000000011/002}', "");
Expect(0, 3888, '\P{^Is_Nv=+00000000011/002}', "");
Expect(1, 3887, '\p{Is_Nv=5.500}', "");
Expect(0, 3887, '\p{^Is_Nv=5.500}', "");
Expect(0, 3887, '\P{Is_Nv=5.500}', "");
Expect(1, 3887, '\P{^Is_Nv=5.500}', "");
Expect(0, 3888, '\p{Is_Nv=5.500}', "");
Expect(1, 3888, '\p{^Is_Nv=5.500}', "");
Expect(1, 3888, '\P{Is_Nv=5.500}', "");
Expect(0, 3888, '\P{^Is_Nv=5.500}', "");
Error('\p{Numeric_Value=/a/ 0007_0}');
Error('\P{Numeric_Value=/a/ 0007_0}');
Expect(1, 119663, '\p{Numeric_Value=00000000070}', "");
Expect(0, 119663, '\p{^Numeric_Value=00000000070}', "");
Expect(0, 119663, '\P{Numeric_Value=00000000070}', "");
Expect(1, 119663, '\P{^Numeric_Value=00000000070}', "");
Expect(0, 119664, '\p{Numeric_Value=00000000070}', "");
Expect(1, 119664, '\p{^Numeric_Value=00000000070}', "");
Expect(1, 119664, '\P{Numeric_Value=00000000070}', "");
Expect(0, 119664, '\P{^Numeric_Value=00000000070}', "");
Error('\p{Nv=- +000007_0:=}');
Error('\P{Nv=- +000007_0:=}');
Expect(1, 119663, '\p{Nv=+007_0}', "");
Expect(0, 119663, '\p{^Nv=+007_0}', "");
Expect(0, 119663, '\P{Nv=+007_0}', "");
Expect(1, 119663, '\P{^Nv=+007_0}', "");
Expect(0, 119664, '\p{Nv=+007_0}', "");
Expect(1, 119664, '\p{^Nv=+007_0}', "");
Expect(1, 119664, '\P{Nv=+007_0}', "");
Expect(0, 119664, '\P{^Nv=+007_0}', "");
Error('\p{Is_Numeric_Value=_:=00000000070}');
Error('\P{Is_Numeric_Value=_:=00000000070}');
Expect(1, 119663, '\p{Is_Numeric_Value=7_0}', "");
Expect(0, 119663, '\p{^Is_Numeric_Value=7_0}', "");
Expect(0, 119663, '\P{Is_Numeric_Value=7_0}', "");
Expect(1, 119663, '\P{^Is_Numeric_Value=7_0}', "");
Expect(0, 119664, '\p{Is_Numeric_Value=7_0}', "");
Expect(1, 119664, '\p{^Is_Numeric_Value=7_0}', "");
Expect(1, 119664, '\P{Is_Numeric_Value=7_0}', "");
Expect(0, 119664, '\P{^Is_Numeric_Value=7_0}', "");
Error('\p{Is_Nv=-:=00_00_00_00_070}');
Error('\P{Is_Nv=-:=00_00_00_00_070}');
Expect(1, 119663, '\p{Is_Nv=000_007_0}', "");
Expect(0, 119663, '\p{^Is_Nv=000_007_0}', "");
Expect(0, 119663, '\P{Is_Nv=000_007_0}', "");
Expect(1, 119663, '\P{^Is_Nv=000_007_0}', "");
Expect(0, 119664, '\p{Is_Nv=000_007_0}', "");
Expect(1, 119664, '\p{^Is_Nv=000_007_0}', "");
Expect(1, 119664, '\P{Is_Nv=000_007_0}', "");
Expect(0, 119664, '\P{^Is_Nv=000_007_0}', "");
Error('\p{Numeric_Value=:= _0000_0000_02}');
Error('\P{Numeric_Value=:= _0000_0000_02}');
Expect(1, 140176, '\p{Numeric_Value:+000002}', "");
Expect(0, 140176, '\p{^Numeric_Value:+000002}', "");
Expect(0, 140176, '\P{Numeric_Value:+000002}', "");
Expect(1, 140176, '\P{^Numeric_Value:+000002}', "");
Expect(0, 140177, '\p{Numeric_Value:+000002}', "");
Expect(1, 140177, '\p{^Numeric_Value:+000002}', "");
Expect(1, 140177, '\P{Numeric_Value:+000002}', "");
Expect(0, 140177, '\P{^Numeric_Value:+000002}', "");
Error('\p{Nv= 02/a/}');
Error('\P{Nv= 02/a/}');
Expect(1, 140176, '\p{Nv=000_2}', "");
Expect(0, 140176, '\p{^Nv=000_2}', "");
Expect(0, 140176, '\P{Nv=000_2}', "");
Expect(1, 140176, '\P{^Nv=000_2}', "");
Expect(0, 140177, '\p{Nv=000_2}', "");
Expect(1, 140177, '\p{^Nv=000_2}', "");
Expect(1, 140177, '\P{Nv=000_2}', "");
Expect(0, 140177, '\P{^Nv=000_2}', "");
Error('\p{Is_Numeric_Value=_:=000000_2}');
Error('\P{Is_Numeric_Value=_:=000000_2}');
Expect(1, 140176, '\p{Is_Numeric_Value:	0_0_0_0_0_0_002}', "");
Expect(0, 140176, '\p{^Is_Numeric_Value:	0_0_0_0_0_0_002}', "");
Expect(0, 140176, '\P{Is_Numeric_Value:	0_0_0_0_0_0_002}', "");
Expect(1, 140176, '\P{^Is_Numeric_Value:	0_0_0_0_0_0_002}', "");
Expect(0, 140177, '\p{Is_Numeric_Value:	0_0_0_0_0_0_002}', "");
Expect(1, 140177, '\p{^Is_Numeric_Value:	0_0_0_0_0_0_002}', "");
Expect(1, 140177, '\P{Is_Numeric_Value:	0_0_0_0_0_0_002}', "");
Expect(0, 140177, '\P{^Is_Numeric_Value:	0_0_0_0_0_0_002}', "");
Error('\p{Is_Nv= _2/a/}');
Error('\P{Is_Nv= _2/a/}');
Expect(1, 140176, '\p{Is_Nv:   0_0_0_02}', "");
Expect(0, 140176, '\p{^Is_Nv:   0_0_0_02}', "");
Expect(0, 140176, '\P{Is_Nv:   0_0_0_02}', "");
Expect(1, 140176, '\P{^Is_Nv:   0_0_0_02}', "");
Expect(0, 140177, '\p{Is_Nv:   0_0_0_02}', "");
Expect(1, 140177, '\p{^Is_Nv:   0_0_0_02}', "");
Expect(1, 140177, '\P{Is_Nv:   0_0_0_02}', "");
Expect(0, 140177, '\P{^Is_Nv:   0_0_0_02}', "");
Error('\p{Numeric_Value:	 _017/a/}');
Error('\P{Numeric_Value:	 _017/a/}');
Expect(1, 9457, '\p{Numeric_Value=0_0_0_0_0_0_0_017}', "");
Expect(0, 9457, '\p{^Numeric_Value=0_0_0_0_0_0_0_017}', "");
Expect(0, 9457, '\P{Numeric_Value=0_0_0_0_0_0_0_017}', "");
Expect(1, 9457, '\P{^Numeric_Value=0_0_0_0_0_0_0_017}', "");
Expect(0, 9458, '\p{Numeric_Value=0_0_0_0_0_0_0_017}', "");
Expect(1, 9458, '\p{^Numeric_Value=0_0_0_0_0_0_0_017}', "");
Expect(1, 9458, '\P{Numeric_Value=0_0_0_0_0_0_0_017}', "");
Expect(0, 9458, '\P{^Numeric_Value=0_0_0_0_0_0_0_017}', "");
Error('\p{Nv= -01_7:=}');
Error('\P{Nv= -01_7:=}');
Expect(1, 9457, '\p{Nv=+0_0_0_017}', "");
Expect(0, 9457, '\p{^Nv=+0_0_0_017}', "");
Expect(0, 9457, '\P{Nv=+0_0_0_017}', "");
Expect(1, 9457, '\P{^Nv=+0_0_0_017}', "");
Expect(0, 9458, '\p{Nv=+0_0_0_017}', "");
Expect(1, 9458, '\p{^Nv=+0_0_0_017}', "");
Expect(1, 9458, '\P{Nv=+0_0_0_017}', "");
Expect(0, 9458, '\P{^Nv=+0_0_0_017}', "");
Error('\p{Is_Numeric_Value=-:=+00000017}');
Error('\P{Is_Numeric_Value=-:=+00000017}');
Expect(1, 9457, '\p{Is_Numeric_Value=00017}', "");
Expect(0, 9457, '\p{^Is_Numeric_Value=00017}', "");
Expect(0, 9457, '\P{Is_Numeric_Value=00017}', "");
Expect(1, 9457, '\P{^Is_Numeric_Value=00017}', "");
Expect(0, 9458, '\p{Is_Numeric_Value=00017}', "");
Expect(1, 9458, '\p{^Is_Numeric_Value=00017}', "");
Expect(1, 9458, '\P{Is_Numeric_Value=00017}', "");
Expect(0, 9458, '\P{^Is_Numeric_Value=00017}', "");
Error('\p{Is_Nv= /a/0001_7}');
Error('\P{Is_Nv= /a/0001_7}');
Expect(1, 9457, '\p{Is_Nv=00001_7}', "");
Expect(0, 9457, '\p{^Is_Nv=00001_7}', "");
Expect(0, 9457, '\P{Is_Nv=00001_7}', "");
Expect(1, 9457, '\P{^Is_Nv=00001_7}', "");
Expect(0, 9458, '\p{Is_Nv=00001_7}', "");
Expect(1, 9458, '\p{^Is_Nv=00001_7}', "");
Expect(1, 9458, '\P{Is_Nv=00001_7}', "");
Expect(0, 9458, '\P{^Is_Nv=00001_7}', "");
Error('\p{Numeric_Value= _00_00_00_20_00_0:=}');
Error('\P{Numeric_Value= _00_00_00_20_00_0:=}');
Expect(1, 65836, '\p{Numeric_Value=20000}', "");
Expect(0, 65836, '\p{^Numeric_Value=20000}', "");
Expect(0, 65836, '\P{Numeric_Value=20000}', "");
Expect(1, 65836, '\P{^Numeric_Value=20000}', "");
Expect(0, 65837, '\p{Numeric_Value=20000}', "");
Expect(1, 65837, '\p{^Numeric_Value=20000}', "");
Expect(1, 65837, '\P{Numeric_Value=20000}', "");
Expect(0, 65837, '\P{^Numeric_Value=20000}', "");
Error('\p{Nv=:= -00002000_0}');
Error('\P{Nv=:= -00002000_0}');
Expect(1, 65836, '\p{Nv=+00000020000}', "");
Expect(0, 65836, '\p{^Nv=+00000020000}', "");
Expect(0, 65836, '\P{Nv=+00000020000}', "");
Expect(1, 65836, '\P{^Nv=+00000020000}', "");
Expect(0, 65837, '\p{Nv=+00000020000}', "");
Expect(1, 65837, '\p{^Nv=+00000020000}', "");
Expect(1, 65837, '\P{Nv=+00000020000}', "");
Expect(0, 65837, '\P{^Nv=+00000020000}', "");
Error('\p{Is_Numeric_Value= :=0020000}');
Error('\P{Is_Numeric_Value= :=0020000}');
Expect(1, 65836, '\p{Is_Numeric_Value=+020000}', "");
Expect(0, 65836, '\p{^Is_Numeric_Value=+020000}', "");
Expect(0, 65836, '\P{Is_Numeric_Value=+020000}', "");
Expect(1, 65836, '\P{^Is_Numeric_Value=+020000}', "");
Expect(0, 65837, '\p{Is_Numeric_Value=+020000}', "");
Expect(1, 65837, '\p{^Is_Numeric_Value=+020000}', "");
Expect(1, 65837, '\P{Is_Numeric_Value=+020000}', "");
Expect(0, 65837, '\P{^Is_Numeric_Value=+020000}', "");
Error('\p{Is_Nv::=__00000020000}');
Error('\P{Is_Nv::=__00000020000}');
Expect(1, 65836, '\p{Is_Nv=+000000020000}', "");
Expect(0, 65836, '\p{^Is_Nv=+000000020000}', "");
Expect(0, 65836, '\P{Is_Nv=+000000020000}', "");
Expect(1, 65836, '\P{^Is_Nv=+000000020000}', "");
Expect(0, 65837, '\p{Is_Nv=+000000020000}', "");
Expect(1, 65837, '\p{^Is_Nv=+000000020000}', "");
Expect(1, 65837, '\P{Is_Nv=+000000020000}', "");
Expect(0, 65837, '\P{^Is_Nv=+000000020000}', "");
Error('\p{Numeric_Value=	_0001:=}');
Error('\P{Numeric_Value=	_0001:=}');
Expect(1, 133418, '\p{Numeric_Value=0_0_0_0_0001}', "");
Expect(0, 133418, '\p{^Numeric_Value=0_0_0_0_0001}', "");
Expect(0, 133418, '\P{Numeric_Value=0_0_0_0_0001}', "");
Expect(1, 133418, '\P{^Numeric_Value=0_0_0_0_0001}', "");
Expect(0, 133419, '\p{Numeric_Value=0_0_0_0_0001}', "");
Expect(1, 133419, '\p{^Numeric_Value=0_0_0_0_0001}', "");
Expect(1, 133419, '\P{Numeric_Value=0_0_0_0_0001}', "");
Expect(0, 133419, '\P{^Numeric_Value=0_0_0_0_0001}', "");
Error('\p{Nv=-:=000000001}');
Error('\P{Nv=-:=000000001}');
Expect(1, 133418, '\p{Nv=00_00_00_01}', "");
Expect(0, 133418, '\p{^Nv=00_00_00_01}', "");
Expect(0, 133418, '\P{Nv=00_00_00_01}', "");
Expect(1, 133418, '\P{^Nv=00_00_00_01}', "");
Expect(0, 133419, '\p{Nv=00_00_00_01}', "");
Expect(1, 133419, '\p{^Nv=00_00_00_01}', "");
Expect(1, 133419, '\P{Nv=00_00_00_01}', "");
Expect(0, 133419, '\P{^Nv=00_00_00_01}', "");
Error('\p{Is_Numeric_Value=-	000000_1:=}');
Error('\P{Is_Numeric_Value=-	000000_1:=}');
Expect(1, 133418, '\p{Is_Numeric_Value=+1}', "");
Expect(0, 133418, '\p{^Is_Numeric_Value=+1}', "");
Expect(0, 133418, '\P{Is_Numeric_Value=+1}', "");
Expect(1, 133418, '\P{^Is_Numeric_Value=+1}', "");
Expect(0, 133419, '\p{Is_Numeric_Value=+1}', "");
Expect(1, 133419, '\p{^Is_Numeric_Value=+1}', "");
Expect(1, 133419, '\P{Is_Numeric_Value=+1}', "");
Expect(0, 133419, '\P{^Is_Numeric_Value=+1}', "");
Error('\p{Is_Nv=-/a/+0_0_0_0_0_01}');
Error('\P{Is_Nv=-/a/+0_0_0_0_0_01}');
Expect(1, 133418, '\p{Is_Nv=0001}', "");
Expect(0, 133418, '\p{^Is_Nv=0001}', "");
Expect(0, 133418, '\P{Is_Nv=0001}', "");
Expect(1, 133418, '\P{^Is_Nv=0001}', "");
Expect(0, 133419, '\p{Is_Nv=0001}', "");
Expect(1, 133419, '\p{^Is_Nv=0001}', "");
Expect(1, 133419, '\P{Is_Nv=0001}', "");
Expect(0, 133419, '\P{^Is_Nv=0001}', "");
Error('\p{Numeric_Value=-:=0006000}');
Error('\P{Numeric_Value=-:=0006000}');
Expect(1, 65831, '\p{Numeric_Value=+06000}', "");
Expect(0, 65831, '\p{^Numeric_Value=+06000}', "");
Expect(0, 65831, '\P{Numeric_Value=+06000}', "");
Expect(1, 65831, '\P{^Numeric_Value=+06000}', "");
Expect(0, 65832, '\p{Numeric_Value=+06000}', "");
Expect(1, 65832, '\p{^Numeric_Value=+06000}', "");
Expect(1, 65832, '\P{Numeric_Value=+06000}', "");
Expect(0, 65832, '\P{^Numeric_Value=+06000}', "");
Error('\p{Nv=/a/	_000006000}');
Error('\P{Nv=/a/	_000006000}');
Expect(1, 65831, '\p{Nv=0000006000}', "");
Expect(0, 65831, '\p{^Nv=0000006000}', "");
Expect(0, 65831, '\P{Nv=0000006000}', "");
Expect(1, 65831, '\P{^Nv=0000006000}', "");
Expect(0, 65832, '\p{Nv=0000006000}', "");
Expect(1, 65832, '\p{^Nv=0000006000}', "");
Expect(1, 65832, '\P{Nv=0000006000}', "");
Expect(0, 65832, '\P{^Nv=0000006000}', "");
Error('\p{Is_Numeric_Value=:=	 +00006000}');
Error('\P{Is_Numeric_Value=:=	 +00006000}');
Expect(1, 65831, '\p{Is_Numeric_Value:   0000000006000}', "");
Expect(0, 65831, '\p{^Is_Numeric_Value:   0000000006000}', "");
Expect(0, 65831, '\P{Is_Numeric_Value:   0000000006000}', "");
Expect(1, 65831, '\P{^Is_Numeric_Value:   0000000006000}', "");
Expect(0, 65832, '\p{Is_Numeric_Value:   0000000006000}', "");
Expect(1, 65832, '\p{^Is_Numeric_Value:   0000000006000}', "");
Expect(1, 65832, '\P{Is_Numeric_Value:   0000000006000}', "");
Expect(0, 65832, '\P{^Is_Numeric_Value:   0000000006000}', "");
Error('\p{Is_Nv= /a/+000000006000}');
Error('\P{Is_Nv= /a/+000000006000}');
Expect(1, 65831, '\p{Is_Nv=000_000_060_00}', "");
Expect(0, 65831, '\p{^Is_Nv=000_000_060_00}', "");
Expect(0, 65831, '\P{Is_Nv=000_000_060_00}', "");
Expect(1, 65831, '\P{^Is_Nv=000_000_060_00}', "");
Expect(0, 65832, '\p{Is_Nv=000_000_060_00}', "");
Expect(1, 65832, '\p{^Is_Nv=000_000_060_00}', "");
Expect(1, 65832, '\P{Is_Nv=000_000_060_00}', "");
Expect(0, 65832, '\P{^Is_Nv=000_000_060_00}', "");
Error('\p{Numeric_Value=	-0_0_0_0_0_0_0030:=}');
Error('\P{Numeric_Value=	-0_0_0_0_0_0_0030:=}');
Expect(1, 133507, '\p{Numeric_Value=3_0}', "");
Expect(0, 133507, '\p{^Numeric_Value=3_0}', "");
Expect(0, 133507, '\P{Numeric_Value=3_0}', "");
Expect(1, 133507, '\P{^Numeric_Value=3_0}', "");
Expect(0, 133508, '\p{Numeric_Value=3_0}', "");
Expect(1, 133508, '\p{^Numeric_Value=3_0}', "");
Expect(1, 133508, '\P{Numeric_Value=3_0}', "");
Expect(0, 133508, '\P{^Numeric_Value=3_0}', "");
Error('\p{Nv=	/a/000_000_000_30}');
Error('\P{Nv=	/a/000_000_000_30}');
Expect(1, 133507, '\p{Nv=00_00_03_0}', "");
Expect(0, 133507, '\p{^Nv=00_00_03_0}', "");
Expect(0, 133507, '\P{Nv=00_00_03_0}', "");
Expect(1, 133507, '\P{^Nv=00_00_03_0}', "");
Expect(0, 133508, '\p{Nv=00_00_03_0}', "");
Expect(1, 133508, '\p{^Nv=00_00_03_0}', "");
Expect(1, 133508, '\P{Nv=00_00_03_0}', "");
Expect(0, 133508, '\P{^Nv=00_00_03_0}', "");
Error('\p{Is_Numeric_Value=:=-00000030}');
Error('\P{Is_Numeric_Value=:=-00000030}');
Expect(1, 133507, '\p{Is_Numeric_Value:00000030}', "");
Expect(0, 133507, '\p{^Is_Numeric_Value:00000030}', "");
Expect(0, 133507, '\P{Is_Numeric_Value:00000030}', "");
Expect(1, 133507, '\P{^Is_Numeric_Value:00000030}', "");
Expect(0, 133508, '\p{Is_Numeric_Value:00000030}', "");
Expect(1, 133508, '\p{^Is_Numeric_Value:00000030}', "");
Expect(1, 133508, '\P{Is_Numeric_Value:00000030}', "");
Expect(0, 133508, '\P{^Is_Numeric_Value:00000030}', "");
Error('\p{Is_Nv=	/a/00_03_0}');
Error('\P{Is_Nv=	/a/00_03_0}');
Expect(1, 133507, '\p{Is_Nv=+0000_0000_30}', "");
Expect(0, 133507, '\p{^Is_Nv=+0000_0000_30}', "");
Expect(0, 133507, '\P{Is_Nv=+0000_0000_30}', "");
Expect(1, 133507, '\P{^Is_Nv=+0000_0000_30}', "");
Expect(0, 133508, '\p{Is_Nv=+0000_0000_30}', "");
Expect(1, 133508, '\p{^Is_Nv=+0000_0000_30}', "");
Expect(1, 133508, '\P{Is_Nv=+0000_0000_30}', "");
Expect(0, 133508, '\P{^Is_Nv=+0000_0000_30}', "");
Error('\p{Numeric_Value=:=-00002/003}');
Error('\P{Numeric_Value=:=-00002/003}');
Expect(1, 74846, '\p{Numeric_Value=000000002/00000003}', "");
Expect(0, 74846, '\p{^Numeric_Value=000000002/00000003}', "");
Expect(0, 74846, '\P{Numeric_Value=000000002/00000003}', "");
Expect(1, 74846, '\P{^Numeric_Value=000000002/00000003}', "");
Expect(0, 74847, '\p{Numeric_Value=000000002/00000003}', "");
Expect(1, 74847, '\p{^Numeric_Value=000000002/00000003}', "");
Expect(1, 74847, '\P{Numeric_Value=000000002/00000003}', "");
Expect(0, 74847, '\P{^Numeric_Value=000000002/00000003}', "");
Error('\p{Numeric_Value=0.7}');
Error('\P{Numeric_Value=0.7}');
Error('\p{Numeric_Value:0.67}');
Error('\P{Numeric_Value:0.67}');
Expect(1, 74846, '\p{Numeric_Value=0.667}', "");
Expect(0, 74846, '\p{^Numeric_Value=0.667}', "");
Expect(0, 74846, '\P{Numeric_Value=0.667}', "");
Expect(1, 74846, '\P{^Numeric_Value=0.667}', "");
Expect(0, 74847, '\p{Numeric_Value=0.667}', "");
Expect(1, 74847, '\p{^Numeric_Value=0.667}', "");
Expect(1, 74847, '\P{Numeric_Value=0.667}', "");
Expect(0, 74847, '\P{^Numeric_Value=0.667}', "");
Error('\p{Nv=/a/ 000002/003}');
Error('\P{Nv=/a/ 000002/003}');
Expect(1, 74846, '\p{Nv=000002/0000003}', "");
Expect(0, 74846, '\p{^Nv=000002/0000003}', "");
Expect(0, 74846, '\P{Nv=000002/0000003}', "");
Expect(1, 74846, '\P{^Nv=000002/0000003}', "");
Expect(0, 74847, '\p{Nv=000002/0000003}', "");
Expect(1, 74847, '\p{^Nv=000002/0000003}', "");
Expect(1, 74847, '\P{Nv=000002/0000003}', "");
Expect(0, 74847, '\P{^Nv=000002/0000003}', "");
Error('\p{Nv=0.7}');
Error('\P{Nv=0.7}');
Error('\p{Nv=0.67}');
Error('\P{Nv=0.67}');
Expect(1, 74846, '\p{Nv=0.667}', "");
Expect(0, 74846, '\p{^Nv=0.667}', "");
Expect(0, 74846, '\P{Nv=0.667}', "");
Expect(1, 74846, '\P{^Nv=0.667}', "");
Expect(0, 74847, '\p{Nv=0.667}', "");
Expect(1, 74847, '\p{^Nv=0.667}', "");
Expect(1, 74847, '\P{Nv=0.667}', "");
Expect(0, 74847, '\P{^Nv=0.667}', "");
Error('\p{Is_Numeric_Value= 	+000002/3/a/}');
Error('\P{Is_Numeric_Value= 	+000002/3/a/}');
Expect(1, 74846, '\p{Is_Numeric_Value:   +00000002/0003}', "");
Expect(0, 74846, '\p{^Is_Numeric_Value:   +00000002/0003}', "");
Expect(0, 74846, '\P{Is_Numeric_Value:   +00000002/0003}', "");
Expect(1, 74846, '\P{^Is_Numeric_Value:   +00000002/0003}', "");
Expect(0, 74847, '\p{Is_Numeric_Value:   +00000002/0003}', "");
Expect(1, 74847, '\p{^Is_Numeric_Value:   +00000002/0003}', "");
Expect(1, 74847, '\P{Is_Numeric_Value:   +00000002/0003}', "");
Expect(0, 74847, '\P{^Is_Numeric_Value:   +00000002/0003}', "");
Error('\p{Is_Numeric_Value=0.7}');
Error('\P{Is_Numeric_Value=0.7}');
Error('\p{Is_Numeric_Value=0.67}');
Error('\P{Is_Numeric_Value=0.67}');
Expect(1, 74846, '\p{Is_Numeric_Value=0.667}', "");
Expect(0, 74846, '\p{^Is_Numeric_Value=0.667}', "");
Expect(0, 74846, '\P{Is_Numeric_Value=0.667}', "");
Expect(1, 74846, '\P{^Is_Numeric_Value=0.667}', "");
Expect(0, 74847, '\p{Is_Numeric_Value=0.667}', "");
Expect(1, 74847, '\p{^Is_Numeric_Value=0.667}', "");
Expect(1, 74847, '\P{Is_Numeric_Value=0.667}', "");
Expect(0, 74847, '\P{^Is_Numeric_Value=0.667}', "");
Error('\p{Is_Nv:/a/_+000000002/0003}');
Error('\P{Is_Nv:/a/_+000000002/0003}');
Expect(1, 74846, '\p{Is_Nv=+000000002/0003}', "");
Expect(0, 74846, '\p{^Is_Nv=+000000002/0003}', "");
Expect(0, 74846, '\P{Is_Nv=+000000002/0003}', "");
Expect(1, 74846, '\P{^Is_Nv=+000000002/0003}', "");
Expect(0, 74847, '\p{Is_Nv=+000000002/0003}', "");
Expect(1, 74847, '\p{^Is_Nv=+000000002/0003}', "");
Expect(1, 74847, '\P{Is_Nv=+000000002/0003}', "");
Expect(0, 74847, '\P{^Is_Nv=+000000002/0003}', "");
Error('\p{Is_Nv:0.7}');
Error('\P{Is_Nv:0.7}');
Error('\p{Is_Nv=0.67}');
Error('\P{Is_Nv=0.67}');
Expect(1, 74846, '\p{Is_Nv=0.667}', "");
Expect(0, 74846, '\p{^Is_Nv=0.667}', "");
Expect(0, 74846, '\P{Is_Nv=0.667}', "");
Expect(1, 74846, '\P{^Is_Nv=0.667}', "");
Expect(0, 74847, '\p{Is_Nv=0.667}', "");
Expect(1, 74847, '\p{^Is_Nv=0.667}', "");
Expect(1, 74847, '\P{Is_Nv=0.667}', "");
Expect(0, 74847, '\P{^Is_Nv=0.667}', "");
Error('\p{Numeric_Value=	/a/00_00_00_04_000}');
Error('\P{Numeric_Value=	/a/00_00_00_04_000}');
Expect(1, 65829, '\p{Numeric_Value=00000400_0}', "");
Expect(0, 65829, '\p{^Numeric_Value=00000400_0}', "");
Expect(0, 65829, '\P{Numeric_Value=00000400_0}', "");
Expect(1, 65829, '\P{^Numeric_Value=00000400_0}', "");
Expect(0, 65830, '\p{Numeric_Value=00000400_0}', "");
Expect(1, 65830, '\p{^Numeric_Value=00000400_0}', "");
Expect(1, 65830, '\P{Numeric_Value=00000400_0}', "");
Expect(0, 65830, '\P{^Numeric_Value=00000400_0}', "");
Error('\p{Nv=_/a/04000}');
Error('\P{Nv=_/a/04000}');
Expect(1, 65829, '\p{Nv=004000}', "");
Expect(0, 65829, '\p{^Nv=004000}', "");
Expect(0, 65829, '\P{Nv=004000}', "");
Expect(1, 65829, '\P{^Nv=004000}', "");
Expect(0, 65830, '\p{Nv=004000}', "");
Expect(1, 65830, '\p{^Nv=004000}', "");
Expect(1, 65830, '\P{Nv=004000}', "");
Expect(0, 65830, '\P{^Nv=004000}', "");
Error('\p{Is_Numeric_Value=-	+004000:=}');
Error('\P{Is_Numeric_Value=-	+004000:=}');
Expect(1, 65829, '\p{Is_Numeric_Value=0000004000}', "");
Expect(0, 65829, '\p{^Is_Numeric_Value=0000004000}', "");
Expect(0, 65829, '\P{Is_Numeric_Value=0000004000}', "");
Expect(1, 65829, '\P{^Is_Numeric_Value=0000004000}', "");
Expect(0, 65830, '\p{Is_Numeric_Value=0000004000}', "");
Expect(1, 65830, '\p{^Is_Numeric_Value=0000004000}', "");
Expect(1, 65830, '\P{Is_Numeric_Value=0000004000}', "");
Expect(0, 65830, '\P{^Is_Numeric_Value=0000004000}', "");
Error('\p{Is_Nv:    /a/00400_0}');
Error('\P{Is_Nv:    /a/00400_0}');
Expect(1, 65829, '\p{Is_Nv=004000}', "");
Expect(0, 65829, '\p{^Is_Nv=004000}', "");
Expect(0, 65829, '\P{Is_Nv=004000}', "");
Expect(1, 65829, '\P{^Is_Nv=004000}', "");
Expect(0, 65830, '\p{Is_Nv=004000}', "");
Expect(1, 65830, '\p{^Is_Nv=004000}', "");
Expect(1, 65830, '\P{Is_Nv=004000}', "");
Expect(0, 65830, '\P{^Is_Nv=004000}', "");
Error('\p{Numeric_Value= 	00000000100:=}');
Error('\P{Numeric_Value= 	00000000100:=}');
Expect(1, 69732, '\p{Numeric_Value=0_0_0_0_0_0_00100}', "");
Expect(0, 69732, '\p{^Numeric_Value=0_0_0_0_0_0_00100}', "");
Expect(0, 69732, '\P{Numeric_Value=0_0_0_0_0_0_00100}', "");
Expect(1, 69732, '\P{^Numeric_Value=0_0_0_0_0_0_00100}', "");
Expect(0, 69733, '\p{Numeric_Value=0_0_0_0_0_0_00100}', "");
Expect(1, 69733, '\p{^Numeric_Value=0_0_0_0_0_0_00100}', "");
Expect(1, 69733, '\P{Numeric_Value=0_0_0_0_0_0_00100}', "");
Expect(0, 69733, '\P{^Numeric_Value=0_0_0_0_0_0_00100}', "");
Error('\p{Nv=:=- 0_1_00}');
Error('\P{Nv=:=- 0_1_00}');
Expect(1, 69732, '\p{Nv=00100}', "");
Expect(0, 69732, '\p{^Nv=00100}', "");
Expect(0, 69732, '\P{Nv=00100}', "");
Expect(1, 69732, '\P{^Nv=00100}', "");
Expect(0, 69733, '\p{Nv=00100}', "");
Expect(1, 69733, '\p{^Nv=00100}', "");
Expect(1, 69733, '\P{Nv=00100}', "");
Expect(0, 69733, '\P{^Nv=00100}', "");
Error('\p{Is_Numeric_Value=_:=+00000000100}');
Error('\P{Is_Numeric_Value=_:=+00000000100}');
Expect(1, 69732, '\p{Is_Numeric_Value=0_1_00}', "");
Expect(0, 69732, '\p{^Is_Numeric_Value=0_1_00}', "");
Expect(0, 69732, '\P{Is_Numeric_Value=0_1_00}', "");
Expect(1, 69732, '\P{^Is_Numeric_Value=0_1_00}', "");
Expect(0, 69733, '\p{Is_Numeric_Value=0_1_00}', "");
Expect(1, 69733, '\p{^Is_Numeric_Value=0_1_00}', "");
Expect(1, 69733, '\P{Is_Numeric_Value=0_1_00}', "");
Expect(0, 69733, '\P{^Is_Numeric_Value=0_1_00}', "");
Error('\p{Is_Nv= :=+00100}');
Error('\P{Is_Nv= :=+00100}');
Expect(1, 69732, '\p{Is_Nv=100}', "");
Expect(0, 69732, '\p{^Is_Nv=100}', "");
Expect(0, 69732, '\P{Is_Nv=100}', "");
Expect(1, 69732, '\P{^Is_Nv=100}', "");
Expect(0, 69733, '\p{Is_Nv=100}', "");
Expect(1, 69733, '\p{^Is_Nv=100}', "");
Expect(1, 69733, '\P{Is_Nv=100}', "");
Expect(0, 69733, '\P{^Is_Nv=100}', "");
Error('\p{Numeric_Value=	0_0_0_0_0_0_0300:=}');
Error('\P{Numeric_Value=	0_0_0_0_0_0_0300:=}');
Expect(1, 69236, '\p{Numeric_Value=+0_0_0_300}', "");
Expect(0, 69236, '\p{^Numeric_Value=+0_0_0_300}', "");
Expect(0, 69236, '\P{Numeric_Value=+0_0_0_300}', "");
Expect(1, 69236, '\P{^Numeric_Value=+0_0_0_300}', "");
Expect(0, 69237, '\p{Numeric_Value=+0_0_0_300}', "");
Expect(1, 69237, '\p{^Numeric_Value=+0_0_0_300}', "");
Expect(1, 69237, '\P{Numeric_Value=+0_0_0_300}', "");
Expect(0, 69237, '\P{^Numeric_Value=+0_0_0_300}', "");
Error('\p{Nv: _/a/0_0_0_0_0_0_00300}');
Error('\P{Nv: _/a/0_0_0_0_0_0_00300}');
Expect(1, 69236, '\p{Nv=0_3_00}', "");
Expect(0, 69236, '\p{^Nv=0_3_00}', "");
Expect(0, 69236, '\P{Nv=0_3_00}', "");
Expect(1, 69236, '\P{^Nv=0_3_00}', "");
Expect(0, 69237, '\p{Nv=0_3_00}', "");
Expect(1, 69237, '\p{^Nv=0_3_00}', "");
Expect(1, 69237, '\P{Nv=0_3_00}', "");
Expect(0, 69237, '\P{^Nv=0_3_00}', "");
Error('\p{Is_Numeric_Value= 	+0030_0/a/}');
Error('\P{Is_Numeric_Value= 	+0030_0/a/}');
Expect(1, 69236, '\p{Is_Numeric_Value:+00_00_00_30_0}', "");
Expect(0, 69236, '\p{^Is_Numeric_Value:+00_00_00_30_0}', "");
Expect(0, 69236, '\P{Is_Numeric_Value:+00_00_00_30_0}', "");
Expect(1, 69236, '\P{^Is_Numeric_Value:+00_00_00_30_0}', "");
Expect(0, 69237, '\p{Is_Numeric_Value:+00_00_00_30_0}', "");
Expect(1, 69237, '\p{^Is_Numeric_Value:+00_00_00_30_0}', "");
Expect(1, 69237, '\P{Is_Numeric_Value:+00_00_00_30_0}', "");
Expect(0, 69237, '\P{^Is_Numeric_Value:+00_00_00_30_0}', "");
Error('\p{Is_Nv=:=0_0_0_3_00}');
Error('\P{Is_Nv=:=0_0_0_3_00}');
Expect(1, 69236, '\p{Is_Nv=+00000030_0}', "");
Expect(0, 69236, '\p{^Is_Nv=+00000030_0}', "");
Expect(0, 69236, '\P{Is_Nv=+00000030_0}', "");
Expect(1, 69236, '\P{^Is_Nv=+00000030_0}', "");
Expect(0, 69237, '\p{Is_Nv=+00000030_0}', "");
Expect(1, 69237, '\p{^Is_Nv=+00000030_0}', "");
Expect(1, 69237, '\P{Is_Nv=+00000030_0}', "");
Expect(0, 69237, '\P{^Is_Nv=+00000030_0}', "");
Error('\p{Numeric_Value=0000001/0000008:=}');
Error('\P{Numeric_Value=0000001/0000008:=}');
Expect(1, 74847, '\p{Numeric_Value=0001/0008}', "");
Expect(0, 74847, '\p{^Numeric_Value=0001/0008}', "");
Expect(0, 74847, '\P{Numeric_Value=0001/0008}', "");
Expect(1, 74847, '\P{^Numeric_Value=0001/0008}', "");
Expect(0, 74848, '\p{Numeric_Value=0001/0008}', "");
Expect(1, 74848, '\p{^Numeric_Value=0001/0008}', "");
Expect(1, 74848, '\P{Numeric_Value=0001/0008}', "");
Expect(0, 74848, '\P{^Numeric_Value=0001/0008}', "");
Error('\p{Numeric_Value=0.12}');
Error('\P{Numeric_Value=0.12}');
Expect(1, 74847, '\p{Numeric_Value=0.125}', "");
Expect(0, 74847, '\p{^Numeric_Value=0.125}', "");
Expect(0, 74847, '\P{Numeric_Value=0.125}', "");
Expect(1, 74847, '\P{^Numeric_Value=0.125}', "");
Expect(0, 74848, '\p{Numeric_Value=0.125}', "");
Expect(1, 74848, '\p{^Numeric_Value=0.125}', "");
Expect(1, 74848, '\P{Numeric_Value=0.125}', "");
Expect(0, 74848, '\P{^Numeric_Value=0.125}', "");
Error('\p{Nv=:=_-001/000000008}');
Error('\P{Nv=:=_-001/000000008}');
Expect(1, 74847, '\p{Nv: +000000001/0000000008}', "");
Expect(0, 74847, '\p{^Nv: +000000001/0000000008}', "");
Expect(0, 74847, '\P{Nv: +000000001/0000000008}', "");
Expect(1, 74847, '\P{^Nv: +000000001/0000000008}', "");
Expect(0, 74848, '\p{Nv: +000000001/0000000008}', "");
Expect(1, 74848, '\p{^Nv: +000000001/0000000008}', "");
Expect(1, 74848, '\P{Nv: +000000001/0000000008}', "");
Expect(0, 74848, '\P{^Nv: +000000001/0000000008}', "");
Error('\p{Nv:0.12}');
Error('\P{Nv:0.12}');
Expect(1, 74847, '\p{Nv=0.125}', "");
Expect(0, 74847, '\p{^Nv=0.125}', "");
Expect(0, 74847, '\P{Nv=0.125}', "");
Expect(1, 74847, '\P{^Nv=0.125}', "");
Expect(0, 74848, '\p{Nv=0.125}', "");
Expect(1, 74848, '\p{^Nv=0.125}', "");
Expect(1, 74848, '\P{Nv=0.125}', "");
Expect(0, 74848, '\P{^Nv=0.125}', "");
Error('\p{Is_Numeric_Value=/a/ _+000001/000000008}');
Error('\P{Is_Numeric_Value=/a/ _+000001/000000008}');
Expect(1, 74847, '\p{Is_Numeric_Value=01/000000008}', "");
Expect(0, 74847, '\p{^Is_Numeric_Value=01/000000008}', "");
Expect(0, 74847, '\P{Is_Numeric_Value=01/000000008}', "");
Expect(1, 74847, '\P{^Is_Numeric_Value=01/000000008}', "");
Expect(0, 74848, '\p{Is_Numeric_Value=01/000000008}', "");
Expect(1, 74848, '\p{^Is_Numeric_Value=01/000000008}', "");
Expect(1, 74848, '\P{Is_Numeric_Value=01/000000008}', "");
Expect(0, 74848, '\P{^Is_Numeric_Value=01/000000008}', "");
Error('\p{Is_Numeric_Value=0.12}');
Error('\P{Is_Numeric_Value=0.12}');
Expect(1, 74847, '\p{Is_Numeric_Value=0.125}', "");
Expect(0, 74847, '\p{^Is_Numeric_Value=0.125}', "");
Expect(0, 74847, '\P{Is_Numeric_Value=0.125}', "");
Expect(1, 74847, '\P{^Is_Numeric_Value=0.125}', "");
Expect(0, 74848, '\p{Is_Numeric_Value=0.125}', "");
Expect(1, 74848, '\p{^Is_Numeric_Value=0.125}', "");
Expect(1, 74848, '\P{Is_Numeric_Value=0.125}', "");
Expect(0, 74848, '\P{^Is_Numeric_Value=0.125}', "");
Error('\p{Is_Nv=/a/__000001/00008}');
Error('\P{Is_Nv=/a/__000001/00008}');
Expect(1, 74847, '\p{Is_Nv=01/0008}', "");
Expect(0, 74847, '\p{^Is_Nv=01/0008}', "");
Expect(0, 74847, '\P{Is_Nv=01/0008}', "");
Expect(1, 74847, '\P{^Is_Nv=01/0008}', "");
Expect(0, 74848, '\p{Is_Nv=01/0008}', "");
Expect(1, 74848, '\p{^Is_Nv=01/0008}', "");
Expect(1, 74848, '\P{Is_Nv=01/0008}', "");
Expect(0, 74848, '\P{^Is_Nv=01/0008}', "");
Error('\p{Is_Nv=0.12}');
Error('\P{Is_Nv=0.12}');
Expect(1, 74847, '\p{Is_Nv=0.125}', "");
Expect(0, 74847, '\p{^Is_Nv=0.125}', "");
Expect(0, 74847, '\P{Is_Nv=0.125}', "");
Expect(1, 74847, '\P{^Is_Nv=0.125}', "");
Expect(0, 74848, '\p{Is_Nv=0.125}', "");
Expect(1, 74848, '\p{^Is_Nv=0.125}', "");
Expect(1, 74848, '\P{Is_Nv=0.125}', "");
Expect(0, 74848, '\P{^Is_Nv=0.125}', "");
Error('\p{Numeric_Value= :=+05/6}');
Error('\P{Numeric_Value= :=+05/6}');
Expect(1, 74844, '\p{Numeric_Value=000000005/006}', "");
Expect(0, 74844, '\p{^Numeric_Value=000000005/006}', "");
Expect(0, 74844, '\P{Numeric_Value=000000005/006}', "");
Expect(1, 74844, '\P{^Numeric_Value=000000005/006}', "");
Expect(0, 74845, '\p{Numeric_Value=000000005/006}', "");
Expect(1, 74845, '\p{^Numeric_Value=000000005/006}', "");
Expect(1, 74845, '\P{Numeric_Value=000000005/006}', "");
Expect(0, 74845, '\P{^Numeric_Value=000000005/006}', "");
Error('\p{Numeric_Value=0.83}');
Error('\P{Numeric_Value=0.83}');
Expect(1, 74844, '\p{Numeric_Value=0.833}', "");
Expect(0, 74844, '\p{^Numeric_Value=0.833}', "");
Expect(0, 74844, '\P{Numeric_Value=0.833}', "");
Expect(1, 74844, '\P{^Numeric_Value=0.833}', "");
Expect(0, 74845, '\p{Numeric_Value=0.833}', "");
Expect(1, 74845, '\p{^Numeric_Value=0.833}', "");
Expect(1, 74845, '\P{Numeric_Value=0.833}', "");
Expect(0, 74845, '\P{^Numeric_Value=0.833}', "");
Error('\p{Nv:   :=00000005/0000006}');
Error('\P{Nv:   :=00000005/0000006}');
Expect(1, 74844, '\p{Nv=00000005/6}', "");
Expect(0, 74844, '\p{^Nv=00000005/6}', "");
Expect(0, 74844, '\P{Nv=00000005/6}', "");
Expect(1, 74844, '\P{^Nv=00000005/6}', "");
Expect(0, 74845, '\p{Nv=00000005/6}', "");
Expect(1, 74845, '\p{^Nv=00000005/6}', "");
Expect(1, 74845, '\P{Nv=00000005/6}', "");
Expect(0, 74845, '\P{^Nv=00000005/6}', "");
Error('\p{Nv:	0.83}');
Error('\P{Nv:	0.83}');
Expect(1, 74844, '\p{Nv:   0.833}', "");
Expect(0, 74844, '\p{^Nv:   0.833}', "");
Expect(0, 74844, '\P{Nv:   0.833}', "");
Expect(1, 74844, '\P{^Nv:   0.833}', "");
Expect(0, 74845, '\p{Nv:   0.833}', "");
Expect(1, 74845, '\p{^Nv:   0.833}', "");
Expect(1, 74845, '\P{Nv:   0.833}', "");
Expect(0, 74845, '\P{^Nv:   0.833}', "");
Error('\p{Is_Numeric_Value:	_5/00006:=}');
Error('\P{Is_Numeric_Value:	_5/00006:=}');
Expect(1, 74844, '\p{Is_Numeric_Value:005/0000006}', "");
Expect(0, 74844, '\p{^Is_Numeric_Value:005/0000006}', "");
Expect(0, 74844, '\P{Is_Numeric_Value:005/0000006}', "");
Expect(1, 74844, '\P{^Is_Numeric_Value:005/0000006}', "");
Expect(0, 74845, '\p{Is_Numeric_Value:005/0000006}', "");
Expect(1, 74845, '\p{^Is_Numeric_Value:005/0000006}', "");
Expect(1, 74845, '\P{Is_Numeric_Value:005/0000006}', "");
Expect(0, 74845, '\P{^Is_Numeric_Value:005/0000006}', "");
Error('\p{Is_Numeric_Value=0.83}');
Error('\P{Is_Numeric_Value=0.83}');
Expect(1, 74844, '\p{Is_Numeric_Value=0.833}', "");
Expect(0, 74844, '\p{^Is_Numeric_Value=0.833}', "");
Expect(0, 74844, '\P{Is_Numeric_Value=0.833}', "");
Expect(1, 74844, '\P{^Is_Numeric_Value=0.833}', "");
Expect(0, 74845, '\p{Is_Numeric_Value=0.833}', "");
Expect(1, 74845, '\p{^Is_Numeric_Value=0.833}', "");
Expect(1, 74845, '\P{Is_Numeric_Value=0.833}', "");
Expect(0, 74845, '\P{^Is_Numeric_Value=0.833}', "");
Error('\p{Is_Nv=- 05/0006/a/}');
Error('\P{Is_Nv=- 05/0006/a/}');
Expect(1, 74844, '\p{Is_Nv=05/0006}', "");
Expect(0, 74844, '\p{^Is_Nv=05/0006}', "");
Expect(0, 74844, '\P{Is_Nv=05/0006}', "");
Expect(1, 74844, '\P{^Is_Nv=05/0006}', "");
Expect(0, 74845, '\p{Is_Nv=05/0006}', "");
Expect(1, 74845, '\p{^Is_Nv=05/0006}', "");
Expect(1, 74845, '\P{Is_Nv=05/0006}', "");
Expect(0, 74845, '\P{^Is_Nv=05/0006}', "");
Error('\p{Is_Nv:   0.83}');
Error('\P{Is_Nv:   0.83}');
Expect(1, 74844, '\p{Is_Nv=0.833}', "");
Expect(0, 74844, '\p{^Is_Nv=0.833}', "");
Expect(0, 74844, '\P{Is_Nv=0.833}', "");
Expect(1, 74844, '\P{^Is_Nv=0.833}', "");
Expect(0, 74845, '\p{Is_Nv=0.833}', "");
Expect(1, 74845, '\p{^Is_Nv=0.833}', "");
Expect(1, 74845, '\P{Is_Nv=0.833}', "");
Expect(0, 74845, '\P{^Is_Nv=0.833}', "");
Error('\p{Numeric_Value=-/a/00000000025}');
Error('\P{Numeric_Value=-/a/00000000025}');
Expect(1, 12885, '\p{Numeric_Value=00000025}', "");
Expect(0, 12885, '\p{^Numeric_Value=00000025}', "");
Expect(0, 12885, '\P{Numeric_Value=00000025}', "");
Expect(1, 12885, '\P{^Numeric_Value=00000025}', "");
Expect(0, 12886, '\p{Numeric_Value=00000025}', "");
Expect(1, 12886, '\p{^Numeric_Value=00000025}', "");
Expect(1, 12886, '\P{Numeric_Value=00000025}', "");
Expect(0, 12886, '\P{^Numeric_Value=00000025}', "");
Error('\p{Nv=-/a/0000025}');
Error('\P{Nv=-/a/0000025}');
Expect(1, 12885, '\p{Nv=+025}', "");
Expect(0, 12885, '\p{^Nv=+025}', "");
Expect(0, 12885, '\P{Nv=+025}', "");
Expect(1, 12885, '\P{^Nv=+025}', "");
Expect(0, 12886, '\p{Nv=+025}', "");
Expect(1, 12886, '\p{^Nv=+025}', "");
Expect(1, 12886, '\P{Nv=+025}', "");
Expect(0, 12886, '\P{^Nv=+025}', "");
Error('\p{Is_Numeric_Value= :=0_0_0_0_0_0_0025}');
Error('\P{Is_Numeric_Value= :=0_0_0_0_0_0_0025}');
Expect(1, 12885, '\p{Is_Numeric_Value=0000025}', "");
Expect(0, 12885, '\p{^Is_Numeric_Value=0000025}', "");
Expect(0, 12885, '\P{Is_Numeric_Value=0000025}', "");
Expect(1, 12885, '\P{^Is_Numeric_Value=0000025}', "");
Expect(0, 12886, '\p{Is_Numeric_Value=0000025}', "");
Expect(1, 12886, '\p{^Is_Numeric_Value=0000025}', "");
Expect(1, 12886, '\P{Is_Numeric_Value=0000025}', "");
Expect(0, 12886, '\P{^Is_Numeric_Value=0000025}', "");
Error('\p{Is_Nv=:=+000_000_000_25}');
Error('\P{Is_Nv=:=+000_000_000_25}');
Expect(1, 12885, '\p{Is_Nv:+00000_00002_5}', "");
Expect(0, 12885, '\p{^Is_Nv:+00000_00002_5}', "");
Expect(0, 12885, '\P{Is_Nv:+00000_00002_5}', "");
Expect(1, 12885, '\P{^Is_Nv:+00000_00002_5}', "");
Expect(0, 12886, '\p{Is_Nv:+00000_00002_5}', "");
Expect(1, 12886, '\p{^Is_Nv:+00000_00002_5}', "");
Expect(1, 12886, '\P{Is_Nv:+00000_00002_5}', "");
Expect(0, 12886, '\P{^Is_Nv:+00000_00002_5}', "");
Error('\p{Numeric_Value=	/a/0_0_0_0_28}');
Error('\P{Numeric_Value=	/a/0_0_0_0_28}');
Expect(1, 12888, '\p{Numeric_Value=000000028}', "");
Expect(0, 12888, '\p{^Numeric_Value=000000028}', "");
Expect(0, 12888, '\P{Numeric_Value=000000028}', "");
Expect(1, 12888, '\P{^Numeric_Value=000000028}', "");
Expect(0, 12889, '\p{Numeric_Value=000000028}', "");
Expect(1, 12889, '\p{^Numeric_Value=000000028}', "");
Expect(1, 12889, '\P{Numeric_Value=000000028}', "");
Expect(0, 12889, '\P{^Numeric_Value=000000028}', "");
Error('\p{Nv=_-028:=}');
Error('\P{Nv=_-028:=}');
Expect(1, 12888, '\p{Nv=0000000028}', "");
Expect(0, 12888, '\p{^Nv=0000000028}', "");
Expect(0, 12888, '\P{Nv=0000000028}', "");
Expect(1, 12888, '\P{^Nv=0000000028}', "");
Expect(0, 12889, '\p{Nv=0000000028}', "");
Expect(1, 12889, '\p{^Nv=0000000028}', "");
Expect(1, 12889, '\P{Nv=0000000028}', "");
Expect(0, 12889, '\P{^Nv=0000000028}', "");
Error('\p{Is_Numeric_Value: -+028/a/}');
Error('\P{Is_Numeric_Value: -+028/a/}');
Expect(1, 12888, '\p{Is_Numeric_Value: 0000000028}', "");
Expect(0, 12888, '\p{^Is_Numeric_Value: 0000000028}', "");
Expect(0, 12888, '\P{Is_Numeric_Value: 0000000028}', "");
Expect(1, 12888, '\P{^Is_Numeric_Value: 0000000028}', "");
Expect(0, 12889, '\p{Is_Numeric_Value: 0000000028}', "");
Expect(1, 12889, '\p{^Is_Numeric_Value: 0000000028}', "");
Expect(1, 12889, '\P{Is_Numeric_Value: 0000000028}', "");
Expect(0, 12889, '\P{^Is_Numeric_Value: 0000000028}', "");
Error('\p{Is_Nv:	/a/+0002_8}');
Error('\P{Is_Nv:	/a/+0002_8}');
Expect(1, 12888, '\p{Is_Nv=0_0_0_028}', "");
Expect(0, 12888, '\p{^Is_Nv=0_0_0_028}', "");
Expect(0, 12888, '\P{Is_Nv=0_0_0_028}', "");
Expect(1, 12888, '\P{^Is_Nv=0_0_0_028}', "");
Expect(0, 12889, '\p{Is_Nv=0_0_0_028}', "");
Expect(1, 12889, '\p{^Is_Nv=0_0_0_028}', "");
Expect(1, 12889, '\P{Is_Nv=0_0_0_028}', "");
Expect(0, 12889, '\P{^Is_Nv=0_0_0_028}', "");
Error('\p{Numeric_Value=:=_0000007/2}');
Error('\P{Numeric_Value=:=_0000007/2}');
Expect(1, 3885, '\p{Numeric_Value=007/2}', "");
Expect(0, 3885, '\p{^Numeric_Value=007/2}', "");
Expect(0, 3885, '\P{Numeric_Value=007/2}', "");
Expect(1, 3885, '\P{^Numeric_Value=007/2}', "");
Expect(0, 3886, '\p{Numeric_Value=007/2}', "");
Expect(1, 3886, '\p{^Numeric_Value=007/2}', "");
Expect(1, 3886, '\P{Numeric_Value=007/2}', "");
Expect(0, 3886, '\P{^Numeric_Value=007/2}', "");
Expect(1, 3885, '\p{Numeric_Value=3.500}', "");
Expect(0, 3885, '\p{^Numeric_Value=3.500}', "");
Expect(0, 3885, '\P{Numeric_Value=3.500}', "");
Expect(1, 3885, '\P{^Numeric_Value=3.500}', "");
Expect(0, 3886, '\p{Numeric_Value=3.500}', "");
Expect(1, 3886, '\p{^Numeric_Value=3.500}', "");
Expect(1, 3886, '\P{Numeric_Value=3.500}', "");
Expect(0, 3886, '\P{^Numeric_Value=3.500}', "");
Error('\p{Nv:   __+07/0000000002/a/}');
Error('\P{Nv:   __+07/0000000002/a/}');
Expect(1, 3885, '\p{Nv=00000007/000000002}', "");
Expect(0, 3885, '\p{^Nv=00000007/000000002}', "");
Expect(0, 3885, '\P{Nv=00000007/000000002}', "");
Expect(1, 3885, '\P{^Nv=00000007/000000002}', "");
Expect(0, 3886, '\p{Nv=00000007/000000002}', "");
Expect(1, 3886, '\p{^Nv=00000007/000000002}', "");
Expect(1, 3886, '\P{Nv=00000007/000000002}', "");
Expect(0, 3886, '\P{^Nv=00000007/000000002}', "");
Expect(1, 3885, '\p{Nv=3.500}', "");
Expect(0, 3885, '\p{^Nv=3.500}', "");
Expect(0, 3885, '\P{Nv=3.500}', "");
Expect(1, 3885, '\P{^Nv=3.500}', "");
Expect(0, 3886, '\p{Nv=3.500}', "");
Expect(1, 3886, '\p{^Nv=3.500}', "");
Expect(1, 3886, '\P{Nv=3.500}', "");
Expect(0, 3886, '\P{^Nv=3.500}', "");
Error('\p{Is_Numeric_Value= -+000000007/00002:=}');
Error('\P{Is_Numeric_Value= -+000000007/00002:=}');
Expect(1, 3885, '\p{Is_Numeric_Value=0000007/2}', "");
Expect(0, 3885, '\p{^Is_Numeric_Value=0000007/2}', "");
Expect(0, 3885, '\P{Is_Numeric_Value=0000007/2}', "");
Expect(1, 3885, '\P{^Is_Numeric_Value=0000007/2}', "");
Expect(0, 3886, '\p{Is_Numeric_Value=0000007/2}', "");
Expect(1, 3886, '\p{^Is_Numeric_Value=0000007/2}', "");
Expect(1, 3886, '\P{Is_Numeric_Value=0000007/2}', "");
Expect(0, 3886, '\P{^Is_Numeric_Value=0000007/2}', "");
Expect(1, 3885, '\p{Is_Numeric_Value=3.500}', "");
Expect(0, 3885, '\p{^Is_Numeric_Value=3.500}', "");
Expect(0, 3885, '\P{Is_Numeric_Value=3.500}', "");
Expect(1, 3885, '\P{^Is_Numeric_Value=3.500}', "");
Expect(0, 3886, '\p{Is_Numeric_Value=3.500}', "");
Expect(1, 3886, '\p{^Is_Numeric_Value=3.500}', "");
Expect(1, 3886, '\P{Is_Numeric_Value=3.500}', "");
Expect(0, 3886, '\P{^Is_Numeric_Value=3.500}', "");
Error('\p{Is_Nv=-:=00000007/0000000002}');
Error('\P{Is_Nv=-:=00000007/0000000002}');
Expect(1, 3885, '\p{Is_Nv=+00007/00002}', "");
Expect(0, 3885, '\p{^Is_Nv=+00007/00002}', "");
Expect(0, 3885, '\P{Is_Nv=+00007/00002}', "");
Expect(1, 3885, '\P{^Is_Nv=+00007/00002}', "");
Expect(0, 3886, '\p{Is_Nv=+00007/00002}', "");
Expect(1, 3886, '\p{^Is_Nv=+00007/00002}', "");
Expect(1, 3886, '\P{Is_Nv=+00007/00002}', "");
Expect(0, 3886, '\P{^Is_Nv=+00007/00002}', "");
Expect(1, 3885, '\p{Is_Nv: 3.500}', "");
Expect(0, 3885, '\p{^Is_Nv: 3.500}', "");
Expect(0, 3885, '\P{Is_Nv: 3.500}', "");
Expect(1, 3885, '\P{^Is_Nv: 3.500}', "");
Expect(0, 3886, '\p{Is_Nv: 3.500}', "");
Expect(1, 3886, '\p{^Is_Nv: 3.500}', "");
Expect(1, 3886, '\P{Is_Nv: 3.500}', "");
Expect(0, 3886, '\P{^Is_Nv: 3.500}', "");
Error('\p{Numeric_Value:   /a/ -000009/0000000002}');
Error('\P{Numeric_Value:   /a/ -000009/0000000002}');
Expect(1, 3886, '\p{Numeric_Value: 000000009/000002}', "");
Expect(0, 3886, '\p{^Numeric_Value: 000000009/000002}', "");
Expect(0, 3886, '\P{Numeric_Value: 000000009/000002}', "");
Expect(1, 3886, '\P{^Numeric_Value: 000000009/000002}', "");
Expect(0, 3887, '\p{Numeric_Value: 000000009/000002}', "");
Expect(1, 3887, '\p{^Numeric_Value: 000000009/000002}', "");
Expect(1, 3887, '\P{Numeric_Value: 000000009/000002}', "");
Expect(0, 3887, '\P{^Numeric_Value: 000000009/000002}', "");
Expect(1, 3886, '\p{Numeric_Value=4.500}', "");
Expect(0, 3886, '\p{^Numeric_Value=4.500}', "");
Expect(0, 3886, '\P{Numeric_Value=4.500}', "");
Expect(1, 3886, '\P{^Numeric_Value=4.500}', "");
Expect(0, 3887, '\p{Numeric_Value=4.500}', "");
Expect(1, 3887, '\p{^Numeric_Value=4.500}', "");
Expect(1, 3887, '\P{Numeric_Value=4.500}', "");
Expect(0, 3887, '\P{^Numeric_Value=4.500}', "");
Error('\p{Nv=-09/02:=}');
Error('\P{Nv=-09/02:=}');
Expect(1, 3886, '\p{Nv=+00009/00002}', "");
Expect(0, 3886, '\p{^Nv=+00009/00002}', "");
Expect(0, 3886, '\P{Nv=+00009/00002}', "");
Expect(1, 3886, '\P{^Nv=+00009/00002}', "");
Expect(0, 3887, '\p{Nv=+00009/00002}', "");
Expect(1, 3887, '\p{^Nv=+00009/00002}', "");
Expect(1, 3887, '\P{Nv=+00009/00002}', "");
Expect(0, 3887, '\P{^Nv=+00009/00002}', "");
Expect(1, 3886, '\p{Nv=4.500}', "");
Expect(0, 3886, '\p{^Nv=4.500}', "");
Expect(0, 3886, '\P{Nv=4.500}', "");
Expect(1, 3886, '\P{^Nv=4.500}', "");
Expect(0, 3887, '\p{Nv=4.500}', "");
Expect(1, 3887, '\p{^Nv=4.500}', "");
Expect(1, 3887, '\P{Nv=4.500}', "");
Expect(0, 3887, '\P{^Nv=4.500}', "");
Error('\p{Is_Numeric_Value= 	0000000009/000000002:=}');
Error('\P{Is_Numeric_Value= 	0000000009/000000002:=}');
Expect(1, 3886, '\p{Is_Numeric_Value=+00009/0000000002}', "");
Expect(0, 3886, '\p{^Is_Numeric_Value=+00009/0000000002}', "");
Expect(0, 3886, '\P{Is_Numeric_Value=+00009/0000000002}', "");
Expect(1, 3886, '\P{^Is_Numeric_Value=+00009/0000000002}', "");
Expect(0, 3887, '\p{Is_Numeric_Value=+00009/0000000002}', "");
Expect(1, 3887, '\p{^Is_Numeric_Value=+00009/0000000002}', "");
Expect(1, 3887, '\P{Is_Numeric_Value=+00009/0000000002}', "");
Expect(0, 3887, '\P{^Is_Numeric_Value=+00009/0000000002}', "");
Expect(1, 3886, '\p{Is_Numeric_Value=4.500}', "");
Expect(0, 3886, '\p{^Is_Numeric_Value=4.500}', "");
Expect(0, 3886, '\P{Is_Numeric_Value=4.500}', "");
Expect(1, 3886, '\P{^Is_Numeric_Value=4.500}', "");
Expect(0, 3887, '\p{Is_Numeric_Value=4.500}', "");
Expect(1, 3887, '\p{^Is_Numeric_Value=4.500}', "");
Expect(1, 3887, '\P{Is_Numeric_Value=4.500}', "");
Expect(0, 3887, '\P{^Is_Numeric_Value=4.500}', "");
Error('\p{Is_Nv=/a/ +000000009/0002}');
Error('\P{Is_Nv=/a/ +000000009/0002}');
Expect(1, 3886, '\p{Is_Nv:   00009/0000000002}', "");
Expect(0, 3886, '\p{^Is_Nv:   00009/0000000002}', "");
Expect(0, 3886, '\P{Is_Nv:   00009/0000000002}', "");
Expect(1, 3886, '\P{^Is_Nv:   00009/0000000002}', "");
Expect(0, 3887, '\p{Is_Nv:   00009/0000000002}', "");
Expect(1, 3887, '\p{^Is_Nv:   00009/0000000002}', "");
Expect(1, 3887, '\P{Is_Nv:   00009/0000000002}', "");
Expect(0, 3887, '\P{^Is_Nv:   00009/0000000002}', "");
Expect(1, 3886, '\p{Is_Nv=4.500}', "");
Expect(0, 3886, '\p{^Is_Nv=4.500}', "");
Expect(0, 3886, '\P{Is_Nv=4.500}', "");
Expect(1, 3886, '\P{^Is_Nv=4.500}', "");
Expect(0, 3887, '\p{Is_Nv=4.500}', "");
Expect(1, 3887, '\p{^Is_Nv=4.500}', "");
Expect(1, 3887, '\P{Is_Nv=4.500}', "");
Expect(0, 3887, '\P{^Is_Nv=4.500}', "");
Error('\p{Numeric_Value:   -:=000040}');
Error('\P{Numeric_Value:   -:=000040}');
Expect(1, 133532, '\p{Numeric_Value=0_0_0_0_0_0_40}', "");
Expect(0, 133532, '\p{^Numeric_Value=0_0_0_0_0_0_40}', "");
Expect(0, 133532, '\P{Numeric_Value=0_0_0_0_0_0_40}', "");
Expect(1, 133532, '\P{^Numeric_Value=0_0_0_0_0_0_40}', "");
Expect(0, 133533, '\p{Numeric_Value=0_0_0_0_0_0_40}', "");
Expect(1, 133533, '\p{^Numeric_Value=0_0_0_0_0_0_40}', "");
Expect(1, 133533, '\P{Numeric_Value=0_0_0_0_0_0_40}', "");
Expect(0, 133533, '\P{^Numeric_Value=0_0_0_0_0_0_40}', "");
Error('\p{Nv=:=_	00040}');
Error('\P{Nv=:=_	00040}');
Expect(1, 133532, '\p{Nv=+0_0_0_0_0_40}', "");
Expect(0, 133532, '\p{^Nv=+0_0_0_0_0_40}', "");
Expect(0, 133532, '\P{Nv=+0_0_0_0_0_40}', "");
Expect(1, 133532, '\P{^Nv=+0_0_0_0_0_40}', "");
Expect(0, 133533, '\p{Nv=+0_0_0_0_0_40}', "");
Expect(1, 133533, '\p{^Nv=+0_0_0_0_0_40}', "");
Expect(1, 133533, '\P{Nv=+0_0_0_0_0_40}', "");
Expect(0, 133533, '\P{^Nv=+0_0_0_0_0_40}', "");
Error('\p{Is_Numeric_Value=:=- 00040}');
Error('\P{Is_Numeric_Value=:=- 00040}');
Expect(1, 133532, '\p{Is_Numeric_Value=4_0}', "");
Expect(0, 133532, '\p{^Is_Numeric_Value=4_0}', "");
Expect(0, 133532, '\P{Is_Numeric_Value=4_0}', "");
Expect(1, 133532, '\P{^Is_Numeric_Value=4_0}', "");
Expect(0, 133533, '\p{Is_Numeric_Value=4_0}', "");
Expect(1, 133533, '\p{^Is_Numeric_Value=4_0}', "");
Expect(1, 133533, '\P{Is_Numeric_Value=4_0}', "");
Expect(0, 133533, '\P{^Is_Numeric_Value=4_0}', "");
Error('\p{Is_Nv:    -00000040/a/}');
Error('\P{Is_Nv:    -00000040/a/}');
Expect(1, 133532, '\p{Is_Nv=+0_0_0_0_0_0_0_0040}', "");
Expect(0, 133532, '\p{^Is_Nv=+0_0_0_0_0_0_0_0040}', "");
Expect(0, 133532, '\P{Is_Nv=+0_0_0_0_0_0_0_0040}', "");
Expect(1, 133532, '\P{^Is_Nv=+0_0_0_0_0_0_0_0040}', "");
Expect(0, 133533, '\p{Is_Nv=+0_0_0_0_0_0_0_0040}', "");
Expect(1, 133533, '\p{^Is_Nv=+0_0_0_0_0_0_0_0040}', "");
Expect(1, 133533, '\P{Is_Nv=+0_0_0_0_0_0_0_0040}', "");
Expect(0, 133533, '\P{^Is_Nv=+0_0_0_0_0_0_0_0040}', "");
Error('\p{Numeric_Value= :=0_0_0_0_0_0_0_0_3_000}');
Error('\P{Numeric_Value= :=0_0_0_0_0_0_0_0_3_000}');
Expect(1, 65828, '\p{Numeric_Value=+00_00_00_00_03_000}', "");
Expect(0, 65828, '\p{^Numeric_Value=+00_00_00_00_03_000}', "");
Expect(0, 65828, '\P{Numeric_Value=+00_00_00_00_03_000}', "");
Expect(1, 65828, '\P{^Numeric_Value=+00_00_00_00_03_000}', "");
Expect(0, 65829, '\p{Numeric_Value=+00_00_00_00_03_000}', "");
Expect(1, 65829, '\p{^Numeric_Value=+00_00_00_00_03_000}', "");
Expect(1, 65829, '\P{Numeric_Value=+00_00_00_00_03_000}', "");
Expect(0, 65829, '\P{^Numeric_Value=+00_00_00_00_03_000}', "");
Error('\p{Nv=_/a/0_0_0_0_0_0_0_0_3_0_00}');
Error('\P{Nv=_/a/0_0_0_0_0_0_0_0_3_0_00}');
Expect(1, 65828, '\p{Nv=0000003000}', "");
Expect(0, 65828, '\p{^Nv=0000003000}', "");
Expect(0, 65828, '\P{Nv=0000003000}', "");
Expect(1, 65828, '\P{^Nv=0000003000}', "");
Expect(0, 65829, '\p{Nv=0000003000}', "");
Expect(1, 65829, '\p{^Nv=0000003000}', "");
Expect(1, 65829, '\P{Nv=0000003000}', "");
Expect(0, 65829, '\P{^Nv=0000003000}', "");
Error('\p{Is_Numeric_Value=-:=+00_00_03_00_0}');
Error('\P{Is_Numeric_Value=-:=+00_00_03_00_0}');
Expect(1, 65828, '\p{Is_Numeric_Value=+0000000003000}', "");
Expect(0, 65828, '\p{^Is_Numeric_Value=+0000000003000}', "");
Expect(0, 65828, '\P{Is_Numeric_Value=+0000000003000}', "");
Expect(1, 65828, '\P{^Is_Numeric_Value=+0000000003000}', "");
Expect(0, 65829, '\p{Is_Numeric_Value=+0000000003000}', "");
Expect(1, 65829, '\p{^Is_Numeric_Value=+0000000003000}', "");
Expect(1, 65829, '\P{Is_Numeric_Value=+0000000003000}', "");
Expect(0, 65829, '\P{^Is_Numeric_Value=+0000000003000}', "");
Error('\p{Is_Nv=/a/	+00003000}');
Error('\P{Is_Nv=/a/	+00003000}');
Expect(1, 65828, '\p{Is_Nv:	000003000}', "");
Expect(0, 65828, '\p{^Is_Nv:	000003000}', "");
Expect(0, 65828, '\P{Is_Nv:	000003000}', "");
Expect(1, 65828, '\P{^Is_Nv:	000003000}', "");
Expect(0, 65829, '\p{Is_Nv:	000003000}', "");
Expect(1, 65829, '\p{^Is_Nv:	000003000}', "");
Expect(1, 65829, '\P{Is_Nv:	000003000}', "");
Expect(0, 65829, '\P{^Is_Nv:	000003000}', "");
Error('\p{Numeric_Value:	-0000000800/a/}');
Error('\P{Numeric_Value:	-0000000800/a/}');
Expect(1, 69241, '\p{Numeric_Value=0_0_0_0_8_00}', "");
Expect(0, 69241, '\p{^Numeric_Value=0_0_0_0_8_00}', "");
Expect(0, 69241, '\P{Numeric_Value=0_0_0_0_8_00}', "");
Expect(1, 69241, '\P{^Numeric_Value=0_0_0_0_8_00}', "");
Expect(0, 69242, '\p{Numeric_Value=0_0_0_0_8_00}', "");
Expect(1, 69242, '\p{^Numeric_Value=0_0_0_0_8_00}', "");
Expect(1, 69242, '\P{Numeric_Value=0_0_0_0_8_00}', "");
Expect(0, 69242, '\P{^Numeric_Value=0_0_0_0_8_00}', "");
Error('\p{Nv:	 080_0/a/}');
Error('\P{Nv:	 080_0/a/}');
Expect(1, 69241, '\p{Nv:+800}', "");
Expect(0, 69241, '\p{^Nv:+800}', "");
Expect(0, 69241, '\P{Nv:+800}', "");
Expect(1, 69241, '\P{^Nv:+800}', "");
Expect(0, 69242, '\p{Nv:+800}', "");
Expect(1, 69242, '\p{^Nv:+800}', "");
Expect(1, 69242, '\P{Nv:+800}', "");
Expect(0, 69242, '\P{^Nv:+800}', "");
Error('\p{Is_Numeric_Value=/a/	+00800}');
Error('\P{Is_Numeric_Value=/a/	+00800}');
Expect(1, 69241, '\p{Is_Numeric_Value=+0800}', "");
Expect(0, 69241, '\p{^Is_Numeric_Value=+0800}', "");
Expect(0, 69241, '\P{Is_Numeric_Value=+0800}', "");
Expect(1, 69241, '\P{^Is_Numeric_Value=+0800}', "");
Expect(0, 69242, '\p{Is_Numeric_Value=+0800}', "");
Expect(1, 69242, '\p{^Is_Numeric_Value=+0800}', "");
Expect(1, 69242, '\P{Is_Numeric_Value=+0800}', "");
Expect(0, 69242, '\P{^Is_Numeric_Value=+0800}', "");
Error('\p{Is_Nv=	_0_0_0_0_0_0_0_0_8_00:=}');
Error('\P{Is_Nv=	_0_0_0_0_0_0_0_0_8_00:=}');
Expect(1, 69241, '\p{Is_Nv: +000000000800}', "");
Expect(0, 69241, '\p{^Is_Nv: +000000000800}', "");
Expect(0, 69241, '\P{Is_Nv: +000000000800}', "");
Expect(1, 69241, '\P{^Is_Nv: +000000000800}', "");
Expect(0, 69242, '\p{Is_Nv: +000000000800}', "");
Expect(1, 69242, '\p{^Is_Nv: +000000000800}', "");
Expect(1, 69242, '\P{Is_Nv: +000000000800}', "");
Expect(0, 69242, '\P{^Is_Nv: +000000000800}', "");
Error('\p{Numeric_Value=/a/_	+000000009000}');
Error('\P{Numeric_Value=/a/_	+000000009000}');
Expect(1, 65834, '\p{Numeric_Value=900_0}', "");
Expect(0, 65834, '\p{^Numeric_Value=900_0}', "");
Expect(0, 65834, '\P{Numeric_Value=900_0}', "");
Expect(1, 65834, '\P{^Numeric_Value=900_0}', "");
Expect(0, 65835, '\p{Numeric_Value=900_0}', "");
Expect(1, 65835, '\p{^Numeric_Value=900_0}', "");
Expect(1, 65835, '\P{Numeric_Value=900_0}', "");
Expect(0, 65835, '\P{^Numeric_Value=900_0}', "");
Error('\p{Nv= _9000/a/}');
Error('\P{Nv= _9000/a/}');
Expect(1, 65834, '\p{Nv=0_0_0_0_0_0_0_0_0_9_0_00}', "");
Expect(0, 65834, '\p{^Nv=0_0_0_0_0_0_0_0_0_9_0_00}', "");
Expect(0, 65834, '\P{Nv=0_0_0_0_0_0_0_0_0_9_0_00}', "");
Expect(1, 65834, '\P{^Nv=0_0_0_0_0_0_0_0_0_9_0_00}', "");
Expect(0, 65835, '\p{Nv=0_0_0_0_0_0_0_0_0_9_0_00}', "");
Expect(1, 65835, '\p{^Nv=0_0_0_0_0_0_0_0_0_9_0_00}', "");
Expect(1, 65835, '\P{Nv=0_0_0_0_0_0_0_0_0_9_0_00}', "");
Expect(0, 65835, '\P{^Nv=0_0_0_0_0_0_0_0_0_9_0_00}', "");
Error('\p{Is_Numeric_Value=	/a/+09000}');
Error('\P{Is_Numeric_Value=	/a/+09000}');
Expect(1, 65834, '\p{Is_Numeric_Value=0_0_0_0_0_9000}', "");
Expect(0, 65834, '\p{^Is_Numeric_Value=0_0_0_0_0_9000}', "");
Expect(0, 65834, '\P{Is_Numeric_Value=0_0_0_0_0_9000}', "");
Expect(1, 65834, '\P{^Is_Numeric_Value=0_0_0_0_0_9000}', "");
Expect(0, 65835, '\p{Is_Numeric_Value=0_0_0_0_0_9000}', "");
Expect(1, 65835, '\p{^Is_Numeric_Value=0_0_0_0_0_9000}', "");
Expect(1, 65835, '\P{Is_Numeric_Value=0_0_0_0_0_9000}', "");
Expect(0, 65835, '\P{^Is_Numeric_Value=0_0_0_0_0_9000}', "");
Error('\p{Is_Nv=/a/-+000009000}');
Error('\P{Is_Nv=/a/-+000009000}');
Expect(1, 65834, '\p{Is_Nv=+9000}', "");
Expect(0, 65834, '\p{^Is_Nv=+9000}', "");
Expect(0, 65834, '\P{Is_Nv=+9000}', "");
Expect(1, 65834, '\P{^Is_Nv=+9000}', "");
Expect(0, 65835, '\p{Is_Nv=+9000}', "");
Expect(1, 65835, '\p{^Is_Nv=+9000}', "");
Expect(1, 65835, '\P{Is_Nv=+9000}', "");
Expect(0, 65835, '\P{^Is_Nv=+9000}', "");
Error('\p{Numeric_Value= /a/0_0_0_0_0_0_0_0_014}');
Error('\P{Numeric_Value= /a/0_0_0_0_0_0_0_0_014}');
Expect(1, 9454, '\p{Numeric_Value=+14}', "");
Expect(0, 9454, '\p{^Numeric_Value=+14}', "");
Expect(0, 9454, '\P{Numeric_Value=+14}', "");
Expect(1, 9454, '\P{^Numeric_Value=+14}', "");
Expect(0, 9455, '\p{Numeric_Value=+14}', "");
Expect(1, 9455, '\p{^Numeric_Value=+14}', "");
Expect(1, 9455, '\P{Numeric_Value=+14}', "");
Expect(0, 9455, '\P{^Numeric_Value=+14}', "");
Error('\p{Nv= /a/00014}');
Error('\P{Nv= /a/00014}');
Expect(1, 9454, '\p{Nv=0014}', "");
Expect(0, 9454, '\p{^Nv=0014}', "");
Expect(0, 9454, '\P{Nv=0014}', "");
Expect(1, 9454, '\P{^Nv=0014}', "");
Expect(0, 9455, '\p{Nv=0014}', "");
Expect(1, 9455, '\p{^Nv=0014}', "");
Expect(1, 9455, '\P{Nv=0014}', "");
Expect(0, 9455, '\P{^Nv=0014}', "");
Error('\p{Is_Numeric_Value: :=_00014}');
Error('\P{Is_Numeric_Value: :=_00014}');
Expect(1, 9454, '\p{Is_Numeric_Value=+0014}', "");
Expect(0, 9454, '\p{^Is_Numeric_Value=+0014}', "");
Expect(0, 9454, '\P{Is_Numeric_Value=+0014}', "");
Expect(1, 9454, '\P{^Is_Numeric_Value=+0014}', "");
Expect(0, 9455, '\p{Is_Numeric_Value=+0014}', "");
Expect(1, 9455, '\p{^Is_Numeric_Value=+0014}', "");
Expect(1, 9455, '\P{Is_Numeric_Value=+0014}', "");
Expect(0, 9455, '\P{^Is_Numeric_Value=+0014}', "");
Error('\p{Is_Nv:   /a/_ 0_0_0_0_0_0_00014}');
Error('\P{Is_Nv:   /a/_ 0_0_0_0_0_0_00014}');
Expect(1, 9454, '\p{Is_Nv=0001_4}', "");
Expect(0, 9454, '\p{^Is_Nv=0001_4}', "");
Expect(0, 9454, '\P{Is_Nv=0001_4}', "");
Expect(1, 9454, '\P{^Is_Nv=0001_4}', "");
Expect(0, 9455, '\p{Is_Nv=0001_4}', "");
Expect(1, 9455, '\p{^Is_Nv=0001_4}', "");
Expect(1, 9455, '\P{Is_Nv=0001_4}', "");
Expect(0, 9455, '\P{^Is_Nv=0001_4}', "");
Error('\p{Numeric_Value=_-000000060000/a/}');
Error('\P{Numeric_Value=_-000000060000/a/}');
Expect(1, 65840, '\p{Numeric_Value=0000_0600_00}', "");
Expect(0, 65840, '\p{^Numeric_Value=0000_0600_00}', "");
Expect(0, 65840, '\P{Numeric_Value=0000_0600_00}', "");
Expect(1, 65840, '\P{^Numeric_Value=0000_0600_00}', "");
Expect(0, 65841, '\p{Numeric_Value=0000_0600_00}', "");
Expect(1, 65841, '\p{^Numeric_Value=0000_0600_00}', "");
Expect(1, 65841, '\P{Numeric_Value=0000_0600_00}', "");
Expect(0, 65841, '\P{^Numeric_Value=0000_0600_00}', "");
Error('\p{Nv=:=	0_6_0_0_00}');
Error('\P{Nv=:=	0_6_0_0_00}');
Expect(1, 65840, '\p{Nv=+0000000060000}', "");
Expect(0, 65840, '\p{^Nv=+0000000060000}', "");
Expect(0, 65840, '\P{Nv=+0000000060000}', "");
Expect(1, 65840, '\P{^Nv=+0000000060000}', "");
Expect(0, 65841, '\p{Nv=+0000000060000}', "");
Expect(1, 65841, '\p{^Nv=+0000000060000}', "");
Expect(1, 65841, '\P{Nv=+0000000060000}', "");
Expect(0, 65841, '\P{^Nv=+0000000060000}', "");
Error('\p{Is_Numeric_Value=	-0000060000/a/}');
Error('\P{Is_Numeric_Value=	-0000060000/a/}');
Expect(1, 65840, '\p{Is_Numeric_Value=00_60_00_0}', "");
Expect(0, 65840, '\p{^Is_Numeric_Value=00_60_00_0}', "");
Expect(0, 65840, '\P{Is_Numeric_Value=00_60_00_0}', "");
Expect(1, 65840, '\P{^Is_Numeric_Value=00_60_00_0}', "");
Expect(0, 65841, '\p{Is_Numeric_Value=00_60_00_0}', "");
Expect(1, 65841, '\p{^Is_Numeric_Value=00_60_00_0}', "");
Expect(1, 65841, '\P{Is_Numeric_Value=00_60_00_0}', "");
Expect(0, 65841, '\P{^Is_Numeric_Value=00_60_00_0}', "");
Error('\p{Is_Nv=--0_0_0_0_0_0_0_0_060000/a/}');
Error('\P{Is_Nv=--0_0_0_0_0_0_0_0_060000/a/}');
Expect(1, 65840, '\p{Is_Nv=0_0_0_0_0_0_0_6_0_000}', "");
Expect(0, 65840, '\p{^Is_Nv=0_0_0_0_0_0_0_6_0_000}', "");
Expect(0, 65840, '\P{Is_Nv=0_0_0_0_0_0_0_6_0_000}', "");
Expect(1, 65840, '\P{^Is_Nv=0_0_0_0_0_0_0_6_0_000}', "");
Expect(0, 65841, '\p{Is_Nv=0_0_0_0_0_0_0_6_0_000}', "");
Expect(1, 65841, '\p{^Is_Nv=0_0_0_0_0_0_0_6_0_000}', "");
Expect(1, 65841, '\P{Is_Nv=0_0_0_0_0_0_0_6_0_000}', "");
Expect(0, 65841, '\P{^Is_Nv=0_0_0_0_0_0_0_6_0_000}', "");
Error('\p{Numeric_Value=	/a/049}');
Error('\P{Numeric_Value=	/a/049}');
Expect(1, 12990, '\p{Numeric_Value=0000000049}', "");
Expect(0, 12990, '\p{^Numeric_Value=0000000049}', "");
Expect(0, 12990, '\P{Numeric_Value=0000000049}', "");
Expect(1, 12990, '\P{^Numeric_Value=0000000049}', "");
Expect(0, 12991, '\p{Numeric_Value=0000000049}', "");
Expect(1, 12991, '\p{^Numeric_Value=0000000049}', "");
Expect(1, 12991, '\P{Numeric_Value=0000000049}', "");
Expect(0, 12991, '\P{^Numeric_Value=0000000049}', "");
Error('\p{Nv=:=-000049}');
Error('\P{Nv=:=-000049}');
Expect(1, 12990, '\p{Nv=000000049}', "");
Expect(0, 12990, '\p{^Nv=000000049}', "");
Expect(0, 12990, '\P{Nv=000000049}', "");
Expect(1, 12990, '\P{^Nv=000000049}', "");
Expect(0, 12991, '\p{Nv=000000049}', "");
Expect(1, 12991, '\p{^Nv=000000049}', "");
Expect(1, 12991, '\P{Nv=000000049}', "");
Expect(0, 12991, '\P{^Nv=000000049}', "");
Error('\p{Is_Numeric_Value= :=+004_9}');
Error('\P{Is_Numeric_Value= :=+004_9}');
Expect(1, 12990, '\p{Is_Numeric_Value=00049}', "");
Expect(0, 12990, '\p{^Is_Numeric_Value=00049}', "");
Expect(0, 12990, '\P{Is_Numeric_Value=00049}', "");
Expect(1, 12990, '\P{^Is_Numeric_Value=00049}', "");
Expect(0, 12991, '\p{Is_Numeric_Value=00049}', "");
Expect(1, 12991, '\p{^Is_Numeric_Value=00049}', "");
Expect(1, 12991, '\P{Is_Numeric_Value=00049}', "");
Expect(0, 12991, '\P{^Is_Numeric_Value=00049}', "");
Error('\p{Is_Nv:	+00_00_00_00_049:=}');
Error('\P{Is_Nv:	+00_00_00_00_049:=}');
Expect(1, 12990, '\p{Is_Nv=4_9}', "");
Expect(0, 12990, '\p{^Is_Nv=4_9}', "");
Expect(0, 12990, '\P{Is_Nv=4_9}', "");
Expect(1, 12990, '\P{^Is_Nv=4_9}', "");
Expect(0, 12991, '\p{Is_Nv=4_9}', "");
Expect(1, 12991, '\p{^Is_Nv=4_9}', "");
Expect(1, 12991, '\P{Is_Nv=4_9}', "");
Expect(0, 12991, '\P{^Is_Nv=4_9}', "");
Error('\p{Numeric_Value=/a/	_000024}');
Error('\P{Numeric_Value=/a/	_000024}');
Expect(1, 12884, '\p{Numeric_Value=0000024}', "");
Expect(0, 12884, '\p{^Numeric_Value=0000024}', "");
Expect(0, 12884, '\P{Numeric_Value=0000024}', "");
Expect(1, 12884, '\P{^Numeric_Value=0000024}', "");
Expect(0, 12885, '\p{Numeric_Value=0000024}', "");
Expect(1, 12885, '\p{^Numeric_Value=0000024}', "");
Expect(1, 12885, '\P{Numeric_Value=0000024}', "");
Expect(0, 12885, '\P{^Numeric_Value=0000024}', "");
Error('\p{Nv= /a/00000000024}');
Error('\P{Nv= /a/00000000024}');
Expect(1, 12884, '\p{Nv=+0_0_0_0_0_0_024}', "");
Expect(0, 12884, '\p{^Nv=+0_0_0_0_0_0_024}', "");
Expect(0, 12884, '\P{Nv=+0_0_0_0_0_0_024}', "");
Expect(1, 12884, '\P{^Nv=+0_0_0_0_0_0_024}', "");
Expect(0, 12885, '\p{Nv=+0_0_0_0_0_0_024}', "");
Expect(1, 12885, '\p{^Nv=+0_0_0_0_0_0_024}', "");
Expect(1, 12885, '\P{Nv=+0_0_0_0_0_0_024}', "");
Expect(0, 12885, '\P{^Nv=+0_0_0_0_0_0_024}', "");
Error('\p{Is_Numeric_Value:	_000_000_000_24:=}');
Error('\P{Is_Numeric_Value:	_000_000_000_24:=}');
Expect(1, 12884, '\p{Is_Numeric_Value=00000024}', "");
Expect(0, 12884, '\p{^Is_Numeric_Value=00000024}', "");
Expect(0, 12884, '\P{Is_Numeric_Value=00000024}', "");
Expect(1, 12884, '\P{^Is_Numeric_Value=00000024}', "");
Expect(0, 12885, '\p{Is_Numeric_Value=00000024}', "");
Expect(1, 12885, '\p{^Is_Numeric_Value=00000024}', "");
Expect(1, 12885, '\P{Is_Numeric_Value=00000024}', "");
Expect(0, 12885, '\P{^Is_Numeric_Value=00000024}', "");
Error('\p{Is_Nv=/a/--0000_0000_24}');
Error('\P{Is_Nv=/a/--0000_0000_24}');
Expect(1, 12884, '\p{Is_Nv=+24}', "");
Expect(0, 12884, '\p{^Is_Nv=+24}', "");
Expect(0, 12884, '\P{Is_Nv=+24}', "");
Expect(1, 12884, '\P{^Is_Nv=+24}', "");
Expect(0, 12885, '\p{Is_Nv=+24}', "");
Expect(1, 12885, '\p{^Is_Nv=+24}', "");
Expect(1, 12885, '\P{Is_Nv=+24}', "");
Expect(0, 12885, '\P{^Is_Nv=+24}', "");
Error('\p{Numeric_Value=-:=+0000001/0000000004}');
Error('\P{Numeric_Value=-:=+0000001/0000000004}');
Expect(1, 74850, '\p{Numeric_Value=+000000001/0000004}', "");
Expect(0, 74850, '\p{^Numeric_Value=+000000001/0000004}', "");
Expect(0, 74850, '\P{Numeric_Value=+000000001/0000004}', "");
Expect(1, 74850, '\P{^Numeric_Value=+000000001/0000004}', "");
Expect(0, 74851, '\p{Numeric_Value=+000000001/0000004}', "");
Expect(1, 74851, '\p{^Numeric_Value=+000000001/0000004}', "");
Expect(1, 74851, '\P{Numeric_Value=+000000001/0000004}', "");
Expect(0, 74851, '\P{^Numeric_Value=+000000001/0000004}', "");
Expect(1, 74850, '\p{Numeric_Value=0.250}', "");
Expect(0, 74850, '\p{^Numeric_Value=0.250}', "");
Expect(0, 74850, '\P{Numeric_Value=0.250}', "");
Expect(1, 74850, '\P{^Numeric_Value=0.250}', "");
Expect(0, 74851, '\p{Numeric_Value=0.250}', "");
Expect(1, 74851, '\p{^Numeric_Value=0.250}', "");
Expect(1, 74851, '\P{Numeric_Value=0.250}', "");
Expect(0, 74851, '\P{^Numeric_Value=0.250}', "");
Error('\p{Nv= 	0001/4:=}');
Error('\P{Nv= 	0001/4:=}');
Expect(1, 74850, '\p{Nv: 00000001/000004}', "");
Expect(0, 74850, '\p{^Nv: 00000001/000004}', "");
Expect(0, 74850, '\P{Nv: 00000001/000004}', "");
Expect(1, 74850, '\P{^Nv: 00000001/000004}', "");
Expect(0, 74851, '\p{Nv: 00000001/000004}', "");
Expect(1, 74851, '\p{^Nv: 00000001/000004}', "");
Expect(1, 74851, '\P{Nv: 00000001/000004}', "");
Expect(0, 74851, '\P{^Nv: 00000001/000004}', "");
Expect(1, 74850, '\p{Nv=0.250}', "");
Expect(0, 74850, '\p{^Nv=0.250}', "");
Expect(0, 74850, '\P{Nv=0.250}', "");
Expect(1, 74850, '\P{^Nv=0.250}', "");
Expect(0, 74851, '\p{Nv=0.250}', "");
Expect(1, 74851, '\p{^Nv=0.250}', "");
Expect(1, 74851, '\P{Nv=0.250}', "");
Expect(0, 74851, '\P{^Nv=0.250}', "");
Error('\p{Is_Numeric_Value=/a/_+0000000001/00000004}');
Error('\P{Is_Numeric_Value=/a/_+0000000001/00000004}');
Expect(1, 74850, '\p{Is_Numeric_Value=000001/00000004}', "");
Expect(0, 74850, '\p{^Is_Numeric_Value=000001/00000004}', "");
Expect(0, 74850, '\P{Is_Numeric_Value=000001/00000004}', "");
Expect(1, 74850, '\P{^Is_Numeric_Value=000001/00000004}', "");
Expect(0, 74851, '\p{Is_Numeric_Value=000001/00000004}', "");
Expect(1, 74851, '\p{^Is_Numeric_Value=000001/00000004}', "");
Expect(1, 74851, '\P{Is_Numeric_Value=000001/00000004}', "");
Expect(0, 74851, '\P{^Is_Numeric_Value=000001/00000004}', "");
Expect(1, 74850, '\p{Is_Numeric_Value=0.250}', "");
Expect(0, 74850, '\p{^Is_Numeric_Value=0.250}', "");
Expect(0, 74850, '\P{Is_Numeric_Value=0.250}', "");
Expect(1, 74850, '\P{^Is_Numeric_Value=0.250}', "");
Expect(0, 74851, '\p{Is_Numeric_Value=0.250}', "");
Expect(1, 74851, '\p{^Is_Numeric_Value=0.250}', "");
Expect(1, 74851, '\P{Is_Numeric_Value=0.250}', "");
Expect(0, 74851, '\P{^Is_Numeric_Value=0.250}', "");
Error('\p{Is_Nv=- 000001/4:=}');
Error('\P{Is_Nv=- 000001/4:=}');
Expect(1, 74850, '\p{Is_Nv=000000001/04}', "");
Expect(0, 74850, '\p{^Is_Nv=000000001/04}', "");
Expect(0, 74850, '\P{Is_Nv=000000001/04}', "");
Expect(1, 74850, '\P{^Is_Nv=000000001/04}', "");
Expect(0, 74851, '\p{Is_Nv=000000001/04}', "");
Expect(1, 74851, '\p{^Is_Nv=000000001/04}', "");
Expect(1, 74851, '\P{Is_Nv=000000001/04}', "");
Expect(0, 74851, '\P{^Is_Nv=000000001/04}', "");
Expect(1, 74850, '\p{Is_Nv: 0.250}', "");
Expect(0, 74850, '\p{^Is_Nv: 0.250}', "");
Expect(0, 74850, '\P{Is_Nv: 0.250}', "");
Expect(1, 74850, '\P{^Is_Nv: 0.250}', "");
Expect(0, 74851, '\p{Is_Nv: 0.250}', "");
Expect(1, 74851, '\p{^Is_Nv: 0.250}', "");
Expect(1, 74851, '\P{Is_Nv: 0.250}', "");
Expect(0, 74851, '\P{^Is_Nv: 0.250}', "");
Error('\p{Numeric_Value=- +90000:=}');
Error('\P{Numeric_Value=- +90000:=}');
Expect(1, 65843, '\p{Numeric_Value=+000090000}', "");
Expect(0, 65843, '\p{^Numeric_Value=+000090000}', "");
Expect(0, 65843, '\P{Numeric_Value=+000090000}', "");
Expect(1, 65843, '\P{^Numeric_Value=+000090000}', "");
Expect(0, 65844, '\p{Numeric_Value=+000090000}', "");
Expect(1, 65844, '\p{^Numeric_Value=+000090000}', "");
Expect(1, 65844, '\P{Numeric_Value=+000090000}', "");
Expect(0, 65844, '\P{^Numeric_Value=+000090000}', "");
Error('\p{Nv=/a/ _0_0_0_9_0000}');
Error('\P{Nv=/a/ _0_0_0_9_0000}');
Expect(1, 65843, '\p{Nv=000000_009000_0}', "");
Expect(0, 65843, '\p{^Nv=000000_009000_0}', "");
Expect(0, 65843, '\P{Nv=000000_009000_0}', "");
Expect(1, 65843, '\P{^Nv=000000_009000_0}', "");
Expect(0, 65844, '\p{Nv=000000_009000_0}', "");
Expect(1, 65844, '\p{^Nv=000000_009000_0}', "");
Expect(1, 65844, '\P{Nv=000000_009000_0}', "");
Expect(0, 65844, '\P{^Nv=000000_009000_0}', "");
Error('\p{Is_Numeric_Value=/a/-_+09_00_00}');
Error('\P{Is_Numeric_Value=/a/-_+09_00_00}');
Expect(1, 65843, '\p{Is_Numeric_Value=+00000000090000}', "");
Expect(0, 65843, '\p{^Is_Numeric_Value=+00000000090000}', "");
Expect(0, 65843, '\P{Is_Numeric_Value=+00000000090000}', "");
Expect(1, 65843, '\P{^Is_Numeric_Value=+00000000090000}', "");
Expect(0, 65844, '\p{Is_Numeric_Value=+00000000090000}', "");
Expect(1, 65844, '\p{^Is_Numeric_Value=+00000000090000}', "");
Expect(1, 65844, '\P{Is_Numeric_Value=+00000000090000}', "");
Expect(0, 65844, '\P{^Is_Numeric_Value=+00000000090000}', "");
Error('\p{Is_Nv=--+0090000:=}');
Error('\P{Is_Nv=--+0090000:=}');
Expect(1, 65843, '\p{Is_Nv=0000000090000}', "");
Expect(0, 65843, '\p{^Is_Nv=0000000090000}', "");
Expect(0, 65843, '\P{Is_Nv=0000000090000}', "");
Expect(1, 65843, '\P{^Is_Nv=0000000090000}', "");
Expect(0, 65844, '\p{Is_Nv=0000000090000}', "");
Expect(1, 65844, '\p{^Is_Nv=0000000090000}', "");
Expect(1, 65844, '\P{Is_Nv=0000000090000}', "");
Expect(0, 65844, '\P{^Is_Nv=0000000090000}', "");
Error('\p{Numeric_Value=:=-05000_0}');
Error('\P{Numeric_Value=:=-05000_0}');
Expect(1, 65878, '\p{Numeric_Value=+00_05_00_00}', "");
Expect(0, 65878, '\p{^Numeric_Value=+00_05_00_00}', "");
Expect(0, 65878, '\P{Numeric_Value=+00_05_00_00}', "");
Expect(1, 65878, '\P{^Numeric_Value=+00_05_00_00}', "");
Expect(0, 65879, '\p{Numeric_Value=+00_05_00_00}', "");
Expect(1, 65879, '\p{^Numeric_Value=+00_05_00_00}', "");
Expect(1, 65879, '\P{Numeric_Value=+00_05_00_00}', "");
Expect(0, 65879, '\P{^Numeric_Value=+00_05_00_00}', "");
Error('\p{Nv= :=00_05_00_00}');
Error('\P{Nv= :=00_05_00_00}');
Expect(1, 65878, '\p{Nv=000000050000}', "");
Expect(0, 65878, '\p{^Nv=000000050000}', "");
Expect(0, 65878, '\P{Nv=000000050000}', "");
Expect(1, 65878, '\P{^Nv=000000050000}', "");
Expect(0, 65879, '\p{Nv=000000050000}', "");
Expect(1, 65879, '\p{^Nv=000000050000}', "");
Expect(1, 65879, '\P{Nv=000000050000}', "");
Expect(0, 65879, '\P{^Nv=000000050000}', "");
Error('\p{Is_Numeric_Value=-/a/+00000050000}');
Error('\P{Is_Numeric_Value=-/a/+00000050000}');
Expect(1, 65878, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_5_0_000}', "");
Expect(0, 65878, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_5_0_000}', "");
Expect(0, 65878, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_5_0_000}', "");
Expect(1, 65878, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_5_0_000}', "");
Expect(0, 65879, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_5_0_000}', "");
Expect(1, 65879, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_5_0_000}', "");
Expect(1, 65879, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_5_0_000}', "");
Expect(0, 65879, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_5_0_000}', "");
Error('\p{Is_Nv=:=-	0_0_5_0_0_00}');
Error('\P{Is_Nv=:=-	0_0_5_0_0_00}');
Expect(1, 65878, '\p{Is_Nv=05000_0}', "");
Expect(0, 65878, '\p{^Is_Nv=05000_0}', "");
Expect(0, 65878, '\P{Is_Nv=05000_0}', "");
Expect(1, 65878, '\P{^Is_Nv=05000_0}', "");
Expect(0, 65879, '\p{Is_Nv=05000_0}', "");
Expect(1, 65879, '\p{^Is_Nv=05000_0}', "");
Expect(1, 65879, '\P{Is_Nv=05000_0}', "");
Expect(0, 65879, '\P{^Is_Nv=05000_0}', "");
Error('\p{Numeric_Value=  +00_00_00_40_00_0/a/}');
Error('\P{Numeric_Value=  +00_00_00_40_00_0/a/}');
Expect(1, 65838, '\p{Numeric_Value=040000}', "");
Expect(0, 65838, '\p{^Numeric_Value=040000}', "");
Expect(0, 65838, '\P{Numeric_Value=040000}', "");
Expect(1, 65838, '\P{^Numeric_Value=040000}', "");
Expect(0, 65839, '\p{Numeric_Value=040000}', "");
Expect(1, 65839, '\p{^Numeric_Value=040000}', "");
Expect(1, 65839, '\P{Numeric_Value=040000}', "");
Expect(0, 65839, '\P{^Numeric_Value=040000}', "");
Error('\p{Nv=/a/0_0_0_0_0_4_0_000}');
Error('\P{Nv=/a/0_0_0_0_0_4_0_000}');
Expect(1, 65838, '\p{Nv: 40000}', "");
Expect(0, 65838, '\p{^Nv: 40000}', "");
Expect(0, 65838, '\P{Nv: 40000}', "");
Expect(1, 65838, '\P{^Nv: 40000}', "");
Expect(0, 65839, '\p{Nv: 40000}', "");
Expect(1, 65839, '\p{^Nv: 40000}', "");
Expect(1, 65839, '\P{Nv: 40000}', "");
Expect(0, 65839, '\P{^Nv: 40000}', "");
Error('\p{Is_Numeric_Value=/a/ +000040000}');
Error('\P{Is_Numeric_Value=/a/ +000040000}');
Expect(1, 65838, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_0_0_4_0000}', "");
Expect(0, 65838, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_0_0_4_0000}', "");
Expect(0, 65838, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_0_0_4_0000}', "");
Expect(1, 65838, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_0_0_4_0000}', "");
Expect(0, 65839, '\p{Is_Numeric_Value=0_0_0_0_0_0_0_0_0_4_0000}', "");
Expect(1, 65839, '\p{^Is_Numeric_Value=0_0_0_0_0_0_0_0_0_4_0000}', "");
Expect(1, 65839, '\P{Is_Numeric_Value=0_0_0_0_0_0_0_0_0_4_0000}', "");
Expect(0, 65839, '\P{^Is_Numeric_Value=0_0_0_0_0_0_0_0_0_4_0000}', "");
Error('\p{Is_Nv=:=00040000}');
Error('\P{Is_Nv=:=00040000}');
Expect(1, 65838, '\p{Is_Nv=0_0_0_4_0000}', "");
Expect(0, 65838, '\p{^Is_Nv=0_0_0_4_0000}', "");
Expect(0, 65838, '\P{Is_Nv=0_0_0_4_0000}', "");
Expect(1, 65838, '\P{^Is_Nv=0_0_0_4_0000}', "");
Expect(0, 65839, '\p{Is_Nv=0_0_0_4_0000}', "");
Expect(1, 65839, '\p{^Is_Nv=0_0_0_4_0000}', "");
Expect(1, 65839, '\P{Is_Nv=0_0_0_4_0000}', "");
Expect(0, 65839, '\P{^Is_Nv=0_0_0_4_0000}', "");
Error('\p{Numeric_Value=:=-0001/00002}');
Error('\P{Numeric_Value=:=-0001/00002}');
Expect(1, 69243, '\p{Numeric_Value=+000001/000000002}', "");
Expect(0, 69243, '\p{^Numeric_Value=+000001/000000002}', "");
Expect(0, 69243, '\P{Numeric_Value=+000001/000000002}', "");
Expect(1, 69243, '\P{^Numeric_Value=+000001/000000002}', "");
Expect(0, 69244, '\p{Numeric_Value=+000001/000000002}', "");
Expect(1, 69244, '\p{^Numeric_Value=+000001/000000002}', "");
Expect(1, 69244, '\P{Numeric_Value=+000001/000000002}', "");
Expect(0, 69244, '\P{^Numeric_Value=+000001/000000002}', "");
Expect(1, 69243, '\p{Numeric_Value=0.500}', "");
Expect(0, 69243, '\p{^Numeric_Value=0.500}', "");
Expect(0, 69243, '\P{Numeric_Value=0.500}', "");
Expect(1, 69243, '\P{^Numeric_Value=0.500}', "");
Expect(0, 69244, '\p{Numeric_Value=0.500}', "");
Expect(1, 69244, '\p{^Numeric_Value=0.500}', "");
Expect(1, 69244, '\P{Numeric_Value=0.500}', "");
Expect(0, 69244, '\P{^Numeric_Value=0.500}', "");
Error('\p{Nv=	/a/+000000001/00000002}');
Error('\P{Nv=	/a/+000000001/00000002}');
Expect(1, 69243, '\p{Nv:	+00001/0002}', "");
Expect(0, 69243, '\p{^Nv:	+00001/0002}', "");
Expect(0, 69243, '\P{Nv:	+00001/0002}', "");
Expect(1, 69243, '\P{^Nv:	+00001/0002}', "");
Expect(0, 69244, '\p{Nv:	+00001/0002}', "");
Expect(1, 69244, '\p{^Nv:	+00001/0002}', "");
Expect(1, 69244, '\P{Nv:	+00001/0002}', "");
Expect(0, 69244, '\P{^Nv:	+00001/0002}', "");
Expect(1, 69243, '\p{Nv=0.500}', "");
Expect(0, 69243, '\p{^Nv=0.500}', "");
Expect(0, 69243, '\P{Nv=0.500}', "");
Expect(1, 69243, '\P{^Nv=0.500}', "");
Expect(0, 69244, '\p{Nv=0.500}', "");
Expect(1, 69244, '\p{^Nv=0.500}', "");
Expect(1, 69244, '\P{Nv=0.500}', "");
Expect(0, 69244, '\P{^Nv=0.500}', "");
Error('\p{Is_Numeric_Value: :=_0000001/2}');
Error('\P{Is_Numeric_Value: :=_0000001/2}');
Expect(1, 69243, '\p{Is_Numeric_Value=+1/000002}', "");
Expect(0, 69243, '\p{^Is_Numeric_Value=+1/000002}', "");
Expect(0, 69243, '\P{Is_Numeric_Value=+1/000002}', "");
Expect(1, 69243, '\P{^Is_Numeric_Value=+1/000002}', "");
Expect(0, 69244, '\p{Is_Numeric_Value=+1/000002}', "");
Expect(1, 69244, '\p{^Is_Numeric_Value=+1/000002}', "");
Expect(1, 69244, '\P{Is_Numeric_Value=+1/000002}', "");
Expect(0, 69244, '\P{^Is_Numeric_Value=+1/000002}', "");
Expect(1, 69243, '\p{Is_Numeric_Value=0.500}', "");
Expect(0, 69243, '\p{^Is_Numeric_Value=0.500}', "");
Expect(0, 69243, '\P{Is_Numeric_Value=0.500}', "");
Expect(1, 69243, '\P{^Is_Numeric_Value=0.500}', "");
Expect(0, 69244, '\p{Is_Numeric_Value=0.500}', "");
Expect(1, 69244, '\p{^Is_Numeric_Value=0.500}', "");
Expect(1, 69244, '\P{Is_Numeric_Value=0.500}', "");
Expect(0, 69244, '\P{^Is_Numeric_Value=0.500}', "");
Error('\p{Is_Nv=:=0000000001/002}');
Error('\P{Is_Nv=:=0000000001/002}');
Expect(1, 69243, '\p{Is_Nv=+0000000001/002}', "");
Expect(0, 69243, '\p{^Is_Nv=+0000000001/002}', "");
Expect(0, 69243, '\P{Is_Nv=+0000000001/002}', "");
Expect(1, 69243, '\P{^Is_Nv=+0000000001/002}', "");
Expect(0, 69244, '\p{Is_Nv=+0000000001/002}', "");
Expect(1, 69244, '\p{^Is_Nv=+0000000001/002}', "");
Expect(1, 69244, '\P{Is_Nv=+0000000001/002}', "");
Expect(0, 69244, '\P{^Is_Nv=+0000000001/002}', "");
Expect(1, 69243, '\p{Is_Nv=0.500}', "");
Expect(0, 69243, '\p{^Is_Nv=0.500}', "");
Expect(0, 69243, '\P{Is_Nv=0.500}', "");
Expect(1, 69243, '\P{^Is_Nv=0.500}', "");
Expect(0, 69244, '\p{Is_Nv=0.500}', "");
Expect(1, 69244, '\p{^Is_Nv=0.500}', "");
Expect(1, 69244, '\P{Is_Nv=0.500}', "");
Expect(0, 69244, '\P{^Is_Nv=0.500}', "");
Error('\p{Numeric_Value= :=000000007000}');
Error('\P{Numeric_Value= :=000000007000}');
Expect(1, 65832, '\p{Numeric_Value=+07000}', "");
Expect(0, 65832, '\p{^Numeric_Value=+07000}', "");
Expect(0, 65832, '\P{Numeric_Value=+07000}', "");
Expect(1, 65832, '\P{^Numeric_Value=+07000}', "");
Expect(0, 65833, '\p{Numeric_Value=+07000}', "");
Expect(1, 65833, '\p{^Numeric_Value=+07000}', "");
Expect(1, 65833, '\P{Numeric_Value=+07000}', "");
Expect(0, 65833, '\P{^Numeric_Value=+07000}', "");
Error('\p{Nv=-/a/+0000000007000}');
Error('\P{Nv=-/a/+0000000007000}');
Expect(1, 65832, '\p{Nv: +000_000_700_0}', "");
Expect(0, 65832, '\p{^Nv: +000_000_700_0}', "");
Expect(0, 65832, '\P{Nv: +000_000_700_0}', "");
Expect(1, 65832, '\P{^Nv: +000_000_700_0}', "");
Expect(0, 65833, '\p{Nv: +000_000_700_0}', "");
Expect(1, 65833, '\p{^Nv: +000_000_700_0}', "");
Expect(1, 65833, '\P{Nv: +000_000_700_0}', "");
Expect(0, 65833, '\P{^Nv: +000_000_700_0}', "");
Error('\p{Is_Numeric_Value=	0_7_0_00/a/}');
Error('\P{Is_Numeric_Value=	0_7_0_00/a/}');
Expect(1, 65832, '\p{Is_Numeric_Value=00_00_00_07_000}', "");
Expect(0, 65832, '\p{^Is_Numeric_Value=00_00_00_07_000}', "");
Expect(0, 65832, '\P{Is_Numeric_Value=00_00_00_07_000}', "");
Expect(1, 65832, '\P{^Is_Numeric_Value=00_00_00_07_000}', "");
Expect(0, 65833, '\p{Is_Numeric_Value=00_00_00_07_000}', "");
Expect(1, 65833, '\p{^Is_Numeric_Value=00_00_00_07_000}', "");
Expect(1, 65833, '\P{Is_Numeric_Value=00_00_00_07_000}', "");
Expect(0, 65833, '\P{^Is_Numeric_Value=00_00_00_07_000}', "");
Error('\p{Is_Nv=/a/_ +00_00_00_07_000}');
Error('\P{Is_Nv=/a/_ +00_00_00_07_000}');
Expect(1, 65832, '\p{Is_Nv=00_00_00_00_07_000}', "");
Expect(0, 65832, '\p{^Is_Nv=00_00_00_00_07_000}', "");
Expect(0, 65832, '\P{Is_Nv=00_00_00_00_07_000}', "");
Expect(1, 65832, '\P{^Is_Nv=00_00_00_00_07_000}', "");
Expect(0, 65833, '\p{Is_Nv=00_00_00_00_07_000}', "");
Expect(1, 65833, '\p{^Is_Nv=00_00_00_00_07_000}', "");
Expect(1, 65833, '\P{Is_Nv=00_00_00_00_07_000}', "");
Expect(0, 65833, '\P{^Is_Nv=00_00_00_00_07_000}', "");
Error('\p{Numeric_Value=/a/	 00000003/005}');
Error('\P{Numeric_Value=/a/	 00000003/005}');
Expect(1, 8535, '\p{Numeric_Value=+03/0000000005}', "");
Expect(0, 8535, '\p{^Numeric_Value=+03/0000000005}', "");
Expect(0, 8535, '\P{Numeric_Value=+03/0000000005}', "");
Expect(1, 8535, '\P{^Numeric_Value=+03/0000000005}', "");
Expect(0, 8536, '\p{Numeric_Value=+03/0000000005}', "");
Expect(1, 8536, '\p{^Numeric_Value=+03/0000000005}', "");
Expect(1, 8536, '\P{Numeric_Value=+03/0000000005}', "");
Expect(0, 8536, '\P{^Numeric_Value=+03/0000000005}', "");
Expect(1, 8535, '\p{Numeric_Value:   0.600}', "");
Expect(0, 8535, '\p{^Numeric_Value:   0.600}', "");
Expect(0, 8535, '\P{Numeric_Value:   0.600}', "");
Expect(1, 8535, '\P{^Numeric_Value:   0.600}', "");
Expect(0, 8536, '\p{Numeric_Value:   0.600}', "");
Expect(1, 8536, '\p{^Numeric_Value:   0.600}', "");
Expect(1, 8536, '\P{Numeric_Value:   0.600}', "");
Expect(0, 8536, '\P{^Numeric_Value:   0.600}', "");
Error('\p{Nv=_-00000003/00000005:=}');
Error('\P{Nv=_-00000003/00000005:=}');
Expect(1, 8535, '\p{Nv=0003/05}', "");
Expect(0, 8535, '\p{^Nv=0003/05}', "");
Expect(0, 8535, '\P{Nv=0003/05}', "");
Expect(1, 8535, '\P{^Nv=0003/05}', "");
Expect(0, 8536, '\p{Nv=0003/05}', "");
Expect(1, 8536, '\p{^Nv=0003/05}', "");
Expect(1, 8536, '\P{Nv=0003/05}', "");
Expect(0, 8536, '\P{^Nv=0003/05}', "");
Expect(1, 8535, '\p{Nv=0.600}', "");
Expect(0, 8535, '\p{^Nv=0.600}', "");
Expect(0, 8535, '\P{Nv=0.600}', "");
Expect(1, 8535, '\P{^Nv=0.600}', "");
Expect(0, 8536, '\p{Nv=0.600}', "");
Expect(1, 8536, '\p{^Nv=0.600}', "");
Expect(1, 8536, '\P{Nv=0.600}', "");
Expect(0, 8536, '\P{^Nv=0.600}', "");
Error('\p{Is_Numeric_Value=-0000000003/5/a/}');
Error('\P{Is_Numeric_Value=-0000000003/5/a/}');
Expect(1, 8535, '\p{Is_Numeric_Value=003/05}', "");
Expect(0, 8535, '\p{^Is_Numeric_Value=003/05}', "");
Expect(0, 8535, '\P{Is_Numeric_Value=003/05}', "");
Expect(1, 8535, '\P{^Is_Numeric_Value=003/05}', "");
Expect(0, 8536, '\p{Is_Numeric_Value=003/05}', "");
Expect(1, 8536, '\p{^Is_Numeric_Value=003/05}', "");
Expect(1, 8536, '\P{Is_Numeric_Value=003/05}', "");
Expect(0, 8536, '\P{^Is_Numeric_Value=003/05}', "");
Expect(1, 8535, '\p{Is_Numeric_Value=0.600}', "");
Expect(0, 8535, '\p{^Is_Numeric_Value=0.600}', "");
Expect(0, 8535, '\P{Is_Numeric_Value=0.600}', "");
Expect(1, 8535, '\P{^Is_Numeric_Value=0.600}', "");
Expect(0, 8536, '\p{Is_Numeric_Value=0.600}', "");
Expect(1, 8536, '\p{^Is_Numeric_Value=0.600}', "");
Expect(1, 8536, '\P{Is_Numeric_Value=0.600}', "");
Expect(0, 8536, '\P{^Is_Numeric_Value=0.600}', "");
Error('\p{Is_Nv= /a/000003/0000005}');
Error('\P{Is_Nv= /a/000003/0000005}');
Expect(1, 8535, '\p{Is_Nv: 0003/0000000005}', "");
Expect(0, 8535, '\p{^Is_Nv: 0003/0000000005}', "");
Expect(0, 8535, '\P{Is_Nv: 0003/0000000005}', "");
Expect(1, 8535, '\P{^Is_Nv: 0003/0000000005}', "");
Expect(0, 8536, '\p{Is_Nv: 0003/0000000005}', "");
Expect(1, 8536, '\p{^Is_Nv: 0003/0000000005}', "");
Expect(1, 8536, '\P{Is_Nv: 0003/0000000005}', "");
Expect(0, 8536, '\P{^Is_Nv: 0003/0000000005}', "");
Expect(1, 8535, '\p{Is_Nv=0.600}', "");
Expect(0, 8535, '\p{^Is_Nv=0.600}', "");
Expect(0, 8535, '\P{Is_Nv=0.600}', "");
Expect(1, 8535, '\P{^Is_Nv=0.600}', "");
Expect(0, 8536, '\p{Is_Nv=0.600}', "");
Expect(1, 8536, '\p{^Is_Nv=0.600}', "");
Expect(1, 8536, '\P{Is_Nv=0.600}', "");
Expect(0, 8536, '\P{^Is_Nv=0.600}', "");
Error('\p{Numeric_Value=-/a/+0017/0000000002}');
Error('\P{Numeric_Value=-/a/+0017/0000000002}');
Expect(1, 3890, '\p{Numeric_Value:	000000017/0002}', "");
Expect(0, 3890, '\p{^Numeric_Value:	000000017/0002}', "");
Expect(0, 3890, '\P{Numeric_Value:	000000017/0002}', "");
Expect(1, 3890, '\P{^Numeric_Value:	000000017/0002}', "");
Expect(0, 3891, '\p{Numeric_Value:	000000017/0002}', "");
Expect(1, 3891, '\p{^Numeric_Value:	000000017/0002}', "");
Expect(1, 3891, '\P{Numeric_Value:	000000017/0002}', "");
Expect(0, 3891, '\P{^Numeric_Value:	000000017/0002}', "");
Expect(1, 3890, '\p{Numeric_Value=8.500}', "");
Expect(0, 3890, '\p{^Numeric_Value=8.500}', "");
Expect(0, 3890, '\P{Numeric_Value=8.500}', "");
Expect(1, 3890, '\P{^Numeric_Value=8.500}', "");
Expect(0, 3891, '\p{Numeric_Value=8.500}', "");
Expect(1, 3891, '\p{^Numeric_Value=8.500}', "");
Expect(1, 3891, '\P{Numeric_Value=8.500}', "");
Expect(0, 3891, '\P{^Numeric_Value=8.500}', "");
Error('\p{Nv=	000017/02:=}');
Error('\P{Nv=	000017/02:=}');
Expect(1, 3890, '\p{Nv=00000000017/0002}', "");
Expect(0, 3890, '\p{^Nv=00000000017/0002}', "");
Expect(0, 3890, '\P{Nv=00000000017/0002}', "");
Expect(1, 3890, '\P{^Nv=00000000017/0002}', "");
Expect(0, 3891, '\p{Nv=00000000017/0002}', "");
Expect(1, 3891, '\p{^Nv=00000000017/0002}', "");
Expect(1, 3891, '\P{Nv=00000000017/0002}', "");
Expect(0, 3891, '\P{^Nv=00000000017/0002}', "");
Expect(1, 3890, '\p{Nv=8.500}', "");
Expect(0, 3890, '\p{^Nv=8.500}', "");
Expect(0, 3890, '\P{Nv=8.500}', "");
Expect(1, 3890, '\P{^Nv=8.500}', "");
Expect(0, 3891, '\p{Nv=8.500}', "");
Expect(1, 3891, '\p{^Nv=8.500}', "");
Expect(1, 3891, '\P{Nv=8.500}', "");
Expect(0, 3891, '\P{^Nv=8.500}', "");
Error('\p{Is_Numeric_Value=:=	-00017/0000002}');
Error('\P{Is_Numeric_Value=:=	-00017/0000002}');
Expect(1, 3890, '\p{Is_Numeric_Value=+000000017/0000000002}', "");
Expect(0, 3890, '\p{^Is_Numeric_Value=+000000017/0000000002}', "");
Expect(0, 3890, '\P{Is_Numeric_Value=+000000017/0000000002}', "");
Expect(1, 3890, '\P{^Is_Numeric_Value=+000000017/0000000002}', "");
Expect(0, 3891, '\p{Is_Numeric_Value=+000000017/0000000002}', "");
Expect(1, 3891, '\p{^Is_Numeric_Value=+000000017/0000000002}', "");
Expect(1, 3891, '\P{Is_Numeric_Value=+000000017/0000000002}', "");
Expect(0, 3891, '\P{^Is_Numeric_Value=+000000017/0000000002}', "");
Expect(1, 3890, '\p{Is_Numeric_Value=8.500}', "");
Expect(0, 3890, '\p{^Is_Numeric_Value=8.500}', "");
Expect(0, 3890, '\P{Is_Numeric_Value=8.500}', "");
Expect(1, 3890, '\P{^Is_Numeric_Value=8.500}', "");
Expect(0, 3891, '\p{Is_Numeric_Value=8.500}', "");
Expect(1, 3891, '\p{^Is_Numeric_Value=8.500}', "");
Expect(1, 3891, '\P{Is_Numeric_Value=8.500}', "");
Expect(0, 3891, '\P{^Is_Numeric_Value=8.500}', "");
Error('\p{Is_Nv=-	0017/002:=}');
Error('\P{Is_Nv=-	0017/002:=}');
Expect(1, 3890, '\p{Is_Nv=0017/0000002}', "");
Expect(0, 3890, '\p{^Is_Nv=0017/0000002}', "");
Expect(0, 3890, '\P{Is_Nv=0017/0000002}', "");
Expect(1, 3890, '\P{^Is_Nv=0017/0000002}', "");
Expect(0, 3891, '\p{Is_Nv=0017/0000002}', "");
Expect(1, 3891, '\p{^Is_Nv=0017/0000002}', "");
Expect(1, 3891, '\P{Is_Nv=0017/0000002}', "");
Expect(0, 3891, '\P{^Is_Nv=0017/0000002}', "");
Expect(1, 3890, '\p{Is_Nv=8.500}', "");
Expect(0, 3890, '\p{^Is_Nv=8.500}', "");
Expect(0, 3890, '\P{Is_Nv=8.500}', "");
Expect(1, 3890, '\P{^Is_Nv=8.500}', "");
Expect(0, 3891, '\p{Is_Nv=8.500}', "");
Expect(1, 3891, '\p{^Is_Nv=8.500}', "");
Expect(1, 3891, '\P{Is_Nv=8.500}', "");
Expect(0, 3891, '\P{^Is_Nv=8.500}', "");
Error('\p{Numeric_Value=:=- 000000042}');
Error('\P{Numeric_Value=:=- 000000042}');
Expect(1, 12983, '\p{Numeric_Value=00000000042}', "");
Expect(0, 12983, '\p{^Numeric_Value=00000000042}', "");
Expect(0, 12983, '\P{Numeric_Value=00000000042}', "");
Expect(1, 12983, '\P{^Numeric_Value=00000000042}', "");
Expect(0, 12984, '\p{Numeric_Value=00000000042}', "");
Expect(1, 12984, '\p{^Numeric_Value=00000000042}', "");
Expect(1, 12984, '\P{Numeric_Value=00000000042}', "");
Expect(0, 12984, '\P{^Numeric_Value=00000000042}', "");
Error('\p{Nv=_-00_00_00_00_042/a/}');
Error('\P{Nv=_-00_00_00_00_042/a/}');
Expect(1, 12983, '\p{Nv: 00000000042}', "");
Expect(0, 12983, '\p{^Nv: 00000000042}', "");
Expect(0, 12983, '\P{Nv: 00000000042}', "");
Expect(1, 12983, '\P{^Nv: 00000000042}', "");
Expect(0, 12984, '\p{Nv: 00000000042}', "");
Expect(1, 12984, '\p{^Nv: 00000000042}', "");
Expect(1, 12984, '\P{Nv: 00000000042}', "");
Expect(0, 12984, '\P{^Nv: 00000000042}', "");
Error('\p{Is_Numeric_Value:	/a/000042}');
Error('\P{Is_Numeric_Value:	/a/000042}');
Expect(1, 12983, '\p{Is_Numeric_Value=00042}', "");
Expect(0, 12983, '\p{^Is_Numeric_Value=00042}', "");
Expect(0, 12983, '\P{Is_Numeric_Value=00042}', "");
Expect(1, 12983, '\P{^Is_Numeric_Value=00042}', "");
Expect(0, 12984, '\p{Is_Numeric_Value=00042}', "");
Expect(1, 12984, '\p{^Is_Numeric_Value=00042}', "");
Expect(1, 12984, '\P{Is_Numeric_Value=00042}', "");
Expect(0, 12984, '\P{^Is_Numeric_Value=00042}', "");
Error('\p{Is_Nv=:=-	000042}');
Error('\P{Is_Nv=:=-	000042}');
Expect(1, 12983, '\p{Is_Nv=+04_2}', "");
Expect(0, 12983, '\p{^Is_Nv=+04_2}', "");
Expect(0, 12983, '\P{Is_Nv=+04_2}', "");
Expect(1, 12983, '\P{^Is_Nv=+04_2}', "");
Expect(0, 12984, '\p{Is_Nv=+04_2}', "");
Expect(1, 12984, '\p{^Is_Nv=+04_2}', "");
Expect(1, 12984, '\P{Is_Nv=+04_2}', "");
Expect(0, 12984, '\P{^Is_Nv=+04_2}', "");
Error('\p{Numeric_Value:-_22/a/}');
Error('\P{Numeric_Value:-_22/a/}');
Expect(1, 12882, '\p{Numeric_Value=0000022}', "");
Expect(0, 12882, '\p{^Numeric_Value=0000022}', "");
Expect(0, 12882, '\P{Numeric_Value=0000022}', "");
Expect(1, 12882, '\P{^Numeric_Value=0000022}', "");
Expect(0, 12883, '\p{Numeric_Value=0000022}', "");
Expect(1, 12883, '\p{^Numeric_Value=0000022}', "");
Expect(1, 12883, '\P{Numeric_Value=0000022}', "");
Expect(0, 12883, '\P{^Numeric_Value=0000022}', "");
Error('\p{Nv=__00000022:=}');
Error('\P{Nv=__00000022:=}');
Expect(1, 12882, '\p{Nv=0000000022}', "");
Expect(0, 12882, '\p{^Nv=0000000022}', "");
Expect(0, 12882, '\P{Nv=0000000022}', "");
Expect(1, 12882, '\P{^Nv=0000000022}', "");
Expect(0, 12883, '\p{Nv=0000000022}', "");
Expect(1, 12883, '\p{^Nv=0000000022}', "");
Expect(1, 12883, '\P{Nv=0000000022}', "");
Expect(0, 12883, '\P{^Nv=0000000022}', "");
Error('\p{Is_Numeric_Value=0022:=}');
Error('\P{Is_Numeric_Value=0022:=}');
Expect(1, 12882, '\p{Is_Numeric_Value=000000022}', "");
Expect(0, 12882, '\p{^Is_Numeric_Value=000000022}', "");
Expect(0, 12882, '\P{Is_Numeric_Value=000000022}', "");
Expect(1, 12882, '\P{^Is_Numeric_Value=000000022}', "");
Expect(0, 12883, '\p{Is_Numeric_Value=000000022}', "");
Expect(1, 12883, '\p{^Is_Numeric_Value=000000022}', "");
Expect(1, 12883, '\P{Is_Numeric_Value=000000022}', "");
Expect(0, 12883, '\P{^Is_Numeric_Value=000000022}', "");
Error('\p{Is_Nv=_+0022/a/}');
Error('\P{Is_Nv=_+0022/a/}');
Expect(1, 12882, '\p{Is_Nv=22}', "");
Expect(0, 12882, '\p{^Is_Nv=22}', "");
Expect(0, 12882, '\P{Is_Nv=22}', "");
Expect(1, 12882, '\P{^Is_Nv=22}', "");
Expect(0, 12883, '\p{Is_Nv=22}', "");
Expect(1, 12883, '\p{^Is_Nv=22}', "");
Expect(1, 12883, '\P{Is_Nv=22}', "");
Expect(0, 12883, '\P{^Is_Nv=22}', "");
Error('\p{Numeric_Value=00001/000000009/a/}');
Error('\P{Numeric_Value=00001/000000009/a/}');
Expect(1, 8529, '\p{Numeric_Value=+0001/009}', "");
Expect(0, 8529, '\p{^Numeric_Value=+0001/009}', "");
Expect(0, 8529, '\P{Numeric_Value=+0001/009}', "");
Expect(1, 8529, '\P{^Numeric_Value=+0001/009}', "");
Expect(0, 8530, '\p{Numeric_Value=+0001/009}', "");
Expect(1, 8530, '\p{^Numeric_Value=+0001/009}', "");
Expect(1, 8530, '\P{Numeric_Value=+0001/009}', "");
Expect(0, 8530, '\P{^Numeric_Value=+0001/009}', "");
Error('\p{Numeric_Value=0.11}');
Error('\P{Numeric_Value=0.11}');
Expect(1, 8529, '\p{Numeric_Value:	0.111}', "");
Expect(0, 8529, '\p{^Numeric_Value:	0.111}', "");
Expect(0, 8529, '\P{Numeric_Value:	0.111}', "");
Expect(1, 8529, '\P{^Numeric_Value:	0.111}', "");
Expect(0, 8530, '\p{Numeric_Value:	0.111}', "");
Expect(1, 8530, '\p{^Numeric_Value:	0.111}', "");
Expect(1, 8530, '\P{Numeric_Value:	0.111}', "");
Expect(0, 8530, '\P{^Numeric_Value:	0.111}', "");
Error('\p{Nv=/a/--0000000001/000000009}');
Error('\P{Nv=/a/--0000000001/000000009}');
Expect(1, 8529, '\p{Nv=0000000001/0000009}', "");
Expect(0, 8529, '\p{^Nv=0000000001/0000009}', "");
Expect(0, 8529, '\P{Nv=0000000001/0000009}', "");
Expect(1, 8529, '\P{^Nv=0000000001/0000009}', "");
Expect(0, 8530, '\p{Nv=0000000001/0000009}', "");
Expect(1, 8530, '\p{^Nv=0000000001/0000009}', "");
Expect(1, 8530, '\P{Nv=0000000001/0000009}', "");
Expect(0, 8530, '\P{^Nv=0000000001/0000009}', "");
Error('\p{Nv=0.11}');
Error('\P{Nv=0.11}');
Expect(1, 8529, '\p{Nv=0.111}', "");
Expect(0, 8529, '\p{^Nv=0.111}', "");
Expect(0, 8529, '\P{Nv=0.111}', "");
Expect(1, 8529, '\P{^Nv=0.111}', "");
Expect(0, 8530, '\p{Nv=0.111}', "");
Expect(1, 8530, '\p{^Nv=0.111}', "");
Expect(1, 8530, '\P{Nv=0.111}', "");
Expect(0, 8530, '\P{^Nv=0.111}', "");
Error('\p{Is_Numeric_Value=- +0001/00009:=}');
Error('\P{Is_Numeric_Value=- +0001/00009:=}');
Expect(1, 8529, '\p{Is_Numeric_Value=0000001/000009}', "");
Expect(0, 8529, '\p{^Is_Numeric_Value=0000001/000009}', "");
Expect(0, 8529, '\P{Is_Numeric_Value=0000001/000009}', "");
Expect(1, 8529, '\P{^Is_Numeric_Value=0000001/000009}', "");
Expect(0, 8530, '\p{Is_Numeric_Value=0000001/000009}', "");
Expect(1, 8530, '\p{^Is_Numeric_Value=0000001/000009}', "");
Expect(1, 8530, '\P{Is_Numeric_Value=0000001/000009}', "");
Expect(0, 8530, '\P{^Is_Numeric_Value=0000001/000009}', "");
Error('\p{Is_Numeric_Value=0.11}');
Error('\P{Is_Numeric_Value=0.11}');
Expect(1, 8529, '\p{Is_Numeric_Value=0.111}', "");
Expect(0, 8529, '\p{^Is_Numeric_Value=0.111}', "");
Expect(0, 8529, '\P{Is_Numeric_Value=0.111}', "");
Expect(1, 8529, '\P{^Is_Numeric_Value=0.111}', "");
Expect(0, 8530, '\p{Is_Numeric_Value=0.111}', "");
Expect(1, 8530, '\p{^Is_Numeric_Value=0.111}', "");
Expect(1, 8530, '\P{Is_Numeric_Value=0.111}', "");
Expect(0, 8530, '\P{^Is_Numeric_Value=0.111}', "");
Error('\p{Is_Nv=/a/	-001/00009}');
Error('\P{Is_Nv=/a/	-001/00009}');
Expect(1, 8529, '\p{Is_Nv=000000001/0000000009}', "");
Expect(0, 8529, '\p{^Is_Nv=000000001/0000000009}', "");
Expect(0, 8529, '\P{Is_Nv=000000001/0000000009}', "");
Expect(1, 8529, '\P{^Is_Nv=000000001/0000000009}', "");
Expect(0, 8530, '\p{Is_Nv=000000001/0000000009}', "");
Expect(1, 8530, '\p{^Is_Nv=000000001/0000000009}', "");
Expect(1, 8530, '\P{Is_Nv=000000001/0000000009}', "");
Expect(0, 8530, '\P{^Is_Nv=000000001/0000000009}', "");
Error('\p{Is_Nv:	0.11}');
Error('\P{Is_Nv:	0.11}');
Expect(1, 8529, '\p{Is_Nv=0.111}', "");
Expect(0, 8529, '\p{^Is_Nv=0.111}', "");
Expect(0, 8529, '\P{Is_Nv=0.111}', "");
Expect(1, 8529, '\P{^Is_Nv=0.111}', "");
Expect(0, 8530, '\p{Is_Nv=0.111}', "");
Expect(1, 8530, '\p{^Is_Nv=0.111}', "");
Expect(1, 8530, '\P{Is_Nv=0.111}', "");
Expect(0, 8530, '\P{^Is_Nv=0.111}', "");
Error('\p{Numeric_Value=/a/0_0_9_00}');
Error('\P{Numeric_Value=/a/0_0_9_00}');
Expect(1, 69242, '\p{Numeric_Value=000000900}', "");
Expect(0, 69242, '\p{^Numeric_Value=000000900}', "");
Expect(0, 69242, '\P{Numeric_Value=000000900}', "");
Expect(1, 69242, '\P{^Numeric_Value=000000900}', "");
Expect(0, 69243, '\p{Numeric_Value=000000900}', "");
Expect(1, 69243, '\p{^Numeric_Value=000000900}', "");
Expect(1, 69243, '\P{Numeric_Value=000000900}', "");
Expect(0, 69243, '\P{^Numeric_Value=000000900}', "");
Error('\p{Nv=/a/00000_00090_0}');
Error('\P{Nv=/a/00000_00090_0}');
Expect(1, 69242, '\p{Nv=00900}', "");
Expect(0, 69242, '\p{^Nv=00900}', "");
Expect(0, 69242, '\P{Nv=00900}', "");
Expect(1, 69242, '\P{^Nv=00900}', "");
Expect(0, 69243, '\p{Nv=00900}', "");
Expect(1, 69243, '\p{^Nv=00900}', "");
Expect(1, 69243, '\P{Nv=00900}', "");
Expect(0, 69243, '\P{^Nv=00900}', "");
Error('\p{Is_Numeric_Value=	+000000000900/a/}');
Error('\P{Is_Numeric_Value=	+000000000900/a/}');
Expect(1, 69242, '\p{Is_Numeric_Value=00000000900}', "");
Expect(0, 69242, '\p{^Is_Numeric_Value=00000000900}', "");
Expect(0, 69242, '\P{Is_Numeric_Value=00000000900}', "");
Expect(1, 69242, '\P{^Is_Numeric_Value=00000000900}', "");
Expect(0, 69243, '\p{Is_Numeric_Value=00000000900}', "");
Expect(1, 69243, '\p{^Is_Numeric_Value=00000000900}', "");
Expect(1, 69243, '\P{Is_Numeric_Value=00000000900}', "");
Expect(0, 69243, '\P{^Is_Numeric_Value=00000000900}', "");
Error('\p{Is_Nv=_0900:=}');
Error('\P{Is_Nv=_0900:=}');
Expect(1, 69242, '\p{Is_Nv: 0_0_0_9_00}', "");
Expect(0, 69242, '\p{^Is_Nv: 0_0_0_9_00}', "");
Expect(0, 69242, '\P{Is_Nv: 0_0_0_9_00}', "");
Expect(1, 69242, '\P{^Is_Nv: 0_0_0_9_00}', "");
Expect(0, 69243, '\p{Is_Nv: 0_0_0_9_00}', "");
Expect(1, 69243, '\p{^Is_Nv: 0_0_0_9_00}', "");
Expect(1, 69243, '\P{Is_Nv: 0_0_0_9_00}', "");
Expect(0, 69243, '\P{^Is_Nv: 0_0_0_9_00}', "");
Error('\p{Numeric_Value=:=+0000000002/05}');
Error('\P{Numeric_Value=:=+0000000002/05}');
Expect(1, 8534, '\p{Numeric_Value=0002/0000000005}', "");
Expect(0, 8534, '\p{^Numeric_Value=0002/0000000005}', "");
Expect(0, 8534, '\P{Numeric_Value=0002/0000000005}', "");
Expect(1, 8534, '\P{^Numeric_Value=0002/0000000005}', "");
Expect(0, 8535, '\p{Numeric_Value=0002/0000000005}', "");
Expect(1, 8535, '\p{^Numeric_Value=0002/0000000005}', "");
Expect(1, 8535, '\P{Numeric_Value=0002/0000000005}', "");
Expect(0, 8535, '\P{^Numeric_Value=0002/0000000005}', "");
Expect(1, 8534, '\p{Numeric_Value=0.400}', "");
Expect(0, 8534, '\p{^Numeric_Value=0.400}', "");
Expect(0, 8534, '\P{Numeric_Value=0.400}', "");
Expect(1, 8534, '\P{^Numeric_Value=0.400}', "");
Expect(0, 8535, '\p{Numeric_Value=0.400}', "");
Expect(1, 8535, '\p{^Numeric_Value=0.400}', "");
Expect(1, 8535, '\P{Numeric_Value=0.400}', "");
Expect(0, 8535, '\P{^Numeric_Value=0.400}', "");
Error('\p{Nv=_:=002/005}');
Error('\P{Nv=_:=002/005}');
Expect(1, 8534, '\p{Nv=000002/005}', "");
Expect(0, 8534, '\p{^Nv=000002/005}', "");
Expect(0, 8534, '\P{Nv=000002/005}', "");
Expect(1, 8534, '\P{^Nv=000002/005}', "");
Expect(0, 8535, '\p{Nv=000002/005}', "");
Expect(1, 8535, '\p{^Nv=000002/005}', "");
Expect(1, 8535, '\P{Nv=000002/005}', "");
Expect(0, 8535, '\P{^Nv=000002/005}', "");
Expect(1, 8534, '\p{Nv:   0.400}', "");
Expect(0, 8534, '\p{^Nv:   0.400}', "");
Expect(0, 8534, '\P{Nv:   0.400}', "");
Expect(1, 8534, '\P{^Nv:   0.400}', "");
Expect(0, 8535, '\p{Nv:   0.400}', "");
Expect(1, 8535, '\p{^Nv:   0.400}', "");
Expect(1, 8535, '\P{Nv:   0.400}', "");
Expect(0, 8535, '\P{^Nv:   0.400}', "");
Error('\p{Is_Numeric_Value=-000000002/05:=}');
Error('\P{Is_Numeric_Value=-000000002/05:=}');
Expect(1, 8534, '\p{Is_Numeric_Value=00002/00005}', "");
Expect(0, 8534, '\p{^Is_Numeric_Value=00002/00005}', "");
Expect(0, 8534, '\P{Is_Numeric_Value=00002/00005}', "");
Expect(1, 8534, '\P{^Is_Numeric_Value=00002/00005}', "");
Expect(0, 8535, '\p{Is_Numeric_Value=00002/00005}', "");
Expect(1, 8535, '\p{^Is_Numeric_Value=00002/00005}', "");
Expect(1, 8535, '\P{Is_Numeric_Value=00002/00005}', "");
Expect(0, 8535, '\P{^Is_Numeric_Value=00002/00005}', "");
Expect(1, 8534, '\p{Is_Numeric_Value=0.400}', "");
Expect(0, 8534, '\p{^Is_Numeric_Value=0.400}', "");
Expect(0, 8534, '\P{Is_Numeric_Value=0.400}', "");
Expect(1, 8534, '\P{^Is_Numeric_Value=0.400}', "");
Expect(0, 8535, '\p{Is_Numeric_Value=0.400}', "");
Expect(1, 8535, '\p{^Is_Numeric_Value=0.400}', "");
Expect(1, 8535, '\P{Is_Numeric_Value=0.400}', "");
Expect(0, 8535, '\P{^Is_Numeric_Value=0.400}', "");
Error('\p{Is_Nv= /a/+02/0005}');
Error('\P{Is_Nv= /a/+02/0005}');
Expect(1, 8534, '\p{Is_Nv=00002/0000000005}', "");
Expect(0, 8534, '\p{^Is_Nv=00002/0000000005}', "");
Expect(0, 8534, '\P{Is_Nv=00002/0000000005}', "");
Expect(1, 8534, '\P{^Is_Nv=00002/0000000005}', "");
Expect(0, 8535, '\p{Is_Nv=00002/0000000005}', "");
Expect(1, 8535, '\p{^Is_Nv=00002/0000000005}', "");
Expect(1, 8535, '\P{Is_Nv=00002/0000000005}', "");
Expect(0, 8535, '\P{^Is_Nv=00002/0000000005}', "");
Expect(1, 8534, '\p{Is_Nv=0.400}', "");
Expect(0, 8534, '\p{^Is_Nv=0.400}', "");
Expect(0, 8534, '\P{Is_Nv=0.400}', "");
Expect(1, 8534, '\P{^Is_Nv=0.400}', "");
Expect(0, 8535, '\p{Is_Nv=0.400}', "");
Expect(1, 8535, '\p{^Is_Nv=0.400}', "");
Expect(1, 8535, '\P{Is_Nv=0.400}', "");
Expect(0, 8535, '\P{^Is_Nv=0.400}', "");
Error('\p{Numeric_Value=-:=4_6}');
Error('\P{Numeric_Value=-:=4_6}');
Expect(1, 12987, '\p{Numeric_Value=0000000046}', "");
Expect(0, 12987, '\p{^Numeric_Value=0000000046}', "");
Expect(0, 12987, '\P{Numeric_Value=0000000046}', "");
Expect(1, 12987, '\P{^Numeric_Value=0000000046}', "");
Expect(0, 12988, '\p{Numeric_Value=0000000046}', "");
Expect(1, 12988, '\p{^Numeric_Value=0000000046}', "");
Expect(1, 12988, '\P{Numeric_Value=0000000046}', "");
Expect(0, 12988, '\P{^Numeric_Value=0000000046}', "");
Error('\p{Nv= 00_00_00_04_6/a/}');
Error('\P{Nv= 00_00_00_04_6/a/}');
Expect(1, 12987, '\p{Nv=0_0_0_0_0_0_0_046}', "");
Expect(0, 12987, '\p{^Nv=0_0_0_0_0_0_0_046}', "");
Expect(0, 12987, '\P{Nv=0_0_0_0_0_0_0_046}', "");
Expect(1, 12987, '\P{^Nv=0_0_0_0_0_0_0_046}', "");
Expect(0, 12988, '\p{Nv=0_0_0_0_0_0_0_046}', "");
Expect(1, 12988, '\p{^Nv=0_0_0_0_0_0_0_046}', "");
Expect(1, 12988, '\P{Nv=0_0_0_0_0_0_0_046}', "");
Expect(0, 12988, '\P{^Nv=0_0_0_0_0_0_0_046}', "");
Error('\p{Is_Numeric_Value=-	0000046:=}');
Error('\P{Is_Numeric_Value=-	0000046:=}');
Expect(1, 12987, '\p{Is_Numeric_Value=00000046}', "");
Expect(0, 12987, '\p{^Is_Numeric_Value=00000046}', "");
Expect(0, 12987, '\P{Is_Numeric_Value=00000046}', "");
Expect(1, 12987, '\P{^Is_Numeric_Value=00000046}', "");
Expect(0, 12988, '\p{Is_Numeric_Value=00000046}', "");
Expect(1, 12988, '\p{^Is_Numeric_Value=00000046}', "");
Expect(1, 12988, '\P{Is_Numeric_Value=00000046}', "");
Expect(0, 12988, '\P{^Is_Numeric_Value=00000046}', "");
Error('\p{Is_Nv=_-0046:=}');
Error('\P{Is_Nv=_-0046:=}');
Expect(1, 12987, '\p{Is_Nv=04_6}', "");
Expect(0, 12987, '\p{^Is_Nv=04_6}', "");
Expect(0, 12987, '\P{Is_Nv=04_6}', "");
Expect(1, 12987, '\P{^Is_Nv=04_6}', "");
Expect(0, 12988, '\p{Is_Nv=04_6}', "");
Expect(1, 12988, '\p{^Is_Nv=04_6}', "");
Expect(1, 12988, '\P{Is_Nv=04_6}', "");
Expect(0, 12988, '\P{^Is_Nv=04_6}', "");
Error('\p{Numeric_Value=:=	0}');
Error('\P{Numeric_Value=:=	0}');
Expect(1, 127233, '\p{Numeric_Value=+000000}', "");
Expect(0, 127233, '\p{^Numeric_Value=+000000}', "");
Expect(0, 127233, '\P{Numeric_Value=+000000}', "");
Expect(1, 127233, '\P{^Numeric_Value=+000000}', "");
Expect(0, 127234, '\p{Numeric_Value=+000000}', "");
Expect(1, 127234, '\p{^Numeric_Value=+000000}', "");
Expect(1, 127234, '\P{Numeric_Value=+000000}', "");
Expect(0, 127234, '\P{^Numeric_Value=+000000}', "");
Error('\p{Nv=-/a/0000}');
Error('\P{Nv=-/a/0000}');
Expect(1, 127233, '\p{Nv=0000000000}', "");
Expect(0, 127233, '\p{^Nv=0000000000}', "");
Expect(0, 127233, '\P{Nv=0000000000}', "");
Expect(1, 127233, '\P{^Nv=0000000000}', "");
Expect(0, 127234, '\p{Nv=0000000000}', "");
Expect(1, 127234, '\p{^Nv=0000000000}', "");
Expect(1, 127234, '\P{Nv=0000000000}', "");
Expect(0, 127234, '\P{^Nv=0000000000}', "");
Error('\p{Is_Numeric_Value: /a/000000_0}');
Error('\P{Is_Numeric_Value: /a/000000_0}');
Expect(1, 127233, '\p{Is_Numeric_Value=0}', "");
Expect(0, 127233, '\p{^Is_Numeric_Value=0}', "");
Expect(0, 127233, '\P{Is_Numeric_Value=0}', "");
Expect(1, 127233, '\P{^Is_Numeric_Value=0}', "");
Expect(0, 127234, '\p{Is_Numeric_Value=0}', "");
Expect(1, 127234, '\p{^Is_Numeric_Value=0}', "");
Expect(1, 127234, '\P{Is_Numeric_Value=0}', "");
Expect(0, 127234, '\P{^Is_Numeric_Value=0}', "");
Error('\p{Is_Nv=	/a/+0_0_0_000}');
Error('\P{Is_Nv=	/a/+0_0_0_000}');
Expect(1, 127233, '\p{Is_Nv=0_0_0_00}', "");
Expect(0, 127233, '\p{^Is_Nv=0_0_0_00}', "");
Expect(0, 127233, '\P{Is_Nv=0_0_0_00}', "");
Expect(1, 127233, '\P{^Is_Nv=0_0_0_00}', "");
Expect(0, 127234, '\p{Is_Nv=0_0_0_00}', "");
Expect(1, 127234, '\p{^Is_Nv=0_0_0_00}', "");
Expect(1, 127234, '\P{Is_Nv=0_0_0_00}', "");
Expect(0, 127234, '\P{^Is_Nv=0_0_0_00}', "");
Error('\p{Numeric_Value=-/a/000_000_002_3}');
Error('\P{Numeric_Value=-/a/000_000_002_3}');
Expect(1, 12883, '\p{Numeric_Value=000000023}', "");
Expect(0, 12883, '\p{^Numeric_Value=000000023}', "");
Expect(0, 12883, '\P{Numeric_Value=000000023}', "");
Expect(1, 12883, '\P{^Numeric_Value=000000023}', "");
Expect(0, 12884, '\p{Numeric_Value=000000023}', "");
Expect(1, 12884, '\p{^Numeric_Value=000000023}', "");
Expect(1, 12884, '\P{Numeric_Value=000000023}', "");
Expect(0, 12884, '\P{^Numeric_Value=000000023}', "");
Error('\p{Nv=_ 0_0_0_0_0_0_0_0_0_23/a/}');
Error('\P{Nv=_ 0_0_0_0_0_0_0_0_0_23/a/}');
Expect(1, 12883, '\p{Nv: 0023}', "");
Expect(0, 12883, '\p{^Nv: 0023}', "");
Expect(0, 12883, '\P{Nv: 0023}', "");
Expect(1, 12883, '\P{^Nv: 0023}', "");
Expect(0, 12884, '\p{Nv: 0023}', "");
Expect(1, 12884, '\p{^Nv: 0023}', "");
Expect(1, 12884, '\P{Nv: 0023}', "");
Expect(0, 12884, '\P{^Nv: 0023}', "");
Error('\p{Is_Numeric_Value=/a/-2_3}');
Error('\P{Is_Numeric_Value=/a/-2_3}');
Expect(1, 12883, '\p{Is_Numeric_Value:   00000023}', "");
Expect(0, 12883, '\p{^Is_Numeric_Value:   00000023}', "");
Expect(0, 12883, '\P{Is_Numeric_Value:   00000023}', "");
Expect(1, 12883, '\P{^Is_Numeric_Value:   00000023}', "");
Expect(0, 12884, '\p{Is_Numeric_Value:   00000023}', "");
Expect(1, 12884, '\p{^Is_Numeric_Value:   00000023}', "");
Expect(1, 12884, '\P{Is_Numeric_Value:   00000023}', "");
Expect(0, 12884, '\P{^Is_Numeric_Value:   00000023}', "");
Error('\p{Is_Nv=		00000023:=}');
Error('\P{Is_Nv=		00000023:=}');
Expect(1, 12883, '\p{Is_Nv=002_3}', "");
Expect(0, 12883, '\p{^Is_Nv=002_3}', "");
Expect(0, 12883, '\P{Is_Nv=002_3}', "");
Expect(1, 12883, '\P{^Is_Nv=002_3}', "");
Expect(0, 12884, '\p{Is_Nv=002_3}', "");
Expect(1, 12884, '\p{^Is_Nv=002_3}', "");
Expect(1, 12884, '\P{Is_Nv=002_3}', "");
Expect(0, 12884, '\P{^Is_Nv=002_3}', "");
Error('\p{Numeric_Value=-01/00007/a/}');
Error('\P{Numeric_Value=-01/00007/a/}');
Expect(1, 8528, '\p{Numeric_Value=01/0000000007}', "");
Expect(0, 8528, '\p{^Numeric_Value=01/0000000007}', "");
Expect(0, 8528, '\P{Numeric_Value=01/0000000007}', "");
Expect(1, 8528, '\P{^Numeric_Value=01/0000000007}', "");
Expect(0, 8529, '\p{Numeric_Value=01/0000000007}', "");
Expect(1, 8529, '\p{^Numeric_Value=01/0000000007}', "");
Expect(1, 8529, '\P{Numeric_Value=01/0000000007}', "");
Expect(0, 8529, '\P{^Numeric_Value=01/0000000007}', "");
Error('\p{Numeric_Value=0.14}');
Error('\P{Numeric_Value=0.14}');
Expect(1, 8528, '\p{Numeric_Value=0.143}', "");
Expect(0, 8528, '\p{^Numeric_Value=0.143}', "");
Expect(0, 8528, '\P{Numeric_Value=0.143}', "");
Expect(1, 8528, '\P{^Numeric_Value=0.143}', "");
Expect(0, 8529, '\p{Numeric_Value=0.143}', "");
Expect(1, 8529, '\p{^Numeric_Value=0.143}', "");
Expect(1, 8529, '\P{Numeric_Value=0.143}', "");
Expect(0, 8529, '\P{^Numeric_Value=0.143}', "");
Error('\p{Nv=__00001/007/a/}');
Error('\P{Nv=__00001/007/a/}');
Expect(1, 8528, '\p{Nv=000000001/007}', "");
Expect(0, 8528, '\p{^Nv=000000001/007}', "");
Expect(0, 8528, '\P{Nv=000000001/007}', "");
Expect(1, 8528, '\P{^Nv=000000001/007}', "");
Expect(0, 8529, '\p{Nv=000000001/007}', "");
Expect(1, 8529, '\p{^Nv=000000001/007}', "");
Expect(1, 8529, '\P{Nv=000000001/007}', "");
Expect(0, 8529, '\P{^Nv=000000001/007}', "");
Error('\p{Nv=0.14}');
Error('\P{Nv=0.14}');
Expect(1, 8528, '\p{Nv=0.143}', "");
Expect(0, 8528, '\p{^Nv=0.143}', "");
Expect(0, 8528, '\P{Nv=0.143}', "");
Expect(1, 8528, '\P{^Nv=0.143}', "");
Expect(0, 8529, '\p{Nv=0.143}', "");
Expect(1, 8529, '\p{^Nv=0.143}', "");
Expect(1, 8529, '\P{Nv=0.143}', "");
Expect(0, 8529, '\P{^Nv=0.143}', "");
Error('\p{Is_Numeric_Value=/a/_	00001/00007}');
Error('\P{Is_Numeric_Value=/a/_	00001/00007}');
Expect(1, 8528, '\p{Is_Numeric_Value=001/07}', "");
Expect(0, 8528, '\p{^Is_Numeric_Value=001/07}', "");
Expect(0, 8528, '\P{Is_Numeric_Value=001/07}', "");
Expect(1, 8528, '\P{^Is_Numeric_Value=001/07}', "");
Expect(0, 8529, '\p{Is_Numeric_Value=001/07}', "");
Expect(1, 8529, '\p{^Is_Numeric_Value=001/07}', "");
Expect(1, 8529, '\P{Is_Numeric_Value=001/07}', "");
Expect(0, 8529, '\P{^Is_Numeric_Value=001/07}', "");
Error('\p{Is_Numeric_Value=0.14}');
Error('\P{Is_Numeric_Value=0.14}');
Expect(1, 8528, '\p{Is_Numeric_Value=0.143}', "");
Expect(0, 8528, '\p{^Is_Numeric_Value=0.143}', "");
Expect(0, 8528, '\P{Is_Numeric_Value=0.143}', "");
Expect(1, 8528, '\P{^Is_Numeric_Value=0.143}', "");
Expect(0, 8529, '\p{Is_Numeric_Value=0.143}', "");
Expect(1, 8529, '\p{^Is_Numeric_Value=0.143}', "");
Expect(1, 8529, '\P{Is_Numeric_Value=0.143}', "");
Expect(0, 8529, '\P{^Is_Numeric_Value=0.143}', "");
Error('\p{Is_Nv:   :=-_0000001/7}');
Error('\P{Is_Nv:   :=-_0000001/7}');
Expect(1, 8528, '\p{Is_Nv=1/000007}', "");
Expect(0, 8528, '\p{^Is_Nv=1/000007}', "");
Expect(0, 8528, '\P{Is_Nv=1/000007}', "");
Expect(1, 8528, '\P{^Is_Nv=1/000007}', "");
Expect(0, 8529, '\p{Is_Nv=1/000007}', "");
Expect(1, 8529, '\p{^Is_Nv=1/000007}', "");
Expect(1, 8529, '\P{Is_Nv=1/000007}', "");
Expect(0, 8529, '\P{^Is_Nv=1/000007}', "");
Error('\p{Is_Nv=0.14}');
Error('\P{Is_Nv=0.14}');
Expect(1, 8528, '\p{Is_Nv=0.143}', "");
Expect(0, 8528, '\p{^Is_Nv=0.143}', "");
Expect(0, 8528, '\P{Is_Nv=0.143}', "");
Expect(1, 8528, '\P{^Is_Nv=0.143}', "");
Expect(0, 8529, '\p{Is_Nv=0.143}', "");
Expect(1, 8529, '\p{^Is_Nv=0.143}', "");
Expect(1, 8529, '\P{Is_Nv=0.143}', "");
Expect(0, 8529, '\P{^Is_Nv=0.143}', "");
Error('\p{Numeric_Value:   	+000000004/005/a/}');
Error('\P{Numeric_Value:   	+000000004/005/a/}');
Expect(1, 8536, '\p{Numeric_Value=004/0000005}', "");
Expect(0, 8536, '\p{^Numeric_Value=004/0000005}', "");
Expect(0, 8536, '\P{Numeric_Value=004/0000005}', "");
Expect(1, 8536, '\P{^Numeric_Value=004/0000005}', "");
Expect(0, 8537, '\p{Numeric_Value=004/0000005}', "");
Expect(1, 8537, '\p{^Numeric_Value=004/0000005}', "");
Expect(1, 8537, '\P{Numeric_Value=004/0000005}', "");
Expect(0, 8537, '\P{^Numeric_Value=004/0000005}', "");
Expect(1, 8536, '\p{Numeric_Value=0.800}', "");
Expect(0, 8536, '\p{^Numeric_Value=0.800}', "");
Expect(0, 8536, '\P{Numeric_Value=0.800}', "");
Expect(1, 8536, '\P{^Numeric_Value=0.800}', "");
Expect(0, 8537, '\p{Numeric_Value=0.800}', "");
Expect(1, 8537, '\p{^Numeric_Value=0.800}', "");
Expect(1, 8537, '\P{Numeric_Value=0.800}', "");
Expect(0, 8537, '\P{^Numeric_Value=0.800}', "");
Error('\p{Nv=:=_000004/5}');
Error('\P{Nv=:=_000004/5}');
Expect(1, 8536, '\p{Nv=0004/00005}', "");
Expect(0, 8536, '\p{^Nv=0004/00005}', "");
Expect(0, 8536, '\P{Nv=0004/00005}', "");
Expect(1, 8536, '\P{^Nv=0004/00005}', "");
Expect(0, 8537, '\p{Nv=0004/00005}', "");
Expect(1, 8537, '\p{^Nv=0004/00005}', "");
Expect(1, 8537, '\P{Nv=0004/00005}', "");
Expect(0, 8537, '\P{^Nv=0004/00005}', "");
Expect(1, 8536, '\p{Nv=0.800}', "");
Expect(0, 8536, '\p{^Nv=0.800}', "");
Expect(0, 8536, '\P{Nv=0.800}', "");
Expect(1, 8536, '\P{^Nv=0.800}', "");
Expect(0, 8537, '\p{Nv=0.800}', "");
Expect(1, 8537, '\p{^Nv=0.800}', "");
Expect(1, 8537, '\P{Nv=0.800}', "");
Expect(0, 8537, '\P{^Nv=0.800}', "");
Error('\p{Is_Numeric_Value: /a/_ +004/0005}');
Error('\P{Is_Numeric_Value: /a/_ +004/0005}');
Expect(1, 8536, '\p{Is_Numeric_Value=004/005}', "");
Expect(0, 8536, '\p{^Is_Numeric_Value=004/005}', "");
Expect(0, 8536, '\P{Is_Numeric_Value=004/005}', "");
Expect(1, 8536, '\P{^Is_Numeric_Value=004/005}', "");
Expect(0, 8537, '\p{Is_Numeric_Value=004/005}', "");
Expect(1, 8537, '\p{^Is_Numeric_Value=004/005}', "");
Expect(1, 8537, '\P{Is_Numeric_Value=004/005}', "");
Expect(0, 8537, '\P{^Is_Numeric_Value=004/005}', "");
Expect(1, 8536, '\p{Is_Numeric_Value=0.800}', "");
Expect(0, 8536, '\p{^Is_Numeric_Value=0.800}', "");
Expect(0, 8536, '\P{Is_Numeric_Value=0.800}', "");
Expect(1, 8536, '\P{^Is_Numeric_Value=0.800}', "");
Expect(0, 8537, '\p{Is_Numeric_Value=0.800}', "");
Expect(1, 8537, '\p{^Is_Numeric_Value=0.800}', "");
Expect(1, 8537, '\P{Is_Numeric_Value=0.800}', "");
Expect(0, 8537, '\P{^Is_Numeric_Value=0.800}', "");
Error('\p{Is_Nv=:=0004/00005}');
Error('\P{Is_Nv=:=0004/00005}');
Expect(1, 8536, '\p{Is_Nv=+4/00005}', "");
Expect(0, 8536, '\p{^Is_Nv=+4/00005}', "");
Expect(0, 8536, '\P{Is_Nv=+4/00005}', "");
Expect(1, 8536, '\P{^Is_Nv=+4/00005}', "");
Expect(0, 8537, '\p{Is_Nv=+4/00005}', "");
Expect(1, 8537, '\p{^Is_Nv=+4/00005}', "");
Expect(1, 8537, '\P{Is_Nv=+4/00005}', "");
Expect(0, 8537, '\P{^Is_Nv=+4/00005}', "");
Expect(1, 8536, '\p{Is_Nv=0.800}', "");
Expect(0, 8536, '\p{^Is_Nv=0.800}', "");
Expect(0, 8536, '\P{Is_Nv=0.800}', "");
Expect(1, 8536, '\P{^Is_Nv=0.800}', "");
Expect(0, 8537, '\p{Is_Nv=0.800}', "");
Expect(1, 8537, '\p{^Is_Nv=0.800}', "");
Expect(1, 8537, '\P{Is_Nv=0.800}', "");
Expect(0, 8537, '\P{^Is_Nv=0.800}', "");
Error('\p{Numeric_Value:	 /a/0_0_0_0_0_0_0_0013}');
Error('\P{Numeric_Value:	 /a/0_0_0_0_0_0_0_0013}');
Expect(1, 9453, '\p{Numeric_Value=0_0_13}', "");
Expect(0, 9453, '\p{^Numeric_Value=0_0_13}', "");
Expect(0, 9453, '\P{Numeric_Value=0_0_13}', "");
Expect(1, 9453, '\P{^Numeric_Value=0_0_13}', "");
Expect(0, 9454, '\p{Numeric_Value=0_0_13}', "");
Expect(1, 9454, '\p{^Numeric_Value=0_0_13}', "");
Expect(1, 9454, '\P{Numeric_Value=0_0_13}', "");
Expect(0, 9454, '\P{^Numeric_Value=0_0_13}', "");
Error('\p{Nv=/a/-_+000013}');
Error('\P{Nv=/a/-_+000013}');
Expect(1, 9453, '\p{Nv=+000000013}', "");
Expect(0, 9453, '\p{^Nv=+000000013}', "");
Expect(0, 9453, '\P{Nv=+000000013}', "");
Expect(1, 9453, '\P{^Nv=+000000013}', "");
Expect(0, 9454, '\p{Nv=+000000013}', "");
Expect(1, 9454, '\p{^Nv=+000000013}', "");
Expect(1, 9454, '\P{Nv=+000000013}', "");
Expect(0, 9454, '\P{^Nv=+000000013}', "");
Error('\p{Is_Numeric_Value= +00_01_3:=}');
Error('\P{Is_Numeric_Value= +00_01_3:=}');
Expect(1, 9453, '\p{Is_Numeric_Value=+0000013}', "");
Expect(0, 9453, '\p{^Is_Numeric_Value=+0000013}', "");
Expect(0, 9453, '\P{Is_Numeric_Value=+0000013}', "");
Expect(1, 9453, '\P{^Is_Numeric_Value=+0000013}', "");
Expect(0, 9454, '\p{Is_Numeric_Value=+0000013}', "");
Expect(1, 9454, '\p{^Is_Numeric_Value=+0000013}', "");
Expect(1, 9454, '\P{Is_Numeric_Value=+0000013}', "");
Expect(0, 9454, '\P{^Is_Numeric_Value=+0000013}', "");
Error('\p{Is_Nv:   _0_0_0_0_0_013:=}');
Error('\P{Is_Nv:   _0_0_0_0_0_013:=}');
Expect(1, 9453, '\p{Is_Nv=+0_0_0_0_0_0_0_0_013}', "");
Expect(0, 9453, '\p{^Is_Nv=+0_0_0_0_0_0_0_0_013}', "");
Expect(0, 9453, '\P{Is_Nv=+0_0_0_0_0_0_0_0_013}', "");
Expect(1, 9453, '\P{^Is_Nv=+0_0_0_0_0_0_0_0_013}', "");
Expect(0, 9454, '\p{Is_Nv=+0_0_0_0_0_0_0_0_013}', "");
Expect(1, 9454, '\p{^Is_Nv=+0_0_0_0_0_0_0_0_013}', "");
Expect(1, 9454, '\P{Is_Nv=+0_0_0_0_0_0_0_0_013}', "");
Expect(0, 9454, '\P{^Is_Nv=+0_0_0_0_0_0_0_0_013}', "");
Error('\p{Numeric_Value=	00006/a/}');
Error('\P{Numeric_Value=	00006/a/}');
Expect(1, 133866, '\p{Numeric_Value=000_000_6}', "");
Expect(0, 133866, '\p{^Numeric_Value=000_000_6}', "");
Expect(0, 133866, '\P{Numeric_Value=000_000_6}', "");
Expect(1, 133866, '\P{^Numeric_Value=000_000_6}', "");
Expect(0, 133867, '\p{Numeric_Value=000_000_6}', "");
Expect(1, 133867, '\p{^Numeric_Value=000_000_6}', "");
Expect(1, 133867, '\P{Numeric_Value=000_000_6}', "");
Expect(0, 133867, '\P{^Numeric_Value=000_000_6}', "");
Error('\p{Nv=_:=006}');
Error('\P{Nv=_:=006}');
Expect(1, 133866, '\p{Nv:   +00000006}', "");
Expect(0, 133866, '\p{^Nv:   +00000006}', "");
Expect(0, 133866, '\P{Nv:   +00000006}', "");
Expect(1, 133866, '\P{^Nv:   +00000006}', "");
Expect(0, 133867, '\p{Nv:   +00000006}', "");
Expect(1, 133867, '\p{^Nv:   +00000006}', "");
Expect(1, 133867, '\P{Nv:   +00000006}', "");
Expect(0, 133867, '\P{^Nv:   +00000006}', "");
Error('\p{Is_Numeric_Value=:=__0_0_06}');
Error('\P{Is_Numeric_Value=:=__0_0_06}');
Expect(1, 133866, '\p{Is_Numeric_Value:   0_0_0_0_0_06}', "");
Expect(0, 133866, '\p{^Is_Numeric_Value:   0_0_0_0_0_06}', "");
Expect(0, 133866, '\P{Is_Numeric_Value:   0_0_0_0_0_06}', "");
Expect(1, 133866, '\P{^Is_Numeric_Value:   0_0_0_0_0_06}', "");
Expect(0, 133867, '\p{Is_Numeric_Value:   0_0_0_0_0_06}', "");
Expect(1, 133867, '\p{^Is_Numeric_Value:   0_0_0_0_0_06}', "");
Expect(1, 133867, '\P{Is_Numeric_Value:   0_0_0_0_0_06}', "");
Expect(0, 133867, '\P{^Is_Numeric_Value:   0_0_0_0_0_06}', "");
Error('\p{Is_Nv=/a/	-0_6}');
Error('\P{Is_Nv=/a/	-0_6}');
Expect(1, 133866, '\p{Is_Nv=0000_6}', "");
Expect(0, 133866, '\p{^Is_Nv=0000_6}', "");
Expect(0, 133866, '\P{Is_Nv=0000_6}', "");
Expect(1, 133866, '\P{^Is_Nv=0000_6}', "");
Expect(0, 133867, '\p{Is_Nv=0000_6}', "");
Expect(1, 133867, '\p{^Is_Nv=0000_6}', "");
Expect(1, 133867, '\P{Is_Nv=0000_6}', "");
Expect(0, 133867, '\P{^Is_Nv=0000_6}', "");
Error('\p{Numeric_Value=_	7000_0/a/}');
Error('\P{Numeric_Value=_	7000_0/a/}');
Expect(1, 65841, '\p{Numeric_Value=070000}', "");
Expect(0, 65841, '\p{^Numeric_Value=070000}', "");
Expect(0, 65841, '\P{Numeric_Value=070000}', "");
Expect(1, 65841, '\P{^Numeric_Value=070000}', "");
Expect(0, 65842, '\p{Numeric_Value=070000}', "");
Expect(1, 65842, '\p{^Numeric_Value=070000}', "");
Expect(1, 65842, '\P{Numeric_Value=070000}', "");
Expect(0, 65842, '\P{^Numeric_Value=070000}', "");
Error('\p{Nv=-	0_7_0_0_00/a/}');
Error('\P{Nv=-	0_7_0_0_00/a/}');
Expect(1, 65841, '\p{Nv:   0000000070000}', "");
Expect(0, 65841, '\p{^Nv:   0000000070000}', "");
Expect(0, 65841, '\P{Nv:   0000000070000}', "");
Expect(1, 65841, '\P{^Nv:   0000000070000}', "");
Expect(0, 65842, '\p{Nv:   0000000070000}', "");
Expect(1, 65842, '\p{^Nv:   0000000070000}', "");
Expect(1, 65842, '\P{Nv:   0000000070000}', "");
Expect(0, 65842, '\P{^Nv:   0000000070000}', "");
Error('\p{Is_Numeric_Value=	/a/70000}');
Error('\P{Is_Numeric_Value=	/a/70000}');
Expect(1, 65841, '\p{Is_Numeric_Value:7_0_0_00}', "");
Expect(0, 65841, '\p{^Is_Numeric_Value:7_0_0_00}', "");
Expect(0, 65841, '\P{Is_Numeric_Value:7_0_0_00}', "");
Expect(1, 65841, '\P{^Is_Numeric_Value:7_0_0_00}', "");
Expect(0, 65842, '\p{Is_Numeric_Value:7_0_0_00}', "");
Expect(1, 65842, '\p{^Is_Numeric_Value:7_0_0_00}', "");
Expect(1, 65842, '\P{Is_Numeric_Value:7_0_0_00}', "");
Expect(0, 65842, '\P{^Is_Numeric_Value:7_0_0_00}', "");
Error('\p{Is_Nv=/a/- 7000_0}');
Error('\P{Is_Nv=/a/- 7000_0}');
Expect(1, 65841, '\p{Is_Nv=+070000}', "");
Expect(0, 65841, '\p{^Is_Nv=+070000}', "");
Expect(0, 65841, '\P{Is_Nv=+070000}', "");
Expect(1, 65841, '\P{^Is_Nv=+070000}', "");
Expect(0, 65842, '\p{Is_Nv=+070000}', "");
Expect(1, 65842, '\p{^Is_Nv=+070000}', "");
Expect(1, 65842, '\P{Is_Nv=+070000}', "");
Expect(0, 65842, '\P{^Is_Nv=+070000}', "");
Error('\p{Numeric_Value=:=-+0036}');
Error('\P{Numeric_Value=:=-+0036}');
Expect(1, 12977, '\p{Numeric_Value=0000_0000_36}', "");
Expect(0, 12977, '\p{^Numeric_Value=0000_0000_36}', "");
Expect(0, 12977, '\P{Numeric_Value=0000_0000_36}', "");
Expect(1, 12977, '\P{^Numeric_Value=0000_0000_36}', "");
Expect(0, 12978, '\p{Numeric_Value=0000_0000_36}', "");
Expect(1, 12978, '\p{^Numeric_Value=0000_0000_36}', "");
Expect(1, 12978, '\P{Numeric_Value=0000_0000_36}', "");
Expect(0, 12978, '\P{^Numeric_Value=0000_0000_36}', "");
Error('\p{Nv: -000000036:=}');
Error('\P{Nv: -000000036:=}');
Expect(1, 12977, '\p{Nv=0_0_0_0_0_00036}', "");
Expect(0, 12977, '\p{^Nv=0_0_0_0_0_00036}', "");
Expect(0, 12977, '\P{Nv=0_0_0_0_0_00036}', "");
Expect(1, 12977, '\P{^Nv=0_0_0_0_0_00036}', "");
Expect(0, 12978, '\p{Nv=0_0_0_0_0_00036}', "");
Expect(1, 12978, '\p{^Nv=0_0_0_0_0_00036}', "");
Expect(1, 12978, '\P{Nv=0_0_0_0_0_00036}', "");
Expect(0, 12978, '\P{^Nv=0_0_0_0_0_00036}', "");
Error('\p{Is_Numeric_Value=:=_ 03_6}');
Error('\P{Is_Numeric_Value=:=_ 03_6}');
Expect(1, 12977, '\p{Is_Numeric_Value=+03_6}', "");
Expect(0, 12977, '\p{^Is_Numeric_Value=+03_6}', "");
Expect(0, 12977, '\P{Is_Numeric_Value=+03_6}', "");
Expect(1, 12977, '\P{^Is_Numeric_Value=+03_6}', "");
Expect(0, 12978, '\p{Is_Numeric_Value=+03_6}', "");
Expect(1, 12978, '\p{^Is_Numeric_Value=+03_6}', "");
Expect(1, 12978, '\P{Is_Numeric_Value=+03_6}', "");
Expect(0, 12978, '\P{^Is_Numeric_Value=+03_6}', "");
Error('\p{Is_Nv=_000003_6:=}');
Error('\P{Is_Nv=_000003_6:=}');
Expect(1, 12977, '\p{Is_Nv=0_0_0_0_0_0036}', "");
Expect(0, 12977, '\p{^Is_Nv=0_0_0_0_0_0036}', "");
Expect(0, 12977, '\P{Is_Nv=0_0_0_0_0_0036}', "");
Expect(1, 12977, '\P{^Is_Nv=0_0_0_0_0_0036}', "");
Expect(0, 12978, '\p{Is_Nv=0_0_0_0_0_0036}', "");
Expect(1, 12978, '\p{^Is_Nv=0_0_0_0_0_0036}', "");
Expect(1, 12978, '\P{Is_Nv=0_0_0_0_0_0036}', "");
Expect(0, 12978, '\P{^Is_Nv=0_0_0_0_0_0036}', "");
Error('\p{Numeric_Value=-/a/+0000000003}');
Error('\P{Numeric_Value=-/a/+0000000003}');
Expect(1, 146203, '\p{Numeric_Value=+00_00_3}', "");
Expect(0, 146203, '\p{^Numeric_Value=+00_00_3}', "");
Expect(0, 146203, '\P{Numeric_Value=+00_00_3}', "");
Expect(1, 146203, '\P{^Numeric_Value=+00_00_3}', "");
Expect(0, 146204, '\p{Numeric_Value=+00_00_3}', "");
Expect(1, 146204, '\p{^Numeric_Value=+00_00_3}', "");
Expect(1, 146204, '\P{Numeric_Value=+00_00_3}', "");
Expect(0, 146204, '\P{^Numeric_Value=+00_00_3}', "");
Error('\p{Nv:	-00_3/a/}');
Error('\P{Nv:	-00_3/a/}');
Expect(1, 146203, '\p{Nv=+000000003}', "");
Expect(0, 146203, '\p{^Nv=+000000003}', "");
Expect(0, 146203, '\P{Nv=+000000003}', "");
Expect(1, 146203, '\P{^Nv=+000000003}', "");
Expect(0, 146204, '\p{Nv=+000000003}', "");
Expect(1, 146204, '\p{^Nv=+000000003}', "");
Expect(1, 146204, '\P{Nv=+000000003}', "");
Expect(0, 146204, '\P{^Nv=+000000003}', "");
Error('\p{Is_Numeric_Value=_:=+00000003}');
Error('\P{Is_Numeric_Value=_:=+00000003}');
Expect(1, 146203, '\p{Is_Numeric_Value=00_3}', "");
Expect(0, 146203, '\p{^Is_Numeric_Value=00_3}', "");
Expect(0, 146203, '\P{Is_Numeric_Value=00_3}', "");
Expect(1, 146203, '\P{^Is_Numeric_Value=00_3}', "");
Expect(0, 146204, '\p{Is_Numeric_Value=00_3}', "");
Expect(1, 146204, '\p{^Is_Numeric_Value=00_3}', "");
Expect(1, 146204, '\P{Is_Numeric_Value=00_3}', "");
Expect(0, 146204, '\P{^Is_Numeric_Value=00_3}', "");
Error('\p{Is_Nv=_ 00_00_03:=}');
Error('\P{Is_Nv=_ 00_00_03:=}');
Expect(1, 146203, '\p{Is_Nv=00_00_03}', "");
Expect(0, 146203, '\p{^Is_Nv=00_00_03}', "");
Expect(0, 146203, '\P{Is_Nv=00_00_03}', "");
Expect(1, 146203, '\P{^Is_Nv=00_00_03}', "");
Expect(0, 146204, '\p{Is_Nv=00_00_03}', "");
Expect(1, 146204, '\p{^Is_Nv=00_00_03}', "");
Expect(1, 146204, '\P{Is_Nv=00_00_03}', "");
Expect(0, 146204, '\P{^Is_Nv=00_00_03}', "");
Error('\p{Numeric_Value=+000000005000:=}');
Error('\P{Numeric_Value=+000000005000:=}');
Expect(1, 65906, '\p{Numeric_Value=500_0}', "");
Expect(0, 65906, '\p{^Numeric_Value=500_0}', "");
Expect(0, 65906, '\P{Numeric_Value=500_0}', "");
Expect(1, 65906, '\P{^Numeric_Value=500_0}', "");
Expect(0, 65907, '\p{Numeric_Value=500_0}', "");
Expect(1, 65907, '\p{^Numeric_Value=500_0}', "");
Expect(1, 65907, '\P{Numeric_Value=500_0}', "");
Expect(0, 65907, '\P{^Numeric_Value=500_0}', "");
Error('\p{Nv=_ +0500_0:=}');
Error('\P{Nv=_ +0500_0:=}');
Expect(1, 65906, '\p{Nv:0000500_0}', "");
Expect(0, 65906, '\p{^Nv:0000500_0}', "");
Expect(0, 65906, '\P{Nv:0000500_0}', "");
Expect(1, 65906, '\P{^Nv:0000500_0}', "");
Expect(0, 65907, '\p{Nv:0000500_0}', "");
Expect(1, 65907, '\p{^Nv:0000500_0}', "");
Expect(1, 65907, '\P{Nv:0000500_0}', "");
Expect(0, 65907, '\P{^Nv:0000500_0}', "");
Error('\p{Is_Numeric_Value=:=		05_00_0}');
Error('\P{Is_Numeric_Value=:=		05_00_0}');
Expect(1, 65906, '\p{Is_Numeric_Value=005000}', "");
Expect(0, 65906, '\p{^Is_Numeric_Value=005000}', "");
Expect(0, 65906, '\P{Is_Numeric_Value=005000}', "");
Expect(1, 65906, '\P{^Is_Numeric_Value=005000}', "");
Expect(0, 65907, '\p{Is_Numeric_Value=005000}', "");
Expect(1, 65907, '\p{^Is_Numeric_Value=005000}', "");
Expect(1, 65907, '\P{Is_Numeric_Value=005000}', "");
Expect(0, 65907, '\P{^Is_Numeric_Value=005000}', "");
Error('\p{Is_Nv=/a/	0000000005000}');
Error('\P{Is_Nv=/a/	0000000005000}');
Expect(1, 65906, '\p{Is_Nv=+00000005000}', "");
Expect(0, 65906, '\p{^Is_Nv=+00000005000}', "");
Expect(0, 65906, '\P{Is_Nv=+00000005000}', "");
Expect(1, 65906, '\P{^Is_Nv=+00000005000}', "");
Expect(0, 65907, '\p{Is_Nv=+00000005000}', "");
Expect(1, 65907, '\p{^Is_Nv=+00000005000}', "");
Expect(1, 65907, '\P{Is_Nv=+00000005000}', "");
Expect(0, 65907, '\P{^Is_Nv=+00000005000}', "");
Error('\p{Numeric_Value=:=00000009}');
Error('\P{Numeric_Value=:=00000009}');
Expect(1, 194704, '\p{Numeric_Value: 9}', "");
Expect(0, 194704, '\p{^Numeric_Value: 9}', "");
Expect(0, 194704, '\P{Numeric_Value: 9}', "");
Expect(1, 194704, '\P{^Numeric_Value: 9}', "");
Expect(0, 194705, '\p{Numeric_Value: 9}', "");
Expect(1, 194705, '\p{^Numeric_Value: 9}', "");
Expect(1, 194705, '\P{Numeric_Value: 9}', "");
Expect(0, 194705, '\P{^Numeric_Value: 9}', "");
Error('\p{Nv=	/a/009}');
Error('\P{Nv=	/a/009}');
Expect(1, 194704, '\p{Nv:009}', "");
Expect(0, 194704, '\p{^Nv:009}', "");
Expect(0, 194704, '\P{Nv:009}', "");
Expect(1, 194704, '\P{^Nv:009}', "");
Expect(0, 194705, '\p{Nv:009}', "");
Expect(1, 194705, '\p{^Nv:009}', "");
Expect(1, 194705, '\P{Nv:009}', "");
Expect(0, 194705, '\P{^Nv:009}', "");
Error('\p{Is_Numeric_Value=-00009/a/}');
Error('\P{Is_Numeric_Value=-00009/a/}');
Expect(1, 194704, '\p{Is_Numeric_Value=0_9}', "");
Expect(0, 194704, '\p{^Is_Numeric_Value=0_9}', "");
Expect(0, 194704, '\P{Is_Numeric_Value=0_9}', "");
Expect(1, 194704, '\P{^Is_Numeric_Value=0_9}', "");
Expect(0, 194705, '\p{Is_Numeric_Value=0_9}', "");
Expect(1, 194705, '\p{^Is_Numeric_Value=0_9}', "");
Expect(1, 194705, '\P{Is_Numeric_Value=0_9}', "");
Expect(0, 194705, '\P{^Is_Numeric_Value=0_9}', "");
Error('\p{Is_Nv=/a/ -0_0_0_0_0_0_0_09}');
Error('\P{Is_Nv=/a/ -0_0_0_0_0_0_0_09}');
Expect(1, 194704, '\p{Is_Nv=000_000_9}', "");
Expect(0, 194704, '\p{^Is_Nv=000_000_9}', "");
Expect(0, 194704, '\P{Is_Nv=000_000_9}', "");
Expect(1, 194704, '\P{^Is_Nv=000_000_9}', "");
Expect(0, 194705, '\p{Is_Nv=000_000_9}', "");
Expect(1, 194705, '\p{^Is_Nv=000_000_9}', "");
Expect(1, 194705, '\P{Is_Nv=000_000_9}', "");
Expect(0, 194705, '\P{^Is_Nv=000_000_9}', "");
Error('\p{Numeric_Value=:=- 01000_0}');
Error('\P{Numeric_Value=:=- 01000_0}');
Expect(1, 67679, '\p{Numeric_Value=0_0_0_0_0_0_10000}', "");
Expect(0, 67679, '\p{^Numeric_Value=0_0_0_0_0_0_10000}', "");
Expect(0, 67679, '\P{Numeric_Value=0_0_0_0_0_0_10000}', "");
Expect(1, 67679, '\P{^Numeric_Value=0_0_0_0_0_0_10000}', "");
Expect(0, 67680, '\p{Numeric_Value=0_0_0_0_0_0_10000}', "");
Expect(1, 67680, '\p{^Numeric_Value=0_0_0_0_0_0_10000}', "");
Expect(1, 67680, '\P{Numeric_Value=0_0_0_0_0_0_10000}', "");
Expect(0, 67680, '\P{^Numeric_Value=0_0_0_0_0_0_10000}', "");
Error('\p{Nv=-/a/00000000010000}');
Error('\P{Nv=-/a/00000000010000}');
Expect(1, 67679, '\p{Nv=010000}', "");
Expect(0, 67679, '\p{^Nv=010000}', "");
Expect(0, 67679, '\P{Nv=010000}', "");
Expect(1, 67679, '\P{^Nv=010000}', "");
Expect(0, 67680, '\p{Nv=010000}', "");
Expect(1, 67680, '\p{^Nv=010000}', "");
Expect(1, 67680, '\P{Nv=010000}', "");
Expect(0, 67680, '\P{^Nv=010000}', "");
Error('\p{Is_Numeric_Value=_:=01000_0}');
Error('\P{Is_Numeric_Value=_:=01000_0}');
Expect(1, 67679, '\p{Is_Numeric_Value=0_0_0_0_1_0_000}', "");
Expect(0, 67679, '\p{^Is_Numeric_Value=0_0_0_0_1_0_000}', "");
Expect(0, 67679, '\P{Is_Numeric_Value=0_0_0_0_1_0_000}', "");
Expect(1, 67679, '\P{^Is_Numeric_Value=0_0_0_0_1_0_000}', "");
Expect(0, 67680, '\p{Is_Numeric_Value=0_0_0_0_1_0_000}', "");
Expect(1, 67680, '\p{^Is_Numeric_Value=0_0_0_0_1_0_000}', "");
Expect(1, 67680, '\P{Is_Numeric_Value=0_0_0_0_1_0_000}', "");
Expect(0, 67680, '\P{^Is_Numeric_Value=0_0_0_0_1_0_000}', "");
Error('\p{Is_Nv=:= 00000000010000}');
Error('\P{Is_Nv=:= 00000000010000}');
Expect(1, 67679, '\p{Is_Nv=+000010000}', "");
Expect(0, 67679, '\p{^Is_Nv=+000010000}', "");
Expect(0, 67679, '\P{Is_Nv=+000010000}', "");
Expect(1, 67679, '\P{^Is_Nv=+000010000}', "");
Expect(0, 67680, '\p{Is_Nv=+000010000}', "");
Expect(1, 67680, '\p{^Is_Nv=+000010000}', "");
Expect(1, 67680, '\P{Is_Nv=+000010000}', "");
Expect(0, 67680, '\P{^Is_Nv=+000010000}', "");
Error('\p{Numeric_Value=	:=0047}');
Error('\P{Numeric_Value=	:=0047}');
Expect(1, 12988, '\p{Numeric_Value=+000047}', "");
Expect(0, 12988, '\p{^Numeric_Value=+000047}', "");
Expect(0, 12988, '\P{Numeric_Value=+000047}', "");
Expect(1, 12988, '\P{^Numeric_Value=+000047}', "");
Expect(0, 12989, '\p{Numeric_Value=+000047}', "");
Expect(1, 12989, '\p{^Numeric_Value=+000047}', "");
Expect(1, 12989, '\P{Numeric_Value=+000047}', "");
Expect(0, 12989, '\P{^Numeric_Value=+000047}', "");
Error('\p{Nv=/a/	047}');
Error('\P{Nv=/a/	047}');
Expect(1, 12988, '\p{Nv=0000000047}', "");
Expect(0, 12988, '\p{^Nv=0000000047}', "");
Expect(0, 12988, '\P{Nv=0000000047}', "");
Expect(1, 12988, '\P{^Nv=0000000047}', "");
Expect(0, 12989, '\p{Nv=0000000047}', "");
Expect(1, 12989, '\p{^Nv=0000000047}', "");
Expect(1, 12989, '\P{Nv=0000000047}', "");
Expect(0, 12989, '\P{^Nv=0000000047}', "");
Error('\p{Is_Numeric_Value=	-4_7:=}');
Error('\P{Is_Numeric_Value=	-4_7:=}');
Expect(1, 12988, '\p{Is_Numeric_Value=0_0_0_0_0_0047}', "");
Expect(0, 12988, '\p{^Is_Numeric_Value=0_0_0_0_0_0047}', "");
Expect(0, 12988, '\P{Is_Numeric_Value=0_0_0_0_0_0047}', "");
Expect(1, 12988, '\P{^Is_Numeric_Value=0_0_0_0_0_0047}', "");
Expect(0, 12989, '\p{Is_Numeric_Value=0_0_0_0_0_0047}', "");
Expect(1, 12989, '\p{^Is_Numeric_Value=0_0_0_0_0_0047}', "");
Expect(1, 12989, '\P{Is_Numeric_Value=0_0_0_0_0_0047}', "");
Expect(0, 12989, '\P{^Is_Numeric_Value=0_0_0_0_0_0047}', "");
Error('\p{Is_Nv= -+000047/a/}');
Error('\P{Is_Nv= -+000047/a/}');
Expect(1, 12988, '\p{Is_Nv=04_7}', "");
Expect(0, 12988, '\p{^Is_Nv=04_7}', "");
Expect(0, 12988, '\P{Is_Nv=04_7}', "");
Expect(1, 12988, '\P{^Is_Nv=04_7}', "");
Expect(0, 12989, '\p{Is_Nv=04_7}', "");
Expect(1, 12989, '\p{^Is_Nv=04_7}', "");
Expect(1, 12989, '\P{Is_Nv=04_7}', "");
Expect(0, 12989, '\P{^Is_Nv=04_7}', "");
Error('\p{Numeric_Value: _:=000001/006}');
Error('\P{Numeric_Value: _:=000001/006}');
Expect(1, 74849, '\p{Numeric_Value:	0000000001/0006}', "");
Expect(0, 74849, '\p{^Numeric_Value:	0000000001/0006}', "");
Expect(0, 74849, '\P{Numeric_Value:	0000000001/0006}', "");
Expect(1, 74849, '\P{^Numeric_Value:	0000000001/0006}', "");
Expect(0, 74850, '\p{Numeric_Value:	0000000001/0006}', "");
Expect(1, 74850, '\p{^Numeric_Value:	0000000001/0006}', "");
Expect(1, 74850, '\P{Numeric_Value:	0000000001/0006}', "");
Expect(0, 74850, '\P{^Numeric_Value:	0000000001/0006}', "");
Error('\p{Numeric_Value=0.17}');
Error('\P{Numeric_Value=0.17}');
Expect(1, 74849, '\p{Numeric_Value:	0.167}', "");
Expect(0, 74849, '\p{^Numeric_Value:	0.167}', "");
Expect(0, 74849, '\P{Numeric_Value:	0.167}', "");
Expect(1, 74849, '\P{^Numeric_Value:	0.167}', "");
Expect(0, 74850, '\p{Numeric_Value:	0.167}', "");
Expect(1, 74850, '\p{^Numeric_Value:	0.167}', "");
Expect(1, 74850, '\P{Numeric_Value:	0.167}', "");
Expect(0, 74850, '\P{^Numeric_Value:	0.167}', "");
Error('\p{Nv=_000000001/00000006:=}');
Error('\P{Nv=_000000001/00000006:=}');
Expect(1, 74849, '\p{Nv=0000000001/00000006}', "");
Expect(0, 74849, '\p{^Nv=0000000001/00000006}', "");
Expect(0, 74849, '\P{Nv=0000000001/00000006}', "");
Expect(1, 74849, '\P{^Nv=0000000001/00000006}', "");
Expect(0, 74850, '\p{Nv=0000000001/00000006}', "");
Expect(1, 74850, '\p{^Nv=0000000001/00000006}', "");
Expect(1, 74850, '\P{Nv=0000000001/00000006}', "");
Expect(0, 74850, '\P{^Nv=0000000001/00000006}', "");
Error('\p{Nv:	0.17}');
Error('\P{Nv:	0.17}');
Expect(1, 74849, '\p{Nv=0.167}', "");
Expect(0, 74849, '\p{^Nv=0.167}', "");
Expect(0, 74849, '\P{Nv=0.167}', "");
Expect(1, 74849, '\P{^Nv=0.167}', "");
Expect(0, 74850, '\p{Nv=0.167}', "");
Expect(1, 74850, '\p{^Nv=0.167}', "");
Expect(1, 74850, '\P{Nv=0.167}', "");
Expect(0, 74850, '\P{^Nv=0.167}', "");
Error('\p{Is_Numeric_Value=/a/ 0001/0000000006}');
Error('\P{Is_Numeric_Value=/a/ 0001/0000000006}');
Expect(1, 74849, '\p{Is_Numeric_Value:   0001/006}', "");
Expect(0, 74849, '\p{^Is_Numeric_Value:   0001/006}', "");
Expect(0, 74849, '\P{Is_Numeric_Value:   0001/006}', "");
Expect(1, 74849, '\P{^Is_Numeric_Value:   0001/006}', "");
Expect(0, 74850, '\p{Is_Numeric_Value:   0001/006}', "");
Expect(1, 74850, '\p{^Is_Numeric_Value:   0001/006}', "");
Expect(1, 74850, '\P{Is_Numeric_Value:   0001/006}', "");
Expect(0, 74850, '\P{^Is_Numeric_Value:   0001/006}', "");
Error('\p{Is_Numeric_Value:   0.17}');
Error('\P{Is_Numeric_Value:   0.17}');
Expect(1, 74849, '\p{Is_Numeric_Value=0.167}', "");
Expect(0, 74849, '\p{^Is_Numeric_Value=0.167}', "");
Expect(0, 74849, '\P{Is_Numeric_Value=0.167}', "");
Expect(1, 74849, '\P{^Is_Numeric_Value=0.167}', "");
Expect(0, 74850, '\p{Is_Numeric_Value=0.167}', "");
Expect(1, 74850, '\p{^Is_Numeric_Value=0.167}', "");
Expect(1, 74850, '\P{Is_Numeric_Value=0.167}', "");
Expect(0, 74850, '\P{^Is_Numeric_Value=0.167}', "");
Error('\p{Is_Nv:   _ 01/0006/a/}');
Error('\P{Is_Nv:   _ 01/0006/a/}');
Expect(1, 74849, '\p{Is_Nv=+0000000001/00006}', "");
Expect(0, 74849, '\p{^Is_Nv=+0000000001/00006}', "");
Expect(0, 74849, '\P{Is_Nv=+0000000001/00006}', "");
Expect(1, 74849, '\P{^Is_Nv=+0000000001/00006}', "");
Expect(0, 74850, '\p{Is_Nv=+0000000001/00006}', "");
Expect(1, 74850, '\p{^Is_Nv=+0000000001/00006}', "");
Expect(1, 74850, '\P{Is_Nv=+0000000001/00006}', "");
Expect(0, 74850, '\P{^Is_Nv=+0000000001/00006}', "");
Error('\p{Is_Nv=0.17}');
Error('\P{Is_Nv=0.17}');
Expect(1, 74849, '\p{Is_Nv=0.167}', "");
Expect(0, 74849, '\p{^Is_Nv=0.167}', "");
Expect(0, 74849, '\P{Is_Nv=0.167}', "");
Expect(1, 74849, '\P{^Is_Nv=0.167}', "");
Expect(0, 74850, '\p{Is_Nv=0.167}', "");
Expect(1, 74850, '\p{^Is_Nv=0.167}', "");
Expect(1, 74850, '\P{Is_Nv=0.167}', "");
Expect(0, 74850, '\P{^Is_Nv=0.167}', "");
Error('\p{Numeric_Value=	/a/003_8}');
Error('\P{Numeric_Value=	/a/003_8}');
Expect(1, 12979, '\p{Numeric_Value=00_00_00_03_8}', "");
Expect(0, 12979, '\p{^Numeric_Value=00_00_00_03_8}', "");
Expect(0, 12979, '\P{Numeric_Value=00_00_00_03_8}', "");
Expect(1, 12979, '\P{^Numeric_Value=00_00_00_03_8}', "");
Expect(0, 12980, '\p{Numeric_Value=00_00_00_03_8}', "");
Expect(1, 12980, '\p{^Numeric_Value=00_00_00_03_8}', "");
Expect(1, 12980, '\P{Numeric_Value=00_00_00_03_8}', "");
Expect(0, 12980, '\P{^Numeric_Value=00_00_00_03_8}', "");
Error('\p{Nv: := _00_00_03_8}');
Error('\P{Nv: := _00_00_03_8}');
Expect(1, 12979, '\p{Nv=00_00_03_8}', "");
Expect(0, 12979, '\p{^Nv=00_00_03_8}', "");
Expect(0, 12979, '\P{Nv=00_00_03_8}', "");
Expect(1, 12979, '\P{^Nv=00_00_03_8}', "");
Expect(0, 12980, '\p{Nv=00_00_03_8}', "");
Expect(1, 12980, '\p{^Nv=00_00_03_8}', "");
Expect(1, 12980, '\P{Nv=00_00_03_8}', "");
Expect(0, 12980, '\P{^Nv=00_00_03_8}', "");
Error('\p{Is_Numeric_Value=/a/ -00_00_00_00_03_8}');
Error('\P{Is_Numeric_Value=/a/ -00_00_00_00_03_8}');
Expect(1, 12979, '\p{Is_Numeric_Value=00000038}', "");
Expect(0, 12979, '\p{^Is_Numeric_Value=00000038}', "");
Expect(0, 12979, '\P{Is_Numeric_Value=00000038}', "");
Expect(1, 12979, '\P{^Is_Numeric_Value=00000038}', "");
Expect(0, 12980, '\p{Is_Numeric_Value=00000038}', "");
Expect(1, 12980, '\p{^Is_Numeric_Value=00000038}', "");
Expect(1, 12980, '\P{Is_Numeric_Value=00000038}', "");
Expect(0, 12980, '\P{^Is_Numeric_Value=00000038}', "");
Error('\p{Is_Nv= 003_8/a/}');
Error('\P{Is_Nv= 003_8/a/}');
Expect(1, 12979, '\p{Is_Nv=0_0_0_0_038}', "");
Expect(0, 12979, '\p{^Is_Nv=0_0_0_0_038}', "");
Expect(0, 12979, '\P{Is_Nv=0_0_0_0_038}', "");
Expect(1, 12979, '\P{^Is_Nv=0_0_0_0_038}', "");
Expect(0, 12980, '\p{Is_Nv=0_0_0_0_038}', "");
Expect(1, 12980, '\p{^Is_Nv=0_0_0_0_038}', "");
Expect(1, 12980, '\P{Is_Nv=0_0_0_0_038}', "");
Expect(0, 12980, '\P{^Is_Nv=0_0_0_0_038}', "");
Error('\p{Numeric_Value=_:=+8}');
Error('\P{Numeric_Value=_:=+8}');
Expect(1, 127241, '\p{Numeric_Value=+0_0_08}', "");
Expect(0, 127241, '\p{^Numeric_Value=+0_0_08}', "");
Expect(0, 127241, '\P{Numeric_Value=+0_0_08}', "");
Expect(1, 127241, '\P{^Numeric_Value=+0_0_08}', "");
Expect(0, 127242, '\p{Numeric_Value=+0_0_08}', "");
Expect(1, 127242, '\p{^Numeric_Value=+0_0_08}', "");
Expect(1, 127242, '\P{Numeric_Value=+0_0_08}', "");
Expect(0, 127242, '\P{^Numeric_Value=+0_0_08}', "");
Error('\p{Nv=:=-_0_0_0_08}');
Error('\P{Nv=:=-_0_0_0_08}');
Expect(1, 127241, '\p{Nv=08}', "");
Expect(0, 127241, '\p{^Nv=08}', "");
Expect(0, 127241, '\P{Nv=08}', "");
Expect(1, 127241, '\P{^Nv=08}', "");
Expect(0, 127242, '\p{Nv=08}', "");
Expect(1, 127242, '\p{^Nv=08}', "");
Expect(1, 127242, '\P{Nv=08}', "");
Expect(0, 127242, '\P{^Nv=08}', "");
Error('\p{Is_Numeric_Value=:=	_000000008}');
Error('\P{Is_Numeric_Value=:=	_000000008}');
Expect(1, 127241, '\p{Is_Numeric_Value=008}', "");
Expect(0, 127241, '\p{^Is_Numeric_Value=008}', "");
Expect(0, 127241, '\P{Is_Numeric_Value=008}', "");
Expect(1, 127241, '\P{^Is_Numeric_Value=008}', "");
Expect(0, 127242, '\p{Is_Numeric_Value=008}', "");
Expect(1, 127242, '\p{^Is_Numeric_Value=008}', "");
Expect(1, 127242, '\P{Is_Numeric_Value=008}', "");
Expect(0, 127242, '\P{^Is_Numeric_Value=008}', "");
Error('\p{Is_Nv=/a/ -0_8}');
Error('\P{Is_Nv=/a/ -0_8}');
Expect(1, 127241, '\p{Is_Nv=0_0_0_0_08}', "");
Expect(0, 127241, '\p{^Is_Nv=0_0_0_0_08}', "");
Expect(0, 127241, '\P{Is_Nv=0_0_0_0_08}', "");
Expect(1, 127241, '\P{^Is_Nv=0_0_0_0_08}', "");
Expect(0, 127242, '\p{Is_Nv=0_0_0_0_08}', "");
Expect(1, 127242, '\p{^Is_Nv=0_0_0_0_08}', "");
Expect(1, 127242, '\P{Is_Nv=0_0_0_0_08}', "");
Expect(0, 127242, '\P{^Is_Nv=0_0_0_0_08}', "");
Error('\p{Numeric_Value:   /a/_ 0000000003/4}');
Error('\P{Numeric_Value:   /a/_ 0000000003/4}');
Expect(1, 65912, '\p{Numeric_Value=0003/04}', "");
Expect(0, 65912, '\p{^Numeric_Value=0003/04}', "");
Expect(0, 65912, '\P{Numeric_Value=0003/04}', "");
Expect(1, 65912, '\P{^Numeric_Value=0003/04}', "");
Expect(0, 65913, '\p{Numeric_Value=0003/04}', "");
Expect(1, 65913, '\p{^Numeric_Value=0003/04}', "");
Expect(1, 65913, '\P{Numeric_Value=0003/04}', "");
Expect(0, 65913, '\P{^Numeric_Value=0003/04}', "");
Expect(1, 65912, '\p{Numeric_Value: 0.750}', "");
Expect(0, 65912, '\p{^Numeric_Value: 0.750}', "");
Expect(0, 65912, '\P{Numeric_Value: 0.750}', "");
Expect(1, 65912, '\P{^Numeric_Value: 0.750}', "");
Expect(0, 65913, '\p{Numeric_Value: 0.750}', "");
Expect(1, 65913, '\p{^Numeric_Value: 0.750}', "");
Expect(1, 65913, '\P{Numeric_Value: 0.750}', "");
Expect(0, 65913, '\P{^Numeric_Value: 0.750}', "");
Error('\p{Nv: __003/000000004:=}');
Error('\P{Nv: __003/000000004:=}');
Expect(1, 65912, '\p{Nv:	+003/0000004}', "");
Expect(0, 65912, '\p{^Nv:	+003/0000004}', "");
Expect(0, 65912, '\P{Nv:	+003/0000004}', "");
Expect(1, 65912, '\P{^Nv:	+003/0000004}', "");
Expect(0, 65913, '\p{Nv:	+003/0000004}', "");
Expect(1, 65913, '\p{^Nv:	+003/0000004}', "");
Expect(1, 65913, '\P{Nv:	+003/0000004}', "");
Expect(0, 65913, '\P{^Nv:	+003/0000004}', "");
Expect(1, 65912, '\p{Nv=0.750}', "");
Expect(0, 65912, '\p{^Nv=0.750}', "");
Expect(0, 65912, '\P{Nv=0.750}', "");
Expect(1, 65912, '\P{^Nv=0.750}', "");
Expect(0, 65913, '\p{Nv=0.750}', "");
Expect(1, 65913, '\p{^Nv=0.750}', "");
Expect(1, 65913, '\P{Nv=0.750}', "");
Expect(0, 65913, '\P{^Nv=0.750}', "");
Error('\p{Is_Numeric_Value= 	0000000003/0000004/a/}');
Error('\P{Is_Numeric_Value= 	0000000003/0000004/a/}');
Expect(1, 65912, '\p{Is_Numeric_Value=00003/0000004}', "");
Expect(0, 65912, '\p{^Is_Numeric_Value=00003/0000004}', "");
Expect(0, 65912, '\P{Is_Numeric_Value=00003/0000004}', "");
Expect(1, 65912, '\P{^Is_Numeric_Value=00003/0000004}', "");
Expect(0, 65913, '\p{Is_Numeric_Value=00003/0000004}', "");
Expect(1, 65913, '\p{^Is_Numeric_Value=00003/0000004}', "");
Expect(1, 65913, '\P{Is_Numeric_Value=00003/0000004}', "");
Expect(0, 65913, '\P{^Is_Numeric_Value=00003/0000004}', "");
Expect(1, 65912, '\p{Is_Numeric_Value=0.750}', "");
Expect(0, 65912, '\p{^Is_Numeric_Value=0.750}', "");
Expect(0, 65912, '\P{Is_Numeric_Value=0.750}', "");
Expect(1, 65912, '\P{^Is_Numeric_Value=0.750}', "");
Expect(0, 65913, '\p{Is_Numeric_Value=0.750}', "");
Expect(1, 65913, '\p{^Is_Numeric_Value=0.750}', "");
Expect(1, 65913, '\P{Is_Numeric_Value=0.750}', "");
Expect(0, 65913, '\P{^Is_Numeric_Value=0.750}', "");
Error('\p{Is_Nv=/a/ 	+0000000003/0000000004}');
Error('\P{Is_Nv=/a/ 	+0000000003/0000000004}');
Expect(1, 65912, '\p{Is_Nv=0000003/000000004}', "");
Expect(0, 65912, '\p{^Is_Nv=0000003/000000004}', "");
Expect(0, 65912, '\P{Is_Nv=0000003/000000004}', "");
Expect(1, 65912, '\P{^Is_Nv=0000003/000000004}', "");
Expect(0, 65913, '\p{Is_Nv=0000003/000000004}', "");
Expect(1, 65913, '\p{^Is_Nv=0000003/000000004}', "");
Expect(1, 65913, '\P{Is_Nv=0000003/000000004}', "");
Expect(0, 65913, '\P{^Is_Nv=0000003/000000004}', "");
Expect(1, 65912, '\p{Is_Nv=0.750}', "");
Expect(0, 65912, '\p{^Is_Nv=0.750}', "");
Expect(0, 65912, '\P{Is_Nv=0.750}', "");
Expect(1, 65912, '\P{^Is_Nv=0.750}', "");
Expect(0, 65913, '\p{Is_Nv=0.750}', "");
Expect(1, 65913, '\p{^Is_Nv=0.750}', "");
Expect(1, 65913, '\P{Is_Nv=0.750}', "");
Expect(0, 65913, '\P{^Is_Nv=0.750}', "");
Error('\p{Numeric_Value=/a/	 4}');
Error('\P{Numeric_Value=/a/	 4}');
Expect(1, 156269, '\p{Numeric_Value=04}', "");
Expect(0, 156269, '\p{^Numeric_Value=04}', "");
Expect(0, 156269, '\P{Numeric_Value=04}', "");
Expect(1, 156269, '\P{^Numeric_Value=04}', "");
Expect(0, 156270, '\p{Numeric_Value=04}', "");
Expect(1, 156270, '\p{^Numeric_Value=04}', "");
Expect(1, 156270, '\P{Numeric_Value=04}', "");
Expect(0, 156270, '\P{^Numeric_Value=04}', "");
Error('\p{Nv=_/a/000004}');
Error('\P{Nv=_/a/000004}');
Expect(1, 156269, '\p{Nv=04}', "");
Expect(0, 156269, '\p{^Nv=04}', "");
Expect(0, 156269, '\P{Nv=04}', "");
Expect(1, 156269, '\P{^Nv=04}', "");
Expect(0, 156270, '\p{Nv=04}', "");
Expect(1, 156270, '\p{^Nv=04}', "");
Expect(1, 156270, '\P{Nv=04}', "");
Expect(0, 156270, '\P{^Nv=04}', "");
Error('\p{Is_Numeric_Value=:= 0004}');
Error('\P{Is_Numeric_Value=:= 0004}');
Expect(1, 156269, '\p{Is_Numeric_Value=+00000004}', "");
Expect(0, 156269, '\p{^Is_Numeric_Value=+00000004}', "");
Expect(0, 156269, '\P{Is_Numeric_Value=+00000004}', "");
Expect(1, 156269, '\P{^Is_Numeric_Value=+00000004}', "");
Expect(0, 156270, '\p{Is_Numeric_Value=+00000004}', "");
Expect(1, 156270, '\p{^Is_Numeric_Value=+00000004}', "");
Expect(1, 156270, '\P{Is_Numeric_Value=+00000004}', "");
Expect(0, 156270, '\P{^Is_Numeric_Value=+00000004}', "");
Error('\p{Is_Nv=--+0_0_0_004/a/}');
Error('\P{Is_Nv=--+0_0_0_004/a/}');
Expect(1, 156269, '\p{Is_Nv=000000004}', "");
Expect(0, 156269, '\p{^Is_Nv=000000004}', "");
Expect(0, 156269, '\P{Is_Nv=000000004}', "");
Expect(1, 156269, '\P{^Is_Nv=000000004}', "");
Expect(0, 156270, '\p{Is_Nv=000000004}', "");
Expect(1, 156270, '\p{^Is_Nv=000000004}', "");
Expect(1, 156270, '\P{Is_Nv=000000004}', "");
Expect(0, 156270, '\P{^Is_Nv=000000004}', "");
Error('\p{Numeric_Value=_	0_0_0_0_0_034:=}');
Error('\P{Numeric_Value=_	0_0_0_0_0_034:=}');
Expect(1, 12894, '\p{Numeric_Value=+003_4}', "");
Expect(0, 12894, '\p{^Numeric_Value=+003_4}', "");
Expect(0, 12894, '\P{Numeric_Value=+003_4}', "");
Expect(1, 12894, '\P{^Numeric_Value=+003_4}', "");
Expect(0, 12895, '\p{Numeric_Value=+003_4}', "");
Expect(1, 12895, '\p{^Numeric_Value=+003_4}', "");
Expect(1, 12895, '\P{Numeric_Value=+003_4}', "");
Expect(0, 12895, '\P{^Numeric_Value=+003_4}', "");
Error('\p{Nv=/a/  +0000034}');
Error('\P{Nv=/a/  +0000034}');
Expect(1, 12894, '\p{Nv=0_0_0_0_0_0034}', "");
Expect(0, 12894, '\p{^Nv=0_0_0_0_0_0034}', "");
Expect(0, 12894, '\P{Nv=0_0_0_0_0_0034}', "");
Expect(1, 12894, '\P{^Nv=0_0_0_0_0_0034}', "");
Expect(0, 12895, '\p{Nv=0_0_0_0_0_0034}', "");
Expect(1, 12895, '\p{^Nv=0_0_0_0_0_0034}', "");
Expect(1, 12895, '\P{Nv=0_0_0_0_0_0034}', "");
Expect(0, 12895, '\P{^Nv=0_0_0_0_0_0034}', "");
Error('\p{Is_Numeric_Value=--000_003_4/a/}');
Error('\P{Is_Numeric_Value=--000_003_4/a/}');
Expect(1, 12894, '\p{Is_Numeric_Value=0000_0000_34}', "");
Expect(0, 12894, '\p{^Is_Numeric_Value=0000_0000_34}', "");
Expect(0, 12894, '\P{Is_Numeric_Value=0000_0000_34}', "");
Expect(1, 12894, '\P{^Is_Numeric_Value=0000_0000_34}', "");
Expect(0, 12895, '\p{Is_Numeric_Value=0000_0000_34}', "");
Expect(1, 12895, '\p{^Is_Numeric_Value=0000_0000_34}', "");
Expect(1, 12895, '\P{Is_Numeric_Value=0000_0000_34}', "");
Expect(0, 12895, '\P{^Is_Numeric_Value=0000_0000_34}', "");
Error('\p{Is_Nv=_-0000000034:=}');
Error('\P{Is_Nv=_-0000000034:=}');
Expect(1, 12894, '\p{Is_Nv=00000000034}', "");
Expect(0, 12894, '\p{^Is_Nv=00000000034}', "");
Expect(0, 12894, '\P{Is_Nv=00000000034}', "");
Expect(1, 12894, '\P{^Is_Nv=00000000034}', "");
Expect(0, 12895, '\p{Is_Nv=00000000034}', "");
Expect(1, 12895, '\p{^Is_Nv=00000000034}', "");
Expect(1, 12895, '\P{Is_Nv=00000000034}', "");
Expect(0, 12895, '\P{^Is_Nv=00000000034}', "");
Error('\p{Numeric_Value=	-0_0_0_0_0_0037/a/}');
Error('\P{Numeric_Value=	-0_0_0_0_0_0037/a/}');
Expect(1, 12978, '\p{Numeric_Value=0_0_0_0_0_037}', "");
Expect(0, 12978, '\p{^Numeric_Value=0_0_0_0_0_037}', "");
Expect(0, 12978, '\P{Numeric_Value=0_0_0_0_0_037}', "");
Expect(1, 12978, '\P{^Numeric_Value=0_0_0_0_0_037}', "");
Expect(0, 12979, '\p{Numeric_Value=0_0_0_0_0_037}', "");
Expect(1, 12979, '\p{^Numeric_Value=0_0_0_0_0_037}', "");
Expect(1, 12979, '\P{Numeric_Value=0_0_0_0_0_037}', "");
Expect(0, 12979, '\P{^Numeric_Value=0_0_0_0_0_037}', "");
Error('\p{Nv=-:=0_0_0_037}');
Error('\P{Nv=-:=0_0_0_037}');
Expect(1, 12978, '\p{Nv=+000003_7}', "");
Expect(0, 12978, '\p{^Nv=+000003_7}', "");
Expect(0, 12978, '\P{Nv=+000003_7}', "");
Expect(1, 12978, '\P{^Nv=+000003_7}', "");
Expect(0, 12979, '\p{Nv=+000003_7}', "");
Expect(1, 12979, '\p{^Nv=+000003_7}', "");
Expect(1, 12979, '\P{Nv=+000003_7}', "");
Expect(0, 12979, '\P{^Nv=+000003_7}', "");
Error('\p{Is_Numeric_Value=/a/	+00_00_00_037}');
Error('\P{Is_Numeric_Value=/a/	+00_00_00_037}');
Expect(1, 12978, '\p{Is_Numeric_Value=+0000000037}', "");
Expect(0, 12978, '\p{^Is_Numeric_Value=+0000000037}', "");
Expect(0, 12978, '\P{Is_Numeric_Value=+0000000037}', "");
Expect(1, 12978, '\P{^Is_Numeric_Value=+0000000037}', "");
Expect(0, 12979, '\p{Is_Numeric_Value=+0000000037}', "");
Expect(1, 12979, '\p{^Is_Numeric_Value=+0000000037}', "");
Expect(1, 12979, '\P{Is_Numeric_Value=+0000000037}', "");
Expect(0, 12979, '\P{^Is_Numeric_Value=+0000000037}', "");
Error('\p{Is_Nv=/a/	+0_0_0_0_0_0_00037}');
Error('\P{Is_Nv=/a/	+0_0_0_0_0_0_00037}');
Expect(1, 12978, '\p{Is_Nv=+037}', "");
Expect(0, 12978, '\p{^Is_Nv=+037}', "");
Expect(0, 12978, '\P{Is_Nv=+037}', "");
Expect(1, 12978, '\P{^Is_Nv=+037}', "");
Expect(0, 12979, '\p{Is_Nv=+037}', "");
Expect(1, 12979, '\p{^Is_Nv=+037}', "");
Expect(1, 12979, '\P{Is_Nv=+037}', "");
Expect(0, 12979, '\P{^Is_Nv=+037}', "");
Error('\p{Numeric_Value=/a/_ 00005/02}');
Error('\P{Numeric_Value=/a/_ 00005/02}');
Expect(1, 3884, '\p{Numeric_Value=+000000005/0002}', "");
Expect(0, 3884, '\p{^Numeric_Value=+000000005/0002}', "");
Expect(0, 3884, '\P{Numeric_Value=+000000005/0002}', "");
Expect(1, 3884, '\P{^Numeric_Value=+000000005/0002}', "");
Expect(0, 3885, '\p{Numeric_Value=+000000005/0002}', "");
Expect(1, 3885, '\p{^Numeric_Value=+000000005/0002}', "");
Expect(1, 3885, '\P{Numeric_Value=+000000005/0002}', "");
Expect(0, 3885, '\P{^Numeric_Value=+000000005/0002}', "");
Expect(1, 3884, '\p{Numeric_Value=2.500}', "");
Expect(0, 3884, '\p{^Numeric_Value=2.500}', "");
Expect(0, 3884, '\P{Numeric_Value=2.500}', "");
Expect(1, 3884, '\P{^Numeric_Value=2.500}', "");
Expect(0, 3885, '\p{Numeric_Value=2.500}', "");
Expect(1, 3885, '\p{^Numeric_Value=2.500}', "");
Expect(1, 3885, '\P{Numeric_Value=2.500}', "");
Expect(0, 3885, '\P{^Numeric_Value=2.500}', "");
Error('\p{Nv=:=	_00005/2}');
Error('\P{Nv=:=	_00005/2}');
Expect(1, 3884, '\p{Nv=5/00000002}', "");
Expect(0, 3884, '\p{^Nv=5/00000002}', "");
Expect(0, 3884, '\P{Nv=5/00000002}', "");
Expect(1, 3884, '\P{^Nv=5/00000002}', "");
Expect(0, 3885, '\p{Nv=5/00000002}', "");
Expect(1, 3885, '\p{^Nv=5/00000002}', "");
Expect(1, 3885, '\P{Nv=5/00000002}', "");
Expect(0, 3885, '\P{^Nv=5/00000002}', "");
Expect(1, 3884, '\p{Nv=2.500}', "");
Expect(0, 3884, '\p{^Nv=2.500}', "");
Expect(0, 3884, '\P{Nv=2.500}', "");
Expect(1, 3884, '\P{^Nv=2.500}', "");
Expect(0, 3885, '\p{Nv=2.500}', "");
Expect(1, 3885, '\p{^Nv=2.500}', "");
Expect(1, 3885, '\P{Nv=2.500}', "");
Expect(0, 3885, '\P{^Nv=2.500}', "");
Error('\p{Is_Numeric_Value=/a/5/00000002}');
Error('\P{Is_Numeric_Value=/a/5/00000002}');
Expect(1, 3884, '\p{Is_Numeric_Value: 05/00000002}', "");
Expect(0, 3884, '\p{^Is_Numeric_Value: 05/00000002}', "");
Expect(0, 3884, '\P{Is_Numeric_Value: 05/00000002}', "");
Expect(1, 3884, '\P{^Is_Numeric_Value: 05/00000002}', "");
Expect(0, 3885, '\p{Is_Numeric_Value: 05/00000002}', "");
Expect(1, 3885, '\p{^Is_Numeric_Value: 05/00000002}', "");
Expect(1, 3885, '\P{Is_Numeric_Value: 05/00000002}', "");
Expect(0, 3885, '\P{^Is_Numeric_Value: 05/00000002}', "");
Expect(1, 3884, '\p{Is_Numeric_Value=2.500}', "");
Expect(0, 3884, '\p{^Is_Numeric_Value=2.500}', "");
Expect(0, 3884, '\P{Is_Numeric_Value=2.500}', "");
Expect(1, 3884, '\P{^Is_Numeric_Value=2.500}', "");
Expect(0, 3885, '\p{Is_Numeric_Value=2.500}', "");
Expect(1, 3885, '\p{^Is_Numeric_Value=2.500}', "");
Expect(1, 3885, '\P{Is_Numeric_Value=2.500}', "");
Expect(0, 3885, '\P{^Is_Numeric_Value=2.500}', "");
Error('\p{Is_Nv=	 +00000005/00000002:=}');
Error('\P{Is_Nv=	 +00000005/00000002:=}');
Expect(1, 3884, '\p{Is_Nv: 5/00002}', "");
Expect(0, 3884, '\p{^Is_Nv: 5/00002}', "");
Expect(0, 3884, '\P{Is_Nv: 5/00002}', "");
Expect(1, 3884, '\P{^Is_Nv: 5/00002}', "");
Expect(0, 3885, '\p{Is_Nv: 5/00002}', "");
Expect(1, 3885, '\p{^Is_Nv: 5/00002}', "");
Expect(1, 3885, '\P{Is_Nv: 5/00002}', "");
Expect(0, 3885, '\P{^Is_Nv: 5/00002}', "");
Expect(1, 3884, '\p{Is_Nv=2.500}', "");
Expect(0, 3884, '\p{^Is_Nv=2.500}', "");
Expect(0, 3884, '\P{Is_Nv=2.500}', "");
Expect(1, 3884, '\P{^Is_Nv=2.500}', "");
Expect(0, 3885, '\p{Is_Nv=2.500}', "");
Expect(1, 3885, '\p{^Is_Nv=2.500}', "");
Expect(1, 3885, '\P{Is_Nv=2.500}', "");
Expect(0, 3885, '\P{^Is_Nv=2.500}', "");
Error('\p{Numeric_Value= 0_0_0_0_0_0_0_1_000:=}');
Error('\P{Numeric_Value= 0_0_0_0_0_0_0_1_000:=}');
Expect(1, 69733, '\p{Numeric_Value=0000001000}', "");
Expect(0, 69733, '\p{^Numeric_Value=0000001000}', "");
Expect(0, 69733, '\P{Numeric_Value=0000001000}', "");
Expect(1, 69733, '\P{^Numeric_Value=0000001000}', "");
Expect(0, 69734, '\p{Numeric_Value=0000001000}', "");
Expect(1, 69734, '\p{^Numeric_Value=0000001000}', "");
Expect(1, 69734, '\P{Numeric_Value=0000001000}', "");
Expect(0, 69734, '\P{^Numeric_Value=0000001000}', "");
Error('\p{Nv=:=_01000}');
Error('\P{Nv=:=_01000}');
Expect(1, 69733, '\p{Nv=0000000001000}', "");
Expect(0, 69733, '\p{^Nv=0000000001000}', "");
Expect(0, 69733, '\P{Nv=0000000001000}', "");
Expect(1, 69733, '\P{^Nv=0000000001000}', "");
Expect(0, 69734, '\p{Nv=0000000001000}', "");
Expect(1, 69734, '\p{^Nv=0000000001000}', "");
Expect(1, 69734, '\P{Nv=0000000001000}', "");
Expect(0, 69734, '\P{^Nv=0000000001000}', "");
Error('\p{Is_Numeric_Value=_	000_000_001_000/a/}');
Error('\P{Is_Numeric_Value=_	000_000_001_000/a/}');
Expect(1, 69733, '\p{Is_Numeric_Value=+0_0_0_1_0_00}', "");
Expect(0, 69733, '\p{^Is_Numeric_Value=+0_0_0_1_0_00}', "");
Expect(0, 69733, '\P{Is_Numeric_Value=+0_0_0_1_0_00}', "");
Expect(1, 69733, '\P{^Is_Numeric_Value=+0_0_0_1_0_00}', "");
Expect(0, 69734, '\p{Is_Numeric_Value=+0_0_0_1_0_00}', "");
Expect(1, 69734, '\p{^Is_Numeric_Value=+0_0_0_1_0_00}', "");
Expect(1, 69734, '\P{Is_Numeric_Value=+0_0_0_1_0_00}', "");
Expect(0, 69734, '\P{^Is_Numeric_Value=+0_0_0_1_0_00}', "");
Error('\p{Is_Nv:   :=+00001000}');
Error('\P{Is_Nv:   :=+00001000}');
Expect(1, 69733, '\p{Is_Nv=0000000001000}', "");
Expect(0, 69733, '\p{^Is_Nv=0000000001000}', "");
Expect(0, 69733, '\P{Is_Nv=0000000001000}', "");
Expect(1, 69733, '\P{^Is_Nv=0000000001000}', "");
Expect(0, 69734, '\p{Is_Nv=0000000001000}', "");
Expect(1, 69734, '\p{^Is_Nv=0000000001000}', "");
Expect(1, 69734, '\P{Is_Nv=0000000001000}', "");
Expect(0, 69734, '\P{^Is_Nv=0000000001000}', "");
Error('\p{Numeric_Value=:=_	+00043}');
Error('\P{Numeric_Value=:=_	+00043}');
Expect(1, 12984, '\p{Numeric_Value=+0004_3}', "");
Expect(0, 12984, '\p{^Numeric_Value=+0004_3}', "");
Expect(0, 12984, '\P{Numeric_Value=+0004_3}', "");
Expect(1, 12984, '\P{^Numeric_Value=+0004_3}', "");
Expect(0, 12985, '\p{Numeric_Value=+0004_3}', "");
Expect(1, 12985, '\p{^Numeric_Value=+0004_3}', "");
Expect(1, 12985, '\P{Numeric_Value=+0004_3}', "");
Expect(0, 12985, '\P{^Numeric_Value=+0004_3}', "");
Error('\p{Nv=_	+00_04_3:=}');
Error('\P{Nv=_	+00_04_3:=}');
Expect(1, 12984, '\p{Nv=+000043}', "");
Expect(0, 12984, '\p{^Nv=+000043}', "");
Expect(0, 12984, '\P{Nv=+000043}', "");
Expect(1, 12984, '\P{^Nv=+000043}', "");
Expect(0, 12985, '\p{Nv=+000043}', "");
Expect(1, 12985, '\p{^Nv=+000043}', "");
Expect(1, 12985, '\P{Nv=+000043}', "");
Expect(0, 12985, '\P{^Nv=+000043}', "");
Error('\p{Is_Numeric_Value=/a/+0000000043}');
Error('\P{Is_Numeric_Value=/a/+0000000043}');
Expect(1, 12984, '\p{Is_Numeric_Value=000_000_43}', "");
Expect(0, 12984, '\p{^Is_Numeric_Value=000_000_43}', "");
Expect(0, 12984, '\P{Is_Numeric_Value=000_000_43}', "");
Expect(1, 12984, '\P{^Is_Numeric_Value=000_000_43}', "");
Expect(0, 12985, '\p{Is_Numeric_Value=000_000_43}', "");
Expect(1, 12985, '\p{^Is_Numeric_Value=000_000_43}', "");
Expect(1, 12985, '\P{Is_Numeric_Value=000_000_43}', "");
Expect(0, 12985, '\P{^Is_Numeric_Value=000_000_43}', "");
Error('\p{Is_Nv= /a/043}');
Error('\P{Is_Nv= /a/043}');
Expect(1, 12984, '\p{Is_Nv=04_3}', "");
Expect(0, 12984, '\p{^Is_Nv=04_3}', "");
Expect(0, 12984, '\P{Is_Nv=04_3}', "");
Expect(1, 12984, '\P{^Is_Nv=04_3}', "");
Expect(0, 12985, '\p{Is_Nv=04_3}', "");
Expect(1, 12985, '\p{^Is_Nv=04_3}', "");
Expect(1, 12985, '\P{Is_Nv=04_3}', "");
Expect(0, 12985, '\P{^Is_Nv=04_3}', "");
Error('\p{Numeric_Value= :=00000005}');
Error('\P{Numeric_Value= :=00000005}');
Expect(1, 131361, '\p{Numeric_Value=+0_5}', "");
Expect(0, 131361, '\p{^Numeric_Value=+0_5}', "");
Expect(0, 131361, '\P{Numeric_Value=+0_5}', "");
Expect(1, 131361, '\P{^Numeric_Value=+0_5}', "");
Expect(0, 131362, '\p{Numeric_Value=+0_5}', "");
Expect(1, 131362, '\p{^Numeric_Value=+0_5}', "");
Expect(1, 131362, '\P{Numeric_Value=+0_5}', "");
Expect(0, 131362, '\P{^Numeric_Value=+0_5}', "");
Error('\p{Nv= 00_5/a/}');
Error('\P{Nv= 00_5/a/}');
Expect(1, 131361, '\p{Nv:   5}', "");
Expect(0, 131361, '\p{^Nv:   5}', "");
Expect(0, 131361, '\P{Nv:   5}', "");
Expect(1, 131361, '\P{^Nv:   5}', "");
Expect(0, 131362, '\p{Nv:   5}', "");
Expect(1, 131362, '\p{^Nv:   5}', "");
Expect(1, 131362, '\P{Nv:   5}', "");
Expect(0, 131362, '\P{^Nv:   5}', "");
Error('\p{Is_Numeric_Value=-	0_0_0_0_0_0005:=}');
Error('\P{Is_Numeric_Value=-	0_0_0_0_0_0005:=}');
Expect(1, 131361, '\p{Is_Numeric_Value=000000005}', "");
Expect(0, 131361, '\p{^Is_Numeric_Value=000000005}', "");
Expect(0, 131361, '\P{Is_Numeric_Value=000000005}', "");
Expect(1, 131361, '\P{^Is_Numeric_Value=000000005}', "");
Expect(0, 131362, '\p{Is_Numeric_Value=000000005}', "");
Expect(1, 131362, '\p{^Is_Numeric_Value=000000005}', "");
Expect(1, 131362, '\P{Is_Numeric_Value=000000005}', "");
Expect(0, 131362, '\P{^Is_Numeric_Value=000000005}', "");
Error('\p{Is_Nv=/a/ -005}');
Error('\P{Is_Nv=/a/ -005}');
Expect(1, 131361, '\p{Is_Nv:	0_0_0_0_05}', "");
Expect(0, 131361, '\p{^Is_Nv:	0_0_0_0_05}', "");
Expect(0, 131361, '\P{Is_Nv:	0_0_0_0_05}', "");
Expect(1, 131361, '\P{^Is_Nv:	0_0_0_0_05}', "");
Expect(0, 131362, '\p{Is_Nv:	0_0_0_0_05}', "");
Expect(1, 131362, '\p{^Is_Nv:	0_0_0_0_05}', "");
Expect(1, 131362, '\P{Is_Nv:	0_0_0_0_05}', "");
Expect(0, 131362, '\P{^Is_Nv:	0_0_0_0_05}', "");
Error('\p{Word_Break=-Newline:=}');
Error('\P{Word_Break=-Newline:=}');
Expect(1, 8233, '\p{Word_Break=newline}', "");
Expect(0, 8233, '\p{^Word_Break=newline}', "");
Expect(0, 8233, '\P{Word_Break=newline}', "");
Expect(1, 8233, '\P{^Word_Break=newline}', "");
Expect(0, 8234, '\p{Word_Break=newline}', "");
Expect(1, 8234, '\p{^Word_Break=newline}', "");
Expect(1, 8234, '\P{Word_Break=newline}', "");
Expect(0, 8234, '\P{^Word_Break=newline}', "");
Expect(1, 8233, '\p{Word_Break=_-NEWLINE}', "");
Expect(0, 8233, '\p{^Word_Break=_-NEWLINE}', "");
Expect(0, 8233, '\P{Word_Break=_-NEWLINE}', "");
Expect(1, 8233, '\P{^Word_Break=_-NEWLINE}', "");
Expect(0, 8234, '\p{Word_Break=_-NEWLINE}', "");
Expect(1, 8234, '\p{^Word_Break=_-NEWLINE}', "");
Expect(1, 8234, '\P{Word_Break=_-NEWLINE}', "");
Expect(0, 8234, '\P{^Word_Break=_-NEWLINE}', "");
Error('\p{WB=_/a/NL}');
Error('\P{WB=_/a/NL}');
Expect(1, 8233, '\p{WB=nl}', "");
Expect(0, 8233, '\p{^WB=nl}', "");
Expect(0, 8233, '\P{WB=nl}', "");
Expect(1, 8233, '\P{^WB=nl}', "");
Expect(0, 8234, '\p{WB=nl}', "");
Expect(1, 8234, '\p{^WB=nl}', "");
Expect(1, 8234, '\P{WB=nl}', "");
Expect(0, 8234, '\P{^WB=nl}', "");
Expect(1, 8233, '\p{WB=_ nl}', "");
Expect(0, 8233, '\p{^WB=_ nl}', "");
Expect(0, 8233, '\P{WB=_ nl}', "");
Expect(1, 8233, '\P{^WB=_ nl}', "");
Expect(0, 8234, '\p{WB=_ nl}', "");
Expect(1, 8234, '\p{^WB=_ nl}', "");
Expect(1, 8234, '\P{WB=_ nl}', "");
Expect(0, 8234, '\P{^WB=_ nl}', "");
Error('\p{Is_Word_Break=/a/		Newline}');
Error('\P{Is_Word_Break=/a/		Newline}');
Expect(1, 8233, '\p{Is_Word_Break=newline}', "");
Expect(0, 8233, '\p{^Is_Word_Break=newline}', "");
Expect(0, 8233, '\P{Is_Word_Break=newline}', "");
Expect(1, 8233, '\P{^Is_Word_Break=newline}', "");
Expect(0, 8234, '\p{Is_Word_Break=newline}', "");
Expect(1, 8234, '\p{^Is_Word_Break=newline}', "");
Expect(1, 8234, '\P{Is_Word_Break=newline}', "");
Expect(0, 8234, '\P{^Is_Word_Break=newline}', "");
Expect(1, 8233, '\p{Is_Word_Break=	_Newline}', "");
Expect(0, 8233, '\p{^Is_Word_Break=	_Newline}', "");
Expect(0, 8233, '\P{Is_Word_Break=	_Newline}', "");
Expect(1, 8233, '\P{^Is_Word_Break=	_Newline}', "");
Expect(0, 8234, '\p{Is_Word_Break=	_Newline}', "");
Expect(1, 8234, '\p{^Is_Word_Break=	_Newline}', "");
Expect(1, 8234, '\P{Is_Word_Break=	_Newline}', "");
Expect(0, 8234, '\P{^Is_Word_Break=	_Newline}', "");
Error('\p{Is_WB=/a/__NL}');
Error('\P{Is_WB=/a/__NL}');
Expect(1, 8233, '\p{Is_WB=nl}', "");
Expect(0, 8233, '\p{^Is_WB=nl}', "");
Expect(0, 8233, '\P{Is_WB=nl}', "");
Expect(1, 8233, '\P{^Is_WB=nl}', "");
Expect(0, 8234, '\p{Is_WB=nl}', "");
Expect(1, 8234, '\p{^Is_WB=nl}', "");
Expect(1, 8234, '\P{Is_WB=nl}', "");
Expect(0, 8234, '\P{^Is_WB=nl}', "");
Expect(1, 8233, '\p{Is_WB=_NL}', "");
Expect(0, 8233, '\p{^Is_WB=_NL}', "");
Expect(0, 8233, '\P{Is_WB=_NL}', "");
Expect(1, 8233, '\P{^Is_WB=_NL}', "");
Expect(0, 8234, '\p{Is_WB=_NL}', "");
Expect(1, 8234, '\p{^Is_WB=_NL}', "");
Expect(1, 8234, '\P{Is_WB=_NL}', "");
Expect(0, 8234, '\P{^Is_WB=_NL}', "");
Error('\p{Word_Break=	/a/ALETTER}');
Error('\P{Word_Break=	/a/ALETTER}');
Expect(1, 126651, '\p{Word_Break=aletter}', "");
Expect(0, 126651, '\p{^Word_Break=aletter}', "");
Expect(0, 126651, '\P{Word_Break=aletter}', "");
Expect(1, 126651, '\P{^Word_Break=aletter}', "");
Expect(0, 126652, '\p{Word_Break=aletter}', "");
Expect(1, 126652, '\p{^Word_Break=aletter}', "");
Expect(1, 126652, '\P{Word_Break=aletter}', "");
Expect(0, 126652, '\P{^Word_Break=aletter}', "");
Expect(1, 126651, '\p{Word_Break= -ALetter}', "");
Expect(0, 126651, '\p{^Word_Break= -ALetter}', "");
Expect(0, 126651, '\P{Word_Break= -ALetter}', "");
Expect(1, 126651, '\P{^Word_Break= -ALetter}', "");
Expect(0, 126652, '\p{Word_Break= -ALetter}', "");
Expect(1, 126652, '\p{^Word_Break= -ALetter}', "");
Expect(1, 126652, '\P{Word_Break= -ALetter}', "");
Expect(0, 126652, '\P{^Word_Break= -ALetter}', "");
Error('\p{WB=/a/	le}');
Error('\P{WB=/a/	le}');
Expect(1, 126651, '\p{WB=le}', "");
Expect(0, 126651, '\p{^WB=le}', "");
Expect(0, 126651, '\P{WB=le}', "");
Expect(1, 126651, '\P{^WB=le}', "");
Expect(0, 126652, '\p{WB=le}', "");
Expect(1, 126652, '\p{^WB=le}', "");
Expect(1, 126652, '\P{WB=le}', "");
Expect(0, 126652, '\P{^WB=le}', "");
Expect(1, 126651, '\p{WB=- le}', "");
Expect(0, 126651, '\p{^WB=- le}', "");
Expect(0, 126651, '\P{WB=- le}', "");
Expect(1, 126651, '\P{^WB=- le}', "");
Expect(0, 126652, '\p{WB=- le}', "");
Expect(1, 126652, '\p{^WB=- le}', "");
Expect(1, 126652, '\P{WB=- le}', "");
Expect(0, 126652, '\P{^WB=- le}', "");
Error('\p{Is_Word_Break=:=	ALetter}');
Error('\P{Is_Word_Break=:=	ALetter}');
Expect(1, 126651, '\p{Is_Word_Break=aletter}', "");
Expect(0, 126651, '\p{^Is_Word_Break=aletter}', "");
Expect(0, 126651, '\P{Is_Word_Break=aletter}', "");
Expect(1, 126651, '\P{^Is_Word_Break=aletter}', "");
Expect(0, 126652, '\p{Is_Word_Break=aletter}', "");
Expect(1, 126652, '\p{^Is_Word_Break=aletter}', "");
Expect(1, 126652, '\P{Is_Word_Break=aletter}', "");
Expect(0, 126652, '\P{^Is_Word_Break=aletter}', "");
Expect(1, 126651, '\p{Is_Word_Break=- aletter}', "");
Expect(0, 126651, '\p{^Is_Word_Break=- aletter}', "");
Expect(0, 126651, '\P{Is_Word_Break=- aletter}', "");
Expect(1, 126651, '\P{^Is_Word_Break=- aletter}', "");
Expect(0, 126652, '\p{Is_Word_Break=- aletter}', "");
Expect(1, 126652, '\p{^Is_Word_Break=- aletter}', "");
Expect(1, 126652, '\P{Is_Word_Break=- aletter}', "");
Expect(0, 126652, '\P{^Is_Word_Break=- aletter}', "");
Error('\p{Is_WB: /a/_LE}');
Error('\P{Is_WB: /a/_LE}');
Expect(1, 126651, '\p{Is_WB:   le}', "");
Expect(0, 126651, '\p{^Is_WB:   le}', "");
Expect(0, 126651, '\P{Is_WB:   le}', "");
Expect(1, 126651, '\P{^Is_WB:   le}', "");
Expect(0, 126652, '\p{Is_WB:   le}', "");
Expect(1, 126652, '\p{^Is_WB:   le}', "");
Expect(1, 126652, '\P{Is_WB:   le}', "");
Expect(0, 126652, '\P{^Is_WB:   le}', "");
Expect(1, 126651, '\p{Is_WB=_LE}', "");
Expect(0, 126651, '\p{^Is_WB=_LE}', "");
Expect(0, 126651, '\P{Is_WB=_LE}', "");
Expect(1, 126651, '\P{^Is_WB=_LE}', "");
Expect(0, 126652, '\p{Is_WB=_LE}', "");
Expect(1, 126652, '\p{^Is_WB=_LE}', "");
Expect(1, 126652, '\P{Is_WB=_LE}', "");
Expect(0, 126652, '\P{^Is_WB=_LE}', "");
Error('\p{Word_Break=__CR/a/}');
Error('\P{Word_Break=__CR/a/}');
Expect(1, 13, '\p{Word_Break=cr}', "");
Expect(0, 13, '\p{^Word_Break=cr}', "");
Expect(0, 13, '\P{Word_Break=cr}', "");
Expect(1, 13, '\P{^Word_Break=cr}', "");
Expect(0, 14, '\p{Word_Break=cr}', "");
Expect(1, 14, '\p{^Word_Break=cr}', "");
Expect(1, 14, '\P{Word_Break=cr}', "");
Expect(0, 14, '\P{^Word_Break=cr}', "");
Expect(1, 13, '\p{Word_Break: _CR}', "");
Expect(0, 13, '\p{^Word_Break: _CR}', "");
Expect(0, 13, '\P{Word_Break: _CR}', "");
Expect(1, 13, '\P{^Word_Break: _CR}', "");
Expect(0, 14, '\p{Word_Break: _CR}', "");
Expect(1, 14, '\p{^Word_Break: _CR}', "");
Expect(1, 14, '\P{Word_Break: _CR}', "");
Expect(0, 14, '\P{^Word_Break: _CR}', "");
Error('\p{WB=	:=CR}');
Error('\P{WB=	:=CR}');
Expect(1, 13, '\p{WB=cr}', "");
Expect(0, 13, '\p{^WB=cr}', "");
Expect(0, 13, '\P{WB=cr}', "");
Expect(1, 13, '\P{^WB=cr}', "");
Expect(0, 14, '\p{WB=cr}', "");
Expect(1, 14, '\p{^WB=cr}', "");
Expect(1, 14, '\P{WB=cr}', "");
Expect(0, 14, '\P{^WB=cr}', "");
Expect(1, 13, '\p{WB=	cr}', "");
Expect(0, 13, '\p{^WB=	cr}', "");
Expect(0, 13, '\P{WB=	cr}', "");
Expect(1, 13, '\P{^WB=	cr}', "");
Expect(0, 14, '\p{WB=	cr}', "");
Expect(1, 14, '\p{^WB=	cr}', "");
Expect(1, 14, '\P{WB=	cr}', "");
Expect(0, 14, '\P{^WB=	cr}', "");
Error('\p{Is_Word_Break=-/a/CR}');
Error('\P{Is_Word_Break=-/a/CR}');
Expect(1, 13, '\p{Is_Word_Break=cr}', "");
Expect(0, 13, '\p{^Is_Word_Break=cr}', "");
Expect(0, 13, '\P{Is_Word_Break=cr}', "");
Expect(1, 13, '\P{^Is_Word_Break=cr}', "");
Expect(0, 14, '\p{Is_Word_Break=cr}', "");
Expect(1, 14, '\p{^Is_Word_Break=cr}', "");
Expect(1, 14, '\P{Is_Word_Break=cr}', "");
Expect(0, 14, '\P{^Is_Word_Break=cr}', "");
Expect(1, 13, '\p{Is_Word_Break=	cr}', "");
Expect(0, 13, '\p{^Is_Word_Break=	cr}', "");
Expect(0, 13, '\P{Is_Word_Break=	cr}', "");
Expect(1, 13, '\P{^Is_Word_Break=	cr}', "");
Expect(0, 14, '\p{Is_Word_Break=	cr}', "");
Expect(1, 14, '\p{^Is_Word_Break=	cr}', "");
Expect(1, 14, '\P{Is_Word_Break=	cr}', "");
Expect(0, 14, '\P{^Is_Word_Break=	cr}', "");
Error('\p{Is_WB=_/a/CR}');
Error('\P{Is_WB=_/a/CR}');
Expect(1, 13, '\p{Is_WB:cr}', "");
Expect(0, 13, '\p{^Is_WB:cr}', "");
Expect(0, 13, '\P{Is_WB:cr}', "");
Expect(1, 13, '\P{^Is_WB:cr}', "");
Expect(0, 14, '\p{Is_WB:cr}', "");
Expect(1, 14, '\p{^Is_WB:cr}', "");
Expect(1, 14, '\P{Is_WB:cr}', "");
Expect(0, 14, '\P{^Is_WB:cr}', "");
Expect(1, 13, '\p{Is_WB=- CR}', "");
Expect(0, 13, '\p{^Is_WB=- CR}', "");
Expect(0, 13, '\P{Is_WB=- CR}', "");
Expect(1, 13, '\P{^Is_WB=- CR}', "");
Expect(0, 14, '\p{Is_WB=- CR}', "");
Expect(1, 14, '\p{^Is_WB=- CR}', "");
Expect(1, 14, '\P{Is_WB=- CR}', "");
Expect(0, 14, '\P{^Is_WB=- CR}', "");
Error('\p{Word_Break=	 other/a/}');
Error('\P{Word_Break=	 other/a/}');
Expect(1, 918000, '\p{Word_Break=other}', "");
Expect(0, 918000, '\p{^Word_Break=other}', "");
Expect(0, 918000, '\P{Word_Break=other}', "");
Expect(1, 918000, '\P{^Word_Break=other}', "");
Expect(0, 917999, '\p{Word_Break=other}', "");
Expect(1, 917999, '\p{^Word_Break=other}', "");
Expect(1, 917999, '\P{Word_Break=other}', "");
Expect(0, 917999, '\P{^Word_Break=other}', "");
Expect(1, 918000, '\p{Word_Break=-other}', "");
Expect(0, 918000, '\p{^Word_Break=-other}', "");
Expect(0, 918000, '\P{Word_Break=-other}', "");
Expect(1, 918000, '\P{^Word_Break=-other}', "");
Expect(0, 917999, '\p{Word_Break=-other}', "");
Expect(1, 917999, '\p{^Word_Break=-other}', "");
Expect(1, 917999, '\P{Word_Break=-other}', "");
Expect(0, 917999, '\P{^Word_Break=-other}', "");
Error('\p{WB=_XX/a/}');
Error('\P{WB=_XX/a/}');
Expect(1, 918000, '\p{WB=xx}', "");
Expect(0, 918000, '\p{^WB=xx}', "");
Expect(0, 918000, '\P{WB=xx}', "");
Expect(1, 918000, '\P{^WB=xx}', "");
Expect(0, 917999, '\p{WB=xx}', "");
Expect(1, 917999, '\p{^WB=xx}', "");
Expect(1, 917999, '\P{WB=xx}', "");
Expect(0, 917999, '\P{^WB=xx}', "");
Expect(1, 918000, '\p{WB=_XX}', "");
Expect(0, 918000, '\p{^WB=_XX}', "");
Expect(0, 918000, '\P{WB=_XX}', "");
Expect(1, 918000, '\P{^WB=_XX}', "");
Expect(0, 917999, '\p{WB=_XX}', "");
Expect(1, 917999, '\p{^WB=_XX}', "");
Expect(1, 917999, '\P{WB=_XX}', "");
Expect(0, 917999, '\P{^WB=_XX}', "");
Error('\p{Is_Word_Break=:=_-Other}');
Error('\P{Is_Word_Break=:=_-Other}');
Expect(1, 918000, '\p{Is_Word_Break=other}', "");
Expect(0, 918000, '\p{^Is_Word_Break=other}', "");
Expect(0, 918000, '\P{Is_Word_Break=other}', "");
Expect(1, 918000, '\P{^Is_Word_Break=other}', "");
Expect(0, 917999, '\p{Is_Word_Break=other}', "");
Expect(1, 917999, '\p{^Is_Word_Break=other}', "");
Expect(1, 917999, '\P{Is_Word_Break=other}', "");
Expect(0, 917999, '\P{^Is_Word_Break=other}', "");
Expect(1, 918000, '\p{Is_Word_Break=- Other}', "");
Expect(0, 918000, '\p{^Is_Word_Break=- Other}', "");
Expect(0, 918000, '\P{Is_Word_Break=- Other}', "");
Expect(1, 918000, '\P{^Is_Word_Break=- Other}', "");
Expect(0, 917999, '\p{Is_Word_Break=- Other}', "");
Expect(1, 917999, '\p{^Is_Word_Break=- Other}', "");
Expect(1, 917999, '\P{Is_Word_Break=- Other}', "");
Expect(0, 917999, '\P{^Is_Word_Break=- Other}', "");
Error('\p{Is_WB=-/a/xx}');
Error('\P{Is_WB=-/a/xx}');
Expect(1, 918000, '\p{Is_WB=xx}', "");
Expect(0, 918000, '\p{^Is_WB=xx}', "");
Expect(0, 918000, '\P{Is_WB=xx}', "");
Expect(1, 918000, '\P{^Is_WB=xx}', "");
Expect(0, 917999, '\p{Is_WB=xx}', "");
Expect(1, 917999, '\p{^Is_WB=xx}', "");
Expect(1, 917999, '\P{Is_WB=xx}', "");
Expect(0, 917999, '\P{^Is_WB=xx}', "");
Expect(1, 918000, '\p{Is_WB=-XX}', "");
Expect(0, 918000, '\p{^Is_WB=-XX}', "");
Expect(0, 918000, '\P{Is_WB=-XX}', "");
Expect(1, 918000, '\P{^Is_WB=-XX}', "");
Expect(0, 917999, '\p{Is_WB=-XX}', "");
Expect(1, 917999, '\p{^Is_WB=-XX}', "");
Expect(1, 917999, '\P{Is_WB=-XX}', "");
Expect(0, 917999, '\P{^Is_WB=-XX}', "");
Error('\p{Word_Break=/a/__Numeric}');
Error('\P{Word_Break=/a/__Numeric}');
Expect(1, 120831, '\p{Word_Break=numeric}', "");
Expect(0, 120831, '\p{^Word_Break=numeric}', "");
Expect(0, 120831, '\P{Word_Break=numeric}', "");
Expect(1, 120831, '\P{^Word_Break=numeric}', "");
Expect(0, 120832, '\p{Word_Break=numeric}', "");
Expect(1, 120832, '\p{^Word_Break=numeric}', "");
Expect(1, 120832, '\P{Word_Break=numeric}', "");
Expect(0, 120832, '\P{^Word_Break=numeric}', "");
Expect(1, 120831, '\p{Word_Break: 	numeric}', "");
Expect(0, 120831, '\p{^Word_Break: 	numeric}', "");
Expect(0, 120831, '\P{Word_Break: 	numeric}', "");
Expect(1, 120831, '\P{^Word_Break: 	numeric}', "");
Expect(0, 120832, '\p{Word_Break: 	numeric}', "");
Expect(1, 120832, '\p{^Word_Break: 	numeric}', "");
Expect(1, 120832, '\P{Word_Break: 	numeric}', "");
Expect(0, 120832, '\P{^Word_Break: 	numeric}', "");
Error('\p{WB= /a/nu}');
Error('\P{WB= /a/nu}');
Expect(1, 120831, '\p{WB=nu}', "");
Expect(0, 120831, '\p{^WB=nu}', "");
Expect(0, 120831, '\P{WB=nu}', "");
Expect(1, 120831, '\P{^WB=nu}', "");
Expect(0, 120832, '\p{WB=nu}', "");
Expect(1, 120832, '\p{^WB=nu}', "");
Expect(1, 120832, '\P{WB=nu}', "");
Expect(0, 120832, '\P{^WB=nu}', "");
Expect(1, 120831, '\p{WB:   		NU}', "");
Expect(0, 120831, '\p{^WB:   		NU}', "");
Expect(0, 120831, '\P{WB:   		NU}', "");
Expect(1, 120831, '\P{^WB:   		NU}', "");
Expect(0, 120832, '\p{WB:   		NU}', "");
Expect(1, 120832, '\p{^WB:   		NU}', "");
Expect(1, 120832, '\P{WB:   		NU}', "");
Expect(0, 120832, '\P{^WB:   		NU}', "");
Error('\p{Is_Word_Break=/a/		Numeric}');
Error('\P{Is_Word_Break=/a/		Numeric}');
Expect(1, 120831, '\p{Is_Word_Break=numeric}', "");
Expect(0, 120831, '\p{^Is_Word_Break=numeric}', "");
Expect(0, 120831, '\P{Is_Word_Break=numeric}', "");
Expect(1, 120831, '\P{^Is_Word_Break=numeric}', "");
Expect(0, 120832, '\p{Is_Word_Break=numeric}', "");
Expect(1, 120832, '\p{^Is_Word_Break=numeric}', "");
Expect(1, 120832, '\P{Is_Word_Break=numeric}', "");
Expect(0, 120832, '\P{^Is_Word_Break=numeric}', "");
Expect(1, 120831, '\p{Is_Word_Break=-Numeric}', "");
Expect(0, 120831, '\p{^Is_Word_Break=-Numeric}', "");
Expect(0, 120831, '\P{Is_Word_Break=-Numeric}', "");
Expect(1, 120831, '\P{^Is_Word_Break=-Numeric}', "");
Expect(0, 120832, '\p{Is_Word_Break=-Numeric}', "");
Expect(1, 120832, '\p{^Is_Word_Break=-Numeric}', "");
Expect(1, 120832, '\P{Is_Word_Break=-Numeric}', "");
Expect(0, 120832, '\P{^Is_Word_Break=-Numeric}', "");
Error('\p{Is_WB=-NU/a/}');
Error('\P{Is_WB=-NU/a/}');
Expect(1, 120831, '\p{Is_WB=nu}', "");
Expect(0, 120831, '\p{^Is_WB=nu}', "");
Expect(0, 120831, '\P{Is_WB=nu}', "");
Expect(1, 120831, '\P{^Is_WB=nu}', "");
Expect(0, 120832, '\p{Is_WB=nu}', "");
Expect(1, 120832, '\p{^Is_WB=nu}', "");
Expect(1, 120832, '\P{Is_WB=nu}', "");
Expect(0, 120832, '\P{^Is_WB=nu}', "");
Expect(1, 120831, '\p{Is_WB=- NU}', "");
Expect(0, 120831, '\p{^Is_WB=- NU}', "");
Expect(0, 120831, '\P{Is_WB=- NU}', "");
Expect(1, 120831, '\P{^Is_WB=- NU}', "");
Expect(0, 120832, '\p{Is_WB=- NU}', "");
Expect(1, 120832, '\p{^Is_WB=- NU}', "");
Expect(1, 120832, '\P{Is_WB=- NU}', "");
Expect(0, 120832, '\P{^Is_WB=- NU}', "");
Error('\p{Word_Break:-/a/format}');
Error('\P{Word_Break:-/a/format}');
Expect(1, 917631, '\p{Word_Break=format}', "");
Expect(0, 917631, '\p{^Word_Break=format}', "");
Expect(0, 917631, '\P{Word_Break=format}', "");
Expect(1, 917631, '\P{^Word_Break=format}', "");
Expect(0, 917632, '\p{Word_Break=format}', "");
Expect(1, 917632, '\p{^Word_Break=format}', "");
Expect(1, 917632, '\P{Word_Break=format}', "");
Expect(0, 917632, '\P{^Word_Break=format}', "");
Expect(1, 917631, '\p{Word_Break: -format}', "");
Expect(0, 917631, '\p{^Word_Break: -format}', "");
Expect(0, 917631, '\P{Word_Break: -format}', "");
Expect(1, 917631, '\P{^Word_Break: -format}', "");
Expect(0, 917632, '\p{Word_Break: -format}', "");
Expect(1, 917632, '\p{^Word_Break: -format}', "");
Expect(1, 917632, '\P{Word_Break: -format}', "");
Expect(0, 917632, '\P{^Word_Break: -format}', "");
Error('\p{WB= _fo/a/}');
Error('\P{WB= _fo/a/}');
Expect(1, 917631, '\p{WB=fo}', "");
Expect(0, 917631, '\p{^WB=fo}', "");
Expect(0, 917631, '\P{WB=fo}', "");
Expect(1, 917631, '\P{^WB=fo}', "");
Expect(0, 917632, '\p{WB=fo}', "");
Expect(1, 917632, '\p{^WB=fo}', "");
Expect(1, 917632, '\P{WB=fo}', "");
Expect(0, 917632, '\P{^WB=fo}', "");
Expect(1, 917631, '\p{WB= FO}', "");
Expect(0, 917631, '\p{^WB= FO}', "");
Expect(0, 917631, '\P{WB= FO}', "");
Expect(1, 917631, '\P{^WB= FO}', "");
Expect(0, 917632, '\p{WB= FO}', "");
Expect(1, 917632, '\p{^WB= FO}', "");
Expect(1, 917632, '\P{WB= FO}', "");
Expect(0, 917632, '\P{^WB= FO}', "");
Error('\p{Is_Word_Break:   -_Format/a/}');
Error('\P{Is_Word_Break:   -_Format/a/}');
Expect(1, 917631, '\p{Is_Word_Break=format}', "");
Expect(0, 917631, '\p{^Is_Word_Break=format}', "");
Expect(0, 917631, '\P{Is_Word_Break=format}', "");
Expect(1, 917631, '\P{^Is_Word_Break=format}', "");
Expect(0, 917632, '\p{Is_Word_Break=format}', "");
Expect(1, 917632, '\p{^Is_Word_Break=format}', "");
Expect(1, 917632, '\P{Is_Word_Break=format}', "");
Expect(0, 917632, '\P{^Is_Word_Break=format}', "");
Expect(1, 917631, '\p{Is_Word_Break=	format}', "");
Expect(0, 917631, '\p{^Is_Word_Break=	format}', "");
Expect(0, 917631, '\P{Is_Word_Break=	format}', "");
Expect(1, 917631, '\P{^Is_Word_Break=	format}', "");
Expect(0, 917632, '\p{Is_Word_Break=	format}', "");
Expect(1, 917632, '\p{^Is_Word_Break=	format}', "");
Expect(1, 917632, '\P{Is_Word_Break=	format}', "");
Expect(0, 917632, '\P{^Is_Word_Break=	format}', "");
Error('\p{Is_WB=	FO:=}');
Error('\P{Is_WB=	FO:=}');
Expect(1, 917631, '\p{Is_WB=fo}', "");
Expect(0, 917631, '\p{^Is_WB=fo}', "");
Expect(0, 917631, '\P{Is_WB=fo}', "");
Expect(1, 917631, '\P{^Is_WB=fo}', "");
Expect(0, 917632, '\p{Is_WB=fo}', "");
Expect(1, 917632, '\p{^Is_WB=fo}', "");
Expect(1, 917632, '\P{Is_WB=fo}', "");
Expect(0, 917632, '\P{^Is_WB=fo}', "");
Expect(1, 917631, '\p{Is_WB=__FO}', "");
Expect(0, 917631, '\p{^Is_WB=__FO}', "");
Expect(0, 917631, '\P{Is_WB=__FO}', "");
Expect(1, 917631, '\P{^Is_WB=__FO}', "");
Expect(0, 917632, '\p{Is_WB=__FO}', "");
Expect(1, 917632, '\p{^Is_WB=__FO}', "");
Expect(1, 917632, '\P{Is_WB=__FO}', "");
Expect(0, 917632, '\P{^Is_WB=__FO}', "");
Error('\p{Word_Break=		extendnumlet:=}');
Error('\P{Word_Break=		extendnumlet:=}');
Expect(1, 65343, '\p{Word_Break=extendnumlet}', "");
Expect(0, 65343, '\p{^Word_Break=extendnumlet}', "");
Expect(0, 65343, '\P{Word_Break=extendnumlet}', "");
Expect(1, 65343, '\P{^Word_Break=extendnumlet}', "");
Expect(0, 65344, '\p{Word_Break=extendnumlet}', "");
Expect(1, 65344, '\p{^Word_Break=extendnumlet}', "");
Expect(1, 65344, '\P{Word_Break=extendnumlet}', "");
Expect(0, 65344, '\P{^Word_Break=extendnumlet}', "");
Expect(1, 65343, '\p{Word_Break=_-ExtendNumLet}', "");
Expect(0, 65343, '\p{^Word_Break=_-ExtendNumLet}', "");
Expect(0, 65343, '\P{Word_Break=_-ExtendNumLet}', "");
Expect(1, 65343, '\P{^Word_Break=_-ExtendNumLet}', "");
Expect(0, 65344, '\p{Word_Break=_-ExtendNumLet}', "");
Expect(1, 65344, '\p{^Word_Break=_-ExtendNumLet}', "");
Expect(1, 65344, '\P{Word_Break=_-ExtendNumLet}', "");
Expect(0, 65344, '\P{^Word_Break=_-ExtendNumLet}', "");
Error('\p{WB:	EX/a/}');
Error('\P{WB:	EX/a/}');
Expect(1, 65343, '\p{WB=ex}', "");
Expect(0, 65343, '\p{^WB=ex}', "");
Expect(0, 65343, '\P{WB=ex}', "");
Expect(1, 65343, '\P{^WB=ex}', "");
Expect(0, 65344, '\p{WB=ex}', "");
Expect(1, 65344, '\p{^WB=ex}', "");
Expect(1, 65344, '\P{WB=ex}', "");
Expect(0, 65344, '\P{^WB=ex}', "");
Expect(1, 65343, '\p{WB=	_EX}', "");
Expect(0, 65343, '\p{^WB=	_EX}', "");
Expect(0, 65343, '\P{WB=	_EX}', "");
Expect(1, 65343, '\P{^WB=	_EX}', "");
Expect(0, 65344, '\p{WB=	_EX}', "");
Expect(1, 65344, '\p{^WB=	_EX}', "");
Expect(1, 65344, '\P{WB=	_EX}', "");
Expect(0, 65344, '\P{^WB=	_EX}', "");
Error('\p{Is_Word_Break= _ExtendNumLet:=}');
Error('\P{Is_Word_Break= _ExtendNumLet:=}');
Expect(1, 65343, '\p{Is_Word_Break:	extendnumlet}', "");
Expect(0, 65343, '\p{^Is_Word_Break:	extendnumlet}', "");
Expect(0, 65343, '\P{Is_Word_Break:	extendnumlet}', "");
Expect(1, 65343, '\P{^Is_Word_Break:	extendnumlet}', "");
Expect(0, 65344, '\p{Is_Word_Break:	extendnumlet}', "");
Expect(1, 65344, '\p{^Is_Word_Break:	extendnumlet}', "");
Expect(1, 65344, '\P{Is_Word_Break:	extendnumlet}', "");
Expect(0, 65344, '\P{^Is_Word_Break:	extendnumlet}', "");
Expect(1, 65343, '\p{Is_Word_Break:     EXTENDNUMLET}', "");
Expect(0, 65343, '\p{^Is_Word_Break:     EXTENDNUMLET}', "");
Expect(0, 65343, '\P{Is_Word_Break:     EXTENDNUMLET}', "");
Expect(1, 65343, '\P{^Is_Word_Break:     EXTENDNUMLET}', "");
Expect(0, 65344, '\p{Is_Word_Break:     EXTENDNUMLET}', "");
Expect(1, 65344, '\p{^Is_Word_Break:     EXTENDNUMLET}', "");
Expect(1, 65344, '\P{Is_Word_Break:     EXTENDNUMLET}', "");
Expect(0, 65344, '\P{^Is_Word_Break:     EXTENDNUMLET}', "");
Error('\p{Is_WB=/a/	 ex}');
Error('\P{Is_WB=/a/	 ex}');
Expect(1, 65343, '\p{Is_WB=ex}', "");
Expect(0, 65343, '\p{^Is_WB=ex}', "");
Expect(0, 65343, '\P{Is_WB=ex}', "");
Expect(1, 65343, '\P{^Is_WB=ex}', "");
Expect(0, 65344, '\p{Is_WB=ex}', "");
Expect(1, 65344, '\p{^Is_WB=ex}', "");
Expect(1, 65344, '\P{Is_WB=ex}', "");
Expect(0, 65344, '\P{^Is_WB=ex}', "");
Expect(1, 65343, '\p{Is_WB=-_EX}', "");
Expect(0, 65343, '\p{^Is_WB=-_EX}', "");
Expect(0, 65343, '\P{Is_WB=-_EX}', "");
Expect(1, 65343, '\P{^Is_WB=-_EX}', "");
Expect(0, 65344, '\p{Is_WB=-_EX}', "");
Expect(1, 65344, '\p{^Is_WB=-_EX}', "");
Expect(1, 65344, '\P{Is_WB=-_EX}', "");
Expect(0, 65344, '\P{^Is_WB=-_EX}', "");
Error('\p{Word_Break::=-Extend}');
Error('\P{Word_Break::=-Extend}');
Expect(1, 917999, '\p{Word_Break=extend}', "");
Expect(0, 917999, '\p{^Word_Break=extend}', "");
Expect(0, 917999, '\P{Word_Break=extend}', "");
Expect(1, 917999, '\P{^Word_Break=extend}', "");
Expect(0, 918000, '\p{Word_Break=extend}', "");
Expect(1, 918000, '\p{^Word_Break=extend}', "");
Expect(1, 918000, '\P{Word_Break=extend}', "");
Expect(0, 918000, '\P{^Word_Break=extend}', "");
Expect(1, 917999, '\p{Word_Break:  _Extend}', "");
Expect(0, 917999, '\p{^Word_Break:  _Extend}', "");
Expect(0, 917999, '\P{Word_Break:  _Extend}', "");
Expect(1, 917999, '\P{^Word_Break:  _Extend}', "");
Expect(0, 918000, '\p{Word_Break:  _Extend}', "");
Expect(1, 918000, '\p{^Word_Break:  _Extend}', "");
Expect(1, 918000, '\P{Word_Break:  _Extend}', "");
Expect(0, 918000, '\P{^Word_Break:  _Extend}', "");
Error('\p{WB=		EXTEND/a/}');
Error('\P{WB=		EXTEND/a/}');
Expect(1, 917999, '\p{WB=extend}', "");
Expect(0, 917999, '\p{^WB=extend}', "");
Expect(0, 917999, '\P{WB=extend}', "");
Expect(1, 917999, '\P{^WB=extend}', "");
Expect(0, 918000, '\p{WB=extend}', "");
Expect(1, 918000, '\p{^WB=extend}', "");
Expect(1, 918000, '\P{WB=extend}', "");
Expect(0, 918000, '\P{^WB=extend}', "");
Expect(1, 917999, '\p{WB:	 -Extend}', "");
Expect(0, 917999, '\p{^WB:	 -Extend}', "");
Expect(0, 917999, '\P{WB:	 -Extend}', "");
Expect(1, 917999, '\P{^WB:	 -Extend}', "");
Expect(0, 918000, '\p{WB:	 -Extend}', "");
Expect(1, 918000, '\p{^WB:	 -Extend}', "");
Expect(1, 918000, '\P{WB:	 -Extend}', "");
Expect(0, 918000, '\P{^WB:	 -Extend}', "");
Error('\p{Is_Word_Break=	/a/Extend}');
Error('\P{Is_Word_Break=	/a/Extend}');
Expect(1, 917999, '\p{Is_Word_Break=extend}', "");
Expect(0, 917999, '\p{^Is_Word_Break=extend}', "");
Expect(0, 917999, '\P{Is_Word_Break=extend}', "");
Expect(1, 917999, '\P{^Is_Word_Break=extend}', "");
Expect(0, 918000, '\p{Is_Word_Break=extend}', "");
Expect(1, 918000, '\p{^Is_Word_Break=extend}', "");
Expect(1, 918000, '\P{Is_Word_Break=extend}', "");
Expect(0, 918000, '\P{^Is_Word_Break=extend}', "");
Expect(1, 917999, '\p{Is_Word_Break= _extend}', "");
Expect(0, 917999, '\p{^Is_Word_Break= _extend}', "");
Expect(0, 917999, '\P{Is_Word_Break= _extend}', "");
Expect(1, 917999, '\P{^Is_Word_Break= _extend}', "");
Expect(0, 918000, '\p{Is_Word_Break= _extend}', "");
Expect(1, 918000, '\p{^Is_Word_Break= _extend}', "");
Expect(1, 918000, '\P{Is_Word_Break= _extend}', "");
Expect(0, 918000, '\P{^Is_Word_Break= _extend}', "");
Error('\p{Is_WB=/a/-EXTEND}');
Error('\P{Is_WB=/a/-EXTEND}');
Expect(1, 917999, '\p{Is_WB=extend}', "");
Expect(0, 917999, '\p{^Is_WB=extend}', "");
Expect(0, 917999, '\P{Is_WB=extend}', "");
Expect(1, 917999, '\P{^Is_WB=extend}', "");
Expect(0, 918000, '\p{Is_WB=extend}', "");
Expect(1, 918000, '\p{^Is_WB=extend}', "");
Expect(1, 918000, '\P{Is_WB=extend}', "");
Expect(0, 918000, '\P{^Is_WB=extend}', "");
Expect(1, 917999, '\p{Is_WB=--Extend}', "");
Expect(0, 917999, '\p{^Is_WB=--Extend}', "");
Expect(0, 917999, '\P{Is_WB=--Extend}', "");
Expect(1, 917999, '\P{^Is_WB=--Extend}', "");
Expect(0, 918000, '\p{Is_WB=--Extend}', "");
Expect(1, 918000, '\p{^Is_WB=--Extend}', "");
Expect(1, 918000, '\P{Is_WB=--Extend}', "");
Expect(0, 918000, '\P{^Is_WB=--Extend}', "");
Error('\p{Word_Break=/a/  midletter}');
Error('\P{Word_Break=/a/  midletter}');
Expect(1, 65306, '\p{Word_Break=midletter}', "");
Expect(0, 65306, '\p{^Word_Break=midletter}', "");
Expect(0, 65306, '\P{Word_Break=midletter}', "");
Expect(1, 65306, '\P{^Word_Break=midletter}', "");
Expect(0, 65307, '\p{Word_Break=midletter}', "");
Expect(1, 65307, '\p{^Word_Break=midletter}', "");
Expect(1, 65307, '\P{Word_Break=midletter}', "");
Expect(0, 65307, '\P{^Word_Break=midletter}', "");
Expect(1, 65306, '\p{Word_Break=MidLetter}', "");
Expect(0, 65306, '\p{^Word_Break=MidLetter}', "");
Expect(0, 65306, '\P{Word_Break=MidLetter}', "");
Expect(1, 65306, '\P{^Word_Break=MidLetter}', "");
Expect(0, 65307, '\p{Word_Break=MidLetter}', "");
Expect(1, 65307, '\p{^Word_Break=MidLetter}', "");
Expect(1, 65307, '\P{Word_Break=MidLetter}', "");
Expect(0, 65307, '\P{^Word_Break=MidLetter}', "");
Error('\p{WB=:=_	ML}');
Error('\P{WB=:=_	ML}');
Expect(1, 65306, '\p{WB=ml}', "");
Expect(0, 65306, '\p{^WB=ml}', "");
Expect(0, 65306, '\P{WB=ml}', "");
Expect(1, 65306, '\P{^WB=ml}', "");
Expect(0, 65307, '\p{WB=ml}', "");
Expect(1, 65307, '\p{^WB=ml}', "");
Expect(1, 65307, '\P{WB=ml}', "");
Expect(0, 65307, '\P{^WB=ml}', "");
Expect(1, 65306, '\p{WB=	ML}', "");
Expect(0, 65306, '\p{^WB=	ML}', "");
Expect(0, 65306, '\P{WB=	ML}', "");
Expect(1, 65306, '\P{^WB=	ML}', "");
Expect(0, 65307, '\p{WB=	ML}', "");
Expect(1, 65307, '\p{^WB=	ML}', "");
Expect(1, 65307, '\P{WB=	ML}', "");
Expect(0, 65307, '\P{^WB=	ML}', "");
Error('\p{Is_Word_Break=/a/_ midletter}');
Error('\P{Is_Word_Break=/a/_ midletter}');
Expect(1, 65306, '\p{Is_Word_Break=midletter}', "");
Expect(0, 65306, '\p{^Is_Word_Break=midletter}', "");
Expect(0, 65306, '\P{Is_Word_Break=midletter}', "");
Expect(1, 65306, '\P{^Is_Word_Break=midletter}', "");
Expect(0, 65307, '\p{Is_Word_Break=midletter}', "");
Expect(1, 65307, '\p{^Is_Word_Break=midletter}', "");
Expect(1, 65307, '\P{Is_Word_Break=midletter}', "");
Expect(0, 65307, '\P{^Is_Word_Break=midletter}', "");
Expect(1, 65306, '\p{Is_Word_Break=MIDLETTER}', "");
Expect(0, 65306, '\p{^Is_Word_Break=MIDLETTER}', "");
Expect(0, 65306, '\P{Is_Word_Break=MIDLETTER}', "");
Expect(1, 65306, '\P{^Is_Word_Break=MIDLETTER}', "");
Expect(0, 65307, '\p{Is_Word_Break=MIDLETTER}', "");
Expect(1, 65307, '\p{^Is_Word_Break=MIDLETTER}', "");
Expect(1, 65307, '\P{Is_Word_Break=MIDLETTER}', "");
Expect(0, 65307, '\P{^Is_Word_Break=MIDLETTER}', "");
Error('\p{Is_WB:-/a/ML}');
Error('\P{Is_WB:-/a/ML}');
Expect(1, 65306, '\p{Is_WB=ml}', "");
Expect(0, 65306, '\p{^Is_WB=ml}', "");
Expect(0, 65306, '\P{Is_WB=ml}', "");
Expect(1, 65306, '\P{^Is_WB=ml}', "");
Expect(0, 65307, '\p{Is_WB=ml}', "");
Expect(1, 65307, '\p{^Is_WB=ml}', "");
Expect(1, 65307, '\P{Is_WB=ml}', "");
Expect(0, 65307, '\P{^Is_WB=ml}', "");
Expect(1, 65306, '\p{Is_WB=_ ml}', "");
Expect(0, 65306, '\p{^Is_WB=_ ml}', "");
Expect(0, 65306, '\P{Is_WB=_ ml}', "");
Expect(1, 65306, '\P{^Is_WB=_ ml}', "");
Expect(0, 65307, '\p{Is_WB=_ ml}', "");
Expect(1, 65307, '\p{^Is_WB=_ ml}', "");
Expect(1, 65307, '\P{Is_WB=_ ml}', "");
Expect(0, 65307, '\P{^Is_WB=_ ml}', "");
Error('\p{Word_Break= /a/MIDNUMLET}');
Error('\P{Word_Break= /a/MIDNUMLET}');
Expect(1, 65294, '\p{Word_Break=midnumlet}', "");
Expect(0, 65294, '\p{^Word_Break=midnumlet}', "");
Expect(0, 65294, '\P{Word_Break=midnumlet}', "");
Expect(1, 65294, '\P{^Word_Break=midnumlet}', "");
Expect(0, 65295, '\p{Word_Break=midnumlet}', "");
Expect(1, 65295, '\p{^Word_Break=midnumlet}', "");
Expect(1, 65295, '\P{Word_Break=midnumlet}', "");
Expect(0, 65295, '\P{^Word_Break=midnumlet}', "");
Expect(1, 65294, '\p{Word_Break=--midnumlet}', "");
Expect(0, 65294, '\p{^Word_Break=--midnumlet}', "");
Expect(0, 65294, '\P{Word_Break=--midnumlet}', "");
Expect(1, 65294, '\P{^Word_Break=--midnumlet}', "");
Expect(0, 65295, '\p{Word_Break=--midnumlet}', "");
Expect(1, 65295, '\p{^Word_Break=--midnumlet}', "");
Expect(1, 65295, '\P{Word_Break=--midnumlet}', "");
Expect(0, 65295, '\P{^Word_Break=--midnumlet}', "");
Error('\p{WB=:=-	MB}');
Error('\P{WB=:=-	MB}');
Expect(1, 65294, '\p{WB=mb}', "");
Expect(0, 65294, '\p{^WB=mb}', "");
Expect(0, 65294, '\P{WB=mb}', "");
Expect(1, 65294, '\P{^WB=mb}', "");
Expect(0, 65295, '\p{WB=mb}', "");
Expect(1, 65295, '\p{^WB=mb}', "");
Expect(1, 65295, '\P{WB=mb}', "");
Expect(0, 65295, '\P{^WB=mb}', "");
Expect(1, 65294, '\p{WB=MB}', "");
Expect(0, 65294, '\p{^WB=MB}', "");
Expect(0, 65294, '\P{WB=MB}', "");
Expect(1, 65294, '\P{^WB=MB}', "");
Expect(0, 65295, '\p{WB=MB}', "");
Expect(1, 65295, '\p{^WB=MB}', "");
Expect(1, 65295, '\P{WB=MB}', "");
Expect(0, 65295, '\P{^WB=MB}', "");
Error('\p{Is_Word_Break:   /a/_midnumlet}');
Error('\P{Is_Word_Break:   /a/_midnumlet}');
Expect(1, 65294, '\p{Is_Word_Break=midnumlet}', "");
Expect(0, 65294, '\p{^Is_Word_Break=midnumlet}', "");
Expect(0, 65294, '\P{Is_Word_Break=midnumlet}', "");
Expect(1, 65294, '\P{^Is_Word_Break=midnumlet}', "");
Expect(0, 65295, '\p{Is_Word_Break=midnumlet}', "");
Expect(1, 65295, '\p{^Is_Word_Break=midnumlet}', "");
Expect(1, 65295, '\P{Is_Word_Break=midnumlet}', "");
Expect(0, 65295, '\P{^Is_Word_Break=midnumlet}', "");
Expect(1, 65294, '\p{Is_Word_Break= 	MidNumLet}', "");
Expect(0, 65294, '\p{^Is_Word_Break= 	MidNumLet}', "");
Expect(0, 65294, '\P{Is_Word_Break= 	MidNumLet}', "");
Expect(1, 65294, '\P{^Is_Word_Break= 	MidNumLet}', "");
Expect(0, 65295, '\p{Is_Word_Break= 	MidNumLet}', "");
Expect(1, 65295, '\p{^Is_Word_Break= 	MidNumLet}', "");
Expect(1, 65295, '\P{Is_Word_Break= 	MidNumLet}', "");
Expect(0, 65295, '\P{^Is_Word_Break= 	MidNumLet}', "");
Error('\p{Is_WB=_:=MB}');
Error('\P{Is_WB=_:=MB}');
Expect(1, 65294, '\p{Is_WB:	mb}', "");
Expect(0, 65294, '\p{^Is_WB:	mb}', "");
Expect(0, 65294, '\P{Is_WB:	mb}', "");
Expect(1, 65294, '\P{^Is_WB:	mb}', "");
Expect(0, 65295, '\p{Is_WB:	mb}', "");
Expect(1, 65295, '\p{^Is_WB:	mb}', "");
Expect(1, 65295, '\P{Is_WB:	mb}', "");
Expect(0, 65295, '\P{^Is_WB:	mb}', "");
Expect(1, 65294, '\p{Is_WB=_MB}', "");
Expect(0, 65294, '\p{^Is_WB=_MB}', "");
Expect(0, 65294, '\P{Is_WB=_MB}', "");
Expect(1, 65294, '\P{^Is_WB=_MB}', "");
Expect(0, 65295, '\p{Is_WB=_MB}', "");
Expect(1, 65295, '\p{^Is_WB=_MB}', "");
Expect(1, 65295, '\P{Is_WB=_MB}', "");
Expect(0, 65295, '\P{^Is_WB=_MB}', "");
Error('\p{Word_Break=/a/_MIDNUM}');
Error('\P{Word_Break=/a/_MIDNUM}');
Expect(1, 65307, '\p{Word_Break=midnum}', "");
Expect(0, 65307, '\p{^Word_Break=midnum}', "");
Expect(0, 65307, '\P{Word_Break=midnum}', "");
Expect(1, 65307, '\P{^Word_Break=midnum}', "");
Expect(0, 65308, '\p{Word_Break=midnum}', "");
Expect(1, 65308, '\p{^Word_Break=midnum}', "");
Expect(1, 65308, '\P{Word_Break=midnum}', "");
Expect(0, 65308, '\P{^Word_Break=midnum}', "");
Error('\p{WB= :=MN}');
Error('\P{WB= :=MN}');
Expect(1, 65307, '\p{WB=mn}', "");
Expect(0, 65307, '\p{^WB=mn}', "");
Expect(0, 65307, '\P{WB=mn}', "");
Expect(1, 65307, '\P{^WB=mn}', "");
Expect(0, 65308, '\p{WB=mn}', "");
Expect(1, 65308, '\p{^WB=mn}', "");
Expect(1, 65308, '\P{WB=mn}', "");
Expect(0, 65308, '\P{^WB=mn}', "");
Expect(1, 65307, '\p{WB=_ MN}', "");
Expect(0, 65307, '\p{^WB=_ MN}', "");
Expect(0, 65307, '\P{WB=_ MN}', "");
Expect(1, 65307, '\P{^WB=_ MN}', "");
Expect(0, 65308, '\p{WB=_ MN}', "");
Expect(1, 65308, '\p{^WB=_ MN}', "");
Expect(1, 65308, '\P{WB=_ MN}', "");
Expect(0, 65308, '\P{^WB=_ MN}', "");
Error('\p{Is_Word_Break=/a/	 MIDNUM}');
Error('\P{Is_Word_Break=/a/	 MIDNUM}');
Expect(1, 65307, '\p{Is_Word_Break=midnum}', "");
Expect(0, 65307, '\p{^Is_Word_Break=midnum}', "");
Expect(0, 65307, '\P{Is_Word_Break=midnum}', "");
Expect(1, 65307, '\P{^Is_Word_Break=midnum}', "");
Expect(0, 65308, '\p{Is_Word_Break=midnum}', "");
Expect(1, 65308, '\p{^Is_Word_Break=midnum}', "");
Expect(1, 65308, '\P{Is_Word_Break=midnum}', "");
Expect(0, 65308, '\P{^Is_Word_Break=midnum}', "");
Expect(1, 65307, '\p{Is_Word_Break=--MIDNUM}', "");
Expect(0, 65307, '\p{^Is_Word_Break=--MIDNUM}', "");
Expect(0, 65307, '\P{Is_Word_Break=--MIDNUM}', "");
Expect(1, 65307, '\P{^Is_Word_Break=--MIDNUM}', "");
Expect(0, 65308, '\p{Is_Word_Break=--MIDNUM}', "");
Expect(1, 65308, '\p{^Is_Word_Break=--MIDNUM}', "");
Expect(1, 65308, '\P{Is_Word_Break=--MIDNUM}', "");
Expect(0, 65308, '\P{^Is_Word_Break=--MIDNUM}', "");
Error('\p{Is_WB=:=_ MN}');
Error('\P{Is_WB=:=_ MN}');
Expect(1, 65307, '\p{Is_WB=mn}', "");
Expect(0, 65307, '\p{^Is_WB=mn}', "");
Expect(0, 65307, '\P{Is_WB=mn}', "");
Expect(1, 65307, '\P{^Is_WB=mn}', "");
Expect(0, 65308, '\p{Is_WB=mn}', "");
Expect(1, 65308, '\p{^Is_WB=mn}', "");
Expect(1, 65308, '\P{Is_WB=mn}', "");
Expect(0, 65308, '\P{^Is_WB=mn}', "");
Expect(1, 65307, '\p{Is_WB:	 _mn}', "");
Expect(0, 65307, '\p{^Is_WB:	 _mn}', "");
Expect(0, 65307, '\P{Is_WB:	 _mn}', "");
Expect(1, 65307, '\P{^Is_WB:	 _mn}', "");
Expect(0, 65308, '\p{Is_WB:	 _mn}', "");
Expect(1, 65308, '\p{^Is_WB:	 _mn}', "");
Expect(1, 65308, '\P{Is_WB:	 _mn}', "");
Expect(0, 65308, '\P{^Is_WB:	 _mn}', "");
Error('\p{Word_Break=:=-Katakana}');
Error('\P{Word_Break=:=-Katakana}');
Expect(1, 110592, '\p{Word_Break=katakana}', "");
Expect(0, 110592, '\p{^Word_Break=katakana}', "");
Expect(0, 110592, '\P{Word_Break=katakana}', "");
Expect(1, 110592, '\P{^Word_Break=katakana}', "");
Expect(0, 110593, '\p{Word_Break=katakana}', "");
Expect(1, 110593, '\p{^Word_Break=katakana}', "");
Expect(1, 110593, '\P{Word_Break=katakana}', "");
Expect(0, 110593, '\P{^Word_Break=katakana}', "");
Expect(1, 110592, '\p{Word_Break=	Katakana}', "");
Expect(0, 110592, '\p{^Word_Break=	Katakana}', "");
Expect(0, 110592, '\P{Word_Break=	Katakana}', "");
Expect(1, 110592, '\P{^Word_Break=	Katakana}', "");
Expect(0, 110593, '\p{Word_Break=	Katakana}', "");
Expect(1, 110593, '\p{^Word_Break=	Katakana}', "");
Expect(1, 110593, '\P{Word_Break=	Katakana}', "");
Expect(0, 110593, '\P{^Word_Break=	Katakana}', "");
Error('\p{WB=/a/__KA}');
Error('\P{WB=/a/__KA}');
Expect(1, 110592, '\p{WB:ka}', "");
Expect(0, 110592, '\p{^WB:ka}', "");
Expect(0, 110592, '\P{WB:ka}', "");
Expect(1, 110592, '\P{^WB:ka}', "");
Expect(0, 110593, '\p{WB:ka}', "");
Expect(1, 110593, '\p{^WB:ka}', "");
Expect(1, 110593, '\P{WB:ka}', "");
Expect(0, 110593, '\P{^WB:ka}', "");
Expect(1, 110592, '\p{WB=	-ka}', "");
Expect(0, 110592, '\p{^WB=	-ka}', "");
Expect(0, 110592, '\P{WB=	-ka}', "");
Expect(1, 110592, '\P{^WB=	-ka}', "");
Expect(0, 110593, '\p{WB=	-ka}', "");
Expect(1, 110593, '\p{^WB=	-ka}', "");
Expect(1, 110593, '\P{WB=	-ka}', "");
Expect(0, 110593, '\P{^WB=	-ka}', "");
Error('\p{Is_Word_Break: :=	katakana}');
Error('\P{Is_Word_Break: :=	katakana}');
Expect(1, 110592, '\p{Is_Word_Break: katakana}', "");
Expect(0, 110592, '\p{^Is_Word_Break: katakana}', "");
Expect(0, 110592, '\P{Is_Word_Break: katakana}', "");
Expect(1, 110592, '\P{^Is_Word_Break: katakana}', "");
Expect(0, 110593, '\p{Is_Word_Break: katakana}', "");
Expect(1, 110593, '\p{^Is_Word_Break: katakana}', "");
Expect(1, 110593, '\P{Is_Word_Break: katakana}', "");
Expect(0, 110593, '\P{^Is_Word_Break: katakana}', "");
Expect(1, 110592, '\p{Is_Word_Break=-Katakana}', "");
Expect(0, 110592, '\p{^Is_Word_Break=-Katakana}', "");
Expect(0, 110592, '\P{Is_Word_Break=-Katakana}', "");
Expect(1, 110592, '\P{^Is_Word_Break=-Katakana}', "");
Expect(0, 110593, '\p{Is_Word_Break=-Katakana}', "");
Expect(1, 110593, '\p{^Is_Word_Break=-Katakana}', "");
Expect(1, 110593, '\P{Is_Word_Break=-Katakana}', "");
Expect(0, 110593, '\P{^Is_Word_Break=-Katakana}', "");
Error('\p{Is_WB=/a/-_KA}');
Error('\P{Is_WB=/a/-_KA}');
Expect(1, 110592, '\p{Is_WB=ka}', "");
Expect(0, 110592, '\p{^Is_WB=ka}', "");
Expect(0, 110592, '\P{Is_WB=ka}', "");
Expect(1, 110592, '\P{^Is_WB=ka}', "");
Expect(0, 110593, '\p{Is_WB=ka}', "");
Expect(1, 110593, '\p{^Is_WB=ka}', "");
Expect(1, 110593, '\P{Is_WB=ka}', "");
Expect(0, 110593, '\P{^Is_WB=ka}', "");
Expect(1, 110592, '\p{Is_WB=	KA}', "");
Expect(0, 110592, '\p{^Is_WB=	KA}', "");
Expect(0, 110592, '\P{Is_WB=	KA}', "");
Expect(1, 110592, '\P{^Is_WB=	KA}', "");
Expect(0, 110593, '\p{Is_WB=	KA}', "");
Expect(1, 110593, '\p{^Is_WB=	KA}', "");
Expect(1, 110593, '\P{Is_WB=	KA}', "");
Expect(0, 110593, '\P{^Is_WB=	KA}', "");
Error('\p{Word_Break=/a/	LF}');
Error('\P{Word_Break=/a/	LF}');
Expect(1, 10, '\p{Word_Break=lf}', "");
Expect(0, 10, '\p{^Word_Break=lf}', "");
Expect(0, 10, '\P{Word_Break=lf}', "");
Expect(1, 10, '\P{^Word_Break=lf}', "");
Expect(0, 11, '\p{Word_Break=lf}', "");
Expect(1, 11, '\p{^Word_Break=lf}', "");
Expect(1, 11, '\P{Word_Break=lf}', "");
Expect(0, 11, '\P{^Word_Break=lf}', "");
Expect(1, 10, '\p{Word_Break=_	LF}', "");
Expect(0, 10, '\p{^Word_Break=_	LF}', "");
Expect(0, 10, '\P{Word_Break=_	LF}', "");
Expect(1, 10, '\P{^Word_Break=_	LF}', "");
Expect(0, 11, '\p{Word_Break=_	LF}', "");
Expect(1, 11, '\p{^Word_Break=_	LF}', "");
Expect(1, 11, '\P{Word_Break=_	LF}', "");
Expect(0, 11, '\P{^Word_Break=_	LF}', "");
Error('\p{WB=-_LF/a/}');
Error('\P{WB=-_LF/a/}');
Expect(1, 10, '\p{WB:   lf}', "");
Expect(0, 10, '\p{^WB:   lf}', "");
Expect(0, 10, '\P{WB:   lf}', "");
Expect(1, 10, '\P{^WB:   lf}', "");
Expect(0, 11, '\p{WB:   lf}', "");
Expect(1, 11, '\p{^WB:   lf}', "");
Expect(1, 11, '\P{WB:   lf}', "");
Expect(0, 11, '\P{^WB:   lf}', "");
Expect(1, 10, '\p{WB=-lf}', "");
Expect(0, 10, '\p{^WB=-lf}', "");
Expect(0, 10, '\P{WB=-lf}', "");
Expect(1, 10, '\P{^WB=-lf}', "");
Expect(0, 11, '\p{WB=-lf}', "");
Expect(1, 11, '\p{^WB=-lf}', "");
Expect(1, 11, '\P{WB=-lf}', "");
Expect(0, 11, '\P{^WB=-lf}', "");
Error('\p{Is_Word_Break: :=	_LF}');
Error('\P{Is_Word_Break: :=	_LF}');
Expect(1, 10, '\p{Is_Word_Break=lf}', "");
Expect(0, 10, '\p{^Is_Word_Break=lf}', "");
Expect(0, 10, '\P{Is_Word_Break=lf}', "");
Expect(1, 10, '\P{^Is_Word_Break=lf}', "");
Expect(0, 11, '\p{Is_Word_Break=lf}', "");
Expect(1, 11, '\p{^Is_Word_Break=lf}', "");
Expect(1, 11, '\P{Is_Word_Break=lf}', "");
Expect(0, 11, '\P{^Is_Word_Break=lf}', "");
Expect(1, 10, '\p{Is_Word_Break=- LF}', "");
Expect(0, 10, '\p{^Is_Word_Break=- LF}', "");
Expect(0, 10, '\P{Is_Word_Break=- LF}', "");
Expect(1, 10, '\P{^Is_Word_Break=- LF}', "");
Expect(0, 11, '\p{Is_Word_Break=- LF}', "");
Expect(1, 11, '\p{^Is_Word_Break=- LF}', "");
Expect(1, 11, '\P{Is_Word_Break=- LF}', "");
Expect(0, 11, '\P{^Is_Word_Break=- LF}', "");
Error('\p{Is_WB:/a/LF}');
Error('\P{Is_WB:/a/LF}');
Expect(1, 10, '\p{Is_WB=lf}', "");
Expect(0, 10, '\p{^Is_WB=lf}', "");
Expect(0, 10, '\P{Is_WB=lf}', "");
Expect(1, 10, '\P{^Is_WB=lf}', "");
Expect(0, 11, '\p{Is_WB=lf}', "");
Expect(1, 11, '\p{^Is_WB=lf}', "");
Expect(1, 11, '\P{Is_WB=lf}', "");
Expect(0, 11, '\P{^Is_WB=lf}', "");
Expect(1, 10, '\p{Is_WB= _LF}', "");
Expect(0, 10, '\p{^Is_WB= _LF}', "");
Expect(0, 10, '\P{Is_WB= _LF}', "");
Expect(1, 10, '\P{^Is_WB= _LF}', "");
Expect(0, 11, '\p{Is_WB= _LF}', "");
Expect(1, 11, '\p{^Is_WB= _LF}', "");
Expect(1, 11, '\P{Is_WB= _LF}', "");
Expect(0, 11, '\P{^Is_WB= _LF}', "");
Error('\p{Hangul_Syllable_Type=/a/Trailing_Jamo}');
Error('\P{Hangul_Syllable_Type=/a/Trailing_Jamo}');
Expect(1, 55291, '\p{Hangul_Syllable_Type=trailingjamo}', "");
Expect(0, 55291, '\p{^Hangul_Syllable_Type=trailingjamo}', "");
Expect(0, 55291, '\P{Hangul_Syllable_Type=trailingjamo}', "");
Expect(1, 55291, '\P{^Hangul_Syllable_Type=trailingjamo}', "");
Expect(0, 55292, '\p{Hangul_Syllable_Type=trailingjamo}', "");
Expect(1, 55292, '\p{^Hangul_Syllable_Type=trailingjamo}', "");
Expect(1, 55292, '\P{Hangul_Syllable_Type=trailingjamo}', "");
Expect(0, 55292, '\P{^Hangul_Syllable_Type=trailingjamo}', "");
Expect(1, 55291, '\p{Hangul_Syllable_Type=_-TRAILING_Jamo}', "");
Expect(0, 55291, '\p{^Hangul_Syllable_Type=_-TRAILING_Jamo}', "");
Expect(0, 55291, '\P{Hangul_Syllable_Type=_-TRAILING_Jamo}', "");
Expect(1, 55291, '\P{^Hangul_Syllable_Type=_-TRAILING_Jamo}', "");
Expect(0, 55292, '\p{Hangul_Syllable_Type=_-TRAILING_Jamo}', "");
Expect(1, 55292, '\p{^Hangul_Syllable_Type=_-TRAILING_Jamo}', "");
Expect(1, 55292, '\P{Hangul_Syllable_Type=_-TRAILING_Jamo}', "");
Expect(0, 55292, '\P{^Hangul_Syllable_Type=_-TRAILING_Jamo}', "");
Error('\p{Hst=- T/a/}');
Error('\P{Hst=- T/a/}');
Expect(1, 55291, '\p{Hst=t}', "");
Expect(0, 55291, '\p{^Hst=t}', "");
Expect(0, 55291, '\P{Hst=t}', "");
Expect(1, 55291, '\P{^Hst=t}', "");
Expect(0, 55292, '\p{Hst=t}', "");
Expect(1, 55292, '\p{^Hst=t}', "");
Expect(1, 55292, '\P{Hst=t}', "");
Expect(0, 55292, '\P{^Hst=t}', "");
Expect(1, 55291, '\p{Hst=_t}', "");
Expect(0, 55291, '\p{^Hst=_t}', "");
Expect(0, 55291, '\P{Hst=_t}', "");
Expect(1, 55291, '\P{^Hst=_t}', "");
Expect(0, 55292, '\p{Hst=_t}', "");
Expect(1, 55292, '\p{^Hst=_t}', "");
Expect(1, 55292, '\P{Hst=_t}', "");
Expect(0, 55292, '\P{^Hst=_t}', "");
Error('\p{Is_Hangul_Syllable_Type:   -/a/Trailing_Jamo}');
Error('\P{Is_Hangul_Syllable_Type:   -/a/Trailing_Jamo}');
Expect(1, 55291, '\p{Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(0, 55291, '\p{^Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(0, 55291, '\P{Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(1, 55291, '\P{^Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(0, 55292, '\p{Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(1, 55292, '\p{^Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(1, 55292, '\P{Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(0, 55292, '\P{^Is_Hangul_Syllable_Type=trailingjamo}', "");
Expect(1, 55291, '\p{Is_Hangul_Syllable_Type=-TRAILING_JAMO}', "");
Expect(0, 55291, '\p{^Is_Hangul_Syllable_Type=-TRAILING_JAMO}', "");
Expect(0, 55291, '\P{Is_Hangul_Syllable_Type=-TRAILING_JAMO}', "");
Expect(1, 55291, '\P{^Is_Hangul_Syllable_Type=-TRAILING_JAMO}', "");
Expect(0, 55292, '\p{Is_Hangul_Syllable_Type=-TRAILING_JAMO}', "");
Expect(1, 55292, '\p{^Is_Hangul_Syllable_Type=-TRAILING_JAMO}', "");
Expect(1, 55292, '\P{Is_Hangul_Syllable_Type=-TRAILING_JAMO}', "");
Expect(0, 55292, '\P{^Is_Hangul_Syllable_Type=-TRAILING_JAMO}', "");
Error('\p{Is_Hst=:=__T}');
Error('\P{Is_Hst=:=__T}');
Expect(1, 55291, '\p{Is_Hst=t}', "");
Expect(0, 55291, '\p{^Is_Hst=t}', "");
Expect(0, 55291, '\P{Is_Hst=t}', "");
Expect(1, 55291, '\P{^Is_Hst=t}', "");
Expect(0, 55292, '\p{Is_Hst=t}', "");
Expect(1, 55292, '\p{^Is_Hst=t}', "");
Expect(1, 55292, '\P{Is_Hst=t}', "");
Expect(0, 55292, '\P{^Is_Hst=t}', "");
Expect(1, 55291, '\p{Is_Hst=- t}', "");
Expect(0, 55291, '\p{^Is_Hst=- t}', "");
Expect(0, 55291, '\P{Is_Hst=- t}', "");
Expect(1, 55291, '\P{^Is_Hst=- t}', "");
Expect(0, 55292, '\p{Is_Hst=- t}', "");
Expect(1, 55292, '\p{^Is_Hst=- t}', "");
Expect(1, 55292, '\P{Is_Hst=- t}', "");
Expect(0, 55292, '\P{^Is_Hst=- t}', "");
Error('\p{Hangul_Syllable_Type:   -/a/LVT_syllable}');
Error('\P{Hangul_Syllable_Type:   -/a/LVT_syllable}');
Expect(1, 55203, '\p{Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55203, '\p{^Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55203, '\P{Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55203, '\P{^Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55204, '\p{Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55204, '\p{^Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55204, '\P{Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55204, '\P{^Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55203, '\p{Hangul_Syllable_Type= _LVT_Syllable}', "");
Expect(0, 55203, '\p{^Hangul_Syllable_Type= _LVT_Syllable}', "");
Expect(0, 55203, '\P{Hangul_Syllable_Type= _LVT_Syllable}', "");
Expect(1, 55203, '\P{^Hangul_Syllable_Type= _LVT_Syllable}', "");
Expect(0, 55204, '\p{Hangul_Syllable_Type= _LVT_Syllable}', "");
Expect(1, 55204, '\p{^Hangul_Syllable_Type= _LVT_Syllable}', "");
Expect(1, 55204, '\P{Hangul_Syllable_Type= _LVT_Syllable}', "");
Expect(0, 55204, '\P{^Hangul_Syllable_Type= _LVT_Syllable}', "");
Error('\p{Hst=/a/LVT}');
Error('\P{Hst=/a/LVT}');
Expect(1, 55203, '\p{Hst=lvt}', "");
Expect(0, 55203, '\p{^Hst=lvt}', "");
Expect(0, 55203, '\P{Hst=lvt}', "");
Expect(1, 55203, '\P{^Hst=lvt}', "");
Expect(0, 55204, '\p{Hst=lvt}', "");
Expect(1, 55204, '\p{^Hst=lvt}', "");
Expect(1, 55204, '\P{Hst=lvt}', "");
Expect(0, 55204, '\P{^Hst=lvt}', "");
Expect(1, 55203, '\p{Hst=	LVT}', "");
Expect(0, 55203, '\p{^Hst=	LVT}', "");
Expect(0, 55203, '\P{Hst=	LVT}', "");
Expect(1, 55203, '\P{^Hst=	LVT}', "");
Expect(0, 55204, '\p{Hst=	LVT}', "");
Expect(1, 55204, '\p{^Hst=	LVT}', "");
Expect(1, 55204, '\P{Hst=	LVT}', "");
Expect(0, 55204, '\P{^Hst=	LVT}', "");
Error('\p{Is_Hangul_Syllable_Type= :=lvt_Syllable}');
Error('\P{Is_Hangul_Syllable_Type= :=lvt_Syllable}');
Expect(1, 55203, '\p{Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55203, '\p{^Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55203, '\P{Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55203, '\P{^Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55204, '\p{Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55204, '\p{^Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55204, '\P{Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(0, 55204, '\P{^Is_Hangul_Syllable_Type=lvtsyllable}', "");
Expect(1, 55203, '\p{Is_Hangul_Syllable_Type=	lvt_Syllable}', "");
Expect(0, 55203, '\p{^Is_Hangul_Syllable_Type=	lvt_Syllable}', "");
Expect(0, 55203, '\P{Is_Hangul_Syllable_Type=	lvt_Syllable}', "");
Expect(1, 55203, '\P{^Is_Hangul_Syllable_Type=	lvt_Syllable}', "");
Expect(0, 55204, '\p{Is_Hangul_Syllable_Type=	lvt_Syllable}', "");
Expect(1, 55204, '\p{^Is_Hangul_Syllable_Type=	lvt_Syllable}', "");
Expect(1, 55204, '\P{Is_Hangul_Syllable_Type=	lvt_Syllable}', "");
Expect(0, 55204, '\P{^Is_Hangul_Syllable_Type=	lvt_Syllable}', "");
Error('\p{Is_Hst= lvt:=}');
Error('\P{Is_Hst= lvt:=}');
Expect(1, 55203, '\p{Is_Hst:   lvt}', "");
Expect(0, 55203, '\p{^Is_Hst:   lvt}', "");
Expect(0, 55203, '\P{Is_Hst:   lvt}', "");
Expect(1, 55203, '\P{^Is_Hst:   lvt}', "");
Expect(0, 55204, '\p{Is_Hst:   lvt}', "");
Expect(1, 55204, '\p{^Is_Hst:   lvt}', "");
Expect(1, 55204, '\P{Is_Hst:   lvt}', "");
Expect(0, 55204, '\P{^Is_Hst:   lvt}', "");
Expect(1, 55203, '\p{Is_Hst=_-LVT}', "");
Expect(0, 55203, '\p{^Is_Hst=_-LVT}', "");
Expect(0, 55203, '\P{Is_Hst=_-LVT}', "");
Expect(1, 55203, '\P{^Is_Hst=_-LVT}', "");
Expect(0, 55204, '\p{Is_Hst=_-LVT}', "");
Expect(1, 55204, '\p{^Is_Hst=_-LVT}', "");
Expect(1, 55204, '\P{Is_Hst=_-LVT}', "");
Expect(0, 55204, '\P{^Is_Hst=_-LVT}', "");
Error('\p{Hangul_Syllable_Type=_ Vowel_Jamo:=}');
Error('\P{Hangul_Syllable_Type=_ Vowel_Jamo:=}');
Expect(1, 55238, '\p{Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55238, '\p{^Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55238, '\P{Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55238, '\P{^Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55239, '\p{Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55239, '\p{^Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55239, '\P{Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55239, '\P{^Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55238, '\p{Hangul_Syllable_Type=_Vowel_Jamo}', "");
Expect(0, 55238, '\p{^Hangul_Syllable_Type=_Vowel_Jamo}', "");
Expect(0, 55238, '\P{Hangul_Syllable_Type=_Vowel_Jamo}', "");
Expect(1, 55238, '\P{^Hangul_Syllable_Type=_Vowel_Jamo}', "");
Expect(0, 55239, '\p{Hangul_Syllable_Type=_Vowel_Jamo}', "");
Expect(1, 55239, '\p{^Hangul_Syllable_Type=_Vowel_Jamo}', "");
Expect(1, 55239, '\P{Hangul_Syllable_Type=_Vowel_Jamo}', "");
Expect(0, 55239, '\P{^Hangul_Syllable_Type=_Vowel_Jamo}', "");
Error('\p{Hst=_V/a/}');
Error('\P{Hst=_V/a/}');
Expect(1, 55238, '\p{Hst=v}', "");
Expect(0, 55238, '\p{^Hst=v}', "");
Expect(0, 55238, '\P{Hst=v}', "");
Expect(1, 55238, '\P{^Hst=v}', "");
Expect(0, 55239, '\p{Hst=v}', "");
Expect(1, 55239, '\p{^Hst=v}', "");
Expect(1, 55239, '\P{Hst=v}', "");
Expect(0, 55239, '\P{^Hst=v}', "");
Expect(1, 55238, '\p{Hst= -v}', "");
Expect(0, 55238, '\p{^Hst= -v}', "");
Expect(0, 55238, '\P{Hst= -v}', "");
Expect(1, 55238, '\P{^Hst= -v}', "");
Expect(0, 55239, '\p{Hst= -v}', "");
Expect(1, 55239, '\p{^Hst= -v}', "");
Expect(1, 55239, '\P{Hst= -v}', "");
Expect(0, 55239, '\P{^Hst= -v}', "");
Error('\p{Is_Hangul_Syllable_Type=	/a/Vowel_JAMO}');
Error('\P{Is_Hangul_Syllable_Type=	/a/Vowel_JAMO}');
Expect(1, 55238, '\p{Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55238, '\p{^Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55238, '\P{Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55238, '\P{^Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55239, '\p{Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55239, '\p{^Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55239, '\P{Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(0, 55239, '\P{^Is_Hangul_Syllable_Type=voweljamo}', "");
Expect(1, 55238, '\p{Is_Hangul_Syllable_Type=- Vowel_Jamo}', "");
Expect(0, 55238, '\p{^Is_Hangul_Syllable_Type=- Vowel_Jamo}', "");
Expect(0, 55238, '\P{Is_Hangul_Syllable_Type=- Vowel_Jamo}', "");
Expect(1, 55238, '\P{^Is_Hangul_Syllable_Type=- Vowel_Jamo}', "");
Expect(0, 55239, '\p{Is_Hangul_Syllable_Type=- Vowel_Jamo}', "");
Expect(1, 55239, '\p{^Is_Hangul_Syllable_Type=- Vowel_Jamo}', "");
Expect(1, 55239, '\P{Is_Hangul_Syllable_Type=- Vowel_Jamo}', "");
Expect(0, 55239, '\P{^Is_Hangul_Syllable_Type=- Vowel_Jamo}', "");
Error('\p{Is_Hst= v/a/}');
Error('\P{Is_Hst= v/a/}');
Expect(1, 55238, '\p{Is_Hst=v}', "");
Expect(0, 55238, '\p{^Is_Hst=v}', "");
Expect(0, 55238, '\P{Is_Hst=v}', "");
Expect(1, 55238, '\P{^Is_Hst=v}', "");
Expect(0, 55239, '\p{Is_Hst=v}', "");
Expect(1, 55239, '\p{^Is_Hst=v}', "");
Expect(1, 55239, '\P{Is_Hst=v}', "");
Expect(0, 55239, '\P{^Is_Hst=v}', "");
Expect(1, 55238, '\p{Is_Hst=	 V}', "");
Expect(0, 55238, '\p{^Is_Hst=	 V}', "");
Expect(0, 55238, '\P{Is_Hst=	 V}', "");
Expect(1, 55238, '\P{^Is_Hst=	 V}', "");
Expect(0, 55239, '\p{Is_Hst=	 V}', "");
Expect(1, 55239, '\p{^Is_Hst=	 V}', "");
Expect(1, 55239, '\P{Is_Hst=	 V}', "");
Expect(0, 55239, '\P{^Is_Hst=	 V}', "");
Error('\p{Hangul_Syllable_Type= _LV_SYLLABLE:=}');
Error('\P{Hangul_Syllable_Type= _LV_SYLLABLE:=}');
Expect(1, 55176, '\p{Hangul_Syllable_Type=lvsyllable}', "");
Expect(0, 55176, '\p{^Hangul_Syllable_Type=lvsyllable}', "");
Expect(0, 55176, '\P{Hangul_Syllable_Type=lvsyllable}', "");
Expect(1, 55176, '\P{^Hangul_Syllable_Type=lvsyllable}', "");
Expect(0, 55177, '\p{Hangul_Syllable_Type=lvsyllable}', "");
Expect(1, 55177, '\p{^Hangul_Syllable_Type=lvsyllable}', "");
Expect(1, 55177, '\P{Hangul_Syllable_Type=lvsyllable}', "");
Expect(0, 55177, '\P{^Hangul_Syllable_Type=lvsyllable}', "");
Expect(1, 55176, '\p{Hangul_Syllable_Type=LV_Syllable}', "");
Expect(0, 55176, '\p{^Hangul_Syllable_Type=LV_Syllable}', "");
Expect(0, 55176, '\P{Hangul_Syllable_Type=LV_Syllable}', "");
Expect(1, 55176, '\P{^Hangul_Syllable_Type=LV_Syllable}', "");
Expect(0, 55177, '\p{Hangul_Syllable_Type=LV_Syllable}', "");
Expect(1, 55177, '\p{^Hangul_Syllable_Type=LV_Syllable}', "");
Expect(1, 55177, '\P{Hangul_Syllable_Type=LV_Syllable}', "");
Expect(0, 55177, '\P{^Hangul_Syllable_Type=LV_Syllable}', "");
Error('\p{Hst= 	lv/a/}');
Error('\P{Hst= 	lv/a/}');
Expect(1, 55176, '\p{Hst=lv}', "");
Expect(0, 55176, '\p{^Hst=lv}', "");
Expect(0, 55176, '\P{Hst=lv}', "");
Expect(1, 55176, '\P{^Hst=lv}', "");
Expect(0, 55177, '\p{Hst=lv}', "");
Expect(1, 55177, '\p{^Hst=lv}', "");
Expect(1, 55177, '\P{Hst=lv}', "");
Expect(0, 55177, '\P{^Hst=lv}', "");
Expect(1, 55176, '\p{Hst= -LV}', "");
Expect(0, 55176, '\p{^Hst= -LV}', "");
Expect(0, 55176, '\P{Hst= -LV}', "");
Expect(1, 55176, '\P{^Hst= -LV}', "");
Expect(0, 55177, '\p{Hst= -LV}', "");
Expect(1, 55177, '\p{^Hst= -LV}', "");
Expect(1, 55177, '\P{Hst= -LV}', "");
Expect(0, 55177, '\P{^Hst= -LV}', "");
Error('\p{Is_Hangul_Syllable_Type=:=	 LV_syllable}');
Error('\P{Is_Hangul_Syllable_Type=:=	 LV_syllable}');
Expect(1, 55176, '\p{Is_Hangul_Syllable_Type=lvsyllable}', "");
Expect(0, 55176, '\p{^Is_Hangul_Syllable_Type=lvsyllable}', "");
Expect(0, 55176, '\P{Is_Hangul_Syllable_Type=lvsyllable}', "");
Expect(1, 55176, '\P{^Is_Hangul_Syllable_Type=lvsyllable}', "");
Expect(0, 55177, '\p{Is_Hangul_Syllable_Type=lvsyllable}', "");
Expect(1, 55177, '\p{^Is_Hangul_Syllable_Type=lvsyllable}', "");
Expect(1, 55177, '\P{Is_Hangul_Syllable_Type=lvsyllable}', "");
Expect(0, 55177, '\P{^Is_Hangul_Syllable_Type=lvsyllable}', "");
Expect(1, 55176, '\p{Is_Hangul_Syllable_Type=	LV_Syllable}', "");
Expect(0, 55176, '\p{^Is_Hangul_Syllable_Type=	LV_Syllable}', "");
Expect(0, 55176, '\P{Is_Hangul_Syllable_Type=	LV_Syllable}', "");
Expect(1, 55176, '\P{^Is_Hangul_Syllable_Type=	LV_Syllable}', "");
Expect(0, 55177, '\p{Is_Hangul_Syllable_Type=	LV_Syllable}', "");
Expect(1, 55177, '\p{^Is_Hangul_Syllable_Type=	LV_Syllable}', "");
Expect(1, 55177, '\P{Is_Hangul_Syllable_Type=	LV_Syllable}', "");
Expect(0, 55177, '\P{^Is_Hangul_Syllable_Type=	LV_Syllable}', "");
Error('\p{Is_Hst=:=-lv}');
Error('\P{Is_Hst=:=-lv}');
Expect(1, 55176, '\p{Is_Hst=lv}', "");
Expect(0, 55176, '\p{^Is_Hst=lv}', "");
Expect(0, 55176, '\P{Is_Hst=lv}', "");
Expect(1, 55176, '\P{^Is_Hst=lv}', "");
Expect(0, 55177, '\p{Is_Hst=lv}', "");
Expect(1, 55177, '\p{^Is_Hst=lv}', "");
Expect(1, 55177, '\P{Is_Hst=lv}', "");
Expect(0, 55177, '\P{^Is_Hst=lv}', "");
Expect(1, 55176, '\p{Is_Hst=	_LV}', "");
Expect(0, 55176, '\p{^Is_Hst=	_LV}', "");
Expect(0, 55176, '\P{Is_Hst=	_LV}', "");
Expect(1, 55176, '\P{^Is_Hst=	_LV}', "");
Expect(0, 55177, '\p{Is_Hst=	_LV}', "");
Expect(1, 55177, '\p{^Is_Hst=	_LV}', "");
Expect(1, 55177, '\P{Is_Hst=	_LV}', "");
Expect(0, 55177, '\P{^Is_Hst=	_LV}', "");
Error('\p{Hangul_Syllable_Type=/a/	Not_Applicable}');
Error('\P{Hangul_Syllable_Type=/a/	Not_Applicable}');
Expect(1, 55292, '\p{Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55292, '\p{^Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55292, '\P{Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55292, '\P{^Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55291, '\p{Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55291, '\p{^Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55291, '\P{Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55291, '\P{^Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55292, '\p{Hangul_Syllable_Type: - Not_APPLICABLE}', "");
Expect(0, 55292, '\p{^Hangul_Syllable_Type: - Not_APPLICABLE}', "");
Expect(0, 55292, '\P{Hangul_Syllable_Type: - Not_APPLICABLE}', "");
Expect(1, 55292, '\P{^Hangul_Syllable_Type: - Not_APPLICABLE}', "");
Expect(0, 55291, '\p{Hangul_Syllable_Type: - Not_APPLICABLE}', "");
Expect(1, 55291, '\p{^Hangul_Syllable_Type: - Not_APPLICABLE}', "");
Expect(1, 55291, '\P{Hangul_Syllable_Type: - Not_APPLICABLE}', "");
Expect(0, 55291, '\P{^Hangul_Syllable_Type: - Not_APPLICABLE}', "");
Error('\p{Hst=_-NA/a/}');
Error('\P{Hst=_-NA/a/}');
Expect(1, 55292, '\p{Hst=na}', "");
Expect(0, 55292, '\p{^Hst=na}', "");
Expect(0, 55292, '\P{Hst=na}', "");
Expect(1, 55292, '\P{^Hst=na}', "");
Expect(0, 55291, '\p{Hst=na}', "");
Expect(1, 55291, '\p{^Hst=na}', "");
Expect(1, 55291, '\P{Hst=na}', "");
Expect(0, 55291, '\P{^Hst=na}', "");
Expect(1, 55292, '\p{Hst=-NA}', "");
Expect(0, 55292, '\p{^Hst=-NA}', "");
Expect(0, 55292, '\P{Hst=-NA}', "");
Expect(1, 55292, '\P{^Hst=-NA}', "");
Expect(0, 55291, '\p{Hst=-NA}', "");
Expect(1, 55291, '\p{^Hst=-NA}', "");
Expect(1, 55291, '\P{Hst=-NA}', "");
Expect(0, 55291, '\P{^Hst=-NA}', "");
Error('\p{Is_Hangul_Syllable_Type=:=-NOT_Applicable}');
Error('\P{Is_Hangul_Syllable_Type=:=-NOT_Applicable}');
Expect(1, 55292, '\p{Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55292, '\p{^Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55292, '\P{Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55292, '\P{^Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55291, '\p{Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55291, '\p{^Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55291, '\P{Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(0, 55291, '\P{^Is_Hangul_Syllable_Type=notapplicable}', "");
Expect(1, 55292, '\p{Is_Hangul_Syllable_Type= 	NOT_applicable}', "");
Expect(0, 55292, '\p{^Is_Hangul_Syllable_Type= 	NOT_applicable}', "");
Expect(0, 55292, '\P{Is_Hangul_Syllable_Type= 	NOT_applicable}', "");
Expect(1, 55292, '\P{^Is_Hangul_Syllable_Type= 	NOT_applicable}', "");
Expect(0, 55291, '\p{Is_Hangul_Syllable_Type= 	NOT_applicable}', "");
Expect(1, 55291, '\p{^Is_Hangul_Syllable_Type= 	NOT_applicable}', "");
Expect(1, 55291, '\P{Is_Hangul_Syllable_Type= 	NOT_applicable}', "");
Expect(0, 55291, '\P{^Is_Hangul_Syllable_Type= 	NOT_applicable}', "");
Error('\p{Is_Hst=	NA:=}');
Error('\P{Is_Hst=	NA:=}');
Expect(1, 55292, '\p{Is_Hst=na}', "");
Expect(0, 55292, '\p{^Is_Hst=na}', "");
Expect(0, 55292, '\P{Is_Hst=na}', "");
Expect(1, 55292, '\P{^Is_Hst=na}', "");
Expect(0, 55291, '\p{Is_Hst=na}', "");
Expect(1, 55291, '\p{^Is_Hst=na}', "");
Expect(1, 55291, '\P{Is_Hst=na}', "");
Expect(0, 55291, '\P{^Is_Hst=na}', "");
Expect(1, 55292, '\p{Is_Hst:	_	NA}', "");
Expect(0, 55292, '\p{^Is_Hst:	_	NA}', "");
Expect(0, 55292, '\P{Is_Hst:	_	NA}', "");
Expect(1, 55292, '\P{^Is_Hst:	_	NA}', "");
Expect(0, 55291, '\p{Is_Hst:	_	NA}', "");
Expect(1, 55291, '\p{^Is_Hst:	_	NA}', "");
Expect(1, 55291, '\P{Is_Hst:	_	NA}', "");
Expect(0, 55291, '\P{^Is_Hst:	_	NA}', "");
Error('\p{Hangul_Syllable_Type=/a/Leading_Jamo}');
Error('\P{Hangul_Syllable_Type=/a/Leading_Jamo}');
Expect(1, 43388, '\p{Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43388, '\p{^Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43388, '\P{Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43388, '\P{^Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43389, '\p{Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43389, '\p{^Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43389, '\P{Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43389, '\P{^Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43388, '\p{Hangul_Syllable_Type= 	leading_Jamo}', "");
Expect(0, 43388, '\p{^Hangul_Syllable_Type= 	leading_Jamo}', "");
Expect(0, 43388, '\P{Hangul_Syllable_Type= 	leading_Jamo}', "");
Expect(1, 43388, '\P{^Hangul_Syllable_Type= 	leading_Jamo}', "");
Expect(0, 43389, '\p{Hangul_Syllable_Type= 	leading_Jamo}', "");
Expect(1, 43389, '\p{^Hangul_Syllable_Type= 	leading_Jamo}', "");
Expect(1, 43389, '\P{Hangul_Syllable_Type= 	leading_Jamo}', "");
Expect(0, 43389, '\P{^Hangul_Syllable_Type= 	leading_Jamo}', "");
Error('\p{Hst= /a/L}');
Error('\P{Hst= /a/L}');
Expect(1, 43388, '\p{Hst=l}', "");
Expect(0, 43388, '\p{^Hst=l}', "");
Expect(0, 43388, '\P{Hst=l}', "");
Expect(1, 43388, '\P{^Hst=l}', "");
Expect(0, 43389, '\p{Hst=l}', "");
Expect(1, 43389, '\p{^Hst=l}', "");
Expect(1, 43389, '\P{Hst=l}', "");
Expect(0, 43389, '\P{^Hst=l}', "");
Expect(1, 43388, '\p{Hst=_L}', "");
Expect(0, 43388, '\p{^Hst=_L}', "");
Expect(0, 43388, '\P{Hst=_L}', "");
Expect(1, 43388, '\P{^Hst=_L}', "");
Expect(0, 43389, '\p{Hst=_L}', "");
Expect(1, 43389, '\p{^Hst=_L}', "");
Expect(1, 43389, '\P{Hst=_L}', "");
Expect(0, 43389, '\P{^Hst=_L}', "");
Error('\p{Is_Hangul_Syllable_Type=	:=leading_jamo}');
Error('\P{Is_Hangul_Syllable_Type=	:=leading_jamo}');
Expect(1, 43388, '\p{Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43388, '\p{^Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43388, '\P{Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43388, '\P{^Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43389, '\p{Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43389, '\p{^Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43389, '\P{Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(0, 43389, '\P{^Is_Hangul_Syllable_Type=leadingjamo}', "");
Expect(1, 43388, '\p{Is_Hangul_Syllable_Type=	Leading_Jamo}', "");
Expect(0, 43388, '\p{^Is_Hangul_Syllable_Type=	Leading_Jamo}', "");
Expect(0, 43388, '\P{Is_Hangul_Syllable_Type=	Leading_Jamo}', "");
Expect(1, 43388, '\P{^Is_Hangul_Syllable_Type=	Leading_Jamo}', "");
Expect(0, 43389, '\p{Is_Hangul_Syllable_Type=	Leading_Jamo}', "");
Expect(1, 43389, '\p{^Is_Hangul_Syllable_Type=	Leading_Jamo}', "");
Expect(1, 43389, '\P{Is_Hangul_Syllable_Type=	Leading_Jamo}', "");
Expect(0, 43389, '\P{^Is_Hangul_Syllable_Type=	Leading_Jamo}', "");
Error('\p{Is_Hst=:=_L}');
Error('\P{Is_Hst=:=_L}');
Expect(1, 43388, '\p{Is_Hst=l}', "");
Expect(0, 43388, '\p{^Is_Hst=l}', "");
Expect(0, 43388, '\P{Is_Hst=l}', "");
Expect(1, 43388, '\P{^Is_Hst=l}', "");
Expect(0, 43389, '\p{Is_Hst=l}', "");
Expect(1, 43389, '\p{^Is_Hst=l}', "");
Expect(1, 43389, '\P{Is_Hst=l}', "");
Expect(0, 43389, '\P{^Is_Hst=l}', "");
Expect(1, 43388, '\p{Is_Hst=-	L}', "");
Expect(0, 43388, '\p{^Is_Hst=-	L}', "");
Expect(0, 43388, '\P{Is_Hst=-	L}', "");
Expect(1, 43388, '\P{^Is_Hst=-	L}', "");
Expect(0, 43389, '\p{Is_Hst=-	L}', "");
Expect(1, 43389, '\p{^Is_Hst=-	L}', "");
Expect(1, 43389, '\P{Is_Hst=-	L}', "");
Expect(0, 43389, '\P{^Is_Hst=-	L}', "");
Error('\p{Changes_When_Titlecased= YES/a/}');
Error('\P{Changes_When_Titlecased= YES/a/}');
Expect(1, 66639, '\p{Changes_When_Titlecased=yes}', "");
Expect(0, 66639, '\p{^Changes_When_Titlecased=yes}', "");
Expect(0, 66639, '\P{Changes_When_Titlecased=yes}', "");
Expect(1, 66639, '\P{^Changes_When_Titlecased=yes}', "");
Expect(0, 66640, '\p{Changes_When_Titlecased=yes}', "");
Expect(1, 66640, '\p{^Changes_When_Titlecased=yes}', "");
Expect(1, 66640, '\P{Changes_When_Titlecased=yes}', "");
Expect(0, 66640, '\P{^Changes_When_Titlecased=yes}', "");
Expect(1, 66639, '\p{Changes_When_Titlecased: --YES}', "");
Expect(0, 66639, '\p{^Changes_When_Titlecased: --YES}', "");
Expect(0, 66639, '\P{Changes_When_Titlecased: --YES}', "");
Expect(1, 66639, '\P{^Changes_When_Titlecased: --YES}', "");
Expect(0, 66640, '\p{Changes_When_Titlecased: --YES}', "");
Expect(1, 66640, '\p{^Changes_When_Titlecased: --YES}', "");
Expect(1, 66640, '\P{Changes_When_Titlecased: --YES}', "");
Expect(0, 66640, '\P{^Changes_When_Titlecased: --YES}', "");
Error('\p{CWT=/a/	-y}');
Error('\P{CWT=/a/	-y}');
Expect(1, 66639, '\p{CWT=y}', "");
Expect(0, 66639, '\p{^CWT=y}', "");
Expect(0, 66639, '\P{CWT=y}', "");
Expect(1, 66639, '\P{^CWT=y}', "");
Expect(0, 66640, '\p{CWT=y}', "");
Expect(1, 66640, '\p{^CWT=y}', "");
Expect(1, 66640, '\P{CWT=y}', "");
Expect(0, 66640, '\P{^CWT=y}', "");
Expect(1, 66639, '\p{CWT=-Y}', "");
Expect(0, 66639, '\p{^CWT=-Y}', "");
Expect(0, 66639, '\P{CWT=-Y}', "");
Expect(1, 66639, '\P{^CWT=-Y}', "");
Expect(0, 66640, '\p{CWT=-Y}', "");
Expect(1, 66640, '\p{^CWT=-Y}', "");
Expect(1, 66640, '\P{CWT=-Y}', "");
Expect(0, 66640, '\P{^CWT=-Y}', "");
Error('\p{Is_Changes_When_Titlecased=/a/T}');
Error('\P{Is_Changes_When_Titlecased=/a/T}');
Expect(1, 66639, '\p{Is_Changes_When_Titlecased=t}', "");
Expect(0, 66639, '\p{^Is_Changes_When_Titlecased=t}', "");
Expect(0, 66639, '\P{Is_Changes_When_Titlecased=t}', "");
Expect(1, 66639, '\P{^Is_Changes_When_Titlecased=t}', "");
Expect(0, 66640, '\p{Is_Changes_When_Titlecased=t}', "");
Expect(1, 66640, '\p{^Is_Changes_When_Titlecased=t}', "");
Expect(1, 66640, '\P{Is_Changes_When_Titlecased=t}', "");
Expect(0, 66640, '\P{^Is_Changes_When_Titlecased=t}', "");
Expect(1, 66639, '\p{Is_Changes_When_Titlecased= _T}', "");
Expect(0, 66639, '\p{^Is_Changes_When_Titlecased= _T}', "");
Expect(0, 66639, '\P{Is_Changes_When_Titlecased= _T}', "");
Expect(1, 66639, '\P{^Is_Changes_When_Titlecased= _T}', "");
Expect(0, 66640, '\p{Is_Changes_When_Titlecased= _T}', "");
Expect(1, 66640, '\p{^Is_Changes_When_Titlecased= _T}', "");
Expect(1, 66640, '\P{Is_Changes_When_Titlecased= _T}', "");
Expect(0, 66640, '\P{^Is_Changes_When_Titlecased= _T}', "");
Error('\p{Is_CWT=-TRUE/a/}');
Error('\P{Is_CWT=-TRUE/a/}');
Expect(1, 66639, '\p{Is_CWT=true}', "");
Expect(0, 66639, '\p{^Is_CWT=true}', "");
Expect(0, 66639, '\P{Is_CWT=true}', "");
Expect(1, 66639, '\P{^Is_CWT=true}', "");
Expect(0, 66640, '\p{Is_CWT=true}', "");
Expect(1, 66640, '\p{^Is_CWT=true}', "");
Expect(1, 66640, '\P{Is_CWT=true}', "");
Expect(0, 66640, '\P{^Is_CWT=true}', "");
Expect(1, 66639, '\p{Is_CWT=	TRUE}', "");
Expect(0, 66639, '\p{^Is_CWT=	TRUE}', "");
Expect(0, 66639, '\P{Is_CWT=	TRUE}', "");
Expect(1, 66639, '\P{^Is_CWT=	TRUE}', "");
Expect(0, 66640, '\p{Is_CWT=	TRUE}', "");
Expect(1, 66640, '\p{^Is_CWT=	TRUE}', "");
Expect(1, 66640, '\P{Is_CWT=	TRUE}', "");
Expect(0, 66640, '\P{^Is_CWT=	TRUE}', "");
Error('\p{Changes_When_Titlecased=:=- No}');
Error('\P{Changes_When_Titlecased=:=- No}');
Expect(1, 66640, '\p{Changes_When_Titlecased: no}', "");
Expect(0, 66640, '\p{^Changes_When_Titlecased: no}', "");
Expect(0, 66640, '\P{Changes_When_Titlecased: no}', "");
Expect(1, 66640, '\P{^Changes_When_Titlecased: no}', "");
Expect(0, 66639, '\p{Changes_When_Titlecased: no}', "");
Expect(1, 66639, '\p{^Changes_When_Titlecased: no}', "");
Expect(1, 66639, '\P{Changes_When_Titlecased: no}', "");
Expect(0, 66639, '\P{^Changes_When_Titlecased: no}', "");
Expect(1, 66640, '\p{Changes_When_Titlecased=	 No}', "");
Expect(0, 66640, '\p{^Changes_When_Titlecased=	 No}', "");
Expect(0, 66640, '\P{Changes_When_Titlecased=	 No}', "");
Expect(1, 66640, '\P{^Changes_When_Titlecased=	 No}', "");
Expect(0, 66639, '\p{Changes_When_Titlecased=	 No}', "");
Expect(1, 66639, '\p{^Changes_When_Titlecased=	 No}', "");
Expect(1, 66639, '\P{Changes_When_Titlecased=	 No}', "");
Expect(0, 66639, '\P{^Changes_When_Titlecased=	 No}', "");
Error('\p{CWT=:=-n}');
Error('\P{CWT=:=-n}');
Expect(1, 66640, '\p{CWT=n}', "");
Expect(0, 66640, '\p{^CWT=n}', "");
Expect(0, 66640, '\P{CWT=n}', "");
Expect(1, 66640, '\P{^CWT=n}', "");
Expect(0, 66639, '\p{CWT=n}', "");
Expect(1, 66639, '\p{^CWT=n}', "");
Expect(1, 66639, '\P{CWT=n}', "");
Expect(0, 66639, '\P{^CWT=n}', "");
Expect(1, 66640, '\p{CWT=__N}', "");
Expect(0, 66640, '\p{^CWT=__N}', "");
Expect(0, 66640, '\P{CWT=__N}', "");
Expect(1, 66640, '\P{^CWT=__N}', "");
Expect(0, 66639, '\p{CWT=__N}', "");
Expect(1, 66639, '\p{^CWT=__N}', "");
Expect(1, 66639, '\P{CWT=__N}', "");
Expect(0, 66639, '\P{^CWT=__N}', "");
Error('\p{Is_Changes_When_Titlecased=/a/_	f}');
Error('\P{Is_Changes_When_Titlecased=/a/_	f}');
Expect(1, 66640, '\p{Is_Changes_When_Titlecased=f}', "");
Expect(0, 66640, '\p{^Is_Changes_When_Titlecased=f}', "");
Expect(0, 66640, '\P{Is_Changes_When_Titlecased=f}', "");
Expect(1, 66640, '\P{^Is_Changes_When_Titlecased=f}', "");
Expect(0, 66639, '\p{Is_Changes_When_Titlecased=f}', "");
Expect(1, 66639, '\p{^Is_Changes_When_Titlecased=f}', "");
Expect(1, 66639, '\P{Is_Changes_When_Titlecased=f}', "");
Expect(0, 66639, '\P{^Is_Changes_When_Titlecased=f}', "");
Expect(1, 66640, '\p{Is_Changes_When_Titlecased= F}', "");
Expect(0, 66640, '\p{^Is_Changes_When_Titlecased= F}', "");
Expect(0, 66640, '\P{Is_Changes_When_Titlecased= F}', "");
Expect(1, 66640, '\P{^Is_Changes_When_Titlecased= F}', "");
Expect(0, 66639, '\p{Is_Changes_When_Titlecased= F}', "");
Expect(1, 66639, '\p{^Is_Changes_When_Titlecased= F}', "");
Expect(1, 66639, '\P{Is_Changes_When_Titlecased= F}', "");
Expect(0, 66639, '\P{^Is_Changes_When_Titlecased= F}', "");
Error('\p{Is_CWT=_	false/a/}');
Error('\P{Is_CWT=_	false/a/}');
Expect(1, 66640, '\p{Is_CWT=false}', "");
Expect(0, 66640, '\p{^Is_CWT=false}', "");
Expect(0, 66640, '\P{Is_CWT=false}', "");
Expect(1, 66640, '\P{^Is_CWT=false}', "");
Expect(0, 66639, '\p{Is_CWT=false}', "");
Expect(1, 66639, '\p{^Is_CWT=false}', "");
Expect(1, 66639, '\P{Is_CWT=false}', "");
Expect(0, 66639, '\P{^Is_CWT=false}', "");
Expect(1, 66640, '\p{Is_CWT=-False}', "");
Expect(0, 66640, '\p{^Is_CWT=-False}', "");
Expect(0, 66640, '\P{Is_CWT=-False}', "");
Expect(1, 66640, '\P{^Is_CWT=-False}', "");
Expect(0, 66639, '\p{Is_CWT=-False}', "");
Expect(1, 66639, '\p{^Is_CWT=-False}', "");
Expect(1, 66639, '\P{Is_CWT=-False}', "");
Expect(0, 66639, '\P{^Is_CWT=-False}', "");
Error('\p{Numeric_Type=_/a/NONE}');
Error('\P{Numeric_Type=_/a/NONE}');
Expect(1, 194705, '\p{Numeric_Type=none}', "");
Expect(0, 194705, '\p{^Numeric_Type=none}', "");
Expect(0, 194705, '\P{Numeric_Type=none}', "");
Expect(1, 194705, '\P{^Numeric_Type=none}', "");
Expect(0, 194704, '\p{Numeric_Type=none}', "");
Expect(1, 194704, '\p{^Numeric_Type=none}', "");
Expect(1, 194704, '\P{Numeric_Type=none}', "");
Expect(0, 194704, '\P{^Numeric_Type=none}', "");
Expect(1, 194705, '\p{Numeric_Type= _None}', "");
Expect(0, 194705, '\p{^Numeric_Type= _None}', "");
Expect(0, 194705, '\P{Numeric_Type= _None}', "");
Expect(1, 194705, '\P{^Numeric_Type= _None}', "");
Expect(0, 194704, '\p{Numeric_Type= _None}', "");
Expect(1, 194704, '\p{^Numeric_Type= _None}', "");
Expect(1, 194704, '\P{Numeric_Type= _None}', "");
Expect(0, 194704, '\P{^Numeric_Type= _None}', "");
Error('\p{Nt=	none:=}');
Error('\P{Nt=	none:=}');
Expect(1, 194705, '\p{Nt=none}', "");
Expect(0, 194705, '\p{^Nt=none}', "");
Expect(0, 194705, '\P{Nt=none}', "");
Expect(1, 194705, '\P{^Nt=none}', "");
Expect(0, 194704, '\p{Nt=none}', "");
Expect(1, 194704, '\p{^Nt=none}', "");
Expect(1, 194704, '\P{Nt=none}', "");
Expect(0, 194704, '\P{^Nt=none}', "");
Expect(1, 194705, '\p{Nt=-NONE}', "");
Expect(0, 194705, '\p{^Nt=-NONE}', "");
Expect(0, 194705, '\P{Nt=-NONE}', "");
Expect(1, 194705, '\P{^Nt=-NONE}', "");
Expect(0, 194704, '\p{Nt=-NONE}', "");
Expect(1, 194704, '\p{^Nt=-NONE}', "");
Expect(1, 194704, '\P{Nt=-NONE}', "");
Expect(0, 194704, '\P{^Nt=-NONE}', "");
Error('\p{Is_Numeric_Type=/a/NONE}');
Error('\P{Is_Numeric_Type=/a/NONE}');
Expect(1, 194705, '\p{Is_Numeric_Type=none}', "");
Expect(0, 194705, '\p{^Is_Numeric_Type=none}', "");
Expect(0, 194705, '\P{Is_Numeric_Type=none}', "");
Expect(1, 194705, '\P{^Is_Numeric_Type=none}', "");
Expect(0, 194704, '\p{Is_Numeric_Type=none}', "");
Expect(1, 194704, '\p{^Is_Numeric_Type=none}', "");
Expect(1, 194704, '\P{Is_Numeric_Type=none}', "");
Expect(0, 194704, '\P{^Is_Numeric_Type=none}', "");
Expect(1, 194705, '\p{Is_Numeric_Type=  None}', "");
Expect(0, 194705, '\p{^Is_Numeric_Type=  None}', "");
Expect(0, 194705, '\P{Is_Numeric_Type=  None}', "");
Expect(1, 194705, '\P{^Is_Numeric_Type=  None}', "");
Expect(0, 194704, '\p{Is_Numeric_Type=  None}', "");
Expect(1, 194704, '\p{^Is_Numeric_Type=  None}', "");
Expect(1, 194704, '\P{Is_Numeric_Type=  None}', "");
Expect(0, 194704, '\P{^Is_Numeric_Type=  None}', "");
Error('\p{Is_Nt= /a/None}');
Error('\P{Is_Nt= /a/None}');
Expect(1, 194705, '\p{Is_Nt:   none}', "");
Expect(0, 194705, '\p{^Is_Nt:   none}', "");
Expect(0, 194705, '\P{Is_Nt:   none}', "");
Expect(1, 194705, '\P{^Is_Nt:   none}', "");
Expect(0, 194704, '\p{Is_Nt:   none}', "");
Expect(1, 194704, '\p{^Is_Nt:   none}', "");
Expect(1, 194704, '\P{Is_Nt:   none}', "");
Expect(0, 194704, '\P{^Is_Nt:   none}', "");
Expect(1, 194705, '\p{Is_Nt=_	None}', "");
Expect(0, 194705, '\p{^Is_Nt=_	None}', "");
Expect(0, 194705, '\P{Is_Nt=_	None}', "");
Expect(1, 194705, '\P{^Is_Nt=_	None}', "");
Expect(0, 194704, '\p{Is_Nt=_	None}', "");
Expect(1, 194704, '\p{^Is_Nt=_	None}', "");
Expect(1, 194704, '\P{Is_Nt=_	None}', "");
Expect(0, 194704, '\P{^Is_Nt=_	None}', "");
Error('\p{Numeric_Type=Digit/a/}');
Error('\P{Numeric_Type=Digit/a/}');
Expect(1, 127242, '\p{Numeric_Type:digit}', "");
Expect(0, 127242, '\p{^Numeric_Type:digit}', "");
Expect(0, 127242, '\P{Numeric_Type:digit}', "");
Expect(1, 127242, '\P{^Numeric_Type:digit}', "");
Expect(0, 127243, '\p{Numeric_Type:digit}', "");
Expect(1, 127243, '\p{^Numeric_Type:digit}', "");
Expect(1, 127243, '\P{Numeric_Type:digit}', "");
Expect(0, 127243, '\P{^Numeric_Type:digit}', "");
Expect(1, 127242, '\p{Numeric_Type=_Digit}', "");
Expect(0, 127242, '\p{^Numeric_Type=_Digit}', "");
Expect(0, 127242, '\P{Numeric_Type=_Digit}', "");
Expect(1, 127242, '\P{^Numeric_Type=_Digit}', "");
Expect(0, 127243, '\p{Numeric_Type=_Digit}', "");
Expect(1, 127243, '\p{^Numeric_Type=_Digit}', "");
Expect(1, 127243, '\P{Numeric_Type=_Digit}', "");
Expect(0, 127243, '\P{^Numeric_Type=_Digit}', "");
Error('\p{Nt=:= Di}');
Error('\P{Nt=:= Di}');
Expect(1, 127242, '\p{Nt=di}', "");
Expect(0, 127242, '\p{^Nt=di}', "");
Expect(0, 127242, '\P{Nt=di}', "");
Expect(1, 127242, '\P{^Nt=di}', "");
Expect(0, 127243, '\p{Nt=di}', "");
Expect(1, 127243, '\p{^Nt=di}', "");
Expect(1, 127243, '\P{Nt=di}', "");
Expect(0, 127243, '\P{^Nt=di}', "");
Expect(1, 127242, '\p{Nt= 	di}', "");
Expect(0, 127242, '\p{^Nt= 	di}', "");
Expect(0, 127242, '\P{Nt= 	di}', "");
Expect(1, 127242, '\P{^Nt= 	di}', "");
Expect(0, 127243, '\p{Nt= 	di}', "");
Expect(1, 127243, '\p{^Nt= 	di}', "");
Expect(1, 127243, '\P{Nt= 	di}', "");
Expect(0, 127243, '\P{^Nt= 	di}', "");
Error('\p{Is_Numeric_Type=/a/- DIGIT}');
Error('\P{Is_Numeric_Type=/a/- DIGIT}');
Expect(1, 127242, '\p{Is_Numeric_Type=digit}', "");
Expect(0, 127242, '\p{^Is_Numeric_Type=digit}', "");
Expect(0, 127242, '\P{Is_Numeric_Type=digit}', "");
Expect(1, 127242, '\P{^Is_Numeric_Type=digit}', "");
Expect(0, 127243, '\p{Is_Numeric_Type=digit}', "");
Expect(1, 127243, '\p{^Is_Numeric_Type=digit}', "");
Expect(1, 127243, '\P{Is_Numeric_Type=digit}', "");
Expect(0, 127243, '\P{^Is_Numeric_Type=digit}', "");
Expect(1, 127242, '\p{Is_Numeric_Type=_Digit}', "");
Expect(0, 127242, '\p{^Is_Numeric_Type=_Digit}', "");
Expect(0, 127242, '\P{Is_Numeric_Type=_Digit}', "");
Expect(1, 127242, '\P{^Is_Numeric_Type=_Digit}', "");
Expect(0, 127243, '\p{Is_Numeric_Type=_Digit}', "");
Expect(1, 127243, '\p{^Is_Numeric_Type=_Digit}', "");
Expect(1, 127243, '\P{Is_Numeric_Type=_Digit}', "");
Expect(0, 127243, '\P{^Is_Numeric_Type=_Digit}', "");
Error('\p{Is_Nt=  Di:=}');
Error('\P{Is_Nt=  Di:=}');
Expect(1, 127242, '\p{Is_Nt=di}', "");
Expect(0, 127242, '\p{^Is_Nt=di}', "");
Expect(0, 127242, '\P{Is_Nt=di}', "");
Expect(1, 127242, '\P{^Is_Nt=di}', "");
Expect(0, 127243, '\p{Is_Nt=di}', "");
Expect(1, 127243, '\p{^Is_Nt=di}', "");
Expect(1, 127243, '\P{Is_Nt=di}', "");
Expect(0, 127243, '\P{^Is_Nt=di}', "");
Expect(1, 127242, '\p{Is_Nt=	-DI}', "");
Expect(0, 127242, '\p{^Is_Nt=	-DI}', "");
Expect(0, 127242, '\P{Is_Nt=	-DI}', "");
Expect(1, 127242, '\P{^Is_Nt=	-DI}', "");
Expect(0, 127243, '\p{Is_Nt=	-DI}', "");
Expect(1, 127243, '\p{^Is_Nt=	-DI}', "");
Expect(1, 127243, '\P{Is_Nt=	-DI}', "");
Expect(0, 127243, '\P{^Is_Nt=	-DI}', "");
Error('\p{Numeric_Type=:=	_decimal}');
Error('\P{Numeric_Type=:=	_decimal}');
Expect(1, 120831, '\p{Numeric_Type=decimal}', "");
Expect(0, 120831, '\p{^Numeric_Type=decimal}', "");
Expect(0, 120831, '\P{Numeric_Type=decimal}', "");
Expect(1, 120831, '\P{^Numeric_Type=decimal}', "");
Expect(0, 120832, '\p{Numeric_Type=decimal}', "");
Expect(1, 120832, '\p{^Numeric_Type=decimal}', "");
Expect(1, 120832, '\P{Numeric_Type=decimal}', "");
Expect(0, 120832, '\P{^Numeric_Type=decimal}', "");
Expect(1, 120831, '\p{Numeric_Type= 	Decimal}', "");
Expect(0, 120831, '\p{^Numeric_Type= 	Decimal}', "");
Expect(0, 120831, '\P{Numeric_Type= 	Decimal}', "");
Expect(1, 120831, '\P{^Numeric_Type= 	Decimal}', "");
Expect(0, 120832, '\p{Numeric_Type= 	Decimal}', "");
Expect(1, 120832, '\p{^Numeric_Type= 	Decimal}', "");
Expect(1, 120832, '\P{Numeric_Type= 	Decimal}', "");
Expect(0, 120832, '\P{^Numeric_Type= 	Decimal}', "");
Error('\p{Nt=/a/ -De}');
Error('\P{Nt=/a/ -De}');
Expect(1, 120831, '\p{Nt=de}', "");
Expect(0, 120831, '\p{^Nt=de}', "");
Expect(0, 120831, '\P{Nt=de}', "");
Expect(1, 120831, '\P{^Nt=de}', "");
Expect(0, 120832, '\p{Nt=de}', "");
Expect(1, 120832, '\p{^Nt=de}', "");
Expect(1, 120832, '\P{Nt=de}', "");
Expect(0, 120832, '\P{^Nt=de}', "");
Expect(1, 120831, '\p{Nt=	De}', "");
Expect(0, 120831, '\p{^Nt=	De}', "");
Expect(0, 120831, '\P{Nt=	De}', "");
Expect(1, 120831, '\P{^Nt=	De}', "");
Expect(0, 120832, '\p{Nt=	De}', "");
Expect(1, 120832, '\p{^Nt=	De}', "");
Expect(1, 120832, '\P{Nt=	De}', "");
Expect(0, 120832, '\P{^Nt=	De}', "");
Error('\p{Is_Numeric_Type=	/a/DECIMAL}');
Error('\P{Is_Numeric_Type=	/a/DECIMAL}');
Expect(1, 120831, '\p{Is_Numeric_Type=decimal}', "");
Expect(0, 120831, '\p{^Is_Numeric_Type=decimal}', "");
Expect(0, 120831, '\P{Is_Numeric_Type=decimal}', "");
Expect(1, 120831, '\P{^Is_Numeric_Type=decimal}', "");
Expect(0, 120832, '\p{Is_Numeric_Type=decimal}', "");
Expect(1, 120832, '\p{^Is_Numeric_Type=decimal}', "");
Expect(1, 120832, '\P{Is_Numeric_Type=decimal}', "");
Expect(0, 120832, '\P{^Is_Numeric_Type=decimal}', "");
Expect(1, 120831, '\p{Is_Numeric_Type=_decimal}', "");
Expect(0, 120831, '\p{^Is_Numeric_Type=_decimal}', "");
Expect(0, 120831, '\P{Is_Numeric_Type=_decimal}', "");
Expect(1, 120831, '\P{^Is_Numeric_Type=_decimal}', "");
Expect(0, 120832, '\p{Is_Numeric_Type=_decimal}', "");
Expect(1, 120832, '\p{^Is_Numeric_Type=_decimal}', "");
Expect(1, 120832, '\P{Is_Numeric_Type=_decimal}', "");
Expect(0, 120832, '\P{^Is_Numeric_Type=_decimal}', "");
Error('\p{Is_Nt=:=-_DE}');
Error('\P{Is_Nt=:=-_DE}');
Expect(1, 120831, '\p{Is_Nt:   de}', "");
Expect(0, 120831, '\p{^Is_Nt:   de}', "");
Expect(0, 120831, '\P{Is_Nt:   de}', "");
Expect(1, 120831, '\P{^Is_Nt:   de}', "");
Expect(0, 120832, '\p{Is_Nt:   de}', "");
Expect(1, 120832, '\p{^Is_Nt:   de}', "");
Expect(1, 120832, '\P{Is_Nt:   de}', "");
Expect(0, 120832, '\P{^Is_Nt:   de}', "");
Expect(1, 120831, '\p{Is_Nt:	De}', "");
Expect(0, 120831, '\p{^Is_Nt:	De}', "");
Expect(0, 120831, '\P{Is_Nt:	De}', "");
Expect(1, 120831, '\P{^Is_Nt:	De}', "");
Expect(0, 120832, '\p{Is_Nt:	De}', "");
Expect(1, 120832, '\p{^Is_Nt:	De}', "");
Expect(1, 120832, '\P{Is_Nt:	De}', "");
Expect(0, 120832, '\P{^Is_Nt:	De}', "");
Error('\p{Numeric_Type= _NUMERIC:=}');
Error('\P{Numeric_Type= _NUMERIC:=}');
Expect(1, 194704, '\p{Numeric_Type=numeric}', "");
Expect(0, 194704, '\p{^Numeric_Type=numeric}', "");
Expect(0, 194704, '\P{Numeric_Type=numeric}', "");
Expect(1, 194704, '\P{^Numeric_Type=numeric}', "");
Expect(0, 194705, '\p{Numeric_Type=numeric}', "");
Expect(1, 194705, '\p{^Numeric_Type=numeric}', "");
Expect(1, 194705, '\P{Numeric_Type=numeric}', "");
Expect(0, 194705, '\P{^Numeric_Type=numeric}', "");
Expect(1, 194704, '\p{Numeric_Type=-_numeric}', "");
Expect(0, 194704, '\p{^Numeric_Type=-_numeric}', "");
Expect(0, 194704, '\P{Numeric_Type=-_numeric}', "");
Expect(1, 194704, '\P{^Numeric_Type=-_numeric}', "");
Expect(0, 194705, '\p{Numeric_Type=-_numeric}', "");
Expect(1, 194705, '\p{^Numeric_Type=-_numeric}', "");
Expect(1, 194705, '\P{Numeric_Type=-_numeric}', "");
Expect(0, 194705, '\P{^Numeric_Type=-_numeric}', "");
Error('\p{Nt=:= nu}');
Error('\P{Nt=:= nu}');
Expect(1, 194704, '\p{Nt=nu}', "");
Expect(0, 194704, '\p{^Nt=nu}', "");
Expect(0, 194704, '\P{Nt=nu}', "");
Expect(1, 194704, '\P{^Nt=nu}', "");
Expect(0, 194705, '\p{Nt=nu}', "");
Expect(1, 194705, '\p{^Nt=nu}', "");
Expect(1, 194705, '\P{Nt=nu}', "");
Expect(0, 194705, '\P{^Nt=nu}', "");
Expect(1, 194704, '\p{Nt= _Nu}', "");
Expect(0, 194704, '\p{^Nt= _Nu}', "");
Expect(0, 194704, '\P{Nt= _Nu}', "");
Expect(1, 194704, '\P{^Nt= _Nu}', "");
Expect(0, 194705, '\p{Nt= _Nu}', "");
Expect(1, 194705, '\p{^Nt= _Nu}', "");
Expect(1, 194705, '\P{Nt= _Nu}', "");
Expect(0, 194705, '\P{^Nt= _Nu}', "");
Error('\p{Is_Numeric_Type= Numeric:=}');
Error('\P{Is_Numeric_Type= Numeric:=}');
Expect(1, 194704, '\p{Is_Numeric_Type=numeric}', "");
Expect(0, 194704, '\p{^Is_Numeric_Type=numeric}', "");
Expect(0, 194704, '\P{Is_Numeric_Type=numeric}', "");
Expect(1, 194704, '\P{^Is_Numeric_Type=numeric}', "");
Expect(0, 194705, '\p{Is_Numeric_Type=numeric}', "");
Expect(1, 194705, '\p{^Is_Numeric_Type=numeric}', "");
Expect(1, 194705, '\P{Is_Numeric_Type=numeric}', "");
Expect(0, 194705, '\P{^Is_Numeric_Type=numeric}', "");
Expect(1, 194704, '\p{Is_Numeric_Type=	 Numeric}', "");
Expect(0, 194704, '\p{^Is_Numeric_Type=	 Numeric}', "");
Expect(0, 194704, '\P{Is_Numeric_Type=	 Numeric}', "");
Expect(1, 194704, '\P{^Is_Numeric_Type=	 Numeric}', "");
Expect(0, 194705, '\p{Is_Numeric_Type=	 Numeric}', "");
Expect(1, 194705, '\p{^Is_Numeric_Type=	 Numeric}', "");
Expect(1, 194705, '\P{Is_Numeric_Type=	 Numeric}', "");
Expect(0, 194705, '\P{^Is_Numeric_Type=	 Numeric}', "");
Error('\p{Is_Nt= /a/NU}');
Error('\P{Is_Nt= /a/NU}');
Expect(1, 194704, '\p{Is_Nt=nu}', "");
Expect(0, 194704, '\p{^Is_Nt=nu}', "");
Expect(0, 194704, '\P{Is_Nt=nu}', "");
Expect(1, 194704, '\P{^Is_Nt=nu}', "");
Expect(0, 194705, '\p{Is_Nt=nu}', "");
Expect(1, 194705, '\p{^Is_Nt=nu}', "");
Expect(1, 194705, '\P{Is_Nt=nu}', "");
Expect(0, 194705, '\P{^Is_Nt=nu}', "");
Expect(1, 194704, '\p{Is_Nt=_-nu}', "");
Expect(0, 194704, '\p{^Is_Nt=_-nu}', "");
Expect(0, 194704, '\P{Is_Nt=_-nu}', "");
Expect(1, 194704, '\P{^Is_Nt=_-nu}', "");
Expect(0, 194705, '\p{Is_Nt=_-nu}', "");
Expect(1, 194705, '\p{^Is_Nt=_-nu}', "");
Expect(1, 194705, '\P{Is_Nt=_-nu}', "");
Expect(0, 194705, '\P{^Is_Nt=_-nu}', "");
Error('\p{Lowercase= 	Yes:=}');
Error('\P{Lowercase= 	Yes:=}');
Expect(1, 120779, '\p{Lowercase:	yes}', "");
Expect(0, 120779, '\p{^Lowercase:	yes}', "");
Expect(0, 120779, '\P{Lowercase:	yes}', "");
Expect(1, 120779, '\P{^Lowercase:	yes}', "");
Expect(0, 120780, '\p{Lowercase:	yes}', "");
Expect(1, 120780, '\p{^Lowercase:	yes}', "");
Expect(1, 120780, '\P{Lowercase:	yes}', "");
Expect(0, 120780, '\P{^Lowercase:	yes}', "");
Expect(1, 120779, '\p{Lowercase=__Yes}', "");
Expect(0, 120779, '\p{^Lowercase=__Yes}', "");
Expect(0, 120779, '\P{Lowercase=__Yes}', "");
Expect(1, 120779, '\P{^Lowercase=__Yes}', "");
Expect(0, 120780, '\p{Lowercase=__Yes}', "");
Expect(1, 120780, '\p{^Lowercase=__Yes}', "");
Expect(1, 120780, '\P{Lowercase=__Yes}', "");
Expect(0, 120780, '\P{^Lowercase=__Yes}', "");
Error('\p{Lower=:=_y}');
Error('\P{Lower=:=_y}');
Expect(1, 120779, '\p{Lower=y}', "");
Expect(0, 120779, '\p{^Lower=y}', "");
Expect(0, 120779, '\P{Lower=y}', "");
Expect(1, 120779, '\P{^Lower=y}', "");
Expect(0, 120780, '\p{Lower=y}', "");
Expect(1, 120780, '\p{^Lower=y}', "");
Expect(1, 120780, '\P{Lower=y}', "");
Expect(0, 120780, '\P{^Lower=y}', "");
Expect(1, 120779, '\p{Lower=  Y}', "");
Expect(0, 120779, '\p{^Lower=  Y}', "");
Expect(0, 120779, '\P{Lower=  Y}', "");
Expect(1, 120779, '\P{^Lower=  Y}', "");
Expect(0, 120780, '\p{Lower=  Y}', "");
Expect(1, 120780, '\p{^Lower=  Y}', "");
Expect(1, 120780, '\P{Lower=  Y}', "");
Expect(0, 120780, '\P{^Lower=  Y}', "");
Error('\p{Is_Lowercase=/a/ -t}');
Error('\P{Is_Lowercase=/a/ -t}');
Expect(1, 120779, '\p{Is_Lowercase: t}', "");
Expect(0, 120779, '\p{^Is_Lowercase: t}', "");
Expect(0, 120779, '\P{Is_Lowercase: t}', "");
Expect(1, 120779, '\P{^Is_Lowercase: t}', "");
Expect(0, 120780, '\p{Is_Lowercase: t}', "");
Expect(1, 120780, '\p{^Is_Lowercase: t}', "");
Expect(1, 120780, '\P{Is_Lowercase: t}', "");
Expect(0, 120780, '\P{^Is_Lowercase: t}', "");
Expect(1, 120779, '\p{Is_Lowercase=	T}', "");
Expect(0, 120779, '\p{^Is_Lowercase=	T}', "");
Expect(0, 120779, '\P{Is_Lowercase=	T}', "");
Expect(1, 120779, '\P{^Is_Lowercase=	T}', "");
Expect(0, 120780, '\p{Is_Lowercase=	T}', "");
Expect(1, 120780, '\p{^Is_Lowercase=	T}', "");
Expect(1, 120780, '\P{Is_Lowercase=	T}', "");
Expect(0, 120780, '\P{^Is_Lowercase=	T}', "");
Error('\p{Is_Lower=:=	-TRUE}');
Error('\P{Is_Lower=:=	-TRUE}');
Expect(1, 120779, '\p{Is_Lower=true}', "");
Expect(0, 120779, '\p{^Is_Lower=true}', "");
Expect(0, 120779, '\P{Is_Lower=true}', "");
Expect(1, 120779, '\P{^Is_Lower=true}', "");
Expect(0, 120780, '\p{Is_Lower=true}', "");
Expect(1, 120780, '\p{^Is_Lower=true}', "");
Expect(1, 120780, '\P{Is_Lower=true}', "");
Expect(0, 120780, '\P{^Is_Lower=true}', "");
Expect(1, 120779, '\p{Is_Lower=- True}', "");
Expect(0, 120779, '\p{^Is_Lower=- True}', "");
Expect(0, 120779, '\P{Is_Lower=- True}', "");
Expect(1, 120779, '\P{^Is_Lower=- True}', "");
Expect(0, 120780, '\p{Is_Lower=- True}', "");
Expect(1, 120780, '\p{^Is_Lower=- True}', "");
Expect(1, 120780, '\P{Is_Lower=- True}', "");
Expect(0, 120780, '\P{^Is_Lower=- True}', "");
Error('\p{Lowercase=/a/no}');
Error('\P{Lowercase=/a/no}');
Expect(1, 120780, '\p{Lowercase=no}', "");
Expect(0, 120780, '\p{^Lowercase=no}', "");
Expect(0, 120780, '\P{Lowercase=no}', "");
Expect(1, 120780, '\P{^Lowercase=no}', "");
Expect(0, 120779, '\p{Lowercase=no}', "");
Expect(1, 120779, '\p{^Lowercase=no}', "");
Expect(1, 120779, '\P{Lowercase=no}', "");
Expect(0, 120779, '\P{^Lowercase=no}', "");
Expect(1, 120780, '\p{Lowercase=- NO}', "");
Expect(0, 120780, '\p{^Lowercase=- NO}', "");
Expect(0, 120780, '\P{Lowercase=- NO}', "");
Expect(1, 120780, '\P{^Lowercase=- NO}', "");
Expect(0, 120779, '\p{Lowercase=- NO}', "");
Expect(1, 120779, '\p{^Lowercase=- NO}', "");
Expect(1, 120779, '\P{Lowercase=- NO}', "");
Expect(0, 120779, '\P{^Lowercase=- NO}', "");
Error('\p{Lower=_-n:=}');
Error('\P{Lower=_-n:=}');
Expect(1, 120780, '\p{Lower=n}', "");
Expect(0, 120780, '\p{^Lower=n}', "");
Expect(0, 120780, '\P{Lower=n}', "");
Expect(1, 120780, '\P{^Lower=n}', "");
Expect(0, 120779, '\p{Lower=n}', "");
Expect(1, 120779, '\p{^Lower=n}', "");
Expect(1, 120779, '\P{Lower=n}', "");
Expect(0, 120779, '\P{^Lower=n}', "");
Expect(1, 120780, '\p{Lower=-_N}', "");
Expect(0, 120780, '\p{^Lower=-_N}', "");
Expect(0, 120780, '\P{Lower=-_N}', "");
Expect(1, 120780, '\P{^Lower=-_N}', "");
Expect(0, 120779, '\p{Lower=-_N}', "");
Expect(1, 120779, '\p{^Lower=-_N}', "");
Expect(1, 120779, '\P{Lower=-_N}', "");
Expect(0, 120779, '\P{^Lower=-_N}', "");
Error('\p{Is_Lowercase=  F:=}');
Error('\P{Is_Lowercase=  F:=}');
Expect(1, 120780, '\p{Is_Lowercase=f}', "");
Expect(0, 120780, '\p{^Is_Lowercase=f}', "");
Expect(0, 120780, '\P{Is_Lowercase=f}', "");
Expect(1, 120780, '\P{^Is_Lowercase=f}', "");
Expect(0, 120779, '\p{Is_Lowercase=f}', "");
Expect(1, 120779, '\p{^Is_Lowercase=f}', "");
Expect(1, 120779, '\P{Is_Lowercase=f}', "");
Expect(0, 120779, '\P{^Is_Lowercase=f}', "");
Expect(1, 120780, '\p{Is_Lowercase=  F}', "");
Expect(0, 120780, '\p{^Is_Lowercase=  F}', "");
Expect(0, 120780, '\P{Is_Lowercase=  F}', "");
Expect(1, 120780, '\P{^Is_Lowercase=  F}', "");
Expect(0, 120779, '\p{Is_Lowercase=  F}', "");
Expect(1, 120779, '\p{^Is_Lowercase=  F}', "");
Expect(1, 120779, '\P{Is_Lowercase=  F}', "");
Expect(0, 120779, '\P{^Is_Lowercase=  F}', "");
Error('\p{Is_Lower=__FALSE:=}');
Error('\P{Is_Lower=__FALSE:=}');
Expect(1, 120780, '\p{Is_Lower=false}', "");
Expect(0, 120780, '\p{^Is_Lower=false}', "");
Expect(0, 120780, '\P{Is_Lower=false}', "");
Expect(1, 120780, '\P{^Is_Lower=false}', "");
Expect(0, 120779, '\p{Is_Lower=false}', "");
Expect(1, 120779, '\p{^Is_Lower=false}', "");
Expect(1, 120779, '\P{Is_Lower=false}', "");
Expect(0, 120779, '\P{^Is_Lower=false}', "");
Expect(1, 120780, '\p{Is_Lower=_	FALSE}', "");
Expect(0, 120780, '\p{^Is_Lower=_	FALSE}', "");
Expect(0, 120780, '\P{Is_Lower=_	FALSE}', "");
Expect(1, 120780, '\P{^Is_Lower=_	FALSE}', "");
Expect(0, 120779, '\p{Is_Lower=_	FALSE}', "");
Expect(1, 120779, '\p{^Is_Lower=_	FALSE}', "");
Expect(1, 120779, '\P{Is_Lower=_	FALSE}', "");
Expect(0, 120779, '\P{^Is_Lower=_	FALSE}', "");
Error('\p{Canonical_Combining_Class: /a/__CCC32}');
Error('\P{Canonical_Combining_Class: /a/__CCC32}');
Expect(1, 1616, '\p{Canonical_Combining_Class=ccc32}', "");
Expect(0, 1616, '\p{^Canonical_Combining_Class=ccc32}', "");
Expect(0, 1616, '\P{Canonical_Combining_Class=ccc32}', "");
Expect(1, 1616, '\P{^Canonical_Combining_Class=ccc32}', "");
Expect(0, 1617, '\p{Canonical_Combining_Class=ccc32}', "");
Expect(1, 1617, '\p{^Canonical_Combining_Class=ccc32}', "");
Expect(1, 1617, '\P{Canonical_Combining_Class=ccc32}', "");
Expect(0, 1617, '\P{^Canonical_Combining_Class=ccc32}', "");
Expect(1, 1616, '\p{Canonical_Combining_Class=_ CCC32}', "");
Expect(0, 1616, '\p{^Canonical_Combining_Class=_ CCC32}', "");
Expect(0, 1616, '\P{Canonical_Combining_Class=_ CCC32}', "");
Expect(1, 1616, '\P{^Canonical_Combining_Class=_ CCC32}', "");
Expect(0, 1617, '\p{Canonical_Combining_Class=_ CCC32}', "");
Expect(1, 1617, '\p{^Canonical_Combining_Class=_ CCC32}', "");
Expect(1, 1617, '\P{Canonical_Combining_Class=_ CCC32}', "");
Expect(0, 1617, '\P{^Canonical_Combining_Class=_ CCC32}', "");
Error('\p{Ccc=__3_2:=}');
Error('\P{Ccc=__3_2:=}');
Expect(1, 1616, '\p{Ccc=000000032}', "");
Expect(0, 1616, '\p{^Ccc=000000032}', "");
Expect(0, 1616, '\P{Ccc=000000032}', "");
Expect(1, 1616, '\P{^Ccc=000000032}', "");
Expect(0, 1617, '\p{Ccc=000000032}', "");
Expect(1, 1617, '\p{^Ccc=000000032}', "");
Expect(1, 1617, '\P{Ccc=000000032}', "");
Expect(0, 1617, '\P{^Ccc=000000032}', "");
Error('\p{Is_Canonical_Combining_Class=--ccc32:=}');
Error('\P{Is_Canonical_Combining_Class=--ccc32:=}');
Expect(1, 1616, '\p{Is_Canonical_Combining_Class=ccc32}', "");
Expect(0, 1616, '\p{^Is_Canonical_Combining_Class=ccc32}', "");
Expect(0, 1616, '\P{Is_Canonical_Combining_Class=ccc32}', "");
Expect(1, 1616, '\P{^Is_Canonical_Combining_Class=ccc32}', "");
Expect(0, 1617, '\p{Is_Canonical_Combining_Class=ccc32}', "");
Expect(1, 1617, '\p{^Is_Canonical_Combining_Class=ccc32}', "");
Expect(1, 1617, '\P{Is_Canonical_Combining_Class=ccc32}', "");
Expect(0, 1617, '\P{^Is_Canonical_Combining_Class=ccc32}', "");
Expect(1, 1616, '\p{Is_Canonical_Combining_Class=_CCC32}', "");
Expect(0, 1616, '\p{^Is_Canonical_Combining_Class=_CCC32}', "");
Expect(0, 1616, '\P{Is_Canonical_Combining_Class=_CCC32}', "");
Expect(1, 1616, '\P{^Is_Canonical_Combining_Class=_CCC32}', "");
Expect(0, 1617, '\p{Is_Canonical_Combining_Class=_CCC32}', "");
Expect(1, 1617, '\p{^Is_Canonical_Combining_Class=_CCC32}', "");
Expect(1, 1617, '\P{Is_Canonical_Combining_Class=_CCC32}', "");
Expect(0, 1617, '\P{^Is_Canonical_Combining_Class=_CCC32}', "");
Error('\p{Is_Ccc=/a/	 +0000000003_2}');
Error('\P{Is_Ccc=/a/	 +0000000003_2}');
Expect(1, 1616, '\p{Is_Ccc=32}', "");
Expect(0, 1616, '\p{^Is_Ccc=32}', "");
Expect(0, 1616, '\P{Is_Ccc=32}', "");
Expect(1, 1616, '\P{^Is_Ccc=32}', "");
Expect(0, 1617, '\p{Is_Ccc=32}', "");
Expect(1, 1617, '\p{^Is_Ccc=32}', "");
Expect(1, 1617, '\P{Is_Ccc=32}', "");
Expect(0, 1617, '\P{^Is_Ccc=32}', "");
Error('\p{Canonical_Combining_Class:   _	CCC118/a/}');
Error('\P{Canonical_Combining_Class:   _	CCC118/a/}');
Expect(1, 3769, '\p{Canonical_Combining_Class=ccc118}', "");
Expect(0, 3769, '\p{^Canonical_Combining_Class=ccc118}', "");
Expect(0, 3769, '\P{Canonical_Combining_Class=ccc118}', "");
Expect(1, 3769, '\P{^Canonical_Combining_Class=ccc118}', "");
Expect(0, 3770, '\p{Canonical_Combining_Class=ccc118}', "");
Expect(1, 3770, '\p{^Canonical_Combining_Class=ccc118}', "");
Expect(1, 3770, '\P{Canonical_Combining_Class=ccc118}', "");
Expect(0, 3770, '\P{^Canonical_Combining_Class=ccc118}', "");
Expect(1, 3769, '\p{Canonical_Combining_Class=	CCC118}', "");
Expect(0, 3769, '\p{^Canonical_Combining_Class=	CCC118}', "");
Expect(0, 3769, '\P{Canonical_Combining_Class=	CCC118}', "");
Expect(1, 3769, '\P{^Canonical_Combining_Class=	CCC118}', "");
Expect(0, 3770, '\p{Canonical_Combining_Class=	CCC118}', "");
Expect(1, 3770, '\p{^Canonical_Combining_Class=	CCC118}', "");
Expect(1, 3770, '\P{Canonical_Combining_Class=	CCC118}', "");
Expect(0, 3770, '\P{^Canonical_Combining_Class=	CCC118}', "");
Error('\p{Ccc:		_00_00_01_18/a/}');
Error('\P{Ccc:		_00_00_01_18/a/}');
Expect(1, 3769, '\p{Ccc=11_8}', "");
Expect(0, 3769, '\p{^Ccc=11_8}', "");
Expect(0, 3769, '\P{Ccc=11_8}', "");
Expect(1, 3769, '\P{^Ccc=11_8}', "");
Expect(0, 3770, '\p{Ccc=11_8}', "");
Expect(1, 3770, '\p{^Ccc=11_8}', "");
Expect(1, 3770, '\P{Ccc=11_8}', "");
Expect(0, 3770, '\P{^Ccc=11_8}', "");
Error('\p{Is_Canonical_Combining_Class=/a/-_ccc118}');
Error('\P{Is_Canonical_Combining_Class=/a/-_ccc118}');
Expect(1, 3769, '\p{Is_Canonical_Combining_Class=ccc118}', "");
Expect(0, 3769, '\p{^Is_Canonical_Combining_Class=ccc118}', "");
Expect(0, 3769, '\P{Is_Canonical_Combining_Class=ccc118}', "");
Expect(1, 3769, '\P{^Is_Canonical_Combining_Class=ccc118}', "");
Expect(0, 3770, '\p{Is_Canonical_Combining_Class=ccc118}', "");
Expect(1, 3770, '\p{^Is_Canonical_Combining_Class=ccc118}', "");
Expect(1, 3770, '\P{Is_Canonical_Combining_Class=ccc118}', "");
Expect(0, 3770, '\P{^Is_Canonical_Combining_Class=ccc118}', "");
Expect(1, 3769, '\p{Is_Canonical_Combining_Class=	ccc118}', "");
Expect(0, 3769, '\p{^Is_Canonical_Combining_Class=	ccc118}', "");
Expect(0, 3769, '\P{Is_Canonical_Combining_Class=	ccc118}', "");
Expect(1, 3769, '\P{^Is_Canonical_Combining_Class=	ccc118}', "");
Expect(0, 3770, '\p{Is_Canonical_Combining_Class=	ccc118}', "");
Expect(1, 3770, '\p{^Is_Canonical_Combining_Class=	ccc118}', "");
Expect(1, 3770, '\P{Is_Canonical_Combining_Class=	ccc118}', "");
Expect(0, 3770, '\P{^Is_Canonical_Combining_Class=	ccc118}', "");
Error('\p{Is_Ccc=-_00011_8/a/}');
Error('\P{Is_Ccc=-_00011_8/a/}');
Expect(1, 3769, '\p{Is_Ccc=+0118}', "");
Expect(0, 3769, '\p{^Is_Ccc=+0118}', "");
Expect(0, 3769, '\P{Is_Ccc=+0118}', "");
Expect(1, 3769, '\P{^Is_Ccc=+0118}', "");
Expect(0, 3770, '\p{Is_Ccc=+0118}', "");
Expect(1, 3770, '\p{^Is_Ccc=+0118}', "");
Expect(1, 3770, '\P{Is_Ccc=+0118}', "");
Expect(0, 3770, '\P{^Is_Ccc=+0118}', "");
Error('\p{Canonical_Combining_Class=:=  Double_ABOVE}');
Error('\P{Canonical_Combining_Class=:=  Double_ABOVE}');
Expect(1, 7629, '\p{Canonical_Combining_Class=doubleabove}', "");
Expect(0, 7629, '\p{^Canonical_Combining_Class=doubleabove}', "");
Expect(0, 7629, '\P{Canonical_Combining_Class=doubleabove}', "");
Expect(1, 7629, '\P{^Canonical_Combining_Class=doubleabove}', "");
Expect(0, 7630, '\p{Canonical_Combining_Class=doubleabove}', "");
Expect(1, 7630, '\p{^Canonical_Combining_Class=doubleabove}', "");
Expect(1, 7630, '\P{Canonical_Combining_Class=doubleabove}', "");
Expect(0, 7630, '\P{^Canonical_Combining_Class=doubleabove}', "");
Expect(1, 7629, '\p{Canonical_Combining_Class=_	Double_above}', "");
Expect(0, 7629, '\p{^Canonical_Combining_Class=_	Double_above}', "");
Expect(0, 7629, '\P{Canonical_Combining_Class=_	Double_above}', "");
Expect(1, 7629, '\P{^Canonical_Combining_Class=_	Double_above}', "");
Expect(0, 7630, '\p{Canonical_Combining_Class=_	Double_above}', "");
Expect(1, 7630, '\p{^Canonical_Combining_Class=_	Double_above}', "");
Expect(1, 7630, '\P{Canonical_Combining_Class=_	Double_above}', "");
Expect(0, 7630, '\P{^Canonical_Combining_Class=_	Double_above}', "");
Error('\p{Ccc=/a/-_DA}');
Error('\P{Ccc=/a/-_DA}');
Expect(1, 7629, '\p{Ccc=da}', "");
Expect(0, 7629, '\p{^Ccc=da}', "");
Expect(0, 7629, '\P{Ccc=da}', "");
Expect(1, 7629, '\P{^Ccc=da}', "");
Expect(0, 7630, '\p{Ccc=da}', "");
Expect(1, 7630, '\p{^Ccc=da}', "");
Expect(1, 7630, '\P{Ccc=da}', "");
Expect(0, 7630, '\P{^Ccc=da}', "");
Expect(1, 7629, '\p{Ccc=  DA}', "");
Expect(0, 7629, '\p{^Ccc=  DA}', "");
Expect(0, 7629, '\P{Ccc=  DA}', "");
Expect(1, 7629, '\P{^Ccc=  DA}', "");
Expect(0, 7630, '\p{Ccc=  DA}', "");
Expect(1, 7630, '\p{^Ccc=  DA}', "");
Expect(1, 7630, '\P{Ccc=  DA}', "");
Expect(0, 7630, '\P{^Ccc=  DA}', "");
Error('\p{Is_Canonical_Combining_Class=/a/ +00234}');
Error('\P{Is_Canonical_Combining_Class=/a/ +00234}');
Expect(1, 7629, '\p{Is_Canonical_Combining_Class: 00_23_4}', "");
Expect(0, 7629, '\p{^Is_Canonical_Combining_Class: 00_23_4}', "");
Expect(0, 7629, '\P{Is_Canonical_Combining_Class: 00_23_4}', "");
Expect(1, 7629, '\P{^Is_Canonical_Combining_Class: 00_23_4}', "");
Expect(0, 7630, '\p{Is_Canonical_Combining_Class: 00_23_4}', "");
Expect(1, 7630, '\p{^Is_Canonical_Combining_Class: 00_23_4}', "");
Expect(1, 7630, '\P{Is_Canonical_Combining_Class: 00_23_4}', "");
Expect(0, 7630, '\P{^Is_Canonical_Combining_Class: 00_23_4}', "");
Error('\p{Is_Ccc=_	double_above:=}');
Error('\P{Is_Ccc=_	double_above:=}');
Expect(1, 7629, '\p{Is_Ccc=doubleabove}', "");
Expect(0, 7629, '\p{^Is_Ccc=doubleabove}', "");
Expect(0, 7629, '\P{Is_Ccc=doubleabove}', "");
Expect(1, 7629, '\P{^Is_Ccc=doubleabove}', "");
Expect(0, 7630, '\p{Is_Ccc=doubleabove}', "");
Expect(1, 7630, '\p{^Is_Ccc=doubleabove}', "");
Expect(1, 7630, '\P{Is_Ccc=doubleabove}', "");
Expect(0, 7630, '\P{^Is_Ccc=doubleabove}', "");
Expect(1, 7629, '\p{Is_Ccc:  Double_above}', "");
Expect(0, 7629, '\p{^Is_Ccc:  Double_above}', "");
Expect(0, 7629, '\P{Is_Ccc:  Double_above}', "");
Expect(1, 7629, '\P{^Is_Ccc:  Double_above}', "");
Expect(0, 7630, '\p{Is_Ccc:  Double_above}', "");
Expect(1, 7630, '\p{^Is_Ccc:  Double_above}', "");
Expect(1, 7630, '\P{Is_Ccc:  Double_above}', "");
Expect(0, 7630, '\P{^Is_Ccc:  Double_above}', "");
Error('\p{Canonical_Combining_Class=:= ATTACHED_Below_LEFT}');
Error('\P{Canonical_Combining_Class=:= ATTACHED_Below_LEFT}');
Expect(0, 1, '\p{Canonical_Combining_Class=attachedbelowleft}', "");
Expect(1, 1, '\p{^Canonical_Combining_Class=attachedbelowleft}', "");
Expect(1, 1, '\P{Canonical_Combining_Class=attachedbelowleft}', "");
Expect(0, 1, '\P{^Canonical_Combining_Class=attachedbelowleft}', "");
Expect(0, 1, '\p{Canonical_Combining_Class=_attached_Below_Left}', "");
Expect(1, 1, '\p{^Canonical_Combining_Class=_attached_Below_Left}', "");
Expect(1, 1, '\P{Canonical_Combining_Class=_attached_Below_Left}', "");
Expect(0, 1, '\P{^Canonical_Combining_Class=_attached_Below_Left}', "");
Error('\p{Ccc=--ATBL/a/}');
Error('\P{Ccc=--ATBL/a/}');
Expect(0, 1, '\p{Ccc=atbl}', "");
Expect(1, 1, '\p{^Ccc=atbl}', "");
Expect(1, 1, '\P{Ccc=atbl}', "");
Expect(0, 1, '\P{^Ccc=atbl}', "");
Expect(0, 1, '\p{Ccc= _ATBL}', "");
Expect(1, 1, '\p{^Ccc= _ATBL}', "");
Expect(1, 1, '\P{Ccc= _ATBL}', "");
Expect(0, 1, '\P{^Ccc= _ATBL}', "");
Error('\p{Is_Canonical_Combining_Class=__000000200/a/}');
Error('\P{Is_Canonical_Combining_Class=__000000200/a/}');
Expect(0, 1, '\p{Is_Canonical_Combining_Class=20_0}', "");
Expect(1, 1, '\p{^Is_Canonical_Combining_Class=20_0}', "");
Expect(1, 1, '\P{Is_Canonical_Combining_Class=20_0}', "");
Expect(0, 1, '\P{^Is_Canonical_Combining_Class=20_0}', "");
Error('\p{Is_Ccc=/a/ -attached_BELOW_Left}');
Error('\P{Is_Ccc=/a/ -attached_BELOW_Left}');
Expect(0, 1, '\p{Is_Ccc=attachedbelowleft}', "");
Expect(1, 1, '\p{^Is_Ccc=attachedbelowleft}', "");
Expect(1, 1, '\P{Is_Ccc=attachedbelowleft}', "");
Expect(0, 1, '\P{^Is_Ccc=attachedbelowleft}', "");
Expect(0, 1, '\p{Is_Ccc: 	-attached_below_LEFT}', "");
Expect(1, 1, '\p{^Is_Ccc: 	-attached_below_LEFT}', "");
Expect(1, 1, '\P{Is_Ccc: 	-attached_below_LEFT}', "");
Expect(0, 1, '\P{^Is_Ccc: 	-attached_below_LEFT}', "");
Error('\p{Canonical_Combining_Class=CCC18:=}');
Error('\P{Canonical_Combining_Class=CCC18:=}');
Expect(1, 1479, '\p{Canonical_Combining_Class=ccc18}', "");
Expect(0, 1479, '\p{^Canonical_Combining_Class=ccc18}', "");
Expect(0, 1479, '\P{Canonical_Combining_Class=ccc18}', "");
Expect(1, 1479, '\P{^Canonical_Combining_Class=ccc18}', "");
Expect(0, 1480, '\p{Canonical_Combining_Class=ccc18}', "");
Expect(1, 1480, '\p{^Canonical_Combining_Class=ccc18}', "");
Expect(1, 1480, '\P{Canonical_Combining_Class=ccc18}', "");
Expect(0, 1480, '\P{^Canonical_Combining_Class=ccc18}', "");
Expect(1, 1479, '\p{Canonical_Combining_Class=  CCC18}', "");
Expect(0, 1479, '\p{^Canonical_Combining_Class=  CCC18}', "");
Expect(0, 1479, '\P{Canonical_Combining_Class=  CCC18}', "");
Expect(1, 1479, '\P{^Canonical_Combining_Class=  CCC18}', "");
Expect(0, 1480, '\p{Canonical_Combining_Class=  CCC18}', "");
Expect(1, 1480, '\p{^Canonical_Combining_Class=  CCC18}', "");
Expect(1, 1480, '\P{Canonical_Combining_Class=  CCC18}', "");
Expect(0, 1480, '\P{^Canonical_Combining_Class=  CCC18}', "");
Error('\p{Ccc=	/a/00000000018}');
Error('\P{Ccc=	/a/00000000018}');
Expect(1, 1479, '\p{Ccc:   00001_8}', "");
Expect(0, 1479, '\p{^Ccc:   00001_8}', "");
Expect(0, 1479, '\P{Ccc:   00001_8}', "");
Expect(1, 1479, '\P{^Ccc:   00001_8}', "");
Expect(0, 1480, '\p{Ccc:   00001_8}', "");
Expect(1, 1480, '\p{^Ccc:   00001_8}', "");
Expect(1, 1480, '\P{Ccc:   00001_8}', "");
Expect(0, 1480, '\P{^Ccc:   00001_8}', "");
Error('\p{Is_Canonical_Combining_Class= 	CCC18:=}');
Error('\P{Is_Canonical_Combining_Class= 	CCC18:=}');
Expect(1, 1479, '\p{Is_Canonical_Combining_Class=ccc18}', "");
Expect(0, 1479, '\p{^Is_Canonical_Combining_Class=ccc18}', "");
Expect(0, 1479, '\P{Is_Canonical_Combining_Class=ccc18}', "");
Expect(1, 1479, '\P{^Is_Canonical_Combining_Class=ccc18}', "");
Expect(0, 1480, '\p{Is_Canonical_Combining_Class=ccc18}', "");
Expect(1, 1480, '\p{^Is_Canonical_Combining_Class=ccc18}', "");
Expect(1, 1480, '\P{Is_Canonical_Combining_Class=ccc18}', "");
Expect(0, 1480, '\P{^Is_Canonical_Combining_Class=ccc18}', "");
Expect(1, 1479, '\p{Is_Canonical_Combining_Class=-ccc18}', "");
Expect(0, 1479, '\p{^Is_Canonical_Combining_Class=-ccc18}', "");
Expect(0, 1479, '\P{Is_Canonical_Combining_Class=-ccc18}', "");
Expect(1, 1479, '\P{^Is_Canonical_Combining_Class=-ccc18}', "");
Expect(0, 1480, '\p{Is_Canonical_Combining_Class=-ccc18}', "");
Expect(1, 1480, '\p{^Is_Canonical_Combining_Class=-ccc18}', "");
Expect(1, 1480, '\P{Is_Canonical_Combining_Class=-ccc18}', "");
Expect(0, 1480, '\P{^Is_Canonical_Combining_Class=-ccc18}', "");
Error('\p{Is_Ccc=/a/-	00_00_00_00_018}');
Error('\P{Is_Ccc=/a/-	00_00_00_00_018}');
Expect(1, 1479, '\p{Is_Ccc:   +0001_8}', "");
Expect(0, 1479, '\p{^Is_Ccc:   +0001_8}', "");
Expect(0, 1479, '\P{Is_Ccc:   +0001_8}', "");
Expect(1, 1479, '\P{^Is_Ccc:   +0001_8}', "");
Expect(0, 1480, '\p{Is_Ccc:   +0001_8}', "");
Expect(1, 1480, '\p{^Is_Ccc:   +0001_8}', "");
Expect(1, 1480, '\P{Is_Ccc:   +0001_8}', "");
Expect(0, 1480, '\P{^Is_Ccc:   +0001_8}', "");
Error('\p{Canonical_Combining_Class=/a/  ccc16}');
Error('\P{Canonical_Combining_Class=/a/  ccc16}');
Expect(1, 1462, '\p{Canonical_Combining_Class=ccc16}', "");
Expect(0, 1462, '\p{^Canonical_Combining_Class=ccc16}', "");
Expect(0, 1462, '\P{Canonical_Combining_Class=ccc16}', "");
Expect(1, 1462, '\P{^Canonical_Combining_Class=ccc16}', "");
Expect(0, 1463, '\p{Canonical_Combining_Class=ccc16}', "");
Expect(1, 1463, '\p{^Canonical_Combining_Class=ccc16}', "");
Expect(1, 1463, '\P{Canonical_Combining_Class=ccc16}', "");
Expect(0, 1463, '\P{^Canonical_Combining_Class=ccc16}', "");
Expect(1, 1462, '\p{Canonical_Combining_Class=_-ccc16}', "");
Expect(0, 1462, '\p{^Canonical_Combining_Class=_-ccc16}', "");
Expect(0, 1462, '\P{Canonical_Combining_Class=_-ccc16}', "");
Expect(1, 1462, '\P{^Canonical_Combining_Class=_-ccc16}', "");
Expect(0, 1463, '\p{Canonical_Combining_Class=_-ccc16}', "");
Expect(1, 1463, '\p{^Canonical_Combining_Class=_-ccc16}', "");
Expect(1, 1463, '\P{Canonical_Combining_Class=_-ccc16}', "");
Expect(0, 1463, '\P{^Canonical_Combining_Class=_-ccc16}', "");
Error('\p{Ccc=/a/__00016}');
Error('\P{Ccc=/a/__00016}');
Expect(1, 1462, '\p{Ccc=000000016}', "");
Expect(0, 1462, '\p{^Ccc=000000016}', "");
Expect(0, 1462, '\P{Ccc=000000016}', "");
Expect(1, 1462, '\P{^Ccc=000000016}', "");
Expect(0, 1463, '\p{Ccc=000000016}', "");
Expect(1, 1463, '\p{^Ccc=000000016}', "");
Expect(1, 1463, '\P{Ccc=000000016}', "");
Expect(0, 1463, '\P{^Ccc=000000016}', "");
Error('\p{Is_Canonical_Combining_Class=/a/ccc16}');
Error('\P{Is_Canonical_Combining_Class=/a/ccc16}');
Expect(1, 1462, '\p{Is_Canonical_Combining_Class=ccc16}', "");
Expect(0, 1462, '\p{^Is_Canonical_Combining_Class=ccc16}', "");
Expect(0, 1462, '\P{Is_Canonical_Combining_Class=ccc16}', "");
Expect(1, 1462, '\P{^Is_Canonical_Combining_Class=ccc16}', "");
Expect(0, 1463, '\p{Is_Canonical_Combining_Class=ccc16}', "");
Expect(1, 1463, '\p{^Is_Canonical_Combining_Class=ccc16}', "");
Expect(1, 1463, '\P{Is_Canonical_Combining_Class=ccc16}', "");
Expect(0, 1463, '\P{^Is_Canonical_Combining_Class=ccc16}', "");
Expect(1, 1462, '\p{Is_Canonical_Combining_Class=	_CCC16}', "");
Expect(0, 1462, '\p{^Is_Canonical_Combining_Class=	_CCC16}', "");
Expect(0, 1462, '\P{Is_Canonical_Combining_Class=	_CCC16}', "");
Expect(1, 1462, '\P{^Is_Canonical_Combining_Class=	_CCC16}', "");
Expect(0, 1463, '\p{Is_Canonical_Combining_Class=	_CCC16}', "");
Expect(1, 1463, '\p{^Is_Canonical_Combining_Class=	_CCC16}', "");
Expect(1, 1463, '\P{Is_Canonical_Combining_Class=	_CCC16}', "");
Expect(0, 1463, '\P{^Is_Canonical_Combining_Class=	_CCC16}', "");
Error('\p{Is_Ccc=-:=16}');
Error('\P{Is_Ccc=-:=16}');
Expect(1, 1462, '\p{Is_Ccc=016}', "");
Expect(0, 1462, '\p{^Is_Ccc=016}', "");
Expect(0, 1462, '\P{Is_Ccc=016}', "");
Expect(1, 1462, '\P{^Is_Ccc=016}', "");
Expect(0, 1463, '\p{Is_Ccc=016}', "");
Expect(1, 1463, '\p{^Is_Ccc=016}', "");
Expect(1, 1463, '\P{Is_Ccc=016}', "");
Expect(0, 1463, '\P{^Is_Ccc=016}', "");
Error('\p{Canonical_Combining_Class=/a/ 	ccc84}');
Error('\P{Canonical_Combining_Class=/a/ 	ccc84}');
Expect(1, 3157, '\p{Canonical_Combining_Class=ccc84}', "");
Expect(0, 3157, '\p{^Canonical_Combining_Class=ccc84}', "");
Expect(0, 3157, '\P{Canonical_Combining_Class=ccc84}', "");
Expect(1, 3157, '\P{^Canonical_Combining_Class=ccc84}', "");
Expect(0, 3158, '\p{Canonical_Combining_Class=ccc84}', "");
Expect(1, 3158, '\p{^Canonical_Combining_Class=ccc84}', "");
Expect(1, 3158, '\P{Canonical_Combining_Class=ccc84}', "");
Expect(0, 3158, '\P{^Canonical_Combining_Class=ccc84}', "");
Expect(1, 3157, '\p{Canonical_Combining_Class:   _CCC84}', "");
Expect(0, 3157, '\p{^Canonical_Combining_Class:   _CCC84}', "");
Expect(0, 3157, '\P{Canonical_Combining_Class:   _CCC84}', "");
Expect(1, 3157, '\P{^Canonical_Combining_Class:   _CCC84}', "");
Expect(0, 3158, '\p{Canonical_Combining_Class:   _CCC84}', "");
Expect(1, 3158, '\p{^Canonical_Combining_Class:   _CCC84}', "");
Expect(1, 3158, '\P{Canonical_Combining_Class:   _CCC84}', "");
Expect(0, 3158, '\P{^Canonical_Combining_Class:   _CCC84}', "");
Error('\p{Ccc:/a/00000_00008_4}');
Error('\P{Ccc:/a/00000_00008_4}');
Expect(1, 3157, '\p{Ccc=0084}', "");
Expect(0, 3157, '\p{^Ccc=0084}', "");
Expect(0, 3157, '\P{Ccc=0084}', "");
Expect(1, 3157, '\P{^Ccc=0084}', "");
Expect(0, 3158, '\p{Ccc=0084}', "");
Expect(1, 3158, '\p{^Ccc=0084}', "");
Expect(1, 3158, '\P{Ccc=0084}', "");
Expect(0, 3158, '\P{^Ccc=0084}', "");
Error('\p{Is_Canonical_Combining_Class:   	CCC84/a/}');
Error('\P{Is_Canonical_Combining_Class:   	CCC84/a/}');
Expect(1, 3157, '\p{Is_Canonical_Combining_Class=ccc84}', "");
Expect(0, 3157, '\p{^Is_Canonical_Combining_Class=ccc84}', "");
Expect(0, 3157, '\P{Is_Canonical_Combining_Class=ccc84}', "");
Expect(1, 3157, '\P{^Is_Canonical_Combining_Class=ccc84}', "");
Expect(0, 3158, '\p{Is_Canonical_Combining_Class=ccc84}', "");
Expect(1, 3158, '\p{^Is_Canonical_Combining_Class=ccc84}', "");
Expect(1, 3158, '\P{Is_Canonical_Combining_Class=ccc84}', "");
Expect(0, 3158, '\P{^Is_Canonical_Combining_Class=ccc84}', "");
Expect(1, 3157, '\p{Is_Canonical_Combining_Class=--CCC84}', "");
Expect(0, 3157, '\p{^Is_Canonical_Combining_Class=--CCC84}', "");
Expect(0, 3157, '\P{Is_Canonical_Combining_Class=--CCC84}', "");
Expect(1, 3157, '\P{^Is_Canonical_Combining_Class=--CCC84}', "");
Expect(0, 3158, '\p{Is_Canonical_Combining_Class=--CCC84}', "");
Expect(1, 3158, '\p{^Is_Canonical_Combining_Class=--CCC84}', "");
Expect(1, 3158, '\P{Is_Canonical_Combining_Class=--CCC84}', "");
Expect(0, 3158, '\P{^Is_Canonical_Combining_Class=--CCC84}', "");
Error('\p{Is_Ccc=	0084/a/}');
Error('\P{Is_Ccc=	0084/a/}');
Expect(1, 3157, '\p{Is_Ccc=+0000000084}', "");
Expect(0, 3157, '\p{^Is_Ccc=+0000000084}', "");
Expect(0, 3157, '\P{Is_Ccc=+0000000084}', "");
Expect(1, 3157, '\P{^Is_Ccc=+0000000084}', "");
Expect(0, 3158, '\p{Is_Ccc=+0000000084}', "");
Expect(1, 3158, '\p{^Is_Ccc=+0000000084}', "");
Expect(1, 3158, '\P{Is_Ccc=+0000000084}', "");
Expect(0, 3158, '\P{^Is_Ccc=+0000000084}', "");
Error('\p{Canonical_Combining_Class=	:=ccc27}');
Error('\P{Canonical_Combining_Class=	:=ccc27}');
Expect(1, 2288, '\p{Canonical_Combining_Class=ccc27}', "");
Expect(0, 2288, '\p{^Canonical_Combining_Class=ccc27}', "");
Expect(0, 2288, '\P{Canonical_Combining_Class=ccc27}', "");
Expect(1, 2288, '\P{^Canonical_Combining_Class=ccc27}', "");
Expect(0, 2289, '\p{Canonical_Combining_Class=ccc27}', "");
Expect(1, 2289, '\p{^Canonical_Combining_Class=ccc27}', "");
Expect(1, 2289, '\P{Canonical_Combining_Class=ccc27}', "");
Expect(0, 2289, '\P{^Canonical_Combining_Class=ccc27}', "");
Expect(1, 2288, '\p{Canonical_Combining_Class=- ccc27}', "");
Expect(0, 2288, '\p{^Canonical_Combining_Class=- ccc27}', "");
Expect(0, 2288, '\P{Canonical_Combining_Class=- ccc27}', "");
Expect(1, 2288, '\P{^Canonical_Combining_Class=- ccc27}', "");
Expect(0, 2289, '\p{Canonical_Combining_Class=- ccc27}', "");
Expect(1, 2289, '\p{^Canonical_Combining_Class=- ccc27}', "");
Expect(1, 2289, '\P{Canonical_Combining_Class=- ccc27}', "");
Expect(0, 2289, '\P{^Canonical_Combining_Class=- ccc27}', "");
Error('\p{Ccc=_:=+00_00_27}');
Error('\P{Ccc=_:=+00_00_27}');
Expect(1, 2288, '\p{Ccc=0000000027}', "");
Expect(0, 2288, '\p{^Ccc=0000000027}', "");
Expect(0, 2288, '\P{Ccc=0000000027}', "");
Expect(1, 2288, '\P{^Ccc=0000000027}', "");
Expect(0, 2289, '\p{Ccc=0000000027}', "");
Expect(1, 2289, '\p{^Ccc=0000000027}', "");
Expect(1, 2289, '\P{Ccc=0000000027}', "");
Expect(0, 2289, '\P{^Ccc=0000000027}', "");
Error('\p{Is_Canonical_Combining_Class=/a/ -CCC27}');
Error('\P{Is_Canonical_Combining_Class=/a/ -CCC27}');
Expect(1, 2288, '\p{Is_Canonical_Combining_Class=ccc27}', "");
Expect(0, 2288, '\p{^Is_Canonical_Combining_Class=ccc27}', "");
Expect(0, 2288, '\P{Is_Canonical_Combining_Class=ccc27}', "");
Expect(1, 2288, '\P{^Is_Canonical_Combining_Class=ccc27}', "");
Expect(0, 2289, '\p{Is_Canonical_Combining_Class=ccc27}', "");
Expect(1, 2289, '\p{^Is_Canonical_Combining_Class=ccc27}', "");
Expect(1, 2289, '\P{Is_Canonical_Combining_Class=ccc27}', "");
Expect(0, 2289, '\P{^Is_Canonical_Combining_Class=ccc27}', "");
Expect(1, 2288, '\p{Is_Canonical_Combining_Class=-_CCC27}', "");
Expect(0, 2288, '\p{^Is_Canonical_Combining_Class=-_CCC27}', "");
Expect(0, 2288, '\P{Is_Canonical_Combining_Class=-_CCC27}', "");
Expect(1, 2288, '\P{^Is_Canonical_Combining_Class=-_CCC27}', "");
Expect(0, 2289, '\p{Is_Canonical_Combining_Class=-_CCC27}', "");
Expect(1, 2289, '\p{^Is_Canonical_Combining_Class=-_CCC27}', "");
Expect(1, 2289, '\P{Is_Canonical_Combining_Class=-_CCC27}', "");
Expect(0, 2289, '\P{^Is_Canonical_Combining_Class=-_CCC27}', "");
Error('\p{Is_Ccc=_/a/02_7}');
Error('\P{Is_Ccc=_/a/02_7}');
Expect(1, 2288, '\p{Is_Ccc=00027}', "");
Expect(0, 2288, '\p{^Is_Ccc=00027}', "");
Expect(0, 2288, '\P{Is_Ccc=00027}', "");
Expect(1, 2288, '\P{^Is_Ccc=00027}', "");
Expect(0, 2289, '\p{Is_Ccc=00027}', "");
Expect(1, 2289, '\p{^Is_Ccc=00027}', "");
Expect(1, 2289, '\P{Is_Ccc=00027}', "");
Expect(0, 2289, '\P{^Is_Ccc=00027}', "");
Error('\p{Canonical_Combining_Class:_ double_below:=}');
Error('\P{Canonical_Combining_Class:_ double_below:=}');
Expect(1, 7676, '\p{Canonical_Combining_Class=doublebelow}', "");
Expect(0, 7676, '\p{^Canonical_Combining_Class=doublebelow}', "");
Expect(0, 7676, '\P{Canonical_Combining_Class=doublebelow}', "");
Expect(1, 7676, '\P{^Canonical_Combining_Class=doublebelow}', "");
Expect(0, 7677, '\p{Canonical_Combining_Class=doublebelow}', "");
Expect(1, 7677, '\p{^Canonical_Combining_Class=doublebelow}', "");
Expect(1, 7677, '\P{Canonical_Combining_Class=doublebelow}', "");
Expect(0, 7677, '\P{^Canonical_Combining_Class=doublebelow}', "");
Expect(1, 7676, '\p{Canonical_Combining_Class=_-Double_BELOW}', "");
Expect(0, 7676, '\p{^Canonical_Combining_Class=_-Double_BELOW}', "");
Expect(0, 7676, '\P{Canonical_Combining_Class=_-Double_BELOW}', "");
Expect(1, 7676, '\P{^Canonical_Combining_Class=_-Double_BELOW}', "");
Expect(0, 7677, '\p{Canonical_Combining_Class=_-Double_BELOW}', "");
Expect(1, 7677, '\p{^Canonical_Combining_Class=_-Double_BELOW}', "");
Expect(1, 7677, '\P{Canonical_Combining_Class=_-Double_BELOW}', "");
Expect(0, 7677, '\P{^Canonical_Combining_Class=_-Double_BELOW}', "");
Error('\p{Ccc=:=_DB}');
Error('\P{Ccc=:=_DB}');
Expect(1, 7676, '\p{Ccc=db}', "");
Expect(0, 7676, '\p{^Ccc=db}', "");
Expect(0, 7676, '\P{Ccc=db}', "");
Expect(1, 7676, '\P{^Ccc=db}', "");
Expect(0, 7677, '\p{Ccc=db}', "");
Expect(1, 7677, '\p{^Ccc=db}', "");
Expect(1, 7677, '\P{Ccc=db}', "");
Expect(0, 7677, '\P{^Ccc=db}', "");
Expect(1, 7676, '\p{Ccc=-_DB}', "");
Expect(0, 7676, '\p{^Ccc=-_DB}', "");
Expect(0, 7676, '\P{Ccc=-_DB}', "");
Expect(1, 7676, '\P{^Ccc=-_DB}', "");
Expect(0, 7677, '\p{Ccc=-_DB}', "");
Expect(1, 7677, '\p{^Ccc=-_DB}', "");
Expect(1, 7677, '\P{Ccc=-_DB}', "");
Expect(0, 7677, '\P{^Ccc=-_DB}', "");
Error('\p{Is_Canonical_Combining_Class:/a/ -00_00_23_3}');
Error('\P{Is_Canonical_Combining_Class:/a/ -00_00_23_3}');
Expect(1, 7676, '\p{Is_Canonical_Combining_Class=023_3}', "");
Expect(0, 7676, '\p{^Is_Canonical_Combining_Class=023_3}', "");
Expect(0, 7676, '\P{Is_Canonical_Combining_Class=023_3}', "");
Expect(1, 7676, '\P{^Is_Canonical_Combining_Class=023_3}', "");
Expect(0, 7677, '\p{Is_Canonical_Combining_Class=023_3}', "");
Expect(1, 7677, '\p{^Is_Canonical_Combining_Class=023_3}', "");
Expect(1, 7677, '\P{Is_Canonical_Combining_Class=023_3}', "");
Expect(0, 7677, '\P{^Is_Canonical_Combining_Class=023_3}', "");
Error('\p{Is_Ccc=:=__Double_BELOW}');
Error('\P{Is_Ccc=:=__Double_BELOW}');
Expect(1, 7676, '\p{Is_Ccc=doublebelow}', "");
Expect(0, 7676, '\p{^Is_Ccc=doublebelow}', "");
Expect(0, 7676, '\P{Is_Ccc=doublebelow}', "");
Expect(1, 7676, '\P{^Is_Ccc=doublebelow}', "");
Expect(0, 7677, '\p{Is_Ccc=doublebelow}', "");
Expect(1, 7677, '\p{^Is_Ccc=doublebelow}', "");
Expect(1, 7677, '\P{Is_Ccc=doublebelow}', "");
Expect(0, 7677, '\P{^Is_Ccc=doublebelow}', "");
Expect(1, 7676, '\p{Is_Ccc= _double_below}', "");
Expect(0, 7676, '\p{^Is_Ccc= _double_below}', "");
Expect(0, 7676, '\P{Is_Ccc= _double_below}', "");
Expect(1, 7676, '\P{^Is_Ccc= _double_below}', "");
Expect(0, 7677, '\p{Is_Ccc= _double_below}', "");
Expect(1, 7677, '\p{^Is_Ccc= _double_below}', "");
Expect(1, 7677, '\P{Is_Ccc= _double_below}', "");
Expect(0, 7677, '\P{^Is_Ccc= _double_below}', "");
Error('\p{Canonical_Combining_Class::=	below_Right}');
Error('\P{Canonical_Combining_Class::=	below_Right}');
Expect(1, 12333, '\p{Canonical_Combining_Class=belowright}', "");
Expect(0, 12333, '\p{^Canonical_Combining_Class=belowright}', "");
Expect(0, 12333, '\P{Canonical_Combining_Class=belowright}', "");
Expect(1, 12333, '\P{^Canonical_Combining_Class=belowright}', "");
Expect(0, 12334, '\p{Canonical_Combining_Class=belowright}', "");
Expect(1, 12334, '\p{^Canonical_Combining_Class=belowright}', "");
Expect(1, 12334, '\P{Canonical_Combining_Class=belowright}', "");
Expect(0, 12334, '\P{^Canonical_Combining_Class=belowright}', "");
Expect(1, 12333, '\p{Canonical_Combining_Class=	_BELOW_Right}', "");
Expect(0, 12333, '\p{^Canonical_Combining_Class=	_BELOW_Right}', "");
Expect(0, 12333, '\P{Canonical_Combining_Class=	_BELOW_Right}', "");
Expect(1, 12333, '\P{^Canonical_Combining_Class=	_BELOW_Right}', "");
Expect(0, 12334, '\p{Canonical_Combining_Class=	_BELOW_Right}', "");
Expect(1, 12334, '\p{^Canonical_Combining_Class=	_BELOW_Right}', "");
Expect(1, 12334, '\P{Canonical_Combining_Class=	_BELOW_Right}', "");
Expect(0, 12334, '\P{^Canonical_Combining_Class=	_BELOW_Right}', "");
Error('\p{Ccc=		BR/a/}');
Error('\P{Ccc=		BR/a/}');
Expect(1, 12333, '\p{Ccc=br}', "");
Expect(0, 12333, '\p{^Ccc=br}', "");
Expect(0, 12333, '\P{Ccc=br}', "");
Expect(1, 12333, '\P{^Ccc=br}', "");
Expect(0, 12334, '\p{Ccc=br}', "");
Expect(1, 12334, '\p{^Ccc=br}', "");
Expect(1, 12334, '\P{Ccc=br}', "");
Expect(0, 12334, '\P{^Ccc=br}', "");
Expect(1, 12333, '\p{Ccc=	BR}', "");
Expect(0, 12333, '\p{^Ccc=	BR}', "");
Expect(0, 12333, '\P{Ccc=	BR}', "");
Expect(1, 12333, '\P{^Ccc=	BR}', "");
Expect(0, 12334, '\p{Ccc=	BR}', "");
Expect(1, 12334, '\p{^Ccc=	BR}', "");
Expect(1, 12334, '\P{Ccc=	BR}', "");
Expect(0, 12334, '\P{^Ccc=	BR}', "");
Error('\p{Is_Canonical_Combining_Class= /a/+00222}');
Error('\P{Is_Canonical_Combining_Class= /a/+00222}');
Expect(1, 12333, '\p{Is_Canonical_Combining_Class=+000000222}', "");
Expect(0, 12333, '\p{^Is_Canonical_Combining_Class=+000000222}', "");
Expect(0, 12333, '\P{Is_Canonical_Combining_Class=+000000222}', "");
Expect(1, 12333, '\P{^Is_Canonical_Combining_Class=+000000222}', "");
Expect(0, 12334, '\p{Is_Canonical_Combining_Class=+000000222}', "");
Expect(1, 12334, '\p{^Is_Canonical_Combining_Class=+000000222}', "");
Expect(1, 12334, '\P{Is_Canonical_Combining_Class=+000000222}', "");
Expect(0, 12334, '\P{^Is_Canonical_Combining_Class=+000000222}', "");
Error('\p{Is_Ccc=	-below_right/a/}');
Error('\P{Is_Ccc=	-below_right/a/}');
Expect(1, 12333, '\p{Is_Ccc=belowright}', "");
Expect(0, 12333, '\p{^Is_Ccc=belowright}', "");
Expect(0, 12333, '\P{Is_Ccc=belowright}', "");
Expect(1, 12333, '\P{^Is_Ccc=belowright}', "");
Expect(0, 12334, '\p{Is_Ccc=belowright}', "");
Expect(1, 12334, '\p{^Is_Ccc=belowright}', "");
Expect(1, 12334, '\P{Is_Ccc=belowright}', "");
Expect(0, 12334, '\P{^Is_Ccc=belowright}', "");
Expect(1, 12333, '\p{Is_Ccc=-BELOW_Right}', "");
Expect(0, 12333, '\p{^Is_Ccc=-BELOW_Right}', "");
Expect(0, 12333, '\P{Is_Ccc=-BELOW_Right}', "");
Expect(1, 12333, '\P{^Is_Ccc=-BELOW_Right}', "");
Expect(0, 12334, '\p{Is_Ccc=-BELOW_Right}', "");
Expect(1, 12334, '\p{^Is_Ccc=-BELOW_Right}', "");
Expect(1, 12334, '\P{Is_Ccc=-BELOW_Right}', "");
Expect(0, 12334, '\P{^Is_Ccc=-BELOW_Right}', "");
Error('\p{Canonical_Combining_Class=-:=CCC19}');
Error('\P{Canonical_Combining_Class=-:=CCC19}');
Expect(1, 1466, '\p{Canonical_Combining_Class=ccc19}', "");
Expect(0, 1466, '\p{^Canonical_Combining_Class=ccc19}', "");
Expect(0, 1466, '\P{Canonical_Combining_Class=ccc19}', "");
Expect(1, 1466, '\P{^Canonical_Combining_Class=ccc19}', "");
Expect(0, 1467, '\p{Canonical_Combining_Class=ccc19}', "");
Expect(1, 1467, '\p{^Canonical_Combining_Class=ccc19}', "");
Expect(1, 1467, '\P{Canonical_Combining_Class=ccc19}', "");
Expect(0, 1467, '\P{^Canonical_Combining_Class=ccc19}', "");
Expect(1, 1466, '\p{Canonical_Combining_Class=_CCC19}', "");
Expect(0, 1466, '\p{^Canonical_Combining_Class=_CCC19}', "");
Expect(0, 1466, '\P{Canonical_Combining_Class=_CCC19}', "");
Expect(1, 1466, '\P{^Canonical_Combining_Class=_CCC19}', "");
Expect(0, 1467, '\p{Canonical_Combining_Class=_CCC19}', "");
Expect(1, 1467, '\p{^Canonical_Combining_Class=_CCC19}', "");
Expect(1, 1467, '\P{Canonical_Combining_Class=_CCC19}', "");
Expect(0, 1467, '\P{^Canonical_Combining_Class=_CCC19}', "");
Error('\p{Ccc=:= -1_9}');
Error('\P{Ccc=:= -1_9}');
Expect(1, 1466, '\p{Ccc=1_9}', "");
Expect(0, 1466, '\p{^Ccc=1_9}', "");
Expect(0, 1466, '\P{Ccc=1_9}', "");
Expect(1, 1466, '\P{^Ccc=1_9}', "");
Expect(0, 1467, '\p{Ccc=1_9}', "");
Expect(1, 1467, '\p{^Ccc=1_9}', "");
Expect(1, 1467, '\P{Ccc=1_9}', "");
Expect(0, 1467, '\P{^Ccc=1_9}', "");
Error('\p{Is_Canonical_Combining_Class=-:=CCC19}');
Error('\P{Is_Canonical_Combining_Class=-:=CCC19}');
Expect(1, 1466, '\p{Is_Canonical_Combining_Class=ccc19}', "");
Expect(0, 1466, '\p{^Is_Canonical_Combining_Class=ccc19}', "");
Expect(0, 1466, '\P{Is_Canonical_Combining_Class=ccc19}', "");
Expect(1, 1466, '\P{^Is_Canonical_Combining_Class=ccc19}', "");
Expect(0, 1467, '\p{Is_Canonical_Combining_Class=ccc19}', "");
Expect(1, 1467, '\p{^Is_Canonical_Combining_Class=ccc19}', "");
Expect(1, 1467, '\P{Is_Canonical_Combining_Class=ccc19}', "");
Expect(0, 1467, '\P{^Is_Canonical_Combining_Class=ccc19}', "");
Expect(1, 1466, '\p{Is_Canonical_Combining_Class=	CCC19}', "");
Expect(0, 1466, '\p{^Is_Canonical_Combining_Class=	CCC19}', "");
Expect(0, 1466, '\P{Is_Canonical_Combining_Class=	CCC19}', "");
Expect(1, 1466, '\P{^Is_Canonical_Combining_Class=	CCC19}', "");
Expect(0, 1467, '\p{Is_Canonical_Combining_Class=	CCC19}', "");
Expect(1, 1467, '\p{^Is_Canonical_Combining_Class=	CCC19}', "");
Expect(1, 1467, '\P{Is_Canonical_Combining_Class=	CCC19}', "");
Expect(0, 1467, '\P{^Is_Canonical_Combining_Class=	CCC19}', "");
Error('\p{Is_Ccc::=	_00000019}');
Error('\P{Is_Ccc::=	_00000019}');
Expect(1, 1466, '\p{Is_Ccc:	0_0_0_0_0_19}', "");
Expect(0, 1466, '\p{^Is_Ccc:	0_0_0_0_0_19}', "");
Expect(0, 1466, '\P{Is_Ccc:	0_0_0_0_0_19}', "");
Expect(1, 1466, '\P{^Is_Ccc:	0_0_0_0_0_19}', "");
Expect(0, 1467, '\p{Is_Ccc:	0_0_0_0_0_19}', "");
Expect(1, 1467, '\p{^Is_Ccc:	0_0_0_0_0_19}', "");
Expect(1, 1467, '\P{Is_Ccc:	0_0_0_0_0_19}', "");
Expect(0, 1467, '\P{^Is_Ccc:	0_0_0_0_0_19}', "");
Error('\p{Canonical_Combining_Class=:=Kana_VOICING}');
Error('\P{Canonical_Combining_Class=:=Kana_VOICING}');
Expect(1, 12442, '\p{Canonical_Combining_Class=kanavoicing}', "");
Expect(0, 12442, '\p{^Canonical_Combining_Class=kanavoicing}', "");
Expect(0, 12442, '\P{Canonical_Combining_Class=kanavoicing}', "");
Expect(1, 12442, '\P{^Canonical_Combining_Class=kanavoicing}', "");
Expect(0, 12443, '\p{Canonical_Combining_Class=kanavoicing}', "");
Expect(1, 12443, '\p{^Canonical_Combining_Class=kanavoicing}', "");
Expect(1, 12443, '\P{Canonical_Combining_Class=kanavoicing}', "");
Expect(0, 12443, '\P{^Canonical_Combining_Class=kanavoicing}', "");
Expect(1, 12442, '\p{Canonical_Combining_Class= -Kana_VOICING}', "");
Expect(0, 12442, '\p{^Canonical_Combining_Class= -Kana_VOICING}', "");
Expect(0, 12442, '\P{Canonical_Combining_Class= -Kana_VOICING}', "");
Expect(1, 12442, '\P{^Canonical_Combining_Class= -Kana_VOICING}', "");
Expect(0, 12443, '\p{Canonical_Combining_Class= -Kana_VOICING}', "");
Expect(1, 12443, '\p{^Canonical_Combining_Class= -Kana_VOICING}', "");
Expect(1, 12443, '\P{Canonical_Combining_Class= -Kana_VOICING}', "");
Expect(0, 12443, '\P{^Canonical_Combining_Class= -Kana_VOICING}', "");
Error('\p{Ccc=:=KV}');
Error('\P{Ccc=:=KV}');
Expect(1, 12442, '\p{Ccc=kv}', "");
Expect(0, 12442, '\p{^Ccc=kv}', "");
Expect(0, 12442, '\P{Ccc=kv}', "");
Expect(1, 12442, '\P{^Ccc=kv}', "");
Expect(0, 12443, '\p{Ccc=kv}', "");
Expect(1, 12443, '\p{^Ccc=kv}', "");
Expect(1, 12443, '\P{Ccc=kv}', "");
Expect(0, 12443, '\P{^Ccc=kv}', "");
Expect(1, 12442, '\p{Ccc: _KV}', "");
Expect(0, 12442, '\p{^Ccc: _KV}', "");
Expect(0, 12442, '\P{Ccc: _KV}', "");
Expect(1, 12442, '\P{^Ccc: _KV}', "");
Expect(0, 12443, '\p{Ccc: _KV}', "");
Expect(1, 12443, '\p{^Ccc: _KV}', "");
Expect(1, 12443, '\P{Ccc: _KV}', "");
Expect(0, 12443, '\P{^Ccc: _KV}', "");
Error('\p{Is_Canonical_Combining_Class= /a/+0000008}');
Error('\P{Is_Canonical_Combining_Class= /a/+0000008}');
Expect(1, 12442, '\p{Is_Canonical_Combining_Class=000000008}', "");
Expect(0, 12442, '\p{^Is_Canonical_Combining_Class=000000008}', "");
Expect(0, 12442, '\P{Is_Canonical_Combining_Class=000000008}', "");
Expect(1, 12442, '\P{^Is_Canonical_Combining_Class=000000008}', "");
Expect(0, 12443, '\p{Is_Canonical_Combining_Class=000000008}', "");
Expect(1, 12443, '\p{^Is_Canonical_Combining_Class=000000008}', "");
Expect(1, 12443, '\P{Is_Canonical_Combining_Class=000000008}', "");
Expect(0, 12443, '\P{^Is_Canonical_Combining_Class=000000008}', "");
Error('\p{Is_Ccc=-/a/kana_Voicing}');
Error('\P{Is_Ccc=-/a/kana_Voicing}');
Expect(1, 12442, '\p{Is_Ccc=kanavoicing}', "");
Expect(0, 12442, '\p{^Is_Ccc=kanavoicing}', "");
Expect(0, 12442, '\P{Is_Ccc=kanavoicing}', "");
Expect(1, 12442, '\P{^Is_Ccc=kanavoicing}', "");
Expect(0, 12443, '\p{Is_Ccc=kanavoicing}', "");
Expect(1, 12443, '\p{^Is_Ccc=kanavoicing}', "");
Expect(1, 12443, '\P{Is_Ccc=kanavoicing}', "");
Expect(0, 12443, '\P{^Is_Ccc=kanavoicing}', "");
Expect(1, 12442, '\p{Is_Ccc=_	Kana_VOICING}', "");
Expect(0, 12442, '\p{^Is_Ccc=_	Kana_VOICING}', "");
Expect(0, 12442, '\P{Is_Ccc=_	Kana_VOICING}', "");
Expect(1, 12442, '\P{^Is_Ccc=_	Kana_VOICING}', "");
Expect(0, 12443, '\p{Is_Ccc=_	Kana_VOICING}', "");
Expect(1, 12443, '\p{^Is_Ccc=_	Kana_VOICING}', "");
Expect(1, 12443, '\P{Is_Ccc=_	Kana_VOICING}', "");
Expect(0, 12443, '\P{^Is_Ccc=_	Kana_VOICING}', "");
Error('\p{Canonical_Combining_Class: :=-BELOW}');
Error('\P{Canonical_Combining_Class: :=-BELOW}');
Expect(1, 119179, '\p{Canonical_Combining_Class=below}', "");
Expect(0, 119179, '\p{^Canonical_Combining_Class=below}', "");
Expect(0, 119179, '\P{Canonical_Combining_Class=below}', "");
Expect(1, 119179, '\P{^Canonical_Combining_Class=below}', "");
Expect(0, 119180, '\p{Canonical_Combining_Class=below}', "");
Expect(1, 119180, '\p{^Canonical_Combining_Class=below}', "");
Expect(1, 119180, '\P{Canonical_Combining_Class=below}', "");
Expect(0, 119180, '\P{^Canonical_Combining_Class=below}', "");
Expect(1, 119179, '\p{Canonical_Combining_Class= 	below}', "");
Expect(0, 119179, '\p{^Canonical_Combining_Class= 	below}', "");
Expect(0, 119179, '\P{Canonical_Combining_Class= 	below}', "");
Expect(1, 119179, '\P{^Canonical_Combining_Class= 	below}', "");
Expect(0, 119180, '\p{Canonical_Combining_Class= 	below}', "");
Expect(1, 119180, '\p{^Canonical_Combining_Class= 	below}', "");
Expect(1, 119180, '\P{Canonical_Combining_Class= 	below}', "");
Expect(0, 119180, '\P{^Canonical_Combining_Class= 	below}', "");
Error('\p{Ccc=/a/_-B}');
Error('\P{Ccc=/a/_-B}');
Expect(1, 119179, '\p{Ccc=b}', "");
Expect(0, 119179, '\p{^Ccc=b}', "");
Expect(0, 119179, '\P{Ccc=b}', "");
Expect(1, 119179, '\P{^Ccc=b}', "");
Expect(0, 119180, '\p{Ccc=b}', "");
Expect(1, 119180, '\p{^Ccc=b}', "");
Expect(1, 119180, '\P{Ccc=b}', "");
Expect(0, 119180, '\P{^Ccc=b}', "");
Error('\p{Is_Canonical_Combining_Class: /a/	_+0000220}');
Error('\P{Is_Canonical_Combining_Class: /a/	_+0000220}');
Expect(1, 119179, '\p{Is_Canonical_Combining_Class: 00000000220}', "");
Expect(0, 119179, '\p{^Is_Canonical_Combining_Class: 00000000220}', "");
Expect(0, 119179, '\P{Is_Canonical_Combining_Class: 00000000220}', "");
Expect(1, 119179, '\P{^Is_Canonical_Combining_Class: 00000000220}', "");
Expect(0, 119180, '\p{Is_Canonical_Combining_Class: 00000000220}', "");
Expect(1, 119180, '\p{^Is_Canonical_Combining_Class: 00000000220}', "");
Expect(1, 119180, '\P{Is_Canonical_Combining_Class: 00000000220}', "");
Expect(0, 119180, '\P{^Is_Canonical_Combining_Class: 00000000220}', "");
Error('\p{Is_Ccc=/a/ BELOW}');
Error('\P{Is_Ccc=/a/ BELOW}');
Expect(1, 119179, '\p{Is_Ccc=below}', "");
Expect(0, 119179, '\p{^Is_Ccc=below}', "");
Expect(0, 119179, '\P{Is_Ccc=below}', "");
Expect(1, 119179, '\P{^Is_Ccc=below}', "");
Expect(0, 119180, '\p{Is_Ccc=below}', "");
Expect(1, 119180, '\p{^Is_Ccc=below}', "");
Expect(1, 119180, '\P{Is_Ccc=below}', "");
Expect(0, 119180, '\P{^Is_Ccc=below}', "");
Expect(1, 119179, '\p{Is_Ccc=	Below}', "");
Expect(0, 119179, '\p{^Is_Ccc=	Below}', "");
Expect(0, 119179, '\P{Is_Ccc=	Below}', "");
Expect(1, 119179, '\P{^Is_Ccc=	Below}', "");
Expect(0, 119180, '\p{Is_Ccc=	Below}', "");
Expect(1, 119180, '\p{^Is_Ccc=	Below}', "");
Expect(1, 119180, '\P{Is_Ccc=	Below}', "");
Expect(0, 119180, '\P{^Is_Ccc=	Below}', "");
Error('\p{Canonical_Combining_Class=-:=ccc20}');
Error('\P{Canonical_Combining_Class=-:=ccc20}');
Expect(1, 1467, '\p{Canonical_Combining_Class=ccc20}', "");
Expect(0, 1467, '\p{^Canonical_Combining_Class=ccc20}', "");
Expect(0, 1467, '\P{Canonical_Combining_Class=ccc20}', "");
Expect(1, 1467, '\P{^Canonical_Combining_Class=ccc20}', "");
Expect(0, 1468, '\p{Canonical_Combining_Class=ccc20}', "");
Expect(1, 1468, '\p{^Canonical_Combining_Class=ccc20}', "");
Expect(1, 1468, '\P{Canonical_Combining_Class=ccc20}', "");
Expect(0, 1468, '\P{^Canonical_Combining_Class=ccc20}', "");
Expect(1, 1467, '\p{Canonical_Combining_Class=_CCC20}', "");
Expect(0, 1467, '\p{^Canonical_Combining_Class=_CCC20}', "");
Expect(0, 1467, '\P{Canonical_Combining_Class=_CCC20}', "");
Expect(1, 1467, '\P{^Canonical_Combining_Class=_CCC20}', "");
Expect(0, 1468, '\p{Canonical_Combining_Class=_CCC20}', "");
Expect(1, 1468, '\p{^Canonical_Combining_Class=_CCC20}', "");
Expect(1, 1468, '\P{Canonical_Combining_Class=_CCC20}', "");
Expect(0, 1468, '\P{^Canonical_Combining_Class=_CCC20}', "");
Error('\p{Ccc=	00020/a/}');
Error('\P{Ccc=	00020/a/}');
Expect(1, 1467, '\p{Ccc=0000002_0}', "");
Expect(0, 1467, '\p{^Ccc=0000002_0}', "");
Expect(0, 1467, '\P{Ccc=0000002_0}', "");
Expect(1, 1467, '\P{^Ccc=0000002_0}', "");
Expect(0, 1468, '\p{Ccc=0000002_0}', "");
Expect(1, 1468, '\p{^Ccc=0000002_0}', "");
Expect(1, 1468, '\P{Ccc=0000002_0}', "");
Expect(0, 1468, '\P{^Ccc=0000002_0}', "");
Error('\p{Is_Canonical_Combining_Class=:=__ccc20}');
Error('\P{Is_Canonical_Combining_Class=:=__ccc20}');
Expect(1, 1467, '\p{Is_Canonical_Combining_Class:	ccc20}', "");
Expect(0, 1467, '\p{^Is_Canonical_Combining_Class:	ccc20}', "");
Expect(0, 1467, '\P{Is_Canonical_Combining_Class:	ccc20}', "");
Expect(1, 1467, '\P{^Is_Canonical_Combining_Class:	ccc20}', "");
Expect(0, 1468, '\p{Is_Canonical_Combining_Class:	ccc20}', "");
Expect(1, 1468, '\p{^Is_Canonical_Combining_Class:	ccc20}', "");
Expect(1, 1468, '\P{Is_Canonical_Combining_Class:	ccc20}', "");
Expect(0, 1468, '\P{^Is_Canonical_Combining_Class:	ccc20}', "");
Expect(1, 1467, '\p{Is_Canonical_Combining_Class:-	CCC20}', "");
Expect(0, 1467, '\p{^Is_Canonical_Combining_Class:-	CCC20}', "");
Expect(0, 1467, '\P{Is_Canonical_Combining_Class:-	CCC20}', "");
Expect(1, 1467, '\P{^Is_Canonical_Combining_Class:-	CCC20}', "");
Expect(0, 1468, '\p{Is_Canonical_Combining_Class:-	CCC20}', "");
Expect(1, 1468, '\p{^Is_Canonical_Combining_Class:-	CCC20}', "");
Expect(1, 1468, '\P{Is_Canonical_Combining_Class:-	CCC20}', "");
Expect(0, 1468, '\P{^Is_Canonical_Combining_Class:-	CCC20}', "");
Error('\p{Is_Ccc: _0000000020:=}');
Error('\P{Is_Ccc: _0000000020:=}');
Expect(1, 1467, '\p{Is_Ccc=0000020}', "");
Expect(0, 1467, '\p{^Is_Ccc=0000020}', "");
Expect(0, 1467, '\P{Is_Ccc=0000020}', "");
Expect(1, 1467, '\P{^Is_Ccc=0000020}', "");
Expect(0, 1468, '\p{Is_Ccc=0000020}', "");
Expect(1, 1468, '\p{^Is_Ccc=0000020}', "");
Expect(1, 1468, '\P{Is_Ccc=0000020}', "");
Expect(0, 1468, '\P{^Is_Ccc=0000020}', "");
Error('\p{Canonical_Combining_Class=-_CCC13:=}');
Error('\P{Canonical_Combining_Class=-_CCC13:=}');
Expect(1, 1459, '\p{Canonical_Combining_Class=ccc13}', "");
Expect(0, 1459, '\p{^Canonical_Combining_Class=ccc13}', "");
Expect(0, 1459, '\P{Canonical_Combining_Class=ccc13}', "");
Expect(1, 1459, '\P{^Canonical_Combining_Class=ccc13}', "");
Expect(0, 1460, '\p{Canonical_Combining_Class=ccc13}', "");
Expect(1, 1460, '\p{^Canonical_Combining_Class=ccc13}', "");
Expect(1, 1460, '\P{Canonical_Combining_Class=ccc13}', "");
Expect(0, 1460, '\P{^Canonical_Combining_Class=ccc13}', "");
Expect(1, 1459, '\p{Canonical_Combining_Class=	CCC13}', "");
Expect(0, 1459, '\p{^Canonical_Combining_Class=	CCC13}', "");
Expect(0, 1459, '\P{Canonical_Combining_Class=	CCC13}', "");
Expect(1, 1459, '\P{^Canonical_Combining_Class=	CCC13}', "");
Expect(0, 1460, '\p{Canonical_Combining_Class=	CCC13}', "");
Expect(1, 1460, '\p{^Canonical_Combining_Class=	CCC13}', "");
Expect(1, 1460, '\P{Canonical_Combining_Class=	CCC13}', "");
Expect(0, 1460, '\P{^Canonical_Combining_Class=	CCC13}', "");
Error('\p{Ccc=-/a/0000013}');
Error('\P{Ccc=-/a/0000013}');
Expect(1, 1459, '\p{Ccc=00000000013}', "");
Expect(0, 1459, '\p{^Ccc=00000000013}', "");
Expect(0, 1459, '\P{Ccc=00000000013}', "");
Expect(1, 1459, '\P{^Ccc=00000000013}', "");
Expect(0, 1460, '\p{Ccc=00000000013}', "");
Expect(1, 1460, '\p{^Ccc=00000000013}', "");
Expect(1, 1460, '\P{Ccc=00000000013}', "");
Expect(0, 1460, '\P{^Ccc=00000000013}', "");
Error('\p{Is_Canonical_Combining_Class=/a/ _CCC13}');
Error('\P{Is_Canonical_Combining_Class=/a/ _CCC13}');
Expect(1, 1459, '\p{Is_Canonical_Combining_Class=ccc13}', "");
Expect(0, 1459, '\p{^Is_Canonical_Combining_Class=ccc13}', "");
Expect(0, 1459, '\P{Is_Canonical_Combining_Class=ccc13}', "");
Expect(1, 1459, '\P{^Is_Canonical_Combining_Class=ccc13}', "");
Expect(0, 1460, '\p{Is_Canonical_Combining_Class=ccc13}', "");
Expect(1, 1460, '\p{^Is_Canonical_Combining_Class=ccc13}', "");
Expect(1, 1460, '\P{Is_Canonical_Combining_Class=ccc13}', "");
Expect(0, 1460, '\P{^Is_Canonical_Combining_Class=ccc13}', "");
Expect(1, 1459, '\p{Is_Canonical_Combining_Class= -CCC13}', "");
Expect(0, 1459, '\p{^Is_Canonical_Combining_Class= -CCC13}', "");
Expect(0, 1459, '\P{Is_Canonical_Combining_Class= -CCC13}', "");
Expect(1, 1459, '\P{^Is_Canonical_Combining_Class= -CCC13}', "");
Expect(0, 1460, '\p{Is_Canonical_Combining_Class= -CCC13}', "");
Expect(1, 1460, '\p{^Is_Canonical_Combining_Class= -CCC13}', "");
Expect(1, 1460, '\P{Is_Canonical_Combining_Class= -CCC13}', "");
Expect(0, 1460, '\P{^Is_Canonical_Combining_Class= -CCC13}', "");
Error('\p{Is_Ccc=:=	-+0_0_0_0_0_0_00013}');
Error('\P{Is_Ccc=:=	-+0_0_0_0_0_0_00013}');
Expect(1, 1459, '\p{Is_Ccc=0001_3}', "");
Expect(0, 1459, '\p{^Is_Ccc=0001_3}', "");
Expect(0, 1459, '\P{Is_Ccc=0001_3}', "");
Expect(1, 1459, '\P{^Is_Ccc=0001_3}', "");
Expect(0, 1460, '\p{Is_Ccc=0001_3}', "");
Expect(1, 1460, '\p{^Is_Ccc=0001_3}', "");
Expect(1, 1460, '\P{Is_Ccc=0001_3}', "");
Expect(0, 1460, '\P{^Is_Ccc=0001_3}', "");
Error('\p{Canonical_Combining_Class= 	CCC91:=}');
Error('\P{Canonical_Combining_Class= 	CCC91:=}');
Expect(1, 3158, '\p{Canonical_Combining_Class:   ccc91}', "");
Expect(0, 3158, '\p{^Canonical_Combining_Class:   ccc91}', "");
Expect(0, 3158, '\P{Canonical_Combining_Class:   ccc91}', "");
Expect(1, 3158, '\P{^Canonical_Combining_Class:   ccc91}', "");
Expect(0, 3159, '\p{Canonical_Combining_Class:   ccc91}', "");
Expect(1, 3159, '\p{^Canonical_Combining_Class:   ccc91}', "");
Expect(1, 3159, '\P{Canonical_Combining_Class:   ccc91}', "");
Expect(0, 3159, '\P{^Canonical_Combining_Class:   ccc91}', "");
Expect(1, 3158, '\p{Canonical_Combining_Class= 	CCC91}', "");
Expect(0, 3158, '\p{^Canonical_Combining_Class= 	CCC91}', "");
Expect(0, 3158, '\P{Canonical_Combining_Class= 	CCC91}', "");
Expect(1, 3158, '\P{^Canonical_Combining_Class= 	CCC91}', "");
Expect(0, 3159, '\p{Canonical_Combining_Class= 	CCC91}', "");
Expect(1, 3159, '\p{^Canonical_Combining_Class= 	CCC91}', "");
Expect(1, 3159, '\P{Canonical_Combining_Class= 	CCC91}', "");
Expect(0, 3159, '\P{^Canonical_Combining_Class= 	CCC91}', "");
Error('\p{Ccc=_+91/a/}');
Error('\P{Ccc=_+91/a/}');
Expect(1, 3158, '\p{Ccc=009_1}', "");
Expect(0, 3158, '\p{^Ccc=009_1}', "");
Expect(0, 3158, '\P{Ccc=009_1}', "");
Expect(1, 3158, '\P{^Ccc=009_1}', "");
Expect(0, 3159, '\p{Ccc=009_1}', "");
Expect(1, 3159, '\p{^Ccc=009_1}', "");
Expect(1, 3159, '\P{Ccc=009_1}', "");
Expect(0, 3159, '\P{^Ccc=009_1}', "");
Error('\p{Is_Canonical_Combining_Class=_/a/CCC91}');
Error('\P{Is_Canonical_Combining_Class=_/a/CCC91}');
Expect(1, 3158, '\p{Is_Canonical_Combining_Class=ccc91}', "");
Expect(0, 3158, '\p{^Is_Canonical_Combining_Class=ccc91}', "");
Expect(0, 3158, '\P{Is_Canonical_Combining_Class=ccc91}', "");
Expect(1, 3158, '\P{^Is_Canonical_Combining_Class=ccc91}', "");
Expect(0, 3159, '\p{Is_Canonical_Combining_Class=ccc91}', "");
Expect(1, 3159, '\p{^Is_Canonical_Combining_Class=ccc91}', "");
Expect(1, 3159, '\P{Is_Canonical_Combining_Class=ccc91}', "");
Expect(0, 3159, '\P{^Is_Canonical_Combining_Class=ccc91}', "");
Expect(1, 3158, '\p{Is_Canonical_Combining_Class:	_ CCC91}', "");
Expect(0, 3158, '\p{^Is_Canonical_Combining_Class:	_ CCC91}', "");
Expect(0, 3158, '\P{Is_Canonical_Combining_Class:	_ CCC91}', "");
Expect(1, 3158, '\P{^Is_Canonical_Combining_Class:	_ CCC91}', "");
Expect(0, 3159, '\p{Is_Canonical_Combining_Class:	_ CCC91}', "");
Expect(1, 3159, '\p{^Is_Canonical_Combining_Class:	_ CCC91}', "");
Expect(1, 3159, '\P{Is_Canonical_Combining_Class:	_ CCC91}', "");
Expect(0, 3159, '\P{^Is_Canonical_Combining_Class:	_ CCC91}', "");
Error('\p{Is_Ccc:	 /a/0000091}');
Error('\P{Is_Ccc:	 /a/0000091}');
Expect(1, 3158, '\p{Is_Ccc:0000009_1}', "");
Expect(0, 3158, '\p{^Is_Ccc:0000009_1}', "");
Expect(0, 3158, '\P{Is_Ccc:0000009_1}', "");
Expect(1, 3158, '\P{^Is_Ccc:0000009_1}', "");
Expect(0, 3159, '\p{Is_Ccc:0000009_1}', "");
Expect(1, 3159, '\p{^Is_Ccc:0000009_1}', "");
Expect(1, 3159, '\P{Is_Ccc:0000009_1}', "");
Expect(0, 3159, '\P{^Is_Ccc:0000009_1}', "");
Error('\p{Canonical_Combining_Class=:=-ccc11}');
Error('\P{Canonical_Combining_Class=:=-ccc11}');
Expect(1, 1457, '\p{Canonical_Combining_Class=ccc11}', "");
Expect(0, 1457, '\p{^Canonical_Combining_Class=ccc11}', "");
Expect(0, 1457, '\P{Canonical_Combining_Class=ccc11}', "");
Expect(1, 1457, '\P{^Canonical_Combining_Class=ccc11}', "");
Expect(0, 1458, '\p{Canonical_Combining_Class=ccc11}', "");
Expect(1, 1458, '\p{^Canonical_Combining_Class=ccc11}', "");
Expect(1, 1458, '\P{Canonical_Combining_Class=ccc11}', "");
Expect(0, 1458, '\P{^Canonical_Combining_Class=ccc11}', "");
Expect(1, 1457, '\p{Canonical_Combining_Class=__CCC11}', "");
Expect(0, 1457, '\p{^Canonical_Combining_Class=__CCC11}', "");
Expect(0, 1457, '\P{Canonical_Combining_Class=__CCC11}', "");
Expect(1, 1457, '\P{^Canonical_Combining_Class=__CCC11}', "");
Expect(0, 1458, '\p{Canonical_Combining_Class=__CCC11}', "");
Expect(1, 1458, '\p{^Canonical_Combining_Class=__CCC11}', "");
Expect(1, 1458, '\P{Canonical_Combining_Class=__CCC11}', "");
Expect(0, 1458, '\P{^Canonical_Combining_Class=__CCC11}', "");
Error('\p{Ccc=+00000011/a/}');
Error('\P{Ccc=+00000011/a/}');
Expect(1, 1457, '\p{Ccc=0_0_0_0_0_11}', "");
Expect(0, 1457, '\p{^Ccc=0_0_0_0_0_11}', "");
Expect(0, 1457, '\P{Ccc=0_0_0_0_0_11}', "");
Expect(1, 1457, '\P{^Ccc=0_0_0_0_0_11}', "");
Expect(0, 1458, '\p{Ccc=0_0_0_0_0_11}', "");
Expect(1, 1458, '\p{^Ccc=0_0_0_0_0_11}', "");
Expect(1, 1458, '\P{Ccc=0_0_0_0_0_11}', "");
Expect(0, 1458, '\P{^Ccc=0_0_0_0_0_11}', "");
Error('\p{Is_Canonical_Combining_Class=	CCC11:=}');
Error('\P{Is_Canonical_Combining_Class=	CCC11:=}');
Expect(1, 1457, '\p{Is_Canonical_Combining_Class: ccc11}', "");
Expect(0, 1457, '\p{^Is_Canonical_Combining_Class: ccc11}', "");
Expect(0, 1457, '\P{Is_Canonical_Combining_Class: ccc11}', "");
Expect(1, 1457, '\P{^Is_Canonical_Combining_Class: ccc11}', "");
Expect(0, 1458, '\p{Is_Canonical_Combining_Class: ccc11}', "");
Expect(1, 1458, '\p{^Is_Canonical_Combining_Class: ccc11}', "");
Expect(1, 1458, '\P{Is_Canonical_Combining_Class: ccc11}', "");
Expect(0, 1458, '\P{^Is_Canonical_Combining_Class: ccc11}', "");
Expect(1, 1457, '\p{Is_Canonical_Combining_Class=--ccc11}', "");
Expect(0, 1457, '\p{^Is_Canonical_Combining_Class=--ccc11}', "");
Expect(0, 1457, '\P{Is_Canonical_Combining_Class=--ccc11}', "");
Expect(1, 1457, '\P{^Is_Canonical_Combining_Class=--ccc11}', "");
Expect(0, 1458, '\p{Is_Canonical_Combining_Class=--ccc11}', "");
Expect(1, 1458, '\p{^Is_Canonical_Combining_Class=--ccc11}', "");
Expect(1, 1458, '\P{Is_Canonical_Combining_Class=--ccc11}', "");
Expect(0, 1458, '\P{^Is_Canonical_Combining_Class=--ccc11}', "");
Error('\p{Is_Ccc=	:=11}');
Error('\P{Is_Ccc=	:=11}');
Expect(1, 1457, '\p{Is_Ccc=0000011}', "");
Expect(0, 1457, '\p{^Is_Ccc=0000011}', "");
Expect(0, 1457, '\P{Is_Ccc=0000011}', "");
Expect(1, 1457, '\P{^Is_Ccc=0000011}', "");
Expect(0, 1458, '\p{Is_Ccc=0000011}', "");
Expect(1, 1458, '\p{^Is_Ccc=0000011}', "");
Expect(1, 1458, '\P{Is_Ccc=0000011}', "");
Expect(0, 1458, '\P{^Is_Ccc=0000011}', "");
Error('\p{Canonical_Combining_Class= /a/ATTACHED_Above_Right}');
Error('\P{Canonical_Combining_Class= /a/ATTACHED_Above_Right}');
Expect(1, 119154, '\p{Canonical_Combining_Class=attachedaboveright}', "");
Expect(0, 119154, '\p{^Canonical_Combining_Class=attachedaboveright}', "");
Expect(0, 119154, '\P{Canonical_Combining_Class=attachedaboveright}', "");
Expect(1, 119154, '\P{^Canonical_Combining_Class=attachedaboveright}', "");
Expect(0, 119155, '\p{Canonical_Combining_Class=attachedaboveright}', "");
Expect(1, 119155, '\p{^Canonical_Combining_Class=attachedaboveright}', "");
Expect(1, 119155, '\P{Canonical_Combining_Class=attachedaboveright}', "");
Expect(0, 119155, '\P{^Canonical_Combining_Class=attachedaboveright}', "");
Expect(1, 119154, '\p{Canonical_Combining_Class=	ATTACHED_above_Right}', "");
Expect(0, 119154, '\p{^Canonical_Combining_Class=	ATTACHED_above_Right}', "");
Expect(0, 119154, '\P{Canonical_Combining_Class=	ATTACHED_above_Right}', "");
Expect(1, 119154, '\P{^Canonical_Combining_Class=	ATTACHED_above_Right}', "");
Expect(0, 119155, '\p{Canonical_Combining_Class=	ATTACHED_above_Right}', "");
Expect(1, 119155, '\p{^Canonical_Combining_Class=	ATTACHED_above_Right}', "");
Expect(1, 119155, '\P{Canonical_Combining_Class=	ATTACHED_above_Right}', "");
Expect(0, 119155, '\P{^Canonical_Combining_Class=	ATTACHED_above_Right}', "");
Error('\p{Ccc=  ATAR/a/}');
Error('\P{Ccc=  ATAR/a/}');
Expect(1, 119154, '\p{Ccc=atar}', "");
Expect(0, 119154, '\p{^Ccc=atar}', "");
Expect(0, 119154, '\P{Ccc=atar}', "");
Expect(1, 119154, '\P{^Ccc=atar}', "");
Expect(0, 119155, '\p{Ccc=atar}', "");
Expect(1, 119155, '\p{^Ccc=atar}', "");
Expect(1, 119155, '\P{Ccc=atar}', "");
Expect(0, 119155, '\P{^Ccc=atar}', "");
Expect(1, 119154, '\p{Ccc=_-atar}', "");
Expect(0, 119154, '\p{^Ccc=_-atar}', "");
Expect(0, 119154, '\P{Ccc=_-atar}', "");
Expect(1, 119154, '\P{^Ccc=_-atar}', "");
Expect(0, 119155, '\p{Ccc=_-atar}', "");
Expect(1, 119155, '\p{^Ccc=_-atar}', "");
Expect(1, 119155, '\P{Ccc=_-atar}', "");
Expect(0, 119155, '\P{^Ccc=_-atar}', "");
Error('\p{Is_Canonical_Combining_Class:/a/-_0_0_0_0_0_216}');
Error('\P{Is_Canonical_Combining_Class:/a/-_0_0_0_0_0_216}');
Expect(1, 119154, '\p{Is_Canonical_Combining_Class=00_21_6}', "");
Expect(0, 119154, '\p{^Is_Canonical_Combining_Class=00_21_6}', "");
Expect(0, 119154, '\P{Is_Canonical_Combining_Class=00_21_6}', "");
Expect(1, 119154, '\P{^Is_Canonical_Combining_Class=00_21_6}', "");
Expect(0, 119155, '\p{Is_Canonical_Combining_Class=00_21_6}', "");
Expect(1, 119155, '\p{^Is_Canonical_Combining_Class=00_21_6}', "");
Expect(1, 119155, '\P{Is_Canonical_Combining_Class=00_21_6}', "");
Expect(0, 119155, '\P{^Is_Canonical_Combining_Class=00_21_6}', "");
Error('\p{Is_Ccc=__ATTACHED_above_Right/a/}');
Error('\P{Is_Ccc=__ATTACHED_above_Right/a/}');
Expect(1, 119154, '\p{Is_Ccc=attachedaboveright}', "");
Expect(0, 119154, '\p{^Is_Ccc=attachedaboveright}', "");
Expect(0, 119154, '\P{Is_Ccc=attachedaboveright}', "");
Expect(1, 119154, '\P{^Is_Ccc=attachedaboveright}', "");
Expect(0, 119155, '\p{Is_Ccc=attachedaboveright}', "");
Expect(1, 119155, '\p{^Is_Ccc=attachedaboveright}', "");
Expect(1, 119155, '\P{Is_Ccc=attachedaboveright}', "");
Expect(0, 119155, '\P{^Is_Ccc=attachedaboveright}', "");
Expect(1, 119154, '\p{Is_Ccc=_ ATTACHED_Above_Right}', "");
Expect(0, 119154, '\p{^Is_Ccc=_ ATTACHED_Above_Right}', "");
Expect(0, 119154, '\P{Is_Ccc=_ ATTACHED_Above_Right}', "");
Expect(1, 119154, '\P{^Is_Ccc=_ ATTACHED_Above_Right}', "");
Expect(0, 119155, '\p{Is_Ccc=_ ATTACHED_Above_Right}', "");
Expect(1, 119155, '\p{^Is_Ccc=_ ATTACHED_Above_Right}', "");
Expect(1, 119155, '\P{Is_Ccc=_ ATTACHED_Above_Right}', "");
Expect(0, 119155, '\P{^Is_Ccc=_ ATTACHED_Above_Right}', "");
Error('\p{Canonical_Combining_Class=_:=ccc103}');
Error('\P{Canonical_Combining_Class=_:=ccc103}');
Expect(1, 3641, '\p{Canonical_Combining_Class=ccc103}', "");
Expect(0, 3641, '\p{^Canonical_Combining_Class=ccc103}', "");
Expect(0, 3641, '\P{Canonical_Combining_Class=ccc103}', "");
Expect(1, 3641, '\P{^Canonical_Combining_Class=ccc103}', "");
Expect(0, 3642, '\p{Canonical_Combining_Class=ccc103}', "");
Expect(1, 3642, '\p{^Canonical_Combining_Class=ccc103}', "");
Expect(1, 3642, '\P{Canonical_Combining_Class=ccc103}', "");
Expect(0, 3642, '\P{^Canonical_Combining_Class=ccc103}', "");
Expect(1, 3641, '\p{Canonical_Combining_Class=_-CCC103}', "");
Expect(0, 3641, '\p{^Canonical_Combining_Class=_-CCC103}', "");
Expect(0, 3641, '\P{Canonical_Combining_Class=_-CCC103}', "");
Expect(1, 3641, '\P{^Canonical_Combining_Class=_-CCC103}', "");
Expect(0, 3642, '\p{Canonical_Combining_Class=_-CCC103}', "");
Expect(1, 3642, '\p{^Canonical_Combining_Class=_-CCC103}', "");
Expect(1, 3642, '\P{Canonical_Combining_Class=_-CCC103}', "");
Expect(0, 3642, '\P{^Canonical_Combining_Class=_-CCC103}', "");
Error('\p{Ccc:	:=_ 00_00_01_03}');
Error('\P{Ccc:	:=_ 00_00_01_03}');
Expect(1, 3641, '\p{Ccc:	000000103}', "");
Expect(0, 3641, '\p{^Ccc:	000000103}', "");
Expect(0, 3641, '\P{Ccc:	000000103}', "");
Expect(1, 3641, '\P{^Ccc:	000000103}', "");
Expect(0, 3642, '\p{Ccc:	000000103}', "");
Expect(1, 3642, '\p{^Ccc:	000000103}', "");
Expect(1, 3642, '\P{Ccc:	000000103}', "");
Expect(0, 3642, '\P{^Ccc:	000000103}', "");
Error('\p{Is_Canonical_Combining_Class=:=ccc103}');
Error('\P{Is_Canonical_Combining_Class=:=ccc103}');
Expect(1, 3641, '\p{Is_Canonical_Combining_Class=ccc103}', "");
Expect(0, 3641, '\p{^Is_Canonical_Combining_Class=ccc103}', "");
Expect(0, 3641, '\P{Is_Canonical_Combining_Class=ccc103}', "");
Expect(1, 3641, '\P{^Is_Canonical_Combining_Class=ccc103}', "");
Expect(0, 3642, '\p{Is_Canonical_Combining_Class=ccc103}', "");
Expect(1, 3642, '\p{^Is_Canonical_Combining_Class=ccc103}', "");
Expect(1, 3642, '\P{Is_Canonical_Combining_Class=ccc103}', "");
Expect(0, 3642, '\P{^Is_Canonical_Combining_Class=ccc103}', "");
Expect(1, 3641, '\p{Is_Canonical_Combining_Class=-CCC103}', "");
Expect(0, 3641, '\p{^Is_Canonical_Combining_Class=-CCC103}', "");
Expect(0, 3641, '\P{Is_Canonical_Combining_Class=-CCC103}', "");
Expect(1, 3641, '\P{^Is_Canonical_Combining_Class=-CCC103}', "");
Expect(0, 3642, '\p{Is_Canonical_Combining_Class=-CCC103}', "");
Expect(1, 3642, '\p{^Is_Canonical_Combining_Class=-CCC103}', "");
Expect(1, 3642, '\P{Is_Canonical_Combining_Class=-CCC103}', "");
Expect(0, 3642, '\P{^Is_Canonical_Combining_Class=-CCC103}', "");
Error('\p{Is_Ccc=/a/+0_0_0_103}');
Error('\P{Is_Ccc=/a/+0_0_0_103}');
Expect(1, 3641, '\p{Is_Ccc=000000103}', "");
Expect(0, 3641, '\p{^Is_Ccc=000000103}', "");
Expect(0, 3641, '\P{Is_Ccc=000000103}', "");
Expect(1, 3641, '\P{^Is_Ccc=000000103}', "");
Expect(0, 3642, '\p{Is_Ccc=000000103}', "");
Expect(1, 3642, '\p{^Is_Ccc=000000103}', "");
Expect(1, 3642, '\P{Is_Ccc=000000103}', "");
Expect(0, 3642, '\P{^Is_Ccc=000000103}', "");
Error('\p{Canonical_Combining_Class=:=-CCC29}');
Error('\P{Canonical_Combining_Class=:=-CCC29}');
Expect(1, 2290, '\p{Canonical_Combining_Class=ccc29}', "");
Expect(0, 2290, '\p{^Canonical_Combining_Class=ccc29}', "");
Expect(0, 2290, '\P{Canonical_Combining_Class=ccc29}', "");
Expect(1, 2290, '\P{^Canonical_Combining_Class=ccc29}', "");
Expect(0, 2291, '\p{Canonical_Combining_Class=ccc29}', "");
Expect(1, 2291, '\p{^Canonical_Combining_Class=ccc29}', "");
Expect(1, 2291, '\P{Canonical_Combining_Class=ccc29}', "");
Expect(0, 2291, '\P{^Canonical_Combining_Class=ccc29}', "");
Expect(1, 2290, '\p{Canonical_Combining_Class=_CCC29}', "");
Expect(0, 2290, '\p{^Canonical_Combining_Class=_CCC29}', "");
Expect(0, 2290, '\P{Canonical_Combining_Class=_CCC29}', "");
Expect(1, 2290, '\P{^Canonical_Combining_Class=_CCC29}', "");
Expect(0, 2291, '\p{Canonical_Combining_Class=_CCC29}', "");
Expect(1, 2291, '\p{^Canonical_Combining_Class=_CCC29}', "");
Expect(1, 2291, '\P{Canonical_Combining_Class=_CCC29}', "");
Expect(0, 2291, '\P{^Canonical_Combining_Class=_CCC29}', "");
Error('\p{Ccc=:=_	+00029}');
Error('\P{Ccc=:=_	+00029}');
Expect(1, 2290, '\p{Ccc=+0000000029}', "");
Expect(0, 2290, '\p{^Ccc=+0000000029}', "");
Expect(0, 2290, '\P{Ccc=+0000000029}', "");
Expect(1, 2290, '\P{^Ccc=+0000000029}', "");
Expect(0, 2291, '\p{Ccc=+0000000029}', "");
Expect(1, 2291, '\p{^Ccc=+0000000029}', "");
Expect(1, 2291, '\P{Ccc=+0000000029}', "");
Expect(0, 2291, '\P{^Ccc=+0000000029}', "");
Error('\p{Is_Canonical_Combining_Class=/a/ _CCC29}');
Error('\P{Is_Canonical_Combining_Class=/a/ _CCC29}');
Expect(1, 2290, '\p{Is_Canonical_Combining_Class=ccc29}', "");
Expect(0, 2290, '\p{^Is_Canonical_Combining_Class=ccc29}', "");
Expect(0, 2290, '\P{Is_Canonical_Combining_Class=ccc29}', "");
Expect(1, 2290, '\P{^Is_Canonical_Combining_Class=ccc29}', "");
Expect(0, 2291, '\p{Is_Canonical_Combining_Class=ccc29}', "");
Expect(1, 2291, '\p{^Is_Canonical_Combining_Class=ccc29}', "");
Expect(1, 2291, '\P{Is_Canonical_Combining_Class=ccc29}', "");
Expect(0, 2291, '\P{^Is_Canonical_Combining_Class=ccc29}', "");
Expect(1, 2290, '\p{Is_Canonical_Combining_Class= _CCC29}', "");
Expect(0, 2290, '\p{^Is_Canonical_Combining_Class= _CCC29}', "");
Expect(0, 2290, '\P{Is_Canonical_Combining_Class= _CCC29}', "");
Expect(1, 2290, '\P{^Is_Canonical_Combining_Class= _CCC29}', "");
Expect(0, 2291, '\p{Is_Canonical_Combining_Class= _CCC29}', "");
Expect(1, 2291, '\p{^Is_Canonical_Combining_Class= _CCC29}', "");
Expect(1, 2291, '\P{Is_Canonical_Combining_Class= _CCC29}', "");
Expect(0, 2291, '\P{^Is_Canonical_Combining_Class= _CCC29}', "");
Error('\p{Is_Ccc=	:=+000029}');
Error('\P{Is_Ccc=	:=+000029}');
Expect(1, 2290, '\p{Is_Ccc=000000029}', "");
Expect(0, 2290, '\p{^Is_Ccc=000000029}', "");
Expect(0, 2290, '\P{Is_Ccc=000000029}', "");
Expect(1, 2290, '\P{^Is_Ccc=000000029}', "");
Expect(0, 2291, '\p{Is_Ccc=000000029}', "");
Expect(1, 2291, '\p{^Is_Ccc=000000029}', "");
Expect(1, 2291, '\P{Is_Ccc=000000029}', "");
Expect(0, 2291, '\P{^Is_Ccc=000000029}', "");
Error('\p{Canonical_Combining_Class= :=CCC31}');
Error('\P{Canonical_Combining_Class= :=CCC31}');
Expect(1, 1615, '\p{Canonical_Combining_Class=ccc31}', "");
Expect(0, 1615, '\p{^Canonical_Combining_Class=ccc31}', "");
Expect(0, 1615, '\P{Canonical_Combining_Class=ccc31}', "");
Expect(1, 1615, '\P{^Canonical_Combining_Class=ccc31}', "");
Expect(0, 1616, '\p{Canonical_Combining_Class=ccc31}', "");
Expect(1, 1616, '\p{^Canonical_Combining_Class=ccc31}', "");
Expect(1, 1616, '\P{Canonical_Combining_Class=ccc31}', "");
Expect(0, 1616, '\P{^Canonical_Combining_Class=ccc31}', "");
Expect(1, 1615, '\p{Canonical_Combining_Class=	 CCC31}', "");
Expect(0, 1615, '\p{^Canonical_Combining_Class=	 CCC31}', "");
Expect(0, 1615, '\P{Canonical_Combining_Class=	 CCC31}', "");
Expect(1, 1615, '\P{^Canonical_Combining_Class=	 CCC31}', "");
Expect(0, 1616, '\p{Canonical_Combining_Class=	 CCC31}', "");
Expect(1, 1616, '\p{^Canonical_Combining_Class=	 CCC31}', "");
Expect(1, 1616, '\P{Canonical_Combining_Class=	 CCC31}', "");
Expect(0, 1616, '\P{^Canonical_Combining_Class=	 CCC31}', "");
Error('\p{Ccc:	 :=31}');
Error('\P{Ccc:	 :=31}');
Expect(1, 1615, '\p{Ccc=31}', "");
Expect(0, 1615, '\p{^Ccc=31}', "");
Expect(0, 1615, '\P{Ccc=31}', "");
Expect(1, 1615, '\P{^Ccc=31}', "");
Expect(0, 1616, '\p{Ccc=31}', "");
Expect(1, 1616, '\p{^Ccc=31}', "");
Expect(1, 1616, '\P{Ccc=31}', "");
Expect(0, 1616, '\P{^Ccc=31}', "");
Error('\p{Is_Canonical_Combining_Class= _CCC31:=}');
Error('\P{Is_Canonical_Combining_Class= _CCC31:=}');
Expect(1, 1615, '\p{Is_Canonical_Combining_Class=ccc31}', "");
Expect(0, 1615, '\p{^Is_Canonical_Combining_Class=ccc31}', "");
Expect(0, 1615, '\P{Is_Canonical_Combining_Class=ccc31}', "");
Expect(1, 1615, '\P{^Is_Canonical_Combining_Class=ccc31}', "");
Expect(0, 1616, '\p{Is_Canonical_Combining_Class=ccc31}', "");
Expect(1, 1616, '\p{^Is_Canonical_Combining_Class=ccc31}', "");
Expect(1, 1616, '\P{Is_Canonical_Combining_Class=ccc31}', "");
Expect(0, 1616, '\P{^Is_Canonical_Combining_Class=ccc31}', "");
Expect(1, 1615, '\p{Is_Canonical_Combining_Class=__CCC31}', "");
Expect(0, 1615, '\p{^Is_Canonical_Combining_Class=__CCC31}', "");
Expect(0, 1615, '\P{Is_Canonical_Combining_Class=__CCC31}', "");
Expect(1, 1615, '\P{^Is_Canonical_Combining_Class=__CCC31}', "");
Expect(0, 1616, '\p{Is_Canonical_Combining_Class=__CCC31}', "");
Expect(1, 1616, '\p{^Is_Canonical_Combining_Class=__CCC31}', "");
Expect(1, 1616, '\P{Is_Canonical_Combining_Class=__CCC31}', "");
Expect(0, 1616, '\P{^Is_Canonical_Combining_Class=__CCC31}', "");
Error('\p{Is_Ccc= :=+00000000031}');
Error('\P{Is_Ccc= :=+00000000031}');
Expect(1, 1615, '\p{Is_Ccc=0_0_0_31}', "");
Expect(0, 1615, '\p{^Is_Ccc=0_0_0_31}', "");
Expect(0, 1615, '\P{Is_Ccc=0_0_0_31}', "");
Expect(1, 1615, '\P{^Is_Ccc=0_0_0_31}', "");
Expect(0, 1616, '\p{Is_Ccc=0_0_0_31}', "");
Expect(1, 1616, '\p{^Is_Ccc=0_0_0_31}', "");
Expect(1, 1616, '\P{Is_Ccc=0_0_0_31}', "");
Expect(0, 1616, '\P{^Is_Ccc=0_0_0_31}', "");
Error('\p{Canonical_Combining_Class=/a/	 CCC35}');
Error('\P{Canonical_Combining_Class=/a/	 CCC35}');
Expect(1, 1648, '\p{Canonical_Combining_Class=ccc35}', "");
Expect(0, 1648, '\p{^Canonical_Combining_Class=ccc35}', "");
Expect(0, 1648, '\P{Canonical_Combining_Class=ccc35}', "");
Expect(1, 1648, '\P{^Canonical_Combining_Class=ccc35}', "");
Expect(0, 1649, '\p{Canonical_Combining_Class=ccc35}', "");
Expect(1, 1649, '\p{^Canonical_Combining_Class=ccc35}', "");
Expect(1, 1649, '\P{Canonical_Combining_Class=ccc35}', "");
Expect(0, 1649, '\P{^Canonical_Combining_Class=ccc35}', "");
Expect(1, 1648, '\p{Canonical_Combining_Class=	_CCC35}', "");
Expect(0, 1648, '\p{^Canonical_Combining_Class=	_CCC35}', "");
Expect(0, 1648, '\P{Canonical_Combining_Class=	_CCC35}', "");
Expect(1, 1648, '\P{^Canonical_Combining_Class=	_CCC35}', "");
Expect(0, 1649, '\p{Canonical_Combining_Class=	_CCC35}', "");
Expect(1, 1649, '\p{^Canonical_Combining_Class=	_CCC35}', "");
Expect(1, 1649, '\P{Canonical_Combining_Class=	_CCC35}', "");
Expect(0, 1649, '\P{^Canonical_Combining_Class=	_CCC35}', "");
Error('\p{Ccc=_-00000000035/a/}');
Error('\P{Ccc=_-00000000035/a/}');
Expect(1, 1648, '\p{Ccc=+0_0_0_35}', "");
Expect(0, 1648, '\p{^Ccc=+0_0_0_35}', "");
Expect(0, 1648, '\P{Ccc=+0_0_0_35}', "");
Expect(1, 1648, '\P{^Ccc=+0_0_0_35}', "");
Expect(0, 1649, '\p{Ccc=+0_0_0_35}', "");
Expect(1, 1649, '\p{^Ccc=+0_0_0_35}', "");
Expect(1, 1649, '\P{Ccc=+0_0_0_35}', "");
Expect(0, 1649, '\P{^Ccc=+0_0_0_35}', "");
Error('\p{Is_Canonical_Combining_Class=/a/_-ccc35}');
Error('\P{Is_Canonical_Combining_Class=/a/_-ccc35}');
Expect(1, 1648, '\p{Is_Canonical_Combining_Class=ccc35}', "");
Expect(0, 1648, '\p{^Is_Canonical_Combining_Class=ccc35}', "");
Expect(0, 1648, '\P{Is_Canonical_Combining_Class=ccc35}', "");
Expect(1, 1648, '\P{^Is_Canonical_Combining_Class=ccc35}', "");
Expect(0, 1649, '\p{Is_Canonical_Combining_Class=ccc35}', "");
Expect(1, 1649, '\p{^Is_Canonical_Combining_Class=ccc35}', "");
Expect(1, 1649, '\P{Is_Canonical_Combining_Class=ccc35}', "");
Expect(0, 1649, '\P{^Is_Canonical_Combining_Class=ccc35}', "");
Expect(1, 1648, '\p{Is_Canonical_Combining_Class=	_CCC35}', "");
Expect(0, 1648, '\p{^Is_Canonical_Combining_Class=	_CCC35}', "");
Expect(0, 1648, '\P{Is_Canonical_Combining_Class=	_CCC35}', "");
Expect(1, 1648, '\P{^Is_Canonical_Combining_Class=	_CCC35}', "");
Expect(0, 1649, '\p{Is_Canonical_Combining_Class=	_CCC35}', "");
Expect(1, 1649, '\p{^Is_Canonical_Combining_Class=	_CCC35}', "");
Expect(1, 1649, '\P{Is_Canonical_Combining_Class=	_CCC35}', "");
Expect(0, 1649, '\P{^Is_Canonical_Combining_Class=	_CCC35}', "");
Error('\p{Is_Ccc=	-00000000035:=}');
Error('\P{Is_Ccc=	-00000000035:=}');
Expect(1, 1648, '\p{Is_Ccc=35}', "");
Expect(0, 1648, '\p{^Is_Ccc=35}', "");
Expect(0, 1648, '\P{Is_Ccc=35}', "");
Expect(1, 1648, '\P{^Is_Ccc=35}', "");
Expect(0, 1649, '\p{Is_Ccc=35}', "");
Expect(1, 1649, '\p{^Is_Ccc=35}', "");
Expect(1, 1649, '\P{Is_Ccc=35}', "");
Expect(0, 1649, '\P{^Is_Ccc=35}', "");
Error('\p{Canonical_Combining_Class= 	IOTA_Subscript/a/}');
Error('\P{Canonical_Combining_Class= 	IOTA_Subscript/a/}');
Expect(1, 837, '\p{Canonical_Combining_Class=iotasubscript}', "");
Expect(0, 837, '\p{^Canonical_Combining_Class=iotasubscript}', "");
Expect(0, 837, '\P{Canonical_Combining_Class=iotasubscript}', "");
Expect(1, 837, '\P{^Canonical_Combining_Class=iotasubscript}', "");
Expect(0, 838, '\p{Canonical_Combining_Class=iotasubscript}', "");
Expect(1, 838, '\p{^Canonical_Combining_Class=iotasubscript}', "");
Expect(1, 838, '\P{Canonical_Combining_Class=iotasubscript}', "");
Expect(0, 838, '\P{^Canonical_Combining_Class=iotasubscript}', "");
Expect(1, 837, '\p{Canonical_Combining_Class=Iota_Subscript}', "");
Expect(0, 837, '\p{^Canonical_Combining_Class=Iota_Subscript}', "");
Expect(0, 837, '\P{Canonical_Combining_Class=Iota_Subscript}', "");
Expect(1, 837, '\P{^Canonical_Combining_Class=Iota_Subscript}', "");
Expect(0, 838, '\p{Canonical_Combining_Class=Iota_Subscript}', "");
Expect(1, 838, '\p{^Canonical_Combining_Class=Iota_Subscript}', "");
Expect(1, 838, '\P{Canonical_Combining_Class=Iota_Subscript}', "");
Expect(0, 838, '\P{^Canonical_Combining_Class=Iota_Subscript}', "");
Error('\p{Ccc=-/a/IS}');
Error('\P{Ccc=-/a/IS}');
Expect(1, 837, '\p{Ccc=is}', "");
Expect(0, 837, '\p{^Ccc=is}', "");
Expect(0, 837, '\P{Ccc=is}', "");
Expect(1, 837, '\P{^Ccc=is}', "");
Expect(0, 838, '\p{Ccc=is}', "");
Expect(1, 838, '\p{^Ccc=is}', "");
Expect(1, 838, '\P{Ccc=is}', "");
Expect(0, 838, '\P{^Ccc=is}', "");
Expect(1, 837, '\p{Ccc=-is}', "");
Expect(0, 837, '\p{^Ccc=-is}', "");
Expect(0, 837, '\P{Ccc=-is}', "");
Expect(1, 837, '\P{^Ccc=-is}', "");
Expect(0, 838, '\p{Ccc=-is}', "");
Expect(1, 838, '\p{^Ccc=-is}', "");
Expect(1, 838, '\P{Ccc=-is}', "");
Expect(0, 838, '\P{^Ccc=-is}', "");
Error('\p{Is_Canonical_Combining_Class=:= 	+240}');
Error('\P{Is_Canonical_Combining_Class=:= 	+240}');
Expect(1, 837, '\p{Is_Canonical_Combining_Class: 00240}', "");
Expect(0, 837, '\p{^Is_Canonical_Combining_Class: 00240}', "");
Expect(0, 837, '\P{Is_Canonical_Combining_Class: 00240}', "");
Expect(1, 837, '\P{^Is_Canonical_Combining_Class: 00240}', "");
Expect(0, 838, '\p{Is_Canonical_Combining_Class: 00240}', "");
Expect(1, 838, '\p{^Is_Canonical_Combining_Class: 00240}', "");
Expect(1, 838, '\P{Is_Canonical_Combining_Class: 00240}', "");
Expect(0, 838, '\P{^Is_Canonical_Combining_Class: 00240}', "");
Error('\p{Is_Ccc=/a/IOTA_Subscript}');
Error('\P{Is_Ccc=/a/IOTA_Subscript}');
Expect(1, 837, '\p{Is_Ccc=iotasubscript}', "");
Expect(0, 837, '\p{^Is_Ccc=iotasubscript}', "");
Expect(0, 837, '\P{Is_Ccc=iotasubscript}', "");
Expect(1, 837, '\P{^Is_Ccc=iotasubscript}', "");
Expect(0, 838, '\p{Is_Ccc=iotasubscript}', "");
Expect(1, 838, '\p{^Is_Ccc=iotasubscript}', "");
Expect(1, 838, '\P{Is_Ccc=iotasubscript}', "");
Expect(0, 838, '\P{^Is_Ccc=iotasubscript}', "");
Expect(1, 837, '\p{Is_Ccc=	iota_Subscript}', "");
Expect(0, 837, '\p{^Is_Ccc=	iota_Subscript}', "");
Expect(0, 837, '\P{Is_Ccc=	iota_Subscript}', "");
Expect(1, 837, '\P{^Is_Ccc=	iota_Subscript}', "");
Expect(0, 838, '\p{Is_Ccc=	iota_Subscript}', "");
Expect(1, 838, '\p{^Is_Ccc=	iota_Subscript}', "");
Expect(1, 838, '\P{Is_Ccc=	iota_Subscript}', "");
Expect(0, 838, '\P{^Is_Ccc=	iota_Subscript}', "");
Error('\p{Canonical_Combining_Class=/a/	-Above_Right}');
Error('\P{Canonical_Combining_Class=/a/	-Above_Right}');
Expect(1, 12332, '\p{Canonical_Combining_Class=aboveright}', "");
Expect(0, 12332, '\p{^Canonical_Combining_Class=aboveright}', "");
Expect(0, 12332, '\P{Canonical_Combining_Class=aboveright}', "");
Expect(1, 12332, '\P{^Canonical_Combining_Class=aboveright}', "");
Expect(0, 12333, '\p{Canonical_Combining_Class=aboveright}', "");
Expect(1, 12333, '\p{^Canonical_Combining_Class=aboveright}', "");
Expect(1, 12333, '\P{Canonical_Combining_Class=aboveright}', "");
Expect(0, 12333, '\P{^Canonical_Combining_Class=aboveright}', "");
Expect(1, 12332, '\p{Canonical_Combining_Class=-	Above_Right}', "");
Expect(0, 12332, '\p{^Canonical_Combining_Class=-	Above_Right}', "");
Expect(0, 12332, '\P{Canonical_Combining_Class=-	Above_Right}', "");
Expect(1, 12332, '\P{^Canonical_Combining_Class=-	Above_Right}', "");
Expect(0, 12333, '\p{Canonical_Combining_Class=-	Above_Right}', "");
Expect(1, 12333, '\p{^Canonical_Combining_Class=-	Above_Right}', "");
Expect(1, 12333, '\P{Canonical_Combining_Class=-	Above_Right}', "");
Expect(0, 12333, '\P{^Canonical_Combining_Class=-	Above_Right}', "");
Error('\p{Ccc=:=_	AR}');
Error('\P{Ccc=:=_	AR}');
Expect(1, 12332, '\p{Ccc: ar}', "");
Expect(0, 12332, '\p{^Ccc: ar}', "");
Expect(0, 12332, '\P{Ccc: ar}', "");
Expect(1, 12332, '\P{^Ccc: ar}', "");
Expect(0, 12333, '\p{Ccc: ar}', "");
Expect(1, 12333, '\p{^Ccc: ar}', "");
Expect(1, 12333, '\P{Ccc: ar}', "");
Expect(0, 12333, '\P{^Ccc: ar}', "");
Expect(1, 12332, '\p{Ccc= AR}', "");
Expect(0, 12332, '\p{^Ccc= AR}', "");
Expect(0, 12332, '\P{Ccc= AR}', "");
Expect(1, 12332, '\P{^Ccc= AR}', "");
Expect(0, 12333, '\p{Ccc= AR}', "");
Expect(1, 12333, '\p{^Ccc= AR}', "");
Expect(1, 12333, '\P{Ccc= AR}', "");
Expect(0, 12333, '\P{^Ccc= AR}', "");
Error('\p{Is_Canonical_Combining_Class=-:=0000000232}');
Error('\P{Is_Canonical_Combining_Class=-:=0000000232}');
Expect(1, 12332, '\p{Is_Canonical_Combining_Class=+000_023_2}', "");
Expect(0, 12332, '\p{^Is_Canonical_Combining_Class=+000_023_2}', "");
Expect(0, 12332, '\P{Is_Canonical_Combining_Class=+000_023_2}', "");
Expect(1, 12332, '\P{^Is_Canonical_Combining_Class=+000_023_2}', "");
Expect(0, 12333, '\p{Is_Canonical_Combining_Class=+000_023_2}', "");
Expect(1, 12333, '\p{^Is_Canonical_Combining_Class=+000_023_2}', "");
Expect(1, 12333, '\P{Is_Canonical_Combining_Class=+000_023_2}', "");
Expect(0, 12333, '\P{^Is_Canonical_Combining_Class=+000_023_2}', "");
Error('\p{Is_Ccc:   	/a/ABOVE_Right}');
Error('\P{Is_Ccc:   	/a/ABOVE_Right}');
Expect(1, 12332, '\p{Is_Ccc=aboveright}', "");
Expect(0, 12332, '\p{^Is_Ccc=aboveright}', "");
Expect(0, 12332, '\P{Is_Ccc=aboveright}', "");
Expect(1, 12332, '\P{^Is_Ccc=aboveright}', "");
Expect(0, 12333, '\p{Is_Ccc=aboveright}', "");
Expect(1, 12333, '\p{^Is_Ccc=aboveright}', "");
Expect(1, 12333, '\P{Is_Ccc=aboveright}', "");
Expect(0, 12333, '\P{^Is_Ccc=aboveright}', "");
Expect(1, 12332, '\p{Is_Ccc:   	 Above_right}', "");
Expect(0, 12332, '\p{^Is_Ccc:   	 Above_right}', "");
Expect(0, 12332, '\P{Is_Ccc:   	 Above_right}', "");
Expect(1, 12332, '\P{^Is_Ccc:   	 Above_right}', "");
Expect(0, 12333, '\p{Is_Ccc:   	 Above_right}', "");
Expect(1, 12333, '\p{^Is_Ccc:   	 Above_right}', "");
Expect(1, 12333, '\P{Is_Ccc:   	 Above_right}', "");
Expect(0, 12333, '\P{^Is_Ccc:   	 Above_right}', "");
Error('\p{Canonical_Combining_Class=:=	CCC21}');
Error('\P{Canonical_Combining_Class=:=	CCC21}');
Expect(1, 1468, '\p{Canonical_Combining_Class=ccc21}', "");
Expect(0, 1468, '\p{^Canonical_Combining_Class=ccc21}', "");
Expect(0, 1468, '\P{Canonical_Combining_Class=ccc21}', "");
Expect(1, 1468, '\P{^Canonical_Combining_Class=ccc21}', "");
Expect(0, 1469, '\p{Canonical_Combining_Class=ccc21}', "");
Expect(1, 1469, '\p{^Canonical_Combining_Class=ccc21}', "");
Expect(1, 1469, '\P{Canonical_Combining_Class=ccc21}', "");
Expect(0, 1469, '\P{^Canonical_Combining_Class=ccc21}', "");
Expect(1, 1468, '\p{Canonical_Combining_Class= -CCC21}', "");
Expect(0, 1468, '\p{^Canonical_Combining_Class= -CCC21}', "");
Expect(0, 1468, '\P{Canonical_Combining_Class= -CCC21}', "");
Expect(1, 1468, '\P{^Canonical_Combining_Class= -CCC21}', "");
Expect(0, 1469, '\p{Canonical_Combining_Class= -CCC21}', "");
Expect(1, 1469, '\p{^Canonical_Combining_Class= -CCC21}', "");
Expect(1, 1469, '\P{Canonical_Combining_Class= -CCC21}', "");
Expect(0, 1469, '\P{^Canonical_Combining_Class= -CCC21}', "");
Error('\p{Ccc=_ 0000021:=}');
Error('\P{Ccc=_ 0000021:=}');
Expect(1, 1468, '\p{Ccc=+0_0_21}', "");
Expect(0, 1468, '\p{^Ccc=+0_0_21}', "");
Expect(0, 1468, '\P{Ccc=+0_0_21}', "");
Expect(1, 1468, '\P{^Ccc=+0_0_21}', "");
Expect(0, 1469, '\p{Ccc=+0_0_21}', "");
Expect(1, 1469, '\p{^Ccc=+0_0_21}', "");
Expect(1, 1469, '\P{Ccc=+0_0_21}', "");
Expect(0, 1469, '\P{^Ccc=+0_0_21}', "");
Error('\p{Is_Canonical_Combining_Class=_-CCC21/a/}');
Error('\P{Is_Canonical_Combining_Class=_-CCC21/a/}');
Expect(1, 1468, '\p{Is_Canonical_Combining_Class=ccc21}', "");
Expect(0, 1468, '\p{^Is_Canonical_Combining_Class=ccc21}', "");
Expect(0, 1468, '\P{Is_Canonical_Combining_Class=ccc21}', "");
Expect(1, 1468, '\P{^Is_Canonical_Combining_Class=ccc21}', "");
Expect(0, 1469, '\p{Is_Canonical_Combining_Class=ccc21}', "");
Expect(1, 1469, '\p{^Is_Canonical_Combining_Class=ccc21}', "");
Expect(1, 1469, '\P{Is_Canonical_Combining_Class=ccc21}', "");
Expect(0, 1469, '\P{^Is_Canonical_Combining_Class=ccc21}', "");
Expect(1, 1468, '\p{Is_Canonical_Combining_Class:	 CCC21}', "");
Expect(0, 1468, '\p{^Is_Canonical_Combining_Class:	 CCC21}', "");
Expect(0, 1468, '\P{Is_Canonical_Combining_Class:	 CCC21}', "");
Expect(1, 1468, '\P{^Is_Canonical_Combining_Class:	 CCC21}', "");
Expect(0, 1469, '\p{Is_Canonical_Combining_Class:	 CCC21}', "");
Expect(1, 1469, '\p{^Is_Canonical_Combining_Class:	 CCC21}', "");
Expect(1, 1469, '\P{Is_Canonical_Combining_Class:	 CCC21}', "");
Expect(0, 1469, '\P{^Is_Canonical_Combining_Class:	 CCC21}', "");
Error('\p{Is_Ccc= -00_02_1:=}');
Error('\P{Is_Ccc= -00_02_1:=}');
Expect(1, 1468, '\p{Is_Ccc=0000000021}', "");
Expect(0, 1468, '\p{^Is_Ccc=0000000021}', "");
Expect(0, 1468, '\P{Is_Ccc=0000000021}', "");
Expect(1, 1468, '\P{^Is_Ccc=0000000021}', "");
Expect(0, 1469, '\p{Is_Ccc=0000000021}', "");
Expect(1, 1469, '\p{^Is_Ccc=0000000021}', "");
Expect(1, 1469, '\P{Is_Ccc=0000000021}', "");
Expect(0, 1469, '\P{^Is_Ccc=0000000021}', "");
Error('\p{Canonical_Combining_Class=	RIGHT:=}');
Error('\P{Canonical_Combining_Class=	RIGHT:=}');
Expect(1, 119149, '\p{Canonical_Combining_Class=right}', "");
Expect(0, 119149, '\p{^Canonical_Combining_Class=right}', "");
Expect(0, 119149, '\P{Canonical_Combining_Class=right}', "");
Expect(1, 119149, '\P{^Canonical_Combining_Class=right}', "");
Expect(0, 119150, '\p{Canonical_Combining_Class=right}', "");
Expect(1, 119150, '\p{^Canonical_Combining_Class=right}', "");
Expect(1, 119150, '\P{Canonical_Combining_Class=right}', "");
Expect(0, 119150, '\P{^Canonical_Combining_Class=right}', "");
Expect(1, 119149, '\p{Canonical_Combining_Class=_right}', "");
Expect(0, 119149, '\p{^Canonical_Combining_Class=_right}', "");
Expect(0, 119149, '\P{Canonical_Combining_Class=_right}', "");
Expect(1, 119149, '\P{^Canonical_Combining_Class=_right}', "");
Expect(0, 119150, '\p{Canonical_Combining_Class=_right}', "");
Expect(1, 119150, '\p{^Canonical_Combining_Class=_right}', "");
Expect(1, 119150, '\P{Canonical_Combining_Class=_right}', "");
Expect(0, 119150, '\P{^Canonical_Combining_Class=_right}', "");
Error('\p{Ccc=/a/_-R}');
Error('\P{Ccc=/a/_-R}');
Expect(1, 119149, '\p{Ccc=r}', "");
Expect(0, 119149, '\p{^Ccc=r}', "");
Expect(0, 119149, '\P{Ccc=r}', "");
Expect(1, 119149, '\P{^Ccc=r}', "");
Expect(0, 119150, '\p{Ccc=r}', "");
Expect(1, 119150, '\p{^Ccc=r}', "");
Expect(1, 119150, '\P{Ccc=r}', "");
Expect(0, 119150, '\P{^Ccc=r}', "");
Expect(1, 119149, '\p{Ccc=		R}', "");
Expect(0, 119149, '\p{^Ccc=		R}', "");
Expect(0, 119149, '\P{Ccc=		R}', "");
Expect(1, 119149, '\P{^Ccc=		R}', "");
Expect(0, 119150, '\p{Ccc=		R}', "");
Expect(1, 119150, '\p{^Ccc=		R}', "");
Expect(1, 119150, '\P{Ccc=		R}', "");
Expect(0, 119150, '\P{^Ccc=		R}', "");
Error('\p{Is_Canonical_Combining_Class=_	00000226/a/}');
Error('\P{Is_Canonical_Combining_Class=_	00000226/a/}');
Expect(1, 119149, '\p{Is_Canonical_Combining_Class=+0_0_0_0_2_26}', "");
Expect(0, 119149, '\p{^Is_Canonical_Combining_Class=+0_0_0_0_2_26}', "");
Expect(0, 119149, '\P{Is_Canonical_Combining_Class=+0_0_0_0_2_26}', "");
Expect(1, 119149, '\P{^Is_Canonical_Combining_Class=+0_0_0_0_2_26}', "");
Expect(0, 119150, '\p{Is_Canonical_Combining_Class=+0_0_0_0_2_26}', "");
Expect(1, 119150, '\p{^Is_Canonical_Combining_Class=+0_0_0_0_2_26}', "");
Expect(1, 119150, '\P{Is_Canonical_Combining_Class=+0_0_0_0_2_26}', "");
Expect(0, 119150, '\P{^Is_Canonical_Combining_Class=+0_0_0_0_2_26}', "");
Error('\p{Is_Ccc=RIGHT:=}');
Error('\P{Is_Ccc=RIGHT:=}');
Expect(1, 119149, '\p{Is_Ccc:   right}', "");
Expect(0, 119149, '\p{^Is_Ccc:   right}', "");
Expect(0, 119149, '\P{Is_Ccc:   right}', "");
Expect(1, 119149, '\P{^Is_Ccc:   right}', "");
Expect(0, 119150, '\p{Is_Ccc:   right}', "");
Expect(1, 119150, '\p{^Is_Ccc:   right}', "");
Expect(1, 119150, '\P{Is_Ccc:   right}', "");
Expect(0, 119150, '\P{^Is_Ccc:   right}', "");
Expect(1, 119149, '\p{Is_Ccc= RIGHT}', "");
Expect(0, 119149, '\p{^Is_Ccc= RIGHT}', "");
Expect(0, 119149, '\P{Is_Ccc= RIGHT}', "");
Expect(1, 119149, '\P{^Is_Ccc= RIGHT}', "");
Expect(0, 119150, '\p{Is_Ccc= RIGHT}', "");
Expect(1, 119150, '\p{^Is_Ccc= RIGHT}', "");
Expect(1, 119150, '\P{Is_Ccc= RIGHT}', "");
Expect(0, 119150, '\P{^Is_Ccc= RIGHT}', "");
Error('\p{Canonical_Combining_Class=:=--CCC15}');
Error('\P{Canonical_Combining_Class=:=--CCC15}');
Expect(1, 1461, '\p{Canonical_Combining_Class=ccc15}', "");
Expect(0, 1461, '\p{^Canonical_Combining_Class=ccc15}', "");
Expect(0, 1461, '\P{Canonical_Combining_Class=ccc15}', "");
Expect(1, 1461, '\P{^Canonical_Combining_Class=ccc15}', "");
Expect(0, 1462, '\p{Canonical_Combining_Class=ccc15}', "");
Expect(1, 1462, '\p{^Canonical_Combining_Class=ccc15}', "");
Expect(1, 1462, '\P{Canonical_Combining_Class=ccc15}', "");
Expect(0, 1462, '\P{^Canonical_Combining_Class=ccc15}', "");
Expect(1, 1461, '\p{Canonical_Combining_Class=-CCC15}', "");
Expect(0, 1461, '\p{^Canonical_Combining_Class=-CCC15}', "");
Expect(0, 1461, '\P{Canonical_Combining_Class=-CCC15}', "");
Expect(1, 1461, '\P{^Canonical_Combining_Class=-CCC15}', "");
Expect(0, 1462, '\p{Canonical_Combining_Class=-CCC15}', "");
Expect(1, 1462, '\p{^Canonical_Combining_Class=-CCC15}', "");
Expect(1, 1462, '\P{Canonical_Combining_Class=-CCC15}', "");
Expect(0, 1462, '\P{^Canonical_Combining_Class=-CCC15}', "");
Error('\p{Ccc=:=00000000015}');
Error('\P{Ccc=:=00000000015}');
Expect(1, 1461, '\p{Ccc=+1_5}', "");
Expect(0, 1461, '\p{^Ccc=+1_5}', "");
Expect(0, 1461, '\P{Ccc=+1_5}', "");
Expect(1, 1461, '\P{^Ccc=+1_5}', "");
Expect(0, 1462, '\p{Ccc=+1_5}', "");
Expect(1, 1462, '\p{^Ccc=+1_5}', "");
Expect(1, 1462, '\P{Ccc=+1_5}', "");
Expect(0, 1462, '\P{^Ccc=+1_5}', "");
Error('\p{Is_Canonical_Combining_Class= CCC15/a/}');
Error('\P{Is_Canonical_Combining_Class= CCC15/a/}');
Expect(1, 1461, '\p{Is_Canonical_Combining_Class=ccc15}', "");
Expect(0, 1461, '\p{^Is_Canonical_Combining_Class=ccc15}', "");
Expect(0, 1461, '\P{Is_Canonical_Combining_Class=ccc15}', "");
Expect(1, 1461, '\P{^Is_Canonical_Combining_Class=ccc15}', "");
Expect(0, 1462, '\p{Is_Canonical_Combining_Class=ccc15}', "");
Expect(1, 1462, '\p{^Is_Canonical_Combining_Class=ccc15}', "");
Expect(1, 1462, '\P{Is_Canonical_Combining_Class=ccc15}', "");
Expect(0, 1462, '\P{^Is_Canonical_Combining_Class=ccc15}', "");
Expect(1, 1461, '\p{Is_Canonical_Combining_Class=_ CCC15}', "");
Expect(0, 1461, '\p{^Is_Canonical_Combining_Class=_ CCC15}', "");
Expect(0, 1461, '\P{Is_Canonical_Combining_Class=_ CCC15}', "");
Expect(1, 1461, '\P{^Is_Canonical_Combining_Class=_ CCC15}', "");
Expect(0, 1462, '\p{Is_Canonical_Combining_Class=_ CCC15}', "");
Expect(1, 1462, '\p{^Is_Canonical_Combining_Class=_ CCC15}', "");
Expect(1, 1462, '\P{Is_Canonical_Combining_Class=_ CCC15}', "");
Expect(0, 1462, '\P{^Is_Canonical_Combining_Class=_ CCC15}', "");
Error('\p{Is_Ccc=:= 0001_5}');
Error('\P{Is_Ccc=:= 0001_5}');
Expect(1, 1461, '\p{Is_Ccc=0_0_0_0_0_15}', "");
Expect(0, 1461, '\p{^Is_Ccc=0_0_0_0_0_15}', "");
Expect(0, 1461, '\P{Is_Ccc=0_0_0_0_0_15}', "");
Expect(1, 1461, '\P{^Is_Ccc=0_0_0_0_0_15}', "");
Expect(0, 1462, '\p{Is_Ccc=0_0_0_0_0_15}', "");
Expect(1, 1462, '\p{^Is_Ccc=0_0_0_0_0_15}', "");
Expect(1, 1462, '\P{Is_Ccc=0_0_0_0_0_15}', "");
Expect(0, 1462, '\P{^Is_Ccc=0_0_0_0_0_15}', "");
Error('\p{Canonical_Combining_Class=:=	Below_Left}');
Error('\P{Canonical_Combining_Class=:=	Below_Left}');
Expect(1, 12330, '\p{Canonical_Combining_Class=belowleft}', "");
Expect(0, 12330, '\p{^Canonical_Combining_Class=belowleft}', "");
Expect(0, 12330, '\P{Canonical_Combining_Class=belowleft}', "");
Expect(1, 12330, '\P{^Canonical_Combining_Class=belowleft}', "");
Expect(0, 12331, '\p{Canonical_Combining_Class=belowleft}', "");
Expect(1, 12331, '\p{^Canonical_Combining_Class=belowleft}', "");
Expect(1, 12331, '\P{Canonical_Combining_Class=belowleft}', "");
Expect(0, 12331, '\P{^Canonical_Combining_Class=belowleft}', "");
Expect(1, 12330, '\p{Canonical_Combining_Class= Below_LEFT}', "");
Expect(0, 12330, '\p{^Canonical_Combining_Class= Below_LEFT}', "");
Expect(0, 12330, '\P{Canonical_Combining_Class= Below_LEFT}', "");
Expect(1, 12330, '\P{^Canonical_Combining_Class= Below_LEFT}', "");
Expect(0, 12331, '\p{Canonical_Combining_Class= Below_LEFT}', "");
Expect(1, 12331, '\p{^Canonical_Combining_Class= Below_LEFT}', "");
Expect(1, 12331, '\P{Canonical_Combining_Class= Below_LEFT}', "");
Expect(0, 12331, '\P{^Canonical_Combining_Class= Below_LEFT}', "");
Error('\p{Ccc=/a/--BL}');
Error('\P{Ccc=/a/--BL}');
Expect(1, 12330, '\p{Ccc: bl}', "");
Expect(0, 12330, '\p{^Ccc: bl}', "");
Expect(0, 12330, '\P{Ccc: bl}', "");
Expect(1, 12330, '\P{^Ccc: bl}', "");
Expect(0, 12331, '\p{Ccc: bl}', "");
Expect(1, 12331, '\p{^Ccc: bl}', "");
Expect(1, 12331, '\P{Ccc: bl}', "");
Expect(0, 12331, '\P{^Ccc: bl}', "");
Expect(1, 12330, '\p{Ccc=-	BL}', "");
Expect(0, 12330, '\p{^Ccc=-	BL}', "");
Expect(0, 12330, '\P{Ccc=-	BL}', "");
Expect(1, 12330, '\P{^Ccc=-	BL}', "");
Expect(0, 12331, '\p{Ccc=-	BL}', "");
Expect(1, 12331, '\p{^Ccc=-	BL}', "");
Expect(1, 12331, '\P{Ccc=-	BL}', "");
Expect(0, 12331, '\P{^Ccc=-	BL}', "");
Error('\p{Is_Canonical_Combining_Class=--21_8/a/}');
Error('\P{Is_Canonical_Combining_Class=--21_8/a/}');
Expect(1, 12330, '\p{Is_Canonical_Combining_Class=0021_8}', "");
Expect(0, 12330, '\p{^Is_Canonical_Combining_Class=0021_8}', "");
Expect(0, 12330, '\P{Is_Canonical_Combining_Class=0021_8}', "");
Expect(1, 12330, '\P{^Is_Canonical_Combining_Class=0021_8}', "");
Expect(0, 12331, '\p{Is_Canonical_Combining_Class=0021_8}', "");
Expect(1, 12331, '\p{^Is_Canonical_Combining_Class=0021_8}', "");
Expect(1, 12331, '\P{Is_Canonical_Combining_Class=0021_8}', "");
Expect(0, 12331, '\P{^Is_Canonical_Combining_Class=0021_8}', "");
Error('\p{Is_Ccc=-Below_left/a/}');
Error('\P{Is_Ccc=-Below_left/a/}');
Expect(1, 12330, '\p{Is_Ccc=belowleft}', "");
Expect(0, 12330, '\p{^Is_Ccc=belowleft}', "");
Expect(0, 12330, '\P{Is_Ccc=belowleft}', "");
Expect(1, 12330, '\P{^Is_Ccc=belowleft}', "");
Expect(0, 12331, '\p{Is_Ccc=belowleft}', "");
Expect(1, 12331, '\p{^Is_Ccc=belowleft}', "");
Expect(1, 12331, '\P{Is_Ccc=belowleft}', "");
Expect(0, 12331, '\P{^Is_Ccc=belowleft}', "");
Expect(1, 12330, '\p{Is_Ccc=_ Below_LEFT}', "");
Expect(0, 12330, '\p{^Is_Ccc=_ Below_LEFT}', "");
Expect(0, 12330, '\P{Is_Ccc=_ Below_LEFT}', "");
Expect(1, 12330, '\P{^Is_Ccc=_ Below_LEFT}', "");
Expect(0, 12331, '\p{Is_Ccc=_ Below_LEFT}', "");
Expect(1, 12331, '\p{^Is_Ccc=_ Below_LEFT}', "");
Expect(1, 12331, '\P{Is_Ccc=_ Below_LEFT}', "");
Expect(0, 12331, '\P{^Is_Ccc=_ Below_LEFT}', "");
Error('\p{Canonical_Combining_Class=_	CCC23/a/}');
Error('\P{Canonical_Combining_Class=_	CCC23/a/}');
Expect(1, 1471, '\p{Canonical_Combining_Class=ccc23}', "");
Expect(0, 1471, '\p{^Canonical_Combining_Class=ccc23}', "");
Expect(0, 1471, '\P{Canonical_Combining_Class=ccc23}', "");
Expect(1, 1471, '\P{^Canonical_Combining_Class=ccc23}', "");
Expect(0, 1472, '\p{Canonical_Combining_Class=ccc23}', "");
Expect(1, 1472, '\p{^Canonical_Combining_Class=ccc23}', "");
Expect(1, 1472, '\P{Canonical_Combining_Class=ccc23}', "");
Expect(0, 1472, '\P{^Canonical_Combining_Class=ccc23}', "");
Expect(1, 1471, '\p{Canonical_Combining_Class=	 CCC23}', "");
Expect(0, 1471, '\p{^Canonical_Combining_Class=	 CCC23}', "");
Expect(0, 1471, '\P{Canonical_Combining_Class=	 CCC23}', "");
Expect(1, 1471, '\P{^Canonical_Combining_Class=	 CCC23}', "");
Expect(0, 1472, '\p{Canonical_Combining_Class=	 CCC23}', "");
Expect(1, 1472, '\p{^Canonical_Combining_Class=	 CCC23}', "");
Expect(1, 1472, '\P{Canonical_Combining_Class=	 CCC23}', "");
Expect(0, 1472, '\P{^Canonical_Combining_Class=	 CCC23}', "");
Error('\p{Ccc= 	00000023/a/}');
Error('\P{Ccc= 	00000023/a/}');
Expect(1, 1471, '\p{Ccc=+000023}', "");
Expect(0, 1471, '\p{^Ccc=+000023}', "");
Expect(0, 1471, '\P{Ccc=+000023}', "");
Expect(1, 1471, '\P{^Ccc=+000023}', "");
Expect(0, 1472, '\p{Ccc=+000023}', "");
Expect(1, 1472, '\p{^Ccc=+000023}', "");
Expect(1, 1472, '\P{Ccc=+000023}', "");
Expect(0, 1472, '\P{^Ccc=+000023}', "");
Error('\p{Is_Canonical_Combining_Class=	 ccc23/a/}');
Error('\P{Is_Canonical_Combining_Class=	 ccc23/a/}');
Expect(1, 1471, '\p{Is_Canonical_Combining_Class=ccc23}', "");
Expect(0, 1471, '\p{^Is_Canonical_Combining_Class=ccc23}', "");
Expect(0, 1471, '\P{Is_Canonical_Combining_Class=ccc23}', "");
Expect(1, 1471, '\P{^Is_Canonical_Combining_Class=ccc23}', "");
Expect(0, 1472, '\p{Is_Canonical_Combining_Class=ccc23}', "");
Expect(1, 1472, '\p{^Is_Canonical_Combining_Class=ccc23}', "");
Expect(1, 1472, '\P{Is_Canonical_Combining_Class=ccc23}', "");
Expect(0, 1472, '\P{^Is_Canonical_Combining_Class=ccc23}', "");
Expect(1, 1471, '\p{Is_Canonical_Combining_Class=	CCC23}', "");
Expect(0, 1471, '\p{^Is_Canonical_Combining_Class=	CCC23}', "");
Expect(0, 1471, '\P{Is_Canonical_Combining_Class=	CCC23}', "");
Expect(1, 1471, '\P{^Is_Canonical_Combining_Class=	CCC23}', "");
Expect(0, 1472, '\p{Is_Canonical_Combining_Class=	CCC23}', "");
Expect(1, 1472, '\p{^Is_Canonical_Combining_Class=	CCC23}', "");
Expect(1, 1472, '\P{Is_Canonical_Combining_Class=	CCC23}', "");
Expect(0, 1472, '\P{^Is_Canonical_Combining_Class=	CCC23}', "");
Error('\p{Is_Ccc= -000023/a/}');
Error('\P{Is_Ccc= -000023/a/}');
Expect(1, 1471, '\p{Is_Ccc=+02_3}', "");
Expect(0, 1471, '\p{^Is_Ccc=+02_3}', "");
Expect(0, 1471, '\P{Is_Ccc=+02_3}', "");
Expect(1, 1471, '\P{^Is_Ccc=+02_3}', "");
Expect(0, 1472, '\p{Is_Ccc=+02_3}', "");
Expect(1, 1472, '\p{^Is_Ccc=+02_3}', "");
Expect(1, 1472, '\P{Is_Ccc=+02_3}', "");
Expect(0, 1472, '\P{^Is_Ccc=+02_3}', "");
Error('\p{Canonical_Combining_Class=/a/ CCC132}');
Error('\P{Canonical_Combining_Class=/a/ CCC132}');
Expect(1, 3956, '\p{Canonical_Combining_Class=ccc132}', "");
Expect(0, 3956, '\p{^Canonical_Combining_Class=ccc132}', "");
Expect(0, 3956, '\P{Canonical_Combining_Class=ccc132}', "");
Expect(1, 3956, '\P{^Canonical_Combining_Class=ccc132}', "");
Expect(0, 3957, '\p{Canonical_Combining_Class=ccc132}', "");
Expect(1, 3957, '\p{^Canonical_Combining_Class=ccc132}', "");
Expect(1, 3957, '\P{Canonical_Combining_Class=ccc132}', "");
Expect(0, 3957, '\P{^Canonical_Combining_Class=ccc132}', "");
Expect(1, 3956, '\p{Canonical_Combining_Class=  CCC132}', "");
Expect(0, 3956, '\p{^Canonical_Combining_Class=  CCC132}', "");
Expect(0, 3956, '\P{Canonical_Combining_Class=  CCC132}', "");
Expect(1, 3956, '\P{^Canonical_Combining_Class=  CCC132}', "");
Expect(0, 3957, '\p{Canonical_Combining_Class=  CCC132}', "");
Expect(1, 3957, '\p{^Canonical_Combining_Class=  CCC132}', "");
Expect(1, 3957, '\P{Canonical_Combining_Class=  CCC132}', "");
Expect(0, 3957, '\P{^Canonical_Combining_Class=  CCC132}', "");
Error('\p{Ccc=-0_0_0_0_0_0_0_132/a/}');
Error('\P{Ccc=-0_0_0_0_0_0_0_132/a/}');
Expect(1, 3956, '\p{Ccc: 00_00_00_01_32}', "");
Expect(0, 3956, '\p{^Ccc: 00_00_00_01_32}', "");
Expect(0, 3956, '\P{Ccc: 00_00_00_01_32}', "");
Expect(1, 3956, '\P{^Ccc: 00_00_00_01_32}', "");
Expect(0, 3957, '\p{Ccc: 00_00_00_01_32}', "");
Expect(1, 3957, '\p{^Ccc: 00_00_00_01_32}', "");
Expect(1, 3957, '\P{Ccc: 00_00_00_01_32}', "");
Expect(0, 3957, '\P{^Ccc: 00_00_00_01_32}', "");
Error('\p{Is_Canonical_Combining_Class=_-CCC132/a/}');
Error('\P{Is_Canonical_Combining_Class=_-CCC132/a/}');
Expect(1, 3956, '\p{Is_Canonical_Combining_Class=ccc132}', "");
Expect(0, 3956, '\p{^Is_Canonical_Combining_Class=ccc132}', "");
Expect(0, 3956, '\P{Is_Canonical_Combining_Class=ccc132}', "");
Expect(1, 3956, '\P{^Is_Canonical_Combining_Class=ccc132}', "");
Expect(0, 3957, '\p{Is_Canonical_Combining_Class=ccc132}', "");
Expect(1, 3957, '\p{^Is_Canonical_Combining_Class=ccc132}', "");
Expect(1, 3957, '\P{Is_Canonical_Combining_Class=ccc132}', "");
Expect(0, 3957, '\P{^Is_Canonical_Combining_Class=ccc132}', "");
Expect(1, 3956, '\p{Is_Canonical_Combining_Class= 	ccc132}', "");
Expect(0, 3956, '\p{^Is_Canonical_Combining_Class= 	ccc132}', "");
Expect(0, 3956, '\P{Is_Canonical_Combining_Class= 	ccc132}', "");
Expect(1, 3956, '\P{^Is_Canonical_Combining_Class= 	ccc132}', "");
Expect(0, 3957, '\p{Is_Canonical_Combining_Class= 	ccc132}', "");
Expect(1, 3957, '\p{^Is_Canonical_Combining_Class= 	ccc132}', "");
Expect(1, 3957, '\P{Is_Canonical_Combining_Class= 	ccc132}', "");
Expect(0, 3957, '\P{^Is_Canonical_Combining_Class= 	ccc132}', "");
Error('\p{Is_Ccc= /a/+0_0_0_1_32}');
Error('\P{Is_Ccc= /a/+0_0_0_1_32}');
Expect(1, 3956, '\p{Is_Ccc=+000_000_013_2}', "");
Expect(0, 3956, '\p{^Is_Ccc=+000_000_013_2}', "");
Expect(0, 3956, '\P{Is_Ccc=+000_000_013_2}', "");
Expect(1, 3956, '\P{^Is_Ccc=+000_000_013_2}', "");
Expect(0, 3957, '\p{Is_Ccc=+000_000_013_2}', "");
Expect(1, 3957, '\p{^Is_Ccc=+000_000_013_2}', "");
Expect(1, 3957, '\P{Is_Ccc=+000_000_013_2}', "");
Expect(0, 3957, '\P{^Is_Ccc=+000_000_013_2}', "");
Error('\p{Canonical_Combining_Class=--Left:=}');
Error('\P{Canonical_Combining_Class=--Left:=}');
Expect(1, 12335, '\p{Canonical_Combining_Class=left}', "");
Expect(0, 12335, '\p{^Canonical_Combining_Class=left}', "");
Expect(0, 12335, '\P{Canonical_Combining_Class=left}', "");
Expect(1, 12335, '\P{^Canonical_Combining_Class=left}', "");
Expect(0, 12336, '\p{Canonical_Combining_Class=left}', "");
Expect(1, 12336, '\p{^Canonical_Combining_Class=left}', "");
Expect(1, 12336, '\P{Canonical_Combining_Class=left}', "");
Expect(0, 12336, '\P{^Canonical_Combining_Class=left}', "");
Expect(1, 12335, '\p{Canonical_Combining_Class:   	Left}', "");
Expect(0, 12335, '\p{^Canonical_Combining_Class:   	Left}', "");
Expect(0, 12335, '\P{Canonical_Combining_Class:   	Left}', "");
Expect(1, 12335, '\P{^Canonical_Combining_Class:   	Left}', "");
Expect(0, 12336, '\p{Canonical_Combining_Class:   	Left}', "");
Expect(1, 12336, '\p{^Canonical_Combining_Class:   	Left}', "");
Expect(1, 12336, '\P{Canonical_Combining_Class:   	Left}', "");
Expect(0, 12336, '\P{^Canonical_Combining_Class:   	Left}', "");
Error('\p{Ccc=	:=L}');
Error('\P{Ccc=	:=L}');
Expect(1, 12335, '\p{Ccc=l}', "");
Expect(0, 12335, '\p{^Ccc=l}', "");
Expect(0, 12335, '\P{Ccc=l}', "");
Expect(1, 12335, '\P{^Ccc=l}', "");
Expect(0, 12336, '\p{Ccc=l}', "");
Expect(1, 12336, '\p{^Ccc=l}', "");
Expect(1, 12336, '\P{Ccc=l}', "");
Expect(0, 12336, '\P{^Ccc=l}', "");
Expect(1, 12335, '\p{Ccc=  l}', "");
Expect(0, 12335, '\p{^Ccc=  l}', "");
Expect(0, 12335, '\P{Ccc=  l}', "");
Expect(1, 12335, '\P{^Ccc=  l}', "");
Expect(0, 12336, '\p{Ccc=  l}', "");
Expect(1, 12336, '\p{^Ccc=  l}', "");
Expect(1, 12336, '\P{Ccc=  l}', "");
Expect(0, 12336, '\P{^Ccc=  l}', "");
Error('\p{Is_Canonical_Combining_Class=	_0224/a/}');
Error('\P{Is_Canonical_Combining_Class=	_0224/a/}');
Expect(1, 12335, '\p{Is_Canonical_Combining_Class=22_4}', "");
Expect(0, 12335, '\p{^Is_Canonical_Combining_Class=22_4}', "");
Expect(0, 12335, '\P{Is_Canonical_Combining_Class=22_4}', "");
Expect(1, 12335, '\P{^Is_Canonical_Combining_Class=22_4}', "");
Expect(0, 12336, '\p{Is_Canonical_Combining_Class=22_4}', "");
Expect(1, 12336, '\p{^Is_Canonical_Combining_Class=22_4}', "");
Expect(1, 12336, '\P{Is_Canonical_Combining_Class=22_4}', "");
Expect(0, 12336, '\P{^Is_Canonical_Combining_Class=22_4}', "");
Error('\p{Is_Ccc=/a/_ LEFT}');
Error('\P{Is_Ccc=/a/_ LEFT}');
Expect(1, 12335, '\p{Is_Ccc=left}', "");
Expect(0, 12335, '\p{^Is_Ccc=left}', "");
Expect(0, 12335, '\P{Is_Ccc=left}', "");
Expect(1, 12335, '\P{^Is_Ccc=left}', "");
Expect(0, 12336, '\p{Is_Ccc=left}', "");
Expect(1, 12336, '\p{^Is_Ccc=left}', "");
Expect(1, 12336, '\P{Is_Ccc=left}', "");
Expect(0, 12336, '\P{^Is_Ccc=left}', "");
Expect(1, 12335, '\p{Is_Ccc=  LEFT}', "");
Expect(0, 12335, '\p{^Is_Ccc=  LEFT}', "");
Expect(0, 12335, '\P{Is_Ccc=  LEFT}', "");
Expect(1, 12335, '\P{^Is_Ccc=  LEFT}', "");
Expect(0, 12336, '\p{Is_Ccc=  LEFT}', "");
Expect(1, 12336, '\p{^Is_Ccc=  LEFT}', "");
Expect(1, 12336, '\P{Is_Ccc=  LEFT}', "");
Expect(0, 12336, '\P{^Is_Ccc=  LEFT}', "");
Error('\p{Canonical_Combining_Class=:=	-CCC36}');
Error('\P{Canonical_Combining_Class=:=	-CCC36}');
Expect(1, 1809, '\p{Canonical_Combining_Class:ccc36}', "");
Expect(0, 1809, '\p{^Canonical_Combining_Class:ccc36}', "");
Expect(0, 1809, '\P{Canonical_Combining_Class:ccc36}', "");
Expect(1, 1809, '\P{^Canonical_Combining_Class:ccc36}', "");
Expect(0, 1810, '\p{Canonical_Combining_Class:ccc36}', "");
Expect(1, 1810, '\p{^Canonical_Combining_Class:ccc36}', "");
Expect(1, 1810, '\P{Canonical_Combining_Class:ccc36}', "");
Expect(0, 1810, '\P{^Canonical_Combining_Class:ccc36}', "");
Expect(1, 1809, '\p{Canonical_Combining_Class=	ccc36}', "");
Expect(0, 1809, '\p{^Canonical_Combining_Class=	ccc36}', "");
Expect(0, 1809, '\P{Canonical_Combining_Class=	ccc36}', "");
Expect(1, 1809, '\P{^Canonical_Combining_Class=	ccc36}', "");
Expect(0, 1810, '\p{Canonical_Combining_Class=	ccc36}', "");
Expect(1, 1810, '\p{^Canonical_Combining_Class=	ccc36}', "");
Expect(1, 1810, '\P{Canonical_Combining_Class=	ccc36}', "");
Expect(0, 1810, '\P{^Canonical_Combining_Class=	ccc36}', "");
Error('\p{Ccc=/a/00000003_6}');
Error('\P{Ccc=/a/00000003_6}');
Expect(1, 1809, '\p{Ccc=00_00_00_36}', "");
Expect(0, 1809, '\p{^Ccc=00_00_00_36}', "");
Expect(0, 1809, '\P{Ccc=00_00_00_36}', "");
Expect(1, 1809, '\P{^Ccc=00_00_00_36}', "");
Expect(0, 1810, '\p{Ccc=00_00_00_36}', "");
Expect(1, 1810, '\p{^Ccc=00_00_00_36}', "");
Expect(1, 1810, '\P{Ccc=00_00_00_36}', "");
Expect(0, 1810, '\P{^Ccc=00_00_00_36}', "");
Error('\p{Is_Canonical_Combining_Class=:=CCC36}');
Error('\P{Is_Canonical_Combining_Class=:=CCC36}');
Expect(1, 1809, '\p{Is_Canonical_Combining_Class: ccc36}', "");
Expect(0, 1809, '\p{^Is_Canonical_Combining_Class: ccc36}', "");
Expect(0, 1809, '\P{Is_Canonical_Combining_Class: ccc36}', "");
Expect(1, 1809, '\P{^Is_Canonical_Combining_Class: ccc36}', "");
Expect(0, 1810, '\p{Is_Canonical_Combining_Class: ccc36}', "");
Expect(1, 1810, '\p{^Is_Canonical_Combining_Class: ccc36}', "");
Expect(1, 1810, '\P{Is_Canonical_Combining_Class: ccc36}', "");
Expect(0, 1810, '\P{^Is_Canonical_Combining_Class: ccc36}', "");
Expect(1, 1809, '\p{Is_Canonical_Combining_Class= _CCC36}', "");
Expect(0, 1809, '\p{^Is_Canonical_Combining_Class= _CCC36}', "");
Expect(0, 1809, '\P{Is_Canonical_Combining_Class= _CCC36}', "");
Expect(1, 1809, '\P{^Is_Canonical_Combining_Class= _CCC36}', "");
Expect(0, 1810, '\p{Is_Canonical_Combining_Class= _CCC36}', "");
Expect(1, 1810, '\p{^Is_Canonical_Combining_Class= _CCC36}', "");
Expect(1, 1810, '\P{Is_Canonical_Combining_Class= _CCC36}', "");
Expect(0, 1810, '\P{^Is_Canonical_Combining_Class= _CCC36}', "");
Error('\p{Is_Ccc=	+000_000_003_6/a/}');
Error('\P{Is_Ccc=	+000_000_003_6/a/}');
Expect(1, 1809, '\p{Is_Ccc=00000_00003_6}', "");
Expect(0, 1809, '\p{^Is_Ccc=00000_00003_6}', "");
Expect(0, 1809, '\P{Is_Ccc=00000_00003_6}', "");
Expect(1, 1809, '\P{^Is_Ccc=00000_00003_6}', "");
Expect(0, 1810, '\p{Is_Ccc=00000_00003_6}', "");
Expect(1, 1810, '\p{^Is_Ccc=00000_00003_6}', "");
Expect(1, 1810, '\P{Is_Ccc=00000_00003_6}', "");
Expect(0, 1810, '\P{^Is_Ccc=00000_00003_6}', "");
Error('\p{Canonical_Combining_Class=		ccc26:=}');
Error('\P{Canonical_Combining_Class=		ccc26:=}');
Expect(1, 64286, '\p{Canonical_Combining_Class=ccc26}', "");
Expect(0, 64286, '\p{^Canonical_Combining_Class=ccc26}', "");
Expect(0, 64286, '\P{Canonical_Combining_Class=ccc26}', "");
Expect(1, 64286, '\P{^Canonical_Combining_Class=ccc26}', "");
Expect(0, 64287, '\p{Canonical_Combining_Class=ccc26}', "");
Expect(1, 64287, '\p{^Canonical_Combining_Class=ccc26}', "");
Expect(1, 64287, '\P{Canonical_Combining_Class=ccc26}', "");
Expect(0, 64287, '\P{^Canonical_Combining_Class=ccc26}', "");
Error('\p{Ccc=:=_000000026}');
Error('\P{Ccc=:=_000000026}');
Expect(1, 64286, '\p{Ccc=000002_6}', "");
Expect(0, 64286, '\p{^Ccc=000002_6}', "");
Expect(0, 64286, '\P{Ccc=000002_6}', "");
Expect(1, 64286, '\P{^Ccc=000002_6}', "");
Expect(0, 64287, '\p{Ccc=000002_6}', "");
Expect(1, 64287, '\p{^Ccc=000002_6}', "");
Expect(1, 64287, '\P{Ccc=000002_6}', "");
Expect(0, 64287, '\P{^Ccc=000002_6}', "");
Error('\p{Is_Canonical_Combining_Class= :=ccc26}');
Error('\P{Is_Canonical_Combining_Class= :=ccc26}');
Expect(1, 64286, '\p{Is_Canonical_Combining_Class: ccc26}', "");
Expect(0, 64286, '\p{^Is_Canonical_Combining_Class: ccc26}', "");
Expect(0, 64286, '\P{Is_Canonical_Combining_Class: ccc26}', "");
Expect(1, 64286, '\P{^Is_Canonical_Combining_Class: ccc26}', "");
Expect(0, 64287, '\p{Is_Canonical_Combining_Class: ccc26}', "");
Expect(1, 64287, '\p{^Is_Canonical_Combining_Class: ccc26}', "");
Expect(1, 64287, '\P{Is_Canonical_Combining_Class: ccc26}', "");
Expect(0, 64287, '\P{^Is_Canonical_Combining_Class: ccc26}', "");
Expect(1, 64286, '\p{Is_Canonical_Combining_Class=_-CCC26}', "");
Expect(0, 64286, '\p{^Is_Canonical_Combining_Class=_-CCC26}', "");
Expect(0, 64286, '\P{Is_Canonical_Combining_Class=_-CCC26}', "");
Expect(1, 64286, '\P{^Is_Canonical_Combining_Class=_-CCC26}', "");
Expect(0, 64287, '\p{Is_Canonical_Combining_Class=_-CCC26}', "");
Expect(1, 64287, '\p{^Is_Canonical_Combining_Class=_-CCC26}', "");
Expect(1, 64287, '\P{Is_Canonical_Combining_Class=_-CCC26}', "");
Expect(0, 64287, '\P{^Is_Canonical_Combining_Class=_-CCC26}', "");
Error('\p{Is_Ccc=_/a/02_6}');
Error('\P{Is_Ccc=_/a/02_6}');
Expect(1, 64286, '\p{Is_Ccc=+00026}', "");
Expect(0, 64286, '\p{^Is_Ccc=+00026}', "");
Expect(0, 64286, '\P{Is_Ccc=+00026}', "");
Expect(1, 64286, '\P{^Is_Ccc=+00026}', "");
Expect(0, 64287, '\p{Is_Ccc=+00026}', "");
Expect(1, 64287, '\p{^Is_Ccc=+00026}', "");
Expect(1, 64287, '\P{Is_Ccc=+00026}', "");
Expect(0, 64287, '\P{^Is_Ccc=+00026}', "");
Error('\p{Canonical_Combining_Class=-:=ABOVE}');
Error('\P{Canonical_Combining_Class=-:=ABOVE}');
Expect(1, 119364, '\p{Canonical_Combining_Class:	above}', "");
Expect(0, 119364, '\p{^Canonical_Combining_Class:	above}', "");
Expect(0, 119364, '\P{Canonical_Combining_Class:	above}', "");
Expect(1, 119364, '\P{^Canonical_Combining_Class:	above}', "");
Expect(0, 119365, '\p{Canonical_Combining_Class:	above}', "");
Expect(1, 119365, '\p{^Canonical_Combining_Class:	above}', "");
Expect(1, 119365, '\P{Canonical_Combining_Class:	above}', "");
Expect(0, 119365, '\P{^Canonical_Combining_Class:	above}', "");
Expect(1, 119364, '\p{Canonical_Combining_Class=-Above}', "");
Expect(0, 119364, '\p{^Canonical_Combining_Class=-Above}', "");
Expect(0, 119364, '\P{Canonical_Combining_Class=-Above}', "");
Expect(1, 119364, '\P{^Canonical_Combining_Class=-Above}', "");
Expect(0, 119365, '\p{Canonical_Combining_Class=-Above}', "");
Expect(1, 119365, '\p{^Canonical_Combining_Class=-Above}', "");
Expect(1, 119365, '\P{Canonical_Combining_Class=-Above}', "");
Expect(0, 119365, '\P{^Canonical_Combining_Class=-Above}', "");
Error('\p{Ccc= /a/A}');
Error('\P{Ccc= /a/A}');
Expect(1, 119364, '\p{Ccc=a}', "");
Expect(0, 119364, '\p{^Ccc=a}', "");
Expect(0, 119364, '\P{Ccc=a}', "");
Expect(1, 119364, '\P{^Ccc=a}', "");
Expect(0, 119365, '\p{Ccc=a}', "");
Expect(1, 119365, '\p{^Ccc=a}', "");
Expect(1, 119365, '\P{Ccc=a}', "");
Expect(0, 119365, '\P{^Ccc=a}', "");
Expect(1, 119364, '\p{Ccc=	 A}', "");
Expect(0, 119364, '\p{^Ccc=	 A}', "");
Expect(0, 119364, '\P{Ccc=	 A}', "");
Expect(1, 119364, '\P{^Ccc=	 A}', "");
Expect(0, 119365, '\p{Ccc=	 A}', "");
Expect(1, 119365, '\p{^Ccc=	 A}', "");
Expect(1, 119365, '\P{Ccc=	 A}', "");
Expect(0, 119365, '\P{^Ccc=	 A}', "");
Error('\p{Is_Canonical_Combining_Class=	:=00_02_30}');
Error('\P{Is_Canonical_Combining_Class=	:=00_02_30}');
Expect(1, 119364, '\p{Is_Canonical_Combining_Class:	0230}', "");
Expect(0, 119364, '\p{^Is_Canonical_Combining_Class:	0230}', "");
Expect(0, 119364, '\P{Is_Canonical_Combining_Class:	0230}', "");
Expect(1, 119364, '\P{^Is_Canonical_Combining_Class:	0230}', "");
Expect(0, 119365, '\p{Is_Canonical_Combining_Class:	0230}', "");
Expect(1, 119365, '\p{^Is_Canonical_Combining_Class:	0230}', "");
Expect(1, 119365, '\P{Is_Canonical_Combining_Class:	0230}', "");
Expect(0, 119365, '\P{^Is_Canonical_Combining_Class:	0230}', "");
Error('\p{Is_Ccc=:=_ ABOVE}');
Error('\P{Is_Ccc=:=_ ABOVE}');
Expect(1, 119364, '\p{Is_Ccc=above}', "");
Expect(0, 119364, '\p{^Is_Ccc=above}', "");
Expect(0, 119364, '\P{Is_Ccc=above}', "");
Expect(1, 119364, '\P{^Is_Ccc=above}', "");
Expect(0, 119365, '\p{Is_Ccc=above}', "");
Expect(1, 119365, '\p{^Is_Ccc=above}', "");
Expect(1, 119365, '\P{Is_Ccc=above}', "");
Expect(0, 119365, '\P{^Is_Ccc=above}', "");
Expect(1, 119364, '\p{Is_Ccc=Above}', "");
Expect(0, 119364, '\p{^Is_Ccc=Above}', "");
Expect(0, 119364, '\P{Is_Ccc=Above}', "");
Expect(1, 119364, '\P{^Is_Ccc=Above}', "");
Expect(0, 119365, '\p{Is_Ccc=Above}', "");
Expect(1, 119365, '\p{^Is_Ccc=Above}', "");
Expect(1, 119365, '\P{Is_Ccc=Above}', "");
Expect(0, 119365, '\P{^Is_Ccc=Above}', "");
Error('\p{Canonical_Combining_Class=-_CCC129:=}');
Error('\P{Canonical_Combining_Class=-_CCC129:=}');
Expect(1, 3953, '\p{Canonical_Combining_Class=ccc129}', "");
Expect(0, 3953, '\p{^Canonical_Combining_Class=ccc129}', "");
Expect(0, 3953, '\P{Canonical_Combining_Class=ccc129}', "");
Expect(1, 3953, '\P{^Canonical_Combining_Class=ccc129}', "");
Expect(0, 3954, '\p{Canonical_Combining_Class=ccc129}', "");
Expect(1, 3954, '\p{^Canonical_Combining_Class=ccc129}', "");
Expect(1, 3954, '\P{Canonical_Combining_Class=ccc129}', "");
Expect(0, 3954, '\P{^Canonical_Combining_Class=ccc129}', "");
Expect(1, 3953, '\p{Canonical_Combining_Class=CCC129}', "");
Expect(0, 3953, '\p{^Canonical_Combining_Class=CCC129}', "");
Expect(0, 3953, '\P{Canonical_Combining_Class=CCC129}', "");
Expect(1, 3953, '\P{^Canonical_Combining_Class=CCC129}', "");
Expect(0, 3954, '\p{Canonical_Combining_Class=CCC129}', "");
Expect(1, 3954, '\p{^Canonical_Combining_Class=CCC129}', "");
Expect(1, 3954, '\P{Canonical_Combining_Class=CCC129}', "");
Expect(0, 3954, '\P{^Canonical_Combining_Class=CCC129}', "");
Error('\p{Ccc=-:=12_9}');
Error('\P{Ccc=-:=12_9}');
Expect(1, 3953, '\p{Ccc=0129}', "");
Expect(0, 3953, '\p{^Ccc=0129}', "");
Expect(0, 3953, '\P{Ccc=0129}', "");
Expect(1, 3953, '\P{^Ccc=0129}', "");
Expect(0, 3954, '\p{Ccc=0129}', "");
Expect(1, 3954, '\p{^Ccc=0129}', "");
Expect(1, 3954, '\P{Ccc=0129}', "");
Expect(0, 3954, '\P{^Ccc=0129}', "");
Error('\p{Is_Canonical_Combining_Class=_	CCC129/a/}');
Error('\P{Is_Canonical_Combining_Class=_	CCC129/a/}');
Expect(1, 3953, '\p{Is_Canonical_Combining_Class=ccc129}', "");
Expect(0, 3953, '\p{^Is_Canonical_Combining_Class=ccc129}', "");
Expect(0, 3953, '\P{Is_Canonical_Combining_Class=ccc129}', "");
Expect(1, 3953, '\P{^Is_Canonical_Combining_Class=ccc129}', "");
Expect(0, 3954, '\p{Is_Canonical_Combining_Class=ccc129}', "");
Expect(1, 3954, '\p{^Is_Canonical_Combining_Class=ccc129}', "");
Expect(1, 3954, '\P{Is_Canonical_Combining_Class=ccc129}', "");
Expect(0, 3954, '\P{^Is_Canonical_Combining_Class=ccc129}', "");
Expect(1, 3953, '\p{Is_Canonical_Combining_Class:	__CCC129}', "");
Expect(0, 3953, '\p{^Is_Canonical_Combining_Class:	__CCC129}', "");
Expect(0, 3953, '\P{Is_Canonical_Combining_Class:	__CCC129}', "");
Expect(1, 3953, '\P{^Is_Canonical_Combining_Class:	__CCC129}', "");
Expect(0, 3954, '\p{Is_Canonical_Combining_Class:	__CCC129}', "");
Expect(1, 3954, '\p{^Is_Canonical_Combining_Class:	__CCC129}', "");
Expect(1, 3954, '\P{Is_Canonical_Combining_Class:	__CCC129}', "");
Expect(0, 3954, '\P{^Is_Canonical_Combining_Class:	__CCC129}', "");
Error('\p{Is_Ccc=_	+0000000129/a/}');
Error('\P{Is_Ccc=_	+0000000129/a/}');
Expect(1, 3953, '\p{Is_Ccc=+000000129}', "");
Expect(0, 3953, '\p{^Is_Ccc=+000000129}', "");
Expect(0, 3953, '\P{Is_Ccc=+000000129}', "");
Expect(1, 3953, '\P{^Is_Ccc=+000000129}', "");
Expect(0, 3954, '\p{Is_Ccc=+000000129}', "");
Expect(1, 3954, '\p{^Is_Ccc=+000000129}', "");
Expect(1, 3954, '\P{Is_Ccc=+000000129}', "");
Expect(0, 3954, '\P{^Is_Ccc=+000000129}', "");
Error('\p{Canonical_Combining_Class= /a/ccc17}');
Error('\P{Canonical_Combining_Class= /a/ccc17}');
Expect(1, 1463, '\p{Canonical_Combining_Class=ccc17}', "");
Expect(0, 1463, '\p{^Canonical_Combining_Class=ccc17}', "");
Expect(0, 1463, '\P{Canonical_Combining_Class=ccc17}', "");
Expect(1, 1463, '\P{^Canonical_Combining_Class=ccc17}', "");
Expect(0, 1464, '\p{Canonical_Combining_Class=ccc17}', "");
Expect(1, 1464, '\p{^Canonical_Combining_Class=ccc17}', "");
Expect(1, 1464, '\P{Canonical_Combining_Class=ccc17}', "");
Expect(0, 1464, '\P{^Canonical_Combining_Class=ccc17}', "");
Expect(1, 1463, '\p{Canonical_Combining_Class=CCC17}', "");
Expect(0, 1463, '\p{^Canonical_Combining_Class=CCC17}', "");
Expect(0, 1463, '\P{Canonical_Combining_Class=CCC17}', "");
Expect(1, 1463, '\P{^Canonical_Combining_Class=CCC17}', "");
Expect(0, 1464, '\p{Canonical_Combining_Class=CCC17}', "");
Expect(1, 1464, '\p{^Canonical_Combining_Class=CCC17}', "");
Expect(1, 1464, '\P{Canonical_Combining_Class=CCC17}', "");
Expect(0, 1464, '\P{^Canonical_Combining_Class=CCC17}', "");
Error('\p{Ccc=:=--001_7}');
Error('\P{Ccc=:=--001_7}');
Expect(1, 1463, '\p{Ccc=+0017}', "");
Expect(0, 1463, '\p{^Ccc=+0017}', "");
Expect(0, 1463, '\P{Ccc=+0017}', "");
Expect(1, 1463, '\P{^Ccc=+0017}', "");
Expect(0, 1464, '\p{Ccc=+0017}', "");
Expect(1, 1464, '\p{^Ccc=+0017}', "");
Expect(1, 1464, '\P{Ccc=+0017}', "");
Expect(0, 1464, '\P{^Ccc=+0017}', "");
Error('\p{Is_Canonical_Combining_Class=:=	CCC17}');
Error('\P{Is_Canonical_Combining_Class=:=	CCC17}');
Expect(1, 1463, '\p{Is_Canonical_Combining_Class=ccc17}', "");
Expect(0, 1463, '\p{^Is_Canonical_Combining_Class=ccc17}', "");
Expect(0, 1463, '\P{Is_Canonical_Combining_Class=ccc17}', "");
Expect(1, 1463, '\P{^Is_Canonical_Combining_Class=ccc17}', "");
Expect(0, 1464, '\p{Is_Canonical_Combining_Class=ccc17}', "");
Expect(1, 1464, '\p{^Is_Canonical_Combining_Class=ccc17}', "");
Expect(1, 1464, '\P{Is_Canonical_Combining_Class=ccc17}', "");
Expect(0, 1464, '\P{^Is_Canonical_Combining_Class=ccc17}', "");
Expect(1, 1463, '\p{Is_Canonical_Combining_Class=		CCC17}', "");
Expect(0, 1463, '\p{^Is_Canonical_Combining_Class=		CCC17}', "");
Expect(0, 1463, '\P{Is_Canonical_Combining_Class=		CCC17}', "");
Expect(1, 1463, '\P{^Is_Canonical_Combining_Class=		CCC17}', "");
Expect(0, 1464, '\p{Is_Canonical_Combining_Class=		CCC17}', "");
Expect(1, 1464, '\p{^Is_Canonical_Combining_Class=		CCC17}', "");
Expect(1, 1464, '\P{Is_Canonical_Combining_Class=		CCC17}', "");
Expect(0, 1464, '\P{^Is_Canonical_Combining_Class=		CCC17}', "");
Error('\p{Is_Ccc:  _+01_7/a/}');
Error('\P{Is_Ccc:  _+01_7/a/}');
Expect(1, 1463, '\p{Is_Ccc: 00000017}', "");
Expect(0, 1463, '\p{^Is_Ccc: 00000017}', "");
Expect(0, 1463, '\P{Is_Ccc: 00000017}', "");
Expect(1, 1463, '\P{^Is_Ccc: 00000017}', "");
Expect(0, 1464, '\p{Is_Ccc: 00000017}', "");
Expect(1, 1464, '\p{^Is_Ccc: 00000017}', "");
Expect(1, 1464, '\P{Is_Ccc: 00000017}', "");
Expect(0, 1464, '\P{^Is_Ccc: 00000017}', "");
Error('\p{Canonical_Combining_Class=_/a/NOT_reordered}');
Error('\P{Canonical_Combining_Class=_/a/NOT_reordered}');
Expect(1, 119365, '\p{Canonical_Combining_Class=notreordered}', "");
Expect(0, 119365, '\p{^Canonical_Combining_Class=notreordered}', "");
Expect(0, 119365, '\P{Canonical_Combining_Class=notreordered}', "");
Expect(1, 119365, '\P{^Canonical_Combining_Class=notreordered}', "");
Expect(0, 119364, '\p{Canonical_Combining_Class=notreordered}', "");
Expect(1, 119364, '\p{^Canonical_Combining_Class=notreordered}', "");
Expect(1, 119364, '\P{Canonical_Combining_Class=notreordered}', "");
Expect(0, 119364, '\P{^Canonical_Combining_Class=notreordered}', "");
Expect(1, 119365, '\p{Canonical_Combining_Class=	 Not_REORDERED}', "");
Expect(0, 119365, '\p{^Canonical_Combining_Class=	 Not_REORDERED}', "");
Expect(0, 119365, '\P{Canonical_Combining_Class=	 Not_REORDERED}', "");
Expect(1, 119365, '\P{^Canonical_Combining_Class=	 Not_REORDERED}', "");
Expect(0, 119364, '\p{Canonical_Combining_Class=	 Not_REORDERED}', "");
Expect(1, 119364, '\p{^Canonical_Combining_Class=	 Not_REORDERED}', "");
Expect(1, 119364, '\P{Canonical_Combining_Class=	 Not_REORDERED}', "");
Expect(0, 119364, '\P{^Canonical_Combining_Class=	 Not_REORDERED}', "");
Error('\p{Ccc=--NR:=}');
Error('\P{Ccc=--NR:=}');
Expect(1, 119365, '\p{Ccc=nr}', "");
Expect(0, 119365, '\p{^Ccc=nr}', "");
Expect(0, 119365, '\P{Ccc=nr}', "");
Expect(1, 119365, '\P{^Ccc=nr}', "");
Expect(0, 119364, '\p{Ccc=nr}', "");
Expect(1, 119364, '\p{^Ccc=nr}', "");
Expect(1, 119364, '\P{Ccc=nr}', "");
Expect(0, 119364, '\P{^Ccc=nr}', "");
Expect(1, 119365, '\p{Ccc=_-NR}', "");
Expect(0, 119365, '\p{^Ccc=_-NR}', "");
Expect(0, 119365, '\P{Ccc=_-NR}', "");
Expect(1, 119365, '\P{^Ccc=_-NR}', "");
Expect(0, 119364, '\p{Ccc=_-NR}', "");
Expect(1, 119364, '\p{^Ccc=_-NR}', "");
Expect(1, 119364, '\P{Ccc=_-NR}', "");
Expect(0, 119364, '\P{^Ccc=_-NR}', "");
Error('\p{Is_Canonical_Combining_Class:	:=000000}');
Error('\P{Is_Canonical_Combining_Class:	:=000000}');
Expect(1, 119365, '\p{Is_Canonical_Combining_Class=000000}', "");
Expect(0, 119365, '\p{^Is_Canonical_Combining_Class=000000}', "");
Expect(0, 119365, '\P{Is_Canonical_Combining_Class=000000}', "");
Expect(1, 119365, '\P{^Is_Canonical_Combining_Class=000000}', "");
Expect(0, 119364, '\p{Is_Canonical_Combining_Class=000000}', "");
Expect(1, 119364, '\p{^Is_Canonical_Combining_Class=000000}', "");
Expect(1, 119364, '\P{Is_Canonical_Combining_Class=000000}', "");
Expect(0, 119364, '\P{^Is_Canonical_Combining_Class=000000}', "");
Error('\p{Is_Ccc: /a/-_Not_Reordered}');
Error('\P{Is_Ccc: /a/-_Not_Reordered}');
Expect(1, 119365, '\p{Is_Ccc: notreordered}', "");
Expect(0, 119365, '\p{^Is_Ccc: notreordered}', "");
Expect(0, 119365, '\P{Is_Ccc: notreordered}', "");
Expect(1, 119365, '\P{^Is_Ccc: notreordered}', "");
Expect(0, 119364, '\p{Is_Ccc: notreordered}', "");
Expect(1, 119364, '\p{^Is_Ccc: notreordered}', "");
Expect(1, 119364, '\P{Is_Ccc: notreordered}', "");
Expect(0, 119364, '\P{^Is_Ccc: notreordered}', "");
Expect(1, 119365, '\p{Is_Ccc: 	_NOT_Reordered}', "");
Expect(0, 119365, '\p{^Is_Ccc: 	_NOT_Reordered}', "");
Expect(0, 119365, '\P{Is_Ccc: 	_NOT_Reordered}', "");
Expect(1, 119365, '\P{^Is_Ccc: 	_NOT_Reordered}', "");
Expect(0, 119364, '\p{Is_Ccc: 	_NOT_Reordered}', "");
Expect(1, 119364, '\p{^Is_Ccc: 	_NOT_Reordered}', "");
Expect(1, 119364, '\P{Is_Ccc: 	_NOT_Reordered}', "");
Expect(0, 119364, '\P{^Is_Ccc: 	_NOT_Reordered}', "");
Error('\p{Canonical_Combining_Class=:=-CCC22}');
Error('\P{Canonical_Combining_Class=:=-CCC22}');
Expect(1, 1469, '\p{Canonical_Combining_Class=ccc22}', "");
Expect(0, 1469, '\p{^Canonical_Combining_Class=ccc22}', "");
Expect(0, 1469, '\P{Canonical_Combining_Class=ccc22}', "");
Expect(1, 1469, '\P{^Canonical_Combining_Class=ccc22}', "");
Expect(0, 1470, '\p{Canonical_Combining_Class=ccc22}', "");
Expect(1, 1470, '\p{^Canonical_Combining_Class=ccc22}', "");
Expect(1, 1470, '\P{Canonical_Combining_Class=ccc22}', "");
Expect(0, 1470, '\P{^Canonical_Combining_Class=ccc22}', "");
Expect(1, 1469, '\p{Canonical_Combining_Class=- CCC22}', "");
Expect(0, 1469, '\p{^Canonical_Combining_Class=- CCC22}', "");
Expect(0, 1469, '\P{Canonical_Combining_Class=- CCC22}', "");
Expect(1, 1469, '\P{^Canonical_Combining_Class=- CCC22}', "");
Expect(0, 1470, '\p{Canonical_Combining_Class=- CCC22}', "");
Expect(1, 1470, '\p{^Canonical_Combining_Class=- CCC22}', "");
Expect(1, 1470, '\P{Canonical_Combining_Class=- CCC22}', "");
Expect(0, 1470, '\P{^Canonical_Combining_Class=- CCC22}', "");
Error('\p{Ccc= 	2_2/a/}');
Error('\P{Ccc= 	2_2/a/}');
Expect(1, 1469, '\p{Ccc=+0_0_0_0_0_0_0_22}', "");
Expect(0, 1469, '\p{^Ccc=+0_0_0_0_0_0_0_22}', "");
Expect(0, 1469, '\P{Ccc=+0_0_0_0_0_0_0_22}', "");
Expect(1, 1469, '\P{^Ccc=+0_0_0_0_0_0_0_22}', "");
Expect(0, 1470, '\p{Ccc=+0_0_0_0_0_0_0_22}', "");
Expect(1, 1470, '\p{^Ccc=+0_0_0_0_0_0_0_22}', "");
Expect(1, 1470, '\P{Ccc=+0_0_0_0_0_0_0_22}', "");
Expect(0, 1470, '\P{^Ccc=+0_0_0_0_0_0_0_22}', "");
Error('\p{Is_Canonical_Combining_Class=_:=CCC22}');
Error('\P{Is_Canonical_Combining_Class=_:=CCC22}');
Expect(1, 1469, '\p{Is_Canonical_Combining_Class=ccc22}', "");
Expect(0, 1469, '\p{^Is_Canonical_Combining_Class=ccc22}', "");
Expect(0, 1469, '\P{Is_Canonical_Combining_Class=ccc22}', "");
Expect(1, 1469, '\P{^Is_Canonical_Combining_Class=ccc22}', "");
Expect(0, 1470, '\p{Is_Canonical_Combining_Class=ccc22}', "");
Expect(1, 1470, '\p{^Is_Canonical_Combining_Class=ccc22}', "");
Expect(1, 1470, '\P{Is_Canonical_Combining_Class=ccc22}', "");
Expect(0, 1470, '\P{^Is_Canonical_Combining_Class=ccc22}', "");
Expect(1, 1469, '\p{Is_Canonical_Combining_Class=--ccc22}', "");
Expect(0, 1469, '\p{^Is_Canonical_Combining_Class=--ccc22}', "");
Expect(0, 1469, '\P{Is_Canonical_Combining_Class=--ccc22}', "");
Expect(1, 1469, '\P{^Is_Canonical_Combining_Class=--ccc22}', "");
Expect(0, 1470, '\p{Is_Canonical_Combining_Class=--ccc22}', "");
Expect(1, 1470, '\p{^Is_Canonical_Combining_Class=--ccc22}', "");
Expect(1, 1470, '\P{Is_Canonical_Combining_Class=--ccc22}', "");
Expect(0, 1470, '\P{^Is_Canonical_Combining_Class=--ccc22}', "");
Error('\p{Is_Ccc= :=000000022}');
Error('\P{Is_Ccc= :=000000022}');
Expect(1, 1469, '\p{Is_Ccc=00000000022}', "");
Expect(0, 1469, '\p{^Is_Ccc=00000000022}', "");
Expect(0, 1469, '\P{Is_Ccc=00000000022}', "");
Expect(1, 1469, '\P{^Is_Ccc=00000000022}', "");
Expect(0, 1470, '\p{Is_Ccc=00000000022}', "");
Expect(1, 1470, '\p{^Is_Ccc=00000000022}', "");
Expect(1, 1470, '\P{Is_Ccc=00000000022}', "");
Expect(0, 1470, '\P{^Is_Ccc=00000000022}', "");
Error('\p{Canonical_Combining_Class=	ABOVE_left:=}');
Error('\P{Canonical_Combining_Class=	ABOVE_left:=}');
Expect(1, 12331, '\p{Canonical_Combining_Class=aboveleft}', "");
Expect(0, 12331, '\p{^Canonical_Combining_Class=aboveleft}', "");
Expect(0, 12331, '\P{Canonical_Combining_Class=aboveleft}', "");
Expect(1, 12331, '\P{^Canonical_Combining_Class=aboveleft}', "");
Expect(0, 12332, '\p{Canonical_Combining_Class=aboveleft}', "");
Expect(1, 12332, '\p{^Canonical_Combining_Class=aboveleft}', "");
Expect(1, 12332, '\P{Canonical_Combining_Class=aboveleft}', "");
Expect(0, 12332, '\P{^Canonical_Combining_Class=aboveleft}', "");
Expect(1, 12331, '\p{Canonical_Combining_Class=--ABOVE_left}', "");
Expect(0, 12331, '\p{^Canonical_Combining_Class=--ABOVE_left}', "");
Expect(0, 12331, '\P{Canonical_Combining_Class=--ABOVE_left}', "");
Expect(1, 12331, '\P{^Canonical_Combining_Class=--ABOVE_left}', "");
Expect(0, 12332, '\p{Canonical_Combining_Class=--ABOVE_left}', "");
Expect(1, 12332, '\p{^Canonical_Combining_Class=--ABOVE_left}', "");
Expect(1, 12332, '\P{Canonical_Combining_Class=--ABOVE_left}', "");
Expect(0, 12332, '\P{^Canonical_Combining_Class=--ABOVE_left}', "");
Error('\p{Ccc=/a/-_AL}');
Error('\P{Ccc=/a/-_AL}');
Expect(1, 12331, '\p{Ccc=al}', "");
Expect(0, 12331, '\p{^Ccc=al}', "");
Expect(0, 12331, '\P{Ccc=al}', "");
Expect(1, 12331, '\P{^Ccc=al}', "");
Expect(0, 12332, '\p{Ccc=al}', "");
Expect(1, 12332, '\p{^Ccc=al}', "");
Expect(1, 12332, '\P{Ccc=al}', "");
Expect(0, 12332, '\P{^Ccc=al}', "");
Expect(1, 12331, '\p{Ccc=	AL}', "");
Expect(0, 12331, '\p{^Ccc=	AL}', "");
Expect(0, 12331, '\P{Ccc=	AL}', "");
Expect(1, 12331, '\P{^Ccc=	AL}', "");
Expect(0, 12332, '\p{Ccc=	AL}', "");
Expect(1, 12332, '\p{^Ccc=	AL}', "");
Expect(1, 12332, '\P{Ccc=	AL}', "");
Expect(0, 12332, '\P{^Ccc=	AL}', "");
Error('\p{Is_Canonical_Combining_Class=-/a/000_022_8}');
Error('\P{Is_Canonical_Combining_Class=-/a/000_022_8}');
Expect(1, 12331, '\p{Is_Canonical_Combining_Class=0022_8}', "");
Expect(0, 12331, '\p{^Is_Canonical_Combining_Class=0022_8}', "");
Expect(0, 12331, '\P{Is_Canonical_Combining_Class=0022_8}', "");
Expect(1, 12331, '\P{^Is_Canonical_Combining_Class=0022_8}', "");
Expect(0, 12332, '\p{Is_Canonical_Combining_Class=0022_8}', "");
Expect(1, 12332, '\p{^Is_Canonical_Combining_Class=0022_8}', "");
Expect(1, 12332, '\P{Is_Canonical_Combining_Class=0022_8}', "");
Expect(0, 12332, '\P{^Is_Canonical_Combining_Class=0022_8}', "");
Error('\p{Is_Ccc=-/a/above_LEFT}');
Error('\P{Is_Ccc=-/a/above_LEFT}');
Expect(1, 12331, '\p{Is_Ccc=aboveleft}', "");
Expect(0, 12331, '\p{^Is_Ccc=aboveleft}', "");
Expect(0, 12331, '\P{Is_Ccc=aboveleft}', "");
Expect(1, 12331, '\P{^Is_Ccc=aboveleft}', "");
Expect(0, 12332, '\p{Is_Ccc=aboveleft}', "");
Expect(1, 12332, '\p{^Is_Ccc=aboveleft}', "");
Expect(1, 12332, '\P{Is_Ccc=aboveleft}', "");
Expect(0, 12332, '\P{^Is_Ccc=aboveleft}', "");
Expect(1, 12331, '\p{Is_Ccc=		Above_Left}', "");
Expect(0, 12331, '\p{^Is_Ccc=		Above_Left}', "");
Expect(0, 12331, '\P{Is_Ccc=		Above_Left}', "");
Expect(1, 12331, '\P{^Is_Ccc=		Above_Left}', "");
Expect(0, 12332, '\p{Is_Ccc=		Above_Left}', "");
Expect(1, 12332, '\p{^Is_Ccc=		Above_Left}', "");
Expect(1, 12332, '\P{Is_Ccc=		Above_Left}', "");
Expect(0, 12332, '\P{^Is_Ccc=		Above_Left}', "");
Error('\p{Canonical_Combining_Class=:=	OVERLAY}');
Error('\P{Canonical_Combining_Class=:=	OVERLAY}');
Expect(1, 119145, '\p{Canonical_Combining_Class=overlay}', "");
Expect(0, 119145, '\p{^Canonical_Combining_Class=overlay}', "");
Expect(0, 119145, '\P{Canonical_Combining_Class=overlay}', "");
Expect(1, 119145, '\P{^Canonical_Combining_Class=overlay}', "");
Expect(0, 119146, '\p{Canonical_Combining_Class=overlay}', "");
Expect(1, 119146, '\p{^Canonical_Combining_Class=overlay}', "");
Expect(1, 119146, '\P{Canonical_Combining_Class=overlay}', "");
Expect(0, 119146, '\P{^Canonical_Combining_Class=overlay}', "");
Expect(1, 119145, '\p{Canonical_Combining_Class=_-Overlay}', "");
Expect(0, 119145, '\p{^Canonical_Combining_Class=_-Overlay}', "");
Expect(0, 119145, '\P{Canonical_Combining_Class=_-Overlay}', "");
Expect(1, 119145, '\P{^Canonical_Combining_Class=_-Overlay}', "");
Expect(0, 119146, '\p{Canonical_Combining_Class=_-Overlay}', "");
Expect(1, 119146, '\p{^Canonical_Combining_Class=_-Overlay}', "");
Expect(1, 119146, '\P{Canonical_Combining_Class=_-Overlay}', "");
Expect(0, 119146, '\P{^Canonical_Combining_Class=_-Overlay}', "");
Error('\p{Ccc=/a/_OV}');
Error('\P{Ccc=/a/_OV}');
Expect(1, 119145, '\p{Ccc=ov}', "");
Expect(0, 119145, '\p{^Ccc=ov}', "");
Expect(0, 119145, '\P{Ccc=ov}', "");
Expect(1, 119145, '\P{^Ccc=ov}', "");
Expect(0, 119146, '\p{Ccc=ov}', "");
Expect(1, 119146, '\p{^Ccc=ov}', "");
Expect(1, 119146, '\P{Ccc=ov}', "");
Expect(0, 119146, '\P{^Ccc=ov}', "");
Expect(1, 119145, '\p{Ccc=__OV}', "");
Expect(0, 119145, '\p{^Ccc=__OV}', "");
Expect(0, 119145, '\P{Ccc=__OV}', "");
Expect(1, 119145, '\P{^Ccc=__OV}', "");
Expect(0, 119146, '\p{Ccc=__OV}', "");
Expect(1, 119146, '\p{^Ccc=__OV}', "");
Expect(1, 119146, '\P{Ccc=__OV}', "");
Expect(0, 119146, '\P{^Ccc=__OV}', "");
Error('\p{Is_Canonical_Combining_Class=/a/-00001}');
Error('\P{Is_Canonical_Combining_Class=/a/-00001}');
Expect(1, 119145, '\p{Is_Canonical_Combining_Class=0_0_0_001}', "");
Expect(0, 119145, '\p{^Is_Canonical_Combining_Class=0_0_0_001}', "");
Expect(0, 119145, '\P{Is_Canonical_Combining_Class=0_0_0_001}', "");
Expect(1, 119145, '\P{^Is_Canonical_Combining_Class=0_0_0_001}', "");
Expect(0, 119146, '\p{Is_Canonical_Combining_Class=0_0_0_001}', "");
Expect(1, 119146, '\p{^Is_Canonical_Combining_Class=0_0_0_001}', "");
Expect(1, 119146, '\P{Is_Canonical_Combining_Class=0_0_0_001}', "");
Expect(0, 119146, '\P{^Is_Canonical_Combining_Class=0_0_0_001}', "");
Error('\p{Is_Ccc=	_OVERLAY:=}');
Error('\P{Is_Ccc=	_OVERLAY:=}');
Expect(1, 119145, '\p{Is_Ccc=overlay}', "");
Expect(0, 119145, '\p{^Is_Ccc=overlay}', "");
Expect(0, 119145, '\P{Is_Ccc=overlay}', "");
Expect(1, 119145, '\P{^Is_Ccc=overlay}', "");
Expect(0, 119146, '\p{Is_Ccc=overlay}', "");
Expect(1, 119146, '\p{^Is_Ccc=overlay}', "");
Expect(1, 119146, '\P{Is_Ccc=overlay}', "");
Expect(0, 119146, '\P{^Is_Ccc=overlay}', "");
Expect(1, 119145, '\p{Is_Ccc=_overlay}', "");
Expect(0, 119145, '\p{^Is_Ccc=_overlay}', "");
Expect(0, 119145, '\P{Is_Ccc=_overlay}', "");
Expect(1, 119145, '\P{^Is_Ccc=_overlay}', "");
Expect(0, 119146, '\p{Is_Ccc=_overlay}', "");
Expect(1, 119146, '\p{^Is_Ccc=_overlay}', "");
Expect(1, 119146, '\P{Is_Ccc=_overlay}', "");
Expect(0, 119146, '\P{^Is_Ccc=_overlay}', "");
Error('\p{Canonical_Combining_Class=	-virama:=}');
Error('\P{Canonical_Combining_Class=	-virama:=}');
Expect(1, 71350, '\p{Canonical_Combining_Class=virama}', "");
Expect(0, 71350, '\p{^Canonical_Combining_Class=virama}', "");
Expect(0, 71350, '\P{Canonical_Combining_Class=virama}', "");
Expect(1, 71350, '\P{^Canonical_Combining_Class=virama}', "");
Expect(0, 71351, '\p{Canonical_Combining_Class=virama}', "");
Expect(1, 71351, '\p{^Canonical_Combining_Class=virama}', "");
Expect(1, 71351, '\P{Canonical_Combining_Class=virama}', "");
Expect(0, 71351, '\P{^Canonical_Combining_Class=virama}', "");
Expect(1, 71350, '\p{Canonical_Combining_Class=  VIRAMA}', "");
Expect(0, 71350, '\p{^Canonical_Combining_Class=  VIRAMA}', "");
Expect(0, 71350, '\P{Canonical_Combining_Class=  VIRAMA}', "");
Expect(1, 71350, '\P{^Canonical_Combining_Class=  VIRAMA}', "");
Expect(0, 71351, '\p{Canonical_Combining_Class=  VIRAMA}', "");
Expect(1, 71351, '\p{^Canonical_Combining_Class=  VIRAMA}', "");
Expect(1, 71351, '\P{Canonical_Combining_Class=  VIRAMA}', "");
Expect(0, 71351, '\P{^Canonical_Combining_Class=  VIRAMA}', "");
Error('\p{Ccc=:=VR}');
Error('\P{Ccc=:=VR}');
Expect(1, 71350, '\p{Ccc: vr}', "");
Expect(0, 71350, '\p{^Ccc: vr}', "");
Expect(0, 71350, '\P{Ccc: vr}', "");
Expect(1, 71350, '\P{^Ccc: vr}', "");
Expect(0, 71351, '\p{Ccc: vr}', "");
Expect(1, 71351, '\p{^Ccc: vr}', "");
Expect(1, 71351, '\P{Ccc: vr}', "");
Expect(0, 71351, '\P{^Ccc: vr}', "");
Expect(1, 71350, '\p{Ccc:  -VR}', "");
Expect(0, 71350, '\p{^Ccc:  -VR}', "");
Expect(0, 71350, '\P{Ccc:  -VR}', "");
Expect(1, 71350, '\P{^Ccc:  -VR}', "");
Expect(0, 71351, '\p{Ccc:  -VR}', "");
Expect(1, 71351, '\p{^Ccc:  -VR}', "");
Expect(1, 71351, '\P{Ccc:  -VR}', "");
Expect(0, 71351, '\P{^Ccc:  -VR}', "");
Error('\p{Is_Canonical_Combining_Class=/a/_	+00000000_9}');
Error('\P{Is_Canonical_Combining_Class=/a/_	+00000000_9}');
Expect(1, 71350, '\p{Is_Canonical_Combining_Class=0_0_0_0_09}', "");
Expect(0, 71350, '\p{^Is_Canonical_Combining_Class=0_0_0_0_09}', "");
Expect(0, 71350, '\P{Is_Canonical_Combining_Class=0_0_0_0_09}', "");
Expect(1, 71350, '\P{^Is_Canonical_Combining_Class=0_0_0_0_09}', "");
Expect(0, 71351, '\p{Is_Canonical_Combining_Class=0_0_0_0_09}', "");
Expect(1, 71351, '\p{^Is_Canonical_Combining_Class=0_0_0_0_09}', "");
Expect(1, 71351, '\P{Is_Canonical_Combining_Class=0_0_0_0_09}', "");
Expect(0, 71351, '\P{^Is_Canonical_Combining_Class=0_0_0_0_09}', "");
Error('\p{Is_Ccc= :=Virama}');
Error('\P{Is_Ccc= :=Virama}');
Expect(1, 71350, '\p{Is_Ccc=virama}', "");
Expect(0, 71350, '\p{^Is_Ccc=virama}', "");
Expect(0, 71350, '\P{Is_Ccc=virama}', "");
Expect(1, 71350, '\P{^Is_Ccc=virama}', "");
Expect(0, 71351, '\p{Is_Ccc=virama}', "");
Expect(1, 71351, '\p{^Is_Ccc=virama}', "");
Expect(1, 71351, '\P{Is_Ccc=virama}', "");
Expect(0, 71351, '\P{^Is_Ccc=virama}', "");
Expect(1, 71350, '\p{Is_Ccc= _Virama}', "");
Expect(0, 71350, '\p{^Is_Ccc= _Virama}', "");
Expect(0, 71350, '\P{Is_Ccc= _Virama}', "");
Expect(1, 71350, '\P{^Is_Ccc= _Virama}', "");
Expect(0, 71351, '\p{Is_Ccc= _Virama}', "");
Expect(1, 71351, '\p{^Is_Ccc= _Virama}', "");
Expect(1, 71351, '\P{Is_Ccc= _Virama}', "");
Expect(0, 71351, '\P{^Is_Ccc= _Virama}', "");
Error('\p{Canonical_Combining_Class=-attached_Below/a/}');
Error('\P{Canonical_Combining_Class=-attached_Below/a/}');
Expect(1, 7632, '\p{Canonical_Combining_Class=attachedbelow}', "");
Expect(0, 7632, '\p{^Canonical_Combining_Class=attachedbelow}', "");
Expect(0, 7632, '\P{Canonical_Combining_Class=attachedbelow}', "");
Expect(1, 7632, '\P{^Canonical_Combining_Class=attachedbelow}', "");
Expect(0, 7633, '\p{Canonical_Combining_Class=attachedbelow}', "");
Expect(1, 7633, '\p{^Canonical_Combining_Class=attachedbelow}', "");
Expect(1, 7633, '\P{Canonical_Combining_Class=attachedbelow}', "");
Expect(0, 7633, '\P{^Canonical_Combining_Class=attachedbelow}', "");
Expect(1, 7632, '\p{Canonical_Combining_Class:  _ATTACHED_Below}', "");
Expect(0, 7632, '\p{^Canonical_Combining_Class:  _ATTACHED_Below}', "");
Expect(0, 7632, '\P{Canonical_Combining_Class:  _ATTACHED_Below}', "");
Expect(1, 7632, '\P{^Canonical_Combining_Class:  _ATTACHED_Below}', "");
Expect(0, 7633, '\p{Canonical_Combining_Class:  _ATTACHED_Below}', "");
Expect(1, 7633, '\p{^Canonical_Combining_Class:  _ATTACHED_Below}', "");
Expect(1, 7633, '\P{Canonical_Combining_Class:  _ATTACHED_Below}', "");
Expect(0, 7633, '\P{^Canonical_Combining_Class:  _ATTACHED_Below}', "");
Error('\p{Ccc=/a/ -ATB}');
Error('\P{Ccc=/a/ -ATB}');
Expect(1, 7632, '\p{Ccc=atb}', "");
Expect(0, 7632, '\p{^Ccc=atb}', "");
Expect(0, 7632, '\P{Ccc=atb}', "");
Expect(1, 7632, '\P{^Ccc=atb}', "");
Expect(0, 7633, '\p{Ccc=atb}', "");
Expect(1, 7633, '\p{^Ccc=atb}', "");
Expect(1, 7633, '\P{Ccc=atb}', "");
Expect(0, 7633, '\P{^Ccc=atb}', "");
Expect(1, 7632, '\p{Ccc: --ATB}', "");
Expect(0, 7632, '\p{^Ccc: --ATB}', "");
Expect(0, 7632, '\P{Ccc: --ATB}', "");
Expect(1, 7632, '\P{^Ccc: --ATB}', "");
Expect(0, 7633, '\p{Ccc: --ATB}', "");
Expect(1, 7633, '\p{^Ccc: --ATB}', "");
Expect(1, 7633, '\P{Ccc: --ATB}', "");
Expect(0, 7633, '\P{^Ccc: --ATB}', "");
Error('\p{Is_Canonical_Combining_Class=:=	+00_00_00_00_02_02}');
Error('\P{Is_Canonical_Combining_Class=:=	+00_00_00_00_02_02}');
Expect(1, 7632, '\p{Is_Canonical_Combining_Class=+0000000202}', "");
Expect(0, 7632, '\p{^Is_Canonical_Combining_Class=+0000000202}', "");
Expect(0, 7632, '\P{Is_Canonical_Combining_Class=+0000000202}', "");
Expect(1, 7632, '\P{^Is_Canonical_Combining_Class=+0000000202}', "");
Expect(0, 7633, '\p{Is_Canonical_Combining_Class=+0000000202}', "");
Expect(1, 7633, '\p{^Is_Canonical_Combining_Class=+0000000202}', "");
Expect(1, 7633, '\P{Is_Canonical_Combining_Class=+0000000202}', "");
Expect(0, 7633, '\P{^Is_Canonical_Combining_Class=+0000000202}', "");
Error('\p{Is_Ccc=-/a/Attached_Below}');
Error('\P{Is_Ccc=-/a/Attached_Below}');
Expect(1, 7632, '\p{Is_Ccc=attachedbelow}', "");
Expect(0, 7632, '\p{^Is_Ccc=attachedbelow}', "");
Expect(0, 7632, '\P{Is_Ccc=attachedbelow}', "");
Expect(1, 7632, '\P{^Is_Ccc=attachedbelow}', "");
Expect(0, 7633, '\p{Is_Ccc=attachedbelow}', "");
Expect(1, 7633, '\p{^Is_Ccc=attachedbelow}', "");
Expect(1, 7633, '\P{Is_Ccc=attachedbelow}', "");
Expect(0, 7633, '\P{^Is_Ccc=attachedbelow}', "");
Expect(1, 7632, '\p{Is_Ccc= attached_below}', "");
Expect(0, 7632, '\p{^Is_Ccc= attached_below}', "");
Expect(0, 7632, '\P{Is_Ccc= attached_below}', "");
Expect(1, 7632, '\P{^Is_Ccc= attached_below}', "");
Expect(0, 7633, '\p{Is_Ccc= attached_below}', "");
Expect(1, 7633, '\p{^Is_Ccc= attached_below}', "");
Expect(1, 7633, '\P{Is_Ccc= attached_below}', "");
Expect(0, 7633, '\P{^Is_Ccc= attached_below}', "");
Error('\p{Canonical_Combining_Class=_CCC14/a/}');
Error('\P{Canonical_Combining_Class=_CCC14/a/}');
Expect(1, 1460, '\p{Canonical_Combining_Class=ccc14}', "");
Expect(0, 1460, '\p{^Canonical_Combining_Class=ccc14}', "");
Expect(0, 1460, '\P{Canonical_Combining_Class=ccc14}', "");
Expect(1, 1460, '\P{^Canonical_Combining_Class=ccc14}', "");
Expect(0, 1461, '\p{Canonical_Combining_Class=ccc14}', "");
Expect(1, 1461, '\p{^Canonical_Combining_Class=ccc14}', "");
Expect(1, 1461, '\P{Canonical_Combining_Class=ccc14}', "");
Expect(0, 1461, '\P{^Canonical_Combining_Class=ccc14}', "");
Expect(1, 1460, '\p{Canonical_Combining_Class=-CCC14}', "");
Expect(0, 1460, '\p{^Canonical_Combining_Class=-CCC14}', "");
Expect(0, 1460, '\P{Canonical_Combining_Class=-CCC14}', "");
Expect(1, 1460, '\P{^Canonical_Combining_Class=-CCC14}', "");
Expect(0, 1461, '\p{Canonical_Combining_Class=-CCC14}', "");
Expect(1, 1461, '\p{^Canonical_Combining_Class=-CCC14}', "");
Expect(1, 1461, '\P{Canonical_Combining_Class=-CCC14}', "");
Expect(0, 1461, '\P{^Canonical_Combining_Class=-CCC14}', "");
Error('\p{Ccc=-_+00_01_4:=}');
Error('\P{Ccc=-_+00_01_4:=}');
Expect(1, 1460, '\p{Ccc=+014}', "");
Expect(0, 1460, '\p{^Ccc=+014}', "");
Expect(0, 1460, '\P{Ccc=+014}', "");
Expect(1, 1460, '\P{^Ccc=+014}', "");
Expect(0, 1461, '\p{Ccc=+014}', "");
Expect(1, 1461, '\p{^Ccc=+014}', "");
Expect(1, 1461, '\P{Ccc=+014}', "");
Expect(0, 1461, '\P{^Ccc=+014}', "");
Error('\p{Is_Canonical_Combining_Class=/a/		CCC14}');
Error('\P{Is_Canonical_Combining_Class=/a/		CCC14}');
Expect(1, 1460, '\p{Is_Canonical_Combining_Class=ccc14}', "");
Expect(0, 1460, '\p{^Is_Canonical_Combining_Class=ccc14}', "");
Expect(0, 1460, '\P{Is_Canonical_Combining_Class=ccc14}', "");
Expect(1, 1460, '\P{^Is_Canonical_Combining_Class=ccc14}', "");
Expect(0, 1461, '\p{Is_Canonical_Combining_Class=ccc14}', "");
Expect(1, 1461, '\p{^Is_Canonical_Combining_Class=ccc14}', "");
Expect(1, 1461, '\P{Is_Canonical_Combining_Class=ccc14}', "");
Expect(0, 1461, '\P{^Is_Canonical_Combining_Class=ccc14}', "");
Expect(1, 1460, '\p{Is_Canonical_Combining_Class=_-ccc14}', "");
Expect(0, 1460, '\p{^Is_Canonical_Combining_Class=_-ccc14}', "");
Expect(0, 1460, '\P{Is_Canonical_Combining_Class=_-ccc14}', "");
Expect(1, 1460, '\P{^Is_Canonical_Combining_Class=_-ccc14}', "");
Expect(0, 1461, '\p{Is_Canonical_Combining_Class=_-ccc14}', "");
Expect(1, 1461, '\p{^Is_Canonical_Combining_Class=_-ccc14}', "");
Expect(1, 1461, '\P{Is_Canonical_Combining_Class=_-ccc14}', "");
Expect(0, 1461, '\P{^Is_Canonical_Combining_Class=_-ccc14}', "");
Error('\p{Is_Ccc=- 0_0_0_0_0_0_00014:=}');
Error('\P{Is_Ccc=- 0_0_0_0_0_0_00014:=}');
Expect(1, 1460, '\p{Is_Ccc=+0000014}', "");
Expect(0, 1460, '\p{^Is_Ccc=+0000014}', "");
Expect(0, 1460, '\P{Is_Ccc=+0000014}', "");
Expect(1, 1460, '\P{^Is_Ccc=+0000014}', "");
Expect(0, 1461, '\p{Is_Ccc=+0000014}', "");
Expect(1, 1461, '\p{^Is_Ccc=+0000014}', "");
Expect(1, 1461, '\P{Is_Ccc=+0000014}', "");
Expect(0, 1461, '\P{^Is_Ccc=+0000014}', "");
Error('\p{Canonical_Combining_Class=:= ccc24}');
Error('\P{Canonical_Combining_Class=:= ccc24}');
Expect(1, 1473, '\p{Canonical_Combining_Class=ccc24}', "");
Expect(0, 1473, '\p{^Canonical_Combining_Class=ccc24}', "");
Expect(0, 1473, '\P{Canonical_Combining_Class=ccc24}', "");
Expect(1, 1473, '\P{^Canonical_Combining_Class=ccc24}', "");
Expect(0, 1474, '\p{Canonical_Combining_Class=ccc24}', "");
Expect(1, 1474, '\p{^Canonical_Combining_Class=ccc24}', "");
Expect(1, 1474, '\P{Canonical_Combining_Class=ccc24}', "");
Expect(0, 1474, '\P{^Canonical_Combining_Class=ccc24}', "");
Expect(1, 1473, '\p{Canonical_Combining_Class=CCC24}', "");
Expect(0, 1473, '\p{^Canonical_Combining_Class=CCC24}', "");
Expect(0, 1473, '\P{Canonical_Combining_Class=CCC24}', "");
Expect(1, 1473, '\P{^Canonical_Combining_Class=CCC24}', "");
Expect(0, 1474, '\p{Canonical_Combining_Class=CCC24}', "");
Expect(1, 1474, '\p{^Canonical_Combining_Class=CCC24}', "");
Expect(1, 1474, '\P{Canonical_Combining_Class=CCC24}', "");
Expect(0, 1474, '\P{^Canonical_Combining_Class=CCC24}', "");
Error('\p{Ccc=:=__00000000024}');
Error('\P{Ccc=:=__00000000024}');
Expect(1, 1473, '\p{Ccc=+00000024}', "");
Expect(0, 1473, '\p{^Ccc=+00000024}', "");
Expect(0, 1473, '\P{Ccc=+00000024}', "");
Expect(1, 1473, '\P{^Ccc=+00000024}', "");
Expect(0, 1474, '\p{Ccc=+00000024}', "");
Expect(1, 1474, '\p{^Ccc=+00000024}', "");
Expect(1, 1474, '\P{Ccc=+00000024}', "");
Expect(0, 1474, '\P{^Ccc=+00000024}', "");
Error('\p{Is_Canonical_Combining_Class:   _	CCC24/a/}');
Error('\P{Is_Canonical_Combining_Class:   _	CCC24/a/}');
Expect(1, 1473, '\p{Is_Canonical_Combining_Class=ccc24}', "");
Expect(0, 1473, '\p{^Is_Canonical_Combining_Class=ccc24}', "");
Expect(0, 1473, '\P{Is_Canonical_Combining_Class=ccc24}', "");
Expect(1, 1473, '\P{^Is_Canonical_Combining_Class=ccc24}', "");
Expect(0, 1474, '\p{Is_Canonical_Combining_Class=ccc24}', "");
Expect(1, 1474, '\p{^Is_Canonical_Combining_Class=ccc24}', "");
Expect(1, 1474, '\P{Is_Canonical_Combining_Class=ccc24}', "");
Expect(0, 1474, '\P{^Is_Canonical_Combining_Class=ccc24}', "");
Expect(1, 1473, '\p{Is_Canonical_Combining_Class=-ccc24}', "");
Expect(0, 1473, '\p{^Is_Canonical_Combining_Class=-ccc24}', "");
Expect(0, 1473, '\P{Is_Canonical_Combining_Class=-ccc24}', "");
Expect(1, 1473, '\P{^Is_Canonical_Combining_Class=-ccc24}', "");
Expect(0, 1474, '\p{Is_Canonical_Combining_Class=-ccc24}', "");
Expect(1, 1474, '\p{^Is_Canonical_Combining_Class=-ccc24}', "");
Expect(1, 1474, '\P{Is_Canonical_Combining_Class=-ccc24}', "");
Expect(0, 1474, '\P{^Is_Canonical_Combining_Class=-ccc24}', "");
Error('\p{Is_Ccc=/a/00_00_00_024}');
Error('\P{Is_Ccc=/a/00_00_00_024}');
Expect(1, 1473, '\p{Is_Ccc=0_0_24}', "");
Expect(0, 1473, '\p{^Is_Ccc=0_0_24}', "");
Expect(0, 1473, '\P{Is_Ccc=0_0_24}', "");
Expect(1, 1473, '\P{^Is_Ccc=0_0_24}', "");
Expect(0, 1474, '\p{Is_Ccc=0_0_24}', "");
Expect(1, 1474, '\p{^Is_Ccc=0_0_24}', "");
Expect(1, 1474, '\P{Is_Ccc=0_0_24}', "");
Expect(0, 1474, '\P{^Is_Ccc=0_0_24}', "");
Error('\p{Canonical_Combining_Class:	-_ccc25/a/}');
Error('\P{Canonical_Combining_Class:	-_ccc25/a/}');
Expect(1, 1474, '\p{Canonical_Combining_Class=ccc25}', "");
Expect(0, 1474, '\p{^Canonical_Combining_Class=ccc25}', "");
Expect(0, 1474, '\P{Canonical_Combining_Class=ccc25}', "");
Expect(1, 1474, '\P{^Canonical_Combining_Class=ccc25}', "");
Expect(0, 1475, '\p{Canonical_Combining_Class=ccc25}', "");
Expect(1, 1475, '\p{^Canonical_Combining_Class=ccc25}', "");
Expect(1, 1475, '\P{Canonical_Combining_Class=ccc25}', "");
Expect(0, 1475, '\P{^Canonical_Combining_Class=ccc25}', "");
Expect(1, 1474, '\p{Canonical_Combining_Class=-CCC25}', "");
Expect(0, 1474, '\p{^Canonical_Combining_Class=-CCC25}', "");
Expect(0, 1474, '\P{Canonical_Combining_Class=-CCC25}', "");
Expect(1, 1474, '\P{^Canonical_Combining_Class=-CCC25}', "");
Expect(0, 1475, '\p{Canonical_Combining_Class=-CCC25}', "");
Expect(1, 1475, '\p{^Canonical_Combining_Class=-CCC25}', "");
Expect(1, 1475, '\P{Canonical_Combining_Class=-CCC25}', "");
Expect(0, 1475, '\P{^Canonical_Combining_Class=-CCC25}', "");
Error('\p{Ccc:-:=2_5}');
Error('\P{Ccc:-:=2_5}');
Expect(1, 1474, '\p{Ccc=0025}', "");
Expect(0, 1474, '\p{^Ccc=0025}', "");
Expect(0, 1474, '\P{Ccc=0025}', "");
Expect(1, 1474, '\P{^Ccc=0025}', "");
Expect(0, 1475, '\p{Ccc=0025}', "");
Expect(1, 1475, '\p{^Ccc=0025}', "");
Expect(1, 1475, '\P{Ccc=0025}', "");
Expect(0, 1475, '\P{^Ccc=0025}', "");
Error('\p{Is_Canonical_Combining_Class=/a/--CCC25}');
Error('\P{Is_Canonical_Combining_Class=/a/--CCC25}');
Expect(1, 1474, '\p{Is_Canonical_Combining_Class=ccc25}', "");
Expect(0, 1474, '\p{^Is_Canonical_Combining_Class=ccc25}', "");
Expect(0, 1474, '\P{Is_Canonical_Combining_Class=ccc25}', "");
Expect(1, 1474, '\P{^Is_Canonical_Combining_Class=ccc25}', "");
Expect(0, 1475, '\p{Is_Canonical_Combining_Class=ccc25}', "");
Expect(1, 1475, '\p{^Is_Canonical_Combining_Class=ccc25}', "");
Expect(1, 1475, '\P{Is_Canonical_Combining_Class=ccc25}', "");
Expect(0, 1475, '\P{^Is_Canonical_Combining_Class=ccc25}', "");
Expect(1, 1474, '\p{Is_Canonical_Combining_Class=	-CCC25}', "");
Expect(0, 1474, '\p{^Is_Canonical_Combining_Class=	-CCC25}', "");
Expect(0, 1474, '\P{Is_Canonical_Combining_Class=	-CCC25}', "");
Expect(1, 1474, '\P{^Is_Canonical_Combining_Class=	-CCC25}', "");
Expect(0, 1475, '\p{Is_Canonical_Combining_Class=	-CCC25}', "");
Expect(1, 1475, '\p{^Is_Canonical_Combining_Class=	-CCC25}', "");
Expect(1, 1475, '\P{Is_Canonical_Combining_Class=	-CCC25}', "");
Expect(0, 1475, '\P{^Is_Canonical_Combining_Class=	-CCC25}', "");
Error('\p{Is_Ccc:   -/a/+000025}');
Error('\P{Is_Ccc:   -/a/+000025}');
Expect(1, 1474, '\p{Is_Ccc=000025}', "");
Expect(0, 1474, '\p{^Is_Ccc=000025}', "");
Expect(0, 1474, '\P{Is_Ccc=000025}', "");
Expect(1, 1474, '\P{^Is_Ccc=000025}', "");
Expect(0, 1475, '\p{Is_Ccc=000025}', "");
Expect(1, 1475, '\p{^Is_Ccc=000025}', "");
Expect(1, 1475, '\P{Is_Ccc=000025}', "");
Expect(0, 1475, '\P{^Is_Ccc=000025}', "");
Error('\p{Canonical_Combining_Class=- ccc33:=}');
Error('\P{Canonical_Combining_Class=- ccc33:=}');
Expect(1, 1617, '\p{Canonical_Combining_Class=ccc33}', "");
Expect(0, 1617, '\p{^Canonical_Combining_Class=ccc33}', "");
Expect(0, 1617, '\P{Canonical_Combining_Class=ccc33}', "");
Expect(1, 1617, '\P{^Canonical_Combining_Class=ccc33}', "");
Expect(0, 1618, '\p{Canonical_Combining_Class=ccc33}', "");
Expect(1, 1618, '\p{^Canonical_Combining_Class=ccc33}', "");
Expect(1, 1618, '\P{Canonical_Combining_Class=ccc33}', "");
Expect(0, 1618, '\P{^Canonical_Combining_Class=ccc33}', "");
Expect(1, 1617, '\p{Canonical_Combining_Class=_	CCC33}', "");
Expect(0, 1617, '\p{^Canonical_Combining_Class=_	CCC33}', "");
Expect(0, 1617, '\P{Canonical_Combining_Class=_	CCC33}', "");
Expect(1, 1617, '\P{^Canonical_Combining_Class=_	CCC33}', "");
Expect(0, 1618, '\p{Canonical_Combining_Class=_	CCC33}', "");
Expect(1, 1618, '\p{^Canonical_Combining_Class=_	CCC33}', "");
Expect(1, 1618, '\P{Canonical_Combining_Class=_	CCC33}', "");
Expect(0, 1618, '\P{^Canonical_Combining_Class=_	CCC33}', "");
Error('\p{Ccc= /a/0_0_33}');
Error('\P{Ccc= /a/0_0_33}');
Expect(1, 1617, '\p{Ccc:	0_0_0_0_0_0_0_033}', "");
Expect(0, 1617, '\p{^Ccc:	0_0_0_0_0_0_0_033}', "");
Expect(0, 1617, '\P{Ccc:	0_0_0_0_0_0_0_033}', "");
Expect(1, 1617, '\P{^Ccc:	0_0_0_0_0_0_0_033}', "");
Expect(0, 1618, '\p{Ccc:	0_0_0_0_0_0_0_033}', "");
Expect(1, 1618, '\p{^Ccc:	0_0_0_0_0_0_0_033}', "");
Expect(1, 1618, '\P{Ccc:	0_0_0_0_0_0_0_033}', "");
Expect(0, 1618, '\P{^Ccc:	0_0_0_0_0_0_0_033}', "");
Error('\p{Is_Canonical_Combining_Class=	_CCC33:=}');
Error('\P{Is_Canonical_Combining_Class=	_CCC33:=}');
Expect(1, 1617, '\p{Is_Canonical_Combining_Class=ccc33}', "");
Expect(0, 1617, '\p{^Is_Canonical_Combining_Class=ccc33}', "");
Expect(0, 1617, '\P{Is_Canonical_Combining_Class=ccc33}', "");
Expect(1, 1617, '\P{^Is_Canonical_Combining_Class=ccc33}', "");
Expect(0, 1618, '\p{Is_Canonical_Combining_Class=ccc33}', "");
Expect(1, 1618, '\p{^Is_Canonical_Combining_Class=ccc33}', "");
Expect(1, 1618, '\P{Is_Canonical_Combining_Class=ccc33}', "");
Expect(0, 1618, '\P{^Is_Canonical_Combining_Class=ccc33}', "");
Expect(1, 1617, '\p{Is_Canonical_Combining_Class= _CCC33}', "");
Expect(0, 1617, '\p{^Is_Canonical_Combining_Class= _CCC33}', "");
Expect(0, 1617, '\P{Is_Canonical_Combining_Class= _CCC33}', "");
Expect(1, 1617, '\P{^Is_Canonical_Combining_Class= _CCC33}', "");
Expect(0, 1618, '\p{Is_Canonical_Combining_Class= _CCC33}', "");
Expect(1, 1618, '\p{^Is_Canonical_Combining_Class= _CCC33}', "");
Expect(1, 1618, '\P{Is_Canonical_Combining_Class= _CCC33}', "");
Expect(0, 1618, '\P{^Is_Canonical_Combining_Class= _CCC33}', "");
Error('\p{Is_Ccc:   -:=00000000033}');
Error('\P{Is_Ccc:   -:=00000000033}');
Expect(1, 1617, '\p{Is_Ccc=033}', "");
Expect(0, 1617, '\p{^Is_Ccc=033}', "");
Expect(0, 1617, '\P{Is_Ccc=033}', "");
Expect(1, 1617, '\P{^Is_Ccc=033}', "");
Expect(0, 1618, '\p{Is_Ccc=033}', "");
Expect(1, 1618, '\p{^Is_Ccc=033}', "");
Expect(1, 1618, '\P{Is_Ccc=033}', "");
Expect(0, 1618, '\P{^Is_Ccc=033}', "");
Error('\p{Canonical_Combining_Class=_ ccc10/a/}');
Error('\P{Canonical_Combining_Class=_ ccc10/a/}');
Expect(1, 1456, '\p{Canonical_Combining_Class:	ccc10}', "");
Expect(0, 1456, '\p{^Canonical_Combining_Class:	ccc10}', "");
Expect(0, 1456, '\P{Canonical_Combining_Class:	ccc10}', "");
Expect(1, 1456, '\P{^Canonical_Combining_Class:	ccc10}', "");
Expect(0, 1457, '\p{Canonical_Combining_Class:	ccc10}', "");
Expect(1, 1457, '\p{^Canonical_Combining_Class:	ccc10}', "");
Expect(1, 1457, '\P{Canonical_Combining_Class:	ccc10}', "");
Expect(0, 1457, '\P{^Canonical_Combining_Class:	ccc10}', "");
Expect(1, 1456, '\p{Canonical_Combining_Class=-_ccc10}', "");
Expect(0, 1456, '\p{^Canonical_Combining_Class=-_ccc10}', "");
Expect(0, 1456, '\P{Canonical_Combining_Class=-_ccc10}', "");
Expect(1, 1456, '\P{^Canonical_Combining_Class=-_ccc10}', "");
Expect(0, 1457, '\p{Canonical_Combining_Class=-_ccc10}', "");
Expect(1, 1457, '\p{^Canonical_Combining_Class=-_ccc10}', "");
Expect(1, 1457, '\P{Canonical_Combining_Class=-_ccc10}', "");
Expect(0, 1457, '\P{^Canonical_Combining_Class=-_ccc10}', "");
Error('\p{Ccc=_	+00_00_00_010:=}');
Error('\P{Ccc=_	+00_00_00_010:=}');
Expect(1, 1456, '\p{Ccc=+000_000_000_10}', "");
Expect(0, 1456, '\p{^Ccc=+000_000_000_10}', "");
Expect(0, 1456, '\P{Ccc=+000_000_000_10}', "");
Expect(1, 1456, '\P{^Ccc=+000_000_000_10}', "");
Expect(0, 1457, '\p{Ccc=+000_000_000_10}', "");
Expect(1, 1457, '\p{^Ccc=+000_000_000_10}', "");
Expect(1, 1457, '\P{Ccc=+000_000_000_10}', "");
Expect(0, 1457, '\P{^Ccc=+000_000_000_10}', "");
Error('\p{Is_Canonical_Combining_Class=-CCC10:=}');
Error('\P{Is_Canonical_Combining_Class=-CCC10:=}');
Expect(1, 1456, '\p{Is_Canonical_Combining_Class=ccc10}', "");
Expect(0, 1456, '\p{^Is_Canonical_Combining_Class=ccc10}', "");
Expect(0, 1456, '\P{Is_Canonical_Combining_Class=ccc10}', "");
Expect(1, 1456, '\P{^Is_Canonical_Combining_Class=ccc10}', "");
Expect(0, 1457, '\p{Is_Canonical_Combining_Class=ccc10}', "");
Expect(1, 1457, '\p{^Is_Canonical_Combining_Class=ccc10}', "");
Expect(1, 1457, '\P{Is_Canonical_Combining_Class=ccc10}', "");
Expect(0, 1457, '\P{^Is_Canonical_Combining_Class=ccc10}', "");
Expect(1, 1456, '\p{Is_Canonical_Combining_Class= CCC10}', "");
Expect(0, 1456, '\p{^Is_Canonical_Combining_Class= CCC10}', "");
Expect(0, 1456, '\P{Is_Canonical_Combining_Class= CCC10}', "");
Expect(1, 1456, '\P{^Is_Canonical_Combining_Class= CCC10}', "");
Expect(0, 1457, '\p{Is_Canonical_Combining_Class= CCC10}', "");
Expect(1, 1457, '\p{^Is_Canonical_Combining_Class= CCC10}', "");
Expect(1, 1457, '\P{Is_Canonical_Combining_Class= CCC10}', "");
Expect(0, 1457, '\P{^Is_Canonical_Combining_Class= CCC10}', "");
Error('\p{Is_Ccc=:=	-00_00_00_010}');
Error('\P{Is_Ccc=:=	-00_00_00_010}');
Expect(1, 1456, '\p{Is_Ccc:	+0_0_0_0_010}', "");
Expect(0, 1456, '\p{^Is_Ccc:	+0_0_0_0_010}', "");
Expect(0, 1456, '\P{Is_Ccc:	+0_0_0_0_010}', "");
Expect(1, 1456, '\P{^Is_Ccc:	+0_0_0_0_010}', "");
Expect(0, 1457, '\p{Is_Ccc:	+0_0_0_0_010}', "");
Expect(1, 1457, '\p{^Is_Ccc:	+0_0_0_0_010}', "");
Expect(1, 1457, '\P{Is_Ccc:	+0_0_0_0_010}', "");
Expect(0, 1457, '\P{^Is_Ccc:	+0_0_0_0_010}', "");
Error('\p{Canonical_Combining_Class=	Nukta:=}');
Error('\P{Canonical_Combining_Class=	Nukta:=}');
Expect(1, 71351, '\p{Canonical_Combining_Class=nukta}', "");
Expect(0, 71351, '\p{^Canonical_Combining_Class=nukta}', "");
Expect(0, 71351, '\P{Canonical_Combining_Class=nukta}', "");
Expect(1, 71351, '\P{^Canonical_Combining_Class=nukta}', "");
Expect(0, 71352, '\p{Canonical_Combining_Class=nukta}', "");
Expect(1, 71352, '\p{^Canonical_Combining_Class=nukta}', "");
Expect(1, 71352, '\P{Canonical_Combining_Class=nukta}', "");
Expect(0, 71352, '\P{^Canonical_Combining_Class=nukta}', "");
Expect(1, 71351, '\p{Canonical_Combining_Class=	-NUKTA}', "");
Expect(0, 71351, '\p{^Canonical_Combining_Class=	-NUKTA}', "");
Expect(0, 71351, '\P{Canonical_Combining_Class=	-NUKTA}', "");
Expect(1, 71351, '\P{^Canonical_Combining_Class=	-NUKTA}', "");
Expect(0, 71352, '\p{Canonical_Combining_Class=	-NUKTA}', "");
Expect(1, 71352, '\p{^Canonical_Combining_Class=	-NUKTA}', "");
Expect(1, 71352, '\P{Canonical_Combining_Class=	-NUKTA}', "");
Expect(0, 71352, '\P{^Canonical_Combining_Class=	-NUKTA}', "");
Error('\p{Ccc=- NK/a/}');
Error('\P{Ccc=- NK/a/}');
Expect(1, 71351, '\p{Ccc=nk}', "");
Expect(0, 71351, '\p{^Ccc=nk}', "");
Expect(0, 71351, '\P{Ccc=nk}', "");
Expect(1, 71351, '\P{^Ccc=nk}', "");
Expect(0, 71352, '\p{Ccc=nk}', "");
Expect(1, 71352, '\p{^Ccc=nk}', "");
Expect(1, 71352, '\P{Ccc=nk}', "");
Expect(0, 71352, '\P{^Ccc=nk}', "");
Expect(1, 71351, '\p{Ccc= NK}', "");
Expect(0, 71351, '\p{^Ccc= NK}', "");
Expect(0, 71351, '\P{Ccc= NK}', "");
Expect(1, 71351, '\P{^Ccc= NK}', "");
Expect(0, 71352, '\p{Ccc= NK}', "");
Expect(1, 71352, '\p{^Ccc= NK}', "");
Expect(1, 71352, '\P{Ccc= NK}', "");
Expect(0, 71352, '\P{^Ccc= NK}', "");
Error('\p{Is_Canonical_Combining_Class=_/a/0000007}');
Error('\P{Is_Canonical_Combining_Class=_/a/0000007}');
Expect(1, 71351, '\p{Is_Canonical_Combining_Class=7}', "");
Expect(0, 71351, '\p{^Is_Canonical_Combining_Class=7}', "");
Expect(0, 71351, '\P{Is_Canonical_Combining_Class=7}', "");
Expect(1, 71351, '\P{^Is_Canonical_Combining_Class=7}', "");
Expect(0, 71352, '\p{Is_Canonical_Combining_Class=7}', "");
Expect(1, 71352, '\p{^Is_Canonical_Combining_Class=7}', "");
Expect(1, 71352, '\P{Is_Canonical_Combining_Class=7}', "");
Expect(0, 71352, '\P{^Is_Canonical_Combining_Class=7}', "");
Error('\p{Is_Ccc= nukta:=}');
Error('\P{Is_Ccc= nukta:=}');
Expect(1, 71351, '\p{Is_Ccc=nukta}', "");
Expect(0, 71351, '\p{^Is_Ccc=nukta}', "");
Expect(0, 71351, '\P{Is_Ccc=nukta}', "");
Expect(1, 71351, '\P{^Is_Ccc=nukta}', "");
Expect(0, 71352, '\p{Is_Ccc=nukta}', "");
Expect(1, 71352, '\p{^Is_Ccc=nukta}', "");
Expect(1, 71352, '\P{Is_Ccc=nukta}', "");
Expect(0, 71352, '\P{^Is_Ccc=nukta}', "");
Expect(1, 71351, '\p{Is_Ccc=- nukta}', "");
Expect(0, 71351, '\p{^Is_Ccc=- nukta}', "");
Expect(0, 71351, '\P{Is_Ccc=- nukta}', "");
Expect(1, 71351, '\P{^Is_Ccc=- nukta}', "");
Expect(0, 71352, '\p{Is_Ccc=- nukta}', "");
Expect(1, 71352, '\p{^Is_Ccc=- nukta}', "");
Expect(1, 71352, '\P{Is_Ccc=- nukta}', "");
Expect(0, 71352, '\P{^Is_Ccc=- nukta}', "");
Error('\p{Canonical_Combining_Class=-:=CCC34}');
Error('\P{Canonical_Combining_Class=-:=CCC34}');
Expect(1, 1618, '\p{Canonical_Combining_Class=ccc34}', "");
Expect(0, 1618, '\p{^Canonical_Combining_Class=ccc34}', "");
Expect(0, 1618, '\P{Canonical_Combining_Class=ccc34}', "");
Expect(1, 1618, '\P{^Canonical_Combining_Class=ccc34}', "");
Expect(0, 1619, '\p{Canonical_Combining_Class=ccc34}', "");
Expect(1, 1619, '\p{^Canonical_Combining_Class=ccc34}', "");
Expect(1, 1619, '\P{Canonical_Combining_Class=ccc34}', "");
Expect(0, 1619, '\P{^Canonical_Combining_Class=ccc34}', "");
Expect(1, 1618, '\p{Canonical_Combining_Class=_CCC34}', "");
Expect(0, 1618, '\p{^Canonical_Combining_Class=_CCC34}', "");
Expect(0, 1618, '\P{Canonical_Combining_Class=_CCC34}', "");
Expect(1, 1618, '\P{^Canonical_Combining_Class=_CCC34}', "");
Expect(0, 1619, '\p{Canonical_Combining_Class=_CCC34}', "");
Expect(1, 1619, '\p{^Canonical_Combining_Class=_CCC34}', "");
Expect(1, 1619, '\P{Canonical_Combining_Class=_CCC34}', "");
Expect(0, 1619, '\P{^Canonical_Combining_Class=_CCC34}', "");
Error('\p{Ccc:	 :=0_0_0_0_0_34}');
Error('\P{Ccc:	 :=0_0_0_0_0_34}');
Expect(1, 1618, '\p{Ccc=3_4}', "");
Expect(0, 1618, '\p{^Ccc=3_4}', "");
Expect(0, 1618, '\P{Ccc=3_4}', "");
Expect(1, 1618, '\P{^Ccc=3_4}', "");
Expect(0, 1619, '\p{Ccc=3_4}', "");
Expect(1, 1619, '\p{^Ccc=3_4}', "");
Expect(1, 1619, '\P{Ccc=3_4}', "");
Expect(0, 1619, '\P{^Ccc=3_4}', "");
Error('\p{Is_Canonical_Combining_Class=/a/--ccc34}');
Error('\P{Is_Canonical_Combining_Class=/a/--ccc34}');
Expect(1, 1618, '\p{Is_Canonical_Combining_Class=ccc34}', "");
Expect(0, 1618, '\p{^Is_Canonical_Combining_Class=ccc34}', "");
Expect(0, 1618, '\P{Is_Canonical_Combining_Class=ccc34}', "");
Expect(1, 1618, '\P{^Is_Canonical_Combining_Class=ccc34}', "");
Expect(0, 1619, '\p{Is_Canonical_Combining_Class=ccc34}', "");
Expect(1, 1619, '\p{^Is_Canonical_Combining_Class=ccc34}', "");
Expect(1, 1619, '\P{Is_Canonical_Combining_Class=ccc34}', "");
Expect(0, 1619, '\P{^Is_Canonical_Combining_Class=ccc34}', "");
Expect(1, 1618, '\p{Is_Canonical_Combining_Class=_ccc34}', "");
Expect(0, 1618, '\p{^Is_Canonical_Combining_Class=_ccc34}', "");
Expect(0, 1618, '\P{Is_Canonical_Combining_Class=_ccc34}', "");
Expect(1, 1618, '\P{^Is_Canonical_Combining_Class=_ccc34}', "");
Expect(0, 1619, '\p{Is_Canonical_Combining_Class=_ccc34}', "");
Expect(1, 1619, '\p{^Is_Canonical_Combining_Class=_ccc34}', "");
Expect(1, 1619, '\P{Is_Canonical_Combining_Class=_ccc34}', "");
Expect(0, 1619, '\P{^Is_Canonical_Combining_Class=_ccc34}', "");
Error('\p{Is_Ccc=-034:=}');
Error('\P{Is_Ccc=-034:=}');
Expect(1, 1618, '\p{Is_Ccc: +00_00_00_034}', "");
Expect(0, 1618, '\p{^Is_Ccc: +00_00_00_034}', "");
Expect(0, 1618, '\P{Is_Ccc: +00_00_00_034}', "");
Expect(1, 1618, '\P{^Is_Ccc: +00_00_00_034}', "");
Expect(0, 1619, '\p{Is_Ccc: +00_00_00_034}', "");
Expect(1, 1619, '\p{^Is_Ccc: +00_00_00_034}', "");
Expect(1, 1619, '\P{Is_Ccc: +00_00_00_034}', "");
Expect(0, 1619, '\P{^Is_Ccc: +00_00_00_034}', "");
Error('\p{Canonical_Combining_Class=/a/-_CCC107}');
Error('\P{Canonical_Combining_Class=/a/-_CCC107}');
Expect(1, 3659, '\p{Canonical_Combining_Class=ccc107}', "");
Expect(0, 3659, '\p{^Canonical_Combining_Class=ccc107}', "");
Expect(0, 3659, '\P{Canonical_Combining_Class=ccc107}', "");
Expect(1, 3659, '\P{^Canonical_Combining_Class=ccc107}', "");
Expect(0, 3660, '\p{Canonical_Combining_Class=ccc107}', "");
Expect(1, 3660, '\p{^Canonical_Combining_Class=ccc107}', "");
Expect(1, 3660, '\P{Canonical_Combining_Class=ccc107}', "");
Expect(0, 3660, '\P{^Canonical_Combining_Class=ccc107}', "");
Expect(1, 3659, '\p{Canonical_Combining_Class=_-CCC107}', "");
Expect(0, 3659, '\p{^Canonical_Combining_Class=_-CCC107}', "");
Expect(0, 3659, '\P{Canonical_Combining_Class=_-CCC107}', "");
Expect(1, 3659, '\P{^Canonical_Combining_Class=_-CCC107}', "");
Expect(0, 3660, '\p{Canonical_Combining_Class=_-CCC107}', "");
Expect(1, 3660, '\p{^Canonical_Combining_Class=_-CCC107}', "");
Expect(1, 3660, '\P{Canonical_Combining_Class=_-CCC107}', "");
Expect(0, 3660, '\P{^Canonical_Combining_Class=_-CCC107}', "");
Error('\p{Ccc=_:=0000000107}');
Error('\P{Ccc=_:=0000000107}');
Expect(1, 3659, '\p{Ccc=+00_01_07}', "");
Expect(0, 3659, '\p{^Ccc=+00_01_07}', "");
Expect(0, 3659, '\P{Ccc=+00_01_07}', "");
Expect(1, 3659, '\P{^Ccc=+00_01_07}', "");
Expect(0, 3660, '\p{Ccc=+00_01_07}', "");
Expect(1, 3660, '\p{^Ccc=+00_01_07}', "");
Expect(1, 3660, '\P{Ccc=+00_01_07}', "");
Expect(0, 3660, '\P{^Ccc=+00_01_07}', "");
Error('\p{Is_Canonical_Combining_Class=	:=CCC107}');
Error('\P{Is_Canonical_Combining_Class=	:=CCC107}');
Expect(1, 3659, '\p{Is_Canonical_Combining_Class:ccc107}', "");
Expect(0, 3659, '\p{^Is_Canonical_Combining_Class:ccc107}', "");
Expect(0, 3659, '\P{Is_Canonical_Combining_Class:ccc107}', "");
Expect(1, 3659, '\P{^Is_Canonical_Combining_Class:ccc107}', "");
Expect(0, 3660, '\p{Is_Canonical_Combining_Class:ccc107}', "");
Expect(1, 3660, '\p{^Is_Canonical_Combining_Class:ccc107}', "");
Expect(1, 3660, '\P{Is_Canonical_Combining_Class:ccc107}', "");
Expect(0, 3660, '\P{^Is_Canonical_Combining_Class:ccc107}', "");
Expect(1, 3659, '\p{Is_Canonical_Combining_Class:	 	CCC107}', "");
Expect(0, 3659, '\p{^Is_Canonical_Combining_Class:	 	CCC107}', "");
Expect(0, 3659, '\P{Is_Canonical_Combining_Class:	 	CCC107}', "");
Expect(1, 3659, '\P{^Is_Canonical_Combining_Class:	 	CCC107}', "");
Expect(0, 3660, '\p{Is_Canonical_Combining_Class:	 	CCC107}', "");
Expect(1, 3660, '\p{^Is_Canonical_Combining_Class:	 	CCC107}', "");
Expect(1, 3660, '\P{Is_Canonical_Combining_Class:	 	CCC107}', "");
Expect(0, 3660, '\P{^Is_Canonical_Combining_Class:	 	CCC107}', "");
Error('\p{Is_Ccc=-0_0_0_0_0_1_07/a/}');
Error('\P{Is_Ccc=-0_0_0_0_0_1_07/a/}');
Expect(1, 3659, '\p{Is_Ccc=0000107}', "");
Expect(0, 3659, '\p{^Is_Ccc=0000107}', "");
Expect(0, 3659, '\P{Is_Ccc=0000107}', "");
Expect(1, 3659, '\P{^Is_Ccc=0000107}', "");
Expect(0, 3660, '\p{Is_Ccc=0000107}', "");
Expect(1, 3660, '\p{^Is_Ccc=0000107}', "");
Expect(1, 3660, '\P{Is_Ccc=0000107}', "");
Expect(0, 3660, '\P{^Is_Ccc=0000107}', "");
Error('\p{Canonical_Combining_Class=:=	_CCC122}');
Error('\P{Canonical_Combining_Class=:=	_CCC122}');
Expect(1, 3787, '\p{Canonical_Combining_Class=ccc122}', "");
Expect(0, 3787, '\p{^Canonical_Combining_Class=ccc122}', "");
Expect(0, 3787, '\P{Canonical_Combining_Class=ccc122}', "");
Expect(1, 3787, '\P{^Canonical_Combining_Class=ccc122}', "");
Expect(0, 3788, '\p{Canonical_Combining_Class=ccc122}', "");
Expect(1, 3788, '\p{^Canonical_Combining_Class=ccc122}', "");
Expect(1, 3788, '\P{Canonical_Combining_Class=ccc122}', "");
Expect(0, 3788, '\P{^Canonical_Combining_Class=ccc122}', "");
Expect(1, 3787, '\p{Canonical_Combining_Class:	_CCC122}', "");
Expect(0, 3787, '\p{^Canonical_Combining_Class:	_CCC122}', "");
Expect(0, 3787, '\P{Canonical_Combining_Class:	_CCC122}', "");
Expect(1, 3787, '\P{^Canonical_Combining_Class:	_CCC122}', "");
Expect(0, 3788, '\p{Canonical_Combining_Class:	_CCC122}', "");
Expect(1, 3788, '\p{^Canonical_Combining_Class:	_CCC122}', "");
Expect(1, 3788, '\P{Canonical_Combining_Class:	_CCC122}', "");
Expect(0, 3788, '\P{^Canonical_Combining_Class:	_CCC122}', "");
Error('\p{Ccc= 00_00_12_2:=}');
Error('\P{Ccc= 00_00_12_2:=}');
Expect(1, 3787, '\p{Ccc=0_0_0_0_0_0_00122}', "");
Expect(0, 3787, '\p{^Ccc=0_0_0_0_0_0_00122}', "");
Expect(0, 3787, '\P{Ccc=0_0_0_0_0_0_00122}', "");
Expect(1, 3787, '\P{^Ccc=0_0_0_0_0_0_00122}', "");
Expect(0, 3788, '\p{Ccc=0_0_0_0_0_0_00122}', "");
Expect(1, 3788, '\p{^Ccc=0_0_0_0_0_0_00122}', "");
Expect(1, 3788, '\P{Ccc=0_0_0_0_0_0_00122}', "");
Expect(0, 3788, '\P{^Ccc=0_0_0_0_0_0_00122}', "");
Error('\p{Is_Canonical_Combining_Class: := _CCC122}');
Error('\P{Is_Canonical_Combining_Class: := _CCC122}');
Expect(1, 3787, '\p{Is_Canonical_Combining_Class=ccc122}', "");
Expect(0, 3787, '\p{^Is_Canonical_Combining_Class=ccc122}', "");
Expect(0, 3787, '\P{Is_Canonical_Combining_Class=ccc122}', "");
Expect(1, 3787, '\P{^Is_Canonical_Combining_Class=ccc122}', "");
Expect(0, 3788, '\p{Is_Canonical_Combining_Class=ccc122}', "");
Expect(1, 3788, '\p{^Is_Canonical_Combining_Class=ccc122}', "");
Expect(1, 3788, '\P{Is_Canonical_Combining_Class=ccc122}', "");
Expect(0, 3788, '\P{^Is_Canonical_Combining_Class=ccc122}', "");
Expect(1, 3787, '\p{Is_Canonical_Combining_Class= -CCC122}', "");
Expect(0, 3787, '\p{^Is_Canonical_Combining_Class= -CCC122}', "");
Expect(0, 3787, '\P{Is_Canonical_Combining_Class= -CCC122}', "");
Expect(1, 3787, '\P{^Is_Canonical_Combining_Class= -CCC122}', "");
Expect(0, 3788, '\p{Is_Canonical_Combining_Class= -CCC122}', "");
Expect(1, 3788, '\p{^Is_Canonical_Combining_Class= -CCC122}', "");
Expect(1, 3788, '\P{Is_Canonical_Combining_Class= -CCC122}', "");
Expect(0, 3788, '\P{^Is_Canonical_Combining_Class= -CCC122}', "");
Error('\p{Is_Ccc=/a/	-000000000122}');
Error('\P{Is_Ccc=/a/	-000000000122}');
Expect(1, 3787, '\p{Is_Ccc=00122}', "");
Expect(0, 3787, '\p{^Is_Ccc=00122}', "");
Expect(0, 3787, '\P{Is_Ccc=00122}', "");
Expect(1, 3787, '\P{^Is_Ccc=00122}', "");
Expect(0, 3788, '\p{Is_Ccc=00122}', "");
Expect(1, 3788, '\p{^Is_Ccc=00122}', "");
Expect(1, 3788, '\P{Is_Ccc=00122}', "");
Expect(0, 3788, '\P{^Is_Ccc=00122}', "");
Error('\p{Canonical_Combining_Class=/a/ CCC130}');
Error('\P{Canonical_Combining_Class=/a/ CCC130}');
Expect(1, 3968, '\p{Canonical_Combining_Class=ccc130}', "");
Expect(0, 3968, '\p{^Canonical_Combining_Class=ccc130}', "");
Expect(0, 3968, '\P{Canonical_Combining_Class=ccc130}', "");
Expect(1, 3968, '\P{^Canonical_Combining_Class=ccc130}', "");
Expect(0, 3969, '\p{Canonical_Combining_Class=ccc130}', "");
Expect(1, 3969, '\p{^Canonical_Combining_Class=ccc130}', "");
Expect(1, 3969, '\P{Canonical_Combining_Class=ccc130}', "");
Expect(0, 3969, '\P{^Canonical_Combining_Class=ccc130}', "");
Expect(1, 3968, '\p{Canonical_Combining_Class: -	CCC130}', "");
Expect(0, 3968, '\p{^Canonical_Combining_Class: -	CCC130}', "");
Expect(0, 3968, '\P{Canonical_Combining_Class: -	CCC130}', "");
Expect(1, 3968, '\P{^Canonical_Combining_Class: -	CCC130}', "");
Expect(0, 3969, '\p{Canonical_Combining_Class: -	CCC130}', "");
Expect(1, 3969, '\p{^Canonical_Combining_Class: -	CCC130}', "");
Expect(1, 3969, '\P{Canonical_Combining_Class: -	CCC130}', "");
Expect(0, 3969, '\P{^Canonical_Combining_Class: -	CCC130}', "");
Error('\p{Ccc=-	+0_0_0_0_0_0_0_0_1_30/a/}');
Error('\P{Ccc=-	+0_0_0_0_0_0_0_0_1_30/a/}');
Expect(1, 3968, '\p{Ccc=000000000130}', "");
Expect(0, 3968, '\p{^Ccc=000000000130}', "");
Expect(0, 3968, '\P{Ccc=000000000130}', "");
Expect(1, 3968, '\P{^Ccc=000000000130}', "");
Expect(0, 3969, '\p{Ccc=000000000130}', "");
Expect(1, 3969, '\p{^Ccc=000000000130}', "");
Expect(1, 3969, '\P{Ccc=000000000130}', "");
Expect(0, 3969, '\P{^Ccc=000000000130}', "");
Error('\p{Is_Canonical_Combining_Class=_ CCC130/a/}');
Error('\P{Is_Canonical_Combining_Class=_ CCC130/a/}');
Expect(1, 3968, '\p{Is_Canonical_Combining_Class=ccc130}', "");
Expect(0, 3968, '\p{^Is_Canonical_Combining_Class=ccc130}', "");
Expect(0, 3968, '\P{Is_Canonical_Combining_Class=ccc130}', "");
Expect(1, 3968, '\P{^Is_Canonical_Combining_Class=ccc130}', "");
Expect(0, 3969, '\p{Is_Canonical_Combining_Class=ccc130}', "");
Expect(1, 3969, '\p{^Is_Canonical_Combining_Class=ccc130}', "");
Expect(1, 3969, '\P{Is_Canonical_Combining_Class=ccc130}', "");
Expect(0, 3969, '\P{^Is_Canonical_Combining_Class=ccc130}', "");
Expect(1, 3968, '\p{Is_Canonical_Combining_Class=		CCC130}', "");
Expect(0, 3968, '\p{^Is_Canonical_Combining_Class=		CCC130}', "");
Expect(0, 3968, '\P{Is_Canonical_Combining_Class=		CCC130}', "");
Expect(1, 3968, '\P{^Is_Canonical_Combining_Class=		CCC130}', "");
Expect(0, 3969, '\p{Is_Canonical_Combining_Class=		CCC130}', "");
Expect(1, 3969, '\p{^Is_Canonical_Combining_Class=		CCC130}', "");
Expect(1, 3969, '\P{Is_Canonical_Combining_Class=		CCC130}', "");
Expect(0, 3969, '\P{^Is_Canonical_Combining_Class=		CCC130}', "");
Error('\p{Is_Ccc=/a/-	+000000000130}');
Error('\P{Is_Ccc=/a/-	+000000000130}');
Expect(1, 3968, '\p{Is_Ccc=+0000000130}', "");
Expect(0, 3968, '\p{^Is_Ccc=+0000000130}', "");
Expect(0, 3968, '\P{Is_Ccc=+0000000130}', "");
Expect(1, 3968, '\P{^Is_Ccc=+0000000130}', "");
Expect(0, 3969, '\p{Is_Ccc=+0000000130}', "");
Expect(1, 3969, '\p{^Is_Ccc=+0000000130}', "");
Expect(1, 3969, '\P{Is_Ccc=+0000000130}', "");
Expect(0, 3969, '\P{^Is_Ccc=+0000000130}', "");
Error('\p{Canonical_Combining_Class=	 CCC28:=}');
Error('\P{Canonical_Combining_Class=	 CCC28:=}');
Expect(1, 2289, '\p{Canonical_Combining_Class=ccc28}', "");
Expect(0, 2289, '\p{^Canonical_Combining_Class=ccc28}', "");
Expect(0, 2289, '\P{Canonical_Combining_Class=ccc28}', "");
Expect(1, 2289, '\P{^Canonical_Combining_Class=ccc28}', "");
Expect(0, 2290, '\p{Canonical_Combining_Class=ccc28}', "");
Expect(1, 2290, '\p{^Canonical_Combining_Class=ccc28}', "");
Expect(1, 2290, '\P{Canonical_Combining_Class=ccc28}', "");
Expect(0, 2290, '\P{^Canonical_Combining_Class=ccc28}', "");
Expect(1, 2289, '\p{Canonical_Combining_Class=__CCC28}', "");
Expect(0, 2289, '\p{^Canonical_Combining_Class=__CCC28}', "");
Expect(0, 2289, '\P{Canonical_Combining_Class=__CCC28}', "");
Expect(1, 2289, '\P{^Canonical_Combining_Class=__CCC28}', "");
Expect(0, 2290, '\p{Canonical_Combining_Class=__CCC28}', "");
Expect(1, 2290, '\p{^Canonical_Combining_Class=__CCC28}', "");
Expect(1, 2290, '\P{Canonical_Combining_Class=__CCC28}', "");
Expect(0, 2290, '\P{^Canonical_Combining_Class=__CCC28}', "");
Error('\p{Ccc=:=	00028}');
Error('\P{Ccc=:=	00028}');
Expect(1, 2289, '\p{Ccc=00_00_28}', "");
Expect(0, 2289, '\p{^Ccc=00_00_28}', "");
Expect(0, 2289, '\P{Ccc=00_00_28}', "");
Expect(1, 2289, '\P{^Ccc=00_00_28}', "");
Expect(0, 2290, '\p{Ccc=00_00_28}', "");
Expect(1, 2290, '\p{^Ccc=00_00_28}', "");
Expect(1, 2290, '\P{Ccc=00_00_28}', "");
Expect(0, 2290, '\P{^Ccc=00_00_28}', "");
Error('\p{Is_Canonical_Combining_Class=ccc28/a/}');
Error('\P{Is_Canonical_Combining_Class=ccc28/a/}');
Expect(1, 2289, '\p{Is_Canonical_Combining_Class=ccc28}', "");
Expect(0, 2289, '\p{^Is_Canonical_Combining_Class=ccc28}', "");
Expect(0, 2289, '\P{Is_Canonical_Combining_Class=ccc28}', "");
Expect(1, 2289, '\P{^Is_Canonical_Combining_Class=ccc28}', "");
Expect(0, 2290, '\p{Is_Canonical_Combining_Class=ccc28}', "");
Expect(1, 2290, '\p{^Is_Canonical_Combining_Class=ccc28}', "");
Expect(1, 2290, '\P{Is_Canonical_Combining_Class=ccc28}', "");
Expect(0, 2290, '\P{^Is_Canonical_Combining_Class=ccc28}', "");
Expect(1, 2289, '\p{Is_Canonical_Combining_Class=__CCC28}', "");
Expect(0, 2289, '\p{^Is_Canonical_Combining_Class=__CCC28}', "");
Expect(0, 2289, '\P{Is_Canonical_Combining_Class=__CCC28}', "");
Expect(1, 2289, '\P{^Is_Canonical_Combining_Class=__CCC28}', "");
Expect(0, 2290, '\p{Is_Canonical_Combining_Class=__CCC28}', "");
Expect(1, 2290, '\p{^Is_Canonical_Combining_Class=__CCC28}', "");
Expect(1, 2290, '\P{Is_Canonical_Combining_Class=__CCC28}', "");
Expect(0, 2290, '\P{^Is_Canonical_Combining_Class=__CCC28}', "");
Error('\p{Is_Ccc= :=+000028}');
Error('\P{Is_Ccc= :=+000028}');
Expect(1, 2289, '\p{Is_Ccc=+00_02_8}', "");
Expect(0, 2289, '\p{^Is_Ccc=+00_02_8}', "");
Expect(0, 2289, '\P{Is_Ccc=+00_02_8}', "");
Expect(1, 2289, '\P{^Is_Ccc=+00_02_8}', "");
Expect(0, 2290, '\p{Is_Ccc=+00_02_8}', "");
Expect(1, 2290, '\p{^Is_Ccc=+00_02_8}', "");
Expect(1, 2290, '\P{Is_Ccc=+00_02_8}', "");
Expect(0, 2290, '\P{^Is_Ccc=+00_02_8}', "");
Error('\p{Canonical_Combining_Class=_ CCC12:=}');
Error('\P{Canonical_Combining_Class=_ CCC12:=}');
Expect(1, 1458, '\p{Canonical_Combining_Class:	ccc12}', "");
Expect(0, 1458, '\p{^Canonical_Combining_Class:	ccc12}', "");
Expect(0, 1458, '\P{Canonical_Combining_Class:	ccc12}', "");
Expect(1, 1458, '\P{^Canonical_Combining_Class:	ccc12}', "");
Expect(0, 1459, '\p{Canonical_Combining_Class:	ccc12}', "");
Expect(1, 1459, '\p{^Canonical_Combining_Class:	ccc12}', "");
Expect(1, 1459, '\P{Canonical_Combining_Class:	ccc12}', "");
Expect(0, 1459, '\P{^Canonical_Combining_Class:	ccc12}', "");
Expect(1, 1458, '\p{Canonical_Combining_Class=  CCC12}', "");
Expect(0, 1458, '\p{^Canonical_Combining_Class=  CCC12}', "");
Expect(0, 1458, '\P{Canonical_Combining_Class=  CCC12}', "");
Expect(1, 1458, '\P{^Canonical_Combining_Class=  CCC12}', "");
Expect(0, 1459, '\p{Canonical_Combining_Class=  CCC12}', "");
Expect(1, 1459, '\p{^Canonical_Combining_Class=  CCC12}', "");
Expect(1, 1459, '\P{Canonical_Combining_Class=  CCC12}', "");
Expect(0, 1459, '\P{^Canonical_Combining_Class=  CCC12}', "");
Error('\p{Ccc:  /a/0_0_0_0_0_0_0_0_012}');
Error('\P{Ccc:  /a/0_0_0_0_0_0_0_0_012}');
Expect(1, 1458, '\p{Ccc=+1_2}', "");
Expect(0, 1458, '\p{^Ccc=+1_2}', "");
Expect(0, 1458, '\P{Ccc=+1_2}', "");
Expect(1, 1458, '\P{^Ccc=+1_2}', "");
Expect(0, 1459, '\p{Ccc=+1_2}', "");
Expect(1, 1459, '\p{^Ccc=+1_2}', "");
Expect(1, 1459, '\P{Ccc=+1_2}', "");
Expect(0, 1459, '\P{^Ccc=+1_2}', "");
Error('\p{Is_Canonical_Combining_Class=:=_CCC12}');
Error('\P{Is_Canonical_Combining_Class=:=_CCC12}');
Expect(1, 1458, '\p{Is_Canonical_Combining_Class=ccc12}', "");
Expect(0, 1458, '\p{^Is_Canonical_Combining_Class=ccc12}', "");
Expect(0, 1458, '\P{Is_Canonical_Combining_Class=ccc12}', "");
Expect(1, 1458, '\P{^Is_Canonical_Combining_Class=ccc12}', "");
Expect(0, 1459, '\p{Is_Canonical_Combining_Class=ccc12}', "");
Expect(1, 1459, '\p{^Is_Canonical_Combining_Class=ccc12}', "");
Expect(1, 1459, '\P{Is_Canonical_Combining_Class=ccc12}', "");
Expect(0, 1459, '\P{^Is_Canonical_Combining_Class=ccc12}', "");
Expect(1, 1458, '\p{Is_Canonical_Combining_Class=	_CCC12}', "");
Expect(0, 1458, '\p{^Is_Canonical_Combining_Class=	_CCC12}', "");
Expect(0, 1458, '\P{Is_Canonical_Combining_Class=	_CCC12}', "");
Expect(1, 1458, '\P{^Is_Canonical_Combining_Class=	_CCC12}', "");
Expect(0, 1459, '\p{Is_Canonical_Combining_Class=	_CCC12}', "");
Expect(1, 1459, '\p{^Is_Canonical_Combining_Class=	_CCC12}', "");
Expect(1, 1459, '\P{Is_Canonical_Combining_Class=	_CCC12}', "");
Expect(0, 1459, '\P{^Is_Canonical_Combining_Class=	_CCC12}', "");
Error('\p{Is_Ccc=	+0000000012:=}');
Error('\P{Is_Ccc=	+0000000012:=}');
Expect(1, 1458, '\p{Is_Ccc=+00_01_2}', "");
Expect(0, 1458, '\p{^Is_Ccc=+00_01_2}', "");
Expect(0, 1458, '\P{Is_Ccc=+00_01_2}', "");
Expect(1, 1458, '\P{^Is_Ccc=+00_01_2}', "");
Expect(0, 1459, '\p{Is_Ccc=+00_01_2}', "");
Expect(1, 1459, '\p{^Is_Ccc=+00_01_2}', "");
Expect(1, 1459, '\P{Is_Ccc=+00_01_2}', "");
Expect(0, 1459, '\P{^Is_Ccc=+00_01_2}', "");
Error('\p{Canonical_Combining_Class: /a/-Attached_Above}');
Error('\P{Canonical_Combining_Class: /a/-Attached_Above}');
Expect(1, 7630, '\p{Canonical_Combining_Class=attachedabove}', "");
Expect(0, 7630, '\p{^Canonical_Combining_Class=attachedabove}', "");
Expect(0, 7630, '\P{Canonical_Combining_Class=attachedabove}', "");
Expect(1, 7630, '\P{^Canonical_Combining_Class=attachedabove}', "");
Expect(0, 7631, '\p{Canonical_Combining_Class=attachedabove}', "");
Expect(1, 7631, '\p{^Canonical_Combining_Class=attachedabove}', "");
Expect(1, 7631, '\P{Canonical_Combining_Class=attachedabove}', "");
Expect(0, 7631, '\P{^Canonical_Combining_Class=attachedabove}', "");
Expect(1, 7630, '\p{Canonical_Combining_Class=__Attached_above}', "");
Expect(0, 7630, '\p{^Canonical_Combining_Class=__Attached_above}', "");
Expect(0, 7630, '\P{Canonical_Combining_Class=__Attached_above}', "");
Expect(1, 7630, '\P{^Canonical_Combining_Class=__Attached_above}', "");
Expect(0, 7631, '\p{Canonical_Combining_Class=__Attached_above}', "");
Expect(1, 7631, '\p{^Canonical_Combining_Class=__Attached_above}', "");
Expect(1, 7631, '\P{Canonical_Combining_Class=__Attached_above}', "");
Expect(0, 7631, '\P{^Canonical_Combining_Class=__Attached_above}', "");
Error('\p{Ccc=_ATA:=}');
Error('\P{Ccc=_ATA:=}');
Expect(1, 7630, '\p{Ccc:ata}', "");
Expect(0, 7630, '\p{^Ccc:ata}', "");
Expect(0, 7630, '\P{Ccc:ata}', "");
Expect(1, 7630, '\P{^Ccc:ata}', "");
Expect(0, 7631, '\p{Ccc:ata}', "");
Expect(1, 7631, '\p{^Ccc:ata}', "");
Expect(1, 7631, '\P{Ccc:ata}', "");
Expect(0, 7631, '\P{^Ccc:ata}', "");
Expect(1, 7630, '\p{Ccc= 	ATA}', "");
Expect(0, 7630, '\p{^Ccc= 	ATA}', "");
Expect(0, 7630, '\P{Ccc= 	ATA}', "");
Expect(1, 7630, '\P{^Ccc= 	ATA}', "");
Expect(0, 7631, '\p{Ccc= 	ATA}', "");
Expect(1, 7631, '\p{^Ccc= 	ATA}', "");
Expect(1, 7631, '\P{Ccc= 	ATA}', "");
Expect(0, 7631, '\P{^Ccc= 	ATA}', "");
Error('\p{Is_Canonical_Combining_Class=_:=0_0_0_0_0_0_0214}');
Error('\P{Is_Canonical_Combining_Class=_:=0_0_0_0_0_0_0214}');
Expect(1, 7630, '\p{Is_Canonical_Combining_Class=+0_0_0_0_214}', "");
Expect(0, 7630, '\p{^Is_Canonical_Combining_Class=+0_0_0_0_214}', "");
Expect(0, 7630, '\P{Is_Canonical_Combining_Class=+0_0_0_0_214}', "");
Expect(1, 7630, '\P{^Is_Canonical_Combining_Class=+0_0_0_0_214}', "");
Expect(0, 7631, '\p{Is_Canonical_Combining_Class=+0_0_0_0_214}', "");
Expect(1, 7631, '\p{^Is_Canonical_Combining_Class=+0_0_0_0_214}', "");
Expect(1, 7631, '\P{Is_Canonical_Combining_Class=+0_0_0_0_214}', "");
Expect(0, 7631, '\P{^Is_Canonical_Combining_Class=+0_0_0_0_214}', "");
Error('\p{Is_Ccc=	-attached_Above:=}');
Error('\P{Is_Ccc=	-attached_Above:=}');
Expect(1, 7630, '\p{Is_Ccc=attachedabove}', "");
Expect(0, 7630, '\p{^Is_Ccc=attachedabove}', "");
Expect(0, 7630, '\P{Is_Ccc=attachedabove}', "");
Expect(1, 7630, '\P{^Is_Ccc=attachedabove}', "");
Expect(0, 7631, '\p{Is_Ccc=attachedabove}', "");
Expect(1, 7631, '\p{^Is_Ccc=attachedabove}', "");
Expect(1, 7631, '\P{Is_Ccc=attachedabove}', "");
Expect(0, 7631, '\P{^Is_Ccc=attachedabove}', "");
Expect(1, 7630, '\p{Is_Ccc:   --Attached_Above}', "");
Expect(0, 7630, '\p{^Is_Ccc:   --Attached_Above}', "");
Expect(0, 7630, '\P{Is_Ccc:   --Attached_Above}', "");
Expect(1, 7630, '\P{^Is_Ccc:   --Attached_Above}', "");
Expect(0, 7631, '\p{Is_Ccc:   --Attached_Above}', "");
Expect(1, 7631, '\p{^Is_Ccc:   --Attached_Above}', "");
Expect(1, 7631, '\P{Is_Ccc:   --Attached_Above}', "");
Expect(0, 7631, '\P{^Is_Ccc:   --Attached_Above}', "");
Error('\p{Canonical_Combining_Class=	-CCC30:=}');
Error('\P{Canonical_Combining_Class=	-CCC30:=}');
Expect(1, 1614, '\p{Canonical_Combining_Class=ccc30}', "");
Expect(0, 1614, '\p{^Canonical_Combining_Class=ccc30}', "");
Expect(0, 1614, '\P{Canonical_Combining_Class=ccc30}', "");
Expect(1, 1614, '\P{^Canonical_Combining_Class=ccc30}', "");
Expect(0, 1615, '\p{Canonical_Combining_Class=ccc30}', "");
Expect(1, 1615, '\p{^Canonical_Combining_Class=ccc30}', "");
Expect(1, 1615, '\P{Canonical_Combining_Class=ccc30}', "");
Expect(0, 1615, '\P{^Canonical_Combining_Class=ccc30}', "");
Expect(1, 1614, '\p{Canonical_Combining_Class=	 ccc30}', "");
Expect(0, 1614, '\p{^Canonical_Combining_Class=	 ccc30}', "");
Expect(0, 1614, '\P{Canonical_Combining_Class=	 ccc30}', "");
Expect(1, 1614, '\P{^Canonical_Combining_Class=	 ccc30}', "");
Expect(0, 1615, '\p{Canonical_Combining_Class=	 ccc30}', "");
Expect(1, 1615, '\p{^Canonical_Combining_Class=	 ccc30}', "");
Expect(1, 1615, '\P{Canonical_Combining_Class=	 ccc30}', "");
Expect(0, 1615, '\P{^Canonical_Combining_Class=	 ccc30}', "");
Error('\p{Ccc=/a/0_0_0_0_0_0_0_0_0_30}');
Error('\P{Ccc=/a/0_0_0_0_0_0_0_0_0_30}');
Expect(1, 1614, '\p{Ccc=0003_0}', "");
Expect(0, 1614, '\p{^Ccc=0003_0}', "");
Expect(0, 1614, '\P{Ccc=0003_0}', "");
Expect(1, 1614, '\P{^Ccc=0003_0}', "");
Expect(0, 1615, '\p{Ccc=0003_0}', "");
Expect(1, 1615, '\p{^Ccc=0003_0}', "");
Expect(1, 1615, '\P{Ccc=0003_0}', "");
Expect(0, 1615, '\P{^Ccc=0003_0}', "");
Error('\p{Is_Canonical_Combining_Class=	:=CCC30}');
Error('\P{Is_Canonical_Combining_Class=	:=CCC30}');
Expect(1, 1614, '\p{Is_Canonical_Combining_Class=ccc30}', "");
Expect(0, 1614, '\p{^Is_Canonical_Combining_Class=ccc30}', "");
Expect(0, 1614, '\P{Is_Canonical_Combining_Class=ccc30}', "");
Expect(1, 1614, '\P{^Is_Canonical_Combining_Class=ccc30}', "");
Expect(0, 1615, '\p{Is_Canonical_Combining_Class=ccc30}', "");
Expect(1, 1615, '\p{^Is_Canonical_Combining_Class=ccc30}', "");
Expect(1, 1615, '\P{Is_Canonical_Combining_Class=ccc30}', "");
Expect(0, 1615, '\P{^Is_Canonical_Combining_Class=ccc30}', "");
Expect(1, 1614, '\p{Is_Canonical_Combining_Class=_ccc30}', "");
Expect(0, 1614, '\p{^Is_Canonical_Combining_Class=_ccc30}', "");
Expect(0, 1614, '\P{Is_Canonical_Combining_Class=_ccc30}', "");
Expect(1, 1614, '\P{^Is_Canonical_Combining_Class=_ccc30}', "");
Expect(0, 1615, '\p{Is_Canonical_Combining_Class=_ccc30}', "");
Expect(1, 1615, '\p{^Is_Canonical_Combining_Class=_ccc30}', "");
Expect(1, 1615, '\P{Is_Canonical_Combining_Class=_ccc30}', "");
Expect(0, 1615, '\P{^Is_Canonical_Combining_Class=_ccc30}', "");
Error('\p{Is_Ccc=:=+0_0_0_0_0_0_0_0_0_30}');
Error('\P{Is_Ccc=:=+0_0_0_0_0_0_0_0_0_30}');
Expect(1, 1614, '\p{Is_Ccc=0_0_0_0_0_030}', "");
Expect(0, 1614, '\p{^Is_Ccc=0_0_0_0_0_030}', "");
Expect(0, 1614, '\P{Is_Ccc=0_0_0_0_0_030}', "");
Expect(1, 1614, '\P{^Is_Ccc=0_0_0_0_0_030}', "");
Expect(0, 1615, '\p{Is_Ccc=0_0_0_0_0_030}', "");
Expect(1, 1615, '\p{^Is_Ccc=0_0_0_0_0_030}', "");
Expect(1, 1615, '\P{Is_Ccc=0_0_0_0_0_030}', "");
Expect(0, 1615, '\P{^Is_Ccc=0_0_0_0_0_030}', "");
Error('\p{Logical_Order_Exception=-:=yes}');
Error('\P{Logical_Order_Exception=-:=yes}');
Expect(1, 43708, '\p{Logical_Order_Exception: yes}', "");
Expect(0, 43708, '\p{^Logical_Order_Exception: yes}', "");
Expect(0, 43708, '\P{Logical_Order_Exception: yes}', "");
Expect(1, 43708, '\P{^Logical_Order_Exception: yes}', "");
Expect(0, 43709, '\p{Logical_Order_Exception: yes}', "");
Expect(1, 43709, '\p{^Logical_Order_Exception: yes}', "");
Expect(1, 43709, '\P{Logical_Order_Exception: yes}', "");
Expect(0, 43709, '\P{^Logical_Order_Exception: yes}', "");
Expect(1, 43708, '\p{Logical_Order_Exception=Yes}', "");
Expect(0, 43708, '\p{^Logical_Order_Exception=Yes}', "");
Expect(0, 43708, '\P{Logical_Order_Exception=Yes}', "");
Expect(1, 43708, '\P{^Logical_Order_Exception=Yes}', "");
Expect(0, 43709, '\p{Logical_Order_Exception=Yes}', "");
Expect(1, 43709, '\p{^Logical_Order_Exception=Yes}', "");
Expect(1, 43709, '\P{Logical_Order_Exception=Yes}', "");
Expect(0, 43709, '\P{^Logical_Order_Exception=Yes}', "");
Error('\p{LOE=_:=y}');
Error('\P{LOE=_:=y}');
Expect(1, 43708, '\p{LOE: y}', "");
Expect(0, 43708, '\p{^LOE: y}', "");
Expect(0, 43708, '\P{LOE: y}', "");
Expect(1, 43708, '\P{^LOE: y}', "");
Expect(0, 43709, '\p{LOE: y}', "");
Expect(1, 43709, '\p{^LOE: y}', "");
Expect(1, 43709, '\P{LOE: y}', "");
Expect(0, 43709, '\P{^LOE: y}', "");
Expect(1, 43708, '\p{LOE: 	-Y}', "");
Expect(0, 43708, '\p{^LOE: 	-Y}', "");
Expect(0, 43708, '\P{LOE: 	-Y}', "");
Expect(1, 43708, '\P{^LOE: 	-Y}', "");
Expect(0, 43709, '\p{LOE: 	-Y}', "");
Expect(1, 43709, '\p{^LOE: 	-Y}', "");
Expect(1, 43709, '\P{LOE: 	-Y}', "");
Expect(0, 43709, '\P{^LOE: 	-Y}', "");
Error('\p{Is_Logical_Order_Exception=	/a/T}');
Error('\P{Is_Logical_Order_Exception=	/a/T}');
Expect(1, 43708, '\p{Is_Logical_Order_Exception=t}', "");
Expect(0, 43708, '\p{^Is_Logical_Order_Exception=t}', "");
Expect(0, 43708, '\P{Is_Logical_Order_Exception=t}', "");
Expect(1, 43708, '\P{^Is_Logical_Order_Exception=t}', "");
Expect(0, 43709, '\p{Is_Logical_Order_Exception=t}', "");
Expect(1, 43709, '\p{^Is_Logical_Order_Exception=t}', "");
Expect(1, 43709, '\P{Is_Logical_Order_Exception=t}', "");
Expect(0, 43709, '\P{^Is_Logical_Order_Exception=t}', "");
Expect(1, 43708, '\p{Is_Logical_Order_Exception=		T}', "");
Expect(0, 43708, '\p{^Is_Logical_Order_Exception=		T}', "");
Expect(0, 43708, '\P{Is_Logical_Order_Exception=		T}', "");
Expect(1, 43708, '\P{^Is_Logical_Order_Exception=		T}', "");
Expect(0, 43709, '\p{Is_Logical_Order_Exception=		T}', "");
Expect(1, 43709, '\p{^Is_Logical_Order_Exception=		T}', "");
Expect(1, 43709, '\P{Is_Logical_Order_Exception=		T}', "");
Expect(0, 43709, '\P{^Is_Logical_Order_Exception=		T}', "");
Error('\p{Is_LOE=/a/_true}');
Error('\P{Is_LOE=/a/_true}');
Expect(1, 43708, '\p{Is_LOE:true}', "");
Expect(0, 43708, '\p{^Is_LOE:true}', "");
Expect(0, 43708, '\P{Is_LOE:true}', "");
Expect(1, 43708, '\P{^Is_LOE:true}', "");
Expect(0, 43709, '\p{Is_LOE:true}', "");
Expect(1, 43709, '\p{^Is_LOE:true}', "");
Expect(1, 43709, '\P{Is_LOE:true}', "");
Expect(0, 43709, '\P{^Is_LOE:true}', "");
Expect(1, 43708, '\p{Is_LOE=_True}', "");
Expect(0, 43708, '\p{^Is_LOE=_True}', "");
Expect(0, 43708, '\P{Is_LOE=_True}', "");
Expect(1, 43708, '\P{^Is_LOE=_True}', "");
Expect(0, 43709, '\p{Is_LOE=_True}', "");
Expect(1, 43709, '\p{^Is_LOE=_True}', "");
Expect(1, 43709, '\P{Is_LOE=_True}', "");
Expect(0, 43709, '\P{^Is_LOE=_True}', "");
Error('\p{Logical_Order_Exception=/a/No}');
Error('\P{Logical_Order_Exception=/a/No}');
Expect(1, 43709, '\p{Logical_Order_Exception=no}', "");
Expect(0, 43709, '\p{^Logical_Order_Exception=no}', "");
Expect(0, 43709, '\P{Logical_Order_Exception=no}', "");
Expect(1, 43709, '\P{^Logical_Order_Exception=no}', "");
Expect(0, 43708, '\p{Logical_Order_Exception=no}', "");
Expect(1, 43708, '\p{^Logical_Order_Exception=no}', "");
Expect(1, 43708, '\P{Logical_Order_Exception=no}', "");
Expect(0, 43708, '\P{^Logical_Order_Exception=no}', "");
Expect(1, 43709, '\p{Logical_Order_Exception:		no}', "");
Expect(0, 43709, '\p{^Logical_Order_Exception:		no}', "");
Expect(0, 43709, '\P{Logical_Order_Exception:		no}', "");
Expect(1, 43709, '\P{^Logical_Order_Exception:		no}', "");
Expect(0, 43708, '\p{Logical_Order_Exception:		no}', "");
Expect(1, 43708, '\p{^Logical_Order_Exception:		no}', "");
Expect(1, 43708, '\P{Logical_Order_Exception:		no}', "");
Expect(0, 43708, '\P{^Logical_Order_Exception:		no}', "");
Error('\p{LOE=__N/a/}');
Error('\P{LOE=__N/a/}');
Expect(1, 43709, '\p{LOE: n}', "");
Expect(0, 43709, '\p{^LOE: n}', "");
Expect(0, 43709, '\P{LOE: n}', "");
Expect(1, 43709, '\P{^LOE: n}', "");
Expect(0, 43708, '\p{LOE: n}', "");
Expect(1, 43708, '\p{^LOE: n}', "");
Expect(1, 43708, '\P{LOE: n}', "");
Expect(0, 43708, '\P{^LOE: n}', "");
Expect(1, 43709, '\p{LOE=		n}', "");
Expect(0, 43709, '\p{^LOE=		n}', "");
Expect(0, 43709, '\P{LOE=		n}', "");
Expect(1, 43709, '\P{^LOE=		n}', "");
Expect(0, 43708, '\p{LOE=		n}', "");
Expect(1, 43708, '\p{^LOE=		n}', "");
Expect(1, 43708, '\P{LOE=		n}', "");
Expect(0, 43708, '\P{^LOE=		n}', "");
Error('\p{Is_Logical_Order_Exception=:=--F}');
Error('\P{Is_Logical_Order_Exception=:=--F}');
Expect(1, 43709, '\p{Is_Logical_Order_Exception=f}', "");
Expect(0, 43709, '\p{^Is_Logical_Order_Exception=f}', "");
Expect(0, 43709, '\P{Is_Logical_Order_Exception=f}', "");
Expect(1, 43709, '\P{^Is_Logical_Order_Exception=f}', "");
Expect(0, 43708, '\p{Is_Logical_Order_Exception=f}', "");
Expect(1, 43708, '\p{^Is_Logical_Order_Exception=f}', "");
Expect(1, 43708, '\P{Is_Logical_Order_Exception=f}', "");
Expect(0, 43708, '\P{^Is_Logical_Order_Exception=f}', "");
Expect(1, 43709, '\p{Is_Logical_Order_Exception=-F}', "");
Expect(0, 43709, '\p{^Is_Logical_Order_Exception=-F}', "");
Expect(0, 43709, '\P{Is_Logical_Order_Exception=-F}', "");
Expect(1, 43709, '\P{^Is_Logical_Order_Exception=-F}', "");
Expect(0, 43708, '\p{Is_Logical_Order_Exception=-F}', "");
Expect(1, 43708, '\p{^Is_Logical_Order_Exception=-F}', "");
Expect(1, 43708, '\P{Is_Logical_Order_Exception=-F}', "");
Expect(0, 43708, '\P{^Is_Logical_Order_Exception=-F}', "");
Error('\p{Is_LOE=:=-_false}');
Error('\P{Is_LOE=:=-_false}');
Expect(1, 43709, '\p{Is_LOE=false}', "");
Expect(0, 43709, '\p{^Is_LOE=false}', "");
Expect(0, 43709, '\P{Is_LOE=false}', "");
Expect(1, 43709, '\P{^Is_LOE=false}', "");
Expect(0, 43708, '\p{Is_LOE=false}', "");
Expect(1, 43708, '\p{^Is_LOE=false}', "");
Expect(1, 43708, '\P{Is_LOE=false}', "");
Expect(0, 43708, '\P{^Is_LOE=false}', "");
Expect(1, 43709, '\p{Is_LOE= false}', "");
Expect(0, 43709, '\p{^Is_LOE= false}', "");
Expect(0, 43709, '\P{Is_LOE= false}', "");
Expect(1, 43709, '\P{^Is_LOE= false}', "");
Expect(0, 43708, '\p{Is_LOE= false}', "");
Expect(1, 43708, '\p{^Is_LOE= false}', "");
Expect(1, 43708, '\P{Is_LOE= false}', "");
Expect(0, 43708, '\P{^Is_LOE= false}', "");
Error('\p{Other_Uppercase=Yes}');
Error('\P{Other_Uppercase=Yes}');
Error('\p{OUpper=Y}');
Error('\P{OUpper=Y}');
Error('\p{Is_Other_Uppercase=T}');
Error('\P{Is_Other_Uppercase=T}');
Error('\p{Is_OUpper: True}');
Error('\P{Is_OUpper: True}');
Error('\p{Other_Uppercase:	No}');
Error('\P{Other_Uppercase:	No}');
Error('\p{OUpper=N}');
Error('\P{OUpper=N}');
Error('\p{Is_Other_Uppercase=F}');
Error('\P{Is_Other_Uppercase=F}');
Error('\p{Is_OUpper=False}');
Error('\P{Is_OUpper=False}');
Error('\p{Noncharacter_Code_Point=_/a/YES}');
Error('\P{Noncharacter_Code_Point=_/a/YES}');
Expect(1, 1114111, '\p{Noncharacter_Code_Point=yes}', "");
Expect(0, 1114111, '\p{^Noncharacter_Code_Point=yes}', "");
Expect(0, 1114111, '\P{Noncharacter_Code_Point=yes}', "");
Expect(1, 1114111, '\P{^Noncharacter_Code_Point=yes}', "");
Expect(0, 1114109, '\p{Noncharacter_Code_Point=yes}', "");
Expect(1, 1114109, '\p{^Noncharacter_Code_Point=yes}', "");
Expect(1, 1114109, '\P{Noncharacter_Code_Point=yes}', "");
Expect(0, 1114109, '\P{^Noncharacter_Code_Point=yes}', "");
Expect(1, 1114111, '\p{Noncharacter_Code_Point=-yes}', "");
Expect(0, 1114111, '\p{^Noncharacter_Code_Point=-yes}', "");
Expect(0, 1114111, '\P{Noncharacter_Code_Point=-yes}', "");
Expect(1, 1114111, '\P{^Noncharacter_Code_Point=-yes}', "");
Expect(0, 1114109, '\p{Noncharacter_Code_Point=-yes}', "");
Expect(1, 1114109, '\p{^Noncharacter_Code_Point=-yes}', "");
Expect(1, 1114109, '\P{Noncharacter_Code_Point=-yes}', "");
Expect(0, 1114109, '\P{^Noncharacter_Code_Point=-yes}', "");
Error('\p{NChar=-y:=}');
Error('\P{NChar=-y:=}');
Expect(1, 1114111, '\p{NChar=y}', "");
Expect(0, 1114111, '\p{^NChar=y}', "");
Expect(0, 1114111, '\P{NChar=y}', "");
Expect(1, 1114111, '\P{^NChar=y}', "");
Expect(0, 1114109, '\p{NChar=y}', "");
Expect(1, 1114109, '\p{^NChar=y}', "");
Expect(1, 1114109, '\P{NChar=y}', "");
Expect(0, 1114109, '\P{^NChar=y}', "");
Expect(1, 1114111, '\p{NChar=	 Y}', "");
Expect(0, 1114111, '\p{^NChar=	 Y}', "");
Expect(0, 1114111, '\P{NChar=	 Y}', "");
Expect(1, 1114111, '\P{^NChar=	 Y}', "");
Expect(0, 1114109, '\p{NChar=	 Y}', "");
Expect(1, 1114109, '\p{^NChar=	 Y}', "");
Expect(1, 1114109, '\P{NChar=	 Y}', "");
Expect(0, 1114109, '\P{^NChar=	 Y}', "");
Error('\p{Is_Noncharacter_Code_Point=	/a/T}');
Error('\P{Is_Noncharacter_Code_Point=	/a/T}');
Expect(1, 1114111, '\p{Is_Noncharacter_Code_Point=t}', "");
Expect(0, 1114111, '\p{^Is_Noncharacter_Code_Point=t}', "");
Expect(0, 1114111, '\P{Is_Noncharacter_Code_Point=t}', "");
Expect(1, 1114111, '\P{^Is_Noncharacter_Code_Point=t}', "");
Expect(0, 1114109, '\p{Is_Noncharacter_Code_Point=t}', "");
Expect(1, 1114109, '\p{^Is_Noncharacter_Code_Point=t}', "");
Expect(1, 1114109, '\P{Is_Noncharacter_Code_Point=t}', "");
Expect(0, 1114109, '\P{^Is_Noncharacter_Code_Point=t}', "");
Expect(1, 1114111, '\p{Is_Noncharacter_Code_Point:	 T}', "");
Expect(0, 1114111, '\p{^Is_Noncharacter_Code_Point:	 T}', "");
Expect(0, 1114111, '\P{Is_Noncharacter_Code_Point:	 T}', "");
Expect(1, 1114111, '\P{^Is_Noncharacter_Code_Point:	 T}', "");
Expect(0, 1114109, '\p{Is_Noncharacter_Code_Point:	 T}', "");
Expect(1, 1114109, '\p{^Is_Noncharacter_Code_Point:	 T}', "");
Expect(1, 1114109, '\P{Is_Noncharacter_Code_Point:	 T}', "");
Expect(0, 1114109, '\P{^Is_Noncharacter_Code_Point:	 T}', "");
Error('\p{Is_NChar=-/a/True}');
Error('\P{Is_NChar=-/a/True}');
Expect(1, 1114111, '\p{Is_NChar=true}', "");
Expect(0, 1114111, '\p{^Is_NChar=true}', "");
Expect(0, 1114111, '\P{Is_NChar=true}', "");
Expect(1, 1114111, '\P{^Is_NChar=true}', "");
Expect(0, 1114109, '\p{Is_NChar=true}', "");
Expect(1, 1114109, '\p{^Is_NChar=true}', "");
Expect(1, 1114109, '\P{Is_NChar=true}', "");
Expect(0, 1114109, '\P{^Is_NChar=true}', "");
Expect(1, 1114111, '\p{Is_NChar=	_True}', "");
Expect(0, 1114111, '\p{^Is_NChar=	_True}', "");
Expect(0, 1114111, '\P{Is_NChar=	_True}', "");
Expect(1, 1114111, '\P{^Is_NChar=	_True}', "");
Expect(0, 1114109, '\p{Is_NChar=	_True}', "");
Expect(1, 1114109, '\p{^Is_NChar=	_True}', "");
Expect(1, 1114109, '\P{Is_NChar=	_True}', "");
Expect(0, 1114109, '\P{^Is_NChar=	_True}', "");
Error('\p{Noncharacter_Code_Point=_-no/a/}');
Error('\P{Noncharacter_Code_Point=_-no/a/}');
Expect(1, 1114109, '\p{Noncharacter_Code_Point=no}', "");
Expect(0, 1114109, '\p{^Noncharacter_Code_Point=no}', "");
Expect(0, 1114109, '\P{Noncharacter_Code_Point=no}', "");
Expect(1, 1114109, '\P{^Noncharacter_Code_Point=no}', "");
Expect(0, 1114111, '\p{Noncharacter_Code_Point=no}', "");
Expect(1, 1114111, '\p{^Noncharacter_Code_Point=no}', "");
Expect(1, 1114111, '\P{Noncharacter_Code_Point=no}', "");
Expect(0, 1114111, '\P{^Noncharacter_Code_Point=no}', "");
Expect(1, 1114109, '\p{Noncharacter_Code_Point=_	No}', "");
Expect(0, 1114109, '\p{^Noncharacter_Code_Point=_	No}', "");
Expect(0, 1114109, '\P{Noncharacter_Code_Point=_	No}', "");
Expect(1, 1114109, '\P{^Noncharacter_Code_Point=_	No}', "");
Expect(0, 1114111, '\p{Noncharacter_Code_Point=_	No}', "");
Expect(1, 1114111, '\p{^Noncharacter_Code_Point=_	No}', "");
Expect(1, 1114111, '\P{Noncharacter_Code_Point=_	No}', "");
Expect(0, 1114111, '\P{^Noncharacter_Code_Point=_	No}', "");
Error('\p{NChar= 	n/a/}');
Error('\P{NChar= 	n/a/}');
Expect(1, 1114109, '\p{NChar=n}', "");
Expect(0, 1114109, '\p{^NChar=n}', "");
Expect(0, 1114109, '\P{NChar=n}', "");
Expect(1, 1114109, '\P{^NChar=n}', "");
Expect(0, 1114111, '\p{NChar=n}', "");
Expect(1, 1114111, '\p{^NChar=n}', "");
Expect(1, 1114111, '\P{NChar=n}', "");
Expect(0, 1114111, '\P{^NChar=n}', "");
Expect(1, 1114109, '\p{NChar= -N}', "");
Expect(0, 1114109, '\p{^NChar= -N}', "");
Expect(0, 1114109, '\P{NChar= -N}', "");
Expect(1, 1114109, '\P{^NChar= -N}', "");
Expect(0, 1114111, '\p{NChar= -N}', "");
Expect(1, 1114111, '\p{^NChar= -N}', "");
Expect(1, 1114111, '\P{NChar= -N}', "");
Expect(0, 1114111, '\P{^NChar= -N}', "");
Error('\p{Is_Noncharacter_Code_Point: --F:=}');
Error('\P{Is_Noncharacter_Code_Point: --F:=}');
Expect(1, 1114109, '\p{Is_Noncharacter_Code_Point=f}', "");
Expect(0, 1114109, '\p{^Is_Noncharacter_Code_Point=f}', "");
Expect(0, 1114109, '\P{Is_Noncharacter_Code_Point=f}', "");
Expect(1, 1114109, '\P{^Is_Noncharacter_Code_Point=f}', "");
Expect(0, 1114111, '\p{Is_Noncharacter_Code_Point=f}', "");
Expect(1, 1114111, '\p{^Is_Noncharacter_Code_Point=f}', "");
Expect(1, 1114111, '\P{Is_Noncharacter_Code_Point=f}', "");
Expect(0, 1114111, '\P{^Is_Noncharacter_Code_Point=f}', "");
Expect(1, 1114109, '\p{Is_Noncharacter_Code_Point=  F}', "");
Expect(0, 1114109, '\p{^Is_Noncharacter_Code_Point=  F}', "");
Expect(0, 1114109, '\P{Is_Noncharacter_Code_Point=  F}', "");
Expect(1, 1114109, '\P{^Is_Noncharacter_Code_Point=  F}', "");
Expect(0, 1114111, '\p{Is_Noncharacter_Code_Point=  F}', "");
Expect(1, 1114111, '\p{^Is_Noncharacter_Code_Point=  F}', "");
Expect(1, 1114111, '\P{Is_Noncharacter_Code_Point=  F}', "");
Expect(0, 1114111, '\P{^Is_Noncharacter_Code_Point=  F}', "");
Error('\p{Is_NChar=-:=False}');
Error('\P{Is_NChar=-:=False}');
Expect(1, 1114109, '\p{Is_NChar=false}', "");
Expect(0, 1114109, '\p{^Is_NChar=false}', "");
Expect(0, 1114109, '\P{Is_NChar=false}', "");
Expect(1, 1114109, '\P{^Is_NChar=false}', "");
Expect(0, 1114111, '\p{Is_NChar=false}', "");
Expect(1, 1114111, '\p{^Is_NChar=false}', "");
Expect(1, 1114111, '\P{Is_NChar=false}', "");
Expect(0, 1114111, '\P{^Is_NChar=false}', "");
Expect(1, 1114109, '\p{Is_NChar=_	FALSE}', "");
Expect(0, 1114109, '\p{^Is_NChar=_	FALSE}', "");
Expect(0, 1114109, '\P{Is_NChar=_	FALSE}', "");
Expect(1, 1114109, '\P{^Is_NChar=_	FALSE}', "");
Expect(0, 1114111, '\p{Is_NChar=_	FALSE}', "");
Expect(1, 1114111, '\p{^Is_NChar=_	FALSE}', "");
Expect(1, 1114111, '\P{Is_NChar=_	FALSE}', "");
Expect(0, 1114111, '\P{^Is_NChar=_	FALSE}', "");
Error('\p{Unified_Ideograph=:= Yes}');
Error('\P{Unified_Ideograph=:= Yes}');
Expect(1, 178205, '\p{Unified_Ideograph=yes}', "");
Expect(0, 178205, '\p{^Unified_Ideograph=yes}', "");
Expect(0, 178205, '\P{Unified_Ideograph=yes}', "");
Expect(1, 178205, '\P{^Unified_Ideograph=yes}', "");
Expect(0, 178206, '\p{Unified_Ideograph=yes}', "");
Expect(1, 178206, '\p{^Unified_Ideograph=yes}', "");
Expect(1, 178206, '\P{Unified_Ideograph=yes}', "");
Expect(0, 178206, '\P{^Unified_Ideograph=yes}', "");
Expect(1, 178205, '\p{Unified_Ideograph=-yes}', "");
Expect(0, 178205, '\p{^Unified_Ideograph=-yes}', "");
Expect(0, 178205, '\P{Unified_Ideograph=-yes}', "");
Expect(1, 178205, '\P{^Unified_Ideograph=-yes}', "");
Expect(0, 178206, '\p{Unified_Ideograph=-yes}', "");
Expect(1, 178206, '\p{^Unified_Ideograph=-yes}', "");
Expect(1, 178206, '\P{Unified_Ideograph=-yes}', "");
Expect(0, 178206, '\P{^Unified_Ideograph=-yes}', "");
Error('\p{UIdeo=	:=Y}');
Error('\P{UIdeo=	:=Y}');
Expect(1, 178205, '\p{UIdeo=y}', "");
Expect(0, 178205, '\p{^UIdeo=y}', "");
Expect(0, 178205, '\P{UIdeo=y}', "");
Expect(1, 178205, '\P{^UIdeo=y}', "");
Expect(0, 178206, '\p{UIdeo=y}', "");
Expect(1, 178206, '\p{^UIdeo=y}', "");
Expect(1, 178206, '\P{UIdeo=y}', "");
Expect(0, 178206, '\P{^UIdeo=y}', "");
Expect(1, 178205, '\p{UIdeo=	Y}', "");
Expect(0, 178205, '\p{^UIdeo=	Y}', "");
Expect(0, 178205, '\P{UIdeo=	Y}', "");
Expect(1, 178205, '\P{^UIdeo=	Y}', "");
Expect(0, 178206, '\p{UIdeo=	Y}', "");
Expect(1, 178206, '\p{^UIdeo=	Y}', "");
Expect(1, 178206, '\P{UIdeo=	Y}', "");
Expect(0, 178206, '\P{^UIdeo=	Y}', "");
Error('\p{Is_Unified_Ideograph=_-t:=}');
Error('\P{Is_Unified_Ideograph=_-t:=}');
Expect(1, 178205, '\p{Is_Unified_Ideograph:   t}', "");
Expect(0, 178205, '\p{^Is_Unified_Ideograph:   t}', "");
Expect(0, 178205, '\P{Is_Unified_Ideograph:   t}', "");
Expect(1, 178205, '\P{^Is_Unified_Ideograph:   t}', "");
Expect(0, 178206, '\p{Is_Unified_Ideograph:   t}', "");
Expect(1, 178206, '\p{^Is_Unified_Ideograph:   t}', "");
Expect(1, 178206, '\P{Is_Unified_Ideograph:   t}', "");
Expect(0, 178206, '\P{^Is_Unified_Ideograph:   t}', "");
Expect(1, 178205, '\p{Is_Unified_Ideograph=_-t}', "");
Expect(0, 178205, '\p{^Is_Unified_Ideograph=_-t}', "");
Expect(0, 178205, '\P{Is_Unified_Ideograph=_-t}', "");
Expect(1, 178205, '\P{^Is_Unified_Ideograph=_-t}', "");
Expect(0, 178206, '\p{Is_Unified_Ideograph=_-t}', "");
Expect(1, 178206, '\p{^Is_Unified_Ideograph=_-t}', "");
Expect(1, 178206, '\P{Is_Unified_Ideograph=_-t}', "");
Expect(0, 178206, '\P{^Is_Unified_Ideograph=_-t}', "");
Error('\p{Is_UIdeo=_-true/a/}');
Error('\P{Is_UIdeo=_-true/a/}');
Expect(1, 178205, '\p{Is_UIdeo=true}', "");
Expect(0, 178205, '\p{^Is_UIdeo=true}', "");
Expect(0, 178205, '\P{Is_UIdeo=true}', "");
Expect(1, 178205, '\P{^Is_UIdeo=true}', "");
Expect(0, 178206, '\p{Is_UIdeo=true}', "");
Expect(1, 178206, '\p{^Is_UIdeo=true}', "");
Expect(1, 178206, '\P{Is_UIdeo=true}', "");
Expect(0, 178206, '\P{^Is_UIdeo=true}', "");
Expect(1, 178205, '\p{Is_UIdeo= true}', "");
Expect(0, 178205, '\p{^Is_UIdeo= true}', "");
Expect(0, 178205, '\P{Is_UIdeo= true}', "");
Expect(1, 178205, '\P{^Is_UIdeo= true}', "");
Expect(0, 178206, '\p{Is_UIdeo= true}', "");
Expect(1, 178206, '\p{^Is_UIdeo= true}', "");
Expect(1, 178206, '\P{Is_UIdeo= true}', "");
Expect(0, 178206, '\P{^Is_UIdeo= true}', "");
Error('\p{Unified_Ideograph=_:=no}');
Error('\P{Unified_Ideograph=_:=no}');
Expect(1, 178206, '\p{Unified_Ideograph=no}', "");
Expect(0, 178206, '\p{^Unified_Ideograph=no}', "");
Expect(0, 178206, '\P{Unified_Ideograph=no}', "");
Expect(1, 178206, '\P{^Unified_Ideograph=no}', "");
Expect(0, 178205, '\p{Unified_Ideograph=no}', "");
Expect(1, 178205, '\p{^Unified_Ideograph=no}', "");
Expect(1, 178205, '\P{Unified_Ideograph=no}', "");
Expect(0, 178205, '\P{^Unified_Ideograph=no}', "");
Expect(1, 178206, '\p{Unified_Ideograph= 	No}', "");
Expect(0, 178206, '\p{^Unified_Ideograph= 	No}', "");
Expect(0, 178206, '\P{Unified_Ideograph= 	No}', "");
Expect(1, 178206, '\P{^Unified_Ideograph= 	No}', "");
Expect(0, 178205, '\p{Unified_Ideograph= 	No}', "");
Expect(1, 178205, '\p{^Unified_Ideograph= 	No}', "");
Expect(1, 178205, '\P{Unified_Ideograph= 	No}', "");
Expect(0, 178205, '\P{^Unified_Ideograph= 	No}', "");
Error('\p{UIdeo=--N:=}');
Error('\P{UIdeo=--N:=}');
Expect(1, 178206, '\p{UIdeo=n}', "");
Expect(0, 178206, '\p{^UIdeo=n}', "");
Expect(0, 178206, '\P{UIdeo=n}', "");
Expect(1, 178206, '\P{^UIdeo=n}', "");
Expect(0, 178205, '\p{UIdeo=n}', "");
Expect(1, 178205, '\p{^UIdeo=n}', "");
Expect(1, 178205, '\P{UIdeo=n}', "");
Expect(0, 178205, '\P{^UIdeo=n}', "");
Expect(1, 178206, '\p{UIdeo= N}', "");
Expect(0, 178206, '\p{^UIdeo= N}', "");
Expect(0, 178206, '\P{UIdeo= N}', "");
Expect(1, 178206, '\P{^UIdeo= N}', "");
Expect(0, 178205, '\p{UIdeo= N}', "");
Expect(1, 178205, '\p{^UIdeo= N}', "");
Expect(1, 178205, '\P{UIdeo= N}', "");
Expect(0, 178205, '\P{^UIdeo= N}', "");
Error('\p{Is_Unified_Ideograph= F/a/}');
Error('\P{Is_Unified_Ideograph= F/a/}');
Expect(1, 178206, '\p{Is_Unified_Ideograph=f}', "");
Expect(0, 178206, '\p{^Is_Unified_Ideograph=f}', "");
Expect(0, 178206, '\P{Is_Unified_Ideograph=f}', "");
Expect(1, 178206, '\P{^Is_Unified_Ideograph=f}', "");
Expect(0, 178205, '\p{Is_Unified_Ideograph=f}', "");
Expect(1, 178205, '\p{^Is_Unified_Ideograph=f}', "");
Expect(1, 178205, '\P{Is_Unified_Ideograph=f}', "");
Expect(0, 178205, '\P{^Is_Unified_Ideograph=f}', "");
Expect(1, 178206, '\p{Is_Unified_Ideograph=_	F}', "");
Expect(0, 178206, '\p{^Is_Unified_Ideograph=_	F}', "");
Expect(0, 178206, '\P{Is_Unified_Ideograph=_	F}', "");
Expect(1, 178206, '\P{^Is_Unified_Ideograph=_	F}', "");
Expect(0, 178205, '\p{Is_Unified_Ideograph=_	F}', "");
Expect(1, 178205, '\p{^Is_Unified_Ideograph=_	F}', "");
Expect(1, 178205, '\P{Is_Unified_Ideograph=_	F}', "");
Expect(0, 178205, '\P{^Is_Unified_Ideograph=_	F}', "");
Error('\p{Is_UIdeo=_:=FALSE}');
Error('\P{Is_UIdeo=_:=FALSE}');
Expect(1, 178206, '\p{Is_UIdeo=false}', "");
Expect(0, 178206, '\p{^Is_UIdeo=false}', "");
Expect(0, 178206, '\P{Is_UIdeo=false}', "");
Expect(1, 178206, '\P{^Is_UIdeo=false}', "");
Expect(0, 178205, '\p{Is_UIdeo=false}', "");
Expect(1, 178205, '\p{^Is_UIdeo=false}', "");
Expect(1, 178205, '\P{Is_UIdeo=false}', "");
Expect(0, 178205, '\P{^Is_UIdeo=false}', "");
Expect(1, 178206, '\p{Is_UIdeo=	-FALSE}', "");
Expect(0, 178206, '\p{^Is_UIdeo=	-FALSE}', "");
Expect(0, 178206, '\P{Is_UIdeo=	-FALSE}', "");
Expect(1, 178206, '\P{^Is_UIdeo=	-FALSE}', "");
Expect(0, 178205, '\p{Is_UIdeo=	-FALSE}', "");
Expect(1, 178205, '\p{^Is_UIdeo=	-FALSE}', "");
Expect(1, 178205, '\P{Is_UIdeo=	-FALSE}', "");
Expect(0, 178205, '\P{^Is_UIdeo=	-FALSE}', "");
Error('\p{Hyphen=/a/	yes}');
Error('\P{Hyphen=/a/	yes}');
Expect(1, 65381, '\p{Hyphen:   yes}', 'deprecated');
Expect(0, 65381, '\p{^Hyphen:   yes}', 'deprecated');
Expect(0, 65381, '\P{Hyphen:   yes}', 'deprecated');
Expect(1, 65381, '\P{^Hyphen:   yes}', 'deprecated');
Expect(0, 65382, '\p{Hyphen:   yes}', 'deprecated');
Expect(1, 65382, '\p{^Hyphen:   yes}', 'deprecated');
Expect(1, 65382, '\P{Hyphen:   yes}', 'deprecated');
Expect(0, 65382, '\P{^Hyphen:   yes}', 'deprecated');
Expect(1, 65381, '\p{Hyphen:	 	Yes}', 'deprecated');
Expect(0, 65381, '\p{^Hyphen:	 	Yes}', 'deprecated');
Expect(0, 65381, '\P{Hyphen:	 	Yes}', 'deprecated');
Expect(1, 65381, '\P{^Hyphen:	 	Yes}', 'deprecated');
Expect(0, 65382, '\p{Hyphen:	 	Yes}', 'deprecated');
Expect(1, 65382, '\p{^Hyphen:	 	Yes}', 'deprecated');
Expect(1, 65382, '\P{Hyphen:	 	Yes}', 'deprecated');
Expect(0, 65382, '\P{^Hyphen:	 	Yes}', 'deprecated');
Error('\p{Is_Hyphen=:=_Y}');
Error('\P{Is_Hyphen=:=_Y}');
Expect(1, 65381, '\p{Is_Hyphen=y}', 'deprecated');
Expect(0, 65381, '\p{^Is_Hyphen=y}', 'deprecated');
Expect(0, 65381, '\P{Is_Hyphen=y}', 'deprecated');
Expect(1, 65381, '\P{^Is_Hyphen=y}', 'deprecated');
Expect(0, 65382, '\p{Is_Hyphen=y}', 'deprecated');
Expect(1, 65382, '\p{^Is_Hyphen=y}', 'deprecated');
Expect(1, 65382, '\P{Is_Hyphen=y}', 'deprecated');
Expect(0, 65382, '\P{^Is_Hyphen=y}', 'deprecated');
Expect(1, 65381, '\p{Is_Hyphen= _Y}', 'deprecated');
Expect(0, 65381, '\p{^Is_Hyphen= _Y}', 'deprecated');
Expect(0, 65381, '\P{Is_Hyphen= _Y}', 'deprecated');
Expect(1, 65381, '\P{^Is_Hyphen= _Y}', 'deprecated');
Expect(0, 65382, '\p{Is_Hyphen= _Y}', 'deprecated');
Expect(1, 65382, '\p{^Is_Hyphen= _Y}', 'deprecated');
Expect(1, 65382, '\P{Is_Hyphen= _Y}', 'deprecated');
Expect(0, 65382, '\P{^Is_Hyphen= _Y}', 'deprecated');
Error('\p{Hyphen=-T/a/}');
Error('\P{Hyphen=-T/a/}');
Expect(1, 65381, '\p{Hyphen:	t}', 'deprecated');
Expect(0, 65381, '\p{^Hyphen:	t}', 'deprecated');
Expect(0, 65381, '\P{Hyphen:	t}', 'deprecated');
Expect(1, 65381, '\P{^Hyphen:	t}', 'deprecated');
Expect(0, 65382, '\p{Hyphen:	t}', 'deprecated');
Expect(1, 65382, '\p{^Hyphen:	t}', 'deprecated');
Expect(1, 65382, '\P{Hyphen:	t}', 'deprecated');
Expect(0, 65382, '\P{^Hyphen:	t}', 'deprecated');
Expect(1, 65381, '\p{Hyphen=_ t}', 'deprecated');
Expect(0, 65381, '\p{^Hyphen=_ t}', 'deprecated');
Expect(0, 65381, '\P{Hyphen=_ t}', 'deprecated');
Expect(1, 65381, '\P{^Hyphen=_ t}', 'deprecated');
Expect(0, 65382, '\p{Hyphen=_ t}', 'deprecated');
Expect(1, 65382, '\p{^Hyphen=_ t}', 'deprecated');
Expect(1, 65382, '\P{Hyphen=_ t}', 'deprecated');
Expect(0, 65382, '\P{^Hyphen=_ t}', 'deprecated');
Error('\p{Is_Hyphen=	:=True}');
Error('\P{Is_Hyphen=	:=True}');
Expect(1, 65381, '\p{Is_Hyphen=true}', 'deprecated');
Expect(0, 65381, '\p{^Is_Hyphen=true}', 'deprecated');
Expect(0, 65381, '\P{Is_Hyphen=true}', 'deprecated');
Expect(1, 65381, '\P{^Is_Hyphen=true}', 'deprecated');
Expect(0, 65382, '\p{Is_Hyphen=true}', 'deprecated');
Expect(1, 65382, '\p{^Is_Hyphen=true}', 'deprecated');
Expect(1, 65382, '\P{Is_Hyphen=true}', 'deprecated');
Expect(0, 65382, '\P{^Is_Hyphen=true}', 'deprecated');
Expect(1, 65381, '\p{Is_Hyphen= 	True}', 'deprecated');
Expect(0, 65381, '\p{^Is_Hyphen= 	True}', 'deprecated');
Expect(0, 65381, '\P{Is_Hyphen= 	True}', 'deprecated');
Expect(1, 65381, '\P{^Is_Hyphen= 	True}', 'deprecated');
Expect(0, 65382, '\p{Is_Hyphen= 	True}', 'deprecated');
Expect(1, 65382, '\p{^Is_Hyphen= 	True}', 'deprecated');
Expect(1, 65382, '\P{Is_Hyphen= 	True}', 'deprecated');
Expect(0, 65382, '\P{^Is_Hyphen= 	True}', 'deprecated');
Error('\p{Hyphen=	No:=}');
Error('\P{Hyphen=	No:=}');
Expect(1, 65382, '\p{Hyphen=no}', 'deprecated');
Expect(0, 65382, '\p{^Hyphen=no}', 'deprecated');
Expect(0, 65382, '\P{Hyphen=no}', 'deprecated');
Expect(1, 65382, '\P{^Hyphen=no}', 'deprecated');
Expect(0, 65381, '\p{Hyphen=no}', 'deprecated');
Expect(1, 65381, '\p{^Hyphen=no}', 'deprecated');
Expect(1, 65381, '\P{Hyphen=no}', 'deprecated');
Expect(0, 65381, '\P{^Hyphen=no}', 'deprecated');
Expect(1, 65382, '\p{Hyphen=- NO}', 'deprecated');
Expect(0, 65382, '\p{^Hyphen=- NO}', 'deprecated');
Expect(0, 65382, '\P{Hyphen=- NO}', 'deprecated');
Expect(1, 65382, '\P{^Hyphen=- NO}', 'deprecated');
Expect(0, 65381, '\p{Hyphen=- NO}', 'deprecated');
Expect(1, 65381, '\p{^Hyphen=- NO}', 'deprecated');
Expect(1, 65381, '\P{Hyphen=- NO}', 'deprecated');
Expect(0, 65381, '\P{^Hyphen=- NO}', 'deprecated');
Error('\p{Is_Hyphen=/a/_N}');
Error('\P{Is_Hyphen=/a/_N}');
Expect(1, 65382, '\p{Is_Hyphen:	n}', 'deprecated');
Expect(0, 65382, '\p{^Is_Hyphen:	n}', 'deprecated');
Expect(0, 65382, '\P{Is_Hyphen:	n}', 'deprecated');
Expect(1, 65382, '\P{^Is_Hyphen:	n}', 'deprecated');
Expect(0, 65381, '\p{Is_Hyphen:	n}', 'deprecated');
Expect(1, 65381, '\p{^Is_Hyphen:	n}', 'deprecated');
Expect(1, 65381, '\P{Is_Hyphen:	n}', 'deprecated');
Expect(0, 65381, '\P{^Is_Hyphen:	n}', 'deprecated');
Expect(1, 65382, '\p{Is_Hyphen:	_ n}', 'deprecated');
Expect(0, 65382, '\p{^Is_Hyphen:	_ n}', 'deprecated');
Expect(0, 65382, '\P{Is_Hyphen:	_ n}', 'deprecated');
Expect(1, 65382, '\P{^Is_Hyphen:	_ n}', 'deprecated');
Expect(0, 65381, '\p{Is_Hyphen:	_ n}', 'deprecated');
Expect(1, 65381, '\p{^Is_Hyphen:	_ n}', 'deprecated');
Expect(1, 65381, '\P{Is_Hyphen:	_ n}', 'deprecated');
Expect(0, 65381, '\P{^Is_Hyphen:	_ n}', 'deprecated');
Error('\p{Hyphen=:= -F}');
Error('\P{Hyphen=:= -F}');
Expect(1, 65382, '\p{Hyphen=f}', 'deprecated');
Expect(0, 65382, '\p{^Hyphen=f}', 'deprecated');
Expect(0, 65382, '\P{Hyphen=f}', 'deprecated');
Expect(1, 65382, '\P{^Hyphen=f}', 'deprecated');
Expect(0, 65381, '\p{Hyphen=f}', 'deprecated');
Expect(1, 65381, '\p{^Hyphen=f}', 'deprecated');
Expect(1, 65381, '\P{Hyphen=f}', 'deprecated');
Expect(0, 65381, '\P{^Hyphen=f}', 'deprecated');
Expect(1, 65382, '\p{Hyphen=_F}', 'deprecated');
Expect(0, 65382, '\p{^Hyphen=_F}', 'deprecated');
Expect(0, 65382, '\P{Hyphen=_F}', 'deprecated');
Expect(1, 65382, '\P{^Hyphen=_F}', 'deprecated');
Expect(0, 65381, '\p{Hyphen=_F}', 'deprecated');
Expect(1, 65381, '\p{^Hyphen=_F}', 'deprecated');
Expect(1, 65381, '\P{Hyphen=_F}', 'deprecated');
Expect(0, 65381, '\P{^Hyphen=_F}', 'deprecated');
Error('\p{Is_Hyphen=  False/a/}');
Error('\P{Is_Hyphen=  False/a/}');
Expect(1, 65382, '\p{Is_Hyphen:false}', 'deprecated');
Expect(0, 65382, '\p{^Is_Hyphen:false}', 'deprecated');
Expect(0, 65382, '\P{Is_Hyphen:false}', 'deprecated');
Expect(1, 65382, '\P{^Is_Hyphen:false}', 'deprecated');
Expect(0, 65381, '\p{Is_Hyphen:false}', 'deprecated');
Expect(1, 65381, '\p{^Is_Hyphen:false}', 'deprecated');
Expect(1, 65381, '\P{Is_Hyphen:false}', 'deprecated');
Expect(0, 65381, '\P{^Is_Hyphen:false}', 'deprecated');
Expect(1, 65382, '\p{Is_Hyphen=	FALSE}', 'deprecated');
Expect(0, 65382, '\p{^Is_Hyphen=	FALSE}', 'deprecated');
Expect(0, 65382, '\P{Is_Hyphen=	FALSE}', 'deprecated');
Expect(1, 65382, '\P{^Is_Hyphen=	FALSE}', 'deprecated');
Expect(0, 65381, '\p{Is_Hyphen=	FALSE}', 'deprecated');
Expect(1, 65381, '\p{^Is_Hyphen=	FALSE}', 'deprecated');
Expect(1, 65381, '\P{Is_Hyphen=	FALSE}', 'deprecated');
Expect(0, 65381, '\P{^Is_Hyphen=	FALSE}', 'deprecated');
Error('\p{Diacritic:	:=YES}');
Error('\P{Diacritic:	:=YES}');
Expect(1, 119213, '\p{Diacritic=yes}', "");
Expect(0, 119213, '\p{^Diacritic=yes}', "");
Expect(0, 119213, '\P{Diacritic=yes}', "");
Expect(1, 119213, '\P{^Diacritic=yes}', "");
Expect(0, 119214, '\p{Diacritic=yes}', "");
Expect(1, 119214, '\p{^Diacritic=yes}', "");
Expect(1, 119214, '\P{Diacritic=yes}', "");
Expect(0, 119214, '\P{^Diacritic=yes}', "");
Expect(1, 119213, '\p{Diacritic=_Yes}', "");
Expect(0, 119213, '\p{^Diacritic=_Yes}', "");
Expect(0, 119213, '\P{Diacritic=_Yes}', "");
Expect(1, 119213, '\P{^Diacritic=_Yes}', "");
Expect(0, 119214, '\p{Diacritic=_Yes}', "");
Expect(1, 119214, '\p{^Diacritic=_Yes}', "");
Expect(1, 119214, '\P{Diacritic=_Yes}', "");
Expect(0, 119214, '\P{^Diacritic=_Yes}', "");
Error('\p{Dia=-/a/Y}');
Error('\P{Dia=-/a/Y}');
Expect(1, 119213, '\p{Dia=y}', "");
Expect(0, 119213, '\p{^Dia=y}', "");
Expect(0, 119213, '\P{Dia=y}', "");
Expect(1, 119213, '\P{^Dia=y}', "");
Expect(0, 119214, '\p{Dia=y}', "");
Expect(1, 119214, '\p{^Dia=y}', "");
Expect(1, 119214, '\P{Dia=y}', "");
Expect(0, 119214, '\P{^Dia=y}', "");
Expect(1, 119213, '\p{Dia= Y}', "");
Expect(0, 119213, '\p{^Dia= Y}', "");
Expect(0, 119213, '\P{Dia= Y}', "");
Expect(1, 119213, '\P{^Dia= Y}', "");
Expect(0, 119214, '\p{Dia= Y}', "");
Expect(1, 119214, '\p{^Dia= Y}', "");
Expect(1, 119214, '\P{Dia= Y}', "");
Expect(0, 119214, '\P{^Dia= Y}', "");
Error('\p{Is_Diacritic=  t/a/}');
Error('\P{Is_Diacritic=  t/a/}');
Expect(1, 119213, '\p{Is_Diacritic=t}', "");
Expect(0, 119213, '\p{^Is_Diacritic=t}', "");
Expect(0, 119213, '\P{Is_Diacritic=t}', "");
Expect(1, 119213, '\P{^Is_Diacritic=t}', "");
Expect(0, 119214, '\p{Is_Diacritic=t}', "");
Expect(1, 119214, '\p{^Is_Diacritic=t}', "");
Expect(1, 119214, '\P{Is_Diacritic=t}', "");
Expect(0, 119214, '\P{^Is_Diacritic=t}', "");
Expect(1, 119213, '\p{Is_Diacritic=-T}', "");
Expect(0, 119213, '\p{^Is_Diacritic=-T}', "");
Expect(0, 119213, '\P{Is_Diacritic=-T}', "");
Expect(1, 119213, '\P{^Is_Diacritic=-T}', "");
Expect(0, 119214, '\p{Is_Diacritic=-T}', "");
Expect(1, 119214, '\p{^Is_Diacritic=-T}', "");
Expect(1, 119214, '\P{Is_Diacritic=-T}', "");
Expect(0, 119214, '\P{^Is_Diacritic=-T}', "");
Error('\p{Is_Dia=	/a/TRUE}');
Error('\P{Is_Dia=	/a/TRUE}');
Expect(1, 119213, '\p{Is_Dia=true}', "");
Expect(0, 119213, '\p{^Is_Dia=true}', "");
Expect(0, 119213, '\P{Is_Dia=true}', "");
Expect(1, 119213, '\P{^Is_Dia=true}', "");
Expect(0, 119214, '\p{Is_Dia=true}', "");
Expect(1, 119214, '\p{^Is_Dia=true}', "");
Expect(1, 119214, '\P{Is_Dia=true}', "");
Expect(0, 119214, '\P{^Is_Dia=true}', "");
Expect(1, 119213, '\p{Is_Dia=- true}', "");
Expect(0, 119213, '\p{^Is_Dia=- true}', "");
Expect(0, 119213, '\P{Is_Dia=- true}', "");
Expect(1, 119213, '\P{^Is_Dia=- true}', "");
Expect(0, 119214, '\p{Is_Dia=- true}', "");
Expect(1, 119214, '\p{^Is_Dia=- true}', "");
Expect(1, 119214, '\P{Is_Dia=- true}', "");
Expect(0, 119214, '\P{^Is_Dia=- true}', "");
Error('\p{Diacritic=:= No}');
Error('\P{Diacritic=:= No}');
Expect(1, 119214, '\p{Diacritic=no}', "");
Expect(0, 119214, '\p{^Diacritic=no}', "");
Expect(0, 119214, '\P{Diacritic=no}', "");
Expect(1, 119214, '\P{^Diacritic=no}', "");
Expect(0, 119213, '\p{Diacritic=no}', "");
Expect(1, 119213, '\p{^Diacritic=no}', "");
Expect(1, 119213, '\P{Diacritic=no}', "");
Expect(0, 119213, '\P{^Diacritic=no}', "");
Expect(1, 119214, '\p{Diacritic=-no}', "");
Expect(0, 119214, '\p{^Diacritic=-no}', "");
Expect(0, 119214, '\P{Diacritic=-no}', "");
Expect(1, 119214, '\P{^Diacritic=-no}', "");
Expect(0, 119213, '\p{Diacritic=-no}', "");
Expect(1, 119213, '\p{^Diacritic=-no}', "");
Expect(1, 119213, '\P{Diacritic=-no}', "");
Expect(0, 119213, '\P{^Diacritic=-no}', "");
Error('\p{Dia:   :=_-N}');
Error('\P{Dia:   :=_-N}');
Expect(1, 119214, '\p{Dia=n}', "");
Expect(0, 119214, '\p{^Dia=n}', "");
Expect(0, 119214, '\P{Dia=n}', "");
Expect(1, 119214, '\P{^Dia=n}', "");
Expect(0, 119213, '\p{Dia=n}', "");
Expect(1, 119213, '\p{^Dia=n}', "");
Expect(1, 119213, '\P{Dia=n}', "");
Expect(0, 119213, '\P{^Dia=n}', "");
Expect(1, 119214, '\p{Dia=_N}', "");
Expect(0, 119214, '\p{^Dia=_N}', "");
Expect(0, 119214, '\P{Dia=_N}', "");
Expect(1, 119214, '\P{^Dia=_N}', "");
Expect(0, 119213, '\p{Dia=_N}', "");
Expect(1, 119213, '\p{^Dia=_N}', "");
Expect(1, 119213, '\P{Dia=_N}', "");
Expect(0, 119213, '\P{^Dia=_N}', "");
Error('\p{Is_Diacritic=	-F:=}');
Error('\P{Is_Diacritic=	-F:=}');
Expect(1, 119214, '\p{Is_Diacritic=f}', "");
Expect(0, 119214, '\p{^Is_Diacritic=f}', "");
Expect(0, 119214, '\P{Is_Diacritic=f}', "");
Expect(1, 119214, '\P{^Is_Diacritic=f}', "");
Expect(0, 119213, '\p{Is_Diacritic=f}', "");
Expect(1, 119213, '\p{^Is_Diacritic=f}', "");
Expect(1, 119213, '\P{Is_Diacritic=f}', "");
Expect(0, 119213, '\P{^Is_Diacritic=f}', "");
Expect(1, 119214, '\p{Is_Diacritic=__f}', "");
Expect(0, 119214, '\p{^Is_Diacritic=__f}', "");
Expect(0, 119214, '\P{Is_Diacritic=__f}', "");
Expect(1, 119214, '\P{^Is_Diacritic=__f}', "");
Expect(0, 119213, '\p{Is_Diacritic=__f}', "");
Expect(1, 119213, '\p{^Is_Diacritic=__f}', "");
Expect(1, 119213, '\P{Is_Diacritic=__f}', "");
Expect(0, 119213, '\P{^Is_Diacritic=__f}', "");
Error('\p{Is_Dia=/a/-False}');
Error('\P{Is_Dia=/a/-False}');
Expect(1, 119214, '\p{Is_Dia=false}', "");
Expect(0, 119214, '\p{^Is_Dia=false}', "");
Expect(0, 119214, '\P{Is_Dia=false}', "");
Expect(1, 119214, '\P{^Is_Dia=false}', "");
Expect(0, 119213, '\p{Is_Dia=false}', "");
Expect(1, 119213, '\p{^Is_Dia=false}', "");
Expect(1, 119213, '\P{Is_Dia=false}', "");
Expect(0, 119213, '\P{^Is_Dia=false}', "");
Expect(1, 119214, '\p{Is_Dia=-_False}', "");
Expect(0, 119214, '\p{^Is_Dia=-_False}', "");
Expect(0, 119214, '\P{Is_Dia=-_False}', "");
Expect(1, 119214, '\P{^Is_Dia=-_False}', "");
Expect(0, 119213, '\p{Is_Dia=-_False}', "");
Expect(1, 119213, '\p{^Is_Dia=-_False}', "");
Expect(1, 119213, '\P{Is_Dia=-_False}', "");
Expect(0, 119213, '\P{^Is_Dia=-_False}', "");
Error('\p{Ideographic=	_Yes/a/}');
Error('\P{Ideographic=	_Yes/a/}');
Expect(1, 195101, '\p{Ideographic=yes}', "");
Expect(0, 195101, '\p{^Ideographic=yes}', "");
Expect(0, 195101, '\P{Ideographic=yes}', "");
Expect(1, 195101, '\P{^Ideographic=yes}', "");
Expect(0, 195102, '\p{Ideographic=yes}', "");
Expect(1, 195102, '\p{^Ideographic=yes}', "");
Expect(1, 195102, '\P{Ideographic=yes}', "");
Expect(0, 195102, '\P{^Ideographic=yes}', "");
Expect(1, 195101, '\p{Ideographic=_Yes}', "");
Expect(0, 195101, '\p{^Ideographic=_Yes}', "");
Expect(0, 195101, '\P{Ideographic=_Yes}', "");
Expect(1, 195101, '\P{^Ideographic=_Yes}', "");
Expect(0, 195102, '\p{Ideographic=_Yes}', "");
Expect(1, 195102, '\p{^Ideographic=_Yes}', "");
Expect(1, 195102, '\P{Ideographic=_Yes}', "");
Expect(0, 195102, '\P{^Ideographic=_Yes}', "");
Error('\p{Ideo: -Y/a/}');
Error('\P{Ideo: -Y/a/}');
Expect(1, 195101, '\p{Ideo=y}', "");
Expect(0, 195101, '\p{^Ideo=y}', "");
Expect(0, 195101, '\P{Ideo=y}', "");
Expect(1, 195101, '\P{^Ideo=y}', "");
Expect(0, 195102, '\p{Ideo=y}', "");
Expect(1, 195102, '\p{^Ideo=y}', "");
Expect(1, 195102, '\P{Ideo=y}', "");
Expect(0, 195102, '\P{^Ideo=y}', "");
Expect(1, 195101, '\p{Ideo=- Y}', "");
Expect(0, 195101, '\p{^Ideo=- Y}', "");
Expect(0, 195101, '\P{Ideo=- Y}', "");
Expect(1, 195101, '\P{^Ideo=- Y}', "");
Expect(0, 195102, '\p{Ideo=- Y}', "");
Expect(1, 195102, '\p{^Ideo=- Y}', "");
Expect(1, 195102, '\P{Ideo=- Y}', "");
Expect(0, 195102, '\P{^Ideo=- Y}', "");
Error('\p{Is_Ideographic=/a/ -T}');
Error('\P{Is_Ideographic=/a/ -T}');
Expect(1, 195101, '\p{Is_Ideographic=t}', "");
Expect(0, 195101, '\p{^Is_Ideographic=t}', "");
Expect(0, 195101, '\P{Is_Ideographic=t}', "");
Expect(1, 195101, '\P{^Is_Ideographic=t}', "");
Expect(0, 195102, '\p{Is_Ideographic=t}', "");
Expect(1, 195102, '\p{^Is_Ideographic=t}', "");
Expect(1, 195102, '\P{Is_Ideographic=t}', "");
Expect(0, 195102, '\P{^Is_Ideographic=t}', "");
Expect(1, 195101, '\p{Is_Ideographic:	 _T}', "");
Expect(0, 195101, '\p{^Is_Ideographic:	 _T}', "");
Expect(0, 195101, '\P{Is_Ideographic:	 _T}', "");
Expect(1, 195101, '\P{^Is_Ideographic:	 _T}', "");
Expect(0, 195102, '\p{Is_Ideographic:	 _T}', "");
Expect(1, 195102, '\p{^Is_Ideographic:	 _T}', "");
Expect(1, 195102, '\P{Is_Ideographic:	 _T}', "");
Expect(0, 195102, '\P{^Is_Ideographic:	 _T}', "");
Error('\p{Is_Ideo=_/a/True}');
Error('\P{Is_Ideo=_/a/True}');
Expect(1, 195101, '\p{Is_Ideo:true}', "");
Expect(0, 195101, '\p{^Is_Ideo:true}', "");
Expect(0, 195101, '\P{Is_Ideo:true}', "");
Expect(1, 195101, '\P{^Is_Ideo:true}', "");
Expect(0, 195102, '\p{Is_Ideo:true}', "");
Expect(1, 195102, '\p{^Is_Ideo:true}', "");
Expect(1, 195102, '\P{Is_Ideo:true}', "");
Expect(0, 195102, '\P{^Is_Ideo:true}', "");
Expect(1, 195101, '\p{Is_Ideo=	-True}', "");
Expect(0, 195101, '\p{^Is_Ideo=	-True}', "");
Expect(0, 195101, '\P{Is_Ideo=	-True}', "");
Expect(1, 195101, '\P{^Is_Ideo=	-True}', "");
Expect(0, 195102, '\p{Is_Ideo=	-True}', "");
Expect(1, 195102, '\p{^Is_Ideo=	-True}', "");
Expect(1, 195102, '\P{Is_Ideo=	-True}', "");
Expect(0, 195102, '\P{^Is_Ideo=	-True}', "");
Error('\p{Ideographic= 	NO:=}');
Error('\P{Ideographic= 	NO:=}');
Expect(1, 195102, '\p{Ideographic:   no}', "");
Expect(0, 195102, '\p{^Ideographic:   no}', "");
Expect(0, 195102, '\P{Ideographic:   no}', "");
Expect(1, 195102, '\P{^Ideographic:   no}', "");
Expect(0, 195101, '\p{Ideographic:   no}', "");
Expect(1, 195101, '\p{^Ideographic:   no}', "");
Expect(1, 195101, '\P{Ideographic:   no}', "");
Expect(0, 195101, '\P{^Ideographic:   no}', "");
Expect(1, 195102, '\p{Ideographic:    	no}', "");
Expect(0, 195102, '\p{^Ideographic:    	no}', "");
Expect(0, 195102, '\P{Ideographic:    	no}', "");
Expect(1, 195102, '\P{^Ideographic:    	no}', "");
Expect(0, 195101, '\p{Ideographic:    	no}', "");
Expect(1, 195101, '\p{^Ideographic:    	no}', "");
Expect(1, 195101, '\P{Ideographic:    	no}', "");
Expect(0, 195101, '\P{^Ideographic:    	no}', "");
Error('\p{Ideo=/a/	N}');
Error('\P{Ideo=/a/	N}');
Expect(1, 195102, '\p{Ideo=n}', "");
Expect(0, 195102, '\p{^Ideo=n}', "");
Expect(0, 195102, '\P{Ideo=n}', "");
Expect(1, 195102, '\P{^Ideo=n}', "");
Expect(0, 195101, '\p{Ideo=n}', "");
Expect(1, 195101, '\p{^Ideo=n}', "");
Expect(1, 195101, '\P{Ideo=n}', "");
Expect(0, 195101, '\P{^Ideo=n}', "");
Expect(1, 195102, '\p{Ideo= _N}', "");
Expect(0, 195102, '\p{^Ideo= _N}', "");
Expect(0, 195102, '\P{Ideo= _N}', "");
Expect(1, 195102, '\P{^Ideo= _N}', "");
Expect(0, 195101, '\p{Ideo= _N}', "");
Expect(1, 195101, '\p{^Ideo= _N}', "");
Expect(1, 195101, '\P{Ideo= _N}', "");
Expect(0, 195101, '\P{^Ideo= _N}', "");
Error('\p{Is_Ideographic=/a/- f}');
Error('\P{Is_Ideographic=/a/- f}');
Expect(1, 195102, '\p{Is_Ideographic=f}', "");
Expect(0, 195102, '\p{^Is_Ideographic=f}', "");
Expect(0, 195102, '\P{Is_Ideographic=f}', "");
Expect(1, 195102, '\P{^Is_Ideographic=f}', "");
Expect(0, 195101, '\p{Is_Ideographic=f}', "");
Expect(1, 195101, '\p{^Is_Ideographic=f}', "");
Expect(1, 195101, '\P{Is_Ideographic=f}', "");
Expect(0, 195101, '\P{^Is_Ideographic=f}', "");
Expect(1, 195102, '\p{Is_Ideographic=	-f}', "");
Expect(0, 195102, '\p{^Is_Ideographic=	-f}', "");
Expect(0, 195102, '\P{Is_Ideographic=	-f}', "");
Expect(1, 195102, '\P{^Is_Ideographic=	-f}', "");
Expect(0, 195101, '\p{Is_Ideographic=	-f}', "");
Expect(1, 195101, '\p{^Is_Ideographic=	-f}', "");
Expect(1, 195101, '\P{Is_Ideographic=	-f}', "");
Expect(0, 195101, '\P{^Is_Ideographic=	-f}', "");
Error('\p{Is_Ideo=/a/_ false}');
Error('\P{Is_Ideo=/a/_ false}');
Expect(1, 195102, '\p{Is_Ideo=false}', "");
Expect(0, 195102, '\p{^Is_Ideo=false}', "");
Expect(0, 195102, '\P{Is_Ideo=false}', "");
Expect(1, 195102, '\P{^Is_Ideo=false}', "");
Expect(0, 195101, '\p{Is_Ideo=false}', "");
Expect(1, 195101, '\p{^Is_Ideo=false}', "");
Expect(1, 195101, '\P{Is_Ideo=false}', "");
Expect(0, 195101, '\P{^Is_Ideo=false}', "");
Expect(1, 195102, '\p{Is_Ideo: _	False}', "");
Expect(0, 195102, '\p{^Is_Ideo: _	False}', "");
Expect(0, 195102, '\P{Is_Ideo: _	False}', "");
Expect(1, 195102, '\P{^Is_Ideo: _	False}', "");
Expect(0, 195101, '\p{Is_Ideo: _	False}', "");
Expect(1, 195101, '\p{^Is_Ideo: _	False}', "");
Expect(1, 195101, '\P{Is_Ideo: _	False}', "");
Expect(0, 195101, '\P{^Is_Ideo: _	False}', "");
Error('\p{Other_ID_Continue=Yes}');
Error('\P{Other_ID_Continue=Yes}');
Error('\p{OIDC=Y}');
Error('\P{OIDC=Y}');
Error('\p{Is_Other_ID_Continue=T}');
Error('\P{Is_Other_ID_Continue=T}');
Error('\p{Is_OIDC=True}');
Error('\P{Is_OIDC=True}');
Error('\p{Other_ID_Continue=No}');
Error('\P{Other_ID_Continue=No}');
Error('\p{OIDC=N}');
Error('\P{OIDC=N}');
Error('\p{Is_Other_ID_Continue=F}');
Error('\P{Is_Other_ID_Continue=F}');
Error('\p{Is_OIDC=False}');
Error('\P{Is_OIDC=False}');
Error('\p{Joining_Type=/a/	-Transparent}');
Error('\P{Joining_Type=/a/	-Transparent}');
Expect(1, 917999, '\p{Joining_Type=transparent}', "");
Expect(0, 917999, '\p{^Joining_Type=transparent}', "");
Expect(0, 917999, '\P{Joining_Type=transparent}', "");
Expect(1, 917999, '\P{^Joining_Type=transparent}', "");
Expect(0, 918000, '\p{Joining_Type=transparent}', "");
Expect(1, 918000, '\p{^Joining_Type=transparent}', "");
Expect(1, 918000, '\P{Joining_Type=transparent}', "");
Expect(0, 918000, '\P{^Joining_Type=transparent}', "");
Expect(1, 917999, '\p{Joining_Type=_-Transparent}', "");
Expect(0, 917999, '\p{^Joining_Type=_-Transparent}', "");
Expect(0, 917999, '\P{Joining_Type=_-Transparent}', "");
Expect(1, 917999, '\P{^Joining_Type=_-Transparent}', "");
Expect(0, 918000, '\p{Joining_Type=_-Transparent}', "");
Expect(1, 918000, '\p{^Joining_Type=_-Transparent}', "");
Expect(1, 918000, '\P{Joining_Type=_-Transparent}', "");
Expect(0, 918000, '\P{^Joining_Type=_-Transparent}', "");
Error('\p{Jt=	 T:=}');
Error('\P{Jt=	 T:=}');
Expect(1, 917999, '\p{Jt=t}', "");
Expect(0, 917999, '\p{^Jt=t}', "");
Expect(0, 917999, '\P{Jt=t}', "");
Expect(1, 917999, '\P{^Jt=t}', "");
Expect(0, 918000, '\p{Jt=t}', "");
Expect(1, 918000, '\p{^Jt=t}', "");
Expect(1, 918000, '\P{Jt=t}', "");
Expect(0, 918000, '\P{^Jt=t}', "");
Expect(1, 917999, '\p{Jt=_-T}', "");
Expect(0, 917999, '\p{^Jt=_-T}', "");
Expect(0, 917999, '\P{Jt=_-T}', "");
Expect(1, 917999, '\P{^Jt=_-T}', "");
Expect(0, 918000, '\p{Jt=_-T}', "");
Expect(1, 918000, '\p{^Jt=_-T}', "");
Expect(1, 918000, '\P{Jt=_-T}', "");
Expect(0, 918000, '\P{^Jt=_-T}', "");
Error('\p{Is_Joining_Type=/a/- TRANSPARENT}');
Error('\P{Is_Joining_Type=/a/- TRANSPARENT}');
Expect(1, 917999, '\p{Is_Joining_Type=transparent}', "");
Expect(0, 917999, '\p{^Is_Joining_Type=transparent}', "");
Expect(0, 917999, '\P{Is_Joining_Type=transparent}', "");
Expect(1, 917999, '\P{^Is_Joining_Type=transparent}', "");
Expect(0, 918000, '\p{Is_Joining_Type=transparent}', "");
Expect(1, 918000, '\p{^Is_Joining_Type=transparent}', "");
Expect(1, 918000, '\P{Is_Joining_Type=transparent}', "");
Expect(0, 918000, '\P{^Is_Joining_Type=transparent}', "");
Expect(1, 917999, '\p{Is_Joining_Type=_transparent}', "");
Expect(0, 917999, '\p{^Is_Joining_Type=_transparent}', "");
Expect(0, 917999, '\P{Is_Joining_Type=_transparent}', "");
Expect(1, 917999, '\P{^Is_Joining_Type=_transparent}', "");
Expect(0, 918000, '\p{Is_Joining_Type=_transparent}', "");
Expect(1, 918000, '\p{^Is_Joining_Type=_transparent}', "");
Expect(1, 918000, '\P{Is_Joining_Type=_transparent}', "");
Expect(0, 918000, '\P{^Is_Joining_Type=_transparent}', "");
Error('\p{Is_Jt=/a/t}');
Error('\P{Is_Jt=/a/t}');
Expect(1, 917999, '\p{Is_Jt=t}', "");
Expect(0, 917999, '\p{^Is_Jt=t}', "");
Expect(0, 917999, '\P{Is_Jt=t}', "");
Expect(1, 917999, '\P{^Is_Jt=t}', "");
Expect(0, 918000, '\p{Is_Jt=t}', "");
Expect(1, 918000, '\p{^Is_Jt=t}', "");
Expect(1, 918000, '\P{Is_Jt=t}', "");
Expect(0, 918000, '\P{^Is_Jt=t}', "");
Expect(1, 917999, '\p{Is_Jt=--T}', "");
Expect(0, 917999, '\p{^Is_Jt=--T}', "");
Expect(0, 917999, '\P{Is_Jt=--T}', "");
Expect(1, 917999, '\P{^Is_Jt=--T}', "");
Expect(0, 918000, '\p{Is_Jt=--T}', "");
Expect(1, 918000, '\p{^Is_Jt=--T}', "");
Expect(1, 918000, '\P{Is_Jt=--T}', "");
Expect(0, 918000, '\P{^Is_Jt=--T}', "");
Error('\p{Joining_Type=:=_ Right_Joining}');
Error('\P{Joining_Type=:=_ Right_Joining}');
Expect(1, 2220, '\p{Joining_Type=rightjoining}', "");
Expect(0, 2220, '\p{^Joining_Type=rightjoining}', "");
Expect(0, 2220, '\P{Joining_Type=rightjoining}', "");
Expect(1, 2220, '\P{^Joining_Type=rightjoining}', "");
Expect(0, 2221, '\p{Joining_Type=rightjoining}', "");
Expect(1, 2221, '\p{^Joining_Type=rightjoining}', "");
Expect(1, 2221, '\P{Joining_Type=rightjoining}', "");
Expect(0, 2221, '\P{^Joining_Type=rightjoining}', "");
Expect(1, 2220, '\p{Joining_Type=_right_joining}', "");
Expect(0, 2220, '\p{^Joining_Type=_right_joining}', "");
Expect(0, 2220, '\P{Joining_Type=_right_joining}', "");
Expect(1, 2220, '\P{^Joining_Type=_right_joining}', "");
Expect(0, 2221, '\p{Joining_Type=_right_joining}', "");
Expect(1, 2221, '\p{^Joining_Type=_right_joining}', "");
Expect(1, 2221, '\P{Joining_Type=_right_joining}', "");
Expect(0, 2221, '\P{^Joining_Type=_right_joining}', "");
Error('\p{Jt=	-r:=}');
Error('\P{Jt=	-r:=}');
Expect(1, 2220, '\p{Jt=r}', "");
Expect(0, 2220, '\p{^Jt=r}', "");
Expect(0, 2220, '\P{Jt=r}', "");
Expect(1, 2220, '\P{^Jt=r}', "");
Expect(0, 2221, '\p{Jt=r}', "");
Expect(1, 2221, '\p{^Jt=r}', "");
Expect(1, 2221, '\P{Jt=r}', "");
Expect(0, 2221, '\P{^Jt=r}', "");
Expect(1, 2220, '\p{Jt:     r}', "");
Expect(0, 2220, '\p{^Jt:     r}', "");
Expect(0, 2220, '\P{Jt:     r}', "");
Expect(1, 2220, '\P{^Jt:     r}', "");
Expect(0, 2221, '\p{Jt:     r}', "");
Expect(1, 2221, '\p{^Jt:     r}', "");
Expect(1, 2221, '\P{Jt:     r}', "");
Expect(0, 2221, '\P{^Jt:     r}', "");
Error('\p{Is_Joining_Type:	-right_JOINING/a/}');
Error('\P{Is_Joining_Type:	-right_JOINING/a/}');
Expect(1, 2220, '\p{Is_Joining_Type=rightjoining}', "");
Expect(0, 2220, '\p{^Is_Joining_Type=rightjoining}', "");
Expect(0, 2220, '\P{Is_Joining_Type=rightjoining}', "");
Expect(1, 2220, '\P{^Is_Joining_Type=rightjoining}', "");
Expect(0, 2221, '\p{Is_Joining_Type=rightjoining}', "");
Expect(1, 2221, '\p{^Is_Joining_Type=rightjoining}', "");
Expect(1, 2221, '\P{Is_Joining_Type=rightjoining}', "");
Expect(0, 2221, '\P{^Is_Joining_Type=rightjoining}', "");
Expect(1, 2220, '\p{Is_Joining_Type: _Right_Joining}', "");
Expect(0, 2220, '\p{^Is_Joining_Type: _Right_Joining}', "");
Expect(0, 2220, '\P{Is_Joining_Type: _Right_Joining}', "");
Expect(1, 2220, '\P{^Is_Joining_Type: _Right_Joining}', "");
Expect(0, 2221, '\p{Is_Joining_Type: _Right_Joining}', "");
Expect(1, 2221, '\p{^Is_Joining_Type: _Right_Joining}', "");
Expect(1, 2221, '\P{Is_Joining_Type: _Right_Joining}', "");
Expect(0, 2221, '\P{^Is_Joining_Type: _Right_Joining}', "");
Error('\p{Is_Jt= -r/a/}');
Error('\P{Is_Jt= -r/a/}');
Expect(1, 2220, '\p{Is_Jt=r}', "");
Expect(0, 2220, '\p{^Is_Jt=r}', "");
Expect(0, 2220, '\P{Is_Jt=r}', "");
Expect(1, 2220, '\P{^Is_Jt=r}', "");
Expect(0, 2221, '\p{Is_Jt=r}', "");
Expect(1, 2221, '\p{^Is_Jt=r}', "");
Expect(1, 2221, '\P{Is_Jt=r}', "");
Expect(0, 2221, '\P{^Is_Jt=r}', "");
Expect(1, 2220, '\p{Is_Jt=R}', "");
Expect(0, 2220, '\p{^Is_Jt=R}', "");
Expect(0, 2220, '\P{Is_Jt=R}', "");
Expect(1, 2220, '\P{^Is_Jt=R}', "");
Expect(0, 2221, '\p{Is_Jt=R}', "");
Expect(1, 2221, '\p{^Is_Jt=R}', "");
Expect(1, 2221, '\P{Is_Jt=R}', "");
Expect(0, 2221, '\P{^Is_Jt=R}', "");
Error('\p{Joining_Type=:=		JOIN_CAUSING}');
Error('\P{Joining_Type=:=		JOIN_CAUSING}');
Expect(1, 8205, '\p{Joining_Type=joincausing}', "");
Expect(0, 8205, '\p{^Joining_Type=joincausing}', "");
Expect(0, 8205, '\P{Joining_Type=joincausing}', "");
Expect(1, 8205, '\P{^Joining_Type=joincausing}', "");
Expect(0, 8206, '\p{Joining_Type=joincausing}', "");
Expect(1, 8206, '\p{^Joining_Type=joincausing}', "");
Expect(1, 8206, '\P{Joining_Type=joincausing}', "");
Expect(0, 8206, '\P{^Joining_Type=joincausing}', "");
Expect(1, 8205, '\p{Joining_Type=	Join_causing}', "");
Expect(0, 8205, '\p{^Joining_Type=	Join_causing}', "");
Expect(0, 8205, '\P{Joining_Type=	Join_causing}', "");
Expect(1, 8205, '\P{^Joining_Type=	Join_causing}', "");
Expect(0, 8206, '\p{Joining_Type=	Join_causing}', "");
Expect(1, 8206, '\p{^Joining_Type=	Join_causing}', "");
Expect(1, 8206, '\P{Joining_Type=	Join_causing}', "");
Expect(0, 8206, '\P{^Joining_Type=	Join_causing}', "");
Error('\p{Jt=  C/a/}');
Error('\P{Jt=  C/a/}');
Expect(1, 8205, '\p{Jt=c}', "");
Expect(0, 8205, '\p{^Jt=c}', "");
Expect(0, 8205, '\P{Jt=c}', "");
Expect(1, 8205, '\P{^Jt=c}', "");
Expect(0, 8206, '\p{Jt=c}', "");
Expect(1, 8206, '\p{^Jt=c}', "");
Expect(1, 8206, '\P{Jt=c}', "");
Expect(0, 8206, '\P{^Jt=c}', "");
Expect(1, 8205, '\p{Jt=-c}', "");
Expect(0, 8205, '\p{^Jt=-c}', "");
Expect(0, 8205, '\P{Jt=-c}', "");
Expect(1, 8205, '\P{^Jt=-c}', "");
Expect(0, 8206, '\p{Jt=-c}', "");
Expect(1, 8206, '\p{^Jt=-c}', "");
Expect(1, 8206, '\P{Jt=-c}', "");
Expect(0, 8206, '\P{^Jt=-c}', "");
Error('\p{Is_Joining_Type=-_JOIN_Causing:=}');
Error('\P{Is_Joining_Type=-_JOIN_Causing:=}');
Expect(1, 8205, '\p{Is_Joining_Type=joincausing}', "");
Expect(0, 8205, '\p{^Is_Joining_Type=joincausing}', "");
Expect(0, 8205, '\P{Is_Joining_Type=joincausing}', "");
Expect(1, 8205, '\P{^Is_Joining_Type=joincausing}', "");
Expect(0, 8206, '\p{Is_Joining_Type=joincausing}', "");
Expect(1, 8206, '\p{^Is_Joining_Type=joincausing}', "");
Expect(1, 8206, '\P{Is_Joining_Type=joincausing}', "");
Expect(0, 8206, '\P{^Is_Joining_Type=joincausing}', "");
Expect(1, 8205, '\p{Is_Joining_Type=__Join_causing}', "");
Expect(0, 8205, '\p{^Is_Joining_Type=__Join_causing}', "");
Expect(0, 8205, '\P{Is_Joining_Type=__Join_causing}', "");
Expect(1, 8205, '\P{^Is_Joining_Type=__Join_causing}', "");
Expect(0, 8206, '\p{Is_Joining_Type=__Join_causing}', "");
Expect(1, 8206, '\p{^Is_Joining_Type=__Join_causing}', "");
Expect(1, 8206, '\P{Is_Joining_Type=__Join_causing}', "");
Expect(0, 8206, '\P{^Is_Joining_Type=__Join_causing}', "");
Error('\p{Is_Jt=:=	c}');
Error('\P{Is_Jt=:=	c}');
Expect(1, 8205, '\p{Is_Jt=c}', "");
Expect(0, 8205, '\p{^Is_Jt=c}', "");
Expect(0, 8205, '\P{Is_Jt=c}', "");
Expect(1, 8205, '\P{^Is_Jt=c}', "");
Expect(0, 8206, '\p{Is_Jt=c}', "");
Expect(1, 8206, '\p{^Is_Jt=c}', "");
Expect(1, 8206, '\P{Is_Jt=c}', "");
Expect(0, 8206, '\P{^Is_Jt=c}', "");
Expect(1, 8205, '\p{Is_Jt=	C}', "");
Expect(0, 8205, '\p{^Is_Jt=	C}', "");
Expect(0, 8205, '\P{Is_Jt=	C}', "");
Expect(1, 8205, '\P{^Is_Jt=	C}', "");
Expect(0, 8206, '\p{Is_Jt=	C}', "");
Expect(1, 8206, '\p{^Is_Jt=	C}', "");
Expect(1, 8206, '\P{Is_Jt=	C}', "");
Expect(0, 8206, '\P{^Is_Jt=	C}', "");
Error('\p{Joining_Type=		Non_Joining:=}');
Error('\P{Joining_Type=		Non_Joining:=}');
Expect(1, 918000, '\p{Joining_Type=nonjoining}', "");
Expect(0, 918000, '\p{^Joining_Type=nonjoining}', "");
Expect(0, 918000, '\P{Joining_Type=nonjoining}', "");
Expect(1, 918000, '\P{^Joining_Type=nonjoining}', "");
Expect(0, 917999, '\p{Joining_Type=nonjoining}', "");
Expect(1, 917999, '\p{^Joining_Type=nonjoining}', "");
Expect(1, 917999, '\P{Joining_Type=nonjoining}', "");
Expect(0, 917999, '\P{^Joining_Type=nonjoining}', "");
Expect(1, 918000, '\p{Joining_Type= 	Non_joining}', "");
Expect(0, 918000, '\p{^Joining_Type= 	Non_joining}', "");
Expect(0, 918000, '\P{Joining_Type= 	Non_joining}', "");
Expect(1, 918000, '\P{^Joining_Type= 	Non_joining}', "");
Expect(0, 917999, '\p{Joining_Type= 	Non_joining}', "");
Expect(1, 917999, '\p{^Joining_Type= 	Non_joining}', "");
Expect(1, 917999, '\P{Joining_Type= 	Non_joining}', "");
Expect(0, 917999, '\P{^Joining_Type= 	Non_joining}', "");
Error('\p{Jt=/a/_U}');
Error('\P{Jt=/a/_U}');
Expect(1, 918000, '\p{Jt: u}', "");
Expect(0, 918000, '\p{^Jt: u}', "");
Expect(0, 918000, '\P{Jt: u}', "");
Expect(1, 918000, '\P{^Jt: u}', "");
Expect(0, 917999, '\p{Jt: u}', "");
Expect(1, 917999, '\p{^Jt: u}', "");
Expect(1, 917999, '\P{Jt: u}', "");
Expect(0, 917999, '\P{^Jt: u}', "");
Expect(1, 918000, '\p{Jt=		u}', "");
Expect(0, 918000, '\p{^Jt=		u}', "");
Expect(0, 918000, '\P{Jt=		u}', "");
Expect(1, 918000, '\P{^Jt=		u}', "");
Expect(0, 917999, '\p{Jt=		u}', "");
Expect(1, 917999, '\p{^Jt=		u}', "");
Expect(1, 917999, '\P{Jt=		u}', "");
Expect(0, 917999, '\P{^Jt=		u}', "");
Error('\p{Is_Joining_Type=:=_non_JOINING}');
Error('\P{Is_Joining_Type=:=_non_JOINING}');
Expect(1, 918000, '\p{Is_Joining_Type=nonjoining}', "");
Expect(0, 918000, '\p{^Is_Joining_Type=nonjoining}', "");
Expect(0, 918000, '\P{Is_Joining_Type=nonjoining}', "");
Expect(1, 918000, '\P{^Is_Joining_Type=nonjoining}', "");
Expect(0, 917999, '\p{Is_Joining_Type=nonjoining}', "");
Expect(1, 917999, '\p{^Is_Joining_Type=nonjoining}', "");
Expect(1, 917999, '\P{Is_Joining_Type=nonjoining}', "");
Expect(0, 917999, '\P{^Is_Joining_Type=nonjoining}', "");
Expect(1, 918000, '\p{Is_Joining_Type= non_joining}', "");
Expect(0, 918000, '\p{^Is_Joining_Type= non_joining}', "");
Expect(0, 918000, '\P{Is_Joining_Type= non_joining}', "");
Expect(1, 918000, '\P{^Is_Joining_Type= non_joining}', "");
Expect(0, 917999, '\p{Is_Joining_Type= non_joining}', "");
Expect(1, 917999, '\p{^Is_Joining_Type= non_joining}', "");
Expect(1, 917999, '\P{Is_Joining_Type= non_joining}', "");
Expect(0, 917999, '\P{^Is_Joining_Type= non_joining}', "");
Error('\p{Is_Jt=-:=U}');
Error('\P{Is_Jt=-:=U}');
Expect(1, 918000, '\p{Is_Jt=u}', "");
Expect(0, 918000, '\p{^Is_Jt=u}', "");
Expect(0, 918000, '\P{Is_Jt=u}', "");
Expect(1, 918000, '\P{^Is_Jt=u}', "");
Expect(0, 917999, '\p{Is_Jt=u}', "");
Expect(1, 917999, '\p{^Is_Jt=u}', "");
Expect(1, 917999, '\P{Is_Jt=u}', "");
Expect(0, 917999, '\P{^Is_Jt=u}', "");
Expect(1, 918000, '\p{Is_Jt=-_u}', "");
Expect(0, 918000, '\p{^Is_Jt=-_u}', "");
Expect(0, 918000, '\P{Is_Jt=-_u}', "");
Expect(1, 918000, '\P{^Is_Jt=-_u}', "");
Expect(0, 917999, '\p{Is_Jt=-_u}', "");
Expect(1, 917999, '\p{^Is_Jt=-_u}', "");
Expect(1, 917999, '\P{Is_Jt=-_u}', "");
Expect(0, 917999, '\P{^Is_Jt=-_u}', "");
Error('\p{Joining_Type= _dual_Joining:=}');
Error('\P{Joining_Type= _dual_Joining:=}');
Expect(1, 2217, '\p{Joining_Type=dualjoining}', "");
Expect(0, 2217, '\p{^Joining_Type=dualjoining}', "");
Expect(0, 2217, '\P{Joining_Type=dualjoining}', "");
Expect(1, 2217, '\P{^Joining_Type=dualjoining}', "");
Expect(0, 2218, '\p{Joining_Type=dualjoining}', "");
Expect(1, 2218, '\p{^Joining_Type=dualjoining}', "");
Expect(1, 2218, '\P{Joining_Type=dualjoining}', "");
Expect(0, 2218, '\P{^Joining_Type=dualjoining}', "");
Expect(1, 2217, '\p{Joining_Type=-DUAL_Joining}', "");
Expect(0, 2217, '\p{^Joining_Type=-DUAL_Joining}', "");
Expect(0, 2217, '\P{Joining_Type=-DUAL_Joining}', "");
Expect(1, 2217, '\P{^Joining_Type=-DUAL_Joining}', "");
Expect(0, 2218, '\p{Joining_Type=-DUAL_Joining}', "");
Expect(1, 2218, '\p{^Joining_Type=-DUAL_Joining}', "");
Expect(1, 2218, '\P{Joining_Type=-DUAL_Joining}', "");
Expect(0, 2218, '\P{^Joining_Type=-DUAL_Joining}', "");
Error('\p{Jt=_d/a/}');
Error('\P{Jt=_d/a/}');
Expect(1, 2217, '\p{Jt=d}', "");
Expect(0, 2217, '\p{^Jt=d}', "");
Expect(0, 2217, '\P{Jt=d}', "");
Expect(1, 2217, '\P{^Jt=d}', "");
Expect(0, 2218, '\p{Jt=d}', "");
Expect(1, 2218, '\p{^Jt=d}', "");
Expect(1, 2218, '\P{Jt=d}', "");
Expect(0, 2218, '\P{^Jt=d}', "");
Expect(1, 2217, '\p{Jt= D}', "");
Expect(0, 2217, '\p{^Jt= D}', "");
Expect(0, 2217, '\P{Jt= D}', "");
Expect(1, 2217, '\P{^Jt= D}', "");
Expect(0, 2218, '\p{Jt= D}', "");
Expect(1, 2218, '\p{^Jt= D}', "");
Expect(1, 2218, '\P{Jt= D}', "");
Expect(0, 2218, '\P{^Jt= D}', "");
Error('\p{Is_Joining_Type=/a/-dual_joining}');
Error('\P{Is_Joining_Type=/a/-dual_joining}');
Expect(1, 2217, '\p{Is_Joining_Type:	dualjoining}', "");
Expect(0, 2217, '\p{^Is_Joining_Type:	dualjoining}', "");
Expect(0, 2217, '\P{Is_Joining_Type:	dualjoining}', "");
Expect(1, 2217, '\P{^Is_Joining_Type:	dualjoining}', "");
Expect(0, 2218, '\p{Is_Joining_Type:	dualjoining}', "");
Expect(1, 2218, '\p{^Is_Joining_Type:	dualjoining}', "");
Expect(1, 2218, '\P{Is_Joining_Type:	dualjoining}', "");
Expect(0, 2218, '\P{^Is_Joining_Type:	dualjoining}', "");
Expect(1, 2217, '\p{Is_Joining_Type=_DUAL_Joining}', "");
Expect(0, 2217, '\p{^Is_Joining_Type=_DUAL_Joining}', "");
Expect(0, 2217, '\P{Is_Joining_Type=_DUAL_Joining}', "");
Expect(1, 2217, '\P{^Is_Joining_Type=_DUAL_Joining}', "");
Expect(0, 2218, '\p{Is_Joining_Type=_DUAL_Joining}', "");
Expect(1, 2218, '\p{^Is_Joining_Type=_DUAL_Joining}', "");
Expect(1, 2218, '\P{Is_Joining_Type=_DUAL_Joining}', "");
Expect(0, 2218, '\P{^Is_Joining_Type=_DUAL_Joining}', "");
Error('\p{Is_Jt=	D:=}');
Error('\P{Is_Jt=	D:=}');
Expect(1, 2217, '\p{Is_Jt=d}', "");
Expect(0, 2217, '\p{^Is_Jt=d}', "");
Expect(0, 2217, '\P{Is_Jt=d}', "");
Expect(1, 2217, '\P{^Is_Jt=d}', "");
Expect(0, 2218, '\p{Is_Jt=d}', "");
Expect(1, 2218, '\p{^Is_Jt=d}', "");
Expect(1, 2218, '\P{Is_Jt=d}', "");
Expect(0, 2218, '\P{^Is_Jt=d}', "");
Expect(1, 2217, '\p{Is_Jt=--D}', "");
Expect(0, 2217, '\p{^Is_Jt=--D}', "");
Expect(0, 2217, '\P{Is_Jt=--D}', "");
Expect(1, 2217, '\P{^Is_Jt=--D}', "");
Expect(0, 2218, '\p{Is_Jt=--D}', "");
Expect(1, 2218, '\p{^Is_Jt=--D}', "");
Expect(1, 2218, '\P{Is_Jt=--D}', "");
Expect(0, 2218, '\P{^Is_Jt=--D}', "");
Error('\p{Joining_Type=	_Left_Joining:=}');
Error('\P{Joining_Type=	_Left_Joining:=}');
Expect(0, 1, '\p{Joining_Type=leftjoining}', "");
Expect(1, 1, '\p{^Joining_Type=leftjoining}', "");
Expect(1, 1, '\P{Joining_Type=leftjoining}', "");
Expect(0, 1, '\P{^Joining_Type=leftjoining}', "");
Expect(0, 1, '\p{Joining_Type= _LEFT_Joining}', "");
Expect(1, 1, '\p{^Joining_Type= _LEFT_Joining}', "");
Expect(1, 1, '\P{Joining_Type= _LEFT_Joining}', "");
Expect(0, 1, '\P{^Joining_Type= _LEFT_Joining}', "");
Error('\p{Jt=--l/a/}');
Error('\P{Jt=--l/a/}');
Expect(0, 1, '\p{Jt:   l}', "");
Expect(1, 1, '\p{^Jt:   l}', "");
Expect(1, 1, '\P{Jt:   l}', "");
Expect(0, 1, '\P{^Jt:   l}', "");
Expect(0, 1, '\p{Jt= _L}', "");
Expect(1, 1, '\p{^Jt= _L}', "");
Expect(1, 1, '\P{Jt= _L}', "");
Expect(0, 1, '\P{^Jt= _L}', "");
Error('\p{Is_Joining_Type:-Left_Joining/a/}');
Error('\P{Is_Joining_Type:-Left_Joining/a/}');
Expect(0, 1, '\p{Is_Joining_Type=leftjoining}', "");
Expect(1, 1, '\p{^Is_Joining_Type=leftjoining}', "");
Expect(1, 1, '\P{Is_Joining_Type=leftjoining}', "");
Expect(0, 1, '\P{^Is_Joining_Type=leftjoining}', "");
Expect(0, 1, '\p{Is_Joining_Type=-_LEFT_Joining}', "");
Expect(1, 1, '\p{^Is_Joining_Type=-_LEFT_Joining}', "");
Expect(1, 1, '\P{Is_Joining_Type=-_LEFT_Joining}', "");
Expect(0, 1, '\P{^Is_Joining_Type=-_LEFT_Joining}', "");
Error('\p{Is_Jt=-:=L}');
Error('\P{Is_Jt=-:=L}');
Expect(0, 1, '\p{Is_Jt=l}', "");
Expect(1, 1, '\p{^Is_Jt=l}', "");
Expect(1, 1, '\P{Is_Jt=l}', "");
Expect(0, 1, '\P{^Is_Jt=l}', "");
Expect(0, 1, '\p{Is_Jt=L}', "");
Expect(1, 1, '\p{^Is_Jt=L}', "");
Expect(1, 1, '\P{Is_Jt=L}', "");
Expect(0, 1, '\P{^Is_Jt=L}', "");
Error('\p{IDS_Binary_Operator:    :=yes}');
Error('\P{IDS_Binary_Operator:    :=yes}');
Expect(1, 12283, '\p{IDS_Binary_Operator=yes}', "");
Expect(0, 12283, '\p{^IDS_Binary_Operator=yes}', "");
Expect(0, 12283, '\P{IDS_Binary_Operator=yes}', "");
Expect(1, 12283, '\P{^IDS_Binary_Operator=yes}', "");
Expect(0, 12284, '\p{IDS_Binary_Operator=yes}', "");
Expect(1, 12284, '\p{^IDS_Binary_Operator=yes}', "");
Expect(1, 12284, '\P{IDS_Binary_Operator=yes}', "");
Expect(0, 12284, '\P{^IDS_Binary_Operator=yes}', "");
Expect(1, 12283, '\p{IDS_Binary_Operator=__Yes}', "");
Expect(0, 12283, '\p{^IDS_Binary_Operator=__Yes}', "");
Expect(0, 12283, '\P{IDS_Binary_Operator=__Yes}', "");
Expect(1, 12283, '\P{^IDS_Binary_Operator=__Yes}', "");
Expect(0, 12284, '\p{IDS_Binary_Operator=__Yes}', "");
Expect(1, 12284, '\p{^IDS_Binary_Operator=__Yes}', "");
Expect(1, 12284, '\P{IDS_Binary_Operator=__Yes}', "");
Expect(0, 12284, '\P{^IDS_Binary_Operator=__Yes}', "");
Error('\p{IDSB:	 :=Y}');
Error('\P{IDSB:	 :=Y}');
Expect(1, 12283, '\p{IDSB=y}', "");
Expect(0, 12283, '\p{^IDSB=y}', "");
Expect(0, 12283, '\P{IDSB=y}', "");
Expect(1, 12283, '\P{^IDSB=y}', "");
Expect(0, 12284, '\p{IDSB=y}', "");
Expect(1, 12284, '\p{^IDSB=y}', "");
Expect(1, 12284, '\P{IDSB=y}', "");
Expect(0, 12284, '\P{^IDSB=y}', "");
Expect(1, 12283, '\p{IDSB=_	Y}', "");
Expect(0, 12283, '\p{^IDSB=_	Y}', "");
Expect(0, 12283, '\P{IDSB=_	Y}', "");
Expect(1, 12283, '\P{^IDSB=_	Y}', "");
Expect(0, 12284, '\p{IDSB=_	Y}', "");
Expect(1, 12284, '\p{^IDSB=_	Y}', "");
Expect(1, 12284, '\P{IDSB=_	Y}', "");
Expect(0, 12284, '\P{^IDSB=_	Y}', "");
Error('\p{Is_IDS_Binary_Operator=T:=}');
Error('\P{Is_IDS_Binary_Operator=T:=}');
Expect(1, 12283, '\p{Is_IDS_Binary_Operator=t}', "");
Expect(0, 12283, '\p{^Is_IDS_Binary_Operator=t}', "");
Expect(0, 12283, '\P{Is_IDS_Binary_Operator=t}', "");
Expect(1, 12283, '\P{^Is_IDS_Binary_Operator=t}', "");
Expect(0, 12284, '\p{Is_IDS_Binary_Operator=t}', "");
Expect(1, 12284, '\p{^Is_IDS_Binary_Operator=t}', "");
Expect(1, 12284, '\P{Is_IDS_Binary_Operator=t}', "");
Expect(0, 12284, '\P{^Is_IDS_Binary_Operator=t}', "");
Expect(1, 12283, '\p{Is_IDS_Binary_Operator= T}', "");
Expect(0, 12283, '\p{^Is_IDS_Binary_Operator= T}', "");
Expect(0, 12283, '\P{Is_IDS_Binary_Operator= T}', "");
Expect(1, 12283, '\P{^Is_IDS_Binary_Operator= T}', "");
Expect(0, 12284, '\p{Is_IDS_Binary_Operator= T}', "");
Expect(1, 12284, '\p{^Is_IDS_Binary_Operator= T}', "");
Expect(1, 12284, '\P{Is_IDS_Binary_Operator= T}', "");
Expect(0, 12284, '\P{^Is_IDS_Binary_Operator= T}', "");
Error('\p{Is_IDSB= /a/TRUE}');
Error('\P{Is_IDSB= /a/TRUE}');
Expect(1, 12283, '\p{Is_IDSB=true}', "");
Expect(0, 12283, '\p{^Is_IDSB=true}', "");
Expect(0, 12283, '\P{Is_IDSB=true}', "");
Expect(1, 12283, '\P{^Is_IDSB=true}', "");
Expect(0, 12284, '\p{Is_IDSB=true}', "");
Expect(1, 12284, '\p{^Is_IDSB=true}', "");
Expect(1, 12284, '\P{Is_IDSB=true}', "");
Expect(0, 12284, '\P{^Is_IDSB=true}', "");
Expect(1, 12283, '\p{Is_IDSB= True}', "");
Expect(0, 12283, '\p{^Is_IDSB= True}', "");
Expect(0, 12283, '\P{Is_IDSB= True}', "");
Expect(1, 12283, '\P{^Is_IDSB= True}', "");
Expect(0, 12284, '\p{Is_IDSB= True}', "");
Expect(1, 12284, '\p{^Is_IDSB= True}', "");
Expect(1, 12284, '\P{Is_IDSB= True}', "");
Expect(0, 12284, '\P{^Is_IDSB= True}', "");
Error('\p{IDS_Binary_Operator= :=No}');
Error('\P{IDS_Binary_Operator= :=No}');
Expect(1, 12284, '\p{IDS_Binary_Operator:   no}', "");
Expect(0, 12284, '\p{^IDS_Binary_Operator:   no}', "");
Expect(0, 12284, '\P{IDS_Binary_Operator:   no}', "");
Expect(1, 12284, '\P{^IDS_Binary_Operator:   no}', "");
Expect(0, 12283, '\p{IDS_Binary_Operator:   no}', "");
Expect(1, 12283, '\p{^IDS_Binary_Operator:   no}', "");
Expect(1, 12283, '\P{IDS_Binary_Operator:   no}', "");
Expect(0, 12283, '\P{^IDS_Binary_Operator:   no}', "");
Expect(1, 12284, '\p{IDS_Binary_Operator= -no}', "");
Expect(0, 12284, '\p{^IDS_Binary_Operator= -no}', "");
Expect(0, 12284, '\P{IDS_Binary_Operator= -no}', "");
Expect(1, 12284, '\P{^IDS_Binary_Operator= -no}', "");
Expect(0, 12283, '\p{IDS_Binary_Operator= -no}', "");
Expect(1, 12283, '\p{^IDS_Binary_Operator= -no}', "");
Expect(1, 12283, '\P{IDS_Binary_Operator= -no}', "");
Expect(0, 12283, '\P{^IDS_Binary_Operator= -no}', "");
Error('\p{IDSB=__N/a/}');
Error('\P{IDSB=__N/a/}');
Expect(1, 12284, '\p{IDSB=n}', "");
Expect(0, 12284, '\p{^IDSB=n}', "");
Expect(0, 12284, '\P{IDSB=n}', "");
Expect(1, 12284, '\P{^IDSB=n}', "");
Expect(0, 12283, '\p{IDSB=n}', "");
Expect(1, 12283, '\p{^IDSB=n}', "");
Expect(1, 12283, '\P{IDSB=n}', "");
Expect(0, 12283, '\P{^IDSB=n}', "");
Expect(1, 12284, '\p{IDSB:	-_N}', "");
Expect(0, 12284, '\p{^IDSB:	-_N}', "");
Expect(0, 12284, '\P{IDSB:	-_N}', "");
Expect(1, 12284, '\P{^IDSB:	-_N}', "");
Expect(0, 12283, '\p{IDSB:	-_N}', "");
Expect(1, 12283, '\p{^IDSB:	-_N}', "");
Expect(1, 12283, '\P{IDSB:	-_N}', "");
Expect(0, 12283, '\P{^IDSB:	-_N}', "");
Error('\p{Is_IDS_Binary_Operator:   :=- F}');
Error('\P{Is_IDS_Binary_Operator:   :=- F}');
Expect(1, 12284, '\p{Is_IDS_Binary_Operator=f}', "");
Expect(0, 12284, '\p{^Is_IDS_Binary_Operator=f}', "");
Expect(0, 12284, '\P{Is_IDS_Binary_Operator=f}', "");
Expect(1, 12284, '\P{^Is_IDS_Binary_Operator=f}', "");
Expect(0, 12283, '\p{Is_IDS_Binary_Operator=f}', "");
Expect(1, 12283, '\p{^Is_IDS_Binary_Operator=f}', "");
Expect(1, 12283, '\P{Is_IDS_Binary_Operator=f}', "");
Expect(0, 12283, '\P{^Is_IDS_Binary_Operator=f}', "");
Expect(1, 12284, '\p{Is_IDS_Binary_Operator:   _F}', "");
Expect(0, 12284, '\p{^Is_IDS_Binary_Operator:   _F}', "");
Expect(0, 12284, '\P{Is_IDS_Binary_Operator:   _F}', "");
Expect(1, 12284, '\P{^Is_IDS_Binary_Operator:   _F}', "");
Expect(0, 12283, '\p{Is_IDS_Binary_Operator:   _F}', "");
Expect(1, 12283, '\p{^Is_IDS_Binary_Operator:   _F}', "");
Expect(1, 12283, '\P{Is_IDS_Binary_Operator:   _F}', "");
Expect(0, 12283, '\P{^Is_IDS_Binary_Operator:   _F}', "");
Error('\p{Is_IDSB=_/a/false}');
Error('\P{Is_IDSB=_/a/false}');
Expect(1, 12284, '\p{Is_IDSB=false}', "");
Expect(0, 12284, '\p{^Is_IDSB=false}', "");
Expect(0, 12284, '\P{Is_IDSB=false}', "");
Expect(1, 12284, '\P{^Is_IDSB=false}', "");
Expect(0, 12283, '\p{Is_IDSB=false}', "");
Expect(1, 12283, '\p{^Is_IDSB=false}', "");
Expect(1, 12283, '\P{Is_IDSB=false}', "");
Expect(0, 12283, '\P{^Is_IDSB=false}', "");
Expect(1, 12284, '\p{Is_IDSB=--False}', "");
Expect(0, 12284, '\p{^Is_IDSB=--False}', "");
Expect(0, 12284, '\P{Is_IDSB=--False}', "");
Expect(1, 12284, '\P{^Is_IDSB=--False}', "");
Expect(0, 12283, '\p{Is_IDSB=--False}', "");
Expect(1, 12283, '\p{^Is_IDSB=--False}', "");
Expect(1, 12283, '\P{Is_IDSB=--False}', "");
Expect(0, 12283, '\P{^Is_IDSB=--False}', "");
Error('\p{Soft_Dotted: := _yes}');
Error('\P{Soft_Dotted: := _yes}');
Expect(1, 120467, '\p{Soft_Dotted=yes}', "");
Expect(0, 120467, '\p{^Soft_Dotted=yes}', "");
Expect(0, 120467, '\P{Soft_Dotted=yes}', "");
Expect(1, 120467, '\P{^Soft_Dotted=yes}', "");
Expect(0, 120468, '\p{Soft_Dotted=yes}', "");
Expect(1, 120468, '\p{^Soft_Dotted=yes}', "");
Expect(1, 120468, '\P{Soft_Dotted=yes}', "");
Expect(0, 120468, '\P{^Soft_Dotted=yes}', "");
Expect(1, 120467, '\p{Soft_Dotted= -Yes}', "");
Expect(0, 120467, '\p{^Soft_Dotted= -Yes}', "");
Expect(0, 120467, '\P{Soft_Dotted= -Yes}', "");
Expect(1, 120467, '\P{^Soft_Dotted= -Yes}', "");
Expect(0, 120468, '\p{Soft_Dotted= -Yes}', "");
Expect(1, 120468, '\p{^Soft_Dotted= -Yes}', "");
Expect(1, 120468, '\P{Soft_Dotted= -Yes}', "");
Expect(0, 120468, '\P{^Soft_Dotted= -Yes}', "");
Error('\p{SD=/a/_-Y}');
Error('\P{SD=/a/_-Y}');
Expect(1, 120467, '\p{SD=y}', "");
Expect(0, 120467, '\p{^SD=y}', "");
Expect(0, 120467, '\P{SD=y}', "");
Expect(1, 120467, '\P{^SD=y}', "");
Expect(0, 120468, '\p{SD=y}', "");
Expect(1, 120468, '\p{^SD=y}', "");
Expect(1, 120468, '\P{SD=y}', "");
Expect(0, 120468, '\P{^SD=y}', "");
Expect(1, 120467, '\p{SD= _y}', "");
Expect(0, 120467, '\p{^SD= _y}', "");
Expect(0, 120467, '\P{SD= _y}', "");
Expect(1, 120467, '\P{^SD= _y}', "");
Expect(0, 120468, '\p{SD= _y}', "");
Expect(1, 120468, '\p{^SD= _y}', "");
Expect(1, 120468, '\P{SD= _y}', "");
Expect(0, 120468, '\P{^SD= _y}', "");
Error('\p{Is_Soft_Dotted=:=_-T}');
Error('\P{Is_Soft_Dotted=:=_-T}');
Expect(1, 120467, '\p{Is_Soft_Dotted=t}', "");
Expect(0, 120467, '\p{^Is_Soft_Dotted=t}', "");
Expect(0, 120467, '\P{Is_Soft_Dotted=t}', "");
Expect(1, 120467, '\P{^Is_Soft_Dotted=t}', "");
Expect(0, 120468, '\p{Is_Soft_Dotted=t}', "");
Expect(1, 120468, '\p{^Is_Soft_Dotted=t}', "");
Expect(1, 120468, '\P{Is_Soft_Dotted=t}', "");
Expect(0, 120468, '\P{^Is_Soft_Dotted=t}', "");
Expect(1, 120467, '\p{Is_Soft_Dotted= -T}', "");
Expect(0, 120467, '\p{^Is_Soft_Dotted= -T}', "");
Expect(0, 120467, '\P{Is_Soft_Dotted= -T}', "");
Expect(1, 120467, '\P{^Is_Soft_Dotted= -T}', "");
Expect(0, 120468, '\p{Is_Soft_Dotted= -T}', "");
Expect(1, 120468, '\p{^Is_Soft_Dotted= -T}', "");
Expect(1, 120468, '\P{Is_Soft_Dotted= -T}', "");
Expect(0, 120468, '\P{^Is_Soft_Dotted= -T}', "");
Error('\p{Is_SD=/a/True}');
Error('\P{Is_SD=/a/True}');
Expect(1, 120467, '\p{Is_SD:	true}', "");
Expect(0, 120467, '\p{^Is_SD:	true}', "");
Expect(0, 120467, '\P{Is_SD:	true}', "");
Expect(1, 120467, '\P{^Is_SD:	true}', "");
Expect(0, 120468, '\p{Is_SD:	true}', "");
Expect(1, 120468, '\p{^Is_SD:	true}', "");
Expect(1, 120468, '\P{Is_SD:	true}', "");
Expect(0, 120468, '\P{^Is_SD:	true}', "");
Expect(1, 120467, '\p{Is_SD=_true}', "");
Expect(0, 120467, '\p{^Is_SD=_true}', "");
Expect(0, 120467, '\P{Is_SD=_true}', "");
Expect(1, 120467, '\P{^Is_SD=_true}', "");
Expect(0, 120468, '\p{Is_SD=_true}', "");
Expect(1, 120468, '\p{^Is_SD=_true}', "");
Expect(1, 120468, '\P{Is_SD=_true}', "");
Expect(0, 120468, '\P{^Is_SD=_true}', "");
Error('\p{Soft_Dotted=-:=No}');
Error('\P{Soft_Dotted=-:=No}');
Expect(1, 120468, '\p{Soft_Dotted=no}', "");
Expect(0, 120468, '\p{^Soft_Dotted=no}', "");
Expect(0, 120468, '\P{Soft_Dotted=no}', "");
Expect(1, 120468, '\P{^Soft_Dotted=no}', "");
Expect(0, 120467, '\p{Soft_Dotted=no}', "");
Expect(1, 120467, '\p{^Soft_Dotted=no}', "");
Expect(1, 120467, '\P{Soft_Dotted=no}', "");
Expect(0, 120467, '\P{^Soft_Dotted=no}', "");
Expect(1, 120468, '\p{Soft_Dotted= 	No}', "");
Expect(0, 120468, '\p{^Soft_Dotted= 	No}', "");
Expect(0, 120468, '\P{Soft_Dotted= 	No}', "");
Expect(1, 120468, '\P{^Soft_Dotted= 	No}', "");
Expect(0, 120467, '\p{Soft_Dotted= 	No}', "");
Expect(1, 120467, '\p{^Soft_Dotted= 	No}', "");
Expect(1, 120467, '\P{Soft_Dotted= 	No}', "");
Expect(0, 120467, '\P{^Soft_Dotted= 	No}', "");
Error('\p{SD=  N:=}');
Error('\P{SD=  N:=}');
Expect(1, 120468, '\p{SD=n}', "");
Expect(0, 120468, '\p{^SD=n}', "");
Expect(0, 120468, '\P{SD=n}', "");
Expect(1, 120468, '\P{^SD=n}', "");
Expect(0, 120467, '\p{SD=n}', "");
Expect(1, 120467, '\p{^SD=n}', "");
Expect(1, 120467, '\P{SD=n}', "");
Expect(0, 120467, '\P{^SD=n}', "");
Expect(1, 120468, '\p{SD:_N}', "");
Expect(0, 120468, '\p{^SD:_N}', "");
Expect(0, 120468, '\P{SD:_N}', "");
Expect(1, 120468, '\P{^SD:_N}', "");
Expect(0, 120467, '\p{SD:_N}', "");
Expect(1, 120467, '\p{^SD:_N}', "");
Expect(1, 120467, '\P{SD:_N}', "");
Expect(0, 120467, '\P{^SD:_N}', "");
Error('\p{Is_Soft_Dotted= :=F}');
Error('\P{Is_Soft_Dotted= :=F}');
Expect(1, 120468, '\p{Is_Soft_Dotted=f}', "");
Expect(0, 120468, '\p{^Is_Soft_Dotted=f}', "");
Expect(0, 120468, '\P{Is_Soft_Dotted=f}', "");
Expect(1, 120468, '\P{^Is_Soft_Dotted=f}', "");
Expect(0, 120467, '\p{Is_Soft_Dotted=f}', "");
Expect(1, 120467, '\p{^Is_Soft_Dotted=f}', "");
Expect(1, 120467, '\P{Is_Soft_Dotted=f}', "");
Expect(0, 120467, '\P{^Is_Soft_Dotted=f}', "");
Expect(1, 120468, '\p{Is_Soft_Dotted=-F}', "");
Expect(0, 120468, '\p{^Is_Soft_Dotted=-F}', "");
Expect(0, 120468, '\P{Is_Soft_Dotted=-F}', "");
Expect(1, 120468, '\P{^Is_Soft_Dotted=-F}', "");
Expect(0, 120467, '\p{Is_Soft_Dotted=-F}', "");
Expect(1, 120467, '\p{^Is_Soft_Dotted=-F}', "");
Expect(1, 120467, '\P{Is_Soft_Dotted=-F}', "");
Expect(0, 120467, '\P{^Is_Soft_Dotted=-F}', "");
Error('\p{Is_SD=	:=false}');
Error('\P{Is_SD=	:=false}');
Expect(1, 120468, '\p{Is_SD=false}', "");
Expect(0, 120468, '\p{^Is_SD=false}', "");
Expect(0, 120468, '\P{Is_SD=false}', "");
Expect(1, 120468, '\P{^Is_SD=false}', "");
Expect(0, 120467, '\p{Is_SD=false}', "");
Expect(1, 120467, '\p{^Is_SD=false}', "");
Expect(1, 120467, '\P{Is_SD=false}', "");
Expect(0, 120467, '\P{^Is_SD=false}', "");
Expect(1, 120468, '\p{Is_SD=_False}', "");
Expect(0, 120468, '\p{^Is_SD=_False}', "");
Expect(0, 120468, '\P{Is_SD=_False}', "");
Expect(1, 120468, '\P{^Is_SD=_False}', "");
Expect(0, 120467, '\p{Is_SD=_False}', "");
Expect(1, 120467, '\p{^Is_SD=_False}', "");
Expect(1, 120467, '\P{Is_SD=_False}', "");
Expect(0, 120467, '\P{^Is_SD=_False}', "");
Error('\p{Uppercase=/a/	 Yes}');
Error('\P{Uppercase=/a/	 Yes}');
Expect(1, 120778, '\p{Uppercase=yes}', "");
Expect(0, 120778, '\p{^Uppercase=yes}', "");
Expect(0, 120778, '\P{Uppercase=yes}', "");
Expect(1, 120778, '\P{^Uppercase=yes}', "");
Expect(0, 120779, '\p{Uppercase=yes}', "");
Expect(1, 120779, '\p{^Uppercase=yes}', "");
Expect(1, 120779, '\P{Uppercase=yes}', "");
Expect(0, 120779, '\P{^Uppercase=yes}', "");
Expect(1, 120778, '\p{Uppercase=-_YES}', "");
Expect(0, 120778, '\p{^Uppercase=-_YES}', "");
Expect(0, 120778, '\P{Uppercase=-_YES}', "");
Expect(1, 120778, '\P{^Uppercase=-_YES}', "");
Expect(0, 120779, '\p{Uppercase=-_YES}', "");
Expect(1, 120779, '\p{^Uppercase=-_YES}', "");
Expect(1, 120779, '\P{Uppercase=-_YES}', "");
Expect(0, 120779, '\P{^Uppercase=-_YES}', "");
Error('\p{Upper=:=Y}');
Error('\P{Upper=:=Y}');
Expect(1, 120778, '\p{Upper=y}', "");
Expect(0, 120778, '\p{^Upper=y}', "");
Expect(0, 120778, '\P{Upper=y}', "");
Expect(1, 120778, '\P{^Upper=y}', "");
Expect(0, 120779, '\p{Upper=y}', "");
Expect(1, 120779, '\p{^Upper=y}', "");
Expect(1, 120779, '\P{Upper=y}', "");
Expect(0, 120779, '\P{^Upper=y}', "");
Error('\p{Is_Uppercase=:=		T}');
Error('\P{Is_Uppercase=:=		T}');
Expect(1, 120778, '\p{Is_Uppercase:   t}', "");
Expect(0, 120778, '\p{^Is_Uppercase:   t}', "");
Expect(0, 120778, '\P{Is_Uppercase:   t}', "");
Expect(1, 120778, '\P{^Is_Uppercase:   t}', "");
Expect(0, 120779, '\p{Is_Uppercase:   t}', "");
Expect(1, 120779, '\p{^Is_Uppercase:   t}', "");
Expect(1, 120779, '\P{Is_Uppercase:   t}', "");
Expect(0, 120779, '\P{^Is_Uppercase:   t}', "");
Expect(1, 120778, '\p{Is_Uppercase=_	T}', "");
Expect(0, 120778, '\p{^Is_Uppercase=_	T}', "");
Expect(0, 120778, '\P{Is_Uppercase=_	T}', "");
Expect(1, 120778, '\P{^Is_Uppercase=_	T}', "");
Expect(0, 120779, '\p{Is_Uppercase=_	T}', "");
Expect(1, 120779, '\p{^Is_Uppercase=_	T}', "");
Expect(1, 120779, '\P{Is_Uppercase=_	T}', "");
Expect(0, 120779, '\P{^Is_Uppercase=_	T}', "");
Error('\p{Is_Upper=_/a/True}');
Error('\P{Is_Upper=_/a/True}');
Expect(1, 120778, '\p{Is_Upper=true}', "");
Expect(0, 120778, '\p{^Is_Upper=true}', "");
Expect(0, 120778, '\P{Is_Upper=true}', "");
Expect(1, 120778, '\P{^Is_Upper=true}', "");
Expect(0, 120779, '\p{Is_Upper=true}', "");
Expect(1, 120779, '\p{^Is_Upper=true}', "");
Expect(1, 120779, '\P{Is_Upper=true}', "");
Expect(0, 120779, '\P{^Is_Upper=true}', "");
Expect(1, 120778, '\p{Is_Upper=	_True}', "");
Expect(0, 120778, '\p{^Is_Upper=	_True}', "");
Expect(0, 120778, '\P{Is_Upper=	_True}', "");
Expect(1, 120778, '\P{^Is_Upper=	_True}', "");
Expect(0, 120779, '\p{Is_Upper=	_True}', "");
Expect(1, 120779, '\p{^Is_Upper=	_True}', "");
Expect(1, 120779, '\P{Is_Upper=	_True}', "");
Expect(0, 120779, '\P{^Is_Upper=	_True}', "");
Error('\p{Uppercase=/a/no}');
Error('\P{Uppercase=/a/no}');
Expect(1, 120779, '\p{Uppercase=no}', "");
Expect(0, 120779, '\p{^Uppercase=no}', "");
Expect(0, 120779, '\P{Uppercase=no}', "");
Expect(1, 120779, '\P{^Uppercase=no}', "");
Expect(0, 120778, '\p{Uppercase=no}', "");
Expect(1, 120778, '\p{^Uppercase=no}', "");
Expect(1, 120778, '\P{Uppercase=no}', "");
Expect(0, 120778, '\P{^Uppercase=no}', "");
Expect(1, 120779, '\p{Uppercase=__NO}', "");
Expect(0, 120779, '\p{^Uppercase=__NO}', "");
Expect(0, 120779, '\P{Uppercase=__NO}', "");
Expect(1, 120779, '\P{^Uppercase=__NO}', "");
Expect(0, 120778, '\p{Uppercase=__NO}', "");
Expect(1, 120778, '\p{^Uppercase=__NO}', "");
Expect(1, 120778, '\P{Uppercase=__NO}', "");
Expect(0, 120778, '\P{^Uppercase=__NO}', "");
Error('\p{Upper=:=N}');
Error('\P{Upper=:=N}');
Expect(1, 120779, '\p{Upper=n}', "");
Expect(0, 120779, '\p{^Upper=n}', "");
Expect(0, 120779, '\P{Upper=n}', "");
Expect(1, 120779, '\P{^Upper=n}', "");
Expect(0, 120778, '\p{Upper=n}', "");
Expect(1, 120778, '\p{^Upper=n}', "");
Expect(1, 120778, '\P{Upper=n}', "");
Expect(0, 120778, '\P{^Upper=n}', "");
Expect(1, 120779, '\p{Upper= -N}', "");
Expect(0, 120779, '\p{^Upper= -N}', "");
Expect(0, 120779, '\P{Upper= -N}', "");
Expect(1, 120779, '\P{^Upper= -N}', "");
Expect(0, 120778, '\p{Upper= -N}', "");
Expect(1, 120778, '\p{^Upper= -N}', "");
Expect(1, 120778, '\P{Upper= -N}', "");
Expect(0, 120778, '\P{^Upper= -N}', "");
Error('\p{Is_Uppercase=:=F}');
Error('\P{Is_Uppercase=:=F}');
Expect(1, 120779, '\p{Is_Uppercase:	f}', "");
Expect(0, 120779, '\p{^Is_Uppercase:	f}', "");
Expect(0, 120779, '\P{Is_Uppercase:	f}', "");
Expect(1, 120779, '\P{^Is_Uppercase:	f}', "");
Expect(0, 120778, '\p{Is_Uppercase:	f}', "");
Expect(1, 120778, '\p{^Is_Uppercase:	f}', "");
Expect(1, 120778, '\P{Is_Uppercase:	f}', "");
Expect(0, 120778, '\P{^Is_Uppercase:	f}', "");
Expect(1, 120779, '\p{Is_Uppercase:     F}', "");
Expect(0, 120779, '\p{^Is_Uppercase:     F}', "");
Expect(0, 120779, '\P{Is_Uppercase:     F}', "");
Expect(1, 120779, '\P{^Is_Uppercase:     F}', "");
Expect(0, 120778, '\p{Is_Uppercase:     F}', "");
Expect(1, 120778, '\p{^Is_Uppercase:     F}', "");
Expect(1, 120778, '\P{Is_Uppercase:     F}', "");
Expect(0, 120778, '\P{^Is_Uppercase:     F}', "");
Error('\p{Is_Upper=:=False}');
Error('\P{Is_Upper=:=False}');
Expect(1, 120779, '\p{Is_Upper=false}', "");
Expect(0, 120779, '\p{^Is_Upper=false}', "");
Expect(0, 120779, '\P{Is_Upper=false}', "");
Expect(1, 120779, '\P{^Is_Upper=false}', "");
Expect(0, 120778, '\p{Is_Upper=false}', "");
Expect(1, 120778, '\p{^Is_Upper=false}', "");
Expect(1, 120778, '\P{Is_Upper=false}', "");
Expect(0, 120778, '\P{^Is_Upper=false}', "");
Expect(1, 120779, '\p{Is_Upper= 	False}', "");
Expect(0, 120779, '\p{^Is_Upper= 	False}', "");
Expect(0, 120779, '\P{Is_Upper= 	False}', "");
Expect(1, 120779, '\P{^Is_Upper= 	False}', "");
Expect(0, 120778, '\p{Is_Upper= 	False}', "");
Expect(1, 120778, '\p{^Is_Upper= 	False}', "");
Expect(1, 120778, '\P{Is_Upper= 	False}', "");
Expect(0, 120778, '\P{^Is_Upper= 	False}', "");
Error('\p{Changes_When_Casefolded=:=_-Yes}');
Error('\P{Changes_When_Casefolded=:=_-Yes}');
Expect(1, 66599, '\p{Changes_When_Casefolded=yes}', "");
Expect(0, 66599, '\p{^Changes_When_Casefolded=yes}', "");
Expect(0, 66599, '\P{Changes_When_Casefolded=yes}', "");
Expect(1, 66599, '\P{^Changes_When_Casefolded=yes}', "");
Expect(0, 66600, '\p{Changes_When_Casefolded=yes}', "");
Expect(1, 66600, '\p{^Changes_When_Casefolded=yes}', "");
Expect(1, 66600, '\P{Changes_When_Casefolded=yes}', "");
Expect(0, 66600, '\P{^Changes_When_Casefolded=yes}', "");
Expect(1, 66599, '\p{Changes_When_Casefolded=__Yes}', "");
Expect(0, 66599, '\p{^Changes_When_Casefolded=__Yes}', "");
Expect(0, 66599, '\P{Changes_When_Casefolded=__Yes}', "");
Expect(1, 66599, '\P{^Changes_When_Casefolded=__Yes}', "");
Expect(0, 66600, '\p{Changes_When_Casefolded=__Yes}', "");
Expect(1, 66600, '\p{^Changes_When_Casefolded=__Yes}', "");
Expect(1, 66600, '\P{Changes_When_Casefolded=__Yes}', "");
Expect(0, 66600, '\P{^Changes_When_Casefolded=__Yes}', "");
Error('\p{CWCF= Y/a/}');
Error('\P{CWCF= Y/a/}');
Expect(1, 66599, '\p{CWCF=y}', "");
Expect(0, 66599, '\p{^CWCF=y}', "");
Expect(0, 66599, '\P{CWCF=y}', "");
Expect(1, 66599, '\P{^CWCF=y}', "");
Expect(0, 66600, '\p{CWCF=y}', "");
Expect(1, 66600, '\p{^CWCF=y}', "");
Expect(1, 66600, '\P{CWCF=y}', "");
Expect(0, 66600, '\P{^CWCF=y}', "");
Expect(1, 66599, '\p{CWCF=--Y}', "");
Expect(0, 66599, '\p{^CWCF=--Y}', "");
Expect(0, 66599, '\P{CWCF=--Y}', "");
Expect(1, 66599, '\P{^CWCF=--Y}', "");
Expect(0, 66600, '\p{CWCF=--Y}', "");
Expect(1, 66600, '\p{^CWCF=--Y}', "");
Expect(1, 66600, '\P{CWCF=--Y}', "");
Expect(0, 66600, '\P{^CWCF=--Y}', "");
Error('\p{Is_Changes_When_Casefolded= :=T}');
Error('\P{Is_Changes_When_Casefolded= :=T}');
Expect(1, 66599, '\p{Is_Changes_When_Casefolded=t}', "");
Expect(0, 66599, '\p{^Is_Changes_When_Casefolded=t}', "");
Expect(0, 66599, '\P{Is_Changes_When_Casefolded=t}', "");
Expect(1, 66599, '\P{^Is_Changes_When_Casefolded=t}', "");
Expect(0, 66600, '\p{Is_Changes_When_Casefolded=t}', "");
Expect(1, 66600, '\p{^Is_Changes_When_Casefolded=t}', "");
Expect(1, 66600, '\P{Is_Changes_When_Casefolded=t}', "");
Expect(0, 66600, '\P{^Is_Changes_When_Casefolded=t}', "");
Error('\p{Is_CWCF:  /a/true}');
Error('\P{Is_CWCF:  /a/true}');
Expect(1, 66599, '\p{Is_CWCF=true}', "");
Expect(0, 66599, '\p{^Is_CWCF=true}', "");
Expect(0, 66599, '\P{Is_CWCF=true}', "");
Expect(1, 66599, '\P{^Is_CWCF=true}', "");
Expect(0, 66600, '\p{Is_CWCF=true}', "");
Expect(1, 66600, '\p{^Is_CWCF=true}', "");
Expect(1, 66600, '\P{Is_CWCF=true}', "");
Expect(0, 66600, '\P{^Is_CWCF=true}', "");
Expect(1, 66599, '\p{Is_CWCF=_-TRUE}', "");
Expect(0, 66599, '\p{^Is_CWCF=_-TRUE}', "");
Expect(0, 66599, '\P{Is_CWCF=_-TRUE}', "");
Expect(1, 66599, '\P{^Is_CWCF=_-TRUE}', "");
Expect(0, 66600, '\p{Is_CWCF=_-TRUE}', "");
Expect(1, 66600, '\p{^Is_CWCF=_-TRUE}', "");
Expect(1, 66600, '\P{Is_CWCF=_-TRUE}', "");
Expect(0, 66600, '\P{^Is_CWCF=_-TRUE}', "");
Error('\p{Changes_When_Casefolded:	 :=NO}');
Error('\P{Changes_When_Casefolded:	 :=NO}');
Expect(1, 66600, '\p{Changes_When_Casefolded=no}', "");
Expect(0, 66600, '\p{^Changes_When_Casefolded=no}', "");
Expect(0, 66600, '\P{Changes_When_Casefolded=no}', "");
Expect(1, 66600, '\P{^Changes_When_Casefolded=no}', "");
Expect(0, 66599, '\p{Changes_When_Casefolded=no}', "");
Expect(1, 66599, '\p{^Changes_When_Casefolded=no}', "");
Expect(1, 66599, '\P{Changes_When_Casefolded=no}', "");
Expect(0, 66599, '\P{^Changes_When_Casefolded=no}', "");
Expect(1, 66600, '\p{Changes_When_Casefolded=  NO}', "");
Expect(0, 66600, '\p{^Changes_When_Casefolded=  NO}', "");
Expect(0, 66600, '\P{Changes_When_Casefolded=  NO}', "");
Expect(1, 66600, '\P{^Changes_When_Casefolded=  NO}', "");
Expect(0, 66599, '\p{Changes_When_Casefolded=  NO}', "");
Expect(1, 66599, '\p{^Changes_When_Casefolded=  NO}', "");
Expect(1, 66599, '\P{Changes_When_Casefolded=  NO}', "");
Expect(0, 66599, '\P{^Changes_When_Casefolded=  NO}', "");
Error('\p{CWCF=_	n:=}');
Error('\P{CWCF=_	n:=}');
Expect(1, 66600, '\p{CWCF:   n}', "");
Expect(0, 66600, '\p{^CWCF:   n}', "");
Expect(0, 66600, '\P{CWCF:   n}', "");
Expect(1, 66600, '\P{^CWCF:   n}', "");
Expect(0, 66599, '\p{CWCF:   n}', "");
Expect(1, 66599, '\p{^CWCF:   n}', "");
Expect(1, 66599, '\P{CWCF:   n}', "");
Expect(0, 66599, '\P{^CWCF:   n}', "");
Expect(1, 66600, '\p{CWCF=_ N}', "");
Expect(0, 66600, '\p{^CWCF=_ N}', "");
Expect(0, 66600, '\P{CWCF=_ N}', "");
Expect(1, 66600, '\P{^CWCF=_ N}', "");
Expect(0, 66599, '\p{CWCF=_ N}', "");
Expect(1, 66599, '\p{^CWCF=_ N}', "");
Expect(1, 66599, '\P{CWCF=_ N}', "");
Expect(0, 66599, '\P{^CWCF=_ N}', "");
Error('\p{Is_Changes_When_Casefolded=	/a/F}');
Error('\P{Is_Changes_When_Casefolded=	/a/F}');
Expect(1, 66600, '\p{Is_Changes_When_Casefolded=f}', "");
Expect(0, 66600, '\p{^Is_Changes_When_Casefolded=f}', "");
Expect(0, 66600, '\P{Is_Changes_When_Casefolded=f}', "");
Expect(1, 66600, '\P{^Is_Changes_When_Casefolded=f}', "");
Expect(0, 66599, '\p{Is_Changes_When_Casefolded=f}', "");
Expect(1, 66599, '\p{^Is_Changes_When_Casefolded=f}', "");
Expect(1, 66599, '\P{Is_Changes_When_Casefolded=f}', "");
Expect(0, 66599, '\P{^Is_Changes_When_Casefolded=f}', "");
Expect(1, 66600, '\p{Is_Changes_When_Casefolded=_f}', "");
Expect(0, 66600, '\p{^Is_Changes_When_Casefolded=_f}', "");
Expect(0, 66600, '\P{Is_Changes_When_Casefolded=_f}', "");
Expect(1, 66600, '\P{^Is_Changes_When_Casefolded=_f}', "");
Expect(0, 66599, '\p{Is_Changes_When_Casefolded=_f}', "");
Expect(1, 66599, '\p{^Is_Changes_When_Casefolded=_f}', "");
Expect(1, 66599, '\P{Is_Changes_When_Casefolded=_f}', "");
Expect(0, 66599, '\P{^Is_Changes_When_Casefolded=_f}', "");
Error('\p{Is_CWCF: 	False/a/}');
Error('\P{Is_CWCF: 	False/a/}');
Expect(1, 66600, '\p{Is_CWCF=false}', "");
Expect(0, 66600, '\p{^Is_CWCF=false}', "");
Expect(0, 66600, '\P{Is_CWCF=false}', "");
Expect(1, 66600, '\P{^Is_CWCF=false}', "");
Expect(0, 66599, '\p{Is_CWCF=false}', "");
Expect(1, 66599, '\p{^Is_CWCF=false}', "");
Expect(1, 66599, '\P{Is_CWCF=false}', "");
Expect(0, 66599, '\P{^Is_CWCF=false}', "");
Expect(1, 66600, '\p{Is_CWCF=	-false}', "");
Expect(0, 66600, '\p{^Is_CWCF=	-false}', "");
Expect(0, 66600, '\P{Is_CWCF=	-false}', "");
Expect(1, 66600, '\P{^Is_CWCF=	-false}', "");
Expect(0, 66599, '\p{Is_CWCF=	-false}', "");
Expect(1, 66599, '\p{^Is_CWCF=	-false}', "");
Expect(1, 66599, '\P{Is_CWCF=	-false}', "");
Expect(0, 66599, '\P{^Is_CWCF=	-false}', "");
Error('\p{Cased=	:=YES}');
Error('\P{Cased=	:=YES}');
Expect(1, 120779, '\p{Cased=yes}', "");
Expect(0, 120779, '\p{^Cased=yes}', "");
Expect(0, 120779, '\P{Cased=yes}', "");
Expect(1, 120779, '\P{^Cased=yes}', "");
Expect(0, 120780, '\p{Cased=yes}', "");
Expect(1, 120780, '\p{^Cased=yes}', "");
Expect(1, 120780, '\P{Cased=yes}', "");
Expect(0, 120780, '\P{^Cased=yes}', "");
Expect(1, 120779, '\p{Cased= _YES}', "");
Expect(0, 120779, '\p{^Cased= _YES}', "");
Expect(0, 120779, '\P{Cased= _YES}', "");
Expect(1, 120779, '\P{^Cased= _YES}', "");
Expect(0, 120780, '\p{Cased= _YES}', "");
Expect(1, 120780, '\p{^Cased= _YES}', "");
Expect(1, 120780, '\P{Cased= _YES}', "");
Expect(0, 120780, '\P{^Cased= _YES}', "");
Error('\p{Is_Cased=_	Y:=}');
Error('\P{Is_Cased=_	Y:=}');
Expect(1, 120779, '\p{Is_Cased=y}', "");
Expect(0, 120779, '\p{^Is_Cased=y}', "");
Expect(0, 120779, '\P{Is_Cased=y}', "");
Expect(1, 120779, '\P{^Is_Cased=y}', "");
Expect(0, 120780, '\p{Is_Cased=y}', "");
Expect(1, 120780, '\p{^Is_Cased=y}', "");
Expect(1, 120780, '\P{Is_Cased=y}', "");
Expect(0, 120780, '\P{^Is_Cased=y}', "");
Expect(1, 120779, '\p{Is_Cased: 	Y}', "");
Expect(0, 120779, '\p{^Is_Cased: 	Y}', "");
Expect(0, 120779, '\P{Is_Cased: 	Y}', "");
Expect(1, 120779, '\P{^Is_Cased: 	Y}', "");
Expect(0, 120780, '\p{Is_Cased: 	Y}', "");
Expect(1, 120780, '\p{^Is_Cased: 	Y}', "");
Expect(1, 120780, '\P{Is_Cased: 	Y}', "");
Expect(0, 120780, '\P{^Is_Cased: 	Y}', "");
Error('\p{Cased=:=-_T}');
Error('\P{Cased=:=-_T}');
Expect(1, 120779, '\p{Cased=t}', "");
Expect(0, 120779, '\p{^Cased=t}', "");
Expect(0, 120779, '\P{Cased=t}', "");
Expect(1, 120779, '\P{^Cased=t}', "");
Expect(0, 120780, '\p{Cased=t}', "");
Expect(1, 120780, '\p{^Cased=t}', "");
Expect(1, 120780, '\P{Cased=t}', "");
Expect(0, 120780, '\P{^Cased=t}', "");
Expect(1, 120779, '\p{Cased= T}', "");
Expect(0, 120779, '\p{^Cased= T}', "");
Expect(0, 120779, '\P{Cased= T}', "");
Expect(1, 120779, '\P{^Cased= T}', "");
Expect(0, 120780, '\p{Cased= T}', "");
Expect(1, 120780, '\p{^Cased= T}', "");
Expect(1, 120780, '\P{Cased= T}', "");
Expect(0, 120780, '\P{^Cased= T}', "");
Error('\p{Is_Cased= /a/True}');
Error('\P{Is_Cased= /a/True}');
Expect(1, 120779, '\p{Is_Cased:true}', "");
Expect(0, 120779, '\p{^Is_Cased:true}', "");
Expect(0, 120779, '\P{Is_Cased:true}', "");
Expect(1, 120779, '\P{^Is_Cased:true}', "");
Expect(0, 120780, '\p{Is_Cased:true}', "");
Expect(1, 120780, '\p{^Is_Cased:true}', "");
Expect(1, 120780, '\P{Is_Cased:true}', "");
Expect(0, 120780, '\P{^Is_Cased:true}', "");
Expect(1, 120779, '\p{Is_Cased=	TRUE}', "");
Expect(0, 120779, '\p{^Is_Cased=	TRUE}', "");
Expect(0, 120779, '\P{Is_Cased=	TRUE}', "");
Expect(1, 120779, '\P{^Is_Cased=	TRUE}', "");
Expect(0, 120780, '\p{Is_Cased=	TRUE}', "");
Expect(1, 120780, '\p{^Is_Cased=	TRUE}', "");
Expect(1, 120780, '\P{Is_Cased=	TRUE}', "");
Expect(0, 120780, '\P{^Is_Cased=	TRUE}', "");
Error('\p{Cased=	:=No}');
Error('\P{Cased=	:=No}');
Expect(1, 120780, '\p{Cased=no}', "");
Expect(0, 120780, '\p{^Cased=no}', "");
Expect(0, 120780, '\P{Cased=no}', "");
Expect(1, 120780, '\P{^Cased=no}', "");
Expect(0, 120779, '\p{Cased=no}', "");
Expect(1, 120779, '\p{^Cased=no}', "");
Expect(1, 120779, '\P{Cased=no}', "");
Expect(0, 120779, '\P{^Cased=no}', "");
Expect(1, 120780, '\p{Cased=		NO}', "");
Expect(0, 120780, '\p{^Cased=		NO}', "");
Expect(0, 120780, '\P{Cased=		NO}', "");
Expect(1, 120780, '\P{^Cased=		NO}', "");
Expect(0, 120779, '\p{Cased=		NO}', "");
Expect(1, 120779, '\p{^Cased=		NO}', "");
Expect(1, 120779, '\P{Cased=		NO}', "");
Expect(0, 120779, '\P{^Cased=		NO}', "");
Error('\p{Is_Cased=	 n/a/}');
Error('\P{Is_Cased=	 n/a/}');
Expect(1, 120780, '\p{Is_Cased=n}', "");
Expect(0, 120780, '\p{^Is_Cased=n}', "");
Expect(0, 120780, '\P{Is_Cased=n}', "");
Expect(1, 120780, '\P{^Is_Cased=n}', "");
Expect(0, 120779, '\p{Is_Cased=n}', "");
Expect(1, 120779, '\p{^Is_Cased=n}', "");
Expect(1, 120779, '\P{Is_Cased=n}', "");
Expect(0, 120779, '\P{^Is_Cased=n}', "");
Expect(1, 120780, '\p{Is_Cased=	-N}', "");
Expect(0, 120780, '\p{^Is_Cased=	-N}', "");
Expect(0, 120780, '\P{Is_Cased=	-N}', "");
Expect(1, 120780, '\P{^Is_Cased=	-N}', "");
Expect(0, 120779, '\p{Is_Cased=	-N}', "");
Expect(1, 120779, '\p{^Is_Cased=	-N}', "");
Expect(1, 120779, '\P{Is_Cased=	-N}', "");
Expect(0, 120779, '\P{^Is_Cased=	-N}', "");
Error('\p{Cased=/a/	-F}');
Error('\P{Cased=/a/	-F}');
Expect(1, 120780, '\p{Cased: f}', "");
Expect(0, 120780, '\p{^Cased: f}', "");
Expect(0, 120780, '\P{Cased: f}', "");
Expect(1, 120780, '\P{^Cased: f}', "");
Expect(0, 120779, '\p{Cased: f}', "");
Expect(1, 120779, '\p{^Cased: f}', "");
Expect(1, 120779, '\P{Cased: f}', "");
Expect(0, 120779, '\P{^Cased: f}', "");
Expect(1, 120780, '\p{Cased=_f}', "");
Expect(0, 120780, '\p{^Cased=_f}', "");
Expect(0, 120780, '\P{Cased=_f}', "");
Expect(1, 120780, '\P{^Cased=_f}', "");
Expect(0, 120779, '\p{Cased=_f}', "");
Expect(1, 120779, '\p{^Cased=_f}', "");
Expect(1, 120779, '\P{Cased=_f}', "");
Expect(0, 120779, '\P{^Cased=_f}', "");
Error('\p{Is_Cased=/a/_	FALSE}');
Error('\P{Is_Cased=/a/_	FALSE}');
Expect(1, 120780, '\p{Is_Cased=false}', "");
Expect(0, 120780, '\p{^Is_Cased=false}', "");
Expect(0, 120780, '\P{Is_Cased=false}', "");
Expect(1, 120780, '\P{^Is_Cased=false}', "");
Expect(0, 120779, '\p{Is_Cased=false}', "");
Expect(1, 120779, '\p{^Is_Cased=false}', "");
Expect(1, 120779, '\P{Is_Cased=false}', "");
Expect(0, 120779, '\P{^Is_Cased=false}', "");
Expect(1, 120780, '\p{Is_Cased=	 False}', "");
Expect(0, 120780, '\p{^Is_Cased=	 False}', "");
Expect(0, 120780, '\P{Is_Cased=	 False}', "");
Expect(1, 120780, '\P{^Is_Cased=	 False}', "");
Expect(0, 120779, '\p{Is_Cased=	 False}', "");
Expect(1, 120779, '\p{^Is_Cased=	 False}', "");
Expect(1, 120779, '\P{Is_Cased=	 False}', "");
Expect(0, 120779, '\P{^Is_Cased=	 False}', "");
Error('\p{Age: := _V4_1}');
Error('\P{Age: := _V4_1}');
Expect(1, 120485, '\p{Age=v41}', "");
Expect(0, 120485, '\p{^Age=v41}', "");
Expect(0, 120485, '\P{Age=v41}', "");
Expect(1, 120485, '\P{^Age=v41}', "");
Expect(0, 120486, '\p{Age=v41}', "");
Expect(1, 120486, '\p{^Age=v41}', "");
Expect(1, 120486, '\P{Age=v41}', "");
Expect(0, 120486, '\P{^Age=v41}', "");
Expect(1, 120485, '\p{Age=_	v4_1}', "");
Expect(0, 120485, '\p{^Age=_	v4_1}', "");
Expect(0, 120485, '\P{Age=_	v4_1}', "");
Expect(1, 120485, '\P{^Age=_	v4_1}', "");
Expect(0, 120486, '\p{Age=_	v4_1}', "");
Expect(1, 120486, '\p{^Age=_	v4_1}', "");
Expect(1, 120486, '\P{Age=_	v4_1}', "");
Expect(0, 120486, '\P{^Age=_	v4_1}', "");
Error('\p{Is_Age=/a/_-0_0_0_0_0_0_0_0_04.1}');
Error('\P{Is_Age=/a/_-0_0_0_0_0_0_0_0_04.1}');
Expect(1, 120485, '\p{Is_Age=0_0_04.1}', "");
Expect(0, 120485, '\p{^Is_Age=0_0_04.1}', "");
Expect(0, 120485, '\P{Is_Age=0_0_04.1}', "");
Expect(1, 120485, '\P{^Is_Age=0_0_04.1}', "");
Expect(0, 120486, '\p{Is_Age=0_0_04.1}', "");
Expect(1, 120486, '\p{^Is_Age=0_0_04.1}', "");
Expect(1, 120486, '\P{Is_Age=0_0_04.1}', "");
Expect(0, 120486, '\P{^Is_Age=0_0_04.1}', "");
Error('\p{Age= :=v5_0}');
Error('\P{Age= :=v5_0}');
Expect(1, 120779, '\p{Age=v50}', "");
Expect(0, 120779, '\p{^Age=v50}', "");
Expect(0, 120779, '\P{Age=v50}', "");
Expect(1, 120779, '\P{^Age=v50}', "");
Expect(0, 120780, '\p{Age=v50}', "");
Expect(1, 120780, '\p{^Age=v50}', "");
Expect(1, 120780, '\P{Age=v50}', "");
Expect(0, 120780, '\P{^Age=v50}', "");
Expect(1, 120779, '\p{Age= v5_0}', "");
Expect(0, 120779, '\p{^Age= v5_0}', "");
Expect(0, 120779, '\P{Age= v5_0}', "");
Expect(1, 120779, '\P{^Age= v5_0}', "");
Expect(0, 120780, '\p{Age= v5_0}', "");
Expect(1, 120780, '\p{^Age= v5_0}', "");
Expect(1, 120780, '\P{Age= v5_0}', "");
Expect(0, 120780, '\P{^Age= v5_0}', "");
Error('\p{Is_Age= 00000000_5.0:=}');
Error('\P{Is_Age= 00000000_5.0:=}');
Expect(1, 120779, '\p{Is_Age=0000005.0}', "");
Expect(0, 120779, '\p{^Is_Age=0000005.0}', "");
Expect(0, 120779, '\P{Is_Age=0000005.0}', "");
Expect(1, 120779, '\P{^Is_Age=0000005.0}', "");
Expect(0, 120780, '\p{Is_Age=0000005.0}', "");
Expect(1, 120780, '\p{^Is_Age=0000005.0}', "");
Expect(1, 120780, '\P{Is_Age=0000005.0}', "");
Expect(0, 120780, '\P{^Is_Age=0000005.0}', "");
Error('\p{Age=/a/ -V3_0}');
Error('\P{Age=/a/ -V3_0}');
Expect(1, 65531, '\p{Age:	v30}', "");
Expect(0, 65531, '\p{^Age:	v30}', "");
Expect(0, 65531, '\P{Age:	v30}', "");
Expect(1, 65531, '\P{^Age:	v30}', "");
Expect(0, 65532, '\p{Age:	v30}', "");
Expect(1, 65532, '\p{^Age:	v30}', "");
Expect(1, 65532, '\P{Age:	v30}', "");
Expect(0, 65532, '\P{^Age:	v30}', "");
Expect(1, 65531, '\p{Age=_	V3_0}', "");
Expect(0, 65531, '\p{^Age=_	V3_0}', "");
Expect(0, 65531, '\P{Age=_	V3_0}', "");
Expect(1, 65531, '\P{^Age=_	V3_0}', "");
Expect(0, 65532, '\p{Age=_	V3_0}', "");
Expect(1, 65532, '\p{^Age=_	V3_0}', "");
Expect(1, 65532, '\P{Age=_	V3_0}', "");
Expect(0, 65532, '\P{^Age=_	V3_0}', "");
Error('\p{Is_Age=:=__0_0_0_0_0_00003.0}');
Error('\P{Is_Age=:=__0_0_0_0_0_00003.0}');
Expect(1, 65531, '\p{Is_Age=0_3.0}', "");
Expect(0, 65531, '\p{^Is_Age=0_3.0}', "");
Expect(0, 65531, '\P{Is_Age=0_3.0}', "");
Expect(1, 65531, '\P{^Is_Age=0_3.0}', "");
Expect(0, 65532, '\p{Is_Age=0_3.0}', "");
Expect(1, 65532, '\p{^Is_Age=0_3.0}', "");
Expect(1, 65532, '\P{Is_Age=0_3.0}', "");
Expect(0, 65532, '\P{^Is_Age=0_3.0}', "");
Error('\p{Age=:=V2_1}');
Error('\P{Age=:=V2_1}');
Expect(1, 65532, '\p{Age=v21}', "");
Expect(0, 65532, '\p{^Age=v21}', "");
Expect(0, 65532, '\P{Age=v21}', "");
Expect(1, 65532, '\P{^Age=v21}', "");
Expect(0, 65533, '\p{Age=v21}', "");
Expect(1, 65533, '\p{^Age=v21}', "");
Expect(1, 65533, '\P{Age=v21}', "");
Expect(0, 65533, '\P{^Age=v21}', "");
Expect(1, 65532, '\p{Age= -v2_1}', "");
Expect(0, 65532, '\p{^Age= -v2_1}', "");
Expect(0, 65532, '\P{Age= -v2_1}', "");
Expect(1, 65532, '\P{^Age= -v2_1}', "");
Expect(0, 65533, '\p{Age= -v2_1}', "");
Expect(1, 65533, '\p{^Age= -v2_1}', "");
Expect(1, 65533, '\P{Age= -v2_1}', "");
Expect(0, 65533, '\P{^Age= -v2_1}', "");
Error('\p{Is_Age=-+0_0_02.1/a/}');
Error('\P{Is_Age=-+0_0_02.1/a/}');
Expect(1, 65532, '\p{Is_Age=0002.1}', "");
Expect(0, 65532, '\p{^Is_Age=0002.1}', "");
Expect(0, 65532, '\P{Is_Age=0002.1}', "");
Expect(1, 65532, '\P{^Is_Age=0002.1}', "");
Expect(0, 65533, '\p{Is_Age=0002.1}', "");
Expect(1, 65533, '\p{^Is_Age=0002.1}', "");
Expect(1, 65533, '\P{Is_Age=0002.1}', "");
Expect(0, 65533, '\P{^Is_Age=0002.1}', "");
Error('\p{Age=/a/	 V3_2}');
Error('\P{Age=/a/	 V3_2}');
Expect(1, 65376, '\p{Age:v32}', "");
Expect(0, 65376, '\p{^Age:v32}', "");
Expect(0, 65376, '\P{Age:v32}', "");
Expect(1, 65376, '\P{^Age:v32}', "");
Expect(0, 65377, '\p{Age:v32}', "");
Expect(1, 65377, '\p{^Age:v32}', "");
Expect(1, 65377, '\P{Age:v32}', "");
Expect(0, 65377, '\P{^Age:v32}', "");
Expect(1, 65376, '\p{Age= V3_2}', "");
Expect(0, 65376, '\p{^Age= V3_2}', "");
Expect(0, 65376, '\P{Age= V3_2}', "");
Expect(1, 65376, '\P{^Age= V3_2}', "");
Expect(0, 65377, '\p{Age= V3_2}', "");
Expect(1, 65377, '\p{^Age= V3_2}', "");
Expect(1, 65377, '\P{Age= V3_2}', "");
Expect(0, 65377, '\P{^Age= V3_2}', "");
Error('\p{Is_Age=/a/0_0_03.2}');
Error('\P{Is_Age=/a/0_0_03.2}');
Expect(1, 65376, '\p{Is_Age=000000003.2}', "");
Expect(0, 65376, '\p{^Is_Age=000000003.2}', "");
Expect(0, 65376, '\P{Is_Age=000000003.2}', "");
Expect(1, 65376, '\P{^Is_Age=000000003.2}', "");
Expect(0, 65377, '\p{Is_Age=000000003.2}', "");
Expect(1, 65377, '\p{^Is_Age=000000003.2}', "");
Expect(1, 65377, '\P{Is_Age=000000003.2}', "");
Expect(0, 65377, '\P{^Is_Age=000000003.2}', "");
Error('\p{Age=:=-unassigned}');
Error('\P{Age=:=-unassigned}');
Expect(1, 983037, '\p{Age=unassigned}', "");
Expect(0, 983037, '\p{^Age=unassigned}', "");
Expect(0, 983037, '\P{Age=unassigned}', "");
Expect(1, 983037, '\P{^Age=unassigned}', "");
Expect(0, 983040, '\p{Age=unassigned}', "");
Expect(1, 983040, '\p{^Age=unassigned}', "");
Expect(1, 983040, '\P{Age=unassigned}', "");
Expect(0, 983040, '\P{^Age=unassigned}', "");
Expect(1, 983037, '\p{Age=	 UNASSIGNED}', "");
Expect(0, 983037, '\p{^Age=	 UNASSIGNED}', "");
Expect(0, 983037, '\P{Age=	 UNASSIGNED}', "");
Expect(1, 983037, '\P{^Age=	 UNASSIGNED}', "");
Expect(0, 983040, '\p{Age=	 UNASSIGNED}', "");
Expect(1, 983040, '\p{^Age=	 UNASSIGNED}', "");
Expect(1, 983040, '\P{Age=	 UNASSIGNED}', "");
Expect(0, 983040, '\P{^Age=	 UNASSIGNED}', "");
Error('\p{Is_Age=-:=NA}');
Error('\P{Is_Age=-:=NA}');
Expect(1, 983037, '\p{Is_Age=na}', "");
Expect(0, 983037, '\p{^Is_Age=na}', "");
Expect(0, 983037, '\P{Is_Age=na}', "");
Expect(1, 983037, '\P{^Is_Age=na}', "");
Expect(0, 983040, '\p{Is_Age=na}', "");
Expect(1, 983040, '\p{^Is_Age=na}', "");
Expect(1, 983040, '\P{Is_Age=na}', "");
Expect(0, 983040, '\P{^Is_Age=na}', "");
Expect(1, 983037, '\p{Is_Age=	NA}', "");
Expect(0, 983037, '\p{^Is_Age=	NA}', "");
Expect(0, 983037, '\P{Is_Age=	NA}', "");
Expect(1, 983037, '\P{^Is_Age=	NA}', "");
Expect(0, 983040, '\p{Is_Age=	NA}', "");
Expect(1, 983040, '\p{^Is_Age=	NA}', "");
Expect(1, 983040, '\P{Is_Age=	NA}', "");
Expect(0, 983040, '\P{^Is_Age=	NA}', "");
Error('\p{Age= -V4_0:=}');
Error('\P{Age= -V4_0:=}');
Expect(1, 917999, '\p{Age=v40}', "");
Expect(0, 917999, '\p{^Age=v40}', "");
Expect(0, 917999, '\P{Age=v40}', "");
Expect(1, 917999, '\P{^Age=v40}', "");
Expect(0, 918000, '\p{Age=v40}', "");
Expect(1, 918000, '\p{^Age=v40}', "");
Expect(1, 918000, '\P{Age=v40}', "");
Expect(0, 918000, '\P{^Age=v40}', "");
Expect(1, 917999, '\p{Age=-V4_0}', "");
Expect(0, 917999, '\p{^Age=-V4_0}', "");
Expect(0, 917999, '\P{Age=-V4_0}', "");
Expect(1, 917999, '\P{^Age=-V4_0}', "");
Expect(0, 918000, '\p{Age=-V4_0}', "");
Expect(1, 918000, '\p{^Age=-V4_0}', "");
Expect(1, 918000, '\P{Age=-V4_0}', "");
Expect(0, 918000, '\P{^Age=-V4_0}', "");
Error('\p{Is_Age:   /a/00_4.0}');
Error('\P{Is_Age:   /a/00_4.0}');
Expect(1, 917999, '\p{Is_Age: 000004.0}', "");
Expect(0, 917999, '\p{^Is_Age: 000004.0}', "");
Expect(0, 917999, '\P{Is_Age: 000004.0}', "");
Expect(1, 917999, '\P{^Is_Age: 000004.0}', "");
Expect(0, 918000, '\p{Is_Age: 000004.0}', "");
Expect(1, 918000, '\p{^Is_Age: 000004.0}', "");
Expect(1, 918000, '\P{Is_Age: 000004.0}', "");
Expect(0, 918000, '\P{^Is_Age: 000004.0}', "");
Error('\p{Age=/a/ -V5_2}');
Error('\P{Age=/a/ -V5_2}');
Expect(1, 177972, '\p{Age=v52}', "");
Expect(0, 177972, '\p{^Age=v52}', "");
Expect(0, 177972, '\P{Age=v52}', "");
Expect(1, 177972, '\P{^Age=v52}', "");
Expect(0, 177973, '\p{Age=v52}', "");
Expect(1, 177973, '\p{^Age=v52}', "");
Expect(1, 177973, '\P{Age=v52}', "");
Expect(0, 177973, '\P{^Age=v52}', "");
Expect(1, 177972, '\p{Age=-V5_2}', "");
Expect(0, 177972, '\p{^Age=-V5_2}', "");
Expect(0, 177972, '\P{Age=-V5_2}', "");
Expect(1, 177972, '\P{^Age=-V5_2}', "");
Expect(0, 177973, '\p{Age=-V5_2}', "");
Expect(1, 177973, '\p{^Age=-V5_2}', "");
Expect(1, 177973, '\P{Age=-V5_2}', "");
Expect(0, 177973, '\P{^Age=-V5_2}', "");
Error('\p{Is_Age= 0000000_5.2/a/}');
Error('\P{Is_Age= 0000000_5.2/a/}');
Expect(1, 177972, '\p{Is_Age: +0005.2}', "");
Expect(0, 177972, '\p{^Is_Age: +0005.2}', "");
Expect(0, 177972, '\P{Is_Age: +0005.2}', "");
Expect(1, 177972, '\P{^Is_Age: +0005.2}', "");
Expect(0, 177973, '\p{Is_Age: +0005.2}', "");
Expect(1, 177973, '\p{^Is_Age: +0005.2}', "");
Expect(1, 177973, '\P{Is_Age: +0005.2}', "");
Expect(0, 177973, '\P{^Is_Age: +0005.2}', "");
Error('\p{Age:		/a/V2_0}');
Error('\P{Age:		/a/V2_0}');
Expect(1, 983040, '\p{Age=v20}', "");
Expect(0, 983040, '\p{^Age=v20}', "");
Expect(0, 983040, '\P{Age=v20}', "");
Expect(1, 983040, '\P{^Age=v20}', "");
Expect(0, 983037, '\p{Age=v20}', "");
Expect(1, 983037, '\p{^Age=v20}', "");
Expect(1, 983037, '\P{Age=v20}', "");
Expect(0, 983037, '\P{^Age=v20}', "");
Expect(1, 983040, '\p{Age=		V2_0}', "");
Expect(0, 983040, '\p{^Age=		V2_0}', "");
Expect(0, 983040, '\P{Age=		V2_0}', "");
Expect(1, 983040, '\P{^Age=		V2_0}', "");
Expect(0, 983037, '\p{Age=		V2_0}', "");
Expect(1, 983037, '\p{^Age=		V2_0}', "");
Expect(1, 983037, '\P{Age=		V2_0}', "");
Expect(0, 983037, '\P{^Age=		V2_0}', "");
Error('\p{Is_Age=		0000000002.0/a/}');
Error('\P{Is_Age=		0000000002.0/a/}');
Expect(1, 983040, '\p{Is_Age=+00000_2.0}', "");
Expect(0, 983040, '\p{^Is_Age=+00000_2.0}', "");
Expect(0, 983040, '\P{Is_Age=+00000_2.0}', "");
Expect(1, 983040, '\P{^Is_Age=+00000_2.0}', "");
Expect(0, 983037, '\p{Is_Age=+00000_2.0}', "");
Expect(1, 983037, '\p{^Is_Age=+00000_2.0}', "");
Expect(1, 983037, '\P{Is_Age=+00000_2.0}', "");
Expect(0, 983037, '\P{^Is_Age=+00000_2.0}', "");
Error('\p{Age=/a/ _V6_1}');
Error('\P{Age=/a/ _V6_1}');
Expect(1, 128564, '\p{Age=v61}', "");
Expect(0, 128564, '\p{^Age=v61}', "");
Expect(0, 128564, '\P{Age=v61}', "");
Expect(1, 128564, '\P{^Age=v61}', "");
Expect(0, 128565, '\p{Age=v61}', "");
Expect(1, 128565, '\p{^Age=v61}', "");
Expect(1, 128565, '\P{Age=v61}', "");
Expect(0, 128565, '\P{^Age=v61}', "");
Expect(1, 128564, '\p{Age: - v6_1}', "");
Expect(0, 128564, '\p{^Age: - v6_1}', "");
Expect(0, 128564, '\P{Age: - v6_1}', "");
Expect(1, 128564, '\P{^Age: - v6_1}', "");
Expect(0, 128565, '\p{Age: - v6_1}', "");
Expect(1, 128565, '\p{^Age: - v6_1}', "");
Expect(1, 128565, '\P{Age: - v6_1}', "");
Expect(0, 128565, '\P{^Age: - v6_1}', "");
Error('\p{Is_Age=	:=0_6.1}');
Error('\P{Is_Age=	:=0_6.1}');
Expect(1, 128564, '\p{Is_Age=06.1}', "");
Expect(0, 128564, '\p{^Is_Age=06.1}', "");
Expect(0, 128564, '\P{Is_Age=06.1}', "");
Expect(1, 128564, '\P{^Is_Age=06.1}', "");
Expect(0, 128565, '\p{Is_Age=06.1}', "");
Expect(1, 128565, '\p{^Is_Age=06.1}', "");
Expect(1, 128565, '\P{Is_Age=06.1}', "");
Expect(0, 128565, '\P{^Is_Age=06.1}', "");
Error('\p{Age: _	V6_0/a/}');
Error('\P{Age: _	V6_0/a/}');
Expect(1, 178205, '\p{Age=v60}', "");
Expect(0, 178205, '\p{^Age=v60}', "");
Expect(0, 178205, '\P{Age=v60}', "");
Expect(1, 178205, '\P{^Age=v60}', "");
Expect(0, 178206, '\p{Age=v60}', "");
Expect(1, 178206, '\p{^Age=v60}', "");
Expect(1, 178206, '\P{Age=v60}', "");
Expect(0, 178206, '\P{^Age=v60}', "");
Expect(1, 178205, '\p{Age= _V6_0}', "");
Expect(0, 178205, '\p{^Age= _V6_0}', "");
Expect(0, 178205, '\P{Age= _V6_0}', "");
Expect(1, 178205, '\P{^Age= _V6_0}', "");
Expect(0, 178206, '\p{Age= _V6_0}', "");
Expect(1, 178206, '\p{^Age= _V6_0}', "");
Expect(1, 178206, '\P{Age= _V6_0}', "");
Expect(0, 178206, '\P{^Age= _V6_0}', "");
Error('\p{Is_Age=-/a/00_6.0}');
Error('\P{Is_Age=-/a/00_6.0}');
Expect(1, 178205, '\p{Is_Age=+00_00_00_00_6.0}', "");
Expect(0, 178205, '\p{^Is_Age=+00_00_00_00_6.0}', "");
Expect(0, 178205, '\P{Is_Age=+00_00_00_00_6.0}', "");
Expect(1, 178205, '\P{^Is_Age=+00_00_00_00_6.0}', "");
Expect(0, 178206, '\p{Is_Age=+00_00_00_00_6.0}', "");
Expect(1, 178206, '\p{^Is_Age=+00_00_00_00_6.0}', "");
Expect(1, 178206, '\P{Is_Age=+00_00_00_00_6.0}', "");
Expect(0, 178206, '\P{^Is_Age=+00_00_00_00_6.0}', "");
Error('\p{Age= :=V1_1}');
Error('\P{Age= :=V1_1}');
Expect(1, 65533, '\p{Age=v11}', "");
Expect(0, 65533, '\p{^Age=v11}', "");
Expect(0, 65533, '\P{Age=v11}', "");
Expect(1, 65533, '\P{^Age=v11}', "");
Expect(0, 65536, '\p{Age=v11}', "");
Expect(1, 65536, '\p{^Age=v11}', "");
Expect(1, 65536, '\P{Age=v11}', "");
Expect(0, 65536, '\P{^Age=v11}', "");
Expect(1, 65533, '\p{Age=  V1_1}', "");
Expect(0, 65533, '\p{^Age=  V1_1}', "");
Expect(0, 65533, '\P{Age=  V1_1}', "");
Expect(1, 65533, '\P{^Age=  V1_1}', "");
Expect(0, 65536, '\p{Age=  V1_1}', "");
Expect(1, 65536, '\p{^Age=  V1_1}', "");
Expect(1, 65536, '\P{Age=  V1_1}', "");
Expect(0, 65536, '\P{^Age=  V1_1}', "");
Error('\p{Is_Age=	 +001.1/a/}');
Error('\P{Is_Age=	 +001.1/a/}');
Expect(1, 65533, '\p{Is_Age=0_1.1}', "");
Expect(0, 65533, '\p{^Is_Age=0_1.1}', "");
Expect(0, 65533, '\P{Is_Age=0_1.1}', "");
Expect(1, 65533, '\P{^Is_Age=0_1.1}', "");
Expect(0, 65536, '\p{Is_Age=0_1.1}', "");
Expect(1, 65536, '\p{^Is_Age=0_1.1}', "");
Expect(1, 65536, '\P{Is_Age=0_1.1}', "");
Expect(0, 65536, '\P{^Is_Age=0_1.1}', "");
Error('\p{Age=:=__V5_1}');
Error('\P{Age=:=__V5_1}');
Expect(1, 127123, '\p{Age:   v51}', "");
Expect(0, 127123, '\p{^Age:   v51}', "");
Expect(0, 127123, '\P{Age:   v51}', "");
Expect(1, 127123, '\P{^Age:   v51}', "");
Expect(0, 127124, '\p{Age:   v51}', "");
Expect(1, 127124, '\p{^Age:   v51}', "");
Expect(1, 127124, '\P{Age:   v51}', "");
Expect(0, 127124, '\P{^Age:   v51}', "");
Expect(1, 127123, '\p{Age=-V5_1}', "");
Expect(0, 127123, '\p{^Age=-V5_1}', "");
Expect(0, 127123, '\P{Age=-V5_1}', "");
Expect(1, 127123, '\P{^Age=-V5_1}', "");
Expect(0, 127124, '\p{Age=-V5_1}', "");
Expect(1, 127124, '\p{^Age=-V5_1}', "");
Expect(1, 127124, '\P{Age=-V5_1}', "");
Expect(0, 127124, '\P{^Age=-V5_1}', "");
Error('\p{Is_Age=/a/	000000000_5.1}');
Error('\P{Is_Age=/a/	000000000_5.1}');
Expect(1, 127123, '\p{Is_Age=0000_0000_5.1}', "");
Expect(0, 127123, '\p{^Is_Age=0000_0000_5.1}', "");
Expect(0, 127123, '\P{Is_Age=0000_0000_5.1}', "");
Expect(1, 127123, '\P{^Is_Age=0000_0000_5.1}', "");
Expect(0, 127124, '\p{Is_Age=0000_0000_5.1}', "");
Expect(1, 127124, '\p{^Is_Age=0000_0000_5.1}', "");
Expect(1, 127124, '\P{Is_Age=0000_0000_5.1}', "");
Expect(0, 127124, '\P{^Is_Age=0000_0000_5.1}', "");
Error('\p{Age=:= 	V3_1}');
Error('\P{Age=:= 	V3_1}');
Expect(1, 917631, '\p{Age=v31}', "");
Expect(0, 917631, '\p{^Age=v31}', "");
Expect(0, 917631, '\P{Age=v31}', "");
Expect(1, 917631, '\P{^Age=v31}', "");
Expect(0, 917632, '\p{Age=v31}', "");
Expect(1, 917632, '\p{^Age=v31}', "");
Expect(1, 917632, '\P{Age=v31}', "");
Expect(0, 917632, '\P{^Age=v31}', "");
Expect(1, 917631, '\p{Age= 	V3_1}', "");
Expect(0, 917631, '\p{^Age= 	V3_1}', "");
Expect(0, 917631, '\P{Age= 	V3_1}', "");
Expect(1, 917631, '\P{^Age= 	V3_1}', "");
Expect(0, 917632, '\p{Age= 	V3_1}', "");
Expect(1, 917632, '\p{^Age= 	V3_1}', "");
Expect(1, 917632, '\P{Age= 	V3_1}', "");
Expect(0, 917632, '\P{^Age= 	V3_1}', "");
Error('\p{Is_Age=	0_0_03.1/a/}');
Error('\P{Is_Age=	0_0_03.1/a/}');
Expect(1, 917631, '\p{Is_Age=0_0_0_03.1}', "");
Expect(0, 917631, '\p{^Is_Age=0_0_0_03.1}', "");
Expect(0, 917631, '\P{Is_Age=0_0_0_03.1}', "");
Expect(1, 917631, '\P{^Is_Age=0_0_0_03.1}', "");
Expect(0, 917632, '\p{Is_Age=0_0_0_03.1}', "");
Expect(1, 917632, '\p{^Is_Age=0_0_0_03.1}', "");
Expect(1, 917632, '\P{Is_Age=0_0_0_03.1}', "");
Expect(0, 917632, '\P{^Is_Age=0_0_0_03.1}', "");
Error('\p{Deprecated=/a/	_Yes}');
Error('\P{Deprecated=/a/	_Yes}');
Expect(1, 917631, '\p{Deprecated=yes}', "");
Expect(0, 917631, '\p{^Deprecated=yes}', "");
Expect(0, 917631, '\P{Deprecated=yes}', "");
Expect(1, 917631, '\P{^Deprecated=yes}', "");
Expect(0, 917632, '\p{Deprecated=yes}', "");
Expect(1, 917632, '\p{^Deprecated=yes}', "");
Expect(1, 917632, '\P{Deprecated=yes}', "");
Expect(0, 917632, '\P{^Deprecated=yes}', "");
Expect(1, 917631, '\p{Deprecated=_YES}', "");
Expect(0, 917631, '\p{^Deprecated=_YES}', "");
Expect(0, 917631, '\P{Deprecated=_YES}', "");
Expect(1, 917631, '\P{^Deprecated=_YES}', "");
Expect(0, 917632, '\p{Deprecated=_YES}', "");
Expect(1, 917632, '\p{^Deprecated=_YES}', "");
Expect(1, 917632, '\P{Deprecated=_YES}', "");
Expect(0, 917632, '\P{^Deprecated=_YES}', "");
Error('\p{Dep=_ Y/a/}');
Error('\P{Dep=_ Y/a/}');
Expect(1, 917631, '\p{Dep=y}', "");
Expect(0, 917631, '\p{^Dep=y}', "");
Expect(0, 917631, '\P{Dep=y}', "");
Expect(1, 917631, '\P{^Dep=y}', "");
Expect(0, 917632, '\p{Dep=y}', "");
Expect(1, 917632, '\p{^Dep=y}', "");
Expect(1, 917632, '\P{Dep=y}', "");
Expect(0, 917632, '\P{^Dep=y}', "");
Expect(1, 917631, '\p{Dep=_Y}', "");
Expect(0, 917631, '\p{^Dep=_Y}', "");
Expect(0, 917631, '\P{Dep=_Y}', "");
Expect(1, 917631, '\P{^Dep=_Y}', "");
Expect(0, 917632, '\p{Dep=_Y}', "");
Expect(1, 917632, '\p{^Dep=_Y}', "");
Expect(1, 917632, '\P{Dep=_Y}', "");
Expect(0, 917632, '\P{^Dep=_Y}', "");
Error('\p{Is_Deprecated= :=t}');
Error('\P{Is_Deprecated= :=t}');
Expect(1, 917631, '\p{Is_Deprecated=t}', "");
Expect(0, 917631, '\p{^Is_Deprecated=t}', "");
Expect(0, 917631, '\P{Is_Deprecated=t}', "");
Expect(1, 917631, '\P{^Is_Deprecated=t}', "");
Expect(0, 917632, '\p{Is_Deprecated=t}', "");
Expect(1, 917632, '\p{^Is_Deprecated=t}', "");
Expect(1, 917632, '\P{Is_Deprecated=t}', "");
Expect(0, 917632, '\P{^Is_Deprecated=t}', "");
Expect(1, 917631, '\p{Is_Deprecated=	 T}', "");
Expect(0, 917631, '\p{^Is_Deprecated=	 T}', "");
Expect(0, 917631, '\P{Is_Deprecated=	 T}', "");
Expect(1, 917631, '\P{^Is_Deprecated=	 T}', "");
Expect(0, 917632, '\p{Is_Deprecated=	 T}', "");
Expect(1, 917632, '\p{^Is_Deprecated=	 T}', "");
Expect(1, 917632, '\P{Is_Deprecated=	 T}', "");
Expect(0, 917632, '\P{^Is_Deprecated=	 T}', "");
Error('\p{Is_Dep=_:=true}');
Error('\P{Is_Dep=_:=true}');
Expect(1, 917631, '\p{Is_Dep=true}', "");
Expect(0, 917631, '\p{^Is_Dep=true}', "");
Expect(0, 917631, '\P{Is_Dep=true}', "");
Expect(1, 917631, '\P{^Is_Dep=true}', "");
Expect(0, 917632, '\p{Is_Dep=true}', "");
Expect(1, 917632, '\p{^Is_Dep=true}', "");
Expect(1, 917632, '\P{Is_Dep=true}', "");
Expect(0, 917632, '\P{^Is_Dep=true}', "");
Expect(1, 917631, '\p{Is_Dep=	_TRUE}', "");
Expect(0, 917631, '\p{^Is_Dep=	_TRUE}', "");
Expect(0, 917631, '\P{Is_Dep=	_TRUE}', "");
Expect(1, 917631, '\P{^Is_Dep=	_TRUE}', "");
Expect(0, 917632, '\p{Is_Dep=	_TRUE}', "");
Expect(1, 917632, '\p{^Is_Dep=	_TRUE}', "");
Expect(1, 917632, '\P{Is_Dep=	_TRUE}', "");
Expect(0, 917632, '\P{^Is_Dep=	_TRUE}', "");
Error('\p{Deprecated=/a/-	No}');
Error('\P{Deprecated=/a/-	No}');
Expect(1, 917632, '\p{Deprecated=no}', "");
Expect(0, 917632, '\p{^Deprecated=no}', "");
Expect(0, 917632, '\P{Deprecated=no}', "");
Expect(1, 917632, '\P{^Deprecated=no}', "");
Expect(0, 917631, '\p{Deprecated=no}', "");
Expect(1, 917631, '\p{^Deprecated=no}', "");
Expect(1, 917631, '\P{Deprecated=no}', "");
Expect(0, 917631, '\P{^Deprecated=no}', "");
Expect(1, 917632, '\p{Deprecated= 	NO}', "");
Expect(0, 917632, '\p{^Deprecated= 	NO}', "");
Expect(0, 917632, '\P{Deprecated= 	NO}', "");
Expect(1, 917632, '\P{^Deprecated= 	NO}', "");
Expect(0, 917631, '\p{Deprecated= 	NO}', "");
Expect(1, 917631, '\p{^Deprecated= 	NO}', "");
Expect(1, 917631, '\P{Deprecated= 	NO}', "");
Expect(0, 917631, '\P{^Deprecated= 	NO}', "");
Error('\p{Dep=/a/ 	N}');
Error('\P{Dep=/a/ 	N}');
Expect(1, 917632, '\p{Dep=n}', "");
Expect(0, 917632, '\p{^Dep=n}', "");
Expect(0, 917632, '\P{Dep=n}', "");
Expect(1, 917632, '\P{^Dep=n}', "");
Expect(0, 917631, '\p{Dep=n}', "");
Expect(1, 917631, '\p{^Dep=n}', "");
Expect(1, 917631, '\P{Dep=n}', "");
Expect(0, 917631, '\P{^Dep=n}', "");
Expect(1, 917632, '\p{Dep=_ N}', "");
Expect(0, 917632, '\p{^Dep=_ N}', "");
Expect(0, 917632, '\P{Dep=_ N}', "");
Expect(1, 917632, '\P{^Dep=_ N}', "");
Expect(0, 917631, '\p{Dep=_ N}', "");
Expect(1, 917631, '\p{^Dep=_ N}', "");
Expect(1, 917631, '\P{Dep=_ N}', "");
Expect(0, 917631, '\P{^Dep=_ N}', "");
Error('\p{Is_Deprecated=/a/f}');
Error('\P{Is_Deprecated=/a/f}');
Expect(1, 917632, '\p{Is_Deprecated=f}', "");
Expect(0, 917632, '\p{^Is_Deprecated=f}', "");
Expect(0, 917632, '\P{Is_Deprecated=f}', "");
Expect(1, 917632, '\P{^Is_Deprecated=f}', "");
Expect(0, 917631, '\p{Is_Deprecated=f}', "");
Expect(1, 917631, '\p{^Is_Deprecated=f}', "");
Expect(1, 917631, '\P{Is_Deprecated=f}', "");
Expect(0, 917631, '\P{^Is_Deprecated=f}', "");
Expect(1, 917632, '\p{Is_Deprecated=_-F}', "");
Expect(0, 917632, '\p{^Is_Deprecated=_-F}', "");
Expect(0, 917632, '\P{Is_Deprecated=_-F}', "");
Expect(1, 917632, '\P{^Is_Deprecated=_-F}', "");
Expect(0, 917631, '\p{Is_Deprecated=_-F}', "");
Expect(1, 917631, '\p{^Is_Deprecated=_-F}', "");
Expect(1, 917631, '\P{Is_Deprecated=_-F}', "");
Expect(0, 917631, '\P{^Is_Deprecated=_-F}', "");
Error('\p{Is_Dep=/a/	-FALSE}');
Error('\P{Is_Dep=/a/	-FALSE}');
Expect(1, 917632, '\p{Is_Dep=false}', "");
Expect(0, 917632, '\p{^Is_Dep=false}', "");
Expect(0, 917632, '\P{Is_Dep=false}', "");
Expect(1, 917632, '\P{^Is_Dep=false}', "");
Expect(0, 917631, '\p{Is_Dep=false}', "");
Expect(1, 917631, '\p{^Is_Dep=false}', "");
Expect(1, 917631, '\P{Is_Dep=false}', "");
Expect(0, 917631, '\P{^Is_Dep=false}', "");
Expect(1, 917632, '\p{Is_Dep=  false}', "");
Expect(0, 917632, '\p{^Is_Dep=  false}', "");
Expect(0, 917632, '\P{Is_Dep=  false}', "");
Expect(1, 917632, '\P{^Is_Dep=  false}', "");
Expect(0, 917631, '\p{Is_Dep=  false}', "");
Expect(1, 917631, '\p{^Is_Dep=  false}', "");
Expect(1, 917631, '\P{Is_Dep=  false}', "");
Expect(0, 917631, '\P{^Is_Dep=  false}', "");
Error('\p{Case_Ignorable=:= Yes}');
Error('\P{Case_Ignorable=:= Yes}');
Expect(1, 917999, '\p{Case_Ignorable=yes}', "");
Expect(0, 917999, '\p{^Case_Ignorable=yes}', "");
Expect(0, 917999, '\P{Case_Ignorable=yes}', "");
Expect(1, 917999, '\P{^Case_Ignorable=yes}', "");
Expect(0, 918000, '\p{Case_Ignorable=yes}', "");
Expect(1, 918000, '\p{^Case_Ignorable=yes}', "");
Expect(1, 918000, '\P{Case_Ignorable=yes}', "");
Expect(0, 918000, '\P{^Case_Ignorable=yes}', "");
Expect(1, 917999, '\p{Case_Ignorable=-Yes}', "");
Expect(0, 917999, '\p{^Case_Ignorable=-Yes}', "");
Expect(0, 917999, '\P{Case_Ignorable=-Yes}', "");
Expect(1, 917999, '\P{^Case_Ignorable=-Yes}', "");
Expect(0, 918000, '\p{Case_Ignorable=-Yes}', "");
Expect(1, 918000, '\p{^Case_Ignorable=-Yes}', "");
Expect(1, 918000, '\P{Case_Ignorable=-Yes}', "");
Expect(0, 918000, '\P{^Case_Ignorable=-Yes}', "");
Error('\p{CI=:=__Y}');
Error('\P{CI=:=__Y}');
Expect(1, 917999, '\p{CI=y}', "");
Expect(0, 917999, '\p{^CI=y}', "");
Expect(0, 917999, '\P{CI=y}', "");
Expect(1, 917999, '\P{^CI=y}', "");
Expect(0, 918000, '\p{CI=y}', "");
Expect(1, 918000, '\p{^CI=y}', "");
Expect(1, 918000, '\P{CI=y}', "");
Expect(0, 918000, '\P{^CI=y}', "");
Expect(1, 917999, '\p{CI: 	Y}', "");
Expect(0, 917999, '\p{^CI: 	Y}', "");
Expect(0, 917999, '\P{CI: 	Y}', "");
Expect(1, 917999, '\P{^CI: 	Y}', "");
Expect(0, 918000, '\p{CI: 	Y}', "");
Expect(1, 918000, '\p{^CI: 	Y}', "");
Expect(1, 918000, '\P{CI: 	Y}', "");
Expect(0, 918000, '\P{^CI: 	Y}', "");
Error('\p{Is_Case_Ignorable=-:=T}');
Error('\P{Is_Case_Ignorable=-:=T}');
Expect(1, 917999, '\p{Is_Case_Ignorable=t}', "");
Expect(0, 917999, '\p{^Is_Case_Ignorable=t}', "");
Expect(0, 917999, '\P{Is_Case_Ignorable=t}', "");
Expect(1, 917999, '\P{^Is_Case_Ignorable=t}', "");
Expect(0, 918000, '\p{Is_Case_Ignorable=t}', "");
Expect(1, 918000, '\p{^Is_Case_Ignorable=t}', "");
Expect(1, 918000, '\P{Is_Case_Ignorable=t}', "");
Expect(0, 918000, '\P{^Is_Case_Ignorable=t}', "");
Expect(1, 917999, '\p{Is_Case_Ignorable=__T}', "");
Expect(0, 917999, '\p{^Is_Case_Ignorable=__T}', "");
Expect(0, 917999, '\P{Is_Case_Ignorable=__T}', "");
Expect(1, 917999, '\P{^Is_Case_Ignorable=__T}', "");
Expect(0, 918000, '\p{Is_Case_Ignorable=__T}', "");
Expect(1, 918000, '\p{^Is_Case_Ignorable=__T}', "");
Expect(1, 918000, '\P{Is_Case_Ignorable=__T}', "");
Expect(0, 918000, '\P{^Is_Case_Ignorable=__T}', "");
Error('\p{Is_CI=_-True/a/}');
Error('\P{Is_CI=_-True/a/}');
Expect(1, 917999, '\p{Is_CI=true}', "");
Expect(0, 917999, '\p{^Is_CI=true}', "");
Expect(0, 917999, '\P{Is_CI=true}', "");
Expect(1, 917999, '\P{^Is_CI=true}', "");
Expect(0, 918000, '\p{Is_CI=true}', "");
Expect(1, 918000, '\p{^Is_CI=true}', "");
Expect(1, 918000, '\P{Is_CI=true}', "");
Expect(0, 918000, '\P{^Is_CI=true}', "");
Expect(1, 917999, '\p{Is_CI: __TRUE}', "");
Expect(0, 917999, '\p{^Is_CI: __TRUE}', "");
Expect(0, 917999, '\P{Is_CI: __TRUE}', "");
Expect(1, 917999, '\P{^Is_CI: __TRUE}', "");
Expect(0, 918000, '\p{Is_CI: __TRUE}', "");
Expect(1, 918000, '\p{^Is_CI: __TRUE}', "");
Expect(1, 918000, '\P{Is_CI: __TRUE}', "");
Expect(0, 918000, '\P{^Is_CI: __TRUE}', "");
Error('\p{Case_Ignorable=_/a/NO}');
Error('\P{Case_Ignorable=_/a/NO}');
Expect(1, 918000, '\p{Case_Ignorable=no}', "");
Expect(0, 918000, '\p{^Case_Ignorable=no}', "");
Expect(0, 918000, '\P{Case_Ignorable=no}', "");
Expect(1, 918000, '\P{^Case_Ignorable=no}', "");
Expect(0, 917999, '\p{Case_Ignorable=no}', "");
Expect(1, 917999, '\p{^Case_Ignorable=no}', "");
Expect(1, 917999, '\P{Case_Ignorable=no}', "");
Expect(0, 917999, '\P{^Case_Ignorable=no}', "");
Expect(1, 918000, '\p{Case_Ignorable=_ No}', "");
Expect(0, 918000, '\p{^Case_Ignorable=_ No}', "");
Expect(0, 918000, '\P{Case_Ignorable=_ No}', "");
Expect(1, 918000, '\P{^Case_Ignorable=_ No}', "");
Expect(0, 917999, '\p{Case_Ignorable=_ No}', "");
Expect(1, 917999, '\p{^Case_Ignorable=_ No}', "");
Expect(1, 917999, '\P{Case_Ignorable=_ No}', "");
Expect(0, 917999, '\P{^Case_Ignorable=_ No}', "");
Error('\p{CI=	N/a/}');
Error('\P{CI=	N/a/}');
Expect(1, 918000, '\p{CI=n}', "");
Expect(0, 918000, '\p{^CI=n}', "");
Expect(0, 918000, '\P{CI=n}', "");
Expect(1, 918000, '\P{^CI=n}', "");
Expect(0, 917999, '\p{CI=n}', "");
Expect(1, 917999, '\p{^CI=n}', "");
Expect(1, 917999, '\P{CI=n}', "");
Expect(0, 917999, '\P{^CI=n}', "");
Expect(1, 918000, '\p{CI=- N}', "");
Expect(0, 918000, '\p{^CI=- N}', "");
Expect(0, 918000, '\P{CI=- N}', "");
Expect(1, 918000, '\P{^CI=- N}', "");
Expect(0, 917999, '\p{CI=- N}', "");
Expect(1, 917999, '\p{^CI=- N}', "");
Expect(1, 917999, '\P{CI=- N}', "");
Expect(0, 917999, '\P{^CI=- N}', "");
Error('\p{Is_Case_Ignorable=/a/		F}');
Error('\P{Is_Case_Ignorable=/a/		F}');
Expect(1, 918000, '\p{Is_Case_Ignorable=f}', "");
Expect(0, 918000, '\p{^Is_Case_Ignorable=f}', "");
Expect(0, 918000, '\P{Is_Case_Ignorable=f}', "");
Expect(1, 918000, '\P{^Is_Case_Ignorable=f}', "");
Expect(0, 917999, '\p{Is_Case_Ignorable=f}', "");
Expect(1, 917999, '\p{^Is_Case_Ignorable=f}', "");
Expect(1, 917999, '\P{Is_Case_Ignorable=f}', "");
Expect(0, 917999, '\P{^Is_Case_Ignorable=f}', "");
Expect(1, 918000, '\p{Is_Case_Ignorable=__F}', "");
Expect(0, 918000, '\p{^Is_Case_Ignorable=__F}', "");
Expect(0, 918000, '\P{Is_Case_Ignorable=__F}', "");
Expect(1, 918000, '\P{^Is_Case_Ignorable=__F}', "");
Expect(0, 917999, '\p{Is_Case_Ignorable=__F}', "");
Expect(1, 917999, '\p{^Is_Case_Ignorable=__F}', "");
Expect(1, 917999, '\P{Is_Case_Ignorable=__F}', "");
Expect(0, 917999, '\P{^Is_Case_Ignorable=__F}', "");
Error('\p{Is_CI=/a/False}');
Error('\P{Is_CI=/a/False}');
Expect(1, 918000, '\p{Is_CI=false}', "");
Expect(0, 918000, '\p{^Is_CI=false}', "");
Expect(0, 918000, '\P{Is_CI=false}', "");
Expect(1, 918000, '\P{^Is_CI=false}', "");
Expect(0, 917999, '\p{Is_CI=false}', "");
Expect(1, 917999, '\p{^Is_CI=false}', "");
Expect(1, 917999, '\P{Is_CI=false}', "");
Expect(0, 917999, '\P{^Is_CI=false}', "");
Expect(1, 918000, '\p{Is_CI= -false}', "");
Expect(0, 918000, '\p{^Is_CI= -false}', "");
Expect(0, 918000, '\P{Is_CI= -false}', "");
Expect(1, 918000, '\P{^Is_CI= -false}', "");
Expect(0, 917999, '\p{Is_CI= -false}', "");
Expect(1, 917999, '\p{^Is_CI= -false}', "");
Expect(1, 917999, '\P{Is_CI= -false}', "");
Expect(0, 917999, '\P{^Is_CI= -false}', "");
Error('\p{NFC_Quick_Check= /a/Yes}');
Error('\P{NFC_Quick_Check= /a/Yes}');
Expect(1, 195102, '\p{NFC_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^NFC_Quick_Check=yes}', "");
Expect(0, 195102, '\P{NFC_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^NFC_Quick_Check=yes}', "");
Expect(0, 195101, '\p{NFC_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^NFC_Quick_Check=yes}', "");
Expect(1, 195101, '\P{NFC_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^NFC_Quick_Check=yes}', "");
Expect(1, 195102, '\p{NFC_Quick_Check= Yes}', "");
Expect(0, 195102, '\p{^NFC_Quick_Check= Yes}', "");
Expect(0, 195102, '\P{NFC_Quick_Check= Yes}', "");
Expect(1, 195102, '\P{^NFC_Quick_Check= Yes}', "");
Expect(0, 195101, '\p{NFC_Quick_Check= Yes}', "");
Expect(1, 195101, '\p{^NFC_Quick_Check= Yes}', "");
Expect(1, 195101, '\P{NFC_Quick_Check= Yes}', "");
Expect(0, 195101, '\P{^NFC_Quick_Check= Yes}', "");
Error('\p{NFC_QC=_ y/a/}');
Error('\P{NFC_QC=_ y/a/}');
Expect(1, 195102, '\p{NFC_QC=y}', "");
Expect(0, 195102, '\p{^NFC_QC=y}', "");
Expect(0, 195102, '\P{NFC_QC=y}', "");
Expect(1, 195102, '\P{^NFC_QC=y}', "");
Expect(0, 195101, '\p{NFC_QC=y}', "");
Expect(1, 195101, '\p{^NFC_QC=y}', "");
Expect(1, 195101, '\P{NFC_QC=y}', "");
Expect(0, 195101, '\P{^NFC_QC=y}', "");
Expect(1, 195102, '\p{NFC_QC=_ Y}', "");
Expect(0, 195102, '\p{^NFC_QC=_ Y}', "");
Expect(0, 195102, '\P{NFC_QC=_ Y}', "");
Expect(1, 195102, '\P{^NFC_QC=_ Y}', "");
Expect(0, 195101, '\p{NFC_QC=_ Y}', "");
Expect(1, 195101, '\p{^NFC_QC=_ Y}', "");
Expect(1, 195101, '\P{NFC_QC=_ Y}', "");
Expect(0, 195101, '\P{^NFC_QC=_ Y}', "");
Error('\p{Is_NFC_Quick_Check=/a/	 YES}');
Error('\P{Is_NFC_Quick_Check=/a/	 YES}');
Expect(1, 195102, '\p{Is_NFC_Quick_Check=yes}', "");
Expect(0, 195102, '\p{^Is_NFC_Quick_Check=yes}', "");
Expect(0, 195102, '\P{Is_NFC_Quick_Check=yes}', "");
Expect(1, 195102, '\P{^Is_NFC_Quick_Check=yes}', "");
Expect(0, 195101, '\p{Is_NFC_Quick_Check=yes}', "");
Expect(1, 195101, '\p{^Is_NFC_Quick_Check=yes}', "");
Expect(1, 195101, '\P{Is_NFC_Quick_Check=yes}', "");
Expect(0, 195101, '\P{^Is_NFC_Quick_Check=yes}', "");
Expect(1, 195102, '\p{Is_NFC_Quick_Check= 	yes}', "");
Expect(0, 195102, '\p{^Is_NFC_Quick_Check= 	yes}', "");
Expect(0, 195102, '\P{Is_NFC_Quick_Check= 	yes}', "");
Expect(1, 195102, '\P{^Is_NFC_Quick_Check= 	yes}', "");
Expect(0, 195101, '\p{Is_NFC_Quick_Check= 	yes}', "");
Expect(1, 195101, '\p{^Is_NFC_Quick_Check= 	yes}', "");
Expect(1, 195101, '\P{Is_NFC_Quick_Check= 	yes}', "");
Expect(0, 195101, '\P{^Is_NFC_Quick_Check= 	yes}', "");
Error('\p{Is_NFC_QC= Y:=}');
Error('\P{Is_NFC_QC= Y:=}');
Expect(1, 195102, '\p{Is_NFC_QC=y}', "");
Expect(0, 195102, '\p{^Is_NFC_QC=y}', "");
Expect(0, 195102, '\P{Is_NFC_QC=y}', "");
Expect(1, 195102, '\P{^Is_NFC_QC=y}', "");
Expect(0, 195101, '\p{Is_NFC_QC=y}', "");
Expect(1, 195101, '\p{^Is_NFC_QC=y}', "");
Expect(1, 195101, '\P{Is_NFC_QC=y}', "");
Expect(0, 195101, '\P{^Is_NFC_QC=y}', "");
Expect(1, 195102, '\p{Is_NFC_QC=	Y}', "");
Expect(0, 195102, '\p{^Is_NFC_QC=	Y}', "");
Expect(0, 195102, '\P{Is_NFC_QC=	Y}', "");
Expect(1, 195102, '\P{^Is_NFC_QC=	Y}', "");
Expect(0, 195101, '\p{Is_NFC_QC=	Y}', "");
Expect(1, 195101, '\p{^Is_NFC_QC=	Y}', "");
Expect(1, 195101, '\P{Is_NFC_QC=	Y}', "");
Expect(0, 195101, '\P{^Is_NFC_QC=	Y}', "");
Error('\p{NFC_Quick_Check=-:=no}');
Error('\P{NFC_Quick_Check=-:=no}');
Expect(1, 195101, '\p{NFC_Quick_Check: no}', "");
Expect(0, 195101, '\p{^NFC_Quick_Check: no}', "");
Expect(0, 195101, '\P{NFC_Quick_Check: no}', "");
Expect(1, 195101, '\P{^NFC_Quick_Check: no}', "");
Expect(0, 195102, '\p{NFC_Quick_Check: no}', "");
Expect(1, 195102, '\p{^NFC_Quick_Check: no}', "");
Expect(1, 195102, '\P{NFC_Quick_Check: no}', "");
Expect(0, 195102, '\P{^NFC_Quick_Check: no}', "");
Expect(1, 195101, '\p{NFC_Quick_Check= No}', "");
Expect(0, 195101, '\p{^NFC_Quick_Check= No}', "");
Expect(0, 195101, '\P{NFC_Quick_Check= No}', "");
Expect(1, 195101, '\P{^NFC_Quick_Check= No}', "");
Expect(0, 195102, '\p{NFC_Quick_Check= No}', "");
Expect(1, 195102, '\p{^NFC_Quick_Check= No}', "");
Expect(1, 195102, '\P{NFC_Quick_Check= No}', "");
Expect(0, 195102, '\P{^NFC_Quick_Check= No}', "");
Error('\p{NFC_QC=/a/_N}');
Error('\P{NFC_QC=/a/_N}');
Expect(1, 195101, '\p{NFC_QC=n}', "");
Expect(0, 195101, '\p{^NFC_QC=n}', "");
Expect(0, 195101, '\P{NFC_QC=n}', "");
Expect(1, 195101, '\P{^NFC_QC=n}', "");
Expect(0, 195102, '\p{NFC_QC=n}', "");
Expect(1, 195102, '\p{^NFC_QC=n}', "");
Expect(1, 195102, '\P{NFC_QC=n}', "");
Expect(0, 195102, '\P{^NFC_QC=n}', "");
Expect(1, 195101, '\p{NFC_QC=-N}', "");
Expect(0, 195101, '\p{^NFC_QC=-N}', "");
Expect(0, 195101, '\P{NFC_QC=-N}', "");
Expect(1, 195101, '\P{^NFC_QC=-N}', "");
Expect(0, 195102, '\p{NFC_QC=-N}', "");
Expect(1, 195102, '\p{^NFC_QC=-N}', "");
Expect(1, 195102, '\P{NFC_QC=-N}', "");
Expect(0, 195102, '\P{^NFC_QC=-N}', "");
Error('\p{Is_NFC_Quick_Check= -No/a/}');
Error('\P{Is_NFC_Quick_Check= -No/a/}');
Expect(1, 195101, '\p{Is_NFC_Quick_Check=no}', "");
Expect(0, 195101, '\p{^Is_NFC_Quick_Check=no}', "");
Expect(0, 195101, '\P{Is_NFC_Quick_Check=no}', "");
Expect(1, 195101, '\P{^Is_NFC_Quick_Check=no}', "");
Expect(0, 195102, '\p{Is_NFC_Quick_Check=no}', "");
Expect(1, 195102, '\p{^Is_NFC_Quick_Check=no}', "");
Expect(1, 195102, '\P{Is_NFC_Quick_Check=no}', "");
Expect(0, 195102, '\P{^Is_NFC_Quick_Check=no}', "");
Expect(1, 195101, '\p{Is_NFC_Quick_Check=--NO}', "");
Expect(0, 195101, '\p{^Is_NFC_Quick_Check=--NO}', "");
Expect(0, 195101, '\P{Is_NFC_Quick_Check=--NO}', "");
Expect(1, 195101, '\P{^Is_NFC_Quick_Check=--NO}', "");
Expect(0, 195102, '\p{Is_NFC_Quick_Check=--NO}', "");
Expect(1, 195102, '\p{^Is_NFC_Quick_Check=--NO}', "");
Expect(1, 195102, '\P{Is_NFC_Quick_Check=--NO}', "");
Expect(0, 195102, '\P{^Is_NFC_Quick_Check=--NO}', "");
Error('\p{Is_NFC_QC=- N/a/}');
Error('\P{Is_NFC_QC=- N/a/}');
Expect(1, 195101, '\p{Is_NFC_QC=n}', "");
Expect(0, 195101, '\p{^Is_NFC_QC=n}', "");
Expect(0, 195101, '\P{Is_NFC_QC=n}', "");
Expect(1, 195101, '\P{^Is_NFC_QC=n}', "");
Expect(0, 195102, '\p{Is_NFC_QC=n}', "");
Expect(1, 195102, '\p{^Is_NFC_QC=n}', "");
Expect(1, 195102, '\P{Is_NFC_QC=n}', "");
Expect(0, 195102, '\P{^Is_NFC_QC=n}', "");
Expect(1, 195101, '\p{Is_NFC_QC= 	N}', "");
Expect(0, 195101, '\p{^Is_NFC_QC= 	N}', "");
Expect(0, 195101, '\P{Is_NFC_QC= 	N}', "");
Expect(1, 195101, '\P{^Is_NFC_QC= 	N}', "");
Expect(0, 195102, '\p{Is_NFC_QC= 	N}', "");
Expect(1, 195102, '\p{^Is_NFC_QC= 	N}', "");
Expect(1, 195102, '\P{Is_NFC_QC= 	N}', "");
Expect(0, 195102, '\P{^Is_NFC_QC= 	N}', "");
Error('\p{NFC_Quick_Check=	_maybe/a/}');
Error('\P{NFC_Quick_Check=	_maybe/a/}');
Expect(1, 69927, '\p{NFC_Quick_Check=maybe}', "");
Expect(0, 69927, '\p{^NFC_Quick_Check=maybe}', "");
Expect(0, 69927, '\P{NFC_Quick_Check=maybe}', "");
Expect(1, 69927, '\P{^NFC_Quick_Check=maybe}', "");
Expect(0, 69928, '\p{NFC_Quick_Check=maybe}', "");
Expect(1, 69928, '\p{^NFC_Quick_Check=maybe}', "");
Expect(1, 69928, '\P{NFC_Quick_Check=maybe}', "");
Expect(0, 69928, '\P{^NFC_Quick_Check=maybe}', "");
Expect(1, 69927, '\p{NFC_Quick_Check=	_maybe}', "");
Expect(0, 69927, '\p{^NFC_Quick_Check=	_maybe}', "");
Expect(0, 69927, '\P{NFC_Quick_Check=	_maybe}', "");
Expect(1, 69927, '\P{^NFC_Quick_Check=	_maybe}', "");
Expect(0, 69928, '\p{NFC_Quick_Check=	_maybe}', "");
Expect(1, 69928, '\p{^NFC_Quick_Check=	_maybe}', "");
Expect(1, 69928, '\P{NFC_Quick_Check=	_maybe}', "");
Expect(0, 69928, '\P{^NFC_Quick_Check=	_maybe}', "");
Error('\p{NFC_QC= m:=}');
Error('\P{NFC_QC= m:=}');
Expect(1, 69927, '\p{NFC_QC:m}', "");
Expect(0, 69927, '\p{^NFC_QC:m}', "");
Expect(0, 69927, '\P{NFC_QC:m}', "");
Expect(1, 69927, '\P{^NFC_QC:m}', "");
Expect(0, 69928, '\p{NFC_QC:m}', "");
Expect(1, 69928, '\p{^NFC_QC:m}', "");
Expect(1, 69928, '\P{NFC_QC:m}', "");
Expect(0, 69928, '\P{^NFC_QC:m}', "");
Expect(1, 69927, '\p{NFC_QC=	_M}', "");
Expect(0, 69927, '\p{^NFC_QC=	_M}', "");
Expect(0, 69927, '\P{NFC_QC=	_M}', "");
Expect(1, 69927, '\P{^NFC_QC=	_M}', "");
Expect(0, 69928, '\p{NFC_QC=	_M}', "");
Expect(1, 69928, '\p{^NFC_QC=	_M}', "");
Expect(1, 69928, '\P{NFC_QC=	_M}', "");
Expect(0, 69928, '\P{^NFC_QC=	_M}', "");
Error('\p{Is_NFC_Quick_Check=/a/Maybe}');
Error('\P{Is_NFC_Quick_Check=/a/Maybe}');
Expect(1, 69927, '\p{Is_NFC_Quick_Check:	maybe}', "");
Expect(0, 69927, '\p{^Is_NFC_Quick_Check:	maybe}', "");
Expect(0, 69927, '\P{Is_NFC_Quick_Check:	maybe}', "");
Expect(1, 69927, '\P{^Is_NFC_Quick_Check:	maybe}', "");
Expect(0, 69928, '\p{Is_NFC_Quick_Check:	maybe}', "");
Expect(1, 69928, '\p{^Is_NFC_Quick_Check:	maybe}', "");
Expect(1, 69928, '\P{Is_NFC_Quick_Check:	maybe}', "");
Expect(0, 69928, '\P{^Is_NFC_Quick_Check:	maybe}', "");
Expect(1, 69927, '\p{Is_NFC_Quick_Check=_ MAYBE}', "");
Expect(0, 69927, '\p{^Is_NFC_Quick_Check=_ MAYBE}', "");
Expect(0, 69927, '\P{Is_NFC_Quick_Check=_ MAYBE}', "");
Expect(1, 69927, '\P{^Is_NFC_Quick_Check=_ MAYBE}', "");
Expect(0, 69928, '\p{Is_NFC_Quick_Check=_ MAYBE}', "");
Expect(1, 69928, '\p{^Is_NFC_Quick_Check=_ MAYBE}', "");
Expect(1, 69928, '\P{Is_NFC_Quick_Check=_ MAYBE}', "");
Expect(0, 69928, '\P{^Is_NFC_Quick_Check=_ MAYBE}', "");
Error('\p{Is_NFC_QC=/a/--M}');
Error('\P{Is_NFC_QC=/a/--M}');
Expect(1, 69927, '\p{Is_NFC_QC=m}', "");
Expect(0, 69927, '\p{^Is_NFC_QC=m}', "");
Expect(0, 69927, '\P{Is_NFC_QC=m}', "");
Expect(1, 69927, '\P{^Is_NFC_QC=m}', "");
Expect(0, 69928, '\p{Is_NFC_QC=m}', "");
Expect(1, 69928, '\p{^Is_NFC_QC=m}', "");
Expect(1, 69928, '\P{Is_NFC_QC=m}', "");
Expect(0, 69928, '\P{^Is_NFC_QC=m}', "");
Expect(1, 69927, '\p{Is_NFC_QC= M}', "");
Expect(0, 69927, '\p{^Is_NFC_QC= M}', "");
Expect(0, 69927, '\P{Is_NFC_QC= M}', "");
Expect(1, 69927, '\P{^Is_NFC_QC= M}', "");
Expect(0, 69928, '\p{Is_NFC_QC= M}', "");
Expect(1, 69928, '\p{^Is_NFC_QC= M}', "");
Expect(1, 69928, '\P{Is_NFC_QC= M}', "");
Expect(0, 69928, '\P{^Is_NFC_QC= M}', "");
Error('\p{Other_Grapheme_Extend:   Yes}');
Error('\P{Other_Grapheme_Extend:   Yes}');
Error('\p{OGr_Ext=Y}');
Error('\P{OGr_Ext=Y}');
Error('\p{Is_Other_Grapheme_Extend=T}');
Error('\P{Is_Other_Grapheme_Extend=T}');
Error('\p{Is_OGr_Ext=True}');
Error('\P{Is_OGr_Ext=True}');
Error('\p{Other_Grapheme_Extend=No}');
Error('\P{Other_Grapheme_Extend=No}');
Error('\p{OGr_Ext:   N}');
Error('\P{OGr_Ext:   N}');
Error('\p{Is_Other_Grapheme_Extend=F}');
Error('\P{Is_Other_Grapheme_Extend=F}');
Error('\p{Is_OGr_Ext:False}');
Error('\P{Is_OGr_Ext:False}');
Error('\p{	:=gothic}');
Error('\P{	:=gothic}');
Expect(1, 66378, '\p{gothic}', "");
Expect(0, 66378, '\p{^gothic}', "");
Expect(0, 66378, '\P{gothic}', "");
Expect(1, 66378, '\P{^gothic}', "");
Expect(0, 66379, '\p{gothic}', "");
Expect(1, 66379, '\p{^gothic}', "");
Expect(1, 66379, '\P{gothic}', "");
Expect(0, 66379, '\P{^gothic}', "");
Expect(1, 66378, '\p{__Gothic}', "");
Expect(0, 66378, '\p{^__Gothic}', "");
Expect(0, 66378, '\P{__Gothic}', "");
Expect(1, 66378, '\P{^__Gothic}', "");
Expect(0, 66379, '\p{__Gothic}', "");
Expect(1, 66379, '\p{^__Gothic}', "");
Expect(1, 66379, '\P{__Gothic}', "");
Expect(0, 66379, '\P{^__Gothic}', "");
Error('\p{-IS_Gothic/a/}');
Error('\P{-IS_Gothic/a/}');
Expect(1, 66378, '\p{isgothic}', "");
Expect(0, 66378, '\p{^isgothic}', "");
Expect(0, 66378, '\P{isgothic}', "");
Expect(1, 66378, '\P{^isgothic}', "");
Expect(0, 66379, '\p{isgothic}', "");
Expect(1, 66379, '\p{^isgothic}', "");
Expect(1, 66379, '\P{isgothic}', "");
Expect(0, 66379, '\P{^isgothic}', "");
Expect(1, 66378, '\p{- is_Gothic}', "");
Expect(0, 66378, '\p{^- is_Gothic}', "");
Expect(0, 66378, '\P{- is_Gothic}', "");
Expect(1, 66378, '\P{^- is_Gothic}', "");
Expect(0, 66379, '\p{- is_Gothic}', "");
Expect(1, 66379, '\p{^- is_Gothic}', "");
Expect(1, 66379, '\P{- is_Gothic}', "");
Expect(0, 66379, '\P{^- is_Gothic}', "");
Error('\p{:=		Goth}');
Error('\P{:=		Goth}');
Expect(1, 66378, '\p{goth}', "");
Expect(0, 66378, '\p{^goth}', "");
Expect(0, 66378, '\P{goth}', "");
Expect(1, 66378, '\P{^goth}', "");
Expect(0, 66379, '\p{goth}', "");
Expect(1, 66379, '\p{^goth}', "");
Expect(1, 66379, '\P{goth}', "");
Expect(0, 66379, '\P{^goth}', "");
Expect(1, 66378, '\p{_Goth}', "");
Expect(0, 66378, '\p{^_Goth}', "");
Expect(0, 66378, '\P{_Goth}', "");
Expect(1, 66378, '\P{^_Goth}', "");
Expect(0, 66379, '\p{_Goth}', "");
Expect(1, 66379, '\p{^_Goth}', "");
Expect(1, 66379, '\P{_Goth}', "");
Expect(0, 66379, '\P{^_Goth}', "");
Error('\p{/a/-is_Goth}');
Error('\P{/a/-is_Goth}');
Expect(1, 66378, '\p{isgoth}', "");
Expect(0, 66378, '\p{^isgoth}', "");
Expect(0, 66378, '\P{isgoth}', "");
Expect(1, 66378, '\P{^isgoth}', "");
Expect(0, 66379, '\p{isgoth}', "");
Expect(1, 66379, '\p{^isgoth}', "");
Expect(1, 66379, '\P{isgoth}', "");
Expect(0, 66379, '\P{^isgoth}', "");
Expect(1, 66378, '\p{-is_Goth}', "");
Expect(0, 66378, '\p{^-is_Goth}', "");
Expect(0, 66378, '\P{-is_Goth}', "");
Expect(1, 66378, '\P{^-is_Goth}', "");
Expect(0, 66379, '\p{-is_Goth}', "");
Expect(1, 66379, '\p{^-is_Goth}', "");
Expect(1, 66379, '\P{-is_Goth}', "");
Expect(0, 66379, '\P{^-is_Goth}', "");
Error('\p{_ Blank/a/}');
Error('\P{_ Blank/a/}');
Expect(1, 12288, '\p{blank}', "");
Expect(0, 12288, '\p{^blank}', "");
Expect(0, 12288, '\P{blank}', "");
Expect(1, 12288, '\P{^blank}', "");
Expect(0, 12289, '\p{blank}', "");
Expect(1, 12289, '\p{^blank}', "");
Expect(1, 12289, '\P{blank}', "");
Expect(0, 12289, '\P{^blank}', "");
Expect(1, 12288, '\p{	Blank}', "");
Expect(0, 12288, '\p{^	Blank}', "");
Expect(0, 12288, '\P{	Blank}', "");
Expect(1, 12288, '\P{^	Blank}', "");
Expect(0, 12289, '\p{	Blank}', "");
Expect(1, 12289, '\p{^	Blank}', "");
Expect(1, 12289, '\P{	Blank}', "");
Expect(0, 12289, '\P{^	Blank}', "");
Error('\p{ /a/HorizSpace}');
Error('\P{ /a/HorizSpace}');
Expect(1, 12288, '\p{horizspace}', "");
Expect(0, 12288, '\p{^horizspace}', "");
Expect(0, 12288, '\P{horizspace}', "");
Expect(1, 12288, '\P{^horizspace}', "");
Expect(0, 12289, '\p{horizspace}', "");
Expect(1, 12289, '\p{^horizspace}', "");
Expect(1, 12289, '\P{horizspace}', "");
Expect(0, 12289, '\P{^horizspace}', "");
Expect(1, 12288, '\p{HorizSpace}', "");
Expect(0, 12288, '\p{^HorizSpace}', "");
Expect(0, 12288, '\P{HorizSpace}', "");
Expect(1, 12288, '\P{^HorizSpace}', "");
Expect(0, 12289, '\p{HorizSpace}', "");
Expect(1, 12289, '\p{^HorizSpace}', "");
Expect(1, 12289, '\P{HorizSpace}', "");
Expect(0, 12289, '\P{^HorizSpace}', "");
Error('\p{ _XPosixBlank:=}');
Error('\P{ _XPosixBlank:=}');
Expect(1, 12288, '\p{xposixblank}', "");
Expect(0, 12288, '\p{^xposixblank}', "");
Expect(0, 12288, '\P{xposixblank}', "");
Expect(1, 12288, '\P{^xposixblank}', "");
Expect(0, 12289, '\p{xposixblank}', "");
Expect(1, 12289, '\p{^xposixblank}', "");
Expect(1, 12289, '\P{xposixblank}', "");
Expect(0, 12289, '\P{^xposixblank}', "");
Expect(1, 12288, '\p{	xposixblank}', "");
Expect(0, 12288, '\p{^	xposixblank}', "");
Expect(0, 12288, '\P{	xposixblank}', "");
Expect(1, 12288, '\P{^	xposixblank}', "");
Expect(0, 12289, '\p{	xposixblank}', "");
Expect(1, 12289, '\p{^	xposixblank}', "");
Expect(1, 12289, '\P{	xposixblank}', "");
Expect(0, 12289, '\P{^	xposixblank}', "");
Error('\p{- Is_blank:=}');
Error('\P{- Is_blank:=}');
Expect(1, 12288, '\p{isblank}', "");
Expect(0, 12288, '\p{^isblank}', "");
Expect(0, 12288, '\P{isblank}', "");
Expect(1, 12288, '\P{^isblank}', "");
Expect(0, 12289, '\p{isblank}', "");
Expect(1, 12289, '\p{^isblank}', "");
Expect(1, 12289, '\P{isblank}', "");
Expect(0, 12289, '\P{^isblank}', "");
Expect(1, 12288, '\p{ is_Blank}', "");
Expect(0, 12288, '\p{^ is_Blank}', "");
Expect(0, 12288, '\P{ is_Blank}', "");
Expect(1, 12288, '\P{^ is_Blank}', "");
Expect(0, 12289, '\p{ is_Blank}', "");
Expect(1, 12289, '\p{^ is_Blank}', "");
Expect(1, 12289, '\P{ is_Blank}', "");
Expect(0, 12289, '\P{^ is_Blank}', "");
Error('\p{  Is_horizspace:=}');
Error('\P{  Is_horizspace:=}');
Expect(1, 12288, '\p{ishorizspace}', "");
Expect(0, 12288, '\p{^ishorizspace}', "");
Expect(0, 12288, '\P{ishorizspace}', "");
Expect(1, 12288, '\P{^ishorizspace}', "");
Expect(0, 12289, '\p{ishorizspace}', "");
Expect(1, 12289, '\p{^ishorizspace}', "");
Expect(1, 12289, '\P{ishorizspace}', "");
Expect(0, 12289, '\P{^ishorizspace}', "");
Expect(1, 12288, '\p{-is_HorizSpace}', "");
Expect(0, 12288, '\p{^-is_HorizSpace}', "");
Expect(0, 12288, '\P{-is_HorizSpace}', "");
Expect(1, 12288, '\P{^-is_HorizSpace}', "");
Expect(0, 12289, '\p{-is_HorizSpace}', "");
Expect(1, 12289, '\p{^-is_HorizSpace}', "");
Expect(1, 12289, '\P{-is_HorizSpace}', "");
Expect(0, 12289, '\P{^-is_HorizSpace}', "");
Error('\p{:=--is_XPOSIXBLANK}');
Error('\P{:=--is_XPOSIXBLANK}');
Expect(1, 12288, '\p{isxposixblank}', "");
Expect(0, 12288, '\p{^isxposixblank}', "");
Expect(0, 12288, '\P{isxposixblank}', "");
Expect(1, 12288, '\P{^isxposixblank}', "");
Expect(0, 12289, '\p{isxposixblank}', "");
Expect(1, 12289, '\p{^isxposixblank}', "");
Expect(1, 12289, '\P{isxposixblank}', "");
Expect(0, 12289, '\P{^isxposixblank}', "");
Expect(1, 12288, '\p{- IS_xposixblank}', "");
Expect(0, 12288, '\p{^- IS_xposixblank}', "");
Expect(0, 12288, '\P{- IS_xposixblank}', "");
Expect(1, 12288, '\P{^- IS_xposixblank}', "");
Expect(0, 12289, '\p{- IS_xposixblank}', "");
Expect(1, 12289, '\p{^- IS_xposixblank}', "");
Expect(1, 12289, '\P{- IS_xposixblank}', "");
Expect(0, 12289, '\P{^- IS_xposixblank}', "");
Error('\p{:= -Meroitic_CURSIVE}');
Error('\P{:= -Meroitic_CURSIVE}');
Expect(1, 68031, '\p{meroiticcursive}', "");
Expect(0, 68031, '\p{^meroiticcursive}', "");
Expect(0, 68031, '\P{meroiticcursive}', "");
Expect(1, 68031, '\P{^meroiticcursive}', "");
Expect(0, 68032, '\p{meroiticcursive}', "");
Expect(1, 68032, '\p{^meroiticcursive}', "");
Expect(1, 68032, '\P{meroiticcursive}', "");
Expect(0, 68032, '\P{^meroiticcursive}', "");
Expect(1, 68031, '\p{_ Meroitic_cursive}', "");
Expect(0, 68031, '\p{^_ Meroitic_cursive}', "");
Expect(0, 68031, '\P{_ Meroitic_cursive}', "");
Expect(1, 68031, '\P{^_ Meroitic_cursive}', "");
Expect(0, 68032, '\p{_ Meroitic_cursive}', "");
Expect(1, 68032, '\p{^_ Meroitic_cursive}', "");
Expect(1, 68032, '\P{_ Meroitic_cursive}', "");
Expect(0, 68032, '\P{^_ Meroitic_cursive}', "");
Error('\p{/a/-	Is_MEROITIC_CURSIVE}');
Error('\P{/a/-	Is_MEROITIC_CURSIVE}');
Expect(1, 68031, '\p{ismeroiticcursive}', "");
Expect(0, 68031, '\p{^ismeroiticcursive}', "");
Expect(0, 68031, '\P{ismeroiticcursive}', "");
Expect(1, 68031, '\P{^ismeroiticcursive}', "");
Expect(0, 68032, '\p{ismeroiticcursive}', "");
Expect(1, 68032, '\p{^ismeroiticcursive}', "");
Expect(1, 68032, '\P{ismeroiticcursive}', "");
Expect(0, 68032, '\P{^ismeroiticcursive}', "");
Expect(1, 68031, '\p{_ Is_meroitic_CURSIVE}', "");
Expect(0, 68031, '\p{^_ Is_meroitic_CURSIVE}', "");
Expect(0, 68031, '\P{_ Is_meroitic_CURSIVE}', "");
Expect(1, 68031, '\P{^_ Is_meroitic_CURSIVE}', "");
Expect(0, 68032, '\p{_ Is_meroitic_CURSIVE}', "");
Expect(1, 68032, '\p{^_ Is_meroitic_CURSIVE}', "");
Expect(1, 68032, '\P{_ Is_meroitic_CURSIVE}', "");
Expect(0, 68032, '\P{^_ Is_meroitic_CURSIVE}', "");
Error('\p{/a/MERC}');
Error('\P{/a/MERC}');
Expect(1, 68031, '\p{merc}', "");
Expect(0, 68031, '\p{^merc}', "");
Expect(0, 68031, '\P{merc}', "");
Expect(1, 68031, '\P{^merc}', "");
Expect(0, 68032, '\p{merc}', "");
Expect(1, 68032, '\p{^merc}', "");
Expect(1, 68032, '\P{merc}', "");
Expect(0, 68032, '\P{^merc}', "");
Expect(1, 68031, '\p{_	merc}', "");
Expect(0, 68031, '\p{^_	merc}', "");
Expect(0, 68031, '\P{_	merc}', "");
Expect(1, 68031, '\P{^_	merc}', "");
Expect(0, 68032, '\p{_	merc}', "");
Expect(1, 68032, '\p{^_	merc}', "");
Expect(1, 68032, '\P{_	merc}', "");
Expect(0, 68032, '\P{^_	merc}', "");
Error('\p{:= is_Merc}');
Error('\P{:= is_Merc}');
Expect(1, 68031, '\p{ismerc}', "");
Expect(0, 68031, '\p{^ismerc}', "");
Expect(0, 68031, '\P{ismerc}', "");
Expect(1, 68031, '\P{^ismerc}', "");
Expect(0, 68032, '\p{ismerc}', "");
Expect(1, 68032, '\p{^ismerc}', "");
Expect(1, 68032, '\P{ismerc}', "");
Expect(0, 68032, '\P{^ismerc}', "");
Expect(1, 68031, '\p{  is_MERC}', "");
Expect(0, 68031, '\p{^  is_MERC}', "");
Expect(0, 68031, '\P{  is_MERC}', "");
Expect(1, 68031, '\P{^  is_MERC}', "");
Expect(0, 68032, '\p{  is_MERC}', "");
Expect(1, 68032, '\p{^  is_MERC}', "");
Expect(1, 68032, '\P{  is_MERC}', "");
Expect(0, 68032, '\P{^  is_MERC}', "");
Error('\p{/a/-combining_DIACRITICAL_MARKS}');
Error('\P{/a/-combining_DIACRITICAL_MARKS}');
Expect(1, 879, '\p{combiningdiacriticalmarks}', "");
Expect(0, 879, '\p{^combiningdiacriticalmarks}', "");
Expect(0, 879, '\P{combiningdiacriticalmarks}', "");
Expect(1, 879, '\P{^combiningdiacriticalmarks}', "");
Expect(0, 880, '\p{combiningdiacriticalmarks}', "");
Expect(1, 880, '\p{^combiningdiacriticalmarks}', "");
Expect(1, 880, '\P{combiningdiacriticalmarks}', "");
Expect(0, 880, '\P{^combiningdiacriticalmarks}', "");
Expect(1, 879, '\p{_COMBINING_Diacritical_MARKS}', "");
Expect(0, 879, '\p{^_COMBINING_Diacritical_MARKS}', "");
Expect(0, 879, '\P{_COMBINING_Diacritical_MARKS}', "");
Expect(1, 879, '\P{^_COMBINING_Diacritical_MARKS}', "");
Expect(0, 880, '\p{_COMBINING_Diacritical_MARKS}', "");
Expect(1, 880, '\p{^_COMBINING_Diacritical_MARKS}', "");
Expect(1, 880, '\P{_COMBINING_Diacritical_MARKS}', "");
Expect(0, 880, '\P{^_COMBINING_Diacritical_MARKS}', "");
Error('\p{:=IS_Combining_diacritical_Marks}');
Error('\P{:=IS_Combining_diacritical_Marks}');
Expect(1, 879, '\p{iscombiningdiacriticalmarks}', "");
Expect(0, 879, '\p{^iscombiningdiacriticalmarks}', "");
Expect(0, 879, '\P{iscombiningdiacriticalmarks}', "");
Expect(1, 879, '\P{^iscombiningdiacriticalmarks}', "");
Expect(0, 880, '\p{iscombiningdiacriticalmarks}', "");
Expect(1, 880, '\p{^iscombiningdiacriticalmarks}', "");
Expect(1, 880, '\P{iscombiningdiacriticalmarks}', "");
Expect(0, 880, '\P{^iscombiningdiacriticalmarks}', "");
Expect(1, 879, '\p{  Is_combining_Diacritical_MARKS}', "");
Expect(0, 879, '\p{^  Is_combining_Diacritical_MARKS}', "");
Expect(0, 879, '\P{  Is_combining_Diacritical_MARKS}', "");
Expect(1, 879, '\P{^  Is_combining_Diacritical_MARKS}', "");
Expect(0, 880, '\p{  Is_combining_Diacritical_MARKS}', "");
Expect(1, 880, '\p{^  Is_combining_Diacritical_MARKS}', "");
Expect(1, 880, '\P{  Is_combining_Diacritical_MARKS}', "");
Expect(0, 880, '\P{^  Is_combining_Diacritical_MARKS}', "");
Error('\p{/a/__In_combining_Diacritical_Marks}');
Error('\P{/a/__In_combining_Diacritical_Marks}');
Expect(1, 879, '\p{incombiningdiacriticalmarks}', "");
Expect(0, 879, '\p{^incombiningdiacriticalmarks}', "");
Expect(0, 879, '\P{incombiningdiacriticalmarks}', "");
Expect(1, 879, '\P{^incombiningdiacriticalmarks}', "");
Expect(0, 880, '\p{incombiningdiacriticalmarks}', "");
Expect(1, 880, '\p{^incombiningdiacriticalmarks}', "");
Expect(1, 880, '\P{incombiningdiacriticalmarks}', "");
Expect(0, 880, '\P{^incombiningdiacriticalmarks}', "");
Expect(1, 879, '\p{-IN_Combining_Diacritical_marks}', "");
Expect(0, 879, '\p{^-IN_Combining_Diacritical_marks}', "");
Expect(0, 879, '\P{-IN_Combining_Diacritical_marks}', "");
Expect(1, 879, '\P{^-IN_Combining_Diacritical_marks}', "");
Expect(0, 880, '\p{-IN_Combining_Diacritical_marks}', "");
Expect(1, 880, '\p{^-IN_Combining_Diacritical_marks}', "");
Expect(1, 880, '\P{-IN_Combining_Diacritical_marks}', "");
Expect(0, 880, '\P{^-IN_Combining_Diacritical_marks}', "");
Error('\p{ Diacriticals:=}');
Error('\P{ Diacriticals:=}');
Expect(1, 879, '\p{diacriticals}', "");
Expect(0, 879, '\p{^diacriticals}', "");
Expect(0, 879, '\P{diacriticals}', "");
Expect(1, 879, '\P{^diacriticals}', "");
Expect(0, 880, '\p{diacriticals}', "");
Expect(1, 880, '\p{^diacriticals}', "");
Expect(1, 880, '\P{diacriticals}', "");
Expect(0, 880, '\P{^diacriticals}', "");
Expect(1, 879, '\p{	Diacriticals}', "");
Expect(0, 879, '\p{^	Diacriticals}', "");
Expect(0, 879, '\P{	Diacriticals}', "");
Expect(1, 879, '\P{^	Diacriticals}', "");
Expect(0, 880, '\p{	Diacriticals}', "");
Expect(1, 880, '\p{^	Diacriticals}', "");
Expect(1, 880, '\P{	Diacriticals}', "");
Expect(0, 880, '\P{^	Diacriticals}', "");
Error('\p{__Is_Diacriticals/a/}');
Error('\P{__Is_Diacriticals/a/}');
Expect(1, 879, '\p{isdiacriticals}', "");
Expect(0, 879, '\p{^isdiacriticals}', "");
Expect(0, 879, '\P{isdiacriticals}', "");
Expect(1, 879, '\P{^isdiacriticals}', "");
Expect(0, 880, '\p{isdiacriticals}', "");
Expect(1, 880, '\p{^isdiacriticals}', "");
Expect(1, 880, '\P{isdiacriticals}', "");
Expect(0, 880, '\P{^isdiacriticals}', "");
Expect(1, 879, '\p{	_IS_DIACRITICALS}', "");
Expect(0, 879, '\p{^	_IS_DIACRITICALS}', "");
Expect(0, 879, '\P{	_IS_DIACRITICALS}', "");
Expect(1, 879, '\P{^	_IS_DIACRITICALS}', "");
Expect(0, 880, '\p{	_IS_DIACRITICALS}', "");
Expect(1, 880, '\p{^	_IS_DIACRITICALS}', "");
Expect(1, 880, '\P{	_IS_DIACRITICALS}', "");
Expect(0, 880, '\P{^	_IS_DIACRITICALS}', "");
Error('\p{:=--In_Diacriticals}');
Error('\P{:=--In_Diacriticals}');
Expect(1, 879, '\p{indiacriticals}', "");
Expect(0, 879, '\p{^indiacriticals}', "");
Expect(0, 879, '\P{indiacriticals}', "");
Expect(1, 879, '\P{^indiacriticals}', "");
Expect(0, 880, '\p{indiacriticals}', "");
Expect(1, 880, '\p{^indiacriticals}', "");
Expect(1, 880, '\P{indiacriticals}', "");
Expect(0, 880, '\P{^indiacriticals}', "");
Expect(1, 879, '\p{In_Diacriticals}', "");
Expect(0, 879, '\p{^In_Diacriticals}', "");
Expect(0, 879, '\P{In_Diacriticals}', "");
Expect(1, 879, '\P{^In_Diacriticals}', "");
Expect(0, 880, '\p{In_Diacriticals}', "");
Expect(1, 880, '\p{^In_Diacriticals}', "");
Expect(1, 880, '\P{In_Diacriticals}', "");
Expect(0, 880, '\P{^In_Diacriticals}', "");
Error('\p{-/a/Inscriptional_Parthian}');
Error('\P{-/a/Inscriptional_Parthian}');
Expect(1, 68447, '\p{inscriptionalparthian}', "");
Expect(0, 68447, '\p{^inscriptionalparthian}', "");
Expect(0, 68447, '\P{inscriptionalparthian}', "");
Expect(1, 68447, '\P{^inscriptionalparthian}', "");
Expect(0, 68448, '\p{inscriptionalparthian}', "");
Expect(1, 68448, '\p{^inscriptionalparthian}', "");
Expect(1, 68448, '\P{inscriptionalparthian}', "");
Expect(0, 68448, '\P{^inscriptionalparthian}', "");
Expect(1, 68447, '\p{ -INSCRIPTIONAL_Parthian}', "");
Expect(0, 68447, '\p{^ -INSCRIPTIONAL_Parthian}', "");
Expect(0, 68447, '\P{ -INSCRIPTIONAL_Parthian}', "");
Expect(1, 68447, '\P{^ -INSCRIPTIONAL_Parthian}', "");
Expect(0, 68448, '\p{ -INSCRIPTIONAL_Parthian}', "");
Expect(1, 68448, '\p{^ -INSCRIPTIONAL_Parthian}', "");
Expect(1, 68448, '\P{ -INSCRIPTIONAL_Parthian}', "");
Expect(0, 68448, '\P{^ -INSCRIPTIONAL_Parthian}', "");
Error('\p{ /a/is_Inscriptional_PARTHIAN}');
Error('\P{ /a/is_Inscriptional_PARTHIAN}');
Expect(1, 68447, '\p{isinscriptionalparthian}', "");
Expect(0, 68447, '\p{^isinscriptionalparthian}', "");
Expect(0, 68447, '\P{isinscriptionalparthian}', "");
Expect(1, 68447, '\P{^isinscriptionalparthian}', "");
Expect(0, 68448, '\p{isinscriptionalparthian}', "");
Expect(1, 68448, '\p{^isinscriptionalparthian}', "");
Expect(1, 68448, '\P{isinscriptionalparthian}', "");
Expect(0, 68448, '\P{^isinscriptionalparthian}', "");
Expect(1, 68447, '\p{__is_INSCRIPTIONAL_PARTHIAN}', "");
Expect(0, 68447, '\p{^__is_INSCRIPTIONAL_PARTHIAN}', "");
Expect(0, 68447, '\P{__is_INSCRIPTIONAL_PARTHIAN}', "");
Expect(1, 68447, '\P{^__is_INSCRIPTIONAL_PARTHIAN}', "");
Expect(0, 68448, '\p{__is_INSCRIPTIONAL_PARTHIAN}', "");
Expect(1, 68448, '\p{^__is_INSCRIPTIONAL_PARTHIAN}', "");
Expect(1, 68448, '\P{__is_INSCRIPTIONAL_PARTHIAN}', "");
Expect(0, 68448, '\P{^__is_INSCRIPTIONAL_PARTHIAN}', "");
Error('\p{:=_-prti}');
Error('\P{:=_-prti}');
Expect(1, 68447, '\p{prti}', "");
Expect(0, 68447, '\p{^prti}', "");
Expect(0, 68447, '\P{prti}', "");
Expect(1, 68447, '\P{^prti}', "");
Expect(0, 68448, '\p{prti}', "");
Expect(1, 68448, '\p{^prti}', "");
Expect(1, 68448, '\P{prti}', "");
Expect(0, 68448, '\P{^prti}', "");
Expect(1, 68447, '\p{	 PRTI}', "");
Expect(0, 68447, '\p{^	 PRTI}', "");
Expect(0, 68447, '\P{	 PRTI}', "");
Expect(1, 68447, '\P{^	 PRTI}', "");
Expect(0, 68448, '\p{	 PRTI}', "");
Expect(1, 68448, '\p{^	 PRTI}', "");
Expect(1, 68448, '\P{	 PRTI}', "");
Expect(0, 68448, '\P{^	 PRTI}', "");
Error('\p{:=	_Is_PRTI}');
Error('\P{:=	_Is_PRTI}');
Expect(1, 68447, '\p{isprti}', "");
Expect(0, 68447, '\p{^isprti}', "");
Expect(0, 68447, '\P{isprti}', "");
Expect(1, 68447, '\P{^isprti}', "");
Expect(0, 68448, '\p{isprti}', "");
Expect(1, 68448, '\p{^isprti}', "");
Expect(1, 68448, '\P{isprti}', "");
Expect(0, 68448, '\P{^isprti}', "");
Expect(1, 68447, '\p{ Is_Prti}', "");
Expect(0, 68447, '\p{^ Is_Prti}', "");
Expect(0, 68447, '\P{ Is_Prti}', "");
Expect(1, 68447, '\P{^ Is_Prti}', "");
Expect(0, 68448, '\p{ Is_Prti}', "");
Expect(1, 68448, '\p{^ Is_Prti}', "");
Expect(1, 68448, '\P{ Is_Prti}', "");
Expect(0, 68448, '\P{^ Is_Prti}', "");
Error('\p{_-Devanagari_Extended:=}');
Error('\P{_-Devanagari_Extended:=}');
Expect(1, 43263, '\p{devanagariextended}', "");
Expect(0, 43263, '\p{^devanagariextended}', "");
Expect(0, 43263, '\P{devanagariextended}', "");
Expect(1, 43263, '\P{^devanagariextended}', "");
Expect(0, 43264, '\p{devanagariextended}', "");
Expect(1, 43264, '\p{^devanagariextended}', "");
Expect(1, 43264, '\P{devanagariextended}', "");
Expect(0, 43264, '\P{^devanagariextended}', "");
Expect(1, 43263, '\p{--DEVANAGARI_EXTENDED}', "");
Expect(0, 43263, '\p{^--DEVANAGARI_EXTENDED}', "");
Expect(0, 43263, '\P{--DEVANAGARI_EXTENDED}', "");
Expect(1, 43263, '\P{^--DEVANAGARI_EXTENDED}', "");
Expect(0, 43264, '\p{--DEVANAGARI_EXTENDED}', "");
Expect(1, 43264, '\p{^--DEVANAGARI_EXTENDED}', "");
Expect(1, 43264, '\P{--DEVANAGARI_EXTENDED}', "");
Expect(0, 43264, '\P{^--DEVANAGARI_EXTENDED}', "");
Error('\p{:=_-is_Devanagari_extended}');
Error('\P{:=_-is_Devanagari_extended}');
Expect(1, 43263, '\p{isdevanagariextended}', "");
Expect(0, 43263, '\p{^isdevanagariextended}', "");
Expect(0, 43263, '\P{isdevanagariextended}', "");
Expect(1, 43263, '\P{^isdevanagariextended}', "");
Expect(0, 43264, '\p{isdevanagariextended}', "");
Expect(1, 43264, '\p{^isdevanagariextended}', "");
Expect(1, 43264, '\P{isdevanagariextended}', "");
Expect(0, 43264, '\P{^isdevanagariextended}', "");
Expect(1, 43263, '\p{  Is_DEVANAGARI_EXTENDED}', "");
Expect(0, 43263, '\p{^  Is_DEVANAGARI_EXTENDED}', "");
Expect(0, 43263, '\P{  Is_DEVANAGARI_EXTENDED}', "");
Expect(1, 43263, '\P{^  Is_DEVANAGARI_EXTENDED}', "");
Expect(0, 43264, '\p{  Is_DEVANAGARI_EXTENDED}', "");
Expect(1, 43264, '\p{^  Is_DEVANAGARI_EXTENDED}', "");
Expect(1, 43264, '\P{  Is_DEVANAGARI_EXTENDED}', "");
Expect(0, 43264, '\P{^  Is_DEVANAGARI_EXTENDED}', "");
Error('\p{:=	-In_DEVANAGARI_Extended}');
Error('\P{:=	-In_DEVANAGARI_Extended}');
Expect(1, 43263, '\p{indevanagariextended}', "");
Expect(0, 43263, '\p{^indevanagariextended}', "");
Expect(0, 43263, '\P{indevanagariextended}', "");
Expect(1, 43263, '\P{^indevanagariextended}', "");
Expect(0, 43264, '\p{indevanagariextended}', "");
Expect(1, 43264, '\p{^indevanagariextended}', "");
Expect(1, 43264, '\P{indevanagariextended}', "");
Expect(0, 43264, '\P{^indevanagariextended}', "");
Expect(1, 43263, '\p{_ In_devanagari_EXTENDED}', "");
Expect(0, 43263, '\p{^_ In_devanagari_EXTENDED}', "");
Expect(0, 43263, '\P{_ In_devanagari_EXTENDED}', "");
Expect(1, 43263, '\P{^_ In_devanagari_EXTENDED}', "");
Expect(0, 43264, '\p{_ In_devanagari_EXTENDED}', "");
Expect(1, 43264, '\p{^_ In_devanagari_EXTENDED}', "");
Expect(1, 43264, '\P{_ In_devanagari_EXTENDED}', "");
Expect(0, 43264, '\P{^_ In_devanagari_EXTENDED}', "");
Error('\p{	/a/devanagari_Ext}');
Error('\P{	/a/devanagari_Ext}');
Expect(1, 43263, '\p{devanagariext}', "");
Expect(0, 43263, '\p{^devanagariext}', "");
Expect(0, 43263, '\P{devanagariext}', "");
Expect(1, 43263, '\P{^devanagariext}', "");
Expect(0, 43264, '\p{devanagariext}', "");
Expect(1, 43264, '\p{^devanagariext}', "");
Expect(1, 43264, '\P{devanagariext}', "");
Expect(0, 43264, '\P{^devanagariext}', "");
Expect(1, 43263, '\p{	 Devanagari_Ext}', "");
Expect(0, 43263, '\p{^	 Devanagari_Ext}', "");
Expect(0, 43263, '\P{	 Devanagari_Ext}', "");
Expect(1, 43263, '\P{^	 Devanagari_Ext}', "");
Expect(0, 43264, '\p{	 Devanagari_Ext}', "");
Expect(1, 43264, '\p{^	 Devanagari_Ext}', "");
Expect(1, 43264, '\P{	 Devanagari_Ext}', "");
Expect(0, 43264, '\P{^	 Devanagari_Ext}', "");
Error('\p{:=_-Is_Devanagari_Ext}');
Error('\P{:=_-Is_Devanagari_Ext}');
Expect(1, 43263, '\p{isdevanagariext}', "");
Expect(0, 43263, '\p{^isdevanagariext}', "");
Expect(0, 43263, '\P{isdevanagariext}', "");
Expect(1, 43263, '\P{^isdevanagariext}', "");
Expect(0, 43264, '\p{isdevanagariext}', "");
Expect(1, 43264, '\p{^isdevanagariext}', "");
Expect(1, 43264, '\P{isdevanagariext}', "");
Expect(0, 43264, '\P{^isdevanagariext}', "");
Expect(1, 43263, '\p{ Is_DEVANAGARI_EXT}', "");
Expect(0, 43263, '\p{^ Is_DEVANAGARI_EXT}', "");
Expect(0, 43263, '\P{ Is_DEVANAGARI_EXT}', "");
Expect(1, 43263, '\P{^ Is_DEVANAGARI_EXT}', "");
Expect(0, 43264, '\p{ Is_DEVANAGARI_EXT}', "");
Expect(1, 43264, '\p{^ Is_DEVANAGARI_EXT}', "");
Expect(1, 43264, '\P{ Is_DEVANAGARI_EXT}', "");
Expect(0, 43264, '\P{^ Is_DEVANAGARI_EXT}', "");
Error('\p{:=_	IN_Devanagari_Ext}');
Error('\P{:=_	IN_Devanagari_Ext}');
Expect(1, 43263, '\p{indevanagariext}', "");
Expect(0, 43263, '\p{^indevanagariext}', "");
Expect(0, 43263, '\P{indevanagariext}', "");
Expect(1, 43263, '\P{^indevanagariext}', "");
Expect(0, 43264, '\p{indevanagariext}', "");
Expect(1, 43264, '\p{^indevanagariext}', "");
Expect(1, 43264, '\P{indevanagariext}', "");
Expect(0, 43264, '\P{^indevanagariext}', "");
Expect(1, 43263, '\p{- in_DEVANAGARI_Ext}', "");
Expect(0, 43263, '\p{^- in_DEVANAGARI_Ext}', "");
Expect(0, 43263, '\P{- in_DEVANAGARI_Ext}', "");
Expect(1, 43263, '\P{^- in_DEVANAGARI_Ext}', "");
Expect(0, 43264, '\p{- in_DEVANAGARI_Ext}', "");
Expect(1, 43264, '\p{^- in_DEVANAGARI_Ext}', "");
Expect(1, 43264, '\P{- in_DEVANAGARI_Ext}', "");
Expect(0, 43264, '\P{^- in_DEVANAGARI_Ext}', "");
Error('\p{	Ethiopic_SUPPLEMENT:=}');
Error('\P{	Ethiopic_SUPPLEMENT:=}');
Expect(1, 5023, '\p{ethiopicsupplement}', "");
Expect(0, 5023, '\p{^ethiopicsupplement}', "");
Expect(0, 5023, '\P{ethiopicsupplement}', "");
Expect(1, 5023, '\P{^ethiopicsupplement}', "");
Expect(0, 5024, '\p{ethiopicsupplement}', "");
Expect(1, 5024, '\p{^ethiopicsupplement}', "");
Expect(1, 5024, '\P{ethiopicsupplement}', "");
Expect(0, 5024, '\P{^ethiopicsupplement}', "");
Expect(1, 5023, '\p{-ethiopic_Supplement}', "");
Expect(0, 5023, '\p{^-ethiopic_Supplement}', "");
Expect(0, 5023, '\P{-ethiopic_Supplement}', "");
Expect(1, 5023, '\P{^-ethiopic_Supplement}', "");
Expect(0, 5024, '\p{-ethiopic_Supplement}', "");
Expect(1, 5024, '\p{^-ethiopic_Supplement}', "");
Expect(1, 5024, '\P{-ethiopic_Supplement}', "");
Expect(0, 5024, '\P{^-ethiopic_Supplement}', "");
Error('\p{	/a/IS_Ethiopic_supplement}');
Error('\P{	/a/IS_Ethiopic_supplement}');
Expect(1, 5023, '\p{isethiopicsupplement}', "");
Expect(0, 5023, '\p{^isethiopicsupplement}', "");
Expect(0, 5023, '\P{isethiopicsupplement}', "");
Expect(1, 5023, '\P{^isethiopicsupplement}', "");
Expect(0, 5024, '\p{isethiopicsupplement}', "");
Expect(1, 5024, '\p{^isethiopicsupplement}', "");
Expect(1, 5024, '\P{isethiopicsupplement}', "");
Expect(0, 5024, '\P{^isethiopicsupplement}', "");
Expect(1, 5023, '\p{-Is_Ethiopic_Supplement}', "");
Expect(0, 5023, '\p{^-Is_Ethiopic_Supplement}', "");
Expect(0, 5023, '\P{-Is_Ethiopic_Supplement}', "");
Expect(1, 5023, '\P{^-Is_Ethiopic_Supplement}', "");
Expect(0, 5024, '\p{-Is_Ethiopic_Supplement}', "");
Expect(1, 5024, '\p{^-Is_Ethiopic_Supplement}', "");
Expect(1, 5024, '\P{-Is_Ethiopic_Supplement}', "");
Expect(0, 5024, '\P{^-Is_Ethiopic_Supplement}', "");
Error('\p{:=	 In_Ethiopic_supplement}');
Error('\P{:=	 In_Ethiopic_supplement}');
Expect(1, 5023, '\p{inethiopicsupplement}', "");
Expect(0, 5023, '\p{^inethiopicsupplement}', "");
Expect(0, 5023, '\P{inethiopicsupplement}', "");
Expect(1, 5023, '\P{^inethiopicsupplement}', "");
Expect(0, 5024, '\p{inethiopicsupplement}', "");
Expect(1, 5024, '\p{^inethiopicsupplement}', "");
Expect(1, 5024, '\P{inethiopicsupplement}', "");
Expect(0, 5024, '\P{^inethiopicsupplement}', "");
Expect(1, 5023, '\p{		In_Ethiopic_supplement}', "");
Expect(0, 5023, '\p{^		In_Ethiopic_supplement}', "");
Expect(0, 5023, '\P{		In_Ethiopic_supplement}', "");
Expect(1, 5023, '\P{^		In_Ethiopic_supplement}', "");
Expect(0, 5024, '\p{		In_Ethiopic_supplement}', "");
Expect(1, 5024, '\p{^		In_Ethiopic_supplement}', "");
Expect(1, 5024, '\P{		In_Ethiopic_supplement}', "");
Expect(0, 5024, '\P{^		In_Ethiopic_supplement}', "");
Error('\p{/a/-_Ethiopic_Sup}');
Error('\P{/a/-_Ethiopic_Sup}');
Expect(1, 5023, '\p{ethiopicsup}', "");
Expect(0, 5023, '\p{^ethiopicsup}', "");
Expect(0, 5023, '\P{ethiopicsup}', "");
Expect(1, 5023, '\P{^ethiopicsup}', "");
Expect(0, 5024, '\p{ethiopicsup}', "");
Expect(1, 5024, '\p{^ethiopicsup}', "");
Expect(1, 5024, '\P{ethiopicsup}', "");
Expect(0, 5024, '\P{^ethiopicsup}', "");
Expect(1, 5023, '\p{	-ethiopic_Sup}', "");
Expect(0, 5023, '\p{^	-ethiopic_Sup}', "");
Expect(0, 5023, '\P{	-ethiopic_Sup}', "");
Expect(1, 5023, '\P{^	-ethiopic_Sup}', "");
Expect(0, 5024, '\p{	-ethiopic_Sup}', "");
Expect(1, 5024, '\p{^	-ethiopic_Sup}', "");
Expect(1, 5024, '\P{	-ethiopic_Sup}', "");
Expect(0, 5024, '\P{^	-ethiopic_Sup}', "");
Error('\p{/a/_Is_ETHIOPIC_Sup}');
Error('\P{/a/_Is_ETHIOPIC_Sup}');
Expect(1, 5023, '\p{isethiopicsup}', "");
Expect(0, 5023, '\p{^isethiopicsup}', "");
Expect(0, 5023, '\P{isethiopicsup}', "");
Expect(1, 5023, '\P{^isethiopicsup}', "");
Expect(0, 5024, '\p{isethiopicsup}', "");
Expect(1, 5024, '\p{^isethiopicsup}', "");
Expect(1, 5024, '\P{isethiopicsup}', "");
Expect(0, 5024, '\P{^isethiopicsup}', "");
Expect(1, 5023, '\p{_-IS_ethiopic_Sup}', "");
Expect(0, 5023, '\p{^_-IS_ethiopic_Sup}', "");
Expect(0, 5023, '\P{_-IS_ethiopic_Sup}', "");
Expect(1, 5023, '\P{^_-IS_ethiopic_Sup}', "");
Expect(0, 5024, '\p{_-IS_ethiopic_Sup}', "");
Expect(1, 5024, '\p{^_-IS_ethiopic_Sup}', "");
Expect(1, 5024, '\P{_-IS_ethiopic_Sup}', "");
Expect(0, 5024, '\P{^_-IS_ethiopic_Sup}', "");
Error('\p{/a/	_In_ethiopic_sup}');
Error('\P{/a/	_In_ethiopic_sup}');
Expect(1, 5023, '\p{inethiopicsup}', "");
Expect(0, 5023, '\p{^inethiopicsup}', "");
Expect(0, 5023, '\P{inethiopicsup}', "");
Expect(1, 5023, '\P{^inethiopicsup}', "");
Expect(0, 5024, '\p{inethiopicsup}', "");
Expect(1, 5024, '\p{^inethiopicsup}', "");
Expect(1, 5024, '\P{inethiopicsup}', "");
Expect(0, 5024, '\P{^inethiopicsup}', "");
Expect(1, 5023, '\p{-In_Ethiopic_Sup}', "");
Expect(0, 5023, '\p{^-In_Ethiopic_Sup}', "");
Expect(0, 5023, '\P{-In_Ethiopic_Sup}', "");
Expect(1, 5023, '\P{^-In_Ethiopic_Sup}', "");
Expect(0, 5024, '\p{-In_Ethiopic_Sup}', "");
Expect(1, 5024, '\p{^-In_Ethiopic_Sup}', "");
Expect(1, 5024, '\P{-In_Ethiopic_Sup}', "");
Expect(0, 5024, '\P{^-In_Ethiopic_Sup}', "");
Error('\p{--Dingbats/a/}');
Error('\P{--Dingbats/a/}');
Expect(1, 10175, '\p{dingbats}', "");
Expect(0, 10175, '\p{^dingbats}', "");
Expect(0, 10175, '\P{dingbats}', "");
Expect(1, 10175, '\P{^dingbats}', "");
Expect(0, 10176, '\p{dingbats}', "");
Expect(1, 10176, '\p{^dingbats}', "");
Expect(1, 10176, '\P{dingbats}', "");
Expect(0, 10176, '\P{^dingbats}', "");
Expect(1, 10175, '\p{_Dingbats}', "");
Expect(0, 10175, '\p{^_Dingbats}', "");
Expect(0, 10175, '\P{_Dingbats}', "");
Expect(1, 10175, '\P{^_Dingbats}', "");
Expect(0, 10176, '\p{_Dingbats}', "");
Expect(1, 10176, '\p{^_Dingbats}', "");
Expect(1, 10176, '\P{_Dingbats}', "");
Expect(0, 10176, '\P{^_Dingbats}', "");
Error('\p{/a/ -Is_dingbats}');
Error('\P{/a/ -Is_dingbats}');
Expect(1, 10175, '\p{isdingbats}', "");
Expect(0, 10175, '\p{^isdingbats}', "");
Expect(0, 10175, '\P{isdingbats}', "");
Expect(1, 10175, '\P{^isdingbats}', "");
Expect(0, 10176, '\p{isdingbats}', "");
Expect(1, 10176, '\p{^isdingbats}', "");
Expect(1, 10176, '\P{isdingbats}', "");
Expect(0, 10176, '\P{^isdingbats}', "");
Expect(1, 10175, '\p{-	is_dingbats}', "");
Expect(0, 10175, '\p{^-	is_dingbats}', "");
Expect(0, 10175, '\P{-	is_dingbats}', "");
Expect(1, 10175, '\P{^-	is_dingbats}', "");
Expect(0, 10176, '\p{-	is_dingbats}', "");
Expect(1, 10176, '\p{^-	is_dingbats}', "");
Expect(1, 10176, '\P{-	is_dingbats}', "");
Expect(0, 10176, '\P{^-	is_dingbats}', "");
Error('\p{ In_dingbats/a/}');
Error('\P{ In_dingbats/a/}');
Expect(1, 10175, '\p{indingbats}', "");
Expect(0, 10175, '\p{^indingbats}', "");
Expect(0, 10175, '\P{indingbats}', "");
Expect(1, 10175, '\P{^indingbats}', "");
Expect(0, 10176, '\p{indingbats}', "");
Expect(1, 10176, '\p{^indingbats}', "");
Expect(1, 10176, '\P{indingbats}', "");
Expect(0, 10176, '\P{^indingbats}', "");
Expect(1, 10175, '\p{--In_DINGBATS}', "");
Expect(0, 10175, '\p{^--In_DINGBATS}', "");
Expect(0, 10175, '\P{--In_DINGBATS}', "");
Expect(1, 10175, '\P{^--In_DINGBATS}', "");
Expect(0, 10176, '\p{--In_DINGBATS}', "");
Expect(1, 10176, '\p{^--In_DINGBATS}', "");
Expect(1, 10176, '\P{--In_DINGBATS}', "");
Expect(0, 10176, '\P{^--In_DINGBATS}', "");
Error('\p{/a/	Logical_ORDER_Exception}');
Error('\P{/a/	Logical_ORDER_Exception}');
Expect(1, 43708, '\p{logicalorderexception}', "");
Expect(0, 43708, '\p{^logicalorderexception}', "");
Expect(0, 43708, '\P{logicalorderexception}', "");
Expect(1, 43708, '\P{^logicalorderexception}', "");
Expect(0, 43709, '\p{logicalorderexception}', "");
Expect(1, 43709, '\p{^logicalorderexception}', "");
Expect(1, 43709, '\P{logicalorderexception}', "");
Expect(0, 43709, '\P{^logicalorderexception}', "");
Expect(1, 43708, '\p{ -LOGICAL_order_EXCEPTION}', "");
Expect(0, 43708, '\p{^ -LOGICAL_order_EXCEPTION}', "");
Expect(0, 43708, '\P{ -LOGICAL_order_EXCEPTION}', "");
Expect(1, 43708, '\P{^ -LOGICAL_order_EXCEPTION}', "");
Expect(0, 43709, '\p{ -LOGICAL_order_EXCEPTION}', "");
Expect(1, 43709, '\p{^ -LOGICAL_order_EXCEPTION}', "");
Expect(1, 43709, '\P{ -LOGICAL_order_EXCEPTION}', "");
Expect(0, 43709, '\P{^ -LOGICAL_order_EXCEPTION}', "");
Error('\p{-_Is_logical_Order_Exception:=}');
Error('\P{-_Is_logical_Order_Exception:=}');
Expect(1, 43708, '\p{islogicalorderexception}', "");
Expect(0, 43708, '\p{^islogicalorderexception}', "");
Expect(0, 43708, '\P{islogicalorderexception}', "");
Expect(1, 43708, '\P{^islogicalorderexception}', "");
Expect(0, 43709, '\p{islogicalorderexception}', "");
Expect(1, 43709, '\p{^islogicalorderexception}', "");
Expect(1, 43709, '\P{islogicalorderexception}', "");
Expect(0, 43709, '\P{^islogicalorderexception}', "");
Expect(1, 43708, '\p{-Is_Logical_ORDER_Exception}', "");
Expect(0, 43708, '\p{^-Is_Logical_ORDER_Exception}', "");
Expect(0, 43708, '\P{-Is_Logical_ORDER_Exception}', "");
Expect(1, 43708, '\P{^-Is_Logical_ORDER_Exception}', "");
Expect(0, 43709, '\p{-Is_Logical_ORDER_Exception}', "");
Expect(1, 43709, '\p{^-Is_Logical_ORDER_Exception}', "");
Expect(1, 43709, '\P{-Is_Logical_ORDER_Exception}', "");
Expect(0, 43709, '\P{^-Is_Logical_ORDER_Exception}', "");
Error('\p{/a/LOE}');
Error('\P{/a/LOE}');
Expect(1, 43708, '\p{loe}', "");
Expect(0, 43708, '\p{^loe}', "");
Expect(0, 43708, '\P{loe}', "");
Expect(1, 43708, '\P{^loe}', "");
Expect(0, 43709, '\p{loe}', "");
Expect(1, 43709, '\p{^loe}', "");
Expect(1, 43709, '\P{loe}', "");
Expect(0, 43709, '\P{^loe}', "");
Expect(1, 43708, '\p{	-loe}', "");
Expect(0, 43708, '\p{^	-loe}', "");
Expect(0, 43708, '\P{	-loe}', "");
Expect(1, 43708, '\P{^	-loe}', "");
Expect(0, 43709, '\p{	-loe}', "");
Expect(1, 43709, '\p{^	-loe}', "");
Expect(1, 43709, '\P{	-loe}', "");
Expect(0, 43709, '\P{^	-loe}', "");
Error('\p{-Is_LOE:=}');
Error('\P{-Is_LOE:=}');
Expect(1, 43708, '\p{isloe}', "");
Expect(0, 43708, '\p{^isloe}', "");
Expect(0, 43708, '\P{isloe}', "");
Expect(1, 43708, '\P{^isloe}', "");
Expect(0, 43709, '\p{isloe}', "");
Expect(1, 43709, '\p{^isloe}', "");
Expect(1, 43709, '\P{isloe}', "");
Expect(0, 43709, '\P{^isloe}', "");
Expect(1, 43708, '\p{  IS_loe}', "");
Expect(0, 43708, '\p{^  IS_loe}', "");
Expect(0, 43708, '\P{  IS_loe}', "");
Expect(1, 43708, '\P{^  IS_loe}', "");
Expect(0, 43709, '\p{  IS_loe}', "");
Expect(1, 43709, '\p{^  IS_loe}', "");
Expect(1, 43709, '\P{  IS_loe}', "");
Expect(0, 43709, '\P{^  IS_loe}', "");
Error('\p{/a/In_Lycian}');
Error('\P{/a/In_Lycian}');
Expect(1, 66207, '\p{inlycian}', "");
Expect(0, 66207, '\p{^inlycian}', "");
Expect(0, 66207, '\P{inlycian}', "");
Expect(1, 66207, '\P{^inlycian}', "");
Expect(0, 66208, '\p{inlycian}', "");
Expect(1, 66208, '\p{^inlycian}', "");
Expect(1, 66208, '\P{inlycian}', "");
Expect(0, 66208, '\P{^inlycian}', "");
Expect(1, 66207, '\p{	In_LYCIAN}', "");
Expect(0, 66207, '\p{^	In_LYCIAN}', "");
Expect(0, 66207, '\P{	In_LYCIAN}', "");
Expect(1, 66207, '\P{^	In_LYCIAN}', "");
Expect(0, 66208, '\p{	In_LYCIAN}', "");
Expect(1, 66208, '\p{^	In_LYCIAN}', "");
Expect(1, 66208, '\P{	In_LYCIAN}', "");
Expect(0, 66208, '\P{^	In_LYCIAN}', "");
Error('\p{_In_lycian:=}');
Error('\P{_In_lycian:=}');
Expect(1, 66207, '\p{-_In_lycian}', "");
Expect(0, 66207, '\p{^-_In_lycian}', "");
Expect(0, 66207, '\P{-_In_lycian}', "");
Expect(1, 66207, '\P{^-_In_lycian}', "");
Expect(0, 66208, '\p{-_In_lycian}', "");
Expect(1, 66208, '\p{^-_In_lycian}', "");
Expect(1, 66208, '\P{-_In_lycian}', "");
Expect(0, 66208, '\P{^-_In_lycian}', "");
Error('\p{/a/_ vai}');
Error('\P{/a/_ vai}');
Expect(1, 42539, '\p{vai}', "");
Expect(0, 42539, '\p{^vai}', "");
Expect(0, 42539, '\P{vai}', "");
Expect(1, 42539, '\P{^vai}', "");
Expect(0, 42540, '\p{vai}', "");
Expect(1, 42540, '\p{^vai}', "");
Expect(1, 42540, '\P{vai}', "");
Expect(0, 42540, '\P{^vai}', "");
Expect(1, 42539, '\p{_	VAI}', "");
Expect(0, 42539, '\p{^_	VAI}', "");
Expect(0, 42539, '\P{_	VAI}', "");
Expect(1, 42539, '\P{^_	VAI}', "");
Expect(0, 42540, '\p{_	VAI}', "");
Expect(1, 42540, '\p{^_	VAI}', "");
Expect(1, 42540, '\P{_	VAI}', "");
Expect(0, 42540, '\P{^_	VAI}', "");
Error('\p{:=	Is_VAI}');
Error('\P{:=	Is_VAI}');
Expect(1, 42539, '\p{isvai}', "");
Expect(0, 42539, '\p{^isvai}', "");
Expect(0, 42539, '\P{isvai}', "");
Expect(1, 42539, '\P{^isvai}', "");
Expect(0, 42540, '\p{isvai}', "");
Expect(1, 42540, '\p{^isvai}', "");
Expect(1, 42540, '\P{isvai}', "");
Expect(0, 42540, '\P{^isvai}', "");
Expect(1, 42539, '\p{-IS_Vai}', "");
Expect(0, 42539, '\p{^-IS_Vai}', "");
Expect(0, 42539, '\P{-IS_Vai}', "");
Expect(1, 42539, '\P{^-IS_Vai}', "");
Expect(0, 42540, '\p{-IS_Vai}', "");
Expect(1, 42540, '\p{^-IS_Vai}', "");
Expect(1, 42540, '\P{-IS_Vai}', "");
Expect(0, 42540, '\P{^-IS_Vai}', "");
Error('\p{/a/__Vaii}');
Error('\P{/a/__Vaii}');
Expect(1, 42539, '\p{vaii}', "");
Expect(0, 42539, '\p{^vaii}', "");
Expect(0, 42539, '\P{vaii}', "");
Expect(1, 42539, '\P{^vaii}', "");
Expect(0, 42540, '\p{vaii}', "");
Expect(1, 42540, '\p{^vaii}', "");
Expect(1, 42540, '\P{vaii}', "");
Expect(0, 42540, '\P{^vaii}', "");
Expect(1, 42539, '\p{	Vaii}', "");
Expect(0, 42539, '\p{^	Vaii}', "");
Expect(0, 42539, '\P{	Vaii}', "");
Expect(1, 42539, '\P{^	Vaii}', "");
Expect(0, 42540, '\p{	Vaii}', "");
Expect(1, 42540, '\p{^	Vaii}', "");
Expect(1, 42540, '\P{	Vaii}', "");
Expect(0, 42540, '\P{^	Vaii}', "");
Error('\p{ 	Is_Vaii:=}');
Error('\P{ 	Is_Vaii:=}');
Expect(1, 42539, '\p{isvaii}', "");
Expect(0, 42539, '\p{^isvaii}', "");
Expect(0, 42539, '\P{isvaii}', "");
Expect(1, 42539, '\P{^isvaii}', "");
Expect(0, 42540, '\p{isvaii}', "");
Expect(1, 42540, '\p{^isvaii}', "");
Expect(1, 42540, '\P{isvaii}', "");
Expect(0, 42540, '\P{^isvaii}', "");
Expect(1, 42539, '\p{ IS_VAII}', "");
Expect(0, 42539, '\p{^ IS_VAII}', "");
Expect(0, 42539, '\P{ IS_VAII}', "");
Expect(1, 42539, '\P{^ IS_VAII}', "");
Expect(0, 42540, '\p{ IS_VAII}', "");
Expect(1, 42540, '\p{^ IS_VAII}', "");
Expect(1, 42540, '\P{ IS_VAII}', "");
Expect(0, 42540, '\P{^ IS_VAII}', "");
Error('\p{__spacing_mark/a/}');
Error('\P{__spacing_mark/a/}');
Expect(1, 119154, '\p{spacingmark}', "");
Expect(0, 119154, '\p{^spacingmark}', "");
Expect(0, 119154, '\P{spacingmark}', "");
Expect(1, 119154, '\P{^spacingmark}', "");
Expect(0, 119155, '\p{spacingmark}', "");
Expect(1, 119155, '\p{^spacingmark}', "");
Expect(1, 119155, '\P{spacingmark}', "");
Expect(0, 119155, '\P{^spacingmark}', "");
Expect(1, 119154, '\p{	Spacing_MARK}', "");
Expect(0, 119154, '\p{^	Spacing_MARK}', "");
Expect(0, 119154, '\P{	Spacing_MARK}', "");
Expect(1, 119154, '\P{^	Spacing_MARK}', "");
Expect(0, 119155, '\p{	Spacing_MARK}', "");
Expect(1, 119155, '\p{^	Spacing_MARK}', "");
Expect(1, 119155, '\P{	Spacing_MARK}', "");
Expect(0, 119155, '\P{^	Spacing_MARK}', "");
Error('\p{:=_ Is_Spacing_mark}');
Error('\P{:=_ Is_Spacing_mark}');
Expect(1, 119154, '\p{isspacingmark}', "");
Expect(0, 119154, '\p{^isspacingmark}', "");
Expect(0, 119154, '\P{isspacingmark}', "");
Expect(1, 119154, '\P{^isspacingmark}', "");
Expect(0, 119155, '\p{isspacingmark}', "");
Expect(1, 119155, '\p{^isspacingmark}', "");
Expect(1, 119155, '\P{isspacingmark}', "");
Expect(0, 119155, '\P{^isspacingmark}', "");
Expect(1, 119154, '\p{--Is_Spacing_Mark}', "");
Expect(0, 119154, '\p{^--Is_Spacing_Mark}', "");
Expect(0, 119154, '\P{--Is_Spacing_Mark}', "");
Expect(1, 119154, '\P{^--Is_Spacing_Mark}', "");
Expect(0, 119155, '\p{--Is_Spacing_Mark}', "");
Expect(1, 119155, '\p{^--Is_Spacing_Mark}', "");
Expect(1, 119155, '\P{--Is_Spacing_Mark}', "");
Expect(0, 119155, '\P{^--Is_Spacing_Mark}', "");
Error('\p{_Mc:=}');
Error('\P{_Mc:=}');
Expect(1, 119154, '\p{mc}', "");
Expect(0, 119154, '\p{^mc}', "");
Expect(0, 119154, '\P{mc}', "");
Expect(1, 119154, '\P{^mc}', "");
Expect(0, 119155, '\p{mc}', "");
Expect(1, 119155, '\p{^mc}', "");
Expect(1, 119155, '\P{mc}', "");
Expect(0, 119155, '\P{^mc}', "");
Expect(1, 119154, '\p{- mc}', "");
Expect(0, 119154, '\p{^- mc}', "");
Expect(0, 119154, '\P{- mc}', "");
Expect(1, 119154, '\P{^- mc}', "");
Expect(0, 119155, '\p{- mc}', "");
Expect(1, 119155, '\p{^- mc}', "");
Expect(1, 119155, '\P{- mc}', "");
Expect(0, 119155, '\P{^- mc}', "");
Error('\p{	 is_mc:=}');
Error('\P{	 is_mc:=}');
Expect(1, 119154, '\p{ismc}', "");
Expect(0, 119154, '\p{^ismc}', "");
Expect(0, 119154, '\P{ismc}', "");
Expect(1, 119154, '\P{^ismc}', "");
Expect(0, 119155, '\p{ismc}', "");
Expect(1, 119155, '\p{^ismc}', "");
Expect(1, 119155, '\P{ismc}', "");
Expect(0, 119155, '\P{^ismc}', "");
Expect(1, 119154, '\p{-_Is_MC}', "");
Expect(0, 119154, '\p{^-_Is_MC}', "");
Expect(0, 119154, '\P{-_Is_MC}', "");
Expect(1, 119154, '\P{^-_Is_MC}', "");
Expect(0, 119155, '\p{-_Is_MC}', "");
Expect(1, 119155, '\p{^-_Is_MC}', "");
Expect(1, 119155, '\P{-_Is_MC}', "");
Expect(0, 119155, '\P{^-_Is_MC}', "");
Error('\p{	Lower/a/}');
Error('\P{	Lower/a/}');
Expect(1, 120779, '\p{lower}', "");
Expect(0, 120779, '\p{^lower}', "");
Expect(0, 120779, '\P{lower}', "");
Expect(1, 120779, '\P{^lower}', "");
Expect(0, 120780, '\p{lower}', "");
Expect(1, 120780, '\p{^lower}', "");
Expect(1, 120780, '\P{lower}', "");
Expect(0, 120780, '\P{^lower}', "");
Expect(1, 120779, '\p{_Lower}', "");
Expect(0, 120779, '\p{^_Lower}', "");
Expect(0, 120779, '\P{_Lower}', "");
Expect(1, 120779, '\P{^_Lower}', "");
Expect(0, 120780, '\p{_Lower}', "");
Expect(1, 120780, '\p{^_Lower}', "");
Expect(1, 120780, '\P{_Lower}', "");
Expect(0, 120780, '\P{^_Lower}', "");
Error('\p{ :=XPOSIXLOWER}');
Error('\P{ :=XPOSIXLOWER}');
Expect(1, 120779, '\p{xposixlower}', "");
Expect(0, 120779, '\p{^xposixlower}', "");
Expect(0, 120779, '\P{xposixlower}', "");
Expect(1, 120779, '\P{^xposixlower}', "");
Expect(0, 120780, '\p{xposixlower}', "");
Expect(1, 120780, '\p{^xposixlower}', "");
Expect(1, 120780, '\P{xposixlower}', "");
Expect(0, 120780, '\P{^xposixlower}', "");
Expect(1, 120779, '\p{--XPosixLower}', "");
Expect(0, 120779, '\p{^--XPosixLower}', "");
Expect(0, 120779, '\P{--XPosixLower}', "");
Expect(1, 120779, '\P{^--XPosixLower}', "");
Expect(0, 120780, '\p{--XPosixLower}', "");
Expect(1, 120780, '\p{^--XPosixLower}', "");
Expect(1, 120780, '\P{--XPosixLower}', "");
Expect(0, 120780, '\P{^--XPosixLower}', "");
Error('\p{:=  is_lower}');
Error('\P{:=  is_lower}');
Expect(1, 120779, '\p{islower}', "");
Expect(0, 120779, '\p{^islower}', "");
Expect(0, 120779, '\P{islower}', "");
Expect(1, 120779, '\P{^islower}', "");
Expect(0, 120780, '\p{islower}', "");
Expect(1, 120780, '\p{^islower}', "");
Expect(1, 120780, '\P{islower}', "");
Expect(0, 120780, '\P{^islower}', "");
Expect(1, 120779, '\p{--Is_LOWER}', "");
Expect(0, 120779, '\p{^--Is_LOWER}', "");
Expect(0, 120779, '\P{--Is_LOWER}', "");
Expect(1, 120779, '\P{^--Is_LOWER}', "");
Expect(0, 120780, '\p{--Is_LOWER}', "");
Expect(1, 120780, '\p{^--Is_LOWER}', "");
Expect(1, 120780, '\P{--Is_LOWER}', "");
Expect(0, 120780, '\P{^--Is_LOWER}', "");
Error('\p{:= -Is_XPosixLower}');
Error('\P{:= -Is_XPosixLower}');
Expect(1, 120779, '\p{isxposixlower}', "");
Expect(0, 120779, '\p{^isxposixlower}', "");
Expect(0, 120779, '\P{isxposixlower}', "");
Expect(1, 120779, '\P{^isxposixlower}', "");
Expect(0, 120780, '\p{isxposixlower}', "");
Expect(1, 120780, '\p{^isxposixlower}', "");
Expect(1, 120780, '\P{isxposixlower}', "");
Expect(0, 120780, '\P{^isxposixlower}', "");
Expect(1, 120779, '\p{Is_XPosixLower}', "");
Expect(0, 120779, '\p{^Is_XPosixLower}', "");
Expect(0, 120779, '\P{Is_XPosixLower}', "");
Expect(1, 120779, '\P{^Is_XPosixLower}', "");
Expect(0, 120780, '\p{Is_XPosixLower}', "");
Expect(1, 120780, '\p{^Is_XPosixLower}', "");
Expect(1, 120780, '\P{Is_XPosixLower}', "");
Expect(0, 120780, '\P{^Is_XPosixLower}', "");
Error('\p{ /a/Lowercase}');
Error('\P{ /a/Lowercase}');
Expect(1, 120779, '\p{lowercase}', "");
Expect(0, 120779, '\p{^lowercase}', "");
Expect(0, 120779, '\P{lowercase}', "");
Expect(1, 120779, '\P{^lowercase}', "");
Expect(0, 120780, '\p{lowercase}', "");
Expect(1, 120780, '\p{^lowercase}', "");
Expect(1, 120780, '\P{lowercase}', "");
Expect(0, 120780, '\P{^lowercase}', "");
Expect(1, 120779, '\p{-	Lowercase}', "");
Expect(0, 120779, '\p{^-	Lowercase}', "");
Expect(0, 120779, '\P{-	Lowercase}', "");
Expect(1, 120779, '\P{^-	Lowercase}', "");
Expect(0, 120780, '\p{-	Lowercase}', "");
Expect(1, 120780, '\p{^-	Lowercase}', "");
Expect(1, 120780, '\P{-	Lowercase}', "");
Expect(0, 120780, '\P{^-	Lowercase}', "");
Error('\p{	 is_lowercase:=}');
Error('\P{	 is_lowercase:=}');
Expect(1, 120779, '\p{islowercase}', "");
Expect(0, 120779, '\p{^islowercase}', "");
Expect(0, 120779, '\P{islowercase}', "");
Expect(1, 120779, '\P{^islowercase}', "");
Expect(0, 120780, '\p{islowercase}', "");
Expect(1, 120780, '\p{^islowercase}', "");
Expect(1, 120780, '\P{islowercase}', "");
Expect(0, 120780, '\P{^islowercase}', "");
Expect(1, 120779, '\p{ Is_Lowercase}', "");
Expect(0, 120779, '\p{^ Is_Lowercase}', "");
Expect(0, 120779, '\P{ Is_Lowercase}', "");
Expect(1, 120779, '\P{^ Is_Lowercase}', "");
Expect(0, 120780, '\p{ Is_Lowercase}', "");
Expect(1, 120780, '\p{^ Is_Lowercase}', "");
Expect(1, 120780, '\P{ Is_Lowercase}', "");
Expect(0, 120780, '\P{^ Is_Lowercase}', "");
Error('\p{:=-Ethiopic_Extended_A}');
Error('\P{:=-Ethiopic_Extended_A}');
Expect(1, 43823, '\p{ethiopicextendeda}', "");
Expect(0, 43823, '\p{^ethiopicextendeda}', "");
Expect(0, 43823, '\P{ethiopicextendeda}', "");
Expect(1, 43823, '\P{^ethiopicextendeda}', "");
Expect(0, 43824, '\p{ethiopicextendeda}', "");
Expect(1, 43824, '\p{^ethiopicextendeda}', "");
Expect(1, 43824, '\P{ethiopicextendeda}', "");
Expect(0, 43824, '\P{^ethiopicextendeda}', "");
Expect(1, 43823, '\p{	ethiopic_extended_A}', "");
Expect(0, 43823, '\p{^	ethiopic_extended_A}', "");
Expect(0, 43823, '\P{	ethiopic_extended_A}', "");
Expect(1, 43823, '\P{^	ethiopic_extended_A}', "");
Expect(0, 43824, '\p{	ethiopic_extended_A}', "");
Expect(1, 43824, '\p{^	ethiopic_extended_A}', "");
Expect(1, 43824, '\P{	ethiopic_extended_A}', "");
Expect(0, 43824, '\P{^	ethiopic_extended_A}', "");
Error('\p{/a/ Is_Ethiopic_extended_A}');
Error('\P{/a/ Is_Ethiopic_extended_A}');
Expect(1, 43823, '\p{isethiopicextendeda}', "");
Expect(0, 43823, '\p{^isethiopicextendeda}', "");
Expect(0, 43823, '\P{isethiopicextendeda}', "");
Expect(1, 43823, '\P{^isethiopicextendeda}', "");
Expect(0, 43824, '\p{isethiopicextendeda}', "");
Expect(1, 43824, '\p{^isethiopicextendeda}', "");
Expect(1, 43824, '\P{isethiopicextendeda}', "");
Expect(0, 43824, '\P{^isethiopicextendeda}', "");
Expect(1, 43823, '\p{ _Is_Ethiopic_Extended_A}', "");
Expect(0, 43823, '\p{^ _Is_Ethiopic_Extended_A}', "");
Expect(0, 43823, '\P{ _Is_Ethiopic_Extended_A}', "");
Expect(1, 43823, '\P{^ _Is_Ethiopic_Extended_A}', "");
Expect(0, 43824, '\p{ _Is_Ethiopic_Extended_A}', "");
Expect(1, 43824, '\p{^ _Is_Ethiopic_Extended_A}', "");
Expect(1, 43824, '\P{ _Is_Ethiopic_Extended_A}', "");
Expect(0, 43824, '\P{^ _Is_Ethiopic_Extended_A}', "");
Error('\p{:=		in_Ethiopic_EXTENDED_A}');
Error('\P{:=		in_Ethiopic_EXTENDED_A}');
Expect(1, 43823, '\p{inethiopicextendeda}', "");
Expect(0, 43823, '\p{^inethiopicextendeda}', "");
Expect(0, 43823, '\P{inethiopicextendeda}', "");
Expect(1, 43823, '\P{^inethiopicextendeda}', "");
Expect(0, 43824, '\p{inethiopicextendeda}', "");
Expect(1, 43824, '\p{^inethiopicextendeda}', "");
Expect(1, 43824, '\P{inethiopicextendeda}', "");
Expect(0, 43824, '\P{^inethiopicextendeda}', "");
Expect(1, 43823, '\p{ IN_ethiopic_Extended_A}', "");
Expect(0, 43823, '\p{^ IN_ethiopic_Extended_A}', "");
Expect(0, 43823, '\P{ IN_ethiopic_Extended_A}', "");
Expect(1, 43823, '\P{^ IN_ethiopic_Extended_A}', "");
Expect(0, 43824, '\p{ IN_ethiopic_Extended_A}', "");
Expect(1, 43824, '\p{^ IN_ethiopic_Extended_A}', "");
Expect(1, 43824, '\P{ IN_ethiopic_Extended_A}', "");
Expect(0, 43824, '\P{^ IN_ethiopic_Extended_A}', "");
Error('\p{	:=ETHIOPIC_EXT_A}');
Error('\P{	:=ETHIOPIC_EXT_A}');
Expect(1, 43823, '\p{ethiopicexta}', "");
Expect(0, 43823, '\p{^ethiopicexta}', "");
Expect(0, 43823, '\P{ethiopicexta}', "");
Expect(1, 43823, '\P{^ethiopicexta}', "");
Expect(0, 43824, '\p{ethiopicexta}', "");
Expect(1, 43824, '\p{^ethiopicexta}', "");
Expect(1, 43824, '\P{ethiopicexta}', "");
Expect(0, 43824, '\P{^ethiopicexta}', "");
Expect(1, 43823, '\p{-Ethiopic_ext_A}', "");
Expect(0, 43823, '\p{^-Ethiopic_ext_A}', "");
Expect(0, 43823, '\P{-Ethiopic_ext_A}', "");
Expect(1, 43823, '\P{^-Ethiopic_ext_A}', "");
Expect(0, 43824, '\p{-Ethiopic_ext_A}', "");
Expect(1, 43824, '\p{^-Ethiopic_ext_A}', "");
Expect(1, 43824, '\P{-Ethiopic_ext_A}', "");
Expect(0, 43824, '\P{^-Ethiopic_ext_A}', "");
Error('\p{ _Is_Ethiopic_Ext_A:=}');
Error('\P{ _Is_Ethiopic_Ext_A:=}');
Expect(1, 43823, '\p{isethiopicexta}', "");
Expect(0, 43823, '\p{^isethiopicexta}', "");
Expect(0, 43823, '\P{isethiopicexta}', "");
Expect(1, 43823, '\P{^isethiopicexta}', "");
Expect(0, 43824, '\p{isethiopicexta}', "");
Expect(1, 43824, '\p{^isethiopicexta}', "");
Expect(1, 43824, '\P{isethiopicexta}', "");
Expect(0, 43824, '\P{^isethiopicexta}', "");
Expect(1, 43823, '\p{_Is_Ethiopic_ext_a}', "");
Expect(0, 43823, '\p{^_Is_Ethiopic_ext_a}', "");
Expect(0, 43823, '\P{_Is_Ethiopic_ext_a}', "");
Expect(1, 43823, '\P{^_Is_Ethiopic_ext_a}', "");
Expect(0, 43824, '\p{_Is_Ethiopic_ext_a}', "");
Expect(1, 43824, '\p{^_Is_Ethiopic_ext_a}', "");
Expect(1, 43824, '\P{_Is_Ethiopic_ext_a}', "");
Expect(0, 43824, '\P{^_Is_Ethiopic_ext_a}', "");
Error('\p{/a/ -in_ethiopic_ext_A}');
Error('\P{/a/ -in_ethiopic_ext_A}');
Expect(1, 43823, '\p{inethiopicexta}', "");
Expect(0, 43823, '\p{^inethiopicexta}', "");
Expect(0, 43823, '\P{inethiopicexta}', "");
Expect(1, 43823, '\P{^inethiopicexta}', "");
Expect(0, 43824, '\p{inethiopicexta}', "");
Expect(1, 43824, '\p{^inethiopicexta}', "");
Expect(1, 43824, '\P{inethiopicexta}', "");
Expect(0, 43824, '\P{^inethiopicexta}', "");
Expect(1, 43823, '\p{--IN_Ethiopic_Ext_A}', "");
Expect(0, 43823, '\p{^--IN_Ethiopic_Ext_A}', "");
Expect(0, 43823, '\P{--IN_Ethiopic_Ext_A}', "");
Expect(1, 43823, '\P{^--IN_Ethiopic_Ext_A}', "");
Expect(0, 43824, '\p{--IN_Ethiopic_Ext_A}', "");
Expect(1, 43824, '\p{^--IN_Ethiopic_Ext_A}', "");
Expect(1, 43824, '\P{--IN_Ethiopic_Ext_A}', "");
Expect(0, 43824, '\P{^--IN_Ethiopic_Ext_A}', "");
Error('\p{Expands_On_NFKD}');
Error('\P{Expands_On_NFKD}');
Error('\p{Is_Expands_On_NFKD}');
Error('\P{Is_Expands_On_NFKD}');
Error('\p{XO_NFKD}');
Error('\P{XO_NFKD}');
Error('\p{Is_XO_NFKD}');
Error('\P{Is_XO_NFKD}');
Error('\p{-	CLOSE_Punctuation/a/}');
Error('\P{-	CLOSE_Punctuation/a/}');
Expect(1, 65379, '\p{closepunctuation}', "");
Expect(0, 65379, '\p{^closepunctuation}', "");
Expect(0, 65379, '\P{closepunctuation}', "");
Expect(1, 65379, '\P{^closepunctuation}', "");
Expect(0, 65380, '\p{closepunctuation}', "");
Expect(1, 65380, '\p{^closepunctuation}', "");
Expect(1, 65380, '\P{closepunctuation}', "");
Expect(0, 65380, '\P{^closepunctuation}', "");
Expect(1, 65379, '\p{	close_punctuation}', "");
Expect(0, 65379, '\p{^	close_punctuation}', "");
Expect(0, 65379, '\P{	close_punctuation}', "");
Expect(1, 65379, '\P{^	close_punctuation}', "");
Expect(0, 65380, '\p{	close_punctuation}', "");
Expect(1, 65380, '\p{^	close_punctuation}', "");
Expect(1, 65380, '\P{	close_punctuation}', "");
Expect(0, 65380, '\P{^	close_punctuation}', "");
Error('\p{/a/-Is_Close_Punctuation}');
Error('\P{/a/-Is_Close_Punctuation}');
Expect(1, 65379, '\p{isclosepunctuation}', "");
Expect(0, 65379, '\p{^isclosepunctuation}', "");
Expect(0, 65379, '\P{isclosepunctuation}', "");
Expect(1, 65379, '\P{^isclosepunctuation}', "");
Expect(0, 65380, '\p{isclosepunctuation}', "");
Expect(1, 65380, '\p{^isclosepunctuation}', "");
Expect(1, 65380, '\P{isclosepunctuation}', "");
Expect(0, 65380, '\P{^isclosepunctuation}', "");
Expect(1, 65379, '\p{_ Is_close_punctuation}', "");
Expect(0, 65379, '\p{^_ Is_close_punctuation}', "");
Expect(0, 65379, '\P{_ Is_close_punctuation}', "");
Expect(1, 65379, '\P{^_ Is_close_punctuation}', "");
Expect(0, 65380, '\p{_ Is_close_punctuation}', "");
Expect(1, 65380, '\p{^_ Is_close_punctuation}', "");
Expect(1, 65380, '\P{_ Is_close_punctuation}', "");
Expect(0, 65380, '\P{^_ Is_close_punctuation}', "");
Error('\p{/a/ Pe}');
Error('\P{/a/ Pe}');
Expect(1, 65379, '\p{pe}', "");
Expect(0, 65379, '\p{^pe}', "");
Expect(0, 65379, '\P{pe}', "");
Expect(1, 65379, '\P{^pe}', "");
Expect(0, 65380, '\p{pe}', "");
Expect(1, 65380, '\p{^pe}', "");
Expect(1, 65380, '\P{pe}', "");
Expect(0, 65380, '\P{^pe}', "");
Expect(1, 65379, '\p{	-Pe}', "");
Expect(0, 65379, '\p{^	-Pe}', "");
Expect(0, 65379, '\P{	-Pe}', "");
Expect(1, 65379, '\P{^	-Pe}', "");
Expect(0, 65380, '\p{	-Pe}', "");
Expect(1, 65380, '\p{^	-Pe}', "");
Expect(1, 65380, '\P{	-Pe}', "");
Expect(0, 65380, '\P{^	-Pe}', "");
Error('\p{-/a/IS_pe}');
Error('\P{-/a/IS_pe}');
Expect(1, 65379, '\p{ispe}', "");
Expect(0, 65379, '\p{^ispe}', "");
Expect(0, 65379, '\P{ispe}', "");
Expect(1, 65379, '\P{^ispe}', "");
Expect(0, 65380, '\p{ispe}', "");
Expect(1, 65380, '\p{^ispe}', "");
Expect(1, 65380, '\P{ispe}', "");
Expect(0, 65380, '\P{^ispe}', "");
Expect(1, 65379, '\p{	_is_PE}', "");
Expect(0, 65379, '\p{^	_is_PE}', "");
Expect(0, 65379, '\P{	_is_PE}', "");
Expect(1, 65379, '\P{^	_is_PE}', "");
Expect(0, 65380, '\p{	_is_PE}', "");
Expect(1, 65380, '\p{^	_is_PE}', "");
Expect(1, 65380, '\P{	_is_PE}', "");
Expect(0, 65380, '\P{^	_is_PE}', "");
Error('\p{Expands_On_NFKC}');
Error('\P{Expands_On_NFKC}');
Error('\p{Is_Expands_On_NFKC}');
Error('\P{Is_Expands_On_NFKC}');
Error('\p{XO_NFKC}');
Error('\P{XO_NFKC}');
Error('\p{Is_XO_NFKC}');
Error('\P{Is_XO_NFKC}');
Error('\p{-/a/unknown}');
Error('\P{-/a/unknown}');
Expect(1, 918000, '\p{unknown}', "");
Expect(0, 918000, '\p{^unknown}', "");
Expect(0, 918000, '\P{unknown}', "");
Expect(1, 918000, '\P{^unknown}', "");
Expect(0, 917999, '\p{unknown}', "");
Expect(1, 917999, '\p{^unknown}', "");
Expect(1, 917999, '\P{unknown}', "");
Expect(0, 917999, '\P{^unknown}', "");
Expect(1, 918000, '\p{- unknown}', "");
Expect(0, 918000, '\p{^- unknown}', "");
Expect(0, 918000, '\P{- unknown}', "");
Expect(1, 918000, '\P{^- unknown}', "");
Expect(0, 917999, '\p{- unknown}', "");
Expect(1, 917999, '\p{^- unknown}', "");
Expect(1, 917999, '\P{- unknown}', "");
Expect(0, 917999, '\P{^- unknown}', "");
Error('\p{:=--is_UNKNOWN}');
Error('\P{:=--is_UNKNOWN}');
Expect(1, 918000, '\p{isunknown}', "");
Expect(0, 918000, '\p{^isunknown}', "");
Expect(0, 918000, '\P{isunknown}', "");
Expect(1, 918000, '\P{^isunknown}', "");
Expect(0, 917999, '\p{isunknown}', "");
Expect(1, 917999, '\p{^isunknown}', "");
Expect(1, 917999, '\P{isunknown}', "");
Expect(0, 917999, '\P{^isunknown}', "");
Expect(1, 918000, '\p{ is_unknown}', "");
Expect(0, 918000, '\p{^ is_unknown}', "");
Expect(0, 918000, '\P{ is_unknown}', "");
Expect(1, 918000, '\P{^ is_unknown}', "");
Expect(0, 917999, '\p{ is_unknown}', "");
Expect(1, 917999, '\p{^ is_unknown}', "");
Expect(1, 917999, '\P{ is_unknown}', "");
Expect(0, 917999, '\P{^ is_unknown}', "");
Error('\p{/a/		zzzz}');
Error('\P{/a/		zzzz}');
Expect(1, 918000, '\p{zzzz}', "");
Expect(0, 918000, '\p{^zzzz}', "");
Expect(0, 918000, '\P{zzzz}', "");
Expect(1, 918000, '\P{^zzzz}', "");
Expect(0, 917999, '\p{zzzz}', "");
Expect(1, 917999, '\p{^zzzz}', "");
Expect(1, 917999, '\P{zzzz}', "");
Expect(0, 917999, '\P{^zzzz}', "");
Expect(1, 918000, '\p{ZZZZ}', "");
Expect(0, 918000, '\p{^ZZZZ}', "");
Expect(0, 918000, '\P{ZZZZ}', "");
Expect(1, 918000, '\P{^ZZZZ}', "");
Expect(0, 917999, '\p{ZZZZ}', "");
Expect(1, 917999, '\p{^ZZZZ}', "");
Expect(1, 917999, '\P{ZZZZ}', "");
Expect(0, 917999, '\P{^ZZZZ}', "");
Error('\p{-/a/is_zzzz}');
Error('\P{-/a/is_zzzz}');
Expect(1, 918000, '\p{iszzzz}', "");
Expect(0, 918000, '\p{^iszzzz}', "");
Expect(0, 918000, '\P{iszzzz}', "");
Expect(1, 918000, '\P{^iszzzz}', "");
Expect(0, 917999, '\p{iszzzz}', "");
Expect(1, 917999, '\p{^iszzzz}', "");
Expect(1, 917999, '\P{iszzzz}', "");
Expect(0, 917999, '\P{^iszzzz}', "");
Expect(1, 918000, '\p{_is_Zzzz}', "");
Expect(0, 918000, '\p{^_is_Zzzz}', "");
Expect(0, 918000, '\P{_is_Zzzz}', "");
Expect(1, 918000, '\P{^_is_Zzzz}', "");
Expect(0, 917999, '\p{_is_Zzzz}', "");
Expect(1, 917999, '\p{^_is_Zzzz}', "");
Expect(1, 917999, '\P{_is_Zzzz}', "");
Expect(0, 917999, '\P{^_is_Zzzz}', "");
Error('\p{Cyrillic_SUPPLEMENT/a/}');
Error('\P{Cyrillic_SUPPLEMENT/a/}');
Expect(1, 1327, '\p{cyrillicsupplement}', "");
Expect(0, 1327, '\p{^cyrillicsupplement}', "");
Expect(0, 1327, '\P{cyrillicsupplement}', "");
Expect(1, 1327, '\P{^cyrillicsupplement}', "");
Expect(0, 1328, '\p{cyrillicsupplement}', "");
Expect(1, 1328, '\p{^cyrillicsupplement}', "");
Expect(1, 1328, '\P{cyrillicsupplement}', "");
Expect(0, 1328, '\P{^cyrillicsupplement}', "");
Expect(1, 1327, '\p{	-CYRILLIC_SUPPLEMENT}', "");
Expect(0, 1327, '\p{^	-CYRILLIC_SUPPLEMENT}', "");
Expect(0, 1327, '\P{	-CYRILLIC_SUPPLEMENT}', "");
Expect(1, 1327, '\P{^	-CYRILLIC_SUPPLEMENT}', "");
Expect(0, 1328, '\p{	-CYRILLIC_SUPPLEMENT}', "");
Expect(1, 1328, '\p{^	-CYRILLIC_SUPPLEMENT}', "");
Expect(1, 1328, '\P{	-CYRILLIC_SUPPLEMENT}', "");
Expect(0, 1328, '\P{^	-CYRILLIC_SUPPLEMENT}', "");
Error('\p{__IS_cyrillic_Supplement:=}');
Error('\P{__IS_cyrillic_Supplement:=}');
Expect(1, 1327, '\p{iscyrillicsupplement}', "");
Expect(0, 1327, '\p{^iscyrillicsupplement}', "");
Expect(0, 1327, '\P{iscyrillicsupplement}', "");
Expect(1, 1327, '\P{^iscyrillicsupplement}', "");
Expect(0, 1328, '\p{iscyrillicsupplement}', "");
Expect(1, 1328, '\p{^iscyrillicsupplement}', "");
Expect(1, 1328, '\P{iscyrillicsupplement}', "");
Expect(0, 1328, '\P{^iscyrillicsupplement}', "");
Expect(1, 1327, '\p{_-Is_Cyrillic_supplement}', "");
Expect(0, 1327, '\p{^_-Is_Cyrillic_supplement}', "");
Expect(0, 1327, '\P{_-Is_Cyrillic_supplement}', "");
Expect(1, 1327, '\P{^_-Is_Cyrillic_supplement}', "");
Expect(0, 1328, '\p{_-Is_Cyrillic_supplement}', "");
Expect(1, 1328, '\p{^_-Is_Cyrillic_supplement}', "");
Expect(1, 1328, '\P{_-Is_Cyrillic_supplement}', "");
Expect(0, 1328, '\P{^_-Is_Cyrillic_supplement}', "");
Error('\p{-_In_cyrillic_Supplement:=}');
Error('\P{-_In_cyrillic_Supplement:=}');
Expect(1, 1327, '\p{incyrillicsupplement}', "");
Expect(0, 1327, '\p{^incyrillicsupplement}', "");
Expect(0, 1327, '\P{incyrillicsupplement}', "");
Expect(1, 1327, '\P{^incyrillicsupplement}', "");
Expect(0, 1328, '\p{incyrillicsupplement}', "");
Expect(1, 1328, '\p{^incyrillicsupplement}', "");
Expect(1, 1328, '\P{incyrillicsupplement}', "");
Expect(0, 1328, '\P{^incyrillicsupplement}', "");
Expect(1, 1327, '\p{ -in_cyrillic_supplement}', "");
Expect(0, 1327, '\p{^ -in_cyrillic_supplement}', "");
Expect(0, 1327, '\P{ -in_cyrillic_supplement}', "");
Expect(1, 1327, '\P{^ -in_cyrillic_supplement}', "");
Expect(0, 1328, '\p{ -in_cyrillic_supplement}', "");
Expect(1, 1328, '\p{^ -in_cyrillic_supplement}', "");
Expect(1, 1328, '\P{ -in_cyrillic_supplement}', "");
Expect(0, 1328, '\P{^ -in_cyrillic_supplement}', "");
Error('\p{/a/Cyrillic_Sup}');
Error('\P{/a/Cyrillic_Sup}');
Expect(1, 1327, '\p{cyrillicsup}', "");
Expect(0, 1327, '\p{^cyrillicsup}', "");
Expect(0, 1327, '\P{cyrillicsup}', "");
Expect(1, 1327, '\P{^cyrillicsup}', "");
Expect(0, 1328, '\p{cyrillicsup}', "");
Expect(1, 1328, '\p{^cyrillicsup}', "");
Expect(1, 1328, '\P{cyrillicsup}', "");
Expect(0, 1328, '\P{^cyrillicsup}', "");
Expect(1, 1327, '\p{-cyrillic_sup}', "");
Expect(0, 1327, '\p{^-cyrillic_sup}', "");
Expect(0, 1327, '\P{-cyrillic_sup}', "");
Expect(1, 1327, '\P{^-cyrillic_sup}', "");
Expect(0, 1328, '\p{-cyrillic_sup}', "");
Expect(1, 1328, '\p{^-cyrillic_sup}', "");
Expect(1, 1328, '\P{-cyrillic_sup}', "");
Expect(0, 1328, '\P{^-cyrillic_sup}', "");
Error('\p{/a/ 	IS_cyrillic_SUP}');
Error('\P{/a/ 	IS_cyrillic_SUP}');
Expect(1, 1327, '\p{iscyrillicsup}', "");
Expect(0, 1327, '\p{^iscyrillicsup}', "");
Expect(0, 1327, '\P{iscyrillicsup}', "");
Expect(1, 1327, '\P{^iscyrillicsup}', "");
Expect(0, 1328, '\p{iscyrillicsup}', "");
Expect(1, 1328, '\p{^iscyrillicsup}', "");
Expect(1, 1328, '\P{iscyrillicsup}', "");
Expect(0, 1328, '\P{^iscyrillicsup}', "");
Expect(1, 1327, '\p{-is_CYRILLIC_Sup}', "");
Expect(0, 1327, '\p{^-is_CYRILLIC_Sup}', "");
Expect(0, 1327, '\P{-is_CYRILLIC_Sup}', "");
Expect(1, 1327, '\P{^-is_CYRILLIC_Sup}', "");
Expect(0, 1328, '\p{-is_CYRILLIC_Sup}', "");
Expect(1, 1328, '\p{^-is_CYRILLIC_Sup}', "");
Expect(1, 1328, '\P{-is_CYRILLIC_Sup}', "");
Expect(0, 1328, '\P{^-is_CYRILLIC_Sup}', "");
Error('\p{	In_Cyrillic_SUP/a/}');
Error('\P{	In_Cyrillic_SUP/a/}');
Expect(1, 1327, '\p{incyrillicsup}', "");
Expect(0, 1327, '\p{^incyrillicsup}', "");
Expect(0, 1327, '\P{incyrillicsup}', "");
Expect(1, 1327, '\P{^incyrillicsup}', "");
Expect(0, 1328, '\p{incyrillicsup}', "");
Expect(1, 1328, '\p{^incyrillicsup}', "");
Expect(1, 1328, '\P{incyrillicsup}', "");
Expect(0, 1328, '\P{^incyrillicsup}', "");
Expect(1, 1327, '\p{ -In_Cyrillic_Sup}', "");
Expect(0, 1327, '\p{^ -In_Cyrillic_Sup}', "");
Expect(0, 1327, '\P{ -In_Cyrillic_Sup}', "");
Expect(1, 1327, '\P{^ -In_Cyrillic_Sup}', "");
Expect(0, 1328, '\p{ -In_Cyrillic_Sup}', "");
Expect(1, 1328, '\p{^ -In_Cyrillic_Sup}', "");
Expect(1, 1328, '\P{ -In_Cyrillic_Sup}', "");
Expect(0, 1328, '\P{^ -In_Cyrillic_Sup}', "");
Error('\p{	_Cyrillic_supplementary/a/}');
Error('\P{	_Cyrillic_supplementary/a/}');
Expect(1, 1327, '\p{cyrillicsupplementary}', "");
Expect(0, 1327, '\p{^cyrillicsupplementary}', "");
Expect(0, 1327, '\P{cyrillicsupplementary}', "");
Expect(1, 1327, '\P{^cyrillicsupplementary}', "");
Expect(0, 1328, '\p{cyrillicsupplementary}', "");
Expect(1, 1328, '\p{^cyrillicsupplementary}', "");
Expect(1, 1328, '\P{cyrillicsupplementary}', "");
Expect(0, 1328, '\P{^cyrillicsupplementary}', "");
Expect(1, 1327, '\p{	cyrillic_Supplementary}', "");
Expect(0, 1327, '\p{^	cyrillic_Supplementary}', "");
Expect(0, 1327, '\P{	cyrillic_Supplementary}', "");
Expect(1, 1327, '\P{^	cyrillic_Supplementary}', "");
Expect(0, 1328, '\p{	cyrillic_Supplementary}', "");
Expect(1, 1328, '\p{^	cyrillic_Supplementary}', "");
Expect(1, 1328, '\P{	cyrillic_Supplementary}', "");
Expect(0, 1328, '\P{^	cyrillic_Supplementary}', "");
Error('\p{-	is_cyrillic_Supplementary:=}');
Error('\P{-	is_cyrillic_Supplementary:=}');
Expect(1, 1327, '\p{iscyrillicsupplementary}', "");
Expect(0, 1327, '\p{^iscyrillicsupplementary}', "");
Expect(0, 1327, '\P{iscyrillicsupplementary}', "");
Expect(1, 1327, '\P{^iscyrillicsupplementary}', "");
Expect(0, 1328, '\p{iscyrillicsupplementary}', "");
Expect(1, 1328, '\p{^iscyrillicsupplementary}', "");
Expect(1, 1328, '\P{iscyrillicsupplementary}', "");
Expect(0, 1328, '\P{^iscyrillicsupplementary}', "");
Expect(1, 1327, '\p{	-Is_Cyrillic_Supplementary}', "");
Expect(0, 1327, '\p{^	-Is_Cyrillic_Supplementary}', "");
Expect(0, 1327, '\P{	-Is_Cyrillic_Supplementary}', "");
Expect(1, 1327, '\P{^	-Is_Cyrillic_Supplementary}', "");
Expect(0, 1328, '\p{	-Is_Cyrillic_Supplementary}', "");
Expect(1, 1328, '\p{^	-Is_Cyrillic_Supplementary}', "");
Expect(1, 1328, '\P{	-Is_Cyrillic_Supplementary}', "");
Expect(0, 1328, '\P{^	-Is_Cyrillic_Supplementary}', "");
Error('\p{_In_Cyrillic_Supplementary/a/}');
Error('\P{_In_Cyrillic_Supplementary/a/}');
Expect(1, 1327, '\p{incyrillicsupplementary}', "");
Expect(0, 1327, '\p{^incyrillicsupplementary}', "");
Expect(0, 1327, '\P{incyrillicsupplementary}', "");
Expect(1, 1327, '\P{^incyrillicsupplementary}', "");
Expect(0, 1328, '\p{incyrillicsupplementary}', "");
Expect(1, 1328, '\p{^incyrillicsupplementary}', "");
Expect(1, 1328, '\P{incyrillicsupplementary}', "");
Expect(0, 1328, '\P{^incyrillicsupplementary}', "");
Expect(1, 1327, '\p{ 	In_CYRILLIC_SUPPLEMENTARY}', "");
Expect(0, 1327, '\p{^ 	In_CYRILLIC_SUPPLEMENTARY}', "");
Expect(0, 1327, '\P{ 	In_CYRILLIC_SUPPLEMENTARY}', "");
Expect(1, 1327, '\P{^ 	In_CYRILLIC_SUPPLEMENTARY}', "");
Expect(0, 1328, '\p{ 	In_CYRILLIC_SUPPLEMENTARY}', "");
Expect(1, 1328, '\p{^ 	In_CYRILLIC_SUPPLEMENTARY}', "");
Expect(1, 1328, '\P{ 	In_CYRILLIC_SUPPLEMENTARY}', "");
Expect(0, 1328, '\P{^ 	In_CYRILLIC_SUPPLEMENTARY}', "");
Error('\p{_-in_brahmi:=}');
Error('\P{_-in_brahmi:=}');
Expect(1, 69759, '\p{inbrahmi}', "");
Expect(0, 69759, '\p{^inbrahmi}', "");
Expect(0, 69759, '\P{inbrahmi}', "");
Expect(1, 69759, '\P{^inbrahmi}', "");
Expect(0, 69760, '\p{inbrahmi}', "");
Expect(1, 69760, '\p{^inbrahmi}', "");
Expect(1, 69760, '\P{inbrahmi}', "");
Expect(0, 69760, '\P{^inbrahmi}', "");
Expect(1, 69759, '\p{_	In_Brahmi}', "");
Expect(0, 69759, '\p{^_	In_Brahmi}', "");
Expect(0, 69759, '\P{_	In_Brahmi}', "");
Expect(1, 69759, '\P{^_	In_Brahmi}', "");
Expect(0, 69760, '\p{_	In_Brahmi}', "");
Expect(1, 69760, '\p{^_	In_Brahmi}', "");
Expect(1, 69760, '\P{_	In_Brahmi}', "");
Expect(0, 69760, '\P{^_	In_Brahmi}', "");
Error('\p{:=- IN_BRAHMI}');
Error('\P{:=- IN_BRAHMI}');
Expect(1, 69759, '\p{	IN_BRAHMI}', "");
Expect(0, 69759, '\p{^	IN_BRAHMI}', "");
Expect(0, 69759, '\P{	IN_BRAHMI}', "");
Expect(1, 69759, '\P{^	IN_BRAHMI}', "");
Expect(0, 69760, '\p{	IN_BRAHMI}', "");
Expect(1, 69760, '\p{^	IN_BRAHMI}', "");
Expect(1, 69760, '\P{	IN_BRAHMI}', "");
Expect(0, 69760, '\P{^	IN_BRAHMI}', "");
Error('\p{:=-	linear_b_Ideograms}');
Error('\P{:=-	linear_b_Ideograms}');
Expect(1, 65791, '\p{linearbideograms}', "");
Expect(0, 65791, '\p{^linearbideograms}', "");
Expect(0, 65791, '\P{linearbideograms}', "");
Expect(1, 65791, '\P{^linearbideograms}', "");
Expect(0, 65792, '\p{linearbideograms}', "");
Expect(1, 65792, '\p{^linearbideograms}', "");
Expect(1, 65792, '\P{linearbideograms}', "");
Expect(0, 65792, '\P{^linearbideograms}', "");
Expect(1, 65791, '\p{	linear_B_IDEOGRAMS}', "");
Expect(0, 65791, '\p{^	linear_B_IDEOGRAMS}', "");
Expect(0, 65791, '\P{	linear_B_IDEOGRAMS}', "");
Expect(1, 65791, '\P{^	linear_B_IDEOGRAMS}', "");
Expect(0, 65792, '\p{	linear_B_IDEOGRAMS}', "");
Expect(1, 65792, '\p{^	linear_B_IDEOGRAMS}', "");
Expect(1, 65792, '\P{	linear_B_IDEOGRAMS}', "");
Expect(0, 65792, '\P{^	linear_B_IDEOGRAMS}', "");
Error('\p{ :=is_Linear_B_IDEOGRAMS}');
Error('\P{ :=is_Linear_B_IDEOGRAMS}');
Expect(1, 65791, '\p{islinearbideograms}', "");
Expect(0, 65791, '\p{^islinearbideograms}', "");
Expect(0, 65791, '\P{islinearbideograms}', "");
Expect(1, 65791, '\P{^islinearbideograms}', "");
Expect(0, 65792, '\p{islinearbideograms}', "");
Expect(1, 65792, '\p{^islinearbideograms}', "");
Expect(1, 65792, '\P{islinearbideograms}', "");
Expect(0, 65792, '\P{^islinearbideograms}', "");
Expect(1, 65791, '\p{_is_linear_B_Ideograms}', "");
Expect(0, 65791, '\p{^_is_linear_B_Ideograms}', "");
Expect(0, 65791, '\P{_is_linear_B_Ideograms}', "");
Expect(1, 65791, '\P{^_is_linear_B_Ideograms}', "");
Expect(0, 65792, '\p{_is_linear_B_Ideograms}', "");
Expect(1, 65792, '\p{^_is_linear_B_Ideograms}', "");
Expect(1, 65792, '\P{_is_linear_B_Ideograms}', "");
Expect(0, 65792, '\P{^_is_linear_B_Ideograms}', "");
Error('\p{		IN_linear_B_IDEOGRAMS:=}');
Error('\P{		IN_linear_B_IDEOGRAMS:=}');
Expect(1, 65791, '\p{inlinearbideograms}', "");
Expect(0, 65791, '\p{^inlinearbideograms}', "");
Expect(0, 65791, '\P{inlinearbideograms}', "");
Expect(1, 65791, '\P{^inlinearbideograms}', "");
Expect(0, 65792, '\p{inlinearbideograms}', "");
Expect(1, 65792, '\p{^inlinearbideograms}', "");
Expect(1, 65792, '\P{inlinearbideograms}', "");
Expect(0, 65792, '\P{^inlinearbideograms}', "");
Expect(1, 65791, '\p{ -In_Linear_b_ideograms}', "");
Expect(0, 65791, '\p{^ -In_Linear_b_ideograms}', "");
Expect(0, 65791, '\P{ -In_Linear_b_ideograms}', "");
Expect(1, 65791, '\P{^ -In_Linear_b_ideograms}', "");
Expect(0, 65792, '\p{ -In_Linear_b_ideograms}', "");
Expect(1, 65792, '\p{^ -In_Linear_b_ideograms}', "");
Expect(1, 65792, '\P{ -In_Linear_b_ideograms}', "");
Expect(0, 65792, '\P{^ -In_Linear_b_ideograms}', "");
Error('\p{		Sundanese_supplement/a/}');
Error('\P{		Sundanese_supplement/a/}');
Expect(1, 7375, '\p{sundanesesupplement}', "");
Expect(0, 7375, '\p{^sundanesesupplement}', "");
Expect(0, 7375, '\P{sundanesesupplement}', "");
Expect(1, 7375, '\P{^sundanesesupplement}', "");
Expect(0, 7376, '\p{sundanesesupplement}', "");
Expect(1, 7376, '\p{^sundanesesupplement}', "");
Expect(1, 7376, '\P{sundanesesupplement}', "");
Expect(0, 7376, '\P{^sundanesesupplement}', "");
Expect(1, 7375, '\p{- Sundanese_Supplement}', "");
Expect(0, 7375, '\p{^- Sundanese_Supplement}', "");
Expect(0, 7375, '\P{- Sundanese_Supplement}', "");
Expect(1, 7375, '\P{^- Sundanese_Supplement}', "");
Expect(0, 7376, '\p{- Sundanese_Supplement}', "");
Expect(1, 7376, '\p{^- Sundanese_Supplement}', "");
Expect(1, 7376, '\P{- Sundanese_Supplement}', "");
Expect(0, 7376, '\P{^- Sundanese_Supplement}', "");
Error('\p{/a/	IS_Sundanese_Supplement}');
Error('\P{/a/	IS_Sundanese_Supplement}');
Expect(1, 7375, '\p{issundanesesupplement}', "");
Expect(0, 7375, '\p{^issundanesesupplement}', "");
Expect(0, 7375, '\P{issundanesesupplement}', "");
Expect(1, 7375, '\P{^issundanesesupplement}', "");
Expect(0, 7376, '\p{issundanesesupplement}', "");
Expect(1, 7376, '\p{^issundanesesupplement}', "");
Expect(1, 7376, '\P{issundanesesupplement}', "");
Expect(0, 7376, '\P{^issundanesesupplement}', "");
Expect(1, 7375, '\p{_ Is_Sundanese_supplement}', "");
Expect(0, 7375, '\p{^_ Is_Sundanese_supplement}', "");
Expect(0, 7375, '\P{_ Is_Sundanese_supplement}', "");
Expect(1, 7375, '\P{^_ Is_Sundanese_supplement}', "");
Expect(0, 7376, '\p{_ Is_Sundanese_supplement}', "");
Expect(1, 7376, '\p{^_ Is_Sundanese_supplement}', "");
Expect(1, 7376, '\P{_ Is_Sundanese_supplement}', "");
Expect(0, 7376, '\P{^_ Is_Sundanese_supplement}', "");
Error('\p{ In_Sundanese_supplement/a/}');
Error('\P{ In_Sundanese_supplement/a/}');
Expect(1, 7375, '\p{insundanesesupplement}', "");
Expect(0, 7375, '\p{^insundanesesupplement}', "");
Expect(0, 7375, '\P{insundanesesupplement}', "");
Expect(1, 7375, '\P{^insundanesesupplement}', "");
Expect(0, 7376, '\p{insundanesesupplement}', "");
Expect(1, 7376, '\p{^insundanesesupplement}', "");
Expect(1, 7376, '\P{insundanesesupplement}', "");
Expect(0, 7376, '\P{^insundanesesupplement}', "");
Expect(1, 7375, '\p{- in_Sundanese_SUPPLEMENT}', "");
Expect(0, 7375, '\p{^- in_Sundanese_SUPPLEMENT}', "");
Expect(0, 7375, '\P{- in_Sundanese_SUPPLEMENT}', "");
Expect(1, 7375, '\P{^- in_Sundanese_SUPPLEMENT}', "");
Expect(0, 7376, '\p{- in_Sundanese_SUPPLEMENT}', "");
Expect(1, 7376, '\p{^- in_Sundanese_SUPPLEMENT}', "");
Expect(1, 7376, '\P{- in_Sundanese_SUPPLEMENT}', "");
Expect(0, 7376, '\P{^- in_Sundanese_SUPPLEMENT}', "");
Error('\p{  sundanese_Sup/a/}');
Error('\P{  sundanese_Sup/a/}');
Expect(1, 7375, '\p{sundanesesup}', "");
Expect(0, 7375, '\p{^sundanesesup}', "");
Expect(0, 7375, '\P{sundanesesup}', "");
Expect(1, 7375, '\P{^sundanesesup}', "");
Expect(0, 7376, '\p{sundanesesup}', "");
Expect(1, 7376, '\p{^sundanesesup}', "");
Expect(1, 7376, '\P{sundanesesup}', "");
Expect(0, 7376, '\P{^sundanesesup}', "");
Expect(1, 7375, '\p{-Sundanese_SUP}', "");
Expect(0, 7375, '\p{^-Sundanese_SUP}', "");
Expect(0, 7375, '\P{-Sundanese_SUP}', "");
Expect(1, 7375, '\P{^-Sundanese_SUP}', "");
Expect(0, 7376, '\p{-Sundanese_SUP}', "");
Expect(1, 7376, '\p{^-Sundanese_SUP}', "");
Expect(1, 7376, '\P{-Sundanese_SUP}', "");
Expect(0, 7376, '\P{^-Sundanese_SUP}', "");
Error('\p{:= _IS_sundanese_sup}');
Error('\P{:= _IS_sundanese_sup}');
Expect(1, 7375, '\p{issundanesesup}', "");
Expect(0, 7375, '\p{^issundanesesup}', "");
Expect(0, 7375, '\P{issundanesesup}', "");
Expect(1, 7375, '\P{^issundanesesup}', "");
Expect(0, 7376, '\p{issundanesesup}', "");
Expect(1, 7376, '\p{^issundanesesup}', "");
Expect(1, 7376, '\P{issundanesesup}', "");
Expect(0, 7376, '\P{^issundanesesup}', "");
Expect(1, 7375, '\p{_ IS_Sundanese_SUP}', "");
Expect(0, 7375, '\p{^_ IS_Sundanese_SUP}', "");
Expect(0, 7375, '\P{_ IS_Sundanese_SUP}', "");
Expect(1, 7375, '\P{^_ IS_Sundanese_SUP}', "");
Expect(0, 7376, '\p{_ IS_Sundanese_SUP}', "");
Expect(1, 7376, '\p{^_ IS_Sundanese_SUP}', "");
Expect(1, 7376, '\P{_ IS_Sundanese_SUP}', "");
Expect(0, 7376, '\P{^_ IS_Sundanese_SUP}', "");
Error('\p{  IN_Sundanese_sup:=}');
Error('\P{  IN_Sundanese_sup:=}');
Expect(1, 7375, '\p{insundanesesup}', "");
Expect(0, 7375, '\p{^insundanesesup}', "");
Expect(0, 7375, '\P{insundanesesup}', "");
Expect(1, 7375, '\P{^insundanesesup}', "");
Expect(0, 7376, '\p{insundanesesup}', "");
Expect(1, 7376, '\p{^insundanesesup}', "");
Expect(1, 7376, '\P{insundanesesup}', "");
Expect(0, 7376, '\P{^insundanesesup}', "");
Expect(1, 7375, '\p{ 	in_SUNDANESE_SUP}', "");
Expect(0, 7375, '\p{^ 	in_SUNDANESE_SUP}', "");
Expect(0, 7375, '\P{ 	in_SUNDANESE_SUP}', "");
Expect(1, 7375, '\P{^ 	in_SUNDANESE_SUP}', "");
Expect(0, 7376, '\p{ 	in_SUNDANESE_SUP}', "");
Expect(1, 7376, '\p{^ 	in_SUNDANESE_SUP}', "");
Expect(1, 7376, '\P{ 	in_SUNDANESE_SUP}', "");
Expect(0, 7376, '\P{^ 	in_SUNDANESE_SUP}', "");
Error('\p{	Mark/a/}');
Error('\P{	Mark/a/}');
Expect(1, 917999, '\p{mark}', "");
Expect(0, 917999, '\p{^mark}', "");
Expect(0, 917999, '\P{mark}', "");
Expect(1, 917999, '\P{^mark}', "");
Expect(0, 918000, '\p{mark}', "");
Expect(1, 918000, '\p{^mark}', "");
Expect(1, 918000, '\P{mark}', "");
Expect(0, 918000, '\P{^mark}', "");
Expect(1, 917999, '\p{-Mark}', "");
Expect(0, 917999, '\p{^-Mark}', "");
Expect(0, 917999, '\P{-Mark}', "");
Expect(1, 917999, '\P{^-Mark}', "");
Expect(0, 918000, '\p{-Mark}', "");
Expect(1, 918000, '\p{^-Mark}', "");
Expect(1, 918000, '\P{-Mark}', "");
Expect(0, 918000, '\P{^-Mark}', "");
Error('\p{:= IS_MARK}');
Error('\P{:= IS_MARK}');
Expect(1, 917999, '\p{ismark}', "");
Expect(0, 917999, '\p{^ismark}', "");
Expect(0, 917999, '\P{ismark}', "");
Expect(1, 917999, '\P{^ismark}', "");
Expect(0, 918000, '\p{ismark}', "");
Expect(1, 918000, '\p{^ismark}', "");
Expect(1, 918000, '\P{ismark}', "");
Expect(0, 918000, '\P{^ismark}', "");
Expect(1, 917999, '\p{_ Is_Mark}', "");
Expect(0, 917999, '\p{^_ Is_Mark}', "");
Expect(0, 917999, '\P{_ Is_Mark}', "");
Expect(1, 917999, '\P{^_ Is_Mark}', "");
Expect(0, 918000, '\p{_ Is_Mark}', "");
Expect(1, 918000, '\p{^_ Is_Mark}', "");
Expect(1, 918000, '\P{_ Is_Mark}', "");
Expect(0, 918000, '\P{^_ Is_Mark}', "");
Error('\p{ -M/a/}');
Error('\P{ -M/a/}');
Expect(1, 917999, '\p{m}', "");
Expect(0, 917999, '\p{^m}', "");
Expect(0, 917999, '\P{m}', "");
Expect(1, 917999, '\P{^m}', "");
Expect(0, 918000, '\p{m}', "");
Expect(1, 918000, '\p{^m}', "");
Expect(1, 918000, '\P{m}', "");
Expect(0, 918000, '\P{^m}', "");
Expect(1, 917999, '\p{ M}', "");
Expect(0, 917999, '\p{^ M}', "");
Expect(0, 917999, '\P{ M}', "");
Expect(1, 917999, '\P{^ M}', "");
Expect(0, 918000, '\p{ M}', "");
Expect(1, 918000, '\p{^ M}', "");
Expect(1, 918000, '\P{ M}', "");
Expect(0, 918000, '\P{^ M}', "");
Error('\p{-:=Is_m}');
Error('\P{-:=Is_m}');
Expect(1, 917999, '\p{ism}', "");
Expect(0, 917999, '\p{^ism}', "");
Expect(0, 917999, '\P{ism}', "");
Expect(1, 917999, '\P{^ism}', "");
Expect(0, 918000, '\p{ism}', "");
Expect(1, 918000, '\p{^ism}', "");
Expect(1, 918000, '\P{ism}', "");
Expect(0, 918000, '\P{^ism}', "");
Expect(1, 917999, '\p{ -is_M}', "");
Expect(0, 917999, '\p{^ -is_M}', "");
Expect(0, 917999, '\P{ -is_M}', "");
Expect(1, 917999, '\P{^ -is_M}', "");
Expect(0, 918000, '\p{ -is_M}', "");
Expect(1, 918000, '\p{^ -is_M}', "");
Expect(1, 918000, '\P{ -is_M}', "");
Expect(0, 918000, '\P{^ -is_M}', "");
Error('\p{/a/_Combining_MARK}');
Error('\P{/a/_Combining_MARK}');
Expect(1, 917999, '\p{combiningmark}', "");
Expect(0, 917999, '\p{^combiningmark}', "");
Expect(0, 917999, '\P{combiningmark}', "");
Expect(1, 917999, '\P{^combiningmark}', "");
Expect(0, 918000, '\p{combiningmark}', "");
Expect(1, 918000, '\p{^combiningmark}', "");
Expect(1, 918000, '\P{combiningmark}', "");
Expect(0, 918000, '\P{^combiningmark}', "");
Expect(1, 917999, '\p{	_combining_MARK}', "");
Expect(0, 917999, '\p{^	_combining_MARK}', "");
Expect(0, 917999, '\P{	_combining_MARK}', "");
Expect(1, 917999, '\P{^	_combining_MARK}', "");
Expect(0, 918000, '\p{	_combining_MARK}', "");
Expect(1, 918000, '\p{^	_combining_MARK}', "");
Expect(1, 918000, '\P{	_combining_MARK}', "");
Expect(0, 918000, '\P{^	_combining_MARK}', "");
Error('\p{:=--is_Combining_Mark}');
Error('\P{:=--is_Combining_Mark}');
Expect(1, 917999, '\p{iscombiningmark}', "");
Expect(0, 917999, '\p{^iscombiningmark}', "");
Expect(0, 917999, '\P{iscombiningmark}', "");
Expect(1, 917999, '\P{^iscombiningmark}', "");
Expect(0, 918000, '\p{iscombiningmark}', "");
Expect(1, 918000, '\p{^iscombiningmark}', "");
Expect(1, 918000, '\P{iscombiningmark}', "");
Expect(0, 918000, '\P{^iscombiningmark}', "");
Expect(1, 917999, '\p{ -Is_Combining_Mark}', "");
Expect(0, 917999, '\p{^ -Is_Combining_Mark}', "");
Expect(0, 917999, '\P{ -Is_Combining_Mark}', "");
Expect(1, 917999, '\P{^ -Is_Combining_Mark}', "");
Expect(0, 918000, '\p{ -Is_Combining_Mark}', "");
Expect(1, 918000, '\p{^ -Is_Combining_Mark}', "");
Expect(1, 918000, '\P{ -Is_Combining_Mark}', "");
Expect(0, 918000, '\P{^ -Is_Combining_Mark}', "");
Error('\p{-/a/ENCLOSED_alphanumerics}');
Error('\P{-/a/ENCLOSED_alphanumerics}');
Expect(1, 9471, '\p{enclosedalphanumerics}', "");
Expect(0, 9471, '\p{^enclosedalphanumerics}', "");
Expect(0, 9471, '\P{enclosedalphanumerics}', "");
Expect(1, 9471, '\P{^enclosedalphanumerics}', "");
Expect(0, 9472, '\p{enclosedalphanumerics}', "");
Expect(1, 9472, '\p{^enclosedalphanumerics}', "");
Expect(1, 9472, '\P{enclosedalphanumerics}', "");
Expect(0, 9472, '\P{^enclosedalphanumerics}', "");
Expect(1, 9471, '\p{ -Enclosed_Alphanumerics}', "");
Expect(0, 9471, '\p{^ -Enclosed_Alphanumerics}', "");
Expect(0, 9471, '\P{ -Enclosed_Alphanumerics}', "");
Expect(1, 9471, '\P{^ -Enclosed_Alphanumerics}', "");
Expect(0, 9472, '\p{ -Enclosed_Alphanumerics}', "");
Expect(1, 9472, '\p{^ -Enclosed_Alphanumerics}', "");
Expect(1, 9472, '\P{ -Enclosed_Alphanumerics}', "");
Expect(0, 9472, '\P{^ -Enclosed_Alphanumerics}', "");
Error('\p{/a/	_is_enclosed_alphanumerics}');
Error('\P{/a/	_is_enclosed_alphanumerics}');
Expect(1, 9471, '\p{isenclosedalphanumerics}', "");
Expect(0, 9471, '\p{^isenclosedalphanumerics}', "");
Expect(0, 9471, '\P{isenclosedalphanumerics}', "");
Expect(1, 9471, '\P{^isenclosedalphanumerics}', "");
Expect(0, 9472, '\p{isenclosedalphanumerics}', "");
Expect(1, 9472, '\p{^isenclosedalphanumerics}', "");
Expect(1, 9472, '\P{isenclosedalphanumerics}', "");
Expect(0, 9472, '\P{^isenclosedalphanumerics}', "");
Expect(1, 9471, '\p{--Is_ENCLOSED_ALPHANUMERICS}', "");
Expect(0, 9471, '\p{^--Is_ENCLOSED_ALPHANUMERICS}', "");
Expect(0, 9471, '\P{--Is_ENCLOSED_ALPHANUMERICS}', "");
Expect(1, 9471, '\P{^--Is_ENCLOSED_ALPHANUMERICS}', "");
Expect(0, 9472, '\p{--Is_ENCLOSED_ALPHANUMERICS}', "");
Expect(1, 9472, '\p{^--Is_ENCLOSED_ALPHANUMERICS}', "");
Expect(1, 9472, '\P{--Is_ENCLOSED_ALPHANUMERICS}', "");
Expect(0, 9472, '\P{^--Is_ENCLOSED_ALPHANUMERICS}', "");
Error('\p{/a/_In_Enclosed_Alphanumerics}');
Error('\P{/a/_In_Enclosed_Alphanumerics}');
Expect(1, 9471, '\p{inenclosedalphanumerics}', "");
Expect(0, 9471, '\p{^inenclosedalphanumerics}', "");
Expect(0, 9471, '\P{inenclosedalphanumerics}', "");
Expect(1, 9471, '\P{^inenclosedalphanumerics}', "");
Expect(0, 9472, '\p{inenclosedalphanumerics}', "");
Expect(1, 9472, '\p{^inenclosedalphanumerics}', "");
Expect(1, 9472, '\P{inenclosedalphanumerics}', "");
Expect(0, 9472, '\P{^inenclosedalphanumerics}', "");
Expect(1, 9471, '\p{_In_Enclosed_Alphanumerics}', "");
Expect(0, 9471, '\p{^_In_Enclosed_Alphanumerics}', "");
Expect(0, 9471, '\P{_In_Enclosed_Alphanumerics}', "");
Expect(1, 9471, '\P{^_In_Enclosed_Alphanumerics}', "");
Expect(0, 9472, '\p{_In_Enclosed_Alphanumerics}', "");
Expect(1, 9472, '\p{^_In_Enclosed_Alphanumerics}', "");
Expect(1, 9472, '\P{_In_Enclosed_Alphanumerics}', "");
Expect(0, 9472, '\P{^_In_Enclosed_Alphanumerics}', "");
Error('\p{_/a/ENCLOSED_ALPHANUM}');
Error('\P{_/a/ENCLOSED_ALPHANUM}');
Expect(1, 9471, '\p{enclosedalphanum}', "");
Expect(0, 9471, '\p{^enclosedalphanum}', "");
Expect(0, 9471, '\P{enclosedalphanum}', "");
Expect(1, 9471, '\P{^enclosedalphanum}', "");
Expect(0, 9472, '\p{enclosedalphanum}', "");
Expect(1, 9472, '\p{^enclosedalphanum}', "");
Expect(1, 9472, '\P{enclosedalphanum}', "");
Expect(0, 9472, '\P{^enclosedalphanum}', "");
Expect(1, 9471, '\p{-Enclosed_ALPHANUM}', "");
Expect(0, 9471, '\p{^-Enclosed_ALPHANUM}', "");
Expect(0, 9471, '\P{-Enclosed_ALPHANUM}', "");
Expect(1, 9471, '\P{^-Enclosed_ALPHANUM}', "");
Expect(0, 9472, '\p{-Enclosed_ALPHANUM}', "");
Expect(1, 9472, '\p{^-Enclosed_ALPHANUM}', "");
Expect(1, 9472, '\P{-Enclosed_ALPHANUM}', "");
Expect(0, 9472, '\P{^-Enclosed_ALPHANUM}', "");
Error('\p{Is_ENCLOSED_Alphanum/a/}');
Error('\P{Is_ENCLOSED_Alphanum/a/}');
Expect(1, 9471, '\p{isenclosedalphanum}', "");
Expect(0, 9471, '\p{^isenclosedalphanum}', "");
Expect(0, 9471, '\P{isenclosedalphanum}', "");
Expect(1, 9471, '\P{^isenclosedalphanum}', "");
Expect(0, 9472, '\p{isenclosedalphanum}', "");
Expect(1, 9472, '\p{^isenclosedalphanum}', "");
Expect(1, 9472, '\P{isenclosedalphanum}', "");
Expect(0, 9472, '\P{^isenclosedalphanum}', "");
Expect(1, 9471, '\p{-	is_ENCLOSED_alphanum}', "");
Expect(0, 9471, '\p{^-	is_ENCLOSED_alphanum}', "");
Expect(0, 9471, '\P{-	is_ENCLOSED_alphanum}', "");
Expect(1, 9471, '\P{^-	is_ENCLOSED_alphanum}', "");
Expect(0, 9472, '\p{-	is_ENCLOSED_alphanum}', "");
Expect(1, 9472, '\p{^-	is_ENCLOSED_alphanum}', "");
Expect(1, 9472, '\P{-	is_ENCLOSED_alphanum}', "");
Expect(0, 9472, '\P{^-	is_ENCLOSED_alphanum}', "");
Error('\p{-:=IN_Enclosed_ALPHANUM}');
Error('\P{-:=IN_Enclosed_ALPHANUM}');
Expect(1, 9471, '\p{inenclosedalphanum}', "");
Expect(0, 9471, '\p{^inenclosedalphanum}', "");
Expect(0, 9471, '\P{inenclosedalphanum}', "");
Expect(1, 9471, '\P{^inenclosedalphanum}', "");
Expect(0, 9472, '\p{inenclosedalphanum}', "");
Expect(1, 9472, '\p{^inenclosedalphanum}', "");
Expect(1, 9472, '\P{inenclosedalphanum}', "");
Expect(0, 9472, '\P{^inenclosedalphanum}', "");
Expect(1, 9471, '\p{__in_Enclosed_alphanum}', "");
Expect(0, 9471, '\p{^__in_Enclosed_alphanum}', "");
Expect(0, 9471, '\P{__in_Enclosed_alphanum}', "");
Expect(1, 9471, '\P{^__in_Enclosed_alphanum}', "");
Expect(0, 9472, '\p{__in_Enclosed_alphanum}', "");
Expect(1, 9472, '\p{^__in_Enclosed_alphanum}', "");
Expect(1, 9472, '\P{__in_Enclosed_alphanum}', "");
Expect(0, 9472, '\P{^__in_Enclosed_alphanum}', "");
Error('\p{ :=MEETEI_Mayek}');
Error('\P{ :=MEETEI_Mayek}');
Expect(1, 44025, '\p{meeteimayek}', "");
Expect(0, 44025, '\p{^meeteimayek}', "");
Expect(0, 44025, '\P{meeteimayek}', "");
Expect(1, 44025, '\P{^meeteimayek}', "");
Expect(0, 44026, '\p{meeteimayek}', "");
Expect(1, 44026, '\p{^meeteimayek}', "");
Expect(1, 44026, '\P{meeteimayek}', "");
Expect(0, 44026, '\P{^meeteimayek}', "");
Expect(1, 44025, '\p{-	meetei_Mayek}', "");
Expect(0, 44025, '\p{^-	meetei_Mayek}', "");
Expect(0, 44025, '\P{-	meetei_Mayek}', "");
Expect(1, 44025, '\P{^-	meetei_Mayek}', "");
Expect(0, 44026, '\p{-	meetei_Mayek}', "");
Expect(1, 44026, '\p{^-	meetei_Mayek}', "");
Expect(1, 44026, '\P{-	meetei_Mayek}', "");
Expect(0, 44026, '\P{^-	meetei_Mayek}', "");
Error('\p{_/a/Is_meetei_mayek}');
Error('\P{_/a/Is_meetei_mayek}');
Expect(1, 44025, '\p{ismeeteimayek}', "");
Expect(0, 44025, '\p{^ismeeteimayek}', "");
Expect(0, 44025, '\P{ismeeteimayek}', "");
Expect(1, 44025, '\P{^ismeeteimayek}', "");
Expect(0, 44026, '\p{ismeeteimayek}', "");
Expect(1, 44026, '\p{^ismeeteimayek}', "");
Expect(1, 44026, '\P{ismeeteimayek}', "");
Expect(0, 44026, '\P{^ismeeteimayek}', "");
Expect(1, 44025, '\p{-IS_Meetei_MAYEK}', "");
Expect(0, 44025, '\p{^-IS_Meetei_MAYEK}', "");
Expect(0, 44025, '\P{-IS_Meetei_MAYEK}', "");
Expect(1, 44025, '\P{^-IS_Meetei_MAYEK}', "");
Expect(0, 44026, '\p{-IS_Meetei_MAYEK}', "");
Expect(1, 44026, '\p{^-IS_Meetei_MAYEK}', "");
Expect(1, 44026, '\P{-IS_Meetei_MAYEK}', "");
Expect(0, 44026, '\P{^-IS_Meetei_MAYEK}', "");
Error('\p{	:=Mtei}');
Error('\P{	:=Mtei}');
Expect(1, 44025, '\p{mtei}', "");
Expect(0, 44025, '\p{^mtei}', "");
Expect(0, 44025, '\P{mtei}', "");
Expect(1, 44025, '\P{^mtei}', "");
Expect(0, 44026, '\p{mtei}', "");
Expect(1, 44026, '\p{^mtei}', "");
Expect(1, 44026, '\P{mtei}', "");
Expect(0, 44026, '\P{^mtei}', "");
Expect(1, 44025, '\p{-MTEI}', "");
Expect(0, 44025, '\p{^-MTEI}', "");
Expect(0, 44025, '\P{-MTEI}', "");
Expect(1, 44025, '\P{^-MTEI}', "");
Expect(0, 44026, '\p{-MTEI}', "");
Expect(1, 44026, '\p{^-MTEI}', "");
Expect(1, 44026, '\P{-MTEI}', "");
Expect(0, 44026, '\P{^-MTEI}', "");
Error('\p{_	IS_MTEI:=}');
Error('\P{_	IS_MTEI:=}');
Expect(1, 44025, '\p{ismtei}', "");
Expect(0, 44025, '\p{^ismtei}', "");
Expect(0, 44025, '\P{ismtei}', "");
Expect(1, 44025, '\P{^ismtei}', "");
Expect(0, 44026, '\p{ismtei}', "");
Expect(1, 44026, '\p{^ismtei}', "");
Expect(1, 44026, '\P{ismtei}', "");
Expect(0, 44026, '\P{^ismtei}', "");
Expect(1, 44025, '\p{__Is_Mtei}', "");
Expect(0, 44025, '\p{^__Is_Mtei}', "");
Expect(0, 44025, '\P{__Is_Mtei}', "");
Expect(1, 44025, '\P{^__Is_Mtei}', "");
Expect(0, 44026, '\p{__Is_Mtei}', "");
Expect(1, 44026, '\p{^__Is_Mtei}', "");
Expect(1, 44026, '\P{__Is_Mtei}', "");
Expect(0, 44026, '\P{^__Is_Mtei}', "");
Error('\p{:=-extender}');
Error('\P{:=-extender}');
Expect(1, 65392, '\p{extender}', "");
Expect(0, 65392, '\p{^extender}', "");
Expect(0, 65392, '\P{extender}', "");
Expect(1, 65392, '\P{^extender}', "");
Expect(0, 65393, '\p{extender}', "");
Expect(1, 65393, '\p{^extender}', "");
Expect(1, 65393, '\P{extender}', "");
Expect(0, 65393, '\P{^extender}', "");
Expect(1, 65392, '\p{	extender}', "");
Expect(0, 65392, '\p{^	extender}', "");
Expect(0, 65392, '\P{	extender}', "");
Expect(1, 65392, '\P{^	extender}', "");
Expect(0, 65393, '\p{	extender}', "");
Expect(1, 65393, '\p{^	extender}', "");
Expect(1, 65393, '\P{	extender}', "");
Expect(0, 65393, '\P{^	extender}', "");
Error('\p{/a/_	IS_EXTENDER}');
Error('\P{/a/_	IS_EXTENDER}');
Expect(1, 65392, '\p{isextender}', "");
Expect(0, 65392, '\p{^isextender}', "");
Expect(0, 65392, '\P{isextender}', "");
Expect(1, 65392, '\P{^isextender}', "");
Expect(0, 65393, '\p{isextender}', "");
Expect(1, 65393, '\p{^isextender}', "");
Expect(1, 65393, '\P{isextender}', "");
Expect(0, 65393, '\P{^isextender}', "");
Expect(1, 65392, '\p{- Is_Extender}', "");
Expect(0, 65392, '\p{^- Is_Extender}', "");
Expect(0, 65392, '\P{- Is_Extender}', "");
Expect(1, 65392, '\P{^- Is_Extender}', "");
Expect(0, 65393, '\p{- Is_Extender}', "");
Expect(1, 65393, '\p{^- Is_Extender}', "");
Expect(1, 65393, '\P{- Is_Extender}', "");
Expect(0, 65393, '\P{^- Is_Extender}', "");
Error('\p{-EXT:=}');
Error('\P{-EXT:=}');
Expect(1, 65392, '\p{ext}', "");
Expect(0, 65392, '\p{^ext}', "");
Expect(0, 65392, '\P{ext}', "");
Expect(1, 65392, '\P{^ext}', "");
Expect(0, 65393, '\p{ext}', "");
Expect(1, 65393, '\p{^ext}', "");
Expect(1, 65393, '\P{ext}', "");
Expect(0, 65393, '\P{^ext}', "");
Expect(1, 65392, '\p{-	ext}', "");
Expect(0, 65392, '\p{^-	ext}', "");
Expect(0, 65392, '\P{-	ext}', "");
Expect(1, 65392, '\P{^-	ext}', "");
Expect(0, 65393, '\p{-	ext}', "");
Expect(1, 65393, '\p{^-	ext}', "");
Expect(1, 65393, '\P{-	ext}', "");
Expect(0, 65393, '\P{^-	ext}', "");
Error('\p{-/a/is_Ext}');
Error('\P{-/a/is_Ext}');
Expect(1, 65392, '\p{isext}', "");
Expect(0, 65392, '\p{^isext}', "");
Expect(0, 65392, '\P{isext}', "");
Expect(1, 65392, '\P{^isext}', "");
Expect(0, 65393, '\p{isext}', "");
Expect(1, 65393, '\p{^isext}', "");
Expect(1, 65393, '\P{isext}', "");
Expect(0, 65393, '\P{^isext}', "");
Expect(1, 65392, '\p{		Is_Ext}', "");
Expect(0, 65392, '\p{^		Is_Ext}', "");
Expect(0, 65392, '\P{		Is_Ext}', "");
Expect(1, 65392, '\P{^		Is_Ext}', "");
Expect(0, 65393, '\p{		Is_Ext}', "");
Expect(1, 65393, '\p{^		Is_Ext}', "");
Expect(1, 65393, '\P{		Is_Ext}', "");
Expect(0, 65393, '\P{^		Is_Ext}', "");
Error('\p{-:=Composition_EXCLUSION}');
Error('\P{-:=Composition_EXCLUSION}');
Expect(1, 119232, '\p{compositionexclusion}', "");
Expect(0, 119232, '\p{^compositionexclusion}', "");
Expect(0, 119232, '\P{compositionexclusion}', "");
Expect(1, 119232, '\P{^compositionexclusion}', "");
Expect(0, 119233, '\p{compositionexclusion}', "");
Expect(1, 119233, '\p{^compositionexclusion}', "");
Expect(1, 119233, '\P{compositionexclusion}', "");
Expect(0, 119233, '\P{^compositionexclusion}', "");
Expect(1, 119232, '\p{ _Composition_Exclusion}', "");
Expect(0, 119232, '\p{^ _Composition_Exclusion}', "");
Expect(0, 119232, '\P{ _Composition_Exclusion}', "");
Expect(1, 119232, '\P{^ _Composition_Exclusion}', "");
Expect(0, 119233, '\p{ _Composition_Exclusion}', "");
Expect(1, 119233, '\p{^ _Composition_Exclusion}', "");
Expect(1, 119233, '\P{ _Composition_Exclusion}', "");
Expect(0, 119233, '\P{^ _Composition_Exclusion}', "");
Error('\p{:= Is_COMPOSITION_Exclusion}');
Error('\P{:= Is_COMPOSITION_Exclusion}');
Expect(1, 119232, '\p{iscompositionexclusion}', "");
Expect(0, 119232, '\p{^iscompositionexclusion}', "");
Expect(0, 119232, '\P{iscompositionexclusion}', "");
Expect(1, 119232, '\P{^iscompositionexclusion}', "");
Expect(0, 119233, '\p{iscompositionexclusion}', "");
Expect(1, 119233, '\p{^iscompositionexclusion}', "");
Expect(1, 119233, '\P{iscompositionexclusion}', "");
Expect(0, 119233, '\P{^iscompositionexclusion}', "");
Expect(1, 119232, '\p{	IS_Composition_exclusion}', "");
Expect(0, 119232, '\p{^	IS_Composition_exclusion}', "");
Expect(0, 119232, '\P{	IS_Composition_exclusion}', "");
Expect(1, 119232, '\P{^	IS_Composition_exclusion}', "");
Expect(0, 119233, '\p{	IS_Composition_exclusion}', "");
Expect(1, 119233, '\p{^	IS_Composition_exclusion}', "");
Expect(1, 119233, '\P{	IS_Composition_exclusion}', "");
Expect(0, 119233, '\P{^	IS_Composition_exclusion}', "");
Error('\p{	_ce/a/}');
Error('\P{	_ce/a/}');
Expect(1, 119232, '\p{ce}', "");
Expect(0, 119232, '\p{^ce}', "");
Expect(0, 119232, '\P{ce}', "");
Expect(1, 119232, '\P{^ce}', "");
Expect(0, 119233, '\p{ce}', "");
Expect(1, 119233, '\p{^ce}', "");
Expect(1, 119233, '\P{ce}', "");
Expect(0, 119233, '\P{^ce}', "");
Expect(1, 119232, '\p{_	CE}', "");
Expect(0, 119232, '\p{^_	CE}', "");
Expect(0, 119232, '\P{_	CE}', "");
Expect(1, 119232, '\P{^_	CE}', "");
Expect(0, 119233, '\p{_	CE}', "");
Expect(1, 119233, '\p{^_	CE}', "");
Expect(1, 119233, '\P{_	CE}', "");
Expect(0, 119233, '\P{^_	CE}', "");
Error('\p{-/a/Is_CE}');
Error('\P{-/a/Is_CE}');
Expect(1, 119232, '\p{isce}', "");
Expect(0, 119232, '\p{^isce}', "");
Expect(0, 119232, '\P{isce}', "");
Expect(1, 119232, '\P{^isce}', "");
Expect(0, 119233, '\p{isce}', "");
Expect(1, 119233, '\p{^isce}', "");
Expect(1, 119233, '\P{isce}', "");
Expect(0, 119233, '\P{^isce}', "");
Expect(1, 119232, '\p{ 	Is_CE}', "");
Expect(0, 119232, '\p{^ 	Is_CE}', "");
Expect(0, 119232, '\P{ 	Is_CE}', "");
Expect(1, 119232, '\P{^ 	Is_CE}', "");
Expect(0, 119233, '\p{ 	Is_CE}', "");
Expect(1, 119233, '\p{^ 	Is_CE}', "");
Expect(1, 119233, '\P{ 	Is_CE}', "");
Expect(0, 119233, '\P{^ 	Is_CE}', "");
Error('\p{	/a/ENCLOSED_CJK_Letters_And_Months}');
Error('\P{	/a/ENCLOSED_CJK_Letters_And_Months}');
Expect(1, 13055, '\p{enclosedcjklettersandmonths}', "");
Expect(0, 13055, '\p{^enclosedcjklettersandmonths}', "");
Expect(0, 13055, '\P{enclosedcjklettersandmonths}', "");
Expect(1, 13055, '\P{^enclosedcjklettersandmonths}', "");
Expect(0, 13056, '\p{enclosedcjklettersandmonths}', "");
Expect(1, 13056, '\p{^enclosedcjklettersandmonths}', "");
Expect(1, 13056, '\P{enclosedcjklettersandmonths}', "");
Expect(0, 13056, '\P{^enclosedcjklettersandmonths}', "");
Expect(1, 13055, '\p{ Enclosed_CJK_LETTERS_And_Months}', "");
Expect(0, 13055, '\p{^ Enclosed_CJK_LETTERS_And_Months}', "");
Expect(0, 13055, '\P{ Enclosed_CJK_LETTERS_And_Months}', "");
Expect(1, 13055, '\P{^ Enclosed_CJK_LETTERS_And_Months}', "");
Expect(0, 13056, '\p{ Enclosed_CJK_LETTERS_And_Months}', "");
Expect(1, 13056, '\p{^ Enclosed_CJK_LETTERS_And_Months}', "");
Expect(1, 13056, '\P{ Enclosed_CJK_LETTERS_And_Months}', "");
Expect(0, 13056, '\P{^ Enclosed_CJK_LETTERS_And_Months}', "");
Error('\p{_/a/Is_ENCLOSED_CJK_LETTERS_And_months}');
Error('\P{_/a/Is_ENCLOSED_CJK_LETTERS_And_months}');
Expect(1, 13055, '\p{isenclosedcjklettersandmonths}', "");
Expect(0, 13055, '\p{^isenclosedcjklettersandmonths}', "");
Expect(0, 13055, '\P{isenclosedcjklettersandmonths}', "");
Expect(1, 13055, '\P{^isenclosedcjklettersandmonths}', "");
Expect(0, 13056, '\p{isenclosedcjklettersandmonths}', "");
Expect(1, 13056, '\p{^isenclosedcjklettersandmonths}', "");
Expect(1, 13056, '\P{isenclosedcjklettersandmonths}', "");
Expect(0, 13056, '\P{^isenclosedcjklettersandmonths}', "");
Expect(1, 13055, '\p{ Is_Enclosed_CJK_Letters_and_Months}', "");
Expect(0, 13055, '\p{^ Is_Enclosed_CJK_Letters_and_Months}', "");
Expect(0, 13055, '\P{ Is_Enclosed_CJK_Letters_and_Months}', "");
Expect(1, 13055, '\P{^ Is_Enclosed_CJK_Letters_and_Months}', "");
Expect(0, 13056, '\p{ Is_Enclosed_CJK_Letters_and_Months}', "");
Expect(1, 13056, '\p{^ Is_Enclosed_CJK_Letters_and_Months}', "");
Expect(1, 13056, '\P{ Is_Enclosed_CJK_Letters_and_Months}', "");
Expect(0, 13056, '\P{^ Is_Enclosed_CJK_Letters_and_Months}', "");
Error('\p{-:=In_Enclosed_CJK_LETTERS_And_months}');
Error('\P{-:=In_Enclosed_CJK_LETTERS_And_months}');
Expect(1, 13055, '\p{inenclosedcjklettersandmonths}', "");
Expect(0, 13055, '\p{^inenclosedcjklettersandmonths}', "");
Expect(0, 13055, '\P{inenclosedcjklettersandmonths}', "");
Expect(1, 13055, '\P{^inenclosedcjklettersandmonths}', "");
Expect(0, 13056, '\p{inenclosedcjklettersandmonths}', "");
Expect(1, 13056, '\p{^inenclosedcjklettersandmonths}', "");
Expect(1, 13056, '\P{inenclosedcjklettersandmonths}', "");
Expect(0, 13056, '\P{^inenclosedcjklettersandmonths}', "");
Expect(1, 13055, '\p{ _in_Enclosed_cjk_Letters_And_months}', "");
Expect(0, 13055, '\p{^ _in_Enclosed_cjk_Letters_And_months}', "");
Expect(0, 13055, '\P{ _in_Enclosed_cjk_Letters_And_months}', "");
Expect(1, 13055, '\P{^ _in_Enclosed_cjk_Letters_And_months}', "");
Expect(0, 13056, '\p{ _in_Enclosed_cjk_Letters_And_months}', "");
Expect(1, 13056, '\p{^ _in_Enclosed_cjk_Letters_And_months}', "");
Expect(1, 13056, '\P{ _in_Enclosed_cjk_Letters_And_months}', "");
Expect(0, 13056, '\P{^ _in_Enclosed_cjk_Letters_And_months}', "");
Error('\p{-	Enclosed_CJK/a/}');
Error('\P{-	Enclosed_CJK/a/}');
Expect(1, 13055, '\p{enclosedcjk}', "");
Expect(0, 13055, '\p{^enclosedcjk}', "");
Expect(0, 13055, '\P{enclosedcjk}', "");
Expect(1, 13055, '\P{^enclosedcjk}', "");
Expect(0, 13056, '\p{enclosedcjk}', "");
Expect(1, 13056, '\p{^enclosedcjk}', "");
Expect(1, 13056, '\P{enclosedcjk}', "");
Expect(0, 13056, '\P{^enclosedcjk}', "");
Expect(1, 13055, '\p{ -enclosed_CJK}', "");
Expect(0, 13055, '\p{^ -enclosed_CJK}', "");
Expect(0, 13055, '\P{ -enclosed_CJK}', "");
Expect(1, 13055, '\P{^ -enclosed_CJK}', "");
Expect(0, 13056, '\p{ -enclosed_CJK}', "");
Expect(1, 13056, '\p{^ -enclosed_CJK}', "");
Expect(1, 13056, '\P{ -enclosed_CJK}', "");
Expect(0, 13056, '\P{^ -enclosed_CJK}', "");
Error('\p{-Is_ENCLOSED_CJK:=}');
Error('\P{-Is_ENCLOSED_CJK:=}');
Expect(1, 13055, '\p{isenclosedcjk}', "");
Expect(0, 13055, '\p{^isenclosedcjk}', "");
Expect(0, 13055, '\P{isenclosedcjk}', "");
Expect(1, 13055, '\P{^isenclosedcjk}', "");
Expect(0, 13056, '\p{isenclosedcjk}', "");
Expect(1, 13056, '\p{^isenclosedcjk}', "");
Expect(1, 13056, '\P{isenclosedcjk}', "");
Expect(0, 13056, '\P{^isenclosedcjk}', "");
Expect(1, 13055, '\p{	 IS_ENCLOSED_cjk}', "");
Expect(0, 13055, '\p{^	 IS_ENCLOSED_cjk}', "");
Expect(0, 13055, '\P{	 IS_ENCLOSED_cjk}', "");
Expect(1, 13055, '\P{^	 IS_ENCLOSED_cjk}', "");
Expect(0, 13056, '\p{	 IS_ENCLOSED_cjk}', "");
Expect(1, 13056, '\p{^	 IS_ENCLOSED_cjk}', "");
Expect(1, 13056, '\P{	 IS_ENCLOSED_cjk}', "");
Expect(0, 13056, '\P{^	 IS_ENCLOSED_cjk}', "");
Error('\p{ :=In_Enclosed_CJK}');
Error('\P{ :=In_Enclosed_CJK}');
Expect(1, 13055, '\p{inenclosedcjk}', "");
Expect(0, 13055, '\p{^inenclosedcjk}', "");
Expect(0, 13055, '\P{inenclosedcjk}', "");
Expect(1, 13055, '\P{^inenclosedcjk}', "");
Expect(0, 13056, '\p{inenclosedcjk}', "");
Expect(1, 13056, '\p{^inenclosedcjk}', "");
Expect(1, 13056, '\P{inenclosedcjk}', "");
Expect(0, 13056, '\P{^inenclosedcjk}', "");
Expect(1, 13055, '\p{-IN_Enclosed_CJK}', "");
Expect(0, 13055, '\p{^-IN_Enclosed_CJK}', "");
Expect(0, 13055, '\P{-IN_Enclosed_CJK}', "");
Expect(1, 13055, '\P{^-IN_Enclosed_CJK}', "");
Expect(0, 13056, '\p{-IN_Enclosed_CJK}', "");
Expect(1, 13056, '\p{^-IN_Enclosed_CJK}', "");
Expect(1, 13056, '\P{-IN_Enclosed_CJK}', "");
Expect(0, 13056, '\P{^-IN_Enclosed_CJK}', "");
Error('\p{	latin_EXTENDED_B/a/}');
Error('\P{	latin_EXTENDED_B/a/}');
Expect(1, 591, '\p{latinextendedb}', "");
Expect(0, 591, '\p{^latinextendedb}', "");
Expect(0, 591, '\P{latinextendedb}', "");
Expect(1, 591, '\P{^latinextendedb}', "");
Expect(0, 592, '\p{latinextendedb}', "");
Expect(1, 592, '\p{^latinextendedb}', "");
Expect(1, 592, '\P{latinextendedb}', "");
Expect(0, 592, '\P{^latinextendedb}', "");
Expect(1, 591, '\p{-	Latin_extended_B}', "");
Expect(0, 591, '\p{^-	Latin_extended_B}', "");
Expect(0, 591, '\P{-	Latin_extended_B}', "");
Expect(1, 591, '\P{^-	Latin_extended_B}', "");
Expect(0, 592, '\p{-	Latin_extended_B}', "");
Expect(1, 592, '\p{^-	Latin_extended_B}', "");
Expect(1, 592, '\P{-	Latin_extended_B}', "");
Expect(0, 592, '\P{^-	Latin_extended_B}', "");
Error('\p{_/a/Is_Latin_Extended_b}');
Error('\P{_/a/Is_Latin_Extended_b}');
Expect(1, 591, '\p{islatinextendedb}', "");
Expect(0, 591, '\p{^islatinextendedb}', "");
Expect(0, 591, '\P{islatinextendedb}', "");
Expect(1, 591, '\P{^islatinextendedb}', "");
Expect(0, 592, '\p{islatinextendedb}', "");
Expect(1, 592, '\p{^islatinextendedb}', "");
Expect(1, 592, '\P{islatinextendedb}', "");
Expect(0, 592, '\P{^islatinextendedb}', "");
Expect(1, 591, '\p{-Is_LATIN_extended_b}', "");
Expect(0, 591, '\p{^-Is_LATIN_extended_b}', "");
Expect(0, 591, '\P{-Is_LATIN_extended_b}', "");
Expect(1, 591, '\P{^-Is_LATIN_extended_b}', "");
Expect(0, 592, '\p{-Is_LATIN_extended_b}', "");
Expect(1, 592, '\p{^-Is_LATIN_extended_b}', "");
Expect(1, 592, '\P{-Is_LATIN_extended_b}', "");
Expect(0, 592, '\P{^-Is_LATIN_extended_b}', "");
Error('\p{/a/_	in_LATIN_Extended_b}');
Error('\P{/a/_	in_LATIN_Extended_b}');
Expect(1, 591, '\p{inlatinextendedb}', "");
Expect(0, 591, '\p{^inlatinextendedb}', "");
Expect(0, 591, '\P{inlatinextendedb}', "");
Expect(1, 591, '\P{^inlatinextendedb}', "");
Expect(0, 592, '\p{inlatinextendedb}', "");
Expect(1, 592, '\p{^inlatinextendedb}', "");
Expect(1, 592, '\P{inlatinextendedb}', "");
Expect(0, 592, '\P{^inlatinextendedb}', "");
Expect(1, 591, '\p{-_in_Latin_extended_B}', "");
Expect(0, 591, '\p{^-_in_Latin_extended_B}', "");
Expect(0, 591, '\P{-_in_Latin_extended_B}', "");
Expect(1, 591, '\P{^-_in_Latin_extended_B}', "");
Expect(0, 592, '\p{-_in_Latin_extended_B}', "");
Expect(1, 592, '\p{^-_in_Latin_extended_B}', "");
Expect(1, 592, '\P{-_in_Latin_extended_B}', "");
Expect(0, 592, '\P{^-_in_Latin_extended_B}', "");
Error('\p{__Latin_Ext_B:=}');
Error('\P{__Latin_Ext_B:=}');
Expect(1, 591, '\p{latinextb}', "");
Expect(0, 591, '\p{^latinextb}', "");
Expect(0, 591, '\P{latinextb}', "");
Expect(1, 591, '\P{^latinextb}', "");
Expect(0, 592, '\p{latinextb}', "");
Expect(1, 592, '\p{^latinextb}', "");
Expect(1, 592, '\P{latinextb}', "");
Expect(0, 592, '\P{^latinextb}', "");
Expect(1, 591, '\p{_LATIN_EXT_b}', "");
Expect(0, 591, '\p{^_LATIN_EXT_b}', "");
Expect(0, 591, '\P{_LATIN_EXT_b}', "");
Expect(1, 591, '\P{^_LATIN_EXT_b}', "");
Expect(0, 592, '\p{_LATIN_EXT_b}', "");
Expect(1, 592, '\p{^_LATIN_EXT_b}', "");
Expect(1, 592, '\P{_LATIN_EXT_b}', "");
Expect(0, 592, '\P{^_LATIN_EXT_b}', "");
Error('\p{	/a/Is_latin_Ext_B}');
Error('\P{	/a/Is_latin_Ext_B}');
Expect(1, 591, '\p{islatinextb}', "");
Expect(0, 591, '\p{^islatinextb}', "");
Expect(0, 591, '\P{islatinextb}', "");
Expect(1, 591, '\P{^islatinextb}', "");
Expect(0, 592, '\p{islatinextb}', "");
Expect(1, 592, '\p{^islatinextb}', "");
Expect(1, 592, '\P{islatinextb}', "");
Expect(0, 592, '\P{^islatinextb}', "");
Expect(1, 591, '\p{ -Is_LATIN_Ext_B}', "");
Expect(0, 591, '\p{^ -Is_LATIN_Ext_B}', "");
Expect(0, 591, '\P{ -Is_LATIN_Ext_B}', "");
Expect(1, 591, '\P{^ -Is_LATIN_Ext_B}', "");
Expect(0, 592, '\p{ -Is_LATIN_Ext_B}', "");
Expect(1, 592, '\p{^ -Is_LATIN_Ext_B}', "");
Expect(1, 592, '\P{ -Is_LATIN_Ext_B}', "");
Expect(0, 592, '\P{^ -Is_LATIN_Ext_B}', "");
Error('\p{:= In_latin_Ext_B}');
Error('\P{:= In_latin_Ext_B}');
Expect(1, 591, '\p{inlatinextb}', "");
Expect(0, 591, '\p{^inlatinextb}', "");
Expect(0, 591, '\P{inlatinextb}', "");
Expect(1, 591, '\P{^inlatinextb}', "");
Expect(0, 592, '\p{inlatinextb}', "");
Expect(1, 592, '\p{^inlatinextb}', "");
Expect(1, 592, '\P{inlatinextb}', "");
Expect(0, 592, '\P{^inlatinextb}', "");
Expect(1, 591, '\p{-_In_LATIN_Ext_B}', "");
Expect(0, 591, '\p{^-_In_LATIN_Ext_B}', "");
Expect(0, 591, '\P{-_In_LATIN_Ext_B}', "");
Expect(1, 591, '\P{^-_In_LATIN_Ext_B}', "");
Expect(0, 592, '\p{-_In_LATIN_Ext_B}', "");
Expect(1, 592, '\p{^-_In_LATIN_Ext_B}', "");
Expect(1, 592, '\P{-_In_LATIN_Ext_B}', "");
Expect(0, 592, '\P{^-_In_LATIN_Ext_B}', "");
Error('\p{:=__Space_Separator}');
Error('\P{:=__Space_Separator}');
Expect(1, 12288, '\p{spaceseparator}', "");
Expect(0, 12288, '\p{^spaceseparator}', "");
Expect(0, 12288, '\P{spaceseparator}', "");
Expect(1, 12288, '\P{^spaceseparator}', "");
Expect(0, 12289, '\p{spaceseparator}', "");
Expect(1, 12289, '\p{^spaceseparator}', "");
Expect(1, 12289, '\P{spaceseparator}', "");
Expect(0, 12289, '\P{^spaceseparator}', "");
Expect(1, 12288, '\p{ Space_Separator}', "");
Expect(0, 12288, '\p{^ Space_Separator}', "");
Expect(0, 12288, '\P{ Space_Separator}', "");
Expect(1, 12288, '\P{^ Space_Separator}', "");
Expect(0, 12289, '\p{ Space_Separator}', "");
Expect(1, 12289, '\p{^ Space_Separator}', "");
Expect(1, 12289, '\P{ Space_Separator}', "");
Expect(0, 12289, '\P{^ Space_Separator}', "");
Error('\p{:=	Is_Space_separator}');
Error('\P{:=	Is_Space_separator}');
Expect(1, 12288, '\p{isspaceseparator}', "");
Expect(0, 12288, '\p{^isspaceseparator}', "");
Expect(0, 12288, '\P{isspaceseparator}', "");
Expect(1, 12288, '\P{^isspaceseparator}', "");
Expect(0, 12289, '\p{isspaceseparator}', "");
Expect(1, 12289, '\p{^isspaceseparator}', "");
Expect(1, 12289, '\P{isspaceseparator}', "");
Expect(0, 12289, '\P{^isspaceseparator}', "");
Expect(1, 12288, '\p{ IS_SPACE_separator}', "");
Expect(0, 12288, '\p{^ IS_SPACE_separator}', "");
Expect(0, 12288, '\P{ IS_SPACE_separator}', "");
Expect(1, 12288, '\P{^ IS_SPACE_separator}', "");
Expect(0, 12289, '\p{ IS_SPACE_separator}', "");
Expect(1, 12289, '\p{^ IS_SPACE_separator}', "");
Expect(1, 12289, '\P{ IS_SPACE_separator}', "");
Expect(0, 12289, '\P{^ IS_SPACE_separator}', "");
Error('\p{_Zs:=}');
Error('\P{_Zs:=}');
Expect(1, 12288, '\p{zs}', "");
Expect(0, 12288, '\p{^zs}', "");
Expect(0, 12288, '\P{zs}', "");
Expect(1, 12288, '\P{^zs}', "");
Expect(0, 12289, '\p{zs}', "");
Expect(1, 12289, '\p{^zs}', "");
Expect(1, 12289, '\P{zs}', "");
Expect(0, 12289, '\P{^zs}', "");
Expect(1, 12288, '\p{ 	zs}', "");
Expect(0, 12288, '\p{^ 	zs}', "");
Expect(0, 12288, '\P{ 	zs}', "");
Expect(1, 12288, '\P{^ 	zs}', "");
Expect(0, 12289, '\p{ 	zs}', "");
Expect(1, 12289, '\p{^ 	zs}', "");
Expect(1, 12289, '\P{ 	zs}', "");
Expect(0, 12289, '\P{^ 	zs}', "");
Error('\p{_	Is_zs/a/}');
Error('\P{_	Is_zs/a/}');
Expect(1, 12288, '\p{iszs}', "");
Expect(0, 12288, '\p{^iszs}', "");
Expect(0, 12288, '\P{iszs}', "");
Expect(1, 12288, '\P{^iszs}', "");
Expect(0, 12289, '\p{iszs}', "");
Expect(1, 12289, '\p{^iszs}', "");
Expect(1, 12289, '\P{iszs}', "");
Expect(0, 12289, '\P{^iszs}', "");
Expect(1, 12288, '\p{ 	Is_Zs}', "");
Expect(0, 12288, '\p{^ 	Is_Zs}', "");
Expect(0, 12288, '\P{ 	Is_Zs}', "");
Expect(1, 12288, '\P{^ 	Is_Zs}', "");
Expect(0, 12289, '\p{ 	Is_Zs}', "");
Expect(1, 12289, '\p{^ 	Is_Zs}', "");
Expect(1, 12289, '\P{ 	Is_Zs}', "");
Expect(0, 12289, '\P{^ 	Is_Zs}', "");
Error('\p{-_Supplemental_Punctuation:=}');
Error('\P{-_Supplemental_Punctuation:=}');
Expect(1, 11903, '\p{supplementalpunctuation}', "");
Expect(0, 11903, '\p{^supplementalpunctuation}', "");
Expect(0, 11903, '\P{supplementalpunctuation}', "");
Expect(1, 11903, '\P{^supplementalpunctuation}', "");
Expect(0, 11904, '\p{supplementalpunctuation}', "");
Expect(1, 11904, '\p{^supplementalpunctuation}', "");
Expect(1, 11904, '\P{supplementalpunctuation}', "");
Expect(0, 11904, '\P{^supplementalpunctuation}', "");
Expect(1, 11903, '\p{ -SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11903, '\p{^ -SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11903, '\P{ -SUPPLEMENTAL_Punctuation}', "");
Expect(1, 11903, '\P{^ -SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11904, '\p{ -SUPPLEMENTAL_Punctuation}', "");
Expect(1, 11904, '\p{^ -SUPPLEMENTAL_Punctuation}', "");
Expect(1, 11904, '\P{ -SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11904, '\P{^ -SUPPLEMENTAL_Punctuation}', "");
Error('\p{/a/ 	IS_SUPPLEMENTAL_punctuation}');
Error('\P{/a/ 	IS_SUPPLEMENTAL_punctuation}');
Expect(1, 11903, '\p{issupplementalpunctuation}', "");
Expect(0, 11903, '\p{^issupplementalpunctuation}', "");
Expect(0, 11903, '\P{issupplementalpunctuation}', "");
Expect(1, 11903, '\P{^issupplementalpunctuation}', "");
Expect(0, 11904, '\p{issupplementalpunctuation}', "");
Expect(1, 11904, '\p{^issupplementalpunctuation}', "");
Expect(1, 11904, '\P{issupplementalpunctuation}', "");
Expect(0, 11904, '\P{^issupplementalpunctuation}', "");
Expect(1, 11903, '\p{	_is_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(0, 11903, '\p{^	_is_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(0, 11903, '\P{	_is_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(1, 11903, '\P{^	_is_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(0, 11904, '\p{	_is_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(1, 11904, '\p{^	_is_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(1, 11904, '\P{	_is_SUPPLEMENTAL_PUNCTUATION}', "");
Expect(0, 11904, '\P{^	_is_SUPPLEMENTAL_PUNCTUATION}', "");
Error('\p{ In_Supplemental_PUNCTUATION/a/}');
Error('\P{ In_Supplemental_PUNCTUATION/a/}');
Expect(1, 11903, '\p{insupplementalpunctuation}', "");
Expect(0, 11903, '\p{^insupplementalpunctuation}', "");
Expect(0, 11903, '\P{insupplementalpunctuation}', "");
Expect(1, 11903, '\P{^insupplementalpunctuation}', "");
Expect(0, 11904, '\p{insupplementalpunctuation}', "");
Expect(1, 11904, '\p{^insupplementalpunctuation}', "");
Expect(1, 11904, '\P{insupplementalpunctuation}', "");
Expect(0, 11904, '\P{^insupplementalpunctuation}', "");
Expect(1, 11903, '\p{_ IN_Supplemental_punctuation}', "");
Expect(0, 11903, '\p{^_ IN_Supplemental_punctuation}', "");
Expect(0, 11903, '\P{_ IN_Supplemental_punctuation}', "");
Expect(1, 11903, '\P{^_ IN_Supplemental_punctuation}', "");
Expect(0, 11904, '\p{_ IN_Supplemental_punctuation}', "");
Expect(1, 11904, '\p{^_ IN_Supplemental_punctuation}', "");
Expect(1, 11904, '\P{_ IN_Supplemental_punctuation}', "");
Expect(0, 11904, '\P{^_ IN_Supplemental_punctuation}', "");
Error('\p{/a/_	Sup_Punctuation}');
Error('\P{/a/_	Sup_Punctuation}');
Expect(1, 11903, '\p{suppunctuation}', "");
Expect(0, 11903, '\p{^suppunctuation}', "");
Expect(0, 11903, '\P{suppunctuation}', "");
Expect(1, 11903, '\P{^suppunctuation}', "");
Expect(0, 11904, '\p{suppunctuation}', "");
Expect(1, 11904, '\p{^suppunctuation}', "");
Expect(1, 11904, '\P{suppunctuation}', "");
Expect(0, 11904, '\P{^suppunctuation}', "");
Expect(1, 11903, '\p{	-SUP_Punctuation}', "");
Expect(0, 11903, '\p{^	-SUP_Punctuation}', "");
Expect(0, 11903, '\P{	-SUP_Punctuation}', "");
Expect(1, 11903, '\P{^	-SUP_Punctuation}', "");
Expect(0, 11904, '\p{	-SUP_Punctuation}', "");
Expect(1, 11904, '\p{^	-SUP_Punctuation}', "");
Expect(1, 11904, '\P{	-SUP_Punctuation}', "");
Expect(0, 11904, '\P{^	-SUP_Punctuation}', "");
Error('\p{	-Is_Sup_Punctuation/a/}');
Error('\P{	-Is_Sup_Punctuation/a/}');
Expect(1, 11903, '\p{issuppunctuation}', "");
Expect(0, 11903, '\p{^issuppunctuation}', "");
Expect(0, 11903, '\P{issuppunctuation}', "");
Expect(1, 11903, '\P{^issuppunctuation}', "");
Expect(0, 11904, '\p{issuppunctuation}', "");
Expect(1, 11904, '\p{^issuppunctuation}', "");
Expect(1, 11904, '\P{issuppunctuation}', "");
Expect(0, 11904, '\P{^issuppunctuation}', "");
Expect(1, 11903, '\p{	IS_Sup_punctuation}', "");
Expect(0, 11903, '\p{^	IS_Sup_punctuation}', "");
Expect(0, 11903, '\P{	IS_Sup_punctuation}', "");
Expect(1, 11903, '\P{^	IS_Sup_punctuation}', "");
Expect(0, 11904, '\p{	IS_Sup_punctuation}', "");
Expect(1, 11904, '\p{^	IS_Sup_punctuation}', "");
Expect(1, 11904, '\P{	IS_Sup_punctuation}', "");
Expect(0, 11904, '\P{^	IS_Sup_punctuation}', "");
Error('\p{:=In_SUP_Punctuation}');
Error('\P{:=In_SUP_Punctuation}');
Expect(1, 11903, '\p{insuppunctuation}', "");
Expect(0, 11903, '\p{^insuppunctuation}', "");
Expect(0, 11903, '\P{insuppunctuation}', "");
Expect(1, 11903, '\P{^insuppunctuation}', "");
Expect(0, 11904, '\p{insuppunctuation}', "");
Expect(1, 11904, '\p{^insuppunctuation}', "");
Expect(1, 11904, '\P{insuppunctuation}', "");
Expect(0, 11904, '\P{^insuppunctuation}', "");
Expect(1, 11903, '\p{_-IN_SUP_Punctuation}', "");
Expect(0, 11903, '\p{^_-IN_SUP_Punctuation}', "");
Expect(0, 11903, '\P{_-IN_SUP_Punctuation}', "");
Expect(1, 11903, '\P{^_-IN_SUP_Punctuation}', "");
Expect(0, 11904, '\p{_-IN_SUP_Punctuation}', "");
Expect(1, 11904, '\p{^_-IN_SUP_Punctuation}', "");
Expect(1, 11904, '\P{_-IN_SUP_Punctuation}', "");
Expect(0, 11904, '\P{^_-IN_SUP_Punctuation}', "");
Error('\p{_:=KATAKANA_Phonetic_EXTENSIONS}');
Error('\P{_:=KATAKANA_Phonetic_EXTENSIONS}');
Expect(1, 12799, '\p{katakanaphoneticextensions}', "");
Expect(0, 12799, '\p{^katakanaphoneticextensions}', "");
Expect(0, 12799, '\P{katakanaphoneticextensions}', "");
Expect(1, 12799, '\P{^katakanaphoneticextensions}', "");
Expect(0, 12800, '\p{katakanaphoneticextensions}', "");
Expect(1, 12800, '\p{^katakanaphoneticextensions}', "");
Expect(1, 12800, '\P{katakanaphoneticextensions}', "");
Expect(0, 12800, '\P{^katakanaphoneticextensions}', "");
Expect(1, 12799, '\p{katakana_Phonetic_Extensions}', "");
Expect(0, 12799, '\p{^katakana_Phonetic_Extensions}', "");
Expect(0, 12799, '\P{katakana_Phonetic_Extensions}', "");
Expect(1, 12799, '\P{^katakana_Phonetic_Extensions}', "");
Expect(0, 12800, '\p{katakana_Phonetic_Extensions}', "");
Expect(1, 12800, '\p{^katakana_Phonetic_Extensions}', "");
Expect(1, 12800, '\P{katakana_Phonetic_Extensions}', "");
Expect(0, 12800, '\P{^katakana_Phonetic_Extensions}', "");
Error('\p{/a/-Is_Katakana_phonetic_Extensions}');
Error('\P{/a/-Is_Katakana_phonetic_Extensions}');
Expect(1, 12799, '\p{iskatakanaphoneticextensions}', "");
Expect(0, 12799, '\p{^iskatakanaphoneticextensions}', "");
Expect(0, 12799, '\P{iskatakanaphoneticextensions}', "");
Expect(1, 12799, '\P{^iskatakanaphoneticextensions}', "");
Expect(0, 12800, '\p{iskatakanaphoneticextensions}', "");
Expect(1, 12800, '\p{^iskatakanaphoneticextensions}', "");
Expect(1, 12800, '\P{iskatakanaphoneticextensions}', "");
Expect(0, 12800, '\P{^iskatakanaphoneticextensions}', "");
Expect(1, 12799, '\p{	IS_Katakana_Phonetic_EXTENSIONS}', "");
Expect(0, 12799, '\p{^	IS_Katakana_Phonetic_EXTENSIONS}', "");
Expect(0, 12799, '\P{	IS_Katakana_Phonetic_EXTENSIONS}', "");
Expect(1, 12799, '\P{^	IS_Katakana_Phonetic_EXTENSIONS}', "");
Expect(0, 12800, '\p{	IS_Katakana_Phonetic_EXTENSIONS}', "");
Expect(1, 12800, '\p{^	IS_Katakana_Phonetic_EXTENSIONS}', "");
Expect(1, 12800, '\P{	IS_Katakana_Phonetic_EXTENSIONS}', "");
Expect(0, 12800, '\P{^	IS_Katakana_Phonetic_EXTENSIONS}', "");
Error('\p{-/a/In_KATAKANA_Phonetic_extensions}');
Error('\P{-/a/In_KATAKANA_Phonetic_extensions}');
Expect(1, 12799, '\p{inkatakanaphoneticextensions}', "");
Expect(0, 12799, '\p{^inkatakanaphoneticextensions}', "");
Expect(0, 12799, '\P{inkatakanaphoneticextensions}', "");
Expect(1, 12799, '\P{^inkatakanaphoneticextensions}', "");
Expect(0, 12800, '\p{inkatakanaphoneticextensions}', "");
Expect(1, 12800, '\p{^inkatakanaphoneticextensions}', "");
Expect(1, 12800, '\P{inkatakanaphoneticextensions}', "");
Expect(0, 12800, '\P{^inkatakanaphoneticextensions}', "");
Expect(1, 12799, '\p{-	in_Katakana_Phonetic_Extensions}', "");
Expect(0, 12799, '\p{^-	in_Katakana_Phonetic_Extensions}', "");
Expect(0, 12799, '\P{-	in_Katakana_Phonetic_Extensions}', "");
Expect(1, 12799, '\P{^-	in_Katakana_Phonetic_Extensions}', "");
Expect(0, 12800, '\p{-	in_Katakana_Phonetic_Extensions}', "");
Expect(1, 12800, '\p{^-	in_Katakana_Phonetic_Extensions}', "");
Expect(1, 12800, '\P{-	in_Katakana_Phonetic_Extensions}', "");
Expect(0, 12800, '\P{^-	in_Katakana_Phonetic_Extensions}', "");
Error('\p{_:=Katakana_ext}');
Error('\P{_:=Katakana_ext}');
Expect(1, 12799, '\p{katakanaext}', "");
Expect(0, 12799, '\p{^katakanaext}', "");
Expect(0, 12799, '\P{katakanaext}', "");
Expect(1, 12799, '\P{^katakanaext}', "");
Expect(0, 12800, '\p{katakanaext}', "");
Expect(1, 12800, '\p{^katakanaext}', "");
Expect(1, 12800, '\P{katakanaext}', "");
Expect(0, 12800, '\P{^katakanaext}', "");
Expect(1, 12799, '\p{-	Katakana_Ext}', "");
Expect(0, 12799, '\p{^-	Katakana_Ext}', "");
Expect(0, 12799, '\P{-	Katakana_Ext}', "");
Expect(1, 12799, '\P{^-	Katakana_Ext}', "");
Expect(0, 12800, '\p{-	Katakana_Ext}', "");
Expect(1, 12800, '\p{^-	Katakana_Ext}', "");
Expect(1, 12800, '\P{-	Katakana_Ext}', "");
Expect(0, 12800, '\P{^-	Katakana_Ext}', "");
Error('\p{	IS_katakana_Ext:=}');
Error('\P{	IS_katakana_Ext:=}');
Expect(1, 12799, '\p{iskatakanaext}', "");
Expect(0, 12799, '\p{^iskatakanaext}', "");
Expect(0, 12799, '\P{iskatakanaext}', "");
Expect(1, 12799, '\P{^iskatakanaext}', "");
Expect(0, 12800, '\p{iskatakanaext}', "");
Expect(1, 12800, '\p{^iskatakanaext}', "");
Expect(1, 12800, '\P{iskatakanaext}', "");
Expect(0, 12800, '\P{^iskatakanaext}', "");
Expect(1, 12799, '\p{_is_KATAKANA_Ext}', "");
Expect(0, 12799, '\p{^_is_KATAKANA_Ext}', "");
Expect(0, 12799, '\P{_is_KATAKANA_Ext}', "");
Expect(1, 12799, '\P{^_is_KATAKANA_Ext}', "");
Expect(0, 12800, '\p{_is_KATAKANA_Ext}', "");
Expect(1, 12800, '\p{^_is_KATAKANA_Ext}', "");
Expect(1, 12800, '\P{_is_KATAKANA_Ext}', "");
Expect(0, 12800, '\P{^_is_KATAKANA_Ext}', "");
Error('\p{:=In_KATAKANA_ext}');
Error('\P{:=In_KATAKANA_ext}');
Expect(1, 12799, '\p{inkatakanaext}', "");
Expect(0, 12799, '\p{^inkatakanaext}', "");
Expect(0, 12799, '\P{inkatakanaext}', "");
Expect(1, 12799, '\P{^inkatakanaext}', "");
Expect(0, 12800, '\p{inkatakanaext}', "");
Expect(1, 12800, '\p{^inkatakanaext}', "");
Expect(1, 12800, '\P{inkatakanaext}', "");
Expect(0, 12800, '\P{^inkatakanaext}', "");
Expect(1, 12799, '\p{	In_Katakana_ext}', "");
Expect(0, 12799, '\p{^	In_Katakana_ext}', "");
Expect(0, 12799, '\P{	In_Katakana_ext}', "");
Expect(1, 12799, '\P{^	In_Katakana_ext}', "");
Expect(0, 12800, '\p{	In_Katakana_ext}', "");
Expect(1, 12800, '\p{^	In_Katakana_ext}', "");
Expect(1, 12800, '\P{	In_Katakana_ext}', "");
Expect(0, 12800, '\P{^	In_Katakana_ext}', "");
Error('\p{:=-modifier_SYMBOL}');
Error('\P{:=-modifier_SYMBOL}');
Expect(1, 65507, '\p{modifiersymbol}', "");
Expect(0, 65507, '\p{^modifiersymbol}', "");
Expect(0, 65507, '\P{modifiersymbol}', "");
Expect(1, 65507, '\P{^modifiersymbol}', "");
Expect(0, 65508, '\p{modifiersymbol}', "");
Expect(1, 65508, '\p{^modifiersymbol}', "");
Expect(1, 65508, '\P{modifiersymbol}', "");
Expect(0, 65508, '\P{^modifiersymbol}', "");
Expect(1, 65507, '\p{_ modifier_Symbol}', "");
Expect(0, 65507, '\p{^_ modifier_Symbol}', "");
Expect(0, 65507, '\P{_ modifier_Symbol}', "");
Expect(1, 65507, '\P{^_ modifier_Symbol}', "");
Expect(0, 65508, '\p{_ modifier_Symbol}', "");
Expect(1, 65508, '\p{^_ modifier_Symbol}', "");
Expect(1, 65508, '\P{_ modifier_Symbol}', "");
Expect(0, 65508, '\P{^_ modifier_Symbol}', "");
Error('\p{-Is_Modifier_Symbol:=}');
Error('\P{-Is_Modifier_Symbol:=}');
Expect(1, 65507, '\p{ismodifiersymbol}', "");
Expect(0, 65507, '\p{^ismodifiersymbol}', "");
Expect(0, 65507, '\P{ismodifiersymbol}', "");
Expect(1, 65507, '\P{^ismodifiersymbol}', "");
Expect(0, 65508, '\p{ismodifiersymbol}', "");
Expect(1, 65508, '\p{^ismodifiersymbol}', "");
Expect(1, 65508, '\P{ismodifiersymbol}', "");
Expect(0, 65508, '\P{^ismodifiersymbol}', "");
Expect(1, 65507, '\p{__IS_modifier_Symbol}', "");
Expect(0, 65507, '\p{^__IS_modifier_Symbol}', "");
Expect(0, 65507, '\P{__IS_modifier_Symbol}', "");
Expect(1, 65507, '\P{^__IS_modifier_Symbol}', "");
Expect(0, 65508, '\p{__IS_modifier_Symbol}', "");
Expect(1, 65508, '\p{^__IS_modifier_Symbol}', "");
Expect(1, 65508, '\P{__IS_modifier_Symbol}', "");
Expect(0, 65508, '\P{^__IS_modifier_Symbol}', "");
Error('\p{:=--SK}');
Error('\P{:=--SK}');
Expect(1, 65507, '\p{sk}', "");
Expect(0, 65507, '\p{^sk}', "");
Expect(0, 65507, '\P{sk}', "");
Expect(1, 65507, '\P{^sk}', "");
Expect(0, 65508, '\p{sk}', "");
Expect(1, 65508, '\p{^sk}', "");
Expect(1, 65508, '\P{sk}', "");
Expect(0, 65508, '\P{^sk}', "");
Expect(1, 65507, '\p{-Sk}', "");
Expect(0, 65507, '\p{^-Sk}', "");
Expect(0, 65507, '\P{-Sk}', "");
Expect(1, 65507, '\P{^-Sk}', "");
Expect(0, 65508, '\p{-Sk}', "");
Expect(1, 65508, '\p{^-Sk}', "");
Expect(1, 65508, '\P{-Sk}', "");
Expect(0, 65508, '\P{^-Sk}', "");
Error('\p{-	Is_Sk:=}');
Error('\P{-	Is_Sk:=}');
Expect(1, 65507, '\p{issk}', "");
Expect(0, 65507, '\p{^issk}', "");
Expect(0, 65507, '\P{issk}', "");
Expect(1, 65507, '\P{^issk}', "");
Expect(0, 65508, '\p{issk}', "");
Expect(1, 65508, '\p{^issk}', "");
Expect(1, 65508, '\P{issk}', "");
Expect(0, 65508, '\P{^issk}', "");
Expect(1, 65507, '\p{	-IS_Sk}', "");
Expect(0, 65507, '\p{^	-IS_Sk}', "");
Expect(0, 65507, '\P{	-IS_Sk}', "");
Expect(1, 65507, '\P{^	-IS_Sk}', "");
Expect(0, 65508, '\p{	-IS_Sk}', "");
Expect(1, 65508, '\p{^	-IS_Sk}', "");
Expect(1, 65508, '\P{	-IS_Sk}', "");
Expect(0, 65508, '\P{^	-IS_Sk}', "");
Error('\p{:=	 ARMENIAN}');
Error('\P{:=	 ARMENIAN}');
Expect(1, 64279, '\p{armenian}', "");
Expect(0, 64279, '\p{^armenian}', "");
Expect(0, 64279, '\P{armenian}', "");
Expect(1, 64279, '\P{^armenian}', "");
Expect(0, 64280, '\p{armenian}', "");
Expect(1, 64280, '\p{^armenian}', "");
Expect(1, 64280, '\P{armenian}', "");
Expect(0, 64280, '\P{^armenian}', "");
Expect(1, 64279, '\p{	Armenian}', "");
Expect(0, 64279, '\p{^	Armenian}', "");
Expect(0, 64279, '\P{	Armenian}', "");
Expect(1, 64279, '\P{^	Armenian}', "");
Expect(0, 64280, '\p{	Armenian}', "");
Expect(1, 64280, '\p{^	Armenian}', "");
Expect(1, 64280, '\P{	Armenian}', "");
Expect(0, 64280, '\P{^	Armenian}', "");
Error('\p{_-Is_armenian/a/}');
Error('\P{_-Is_armenian/a/}');
Expect(1, 64279, '\p{isarmenian}', "");
Expect(0, 64279, '\p{^isarmenian}', "");
Expect(0, 64279, '\P{isarmenian}', "");
Expect(1, 64279, '\P{^isarmenian}', "");
Expect(0, 64280, '\p{isarmenian}', "");
Expect(1, 64280, '\p{^isarmenian}', "");
Expect(1, 64280, '\P{isarmenian}', "");
Expect(0, 64280, '\P{^isarmenian}', "");
Expect(1, 64279, '\p{ is_armenian}', "");
Expect(0, 64279, '\p{^ is_armenian}', "");
Expect(0, 64279, '\P{ is_armenian}', "");
Expect(1, 64279, '\P{^ is_armenian}', "");
Expect(0, 64280, '\p{ is_armenian}', "");
Expect(1, 64280, '\p{^ is_armenian}', "");
Expect(1, 64280, '\P{ is_armenian}', "");
Expect(0, 64280, '\P{^ is_armenian}', "");
Error('\p{__Armn/a/}');
Error('\P{__Armn/a/}');
Expect(1, 64279, '\p{armn}', "");
Expect(0, 64279, '\p{^armn}', "");
Expect(0, 64279, '\P{armn}', "");
Expect(1, 64279, '\P{^armn}', "");
Expect(0, 64280, '\p{armn}', "");
Expect(1, 64280, '\p{^armn}', "");
Expect(1, 64280, '\P{armn}', "");
Expect(0, 64280, '\P{^armn}', "");
Expect(1, 64279, '\p{-_Armn}', "");
Expect(0, 64279, '\p{^-_Armn}', "");
Expect(0, 64279, '\P{-_Armn}', "");
Expect(1, 64279, '\P{^-_Armn}', "");
Expect(0, 64280, '\p{-_Armn}', "");
Expect(1, 64280, '\p{^-_Armn}', "");
Expect(1, 64280, '\P{-_Armn}', "");
Expect(0, 64280, '\P{^-_Armn}', "");
Error('\p{:=	 IS_Armn}');
Error('\P{:=	 IS_Armn}');
Expect(1, 64279, '\p{isarmn}', "");
Expect(0, 64279, '\p{^isarmn}', "");
Expect(0, 64279, '\P{isarmn}', "");
Expect(1, 64279, '\P{^isarmn}', "");
Expect(0, 64280, '\p{isarmn}', "");
Expect(1, 64280, '\p{^isarmn}', "");
Expect(1, 64280, '\P{isarmn}', "");
Expect(0, 64280, '\P{^isarmn}', "");
Expect(1, 64279, '\p{-Is_Armn}', "");
Expect(0, 64279, '\p{^-Is_Armn}', "");
Expect(0, 64279, '\P{-Is_Armn}', "");
Expect(1, 64279, '\P{^-Is_Armn}', "");
Expect(0, 64280, '\p{-Is_Armn}', "");
Expect(1, 64280, '\p{^-Is_Armn}', "");
Expect(1, 64280, '\P{-Is_Armn}', "");
Expect(0, 64280, '\P{^-Is_Armn}', "");
Error('\p{-:=Hyphen}');
Error('\P{-:=Hyphen}');
Expect(1, 65381, '\p{hyphen}', 'deprecated');
Expect(0, 65381, '\p{^hyphen}', 'deprecated');
Expect(0, 65381, '\P{hyphen}', 'deprecated');
Expect(1, 65381, '\P{^hyphen}', 'deprecated');
Expect(0, 65382, '\p{hyphen}', 'deprecated');
Expect(1, 65382, '\p{^hyphen}', 'deprecated');
Expect(1, 65382, '\P{hyphen}', 'deprecated');
Expect(0, 65382, '\P{^hyphen}', 'deprecated');
Expect(1, 65381, '\p{_hyphen}', 'deprecated');
Expect(0, 65381, '\p{^_hyphen}', 'deprecated');
Expect(0, 65381, '\P{_hyphen}', 'deprecated');
Expect(1, 65381, '\P{^_hyphen}', 'deprecated');
Expect(0, 65382, '\p{_hyphen}', 'deprecated');
Expect(1, 65382, '\p{^_hyphen}', 'deprecated');
Expect(1, 65382, '\P{_hyphen}', 'deprecated');
Expect(0, 65382, '\P{^_hyphen}', 'deprecated');
Error('\p{ 	Is_Hyphen/a/}');
Error('\P{ 	Is_Hyphen/a/}');
Expect(1, 65381, '\p{ishyphen}', 'deprecated');
Expect(0, 65381, '\p{^ishyphen}', 'deprecated');
Expect(0, 65381, '\P{ishyphen}', 'deprecated');
Expect(1, 65381, '\P{^ishyphen}', 'deprecated');
Expect(0, 65382, '\p{ishyphen}', 'deprecated');
Expect(1, 65382, '\p{^ishyphen}', 'deprecated');
Expect(1, 65382, '\P{ishyphen}', 'deprecated');
Expect(0, 65382, '\P{^ishyphen}', 'deprecated');
Expect(1, 65381, '\p{	Is_Hyphen}', 'deprecated');
Expect(0, 65381, '\p{^	Is_Hyphen}', 'deprecated');
Expect(0, 65381, '\P{	Is_Hyphen}', 'deprecated');
Expect(1, 65381, '\P{^	Is_Hyphen}', 'deprecated');
Expect(0, 65382, '\p{	Is_Hyphen}', 'deprecated');
Expect(1, 65382, '\p{^	Is_Hyphen}', 'deprecated');
Expect(1, 65382, '\P{	Is_Hyphen}', 'deprecated');
Expect(0, 65382, '\P{^	Is_Hyphen}', 'deprecated');
Error('\p{-_XID_CONTINUE:=}');
Error('\P{-_XID_CONTINUE:=}');
Expect(1, 917999, '\p{xidcontinue}', "");
Expect(0, 917999, '\p{^xidcontinue}', "");
Expect(0, 917999, '\P{xidcontinue}', "");
Expect(1, 917999, '\P{^xidcontinue}', "");
Expect(0, 918000, '\p{xidcontinue}', "");
Expect(1, 918000, '\p{^xidcontinue}', "");
Expect(1, 918000, '\P{xidcontinue}', "");
Expect(0, 918000, '\P{^xidcontinue}', "");
Expect(1, 917999, '\p{	-XID_Continue}', "");
Expect(0, 917999, '\p{^	-XID_Continue}', "");
Expect(0, 917999, '\P{	-XID_Continue}', "");
Expect(1, 917999, '\P{^	-XID_Continue}', "");
Expect(0, 918000, '\p{	-XID_Continue}', "");
Expect(1, 918000, '\p{^	-XID_Continue}', "");
Expect(1, 918000, '\P{	-XID_Continue}', "");
Expect(0, 918000, '\P{^	-XID_Continue}', "");
Error('\p{/a/ IS_XID_Continue}');
Error('\P{/a/ IS_XID_Continue}');
Expect(1, 917999, '\p{isxidcontinue}', "");
Expect(0, 917999, '\p{^isxidcontinue}', "");
Expect(0, 917999, '\P{isxidcontinue}', "");
Expect(1, 917999, '\P{^isxidcontinue}', "");
Expect(0, 918000, '\p{isxidcontinue}', "");
Expect(1, 918000, '\p{^isxidcontinue}', "");
Expect(1, 918000, '\P{isxidcontinue}', "");
Expect(0, 918000, '\P{^isxidcontinue}', "");
Expect(1, 917999, '\p{	Is_xid_Continue}', "");
Expect(0, 917999, '\p{^	Is_xid_Continue}', "");
Expect(0, 917999, '\P{	Is_xid_Continue}', "");
Expect(1, 917999, '\P{^	Is_xid_Continue}', "");
Expect(0, 918000, '\p{	Is_xid_Continue}', "");
Expect(1, 918000, '\p{^	Is_xid_Continue}', "");
Expect(1, 918000, '\P{	Is_xid_Continue}', "");
Expect(0, 918000, '\P{^	Is_xid_Continue}', "");
Error('\p{	:=XIDC}');
Error('\P{	:=XIDC}');
Expect(1, 917999, '\p{xidc}', "");
Expect(0, 917999, '\p{^xidc}', "");
Expect(0, 917999, '\P{xidc}', "");
Expect(1, 917999, '\P{^xidc}', "");
Expect(0, 918000, '\p{xidc}', "");
Expect(1, 918000, '\p{^xidc}', "");
Expect(1, 918000, '\P{xidc}', "");
Expect(0, 918000, '\P{^xidc}', "");
Expect(1, 917999, '\p{XIDC}', "");
Expect(0, 917999, '\p{^XIDC}', "");
Expect(0, 917999, '\P{XIDC}', "");
Expect(1, 917999, '\P{^XIDC}', "");
Expect(0, 918000, '\p{XIDC}', "");
Expect(1, 918000, '\p{^XIDC}', "");
Expect(1, 918000, '\P{XIDC}', "");
Expect(0, 918000, '\P{^XIDC}', "");
Error('\p{:=- Is_XIDC}');
Error('\P{:=- Is_XIDC}');
Expect(1, 917999, '\p{isxidc}', "");
Expect(0, 917999, '\p{^isxidc}', "");
Expect(0, 917999, '\P{isxidc}', "");
Expect(1, 917999, '\P{^isxidc}', "");
Expect(0, 918000, '\p{isxidc}', "");
Expect(1, 918000, '\p{^isxidc}', "");
Expect(1, 918000, '\P{isxidc}', "");
Expect(0, 918000, '\P{^isxidc}', "");
Expect(1, 917999, '\p{-_Is_XIDC}', "");
Expect(0, 917999, '\p{^-_Is_XIDC}', "");
Expect(0, 917999, '\P{-_Is_XIDC}', "");
Expect(1, 917999, '\P{^-_Is_XIDC}', "");
Expect(0, 918000, '\p{-_Is_XIDC}', "");
Expect(1, 918000, '\p{^-_Is_XIDC}', "");
Expect(1, 918000, '\P{-_Is_XIDC}', "");
Expect(0, 918000, '\P{^-_Is_XIDC}', "");
Error('\p{/a/_	Supplementary_PRIVATE_use_AREA_B}');
Error('\P{/a/_	Supplementary_PRIVATE_use_AREA_B}');
Expect(1, 1048576, '\p{supplementaryprivateuseareab}', "");
Expect(0, 1048576, '\p{^supplementaryprivateuseareab}', "");
Expect(0, 1048576, '\P{supplementaryprivateuseareab}', "");
Expect(1, 1048576, '\P{^supplementaryprivateuseareab}', "");
Expect(0, 1, '\p{supplementaryprivateuseareab}', "");
Expect(1, 1, '\p{^supplementaryprivateuseareab}', "");
Expect(1, 1, '\P{supplementaryprivateuseareab}', "");
Expect(0, 1, '\P{^supplementaryprivateuseareab}', "");
Expect(1, 1048576, '\p{ -supplementary_private_Use_Area_B}', "");
Expect(0, 1048576, '\p{^ -supplementary_private_Use_Area_B}', "");
Expect(0, 1048576, '\P{ -supplementary_private_Use_Area_B}', "");
Expect(1, 1048576, '\P{^ -supplementary_private_Use_Area_B}', "");
Expect(0, 1, '\p{ -supplementary_private_Use_Area_B}', "");
Expect(1, 1, '\p{^ -supplementary_private_Use_Area_B}', "");
Expect(1, 1, '\P{ -supplementary_private_Use_Area_B}', "");
Expect(0, 1, '\P{^ -supplementary_private_Use_Area_B}', "");
Error('\p{_is_SUPPLEMENTARY_PRIVATE_use_Area_b/a/}');
Error('\P{_is_SUPPLEMENTARY_PRIVATE_use_Area_b/a/}');
Expect(1, 1048576, '\p{issupplementaryprivateuseareab}', "");
Expect(0, 1048576, '\p{^issupplementaryprivateuseareab}', "");
Expect(0, 1048576, '\P{issupplementaryprivateuseareab}', "");
Expect(1, 1048576, '\P{^issupplementaryprivateuseareab}', "");
Expect(0, 1, '\p{issupplementaryprivateuseareab}', "");
Expect(1, 1, '\p{^issupplementaryprivateuseareab}', "");
Expect(1, 1, '\P{issupplementaryprivateuseareab}', "");
Expect(0, 1, '\P{^issupplementaryprivateuseareab}', "");
Expect(1, 1048576, '\p{	_is_SUPPLEMENTARY_PRIVATE_use_Area_b}', "");
Expect(0, 1048576, '\p{^	_is_SUPPLEMENTARY_PRIVATE_use_Area_b}', "");
Expect(0, 1048576, '\P{	_is_SUPPLEMENTARY_PRIVATE_use_Area_b}', "");
Expect(1, 1048576, '\P{^	_is_SUPPLEMENTARY_PRIVATE_use_Area_b}', "");
Expect(0, 1, '\p{	_is_SUPPLEMENTARY_PRIVATE_use_Area_b}', "");
Expect(1, 1, '\p{^	_is_SUPPLEMENTARY_PRIVATE_use_Area_b}', "");
Expect(1, 1, '\P{	_is_SUPPLEMENTARY_PRIVATE_use_Area_b}', "");
Expect(0, 1, '\P{^	_is_SUPPLEMENTARY_PRIVATE_use_Area_b}', "");
Error('\p{	-IN_Supplementary_Private_USE_Area_B:=}');
Error('\P{	-IN_Supplementary_Private_USE_Area_B:=}');
Expect(1, 1048576, '\p{insupplementaryprivateuseareab}', "");
Expect(0, 1048576, '\p{^insupplementaryprivateuseareab}', "");
Expect(0, 1048576, '\P{insupplementaryprivateuseareab}', "");
Expect(1, 1048576, '\P{^insupplementaryprivateuseareab}', "");
Expect(0, 1, '\p{insupplementaryprivateuseareab}', "");
Expect(1, 1, '\p{^insupplementaryprivateuseareab}', "");
Expect(1, 1, '\P{insupplementaryprivateuseareab}', "");
Expect(0, 1, '\P{^insupplementaryprivateuseareab}', "");
Expect(1, 1048576, '\p{-_in_SUPPLEMENTARY_Private_Use_Area_B}', "");
Expect(0, 1048576, '\p{^-_in_SUPPLEMENTARY_Private_Use_Area_B}', "");
Expect(0, 1048576, '\P{-_in_SUPPLEMENTARY_Private_Use_Area_B}', "");
Expect(1, 1048576, '\P{^-_in_SUPPLEMENTARY_Private_Use_Area_B}', "");
Expect(0, 1, '\p{-_in_SUPPLEMENTARY_Private_Use_Area_B}', "");
Expect(1, 1, '\p{^-_in_SUPPLEMENTARY_Private_Use_Area_B}', "");
Expect(1, 1, '\P{-_in_SUPPLEMENTARY_Private_Use_Area_B}', "");
Expect(0, 1, '\P{^-_in_SUPPLEMENTARY_Private_Use_Area_B}', "");
Error('\p{-:=SUP_pua_B}');
Error('\P{-:=SUP_pua_B}');
Expect(1, 1048576, '\p{suppuab}', "");
Expect(0, 1048576, '\p{^suppuab}', "");
Expect(0, 1048576, '\P{suppuab}', "");
Expect(1, 1048576, '\P{^suppuab}', "");
Expect(0, 1, '\p{suppuab}', "");
Expect(1, 1, '\p{^suppuab}', "");
Expect(1, 1, '\P{suppuab}', "");
Expect(0, 1, '\P{^suppuab}', "");
Expect(1, 1048576, '\p{		Sup_pua_B}', "");
Expect(0, 1048576, '\p{^		Sup_pua_B}', "");
Expect(0, 1048576, '\P{		Sup_pua_B}', "");
Expect(1, 1048576, '\P{^		Sup_pua_B}', "");
Expect(0, 1, '\p{		Sup_pua_B}', "");
Expect(1, 1, '\p{^		Sup_pua_B}', "");
Expect(1, 1, '\P{		Sup_pua_B}', "");
Expect(0, 1, '\P{^		Sup_pua_B}', "");
Error('\p{-/a/Is_sup_PUA_B}');
Error('\P{-/a/Is_sup_PUA_B}');
Expect(1, 1048576, '\p{issuppuab}', "");
Expect(0, 1048576, '\p{^issuppuab}', "");
Expect(0, 1048576, '\P{issuppuab}', "");
Expect(1, 1048576, '\P{^issuppuab}', "");
Expect(0, 1, '\p{issuppuab}', "");
Expect(1, 1, '\p{^issuppuab}', "");
Expect(1, 1, '\P{issuppuab}', "");
Expect(0, 1, '\P{^issuppuab}', "");
Expect(1, 1048576, '\p{	is_Sup_PUA_B}', "");
Expect(0, 1048576, '\p{^	is_Sup_PUA_B}', "");
Expect(0, 1048576, '\P{	is_Sup_PUA_B}', "");
Expect(1, 1048576, '\P{^	is_Sup_PUA_B}', "");
Expect(0, 1, '\p{	is_Sup_PUA_B}', "");
Expect(1, 1, '\p{^	is_Sup_PUA_B}', "");
Expect(1, 1, '\P{	is_Sup_PUA_B}', "");
Expect(0, 1, '\P{^	is_Sup_PUA_B}', "");
Error('\p{/a/-In_sup_PUA_b}');
Error('\P{/a/-In_sup_PUA_b}');
Expect(1, 1048576, '\p{insuppuab}', "");
Expect(0, 1048576, '\p{^insuppuab}', "");
Expect(0, 1048576, '\P{insuppuab}', "");
Expect(1, 1048576, '\P{^insuppuab}', "");
Expect(0, 1, '\p{insuppuab}', "");
Expect(1, 1, '\p{^insuppuab}', "");
Expect(1, 1, '\P{insuppuab}', "");
Expect(0, 1, '\P{^insuppuab}', "");
Expect(1, 1048576, '\p{ 	in_sup_pua_B}', "");
Expect(0, 1048576, '\p{^ 	in_sup_pua_B}', "");
Expect(0, 1048576, '\P{ 	in_sup_pua_B}', "");
Expect(1, 1048576, '\P{^ 	in_sup_pua_B}', "");
Expect(0, 1, '\p{ 	in_sup_pua_B}', "");
Expect(1, 1, '\p{^ 	in_sup_pua_B}', "");
Expect(1, 1, '\P{ 	in_sup_pua_B}', "");
Expect(0, 1, '\P{^ 	in_sup_pua_B}', "");
Error('\p{ :=ARABIC_Presentation_Forms_B}');
Error('\P{ :=ARABIC_Presentation_Forms_B}');
Expect(1, 65279, '\p{arabicpresentationformsb}', "");
Expect(0, 65279, '\p{^arabicpresentationformsb}', "");
Expect(0, 65279, '\P{arabicpresentationformsb}', "");
Expect(1, 65279, '\P{^arabicpresentationformsb}', "");
Expect(0, 65280, '\p{arabicpresentationformsb}', "");
Expect(1, 65280, '\p{^arabicpresentationformsb}', "");
Expect(1, 65280, '\P{arabicpresentationformsb}', "");
Expect(0, 65280, '\P{^arabicpresentationformsb}', "");
Expect(1, 65279, '\p{  Arabic_Presentation_FORMS_B}', "");
Expect(0, 65279, '\p{^  Arabic_Presentation_FORMS_B}', "");
Expect(0, 65279, '\P{  Arabic_Presentation_FORMS_B}', "");
Expect(1, 65279, '\P{^  Arabic_Presentation_FORMS_B}', "");
Expect(0, 65280, '\p{  Arabic_Presentation_FORMS_B}', "");
Expect(1, 65280, '\p{^  Arabic_Presentation_FORMS_B}', "");
Expect(1, 65280, '\P{  Arabic_Presentation_FORMS_B}', "");
Expect(0, 65280, '\P{^  Arabic_Presentation_FORMS_B}', "");
Error('\p{-IS_arabic_PRESENTATION_FORMS_B/a/}');
Error('\P{-IS_arabic_PRESENTATION_FORMS_B/a/}');
Expect(1, 65279, '\p{isarabicpresentationformsb}', "");
Expect(0, 65279, '\p{^isarabicpresentationformsb}', "");
Expect(0, 65279, '\P{isarabicpresentationformsb}', "");
Expect(1, 65279, '\P{^isarabicpresentationformsb}', "");
Expect(0, 65280, '\p{isarabicpresentationformsb}', "");
Expect(1, 65280, '\p{^isarabicpresentationformsb}', "");
Expect(1, 65280, '\P{isarabicpresentationformsb}', "");
Expect(0, 65280, '\P{^isarabicpresentationformsb}', "");
Expect(1, 65279, '\p{-	is_Arabic_Presentation_forms_B}', "");
Expect(0, 65279, '\p{^-	is_Arabic_Presentation_forms_B}', "");
Expect(0, 65279, '\P{-	is_Arabic_Presentation_forms_B}', "");
Expect(1, 65279, '\P{^-	is_Arabic_Presentation_forms_B}', "");
Expect(0, 65280, '\p{-	is_Arabic_Presentation_forms_B}', "");
Expect(1, 65280, '\p{^-	is_Arabic_Presentation_forms_B}', "");
Expect(1, 65280, '\P{-	is_Arabic_Presentation_forms_B}', "");
Expect(0, 65280, '\P{^-	is_Arabic_Presentation_forms_B}', "");
Error('\p{:=	_In_Arabic_Presentation_FORMS_B}');
Error('\P{:=	_In_Arabic_Presentation_FORMS_B}');
Expect(1, 65279, '\p{inarabicpresentationformsb}', "");
Expect(0, 65279, '\p{^inarabicpresentationformsb}', "");
Expect(0, 65279, '\P{inarabicpresentationformsb}', "");
Expect(1, 65279, '\P{^inarabicpresentationformsb}', "");
Expect(0, 65280, '\p{inarabicpresentationformsb}', "");
Expect(1, 65280, '\p{^inarabicpresentationformsb}', "");
Expect(1, 65280, '\P{inarabicpresentationformsb}', "");
Expect(0, 65280, '\P{^inarabicpresentationformsb}', "");
Expect(1, 65279, '\p{  IN_ARABIC_Presentation_Forms_B}', "");
Expect(0, 65279, '\p{^  IN_ARABIC_Presentation_Forms_B}', "");
Expect(0, 65279, '\P{  IN_ARABIC_Presentation_Forms_B}', "");
Expect(1, 65279, '\P{^  IN_ARABIC_Presentation_Forms_B}', "");
Expect(0, 65280, '\p{  IN_ARABIC_Presentation_Forms_B}', "");
Expect(1, 65280, '\p{^  IN_ARABIC_Presentation_Forms_B}', "");
Expect(1, 65280, '\P{  IN_ARABIC_Presentation_Forms_B}', "");
Expect(0, 65280, '\P{^  IN_ARABIC_Presentation_Forms_B}', "");
Error('\p{/a/	arabic_PF_B}');
Error('\P{/a/	arabic_PF_B}');
Expect(1, 65279, '\p{arabicpfb}', "");
Expect(0, 65279, '\p{^arabicpfb}', "");
Expect(0, 65279, '\P{arabicpfb}', "");
Expect(1, 65279, '\P{^arabicpfb}', "");
Expect(0, 65280, '\p{arabicpfb}', "");
Expect(1, 65280, '\p{^arabicpfb}', "");
Expect(1, 65280, '\P{arabicpfb}', "");
Expect(0, 65280, '\P{^arabicpfb}', "");
Expect(1, 65279, '\p{ Arabic_PF_b}', "");
Expect(0, 65279, '\p{^ Arabic_PF_b}', "");
Expect(0, 65279, '\P{ Arabic_PF_b}', "");
Expect(1, 65279, '\P{^ Arabic_PF_b}', "");
Expect(0, 65280, '\p{ Arabic_PF_b}', "");
Expect(1, 65280, '\p{^ Arabic_PF_b}', "");
Expect(1, 65280, '\P{ Arabic_PF_b}', "");
Expect(0, 65280, '\P{^ Arabic_PF_b}', "");
Error('\p{-IS_arabic_pf_B:=}');
Error('\P{-IS_arabic_pf_B:=}');
Expect(1, 65279, '\p{isarabicpfb}', "");
Expect(0, 65279, '\p{^isarabicpfb}', "");
Expect(0, 65279, '\P{isarabicpfb}', "");
Expect(1, 65279, '\P{^isarabicpfb}', "");
Expect(0, 65280, '\p{isarabicpfb}', "");
Expect(1, 65280, '\p{^isarabicpfb}', "");
Expect(1, 65280, '\P{isarabicpfb}', "");
Expect(0, 65280, '\P{^isarabicpfb}', "");
Expect(1, 65279, '\p{--is_ARABIC_PF_B}', "");
Expect(0, 65279, '\p{^--is_ARABIC_PF_B}', "");
Expect(0, 65279, '\P{--is_ARABIC_PF_B}', "");
Expect(1, 65279, '\P{^--is_ARABIC_PF_B}', "");
Expect(0, 65280, '\p{--is_ARABIC_PF_B}', "");
Expect(1, 65280, '\p{^--is_ARABIC_PF_B}', "");
Expect(1, 65280, '\P{--is_ARABIC_PF_B}', "");
Expect(0, 65280, '\P{^--is_ARABIC_PF_B}', "");
Error('\p{:=IN_arabic_PF_b}');
Error('\P{:=IN_arabic_PF_b}');
Expect(1, 65279, '\p{inarabicpfb}', "");
Expect(0, 65279, '\p{^inarabicpfb}', "");
Expect(0, 65279, '\P{inarabicpfb}', "");
Expect(1, 65279, '\P{^inarabicpfb}', "");
Expect(0, 65280, '\p{inarabicpfb}', "");
Expect(1, 65280, '\p{^inarabicpfb}', "");
Expect(1, 65280, '\P{inarabicpfb}', "");
Expect(0, 65280, '\P{^inarabicpfb}', "");
Expect(1, 65279, '\p{ 	In_arabic_PF_B}', "");
Expect(0, 65279, '\p{^ 	In_arabic_PF_B}', "");
Expect(0, 65279, '\P{ 	In_arabic_PF_B}', "");
Expect(1, 65279, '\P{^ 	In_arabic_PF_B}', "");
Expect(0, 65280, '\p{ 	In_arabic_PF_B}', "");
Expect(1, 65280, '\p{^ 	In_arabic_PF_B}', "");
Expect(1, 65280, '\P{ 	In_arabic_PF_B}', "");
Expect(0, 65280, '\P{^ 	In_arabic_PF_B}', "");
Error('\p{_-miscellaneous_Symbols_And_Arrows:=}');
Error('\P{_-miscellaneous_Symbols_And_Arrows:=}');
Expect(1, 11263, '\p{miscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\p{^miscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\P{miscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\P{^miscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\p{miscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\p{^miscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\P{miscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\P{^miscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\p{ Miscellaneous_Symbols_AND_Arrows}', "");
Expect(0, 11263, '\p{^ Miscellaneous_Symbols_AND_Arrows}', "");
Expect(0, 11263, '\P{ Miscellaneous_Symbols_AND_Arrows}', "");
Expect(1, 11263, '\P{^ Miscellaneous_Symbols_AND_Arrows}', "");
Expect(0, 11264, '\p{ Miscellaneous_Symbols_AND_Arrows}', "");
Expect(1, 11264, '\p{^ Miscellaneous_Symbols_AND_Arrows}', "");
Expect(1, 11264, '\P{ Miscellaneous_Symbols_AND_Arrows}', "");
Expect(0, 11264, '\P{^ Miscellaneous_Symbols_AND_Arrows}', "");
Error('\p{:=-_Is_MISCELLANEOUS_Symbols_And_arrows}');
Error('\P{:=-_Is_MISCELLANEOUS_Symbols_And_arrows}');
Expect(1, 11263, '\p{ismiscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\p{^ismiscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\P{ismiscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\P{^ismiscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\p{ismiscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\p{^ismiscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\P{ismiscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\P{^ismiscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\p{_ Is_MISCELLANEOUS_Symbols_And_Arrows}', "");
Expect(0, 11263, '\p{^_ Is_MISCELLANEOUS_Symbols_And_Arrows}', "");
Expect(0, 11263, '\P{_ Is_MISCELLANEOUS_Symbols_And_Arrows}', "");
Expect(1, 11263, '\P{^_ Is_MISCELLANEOUS_Symbols_And_Arrows}', "");
Expect(0, 11264, '\p{_ Is_MISCELLANEOUS_Symbols_And_Arrows}', "");
Expect(1, 11264, '\p{^_ Is_MISCELLANEOUS_Symbols_And_Arrows}', "");
Expect(1, 11264, '\P{_ Is_MISCELLANEOUS_Symbols_And_Arrows}', "");
Expect(0, 11264, '\P{^_ Is_MISCELLANEOUS_Symbols_And_Arrows}', "");
Error('\p{:=__In_MISCELLANEOUS_Symbols_and_Arrows}');
Error('\P{:=__In_MISCELLANEOUS_Symbols_and_Arrows}');
Expect(1, 11263, '\p{inmiscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\p{^inmiscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\P{inmiscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\P{^inmiscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\p{inmiscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\p{^inmiscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\P{inmiscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\P{^inmiscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\p{_ In_Miscellaneous_Symbols_And_arrows}', "");
Expect(0, 11263, '\p{^_ In_Miscellaneous_Symbols_And_arrows}', "");
Expect(0, 11263, '\P{_ In_Miscellaneous_Symbols_And_arrows}', "");
Expect(1, 11263, '\P{^_ In_Miscellaneous_Symbols_And_arrows}', "");
Expect(0, 11264, '\p{_ In_Miscellaneous_Symbols_And_arrows}', "");
Expect(1, 11264, '\p{^_ In_Miscellaneous_Symbols_And_arrows}', "");
Expect(1, 11264, '\P{_ In_Miscellaneous_Symbols_And_arrows}', "");
Expect(0, 11264, '\P{^_ In_Miscellaneous_Symbols_And_arrows}', "");
Error('\p{:= 	Misc_arrows}');
Error('\P{:= 	Misc_arrows}');
Expect(1, 11263, '\p{miscarrows}', "");
Expect(0, 11263, '\p{^miscarrows}', "");
Expect(0, 11263, '\P{miscarrows}', "");
Expect(1, 11263, '\P{^miscarrows}', "");
Expect(0, 11264, '\p{miscarrows}', "");
Expect(1, 11264, '\p{^miscarrows}', "");
Expect(1, 11264, '\P{miscarrows}', "");
Expect(0, 11264, '\P{^miscarrows}', "");
Expect(1, 11263, '\p{--Misc_Arrows}', "");
Expect(0, 11263, '\p{^--Misc_Arrows}', "");
Expect(0, 11263, '\P{--Misc_Arrows}', "");
Expect(1, 11263, '\P{^--Misc_Arrows}', "");
Expect(0, 11264, '\p{--Misc_Arrows}', "");
Expect(1, 11264, '\p{^--Misc_Arrows}', "");
Expect(1, 11264, '\P{--Misc_Arrows}', "");
Expect(0, 11264, '\P{^--Misc_Arrows}', "");
Error('\p{_	Is_MISC_Arrows/a/}');
Error('\P{_	Is_MISC_Arrows/a/}');
Expect(1, 11263, '\p{ismiscarrows}', "");
Expect(0, 11263, '\p{^ismiscarrows}', "");
Expect(0, 11263, '\P{ismiscarrows}', "");
Expect(1, 11263, '\P{^ismiscarrows}', "");
Expect(0, 11264, '\p{ismiscarrows}', "");
Expect(1, 11264, '\p{^ismiscarrows}', "");
Expect(1, 11264, '\P{ismiscarrows}', "");
Expect(0, 11264, '\P{^ismiscarrows}', "");
Expect(1, 11263, '\p{-IS_Misc_ARROWS}', "");
Expect(0, 11263, '\p{^-IS_Misc_ARROWS}', "");
Expect(0, 11263, '\P{-IS_Misc_ARROWS}', "");
Expect(1, 11263, '\P{^-IS_Misc_ARROWS}', "");
Expect(0, 11264, '\p{-IS_Misc_ARROWS}', "");
Expect(1, 11264, '\p{^-IS_Misc_ARROWS}', "");
Expect(1, 11264, '\P{-IS_Misc_ARROWS}', "");
Expect(0, 11264, '\P{^-IS_Misc_ARROWS}', "");
Error('\p{/a/-in_Misc_ARROWS}');
Error('\P{/a/-in_Misc_ARROWS}');
Expect(1, 11263, '\p{inmiscarrows}', "");
Expect(0, 11263, '\p{^inmiscarrows}', "");
Expect(0, 11263, '\P{inmiscarrows}', "");
Expect(1, 11263, '\P{^inmiscarrows}', "");
Expect(0, 11264, '\p{inmiscarrows}', "");
Expect(1, 11264, '\p{^inmiscarrows}', "");
Expect(1, 11264, '\P{inmiscarrows}', "");
Expect(0, 11264, '\P{^inmiscarrows}', "");
Expect(1, 11263, '\p{-	IN_Misc_arrows}', "");
Expect(0, 11263, '\p{^-	IN_Misc_arrows}', "");
Expect(0, 11263, '\P{-	IN_Misc_arrows}', "");
Expect(1, 11263, '\P{^-	IN_Misc_arrows}', "");
Expect(0, 11264, '\p{-	IN_Misc_arrows}', "");
Expect(1, 11264, '\p{^-	IN_Misc_arrows}', "");
Expect(1, 11264, '\P{-	IN_Misc_arrows}', "");
Expect(0, 11264, '\P{^-	IN_Misc_arrows}', "");
Error('\p{		Unified_Canadian_Aboriginal_SYLLABICS/a/}');
Error('\P{		Unified_Canadian_Aboriginal_SYLLABICS/a/}');
Expect(1, 5759, '\p{unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\p{^unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\P{unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\P{^unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\p{unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\p{^unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\P{unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\P{^unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\p{	_unified_CANADIAN_Aboriginal_Syllabics}', "");
Expect(0, 5759, '\p{^	_unified_CANADIAN_Aboriginal_Syllabics}', "");
Expect(0, 5759, '\P{	_unified_CANADIAN_Aboriginal_Syllabics}', "");
Expect(1, 5759, '\P{^	_unified_CANADIAN_Aboriginal_Syllabics}', "");
Expect(0, 5760, '\p{	_unified_CANADIAN_Aboriginal_Syllabics}', "");
Expect(1, 5760, '\p{^	_unified_CANADIAN_Aboriginal_Syllabics}', "");
Expect(1, 5760, '\P{	_unified_CANADIAN_Aboriginal_Syllabics}', "");
Expect(0, 5760, '\P{^	_unified_CANADIAN_Aboriginal_Syllabics}', "");
Error('\p{_/a/IS_Unified_CANADIAN_Aboriginal_Syllabics}');
Error('\P{_/a/IS_Unified_CANADIAN_Aboriginal_Syllabics}');
Expect(1, 5759, '\p{isunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\p{^isunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\P{isunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\P{^isunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\p{isunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\p{^isunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\P{isunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\P{^isunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\p{- IS_Unified_Canadian_Aboriginal_SYLLABICS}', "");
Expect(0, 5759, '\p{^- IS_Unified_Canadian_Aboriginal_SYLLABICS}', "");
Expect(0, 5759, '\P{- IS_Unified_Canadian_Aboriginal_SYLLABICS}', "");
Expect(1, 5759, '\P{^- IS_Unified_Canadian_Aboriginal_SYLLABICS}', "");
Expect(0, 5760, '\p{- IS_Unified_Canadian_Aboriginal_SYLLABICS}', "");
Expect(1, 5760, '\p{^- IS_Unified_Canadian_Aboriginal_SYLLABICS}', "");
Expect(1, 5760, '\P{- IS_Unified_Canadian_Aboriginal_SYLLABICS}', "");
Expect(0, 5760, '\P{^- IS_Unified_Canadian_Aboriginal_SYLLABICS}', "");
Error('\p{:=- in_Unified_canadian_Aboriginal_Syllabics}');
Error('\P{:=- in_Unified_canadian_Aboriginal_Syllabics}');
Expect(1, 5759, '\p{inunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\p{^inunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\P{inunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\P{^inunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\p{inunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\p{^inunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\P{inunifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\P{^inunifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\p{in_Unified_canadian_Aboriginal_Syllabics}', "");
Expect(0, 5759, '\p{^in_Unified_canadian_Aboriginal_Syllabics}', "");
Expect(0, 5759, '\P{in_Unified_canadian_Aboriginal_Syllabics}', "");
Expect(1, 5759, '\P{^in_Unified_canadian_Aboriginal_Syllabics}', "");
Expect(0, 5760, '\p{in_Unified_canadian_Aboriginal_Syllabics}', "");
Expect(1, 5760, '\p{^in_Unified_canadian_Aboriginal_Syllabics}', "");
Expect(1, 5760, '\P{in_Unified_canadian_Aboriginal_Syllabics}', "");
Expect(0, 5760, '\P{^in_Unified_canadian_Aboriginal_Syllabics}', "");
Error('\p{ UCAS/a/}');
Error('\P{ UCAS/a/}');
Expect(1, 5759, '\p{ucas}', "");
Expect(0, 5759, '\p{^ucas}', "");
Expect(0, 5759, '\P{ucas}', "");
Expect(1, 5759, '\P{^ucas}', "");
Expect(0, 5760, '\p{ucas}', "");
Expect(1, 5760, '\p{^ucas}', "");
Expect(1, 5760, '\P{ucas}', "");
Expect(0, 5760, '\P{^ucas}', "");
Expect(1, 5759, '\p{ UCAS}', "");
Expect(0, 5759, '\p{^ UCAS}', "");
Expect(0, 5759, '\P{ UCAS}', "");
Expect(1, 5759, '\P{^ UCAS}', "");
Expect(0, 5760, '\p{ UCAS}', "");
Expect(1, 5760, '\p{^ UCAS}', "");
Expect(1, 5760, '\P{ UCAS}', "");
Expect(0, 5760, '\P{^ UCAS}', "");
Error('\p{:=Is_UCAS}');
Error('\P{:=Is_UCAS}');
Expect(1, 5759, '\p{isucas}', "");
Expect(0, 5759, '\p{^isucas}', "");
Expect(0, 5759, '\P{isucas}', "");
Expect(1, 5759, '\P{^isucas}', "");
Expect(0, 5760, '\p{isucas}', "");
Expect(1, 5760, '\p{^isucas}', "");
Expect(1, 5760, '\P{isucas}', "");
Expect(0, 5760, '\P{^isucas}', "");
Expect(1, 5759, '\p{	Is_UCAS}', "");
Expect(0, 5759, '\p{^	Is_UCAS}', "");
Expect(0, 5759, '\P{	Is_UCAS}', "");
Expect(1, 5759, '\P{^	Is_UCAS}', "");
Expect(0, 5760, '\p{	Is_UCAS}', "");
Expect(1, 5760, '\p{^	Is_UCAS}', "");
Expect(1, 5760, '\P{	Is_UCAS}', "");
Expect(0, 5760, '\P{^	Is_UCAS}', "");
Error('\p{-:=In_UCAS}');
Error('\P{-:=In_UCAS}');
Expect(1, 5759, '\p{inucas}', "");
Expect(0, 5759, '\p{^inucas}', "");
Expect(0, 5759, '\P{inucas}', "");
Expect(1, 5759, '\P{^inucas}', "");
Expect(0, 5760, '\p{inucas}', "");
Expect(1, 5760, '\p{^inucas}', "");
Expect(1, 5760, '\P{inucas}', "");
Expect(0, 5760, '\P{^inucas}', "");
Expect(1, 5759, '\p{-_in_UCAS}', "");
Expect(0, 5759, '\p{^-_in_UCAS}', "");
Expect(0, 5759, '\P{-_in_UCAS}', "");
Expect(1, 5759, '\P{^-_in_UCAS}', "");
Expect(0, 5760, '\p{-_in_UCAS}', "");
Expect(1, 5760, '\p{^-_in_UCAS}', "");
Expect(1, 5760, '\P{-_in_UCAS}', "");
Expect(0, 5760, '\P{^-_in_UCAS}', "");
Error('\p{/a/-	Canadian_Syllabics}');
Error('\P{/a/-	Canadian_Syllabics}');
Expect(1, 5759, '\p{canadiansyllabics}', "");
Expect(0, 5759, '\p{^canadiansyllabics}', "");
Expect(0, 5759, '\P{canadiansyllabics}', "");
Expect(1, 5759, '\P{^canadiansyllabics}', "");
Expect(0, 5760, '\p{canadiansyllabics}', "");
Expect(1, 5760, '\p{^canadiansyllabics}', "");
Expect(1, 5760, '\P{canadiansyllabics}', "");
Expect(0, 5760, '\P{^canadiansyllabics}', "");
Expect(1, 5759, '\p{		Canadian_SYLLABICS}', "");
Expect(0, 5759, '\p{^		Canadian_SYLLABICS}', "");
Expect(0, 5759, '\P{		Canadian_SYLLABICS}', "");
Expect(1, 5759, '\P{^		Canadian_SYLLABICS}', "");
Expect(0, 5760, '\p{		Canadian_SYLLABICS}', "");
Expect(1, 5760, '\p{^		Canadian_SYLLABICS}', "");
Expect(1, 5760, '\P{		Canadian_SYLLABICS}', "");
Expect(0, 5760, '\P{^		Canadian_SYLLABICS}', "");
Error('\p{-:=Is_Canadian_syllabics}');
Error('\P{-:=Is_Canadian_syllabics}');
Expect(1, 5759, '\p{iscanadiansyllabics}', "");
Expect(0, 5759, '\p{^iscanadiansyllabics}', "");
Expect(0, 5759, '\P{iscanadiansyllabics}', "");
Expect(1, 5759, '\P{^iscanadiansyllabics}', "");
Expect(0, 5760, '\p{iscanadiansyllabics}', "");
Expect(1, 5760, '\p{^iscanadiansyllabics}', "");
Expect(1, 5760, '\P{iscanadiansyllabics}', "");
Expect(0, 5760, '\P{^iscanadiansyllabics}', "");
Expect(1, 5759, '\p{ 	Is_Canadian_Syllabics}', "");
Expect(0, 5759, '\p{^ 	Is_Canadian_Syllabics}', "");
Expect(0, 5759, '\P{ 	Is_Canadian_Syllabics}', "");
Expect(1, 5759, '\P{^ 	Is_Canadian_Syllabics}', "");
Expect(0, 5760, '\p{ 	Is_Canadian_Syllabics}', "");
Expect(1, 5760, '\p{^ 	Is_Canadian_Syllabics}', "");
Expect(1, 5760, '\P{ 	Is_Canadian_Syllabics}', "");
Expect(0, 5760, '\P{^ 	Is_Canadian_Syllabics}', "");
Error('\p{:=- In_Canadian_SYLLABICS}');
Error('\P{:=- In_Canadian_SYLLABICS}');
Expect(1, 5759, '\p{incanadiansyllabics}', "");
Expect(0, 5759, '\p{^incanadiansyllabics}', "");
Expect(0, 5759, '\P{incanadiansyllabics}', "");
Expect(1, 5759, '\P{^incanadiansyllabics}', "");
Expect(0, 5760, '\p{incanadiansyllabics}', "");
Expect(1, 5760, '\p{^incanadiansyllabics}', "");
Expect(1, 5760, '\P{incanadiansyllabics}', "");
Expect(0, 5760, '\P{^incanadiansyllabics}', "");
Expect(1, 5759, '\p{-	In_CANADIAN_Syllabics}', "");
Expect(0, 5759, '\p{^-	In_CANADIAN_Syllabics}', "");
Expect(0, 5759, '\P{-	In_CANADIAN_Syllabics}', "");
Expect(1, 5759, '\P{^-	In_CANADIAN_Syllabics}', "");
Expect(0, 5760, '\p{-	In_CANADIAN_Syllabics}', "");
Expect(1, 5760, '\p{^-	In_CANADIAN_Syllabics}', "");
Expect(1, 5760, '\P{-	In_CANADIAN_Syllabics}', "");
Expect(0, 5760, '\P{^-	In_CANADIAN_Syllabics}', "");
Error('\p{	_Egyptian_hieroglyphs:=}');
Error('\P{	_Egyptian_hieroglyphs:=}');
Expect(1, 78894, '\p{egyptianhieroglyphs}', "");
Expect(0, 78894, '\p{^egyptianhieroglyphs}', "");
Expect(0, 78894, '\P{egyptianhieroglyphs}', "");
Expect(1, 78894, '\P{^egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{^egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{^egyptianhieroglyphs}', "");
Expect(1, 78894, '\p{_	Egyptian_Hieroglyphs}', "");
Expect(0, 78894, '\p{^_	Egyptian_Hieroglyphs}', "");
Expect(0, 78894, '\P{_	Egyptian_Hieroglyphs}', "");
Expect(1, 78894, '\P{^_	Egyptian_Hieroglyphs}', "");
Expect(0, 78895, '\p{_	Egyptian_Hieroglyphs}', "");
Expect(1, 78895, '\p{^_	Egyptian_Hieroglyphs}', "");
Expect(1, 78895, '\P{_	Egyptian_Hieroglyphs}', "");
Expect(0, 78895, '\P{^_	Egyptian_Hieroglyphs}', "");
Error('\p{-/a/Is_Egyptian_HIEROGLYPHS}');
Error('\P{-/a/Is_Egyptian_HIEROGLYPHS}');
Expect(1, 78894, '\p{isegyptianhieroglyphs}', "");
Expect(0, 78894, '\p{^isegyptianhieroglyphs}', "");
Expect(0, 78894, '\P{isegyptianhieroglyphs}', "");
Expect(1, 78894, '\P{^isegyptianhieroglyphs}', "");
Expect(0, 78895, '\p{isegyptianhieroglyphs}', "");
Expect(1, 78895, '\p{^isegyptianhieroglyphs}', "");
Expect(1, 78895, '\P{isegyptianhieroglyphs}', "");
Expect(0, 78895, '\P{^isegyptianhieroglyphs}', "");
Expect(1, 78894, '\p{-Is_Egyptian_hieroglyphs}', "");
Expect(0, 78894, '\p{^-Is_Egyptian_hieroglyphs}', "");
Expect(0, 78894, '\P{-Is_Egyptian_hieroglyphs}', "");
Expect(1, 78894, '\P{^-Is_Egyptian_hieroglyphs}', "");
Expect(0, 78895, '\p{-Is_Egyptian_hieroglyphs}', "");
Expect(1, 78895, '\p{^-Is_Egyptian_hieroglyphs}', "");
Expect(1, 78895, '\P{-Is_Egyptian_hieroglyphs}', "");
Expect(0, 78895, '\P{^-Is_Egyptian_hieroglyphs}', "");
Error('\p{/a/__EGYP}');
Error('\P{/a/__EGYP}');
Expect(1, 78894, '\p{egyp}', "");
Expect(0, 78894, '\p{^egyp}', "");
Expect(0, 78894, '\P{egyp}', "");
Expect(1, 78894, '\P{^egyp}', "");
Expect(0, 78895, '\p{egyp}', "");
Expect(1, 78895, '\p{^egyp}', "");
Expect(1, 78895, '\P{egyp}', "");
Expect(0, 78895, '\P{^egyp}', "");
Expect(1, 78894, '\p{-egyp}', "");
Expect(0, 78894, '\p{^-egyp}', "");
Expect(0, 78894, '\P{-egyp}', "");
Expect(1, 78894, '\P{^-egyp}', "");
Expect(0, 78895, '\p{-egyp}', "");
Expect(1, 78895, '\p{^-egyp}', "");
Expect(1, 78895, '\P{-egyp}', "");
Expect(0, 78895, '\P{^-egyp}', "");
Error('\p{- Is_egyp/a/}');
Error('\P{- Is_egyp/a/}');
Expect(1, 78894, '\p{isegyp}', "");
Expect(0, 78894, '\p{^isegyp}', "");
Expect(0, 78894, '\P{isegyp}', "");
Expect(1, 78894, '\P{^isegyp}', "");
Expect(0, 78895, '\p{isegyp}', "");
Expect(1, 78895, '\p{^isegyp}', "");
Expect(1, 78895, '\P{isegyp}', "");
Expect(0, 78895, '\P{^isegyp}', "");
Expect(1, 78894, '\p{-Is_EGYP}', "");
Expect(0, 78894, '\p{^-Is_EGYP}', "");
Expect(0, 78894, '\P{-Is_EGYP}', "");
Expect(1, 78894, '\P{^-Is_EGYP}', "");
Expect(0, 78895, '\p{-Is_EGYP}', "");
Expect(1, 78895, '\p{^-Is_EGYP}', "");
Expect(1, 78895, '\P{-Is_EGYP}', "");
Expect(0, 78895, '\P{^-Is_EGYP}', "");
Error('\p{/a/Aegean_numbers}');
Error('\P{/a/Aegean_numbers}');
Expect(1, 65855, '\p{aegeannumbers}', "");
Expect(0, 65855, '\p{^aegeannumbers}', "");
Expect(0, 65855, '\P{aegeannumbers}', "");
Expect(1, 65855, '\P{^aegeannumbers}', "");
Expect(0, 65856, '\p{aegeannumbers}', "");
Expect(1, 65856, '\p{^aegeannumbers}', "");
Expect(1, 65856, '\P{aegeannumbers}', "");
Expect(0, 65856, '\P{^aegeannumbers}', "");
Expect(1, 65855, '\p{-aegean_Numbers}', "");
Expect(0, 65855, '\p{^-aegean_Numbers}', "");
Expect(0, 65855, '\P{-aegean_Numbers}', "");
Expect(1, 65855, '\P{^-aegean_Numbers}', "");
Expect(0, 65856, '\p{-aegean_Numbers}', "");
Expect(1, 65856, '\p{^-aegean_Numbers}', "");
Expect(1, 65856, '\P{-aegean_Numbers}', "");
Expect(0, 65856, '\P{^-aegean_Numbers}', "");
Error('\p{/a/__Is_Aegean_NUMBERS}');
Error('\P{/a/__Is_Aegean_NUMBERS}');
Expect(1, 65855, '\p{isaegeannumbers}', "");
Expect(0, 65855, '\p{^isaegeannumbers}', "");
Expect(0, 65855, '\P{isaegeannumbers}', "");
Expect(1, 65855, '\P{^isaegeannumbers}', "");
Expect(0, 65856, '\p{isaegeannumbers}', "");
Expect(1, 65856, '\p{^isaegeannumbers}', "");
Expect(1, 65856, '\P{isaegeannumbers}', "");
Expect(0, 65856, '\P{^isaegeannumbers}', "");
Expect(1, 65855, '\p{ -Is_Aegean_NUMBERS}', "");
Expect(0, 65855, '\p{^ -Is_Aegean_NUMBERS}', "");
Expect(0, 65855, '\P{ -Is_Aegean_NUMBERS}', "");
Expect(1, 65855, '\P{^ -Is_Aegean_NUMBERS}', "");
Expect(0, 65856, '\p{ -Is_Aegean_NUMBERS}', "");
Expect(1, 65856, '\p{^ -Is_Aegean_NUMBERS}', "");
Expect(1, 65856, '\P{ -Is_Aegean_NUMBERS}', "");
Expect(0, 65856, '\P{^ -Is_Aegean_NUMBERS}', "");
Error('\p{:=-_in_Aegean_numbers}');
Error('\P{:=-_in_Aegean_numbers}');
Expect(1, 65855, '\p{inaegeannumbers}', "");
Expect(0, 65855, '\p{^inaegeannumbers}', "");
Expect(0, 65855, '\P{inaegeannumbers}', "");
Expect(1, 65855, '\P{^inaegeannumbers}', "");
Expect(0, 65856, '\p{inaegeannumbers}', "");
Expect(1, 65856, '\p{^inaegeannumbers}', "");
Expect(1, 65856, '\P{inaegeannumbers}', "");
Expect(0, 65856, '\P{^inaegeannumbers}', "");
Expect(1, 65855, '\p{  In_Aegean_Numbers}', "");
Expect(0, 65855, '\p{^  In_Aegean_Numbers}', "");
Expect(0, 65855, '\P{  In_Aegean_Numbers}', "");
Expect(1, 65855, '\P{^  In_Aegean_Numbers}', "");
Expect(0, 65856, '\p{  In_Aegean_Numbers}', "");
Expect(1, 65856, '\p{^  In_Aegean_Numbers}', "");
Expect(1, 65856, '\P{  In_Aegean_Numbers}', "");
Expect(0, 65856, '\P{^  In_Aegean_Numbers}', "");
Error('\p{--Alphabetic_presentation_FORMS:=}');
Error('\P{--Alphabetic_presentation_FORMS:=}');
Expect(1, 64335, '\p{alphabeticpresentationforms}', "");
Expect(0, 64335, '\p{^alphabeticpresentationforms}', "");
Expect(0, 64335, '\P{alphabeticpresentationforms}', "");
Expect(1, 64335, '\P{^alphabeticpresentationforms}', "");
Expect(0, 64336, '\p{alphabeticpresentationforms}', "");
Expect(1, 64336, '\p{^alphabeticpresentationforms}', "");
Expect(1, 64336, '\P{alphabeticpresentationforms}', "");
Expect(0, 64336, '\P{^alphabeticpresentationforms}', "");
Expect(1, 64335, '\p{-alphabetic_Presentation_Forms}', "");
Expect(0, 64335, '\p{^-alphabetic_Presentation_Forms}', "");
Expect(0, 64335, '\P{-alphabetic_Presentation_Forms}', "");
Expect(1, 64335, '\P{^-alphabetic_Presentation_Forms}', "");
Expect(0, 64336, '\p{-alphabetic_Presentation_Forms}', "");
Expect(1, 64336, '\p{^-alphabetic_Presentation_Forms}', "");
Expect(1, 64336, '\P{-alphabetic_Presentation_Forms}', "");
Expect(0, 64336, '\P{^-alphabetic_Presentation_Forms}', "");
Error('\p{_	Is_Alphabetic_PRESENTATION_Forms:=}');
Error('\P{_	Is_Alphabetic_PRESENTATION_Forms:=}');
Expect(1, 64335, '\p{isalphabeticpresentationforms}', "");
Expect(0, 64335, '\p{^isalphabeticpresentationforms}', "");
Expect(0, 64335, '\P{isalphabeticpresentationforms}', "");
Expect(1, 64335, '\P{^isalphabeticpresentationforms}', "");
Expect(0, 64336, '\p{isalphabeticpresentationforms}', "");
Expect(1, 64336, '\p{^isalphabeticpresentationforms}', "");
Expect(1, 64336, '\P{isalphabeticpresentationforms}', "");
Expect(0, 64336, '\P{^isalphabeticpresentationforms}', "");
Expect(1, 64335, '\p{_Is_Alphabetic_presentation_forms}', "");
Expect(0, 64335, '\p{^_Is_Alphabetic_presentation_forms}', "");
Expect(0, 64335, '\P{_Is_Alphabetic_presentation_forms}', "");
Expect(1, 64335, '\P{^_Is_Alphabetic_presentation_forms}', "");
Expect(0, 64336, '\p{_Is_Alphabetic_presentation_forms}', "");
Expect(1, 64336, '\p{^_Is_Alphabetic_presentation_forms}', "");
Expect(1, 64336, '\P{_Is_Alphabetic_presentation_forms}', "");
Expect(0, 64336, '\P{^_Is_Alphabetic_presentation_forms}', "");
Error('\p{:=In_Alphabetic_Presentation_Forms}');
Error('\P{:=In_Alphabetic_Presentation_Forms}');
Expect(1, 64335, '\p{inalphabeticpresentationforms}', "");
Expect(0, 64335, '\p{^inalphabeticpresentationforms}', "");
Expect(0, 64335, '\P{inalphabeticpresentationforms}', "");
Expect(1, 64335, '\P{^inalphabeticpresentationforms}', "");
Expect(0, 64336, '\p{inalphabeticpresentationforms}', "");
Expect(1, 64336, '\p{^inalphabeticpresentationforms}', "");
Expect(1, 64336, '\P{inalphabeticpresentationforms}', "");
Expect(0, 64336, '\P{^inalphabeticpresentationforms}', "");
Expect(1, 64335, '\p{ In_ALPHABETIC_Presentation_FORMS}', "");
Expect(0, 64335, '\p{^ In_ALPHABETIC_Presentation_FORMS}', "");
Expect(0, 64335, '\P{ In_ALPHABETIC_Presentation_FORMS}', "");
Expect(1, 64335, '\P{^ In_ALPHABETIC_Presentation_FORMS}', "");
Expect(0, 64336, '\p{ In_ALPHABETIC_Presentation_FORMS}', "");
Expect(1, 64336, '\p{^ In_ALPHABETIC_Presentation_FORMS}', "");
Expect(1, 64336, '\P{ In_ALPHABETIC_Presentation_FORMS}', "");
Expect(0, 64336, '\P{^ In_ALPHABETIC_Presentation_FORMS}', "");
Error('\p{:=alphabetic_PF}');
Error('\P{:=alphabetic_PF}');
Expect(1, 64335, '\p{alphabeticpf}', "");
Expect(0, 64335, '\p{^alphabeticpf}', "");
Expect(0, 64335, '\P{alphabeticpf}', "");
Expect(1, 64335, '\P{^alphabeticpf}', "");
Expect(0, 64336, '\p{alphabeticpf}', "");
Expect(1, 64336, '\p{^alphabeticpf}', "");
Expect(1, 64336, '\P{alphabeticpf}', "");
Expect(0, 64336, '\P{^alphabeticpf}', "");
Expect(1, 64335, '\p{_ ALPHABETIC_PF}', "");
Expect(0, 64335, '\p{^_ ALPHABETIC_PF}', "");
Expect(0, 64335, '\P{_ ALPHABETIC_PF}', "");
Expect(1, 64335, '\P{^_ ALPHABETIC_PF}', "");
Expect(0, 64336, '\p{_ ALPHABETIC_PF}', "");
Expect(1, 64336, '\p{^_ ALPHABETIC_PF}', "");
Expect(1, 64336, '\P{_ ALPHABETIC_PF}', "");
Expect(0, 64336, '\P{^_ ALPHABETIC_PF}', "");
Error('\p{_	is_ALPHABETIC_PF/a/}');
Error('\P{_	is_ALPHABETIC_PF/a/}');
Expect(1, 64335, '\p{isalphabeticpf}', "");
Expect(0, 64335, '\p{^isalphabeticpf}', "");
Expect(0, 64335, '\P{isalphabeticpf}', "");
Expect(1, 64335, '\P{^isalphabeticpf}', "");
Expect(0, 64336, '\p{isalphabeticpf}', "");
Expect(1, 64336, '\p{^isalphabeticpf}', "");
Expect(1, 64336, '\P{isalphabeticpf}', "");
Expect(0, 64336, '\P{^isalphabeticpf}', "");
Expect(1, 64335, '\p{	_Is_Alphabetic_PF}', "");
Expect(0, 64335, '\p{^	_Is_Alphabetic_PF}', "");
Expect(0, 64335, '\P{	_Is_Alphabetic_PF}', "");
Expect(1, 64335, '\P{^	_Is_Alphabetic_PF}', "");
Expect(0, 64336, '\p{	_Is_Alphabetic_PF}', "");
Expect(1, 64336, '\p{^	_Is_Alphabetic_PF}', "");
Expect(1, 64336, '\P{	_Is_Alphabetic_PF}', "");
Expect(0, 64336, '\P{^	_Is_Alphabetic_PF}', "");
Error('\p{__in_alphabetic_PF:=}');
Error('\P{__in_alphabetic_PF:=}');
Expect(1, 64335, '\p{inalphabeticpf}', "");
Expect(0, 64335, '\p{^inalphabeticpf}', "");
Expect(0, 64335, '\P{inalphabeticpf}', "");
Expect(1, 64335, '\P{^inalphabeticpf}', "");
Expect(0, 64336, '\p{inalphabeticpf}', "");
Expect(1, 64336, '\p{^inalphabeticpf}', "");
Expect(1, 64336, '\P{inalphabeticpf}', "");
Expect(0, 64336, '\P{^inalphabeticpf}', "");
Expect(1, 64335, '\p{ -In_ALPHABETIC_PF}', "");
Expect(0, 64335, '\p{^ -In_ALPHABETIC_PF}', "");
Expect(0, 64335, '\P{ -In_ALPHABETIC_PF}', "");
Expect(1, 64335, '\P{^ -In_ALPHABETIC_PF}', "");
Expect(0, 64336, '\p{ -In_ALPHABETIC_PF}', "");
Expect(1, 64336, '\p{^ -In_ALPHABETIC_PF}', "");
Expect(1, 64336, '\P{ -In_ALPHABETIC_PF}', "");
Expect(0, 64336, '\P{^ -In_ALPHABETIC_PF}', "");
Error('\p{	/a/BIDI_control}');
Error('\P{	/a/BIDI_control}');
Expect(1, 8238, '\p{bidicontrol}', "");
Expect(0, 8238, '\p{^bidicontrol}', "");
Expect(0, 8238, '\P{bidicontrol}', "");
Expect(1, 8238, '\P{^bidicontrol}', "");
Expect(0, 8239, '\p{bidicontrol}', "");
Expect(1, 8239, '\p{^bidicontrol}', "");
Expect(1, 8239, '\P{bidicontrol}', "");
Expect(0, 8239, '\P{^bidicontrol}', "");
Expect(1, 8238, '\p{	_Bidi_CONTROL}', "");
Expect(0, 8238, '\p{^	_Bidi_CONTROL}', "");
Expect(0, 8238, '\P{	_Bidi_CONTROL}', "");
Expect(1, 8238, '\P{^	_Bidi_CONTROL}', "");
Expect(0, 8239, '\p{	_Bidi_CONTROL}', "");
Expect(1, 8239, '\p{^	_Bidi_CONTROL}', "");
Expect(1, 8239, '\P{	_Bidi_CONTROL}', "");
Expect(0, 8239, '\P{^	_Bidi_CONTROL}', "");
Error('\p{:=__Is_Bidi_Control}');
Error('\P{:=__Is_Bidi_Control}');
Expect(1, 8238, '\p{isbidicontrol}', "");
Expect(0, 8238, '\p{^isbidicontrol}', "");
Expect(0, 8238, '\P{isbidicontrol}', "");
Expect(1, 8238, '\P{^isbidicontrol}', "");
Expect(0, 8239, '\p{isbidicontrol}', "");
Expect(1, 8239, '\p{^isbidicontrol}', "");
Expect(1, 8239, '\P{isbidicontrol}', "");
Expect(0, 8239, '\P{^isbidicontrol}', "");
Expect(1, 8238, '\p{- Is_BIDI_CONTROL}', "");
Expect(0, 8238, '\p{^- Is_BIDI_CONTROL}', "");
Expect(0, 8238, '\P{- Is_BIDI_CONTROL}', "");
Expect(1, 8238, '\P{^- Is_BIDI_CONTROL}', "");
Expect(0, 8239, '\p{- Is_BIDI_CONTROL}', "");
Expect(1, 8239, '\p{^- Is_BIDI_CONTROL}', "");
Expect(1, 8239, '\P{- Is_BIDI_CONTROL}', "");
Expect(0, 8239, '\P{^- Is_BIDI_CONTROL}', "");
Error('\p{:=-Bidi_C}');
Error('\P{:=-Bidi_C}');
Expect(1, 8238, '\p{bidic}', "");
Expect(0, 8238, '\p{^bidic}', "");
Expect(0, 8238, '\P{bidic}', "");
Expect(1, 8238, '\P{^bidic}', "");
Expect(0, 8239, '\p{bidic}', "");
Expect(1, 8239, '\p{^bidic}', "");
Expect(1, 8239, '\P{bidic}', "");
Expect(0, 8239, '\P{^bidic}', "");
Expect(1, 8238, '\p{ bidi_c}', "");
Expect(0, 8238, '\p{^ bidi_c}', "");
Expect(0, 8238, '\P{ bidi_c}', "");
Expect(1, 8238, '\P{^ bidi_c}', "");
Expect(0, 8239, '\p{ bidi_c}', "");
Expect(1, 8239, '\p{^ bidi_c}', "");
Expect(1, 8239, '\P{ bidi_c}', "");
Expect(0, 8239, '\P{^ bidi_c}', "");
Error('\p{/a/  Is_BIDI_C}');
Error('\P{/a/  Is_BIDI_C}');
Expect(1, 8238, '\p{isbidic}', "");
Expect(0, 8238, '\p{^isbidic}', "");
Expect(0, 8238, '\P{isbidic}', "");
Expect(1, 8238, '\P{^isbidic}', "");
Expect(0, 8239, '\p{isbidic}', "");
Expect(1, 8239, '\p{^isbidic}', "");
Expect(1, 8239, '\P{isbidic}', "");
Expect(0, 8239, '\P{^isbidic}', "");
Expect(1, 8238, '\p{__Is_Bidi_C}', "");
Expect(0, 8238, '\p{^__Is_Bidi_C}', "");
Expect(0, 8238, '\P{__Is_Bidi_C}', "");
Expect(1, 8238, '\P{^__Is_Bidi_C}', "");
Expect(0, 8239, '\p{__Is_Bidi_C}', "");
Expect(1, 8239, '\p{^__Is_Bidi_C}', "");
Expect(1, 8239, '\P{__Is_Bidi_C}', "");
Expect(0, 8239, '\P{^__Is_Bidi_C}', "");
Error('\p{-/a/COMBINING_Diacritical_MARKS_For_SYMBOLS}');
Error('\P{-/a/COMBINING_Diacritical_MARKS_For_SYMBOLS}');
Expect(1, 8447, '\p{combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\p{^combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\P{combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\P{^combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\p{combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\p{^combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\P{combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\P{^combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\p{ -combining_Diacritical_marks_FOR_SYMBOLS}', "");
Expect(0, 8447, '\p{^ -combining_Diacritical_marks_FOR_SYMBOLS}', "");
Expect(0, 8447, '\P{ -combining_Diacritical_marks_FOR_SYMBOLS}', "");
Expect(1, 8447, '\P{^ -combining_Diacritical_marks_FOR_SYMBOLS}', "");
Expect(0, 8448, '\p{ -combining_Diacritical_marks_FOR_SYMBOLS}', "");
Expect(1, 8448, '\p{^ -combining_Diacritical_marks_FOR_SYMBOLS}', "");
Expect(1, 8448, '\P{ -combining_Diacritical_marks_FOR_SYMBOLS}', "");
Expect(0, 8448, '\P{^ -combining_Diacritical_marks_FOR_SYMBOLS}', "");
Error('\p{ :=Is_Combining_Diacritical_Marks_for_Symbols}');
Error('\P{ :=Is_Combining_Diacritical_Marks_for_Symbols}');
Expect(1, 8447, '\p{iscombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\p{^iscombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\P{iscombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\P{^iscombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\p{iscombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\p{^iscombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\P{iscombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\P{^iscombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\p{__IS_combining_Diacritical_Marks_for_symbols}', "");
Expect(0, 8447, '\p{^__IS_combining_Diacritical_Marks_for_symbols}', "");
Expect(0, 8447, '\P{__IS_combining_Diacritical_Marks_for_symbols}', "");
Expect(1, 8447, '\P{^__IS_combining_Diacritical_Marks_for_symbols}', "");
Expect(0, 8448, '\p{__IS_combining_Diacritical_Marks_for_symbols}', "");
Expect(1, 8448, '\p{^__IS_combining_Diacritical_Marks_for_symbols}', "");
Expect(1, 8448, '\P{__IS_combining_Diacritical_Marks_for_symbols}', "");
Expect(0, 8448, '\P{^__IS_combining_Diacritical_Marks_for_symbols}', "");
Error('\p{/a/-In_COMBINING_Diacritical_Marks_FOR_Symbols}');
Error('\P{/a/-In_COMBINING_Diacritical_Marks_FOR_Symbols}');
Expect(1, 8447, '\p{incombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\p{^incombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\P{incombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\P{^incombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\p{incombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\p{^incombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\P{incombiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\P{^incombiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\p{_	IN_Combining_diacritical_marks_For_symbols}', "");
Expect(0, 8447, '\p{^_	IN_Combining_diacritical_marks_For_symbols}', "");
Expect(0, 8447, '\P{_	IN_Combining_diacritical_marks_For_symbols}', "");
Expect(1, 8447, '\P{^_	IN_Combining_diacritical_marks_For_symbols}', "");
Expect(0, 8448, '\p{_	IN_Combining_diacritical_marks_For_symbols}', "");
Expect(1, 8448, '\p{^_	IN_Combining_diacritical_marks_For_symbols}', "");
Expect(1, 8448, '\P{_	IN_Combining_diacritical_marks_For_symbols}', "");
Expect(0, 8448, '\P{^_	IN_Combining_diacritical_marks_For_symbols}', "");
Error('\p{	:=Diacriticals_FOR_SYMBOLS}');
Error('\P{	:=Diacriticals_FOR_SYMBOLS}');
Expect(1, 8447, '\p{diacriticalsforsymbols}', "");
Expect(0, 8447, '\p{^diacriticalsforsymbols}', "");
Expect(0, 8447, '\P{diacriticalsforsymbols}', "");
Expect(1, 8447, '\P{^diacriticalsforsymbols}', "");
Expect(0, 8448, '\p{diacriticalsforsymbols}', "");
Expect(1, 8448, '\p{^diacriticalsforsymbols}', "");
Expect(1, 8448, '\P{diacriticalsforsymbols}', "");
Expect(0, 8448, '\P{^diacriticalsforsymbols}', "");
Expect(1, 8447, '\p{	DIACRITICALS_for_Symbols}', "");
Expect(0, 8447, '\p{^	DIACRITICALS_for_Symbols}', "");
Expect(0, 8447, '\P{	DIACRITICALS_for_Symbols}', "");
Expect(1, 8447, '\P{^	DIACRITICALS_for_Symbols}', "");
Expect(0, 8448, '\p{	DIACRITICALS_for_Symbols}', "");
Expect(1, 8448, '\p{^	DIACRITICALS_for_Symbols}', "");
Expect(1, 8448, '\P{	DIACRITICALS_for_Symbols}', "");
Expect(0, 8448, '\P{^	DIACRITICALS_for_Symbols}', "");
Error('\p{ /a/Is_DIACRITICALS_FOR_Symbols}');
Error('\P{ /a/Is_DIACRITICALS_FOR_Symbols}');
Expect(1, 8447, '\p{isdiacriticalsforsymbols}', "");
Expect(0, 8447, '\p{^isdiacriticalsforsymbols}', "");
Expect(0, 8447, '\P{isdiacriticalsforsymbols}', "");
Expect(1, 8447, '\P{^isdiacriticalsforsymbols}', "");
Expect(0, 8448, '\p{isdiacriticalsforsymbols}', "");
Expect(1, 8448, '\p{^isdiacriticalsforsymbols}', "");
Expect(1, 8448, '\P{isdiacriticalsforsymbols}', "");
Expect(0, 8448, '\P{^isdiacriticalsforsymbols}', "");
Expect(1, 8447, '\p{_Is_DIACRITICALS_FOR_Symbols}', "");
Expect(0, 8447, '\p{^_Is_DIACRITICALS_FOR_Symbols}', "");
Expect(0, 8447, '\P{_Is_DIACRITICALS_FOR_Symbols}', "");
Expect(1, 8447, '\P{^_Is_DIACRITICALS_FOR_Symbols}', "");
Expect(0, 8448, '\p{_Is_DIACRITICALS_FOR_Symbols}', "");
Expect(1, 8448, '\p{^_Is_DIACRITICALS_FOR_Symbols}', "");
Expect(1, 8448, '\P{_Is_DIACRITICALS_FOR_Symbols}', "");
Expect(0, 8448, '\P{^_Is_DIACRITICALS_FOR_Symbols}', "");
Error('\p{/a/	 IN_Diacriticals_for_SYMBOLS}');
Error('\P{/a/	 IN_Diacriticals_for_SYMBOLS}');
Expect(1, 8447, '\p{indiacriticalsforsymbols}', "");
Expect(0, 8447, '\p{^indiacriticalsforsymbols}', "");
Expect(0, 8447, '\P{indiacriticalsforsymbols}', "");
Expect(1, 8447, '\P{^indiacriticalsforsymbols}', "");
Expect(0, 8448, '\p{indiacriticalsforsymbols}', "");
Expect(1, 8448, '\p{^indiacriticalsforsymbols}', "");
Expect(1, 8448, '\P{indiacriticalsforsymbols}', "");
Expect(0, 8448, '\P{^indiacriticalsforsymbols}', "");
Expect(1, 8447, '\p{	_In_Diacriticals_For_SYMBOLS}', "");
Expect(0, 8447, '\p{^	_In_Diacriticals_For_SYMBOLS}', "");
Expect(0, 8447, '\P{	_In_Diacriticals_For_SYMBOLS}', "");
Expect(1, 8447, '\P{^	_In_Diacriticals_For_SYMBOLS}', "");
Expect(0, 8448, '\p{	_In_Diacriticals_For_SYMBOLS}', "");
Expect(1, 8448, '\p{^	_In_Diacriticals_For_SYMBOLS}', "");
Expect(1, 8448, '\P{	_In_Diacriticals_For_SYMBOLS}', "");
Expect(0, 8448, '\P{^	_In_Diacriticals_For_SYMBOLS}', "");
Error('\p{_:=Combining_Marks_For_Symbols}');
Error('\P{_:=Combining_Marks_For_Symbols}');
Expect(1, 8447, '\p{combiningmarksforsymbols}', "");
Expect(0, 8447, '\p{^combiningmarksforsymbols}', "");
Expect(0, 8447, '\P{combiningmarksforsymbols}', "");
Expect(1, 8447, '\P{^combiningmarksforsymbols}', "");
Expect(0, 8448, '\p{combiningmarksforsymbols}', "");
Expect(1, 8448, '\p{^combiningmarksforsymbols}', "");
Expect(1, 8448, '\P{combiningmarksforsymbols}', "");
Expect(0, 8448, '\P{^combiningmarksforsymbols}', "");
Expect(1, 8447, '\p{	combining_MARKS_For_Symbols}', "");
Expect(0, 8447, '\p{^	combining_MARKS_For_Symbols}', "");
Expect(0, 8447, '\P{	combining_MARKS_For_Symbols}', "");
Expect(1, 8447, '\P{^	combining_MARKS_For_Symbols}', "");
Expect(0, 8448, '\p{	combining_MARKS_For_Symbols}', "");
Expect(1, 8448, '\p{^	combining_MARKS_For_Symbols}', "");
Expect(1, 8448, '\P{	combining_MARKS_For_Symbols}', "");
Expect(0, 8448, '\P{^	combining_MARKS_For_Symbols}', "");
Error('\p{ :=is_Combining_Marks_For_symbols}');
Error('\P{ :=is_Combining_Marks_For_symbols}');
Expect(1, 8447, '\p{iscombiningmarksforsymbols}', "");
Expect(0, 8447, '\p{^iscombiningmarksforsymbols}', "");
Expect(0, 8447, '\P{iscombiningmarksforsymbols}', "");
Expect(1, 8447, '\P{^iscombiningmarksforsymbols}', "");
Expect(0, 8448, '\p{iscombiningmarksforsymbols}', "");
Expect(1, 8448, '\p{^iscombiningmarksforsymbols}', "");
Expect(1, 8448, '\P{iscombiningmarksforsymbols}', "");
Expect(0, 8448, '\P{^iscombiningmarksforsymbols}', "");
Expect(1, 8447, '\p{- is_Combining_MARKS_For_Symbols}', "");
Expect(0, 8447, '\p{^- is_Combining_MARKS_For_Symbols}', "");
Expect(0, 8447, '\P{- is_Combining_MARKS_For_Symbols}', "");
Expect(1, 8447, '\P{^- is_Combining_MARKS_For_Symbols}', "");
Expect(0, 8448, '\p{- is_Combining_MARKS_For_Symbols}', "");
Expect(1, 8448, '\p{^- is_Combining_MARKS_For_Symbols}', "");
Expect(1, 8448, '\P{- is_Combining_MARKS_For_Symbols}', "");
Expect(0, 8448, '\P{^- is_Combining_MARKS_For_Symbols}', "");
Error('\p{/a/	-in_Combining_marks_FOR_symbols}');
Error('\P{/a/	-in_Combining_marks_FOR_symbols}');
Expect(1, 8447, '\p{incombiningmarksforsymbols}', "");
Expect(0, 8447, '\p{^incombiningmarksforsymbols}', "");
Expect(0, 8447, '\P{incombiningmarksforsymbols}', "");
Expect(1, 8447, '\P{^incombiningmarksforsymbols}', "");
Expect(0, 8448, '\p{incombiningmarksforsymbols}', "");
Expect(1, 8448, '\p{^incombiningmarksforsymbols}', "");
Expect(1, 8448, '\P{incombiningmarksforsymbols}', "");
Expect(0, 8448, '\P{^incombiningmarksforsymbols}', "");
Expect(1, 8447, '\p{- In_Combining_MARKS_For_Symbols}', "");
Expect(0, 8447, '\p{^- In_Combining_MARKS_For_Symbols}', "");
Expect(0, 8447, '\P{- In_Combining_MARKS_For_Symbols}', "");
Expect(1, 8447, '\P{^- In_Combining_MARKS_For_Symbols}', "");
Expect(0, 8448, '\p{- In_Combining_MARKS_For_Symbols}', "");
Expect(1, 8448, '\p{^- In_Combining_MARKS_For_Symbols}', "");
Expect(1, 8448, '\P{- In_Combining_MARKS_For_Symbols}', "");
Expect(0, 8448, '\P{^- In_Combining_MARKS_For_Symbols}', "");
Error('\p{/a/SMALL_Form_Variants}');
Error('\P{/a/SMALL_Form_Variants}');
Expect(1, 65135, '\p{smallformvariants}', "");
Expect(0, 65135, '\p{^smallformvariants}', "");
Expect(0, 65135, '\P{smallformvariants}', "");
Expect(1, 65135, '\P{^smallformvariants}', "");
Expect(0, 65136, '\p{smallformvariants}', "");
Expect(1, 65136, '\p{^smallformvariants}', "");
Expect(1, 65136, '\P{smallformvariants}', "");
Expect(0, 65136, '\P{^smallformvariants}', "");
Expect(1, 65135, '\p{- Small_form_variants}', "");
Expect(0, 65135, '\p{^- Small_form_variants}', "");
Expect(0, 65135, '\P{- Small_form_variants}', "");
Expect(1, 65135, '\P{^- Small_form_variants}', "");
Expect(0, 65136, '\p{- Small_form_variants}', "");
Expect(1, 65136, '\p{^- Small_form_variants}', "");
Expect(1, 65136, '\P{- Small_form_variants}', "");
Expect(0, 65136, '\P{^- Small_form_variants}', "");
Error('\p{-_Is_Small_FORM_variants/a/}');
Error('\P{-_Is_Small_FORM_variants/a/}');
Expect(1, 65135, '\p{issmallformvariants}', "");
Expect(0, 65135, '\p{^issmallformvariants}', "");
Expect(0, 65135, '\P{issmallformvariants}', "");
Expect(1, 65135, '\P{^issmallformvariants}', "");
Expect(0, 65136, '\p{issmallformvariants}', "");
Expect(1, 65136, '\p{^issmallformvariants}', "");
Expect(1, 65136, '\P{issmallformvariants}', "");
Expect(0, 65136, '\P{^issmallformvariants}', "");
Expect(1, 65135, '\p{_Is_SMALL_FORM_Variants}', "");
Expect(0, 65135, '\p{^_Is_SMALL_FORM_Variants}', "");
Expect(0, 65135, '\P{_Is_SMALL_FORM_Variants}', "");
Expect(1, 65135, '\P{^_Is_SMALL_FORM_Variants}', "");
Expect(0, 65136, '\p{_Is_SMALL_FORM_Variants}', "");
Expect(1, 65136, '\p{^_Is_SMALL_FORM_Variants}', "");
Expect(1, 65136, '\P{_Is_SMALL_FORM_Variants}', "");
Expect(0, 65136, '\P{^_Is_SMALL_FORM_Variants}', "");
Error('\p{	/a/In_Small_Form_variants}');
Error('\P{	/a/In_Small_Form_variants}');
Expect(1, 65135, '\p{insmallformvariants}', "");
Expect(0, 65135, '\p{^insmallformvariants}', "");
Expect(0, 65135, '\P{insmallformvariants}', "");
Expect(1, 65135, '\P{^insmallformvariants}', "");
Expect(0, 65136, '\p{insmallformvariants}', "");
Expect(1, 65136, '\p{^insmallformvariants}', "");
Expect(1, 65136, '\P{insmallformvariants}', "");
Expect(0, 65136, '\P{^insmallformvariants}', "");
Expect(1, 65135, '\p{--In_Small_FORM_Variants}', "");
Expect(0, 65135, '\p{^--In_Small_FORM_Variants}', "");
Expect(0, 65135, '\P{--In_Small_FORM_Variants}', "");
Expect(1, 65135, '\P{^--In_Small_FORM_Variants}', "");
Expect(0, 65136, '\p{--In_Small_FORM_Variants}', "");
Expect(1, 65136, '\p{^--In_Small_FORM_Variants}', "");
Expect(1, 65136, '\P{--In_Small_FORM_Variants}', "");
Expect(0, 65136, '\P{^--In_Small_FORM_Variants}', "");
Error('\p{_/a/Small_Forms}');
Error('\P{_/a/Small_Forms}');
Expect(1, 65135, '\p{smallforms}', "");
Expect(0, 65135, '\p{^smallforms}', "");
Expect(0, 65135, '\P{smallforms}', "");
Expect(1, 65135, '\P{^smallforms}', "");
Expect(0, 65136, '\p{smallforms}', "");
Expect(1, 65136, '\p{^smallforms}', "");
Expect(1, 65136, '\P{smallforms}', "");
Expect(0, 65136, '\P{^smallforms}', "");
Expect(1, 65135, '\p{	Small_Forms}', "");
Expect(0, 65135, '\p{^	Small_Forms}', "");
Expect(0, 65135, '\P{	Small_Forms}', "");
Expect(1, 65135, '\P{^	Small_Forms}', "");
Expect(0, 65136, '\p{	Small_Forms}', "");
Expect(1, 65136, '\p{^	Small_Forms}', "");
Expect(1, 65136, '\P{	Small_Forms}', "");
Expect(0, 65136, '\P{^	Small_Forms}', "");
Error('\p{ IS_Small_Forms:=}');
Error('\P{ IS_Small_Forms:=}');
Expect(1, 65135, '\p{issmallforms}', "");
Expect(0, 65135, '\p{^issmallforms}', "");
Expect(0, 65135, '\P{issmallforms}', "");
Expect(1, 65135, '\P{^issmallforms}', "");
Expect(0, 65136, '\p{issmallforms}', "");
Expect(1, 65136, '\p{^issmallforms}', "");
Expect(1, 65136, '\P{issmallforms}', "");
Expect(0, 65136, '\P{^issmallforms}', "");
Expect(1, 65135, '\p{- is_small_FORMS}', "");
Expect(0, 65135, '\p{^- is_small_FORMS}', "");
Expect(0, 65135, '\P{- is_small_FORMS}', "");
Expect(1, 65135, '\P{^- is_small_FORMS}', "");
Expect(0, 65136, '\p{- is_small_FORMS}', "");
Expect(1, 65136, '\p{^- is_small_FORMS}', "");
Expect(1, 65136, '\P{- is_small_FORMS}', "");
Expect(0, 65136, '\P{^- is_small_FORMS}', "");
Error('\p{:=in_Small_FORMS}');
Error('\P{:=in_Small_FORMS}');
Expect(1, 65135, '\p{insmallforms}', "");
Expect(0, 65135, '\p{^insmallforms}', "");
Expect(0, 65135, '\P{insmallforms}', "");
Expect(1, 65135, '\P{^insmallforms}', "");
Expect(0, 65136, '\p{insmallforms}', "");
Expect(1, 65136, '\p{^insmallforms}', "");
Expect(1, 65136, '\P{insmallforms}', "");
Expect(0, 65136, '\P{^insmallforms}', "");
Expect(1, 65135, '\p{-	In_small_forms}', "");
Expect(0, 65135, '\p{^-	In_small_forms}', "");
Expect(0, 65135, '\P{-	In_small_forms}', "");
Expect(1, 65135, '\P{^-	In_small_forms}', "");
Expect(0, 65136, '\p{-	In_small_forms}', "");
Expect(1, 65136, '\p{^-	In_small_forms}', "");
Expect(1, 65136, '\P{-	In_small_forms}', "");
Expect(0, 65136, '\P{^-	In_small_forms}', "");
Error('\p{-/a/Private_Use_area}');
Error('\P{-/a/Private_Use_area}');
Expect(1, 63743, '\p{privateusearea}', "");
Expect(0, 63743, '\p{^privateusearea}', "");
Expect(0, 63743, '\P{privateusearea}', "");
Expect(1, 63743, '\P{^privateusearea}', "");
Expect(0, 63744, '\p{privateusearea}', "");
Expect(1, 63744, '\p{^privateusearea}', "");
Expect(1, 63744, '\P{privateusearea}', "");
Expect(0, 63744, '\P{^privateusearea}', "");
Expect(1, 63743, '\p{ PRIVATE_Use_area}', "");
Expect(0, 63743, '\p{^ PRIVATE_Use_area}', "");
Expect(0, 63743, '\P{ PRIVATE_Use_area}', "");
Expect(1, 63743, '\P{^ PRIVATE_Use_area}', "");
Expect(0, 63744, '\p{ PRIVATE_Use_area}', "");
Expect(1, 63744, '\p{^ PRIVATE_Use_area}', "");
Expect(1, 63744, '\P{ PRIVATE_Use_area}', "");
Expect(0, 63744, '\P{^ PRIVATE_Use_area}', "");
Error('\p{:=	Is_Private_Use_Area}');
Error('\P{:=	Is_Private_Use_Area}');
Expect(1, 63743, '\p{isprivateusearea}', "");
Expect(0, 63743, '\p{^isprivateusearea}', "");
Expect(0, 63743, '\P{isprivateusearea}', "");
Expect(1, 63743, '\P{^isprivateusearea}', "");
Expect(0, 63744, '\p{isprivateusearea}', "");
Expect(1, 63744, '\p{^isprivateusearea}', "");
Expect(1, 63744, '\P{isprivateusearea}', "");
Expect(0, 63744, '\P{^isprivateusearea}', "");
Expect(1, 63743, '\p{	-Is_PRIVATE_use_Area}', "");
Expect(0, 63743, '\p{^	-Is_PRIVATE_use_Area}', "");
Expect(0, 63743, '\P{	-Is_PRIVATE_use_Area}', "");
Expect(1, 63743, '\P{^	-Is_PRIVATE_use_Area}', "");
Expect(0, 63744, '\p{	-Is_PRIVATE_use_Area}', "");
Expect(1, 63744, '\p{^	-Is_PRIVATE_use_Area}', "");
Expect(1, 63744, '\P{	-Is_PRIVATE_use_Area}', "");
Expect(0, 63744, '\P{^	-Is_PRIVATE_use_Area}', "");
Error('\p{-	in_PRIVATE_USE_Area:=}');
Error('\P{-	in_PRIVATE_USE_Area:=}');
Expect(1, 63743, '\p{inprivateusearea}', "");
Expect(0, 63743, '\p{^inprivateusearea}', "");
Expect(0, 63743, '\P{inprivateusearea}', "");
Expect(1, 63743, '\P{^inprivateusearea}', "");
Expect(0, 63744, '\p{inprivateusearea}', "");
Expect(1, 63744, '\p{^inprivateusearea}', "");
Expect(1, 63744, '\P{inprivateusearea}', "");
Expect(0, 63744, '\P{^inprivateusearea}', "");
Expect(1, 63743, '\p{--In_Private_Use_AREA}', "");
Expect(0, 63743, '\p{^--In_Private_Use_AREA}', "");
Expect(0, 63743, '\P{--In_Private_Use_AREA}', "");
Expect(1, 63743, '\P{^--In_Private_Use_AREA}', "");
Expect(0, 63744, '\p{--In_Private_Use_AREA}', "");
Expect(1, 63744, '\p{^--In_Private_Use_AREA}', "");
Expect(1, 63744, '\P{--In_Private_Use_AREA}', "");
Expect(0, 63744, '\P{^--In_Private_Use_AREA}', "");
Error('\p{/a/  PUA}');
Error('\P{/a/  PUA}');
Expect(1, 63743, '\p{pua}', "");
Expect(0, 63743, '\p{^pua}', "");
Expect(0, 63743, '\P{pua}', "");
Expect(1, 63743, '\P{^pua}', "");
Expect(0, 63744, '\p{pua}', "");
Expect(1, 63744, '\p{^pua}', "");
Expect(1, 63744, '\P{pua}', "");
Expect(0, 63744, '\P{^pua}', "");
Expect(1, 63743, '\p{_ PUA}', "");
Expect(0, 63743, '\p{^_ PUA}', "");
Expect(0, 63743, '\P{_ PUA}', "");
Expect(1, 63743, '\P{^_ PUA}', "");
Expect(0, 63744, '\p{_ PUA}', "");
Expect(1, 63744, '\p{^_ PUA}', "");
Expect(1, 63744, '\P{_ PUA}', "");
Expect(0, 63744, '\P{^_ PUA}', "");
Error('\p{/a/	Is_pua}');
Error('\P{/a/	Is_pua}');
Expect(1, 63743, '\p{ispua}', "");
Expect(0, 63743, '\p{^ispua}', "");
Expect(0, 63743, '\P{ispua}', "");
Expect(1, 63743, '\P{^ispua}', "");
Expect(0, 63744, '\p{ispua}', "");
Expect(1, 63744, '\p{^ispua}', "");
Expect(1, 63744, '\P{ispua}', "");
Expect(0, 63744, '\P{^ispua}', "");
Expect(1, 63743, '\p{	 IS_PUA}', "");
Expect(0, 63743, '\p{^	 IS_PUA}', "");
Expect(0, 63743, '\P{	 IS_PUA}', "");
Expect(1, 63743, '\P{^	 IS_PUA}', "");
Expect(0, 63744, '\p{	 IS_PUA}', "");
Expect(1, 63744, '\p{^	 IS_PUA}', "");
Expect(1, 63744, '\P{	 IS_PUA}', "");
Expect(0, 63744, '\P{^	 IS_PUA}', "");
Error('\p{ :=IN_PUA}');
Error('\P{ :=IN_PUA}');
Expect(1, 63743, '\p{inpua}', "");
Expect(0, 63743, '\p{^inpua}', "");
Expect(0, 63743, '\P{inpua}', "");
Expect(1, 63743, '\P{^inpua}', "");
Expect(0, 63744, '\p{inpua}', "");
Expect(1, 63744, '\p{^inpua}', "");
Expect(1, 63744, '\P{inpua}', "");
Expect(0, 63744, '\P{^inpua}', "");
Expect(1, 63743, '\p{ In_PUA}', "");
Expect(0, 63743, '\p{^ In_PUA}', "");
Expect(0, 63743, '\P{ In_PUA}', "");
Expect(1, 63743, '\P{^ In_PUA}', "");
Expect(0, 63744, '\p{ In_PUA}', "");
Expect(1, 63744, '\p{^ In_PUA}', "");
Expect(1, 63744, '\P{ In_PUA}', "");
Expect(0, 63744, '\P{^ In_PUA}', "");
Error('\p{-/a/in_Private_Use}');
Error('\P{-/a/in_Private_Use}');
Expect(1, 63743, '\p{inprivateuse}', "");
Expect(0, 63743, '\p{^inprivateuse}', "");
Expect(0, 63743, '\P{inprivateuse}', "");
Expect(1, 63743, '\P{^inprivateuse}', "");
Expect(0, 63744, '\p{inprivateuse}', "");
Expect(1, 63744, '\p{^inprivateuse}', "");
Expect(1, 63744, '\P{inprivateuse}', "");
Expect(0, 63744, '\P{^inprivateuse}', "");
Expect(1, 63743, '\p{	IN_PRIVATE_USE}', "");
Expect(0, 63743, '\p{^	IN_PRIVATE_USE}', "");
Expect(0, 63743, '\P{	IN_PRIVATE_USE}', "");
Expect(1, 63743, '\P{^	IN_PRIVATE_USE}', "");
Expect(0, 63744, '\p{	IN_PRIVATE_USE}', "");
Expect(1, 63744, '\p{^	IN_PRIVATE_USE}', "");
Expect(1, 63744, '\P{	IN_PRIVATE_USE}', "");
Expect(0, 63744, '\P{^	IN_PRIVATE_USE}', "");
Error('\p{:=_	TAI_THAM}');
Error('\P{:=_	TAI_THAM}');
Expect(1, 6829, '\p{taitham}', "");
Expect(0, 6829, '\p{^taitham}', "");
Expect(0, 6829, '\P{taitham}', "");
Expect(1, 6829, '\P{^taitham}', "");
Expect(0, 6830, '\p{taitham}', "");
Expect(1, 6830, '\p{^taitham}', "");
Expect(1, 6830, '\P{taitham}', "");
Expect(0, 6830, '\P{^taitham}', "");
Expect(1, 6829, '\p{	TAI_Tham}', "");
Expect(0, 6829, '\p{^	TAI_Tham}', "");
Expect(0, 6829, '\P{	TAI_Tham}', "");
Expect(1, 6829, '\P{^	TAI_Tham}', "");
Expect(0, 6830, '\p{	TAI_Tham}', "");
Expect(1, 6830, '\p{^	TAI_Tham}', "");
Expect(1, 6830, '\P{	TAI_Tham}', "");
Expect(0, 6830, '\P{^	TAI_Tham}', "");
Error('\p{/a/  IS_Tai_tham}');
Error('\P{/a/  IS_Tai_tham}');
Expect(1, 6829, '\p{istaitham}', "");
Expect(0, 6829, '\p{^istaitham}', "");
Expect(0, 6829, '\P{istaitham}', "");
Expect(1, 6829, '\P{^istaitham}', "");
Expect(0, 6830, '\p{istaitham}', "");
Expect(1, 6830, '\p{^istaitham}', "");
Expect(1, 6830, '\P{istaitham}', "");
Expect(0, 6830, '\P{^istaitham}', "");
Expect(1, 6829, '\p{__is_TAI_tham}', "");
Expect(0, 6829, '\p{^__is_TAI_tham}', "");
Expect(0, 6829, '\P{__is_TAI_tham}', "");
Expect(1, 6829, '\P{^__is_TAI_tham}', "");
Expect(0, 6830, '\p{__is_TAI_tham}', "");
Expect(1, 6830, '\p{^__is_TAI_tham}', "");
Expect(1, 6830, '\P{__is_TAI_tham}', "");
Expect(0, 6830, '\P{^__is_TAI_tham}', "");
Error('\p{/a/-_Lana}');
Error('\P{/a/-_Lana}');
Expect(1, 6829, '\p{lana}', "");
Expect(0, 6829, '\p{^lana}', "");
Expect(0, 6829, '\P{lana}', "");
Expect(1, 6829, '\P{^lana}', "");
Expect(0, 6830, '\p{lana}', "");
Expect(1, 6830, '\p{^lana}', "");
Expect(1, 6830, '\P{lana}', "");
Expect(0, 6830, '\P{^lana}', "");
Expect(1, 6829, '\p{	Lana}', "");
Expect(0, 6829, '\p{^	Lana}', "");
Expect(0, 6829, '\P{	Lana}', "");
Expect(1, 6829, '\P{^	Lana}', "");
Expect(0, 6830, '\p{	Lana}', "");
Expect(1, 6830, '\p{^	Lana}', "");
Expect(1, 6830, '\P{	Lana}', "");
Expect(0, 6830, '\P{^	Lana}', "");
Error('\p{_:=Is_Lana}');
Error('\P{_:=Is_Lana}');
Expect(1, 6829, '\p{islana}', "");
Expect(0, 6829, '\p{^islana}', "");
Expect(0, 6829, '\P{islana}', "");
Expect(1, 6829, '\P{^islana}', "");
Expect(0, 6830, '\p{islana}', "");
Expect(1, 6830, '\p{^islana}', "");
Expect(1, 6830, '\P{islana}', "");
Expect(0, 6830, '\P{^islana}', "");
Expect(1, 6829, '\p{_is_LANA}', "");
Expect(0, 6829, '\p{^_is_LANA}', "");
Expect(0, 6829, '\P{_is_LANA}', "");
Expect(1, 6829, '\P{^_is_LANA}', "");
Expect(0, 6830, '\p{_is_LANA}', "");
Expect(1, 6830, '\p{^_is_LANA}', "");
Expect(1, 6830, '\P{_is_LANA}', "");
Expect(0, 6830, '\P{^_is_LANA}', "");
Error('\p{ 	nko:=}');
Error('\P{ 	nko:=}');
Expect(1, 2042, '\p{nko}', "");
Expect(0, 2042, '\p{^nko}', "");
Expect(0, 2042, '\P{nko}', "");
Expect(1, 2042, '\P{^nko}', "");
Expect(0, 2043, '\p{nko}', "");
Expect(1, 2043, '\p{^nko}', "");
Expect(1, 2043, '\P{nko}', "");
Expect(0, 2043, '\P{^nko}', "");
Expect(1, 2042, '\p{ -Nko}', "");
Expect(0, 2042, '\p{^ -Nko}', "");
Expect(0, 2042, '\P{ -Nko}', "");
Expect(1, 2042, '\P{^ -Nko}', "");
Expect(0, 2043, '\p{ -Nko}', "");
Expect(1, 2043, '\p{^ -Nko}', "");
Expect(1, 2043, '\P{ -Nko}', "");
Expect(0, 2043, '\P{^ -Nko}', "");
Error('\p{	_Is_NKO:=}');
Error('\P{	_Is_NKO:=}');
Expect(1, 2042, '\p{isnko}', "");
Expect(0, 2042, '\p{^isnko}', "");
Expect(0, 2042, '\P{isnko}', "");
Expect(1, 2042, '\P{^isnko}', "");
Expect(0, 2043, '\p{isnko}', "");
Expect(1, 2043, '\p{^isnko}', "");
Expect(1, 2043, '\P{isnko}', "");
Expect(0, 2043, '\P{^isnko}', "");
Expect(1, 2042, '\p{Is_Nko}', "");
Expect(0, 2042, '\p{^Is_Nko}', "");
Expect(0, 2042, '\P{Is_Nko}', "");
Expect(1, 2042, '\P{^Is_Nko}', "");
Expect(0, 2043, '\p{Is_Nko}', "");
Expect(1, 2043, '\p{^Is_Nko}', "");
Expect(1, 2043, '\P{Is_Nko}', "");
Expect(0, 2043, '\P{^Is_Nko}', "");
Error('\p{/a/__Nkoo}');
Error('\P{/a/__Nkoo}');
Expect(1, 2042, '\p{nkoo}', "");
Expect(0, 2042, '\p{^nkoo}', "");
Expect(0, 2042, '\P{nkoo}', "");
Expect(1, 2042, '\P{^nkoo}', "");
Expect(0, 2043, '\p{nkoo}', "");
Expect(1, 2043, '\p{^nkoo}', "");
Expect(1, 2043, '\P{nkoo}', "");
Expect(0, 2043, '\P{^nkoo}', "");
Expect(1, 2042, '\p{	 Nkoo}', "");
Expect(0, 2042, '\p{^	 Nkoo}', "");
Expect(0, 2042, '\P{	 Nkoo}', "");
Expect(1, 2042, '\P{^	 Nkoo}', "");
Expect(0, 2043, '\p{	 Nkoo}', "");
Expect(1, 2043, '\p{^	 Nkoo}', "");
Expect(1, 2043, '\P{	 Nkoo}', "");
Expect(0, 2043, '\P{^	 Nkoo}', "");
Error('\p{:=	_Is_nkoo}');
Error('\P{:=	_Is_nkoo}');
Expect(1, 2042, '\p{isnkoo}', "");
Expect(0, 2042, '\p{^isnkoo}', "");
Expect(0, 2042, '\P{isnkoo}', "");
Expect(1, 2042, '\P{^isnkoo}', "");
Expect(0, 2043, '\p{isnkoo}', "");
Expect(1, 2043, '\p{^isnkoo}', "");
Expect(1, 2043, '\P{isnkoo}', "");
Expect(0, 2043, '\P{^isnkoo}', "");
Expect(1, 2042, '\p{	 Is_Nkoo}', "");
Expect(0, 2042, '\p{^	 Is_Nkoo}', "");
Expect(0, 2042, '\P{	 Is_Nkoo}', "");
Expect(1, 2042, '\P{^	 Is_Nkoo}', "");
Expect(0, 2043, '\p{	 Is_Nkoo}', "");
Expect(1, 2043, '\p{^	 Is_Nkoo}', "");
Expect(1, 2043, '\P{	 Is_Nkoo}', "");
Expect(0, 2043, '\P{^	 Is_Nkoo}', "");
Error('\p{ /a/Mathematical_Alphanumeric_Symbols}');
Error('\P{ /a/Mathematical_Alphanumeric_Symbols}');
Expect(1, 120831, '\p{mathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\p{^mathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\P{mathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\P{^mathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\p{mathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\p{^mathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\P{mathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\P{^mathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\p{_	mathematical_Alphanumeric_Symbols}', "");
Expect(0, 120831, '\p{^_	mathematical_Alphanumeric_Symbols}', "");
Expect(0, 120831, '\P{_	mathematical_Alphanumeric_Symbols}', "");
Expect(1, 120831, '\P{^_	mathematical_Alphanumeric_Symbols}', "");
Expect(0, 120832, '\p{_	mathematical_Alphanumeric_Symbols}', "");
Expect(1, 120832, '\p{^_	mathematical_Alphanumeric_Symbols}', "");
Expect(1, 120832, '\P{_	mathematical_Alphanumeric_Symbols}', "");
Expect(0, 120832, '\P{^_	mathematical_Alphanumeric_Symbols}', "");
Error('\p{/a/	 is_Mathematical_ALPHANUMERIC_Symbols}');
Error('\P{/a/	 is_Mathematical_ALPHANUMERIC_Symbols}');
Expect(1, 120831, '\p{ismathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\p{^ismathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\P{ismathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\P{^ismathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\p{ismathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\p{^ismathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\P{ismathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\P{^ismathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\p{-Is_MATHEMATICAL_Alphanumeric_SYMBOLS}', "");
Expect(0, 120831, '\p{^-Is_MATHEMATICAL_Alphanumeric_SYMBOLS}', "");
Expect(0, 120831, '\P{-Is_MATHEMATICAL_Alphanumeric_SYMBOLS}', "");
Expect(1, 120831, '\P{^-Is_MATHEMATICAL_Alphanumeric_SYMBOLS}', "");
Expect(0, 120832, '\p{-Is_MATHEMATICAL_Alphanumeric_SYMBOLS}', "");
Expect(1, 120832, '\p{^-Is_MATHEMATICAL_Alphanumeric_SYMBOLS}', "");
Expect(1, 120832, '\P{-Is_MATHEMATICAL_Alphanumeric_SYMBOLS}', "");
Expect(0, 120832, '\P{^-Is_MATHEMATICAL_Alphanumeric_SYMBOLS}', "");
Error('\p{/a/-in_MATHEMATICAL_Alphanumeric_symbols}');
Error('\P{/a/-in_MATHEMATICAL_Alphanumeric_symbols}');
Expect(1, 120831, '\p{inmathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\p{^inmathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\P{inmathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\P{^inmathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\p{inmathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\p{^inmathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\P{inmathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\P{^inmathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\p{-in_mathematical_alphanumeric_symbols}', "");
Expect(0, 120831, '\p{^-in_mathematical_alphanumeric_symbols}', "");
Expect(0, 120831, '\P{-in_mathematical_alphanumeric_symbols}', "");
Expect(1, 120831, '\P{^-in_mathematical_alphanumeric_symbols}', "");
Expect(0, 120832, '\p{-in_mathematical_alphanumeric_symbols}', "");
Expect(1, 120832, '\p{^-in_mathematical_alphanumeric_symbols}', "");
Expect(1, 120832, '\P{-in_mathematical_alphanumeric_symbols}', "");
Expect(0, 120832, '\P{^-in_mathematical_alphanumeric_symbols}', "");
Error('\p{_/a/Math_Alphanum}');
Error('\P{_/a/Math_Alphanum}');
Expect(1, 120831, '\p{mathalphanum}', "");
Expect(0, 120831, '\p{^mathalphanum}', "");
Expect(0, 120831, '\P{mathalphanum}', "");
Expect(1, 120831, '\P{^mathalphanum}', "");
Expect(0, 120832, '\p{mathalphanum}', "");
Expect(1, 120832, '\p{^mathalphanum}', "");
Expect(1, 120832, '\P{mathalphanum}', "");
Expect(0, 120832, '\P{^mathalphanum}', "");
Expect(1, 120831, '\p{MATH_Alphanum}', "");
Expect(0, 120831, '\p{^MATH_Alphanum}', "");
Expect(0, 120831, '\P{MATH_Alphanum}', "");
Expect(1, 120831, '\P{^MATH_Alphanum}', "");
Expect(0, 120832, '\p{MATH_Alphanum}', "");
Expect(1, 120832, '\p{^MATH_Alphanum}', "");
Expect(1, 120832, '\P{MATH_Alphanum}', "");
Expect(0, 120832, '\P{^MATH_Alphanum}', "");
Error('\p{/a/-Is_math_alphanum}');
Error('\P{/a/-Is_math_alphanum}');
Expect(1, 120831, '\p{ismathalphanum}', "");
Expect(0, 120831, '\p{^ismathalphanum}', "");
Expect(0, 120831, '\P{ismathalphanum}', "");
Expect(1, 120831, '\P{^ismathalphanum}', "");
Expect(0, 120832, '\p{ismathalphanum}', "");
Expect(1, 120832, '\p{^ismathalphanum}', "");
Expect(1, 120832, '\P{ismathalphanum}', "");
Expect(0, 120832, '\P{^ismathalphanum}', "");
Expect(1, 120831, '\p{is_Math_Alphanum}', "");
Expect(0, 120831, '\p{^is_Math_Alphanum}', "");
Expect(0, 120831, '\P{is_Math_Alphanum}', "");
Expect(1, 120831, '\P{^is_Math_Alphanum}', "");
Expect(0, 120832, '\p{is_Math_Alphanum}', "");
Expect(1, 120832, '\p{^is_Math_Alphanum}', "");
Expect(1, 120832, '\P{is_Math_Alphanum}', "");
Expect(0, 120832, '\P{^is_Math_Alphanum}', "");
Error('\p{/a/		in_Math_alphanum}');
Error('\P{/a/		in_Math_alphanum}');
Expect(1, 120831, '\p{inmathalphanum}', "");
Expect(0, 120831, '\p{^inmathalphanum}', "");
Expect(0, 120831, '\P{inmathalphanum}', "");
Expect(1, 120831, '\P{^inmathalphanum}', "");
Expect(0, 120832, '\p{inmathalphanum}', "");
Expect(1, 120832, '\p{^inmathalphanum}', "");
Expect(1, 120832, '\P{inmathalphanum}', "");
Expect(0, 120832, '\P{^inmathalphanum}', "");
Expect(1, 120831, '\p{_-In_Math_alphanum}', "");
Expect(0, 120831, '\p{^_-In_Math_alphanum}', "");
Expect(0, 120831, '\P{_-In_Math_alphanum}', "");
Expect(1, 120831, '\P{^_-In_Math_alphanum}', "");
Expect(0, 120832, '\p{_-In_Math_alphanum}', "");
Expect(1, 120832, '\p{^_-In_Math_alphanum}', "");
Expect(1, 120832, '\P{_-In_Math_alphanum}', "");
Expect(0, 120832, '\P{^_-In_Math_alphanum}', "");
Error('\p{	/a/Lydian}');
Error('\P{	/a/Lydian}');
Expect(1, 67903, '\p{lydian}', "");
Expect(0, 67903, '\p{^lydian}', "");
Expect(0, 67903, '\P{lydian}', "");
Expect(1, 67903, '\P{^lydian}', "");
Expect(0, 67904, '\p{lydian}', "");
Expect(1, 67904, '\p{^lydian}', "");
Expect(1, 67904, '\P{lydian}', "");
Expect(0, 67904, '\P{^lydian}', "");
Expect(1, 67903, '\p{_-Lydian}', "");
Expect(0, 67903, '\p{^_-Lydian}', "");
Expect(0, 67903, '\P{_-Lydian}', "");
Expect(1, 67903, '\P{^_-Lydian}', "");
Expect(0, 67904, '\p{_-Lydian}', "");
Expect(1, 67904, '\p{^_-Lydian}', "");
Expect(1, 67904, '\P{_-Lydian}', "");
Expect(0, 67904, '\P{^_-Lydian}', "");
Error('\p{ /a/Is_lydian}');
Error('\P{ /a/Is_lydian}');
Expect(1, 67903, '\p{islydian}', "");
Expect(0, 67903, '\p{^islydian}', "");
Expect(0, 67903, '\P{islydian}', "");
Expect(1, 67903, '\P{^islydian}', "");
Expect(0, 67904, '\p{islydian}', "");
Expect(1, 67904, '\p{^islydian}', "");
Expect(1, 67904, '\P{islydian}', "");
Expect(0, 67904, '\P{^islydian}', "");
Expect(1, 67903, '\p{		Is_Lydian}', "");
Expect(0, 67903, '\p{^		Is_Lydian}', "");
Expect(0, 67903, '\P{		Is_Lydian}', "");
Expect(1, 67903, '\P{^		Is_Lydian}', "");
Expect(0, 67904, '\p{		Is_Lydian}', "");
Expect(1, 67904, '\p{^		Is_Lydian}', "");
Expect(1, 67904, '\P{		Is_Lydian}', "");
Expect(0, 67904, '\P{^		Is_Lydian}', "");
Error('\p{ /a/lydi}');
Error('\P{ /a/lydi}');
Expect(1, 67903, '\p{lydi}', "");
Expect(0, 67903, '\p{^lydi}', "");
Expect(0, 67903, '\P{lydi}', "");
Expect(1, 67903, '\P{^lydi}', "");
Expect(0, 67904, '\p{lydi}', "");
Expect(1, 67904, '\p{^lydi}', "");
Expect(1, 67904, '\P{lydi}', "");
Expect(0, 67904, '\P{^lydi}', "");
Expect(1, 67903, '\p{-	Lydi}', "");
Expect(0, 67903, '\p{^-	Lydi}', "");
Expect(0, 67903, '\P{-	Lydi}', "");
Expect(1, 67903, '\P{^-	Lydi}', "");
Expect(0, 67904, '\p{-	Lydi}', "");
Expect(1, 67904, '\p{^-	Lydi}', "");
Expect(1, 67904, '\P{-	Lydi}', "");
Expect(0, 67904, '\P{^-	Lydi}', "");
Error('\p{--is_Lydi/a/}');
Error('\P{--is_Lydi/a/}');
Expect(1, 67903, '\p{islydi}', "");
Expect(0, 67903, '\p{^islydi}', "");
Expect(0, 67903, '\P{islydi}', "");
Expect(1, 67903, '\P{^islydi}', "");
Expect(0, 67904, '\p{islydi}', "");
Expect(1, 67904, '\p{^islydi}', "");
Expect(1, 67904, '\P{islydi}', "");
Expect(0, 67904, '\P{^islydi}', "");
Expect(1, 67903, '\p{ -IS_Lydi}', "");
Expect(0, 67903, '\p{^ -IS_Lydi}', "");
Expect(0, 67903, '\P{ -IS_Lydi}', "");
Expect(1, 67903, '\P{^ -IS_Lydi}', "");
Expect(0, 67904, '\p{ -IS_Lydi}', "");
Expect(1, 67904, '\p{^ -IS_Lydi}', "");
Expect(1, 67904, '\P{ -IS_Lydi}', "");
Expect(0, 67904, '\P{^ -IS_Lydi}', "");
Error('\p{:=_Georgian_supplement}');
Error('\P{:=_Georgian_supplement}');
Expect(1, 11567, '\p{georgiansupplement}', "");
Expect(0, 11567, '\p{^georgiansupplement}', "");
Expect(0, 11567, '\P{georgiansupplement}', "");
Expect(1, 11567, '\P{^georgiansupplement}', "");
Expect(0, 11568, '\p{georgiansupplement}', "");
Expect(1, 11568, '\p{^georgiansupplement}', "");
Expect(1, 11568, '\P{georgiansupplement}', "");
Expect(0, 11568, '\P{^georgiansupplement}', "");
Expect(1, 11567, '\p{ 	Georgian_SUPPLEMENT}', "");
Expect(0, 11567, '\p{^ 	Georgian_SUPPLEMENT}', "");
Expect(0, 11567, '\P{ 	Georgian_SUPPLEMENT}', "");
Expect(1, 11567, '\P{^ 	Georgian_SUPPLEMENT}', "");
Expect(0, 11568, '\p{ 	Georgian_SUPPLEMENT}', "");
Expect(1, 11568, '\p{^ 	Georgian_SUPPLEMENT}', "");
Expect(1, 11568, '\P{ 	Georgian_SUPPLEMENT}', "");
Expect(0, 11568, '\P{^ 	Georgian_SUPPLEMENT}', "");
Error('\p{-/a/IS_georgian_SUPPLEMENT}');
Error('\P{-/a/IS_georgian_SUPPLEMENT}');
Expect(1, 11567, '\p{isgeorgiansupplement}', "");
Expect(0, 11567, '\p{^isgeorgiansupplement}', "");
Expect(0, 11567, '\P{isgeorgiansupplement}', "");
Expect(1, 11567, '\P{^isgeorgiansupplement}', "");
Expect(0, 11568, '\p{isgeorgiansupplement}', "");
Expect(1, 11568, '\p{^isgeorgiansupplement}', "");
Expect(1, 11568, '\P{isgeorgiansupplement}', "");
Expect(0, 11568, '\P{^isgeorgiansupplement}', "");
Expect(1, 11567, '\p{_ Is_georgian_SUPPLEMENT}', "");
Expect(0, 11567, '\p{^_ Is_georgian_SUPPLEMENT}', "");
Expect(0, 11567, '\P{_ Is_georgian_SUPPLEMENT}', "");
Expect(1, 11567, '\P{^_ Is_georgian_SUPPLEMENT}', "");
Expect(0, 11568, '\p{_ Is_georgian_SUPPLEMENT}', "");
Expect(1, 11568, '\p{^_ Is_georgian_SUPPLEMENT}', "");
Expect(1, 11568, '\P{_ Is_georgian_SUPPLEMENT}', "");
Expect(0, 11568, '\P{^_ Is_georgian_SUPPLEMENT}', "");
Error('\p{:=	In_georgian_Supplement}');
Error('\P{:=	In_georgian_Supplement}');
Expect(1, 11567, '\p{ingeorgiansupplement}', "");
Expect(0, 11567, '\p{^ingeorgiansupplement}', "");
Expect(0, 11567, '\P{ingeorgiansupplement}', "");
Expect(1, 11567, '\P{^ingeorgiansupplement}', "");
Expect(0, 11568, '\p{ingeorgiansupplement}', "");
Expect(1, 11568, '\p{^ingeorgiansupplement}', "");
Expect(1, 11568, '\P{ingeorgiansupplement}', "");
Expect(0, 11568, '\P{^ingeorgiansupplement}', "");
Expect(1, 11567, '\p{_in_Georgian_supplement}', "");
Expect(0, 11567, '\p{^_in_Georgian_supplement}', "");
Expect(0, 11567, '\P{_in_Georgian_supplement}', "");
Expect(1, 11567, '\P{^_in_Georgian_supplement}', "");
Expect(0, 11568, '\p{_in_Georgian_supplement}', "");
Expect(1, 11568, '\p{^_in_Georgian_supplement}', "");
Expect(1, 11568, '\P{_in_Georgian_supplement}', "");
Expect(0, 11568, '\P{^_in_Georgian_supplement}', "");
Error('\p{/a/ Georgian_sup}');
Error('\P{/a/ Georgian_sup}');
Expect(1, 11567, '\p{georgiansup}', "");
Expect(0, 11567, '\p{^georgiansup}', "");
Expect(0, 11567, '\P{georgiansup}', "");
Expect(1, 11567, '\P{^georgiansup}', "");
Expect(0, 11568, '\p{georgiansup}', "");
Expect(1, 11568, '\p{^georgiansup}', "");
Expect(1, 11568, '\P{georgiansup}', "");
Expect(0, 11568, '\P{^georgiansup}', "");
Expect(1, 11567, '\p{	Georgian_Sup}', "");
Expect(0, 11567, '\p{^	Georgian_Sup}', "");
Expect(0, 11567, '\P{	Georgian_Sup}', "");
Expect(1, 11567, '\P{^	Georgian_Sup}', "");
Expect(0, 11568, '\p{	Georgian_Sup}', "");
Expect(1, 11568, '\p{^	Georgian_Sup}', "");
Expect(1, 11568, '\P{	Georgian_Sup}', "");
Expect(0, 11568, '\P{^	Georgian_Sup}', "");
Error('\p{/a/is_GEORGIAN_sup}');
Error('\P{/a/is_GEORGIAN_sup}');
Expect(1, 11567, '\p{isgeorgiansup}', "");
Expect(0, 11567, '\p{^isgeorgiansup}', "");
Expect(0, 11567, '\P{isgeorgiansup}', "");
Expect(1, 11567, '\P{^isgeorgiansup}', "");
Expect(0, 11568, '\p{isgeorgiansup}', "");
Expect(1, 11568, '\p{^isgeorgiansup}', "");
Expect(1, 11568, '\P{isgeorgiansup}', "");
Expect(0, 11568, '\P{^isgeorgiansup}', "");
Expect(1, 11567, '\p{-_Is_GEORGIAN_sup}', "");
Expect(0, 11567, '\p{^-_Is_GEORGIAN_sup}', "");
Expect(0, 11567, '\P{-_Is_GEORGIAN_sup}', "");
Expect(1, 11567, '\P{^-_Is_GEORGIAN_sup}', "");
Expect(0, 11568, '\p{-_Is_GEORGIAN_sup}', "");
Expect(1, 11568, '\p{^-_Is_GEORGIAN_sup}', "");
Expect(1, 11568, '\P{-_Is_GEORGIAN_sup}', "");
Expect(0, 11568, '\P{^-_Is_GEORGIAN_sup}', "");
Error('\p{/a/-In_GEORGIAN_Sup}');
Error('\P{/a/-In_GEORGIAN_Sup}');
Expect(1, 11567, '\p{ingeorgiansup}', "");
Expect(0, 11567, '\p{^ingeorgiansup}', "");
Expect(0, 11567, '\P{ingeorgiansup}', "");
Expect(1, 11567, '\P{^ingeorgiansup}', "");
Expect(0, 11568, '\p{ingeorgiansup}', "");
Expect(1, 11568, '\p{^ingeorgiansup}', "");
Expect(1, 11568, '\P{ingeorgiansup}', "");
Expect(0, 11568, '\P{^ingeorgiansup}', "");
Expect(1, 11567, '\p{ in_georgian_SUP}', "");
Expect(0, 11567, '\p{^ in_georgian_SUP}', "");
Expect(0, 11567, '\P{ in_georgian_SUP}', "");
Expect(1, 11567, '\P{^ in_georgian_SUP}', "");
Expect(0, 11568, '\p{ in_georgian_SUP}', "");
Expect(1, 11568, '\p{^ in_georgian_SUP}', "");
Expect(1, 11568, '\P{ in_georgian_SUP}', "");
Expect(0, 11568, '\P{^ in_georgian_SUP}', "");
Error('\p{-_MEROITIC_Hieroglyphs:=}');
Error('\P{-_MEROITIC_Hieroglyphs:=}');
Expect(1, 67999, '\p{meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^meroitichieroglyphs}', "");
Expect(0, 67999, '\P{meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^meroitichieroglyphs}', "");
Expect(0, 68000, '\p{meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^meroitichieroglyphs}', "");
Expect(1, 68000, '\P{meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^meroitichieroglyphs}', "");
Expect(1, 67999, '\p{_Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\p{^_Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\P{_Meroitic_Hieroglyphs}', "");
Expect(1, 67999, '\P{^_Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\p{_Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\p{^_Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\P{_Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\P{^_Meroitic_Hieroglyphs}', "");
Error('\p{__Is_meroitic_Hieroglyphs/a/}');
Error('\P{__Is_meroitic_Hieroglyphs/a/}');
Expect(1, 67999, '\p{ismeroitichieroglyphs}', "");
Expect(0, 67999, '\p{^ismeroitichieroglyphs}', "");
Expect(0, 67999, '\P{ismeroitichieroglyphs}', "");
Expect(1, 67999, '\P{^ismeroitichieroglyphs}', "");
Expect(0, 68000, '\p{ismeroitichieroglyphs}', "");
Expect(1, 68000, '\p{^ismeroitichieroglyphs}', "");
Expect(1, 68000, '\P{ismeroitichieroglyphs}', "");
Expect(0, 68000, '\P{^ismeroitichieroglyphs}', "");
Expect(1, 67999, '\p{_IS_Meroitic_hieroglyphs}', "");
Expect(0, 67999, '\p{^_IS_Meroitic_hieroglyphs}', "");
Expect(0, 67999, '\P{_IS_Meroitic_hieroglyphs}', "");
Expect(1, 67999, '\P{^_IS_Meroitic_hieroglyphs}', "");
Expect(0, 68000, '\p{_IS_Meroitic_hieroglyphs}', "");
Expect(1, 68000, '\p{^_IS_Meroitic_hieroglyphs}', "");
Expect(1, 68000, '\P{_IS_Meroitic_hieroglyphs}', "");
Expect(0, 68000, '\P{^_IS_Meroitic_hieroglyphs}', "");
Error('\p{:= -mero}');
Error('\P{:= -mero}');
Expect(1, 67999, '\p{mero}', "");
Expect(0, 67999, '\p{^mero}', "");
Expect(0, 67999, '\P{mero}', "");
Expect(1, 67999, '\P{^mero}', "");
Expect(0, 68000, '\p{mero}', "");
Expect(1, 68000, '\p{^mero}', "");
Expect(1, 68000, '\P{mero}', "");
Expect(0, 68000, '\P{^mero}', "");
Expect(1, 67999, '\p{ 	MERO}', "");
Expect(0, 67999, '\p{^ 	MERO}', "");
Expect(0, 67999, '\P{ 	MERO}', "");
Expect(1, 67999, '\P{^ 	MERO}', "");
Expect(0, 68000, '\p{ 	MERO}', "");
Expect(1, 68000, '\p{^ 	MERO}', "");
Expect(1, 68000, '\P{ 	MERO}', "");
Expect(0, 68000, '\P{^ 	MERO}', "");
Error('\p{- Is_MERO/a/}');
Error('\P{- Is_MERO/a/}');
Expect(1, 67999, '\p{ismero}', "");
Expect(0, 67999, '\p{^ismero}', "");
Expect(0, 67999, '\P{ismero}', "");
Expect(1, 67999, '\P{^ismero}', "");
Expect(0, 68000, '\p{ismero}', "");
Expect(1, 68000, '\p{^ismero}', "");
Expect(1, 68000, '\P{ismero}', "");
Expect(0, 68000, '\P{^ismero}', "");
Expect(1, 67999, '\p{-	is_MERO}', "");
Expect(0, 67999, '\p{^-	is_MERO}', "");
Expect(0, 67999, '\P{-	is_MERO}', "");
Expect(1, 67999, '\P{^-	is_MERO}', "");
Expect(0, 68000, '\p{-	is_MERO}', "");
Expect(1, 68000, '\p{^-	is_MERO}', "");
Expect(1, 68000, '\P{-	is_MERO}', "");
Expect(0, 68000, '\P{^-	is_MERO}', "");
Error('\p{:= LETTER}');
Error('\P{:= LETTER}');
Expect(1, 195101, '\p{letter}', "");
Expect(0, 195101, '\p{^letter}', "");
Expect(0, 195101, '\P{letter}', "");
Expect(1, 195101, '\P{^letter}', "");
Expect(0, 195102, '\p{letter}', "");
Expect(1, 195102, '\p{^letter}', "");
Expect(1, 195102, '\P{letter}', "");
Expect(0, 195102, '\P{^letter}', "");
Expect(1, 195101, '\p{	-letter}', "");
Expect(0, 195101, '\p{^	-letter}', "");
Expect(0, 195101, '\P{	-letter}', "");
Expect(1, 195101, '\P{^	-letter}', "");
Expect(0, 195102, '\p{	-letter}', "");
Expect(1, 195102, '\p{^	-letter}', "");
Expect(1, 195102, '\P{	-letter}', "");
Expect(0, 195102, '\P{^	-letter}', "");
Error('\p{/a/_	Is_LETTER}');
Error('\P{/a/_	Is_LETTER}');
Expect(1, 195101, '\p{isletter}', "");
Expect(0, 195101, '\p{^isletter}', "");
Expect(0, 195101, '\P{isletter}', "");
Expect(1, 195101, '\P{^isletter}', "");
Expect(0, 195102, '\p{isletter}', "");
Expect(1, 195102, '\p{^isletter}', "");
Expect(1, 195102, '\P{isletter}', "");
Expect(0, 195102, '\P{^isletter}', "");
Expect(1, 195101, '\p{_IS_LETTER}', "");
Expect(0, 195101, '\p{^_IS_LETTER}', "");
Expect(0, 195101, '\P{_IS_LETTER}', "");
Expect(1, 195101, '\P{^_IS_LETTER}', "");
Expect(0, 195102, '\p{_IS_LETTER}', "");
Expect(1, 195102, '\p{^_IS_LETTER}', "");
Expect(1, 195102, '\P{_IS_LETTER}', "");
Expect(0, 195102, '\P{^_IS_LETTER}', "");
Error('\p{:=-L}');
Error('\P{:=-L}');
Expect(1, 195101, '\p{l}', "");
Expect(0, 195101, '\p{^l}', "");
Expect(0, 195101, '\P{l}', "");
Expect(1, 195101, '\P{^l}', "");
Expect(0, 195102, '\p{l}', "");
Expect(1, 195102, '\p{^l}', "");
Expect(1, 195102, '\P{l}', "");
Expect(0, 195102, '\P{^l}', "");
Expect(1, 195101, '\p{_-L}', "");
Expect(0, 195101, '\p{^_-L}', "");
Expect(0, 195101, '\P{_-L}', "");
Expect(1, 195101, '\P{^_-L}', "");
Expect(0, 195102, '\p{_-L}', "");
Expect(1, 195102, '\p{^_-L}', "");
Expect(1, 195102, '\P{_-L}', "");
Expect(0, 195102, '\P{^_-L}', "");
Error('\p{:=  Is_L}');
Error('\P{:=  Is_L}');
Expect(1, 195101, '\p{isl}', "");
Expect(0, 195101, '\p{^isl}', "");
Expect(0, 195101, '\P{isl}', "");
Expect(1, 195101, '\P{^isl}', "");
Expect(0, 195102, '\p{isl}', "");
Expect(1, 195102, '\p{^isl}', "");
Expect(1, 195102, '\P{isl}', "");
Expect(0, 195102, '\P{^isl}', "");
Expect(1, 195101, '\p{-is_L}', "");
Expect(0, 195101, '\p{^-is_L}', "");
Expect(0, 195101, '\P{-is_L}', "");
Expect(1, 195101, '\P{^-is_L}', "");
Expect(0, 195102, '\p{-is_L}', "");
Expect(1, 195102, '\p{^-is_L}', "");
Expect(1, 195102, '\P{-is_L}', "");
Expect(0, 195102, '\P{^-is_L}', "");
Error('\p{:=cyrillic_extended_B}');
Error('\P{:=cyrillic_extended_B}');
Expect(1, 42655, '\p{cyrillicextendedb}', "");
Expect(0, 42655, '\p{^cyrillicextendedb}', "");
Expect(0, 42655, '\P{cyrillicextendedb}', "");
Expect(1, 42655, '\P{^cyrillicextendedb}', "");
Expect(0, 42656, '\p{cyrillicextendedb}', "");
Expect(1, 42656, '\p{^cyrillicextendedb}', "");
Expect(1, 42656, '\P{cyrillicextendedb}', "");
Expect(0, 42656, '\P{^cyrillicextendedb}', "");
Expect(1, 42655, '\p{ _Cyrillic_EXTENDED_B}', "");
Expect(0, 42655, '\p{^ _Cyrillic_EXTENDED_B}', "");
Expect(0, 42655, '\P{ _Cyrillic_EXTENDED_B}', "");
Expect(1, 42655, '\P{^ _Cyrillic_EXTENDED_B}', "");
Expect(0, 42656, '\p{ _Cyrillic_EXTENDED_B}', "");
Expect(1, 42656, '\p{^ _Cyrillic_EXTENDED_B}', "");
Expect(1, 42656, '\P{ _Cyrillic_EXTENDED_B}', "");
Expect(0, 42656, '\P{^ _Cyrillic_EXTENDED_B}', "");
Error('\p{ /a/Is_Cyrillic_EXTENDED_B}');
Error('\P{ /a/Is_Cyrillic_EXTENDED_B}');
Expect(1, 42655, '\p{iscyrillicextendedb}', "");
Expect(0, 42655, '\p{^iscyrillicextendedb}', "");
Expect(0, 42655, '\P{iscyrillicextendedb}', "");
Expect(1, 42655, '\P{^iscyrillicextendedb}', "");
Expect(0, 42656, '\p{iscyrillicextendedb}', "");
Expect(1, 42656, '\p{^iscyrillicextendedb}', "");
Expect(1, 42656, '\P{iscyrillicextendedb}', "");
Expect(0, 42656, '\P{^iscyrillicextendedb}', "");
Expect(1, 42655, '\p{	is_cyrillic_Extended_b}', "");
Expect(0, 42655, '\p{^	is_cyrillic_Extended_b}', "");
Expect(0, 42655, '\P{	is_cyrillic_Extended_b}', "");
Expect(1, 42655, '\P{^	is_cyrillic_Extended_b}', "");
Expect(0, 42656, '\p{	is_cyrillic_Extended_b}', "");
Expect(1, 42656, '\p{^	is_cyrillic_Extended_b}', "");
Expect(1, 42656, '\P{	is_cyrillic_Extended_b}', "");
Expect(0, 42656, '\P{^	is_cyrillic_Extended_b}', "");
Error('\p{ in_Cyrillic_Extended_b:=}');
Error('\P{ in_Cyrillic_Extended_b:=}');
Expect(1, 42655, '\p{incyrillicextendedb}', "");
Expect(0, 42655, '\p{^incyrillicextendedb}', "");
Expect(0, 42655, '\P{incyrillicextendedb}', "");
Expect(1, 42655, '\P{^incyrillicextendedb}', "");
Expect(0, 42656, '\p{incyrillicextendedb}', "");
Expect(1, 42656, '\p{^incyrillicextendedb}', "");
Expect(1, 42656, '\P{incyrillicextendedb}', "");
Expect(0, 42656, '\P{^incyrillicextendedb}', "");
Expect(1, 42655, '\p{ _in_CYRILLIC_Extended_B}', "");
Expect(0, 42655, '\p{^ _in_CYRILLIC_Extended_B}', "");
Expect(0, 42655, '\P{ _in_CYRILLIC_Extended_B}', "");
Expect(1, 42655, '\P{^ _in_CYRILLIC_Extended_B}', "");
Expect(0, 42656, '\p{ _in_CYRILLIC_Extended_B}', "");
Expect(1, 42656, '\p{^ _in_CYRILLIC_Extended_B}', "");
Expect(1, 42656, '\P{ _in_CYRILLIC_Extended_B}', "");
Expect(0, 42656, '\P{^ _in_CYRILLIC_Extended_B}', "");
Error('\p{	/a/Cyrillic_Ext_b}');
Error('\P{	/a/Cyrillic_Ext_b}');
Expect(1, 42655, '\p{cyrillicextb}', "");
Expect(0, 42655, '\p{^cyrillicextb}', "");
Expect(0, 42655, '\P{cyrillicextb}', "");
Expect(1, 42655, '\P{^cyrillicextb}', "");
Expect(0, 42656, '\p{cyrillicextb}', "");
Expect(1, 42656, '\p{^cyrillicextb}', "");
Expect(1, 42656, '\P{cyrillicextb}', "");
Expect(0, 42656, '\P{^cyrillicextb}', "");
Expect(1, 42655, '\p{_cyrillic_ext_B}', "");
Expect(0, 42655, '\p{^_cyrillic_ext_B}', "");
Expect(0, 42655, '\P{_cyrillic_ext_B}', "");
Expect(1, 42655, '\P{^_cyrillic_ext_B}', "");
Expect(0, 42656, '\p{_cyrillic_ext_B}', "");
Expect(1, 42656, '\p{^_cyrillic_ext_B}', "");
Expect(1, 42656, '\P{_cyrillic_ext_B}', "");
Expect(0, 42656, '\P{^_cyrillic_ext_B}', "");
Error('\p{:=-	IS_cyrillic_Ext_B}');
Error('\P{:=-	IS_cyrillic_Ext_B}');
Expect(1, 42655, '\p{iscyrillicextb}', "");
Expect(0, 42655, '\p{^iscyrillicextb}', "");
Expect(0, 42655, '\P{iscyrillicextb}', "");
Expect(1, 42655, '\P{^iscyrillicextb}', "");
Expect(0, 42656, '\p{iscyrillicextb}', "");
Expect(1, 42656, '\p{^iscyrillicextb}', "");
Expect(1, 42656, '\P{iscyrillicextb}', "");
Expect(0, 42656, '\P{^iscyrillicextb}', "");
Expect(1, 42655, '\p{_ Is_Cyrillic_Ext_b}', "");
Expect(0, 42655, '\p{^_ Is_Cyrillic_Ext_b}', "");
Expect(0, 42655, '\P{_ Is_Cyrillic_Ext_b}', "");
Expect(1, 42655, '\P{^_ Is_Cyrillic_Ext_b}', "");
Expect(0, 42656, '\p{_ Is_Cyrillic_Ext_b}', "");
Expect(1, 42656, '\p{^_ Is_Cyrillic_Ext_b}', "");
Expect(1, 42656, '\P{_ Is_Cyrillic_Ext_b}', "");
Expect(0, 42656, '\P{^_ Is_Cyrillic_Ext_b}', "");
Error('\p{/a/	In_cyrillic_Ext_B}');
Error('\P{/a/	In_cyrillic_Ext_B}');
Expect(1, 42655, '\p{incyrillicextb}', "");
Expect(0, 42655, '\p{^incyrillicextb}', "");
Expect(0, 42655, '\P{incyrillicextb}', "");
Expect(1, 42655, '\P{^incyrillicextb}', "");
Expect(0, 42656, '\p{incyrillicextb}', "");
Expect(1, 42656, '\p{^incyrillicextb}', "");
Expect(1, 42656, '\P{incyrillicextb}', "");
Expect(0, 42656, '\P{^incyrillicextb}', "");
Expect(1, 42655, '\p{-_In_Cyrillic_Ext_B}', "");
Expect(0, 42655, '\p{^-_In_Cyrillic_Ext_B}', "");
Expect(0, 42655, '\P{-_In_Cyrillic_Ext_B}', "");
Expect(1, 42655, '\P{^-_In_Cyrillic_Ext_B}', "");
Expect(0, 42656, '\p{-_In_Cyrillic_Ext_B}', "");
Expect(1, 42656, '\p{^-_In_Cyrillic_Ext_B}', "");
Expect(1, 42656, '\P{-_In_Cyrillic_Ext_B}', "");
Expect(0, 42656, '\P{^-_In_Cyrillic_Ext_B}', "");
Error('\p{:=Greek_Extended}');
Error('\P{:=Greek_Extended}');
Expect(1, 8191, '\p{greekextended}', "");
Expect(0, 8191, '\p{^greekextended}', "");
Expect(0, 8191, '\P{greekextended}', "");
Expect(1, 8191, '\P{^greekextended}', "");
Expect(0, 8192, '\p{greekextended}', "");
Expect(1, 8192, '\p{^greekextended}', "");
Expect(1, 8192, '\P{greekextended}', "");
Expect(0, 8192, '\P{^greekextended}', "");
Expect(1, 8191, '\p{  greek_EXTENDED}', "");
Expect(0, 8191, '\p{^  greek_EXTENDED}', "");
Expect(0, 8191, '\P{  greek_EXTENDED}', "");
Expect(1, 8191, '\P{^  greek_EXTENDED}', "");
Expect(0, 8192, '\p{  greek_EXTENDED}', "");
Expect(1, 8192, '\p{^  greek_EXTENDED}', "");
Expect(1, 8192, '\P{  greek_EXTENDED}', "");
Expect(0, 8192, '\P{^  greek_EXTENDED}', "");
Error('\p{ /a/Is_greek_Extended}');
Error('\P{ /a/Is_greek_Extended}');
Expect(1, 8191, '\p{isgreekextended}', "");
Expect(0, 8191, '\p{^isgreekextended}', "");
Expect(0, 8191, '\P{isgreekextended}', "");
Expect(1, 8191, '\P{^isgreekextended}', "");
Expect(0, 8192, '\p{isgreekextended}', "");
Expect(1, 8192, '\p{^isgreekextended}', "");
Expect(1, 8192, '\P{isgreekextended}', "");
Expect(0, 8192, '\P{^isgreekextended}', "");
Expect(1, 8191, '\p{  is_Greek_extended}', "");
Expect(0, 8191, '\p{^  is_Greek_extended}', "");
Expect(0, 8191, '\P{  is_Greek_extended}', "");
Expect(1, 8191, '\P{^  is_Greek_extended}', "");
Expect(0, 8192, '\p{  is_Greek_extended}', "");
Expect(1, 8192, '\p{^  is_Greek_extended}', "");
Expect(1, 8192, '\P{  is_Greek_extended}', "");
Expect(0, 8192, '\P{^  is_Greek_extended}', "");
Error('\p{/a/		IN_greek_EXTENDED}');
Error('\P{/a/		IN_greek_EXTENDED}');
Expect(1, 8191, '\p{ingreekextended}', "");
Expect(0, 8191, '\p{^ingreekextended}', "");
Expect(0, 8191, '\P{ingreekextended}', "");
Expect(1, 8191, '\P{^ingreekextended}', "");
Expect(0, 8192, '\p{ingreekextended}', "");
Expect(1, 8192, '\p{^ingreekextended}', "");
Expect(1, 8192, '\P{ingreekextended}', "");
Expect(0, 8192, '\P{^ingreekextended}', "");
Expect(1, 8191, '\p{	In_GREEK_extended}', "");
Expect(0, 8191, '\p{^	In_GREEK_extended}', "");
Expect(0, 8191, '\P{	In_GREEK_extended}', "");
Expect(1, 8191, '\P{^	In_GREEK_extended}', "");
Expect(0, 8192, '\p{	In_GREEK_extended}', "");
Expect(1, 8192, '\p{^	In_GREEK_extended}', "");
Expect(1, 8192, '\P{	In_GREEK_extended}', "");
Expect(0, 8192, '\P{^	In_GREEK_extended}', "");
Error('\p{:=- greek_EXT}');
Error('\P{:=- greek_EXT}');
Expect(1, 8191, '\p{greekext}', "");
Expect(0, 8191, '\p{^greekext}', "");
Expect(0, 8191, '\P{greekext}', "");
Expect(1, 8191, '\P{^greekext}', "");
Expect(0, 8192, '\p{greekext}', "");
Expect(1, 8192, '\p{^greekext}', "");
Expect(1, 8192, '\P{greekext}', "");
Expect(0, 8192, '\P{^greekext}', "");
Expect(1, 8191, '\p{ Greek_Ext}', "");
Expect(0, 8191, '\p{^ Greek_Ext}', "");
Expect(0, 8191, '\P{ Greek_Ext}', "");
Expect(1, 8191, '\P{^ Greek_Ext}', "");
Expect(0, 8192, '\p{ Greek_Ext}', "");
Expect(1, 8192, '\p{^ Greek_Ext}', "");
Expect(1, 8192, '\P{ Greek_Ext}', "");
Expect(0, 8192, '\P{^ Greek_Ext}', "");
Error('\p{-:=is_greek_Ext}');
Error('\P{-:=is_greek_Ext}');
Expect(1, 8191, '\p{isgreekext}', "");
Expect(0, 8191, '\p{^isgreekext}', "");
Expect(0, 8191, '\P{isgreekext}', "");
Expect(1, 8191, '\P{^isgreekext}', "");
Expect(0, 8192, '\p{isgreekext}', "");
Expect(1, 8192, '\p{^isgreekext}', "");
Expect(1, 8192, '\P{isgreekext}', "");
Expect(0, 8192, '\P{^isgreekext}', "");
Expect(1, 8191, '\p{--Is_Greek_ext}', "");
Expect(0, 8191, '\p{^--Is_Greek_ext}', "");
Expect(0, 8191, '\P{--Is_Greek_ext}', "");
Expect(1, 8191, '\P{^--Is_Greek_ext}', "");
Expect(0, 8192, '\p{--Is_Greek_ext}', "");
Expect(1, 8192, '\p{^--Is_Greek_ext}', "");
Expect(1, 8192, '\P{--Is_Greek_ext}', "");
Expect(0, 8192, '\P{^--Is_Greek_ext}', "");
Error('\p{  in_Greek_Ext/a/}');
Error('\P{  in_Greek_Ext/a/}');
Expect(1, 8191, '\p{ingreekext}', "");
Expect(0, 8191, '\p{^ingreekext}', "");
Expect(0, 8191, '\P{ingreekext}', "");
Expect(1, 8191, '\P{^ingreekext}', "");
Expect(0, 8192, '\p{ingreekext}', "");
Expect(1, 8192, '\p{^ingreekext}', "");
Expect(1, 8192, '\P{ingreekext}', "");
Expect(0, 8192, '\P{^ingreekext}', "");
Expect(1, 8191, '\p{--In_Greek_ext}', "");
Expect(0, 8191, '\p{^--In_Greek_ext}', "");
Expect(0, 8191, '\P{--In_Greek_ext}', "");
Expect(1, 8191, '\P{^--In_Greek_ext}', "");
Expect(0, 8192, '\p{--In_Greek_ext}', "");
Expect(1, 8192, '\p{^--In_Greek_ext}', "");
Expect(1, 8192, '\P{--In_Greek_ext}', "");
Expect(0, 8192, '\P{^--In_Greek_ext}', "");
Error('\p{ :=Emoticons}');
Error('\P{ :=Emoticons}');
Expect(1, 128591, '\p{emoticons}', "");
Expect(0, 128591, '\p{^emoticons}', "");
Expect(0, 128591, '\P{emoticons}', "");
Expect(1, 128591, '\P{^emoticons}', "");
Expect(0, 128592, '\p{emoticons}', "");
Expect(1, 128592, '\p{^emoticons}', "");
Expect(1, 128592, '\P{emoticons}', "");
Expect(0, 128592, '\P{^emoticons}', "");
Expect(1, 128591, '\p{	Emoticons}', "");
Expect(0, 128591, '\p{^	Emoticons}', "");
Expect(0, 128591, '\P{	Emoticons}', "");
Expect(1, 128591, '\P{^	Emoticons}', "");
Expect(0, 128592, '\p{	Emoticons}', "");
Expect(1, 128592, '\p{^	Emoticons}', "");
Expect(1, 128592, '\P{	Emoticons}', "");
Expect(0, 128592, '\P{^	Emoticons}', "");
Error('\p{ IS_emoticons:=}');
Error('\P{ IS_emoticons:=}');
Expect(1, 128591, '\p{isemoticons}', "");
Expect(0, 128591, '\p{^isemoticons}', "");
Expect(0, 128591, '\P{isemoticons}', "");
Expect(1, 128591, '\P{^isemoticons}', "");
Expect(0, 128592, '\p{isemoticons}', "");
Expect(1, 128592, '\p{^isemoticons}', "");
Expect(1, 128592, '\P{isemoticons}', "");
Expect(0, 128592, '\P{^isemoticons}', "");
Expect(1, 128591, '\p{ Is_emoticons}', "");
Expect(0, 128591, '\p{^ Is_emoticons}', "");
Expect(0, 128591, '\P{ Is_emoticons}', "");
Expect(1, 128591, '\P{^ Is_emoticons}', "");
Expect(0, 128592, '\p{ Is_emoticons}', "");
Expect(1, 128592, '\p{^ Is_emoticons}', "");
Expect(1, 128592, '\P{ Is_emoticons}', "");
Expect(0, 128592, '\P{^ Is_emoticons}', "");
Error('\p{/a/in_Emoticons}');
Error('\P{/a/in_Emoticons}');
Expect(1, 128591, '\p{inemoticons}', "");
Expect(0, 128591, '\p{^inemoticons}', "");
Expect(0, 128591, '\P{inemoticons}', "");
Expect(1, 128591, '\P{^inemoticons}', "");
Expect(0, 128592, '\p{inemoticons}', "");
Expect(1, 128592, '\p{^inemoticons}', "");
Expect(1, 128592, '\P{inemoticons}', "");
Expect(0, 128592, '\P{^inemoticons}', "");
Expect(1, 128591, '\p{__In_emoticons}', "");
Expect(0, 128591, '\p{^__In_emoticons}', "");
Expect(0, 128591, '\P{__In_emoticons}', "");
Expect(1, 128591, '\P{^__In_emoticons}', "");
Expect(0, 128592, '\p{__In_emoticons}', "");
Expect(1, 128592, '\p{^__In_emoticons}', "");
Expect(1, 128592, '\P{__In_emoticons}', "");
Expect(0, 128592, '\P{^__In_emoticons}', "");
Error('\p{	/a/HEBREW}');
Error('\P{	/a/HEBREW}');
Expect(1, 64335, '\p{hebrew}', "");
Expect(0, 64335, '\p{^hebrew}', "");
Expect(0, 64335, '\P{hebrew}', "");
Expect(1, 64335, '\P{^hebrew}', "");
Expect(0, 64336, '\p{hebrew}', "");
Expect(1, 64336, '\p{^hebrew}', "");
Expect(1, 64336, '\P{hebrew}', "");
Expect(0, 64336, '\P{^hebrew}', "");
Expect(1, 64335, '\p{ _HEBREW}', "");
Expect(0, 64335, '\p{^ _HEBREW}', "");
Expect(0, 64335, '\P{ _HEBREW}', "");
Expect(1, 64335, '\P{^ _HEBREW}', "");
Expect(0, 64336, '\p{ _HEBREW}', "");
Expect(1, 64336, '\p{^ _HEBREW}', "");
Expect(1, 64336, '\P{ _HEBREW}', "");
Expect(0, 64336, '\P{^ _HEBREW}', "");
Error('\p{-:=is_Hebrew}');
Error('\P{-:=is_Hebrew}');
Expect(1, 64335, '\p{ishebrew}', "");
Expect(0, 64335, '\p{^ishebrew}', "");
Expect(0, 64335, '\P{ishebrew}', "");
Expect(1, 64335, '\P{^ishebrew}', "");
Expect(0, 64336, '\p{ishebrew}', "");
Expect(1, 64336, '\p{^ishebrew}', "");
Expect(1, 64336, '\P{ishebrew}', "");
Expect(0, 64336, '\P{^ishebrew}', "");
Expect(1, 64335, '\p{-IS_hebrew}', "");
Expect(0, 64335, '\p{^-IS_hebrew}', "");
Expect(0, 64335, '\P{-IS_hebrew}', "");
Expect(1, 64335, '\P{^-IS_hebrew}', "");
Expect(0, 64336, '\p{-IS_hebrew}', "");
Expect(1, 64336, '\p{^-IS_hebrew}', "");
Expect(1, 64336, '\P{-IS_hebrew}', "");
Expect(0, 64336, '\P{^-IS_hebrew}', "");
Error('\p{/a/ _Hebr}');
Error('\P{/a/ _Hebr}');
Expect(1, 64335, '\p{hebr}', "");
Expect(0, 64335, '\p{^hebr}', "");
Expect(0, 64335, '\P{hebr}', "");
Expect(1, 64335, '\P{^hebr}', "");
Expect(0, 64336, '\p{hebr}', "");
Expect(1, 64336, '\p{^hebr}', "");
Expect(1, 64336, '\P{hebr}', "");
Expect(0, 64336, '\P{^hebr}', "");
Expect(1, 64335, '\p{ Hebr}', "");
Expect(0, 64335, '\p{^ Hebr}', "");
Expect(0, 64335, '\P{ Hebr}', "");
Expect(1, 64335, '\P{^ Hebr}', "");
Expect(0, 64336, '\p{ Hebr}', "");
Expect(1, 64336, '\p{^ Hebr}', "");
Expect(1, 64336, '\P{ Hebr}', "");
Expect(0, 64336, '\P{^ Hebr}', "");
Error('\p{_/a/Is_HEBR}');
Error('\P{_/a/Is_HEBR}');
Expect(1, 64335, '\p{ishebr}', "");
Expect(0, 64335, '\p{^ishebr}', "");
Expect(0, 64335, '\P{ishebr}', "");
Expect(1, 64335, '\P{^ishebr}', "");
Expect(0, 64336, '\p{ishebr}', "");
Expect(1, 64336, '\p{^ishebr}', "");
Expect(1, 64336, '\P{ishebr}', "");
Expect(0, 64336, '\P{^ishebr}', "");
Expect(1, 64335, '\p{-IS_HEBR}', "");
Expect(0, 64335, '\p{^-IS_HEBR}', "");
Expect(0, 64335, '\P{-IS_HEBR}', "");
Expect(1, 64335, '\P{^-IS_HEBR}', "");
Expect(0, 64336, '\p{-IS_HEBR}', "");
Expect(1, 64336, '\p{^-IS_HEBR}', "");
Expect(1, 64336, '\P{-IS_HEBR}', "");
Expect(0, 64336, '\P{^-IS_HEBR}', "");
Error('\p{/a/-_Titlecase}');
Error('\P{/a/-_Titlecase}');
Expect(1, 8188, '\p{titlecase}', "");
Expect(0, 8188, '\p{^titlecase}', "");
Expect(0, 8188, '\P{titlecase}', "");
Expect(1, 8188, '\P{^titlecase}', "");
Expect(0, 8189, '\p{titlecase}', "");
Expect(1, 8189, '\p{^titlecase}', "");
Expect(1, 8189, '\P{titlecase}', "");
Expect(0, 8189, '\P{^titlecase}', "");
Expect(1, 8188, '\p{  Titlecase}', "");
Expect(0, 8188, '\p{^  Titlecase}', "");
Expect(0, 8188, '\P{  Titlecase}', "");
Expect(1, 8188, '\P{^  Titlecase}', "");
Expect(0, 8189, '\p{  Titlecase}', "");
Expect(1, 8189, '\p{^  Titlecase}', "");
Expect(1, 8189, '\P{  Titlecase}', "");
Expect(0, 8189, '\P{^  Titlecase}', "");
Error('\p{	/a/Title}');
Error('\P{	/a/Title}');
Expect(1, 8188, '\p{title}', "");
Expect(0, 8188, '\p{^title}', "");
Expect(0, 8188, '\P{title}', "");
Expect(1, 8188, '\P{^title}', "");
Expect(0, 8189, '\p{title}', "");
Expect(1, 8189, '\p{^title}', "");
Expect(1, 8189, '\P{title}', "");
Expect(0, 8189, '\P{^title}', "");
Expect(1, 8188, '\p{ 	title}', "");
Expect(0, 8188, '\p{^ 	title}', "");
Expect(0, 8188, '\P{ 	title}', "");
Expect(1, 8188, '\P{^ 	title}', "");
Expect(0, 8189, '\p{ 	title}', "");
Expect(1, 8189, '\p{^ 	title}', "");
Expect(1, 8189, '\P{ 	title}', "");
Expect(0, 8189, '\P{^ 	title}', "");
Error('\p{/a/is_Titlecase}');
Error('\P{/a/is_Titlecase}');
Expect(1, 8188, '\p{istitlecase}', "");
Expect(0, 8188, '\p{^istitlecase}', "");
Expect(0, 8188, '\P{istitlecase}', "");
Expect(1, 8188, '\P{^istitlecase}', "");
Expect(0, 8189, '\p{istitlecase}', "");
Expect(1, 8189, '\p{^istitlecase}', "");
Expect(1, 8189, '\P{istitlecase}', "");
Expect(0, 8189, '\P{^istitlecase}', "");
Expect(1, 8188, '\p{ -is_TITLECASE}', "");
Expect(0, 8188, '\p{^ -is_TITLECASE}', "");
Expect(0, 8188, '\P{ -is_TITLECASE}', "");
Expect(1, 8188, '\P{^ -is_TITLECASE}', "");
Expect(0, 8189, '\p{ -is_TITLECASE}', "");
Expect(1, 8189, '\p{^ -is_TITLECASE}', "");
Expect(1, 8189, '\P{ -is_TITLECASE}', "");
Expect(0, 8189, '\P{^ -is_TITLECASE}', "");
Error('\p{/a/-IS_title}');
Error('\P{/a/-IS_title}');
Expect(1, 8188, '\p{istitle}', "");
Expect(0, 8188, '\p{^istitle}', "");
Expect(0, 8188, '\P{istitle}', "");
Expect(1, 8188, '\P{^istitle}', "");
Expect(0, 8189, '\p{istitle}', "");
Expect(1, 8189, '\p{^istitle}', "");
Expect(1, 8189, '\P{istitle}', "");
Expect(0, 8189, '\P{^istitle}', "");
Expect(1, 8188, '\p{-Is_Title}', "");
Expect(0, 8188, '\p{^-Is_Title}', "");
Expect(0, 8188, '\P{-Is_Title}', "");
Expect(1, 8188, '\P{^-Is_Title}', "");
Expect(0, 8189, '\p{-Is_Title}', "");
Expect(1, 8189, '\p{^-Is_Title}', "");
Expect(1, 8189, '\P{-Is_Title}', "");
Expect(0, 8189, '\P{^-Is_Title}', "");
Error('\p{-:=IN_Hiragana}');
Error('\P{-:=IN_Hiragana}');
Expect(1, 12447, '\p{inhiragana}', "");
Expect(0, 12447, '\p{^inhiragana}', "");
Expect(0, 12447, '\P{inhiragana}', "");
Expect(1, 12447, '\P{^inhiragana}', "");
Expect(0, 12448, '\p{inhiragana}', "");
Expect(1, 12448, '\p{^inhiragana}', "");
Expect(1, 12448, '\P{inhiragana}', "");
Expect(0, 12448, '\P{^inhiragana}', "");
Expect(1, 12447, '\p{_IN_hiragana}', "");
Expect(0, 12447, '\p{^_IN_hiragana}', "");
Expect(0, 12447, '\P{_IN_hiragana}', "");
Expect(1, 12447, '\P{^_IN_hiragana}', "");
Expect(0, 12448, '\p{_IN_hiragana}', "");
Expect(1, 12448, '\p{^_IN_hiragana}', "");
Expect(1, 12448, '\P{_IN_hiragana}', "");
Expect(0, 12448, '\P{^_IN_hiragana}', "");
Error('\p{:=	In_HIRAGANA}');
Error('\P{:=	In_HIRAGANA}');
Expect(1, 12447, '\p{-_in_hiragana}', "");
Expect(0, 12447, '\p{^-_in_hiragana}', "");
Expect(0, 12447, '\P{-_in_hiragana}', "");
Expect(1, 12447, '\P{^-_in_hiragana}', "");
Expect(0, 12448, '\p{-_in_hiragana}', "");
Expect(1, 12448, '\p{^-_in_hiragana}', "");
Expect(1, 12448, '\P{-_in_hiragana}', "");
Expect(0, 12448, '\P{^-_in_hiragana}', "");
Error('\p{/a/	_latin}');
Error('\P{/a/	_latin}');
Expect(1, 65370, '\p{latin}', "");
Expect(0, 65370, '\p{^latin}', "");
Expect(0, 65370, '\P{latin}', "");
Expect(1, 65370, '\P{^latin}', "");
Expect(0, 65371, '\p{latin}', "");
Expect(1, 65371, '\p{^latin}', "");
Expect(1, 65371, '\P{latin}', "");
Expect(0, 65371, '\P{^latin}', "");
Expect(1, 65370, '\p{-Latin}', "");
Expect(0, 65370, '\p{^-Latin}', "");
Expect(0, 65370, '\P{-Latin}', "");
Expect(1, 65370, '\P{^-Latin}', "");
Expect(0, 65371, '\p{-Latin}', "");
Expect(1, 65371, '\p{^-Latin}', "");
Expect(1, 65371, '\P{-Latin}', "");
Expect(0, 65371, '\P{^-Latin}', "");
Error('\p{:=_-IS_Latin}');
Error('\P{:=_-IS_Latin}');
Expect(1, 65370, '\p{islatin}', "");
Expect(0, 65370, '\p{^islatin}', "");
Expect(0, 65370, '\P{islatin}', "");
Expect(1, 65370, '\P{^islatin}', "");
Expect(0, 65371, '\p{islatin}', "");
Expect(1, 65371, '\p{^islatin}', "");
Expect(1, 65371, '\P{islatin}', "");
Expect(0, 65371, '\P{^islatin}', "");
Expect(1, 65370, '\p{_ is_LATIN}', "");
Expect(0, 65370, '\p{^_ is_LATIN}', "");
Expect(0, 65370, '\P{_ is_LATIN}', "");
Expect(1, 65370, '\P{^_ is_LATIN}', "");
Expect(0, 65371, '\p{_ is_LATIN}', "");
Expect(1, 65371, '\p{^_ is_LATIN}', "");
Expect(1, 65371, '\P{_ is_LATIN}', "");
Expect(0, 65371, '\P{^_ is_LATIN}', "");
Error('\p{--latn:=}');
Error('\P{--latn:=}');
Expect(1, 65370, '\p{latn}', "");
Expect(0, 65370, '\p{^latn}', "");
Expect(0, 65370, '\P{latn}', "");
Expect(1, 65370, '\P{^latn}', "");
Expect(0, 65371, '\p{latn}', "");
Expect(1, 65371, '\p{^latn}', "");
Expect(1, 65371, '\P{latn}', "");
Expect(0, 65371, '\P{^latn}', "");
Expect(1, 65370, '\p{	-LATN}', "");
Expect(0, 65370, '\p{^	-LATN}', "");
Expect(0, 65370, '\P{	-LATN}', "");
Expect(1, 65370, '\P{^	-LATN}', "");
Expect(0, 65371, '\p{	-LATN}', "");
Expect(1, 65371, '\p{^	-LATN}', "");
Expect(1, 65371, '\P{	-LATN}', "");
Expect(0, 65371, '\P{^	-LATN}', "");
Error('\p{_Is_LATN:=}');
Error('\P{_Is_LATN:=}');
Expect(1, 65370, '\p{islatn}', "");
Expect(0, 65370, '\p{^islatn}', "");
Expect(0, 65370, '\P{islatn}', "");
Expect(1, 65370, '\P{^islatn}', "");
Expect(0, 65371, '\p{islatn}', "");
Expect(1, 65371, '\p{^islatn}', "");
Expect(1, 65371, '\P{islatn}', "");
Expect(0, 65371, '\P{^islatn}', "");
Expect(1, 65370, '\p{ 	is_Latn}', "");
Expect(0, 65370, '\p{^ 	is_Latn}', "");
Expect(0, 65370, '\P{ 	is_Latn}', "");
Expect(1, 65370, '\P{^ 	is_Latn}', "");
Expect(0, 65371, '\p{ 	is_Latn}', "");
Expect(1, 65371, '\p{^ 	is_Latn}', "");
Expect(1, 65371, '\P{ 	is_Latn}', "");
Expect(0, 65371, '\P{^ 	is_Latn}', "");
Error('\p{/a/ 	enclosing_Mark}');
Error('\P{/a/ 	enclosing_Mark}');
Expect(1, 42610, '\p{enclosingmark}', "");
Expect(0, 42610, '\p{^enclosingmark}', "");
Expect(0, 42610, '\P{enclosingmark}', "");
Expect(1, 42610, '\P{^enclosingmark}', "");
Expect(0, 42611, '\p{enclosingmark}', "");
Expect(1, 42611, '\p{^enclosingmark}', "");
Expect(1, 42611, '\P{enclosingmark}', "");
Expect(0, 42611, '\P{^enclosingmark}', "");
Expect(1, 42610, '\p{--Enclosing_MARK}', "");
Expect(0, 42610, '\p{^--Enclosing_MARK}', "");
Expect(0, 42610, '\P{--Enclosing_MARK}', "");
Expect(1, 42610, '\P{^--Enclosing_MARK}', "");
Expect(0, 42611, '\p{--Enclosing_MARK}', "");
Expect(1, 42611, '\p{^--Enclosing_MARK}', "");
Expect(1, 42611, '\P{--Enclosing_MARK}', "");
Expect(0, 42611, '\P{^--Enclosing_MARK}', "");
Error('\p{_Is_Enclosing_Mark:=}');
Error('\P{_Is_Enclosing_Mark:=}');
Expect(1, 42610, '\p{isenclosingmark}', "");
Expect(0, 42610, '\p{^isenclosingmark}', "");
Expect(0, 42610, '\P{isenclosingmark}', "");
Expect(1, 42610, '\P{^isenclosingmark}', "");
Expect(0, 42611, '\p{isenclosingmark}', "");
Expect(1, 42611, '\p{^isenclosingmark}', "");
Expect(1, 42611, '\P{isenclosingmark}', "");
Expect(0, 42611, '\P{^isenclosingmark}', "");
Expect(1, 42610, '\p{--Is_enclosing_Mark}', "");
Expect(0, 42610, '\p{^--Is_enclosing_Mark}', "");
Expect(0, 42610, '\P{--Is_enclosing_Mark}', "");
Expect(1, 42610, '\P{^--Is_enclosing_Mark}', "");
Expect(0, 42611, '\p{--Is_enclosing_Mark}', "");
Expect(1, 42611, '\p{^--Is_enclosing_Mark}', "");
Expect(1, 42611, '\P{--Is_enclosing_Mark}', "");
Expect(0, 42611, '\P{^--Is_enclosing_Mark}', "");
Error('\p{/a/	 Me}');
Error('\P{/a/	 Me}');
Expect(1, 42610, '\p{me}', "");
Expect(0, 42610, '\p{^me}', "");
Expect(0, 42610, '\P{me}', "");
Expect(1, 42610, '\P{^me}', "");
Expect(0, 42611, '\p{me}', "");
Expect(1, 42611, '\p{^me}', "");
Expect(1, 42611, '\P{me}', "");
Expect(0, 42611, '\P{^me}', "");
Expect(1, 42610, '\p{-Me}', "");
Expect(0, 42610, '\p{^-Me}', "");
Expect(0, 42610, '\P{-Me}', "");
Expect(1, 42610, '\P{^-Me}', "");
Expect(0, 42611, '\p{-Me}', "");
Expect(1, 42611, '\p{^-Me}', "");
Expect(1, 42611, '\P{-Me}', "");
Expect(0, 42611, '\P{^-Me}', "");
Error('\p{	:=Is_Me}');
Error('\P{	:=Is_Me}');
Expect(1, 42610, '\p{isme}', "");
Expect(0, 42610, '\p{^isme}', "");
Expect(0, 42610, '\P{isme}', "");
Expect(1, 42610, '\P{^isme}', "");
Expect(0, 42611, '\p{isme}', "");
Expect(1, 42611, '\p{^isme}', "");
Expect(1, 42611, '\P{isme}', "");
Expect(0, 42611, '\P{^isme}', "");
Expect(1, 42610, '\p{-	IS_Me}', "");
Expect(0, 42610, '\p{^-	IS_Me}', "");
Expect(0, 42610, '\P{-	IS_Me}', "");
Expect(1, 42610, '\P{^-	IS_Me}', "");
Expect(0, 42611, '\p{-	IS_Me}', "");
Expect(1, 42611, '\p{^-	IS_Me}', "");
Expect(1, 42611, '\P{-	IS_Me}', "");
Expect(0, 42611, '\P{^-	IS_Me}', "");
Error('\p{_:=Format}');
Error('\P{_:=Format}');
Expect(1, 917631, '\p{format}', "");
Expect(0, 917631, '\p{^format}', "");
Expect(0, 917631, '\P{format}', "");
Expect(1, 917631, '\P{^format}', "");
Expect(0, 917632, '\p{format}', "");
Expect(1, 917632, '\p{^format}', "");
Expect(1, 917632, '\P{format}', "");
Expect(0, 917632, '\P{^format}', "");
Expect(1, 917631, '\p{ 	Format}', "");
Expect(0, 917631, '\p{^ 	Format}', "");
Expect(0, 917631, '\P{ 	Format}', "");
Expect(1, 917631, '\P{^ 	Format}', "");
Expect(0, 917632, '\p{ 	Format}', "");
Expect(1, 917632, '\p{^ 	Format}', "");
Expect(1, 917632, '\P{ 	Format}', "");
Expect(0, 917632, '\P{^ 	Format}', "");
Error('\p{	Is_Format/a/}');
Error('\P{	Is_Format/a/}');
Expect(1, 917631, '\p{isformat}', "");
Expect(0, 917631, '\p{^isformat}', "");
Expect(0, 917631, '\P{isformat}', "");
Expect(1, 917631, '\P{^isformat}', "");
Expect(0, 917632, '\p{isformat}', "");
Expect(1, 917632, '\p{^isformat}', "");
Expect(1, 917632, '\P{isformat}', "");
Expect(0, 917632, '\P{^isformat}', "");
Expect(1, 917631, '\p{ Is_Format}', "");
Expect(0, 917631, '\p{^ Is_Format}', "");
Expect(0, 917631, '\P{ Is_Format}', "");
Expect(1, 917631, '\P{^ Is_Format}', "");
Expect(0, 917632, '\p{ Is_Format}', "");
Expect(1, 917632, '\p{^ Is_Format}', "");
Expect(1, 917632, '\P{ Is_Format}', "");
Expect(0, 917632, '\P{^ Is_Format}', "");
Error('\p{ cf/a/}');
Error('\P{ cf/a/}');
Expect(1, 917631, '\p{cf}', "");
Expect(0, 917631, '\p{^cf}', "");
Expect(0, 917631, '\P{cf}', "");
Expect(1, 917631, '\P{^cf}', "");
Expect(0, 917632, '\p{cf}', "");
Expect(1, 917632, '\p{^cf}', "");
Expect(1, 917632, '\P{cf}', "");
Expect(0, 917632, '\P{^cf}', "");
Expect(1, 917631, '\p{ 	Cf}', "");
Expect(0, 917631, '\p{^ 	Cf}', "");
Expect(0, 917631, '\P{ 	Cf}', "");
Expect(1, 917631, '\P{^ 	Cf}', "");
Expect(0, 917632, '\p{ 	Cf}', "");
Expect(1, 917632, '\p{^ 	Cf}', "");
Expect(1, 917632, '\P{ 	Cf}', "");
Expect(0, 917632, '\P{^ 	Cf}', "");
Error('\p{	:=Is_cf}');
Error('\P{	:=Is_cf}');
Expect(1, 917631, '\p{iscf}', "");
Expect(0, 917631, '\p{^iscf}', "");
Expect(0, 917631, '\P{iscf}', "");
Expect(1, 917631, '\P{^iscf}', "");
Expect(0, 917632, '\p{iscf}', "");
Expect(1, 917632, '\p{^iscf}', "");
Expect(1, 917632, '\P{iscf}', "");
Expect(0, 917632, '\P{^iscf}', "");
Expect(1, 917631, '\p{ 	Is_CF}', "");
Expect(0, 917631, '\p{^ 	Is_CF}', "");
Expect(0, 917631, '\P{ 	Is_CF}', "");
Expect(1, 917631, '\P{^ 	Is_CF}', "");
Expect(0, 917632, '\p{ 	Is_CF}', "");
Expect(1, 917632, '\p{^ 	Is_CF}', "");
Expect(1, 917632, '\P{ 	Is_CF}', "");
Expect(0, 917632, '\P{^ 	Is_CF}', "");
Error('\p{-/a/ALCHEMICAL_Symbols}');
Error('\P{-/a/ALCHEMICAL_Symbols}');
Expect(1, 128895, '\p{alchemicalsymbols}', "");
Expect(0, 128895, '\p{^alchemicalsymbols}', "");
Expect(0, 128895, '\P{alchemicalsymbols}', "");
Expect(1, 128895, '\P{^alchemicalsymbols}', "");
Expect(0, 128896, '\p{alchemicalsymbols}', "");
Expect(1, 128896, '\p{^alchemicalsymbols}', "");
Expect(1, 128896, '\P{alchemicalsymbols}', "");
Expect(0, 128896, '\P{^alchemicalsymbols}', "");
Expect(1, 128895, '\p{-	Alchemical_Symbols}', "");
Expect(0, 128895, '\p{^-	Alchemical_Symbols}', "");
Expect(0, 128895, '\P{-	Alchemical_Symbols}', "");
Expect(1, 128895, '\P{^-	Alchemical_Symbols}', "");
Expect(0, 128896, '\p{-	Alchemical_Symbols}', "");
Expect(1, 128896, '\p{^-	Alchemical_Symbols}', "");
Expect(1, 128896, '\P{-	Alchemical_Symbols}', "");
Expect(0, 128896, '\P{^-	Alchemical_Symbols}', "");
Error('\p{-/a/IS_ALCHEMICAL_Symbols}');
Error('\P{-/a/IS_ALCHEMICAL_Symbols}');
Expect(1, 128895, '\p{isalchemicalsymbols}', "");
Expect(0, 128895, '\p{^isalchemicalsymbols}', "");
Expect(0, 128895, '\P{isalchemicalsymbols}', "");
Expect(1, 128895, '\P{^isalchemicalsymbols}', "");
Expect(0, 128896, '\p{isalchemicalsymbols}', "");
Expect(1, 128896, '\p{^isalchemicalsymbols}', "");
Expect(1, 128896, '\P{isalchemicalsymbols}', "");
Expect(0, 128896, '\P{^isalchemicalsymbols}', "");
Expect(1, 128895, '\p{ Is_ALCHEMICAL_symbols}', "");
Expect(0, 128895, '\p{^ Is_ALCHEMICAL_symbols}', "");
Expect(0, 128895, '\P{ Is_ALCHEMICAL_symbols}', "");
Expect(1, 128895, '\P{^ Is_ALCHEMICAL_symbols}', "");
Expect(0, 128896, '\p{ Is_ALCHEMICAL_symbols}', "");
Expect(1, 128896, '\p{^ Is_ALCHEMICAL_symbols}', "");
Expect(1, 128896, '\P{ Is_ALCHEMICAL_symbols}', "");
Expect(0, 128896, '\P{^ Is_ALCHEMICAL_symbols}', "");
Error('\p{ In_ALCHEMICAL_Symbols:=}');
Error('\P{ In_ALCHEMICAL_Symbols:=}');
Expect(1, 128895, '\p{inalchemicalsymbols}', "");
Expect(0, 128895, '\p{^inalchemicalsymbols}', "");
Expect(0, 128895, '\P{inalchemicalsymbols}', "");
Expect(1, 128895, '\P{^inalchemicalsymbols}', "");
Expect(0, 128896, '\p{inalchemicalsymbols}', "");
Expect(1, 128896, '\p{^inalchemicalsymbols}', "");
Expect(1, 128896, '\P{inalchemicalsymbols}', "");
Expect(0, 128896, '\P{^inalchemicalsymbols}', "");
Expect(1, 128895, '\p{-	In_ALCHEMICAL_Symbols}', "");
Expect(0, 128895, '\p{^-	In_ALCHEMICAL_Symbols}', "");
Expect(0, 128895, '\P{-	In_ALCHEMICAL_Symbols}', "");
Expect(1, 128895, '\P{^-	In_ALCHEMICAL_Symbols}', "");
Expect(0, 128896, '\p{-	In_ALCHEMICAL_Symbols}', "");
Expect(1, 128896, '\p{^-	In_ALCHEMICAL_Symbols}', "");
Expect(1, 128896, '\P{-	In_ALCHEMICAL_Symbols}', "");
Expect(0, 128896, '\P{^-	In_ALCHEMICAL_Symbols}', "");
Error('\p{-_alchemical:=}');
Error('\P{-_alchemical:=}');
Expect(1, 128895, '\p{alchemical}', "");
Expect(0, 128895, '\p{^alchemical}', "");
Expect(0, 128895, '\P{alchemical}', "");
Expect(1, 128895, '\P{^alchemical}', "");
Expect(0, 128896, '\p{alchemical}', "");
Expect(1, 128896, '\p{^alchemical}', "");
Expect(1, 128896, '\P{alchemical}', "");
Expect(0, 128896, '\P{^alchemical}', "");
Expect(1, 128895, '\p{	 Alchemical}', "");
Expect(0, 128895, '\p{^	 Alchemical}', "");
Expect(0, 128895, '\P{	 Alchemical}', "");
Expect(1, 128895, '\P{^	 Alchemical}', "");
Expect(0, 128896, '\p{	 Alchemical}', "");
Expect(1, 128896, '\p{^	 Alchemical}', "");
Expect(1, 128896, '\P{	 Alchemical}', "");
Expect(0, 128896, '\P{^	 Alchemical}', "");
Error('\p{_:=Is_ALCHEMICAL}');
Error('\P{_:=Is_ALCHEMICAL}');
Expect(1, 128895, '\p{isalchemical}', "");
Expect(0, 128895, '\p{^isalchemical}', "");
Expect(0, 128895, '\P{isalchemical}', "");
Expect(1, 128895, '\P{^isalchemical}', "");
Expect(0, 128896, '\p{isalchemical}', "");
Expect(1, 128896, '\p{^isalchemical}', "");
Expect(1, 128896, '\P{isalchemical}', "");
Expect(0, 128896, '\P{^isalchemical}', "");
Expect(1, 128895, '\p{ Is_alchemical}', "");
Expect(0, 128895, '\p{^ Is_alchemical}', "");
Expect(0, 128895, '\P{ Is_alchemical}', "");
Expect(1, 128895, '\P{^ Is_alchemical}', "");
Expect(0, 128896, '\p{ Is_alchemical}', "");
Expect(1, 128896, '\p{^ Is_alchemical}', "");
Expect(1, 128896, '\P{ Is_alchemical}', "");
Expect(0, 128896, '\P{^ Is_alchemical}', "");
Error('\p{	:=In_ALCHEMICAL}');
Error('\P{	:=In_ALCHEMICAL}');
Expect(1, 128895, '\p{inalchemical}', "");
Expect(0, 128895, '\p{^inalchemical}', "");
Expect(0, 128895, '\P{inalchemical}', "");
Expect(1, 128895, '\P{^inalchemical}', "");
Expect(0, 128896, '\p{inalchemical}', "");
Expect(1, 128896, '\p{^inalchemical}', "");
Expect(1, 128896, '\P{inalchemical}', "");
Expect(0, 128896, '\P{^inalchemical}', "");
Expect(1, 128895, '\p{ _In_alchemical}', "");
Expect(0, 128895, '\p{^ _In_alchemical}', "");
Expect(0, 128895, '\P{ _In_alchemical}', "");
Expect(1, 128895, '\P{^ _In_alchemical}', "");
Expect(0, 128896, '\p{ _In_alchemical}', "");
Expect(1, 128896, '\p{^ _In_alchemical}', "");
Expect(1, 128896, '\P{ _In_alchemical}', "");
Expect(0, 128896, '\P{^ _In_alchemical}', "");
Error('\p{:=-	other_NUMBER}');
Error('\P{:=-	other_NUMBER}');
Expect(1, 127242, '\p{othernumber}', "");
Expect(0, 127242, '\p{^othernumber}', "");
Expect(0, 127242, '\P{othernumber}', "");
Expect(1, 127242, '\P{^othernumber}', "");
Expect(0, 127243, '\p{othernumber}', "");
Expect(1, 127243, '\p{^othernumber}', "");
Expect(1, 127243, '\P{othernumber}', "");
Expect(0, 127243, '\P{^othernumber}', "");
Expect(1, 127242, '\p{__Other_NUMBER}', "");
Expect(0, 127242, '\p{^__Other_NUMBER}', "");
Expect(0, 127242, '\P{__Other_NUMBER}', "");
Expect(1, 127242, '\P{^__Other_NUMBER}', "");
Expect(0, 127243, '\p{__Other_NUMBER}', "");
Expect(1, 127243, '\p{^__Other_NUMBER}', "");
Expect(1, 127243, '\P{__Other_NUMBER}', "");
Expect(0, 127243, '\P{^__Other_NUMBER}', "");
Error('\p{-:=IS_OTHER_Number}');
Error('\P{-:=IS_OTHER_Number}');
Expect(1, 127242, '\p{isothernumber}', "");
Expect(0, 127242, '\p{^isothernumber}', "");
Expect(0, 127242, '\P{isothernumber}', "");
Expect(1, 127242, '\P{^isothernumber}', "");
Expect(0, 127243, '\p{isothernumber}', "");
Expect(1, 127243, '\p{^isothernumber}', "");
Expect(1, 127243, '\P{isothernumber}', "");
Expect(0, 127243, '\P{^isothernumber}', "");
Expect(1, 127242, '\p{_IS_Other_NUMBER}', "");
Expect(0, 127242, '\p{^_IS_Other_NUMBER}', "");
Expect(0, 127242, '\P{_IS_Other_NUMBER}', "");
Expect(1, 127242, '\P{^_IS_Other_NUMBER}', "");
Expect(0, 127243, '\p{_IS_Other_NUMBER}', "");
Expect(1, 127243, '\p{^_IS_Other_NUMBER}', "");
Expect(1, 127243, '\P{_IS_Other_NUMBER}', "");
Expect(0, 127243, '\P{^_IS_Other_NUMBER}', "");
Error('\p{/a/- NO}');
Error('\P{/a/- NO}');
Expect(1, 127242, '\p{no}', "");
Expect(0, 127242, '\p{^no}', "");
Expect(0, 127242, '\P{no}', "");
Expect(1, 127242, '\P{^no}', "");
Expect(0, 127243, '\p{no}', "");
Expect(1, 127243, '\p{^no}', "");
Expect(1, 127243, '\P{no}', "");
Expect(0, 127243, '\P{^no}', "");
Expect(1, 127242, '\p{	_no}', "");
Expect(0, 127242, '\p{^	_no}', "");
Expect(0, 127242, '\P{	_no}', "");
Expect(1, 127242, '\P{^	_no}', "");
Expect(0, 127243, '\p{	_no}', "");
Expect(1, 127243, '\p{^	_no}', "");
Expect(1, 127243, '\P{	_no}', "");
Expect(0, 127243, '\P{^	_no}', "");
Error('\p{	 IS_No/a/}');
Error('\P{	 IS_No/a/}');
Expect(1, 127242, '\p{isno}', "");
Expect(0, 127242, '\p{^isno}', "");
Expect(0, 127242, '\P{isno}', "");
Expect(1, 127242, '\P{^isno}', "");
Expect(0, 127243, '\p{isno}', "");
Expect(1, 127243, '\p{^isno}', "");
Expect(1, 127243, '\P{isno}', "");
Expect(0, 127243, '\P{^isno}', "");
Expect(1, 127242, '\p{ _Is_NO}', "");
Expect(0, 127242, '\p{^ _Is_NO}', "");
Expect(0, 127242, '\P{ _Is_NO}', "");
Expect(1, 127242, '\P{^ _Is_NO}', "");
Expect(0, 127243, '\p{ _Is_NO}', "");
Expect(1, 127243, '\p{^ _Is_NO}', "");
Expect(1, 127243, '\P{ _Is_NO}', "");
Expect(0, 127243, '\P{^ _Is_NO}', "");
Error('\p{	 Latin_1_Supplement:=}');
Error('\P{	 Latin_1_Supplement:=}');
Expect(1, 255, '\p{latin1supplement}', "");
Expect(0, 255, '\p{^latin1supplement}', "");
Expect(0, 255, '\P{latin1supplement}', "");
Expect(1, 255, '\P{^latin1supplement}', "");
Expect(0, 256, '\p{latin1supplement}', "");
Expect(1, 256, '\p{^latin1supplement}', "");
Expect(1, 256, '\P{latin1supplement}', "");
Expect(0, 256, '\P{^latin1supplement}', "");
Expect(1, 255, '\p{	_Latin_1_Supplement}', "");
Expect(0, 255, '\p{^	_Latin_1_Supplement}', "");
Expect(0, 255, '\P{	_Latin_1_Supplement}', "");
Expect(1, 255, '\P{^	_Latin_1_Supplement}', "");
Expect(0, 256, '\p{	_Latin_1_Supplement}', "");
Expect(1, 256, '\p{^	_Latin_1_Supplement}', "");
Expect(1, 256, '\P{	_Latin_1_Supplement}', "");
Expect(0, 256, '\P{^	_Latin_1_Supplement}', "");
Error('\p{	Is_Latin_1_Supplement:=}');
Error('\P{	Is_Latin_1_Supplement:=}');
Expect(1, 255, '\p{islatin1supplement}', "");
Expect(0, 255, '\p{^islatin1supplement}', "");
Expect(0, 255, '\P{islatin1supplement}', "");
Expect(1, 255, '\P{^islatin1supplement}', "");
Expect(0, 256, '\p{islatin1supplement}', "");
Expect(1, 256, '\p{^islatin1supplement}', "");
Expect(1, 256, '\P{islatin1supplement}', "");
Expect(0, 256, '\P{^islatin1supplement}', "");
Expect(1, 255, '\p{is_latin_1_SUPPLEMENT}', "");
Expect(0, 255, '\p{^is_latin_1_SUPPLEMENT}', "");
Expect(0, 255, '\P{is_latin_1_SUPPLEMENT}', "");
Expect(1, 255, '\P{^is_latin_1_SUPPLEMENT}', "");
Expect(0, 256, '\p{is_latin_1_SUPPLEMENT}', "");
Expect(1, 256, '\p{^is_latin_1_SUPPLEMENT}', "");
Expect(1, 256, '\P{is_latin_1_SUPPLEMENT}', "");
Expect(0, 256, '\P{^is_latin_1_SUPPLEMENT}', "");
Error('\p{:=	IN_LATIN_1_Supplement}');
Error('\P{:=	IN_LATIN_1_Supplement}');
Expect(1, 255, '\p{inlatin1supplement}', "");
Expect(0, 255, '\p{^inlatin1supplement}', "");
Expect(0, 255, '\P{inlatin1supplement}', "");
Expect(1, 255, '\P{^inlatin1supplement}', "");
Expect(0, 256, '\p{inlatin1supplement}', "");
Expect(1, 256, '\p{^inlatin1supplement}', "");
Expect(1, 256, '\P{inlatin1supplement}', "");
Expect(0, 256, '\P{^inlatin1supplement}', "");
Expect(1, 255, '\p{-in_Latin_1_supplement}', "");
Expect(0, 255, '\p{^-in_Latin_1_supplement}', "");
Expect(0, 255, '\P{-in_Latin_1_supplement}', "");
Expect(1, 255, '\P{^-in_Latin_1_supplement}', "");
Expect(0, 256, '\p{-in_Latin_1_supplement}', "");
Expect(1, 256, '\p{^-in_Latin_1_supplement}', "");
Expect(1, 256, '\P{-in_Latin_1_supplement}', "");
Expect(0, 256, '\P{^-in_Latin_1_supplement}', "");
Error('\p{:= -LATIN_1_SUP}');
Error('\P{:= -LATIN_1_SUP}');
Expect(1, 255, '\p{latin1sup}', "");
Expect(0, 255, '\p{^latin1sup}', "");
Expect(0, 255, '\P{latin1sup}', "");
Expect(1, 255, '\P{^latin1sup}', "");
Expect(0, 256, '\p{latin1sup}', "");
Expect(1, 256, '\p{^latin1sup}', "");
Expect(1, 256, '\P{latin1sup}', "");
Expect(0, 256, '\P{^latin1sup}', "");
Expect(1, 255, '\p{	 Latin_1_SUP}', "");
Expect(0, 255, '\p{^	 Latin_1_SUP}', "");
Expect(0, 255, '\P{	 Latin_1_SUP}', "");
Expect(1, 255, '\P{^	 Latin_1_SUP}', "");
Expect(0, 256, '\p{	 Latin_1_SUP}', "");
Expect(1, 256, '\p{^	 Latin_1_SUP}', "");
Expect(1, 256, '\P{	 Latin_1_SUP}', "");
Expect(0, 256, '\P{^	 Latin_1_SUP}', "");
Error('\p{/a/__IS_latin_1_sup}');
Error('\P{/a/__IS_latin_1_sup}');
Expect(1, 255, '\p{islatin1sup}', "");
Expect(0, 255, '\p{^islatin1sup}', "");
Expect(0, 255, '\P{islatin1sup}', "");
Expect(1, 255, '\P{^islatin1sup}', "");
Expect(0, 256, '\p{islatin1sup}', "");
Expect(1, 256, '\p{^islatin1sup}', "");
Expect(1, 256, '\P{islatin1sup}', "");
Expect(0, 256, '\P{^islatin1sup}', "");
Expect(1, 255, '\p{_is_LATIN_1_Sup}', "");
Expect(0, 255, '\p{^_is_LATIN_1_Sup}', "");
Expect(0, 255, '\P{_is_LATIN_1_Sup}', "");
Expect(1, 255, '\P{^_is_LATIN_1_Sup}', "");
Expect(0, 256, '\p{_is_LATIN_1_Sup}', "");
Expect(1, 256, '\p{^_is_LATIN_1_Sup}', "");
Expect(1, 256, '\P{_is_LATIN_1_Sup}', "");
Expect(0, 256, '\P{^_is_LATIN_1_Sup}', "");
Error('\p{-:=in_LATIN_1_sup}');
Error('\P{-:=in_LATIN_1_sup}');
Expect(1, 255, '\p{inlatin1sup}', "");
Expect(0, 255, '\p{^inlatin1sup}', "");
Expect(0, 255, '\P{inlatin1sup}', "");
Expect(1, 255, '\P{^inlatin1sup}', "");
Expect(0, 256, '\p{inlatin1sup}', "");
Expect(1, 256, '\p{^inlatin1sup}', "");
Expect(1, 256, '\P{inlatin1sup}', "");
Expect(0, 256, '\P{^inlatin1sup}', "");
Expect(1, 255, '\p{_-in_latin_1_sup}', "");
Expect(0, 255, '\p{^_-in_latin_1_sup}', "");
Expect(0, 255, '\P{_-in_latin_1_sup}', "");
Expect(1, 255, '\P{^_-in_latin_1_sup}', "");
Expect(0, 256, '\p{_-in_latin_1_sup}', "");
Expect(1, 256, '\p{^_-in_latin_1_sup}', "");
Expect(1, 256, '\P{_-in_latin_1_sup}', "");
Expect(0, 256, '\P{^_-in_latin_1_sup}', "");
Error('\p{/a/_ Latin_1}');
Error('\P{/a/_ Latin_1}');
Expect(1, 255, '\p{latin1}', "");
Expect(0, 255, '\p{^latin1}', "");
Expect(0, 255, '\P{latin1}', "");
Expect(1, 255, '\P{^latin1}', "");
Expect(0, 256, '\p{latin1}', "");
Expect(1, 256, '\p{^latin1}', "");
Expect(1, 256, '\P{latin1}', "");
Expect(0, 256, '\P{^latin1}', "");
Expect(1, 255, '\p{latin_1}', "");
Expect(0, 255, '\p{^latin_1}', "");
Expect(0, 255, '\P{latin_1}', "");
Expect(1, 255, '\P{^latin_1}', "");
Expect(0, 256, '\p{latin_1}', "");
Expect(1, 256, '\p{^latin_1}', "");
Expect(1, 256, '\P{latin_1}', "");
Expect(0, 256, '\P{^latin_1}', "");
Error('\p{/a/ IS_Latin_1}');
Error('\P{/a/ IS_Latin_1}');
Expect(1, 255, '\p{islatin1}', "");
Expect(0, 255, '\p{^islatin1}', "");
Expect(0, 255, '\P{islatin1}', "");
Expect(1, 255, '\P{^islatin1}', "");
Expect(0, 256, '\p{islatin1}', "");
Expect(1, 256, '\p{^islatin1}', "");
Expect(1, 256, '\P{islatin1}', "");
Expect(0, 256, '\P{^islatin1}', "");
Expect(1, 255, '\p{--Is_Latin_1}', "");
Expect(0, 255, '\p{^--Is_Latin_1}', "");
Expect(0, 255, '\P{--Is_Latin_1}', "");
Expect(1, 255, '\P{^--Is_Latin_1}', "");
Expect(0, 256, '\p{--Is_Latin_1}', "");
Expect(1, 256, '\p{^--Is_Latin_1}', "");
Expect(1, 256, '\P{--Is_Latin_1}', "");
Expect(0, 256, '\P{^--Is_Latin_1}', "");
Error('\p{:=_	IN_Latin_1}');
Error('\P{:=_	IN_Latin_1}');
Expect(1, 255, '\p{inlatin1}', "");
Expect(0, 255, '\p{^inlatin1}', "");
Expect(0, 255, '\P{inlatin1}', "");
Expect(1, 255, '\P{^inlatin1}', "");
Expect(0, 256, '\p{inlatin1}', "");
Expect(1, 256, '\p{^inlatin1}', "");
Expect(1, 256, '\P{inlatin1}', "");
Expect(0, 256, '\P{^inlatin1}', "");
Expect(1, 255, '\p{	-IN_Latin_1}', "");
Expect(0, 255, '\p{^	-IN_Latin_1}', "");
Expect(0, 255, '\P{	-IN_Latin_1}', "");
Expect(1, 255, '\P{^	-IN_Latin_1}', "");
Expect(0, 256, '\p{	-IN_Latin_1}', "");
Expect(1, 256, '\p{^	-IN_Latin_1}', "");
Expect(1, 256, '\P{	-IN_Latin_1}', "");
Expect(0, 256, '\P{^	-IN_Latin_1}', "");
Error('\p{ /a/Tamil}');
Error('\P{ /a/Tamil}');
Expect(1, 3066, '\p{tamil}', "");
Expect(0, 3066, '\p{^tamil}', "");
Expect(0, 3066, '\P{tamil}', "");
Expect(1, 3066, '\P{^tamil}', "");
Expect(0, 3067, '\p{tamil}', "");
Expect(1, 3067, '\p{^tamil}', "");
Expect(1, 3067, '\P{tamil}', "");
Expect(0, 3067, '\P{^tamil}', "");
Expect(1, 3066, '\p{--Tamil}', "");
Expect(0, 3066, '\p{^--Tamil}', "");
Expect(0, 3066, '\P{--Tamil}', "");
Expect(1, 3066, '\P{^--Tamil}', "");
Expect(0, 3067, '\p{--Tamil}', "");
Expect(1, 3067, '\p{^--Tamil}', "");
Expect(1, 3067, '\P{--Tamil}', "");
Expect(0, 3067, '\P{^--Tamil}', "");
Error('\p{	:=Is_Tamil}');
Error('\P{	:=Is_Tamil}');
Expect(1, 3066, '\p{istamil}', "");
Expect(0, 3066, '\p{^istamil}', "");
Expect(0, 3066, '\P{istamil}', "");
Expect(1, 3066, '\P{^istamil}', "");
Expect(0, 3067, '\p{istamil}', "");
Expect(1, 3067, '\p{^istamil}', "");
Expect(1, 3067, '\P{istamil}', "");
Expect(0, 3067, '\P{^istamil}', "");
Expect(1, 3066, '\p{ -Is_Tamil}', "");
Expect(0, 3066, '\p{^ -Is_Tamil}', "");
Expect(0, 3066, '\P{ -Is_Tamil}', "");
Expect(1, 3066, '\P{^ -Is_Tamil}', "");
Expect(0, 3067, '\p{ -Is_Tamil}', "");
Expect(1, 3067, '\p{^ -Is_Tamil}', "");
Expect(1, 3067, '\P{ -Is_Tamil}', "");
Expect(0, 3067, '\P{^ -Is_Tamil}', "");
Error('\p{	:=Taml}');
Error('\P{	:=Taml}');
Expect(1, 3066, '\p{taml}', "");
Expect(0, 3066, '\p{^taml}', "");
Expect(0, 3066, '\P{taml}', "");
Expect(1, 3066, '\P{^taml}', "");
Expect(0, 3067, '\p{taml}', "");
Expect(1, 3067, '\p{^taml}', "");
Expect(1, 3067, '\P{taml}', "");
Expect(0, 3067, '\P{^taml}', "");
Expect(1, 3066, '\p{	 TAML}', "");
Expect(0, 3066, '\p{^	 TAML}', "");
Expect(0, 3066, '\P{	 TAML}', "");
Expect(1, 3066, '\P{^	 TAML}', "");
Expect(0, 3067, '\p{	 TAML}', "");
Expect(1, 3067, '\p{^	 TAML}', "");
Expect(1, 3067, '\P{	 TAML}', "");
Expect(0, 3067, '\P{^	 TAML}', "");
Error('\p{/a/- Is_TAML}');
Error('\P{/a/- Is_TAML}');
Expect(1, 3066, '\p{istaml}', "");
Expect(0, 3066, '\p{^istaml}', "");
Expect(0, 3066, '\P{istaml}', "");
Expect(1, 3066, '\P{^istaml}', "");
Expect(0, 3067, '\p{istaml}', "");
Expect(1, 3067, '\p{^istaml}', "");
Expect(1, 3067, '\P{istaml}', "");
Expect(0, 3067, '\P{^istaml}', "");
Expect(1, 3066, '\p{ 	IS_taml}', "");
Expect(0, 3066, '\p{^ 	IS_taml}', "");
Expect(0, 3066, '\P{ 	IS_taml}', "");
Expect(1, 3066, '\P{^ 	IS_taml}', "");
Expect(0, 3067, '\p{ 	IS_taml}', "");
Expect(1, 3067, '\p{^ 	IS_taml}', "");
Expect(1, 3067, '\P{ 	IS_taml}', "");
Expect(0, 3067, '\P{^ 	IS_taml}', "");
Error('\p{:= Shavian}');
Error('\P{:= Shavian}');
Expect(1, 66687, '\p{shavian}', "");
Expect(0, 66687, '\p{^shavian}', "");
Expect(0, 66687, '\P{shavian}', "");
Expect(1, 66687, '\P{^shavian}', "");
Expect(0, 66688, '\p{shavian}', "");
Expect(1, 66688, '\p{^shavian}', "");
Expect(1, 66688, '\P{shavian}', "");
Expect(0, 66688, '\P{^shavian}', "");
Expect(1, 66687, '\p{	 Shavian}', "");
Expect(0, 66687, '\p{^	 Shavian}', "");
Expect(0, 66687, '\P{	 Shavian}', "");
Expect(1, 66687, '\P{^	 Shavian}', "");
Expect(0, 66688, '\p{	 Shavian}', "");
Expect(1, 66688, '\p{^	 Shavian}', "");
Expect(1, 66688, '\P{	 Shavian}', "");
Expect(0, 66688, '\P{^	 Shavian}', "");
Error('\p{:=is_Shavian}');
Error('\P{:=is_Shavian}');
Expect(1, 66687, '\p{isshavian}', "");
Expect(0, 66687, '\p{^isshavian}', "");
Expect(0, 66687, '\P{isshavian}', "");
Expect(1, 66687, '\P{^isshavian}', "");
Expect(0, 66688, '\p{isshavian}', "");
Expect(1, 66688, '\p{^isshavian}', "");
Expect(1, 66688, '\P{isshavian}', "");
Expect(0, 66688, '\P{^isshavian}', "");
Expect(1, 66687, '\p{-	Is_shavian}', "");
Expect(0, 66687, '\p{^-	Is_shavian}', "");
Expect(0, 66687, '\P{-	Is_shavian}', "");
Expect(1, 66687, '\P{^-	Is_shavian}', "");
Expect(0, 66688, '\p{-	Is_shavian}', "");
Expect(1, 66688, '\p{^-	Is_shavian}', "");
Expect(1, 66688, '\P{-	Is_shavian}', "");
Expect(0, 66688, '\P{^-	Is_shavian}', "");
Error('\p{/a/-_Shaw}');
Error('\P{/a/-_Shaw}');
Expect(1, 66687, '\p{shaw}', "");
Expect(0, 66687, '\p{^shaw}', "");
Expect(0, 66687, '\P{shaw}', "");
Expect(1, 66687, '\P{^shaw}', "");
Expect(0, 66688, '\p{shaw}', "");
Expect(1, 66688, '\p{^shaw}', "");
Expect(1, 66688, '\P{shaw}', "");
Expect(0, 66688, '\P{^shaw}', "");
Expect(1, 66687, '\p{	Shaw}', "");
Expect(0, 66687, '\p{^	Shaw}', "");
Expect(0, 66687, '\P{	Shaw}', "");
Expect(1, 66687, '\P{^	Shaw}', "");
Expect(0, 66688, '\p{	Shaw}', "");
Expect(1, 66688, '\p{^	Shaw}', "");
Expect(1, 66688, '\P{	Shaw}', "");
Expect(0, 66688, '\P{^	Shaw}', "");
Error('\p{:=	-Is_SHAW}');
Error('\P{:=	-Is_SHAW}');
Expect(1, 66687, '\p{isshaw}', "");
Expect(0, 66687, '\p{^isshaw}', "");
Expect(0, 66687, '\P{isshaw}', "");
Expect(1, 66687, '\P{^isshaw}', "");
Expect(0, 66688, '\p{isshaw}', "");
Expect(1, 66688, '\p{^isshaw}', "");
Expect(1, 66688, '\P{isshaw}', "");
Expect(0, 66688, '\P{^isshaw}', "");
Expect(1, 66687, '\p{- Is_shaw}', "");
Expect(0, 66687, '\p{^- Is_shaw}', "");
Expect(0, 66687, '\P{- Is_shaw}', "");
Expect(1, 66687, '\P{^- Is_shaw}', "");
Expect(0, 66688, '\p{- Is_shaw}', "");
Expect(1, 66688, '\p{^- Is_shaw}', "");
Expect(1, 66688, '\P{- Is_shaw}', "");
Expect(0, 66688, '\P{^- Is_shaw}', "");
Error('\p{/a/	IPA_Extensions}');
Error('\P{/a/	IPA_Extensions}');
Expect(1, 687, '\p{ipaextensions}', "");
Expect(0, 687, '\p{^ipaextensions}', "");
Expect(0, 687, '\P{ipaextensions}', "");
Expect(1, 687, '\P{^ipaextensions}', "");
Expect(0, 688, '\p{ipaextensions}', "");
Expect(1, 688, '\p{^ipaextensions}', "");
Expect(1, 688, '\P{ipaextensions}', "");
Expect(0, 688, '\P{^ipaextensions}', "");
Expect(1, 687, '\p{ _IPA_EXTENSIONS}', "");
Expect(0, 687, '\p{^ _IPA_EXTENSIONS}', "");
Expect(0, 687, '\P{ _IPA_EXTENSIONS}', "");
Expect(1, 687, '\P{^ _IPA_EXTENSIONS}', "");
Expect(0, 688, '\p{ _IPA_EXTENSIONS}', "");
Expect(1, 688, '\p{^ _IPA_EXTENSIONS}', "");
Expect(1, 688, '\P{ _IPA_EXTENSIONS}', "");
Expect(0, 688, '\P{^ _IPA_EXTENSIONS}', "");
Error('\p{_IS_IPA_Extensions/a/}');
Error('\P{_IS_IPA_Extensions/a/}');
Expect(1, 687, '\p{isipaextensions}', "");
Expect(0, 687, '\p{^isipaextensions}', "");
Expect(0, 687, '\P{isipaextensions}', "");
Expect(1, 687, '\P{^isipaextensions}', "");
Expect(0, 688, '\p{isipaextensions}', "");
Expect(1, 688, '\p{^isipaextensions}', "");
Expect(1, 688, '\P{isipaextensions}', "");
Expect(0, 688, '\P{^isipaextensions}', "");
Expect(1, 687, '\p{ _Is_IPA_EXTENSIONS}', "");
Expect(0, 687, '\p{^ _Is_IPA_EXTENSIONS}', "");
Expect(0, 687, '\P{ _Is_IPA_EXTENSIONS}', "");
Expect(1, 687, '\P{^ _Is_IPA_EXTENSIONS}', "");
Expect(0, 688, '\p{ _Is_IPA_EXTENSIONS}', "");
Expect(1, 688, '\p{^ _Is_IPA_EXTENSIONS}', "");
Expect(1, 688, '\P{ _Is_IPA_EXTENSIONS}', "");
Expect(0, 688, '\P{^ _Is_IPA_EXTENSIONS}', "");
Error('\p{/a/In_ipa_Extensions}');
Error('\P{/a/In_ipa_Extensions}');
Expect(1, 687, '\p{inipaextensions}', "");
Expect(0, 687, '\p{^inipaextensions}', "");
Expect(0, 687, '\P{inipaextensions}', "");
Expect(1, 687, '\P{^inipaextensions}', "");
Expect(0, 688, '\p{inipaextensions}', "");
Expect(1, 688, '\p{^inipaextensions}', "");
Expect(1, 688, '\P{inipaextensions}', "");
Expect(0, 688, '\P{^inipaextensions}', "");
Expect(1, 687, '\p{--in_ipa_Extensions}', "");
Expect(0, 687, '\p{^--in_ipa_Extensions}', "");
Expect(0, 687, '\P{--in_ipa_Extensions}', "");
Expect(1, 687, '\P{^--in_ipa_Extensions}', "");
Expect(0, 688, '\p{--in_ipa_Extensions}', "");
Expect(1, 688, '\p{^--in_ipa_Extensions}', "");
Expect(1, 688, '\P{--in_ipa_Extensions}', "");
Expect(0, 688, '\P{^--in_ipa_Extensions}', "");
Error('\p{_IPA_Ext:=}');
Error('\P{_IPA_Ext:=}');
Expect(1, 687, '\p{ipaext}', "");
Expect(0, 687, '\p{^ipaext}', "");
Expect(0, 687, '\P{ipaext}', "");
Expect(1, 687, '\P{^ipaext}', "");
Expect(0, 688, '\p{ipaext}', "");
Expect(1, 688, '\p{^ipaext}', "");
Expect(1, 688, '\P{ipaext}', "");
Expect(0, 688, '\P{^ipaext}', "");
Expect(1, 687, '\p{_-IPA_Ext}', "");
Expect(0, 687, '\p{^_-IPA_Ext}', "");
Expect(0, 687, '\P{_-IPA_Ext}', "");
Expect(1, 687, '\P{^_-IPA_Ext}', "");
Expect(0, 688, '\p{_-IPA_Ext}', "");
Expect(1, 688, '\p{^_-IPA_Ext}', "");
Expect(1, 688, '\P{_-IPA_Ext}', "");
Expect(0, 688, '\P{^_-IPA_Ext}', "");
Error('\p{:= _IS_IPA_ext}');
Error('\P{:= _IS_IPA_ext}');
Expect(1, 687, '\p{isipaext}', "");
Expect(0, 687, '\p{^isipaext}', "");
Expect(0, 687, '\P{isipaext}', "");
Expect(1, 687, '\P{^isipaext}', "");
Expect(0, 688, '\p{isipaext}', "");
Expect(1, 688, '\p{^isipaext}', "");
Expect(1, 688, '\P{isipaext}', "");
Expect(0, 688, '\P{^isipaext}', "");
Expect(1, 687, '\p{_ Is_IPA_EXT}', "");
Expect(0, 687, '\p{^_ Is_IPA_EXT}', "");
Expect(0, 687, '\P{_ Is_IPA_EXT}', "");
Expect(1, 687, '\P{^_ Is_IPA_EXT}', "");
Expect(0, 688, '\p{_ Is_IPA_EXT}', "");
Expect(1, 688, '\p{^_ Is_IPA_EXT}', "");
Expect(1, 688, '\P{_ Is_IPA_EXT}', "");
Expect(0, 688, '\P{^_ Is_IPA_EXT}', "");
Error('\p{:=In_IPA_EXT}');
Error('\P{:=In_IPA_EXT}');
Expect(1, 687, '\p{inipaext}', "");
Expect(0, 687, '\p{^inipaext}', "");
Expect(0, 687, '\P{inipaext}', "");
Expect(1, 687, '\P{^inipaext}', "");
Expect(0, 688, '\p{inipaext}', "");
Expect(1, 688, '\p{^inipaext}', "");
Expect(1, 688, '\P{inipaext}', "");
Expect(0, 688, '\P{^inipaext}', "");
Expect(1, 687, '\p{	In_ipa_Ext}', "");
Expect(0, 687, '\p{^	In_ipa_Ext}', "");
Expect(0, 687, '\P{	In_ipa_Ext}', "");
Expect(1, 687, '\P{^	In_ipa_Ext}', "");
Expect(0, 688, '\p{	In_ipa_Ext}', "");
Expect(1, 688, '\p{^	In_ipa_Ext}', "");
Expect(1, 688, '\P{	In_ipa_Ext}', "");
Expect(0, 688, '\P{^	In_ipa_Ext}', "");
Error('\p{hiragana/a/}');
Error('\P{hiragana/a/}');
Expect(1, 127488, '\p{hiragana}', "");
Expect(0, 127488, '\p{^hiragana}', "");
Expect(0, 127488, '\P{hiragana}', "");
Expect(1, 127488, '\P{^hiragana}', "");
Expect(0, 127489, '\p{hiragana}', "");
Expect(1, 127489, '\p{^hiragana}', "");
Expect(1, 127489, '\P{hiragana}', "");
Expect(0, 127489, '\P{^hiragana}', "");
Expect(1, 127488, '\p{--HIRAGANA}', "");
Expect(0, 127488, '\p{^--HIRAGANA}', "");
Expect(0, 127488, '\P{--HIRAGANA}', "");
Expect(1, 127488, '\P{^--HIRAGANA}', "");
Expect(0, 127489, '\p{--HIRAGANA}', "");
Expect(1, 127489, '\p{^--HIRAGANA}', "");
Expect(1, 127489, '\P{--HIRAGANA}', "");
Expect(0, 127489, '\P{^--HIRAGANA}', "");
Error('\p{/a/  Is_Hiragana}');
Error('\P{/a/  Is_Hiragana}');
Expect(1, 127488, '\p{ishiragana}', "");
Expect(0, 127488, '\p{^ishiragana}', "");
Expect(0, 127488, '\P{ishiragana}', "");
Expect(1, 127488, '\P{^ishiragana}', "");
Expect(0, 127489, '\p{ishiragana}', "");
Expect(1, 127489, '\p{^ishiragana}', "");
Expect(1, 127489, '\P{ishiragana}', "");
Expect(0, 127489, '\P{^ishiragana}', "");
Expect(1, 127488, '\p{- is_Hiragana}', "");
Expect(0, 127488, '\p{^- is_Hiragana}', "");
Expect(0, 127488, '\P{- is_Hiragana}', "");
Expect(1, 127488, '\P{^- is_Hiragana}', "");
Expect(0, 127489, '\p{- is_Hiragana}', "");
Expect(1, 127489, '\p{^- is_Hiragana}', "");
Expect(1, 127489, '\P{- is_Hiragana}', "");
Expect(0, 127489, '\P{^- is_Hiragana}', "");
Error('\p{  hira:=}');
Error('\P{  hira:=}');
Expect(1, 127488, '\p{hira}', "");
Expect(0, 127488, '\p{^hira}', "");
Expect(0, 127488, '\P{hira}', "");
Expect(1, 127488, '\P{^hira}', "");
Expect(0, 127489, '\p{hira}', "");
Expect(1, 127489, '\p{^hira}', "");
Expect(1, 127489, '\P{hira}', "");
Expect(0, 127489, '\P{^hira}', "");
Expect(1, 127488, '\p{_hira}', "");
Expect(0, 127488, '\p{^_hira}', "");
Expect(0, 127488, '\P{_hira}', "");
Expect(1, 127488, '\P{^_hira}', "");
Expect(0, 127489, '\p{_hira}', "");
Expect(1, 127489, '\p{^_hira}', "");
Expect(1, 127489, '\P{_hira}', "");
Expect(0, 127489, '\P{^_hira}', "");
Error('\p{_:=Is_Hira}');
Error('\P{_:=Is_Hira}');
Expect(1, 127488, '\p{ishira}', "");
Expect(0, 127488, '\p{^ishira}', "");
Expect(0, 127488, '\P{ishira}', "");
Expect(1, 127488, '\P{^ishira}', "");
Expect(0, 127489, '\p{ishira}', "");
Expect(1, 127489, '\p{^ishira}', "");
Expect(1, 127489, '\P{ishira}', "");
Expect(0, 127489, '\P{^ishira}', "");
Expect(1, 127488, '\p{-	Is_Hira}', "");
Expect(0, 127488, '\p{^-	Is_Hira}', "");
Expect(0, 127488, '\P{-	Is_Hira}', "");
Expect(1, 127488, '\P{^-	Is_Hira}', "");
Expect(0, 127489, '\p{-	Is_Hira}', "");
Expect(1, 127489, '\p{^-	Is_Hira}', "");
Expect(1, 127489, '\P{-	Is_Hira}', "");
Expect(0, 127489, '\P{^-	Is_Hira}', "");
Error('\p{__greek/a/}');
Error('\P{__greek/a/}');
Expect(1, 119365, '\p{greek}', "");
Expect(0, 119365, '\p{^greek}', "");
Expect(0, 119365, '\P{greek}', "");
Expect(1, 119365, '\P{^greek}', "");
Expect(0, 119366, '\p{greek}', "");
Expect(1, 119366, '\p{^greek}', "");
Expect(1, 119366, '\P{greek}', "");
Expect(0, 119366, '\P{^greek}', "");
Expect(1, 119365, '\p{	-greek}', "");
Expect(0, 119365, '\p{^	-greek}', "");
Expect(0, 119365, '\P{	-greek}', "");
Expect(1, 119365, '\P{^	-greek}', "");
Expect(0, 119366, '\p{	-greek}', "");
Expect(1, 119366, '\p{^	-greek}', "");
Expect(1, 119366, '\P{	-greek}', "");
Expect(0, 119366, '\P{^	-greek}', "");
Error('\p{/a/-is_Greek}');
Error('\P{/a/-is_Greek}');
Expect(1, 119365, '\p{isgreek}', "");
Expect(0, 119365, '\p{^isgreek}', "");
Expect(0, 119365, '\P{isgreek}', "");
Expect(1, 119365, '\P{^isgreek}', "");
Expect(0, 119366, '\p{isgreek}', "");
Expect(1, 119366, '\p{^isgreek}', "");
Expect(1, 119366, '\P{isgreek}', "");
Expect(0, 119366, '\P{^isgreek}', "");
Expect(1, 119365, '\p{_	IS_Greek}', "");
Expect(0, 119365, '\p{^_	IS_Greek}', "");
Expect(0, 119365, '\P{_	IS_Greek}', "");
Expect(1, 119365, '\P{^_	IS_Greek}', "");
Expect(0, 119366, '\p{_	IS_Greek}', "");
Expect(1, 119366, '\p{^_	IS_Greek}', "");
Expect(1, 119366, '\P{_	IS_Greek}', "");
Expect(0, 119366, '\P{^_	IS_Greek}', "");
Error('\p{-Grek:=}');
Error('\P{-Grek:=}');
Expect(1, 119365, '\p{grek}', "");
Expect(0, 119365, '\p{^grek}', "");
Expect(0, 119365, '\P{grek}', "");
Expect(1, 119365, '\P{^grek}', "");
Expect(0, 119366, '\p{grek}', "");
Expect(1, 119366, '\p{^grek}', "");
Expect(1, 119366, '\P{grek}', "");
Expect(0, 119366, '\P{^grek}', "");
Expect(1, 119365, '\p{	Grek}', "");
Expect(0, 119365, '\p{^	Grek}', "");
Expect(0, 119365, '\P{	Grek}', "");
Expect(1, 119365, '\P{^	Grek}', "");
Expect(0, 119366, '\p{	Grek}', "");
Expect(1, 119366, '\p{^	Grek}', "");
Expect(1, 119366, '\P{	Grek}', "");
Expect(0, 119366, '\P{^	Grek}', "");
Error('\p{	/a/Is_Grek}');
Error('\P{	/a/Is_Grek}');
Expect(1, 119365, '\p{isgrek}', "");
Expect(0, 119365, '\p{^isgrek}', "");
Expect(0, 119365, '\P{isgrek}', "");
Expect(1, 119365, '\P{^isgrek}', "");
Expect(0, 119366, '\p{isgrek}', "");
Expect(1, 119366, '\p{^isgrek}', "");
Expect(1, 119366, '\P{isgrek}', "");
Expect(0, 119366, '\P{^isgrek}', "");
Expect(1, 119365, '\p{	Is_Grek}', "");
Expect(0, 119365, '\p{^	Is_Grek}', "");
Expect(0, 119365, '\P{	Is_Grek}', "");
Expect(1, 119365, '\P{^	Is_Grek}', "");
Expect(0, 119366, '\p{	Is_Grek}', "");
Expect(1, 119366, '\p{^	Is_Grek}', "");
Expect(1, 119366, '\P{	Is_Grek}', "");
Expect(0, 119366, '\P{^	Is_Grek}', "");
Error('\p{__In_thai/a/}');
Error('\P{__In_thai/a/}');
Expect(1, 3711, '\p{inthai}', "");
Expect(0, 3711, '\p{^inthai}', "");
Expect(0, 3711, '\P{inthai}', "");
Expect(1, 3711, '\P{^inthai}', "");
Expect(0, 3712, '\p{inthai}', "");
Expect(1, 3712, '\p{^inthai}', "");
Expect(1, 3712, '\P{inthai}', "");
Expect(0, 3712, '\P{^inthai}', "");
Expect(1, 3711, '\p{	_IN_Thai}', "");
Expect(0, 3711, '\p{^	_IN_Thai}', "");
Expect(0, 3711, '\P{	_IN_Thai}', "");
Expect(1, 3711, '\P{^	_IN_Thai}', "");
Expect(0, 3712, '\p{	_IN_Thai}', "");
Expect(1, 3712, '\p{^	_IN_Thai}', "");
Expect(1, 3712, '\P{	_IN_Thai}', "");
Expect(0, 3712, '\P{^	_IN_Thai}', "");
Error('\p{	 In_Thai/a/}');
Error('\P{	 In_Thai/a/}');
Expect(1, 3711, '\p{	 IN_Thai}', "");
Expect(0, 3711, '\p{^	 IN_Thai}', "");
Expect(0, 3711, '\P{	 IN_Thai}', "");
Expect(1, 3711, '\P{^	 IN_Thai}', "");
Expect(0, 3712, '\p{	 IN_Thai}', "");
Expect(1, 3712, '\p{^	 IN_Thai}', "");
Expect(1, 3712, '\P{	 IN_Thai}', "");
Expect(0, 3712, '\P{^	 IN_Thai}', "");
Error('\p{	-Line_SEPARATOR/a/}');
Error('\P{	-Line_SEPARATOR/a/}');
Expect(1, 8232, '\p{lineseparator}', "");
Expect(0, 8232, '\p{^lineseparator}', "");
Expect(0, 8232, '\P{lineseparator}', "");
Expect(1, 8232, '\P{^lineseparator}', "");
Expect(0, 8233, '\p{lineseparator}', "");
Expect(1, 8233, '\p{^lineseparator}', "");
Expect(1, 8233, '\P{lineseparator}', "");
Expect(0, 8233, '\P{^lineseparator}', "");
Expect(1, 8232, '\p{	_line_Separator}', "");
Expect(0, 8232, '\p{^	_line_Separator}', "");
Expect(0, 8232, '\P{	_line_Separator}', "");
Expect(1, 8232, '\P{^	_line_Separator}', "");
Expect(0, 8233, '\p{	_line_Separator}', "");
Expect(1, 8233, '\p{^	_line_Separator}', "");
Expect(1, 8233, '\P{	_line_Separator}', "");
Expect(0, 8233, '\P{^	_line_Separator}', "");
Error('\p{	 Is_LINE_Separator:=}');
Error('\P{	 Is_LINE_Separator:=}');
Expect(1, 8232, '\p{islineseparator}', "");
Expect(0, 8232, '\p{^islineseparator}', "");
Expect(0, 8232, '\P{islineseparator}', "");
Expect(1, 8232, '\P{^islineseparator}', "");
Expect(0, 8233, '\p{islineseparator}', "");
Expect(1, 8233, '\p{^islineseparator}', "");
Expect(1, 8233, '\P{islineseparator}', "");
Expect(0, 8233, '\P{^islineseparator}', "");
Expect(1, 8232, '\p{ 	is_line_Separator}', "");
Expect(0, 8232, '\p{^ 	is_line_Separator}', "");
Expect(0, 8232, '\P{ 	is_line_Separator}', "");
Expect(1, 8232, '\P{^ 	is_line_Separator}', "");
Expect(0, 8233, '\p{ 	is_line_Separator}', "");
Expect(1, 8233, '\p{^ 	is_line_Separator}', "");
Expect(1, 8233, '\P{ 	is_line_Separator}', "");
Expect(0, 8233, '\P{^ 	is_line_Separator}', "");
Error('\p{/a/	_Zl}');
Error('\P{/a/	_Zl}');
Expect(1, 8232, '\p{zl}', "");
Expect(0, 8232, '\p{^zl}', "");
Expect(0, 8232, '\P{zl}', "");
Expect(1, 8232, '\P{^zl}', "");
Expect(0, 8233, '\p{zl}', "");
Expect(1, 8233, '\p{^zl}', "");
Expect(1, 8233, '\P{zl}', "");
Expect(0, 8233, '\P{^zl}', "");
Expect(1, 8232, '\p{ Zl}', "");
Expect(0, 8232, '\p{^ Zl}', "");
Expect(0, 8232, '\P{ Zl}', "");
Expect(1, 8232, '\P{^ Zl}', "");
Expect(0, 8233, '\p{ Zl}', "");
Expect(1, 8233, '\p{^ Zl}', "");
Expect(1, 8233, '\P{ Zl}', "");
Expect(0, 8233, '\P{^ Zl}', "");
Error('\p{/a/ is_ZL}');
Error('\P{/a/ is_ZL}');
Expect(1, 8232, '\p{iszl}', "");
Expect(0, 8232, '\p{^iszl}', "");
Expect(0, 8232, '\P{iszl}', "");
Expect(1, 8232, '\P{^iszl}', "");
Expect(0, 8233, '\p{iszl}', "");
Expect(1, 8233, '\p{^iszl}', "");
Expect(1, 8233, '\P{iszl}', "");
Expect(0, 8233, '\P{^iszl}', "");
Expect(1, 8232, '\p{	_is_Zl}', "");
Expect(0, 8232, '\p{^	_is_Zl}', "");
Expect(0, 8232, '\P{	_is_Zl}', "");
Expect(1, 8232, '\P{^	_is_Zl}', "");
Expect(0, 8233, '\p{	_is_Zl}', "");
Expect(1, 8233, '\p{^	_is_Zl}', "");
Expect(1, 8233, '\P{	_is_Zl}', "");
Expect(0, 8233, '\P{^	_is_Zl}', "");
Error('\p{-In_meroitic_HIEROGLYPHS/a/}');
Error('\P{-In_meroitic_HIEROGLYPHS/a/}');
Expect(1, 67999, '\p{inmeroitichieroglyphs}', "");
Expect(0, 67999, '\p{^inmeroitichieroglyphs}', "");
Expect(0, 67999, '\P{inmeroitichieroglyphs}', "");
Expect(1, 67999, '\P{^inmeroitichieroglyphs}', "");
Expect(0, 68000, '\p{inmeroitichieroglyphs}', "");
Expect(1, 68000, '\p{^inmeroitichieroglyphs}', "");
Expect(1, 68000, '\P{inmeroitichieroglyphs}', "");
Expect(0, 68000, '\P{^inmeroitichieroglyphs}', "");
Expect(1, 67999, '\p{-IN_Meroitic_HIEROGLYPHS}', "");
Expect(0, 67999, '\p{^-IN_Meroitic_HIEROGLYPHS}', "");
Expect(0, 67999, '\P{-IN_Meroitic_HIEROGLYPHS}', "");
Expect(1, 67999, '\P{^-IN_Meroitic_HIEROGLYPHS}', "");
Expect(0, 68000, '\p{-IN_Meroitic_HIEROGLYPHS}', "");
Expect(1, 68000, '\p{^-IN_Meroitic_HIEROGLYPHS}', "");
Expect(1, 68000, '\P{-IN_Meroitic_HIEROGLYPHS}', "");
Expect(0, 68000, '\P{^-IN_Meroitic_HIEROGLYPHS}', "");
Error('\p{:=_IN_Meroitic_HIEROGLYPHS}');
Error('\P{:=_IN_Meroitic_HIEROGLYPHS}');
Expect(1, 67999, '\p{_IN_Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\p{^_IN_Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\P{_IN_Meroitic_Hieroglyphs}', "");
Expect(1, 67999, '\P{^_IN_Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\p{_IN_Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\p{^_IN_Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\P{_IN_Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\P{^_IN_Meroitic_Hieroglyphs}', "");
Error('\p{/a/Ol_CHIKI}');
Error('\P{/a/Ol_CHIKI}');
Expect(1, 7295, '\p{olchiki}', "");
Expect(0, 7295, '\p{^olchiki}', "");
Expect(0, 7295, '\P{olchiki}', "");
Expect(1, 7295, '\P{^olchiki}', "");
Expect(0, 7296, '\p{olchiki}', "");
Expect(1, 7296, '\p{^olchiki}', "");
Expect(1, 7296, '\P{olchiki}', "");
Expect(0, 7296, '\P{^olchiki}', "");
Expect(1, 7295, '\p{	-Ol_Chiki}', "");
Expect(0, 7295, '\p{^	-Ol_Chiki}', "");
Expect(0, 7295, '\P{	-Ol_Chiki}', "");
Expect(1, 7295, '\P{^	-Ol_Chiki}', "");
Expect(0, 7296, '\p{	-Ol_Chiki}', "");
Expect(1, 7296, '\p{^	-Ol_Chiki}', "");
Expect(1, 7296, '\P{	-Ol_Chiki}', "");
Expect(0, 7296, '\P{^	-Ol_Chiki}', "");
Error('\p{ /a/IS_ol_CHIKI}');
Error('\P{ /a/IS_ol_CHIKI}');
Expect(1, 7295, '\p{isolchiki}', "");
Expect(0, 7295, '\p{^isolchiki}', "");
Expect(0, 7295, '\P{isolchiki}', "");
Expect(1, 7295, '\P{^isolchiki}', "");
Expect(0, 7296, '\p{isolchiki}', "");
Expect(1, 7296, '\p{^isolchiki}', "");
Expect(1, 7296, '\P{isolchiki}', "");
Expect(0, 7296, '\P{^isolchiki}', "");
Expect(1, 7295, '\p{-Is_ol_chiki}', "");
Expect(0, 7295, '\p{^-Is_ol_chiki}', "");
Expect(0, 7295, '\P{-Is_ol_chiki}', "");
Expect(1, 7295, '\P{^-Is_ol_chiki}', "");
Expect(0, 7296, '\p{-Is_ol_chiki}', "");
Expect(1, 7296, '\p{^-Is_ol_chiki}', "");
Expect(1, 7296, '\P{-Is_ol_chiki}', "");
Expect(0, 7296, '\P{^-Is_ol_chiki}', "");
Error('\p{-	OLCK/a/}');
Error('\P{-	OLCK/a/}');
Expect(1, 7295, '\p{olck}', "");
Expect(0, 7295, '\p{^olck}', "");
Expect(0, 7295, '\P{olck}', "");
Expect(1, 7295, '\P{^olck}', "");
Expect(0, 7296, '\p{olck}', "");
Expect(1, 7296, '\p{^olck}', "");
Expect(1, 7296, '\P{olck}', "");
Expect(0, 7296, '\P{^olck}', "");
Expect(1, 7295, '\p{	_Olck}', "");
Expect(0, 7295, '\p{^	_Olck}', "");
Expect(0, 7295, '\P{	_Olck}', "");
Expect(1, 7295, '\P{^	_Olck}', "");
Expect(0, 7296, '\p{	_Olck}', "");
Expect(1, 7296, '\p{^	_Olck}', "");
Expect(1, 7296, '\P{	_Olck}', "");
Expect(0, 7296, '\P{^	_Olck}', "");
Error('\p{ -Is_olck:=}');
Error('\P{ -Is_olck:=}');
Expect(1, 7295, '\p{isolck}', "");
Expect(0, 7295, '\p{^isolck}', "");
Expect(0, 7295, '\P{isolck}', "");
Expect(1, 7295, '\P{^isolck}', "");
Expect(0, 7296, '\p{isolck}', "");
Expect(1, 7296, '\p{^isolck}', "");
Expect(1, 7296, '\P{isolck}', "");
Expect(0, 7296, '\P{^isolck}', "");
Expect(1, 7295, '\p{	_Is_OLCK}', "");
Expect(0, 7295, '\p{^	_Is_OLCK}', "");
Expect(0, 7295, '\P{	_Is_OLCK}', "");
Expect(1, 7295, '\P{^	_Is_OLCK}', "");
Expect(0, 7296, '\p{	_Is_OLCK}', "");
Expect(1, 7296, '\p{^	_Is_OLCK}', "");
Expect(1, 7296, '\P{	_Is_OLCK}', "");
Expect(0, 7296, '\P{^	_Is_OLCK}', "");
Error('\p{:=	bamum_Supplement}');
Error('\P{:=	bamum_Supplement}');
Expect(1, 92735, '\p{bamumsupplement}', "");
Expect(0, 92735, '\p{^bamumsupplement}', "");
Expect(0, 92735, '\P{bamumsupplement}', "");
Expect(1, 92735, '\P{^bamumsupplement}', "");
Expect(0, 92736, '\p{bamumsupplement}', "");
Expect(1, 92736, '\p{^bamumsupplement}', "");
Expect(1, 92736, '\P{bamumsupplement}', "");
Expect(0, 92736, '\P{^bamumsupplement}', "");
Expect(1, 92735, '\p{bamum_Supplement}', "");
Expect(0, 92735, '\p{^bamum_Supplement}', "");
Expect(0, 92735, '\P{bamum_Supplement}', "");
Expect(1, 92735, '\P{^bamum_Supplement}', "");
Expect(0, 92736, '\p{bamum_Supplement}', "");
Expect(1, 92736, '\p{^bamum_Supplement}', "");
Expect(1, 92736, '\P{bamum_Supplement}', "");
Expect(0, 92736, '\P{^bamum_Supplement}', "");
Error('\p{- is_bamum_SUPPLEMENT:=}');
Error('\P{- is_bamum_SUPPLEMENT:=}');
Expect(1, 92735, '\p{isbamumsupplement}', "");
Expect(0, 92735, '\p{^isbamumsupplement}', "");
Expect(0, 92735, '\P{isbamumsupplement}', "");
Expect(1, 92735, '\P{^isbamumsupplement}', "");
Expect(0, 92736, '\p{isbamumsupplement}', "");
Expect(1, 92736, '\p{^isbamumsupplement}', "");
Expect(1, 92736, '\P{isbamumsupplement}', "");
Expect(0, 92736, '\P{^isbamumsupplement}', "");
Expect(1, 92735, '\p{	_is_Bamum_supplement}', "");
Expect(0, 92735, '\p{^	_is_Bamum_supplement}', "");
Expect(0, 92735, '\P{	_is_Bamum_supplement}', "");
Expect(1, 92735, '\P{^	_is_Bamum_supplement}', "");
Expect(0, 92736, '\p{	_is_Bamum_supplement}', "");
Expect(1, 92736, '\p{^	_is_Bamum_supplement}', "");
Expect(1, 92736, '\P{	_is_Bamum_supplement}', "");
Expect(0, 92736, '\P{^	_is_Bamum_supplement}', "");
Error('\p{/a/ -In_Bamum_SUPPLEMENT}');
Error('\P{/a/ -In_Bamum_SUPPLEMENT}');
Expect(1, 92735, '\p{inbamumsupplement}', "");
Expect(0, 92735, '\p{^inbamumsupplement}', "");
Expect(0, 92735, '\P{inbamumsupplement}', "");
Expect(1, 92735, '\P{^inbamumsupplement}', "");
Expect(0, 92736, '\p{inbamumsupplement}', "");
Expect(1, 92736, '\p{^inbamumsupplement}', "");
Expect(1, 92736, '\P{inbamumsupplement}', "");
Expect(0, 92736, '\P{^inbamumsupplement}', "");
Expect(1, 92735, '\p{- In_Bamum_SUPPLEMENT}', "");
Expect(0, 92735, '\p{^- In_Bamum_SUPPLEMENT}', "");
Expect(0, 92735, '\P{- In_Bamum_SUPPLEMENT}', "");
Expect(1, 92735, '\P{^- In_Bamum_SUPPLEMENT}', "");
Expect(0, 92736, '\p{- In_Bamum_SUPPLEMENT}', "");
Expect(1, 92736, '\p{^- In_Bamum_SUPPLEMENT}', "");
Expect(1, 92736, '\P{- In_Bamum_SUPPLEMENT}', "");
Expect(0, 92736, '\P{^- In_Bamum_SUPPLEMENT}', "");
Error('\p{:=__bamum_sup}');
Error('\P{:=__bamum_sup}');
Expect(1, 92735, '\p{bamumsup}', "");
Expect(0, 92735, '\p{^bamumsup}', "");
Expect(0, 92735, '\P{bamumsup}', "");
Expect(1, 92735, '\P{^bamumsup}', "");
Expect(0, 92736, '\p{bamumsup}', "");
Expect(1, 92736, '\p{^bamumsup}', "");
Expect(1, 92736, '\P{bamumsup}', "");
Expect(0, 92736, '\P{^bamumsup}', "");
Expect(1, 92735, '\p{ -bamum_Sup}', "");
Expect(0, 92735, '\p{^ -bamum_Sup}', "");
Expect(0, 92735, '\P{ -bamum_Sup}', "");
Expect(1, 92735, '\P{^ -bamum_Sup}', "");
Expect(0, 92736, '\p{ -bamum_Sup}', "");
Expect(1, 92736, '\p{^ -bamum_Sup}', "");
Expect(1, 92736, '\P{ -bamum_Sup}', "");
Expect(0, 92736, '\P{^ -bamum_Sup}', "");
Error('\p{:=IS_BAMUM_sup}');
Error('\P{:=IS_BAMUM_sup}');
Expect(1, 92735, '\p{isbamumsup}', "");
Expect(0, 92735, '\p{^isbamumsup}', "");
Expect(0, 92735, '\P{isbamumsup}', "");
Expect(1, 92735, '\P{^isbamumsup}', "");
Expect(0, 92736, '\p{isbamumsup}', "");
Expect(1, 92736, '\p{^isbamumsup}', "");
Expect(1, 92736, '\P{isbamumsup}', "");
Expect(0, 92736, '\P{^isbamumsup}', "");
Expect(1, 92735, '\p{--Is_bamum_SUP}', "");
Expect(0, 92735, '\p{^--Is_bamum_SUP}', "");
Expect(0, 92735, '\P{--Is_bamum_SUP}', "");
Expect(1, 92735, '\P{^--Is_bamum_SUP}', "");
Expect(0, 92736, '\p{--Is_bamum_SUP}', "");
Expect(1, 92736, '\p{^--Is_bamum_SUP}', "");
Expect(1, 92736, '\P{--Is_bamum_SUP}', "");
Expect(0, 92736, '\P{^--Is_bamum_SUP}', "");
Error('\p{	/a/IN_Bamum_sup}');
Error('\P{	/a/IN_Bamum_sup}');
Expect(1, 92735, '\p{inbamumsup}', "");
Expect(0, 92735, '\p{^inbamumsup}', "");
Expect(0, 92735, '\P{inbamumsup}', "");
Expect(1, 92735, '\P{^inbamumsup}', "");
Expect(0, 92736, '\p{inbamumsup}', "");
Expect(1, 92736, '\p{^inbamumsup}', "");
Expect(1, 92736, '\P{inbamumsup}', "");
Expect(0, 92736, '\P{^inbamumsup}', "");
Expect(1, 92735, '\p{  IN_Bamum_Sup}', "");
Expect(0, 92735, '\p{^  IN_Bamum_Sup}', "");
Expect(0, 92735, '\P{  IN_Bamum_Sup}', "");
Expect(1, 92735, '\P{^  IN_Bamum_Sup}', "");
Expect(0, 92736, '\p{  IN_Bamum_Sup}', "");
Expect(1, 92736, '\p{^  IN_Bamum_Sup}', "");
Expect(1, 92736, '\P{  IN_Bamum_Sup}', "");
Expect(0, 92736, '\P{^  IN_Bamum_Sup}', "");
Error('\p{:=		TAI_viet}');
Error('\P{:=		TAI_viet}');
Expect(1, 43743, '\p{taiviet}', "");
Expect(0, 43743, '\p{^taiviet}', "");
Expect(0, 43743, '\P{taiviet}', "");
Expect(1, 43743, '\P{^taiviet}', "");
Expect(0, 43744, '\p{taiviet}', "");
Expect(1, 43744, '\p{^taiviet}', "");
Expect(1, 43744, '\P{taiviet}', "");
Expect(0, 43744, '\P{^taiviet}', "");
Expect(1, 43743, '\p{_ TAI_Viet}', "");
Expect(0, 43743, '\p{^_ TAI_Viet}', "");
Expect(0, 43743, '\P{_ TAI_Viet}', "");
Expect(1, 43743, '\P{^_ TAI_Viet}', "");
Expect(0, 43744, '\p{_ TAI_Viet}', "");
Expect(1, 43744, '\p{^_ TAI_Viet}', "");
Expect(1, 43744, '\P{_ TAI_Viet}', "");
Expect(0, 43744, '\P{^_ TAI_Viet}', "");
Error('\p{:=_Is_TAI_Viet}');
Error('\P{:=_Is_TAI_Viet}');
Expect(1, 43743, '\p{istaiviet}', "");
Expect(0, 43743, '\p{^istaiviet}', "");
Expect(0, 43743, '\P{istaiviet}', "");
Expect(1, 43743, '\P{^istaiviet}', "");
Expect(0, 43744, '\p{istaiviet}', "");
Expect(1, 43744, '\p{^istaiviet}', "");
Expect(1, 43744, '\P{istaiviet}', "");
Expect(0, 43744, '\P{^istaiviet}', "");
Expect(1, 43743, '\p{	is_Tai_Viet}', "");
Expect(0, 43743, '\p{^	is_Tai_Viet}', "");
Expect(0, 43743, '\P{	is_Tai_Viet}', "");
Expect(1, 43743, '\P{^	is_Tai_Viet}', "");
Expect(0, 43744, '\p{	is_Tai_Viet}', "");
Expect(1, 43744, '\p{^	is_Tai_Viet}', "");
Expect(1, 43744, '\P{	is_Tai_Viet}', "");
Expect(0, 43744, '\P{^	is_Tai_Viet}', "");
Error('\p{-_tavt:=}');
Error('\P{-_tavt:=}');
Expect(1, 43743, '\p{tavt}', "");
Expect(0, 43743, '\p{^tavt}', "");
Expect(0, 43743, '\P{tavt}', "");
Expect(1, 43743, '\P{^tavt}', "");
Expect(0, 43744, '\p{tavt}', "");
Expect(1, 43744, '\p{^tavt}', "");
Expect(1, 43744, '\P{tavt}', "");
Expect(0, 43744, '\P{^tavt}', "");
Expect(1, 43743, '\p{-tavt}', "");
Expect(0, 43743, '\p{^-tavt}', "");
Expect(0, 43743, '\P{-tavt}', "");
Expect(1, 43743, '\P{^-tavt}', "");
Expect(0, 43744, '\p{-tavt}', "");
Expect(1, 43744, '\p{^-tavt}', "");
Expect(1, 43744, '\P{-tavt}', "");
Expect(0, 43744, '\P{^-tavt}', "");
Error('\p{ :=Is_TAVT}');
Error('\P{ :=Is_TAVT}');
Expect(1, 43743, '\p{istavt}', "");
Expect(0, 43743, '\p{^istavt}', "");
Expect(0, 43743, '\P{istavt}', "");
Expect(1, 43743, '\P{^istavt}', "");
Expect(0, 43744, '\p{istavt}', "");
Expect(1, 43744, '\p{^istavt}', "");
Expect(1, 43744, '\P{istavt}', "");
Expect(0, 43744, '\P{^istavt}', "");
Expect(1, 43743, '\p{	 Is_TAVT}', "");
Expect(0, 43743, '\p{^	 Is_TAVT}', "");
Expect(0, 43743, '\P{	 Is_TAVT}', "");
Expect(1, 43743, '\P{^	 Is_TAVT}', "");
Expect(0, 43744, '\p{	 Is_TAVT}', "");
Expect(1, 43744, '\p{^	 Is_TAVT}', "");
Expect(1, 43744, '\P{	 Is_TAVT}', "");
Expect(0, 43744, '\P{^	 Is_TAVT}', "");
Error('\p{	thai/a/}');
Error('\P{	thai/a/}');
Expect(1, 3675, '\p{thai}', "");
Expect(0, 3675, '\p{^thai}', "");
Expect(0, 3675, '\P{thai}', "");
Expect(1, 3675, '\P{^thai}', "");
Expect(0, 3676, '\p{thai}', "");
Expect(1, 3676, '\p{^thai}', "");
Expect(1, 3676, '\P{thai}', "");
Expect(0, 3676, '\P{^thai}', "");
Expect(1, 3675, '\p{	 Thai}', "");
Expect(0, 3675, '\p{^	 Thai}', "");
Expect(0, 3675, '\P{	 Thai}', "");
Expect(1, 3675, '\P{^	 Thai}', "");
Expect(0, 3676, '\p{	 Thai}', "");
Expect(1, 3676, '\p{^	 Thai}', "");
Expect(1, 3676, '\P{	 Thai}', "");
Expect(0, 3676, '\P{^	 Thai}', "");
Error('\p{	:=Is_THAI}');
Error('\P{	:=Is_THAI}');
Expect(1, 3675, '\p{isthai}', "");
Expect(0, 3675, '\p{^isthai}', "");
Expect(0, 3675, '\P{isthai}', "");
Expect(1, 3675, '\P{^isthai}', "");
Expect(0, 3676, '\p{isthai}', "");
Expect(1, 3676, '\p{^isthai}', "");
Expect(1, 3676, '\P{isthai}', "");
Expect(0, 3676, '\P{^isthai}', "");
Expect(1, 3675, '\p{ Is_THAI}', "");
Expect(0, 3675, '\p{^ Is_THAI}', "");
Expect(0, 3675, '\P{ Is_THAI}', "");
Expect(1, 3675, '\P{^ Is_THAI}', "");
Expect(0, 3676, '\p{ Is_THAI}', "");
Expect(1, 3676, '\p{^ Is_THAI}', "");
Expect(1, 3676, '\P{ Is_THAI}', "");
Expect(0, 3676, '\P{^ Is_THAI}', "");
Error('\p{-ASCII:=}');
Error('\P{-ASCII:=}');
Expect(1, 127, '\p{ascii}', "");
Expect(0, 127, '\p{^ascii}', "");
Expect(0, 127, '\P{ascii}', "");
Expect(1, 127, '\P{^ascii}', "");
Expect(0, 128, '\p{ascii}', "");
Expect(1, 128, '\p{^ascii}', "");
Expect(1, 128, '\P{ascii}', "");
Expect(0, 128, '\P{^ascii}', "");
Expect(1, 127, '\p{ ASCII}', "");
Expect(0, 127, '\p{^ ASCII}', "");
Expect(0, 127, '\P{ ASCII}', "");
Expect(1, 127, '\P{^ ASCII}', "");
Expect(0, 128, '\p{ ASCII}', "");
Expect(1, 128, '\p{^ ASCII}', "");
Expect(1, 128, '\P{ ASCII}', "");
Expect(0, 128, '\P{^ ASCII}', "");
Error('\p{  is_ASCII:=}');
Error('\P{  is_ASCII:=}');
Expect(1, 127, '\p{isascii}', "");
Expect(0, 127, '\p{^isascii}', "");
Expect(0, 127, '\P{isascii}', "");
Expect(1, 127, '\P{^isascii}', "");
Expect(0, 128, '\p{isascii}', "");
Expect(1, 128, '\p{^isascii}', "");
Expect(1, 128, '\P{isascii}', "");
Expect(0, 128, '\P{^isascii}', "");
Expect(1, 127, '\p{	 Is_ASCII}', "");
Expect(0, 127, '\p{^	 Is_ASCII}', "");
Expect(0, 127, '\P{	 Is_ASCII}', "");
Expect(1, 127, '\P{^	 Is_ASCII}', "");
Expect(0, 128, '\p{	 Is_ASCII}', "");
Expect(1, 128, '\p{^	 Is_ASCII}', "");
Expect(1, 128, '\P{	 Is_ASCII}', "");
Expect(0, 128, '\P{^	 Is_ASCII}', "");
Error('\p{	 BASIC_LATIN:=}');
Error('\P{	 BASIC_LATIN:=}');
Expect(1, 127, '\p{basiclatin}', "");
Expect(0, 127, '\p{^basiclatin}', "");
Expect(0, 127, '\P{basiclatin}', "");
Expect(1, 127, '\P{^basiclatin}', "");
Expect(0, 128, '\p{basiclatin}', "");
Expect(1, 128, '\p{^basiclatin}', "");
Expect(1, 128, '\P{basiclatin}', "");
Expect(0, 128, '\P{^basiclatin}', "");
Expect(1, 127, '\p{	_BASIC_Latin}', "");
Expect(0, 127, '\p{^	_BASIC_Latin}', "");
Expect(0, 127, '\P{	_BASIC_Latin}', "");
Expect(1, 127, '\P{^	_BASIC_Latin}', "");
Expect(0, 128, '\p{	_BASIC_Latin}', "");
Expect(1, 128, '\p{^	_BASIC_Latin}', "");
Expect(1, 128, '\P{	_BASIC_Latin}', "");
Expect(0, 128, '\P{^	_BASIC_Latin}', "");
Error('\p{_/a/is_basic_LATIN}');
Error('\P{_/a/is_basic_LATIN}');
Expect(1, 127, '\p{isbasiclatin}', "");
Expect(0, 127, '\p{^isbasiclatin}', "");
Expect(0, 127, '\P{isbasiclatin}', "");
Expect(1, 127, '\P{^isbasiclatin}', "");
Expect(0, 128, '\p{isbasiclatin}', "");
Expect(1, 128, '\p{^isbasiclatin}', "");
Expect(1, 128, '\P{isbasiclatin}', "");
Expect(0, 128, '\P{^isbasiclatin}', "");
Expect(1, 127, '\p{__is_basic_latin}', "");
Expect(0, 127, '\p{^__is_basic_latin}', "");
Expect(0, 127, '\P{__is_basic_latin}', "");
Expect(1, 127, '\P{^__is_basic_latin}', "");
Expect(0, 128, '\p{__is_basic_latin}', "");
Expect(1, 128, '\p{^__is_basic_latin}', "");
Expect(1, 128, '\P{__is_basic_latin}', "");
Expect(0, 128, '\P{^__is_basic_latin}', "");
Error('\p{ In_basic_latin:=}');
Error('\P{ In_basic_latin:=}');
Expect(1, 127, '\p{inbasiclatin}', "");
Expect(0, 127, '\p{^inbasiclatin}', "");
Expect(0, 127, '\P{inbasiclatin}', "");
Expect(1, 127, '\P{^inbasiclatin}', "");
Expect(0, 128, '\p{inbasiclatin}', "");
Expect(1, 128, '\p{^inbasiclatin}', "");
Expect(1, 128, '\P{inbasiclatin}', "");
Expect(0, 128, '\P{^inbasiclatin}', "");
Expect(1, 127, '\p{-IN_BASIC_LATIN}', "");
Expect(0, 127, '\p{^-IN_BASIC_LATIN}', "");
Expect(0, 127, '\P{-IN_BASIC_LATIN}', "");
Expect(1, 127, '\P{^-IN_BASIC_LATIN}', "");
Expect(0, 128, '\p{-IN_BASIC_LATIN}', "");
Expect(1, 128, '\p{^-IN_BASIC_LATIN}', "");
Expect(1, 128, '\P{-IN_BASIC_LATIN}', "");
Expect(0, 128, '\P{^-IN_BASIC_LATIN}', "");
Error('\p{/a/In_ASCII}');
Error('\P{/a/In_ASCII}');
Expect(1, 127, '\p{inascii}', "");
Expect(0, 127, '\p{^inascii}', "");
Expect(0, 127, '\P{inascii}', "");
Expect(1, 127, '\P{^inascii}', "");
Expect(0, 128, '\p{inascii}', "");
Expect(1, 128, '\p{^inascii}', "");
Expect(1, 128, '\P{inascii}', "");
Expect(0, 128, '\P{^inascii}', "");
Expect(1, 127, '\p{ _In_ASCII}', "");
Expect(0, 127, '\p{^ _In_ASCII}', "");
Expect(0, 127, '\P{ _In_ASCII}', "");
Expect(1, 127, '\P{^ _In_ASCII}', "");
Expect(0, 128, '\p{ _In_ASCII}', "");
Expect(1, 128, '\p{^ _In_ASCII}', "");
Expect(1, 128, '\P{ _In_ASCII}', "");
Expect(0, 128, '\P{^ _In_ASCII}', "");
Error('\p{/a/  Yi_syllables}');
Error('\P{/a/  Yi_syllables}');
Expect(1, 42127, '\p{yisyllables}', "");
Expect(0, 42127, '\p{^yisyllables}', "");
Expect(0, 42127, '\P{yisyllables}', "");
Expect(1, 42127, '\P{^yisyllables}', "");
Expect(0, 42128, '\p{yisyllables}', "");
Expect(1, 42128, '\p{^yisyllables}', "");
Expect(1, 42128, '\P{yisyllables}', "");
Expect(0, 42128, '\P{^yisyllables}', "");
Expect(1, 42127, '\p{_YI_Syllables}', "");
Expect(0, 42127, '\p{^_YI_Syllables}', "");
Expect(0, 42127, '\P{_YI_Syllables}', "");
Expect(1, 42127, '\P{^_YI_Syllables}', "");
Expect(0, 42128, '\p{_YI_Syllables}', "");
Expect(1, 42128, '\p{^_YI_Syllables}', "");
Expect(1, 42128, '\P{_YI_Syllables}', "");
Expect(0, 42128, '\P{^_YI_Syllables}', "");
Error('\p{/a/IS_yi_syllables}');
Error('\P{/a/IS_yi_syllables}');
Expect(1, 42127, '\p{isyisyllables}', "");
Expect(0, 42127, '\p{^isyisyllables}', "");
Expect(0, 42127, '\P{isyisyllables}', "");
Expect(1, 42127, '\P{^isyisyllables}', "");
Expect(0, 42128, '\p{isyisyllables}', "");
Expect(1, 42128, '\p{^isyisyllables}', "");
Expect(1, 42128, '\P{isyisyllables}', "");
Expect(0, 42128, '\P{^isyisyllables}', "");
Expect(1, 42127, '\p{-Is_Yi_Syllables}', "");
Expect(0, 42127, '\p{^-Is_Yi_Syllables}', "");
Expect(0, 42127, '\P{-Is_Yi_Syllables}', "");
Expect(1, 42127, '\P{^-Is_Yi_Syllables}', "");
Expect(0, 42128, '\p{-Is_Yi_Syllables}', "");
Expect(1, 42128, '\p{^-Is_Yi_Syllables}', "");
Expect(1, 42128, '\P{-Is_Yi_Syllables}', "");
Expect(0, 42128, '\P{^-Is_Yi_Syllables}', "");
Error('\p{/a/-_in_Yi_Syllables}');
Error('\P{/a/-_in_Yi_Syllables}');
Expect(1, 42127, '\p{inyisyllables}', "");
Expect(0, 42127, '\p{^inyisyllables}', "");
Expect(0, 42127, '\P{inyisyllables}', "");
Expect(1, 42127, '\P{^inyisyllables}', "");
Expect(0, 42128, '\p{inyisyllables}', "");
Expect(1, 42128, '\p{^inyisyllables}', "");
Expect(1, 42128, '\P{inyisyllables}', "");
Expect(0, 42128, '\P{^inyisyllables}', "");
Expect(1, 42127, '\p{_	In_yi_SYLLABLES}', "");
Expect(0, 42127, '\p{^_	In_yi_SYLLABLES}', "");
Expect(0, 42127, '\P{_	In_yi_SYLLABLES}', "");
Expect(1, 42127, '\P{^_	In_yi_SYLLABLES}', "");
Expect(0, 42128, '\p{_	In_yi_SYLLABLES}', "");
Expect(1, 42128, '\p{^_	In_yi_SYLLABLES}', "");
Expect(1, 42128, '\P{_	In_yi_SYLLABLES}', "");
Expect(0, 42128, '\P{^_	In_yi_SYLLABLES}', "");
Error('\p{	_Punct:=}');
Error('\P{	_Punct:=}');
Expect(1, 74867, '\p{punct}', "");
Expect(0, 74867, '\p{^punct}', "");
Expect(0, 74867, '\P{punct}', "");
Expect(1, 74867, '\P{^punct}', "");
Expect(0, 74868, '\p{punct}', "");
Expect(1, 74868, '\p{^punct}', "");
Expect(1, 74868, '\P{punct}', "");
Expect(0, 74868, '\P{^punct}', "");
Expect(1, 74867, '\p{_ Punct}', "");
Expect(0, 74867, '\p{^_ Punct}', "");
Expect(0, 74867, '\P{_ Punct}', "");
Expect(1, 74867, '\P{^_ Punct}', "");
Expect(0, 74868, '\p{_ Punct}', "");
Expect(1, 74868, '\p{^_ Punct}', "");
Expect(1, 74868, '\P{_ Punct}', "");
Expect(0, 74868, '\P{^_ Punct}', "");
Error('\p{/a/_IS_punct}');
Error('\P{/a/_IS_punct}');
Expect(1, 74867, '\p{ispunct}', "");
Expect(0, 74867, '\p{^ispunct}', "");
Expect(0, 74867, '\P{ispunct}', "");
Expect(1, 74867, '\P{^ispunct}', "");
Expect(0, 74868, '\p{ispunct}', "");
Expect(1, 74868, '\p{^ispunct}', "");
Expect(1, 74868, '\P{ispunct}', "");
Expect(0, 74868, '\P{^ispunct}', "");
Expect(1, 74867, '\p{ is_Punct}', "");
Expect(0, 74867, '\p{^ is_Punct}', "");
Expect(0, 74867, '\P{ is_Punct}', "");
Expect(1, 74867, '\P{^ is_Punct}', "");
Expect(0, 74868, '\p{ is_Punct}', "");
Expect(1, 74868, '\p{^ is_Punct}', "");
Expect(1, 74868, '\P{ is_Punct}', "");
Expect(0, 74868, '\P{^ is_Punct}', "");
Error('\p{ /a/Punctuation}');
Error('\P{ /a/Punctuation}');
Expect(1, 74867, '\p{punctuation}', "");
Expect(0, 74867, '\p{^punctuation}', "");
Expect(0, 74867, '\P{punctuation}', "");
Expect(1, 74867, '\P{^punctuation}', "");
Expect(0, 74868, '\p{punctuation}', "");
Expect(1, 74868, '\p{^punctuation}', "");
Expect(1, 74868, '\P{punctuation}', "");
Expect(0, 74868, '\P{^punctuation}', "");
Expect(1, 74867, '\p{_Punctuation}', "");
Expect(0, 74867, '\p{^_Punctuation}', "");
Expect(0, 74867, '\P{_Punctuation}', "");
Expect(1, 74867, '\P{^_Punctuation}', "");
Expect(0, 74868, '\p{_Punctuation}', "");
Expect(1, 74868, '\p{^_Punctuation}', "");
Expect(1, 74868, '\P{_Punctuation}', "");
Expect(0, 74868, '\P{^_Punctuation}', "");
Error('\p{:=		IS_punctuation}');
Error('\P{:=		IS_punctuation}');
Expect(1, 74867, '\p{ispunctuation}', "");
Expect(0, 74867, '\p{^ispunctuation}', "");
Expect(0, 74867, '\P{ispunctuation}', "");
Expect(1, 74867, '\P{^ispunctuation}', "");
Expect(0, 74868, '\p{ispunctuation}', "");
Expect(1, 74868, '\p{^ispunctuation}', "");
Expect(1, 74868, '\P{ispunctuation}', "");
Expect(0, 74868, '\P{^ispunctuation}', "");
Expect(1, 74867, '\p{		IS_punctuation}', "");
Expect(0, 74867, '\p{^		IS_punctuation}', "");
Expect(0, 74867, '\P{		IS_punctuation}', "");
Expect(1, 74867, '\P{^		IS_punctuation}', "");
Expect(0, 74868, '\p{		IS_punctuation}', "");
Expect(1, 74868, '\p{^		IS_punctuation}', "");
Expect(1, 74868, '\P{		IS_punctuation}', "");
Expect(0, 74868, '\P{^		IS_punctuation}', "");
Error('\p{_/a/P}');
Error('\P{_/a/P}');
Expect(1, 74867, '\p{p}', "");
Expect(0, 74867, '\p{^p}', "");
Expect(0, 74867, '\P{p}', "");
Expect(1, 74867, '\P{^p}', "");
Expect(0, 74868, '\p{p}', "");
Expect(1, 74868, '\p{^p}', "");
Expect(1, 74868, '\P{p}', "");
Expect(0, 74868, '\P{^p}', "");
Expect(1, 74867, '\p{-p}', "");
Expect(0, 74867, '\p{^-p}', "");
Expect(0, 74867, '\P{-p}', "");
Expect(1, 74867, '\P{^-p}', "");
Expect(0, 74868, '\p{-p}', "");
Expect(1, 74868, '\p{^-p}', "");
Expect(1, 74868, '\P{-p}', "");
Expect(0, 74868, '\P{^-p}', "");
Error('\p{-:=Is_P}');
Error('\P{-:=Is_P}');
Expect(1, 74867, '\p{isp}', "");
Expect(0, 74867, '\p{^isp}', "");
Expect(0, 74867, '\P{isp}', "");
Expect(1, 74867, '\P{^isp}', "");
Expect(0, 74868, '\p{isp}', "");
Expect(1, 74868, '\p{^isp}', "");
Expect(1, 74868, '\P{isp}', "");
Expect(0, 74868, '\P{^isp}', "");
Expect(1, 74867, '\p{ -IS_P}', "");
Expect(0, 74867, '\p{^ -IS_P}', "");
Expect(0, 74867, '\P{ -IS_P}', "");
Expect(1, 74867, '\P{^ -IS_P}', "");
Expect(0, 74868, '\p{ -IS_P}', "");
Expect(1, 74868, '\p{^ -IS_P}', "");
Expect(1, 74868, '\P{ -IS_P}', "");
Expect(0, 74868, '\P{^ -IS_P}', "");
Error('\p{Expands_On_NFD}');
Error('\P{Expands_On_NFD}');
Error('\p{Is_Expands_On_NFD}');
Error('\P{Is_Expands_On_NFD}');
Error('\p{XO_NFD}');
Error('\P{XO_NFD}');
Error('\p{Is_XO_NFD}');
Error('\P{Is_XO_NFD}');
Error('\p{:=--Superscripts_And_Subscripts}');
Error('\P{:=--Superscripts_And_Subscripts}');
Expect(1, 8351, '\p{superscriptsandsubscripts}', "");
Expect(0, 8351, '\p{^superscriptsandsubscripts}', "");
Expect(0, 8351, '\P{superscriptsandsubscripts}', "");
Expect(1, 8351, '\P{^superscriptsandsubscripts}', "");
Expect(0, 8352, '\p{superscriptsandsubscripts}', "");
Expect(1, 8352, '\p{^superscriptsandsubscripts}', "");
Expect(1, 8352, '\P{superscriptsandsubscripts}', "");
Expect(0, 8352, '\P{^superscriptsandsubscripts}', "");
Expect(1, 8351, '\p{_	Superscripts_AND_SUBSCRIPTS}', "");
Expect(0, 8351, '\p{^_	Superscripts_AND_SUBSCRIPTS}', "");
Expect(0, 8351, '\P{_	Superscripts_AND_SUBSCRIPTS}', "");
Expect(1, 8351, '\P{^_	Superscripts_AND_SUBSCRIPTS}', "");
Expect(0, 8352, '\p{_	Superscripts_AND_SUBSCRIPTS}', "");
Expect(1, 8352, '\p{^_	Superscripts_AND_SUBSCRIPTS}', "");
Expect(1, 8352, '\P{_	Superscripts_AND_SUBSCRIPTS}', "");
Expect(0, 8352, '\P{^_	Superscripts_AND_SUBSCRIPTS}', "");
Error('\p{:=__is_Superscripts_And_SUBSCRIPTS}');
Error('\P{:=__is_Superscripts_And_SUBSCRIPTS}');
Expect(1, 8351, '\p{issuperscriptsandsubscripts}', "");
Expect(0, 8351, '\p{^issuperscriptsandsubscripts}', "");
Expect(0, 8351, '\P{issuperscriptsandsubscripts}', "");
Expect(1, 8351, '\P{^issuperscriptsandsubscripts}', "");
Expect(0, 8352, '\p{issuperscriptsandsubscripts}', "");
Expect(1, 8352, '\p{^issuperscriptsandsubscripts}', "");
Expect(1, 8352, '\P{issuperscriptsandsubscripts}', "");
Expect(0, 8352, '\P{^issuperscriptsandsubscripts}', "");
Expect(1, 8351, '\p{-IS_Superscripts_AND_subscripts}', "");
Expect(0, 8351, '\p{^-IS_Superscripts_AND_subscripts}', "");
Expect(0, 8351, '\P{-IS_Superscripts_AND_subscripts}', "");
Expect(1, 8351, '\P{^-IS_Superscripts_AND_subscripts}', "");
Expect(0, 8352, '\p{-IS_Superscripts_AND_subscripts}', "");
Expect(1, 8352, '\p{^-IS_Superscripts_AND_subscripts}', "");
Expect(1, 8352, '\P{-IS_Superscripts_AND_subscripts}', "");
Expect(0, 8352, '\P{^-IS_Superscripts_AND_subscripts}', "");
Error('\p{	in_Superscripts_AND_Subscripts:=}');
Error('\P{	in_Superscripts_AND_Subscripts:=}');
Expect(1, 8351, '\p{insuperscriptsandsubscripts}', "");
Expect(0, 8351, '\p{^insuperscriptsandsubscripts}', "");
Expect(0, 8351, '\P{insuperscriptsandsubscripts}', "");
Expect(1, 8351, '\P{^insuperscriptsandsubscripts}', "");
Expect(0, 8352, '\p{insuperscriptsandsubscripts}', "");
Expect(1, 8352, '\p{^insuperscriptsandsubscripts}', "");
Expect(1, 8352, '\P{insuperscriptsandsubscripts}', "");
Expect(0, 8352, '\P{^insuperscriptsandsubscripts}', "");
Expect(1, 8351, '\p{ IN_SUPERSCRIPTS_and_Subscripts}', "");
Expect(0, 8351, '\p{^ IN_SUPERSCRIPTS_and_Subscripts}', "");
Expect(0, 8351, '\P{ IN_SUPERSCRIPTS_and_Subscripts}', "");
Expect(1, 8351, '\P{^ IN_SUPERSCRIPTS_and_Subscripts}', "");
Expect(0, 8352, '\p{ IN_SUPERSCRIPTS_and_Subscripts}', "");
Expect(1, 8352, '\p{^ IN_SUPERSCRIPTS_and_Subscripts}', "");
Expect(1, 8352, '\P{ IN_SUPERSCRIPTS_and_Subscripts}', "");
Expect(0, 8352, '\P{^ IN_SUPERSCRIPTS_and_Subscripts}', "");
Error('\p{_	super_AND_Sub/a/}');
Error('\P{_	super_AND_Sub/a/}');
Expect(1, 8351, '\p{superandsub}', "");
Expect(0, 8351, '\p{^superandsub}', "");
Expect(0, 8351, '\P{superandsub}', "");
Expect(1, 8351, '\P{^superandsub}', "");
Expect(0, 8352, '\p{superandsub}', "");
Expect(1, 8352, '\p{^superandsub}', "");
Expect(1, 8352, '\P{superandsub}', "");
Expect(0, 8352, '\P{^superandsub}', "");
Expect(1, 8351, '\p{__Super_And_sub}', "");
Expect(0, 8351, '\p{^__Super_And_sub}', "");
Expect(0, 8351, '\P{__Super_And_sub}', "");
Expect(1, 8351, '\P{^__Super_And_sub}', "");
Expect(0, 8352, '\p{__Super_And_sub}', "");
Expect(1, 8352, '\p{^__Super_And_sub}', "");
Expect(1, 8352, '\P{__Super_And_sub}', "");
Expect(0, 8352, '\P{^__Super_And_sub}', "");
Error('\p{		is_SUPER_and_Sub/a/}');
Error('\P{		is_SUPER_and_Sub/a/}');
Expect(1, 8351, '\p{issuperandsub}', "");
Expect(0, 8351, '\p{^issuperandsub}', "");
Expect(0, 8351, '\P{issuperandsub}', "");
Expect(1, 8351, '\P{^issuperandsub}', "");
Expect(0, 8352, '\p{issuperandsub}', "");
Expect(1, 8352, '\p{^issuperandsub}', "");
Expect(1, 8352, '\P{issuperandsub}', "");
Expect(0, 8352, '\P{^issuperandsub}', "");
Expect(1, 8351, '\p{ Is_SUPER_And_SUB}', "");
Expect(0, 8351, '\p{^ Is_SUPER_And_SUB}', "");
Expect(0, 8351, '\P{ Is_SUPER_And_SUB}', "");
Expect(1, 8351, '\P{^ Is_SUPER_And_SUB}', "");
Expect(0, 8352, '\p{ Is_SUPER_And_SUB}', "");
Expect(1, 8352, '\p{^ Is_SUPER_And_SUB}', "");
Expect(1, 8352, '\P{ Is_SUPER_And_SUB}', "");
Expect(0, 8352, '\P{^ Is_SUPER_And_SUB}', "");
Error('\p{-	In_Super_And_SUB/a/}');
Error('\P{-	In_Super_And_SUB/a/}');
Expect(1, 8351, '\p{insuperandsub}', "");
Expect(0, 8351, '\p{^insuperandsub}', "");
Expect(0, 8351, '\P{insuperandsub}', "");
Expect(1, 8351, '\P{^insuperandsub}', "");
Expect(0, 8352, '\p{insuperandsub}', "");
Expect(1, 8352, '\p{^insuperandsub}', "");
Expect(1, 8352, '\P{insuperandsub}', "");
Expect(0, 8352, '\P{^insuperandsub}', "");
Expect(1, 8351, '\p{__In_SUPER_And_Sub}', "");
Expect(0, 8351, '\p{^__In_SUPER_And_Sub}', "");
Expect(0, 8351, '\P{__In_SUPER_And_Sub}', "");
Expect(1, 8351, '\P{^__In_SUPER_And_Sub}', "");
Expect(0, 8352, '\p{__In_SUPER_And_Sub}', "");
Expect(1, 8352, '\p{^__In_SUPER_And_Sub}', "");
Expect(1, 8352, '\P{__In_SUPER_And_Sub}', "");
Expect(0, 8352, '\P{^__In_SUPER_And_Sub}', "");
Error('\p{ in_balinese/a/}');
Error('\P{ in_balinese/a/}');
Expect(1, 7039, '\p{inbalinese}', "");
Expect(0, 7039, '\p{^inbalinese}', "");
Expect(0, 7039, '\P{inbalinese}', "");
Expect(1, 7039, '\P{^inbalinese}', "");
Expect(0, 7040, '\p{inbalinese}', "");
Expect(1, 7040, '\p{^inbalinese}', "");
Expect(1, 7040, '\P{inbalinese}', "");
Expect(0, 7040, '\P{^inbalinese}', "");
Expect(1, 7039, '\p{_In_balinese}', "");
Expect(0, 7039, '\p{^_In_balinese}', "");
Expect(0, 7039, '\P{_In_balinese}', "");
Expect(1, 7039, '\P{^_In_balinese}', "");
Expect(0, 7040, '\p{_In_balinese}', "");
Expect(1, 7040, '\p{^_In_balinese}', "");
Expect(1, 7040, '\P{_In_balinese}', "");
Expect(0, 7040, '\P{^_In_balinese}', "");
Error('\p{/a/_in_Balinese}');
Error('\P{/a/_in_Balinese}');
Expect(1, 7039, '\p{_	In_balinese}', "");
Expect(0, 7039, '\p{^_	In_balinese}', "");
Expect(0, 7039, '\P{_	In_balinese}', "");
Expect(1, 7039, '\P{^_	In_balinese}', "");
Expect(0, 7040, '\p{_	In_balinese}', "");
Expect(1, 7040, '\p{^_	In_balinese}', "");
Expect(1, 7040, '\P{_	In_balinese}', "");
Expect(0, 7040, '\P{^_	In_balinese}', "");
Error('\p{:=_Deprecated}');
Error('\P{:=_Deprecated}');
Expect(1, 917631, '\p{deprecated}', "");
Expect(0, 917631, '\p{^deprecated}', "");
Expect(0, 917631, '\P{deprecated}', "");
Expect(1, 917631, '\P{^deprecated}', "");
Expect(0, 917632, '\p{deprecated}', "");
Expect(1, 917632, '\p{^deprecated}', "");
Expect(1, 917632, '\P{deprecated}', "");
Expect(0, 917632, '\P{^deprecated}', "");
Expect(1, 917631, '\p{-DEPRECATED}', "");
Expect(0, 917631, '\p{^-DEPRECATED}', "");
Expect(0, 917631, '\P{-DEPRECATED}', "");
Expect(1, 917631, '\P{^-DEPRECATED}', "");
Expect(0, 917632, '\p{-DEPRECATED}', "");
Expect(1, 917632, '\p{^-DEPRECATED}', "");
Expect(1, 917632, '\P{-DEPRECATED}', "");
Expect(0, 917632, '\P{^-DEPRECATED}', "");
Error('\p{ _is_Deprecated/a/}');
Error('\P{ _is_Deprecated/a/}');
Expect(1, 917631, '\p{isdeprecated}', "");
Expect(0, 917631, '\p{^isdeprecated}', "");
Expect(0, 917631, '\P{isdeprecated}', "");
Expect(1, 917631, '\P{^isdeprecated}', "");
Expect(0, 917632, '\p{isdeprecated}', "");
Expect(1, 917632, '\p{^isdeprecated}', "");
Expect(1, 917632, '\P{isdeprecated}', "");
Expect(0, 917632, '\P{^isdeprecated}', "");
Expect(1, 917631, '\p{_ IS_Deprecated}', "");
Expect(0, 917631, '\p{^_ IS_Deprecated}', "");
Expect(0, 917631, '\P{_ IS_Deprecated}', "");
Expect(1, 917631, '\P{^_ IS_Deprecated}', "");
Expect(0, 917632, '\p{_ IS_Deprecated}', "");
Expect(1, 917632, '\p{^_ IS_Deprecated}', "");
Expect(1, 917632, '\P{_ IS_Deprecated}', "");
Expect(0, 917632, '\P{^_ IS_Deprecated}', "");
Error('\p{:=-_DEP}');
Error('\P{:=-_DEP}');
Expect(1, 917631, '\p{dep}', "");
Expect(0, 917631, '\p{^dep}', "");
Expect(0, 917631, '\P{dep}', "");
Expect(1, 917631, '\P{^dep}', "");
Expect(0, 917632, '\p{dep}', "");
Expect(1, 917632, '\p{^dep}', "");
Expect(1, 917632, '\P{dep}', "");
Expect(0, 917632, '\P{^dep}', "");
Expect(1, 917631, '\p{_ Dep}', "");
Expect(0, 917631, '\p{^_ Dep}', "");
Expect(0, 917631, '\P{_ Dep}', "");
Expect(1, 917631, '\P{^_ Dep}', "");
Expect(0, 917632, '\p{_ Dep}', "");
Expect(1, 917632, '\p{^_ Dep}', "");
Expect(1, 917632, '\P{_ Dep}', "");
Expect(0, 917632, '\P{^_ Dep}', "");
Error('\p{	:=Is_Dep}');
Error('\P{	:=Is_Dep}');
Expect(1, 917631, '\p{isdep}', "");
Expect(0, 917631, '\p{^isdep}', "");
Expect(0, 917631, '\P{isdep}', "");
Expect(1, 917631, '\P{^isdep}', "");
Expect(0, 917632, '\p{isdep}', "");
Expect(1, 917632, '\p{^isdep}', "");
Expect(1, 917632, '\P{isdep}', "");
Expect(0, 917632, '\P{^isdep}', "");
Expect(1, 917631, '\p{_-IS_Dep}', "");
Expect(0, 917631, '\p{^_-IS_Dep}', "");
Expect(0, 917631, '\P{_-IS_Dep}', "");
Expect(1, 917631, '\P{^_-IS_Dep}', "");
Expect(0, 917632, '\p{_-IS_Dep}', "");
Expect(1, 917632, '\p{^_-IS_Dep}', "");
Expect(1, 917632, '\P{_-IS_Dep}', "");
Expect(0, 917632, '\P{^_-IS_Dep}', "");
Error('\p{	CJK_Symbols_and_punctuation/a/}');
Error('\P{	CJK_Symbols_and_punctuation/a/}');
Expect(1, 12351, '\p{cjksymbolsandpunctuation}', "");
Expect(0, 12351, '\p{^cjksymbolsandpunctuation}', "");
Expect(0, 12351, '\P{cjksymbolsandpunctuation}', "");
Expect(1, 12351, '\P{^cjksymbolsandpunctuation}', "");
Expect(0, 12352, '\p{cjksymbolsandpunctuation}', "");
Expect(1, 12352, '\p{^cjksymbolsandpunctuation}', "");
Expect(1, 12352, '\P{cjksymbolsandpunctuation}', "");
Expect(0, 12352, '\P{^cjksymbolsandpunctuation}', "");
Expect(1, 12351, '\p{-	CJK_symbols_AND_punctuation}', "");
Expect(0, 12351, '\p{^-	CJK_symbols_AND_punctuation}', "");
Expect(0, 12351, '\P{-	CJK_symbols_AND_punctuation}', "");
Expect(1, 12351, '\P{^-	CJK_symbols_AND_punctuation}', "");
Expect(0, 12352, '\p{-	CJK_symbols_AND_punctuation}', "");
Expect(1, 12352, '\p{^-	CJK_symbols_AND_punctuation}', "");
Expect(1, 12352, '\P{-	CJK_symbols_AND_punctuation}', "");
Expect(0, 12352, '\P{^-	CJK_symbols_AND_punctuation}', "");
Error('\p{	_Is_CJK_SYMBOLS_And_Punctuation:=}');
Error('\P{	_Is_CJK_SYMBOLS_And_Punctuation:=}');
Expect(1, 12351, '\p{iscjksymbolsandpunctuation}', "");
Expect(0, 12351, '\p{^iscjksymbolsandpunctuation}', "");
Expect(0, 12351, '\P{iscjksymbolsandpunctuation}', "");
Expect(1, 12351, '\P{^iscjksymbolsandpunctuation}', "");
Expect(0, 12352, '\p{iscjksymbolsandpunctuation}', "");
Expect(1, 12352, '\p{^iscjksymbolsandpunctuation}', "");
Expect(1, 12352, '\P{iscjksymbolsandpunctuation}', "");
Expect(0, 12352, '\P{^iscjksymbolsandpunctuation}', "");
Expect(1, 12351, '\p{	is_CJK_Symbols_And_PUNCTUATION}', "");
Expect(0, 12351, '\p{^	is_CJK_Symbols_And_PUNCTUATION}', "");
Expect(0, 12351, '\P{	is_CJK_Symbols_And_PUNCTUATION}', "");
Expect(1, 12351, '\P{^	is_CJK_Symbols_And_PUNCTUATION}', "");
Expect(0, 12352, '\p{	is_CJK_Symbols_And_PUNCTUATION}', "");
Expect(1, 12352, '\p{^	is_CJK_Symbols_And_PUNCTUATION}', "");
Expect(1, 12352, '\P{	is_CJK_Symbols_And_PUNCTUATION}', "");
Expect(0, 12352, '\P{^	is_CJK_Symbols_And_PUNCTUATION}', "");
Error('\p{:=-_In_CJK_Symbols_and_Punctuation}');
Error('\P{:=-_In_CJK_Symbols_and_Punctuation}');
Expect(1, 12351, '\p{incjksymbolsandpunctuation}', "");
Expect(0, 12351, '\p{^incjksymbolsandpunctuation}', "");
Expect(0, 12351, '\P{incjksymbolsandpunctuation}', "");
Expect(1, 12351, '\P{^incjksymbolsandpunctuation}', "");
Expect(0, 12352, '\p{incjksymbolsandpunctuation}', "");
Expect(1, 12352, '\p{^incjksymbolsandpunctuation}', "");
Expect(1, 12352, '\P{incjksymbolsandpunctuation}', "");
Expect(0, 12352, '\P{^incjksymbolsandpunctuation}', "");
Expect(1, 12351, '\p{ In_CJK_SYMBOLS_And_punctuation}', "");
Expect(0, 12351, '\p{^ In_CJK_SYMBOLS_And_punctuation}', "");
Expect(0, 12351, '\P{ In_CJK_SYMBOLS_And_punctuation}', "");
Expect(1, 12351, '\P{^ In_CJK_SYMBOLS_And_punctuation}', "");
Expect(0, 12352, '\p{ In_CJK_SYMBOLS_And_punctuation}', "");
Expect(1, 12352, '\p{^ In_CJK_SYMBOLS_And_punctuation}', "");
Expect(1, 12352, '\P{ In_CJK_SYMBOLS_And_punctuation}', "");
Expect(0, 12352, '\P{^ In_CJK_SYMBOLS_And_punctuation}', "");
Error('\p{/a/_ CJK_SYMBOLS}');
Error('\P{/a/_ CJK_SYMBOLS}');
Expect(1, 12351, '\p{cjksymbols}', "");
Expect(0, 12351, '\p{^cjksymbols}', "");
Expect(0, 12351, '\P{cjksymbols}', "");
Expect(1, 12351, '\P{^cjksymbols}', "");
Expect(0, 12352, '\p{cjksymbols}', "");
Expect(1, 12352, '\p{^cjksymbols}', "");
Expect(1, 12352, '\P{cjksymbols}', "");
Expect(0, 12352, '\P{^cjksymbols}', "");
Expect(1, 12351, '\p{_	CJK_SYMBOLS}', "");
Expect(0, 12351, '\p{^_	CJK_SYMBOLS}', "");
Expect(0, 12351, '\P{_	CJK_SYMBOLS}', "");
Expect(1, 12351, '\P{^_	CJK_SYMBOLS}', "");
Expect(0, 12352, '\p{_	CJK_SYMBOLS}', "");
Expect(1, 12352, '\p{^_	CJK_SYMBOLS}', "");
Expect(1, 12352, '\P{_	CJK_SYMBOLS}', "");
Expect(0, 12352, '\P{^_	CJK_SYMBOLS}', "");
Error('\p{ 	Is_CJK_SYMBOLS:=}');
Error('\P{ 	Is_CJK_SYMBOLS:=}');
Expect(1, 12351, '\p{iscjksymbols}', "");
Expect(0, 12351, '\p{^iscjksymbols}', "");
Expect(0, 12351, '\P{iscjksymbols}', "");
Expect(1, 12351, '\P{^iscjksymbols}', "");
Expect(0, 12352, '\p{iscjksymbols}', "");
Expect(1, 12352, '\p{^iscjksymbols}', "");
Expect(1, 12352, '\P{iscjksymbols}', "");
Expect(0, 12352, '\P{^iscjksymbols}', "");
Expect(1, 12351, '\p{		IS_CJK_SYMBOLS}', "");
Expect(0, 12351, '\p{^		IS_CJK_SYMBOLS}', "");
Expect(0, 12351, '\P{		IS_CJK_SYMBOLS}', "");
Expect(1, 12351, '\P{^		IS_CJK_SYMBOLS}', "");
Expect(0, 12352, '\p{		IS_CJK_SYMBOLS}', "");
Expect(1, 12352, '\p{^		IS_CJK_SYMBOLS}', "");
Expect(1, 12352, '\P{		IS_CJK_SYMBOLS}', "");
Expect(0, 12352, '\P{^		IS_CJK_SYMBOLS}', "");
Error('\p{/a/-in_CJK_SYMBOLS}');
Error('\P{/a/-in_CJK_SYMBOLS}');
Expect(1, 12351, '\p{incjksymbols}', "");
Expect(0, 12351, '\p{^incjksymbols}', "");
Expect(0, 12351, '\P{incjksymbols}', "");
Expect(1, 12351, '\P{^incjksymbols}', "");
Expect(0, 12352, '\p{incjksymbols}', "");
Expect(1, 12352, '\p{^incjksymbols}', "");
Expect(1, 12352, '\P{incjksymbols}', "");
Expect(0, 12352, '\P{^incjksymbols}', "");
Expect(1, 12351, '\p{  IN_CJK_symbols}', "");
Expect(0, 12351, '\p{^  IN_CJK_symbols}', "");
Expect(0, 12351, '\P{  IN_CJK_symbols}', "");
Expect(1, 12351, '\P{^  IN_CJK_symbols}', "");
Expect(0, 12352, '\p{  IN_CJK_symbols}', "");
Expect(1, 12352, '\p{^  IN_CJK_symbols}', "");
Expect(1, 12352, '\P{  IN_CJK_symbols}', "");
Expect(0, 12352, '\P{^  IN_CJK_symbols}', "");
Error('\p{:=_	Myanmar}');
Error('\P{:=_	Myanmar}');
Expect(1, 43643, '\p{myanmar}', "");
Expect(0, 43643, '\p{^myanmar}', "");
Expect(0, 43643, '\P{myanmar}', "");
Expect(1, 43643, '\P{^myanmar}', "");
Expect(0, 43644, '\p{myanmar}', "");
Expect(1, 43644, '\p{^myanmar}', "");
Expect(1, 43644, '\P{myanmar}', "");
Expect(0, 43644, '\P{^myanmar}', "");
Expect(1, 43643, '\p{		myanmar}', "");
Expect(0, 43643, '\p{^		myanmar}', "");
Expect(0, 43643, '\P{		myanmar}', "");
Expect(1, 43643, '\P{^		myanmar}', "");
Expect(0, 43644, '\p{		myanmar}', "");
Expect(1, 43644, '\p{^		myanmar}', "");
Expect(1, 43644, '\P{		myanmar}', "");
Expect(0, 43644, '\P{^		myanmar}', "");
Error('\p{/a/-	IS_Myanmar}');
Error('\P{/a/-	IS_Myanmar}');
Expect(1, 43643, '\p{ismyanmar}', "");
Expect(0, 43643, '\p{^ismyanmar}', "");
Expect(0, 43643, '\P{ismyanmar}', "");
Expect(1, 43643, '\P{^ismyanmar}', "");
Expect(0, 43644, '\p{ismyanmar}', "");
Expect(1, 43644, '\p{^ismyanmar}', "");
Expect(1, 43644, '\P{ismyanmar}', "");
Expect(0, 43644, '\P{^ismyanmar}', "");
Expect(1, 43643, '\p{  is_Myanmar}', "");
Expect(0, 43643, '\p{^  is_Myanmar}', "");
Expect(0, 43643, '\P{  is_Myanmar}', "");
Expect(1, 43643, '\P{^  is_Myanmar}', "");
Expect(0, 43644, '\p{  is_Myanmar}', "");
Expect(1, 43644, '\p{^  is_Myanmar}', "");
Expect(1, 43644, '\P{  is_Myanmar}', "");
Expect(0, 43644, '\P{^  is_Myanmar}', "");
Error('\p{-mymr/a/}');
Error('\P{-mymr/a/}');
Expect(1, 43643, '\p{mymr}', "");
Expect(0, 43643, '\p{^mymr}', "");
Expect(0, 43643, '\P{mymr}', "");
Expect(1, 43643, '\P{^mymr}', "");
Expect(0, 43644, '\p{mymr}', "");
Expect(1, 43644, '\p{^mymr}', "");
Expect(1, 43644, '\P{mymr}', "");
Expect(0, 43644, '\P{^mymr}', "");
Expect(1, 43643, '\p{  MYMR}', "");
Expect(0, 43643, '\p{^  MYMR}', "");
Expect(0, 43643, '\P{  MYMR}', "");
Expect(1, 43643, '\P{^  MYMR}', "");
Expect(0, 43644, '\p{  MYMR}', "");
Expect(1, 43644, '\p{^  MYMR}', "");
Expect(1, 43644, '\P{  MYMR}', "");
Expect(0, 43644, '\P{^  MYMR}', "");
Error('\p{/a/Is_mymr}');
Error('\P{/a/Is_mymr}');
Expect(1, 43643, '\p{ismymr}', "");
Expect(0, 43643, '\p{^ismymr}', "");
Expect(0, 43643, '\P{ismymr}', "");
Expect(1, 43643, '\P{^ismymr}', "");
Expect(0, 43644, '\p{ismymr}', "");
Expect(1, 43644, '\p{^ismymr}', "");
Expect(1, 43644, '\P{ismymr}', "");
Expect(0, 43644, '\P{^ismymr}', "");
Expect(1, 43643, '\p{_ Is_Mymr}', "");
Expect(0, 43643, '\p{^_ Is_Mymr}', "");
Expect(0, 43643, '\P{_ Is_Mymr}', "");
Expect(1, 43643, '\P{^_ Is_Mymr}', "");
Expect(0, 43644, '\p{_ Is_Mymr}', "");
Expect(1, 43644, '\p{^_ Is_Mymr}', "");
Expect(1, 43644, '\P{_ Is_Mymr}', "");
Expect(0, 43644, '\P{^_ Is_Mymr}', "");
Error('\p{	 Initial_PUNCTUATION/a/}');
Error('\P{	 Initial_PUNCTUATION/a/}');
Expect(1, 11808, '\p{initialpunctuation}', "");
Expect(0, 11808, '\p{^initialpunctuation}', "");
Expect(0, 11808, '\P{initialpunctuation}', "");
Expect(1, 11808, '\P{^initialpunctuation}', "");
Expect(0, 11809, '\p{initialpunctuation}', "");
Expect(1, 11809, '\p{^initialpunctuation}', "");
Expect(1, 11809, '\P{initialpunctuation}', "");
Expect(0, 11809, '\P{^initialpunctuation}', "");
Expect(1, 11808, '\p{	-INITIAL_Punctuation}', "");
Expect(0, 11808, '\p{^	-INITIAL_Punctuation}', "");
Expect(0, 11808, '\P{	-INITIAL_Punctuation}', "");
Expect(1, 11808, '\P{^	-INITIAL_Punctuation}', "");
Expect(0, 11809, '\p{	-INITIAL_Punctuation}', "");
Expect(1, 11809, '\p{^	-INITIAL_Punctuation}', "");
Expect(1, 11809, '\P{	-INITIAL_Punctuation}', "");
Expect(0, 11809, '\P{^	-INITIAL_Punctuation}', "");
Error('\p{- IS_Initial_Punctuation:=}');
Error('\P{- IS_Initial_Punctuation:=}');
Expect(1, 11808, '\p{isinitialpunctuation}', "");
Expect(0, 11808, '\p{^isinitialpunctuation}', "");
Expect(0, 11808, '\P{isinitialpunctuation}', "");
Expect(1, 11808, '\P{^isinitialpunctuation}', "");
Expect(0, 11809, '\p{isinitialpunctuation}', "");
Expect(1, 11809, '\p{^isinitialpunctuation}', "");
Expect(1, 11809, '\P{isinitialpunctuation}', "");
Expect(0, 11809, '\P{^isinitialpunctuation}', "");
Expect(1, 11808, '\p{_	IS_Initial_PUNCTUATION}', "");
Expect(0, 11808, '\p{^_	IS_Initial_PUNCTUATION}', "");
Expect(0, 11808, '\P{_	IS_Initial_PUNCTUATION}', "");
Expect(1, 11808, '\P{^_	IS_Initial_PUNCTUATION}', "");
Expect(0, 11809, '\p{_	IS_Initial_PUNCTUATION}', "");
Expect(1, 11809, '\p{^_	IS_Initial_PUNCTUATION}', "");
Expect(1, 11809, '\P{_	IS_Initial_PUNCTUATION}', "");
Expect(0, 11809, '\P{^_	IS_Initial_PUNCTUATION}', "");
Error('\p{:=_pi}');
Error('\P{:=_pi}');
Expect(1, 11808, '\p{pi}', "");
Expect(0, 11808, '\p{^pi}', "");
Expect(0, 11808, '\P{pi}', "");
Expect(1, 11808, '\P{^pi}', "");
Expect(0, 11809, '\p{pi}', "");
Expect(1, 11809, '\p{^pi}', "");
Expect(1, 11809, '\P{pi}', "");
Expect(0, 11809, '\P{^pi}', "");
Expect(1, 11808, '\p{--PI}', "");
Expect(0, 11808, '\p{^--PI}', "");
Expect(0, 11808, '\P{--PI}', "");
Expect(1, 11808, '\P{^--PI}', "");
Expect(0, 11809, '\p{--PI}', "");
Expect(1, 11809, '\p{^--PI}', "");
Expect(1, 11809, '\P{--PI}', "");
Expect(0, 11809, '\P{^--PI}', "");
Error('\p{- Is_PI/a/}');
Error('\P{- Is_PI/a/}');
Expect(1, 11808, '\p{ispi}', "");
Expect(0, 11808, '\p{^ispi}', "");
Expect(0, 11808, '\P{ispi}', "");
Expect(1, 11808, '\P{^ispi}', "");
Expect(0, 11809, '\p{ispi}', "");
Expect(1, 11809, '\p{^ispi}', "");
Expect(1, 11809, '\P{ispi}', "");
Expect(0, 11809, '\P{^ispi}', "");
Expect(1, 11808, '\p{_	is_Pi}', "");
Expect(0, 11808, '\p{^_	is_Pi}', "");
Expect(0, 11808, '\P{_	is_Pi}', "");
Expect(1, 11808, '\P{^_	is_Pi}', "");
Expect(0, 11809, '\p{_	is_Pi}', "");
Expect(1, 11809, '\p{^_	is_Pi}', "");
Expect(1, 11809, '\P{_	is_Pi}', "");
Expect(0, 11809, '\P{^_	is_Pi}', "");
Error('\p{	:=common}');
Error('\P{	:=common}');
Expect(1, 917631, '\p{common}', "");
Expect(0, 917631, '\p{^common}', "");
Expect(0, 917631, '\P{common}', "");
Expect(1, 917631, '\P{^common}', "");
Expect(0, 917632, '\p{common}', "");
Expect(1, 917632, '\p{^common}', "");
Expect(1, 917632, '\P{common}', "");
Expect(0, 917632, '\P{^common}', "");
Expect(1, 917631, '\p{ Common}', "");
Expect(0, 917631, '\p{^ Common}', "");
Expect(0, 917631, '\P{ Common}', "");
Expect(1, 917631, '\P{^ Common}', "");
Expect(0, 917632, '\p{ Common}', "");
Expect(1, 917632, '\p{^ Common}', "");
Expect(1, 917632, '\P{ Common}', "");
Expect(0, 917632, '\P{^ Common}', "");
Error('\p{		Is_common:=}');
Error('\P{		Is_common:=}');
Expect(1, 917631, '\p{iscommon}', "");
Expect(0, 917631, '\p{^iscommon}', "");
Expect(0, 917631, '\P{iscommon}', "");
Expect(1, 917631, '\P{^iscommon}', "");
Expect(0, 917632, '\p{iscommon}', "");
Expect(1, 917632, '\p{^iscommon}', "");
Expect(1, 917632, '\P{iscommon}', "");
Expect(0, 917632, '\P{^iscommon}', "");
Expect(1, 917631, '\p{_	is_common}', "");
Expect(0, 917631, '\p{^_	is_common}', "");
Expect(0, 917631, '\P{_	is_common}', "");
Expect(1, 917631, '\P{^_	is_common}', "");
Expect(0, 917632, '\p{_	is_common}', "");
Expect(1, 917632, '\p{^_	is_common}', "");
Expect(1, 917632, '\P{_	is_common}', "");
Expect(0, 917632, '\P{^_	is_common}', "");
Error('\p{:=Zyyy}');
Error('\P{:=Zyyy}');
Expect(1, 917631, '\p{zyyy}', "");
Expect(0, 917631, '\p{^zyyy}', "");
Expect(0, 917631, '\P{zyyy}', "");
Expect(1, 917631, '\P{^zyyy}', "");
Expect(0, 917632, '\p{zyyy}', "");
Expect(1, 917632, '\p{^zyyy}', "");
Expect(1, 917632, '\P{zyyy}', "");
Expect(0, 917632, '\P{^zyyy}', "");
Expect(1, 917631, '\p{_ Zyyy}', "");
Expect(0, 917631, '\p{^_ Zyyy}', "");
Expect(0, 917631, '\P{_ Zyyy}', "");
Expect(1, 917631, '\P{^_ Zyyy}', "");
Expect(0, 917632, '\p{_ Zyyy}', "");
Expect(1, 917632, '\p{^_ Zyyy}', "");
Expect(1, 917632, '\P{_ Zyyy}', "");
Expect(0, 917632, '\P{^_ Zyyy}', "");
Error('\p{:=is_zyyy}');
Error('\P{:=is_zyyy}');
Expect(1, 917631, '\p{iszyyy}', "");
Expect(0, 917631, '\p{^iszyyy}', "");
Expect(0, 917631, '\P{iszyyy}', "");
Expect(1, 917631, '\P{^iszyyy}', "");
Expect(0, 917632, '\p{iszyyy}', "");
Expect(1, 917632, '\p{^iszyyy}', "");
Expect(1, 917632, '\P{iszyyy}', "");
Expect(0, 917632, '\P{^iszyyy}', "");
Expect(1, 917631, '\p{ IS_ZYYY}', "");
Expect(0, 917631, '\p{^ IS_ZYYY}', "");
Expect(0, 917631, '\P{ IS_ZYYY}', "");
Expect(1, 917631, '\P{^ IS_ZYYY}', "");
Expect(0, 917632, '\p{ IS_ZYYY}', "");
Expect(1, 917632, '\p{^ IS_ZYYY}', "");
Expect(1, 917632, '\P{ IS_ZYYY}', "");
Expect(0, 917632, '\P{^ IS_ZYYY}', "");
Error('\p{	:=HANGUL_Jamo}');
Error('\P{	:=HANGUL_Jamo}');
Expect(1, 4607, '\p{hanguljamo}', "");
Expect(0, 4607, '\p{^hanguljamo}', "");
Expect(0, 4607, '\P{hanguljamo}', "");
Expect(1, 4607, '\P{^hanguljamo}', "");
Expect(0, 4608, '\p{hanguljamo}', "");
Expect(1, 4608, '\p{^hanguljamo}', "");
Expect(1, 4608, '\P{hanguljamo}', "");
Expect(0, 4608, '\P{^hanguljamo}', "");
Expect(1, 4607, '\p{	Hangul_Jamo}', "");
Expect(0, 4607, '\p{^	Hangul_Jamo}', "");
Expect(0, 4607, '\P{	Hangul_Jamo}', "");
Expect(1, 4607, '\P{^	Hangul_Jamo}', "");
Expect(0, 4608, '\p{	Hangul_Jamo}', "");
Expect(1, 4608, '\p{^	Hangul_Jamo}', "");
Expect(1, 4608, '\P{	Hangul_Jamo}', "");
Expect(0, 4608, '\P{^	Hangul_Jamo}', "");
Error('\p{:=-	is_Hangul_Jamo}');
Error('\P{:=-	is_Hangul_Jamo}');
Expect(1, 4607, '\p{ishanguljamo}', "");
Expect(0, 4607, '\p{^ishanguljamo}', "");
Expect(0, 4607, '\P{ishanguljamo}', "");
Expect(1, 4607, '\P{^ishanguljamo}', "");
Expect(0, 4608, '\p{ishanguljamo}', "");
Expect(1, 4608, '\p{^ishanguljamo}', "");
Expect(1, 4608, '\P{ishanguljamo}', "");
Expect(0, 4608, '\P{^ishanguljamo}', "");
Expect(1, 4607, '\p{-Is_HANGUL_JAMO}', "");
Expect(0, 4607, '\p{^-Is_HANGUL_JAMO}', "");
Expect(0, 4607, '\P{-Is_HANGUL_JAMO}', "");
Expect(1, 4607, '\P{^-Is_HANGUL_JAMO}', "");
Expect(0, 4608, '\p{-Is_HANGUL_JAMO}', "");
Expect(1, 4608, '\p{^-Is_HANGUL_JAMO}', "");
Expect(1, 4608, '\P{-Is_HANGUL_JAMO}', "");
Expect(0, 4608, '\P{^-Is_HANGUL_JAMO}', "");
Error('\p{:=--IN_HANGUL_jamo}');
Error('\P{:=--IN_HANGUL_jamo}');
Expect(1, 4607, '\p{inhanguljamo}', "");
Expect(0, 4607, '\p{^inhanguljamo}', "");
Expect(0, 4607, '\P{inhanguljamo}', "");
Expect(1, 4607, '\P{^inhanguljamo}', "");
Expect(0, 4608, '\p{inhanguljamo}', "");
Expect(1, 4608, '\p{^inhanguljamo}', "");
Expect(1, 4608, '\P{inhanguljamo}', "");
Expect(0, 4608, '\P{^inhanguljamo}', "");
Expect(1, 4607, '\p{	_In_hangul_Jamo}', "");
Expect(0, 4607, '\p{^	_In_hangul_Jamo}', "");
Expect(0, 4607, '\P{	_In_hangul_Jamo}', "");
Expect(1, 4607, '\P{^	_In_hangul_Jamo}', "");
Expect(0, 4608, '\p{	_In_hangul_Jamo}', "");
Expect(1, 4608, '\p{^	_In_hangul_Jamo}', "");
Expect(1, 4608, '\P{	_In_hangul_Jamo}', "");
Expect(0, 4608, '\P{^	_In_hangul_Jamo}', "");
Error('\p{/a/-_jamo}');
Error('\P{/a/-_jamo}');
Expect(1, 4607, '\p{jamo}', "");
Expect(0, 4607, '\p{^jamo}', "");
Expect(0, 4607, '\P{jamo}', "");
Expect(1, 4607, '\P{^jamo}', "");
Expect(0, 4608, '\p{jamo}', "");
Expect(1, 4608, '\p{^jamo}', "");
Expect(1, 4608, '\P{jamo}', "");
Expect(0, 4608, '\P{^jamo}', "");
Expect(1, 4607, '\p{  jamo}', "");
Expect(0, 4607, '\p{^  jamo}', "");
Expect(0, 4607, '\P{  jamo}', "");
Expect(1, 4607, '\P{^  jamo}', "");
Expect(0, 4608, '\p{  jamo}', "");
Expect(1, 4608, '\p{^  jamo}', "");
Expect(1, 4608, '\P{  jamo}', "");
Expect(0, 4608, '\P{^  jamo}', "");
Error('\p{:=_	Is_jamo}');
Error('\P{:=_	Is_jamo}');
Expect(1, 4607, '\p{isjamo}', "");
Expect(0, 4607, '\p{^isjamo}', "");
Expect(0, 4607, '\P{isjamo}', "");
Expect(1, 4607, '\P{^isjamo}', "");
Expect(0, 4608, '\p{isjamo}', "");
Expect(1, 4608, '\p{^isjamo}', "");
Expect(1, 4608, '\P{isjamo}', "");
Expect(0, 4608, '\P{^isjamo}', "");
Expect(1, 4607, '\p{  Is_jamo}', "");
Expect(0, 4607, '\p{^  Is_jamo}', "");
Expect(0, 4607, '\P{  Is_jamo}', "");
Expect(1, 4607, '\P{^  Is_jamo}', "");
Expect(0, 4608, '\p{  Is_jamo}', "");
Expect(1, 4608, '\p{^  Is_jamo}', "");
Expect(1, 4608, '\P{  Is_jamo}', "");
Expect(0, 4608, '\P{^  Is_jamo}', "");
Error('\p{/a/		In_jamo}');
Error('\P{/a/		In_jamo}');
Expect(1, 4607, '\p{injamo}', "");
Expect(0, 4607, '\p{^injamo}', "");
Expect(0, 4607, '\P{injamo}', "");
Expect(1, 4607, '\P{^injamo}', "");
Expect(0, 4608, '\p{injamo}', "");
Expect(1, 4608, '\p{^injamo}', "");
Expect(1, 4608, '\P{injamo}', "");
Expect(0, 4608, '\P{^injamo}', "");
Expect(1, 4607, '\p{in_Jamo}', "");
Expect(0, 4607, '\p{^in_Jamo}', "");
Expect(0, 4607, '\P{in_Jamo}', "");
Expect(1, 4607, '\P{^in_Jamo}', "");
Expect(0, 4608, '\p{in_Jamo}', "");
Expect(1, 4608, '\p{^in_Jamo}', "");
Expect(1, 4608, '\P{in_Jamo}', "");
Expect(0, 4608, '\P{^in_Jamo}', "");
Error('\p{/a/_ CHANGES_When_Lowercased}');
Error('\P{/a/_ CHANGES_When_Lowercased}');
Expect(1, 66599, '\p{changeswhenlowercased}', "");
Expect(0, 66599, '\p{^changeswhenlowercased}', "");
Expect(0, 66599, '\P{changeswhenlowercased}', "");
Expect(1, 66599, '\P{^changeswhenlowercased}', "");
Expect(0, 66600, '\p{changeswhenlowercased}', "");
Expect(1, 66600, '\p{^changeswhenlowercased}', "");
Expect(1, 66600, '\P{changeswhenlowercased}', "");
Expect(0, 66600, '\P{^changeswhenlowercased}', "");
Expect(1, 66599, '\p{-	CHANGES_when_LOWERCASED}', "");
Expect(0, 66599, '\p{^-	CHANGES_when_LOWERCASED}', "");
Expect(0, 66599, '\P{-	CHANGES_when_LOWERCASED}', "");
Expect(1, 66599, '\P{^-	CHANGES_when_LOWERCASED}', "");
Expect(0, 66600, '\p{-	CHANGES_when_LOWERCASED}', "");
Expect(1, 66600, '\p{^-	CHANGES_when_LOWERCASED}', "");
Expect(1, 66600, '\P{-	CHANGES_when_LOWERCASED}', "");
Expect(0, 66600, '\P{^-	CHANGES_when_LOWERCASED}', "");
Error('\p{:=	 Is_changes_When_Lowercased}');
Error('\P{:=	 Is_changes_When_Lowercased}');
Expect(1, 66599, '\p{ischangeswhenlowercased}', "");
Expect(0, 66599, '\p{^ischangeswhenlowercased}', "");
Expect(0, 66599, '\P{ischangeswhenlowercased}', "");
Expect(1, 66599, '\P{^ischangeswhenlowercased}', "");
Expect(0, 66600, '\p{ischangeswhenlowercased}', "");
Expect(1, 66600, '\p{^ischangeswhenlowercased}', "");
Expect(1, 66600, '\P{ischangeswhenlowercased}', "");
Expect(0, 66600, '\P{^ischangeswhenlowercased}', "");
Expect(1, 66599, '\p{_ Is_Changes_When_Lowercased}', "");
Expect(0, 66599, '\p{^_ Is_Changes_When_Lowercased}', "");
Expect(0, 66599, '\P{_ Is_Changes_When_Lowercased}', "");
Expect(1, 66599, '\P{^_ Is_Changes_When_Lowercased}', "");
Expect(0, 66600, '\p{_ Is_Changes_When_Lowercased}', "");
Expect(1, 66600, '\p{^_ Is_Changes_When_Lowercased}', "");
Expect(1, 66600, '\P{_ Is_Changes_When_Lowercased}', "");
Expect(0, 66600, '\P{^_ Is_Changes_When_Lowercased}', "");
Error('\p{__CWL:=}');
Error('\P{__CWL:=}');
Expect(1, 66599, '\p{cwl}', "");
Expect(0, 66599, '\p{^cwl}', "");
Expect(0, 66599, '\P{cwl}', "");
Expect(1, 66599, '\P{^cwl}', "");
Expect(0, 66600, '\p{cwl}', "");
Expect(1, 66600, '\p{^cwl}', "");
Expect(1, 66600, '\P{cwl}', "");
Expect(0, 66600, '\P{^cwl}', "");
Expect(1, 66599, '\p{_CWL}', "");
Expect(0, 66599, '\p{^_CWL}', "");
Expect(0, 66599, '\P{_CWL}', "");
Expect(1, 66599, '\P{^_CWL}', "");
Expect(0, 66600, '\p{_CWL}', "");
Expect(1, 66600, '\p{^_CWL}', "");
Expect(1, 66600, '\P{_CWL}', "");
Expect(0, 66600, '\P{^_CWL}', "");
Error('\p{__is_CWL/a/}');
Error('\P{__is_CWL/a/}');
Expect(1, 66599, '\p{iscwl}', "");
Expect(0, 66599, '\p{^iscwl}', "");
Expect(0, 66599, '\P{iscwl}', "");
Expect(1, 66599, '\P{^iscwl}', "");
Expect(0, 66600, '\p{iscwl}', "");
Expect(1, 66600, '\p{^iscwl}', "");
Expect(1, 66600, '\P{iscwl}', "");
Expect(0, 66600, '\P{^iscwl}', "");
Expect(1, 66599, '\p{ IS_CWL}', "");
Expect(0, 66599, '\p{^ IS_CWL}', "");
Expect(0, 66599, '\P{ IS_CWL}', "");
Expect(1, 66599, '\P{^ IS_CWL}', "");
Expect(0, 66600, '\p{ IS_CWL}', "");
Expect(1, 66600, '\p{^ IS_CWL}', "");
Expect(1, 66600, '\P{ IS_CWL}', "");
Expect(0, 66600, '\P{^ IS_CWL}', "");
Error('\p{	Canadian_Aboriginal/a/}');
Error('\P{	Canadian_Aboriginal/a/}');
Expect(1, 6389, '\p{canadianaboriginal}', "");
Expect(0, 6389, '\p{^canadianaboriginal}', "");
Expect(0, 6389, '\P{canadianaboriginal}', "");
Expect(1, 6389, '\P{^canadianaboriginal}', "");
Expect(0, 6390, '\p{canadianaboriginal}', "");
Expect(1, 6390, '\p{^canadianaboriginal}', "");
Expect(1, 6390, '\P{canadianaboriginal}', "");
Expect(0, 6390, '\P{^canadianaboriginal}', "");
Expect(1, 6389, '\p{ 	Canadian_ABORIGINAL}', "");
Expect(0, 6389, '\p{^ 	Canadian_ABORIGINAL}', "");
Expect(0, 6389, '\P{ 	Canadian_ABORIGINAL}', "");
Expect(1, 6389, '\P{^ 	Canadian_ABORIGINAL}', "");
Expect(0, 6390, '\p{ 	Canadian_ABORIGINAL}', "");
Expect(1, 6390, '\p{^ 	Canadian_ABORIGINAL}', "");
Expect(1, 6390, '\P{ 	Canadian_ABORIGINAL}', "");
Expect(0, 6390, '\P{^ 	Canadian_ABORIGINAL}', "");
Error('\p{/a/-_Is_Canadian_Aboriginal}');
Error('\P{/a/-_Is_Canadian_Aboriginal}');
Expect(1, 6389, '\p{iscanadianaboriginal}', "");
Expect(0, 6389, '\p{^iscanadianaboriginal}', "");
Expect(0, 6389, '\P{iscanadianaboriginal}', "");
Expect(1, 6389, '\P{^iscanadianaboriginal}', "");
Expect(0, 6390, '\p{iscanadianaboriginal}', "");
Expect(1, 6390, '\p{^iscanadianaboriginal}', "");
Expect(1, 6390, '\P{iscanadianaboriginal}', "");
Expect(0, 6390, '\P{^iscanadianaboriginal}', "");
Expect(1, 6389, '\p{	-IS_Canadian_Aboriginal}', "");
Expect(0, 6389, '\p{^	-IS_Canadian_Aboriginal}', "");
Expect(0, 6389, '\P{	-IS_Canadian_Aboriginal}', "");
Expect(1, 6389, '\P{^	-IS_Canadian_Aboriginal}', "");
Expect(0, 6390, '\p{	-IS_Canadian_Aboriginal}', "");
Expect(1, 6390, '\p{^	-IS_Canadian_Aboriginal}', "");
Expect(1, 6390, '\P{	-IS_Canadian_Aboriginal}', "");
Expect(0, 6390, '\P{^	-IS_Canadian_Aboriginal}', "");
Error('\p{_ Cans:=}');
Error('\P{_ Cans:=}');
Expect(1, 6389, '\p{cans}', "");
Expect(0, 6389, '\p{^cans}', "");
Expect(0, 6389, '\P{cans}', "");
Expect(1, 6389, '\P{^cans}', "");
Expect(0, 6390, '\p{cans}', "");
Expect(1, 6390, '\p{^cans}', "");
Expect(1, 6390, '\P{cans}', "");
Expect(0, 6390, '\P{^cans}', "");
Expect(1, 6389, '\p{ cans}', "");
Expect(0, 6389, '\p{^ cans}', "");
Expect(0, 6389, '\P{ cans}', "");
Expect(1, 6389, '\P{^ cans}', "");
Expect(0, 6390, '\p{ cans}', "");
Expect(1, 6390, '\p{^ cans}', "");
Expect(1, 6390, '\P{ cans}', "");
Expect(0, 6390, '\P{^ cans}', "");
Error('\p{_/a/is_CANS}');
Error('\P{_/a/is_CANS}');
Expect(1, 6389, '\p{iscans}', "");
Expect(0, 6389, '\p{^iscans}', "");
Expect(0, 6389, '\P{iscans}', "");
Expect(1, 6389, '\P{^iscans}', "");
Expect(0, 6390, '\p{iscans}', "");
Expect(1, 6390, '\p{^iscans}', "");
Expect(1, 6390, '\P{iscans}', "");
Expect(0, 6390, '\P{^iscans}', "");
Expect(1, 6389, '\p{	IS_Cans}', "");
Expect(0, 6389, '\p{^	IS_Cans}', "");
Expect(0, 6389, '\P{	IS_Cans}', "");
Expect(1, 6389, '\P{^	IS_Cans}', "");
Expect(0, 6390, '\p{	IS_Cans}', "");
Expect(1, 6390, '\p{^	IS_Cans}', "");
Expect(1, 6390, '\P{	IS_Cans}', "");
Expect(0, 6390, '\P{^	IS_Cans}', "");
Error('\p{_/a/Pattern_syntax}');
Error('\P{_/a/Pattern_syntax}');
Expect(1, 65094, '\p{patternsyntax}', "");
Expect(0, 65094, '\p{^patternsyntax}', "");
Expect(0, 65094, '\P{patternsyntax}', "");
Expect(1, 65094, '\P{^patternsyntax}', "");
Expect(0, 65095, '\p{patternsyntax}', "");
Expect(1, 65095, '\p{^patternsyntax}', "");
Expect(1, 65095, '\P{patternsyntax}', "");
Expect(0, 65095, '\P{^patternsyntax}', "");
Expect(1, 65094, '\p{-	pattern_Syntax}', "");
Expect(0, 65094, '\p{^-	pattern_Syntax}', "");
Expect(0, 65094, '\P{-	pattern_Syntax}', "");
Expect(1, 65094, '\P{^-	pattern_Syntax}', "");
Expect(0, 65095, '\p{-	pattern_Syntax}', "");
Expect(1, 65095, '\p{^-	pattern_Syntax}', "");
Expect(1, 65095, '\P{-	pattern_Syntax}', "");
Expect(0, 65095, '\P{^-	pattern_Syntax}', "");
Error('\p{/a/_IS_PATTERN_syntax}');
Error('\P{/a/_IS_PATTERN_syntax}');
Expect(1, 65094, '\p{ispatternsyntax}', "");
Expect(0, 65094, '\p{^ispatternsyntax}', "");
Expect(0, 65094, '\P{ispatternsyntax}', "");
Expect(1, 65094, '\P{^ispatternsyntax}', "");
Expect(0, 65095, '\p{ispatternsyntax}', "");
Expect(1, 65095, '\p{^ispatternsyntax}', "");
Expect(1, 65095, '\P{ispatternsyntax}', "");
Expect(0, 65095, '\P{^ispatternsyntax}', "");
Expect(1, 65094, '\p{ -Is_pattern_syntax}', "");
Expect(0, 65094, '\p{^ -Is_pattern_syntax}', "");
Expect(0, 65094, '\P{ -Is_pattern_syntax}', "");
Expect(1, 65094, '\P{^ -Is_pattern_syntax}', "");
Expect(0, 65095, '\p{ -Is_pattern_syntax}', "");
Expect(1, 65095, '\p{^ -Is_pattern_syntax}', "");
Expect(1, 65095, '\P{ -Is_pattern_syntax}', "");
Expect(0, 65095, '\P{^ -Is_pattern_syntax}', "");
Error('\p{:= _Pat_Syn}');
Error('\P{:= _Pat_Syn}');
Expect(1, 65094, '\p{patsyn}', "");
Expect(0, 65094, '\p{^patsyn}', "");
Expect(0, 65094, '\P{patsyn}', "");
Expect(1, 65094, '\P{^patsyn}', "");
Expect(0, 65095, '\p{patsyn}', "");
Expect(1, 65095, '\p{^patsyn}', "");
Expect(1, 65095, '\P{patsyn}', "");
Expect(0, 65095, '\P{^patsyn}', "");
Expect(1, 65094, '\p{ 	Pat_Syn}', "");
Expect(0, 65094, '\p{^ 	Pat_Syn}', "");
Expect(0, 65094, '\P{ 	Pat_Syn}', "");
Expect(1, 65094, '\P{^ 	Pat_Syn}', "");
Expect(0, 65095, '\p{ 	Pat_Syn}', "");
Expect(1, 65095, '\p{^ 	Pat_Syn}', "");
Expect(1, 65095, '\P{ 	Pat_Syn}', "");
Expect(0, 65095, '\P{^ 	Pat_Syn}', "");
Error('\p{-:=IS_PAT_SYN}');
Error('\P{-:=IS_PAT_SYN}');
Expect(1, 65094, '\p{ispatsyn}', "");
Expect(0, 65094, '\p{^ispatsyn}', "");
Expect(0, 65094, '\P{ispatsyn}', "");
Expect(1, 65094, '\P{^ispatsyn}', "");
Expect(0, 65095, '\p{ispatsyn}', "");
Expect(1, 65095, '\p{^ispatsyn}', "");
Expect(1, 65095, '\P{ispatsyn}', "");
Expect(0, 65095, '\P{^ispatsyn}', "");
Expect(1, 65094, '\p{_-is_Pat_syn}', "");
Expect(0, 65094, '\p{^_-is_Pat_syn}', "");
Expect(0, 65094, '\P{_-is_Pat_syn}', "");
Expect(1, 65094, '\P{^_-is_Pat_syn}', "");
Expect(0, 65095, '\p{_-is_Pat_syn}', "");
Expect(1, 65095, '\p{^_-is_Pat_syn}', "");
Expect(1, 65095, '\P{_-is_Pat_syn}', "");
Expect(0, 65095, '\P{^_-is_Pat_syn}', "");
Error('\p{/a/		No_BLOCK}');
Error('\P{/a/		No_BLOCK}');
Expect(1, 918000, '\p{noblock}', "");
Expect(0, 918000, '\p{^noblock}', "");
Expect(0, 918000, '\P{noblock}', "");
Expect(1, 918000, '\P{^noblock}', "");
Expect(0, 983040, '\p{noblock}', "");
Expect(1, 983040, '\p{^noblock}', "");
Expect(1, 983040, '\P{noblock}', "");
Expect(0, 983040, '\P{^noblock}', "");
Expect(1, 918000, '\p{ no_block}', "");
Expect(0, 918000, '\p{^ no_block}', "");
Expect(0, 918000, '\P{ no_block}', "");
Expect(1, 918000, '\P{^ no_block}', "");
Expect(0, 983040, '\p{ no_block}', "");
Expect(1, 983040, '\p{^ no_block}', "");
Expect(1, 983040, '\P{ no_block}', "");
Expect(0, 983040, '\P{^ no_block}', "");
Error('\p{	Is_No_BLOCK:=}');
Error('\P{	Is_No_BLOCK:=}');
Expect(1, 918000, '\p{isnoblock}', "");
Expect(0, 918000, '\p{^isnoblock}', "");
Expect(0, 918000, '\P{isnoblock}', "");
Expect(1, 918000, '\P{^isnoblock}', "");
Expect(0, 983040, '\p{isnoblock}', "");
Expect(1, 983040, '\p{^isnoblock}', "");
Expect(1, 983040, '\P{isnoblock}', "");
Expect(0, 983040, '\P{^isnoblock}', "");
Expect(1, 918000, '\p{_	is_NO_block}', "");
Expect(0, 918000, '\p{^_	is_NO_block}', "");
Expect(0, 918000, '\P{_	is_NO_block}', "");
Expect(1, 918000, '\P{^_	is_NO_block}', "");
Expect(0, 983040, '\p{_	is_NO_block}', "");
Expect(1, 983040, '\p{^_	is_NO_block}', "");
Expect(1, 983040, '\P{_	is_NO_block}', "");
Expect(0, 983040, '\P{^_	is_NO_block}', "");
Error('\p{	:=In_NO_Block}');
Error('\P{	:=In_NO_Block}');
Expect(1, 918000, '\p{innoblock}', "");
Expect(0, 918000, '\p{^innoblock}', "");
Expect(0, 918000, '\P{innoblock}', "");
Expect(1, 918000, '\P{^innoblock}', "");
Expect(0, 983040, '\p{innoblock}', "");
Expect(1, 983040, '\p{^innoblock}', "");
Expect(1, 983040, '\P{innoblock}', "");
Expect(0, 983040, '\P{^innoblock}', "");
Expect(1, 918000, '\p{_-in_No_Block}', "");
Expect(0, 918000, '\p{^_-in_No_Block}', "");
Expect(0, 918000, '\P{_-in_No_Block}', "");
Expect(1, 918000, '\P{^_-in_No_Block}', "");
Expect(0, 983040, '\p{_-in_No_Block}', "");
Expect(1, 983040, '\p{^_-in_No_Block}', "");
Expect(1, 983040, '\P{_-in_No_Block}', "");
Expect(0, 983040, '\P{^_-in_No_Block}', "");
Error('\p{-NB:=}');
Error('\P{-NB:=}');
Expect(1, 918000, '\p{nb}', "");
Expect(0, 918000, '\p{^nb}', "");
Expect(0, 918000, '\P{nb}', "");
Expect(1, 918000, '\P{^nb}', "");
Expect(0, 983040, '\p{nb}', "");
Expect(1, 983040, '\p{^nb}', "");
Expect(1, 983040, '\P{nb}', "");
Expect(0, 983040, '\P{^nb}', "");
Expect(1, 918000, '\p{	NB}', "");
Expect(0, 918000, '\p{^	NB}', "");
Expect(0, 918000, '\P{	NB}', "");
Expect(1, 918000, '\P{^	NB}', "");
Expect(0, 983040, '\p{	NB}', "");
Expect(1, 983040, '\p{^	NB}', "");
Expect(1, 983040, '\P{	NB}', "");
Expect(0, 983040, '\P{^	NB}', "");
Error('\p{_/a/IS_NB}');
Error('\P{_/a/IS_NB}');
Expect(1, 918000, '\p{isnb}', "");
Expect(0, 918000, '\p{^isnb}', "");
Expect(0, 918000, '\P{isnb}', "");
Expect(1, 918000, '\P{^isnb}', "");
Expect(0, 983040, '\p{isnb}', "");
Expect(1, 983040, '\p{^isnb}', "");
Expect(1, 983040, '\P{isnb}', "");
Expect(0, 983040, '\P{^isnb}', "");
Expect(1, 918000, '\p{	Is_NB}', "");
Expect(0, 918000, '\p{^	Is_NB}', "");
Expect(0, 918000, '\P{	Is_NB}', "");
Expect(1, 918000, '\P{^	Is_NB}', "");
Expect(0, 983040, '\p{	Is_NB}', "");
Expect(1, 983040, '\p{^	Is_NB}', "");
Expect(1, 983040, '\P{	Is_NB}', "");
Expect(0, 983040, '\P{^	Is_NB}', "");
Error('\p{:=-	In_NB}');
Error('\P{:=-	In_NB}');
Expect(1, 918000, '\p{innb}', "");
Expect(0, 918000, '\p{^innb}', "");
Expect(0, 918000, '\P{innb}', "");
Expect(1, 918000, '\P{^innb}', "");
Expect(0, 983040, '\p{innb}', "");
Expect(1, 983040, '\p{^innb}', "");
Expect(1, 983040, '\P{innb}', "");
Expect(0, 983040, '\P{^innb}', "");
Expect(1, 918000, '\p{_-In_NB}', "");
Expect(0, 918000, '\p{^_-In_NB}', "");
Expect(0, 918000, '\P{_-In_NB}', "");
Expect(1, 918000, '\P{^_-In_NB}', "");
Expect(0, 983040, '\p{_-In_NB}', "");
Expect(1, 983040, '\p{^_-In_NB}', "");
Expect(1, 983040, '\P{_-In_NB}', "");
Expect(0, 983040, '\P{^_-In_NB}', "");
Error('\p{:=-_Other}');
Error('\P{:=-_Other}');
Expect(1, 918000, '\p{other}', "");
Expect(0, 918000, '\p{^other}', "");
Expect(0, 918000, '\P{other}', "");
Expect(1, 918000, '\P{^other}', "");
Expect(0, 917999, '\p{other}', "");
Expect(1, 917999, '\p{^other}', "");
Expect(1, 917999, '\P{other}', "");
Expect(0, 917999, '\P{^other}', "");
Expect(1, 918000, '\p{_OTHER}', "");
Expect(0, 918000, '\p{^_OTHER}', "");
Expect(0, 918000, '\P{_OTHER}', "");
Expect(1, 918000, '\P{^_OTHER}', "");
Expect(0, 917999, '\p{_OTHER}', "");
Expect(1, 917999, '\p{^_OTHER}', "");
Expect(1, 917999, '\P{_OTHER}', "");
Expect(0, 917999, '\P{^_OTHER}', "");
Error('\p{:= 	Is_Other}');
Error('\P{:= 	Is_Other}');
Expect(1, 918000, '\p{isother}', "");
Expect(0, 918000, '\p{^isother}', "");
Expect(0, 918000, '\P{isother}', "");
Expect(1, 918000, '\P{^isother}', "");
Expect(0, 917999, '\p{isother}', "");
Expect(1, 917999, '\p{^isother}', "");
Expect(1, 917999, '\P{isother}', "");
Expect(0, 917999, '\P{^isother}', "");
Expect(1, 918000, '\p{-IS_OTHER}', "");
Expect(0, 918000, '\p{^-IS_OTHER}', "");
Expect(0, 918000, '\P{-IS_OTHER}', "");
Expect(1, 918000, '\P{^-IS_OTHER}', "");
Expect(0, 917999, '\p{-IS_OTHER}', "");
Expect(1, 917999, '\p{^-IS_OTHER}', "");
Expect(1, 917999, '\P{-IS_OTHER}', "");
Expect(0, 917999, '\P{^-IS_OTHER}', "");
Error('\p{/a/ 	c}');
Error('\P{/a/ 	c}');
Expect(1, 918000, '\p{c}', "");
Expect(0, 918000, '\p{^c}', "");
Expect(0, 918000, '\P{c}', "");
Expect(1, 918000, '\P{^c}', "");
Expect(0, 917999, '\p{c}', "");
Expect(1, 917999, '\p{^c}', "");
Expect(1, 917999, '\P{c}', "");
Expect(0, 917999, '\P{^c}', "");
Expect(1, 918000, '\p{  c}', "");
Expect(0, 918000, '\p{^  c}', "");
Expect(0, 918000, '\P{  c}', "");
Expect(1, 918000, '\P{^  c}', "");
Expect(0, 917999, '\p{  c}', "");
Expect(1, 917999, '\p{^  c}', "");
Expect(1, 917999, '\P{  c}', "");
Expect(0, 917999, '\P{^  c}', "");
Error('\p{	:=is_C}');
Error('\P{	:=is_C}');
Expect(1, 918000, '\p{isc}', "");
Expect(0, 918000, '\p{^isc}', "");
Expect(0, 918000, '\P{isc}', "");
Expect(1, 918000, '\P{^isc}', "");
Expect(0, 917999, '\p{isc}', "");
Expect(1, 917999, '\p{^isc}', "");
Expect(1, 917999, '\P{isc}', "");
Expect(0, 917999, '\P{^isc}', "");
Expect(1, 918000, '\p{  Is_C}', "");
Expect(0, 918000, '\p{^  Is_C}', "");
Expect(0, 918000, '\P{  Is_C}', "");
Expect(1, 918000, '\P{^  Is_C}', "");
Expect(0, 917999, '\p{  Is_C}', "");
Expect(1, 917999, '\p{^  Is_C}', "");
Expect(1, 917999, '\P{  Is_C}', "");
Expect(0, 917999, '\P{^  Is_C}', "");
Error('\p{:=	 VARIATION_selector}');
Error('\P{:=	 VARIATION_selector}');
Expect(1, 917999, '\p{variationselector}', "");
Expect(0, 917999, '\p{^variationselector}', "");
Expect(0, 917999, '\P{variationselector}', "");
Expect(1, 917999, '\P{^variationselector}', "");
Expect(0, 918000, '\p{variationselector}', "");
Expect(1, 918000, '\p{^variationselector}', "");
Expect(1, 918000, '\P{variationselector}', "");
Expect(0, 918000, '\P{^variationselector}', "");
Expect(1, 917999, '\p{-Variation_SELECTOR}', "");
Expect(0, 917999, '\p{^-Variation_SELECTOR}', "");
Expect(0, 917999, '\P{-Variation_SELECTOR}', "");
Expect(1, 917999, '\P{^-Variation_SELECTOR}', "");
Expect(0, 918000, '\p{-Variation_SELECTOR}', "");
Expect(1, 918000, '\p{^-Variation_SELECTOR}', "");
Expect(1, 918000, '\P{-Variation_SELECTOR}', "");
Expect(0, 918000, '\P{^-Variation_SELECTOR}', "");
Error('\p{-/a/Is_variation_selector}');
Error('\P{-/a/Is_variation_selector}');
Expect(1, 917999, '\p{isvariationselector}', "");
Expect(0, 917999, '\p{^isvariationselector}', "");
Expect(0, 917999, '\P{isvariationselector}', "");
Expect(1, 917999, '\P{^isvariationselector}', "");
Expect(0, 918000, '\p{isvariationselector}', "");
Expect(1, 918000, '\p{^isvariationselector}', "");
Expect(1, 918000, '\P{isvariationselector}', "");
Expect(0, 918000, '\P{^isvariationselector}', "");
Expect(1, 917999, '\p{__Is_Variation_Selector}', "");
Expect(0, 917999, '\p{^__Is_Variation_Selector}', "");
Expect(0, 917999, '\P{__Is_Variation_Selector}', "");
Expect(1, 917999, '\P{^__Is_Variation_Selector}', "");
Expect(0, 918000, '\p{__Is_Variation_Selector}', "");
Expect(1, 918000, '\p{^__Is_Variation_Selector}', "");
Expect(1, 918000, '\P{__Is_Variation_Selector}', "");
Expect(0, 918000, '\P{^__Is_Variation_Selector}', "");
Error('\p{  vs/a/}');
Error('\P{  vs/a/}');
Expect(1, 917999, '\p{vs}', "");
Expect(0, 917999, '\p{^vs}', "");
Expect(0, 917999, '\P{vs}', "");
Expect(1, 917999, '\P{^vs}', "");
Expect(0, 918000, '\p{vs}', "");
Expect(1, 918000, '\p{^vs}', "");
Expect(1, 918000, '\P{vs}', "");
Expect(0, 918000, '\P{^vs}', "");
Expect(1, 917999, '\p{_VS}', "");
Expect(0, 917999, '\p{^_VS}', "");
Expect(0, 917999, '\P{_VS}', "");
Expect(1, 917999, '\P{^_VS}', "");
Expect(0, 918000, '\p{_VS}', "");
Expect(1, 918000, '\p{^_VS}', "");
Expect(1, 918000, '\P{_VS}', "");
Expect(0, 918000, '\P{^_VS}', "");
Error('\p{  Is_VS:=}');
Error('\P{  Is_VS:=}');
Expect(1, 917999, '\p{isvs}', "");
Expect(0, 917999, '\p{^isvs}', "");
Expect(0, 917999, '\P{isvs}', "");
Expect(1, 917999, '\P{^isvs}', "");
Expect(0, 918000, '\p{isvs}', "");
Expect(1, 918000, '\p{^isvs}', "");
Expect(1, 918000, '\P{isvs}', "");
Expect(0, 918000, '\P{^isvs}', "");
Expect(1, 917999, '\p{	 is_vs}', "");
Expect(0, 917999, '\p{^	 is_vs}', "");
Expect(0, 917999, '\P{	 is_vs}', "");
Expect(1, 917999, '\P{^	 is_vs}', "");
Expect(0, 918000, '\p{	 is_vs}', "");
Expect(1, 918000, '\p{^	 is_vs}', "");
Expect(1, 918000, '\P{	 is_vs}', "");
Expect(0, 918000, '\P{^	 is_vs}', "");
Error('\p{_-Rumi_Numeral_Symbols/a/}');
Error('\P{_-Rumi_Numeral_Symbols/a/}');
Expect(1, 69247, '\p{ruminumeralsymbols}', "");
Expect(0, 69247, '\p{^ruminumeralsymbols}', "");
Expect(0, 69247, '\P{ruminumeralsymbols}', "");
Expect(1, 69247, '\P{^ruminumeralsymbols}', "");
Expect(0, 69248, '\p{ruminumeralsymbols}', "");
Expect(1, 69248, '\p{^ruminumeralsymbols}', "");
Expect(1, 69248, '\P{ruminumeralsymbols}', "");
Expect(0, 69248, '\P{^ruminumeralsymbols}', "");
Expect(1, 69247, '\p{	rumi_NUMERAL_SYMBOLS}', "");
Expect(0, 69247, '\p{^	rumi_NUMERAL_SYMBOLS}', "");
Expect(0, 69247, '\P{	rumi_NUMERAL_SYMBOLS}', "");
Expect(1, 69247, '\P{^	rumi_NUMERAL_SYMBOLS}', "");
Expect(0, 69248, '\p{	rumi_NUMERAL_SYMBOLS}', "");
Expect(1, 69248, '\p{^	rumi_NUMERAL_SYMBOLS}', "");
Expect(1, 69248, '\P{	rumi_NUMERAL_SYMBOLS}', "");
Expect(0, 69248, '\P{^	rumi_NUMERAL_SYMBOLS}', "");
Error('\p{:=_	IS_Rumi_numeral_symbols}');
Error('\P{:=_	IS_Rumi_numeral_symbols}');
Expect(1, 69247, '\p{isruminumeralsymbols}', "");
Expect(0, 69247, '\p{^isruminumeralsymbols}', "");
Expect(0, 69247, '\P{isruminumeralsymbols}', "");
Expect(1, 69247, '\P{^isruminumeralsymbols}', "");
Expect(0, 69248, '\p{isruminumeralsymbols}', "");
Expect(1, 69248, '\p{^isruminumeralsymbols}', "");
Expect(1, 69248, '\P{isruminumeralsymbols}', "");
Expect(0, 69248, '\P{^isruminumeralsymbols}', "");
Expect(1, 69247, '\p{ -Is_Rumi_NUMERAL_SYMBOLS}', "");
Expect(0, 69247, '\p{^ -Is_Rumi_NUMERAL_SYMBOLS}', "");
Expect(0, 69247, '\P{ -Is_Rumi_NUMERAL_SYMBOLS}', "");
Expect(1, 69247, '\P{^ -Is_Rumi_NUMERAL_SYMBOLS}', "");
Expect(0, 69248, '\p{ -Is_Rumi_NUMERAL_SYMBOLS}', "");
Expect(1, 69248, '\p{^ -Is_Rumi_NUMERAL_SYMBOLS}', "");
Expect(1, 69248, '\P{ -Is_Rumi_NUMERAL_SYMBOLS}', "");
Expect(0, 69248, '\P{^ -Is_Rumi_NUMERAL_SYMBOLS}', "");
Error('\p{:=_-In_Rumi_Numeral_symbols}');
Error('\P{:=_-In_Rumi_Numeral_symbols}');
Expect(1, 69247, '\p{inruminumeralsymbols}', "");
Expect(0, 69247, '\p{^inruminumeralsymbols}', "");
Expect(0, 69247, '\P{inruminumeralsymbols}', "");
Expect(1, 69247, '\P{^inruminumeralsymbols}', "");
Expect(0, 69248, '\p{inruminumeralsymbols}', "");
Expect(1, 69248, '\p{^inruminumeralsymbols}', "");
Expect(1, 69248, '\P{inruminumeralsymbols}', "");
Expect(0, 69248, '\P{^inruminumeralsymbols}', "");
Expect(1, 69247, '\p{	-In_RUMI_Numeral_Symbols}', "");
Expect(0, 69247, '\p{^	-In_RUMI_Numeral_Symbols}', "");
Expect(0, 69247, '\P{	-In_RUMI_Numeral_Symbols}', "");
Expect(1, 69247, '\P{^	-In_RUMI_Numeral_Symbols}', "");
Expect(0, 69248, '\p{	-In_RUMI_Numeral_Symbols}', "");
Expect(1, 69248, '\p{^	-In_RUMI_Numeral_Symbols}', "");
Expect(1, 69248, '\P{	-In_RUMI_Numeral_Symbols}', "");
Expect(0, 69248, '\P{^	-In_RUMI_Numeral_Symbols}', "");
Error('\p{	/a/Rumi}');
Error('\P{	/a/Rumi}');
Expect(1, 69247, '\p{rumi}', "");
Expect(0, 69247, '\p{^rumi}', "");
Expect(0, 69247, '\P{rumi}', "");
Expect(1, 69247, '\P{^rumi}', "");
Expect(0, 69248, '\p{rumi}', "");
Expect(1, 69248, '\p{^rumi}', "");
Expect(1, 69248, '\P{rumi}', "");
Expect(0, 69248, '\P{^rumi}', "");
Expect(1, 69247, '\p{--RUMI}', "");
Expect(0, 69247, '\p{^--RUMI}', "");
Expect(0, 69247, '\P{--RUMI}', "");
Expect(1, 69247, '\P{^--RUMI}', "");
Expect(0, 69248, '\p{--RUMI}', "");
Expect(1, 69248, '\p{^--RUMI}', "");
Expect(1, 69248, '\P{--RUMI}', "");
Expect(0, 69248, '\P{^--RUMI}', "");
Error('\p{__IS_RUMI/a/}');
Error('\P{__IS_RUMI/a/}');
Expect(1, 69247, '\p{isrumi}', "");
Expect(0, 69247, '\p{^isrumi}', "");
Expect(0, 69247, '\P{isrumi}', "");
Expect(1, 69247, '\P{^isrumi}', "");
Expect(0, 69248, '\p{isrumi}', "");
Expect(1, 69248, '\p{^isrumi}', "");
Expect(1, 69248, '\P{isrumi}', "");
Expect(0, 69248, '\P{^isrumi}', "");
Expect(1, 69247, '\p{-Is_Rumi}', "");
Expect(0, 69247, '\p{^-Is_Rumi}', "");
Expect(0, 69247, '\P{-Is_Rumi}', "");
Expect(1, 69247, '\P{^-Is_Rumi}', "");
Expect(0, 69248, '\p{-Is_Rumi}', "");
Expect(1, 69248, '\p{^-Is_Rumi}', "");
Expect(1, 69248, '\P{-Is_Rumi}', "");
Expect(0, 69248, '\P{^-Is_Rumi}', "");
Error('\p{:=	_IN_Rumi}');
Error('\P{:=	_IN_Rumi}');
Expect(1, 69247, '\p{inrumi}', "");
Expect(0, 69247, '\p{^inrumi}', "");
Expect(0, 69247, '\P{inrumi}', "");
Expect(1, 69247, '\P{^inrumi}', "");
Expect(0, 69248, '\p{inrumi}', "");
Expect(1, 69248, '\p{^inrumi}', "");
Expect(1, 69248, '\P{inrumi}', "");
Expect(0, 69248, '\P{^inrumi}', "");
Expect(1, 69247, '\p{in_Rumi}', "");
Expect(0, 69247, '\p{^in_Rumi}', "");
Expect(0, 69247, '\P{in_Rumi}', "");
Expect(1, 69247, '\P{^in_Rumi}', "");
Expect(0, 69248, '\p{in_Rumi}', "");
Expect(1, 69248, '\p{^in_Rumi}', "");
Expect(1, 69248, '\P{in_Rumi}', "");
Expect(0, 69248, '\P{^in_Rumi}', "");
Error('\p{/a/Ethiopic_EXTENDED}');
Error('\P{/a/Ethiopic_EXTENDED}');
Expect(1, 11743, '\p{ethiopicextended}', "");
Expect(0, 11743, '\p{^ethiopicextended}', "");
Expect(0, 11743, '\P{ethiopicextended}', "");
Expect(1, 11743, '\P{^ethiopicextended}', "");
Expect(0, 11744, '\p{ethiopicextended}', "");
Expect(1, 11744, '\p{^ethiopicextended}', "");
Expect(1, 11744, '\P{ethiopicextended}', "");
Expect(0, 11744, '\P{^ethiopicextended}', "");
Expect(1, 11743, '\p{ _Ethiopic_extended}', "");
Expect(0, 11743, '\p{^ _Ethiopic_extended}', "");
Expect(0, 11743, '\P{ _Ethiopic_extended}', "");
Expect(1, 11743, '\P{^ _Ethiopic_extended}', "");
Expect(0, 11744, '\p{ _Ethiopic_extended}', "");
Expect(1, 11744, '\p{^ _Ethiopic_extended}', "");
Expect(1, 11744, '\P{ _Ethiopic_extended}', "");
Expect(0, 11744, '\P{^ _Ethiopic_extended}', "");
Error('\p{  IS_ETHIOPIC_EXTENDED:=}');
Error('\P{  IS_ETHIOPIC_EXTENDED:=}');
Expect(1, 11743, '\p{isethiopicextended}', "");
Expect(0, 11743, '\p{^isethiopicextended}', "");
Expect(0, 11743, '\P{isethiopicextended}', "");
Expect(1, 11743, '\P{^isethiopicextended}', "");
Expect(0, 11744, '\p{isethiopicextended}', "");
Expect(1, 11744, '\p{^isethiopicextended}', "");
Expect(1, 11744, '\P{isethiopicextended}', "");
Expect(0, 11744, '\P{^isethiopicextended}', "");
Expect(1, 11743, '\p{is_Ethiopic_Extended}', "");
Expect(0, 11743, '\p{^is_Ethiopic_Extended}', "");
Expect(0, 11743, '\P{is_Ethiopic_Extended}', "");
Expect(1, 11743, '\P{^is_Ethiopic_Extended}', "");
Expect(0, 11744, '\p{is_Ethiopic_Extended}', "");
Expect(1, 11744, '\p{^is_Ethiopic_Extended}', "");
Expect(1, 11744, '\P{is_Ethiopic_Extended}', "");
Expect(0, 11744, '\P{^is_Ethiopic_Extended}', "");
Error('\p{	In_ETHIOPIC_Extended:=}');
Error('\P{	In_ETHIOPIC_Extended:=}');
Expect(1, 11743, '\p{inethiopicextended}', "");
Expect(0, 11743, '\p{^inethiopicextended}', "");
Expect(0, 11743, '\P{inethiopicextended}', "");
Expect(1, 11743, '\P{^inethiopicextended}', "");
Expect(0, 11744, '\p{inethiopicextended}', "");
Expect(1, 11744, '\p{^inethiopicextended}', "");
Expect(1, 11744, '\P{inethiopicextended}', "");
Expect(0, 11744, '\P{^inethiopicextended}', "");
Expect(1, 11743, '\p{	 IN_Ethiopic_Extended}', "");
Expect(0, 11743, '\p{^	 IN_Ethiopic_Extended}', "");
Expect(0, 11743, '\P{	 IN_Ethiopic_Extended}', "");
Expect(1, 11743, '\P{^	 IN_Ethiopic_Extended}', "");
Expect(0, 11744, '\p{	 IN_Ethiopic_Extended}', "");
Expect(1, 11744, '\p{^	 IN_Ethiopic_Extended}', "");
Expect(1, 11744, '\P{	 IN_Ethiopic_Extended}', "");
Expect(0, 11744, '\P{^	 IN_Ethiopic_Extended}', "");
Error('\p{_:=ETHIOPIC_EXT}');
Error('\P{_:=ETHIOPIC_EXT}');
Expect(1, 11743, '\p{ethiopicext}', "");
Expect(0, 11743, '\p{^ethiopicext}', "");
Expect(0, 11743, '\P{ethiopicext}', "");
Expect(1, 11743, '\P{^ethiopicext}', "");
Expect(0, 11744, '\p{ethiopicext}', "");
Expect(1, 11744, '\p{^ethiopicext}', "");
Expect(1, 11744, '\P{ethiopicext}', "");
Expect(0, 11744, '\P{^ethiopicext}', "");
Expect(1, 11743, '\p{	ETHIOPIC_Ext}', "");
Expect(0, 11743, '\p{^	ETHIOPIC_Ext}', "");
Expect(0, 11743, '\P{	ETHIOPIC_Ext}', "");
Expect(1, 11743, '\P{^	ETHIOPIC_Ext}', "");
Expect(0, 11744, '\p{	ETHIOPIC_Ext}', "");
Expect(1, 11744, '\p{^	ETHIOPIC_Ext}', "");
Expect(1, 11744, '\P{	ETHIOPIC_Ext}', "");
Expect(0, 11744, '\P{^	ETHIOPIC_Ext}', "");
Error('\p{-:=is_Ethiopic_EXT}');
Error('\P{-:=is_Ethiopic_EXT}');
Expect(1, 11743, '\p{isethiopicext}', "");
Expect(0, 11743, '\p{^isethiopicext}', "");
Expect(0, 11743, '\P{isethiopicext}', "");
Expect(1, 11743, '\P{^isethiopicext}', "");
Expect(0, 11744, '\p{isethiopicext}', "");
Expect(1, 11744, '\p{^isethiopicext}', "");
Expect(1, 11744, '\P{isethiopicext}', "");
Expect(0, 11744, '\P{^isethiopicext}', "");
Expect(1, 11743, '\p{-_Is_Ethiopic_EXT}', "");
Expect(0, 11743, '\p{^-_Is_Ethiopic_EXT}', "");
Expect(0, 11743, '\P{-_Is_Ethiopic_EXT}', "");
Expect(1, 11743, '\P{^-_Is_Ethiopic_EXT}', "");
Expect(0, 11744, '\p{-_Is_Ethiopic_EXT}', "");
Expect(1, 11744, '\p{^-_Is_Ethiopic_EXT}', "");
Expect(1, 11744, '\P{-_Is_Ethiopic_EXT}', "");
Expect(0, 11744, '\P{^-_Is_Ethiopic_EXT}', "");
Error('\p{:=	-IN_Ethiopic_Ext}');
Error('\P{:=	-IN_Ethiopic_Ext}');
Expect(1, 11743, '\p{inethiopicext}', "");
Expect(0, 11743, '\p{^inethiopicext}', "");
Expect(0, 11743, '\P{inethiopicext}', "");
Expect(1, 11743, '\P{^inethiopicext}', "");
Expect(0, 11744, '\p{inethiopicext}', "");
Expect(1, 11744, '\p{^inethiopicext}', "");
Expect(1, 11744, '\P{inethiopicext}', "");
Expect(0, 11744, '\P{^inethiopicext}', "");
Expect(1, 11743, '\p{In_ethiopic_EXT}', "");
Expect(0, 11743, '\p{^In_ethiopic_EXT}', "");
Expect(0, 11743, '\P{In_ethiopic_EXT}', "");
Expect(1, 11743, '\P{^In_ethiopic_EXT}', "");
Expect(0, 11744, '\p{In_ethiopic_EXT}', "");
Expect(1, 11744, '\p{^In_ethiopic_EXT}', "");
Expect(1, 11744, '\P{In_ethiopic_EXT}', "");
Expect(0, 11744, '\P{^In_ethiopic_EXT}', "");
Error('\p{/a/	-CYPRIOT_Syllabary}');
Error('\P{/a/	-CYPRIOT_Syllabary}');
Expect(1, 67647, '\p{cypriotsyllabary}', "");
Expect(0, 67647, '\p{^cypriotsyllabary}', "");
Expect(0, 67647, '\P{cypriotsyllabary}', "");
Expect(1, 67647, '\P{^cypriotsyllabary}', "");
Expect(0, 67648, '\p{cypriotsyllabary}', "");
Expect(1, 67648, '\p{^cypriotsyllabary}', "");
Expect(1, 67648, '\P{cypriotsyllabary}', "");
Expect(0, 67648, '\P{^cypriotsyllabary}', "");
Expect(1, 67647, '\p{  CYPRIOT_Syllabary}', "");
Expect(0, 67647, '\p{^  CYPRIOT_Syllabary}', "");
Expect(0, 67647, '\P{  CYPRIOT_Syllabary}', "");
Expect(1, 67647, '\P{^  CYPRIOT_Syllabary}', "");
Expect(0, 67648, '\p{  CYPRIOT_Syllabary}', "");
Expect(1, 67648, '\p{^  CYPRIOT_Syllabary}', "");
Expect(1, 67648, '\P{  CYPRIOT_Syllabary}', "");
Expect(0, 67648, '\P{^  CYPRIOT_Syllabary}', "");
Error('\p{_is_Cypriot_Syllabary:=}');
Error('\P{_is_Cypriot_Syllabary:=}');
Expect(1, 67647, '\p{iscypriotsyllabary}', "");
Expect(0, 67647, '\p{^iscypriotsyllabary}', "");
Expect(0, 67647, '\P{iscypriotsyllabary}', "");
Expect(1, 67647, '\P{^iscypriotsyllabary}', "");
Expect(0, 67648, '\p{iscypriotsyllabary}', "");
Expect(1, 67648, '\p{^iscypriotsyllabary}', "");
Expect(1, 67648, '\P{iscypriotsyllabary}', "");
Expect(0, 67648, '\P{^iscypriotsyllabary}', "");
Expect(1, 67647, '\p{ Is_cypriot_syllabary}', "");
Expect(0, 67647, '\p{^ Is_cypriot_syllabary}', "");
Expect(0, 67647, '\P{ Is_cypriot_syllabary}', "");
Expect(1, 67647, '\P{^ Is_cypriot_syllabary}', "");
Expect(0, 67648, '\p{ Is_cypriot_syllabary}', "");
Expect(1, 67648, '\p{^ Is_cypriot_syllabary}', "");
Expect(1, 67648, '\P{ Is_cypriot_syllabary}', "");
Expect(0, 67648, '\P{^ Is_cypriot_syllabary}', "");
Error('\p{	/a/IN_CYPRIOT_Syllabary}');
Error('\P{	/a/IN_CYPRIOT_Syllabary}');
Expect(1, 67647, '\p{incypriotsyllabary}', "");
Expect(0, 67647, '\p{^incypriotsyllabary}', "");
Expect(0, 67647, '\P{incypriotsyllabary}', "");
Expect(1, 67647, '\P{^incypriotsyllabary}', "");
Expect(0, 67648, '\p{incypriotsyllabary}', "");
Expect(1, 67648, '\p{^incypriotsyllabary}', "");
Expect(1, 67648, '\P{incypriotsyllabary}', "");
Expect(0, 67648, '\P{^incypriotsyllabary}', "");
Expect(1, 67647, '\p{_IN_CYPRIOT_SYLLABARY}', "");
Expect(0, 67647, '\p{^_IN_CYPRIOT_SYLLABARY}', "");
Expect(0, 67647, '\P{_IN_CYPRIOT_SYLLABARY}', "");
Expect(1, 67647, '\P{^_IN_CYPRIOT_SYLLABARY}', "");
Expect(0, 67648, '\p{_IN_CYPRIOT_SYLLABARY}', "");
Expect(1, 67648, '\p{^_IN_CYPRIOT_SYLLABARY}', "");
Expect(1, 67648, '\P{_IN_CYPRIOT_SYLLABARY}', "");
Expect(0, 67648, '\P{^_IN_CYPRIOT_SYLLABARY}', "");
Error('\p{_-Coptic:=}');
Error('\P{_-Coptic:=}');
Expect(1, 11519, '\p{coptic}', "");
Expect(0, 11519, '\p{^coptic}', "");
Expect(0, 11519, '\P{coptic}', "");
Expect(1, 11519, '\P{^coptic}', "");
Expect(0, 11520, '\p{coptic}', "");
Expect(1, 11520, '\p{^coptic}', "");
Expect(1, 11520, '\P{coptic}', "");
Expect(0, 11520, '\P{^coptic}', "");
Expect(1, 11519, '\p{  coptic}', "");
Expect(0, 11519, '\p{^  coptic}', "");
Expect(0, 11519, '\P{  coptic}', "");
Expect(1, 11519, '\P{^  coptic}', "");
Expect(0, 11520, '\p{  coptic}', "");
Expect(1, 11520, '\p{^  coptic}', "");
Expect(1, 11520, '\P{  coptic}', "");
Expect(0, 11520, '\P{^  coptic}', "");
Error('\p{/a/_IS_Coptic}');
Error('\P{/a/_IS_Coptic}');
Expect(1, 11519, '\p{iscoptic}', "");
Expect(0, 11519, '\p{^iscoptic}', "");
Expect(0, 11519, '\P{iscoptic}', "");
Expect(1, 11519, '\P{^iscoptic}', "");
Expect(0, 11520, '\p{iscoptic}', "");
Expect(1, 11520, '\p{^iscoptic}', "");
Expect(1, 11520, '\P{iscoptic}', "");
Expect(0, 11520, '\P{^iscoptic}', "");
Expect(1, 11519, '\p{ _Is_coptic}', "");
Expect(0, 11519, '\p{^ _Is_coptic}', "");
Expect(0, 11519, '\P{ _Is_coptic}', "");
Expect(1, 11519, '\P{^ _Is_coptic}', "");
Expect(0, 11520, '\p{ _Is_coptic}', "");
Expect(1, 11520, '\p{^ _Is_coptic}', "");
Expect(1, 11520, '\P{ _Is_coptic}', "");
Expect(0, 11520, '\P{^ _Is_coptic}', "");
Error('\p{/a/-_COPT}');
Error('\P{/a/-_COPT}');
Expect(1, 11519, '\p{copt}', "");
Expect(0, 11519, '\p{^copt}', "");
Expect(0, 11519, '\P{copt}', "");
Expect(1, 11519, '\P{^copt}', "");
Expect(0, 11520, '\p{copt}', "");
Expect(1, 11520, '\p{^copt}', "");
Expect(1, 11520, '\P{copt}', "");
Expect(0, 11520, '\P{^copt}', "");
Expect(1, 11519, '\p{	Copt}', "");
Expect(0, 11519, '\p{^	Copt}', "");
Expect(0, 11519, '\P{	Copt}', "");
Expect(1, 11519, '\P{^	Copt}', "");
Expect(0, 11520, '\p{	Copt}', "");
Expect(1, 11520, '\p{^	Copt}', "");
Expect(1, 11520, '\P{	Copt}', "");
Expect(0, 11520, '\P{^	Copt}', "");
Error('\p{_:=is_COPT}');
Error('\P{_:=is_COPT}');
Expect(1, 11519, '\p{iscopt}', "");
Expect(0, 11519, '\p{^iscopt}', "");
Expect(0, 11519, '\P{iscopt}', "");
Expect(1, 11519, '\P{^iscopt}', "");
Expect(0, 11520, '\p{iscopt}', "");
Expect(1, 11520, '\p{^iscopt}', "");
Expect(1, 11520, '\P{iscopt}', "");
Expect(0, 11520, '\P{^iscopt}', "");
Expect(1, 11519, '\p{  is_Copt}', "");
Expect(0, 11519, '\p{^  is_Copt}', "");
Expect(0, 11519, '\P{  is_Copt}', "");
Expect(1, 11519, '\P{^  is_Copt}', "");
Expect(0, 11520, '\p{  is_Copt}', "");
Expect(1, 11520, '\p{^  is_Copt}', "");
Expect(1, 11520, '\P{  is_Copt}', "");
Expect(0, 11520, '\P{^  is_Copt}', "");
Error('\p{	:=qaac}');
Error('\P{	:=qaac}');
Expect(1, 11519, '\p{qaac}', "");
Expect(0, 11519, '\p{^qaac}', "");
Expect(0, 11519, '\P{qaac}', "");
Expect(1, 11519, '\P{^qaac}', "");
Expect(0, 11520, '\p{qaac}', "");
Expect(1, 11520, '\p{^qaac}', "");
Expect(1, 11520, '\P{qaac}', "");
Expect(0, 11520, '\P{^qaac}', "");
Expect(1, 11519, '\p{_-Qaac}', "");
Expect(0, 11519, '\p{^_-Qaac}', "");
Expect(0, 11519, '\P{_-Qaac}', "");
Expect(1, 11519, '\P{^_-Qaac}', "");
Expect(0, 11520, '\p{_-Qaac}', "");
Expect(1, 11520, '\p{^_-Qaac}', "");
Expect(1, 11520, '\P{_-Qaac}', "");
Expect(0, 11520, '\P{^_-Qaac}', "");
Error('\p{:=Is_Qaac}');
Error('\P{:=Is_Qaac}');
Expect(1, 11519, '\p{isqaac}', "");
Expect(0, 11519, '\p{^isqaac}', "");
Expect(0, 11519, '\P{isqaac}', "");
Expect(1, 11519, '\P{^isqaac}', "");
Expect(0, 11520, '\p{isqaac}', "");
Expect(1, 11520, '\p{^isqaac}', "");
Expect(1, 11520, '\P{isqaac}', "");
Expect(0, 11520, '\P{^isqaac}', "");
Expect(1, 11519, '\p{is_QAAC}', "");
Expect(0, 11519, '\p{^is_QAAC}', "");
Expect(0, 11519, '\P{is_QAAC}', "");
Expect(1, 11519, '\P{^is_QAAC}', "");
Expect(0, 11520, '\p{is_QAAC}', "");
Expect(1, 11520, '\p{^is_QAAC}', "");
Expect(1, 11520, '\P{is_QAAC}', "");
Expect(0, 11520, '\P{^is_QAAC}', "");
Error('\p{ Supplemental_arrows_b:=}');
Error('\P{ Supplemental_arrows_b:=}');
Expect(1, 10623, '\p{supplementalarrowsb}', "");
Expect(0, 10623, '\p{^supplementalarrowsb}', "");
Expect(0, 10623, '\P{supplementalarrowsb}', "");
Expect(1, 10623, '\P{^supplementalarrowsb}', "");
Expect(0, 10624, '\p{supplementalarrowsb}', "");
Expect(1, 10624, '\p{^supplementalarrowsb}', "");
Expect(1, 10624, '\P{supplementalarrowsb}', "");
Expect(0, 10624, '\P{^supplementalarrowsb}', "");
Expect(1, 10623, '\p{	SUPPLEMENTAL_arrows_B}', "");
Expect(0, 10623, '\p{^	SUPPLEMENTAL_arrows_B}', "");
Expect(0, 10623, '\P{	SUPPLEMENTAL_arrows_B}', "");
Expect(1, 10623, '\P{^	SUPPLEMENTAL_arrows_B}', "");
Expect(0, 10624, '\p{	SUPPLEMENTAL_arrows_B}', "");
Expect(1, 10624, '\p{^	SUPPLEMENTAL_arrows_B}', "");
Expect(1, 10624, '\P{	SUPPLEMENTAL_arrows_B}', "");
Expect(0, 10624, '\P{^	SUPPLEMENTAL_arrows_B}', "");
Error('\p{	/a/Is_Supplemental_ARROWS_b}');
Error('\P{	/a/Is_Supplemental_ARROWS_b}');
Expect(1, 10623, '\p{issupplementalarrowsb}', "");
Expect(0, 10623, '\p{^issupplementalarrowsb}', "");
Expect(0, 10623, '\P{issupplementalarrowsb}', "");
Expect(1, 10623, '\P{^issupplementalarrowsb}', "");
Expect(0, 10624, '\p{issupplementalarrowsb}', "");
Expect(1, 10624, '\p{^issupplementalarrowsb}', "");
Expect(1, 10624, '\P{issupplementalarrowsb}', "");
Expect(0, 10624, '\P{^issupplementalarrowsb}', "");
Expect(1, 10623, '\p{ is_Supplemental_arrows_b}', "");
Expect(0, 10623, '\p{^ is_Supplemental_arrows_b}', "");
Expect(0, 10623, '\P{ is_Supplemental_arrows_b}', "");
Expect(1, 10623, '\P{^ is_Supplemental_arrows_b}', "");
Expect(0, 10624, '\p{ is_Supplemental_arrows_b}', "");
Expect(1, 10624, '\p{^ is_Supplemental_arrows_b}', "");
Expect(1, 10624, '\P{ is_Supplemental_arrows_b}', "");
Expect(0, 10624, '\P{^ is_Supplemental_arrows_b}', "");
Error('\p{_	in_Supplemental_ARROWS_B/a/}');
Error('\P{_	in_Supplemental_ARROWS_B/a/}');
Expect(1, 10623, '\p{insupplementalarrowsb}', "");
Expect(0, 10623, '\p{^insupplementalarrowsb}', "");
Expect(0, 10623, '\P{insupplementalarrowsb}', "");
Expect(1, 10623, '\P{^insupplementalarrowsb}', "");
Expect(0, 10624, '\p{insupplementalarrowsb}', "");
Expect(1, 10624, '\p{^insupplementalarrowsb}', "");
Expect(1, 10624, '\P{insupplementalarrowsb}', "");
Expect(0, 10624, '\P{^insupplementalarrowsb}', "");
Expect(1, 10623, '\p{ -In_Supplemental_Arrows_B}', "");
Expect(0, 10623, '\p{^ -In_Supplemental_Arrows_B}', "");
Expect(0, 10623, '\P{ -In_Supplemental_Arrows_B}', "");
Expect(1, 10623, '\P{^ -In_Supplemental_Arrows_B}', "");
Expect(0, 10624, '\p{ -In_Supplemental_Arrows_B}', "");
Expect(1, 10624, '\p{^ -In_Supplemental_Arrows_B}', "");
Expect(1, 10624, '\P{ -In_Supplemental_Arrows_B}', "");
Expect(0, 10624, '\P{^ -In_Supplemental_Arrows_B}', "");
Error('\p{:=	Sup_Arrows_B}');
Error('\P{:=	Sup_Arrows_B}');
Expect(1, 10623, '\p{suparrowsb}', "");
Expect(0, 10623, '\p{^suparrowsb}', "");
Expect(0, 10623, '\P{suparrowsb}', "");
Expect(1, 10623, '\P{^suparrowsb}', "");
Expect(0, 10624, '\p{suparrowsb}', "");
Expect(1, 10624, '\p{^suparrowsb}', "");
Expect(1, 10624, '\P{suparrowsb}', "");
Expect(0, 10624, '\P{^suparrowsb}', "");
Expect(1, 10623, '\p{	-sup_ARROWS_B}', "");
Expect(0, 10623, '\p{^	-sup_ARROWS_B}', "");
Expect(0, 10623, '\P{	-sup_ARROWS_B}', "");
Expect(1, 10623, '\P{^	-sup_ARROWS_B}', "");
Expect(0, 10624, '\p{	-sup_ARROWS_B}', "");
Expect(1, 10624, '\p{^	-sup_ARROWS_B}', "");
Expect(1, 10624, '\P{	-sup_ARROWS_B}', "");
Expect(0, 10624, '\P{^	-sup_ARROWS_B}', "");
Error('\p{- Is_sup_Arrows_B:=}');
Error('\P{- Is_sup_Arrows_B:=}');
Expect(1, 10623, '\p{issuparrowsb}', "");
Expect(0, 10623, '\p{^issuparrowsb}', "");
Expect(0, 10623, '\P{issuparrowsb}', "");
Expect(1, 10623, '\P{^issuparrowsb}', "");
Expect(0, 10624, '\p{issuparrowsb}', "");
Expect(1, 10624, '\p{^issuparrowsb}', "");
Expect(1, 10624, '\P{issuparrowsb}', "");
Expect(0, 10624, '\P{^issuparrowsb}', "");
Expect(1, 10623, '\p{ 	is_SUP_arrows_B}', "");
Expect(0, 10623, '\p{^ 	is_SUP_arrows_B}', "");
Expect(0, 10623, '\P{ 	is_SUP_arrows_B}', "");
Expect(1, 10623, '\P{^ 	is_SUP_arrows_B}', "");
Expect(0, 10624, '\p{ 	is_SUP_arrows_B}', "");
Expect(1, 10624, '\p{^ 	is_SUP_arrows_B}', "");
Expect(1, 10624, '\P{ 	is_SUP_arrows_B}', "");
Expect(0, 10624, '\P{^ 	is_SUP_arrows_B}', "");
Error('\p{/a/In_SUP_arrows_B}');
Error('\P{/a/In_SUP_arrows_B}');
Expect(1, 10623, '\p{insuparrowsb}', "");
Expect(0, 10623, '\p{^insuparrowsb}', "");
Expect(0, 10623, '\P{insuparrowsb}', "");
Expect(1, 10623, '\P{^insuparrowsb}', "");
Expect(0, 10624, '\p{insuparrowsb}', "");
Expect(1, 10624, '\p{^insuparrowsb}', "");
Expect(1, 10624, '\P{insuparrowsb}', "");
Expect(0, 10624, '\P{^insuparrowsb}', "");
Expect(1, 10623, '\p{_ in_sup_Arrows_B}', "");
Expect(0, 10623, '\p{^_ in_sup_Arrows_B}', "");
Expect(0, 10623, '\P{_ in_sup_Arrows_B}', "");
Expect(1, 10623, '\P{^_ in_sup_Arrows_B}', "");
Expect(0, 10624, '\p{_ in_sup_Arrows_B}', "");
Expect(1, 10624, '\p{^_ in_sup_Arrows_B}', "");
Expect(1, 10624, '\P{_ in_sup_Arrows_B}', "");
Expect(0, 10624, '\P{^_ in_sup_Arrows_B}', "");
Error('\p{Expands_On_NFC}');
Error('\P{Expands_On_NFC}');
Error('\p{Is_Expands_On_NFC}');
Error('\P{Is_Expands_On_NFC}');
Error('\p{XO_NFC}');
Error('\P{XO_NFC}');
Error('\p{Is_XO_NFC}');
Error('\P{Is_XO_NFC}');
Error('\p{:=  CJK_COMPATIBILITY_ideographs}');
Error('\P{:=  CJK_COMPATIBILITY_ideographs}');
Expect(1, 64255, '\p{cjkcompatibilityideographs}', "");
Expect(0, 64255, '\p{^cjkcompatibilityideographs}', "");
Expect(0, 64255, '\P{cjkcompatibilityideographs}', "");
Expect(1, 64255, '\P{^cjkcompatibilityideographs}', "");
Expect(0, 64256, '\p{cjkcompatibilityideographs}', "");
Expect(1, 64256, '\p{^cjkcompatibilityideographs}', "");
Expect(1, 64256, '\P{cjkcompatibilityideographs}', "");
Expect(0, 64256, '\P{^cjkcompatibilityideographs}', "");
Expect(1, 64255, '\p{	-cjk_Compatibility_Ideographs}', "");
Expect(0, 64255, '\p{^	-cjk_Compatibility_Ideographs}', "");
Expect(0, 64255, '\P{	-cjk_Compatibility_Ideographs}', "");
Expect(1, 64255, '\P{^	-cjk_Compatibility_Ideographs}', "");
Expect(0, 64256, '\p{	-cjk_Compatibility_Ideographs}', "");
Expect(1, 64256, '\p{^	-cjk_Compatibility_Ideographs}', "");
Expect(1, 64256, '\P{	-cjk_Compatibility_Ideographs}', "");
Expect(0, 64256, '\P{^	-cjk_Compatibility_Ideographs}', "");
Error('\p{:=  is_cjk_Compatibility_Ideographs}');
Error('\P{:=  is_cjk_Compatibility_Ideographs}');
Expect(1, 64255, '\p{iscjkcompatibilityideographs}', "");
Expect(0, 64255, '\p{^iscjkcompatibilityideographs}', "");
Expect(0, 64255, '\P{iscjkcompatibilityideographs}', "");
Expect(1, 64255, '\P{^iscjkcompatibilityideographs}', "");
Expect(0, 64256, '\p{iscjkcompatibilityideographs}', "");
Expect(1, 64256, '\p{^iscjkcompatibilityideographs}', "");
Expect(1, 64256, '\P{iscjkcompatibilityideographs}', "");
Expect(0, 64256, '\P{^iscjkcompatibilityideographs}', "");
Expect(1, 64255, '\p{ Is_CJK_Compatibility_Ideographs}', "");
Expect(0, 64255, '\p{^ Is_CJK_Compatibility_Ideographs}', "");
Expect(0, 64255, '\P{ Is_CJK_Compatibility_Ideographs}', "");
Expect(1, 64255, '\P{^ Is_CJK_Compatibility_Ideographs}', "");
Expect(0, 64256, '\p{ Is_CJK_Compatibility_Ideographs}', "");
Expect(1, 64256, '\p{^ Is_CJK_Compatibility_Ideographs}', "");
Expect(1, 64256, '\P{ Is_CJK_Compatibility_Ideographs}', "");
Expect(0, 64256, '\P{^ Is_CJK_Compatibility_Ideographs}', "");
Error('\p{_:=IN_CJK_Compatibility_IDEOGRAPHS}');
Error('\P{_:=IN_CJK_Compatibility_IDEOGRAPHS}');
Expect(1, 64255, '\p{incjkcompatibilityideographs}', "");
Expect(0, 64255, '\p{^incjkcompatibilityideographs}', "");
Expect(0, 64255, '\P{incjkcompatibilityideographs}', "");
Expect(1, 64255, '\P{^incjkcompatibilityideographs}', "");
Expect(0, 64256, '\p{incjkcompatibilityideographs}', "");
Expect(1, 64256, '\p{^incjkcompatibilityideographs}', "");
Expect(1, 64256, '\P{incjkcompatibilityideographs}', "");
Expect(0, 64256, '\P{^incjkcompatibilityideographs}', "");
Expect(1, 64255, '\p{  IN_CJK_Compatibility_Ideographs}', "");
Expect(0, 64255, '\p{^  IN_CJK_Compatibility_Ideographs}', "");
Expect(0, 64255, '\P{  IN_CJK_Compatibility_Ideographs}', "");
Expect(1, 64255, '\P{^  IN_CJK_Compatibility_Ideographs}', "");
Expect(0, 64256, '\p{  IN_CJK_Compatibility_Ideographs}', "");
Expect(1, 64256, '\p{^  IN_CJK_Compatibility_Ideographs}', "");
Expect(1, 64256, '\P{  IN_CJK_Compatibility_Ideographs}', "");
Expect(0, 64256, '\P{^  IN_CJK_Compatibility_Ideographs}', "");
Error('\p{_cjk_Compat_IDEOGRAPHS/a/}');
Error('\P{_cjk_Compat_IDEOGRAPHS/a/}');
Expect(1, 64255, '\p{cjkcompatideographs}', "");
Expect(0, 64255, '\p{^cjkcompatideographs}', "");
Expect(0, 64255, '\P{cjkcompatideographs}', "");
Expect(1, 64255, '\P{^cjkcompatideographs}', "");
Expect(0, 64256, '\p{cjkcompatideographs}', "");
Expect(1, 64256, '\p{^cjkcompatideographs}', "");
Expect(1, 64256, '\P{cjkcompatideographs}', "");
Expect(0, 64256, '\P{^cjkcompatideographs}', "");
Expect(1, 64255, '\p{	cjk_compat_Ideographs}', "");
Expect(0, 64255, '\p{^	cjk_compat_Ideographs}', "");
Expect(0, 64255, '\P{	cjk_compat_Ideographs}', "");
Expect(1, 64255, '\P{^	cjk_compat_Ideographs}', "");
Expect(0, 64256, '\p{	cjk_compat_Ideographs}', "");
Expect(1, 64256, '\p{^	cjk_compat_Ideographs}', "");
Expect(1, 64256, '\P{	cjk_compat_Ideographs}', "");
Expect(0, 64256, '\P{^	cjk_compat_Ideographs}', "");
Error('\p{	-Is_cjk_COMPAT_Ideographs:=}');
Error('\P{	-Is_cjk_COMPAT_Ideographs:=}');
Expect(1, 64255, '\p{iscjkcompatideographs}', "");
Expect(0, 64255, '\p{^iscjkcompatideographs}', "");
Expect(0, 64255, '\P{iscjkcompatideographs}', "");
Expect(1, 64255, '\P{^iscjkcompatideographs}', "");
Expect(0, 64256, '\p{iscjkcompatideographs}', "");
Expect(1, 64256, '\p{^iscjkcompatideographs}', "");
Expect(1, 64256, '\P{iscjkcompatideographs}', "");
Expect(0, 64256, '\P{^iscjkcompatideographs}', "");
Expect(1, 64255, '\p{	 IS_CJK_compat_Ideographs}', "");
Expect(0, 64255, '\p{^	 IS_CJK_compat_Ideographs}', "");
Expect(0, 64255, '\P{	 IS_CJK_compat_Ideographs}', "");
Expect(1, 64255, '\P{^	 IS_CJK_compat_Ideographs}', "");
Expect(0, 64256, '\p{	 IS_CJK_compat_Ideographs}', "");
Expect(1, 64256, '\p{^	 IS_CJK_compat_Ideographs}', "");
Expect(1, 64256, '\P{	 IS_CJK_compat_Ideographs}', "");
Expect(0, 64256, '\P{^	 IS_CJK_compat_Ideographs}', "");
Error('\p{_-In_CJK_COMPAT_Ideographs:=}');
Error('\P{_-In_CJK_COMPAT_Ideographs:=}');
Expect(1, 64255, '\p{incjkcompatideographs}', "");
Expect(0, 64255, '\p{^incjkcompatideographs}', "");
Expect(0, 64255, '\P{incjkcompatideographs}', "");
Expect(1, 64255, '\P{^incjkcompatideographs}', "");
Expect(0, 64256, '\p{incjkcompatideographs}', "");
Expect(1, 64256, '\p{^incjkcompatideographs}', "");
Expect(1, 64256, '\P{incjkcompatideographs}', "");
Expect(0, 64256, '\P{^incjkcompatideographs}', "");
Expect(1, 64255, '\p{_	In_CJK_Compat_ideographs}', "");
Expect(0, 64255, '\p{^_	In_CJK_Compat_ideographs}', "");
Expect(0, 64255, '\P{_	In_CJK_Compat_ideographs}', "");
Expect(1, 64255, '\P{^_	In_CJK_Compat_ideographs}', "");
Expect(0, 64256, '\p{_	In_CJK_Compat_ideographs}', "");
Expect(1, 64256, '\p{^_	In_CJK_Compat_ideographs}', "");
Expect(1, 64256, '\P{_	In_CJK_Compat_ideographs}', "");
Expect(0, 64256, '\P{^_	In_CJK_Compat_ideographs}', "");
Error('\p{:=inherited}');
Error('\P{:=inherited}');
Expect(1, 917999, '\p{inherited}', "");
Expect(0, 917999, '\p{^inherited}', "");
Expect(0, 917999, '\P{inherited}', "");
Expect(1, 917999, '\P{^inherited}', "");
Expect(0, 918000, '\p{inherited}', "");
Expect(1, 918000, '\p{^inherited}', "");
Expect(1, 918000, '\P{inherited}', "");
Expect(0, 918000, '\P{^inherited}', "");
Expect(1, 917999, '\p{-inherited}', "");
Expect(0, 917999, '\p{^-inherited}', "");
Expect(0, 917999, '\P{-inherited}', "");
Expect(1, 917999, '\P{^-inherited}', "");
Expect(0, 918000, '\p{-inherited}', "");
Expect(1, 918000, '\p{^-inherited}', "");
Expect(1, 918000, '\P{-inherited}', "");
Expect(0, 918000, '\P{^-inherited}', "");
Error('\p{_Is_INHERITED/a/}');
Error('\P{_Is_INHERITED/a/}');
Expect(1, 917999, '\p{isinherited}', "");
Expect(0, 917999, '\p{^isinherited}', "");
Expect(0, 917999, '\P{isinherited}', "");
Expect(1, 917999, '\P{^isinherited}', "");
Expect(0, 918000, '\p{isinherited}', "");
Expect(1, 918000, '\p{^isinherited}', "");
Expect(1, 918000, '\P{isinherited}', "");
Expect(0, 918000, '\P{^isinherited}', "");
Expect(1, 917999, '\p{	Is_inherited}', "");
Expect(0, 917999, '\p{^	Is_inherited}', "");
Expect(0, 917999, '\P{	Is_inherited}', "");
Expect(1, 917999, '\P{^	Is_inherited}', "");
Expect(0, 918000, '\p{	Is_inherited}', "");
Expect(1, 918000, '\p{^	Is_inherited}', "");
Expect(1, 918000, '\P{	Is_inherited}', "");
Expect(0, 918000, '\P{^	Is_inherited}', "");
Error('\p{:=__Zinh}');
Error('\P{:=__Zinh}');
Expect(1, 917999, '\p{zinh}', "");
Expect(0, 917999, '\p{^zinh}', "");
Expect(0, 917999, '\P{zinh}', "");
Expect(1, 917999, '\P{^zinh}', "");
Expect(0, 918000, '\p{zinh}', "");
Expect(1, 918000, '\p{^zinh}', "");
Expect(1, 918000, '\P{zinh}', "");
Expect(0, 918000, '\P{^zinh}', "");
Expect(1, 917999, '\p{	 ZINH}', "");
Expect(0, 917999, '\p{^	 ZINH}', "");
Expect(0, 917999, '\P{	 ZINH}', "");
Expect(1, 917999, '\P{^	 ZINH}', "");
Expect(0, 918000, '\p{	 ZINH}', "");
Expect(1, 918000, '\p{^	 ZINH}', "");
Expect(1, 918000, '\P{	 ZINH}', "");
Expect(0, 918000, '\P{^	 ZINH}', "");
Error('\p{/a/_	IS_zinh}');
Error('\P{/a/_	IS_zinh}');
Expect(1, 917999, '\p{iszinh}', "");
Expect(0, 917999, '\p{^iszinh}', "");
Expect(0, 917999, '\P{iszinh}', "");
Expect(1, 917999, '\P{^iszinh}', "");
Expect(0, 918000, '\p{iszinh}', "");
Expect(1, 918000, '\p{^iszinh}', "");
Expect(1, 918000, '\P{iszinh}', "");
Expect(0, 918000, '\P{^iszinh}', "");
Expect(1, 917999, '\p{_-Is_zinh}', "");
Expect(0, 917999, '\p{^_-Is_zinh}', "");
Expect(0, 917999, '\P{_-Is_zinh}', "");
Expect(1, 917999, '\P{^_-Is_zinh}', "");
Expect(0, 918000, '\p{_-Is_zinh}', "");
Expect(1, 918000, '\p{^_-Is_zinh}', "");
Expect(1, 918000, '\P{_-Is_zinh}', "");
Expect(0, 918000, '\P{^_-Is_zinh}', "");
Error('\p{:=-_qaai}');
Error('\P{:=-_qaai}');
Expect(1, 917999, '\p{qaai}', "");
Expect(0, 917999, '\p{^qaai}', "");
Expect(0, 917999, '\P{qaai}', "");
Expect(1, 917999, '\P{^qaai}', "");
Expect(0, 918000, '\p{qaai}', "");
Expect(1, 918000, '\p{^qaai}', "");
Expect(1, 918000, '\P{qaai}', "");
Expect(0, 918000, '\P{^qaai}', "");
Expect(1, 917999, '\p{ 	Qaai}', "");
Expect(0, 917999, '\p{^ 	Qaai}', "");
Expect(0, 917999, '\P{ 	Qaai}', "");
Expect(1, 917999, '\P{^ 	Qaai}', "");
Expect(0, 918000, '\p{ 	Qaai}', "");
Expect(1, 918000, '\p{^ 	Qaai}', "");
Expect(1, 918000, '\P{ 	Qaai}', "");
Expect(0, 918000, '\P{^ 	Qaai}', "");
Error('\p{/a/_	Is_qaai}');
Error('\P{/a/_	Is_qaai}');
Expect(1, 917999, '\p{isqaai}', "");
Expect(0, 917999, '\p{^isqaai}', "");
Expect(0, 917999, '\P{isqaai}', "");
Expect(1, 917999, '\P{^isqaai}', "");
Expect(0, 918000, '\p{isqaai}', "");
Expect(1, 918000, '\p{^isqaai}', "");
Expect(1, 918000, '\P{isqaai}', "");
Expect(0, 918000, '\P{^isqaai}', "");
Expect(1, 917999, '\p{ Is_Qaai}', "");
Expect(0, 917999, '\p{^ Is_Qaai}', "");
Expect(0, 917999, '\P{ Is_Qaai}', "");
Expect(1, 917999, '\P{^ Is_Qaai}', "");
Expect(0, 918000, '\p{ Is_Qaai}', "");
Expect(1, 918000, '\p{^ Is_Qaai}', "");
Expect(1, 918000, '\P{ Is_Qaai}', "");
Expect(0, 918000, '\P{^ Is_Qaai}', "");
Error('\p{Other_Math}');
Error('\P{Other_Math}');
Error('\p{Is_Other_Math}');
Error('\P{Is_Other_Math}');
Error('\p{OMath}');
Error('\P{OMath}');
Error('\p{Is_OMath}');
Error('\P{Is_OMath}');
Error('\p{:=-_CJK_Unified_ideographs_extension_C}');
Error('\P{:=-_CJK_Unified_ideographs_extension_C}');
Expect(1, 177983, '\p{cjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\p{^cjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\P{cjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\P{^cjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\p{cjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\p{^cjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\P{cjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\P{^cjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\p{-	CJK_unified_Ideographs_EXTENSION_C}', "");
Expect(0, 177983, '\p{^-	CJK_unified_Ideographs_EXTENSION_C}', "");
Expect(0, 177983, '\P{-	CJK_unified_Ideographs_EXTENSION_C}', "");
Expect(1, 177983, '\P{^-	CJK_unified_Ideographs_EXTENSION_C}', "");
Expect(0, 177984, '\p{-	CJK_unified_Ideographs_EXTENSION_C}', "");
Expect(1, 177984, '\p{^-	CJK_unified_Ideographs_EXTENSION_C}', "");
Expect(1, 177984, '\P{-	CJK_unified_Ideographs_EXTENSION_C}', "");
Expect(0, 177984, '\P{^-	CJK_unified_Ideographs_EXTENSION_C}', "");
Error('\p{-Is_CJK_Unified_ideographs_Extension_C:=}');
Error('\P{-Is_CJK_Unified_ideographs_Extension_C:=}');
Expect(1, 177983, '\p{iscjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\p{^iscjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\P{iscjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\P{^iscjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\p{iscjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\p{^iscjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\P{iscjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\P{^iscjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\p{ is_cjk_unified_IDEOGRAPHS_Extension_C}', "");
Expect(0, 177983, '\p{^ is_cjk_unified_IDEOGRAPHS_Extension_C}', "");
Expect(0, 177983, '\P{ is_cjk_unified_IDEOGRAPHS_Extension_C}', "");
Expect(1, 177983, '\P{^ is_cjk_unified_IDEOGRAPHS_Extension_C}', "");
Expect(0, 177984, '\p{ is_cjk_unified_IDEOGRAPHS_Extension_C}', "");
Expect(1, 177984, '\p{^ is_cjk_unified_IDEOGRAPHS_Extension_C}', "");
Expect(1, 177984, '\P{ is_cjk_unified_IDEOGRAPHS_Extension_C}', "");
Expect(0, 177984, '\P{^ is_cjk_unified_IDEOGRAPHS_Extension_C}', "");
Error('\p{:=_In_cjk_UNIFIED_IDEOGRAPHS_EXTENSION_C}');
Error('\P{:=_In_cjk_UNIFIED_IDEOGRAPHS_EXTENSION_C}');
Expect(1, 177983, '\p{incjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\p{^incjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\P{incjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\P{^incjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\p{incjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\p{^incjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\P{incjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\P{^incjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\p{_In_CJK_UNIFIED_IDEOGRAPHS_Extension_C}', "");
Expect(0, 177983, '\p{^_In_CJK_UNIFIED_IDEOGRAPHS_Extension_C}', "");
Expect(0, 177983, '\P{_In_CJK_UNIFIED_IDEOGRAPHS_Extension_C}', "");
Expect(1, 177983, '\P{^_In_CJK_UNIFIED_IDEOGRAPHS_Extension_C}', "");
Expect(0, 177984, '\p{_In_CJK_UNIFIED_IDEOGRAPHS_Extension_C}', "");
Expect(1, 177984, '\p{^_In_CJK_UNIFIED_IDEOGRAPHS_Extension_C}', "");
Expect(1, 177984, '\P{_In_CJK_UNIFIED_IDEOGRAPHS_Extension_C}', "");
Expect(0, 177984, '\P{^_In_CJK_UNIFIED_IDEOGRAPHS_Extension_C}', "");
Error('\p{:=_ CJK_Ext_c}');
Error('\P{:=_ CJK_Ext_c}');
Expect(1, 177983, '\p{cjkextc}', "");
Expect(0, 177983, '\p{^cjkextc}', "");
Expect(0, 177983, '\P{cjkextc}', "");
Expect(1, 177983, '\P{^cjkextc}', "");
Expect(0, 177984, '\p{cjkextc}', "");
Expect(1, 177984, '\p{^cjkextc}', "");
Expect(1, 177984, '\P{cjkextc}', "");
Expect(0, 177984, '\P{^cjkextc}', "");
Expect(1, 177983, '\p{-	CJK_EXT_c}', "");
Expect(0, 177983, '\p{^-	CJK_EXT_c}', "");
Expect(0, 177983, '\P{-	CJK_EXT_c}', "");
Expect(1, 177983, '\P{^-	CJK_EXT_c}', "");
Expect(0, 177984, '\p{-	CJK_EXT_c}', "");
Expect(1, 177984, '\p{^-	CJK_EXT_c}', "");
Expect(1, 177984, '\P{-	CJK_EXT_c}', "");
Expect(0, 177984, '\P{^-	CJK_EXT_c}', "");
Error('\p{:=Is_cjk_ext_C}');
Error('\P{:=Is_cjk_ext_C}');
Expect(1, 177983, '\p{iscjkextc}', "");
Expect(0, 177983, '\p{^iscjkextc}', "");
Expect(0, 177983, '\P{iscjkextc}', "");
Expect(1, 177983, '\P{^iscjkextc}', "");
Expect(0, 177984, '\p{iscjkextc}', "");
Expect(1, 177984, '\p{^iscjkextc}', "");
Expect(1, 177984, '\P{iscjkextc}', "");
Expect(0, 177984, '\P{^iscjkextc}', "");
Expect(1, 177983, '\p{--Is_CJK_EXT_C}', "");
Expect(0, 177983, '\p{^--Is_CJK_EXT_C}', "");
Expect(0, 177983, '\P{--Is_CJK_EXT_C}', "");
Expect(1, 177983, '\P{^--Is_CJK_EXT_C}', "");
Expect(0, 177984, '\p{--Is_CJK_EXT_C}', "");
Expect(1, 177984, '\p{^--Is_CJK_EXT_C}', "");
Expect(1, 177984, '\P{--Is_CJK_EXT_C}', "");
Expect(0, 177984, '\P{^--Is_CJK_EXT_C}', "");
Error('\p{-:=in_cjk_EXT_c}');
Error('\P{-:=in_cjk_EXT_c}');
Expect(1, 177983, '\p{incjkextc}', "");
Expect(0, 177983, '\p{^incjkextc}', "");
Expect(0, 177983, '\P{incjkextc}', "");
Expect(1, 177983, '\P{^incjkextc}', "");
Expect(0, 177984, '\p{incjkextc}', "");
Expect(1, 177984, '\p{^incjkextc}', "");
Expect(1, 177984, '\P{incjkextc}', "");
Expect(0, 177984, '\P{^incjkextc}', "");
Expect(1, 177983, '\p{--In_CJK_ext_C}', "");
Expect(0, 177983, '\p{^--In_CJK_ext_C}', "");
Expect(0, 177983, '\P{--In_CJK_ext_C}', "");
Expect(1, 177983, '\P{^--In_CJK_ext_C}', "");
Expect(0, 177984, '\p{--In_CJK_ext_C}', "");
Expect(1, 177984, '\p{^--In_CJK_ext_C}', "");
Expect(1, 177984, '\P{--In_CJK_ext_C}', "");
Expect(0, 177984, '\P{^--In_CJK_ext_C}', "");
Error('\p{ latin_Extended_Additional:=}');
Error('\P{ latin_Extended_Additional:=}');
Expect(1, 7935, '\p{latinextendedadditional}', "");
Expect(0, 7935, '\p{^latinextendedadditional}', "");
Expect(0, 7935, '\P{latinextendedadditional}', "");
Expect(1, 7935, '\P{^latinextendedadditional}', "");
Expect(0, 7936, '\p{latinextendedadditional}', "");
Expect(1, 7936, '\p{^latinextendedadditional}', "");
Expect(1, 7936, '\P{latinextendedadditional}', "");
Expect(0, 7936, '\P{^latinextendedadditional}', "");
Expect(1, 7935, '\p{		Latin_Extended_Additional}', "");
Expect(0, 7935, '\p{^		Latin_Extended_Additional}', "");
Expect(0, 7935, '\P{		Latin_Extended_Additional}', "");
Expect(1, 7935, '\P{^		Latin_Extended_Additional}', "");
Expect(0, 7936, '\p{		Latin_Extended_Additional}', "");
Expect(1, 7936, '\p{^		Latin_Extended_Additional}', "");
Expect(1, 7936, '\P{		Latin_Extended_Additional}', "");
Expect(0, 7936, '\P{^		Latin_Extended_Additional}', "");
Error('\p{-is_latin_Extended_ADDITIONAL:=}');
Error('\P{-is_latin_Extended_ADDITIONAL:=}');
Expect(1, 7935, '\p{islatinextendedadditional}', "");
Expect(0, 7935, '\p{^islatinextendedadditional}', "");
Expect(0, 7935, '\P{islatinextendedadditional}', "");
Expect(1, 7935, '\P{^islatinextendedadditional}', "");
Expect(0, 7936, '\p{islatinextendedadditional}', "");
Expect(1, 7936, '\p{^islatinextendedadditional}', "");
Expect(1, 7936, '\P{islatinextendedadditional}', "");
Expect(0, 7936, '\P{^islatinextendedadditional}', "");
Expect(1, 7935, '\p{_Is_latin_extended_Additional}', "");
Expect(0, 7935, '\p{^_Is_latin_extended_Additional}', "");
Expect(0, 7935, '\P{_Is_latin_extended_Additional}', "");
Expect(1, 7935, '\P{^_Is_latin_extended_Additional}', "");
Expect(0, 7936, '\p{_Is_latin_extended_Additional}', "");
Expect(1, 7936, '\p{^_Is_latin_extended_Additional}', "");
Expect(1, 7936, '\P{_Is_latin_extended_Additional}', "");
Expect(0, 7936, '\P{^_Is_latin_extended_Additional}', "");
Error('\p{-/a/In_Latin_EXTENDED_Additional}');
Error('\P{-/a/In_Latin_EXTENDED_Additional}');
Expect(1, 7935, '\p{inlatinextendedadditional}', "");
Expect(0, 7935, '\p{^inlatinextendedadditional}', "");
Expect(0, 7935, '\P{inlatinextendedadditional}', "");
Expect(1, 7935, '\P{^inlatinextendedadditional}', "");
Expect(0, 7936, '\p{inlatinextendedadditional}', "");
Expect(1, 7936, '\p{^inlatinextendedadditional}', "");
Expect(1, 7936, '\P{inlatinextendedadditional}', "");
Expect(0, 7936, '\P{^inlatinextendedadditional}', "");
Expect(1, 7935, '\p{	in_Latin_extended_Additional}', "");
Expect(0, 7935, '\p{^	in_Latin_extended_Additional}', "");
Expect(0, 7935, '\P{	in_Latin_extended_Additional}', "");
Expect(1, 7935, '\P{^	in_Latin_extended_Additional}', "");
Expect(0, 7936, '\p{	in_Latin_extended_Additional}', "");
Expect(1, 7936, '\p{^	in_Latin_extended_Additional}', "");
Expect(1, 7936, '\P{	in_Latin_extended_Additional}', "");
Expect(0, 7936, '\P{^	in_Latin_extended_Additional}', "");
Error('\p{/a/		Latin_Ext_ADDITIONAL}');
Error('\P{/a/		Latin_Ext_ADDITIONAL}');
Expect(1, 7935, '\p{latinextadditional}', "");
Expect(0, 7935, '\p{^latinextadditional}', "");
Expect(0, 7935, '\P{latinextadditional}', "");
Expect(1, 7935, '\P{^latinextadditional}', "");
Expect(0, 7936, '\p{latinextadditional}', "");
Expect(1, 7936, '\p{^latinextadditional}', "");
Expect(1, 7936, '\P{latinextadditional}', "");
Expect(0, 7936, '\P{^latinextadditional}', "");
Expect(1, 7935, '\p{-Latin_Ext_additional}', "");
Expect(0, 7935, '\p{^-Latin_Ext_additional}', "");
Expect(0, 7935, '\P{-Latin_Ext_additional}', "");
Expect(1, 7935, '\P{^-Latin_Ext_additional}', "");
Expect(0, 7936, '\p{-Latin_Ext_additional}', "");
Expect(1, 7936, '\p{^-Latin_Ext_additional}', "");
Expect(1, 7936, '\P{-Latin_Ext_additional}', "");
Expect(0, 7936, '\P{^-Latin_Ext_additional}', "");
Error('\p{ /a/is_LATIN_Ext_ADDITIONAL}');
Error('\P{ /a/is_LATIN_Ext_ADDITIONAL}');
Expect(1, 7935, '\p{islatinextadditional}', "");
Expect(0, 7935, '\p{^islatinextadditional}', "");
Expect(0, 7935, '\P{islatinextadditional}', "");
Expect(1, 7935, '\P{^islatinextadditional}', "");
Expect(0, 7936, '\p{islatinextadditional}', "");
Expect(1, 7936, '\p{^islatinextadditional}', "");
Expect(1, 7936, '\P{islatinextadditional}', "");
Expect(0, 7936, '\P{^islatinextadditional}', "");
Expect(1, 7935, '\p{ IS_LATIN_Ext_Additional}', "");
Expect(0, 7935, '\p{^ IS_LATIN_Ext_Additional}', "");
Expect(0, 7935, '\P{ IS_LATIN_Ext_Additional}', "");
Expect(1, 7935, '\P{^ IS_LATIN_Ext_Additional}', "");
Expect(0, 7936, '\p{ IS_LATIN_Ext_Additional}', "");
Expect(1, 7936, '\p{^ IS_LATIN_Ext_Additional}', "");
Expect(1, 7936, '\P{ IS_LATIN_Ext_Additional}', "");
Expect(0, 7936, '\P{^ IS_LATIN_Ext_Additional}', "");
Error('\p{	_IN_Latin_EXT_ADDITIONAL/a/}');
Error('\P{	_IN_Latin_EXT_ADDITIONAL/a/}');
Expect(1, 7935, '\p{inlatinextadditional}', "");
Expect(0, 7935, '\p{^inlatinextadditional}', "");
Expect(0, 7935, '\P{inlatinextadditional}', "");
Expect(1, 7935, '\P{^inlatinextadditional}', "");
Expect(0, 7936, '\p{inlatinextadditional}', "");
Expect(1, 7936, '\p{^inlatinextadditional}', "");
Expect(1, 7936, '\P{inlatinextadditional}', "");
Expect(0, 7936, '\P{^inlatinextadditional}', "");
Expect(1, 7935, '\p{_ In_Latin_EXT_Additional}', "");
Expect(0, 7935, '\p{^_ In_Latin_EXT_Additional}', "");
Expect(0, 7935, '\P{_ In_Latin_EXT_Additional}', "");
Expect(1, 7935, '\P{^_ In_Latin_EXT_Additional}', "");
Expect(0, 7936, '\p{_ In_Latin_EXT_Additional}', "");
Expect(1, 7936, '\p{^_ In_Latin_EXT_Additional}', "");
Expect(1, 7936, '\P{_ In_Latin_EXT_Additional}', "");
Expect(0, 7936, '\P{^_ In_Latin_EXT_Additional}', "");
Error('\p{_/a/BOPOMOFO_EXTENDED}');
Error('\P{_/a/BOPOMOFO_EXTENDED}');
Expect(1, 12735, '\p{bopomofoextended}', "");
Expect(0, 12735, '\p{^bopomofoextended}', "");
Expect(0, 12735, '\P{bopomofoextended}', "");
Expect(1, 12735, '\P{^bopomofoextended}', "");
Expect(0, 12736, '\p{bopomofoextended}', "");
Expect(1, 12736, '\p{^bopomofoextended}', "");
Expect(1, 12736, '\P{bopomofoextended}', "");
Expect(0, 12736, '\P{^bopomofoextended}', "");
Expect(1, 12735, '\p{ Bopomofo_extended}', "");
Expect(0, 12735, '\p{^ Bopomofo_extended}', "");
Expect(0, 12735, '\P{ Bopomofo_extended}', "");
Expect(1, 12735, '\P{^ Bopomofo_extended}', "");
Expect(0, 12736, '\p{ Bopomofo_extended}', "");
Expect(1, 12736, '\p{^ Bopomofo_extended}', "");
Expect(1, 12736, '\P{ Bopomofo_extended}', "");
Expect(0, 12736, '\P{^ Bopomofo_extended}', "");
Error('\p{:= is_Bopomofo_extended}');
Error('\P{:= is_Bopomofo_extended}');
Expect(1, 12735, '\p{isbopomofoextended}', "");
Expect(0, 12735, '\p{^isbopomofoextended}', "");
Expect(0, 12735, '\P{isbopomofoextended}', "");
Expect(1, 12735, '\P{^isbopomofoextended}', "");
Expect(0, 12736, '\p{isbopomofoextended}', "");
Expect(1, 12736, '\p{^isbopomofoextended}', "");
Expect(1, 12736, '\P{isbopomofoextended}', "");
Expect(0, 12736, '\P{^isbopomofoextended}', "");
Expect(1, 12735, '\p{ 	Is_BOPOMOFO_Extended}', "");
Expect(0, 12735, '\p{^ 	Is_BOPOMOFO_Extended}', "");
Expect(0, 12735, '\P{ 	Is_BOPOMOFO_Extended}', "");
Expect(1, 12735, '\P{^ 	Is_BOPOMOFO_Extended}', "");
Expect(0, 12736, '\p{ 	Is_BOPOMOFO_Extended}', "");
Expect(1, 12736, '\p{^ 	Is_BOPOMOFO_Extended}', "");
Expect(1, 12736, '\P{ 	Is_BOPOMOFO_Extended}', "");
Expect(0, 12736, '\P{^ 	Is_BOPOMOFO_Extended}', "");
Error('\p{:=	_In_Bopomofo_EXTENDED}');
Error('\P{:=	_In_Bopomofo_EXTENDED}');
Expect(1, 12735, '\p{inbopomofoextended}', "");
Expect(0, 12735, '\p{^inbopomofoextended}', "");
Expect(0, 12735, '\P{inbopomofoextended}', "");
Expect(1, 12735, '\P{^inbopomofoextended}', "");
Expect(0, 12736, '\p{inbopomofoextended}', "");
Expect(1, 12736, '\p{^inbopomofoextended}', "");
Expect(1, 12736, '\P{inbopomofoextended}', "");
Expect(0, 12736, '\P{^inbopomofoextended}', "");
Expect(1, 12735, '\p{_-In_BOPOMOFO_extended}', "");
Expect(0, 12735, '\p{^_-In_BOPOMOFO_extended}', "");
Expect(0, 12735, '\P{_-In_BOPOMOFO_extended}', "");
Expect(1, 12735, '\P{^_-In_BOPOMOFO_extended}', "");
Expect(0, 12736, '\p{_-In_BOPOMOFO_extended}', "");
Expect(1, 12736, '\p{^_-In_BOPOMOFO_extended}', "");
Expect(1, 12736, '\P{_-In_BOPOMOFO_extended}', "");
Expect(0, 12736, '\P{^_-In_BOPOMOFO_extended}', "");
Error('\p{ :=Bopomofo_ext}');
Error('\P{ :=Bopomofo_ext}');
Expect(1, 12735, '\p{bopomofoext}', "");
Expect(0, 12735, '\p{^bopomofoext}', "");
Expect(0, 12735, '\P{bopomofoext}', "");
Expect(1, 12735, '\P{^bopomofoext}', "");
Expect(0, 12736, '\p{bopomofoext}', "");
Expect(1, 12736, '\p{^bopomofoext}', "");
Expect(1, 12736, '\P{bopomofoext}', "");
Expect(0, 12736, '\P{^bopomofoext}', "");
Expect(1, 12735, '\p{_-BOPOMOFO_Ext}', "");
Expect(0, 12735, '\p{^_-BOPOMOFO_Ext}', "");
Expect(0, 12735, '\P{_-BOPOMOFO_Ext}', "");
Expect(1, 12735, '\P{^_-BOPOMOFO_Ext}', "");
Expect(0, 12736, '\p{_-BOPOMOFO_Ext}', "");
Expect(1, 12736, '\p{^_-BOPOMOFO_Ext}', "");
Expect(1, 12736, '\P{_-BOPOMOFO_Ext}', "");
Expect(0, 12736, '\P{^_-BOPOMOFO_Ext}', "");
Error('\p{:=__is_Bopomofo_Ext}');
Error('\P{:=__is_Bopomofo_Ext}');
Expect(1, 12735, '\p{isbopomofoext}', "");
Expect(0, 12735, '\p{^isbopomofoext}', "");
Expect(0, 12735, '\P{isbopomofoext}', "");
Expect(1, 12735, '\P{^isbopomofoext}', "");
Expect(0, 12736, '\p{isbopomofoext}', "");
Expect(1, 12736, '\p{^isbopomofoext}', "");
Expect(1, 12736, '\P{isbopomofoext}', "");
Expect(0, 12736, '\P{^isbopomofoext}', "");
Expect(1, 12735, '\p{ is_Bopomofo_Ext}', "");
Expect(0, 12735, '\p{^ is_Bopomofo_Ext}', "");
Expect(0, 12735, '\P{ is_Bopomofo_Ext}', "");
Expect(1, 12735, '\P{^ is_Bopomofo_Ext}', "");
Expect(0, 12736, '\p{ is_Bopomofo_Ext}', "");
Expect(1, 12736, '\p{^ is_Bopomofo_Ext}', "");
Expect(1, 12736, '\P{ is_Bopomofo_Ext}', "");
Expect(0, 12736, '\P{^ is_Bopomofo_Ext}', "");
Error('\p{:=-	In_Bopomofo_EXT}');
Error('\P{:=-	In_Bopomofo_EXT}');
Expect(1, 12735, '\p{inbopomofoext}', "");
Expect(0, 12735, '\p{^inbopomofoext}', "");
Expect(0, 12735, '\P{inbopomofoext}', "");
Expect(1, 12735, '\P{^inbopomofoext}', "");
Expect(0, 12736, '\p{inbopomofoext}', "");
Expect(1, 12736, '\p{^inbopomofoext}', "");
Expect(1, 12736, '\P{inbopomofoext}', "");
Expect(0, 12736, '\P{^inbopomofoext}', "");
Expect(1, 12735, '\p{	 In_Bopomofo_Ext}', "");
Expect(0, 12735, '\p{^	 In_Bopomofo_Ext}', "");
Expect(0, 12735, '\P{	 In_Bopomofo_Ext}', "");
Expect(1, 12735, '\P{^	 In_Bopomofo_Ext}', "");
Expect(0, 12736, '\p{	 In_Bopomofo_Ext}', "");
Expect(1, 12736, '\p{^	 In_Bopomofo_Ext}', "");
Expect(1, 12736, '\P{	 In_Bopomofo_Ext}', "");
Expect(0, 12736, '\P{^	 In_Bopomofo_Ext}', "");
Error('\p{ 	Kannada/a/}');
Error('\P{ 	Kannada/a/}');
Expect(1, 3314, '\p{kannada}', "");
Expect(0, 3314, '\p{^kannada}', "");
Expect(0, 3314, '\P{kannada}', "");
Expect(1, 3314, '\P{^kannada}', "");
Expect(0, 3315, '\p{kannada}', "");
Expect(1, 3315, '\p{^kannada}', "");
Expect(1, 3315, '\P{kannada}', "");
Expect(0, 3315, '\P{^kannada}', "");
Expect(1, 3314, '\p{	_Kannada}', "");
Expect(0, 3314, '\p{^	_Kannada}', "");
Expect(0, 3314, '\P{	_Kannada}', "");
Expect(1, 3314, '\P{^	_Kannada}', "");
Expect(0, 3315, '\p{	_Kannada}', "");
Expect(1, 3315, '\p{^	_Kannada}', "");
Expect(1, 3315, '\P{	_Kannada}', "");
Expect(0, 3315, '\P{^	_Kannada}', "");
Error('\p{		IS_Kannada/a/}');
Error('\P{		IS_Kannada/a/}');
Expect(1, 3314, '\p{iskannada}', "");
Expect(0, 3314, '\p{^iskannada}', "");
Expect(0, 3314, '\P{iskannada}', "");
Expect(1, 3314, '\P{^iskannada}', "");
Expect(0, 3315, '\p{iskannada}', "");
Expect(1, 3315, '\p{^iskannada}', "");
Expect(1, 3315, '\P{iskannada}', "");
Expect(0, 3315, '\P{^iskannada}', "");
Expect(1, 3314, '\p{- Is_Kannada}', "");
Expect(0, 3314, '\p{^- Is_Kannada}', "");
Expect(0, 3314, '\P{- Is_Kannada}', "");
Expect(1, 3314, '\P{^- Is_Kannada}', "");
Expect(0, 3315, '\p{- Is_Kannada}', "");
Expect(1, 3315, '\p{^- Is_Kannada}', "");
Expect(1, 3315, '\P{- Is_Kannada}', "");
Expect(0, 3315, '\P{^- Is_Kannada}', "");
Error('\p{:=- Knda}');
Error('\P{:=- Knda}');
Expect(1, 3314, '\p{knda}', "");
Expect(0, 3314, '\p{^knda}', "");
Expect(0, 3314, '\P{knda}', "");
Expect(1, 3314, '\P{^knda}', "");
Expect(0, 3315, '\p{knda}', "");
Expect(1, 3315, '\p{^knda}', "");
Expect(1, 3315, '\P{knda}', "");
Expect(0, 3315, '\P{^knda}', "");
Expect(1, 3314, '\p{-Knda}', "");
Expect(0, 3314, '\p{^-Knda}', "");
Expect(0, 3314, '\P{-Knda}', "");
Expect(1, 3314, '\P{^-Knda}', "");
Expect(0, 3315, '\p{-Knda}', "");
Expect(1, 3315, '\p{^-Knda}', "");
Expect(1, 3315, '\P{-Knda}', "");
Expect(0, 3315, '\P{^-Knda}', "");
Error('\p{:=-is_knda}');
Error('\P{:=-is_knda}');
Expect(1, 3314, '\p{isknda}', "");
Expect(0, 3314, '\p{^isknda}', "");
Expect(0, 3314, '\P{isknda}', "");
Expect(1, 3314, '\P{^isknda}', "");
Expect(0, 3315, '\p{isknda}', "");
Expect(1, 3315, '\p{^isknda}', "");
Expect(1, 3315, '\P{isknda}', "");
Expect(0, 3315, '\P{^isknda}', "");
Expect(1, 3314, '\p{		is_knda}', "");
Expect(0, 3314, '\p{^		is_knda}', "");
Expect(0, 3314, '\P{		is_knda}', "");
Expect(1, 3314, '\P{^		is_knda}', "");
Expect(0, 3315, '\p{		is_knda}', "");
Expect(1, 3315, '\p{^		is_knda}', "");
Expect(1, 3315, '\P{		is_knda}', "");
Expect(0, 3315, '\P{^		is_knda}', "");
Error('\p{/a/Lao}');
Error('\P{/a/Lao}');
Expect(1, 3807, '\p{lao}', "");
Expect(0, 3807, '\p{^lao}', "");
Expect(0, 3807, '\P{lao}', "");
Expect(1, 3807, '\P{^lao}', "");
Expect(0, 3808, '\p{lao}', "");
Expect(1, 3808, '\p{^lao}', "");
Expect(1, 3808, '\P{lao}', "");
Expect(0, 3808, '\P{^lao}', "");
Expect(1, 3807, '\p{  LAO}', "");
Expect(0, 3807, '\p{^  LAO}', "");
Expect(0, 3807, '\P{  LAO}', "");
Expect(1, 3807, '\P{^  LAO}', "");
Expect(0, 3808, '\p{  LAO}', "");
Expect(1, 3808, '\p{^  LAO}', "");
Expect(1, 3808, '\P{  LAO}', "");
Expect(0, 3808, '\P{^  LAO}', "");
Error('\p{ /a/Is_Lao}');
Error('\P{ /a/Is_Lao}');
Expect(1, 3807, '\p{islao}', "");
Expect(0, 3807, '\p{^islao}', "");
Expect(0, 3807, '\P{islao}', "");
Expect(1, 3807, '\P{^islao}', "");
Expect(0, 3808, '\p{islao}', "");
Expect(1, 3808, '\p{^islao}', "");
Expect(1, 3808, '\P{islao}', "");
Expect(0, 3808, '\P{^islao}', "");
Expect(1, 3807, '\p{ _Is_LAO}', "");
Expect(0, 3807, '\p{^ _Is_LAO}', "");
Expect(0, 3807, '\P{ _Is_LAO}', "");
Expect(1, 3807, '\P{^ _Is_LAO}', "");
Expect(0, 3808, '\p{ _Is_LAO}', "");
Expect(1, 3808, '\p{^ _Is_LAO}', "");
Expect(1, 3808, '\P{ _Is_LAO}', "");
Expect(0, 3808, '\P{^ _Is_LAO}', "");
Error('\p{	Laoo:=}');
Error('\P{	Laoo:=}');
Expect(1, 3807, '\p{laoo}', "");
Expect(0, 3807, '\p{^laoo}', "");
Expect(0, 3807, '\P{laoo}', "");
Expect(1, 3807, '\P{^laoo}', "");
Expect(0, 3808, '\p{laoo}', "");
Expect(1, 3808, '\p{^laoo}', "");
Expect(1, 3808, '\P{laoo}', "");
Expect(0, 3808, '\P{^laoo}', "");
Expect(1, 3807, '\p{__Laoo}', "");
Expect(0, 3807, '\p{^__Laoo}', "");
Expect(0, 3807, '\P{__Laoo}', "");
Expect(1, 3807, '\P{^__Laoo}', "");
Expect(0, 3808, '\p{__Laoo}', "");
Expect(1, 3808, '\p{^__Laoo}', "");
Expect(1, 3808, '\P{__Laoo}', "");
Expect(0, 3808, '\P{^__Laoo}', "");
Error('\p{_/a/Is_laoo}');
Error('\P{_/a/Is_laoo}');
Expect(1, 3807, '\p{islaoo}', "");
Expect(0, 3807, '\p{^islaoo}', "");
Expect(0, 3807, '\P{islaoo}', "");
Expect(1, 3807, '\P{^islaoo}', "");
Expect(0, 3808, '\p{islaoo}', "");
Expect(1, 3808, '\p{^islaoo}', "");
Expect(1, 3808, '\P{islaoo}', "");
Expect(0, 3808, '\P{^islaoo}', "");
Expect(1, 3807, '\p{ Is_laoo}', "");
Expect(0, 3807, '\p{^ Is_laoo}', "");
Expect(0, 3807, '\P{ Is_laoo}', "");
Expect(1, 3807, '\P{^ Is_laoo}', "");
Expect(0, 3808, '\p{ Is_laoo}', "");
Expect(1, 3808, '\p{^ Is_laoo}', "");
Expect(1, 3808, '\P{ Is_laoo}', "");
Expect(0, 3808, '\P{^ Is_laoo}', "");
Error('\p{/a/ _PerlWord}');
Error('\P{/a/ _PerlWord}');
Expect(1, 122, '\p{perlword}', "");
Expect(0, 122, '\p{^perlword}', "");
Expect(0, 122, '\P{perlword}', "");
Expect(1, 122, '\P{^perlword}', "");
Expect(0, 123, '\p{perlword}', "");
Expect(1, 123, '\p{^perlword}', "");
Expect(1, 123, '\P{perlword}', "");
Expect(0, 123, '\P{^perlword}', "");
Expect(1, 122, '\p{__PERLWORD}', "");
Expect(0, 122, '\p{^__PERLWORD}', "");
Expect(0, 122, '\P{__PERLWORD}', "");
Expect(1, 122, '\P{^__PERLWORD}', "");
Expect(0, 123, '\p{__PERLWORD}', "");
Expect(1, 123, '\p{^__PERLWORD}', "");
Expect(1, 123, '\P{__PERLWORD}', "");
Expect(0, 123, '\P{^__PERLWORD}', "");
Error('\p{ _posixword:=}');
Error('\P{ _posixword:=}');
Expect(1, 122, '\p{posixword}', "");
Expect(0, 122, '\p{^posixword}', "");
Expect(0, 122, '\P{posixword}', "");
Expect(1, 122, '\P{^posixword}', "");
Expect(0, 123, '\p{posixword}', "");
Expect(1, 123, '\p{^posixword}', "");
Expect(1, 123, '\P{posixword}', "");
Expect(0, 123, '\P{^posixword}', "");
Expect(1, 122, '\p{		posixword}', "");
Expect(0, 122, '\p{^		posixword}', "");
Expect(0, 122, '\P{		posixword}', "");
Expect(1, 122, '\P{^		posixword}', "");
Expect(0, 123, '\p{		posixword}', "");
Expect(1, 123, '\p{^		posixword}', "");
Expect(1, 123, '\P{		posixword}', "");
Expect(0, 123, '\P{^		posixword}', "");
Error('\p{ /a/Is_PerlWord}');
Error('\P{ /a/Is_PerlWord}');
Expect(1, 122, '\p{isperlword}', "");
Expect(0, 122, '\p{^isperlword}', "");
Expect(0, 122, '\P{isperlword}', "");
Expect(1, 122, '\P{^isperlword}', "");
Expect(0, 123, '\p{isperlword}', "");
Expect(1, 123, '\p{^isperlword}', "");
Expect(1, 123, '\P{isperlword}', "");
Expect(0, 123, '\P{^isperlword}', "");
Expect(1, 122, '\p{_ Is_perlword}', "");
Expect(0, 122, '\p{^_ Is_perlword}', "");
Expect(0, 122, '\P{_ Is_perlword}', "");
Expect(1, 122, '\P{^_ Is_perlword}', "");
Expect(0, 123, '\p{_ Is_perlword}', "");
Expect(1, 123, '\p{^_ Is_perlword}', "");
Expect(1, 123, '\P{_ Is_perlword}', "");
Expect(0, 123, '\P{^_ Is_perlword}', "");
Error('\p{ Is_PosixWord/a/}');
Error('\P{ Is_PosixWord/a/}');
Expect(1, 122, '\p{isposixword}', "");
Expect(0, 122, '\p{^isposixword}', "");
Expect(0, 122, '\P{isposixword}', "");
Expect(1, 122, '\P{^isposixword}', "");
Expect(0, 123, '\p{isposixword}', "");
Expect(1, 123, '\p{^isposixword}', "");
Expect(1, 123, '\P{isposixword}', "");
Expect(0, 123, '\P{^isposixword}', "");
Expect(1, 122, '\p{-Is_PosixWord}', "");
Expect(0, 122, '\p{^-Is_PosixWord}', "");
Expect(0, 122, '\P{-Is_PosixWord}', "");
Expect(1, 122, '\P{^-Is_PosixWord}', "");
Expect(0, 123, '\p{-Is_PosixWord}', "");
Expect(1, 123, '\p{^-Is_PosixWord}', "");
Expect(1, 123, '\P{-Is_PosixWord}', "");
Expect(0, 123, '\P{^-Is_PosixWord}', "");
Error('\p{:= _Digit}');
Error('\P{:= _Digit}');
Expect(1, 120831, '\p{digit}', "");
Expect(0, 120831, '\p{^digit}', "");
Expect(0, 120831, '\P{digit}', "");
Expect(1, 120831, '\P{^digit}', "");
Expect(0, 120832, '\p{digit}', "");
Expect(1, 120832, '\p{^digit}', "");
Expect(1, 120832, '\P{digit}', "");
Expect(0, 120832, '\P{^digit}', "");
Expect(1, 120831, '\p{	Digit}', "");
Expect(0, 120831, '\p{^	Digit}', "");
Expect(0, 120831, '\P{	Digit}', "");
Expect(1, 120831, '\P{^	Digit}', "");
Expect(0, 120832, '\p{	Digit}', "");
Expect(1, 120832, '\p{^	Digit}', "");
Expect(1, 120832, '\P{	Digit}', "");
Expect(0, 120832, '\P{^	Digit}', "");
Error('\p{ :=XPosixDigit}');
Error('\P{ :=XPosixDigit}');
Expect(1, 120831, '\p{xposixdigit}', "");
Expect(0, 120831, '\p{^xposixdigit}', "");
Expect(0, 120831, '\P{xposixdigit}', "");
Expect(1, 120831, '\P{^xposixdigit}', "");
Expect(0, 120832, '\p{xposixdigit}', "");
Expect(1, 120832, '\p{^xposixdigit}', "");
Expect(1, 120832, '\P{xposixdigit}', "");
Expect(0, 120832, '\P{^xposixdigit}', "");
Expect(1, 120831, '\p{_ XPosixDigit}', "");
Expect(0, 120831, '\p{^_ XPosixDigit}', "");
Expect(0, 120831, '\P{_ XPosixDigit}', "");
Expect(1, 120831, '\P{^_ XPosixDigit}', "");
Expect(0, 120832, '\p{_ XPosixDigit}', "");
Expect(1, 120832, '\p{^_ XPosixDigit}', "");
Expect(1, 120832, '\P{_ XPosixDigit}', "");
Expect(0, 120832, '\P{^_ XPosixDigit}', "");
Error('\p{/a/is_digit}');
Error('\P{/a/is_digit}');
Expect(1, 120831, '\p{isdigit}', "");
Expect(0, 120831, '\p{^isdigit}', "");
Expect(0, 120831, '\P{isdigit}', "");
Expect(1, 120831, '\P{^isdigit}', "");
Expect(0, 120832, '\p{isdigit}', "");
Expect(1, 120832, '\p{^isdigit}', "");
Expect(1, 120832, '\P{isdigit}', "");
Expect(0, 120832, '\P{^isdigit}', "");
Expect(1, 120831, '\p{-Is_DIGIT}', "");
Expect(0, 120831, '\p{^-Is_DIGIT}', "");
Expect(0, 120831, '\P{-Is_DIGIT}', "");
Expect(1, 120831, '\P{^-Is_DIGIT}', "");
Expect(0, 120832, '\p{-Is_DIGIT}', "");
Expect(1, 120832, '\p{^-Is_DIGIT}', "");
Expect(1, 120832, '\P{-Is_DIGIT}', "");
Expect(0, 120832, '\P{^-Is_DIGIT}', "");
Error('\p{:=- is_XPOSIXDIGIT}');
Error('\P{:=- is_XPOSIXDIGIT}');
Expect(1, 120831, '\p{isxposixdigit}', "");
Expect(0, 120831, '\p{^isxposixdigit}', "");
Expect(0, 120831, '\P{isxposixdigit}', "");
Expect(1, 120831, '\P{^isxposixdigit}', "");
Expect(0, 120832, '\p{isxposixdigit}', "");
Expect(1, 120832, '\p{^isxposixdigit}', "");
Expect(1, 120832, '\P{isxposixdigit}', "");
Expect(0, 120832, '\P{^isxposixdigit}', "");
Expect(1, 120831, '\p{__IS_XPOSIXDIGIT}', "");
Expect(0, 120831, '\p{^__IS_XPOSIXDIGIT}', "");
Expect(0, 120831, '\P{__IS_XPOSIXDIGIT}', "");
Expect(1, 120831, '\P{^__IS_XPOSIXDIGIT}', "");
Expect(0, 120832, '\p{__IS_XPOSIXDIGIT}', "");
Expect(1, 120832, '\p{^__IS_XPOSIXDIGIT}', "");
Expect(1, 120832, '\P{__IS_XPOSIXDIGIT}', "");
Expect(0, 120832, '\P{^__IS_XPOSIXDIGIT}', "");
Error('\p{_DECIMAL_NUMBER:=}');
Error('\P{_DECIMAL_NUMBER:=}');
Expect(1, 120831, '\p{decimalnumber}', "");
Expect(0, 120831, '\p{^decimalnumber}', "");
Expect(0, 120831, '\P{decimalnumber}', "");
Expect(1, 120831, '\P{^decimalnumber}', "");
Expect(0, 120832, '\p{decimalnumber}', "");
Expect(1, 120832, '\p{^decimalnumber}', "");
Expect(1, 120832, '\P{decimalnumber}', "");
Expect(0, 120832, '\P{^decimalnumber}', "");
Expect(1, 120831, '\p{_	Decimal_Number}', "");
Expect(0, 120831, '\p{^_	Decimal_Number}', "");
Expect(0, 120831, '\P{_	Decimal_Number}', "");
Expect(1, 120831, '\P{^_	Decimal_Number}', "");
Expect(0, 120832, '\p{_	Decimal_Number}', "");
Expect(1, 120832, '\p{^_	Decimal_Number}', "");
Expect(1, 120832, '\P{_	Decimal_Number}', "");
Expect(0, 120832, '\P{^_	Decimal_Number}', "");
Error('\p{__IS_decimal_NUMBER/a/}');
Error('\P{__IS_decimal_NUMBER/a/}');
Expect(1, 120831, '\p{isdecimalnumber}', "");
Expect(0, 120831, '\p{^isdecimalnumber}', "");
Expect(0, 120831, '\P{isdecimalnumber}', "");
Expect(1, 120831, '\P{^isdecimalnumber}', "");
Expect(0, 120832, '\p{isdecimalnumber}', "");
Expect(1, 120832, '\p{^isdecimalnumber}', "");
Expect(1, 120832, '\P{isdecimalnumber}', "");
Expect(0, 120832, '\P{^isdecimalnumber}', "");
Expect(1, 120831, '\p{ 	Is_Decimal_Number}', "");
Expect(0, 120831, '\p{^ 	Is_Decimal_Number}', "");
Expect(0, 120831, '\P{ 	Is_Decimal_Number}', "");
Expect(1, 120831, '\P{^ 	Is_Decimal_Number}', "");
Expect(0, 120832, '\p{ 	Is_Decimal_Number}', "");
Expect(1, 120832, '\p{^ 	Is_Decimal_Number}', "");
Expect(1, 120832, '\P{ 	Is_Decimal_Number}', "");
Expect(0, 120832, '\P{^ 	Is_Decimal_Number}', "");
Error('\p{/a/Nd}');
Error('\P{/a/Nd}');
Expect(1, 120831, '\p{nd}', "");
Expect(0, 120831, '\p{^nd}', "");
Expect(0, 120831, '\P{nd}', "");
Expect(1, 120831, '\P{^nd}', "");
Expect(0, 120832, '\p{nd}', "");
Expect(1, 120832, '\p{^nd}', "");
Expect(1, 120832, '\P{nd}', "");
Expect(0, 120832, '\P{^nd}', "");
Expect(1, 120831, '\p{		nd}', "");
Expect(0, 120831, '\p{^		nd}', "");
Expect(0, 120831, '\P{		nd}', "");
Expect(1, 120831, '\P{^		nd}', "");
Expect(0, 120832, '\p{		nd}', "");
Expect(1, 120832, '\p{^		nd}', "");
Expect(1, 120832, '\P{		nd}', "");
Expect(0, 120832, '\P{^		nd}', "");
Error('\p{:=is_Nd}');
Error('\P{:=is_Nd}');
Expect(1, 120831, '\p{isnd}', "");
Expect(0, 120831, '\p{^isnd}', "");
Expect(0, 120831, '\P{isnd}', "");
Expect(1, 120831, '\P{^isnd}', "");
Expect(0, 120832, '\p{isnd}', "");
Expect(1, 120832, '\p{^isnd}', "");
Expect(1, 120832, '\P{isnd}', "");
Expect(0, 120832, '\P{^isnd}', "");
Expect(1, 120831, '\p{ Is_Nd}', "");
Expect(0, 120831, '\p{^ Is_Nd}', "");
Expect(0, 120831, '\P{ Is_Nd}', "");
Expect(1, 120831, '\P{^ Is_Nd}', "");
Expect(0, 120832, '\p{ Is_Nd}', "");
Expect(1, 120832, '\p{^ Is_Nd}', "");
Expect(1, 120832, '\P{ Is_Nd}', "");
Expect(0, 120832, '\P{^ Is_Nd}', "");
Error('\p{ :=cyrillic_EXTENDED_a}');
Error('\P{ :=cyrillic_EXTENDED_a}');
Expect(1, 11775, '\p{cyrillicextendeda}', "");
Expect(0, 11775, '\p{^cyrillicextendeda}', "");
Expect(0, 11775, '\P{cyrillicextendeda}', "");
Expect(1, 11775, '\P{^cyrillicextendeda}', "");
Expect(0, 11776, '\p{cyrillicextendeda}', "");
Expect(1, 11776, '\p{^cyrillicextendeda}', "");
Expect(1, 11776, '\P{cyrillicextendeda}', "");
Expect(0, 11776, '\P{^cyrillicextendeda}', "");
Expect(1, 11775, '\p{_Cyrillic_EXTENDED_A}', "");
Expect(0, 11775, '\p{^_Cyrillic_EXTENDED_A}', "");
Expect(0, 11775, '\P{_Cyrillic_EXTENDED_A}', "");
Expect(1, 11775, '\P{^_Cyrillic_EXTENDED_A}', "");
Expect(0, 11776, '\p{_Cyrillic_EXTENDED_A}', "");
Expect(1, 11776, '\p{^_Cyrillic_EXTENDED_A}', "");
Expect(1, 11776, '\P{_Cyrillic_EXTENDED_A}', "");
Expect(0, 11776, '\P{^_Cyrillic_EXTENDED_A}', "");
Error('\p{	:=Is_cyrillic_Extended_a}');
Error('\P{	:=Is_cyrillic_Extended_a}');
Expect(1, 11775, '\p{iscyrillicextendeda}', "");
Expect(0, 11775, '\p{^iscyrillicextendeda}', "");
Expect(0, 11775, '\P{iscyrillicextendeda}', "");
Expect(1, 11775, '\P{^iscyrillicextendeda}', "");
Expect(0, 11776, '\p{iscyrillicextendeda}', "");
Expect(1, 11776, '\p{^iscyrillicextendeda}', "");
Expect(1, 11776, '\P{iscyrillicextendeda}', "");
Expect(0, 11776, '\P{^iscyrillicextendeda}', "");
Expect(1, 11775, '\p{		IS_Cyrillic_Extended_A}', "");
Expect(0, 11775, '\p{^		IS_Cyrillic_Extended_A}', "");
Expect(0, 11775, '\P{		IS_Cyrillic_Extended_A}', "");
Expect(1, 11775, '\P{^		IS_Cyrillic_Extended_A}', "");
Expect(0, 11776, '\p{		IS_Cyrillic_Extended_A}', "");
Expect(1, 11776, '\p{^		IS_Cyrillic_Extended_A}', "");
Expect(1, 11776, '\P{		IS_Cyrillic_Extended_A}', "");
Expect(0, 11776, '\P{^		IS_Cyrillic_Extended_A}', "");
Error('\p{_/a/In_CYRILLIC_EXTENDED_A}');
Error('\P{_/a/In_CYRILLIC_EXTENDED_A}');
Expect(1, 11775, '\p{incyrillicextendeda}', "");
Expect(0, 11775, '\p{^incyrillicextendeda}', "");
Expect(0, 11775, '\P{incyrillicextendeda}', "");
Expect(1, 11775, '\P{^incyrillicextendeda}', "");
Expect(0, 11776, '\p{incyrillicextendeda}', "");
Expect(1, 11776, '\p{^incyrillicextendeda}', "");
Expect(1, 11776, '\P{incyrillicextendeda}', "");
Expect(0, 11776, '\P{^incyrillicextendeda}', "");
Expect(1, 11775, '\p{ -IN_cyrillic_extended_a}', "");
Expect(0, 11775, '\p{^ -IN_cyrillic_extended_a}', "");
Expect(0, 11775, '\P{ -IN_cyrillic_extended_a}', "");
Expect(1, 11775, '\P{^ -IN_cyrillic_extended_a}', "");
Expect(0, 11776, '\p{ -IN_cyrillic_extended_a}', "");
Expect(1, 11776, '\p{^ -IN_cyrillic_extended_a}', "");
Expect(1, 11776, '\P{ -IN_cyrillic_extended_a}', "");
Expect(0, 11776, '\P{^ -IN_cyrillic_extended_a}', "");
Error('\p{:=_ Cyrillic_ext_A}');
Error('\P{:=_ Cyrillic_ext_A}');
Expect(1, 11775, '\p{cyrillicexta}', "");
Expect(0, 11775, '\p{^cyrillicexta}', "");
Expect(0, 11775, '\P{cyrillicexta}', "");
Expect(1, 11775, '\P{^cyrillicexta}', "");
Expect(0, 11776, '\p{cyrillicexta}', "");
Expect(1, 11776, '\p{^cyrillicexta}', "");
Expect(1, 11776, '\P{cyrillicexta}', "");
Expect(0, 11776, '\P{^cyrillicexta}', "");
Expect(1, 11775, '\p{_	CYRILLIC_ext_A}', "");
Expect(0, 11775, '\p{^_	CYRILLIC_ext_A}', "");
Expect(0, 11775, '\P{_	CYRILLIC_ext_A}', "");
Expect(1, 11775, '\P{^_	CYRILLIC_ext_A}', "");
Expect(0, 11776, '\p{_	CYRILLIC_ext_A}', "");
Expect(1, 11776, '\p{^_	CYRILLIC_ext_A}', "");
Expect(1, 11776, '\P{_	CYRILLIC_ext_A}', "");
Expect(0, 11776, '\P{^_	CYRILLIC_ext_A}', "");
Error('\p{/a/Is_CYRILLIC_ext_A}');
Error('\P{/a/Is_CYRILLIC_ext_A}');
Expect(1, 11775, '\p{iscyrillicexta}', "");
Expect(0, 11775, '\p{^iscyrillicexta}', "");
Expect(0, 11775, '\P{iscyrillicexta}', "");
Expect(1, 11775, '\P{^iscyrillicexta}', "");
Expect(0, 11776, '\p{iscyrillicexta}', "");
Expect(1, 11776, '\p{^iscyrillicexta}', "");
Expect(1, 11776, '\P{iscyrillicexta}', "");
Expect(0, 11776, '\P{^iscyrillicexta}', "");
Expect(1, 11775, '\p{ Is_Cyrillic_Ext_A}', "");
Expect(0, 11775, '\p{^ Is_Cyrillic_Ext_A}', "");
Expect(0, 11775, '\P{ Is_Cyrillic_Ext_A}', "");
Expect(1, 11775, '\P{^ Is_Cyrillic_Ext_A}', "");
Expect(0, 11776, '\p{ Is_Cyrillic_Ext_A}', "");
Expect(1, 11776, '\p{^ Is_Cyrillic_Ext_A}', "");
Expect(1, 11776, '\P{ Is_Cyrillic_Ext_A}', "");
Expect(0, 11776, '\P{^ Is_Cyrillic_Ext_A}', "");
Error('\p{-_in_Cyrillic_ext_A/a/}');
Error('\P{-_in_Cyrillic_ext_A/a/}');
Expect(1, 11775, '\p{incyrillicexta}', "");
Expect(0, 11775, '\p{^incyrillicexta}', "");
Expect(0, 11775, '\P{incyrillicexta}', "");
Expect(1, 11775, '\P{^incyrillicexta}', "");
Expect(0, 11776, '\p{incyrillicexta}', "");
Expect(1, 11776, '\p{^incyrillicexta}', "");
Expect(1, 11776, '\P{incyrillicexta}', "");
Expect(0, 11776, '\P{^incyrillicexta}', "");
Expect(1, 11775, '\p{	in_CYRILLIC_EXT_A}', "");
Expect(0, 11775, '\p{^	in_CYRILLIC_EXT_A}', "");
Expect(0, 11775, '\P{	in_CYRILLIC_EXT_A}', "");
Expect(1, 11775, '\P{^	in_CYRILLIC_EXT_A}', "");
Expect(0, 11776, '\p{	in_CYRILLIC_EXT_A}', "");
Expect(1, 11776, '\p{^	in_CYRILLIC_EXT_A}', "");
Expect(1, 11776, '\P{	in_CYRILLIC_EXT_A}', "");
Expect(0, 11776, '\P{^	in_CYRILLIC_EXT_A}', "");
Error('\p{:=_MISCELLANEOUS_SYMBOLS_And_PICTOGRAPHS}');
Error('\P{:=_MISCELLANEOUS_SYMBOLS_And_PICTOGRAPHS}');
Expect(1, 128511, '\p{miscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\p{^miscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\P{miscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\P{^miscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\p{miscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\p{^miscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\P{miscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\P{^miscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\p{	-MISCELLANEOUS_SYMBOLS_And_Pictographs}', "");
Expect(0, 128511, '\p{^	-MISCELLANEOUS_SYMBOLS_And_Pictographs}', "");
Expect(0, 128511, '\P{	-MISCELLANEOUS_SYMBOLS_And_Pictographs}', "");
Expect(1, 128511, '\P{^	-MISCELLANEOUS_SYMBOLS_And_Pictographs}', "");
Expect(0, 128512, '\p{	-MISCELLANEOUS_SYMBOLS_And_Pictographs}', "");
Expect(1, 128512, '\p{^	-MISCELLANEOUS_SYMBOLS_And_Pictographs}', "");
Expect(1, 128512, '\P{	-MISCELLANEOUS_SYMBOLS_And_Pictographs}', "");
Expect(0, 128512, '\P{^	-MISCELLANEOUS_SYMBOLS_And_Pictographs}', "");
Error('\p{ -Is_miscellaneous_Symbols_AND_Pictographs/a/}');
Error('\P{ -Is_miscellaneous_Symbols_AND_Pictographs/a/}');
Expect(1, 128511, '\p{ismiscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\p{^ismiscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\P{ismiscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\P{^ismiscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\p{ismiscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\p{^ismiscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\P{ismiscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\P{^ismiscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\p{_is_Miscellaneous_SYMBOLS_and_Pictographs}', "");
Expect(0, 128511, '\p{^_is_Miscellaneous_SYMBOLS_and_Pictographs}', "");
Expect(0, 128511, '\P{_is_Miscellaneous_SYMBOLS_and_Pictographs}', "");
Expect(1, 128511, '\P{^_is_Miscellaneous_SYMBOLS_and_Pictographs}', "");
Expect(0, 128512, '\p{_is_Miscellaneous_SYMBOLS_and_Pictographs}', "");
Expect(1, 128512, '\p{^_is_Miscellaneous_SYMBOLS_and_Pictographs}', "");
Expect(1, 128512, '\P{_is_Miscellaneous_SYMBOLS_and_Pictographs}', "");
Expect(0, 128512, '\P{^_is_Miscellaneous_SYMBOLS_and_Pictographs}', "");
Error('\p{:=	In_MISCELLANEOUS_Symbols_and_Pictographs}');
Error('\P{:=	In_MISCELLANEOUS_Symbols_and_Pictographs}');
Expect(1, 128511, '\p{inmiscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\p{^inmiscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\P{inmiscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\P{^inmiscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\p{inmiscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\p{^inmiscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\P{inmiscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\P{^inmiscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\p{	In_MISCELLANEOUS_Symbols_And_Pictographs}', "");
Expect(0, 128511, '\p{^	In_MISCELLANEOUS_Symbols_And_Pictographs}', "");
Expect(0, 128511, '\P{	In_MISCELLANEOUS_Symbols_And_Pictographs}', "");
Expect(1, 128511, '\P{^	In_MISCELLANEOUS_Symbols_And_Pictographs}', "");
Expect(0, 128512, '\p{	In_MISCELLANEOUS_Symbols_And_Pictographs}', "");
Expect(1, 128512, '\p{^	In_MISCELLANEOUS_Symbols_And_Pictographs}', "");
Expect(1, 128512, '\P{	In_MISCELLANEOUS_Symbols_And_Pictographs}', "");
Expect(0, 128512, '\P{^	In_MISCELLANEOUS_Symbols_And_Pictographs}', "");
Error('\p{/a/__MISC_Pictographs}');
Error('\P{/a/__MISC_Pictographs}');
Expect(1, 128511, '\p{miscpictographs}', "");
Expect(0, 128511, '\p{^miscpictographs}', "");
Expect(0, 128511, '\P{miscpictographs}', "");
Expect(1, 128511, '\P{^miscpictographs}', "");
Expect(0, 128512, '\p{miscpictographs}', "");
Expect(1, 128512, '\p{^miscpictographs}', "");
Expect(1, 128512, '\P{miscpictographs}', "");
Expect(0, 128512, '\P{^miscpictographs}', "");
Expect(1, 128511, '\p{	misc_Pictographs}', "");
Expect(0, 128511, '\p{^	misc_Pictographs}', "");
Expect(0, 128511, '\P{	misc_Pictographs}', "");
Expect(1, 128511, '\P{^	misc_Pictographs}', "");
Expect(0, 128512, '\p{	misc_Pictographs}', "");
Expect(1, 128512, '\p{^	misc_Pictographs}', "");
Expect(1, 128512, '\P{	misc_Pictographs}', "");
Expect(0, 128512, '\P{^	misc_Pictographs}', "");
Error('\p{	/a/is_Misc_Pictographs}');
Error('\P{	/a/is_Misc_Pictographs}');
Expect(1, 128511, '\p{ismiscpictographs}', "");
Expect(0, 128511, '\p{^ismiscpictographs}', "");
Expect(0, 128511, '\P{ismiscpictographs}', "");
Expect(1, 128511, '\P{^ismiscpictographs}', "");
Expect(0, 128512, '\p{ismiscpictographs}', "");
Expect(1, 128512, '\p{^ismiscpictographs}', "");
Expect(1, 128512, '\P{ismiscpictographs}', "");
Expect(0, 128512, '\P{^ismiscpictographs}', "");
Expect(1, 128511, '\p{_	is_MISC_Pictographs}', "");
Expect(0, 128511, '\p{^_	is_MISC_Pictographs}', "");
Expect(0, 128511, '\P{_	is_MISC_Pictographs}', "");
Expect(1, 128511, '\P{^_	is_MISC_Pictographs}', "");
Expect(0, 128512, '\p{_	is_MISC_Pictographs}', "");
Expect(1, 128512, '\p{^_	is_MISC_Pictographs}', "");
Expect(1, 128512, '\P{_	is_MISC_Pictographs}', "");
Expect(0, 128512, '\P{^_	is_MISC_Pictographs}', "");
Error('\p{	 IN_Misc_PICTOGRAPHS/a/}');
Error('\P{	 IN_Misc_PICTOGRAPHS/a/}');
Expect(1, 128511, '\p{inmiscpictographs}', "");
Expect(0, 128511, '\p{^inmiscpictographs}', "");
Expect(0, 128511, '\P{inmiscpictographs}', "");
Expect(1, 128511, '\P{^inmiscpictographs}', "");
Expect(0, 128512, '\p{inmiscpictographs}', "");
Expect(1, 128512, '\p{^inmiscpictographs}', "");
Expect(1, 128512, '\P{inmiscpictographs}', "");
Expect(0, 128512, '\P{^inmiscpictographs}', "");
Expect(1, 128511, '\p{-In_MISC_PICTOGRAPHS}', "");
Expect(0, 128511, '\p{^-In_MISC_PICTOGRAPHS}', "");
Expect(0, 128511, '\P{-In_MISC_PICTOGRAPHS}', "");
Expect(1, 128511, '\P{^-In_MISC_PICTOGRAPHS}', "");
Expect(0, 128512, '\p{-In_MISC_PICTOGRAPHS}', "");
Expect(1, 128512, '\p{^-In_MISC_PICTOGRAPHS}', "");
Expect(1, 128512, '\P{-In_MISC_PICTOGRAPHS}', "");
Expect(0, 128512, '\P{^-In_MISC_PICTOGRAPHS}', "");
Error('\p{:=-	osmanya}');
Error('\P{:=-	osmanya}');
Expect(1, 66729, '\p{osmanya}', "");
Expect(0, 66729, '\p{^osmanya}', "");
Expect(0, 66729, '\P{osmanya}', "");
Expect(1, 66729, '\P{^osmanya}', "");
Expect(0, 66730, '\p{osmanya}', "");
Expect(1, 66730, '\p{^osmanya}', "");
Expect(1, 66730, '\P{osmanya}', "");
Expect(0, 66730, '\P{^osmanya}', "");
Expect(1, 66729, '\p{ Osmanya}', "");
Expect(0, 66729, '\p{^ Osmanya}', "");
Expect(0, 66729, '\P{ Osmanya}', "");
Expect(1, 66729, '\P{^ Osmanya}', "");
Expect(0, 66730, '\p{ Osmanya}', "");
Expect(1, 66730, '\p{^ Osmanya}', "");
Expect(1, 66730, '\P{ Osmanya}', "");
Expect(0, 66730, '\P{^ Osmanya}', "");
Error('\p{/a/	-IS_osmanya}');
Error('\P{/a/	-IS_osmanya}');
Expect(1, 66729, '\p{isosmanya}', "");
Expect(0, 66729, '\p{^isosmanya}', "");
Expect(0, 66729, '\P{isosmanya}', "");
Expect(1, 66729, '\P{^isosmanya}', "");
Expect(0, 66730, '\p{isosmanya}', "");
Expect(1, 66730, '\p{^isosmanya}', "");
Expect(1, 66730, '\P{isosmanya}', "");
Expect(0, 66730, '\P{^isosmanya}', "");
Expect(1, 66729, '\p{--Is_Osmanya}', "");
Expect(0, 66729, '\p{^--Is_Osmanya}', "");
Expect(0, 66729, '\P{--Is_Osmanya}', "");
Expect(1, 66729, '\P{^--Is_Osmanya}', "");
Expect(0, 66730, '\p{--Is_Osmanya}', "");
Expect(1, 66730, '\p{^--Is_Osmanya}', "");
Expect(1, 66730, '\P{--Is_Osmanya}', "");
Expect(0, 66730, '\P{^--Is_Osmanya}', "");
Error('\p{:=osma}');
Error('\P{:=osma}');
Expect(1, 66729, '\p{osma}', "");
Expect(0, 66729, '\p{^osma}', "");
Expect(0, 66729, '\P{osma}', "");
Expect(1, 66729, '\P{^osma}', "");
Expect(0, 66730, '\p{osma}', "");
Expect(1, 66730, '\p{^osma}', "");
Expect(1, 66730, '\P{osma}', "");
Expect(0, 66730, '\P{^osma}', "");
Expect(1, 66729, '\p{-Osma}', "");
Expect(0, 66729, '\p{^-Osma}', "");
Expect(0, 66729, '\P{-Osma}', "");
Expect(1, 66729, '\P{^-Osma}', "");
Expect(0, 66730, '\p{-Osma}', "");
Expect(1, 66730, '\p{^-Osma}', "");
Expect(1, 66730, '\P{-Osma}', "");
Expect(0, 66730, '\P{^-Osma}', "");
Error('\p{/a/ is_OSMA}');
Error('\P{/a/ is_OSMA}');
Expect(1, 66729, '\p{isosma}', "");
Expect(0, 66729, '\p{^isosma}', "");
Expect(0, 66729, '\P{isosma}', "");
Expect(1, 66729, '\P{^isosma}', "");
Expect(0, 66730, '\p{isosma}', "");
Expect(1, 66730, '\p{^isosma}', "");
Expect(1, 66730, '\P{isosma}', "");
Expect(0, 66730, '\P{^isosma}', "");
Expect(1, 66729, '\p{ -is_OSMA}', "");
Expect(0, 66729, '\p{^ -is_OSMA}', "");
Expect(0, 66729, '\P{ -is_OSMA}', "");
Expect(1, 66729, '\P{^ -is_OSMA}', "");
Expect(0, 66730, '\p{ -is_OSMA}', "");
Expect(1, 66730, '\p{^ -is_OSMA}', "");
Expect(1, 66730, '\P{ -is_OSMA}', "");
Expect(0, 66730, '\P{^ -is_OSMA}', "");
Error('\p{:=	_Unified_ideograph}');
Error('\P{:=	_Unified_ideograph}');
Expect(1, 178205, '\p{unifiedideograph}', "");
Expect(0, 178205, '\p{^unifiedideograph}', "");
Expect(0, 178205, '\P{unifiedideograph}', "");
Expect(1, 178205, '\P{^unifiedideograph}', "");
Expect(0, 178206, '\p{unifiedideograph}', "");
Expect(1, 178206, '\p{^unifiedideograph}', "");
Expect(1, 178206, '\P{unifiedideograph}', "");
Expect(0, 178206, '\P{^unifiedideograph}', "");
Expect(1, 178205, '\p{	_Unified_Ideograph}', "");
Expect(0, 178205, '\p{^	_Unified_Ideograph}', "");
Expect(0, 178205, '\P{	_Unified_Ideograph}', "");
Expect(1, 178205, '\P{^	_Unified_Ideograph}', "");
Expect(0, 178206, '\p{	_Unified_Ideograph}', "");
Expect(1, 178206, '\p{^	_Unified_Ideograph}', "");
Expect(1, 178206, '\P{	_Unified_Ideograph}', "");
Expect(0, 178206, '\P{^	_Unified_Ideograph}', "");
Error('\p{ is_UNIFIED_ideograph:=}');
Error('\P{ is_UNIFIED_ideograph:=}');
Expect(1, 178205, '\p{isunifiedideograph}', "");
Expect(0, 178205, '\p{^isunifiedideograph}', "");
Expect(0, 178205, '\P{isunifiedideograph}', "");
Expect(1, 178205, '\P{^isunifiedideograph}', "");
Expect(0, 178206, '\p{isunifiedideograph}', "");
Expect(1, 178206, '\p{^isunifiedideograph}', "");
Expect(1, 178206, '\P{isunifiedideograph}', "");
Expect(0, 178206, '\P{^isunifiedideograph}', "");
Expect(1, 178205, '\p{	Is_unified_ideograph}', "");
Expect(0, 178205, '\p{^	Is_unified_ideograph}', "");
Expect(0, 178205, '\P{	Is_unified_ideograph}', "");
Expect(1, 178205, '\P{^	Is_unified_ideograph}', "");
Expect(0, 178206, '\p{	Is_unified_ideograph}', "");
Expect(1, 178206, '\p{^	Is_unified_ideograph}', "");
Expect(1, 178206, '\P{	Is_unified_ideograph}', "");
Expect(0, 178206, '\P{^	Is_unified_ideograph}', "");
Error('\p{/a/-	uideo}');
Error('\P{/a/-	uideo}');
Expect(1, 178205, '\p{uideo}', "");
Expect(0, 178205, '\p{^uideo}', "");
Expect(0, 178205, '\P{uideo}', "");
Expect(1, 178205, '\P{^uideo}', "");
Expect(0, 178206, '\p{uideo}', "");
Expect(1, 178206, '\p{^uideo}', "");
Expect(1, 178206, '\P{uideo}', "");
Expect(0, 178206, '\P{^uideo}', "");
Expect(1, 178205, '\p{	UIDEO}', "");
Expect(0, 178205, '\p{^	UIDEO}', "");
Expect(0, 178205, '\P{	UIDEO}', "");
Expect(1, 178205, '\P{^	UIDEO}', "");
Expect(0, 178206, '\p{	UIDEO}', "");
Expect(1, 178206, '\p{^	UIDEO}', "");
Expect(1, 178206, '\P{	UIDEO}', "");
Expect(0, 178206, '\P{^	UIDEO}', "");
Error('\p{:= 	Is_UIDEO}');
Error('\P{:= 	Is_UIDEO}');
Expect(1, 178205, '\p{isuideo}', "");
Expect(0, 178205, '\p{^isuideo}', "");
Expect(0, 178205, '\P{isuideo}', "");
Expect(1, 178205, '\P{^isuideo}', "");
Expect(0, 178206, '\p{isuideo}', "");
Expect(1, 178206, '\p{^isuideo}', "");
Expect(1, 178206, '\P{isuideo}', "");
Expect(0, 178206, '\P{^isuideo}', "");
Expect(1, 178205, '\p{-Is_UIDEO}', "");
Expect(0, 178205, '\p{^-Is_UIDEO}', "");
Expect(0, 178205, '\P{-Is_UIDEO}', "");
Expect(1, 178205, '\P{^-Is_UIDEO}', "");
Expect(0, 178206, '\p{-Is_UIDEO}', "");
Expect(1, 178206, '\p{^-Is_UIDEO}', "");
Expect(1, 178206, '\P{-Is_UIDEO}', "");
Expect(0, 178206, '\P{^-Is_UIDEO}', "");
Error('\p{:=-arabic_mathematical_ALPHABETIC_Symbols}');
Error('\P{:=-arabic_mathematical_ALPHABETIC_Symbols}');
Expect(1, 126719, '\p{arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\p{^arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\P{arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\P{^arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\p{arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\p{^arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\P{arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\P{^arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\p{ arabic_MATHEMATICAL_ALPHABETIC_symbols}', "");
Expect(0, 126719, '\p{^ arabic_MATHEMATICAL_ALPHABETIC_symbols}', "");
Expect(0, 126719, '\P{ arabic_MATHEMATICAL_ALPHABETIC_symbols}', "");
Expect(1, 126719, '\P{^ arabic_MATHEMATICAL_ALPHABETIC_symbols}', "");
Expect(0, 126720, '\p{ arabic_MATHEMATICAL_ALPHABETIC_symbols}', "");
Expect(1, 126720, '\p{^ arabic_MATHEMATICAL_ALPHABETIC_symbols}', "");
Expect(1, 126720, '\P{ arabic_MATHEMATICAL_ALPHABETIC_symbols}', "");
Expect(0, 126720, '\P{^ arabic_MATHEMATICAL_ALPHABETIC_symbols}', "");
Error('\p{:= IS_arabic_MATHEMATICAL_Alphabetic_symbols}');
Error('\P{:= IS_arabic_MATHEMATICAL_Alphabetic_symbols}');
Expect(1, 126719, '\p{isarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\p{^isarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\P{isarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\P{^isarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\p{isarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\p{^isarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\P{isarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\P{^isarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\p{  IS_Arabic_Mathematical_Alphabetic_Symbols}', "");
Expect(0, 126719, '\p{^  IS_Arabic_Mathematical_Alphabetic_Symbols}', "");
Expect(0, 126719, '\P{  IS_Arabic_Mathematical_Alphabetic_Symbols}', "");
Expect(1, 126719, '\P{^  IS_Arabic_Mathematical_Alphabetic_Symbols}', "");
Expect(0, 126720, '\p{  IS_Arabic_Mathematical_Alphabetic_Symbols}', "");
Expect(1, 126720, '\p{^  IS_Arabic_Mathematical_Alphabetic_Symbols}', "");
Expect(1, 126720, '\P{  IS_Arabic_Mathematical_Alphabetic_Symbols}', "");
Expect(0, 126720, '\P{^  IS_Arabic_Mathematical_Alphabetic_Symbols}', "");
Error('\p{	/a/IN_ARABIC_Mathematical_Alphabetic_Symbols}');
Error('\P{	/a/IN_ARABIC_Mathematical_Alphabetic_Symbols}');
Expect(1, 126719, '\p{inarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\p{^inarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\P{inarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\P{^inarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\p{inarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\p{^inarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\P{inarabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\P{^inarabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\p{ in_ARABIC_Mathematical_ALPHABETIC_Symbols}', "");
Expect(0, 126719, '\p{^ in_ARABIC_Mathematical_ALPHABETIC_Symbols}', "");
Expect(0, 126719, '\P{ in_ARABIC_Mathematical_ALPHABETIC_Symbols}', "");
Expect(1, 126719, '\P{^ in_ARABIC_Mathematical_ALPHABETIC_Symbols}', "");
Expect(0, 126720, '\p{ in_ARABIC_Mathematical_ALPHABETIC_Symbols}', "");
Expect(1, 126720, '\p{^ in_ARABIC_Mathematical_ALPHABETIC_Symbols}', "");
Expect(1, 126720, '\P{ in_ARABIC_Mathematical_ALPHABETIC_Symbols}', "");
Expect(0, 126720, '\P{^ in_ARABIC_Mathematical_ALPHABETIC_Symbols}', "");
Error('\p{		ARABIC_MATH/a/}');
Error('\P{		ARABIC_MATH/a/}');
Expect(1, 126719, '\p{arabicmath}', "");
Expect(0, 126719, '\p{^arabicmath}', "");
Expect(0, 126719, '\P{arabicmath}', "");
Expect(1, 126719, '\P{^arabicmath}', "");
Expect(0, 126720, '\p{arabicmath}', "");
Expect(1, 126720, '\p{^arabicmath}', "");
Expect(1, 126720, '\P{arabicmath}', "");
Expect(0, 126720, '\P{^arabicmath}', "");
Expect(1, 126719, '\p{- Arabic_MATH}', "");
Expect(0, 126719, '\p{^- Arabic_MATH}', "");
Expect(0, 126719, '\P{- Arabic_MATH}', "");
Expect(1, 126719, '\P{^- Arabic_MATH}', "");
Expect(0, 126720, '\p{- Arabic_MATH}', "");
Expect(1, 126720, '\p{^- Arabic_MATH}', "");
Expect(1, 126720, '\P{- Arabic_MATH}', "");
Expect(0, 126720, '\P{^- Arabic_MATH}', "");
Error('\p{-:=IS_arabic_Math}');
Error('\P{-:=IS_arabic_Math}');
Expect(1, 126719, '\p{isarabicmath}', "");
Expect(0, 126719, '\p{^isarabicmath}', "");
Expect(0, 126719, '\P{isarabicmath}', "");
Expect(1, 126719, '\P{^isarabicmath}', "");
Expect(0, 126720, '\p{isarabicmath}', "");
Expect(1, 126720, '\p{^isarabicmath}', "");
Expect(1, 126720, '\P{isarabicmath}', "");
Expect(0, 126720, '\P{^isarabicmath}', "");
Expect(1, 126719, '\p{-is_ARABIC_MATH}', "");
Expect(0, 126719, '\p{^-is_ARABIC_MATH}', "");
Expect(0, 126719, '\P{-is_ARABIC_MATH}', "");
Expect(1, 126719, '\P{^-is_ARABIC_MATH}', "");
Expect(0, 126720, '\p{-is_ARABIC_MATH}', "");
Expect(1, 126720, '\p{^-is_ARABIC_MATH}', "");
Expect(1, 126720, '\P{-is_ARABIC_MATH}', "");
Expect(0, 126720, '\P{^-is_ARABIC_MATH}', "");
Error('\p{_In_Arabic_MATH/a/}');
Error('\P{_In_Arabic_MATH/a/}');
Expect(1, 126719, '\p{inarabicmath}', "");
Expect(0, 126719, '\p{^inarabicmath}', "");
Expect(0, 126719, '\P{inarabicmath}', "");
Expect(1, 126719, '\P{^inarabicmath}', "");
Expect(0, 126720, '\p{inarabicmath}', "");
Expect(1, 126720, '\p{^inarabicmath}', "");
Expect(1, 126720, '\P{inarabicmath}', "");
Expect(0, 126720, '\P{^inarabicmath}', "");
Expect(1, 126719, '\p{ in_arabic_MATH}', "");
Expect(0, 126719, '\p{^ in_arabic_MATH}', "");
Expect(0, 126719, '\P{ in_arabic_MATH}', "");
Expect(1, 126719, '\P{^ in_arabic_MATH}', "");
Expect(0, 126720, '\p{ in_arabic_MATH}', "");
Expect(1, 126720, '\p{^ in_arabic_MATH}', "");
Expect(1, 126720, '\P{ in_arabic_MATH}', "");
Expect(0, 126720, '\P{^ in_arabic_MATH}', "");
Error('\p{/a/ 	Bopomofo}');
Error('\P{/a/ 	Bopomofo}');
Expect(1, 12730, '\p{bopomofo}', "");
Expect(0, 12730, '\p{^bopomofo}', "");
Expect(0, 12730, '\P{bopomofo}', "");
Expect(1, 12730, '\P{^bopomofo}', "");
Expect(0, 12731, '\p{bopomofo}', "");
Expect(1, 12731, '\p{^bopomofo}', "");
Expect(1, 12731, '\P{bopomofo}', "");
Expect(0, 12731, '\P{^bopomofo}', "");
Expect(1, 12730, '\p{	-Bopomofo}', "");
Expect(0, 12730, '\p{^	-Bopomofo}', "");
Expect(0, 12730, '\P{	-Bopomofo}', "");
Expect(1, 12730, '\P{^	-Bopomofo}', "");
Expect(0, 12731, '\p{	-Bopomofo}', "");
Expect(1, 12731, '\p{^	-Bopomofo}', "");
Expect(1, 12731, '\P{	-Bopomofo}', "");
Expect(0, 12731, '\P{^	-Bopomofo}', "");
Error('\p{	-Is_Bopomofo/a/}');
Error('\P{	-Is_Bopomofo/a/}');
Expect(1, 12730, '\p{isbopomofo}', "");
Expect(0, 12730, '\p{^isbopomofo}', "");
Expect(0, 12730, '\P{isbopomofo}', "");
Expect(1, 12730, '\P{^isbopomofo}', "");
Expect(0, 12731, '\p{isbopomofo}', "");
Expect(1, 12731, '\p{^isbopomofo}', "");
Expect(1, 12731, '\P{isbopomofo}', "");
Expect(0, 12731, '\P{^isbopomofo}', "");
Expect(1, 12730, '\p{ -IS_BOPOMOFO}', "");
Expect(0, 12730, '\p{^ -IS_BOPOMOFO}', "");
Expect(0, 12730, '\P{ -IS_BOPOMOFO}', "");
Expect(1, 12730, '\P{^ -IS_BOPOMOFO}', "");
Expect(0, 12731, '\p{ -IS_BOPOMOFO}', "");
Expect(1, 12731, '\p{^ -IS_BOPOMOFO}', "");
Expect(1, 12731, '\P{ -IS_BOPOMOFO}', "");
Expect(0, 12731, '\P{^ -IS_BOPOMOFO}', "");
Error('\p{	BOPO/a/}');
Error('\P{	BOPO/a/}');
Expect(1, 12730, '\p{bopo}', "");
Expect(0, 12730, '\p{^bopo}', "");
Expect(0, 12730, '\P{bopo}', "");
Expect(1, 12730, '\P{^bopo}', "");
Expect(0, 12731, '\p{bopo}', "");
Expect(1, 12731, '\p{^bopo}', "");
Expect(1, 12731, '\P{bopo}', "");
Expect(0, 12731, '\P{^bopo}', "");
Expect(1, 12730, '\p{	_bopo}', "");
Expect(0, 12730, '\p{^	_bopo}', "");
Expect(0, 12730, '\P{	_bopo}', "");
Expect(1, 12730, '\P{^	_bopo}', "");
Expect(0, 12731, '\p{	_bopo}', "");
Expect(1, 12731, '\p{^	_bopo}', "");
Expect(1, 12731, '\P{	_bopo}', "");
Expect(0, 12731, '\P{^	_bopo}', "");
Error('\p{/a/	_Is_bopo}');
Error('\P{/a/	_Is_bopo}');
Expect(1, 12730, '\p{isbopo}', "");
Expect(0, 12730, '\p{^isbopo}', "");
Expect(0, 12730, '\P{isbopo}', "");
Expect(1, 12730, '\P{^isbopo}', "");
Expect(0, 12731, '\p{isbopo}', "");
Expect(1, 12731, '\p{^isbopo}', "");
Expect(1, 12731, '\P{isbopo}', "");
Expect(0, 12731, '\P{^isbopo}', "");
Expect(1, 12730, '\p{	IS_Bopo}', "");
Expect(0, 12730, '\p{^	IS_Bopo}', "");
Expect(0, 12730, '\P{	IS_Bopo}', "");
Expect(1, 12730, '\P{^	IS_Bopo}', "");
Expect(0, 12731, '\p{	IS_Bopo}', "");
Expect(1, 12731, '\p{^	IS_Bopo}', "");
Expect(1, 12731, '\P{	IS_Bopo}', "");
Expect(0, 12731, '\P{^	IS_Bopo}', "");
Error('\p{ miscellaneous_TECHNICAL:=}');
Error('\P{ miscellaneous_TECHNICAL:=}');
Expect(1, 9215, '\p{miscellaneoustechnical}', "");
Expect(0, 9215, '\p{^miscellaneoustechnical}', "");
Expect(0, 9215, '\P{miscellaneoustechnical}', "");
Expect(1, 9215, '\P{^miscellaneoustechnical}', "");
Expect(0, 9216, '\p{miscellaneoustechnical}', "");
Expect(1, 9216, '\p{^miscellaneoustechnical}', "");
Expect(1, 9216, '\P{miscellaneoustechnical}', "");
Expect(0, 9216, '\P{^miscellaneoustechnical}', "");
Expect(1, 9215, '\p{-	Miscellaneous_Technical}', "");
Expect(0, 9215, '\p{^-	Miscellaneous_Technical}', "");
Expect(0, 9215, '\P{-	Miscellaneous_Technical}', "");
Expect(1, 9215, '\P{^-	Miscellaneous_Technical}', "");
Expect(0, 9216, '\p{-	Miscellaneous_Technical}', "");
Expect(1, 9216, '\p{^-	Miscellaneous_Technical}', "");
Expect(1, 9216, '\P{-	Miscellaneous_Technical}', "");
Expect(0, 9216, '\P{^-	Miscellaneous_Technical}', "");
Error('\p{is_miscellaneous_TECHNICAL:=}');
Error('\P{is_miscellaneous_TECHNICAL:=}');
Expect(1, 9215, '\p{ismiscellaneoustechnical}', "");
Expect(0, 9215, '\p{^ismiscellaneoustechnical}', "");
Expect(0, 9215, '\P{ismiscellaneoustechnical}', "");
Expect(1, 9215, '\P{^ismiscellaneoustechnical}', "");
Expect(0, 9216, '\p{ismiscellaneoustechnical}', "");
Expect(1, 9216, '\p{^ismiscellaneoustechnical}', "");
Expect(1, 9216, '\P{ismiscellaneoustechnical}', "");
Expect(0, 9216, '\P{^ismiscellaneoustechnical}', "");
Expect(1, 9215, '\p{--is_Miscellaneous_TECHNICAL}', "");
Expect(0, 9215, '\p{^--is_Miscellaneous_TECHNICAL}', "");
Expect(0, 9215, '\P{--is_Miscellaneous_TECHNICAL}', "");
Expect(1, 9215, '\P{^--is_Miscellaneous_TECHNICAL}', "");
Expect(0, 9216, '\p{--is_Miscellaneous_TECHNICAL}', "");
Expect(1, 9216, '\p{^--is_Miscellaneous_TECHNICAL}', "");
Expect(1, 9216, '\P{--is_Miscellaneous_TECHNICAL}', "");
Expect(0, 9216, '\P{^--is_Miscellaneous_TECHNICAL}', "");
Error('\p{ :=In_MISCELLANEOUS_technical}');
Error('\P{ :=In_MISCELLANEOUS_technical}');
Expect(1, 9215, '\p{inmiscellaneoustechnical}', "");
Expect(0, 9215, '\p{^inmiscellaneoustechnical}', "");
Expect(0, 9215, '\P{inmiscellaneoustechnical}', "");
Expect(1, 9215, '\P{^inmiscellaneoustechnical}', "");
Expect(0, 9216, '\p{inmiscellaneoustechnical}', "");
Expect(1, 9216, '\p{^inmiscellaneoustechnical}', "");
Expect(1, 9216, '\P{inmiscellaneoustechnical}', "");
Expect(0, 9216, '\P{^inmiscellaneoustechnical}', "");
Expect(1, 9215, '\p{ IN_Miscellaneous_Technical}', "");
Expect(0, 9215, '\p{^ IN_Miscellaneous_Technical}', "");
Expect(0, 9215, '\P{ IN_Miscellaneous_Technical}', "");
Expect(1, 9215, '\P{^ IN_Miscellaneous_Technical}', "");
Expect(0, 9216, '\p{ IN_Miscellaneous_Technical}', "");
Expect(1, 9216, '\p{^ IN_Miscellaneous_Technical}', "");
Expect(1, 9216, '\P{ IN_Miscellaneous_Technical}', "");
Expect(0, 9216, '\P{^ IN_Miscellaneous_Technical}', "");
Error('\p{ MISC_TECHNICAL:=}');
Error('\P{ MISC_TECHNICAL:=}');
Expect(1, 9215, '\p{misctechnical}', "");
Expect(0, 9215, '\p{^misctechnical}', "");
Expect(0, 9215, '\P{misctechnical}', "");
Expect(1, 9215, '\P{^misctechnical}', "");
Expect(0, 9216, '\p{misctechnical}', "");
Expect(1, 9216, '\p{^misctechnical}', "");
Expect(1, 9216, '\P{misctechnical}', "");
Expect(0, 9216, '\P{^misctechnical}', "");
Expect(1, 9215, '\p{	-Misc_TECHNICAL}', "");
Expect(0, 9215, '\p{^	-Misc_TECHNICAL}', "");
Expect(0, 9215, '\P{	-Misc_TECHNICAL}', "");
Expect(1, 9215, '\P{^	-Misc_TECHNICAL}', "");
Expect(0, 9216, '\p{	-Misc_TECHNICAL}', "");
Expect(1, 9216, '\p{^	-Misc_TECHNICAL}', "");
Expect(1, 9216, '\P{	-Misc_TECHNICAL}', "");
Expect(0, 9216, '\P{^	-Misc_TECHNICAL}', "");
Error('\p{:= -IS_misc_TECHNICAL}');
Error('\P{:= -IS_misc_TECHNICAL}');
Expect(1, 9215, '\p{ismisctechnical}', "");
Expect(0, 9215, '\p{^ismisctechnical}', "");
Expect(0, 9215, '\P{ismisctechnical}', "");
Expect(1, 9215, '\P{^ismisctechnical}', "");
Expect(0, 9216, '\p{ismisctechnical}', "");
Expect(1, 9216, '\p{^ismisctechnical}', "");
Expect(1, 9216, '\P{ismisctechnical}', "");
Expect(0, 9216, '\P{^ismisctechnical}', "");
Expect(1, 9215, '\p{ -IS_Misc_technical}', "");
Expect(0, 9215, '\p{^ -IS_Misc_technical}', "");
Expect(0, 9215, '\P{ -IS_Misc_technical}', "");
Expect(1, 9215, '\P{^ -IS_Misc_technical}', "");
Expect(0, 9216, '\p{ -IS_Misc_technical}', "");
Expect(1, 9216, '\p{^ -IS_Misc_technical}', "");
Expect(1, 9216, '\P{ -IS_Misc_technical}', "");
Expect(0, 9216, '\P{^ -IS_Misc_technical}', "");
Error('\p{ In_misc_Technical/a/}');
Error('\P{ In_misc_Technical/a/}');
Expect(1, 9215, '\p{inmisctechnical}', "");
Expect(0, 9215, '\p{^inmisctechnical}', "");
Expect(0, 9215, '\P{inmisctechnical}', "");
Expect(1, 9215, '\P{^inmisctechnical}', "");
Expect(0, 9216, '\p{inmisctechnical}', "");
Expect(1, 9216, '\p{^inmisctechnical}', "");
Expect(1, 9216, '\P{inmisctechnical}', "");
Expect(0, 9216, '\P{^inmisctechnical}', "");
Expect(1, 9215, '\p{-	In_MISC_Technical}', "");
Expect(0, 9215, '\p{^-	In_MISC_Technical}', "");
Expect(0, 9215, '\P{-	In_MISC_Technical}', "");
Expect(1, 9215, '\P{^-	In_MISC_Technical}', "");
Expect(0, 9216, '\p{-	In_MISC_Technical}', "");
Expect(1, 9216, '\p{^-	In_MISC_Technical}', "");
Expect(1, 9216, '\P{-	In_MISC_Technical}', "");
Expect(0, 9216, '\P{^-	In_MISC_Technical}', "");
Error('\p{	_CJK_unified_Ideographs_extension_D/a/}');
Error('\P{	_CJK_unified_Ideographs_extension_D/a/}');
Expect(1, 178207, '\p{cjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\p{^cjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\P{cjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\P{^cjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\p{cjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\p{^cjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\P{cjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\P{^cjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\p{-_CJK_Unified_Ideographs_Extension_D}', "");
Expect(0, 178207, '\p{^-_CJK_Unified_Ideographs_Extension_D}', "");
Expect(0, 178207, '\P{-_CJK_Unified_Ideographs_Extension_D}', "");
Expect(1, 178207, '\P{^-_CJK_Unified_Ideographs_Extension_D}', "");
Expect(0, 178208, '\p{-_CJK_Unified_Ideographs_Extension_D}', "");
Expect(1, 178208, '\p{^-_CJK_Unified_Ideographs_Extension_D}', "");
Expect(1, 178208, '\P{-_CJK_Unified_Ideographs_Extension_D}', "");
Expect(0, 178208, '\P{^-_CJK_Unified_Ideographs_Extension_D}', "");
Error('\p{:=_-is_CJK_Unified_ideographs_extension_D}');
Error('\P{:=_-is_CJK_Unified_ideographs_extension_D}');
Expect(1, 178207, '\p{iscjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\p{^iscjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\P{iscjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\P{^iscjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\p{iscjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\p{^iscjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\P{iscjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\P{^iscjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\p{__is_cjk_Unified_IDEOGRAPHS_EXTENSION_D}', "");
Expect(0, 178207, '\p{^__is_cjk_Unified_IDEOGRAPHS_EXTENSION_D}', "");
Expect(0, 178207, '\P{__is_cjk_Unified_IDEOGRAPHS_EXTENSION_D}', "");
Expect(1, 178207, '\P{^__is_cjk_Unified_IDEOGRAPHS_EXTENSION_D}', "");
Expect(0, 178208, '\p{__is_cjk_Unified_IDEOGRAPHS_EXTENSION_D}', "");
Expect(1, 178208, '\p{^__is_cjk_Unified_IDEOGRAPHS_EXTENSION_D}', "");
Expect(1, 178208, '\P{__is_cjk_Unified_IDEOGRAPHS_EXTENSION_D}', "");
Expect(0, 178208, '\P{^__is_cjk_Unified_IDEOGRAPHS_EXTENSION_D}', "");
Error('\p{:=_-IN_cjk_Unified_Ideographs_Extension_D}');
Error('\P{:=_-IN_cjk_Unified_Ideographs_Extension_D}');
Expect(1, 178207, '\p{incjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\p{^incjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\P{incjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\P{^incjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\p{incjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\p{^incjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\P{incjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\P{^incjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\p{-	IN_CJK_Unified_Ideographs_extension_D}', "");
Expect(0, 178207, '\p{^-	IN_CJK_Unified_Ideographs_extension_D}', "");
Expect(0, 178207, '\P{-	IN_CJK_Unified_Ideographs_extension_D}', "");
Expect(1, 178207, '\P{^-	IN_CJK_Unified_Ideographs_extension_D}', "");
Expect(0, 178208, '\p{-	IN_CJK_Unified_Ideographs_extension_D}', "");
Expect(1, 178208, '\p{^-	IN_CJK_Unified_Ideographs_extension_D}', "");
Expect(1, 178208, '\P{-	IN_CJK_Unified_Ideographs_extension_D}', "");
Expect(0, 178208, '\P{^-	IN_CJK_Unified_Ideographs_extension_D}', "");
Error('\p{-/a/CJK_Ext_d}');
Error('\P{-/a/CJK_Ext_d}');
Expect(1, 178207, '\p{cjkextd}', "");
Expect(0, 178207, '\p{^cjkextd}', "");
Expect(0, 178207, '\P{cjkextd}', "");
Expect(1, 178207, '\P{^cjkextd}', "");
Expect(0, 178208, '\p{cjkextd}', "");
Expect(1, 178208, '\p{^cjkextd}', "");
Expect(1, 178208, '\P{cjkextd}', "");
Expect(0, 178208, '\P{^cjkextd}', "");
Expect(1, 178207, '\p{cjk_EXT_D}', "");
Expect(0, 178207, '\p{^cjk_EXT_D}', "");
Expect(0, 178207, '\P{cjk_EXT_D}', "");
Expect(1, 178207, '\P{^cjk_EXT_D}', "");
Expect(0, 178208, '\p{cjk_EXT_D}', "");
Expect(1, 178208, '\p{^cjk_EXT_D}', "");
Expect(1, 178208, '\P{cjk_EXT_D}', "");
Expect(0, 178208, '\P{^cjk_EXT_D}', "");
Error('\p{	Is_CJK_ext_D:=}');
Error('\P{	Is_CJK_ext_D:=}');
Expect(1, 178207, '\p{iscjkextd}', "");
Expect(0, 178207, '\p{^iscjkextd}', "");
Expect(0, 178207, '\P{iscjkextd}', "");
Expect(1, 178207, '\P{^iscjkextd}', "");
Expect(0, 178208, '\p{iscjkextd}', "");
Expect(1, 178208, '\p{^iscjkextd}', "");
Expect(1, 178208, '\P{iscjkextd}', "");
Expect(0, 178208, '\P{^iscjkextd}', "");
Expect(1, 178207, '\p{ is_CJK_Ext_D}', "");
Expect(0, 178207, '\p{^ is_CJK_Ext_D}', "");
Expect(0, 178207, '\P{ is_CJK_Ext_D}', "");
Expect(1, 178207, '\P{^ is_CJK_Ext_D}', "");
Expect(0, 178208, '\p{ is_CJK_Ext_D}', "");
Expect(1, 178208, '\p{^ is_CJK_Ext_D}', "");
Expect(1, 178208, '\P{ is_CJK_Ext_D}', "");
Expect(0, 178208, '\P{^ is_CJK_Ext_D}', "");
Error('\p{:=__IN_CJK_Ext_D}');
Error('\P{:=__IN_CJK_Ext_D}');
Expect(1, 178207, '\p{incjkextd}', "");
Expect(0, 178207, '\p{^incjkextd}', "");
Expect(0, 178207, '\P{incjkextd}', "");
Expect(1, 178207, '\P{^incjkextd}', "");
Expect(0, 178208, '\p{incjkextd}', "");
Expect(1, 178208, '\p{^incjkextd}', "");
Expect(1, 178208, '\P{incjkextd}', "");
Expect(0, 178208, '\P{^incjkextd}', "");
Expect(1, 178207, '\p{-_in_CJK_Ext_D}', "");
Expect(0, 178207, '\p{^-_in_CJK_Ext_D}', "");
Expect(0, 178207, '\P{-_in_CJK_Ext_D}', "");
Expect(1, 178207, '\P{^-_in_CJK_Ext_D}', "");
Expect(0, 178208, '\p{-_in_CJK_Ext_D}', "");
Expect(1, 178208, '\p{^-_in_CJK_Ext_D}', "");
Expect(1, 178208, '\P{-_in_CJK_Ext_D}', "");
Expect(0, 178208, '\P{^-_in_CJK_Ext_D}', "");
Error('\p{_/a/MAHJONG_Tiles}');
Error('\P{_/a/MAHJONG_Tiles}');
Expect(1, 127023, '\p{mahjongtiles}', "");
Expect(0, 127023, '\p{^mahjongtiles}', "");
Expect(0, 127023, '\P{mahjongtiles}', "");
Expect(1, 127023, '\P{^mahjongtiles}', "");
Expect(0, 127024, '\p{mahjongtiles}', "");
Expect(1, 127024, '\p{^mahjongtiles}', "");
Expect(1, 127024, '\P{mahjongtiles}', "");
Expect(0, 127024, '\P{^mahjongtiles}', "");
Expect(1, 127023, '\p{_	Mahjong_tiles}', "");
Expect(0, 127023, '\p{^_	Mahjong_tiles}', "");
Expect(0, 127023, '\P{_	Mahjong_tiles}', "");
Expect(1, 127023, '\P{^_	Mahjong_tiles}', "");
Expect(0, 127024, '\p{_	Mahjong_tiles}', "");
Expect(1, 127024, '\p{^_	Mahjong_tiles}', "");
Expect(1, 127024, '\P{_	Mahjong_tiles}', "");
Expect(0, 127024, '\P{^_	Mahjong_tiles}', "");
Error('\p{	IS_Mahjong_Tiles:=}');
Error('\P{	IS_Mahjong_Tiles:=}');
Expect(1, 127023, '\p{ismahjongtiles}', "");
Expect(0, 127023, '\p{^ismahjongtiles}', "");
Expect(0, 127023, '\P{ismahjongtiles}', "");
Expect(1, 127023, '\P{^ismahjongtiles}', "");
Expect(0, 127024, '\p{ismahjongtiles}', "");
Expect(1, 127024, '\p{^ismahjongtiles}', "");
Expect(1, 127024, '\P{ismahjongtiles}', "");
Expect(0, 127024, '\P{^ismahjongtiles}', "");
Expect(1, 127023, '\p{	 is_mahjong_tiles}', "");
Expect(0, 127023, '\p{^	 is_mahjong_tiles}', "");
Expect(0, 127023, '\P{	 is_mahjong_tiles}', "");
Expect(1, 127023, '\P{^	 is_mahjong_tiles}', "");
Expect(0, 127024, '\p{	 is_mahjong_tiles}', "");
Expect(1, 127024, '\p{^	 is_mahjong_tiles}', "");
Expect(1, 127024, '\P{	 is_mahjong_tiles}', "");
Expect(0, 127024, '\P{^	 is_mahjong_tiles}', "");
Error('\p{/a/ In_MAHJONG_TILES}');
Error('\P{/a/ In_MAHJONG_TILES}');
Expect(1, 127023, '\p{inmahjongtiles}', "");
Expect(0, 127023, '\p{^inmahjongtiles}', "");
Expect(0, 127023, '\P{inmahjongtiles}', "");
Expect(1, 127023, '\P{^inmahjongtiles}', "");
Expect(0, 127024, '\p{inmahjongtiles}', "");
Expect(1, 127024, '\p{^inmahjongtiles}', "");
Expect(1, 127024, '\P{inmahjongtiles}', "");
Expect(0, 127024, '\P{^inmahjongtiles}', "");
Expect(1, 127023, '\p{_In_mahjong_tiles}', "");
Expect(0, 127023, '\p{^_In_mahjong_tiles}', "");
Expect(0, 127023, '\P{_In_mahjong_tiles}', "");
Expect(1, 127023, '\P{^_In_mahjong_tiles}', "");
Expect(0, 127024, '\p{_In_mahjong_tiles}', "");
Expect(1, 127024, '\p{^_In_mahjong_tiles}', "");
Expect(1, 127024, '\P{_In_mahjong_tiles}', "");
Expect(0, 127024, '\P{^_In_mahjong_tiles}', "");
Error('\p{ :=Mahjong}');
Error('\P{ :=Mahjong}');
Expect(1, 127023, '\p{mahjong}', "");
Expect(0, 127023, '\p{^mahjong}', "");
Expect(0, 127023, '\P{mahjong}', "");
Expect(1, 127023, '\P{^mahjong}', "");
Expect(0, 127024, '\p{mahjong}', "");
Expect(1, 127024, '\p{^mahjong}', "");
Expect(1, 127024, '\P{mahjong}', "");
Expect(0, 127024, '\P{^mahjong}', "");
Expect(1, 127023, '\p{		MAHJONG}', "");
Expect(0, 127023, '\p{^		MAHJONG}', "");
Expect(0, 127023, '\P{		MAHJONG}', "");
Expect(1, 127023, '\P{^		MAHJONG}', "");
Expect(0, 127024, '\p{		MAHJONG}', "");
Expect(1, 127024, '\p{^		MAHJONG}', "");
Expect(1, 127024, '\P{		MAHJONG}', "");
Expect(0, 127024, '\P{^		MAHJONG}', "");
Error('\p{	Is_MAHJONG:=}');
Error('\P{	Is_MAHJONG:=}');
Expect(1, 127023, '\p{ismahjong}', "");
Expect(0, 127023, '\p{^ismahjong}', "");
Expect(0, 127023, '\P{ismahjong}', "");
Expect(1, 127023, '\P{^ismahjong}', "");
Expect(0, 127024, '\p{ismahjong}', "");
Expect(1, 127024, '\p{^ismahjong}', "");
Expect(1, 127024, '\P{ismahjong}', "");
Expect(0, 127024, '\P{^ismahjong}', "");
Expect(1, 127023, '\p{Is_Mahjong}', "");
Expect(0, 127023, '\p{^Is_Mahjong}', "");
Expect(0, 127023, '\P{Is_Mahjong}', "");
Expect(1, 127023, '\P{^Is_Mahjong}', "");
Expect(0, 127024, '\p{Is_Mahjong}', "");
Expect(1, 127024, '\p{^Is_Mahjong}', "");
Expect(1, 127024, '\P{Is_Mahjong}', "");
Expect(0, 127024, '\P{^Is_Mahjong}', "");
Error('\p{-:=In_Mahjong}');
Error('\P{-:=In_Mahjong}');
Expect(1, 127023, '\p{inmahjong}', "");
Expect(0, 127023, '\p{^inmahjong}', "");
Expect(0, 127023, '\P{inmahjong}', "");
Expect(1, 127023, '\P{^inmahjong}', "");
Expect(0, 127024, '\p{inmahjong}', "");
Expect(1, 127024, '\p{^inmahjong}', "");
Expect(1, 127024, '\P{inmahjong}', "");
Expect(0, 127024, '\P{^inmahjong}', "");
Expect(1, 127023, '\p{_ In_MAHJONG}', "");
Expect(0, 127023, '\p{^_ In_MAHJONG}', "");
Expect(0, 127023, '\P{_ In_MAHJONG}', "");
Expect(1, 127023, '\P{^_ In_MAHJONG}', "");
Expect(0, 127024, '\p{_ In_MAHJONG}', "");
Expect(1, 127024, '\p{^_ In_MAHJONG}', "");
Expect(1, 127024, '\P{_ In_MAHJONG}', "");
Expect(0, 127024, '\P{^_ In_MAHJONG}', "");
Error('\p{-miscellaneous_Mathematical_Symbols_B:=}');
Error('\P{-miscellaneous_Mathematical_Symbols_B:=}');
Expect(1, 10751, '\p{miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\p{^miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\P{miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\P{^miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\p{miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\p{^miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\P{miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\P{^miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\p{_miscellaneous_mathematical_Symbols_B}', "");
Expect(0, 10751, '\p{^_miscellaneous_mathematical_Symbols_B}', "");
Expect(0, 10751, '\P{_miscellaneous_mathematical_Symbols_B}', "");
Expect(1, 10751, '\P{^_miscellaneous_mathematical_Symbols_B}', "");
Expect(0, 10752, '\p{_miscellaneous_mathematical_Symbols_B}', "");
Expect(1, 10752, '\p{^_miscellaneous_mathematical_Symbols_B}', "");
Expect(1, 10752, '\P{_miscellaneous_mathematical_Symbols_B}', "");
Expect(0, 10752, '\P{^_miscellaneous_mathematical_Symbols_B}', "");
Error('\p{ _IS_miscellaneous_mathematical_SYMBOLS_B:=}');
Error('\P{ _IS_miscellaneous_mathematical_SYMBOLS_B:=}');
Expect(1, 10751, '\p{ismiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\p{^ismiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\P{ismiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\P{^ismiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\p{ismiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\p{^ismiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\P{ismiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\P{^ismiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\p{--Is_Miscellaneous_Mathematical_SYMBOLS_B}', "");
Expect(0, 10751, '\p{^--Is_Miscellaneous_Mathematical_SYMBOLS_B}', "");
Expect(0, 10751, '\P{--Is_Miscellaneous_Mathematical_SYMBOLS_B}', "");
Expect(1, 10751, '\P{^--Is_Miscellaneous_Mathematical_SYMBOLS_B}', "");
Expect(0, 10752, '\p{--Is_Miscellaneous_Mathematical_SYMBOLS_B}', "");
Expect(1, 10752, '\p{^--Is_Miscellaneous_Mathematical_SYMBOLS_B}', "");
Expect(1, 10752, '\P{--Is_Miscellaneous_Mathematical_SYMBOLS_B}', "");
Expect(0, 10752, '\P{^--Is_Miscellaneous_Mathematical_SYMBOLS_B}', "");
Error('\p{-/a/In_Miscellaneous_MATHEMATICAL_SYMBOLS_b}');
Error('\P{-/a/In_Miscellaneous_MATHEMATICAL_SYMBOLS_b}');
Expect(1, 10751, '\p{inmiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\p{^inmiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\P{inmiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\P{^inmiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\p{inmiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\p{^inmiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\P{inmiscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\P{^inmiscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\p{_	in_MISCELLANEOUS_mathematical_Symbols_b}', "");
Expect(0, 10751, '\p{^_	in_MISCELLANEOUS_mathematical_Symbols_b}', "");
Expect(0, 10751, '\P{_	in_MISCELLANEOUS_mathematical_Symbols_b}', "");
Expect(1, 10751, '\P{^_	in_MISCELLANEOUS_mathematical_Symbols_b}', "");
Expect(0, 10752, '\p{_	in_MISCELLANEOUS_mathematical_Symbols_b}', "");
Expect(1, 10752, '\p{^_	in_MISCELLANEOUS_mathematical_Symbols_b}', "");
Expect(1, 10752, '\P{_	in_MISCELLANEOUS_mathematical_Symbols_b}', "");
Expect(0, 10752, '\P{^_	in_MISCELLANEOUS_mathematical_Symbols_b}', "");
Error('\p{-_misc_MATH_symbols_B:=}');
Error('\P{-_misc_MATH_symbols_B:=}');
Expect(1, 10751, '\p{miscmathsymbolsb}', "");
Expect(0, 10751, '\p{^miscmathsymbolsb}', "");
Expect(0, 10751, '\P{miscmathsymbolsb}', "");
Expect(1, 10751, '\P{^miscmathsymbolsb}', "");
Expect(0, 10752, '\p{miscmathsymbolsb}', "");
Expect(1, 10752, '\p{^miscmathsymbolsb}', "");
Expect(1, 10752, '\P{miscmathsymbolsb}', "");
Expect(0, 10752, '\P{^miscmathsymbolsb}', "");
Expect(1, 10751, '\p{ misc_Math_Symbols_B}', "");
Expect(0, 10751, '\p{^ misc_Math_Symbols_B}', "");
Expect(0, 10751, '\P{ misc_Math_Symbols_B}', "");
Expect(1, 10751, '\P{^ misc_Math_Symbols_B}', "");
Expect(0, 10752, '\p{ misc_Math_Symbols_B}', "");
Expect(1, 10752, '\p{^ misc_Math_Symbols_B}', "");
Expect(1, 10752, '\P{ misc_Math_Symbols_B}', "");
Expect(0, 10752, '\P{^ misc_Math_Symbols_B}', "");
Error('\p{	-Is_Misc_Math_Symbols_B:=}');
Error('\P{	-Is_Misc_Math_Symbols_B:=}');
Expect(1, 10751, '\p{ismiscmathsymbolsb}', "");
Expect(0, 10751, '\p{^ismiscmathsymbolsb}', "");
Expect(0, 10751, '\P{ismiscmathsymbolsb}', "");
Expect(1, 10751, '\P{^ismiscmathsymbolsb}', "");
Expect(0, 10752, '\p{ismiscmathsymbolsb}', "");
Expect(1, 10752, '\p{^ismiscmathsymbolsb}', "");
Expect(1, 10752, '\P{ismiscmathsymbolsb}', "");
Expect(0, 10752, '\P{^ismiscmathsymbolsb}', "");
Expect(1, 10751, '\p{ 	Is_misc_MATH_symbols_B}', "");
Expect(0, 10751, '\p{^ 	Is_misc_MATH_symbols_B}', "");
Expect(0, 10751, '\P{ 	Is_misc_MATH_symbols_B}', "");
Expect(1, 10751, '\P{^ 	Is_misc_MATH_symbols_B}', "");
Expect(0, 10752, '\p{ 	Is_misc_MATH_symbols_B}', "");
Expect(1, 10752, '\p{^ 	Is_misc_MATH_symbols_B}', "");
Expect(1, 10752, '\P{ 	Is_misc_MATH_symbols_B}', "");
Expect(0, 10752, '\P{^ 	Is_misc_MATH_symbols_B}', "");
Error('\p{	IN_misc_Math_symbols_B:=}');
Error('\P{	IN_misc_Math_symbols_B:=}');
Expect(1, 10751, '\p{inmiscmathsymbolsb}', "");
Expect(0, 10751, '\p{^inmiscmathsymbolsb}', "");
Expect(0, 10751, '\P{inmiscmathsymbolsb}', "");
Expect(1, 10751, '\P{^inmiscmathsymbolsb}', "");
Expect(0, 10752, '\p{inmiscmathsymbolsb}', "");
Expect(1, 10752, '\p{^inmiscmathsymbolsb}', "");
Expect(1, 10752, '\P{inmiscmathsymbolsb}', "");
Expect(0, 10752, '\P{^inmiscmathsymbolsb}', "");
Expect(1, 10751, '\p{ _In_Misc_Math_symbols_B}', "");
Expect(0, 10751, '\p{^ _In_Misc_Math_symbols_B}', "");
Expect(0, 10751, '\P{ _In_Misc_Math_symbols_B}', "");
Expect(1, 10751, '\P{^ _In_Misc_Math_symbols_B}', "");
Expect(0, 10752, '\p{ _In_Misc_Math_symbols_B}', "");
Expect(1, 10752, '\p{^ _In_Misc_Math_symbols_B}', "");
Expect(1, 10752, '\P{ _In_Misc_Math_symbols_B}', "");
Expect(0, 10752, '\P{^ _In_Misc_Math_symbols_B}', "");
Error('\p{:=__Enclosed_ideographic_supplement}');
Error('\P{:=__Enclosed_ideographic_supplement}');
Expect(1, 127743, '\p{enclosedideographicsupplement}', "");
Expect(0, 127743, '\p{^enclosedideographicsupplement}', "");
Expect(0, 127743, '\P{enclosedideographicsupplement}', "");
Expect(1, 127743, '\P{^enclosedideographicsupplement}', "");
Expect(0, 127744, '\p{enclosedideographicsupplement}', "");
Expect(1, 127744, '\p{^enclosedideographicsupplement}', "");
Expect(1, 127744, '\P{enclosedideographicsupplement}', "");
Expect(0, 127744, '\P{^enclosedideographicsupplement}', "");
Expect(1, 127743, '\p{ _ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127743, '\p{^ _ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127743, '\P{ _ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(1, 127743, '\P{^ _ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127744, '\p{ _ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(1, 127744, '\p{^ _ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(1, 127744, '\P{ _ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127744, '\P{^ _ENCLOSED_IDEOGRAPHIC_Supplement}', "");
Error('\p{:=		Is_Enclosed_Ideographic_SUPPLEMENT}');
Error('\P{:=		Is_Enclosed_Ideographic_SUPPLEMENT}');
Expect(1, 127743, '\p{isenclosedideographicsupplement}', "");
Expect(0, 127743, '\p{^isenclosedideographicsupplement}', "");
Expect(0, 127743, '\P{isenclosedideographicsupplement}', "");
Expect(1, 127743, '\P{^isenclosedideographicsupplement}', "");
Expect(0, 127744, '\p{isenclosedideographicsupplement}', "");
Expect(1, 127744, '\p{^isenclosedideographicsupplement}', "");
Expect(1, 127744, '\P{isenclosedideographicsupplement}', "");
Expect(0, 127744, '\P{^isenclosedideographicsupplement}', "");
Expect(1, 127743, '\p{_	Is_ENCLOSED_Ideographic_supplement}', "");
Expect(0, 127743, '\p{^_	Is_ENCLOSED_Ideographic_supplement}', "");
Expect(0, 127743, '\P{_	Is_ENCLOSED_Ideographic_supplement}', "");
Expect(1, 127743, '\P{^_	Is_ENCLOSED_Ideographic_supplement}', "");
Expect(0, 127744, '\p{_	Is_ENCLOSED_Ideographic_supplement}', "");
Expect(1, 127744, '\p{^_	Is_ENCLOSED_Ideographic_supplement}', "");
Expect(1, 127744, '\P{_	Is_ENCLOSED_Ideographic_supplement}', "");
Expect(0, 127744, '\P{^_	Is_ENCLOSED_Ideographic_supplement}', "");
Error('\p{:=	 in_Enclosed_Ideographic_SUPPLEMENT}');
Error('\P{:=	 in_Enclosed_Ideographic_SUPPLEMENT}');
Expect(1, 127743, '\p{inenclosedideographicsupplement}', "");
Expect(0, 127743, '\p{^inenclosedideographicsupplement}', "");
Expect(0, 127743, '\P{inenclosedideographicsupplement}', "");
Expect(1, 127743, '\P{^inenclosedideographicsupplement}', "");
Expect(0, 127744, '\p{inenclosedideographicsupplement}', "");
Expect(1, 127744, '\p{^inenclosedideographicsupplement}', "");
Expect(1, 127744, '\P{inenclosedideographicsupplement}', "");
Expect(0, 127744, '\P{^inenclosedideographicsupplement}', "");
Expect(1, 127743, '\p{- In_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(0, 127743, '\p{^- In_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(0, 127743, '\P{- In_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(1, 127743, '\P{^- In_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(0, 127744, '\p{- In_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(1, 127744, '\p{^- In_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(1, 127744, '\P{- In_Enclosed_IDEOGRAPHIC_supplement}', "");
Expect(0, 127744, '\P{^- In_Enclosed_IDEOGRAPHIC_supplement}', "");
Error('\p{	/a/enclosed_Ideographic_Sup}');
Error('\P{	/a/enclosed_Ideographic_Sup}');
Expect(1, 127743, '\p{enclosedideographicsup}', "");
Expect(0, 127743, '\p{^enclosedideographicsup}', "");
Expect(0, 127743, '\P{enclosedideographicsup}', "");
Expect(1, 127743, '\P{^enclosedideographicsup}', "");
Expect(0, 127744, '\p{enclosedideographicsup}', "");
Expect(1, 127744, '\p{^enclosedideographicsup}', "");
Expect(1, 127744, '\P{enclosedideographicsup}', "");
Expect(0, 127744, '\P{^enclosedideographicsup}', "");
Expect(1, 127743, '\p{ Enclosed_IDEOGRAPHIC_SUP}', "");
Expect(0, 127743, '\p{^ Enclosed_IDEOGRAPHIC_SUP}', "");
Expect(0, 127743, '\P{ Enclosed_IDEOGRAPHIC_SUP}', "");
Expect(1, 127743, '\P{^ Enclosed_IDEOGRAPHIC_SUP}', "");
Expect(0, 127744, '\p{ Enclosed_IDEOGRAPHIC_SUP}', "");
Expect(1, 127744, '\p{^ Enclosed_IDEOGRAPHIC_SUP}', "");
Expect(1, 127744, '\P{ Enclosed_IDEOGRAPHIC_SUP}', "");
Expect(0, 127744, '\P{^ Enclosed_IDEOGRAPHIC_SUP}', "");
Error('\p{Is_ENCLOSED_ideographic_Sup:=}');
Error('\P{Is_ENCLOSED_ideographic_Sup:=}');
Expect(1, 127743, '\p{isenclosedideographicsup}', "");
Expect(0, 127743, '\p{^isenclosedideographicsup}', "");
Expect(0, 127743, '\P{isenclosedideographicsup}', "");
Expect(1, 127743, '\P{^isenclosedideographicsup}', "");
Expect(0, 127744, '\p{isenclosedideographicsup}', "");
Expect(1, 127744, '\p{^isenclosedideographicsup}', "");
Expect(1, 127744, '\P{isenclosedideographicsup}', "");
Expect(0, 127744, '\P{^isenclosedideographicsup}', "");
Expect(1, 127743, '\p{	Is_Enclosed_ideographic_SUP}', "");
Expect(0, 127743, '\p{^	Is_Enclosed_ideographic_SUP}', "");
Expect(0, 127743, '\P{	Is_Enclosed_ideographic_SUP}', "");
Expect(1, 127743, '\P{^	Is_Enclosed_ideographic_SUP}', "");
Expect(0, 127744, '\p{	Is_Enclosed_ideographic_SUP}', "");
Expect(1, 127744, '\p{^	Is_Enclosed_ideographic_SUP}', "");
Expect(1, 127744, '\P{	Is_Enclosed_ideographic_SUP}', "");
Expect(0, 127744, '\P{^	Is_Enclosed_ideographic_SUP}', "");
Error('\p{-	In_enclosed_IDEOGRAPHIC_Sup/a/}');
Error('\P{-	In_enclosed_IDEOGRAPHIC_Sup/a/}');
Expect(1, 127743, '\p{inenclosedideographicsup}', "");
Expect(0, 127743, '\p{^inenclosedideographicsup}', "");
Expect(0, 127743, '\P{inenclosedideographicsup}', "");
Expect(1, 127743, '\P{^inenclosedideographicsup}', "");
Expect(0, 127744, '\p{inenclosedideographicsup}', "");
Expect(1, 127744, '\p{^inenclosedideographicsup}', "");
Expect(1, 127744, '\P{inenclosedideographicsup}', "");
Expect(0, 127744, '\P{^inenclosedideographicsup}', "");
Expect(1, 127743, '\p{- IN_Enclosed_Ideographic_Sup}', "");
Expect(0, 127743, '\p{^- IN_Enclosed_Ideographic_Sup}', "");
Expect(0, 127743, '\P{- IN_Enclosed_Ideographic_Sup}', "");
Expect(1, 127743, '\P{^- IN_Enclosed_Ideographic_Sup}', "");
Expect(0, 127744, '\p{- IN_Enclosed_Ideographic_Sup}', "");
Expect(1, 127744, '\p{^- IN_Enclosed_Ideographic_Sup}', "");
Expect(1, 127744, '\P{- IN_Enclosed_Ideographic_Sup}', "");
Expect(0, 127744, '\P{^- IN_Enclosed_Ideographic_Sup}', "");
Error('\p{/a/malayalam}');
Error('\P{/a/malayalam}');
Expect(1, 3455, '\p{malayalam}', "");
Expect(0, 3455, '\p{^malayalam}', "");
Expect(0, 3455, '\P{malayalam}', "");
Expect(1, 3455, '\P{^malayalam}', "");
Expect(0, 3456, '\p{malayalam}', "");
Expect(1, 3456, '\p{^malayalam}', "");
Expect(1, 3456, '\P{malayalam}', "");
Expect(0, 3456, '\P{^malayalam}', "");
Expect(1, 3455, '\p{Malayalam}', "");
Expect(0, 3455, '\p{^Malayalam}', "");
Expect(0, 3455, '\P{Malayalam}', "");
Expect(1, 3455, '\P{^Malayalam}', "");
Expect(0, 3456, '\p{Malayalam}', "");
Expect(1, 3456, '\p{^Malayalam}', "");
Expect(1, 3456, '\P{Malayalam}', "");
Expect(0, 3456, '\P{^Malayalam}', "");
Error('\p{_:=Is_malayalam}');
Error('\P{_:=Is_malayalam}');
Expect(1, 3455, '\p{ismalayalam}', "");
Expect(0, 3455, '\p{^ismalayalam}', "");
Expect(0, 3455, '\P{ismalayalam}', "");
Expect(1, 3455, '\P{^ismalayalam}', "");
Expect(0, 3456, '\p{ismalayalam}', "");
Expect(1, 3456, '\p{^ismalayalam}', "");
Expect(1, 3456, '\P{ismalayalam}', "");
Expect(0, 3456, '\P{^ismalayalam}', "");
Expect(1, 3455, '\p{-Is_malayalam}', "");
Expect(0, 3455, '\p{^-Is_malayalam}', "");
Expect(0, 3455, '\P{-Is_malayalam}', "");
Expect(1, 3455, '\P{^-Is_malayalam}', "");
Expect(0, 3456, '\p{-Is_malayalam}', "");
Expect(1, 3456, '\p{^-Is_malayalam}', "");
Expect(1, 3456, '\P{-Is_malayalam}', "");
Expect(0, 3456, '\P{^-Is_malayalam}', "");
Error('\p{/a/-Mlym}');
Error('\P{/a/-Mlym}');
Expect(1, 3455, '\p{mlym}', "");
Expect(0, 3455, '\p{^mlym}', "");
Expect(0, 3455, '\P{mlym}', "");
Expect(1, 3455, '\P{^mlym}', "");
Expect(0, 3456, '\p{mlym}', "");
Expect(1, 3456, '\p{^mlym}', "");
Expect(1, 3456, '\P{mlym}', "");
Expect(0, 3456, '\P{^mlym}', "");
Expect(1, 3455, '\p{ mlym}', "");
Expect(0, 3455, '\p{^ mlym}', "");
Expect(0, 3455, '\P{ mlym}', "");
Expect(1, 3455, '\P{^ mlym}', "");
Expect(0, 3456, '\p{ mlym}', "");
Expect(1, 3456, '\p{^ mlym}', "");
Expect(1, 3456, '\P{ mlym}', "");
Expect(0, 3456, '\P{^ mlym}', "");
Error('\p{_:=is_Mlym}');
Error('\P{_:=is_Mlym}');
Expect(1, 3455, '\p{ismlym}', "");
Expect(0, 3455, '\p{^ismlym}', "");
Expect(0, 3455, '\P{ismlym}', "");
Expect(1, 3455, '\P{^ismlym}', "");
Expect(0, 3456, '\p{ismlym}', "");
Expect(1, 3456, '\p{^ismlym}', "");
Expect(1, 3456, '\P{ismlym}', "");
Expect(0, 3456, '\P{^ismlym}', "");
Expect(1, 3455, '\p{__is_Mlym}', "");
Expect(0, 3455, '\p{^__is_Mlym}', "");
Expect(0, 3455, '\P{__is_Mlym}', "");
Expect(1, 3455, '\P{^__is_Mlym}', "");
Expect(0, 3456, '\p{__is_Mlym}', "");
Expect(1, 3456, '\p{^__is_Mlym}', "");
Expect(1, 3456, '\P{__is_Mlym}', "");
Expect(0, 3456, '\P{^__is_Mlym}', "");
Error('\p{_-vedic_Extensions/a/}');
Error('\P{_-vedic_Extensions/a/}');
Expect(1, 7423, '\p{vedicextensions}', "");
Expect(0, 7423, '\p{^vedicextensions}', "");
Expect(0, 7423, '\P{vedicextensions}', "");
Expect(1, 7423, '\P{^vedicextensions}', "");
Expect(0, 7424, '\p{vedicextensions}', "");
Expect(1, 7424, '\p{^vedicextensions}', "");
Expect(1, 7424, '\P{vedicextensions}', "");
Expect(0, 7424, '\P{^vedicextensions}', "");
Expect(1, 7423, '\p{ Vedic_EXTENSIONS}', "");
Expect(0, 7423, '\p{^ Vedic_EXTENSIONS}', "");
Expect(0, 7423, '\P{ Vedic_EXTENSIONS}', "");
Expect(1, 7423, '\P{^ Vedic_EXTENSIONS}', "");
Expect(0, 7424, '\p{ Vedic_EXTENSIONS}', "");
Expect(1, 7424, '\p{^ Vedic_EXTENSIONS}', "");
Expect(1, 7424, '\P{ Vedic_EXTENSIONS}', "");
Expect(0, 7424, '\P{^ Vedic_EXTENSIONS}', "");
Error('\p{/a/ 	Is_vedic_Extensions}');
Error('\P{/a/ 	Is_vedic_Extensions}');
Expect(1, 7423, '\p{isvedicextensions}', "");
Expect(0, 7423, '\p{^isvedicextensions}', "");
Expect(0, 7423, '\P{isvedicextensions}', "");
Expect(1, 7423, '\P{^isvedicextensions}', "");
Expect(0, 7424, '\p{isvedicextensions}', "");
Expect(1, 7424, '\p{^isvedicextensions}', "");
Expect(1, 7424, '\P{isvedicextensions}', "");
Expect(0, 7424, '\P{^isvedicextensions}', "");
Expect(1, 7423, '\p{_	IS_VEDIC_Extensions}', "");
Expect(0, 7423, '\p{^_	IS_VEDIC_Extensions}', "");
Expect(0, 7423, '\P{_	IS_VEDIC_Extensions}', "");
Expect(1, 7423, '\P{^_	IS_VEDIC_Extensions}', "");
Expect(0, 7424, '\p{_	IS_VEDIC_Extensions}', "");
Expect(1, 7424, '\p{^_	IS_VEDIC_Extensions}', "");
Expect(1, 7424, '\P{_	IS_VEDIC_Extensions}', "");
Expect(0, 7424, '\P{^_	IS_VEDIC_Extensions}', "");
Error('\p{	/a/In_Vedic_extensions}');
Error('\P{	/a/In_Vedic_extensions}');
Expect(1, 7423, '\p{invedicextensions}', "");
Expect(0, 7423, '\p{^invedicextensions}', "");
Expect(0, 7423, '\P{invedicextensions}', "");
Expect(1, 7423, '\P{^invedicextensions}', "");
Expect(0, 7424, '\p{invedicextensions}', "");
Expect(1, 7424, '\p{^invedicextensions}', "");
Expect(1, 7424, '\P{invedicextensions}', "");
Expect(0, 7424, '\P{^invedicextensions}', "");
Expect(1, 7423, '\p{-In_VEDIC_EXTENSIONS}', "");
Expect(0, 7423, '\p{^-In_VEDIC_EXTENSIONS}', "");
Expect(0, 7423, '\P{-In_VEDIC_EXTENSIONS}', "");
Expect(1, 7423, '\P{^-In_VEDIC_EXTENSIONS}', "");
Expect(0, 7424, '\p{-In_VEDIC_EXTENSIONS}', "");
Expect(1, 7424, '\p{^-In_VEDIC_EXTENSIONS}', "");
Expect(1, 7424, '\P{-In_VEDIC_EXTENSIONS}', "");
Expect(0, 7424, '\P{^-In_VEDIC_EXTENSIONS}', "");
Error('\p{:=VEDIC_ext}');
Error('\P{:=VEDIC_ext}');
Expect(1, 7423, '\p{vedicext}', "");
Expect(0, 7423, '\p{^vedicext}', "");
Expect(0, 7423, '\P{vedicext}', "");
Expect(1, 7423, '\P{^vedicext}', "");
Expect(0, 7424, '\p{vedicext}', "");
Expect(1, 7424, '\p{^vedicext}', "");
Expect(1, 7424, '\P{vedicext}', "");
Expect(0, 7424, '\P{^vedicext}', "");
Expect(1, 7423, '\p{-_Vedic_ext}', "");
Expect(0, 7423, '\p{^-_Vedic_ext}', "");
Expect(0, 7423, '\P{-_Vedic_ext}', "");
Expect(1, 7423, '\P{^-_Vedic_ext}', "");
Expect(0, 7424, '\p{-_Vedic_ext}', "");
Expect(1, 7424, '\p{^-_Vedic_ext}', "");
Expect(1, 7424, '\P{-_Vedic_ext}', "");
Expect(0, 7424, '\P{^-_Vedic_ext}', "");
Error('\p{_	Is_VEDIC_Ext:=}');
Error('\P{_	Is_VEDIC_Ext:=}');
Expect(1, 7423, '\p{isvedicext}', "");
Expect(0, 7423, '\p{^isvedicext}', "");
Expect(0, 7423, '\P{isvedicext}', "");
Expect(1, 7423, '\P{^isvedicext}', "");
Expect(0, 7424, '\p{isvedicext}', "");
Expect(1, 7424, '\p{^isvedicext}', "");
Expect(1, 7424, '\P{isvedicext}', "");
Expect(0, 7424, '\P{^isvedicext}', "");
Expect(1, 7423, '\p{ -Is_vedic_Ext}', "");
Expect(0, 7423, '\p{^ -Is_vedic_Ext}', "");
Expect(0, 7423, '\P{ -Is_vedic_Ext}', "");
Expect(1, 7423, '\P{^ -Is_vedic_Ext}', "");
Expect(0, 7424, '\p{ -Is_vedic_Ext}', "");
Expect(1, 7424, '\p{^ -Is_vedic_Ext}', "");
Expect(1, 7424, '\P{ -Is_vedic_Ext}', "");
Expect(0, 7424, '\P{^ -Is_vedic_Ext}', "");
Error('\p{-:=In_VEDIC_Ext}');
Error('\P{-:=In_VEDIC_Ext}');
Expect(1, 7423, '\p{invedicext}', "");
Expect(0, 7423, '\p{^invedicext}', "");
Expect(0, 7423, '\P{invedicext}', "");
Expect(1, 7423, '\P{^invedicext}', "");
Expect(0, 7424, '\p{invedicext}', "");
Expect(1, 7424, '\p{^invedicext}', "");
Expect(1, 7424, '\P{invedicext}', "");
Expect(0, 7424, '\P{^invedicext}', "");
Expect(1, 7423, '\p{	in_VEDIC_EXT}', "");
Expect(0, 7423, '\p{^	in_VEDIC_EXT}', "");
Expect(0, 7423, '\P{	in_VEDIC_EXT}', "");
Expect(1, 7423, '\P{^	in_VEDIC_EXT}', "");
Expect(0, 7424, '\p{	in_VEDIC_EXT}', "");
Expect(1, 7424, '\p{^	in_VEDIC_EXT}', "");
Expect(1, 7424, '\P{	in_VEDIC_EXT}', "");
Expect(0, 7424, '\P{^	in_VEDIC_EXT}', "");
Error('\p{:=	OLD_Turkic}');
Error('\P{:=	OLD_Turkic}');
Expect(1, 68680, '\p{oldturkic}', "");
Expect(0, 68680, '\p{^oldturkic}', "");
Expect(0, 68680, '\P{oldturkic}', "");
Expect(1, 68680, '\P{^oldturkic}', "");
Expect(0, 68681, '\p{oldturkic}', "");
Expect(1, 68681, '\p{^oldturkic}', "");
Expect(1, 68681, '\P{oldturkic}', "");
Expect(0, 68681, '\P{^oldturkic}', "");
Expect(1, 68680, '\p{  Old_Turkic}', "");
Expect(0, 68680, '\p{^  Old_Turkic}', "");
Expect(0, 68680, '\P{  Old_Turkic}', "");
Expect(1, 68680, '\P{^  Old_Turkic}', "");
Expect(0, 68681, '\p{  Old_Turkic}', "");
Expect(1, 68681, '\p{^  Old_Turkic}', "");
Expect(1, 68681, '\P{  Old_Turkic}', "");
Expect(0, 68681, '\P{^  Old_Turkic}', "");
Error('\p{/a/		IS_OLD_TURKIC}');
Error('\P{/a/		IS_OLD_TURKIC}');
Expect(1, 68680, '\p{isoldturkic}', "");
Expect(0, 68680, '\p{^isoldturkic}', "");
Expect(0, 68680, '\P{isoldturkic}', "");
Expect(1, 68680, '\P{^isoldturkic}', "");
Expect(0, 68681, '\p{isoldturkic}', "");
Expect(1, 68681, '\p{^isoldturkic}', "");
Expect(1, 68681, '\P{isoldturkic}', "");
Expect(0, 68681, '\P{^isoldturkic}', "");
Expect(1, 68680, '\p{_Is_OLD_Turkic}', "");
Expect(0, 68680, '\p{^_Is_OLD_Turkic}', "");
Expect(0, 68680, '\P{_Is_OLD_Turkic}', "");
Expect(1, 68680, '\P{^_Is_OLD_Turkic}', "");
Expect(0, 68681, '\p{_Is_OLD_Turkic}', "");
Expect(1, 68681, '\p{^_Is_OLD_Turkic}', "");
Expect(1, 68681, '\P{_Is_OLD_Turkic}', "");
Expect(0, 68681, '\P{^_Is_OLD_Turkic}', "");
Error('\p{:=	ORKH}');
Error('\P{:=	ORKH}');
Expect(1, 68680, '\p{orkh}', "");
Expect(0, 68680, '\p{^orkh}', "");
Expect(0, 68680, '\P{orkh}', "");
Expect(1, 68680, '\P{^orkh}', "");
Expect(0, 68681, '\p{orkh}', "");
Expect(1, 68681, '\p{^orkh}', "");
Expect(1, 68681, '\P{orkh}', "");
Expect(0, 68681, '\P{^orkh}', "");
Expect(1, 68680, '\p{_-Orkh}', "");
Expect(0, 68680, '\p{^_-Orkh}', "");
Expect(0, 68680, '\P{_-Orkh}', "");
Expect(1, 68680, '\P{^_-Orkh}', "");
Expect(0, 68681, '\p{_-Orkh}', "");
Expect(1, 68681, '\p{^_-Orkh}', "");
Expect(1, 68681, '\P{_-Orkh}', "");
Expect(0, 68681, '\P{^_-Orkh}', "");
Error('\p{-_is_Orkh/a/}');
Error('\P{-_is_Orkh/a/}');
Expect(1, 68680, '\p{isorkh}', "");
Expect(0, 68680, '\p{^isorkh}', "");
Expect(0, 68680, '\P{isorkh}', "");
Expect(1, 68680, '\P{^isorkh}', "");
Expect(0, 68681, '\p{isorkh}', "");
Expect(1, 68681, '\p{^isorkh}', "");
Expect(1, 68681, '\P{isorkh}', "");
Expect(0, 68681, '\P{^isorkh}', "");
Expect(1, 68680, '\p{_ is_Orkh}', "");
Expect(0, 68680, '\p{^_ is_Orkh}', "");
Expect(0, 68680, '\P{_ is_Orkh}', "");
Expect(1, 68680, '\P{^_ is_Orkh}', "");
Expect(0, 68681, '\p{_ is_Orkh}', "");
Expect(1, 68681, '\p{^_ is_Orkh}', "");
Expect(1, 68681, '\P{_ is_Orkh}', "");
Expect(0, 68681, '\P{^_ is_Orkh}', "");
Error('\p{	_Nonspacing_MARK/a/}');
Error('\P{	_Nonspacing_MARK/a/}');
Expect(1, 917999, '\p{nonspacingmark}', "");
Expect(0, 917999, '\p{^nonspacingmark}', "");
Expect(0, 917999, '\P{nonspacingmark}', "");
Expect(1, 917999, '\P{^nonspacingmark}', "");
Expect(0, 918000, '\p{nonspacingmark}', "");
Expect(1, 918000, '\p{^nonspacingmark}', "");
Expect(1, 918000, '\P{nonspacingmark}', "");
Expect(0, 918000, '\P{^nonspacingmark}', "");
Expect(1, 917999, '\p{ nonspacing_Mark}', "");
Expect(0, 917999, '\p{^ nonspacing_Mark}', "");
Expect(0, 917999, '\P{ nonspacing_Mark}', "");
Expect(1, 917999, '\P{^ nonspacing_Mark}', "");
Expect(0, 918000, '\p{ nonspacing_Mark}', "");
Expect(1, 918000, '\p{^ nonspacing_Mark}', "");
Expect(1, 918000, '\P{ nonspacing_Mark}', "");
Expect(0, 918000, '\P{^ nonspacing_Mark}', "");
Error('\p{		Is_NONSPACING_Mark/a/}');
Error('\P{		Is_NONSPACING_Mark/a/}');
Expect(1, 917999, '\p{isnonspacingmark}', "");
Expect(0, 917999, '\p{^isnonspacingmark}', "");
Expect(0, 917999, '\P{isnonspacingmark}', "");
Expect(1, 917999, '\P{^isnonspacingmark}', "");
Expect(0, 918000, '\p{isnonspacingmark}', "");
Expect(1, 918000, '\p{^isnonspacingmark}', "");
Expect(1, 918000, '\P{isnonspacingmark}', "");
Expect(0, 918000, '\P{^isnonspacingmark}', "");
Expect(1, 917999, '\p{ -Is_NONSPACING_mark}', "");
Expect(0, 917999, '\p{^ -Is_NONSPACING_mark}', "");
Expect(0, 917999, '\P{ -Is_NONSPACING_mark}', "");
Expect(1, 917999, '\P{^ -Is_NONSPACING_mark}', "");
Expect(0, 918000, '\p{ -Is_NONSPACING_mark}', "");
Expect(1, 918000, '\p{^ -Is_NONSPACING_mark}', "");
Expect(1, 918000, '\P{ -Is_NONSPACING_mark}', "");
Expect(0, 918000, '\P{^ -Is_NONSPACING_mark}', "");
Error('\p{/a/  Mn}');
Error('\P{/a/  Mn}');
Expect(1, 917999, '\p{mn}', "");
Expect(0, 917999, '\p{^mn}', "");
Expect(0, 917999, '\P{mn}', "");
Expect(1, 917999, '\P{^mn}', "");
Expect(0, 918000, '\p{mn}', "");
Expect(1, 918000, '\p{^mn}', "");
Expect(1, 918000, '\P{mn}', "");
Expect(0, 918000, '\P{^mn}', "");
Expect(1, 917999, '\p{-Mn}', "");
Expect(0, 917999, '\p{^-Mn}', "");
Expect(0, 917999, '\P{-Mn}', "");
Expect(1, 917999, '\P{^-Mn}', "");
Expect(0, 918000, '\p{-Mn}', "");
Expect(1, 918000, '\p{^-Mn}', "");
Expect(1, 918000, '\P{-Mn}', "");
Expect(0, 918000, '\P{^-Mn}', "");
Error('\p{	Is_Mn:=}');
Error('\P{	Is_Mn:=}');
Expect(1, 917999, '\p{ismn}', "");
Expect(0, 917999, '\p{^ismn}', "");
Expect(0, 917999, '\P{ismn}', "");
Expect(1, 917999, '\P{^ismn}', "");
Expect(0, 918000, '\p{ismn}', "");
Expect(1, 918000, '\p{^ismn}', "");
Expect(1, 918000, '\P{ismn}', "");
Expect(0, 918000, '\P{^ismn}', "");
Expect(1, 917999, '\p{ -IS_Mn}', "");
Expect(0, 917999, '\p{^ -IS_Mn}', "");
Expect(0, 917999, '\P{ -IS_Mn}', "");
Expect(1, 917999, '\P{^ -IS_Mn}', "");
Expect(0, 918000, '\p{ -IS_Mn}', "");
Expect(1, 918000, '\p{^ -IS_Mn}', "");
Expect(1, 918000, '\P{ -IS_Mn}', "");
Expect(0, 918000, '\P{^ -IS_Mn}', "");
Error('\p{:=-	Transport_And_MAP_Symbols}');
Error('\P{:=-	Transport_And_MAP_Symbols}');
Expect(1, 128767, '\p{transportandmapsymbols}', "");
Expect(0, 128767, '\p{^transportandmapsymbols}', "");
Expect(0, 128767, '\P{transportandmapsymbols}', "");
Expect(1, 128767, '\P{^transportandmapsymbols}', "");
Expect(0, 128768, '\p{transportandmapsymbols}', "");
Expect(1, 128768, '\p{^transportandmapsymbols}', "");
Expect(1, 128768, '\P{transportandmapsymbols}', "");
Expect(0, 128768, '\P{^transportandmapsymbols}', "");
Expect(1, 128767, '\p{- Transport_and_map_Symbols}', "");
Expect(0, 128767, '\p{^- Transport_and_map_Symbols}', "");
Expect(0, 128767, '\P{- Transport_and_map_Symbols}', "");
Expect(1, 128767, '\P{^- Transport_and_map_Symbols}', "");
Expect(0, 128768, '\p{- Transport_and_map_Symbols}', "");
Expect(1, 128768, '\p{^- Transport_and_map_Symbols}', "");
Expect(1, 128768, '\P{- Transport_and_map_Symbols}', "");
Expect(0, 128768, '\P{^- Transport_and_map_Symbols}', "");
Error('\p{:=	-is_Transport_AND_MAP_symbols}');
Error('\P{:=	-is_Transport_AND_MAP_symbols}');
Expect(1, 128767, '\p{istransportandmapsymbols}', "");
Expect(0, 128767, '\p{^istransportandmapsymbols}', "");
Expect(0, 128767, '\P{istransportandmapsymbols}', "");
Expect(1, 128767, '\P{^istransportandmapsymbols}', "");
Expect(0, 128768, '\p{istransportandmapsymbols}', "");
Expect(1, 128768, '\p{^istransportandmapsymbols}', "");
Expect(1, 128768, '\P{istransportandmapsymbols}', "");
Expect(0, 128768, '\P{^istransportandmapsymbols}', "");
Expect(1, 128767, '\p{_-is_transport_And_map_SYMBOLS}', "");
Expect(0, 128767, '\p{^_-is_transport_And_map_SYMBOLS}', "");
Expect(0, 128767, '\P{_-is_transport_And_map_SYMBOLS}', "");
Expect(1, 128767, '\P{^_-is_transport_And_map_SYMBOLS}', "");
Expect(0, 128768, '\p{_-is_transport_And_map_SYMBOLS}', "");
Expect(1, 128768, '\p{^_-is_transport_And_map_SYMBOLS}', "");
Expect(1, 128768, '\P{_-is_transport_And_map_SYMBOLS}', "");
Expect(0, 128768, '\P{^_-is_transport_And_map_SYMBOLS}', "");
Error('\p{:=In_Transport_And_Map_SYMBOLS}');
Error('\P{:=In_Transport_And_Map_SYMBOLS}');
Expect(1, 128767, '\p{intransportandmapsymbols}', "");
Expect(0, 128767, '\p{^intransportandmapsymbols}', "");
Expect(0, 128767, '\P{intransportandmapsymbols}', "");
Expect(1, 128767, '\P{^intransportandmapsymbols}', "");
Expect(0, 128768, '\p{intransportandmapsymbols}', "");
Expect(1, 128768, '\p{^intransportandmapsymbols}', "");
Expect(1, 128768, '\P{intransportandmapsymbols}', "");
Expect(0, 128768, '\P{^intransportandmapsymbols}', "");
Expect(1, 128767, '\p{	_IN_Transport_AND_Map_symbols}', "");
Expect(0, 128767, '\p{^	_IN_Transport_AND_Map_symbols}', "");
Expect(0, 128767, '\P{	_IN_Transport_AND_Map_symbols}', "");
Expect(1, 128767, '\P{^	_IN_Transport_AND_Map_symbols}', "");
Expect(0, 128768, '\p{	_IN_Transport_AND_Map_symbols}', "");
Expect(1, 128768, '\p{^	_IN_Transport_AND_Map_symbols}', "");
Expect(1, 128768, '\P{	_IN_Transport_AND_Map_symbols}', "");
Expect(0, 128768, '\P{^	_IN_Transport_AND_Map_symbols}', "");
Error('\p{:=Transport_AND_map}');
Error('\P{:=Transport_AND_map}');
Expect(1, 128767, '\p{transportandmap}', "");
Expect(0, 128767, '\p{^transportandmap}', "");
Expect(0, 128767, '\P{transportandmap}', "");
Expect(1, 128767, '\P{^transportandmap}', "");
Expect(0, 128768, '\p{transportandmap}', "");
Expect(1, 128768, '\p{^transportandmap}', "");
Expect(1, 128768, '\P{transportandmap}', "");
Expect(0, 128768, '\P{^transportandmap}', "");
Expect(1, 128767, '\p{ Transport_and_Map}', "");
Expect(0, 128767, '\p{^ Transport_and_Map}', "");
Expect(0, 128767, '\P{ Transport_and_Map}', "");
Expect(1, 128767, '\P{^ Transport_and_Map}', "");
Expect(0, 128768, '\p{ Transport_and_Map}', "");
Expect(1, 128768, '\p{^ Transport_and_Map}', "");
Expect(1, 128768, '\P{ Transport_and_Map}', "");
Expect(0, 128768, '\P{^ Transport_and_Map}', "");
Error('\p{/a/	 Is_Transport_And_Map}');
Error('\P{/a/	 Is_Transport_And_Map}');
Expect(1, 128767, '\p{istransportandmap}', "");
Expect(0, 128767, '\p{^istransportandmap}', "");
Expect(0, 128767, '\P{istransportandmap}', "");
Expect(1, 128767, '\P{^istransportandmap}', "");
Expect(0, 128768, '\p{istransportandmap}', "");
Expect(1, 128768, '\p{^istransportandmap}', "");
Expect(1, 128768, '\P{istransportandmap}', "");
Expect(0, 128768, '\P{^istransportandmap}', "");
Expect(1, 128767, '\p{ 	Is_TRANSPORT_AND_Map}', "");
Expect(0, 128767, '\p{^ 	Is_TRANSPORT_AND_Map}', "");
Expect(0, 128767, '\P{ 	Is_TRANSPORT_AND_Map}', "");
Expect(1, 128767, '\P{^ 	Is_TRANSPORT_AND_Map}', "");
Expect(0, 128768, '\p{ 	Is_TRANSPORT_AND_Map}', "");
Expect(1, 128768, '\p{^ 	Is_TRANSPORT_AND_Map}', "");
Expect(1, 128768, '\P{ 	Is_TRANSPORT_AND_Map}', "");
Expect(0, 128768, '\P{^ 	Is_TRANSPORT_AND_Map}', "");
Error('\p{/a/_IN_Transport_AND_Map}');
Error('\P{/a/_IN_Transport_AND_Map}');
Expect(1, 128767, '\p{intransportandmap}', "");
Expect(0, 128767, '\p{^intransportandmap}', "");
Expect(0, 128767, '\P{intransportandmap}', "");
Expect(1, 128767, '\P{^intransportandmap}', "");
Expect(0, 128768, '\p{intransportandmap}', "");
Expect(1, 128768, '\p{^intransportandmap}', "");
Expect(1, 128768, '\P{intransportandmap}', "");
Expect(0, 128768, '\P{^intransportandmap}', "");
Expect(1, 128767, '\p{		In_transport_And_Map}', "");
Expect(0, 128767, '\p{^		In_transport_And_Map}', "");
Expect(0, 128767, '\P{		In_transport_And_Map}', "");
Expect(1, 128767, '\P{^		In_transport_And_Map}', "");
Expect(0, 128768, '\p{		In_transport_And_Map}', "");
Expect(1, 128768, '\p{^		In_transport_And_Map}', "");
Expect(1, 128768, '\P{		In_transport_And_Map}', "");
Expect(0, 128768, '\P{^		In_transport_And_Map}', "");
Error('\p{:=	LISU}');
Error('\P{:=	LISU}');
Expect(1, 42239, '\p{lisu}', "");
Expect(0, 42239, '\p{^lisu}', "");
Expect(0, 42239, '\P{lisu}', "");
Expect(1, 42239, '\P{^lisu}', "");
Expect(0, 42240, '\p{lisu}', "");
Expect(1, 42240, '\p{^lisu}', "");
Expect(1, 42240, '\P{lisu}', "");
Expect(0, 42240, '\P{^lisu}', "");
Expect(1, 42239, '\p{ Lisu}', "");
Expect(0, 42239, '\p{^ Lisu}', "");
Expect(0, 42239, '\P{ Lisu}', "");
Expect(1, 42239, '\P{^ Lisu}', "");
Expect(0, 42240, '\p{ Lisu}', "");
Expect(1, 42240, '\p{^ Lisu}', "");
Expect(1, 42240, '\P{ Lisu}', "");
Expect(0, 42240, '\P{^ Lisu}', "");
Error('\p{_/a/Is_lisu}');
Error('\P{_/a/Is_lisu}');
Expect(1, 42239, '\p{islisu}', "");
Expect(0, 42239, '\p{^islisu}', "");
Expect(0, 42239, '\P{islisu}', "");
Expect(1, 42239, '\P{^islisu}', "");
Expect(0, 42240, '\p{islisu}', "");
Expect(1, 42240, '\p{^islisu}', "");
Expect(1, 42240, '\P{islisu}', "");
Expect(0, 42240, '\P{^islisu}', "");
Expect(1, 42239, '\p{Is_LISU}', "");
Expect(0, 42239, '\p{^Is_LISU}', "");
Expect(0, 42239, '\P{Is_LISU}', "");
Expect(1, 42239, '\P{^Is_LISU}', "");
Expect(0, 42240, '\p{Is_LISU}', "");
Expect(1, 42240, '\p{^Is_LISU}', "");
Expect(1, 42240, '\P{Is_LISU}', "");
Expect(0, 42240, '\P{^Is_LISU}', "");
Error('\p{	:=Full_Composition_EXCLUSION}');
Error('\P{	:=Full_Composition_EXCLUSION}');
Expect(1, 195101, '\p{fullcompositionexclusion}', "");
Expect(0, 195101, '\p{^fullcompositionexclusion}', "");
Expect(0, 195101, '\P{fullcompositionexclusion}', "");
Expect(1, 195101, '\P{^fullcompositionexclusion}', "");
Expect(0, 195102, '\p{fullcompositionexclusion}', "");
Expect(1, 195102, '\p{^fullcompositionexclusion}', "");
Expect(1, 195102, '\P{fullcompositionexclusion}', "");
Expect(0, 195102, '\P{^fullcompositionexclusion}', "");
Expect(1, 195101, '\p{		FULL_COMPOSITION_Exclusion}', "");
Expect(0, 195101, '\p{^		FULL_COMPOSITION_Exclusion}', "");
Expect(0, 195101, '\P{		FULL_COMPOSITION_Exclusion}', "");
Expect(1, 195101, '\P{^		FULL_COMPOSITION_Exclusion}', "");
Expect(0, 195102, '\p{		FULL_COMPOSITION_Exclusion}', "");
Expect(1, 195102, '\p{^		FULL_COMPOSITION_Exclusion}', "");
Expect(1, 195102, '\P{		FULL_COMPOSITION_Exclusion}', "");
Expect(0, 195102, '\P{^		FULL_COMPOSITION_Exclusion}', "");
Error('\p{  is_full_Composition_exclusion/a/}');
Error('\P{  is_full_Composition_exclusion/a/}');
Expect(1, 195101, '\p{isfullcompositionexclusion}', "");
Expect(0, 195101, '\p{^isfullcompositionexclusion}', "");
Expect(0, 195101, '\P{isfullcompositionexclusion}', "");
Expect(1, 195101, '\P{^isfullcompositionexclusion}', "");
Expect(0, 195102, '\p{isfullcompositionexclusion}', "");
Expect(1, 195102, '\p{^isfullcompositionexclusion}', "");
Expect(1, 195102, '\P{isfullcompositionexclusion}', "");
Expect(0, 195102, '\P{^isfullcompositionexclusion}', "");
Expect(1, 195101, '\p{_ Is_full_Composition_EXCLUSION}', "");
Expect(0, 195101, '\p{^_ Is_full_Composition_EXCLUSION}', "");
Expect(0, 195101, '\P{_ Is_full_Composition_EXCLUSION}', "");
Expect(1, 195101, '\P{^_ Is_full_Composition_EXCLUSION}', "");
Expect(0, 195102, '\p{_ Is_full_Composition_EXCLUSION}', "");
Expect(1, 195102, '\p{^_ Is_full_Composition_EXCLUSION}', "");
Expect(1, 195102, '\P{_ Is_full_Composition_EXCLUSION}', "");
Expect(0, 195102, '\P{^_ Is_full_Composition_EXCLUSION}', "");
Error('\p{:=Comp_Ex}');
Error('\P{:=Comp_Ex}');
Expect(1, 195101, '\p{compex}', "");
Expect(0, 195101, '\p{^compex}', "");
Expect(0, 195101, '\P{compex}', "");
Expect(1, 195101, '\P{^compex}', "");
Expect(0, 195102, '\p{compex}', "");
Expect(1, 195102, '\p{^compex}', "");
Expect(1, 195102, '\P{compex}', "");
Expect(0, 195102, '\P{^compex}', "");
Expect(1, 195101, '\p{	Comp_ex}', "");
Expect(0, 195101, '\p{^	Comp_ex}', "");
Expect(0, 195101, '\P{	Comp_ex}', "");
Expect(1, 195101, '\P{^	Comp_ex}', "");
Expect(0, 195102, '\p{	Comp_ex}', "");
Expect(1, 195102, '\p{^	Comp_ex}', "");
Expect(1, 195102, '\P{	Comp_ex}', "");
Expect(0, 195102, '\P{^	Comp_ex}', "");
Error('\p{:=_is_Comp_ex}');
Error('\P{:=_is_Comp_ex}');
Expect(1, 195101, '\p{iscompex}', "");
Expect(0, 195101, '\p{^iscompex}', "");
Expect(0, 195101, '\P{iscompex}', "");
Expect(1, 195101, '\P{^iscompex}', "");
Expect(0, 195102, '\p{iscompex}', "");
Expect(1, 195102, '\p{^iscompex}', "");
Expect(1, 195102, '\P{iscompex}', "");
Expect(0, 195102, '\P{^iscompex}', "");
Expect(1, 195101, '\p{_	is_COMP_Ex}', "");
Expect(0, 195101, '\p{^_	is_COMP_Ex}', "");
Expect(0, 195101, '\P{_	is_COMP_Ex}', "");
Expect(1, 195101, '\P{^_	is_COMP_Ex}', "");
Expect(0, 195102, '\p{_	is_COMP_Ex}', "");
Expect(1, 195102, '\p{^_	is_COMP_Ex}', "");
Expect(1, 195102, '\P{_	is_COMP_Ex}', "");
Expect(0, 195102, '\P{^_	is_COMP_Ex}', "");
Error('\p{_Join_control/a/}');
Error('\P{_Join_control/a/}');
Expect(1, 8205, '\p{joincontrol}', "");
Expect(0, 8205, '\p{^joincontrol}', "");
Expect(0, 8205, '\P{joincontrol}', "");
Expect(1, 8205, '\P{^joincontrol}', "");
Expect(0, 8206, '\p{joincontrol}', "");
Expect(1, 8206, '\p{^joincontrol}', "");
Expect(1, 8206, '\P{joincontrol}', "");
Expect(0, 8206, '\P{^joincontrol}', "");
Expect(1, 8205, '\p{- JOIN_Control}', "");
Expect(0, 8205, '\p{^- JOIN_Control}', "");
Expect(0, 8205, '\P{- JOIN_Control}', "");
Expect(1, 8205, '\P{^- JOIN_Control}', "");
Expect(0, 8206, '\p{- JOIN_Control}', "");
Expect(1, 8206, '\p{^- JOIN_Control}', "");
Expect(1, 8206, '\P{- JOIN_Control}', "");
Expect(0, 8206, '\P{^- JOIN_Control}', "");
Error('\p{__Is_Join_control:=}');
Error('\P{__Is_Join_control:=}');
Expect(1, 8205, '\p{isjoincontrol}', "");
Expect(0, 8205, '\p{^isjoincontrol}', "");
Expect(0, 8205, '\P{isjoincontrol}', "");
Expect(1, 8205, '\P{^isjoincontrol}', "");
Expect(0, 8206, '\p{isjoincontrol}', "");
Expect(1, 8206, '\p{^isjoincontrol}', "");
Expect(1, 8206, '\P{isjoincontrol}', "");
Expect(0, 8206, '\P{^isjoincontrol}', "");
Expect(1, 8205, '\p{	_Is_Join_Control}', "");
Expect(0, 8205, '\p{^	_Is_Join_Control}', "");
Expect(0, 8205, '\P{	_Is_Join_Control}', "");
Expect(1, 8205, '\P{^	_Is_Join_Control}', "");
Expect(0, 8206, '\p{	_Is_Join_Control}', "");
Expect(1, 8206, '\p{^	_Is_Join_Control}', "");
Expect(1, 8206, '\P{	_Is_Join_Control}', "");
Expect(0, 8206, '\P{^	_Is_Join_Control}', "");
Error('\p{/a/	JOIN_C}');
Error('\P{/a/	JOIN_C}');
Expect(1, 8205, '\p{joinc}', "");
Expect(0, 8205, '\p{^joinc}', "");
Expect(0, 8205, '\P{joinc}', "");
Expect(1, 8205, '\P{^joinc}', "");
Expect(0, 8206, '\p{joinc}', "");
Expect(1, 8206, '\p{^joinc}', "");
Expect(1, 8206, '\P{joinc}', "");
Expect(0, 8206, '\P{^joinc}', "");
Expect(1, 8205, '\p{-Join_c}', "");
Expect(0, 8205, '\p{^-Join_c}', "");
Expect(0, 8205, '\P{-Join_c}', "");
Expect(1, 8205, '\P{^-Join_c}', "");
Expect(0, 8206, '\p{-Join_c}', "");
Expect(1, 8206, '\p{^-Join_c}', "");
Expect(1, 8206, '\P{-Join_c}', "");
Expect(0, 8206, '\P{^-Join_c}', "");
Error('\p{ /a/Is_join_C}');
Error('\P{ /a/Is_join_C}');
Expect(1, 8205, '\p{isjoinc}', "");
Expect(0, 8205, '\p{^isjoinc}', "");
Expect(0, 8205, '\P{isjoinc}', "");
Expect(1, 8205, '\P{^isjoinc}', "");
Expect(0, 8206, '\p{isjoinc}', "");
Expect(1, 8206, '\p{^isjoinc}', "");
Expect(1, 8206, '\P{isjoinc}', "");
Expect(0, 8206, '\P{^isjoinc}', "");
Expect(1, 8205, '\p{_	is_JOIN_C}', "");
Expect(0, 8205, '\p{^_	is_JOIN_C}', "");
Expect(0, 8205, '\P{_	is_JOIN_C}', "");
Expect(1, 8205, '\P{^_	is_JOIN_C}', "");
Expect(0, 8206, '\p{_	is_JOIN_C}', "");
Expect(1, 8206, '\p{^_	is_JOIN_C}', "");
Expect(1, 8206, '\P{_	is_JOIN_C}', "");
Expect(0, 8206, '\P{^_	is_JOIN_C}', "");
Error('\p{:=Upper}');
Error('\P{:=Upper}');
Expect(1, 120778, '\p{upper}', "");
Expect(0, 120778, '\p{^upper}', "");
Expect(0, 120778, '\P{upper}', "");
Expect(1, 120778, '\P{^upper}', "");
Expect(0, 120779, '\p{upper}', "");
Expect(1, 120779, '\p{^upper}', "");
Expect(1, 120779, '\P{upper}', "");
Expect(0, 120779, '\P{^upper}', "");
Expect(1, 120778, '\p{--upper}', "");
Expect(0, 120778, '\p{^--upper}', "");
Expect(0, 120778, '\P{--upper}', "");
Expect(1, 120778, '\P{^--upper}', "");
Expect(0, 120779, '\p{--upper}', "");
Expect(1, 120779, '\p{^--upper}', "");
Expect(1, 120779, '\P{--upper}', "");
Expect(0, 120779, '\P{^--upper}', "");
Error('\p{/a/ _xposixupper}');
Error('\P{/a/ _xposixupper}');
Expect(1, 120778, '\p{xposixupper}', "");
Expect(0, 120778, '\p{^xposixupper}', "");
Expect(0, 120778, '\P{xposixupper}', "");
Expect(1, 120778, '\P{^xposixupper}', "");
Expect(0, 120779, '\p{xposixupper}', "");
Expect(1, 120779, '\p{^xposixupper}', "");
Expect(1, 120779, '\P{xposixupper}', "");
Expect(0, 120779, '\P{^xposixupper}', "");
Expect(1, 120778, '\p{ -XPosixUpper}', "");
Expect(0, 120778, '\p{^ -XPosixUpper}', "");
Expect(0, 120778, '\P{ -XPosixUpper}', "");
Expect(1, 120778, '\P{^ -XPosixUpper}', "");
Expect(0, 120779, '\p{ -XPosixUpper}', "");
Expect(1, 120779, '\p{^ -XPosixUpper}', "");
Expect(1, 120779, '\P{ -XPosixUpper}', "");
Expect(0, 120779, '\P{^ -XPosixUpper}', "");
Error('\p{:=IS_UPPER}');
Error('\P{:=IS_UPPER}');
Expect(1, 120778, '\p{isupper}', "");
Expect(0, 120778, '\p{^isupper}', "");
Expect(0, 120778, '\P{isupper}', "");
Expect(1, 120778, '\P{^isupper}', "");
Expect(0, 120779, '\p{isupper}', "");
Expect(1, 120779, '\p{^isupper}', "");
Expect(1, 120779, '\P{isupper}', "");
Expect(0, 120779, '\P{^isupper}', "");
Expect(1, 120778, '\p{		is_upper}', "");
Expect(0, 120778, '\p{^		is_upper}', "");
Expect(0, 120778, '\P{		is_upper}', "");
Expect(1, 120778, '\P{^		is_upper}', "");
Expect(0, 120779, '\p{		is_upper}', "");
Expect(1, 120779, '\p{^		is_upper}', "");
Expect(1, 120779, '\P{		is_upper}', "");
Expect(0, 120779, '\P{^		is_upper}', "");
Error('\p{	-is_XPosixUpper/a/}');
Error('\P{	-is_XPosixUpper/a/}');
Expect(1, 120778, '\p{isxposixupper}', "");
Expect(0, 120778, '\p{^isxposixupper}', "");
Expect(0, 120778, '\P{isxposixupper}', "");
Expect(1, 120778, '\P{^isxposixupper}', "");
Expect(0, 120779, '\p{isxposixupper}', "");
Expect(1, 120779, '\p{^isxposixupper}', "");
Expect(1, 120779, '\P{isxposixupper}', "");
Expect(0, 120779, '\P{^isxposixupper}', "");
Expect(1, 120778, '\p{ _is_xposixupper}', "");
Expect(0, 120778, '\p{^ _is_xposixupper}', "");
Expect(0, 120778, '\P{ _is_xposixupper}', "");
Expect(1, 120778, '\P{^ _is_xposixupper}', "");
Expect(0, 120779, '\p{ _is_xposixupper}', "");
Expect(1, 120779, '\p{^ _is_xposixupper}', "");
Expect(1, 120779, '\P{ _is_xposixupper}', "");
Expect(0, 120779, '\P{^ _is_xposixupper}', "");
Error('\p{	_UPPERCASE/a/}');
Error('\P{	_UPPERCASE/a/}');
Expect(1, 120778, '\p{uppercase}', "");
Expect(0, 120778, '\p{^uppercase}', "");
Expect(0, 120778, '\P{uppercase}', "");
Expect(1, 120778, '\P{^uppercase}', "");
Expect(0, 120779, '\p{uppercase}', "");
Expect(1, 120779, '\p{^uppercase}', "");
Expect(1, 120779, '\P{uppercase}', "");
Expect(0, 120779, '\P{^uppercase}', "");
Expect(1, 120778, '\p{-	UPPERCASE}', "");
Expect(0, 120778, '\p{^-	UPPERCASE}', "");
Expect(0, 120778, '\P{-	UPPERCASE}', "");
Expect(1, 120778, '\P{^-	UPPERCASE}', "");
Expect(0, 120779, '\p{-	UPPERCASE}', "");
Expect(1, 120779, '\p{^-	UPPERCASE}', "");
Expect(1, 120779, '\P{-	UPPERCASE}', "");
Expect(0, 120779, '\P{^-	UPPERCASE}', "");
Error('\p{-/a/is_Uppercase}');
Error('\P{-/a/is_Uppercase}');
Expect(1, 120778, '\p{isuppercase}', "");
Expect(0, 120778, '\p{^isuppercase}', "");
Expect(0, 120778, '\P{isuppercase}', "");
Expect(1, 120778, '\P{^isuppercase}', "");
Expect(0, 120779, '\p{isuppercase}', "");
Expect(1, 120779, '\p{^isuppercase}', "");
Expect(1, 120779, '\P{isuppercase}', "");
Expect(0, 120779, '\P{^isuppercase}', "");
Expect(1, 120778, '\p{		is_uppercase}', "");
Expect(0, 120778, '\p{^		is_uppercase}', "");
Expect(0, 120778, '\P{		is_uppercase}', "");
Expect(1, 120778, '\P{^		is_uppercase}', "");
Expect(0, 120779, '\p{		is_uppercase}', "");
Expect(1, 120779, '\p{^		is_uppercase}', "");
Expect(1, 120779, '\P{		is_uppercase}', "");
Expect(0, 120779, '\P{^		is_uppercase}', "");
Error('\p{_-Braille_Patterns/a/}');
Error('\P{_-Braille_Patterns/a/}');
Expect(1, 10495, '\p{braillepatterns}', "");
Expect(0, 10495, '\p{^braillepatterns}', "");
Expect(0, 10495, '\P{braillepatterns}', "");
Expect(1, 10495, '\P{^braillepatterns}', "");
Expect(0, 10496, '\p{braillepatterns}', "");
Expect(1, 10496, '\p{^braillepatterns}', "");
Expect(1, 10496, '\P{braillepatterns}', "");
Expect(0, 10496, '\P{^braillepatterns}', "");
Expect(1, 10495, '\p{  Braille_PATTERNS}', "");
Expect(0, 10495, '\p{^  Braille_PATTERNS}', "");
Expect(0, 10495, '\P{  Braille_PATTERNS}', "");
Expect(1, 10495, '\P{^  Braille_PATTERNS}', "");
Expect(0, 10496, '\p{  Braille_PATTERNS}', "");
Expect(1, 10496, '\p{^  Braille_PATTERNS}', "");
Expect(1, 10496, '\P{  Braille_PATTERNS}', "");
Expect(0, 10496, '\P{^  Braille_PATTERNS}', "");
Error('\p{/a/__is_BRAILLE_patterns}');
Error('\P{/a/__is_BRAILLE_patterns}');
Expect(1, 10495, '\p{isbraillepatterns}', "");
Expect(0, 10495, '\p{^isbraillepatterns}', "");
Expect(0, 10495, '\P{isbraillepatterns}', "");
Expect(1, 10495, '\P{^isbraillepatterns}', "");
Expect(0, 10496, '\p{isbraillepatterns}', "");
Expect(1, 10496, '\p{^isbraillepatterns}', "");
Expect(1, 10496, '\P{isbraillepatterns}', "");
Expect(0, 10496, '\P{^isbraillepatterns}', "");
Expect(1, 10495, '\p{__IS_Braille_Patterns}', "");
Expect(0, 10495, '\p{^__IS_Braille_Patterns}', "");
Expect(0, 10495, '\P{__IS_Braille_Patterns}', "");
Expect(1, 10495, '\P{^__IS_Braille_Patterns}', "");
Expect(0, 10496, '\p{__IS_Braille_Patterns}', "");
Expect(1, 10496, '\p{^__IS_Braille_Patterns}', "");
Expect(1, 10496, '\P{__IS_Braille_Patterns}', "");
Expect(0, 10496, '\P{^__IS_Braille_Patterns}', "");
Error('\p{	 in_Braille_Patterns:=}');
Error('\P{	 in_Braille_Patterns:=}');
Expect(1, 10495, '\p{inbraillepatterns}', "");
Expect(0, 10495, '\p{^inbraillepatterns}', "");
Expect(0, 10495, '\P{inbraillepatterns}', "");
Expect(1, 10495, '\P{^inbraillepatterns}', "");
Expect(0, 10496, '\p{inbraillepatterns}', "");
Expect(1, 10496, '\p{^inbraillepatterns}', "");
Expect(1, 10496, '\P{inbraillepatterns}', "");
Expect(0, 10496, '\P{^inbraillepatterns}', "");
Expect(1, 10495, '\p{	in_braille_PATTERNS}', "");
Expect(0, 10495, '\p{^	in_braille_PATTERNS}', "");
Expect(0, 10495, '\P{	in_braille_PATTERNS}', "");
Expect(1, 10495, '\P{^	in_braille_PATTERNS}', "");
Expect(0, 10496, '\p{	in_braille_PATTERNS}', "");
Expect(1, 10496, '\p{^	in_braille_PATTERNS}', "");
Expect(1, 10496, '\P{	in_braille_PATTERNS}', "");
Expect(0, 10496, '\P{^	in_braille_PATTERNS}', "");
Error('\p{/a/ In_BRAILLE}');
Error('\P{/a/ In_BRAILLE}');
Expect(1, 10495, '\p{inbraille}', "");
Expect(0, 10495, '\p{^inbraille}', "");
Expect(0, 10495, '\P{inbraille}', "");
Expect(1, 10495, '\P{^inbraille}', "");
Expect(0, 10496, '\p{inbraille}', "");
Expect(1, 10496, '\p{^inbraille}', "");
Expect(1, 10496, '\P{inbraille}', "");
Expect(0, 10496, '\P{^inbraille}', "");
Expect(1, 10495, '\p{In_braille}', "");
Expect(0, 10495, '\p{^In_braille}', "");
Expect(0, 10495, '\P{In_braille}', "");
Expect(1, 10495, '\P{^In_braille}', "");
Expect(0, 10496, '\p{In_braille}', "");
Expect(1, 10496, '\p{^In_braille}', "");
Expect(1, 10496, '\P{In_braille}', "");
Expect(0, 10496, '\P{^In_braille}', "");
Error('\p{	 Katakana/a/}');
Error('\P{	 Katakana/a/}');
Expect(1, 110592, '\p{katakana}', "");
Expect(0, 110592, '\p{^katakana}', "");
Expect(0, 110592, '\P{katakana}', "");
Expect(1, 110592, '\P{^katakana}', "");
Expect(0, 110593, '\p{katakana}', "");
Expect(1, 110593, '\p{^katakana}', "");
Expect(1, 110593, '\P{katakana}', "");
Expect(0, 110593, '\P{^katakana}', "");
Expect(1, 110592, '\p{ -katakana}', "");
Expect(0, 110592, '\p{^ -katakana}', "");
Expect(0, 110592, '\P{ -katakana}', "");
Expect(1, 110592, '\P{^ -katakana}', "");
Expect(0, 110593, '\p{ -katakana}', "");
Expect(1, 110593, '\p{^ -katakana}', "");
Expect(1, 110593, '\P{ -katakana}', "");
Expect(0, 110593, '\P{^ -katakana}', "");
Error('\p{- Is_katakana/a/}');
Error('\P{- Is_katakana/a/}');
Expect(1, 110592, '\p{iskatakana}', "");
Expect(0, 110592, '\p{^iskatakana}', "");
Expect(0, 110592, '\P{iskatakana}', "");
Expect(1, 110592, '\P{^iskatakana}', "");
Expect(0, 110593, '\p{iskatakana}', "");
Expect(1, 110593, '\p{^iskatakana}', "");
Expect(1, 110593, '\P{iskatakana}', "");
Expect(0, 110593, '\P{^iskatakana}', "");
Expect(1, 110592, '\p{ 	IS_katakana}', "");
Expect(0, 110592, '\p{^ 	IS_katakana}', "");
Expect(0, 110592, '\P{ 	IS_katakana}', "");
Expect(1, 110592, '\P{^ 	IS_katakana}', "");
Expect(0, 110593, '\p{ 	IS_katakana}', "");
Expect(1, 110593, '\p{^ 	IS_katakana}', "");
Expect(1, 110593, '\P{ 	IS_katakana}', "");
Expect(0, 110593, '\P{^ 	IS_katakana}', "");
Error('\p{-/a/Kana}');
Error('\P{-/a/Kana}');
Expect(1, 110592, '\p{kana}', "");
Expect(0, 110592, '\p{^kana}', "");
Expect(0, 110592, '\P{kana}', "");
Expect(1, 110592, '\P{^kana}', "");
Expect(0, 110593, '\p{kana}', "");
Expect(1, 110593, '\p{^kana}', "");
Expect(1, 110593, '\P{kana}', "");
Expect(0, 110593, '\P{^kana}', "");
Expect(1, 110592, '\p{  Kana}', "");
Expect(0, 110592, '\p{^  Kana}', "");
Expect(0, 110592, '\P{  Kana}', "");
Expect(1, 110592, '\P{^  Kana}', "");
Expect(0, 110593, '\p{  Kana}', "");
Expect(1, 110593, '\p{^  Kana}', "");
Expect(1, 110593, '\P{  Kana}', "");
Expect(0, 110593, '\P{^  Kana}', "");
Error('\p{:=-Is_kana}');
Error('\P{:=-Is_kana}');
Expect(1, 110592, '\p{iskana}', "");
Expect(0, 110592, '\p{^iskana}', "");
Expect(0, 110592, '\P{iskana}', "");
Expect(1, 110592, '\P{^iskana}', "");
Expect(0, 110593, '\p{iskana}', "");
Expect(1, 110593, '\p{^iskana}', "");
Expect(1, 110593, '\P{iskana}', "");
Expect(0, 110593, '\P{^iskana}', "");
Expect(1, 110592, '\p{	-Is_Kana}', "");
Expect(0, 110592, '\p{^	-Is_Kana}', "");
Expect(0, 110592, '\P{	-Is_Kana}', "");
Expect(1, 110592, '\P{^	-Is_Kana}', "");
Expect(0, 110593, '\p{	-Is_Kana}', "");
Expect(1, 110593, '\p{^	-Is_Kana}', "");
Expect(1, 110593, '\P{	-Is_Kana}', "");
Expect(0, 110593, '\P{^	-Is_Kana}', "");
Error('\p{_:=pattern_White_space}');
Error('\P{_:=pattern_White_space}');
Expect(1, 8233, '\p{patternwhitespace}', "");
Expect(0, 8233, '\p{^patternwhitespace}', "");
Expect(0, 8233, '\P{patternwhitespace}', "");
Expect(1, 8233, '\P{^patternwhitespace}', "");
Expect(0, 8234, '\p{patternwhitespace}', "");
Expect(1, 8234, '\p{^patternwhitespace}', "");
Expect(1, 8234, '\P{patternwhitespace}', "");
Expect(0, 8234, '\P{^patternwhitespace}', "");
Expect(1, 8233, '\p{--PATTERN_WHITE_SPACE}', "");
Expect(0, 8233, '\p{^--PATTERN_WHITE_SPACE}', "");
Expect(0, 8233, '\P{--PATTERN_WHITE_SPACE}', "");
Expect(1, 8233, '\P{^--PATTERN_WHITE_SPACE}', "");
Expect(0, 8234, '\p{--PATTERN_WHITE_SPACE}', "");
Expect(1, 8234, '\p{^--PATTERN_WHITE_SPACE}', "");
Expect(1, 8234, '\P{--PATTERN_WHITE_SPACE}', "");
Expect(0, 8234, '\P{^--PATTERN_WHITE_SPACE}', "");
Error('\p{-Is_pattern_White_Space/a/}');
Error('\P{-Is_pattern_White_Space/a/}');
Expect(1, 8233, '\p{ispatternwhitespace}', "");
Expect(0, 8233, '\p{^ispatternwhitespace}', "");
Expect(0, 8233, '\P{ispatternwhitespace}', "");
Expect(1, 8233, '\P{^ispatternwhitespace}', "");
Expect(0, 8234, '\p{ispatternwhitespace}', "");
Expect(1, 8234, '\p{^ispatternwhitespace}', "");
Expect(1, 8234, '\P{ispatternwhitespace}', "");
Expect(0, 8234, '\P{^ispatternwhitespace}', "");
Expect(1, 8233, '\p{	-Is_pattern_white_Space}', "");
Expect(0, 8233, '\p{^	-Is_pattern_white_Space}', "");
Expect(0, 8233, '\P{	-Is_pattern_white_Space}', "");
Expect(1, 8233, '\P{^	-Is_pattern_white_Space}', "");
Expect(0, 8234, '\p{	-Is_pattern_white_Space}', "");
Expect(1, 8234, '\p{^	-Is_pattern_white_Space}', "");
Expect(1, 8234, '\P{	-Is_pattern_white_Space}', "");
Expect(0, 8234, '\P{^	-Is_pattern_white_Space}', "");
Error('\p{_ Pat_WS:=}');
Error('\P{_ Pat_WS:=}');
Expect(1, 8233, '\p{patws}', "");
Expect(0, 8233, '\p{^patws}', "");
Expect(0, 8233, '\P{patws}', "");
Expect(1, 8233, '\P{^patws}', "");
Expect(0, 8234, '\p{patws}', "");
Expect(1, 8234, '\p{^patws}', "");
Expect(1, 8234, '\P{patws}', "");
Expect(0, 8234, '\P{^patws}', "");
Expect(1, 8233, '\p{_ Pat_WS}', "");
Expect(0, 8233, '\p{^_ Pat_WS}', "");
Expect(0, 8233, '\P{_ Pat_WS}', "");
Expect(1, 8233, '\P{^_ Pat_WS}', "");
Expect(0, 8234, '\p{_ Pat_WS}', "");
Expect(1, 8234, '\p{^_ Pat_WS}', "");
Expect(1, 8234, '\P{_ Pat_WS}', "");
Expect(0, 8234, '\P{^_ Pat_WS}', "");
Error('\p{_:=is_PAT_WS}');
Error('\P{_:=is_PAT_WS}');
Expect(1, 8233, '\p{ispatws}', "");
Expect(0, 8233, '\p{^ispatws}', "");
Expect(0, 8233, '\P{ispatws}', "");
Expect(1, 8233, '\P{^ispatws}', "");
Expect(0, 8234, '\p{ispatws}', "");
Expect(1, 8234, '\p{^ispatws}', "");
Expect(1, 8234, '\P{ispatws}', "");
Expect(0, 8234, '\P{^ispatws}', "");
Expect(1, 8233, '\p{-Is_Pat_ws}', "");
Expect(0, 8233, '\p{^-Is_Pat_ws}', "");
Expect(0, 8233, '\P{-Is_Pat_ws}', "");
Expect(1, 8233, '\P{^-Is_Pat_ws}', "");
Expect(0, 8234, '\p{-Is_Pat_ws}', "");
Expect(1, 8234, '\p{^-Is_Pat_ws}', "");
Expect(1, 8234, '\P{-Is_Pat_ws}', "");
Expect(0, 8234, '\P{^-Is_Pat_ws}', "");
Error('\p{ noncharacter_CODE_Point:=}');
Error('\P{ noncharacter_CODE_Point:=}');
Expect(1, 1114111, '\p{noncharactercodepoint}', "");
Expect(0, 1114111, '\p{^noncharactercodepoint}', "");
Expect(0, 1114111, '\P{noncharactercodepoint}', "");
Expect(1, 1114111, '\P{^noncharactercodepoint}', "");
Expect(0, 1114109, '\p{noncharactercodepoint}', "");
Expect(1, 1114109, '\p{^noncharactercodepoint}', "");
Expect(1, 1114109, '\P{noncharactercodepoint}', "");
Expect(0, 1114109, '\P{^noncharactercodepoint}', "");
Expect(1, 1114111, '\p{--NONCHARACTER_Code_Point}', "");
Expect(0, 1114111, '\p{^--NONCHARACTER_Code_Point}', "");
Expect(0, 1114111, '\P{--NONCHARACTER_Code_Point}', "");
Expect(1, 1114111, '\P{^--NONCHARACTER_Code_Point}', "");
Expect(0, 1114109, '\p{--NONCHARACTER_Code_Point}', "");
Expect(1, 1114109, '\p{^--NONCHARACTER_Code_Point}', "");
Expect(1, 1114109, '\P{--NONCHARACTER_Code_Point}', "");
Expect(0, 1114109, '\P{^--NONCHARACTER_Code_Point}', "");
Error('\p{/a/_Is_Noncharacter_Code_point}');
Error('\P{/a/_Is_Noncharacter_Code_point}');
Expect(1, 1114111, '\p{isnoncharactercodepoint}', "");
Expect(0, 1114111, '\p{^isnoncharactercodepoint}', "");
Expect(0, 1114111, '\P{isnoncharactercodepoint}', "");
Expect(1, 1114111, '\P{^isnoncharactercodepoint}', "");
Expect(0, 1114109, '\p{isnoncharactercodepoint}', "");
Expect(1, 1114109, '\p{^isnoncharactercodepoint}', "");
Expect(1, 1114109, '\P{isnoncharactercodepoint}', "");
Expect(0, 1114109, '\P{^isnoncharactercodepoint}', "");
Expect(1, 1114111, '\p{  Is_Noncharacter_Code_Point}', "");
Expect(0, 1114111, '\p{^  Is_Noncharacter_Code_Point}', "");
Expect(0, 1114111, '\P{  Is_Noncharacter_Code_Point}', "");
Expect(1, 1114111, '\P{^  Is_Noncharacter_Code_Point}', "");
Expect(0, 1114109, '\p{  Is_Noncharacter_Code_Point}', "");
Expect(1, 1114109, '\p{^  Is_Noncharacter_Code_Point}', "");
Expect(1, 1114109, '\P{  Is_Noncharacter_Code_Point}', "");
Expect(0, 1114109, '\P{^  Is_Noncharacter_Code_Point}', "");
Error('\p{-/a/NCHAR}');
Error('\P{-/a/NCHAR}');
Expect(1, 1114111, '\p{nchar}', "");
Expect(0, 1114111, '\p{^nchar}', "");
Expect(0, 1114111, '\P{nchar}', "");
Expect(1, 1114111, '\P{^nchar}', "");
Expect(0, 1114109, '\p{nchar}', "");
Expect(1, 1114109, '\p{^nchar}', "");
Expect(1, 1114109, '\P{nchar}', "");
Expect(0, 1114109, '\P{^nchar}', "");
Expect(1, 1114111, '\p{-nchar}', "");
Expect(0, 1114111, '\p{^-nchar}', "");
Expect(0, 1114111, '\P{-nchar}', "");
Expect(1, 1114111, '\P{^-nchar}', "");
Expect(0, 1114109, '\p{-nchar}', "");
Expect(1, 1114109, '\p{^-nchar}', "");
Expect(1, 1114109, '\P{-nchar}', "");
Expect(0, 1114109, '\P{^-nchar}', "");
Error('\p{/a/	Is_NCHAR}');
Error('\P{/a/	Is_NCHAR}');
Expect(1, 1114111, '\p{isnchar}', "");
Expect(0, 1114111, '\p{^isnchar}', "");
Expect(0, 1114111, '\P{isnchar}', "");
Expect(1, 1114111, '\P{^isnchar}', "");
Expect(0, 1114109, '\p{isnchar}', "");
Expect(1, 1114109, '\p{^isnchar}', "");
Expect(1, 1114109, '\P{isnchar}', "");
Expect(0, 1114109, '\P{^isnchar}', "");
Expect(1, 1114111, '\p{_-Is_NCHAR}', "");
Expect(0, 1114111, '\p{^_-Is_NCHAR}', "");
Expect(0, 1114111, '\P{_-Is_NCHAR}', "");
Expect(1, 1114111, '\P{^_-Is_NCHAR}', "");
Expect(0, 1114109, '\p{_-Is_NCHAR}', "");
Expect(1, 1114109, '\p{^_-Is_NCHAR}', "");
Expect(1, 1114109, '\P{_-Is_NCHAR}', "");
Expect(0, 1114109, '\P{^_-Is_NCHAR}', "");
Error('\p{/a/-_In_Ogham}');
Error('\P{/a/-_In_Ogham}');
Expect(1, 5791, '\p{inogham}', "");
Expect(0, 5791, '\p{^inogham}', "");
Expect(0, 5791, '\P{inogham}', "");
Expect(1, 5791, '\P{^inogham}', "");
Expect(0, 5792, '\p{inogham}', "");
Expect(1, 5792, '\p{^inogham}', "");
Expect(1, 5792, '\P{inogham}', "");
Expect(0, 5792, '\P{^inogham}', "");
Expect(1, 5791, '\p{	IN_Ogham}', "");
Expect(0, 5791, '\p{^	IN_Ogham}', "");
Expect(0, 5791, '\P{	IN_Ogham}', "");
Expect(1, 5791, '\P{^	IN_Ogham}', "");
Expect(0, 5792, '\p{	IN_Ogham}', "");
Expect(1, 5792, '\p{^	IN_Ogham}', "");
Expect(1, 5792, '\P{	IN_Ogham}', "");
Expect(0, 5792, '\P{^	IN_Ogham}', "");
Error('\p{_/a/IN_Ogham}');
Error('\P{_/a/IN_Ogham}');
Expect(1, 5791, '\p{_in_OGHAM}', "");
Expect(0, 5791, '\p{^_in_OGHAM}', "");
Expect(0, 5791, '\P{_in_OGHAM}', "");
Expect(1, 5791, '\P{^_in_OGHAM}', "");
Expect(0, 5792, '\p{_in_OGHAM}', "");
Expect(1, 5792, '\p{^_in_OGHAM}', "");
Expect(1, 5792, '\P{_in_OGHAM}', "");
Expect(0, 5792, '\P{^_in_OGHAM}', "");
Error('\p{ /a/MIAO}');
Error('\P{ /a/MIAO}');
Expect(1, 94111, '\p{miao}', "");
Expect(0, 94111, '\p{^miao}', "");
Expect(0, 94111, '\P{miao}', "");
Expect(1, 94111, '\P{^miao}', "");
Expect(0, 94112, '\p{miao}', "");
Expect(1, 94112, '\p{^miao}', "");
Expect(1, 94112, '\P{miao}', "");
Expect(0, 94112, '\P{^miao}', "");
Expect(1, 94111, '\p{_-miao}', "");
Expect(0, 94111, '\p{^_-miao}', "");
Expect(0, 94111, '\P{_-miao}', "");
Expect(1, 94111, '\P{^_-miao}', "");
Expect(0, 94112, '\p{_-miao}', "");
Expect(1, 94112, '\p{^_-miao}', "");
Expect(1, 94112, '\P{_-miao}', "");
Expect(0, 94112, '\P{^_-miao}', "");
Error('\p{/a/	-Is_MIAO}');
Error('\P{/a/	-Is_MIAO}');
Expect(1, 94111, '\p{ismiao}', "");
Expect(0, 94111, '\p{^ismiao}', "");
Expect(0, 94111, '\P{ismiao}', "");
Expect(1, 94111, '\P{^ismiao}', "");
Expect(0, 94112, '\p{ismiao}', "");
Expect(1, 94112, '\p{^ismiao}', "");
Expect(1, 94112, '\P{ismiao}', "");
Expect(0, 94112, '\P{^ismiao}', "");
Expect(1, 94111, '\p{ IS_miao}', "");
Expect(0, 94111, '\p{^ IS_miao}', "");
Expect(0, 94111, '\P{ IS_miao}', "");
Expect(1, 94111, '\P{^ IS_miao}', "");
Expect(0, 94112, '\p{ IS_miao}', "");
Expect(1, 94112, '\p{^ IS_miao}', "");
Expect(1, 94112, '\P{ IS_miao}', "");
Expect(0, 94112, '\P{^ IS_miao}', "");
Error('\p{/a/ _plrd}');
Error('\P{/a/ _plrd}');
Expect(1, 94111, '\p{plrd}', "");
Expect(0, 94111, '\p{^plrd}', "");
Expect(0, 94111, '\P{plrd}', "");
Expect(1, 94111, '\P{^plrd}', "");
Expect(0, 94112, '\p{plrd}', "");
Expect(1, 94112, '\p{^plrd}', "");
Expect(1, 94112, '\P{plrd}', "");
Expect(0, 94112, '\P{^plrd}', "");
Expect(1, 94111, '\p{PLRD}', "");
Expect(0, 94111, '\p{^PLRD}', "");
Expect(0, 94111, '\P{PLRD}', "");
Expect(1, 94111, '\P{^PLRD}', "");
Expect(0, 94112, '\p{PLRD}', "");
Expect(1, 94112, '\p{^PLRD}', "");
Expect(1, 94112, '\P{PLRD}', "");
Expect(0, 94112, '\P{^PLRD}', "");
Error('\p{ _IS_PLRD/a/}');
Error('\P{ _IS_PLRD/a/}');
Expect(1, 94111, '\p{isplrd}', "");
Expect(0, 94111, '\p{^isplrd}', "");
Expect(0, 94111, '\P{isplrd}', "");
Expect(1, 94111, '\P{^isplrd}', "");
Expect(0, 94112, '\p{isplrd}', "");
Expect(1, 94112, '\p{^isplrd}', "");
Expect(1, 94112, '\P{isplrd}', "");
Expect(0, 94112, '\P{^isplrd}', "");
Expect(1, 94111, '\p{_	Is_PLRD}', "");
Expect(0, 94111, '\p{^_	Is_PLRD}', "");
Expect(0, 94111, '\P{_	Is_PLRD}', "");
Expect(1, 94111, '\P{^_	Is_PLRD}', "");
Expect(0, 94112, '\p{_	Is_PLRD}', "");
Expect(1, 94112, '\p{^_	Is_PLRD}', "");
Expect(1, 94112, '\P{_	Is_PLRD}', "");
Expect(0, 94112, '\P{^_	Is_PLRD}', "");
Error('\p{_:=phonetic_Extensions_SUPPLEMENT}');
Error('\P{_:=phonetic_Extensions_SUPPLEMENT}');
Expect(1, 7615, '\p{phoneticextensionssupplement}', "");
Expect(0, 7615, '\p{^phoneticextensionssupplement}', "");
Expect(0, 7615, '\P{phoneticextensionssupplement}', "");
Expect(1, 7615, '\P{^phoneticextensionssupplement}', "");
Expect(0, 7616, '\p{phoneticextensionssupplement}', "");
Expect(1, 7616, '\p{^phoneticextensionssupplement}', "");
Expect(1, 7616, '\P{phoneticextensionssupplement}', "");
Expect(0, 7616, '\P{^phoneticextensionssupplement}', "");
Expect(1, 7615, '\p{- PHONETIC_EXTENSIONS_SUPPLEMENT}', "");
Expect(0, 7615, '\p{^- PHONETIC_EXTENSIONS_SUPPLEMENT}', "");
Expect(0, 7615, '\P{- PHONETIC_EXTENSIONS_SUPPLEMENT}', "");
Expect(1, 7615, '\P{^- PHONETIC_EXTENSIONS_SUPPLEMENT}', "");
Expect(0, 7616, '\p{- PHONETIC_EXTENSIONS_SUPPLEMENT}', "");
Expect(1, 7616, '\p{^- PHONETIC_EXTENSIONS_SUPPLEMENT}', "");
Expect(1, 7616, '\P{- PHONETIC_EXTENSIONS_SUPPLEMENT}', "");
Expect(0, 7616, '\P{^- PHONETIC_EXTENSIONS_SUPPLEMENT}', "");
Error('\p{	/a/Is_PHONETIC_Extensions_supplement}');
Error('\P{	/a/Is_PHONETIC_Extensions_supplement}');
Expect(1, 7615, '\p{isphoneticextensionssupplement}', "");
Expect(0, 7615, '\p{^isphoneticextensionssupplement}', "");
Expect(0, 7615, '\P{isphoneticextensionssupplement}', "");
Expect(1, 7615, '\P{^isphoneticextensionssupplement}', "");
Expect(0, 7616, '\p{isphoneticextensionssupplement}', "");
Expect(1, 7616, '\p{^isphoneticextensionssupplement}', "");
Expect(1, 7616, '\P{isphoneticextensionssupplement}', "");
Expect(0, 7616, '\P{^isphoneticextensionssupplement}', "");
Expect(1, 7615, '\p{-is_phonetic_EXTENSIONS_Supplement}', "");
Expect(0, 7615, '\p{^-is_phonetic_EXTENSIONS_Supplement}', "");
Expect(0, 7615, '\P{-is_phonetic_EXTENSIONS_Supplement}', "");
Expect(1, 7615, '\P{^-is_phonetic_EXTENSIONS_Supplement}', "");
Expect(0, 7616, '\p{-is_phonetic_EXTENSIONS_Supplement}', "");
Expect(1, 7616, '\p{^-is_phonetic_EXTENSIONS_Supplement}', "");
Expect(1, 7616, '\P{-is_phonetic_EXTENSIONS_Supplement}', "");
Expect(0, 7616, '\P{^-is_phonetic_EXTENSIONS_Supplement}', "");
Error('\p{ _In_phonetic_extensions_Supplement/a/}');
Error('\P{ _In_phonetic_extensions_Supplement/a/}');
Expect(1, 7615, '\p{inphoneticextensionssupplement}', "");
Expect(0, 7615, '\p{^inphoneticextensionssupplement}', "");
Expect(0, 7615, '\P{inphoneticextensionssupplement}', "");
Expect(1, 7615, '\P{^inphoneticextensionssupplement}', "");
Expect(0, 7616, '\p{inphoneticextensionssupplement}', "");
Expect(1, 7616, '\p{^inphoneticextensionssupplement}', "");
Expect(1, 7616, '\P{inphoneticextensionssupplement}', "");
Expect(0, 7616, '\P{^inphoneticextensionssupplement}', "");
Expect(1, 7615, '\p{ IN_PHONETIC_EXTENSIONS_supplement}', "");
Expect(0, 7615, '\p{^ IN_PHONETIC_EXTENSIONS_supplement}', "");
Expect(0, 7615, '\P{ IN_PHONETIC_EXTENSIONS_supplement}', "");
Expect(1, 7615, '\P{^ IN_PHONETIC_EXTENSIONS_supplement}', "");
Expect(0, 7616, '\p{ IN_PHONETIC_EXTENSIONS_supplement}', "");
Expect(1, 7616, '\p{^ IN_PHONETIC_EXTENSIONS_supplement}', "");
Expect(1, 7616, '\P{ IN_PHONETIC_EXTENSIONS_supplement}', "");
Expect(0, 7616, '\P{^ IN_PHONETIC_EXTENSIONS_supplement}', "");
Error('\p{:=-_PHONETIC_Ext_sup}');
Error('\P{:=-_PHONETIC_Ext_sup}');
Expect(1, 7615, '\p{phoneticextsup}', "");
Expect(0, 7615, '\p{^phoneticextsup}', "");
Expect(0, 7615, '\P{phoneticextsup}', "");
Expect(1, 7615, '\P{^phoneticextsup}', "");
Expect(0, 7616, '\p{phoneticextsup}', "");
Expect(1, 7616, '\p{^phoneticextsup}', "");
Expect(1, 7616, '\P{phoneticextsup}', "");
Expect(0, 7616, '\P{^phoneticextsup}', "");
Expect(1, 7615, '\p{	 Phonetic_Ext_SUP}', "");
Expect(0, 7615, '\p{^	 Phonetic_Ext_SUP}', "");
Expect(0, 7615, '\P{	 Phonetic_Ext_SUP}', "");
Expect(1, 7615, '\P{^	 Phonetic_Ext_SUP}', "");
Expect(0, 7616, '\p{	 Phonetic_Ext_SUP}', "");
Expect(1, 7616, '\p{^	 Phonetic_Ext_SUP}', "");
Expect(1, 7616, '\P{	 Phonetic_Ext_SUP}', "");
Expect(0, 7616, '\P{^	 Phonetic_Ext_SUP}', "");
Error('\p{:=-	Is_Phonetic_Ext_SUP}');
Error('\P{:=-	Is_Phonetic_Ext_SUP}');
Expect(1, 7615, '\p{isphoneticextsup}', "");
Expect(0, 7615, '\p{^isphoneticextsup}', "");
Expect(0, 7615, '\P{isphoneticextsup}', "");
Expect(1, 7615, '\P{^isphoneticextsup}', "");
Expect(0, 7616, '\p{isphoneticextsup}', "");
Expect(1, 7616, '\p{^isphoneticextsup}', "");
Expect(1, 7616, '\P{isphoneticextsup}', "");
Expect(0, 7616, '\P{^isphoneticextsup}', "");
Expect(1, 7615, '\p{- is_Phonetic_Ext_sup}', "");
Expect(0, 7615, '\p{^- is_Phonetic_Ext_sup}', "");
Expect(0, 7615, '\P{- is_Phonetic_Ext_sup}', "");
Expect(1, 7615, '\P{^- is_Phonetic_Ext_sup}', "");
Expect(0, 7616, '\p{- is_Phonetic_Ext_sup}', "");
Expect(1, 7616, '\p{^- is_Phonetic_Ext_sup}', "");
Expect(1, 7616, '\P{- is_Phonetic_Ext_sup}', "");
Expect(0, 7616, '\P{^- is_Phonetic_Ext_sup}', "");
Error('\p{ 	In_phonetic_Ext_SUP/a/}');
Error('\P{ 	In_phonetic_Ext_SUP/a/}');
Expect(1, 7615, '\p{inphoneticextsup}', "");
Expect(0, 7615, '\p{^inphoneticextsup}', "");
Expect(0, 7615, '\P{inphoneticextsup}', "");
Expect(1, 7615, '\P{^inphoneticextsup}', "");
Expect(0, 7616, '\p{inphoneticextsup}', "");
Expect(1, 7616, '\p{^inphoneticextsup}', "");
Expect(1, 7616, '\P{inphoneticextsup}', "");
Expect(0, 7616, '\P{^inphoneticextsup}', "");
Expect(1, 7615, '\p{	In_phonetic_Ext_sup}', "");
Expect(0, 7615, '\p{^	In_phonetic_Ext_sup}', "");
Expect(0, 7615, '\P{	In_phonetic_Ext_sup}', "");
Expect(1, 7615, '\P{^	In_phonetic_Ext_sup}', "");
Expect(0, 7616, '\p{	In_phonetic_Ext_sup}', "");
Expect(1, 7616, '\p{^	In_phonetic_Ext_sup}', "");
Expect(1, 7616, '\P{	In_phonetic_Ext_sup}', "");
Expect(0, 7616, '\P{^	In_phonetic_Ext_sup}', "");
Error('\p{	:=word}');
Error('\P{	:=word}');
Expect(1, 917999, '\p{word}', "");
Expect(0, 917999, '\p{^word}', "");
Expect(0, 917999, '\P{word}', "");
Expect(1, 917999, '\P{^word}', "");
Expect(0, 918000, '\p{word}', "");
Expect(1, 918000, '\p{^word}', "");
Expect(1, 918000, '\P{word}', "");
Expect(0, 918000, '\P{^word}', "");
Expect(1, 917999, '\p{ Word}', "");
Expect(0, 917999, '\p{^ Word}', "");
Expect(0, 917999, '\P{ Word}', "");
Expect(1, 917999, '\P{^ Word}', "");
Expect(0, 918000, '\p{ Word}', "");
Expect(1, 918000, '\p{^ Word}', "");
Expect(1, 918000, '\P{ Word}', "");
Expect(0, 918000, '\P{^ Word}', "");
Error('\p{-:=XPosixWord}');
Error('\P{-:=XPosixWord}');
Expect(1, 917999, '\p{xposixword}', "");
Expect(0, 917999, '\p{^xposixword}', "");
Expect(0, 917999, '\P{xposixword}', "");
Expect(1, 917999, '\P{^xposixword}', "");
Expect(0, 918000, '\p{xposixword}', "");
Expect(1, 918000, '\p{^xposixword}', "");
Expect(1, 918000, '\P{xposixword}', "");
Expect(0, 918000, '\P{^xposixword}', "");
Error('\p{-Is_Word:=}');
Error('\P{-Is_Word:=}');
Expect(1, 917999, '\p{isword}', "");
Expect(0, 917999, '\p{^isword}', "");
Expect(0, 917999, '\P{isword}', "");
Expect(1, 917999, '\P{^isword}', "");
Expect(0, 918000, '\p{isword}', "");
Expect(1, 918000, '\p{^isword}', "");
Expect(1, 918000, '\P{isword}', "");
Expect(0, 918000, '\P{^isword}', "");
Expect(1, 917999, '\p{_Is_WORD}', "");
Expect(0, 917999, '\p{^_Is_WORD}', "");
Expect(0, 917999, '\P{_Is_WORD}', "");
Expect(1, 917999, '\P{^_Is_WORD}', "");
Expect(0, 918000, '\p{_Is_WORD}', "");
Expect(1, 918000, '\p{^_Is_WORD}', "");
Expect(1, 918000, '\P{_Is_WORD}', "");
Expect(0, 918000, '\P{^_Is_WORD}', "");
Error('\p{		is_XPOSIXWORD/a/}');
Error('\P{		is_XPOSIXWORD/a/}');
Expect(1, 917999, '\p{isxposixword}', "");
Expect(0, 917999, '\p{^isxposixword}', "");
Expect(0, 917999, '\P{isxposixword}', "");
Expect(1, 917999, '\P{^isxposixword}', "");
Expect(0, 918000, '\p{isxposixword}', "");
Expect(1, 918000, '\p{^isxposixword}', "");
Expect(1, 918000, '\P{isxposixword}', "");
Expect(0, 918000, '\P{^isxposixword}', "");
Expect(1, 917999, '\p{ -IS_xposixword}', "");
Expect(0, 917999, '\p{^ -IS_xposixword}', "");
Expect(0, 917999, '\P{ -IS_xposixword}', "");
Expect(1, 917999, '\P{^ -IS_xposixword}', "");
Expect(0, 918000, '\p{ -IS_xposixword}', "");
Expect(1, 918000, '\p{^ -IS_xposixword}', "");
Expect(1, 918000, '\P{ -IS_xposixword}', "");
Expect(0, 918000, '\P{^ -IS_xposixword}', "");
Error('\p{	_Old_South_Arabian:=}');
Error('\P{	_Old_South_Arabian:=}');
Expect(1, 68223, '\p{oldsoutharabian}', "");
Expect(0, 68223, '\p{^oldsoutharabian}', "");
Expect(0, 68223, '\P{oldsoutharabian}', "");
Expect(1, 68223, '\P{^oldsoutharabian}', "");
Expect(0, 68224, '\p{oldsoutharabian}', "");
Expect(1, 68224, '\p{^oldsoutharabian}', "");
Expect(1, 68224, '\P{oldsoutharabian}', "");
Expect(0, 68224, '\P{^oldsoutharabian}', "");
Expect(1, 68223, '\p{-	OLD_South_arabian}', "");
Expect(0, 68223, '\p{^-	OLD_South_arabian}', "");
Expect(0, 68223, '\P{-	OLD_South_arabian}', "");
Expect(1, 68223, '\P{^-	OLD_South_arabian}', "");
Expect(0, 68224, '\p{-	OLD_South_arabian}', "");
Expect(1, 68224, '\p{^-	OLD_South_arabian}', "");
Expect(1, 68224, '\P{-	OLD_South_arabian}', "");
Expect(0, 68224, '\P{^-	OLD_South_arabian}', "");
Error('\p{/a/is_old_South_ARABIAN}');
Error('\P{/a/is_old_South_ARABIAN}');
Expect(1, 68223, '\p{isoldsoutharabian}', "");
Expect(0, 68223, '\p{^isoldsoutharabian}', "");
Expect(0, 68223, '\P{isoldsoutharabian}', "");
Expect(1, 68223, '\P{^isoldsoutharabian}', "");
Expect(0, 68224, '\p{isoldsoutharabian}', "");
Expect(1, 68224, '\p{^isoldsoutharabian}', "");
Expect(1, 68224, '\P{isoldsoutharabian}', "");
Expect(0, 68224, '\P{^isoldsoutharabian}', "");
Expect(1, 68223, '\p{	 IS_Old_SOUTH_Arabian}', "");
Expect(0, 68223, '\p{^	 IS_Old_SOUTH_Arabian}', "");
Expect(0, 68223, '\P{	 IS_Old_SOUTH_Arabian}', "");
Expect(1, 68223, '\P{^	 IS_Old_SOUTH_Arabian}', "");
Expect(0, 68224, '\p{	 IS_Old_SOUTH_Arabian}', "");
Expect(1, 68224, '\p{^	 IS_Old_SOUTH_Arabian}', "");
Expect(1, 68224, '\P{	 IS_Old_SOUTH_Arabian}', "");
Expect(0, 68224, '\P{^	 IS_Old_SOUTH_Arabian}', "");
Error('\p{:=Sarb}');
Error('\P{:=Sarb}');
Expect(1, 68223, '\p{sarb}', "");
Expect(0, 68223, '\p{^sarb}', "");
Expect(0, 68223, '\P{sarb}', "");
Expect(1, 68223, '\P{^sarb}', "");
Expect(0, 68224, '\p{sarb}', "");
Expect(1, 68224, '\p{^sarb}', "");
Expect(1, 68224, '\P{sarb}', "");
Expect(0, 68224, '\P{^sarb}', "");
Expect(1, 68223, '\p{_ SARB}', "");
Expect(0, 68223, '\p{^_ SARB}', "");
Expect(0, 68223, '\P{_ SARB}', "");
Expect(1, 68223, '\P{^_ SARB}', "");
Expect(0, 68224, '\p{_ SARB}', "");
Expect(1, 68224, '\p{^_ SARB}', "");
Expect(1, 68224, '\P{_ SARB}', "");
Expect(0, 68224, '\P{^_ SARB}', "");
Error('\p{/a/ 	Is_sarb}');
Error('\P{/a/ 	Is_sarb}');
Expect(1, 68223, '\p{issarb}', "");
Expect(0, 68223, '\p{^issarb}', "");
Expect(0, 68223, '\P{issarb}', "");
Expect(1, 68223, '\P{^issarb}', "");
Expect(0, 68224, '\p{issarb}', "");
Expect(1, 68224, '\p{^issarb}', "");
Expect(1, 68224, '\P{issarb}', "");
Expect(0, 68224, '\P{^issarb}', "");
Expect(1, 68223, '\p{-_is_Sarb}', "");
Expect(0, 68223, '\p{^-_is_Sarb}', "");
Expect(0, 68223, '\P{-_is_Sarb}', "");
Expect(1, 68223, '\P{^-_is_Sarb}', "");
Expect(0, 68224, '\p{-_is_Sarb}', "");
Expect(1, 68224, '\p{^-_is_Sarb}', "");
Expect(1, 68224, '\P{-_is_Sarb}', "");
Expect(0, 68224, '\P{^-_is_Sarb}', "");
Error('\p{-/a/BATAK}');
Error('\P{-/a/BATAK}');
Expect(1, 7167, '\p{batak}', "");
Expect(0, 7167, '\p{^batak}', "");
Expect(0, 7167, '\P{batak}', "");
Expect(1, 7167, '\P{^batak}', "");
Expect(0, 7168, '\p{batak}', "");
Expect(1, 7168, '\p{^batak}', "");
Expect(1, 7168, '\P{batak}', "");
Expect(0, 7168, '\P{^batak}', "");
Expect(1, 7167, '\p{-BATAK}', "");
Expect(0, 7167, '\p{^-BATAK}', "");
Expect(0, 7167, '\P{-BATAK}', "");
Expect(1, 7167, '\P{^-BATAK}', "");
Expect(0, 7168, '\p{-BATAK}', "");
Expect(1, 7168, '\p{^-BATAK}', "");
Expect(1, 7168, '\P{-BATAK}', "");
Expect(0, 7168, '\P{^-BATAK}', "");
Error('\p{ Is_Batak:=}');
Error('\P{ Is_Batak:=}');
Expect(1, 7167, '\p{isbatak}', "");
Expect(0, 7167, '\p{^isbatak}', "");
Expect(0, 7167, '\P{isbatak}', "");
Expect(1, 7167, '\P{^isbatak}', "");
Expect(0, 7168, '\p{isbatak}', "");
Expect(1, 7168, '\p{^isbatak}', "");
Expect(1, 7168, '\P{isbatak}', "");
Expect(0, 7168, '\P{^isbatak}', "");
Expect(1, 7167, '\p{_IS_Batak}', "");
Expect(0, 7167, '\p{^_IS_Batak}', "");
Expect(0, 7167, '\P{_IS_Batak}', "");
Expect(1, 7167, '\P{^_IS_Batak}', "");
Expect(0, 7168, '\p{_IS_Batak}', "");
Expect(1, 7168, '\p{^_IS_Batak}', "");
Expect(1, 7168, '\P{_IS_Batak}', "");
Expect(0, 7168, '\P{^_IS_Batak}', "");
Error('\p{ _Batk/a/}');
Error('\P{ _Batk/a/}');
Expect(1, 7167, '\p{batk}', "");
Expect(0, 7167, '\p{^batk}', "");
Expect(0, 7167, '\P{batk}', "");
Expect(1, 7167, '\P{^batk}', "");
Expect(0, 7168, '\p{batk}', "");
Expect(1, 7168, '\p{^batk}', "");
Expect(1, 7168, '\P{batk}', "");
Expect(0, 7168, '\P{^batk}', "");
Expect(1, 7167, '\p{_ Batk}', "");
Expect(0, 7167, '\p{^_ Batk}', "");
Expect(0, 7167, '\P{_ Batk}', "");
Expect(1, 7167, '\P{^_ Batk}', "");
Expect(0, 7168, '\p{_ Batk}', "");
Expect(1, 7168, '\p{^_ Batk}', "");
Expect(1, 7168, '\P{_ Batk}', "");
Expect(0, 7168, '\P{^_ Batk}', "");
Error('\p{:=- Is_batk}');
Error('\P{:=- Is_batk}');
Expect(1, 7167, '\p{isbatk}', "");
Expect(0, 7167, '\p{^isbatk}', "");
Expect(0, 7167, '\P{isbatk}', "");
Expect(1, 7167, '\P{^isbatk}', "");
Expect(0, 7168, '\p{isbatk}', "");
Expect(1, 7168, '\p{^isbatk}', "");
Expect(1, 7168, '\P{isbatk}', "");
Expect(0, 7168, '\P{^isbatk}', "");
Expect(1, 7167, '\p{-is_Batk}', "");
Expect(0, 7167, '\p{^-is_Batk}', "");
Expect(0, 7167, '\P{-is_Batk}', "");
Expect(1, 7167, '\P{^-is_Batk}', "");
Expect(0, 7168, '\p{-is_Batk}', "");
Expect(1, 7168, '\p{^-is_Batk}', "");
Expect(1, 7168, '\P{-is_Batk}', "");
Expect(0, 7168, '\P{^-is_Batk}', "");
Error('\p{/a/_ ARABIC_extended_A}');
Error('\P{/a/_ ARABIC_extended_A}');
Expect(1, 2303, '\p{arabicextendeda}', "");
Expect(0, 2303, '\p{^arabicextendeda}', "");
Expect(0, 2303, '\P{arabicextendeda}', "");
Expect(1, 2303, '\P{^arabicextendeda}', "");
Expect(0, 2304, '\p{arabicextendeda}', "");
Expect(1, 2304, '\p{^arabicextendeda}', "");
Expect(1, 2304, '\P{arabicextendeda}', "");
Expect(0, 2304, '\P{^arabicextendeda}', "");
Expect(1, 2303, '\p{ Arabic_EXTENDED_A}', "");
Expect(0, 2303, '\p{^ Arabic_EXTENDED_A}', "");
Expect(0, 2303, '\P{ Arabic_EXTENDED_A}', "");
Expect(1, 2303, '\P{^ Arabic_EXTENDED_A}', "");
Expect(0, 2304, '\p{ Arabic_EXTENDED_A}', "");
Expect(1, 2304, '\p{^ Arabic_EXTENDED_A}', "");
Expect(1, 2304, '\P{ Arabic_EXTENDED_A}', "");
Expect(0, 2304, '\P{^ Arabic_EXTENDED_A}', "");
Error('\p{/a/_ Is_arabic_Extended_a}');
Error('\P{/a/_ Is_arabic_Extended_a}');
Expect(1, 2303, '\p{isarabicextendeda}', "");
Expect(0, 2303, '\p{^isarabicextendeda}', "");
Expect(0, 2303, '\P{isarabicextendeda}', "");
Expect(1, 2303, '\P{^isarabicextendeda}', "");
Expect(0, 2304, '\p{isarabicextendeda}', "");
Expect(1, 2304, '\p{^isarabicextendeda}', "");
Expect(1, 2304, '\P{isarabicextendeda}', "");
Expect(0, 2304, '\P{^isarabicextendeda}', "");
Expect(1, 2303, '\p{ Is_Arabic_Extended_A}', "");
Expect(0, 2303, '\p{^ Is_Arabic_Extended_A}', "");
Expect(0, 2303, '\P{ Is_Arabic_Extended_A}', "");
Expect(1, 2303, '\P{^ Is_Arabic_Extended_A}', "");
Expect(0, 2304, '\p{ Is_Arabic_Extended_A}', "");
Expect(1, 2304, '\p{^ Is_Arabic_Extended_A}', "");
Expect(1, 2304, '\P{ Is_Arabic_Extended_A}', "");
Expect(0, 2304, '\P{^ Is_Arabic_Extended_A}', "");
Error('\p{ IN_Arabic_Extended_A/a/}');
Error('\P{ IN_Arabic_Extended_A/a/}');
Expect(1, 2303, '\p{inarabicextendeda}', "");
Expect(0, 2303, '\p{^inarabicextendeda}', "");
Expect(0, 2303, '\P{inarabicextendeda}', "");
Expect(1, 2303, '\P{^inarabicextendeda}', "");
Expect(0, 2304, '\p{inarabicextendeda}', "");
Expect(1, 2304, '\p{^inarabicextendeda}', "");
Expect(1, 2304, '\P{inarabicextendeda}', "");
Expect(0, 2304, '\P{^inarabicextendeda}', "");
Expect(1, 2303, '\p{	-in_arabic_EXTENDED_A}', "");
Expect(0, 2303, '\p{^	-in_arabic_EXTENDED_A}', "");
Expect(0, 2303, '\P{	-in_arabic_EXTENDED_A}', "");
Expect(1, 2303, '\P{^	-in_arabic_EXTENDED_A}', "");
Expect(0, 2304, '\p{	-in_arabic_EXTENDED_A}', "");
Expect(1, 2304, '\p{^	-in_arabic_EXTENDED_A}', "");
Expect(1, 2304, '\P{	-in_arabic_EXTENDED_A}', "");
Expect(0, 2304, '\P{^	-in_arabic_EXTENDED_A}', "");
Error('\p{_/a/Arabic_ext_A}');
Error('\P{_/a/Arabic_ext_A}');
Expect(1, 2303, '\p{arabicexta}', "");
Expect(0, 2303, '\p{^arabicexta}', "");
Expect(0, 2303, '\P{arabicexta}', "");
Expect(1, 2303, '\P{^arabicexta}', "");
Expect(0, 2304, '\p{arabicexta}', "");
Expect(1, 2304, '\p{^arabicexta}', "");
Expect(1, 2304, '\P{arabicexta}', "");
Expect(0, 2304, '\P{^arabicexta}', "");
Expect(1, 2303, '\p{	-ARABIC_Ext_A}', "");
Expect(0, 2303, '\p{^	-ARABIC_Ext_A}', "");
Expect(0, 2303, '\P{	-ARABIC_Ext_A}', "");
Expect(1, 2303, '\P{^	-ARABIC_Ext_A}', "");
Expect(0, 2304, '\p{	-ARABIC_Ext_A}', "");
Expect(1, 2304, '\p{^	-ARABIC_Ext_A}', "");
Expect(1, 2304, '\P{	-ARABIC_Ext_A}', "");
Expect(0, 2304, '\P{^	-ARABIC_Ext_A}', "");
Error('\p{-IS_ARABIC_Ext_a/a/}');
Error('\P{-IS_ARABIC_Ext_a/a/}');
Expect(1, 2303, '\p{isarabicexta}', "");
Expect(0, 2303, '\p{^isarabicexta}', "");
Expect(0, 2303, '\P{isarabicexta}', "");
Expect(1, 2303, '\P{^isarabicexta}', "");
Expect(0, 2304, '\p{isarabicexta}', "");
Expect(1, 2304, '\p{^isarabicexta}', "");
Expect(1, 2304, '\P{isarabicexta}', "");
Expect(0, 2304, '\P{^isarabicexta}', "");
Expect(1, 2303, '\p{  Is_ARABIC_Ext_A}', "");
Expect(0, 2303, '\p{^  Is_ARABIC_Ext_A}', "");
Expect(0, 2303, '\P{  Is_ARABIC_Ext_A}', "");
Expect(1, 2303, '\P{^  Is_ARABIC_Ext_A}', "");
Expect(0, 2304, '\p{  Is_ARABIC_Ext_A}', "");
Expect(1, 2304, '\p{^  Is_ARABIC_Ext_A}', "");
Expect(1, 2304, '\P{  Is_ARABIC_Ext_A}', "");
Expect(0, 2304, '\P{^  Is_ARABIC_Ext_A}', "");
Error('\p{_/a/IN_Arabic_ext_a}');
Error('\P{_/a/IN_Arabic_ext_a}');
Expect(1, 2303, '\p{inarabicexta}', "");
Expect(0, 2303, '\p{^inarabicexta}', "");
Expect(0, 2303, '\P{inarabicexta}', "");
Expect(1, 2303, '\P{^inarabicexta}', "");
Expect(0, 2304, '\p{inarabicexta}', "");
Expect(1, 2304, '\p{^inarabicexta}', "");
Expect(1, 2304, '\P{inarabicexta}', "");
Expect(0, 2304, '\P{^inarabicexta}', "");
Expect(1, 2303, '\p{In_ARABIC_ext_A}', "");
Expect(0, 2303, '\p{^In_ARABIC_ext_A}', "");
Expect(0, 2303, '\P{In_ARABIC_ext_A}', "");
Expect(1, 2303, '\P{^In_ARABIC_ext_A}', "");
Expect(0, 2304, '\p{In_ARABIC_ext_A}', "");
Expect(1, 2304, '\p{^In_ARABIC_ext_A}', "");
Expect(1, 2304, '\P{In_ARABIC_ext_A}', "");
Expect(0, 2304, '\P{^In_ARABIC_ext_A}', "");
Error('\p{/a/	 Domino_tiles}');
Error('\P{/a/	 Domino_tiles}');
Expect(1, 127135, '\p{dominotiles}', "");
Expect(0, 127135, '\p{^dominotiles}', "");
Expect(0, 127135, '\P{dominotiles}', "");
Expect(1, 127135, '\P{^dominotiles}', "");
Expect(0, 127136, '\p{dominotiles}', "");
Expect(1, 127136, '\p{^dominotiles}', "");
Expect(1, 127136, '\P{dominotiles}', "");
Expect(0, 127136, '\P{^dominotiles}', "");
Expect(1, 127135, '\p{-_DOMINO_tiles}', "");
Expect(0, 127135, '\p{^-_DOMINO_tiles}', "");
Expect(0, 127135, '\P{-_DOMINO_tiles}', "");
Expect(1, 127135, '\P{^-_DOMINO_tiles}', "");
Expect(0, 127136, '\p{-_DOMINO_tiles}', "");
Expect(1, 127136, '\p{^-_DOMINO_tiles}', "");
Expect(1, 127136, '\P{-_DOMINO_tiles}', "");
Expect(0, 127136, '\P{^-_DOMINO_tiles}', "");
Error('\p{	/a/IS_Domino_tiles}');
Error('\P{	/a/IS_Domino_tiles}');
Expect(1, 127135, '\p{isdominotiles}', "");
Expect(0, 127135, '\p{^isdominotiles}', "");
Expect(0, 127135, '\P{isdominotiles}', "");
Expect(1, 127135, '\P{^isdominotiles}', "");
Expect(0, 127136, '\p{isdominotiles}', "");
Expect(1, 127136, '\p{^isdominotiles}', "");
Expect(1, 127136, '\P{isdominotiles}', "");
Expect(0, 127136, '\P{^isdominotiles}', "");
Expect(1, 127135, '\p{_	Is_Domino_Tiles}', "");
Expect(0, 127135, '\p{^_	Is_Domino_Tiles}', "");
Expect(0, 127135, '\P{_	Is_Domino_Tiles}', "");
Expect(1, 127135, '\P{^_	Is_Domino_Tiles}', "");
Expect(0, 127136, '\p{_	Is_Domino_Tiles}', "");
Expect(1, 127136, '\p{^_	Is_Domino_Tiles}', "");
Expect(1, 127136, '\P{_	Is_Domino_Tiles}', "");
Expect(0, 127136, '\P{^_	Is_Domino_Tiles}', "");
Error('\p{-/a/In_domino_Tiles}');
Error('\P{-/a/In_domino_Tiles}');
Expect(1, 127135, '\p{indominotiles}', "");
Expect(0, 127135, '\p{^indominotiles}', "");
Expect(0, 127135, '\P{indominotiles}', "");
Expect(1, 127135, '\P{^indominotiles}', "");
Expect(0, 127136, '\p{indominotiles}', "");
Expect(1, 127136, '\p{^indominotiles}', "");
Expect(1, 127136, '\P{indominotiles}', "");
Expect(0, 127136, '\P{^indominotiles}', "");
Expect(1, 127135, '\p{		In_domino_tiles}', "");
Expect(0, 127135, '\p{^		In_domino_tiles}', "");
Expect(0, 127135, '\P{		In_domino_tiles}', "");
Expect(1, 127135, '\P{^		In_domino_tiles}', "");
Expect(0, 127136, '\p{		In_domino_tiles}', "");
Expect(1, 127136, '\p{^		In_domino_tiles}', "");
Expect(1, 127136, '\P{		In_domino_tiles}', "");
Expect(0, 127136, '\P{^		In_domino_tiles}', "");
Error('\p{Domino:=}');
Error('\P{Domino:=}');
Expect(1, 127135, '\p{domino}', "");
Expect(0, 127135, '\p{^domino}', "");
Expect(0, 127135, '\P{domino}', "");
Expect(1, 127135, '\P{^domino}', "");
Expect(0, 127136, '\p{domino}', "");
Expect(1, 127136, '\p{^domino}', "");
Expect(1, 127136, '\P{domino}', "");
Expect(0, 127136, '\P{^domino}', "");
Expect(1, 127135, '\p{	_Domino}', "");
Expect(0, 127135, '\p{^	_Domino}', "");
Expect(0, 127135, '\P{	_Domino}', "");
Expect(1, 127135, '\P{^	_Domino}', "");
Expect(0, 127136, '\p{	_Domino}', "");
Expect(1, 127136, '\p{^	_Domino}', "");
Expect(1, 127136, '\P{	_Domino}', "");
Expect(0, 127136, '\P{^	_Domino}', "");
Error('\p{ IS_Domino:=}');
Error('\P{ IS_Domino:=}');
Expect(1, 127135, '\p{isdomino}', "");
Expect(0, 127135, '\p{^isdomino}', "");
Expect(0, 127135, '\P{isdomino}', "");
Expect(1, 127135, '\P{^isdomino}', "");
Expect(0, 127136, '\p{isdomino}', "");
Expect(1, 127136, '\p{^isdomino}', "");
Expect(1, 127136, '\P{isdomino}', "");
Expect(0, 127136, '\P{^isdomino}', "");
Expect(1, 127135, '\p{_is_Domino}', "");
Expect(0, 127135, '\p{^_is_Domino}', "");
Expect(0, 127135, '\P{_is_Domino}', "");
Expect(1, 127135, '\P{^_is_Domino}', "");
Expect(0, 127136, '\p{_is_Domino}', "");
Expect(1, 127136, '\p{^_is_Domino}', "");
Expect(1, 127136, '\P{_is_Domino}', "");
Expect(0, 127136, '\P{^_is_Domino}', "");
Error('\p{:=__In_Domino}');
Error('\P{:=__In_Domino}');
Expect(1, 127135, '\p{indomino}', "");
Expect(0, 127135, '\p{^indomino}', "");
Expect(0, 127135, '\P{indomino}', "");
Expect(1, 127135, '\P{^indomino}', "");
Expect(0, 127136, '\p{indomino}', "");
Expect(1, 127136, '\p{^indomino}', "");
Expect(1, 127136, '\P{indomino}', "");
Expect(0, 127136, '\P{^indomino}', "");
Expect(1, 127135, '\p{_-In_DOMINO}', "");
Expect(0, 127135, '\p{^_-In_DOMINO}', "");
Expect(0, 127135, '\P{_-In_DOMINO}', "");
Expect(1, 127135, '\P{^_-In_DOMINO}', "");
Expect(0, 127136, '\p{_-In_DOMINO}', "");
Expect(1, 127136, '\p{^_-In_DOMINO}', "");
Expect(1, 127136, '\P{_-In_DOMINO}', "");
Expect(0, 127136, '\P{^_-In_DOMINO}', "");
Error('\p{_-in_Kayah_LI/a/}');
Error('\P{_-in_Kayah_LI/a/}');
Expect(1, 43311, '\p{inkayahli}', "");
Expect(0, 43311, '\p{^inkayahli}', "");
Expect(0, 43311, '\P{inkayahli}', "");
Expect(1, 43311, '\P{^inkayahli}', "");
Expect(0, 43312, '\p{inkayahli}', "");
Expect(1, 43312, '\p{^inkayahli}', "");
Expect(1, 43312, '\P{inkayahli}', "");
Expect(0, 43312, '\P{^inkayahli}', "");
Expect(1, 43311, '\p{-IN_Kayah_LI}', "");
Expect(0, 43311, '\p{^-IN_Kayah_LI}', "");
Expect(0, 43311, '\P{-IN_Kayah_LI}', "");
Expect(1, 43311, '\P{^-IN_Kayah_LI}', "");
Expect(0, 43312, '\p{-IN_Kayah_LI}', "");
Expect(1, 43312, '\p{^-IN_Kayah_LI}', "");
Expect(1, 43312, '\P{-IN_Kayah_LI}', "");
Expect(0, 43312, '\P{^-IN_Kayah_LI}', "");
Error('\p{:=- IN_Kayah_LI}');
Error('\P{:=- IN_Kayah_LI}');
Expect(1, 43311, '\p{- IN_kayah_li}', "");
Expect(0, 43311, '\p{^- IN_kayah_li}', "");
Expect(0, 43311, '\P{- IN_kayah_li}', "");
Expect(1, 43311, '\P{^- IN_kayah_li}', "");
Expect(0, 43312, '\p{- IN_kayah_li}', "");
Expect(1, 43312, '\p{^- IN_kayah_li}', "");
Expect(1, 43312, '\P{- IN_kayah_li}', "");
Expect(0, 43312, '\P{^- IN_kayah_li}', "");
Error('\p{	_uppercase_Letter:=}');
Error('\P{	_uppercase_Letter:=}');
Expect(1, 120778, '\p{uppercaseletter}', "");
Expect(0, 120778, '\p{^uppercaseletter}', "");
Expect(0, 120778, '\P{uppercaseletter}', "");
Expect(1, 120778, '\P{^uppercaseletter}', "");
Expect(0, 120779, '\p{uppercaseletter}', "");
Expect(1, 120779, '\p{^uppercaseletter}', "");
Expect(1, 120779, '\P{uppercaseletter}', "");
Expect(0, 120779, '\P{^uppercaseletter}', "");
Expect(1, 120778, '\p{	Uppercase_Letter}', "");
Expect(0, 120778, '\p{^	Uppercase_Letter}', "");
Expect(0, 120778, '\P{	Uppercase_Letter}', "");
Expect(1, 120778, '\P{^	Uppercase_Letter}', "");
Expect(0, 120779, '\p{	Uppercase_Letter}', "");
Expect(1, 120779, '\p{^	Uppercase_Letter}', "");
Expect(1, 120779, '\P{	Uppercase_Letter}', "");
Expect(0, 120779, '\P{^	Uppercase_Letter}', "");
Error('\p{:=_ is_UPPERCASE_LETTER}');
Error('\P{:=_ is_UPPERCASE_LETTER}');
Expect(1, 120778, '\p{isuppercaseletter}', "");
Expect(0, 120778, '\p{^isuppercaseletter}', "");
Expect(0, 120778, '\P{isuppercaseletter}', "");
Expect(1, 120778, '\P{^isuppercaseletter}', "");
Expect(0, 120779, '\p{isuppercaseletter}', "");
Expect(1, 120779, '\p{^isuppercaseletter}', "");
Expect(1, 120779, '\P{isuppercaseletter}', "");
Expect(0, 120779, '\P{^isuppercaseletter}', "");
Expect(1, 120778, '\p{-Is_Uppercase_Letter}', "");
Expect(0, 120778, '\p{^-Is_Uppercase_Letter}', "");
Expect(0, 120778, '\P{-Is_Uppercase_Letter}', "");
Expect(1, 120778, '\P{^-Is_Uppercase_Letter}', "");
Expect(0, 120779, '\p{-Is_Uppercase_Letter}', "");
Expect(1, 120779, '\p{^-Is_Uppercase_Letter}', "");
Expect(1, 120779, '\P{-Is_Uppercase_Letter}', "");
Expect(0, 120779, '\P{^-Is_Uppercase_Letter}', "");
Error('\p{:=-	Lu}');
Error('\P{:=-	Lu}');
Expect(1, 120778, '\p{lu}', "");
Expect(0, 120778, '\p{^lu}', "");
Expect(0, 120778, '\P{lu}', "");
Expect(1, 120778, '\P{^lu}', "");
Expect(0, 120779, '\p{lu}', "");
Expect(1, 120779, '\p{^lu}', "");
Expect(1, 120779, '\P{lu}', "");
Expect(0, 120779, '\P{^lu}', "");
Expect(1, 120778, '\p{_lu}', "");
Expect(0, 120778, '\p{^_lu}', "");
Expect(0, 120778, '\P{_lu}', "");
Expect(1, 120778, '\P{^_lu}', "");
Expect(0, 120779, '\p{_lu}', "");
Expect(1, 120779, '\p{^_lu}', "");
Expect(1, 120779, '\P{_lu}', "");
Expect(0, 120779, '\P{^_lu}', "");
Error('\p{/a/-	Is_Lu}');
Error('\P{/a/-	Is_Lu}');
Expect(1, 120778, '\p{islu}', "");
Expect(0, 120778, '\p{^islu}', "");
Expect(0, 120778, '\P{islu}', "");
Expect(1, 120778, '\P{^islu}', "");
Expect(0, 120779, '\p{islu}', "");
Expect(1, 120779, '\p{^islu}', "");
Expect(1, 120779, '\P{islu}', "");
Expect(0, 120779, '\P{^islu}', "");
Expect(1, 120778, '\p{-is_LU}', "");
Expect(0, 120778, '\p{^-is_LU}', "");
Expect(0, 120778, '\P{-is_LU}', "");
Expect(1, 120778, '\P{^-is_LU}', "");
Expect(0, 120779, '\p{-is_LU}', "");
Expect(1, 120779, '\p{^-is_LU}', "");
Expect(1, 120779, '\P{-is_LU}', "");
Expect(0, 120779, '\P{^-is_LU}', "");
Error('\p{:=-HALFWIDTH_And_Fullwidth_FORMS}');
Error('\P{:=-HALFWIDTH_And_Fullwidth_FORMS}');
Expect(1, 65519, '\p{halfwidthandfullwidthforms}', "");
Expect(0, 65519, '\p{^halfwidthandfullwidthforms}', "");
Expect(0, 65519, '\P{halfwidthandfullwidthforms}', "");
Expect(1, 65519, '\P{^halfwidthandfullwidthforms}', "");
Expect(0, 65520, '\p{halfwidthandfullwidthforms}', "");
Expect(1, 65520, '\p{^halfwidthandfullwidthforms}', "");
Expect(1, 65520, '\P{halfwidthandfullwidthforms}', "");
Expect(0, 65520, '\P{^halfwidthandfullwidthforms}', "");
Expect(1, 65519, '\p{	halfwidth_AND_FULLWIDTH_FORMS}', "");
Expect(0, 65519, '\p{^	halfwidth_AND_FULLWIDTH_FORMS}', "");
Expect(0, 65519, '\P{	halfwidth_AND_FULLWIDTH_FORMS}', "");
Expect(1, 65519, '\P{^	halfwidth_AND_FULLWIDTH_FORMS}', "");
Expect(0, 65520, '\p{	halfwidth_AND_FULLWIDTH_FORMS}', "");
Expect(1, 65520, '\p{^	halfwidth_AND_FULLWIDTH_FORMS}', "");
Expect(1, 65520, '\P{	halfwidth_AND_FULLWIDTH_FORMS}', "");
Expect(0, 65520, '\P{^	halfwidth_AND_FULLWIDTH_FORMS}', "");
Error('\p{__IS_halfwidth_and_FULLWIDTH_Forms:=}');
Error('\P{__IS_halfwidth_and_FULLWIDTH_Forms:=}');
Expect(1, 65519, '\p{ishalfwidthandfullwidthforms}', "");
Expect(0, 65519, '\p{^ishalfwidthandfullwidthforms}', "");
Expect(0, 65519, '\P{ishalfwidthandfullwidthforms}', "");
Expect(1, 65519, '\P{^ishalfwidthandfullwidthforms}', "");
Expect(0, 65520, '\p{ishalfwidthandfullwidthforms}', "");
Expect(1, 65520, '\p{^ishalfwidthandfullwidthforms}', "");
Expect(1, 65520, '\P{ishalfwidthandfullwidthforms}', "");
Expect(0, 65520, '\P{^ishalfwidthandfullwidthforms}', "");
Expect(1, 65519, '\p{Is_HALFWIDTH_And_Fullwidth_FORMS}', "");
Expect(0, 65519, '\p{^Is_HALFWIDTH_And_Fullwidth_FORMS}', "");
Expect(0, 65519, '\P{Is_HALFWIDTH_And_Fullwidth_FORMS}', "");
Expect(1, 65519, '\P{^Is_HALFWIDTH_And_Fullwidth_FORMS}', "");
Expect(0, 65520, '\p{Is_HALFWIDTH_And_Fullwidth_FORMS}', "");
Expect(1, 65520, '\p{^Is_HALFWIDTH_And_Fullwidth_FORMS}', "");
Expect(1, 65520, '\P{Is_HALFWIDTH_And_Fullwidth_FORMS}', "");
Expect(0, 65520, '\P{^Is_HALFWIDTH_And_Fullwidth_FORMS}', "");
Error('\p{	/a/In_HALFWIDTH_and_fullwidth_Forms}');
Error('\P{	/a/In_HALFWIDTH_and_fullwidth_Forms}');
Expect(1, 65519, '\p{inhalfwidthandfullwidthforms}', "");
Expect(0, 65519, '\p{^inhalfwidthandfullwidthforms}', "");
Expect(0, 65519, '\P{inhalfwidthandfullwidthforms}', "");
Expect(1, 65519, '\P{^inhalfwidthandfullwidthforms}', "");
Expect(0, 65520, '\p{inhalfwidthandfullwidthforms}', "");
Expect(1, 65520, '\p{^inhalfwidthandfullwidthforms}', "");
Expect(1, 65520, '\P{inhalfwidthandfullwidthforms}', "");
Expect(0, 65520, '\P{^inhalfwidthandfullwidthforms}', "");
Expect(1, 65519, '\p{__in_Halfwidth_AND_Fullwidth_forms}', "");
Expect(0, 65519, '\p{^__in_Halfwidth_AND_Fullwidth_forms}', "");
Expect(0, 65519, '\P{__in_Halfwidth_AND_Fullwidth_forms}', "");
Expect(1, 65519, '\P{^__in_Halfwidth_AND_Fullwidth_forms}', "");
Expect(0, 65520, '\p{__in_Halfwidth_AND_Fullwidth_forms}', "");
Expect(1, 65520, '\p{^__in_Halfwidth_AND_Fullwidth_forms}', "");
Expect(1, 65520, '\P{__in_Halfwidth_AND_Fullwidth_forms}', "");
Expect(0, 65520, '\P{^__in_Halfwidth_AND_Fullwidth_forms}', "");
Error('\p{_:=HALF_And_FULL_Forms}');
Error('\P{_:=HALF_And_FULL_Forms}');
Expect(1, 65519, '\p{halfandfullforms}', "");
Expect(0, 65519, '\p{^halfandfullforms}', "");
Expect(0, 65519, '\P{halfandfullforms}', "");
Expect(1, 65519, '\P{^halfandfullforms}', "");
Expect(0, 65520, '\p{halfandfullforms}', "");
Expect(1, 65520, '\p{^halfandfullforms}', "");
Expect(1, 65520, '\P{halfandfullforms}', "");
Expect(0, 65520, '\P{^halfandfullforms}', "");
Expect(1, 65519, '\p{-Half_and_FULL_Forms}', "");
Expect(0, 65519, '\p{^-Half_and_FULL_Forms}', "");
Expect(0, 65519, '\P{-Half_and_FULL_Forms}', "");
Expect(1, 65519, '\P{^-Half_and_FULL_Forms}', "");
Expect(0, 65520, '\p{-Half_and_FULL_Forms}', "");
Expect(1, 65520, '\p{^-Half_and_FULL_Forms}', "");
Expect(1, 65520, '\P{-Half_and_FULL_Forms}', "");
Expect(0, 65520, '\P{^-Half_and_FULL_Forms}', "");
Error('\p{/a/- Is_Half_AND_Full_FORMS}');
Error('\P{/a/- Is_Half_AND_Full_FORMS}');
Expect(1, 65519, '\p{ishalfandfullforms}', "");
Expect(0, 65519, '\p{^ishalfandfullforms}', "");
Expect(0, 65519, '\P{ishalfandfullforms}', "");
Expect(1, 65519, '\P{^ishalfandfullforms}', "");
Expect(0, 65520, '\p{ishalfandfullforms}', "");
Expect(1, 65520, '\p{^ishalfandfullforms}', "");
Expect(1, 65520, '\P{ishalfandfullforms}', "");
Expect(0, 65520, '\P{^ishalfandfullforms}', "");
Expect(1, 65519, '\p{ _IS_HALF_AND_full_forms}', "");
Expect(0, 65519, '\p{^ _IS_HALF_AND_full_forms}', "");
Expect(0, 65519, '\P{ _IS_HALF_AND_full_forms}', "");
Expect(1, 65519, '\P{^ _IS_HALF_AND_full_forms}', "");
Expect(0, 65520, '\p{ _IS_HALF_AND_full_forms}', "");
Expect(1, 65520, '\p{^ _IS_HALF_AND_full_forms}', "");
Expect(1, 65520, '\P{ _IS_HALF_AND_full_forms}', "");
Expect(0, 65520, '\P{^ _IS_HALF_AND_full_forms}', "");
Error('\p{- in_half_AND_FULL_forms:=}');
Error('\P{- in_half_AND_FULL_forms:=}');
Expect(1, 65519, '\p{inhalfandfullforms}', "");
Expect(0, 65519, '\p{^inhalfandfullforms}', "");
Expect(0, 65519, '\P{inhalfandfullforms}', "");
Expect(1, 65519, '\P{^inhalfandfullforms}', "");
Expect(0, 65520, '\p{inhalfandfullforms}', "");
Expect(1, 65520, '\p{^inhalfandfullforms}', "");
Expect(1, 65520, '\P{inhalfandfullforms}', "");
Expect(0, 65520, '\P{^inhalfandfullforms}', "");
Expect(1, 65519, '\p{ _IN_Half_and_Full_Forms}', "");
Expect(0, 65519, '\p{^ _IN_Half_and_Full_Forms}', "");
Expect(0, 65519, '\P{ _IN_Half_and_Full_Forms}', "");
Expect(1, 65519, '\P{^ _IN_Half_and_Full_Forms}', "");
Expect(0, 65520, '\p{ _IN_Half_and_Full_Forms}', "");
Expect(1, 65520, '\p{^ _IN_Half_and_Full_Forms}', "");
Expect(1, 65520, '\P{ _IN_Half_and_Full_Forms}', "");
Expect(0, 65520, '\P{^ _IN_Half_and_Full_Forms}', "");
Error('\p{_-Tai_Le:=}');
Error('\P{_-Tai_Le:=}');
Expect(1, 6516, '\p{taile}', "");
Expect(0, 6516, '\p{^taile}', "");
Expect(0, 6516, '\P{taile}', "");
Expect(1, 6516, '\P{^taile}', "");
Expect(0, 6517, '\p{taile}', "");
Expect(1, 6517, '\p{^taile}', "");
Expect(1, 6517, '\P{taile}', "");
Expect(0, 6517, '\P{^taile}', "");
Expect(1, 6516, '\p{-_Tai_Le}', "");
Expect(0, 6516, '\p{^-_Tai_Le}', "");
Expect(0, 6516, '\P{-_Tai_Le}', "");
Expect(1, 6516, '\P{^-_Tai_Le}', "");
Expect(0, 6517, '\p{-_Tai_Le}', "");
Expect(1, 6517, '\p{^-_Tai_Le}', "");
Expect(1, 6517, '\P{-_Tai_Le}', "");
Expect(0, 6517, '\P{^-_Tai_Le}', "");
Error('\p{_:=is_TAI_Le}');
Error('\P{_:=is_TAI_Le}');
Expect(1, 6516, '\p{istaile}', "");
Expect(0, 6516, '\p{^istaile}', "");
Expect(0, 6516, '\P{istaile}', "");
Expect(1, 6516, '\P{^istaile}', "");
Expect(0, 6517, '\p{istaile}', "");
Expect(1, 6517, '\p{^istaile}', "");
Expect(1, 6517, '\P{istaile}', "");
Expect(0, 6517, '\P{^istaile}', "");
Expect(1, 6516, '\p{-Is_tai_LE}', "");
Expect(0, 6516, '\p{^-Is_tai_LE}', "");
Expect(0, 6516, '\P{-Is_tai_LE}', "");
Expect(1, 6516, '\P{^-Is_tai_LE}', "");
Expect(0, 6517, '\p{-Is_tai_LE}', "");
Expect(1, 6517, '\p{^-Is_tai_LE}', "");
Expect(1, 6517, '\P{-Is_tai_LE}', "");
Expect(0, 6517, '\P{^-Is_tai_LE}', "");
Error('\p{:=	-TALE}');
Error('\P{:=	-TALE}');
Expect(1, 6516, '\p{tale}', "");
Expect(0, 6516, '\p{^tale}', "");
Expect(0, 6516, '\P{tale}', "");
Expect(1, 6516, '\P{^tale}', "");
Expect(0, 6517, '\p{tale}', "");
Expect(1, 6517, '\p{^tale}', "");
Expect(1, 6517, '\P{tale}', "");
Expect(0, 6517, '\P{^tale}', "");
Expect(1, 6516, '\p{	-tale}', "");
Expect(0, 6516, '\p{^	-tale}', "");
Expect(0, 6516, '\P{	-tale}', "");
Expect(1, 6516, '\P{^	-tale}', "");
Expect(0, 6517, '\p{	-tale}', "");
Expect(1, 6517, '\p{^	-tale}', "");
Expect(1, 6517, '\P{	-tale}', "");
Expect(0, 6517, '\P{^	-tale}', "");
Error('\p{-/a/IS_tale}');
Error('\P{-/a/IS_tale}');
Expect(1, 6516, '\p{istale}', "");
Expect(0, 6516, '\p{^istale}', "");
Expect(0, 6516, '\P{istale}', "");
Expect(1, 6516, '\P{^istale}', "");
Expect(0, 6517, '\p{istale}', "");
Expect(1, 6517, '\p{^istale}', "");
Expect(1, 6517, '\P{istale}', "");
Expect(0, 6517, '\P{^istale}', "");
Expect(1, 6516, '\p{IS_tale}', "");
Expect(0, 6516, '\p{^IS_tale}', "");
Expect(0, 6516, '\P{IS_tale}', "");
Expect(1, 6516, '\P{^IS_tale}', "");
Expect(0, 6517, '\p{IS_tale}', "");
Expect(1, 6517, '\p{^IS_tale}', "");
Expect(1, 6517, '\P{IS_tale}', "");
Expect(0, 6517, '\P{^IS_tale}', "");
Error('\p{/a/ -PosixXDigit}');
Error('\P{/a/ -PosixXDigit}');
Expect(1, 102, '\p{posixxdigit}', "");
Expect(0, 102, '\p{^posixxdigit}', "");
Expect(0, 102, '\P{posixxdigit}', "");
Expect(1, 102, '\P{^posixxdigit}', "");
Expect(0, 103, '\p{posixxdigit}', "");
Expect(1, 103, '\p{^posixxdigit}', "");
Expect(1, 103, '\P{posixxdigit}', "");
Expect(0, 103, '\P{^posixxdigit}', "");
Expect(1, 102, '\p{_	PosixXDigit}', "");
Expect(0, 102, '\p{^_	PosixXDigit}', "");
Expect(0, 102, '\P{_	PosixXDigit}', "");
Expect(1, 102, '\P{^_	PosixXDigit}', "");
Expect(0, 103, '\p{_	PosixXDigit}', "");
Expect(1, 103, '\p{^_	PosixXDigit}', "");
Expect(1, 103, '\P{_	PosixXDigit}', "");
Expect(0, 103, '\P{^_	PosixXDigit}', "");
Error('\p{/a/__Is_PosixXDigit}');
Error('\P{/a/__Is_PosixXDigit}');
Expect(1, 102, '\p{isposixxdigit}', "");
Expect(0, 102, '\p{^isposixxdigit}', "");
Expect(0, 102, '\P{isposixxdigit}', "");
Expect(1, 102, '\P{^isposixxdigit}', "");
Expect(0, 103, '\p{isposixxdigit}', "");
Expect(1, 103, '\p{^isposixxdigit}', "");
Expect(1, 103, '\P{isposixxdigit}', "");
Expect(0, 103, '\P{^isposixxdigit}', "");
Expect(1, 102, '\p{_-Is_POSIXXDIGIT}', "");
Expect(0, 102, '\p{^_-Is_POSIXXDIGIT}', "");
Expect(0, 102, '\P{_-Is_POSIXXDIGIT}', "");
Expect(1, 102, '\P{^_-Is_POSIXXDIGIT}', "");
Expect(0, 103, '\p{_-Is_POSIXXDIGIT}', "");
Expect(1, 103, '\p{^_-Is_POSIXXDIGIT}', "");
Expect(1, 103, '\P{_-Is_POSIXXDIGIT}', "");
Expect(0, 103, '\P{^_-Is_POSIXXDIGIT}', "");
Error('\p{/a/	 ASCII_hex_Digit}');
Error('\P{/a/	 ASCII_hex_Digit}');
Expect(1, 102, '\p{asciihexdigit}', "");
Expect(0, 102, '\p{^asciihexdigit}', "");
Expect(0, 102, '\P{asciihexdigit}', "");
Expect(1, 102, '\P{^asciihexdigit}', "");
Expect(0, 103, '\p{asciihexdigit}', "");
Expect(1, 103, '\p{^asciihexdigit}', "");
Expect(1, 103, '\P{asciihexdigit}', "");
Expect(0, 103, '\P{^asciihexdigit}', "");
Expect(1, 102, '\p{_-ascii_hex_DIGIT}', "");
Expect(0, 102, '\p{^_-ascii_hex_DIGIT}', "");
Expect(0, 102, '\P{_-ascii_hex_DIGIT}', "");
Expect(1, 102, '\P{^_-ascii_hex_DIGIT}', "");
Expect(0, 103, '\p{_-ascii_hex_DIGIT}', "");
Expect(1, 103, '\p{^_-ascii_hex_DIGIT}', "");
Expect(1, 103, '\P{_-ascii_hex_DIGIT}', "");
Expect(0, 103, '\P{^_-ascii_hex_DIGIT}', "");
Error('\p{/a/	Is_ASCII_hex_Digit}');
Error('\P{/a/	Is_ASCII_hex_Digit}');
Expect(1, 102, '\p{isasciihexdigit}', "");
Expect(0, 102, '\p{^isasciihexdigit}', "");
Expect(0, 102, '\P{isasciihexdigit}', "");
Expect(1, 102, '\P{^isasciihexdigit}', "");
Expect(0, 103, '\p{isasciihexdigit}', "");
Expect(1, 103, '\p{^isasciihexdigit}', "");
Expect(1, 103, '\P{isasciihexdigit}', "");
Expect(0, 103, '\P{^isasciihexdigit}', "");
Expect(1, 102, '\p{_-IS_ascii_hex_DIGIT}', "");
Expect(0, 102, '\p{^_-IS_ascii_hex_DIGIT}', "");
Expect(0, 102, '\P{_-IS_ascii_hex_DIGIT}', "");
Expect(1, 102, '\P{^_-IS_ascii_hex_DIGIT}', "");
Expect(0, 103, '\p{_-IS_ascii_hex_DIGIT}', "");
Expect(1, 103, '\p{^_-IS_ascii_hex_DIGIT}', "");
Expect(1, 103, '\P{_-IS_ascii_hex_DIGIT}', "");
Expect(0, 103, '\P{^_-IS_ascii_hex_DIGIT}', "");
Error('\p{_/a/ahex}');
Error('\P{_/a/ahex}');
Expect(1, 102, '\p{ahex}', "");
Expect(0, 102, '\p{^ahex}', "");
Expect(0, 102, '\P{ahex}', "");
Expect(1, 102, '\P{^ahex}', "");
Expect(0, 103, '\p{ahex}', "");
Expect(1, 103, '\p{^ahex}', "");
Expect(1, 103, '\P{ahex}', "");
Expect(0, 103, '\P{^ahex}', "");
Expect(1, 102, '\p{	AHEX}', "");
Expect(0, 102, '\p{^	AHEX}', "");
Expect(0, 102, '\P{	AHEX}', "");
Expect(1, 102, '\P{^	AHEX}', "");
Expect(0, 103, '\p{	AHEX}', "");
Expect(1, 103, '\p{^	AHEX}', "");
Expect(1, 103, '\P{	AHEX}', "");
Expect(0, 103, '\P{^	AHEX}', "");
Error('\p{ IS_AHex:=}');
Error('\P{ IS_AHex:=}');
Expect(1, 102, '\p{isahex}', "");
Expect(0, 102, '\p{^isahex}', "");
Expect(0, 102, '\P{isahex}', "");
Expect(1, 102, '\P{^isahex}', "");
Expect(0, 103, '\p{isahex}', "");
Expect(1, 103, '\p{^isahex}', "");
Expect(1, 103, '\P{isahex}', "");
Expect(0, 103, '\P{^isahex}', "");
Expect(1, 102, '\p{  Is_AHex}', "");
Expect(0, 102, '\p{^  Is_AHex}', "");
Expect(0, 102, '\P{  Is_AHex}', "");
Expect(1, 102, '\P{^  Is_AHex}', "");
Expect(0, 103, '\p{  Is_AHex}', "");
Expect(1, 103, '\p{^  Is_AHex}', "");
Expect(1, 103, '\P{  Is_AHex}', "");
Expect(0, 103, '\P{^  Is_AHex}', "");
Error('\p{_/a/COMBINING_half_Marks}');
Error('\P{_/a/COMBINING_half_Marks}');
Expect(1, 65071, '\p{combininghalfmarks}', "");
Expect(0, 65071, '\p{^combininghalfmarks}', "");
Expect(0, 65071, '\P{combininghalfmarks}', "");
Expect(1, 65071, '\P{^combininghalfmarks}', "");
Expect(0, 65072, '\p{combininghalfmarks}', "");
Expect(1, 65072, '\p{^combininghalfmarks}', "");
Expect(1, 65072, '\P{combininghalfmarks}', "");
Expect(0, 65072, '\P{^combininghalfmarks}', "");
Expect(1, 65071, '\p{	_combining_HALF_marks}', "");
Expect(0, 65071, '\p{^	_combining_HALF_marks}', "");
Expect(0, 65071, '\P{	_combining_HALF_marks}', "");
Expect(1, 65071, '\P{^	_combining_HALF_marks}', "");
Expect(0, 65072, '\p{	_combining_HALF_marks}', "");
Expect(1, 65072, '\p{^	_combining_HALF_marks}', "");
Expect(1, 65072, '\P{	_combining_HALF_marks}', "");
Expect(0, 65072, '\P{^	_combining_HALF_marks}', "");
Error('\p{:= Is_combining_half_Marks}');
Error('\P{:= Is_combining_half_Marks}');
Expect(1, 65071, '\p{iscombininghalfmarks}', "");
Expect(0, 65071, '\p{^iscombininghalfmarks}', "");
Expect(0, 65071, '\P{iscombininghalfmarks}', "");
Expect(1, 65071, '\P{^iscombininghalfmarks}', "");
Expect(0, 65072, '\p{iscombininghalfmarks}', "");
Expect(1, 65072, '\p{^iscombininghalfmarks}', "");
Expect(1, 65072, '\P{iscombininghalfmarks}', "");
Expect(0, 65072, '\P{^iscombininghalfmarks}', "");
Expect(1, 65071, '\p{__Is_Combining_Half_MARKS}', "");
Expect(0, 65071, '\p{^__Is_Combining_Half_MARKS}', "");
Expect(0, 65071, '\P{__Is_Combining_Half_MARKS}', "");
Expect(1, 65071, '\P{^__Is_Combining_Half_MARKS}', "");
Expect(0, 65072, '\p{__Is_Combining_Half_MARKS}', "");
Expect(1, 65072, '\p{^__Is_Combining_Half_MARKS}', "");
Expect(1, 65072, '\P{__Is_Combining_Half_MARKS}', "");
Expect(0, 65072, '\P{^__Is_Combining_Half_MARKS}', "");
Error('\p{-	in_Combining_Half_marks:=}');
Error('\P{-	in_Combining_Half_marks:=}');
Expect(1, 65071, '\p{incombininghalfmarks}', "");
Expect(0, 65071, '\p{^incombininghalfmarks}', "");
Expect(0, 65071, '\P{incombininghalfmarks}', "");
Expect(1, 65071, '\P{^incombininghalfmarks}', "");
Expect(0, 65072, '\p{incombininghalfmarks}', "");
Expect(1, 65072, '\p{^incombininghalfmarks}', "");
Expect(1, 65072, '\P{incombininghalfmarks}', "");
Expect(0, 65072, '\P{^incombininghalfmarks}', "");
Expect(1, 65071, '\p{ 	In_COMBINING_Half_MARKS}', "");
Expect(0, 65071, '\p{^ 	In_COMBINING_Half_MARKS}', "");
Expect(0, 65071, '\P{ 	In_COMBINING_Half_MARKS}', "");
Expect(1, 65071, '\P{^ 	In_COMBINING_Half_MARKS}', "");
Expect(0, 65072, '\p{ 	In_COMBINING_Half_MARKS}', "");
Expect(1, 65072, '\p{^ 	In_COMBINING_Half_MARKS}', "");
Expect(1, 65072, '\P{ 	In_COMBINING_Half_MARKS}', "");
Expect(0, 65072, '\P{^ 	In_COMBINING_Half_MARKS}', "");
Error('\p{ _Half_marks/a/}');
Error('\P{ _Half_marks/a/}');
Expect(1, 65071, '\p{halfmarks}', "");
Expect(0, 65071, '\p{^halfmarks}', "");
Expect(0, 65071, '\P{halfmarks}', "");
Expect(1, 65071, '\P{^halfmarks}', "");
Expect(0, 65072, '\p{halfmarks}', "");
Expect(1, 65072, '\p{^halfmarks}', "");
Expect(1, 65072, '\P{halfmarks}', "");
Expect(0, 65072, '\P{^halfmarks}', "");
Expect(1, 65071, '\p{ Half_marks}', "");
Expect(0, 65071, '\p{^ Half_marks}', "");
Expect(0, 65071, '\P{ Half_marks}', "");
Expect(1, 65071, '\P{^ Half_marks}', "");
Expect(0, 65072, '\p{ Half_marks}', "");
Expect(1, 65072, '\p{^ Half_marks}', "");
Expect(1, 65072, '\P{ Half_marks}', "");
Expect(0, 65072, '\P{^ Half_marks}', "");
Error('\p{- Is_Half_Marks/a/}');
Error('\P{- Is_Half_Marks/a/}');
Expect(1, 65071, '\p{ishalfmarks}', "");
Expect(0, 65071, '\p{^ishalfmarks}', "");
Expect(0, 65071, '\P{ishalfmarks}', "");
Expect(1, 65071, '\P{^ishalfmarks}', "");
Expect(0, 65072, '\p{ishalfmarks}', "");
Expect(1, 65072, '\p{^ishalfmarks}', "");
Expect(1, 65072, '\P{ishalfmarks}', "");
Expect(0, 65072, '\P{^ishalfmarks}', "");
Expect(1, 65071, '\p{_-is_Half_MARKS}', "");
Expect(0, 65071, '\p{^_-is_Half_MARKS}', "");
Expect(0, 65071, '\P{_-is_Half_MARKS}', "");
Expect(1, 65071, '\P{^_-is_Half_MARKS}', "");
Expect(0, 65072, '\p{_-is_Half_MARKS}', "");
Expect(1, 65072, '\p{^_-is_Half_MARKS}', "");
Expect(1, 65072, '\P{_-is_Half_MARKS}', "");
Expect(0, 65072, '\P{^_-is_Half_MARKS}', "");
Error('\p{:=In_HALF_Marks}');
Error('\P{:=In_HALF_Marks}');
Expect(1, 65071, '\p{inhalfmarks}', "");
Expect(0, 65071, '\p{^inhalfmarks}', "");
Expect(0, 65071, '\P{inhalfmarks}', "");
Expect(1, 65071, '\P{^inhalfmarks}', "");
Expect(0, 65072, '\p{inhalfmarks}', "");
Expect(1, 65072, '\p{^inhalfmarks}', "");
Expect(1, 65072, '\P{inhalfmarks}', "");
Expect(0, 65072, '\P{^inhalfmarks}', "");
Expect(1, 65071, '\p{		In_half_Marks}', "");
Expect(0, 65071, '\p{^		In_half_Marks}', "");
Expect(0, 65071, '\P{		In_half_Marks}', "");
Expect(1, 65071, '\P{^		In_half_Marks}', "");
Expect(0, 65072, '\p{		In_half_Marks}', "");
Expect(1, 65072, '\p{^		In_half_Marks}', "");
Expect(1, 65072, '\P{		In_half_Marks}', "");
Expect(0, 65072, '\P{^		In_half_Marks}', "");
Error('\p{:=	-Mandaic}');
Error('\P{:=	-Mandaic}');
Expect(1, 2142, '\p{mandaic}', "");
Expect(0, 2142, '\p{^mandaic}', "");
Expect(0, 2142, '\P{mandaic}', "");
Expect(1, 2142, '\P{^mandaic}', "");
Expect(0, 2143, '\p{mandaic}', "");
Expect(1, 2143, '\p{^mandaic}', "");
Expect(1, 2143, '\P{mandaic}', "");
Expect(0, 2143, '\P{^mandaic}', "");
Expect(1, 2142, '\p{	_Mandaic}', "");
Expect(0, 2142, '\p{^	_Mandaic}', "");
Expect(0, 2142, '\P{	_Mandaic}', "");
Expect(1, 2142, '\P{^	_Mandaic}', "");
Expect(0, 2143, '\p{	_Mandaic}', "");
Expect(1, 2143, '\p{^	_Mandaic}', "");
Expect(1, 2143, '\P{	_Mandaic}', "");
Expect(0, 2143, '\P{^	_Mandaic}', "");
Error('\p{/a/--is_mandaic}');
Error('\P{/a/--is_mandaic}');
Expect(1, 2142, '\p{ismandaic}', "");
Expect(0, 2142, '\p{^ismandaic}', "");
Expect(0, 2142, '\P{ismandaic}', "");
Expect(1, 2142, '\P{^ismandaic}', "");
Expect(0, 2143, '\p{ismandaic}', "");
Expect(1, 2143, '\p{^ismandaic}', "");
Expect(1, 2143, '\P{ismandaic}', "");
Expect(0, 2143, '\P{^ismandaic}', "");
Expect(1, 2142, '\p{-	Is_Mandaic}', "");
Expect(0, 2142, '\p{^-	Is_Mandaic}', "");
Expect(0, 2142, '\P{-	Is_Mandaic}', "");
Expect(1, 2142, '\P{^-	Is_Mandaic}', "");
Expect(0, 2143, '\p{-	Is_Mandaic}', "");
Expect(1, 2143, '\p{^-	Is_Mandaic}', "");
Expect(1, 2143, '\P{-	Is_Mandaic}', "");
Expect(0, 2143, '\P{^-	Is_Mandaic}', "");
Error('\p{:=__MAND}');
Error('\P{:=__MAND}');
Expect(1, 2142, '\p{mand}', "");
Expect(0, 2142, '\p{^mand}', "");
Expect(0, 2142, '\P{mand}', "");
Expect(1, 2142, '\P{^mand}', "");
Expect(0, 2143, '\p{mand}', "");
Expect(1, 2143, '\p{^mand}', "");
Expect(1, 2143, '\P{mand}', "");
Expect(0, 2143, '\P{^mand}', "");
Expect(1, 2142, '\p{ mand}', "");
Expect(0, 2142, '\p{^ mand}', "");
Expect(0, 2142, '\P{ mand}', "");
Expect(1, 2142, '\P{^ mand}', "");
Expect(0, 2143, '\p{ mand}', "");
Expect(1, 2143, '\p{^ mand}', "");
Expect(1, 2143, '\P{ mand}', "");
Expect(0, 2143, '\P{^ mand}', "");
Error('\p{/a/-Is_Mand}');
Error('\P{/a/-Is_Mand}');
Expect(1, 2142, '\p{ismand}', "");
Expect(0, 2142, '\p{^ismand}', "");
Expect(0, 2142, '\P{ismand}', "");
Expect(1, 2142, '\P{^ismand}', "");
Expect(0, 2143, '\p{ismand}', "");
Expect(1, 2143, '\p{^ismand}', "");
Expect(1, 2143, '\P{ismand}', "");
Expect(0, 2143, '\P{^ismand}', "");
Expect(1, 2142, '\p{_	IS_Mand}', "");
Expect(0, 2142, '\p{^_	IS_Mand}', "");
Expect(0, 2142, '\P{_	IS_Mand}', "");
Expect(1, 2142, '\P{^_	IS_Mand}', "");
Expect(0, 2143, '\p{_	IS_Mand}', "");
Expect(1, 2143, '\p{^_	IS_Mand}', "");
Expect(1, 2143, '\P{_	IS_Mand}', "");
Expect(0, 2143, '\P{^_	IS_Mand}', "");
Error('\p{:=	_rejang}');
Error('\P{:=	_rejang}');
Expect(1, 43359, '\p{rejang}', "");
Expect(0, 43359, '\p{^rejang}', "");
Expect(0, 43359, '\P{rejang}', "");
Expect(1, 43359, '\P{^rejang}', "");
Expect(0, 43360, '\p{rejang}', "");
Expect(1, 43360, '\p{^rejang}', "");
Expect(1, 43360, '\P{rejang}', "");
Expect(0, 43360, '\P{^rejang}', "");
Expect(1, 43359, '\p{	rejang}', "");
Expect(0, 43359, '\p{^	rejang}', "");
Expect(0, 43359, '\P{	rejang}', "");
Expect(1, 43359, '\P{^	rejang}', "");
Expect(0, 43360, '\p{	rejang}', "");
Expect(1, 43360, '\p{^	rejang}', "");
Expect(1, 43360, '\P{	rejang}', "");
Expect(0, 43360, '\P{^	rejang}', "");
Error('\p{	/a/IS_Rejang}');
Error('\P{	/a/IS_Rejang}');
Expect(1, 43359, '\p{isrejang}', "");
Expect(0, 43359, '\p{^isrejang}', "");
Expect(0, 43359, '\P{isrejang}', "");
Expect(1, 43359, '\P{^isrejang}', "");
Expect(0, 43360, '\p{isrejang}', "");
Expect(1, 43360, '\p{^isrejang}', "");
Expect(1, 43360, '\P{isrejang}', "");
Expect(0, 43360, '\P{^isrejang}', "");
Expect(1, 43359, '\p{_Is_REJANG}', "");
Expect(0, 43359, '\p{^_Is_REJANG}', "");
Expect(0, 43359, '\P{_Is_REJANG}', "");
Expect(1, 43359, '\P{^_Is_REJANG}', "");
Expect(0, 43360, '\p{_Is_REJANG}', "");
Expect(1, 43360, '\p{^_Is_REJANG}', "");
Expect(1, 43360, '\P{_Is_REJANG}', "");
Expect(0, 43360, '\P{^_Is_REJANG}', "");
Error('\p{/a/--RJNG}');
Error('\P{/a/--RJNG}');
Expect(1, 43359, '\p{rjng}', "");
Expect(0, 43359, '\p{^rjng}', "");
Expect(0, 43359, '\P{rjng}', "");
Expect(1, 43359, '\P{^rjng}', "");
Expect(0, 43360, '\p{rjng}', "");
Expect(1, 43360, '\p{^rjng}', "");
Expect(1, 43360, '\P{rjng}', "");
Expect(0, 43360, '\P{^rjng}', "");
Expect(1, 43359, '\p{  RJNG}', "");
Expect(0, 43359, '\p{^  RJNG}', "");
Expect(0, 43359, '\P{  RJNG}', "");
Expect(1, 43359, '\P{^  RJNG}', "");
Expect(0, 43360, '\p{  RJNG}', "");
Expect(1, 43360, '\p{^  RJNG}', "");
Expect(1, 43360, '\P{  RJNG}', "");
Expect(0, 43360, '\P{^  RJNG}', "");
Error('\p{:=__is_Rjng}');
Error('\P{:=__is_Rjng}');
Expect(1, 43359, '\p{isrjng}', "");
Expect(0, 43359, '\p{^isrjng}', "");
Expect(0, 43359, '\P{isrjng}', "");
Expect(1, 43359, '\P{^isrjng}', "");
Expect(0, 43360, '\p{isrjng}', "");
Expect(1, 43360, '\p{^isrjng}', "");
Expect(1, 43360, '\P{isrjng}', "");
Expect(0, 43360, '\P{^isrjng}', "");
Expect(1, 43359, '\p{-	is_rjng}', "");
Expect(0, 43359, '\p{^-	is_rjng}', "");
Expect(0, 43359, '\P{-	is_rjng}', "");
Expect(1, 43359, '\P{^-	is_rjng}', "");
Expect(0, 43360, '\p{-	is_rjng}', "");
Expect(1, 43360, '\p{^-	is_rjng}', "");
Expect(1, 43360, '\P{-	is_rjng}', "");
Expect(0, 43360, '\P{^-	is_rjng}', "");
Error('\p{	:=UNIFIED_Canadian_Aboriginal_Syllabics_EXTENDED}');
Error('\P{	:=UNIFIED_Canadian_Aboriginal_Syllabics_EXTENDED}');
Expect(1, 6399, '\p{unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\p{^unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\P{unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\P{^unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\p{unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\p{^unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\P{unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\P{^unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\p{ _unified_Canadian_Aboriginal_Syllabics_EXTENDED}', "");
Expect(0, 6399, '\p{^ _unified_Canadian_Aboriginal_Syllabics_EXTENDED}', "");
Expect(0, 6399, '\P{ _unified_Canadian_Aboriginal_Syllabics_EXTENDED}', "");
Expect(1, 6399, '\P{^ _unified_Canadian_Aboriginal_Syllabics_EXTENDED}', "");
Expect(0, 6400, '\p{ _unified_Canadian_Aboriginal_Syllabics_EXTENDED}', "");
Expect(1, 6400, '\p{^ _unified_Canadian_Aboriginal_Syllabics_EXTENDED}', "");
Expect(1, 6400, '\P{ _unified_Canadian_Aboriginal_Syllabics_EXTENDED}', "");
Expect(0, 6400, '\P{^ _unified_Canadian_Aboriginal_Syllabics_EXTENDED}', "");
Error('\p{-:=Is_unified_Canadian_ABORIGINAL_Syllabics_extended}');
Error('\P{-:=Is_unified_Canadian_ABORIGINAL_Syllabics_extended}');
Expect(1, 6399, '\p{isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\p{^isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\P{isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\P{^isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\p{isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\p{^isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\P{isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\P{^isunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\p{	is_Unified_Canadian_ABORIGINAL_Syllabics_extended}', "");
Expect(0, 6399, '\p{^	is_Unified_Canadian_ABORIGINAL_Syllabics_extended}', "");
Expect(0, 6399, '\P{	is_Unified_Canadian_ABORIGINAL_Syllabics_extended}', "");
Expect(1, 6399, '\P{^	is_Unified_Canadian_ABORIGINAL_Syllabics_extended}', "");
Expect(0, 6400, '\p{	is_Unified_Canadian_ABORIGINAL_Syllabics_extended}', "");
Expect(1, 6400, '\p{^	is_Unified_Canadian_ABORIGINAL_Syllabics_extended}', "");
Expect(1, 6400, '\P{	is_Unified_Canadian_ABORIGINAL_Syllabics_extended}', "");
Expect(0, 6400, '\P{^	is_Unified_Canadian_ABORIGINAL_Syllabics_extended}', "");
Error('\p{_/a/in_unified_Canadian_Aboriginal_Syllabics_Extended}');
Error('\P{_/a/in_unified_Canadian_Aboriginal_Syllabics_Extended}');
Expect(1, 6399, '\p{inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\p{^inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\P{inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\P{^inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\p{inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\p{^inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\P{inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\P{^inunifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\p{In_Unified_Canadian_ABORIGINAL_SYLLABICS_Extended}', "");
Expect(0, 6399, '\p{^In_Unified_Canadian_ABORIGINAL_SYLLABICS_Extended}', "");
Expect(0, 6399, '\P{In_Unified_Canadian_ABORIGINAL_SYLLABICS_Extended}', "");
Expect(1, 6399, '\P{^In_Unified_Canadian_ABORIGINAL_SYLLABICS_Extended}', "");
Expect(0, 6400, '\p{In_Unified_Canadian_ABORIGINAL_SYLLABICS_Extended}', "");
Expect(1, 6400, '\p{^In_Unified_Canadian_ABORIGINAL_SYLLABICS_Extended}', "");
Expect(1, 6400, '\P{In_Unified_Canadian_ABORIGINAL_SYLLABICS_Extended}', "");
Expect(0, 6400, '\P{^In_Unified_Canadian_ABORIGINAL_SYLLABICS_Extended}', "");
Error('\p{_:=ucas_Ext}');
Error('\P{_:=ucas_Ext}');
Expect(1, 6399, '\p{ucasext}', "");
Expect(0, 6399, '\p{^ucasext}', "");
Expect(0, 6399, '\P{ucasext}', "");
Expect(1, 6399, '\P{^ucasext}', "");
Expect(0, 6400, '\p{ucasext}', "");
Expect(1, 6400, '\p{^ucasext}', "");
Expect(1, 6400, '\P{ucasext}', "");
Expect(0, 6400, '\P{^ucasext}', "");
Expect(1, 6399, '\p{_-ucas_Ext}', "");
Expect(0, 6399, '\p{^_-ucas_Ext}', "");
Expect(0, 6399, '\P{_-ucas_Ext}', "");
Expect(1, 6399, '\P{^_-ucas_Ext}', "");
Expect(0, 6400, '\p{_-ucas_Ext}', "");
Expect(1, 6400, '\p{^_-ucas_Ext}', "");
Expect(1, 6400, '\P{_-ucas_Ext}', "");
Expect(0, 6400, '\P{^_-ucas_Ext}', "");
Error('\p{/a/Is_UCAS_ext}');
Error('\P{/a/Is_UCAS_ext}');
Expect(1, 6399, '\p{isucasext}', "");
Expect(0, 6399, '\p{^isucasext}', "");
Expect(0, 6399, '\P{isucasext}', "");
Expect(1, 6399, '\P{^isucasext}', "");
Expect(0, 6400, '\p{isucasext}', "");
Expect(1, 6400, '\p{^isucasext}', "");
Expect(1, 6400, '\P{isucasext}', "");
Expect(0, 6400, '\P{^isucasext}', "");
Expect(1, 6399, '\p{  IS_ucas_Ext}', "");
Expect(0, 6399, '\p{^  IS_ucas_Ext}', "");
Expect(0, 6399, '\P{  IS_ucas_Ext}', "");
Expect(1, 6399, '\P{^  IS_ucas_Ext}', "");
Expect(0, 6400, '\p{  IS_ucas_Ext}', "");
Expect(1, 6400, '\p{^  IS_ucas_Ext}', "");
Expect(1, 6400, '\P{  IS_ucas_Ext}', "");
Expect(0, 6400, '\P{^  IS_ucas_Ext}', "");
Error('\p{	in_ucas_Ext:=}');
Error('\P{	in_ucas_Ext:=}');
Expect(1, 6399, '\p{inucasext}', "");
Expect(0, 6399, '\p{^inucasext}', "");
Expect(0, 6399, '\P{inucasext}', "");
Expect(1, 6399, '\P{^inucasext}', "");
Expect(0, 6400, '\p{inucasext}', "");
Expect(1, 6400, '\p{^inucasext}', "");
Expect(1, 6400, '\P{inucasext}', "");
Expect(0, 6400, '\P{^inucasext}', "");
Expect(1, 6399, '\p{ _in_UCAS_EXT}', "");
Expect(0, 6399, '\p{^ _in_UCAS_EXT}', "");
Expect(0, 6399, '\P{ _in_UCAS_EXT}', "");
Expect(1, 6399, '\P{^ _in_UCAS_EXT}', "");
Expect(0, 6400, '\p{ _in_UCAS_EXT}', "");
Expect(1, 6400, '\p{^ _in_UCAS_EXT}', "");
Expect(1, 6400, '\P{ _in_UCAS_EXT}', "");
Expect(0, 6400, '\P{^ _in_UCAS_EXT}', "");
Error('\p{-:=GLAGOLITIC}');
Error('\P{-:=GLAGOLITIC}');
Expect(1, 11358, '\p{glagolitic}', "");
Expect(0, 11358, '\p{^glagolitic}', "");
Expect(0, 11358, '\P{glagolitic}', "");
Expect(1, 11358, '\P{^glagolitic}', "");
Expect(0, 11359, '\p{glagolitic}', "");
Expect(1, 11359, '\p{^glagolitic}', "");
Expect(1, 11359, '\P{glagolitic}', "");
Expect(0, 11359, '\P{^glagolitic}', "");
Expect(1, 11358, '\p{_Glagolitic}', "");
Expect(0, 11358, '\p{^_Glagolitic}', "");
Expect(0, 11358, '\P{_Glagolitic}', "");
Expect(1, 11358, '\P{^_Glagolitic}', "");
Expect(0, 11359, '\p{_Glagolitic}', "");
Expect(1, 11359, '\p{^_Glagolitic}', "");
Expect(1, 11359, '\P{_Glagolitic}', "");
Expect(0, 11359, '\P{^_Glagolitic}', "");
Error('\p{:=-is_Glagolitic}');
Error('\P{:=-is_Glagolitic}');
Expect(1, 11358, '\p{isglagolitic}', "");
Expect(0, 11358, '\p{^isglagolitic}', "");
Expect(0, 11358, '\P{isglagolitic}', "");
Expect(1, 11358, '\P{^isglagolitic}', "");
Expect(0, 11359, '\p{isglagolitic}', "");
Expect(1, 11359, '\p{^isglagolitic}', "");
Expect(1, 11359, '\P{isglagolitic}', "");
Expect(0, 11359, '\P{^isglagolitic}', "");
Expect(1, 11358, '\p{  is_glagolitic}', "");
Expect(0, 11358, '\p{^  is_glagolitic}', "");
Expect(0, 11358, '\P{  is_glagolitic}', "");
Expect(1, 11358, '\P{^  is_glagolitic}', "");
Expect(0, 11359, '\p{  is_glagolitic}', "");
Expect(1, 11359, '\p{^  is_glagolitic}', "");
Expect(1, 11359, '\P{  is_glagolitic}', "");
Expect(0, 11359, '\P{^  is_glagolitic}', "");
Error('\p{:= glag}');
Error('\P{:= glag}');
Expect(1, 11358, '\p{glag}', "");
Expect(0, 11358, '\p{^glag}', "");
Expect(0, 11358, '\P{glag}', "");
Expect(1, 11358, '\P{^glag}', "");
Expect(0, 11359, '\p{glag}', "");
Expect(1, 11359, '\p{^glag}', "");
Expect(1, 11359, '\P{glag}', "");
Expect(0, 11359, '\P{^glag}', "");
Expect(1, 11358, '\p{ Glag}', "");
Expect(0, 11358, '\p{^ Glag}', "");
Expect(0, 11358, '\P{ Glag}', "");
Expect(1, 11358, '\P{^ Glag}', "");
Expect(0, 11359, '\p{ Glag}', "");
Expect(1, 11359, '\p{^ Glag}', "");
Expect(1, 11359, '\P{ Glag}', "");
Expect(0, 11359, '\P{^ Glag}', "");
Error('\p{ _IS_Glag/a/}');
Error('\P{ _IS_Glag/a/}');
Expect(1, 11358, '\p{isglag}', "");
Expect(0, 11358, '\p{^isglag}', "");
Expect(0, 11358, '\P{isglag}', "");
Expect(1, 11358, '\P{^isglag}', "");
Expect(0, 11359, '\p{isglag}', "");
Expect(1, 11359, '\p{^isglag}', "");
Expect(1, 11359, '\P{isglag}', "");
Expect(0, 11359, '\P{^isglag}', "");
Expect(1, 11358, '\p{Is_Glag}', "");
Expect(0, 11358, '\p{^Is_Glag}', "");
Expect(0, 11358, '\P{Is_Glag}', "");
Expect(1, 11358, '\P{^Is_Glag}', "");
Expect(0, 11359, '\p{Is_Glag}', "");
Expect(1, 11359, '\p{^Is_Glag}', "");
Expect(1, 11359, '\P{Is_Glag}', "");
Expect(0, 11359, '\P{^Is_Glag}', "");
Error('\p{_:=in_Ol_CHIKI}');
Error('\P{_:=in_Ol_CHIKI}');
Expect(1, 7295, '\p{inolchiki}', "");
Expect(0, 7295, '\p{^inolchiki}', "");
Expect(0, 7295, '\P{inolchiki}', "");
Expect(1, 7295, '\P{^inolchiki}', "");
Expect(0, 7296, '\p{inolchiki}', "");
Expect(1, 7296, '\p{^inolchiki}', "");
Expect(1, 7296, '\P{inolchiki}', "");
Expect(0, 7296, '\P{^inolchiki}', "");
Expect(1, 7295, '\p{--IN_Ol_Chiki}', "");
Expect(0, 7295, '\p{^--IN_Ol_Chiki}', "");
Expect(0, 7295, '\P{--IN_Ol_Chiki}', "");
Expect(1, 7295, '\P{^--IN_Ol_Chiki}', "");
Expect(0, 7296, '\p{--IN_Ol_Chiki}', "");
Expect(1, 7296, '\p{^--IN_Ol_Chiki}', "");
Expect(1, 7296, '\P{--IN_Ol_Chiki}', "");
Expect(0, 7296, '\P{^--IN_Ol_Chiki}', "");
Error('\p{/a/	_IN_OL_Chiki}');
Error('\P{/a/	_IN_OL_Chiki}');
Expect(1, 7295, '\p{	IN_OL_chiki}', "");
Expect(0, 7295, '\p{^	IN_OL_chiki}', "");
Expect(0, 7295, '\P{	IN_OL_chiki}', "");
Expect(1, 7295, '\P{^	IN_OL_chiki}', "");
Expect(0, 7296, '\p{	IN_OL_chiki}', "");
Expect(1, 7296, '\p{^	IN_OL_chiki}', "");
Expect(1, 7296, '\P{	IN_OL_chiki}', "");
Expect(0, 7296, '\P{^	IN_OL_chiki}', "");
Error('\p{/a/CJK_Strokes}');
Error('\P{/a/CJK_Strokes}');
Expect(1, 12783, '\p{cjkstrokes}', "");
Expect(0, 12783, '\p{^cjkstrokes}', "");
Expect(0, 12783, '\P{cjkstrokes}', "");
Expect(1, 12783, '\P{^cjkstrokes}', "");
Expect(0, 12784, '\p{cjkstrokes}', "");
Expect(1, 12784, '\p{^cjkstrokes}', "");
Expect(1, 12784, '\P{cjkstrokes}', "");
Expect(0, 12784, '\P{^cjkstrokes}', "");
Expect(1, 12783, '\p{_ CJK_Strokes}', "");
Expect(0, 12783, '\p{^_ CJK_Strokes}', "");
Expect(0, 12783, '\P{_ CJK_Strokes}', "");
Expect(1, 12783, '\P{^_ CJK_Strokes}', "");
Expect(0, 12784, '\p{_ CJK_Strokes}', "");
Expect(1, 12784, '\p{^_ CJK_Strokes}', "");
Expect(1, 12784, '\P{_ CJK_Strokes}', "");
Expect(0, 12784, '\P{^_ CJK_Strokes}', "");
Error('\p{-:=is_CJK_Strokes}');
Error('\P{-:=is_CJK_Strokes}');
Expect(1, 12783, '\p{iscjkstrokes}', "");
Expect(0, 12783, '\p{^iscjkstrokes}', "");
Expect(0, 12783, '\P{iscjkstrokes}', "");
Expect(1, 12783, '\P{^iscjkstrokes}', "");
Expect(0, 12784, '\p{iscjkstrokes}', "");
Expect(1, 12784, '\p{^iscjkstrokes}', "");
Expect(1, 12784, '\P{iscjkstrokes}', "");
Expect(0, 12784, '\P{^iscjkstrokes}', "");
Expect(1, 12783, '\p{- Is_CJK_Strokes}', "");
Expect(0, 12783, '\p{^- Is_CJK_Strokes}', "");
Expect(0, 12783, '\P{- Is_CJK_Strokes}', "");
Expect(1, 12783, '\P{^- Is_CJK_Strokes}', "");
Expect(0, 12784, '\p{- Is_CJK_Strokes}', "");
Expect(1, 12784, '\p{^- Is_CJK_Strokes}', "");
Expect(1, 12784, '\P{- Is_CJK_Strokes}', "");
Expect(0, 12784, '\P{^- Is_CJK_Strokes}', "");
Error('\p{_ IN_CJK_STROKES:=}');
Error('\P{_ IN_CJK_STROKES:=}');
Expect(1, 12783, '\p{incjkstrokes}', "");
Expect(0, 12783, '\p{^incjkstrokes}', "");
Expect(0, 12783, '\P{incjkstrokes}', "");
Expect(1, 12783, '\P{^incjkstrokes}', "");
Expect(0, 12784, '\p{incjkstrokes}', "");
Expect(1, 12784, '\p{^incjkstrokes}', "");
Expect(1, 12784, '\P{incjkstrokes}', "");
Expect(0, 12784, '\P{^incjkstrokes}', "");
Expect(1, 12783, '\p{_In_CJK_Strokes}', "");
Expect(0, 12783, '\p{^_In_CJK_Strokes}', "");
Expect(0, 12783, '\P{_In_CJK_Strokes}', "");
Expect(1, 12783, '\P{^_In_CJK_Strokes}', "");
Expect(0, 12784, '\p{_In_CJK_Strokes}', "");
Expect(1, 12784, '\p{^_In_CJK_Strokes}', "");
Expect(1, 12784, '\P{_In_CJK_Strokes}', "");
Expect(0, 12784, '\P{^_In_CJK_Strokes}', "");
Error('\p{	_In_Mandaic/a/}');
Error('\P{	_In_Mandaic/a/}');
Expect(1, 2143, '\p{inmandaic}', "");
Expect(0, 2143, '\p{^inmandaic}', "");
Expect(0, 2143, '\P{inmandaic}', "");
Expect(1, 2143, '\P{^inmandaic}', "");
Expect(0, 2144, '\p{inmandaic}', "");
Expect(1, 2144, '\p{^inmandaic}', "");
Expect(1, 2144, '\P{inmandaic}', "");
Expect(0, 2144, '\P{^inmandaic}', "");
Expect(1, 2143, '\p{  In_Mandaic}', "");
Expect(0, 2143, '\p{^  In_Mandaic}', "");
Expect(0, 2143, '\P{  In_Mandaic}', "");
Expect(1, 2143, '\P{^  In_Mandaic}', "");
Expect(0, 2144, '\p{  In_Mandaic}', "");
Expect(1, 2144, '\p{^  In_Mandaic}', "");
Expect(1, 2144, '\P{  In_Mandaic}', "");
Expect(0, 2144, '\P{^  In_Mandaic}', "");
Error('\p{ -In_Mandaic/a/}');
Error('\P{ -In_Mandaic/a/}');
Expect(1, 2143, '\p{  IN_MANDAIC}', "");
Expect(0, 2143, '\p{^  IN_MANDAIC}', "");
Expect(0, 2143, '\P{  IN_MANDAIC}', "");
Expect(1, 2143, '\P{^  IN_MANDAIC}', "");
Expect(0, 2144, '\p{  IN_MANDAIC}', "");
Expect(1, 2144, '\p{^  IN_MANDAIC}', "");
Expect(1, 2144, '\P{  IN_MANDAIC}', "");
Expect(0, 2144, '\P{^  IN_MANDAIC}', "");
Error('\p{- Sinhala:=}');
Error('\P{- Sinhala:=}');
Expect(1, 3572, '\p{sinhala}', "");
Expect(0, 3572, '\p{^sinhala}', "");
Expect(0, 3572, '\P{sinhala}', "");
Expect(1, 3572, '\P{^sinhala}', "");
Expect(0, 3573, '\p{sinhala}', "");
Expect(1, 3573, '\p{^sinhala}', "");
Expect(1, 3573, '\P{sinhala}', "");
Expect(0, 3573, '\P{^sinhala}', "");
Expect(1, 3572, '\p{-Sinhala}', "");
Expect(0, 3572, '\p{^-Sinhala}', "");
Expect(0, 3572, '\P{-Sinhala}', "");
Expect(1, 3572, '\P{^-Sinhala}', "");
Expect(0, 3573, '\p{-Sinhala}', "");
Expect(1, 3573, '\p{^-Sinhala}', "");
Expect(1, 3573, '\P{-Sinhala}', "");
Expect(0, 3573, '\P{^-Sinhala}', "");
Error('\p{:=		is_SINHALA}');
Error('\P{:=		is_SINHALA}');
Expect(1, 3572, '\p{issinhala}', "");
Expect(0, 3572, '\p{^issinhala}', "");
Expect(0, 3572, '\P{issinhala}', "");
Expect(1, 3572, '\P{^issinhala}', "");
Expect(0, 3573, '\p{issinhala}', "");
Expect(1, 3573, '\p{^issinhala}', "");
Expect(1, 3573, '\P{issinhala}', "");
Expect(0, 3573, '\P{^issinhala}', "");
Expect(1, 3572, '\p{__IS_sinhala}', "");
Expect(0, 3572, '\p{^__IS_sinhala}', "");
Expect(0, 3572, '\P{__IS_sinhala}', "");
Expect(1, 3572, '\P{^__IS_sinhala}', "");
Expect(0, 3573, '\p{__IS_sinhala}', "");
Expect(1, 3573, '\p{^__IS_sinhala}', "");
Expect(1, 3573, '\P{__IS_sinhala}', "");
Expect(0, 3573, '\P{^__IS_sinhala}', "");
Error('\p{_:=Sinh}');
Error('\P{_:=Sinh}');
Expect(1, 3572, '\p{sinh}', "");
Expect(0, 3572, '\p{^sinh}', "");
Expect(0, 3572, '\P{sinh}', "");
Expect(1, 3572, '\P{^sinh}', "");
Expect(0, 3573, '\p{sinh}', "");
Expect(1, 3573, '\p{^sinh}', "");
Expect(1, 3573, '\P{sinh}', "");
Expect(0, 3573, '\P{^sinh}', "");
Expect(1, 3572, '\p{__SINH}', "");
Expect(0, 3572, '\p{^__SINH}', "");
Expect(0, 3572, '\P{__SINH}', "");
Expect(1, 3572, '\P{^__SINH}', "");
Expect(0, 3573, '\p{__SINH}', "");
Expect(1, 3573, '\p{^__SINH}', "");
Expect(1, 3573, '\P{__SINH}', "");
Expect(0, 3573, '\P{^__SINH}', "");
Error('\p{/a/_	Is_Sinh}');
Error('\P{/a/_	Is_Sinh}');
Expect(1, 3572, '\p{issinh}', "");
Expect(0, 3572, '\p{^issinh}', "");
Expect(0, 3572, '\P{issinh}', "");
Expect(1, 3572, '\P{^issinh}', "");
Expect(0, 3573, '\p{issinh}', "");
Expect(1, 3573, '\p{^issinh}', "");
Expect(1, 3573, '\P{issinh}', "");
Expect(0, 3573, '\P{^issinh}', "");
Expect(1, 3572, '\p{ 	is_SINH}', "");
Expect(0, 3572, '\p{^ 	is_SINH}', "");
Expect(0, 3572, '\P{ 	is_SINH}', "");
Expect(1, 3572, '\P{^ 	is_SINH}', "");
Expect(0, 3573, '\p{ 	is_SINH}', "");
Expect(1, 3573, '\p{^ 	is_SINH}', "");
Expect(1, 3573, '\P{ 	is_SINH}', "");
Expect(0, 3573, '\P{^ 	is_SINH}', "");
Error('\p{	:=_Perl_IDStart}');
Error('\P{	:=_Perl_IDStart}');
Expect(1, 195101, '\p{_Perl_IDStart}', "");
Expect(0, 195101, '\p{^_Perl_IDStart}', "");
Expect(0, 195101, '\P{_Perl_IDStart}', "");
Expect(1, 195101, '\P{^_Perl_IDStart}', "");
Expect(0, 195102, '\p{_Perl_IDStart}', "");
Expect(1, 195102, '\p{^_Perl_IDStart}', "");
Expect(1, 195102, '\P{_Perl_IDStart}', "");
Expect(0, 195102, '\P{^_Perl_IDStart}', "");
Error('\p{/a/__PERL_IDSTART}');
Error('\P{/a/__PERL_IDSTART}');
Error('\p{_/a/ugaritic}');
Error('\P{_/a/ugaritic}');
Expect(1, 66463, '\p{ugaritic}', "");
Expect(0, 66463, '\p{^ugaritic}', "");
Expect(0, 66463, '\P{ugaritic}', "");
Expect(1, 66463, '\P{^ugaritic}', "");
Expect(0, 66464, '\p{ugaritic}', "");
Expect(1, 66464, '\p{^ugaritic}', "");
Expect(1, 66464, '\P{ugaritic}', "");
Expect(0, 66464, '\P{^ugaritic}', "");
Expect(1, 66463, '\p{_ugaritic}', "");
Expect(0, 66463, '\p{^_ugaritic}', "");
Expect(0, 66463, '\P{_ugaritic}', "");
Expect(1, 66463, '\P{^_ugaritic}', "");
Expect(0, 66464, '\p{_ugaritic}', "");
Expect(1, 66464, '\p{^_ugaritic}', "");
Expect(1, 66464, '\P{_ugaritic}', "");
Expect(0, 66464, '\P{^_ugaritic}', "");
Error('\p{/a/--Is_Ugaritic}');
Error('\P{/a/--Is_Ugaritic}');
Expect(1, 66463, '\p{isugaritic}', "");
Expect(0, 66463, '\p{^isugaritic}', "");
Expect(0, 66463, '\P{isugaritic}', "");
Expect(1, 66463, '\P{^isugaritic}', "");
Expect(0, 66464, '\p{isugaritic}', "");
Expect(1, 66464, '\p{^isugaritic}', "");
Expect(1, 66464, '\P{isugaritic}', "");
Expect(0, 66464, '\P{^isugaritic}', "");
Expect(1, 66463, '\p{	_Is_Ugaritic}', "");
Expect(0, 66463, '\p{^	_Is_Ugaritic}', "");
Expect(0, 66463, '\P{	_Is_Ugaritic}', "");
Expect(1, 66463, '\P{^	_Is_Ugaritic}', "");
Expect(0, 66464, '\p{	_Is_Ugaritic}', "");
Expect(1, 66464, '\p{^	_Is_Ugaritic}', "");
Expect(1, 66464, '\P{	_Is_Ugaritic}', "");
Expect(0, 66464, '\P{^	_Is_Ugaritic}', "");
Error('\p{-:=Ugar}');
Error('\P{-:=Ugar}');
Expect(1, 66463, '\p{ugar}', "");
Expect(0, 66463, '\p{^ugar}', "");
Expect(0, 66463, '\P{ugar}', "");
Expect(1, 66463, '\P{^ugar}', "");
Expect(0, 66464, '\p{ugar}', "");
Expect(1, 66464, '\p{^ugar}', "");
Expect(1, 66464, '\P{ugar}', "");
Expect(0, 66464, '\P{^ugar}', "");
Expect(1, 66463, '\p{_	Ugar}', "");
Expect(0, 66463, '\p{^_	Ugar}', "");
Expect(0, 66463, '\P{_	Ugar}', "");
Expect(1, 66463, '\P{^_	Ugar}', "");
Expect(0, 66464, '\p{_	Ugar}', "");
Expect(1, 66464, '\p{^_	Ugar}', "");
Expect(1, 66464, '\P{_	Ugar}', "");
Expect(0, 66464, '\P{^_	Ugar}', "");
Error('\p{	 Is_Ugar:=}');
Error('\P{	 Is_Ugar:=}');
Expect(1, 66463, '\p{isugar}', "");
Expect(0, 66463, '\p{^isugar}', "");
Expect(0, 66463, '\P{isugar}', "");
Expect(1, 66463, '\P{^isugar}', "");
Expect(0, 66464, '\p{isugar}', "");
Expect(1, 66464, '\p{^isugar}', "");
Expect(1, 66464, '\P{isugar}', "");
Expect(0, 66464, '\P{^isugar}', "");
Expect(1, 66463, '\p{ IS_Ugar}', "");
Expect(0, 66463, '\p{^ IS_Ugar}', "");
Expect(0, 66463, '\P{ IS_Ugar}', "");
Expect(1, 66463, '\P{^ IS_Ugar}', "");
Expect(0, 66464, '\p{ IS_Ugar}', "");
Expect(1, 66464, '\p{^ IS_Ugar}', "");
Expect(1, 66464, '\P{ IS_Ugar}', "");
Expect(0, 66464, '\P{^ IS_Ugar}', "");
Error('\p{XPERLSPACE:=}');
Error('\P{XPERLSPACE:=}');
Expect(1, 12288, '\p{xperlspace}', "");
Expect(0, 12288, '\p{^xperlspace}', "");
Expect(0, 12288, '\P{xperlspace}', "");
Expect(1, 12288, '\P{^xperlspace}', "");
Expect(0, 12289, '\p{xperlspace}', "");
Expect(1, 12289, '\p{^xperlspace}', "");
Expect(1, 12289, '\P{xperlspace}', "");
Expect(0, 12289, '\P{^xperlspace}', "");
Error('\p{	 SPACEPERL:=}');
Error('\P{	 SPACEPERL:=}');
Expect(1, 12288, '\p{spaceperl}', "");
Expect(0, 12288, '\p{^spaceperl}', "");
Expect(0, 12288, '\P{spaceperl}', "");
Expect(1, 12288, '\P{^spaceperl}', "");
Expect(0, 12289, '\p{spaceperl}', "");
Expect(1, 12289, '\p{^spaceperl}', "");
Expect(1, 12289, '\P{spaceperl}', "");
Expect(0, 12289, '\P{^spaceperl}', "");
Expect(1, 12288, '\p{-	SpacePerl}', "");
Expect(0, 12288, '\p{^-	SpacePerl}', "");
Expect(0, 12288, '\P{-	SpacePerl}', "");
Expect(1, 12288, '\P{^-	SpacePerl}', "");
Expect(0, 12289, '\p{-	SpacePerl}', "");
Expect(1, 12289, '\p{^-	SpacePerl}', "");
Expect(1, 12289, '\P{-	SpacePerl}', "");
Expect(0, 12289, '\P{^-	SpacePerl}', "");
Error('\p{/a/is_XPerlSpace}');
Error('\P{/a/is_XPerlSpace}');
Expect(1, 12288, '\p{isxperlspace}', "");
Expect(0, 12288, '\p{^isxperlspace}', "");
Expect(0, 12288, '\P{isxperlspace}', "");
Expect(1, 12288, '\P{^isxperlspace}', "");
Expect(0, 12289, '\p{isxperlspace}', "");
Expect(1, 12289, '\p{^isxperlspace}', "");
Expect(1, 12289, '\P{isxperlspace}', "");
Expect(0, 12289, '\P{^isxperlspace}', "");
Expect(1, 12288, '\p{	_IS_XPERLSPACE}', "");
Expect(0, 12288, '\p{^	_IS_XPERLSPACE}', "");
Expect(0, 12288, '\P{	_IS_XPERLSPACE}', "");
Expect(1, 12288, '\P{^	_IS_XPERLSPACE}', "");
Expect(0, 12289, '\p{	_IS_XPERLSPACE}', "");
Expect(1, 12289, '\p{^	_IS_XPERLSPACE}', "");
Expect(1, 12289, '\P{	_IS_XPERLSPACE}', "");
Expect(0, 12289, '\P{^	_IS_XPERLSPACE}', "");
Error('\p{:=__Is_SPACEPERL}');
Error('\P{:=__Is_SPACEPERL}');
Expect(1, 12288, '\p{isspaceperl}', "");
Expect(0, 12288, '\p{^isspaceperl}', "");
Expect(0, 12288, '\P{isspaceperl}', "");
Expect(1, 12288, '\P{^isspaceperl}', "");
Expect(0, 12289, '\p{isspaceperl}', "");
Expect(1, 12289, '\p{^isspaceperl}', "");
Expect(1, 12289, '\P{isspaceperl}', "");
Expect(0, 12289, '\P{^isspaceperl}', "");
Expect(1, 12288, '\p{_	Is_SpacePerl}', "");
Expect(0, 12288, '\p{^_	Is_SpacePerl}', "");
Expect(0, 12288, '\P{_	Is_SpacePerl}', "");
Expect(1, 12288, '\P{^_	Is_SpacePerl}', "");
Expect(0, 12289, '\p{_	Is_SpacePerl}', "");
Expect(1, 12289, '\p{^_	Is_SpacePerl}', "");
Expect(1, 12289, '\P{_	Is_SpacePerl}', "");
Expect(0, 12289, '\P{^_	Is_SpacePerl}', "");
Error('\p{--ogham/a/}');
Error('\P{--ogham/a/}');
Expect(1, 5788, '\p{ogham}', "");
Expect(0, 5788, '\p{^ogham}', "");
Expect(0, 5788, '\P{ogham}', "");
Expect(1, 5788, '\P{^ogham}', "");
Expect(0, 5789, '\p{ogham}', "");
Expect(1, 5789, '\p{^ogham}', "");
Expect(1, 5789, '\P{ogham}', "");
Expect(0, 5789, '\P{^ogham}', "");
Expect(1, 5788, '\p{ 	Ogham}', "");
Expect(0, 5788, '\p{^ 	Ogham}', "");
Expect(0, 5788, '\P{ 	Ogham}', "");
Expect(1, 5788, '\P{^ 	Ogham}', "");
Expect(0, 5789, '\p{ 	Ogham}', "");
Expect(1, 5789, '\p{^ 	Ogham}', "");
Expect(1, 5789, '\P{ 	Ogham}', "");
Expect(0, 5789, '\P{^ 	Ogham}', "");
Error('\p{is_Ogham:=}');
Error('\P{is_Ogham:=}');
Expect(1, 5788, '\p{isogham}', "");
Expect(0, 5788, '\p{^isogham}', "");
Expect(0, 5788, '\P{isogham}', "");
Expect(1, 5788, '\P{^isogham}', "");
Expect(0, 5789, '\p{isogham}', "");
Expect(1, 5789, '\p{^isogham}', "");
Expect(1, 5789, '\P{isogham}', "");
Expect(0, 5789, '\P{^isogham}', "");
Expect(1, 5788, '\p{-IS_Ogham}', "");
Expect(0, 5788, '\p{^-IS_Ogham}', "");
Expect(0, 5788, '\P{-IS_Ogham}', "");
Expect(1, 5788, '\P{^-IS_Ogham}', "");
Expect(0, 5789, '\p{-IS_Ogham}', "");
Expect(1, 5789, '\p{^-IS_Ogham}', "");
Expect(1, 5789, '\P{-IS_Ogham}', "");
Expect(0, 5789, '\P{^-IS_Ogham}', "");
Error('\p{_:=Ogam}');
Error('\P{_:=Ogam}');
Expect(1, 5788, '\p{ogam}', "");
Expect(0, 5788, '\p{^ogam}', "");
Expect(0, 5788, '\P{ogam}', "");
Expect(1, 5788, '\P{^ogam}', "");
Expect(0, 5789, '\p{ogam}', "");
Expect(1, 5789, '\p{^ogam}', "");
Expect(1, 5789, '\P{ogam}', "");
Expect(0, 5789, '\P{^ogam}', "");
Expect(1, 5788, '\p{__OGAM}', "");
Expect(0, 5788, '\p{^__OGAM}', "");
Expect(0, 5788, '\P{__OGAM}', "");
Expect(1, 5788, '\P{^__OGAM}', "");
Expect(0, 5789, '\p{__OGAM}', "");
Expect(1, 5789, '\p{^__OGAM}', "");
Expect(1, 5789, '\P{__OGAM}', "");
Expect(0, 5789, '\P{^__OGAM}', "");
Error('\p{ 	Is_Ogam:=}');
Error('\P{ 	Is_Ogam:=}');
Expect(1, 5788, '\p{isogam}', "");
Expect(0, 5788, '\p{^isogam}', "");
Expect(0, 5788, '\P{isogam}', "");
Expect(1, 5788, '\P{^isogam}', "");
Expect(0, 5789, '\p{isogam}', "");
Expect(1, 5789, '\p{^isogam}', "");
Expect(1, 5789, '\P{isogam}', "");
Expect(0, 5789, '\P{^isogam}', "");
Expect(1, 5788, '\p{__Is_OGAM}', "");
Expect(0, 5788, '\p{^__Is_OGAM}', "");
Expect(0, 5788, '\P{__Is_OGAM}', "");
Expect(1, 5788, '\P{^__Is_OGAM}', "");
Expect(0, 5789, '\p{__Is_OGAM}', "");
Expect(1, 5789, '\p{^__Is_OGAM}', "");
Expect(1, 5789, '\P{__Is_OGAM}', "");
Expect(0, 5789, '\P{^__Is_OGAM}', "");
Error('\p{/a/-	BENGALI}');
Error('\P{/a/-	BENGALI}');
Expect(1, 2555, '\p{bengali}', "");
Expect(0, 2555, '\p{^bengali}', "");
Expect(0, 2555, '\P{bengali}', "");
Expect(1, 2555, '\P{^bengali}', "");
Expect(0, 2556, '\p{bengali}', "");
Expect(1, 2556, '\p{^bengali}', "");
Expect(1, 2556, '\P{bengali}', "");
Expect(0, 2556, '\P{^bengali}', "");
Expect(1, 2555, '\p{_-bengali}', "");
Expect(0, 2555, '\p{^_-bengali}', "");
Expect(0, 2555, '\P{_-bengali}', "");
Expect(1, 2555, '\P{^_-bengali}', "");
Expect(0, 2556, '\p{_-bengali}', "");
Expect(1, 2556, '\p{^_-bengali}', "");
Expect(1, 2556, '\P{_-bengali}', "");
Expect(0, 2556, '\P{^_-bengali}', "");
Error('\p{	is_Bengali/a/}');
Error('\P{	is_Bengali/a/}');
Expect(1, 2555, '\p{isbengali}', "");
Expect(0, 2555, '\p{^isbengali}', "");
Expect(0, 2555, '\P{isbengali}', "");
Expect(1, 2555, '\P{^isbengali}', "");
Expect(0, 2556, '\p{isbengali}', "");
Expect(1, 2556, '\p{^isbengali}', "");
Expect(1, 2556, '\P{isbengali}', "");
Expect(0, 2556, '\P{^isbengali}', "");
Expect(1, 2555, '\p{-_Is_bengali}', "");
Expect(0, 2555, '\p{^-_Is_bengali}', "");
Expect(0, 2555, '\P{-_Is_bengali}', "");
Expect(1, 2555, '\P{^-_Is_bengali}', "");
Expect(0, 2556, '\p{-_Is_bengali}', "");
Expect(1, 2556, '\p{^-_Is_bengali}', "");
Expect(1, 2556, '\P{-_Is_bengali}', "");
Expect(0, 2556, '\P{^-_Is_bengali}', "");
Error('\p{	_BENG/a/}');
Error('\P{	_BENG/a/}');
Expect(1, 2555, '\p{beng}', "");
Expect(0, 2555, '\p{^beng}', "");
Expect(0, 2555, '\P{beng}', "");
Expect(1, 2555, '\P{^beng}', "");
Expect(0, 2556, '\p{beng}', "");
Expect(1, 2556, '\p{^beng}', "");
Expect(1, 2556, '\P{beng}', "");
Expect(0, 2556, '\P{^beng}', "");
Expect(1, 2555, '\p{--beng}', "");
Expect(0, 2555, '\p{^--beng}', "");
Expect(0, 2555, '\P{--beng}', "");
Expect(1, 2555, '\P{^--beng}', "");
Expect(0, 2556, '\p{--beng}', "");
Expect(1, 2556, '\p{^--beng}', "");
Expect(1, 2556, '\P{--beng}', "");
Expect(0, 2556, '\P{^--beng}', "");
Error('\p{:=-_Is_Beng}');
Error('\P{:=-_Is_Beng}');
Expect(1, 2555, '\p{isbeng}', "");
Expect(0, 2555, '\p{^isbeng}', "");
Expect(0, 2555, '\P{isbeng}', "");
Expect(1, 2555, '\P{^isbeng}', "");
Expect(0, 2556, '\p{isbeng}', "");
Expect(1, 2556, '\p{^isbeng}', "");
Expect(1, 2556, '\P{isbeng}', "");
Expect(0, 2556, '\P{^isbeng}', "");
Expect(1, 2555, '\p{__is_Beng}', "");
Expect(0, 2555, '\p{^__is_Beng}', "");
Expect(0, 2555, '\P{__is_Beng}', "");
Expect(1, 2555, '\P{^__is_Beng}', "");
Expect(0, 2556, '\p{__is_Beng}', "");
Expect(1, 2556, '\p{^__is_Beng}', "");
Expect(1, 2556, '\P{__is_Beng}', "");
Expect(0, 2556, '\P{^__is_Beng}', "");
Error('\p{	 Hangul_COMPATIBILITY_JAMO/a/}');
Error('\P{	 Hangul_COMPATIBILITY_JAMO/a/}');
Expect(1, 12687, '\p{hangulcompatibilityjamo}', "");
Expect(0, 12687, '\p{^hangulcompatibilityjamo}', "");
Expect(0, 12687, '\P{hangulcompatibilityjamo}', "");
Expect(1, 12687, '\P{^hangulcompatibilityjamo}', "");
Expect(0, 12688, '\p{hangulcompatibilityjamo}', "");
Expect(1, 12688, '\p{^hangulcompatibilityjamo}', "");
Expect(1, 12688, '\P{hangulcompatibilityjamo}', "");
Expect(0, 12688, '\P{^hangulcompatibilityjamo}', "");
Expect(1, 12687, '\p{_HANGUL_Compatibility_Jamo}', "");
Expect(0, 12687, '\p{^_HANGUL_Compatibility_Jamo}', "");
Expect(0, 12687, '\P{_HANGUL_Compatibility_Jamo}', "");
Expect(1, 12687, '\P{^_HANGUL_Compatibility_Jamo}', "");
Expect(0, 12688, '\p{_HANGUL_Compatibility_Jamo}', "");
Expect(1, 12688, '\p{^_HANGUL_Compatibility_Jamo}', "");
Expect(1, 12688, '\P{_HANGUL_Compatibility_Jamo}', "");
Expect(0, 12688, '\P{^_HANGUL_Compatibility_Jamo}', "");
Error('\p{/a/_	is_hangul_compatibility_jamo}');
Error('\P{/a/_	is_hangul_compatibility_jamo}');
Expect(1, 12687, '\p{ishangulcompatibilityjamo}', "");
Expect(0, 12687, '\p{^ishangulcompatibilityjamo}', "");
Expect(0, 12687, '\P{ishangulcompatibilityjamo}', "");
Expect(1, 12687, '\P{^ishangulcompatibilityjamo}', "");
Expect(0, 12688, '\p{ishangulcompatibilityjamo}', "");
Expect(1, 12688, '\p{^ishangulcompatibilityjamo}', "");
Expect(1, 12688, '\P{ishangulcompatibilityjamo}', "");
Expect(0, 12688, '\P{^ishangulcompatibilityjamo}', "");
Expect(1, 12687, '\p{ Is_hangul_COMPATIBILITY_jamo}', "");
Expect(0, 12687, '\p{^ Is_hangul_COMPATIBILITY_jamo}', "");
Expect(0, 12687, '\P{ Is_hangul_COMPATIBILITY_jamo}', "");
Expect(1, 12687, '\P{^ Is_hangul_COMPATIBILITY_jamo}', "");
Expect(0, 12688, '\p{ Is_hangul_COMPATIBILITY_jamo}', "");
Expect(1, 12688, '\p{^ Is_hangul_COMPATIBILITY_jamo}', "");
Expect(1, 12688, '\P{ Is_hangul_COMPATIBILITY_jamo}', "");
Expect(0, 12688, '\P{^ Is_hangul_COMPATIBILITY_jamo}', "");
Error('\p{/a/In_Hangul_compatibility_Jamo}');
Error('\P{/a/In_Hangul_compatibility_Jamo}');
Expect(1, 12687, '\p{inhangulcompatibilityjamo}', "");
Expect(0, 12687, '\p{^inhangulcompatibilityjamo}', "");
Expect(0, 12687, '\P{inhangulcompatibilityjamo}', "");
Expect(1, 12687, '\P{^inhangulcompatibilityjamo}', "");
Expect(0, 12688, '\p{inhangulcompatibilityjamo}', "");
Expect(1, 12688, '\p{^inhangulcompatibilityjamo}', "");
Expect(1, 12688, '\P{inhangulcompatibilityjamo}', "");
Expect(0, 12688, '\P{^inhangulcompatibilityjamo}', "");
Expect(1, 12687, '\p{In_Hangul_compatibility_JAMO}', "");
Expect(0, 12687, '\p{^In_Hangul_compatibility_JAMO}', "");
Expect(0, 12687, '\P{In_Hangul_compatibility_JAMO}', "");
Expect(1, 12687, '\P{^In_Hangul_compatibility_JAMO}', "");
Expect(0, 12688, '\p{In_Hangul_compatibility_JAMO}', "");
Expect(1, 12688, '\p{^In_Hangul_compatibility_JAMO}', "");
Expect(1, 12688, '\P{In_Hangul_compatibility_JAMO}', "");
Expect(0, 12688, '\P{^In_Hangul_compatibility_JAMO}', "");
Error('\p{	COMPAT_jamo/a/}');
Error('\P{	COMPAT_jamo/a/}');
Expect(1, 12687, '\p{compatjamo}', "");
Expect(0, 12687, '\p{^compatjamo}', "");
Expect(0, 12687, '\P{compatjamo}', "");
Expect(1, 12687, '\P{^compatjamo}', "");
Expect(0, 12688, '\p{compatjamo}', "");
Expect(1, 12688, '\p{^compatjamo}', "");
Expect(1, 12688, '\P{compatjamo}', "");
Expect(0, 12688, '\P{^compatjamo}', "");
Expect(1, 12687, '\p{ -compat_Jamo}', "");
Expect(0, 12687, '\p{^ -compat_Jamo}', "");
Expect(0, 12687, '\P{ -compat_Jamo}', "");
Expect(1, 12687, '\P{^ -compat_Jamo}', "");
Expect(0, 12688, '\p{ -compat_Jamo}', "");
Expect(1, 12688, '\p{^ -compat_Jamo}', "");
Expect(1, 12688, '\P{ -compat_Jamo}', "");
Expect(0, 12688, '\P{^ -compat_Jamo}', "");
Error('\p{_IS_COMPAT_Jamo/a/}');
Error('\P{_IS_COMPAT_Jamo/a/}');
Expect(1, 12687, '\p{iscompatjamo}', "");
Expect(0, 12687, '\p{^iscompatjamo}', "");
Expect(0, 12687, '\P{iscompatjamo}', "");
Expect(1, 12687, '\P{^iscompatjamo}', "");
Expect(0, 12688, '\p{iscompatjamo}', "");
Expect(1, 12688, '\p{^iscompatjamo}', "");
Expect(1, 12688, '\P{iscompatjamo}', "");
Expect(0, 12688, '\P{^iscompatjamo}', "");
Expect(1, 12687, '\p{	_Is_Compat_Jamo}', "");
Expect(0, 12687, '\p{^	_Is_Compat_Jamo}', "");
Expect(0, 12687, '\P{	_Is_Compat_Jamo}', "");
Expect(1, 12687, '\P{^	_Is_Compat_Jamo}', "");
Expect(0, 12688, '\p{	_Is_Compat_Jamo}', "");
Expect(1, 12688, '\p{^	_Is_Compat_Jamo}', "");
Expect(1, 12688, '\P{	_Is_Compat_Jamo}', "");
Expect(0, 12688, '\P{^	_Is_Compat_Jamo}', "");
Error('\p{:=__In_Compat_Jamo}');
Error('\P{:=__In_Compat_Jamo}');
Expect(1, 12687, '\p{incompatjamo}', "");
Expect(0, 12687, '\p{^incompatjamo}', "");
Expect(0, 12687, '\P{incompatjamo}', "");
Expect(1, 12687, '\P{^incompatjamo}', "");
Expect(0, 12688, '\p{incompatjamo}', "");
Expect(1, 12688, '\p{^incompatjamo}', "");
Expect(1, 12688, '\P{incompatjamo}', "");
Expect(0, 12688, '\P{^incompatjamo}', "");
Expect(1, 12687, '\p{	-IN_Compat_JAMO}', "");
Expect(0, 12687, '\p{^	-IN_Compat_JAMO}', "");
Expect(0, 12687, '\P{	-IN_Compat_JAMO}', "");
Expect(1, 12687, '\P{^	-IN_Compat_JAMO}', "");
Expect(0, 12688, '\p{	-IN_Compat_JAMO}', "");
Expect(1, 12688, '\p{^	-IN_Compat_JAMO}', "");
Expect(1, 12688, '\P{	-IN_Compat_JAMO}', "");
Expect(0, 12688, '\P{^	-IN_Compat_JAMO}', "");
Error('\p{_:=Diacritic}');
Error('\P{_:=Diacritic}');
Expect(1, 119213, '\p{diacritic}', "");
Expect(0, 119213, '\p{^diacritic}', "");
Expect(0, 119213, '\P{diacritic}', "");
Expect(1, 119213, '\P{^diacritic}', "");
Expect(0, 119214, '\p{diacritic}', "");
Expect(1, 119214, '\p{^diacritic}', "");
Expect(1, 119214, '\P{diacritic}', "");
Expect(0, 119214, '\P{^diacritic}', "");
Expect(1, 119213, '\p{-_Diacritic}', "");
Expect(0, 119213, '\p{^-_Diacritic}', "");
Expect(0, 119213, '\P{-_Diacritic}', "");
Expect(1, 119213, '\P{^-_Diacritic}', "");
Expect(0, 119214, '\p{-_Diacritic}', "");
Expect(1, 119214, '\p{^-_Diacritic}', "");
Expect(1, 119214, '\P{-_Diacritic}', "");
Expect(0, 119214, '\P{^-_Diacritic}', "");
Error('\p{_-IS_diacritic:=}');
Error('\P{_-IS_diacritic:=}');
Expect(1, 119213, '\p{isdiacritic}', "");
Expect(0, 119213, '\p{^isdiacritic}', "");
Expect(0, 119213, '\P{isdiacritic}', "");
Expect(1, 119213, '\P{^isdiacritic}', "");
Expect(0, 119214, '\p{isdiacritic}', "");
Expect(1, 119214, '\p{^isdiacritic}', "");
Expect(1, 119214, '\P{isdiacritic}', "");
Expect(0, 119214, '\P{^isdiacritic}', "");
Expect(1, 119213, '\p{	_is_DIACRITIC}', "");
Expect(0, 119213, '\p{^	_is_DIACRITIC}', "");
Expect(0, 119213, '\P{	_is_DIACRITIC}', "");
Expect(1, 119213, '\P{^	_is_DIACRITIC}', "");
Expect(0, 119214, '\p{	_is_DIACRITIC}', "");
Expect(1, 119214, '\p{^	_is_DIACRITIC}', "");
Expect(1, 119214, '\P{	_is_DIACRITIC}', "");
Expect(0, 119214, '\P{^	_is_DIACRITIC}', "");
Error('\p{/a/	Dia}');
Error('\P{/a/	Dia}');
Expect(1, 119213, '\p{dia}', "");
Expect(0, 119213, '\p{^dia}', "");
Expect(0, 119213, '\P{dia}', "");
Expect(1, 119213, '\P{^dia}', "");
Expect(0, 119214, '\p{dia}', "");
Expect(1, 119214, '\p{^dia}', "");
Expect(1, 119214, '\P{dia}', "");
Expect(0, 119214, '\P{^dia}', "");
Expect(1, 119213, '\p{ -Dia}', "");
Expect(0, 119213, '\p{^ -Dia}', "");
Expect(0, 119213, '\P{ -Dia}', "");
Expect(1, 119213, '\P{^ -Dia}', "");
Expect(0, 119214, '\p{ -Dia}', "");
Expect(1, 119214, '\p{^ -Dia}', "");
Expect(1, 119214, '\P{ -Dia}', "");
Expect(0, 119214, '\P{^ -Dia}', "");
Error('\p{	_Is_DIA/a/}');
Error('\P{	_Is_DIA/a/}');
Expect(1, 119213, '\p{isdia}', "");
Expect(0, 119213, '\p{^isdia}', "");
Expect(0, 119213, '\P{isdia}', "");
Expect(1, 119213, '\P{^isdia}', "");
Expect(0, 119214, '\p{isdia}', "");
Expect(1, 119214, '\p{^isdia}', "");
Expect(1, 119214, '\P{isdia}', "");
Expect(0, 119214, '\P{^isdia}', "");
Expect(1, 119213, '\p{	Is_DIA}', "");
Expect(0, 119213, '\p{^	Is_DIA}', "");
Expect(0, 119213, '\P{	Is_DIA}', "");
Expect(1, 119213, '\P{^	Is_DIA}', "");
Expect(0, 119214, '\p{	Is_DIA}', "");
Expect(1, 119214, '\p{^	Is_DIA}', "");
Expect(1, 119214, '\P{	Is_DIA}', "");
Expect(0, 119214, '\P{^	Is_DIA}', "");
Error('\p{:=lowercase_Letter}');
Error('\P{:=lowercase_Letter}');
Expect(1, 120779, '\p{lowercaseletter}', "");
Expect(0, 120779, '\p{^lowercaseletter}', "");
Expect(0, 120779, '\P{lowercaseletter}', "");
Expect(1, 120779, '\P{^lowercaseletter}', "");
Expect(0, 120780, '\p{lowercaseletter}', "");
Expect(1, 120780, '\p{^lowercaseletter}', "");
Expect(1, 120780, '\P{lowercaseletter}', "");
Expect(0, 120780, '\P{^lowercaseletter}', "");
Expect(1, 120779, '\p{	Lowercase_Letter}', "");
Expect(0, 120779, '\p{^	Lowercase_Letter}', "");
Expect(0, 120779, '\P{	Lowercase_Letter}', "");
Expect(1, 120779, '\P{^	Lowercase_Letter}', "");
Expect(0, 120780, '\p{	Lowercase_Letter}', "");
Expect(1, 120780, '\p{^	Lowercase_Letter}', "");
Expect(1, 120780, '\P{	Lowercase_Letter}', "");
Expect(0, 120780, '\P{^	Lowercase_Letter}', "");
Error('\p{	_IS_LOWERCASE_letter/a/}');
Error('\P{	_IS_LOWERCASE_letter/a/}');
Expect(1, 120779, '\p{islowercaseletter}', "");
Expect(0, 120779, '\p{^islowercaseletter}', "");
Expect(0, 120779, '\P{islowercaseletter}', "");
Expect(1, 120779, '\P{^islowercaseletter}', "");
Expect(0, 120780, '\p{islowercaseletter}', "");
Expect(1, 120780, '\p{^islowercaseletter}', "");
Expect(1, 120780, '\P{islowercaseletter}', "");
Expect(0, 120780, '\P{^islowercaseletter}', "");
Expect(1, 120779, '\p{	-is_Lowercase_Letter}', "");
Expect(0, 120779, '\p{^	-is_Lowercase_Letter}', "");
Expect(0, 120779, '\P{	-is_Lowercase_Letter}', "");
Expect(1, 120779, '\P{^	-is_Lowercase_Letter}', "");
Expect(0, 120780, '\p{	-is_Lowercase_Letter}', "");
Expect(1, 120780, '\p{^	-is_Lowercase_Letter}', "");
Expect(1, 120780, '\P{	-is_Lowercase_Letter}', "");
Expect(0, 120780, '\P{^	-is_Lowercase_Letter}', "");
Error('\p{:=_Ll}');
Error('\P{:=_Ll}');
Expect(1, 120779, '\p{ll}', "");
Expect(0, 120779, '\p{^ll}', "");
Expect(0, 120779, '\P{ll}', "");
Expect(1, 120779, '\P{^ll}', "");
Expect(0, 120780, '\p{ll}', "");
Expect(1, 120780, '\p{^ll}', "");
Expect(1, 120780, '\P{ll}', "");
Expect(0, 120780, '\P{^ll}', "");
Expect(1, 120779, '\p{ LL}', "");
Expect(0, 120779, '\p{^ LL}', "");
Expect(0, 120779, '\P{ LL}', "");
Expect(1, 120779, '\P{^ LL}', "");
Expect(0, 120780, '\p{ LL}', "");
Expect(1, 120780, '\p{^ LL}', "");
Expect(1, 120780, '\P{ LL}', "");
Expect(0, 120780, '\P{^ LL}', "");
Error('\p{-:=IS_Ll}');
Error('\P{-:=IS_Ll}');
Expect(1, 120779, '\p{isll}', "");
Expect(0, 120779, '\p{^isll}', "");
Expect(0, 120779, '\P{isll}', "");
Expect(1, 120779, '\P{^isll}', "");
Expect(0, 120780, '\p{isll}', "");
Expect(1, 120780, '\p{^isll}', "");
Expect(1, 120780, '\P{isll}', "");
Expect(0, 120780, '\P{^isll}', "");
Expect(1, 120779, '\p{		Is_Ll}', "");
Expect(0, 120779, '\p{^		Is_Ll}', "");
Expect(0, 120779, '\P{		Is_Ll}', "");
Expect(1, 120779, '\P{^		Is_Ll}', "");
Expect(0, 120780, '\p{		Is_Ll}', "");
Expect(1, 120780, '\p{^		Is_Ll}', "");
Expect(1, 120780, '\P{		Is_Ll}', "");
Expect(0, 120780, '\P{^		Is_Ll}', "");
Error('\p{:=-Musical_Symbols}');
Error('\P{:=-Musical_Symbols}');
Expect(1, 119295, '\p{musicalsymbols}', "");
Expect(0, 119295, '\p{^musicalsymbols}', "");
Expect(0, 119295, '\P{musicalsymbols}', "");
Expect(1, 119295, '\P{^musicalsymbols}', "");
Expect(0, 119296, '\p{musicalsymbols}', "");
Expect(1, 119296, '\p{^musicalsymbols}', "");
Expect(1, 119296, '\P{musicalsymbols}', "");
Expect(0, 119296, '\P{^musicalsymbols}', "");
Expect(1, 119295, '\p{_	Musical_Symbols}', "");
Expect(0, 119295, '\p{^_	Musical_Symbols}', "");
Expect(0, 119295, '\P{_	Musical_Symbols}', "");
Expect(1, 119295, '\P{^_	Musical_Symbols}', "");
Expect(0, 119296, '\p{_	Musical_Symbols}', "");
Expect(1, 119296, '\p{^_	Musical_Symbols}', "");
Expect(1, 119296, '\P{_	Musical_Symbols}', "");
Expect(0, 119296, '\P{^_	Musical_Symbols}', "");
Error('\p{:=- is_Musical_Symbols}');
Error('\P{:=- is_Musical_Symbols}');
Expect(1, 119295, '\p{ismusicalsymbols}', "");
Expect(0, 119295, '\p{^ismusicalsymbols}', "");
Expect(0, 119295, '\P{ismusicalsymbols}', "");
Expect(1, 119295, '\P{^ismusicalsymbols}', "");
Expect(0, 119296, '\p{ismusicalsymbols}', "");
Expect(1, 119296, '\p{^ismusicalsymbols}', "");
Expect(1, 119296, '\P{ismusicalsymbols}', "");
Expect(0, 119296, '\P{^ismusicalsymbols}', "");
Expect(1, 119295, '\p{		IS_Musical_symbols}', "");
Expect(0, 119295, '\p{^		IS_Musical_symbols}', "");
Expect(0, 119295, '\P{		IS_Musical_symbols}', "");
Expect(1, 119295, '\P{^		IS_Musical_symbols}', "");
Expect(0, 119296, '\p{		IS_Musical_symbols}', "");
Expect(1, 119296, '\p{^		IS_Musical_symbols}', "");
Expect(1, 119296, '\P{		IS_Musical_symbols}', "");
Expect(0, 119296, '\P{^		IS_Musical_symbols}', "");
Error('\p{_/a/In_Musical_Symbols}');
Error('\P{_/a/In_Musical_Symbols}');
Expect(1, 119295, '\p{inmusicalsymbols}', "");
Expect(0, 119295, '\p{^inmusicalsymbols}', "");
Expect(0, 119295, '\P{inmusicalsymbols}', "");
Expect(1, 119295, '\P{^inmusicalsymbols}', "");
Expect(0, 119296, '\p{inmusicalsymbols}', "");
Expect(1, 119296, '\p{^inmusicalsymbols}', "");
Expect(1, 119296, '\P{inmusicalsymbols}', "");
Expect(0, 119296, '\P{^inmusicalsymbols}', "");
Expect(1, 119295, '\p{  In_Musical_symbols}', "");
Expect(0, 119295, '\p{^  In_Musical_symbols}', "");
Expect(0, 119295, '\P{  In_Musical_symbols}', "");
Expect(1, 119295, '\P{^  In_Musical_symbols}', "");
Expect(0, 119296, '\p{  In_Musical_symbols}', "");
Expect(1, 119296, '\p{^  In_Musical_symbols}', "");
Expect(1, 119296, '\P{  In_Musical_symbols}', "");
Expect(0, 119296, '\P{^  In_Musical_symbols}', "");
Error('\p{ :=MUSIC}');
Error('\P{ :=MUSIC}');
Expect(1, 119295, '\p{music}', "");
Expect(0, 119295, '\p{^music}', "");
Expect(0, 119295, '\P{music}', "");
Expect(1, 119295, '\P{^music}', "");
Expect(0, 119296, '\p{music}', "");
Expect(1, 119296, '\p{^music}', "");
Expect(1, 119296, '\P{music}', "");
Expect(0, 119296, '\P{^music}', "");
Expect(1, 119295, '\p{		MUSIC}', "");
Expect(0, 119295, '\p{^		MUSIC}', "");
Expect(0, 119295, '\P{		MUSIC}', "");
Expect(1, 119295, '\P{^		MUSIC}', "");
Expect(0, 119296, '\p{		MUSIC}', "");
Expect(1, 119296, '\p{^		MUSIC}', "");
Expect(1, 119296, '\P{		MUSIC}', "");
Expect(0, 119296, '\P{^		MUSIC}', "");
Error('\p{_	Is_Music:=}');
Error('\P{_	Is_Music:=}');
Expect(1, 119295, '\p{ismusic}', "");
Expect(0, 119295, '\p{^ismusic}', "");
Expect(0, 119295, '\P{ismusic}', "");
Expect(1, 119295, '\P{^ismusic}', "");
Expect(0, 119296, '\p{ismusic}', "");
Expect(1, 119296, '\p{^ismusic}', "");
Expect(1, 119296, '\P{ismusic}', "");
Expect(0, 119296, '\P{^ismusic}', "");
Expect(1, 119295, '\p{	 is_Music}', "");
Expect(0, 119295, '\p{^	 is_Music}', "");
Expect(0, 119295, '\P{	 is_Music}', "");
Expect(1, 119295, '\P{^	 is_Music}', "");
Expect(0, 119296, '\p{	 is_Music}', "");
Expect(1, 119296, '\p{^	 is_Music}', "");
Expect(1, 119296, '\P{	 is_Music}', "");
Expect(0, 119296, '\P{^	 is_Music}', "");
Error('\p{:=in_Music}');
Error('\P{:=in_Music}');
Expect(1, 119295, '\p{inmusic}', "");
Expect(0, 119295, '\p{^inmusic}', "");
Expect(0, 119295, '\P{inmusic}', "");
Expect(1, 119295, '\P{^inmusic}', "");
Expect(0, 119296, '\p{inmusic}', "");
Expect(1, 119296, '\p{^inmusic}', "");
Expect(1, 119296, '\P{inmusic}', "");
Expect(0, 119296, '\P{^inmusic}', "");
Expect(1, 119295, '\p{_IN_MUSIC}', "");
Expect(0, 119295, '\p{^_IN_MUSIC}', "");
Expect(0, 119295, '\P{_IN_MUSIC}', "");
Expect(1, 119295, '\P{^_IN_MUSIC}', "");
Expect(0, 119296, '\p{_IN_MUSIC}', "");
Expect(1, 119296, '\p{^_IN_MUSIC}', "");
Expect(1, 119296, '\P{_IN_MUSIC}', "");
Expect(0, 119296, '\P{^_IN_MUSIC}', "");
Error('\p{ 	tags:=}');
Error('\P{ 	tags:=}');
Expect(1, 917631, '\p{tags}', "");
Expect(0, 917631, '\p{^tags}', "");
Expect(0, 917631, '\P{tags}', "");
Expect(1, 917631, '\P{^tags}', "");
Expect(0, 917632, '\p{tags}', "");
Expect(1, 917632, '\p{^tags}', "");
Expect(1, 917632, '\P{tags}', "");
Expect(0, 917632, '\P{^tags}', "");
Expect(1, 917631, '\p{	tags}', "");
Expect(0, 917631, '\p{^	tags}', "");
Expect(0, 917631, '\P{	tags}', "");
Expect(1, 917631, '\P{^	tags}', "");
Expect(0, 917632, '\p{	tags}', "");
Expect(1, 917632, '\p{^	tags}', "");
Expect(1, 917632, '\P{	tags}', "");
Expect(0, 917632, '\P{^	tags}', "");
Error('\p{	/a/Is_tags}');
Error('\P{	/a/Is_tags}');
Expect(1, 917631, '\p{istags}', "");
Expect(0, 917631, '\p{^istags}', "");
Expect(0, 917631, '\P{istags}', "");
Expect(1, 917631, '\P{^istags}', "");
Expect(0, 917632, '\p{istags}', "");
Expect(1, 917632, '\p{^istags}', "");
Expect(1, 917632, '\P{istags}', "");
Expect(0, 917632, '\P{^istags}', "");
Expect(1, 917631, '\p{__Is_tags}', "");
Expect(0, 917631, '\p{^__Is_tags}', "");
Expect(0, 917631, '\P{__Is_tags}', "");
Expect(1, 917631, '\P{^__Is_tags}', "");
Expect(0, 917632, '\p{__Is_tags}', "");
Expect(1, 917632, '\p{^__Is_tags}', "");
Expect(1, 917632, '\P{__Is_tags}', "");
Expect(0, 917632, '\P{^__Is_tags}', "");
Error('\p{/a/  in_Tags}');
Error('\P{/a/  in_Tags}');
Expect(1, 917631, '\p{intags}', "");
Expect(0, 917631, '\p{^intags}', "");
Expect(0, 917631, '\P{intags}', "");
Expect(1, 917631, '\P{^intags}', "");
Expect(0, 917632, '\p{intags}', "");
Expect(1, 917632, '\p{^intags}', "");
Expect(1, 917632, '\P{intags}', "");
Expect(0, 917632, '\P{^intags}', "");
Expect(1, 917631, '\p{_	In_Tags}', "");
Expect(0, 917631, '\p{^_	In_Tags}', "");
Expect(0, 917631, '\P{_	In_Tags}', "");
Expect(1, 917631, '\P{^_	In_Tags}', "");
Expect(0, 917632, '\p{_	In_Tags}', "");
Expect(1, 917632, '\p{^_	In_Tags}', "");
Expect(1, 917632, '\P{_	In_Tags}', "");
Expect(0, 917632, '\P{^_	In_Tags}', "");
Error('\p{/a/	Separator}');
Error('\P{/a/	Separator}');
Expect(1, 12288, '\p{separator}', "");
Expect(0, 12288, '\p{^separator}', "");
Expect(0, 12288, '\P{separator}', "");
Expect(1, 12288, '\P{^separator}', "");
Expect(0, 12289, '\p{separator}', "");
Expect(1, 12289, '\p{^separator}', "");
Expect(1, 12289, '\P{separator}', "");
Expect(0, 12289, '\P{^separator}', "");
Expect(1, 12288, '\p{_Separator}', "");
Expect(0, 12288, '\p{^_Separator}', "");
Expect(0, 12288, '\P{_Separator}', "");
Expect(1, 12288, '\P{^_Separator}', "");
Expect(0, 12289, '\p{_Separator}', "");
Expect(1, 12289, '\p{^_Separator}', "");
Expect(1, 12289, '\P{_Separator}', "");
Expect(0, 12289, '\P{^_Separator}', "");
Error('\p{_IS_SEPARATOR:=}');
Error('\P{_IS_SEPARATOR:=}');
Expect(1, 12288, '\p{isseparator}', "");
Expect(0, 12288, '\p{^isseparator}', "");
Expect(0, 12288, '\P{isseparator}', "");
Expect(1, 12288, '\P{^isseparator}', "");
Expect(0, 12289, '\p{isseparator}', "");
Expect(1, 12289, '\p{^isseparator}', "");
Expect(1, 12289, '\P{isseparator}', "");
Expect(0, 12289, '\P{^isseparator}', "");
Expect(1, 12288, '\p{-_IS_SEPARATOR}', "");
Expect(0, 12288, '\p{^-_IS_SEPARATOR}', "");
Expect(0, 12288, '\P{-_IS_SEPARATOR}', "");
Expect(1, 12288, '\P{^-_IS_SEPARATOR}', "");
Expect(0, 12289, '\p{-_IS_SEPARATOR}', "");
Expect(1, 12289, '\p{^-_IS_SEPARATOR}', "");
Expect(1, 12289, '\P{-_IS_SEPARATOR}', "");
Expect(0, 12289, '\P{^-_IS_SEPARATOR}', "");
Error('\p{:=_	Z}');
Error('\P{:=_	Z}');
Expect(1, 12288, '\p{z}', "");
Expect(0, 12288, '\p{^z}', "");
Expect(0, 12288, '\P{z}', "");
Expect(1, 12288, '\P{^z}', "");
Expect(0, 12289, '\p{z}', "");
Expect(1, 12289, '\p{^z}', "");
Expect(1, 12289, '\P{z}', "");
Expect(0, 12289, '\P{^z}', "");
Expect(1, 12288, '\p{_ Z}', "");
Expect(0, 12288, '\p{^_ Z}', "");
Expect(0, 12288, '\P{_ Z}', "");
Expect(1, 12288, '\P{^_ Z}', "");
Expect(0, 12289, '\p{_ Z}', "");
Expect(1, 12289, '\p{^_ Z}', "");
Expect(1, 12289, '\P{_ Z}', "");
Expect(0, 12289, '\P{^_ Z}', "");
Error('\p{ -Is_Z/a/}');
Error('\P{ -Is_Z/a/}');
Expect(1, 12288, '\p{isz}', "");
Expect(0, 12288, '\p{^isz}', "");
Expect(0, 12288, '\P{isz}', "");
Expect(1, 12288, '\P{^isz}', "");
Expect(0, 12289, '\p{isz}', "");
Expect(1, 12289, '\p{^isz}', "");
Expect(1, 12289, '\P{isz}', "");
Expect(0, 12289, '\P{^isz}', "");
Expect(1, 12288, '\p{	_Is_Z}', "");
Expect(0, 12288, '\p{^	_Is_Z}', "");
Expect(0, 12288, '\P{	_Is_Z}', "");
Expect(1, 12288, '\P{^	_Is_Z}', "");
Expect(0, 12289, '\p{	_Is_Z}', "");
Expect(1, 12289, '\p{^	_Is_Z}', "");
Expect(1, 12289, '\P{	_Is_Z}', "");
Expect(0, 12289, '\P{^	_Is_Z}', "");
Error('\p{_:=Ancient_Greek_MUSICAL_NOTATION}');
Error('\P{_:=Ancient_Greek_MUSICAL_NOTATION}');
Expect(1, 119375, '\p{ancientgreekmusicalnotation}', "");
Expect(0, 119375, '\p{^ancientgreekmusicalnotation}', "");
Expect(0, 119375, '\P{ancientgreekmusicalnotation}', "");
Expect(1, 119375, '\P{^ancientgreekmusicalnotation}', "");
Expect(0, 119376, '\p{ancientgreekmusicalnotation}', "");
Expect(1, 119376, '\p{^ancientgreekmusicalnotation}', "");
Expect(1, 119376, '\P{ancientgreekmusicalnotation}', "");
Expect(0, 119376, '\P{^ancientgreekmusicalnotation}', "");
Expect(1, 119375, '\p{-_ancient_GREEK_MUSICAL_Notation}', "");
Expect(0, 119375, '\p{^-_ancient_GREEK_MUSICAL_Notation}', "");
Expect(0, 119375, '\P{-_ancient_GREEK_MUSICAL_Notation}', "");
Expect(1, 119375, '\P{^-_ancient_GREEK_MUSICAL_Notation}', "");
Expect(0, 119376, '\p{-_ancient_GREEK_MUSICAL_Notation}', "");
Expect(1, 119376, '\p{^-_ancient_GREEK_MUSICAL_Notation}', "");
Expect(1, 119376, '\P{-_ancient_GREEK_MUSICAL_Notation}', "");
Expect(0, 119376, '\P{^-_ancient_GREEK_MUSICAL_Notation}', "");
Error('\p{		is_ancient_greek_MUSICAL_Notation/a/}');
Error('\P{		is_ancient_greek_MUSICAL_Notation/a/}');
Expect(1, 119375, '\p{isancientgreekmusicalnotation}', "");
Expect(0, 119375, '\p{^isancientgreekmusicalnotation}', "");
Expect(0, 119375, '\P{isancientgreekmusicalnotation}', "");
Expect(1, 119375, '\P{^isancientgreekmusicalnotation}', "");
Expect(0, 119376, '\p{isancientgreekmusicalnotation}', "");
Expect(1, 119376, '\p{^isancientgreekmusicalnotation}', "");
Expect(1, 119376, '\P{isancientgreekmusicalnotation}', "");
Expect(0, 119376, '\P{^isancientgreekmusicalnotation}', "");
Expect(1, 119375, '\p{	Is_ANCIENT_Greek_musical_notation}', "");
Expect(0, 119375, '\p{^	Is_ANCIENT_Greek_musical_notation}', "");
Expect(0, 119375, '\P{	Is_ANCIENT_Greek_musical_notation}', "");
Expect(1, 119375, '\P{^	Is_ANCIENT_Greek_musical_notation}', "");
Expect(0, 119376, '\p{	Is_ANCIENT_Greek_musical_notation}', "");
Expect(1, 119376, '\p{^	Is_ANCIENT_Greek_musical_notation}', "");
Expect(1, 119376, '\P{	Is_ANCIENT_Greek_musical_notation}', "");
Expect(0, 119376, '\P{^	Is_ANCIENT_Greek_musical_notation}', "");
Error('\p{--IN_Ancient_GREEK_musical_notation:=}');
Error('\P{--IN_Ancient_GREEK_musical_notation:=}');
Expect(1, 119375, '\p{inancientgreekmusicalnotation}', "");
Expect(0, 119375, '\p{^inancientgreekmusicalnotation}', "");
Expect(0, 119375, '\P{inancientgreekmusicalnotation}', "");
Expect(1, 119375, '\P{^inancientgreekmusicalnotation}', "");
Expect(0, 119376, '\p{inancientgreekmusicalnotation}', "");
Expect(1, 119376, '\p{^inancientgreekmusicalnotation}', "");
Expect(1, 119376, '\P{inancientgreekmusicalnotation}', "");
Expect(0, 119376, '\P{^inancientgreekmusicalnotation}', "");
Expect(1, 119375, '\p{	_IN_Ancient_Greek_Musical_notation}', "");
Expect(0, 119375, '\p{^	_IN_Ancient_Greek_Musical_notation}', "");
Expect(0, 119375, '\P{	_IN_Ancient_Greek_Musical_notation}', "");
Expect(1, 119375, '\P{^	_IN_Ancient_Greek_Musical_notation}', "");
Expect(0, 119376, '\p{	_IN_Ancient_Greek_Musical_notation}', "");
Expect(1, 119376, '\p{^	_IN_Ancient_Greek_Musical_notation}', "");
Expect(1, 119376, '\P{	_IN_Ancient_Greek_Musical_notation}', "");
Expect(0, 119376, '\P{^	_IN_Ancient_Greek_Musical_notation}', "");
Error('\p{	:=ANCIENT_Greek_Music}');
Error('\P{	:=ANCIENT_Greek_Music}');
Expect(1, 119375, '\p{ancientgreekmusic}', "");
Expect(0, 119375, '\p{^ancientgreekmusic}', "");
Expect(0, 119375, '\P{ancientgreekmusic}', "");
Expect(1, 119375, '\P{^ancientgreekmusic}', "");
Expect(0, 119376, '\p{ancientgreekmusic}', "");
Expect(1, 119376, '\p{^ancientgreekmusic}', "");
Expect(1, 119376, '\P{ancientgreekmusic}', "");
Expect(0, 119376, '\P{^ancientgreekmusic}', "");
Expect(1, 119375, '\p{ -Ancient_Greek_Music}', "");
Expect(0, 119375, '\p{^ -Ancient_Greek_Music}', "");
Expect(0, 119375, '\P{ -Ancient_Greek_Music}', "");
Expect(1, 119375, '\P{^ -Ancient_Greek_Music}', "");
Expect(0, 119376, '\p{ -Ancient_Greek_Music}', "");
Expect(1, 119376, '\p{^ -Ancient_Greek_Music}', "");
Expect(1, 119376, '\P{ -Ancient_Greek_Music}', "");
Expect(0, 119376, '\P{^ -Ancient_Greek_Music}', "");
Error('\p{:=Is_Ancient_GREEK_Music}');
Error('\P{:=Is_Ancient_GREEK_Music}');
Expect(1, 119375, '\p{isancientgreekmusic}', "");
Expect(0, 119375, '\p{^isancientgreekmusic}', "");
Expect(0, 119375, '\P{isancientgreekmusic}', "");
Expect(1, 119375, '\P{^isancientgreekmusic}', "");
Expect(0, 119376, '\p{isancientgreekmusic}', "");
Expect(1, 119376, '\p{^isancientgreekmusic}', "");
Expect(1, 119376, '\P{isancientgreekmusic}', "");
Expect(0, 119376, '\P{^isancientgreekmusic}', "");
Expect(1, 119375, '\p{__Is_ANCIENT_Greek_Music}', "");
Expect(0, 119375, '\p{^__Is_ANCIENT_Greek_Music}', "");
Expect(0, 119375, '\P{__Is_ANCIENT_Greek_Music}', "");
Expect(1, 119375, '\P{^__Is_ANCIENT_Greek_Music}', "");
Expect(0, 119376, '\p{__Is_ANCIENT_Greek_Music}', "");
Expect(1, 119376, '\p{^__Is_ANCIENT_Greek_Music}', "");
Expect(1, 119376, '\P{__Is_ANCIENT_Greek_Music}', "");
Expect(0, 119376, '\P{^__Is_ANCIENT_Greek_Music}', "");
Error('\p{_:=IN_Ancient_GREEK_music}');
Error('\P{_:=IN_Ancient_GREEK_music}');
Expect(1, 119375, '\p{inancientgreekmusic}', "");
Expect(0, 119375, '\p{^inancientgreekmusic}', "");
Expect(0, 119375, '\P{inancientgreekmusic}', "");
Expect(1, 119375, '\P{^inancientgreekmusic}', "");
Expect(0, 119376, '\p{inancientgreekmusic}', "");
Expect(1, 119376, '\p{^inancientgreekmusic}', "");
Expect(1, 119376, '\P{inancientgreekmusic}', "");
Expect(0, 119376, '\P{^inancientgreekmusic}', "");
Expect(1, 119375, '\p{--in_Ancient_Greek_Music}', "");
Expect(0, 119375, '\p{^--in_Ancient_Greek_Music}', "");
Expect(0, 119375, '\P{--in_Ancient_Greek_Music}', "");
Expect(1, 119375, '\P{^--in_Ancient_Greek_Music}', "");
Expect(0, 119376, '\p{--in_Ancient_Greek_Music}', "");
Expect(1, 119376, '\p{^--in_Ancient_Greek_Music}', "");
Expect(1, 119376, '\P{--in_Ancient_Greek_Music}', "");
Expect(0, 119376, '\P{^--in_Ancient_Greek_Music}', "");
Error('\p{/a/__Cuneiform_Numbers_And_punctuation}');
Error('\P{/a/__Cuneiform_Numbers_And_punctuation}');
Expect(1, 74879, '\p{cuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\p{^cuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\P{cuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\P{^cuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\p{cuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\p{^cuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\P{cuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\P{^cuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\p{-	cuneiform_numbers_And_Punctuation}', "");
Expect(0, 74879, '\p{^-	cuneiform_numbers_And_Punctuation}', "");
Expect(0, 74879, '\P{-	cuneiform_numbers_And_Punctuation}', "");
Expect(1, 74879, '\P{^-	cuneiform_numbers_And_Punctuation}', "");
Expect(0, 74880, '\p{-	cuneiform_numbers_And_Punctuation}', "");
Expect(1, 74880, '\p{^-	cuneiform_numbers_And_Punctuation}', "");
Expect(1, 74880, '\P{-	cuneiform_numbers_And_Punctuation}', "");
Expect(0, 74880, '\P{^-	cuneiform_numbers_And_Punctuation}', "");
Error('\p{:=Is_cuneiform_NUMBERS_And_Punctuation}');
Error('\P{:=Is_cuneiform_NUMBERS_And_Punctuation}');
Expect(1, 74879, '\p{iscuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\p{^iscuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\P{iscuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\P{^iscuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\p{iscuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\p{^iscuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\P{iscuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\P{^iscuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\p{	 IS_CUNEIFORM_numbers_AND_PUNCTUATION}', "");
Expect(0, 74879, '\p{^	 IS_CUNEIFORM_numbers_AND_PUNCTUATION}', "");
Expect(0, 74879, '\P{	 IS_CUNEIFORM_numbers_AND_PUNCTUATION}', "");
Expect(1, 74879, '\P{^	 IS_CUNEIFORM_numbers_AND_PUNCTUATION}', "");
Expect(0, 74880, '\p{	 IS_CUNEIFORM_numbers_AND_PUNCTUATION}', "");
Expect(1, 74880, '\p{^	 IS_CUNEIFORM_numbers_AND_PUNCTUATION}', "");
Expect(1, 74880, '\P{	 IS_CUNEIFORM_numbers_AND_PUNCTUATION}', "");
Expect(0, 74880, '\P{^	 IS_CUNEIFORM_numbers_AND_PUNCTUATION}', "");
Error('\p{/a/_ IN_cuneiform_numbers_And_Punctuation}');
Error('\P{/a/_ IN_cuneiform_numbers_And_Punctuation}');
Expect(1, 74879, '\p{incuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\p{^incuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\P{incuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\P{^incuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\p{incuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\p{^incuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\P{incuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\P{^incuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\p{__In_cuneiform_Numbers_and_Punctuation}', "");
Expect(0, 74879, '\p{^__In_cuneiform_Numbers_and_Punctuation}', "");
Expect(0, 74879, '\P{__In_cuneiform_Numbers_and_Punctuation}', "");
Expect(1, 74879, '\P{^__In_cuneiform_Numbers_and_Punctuation}', "");
Expect(0, 74880, '\p{__In_cuneiform_Numbers_and_Punctuation}', "");
Expect(1, 74880, '\p{^__In_cuneiform_Numbers_and_Punctuation}', "");
Expect(1, 74880, '\P{__In_cuneiform_Numbers_and_Punctuation}', "");
Expect(0, 74880, '\P{^__In_cuneiform_Numbers_and_Punctuation}', "");
Error('\p{		Cuneiform_Numbers/a/}');
Error('\P{		Cuneiform_Numbers/a/}');
Expect(1, 74879, '\p{cuneiformnumbers}', "");
Expect(0, 74879, '\p{^cuneiformnumbers}', "");
Expect(0, 74879, '\P{cuneiformnumbers}', "");
Expect(1, 74879, '\P{^cuneiformnumbers}', "");
Expect(0, 74880, '\p{cuneiformnumbers}', "");
Expect(1, 74880, '\p{^cuneiformnumbers}', "");
Expect(1, 74880, '\P{cuneiformnumbers}', "");
Expect(0, 74880, '\P{^cuneiformnumbers}', "");
Expect(1, 74879, '\p{_Cuneiform_Numbers}', "");
Expect(0, 74879, '\p{^_Cuneiform_Numbers}', "");
Expect(0, 74879, '\P{_Cuneiform_Numbers}', "");
Expect(1, 74879, '\P{^_Cuneiform_Numbers}', "");
Expect(0, 74880, '\p{_Cuneiform_Numbers}', "");
Expect(1, 74880, '\p{^_Cuneiform_Numbers}', "");
Expect(1, 74880, '\P{_Cuneiform_Numbers}', "");
Expect(0, 74880, '\P{^_Cuneiform_Numbers}', "");
Error('\p{_ is_cuneiform_Numbers:=}');
Error('\P{_ is_cuneiform_Numbers:=}');
Expect(1, 74879, '\p{iscuneiformnumbers}', "");
Expect(0, 74879, '\p{^iscuneiformnumbers}', "");
Expect(0, 74879, '\P{iscuneiformnumbers}', "");
Expect(1, 74879, '\P{^iscuneiformnumbers}', "");
Expect(0, 74880, '\p{iscuneiformnumbers}', "");
Expect(1, 74880, '\p{^iscuneiformnumbers}', "");
Expect(1, 74880, '\P{iscuneiformnumbers}', "");
Expect(0, 74880, '\P{^iscuneiformnumbers}', "");
Expect(1, 74879, '\p{	IS_Cuneiform_NUMBERS}', "");
Expect(0, 74879, '\p{^	IS_Cuneiform_NUMBERS}', "");
Expect(0, 74879, '\P{	IS_Cuneiform_NUMBERS}', "");
Expect(1, 74879, '\P{^	IS_Cuneiform_NUMBERS}', "");
Expect(0, 74880, '\p{	IS_Cuneiform_NUMBERS}', "");
Expect(1, 74880, '\p{^	IS_Cuneiform_NUMBERS}', "");
Expect(1, 74880, '\P{	IS_Cuneiform_NUMBERS}', "");
Expect(0, 74880, '\P{^	IS_Cuneiform_NUMBERS}', "");
Error('\p{	in_Cuneiform_NUMBERS:=}');
Error('\P{	in_Cuneiform_NUMBERS:=}');
Expect(1, 74879, '\p{incuneiformnumbers}', "");
Expect(0, 74879, '\p{^incuneiformnumbers}', "");
Expect(0, 74879, '\P{incuneiformnumbers}', "");
Expect(1, 74879, '\P{^incuneiformnumbers}', "");
Expect(0, 74880, '\p{incuneiformnumbers}', "");
Expect(1, 74880, '\p{^incuneiformnumbers}', "");
Expect(1, 74880, '\P{incuneiformnumbers}', "");
Expect(0, 74880, '\P{^incuneiformnumbers}', "");
Expect(1, 74879, '\p{-	IN_Cuneiform_Numbers}', "");
Expect(0, 74879, '\p{^-	IN_Cuneiform_Numbers}', "");
Expect(0, 74879, '\P{-	IN_Cuneiform_Numbers}', "");
Expect(1, 74879, '\P{^-	IN_Cuneiform_Numbers}', "");
Expect(0, 74880, '\p{-	IN_Cuneiform_Numbers}', "");
Expect(1, 74880, '\p{^-	IN_Cuneiform_Numbers}', "");
Expect(1, 74880, '\P{-	IN_Cuneiform_Numbers}', "");
Expect(0, 74880, '\P{^-	IN_Cuneiform_Numbers}', "");
Error('\p{_ Space:=}');
Error('\P{_ Space:=}');
Expect(1, 12288, '\p{space}', "");
Expect(0, 12288, '\p{^space}', "");
Expect(0, 12288, '\P{space}', "");
Expect(1, 12288, '\P{^space}', "");
Expect(0, 12289, '\p{space}', "");
Expect(1, 12289, '\p{^space}', "");
Expect(1, 12289, '\P{space}', "");
Expect(0, 12289, '\P{^space}', "");
Expect(1, 12288, '\p{-_space}', "");
Expect(0, 12288, '\p{^-_space}', "");
Expect(0, 12288, '\P{-_space}', "");
Expect(1, 12288, '\P{^-_space}', "");
Expect(0, 12289, '\p{-_space}', "");
Expect(1, 12289, '\p{^-_space}', "");
Expect(1, 12289, '\P{-_space}', "");
Expect(0, 12289, '\P{^-_space}', "");
Error('\p{/a/	 XPosixSpace}');
Error('\P{/a/	 XPosixSpace}');
Expect(1, 12288, '\p{xposixspace}', "");
Expect(0, 12288, '\p{^xposixspace}', "");
Expect(0, 12288, '\P{xposixspace}', "");
Expect(1, 12288, '\P{^xposixspace}', "");
Expect(0, 12289, '\p{xposixspace}', "");
Expect(1, 12289, '\p{^xposixspace}', "");
Expect(1, 12289, '\P{xposixspace}', "");
Expect(0, 12289, '\P{^xposixspace}', "");
Expect(1, 12288, '\p{ 	XPosixSpace}', "");
Expect(0, 12288, '\p{^ 	XPosixSpace}', "");
Expect(0, 12288, '\P{ 	XPosixSpace}', "");
Expect(1, 12288, '\P{^ 	XPosixSpace}', "");
Expect(0, 12289, '\p{ 	XPosixSpace}', "");
Expect(1, 12289, '\p{^ 	XPosixSpace}', "");
Expect(1, 12289, '\P{ 	XPosixSpace}', "");
Expect(0, 12289, '\P{^ 	XPosixSpace}', "");
Error('\p{	/a/is_space}');
Error('\P{	/a/is_space}');
Expect(1, 12288, '\p{isspace}', "");
Expect(0, 12288, '\p{^isspace}', "");
Expect(0, 12288, '\P{isspace}', "");
Expect(1, 12288, '\P{^isspace}', "");
Expect(0, 12289, '\p{isspace}', "");
Expect(1, 12289, '\p{^isspace}', "");
Expect(1, 12289, '\P{isspace}', "");
Expect(0, 12289, '\P{^isspace}', "");
Expect(1, 12288, '\p{ 	Is_Space}', "");
Expect(0, 12288, '\p{^ 	Is_Space}', "");
Expect(0, 12288, '\P{ 	Is_Space}', "");
Expect(1, 12288, '\P{^ 	Is_Space}', "");
Expect(0, 12289, '\p{ 	Is_Space}', "");
Expect(1, 12289, '\p{^ 	Is_Space}', "");
Expect(1, 12289, '\P{ 	Is_Space}', "");
Expect(0, 12289, '\P{^ 	Is_Space}', "");
Error('\p{:=	Is_XPosixSpace}');
Error('\P{:=	Is_XPosixSpace}');
Expect(1, 12288, '\p{isxposixspace}', "");
Expect(0, 12288, '\p{^isxposixspace}', "");
Expect(0, 12288, '\P{isxposixspace}', "");
Expect(1, 12288, '\P{^isxposixspace}', "");
Expect(0, 12289, '\p{isxposixspace}', "");
Expect(1, 12289, '\p{^isxposixspace}', "");
Expect(1, 12289, '\P{isxposixspace}', "");
Expect(0, 12289, '\P{^isxposixspace}', "");
Expect(1, 12288, '\p{	-Is_xposixspace}', "");
Expect(0, 12288, '\p{^	-Is_xposixspace}', "");
Expect(0, 12288, '\P{	-Is_xposixspace}', "");
Expect(1, 12288, '\P{^	-Is_xposixspace}', "");
Expect(0, 12289, '\p{	-Is_xposixspace}', "");
Expect(1, 12289, '\p{^	-Is_xposixspace}', "");
Expect(1, 12289, '\P{	-Is_xposixspace}', "");
Expect(0, 12289, '\P{^	-Is_xposixspace}', "");
Error('\p{ _POSIXBLANK/a/}');
Error('\P{ _POSIXBLANK/a/}');
Expect(1, 32, '\p{posixblank}', "");
Expect(0, 32, '\p{^posixblank}', "");
Expect(0, 32, '\P{posixblank}', "");
Expect(1, 32, '\P{^posixblank}', "");
Expect(0, 33, '\p{posixblank}', "");
Expect(1, 33, '\p{^posixblank}', "");
Expect(1, 33, '\P{posixblank}', "");
Expect(0, 33, '\P{^posixblank}', "");
Expect(1, 32, '\p{_PosixBlank}', "");
Expect(0, 32, '\p{^_PosixBlank}', "");
Expect(0, 32, '\P{_PosixBlank}', "");
Expect(1, 32, '\P{^_PosixBlank}', "");
Expect(0, 33, '\p{_PosixBlank}', "");
Expect(1, 33, '\p{^_PosixBlank}', "");
Expect(1, 33, '\P{_PosixBlank}', "");
Expect(0, 33, '\P{^_PosixBlank}', "");
Error('\p{/a/IS_posixblank}');
Error('\P{/a/IS_posixblank}');
Expect(1, 32, '\p{isposixblank}', "");
Expect(0, 32, '\p{^isposixblank}', "");
Expect(0, 32, '\P{isposixblank}', "");
Expect(1, 32, '\P{^isposixblank}', "");
Expect(0, 33, '\p{isposixblank}', "");
Expect(1, 33, '\p{^isposixblank}', "");
Expect(1, 33, '\P{isposixblank}', "");
Expect(0, 33, '\P{^isposixblank}', "");
Expect(1, 32, '\p{	 Is_posixblank}', "");
Expect(0, 32, '\p{^	 Is_posixblank}', "");
Expect(0, 32, '\P{	 Is_posixblank}', "");
Expect(1, 32, '\P{^	 Is_posixblank}', "");
Expect(0, 33, '\p{	 Is_posixblank}', "");
Expect(1, 33, '\p{^	 Is_posixblank}', "");
Expect(1, 33, '\P{	 Is_posixblank}', "");
Expect(0, 33, '\P{^	 Is_posixblank}', "");
Error('\p{-IN_SINHALA/a/}');
Error('\P{-IN_SINHALA/a/}');
Expect(1, 3583, '\p{insinhala}', "");
Expect(0, 3583, '\p{^insinhala}', "");
Expect(0, 3583, '\P{insinhala}', "");
Expect(1, 3583, '\P{^insinhala}', "");
Expect(0, 3584, '\p{insinhala}', "");
Expect(1, 3584, '\p{^insinhala}', "");
Expect(1, 3584, '\P{insinhala}', "");
Expect(0, 3584, '\P{^insinhala}', "");
Expect(1, 3583, '\p{--IN_SINHALA}', "");
Expect(0, 3583, '\p{^--IN_SINHALA}', "");
Expect(0, 3583, '\P{--IN_SINHALA}', "");
Expect(1, 3583, '\P{^--IN_SINHALA}', "");
Expect(0, 3584, '\p{--IN_SINHALA}', "");
Expect(1, 3584, '\p{^--IN_SINHALA}', "");
Expect(1, 3584, '\P{--IN_SINHALA}', "");
Expect(0, 3584, '\P{^--IN_SINHALA}', "");
Error('\p{:=- in_SINHALA}');
Error('\P{:=- in_SINHALA}');
Expect(1, 3583, '\p{-_In_Sinhala}', "");
Expect(0, 3583, '\p{^-_In_Sinhala}', "");
Expect(0, 3583, '\P{-_In_Sinhala}', "");
Expect(1, 3583, '\P{^-_In_Sinhala}', "");
Expect(0, 3584, '\p{-_In_Sinhala}', "");
Expect(1, 3584, '\p{^-_In_Sinhala}', "");
Expect(1, 3584, '\P{-_In_Sinhala}', "");
Expect(0, 3584, '\P{^-_In_Sinhala}', "");
Error('\p{-:=Mongolian}');
Error('\P{-:=Mongolian}');
Expect(1, 6314, '\p{mongolian}', "");
Expect(0, 6314, '\p{^mongolian}', "");
Expect(0, 6314, '\P{mongolian}', "");
Expect(1, 6314, '\P{^mongolian}', "");
Expect(0, 6315, '\p{mongolian}', "");
Expect(1, 6315, '\p{^mongolian}', "");
Expect(1, 6315, '\P{mongolian}', "");
Expect(0, 6315, '\P{^mongolian}', "");
Expect(1, 6314, '\p{- MONGOLIAN}', "");
Expect(0, 6314, '\p{^- MONGOLIAN}', "");
Expect(0, 6314, '\P{- MONGOLIAN}', "");
Expect(1, 6314, '\P{^- MONGOLIAN}', "");
Expect(0, 6315, '\p{- MONGOLIAN}', "");
Expect(1, 6315, '\p{^- MONGOLIAN}', "");
Expect(1, 6315, '\P{- MONGOLIAN}', "");
Expect(0, 6315, '\P{^- MONGOLIAN}', "");
Error('\p{/a/	 Is_Mongolian}');
Error('\P{/a/	 Is_Mongolian}');
Expect(1, 6314, '\p{ismongolian}', "");
Expect(0, 6314, '\p{^ismongolian}', "");
Expect(0, 6314, '\P{ismongolian}', "");
Expect(1, 6314, '\P{^ismongolian}', "");
Expect(0, 6315, '\p{ismongolian}', "");
Expect(1, 6315, '\p{^ismongolian}', "");
Expect(1, 6315, '\P{ismongolian}', "");
Expect(0, 6315, '\P{^ismongolian}', "");
Expect(1, 6314, '\p{ IS_mongolian}', "");
Expect(0, 6314, '\p{^ IS_mongolian}', "");
Expect(0, 6314, '\P{ IS_mongolian}', "");
Expect(1, 6314, '\P{^ IS_mongolian}', "");
Expect(0, 6315, '\p{ IS_mongolian}', "");
Expect(1, 6315, '\p{^ IS_mongolian}', "");
Expect(1, 6315, '\P{ IS_mongolian}', "");
Expect(0, 6315, '\P{^ IS_mongolian}', "");
Error('\p{/a/MONG}');
Error('\P{/a/MONG}');
Expect(1, 6314, '\p{mong}', "");
Expect(0, 6314, '\p{^mong}', "");
Expect(0, 6314, '\P{mong}', "");
Expect(1, 6314, '\P{^mong}', "");
Expect(0, 6315, '\p{mong}', "");
Expect(1, 6315, '\p{^mong}', "");
Expect(1, 6315, '\P{mong}', "");
Expect(0, 6315, '\P{^mong}', "");
Expect(1, 6314, '\p{-_Mong}', "");
Expect(0, 6314, '\p{^-_Mong}', "");
Expect(0, 6314, '\P{-_Mong}', "");
Expect(1, 6314, '\P{^-_Mong}', "");
Expect(0, 6315, '\p{-_Mong}', "");
Expect(1, 6315, '\p{^-_Mong}', "");
Expect(1, 6315, '\P{-_Mong}', "");
Expect(0, 6315, '\P{^-_Mong}', "");
Error('\p{/a/-Is_MONG}');
Error('\P{/a/-Is_MONG}');
Expect(1, 6314, '\p{ismong}', "");
Expect(0, 6314, '\p{^ismong}', "");
Expect(0, 6314, '\P{ismong}', "");
Expect(1, 6314, '\P{^ismong}', "");
Expect(0, 6315, '\p{ismong}', "");
Expect(1, 6315, '\p{^ismong}', "");
Expect(1, 6315, '\P{ismong}', "");
Expect(0, 6315, '\P{^ismong}', "");
Expect(1, 6314, '\p{-	is_Mong}', "");
Expect(0, 6314, '\p{^-	is_Mong}', "");
Expect(0, 6314, '\P{-	is_Mong}', "");
Expect(1, 6314, '\P{^-	is_Mong}', "");
Expect(0, 6315, '\p{-	is_Mong}', "");
Expect(1, 6315, '\p{^-	is_Mong}', "");
Expect(1, 6315, '\P{-	is_Mong}', "");
Expect(0, 6315, '\P{^-	is_Mong}', "");
Error('\p{-	Yijing_Hexagram_Symbols/a/}');
Error('\P{-	Yijing_Hexagram_Symbols/a/}');
Expect(1, 19967, '\p{yijinghexagramsymbols}', "");
Expect(0, 19967, '\p{^yijinghexagramsymbols}', "");
Expect(0, 19967, '\P{yijinghexagramsymbols}', "");
Expect(1, 19967, '\P{^yijinghexagramsymbols}', "");
Expect(0, 19968, '\p{yijinghexagramsymbols}', "");
Expect(1, 19968, '\p{^yijinghexagramsymbols}', "");
Expect(1, 19968, '\P{yijinghexagramsymbols}', "");
Expect(0, 19968, '\P{^yijinghexagramsymbols}', "");
Expect(1, 19967, '\p{		YIJING_Hexagram_Symbols}', "");
Expect(0, 19967, '\p{^		YIJING_Hexagram_Symbols}', "");
Expect(0, 19967, '\P{		YIJING_Hexagram_Symbols}', "");
Expect(1, 19967, '\P{^		YIJING_Hexagram_Symbols}', "");
Expect(0, 19968, '\p{		YIJING_Hexagram_Symbols}', "");
Expect(1, 19968, '\p{^		YIJING_Hexagram_Symbols}', "");
Expect(1, 19968, '\P{		YIJING_Hexagram_Symbols}', "");
Expect(0, 19968, '\P{^		YIJING_Hexagram_Symbols}', "");
Error('\p{_:=Is_YIJING_Hexagram_symbols}');
Error('\P{_:=Is_YIJING_Hexagram_symbols}');
Expect(1, 19967, '\p{isyijinghexagramsymbols}', "");
Expect(0, 19967, '\p{^isyijinghexagramsymbols}', "");
Expect(0, 19967, '\P{isyijinghexagramsymbols}', "");
Expect(1, 19967, '\P{^isyijinghexagramsymbols}', "");
Expect(0, 19968, '\p{isyijinghexagramsymbols}', "");
Expect(1, 19968, '\p{^isyijinghexagramsymbols}', "");
Expect(1, 19968, '\P{isyijinghexagramsymbols}', "");
Expect(0, 19968, '\P{^isyijinghexagramsymbols}', "");
Expect(1, 19967, '\p{  IS_Yijing_Hexagram_Symbols}', "");
Expect(0, 19967, '\p{^  IS_Yijing_Hexagram_Symbols}', "");
Expect(0, 19967, '\P{  IS_Yijing_Hexagram_Symbols}', "");
Expect(1, 19967, '\P{^  IS_Yijing_Hexagram_Symbols}', "");
Expect(0, 19968, '\p{  IS_Yijing_Hexagram_Symbols}', "");
Expect(1, 19968, '\p{^  IS_Yijing_Hexagram_Symbols}', "");
Expect(1, 19968, '\P{  IS_Yijing_Hexagram_Symbols}', "");
Expect(0, 19968, '\P{^  IS_Yijing_Hexagram_Symbols}', "");
Error('\p{ :=IN_YIJING_Hexagram_Symbols}');
Error('\P{ :=IN_YIJING_Hexagram_Symbols}');
Expect(1, 19967, '\p{inyijinghexagramsymbols}', "");
Expect(0, 19967, '\p{^inyijinghexagramsymbols}', "");
Expect(0, 19967, '\P{inyijinghexagramsymbols}', "");
Expect(1, 19967, '\P{^inyijinghexagramsymbols}', "");
Expect(0, 19968, '\p{inyijinghexagramsymbols}', "");
Expect(1, 19968, '\p{^inyijinghexagramsymbols}', "");
Expect(1, 19968, '\P{inyijinghexagramsymbols}', "");
Expect(0, 19968, '\P{^inyijinghexagramsymbols}', "");
Expect(1, 19967, '\p{In_Yijing_Hexagram_symbols}', "");
Expect(0, 19967, '\p{^In_Yijing_Hexagram_symbols}', "");
Expect(0, 19967, '\P{In_Yijing_Hexagram_symbols}', "");
Expect(1, 19967, '\P{^In_Yijing_Hexagram_symbols}', "");
Expect(0, 19968, '\p{In_Yijing_Hexagram_symbols}', "");
Expect(1, 19968, '\p{^In_Yijing_Hexagram_symbols}', "");
Expect(1, 19968, '\P{In_Yijing_Hexagram_symbols}', "");
Expect(0, 19968, '\P{^In_Yijing_Hexagram_symbols}', "");
Error('\p{ :=Yijing}');
Error('\P{ :=Yijing}');
Expect(1, 19967, '\p{yijing}', "");
Expect(0, 19967, '\p{^yijing}', "");
Expect(0, 19967, '\P{yijing}', "");
Expect(1, 19967, '\P{^yijing}', "");
Expect(0, 19968, '\p{yijing}', "");
Expect(1, 19968, '\p{^yijing}', "");
Expect(1, 19968, '\P{yijing}', "");
Expect(0, 19968, '\P{^yijing}', "");
Expect(1, 19967, '\p{_yijing}', "");
Expect(0, 19967, '\p{^_yijing}', "");
Expect(0, 19967, '\P{_yijing}', "");
Expect(1, 19967, '\P{^_yijing}', "");
Expect(0, 19968, '\p{_yijing}', "");
Expect(1, 19968, '\p{^_yijing}', "");
Expect(1, 19968, '\P{_yijing}', "");
Expect(0, 19968, '\P{^_yijing}', "");
Error('\p{  Is_Yijing/a/}');
Error('\P{  Is_Yijing/a/}');
Expect(1, 19967, '\p{isyijing}', "");
Expect(0, 19967, '\p{^isyijing}', "");
Expect(0, 19967, '\P{isyijing}', "");
Expect(1, 19967, '\P{^isyijing}', "");
Expect(0, 19968, '\p{isyijing}', "");
Expect(1, 19968, '\p{^isyijing}', "");
Expect(1, 19968, '\P{isyijing}', "");
Expect(0, 19968, '\P{^isyijing}', "");
Expect(1, 19967, '\p{ Is_YIJING}', "");
Expect(0, 19967, '\p{^ Is_YIJING}', "");
Expect(0, 19967, '\P{ Is_YIJING}', "");
Expect(1, 19967, '\P{^ Is_YIJING}', "");
Expect(0, 19968, '\p{ Is_YIJING}', "");
Expect(1, 19968, '\p{^ Is_YIJING}', "");
Expect(1, 19968, '\P{ Is_YIJING}', "");
Expect(0, 19968, '\P{^ Is_YIJING}', "");
Error('\p{IN_Yijing:=}');
Error('\P{IN_Yijing:=}');
Expect(1, 19967, '\p{inyijing}', "");
Expect(0, 19967, '\p{^inyijing}', "");
Expect(0, 19967, '\P{inyijing}', "");
Expect(1, 19967, '\P{^inyijing}', "");
Expect(0, 19968, '\p{inyijing}', "");
Expect(1, 19968, '\p{^inyijing}', "");
Expect(1, 19968, '\P{inyijing}', "");
Expect(0, 19968, '\P{^inyijing}', "");
Expect(1, 19967, '\p{		in_Yijing}', "");
Expect(0, 19967, '\p{^		in_Yijing}', "");
Expect(0, 19967, '\P{		in_Yijing}', "");
Expect(1, 19967, '\P{^		in_Yijing}', "");
Expect(0, 19968, '\p{		in_Yijing}', "");
Expect(1, 19968, '\p{^		in_Yijing}', "");
Expect(1, 19968, '\P{		in_Yijing}', "");
Expect(0, 19968, '\P{^		in_Yijing}', "");
Error('\p{-:=Cherokee}');
Error('\P{-:=Cherokee}');
Expect(1, 5108, '\p{cherokee}', "");
Expect(0, 5108, '\p{^cherokee}', "");
Expect(0, 5108, '\P{cherokee}', "");
Expect(1, 5108, '\P{^cherokee}', "");
Expect(0, 5109, '\p{cherokee}', "");
Expect(1, 5109, '\p{^cherokee}', "");
Expect(1, 5109, '\P{cherokee}', "");
Expect(0, 5109, '\P{^cherokee}', "");
Expect(1, 5108, '\p{ CHEROKEE}', "");
Expect(0, 5108, '\p{^ CHEROKEE}', "");
Expect(0, 5108, '\P{ CHEROKEE}', "");
Expect(1, 5108, '\P{^ CHEROKEE}', "");
Expect(0, 5109, '\p{ CHEROKEE}', "");
Expect(1, 5109, '\p{^ CHEROKEE}', "");
Expect(1, 5109, '\P{ CHEROKEE}', "");
Expect(0, 5109, '\P{^ CHEROKEE}', "");
Error('\p{_Is_CHEROKEE/a/}');
Error('\P{_Is_CHEROKEE/a/}');
Expect(1, 5108, '\p{ischerokee}', "");
Expect(0, 5108, '\p{^ischerokee}', "");
Expect(0, 5108, '\P{ischerokee}', "");
Expect(1, 5108, '\P{^ischerokee}', "");
Expect(0, 5109, '\p{ischerokee}', "");
Expect(1, 5109, '\p{^ischerokee}', "");
Expect(1, 5109, '\P{ischerokee}', "");
Expect(0, 5109, '\P{^ischerokee}', "");
Expect(1, 5108, '\p{ is_CHEROKEE}', "");
Expect(0, 5108, '\p{^ is_CHEROKEE}', "");
Expect(0, 5108, '\P{ is_CHEROKEE}', "");
Expect(1, 5108, '\P{^ is_CHEROKEE}', "");
Expect(0, 5109, '\p{ is_CHEROKEE}', "");
Expect(1, 5109, '\p{^ is_CHEROKEE}', "");
Expect(1, 5109, '\P{ is_CHEROKEE}', "");
Expect(0, 5109, '\P{^ is_CHEROKEE}', "");
Error('\p{-/a/Cher}');
Error('\P{-/a/Cher}');
Expect(1, 5108, '\p{cher}', "");
Expect(0, 5108, '\p{^cher}', "");
Expect(0, 5108, '\P{cher}', "");
Expect(1, 5108, '\P{^cher}', "");
Expect(0, 5109, '\p{cher}', "");
Expect(1, 5109, '\p{^cher}', "");
Expect(1, 5109, '\P{cher}', "");
Expect(0, 5109, '\P{^cher}', "");
Expect(1, 5108, '\p{	 Cher}', "");
Expect(0, 5108, '\p{^	 Cher}', "");
Expect(0, 5108, '\P{	 Cher}', "");
Expect(1, 5108, '\P{^	 Cher}', "");
Expect(0, 5109, '\p{	 Cher}', "");
Expect(1, 5109, '\p{^	 Cher}', "");
Expect(1, 5109, '\P{	 Cher}', "");
Expect(0, 5109, '\P{^	 Cher}', "");
Error('\p{/a/is_CHER}');
Error('\P{/a/is_CHER}');
Expect(1, 5108, '\p{ischer}', "");
Expect(0, 5108, '\p{^ischer}', "");
Expect(0, 5108, '\P{ischer}', "");
Expect(1, 5108, '\P{^ischer}', "");
Expect(0, 5109, '\p{ischer}', "");
Expect(1, 5109, '\p{^ischer}', "");
Expect(1, 5109, '\P{ischer}', "");
Expect(0, 5109, '\P{^ischer}', "");
Expect(1, 5108, '\p{_is_CHER}', "");
Expect(0, 5108, '\p{^_is_CHER}', "");
Expect(0, 5108, '\P{_is_CHER}', "");
Expect(1, 5108, '\P{^_is_CHER}', "");
Expect(0, 5109, '\p{_is_CHER}', "");
Expect(1, 5109, '\p{^_is_CHER}', "");
Expect(1, 5109, '\P{_is_CHER}', "");
Expect(0, 5109, '\P{^_is_CHER}', "");
Error('\p{- Samaritan/a/}');
Error('\P{- Samaritan/a/}');
Expect(1, 2110, '\p{samaritan}', "");
Expect(0, 2110, '\p{^samaritan}', "");
Expect(0, 2110, '\P{samaritan}', "");
Expect(1, 2110, '\P{^samaritan}', "");
Expect(0, 2111, '\p{samaritan}', "");
Expect(1, 2111, '\p{^samaritan}', "");
Expect(1, 2111, '\P{samaritan}', "");
Expect(0, 2111, '\P{^samaritan}', "");
Expect(1, 2110, '\p{ 	samaritan}', "");
Expect(0, 2110, '\p{^ 	samaritan}', "");
Expect(0, 2110, '\P{ 	samaritan}', "");
Expect(1, 2110, '\P{^ 	samaritan}', "");
Expect(0, 2111, '\p{ 	samaritan}', "");
Expect(1, 2111, '\p{^ 	samaritan}', "");
Expect(1, 2111, '\P{ 	samaritan}', "");
Expect(0, 2111, '\P{^ 	samaritan}', "");
Error('\p{/a/ -IS_SAMARITAN}');
Error('\P{/a/ -IS_SAMARITAN}');
Expect(1, 2110, '\p{issamaritan}', "");
Expect(0, 2110, '\p{^issamaritan}', "");
Expect(0, 2110, '\P{issamaritan}', "");
Expect(1, 2110, '\P{^issamaritan}', "");
Expect(0, 2111, '\p{issamaritan}', "");
Expect(1, 2111, '\p{^issamaritan}', "");
Expect(1, 2111, '\P{issamaritan}', "");
Expect(0, 2111, '\P{^issamaritan}', "");
Expect(1, 2110, '\p{ 	Is_samaritan}', "");
Expect(0, 2110, '\p{^ 	Is_samaritan}', "");
Expect(0, 2110, '\P{ 	Is_samaritan}', "");
Expect(1, 2110, '\P{^ 	Is_samaritan}', "");
Expect(0, 2111, '\p{ 	Is_samaritan}', "");
Expect(1, 2111, '\p{^ 	Is_samaritan}', "");
Expect(1, 2111, '\P{ 	Is_samaritan}', "");
Expect(0, 2111, '\P{^ 	Is_samaritan}', "");
Error('\p{:=		samr}');
Error('\P{:=		samr}');
Expect(1, 2110, '\p{samr}', "");
Expect(0, 2110, '\p{^samr}', "");
Expect(0, 2110, '\P{samr}', "");
Expect(1, 2110, '\P{^samr}', "");
Expect(0, 2111, '\p{samr}', "");
Expect(1, 2111, '\p{^samr}', "");
Expect(1, 2111, '\P{samr}', "");
Expect(0, 2111, '\P{^samr}', "");
Expect(1, 2110, '\p{-Samr}', "");
Expect(0, 2110, '\p{^-Samr}', "");
Expect(0, 2110, '\P{-Samr}', "");
Expect(1, 2110, '\P{^-Samr}', "");
Expect(0, 2111, '\p{-Samr}', "");
Expect(1, 2111, '\p{^-Samr}', "");
Expect(1, 2111, '\P{-Samr}', "");
Expect(0, 2111, '\P{^-Samr}', "");
Error('\p{/a/-_IS_SAMR}');
Error('\P{/a/-_IS_SAMR}');
Expect(1, 2110, '\p{issamr}', "");
Expect(0, 2110, '\p{^issamr}', "");
Expect(0, 2110, '\P{issamr}', "");
Expect(1, 2110, '\P{^issamr}', "");
Expect(0, 2111, '\p{issamr}', "");
Expect(1, 2111, '\p{^issamr}', "");
Expect(1, 2111, '\P{issamr}', "");
Expect(0, 2111, '\P{^issamr}', "");
Expect(1, 2110, '\p{_Is_Samr}', "");
Expect(0, 2110, '\p{^_Is_Samr}', "");
Expect(0, 2110, '\P{_Is_Samr}', "");
Expect(1, 2110, '\P{^_Is_Samr}', "");
Expect(0, 2111, '\p{_Is_Samr}', "");
Expect(1, 2111, '\p{^_Is_Samr}', "");
Expect(1, 2111, '\P{_Is_Samr}', "");
Expect(0, 2111, '\P{^_Is_Samr}', "");
Error('\p{/a/-Other_punctuation}');
Error('\P{/a/-Other_punctuation}');
Expect(1, 74867, '\p{otherpunctuation}', "");
Expect(0, 74867, '\p{^otherpunctuation}', "");
Expect(0, 74867, '\P{otherpunctuation}', "");
Expect(1, 74867, '\P{^otherpunctuation}', "");
Expect(0, 74868, '\p{otherpunctuation}', "");
Expect(1, 74868, '\p{^otherpunctuation}', "");
Expect(1, 74868, '\P{otherpunctuation}', "");
Expect(0, 74868, '\P{^otherpunctuation}', "");
Expect(1, 74867, '\p{  Other_Punctuation}', "");
Expect(0, 74867, '\p{^  Other_Punctuation}', "");
Expect(0, 74867, '\P{  Other_Punctuation}', "");
Expect(1, 74867, '\P{^  Other_Punctuation}', "");
Expect(0, 74868, '\p{  Other_Punctuation}', "");
Expect(1, 74868, '\p{^  Other_Punctuation}', "");
Expect(1, 74868, '\P{  Other_Punctuation}', "");
Expect(0, 74868, '\P{^  Other_Punctuation}', "");
Error('\p{ /a/IS_other_Punctuation}');
Error('\P{ /a/IS_other_Punctuation}');
Expect(1, 74867, '\p{isotherpunctuation}', "");
Expect(0, 74867, '\p{^isotherpunctuation}', "");
Expect(0, 74867, '\P{isotherpunctuation}', "");
Expect(1, 74867, '\P{^isotherpunctuation}', "");
Expect(0, 74868, '\p{isotherpunctuation}', "");
Expect(1, 74868, '\p{^isotherpunctuation}', "");
Expect(1, 74868, '\P{isotherpunctuation}', "");
Expect(0, 74868, '\P{^isotherpunctuation}', "");
Expect(1, 74867, '\p{-Is_Other_Punctuation}', "");
Expect(0, 74867, '\p{^-Is_Other_Punctuation}', "");
Expect(0, 74867, '\P{-Is_Other_Punctuation}', "");
Expect(1, 74867, '\P{^-Is_Other_Punctuation}', "");
Expect(0, 74868, '\p{-Is_Other_Punctuation}', "");
Expect(1, 74868, '\p{^-Is_Other_Punctuation}', "");
Expect(1, 74868, '\P{-Is_Other_Punctuation}', "");
Expect(0, 74868, '\P{^-Is_Other_Punctuation}', "");
Error('\p{ Po/a/}');
Error('\P{ Po/a/}');
Expect(1, 74867, '\p{po}', "");
Expect(0, 74867, '\p{^po}', "");
Expect(0, 74867, '\P{po}', "");
Expect(1, 74867, '\P{^po}', "");
Expect(0, 74868, '\p{po}', "");
Expect(1, 74868, '\p{^po}', "");
Expect(1, 74868, '\P{po}', "");
Expect(0, 74868, '\P{^po}', "");
Expect(1, 74867, '\p{- Po}', "");
Expect(0, 74867, '\p{^- Po}', "");
Expect(0, 74867, '\P{- Po}', "");
Expect(1, 74867, '\P{^- Po}', "");
Expect(0, 74868, '\p{- Po}', "");
Expect(1, 74868, '\p{^- Po}', "");
Expect(1, 74868, '\P{- Po}', "");
Expect(0, 74868, '\P{^- Po}', "");
Error('\p{ /a/IS_Po}');
Error('\P{ /a/IS_Po}');
Expect(1, 74867, '\p{ispo}', "");
Expect(0, 74867, '\p{^ispo}', "");
Expect(0, 74867, '\P{ispo}', "");
Expect(1, 74867, '\P{^ispo}', "");
Expect(0, 74868, '\p{ispo}', "");
Expect(1, 74868, '\p{^ispo}', "");
Expect(1, 74868, '\P{ispo}', "");
Expect(0, 74868, '\P{^ispo}', "");
Expect(1, 74867, '\p{	Is_Po}', "");
Expect(0, 74867, '\p{^	Is_Po}', "");
Expect(0, 74867, '\P{	Is_Po}', "");
Expect(1, 74867, '\P{^	Is_Po}', "");
Expect(0, 74868, '\p{	Is_Po}', "");
Expect(1, 74868, '\p{^	Is_Po}', "");
Expect(1, 74868, '\P{	Is_Po}', "");
Expect(0, 74868, '\P{^	Is_Po}', "");
Error('\p{_:=Case_Ignorable}');
Error('\P{_:=Case_Ignorable}');
Expect(1, 917999, '\p{caseignorable}', "");
Expect(0, 917999, '\p{^caseignorable}', "");
Expect(0, 917999, '\P{caseignorable}', "");
Expect(1, 917999, '\P{^caseignorable}', "");
Expect(0, 918000, '\p{caseignorable}', "");
Expect(1, 918000, '\p{^caseignorable}', "");
Expect(1, 918000, '\P{caseignorable}', "");
Expect(0, 918000, '\P{^caseignorable}', "");
Expect(1, 917999, '\p{-CASE_IGNORABLE}', "");
Expect(0, 917999, '\p{^-CASE_IGNORABLE}', "");
Expect(0, 917999, '\P{-CASE_IGNORABLE}', "");
Expect(1, 917999, '\P{^-CASE_IGNORABLE}', "");
Expect(0, 918000, '\p{-CASE_IGNORABLE}', "");
Expect(1, 918000, '\p{^-CASE_IGNORABLE}', "");
Expect(1, 918000, '\P{-CASE_IGNORABLE}', "");
Expect(0, 918000, '\P{^-CASE_IGNORABLE}', "");
Error('\p{:=	IS_CASE_Ignorable}');
Error('\P{:=	IS_CASE_Ignorable}');
Expect(1, 917999, '\p{iscaseignorable}', "");
Expect(0, 917999, '\p{^iscaseignorable}', "");
Expect(0, 917999, '\P{iscaseignorable}', "");
Expect(1, 917999, '\P{^iscaseignorable}', "");
Expect(0, 918000, '\p{iscaseignorable}', "");
Expect(1, 918000, '\p{^iscaseignorable}', "");
Expect(1, 918000, '\P{iscaseignorable}', "");
Expect(0, 918000, '\P{^iscaseignorable}', "");
Expect(1, 917999, '\p{  is_CASE_Ignorable}', "");
Expect(0, 917999, '\p{^  is_CASE_Ignorable}', "");
Expect(0, 917999, '\P{  is_CASE_Ignorable}', "");
Expect(1, 917999, '\P{^  is_CASE_Ignorable}', "");
Expect(0, 918000, '\p{  is_CASE_Ignorable}', "");
Expect(1, 918000, '\p{^  is_CASE_Ignorable}', "");
Expect(1, 918000, '\P{  is_CASE_Ignorable}', "");
Expect(0, 918000, '\P{^  is_CASE_Ignorable}', "");
Error('\p{-:=ci}');
Error('\P{-:=ci}');
Expect(1, 917999, '\p{ci}', "");
Expect(0, 917999, '\p{^ci}', "");
Expect(0, 917999, '\P{ci}', "");
Expect(1, 917999, '\P{^ci}', "");
Expect(0, 918000, '\p{ci}', "");
Expect(1, 918000, '\p{^ci}', "");
Expect(1, 918000, '\P{ci}', "");
Expect(0, 918000, '\P{^ci}', "");
Expect(1, 917999, '\p{- CI}', "");
Expect(0, 917999, '\p{^- CI}', "");
Expect(0, 917999, '\P{- CI}', "");
Expect(1, 917999, '\P{^- CI}', "");
Expect(0, 918000, '\p{- CI}', "");
Expect(1, 918000, '\p{^- CI}', "");
Expect(1, 918000, '\P{- CI}', "");
Expect(0, 918000, '\P{^- CI}', "");
Error('\p{/a/IS_CI}');
Error('\P{/a/IS_CI}');
Expect(1, 917999, '\p{isci}', "");
Expect(0, 917999, '\p{^isci}', "");
Expect(0, 917999, '\P{isci}', "");
Expect(1, 917999, '\P{^isci}', "");
Expect(0, 918000, '\p{isci}', "");
Expect(1, 918000, '\p{^isci}', "");
Expect(1, 918000, '\P{isci}', "");
Expect(0, 918000, '\P{^isci}', "");
Expect(1, 917999, '\p{_ IS_CI}', "");
Expect(0, 917999, '\p{^_ IS_CI}', "");
Expect(0, 917999, '\P{_ IS_CI}', "");
Expect(1, 917999, '\P{^_ IS_CI}', "");
Expect(0, 918000, '\p{_ IS_CI}', "");
Expect(1, 918000, '\p{^_ IS_CI}', "");
Expect(1, 918000, '\P{_ IS_CI}', "");
Expect(0, 918000, '\P{^_ IS_CI}', "");
Error('\p{-:=_Case_Ignorable}');
Error('\P{-:=_Case_Ignorable}');
Expect(1, 917999, '\p{_Case_ignorable}', "");
Expect(0, 917999, '\p{^_Case_ignorable}', "");
Expect(0, 917999, '\P{_Case_ignorable}', "");
Expect(1, 917999, '\P{^_Case_ignorable}', "");
Expect(0, 918000, '\p{_Case_ignorable}', "");
Expect(1, 918000, '\p{^_Case_ignorable}', "");
Expect(1, 918000, '\P{_Case_ignorable}', "");
Expect(0, 918000, '\P{^_Case_ignorable}', "");
Error('\p{/a/Changes_When_nfkc_casefolded}');
Error('\P{/a/Changes_When_nfkc_casefolded}');
Expect(1, 921599, '\p{changeswhennfkccasefolded}', "");
Expect(0, 921599, '\p{^changeswhennfkccasefolded}', "");
Expect(0, 921599, '\P{changeswhennfkccasefolded}', "");
Expect(1, 921599, '\P{^changeswhennfkccasefolded}', "");
Expect(0, 921600, '\p{changeswhennfkccasefolded}', "");
Expect(1, 921600, '\p{^changeswhennfkccasefolded}', "");
Expect(1, 921600, '\P{changeswhennfkccasefolded}', "");
Expect(0, 921600, '\P{^changeswhennfkccasefolded}', "");
Expect(1, 921599, '\p{  Changes_when_NFKC_casefolded}', "");
Expect(0, 921599, '\p{^  Changes_when_NFKC_casefolded}', "");
Expect(0, 921599, '\P{  Changes_when_NFKC_casefolded}', "");
Expect(1, 921599, '\P{^  Changes_when_NFKC_casefolded}', "");
Expect(0, 921600, '\p{  Changes_when_NFKC_casefolded}', "");
Expect(1, 921600, '\p{^  Changes_when_NFKC_casefolded}', "");
Expect(1, 921600, '\P{  Changes_when_NFKC_casefolded}', "");
Expect(0, 921600, '\P{^  Changes_when_NFKC_casefolded}', "");
Error('\p{_/a/Is_changes_when_NFKC_casefolded}');
Error('\P{_/a/Is_changes_when_NFKC_casefolded}');
Expect(1, 921599, '\p{ischangeswhennfkccasefolded}', "");
Expect(0, 921599, '\p{^ischangeswhennfkccasefolded}', "");
Expect(0, 921599, '\P{ischangeswhennfkccasefolded}', "");
Expect(1, 921599, '\P{^ischangeswhennfkccasefolded}', "");
Expect(0, 921600, '\p{ischangeswhennfkccasefolded}', "");
Expect(1, 921600, '\p{^ischangeswhennfkccasefolded}', "");
Expect(1, 921600, '\P{ischangeswhennfkccasefolded}', "");
Expect(0, 921600, '\P{^ischangeswhennfkccasefolded}', "");
Expect(1, 921599, '\p{_-IS_Changes_when_NFKC_Casefolded}', "");
Expect(0, 921599, '\p{^_-IS_Changes_when_NFKC_Casefolded}', "");
Expect(0, 921599, '\P{_-IS_Changes_when_NFKC_Casefolded}', "");
Expect(1, 921599, '\P{^_-IS_Changes_when_NFKC_Casefolded}', "");
Expect(0, 921600, '\p{_-IS_Changes_when_NFKC_Casefolded}', "");
Expect(1, 921600, '\p{^_-IS_Changes_when_NFKC_Casefolded}', "");
Expect(1, 921600, '\P{_-IS_Changes_when_NFKC_Casefolded}', "");
Expect(0, 921600, '\P{^_-IS_Changes_when_NFKC_Casefolded}', "");
Error('\p{ cwkcf:=}');
Error('\P{ cwkcf:=}');
Expect(1, 921599, '\p{cwkcf}', "");
Expect(0, 921599, '\p{^cwkcf}', "");
Expect(0, 921599, '\P{cwkcf}', "");
Expect(1, 921599, '\P{^cwkcf}', "");
Expect(0, 921600, '\p{cwkcf}', "");
Expect(1, 921600, '\p{^cwkcf}', "");
Expect(1, 921600, '\P{cwkcf}', "");
Expect(0, 921600, '\P{^cwkcf}', "");
Expect(1, 921599, '\p{_	CWKCF}', "");
Expect(0, 921599, '\p{^_	CWKCF}', "");
Expect(0, 921599, '\P{_	CWKCF}', "");
Expect(1, 921599, '\P{^_	CWKCF}', "");
Expect(0, 921600, '\p{_	CWKCF}', "");
Expect(1, 921600, '\p{^_	CWKCF}', "");
Expect(1, 921600, '\P{_	CWKCF}', "");
Expect(0, 921600, '\P{^_	CWKCF}', "");
Error('\p{/a/IS_CWKCF}');
Error('\P{/a/IS_CWKCF}');
Expect(1, 921599, '\p{iscwkcf}', "");
Expect(0, 921599, '\p{^iscwkcf}', "");
Expect(0, 921599, '\P{iscwkcf}', "");
Expect(1, 921599, '\P{^iscwkcf}', "");
Expect(0, 921600, '\p{iscwkcf}', "");
Expect(1, 921600, '\p{^iscwkcf}', "");
Expect(1, 921600, '\P{iscwkcf}', "");
Expect(0, 921600, '\P{^iscwkcf}', "");
Expect(1, 921599, '\p{ is_CWKCF}', "");
Expect(0, 921599, '\p{^ is_CWKCF}', "");
Expect(0, 921599, '\P{ is_CWKCF}', "");
Expect(1, 921599, '\P{^ is_CWKCF}', "");
Expect(0, 921600, '\p{ is_CWKCF}', "");
Expect(1, 921600, '\p{^ is_CWKCF}', "");
Expect(1, 921600, '\P{ is_CWKCF}', "");
Expect(0, 921600, '\P{^ is_CWKCF}', "");
Error('\p{_/a/in_Cherokee}');
Error('\P{_/a/in_Cherokee}');
Expect(1, 5119, '\p{incherokee}', "");
Expect(0, 5119, '\p{^incherokee}', "");
Expect(0, 5119, '\P{incherokee}', "");
Expect(1, 5119, '\P{^incherokee}', "");
Expect(0, 5120, '\p{incherokee}', "");
Expect(1, 5120, '\p{^incherokee}', "");
Expect(1, 5120, '\P{incherokee}', "");
Expect(0, 5120, '\P{^incherokee}', "");
Expect(1, 5119, '\p{ -IN_Cherokee}', "");
Expect(0, 5119, '\p{^ -IN_Cherokee}', "");
Expect(0, 5119, '\P{ -IN_Cherokee}', "");
Expect(1, 5119, '\P{^ -IN_Cherokee}', "");
Expect(0, 5120, '\p{ -IN_Cherokee}', "");
Expect(1, 5120, '\p{^ -IN_Cherokee}', "");
Expect(1, 5120, '\P{ -IN_Cherokee}', "");
Expect(0, 5120, '\P{^ -IN_Cherokee}', "");
Error('\p{ :=In_CHEROKEE}');
Error('\P{ :=In_CHEROKEE}');
Expect(1, 5119, '\p{ -in_Cherokee}', "");
Expect(0, 5119, '\p{^ -in_Cherokee}', "");
Expect(0, 5119, '\P{ -in_Cherokee}', "");
Expect(1, 5119, '\P{^ -in_Cherokee}', "");
Expect(0, 5120, '\p{ -in_Cherokee}', "");
Expect(1, 5120, '\p{^ -in_Cherokee}', "");
Expect(1, 5120, '\P{ -in_Cherokee}', "");
Expect(0, 5120, '\P{^ -in_Cherokee}', "");
Error('\p{:=ETHIOPIC}');
Error('\P{:=ETHIOPIC}');
Expect(1, 43822, '\p{ethiopic}', "");
Expect(0, 43822, '\p{^ethiopic}', "");
Expect(0, 43822, '\P{ethiopic}', "");
Expect(1, 43822, '\P{^ethiopic}', "");
Expect(0, 43823, '\p{ethiopic}', "");
Expect(1, 43823, '\p{^ethiopic}', "");
Expect(1, 43823, '\P{ethiopic}', "");
Expect(0, 43823, '\P{^ethiopic}', "");
Expect(1, 43822, '\p{	ethiopic}', "");
Expect(0, 43822, '\p{^	ethiopic}', "");
Expect(0, 43822, '\P{	ethiopic}', "");
Expect(1, 43822, '\P{^	ethiopic}', "");
Expect(0, 43823, '\p{	ethiopic}', "");
Expect(1, 43823, '\p{^	ethiopic}', "");
Expect(1, 43823, '\P{	ethiopic}', "");
Expect(0, 43823, '\P{^	ethiopic}', "");
Error('\p{/a/	_Is_Ethiopic}');
Error('\P{/a/	_Is_Ethiopic}');
Expect(1, 43822, '\p{isethiopic}', "");
Expect(0, 43822, '\p{^isethiopic}', "");
Expect(0, 43822, '\P{isethiopic}', "");
Expect(1, 43822, '\P{^isethiopic}', "");
Expect(0, 43823, '\p{isethiopic}', "");
Expect(1, 43823, '\p{^isethiopic}', "");
Expect(1, 43823, '\P{isethiopic}', "");
Expect(0, 43823, '\P{^isethiopic}', "");
Expect(1, 43822, '\p{-Is_ethiopic}', "");
Expect(0, 43822, '\p{^-Is_ethiopic}', "");
Expect(0, 43822, '\P{-Is_ethiopic}', "");
Expect(1, 43822, '\P{^-Is_ethiopic}', "");
Expect(0, 43823, '\p{-Is_ethiopic}', "");
Expect(1, 43823, '\p{^-Is_ethiopic}', "");
Expect(1, 43823, '\P{-Is_ethiopic}', "");
Expect(0, 43823, '\P{^-Is_ethiopic}', "");
Error('\p{ /a/Ethi}');
Error('\P{ /a/Ethi}');
Expect(1, 43822, '\p{ethi}', "");
Expect(0, 43822, '\p{^ethi}', "");
Expect(0, 43822, '\P{ethi}', "");
Expect(1, 43822, '\P{^ethi}', "");
Expect(0, 43823, '\p{ethi}', "");
Expect(1, 43823, '\p{^ethi}', "");
Expect(1, 43823, '\P{ethi}', "");
Expect(0, 43823, '\P{^ethi}', "");
Expect(1, 43822, '\p{ ETHI}', "");
Expect(0, 43822, '\p{^ ETHI}', "");
Expect(0, 43822, '\P{ ETHI}', "");
Expect(1, 43822, '\P{^ ETHI}', "");
Expect(0, 43823, '\p{ ETHI}', "");
Expect(1, 43823, '\p{^ ETHI}', "");
Expect(1, 43823, '\P{ ETHI}', "");
Expect(0, 43823, '\P{^ ETHI}', "");
Error('\p{/a/ _Is_ETHI}');
Error('\P{/a/ _Is_ETHI}');
Expect(1, 43822, '\p{isethi}', "");
Expect(0, 43822, '\p{^isethi}', "");
Expect(0, 43822, '\P{isethi}', "");
Expect(1, 43822, '\P{^isethi}', "");
Expect(0, 43823, '\p{isethi}', "");
Expect(1, 43823, '\p{^isethi}', "");
Expect(1, 43823, '\P{isethi}', "");
Expect(0, 43823, '\P{^isethi}', "");
Expect(1, 43822, '\p{ Is_ETHI}', "");
Expect(0, 43822, '\p{^ Is_ETHI}', "");
Expect(0, 43822, '\P{ Is_ETHI}', "");
Expect(1, 43822, '\P{^ Is_ETHI}', "");
Expect(0, 43823, '\p{ Is_ETHI}', "");
Expect(1, 43823, '\p{^ Is_ETHI}', "");
Expect(1, 43823, '\P{ Is_ETHI}', "");
Expect(0, 43823, '\P{^ Is_ETHI}', "");
Error('\p{-_X_Begin:=}');
Error('\P{-_X_Begin:=}');
Expect(1, 921600, '\p{_X_Begin}', "");
Expect(0, 921600, '\p{^_X_Begin}', "");
Expect(0, 921600, '\P{_X_Begin}', "");
Expect(1, 921600, '\P{^_X_Begin}', "");
Expect(0, 921599, '\p{_X_Begin}', "");
Expect(1, 921599, '\p{^_X_Begin}', "");
Expect(1, 921599, '\P{_X_Begin}', "");
Expect(0, 921599, '\P{^_X_Begin}', "");
Error('\p{-:=_X_BEGIN}');
Error('\P{-:=_X_BEGIN}');
Error('\p{	YI/a/}');
Error('\P{	YI/a/}');
Expect(1, 42182, '\p{yi}', "");
Expect(0, 42182, '\p{^yi}', "");
Expect(0, 42182, '\P{yi}', "");
Expect(1, 42182, '\P{^yi}', "");
Expect(0, 42183, '\p{yi}', "");
Expect(1, 42183, '\p{^yi}', "");
Expect(1, 42183, '\P{yi}', "");
Expect(0, 42183, '\P{^yi}', "");
Expect(1, 42182, '\p{_Yi}', "");
Expect(0, 42182, '\p{^_Yi}', "");
Expect(0, 42182, '\P{_Yi}', "");
Expect(1, 42182, '\P{^_Yi}', "");
Expect(0, 42183, '\p{_Yi}', "");
Expect(1, 42183, '\p{^_Yi}', "");
Expect(1, 42183, '\P{_Yi}', "");
Expect(0, 42183, '\P{^_Yi}', "");
Error('\p{ :=Is_Yi}');
Error('\P{ :=Is_Yi}');
Expect(1, 42182, '\p{isyi}', "");
Expect(0, 42182, '\p{^isyi}', "");
Expect(0, 42182, '\P{isyi}', "");
Expect(1, 42182, '\P{^isyi}', "");
Expect(0, 42183, '\p{isyi}', "");
Expect(1, 42183, '\p{^isyi}', "");
Expect(1, 42183, '\P{isyi}', "");
Expect(0, 42183, '\P{^isyi}', "");
Expect(1, 42182, '\p{_Is_Yi}', "");
Expect(0, 42182, '\p{^_Is_Yi}', "");
Expect(0, 42182, '\P{_Is_Yi}', "");
Expect(1, 42182, '\P{^_Is_Yi}', "");
Expect(0, 42183, '\p{_Is_Yi}', "");
Expect(1, 42183, '\p{^_Is_Yi}', "");
Expect(1, 42183, '\P{_Is_Yi}', "");
Expect(0, 42183, '\P{^_Is_Yi}', "");
Error('\p{:=YIII}');
Error('\P{:=YIII}');
Expect(1, 42182, '\p{yiii}', "");
Expect(0, 42182, '\p{^yiii}', "");
Expect(0, 42182, '\P{yiii}', "");
Expect(1, 42182, '\P{^yiii}', "");
Expect(0, 42183, '\p{yiii}', "");
Expect(1, 42183, '\p{^yiii}', "");
Expect(1, 42183, '\P{yiii}', "");
Expect(0, 42183, '\P{^yiii}', "");
Expect(1, 42182, '\p{_-yiii}', "");
Expect(0, 42182, '\p{^_-yiii}', "");
Expect(0, 42182, '\P{_-yiii}', "");
Expect(1, 42182, '\P{^_-yiii}', "");
Expect(0, 42183, '\p{_-yiii}', "");
Expect(1, 42183, '\p{^_-yiii}', "");
Expect(1, 42183, '\P{_-yiii}', "");
Expect(0, 42183, '\P{^_-yiii}', "");
Error('\p{_:=is_yiii}');
Error('\P{_:=is_yiii}');
Expect(1, 42182, '\p{isyiii}', "");
Expect(0, 42182, '\p{^isyiii}', "");
Expect(0, 42182, '\P{isyiii}', "");
Expect(1, 42182, '\P{^isyiii}', "");
Expect(0, 42183, '\p{isyiii}', "");
Expect(1, 42183, '\p{^isyiii}', "");
Expect(1, 42183, '\P{isyiii}', "");
Expect(0, 42183, '\P{^isyiii}', "");
Expect(1, 42182, '\p{__Is_Yiii}', "");
Expect(0, 42182, '\p{^__Is_Yiii}', "");
Expect(0, 42182, '\P{__Is_Yiii}', "");
Expect(1, 42182, '\P{^__Is_Yiii}', "");
Expect(0, 42183, '\p{__Is_Yiii}', "");
Expect(1, 42183, '\p{^__Is_Yiii}', "");
Expect(1, 42183, '\P{__Is_Yiii}', "");
Expect(0, 42183, '\P{^__Is_Yiii}', "");
Error('\p{/a/_-Gurmukhi}');
Error('\P{/a/_-Gurmukhi}');
Expect(1, 2677, '\p{gurmukhi}', "");
Expect(0, 2677, '\p{^gurmukhi}', "");
Expect(0, 2677, '\P{gurmukhi}', "");
Expect(1, 2677, '\P{^gurmukhi}', "");
Expect(0, 2678, '\p{gurmukhi}', "");
Expect(1, 2678, '\p{^gurmukhi}', "");
Expect(1, 2678, '\P{gurmukhi}', "");
Expect(0, 2678, '\P{^gurmukhi}', "");
Expect(1, 2677, '\p{ _Gurmukhi}', "");
Expect(0, 2677, '\p{^ _Gurmukhi}', "");
Expect(0, 2677, '\P{ _Gurmukhi}', "");
Expect(1, 2677, '\P{^ _Gurmukhi}', "");
Expect(0, 2678, '\p{ _Gurmukhi}', "");
Expect(1, 2678, '\p{^ _Gurmukhi}', "");
Expect(1, 2678, '\P{ _Gurmukhi}', "");
Expect(0, 2678, '\P{^ _Gurmukhi}', "");
Error('\p{:=__Is_Gurmukhi}');
Error('\P{:=__Is_Gurmukhi}');
Expect(1, 2677, '\p{isgurmukhi}', "");
Expect(0, 2677, '\p{^isgurmukhi}', "");
Expect(0, 2677, '\P{isgurmukhi}', "");
Expect(1, 2677, '\P{^isgurmukhi}', "");
Expect(0, 2678, '\p{isgurmukhi}', "");
Expect(1, 2678, '\p{^isgurmukhi}', "");
Expect(1, 2678, '\P{isgurmukhi}', "");
Expect(0, 2678, '\P{^isgurmukhi}', "");
Expect(1, 2677, '\p{ IS_gurmukhi}', "");
Expect(0, 2677, '\p{^ IS_gurmukhi}', "");
Expect(0, 2677, '\P{ IS_gurmukhi}', "");
Expect(1, 2677, '\P{^ IS_gurmukhi}', "");
Expect(0, 2678, '\p{ IS_gurmukhi}', "");
Expect(1, 2678, '\p{^ IS_gurmukhi}', "");
Expect(1, 2678, '\P{ IS_gurmukhi}', "");
Expect(0, 2678, '\P{^ IS_gurmukhi}', "");
Error('\p{/a/-Guru}');
Error('\P{/a/-Guru}');
Expect(1, 2677, '\p{guru}', "");
Expect(0, 2677, '\p{^guru}', "");
Expect(0, 2677, '\P{guru}', "");
Expect(1, 2677, '\P{^guru}', "");
Expect(0, 2678, '\p{guru}', "");
Expect(1, 2678, '\p{^guru}', "");
Expect(1, 2678, '\P{guru}', "");
Expect(0, 2678, '\P{^guru}', "");
Expect(1, 2677, '\p{ guru}', "");
Expect(0, 2677, '\p{^ guru}', "");
Expect(0, 2677, '\P{ guru}', "");
Expect(1, 2677, '\P{^ guru}', "");
Expect(0, 2678, '\p{ guru}', "");
Expect(1, 2678, '\p{^ guru}', "");
Expect(1, 2678, '\P{ guru}', "");
Expect(0, 2678, '\P{^ guru}', "");
Error('\p{	_Is_guru:=}');
Error('\P{	_Is_guru:=}');
Expect(1, 2677, '\p{isguru}', "");
Expect(0, 2677, '\p{^isguru}', "");
Expect(0, 2677, '\P{isguru}', "");
Expect(1, 2677, '\P{^isguru}', "");
Expect(0, 2678, '\p{isguru}', "");
Expect(1, 2678, '\p{^isguru}', "");
Expect(1, 2678, '\P{isguru}', "");
Expect(0, 2678, '\P{^isguru}', "");
Expect(1, 2677, '\p{_is_Guru}', "");
Expect(0, 2677, '\p{^_is_Guru}', "");
Expect(0, 2677, '\P{_is_Guru}', "");
Expect(1, 2677, '\P{^_is_Guru}', "");
Expect(0, 2678, '\p{_is_Guru}', "");
Expect(1, 2678, '\p{^_is_Guru}', "");
Expect(1, 2678, '\P{_is_Guru}', "");
Expect(0, 2678, '\P{^_is_Guru}', "");
Error('\p{ /a/Cyrillic}');
Error('\P{ /a/Cyrillic}');
Expect(1, 42655, '\p{cyrillic}', "");
Expect(0, 42655, '\p{^cyrillic}', "");
Expect(0, 42655, '\P{cyrillic}', "");
Expect(1, 42655, '\P{^cyrillic}', "");
Expect(0, 42656, '\p{cyrillic}', "");
Expect(1, 42656, '\p{^cyrillic}', "");
Expect(1, 42656, '\P{cyrillic}', "");
Expect(0, 42656, '\P{^cyrillic}', "");
Expect(1, 42655, '\p{	_Cyrillic}', "");
Expect(0, 42655, '\p{^	_Cyrillic}', "");
Expect(0, 42655, '\P{	_Cyrillic}', "");
Expect(1, 42655, '\P{^	_Cyrillic}', "");
Expect(0, 42656, '\p{	_Cyrillic}', "");
Expect(1, 42656, '\p{^	_Cyrillic}', "");
Expect(1, 42656, '\P{	_Cyrillic}', "");
Expect(0, 42656, '\P{^	_Cyrillic}', "");
Error('\p{:=_is_Cyrillic}');
Error('\P{:=_is_Cyrillic}');
Expect(1, 42655, '\p{iscyrillic}', "");
Expect(0, 42655, '\p{^iscyrillic}', "");
Expect(0, 42655, '\P{iscyrillic}', "");
Expect(1, 42655, '\P{^iscyrillic}', "");
Expect(0, 42656, '\p{iscyrillic}', "");
Expect(1, 42656, '\p{^iscyrillic}', "");
Expect(1, 42656, '\P{iscyrillic}', "");
Expect(0, 42656, '\P{^iscyrillic}', "");
Expect(1, 42655, '\p{__Is_CYRILLIC}', "");
Expect(0, 42655, '\p{^__Is_CYRILLIC}', "");
Expect(0, 42655, '\P{__Is_CYRILLIC}', "");
Expect(1, 42655, '\P{^__Is_CYRILLIC}', "");
Expect(0, 42656, '\p{__Is_CYRILLIC}', "");
Expect(1, 42656, '\p{^__Is_CYRILLIC}', "");
Expect(1, 42656, '\P{__Is_CYRILLIC}', "");
Expect(0, 42656, '\P{^__Is_CYRILLIC}', "");
Error('\p{_/a/CYRL}');
Error('\P{_/a/CYRL}');
Expect(1, 42655, '\p{cyrl}', "");
Expect(0, 42655, '\p{^cyrl}', "");
Expect(0, 42655, '\P{cyrl}', "");
Expect(1, 42655, '\P{^cyrl}', "");
Expect(0, 42656, '\p{cyrl}', "");
Expect(1, 42656, '\p{^cyrl}', "");
Expect(1, 42656, '\P{cyrl}', "");
Expect(0, 42656, '\P{^cyrl}', "");
Expect(1, 42655, '\p{ Cyrl}', "");
Expect(0, 42655, '\p{^ Cyrl}', "");
Expect(0, 42655, '\P{ Cyrl}', "");
Expect(1, 42655, '\P{^ Cyrl}', "");
Expect(0, 42656, '\p{ Cyrl}', "");
Expect(1, 42656, '\p{^ Cyrl}', "");
Expect(1, 42656, '\P{ Cyrl}', "");
Expect(0, 42656, '\P{^ Cyrl}', "");
Error('\p{_:=Is_Cyrl}');
Error('\P{_:=Is_Cyrl}');
Expect(1, 42655, '\p{iscyrl}', "");
Expect(0, 42655, '\p{^iscyrl}', "");
Expect(0, 42655, '\P{iscyrl}', "");
Expect(1, 42655, '\P{^iscyrl}', "");
Expect(0, 42656, '\p{iscyrl}', "");
Expect(1, 42656, '\p{^iscyrl}', "");
Expect(1, 42656, '\P{iscyrl}', "");
Expect(0, 42656, '\P{^iscyrl}', "");
Expect(1, 42655, '\p{is_cyrl}', "");
Expect(0, 42655, '\p{^is_cyrl}', "");
Expect(0, 42655, '\P{is_cyrl}', "");
Expect(1, 42655, '\P{^is_cyrl}', "");
Expect(0, 42656, '\p{is_cyrl}', "");
Expect(1, 42656, '\p{^is_cyrl}', "");
Expect(1, 42656, '\P{is_cyrl}', "");
Expect(0, 42656, '\P{^is_cyrl}', "");
Error('\p{:=__latin_Extended_C}');
Error('\P{:=__latin_Extended_C}');
Expect(1, 11391, '\p{latinextendedc}', "");
Expect(0, 11391, '\p{^latinextendedc}', "");
Expect(0, 11391, '\P{latinextendedc}', "");
Expect(1, 11391, '\P{^latinextendedc}', "");
Expect(0, 11392, '\p{latinextendedc}', "");
Expect(1, 11392, '\p{^latinextendedc}', "");
Expect(1, 11392, '\P{latinextendedc}', "");
Expect(0, 11392, '\P{^latinextendedc}', "");
Expect(1, 11391, '\p{ _LATIN_Extended_c}', "");
Expect(0, 11391, '\p{^ _LATIN_Extended_c}', "");
Expect(0, 11391, '\P{ _LATIN_Extended_c}', "");
Expect(1, 11391, '\P{^ _LATIN_Extended_c}', "");
Expect(0, 11392, '\p{ _LATIN_Extended_c}', "");
Expect(1, 11392, '\p{^ _LATIN_Extended_c}', "");
Expect(1, 11392, '\P{ _LATIN_Extended_c}', "");
Expect(0, 11392, '\P{^ _LATIN_Extended_c}', "");
Error('\p{:=is_Latin_extended_C}');
Error('\P{:=is_Latin_extended_C}');
Expect(1, 11391, '\p{islatinextendedc}', "");
Expect(0, 11391, '\p{^islatinextendedc}', "");
Expect(0, 11391, '\P{islatinextendedc}', "");
Expect(1, 11391, '\P{^islatinextendedc}', "");
Expect(0, 11392, '\p{islatinextendedc}', "");
Expect(1, 11392, '\p{^islatinextendedc}', "");
Expect(1, 11392, '\P{islatinextendedc}', "");
Expect(0, 11392, '\P{^islatinextendedc}', "");
Expect(1, 11391, '\p{-	Is_LATIN_EXTENDED_C}', "");
Expect(0, 11391, '\p{^-	Is_LATIN_EXTENDED_C}', "");
Expect(0, 11391, '\P{-	Is_LATIN_EXTENDED_C}', "");
Expect(1, 11391, '\P{^-	Is_LATIN_EXTENDED_C}', "");
Expect(0, 11392, '\p{-	Is_LATIN_EXTENDED_C}', "");
Expect(1, 11392, '\p{^-	Is_LATIN_EXTENDED_C}', "");
Expect(1, 11392, '\P{-	Is_LATIN_EXTENDED_C}', "");
Expect(0, 11392, '\P{^-	Is_LATIN_EXTENDED_C}', "");
Error('\p{	/a/In_LATIN_EXTENDED_C}');
Error('\P{	/a/In_LATIN_EXTENDED_C}');
Expect(1, 11391, '\p{inlatinextendedc}', "");
Expect(0, 11391, '\p{^inlatinextendedc}', "");
Expect(0, 11391, '\P{inlatinextendedc}', "");
Expect(1, 11391, '\P{^inlatinextendedc}', "");
Expect(0, 11392, '\p{inlatinextendedc}', "");
Expect(1, 11392, '\p{^inlatinextendedc}', "");
Expect(1, 11392, '\P{inlatinextendedc}', "");
Expect(0, 11392, '\P{^inlatinextendedc}', "");
Expect(1, 11391, '\p{in_latin_Extended_c}', "");
Expect(0, 11391, '\p{^in_latin_Extended_c}', "");
Expect(0, 11391, '\P{in_latin_Extended_c}', "");
Expect(1, 11391, '\P{^in_latin_Extended_c}', "");
Expect(0, 11392, '\p{in_latin_Extended_c}', "");
Expect(1, 11392, '\p{^in_latin_Extended_c}', "");
Expect(1, 11392, '\P{in_latin_Extended_c}', "");
Expect(0, 11392, '\P{^in_latin_Extended_c}', "");
Error('\p{/a/-_Latin_Ext_C}');
Error('\P{/a/-_Latin_Ext_C}');
Expect(1, 11391, '\p{latinextc}', "");
Expect(0, 11391, '\p{^latinextc}', "");
Expect(0, 11391, '\P{latinextc}', "");
Expect(1, 11391, '\P{^latinextc}', "");
Expect(0, 11392, '\p{latinextc}', "");
Expect(1, 11392, '\p{^latinextc}', "");
Expect(1, 11392, '\P{latinextc}', "");
Expect(0, 11392, '\P{^latinextc}', "");
Expect(1, 11391, '\p{ 	Latin_ext_C}', "");
Expect(0, 11391, '\p{^ 	Latin_ext_C}', "");
Expect(0, 11391, '\P{ 	Latin_ext_C}', "");
Expect(1, 11391, '\P{^ 	Latin_ext_C}', "");
Expect(0, 11392, '\p{ 	Latin_ext_C}', "");
Expect(1, 11392, '\p{^ 	Latin_ext_C}', "");
Expect(1, 11392, '\P{ 	Latin_ext_C}', "");
Expect(0, 11392, '\P{^ 	Latin_ext_C}', "");
Error('\p{/a/-_IS_Latin_ext_C}');
Error('\P{/a/-_IS_Latin_ext_C}');
Expect(1, 11391, '\p{islatinextc}', "");
Expect(0, 11391, '\p{^islatinextc}', "");
Expect(0, 11391, '\P{islatinextc}', "");
Expect(1, 11391, '\P{^islatinextc}', "");
Expect(0, 11392, '\p{islatinextc}', "");
Expect(1, 11392, '\p{^islatinextc}', "");
Expect(1, 11392, '\P{islatinextc}', "");
Expect(0, 11392, '\P{^islatinextc}', "");
Expect(1, 11391, '\p{-_IS_LATIN_ext_c}', "");
Expect(0, 11391, '\p{^-_IS_LATIN_ext_c}', "");
Expect(0, 11391, '\P{-_IS_LATIN_ext_c}', "");
Expect(1, 11391, '\P{^-_IS_LATIN_ext_c}', "");
Expect(0, 11392, '\p{-_IS_LATIN_ext_c}', "");
Expect(1, 11392, '\p{^-_IS_LATIN_ext_c}', "");
Expect(1, 11392, '\P{-_IS_LATIN_ext_c}', "");
Expect(0, 11392, '\P{^-_IS_LATIN_ext_c}', "");
Error('\p{/a/		in_Latin_EXT_C}');
Error('\P{/a/		in_Latin_EXT_C}');
Expect(1, 11391, '\p{inlatinextc}', "");
Expect(0, 11391, '\p{^inlatinextc}', "");
Expect(0, 11391, '\P{inlatinextc}', "");
Expect(1, 11391, '\P{^inlatinextc}', "");
Expect(0, 11392, '\p{inlatinextc}', "");
Expect(1, 11392, '\p{^inlatinextc}', "");
Expect(1, 11392, '\P{inlatinextc}', "");
Expect(0, 11392, '\P{^inlatinextc}', "");
Expect(1, 11391, '\p{ _in_Latin_EXT_C}', "");
Expect(0, 11391, '\p{^ _in_Latin_EXT_C}', "");
Expect(0, 11391, '\P{ _in_Latin_EXT_C}', "");
Expect(1, 11391, '\P{^ _in_Latin_EXT_C}', "");
Expect(0, 11392, '\p{ _in_Latin_EXT_C}', "");
Expect(1, 11392, '\p{^ _in_Latin_EXT_C}', "");
Expect(1, 11392, '\P{ _in_Latin_EXT_C}', "");
Expect(0, 11392, '\P{^ _in_Latin_EXT_C}', "");
Error('\p{_-_Perl_Quotemeta:=}');
Error('\P{_-_Perl_Quotemeta:=}');
Expect(1, 921599, '\p{_PERL_Quotemeta}', "");
Expect(0, 921599, '\p{^_PERL_Quotemeta}', "");
Expect(0, 921599, '\P{_PERL_Quotemeta}', "");
Expect(1, 921599, '\P{^_PERL_Quotemeta}', "");
Expect(0, 921600, '\p{_PERL_Quotemeta}', "");
Expect(1, 921600, '\p{^_PERL_Quotemeta}', "");
Expect(1, 921600, '\P{_PERL_Quotemeta}', "");
Expect(0, 921600, '\P{^_PERL_Quotemeta}', "");
Error('\p{__Perl_Quotemeta/a/}');
Error('\P{__Perl_Quotemeta/a/}');
Expect(1, 921599, '\p{_Perl_Quotemeta}', "");
Expect(0, 921599, '\p{^_Perl_Quotemeta}', "");
Expect(0, 921599, '\P{_Perl_Quotemeta}', "");
Expect(1, 921599, '\P{^_Perl_Quotemeta}', "");
Expect(0, 921600, '\p{_Perl_Quotemeta}', "");
Expect(1, 921600, '\p{^_Perl_Quotemeta}', "");
Expect(1, 921600, '\P{_Perl_Quotemeta}', "");
Expect(0, 921600, '\P{^_Perl_Quotemeta}', "");
Error('\p{/a/	_In_rejang}');
Error('\P{/a/	_In_rejang}');
Expect(1, 43359, '\p{inrejang}', "");
Expect(0, 43359, '\p{^inrejang}', "");
Expect(0, 43359, '\P{inrejang}', "");
Expect(1, 43359, '\P{^inrejang}', "");
Expect(0, 43360, '\p{inrejang}', "");
Expect(1, 43360, '\p{^inrejang}', "");
Expect(1, 43360, '\P{inrejang}', "");
Expect(0, 43360, '\P{^inrejang}', "");
Expect(1, 43359, '\p{ -In_Rejang}', "");
Expect(0, 43359, '\p{^ -In_Rejang}', "");
Expect(0, 43359, '\P{ -In_Rejang}', "");
Expect(1, 43359, '\P{^ -In_Rejang}', "");
Expect(0, 43360, '\p{ -In_Rejang}', "");
Expect(1, 43360, '\p{^ -In_Rejang}', "");
Expect(1, 43360, '\P{ -In_Rejang}', "");
Expect(0, 43360, '\P{^ -In_Rejang}', "");
Error('\p{:= _in_Rejang}');
Error('\P{:= _in_Rejang}');
Expect(1, 43359, '\p{ -IN_Rejang}', "");
Expect(0, 43359, '\p{^ -IN_Rejang}', "");
Expect(0, 43359, '\P{ -IN_Rejang}', "");
Expect(1, 43359, '\P{^ -IN_Rejang}', "");
Expect(0, 43360, '\p{ -IN_Rejang}', "");
Expect(1, 43360, '\p{^ -IN_Rejang}', "");
Expect(1, 43360, '\P{ -IN_Rejang}', "");
Expect(0, 43360, '\P{^ -IN_Rejang}', "");
Error('\p{-/a/Combining_Diacritical_marks_SUPPLEMENT}');
Error('\P{-/a/Combining_Diacritical_marks_SUPPLEMENT}');
Expect(1, 7679, '\p{combiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\p{^combiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\P{combiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\P{^combiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\p{combiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\p{^combiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\P{combiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\P{^combiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\p{	Combining_Diacritical_Marks_supplement}', "");
Expect(0, 7679, '\p{^	Combining_Diacritical_Marks_supplement}', "");
Expect(0, 7679, '\P{	Combining_Diacritical_Marks_supplement}', "");
Expect(1, 7679, '\P{^	Combining_Diacritical_Marks_supplement}', "");
Expect(0, 7680, '\p{	Combining_Diacritical_Marks_supplement}', "");
Expect(1, 7680, '\p{^	Combining_Diacritical_Marks_supplement}', "");
Expect(1, 7680, '\P{	Combining_Diacritical_Marks_supplement}', "");
Expect(0, 7680, '\P{^	Combining_Diacritical_Marks_supplement}', "");
Error('\p{ /a/Is_combining_Diacritical_MARKS_supplement}');
Error('\P{ /a/Is_combining_Diacritical_MARKS_supplement}');
Expect(1, 7679, '\p{iscombiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\p{^iscombiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\P{iscombiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\P{^iscombiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\p{iscombiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\p{^iscombiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\P{iscombiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\P{^iscombiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\p{-Is_COMBINING_Diacritical_MARKS_Supplement}', "");
Expect(0, 7679, '\p{^-Is_COMBINING_Diacritical_MARKS_Supplement}', "");
Expect(0, 7679, '\P{-Is_COMBINING_Diacritical_MARKS_Supplement}', "");
Expect(1, 7679, '\P{^-Is_COMBINING_Diacritical_MARKS_Supplement}', "");
Expect(0, 7680, '\p{-Is_COMBINING_Diacritical_MARKS_Supplement}', "");
Expect(1, 7680, '\p{^-Is_COMBINING_Diacritical_MARKS_Supplement}', "");
Expect(1, 7680, '\P{-Is_COMBINING_Diacritical_MARKS_Supplement}', "");
Expect(0, 7680, '\P{^-Is_COMBINING_Diacritical_MARKS_Supplement}', "");
Error('\p{/a/In_Combining_Diacritical_MARKS_SUPPLEMENT}');
Error('\P{/a/In_Combining_Diacritical_MARKS_SUPPLEMENT}');
Expect(1, 7679, '\p{incombiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\p{^incombiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\P{incombiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\P{^incombiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\p{incombiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\p{^incombiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\P{incombiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\P{^incombiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\p{-IN_Combining_DIACRITICAL_marks_supplement}', "");
Expect(0, 7679, '\p{^-IN_Combining_DIACRITICAL_marks_supplement}', "");
Expect(0, 7679, '\P{-IN_Combining_DIACRITICAL_marks_supplement}', "");
Expect(1, 7679, '\P{^-IN_Combining_DIACRITICAL_marks_supplement}', "");
Expect(0, 7680, '\p{-IN_Combining_DIACRITICAL_marks_supplement}', "");
Expect(1, 7680, '\p{^-IN_Combining_DIACRITICAL_marks_supplement}', "");
Expect(1, 7680, '\P{-IN_Combining_DIACRITICAL_marks_supplement}', "");
Expect(0, 7680, '\P{^-IN_Combining_DIACRITICAL_marks_supplement}', "");
Error('\p{__diacriticals_sup:=}');
Error('\P{__diacriticals_sup:=}');
Expect(1, 7679, '\p{diacriticalssup}', "");
Expect(0, 7679, '\p{^diacriticalssup}', "");
Expect(0, 7679, '\P{diacriticalssup}', "");
Expect(1, 7679, '\P{^diacriticalssup}', "");
Expect(0, 7680, '\p{diacriticalssup}', "");
Expect(1, 7680, '\p{^diacriticalssup}', "");
Expect(1, 7680, '\P{diacriticalssup}', "");
Expect(0, 7680, '\P{^diacriticalssup}', "");
Expect(1, 7679, '\p{_ Diacriticals_sup}', "");
Expect(0, 7679, '\p{^_ Diacriticals_sup}', "");
Expect(0, 7679, '\P{_ Diacriticals_sup}', "");
Expect(1, 7679, '\P{^_ Diacriticals_sup}', "");
Expect(0, 7680, '\p{_ Diacriticals_sup}', "");
Expect(1, 7680, '\p{^_ Diacriticals_sup}', "");
Expect(1, 7680, '\P{_ Diacriticals_sup}', "");
Expect(0, 7680, '\P{^_ Diacriticals_sup}', "");
Error('\p{/a/_is_diacriticals_Sup}');
Error('\P{/a/_is_diacriticals_Sup}');
Expect(1, 7679, '\p{isdiacriticalssup}', "");
Expect(0, 7679, '\p{^isdiacriticalssup}', "");
Expect(0, 7679, '\P{isdiacriticalssup}', "");
Expect(1, 7679, '\P{^isdiacriticalssup}', "");
Expect(0, 7680, '\p{isdiacriticalssup}', "");
Expect(1, 7680, '\p{^isdiacriticalssup}', "");
Expect(1, 7680, '\P{isdiacriticalssup}', "");
Expect(0, 7680, '\P{^isdiacriticalssup}', "");
Expect(1, 7679, '\p{-Is_Diacriticals_SUP}', "");
Expect(0, 7679, '\p{^-Is_Diacriticals_SUP}', "");
Expect(0, 7679, '\P{-Is_Diacriticals_SUP}', "");
Expect(1, 7679, '\P{^-Is_Diacriticals_SUP}', "");
Expect(0, 7680, '\p{-Is_Diacriticals_SUP}', "");
Expect(1, 7680, '\p{^-Is_Diacriticals_SUP}', "");
Expect(1, 7680, '\P{-Is_Diacriticals_SUP}', "");
Expect(0, 7680, '\P{^-Is_Diacriticals_SUP}', "");
Error('\p{-:=in_DIACRITICALS_sup}');
Error('\P{-:=in_DIACRITICALS_sup}');
Expect(1, 7679, '\p{indiacriticalssup}', "");
Expect(0, 7679, '\p{^indiacriticalssup}', "");
Expect(0, 7679, '\P{indiacriticalssup}', "");
Expect(1, 7679, '\P{^indiacriticalssup}', "");
Expect(0, 7680, '\p{indiacriticalssup}', "");
Expect(1, 7680, '\p{^indiacriticalssup}', "");
Expect(1, 7680, '\P{indiacriticalssup}', "");
Expect(0, 7680, '\P{^indiacriticalssup}', "");
Expect(1, 7679, '\p{  In_DIACRITICALS_SUP}', "");
Expect(0, 7679, '\p{^  In_DIACRITICALS_SUP}', "");
Expect(0, 7679, '\P{  In_DIACRITICALS_SUP}', "");
Expect(1, 7679, '\P{^  In_DIACRITICALS_SUP}', "");
Expect(0, 7680, '\p{  In_DIACRITICALS_SUP}', "");
Expect(1, 7680, '\p{^  In_DIACRITICALS_SUP}', "");
Expect(1, 7680, '\P{  In_DIACRITICALS_SUP}', "");
Expect(0, 7680, '\P{^  In_DIACRITICALS_SUP}', "");
Error('\p{ /a/sterm}');
Error('\P{ /a/sterm}');
Expect(1, 70086, '\p{sterm}', "");
Expect(0, 70086, '\p{^sterm}', "");
Expect(0, 70086, '\P{sterm}', "");
Expect(1, 70086, '\P{^sterm}', "");
Expect(0, 70087, '\p{sterm}', "");
Expect(1, 70087, '\p{^sterm}', "");
Expect(1, 70087, '\P{sterm}', "");
Expect(0, 70087, '\P{^sterm}', "");
Expect(1, 70086, '\p{ 	STerm}', "");
Expect(0, 70086, '\p{^ 	STerm}', "");
Expect(0, 70086, '\P{ 	STerm}', "");
Expect(1, 70086, '\P{^ 	STerm}', "");
Expect(0, 70087, '\p{ 	STerm}', "");
Expect(1, 70087, '\p{^ 	STerm}', "");
Expect(1, 70087, '\P{ 	STerm}', "");
Expect(0, 70087, '\P{^ 	STerm}', "");
Error('\p{/a/is_sterm}');
Error('\P{/a/is_sterm}');
Expect(1, 70086, '\p{issterm}', "");
Expect(0, 70086, '\p{^issterm}', "");
Expect(0, 70086, '\P{issterm}', "");
Expect(1, 70086, '\P{^issterm}', "");
Expect(0, 70087, '\p{issterm}', "");
Expect(1, 70087, '\p{^issterm}', "");
Expect(1, 70087, '\P{issterm}', "");
Expect(0, 70087, '\P{^issterm}', "");
Expect(1, 70086, '\p{	 Is_STerm}', "");
Expect(0, 70086, '\p{^	 Is_STerm}', "");
Expect(0, 70086, '\P{	 Is_STerm}', "");
Expect(1, 70086, '\P{^	 Is_STerm}', "");
Expect(0, 70087, '\p{	 Is_STerm}', "");
Expect(1, 70087, '\p{^	 Is_STerm}', "");
Expect(1, 70087, '\P{	 Is_STerm}', "");
Expect(0, 70087, '\P{^	 Is_STerm}', "");
Error('\p{/a/--IN_Kannada}');
Error('\P{/a/--IN_Kannada}');
Expect(1, 3327, '\p{inkannada}', "");
Expect(0, 3327, '\p{^inkannada}', "");
Expect(0, 3327, '\P{inkannada}', "");
Expect(1, 3327, '\P{^inkannada}', "");
Expect(0, 3328, '\p{inkannada}', "");
Expect(1, 3328, '\p{^inkannada}', "");
Expect(1, 3328, '\P{inkannada}', "");
Expect(0, 3328, '\P{^inkannada}', "");
Expect(1, 3327, '\p{-IN_KANNADA}', "");
Expect(0, 3327, '\p{^-IN_KANNADA}', "");
Expect(0, 3327, '\P{-IN_KANNADA}', "");
Expect(1, 3327, '\P{^-IN_KANNADA}', "");
Expect(0, 3328, '\p{-IN_KANNADA}', "");
Expect(1, 3328, '\p{^-IN_KANNADA}', "");
Expect(1, 3328, '\P{-IN_KANNADA}', "");
Expect(0, 3328, '\P{^-IN_KANNADA}', "");
Error('\p{ /a/In_KANNADA}');
Error('\P{ /a/In_KANNADA}');
Expect(1, 3327, '\p{_	In_kannada}', "");
Expect(0, 3327, '\p{^_	In_kannada}', "");
Expect(0, 3327, '\P{_	In_kannada}', "");
Expect(1, 3327, '\P{^_	In_kannada}', "");
Expect(0, 3328, '\p{_	In_kannada}', "");
Expect(1, 3328, '\p{^_	In_kannada}', "");
Expect(1, 3328, '\P{_	In_kannada}', "");
Expect(0, 3328, '\P{^_	In_kannada}', "");
Error('\p{_:=soft_dotted}');
Error('\P{_:=soft_dotted}');
Expect(1, 120467, '\p{softdotted}', "");
Expect(0, 120467, '\p{^softdotted}', "");
Expect(0, 120467, '\P{softdotted}', "");
Expect(1, 120467, '\P{^softdotted}', "");
Expect(0, 120468, '\p{softdotted}', "");
Expect(1, 120468, '\p{^softdotted}', "");
Expect(1, 120468, '\P{softdotted}', "");
Expect(0, 120468, '\P{^softdotted}', "");
Expect(1, 120467, '\p{	-Soft_Dotted}', "");
Expect(0, 120467, '\p{^	-Soft_Dotted}', "");
Expect(0, 120467, '\P{	-Soft_Dotted}', "");
Expect(1, 120467, '\P{^	-Soft_Dotted}', "");
Expect(0, 120468, '\p{	-Soft_Dotted}', "");
Expect(1, 120468, '\p{^	-Soft_Dotted}', "");
Expect(1, 120468, '\P{	-Soft_Dotted}', "");
Expect(0, 120468, '\P{^	-Soft_Dotted}', "");
Error('\p{:=	-is_SOFT_DOTTED}');
Error('\P{:=	-is_SOFT_DOTTED}');
Expect(1, 120467, '\p{issoftdotted}', "");
Expect(0, 120467, '\p{^issoftdotted}', "");
Expect(0, 120467, '\P{issoftdotted}', "");
Expect(1, 120467, '\P{^issoftdotted}', "");
Expect(0, 120468, '\p{issoftdotted}', "");
Expect(1, 120468, '\p{^issoftdotted}', "");
Expect(1, 120468, '\P{issoftdotted}', "");
Expect(0, 120468, '\P{^issoftdotted}', "");
Expect(1, 120467, '\p{ IS_SOFT_DOTTED}', "");
Expect(0, 120467, '\p{^ IS_SOFT_DOTTED}', "");
Expect(0, 120467, '\P{ IS_SOFT_DOTTED}', "");
Expect(1, 120467, '\P{^ IS_SOFT_DOTTED}', "");
Expect(0, 120468, '\p{ IS_SOFT_DOTTED}', "");
Expect(1, 120468, '\p{^ IS_SOFT_DOTTED}', "");
Expect(1, 120468, '\P{ IS_SOFT_DOTTED}', "");
Expect(0, 120468, '\P{^ IS_SOFT_DOTTED}', "");
Error('\p{/a/	 sd}');
Error('\P{/a/	 sd}');
Expect(1, 120467, '\p{sd}', "");
Expect(0, 120467, '\p{^sd}', "");
Expect(0, 120467, '\P{sd}', "");
Expect(1, 120467, '\P{^sd}', "");
Expect(0, 120468, '\p{sd}', "");
Expect(1, 120468, '\p{^sd}', "");
Expect(1, 120468, '\P{sd}', "");
Expect(0, 120468, '\P{^sd}', "");
Expect(1, 120467, '\p{	SD}', "");
Expect(0, 120467, '\p{^	SD}', "");
Expect(0, 120467, '\P{	SD}', "");
Expect(1, 120467, '\P{^	SD}', "");
Expect(0, 120468, '\p{	SD}', "");
Expect(1, 120468, '\p{^	SD}', "");
Expect(1, 120468, '\P{	SD}', "");
Expect(0, 120468, '\P{^	SD}', "");
Error('\p{__Is_SD:=}');
Error('\P{__Is_SD:=}');
Expect(1, 120467, '\p{issd}', "");
Expect(0, 120467, '\p{^issd}', "");
Expect(0, 120467, '\P{issd}', "");
Expect(1, 120467, '\P{^issd}', "");
Expect(0, 120468, '\p{issd}', "");
Expect(1, 120468, '\p{^issd}', "");
Expect(1, 120468, '\P{issd}', "");
Expect(0, 120468, '\P{^issd}', "");
Expect(1, 120467, '\p{ -IS_SD}', "");
Expect(0, 120467, '\p{^ -IS_SD}', "");
Expect(0, 120467, '\P{ -IS_SD}', "");
Expect(1, 120467, '\P{^ -IS_SD}', "");
Expect(0, 120468, '\p{ -IS_SD}', "");
Expect(1, 120468, '\p{^ -IS_SD}', "");
Expect(1, 120468, '\P{ -IS_SD}', "");
Expect(0, 120468, '\P{^ -IS_SD}', "");
Error('\p{	:=_CanonDCIJ}');
Error('\P{	:=_CanonDCIJ}');
Expect(1, 120467, '\p{_CANONDCIJ}', "");
Expect(0, 120467, '\p{^_CANONDCIJ}', "");
Expect(0, 120467, '\P{_CANONDCIJ}', "");
Expect(1, 120467, '\P{^_CANONDCIJ}', "");
Expect(0, 120468, '\p{_CANONDCIJ}', "");
Expect(1, 120468, '\p{^_CANONDCIJ}', "");
Expect(1, 120468, '\P{_CANONDCIJ}', "");
Expect(0, 120468, '\P{^_CANONDCIJ}', "");
Error('\p{-AVESTAN:=}');
Error('\P{-AVESTAN:=}');
Expect(1, 68415, '\p{avestan}', "");
Expect(0, 68415, '\p{^avestan}', "");
Expect(0, 68415, '\P{avestan}', "");
Expect(1, 68415, '\P{^avestan}', "");
Expect(0, 68416, '\p{avestan}', "");
Expect(1, 68416, '\p{^avestan}', "");
Expect(1, 68416, '\P{avestan}', "");
Expect(0, 68416, '\P{^avestan}', "");
Expect(1, 68415, '\p{  Avestan}', "");
Expect(0, 68415, '\p{^  Avestan}', "");
Expect(0, 68415, '\P{  Avestan}', "");
Expect(1, 68415, '\P{^  Avestan}', "");
Expect(0, 68416, '\p{  Avestan}', "");
Expect(1, 68416, '\p{^  Avestan}', "");
Expect(1, 68416, '\P{  Avestan}', "");
Expect(0, 68416, '\P{^  Avestan}', "");
Error('\p{/a/_-Is_Avestan}');
Error('\P{/a/_-Is_Avestan}');
Expect(1, 68415, '\p{isavestan}', "");
Expect(0, 68415, '\p{^isavestan}', "");
Expect(0, 68415, '\P{isavestan}', "");
Expect(1, 68415, '\P{^isavestan}', "");
Expect(0, 68416, '\p{isavestan}', "");
Expect(1, 68416, '\p{^isavestan}', "");
Expect(1, 68416, '\P{isavestan}', "");
Expect(0, 68416, '\P{^isavestan}', "");
Expect(1, 68415, '\p{-is_avestan}', "");
Expect(0, 68415, '\p{^-is_avestan}', "");
Expect(0, 68415, '\P{-is_avestan}', "");
Expect(1, 68415, '\P{^-is_avestan}', "");
Expect(0, 68416, '\p{-is_avestan}', "");
Expect(1, 68416, '\p{^-is_avestan}', "");
Expect(1, 68416, '\P{-is_avestan}', "");
Expect(0, 68416, '\P{^-is_avestan}', "");
Error('\p{/a/  Avst}');
Error('\P{/a/  Avst}');
Expect(1, 68415, '\p{avst}', "");
Expect(0, 68415, '\p{^avst}', "");
Expect(0, 68415, '\P{avst}', "");
Expect(1, 68415, '\P{^avst}', "");
Expect(0, 68416, '\p{avst}', "");
Expect(1, 68416, '\p{^avst}', "");
Expect(1, 68416, '\P{avst}', "");
Expect(0, 68416, '\P{^avst}', "");
Expect(1, 68415, '\p{ avst}', "");
Expect(0, 68415, '\p{^ avst}', "");
Expect(0, 68415, '\P{ avst}', "");
Expect(1, 68415, '\P{^ avst}', "");
Expect(0, 68416, '\p{ avst}', "");
Expect(1, 68416, '\p{^ avst}', "");
Expect(1, 68416, '\P{ avst}', "");
Expect(0, 68416, '\P{^ avst}', "");
Error('\p{/a/	Is_Avst}');
Error('\P{/a/	Is_Avst}');
Expect(1, 68415, '\p{isavst}', "");
Expect(0, 68415, '\p{^isavst}', "");
Expect(0, 68415, '\P{isavst}', "");
Expect(1, 68415, '\P{^isavst}', "");
Expect(0, 68416, '\p{isavst}', "");
Expect(1, 68416, '\p{^isavst}', "");
Expect(1, 68416, '\P{isavst}', "");
Expect(0, 68416, '\P{^isavst}', "");
Expect(1, 68415, '\p{_ IS_Avst}', "");
Expect(0, 68415, '\p{^_ IS_Avst}', "");
Expect(0, 68415, '\P{_ IS_Avst}', "");
Expect(1, 68415, '\P{^_ IS_Avst}', "");
Expect(0, 68416, '\p{_ IS_Avst}', "");
Expect(1, 68416, '\p{^_ IS_Avst}', "");
Expect(1, 68416, '\P{_ IS_Avst}', "");
Expect(0, 68416, '\P{^_ IS_Avst}', "");
Error('\p{__CONTROL_pictures/a/}');
Error('\P{__CONTROL_pictures/a/}');
Expect(1, 9279, '\p{controlpictures}', "");
Expect(0, 9279, '\p{^controlpictures}', "");
Expect(0, 9279, '\P{controlpictures}', "");
Expect(1, 9279, '\P{^controlpictures}', "");
Expect(0, 9280, '\p{controlpictures}', "");
Expect(1, 9280, '\p{^controlpictures}', "");
Expect(1, 9280, '\P{controlpictures}', "");
Expect(0, 9280, '\P{^controlpictures}', "");
Expect(1, 9279, '\p{_	CONTROL_Pictures}', "");
Expect(0, 9279, '\p{^_	CONTROL_Pictures}', "");
Expect(0, 9279, '\P{_	CONTROL_Pictures}', "");
Expect(1, 9279, '\P{^_	CONTROL_Pictures}', "");
Expect(0, 9280, '\p{_	CONTROL_Pictures}', "");
Expect(1, 9280, '\p{^_	CONTROL_Pictures}', "");
Expect(1, 9280, '\P{_	CONTROL_Pictures}', "");
Expect(0, 9280, '\P{^_	CONTROL_Pictures}', "");
Error('\p{is_Control_PICTURES/a/}');
Error('\P{is_Control_PICTURES/a/}');
Expect(1, 9279, '\p{iscontrolpictures}', "");
Expect(0, 9279, '\p{^iscontrolpictures}', "");
Expect(0, 9279, '\P{iscontrolpictures}', "");
Expect(1, 9279, '\P{^iscontrolpictures}', "");
Expect(0, 9280, '\p{iscontrolpictures}', "");
Expect(1, 9280, '\p{^iscontrolpictures}', "");
Expect(1, 9280, '\P{iscontrolpictures}', "");
Expect(0, 9280, '\P{^iscontrolpictures}', "");
Expect(1, 9279, '\p{		Is_Control_PICTURES}', "");
Expect(0, 9279, '\p{^		Is_Control_PICTURES}', "");
Expect(0, 9279, '\P{		Is_Control_PICTURES}', "");
Expect(1, 9279, '\P{^		Is_Control_PICTURES}', "");
Expect(0, 9280, '\p{		Is_Control_PICTURES}', "");
Expect(1, 9280, '\p{^		Is_Control_PICTURES}', "");
Expect(1, 9280, '\P{		Is_Control_PICTURES}', "");
Expect(0, 9280, '\P{^		Is_Control_PICTURES}', "");
Error('\p{_-In_Control_Pictures/a/}');
Error('\P{_-In_Control_Pictures/a/}');
Expect(1, 9279, '\p{incontrolpictures}', "");
Expect(0, 9279, '\p{^incontrolpictures}', "");
Expect(0, 9279, '\P{incontrolpictures}', "");
Expect(1, 9279, '\P{^incontrolpictures}', "");
Expect(0, 9280, '\p{incontrolpictures}', "");
Expect(1, 9280, '\p{^incontrolpictures}', "");
Expect(1, 9280, '\P{incontrolpictures}', "");
Expect(0, 9280, '\P{^incontrolpictures}', "");
Expect(1, 9279, '\p{_	In_Control_PICTURES}', "");
Expect(0, 9279, '\p{^_	In_Control_PICTURES}', "");
Expect(0, 9279, '\P{_	In_Control_PICTURES}', "");
Expect(1, 9279, '\P{^_	In_Control_PICTURES}', "");
Expect(0, 9280, '\p{_	In_Control_PICTURES}', "");
Expect(1, 9280, '\p{^_	In_Control_PICTURES}', "");
Expect(1, 9280, '\P{_	In_Control_PICTURES}', "");
Expect(0, 9280, '\P{^_	In_Control_PICTURES}', "");
Error('\p{_/a/LATIN_Extended_A}');
Error('\P{_/a/LATIN_Extended_A}');
Expect(1, 383, '\p{latinextendeda}', "");
Expect(0, 383, '\p{^latinextendeda}', "");
Expect(0, 383, '\P{latinextendeda}', "");
Expect(1, 383, '\P{^latinextendeda}', "");
Expect(0, 384, '\p{latinextendeda}', "");
Expect(1, 384, '\p{^latinextendeda}', "");
Expect(1, 384, '\P{latinextendeda}', "");
Expect(0, 384, '\P{^latinextendeda}', "");
Expect(1, 383, '\p{ latin_Extended_a}', "");
Expect(0, 383, '\p{^ latin_Extended_a}', "");
Expect(0, 383, '\P{ latin_Extended_a}', "");
Expect(1, 383, '\P{^ latin_Extended_a}', "");
Expect(0, 384, '\p{ latin_Extended_a}', "");
Expect(1, 384, '\p{^ latin_Extended_a}', "");
Expect(1, 384, '\P{ latin_Extended_a}', "");
Expect(0, 384, '\P{^ latin_Extended_a}', "");
Error('\p{/a/_ Is_Latin_Extended_A}');
Error('\P{/a/_ Is_Latin_Extended_A}');
Expect(1, 383, '\p{islatinextendeda}', "");
Expect(0, 383, '\p{^islatinextendeda}', "");
Expect(0, 383, '\P{islatinextendeda}', "");
Expect(1, 383, '\P{^islatinextendeda}', "");
Expect(0, 384, '\p{islatinextendeda}', "");
Expect(1, 384, '\p{^islatinextendeda}', "");
Expect(1, 384, '\P{islatinextendeda}', "");
Expect(0, 384, '\P{^islatinextendeda}', "");
Expect(1, 383, '\p{is_latin_Extended_A}', "");
Expect(0, 383, '\p{^is_latin_Extended_A}', "");
Expect(0, 383, '\P{is_latin_Extended_A}', "");
Expect(1, 383, '\P{^is_latin_Extended_A}', "");
Expect(0, 384, '\p{is_latin_Extended_A}', "");
Expect(1, 384, '\p{^is_latin_Extended_A}', "");
Expect(1, 384, '\P{is_latin_Extended_A}', "");
Expect(0, 384, '\P{^is_latin_Extended_A}', "");
Error('\p{_	in_Latin_EXTENDED_A/a/}');
Error('\P{_	in_Latin_EXTENDED_A/a/}');
Expect(1, 383, '\p{inlatinextendeda}', "");
Expect(0, 383, '\p{^inlatinextendeda}', "");
Expect(0, 383, '\P{inlatinextendeda}', "");
Expect(1, 383, '\P{^inlatinextendeda}', "");
Expect(0, 384, '\p{inlatinextendeda}', "");
Expect(1, 384, '\p{^inlatinextendeda}', "");
Expect(1, 384, '\P{inlatinextendeda}', "");
Expect(0, 384, '\P{^inlatinextendeda}', "");
Expect(1, 383, '\p{_IN_Latin_Extended_A}', "");
Expect(0, 383, '\p{^_IN_Latin_Extended_A}', "");
Expect(0, 383, '\P{_IN_Latin_Extended_A}', "");
Expect(1, 383, '\P{^_IN_Latin_Extended_A}', "");
Expect(0, 384, '\p{_IN_Latin_Extended_A}', "");
Expect(1, 384, '\p{^_IN_Latin_Extended_A}', "");
Expect(1, 384, '\P{_IN_Latin_Extended_A}', "");
Expect(0, 384, '\P{^_IN_Latin_Extended_A}', "");
Error('\p{	-Latin_Ext_A/a/}');
Error('\P{	-Latin_Ext_A/a/}');
Expect(1, 383, '\p{latinexta}', "");
Expect(0, 383, '\p{^latinexta}', "");
Expect(0, 383, '\P{latinexta}', "");
Expect(1, 383, '\P{^latinexta}', "");
Expect(0, 384, '\p{latinexta}', "");
Expect(1, 384, '\p{^latinexta}', "");
Expect(1, 384, '\P{latinexta}', "");
Expect(0, 384, '\P{^latinexta}', "");
Expect(1, 383, '\p{ latin_Ext_A}', "");
Expect(0, 383, '\p{^ latin_Ext_A}', "");
Expect(0, 383, '\P{ latin_Ext_A}', "");
Expect(1, 383, '\P{^ latin_Ext_A}', "");
Expect(0, 384, '\p{ latin_Ext_A}', "");
Expect(1, 384, '\p{^ latin_Ext_A}', "");
Expect(1, 384, '\P{ latin_Ext_A}', "");
Expect(0, 384, '\P{^ latin_Ext_A}', "");
Error('\p{:=Is_latin_ext_a}');
Error('\P{:=Is_latin_ext_a}');
Expect(1, 383, '\p{islatinexta}', "");
Expect(0, 383, '\p{^islatinexta}', "");
Expect(0, 383, '\P{islatinexta}', "");
Expect(1, 383, '\P{^islatinexta}', "");
Expect(0, 384, '\p{islatinexta}', "");
Expect(1, 384, '\p{^islatinexta}', "");
Expect(1, 384, '\P{islatinexta}', "");
Expect(0, 384, '\P{^islatinexta}', "");
Expect(1, 383, '\p{--Is_Latin_Ext_A}', "");
Expect(0, 383, '\p{^--Is_Latin_Ext_A}', "");
Expect(0, 383, '\P{--Is_Latin_Ext_A}', "");
Expect(1, 383, '\P{^--Is_Latin_Ext_A}', "");
Expect(0, 384, '\p{--Is_Latin_Ext_A}', "");
Expect(1, 384, '\p{^--Is_Latin_Ext_A}', "");
Expect(1, 384, '\P{--Is_Latin_Ext_A}', "");
Expect(0, 384, '\P{^--Is_Latin_Ext_A}', "");
Error('\p{/a/IN_Latin_Ext_A}');
Error('\P{/a/IN_Latin_Ext_A}');
Expect(1, 383, '\p{inlatinexta}', "");
Expect(0, 383, '\p{^inlatinexta}', "");
Expect(0, 383, '\P{inlatinexta}', "");
Expect(1, 383, '\P{^inlatinexta}', "");
Expect(0, 384, '\p{inlatinexta}', "");
Expect(1, 384, '\p{^inlatinexta}', "");
Expect(1, 384, '\P{inlatinexta}', "");
Expect(0, 384, '\P{^inlatinexta}', "");
Expect(1, 383, '\p{in_latin_Ext_a}', "");
Expect(0, 383, '\p{^in_latin_Ext_a}', "");
Expect(0, 383, '\P{in_latin_Ext_a}', "");
Expect(1, 383, '\P{^in_latin_Ext_a}', "");
Expect(0, 384, '\p{in_latin_Ext_a}', "");
Expect(1, 384, '\p{^in_latin_Ext_a}', "");
Expect(1, 384, '\P{in_latin_Ext_a}', "");
Expect(0, 384, '\P{^in_latin_Ext_a}', "");
Error('\p{-_gujarati/a/}');
Error('\P{-_gujarati/a/}');
Expect(1, 2801, '\p{gujarati}', "");
Expect(0, 2801, '\p{^gujarati}', "");
Expect(0, 2801, '\P{gujarati}', "");
Expect(1, 2801, '\P{^gujarati}', "");
Expect(0, 2802, '\p{gujarati}', "");
Expect(1, 2802, '\p{^gujarati}', "");
Expect(1, 2802, '\P{gujarati}', "");
Expect(0, 2802, '\P{^gujarati}', "");
Expect(1, 2801, '\p{	Gujarati}', "");
Expect(0, 2801, '\p{^	Gujarati}', "");
Expect(0, 2801, '\P{	Gujarati}', "");
Expect(1, 2801, '\P{^	Gujarati}', "");
Expect(0, 2802, '\p{	Gujarati}', "");
Expect(1, 2802, '\p{^	Gujarati}', "");
Expect(1, 2802, '\P{	Gujarati}', "");
Expect(0, 2802, '\P{^	Gujarati}', "");
Error('\p{	Is_Gujarati/a/}');
Error('\P{	Is_Gujarati/a/}');
Expect(1, 2801, '\p{isgujarati}', "");
Expect(0, 2801, '\p{^isgujarati}', "");
Expect(0, 2801, '\P{isgujarati}', "");
Expect(1, 2801, '\P{^isgujarati}', "");
Expect(0, 2802, '\p{isgujarati}', "");
Expect(1, 2802, '\p{^isgujarati}', "");
Expect(1, 2802, '\P{isgujarati}', "");
Expect(0, 2802, '\P{^isgujarati}', "");
Expect(1, 2801, '\p{	 Is_GUJARATI}', "");
Expect(0, 2801, '\p{^	 Is_GUJARATI}', "");
Expect(0, 2801, '\P{	 Is_GUJARATI}', "");
Expect(1, 2801, '\P{^	 Is_GUJARATI}', "");
Expect(0, 2802, '\p{	 Is_GUJARATI}', "");
Expect(1, 2802, '\p{^	 Is_GUJARATI}', "");
Expect(1, 2802, '\P{	 Is_GUJARATI}', "");
Expect(0, 2802, '\P{^	 Is_GUJARATI}', "");
Error('\p{/a/Gujr}');
Error('\P{/a/Gujr}');
Expect(1, 2801, '\p{gujr}', "");
Expect(0, 2801, '\p{^gujr}', "");
Expect(0, 2801, '\P{gujr}', "");
Expect(1, 2801, '\P{^gujr}', "");
Expect(0, 2802, '\p{gujr}', "");
Expect(1, 2802, '\p{^gujr}', "");
Expect(1, 2802, '\P{gujr}', "");
Expect(0, 2802, '\P{^gujr}', "");
Expect(1, 2801, '\p{_gujr}', "");
Expect(0, 2801, '\p{^_gujr}', "");
Expect(0, 2801, '\P{_gujr}', "");
Expect(1, 2801, '\P{^_gujr}', "");
Expect(0, 2802, '\p{_gujr}', "");
Expect(1, 2802, '\p{^_gujr}', "");
Expect(1, 2802, '\P{_gujr}', "");
Expect(0, 2802, '\P{^_gujr}', "");
Error('\p{:= Is_GUJR}');
Error('\P{:= Is_GUJR}');
Expect(1, 2801, '\p{isgujr}', "");
Expect(0, 2801, '\p{^isgujr}', "");
Expect(0, 2801, '\P{isgujr}', "");
Expect(1, 2801, '\P{^isgujr}', "");
Expect(0, 2802, '\p{isgujr}', "");
Expect(1, 2802, '\p{^isgujr}', "");
Expect(1, 2802, '\P{isgujr}', "");
Expect(0, 2802, '\P{^isgujr}', "");
Expect(1, 2801, '\p{-	Is_gujr}', "");
Expect(0, 2801, '\p{^-	Is_gujr}', "");
Expect(0, 2801, '\P{-	Is_gujr}', "");
Expect(1, 2801, '\P{^-	Is_gujr}', "");
Expect(0, 2802, '\p{-	Is_gujr}', "");
Expect(1, 2802, '\p{^-	Is_gujr}', "");
Expect(1, 2802, '\P{-	Is_gujr}', "");
Expect(0, 2802, '\P{^-	Is_gujr}', "");
Error('\p{:= _In_oriya}');
Error('\P{:= _In_oriya}');
Expect(1, 2943, '\p{inoriya}', "");
Expect(0, 2943, '\p{^inoriya}', "");
Expect(0, 2943, '\P{inoriya}', "");
Expect(1, 2943, '\P{^inoriya}', "");
Expect(0, 2944, '\p{inoriya}', "");
Expect(1, 2944, '\p{^inoriya}', "");
Expect(1, 2944, '\P{inoriya}', "");
Expect(0, 2944, '\P{^inoriya}', "");
Expect(1, 2943, '\p{	-in_oriya}', "");
Expect(0, 2943, '\p{^	-in_oriya}', "");
Expect(0, 2943, '\P{	-in_oriya}', "");
Expect(1, 2943, '\P{^	-in_oriya}', "");
Expect(0, 2944, '\p{	-in_oriya}', "");
Expect(1, 2944, '\p{^	-in_oriya}', "");
Expect(1, 2944, '\P{	-in_oriya}', "");
Expect(0, 2944, '\P{^	-in_oriya}', "");
Error('\p{	/a/In_ORIYA}');
Error('\P{	/a/In_ORIYA}');
Expect(1, 2943, '\p{in_Oriya}', "");
Expect(0, 2943, '\p{^in_Oriya}', "");
Expect(0, 2943, '\P{in_Oriya}', "");
Expect(1, 2943, '\P{^in_Oriya}', "");
Expect(0, 2944, '\p{in_Oriya}', "");
Expect(1, 2944, '\p{^in_Oriya}', "");
Expect(1, 2944, '\P{in_Oriya}', "");
Expect(0, 2944, '\P{^in_Oriya}', "");
Error('\p{:=	_CJK_Radicals_Supplement}');
Error('\P{:=	_CJK_Radicals_Supplement}');
Expect(1, 12031, '\p{cjkradicalssupplement}', "");
Expect(0, 12031, '\p{^cjkradicalssupplement}', "");
Expect(0, 12031, '\P{cjkradicalssupplement}', "");
Expect(1, 12031, '\P{^cjkradicalssupplement}', "");
Expect(0, 12032, '\p{cjkradicalssupplement}', "");
Expect(1, 12032, '\p{^cjkradicalssupplement}', "");
Expect(1, 12032, '\P{cjkradicalssupplement}', "");
Expect(0, 12032, '\P{^cjkradicalssupplement}', "");
Expect(1, 12031, '\p{--cjk_radicals_Supplement}', "");
Expect(0, 12031, '\p{^--cjk_radicals_Supplement}', "");
Expect(0, 12031, '\P{--cjk_radicals_Supplement}', "");
Expect(1, 12031, '\P{^--cjk_radicals_Supplement}', "");
Expect(0, 12032, '\p{--cjk_radicals_Supplement}', "");
Expect(1, 12032, '\p{^--cjk_radicals_Supplement}', "");
Expect(1, 12032, '\P{--cjk_radicals_Supplement}', "");
Expect(0, 12032, '\P{^--cjk_radicals_Supplement}', "");
Error('\p{__Is_cjk_Radicals_SUPPLEMENT:=}');
Error('\P{__Is_cjk_Radicals_SUPPLEMENT:=}');
Expect(1, 12031, '\p{iscjkradicalssupplement}', "");
Expect(0, 12031, '\p{^iscjkradicalssupplement}', "");
Expect(0, 12031, '\P{iscjkradicalssupplement}', "");
Expect(1, 12031, '\P{^iscjkradicalssupplement}', "");
Expect(0, 12032, '\p{iscjkradicalssupplement}', "");
Expect(1, 12032, '\p{^iscjkradicalssupplement}', "");
Expect(1, 12032, '\P{iscjkradicalssupplement}', "");
Expect(0, 12032, '\P{^iscjkradicalssupplement}', "");
Expect(1, 12031, '\p{--IS_CJK_Radicals_Supplement}', "");
Expect(0, 12031, '\p{^--IS_CJK_Radicals_Supplement}', "");
Expect(0, 12031, '\P{--IS_CJK_Radicals_Supplement}', "");
Expect(1, 12031, '\P{^--IS_CJK_Radicals_Supplement}', "");
Expect(0, 12032, '\p{--IS_CJK_Radicals_Supplement}', "");
Expect(1, 12032, '\p{^--IS_CJK_Radicals_Supplement}', "");
Expect(1, 12032, '\P{--IS_CJK_Radicals_Supplement}', "");
Expect(0, 12032, '\P{^--IS_CJK_Radicals_Supplement}', "");
Error('\p{	 IN_CJK_Radicals_supplement/a/}');
Error('\P{	 IN_CJK_Radicals_supplement/a/}');
Expect(1, 12031, '\p{incjkradicalssupplement}', "");
Expect(0, 12031, '\p{^incjkradicalssupplement}', "");
Expect(0, 12031, '\P{incjkradicalssupplement}', "");
Expect(1, 12031, '\P{^incjkradicalssupplement}', "");
Expect(0, 12032, '\p{incjkradicalssupplement}', "");
Expect(1, 12032, '\p{^incjkradicalssupplement}', "");
Expect(1, 12032, '\P{incjkradicalssupplement}', "");
Expect(0, 12032, '\P{^incjkradicalssupplement}', "");
Expect(1, 12031, '\p{_	In_cjk_Radicals_Supplement}', "");
Expect(0, 12031, '\p{^_	In_cjk_Radicals_Supplement}', "");
Expect(0, 12031, '\P{_	In_cjk_Radicals_Supplement}', "");
Expect(1, 12031, '\P{^_	In_cjk_Radicals_Supplement}', "");
Expect(0, 12032, '\p{_	In_cjk_Radicals_Supplement}', "");
Expect(1, 12032, '\p{^_	In_cjk_Radicals_Supplement}', "");
Expect(1, 12032, '\P{_	In_cjk_Radicals_Supplement}', "");
Expect(0, 12032, '\P{^_	In_cjk_Radicals_Supplement}', "");
Error('\p{/a/		CJK_RADICALS_sup}');
Error('\P{/a/		CJK_RADICALS_sup}');
Expect(1, 12031, '\p{cjkradicalssup}', "");
Expect(0, 12031, '\p{^cjkradicalssup}', "");
Expect(0, 12031, '\P{cjkradicalssup}', "");
Expect(1, 12031, '\P{^cjkradicalssup}', "");
Expect(0, 12032, '\p{cjkradicalssup}', "");
Expect(1, 12032, '\p{^cjkradicalssup}', "");
Expect(1, 12032, '\P{cjkradicalssup}', "");
Expect(0, 12032, '\P{^cjkradicalssup}', "");
Expect(1, 12031, '\p{		CJK_Radicals_sup}', "");
Expect(0, 12031, '\p{^		CJK_Radicals_sup}', "");
Expect(0, 12031, '\P{		CJK_Radicals_sup}', "");
Expect(1, 12031, '\P{^		CJK_Radicals_sup}', "");
Expect(0, 12032, '\p{		CJK_Radicals_sup}', "");
Expect(1, 12032, '\p{^		CJK_Radicals_sup}', "");
Expect(1, 12032, '\P{		CJK_Radicals_sup}', "");
Expect(0, 12032, '\P{^		CJK_Radicals_sup}', "");
Error('\p{-_is_CJK_Radicals_SUP:=}');
Error('\P{-_is_CJK_Radicals_SUP:=}');
Expect(1, 12031, '\p{iscjkradicalssup}', "");
Expect(0, 12031, '\p{^iscjkradicalssup}', "");
Expect(0, 12031, '\P{iscjkradicalssup}', "");
Expect(1, 12031, '\P{^iscjkradicalssup}', "");
Expect(0, 12032, '\p{iscjkradicalssup}', "");
Expect(1, 12032, '\p{^iscjkradicalssup}', "");
Expect(1, 12032, '\P{iscjkradicalssup}', "");
Expect(0, 12032, '\P{^iscjkradicalssup}', "");
Expect(1, 12031, '\p{		Is_CJK_RADICALS_SUP}', "");
Expect(0, 12031, '\p{^		Is_CJK_RADICALS_SUP}', "");
Expect(0, 12031, '\P{		Is_CJK_RADICALS_SUP}', "");
Expect(1, 12031, '\P{^		Is_CJK_RADICALS_SUP}', "");
Expect(0, 12032, '\p{		Is_CJK_RADICALS_SUP}', "");
Expect(1, 12032, '\p{^		Is_CJK_RADICALS_SUP}', "");
Expect(1, 12032, '\P{		Is_CJK_RADICALS_SUP}', "");
Expect(0, 12032, '\P{^		Is_CJK_RADICALS_SUP}', "");
Error('\p{	_in_CJK_Radicals_SUP:=}');
Error('\P{	_in_CJK_Radicals_SUP:=}');
Expect(1, 12031, '\p{incjkradicalssup}', "");
Expect(0, 12031, '\p{^incjkradicalssup}', "");
Expect(0, 12031, '\P{incjkradicalssup}', "");
Expect(1, 12031, '\P{^incjkradicalssup}', "");
Expect(0, 12032, '\p{incjkradicalssup}', "");
Expect(1, 12032, '\p{^incjkradicalssup}', "");
Expect(1, 12032, '\P{incjkradicalssup}', "");
Expect(0, 12032, '\P{^incjkradicalssup}', "");
Expect(1, 12031, '\p{_ In_CJK_RADICALS_Sup}', "");
Expect(0, 12031, '\p{^_ In_CJK_RADICALS_Sup}', "");
Expect(0, 12031, '\P{_ In_CJK_RADICALS_Sup}', "");
Expect(1, 12031, '\P{^_ In_CJK_RADICALS_Sup}', "");
Expect(0, 12032, '\p{_ In_CJK_RADICALS_Sup}', "");
Expect(1, 12032, '\p{^_ In_CJK_RADICALS_Sup}', "");
Expect(1, 12032, '\P{_ In_CJK_RADICALS_Sup}', "");
Expect(0, 12032, '\P{^_ In_CJK_RADICALS_Sup}', "");
Error('\p{ CJK_Unified_Ideographs/a/}');
Error('\P{ CJK_Unified_Ideographs/a/}');
Expect(1, 40959, '\p{cjkunifiedideographs}', "");
Expect(0, 40959, '\p{^cjkunifiedideographs}', "");
Expect(0, 40959, '\P{cjkunifiedideographs}', "");
Expect(1, 40959, '\P{^cjkunifiedideographs}', "");
Expect(0, 40960, '\p{cjkunifiedideographs}', "");
Expect(1, 40960, '\p{^cjkunifiedideographs}', "");
Expect(1, 40960, '\P{cjkunifiedideographs}', "");
Expect(0, 40960, '\P{^cjkunifiedideographs}', "");
Expect(1, 40959, '\p{--CJK_unified_Ideographs}', "");
Expect(0, 40959, '\p{^--CJK_unified_Ideographs}', "");
Expect(0, 40959, '\P{--CJK_unified_Ideographs}', "");
Expect(1, 40959, '\P{^--CJK_unified_Ideographs}', "");
Expect(0, 40960, '\p{--CJK_unified_Ideographs}', "");
Expect(1, 40960, '\p{^--CJK_unified_Ideographs}', "");
Expect(1, 40960, '\P{--CJK_unified_Ideographs}', "");
Expect(0, 40960, '\P{^--CJK_unified_Ideographs}', "");
Error('\p{/a/		is_CJK_Unified_Ideographs}');
Error('\P{/a/		is_CJK_Unified_Ideographs}');
Expect(1, 40959, '\p{iscjkunifiedideographs}', "");
Expect(0, 40959, '\p{^iscjkunifiedideographs}', "");
Expect(0, 40959, '\P{iscjkunifiedideographs}', "");
Expect(1, 40959, '\P{^iscjkunifiedideographs}', "");
Expect(0, 40960, '\p{iscjkunifiedideographs}', "");
Expect(1, 40960, '\p{^iscjkunifiedideographs}', "");
Expect(1, 40960, '\P{iscjkunifiedideographs}', "");
Expect(0, 40960, '\P{^iscjkunifiedideographs}', "");
Expect(1, 40959, '\p{-	is_CJK_Unified_IDEOGRAPHS}', "");
Expect(0, 40959, '\p{^-	is_CJK_Unified_IDEOGRAPHS}', "");
Expect(0, 40959, '\P{-	is_CJK_Unified_IDEOGRAPHS}', "");
Expect(1, 40959, '\P{^-	is_CJK_Unified_IDEOGRAPHS}', "");
Expect(0, 40960, '\p{-	is_CJK_Unified_IDEOGRAPHS}', "");
Expect(1, 40960, '\p{^-	is_CJK_Unified_IDEOGRAPHS}', "");
Expect(1, 40960, '\P{-	is_CJK_Unified_IDEOGRAPHS}', "");
Expect(0, 40960, '\P{^-	is_CJK_Unified_IDEOGRAPHS}', "");
Error('\p{/a/-In_CJK_Unified_IDEOGRAPHS}');
Error('\P{/a/-In_CJK_Unified_IDEOGRAPHS}');
Expect(1, 40959, '\p{incjkunifiedideographs}', "");
Expect(0, 40959, '\p{^incjkunifiedideographs}', "");
Expect(0, 40959, '\P{incjkunifiedideographs}', "");
Expect(1, 40959, '\P{^incjkunifiedideographs}', "");
Expect(0, 40960, '\p{incjkunifiedideographs}', "");
Expect(1, 40960, '\p{^incjkunifiedideographs}', "");
Expect(1, 40960, '\P{incjkunifiedideographs}', "");
Expect(0, 40960, '\P{^incjkunifiedideographs}', "");
Expect(1, 40959, '\p{_in_cjk_Unified_Ideographs}', "");
Expect(0, 40959, '\p{^_in_cjk_Unified_Ideographs}', "");
Expect(0, 40959, '\P{_in_cjk_Unified_Ideographs}', "");
Expect(1, 40959, '\P{^_in_cjk_Unified_Ideographs}', "");
Expect(0, 40960, '\p{_in_cjk_Unified_Ideographs}', "");
Expect(1, 40960, '\p{^_in_cjk_Unified_Ideographs}', "");
Expect(1, 40960, '\P{_in_cjk_Unified_Ideographs}', "");
Expect(0, 40960, '\P{^_in_cjk_Unified_Ideographs}', "");
Error('\p{/a/-_CJK}');
Error('\P{/a/-_CJK}');
Expect(1, 40959, '\p{cjk}', "");
Expect(0, 40959, '\p{^cjk}', "");
Expect(0, 40959, '\P{cjk}', "");
Expect(1, 40959, '\P{^cjk}', "");
Expect(0, 40960, '\p{cjk}', "");
Expect(1, 40960, '\p{^cjk}', "");
Expect(1, 40960, '\P{cjk}', "");
Expect(0, 40960, '\P{^cjk}', "");
Expect(1, 40959, '\p{		CJK}', "");
Expect(0, 40959, '\p{^		CJK}', "");
Expect(0, 40959, '\P{		CJK}', "");
Expect(1, 40959, '\P{^		CJK}', "");
Expect(0, 40960, '\p{		CJK}', "");
Expect(1, 40960, '\p{^		CJK}', "");
Expect(1, 40960, '\P{		CJK}', "");
Expect(0, 40960, '\P{^		CJK}', "");
Error('\p{/a/-	Is_CJK}');
Error('\P{/a/-	Is_CJK}');
Expect(1, 40959, '\p{iscjk}', "");
Expect(0, 40959, '\p{^iscjk}', "");
Expect(0, 40959, '\P{iscjk}', "");
Expect(1, 40959, '\P{^iscjk}', "");
Expect(0, 40960, '\p{iscjk}', "");
Expect(1, 40960, '\p{^iscjk}', "");
Expect(1, 40960, '\P{iscjk}', "");
Expect(0, 40960, '\P{^iscjk}', "");
Expect(1, 40959, '\p{ 	Is_cjk}', "");
Expect(0, 40959, '\p{^ 	Is_cjk}', "");
Expect(0, 40959, '\P{ 	Is_cjk}', "");
Expect(1, 40959, '\P{^ 	Is_cjk}', "");
Expect(0, 40960, '\p{ 	Is_cjk}', "");
Expect(1, 40960, '\p{^ 	Is_cjk}', "");
Expect(1, 40960, '\P{ 	Is_cjk}', "");
Expect(0, 40960, '\P{^ 	Is_cjk}', "");
Error('\p{	/a/IN_cjk}');
Error('\P{	/a/IN_cjk}');
Expect(1, 40959, '\p{incjk}', "");
Expect(0, 40959, '\p{^incjk}', "");
Expect(0, 40959, '\P{incjk}', "");
Expect(1, 40959, '\P{^incjk}', "");
Expect(0, 40960, '\p{incjk}', "");
Expect(1, 40960, '\p{^incjk}', "");
Expect(1, 40960, '\P{incjk}', "");
Expect(0, 40960, '\P{^incjk}', "");
Expect(1, 40959, '\p{__IN_cjk}', "");
Expect(0, 40959, '\p{^__IN_cjk}', "");
Expect(0, 40959, '\P{__IN_cjk}', "");
Expect(1, 40959, '\P{^__IN_cjk}', "");
Expect(0, 40960, '\p{__IN_cjk}', "");
Expect(1, 40960, '\p{^__IN_cjk}', "");
Expect(1, 40960, '\P{__IN_cjk}', "");
Expect(0, 40960, '\P{^__IN_cjk}', "");
Error('\p{-	id_continue:=}');
Error('\P{-	id_continue:=}');
Expect(1, 917999, '\p{idcontinue}', "");
Expect(0, 917999, '\p{^idcontinue}', "");
Expect(0, 917999, '\P{idcontinue}', "");
Expect(1, 917999, '\P{^idcontinue}', "");
Expect(0, 918000, '\p{idcontinue}', "");
Expect(1, 918000, '\p{^idcontinue}', "");
Expect(1, 918000, '\P{idcontinue}', "");
Expect(0, 918000, '\P{^idcontinue}', "");
Expect(1, 917999, '\p{  ID_Continue}', "");
Expect(0, 917999, '\p{^  ID_Continue}', "");
Expect(0, 917999, '\P{  ID_Continue}', "");
Expect(1, 917999, '\P{^  ID_Continue}', "");
Expect(0, 918000, '\p{  ID_Continue}', "");
Expect(1, 918000, '\p{^  ID_Continue}', "");
Expect(1, 918000, '\P{  ID_Continue}', "");
Expect(0, 918000, '\P{^  ID_Continue}', "");
Error('\p{ :=Is_ID_Continue}');
Error('\P{ :=Is_ID_Continue}');
Expect(1, 917999, '\p{isidcontinue}', "");
Expect(0, 917999, '\p{^isidcontinue}', "");
Expect(0, 917999, '\P{isidcontinue}', "");
Expect(1, 917999, '\P{^isidcontinue}', "");
Expect(0, 918000, '\p{isidcontinue}', "");
Expect(1, 918000, '\p{^isidcontinue}', "");
Expect(1, 918000, '\P{isidcontinue}', "");
Expect(0, 918000, '\P{^isidcontinue}', "");
Expect(1, 917999, '\p{Is_ID_continue}', "");
Expect(0, 917999, '\p{^Is_ID_continue}', "");
Expect(0, 917999, '\P{Is_ID_continue}', "");
Expect(1, 917999, '\P{^Is_ID_continue}', "");
Expect(0, 918000, '\p{Is_ID_continue}', "");
Expect(1, 918000, '\p{^Is_ID_continue}', "");
Expect(1, 918000, '\P{Is_ID_continue}', "");
Expect(0, 918000, '\P{^Is_ID_continue}', "");
Error('\p{	 IDC:=}');
Error('\P{	 IDC:=}');
Expect(1, 917999, '\p{idc}', "");
Expect(0, 917999, '\p{^idc}', "");
Expect(0, 917999, '\P{idc}', "");
Expect(1, 917999, '\P{^idc}', "");
Expect(0, 918000, '\p{idc}', "");
Expect(1, 918000, '\p{^idc}', "");
Expect(1, 918000, '\P{idc}', "");
Expect(0, 918000, '\P{^idc}', "");
Expect(1, 917999, '\p{	_IDC}', "");
Expect(0, 917999, '\p{^	_IDC}', "");
Expect(0, 917999, '\P{	_IDC}', "");
Expect(1, 917999, '\P{^	_IDC}', "");
Expect(0, 918000, '\p{	_IDC}', "");
Expect(1, 918000, '\p{^	_IDC}', "");
Expect(1, 918000, '\P{	_IDC}', "");
Expect(0, 918000, '\P{^	_IDC}', "");
Error('\p{- is_IDC:=}');
Error('\P{- is_IDC:=}');
Expect(1, 917999, '\p{isidc}', "");
Expect(0, 917999, '\p{^isidc}', "");
Expect(0, 917999, '\P{isidc}', "");
Expect(1, 917999, '\P{^isidc}', "");
Expect(0, 918000, '\p{isidc}', "");
Expect(1, 918000, '\p{^isidc}', "");
Expect(1, 918000, '\P{isidc}', "");
Expect(0, 918000, '\P{^isidc}', "");
Expect(1, 917999, '\p{-is_idc}', "");
Expect(0, 917999, '\p{^-is_idc}', "");
Expect(0, 917999, '\P{-is_idc}', "");
Expect(1, 917999, '\P{^-is_idc}', "");
Expect(0, 918000, '\p{-is_idc}', "");
Expect(1, 918000, '\p{^-is_idc}', "");
Expect(1, 918000, '\P{-is_idc}', "");
Expect(0, 918000, '\P{^-is_idc}', "");
Error('\p{_:=IN_glagolitic}');
Error('\P{_:=IN_glagolitic}');
Expect(1, 11359, '\p{inglagolitic}', "");
Expect(0, 11359, '\p{^inglagolitic}', "");
Expect(0, 11359, '\P{inglagolitic}', "");
Expect(1, 11359, '\P{^inglagolitic}', "");
Expect(0, 11360, '\p{inglagolitic}', "");
Expect(1, 11360, '\p{^inglagolitic}', "");
Expect(1, 11360, '\P{inglagolitic}', "");
Expect(0, 11360, '\P{^inglagolitic}', "");
Expect(1, 11359, '\p{	In_GLAGOLITIC}', "");
Expect(0, 11359, '\p{^	In_GLAGOLITIC}', "");
Expect(0, 11359, '\P{	In_GLAGOLITIC}', "");
Expect(1, 11359, '\P{^	In_GLAGOLITIC}', "");
Expect(0, 11360, '\p{	In_GLAGOLITIC}', "");
Expect(1, 11360, '\p{^	In_GLAGOLITIC}', "");
Expect(1, 11360, '\P{	In_GLAGOLITIC}', "");
Expect(0, 11360, '\P{^	In_GLAGOLITIC}', "");
Error('\p{_In_GLAGOLITIC/a/}');
Error('\P{_In_GLAGOLITIC/a/}');
Expect(1, 11359, '\p{	_IN_Glagolitic}', "");
Expect(0, 11359, '\p{^	_IN_Glagolitic}', "");
Expect(0, 11359, '\P{	_IN_Glagolitic}', "");
Expect(1, 11359, '\P{^	_IN_Glagolitic}', "");
Expect(0, 11360, '\p{	_IN_Glagolitic}', "");
Expect(1, 11360, '\p{^	_IN_Glagolitic}', "");
Expect(1, 11360, '\P{	_IN_Glagolitic}', "");
Expect(0, 11360, '\P{^	_IN_Glagolitic}', "");
Error('\p{_:=kana_supplement}');
Error('\P{_:=kana_supplement}');
Expect(1, 110847, '\p{kanasupplement}', "");
Expect(0, 110847, '\p{^kanasupplement}', "");
Expect(0, 110847, '\P{kanasupplement}', "");
Expect(1, 110847, '\P{^kanasupplement}', "");
Expect(0, 110848, '\p{kanasupplement}', "");
Expect(1, 110848, '\p{^kanasupplement}', "");
Expect(1, 110848, '\P{kanasupplement}', "");
Expect(0, 110848, '\P{^kanasupplement}', "");
Expect(1, 110847, '\p{--Kana_Supplement}', "");
Expect(0, 110847, '\p{^--Kana_Supplement}', "");
Expect(0, 110847, '\P{--Kana_Supplement}', "");
Expect(1, 110847, '\P{^--Kana_Supplement}', "");
Expect(0, 110848, '\p{--Kana_Supplement}', "");
Expect(1, 110848, '\p{^--Kana_Supplement}', "");
Expect(1, 110848, '\P{--Kana_Supplement}', "");
Expect(0, 110848, '\P{^--Kana_Supplement}', "");
Error('\p{	:=Is_Kana_Supplement}');
Error('\P{	:=Is_Kana_Supplement}');
Expect(1, 110847, '\p{iskanasupplement}', "");
Expect(0, 110847, '\p{^iskanasupplement}', "");
Expect(0, 110847, '\P{iskanasupplement}', "");
Expect(1, 110847, '\P{^iskanasupplement}', "");
Expect(0, 110848, '\p{iskanasupplement}', "");
Expect(1, 110848, '\p{^iskanasupplement}', "");
Expect(1, 110848, '\P{iskanasupplement}', "");
Expect(0, 110848, '\P{^iskanasupplement}', "");
Expect(1, 110847, '\p{- Is_Kana_supplement}', "");
Expect(0, 110847, '\p{^- Is_Kana_supplement}', "");
Expect(0, 110847, '\P{- Is_Kana_supplement}', "");
Expect(1, 110847, '\P{^- Is_Kana_supplement}', "");
Expect(0, 110848, '\p{- Is_Kana_supplement}', "");
Expect(1, 110848, '\p{^- Is_Kana_supplement}', "");
Expect(1, 110848, '\P{- Is_Kana_supplement}', "");
Expect(0, 110848, '\P{^- Is_Kana_supplement}', "");
Error('\p{/a/_in_Kana_SUPPLEMENT}');
Error('\P{/a/_in_Kana_SUPPLEMENT}');
Expect(1, 110847, '\p{inkanasupplement}', "");
Expect(0, 110847, '\p{^inkanasupplement}', "");
Expect(0, 110847, '\P{inkanasupplement}', "");
Expect(1, 110847, '\P{^inkanasupplement}', "");
Expect(0, 110848, '\p{inkanasupplement}', "");
Expect(1, 110848, '\p{^inkanasupplement}', "");
Expect(1, 110848, '\P{inkanasupplement}', "");
Expect(0, 110848, '\P{^inkanasupplement}', "");
Expect(1, 110847, '\p{_ in_Kana_Supplement}', "");
Expect(0, 110847, '\p{^_ in_Kana_Supplement}', "");
Expect(0, 110847, '\P{_ in_Kana_Supplement}', "");
Expect(1, 110847, '\P{^_ in_Kana_Supplement}', "");
Expect(0, 110848, '\p{_ in_Kana_Supplement}', "");
Expect(1, 110848, '\p{^_ in_Kana_Supplement}', "");
Expect(1, 110848, '\P{_ in_Kana_Supplement}', "");
Expect(0, 110848, '\P{^_ in_Kana_Supplement}', "");
Error('\p{--Kana_Sup/a/}');
Error('\P{--Kana_Sup/a/}');
Expect(1, 110847, '\p{kanasup}', "");
Expect(0, 110847, '\p{^kanasup}', "");
Expect(0, 110847, '\P{kanasup}', "");
Expect(1, 110847, '\P{^kanasup}', "");
Expect(0, 110848, '\p{kanasup}', "");
Expect(1, 110848, '\p{^kanasup}', "");
Expect(1, 110848, '\P{kanasup}', "");
Expect(0, 110848, '\P{^kanasup}', "");
Expect(1, 110847, '\p{ Kana_sup}', "");
Expect(0, 110847, '\p{^ Kana_sup}', "");
Expect(0, 110847, '\P{ Kana_sup}', "");
Expect(1, 110847, '\P{^ Kana_sup}', "");
Expect(0, 110848, '\p{ Kana_sup}', "");
Expect(1, 110848, '\p{^ Kana_sup}', "");
Expect(1, 110848, '\P{ Kana_sup}', "");
Expect(0, 110848, '\P{^ Kana_sup}', "");
Error('\p{	/a/IS_Kana_Sup}');
Error('\P{	/a/IS_Kana_Sup}');
Expect(1, 110847, '\p{iskanasup}', "");
Expect(0, 110847, '\p{^iskanasup}', "");
Expect(0, 110847, '\P{iskanasup}', "");
Expect(1, 110847, '\P{^iskanasup}', "");
Expect(0, 110848, '\p{iskanasup}', "");
Expect(1, 110848, '\p{^iskanasup}', "");
Expect(1, 110848, '\P{iskanasup}', "");
Expect(0, 110848, '\P{^iskanasup}', "");
Expect(1, 110847, '\p{	Is_KANA_Sup}', "");
Expect(0, 110847, '\p{^	Is_KANA_Sup}', "");
Expect(0, 110847, '\P{	Is_KANA_Sup}', "");
Expect(1, 110847, '\P{^	Is_KANA_Sup}', "");
Expect(0, 110848, '\p{	Is_KANA_Sup}', "");
Expect(1, 110848, '\p{^	Is_KANA_Sup}', "");
Expect(1, 110848, '\P{	Is_KANA_Sup}', "");
Expect(0, 110848, '\P{^	Is_KANA_Sup}', "");
Error('\p{ IN_kana_Sup:=}');
Error('\P{ IN_kana_Sup:=}');
Expect(1, 110847, '\p{inkanasup}', "");
Expect(0, 110847, '\p{^inkanasup}', "");
Expect(0, 110847, '\P{inkanasup}', "");
Expect(1, 110847, '\P{^inkanasup}', "");
Expect(0, 110848, '\p{inkanasup}', "");
Expect(1, 110848, '\p{^inkanasup}', "");
Expect(1, 110848, '\P{inkanasup}', "");
Expect(0, 110848, '\P{^inkanasup}', "");
Expect(1, 110847, '\p{--In_Kana_Sup}', "");
Expect(0, 110847, '\p{^--In_Kana_Sup}', "");
Expect(0, 110847, '\P{--In_Kana_Sup}', "");
Expect(1, 110847, '\P{^--In_Kana_Sup}', "");
Expect(0, 110848, '\p{--In_Kana_Sup}', "");
Expect(1, 110848, '\p{^--In_Kana_Sup}', "");
Expect(1, 110848, '\P{--In_Kana_Sup}', "");
Expect(0, 110848, '\P{^--In_Kana_Sup}', "");
Error('\p{/a/_-open_PUNCTUATION}');
Error('\P{/a/_-open_PUNCTUATION}');
Expect(1, 65378, '\p{openpunctuation}', "");
Expect(0, 65378, '\p{^openpunctuation}', "");
Expect(0, 65378, '\P{openpunctuation}', "");
Expect(1, 65378, '\P{^openpunctuation}', "");
Expect(0, 65379, '\p{openpunctuation}', "");
Expect(1, 65379, '\p{^openpunctuation}', "");
Expect(1, 65379, '\P{openpunctuation}', "");
Expect(0, 65379, '\P{^openpunctuation}', "");
Expect(1, 65378, '\p{open_Punctuation}', "");
Expect(0, 65378, '\p{^open_Punctuation}', "");
Expect(0, 65378, '\P{open_Punctuation}', "");
Expect(1, 65378, '\P{^open_Punctuation}', "");
Expect(0, 65379, '\p{open_Punctuation}', "");
Expect(1, 65379, '\p{^open_Punctuation}', "");
Expect(1, 65379, '\P{open_Punctuation}', "");
Expect(0, 65379, '\P{^open_Punctuation}', "");
Error('\p{:= _Is_Open_Punctuation}');
Error('\P{:= _Is_Open_Punctuation}');
Expect(1, 65378, '\p{isopenpunctuation}', "");
Expect(0, 65378, '\p{^isopenpunctuation}', "");
Expect(0, 65378, '\P{isopenpunctuation}', "");
Expect(1, 65378, '\P{^isopenpunctuation}', "");
Expect(0, 65379, '\p{isopenpunctuation}', "");
Expect(1, 65379, '\p{^isopenpunctuation}', "");
Expect(1, 65379, '\P{isopenpunctuation}', "");
Expect(0, 65379, '\P{^isopenpunctuation}', "");
Expect(1, 65378, '\p{-is_OPEN_punctuation}', "");
Expect(0, 65378, '\p{^-is_OPEN_punctuation}', "");
Expect(0, 65378, '\P{-is_OPEN_punctuation}', "");
Expect(1, 65378, '\P{^-is_OPEN_punctuation}', "");
Expect(0, 65379, '\p{-is_OPEN_punctuation}', "");
Expect(1, 65379, '\p{^-is_OPEN_punctuation}', "");
Expect(1, 65379, '\P{-is_OPEN_punctuation}', "");
Expect(0, 65379, '\P{^-is_OPEN_punctuation}', "");
Error('\p{:=-	ps}');
Error('\P{:=-	ps}');
Expect(1, 65378, '\p{ps}', "");
Expect(0, 65378, '\p{^ps}', "");
Expect(0, 65378, '\P{ps}', "");
Expect(1, 65378, '\P{^ps}', "");
Expect(0, 65379, '\p{ps}', "");
Expect(1, 65379, '\p{^ps}', "");
Expect(1, 65379, '\P{ps}', "");
Expect(0, 65379, '\P{^ps}', "");
Expect(1, 65378, '\p{ _Ps}', "");
Expect(0, 65378, '\p{^ _Ps}', "");
Expect(0, 65378, '\P{ _Ps}', "");
Expect(1, 65378, '\P{^ _Ps}', "");
Expect(0, 65379, '\p{ _Ps}', "");
Expect(1, 65379, '\p{^ _Ps}', "");
Expect(1, 65379, '\P{ _Ps}', "");
Expect(0, 65379, '\P{^ _Ps}', "");
Error('\p{_/a/IS_PS}');
Error('\P{_/a/IS_PS}');
Expect(1, 65378, '\p{isps}', "");
Expect(0, 65378, '\p{^isps}', "");
Expect(0, 65378, '\P{isps}', "");
Expect(1, 65378, '\P{^isps}', "");
Expect(0, 65379, '\p{isps}', "");
Expect(1, 65379, '\p{^isps}', "");
Expect(1, 65379, '\P{isps}', "");
Expect(0, 65379, '\P{^isps}', "");
Expect(1, 65378, '\p{Is_ps}', "");
Expect(0, 65378, '\p{^Is_ps}', "");
Expect(0, 65378, '\P{Is_ps}', "");
Expect(1, 65378, '\P{^Is_ps}', "");
Expect(0, 65379, '\p{Is_ps}', "");
Expect(1, 65379, '\p{^Is_ps}', "");
Expect(1, 65379, '\P{Is_ps}', "");
Expect(0, 65379, '\P{^Is_ps}', "");
Error('\p{:=	HANGUL}');
Error('\P{:=	HANGUL}');
Expect(1, 65500, '\p{hangul}', "");
Expect(0, 65500, '\p{^hangul}', "");
Expect(0, 65500, '\P{hangul}', "");
Expect(1, 65500, '\P{^hangul}', "");
Expect(0, 65501, '\p{hangul}', "");
Expect(1, 65501, '\p{^hangul}', "");
Expect(1, 65501, '\P{hangul}', "");
Expect(0, 65501, '\P{^hangul}', "");
Expect(1, 65500, '\p{	 Hangul}', "");
Expect(0, 65500, '\p{^	 Hangul}', "");
Expect(0, 65500, '\P{	 Hangul}', "");
Expect(1, 65500, '\P{^	 Hangul}', "");
Expect(0, 65501, '\p{	 Hangul}', "");
Expect(1, 65501, '\p{^	 Hangul}', "");
Expect(1, 65501, '\P{	 Hangul}', "");
Expect(0, 65501, '\P{^	 Hangul}', "");
Error('\p{/a/ _Is_hangul}');
Error('\P{/a/ _Is_hangul}');
Expect(1, 65500, '\p{ishangul}', "");
Expect(0, 65500, '\p{^ishangul}', "");
Expect(0, 65500, '\P{ishangul}', "");
Expect(1, 65500, '\P{^ishangul}', "");
Expect(0, 65501, '\p{ishangul}', "");
Expect(1, 65501, '\p{^ishangul}', "");
Expect(1, 65501, '\P{ishangul}', "");
Expect(0, 65501, '\P{^ishangul}', "");
Expect(1, 65500, '\p{	 Is_Hangul}', "");
Expect(0, 65500, '\p{^	 Is_Hangul}', "");
Expect(0, 65500, '\P{	 Is_Hangul}', "");
Expect(1, 65500, '\P{^	 Is_Hangul}', "");
Expect(0, 65501, '\p{	 Is_Hangul}', "");
Expect(1, 65501, '\p{^	 Is_Hangul}', "");
Expect(1, 65501, '\P{	 Is_Hangul}', "");
Expect(0, 65501, '\P{^	 Is_Hangul}', "");
Error('\p{_	hang:=}');
Error('\P{_	hang:=}');
Expect(1, 65500, '\p{hang}', "");
Expect(0, 65500, '\p{^hang}', "");
Expect(0, 65500, '\P{hang}', "");
Expect(1, 65500, '\P{^hang}', "");
Expect(0, 65501, '\p{hang}', "");
Expect(1, 65501, '\p{^hang}', "");
Expect(1, 65501, '\P{hang}', "");
Expect(0, 65501, '\P{^hang}', "");
Expect(1, 65500, '\p{	_Hang}', "");
Expect(0, 65500, '\p{^	_Hang}', "");
Expect(0, 65500, '\P{	_Hang}', "");
Expect(1, 65500, '\P{^	_Hang}', "");
Expect(0, 65501, '\p{	_Hang}', "");
Expect(1, 65501, '\p{^	_Hang}', "");
Expect(1, 65501, '\P{	_Hang}', "");
Expect(0, 65501, '\P{^	_Hang}', "");
Error('\p{  Is_Hang:=}');
Error('\P{  Is_Hang:=}');
Expect(1, 65500, '\p{ishang}', "");
Expect(0, 65500, '\p{^ishang}', "");
Expect(0, 65500, '\P{ishang}', "");
Expect(1, 65500, '\P{^ishang}', "");
Expect(0, 65501, '\p{ishang}', "");
Expect(1, 65501, '\p{^ishang}', "");
Expect(1, 65501, '\P{ishang}', "");
Expect(0, 65501, '\P{^ishang}', "");
Expect(1, 65500, '\p{	-is_HANG}', "");
Expect(0, 65500, '\p{^	-is_HANG}', "");
Expect(0, 65500, '\P{	-is_HANG}', "");
Expect(1, 65500, '\P{^	-is_HANG}', "");
Expect(0, 65501, '\p{	-is_HANG}', "");
Expect(1, 65501, '\p{^	-is_HANG}', "");
Expect(1, 65501, '\P{	-is_HANG}', "");
Expect(0, 65501, '\P{^	-is_HANG}', "");
Error('\p{:= 	Cypriot}');
Error('\P{:= 	Cypriot}');
Expect(1, 67647, '\p{cypriot}', "");
Expect(0, 67647, '\p{^cypriot}', "");
Expect(0, 67647, '\P{cypriot}', "");
Expect(1, 67647, '\P{^cypriot}', "");
Expect(0, 67648, '\p{cypriot}', "");
Expect(1, 67648, '\p{^cypriot}', "");
Expect(1, 67648, '\P{cypriot}', "");
Expect(0, 67648, '\P{^cypriot}', "");
Expect(1, 67647, '\p{_ Cypriot}', "");
Expect(0, 67647, '\p{^_ Cypriot}', "");
Expect(0, 67647, '\P{_ Cypriot}', "");
Expect(1, 67647, '\P{^_ Cypriot}', "");
Expect(0, 67648, '\p{_ Cypriot}', "");
Expect(1, 67648, '\p{^_ Cypriot}', "");
Expect(1, 67648, '\P{_ Cypriot}', "");
Expect(0, 67648, '\P{^_ Cypriot}', "");
Error('\p{	-IS_CYPRIOT/a/}');
Error('\P{	-IS_CYPRIOT/a/}');
Expect(1, 67647, '\p{iscypriot}', "");
Expect(0, 67647, '\p{^iscypriot}', "");
Expect(0, 67647, '\P{iscypriot}', "");
Expect(1, 67647, '\P{^iscypriot}', "");
Expect(0, 67648, '\p{iscypriot}', "");
Expect(1, 67648, '\p{^iscypriot}', "");
Expect(1, 67648, '\P{iscypriot}', "");
Expect(0, 67648, '\P{^iscypriot}', "");
Expect(1, 67647, '\p{-is_Cypriot}', "");
Expect(0, 67647, '\p{^-is_Cypriot}', "");
Expect(0, 67647, '\P{-is_Cypriot}', "");
Expect(1, 67647, '\P{^-is_Cypriot}', "");
Expect(0, 67648, '\p{-is_Cypriot}', "");
Expect(1, 67648, '\p{^-is_Cypriot}', "");
Expect(1, 67648, '\P{-is_Cypriot}', "");
Expect(0, 67648, '\P{^-is_Cypriot}', "");
Error('\p{:= _Cprt}');
Error('\P{:= _Cprt}');
Expect(1, 67647, '\p{cprt}', "");
Expect(0, 67647, '\p{^cprt}', "");
Expect(0, 67647, '\P{cprt}', "");
Expect(1, 67647, '\P{^cprt}', "");
Expect(0, 67648, '\p{cprt}', "");
Expect(1, 67648, '\p{^cprt}', "");
Expect(1, 67648, '\P{cprt}', "");
Expect(0, 67648, '\P{^cprt}', "");
Expect(1, 67647, '\p{ _CPRT}', "");
Expect(0, 67647, '\p{^ _CPRT}', "");
Expect(0, 67647, '\P{ _CPRT}', "");
Expect(1, 67647, '\P{^ _CPRT}', "");
Expect(0, 67648, '\p{ _CPRT}', "");
Expect(1, 67648, '\p{^ _CPRT}', "");
Expect(1, 67648, '\P{ _CPRT}', "");
Expect(0, 67648, '\P{^ _CPRT}', "");
Error('\p{/a/ Is_cprt}');
Error('\P{/a/ Is_cprt}');
Expect(1, 67647, '\p{iscprt}', "");
Expect(0, 67647, '\p{^iscprt}', "");
Expect(0, 67647, '\P{iscprt}', "");
Expect(1, 67647, '\P{^iscprt}', "");
Expect(0, 67648, '\p{iscprt}', "");
Expect(1, 67648, '\p{^iscprt}', "");
Expect(1, 67648, '\P{iscprt}', "");
Expect(0, 67648, '\P{^iscprt}', "");
Expect(1, 67647, '\p{		Is_Cprt}', "");
Expect(0, 67647, '\p{^		Is_Cprt}', "");
Expect(0, 67647, '\P{		Is_Cprt}', "");
Expect(1, 67647, '\P{^		Is_Cprt}', "");
Expect(0, 67648, '\p{		Is_Cprt}', "");
Expect(1, 67648, '\p{^		Is_Cprt}', "");
Expect(1, 67648, '\P{		Is_Cprt}', "");
Expect(0, 67648, '\P{^		Is_Cprt}', "");
Error('\p{	_byzantine_Musical_Symbols:=}');
Error('\P{	_byzantine_Musical_Symbols:=}');
Expect(1, 119039, '\p{byzantinemusicalsymbols}', "");
Expect(0, 119039, '\p{^byzantinemusicalsymbols}', "");
Expect(0, 119039, '\P{byzantinemusicalsymbols}', "");
Expect(1, 119039, '\P{^byzantinemusicalsymbols}', "");
Expect(0, 119040, '\p{byzantinemusicalsymbols}', "");
Expect(1, 119040, '\p{^byzantinemusicalsymbols}', "");
Expect(1, 119040, '\P{byzantinemusicalsymbols}', "");
Expect(0, 119040, '\P{^byzantinemusicalsymbols}', "");
Expect(1, 119039, '\p{ Byzantine_MUSICAL_SYMBOLS}', "");
Expect(0, 119039, '\p{^ Byzantine_MUSICAL_SYMBOLS}', "");
Expect(0, 119039, '\P{ Byzantine_MUSICAL_SYMBOLS}', "");
Expect(1, 119039, '\P{^ Byzantine_MUSICAL_SYMBOLS}', "");
Expect(0, 119040, '\p{ Byzantine_MUSICAL_SYMBOLS}', "");
Expect(1, 119040, '\p{^ Byzantine_MUSICAL_SYMBOLS}', "");
Expect(1, 119040, '\P{ Byzantine_MUSICAL_SYMBOLS}', "");
Expect(0, 119040, '\P{^ Byzantine_MUSICAL_SYMBOLS}', "");
Error('\p{:=_is_BYZANTINE_Musical_SYMBOLS}');
Error('\P{:=_is_BYZANTINE_Musical_SYMBOLS}');
Expect(1, 119039, '\p{isbyzantinemusicalsymbols}', "");
Expect(0, 119039, '\p{^isbyzantinemusicalsymbols}', "");
Expect(0, 119039, '\P{isbyzantinemusicalsymbols}', "");
Expect(1, 119039, '\P{^isbyzantinemusicalsymbols}', "");
Expect(0, 119040, '\p{isbyzantinemusicalsymbols}', "");
Expect(1, 119040, '\p{^isbyzantinemusicalsymbols}', "");
Expect(1, 119040, '\P{isbyzantinemusicalsymbols}', "");
Expect(0, 119040, '\P{^isbyzantinemusicalsymbols}', "");
Expect(1, 119039, '\p{  IS_BYZANTINE_musical_Symbols}', "");
Expect(0, 119039, '\p{^  IS_BYZANTINE_musical_Symbols}', "");
Expect(0, 119039, '\P{  IS_BYZANTINE_musical_Symbols}', "");
Expect(1, 119039, '\P{^  IS_BYZANTINE_musical_Symbols}', "");
Expect(0, 119040, '\p{  IS_BYZANTINE_musical_Symbols}', "");
Expect(1, 119040, '\p{^  IS_BYZANTINE_musical_Symbols}', "");
Expect(1, 119040, '\P{  IS_BYZANTINE_musical_Symbols}', "");
Expect(0, 119040, '\P{^  IS_BYZANTINE_musical_Symbols}', "");
Error('\p{:=		IN_Byzantine_musical_SYMBOLS}');
Error('\P{:=		IN_Byzantine_musical_SYMBOLS}');
Expect(1, 119039, '\p{inbyzantinemusicalsymbols}', "");
Expect(0, 119039, '\p{^inbyzantinemusicalsymbols}', "");
Expect(0, 119039, '\P{inbyzantinemusicalsymbols}', "");
Expect(1, 119039, '\P{^inbyzantinemusicalsymbols}', "");
Expect(0, 119040, '\p{inbyzantinemusicalsymbols}', "");
Expect(1, 119040, '\p{^inbyzantinemusicalsymbols}', "");
Expect(1, 119040, '\P{inbyzantinemusicalsymbols}', "");
Expect(0, 119040, '\P{^inbyzantinemusicalsymbols}', "");
Expect(1, 119039, '\p{_in_byzantine_Musical_symbols}', "");
Expect(0, 119039, '\p{^_in_byzantine_Musical_symbols}', "");
Expect(0, 119039, '\P{_in_byzantine_Musical_symbols}', "");
Expect(1, 119039, '\P{^_in_byzantine_Musical_symbols}', "");
Expect(0, 119040, '\p{_in_byzantine_Musical_symbols}', "");
Expect(1, 119040, '\p{^_in_byzantine_Musical_symbols}', "");
Expect(1, 119040, '\P{_in_byzantine_Musical_symbols}', "");
Expect(0, 119040, '\P{^_in_byzantine_Musical_symbols}', "");
Error('\p{_/a/byzantine_Music}');
Error('\P{_/a/byzantine_Music}');
Expect(1, 119039, '\p{byzantinemusic}', "");
Expect(0, 119039, '\p{^byzantinemusic}', "");
Expect(0, 119039, '\P{byzantinemusic}', "");
Expect(1, 119039, '\P{^byzantinemusic}', "");
Expect(0, 119040, '\p{byzantinemusic}', "");
Expect(1, 119040, '\p{^byzantinemusic}', "");
Expect(1, 119040, '\P{byzantinemusic}', "");
Expect(0, 119040, '\P{^byzantinemusic}', "");
Expect(1, 119039, '\p{--Byzantine_Music}', "");
Expect(0, 119039, '\p{^--Byzantine_Music}', "");
Expect(0, 119039, '\P{--Byzantine_Music}', "");
Expect(1, 119039, '\P{^--Byzantine_Music}', "");
Expect(0, 119040, '\p{--Byzantine_Music}', "");
Expect(1, 119040, '\p{^--Byzantine_Music}', "");
Expect(1, 119040, '\P{--Byzantine_Music}', "");
Expect(0, 119040, '\P{^--Byzantine_Music}', "");
Error('\p{:=_is_Byzantine_music}');
Error('\P{:=_is_Byzantine_music}');
Expect(1, 119039, '\p{isbyzantinemusic}', "");
Expect(0, 119039, '\p{^isbyzantinemusic}', "");
Expect(0, 119039, '\P{isbyzantinemusic}', "");
Expect(1, 119039, '\P{^isbyzantinemusic}', "");
Expect(0, 119040, '\p{isbyzantinemusic}', "");
Expect(1, 119040, '\p{^isbyzantinemusic}', "");
Expect(1, 119040, '\P{isbyzantinemusic}', "");
Expect(0, 119040, '\P{^isbyzantinemusic}', "");
Expect(1, 119039, '\p{	-is_Byzantine_music}', "");
Expect(0, 119039, '\p{^	-is_Byzantine_music}', "");
Expect(0, 119039, '\P{	-is_Byzantine_music}', "");
Expect(1, 119039, '\P{^	-is_Byzantine_music}', "");
Expect(0, 119040, '\p{	-is_Byzantine_music}', "");
Expect(1, 119040, '\p{^	-is_Byzantine_music}', "");
Expect(1, 119040, '\P{	-is_Byzantine_music}', "");
Expect(0, 119040, '\P{^	-is_Byzantine_music}', "");
Error('\p{:=-In_Byzantine_Music}');
Error('\P{:=-In_Byzantine_Music}');
Expect(1, 119039, '\p{inbyzantinemusic}', "");
Expect(0, 119039, '\p{^inbyzantinemusic}', "");
Expect(0, 119039, '\P{inbyzantinemusic}', "");
Expect(1, 119039, '\P{^inbyzantinemusic}', "");
Expect(0, 119040, '\p{inbyzantinemusic}', "");
Expect(1, 119040, '\p{^inbyzantinemusic}', "");
Expect(1, 119040, '\P{inbyzantinemusic}', "");
Expect(0, 119040, '\P{^inbyzantinemusic}', "");
Expect(1, 119039, '\p{_In_byzantine_MUSIC}', "");
Expect(0, 119039, '\p{^_In_byzantine_MUSIC}', "");
Expect(0, 119039, '\P{_In_byzantine_MUSIC}', "");
Expect(1, 119039, '\P{^_In_byzantine_MUSIC}', "");
Expect(0, 119040, '\p{_In_byzantine_MUSIC}', "");
Expect(1, 119040, '\p{^_In_byzantine_MUSIC}', "");
Expect(1, 119040, '\P{_In_byzantine_MUSIC}', "");
Expect(0, 119040, '\P{^_In_byzantine_MUSIC}', "");
Error('\p{:=_PosixUpper}');
Error('\P{:=_PosixUpper}');
Expect(1, 90, '\p{posixupper}', "");
Expect(0, 90, '\p{^posixupper}', "");
Expect(0, 90, '\P{posixupper}', "");
Expect(1, 90, '\P{^posixupper}', "");
Expect(0, 91, '\p{posixupper}', "");
Expect(1, 91, '\p{^posixupper}', "");
Expect(1, 91, '\P{posixupper}', "");
Expect(0, 91, '\P{^posixupper}', "");
Expect(1, 90, '\p{_-posixupper}', "");
Expect(0, 90, '\p{^_-posixupper}', "");
Expect(0, 90, '\P{_-posixupper}', "");
Expect(1, 90, '\P{^_-posixupper}', "");
Expect(0, 91, '\p{_-posixupper}', "");
Expect(1, 91, '\p{^_-posixupper}', "");
Expect(1, 91, '\P{_-posixupper}', "");
Expect(0, 91, '\P{^_-posixupper}', "");
Error('\p{:=_Is_posixupper}');
Error('\P{:=_Is_posixupper}');
Expect(1, 90, '\p{isposixupper}', "");
Expect(0, 90, '\p{^isposixupper}', "");
Expect(0, 90, '\P{isposixupper}', "");
Expect(1, 90, '\P{^isposixupper}', "");
Expect(0, 91, '\p{isposixupper}', "");
Expect(1, 91, '\p{^isposixupper}', "");
Expect(1, 91, '\P{isposixupper}', "");
Expect(0, 91, '\P{^isposixupper}', "");
Expect(1, 90, '\p{	_IS_posixupper}', "");
Expect(0, 90, '\p{^	_IS_posixupper}', "");
Expect(0, 90, '\P{	_IS_posixupper}', "");
Expect(1, 90, '\P{^	_IS_posixupper}', "");
Expect(0, 91, '\p{	_IS_posixupper}', "");
Expect(1, 91, '\p{^	_IS_posixupper}', "");
Expect(1, 91, '\P{	_IS_posixupper}', "");
Expect(0, 91, '\P{^	_IS_posixupper}', "");
Error('\p{/a/--High_private_Use_Surrogates}');
Error('\P{/a/--High_private_Use_Surrogates}');
Expect(1, 56319, '\p{highprivateusesurrogates}', "");
Expect(0, 56319, '\p{^highprivateusesurrogates}', "");
Expect(0, 56319, '\P{highprivateusesurrogates}', "");
Expect(1, 56319, '\P{^highprivateusesurrogates}', "");
Expect(0, 57344, '\p{highprivateusesurrogates}', "");
Expect(1, 57344, '\p{^highprivateusesurrogates}', "");
Expect(1, 57344, '\P{highprivateusesurrogates}', "");
Expect(0, 57344, '\P{^highprivateusesurrogates}', "");
Expect(1, 56319, '\p{	-High_PRIVATE_Use_Surrogates}', "");
Expect(0, 56319, '\p{^	-High_PRIVATE_Use_Surrogates}', "");
Expect(0, 56319, '\P{	-High_PRIVATE_Use_Surrogates}', "");
Expect(1, 56319, '\P{^	-High_PRIVATE_Use_Surrogates}', "");
Expect(0, 57344, '\p{	-High_PRIVATE_Use_Surrogates}', "");
Expect(1, 57344, '\p{^	-High_PRIVATE_Use_Surrogates}', "");
Expect(1, 57344, '\P{	-High_PRIVATE_Use_Surrogates}', "");
Expect(0, 57344, '\P{^	-High_PRIVATE_Use_Surrogates}', "");
Error('\p{:=	Is_High_Private_use_Surrogates}');
Error('\P{:=	Is_High_Private_use_Surrogates}');
Expect(1, 56319, '\p{ishighprivateusesurrogates}', "");
Expect(0, 56319, '\p{^ishighprivateusesurrogates}', "");
Expect(0, 56319, '\P{ishighprivateusesurrogates}', "");
Expect(1, 56319, '\P{^ishighprivateusesurrogates}', "");
Expect(0, 57344, '\p{ishighprivateusesurrogates}', "");
Expect(1, 57344, '\p{^ishighprivateusesurrogates}', "");
Expect(1, 57344, '\P{ishighprivateusesurrogates}', "");
Expect(0, 57344, '\P{^ishighprivateusesurrogates}', "");
Expect(1, 56319, '\p{		IS_high_PRIVATE_Use_Surrogates}', "");
Expect(0, 56319, '\p{^		IS_high_PRIVATE_Use_Surrogates}', "");
Expect(0, 56319, '\P{		IS_high_PRIVATE_Use_Surrogates}', "");
Expect(1, 56319, '\P{^		IS_high_PRIVATE_Use_Surrogates}', "");
Expect(0, 57344, '\p{		IS_high_PRIVATE_Use_Surrogates}', "");
Expect(1, 57344, '\p{^		IS_high_PRIVATE_Use_Surrogates}', "");
Expect(1, 57344, '\P{		IS_high_PRIVATE_Use_Surrogates}', "");
Expect(0, 57344, '\P{^		IS_high_PRIVATE_Use_Surrogates}', "");
Error('\p{/a/ In_high_PRIVATE_USE_surrogates}');
Error('\P{/a/ In_high_PRIVATE_USE_surrogates}');
Expect(1, 56319, '\p{inhighprivateusesurrogates}', "");
Expect(0, 56319, '\p{^inhighprivateusesurrogates}', "");
Expect(0, 56319, '\P{inhighprivateusesurrogates}', "");
Expect(1, 56319, '\P{^inhighprivateusesurrogates}', "");
Expect(0, 57344, '\p{inhighprivateusesurrogates}', "");
Expect(1, 57344, '\p{^inhighprivateusesurrogates}', "");
Expect(1, 57344, '\P{inhighprivateusesurrogates}', "");
Expect(0, 57344, '\P{^inhighprivateusesurrogates}', "");
Expect(1, 56319, '\p{ -in_High_Private_Use_Surrogates}', "");
Expect(0, 56319, '\p{^ -in_High_Private_Use_Surrogates}', "");
Expect(0, 56319, '\P{ -in_High_Private_Use_Surrogates}', "");
Expect(1, 56319, '\P{^ -in_High_Private_Use_Surrogates}', "");
Expect(0, 57344, '\p{ -in_High_Private_Use_Surrogates}', "");
Expect(1, 57344, '\p{^ -in_High_Private_Use_Surrogates}', "");
Expect(1, 57344, '\P{ -in_High_Private_Use_Surrogates}', "");
Expect(0, 57344, '\P{^ -in_High_Private_Use_Surrogates}', "");
Error('\p{/a/-HIGH_PU_surrogates}');
Error('\P{/a/-HIGH_PU_surrogates}');
Expect(1, 56319, '\p{highpusurrogates}', "");
Expect(0, 56319, '\p{^highpusurrogates}', "");
Expect(0, 56319, '\P{highpusurrogates}', "");
Expect(1, 56319, '\P{^highpusurrogates}', "");
Expect(0, 57344, '\p{highpusurrogates}', "");
Expect(1, 57344, '\p{^highpusurrogates}', "");
Expect(1, 57344, '\P{highpusurrogates}', "");
Expect(0, 57344, '\P{^highpusurrogates}', "");
Expect(1, 56319, '\p{ High_PU_SURROGATES}', "");
Expect(0, 56319, '\p{^ High_PU_SURROGATES}', "");
Expect(0, 56319, '\P{ High_PU_SURROGATES}', "");
Expect(1, 56319, '\P{^ High_PU_SURROGATES}', "");
Expect(0, 57344, '\p{ High_PU_SURROGATES}', "");
Expect(1, 57344, '\p{^ High_PU_SURROGATES}', "");
Expect(1, 57344, '\P{ High_PU_SURROGATES}', "");
Expect(0, 57344, '\P{^ High_PU_SURROGATES}', "");
Error('\p{:=IS_High_pu_SURROGATES}');
Error('\P{:=IS_High_pu_SURROGATES}');
Expect(1, 56319, '\p{ishighpusurrogates}', "");
Expect(0, 56319, '\p{^ishighpusurrogates}', "");
Expect(0, 56319, '\P{ishighpusurrogates}', "");
Expect(1, 56319, '\P{^ishighpusurrogates}', "");
Expect(0, 57344, '\p{ishighpusurrogates}', "");
Expect(1, 57344, '\p{^ishighpusurrogates}', "");
Expect(1, 57344, '\P{ishighpusurrogates}', "");
Expect(0, 57344, '\P{^ishighpusurrogates}', "");
Expect(1, 56319, '\p{__Is_High_PU_SURROGATES}', "");
Expect(0, 56319, '\p{^__Is_High_PU_SURROGATES}', "");
Expect(0, 56319, '\P{__Is_High_PU_SURROGATES}', "");
Expect(1, 56319, '\P{^__Is_High_PU_SURROGATES}', "");
Expect(0, 57344, '\p{__Is_High_PU_SURROGATES}', "");
Expect(1, 57344, '\p{^__Is_High_PU_SURROGATES}', "");
Expect(1, 57344, '\P{__Is_High_PU_SURROGATES}', "");
Expect(0, 57344, '\P{^__Is_High_PU_SURROGATES}', "");
Error('\p{/a/ IN_high_PU_Surrogates}');
Error('\P{/a/ IN_high_PU_Surrogates}');
Expect(1, 56319, '\p{inhighpusurrogates}', "");
Expect(0, 56319, '\p{^inhighpusurrogates}', "");
Expect(0, 56319, '\P{inhighpusurrogates}', "");
Expect(1, 56319, '\P{^inhighpusurrogates}', "");
Expect(0, 57344, '\p{inhighpusurrogates}', "");
Expect(1, 57344, '\p{^inhighpusurrogates}', "");
Expect(1, 57344, '\P{inhighpusurrogates}', "");
Expect(0, 57344, '\P{^inhighpusurrogates}', "");
Expect(1, 56319, '\p{	In_High_pu_Surrogates}', "");
Expect(0, 56319, '\p{^	In_High_pu_Surrogates}', "");
Expect(0, 56319, '\P{	In_High_pu_Surrogates}', "");
Expect(1, 56319, '\P{^	In_High_pu_Surrogates}', "");
Expect(0, 57344, '\p{	In_High_pu_Surrogates}', "");
Expect(1, 57344, '\p{^	In_High_pu_Surrogates}', "");
Expect(1, 57344, '\P{	In_High_pu_Surrogates}', "");
Expect(0, 57344, '\P{^	In_High_pu_Surrogates}', "");
Error('\p{/a/ -xposixpunct}');
Error('\P{/a/ -xposixpunct}');
Expect(1, 74867, '\p{xposixpunct}', "");
Expect(0, 74867, '\p{^xposixpunct}', "");
Expect(0, 74867, '\P{xposixpunct}', "");
Expect(1, 74867, '\P{^xposixpunct}', "");
Expect(0, 74868, '\p{xposixpunct}', "");
Expect(1, 74868, '\p{^xposixpunct}', "");
Expect(1, 74868, '\P{xposixpunct}', "");
Expect(0, 74868, '\P{^xposixpunct}', "");
Expect(1, 74867, '\p{	 xposixpunct}', "");
Expect(0, 74867, '\p{^	 xposixpunct}', "");
Expect(0, 74867, '\P{	 xposixpunct}', "");
Expect(1, 74867, '\P{^	 xposixpunct}', "");
Expect(0, 74868, '\p{	 xposixpunct}', "");
Expect(1, 74868, '\p{^	 xposixpunct}', "");
Expect(1, 74868, '\P{	 xposixpunct}', "");
Expect(0, 74868, '\P{^	 xposixpunct}', "");
Error('\p{	_Is_XPOSIXPUNCT:=}');
Error('\P{	_Is_XPOSIXPUNCT:=}');
Expect(1, 74867, '\p{isxposixpunct}', "");
Expect(0, 74867, '\p{^isxposixpunct}', "");
Expect(0, 74867, '\P{isxposixpunct}', "");
Expect(1, 74867, '\P{^isxposixpunct}', "");
Expect(0, 74868, '\p{isxposixpunct}', "");
Expect(1, 74868, '\p{^isxposixpunct}', "");
Expect(1, 74868, '\P{isxposixpunct}', "");
Expect(0, 74868, '\P{^isxposixpunct}', "");
Expect(1, 74867, '\p{_-Is_XPOSIXPUNCT}', "");
Expect(0, 74867, '\p{^_-Is_XPOSIXPUNCT}', "");
Expect(0, 74867, '\P{_-Is_XPOSIXPUNCT}', "");
Expect(1, 74867, '\P{^_-Is_XPOSIXPUNCT}', "");
Expect(0, 74868, '\p{_-Is_XPOSIXPUNCT}', "");
Expect(1, 74868, '\p{^_-Is_XPOSIXPUNCT}', "");
Expect(1, 74868, '\P{_-Is_XPOSIXPUNCT}', "");
Expect(0, 74868, '\P{^_-Is_XPOSIXPUNCT}', "");
Error('\p{	:=In_ARMENIAN}');
Error('\P{	:=In_ARMENIAN}');
Expect(1, 1423, '\p{inarmenian}', "");
Expect(0, 1423, '\p{^inarmenian}', "");
Expect(0, 1423, '\P{inarmenian}', "");
Expect(1, 1423, '\P{^inarmenian}', "");
Expect(0, 1424, '\p{inarmenian}', "");
Expect(1, 1424, '\p{^inarmenian}', "");
Expect(1, 1424, '\P{inarmenian}', "");
Expect(0, 1424, '\P{^inarmenian}', "");
Expect(1, 1423, '\p{	-In_ARMENIAN}', "");
Expect(0, 1423, '\p{^	-In_ARMENIAN}', "");
Expect(0, 1423, '\P{	-In_ARMENIAN}', "");
Expect(1, 1423, '\P{^	-In_ARMENIAN}', "");
Expect(0, 1424, '\p{	-In_ARMENIAN}', "");
Expect(1, 1424, '\p{^	-In_ARMENIAN}', "");
Expect(1, 1424, '\P{	-In_ARMENIAN}', "");
Expect(0, 1424, '\P{^	-In_ARMENIAN}', "");
Error('\p{-/a/In_Armenian}');
Error('\P{-/a/In_Armenian}');
Expect(1, 1423, '\p{_In_Armenian}', "");
Expect(0, 1423, '\p{^_In_Armenian}', "");
Expect(0, 1423, '\P{_In_Armenian}', "");
Expect(1, 1423, '\P{^_In_Armenian}', "");
Expect(0, 1424, '\p{_In_Armenian}', "");
Expect(1, 1424, '\p{^_In_Armenian}', "");
Expect(1, 1424, '\P{_In_Armenian}', "");
Expect(0, 1424, '\P{^_In_Armenian}', "");
Error('\p{	-PosixAlpha:=}');
Error('\P{	-PosixAlpha:=}');
Expect(1, 122, '\p{posixalpha}', "");
Expect(0, 122, '\p{^posixalpha}', "");
Expect(0, 122, '\P{posixalpha}', "");
Expect(1, 122, '\P{^posixalpha}', "");
Expect(0, 123, '\p{posixalpha}', "");
Expect(1, 123, '\p{^posixalpha}', "");
Expect(1, 123, '\P{posixalpha}', "");
Expect(0, 123, '\P{^posixalpha}', "");
Expect(1, 122, '\p{_	POSIXALPHA}', "");
Expect(0, 122, '\p{^_	POSIXALPHA}', "");
Expect(0, 122, '\P{_	POSIXALPHA}', "");
Expect(1, 122, '\P{^_	POSIXALPHA}', "");
Expect(0, 123, '\p{_	POSIXALPHA}', "");
Expect(1, 123, '\p{^_	POSIXALPHA}', "");
Expect(1, 123, '\P{_	POSIXALPHA}', "");
Expect(0, 123, '\P{^_	POSIXALPHA}', "");
Error('\p{:=__Is_POSIXALPHA}');
Error('\P{:=__Is_POSIXALPHA}');
Expect(1, 122, '\p{isposixalpha}', "");
Expect(0, 122, '\p{^isposixalpha}', "");
Expect(0, 122, '\P{isposixalpha}', "");
Expect(1, 122, '\P{^isposixalpha}', "");
Expect(0, 123, '\p{isposixalpha}', "");
Expect(1, 123, '\p{^isposixalpha}', "");
Expect(1, 123, '\P{isposixalpha}', "");
Expect(0, 123, '\P{^isposixalpha}', "");
Expect(1, 122, '\p{_IS_PosixAlpha}', "");
Expect(0, 122, '\p{^_IS_PosixAlpha}', "");
Expect(0, 122, '\P{_IS_PosixAlpha}', "");
Expect(1, 122, '\P{^_IS_PosixAlpha}', "");
Expect(0, 123, '\p{_IS_PosixAlpha}', "");
Expect(1, 123, '\p{^_IS_PosixAlpha}', "");
Expect(1, 123, '\P{_IS_PosixAlpha}', "");
Expect(0, 123, '\P{^_IS_PosixAlpha}', "");
Error('\p{:=-	Paragraph_SEPARATOR}');
Error('\P{:=-	Paragraph_SEPARATOR}');
Expect(1, 8233, '\p{paragraphseparator}', "");
Expect(0, 8233, '\p{^paragraphseparator}', "");
Expect(0, 8233, '\P{paragraphseparator}', "");
Expect(1, 8233, '\P{^paragraphseparator}', "");
Expect(0, 8234, '\p{paragraphseparator}', "");
Expect(1, 8234, '\p{^paragraphseparator}', "");
Expect(1, 8234, '\P{paragraphseparator}', "");
Expect(0, 8234, '\P{^paragraphseparator}', "");
Expect(1, 8233, '\p{	Paragraph_separator}', "");
Expect(0, 8233, '\p{^	Paragraph_separator}', "");
Expect(0, 8233, '\P{	Paragraph_separator}', "");
Expect(1, 8233, '\P{^	Paragraph_separator}', "");
Expect(0, 8234, '\p{	Paragraph_separator}', "");
Expect(1, 8234, '\p{^	Paragraph_separator}', "");
Expect(1, 8234, '\P{	Paragraph_separator}', "");
Expect(0, 8234, '\P{^	Paragraph_separator}', "");
Error('\p{/a/	 Is_Paragraph_SEPARATOR}');
Error('\P{/a/	 Is_Paragraph_SEPARATOR}');
Expect(1, 8233, '\p{isparagraphseparator}', "");
Expect(0, 8233, '\p{^isparagraphseparator}', "");
Expect(0, 8233, '\P{isparagraphseparator}', "");
Expect(1, 8233, '\P{^isparagraphseparator}', "");
Expect(0, 8234, '\p{isparagraphseparator}', "");
Expect(1, 8234, '\p{^isparagraphseparator}', "");
Expect(1, 8234, '\P{isparagraphseparator}', "");
Expect(0, 8234, '\P{^isparagraphseparator}', "");
Expect(1, 8233, '\p{	_is_Paragraph_separator}', "");
Expect(0, 8233, '\p{^	_is_Paragraph_separator}', "");
Expect(0, 8233, '\P{	_is_Paragraph_separator}', "");
Expect(1, 8233, '\P{^	_is_Paragraph_separator}', "");
Expect(0, 8234, '\p{	_is_Paragraph_separator}', "");
Expect(1, 8234, '\p{^	_is_Paragraph_separator}', "");
Expect(1, 8234, '\P{	_is_Paragraph_separator}', "");
Expect(0, 8234, '\P{^	_is_Paragraph_separator}', "");
Error('\p{- ZP:=}');
Error('\P{- ZP:=}');
Expect(1, 8233, '\p{zp}', "");
Expect(0, 8233, '\p{^zp}', "");
Expect(0, 8233, '\P{zp}', "");
Expect(1, 8233, '\P{^zp}', "");
Expect(0, 8234, '\p{zp}', "");
Expect(1, 8234, '\p{^zp}', "");
Expect(1, 8234, '\P{zp}', "");
Expect(0, 8234, '\P{^zp}', "");
Expect(1, 8233, '\p{--zp}', "");
Expect(0, 8233, '\p{^--zp}', "");
Expect(0, 8233, '\P{--zp}', "");
Expect(1, 8233, '\P{^--zp}', "");
Expect(0, 8234, '\p{--zp}', "");
Expect(1, 8234, '\p{^--zp}', "");
Expect(1, 8234, '\P{--zp}', "");
Expect(0, 8234, '\P{^--zp}', "");
Error('\p{_-is_Zp:=}');
Error('\P{_-is_Zp:=}');
Expect(1, 8233, '\p{iszp}', "");
Expect(0, 8233, '\p{^iszp}', "");
Expect(0, 8233, '\P{iszp}', "");
Expect(1, 8233, '\P{^iszp}', "");
Expect(0, 8234, '\p{iszp}', "");
Expect(1, 8234, '\p{^iszp}', "");
Expect(1, 8234, '\P{iszp}', "");
Expect(0, 8234, '\P{^iszp}', "");
Expect(1, 8233, '\p{-Is_Zp}', "");
Expect(0, 8233, '\p{^-Is_Zp}', "");
Expect(0, 8233, '\P{-Is_Zp}', "");
Expect(1, 8233, '\P{^-Is_Zp}', "");
Expect(0, 8234, '\p{-Is_Zp}', "");
Expect(1, 8234, '\p{^-Is_Zp}', "");
Expect(1, 8234, '\P{-Is_Zp}', "");
Expect(0, 8234, '\P{^-Is_Zp}', "");
Error('\p{:=	Phags_PA}');
Error('\P{:=	Phags_PA}');
Expect(1, 43127, '\p{phagspa}', "");
Expect(0, 43127, '\p{^phagspa}', "");
Expect(0, 43127, '\P{phagspa}', "");
Expect(1, 43127, '\P{^phagspa}', "");
Expect(0, 43128, '\p{phagspa}', "");
Expect(1, 43128, '\p{^phagspa}', "");
Expect(1, 43128, '\P{phagspa}', "");
Expect(0, 43128, '\P{^phagspa}', "");
Expect(1, 43127, '\p{	Phags_PA}', "");
Expect(0, 43127, '\p{^	Phags_PA}', "");
Expect(0, 43127, '\P{	Phags_PA}', "");
Expect(1, 43127, '\P{^	Phags_PA}', "");
Expect(0, 43128, '\p{	Phags_PA}', "");
Expect(1, 43128, '\p{^	Phags_PA}', "");
Expect(1, 43128, '\P{	Phags_PA}', "");
Expect(0, 43128, '\P{^	Phags_PA}', "");
Error('\p{:=_Is_Phags_Pa}');
Error('\P{:=_Is_Phags_Pa}');
Expect(1, 43127, '\p{isphagspa}', "");
Expect(0, 43127, '\p{^isphagspa}', "");
Expect(0, 43127, '\P{isphagspa}', "");
Expect(1, 43127, '\P{^isphagspa}', "");
Expect(0, 43128, '\p{isphagspa}', "");
Expect(1, 43128, '\p{^isphagspa}', "");
Expect(1, 43128, '\P{isphagspa}', "");
Expect(0, 43128, '\P{^isphagspa}', "");
Expect(1, 43127, '\p{ _Is_Phags_PA}', "");
Expect(0, 43127, '\p{^ _Is_Phags_PA}', "");
Expect(0, 43127, '\P{ _Is_Phags_PA}', "");
Expect(1, 43127, '\P{^ _Is_Phags_PA}', "");
Expect(0, 43128, '\p{ _Is_Phags_PA}', "");
Expect(1, 43128, '\p{^ _Is_Phags_PA}', "");
Expect(1, 43128, '\P{ _Is_Phags_PA}', "");
Expect(0, 43128, '\P{^ _Is_Phags_PA}', "");
Error('\p{-_phag/a/}');
Error('\P{-_phag/a/}');
Expect(1, 43127, '\p{phag}', "");
Expect(0, 43127, '\p{^phag}', "");
Expect(0, 43127, '\P{phag}', "");
Expect(1, 43127, '\P{^phag}', "");
Expect(0, 43128, '\p{phag}', "");
Expect(1, 43128, '\p{^phag}', "");
Expect(1, 43128, '\P{phag}', "");
Expect(0, 43128, '\P{^phag}', "");
Expect(1, 43127, '\p{_phag}', "");
Expect(0, 43127, '\p{^_phag}', "");
Expect(0, 43127, '\P{_phag}', "");
Expect(1, 43127, '\P{^_phag}', "");
Expect(0, 43128, '\p{_phag}', "");
Expect(1, 43128, '\p{^_phag}', "");
Expect(1, 43128, '\P{_phag}', "");
Expect(0, 43128, '\P{^_phag}', "");
Error('\p{	:=Is_Phag}');
Error('\P{	:=Is_Phag}');
Expect(1, 43127, '\p{isphag}', "");
Expect(0, 43127, '\p{^isphag}', "");
Expect(0, 43127, '\P{isphag}', "");
Expect(1, 43127, '\P{^isphag}', "");
Expect(0, 43128, '\p{isphag}', "");
Expect(1, 43128, '\p{^isphag}', "");
Expect(1, 43128, '\P{isphag}', "");
Expect(0, 43128, '\P{^isphag}', "");
Expect(1, 43127, '\p{_	Is_phag}', "");
Expect(0, 43127, '\p{^_	Is_phag}', "");
Expect(0, 43127, '\P{_	Is_phag}', "");
Expect(1, 43127, '\P{^_	Is_phag}', "");
Expect(0, 43128, '\p{_	Is_phag}', "");
Expect(1, 43128, '\p{^_	Is_phag}', "");
Expect(1, 43128, '\P{_	Is_phag}', "");
Expect(0, 43128, '\P{^_	Is_phag}', "");
Error('\p{__letter_Number/a/}');
Error('\P{__letter_Number/a/}');
Expect(1, 74850, '\p{letternumber}', "");
Expect(0, 74850, '\p{^letternumber}', "");
Expect(0, 74850, '\P{letternumber}', "");
Expect(1, 74850, '\P{^letternumber}', "");
Expect(0, 74851, '\p{letternumber}', "");
Expect(1, 74851, '\p{^letternumber}', "");
Expect(1, 74851, '\P{letternumber}', "");
Expect(0, 74851, '\P{^letternumber}', "");
Expect(1, 74850, '\p{ Letter_Number}', "");
Expect(0, 74850, '\p{^ Letter_Number}', "");
Expect(0, 74850, '\P{ Letter_Number}', "");
Expect(1, 74850, '\P{^ Letter_Number}', "");
Expect(0, 74851, '\p{ Letter_Number}', "");
Expect(1, 74851, '\p{^ Letter_Number}', "");
Expect(1, 74851, '\P{ Letter_Number}', "");
Expect(0, 74851, '\P{^ Letter_Number}', "");
Error('\p{		IS_LETTER_Number/a/}');
Error('\P{		IS_LETTER_Number/a/}');
Expect(1, 74850, '\p{isletternumber}', "");
Expect(0, 74850, '\p{^isletternumber}', "");
Expect(0, 74850, '\P{isletternumber}', "");
Expect(1, 74850, '\P{^isletternumber}', "");
Expect(0, 74851, '\p{isletternumber}', "");
Expect(1, 74851, '\p{^isletternumber}', "");
Expect(1, 74851, '\P{isletternumber}', "");
Expect(0, 74851, '\P{^isletternumber}', "");
Expect(1, 74850, '\p{IS_Letter_Number}', "");
Expect(0, 74850, '\p{^IS_Letter_Number}', "");
Expect(0, 74850, '\P{IS_Letter_Number}', "");
Expect(1, 74850, '\P{^IS_Letter_Number}', "");
Expect(0, 74851, '\p{IS_Letter_Number}', "");
Expect(1, 74851, '\p{^IS_Letter_Number}', "");
Expect(1, 74851, '\P{IS_Letter_Number}', "");
Expect(0, 74851, '\P{^IS_Letter_Number}', "");
Error('\p{_	Nl:=}');
Error('\P{_	Nl:=}');
Expect(1, 74850, '\p{nl}', "");
Expect(0, 74850, '\p{^nl}', "");
Expect(0, 74850, '\P{nl}', "");
Expect(1, 74850, '\P{^nl}', "");
Expect(0, 74851, '\p{nl}', "");
Expect(1, 74851, '\p{^nl}', "");
Expect(1, 74851, '\P{nl}', "");
Expect(0, 74851, '\P{^nl}', "");
Expect(1, 74850, '\p{ Nl}', "");
Expect(0, 74850, '\p{^ Nl}', "");
Expect(0, 74850, '\P{ Nl}', "");
Expect(1, 74850, '\P{^ Nl}', "");
Expect(0, 74851, '\p{ Nl}', "");
Expect(1, 74851, '\p{^ Nl}', "");
Expect(1, 74851, '\P{ Nl}', "");
Expect(0, 74851, '\P{^ Nl}', "");
Error('\p{ :=is_Nl}');
Error('\P{ :=is_Nl}');
Expect(1, 74850, '\p{isnl}', "");
Expect(0, 74850, '\p{^isnl}', "");
Expect(0, 74850, '\P{isnl}', "");
Expect(1, 74850, '\P{^isnl}', "");
Expect(0, 74851, '\p{isnl}', "");
Expect(1, 74851, '\p{^isnl}', "");
Expect(1, 74851, '\P{isnl}', "");
Expect(0, 74851, '\P{^isnl}', "");
Expect(1, 74850, '\p{	is_Nl}', "");
Expect(0, 74850, '\p{^	is_Nl}', "");
Expect(0, 74850, '\P{	is_Nl}', "");
Expect(1, 74850, '\P{^	is_Nl}', "");
Expect(0, 74851, '\p{	is_Nl}', "");
Expect(1, 74851, '\p{^	is_Nl}', "");
Expect(1, 74851, '\P{	is_Nl}', "");
Expect(0, 74851, '\P{^	is_Nl}', "");
Error('\p{ 	Number_Forms:=}');
Error('\P{ 	Number_Forms:=}');
Expect(1, 8591, '\p{numberforms}', "");
Expect(0, 8591, '\p{^numberforms}', "");
Expect(0, 8591, '\P{numberforms}', "");
Expect(1, 8591, '\P{^numberforms}', "");
Expect(0, 8592, '\p{numberforms}', "");
Expect(1, 8592, '\p{^numberforms}', "");
Expect(1, 8592, '\P{numberforms}', "");
Expect(0, 8592, '\P{^numberforms}', "");
Expect(1, 8591, '\p{ NUMBER_FORMS}', "");
Expect(0, 8591, '\p{^ NUMBER_FORMS}', "");
Expect(0, 8591, '\P{ NUMBER_FORMS}', "");
Expect(1, 8591, '\P{^ NUMBER_FORMS}', "");
Expect(0, 8592, '\p{ NUMBER_FORMS}', "");
Expect(1, 8592, '\p{^ NUMBER_FORMS}', "");
Expect(1, 8592, '\P{ NUMBER_FORMS}', "");
Expect(0, 8592, '\P{^ NUMBER_FORMS}', "");
Error('\p{/a/IS_number_Forms}');
Error('\P{/a/IS_number_Forms}');
Expect(1, 8591, '\p{isnumberforms}', "");
Expect(0, 8591, '\p{^isnumberforms}', "");
Expect(0, 8591, '\P{isnumberforms}', "");
Expect(1, 8591, '\P{^isnumberforms}', "");
Expect(0, 8592, '\p{isnumberforms}', "");
Expect(1, 8592, '\p{^isnumberforms}', "");
Expect(1, 8592, '\P{isnumberforms}', "");
Expect(0, 8592, '\P{^isnumberforms}', "");
Expect(1, 8591, '\p{ -Is_Number_Forms}', "");
Expect(0, 8591, '\p{^ -Is_Number_Forms}', "");
Expect(0, 8591, '\P{ -Is_Number_Forms}', "");
Expect(1, 8591, '\P{^ -Is_Number_Forms}', "");
Expect(0, 8592, '\p{ -Is_Number_Forms}', "");
Expect(1, 8592, '\p{^ -Is_Number_Forms}', "");
Expect(1, 8592, '\P{ -Is_Number_Forms}', "");
Expect(0, 8592, '\P{^ -Is_Number_Forms}', "");
Error('\p{ :=In_Number_FORMS}');
Error('\P{ :=In_Number_FORMS}');
Expect(1, 8591, '\p{innumberforms}', "");
Expect(0, 8591, '\p{^innumberforms}', "");
Expect(0, 8591, '\P{innumberforms}', "");
Expect(1, 8591, '\P{^innumberforms}', "");
Expect(0, 8592, '\p{innumberforms}', "");
Expect(1, 8592, '\p{^innumberforms}', "");
Expect(1, 8592, '\P{innumberforms}', "");
Expect(0, 8592, '\P{^innumberforms}', "");
Expect(1, 8591, '\p{	 In_number_Forms}', "");
Expect(0, 8591, '\p{^	 In_number_Forms}', "");
Expect(0, 8591, '\P{	 In_number_Forms}', "");
Expect(1, 8591, '\P{^	 In_number_Forms}', "");
Expect(0, 8592, '\p{	 In_number_Forms}', "");
Expect(1, 8592, '\p{^	 In_number_Forms}', "");
Expect(1, 8592, '\P{	 In_number_Forms}', "");
Expect(0, 8592, '\P{^	 In_number_Forms}', "");
Error('\p{-:=alnum}');
Error('\P{-:=alnum}');
Expect(1, 195101, '\p{alnum}', "");
Expect(0, 195101, '\p{^alnum}', "");
Expect(0, 195101, '\P{alnum}', "");
Expect(1, 195101, '\P{^alnum}', "");
Expect(0, 195102, '\p{alnum}', "");
Expect(1, 195102, '\p{^alnum}', "");
Expect(1, 195102, '\P{alnum}', "");
Expect(0, 195102, '\P{^alnum}', "");
Expect(1, 195101, '\p{ Alnum}', "");
Expect(0, 195101, '\p{^ Alnum}', "");
Expect(0, 195101, '\P{ Alnum}', "");
Expect(1, 195101, '\P{^ Alnum}', "");
Expect(0, 195102, '\p{ Alnum}', "");
Expect(1, 195102, '\p{^ Alnum}', "");
Expect(1, 195102, '\P{ Alnum}', "");
Expect(0, 195102, '\P{^ Alnum}', "");
Error('\p{_ XPosixAlnum:=}');
Error('\P{_ XPosixAlnum:=}');
Expect(1, 195101, '\p{xposixalnum}', "");
Expect(0, 195101, '\p{^xposixalnum}', "");
Expect(0, 195101, '\P{xposixalnum}', "");
Expect(1, 195101, '\P{^xposixalnum}', "");
Expect(0, 195102, '\p{xposixalnum}', "");
Expect(1, 195102, '\p{^xposixalnum}', "");
Expect(1, 195102, '\P{xposixalnum}', "");
Expect(0, 195102, '\P{^xposixalnum}', "");
Expect(1, 195101, '\p{	XPosixAlnum}', "");
Expect(0, 195101, '\p{^	XPosixAlnum}', "");
Expect(0, 195101, '\P{	XPosixAlnum}', "");
Expect(1, 195101, '\P{^	XPosixAlnum}', "");
Expect(0, 195102, '\p{	XPosixAlnum}', "");
Expect(1, 195102, '\p{^	XPosixAlnum}', "");
Expect(1, 195102, '\P{	XPosixAlnum}', "");
Expect(0, 195102, '\P{^	XPosixAlnum}', "");
Error('\p{ -IS_ALNUM:=}');
Error('\P{ -IS_ALNUM:=}');
Expect(1, 195101, '\p{isalnum}', "");
Expect(0, 195101, '\p{^isalnum}', "");
Expect(0, 195101, '\P{isalnum}', "");
Expect(1, 195101, '\P{^isalnum}', "");
Expect(0, 195102, '\p{isalnum}', "");
Expect(1, 195102, '\p{^isalnum}', "");
Expect(1, 195102, '\P{isalnum}', "");
Expect(0, 195102, '\P{^isalnum}', "");
Expect(1, 195101, '\p{	 is_Alnum}', "");
Expect(0, 195101, '\p{^	 is_Alnum}', "");
Expect(0, 195101, '\P{	 is_Alnum}', "");
Expect(1, 195101, '\P{^	 is_Alnum}', "");
Expect(0, 195102, '\p{	 is_Alnum}', "");
Expect(1, 195102, '\p{^	 is_Alnum}', "");
Expect(1, 195102, '\P{	 is_Alnum}', "");
Expect(0, 195102, '\P{^	 is_Alnum}', "");
Error('\p{:=_ is_XPOSIXALNUM}');
Error('\P{:=_ is_XPOSIXALNUM}');
Expect(1, 195101, '\p{isxposixalnum}', "");
Expect(0, 195101, '\p{^isxposixalnum}', "");
Expect(0, 195101, '\P{isxposixalnum}', "");
Expect(1, 195101, '\P{^isxposixalnum}', "");
Expect(0, 195102, '\p{isxposixalnum}', "");
Expect(1, 195102, '\p{^isxposixalnum}', "");
Expect(1, 195102, '\P{isxposixalnum}', "");
Expect(0, 195102, '\P{^isxposixalnum}', "");
Expect(1, 195101, '\p{ Is_xposixalnum}', "");
Expect(0, 195101, '\p{^ Is_xposixalnum}', "");
Expect(0, 195101, '\P{ Is_xposixalnum}', "");
Expect(1, 195101, '\P{^ Is_xposixalnum}', "");
Expect(0, 195102, '\p{ Is_xposixalnum}', "");
Expect(1, 195102, '\p{^ Is_xposixalnum}', "");
Expect(1, 195102, '\P{ Is_xposixalnum}', "");
Expect(0, 195102, '\P{^ Is_xposixalnum}', "");
Error('\p{_ in_Sundanese:=}');
Error('\P{_ in_Sundanese:=}');
Expect(1, 7103, '\p{insundanese}', "");
Expect(0, 7103, '\p{^insundanese}', "");
Expect(0, 7103, '\P{insundanese}', "");
Expect(1, 7103, '\P{^insundanese}', "");
Expect(0, 7104, '\p{insundanese}', "");
Expect(1, 7104, '\p{^insundanese}', "");
Expect(1, 7104, '\P{insundanese}', "");
Expect(0, 7104, '\P{^insundanese}', "");
Expect(1, 7103, '\p{_-IN_Sundanese}', "");
Expect(0, 7103, '\p{^_-IN_Sundanese}', "");
Expect(0, 7103, '\P{_-IN_Sundanese}', "");
Expect(1, 7103, '\P{^_-IN_Sundanese}', "");
Expect(0, 7104, '\p{_-IN_Sundanese}', "");
Expect(1, 7104, '\p{^_-IN_Sundanese}', "");
Expect(1, 7104, '\P{_-IN_Sundanese}', "");
Expect(0, 7104, '\P{^_-IN_Sundanese}', "");
Error('\p{		In_Sundanese/a/}');
Error('\P{		In_Sundanese/a/}');
Expect(1, 7103, '\p{--in_SUNDANESE}', "");
Expect(0, 7103, '\p{^--in_SUNDANESE}', "");
Expect(0, 7103, '\P{--in_SUNDANESE}', "");
Expect(1, 7103, '\P{^--in_SUNDANESE}', "");
Expect(0, 7104, '\p{--in_SUNDANESE}', "");
Expect(1, 7104, '\p{^--in_SUNDANESE}', "");
Expect(1, 7104, '\P{--in_SUNDANESE}', "");
Expect(0, 7104, '\P{^--in_SUNDANESE}', "");
Error('\p{	:=COMMON_Indic_NUMBER_forms}');
Error('\P{	:=COMMON_Indic_NUMBER_forms}');
Expect(1, 43071, '\p{commonindicnumberforms}', "");
Expect(0, 43071, '\p{^commonindicnumberforms}', "");
Expect(0, 43071, '\P{commonindicnumberforms}', "");
Expect(1, 43071, '\P{^commonindicnumberforms}', "");
Expect(0, 43072, '\p{commonindicnumberforms}', "");
Expect(1, 43072, '\p{^commonindicnumberforms}', "");
Expect(1, 43072, '\P{commonindicnumberforms}', "");
Expect(0, 43072, '\P{^commonindicnumberforms}', "");
Expect(1, 43071, '\p{	-common_INDIC_Number_forms}', "");
Expect(0, 43071, '\p{^	-common_INDIC_Number_forms}', "");
Expect(0, 43071, '\P{	-common_INDIC_Number_forms}', "");
Expect(1, 43071, '\P{^	-common_INDIC_Number_forms}', "");
Expect(0, 43072, '\p{	-common_INDIC_Number_forms}', "");
Expect(1, 43072, '\p{^	-common_INDIC_Number_forms}', "");
Expect(1, 43072, '\P{	-common_INDIC_Number_forms}', "");
Expect(0, 43072, '\P{^	-common_INDIC_Number_forms}', "");
Error('\p{ /a/Is_Common_Indic_Number_forms}');
Error('\P{ /a/Is_Common_Indic_Number_forms}');
Expect(1, 43071, '\p{iscommonindicnumberforms}', "");
Expect(0, 43071, '\p{^iscommonindicnumberforms}', "");
Expect(0, 43071, '\P{iscommonindicnumberforms}', "");
Expect(1, 43071, '\P{^iscommonindicnumberforms}', "");
Expect(0, 43072, '\p{iscommonindicnumberforms}', "");
Expect(1, 43072, '\p{^iscommonindicnumberforms}', "");
Expect(1, 43072, '\P{iscommonindicnumberforms}', "");
Expect(0, 43072, '\P{^iscommonindicnumberforms}', "");
Expect(1, 43071, '\p{ Is_common_Indic_Number_FORMS}', "");
Expect(0, 43071, '\p{^ Is_common_Indic_Number_FORMS}', "");
Expect(0, 43071, '\P{ Is_common_Indic_Number_FORMS}', "");
Expect(1, 43071, '\P{^ Is_common_Indic_Number_FORMS}', "");
Expect(0, 43072, '\p{ Is_common_Indic_Number_FORMS}', "");
Expect(1, 43072, '\p{^ Is_common_Indic_Number_FORMS}', "");
Expect(1, 43072, '\P{ Is_common_Indic_Number_FORMS}', "");
Expect(0, 43072, '\P{^ Is_common_Indic_Number_FORMS}', "");
Error('\p{	_in_COMMON_Indic_NUMBER_Forms/a/}');
Error('\P{	_in_COMMON_Indic_NUMBER_Forms/a/}');
Expect(1, 43071, '\p{incommonindicnumberforms}', "");
Expect(0, 43071, '\p{^incommonindicnumberforms}', "");
Expect(0, 43071, '\P{incommonindicnumberforms}', "");
Expect(1, 43071, '\P{^incommonindicnumberforms}', "");
Expect(0, 43072, '\p{incommonindicnumberforms}', "");
Expect(1, 43072, '\p{^incommonindicnumberforms}', "");
Expect(1, 43072, '\P{incommonindicnumberforms}', "");
Expect(0, 43072, '\P{^incommonindicnumberforms}', "");
Expect(1, 43071, '\p{_In_Common_Indic_number_FORMS}', "");
Expect(0, 43071, '\p{^_In_Common_Indic_number_FORMS}', "");
Expect(0, 43071, '\P{_In_Common_Indic_number_FORMS}', "");
Expect(1, 43071, '\P{^_In_Common_Indic_number_FORMS}', "");
Expect(0, 43072, '\p{_In_Common_Indic_number_FORMS}', "");
Expect(1, 43072, '\p{^_In_Common_Indic_number_FORMS}', "");
Expect(1, 43072, '\P{_In_Common_Indic_number_FORMS}', "");
Expect(0, 43072, '\P{^_In_Common_Indic_number_FORMS}', "");
Error('\p{_Indic_Number_Forms:=}');
Error('\P{_Indic_Number_Forms:=}');
Expect(1, 43071, '\p{indicnumberforms}', "");
Expect(0, 43071, '\p{^indicnumberforms}', "");
Expect(0, 43071, '\P{indicnumberforms}', "");
Expect(1, 43071, '\P{^indicnumberforms}', "");
Expect(0, 43072, '\p{indicnumberforms}', "");
Expect(1, 43072, '\p{^indicnumberforms}', "");
Expect(1, 43072, '\P{indicnumberforms}', "");
Expect(0, 43072, '\P{^indicnumberforms}', "");
Expect(1, 43071, '\p{-_INDIC_Number_FORMS}', "");
Expect(0, 43071, '\p{^-_INDIC_Number_FORMS}', "");
Expect(0, 43071, '\P{-_INDIC_Number_FORMS}', "");
Expect(1, 43071, '\P{^-_INDIC_Number_FORMS}', "");
Expect(0, 43072, '\p{-_INDIC_Number_FORMS}', "");
Expect(1, 43072, '\p{^-_INDIC_Number_FORMS}', "");
Expect(1, 43072, '\P{-_INDIC_Number_FORMS}', "");
Expect(0, 43072, '\P{^-_INDIC_Number_FORMS}', "");
Error('\p{	-IS_indic_number_Forms/a/}');
Error('\P{	-IS_indic_number_Forms/a/}');
Expect(1, 43071, '\p{isindicnumberforms}', "");
Expect(0, 43071, '\p{^isindicnumberforms}', "");
Expect(0, 43071, '\P{isindicnumberforms}', "");
Expect(1, 43071, '\P{^isindicnumberforms}', "");
Expect(0, 43072, '\p{isindicnumberforms}', "");
Expect(1, 43072, '\p{^isindicnumberforms}', "");
Expect(1, 43072, '\P{isindicnumberforms}', "");
Expect(0, 43072, '\P{^isindicnumberforms}', "");
Expect(1, 43071, '\p{ Is_indic_number_Forms}', "");
Expect(0, 43071, '\p{^ Is_indic_number_Forms}', "");
Expect(0, 43071, '\P{ Is_indic_number_Forms}', "");
Expect(1, 43071, '\P{^ Is_indic_number_Forms}', "");
Expect(0, 43072, '\p{ Is_indic_number_Forms}', "");
Expect(1, 43072, '\p{^ Is_indic_number_Forms}', "");
Expect(1, 43072, '\P{ Is_indic_number_Forms}', "");
Expect(0, 43072, '\P{^ Is_indic_number_Forms}', "");
Error('\p{/a/--IN_indic_NUMBER_Forms}');
Error('\P{/a/--IN_indic_NUMBER_Forms}');
Expect(1, 43071, '\p{inindicnumberforms}', "");
Expect(0, 43071, '\p{^inindicnumberforms}', "");
Expect(0, 43071, '\P{inindicnumberforms}', "");
Expect(1, 43071, '\P{^inindicnumberforms}', "");
Expect(0, 43072, '\p{inindicnumberforms}', "");
Expect(1, 43072, '\p{^inindicnumberforms}', "");
Expect(1, 43072, '\P{inindicnumberforms}', "");
Expect(0, 43072, '\P{^inindicnumberforms}', "");
Expect(1, 43071, '\p{--IN_Indic_NUMBER_FORMS}', "");
Expect(0, 43071, '\p{^--IN_Indic_NUMBER_FORMS}', "");
Expect(0, 43071, '\P{--IN_Indic_NUMBER_FORMS}', "");
Expect(1, 43071, '\P{^--IN_Indic_NUMBER_FORMS}', "");
Expect(0, 43072, '\p{--IN_Indic_NUMBER_FORMS}', "");
Expect(1, 43072, '\p{^--IN_Indic_NUMBER_FORMS}', "");
Expect(1, 43072, '\P{--IN_Indic_NUMBER_FORMS}', "");
Expect(0, 43072, '\P{^--IN_Indic_NUMBER_FORMS}', "");
Error('\p{ /a/Alpha}');
Error('\P{ /a/Alpha}');
Expect(1, 195101, '\p{alpha}', "");
Expect(0, 195101, '\p{^alpha}', "");
Expect(0, 195101, '\P{alpha}', "");
Expect(1, 195101, '\P{^alpha}', "");
Expect(0, 195102, '\p{alpha}', "");
Expect(1, 195102, '\p{^alpha}', "");
Expect(1, 195102, '\P{alpha}', "");
Expect(0, 195102, '\P{^alpha}', "");
Expect(1, 195101, '\p{_Alpha}', "");
Expect(0, 195101, '\p{^_Alpha}', "");
Expect(0, 195101, '\P{_Alpha}', "");
Expect(1, 195101, '\P{^_Alpha}', "");
Expect(0, 195102, '\p{_Alpha}', "");
Expect(1, 195102, '\p{^_Alpha}', "");
Expect(1, 195102, '\P{_Alpha}', "");
Expect(0, 195102, '\P{^_Alpha}', "");
Error('\p{/a/ XPOSIXALPHA}');
Error('\P{/a/ XPOSIXALPHA}');
Expect(1, 195101, '\p{xposixalpha}', "");
Expect(0, 195101, '\p{^xposixalpha}', "");
Expect(0, 195101, '\P{xposixalpha}', "");
Expect(1, 195101, '\P{^xposixalpha}', "");
Expect(0, 195102, '\p{xposixalpha}', "");
Expect(1, 195102, '\p{^xposixalpha}', "");
Expect(1, 195102, '\P{xposixalpha}', "");
Expect(0, 195102, '\P{^xposixalpha}', "");
Expect(1, 195101, '\p{_xposixalpha}', "");
Expect(0, 195101, '\p{^_xposixalpha}', "");
Expect(0, 195101, '\P{_xposixalpha}', "");
Expect(1, 195101, '\P{^_xposixalpha}', "");
Expect(0, 195102, '\p{_xposixalpha}', "");
Expect(1, 195102, '\p{^_xposixalpha}', "");
Expect(1, 195102, '\P{_xposixalpha}', "");
Expect(0, 195102, '\P{^_xposixalpha}', "");
Error('\p{/a/-is_alpha}');
Error('\P{/a/-is_alpha}');
Expect(1, 195101, '\p{isalpha}', "");
Expect(0, 195101, '\p{^isalpha}', "");
Expect(0, 195101, '\P{isalpha}', "");
Expect(1, 195101, '\P{^isalpha}', "");
Expect(0, 195102, '\p{isalpha}', "");
Expect(1, 195102, '\p{^isalpha}', "");
Expect(1, 195102, '\P{isalpha}', "");
Expect(0, 195102, '\P{^isalpha}', "");
Expect(1, 195101, '\p{-_Is_Alpha}', "");
Expect(0, 195101, '\p{^-_Is_Alpha}', "");
Expect(0, 195101, '\P{-_Is_Alpha}', "");
Expect(1, 195101, '\P{^-_Is_Alpha}', "");
Expect(0, 195102, '\p{-_Is_Alpha}', "");
Expect(1, 195102, '\p{^-_Is_Alpha}', "");
Expect(1, 195102, '\P{-_Is_Alpha}', "");
Expect(0, 195102, '\P{^-_Is_Alpha}', "");
Error('\p{/a/ -Is_XPosixAlpha}');
Error('\P{/a/ -Is_XPosixAlpha}');
Expect(1, 195101, '\p{isxposixalpha}', "");
Expect(0, 195101, '\p{^isxposixalpha}', "");
Expect(0, 195101, '\P{isxposixalpha}', "");
Expect(1, 195101, '\P{^isxposixalpha}', "");
Expect(0, 195102, '\p{isxposixalpha}', "");
Expect(1, 195102, '\p{^isxposixalpha}', "");
Expect(1, 195102, '\P{isxposixalpha}', "");
Expect(0, 195102, '\P{^isxposixalpha}', "");
Expect(1, 195101, '\p{	-Is_XPosixAlpha}', "");
Expect(0, 195101, '\p{^	-Is_XPosixAlpha}', "");
Expect(0, 195101, '\P{	-Is_XPosixAlpha}', "");
Expect(1, 195101, '\P{^	-Is_XPosixAlpha}', "");
Expect(0, 195102, '\p{	-Is_XPosixAlpha}', "");
Expect(1, 195102, '\p{^	-Is_XPosixAlpha}', "");
Expect(1, 195102, '\P{	-Is_XPosixAlpha}', "");
Expect(0, 195102, '\P{^	-Is_XPosixAlpha}', "");
Error('\p{_/a/Alphabetic}');
Error('\P{_/a/Alphabetic}');
Expect(1, 195101, '\p{alphabetic}', "");
Expect(0, 195101, '\p{^alphabetic}', "");
Expect(0, 195101, '\P{alphabetic}', "");
Expect(1, 195101, '\P{^alphabetic}', "");
Expect(0, 195102, '\p{alphabetic}', "");
Expect(1, 195102, '\p{^alphabetic}', "");
Expect(1, 195102, '\P{alphabetic}', "");
Expect(0, 195102, '\P{^alphabetic}', "");
Expect(1, 195101, '\p{_-Alphabetic}', "");
Expect(0, 195101, '\p{^_-Alphabetic}', "");
Expect(0, 195101, '\P{_-Alphabetic}', "");
Expect(1, 195101, '\P{^_-Alphabetic}', "");
Expect(0, 195102, '\p{_-Alphabetic}', "");
Expect(1, 195102, '\p{^_-Alphabetic}', "");
Expect(1, 195102, '\P{_-Alphabetic}', "");
Expect(0, 195102, '\P{^_-Alphabetic}', "");
Error('\p{/a/ Is_ALPHABETIC}');
Error('\P{/a/ Is_ALPHABETIC}');
Expect(1, 195101, '\p{isalphabetic}', "");
Expect(0, 195101, '\p{^isalphabetic}', "");
Expect(0, 195101, '\P{isalphabetic}', "");
Expect(1, 195101, '\P{^isalphabetic}', "");
Expect(0, 195102, '\p{isalphabetic}', "");
Expect(1, 195102, '\p{^isalphabetic}', "");
Expect(1, 195102, '\P{isalphabetic}', "");
Expect(0, 195102, '\P{^isalphabetic}', "");
Expect(1, 195101, '\p{__Is_alphabetic}', "");
Expect(0, 195101, '\p{^__Is_alphabetic}', "");
Expect(0, 195101, '\P{__Is_alphabetic}', "");
Expect(1, 195101, '\P{^__Is_alphabetic}', "");
Expect(0, 195102, '\p{__Is_alphabetic}', "");
Expect(1, 195102, '\p{^__Is_alphabetic}', "");
Expect(1, 195102, '\P{__Is_alphabetic}', "");
Expect(0, 195102, '\P{^__Is_alphabetic}', "");
Error('\p{_takri:=}');
Error('\P{_takri:=}');
Expect(1, 71369, '\p{takri}', "");
Expect(0, 71369, '\p{^takri}', "");
Expect(0, 71369, '\P{takri}', "");
Expect(1, 71369, '\P{^takri}', "");
Expect(0, 71370, '\p{takri}', "");
Expect(1, 71370, '\p{^takri}', "");
Expect(1, 71370, '\P{takri}', "");
Expect(0, 71370, '\P{^takri}', "");
Expect(1, 71369, '\p{ Takri}', "");
Expect(0, 71369, '\p{^ Takri}', "");
Expect(0, 71369, '\P{ Takri}', "");
Expect(1, 71369, '\P{^ Takri}', "");
Expect(0, 71370, '\p{ Takri}', "");
Expect(1, 71370, '\p{^ Takri}', "");
Expect(1, 71370, '\P{ Takri}', "");
Expect(0, 71370, '\P{^ Takri}', "");
Error('\p{/a/Is_Takri}');
Error('\P{/a/Is_Takri}');
Expect(1, 71369, '\p{istakri}', "");
Expect(0, 71369, '\p{^istakri}', "");
Expect(0, 71369, '\P{istakri}', "");
Expect(1, 71369, '\P{^istakri}', "");
Expect(0, 71370, '\p{istakri}', "");
Expect(1, 71370, '\p{^istakri}', "");
Expect(1, 71370, '\P{istakri}', "");
Expect(0, 71370, '\P{^istakri}', "");
Expect(1, 71369, '\p{_IS_Takri}', "");
Expect(0, 71369, '\p{^_IS_Takri}', "");
Expect(0, 71369, '\P{_IS_Takri}', "");
Expect(1, 71369, '\P{^_IS_Takri}', "");
Expect(0, 71370, '\p{_IS_Takri}', "");
Expect(1, 71370, '\p{^_IS_Takri}', "");
Expect(1, 71370, '\P{_IS_Takri}', "");
Expect(0, 71370, '\P{^_IS_Takri}', "");
Error('\p{/a/	TAKR}');
Error('\P{/a/	TAKR}');
Expect(1, 71369, '\p{takr}', "");
Expect(0, 71369, '\p{^takr}', "");
Expect(0, 71369, '\P{takr}', "");
Expect(1, 71369, '\P{^takr}', "");
Expect(0, 71370, '\p{takr}', "");
Expect(1, 71370, '\p{^takr}', "");
Expect(1, 71370, '\P{takr}', "");
Expect(0, 71370, '\P{^takr}', "");
Expect(1, 71369, '\p{ takr}', "");
Expect(0, 71369, '\p{^ takr}', "");
Expect(0, 71369, '\P{ takr}', "");
Expect(1, 71369, '\P{^ takr}', "");
Expect(0, 71370, '\p{ takr}', "");
Expect(1, 71370, '\p{^ takr}', "");
Expect(1, 71370, '\P{ takr}', "");
Expect(0, 71370, '\P{^ takr}', "");
Error('\p{/a/_is_takr}');
Error('\P{/a/_is_takr}');
Expect(1, 71369, '\p{istakr}', "");
Expect(0, 71369, '\p{^istakr}', "");
Expect(0, 71369, '\P{istakr}', "");
Expect(1, 71369, '\P{^istakr}', "");
Expect(0, 71370, '\p{istakr}', "");
Expect(1, 71370, '\p{^istakr}', "");
Expect(1, 71370, '\P{istakr}', "");
Expect(0, 71370, '\P{^istakr}', "");
Expect(1, 71369, '\p{ -Is_TAKR}', "");
Expect(0, 71369, '\p{^ -Is_TAKR}', "");
Expect(0, 71369, '\P{ -Is_TAKR}', "");
Expect(1, 71369, '\P{^ -Is_TAKR}', "");
Expect(0, 71370, '\p{ -Is_TAKR}', "");
Expect(1, 71370, '\p{^ -Is_TAKR}', "");
Expect(1, 71370, '\P{ -Is_TAKR}', "");
Expect(0, 71370, '\P{^ -Is_TAKR}', "");
Error('\p{-Playing_Cards:=}');
Error('\P{-Playing_Cards:=}');
Expect(1, 127231, '\p{playingcards}', "");
Expect(0, 127231, '\p{^playingcards}', "");
Expect(0, 127231, '\P{playingcards}', "");
Expect(1, 127231, '\P{^playingcards}', "");
Expect(0, 127232, '\p{playingcards}', "");
Expect(1, 127232, '\p{^playingcards}', "");
Expect(1, 127232, '\P{playingcards}', "");
Expect(0, 127232, '\P{^playingcards}', "");
Expect(1, 127231, '\p{ _PLAYING_CARDS}', "");
Expect(0, 127231, '\p{^ _PLAYING_CARDS}', "");
Expect(0, 127231, '\P{ _PLAYING_CARDS}', "");
Expect(1, 127231, '\P{^ _PLAYING_CARDS}', "");
Expect(0, 127232, '\p{ _PLAYING_CARDS}', "");
Expect(1, 127232, '\p{^ _PLAYING_CARDS}', "");
Expect(1, 127232, '\P{ _PLAYING_CARDS}', "");
Expect(0, 127232, '\P{^ _PLAYING_CARDS}', "");
Error('\p{- IS_Playing_Cards/a/}');
Error('\P{- IS_Playing_Cards/a/}');
Expect(1, 127231, '\p{isplayingcards}', "");
Expect(0, 127231, '\p{^isplayingcards}', "");
Expect(0, 127231, '\P{isplayingcards}', "");
Expect(1, 127231, '\P{^isplayingcards}', "");
Expect(0, 127232, '\p{isplayingcards}', "");
Expect(1, 127232, '\p{^isplayingcards}', "");
Expect(1, 127232, '\P{isplayingcards}', "");
Expect(0, 127232, '\P{^isplayingcards}', "");
Expect(1, 127231, '\p{	-is_Playing_Cards}', "");
Expect(0, 127231, '\p{^	-is_Playing_Cards}', "");
Expect(0, 127231, '\P{	-is_Playing_Cards}', "");
Expect(1, 127231, '\P{^	-is_Playing_Cards}', "");
Expect(0, 127232, '\p{	-is_Playing_Cards}', "");
Expect(1, 127232, '\p{^	-is_Playing_Cards}', "");
Expect(1, 127232, '\P{	-is_Playing_Cards}', "");
Expect(0, 127232, '\P{^	-is_Playing_Cards}', "");
Error('\p{ /a/in_Playing_Cards}');
Error('\P{ /a/in_Playing_Cards}');
Expect(1, 127231, '\p{inplayingcards}', "");
Expect(0, 127231, '\p{^inplayingcards}', "");
Expect(0, 127231, '\P{inplayingcards}', "");
Expect(1, 127231, '\P{^inplayingcards}', "");
Expect(0, 127232, '\p{inplayingcards}', "");
Expect(1, 127232, '\p{^inplayingcards}', "");
Expect(1, 127232, '\P{inplayingcards}', "");
Expect(0, 127232, '\P{^inplayingcards}', "");
Expect(1, 127231, '\p{_In_playing_cards}', "");
Expect(0, 127231, '\p{^_In_playing_cards}', "");
Expect(0, 127231, '\P{_In_playing_cards}', "");
Expect(1, 127231, '\P{^_In_playing_cards}', "");
Expect(0, 127232, '\p{_In_playing_cards}', "");
Expect(1, 127232, '\p{^_In_playing_cards}', "");
Expect(1, 127232, '\P{_In_playing_cards}', "");
Expect(0, 127232, '\P{^_In_playing_cards}', "");
Error('\p{  In_Egyptian_Hieroglyphs/a/}');
Error('\P{  In_Egyptian_Hieroglyphs/a/}');
Expect(1, 78895, '\p{inegyptianhieroglyphs}', "");
Expect(0, 78895, '\p{^inegyptianhieroglyphs}', "");
Expect(0, 78895, '\P{inegyptianhieroglyphs}', "");
Expect(1, 78895, '\P{^inegyptianhieroglyphs}', "");
Expect(0, 78896, '\p{inegyptianhieroglyphs}', "");
Expect(1, 78896, '\p{^inegyptianhieroglyphs}', "");
Expect(1, 78896, '\P{inegyptianhieroglyphs}', "");
Expect(0, 78896, '\P{^inegyptianhieroglyphs}', "");
Expect(1, 78895, '\p{ IN_Egyptian_Hieroglyphs}', "");
Expect(0, 78895, '\p{^ IN_Egyptian_Hieroglyphs}', "");
Expect(0, 78895, '\P{ IN_Egyptian_Hieroglyphs}', "");
Expect(1, 78895, '\P{^ IN_Egyptian_Hieroglyphs}', "");
Expect(0, 78896, '\p{ IN_Egyptian_Hieroglyphs}', "");
Expect(1, 78896, '\p{^ IN_Egyptian_Hieroglyphs}', "");
Expect(1, 78896, '\P{ IN_Egyptian_Hieroglyphs}', "");
Expect(0, 78896, '\P{^ IN_Egyptian_Hieroglyphs}', "");
Error('\p{	-in_Egyptian_HIEROGLYPHS:=}');
Error('\P{	-in_Egyptian_HIEROGLYPHS:=}');
Expect(1, 78895, '\p{__in_EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78895, '\p{^__in_EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78895, '\P{__in_EGYPTIAN_Hieroglyphs}', "");
Expect(1, 78895, '\P{^__in_EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78896, '\p{__in_EGYPTIAN_Hieroglyphs}', "");
Expect(1, 78896, '\p{^__in_EGYPTIAN_Hieroglyphs}', "");
Expect(1, 78896, '\P{__in_EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78896, '\P{^__in_EGYPTIAN_Hieroglyphs}', "");
Error('\p{:= -IN_Myanmar}');
Error('\P{:= -IN_Myanmar}');
Expect(1, 4255, '\p{inmyanmar}', "");
Expect(0, 4255, '\p{^inmyanmar}', "");
Expect(0, 4255, '\P{inmyanmar}', "");
Expect(1, 4255, '\P{^inmyanmar}', "");
Expect(0, 4256, '\p{inmyanmar}', "");
Expect(1, 4256, '\p{^inmyanmar}', "");
Expect(1, 4256, '\P{inmyanmar}', "");
Expect(0, 4256, '\P{^inmyanmar}', "");
Expect(1, 4255, '\p{--In_Myanmar}', "");
Expect(0, 4255, '\p{^--In_Myanmar}', "");
Expect(0, 4255, '\P{--In_Myanmar}', "");
Expect(1, 4255, '\P{^--In_Myanmar}', "");
Expect(0, 4256, '\p{--In_Myanmar}', "");
Expect(1, 4256, '\p{^--In_Myanmar}', "");
Expect(1, 4256, '\P{--In_Myanmar}', "");
Expect(0, 4256, '\P{^--In_Myanmar}', "");
Error('\p{  in_Myanmar/a/}');
Error('\P{  in_Myanmar/a/}');
Expect(1, 4255, '\p{_ In_MYANMAR}', "");
Expect(0, 4255, '\p{^_ In_MYANMAR}', "");
Expect(0, 4255, '\P{_ In_MYANMAR}', "");
Expect(1, 4255, '\P{^_ In_MYANMAR}', "");
Expect(0, 4256, '\p{_ In_MYANMAR}', "");
Expect(1, 4256, '\p{^_ In_MYANMAR}', "");
Expect(1, 4256, '\P{_ In_MYANMAR}', "");
Expect(0, 4256, '\P{^_ In_MYANMAR}', "");
Error('\p{/a/-Radical}');
Error('\P{/a/-Radical}');
Expect(1, 12245, '\p{radical}', "");
Expect(0, 12245, '\p{^radical}', "");
Expect(0, 12245, '\P{radical}', "");
Expect(1, 12245, '\P{^radical}', "");
Expect(0, 12246, '\p{radical}', "");
Expect(1, 12246, '\p{^radical}', "");
Expect(1, 12246, '\P{radical}', "");
Expect(0, 12246, '\P{^radical}', "");
Expect(1, 12245, '\p{-	Radical}', "");
Expect(0, 12245, '\p{^-	Radical}', "");
Expect(0, 12245, '\P{-	Radical}', "");
Expect(1, 12245, '\P{^-	Radical}', "");
Expect(0, 12246, '\p{-	Radical}', "");
Expect(1, 12246, '\p{^-	Radical}', "");
Expect(1, 12246, '\P{-	Radical}', "");
Expect(0, 12246, '\P{^-	Radical}', "");
Error('\p{-:=IS_RADICAL}');
Error('\P{-:=IS_RADICAL}');
Expect(1, 12245, '\p{isradical}', "");
Expect(0, 12245, '\p{^isradical}', "");
Expect(0, 12245, '\P{isradical}', "");
Expect(1, 12245, '\P{^isradical}', "");
Expect(0, 12246, '\p{isradical}', "");
Expect(1, 12246, '\p{^isradical}', "");
Expect(1, 12246, '\P{isradical}', "");
Expect(0, 12246, '\P{^isradical}', "");
Expect(1, 12245, '\p{	Is_Radical}', "");
Expect(0, 12245, '\p{^	Is_Radical}', "");
Expect(0, 12245, '\P{	Is_Radical}', "");
Expect(1, 12245, '\P{^	Is_Radical}', "");
Expect(0, 12246, '\p{	Is_Radical}', "");
Expect(1, 12246, '\p{^	Is_Radical}', "");
Expect(1, 12246, '\P{	Is_Radical}', "");
Expect(0, 12246, '\P{^	Is_Radical}', "");
Error('\p{ :=Low_SURROGATES}');
Error('\P{ :=Low_SURROGATES}');
Expect(1, 57343, '\p{lowsurrogates}', "");
Expect(0, 57343, '\p{^lowsurrogates}', "");
Expect(0, 57343, '\P{lowsurrogates}', "");
Expect(1, 57343, '\P{^lowsurrogates}', "");
Expect(0, 57344, '\p{lowsurrogates}', "");
Expect(1, 57344, '\p{^lowsurrogates}', "");
Expect(1, 57344, '\P{lowsurrogates}', "");
Expect(0, 57344, '\P{^lowsurrogates}', "");
Expect(1, 57343, '\p{_	low_Surrogates}', "");
Expect(0, 57343, '\p{^_	low_Surrogates}', "");
Expect(0, 57343, '\P{_	low_Surrogates}', "");
Expect(1, 57343, '\P{^_	low_Surrogates}', "");
Expect(0, 57344, '\p{_	low_Surrogates}', "");
Expect(1, 57344, '\p{^_	low_Surrogates}', "");
Expect(1, 57344, '\P{_	low_Surrogates}', "");
Expect(0, 57344, '\P{^_	low_Surrogates}', "");
Error('\p{		Is_Low_surrogates:=}');
Error('\P{		Is_Low_surrogates:=}');
Expect(1, 57343, '\p{islowsurrogates}', "");
Expect(0, 57343, '\p{^islowsurrogates}', "");
Expect(0, 57343, '\P{islowsurrogates}', "");
Expect(1, 57343, '\P{^islowsurrogates}', "");
Expect(0, 57344, '\p{islowsurrogates}', "");
Expect(1, 57344, '\p{^islowsurrogates}', "");
Expect(1, 57344, '\P{islowsurrogates}', "");
Expect(0, 57344, '\P{^islowsurrogates}', "");
Expect(1, 57343, '\p{	_Is_low_Surrogates}', "");
Expect(0, 57343, '\p{^	_Is_low_Surrogates}', "");
Expect(0, 57343, '\P{	_Is_low_Surrogates}', "");
Expect(1, 57343, '\P{^	_Is_low_Surrogates}', "");
Expect(0, 57344, '\p{	_Is_low_Surrogates}', "");
Expect(1, 57344, '\p{^	_Is_low_Surrogates}', "");
Expect(1, 57344, '\P{	_Is_low_Surrogates}', "");
Expect(0, 57344, '\P{^	_Is_low_Surrogates}', "");
Error('\p{-	In_Low_SURROGATES/a/}');
Error('\P{-	In_Low_SURROGATES/a/}');
Expect(1, 57343, '\p{inlowsurrogates}', "");
Expect(0, 57343, '\p{^inlowsurrogates}', "");
Expect(0, 57343, '\P{inlowsurrogates}', "");
Expect(1, 57343, '\P{^inlowsurrogates}', "");
Expect(0, 57344, '\p{inlowsurrogates}', "");
Expect(1, 57344, '\p{^inlowsurrogates}', "");
Expect(1, 57344, '\P{inlowsurrogates}', "");
Expect(0, 57344, '\P{^inlowsurrogates}', "");
Expect(1, 57343, '\p{- In_Low_SURROGATES}', "");
Expect(0, 57343, '\p{^- In_Low_SURROGATES}', "");
Expect(0, 57343, '\P{- In_Low_SURROGATES}', "");
Expect(1, 57343, '\P{^- In_Low_SURROGATES}', "");
Expect(0, 57344, '\p{- In_Low_SURROGATES}', "");
Expect(1, 57344, '\p{^- In_Low_SURROGATES}', "");
Expect(1, 57344, '\P{- In_Low_SURROGATES}', "");
Expect(0, 57344, '\P{^- In_Low_SURROGATES}', "");
Error('\p{	 In_PHAGS_Pa/a/}');
Error('\P{	 In_PHAGS_Pa/a/}');
Expect(1, 43135, '\p{inphagspa}', "");
Expect(0, 43135, '\p{^inphagspa}', "");
Expect(0, 43135, '\P{inphagspa}', "");
Expect(1, 43135, '\P{^inphagspa}', "");
Expect(0, 43136, '\p{inphagspa}', "");
Expect(1, 43136, '\p{^inphagspa}', "");
Expect(1, 43136, '\P{inphagspa}', "");
Expect(0, 43136, '\P{^inphagspa}', "");
Expect(1, 43135, '\p{	IN_Phags_Pa}', "");
Expect(0, 43135, '\p{^	IN_Phags_Pa}', "");
Expect(0, 43135, '\P{	IN_Phags_Pa}', "");
Expect(1, 43135, '\P{^	IN_Phags_Pa}', "");
Expect(0, 43136, '\p{	IN_Phags_Pa}', "");
Expect(1, 43136, '\p{^	IN_Phags_Pa}', "");
Expect(1, 43136, '\P{	IN_Phags_Pa}', "");
Expect(0, 43136, '\P{^	IN_Phags_Pa}', "");
Error('\p{-In_PHAGS_Pa/a/}');
Error('\P{-In_PHAGS_Pa/a/}');
Expect(1, 43135, '\p{-in_phags_Pa}', "");
Expect(0, 43135, '\p{^-in_phags_Pa}', "");
Expect(0, 43135, '\P{-in_phags_Pa}', "");
Expect(1, 43135, '\P{^-in_phags_Pa}', "");
Expect(0, 43136, '\p{-in_phags_Pa}', "");
Expect(1, 43136, '\p{^-in_phags_Pa}', "");
Expect(1, 43136, '\P{-in_phags_Pa}', "");
Expect(0, 43136, '\P{^-in_phags_Pa}', "");
Error('\p{  imperial_Aramaic/a/}');
Error('\P{  imperial_Aramaic/a/}');
Expect(1, 67679, '\p{imperialaramaic}', "");
Expect(0, 67679, '\p{^imperialaramaic}', "");
Expect(0, 67679, '\P{imperialaramaic}', "");
Expect(1, 67679, '\P{^imperialaramaic}', "");
Expect(0, 67680, '\p{imperialaramaic}', "");
Expect(1, 67680, '\p{^imperialaramaic}', "");
Expect(1, 67680, '\P{imperialaramaic}', "");
Expect(0, 67680, '\P{^imperialaramaic}', "");
Expect(1, 67679, '\p{	_IMPERIAL_Aramaic}', "");
Expect(0, 67679, '\p{^	_IMPERIAL_Aramaic}', "");
Expect(0, 67679, '\P{	_IMPERIAL_Aramaic}', "");
Expect(1, 67679, '\P{^	_IMPERIAL_Aramaic}', "");
Expect(0, 67680, '\p{	_IMPERIAL_Aramaic}', "");
Expect(1, 67680, '\p{^	_IMPERIAL_Aramaic}', "");
Expect(1, 67680, '\P{	_IMPERIAL_Aramaic}', "");
Expect(0, 67680, '\P{^	_IMPERIAL_Aramaic}', "");
Error('\p{		Is_Imperial_aramaic/a/}');
Error('\P{		Is_Imperial_aramaic/a/}');
Expect(1, 67679, '\p{isimperialaramaic}', "");
Expect(0, 67679, '\p{^isimperialaramaic}', "");
Expect(0, 67679, '\P{isimperialaramaic}', "");
Expect(1, 67679, '\P{^isimperialaramaic}', "");
Expect(0, 67680, '\p{isimperialaramaic}', "");
Expect(1, 67680, '\p{^isimperialaramaic}', "");
Expect(1, 67680, '\P{isimperialaramaic}', "");
Expect(0, 67680, '\P{^isimperialaramaic}', "");
Expect(1, 67679, '\p{_	Is_imperial_ARAMAIC}', "");
Expect(0, 67679, '\p{^_	Is_imperial_ARAMAIC}', "");
Expect(0, 67679, '\P{_	Is_imperial_ARAMAIC}', "");
Expect(1, 67679, '\P{^_	Is_imperial_ARAMAIC}', "");
Expect(0, 67680, '\p{_	Is_imperial_ARAMAIC}', "");
Expect(1, 67680, '\p{^_	Is_imperial_ARAMAIC}', "");
Expect(1, 67680, '\P{_	Is_imperial_ARAMAIC}', "");
Expect(0, 67680, '\P{^_	Is_imperial_ARAMAIC}', "");
Error('\p{_:=armi}');
Error('\P{_:=armi}');
Expect(1, 67679, '\p{armi}', "");
Expect(0, 67679, '\p{^armi}', "");
Expect(0, 67679, '\P{armi}', "");
Expect(1, 67679, '\P{^armi}', "");
Expect(0, 67680, '\p{armi}', "");
Expect(1, 67680, '\p{^armi}', "");
Expect(1, 67680, '\P{armi}', "");
Expect(0, 67680, '\P{^armi}', "");
Expect(1, 67679, '\p{		Armi}', "");
Expect(0, 67679, '\p{^		Armi}', "");
Expect(0, 67679, '\P{		Armi}', "");
Expect(1, 67679, '\P{^		Armi}', "");
Expect(0, 67680, '\p{		Armi}', "");
Expect(1, 67680, '\p{^		Armi}', "");
Expect(1, 67680, '\P{		Armi}', "");
Expect(0, 67680, '\P{^		Armi}', "");
Error('\p{:=-is_ARMI}');
Error('\P{:=-is_ARMI}');
Expect(1, 67679, '\p{isarmi}', "");
Expect(0, 67679, '\p{^isarmi}', "");
Expect(0, 67679, '\P{isarmi}', "");
Expect(1, 67679, '\P{^isarmi}', "");
Expect(0, 67680, '\p{isarmi}', "");
Expect(1, 67680, '\p{^isarmi}', "");
Expect(1, 67680, '\P{isarmi}', "");
Expect(0, 67680, '\P{^isarmi}', "");
Expect(1, 67679, '\p{--Is_armi}', "");
Expect(0, 67679, '\p{^--Is_armi}', "");
Expect(0, 67679, '\P{--Is_armi}', "");
Expect(1, 67679, '\P{^--Is_armi}', "");
Expect(0, 67680, '\p{--Is_armi}', "");
Expect(1, 67680, '\p{^--Is_armi}', "");
Expect(1, 67680, '\P{--Is_armi}', "");
Expect(0, 67680, '\P{^--Is_armi}', "");
Error('\p{ :=PerlSpace}');
Error('\P{ :=PerlSpace}');
Expect(1, 32, '\p{perlspace}', "");
Expect(0, 32, '\p{^perlspace}', "");
Expect(0, 32, '\P{perlspace}', "");
Expect(1, 32, '\P{^perlspace}', "");
Expect(0, 33, '\p{perlspace}', "");
Expect(1, 33, '\p{^perlspace}', "");
Expect(1, 33, '\P{perlspace}', "");
Expect(0, 33, '\P{^perlspace}', "");
Expect(1, 32, '\p{_PerlSpace}', "");
Expect(0, 32, '\p{^_PerlSpace}', "");
Expect(0, 32, '\P{_PerlSpace}', "");
Expect(1, 32, '\P{^_PerlSpace}', "");
Expect(0, 33, '\p{_PerlSpace}', "");
Expect(1, 33, '\p{^_PerlSpace}', "");
Expect(1, 33, '\P{_PerlSpace}', "");
Expect(0, 33, '\P{^_PerlSpace}', "");
Error('\p{-Is_perlspace:=}');
Error('\P{-Is_perlspace:=}');
Expect(1, 32, '\p{isperlspace}', "");
Expect(0, 32, '\p{^isperlspace}', "");
Expect(0, 32, '\P{isperlspace}', "");
Expect(1, 32, '\P{^isperlspace}', "");
Expect(0, 33, '\p{isperlspace}', "");
Expect(1, 33, '\p{^isperlspace}', "");
Expect(1, 33, '\P{isperlspace}', "");
Expect(0, 33, '\P{^isperlspace}', "");
Expect(1, 32, '\p{-_Is_PERLSPACE}', "");
Expect(0, 32, '\p{^-_Is_PERLSPACE}', "");
Expect(0, 32, '\P{-_Is_PERLSPACE}', "");
Expect(1, 32, '\P{^-_Is_PERLSPACE}', "");
Expect(0, 33, '\p{-_Is_PERLSPACE}', "");
Expect(1, 33, '\p{^-_Is_PERLSPACE}', "");
Expect(1, 33, '\P{-_Is_PERLSPACE}', "");
Expect(0, 33, '\P{^-_Is_PERLSPACE}', "");
Error('\p{/a/__CNTRL}');
Error('\P{/a/__CNTRL}');
Expect(1, 159, '\p{cntrl}', "");
Expect(0, 159, '\p{^cntrl}', "");
Expect(0, 159, '\P{cntrl}', "");
Expect(1, 159, '\P{^cntrl}', "");
Expect(0, 160, '\p{cntrl}', "");
Expect(1, 160, '\p{^cntrl}', "");
Expect(1, 160, '\P{cntrl}', "");
Expect(0, 160, '\P{^cntrl}', "");
Expect(1, 159, '\p{_	Cntrl}', "");
Expect(0, 159, '\p{^_	Cntrl}', "");
Expect(0, 159, '\P{_	Cntrl}', "");
Expect(1, 159, '\P{^_	Cntrl}', "");
Expect(0, 160, '\p{_	Cntrl}', "");
Expect(1, 160, '\p{^_	Cntrl}', "");
Expect(1, 160, '\P{_	Cntrl}', "");
Expect(0, 160, '\P{^_	Cntrl}', "");
Error('\p{-:=XPosixCntrl}');
Error('\P{-:=XPosixCntrl}');
Expect(1, 159, '\p{xposixcntrl}', "");
Expect(0, 159, '\p{^xposixcntrl}', "");
Expect(0, 159, '\P{xposixcntrl}', "");
Expect(1, 159, '\P{^xposixcntrl}', "");
Expect(0, 160, '\p{xposixcntrl}', "");
Expect(1, 160, '\p{^xposixcntrl}', "");
Expect(1, 160, '\P{xposixcntrl}', "");
Expect(0, 160, '\P{^xposixcntrl}', "");
Expect(1, 159, '\p{_ XPOSIXCNTRL}', "");
Expect(0, 159, '\p{^_ XPOSIXCNTRL}', "");
Expect(0, 159, '\P{_ XPOSIXCNTRL}', "");
Expect(1, 159, '\P{^_ XPOSIXCNTRL}', "");
Expect(0, 160, '\p{_ XPOSIXCNTRL}', "");
Expect(1, 160, '\p{^_ XPOSIXCNTRL}', "");
Expect(1, 160, '\P{_ XPOSIXCNTRL}', "");
Expect(0, 160, '\P{^_ XPOSIXCNTRL}', "");
Error('\p{:=-_Is_CNTRL}');
Error('\P{:=-_Is_CNTRL}');
Expect(1, 159, '\p{iscntrl}', "");
Expect(0, 159, '\p{^iscntrl}', "");
Expect(0, 159, '\P{iscntrl}', "");
Expect(1, 159, '\P{^iscntrl}', "");
Expect(0, 160, '\p{iscntrl}', "");
Expect(1, 160, '\p{^iscntrl}', "");
Expect(1, 160, '\P{iscntrl}', "");
Expect(0, 160, '\P{^iscntrl}', "");
Expect(1, 159, '\p{ is_Cntrl}', "");
Expect(0, 159, '\p{^ is_Cntrl}', "");
Expect(0, 159, '\P{ is_Cntrl}', "");
Expect(1, 159, '\P{^ is_Cntrl}', "");
Expect(0, 160, '\p{ is_Cntrl}', "");
Expect(1, 160, '\p{^ is_Cntrl}', "");
Expect(1, 160, '\P{ is_Cntrl}', "");
Expect(0, 160, '\P{^ is_Cntrl}', "");
Error('\p{	/a/IS_XPosixCntrl}');
Error('\P{	/a/IS_XPosixCntrl}');
Expect(1, 159, '\p{isxposixcntrl}', "");
Expect(0, 159, '\p{^isxposixcntrl}', "");
Expect(0, 159, '\P{isxposixcntrl}', "");
Expect(1, 159, '\P{^isxposixcntrl}', "");
Expect(0, 160, '\p{isxposixcntrl}', "");
Expect(1, 160, '\p{^isxposixcntrl}', "");
Expect(1, 160, '\P{isxposixcntrl}', "");
Expect(0, 160, '\P{^isxposixcntrl}', "");
Expect(1, 159, '\p{Is_XPOSIXCNTRL}', "");
Expect(0, 159, '\p{^Is_XPOSIXCNTRL}', "");
Expect(0, 159, '\P{Is_XPOSIXCNTRL}', "");
Expect(1, 159, '\P{^Is_XPOSIXCNTRL}', "");
Expect(0, 160, '\p{Is_XPOSIXCNTRL}', "");
Expect(1, 160, '\p{^Is_XPOSIXCNTRL}', "");
Expect(1, 160, '\P{Is_XPOSIXCNTRL}', "");
Expect(0, 160, '\P{^Is_XPOSIXCNTRL}', "");
Error('\p{_/a/control}');
Error('\P{_/a/control}');
Expect(1, 159, '\p{control}', "");
Expect(0, 159, '\p{^control}', "");
Expect(0, 159, '\P{control}', "");
Expect(1, 159, '\P{^control}', "");
Expect(0, 160, '\p{control}', "");
Expect(1, 160, '\p{^control}', "");
Expect(1, 160, '\P{control}', "");
Expect(0, 160, '\P{^control}', "");
Expect(1, 159, '\p{		Control}', "");
Expect(0, 159, '\p{^		Control}', "");
Expect(0, 159, '\P{		Control}', "");
Expect(1, 159, '\P{^		Control}', "");
Expect(0, 160, '\p{		Control}', "");
Expect(1, 160, '\p{^		Control}', "");
Expect(1, 160, '\P{		Control}', "");
Expect(0, 160, '\P{^		Control}', "");
Error('\p{_:=is_control}');
Error('\P{_:=is_control}');
Expect(1, 159, '\p{iscontrol}', "");
Expect(0, 159, '\p{^iscontrol}', "");
Expect(0, 159, '\P{iscontrol}', "");
Expect(1, 159, '\P{^iscontrol}', "");
Expect(0, 160, '\p{iscontrol}', "");
Expect(1, 160, '\p{^iscontrol}', "");
Expect(1, 160, '\P{iscontrol}', "");
Expect(0, 160, '\P{^iscontrol}', "");
Expect(1, 159, '\p{ IS_Control}', "");
Expect(0, 159, '\p{^ IS_Control}', "");
Expect(0, 159, '\P{ IS_Control}', "");
Expect(1, 159, '\P{^ IS_Control}', "");
Expect(0, 160, '\p{ IS_Control}', "");
Expect(1, 160, '\p{^ IS_Control}', "");
Expect(1, 160, '\P{ IS_Control}', "");
Expect(0, 160, '\P{^ IS_Control}', "");
Error('\p{/a/Cc}');
Error('\P{/a/Cc}');
Expect(1, 159, '\p{cc}', "");
Expect(0, 159, '\p{^cc}', "");
Expect(0, 159, '\P{cc}', "");
Expect(1, 159, '\P{^cc}', "");
Expect(0, 160, '\p{cc}', "");
Expect(1, 160, '\p{^cc}', "");
Expect(1, 160, '\P{cc}', "");
Expect(0, 160, '\P{^cc}', "");
Expect(1, 159, '\p{ -CC}', "");
Expect(0, 159, '\p{^ -CC}', "");
Expect(0, 159, '\P{ -CC}', "");
Expect(1, 159, '\P{^ -CC}', "");
Expect(0, 160, '\p{ -CC}', "");
Expect(1, 160, '\p{^ -CC}', "");
Expect(1, 160, '\P{ -CC}', "");
Expect(0, 160, '\P{^ -CC}', "");
Error('\p{/a/_Is_CC}');
Error('\P{/a/_Is_CC}');
Expect(1, 159, '\p{iscc}', "");
Expect(0, 159, '\p{^iscc}', "");
Expect(0, 159, '\P{iscc}', "");
Expect(1, 159, '\P{^iscc}', "");
Expect(0, 160, '\p{iscc}', "");
Expect(1, 160, '\p{^iscc}', "");
Expect(1, 160, '\P{iscc}', "");
Expect(0, 160, '\P{^iscc}', "");
Expect(1, 159, '\p{	_is_CC}', "");
Expect(0, 159, '\p{^	_is_CC}', "");
Expect(0, 159, '\P{	_is_CC}', "");
Expect(1, 159, '\P{^	_is_CC}', "");
Expect(0, 160, '\p{	_is_CC}', "");
Expect(1, 160, '\p{^	_is_CC}', "");
Expect(1, 160, '\P{	_is_CC}', "");
Expect(0, 160, '\P{^	_is_CC}', "");
Error('\p{-IDS_Binary_Operator:=}');
Error('\P{-IDS_Binary_Operator:=}');
Expect(1, 12283, '\p{idsbinaryoperator}', "");
Expect(0, 12283, '\p{^idsbinaryoperator}', "");
Expect(0, 12283, '\P{idsbinaryoperator}', "");
Expect(1, 12283, '\P{^idsbinaryoperator}', "");
Expect(0, 12284, '\p{idsbinaryoperator}', "");
Expect(1, 12284, '\p{^idsbinaryoperator}', "");
Expect(1, 12284, '\P{idsbinaryoperator}', "");
Expect(0, 12284, '\P{^idsbinaryoperator}', "");
Expect(1, 12283, '\p{ _IDS_Binary_Operator}', "");
Expect(0, 12283, '\p{^ _IDS_Binary_Operator}', "");
Expect(0, 12283, '\P{ _IDS_Binary_Operator}', "");
Expect(1, 12283, '\P{^ _IDS_Binary_Operator}', "");
Expect(0, 12284, '\p{ _IDS_Binary_Operator}', "");
Expect(1, 12284, '\p{^ _IDS_Binary_Operator}', "");
Expect(1, 12284, '\P{ _IDS_Binary_Operator}', "");
Expect(0, 12284, '\P{^ _IDS_Binary_Operator}', "");
Error('\p{_ IS_IDS_binary_Operator:=}');
Error('\P{_ IS_IDS_binary_Operator:=}');
Expect(1, 12283, '\p{isidsbinaryoperator}', "");
Expect(0, 12283, '\p{^isidsbinaryoperator}', "");
Expect(0, 12283, '\P{isidsbinaryoperator}', "");
Expect(1, 12283, '\P{^isidsbinaryoperator}', "");
Expect(0, 12284, '\p{isidsbinaryoperator}', "");
Expect(1, 12284, '\p{^isidsbinaryoperator}', "");
Expect(1, 12284, '\P{isidsbinaryoperator}', "");
Expect(0, 12284, '\P{^isidsbinaryoperator}', "");
Expect(1, 12283, '\p{_Is_IDS_binary_OPERATOR}', "");
Expect(0, 12283, '\p{^_Is_IDS_binary_OPERATOR}', "");
Expect(0, 12283, '\P{_Is_IDS_binary_OPERATOR}', "");
Expect(1, 12283, '\P{^_Is_IDS_binary_OPERATOR}', "");
Expect(0, 12284, '\p{_Is_IDS_binary_OPERATOR}', "");
Expect(1, 12284, '\p{^_Is_IDS_binary_OPERATOR}', "");
Expect(1, 12284, '\P{_Is_IDS_binary_OPERATOR}', "");
Expect(0, 12284, '\P{^_Is_IDS_binary_OPERATOR}', "");
Error('\p{:=_IDSB}');
Error('\P{:=_IDSB}');
Expect(1, 12283, '\p{idsb}', "");
Expect(0, 12283, '\p{^idsb}', "");
Expect(0, 12283, '\P{idsb}', "");
Expect(1, 12283, '\P{^idsb}', "");
Expect(0, 12284, '\p{idsb}', "");
Expect(1, 12284, '\p{^idsb}', "");
Expect(1, 12284, '\P{idsb}', "");
Expect(0, 12284, '\P{^idsb}', "");
Expect(1, 12283, '\p{  IDSB}', "");
Expect(0, 12283, '\p{^  IDSB}', "");
Expect(0, 12283, '\P{  IDSB}', "");
Expect(1, 12283, '\P{^  IDSB}', "");
Expect(0, 12284, '\p{  IDSB}', "");
Expect(1, 12284, '\p{^  IDSB}', "");
Expect(1, 12284, '\P{  IDSB}', "");
Expect(0, 12284, '\P{^  IDSB}', "");
Error('\p{Is_IDSB/a/}');
Error('\P{Is_IDSB/a/}');
Expect(1, 12283, '\p{isidsb}', "");
Expect(0, 12283, '\p{^isidsb}', "");
Expect(0, 12283, '\P{isidsb}', "");
Expect(1, 12283, '\P{^isidsb}', "");
Expect(0, 12284, '\p{isidsb}', "");
Expect(1, 12284, '\p{^isidsb}', "");
Expect(1, 12284, '\P{isidsb}', "");
Expect(0, 12284, '\P{^isidsb}', "");
Expect(1, 12283, '\p{ -Is_IDSB}', "");
Expect(0, 12283, '\p{^ -Is_IDSB}', "");
Expect(0, 12283, '\P{ -Is_IDSB}', "");
Expect(1, 12283, '\P{^ -Is_IDSB}', "");
Expect(0, 12284, '\p{ -Is_IDSB}', "");
Expect(1, 12284, '\p{^ -Is_IDSB}', "");
Expect(1, 12284, '\P{ -Is_IDSB}', "");
Expect(0, 12284, '\P{^ -Is_IDSB}', "");
Error('\p{_Changes_when_Titlecased:=}');
Error('\P{_Changes_when_Titlecased:=}');
Expect(1, 66639, '\p{changeswhentitlecased}', "");
Expect(0, 66639, '\p{^changeswhentitlecased}', "");
Expect(0, 66639, '\P{changeswhentitlecased}', "");
Expect(1, 66639, '\P{^changeswhentitlecased}', "");
Expect(0, 66640, '\p{changeswhentitlecased}', "");
Expect(1, 66640, '\p{^changeswhentitlecased}', "");
Expect(1, 66640, '\P{changeswhentitlecased}', "");
Expect(0, 66640, '\P{^changeswhentitlecased}', "");
Expect(1, 66639, '\p{_-Changes_When_titlecased}', "");
Expect(0, 66639, '\p{^_-Changes_When_titlecased}', "");
Expect(0, 66639, '\P{_-Changes_When_titlecased}', "");
Expect(1, 66639, '\P{^_-Changes_When_titlecased}', "");
Expect(0, 66640, '\p{_-Changes_When_titlecased}', "");
Expect(1, 66640, '\p{^_-Changes_When_titlecased}', "");
Expect(1, 66640, '\P{_-Changes_When_titlecased}', "");
Expect(0, 66640, '\P{^_-Changes_When_titlecased}', "");
Error('\p{  is_Changes_When_TITLECASED:=}');
Error('\P{  is_Changes_When_TITLECASED:=}');
Expect(1, 66639, '\p{ischangeswhentitlecased}', "");
Expect(0, 66639, '\p{^ischangeswhentitlecased}', "");
Expect(0, 66639, '\P{ischangeswhentitlecased}', "");
Expect(1, 66639, '\P{^ischangeswhentitlecased}', "");
Expect(0, 66640, '\p{ischangeswhentitlecased}', "");
Expect(1, 66640, '\p{^ischangeswhentitlecased}', "");
Expect(1, 66640, '\P{ischangeswhentitlecased}', "");
Expect(0, 66640, '\P{^ischangeswhentitlecased}', "");
Expect(1, 66639, '\p{ Is_CHANGES_when_titlecased}', "");
Expect(0, 66639, '\p{^ Is_CHANGES_when_titlecased}', "");
Expect(0, 66639, '\P{ Is_CHANGES_when_titlecased}', "");
Expect(1, 66639, '\P{^ Is_CHANGES_when_titlecased}', "");
Expect(0, 66640, '\p{ Is_CHANGES_when_titlecased}', "");
Expect(1, 66640, '\p{^ Is_CHANGES_when_titlecased}', "");
Expect(1, 66640, '\P{ Is_CHANGES_when_titlecased}', "");
Expect(0, 66640, '\P{^ Is_CHANGES_when_titlecased}', "");
Error('\p{	_CWT:=}');
Error('\P{	_CWT:=}');
Expect(1, 66639, '\p{cwt}', "");
Expect(0, 66639, '\p{^cwt}', "");
Expect(0, 66639, '\P{cwt}', "");
Expect(1, 66639, '\P{^cwt}', "");
Expect(0, 66640, '\p{cwt}', "");
Expect(1, 66640, '\p{^cwt}', "");
Expect(1, 66640, '\P{cwt}', "");
Expect(0, 66640, '\P{^cwt}', "");
Expect(1, 66639, '\p{_-CWT}', "");
Expect(0, 66639, '\p{^_-CWT}', "");
Expect(0, 66639, '\P{_-CWT}', "");
Expect(1, 66639, '\P{^_-CWT}', "");
Expect(0, 66640, '\p{_-CWT}', "");
Expect(1, 66640, '\p{^_-CWT}', "");
Expect(1, 66640, '\P{_-CWT}', "");
Expect(0, 66640, '\P{^_-CWT}', "");
Error('\p{_IS_cwt:=}');
Error('\P{_IS_cwt:=}');
Expect(1, 66639, '\p{iscwt}', "");
Expect(0, 66639, '\p{^iscwt}', "");
Expect(0, 66639, '\P{iscwt}', "");
Expect(1, 66639, '\P{^iscwt}', "");
Expect(0, 66640, '\p{iscwt}', "");
Expect(1, 66640, '\p{^iscwt}', "");
Expect(1, 66640, '\P{iscwt}', "");
Expect(0, 66640, '\P{^iscwt}', "");
Expect(1, 66639, '\p{__IS_CWT}', "");
Expect(0, 66639, '\p{^__IS_CWT}', "");
Expect(0, 66639, '\P{__IS_CWT}', "");
Expect(1, 66639, '\P{^__IS_CWT}', "");
Expect(0, 66640, '\p{__IS_CWT}', "");
Expect(1, 66640, '\p{^__IS_CWT}', "");
Expect(1, 66640, '\P{__IS_CWT}', "");
Expect(0, 66640, '\P{^__IS_CWT}', "");
Error('\p{:=_PosixLower}');
Error('\P{:=_PosixLower}');
Expect(1, 122, '\p{posixlower}', "");
Expect(0, 122, '\p{^posixlower}', "");
Expect(0, 122, '\P{posixlower}', "");
Expect(1, 122, '\P{^posixlower}', "");
Expect(0, 123, '\p{posixlower}', "");
Expect(1, 123, '\p{^posixlower}', "");
Expect(1, 123, '\P{posixlower}', "");
Expect(0, 123, '\P{^posixlower}', "");
Expect(1, 122, '\p{_-POSIXLOWER}', "");
Expect(0, 122, '\p{^_-POSIXLOWER}', "");
Expect(0, 122, '\P{_-POSIXLOWER}', "");
Expect(1, 122, '\P{^_-POSIXLOWER}', "");
Expect(0, 123, '\p{_-POSIXLOWER}', "");
Expect(1, 123, '\p{^_-POSIXLOWER}', "");
Expect(1, 123, '\P{_-POSIXLOWER}', "");
Expect(0, 123, '\P{^_-POSIXLOWER}', "");
Error('\p{-is_POSIXLOWER:=}');
Error('\P{-is_POSIXLOWER:=}');
Expect(1, 122, '\p{isposixlower}', "");
Expect(0, 122, '\p{^isposixlower}', "");
Expect(0, 122, '\P{isposixlower}', "");
Expect(1, 122, '\P{^isposixlower}', "");
Expect(0, 123, '\p{isposixlower}', "");
Expect(1, 123, '\p{^isposixlower}', "");
Expect(1, 123, '\P{isposixlower}', "");
Expect(0, 123, '\P{^isposixlower}', "");
Expect(1, 122, '\p{	_Is_PosixLower}', "");
Expect(0, 122, '\p{^	_Is_PosixLower}', "");
Expect(0, 122, '\P{	_Is_PosixLower}', "");
Expect(1, 122, '\P{^	_Is_PosixLower}', "");
Expect(0, 123, '\p{	_Is_PosixLower}', "");
Expect(1, 123, '\p{^	_Is_PosixLower}', "");
Expect(1, 123, '\P{	_Is_PosixLower}', "");
Expect(0, 123, '\P{^	_Is_PosixLower}', "");
Error('\p{_:=Supplemental_ARROWS_a}');
Error('\P{_:=Supplemental_ARROWS_a}');
Expect(1, 10239, '\p{supplementalarrowsa}', "");
Expect(0, 10239, '\p{^supplementalarrowsa}', "");
Expect(0, 10239, '\P{supplementalarrowsa}', "");
Expect(1, 10239, '\P{^supplementalarrowsa}', "");
Expect(0, 10240, '\p{supplementalarrowsa}', "");
Expect(1, 10240, '\p{^supplementalarrowsa}', "");
Expect(1, 10240, '\P{supplementalarrowsa}', "");
Expect(0, 10240, '\P{^supplementalarrowsa}', "");
Expect(1, 10239, '\p{_	Supplemental_arrows_a}', "");
Expect(0, 10239, '\p{^_	Supplemental_arrows_a}', "");
Expect(0, 10239, '\P{_	Supplemental_arrows_a}', "");
Expect(1, 10239, '\P{^_	Supplemental_arrows_a}', "");
Expect(0, 10240, '\p{_	Supplemental_arrows_a}', "");
Expect(1, 10240, '\p{^_	Supplemental_arrows_a}', "");
Expect(1, 10240, '\P{_	Supplemental_arrows_a}', "");
Expect(0, 10240, '\P{^_	Supplemental_arrows_a}', "");
Error('\p{ _is_Supplemental_ARROWS_A:=}');
Error('\P{ _is_Supplemental_ARROWS_A:=}');
Expect(1, 10239, '\p{issupplementalarrowsa}', "");
Expect(0, 10239, '\p{^issupplementalarrowsa}', "");
Expect(0, 10239, '\P{issupplementalarrowsa}', "");
Expect(1, 10239, '\P{^issupplementalarrowsa}', "");
Expect(0, 10240, '\p{issupplementalarrowsa}', "");
Expect(1, 10240, '\p{^issupplementalarrowsa}', "");
Expect(1, 10240, '\P{issupplementalarrowsa}', "");
Expect(0, 10240, '\P{^issupplementalarrowsa}', "");
Expect(1, 10239, '\p{-Is_SUPPLEMENTAL_ARROWS_A}', "");
Expect(0, 10239, '\p{^-Is_SUPPLEMENTAL_ARROWS_A}', "");
Expect(0, 10239, '\P{-Is_SUPPLEMENTAL_ARROWS_A}', "");
Expect(1, 10239, '\P{^-Is_SUPPLEMENTAL_ARROWS_A}', "");
Expect(0, 10240, '\p{-Is_SUPPLEMENTAL_ARROWS_A}', "");
Expect(1, 10240, '\p{^-Is_SUPPLEMENTAL_ARROWS_A}', "");
Expect(1, 10240, '\P{-Is_SUPPLEMENTAL_ARROWS_A}', "");
Expect(0, 10240, '\P{^-Is_SUPPLEMENTAL_ARROWS_A}', "");
Error('\p{ IN_supplemental_ARROWS_A/a/}');
Error('\P{ IN_supplemental_ARROWS_A/a/}');
Expect(1, 10239, '\p{insupplementalarrowsa}', "");
Expect(0, 10239, '\p{^insupplementalarrowsa}', "");
Expect(0, 10239, '\P{insupplementalarrowsa}', "");
Expect(1, 10239, '\P{^insupplementalarrowsa}', "");
Expect(0, 10240, '\p{insupplementalarrowsa}', "");
Expect(1, 10240, '\p{^insupplementalarrowsa}', "");
Expect(1, 10240, '\P{insupplementalarrowsa}', "");
Expect(0, 10240, '\P{^insupplementalarrowsa}', "");
Expect(1, 10239, '\p{_-IN_supplemental_Arrows_A}', "");
Expect(0, 10239, '\p{^_-IN_supplemental_Arrows_A}', "");
Expect(0, 10239, '\P{_-IN_supplemental_Arrows_A}', "");
Expect(1, 10239, '\P{^_-IN_supplemental_Arrows_A}', "");
Expect(0, 10240, '\p{_-IN_supplemental_Arrows_A}', "");
Expect(1, 10240, '\p{^_-IN_supplemental_Arrows_A}', "");
Expect(1, 10240, '\P{_-IN_supplemental_Arrows_A}', "");
Expect(0, 10240, '\P{^_-IN_supplemental_Arrows_A}', "");
Error('\p{-_sup_Arrows_A/a/}');
Error('\P{-_sup_Arrows_A/a/}');
Expect(1, 10239, '\p{suparrowsa}', "");
Expect(0, 10239, '\p{^suparrowsa}', "");
Expect(0, 10239, '\P{suparrowsa}', "");
Expect(1, 10239, '\P{^suparrowsa}', "");
Expect(0, 10240, '\p{suparrowsa}', "");
Expect(1, 10240, '\p{^suparrowsa}', "");
Expect(1, 10240, '\P{suparrowsa}', "");
Expect(0, 10240, '\P{^suparrowsa}', "");
Expect(1, 10239, '\p{ _sup_Arrows_a}', "");
Expect(0, 10239, '\p{^ _sup_Arrows_a}', "");
Expect(0, 10239, '\P{ _sup_Arrows_a}', "");
Expect(1, 10239, '\P{^ _sup_Arrows_a}', "");
Expect(0, 10240, '\p{ _sup_Arrows_a}', "");
Expect(1, 10240, '\p{^ _sup_Arrows_a}', "");
Expect(1, 10240, '\P{ _sup_Arrows_a}', "");
Expect(0, 10240, '\P{^ _sup_Arrows_a}', "");
Error('\p{/a/-	is_SUP_Arrows_A}');
Error('\P{/a/-	is_SUP_Arrows_A}');
Expect(1, 10239, '\p{issuparrowsa}', "");
Expect(0, 10239, '\p{^issuparrowsa}', "");
Expect(0, 10239, '\P{issuparrowsa}', "");
Expect(1, 10239, '\P{^issuparrowsa}', "");
Expect(0, 10240, '\p{issuparrowsa}', "");
Expect(1, 10240, '\p{^issuparrowsa}', "");
Expect(1, 10240, '\P{issuparrowsa}', "");
Expect(0, 10240, '\P{^issuparrowsa}', "");
Expect(1, 10239, '\p{- is_SUP_Arrows_a}', "");
Expect(0, 10239, '\p{^- is_SUP_Arrows_a}', "");
Expect(0, 10239, '\P{- is_SUP_Arrows_a}', "");
Expect(1, 10239, '\P{^- is_SUP_Arrows_a}', "");
Expect(0, 10240, '\p{- is_SUP_Arrows_a}', "");
Expect(1, 10240, '\p{^- is_SUP_Arrows_a}', "");
Expect(1, 10240, '\P{- is_SUP_Arrows_a}', "");
Expect(0, 10240, '\P{^- is_SUP_Arrows_a}', "");
Error('\p{	 IN_Sup_Arrows_A:=}');
Error('\P{	 IN_Sup_Arrows_A:=}');
Expect(1, 10239, '\p{insuparrowsa}', "");
Expect(0, 10239, '\p{^insuparrowsa}', "");
Expect(0, 10239, '\P{insuparrowsa}', "");
Expect(1, 10239, '\P{^insuparrowsa}', "");
Expect(0, 10240, '\p{insuparrowsa}', "");
Expect(1, 10240, '\p{^insuparrowsa}', "");
Expect(1, 10240, '\P{insuparrowsa}', "");
Expect(0, 10240, '\P{^insuparrowsa}', "");
Expect(1, 10239, '\p{-In_Sup_Arrows_a}', "");
Expect(0, 10239, '\p{^-In_Sup_Arrows_a}', "");
Expect(0, 10239, '\P{-In_Sup_Arrows_a}', "");
Expect(1, 10239, '\P{^-In_Sup_Arrows_a}', "");
Expect(0, 10240, '\p{-In_Sup_Arrows_a}', "");
Expect(1, 10240, '\p{^-In_Sup_Arrows_a}', "");
Expect(1, 10240, '\P{-In_Sup_Arrows_a}', "");
Expect(0, 10240, '\P{^-In_Sup_Arrows_a}', "");
Error('\p{Other_Grapheme_Extend}');
Error('\P{Other_Grapheme_Extend}');
Error('\p{Is_Other_Grapheme_Extend}');
Error('\P{Is_Other_Grapheme_Extend}');
Error('\p{OGr_Ext}');
Error('\P{OGr_Ext}');
Error('\p{Is_OGr_Ext}');
Error('\P{Is_OGr_Ext}');
Error('\p{:=	 OLD_Persian}');
Error('\P{:=	 OLD_Persian}');
Expect(1, 66517, '\p{oldpersian}', "");
Expect(0, 66517, '\p{^oldpersian}', "");
Expect(0, 66517, '\P{oldpersian}', "");
Expect(1, 66517, '\P{^oldpersian}', "");
Expect(0, 66518, '\p{oldpersian}', "");
Expect(1, 66518, '\p{^oldpersian}', "");
Expect(1, 66518, '\P{oldpersian}', "");
Expect(0, 66518, '\P{^oldpersian}', "");
Expect(1, 66517, '\p{	OLD_Persian}', "");
Expect(0, 66517, '\p{^	OLD_Persian}', "");
Expect(0, 66517, '\P{	OLD_Persian}', "");
Expect(1, 66517, '\P{^	OLD_Persian}', "");
Expect(0, 66518, '\p{	OLD_Persian}', "");
Expect(1, 66518, '\p{^	OLD_Persian}', "");
Expect(1, 66518, '\P{	OLD_Persian}', "");
Expect(0, 66518, '\P{^	OLD_Persian}', "");
Error('\p{ :=IS_OLD_Persian}');
Error('\P{ :=IS_OLD_Persian}');
Expect(1, 66517, '\p{isoldpersian}', "");
Expect(0, 66517, '\p{^isoldpersian}', "");
Expect(0, 66517, '\P{isoldpersian}', "");
Expect(1, 66517, '\P{^isoldpersian}', "");
Expect(0, 66518, '\p{isoldpersian}', "");
Expect(1, 66518, '\p{^isoldpersian}', "");
Expect(1, 66518, '\P{isoldpersian}', "");
Expect(0, 66518, '\P{^isoldpersian}', "");
Expect(1, 66517, '\p{- IS_Old_PERSIAN}', "");
Expect(0, 66517, '\p{^- IS_Old_PERSIAN}', "");
Expect(0, 66517, '\P{- IS_Old_PERSIAN}', "");
Expect(1, 66517, '\P{^- IS_Old_PERSIAN}', "");
Expect(0, 66518, '\p{- IS_Old_PERSIAN}', "");
Expect(1, 66518, '\p{^- IS_Old_PERSIAN}', "");
Expect(1, 66518, '\P{- IS_Old_PERSIAN}', "");
Expect(0, 66518, '\P{^- IS_Old_PERSIAN}', "");
Error('\p{- Xpeo/a/}');
Error('\P{- Xpeo/a/}');
Expect(1, 66517, '\p{xpeo}', "");
Expect(0, 66517, '\p{^xpeo}', "");
Expect(0, 66517, '\P{xpeo}', "");
Expect(1, 66517, '\P{^xpeo}', "");
Expect(0, 66518, '\p{xpeo}', "");
Expect(1, 66518, '\p{^xpeo}', "");
Expect(1, 66518, '\P{xpeo}', "");
Expect(0, 66518, '\P{^xpeo}', "");
Expect(1, 66517, '\p{	_XPEO}', "");
Expect(0, 66517, '\p{^	_XPEO}', "");
Expect(0, 66517, '\P{	_XPEO}', "");
Expect(1, 66517, '\P{^	_XPEO}', "");
Expect(0, 66518, '\p{	_XPEO}', "");
Expect(1, 66518, '\p{^	_XPEO}', "");
Expect(1, 66518, '\P{	_XPEO}', "");
Expect(0, 66518, '\P{^	_XPEO}', "");
Error('\p{- IS_Xpeo/a/}');
Error('\P{- IS_Xpeo/a/}');
Expect(1, 66517, '\p{isxpeo}', "");
Expect(0, 66517, '\p{^isxpeo}', "");
Expect(0, 66517, '\P{isxpeo}', "");
Expect(1, 66517, '\P{^isxpeo}', "");
Expect(0, 66518, '\p{isxpeo}', "");
Expect(1, 66518, '\p{^isxpeo}', "");
Expect(1, 66518, '\P{isxpeo}', "");
Expect(0, 66518, '\P{^isxpeo}', "");
Expect(1, 66517, '\p{-	IS_Xpeo}', "");
Expect(0, 66517, '\p{^-	IS_Xpeo}', "");
Expect(0, 66517, '\P{-	IS_Xpeo}', "");
Expect(1, 66517, '\P{^-	IS_Xpeo}', "");
Expect(0, 66518, '\p{-	IS_Xpeo}', "");
Expect(1, 66518, '\p{^-	IS_Xpeo}', "");
Expect(1, 66518, '\P{-	IS_Xpeo}', "");
Expect(0, 66518, '\P{^-	IS_Xpeo}', "");
Error('\p{ /a/Phaistos_DISC}');
Error('\P{ /a/Phaistos_DISC}');
Expect(1, 66047, '\p{phaistosdisc}', "");
Expect(0, 66047, '\p{^phaistosdisc}', "");
Expect(0, 66047, '\P{phaistosdisc}', "");
Expect(1, 66047, '\P{^phaistosdisc}', "");
Expect(0, 66048, '\p{phaistosdisc}', "");
Expect(1, 66048, '\p{^phaistosdisc}', "");
Expect(1, 66048, '\P{phaistosdisc}', "");
Expect(0, 66048, '\P{^phaistosdisc}', "");
Expect(1, 66047, '\p{	-PHAISTOS_Disc}', "");
Expect(0, 66047, '\p{^	-PHAISTOS_Disc}', "");
Expect(0, 66047, '\P{	-PHAISTOS_Disc}', "");
Expect(1, 66047, '\P{^	-PHAISTOS_Disc}', "");
Expect(0, 66048, '\p{	-PHAISTOS_Disc}', "");
Expect(1, 66048, '\p{^	-PHAISTOS_Disc}', "");
Expect(1, 66048, '\P{	-PHAISTOS_Disc}', "");
Expect(0, 66048, '\P{^	-PHAISTOS_Disc}', "");
Error('\p{-_Is_phaistos_disc/a/}');
Error('\P{-_Is_phaistos_disc/a/}');
Expect(1, 66047, '\p{isphaistosdisc}', "");
Expect(0, 66047, '\p{^isphaistosdisc}', "");
Expect(0, 66047, '\P{isphaistosdisc}', "");
Expect(1, 66047, '\P{^isphaistosdisc}', "");
Expect(0, 66048, '\p{isphaistosdisc}', "");
Expect(1, 66048, '\p{^isphaistosdisc}', "");
Expect(1, 66048, '\P{isphaistosdisc}', "");
Expect(0, 66048, '\P{^isphaistosdisc}', "");
Expect(1, 66047, '\p{_	is_PHAISTOS_DISC}', "");
Expect(0, 66047, '\p{^_	is_PHAISTOS_DISC}', "");
Expect(0, 66047, '\P{_	is_PHAISTOS_DISC}', "");
Expect(1, 66047, '\P{^_	is_PHAISTOS_DISC}', "");
Expect(0, 66048, '\p{_	is_PHAISTOS_DISC}', "");
Expect(1, 66048, '\p{^_	is_PHAISTOS_DISC}', "");
Expect(1, 66048, '\P{_	is_PHAISTOS_DISC}', "");
Expect(0, 66048, '\P{^_	is_PHAISTOS_DISC}', "");
Error('\p{/a/in_Phaistos_Disc}');
Error('\P{/a/in_Phaistos_Disc}');
Expect(1, 66047, '\p{inphaistosdisc}', "");
Expect(0, 66047, '\p{^inphaistosdisc}', "");
Expect(0, 66047, '\P{inphaistosdisc}', "");
Expect(1, 66047, '\P{^inphaistosdisc}', "");
Expect(0, 66048, '\p{inphaistosdisc}', "");
Expect(1, 66048, '\p{^inphaistosdisc}', "");
Expect(1, 66048, '\P{inphaistosdisc}', "");
Expect(0, 66048, '\P{^inphaistosdisc}', "");
Expect(1, 66047, '\p{		IN_PHAISTOS_DISC}', "");
Expect(0, 66047, '\p{^		IN_PHAISTOS_DISC}', "");
Expect(0, 66047, '\P{		IN_PHAISTOS_DISC}', "");
Expect(1, 66047, '\P{^		IN_PHAISTOS_DISC}', "");
Expect(0, 66048, '\p{		IN_PHAISTOS_DISC}', "");
Expect(1, 66048, '\p{^		IN_PHAISTOS_DISC}', "");
Expect(1, 66048, '\P{		IN_PHAISTOS_DISC}', "");
Expect(0, 66048, '\P{^		IN_PHAISTOS_DISC}', "");
Error('\p{:= _PHAISTOS}');
Error('\P{:= _PHAISTOS}');
Expect(1, 66047, '\p{phaistos}', "");
Expect(0, 66047, '\p{^phaistos}', "");
Expect(0, 66047, '\P{phaistos}', "");
Expect(1, 66047, '\P{^phaistos}', "");
Expect(0, 66048, '\p{phaistos}', "");
Expect(1, 66048, '\p{^phaistos}', "");
Expect(1, 66048, '\P{phaistos}', "");
Expect(0, 66048, '\P{^phaistos}', "");
Expect(1, 66047, '\p{_ PHAISTOS}', "");
Expect(0, 66047, '\p{^_ PHAISTOS}', "");
Expect(0, 66047, '\P{_ PHAISTOS}', "");
Expect(1, 66047, '\P{^_ PHAISTOS}', "");
Expect(0, 66048, '\p{_ PHAISTOS}', "");
Expect(1, 66048, '\p{^_ PHAISTOS}', "");
Expect(1, 66048, '\P{_ PHAISTOS}', "");
Expect(0, 66048, '\P{^_ PHAISTOS}', "");
Error('\p{:=--Is_PHAISTOS}');
Error('\P{:=--Is_PHAISTOS}');
Expect(1, 66047, '\p{isphaistos}', "");
Expect(0, 66047, '\p{^isphaistos}', "");
Expect(0, 66047, '\P{isphaistos}', "");
Expect(1, 66047, '\P{^isphaistos}', "");
Expect(0, 66048, '\p{isphaistos}', "");
Expect(1, 66048, '\p{^isphaistos}', "");
Expect(1, 66048, '\P{isphaistos}', "");
Expect(0, 66048, '\P{^isphaistos}', "");
Expect(1, 66047, '\p{-IS_Phaistos}', "");
Expect(0, 66047, '\p{^-IS_Phaistos}', "");
Expect(0, 66047, '\P{-IS_Phaistos}', "");
Expect(1, 66047, '\P{^-IS_Phaistos}', "");
Expect(0, 66048, '\p{-IS_Phaistos}', "");
Expect(1, 66048, '\p{^-IS_Phaistos}', "");
Expect(1, 66048, '\P{-IS_Phaistos}', "");
Expect(0, 66048, '\P{^-IS_Phaistos}', "");
Error('\p{  In_Phaistos/a/}');
Error('\P{  In_Phaistos/a/}');
Expect(1, 66047, '\p{inphaistos}', "");
Expect(0, 66047, '\p{^inphaistos}', "");
Expect(0, 66047, '\P{inphaistos}', "");
Expect(1, 66047, '\P{^inphaistos}', "");
Expect(0, 66048, '\p{inphaistos}', "");
Expect(1, 66048, '\p{^inphaistos}', "");
Expect(1, 66048, '\P{inphaistos}', "");
Expect(0, 66048, '\P{^inphaistos}', "");
Expect(1, 66047, '\p{ -IN_Phaistos}', "");
Expect(0, 66047, '\p{^ -IN_Phaistos}', "");
Expect(0, 66047, '\P{ -IN_Phaistos}', "");
Expect(1, 66047, '\P{^ -IN_Phaistos}', "");
Expect(0, 66048, '\p{ -IN_Phaistos}', "");
Expect(1, 66048, '\p{^ -IN_Phaistos}', "");
Expect(1, 66048, '\P{ -IN_Phaistos}', "");
Expect(0, 66048, '\P{^ -IN_Phaistos}', "");
Error('\p{ :=CJK_unified_ideographs_Extension_A}');
Error('\P{ :=CJK_unified_ideographs_Extension_A}');
Expect(1, 19903, '\p{cjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\p{^cjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\P{cjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\P{^cjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\p{cjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\p{^cjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\P{cjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\P{^cjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\p{- CJK_Unified_IDEOGRAPHS_extension_A}', "");
Expect(0, 19903, '\p{^- CJK_Unified_IDEOGRAPHS_extension_A}', "");
Expect(0, 19903, '\P{- CJK_Unified_IDEOGRAPHS_extension_A}', "");
Expect(1, 19903, '\P{^- CJK_Unified_IDEOGRAPHS_extension_A}', "");
Expect(0, 19904, '\p{- CJK_Unified_IDEOGRAPHS_extension_A}', "");
Expect(1, 19904, '\p{^- CJK_Unified_IDEOGRAPHS_extension_A}', "");
Expect(1, 19904, '\P{- CJK_Unified_IDEOGRAPHS_extension_A}', "");
Expect(0, 19904, '\P{^- CJK_Unified_IDEOGRAPHS_extension_A}', "");
Error('\p{:=Is_CJK_Unified_Ideographs_EXTENSION_a}');
Error('\P{:=Is_CJK_Unified_Ideographs_EXTENSION_a}');
Expect(1, 19903, '\p{iscjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\p{^iscjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\P{iscjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\P{^iscjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\p{iscjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\p{^iscjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\P{iscjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\P{^iscjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\p{is_CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19903, '\p{^is_CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19903, '\P{is_CJK_unified_Ideographs_extension_A}', "");
Expect(1, 19903, '\P{^is_CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19904, '\p{is_CJK_unified_Ideographs_extension_A}', "");
Expect(1, 19904, '\p{^is_CJK_unified_Ideographs_extension_A}', "");
Expect(1, 19904, '\P{is_CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19904, '\P{^is_CJK_unified_Ideographs_extension_A}', "");
Error('\p{ In_CJK_unified_IDEOGRAPHS_extension_A/a/}');
Error('\P{ In_CJK_unified_IDEOGRAPHS_extension_A/a/}');
Expect(1, 19903, '\p{incjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\p{^incjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\P{incjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\P{^incjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\p{incjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\p{^incjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\P{incjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\P{^incjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\p{ _in_CJK_unified_Ideographs_EXTENSION_A}', "");
Expect(0, 19903, '\p{^ _in_CJK_unified_Ideographs_EXTENSION_A}', "");
Expect(0, 19903, '\P{ _in_CJK_unified_Ideographs_EXTENSION_A}', "");
Expect(1, 19903, '\P{^ _in_CJK_unified_Ideographs_EXTENSION_A}', "");
Expect(0, 19904, '\p{ _in_CJK_unified_Ideographs_EXTENSION_A}', "");
Expect(1, 19904, '\p{^ _in_CJK_unified_Ideographs_EXTENSION_A}', "");
Expect(1, 19904, '\P{ _in_CJK_unified_Ideographs_EXTENSION_A}', "");
Expect(0, 19904, '\P{^ _in_CJK_unified_Ideographs_EXTENSION_A}', "");
Error('\p{/a/-	CJK_Ext_A}');
Error('\P{/a/-	CJK_Ext_A}');
Expect(1, 19903, '\p{cjkexta}', "");
Expect(0, 19903, '\p{^cjkexta}', "");
Expect(0, 19903, '\P{cjkexta}', "");
Expect(1, 19903, '\P{^cjkexta}', "");
Expect(0, 19904, '\p{cjkexta}', "");
Expect(1, 19904, '\p{^cjkexta}', "");
Expect(1, 19904, '\P{cjkexta}', "");
Expect(0, 19904, '\P{^cjkexta}', "");
Expect(1, 19903, '\p{ -CJK_Ext_A}', "");
Expect(0, 19903, '\p{^ -CJK_Ext_A}', "");
Expect(0, 19903, '\P{ -CJK_Ext_A}', "");
Expect(1, 19903, '\P{^ -CJK_Ext_A}', "");
Expect(0, 19904, '\p{ -CJK_Ext_A}', "");
Expect(1, 19904, '\p{^ -CJK_Ext_A}', "");
Expect(1, 19904, '\P{ -CJK_Ext_A}', "");
Expect(0, 19904, '\P{^ -CJK_Ext_A}', "");
Error('\p{ /a/IS_cjk_Ext_A}');
Error('\P{ /a/IS_cjk_Ext_A}');
Expect(1, 19903, '\p{iscjkexta}', "");
Expect(0, 19903, '\p{^iscjkexta}', "");
Expect(0, 19903, '\P{iscjkexta}', "");
Expect(1, 19903, '\P{^iscjkexta}', "");
Expect(0, 19904, '\p{iscjkexta}', "");
Expect(1, 19904, '\p{^iscjkexta}', "");
Expect(1, 19904, '\P{iscjkexta}', "");
Expect(0, 19904, '\P{^iscjkexta}', "");
Expect(1, 19903, '\p{_	Is_CJK_Ext_A}', "");
Expect(0, 19903, '\p{^_	Is_CJK_Ext_A}', "");
Expect(0, 19903, '\P{_	Is_CJK_Ext_A}', "");
Expect(1, 19903, '\P{^_	Is_CJK_Ext_A}', "");
Expect(0, 19904, '\p{_	Is_CJK_Ext_A}', "");
Expect(1, 19904, '\p{^_	Is_CJK_Ext_A}', "");
Expect(1, 19904, '\P{_	Is_CJK_Ext_A}', "");
Expect(0, 19904, '\P{^_	Is_CJK_Ext_A}', "");
Error('\p{-in_CJK_ext_A/a/}');
Error('\P{-in_CJK_ext_A/a/}');
Expect(1, 19903, '\p{incjkexta}', "");
Expect(0, 19903, '\p{^incjkexta}', "");
Expect(0, 19903, '\P{incjkexta}', "");
Expect(1, 19903, '\P{^incjkexta}', "");
Expect(0, 19904, '\p{incjkexta}', "");
Expect(1, 19904, '\p{^incjkexta}', "");
Expect(1, 19904, '\P{incjkexta}', "");
Expect(0, 19904, '\P{^incjkexta}', "");
Expect(1, 19903, '\p{	_in_CJK_ext_a}', "");
Expect(0, 19903, '\p{^	_in_CJK_ext_a}', "");
Expect(0, 19903, '\P{	_in_CJK_ext_a}', "");
Expect(1, 19903, '\P{^	_in_CJK_ext_a}', "");
Expect(0, 19904, '\p{	_in_CJK_ext_a}', "");
Expect(1, 19904, '\p{^	_in_CJK_ext_a}', "");
Expect(1, 19904, '\P{	_in_CJK_ext_a}', "");
Expect(0, 19904, '\P{^	_in_CJK_ext_a}', "");
Error('\p{:= _supplementary_Private_USE_Area_A}');
Error('\P{:= _supplementary_Private_USE_Area_A}');
Expect(1, 983040, '\p{supplementaryprivateuseareaa}', "");
Expect(0, 983040, '\p{^supplementaryprivateuseareaa}', "");
Expect(0, 983040, '\P{supplementaryprivateuseareaa}', "");
Expect(1, 983040, '\P{^supplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\p{supplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\p{^supplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\P{supplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\P{^supplementaryprivateuseareaa}', "");
Expect(1, 983040, '\p{ Supplementary_PRIVATE_USE_Area_A}', "");
Expect(0, 983040, '\p{^ Supplementary_PRIVATE_USE_Area_A}', "");
Expect(0, 983040, '\P{ Supplementary_PRIVATE_USE_Area_A}', "");
Expect(1, 983040, '\P{^ Supplementary_PRIVATE_USE_Area_A}', "");
Expect(0, 1048576, '\p{ Supplementary_PRIVATE_USE_Area_A}', "");
Expect(1, 1048576, '\p{^ Supplementary_PRIVATE_USE_Area_A}', "");
Expect(1, 1048576, '\P{ Supplementary_PRIVATE_USE_Area_A}', "");
Expect(0, 1048576, '\P{^ Supplementary_PRIVATE_USE_Area_A}', "");
Error('\p{/a/_-is_SUPPLEMENTARY_Private_Use_AREA_A}');
Error('\P{/a/_-is_SUPPLEMENTARY_Private_Use_AREA_A}');
Expect(1, 983040, '\p{issupplementaryprivateuseareaa}', "");
Expect(0, 983040, '\p{^issupplementaryprivateuseareaa}', "");
Expect(0, 983040, '\P{issupplementaryprivateuseareaa}', "");
Expect(1, 983040, '\P{^issupplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\p{issupplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\p{^issupplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\P{issupplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\P{^issupplementaryprivateuseareaa}', "");
Expect(1, 983040, '\p{	 is_Supplementary_Private_use_Area_a}', "");
Expect(0, 983040, '\p{^	 is_Supplementary_Private_use_Area_a}', "");
Expect(0, 983040, '\P{	 is_Supplementary_Private_use_Area_a}', "");
Expect(1, 983040, '\P{^	 is_Supplementary_Private_use_Area_a}', "");
Expect(0, 1048576, '\p{	 is_Supplementary_Private_use_Area_a}', "");
Expect(1, 1048576, '\p{^	 is_Supplementary_Private_use_Area_a}', "");
Expect(1, 1048576, '\P{	 is_Supplementary_Private_use_Area_a}', "");
Expect(0, 1048576, '\P{^	 is_Supplementary_Private_use_Area_a}', "");
Error('\p{/a/In_Supplementary_private_use_AREA_A}');
Error('\P{/a/In_Supplementary_private_use_AREA_A}');
Expect(1, 983040, '\p{insupplementaryprivateuseareaa}', "");
Expect(0, 983040, '\p{^insupplementaryprivateuseareaa}', "");
Expect(0, 983040, '\P{insupplementaryprivateuseareaa}', "");
Expect(1, 983040, '\P{^insupplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\p{insupplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\p{^insupplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\P{insupplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\P{^insupplementaryprivateuseareaa}', "");
Expect(1, 983040, '\p{_In_Supplementary_Private_use_Area_A}', "");
Expect(0, 983040, '\p{^_In_Supplementary_Private_use_Area_A}', "");
Expect(0, 983040, '\P{_In_Supplementary_Private_use_Area_A}', "");
Expect(1, 983040, '\P{^_In_Supplementary_Private_use_Area_A}', "");
Expect(0, 1048576, '\p{_In_Supplementary_Private_use_Area_A}', "");
Expect(1, 1048576, '\p{^_In_Supplementary_Private_use_Area_A}', "");
Expect(1, 1048576, '\P{_In_Supplementary_Private_use_Area_A}', "");
Expect(0, 1048576, '\P{^_In_Supplementary_Private_use_Area_A}', "");
Error('\p{	/a/Sup_PUA_a}');
Error('\P{	/a/Sup_PUA_a}');
Expect(1, 983040, '\p{suppuaa}', "");
Expect(0, 983040, '\p{^suppuaa}', "");
Expect(0, 983040, '\P{suppuaa}', "");
Expect(1, 983040, '\P{^suppuaa}', "");
Expect(0, 1048576, '\p{suppuaa}', "");
Expect(1, 1048576, '\p{^suppuaa}', "");
Expect(1, 1048576, '\P{suppuaa}', "");
Expect(0, 1048576, '\P{^suppuaa}', "");
Expect(1, 983040, '\p{ 	Sup_PUA_A}', "");
Expect(0, 983040, '\p{^ 	Sup_PUA_A}', "");
Expect(0, 983040, '\P{ 	Sup_PUA_A}', "");
Expect(1, 983040, '\P{^ 	Sup_PUA_A}', "");
Expect(0, 1048576, '\p{ 	Sup_PUA_A}', "");
Expect(1, 1048576, '\p{^ 	Sup_PUA_A}', "");
Expect(1, 1048576, '\P{ 	Sup_PUA_A}', "");
Expect(0, 1048576, '\P{^ 	Sup_PUA_A}', "");
Error('\p{	/a/Is_Sup_PUA_A}');
Error('\P{	/a/Is_Sup_PUA_A}');
Expect(1, 983040, '\p{issuppuaa}', "");
Expect(0, 983040, '\p{^issuppuaa}', "");
Expect(0, 983040, '\P{issuppuaa}', "");
Expect(1, 983040, '\P{^issuppuaa}', "");
Expect(0, 1048576, '\p{issuppuaa}', "");
Expect(1, 1048576, '\p{^issuppuaa}', "");
Expect(1, 1048576, '\P{issuppuaa}', "");
Expect(0, 1048576, '\P{^issuppuaa}', "");
Expect(1, 983040, '\p{--IS_Sup_PUA_A}', "");
Expect(0, 983040, '\p{^--IS_Sup_PUA_A}', "");
Expect(0, 983040, '\P{--IS_Sup_PUA_A}', "");
Expect(1, 983040, '\P{^--IS_Sup_PUA_A}', "");
Expect(0, 1048576, '\p{--IS_Sup_PUA_A}', "");
Expect(1, 1048576, '\p{^--IS_Sup_PUA_A}', "");
Expect(1, 1048576, '\P{--IS_Sup_PUA_A}', "");
Expect(0, 1048576, '\P{^--IS_Sup_PUA_A}', "");
Error('\p{-_In_SUP_PUA_A:=}');
Error('\P{-_In_SUP_PUA_A:=}');
Expect(1, 983040, '\p{insuppuaa}', "");
Expect(0, 983040, '\p{^insuppuaa}', "");
Expect(0, 983040, '\P{insuppuaa}', "");
Expect(1, 983040, '\P{^insuppuaa}', "");
Expect(0, 1048576, '\p{insuppuaa}', "");
Expect(1, 1048576, '\p{^insuppuaa}', "");
Expect(1, 1048576, '\P{insuppuaa}', "");
Expect(0, 1048576, '\P{^insuppuaa}', "");
Expect(1, 983040, '\p{ in_Sup_PUA_A}', "");
Expect(0, 983040, '\p{^ in_Sup_PUA_A}', "");
Expect(0, 983040, '\P{ in_Sup_PUA_A}', "");
Expect(1, 983040, '\P{^ in_Sup_PUA_A}', "");
Expect(0, 1048576, '\p{ in_Sup_PUA_A}', "");
Expect(1, 1048576, '\p{^ in_Sup_PUA_A}', "");
Expect(1, 1048576, '\P{ in_Sup_PUA_A}', "");
Expect(0, 1048576, '\P{^ in_Sup_PUA_A}', "");
Error('\p{_:=PHONETIC_Extensions}');
Error('\P{_:=PHONETIC_Extensions}');
Expect(1, 7551, '\p{phoneticextensions}', "");
Expect(0, 7551, '\p{^phoneticextensions}', "");
Expect(0, 7551, '\P{phoneticextensions}', "");
Expect(1, 7551, '\P{^phoneticextensions}', "");
Expect(0, 7552, '\p{phoneticextensions}', "");
Expect(1, 7552, '\p{^phoneticextensions}', "");
Expect(1, 7552, '\P{phoneticextensions}', "");
Expect(0, 7552, '\P{^phoneticextensions}', "");
Expect(1, 7551, '\p{- Phonetic_extensions}', "");
Expect(0, 7551, '\p{^- Phonetic_extensions}', "");
Expect(0, 7551, '\P{- Phonetic_extensions}', "");
Expect(1, 7551, '\P{^- Phonetic_extensions}', "");
Expect(0, 7552, '\p{- Phonetic_extensions}', "");
Expect(1, 7552, '\p{^- Phonetic_extensions}', "");
Expect(1, 7552, '\P{- Phonetic_extensions}', "");
Expect(0, 7552, '\P{^- Phonetic_extensions}', "");
Error('\p{-/a/IS_Phonetic_Extensions}');
Error('\P{-/a/IS_Phonetic_Extensions}');
Expect(1, 7551, '\p{isphoneticextensions}', "");
Expect(0, 7551, '\p{^isphoneticextensions}', "");
Expect(0, 7551, '\P{isphoneticextensions}', "");
Expect(1, 7551, '\P{^isphoneticextensions}', "");
Expect(0, 7552, '\p{isphoneticextensions}', "");
Expect(1, 7552, '\p{^isphoneticextensions}', "");
Expect(1, 7552, '\P{isphoneticextensions}', "");
Expect(0, 7552, '\P{^isphoneticextensions}', "");
Expect(1, 7551, '\p{- IS_Phonetic_Extensions}', "");
Expect(0, 7551, '\p{^- IS_Phonetic_Extensions}', "");
Expect(0, 7551, '\P{- IS_Phonetic_Extensions}', "");
Expect(1, 7551, '\P{^- IS_Phonetic_Extensions}', "");
Expect(0, 7552, '\p{- IS_Phonetic_Extensions}', "");
Expect(1, 7552, '\p{^- IS_Phonetic_Extensions}', "");
Expect(1, 7552, '\P{- IS_Phonetic_Extensions}', "");
Expect(0, 7552, '\P{^- IS_Phonetic_Extensions}', "");
Error('\p{/a/-in_Phonetic_extensions}');
Error('\P{/a/-in_Phonetic_extensions}');
Expect(1, 7551, '\p{inphoneticextensions}', "");
Expect(0, 7551, '\p{^inphoneticextensions}', "");
Expect(0, 7551, '\P{inphoneticextensions}', "");
Expect(1, 7551, '\P{^inphoneticextensions}', "");
Expect(0, 7552, '\p{inphoneticextensions}', "");
Expect(1, 7552, '\p{^inphoneticextensions}', "");
Expect(1, 7552, '\P{inphoneticextensions}', "");
Expect(0, 7552, '\P{^inphoneticextensions}', "");
Expect(1, 7551, '\p{-IN_Phonetic_extensions}', "");
Expect(0, 7551, '\p{^-IN_Phonetic_extensions}', "");
Expect(0, 7551, '\P{-IN_Phonetic_extensions}', "");
Expect(1, 7551, '\P{^-IN_Phonetic_extensions}', "");
Expect(0, 7552, '\p{-IN_Phonetic_extensions}', "");
Expect(1, 7552, '\p{^-IN_Phonetic_extensions}', "");
Expect(1, 7552, '\P{-IN_Phonetic_extensions}', "");
Expect(0, 7552, '\P{^-IN_Phonetic_extensions}', "");
Error('\p{:=_ PHONETIC_Ext}');
Error('\P{:=_ PHONETIC_Ext}');
Expect(1, 7551, '\p{phoneticext}', "");
Expect(0, 7551, '\p{^phoneticext}', "");
Expect(0, 7551, '\P{phoneticext}', "");
Expect(1, 7551, '\P{^phoneticext}', "");
Expect(0, 7552, '\p{phoneticext}', "");
Expect(1, 7552, '\p{^phoneticext}', "");
Expect(1, 7552, '\P{phoneticext}', "");
Expect(0, 7552, '\P{^phoneticext}', "");
Expect(1, 7551, '\p{Phonetic_EXT}', "");
Expect(0, 7551, '\p{^Phonetic_EXT}', "");
Expect(0, 7551, '\P{Phonetic_EXT}', "");
Expect(1, 7551, '\P{^Phonetic_EXT}', "");
Expect(0, 7552, '\p{Phonetic_EXT}', "");
Expect(1, 7552, '\p{^Phonetic_EXT}', "");
Expect(1, 7552, '\P{Phonetic_EXT}', "");
Expect(0, 7552, '\P{^Phonetic_EXT}', "");
Error('\p{:=_Is_Phonetic_ext}');
Error('\P{:=_Is_Phonetic_ext}');
Expect(1, 7551, '\p{isphoneticext}', "");
Expect(0, 7551, '\p{^isphoneticext}', "");
Expect(0, 7551, '\P{isphoneticext}', "");
Expect(1, 7551, '\P{^isphoneticext}', "");
Expect(0, 7552, '\p{isphoneticext}', "");
Expect(1, 7552, '\p{^isphoneticext}', "");
Expect(1, 7552, '\P{isphoneticext}', "");
Expect(0, 7552, '\P{^isphoneticext}', "");
Expect(1, 7551, '\p{_Is_Phonetic_EXT}', "");
Expect(0, 7551, '\p{^_Is_Phonetic_EXT}', "");
Expect(0, 7551, '\P{_Is_Phonetic_EXT}', "");
Expect(1, 7551, '\P{^_Is_Phonetic_EXT}', "");
Expect(0, 7552, '\p{_Is_Phonetic_EXT}', "");
Expect(1, 7552, '\p{^_Is_Phonetic_EXT}', "");
Expect(1, 7552, '\P{_Is_Phonetic_EXT}', "");
Expect(0, 7552, '\P{^_Is_Phonetic_EXT}', "");
Error('\p{	/a/in_Phonetic_EXT}');
Error('\P{	/a/in_Phonetic_EXT}');
Expect(1, 7551, '\p{inphoneticext}', "");
Expect(0, 7551, '\p{^inphoneticext}', "");
Expect(0, 7551, '\P{inphoneticext}', "");
Expect(1, 7551, '\P{^inphoneticext}', "");
Expect(0, 7552, '\p{inphoneticext}', "");
Expect(1, 7552, '\p{^inphoneticext}', "");
Expect(1, 7552, '\P{inphoneticext}', "");
Expect(0, 7552, '\P{^inphoneticext}', "");
Expect(1, 7551, '\p{	-In_Phonetic_ext}', "");
Expect(0, 7551, '\p{^	-In_Phonetic_ext}', "");
Expect(0, 7551, '\P{	-In_Phonetic_ext}', "");
Expect(1, 7551, '\P{^	-In_Phonetic_ext}', "");
Expect(0, 7552, '\p{	-In_Phonetic_ext}', "");
Expect(1, 7552, '\p{^	-In_Phonetic_ext}', "");
Expect(1, 7552, '\P{	-In_Phonetic_ext}', "");
Expect(0, 7552, '\P{^	-In_Phonetic_ext}', "");
Error('\p{__New_TAI_Lue:=}');
Error('\P{__New_TAI_Lue:=}');
Expect(1, 6623, '\p{newtailue}', "");
Expect(0, 6623, '\p{^newtailue}', "");
Expect(0, 6623, '\P{newtailue}', "");
Expect(1, 6623, '\P{^newtailue}', "");
Expect(0, 6624, '\p{newtailue}', "");
Expect(1, 6624, '\p{^newtailue}', "");
Expect(1, 6624, '\P{newtailue}', "");
Expect(0, 6624, '\P{^newtailue}', "");
Expect(1, 6623, '\p{_-new_TAI_lue}', "");
Expect(0, 6623, '\p{^_-new_TAI_lue}', "");
Expect(0, 6623, '\P{_-new_TAI_lue}', "");
Expect(1, 6623, '\P{^_-new_TAI_lue}', "");
Expect(0, 6624, '\p{_-new_TAI_lue}', "");
Expect(1, 6624, '\p{^_-new_TAI_lue}', "");
Expect(1, 6624, '\P{_-new_TAI_lue}', "");
Expect(0, 6624, '\P{^_-new_TAI_lue}', "");
Error('\p{/a/	Is_new_Tai_Lue}');
Error('\P{/a/	Is_new_Tai_Lue}');
Expect(1, 6623, '\p{isnewtailue}', "");
Expect(0, 6623, '\p{^isnewtailue}', "");
Expect(0, 6623, '\P{isnewtailue}', "");
Expect(1, 6623, '\P{^isnewtailue}', "");
Expect(0, 6624, '\p{isnewtailue}', "");
Expect(1, 6624, '\p{^isnewtailue}', "");
Expect(1, 6624, '\P{isnewtailue}', "");
Expect(0, 6624, '\P{^isnewtailue}', "");
Expect(1, 6623, '\p{ is_New_Tai_Lue}', "");
Expect(0, 6623, '\p{^ is_New_Tai_Lue}', "");
Expect(0, 6623, '\P{ is_New_Tai_Lue}', "");
Expect(1, 6623, '\P{^ is_New_Tai_Lue}', "");
Expect(0, 6624, '\p{ is_New_Tai_Lue}', "");
Expect(1, 6624, '\p{^ is_New_Tai_Lue}', "");
Expect(1, 6624, '\P{ is_New_Tai_Lue}', "");
Expect(0, 6624, '\P{^ is_New_Tai_Lue}', "");
Error('\p{-Talu:=}');
Error('\P{-Talu:=}');
Expect(1, 6623, '\p{talu}', "");
Expect(0, 6623, '\p{^talu}', "");
Expect(0, 6623, '\P{talu}', "");
Expect(1, 6623, '\P{^talu}', "");
Expect(0, 6624, '\p{talu}', "");
Expect(1, 6624, '\p{^talu}', "");
Expect(1, 6624, '\P{talu}', "");
Expect(0, 6624, '\P{^talu}', "");
Expect(1, 6623, '\p{	 TALU}', "");
Expect(0, 6623, '\p{^	 TALU}', "");
Expect(0, 6623, '\P{	 TALU}', "");
Expect(1, 6623, '\P{^	 TALU}', "");
Expect(0, 6624, '\p{	 TALU}', "");
Expect(1, 6624, '\p{^	 TALU}', "");
Expect(1, 6624, '\P{	 TALU}', "");
Expect(0, 6624, '\P{^	 TALU}', "");
Error('\p{ /a/Is_Talu}');
Error('\P{ /a/Is_Talu}');
Expect(1, 6623, '\p{istalu}', "");
Expect(0, 6623, '\p{^istalu}', "");
Expect(0, 6623, '\P{istalu}', "");
Expect(1, 6623, '\P{^istalu}', "");
Expect(0, 6624, '\p{istalu}', "");
Expect(1, 6624, '\p{^istalu}', "");
Expect(1, 6624, '\P{istalu}', "");
Expect(0, 6624, '\P{^istalu}', "");
Expect(1, 6623, '\p{_IS_TALU}', "");
Expect(0, 6623, '\p{^_IS_TALU}', "");
Expect(0, 6623, '\P{_IS_TALU}', "");
Expect(1, 6623, '\P{^_IS_TALU}', "");
Expect(0, 6624, '\p{_IS_TALU}', "");
Expect(1, 6624, '\p{^_IS_TALU}', "");
Expect(1, 6624, '\P{_IS_TALU}', "");
Expect(0, 6624, '\P{^_IS_TALU}', "");
Error('\p{		UNASSIGNED:=}');
Error('\P{		UNASSIGNED:=}');
Expect(1, 918000, '\p{unassigned}', "");
Expect(0, 918000, '\p{^unassigned}', "");
Expect(0, 918000, '\P{unassigned}', "");
Expect(1, 918000, '\P{^unassigned}', "");
Expect(0, 1114109, '\p{unassigned}', "");
Expect(1, 1114109, '\p{^unassigned}', "");
Expect(1, 1114109, '\P{unassigned}', "");
Expect(0, 1114109, '\P{^unassigned}', "");
Expect(1, 918000, '\p{__UNASSIGNED}', "");
Expect(0, 918000, '\p{^__UNASSIGNED}', "");
Expect(0, 918000, '\P{__UNASSIGNED}', "");
Expect(1, 918000, '\P{^__UNASSIGNED}', "");
Expect(0, 1114109, '\p{__UNASSIGNED}', "");
Expect(1, 1114109, '\p{^__UNASSIGNED}', "");
Expect(1, 1114109, '\P{__UNASSIGNED}', "");
Expect(0, 1114109, '\P{^__UNASSIGNED}', "");
Error('\p{- Is_unassigned:=}');
Error('\P{- Is_unassigned:=}');
Expect(1, 918000, '\p{isunassigned}', "");
Expect(0, 918000, '\p{^isunassigned}', "");
Expect(0, 918000, '\P{isunassigned}', "");
Expect(1, 918000, '\P{^isunassigned}', "");
Expect(0, 1114109, '\p{isunassigned}', "");
Expect(1, 1114109, '\p{^isunassigned}', "");
Expect(1, 1114109, '\P{isunassigned}', "");
Expect(0, 1114109, '\P{^isunassigned}', "");
Expect(1, 918000, '\p{ _IS_Unassigned}', "");
Expect(0, 918000, '\p{^ _IS_Unassigned}', "");
Expect(0, 918000, '\P{ _IS_Unassigned}', "");
Expect(1, 918000, '\P{^ _IS_Unassigned}', "");
Expect(0, 1114109, '\p{ _IS_Unassigned}', "");
Expect(1, 1114109, '\p{^ _IS_Unassigned}', "");
Expect(1, 1114109, '\P{ _IS_Unassigned}', "");
Expect(0, 1114109, '\P{^ _IS_Unassigned}', "");
Error('\p{	CN:=}');
Error('\P{	CN:=}');
Expect(1, 918000, '\p{cn}', "");
Expect(0, 918000, '\p{^cn}', "");
Expect(0, 918000, '\P{cn}', "");
Expect(1, 918000, '\P{^cn}', "");
Expect(0, 1114109, '\p{cn}', "");
Expect(1, 1114109, '\p{^cn}', "");
Expect(1, 1114109, '\P{cn}', "");
Expect(0, 1114109, '\P{^cn}', "");
Expect(1, 918000, '\p{	 Cn}', "");
Expect(0, 918000, '\p{^	 Cn}', "");
Expect(0, 918000, '\P{	 Cn}', "");
Expect(1, 918000, '\P{^	 Cn}', "");
Expect(0, 1114109, '\p{	 Cn}', "");
Expect(1, 1114109, '\p{^	 Cn}', "");
Expect(1, 1114109, '\P{	 Cn}', "");
Expect(0, 1114109, '\P{^	 Cn}', "");
Error('\p{/a/-_IS_CN}');
Error('\P{/a/-_IS_CN}');
Expect(1, 918000, '\p{iscn}', "");
Expect(0, 918000, '\p{^iscn}', "");
Expect(0, 918000, '\P{iscn}', "");
Expect(1, 918000, '\P{^iscn}', "");
Expect(0, 1114109, '\p{iscn}', "");
Expect(1, 1114109, '\p{^iscn}', "");
Expect(1, 1114109, '\P{iscn}', "");
Expect(0, 1114109, '\P{^iscn}', "");
Expect(1, 918000, '\p{	_Is_cn}', "");
Expect(0, 918000, '\p{^	_Is_cn}', "");
Expect(0, 918000, '\P{	_Is_cn}', "");
Expect(1, 918000, '\P{^	_Is_cn}', "");
Expect(0, 1114109, '\p{	_Is_cn}', "");
Expect(1, 1114109, '\p{^	_Is_cn}', "");
Expect(1, 1114109, '\P{	_Is_cn}', "");
Expect(0, 1114109, '\P{^	_Is_cn}', "");
Error('\p{_-LEPCHA/a/}');
Error('\P{_-LEPCHA/a/}');
Expect(1, 7247, '\p{lepcha}', "");
Expect(0, 7247, '\p{^lepcha}', "");
Expect(0, 7247, '\P{lepcha}', "");
Expect(1, 7247, '\P{^lepcha}', "");
Expect(0, 7248, '\p{lepcha}', "");
Expect(1, 7248, '\p{^lepcha}', "");
Expect(1, 7248, '\P{lepcha}', "");
Expect(0, 7248, '\P{^lepcha}', "");
Expect(1, 7247, '\p{  lepcha}', "");
Expect(0, 7247, '\p{^  lepcha}', "");
Expect(0, 7247, '\P{  lepcha}', "");
Expect(1, 7247, '\P{^  lepcha}', "");
Expect(0, 7248, '\p{  lepcha}', "");
Expect(1, 7248, '\p{^  lepcha}', "");
Expect(1, 7248, '\P{  lepcha}', "");
Expect(0, 7248, '\P{^  lepcha}', "");
Error('\p{-_Is_lepcha/a/}');
Error('\P{-_Is_lepcha/a/}');
Expect(1, 7247, '\p{islepcha}', "");
Expect(0, 7247, '\p{^islepcha}', "");
Expect(0, 7247, '\P{islepcha}', "");
Expect(1, 7247, '\P{^islepcha}', "");
Expect(0, 7248, '\p{islepcha}', "");
Expect(1, 7248, '\p{^islepcha}', "");
Expect(1, 7248, '\P{islepcha}', "");
Expect(0, 7248, '\P{^islepcha}', "");
Expect(1, 7247, '\p{	Is_LEPCHA}', "");
Expect(0, 7247, '\p{^	Is_LEPCHA}', "");
Expect(0, 7247, '\P{	Is_LEPCHA}', "");
Expect(1, 7247, '\P{^	Is_LEPCHA}', "");
Expect(0, 7248, '\p{	Is_LEPCHA}', "");
Expect(1, 7248, '\p{^	Is_LEPCHA}', "");
Expect(1, 7248, '\P{	Is_LEPCHA}', "");
Expect(0, 7248, '\P{^	Is_LEPCHA}', "");
Error('\p{_-LEPC:=}');
Error('\P{_-LEPC:=}');
Expect(1, 7247, '\p{lepc}', "");
Expect(0, 7247, '\p{^lepc}', "");
Expect(0, 7247, '\P{lepc}', "");
Expect(1, 7247, '\P{^lepc}', "");
Expect(0, 7248, '\p{lepc}', "");
Expect(1, 7248, '\p{^lepc}', "");
Expect(1, 7248, '\P{lepc}', "");
Expect(0, 7248, '\P{^lepc}', "");
Expect(1, 7247, '\p{		lepc}', "");
Expect(0, 7247, '\p{^		lepc}', "");
Expect(0, 7247, '\P{		lepc}', "");
Expect(1, 7247, '\P{^		lepc}', "");
Expect(0, 7248, '\p{		lepc}', "");
Expect(1, 7248, '\p{^		lepc}', "");
Expect(1, 7248, '\P{		lepc}', "");
Expect(0, 7248, '\P{^		lepc}', "");
Error('\p{ :=IS_Lepc}');
Error('\P{ :=IS_Lepc}');
Expect(1, 7247, '\p{islepc}', "");
Expect(0, 7247, '\p{^islepc}', "");
Expect(0, 7247, '\P{islepc}', "");
Expect(1, 7247, '\P{^islepc}', "");
Expect(0, 7248, '\p{islepc}', "");
Expect(1, 7248, '\p{^islepc}', "");
Expect(1, 7248, '\P{islepc}', "");
Expect(0, 7248, '\P{^islepc}', "");
Expect(1, 7247, '\p{_Is_Lepc}', "");
Expect(0, 7247, '\p{^_Is_Lepc}', "");
Expect(0, 7247, '\P{_Is_Lepc}', "");
Expect(1, 7247, '\P{^_Is_Lepc}', "");
Expect(0, 7248, '\p{_Is_Lepc}', "");
Expect(1, 7248, '\p{^_Is_Lepc}', "");
Expect(1, 7248, '\P{_Is_Lepc}', "");
Expect(0, 7248, '\P{^_Is_Lepc}', "");
Error('\p{/a/-	Ideographic_Description_characters}');
Error('\P{/a/-	Ideographic_Description_characters}');
Expect(1, 12287, '\p{ideographicdescriptioncharacters}', "");
Expect(0, 12287, '\p{^ideographicdescriptioncharacters}', "");
Expect(0, 12287, '\P{ideographicdescriptioncharacters}', "");
Expect(1, 12287, '\P{^ideographicdescriptioncharacters}', "");
Expect(0, 12288, '\p{ideographicdescriptioncharacters}', "");
Expect(1, 12288, '\p{^ideographicdescriptioncharacters}', "");
Expect(1, 12288, '\P{ideographicdescriptioncharacters}', "");
Expect(0, 12288, '\P{^ideographicdescriptioncharacters}', "");
Expect(1, 12287, '\p{ 	ideographic_DESCRIPTION_Characters}', "");
Expect(0, 12287, '\p{^ 	ideographic_DESCRIPTION_Characters}', "");
Expect(0, 12287, '\P{ 	ideographic_DESCRIPTION_Characters}', "");
Expect(1, 12287, '\P{^ 	ideographic_DESCRIPTION_Characters}', "");
Expect(0, 12288, '\p{ 	ideographic_DESCRIPTION_Characters}', "");
Expect(1, 12288, '\p{^ 	ideographic_DESCRIPTION_Characters}', "");
Expect(1, 12288, '\P{ 	ideographic_DESCRIPTION_Characters}', "");
Expect(0, 12288, '\P{^ 	ideographic_DESCRIPTION_Characters}', "");
Error('\p{/a/	Is_ideographic_Description_characters}');
Error('\P{/a/	Is_ideographic_Description_characters}');
Expect(1, 12287, '\p{isideographicdescriptioncharacters}', "");
Expect(0, 12287, '\p{^isideographicdescriptioncharacters}', "");
Expect(0, 12287, '\P{isideographicdescriptioncharacters}', "");
Expect(1, 12287, '\P{^isideographicdescriptioncharacters}', "");
Expect(0, 12288, '\p{isideographicdescriptioncharacters}', "");
Expect(1, 12288, '\p{^isideographicdescriptioncharacters}', "");
Expect(1, 12288, '\P{isideographicdescriptioncharacters}', "");
Expect(0, 12288, '\P{^isideographicdescriptioncharacters}', "");
Expect(1, 12287, '\p{ _Is_Ideographic_DESCRIPTION_CHARACTERS}', "");
Expect(0, 12287, '\p{^ _Is_Ideographic_DESCRIPTION_CHARACTERS}', "");
Expect(0, 12287, '\P{ _Is_Ideographic_DESCRIPTION_CHARACTERS}', "");
Expect(1, 12287, '\P{^ _Is_Ideographic_DESCRIPTION_CHARACTERS}', "");
Expect(0, 12288, '\p{ _Is_Ideographic_DESCRIPTION_CHARACTERS}', "");
Expect(1, 12288, '\p{^ _Is_Ideographic_DESCRIPTION_CHARACTERS}', "");
Expect(1, 12288, '\P{ _Is_Ideographic_DESCRIPTION_CHARACTERS}', "");
Expect(0, 12288, '\P{^ _Is_Ideographic_DESCRIPTION_CHARACTERS}', "");
Error('\p{-:=In_Ideographic_Description_Characters}');
Error('\P{-:=In_Ideographic_Description_Characters}');
Expect(1, 12287, '\p{inideographicdescriptioncharacters}', "");
Expect(0, 12287, '\p{^inideographicdescriptioncharacters}', "");
Expect(0, 12287, '\P{inideographicdescriptioncharacters}', "");
Expect(1, 12287, '\P{^inideographicdescriptioncharacters}', "");
Expect(0, 12288, '\p{inideographicdescriptioncharacters}', "");
Expect(1, 12288, '\p{^inideographicdescriptioncharacters}', "");
Expect(1, 12288, '\P{inideographicdescriptioncharacters}', "");
Expect(0, 12288, '\P{^inideographicdescriptioncharacters}', "");
Expect(1, 12287, '\p{ In_IDEOGRAPHIC_Description_Characters}', "");
Expect(0, 12287, '\p{^ In_IDEOGRAPHIC_Description_Characters}', "");
Expect(0, 12287, '\P{ In_IDEOGRAPHIC_Description_Characters}', "");
Expect(1, 12287, '\P{^ In_IDEOGRAPHIC_Description_Characters}', "");
Expect(0, 12288, '\p{ In_IDEOGRAPHIC_Description_Characters}', "");
Expect(1, 12288, '\p{^ In_IDEOGRAPHIC_Description_Characters}', "");
Expect(1, 12288, '\P{ In_IDEOGRAPHIC_Description_Characters}', "");
Expect(0, 12288, '\P{^ In_IDEOGRAPHIC_Description_Characters}', "");
Error('\p{/a/	 IN_IDC}');
Error('\P{/a/	 IN_IDC}');
Expect(1, 12287, '\p{inidc}', "");
Expect(0, 12287, '\p{^inidc}', "");
Expect(0, 12287, '\P{inidc}', "");
Expect(1, 12287, '\P{^inidc}', "");
Expect(0, 12288, '\p{inidc}', "");
Expect(1, 12288, '\p{^inidc}', "");
Expect(1, 12288, '\P{inidc}', "");
Expect(0, 12288, '\P{^inidc}', "");
Expect(1, 12287, '\p{-_in_IDC}', "");
Expect(0, 12287, '\p{^-_in_IDC}', "");
Expect(0, 12287, '\P{-_in_IDC}', "");
Expect(1, 12287, '\P{^-_in_IDC}', "");
Expect(0, 12288, '\p{-_in_IDC}', "");
Expect(1, 12288, '\p{^-_in_IDC}', "");
Expect(1, 12288, '\P{-_in_IDC}', "");
Expect(0, 12288, '\P{^-_in_IDC}', "");
Error('\p{__ORIYA/a/}');
Error('\P{__ORIYA/a/}');
Expect(1, 2935, '\p{oriya}', "");
Expect(0, 2935, '\p{^oriya}', "");
Expect(0, 2935, '\P{oriya}', "");
Expect(1, 2935, '\P{^oriya}', "");
Expect(0, 2936, '\p{oriya}', "");
Expect(1, 2936, '\p{^oriya}', "");
Expect(1, 2936, '\P{oriya}', "");
Expect(0, 2936, '\P{^oriya}', "");
Expect(1, 2935, '\p{  ORIYA}', "");
Expect(0, 2935, '\p{^  ORIYA}', "");
Expect(0, 2935, '\P{  ORIYA}', "");
Expect(1, 2935, '\P{^  ORIYA}', "");
Expect(0, 2936, '\p{  ORIYA}', "");
Expect(1, 2936, '\p{^  ORIYA}', "");
Expect(1, 2936, '\P{  ORIYA}', "");
Expect(0, 2936, '\P{^  ORIYA}', "");
Error('\p{ -is_oriya/a/}');
Error('\P{ -is_oriya/a/}');
Expect(1, 2935, '\p{isoriya}', "");
Expect(0, 2935, '\p{^isoriya}', "");
Expect(0, 2935, '\P{isoriya}', "");
Expect(1, 2935, '\P{^isoriya}', "");
Expect(0, 2936, '\p{isoriya}', "");
Expect(1, 2936, '\p{^isoriya}', "");
Expect(1, 2936, '\P{isoriya}', "");
Expect(0, 2936, '\P{^isoriya}', "");
Expect(1, 2935, '\p{_ Is_Oriya}', "");
Expect(0, 2935, '\p{^_ Is_Oriya}', "");
Expect(0, 2935, '\P{_ Is_Oriya}', "");
Expect(1, 2935, '\P{^_ Is_Oriya}', "");
Expect(0, 2936, '\p{_ Is_Oriya}', "");
Expect(1, 2936, '\p{^_ Is_Oriya}', "");
Expect(1, 2936, '\P{_ Is_Oriya}', "");
Expect(0, 2936, '\P{^_ Is_Oriya}', "");
Error('\p{/a/ orya}');
Error('\P{/a/ orya}');
Expect(1, 2935, '\p{orya}', "");
Expect(0, 2935, '\p{^orya}', "");
Expect(0, 2935, '\P{orya}', "");
Expect(1, 2935, '\P{^orya}', "");
Expect(0, 2936, '\p{orya}', "");
Expect(1, 2936, '\p{^orya}', "");
Expect(1, 2936, '\P{orya}', "");
Expect(0, 2936, '\P{^orya}', "");
Expect(1, 2935, '\p{_ Orya}', "");
Expect(0, 2935, '\p{^_ Orya}', "");
Expect(0, 2935, '\P{_ Orya}', "");
Expect(1, 2935, '\P{^_ Orya}', "");
Expect(0, 2936, '\p{_ Orya}', "");
Expect(1, 2936, '\p{^_ Orya}', "");
Expect(1, 2936, '\P{_ Orya}', "");
Expect(0, 2936, '\P{^_ Orya}', "");
Error('\p{:=-_Is_orya}');
Error('\P{:=-_Is_orya}');
Expect(1, 2935, '\p{isorya}', "");
Expect(0, 2935, '\p{^isorya}', "");
Expect(0, 2935, '\P{isorya}', "");
Expect(1, 2935, '\P{^isorya}', "");
Expect(0, 2936, '\p{isorya}', "");
Expect(1, 2936, '\p{^isorya}', "");
Expect(1, 2936, '\P{isorya}', "");
Expect(0, 2936, '\P{^isorya}', "");
Expect(1, 2935, '\p{	-is_orya}', "");
Expect(0, 2935, '\p{^	-is_orya}', "");
Expect(0, 2935, '\P{	-is_orya}', "");
Expect(1, 2935, '\P{^	-is_orya}', "");
Expect(0, 2936, '\p{	-is_orya}', "");
Expect(1, 2936, '\p{^	-is_orya}', "");
Expect(1, 2936, '\P{	-is_orya}', "");
Expect(0, 2936, '\P{^	-is_orya}', "");
Error('\p{/a/	 Grapheme_Extend}');
Error('\P{/a/	 Grapheme_Extend}');
Expect(1, 917999, '\p{graphemeextend}', "");
Expect(0, 917999, '\p{^graphemeextend}', "");
Expect(0, 917999, '\P{graphemeextend}', "");
Expect(1, 917999, '\P{^graphemeextend}', "");
Expect(0, 918000, '\p{graphemeextend}', "");
Expect(1, 918000, '\p{^graphemeextend}', "");
Expect(1, 918000, '\P{graphemeextend}', "");
Expect(0, 918000, '\P{^graphemeextend}', "");
Expect(1, 917999, '\p{_grapheme_Extend}', "");
Expect(0, 917999, '\p{^_grapheme_Extend}', "");
Expect(0, 917999, '\P{_grapheme_Extend}', "");
Expect(1, 917999, '\P{^_grapheme_Extend}', "");
Expect(0, 918000, '\p{_grapheme_Extend}', "");
Expect(1, 918000, '\p{^_grapheme_Extend}', "");
Expect(1, 918000, '\P{_grapheme_Extend}', "");
Expect(0, 918000, '\P{^_grapheme_Extend}', "");
Error('\p{/a/-	Is_GRAPHEME_extend}');
Error('\P{/a/-	Is_GRAPHEME_extend}');
Expect(1, 917999, '\p{isgraphemeextend}', "");
Expect(0, 917999, '\p{^isgraphemeextend}', "");
Expect(0, 917999, '\P{isgraphemeextend}', "");
Expect(1, 917999, '\P{^isgraphemeextend}', "");
Expect(0, 918000, '\p{isgraphemeextend}', "");
Expect(1, 918000, '\p{^isgraphemeextend}', "");
Expect(1, 918000, '\P{isgraphemeextend}', "");
Expect(0, 918000, '\P{^isgraphemeextend}', "");
Expect(1, 917999, '\p{ IS_GRAPHEME_extend}', "");
Expect(0, 917999, '\p{^ IS_GRAPHEME_extend}', "");
Expect(0, 917999, '\P{ IS_GRAPHEME_extend}', "");
Expect(1, 917999, '\P{^ IS_GRAPHEME_extend}', "");
Expect(0, 918000, '\p{ IS_GRAPHEME_extend}', "");
Expect(1, 918000, '\p{^ IS_GRAPHEME_extend}', "");
Expect(1, 918000, '\P{ IS_GRAPHEME_extend}', "");
Expect(0, 918000, '\P{^ IS_GRAPHEME_extend}', "");
Error('\p{:=_	GR_EXT}');
Error('\P{:=_	GR_EXT}');
Expect(1, 917999, '\p{grext}', "");
Expect(0, 917999, '\p{^grext}', "");
Expect(0, 917999, '\P{grext}', "");
Expect(1, 917999, '\P{^grext}', "");
Expect(0, 918000, '\p{grext}', "");
Expect(1, 918000, '\p{^grext}', "");
Expect(1, 918000, '\P{grext}', "");
Expect(0, 918000, '\P{^grext}', "");
Expect(1, 917999, '\p{_-GR_EXT}', "");
Expect(0, 917999, '\p{^_-GR_EXT}', "");
Expect(0, 917999, '\P{_-GR_EXT}', "");
Expect(1, 917999, '\P{^_-GR_EXT}', "");
Expect(0, 918000, '\p{_-GR_EXT}', "");
Expect(1, 918000, '\p{^_-GR_EXT}', "");
Expect(1, 918000, '\P{_-GR_EXT}', "");
Expect(0, 918000, '\P{^_-GR_EXT}', "");
Error('\p{--Is_GR_ext:=}');
Error('\P{--Is_GR_ext:=}');
Expect(1, 917999, '\p{isgrext}', "");
Expect(0, 917999, '\p{^isgrext}', "");
Expect(0, 917999, '\P{isgrext}', "");
Expect(1, 917999, '\P{^isgrext}', "");
Expect(0, 918000, '\p{isgrext}', "");
Expect(1, 918000, '\p{^isgrext}', "");
Expect(1, 918000, '\P{isgrext}', "");
Expect(0, 918000, '\P{^isgrext}', "");
Expect(1, 917999, '\p{ _Is_Gr_EXT}', "");
Expect(0, 917999, '\p{^ _Is_Gr_EXT}', "");
Expect(0, 917999, '\P{ _Is_Gr_EXT}', "");
Expect(1, 917999, '\P{^ _Is_Gr_EXT}', "");
Expect(0, 918000, '\p{ _Is_Gr_EXT}', "");
Expect(1, 918000, '\p{^ _Is_Gr_EXT}', "");
Expect(1, 918000, '\P{ _Is_Gr_EXT}', "");
Expect(0, 918000, '\P{^ _Is_Gr_EXT}', "");
Error('\p{ Default_Ignorable_Code_POINT:=}');
Error('\P{ Default_Ignorable_Code_POINT:=}');
Expect(1, 921599, '\p{defaultignorablecodepoint}', "");
Expect(0, 921599, '\p{^defaultignorablecodepoint}', "");
Expect(0, 921599, '\P{defaultignorablecodepoint}', "");
Expect(1, 921599, '\P{^defaultignorablecodepoint}', "");
Expect(0, 921600, '\p{defaultignorablecodepoint}', "");
Expect(1, 921600, '\p{^defaultignorablecodepoint}', "");
Expect(1, 921600, '\P{defaultignorablecodepoint}', "");
Expect(0, 921600, '\P{^defaultignorablecodepoint}', "");
Expect(1, 921599, '\p{-_Default_Ignorable_code_POINT}', "");
Expect(0, 921599, '\p{^-_Default_Ignorable_code_POINT}', "");
Expect(0, 921599, '\P{-_Default_Ignorable_code_POINT}', "");
Expect(1, 921599, '\P{^-_Default_Ignorable_code_POINT}', "");
Expect(0, 921600, '\p{-_Default_Ignorable_code_POINT}', "");
Expect(1, 921600, '\p{^-_Default_Ignorable_code_POINT}', "");
Expect(1, 921600, '\P{-_Default_Ignorable_code_POINT}', "");
Expect(0, 921600, '\P{^-_Default_Ignorable_code_POINT}', "");
Error('\p{-Is_DEFAULT_IGNORABLE_Code_Point/a/}');
Error('\P{-Is_DEFAULT_IGNORABLE_Code_Point/a/}');
Expect(1, 921599, '\p{isdefaultignorablecodepoint}', "");
Expect(0, 921599, '\p{^isdefaultignorablecodepoint}', "");
Expect(0, 921599, '\P{isdefaultignorablecodepoint}', "");
Expect(1, 921599, '\P{^isdefaultignorablecodepoint}', "");
Expect(0, 921600, '\p{isdefaultignorablecodepoint}', "");
Expect(1, 921600, '\p{^isdefaultignorablecodepoint}', "");
Expect(1, 921600, '\P{isdefaultignorablecodepoint}', "");
Expect(0, 921600, '\P{^isdefaultignorablecodepoint}', "");
Expect(1, 921599, '\p{_	Is_Default_Ignorable_code_Point}', "");
Expect(0, 921599, '\p{^_	Is_Default_Ignorable_code_Point}', "");
Expect(0, 921599, '\P{_	Is_Default_Ignorable_code_Point}', "");
Expect(1, 921599, '\P{^_	Is_Default_Ignorable_code_Point}', "");
Expect(0, 921600, '\p{_	Is_Default_Ignorable_code_Point}', "");
Expect(1, 921600, '\p{^_	Is_Default_Ignorable_code_Point}', "");
Expect(1, 921600, '\P{_	Is_Default_Ignorable_code_Point}', "");
Expect(0, 921600, '\P{^_	Is_Default_Ignorable_code_Point}', "");
Error('\p{/a/  di}');
Error('\P{/a/  di}');
Expect(1, 921599, '\p{di}', "");
Expect(0, 921599, '\p{^di}', "");
Expect(0, 921599, '\P{di}', "");
Expect(1, 921599, '\P{^di}', "");
Expect(0, 921600, '\p{di}', "");
Expect(1, 921600, '\p{^di}', "");
Expect(1, 921600, '\P{di}', "");
Expect(0, 921600, '\P{^di}', "");
Expect(1, 921599, '\p{	-DI}', "");
Expect(0, 921599, '\p{^	-DI}', "");
Expect(0, 921599, '\P{	-DI}', "");
Expect(1, 921599, '\P{^	-DI}', "");
Expect(0, 921600, '\p{	-DI}', "");
Expect(1, 921600, '\p{^	-DI}', "");
Expect(1, 921600, '\P{	-DI}', "");
Expect(0, 921600, '\P{^	-DI}', "");
Error('\p{-:=Is_DI}');
Error('\P{-:=Is_DI}');
Expect(1, 921599, '\p{isdi}', "");
Expect(0, 921599, '\p{^isdi}', "");
Expect(0, 921599, '\P{isdi}', "");
Expect(1, 921599, '\P{^isdi}', "");
Expect(0, 921600, '\p{isdi}', "");
Expect(1, 921600, '\p{^isdi}', "");
Expect(1, 921600, '\P{isdi}', "");
Expect(0, 921600, '\P{^isdi}', "");
Expect(1, 921599, '\p{-	IS_DI}', "");
Expect(0, 921599, '\p{^-	IS_DI}', "");
Expect(0, 921599, '\P{-	IS_DI}', "");
Expect(1, 921599, '\P{^-	IS_DI}', "");
Expect(0, 921600, '\p{-	IS_DI}', "");
Expect(1, 921600, '\p{^-	IS_DI}', "");
Expect(1, 921600, '\P{-	IS_DI}', "");
Expect(0, 921600, '\P{^-	IS_DI}', "");
Error('\p{-/a/ID_Start}');
Error('\P{-/a/ID_Start}');
Expect(1, 195101, '\p{idstart}', "");
Expect(0, 195101, '\p{^idstart}', "");
Expect(0, 195101, '\P{idstart}', "");
Expect(1, 195101, '\P{^idstart}', "");
Expect(0, 195102, '\p{idstart}', "");
Expect(1, 195102, '\p{^idstart}', "");
Expect(1, 195102, '\P{idstart}', "");
Expect(0, 195102, '\P{^idstart}', "");
Expect(1, 195101, '\p{-	ID_Start}', "");
Expect(0, 195101, '\p{^-	ID_Start}', "");
Expect(0, 195101, '\P{-	ID_Start}', "");
Expect(1, 195101, '\P{^-	ID_Start}', "");
Expect(0, 195102, '\p{-	ID_Start}', "");
Expect(1, 195102, '\p{^-	ID_Start}', "");
Expect(1, 195102, '\P{-	ID_Start}', "");
Expect(0, 195102, '\P{^-	ID_Start}', "");
Error('\p{/a/--IS_ID_Start}');
Error('\P{/a/--IS_ID_Start}');
Expect(1, 195101, '\p{isidstart}', "");
Expect(0, 195101, '\p{^isidstart}', "");
Expect(0, 195101, '\P{isidstart}', "");
Expect(1, 195101, '\P{^isidstart}', "");
Expect(0, 195102, '\p{isidstart}', "");
Expect(1, 195102, '\p{^isidstart}', "");
Expect(1, 195102, '\P{isidstart}', "");
Expect(0, 195102, '\P{^isidstart}', "");
Expect(1, 195101, '\p{ -Is_ID_Start}', "");
Expect(0, 195101, '\p{^ -Is_ID_Start}', "");
Expect(0, 195101, '\P{ -Is_ID_Start}', "");
Expect(1, 195101, '\P{^ -Is_ID_Start}', "");
Expect(0, 195102, '\p{ -Is_ID_Start}', "");
Expect(1, 195102, '\p{^ -Is_ID_Start}', "");
Expect(1, 195102, '\P{ -Is_ID_Start}', "");
Expect(0, 195102, '\P{^ -Is_ID_Start}', "");
Error('\p{	ids:=}');
Error('\P{	ids:=}');
Expect(1, 195101, '\p{ids}', "");
Expect(0, 195101, '\p{^ids}', "");
Expect(0, 195101, '\P{ids}', "");
Expect(1, 195101, '\P{^ids}', "");
Expect(0, 195102, '\p{ids}', "");
Expect(1, 195102, '\p{^ids}', "");
Expect(1, 195102, '\P{ids}', "");
Expect(0, 195102, '\P{^ids}', "");
Expect(1, 195101, '\p{--IDS}', "");
Expect(0, 195101, '\p{^--IDS}', "");
Expect(0, 195101, '\P{--IDS}', "");
Expect(1, 195101, '\P{^--IDS}', "");
Expect(0, 195102, '\p{--IDS}', "");
Expect(1, 195102, '\p{^--IDS}', "");
Expect(1, 195102, '\P{--IDS}', "");
Expect(0, 195102, '\P{^--IDS}', "");
Error('\p{-Is_ids:=}');
Error('\P{-Is_ids:=}');
Expect(1, 195101, '\p{isids}', "");
Expect(0, 195101, '\p{^isids}', "");
Expect(0, 195101, '\P{isids}', "");
Expect(1, 195101, '\P{^isids}', "");
Expect(0, 195102, '\p{isids}', "");
Expect(1, 195102, '\p{^isids}', "");
Expect(1, 195102, '\P{isids}', "");
Expect(0, 195102, '\P{^isids}', "");
Expect(1, 195101, '\p{	_Is_IDS}', "");
Expect(0, 195101, '\p{^	_Is_IDS}', "");
Expect(0, 195101, '\P{	_Is_IDS}', "");
Expect(1, 195101, '\P{^	_Is_IDS}', "");
Expect(0, 195102, '\p{	_Is_IDS}', "");
Expect(1, 195102, '\p{^	_Is_IDS}', "");
Expect(1, 195102, '\P{	_Is_IDS}', "");
Expect(0, 195102, '\P{^	_Is_IDS}', "");
Error('\p{:=--latin_Extended_d}');
Error('\P{:=--latin_Extended_d}');
Expect(1, 43007, '\p{latinextendedd}', "");
Expect(0, 43007, '\p{^latinextendedd}', "");
Expect(0, 43007, '\P{latinextendedd}', "");
Expect(1, 43007, '\P{^latinextendedd}', "");
Expect(0, 43008, '\p{latinextendedd}', "");
Expect(1, 43008, '\p{^latinextendedd}', "");
Expect(1, 43008, '\P{latinextendedd}', "");
Expect(0, 43008, '\P{^latinextendedd}', "");
Expect(1, 43007, '\p{_ Latin_Extended_D}', "");
Expect(0, 43007, '\p{^_ Latin_Extended_D}', "");
Expect(0, 43007, '\P{_ Latin_Extended_D}', "");
Expect(1, 43007, '\P{^_ Latin_Extended_D}', "");
Expect(0, 43008, '\p{_ Latin_Extended_D}', "");
Expect(1, 43008, '\p{^_ Latin_Extended_D}', "");
Expect(1, 43008, '\P{_ Latin_Extended_D}', "");
Expect(0, 43008, '\P{^_ Latin_Extended_D}', "");
Error('\p{/a/ -Is_LATIN_Extended_D}');
Error('\P{/a/ -Is_LATIN_Extended_D}');
Expect(1, 43007, '\p{islatinextendedd}', "");
Expect(0, 43007, '\p{^islatinextendedd}', "");
Expect(0, 43007, '\P{islatinextendedd}', "");
Expect(1, 43007, '\P{^islatinextendedd}', "");
Expect(0, 43008, '\p{islatinextendedd}', "");
Expect(1, 43008, '\p{^islatinextendedd}', "");
Expect(1, 43008, '\P{islatinextendedd}', "");
Expect(0, 43008, '\P{^islatinextendedd}', "");
Expect(1, 43007, '\p{	is_Latin_EXTENDED_D}', "");
Expect(0, 43007, '\p{^	is_Latin_EXTENDED_D}', "");
Expect(0, 43007, '\P{	is_Latin_EXTENDED_D}', "");
Expect(1, 43007, '\P{^	is_Latin_EXTENDED_D}', "");
Expect(0, 43008, '\p{	is_Latin_EXTENDED_D}', "");
Expect(1, 43008, '\p{^	is_Latin_EXTENDED_D}', "");
Expect(1, 43008, '\P{	is_Latin_EXTENDED_D}', "");
Expect(0, 43008, '\P{^	is_Latin_EXTENDED_D}', "");
Error('\p{:=_In_latin_extended_d}');
Error('\P{:=_In_latin_extended_d}');
Expect(1, 43007, '\p{inlatinextendedd}', "");
Expect(0, 43007, '\p{^inlatinextendedd}', "");
Expect(0, 43007, '\P{inlatinextendedd}', "");
Expect(1, 43007, '\P{^inlatinextendedd}', "");
Expect(0, 43008, '\p{inlatinextendedd}', "");
Expect(1, 43008, '\p{^inlatinextendedd}', "");
Expect(1, 43008, '\P{inlatinextendedd}', "");
Expect(0, 43008, '\P{^inlatinextendedd}', "");
Expect(1, 43007, '\p{ _In_Latin_extended_D}', "");
Expect(0, 43007, '\p{^ _In_Latin_extended_D}', "");
Expect(0, 43007, '\P{ _In_Latin_extended_D}', "");
Expect(1, 43007, '\P{^ _In_Latin_extended_D}', "");
Expect(0, 43008, '\p{ _In_Latin_extended_D}', "");
Expect(1, 43008, '\p{^ _In_Latin_extended_D}', "");
Expect(1, 43008, '\P{ _In_Latin_extended_D}', "");
Expect(0, 43008, '\P{^ _In_Latin_extended_D}', "");
Error('\p{/a/		Latin_ext_d}');
Error('\P{/a/		Latin_ext_d}');
Expect(1, 43007, '\p{latinextd}', "");
Expect(0, 43007, '\p{^latinextd}', "");
Expect(0, 43007, '\P{latinextd}', "");
Expect(1, 43007, '\P{^latinextd}', "");
Expect(0, 43008, '\p{latinextd}', "");
Expect(1, 43008, '\p{^latinextd}', "");
Expect(1, 43008, '\P{latinextd}', "");
Expect(0, 43008, '\P{^latinextd}', "");
Expect(1, 43007, '\p{--Latin_Ext_d}', "");
Expect(0, 43007, '\p{^--Latin_Ext_d}', "");
Expect(0, 43007, '\P{--Latin_Ext_d}', "");
Expect(1, 43007, '\P{^--Latin_Ext_d}', "");
Expect(0, 43008, '\p{--Latin_Ext_d}', "");
Expect(1, 43008, '\p{^--Latin_Ext_d}', "");
Expect(1, 43008, '\P{--Latin_Ext_d}', "");
Expect(0, 43008, '\P{^--Latin_Ext_d}', "");
Error('\p{ 	is_LATIN_EXT_D/a/}');
Error('\P{ 	is_LATIN_EXT_D/a/}');
Expect(1, 43007, '\p{islatinextd}', "");
Expect(0, 43007, '\p{^islatinextd}', "");
Expect(0, 43007, '\P{islatinextd}', "");
Expect(1, 43007, '\P{^islatinextd}', "");
Expect(0, 43008, '\p{islatinextd}', "");
Expect(1, 43008, '\p{^islatinextd}', "");
Expect(1, 43008, '\P{islatinextd}', "");
Expect(0, 43008, '\P{^islatinextd}', "");
Expect(1, 43007, '\p{-_Is_Latin_EXT_D}', "");
Expect(0, 43007, '\p{^-_Is_Latin_EXT_D}', "");
Expect(0, 43007, '\P{-_Is_Latin_EXT_D}', "");
Expect(1, 43007, '\P{^-_Is_Latin_EXT_D}', "");
Expect(0, 43008, '\p{-_Is_Latin_EXT_D}', "");
Expect(1, 43008, '\p{^-_Is_Latin_EXT_D}', "");
Expect(1, 43008, '\P{-_Is_Latin_EXT_D}', "");
Expect(0, 43008, '\P{^-_Is_Latin_EXT_D}', "");
Error('\p{	in_LATIN_EXT_D:=}');
Error('\P{	in_LATIN_EXT_D:=}');
Expect(1, 43007, '\p{inlatinextd}', "");
Expect(0, 43007, '\p{^inlatinextd}', "");
Expect(0, 43007, '\P{inlatinextd}', "");
Expect(1, 43007, '\P{^inlatinextd}', "");
Expect(0, 43008, '\p{inlatinextd}', "");
Expect(1, 43008, '\p{^inlatinextd}', "");
Expect(1, 43008, '\P{inlatinextd}', "");
Expect(0, 43008, '\P{^inlatinextd}', "");
Expect(1, 43007, '\p{_In_Latin_EXT_D}', "");
Expect(0, 43007, '\p{^_In_Latin_EXT_D}', "");
Expect(0, 43007, '\P{_In_Latin_EXT_D}', "");
Expect(1, 43007, '\P{^_In_Latin_EXT_D}', "");
Expect(0, 43008, '\p{_In_Latin_EXT_D}', "");
Expect(1, 43008, '\p{^_In_Latin_EXT_D}', "");
Expect(1, 43008, '\P{_In_Latin_EXT_D}', "");
Expect(0, 43008, '\P{^_In_Latin_EXT_D}', "");
Error('\p{	_Geometric_shapes/a/}');
Error('\P{	_Geometric_shapes/a/}');
Expect(1, 9727, '\p{geometricshapes}', "");
Expect(0, 9727, '\p{^geometricshapes}', "");
Expect(0, 9727, '\P{geometricshapes}', "");
Expect(1, 9727, '\P{^geometricshapes}', "");
Expect(0, 9728, '\p{geometricshapes}', "");
Expect(1, 9728, '\p{^geometricshapes}', "");
Expect(1, 9728, '\P{geometricshapes}', "");
Expect(0, 9728, '\P{^geometricshapes}', "");
Expect(1, 9727, '\p{	 Geometric_shapes}', "");
Expect(0, 9727, '\p{^	 Geometric_shapes}', "");
Expect(0, 9727, '\P{	 Geometric_shapes}', "");
Expect(1, 9727, '\P{^	 Geometric_shapes}', "");
Expect(0, 9728, '\p{	 Geometric_shapes}', "");
Expect(1, 9728, '\p{^	 Geometric_shapes}', "");
Expect(1, 9728, '\P{	 Geometric_shapes}', "");
Expect(0, 9728, '\P{^	 Geometric_shapes}', "");
Error('\p{/a/ 	Is_Geometric_Shapes}');
Error('\P{/a/ 	Is_Geometric_Shapes}');
Expect(1, 9727, '\p{isgeometricshapes}', "");
Expect(0, 9727, '\p{^isgeometricshapes}', "");
Expect(0, 9727, '\P{isgeometricshapes}', "");
Expect(1, 9727, '\P{^isgeometricshapes}', "");
Expect(0, 9728, '\p{isgeometricshapes}', "");
Expect(1, 9728, '\p{^isgeometricshapes}', "");
Expect(1, 9728, '\P{isgeometricshapes}', "");
Expect(0, 9728, '\P{^isgeometricshapes}', "");
Expect(1, 9727, '\p{ IS_geometric_Shapes}', "");
Expect(0, 9727, '\p{^ IS_geometric_Shapes}', "");
Expect(0, 9727, '\P{ IS_geometric_Shapes}', "");
Expect(1, 9727, '\P{^ IS_geometric_Shapes}', "");
Expect(0, 9728, '\p{ IS_geometric_Shapes}', "");
Expect(1, 9728, '\p{^ IS_geometric_Shapes}', "");
Expect(1, 9728, '\P{ IS_geometric_Shapes}', "");
Expect(0, 9728, '\P{^ IS_geometric_Shapes}', "");
Error('\p{	In_GEOMETRIC_Shapes/a/}');
Error('\P{	In_GEOMETRIC_Shapes/a/}');
Expect(1, 9727, '\p{ingeometricshapes}', "");
Expect(0, 9727, '\p{^ingeometricshapes}', "");
Expect(0, 9727, '\P{ingeometricshapes}', "");
Expect(1, 9727, '\P{^ingeometricshapes}', "");
Expect(0, 9728, '\p{ingeometricshapes}', "");
Expect(1, 9728, '\p{^ingeometricshapes}', "");
Expect(1, 9728, '\P{ingeometricshapes}', "");
Expect(0, 9728, '\P{^ingeometricshapes}', "");
Expect(1, 9727, '\p{_in_Geometric_SHAPES}', "");
Expect(0, 9727, '\p{^_in_Geometric_SHAPES}', "");
Expect(0, 9727, '\P{_in_Geometric_SHAPES}', "");
Expect(1, 9727, '\P{^_in_Geometric_SHAPES}', "");
Expect(0, 9728, '\p{_in_Geometric_SHAPES}', "");
Expect(1, 9728, '\p{^_in_Geometric_SHAPES}', "");
Expect(1, 9728, '\P{_in_Geometric_SHAPES}', "");
Expect(0, 9728, '\P{^_in_Geometric_SHAPES}', "");
Error('\p{  MATHEMATICAL_Operators/a/}');
Error('\P{  MATHEMATICAL_Operators/a/}');
Expect(1, 8959, '\p{mathematicaloperators}', "");
Expect(0, 8959, '\p{^mathematicaloperators}', "");
Expect(0, 8959, '\P{mathematicaloperators}', "");
Expect(1, 8959, '\P{^mathematicaloperators}', "");
Expect(0, 8960, '\p{mathematicaloperators}', "");
Expect(1, 8960, '\p{^mathematicaloperators}', "");
Expect(1, 8960, '\P{mathematicaloperators}', "");
Expect(0, 8960, '\P{^mathematicaloperators}', "");
Expect(1, 8959, '\p{	 Mathematical_Operators}', "");
Expect(0, 8959, '\p{^	 Mathematical_Operators}', "");
Expect(0, 8959, '\P{	 Mathematical_Operators}', "");
Expect(1, 8959, '\P{^	 Mathematical_Operators}', "");
Expect(0, 8960, '\p{	 Mathematical_Operators}', "");
Expect(1, 8960, '\p{^	 Mathematical_Operators}', "");
Expect(1, 8960, '\P{	 Mathematical_Operators}', "");
Expect(0, 8960, '\P{^	 Mathematical_Operators}', "");
Error('\p{	/a/is_Mathematical_Operators}');
Error('\P{	/a/is_Mathematical_Operators}');
Expect(1, 8959, '\p{ismathematicaloperators}', "");
Expect(0, 8959, '\p{^ismathematicaloperators}', "");
Expect(0, 8959, '\P{ismathematicaloperators}', "");
Expect(1, 8959, '\P{^ismathematicaloperators}', "");
Expect(0, 8960, '\p{ismathematicaloperators}', "");
Expect(1, 8960, '\p{^ismathematicaloperators}', "");
Expect(1, 8960, '\P{ismathematicaloperators}', "");
Expect(0, 8960, '\P{^ismathematicaloperators}', "");
Expect(1, 8959, '\p{-is_Mathematical_Operators}', "");
Expect(0, 8959, '\p{^-is_Mathematical_Operators}', "");
Expect(0, 8959, '\P{-is_Mathematical_Operators}', "");
Expect(1, 8959, '\P{^-is_Mathematical_Operators}', "");
Expect(0, 8960, '\p{-is_Mathematical_Operators}', "");
Expect(1, 8960, '\p{^-is_Mathematical_Operators}', "");
Expect(1, 8960, '\P{-is_Mathematical_Operators}', "");
Expect(0, 8960, '\P{^-is_Mathematical_Operators}', "");
Error('\p{:= In_Mathematical_Operators}');
Error('\P{:= In_Mathematical_Operators}');
Expect(1, 8959, '\p{inmathematicaloperators}', "");
Expect(0, 8959, '\p{^inmathematicaloperators}', "");
Expect(0, 8959, '\P{inmathematicaloperators}', "");
Expect(1, 8959, '\P{^inmathematicaloperators}', "");
Expect(0, 8960, '\p{inmathematicaloperators}', "");
Expect(1, 8960, '\p{^inmathematicaloperators}', "");
Expect(1, 8960, '\P{inmathematicaloperators}', "");
Expect(0, 8960, '\P{^inmathematicaloperators}', "");
Expect(1, 8959, '\p{_-In_Mathematical_operators}', "");
Expect(0, 8959, '\p{^_-In_Mathematical_operators}', "");
Expect(0, 8959, '\P{_-In_Mathematical_operators}', "");
Expect(1, 8959, '\P{^_-In_Mathematical_operators}', "");
Expect(0, 8960, '\p{_-In_Mathematical_operators}', "");
Expect(1, 8960, '\p{^_-In_Mathematical_operators}', "");
Expect(1, 8960, '\P{_-In_Mathematical_operators}', "");
Expect(0, 8960, '\P{^_-In_Mathematical_operators}', "");
Error('\p{_/a/Math_OPERATORS}');
Error('\P{_/a/Math_OPERATORS}');
Expect(1, 8959, '\p{mathoperators}', "");
Expect(0, 8959, '\p{^mathoperators}', "");
Expect(0, 8959, '\P{mathoperators}', "");
Expect(1, 8959, '\P{^mathoperators}', "");
Expect(0, 8960, '\p{mathoperators}', "");
Expect(1, 8960, '\p{^mathoperators}', "");
Expect(1, 8960, '\P{mathoperators}', "");
Expect(0, 8960, '\P{^mathoperators}', "");
Expect(1, 8959, '\p{_-MATH_OPERATORS}', "");
Expect(0, 8959, '\p{^_-MATH_OPERATORS}', "");
Expect(0, 8959, '\P{_-MATH_OPERATORS}', "");
Expect(1, 8959, '\P{^_-MATH_OPERATORS}', "");
Expect(0, 8960, '\p{_-MATH_OPERATORS}', "");
Expect(1, 8960, '\p{^_-MATH_OPERATORS}', "");
Expect(1, 8960, '\P{_-MATH_OPERATORS}', "");
Expect(0, 8960, '\P{^_-MATH_OPERATORS}', "");
Error('\p{_:=Is_MATH_Operators}');
Error('\P{_:=Is_MATH_Operators}');
Expect(1, 8959, '\p{ismathoperators}', "");
Expect(0, 8959, '\p{^ismathoperators}', "");
Expect(0, 8959, '\P{ismathoperators}', "");
Expect(1, 8959, '\P{^ismathoperators}', "");
Expect(0, 8960, '\p{ismathoperators}', "");
Expect(1, 8960, '\p{^ismathoperators}', "");
Expect(1, 8960, '\P{ismathoperators}', "");
Expect(0, 8960, '\P{^ismathoperators}', "");
Expect(1, 8959, '\p{-is_MATH_OPERATORS}', "");
Expect(0, 8959, '\p{^-is_MATH_OPERATORS}', "");
Expect(0, 8959, '\P{-is_MATH_OPERATORS}', "");
Expect(1, 8959, '\P{^-is_MATH_OPERATORS}', "");
Expect(0, 8960, '\p{-is_MATH_OPERATORS}', "");
Expect(1, 8960, '\p{^-is_MATH_OPERATORS}', "");
Expect(1, 8960, '\P{-is_MATH_OPERATORS}', "");
Expect(0, 8960, '\P{^-is_MATH_OPERATORS}', "");
Error('\p{:= 	In_Math_Operators}');
Error('\P{:= 	In_Math_Operators}');
Expect(1, 8959, '\p{inmathoperators}', "");
Expect(0, 8959, '\p{^inmathoperators}', "");
Expect(0, 8959, '\P{inmathoperators}', "");
Expect(1, 8959, '\P{^inmathoperators}', "");
Expect(0, 8960, '\p{inmathoperators}', "");
Expect(1, 8960, '\p{^inmathoperators}', "");
Expect(1, 8960, '\P{inmathoperators}', "");
Expect(0, 8960, '\P{^inmathoperators}', "");
Expect(1, 8959, '\p{In_math_Operators}', "");
Expect(0, 8959, '\p{^In_math_Operators}', "");
Expect(0, 8959, '\P{In_math_Operators}', "");
Expect(1, 8959, '\P{^In_math_Operators}', "");
Expect(0, 8960, '\p{In_math_Operators}', "");
Expect(1, 8960, '\p{^In_math_Operators}', "");
Expect(1, 8960, '\P{In_math_Operators}', "");
Expect(0, 8960, '\P{^In_math_Operators}', "");
Error('\p{	:=Buginese}');
Error('\P{	:=Buginese}');
Expect(1, 6687, '\p{buginese}', "");
Expect(0, 6687, '\p{^buginese}', "");
Expect(0, 6687, '\P{buginese}', "");
Expect(1, 6687, '\P{^buginese}', "");
Expect(0, 6688, '\p{buginese}', "");
Expect(1, 6688, '\p{^buginese}', "");
Expect(1, 6688, '\P{buginese}', "");
Expect(0, 6688, '\P{^buginese}', "");
Expect(1, 6687, '\p{ -Buginese}', "");
Expect(0, 6687, '\p{^ -Buginese}', "");
Expect(0, 6687, '\P{ -Buginese}', "");
Expect(1, 6687, '\P{^ -Buginese}', "");
Expect(0, 6688, '\p{ -Buginese}', "");
Expect(1, 6688, '\p{^ -Buginese}', "");
Expect(1, 6688, '\P{ -Buginese}', "");
Expect(0, 6688, '\P{^ -Buginese}', "");
Error('\p{-is_buginese:=}');
Error('\P{-is_buginese:=}');
Expect(1, 6687, '\p{isbuginese}', "");
Expect(0, 6687, '\p{^isbuginese}', "");
Expect(0, 6687, '\P{isbuginese}', "");
Expect(1, 6687, '\P{^isbuginese}', "");
Expect(0, 6688, '\p{isbuginese}', "");
Expect(1, 6688, '\p{^isbuginese}', "");
Expect(1, 6688, '\P{isbuginese}', "");
Expect(0, 6688, '\P{^isbuginese}', "");
Expect(1, 6687, '\p{_-IS_Buginese}', "");
Expect(0, 6687, '\p{^_-IS_Buginese}', "");
Expect(0, 6687, '\P{_-IS_Buginese}', "");
Expect(1, 6687, '\P{^_-IS_Buginese}', "");
Expect(0, 6688, '\p{_-IS_Buginese}', "");
Expect(1, 6688, '\p{^_-IS_Buginese}', "");
Expect(1, 6688, '\P{_-IS_Buginese}', "");
Expect(0, 6688, '\P{^_-IS_Buginese}', "");
Error('\p{ :=bugi}');
Error('\P{ :=bugi}');
Expect(1, 6687, '\p{bugi}', "");
Expect(0, 6687, '\p{^bugi}', "");
Expect(0, 6687, '\P{bugi}', "");
Expect(1, 6687, '\P{^bugi}', "");
Expect(0, 6688, '\p{bugi}', "");
Expect(1, 6688, '\p{^bugi}', "");
Expect(1, 6688, '\P{bugi}', "");
Expect(0, 6688, '\P{^bugi}', "");
Expect(1, 6687, '\p{  Bugi}', "");
Expect(0, 6687, '\p{^  Bugi}', "");
Expect(0, 6687, '\P{  Bugi}', "");
Expect(1, 6687, '\P{^  Bugi}', "");
Expect(0, 6688, '\p{  Bugi}', "");
Expect(1, 6688, '\p{^  Bugi}', "");
Expect(1, 6688, '\P{  Bugi}', "");
Expect(0, 6688, '\P{^  Bugi}', "");
Error('\p{-_is_Bugi:=}');
Error('\P{-_is_Bugi:=}');
Expect(1, 6687, '\p{isbugi}', "");
Expect(0, 6687, '\p{^isbugi}', "");
Expect(0, 6687, '\P{isbugi}', "");
Expect(1, 6687, '\P{^isbugi}', "");
Expect(0, 6688, '\p{isbugi}', "");
Expect(1, 6688, '\p{^isbugi}', "");
Expect(1, 6688, '\P{isbugi}', "");
Expect(0, 6688, '\P{^isbugi}', "");
Expect(1, 6687, '\p{_IS_Bugi}', "");
Expect(0, 6687, '\p{^_IS_Bugi}', "");
Expect(0, 6687, '\P{_IS_Bugi}', "");
Expect(1, 6687, '\P{^_IS_Bugi}', "");
Expect(0, 6688, '\p{_IS_Bugi}', "");
Expect(1, 6688, '\p{^_IS_Bugi}', "");
Expect(1, 6688, '\P{_IS_Bugi}', "");
Expect(0, 6688, '\P{^_IS_Bugi}', "");
Error('\p{:=--Enclosed_Alphanumeric_Supplement}');
Error('\P{:=--Enclosed_Alphanumeric_Supplement}');
Expect(1, 127487, '\p{enclosedalphanumericsupplement}', "");
Expect(0, 127487, '\p{^enclosedalphanumericsupplement}', "");
Expect(0, 127487, '\P{enclosedalphanumericsupplement}', "");
Expect(1, 127487, '\P{^enclosedalphanumericsupplement}', "");
Expect(0, 127488, '\p{enclosedalphanumericsupplement}', "");
Expect(1, 127488, '\p{^enclosedalphanumericsupplement}', "");
Expect(1, 127488, '\P{enclosedalphanumericsupplement}', "");
Expect(0, 127488, '\P{^enclosedalphanumericsupplement}', "");
Expect(1, 127487, '\p{-	Enclosed_Alphanumeric_SUPPLEMENT}', "");
Expect(0, 127487, '\p{^-	Enclosed_Alphanumeric_SUPPLEMENT}', "");
Expect(0, 127487, '\P{-	Enclosed_Alphanumeric_SUPPLEMENT}', "");
Expect(1, 127487, '\P{^-	Enclosed_Alphanumeric_SUPPLEMENT}', "");
Expect(0, 127488, '\p{-	Enclosed_Alphanumeric_SUPPLEMENT}', "");
Expect(1, 127488, '\p{^-	Enclosed_Alphanumeric_SUPPLEMENT}', "");
Expect(1, 127488, '\P{-	Enclosed_Alphanumeric_SUPPLEMENT}', "");
Expect(0, 127488, '\P{^-	Enclosed_Alphanumeric_SUPPLEMENT}', "");
Error('\p{Is_enclosed_Alphanumeric_SUPPLEMENT/a/}');
Error('\P{Is_enclosed_Alphanumeric_SUPPLEMENT/a/}');
Expect(1, 127487, '\p{isenclosedalphanumericsupplement}', "");
Expect(0, 127487, '\p{^isenclosedalphanumericsupplement}', "");
Expect(0, 127487, '\P{isenclosedalphanumericsupplement}', "");
Expect(1, 127487, '\P{^isenclosedalphanumericsupplement}', "");
Expect(0, 127488, '\p{isenclosedalphanumericsupplement}', "");
Expect(1, 127488, '\p{^isenclosedalphanumericsupplement}', "");
Expect(1, 127488, '\P{isenclosedalphanumericsupplement}', "");
Expect(0, 127488, '\P{^isenclosedalphanumericsupplement}', "");
Expect(1, 127487, '\p{	Is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(0, 127487, '\p{^	Is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(0, 127487, '\P{	Is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(1, 127487, '\P{^	Is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(0, 127488, '\p{	Is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(1, 127488, '\p{^	Is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(1, 127488, '\P{	Is_enclosed_ALPHANUMERIC_supplement}', "");
Expect(0, 127488, '\P{^	Is_enclosed_ALPHANUMERIC_supplement}', "");
Error('\p{- In_ENCLOSED_alphanumeric_supplement:=}');
Error('\P{- In_ENCLOSED_alphanumeric_supplement:=}');
Expect(1, 127487, '\p{inenclosedalphanumericsupplement}', "");
Expect(0, 127487, '\p{^inenclosedalphanumericsupplement}', "");
Expect(0, 127487, '\P{inenclosedalphanumericsupplement}', "");
Expect(1, 127487, '\P{^inenclosedalphanumericsupplement}', "");
Expect(0, 127488, '\p{inenclosedalphanumericsupplement}', "");
Expect(1, 127488, '\p{^inenclosedalphanumericsupplement}', "");
Expect(1, 127488, '\P{inenclosedalphanumericsupplement}', "");
Expect(0, 127488, '\P{^inenclosedalphanumericsupplement}', "");
Expect(1, 127487, '\p{ 	In_Enclosed_ALPHANUMERIC_supplement}', "");
Expect(0, 127487, '\p{^ 	In_Enclosed_ALPHANUMERIC_supplement}', "");
Expect(0, 127487, '\P{ 	In_Enclosed_ALPHANUMERIC_supplement}', "");
Expect(1, 127487, '\P{^ 	In_Enclosed_ALPHANUMERIC_supplement}', "");
Expect(0, 127488, '\p{ 	In_Enclosed_ALPHANUMERIC_supplement}', "");
Expect(1, 127488, '\p{^ 	In_Enclosed_ALPHANUMERIC_supplement}', "");
Expect(1, 127488, '\P{ 	In_Enclosed_ALPHANUMERIC_supplement}', "");
Expect(0, 127488, '\P{^ 	In_Enclosed_ALPHANUMERIC_supplement}', "");
Error('\p{ /a/Enclosed_Alphanum_SUP}');
Error('\P{ /a/Enclosed_Alphanum_SUP}');
Expect(1, 127487, '\p{enclosedalphanumsup}', "");
Expect(0, 127487, '\p{^enclosedalphanumsup}', "");
Expect(0, 127487, '\P{enclosedalphanumsup}', "");
Expect(1, 127487, '\P{^enclosedalphanumsup}', "");
Expect(0, 127488, '\p{enclosedalphanumsup}', "");
Expect(1, 127488, '\p{^enclosedalphanumsup}', "");
Expect(1, 127488, '\P{enclosedalphanumsup}', "");
Expect(0, 127488, '\P{^enclosedalphanumsup}', "");
Expect(1, 127487, '\p{ _Enclosed_Alphanum_Sup}', "");
Expect(0, 127487, '\p{^ _Enclosed_Alphanum_Sup}', "");
Expect(0, 127487, '\P{ _Enclosed_Alphanum_Sup}', "");
Expect(1, 127487, '\P{^ _Enclosed_Alphanum_Sup}', "");
Expect(0, 127488, '\p{ _Enclosed_Alphanum_Sup}', "");
Expect(1, 127488, '\p{^ _Enclosed_Alphanum_Sup}', "");
Expect(1, 127488, '\P{ _Enclosed_Alphanum_Sup}', "");
Expect(0, 127488, '\P{^ _Enclosed_Alphanum_Sup}', "");
Error('\p{	Is_Enclosed_ALPHANUM_Sup:=}');
Error('\P{	Is_Enclosed_ALPHANUM_Sup:=}');
Expect(1, 127487, '\p{isenclosedalphanumsup}', "");
Expect(0, 127487, '\p{^isenclosedalphanumsup}', "");
Expect(0, 127487, '\P{isenclosedalphanumsup}', "");
Expect(1, 127487, '\P{^isenclosedalphanumsup}', "");
Expect(0, 127488, '\p{isenclosedalphanumsup}', "");
Expect(1, 127488, '\p{^isenclosedalphanumsup}', "");
Expect(1, 127488, '\P{isenclosedalphanumsup}', "");
Expect(0, 127488, '\P{^isenclosedalphanumsup}', "");
Expect(1, 127487, '\p{_-Is_Enclosed_alphanum_Sup}', "");
Expect(0, 127487, '\p{^_-Is_Enclosed_alphanum_Sup}', "");
Expect(0, 127487, '\P{_-Is_Enclosed_alphanum_Sup}', "");
Expect(1, 127487, '\P{^_-Is_Enclosed_alphanum_Sup}', "");
Expect(0, 127488, '\p{_-Is_Enclosed_alphanum_Sup}', "");
Expect(1, 127488, '\p{^_-Is_Enclosed_alphanum_Sup}', "");
Expect(1, 127488, '\P{_-Is_Enclosed_alphanum_Sup}', "");
Expect(0, 127488, '\P{^_-Is_Enclosed_alphanum_Sup}', "");
Error('\p{ :=In_Enclosed_Alphanum_sup}');
Error('\P{ :=In_Enclosed_Alphanum_sup}');
Expect(1, 127487, '\p{inenclosedalphanumsup}', "");
Expect(0, 127487, '\p{^inenclosedalphanumsup}', "");
Expect(0, 127487, '\P{inenclosedalphanumsup}', "");
Expect(1, 127487, '\P{^inenclosedalphanumsup}', "");
Expect(0, 127488, '\p{inenclosedalphanumsup}', "");
Expect(1, 127488, '\p{^inenclosedalphanumsup}', "");
Expect(1, 127488, '\P{inenclosedalphanumsup}', "");
Expect(0, 127488, '\P{^inenclosedalphanumsup}', "");
Expect(1, 127487, '\p{_in_ENCLOSED_ALPHANUM_sup}', "");
Expect(0, 127487, '\p{^_in_ENCLOSED_ALPHANUM_sup}', "");
Expect(0, 127487, '\P{_in_ENCLOSED_ALPHANUM_sup}', "");
Expect(1, 127487, '\P{^_in_ENCLOSED_ALPHANUM_sup}', "");
Expect(0, 127488, '\p{_in_ENCLOSED_ALPHANUM_sup}', "");
Expect(1, 127488, '\p{^_in_ENCLOSED_ALPHANUM_sup}', "");
Expect(1, 127488, '\P{_in_ENCLOSED_ALPHANUM_sup}', "");
Expect(0, 127488, '\P{^_in_ENCLOSED_ALPHANUM_sup}', "");
Error('\p{:=cjk_COMPATIBILITY}');
Error('\P{:=cjk_COMPATIBILITY}');
Expect(1, 13311, '\p{cjkcompatibility}', "");
Expect(0, 13311, '\p{^cjkcompatibility}', "");
Expect(0, 13311, '\P{cjkcompatibility}', "");
Expect(1, 13311, '\P{^cjkcompatibility}', "");
Expect(0, 13312, '\p{cjkcompatibility}', "");
Expect(1, 13312, '\p{^cjkcompatibility}', "");
Expect(1, 13312, '\P{cjkcompatibility}', "");
Expect(0, 13312, '\P{^cjkcompatibility}', "");
Expect(1, 13311, '\p{ CJK_COMPATIBILITY}', "");
Expect(0, 13311, '\p{^ CJK_COMPATIBILITY}', "");
Expect(0, 13311, '\P{ CJK_COMPATIBILITY}', "");
Expect(1, 13311, '\P{^ CJK_COMPATIBILITY}', "");
Expect(0, 13312, '\p{ CJK_COMPATIBILITY}', "");
Expect(1, 13312, '\p{^ CJK_COMPATIBILITY}', "");
Expect(1, 13312, '\P{ CJK_COMPATIBILITY}', "");
Expect(0, 13312, '\P{^ CJK_COMPATIBILITY}', "");
Error('\p{-Is_cjk_COMPATIBILITY:=}');
Error('\P{-Is_cjk_COMPATIBILITY:=}');
Expect(1, 13311, '\p{iscjkcompatibility}', "");
Expect(0, 13311, '\p{^iscjkcompatibility}', "");
Expect(0, 13311, '\P{iscjkcompatibility}', "");
Expect(1, 13311, '\P{^iscjkcompatibility}', "");
Expect(0, 13312, '\p{iscjkcompatibility}', "");
Expect(1, 13312, '\p{^iscjkcompatibility}', "");
Expect(1, 13312, '\P{iscjkcompatibility}', "");
Expect(0, 13312, '\P{^iscjkcompatibility}', "");
Expect(1, 13311, '\p{__Is_CJK_Compatibility}', "");
Expect(0, 13311, '\p{^__Is_CJK_Compatibility}', "");
Expect(0, 13311, '\P{__Is_CJK_Compatibility}', "");
Expect(1, 13311, '\P{^__Is_CJK_Compatibility}', "");
Expect(0, 13312, '\p{__Is_CJK_Compatibility}', "");
Expect(1, 13312, '\p{^__Is_CJK_Compatibility}', "");
Expect(1, 13312, '\P{__Is_CJK_Compatibility}', "");
Expect(0, 13312, '\P{^__Is_CJK_Compatibility}', "");
Error('\p{/a/In_CJK_Compatibility}');
Error('\P{/a/In_CJK_Compatibility}');
Expect(1, 13311, '\p{incjkcompatibility}', "");
Expect(0, 13311, '\p{^incjkcompatibility}', "");
Expect(0, 13311, '\P{incjkcompatibility}', "");
Expect(1, 13311, '\P{^incjkcompatibility}', "");
Expect(0, 13312, '\p{incjkcompatibility}', "");
Expect(1, 13312, '\p{^incjkcompatibility}', "");
Expect(1, 13312, '\P{incjkcompatibility}', "");
Expect(0, 13312, '\P{^incjkcompatibility}', "");
Expect(1, 13311, '\p{	In_CJK_Compatibility}', "");
Expect(0, 13311, '\p{^	In_CJK_Compatibility}', "");
Expect(0, 13311, '\P{	In_CJK_Compatibility}', "");
Expect(1, 13311, '\P{^	In_CJK_Compatibility}', "");
Expect(0, 13312, '\p{	In_CJK_Compatibility}', "");
Expect(1, 13312, '\p{^	In_CJK_Compatibility}', "");
Expect(1, 13312, '\P{	In_CJK_Compatibility}', "");
Expect(0, 13312, '\P{^	In_CJK_Compatibility}', "");
Error('\p{_/a/CJK_Compat}');
Error('\P{_/a/CJK_Compat}');
Expect(1, 13311, '\p{cjkcompat}', "");
Expect(0, 13311, '\p{^cjkcompat}', "");
Expect(0, 13311, '\P{cjkcompat}', "");
Expect(1, 13311, '\P{^cjkcompat}', "");
Expect(0, 13312, '\p{cjkcompat}', "");
Expect(1, 13312, '\p{^cjkcompat}', "");
Expect(1, 13312, '\P{cjkcompat}', "");
Expect(0, 13312, '\P{^cjkcompat}', "");
Expect(1, 13311, '\p{_-cjk_compat}', "");
Expect(0, 13311, '\p{^_-cjk_compat}', "");
Expect(0, 13311, '\P{_-cjk_compat}', "");
Expect(1, 13311, '\P{^_-cjk_compat}', "");
Expect(0, 13312, '\p{_-cjk_compat}', "");
Expect(1, 13312, '\p{^_-cjk_compat}', "");
Expect(1, 13312, '\P{_-cjk_compat}', "");
Expect(0, 13312, '\P{^_-cjk_compat}', "");
Error('\p{/a/IS_CJK_Compat}');
Error('\P{/a/IS_CJK_Compat}');
Expect(1, 13311, '\p{iscjkcompat}', "");
Expect(0, 13311, '\p{^iscjkcompat}', "");
Expect(0, 13311, '\P{iscjkcompat}', "");
Expect(1, 13311, '\P{^iscjkcompat}', "");
Expect(0, 13312, '\p{iscjkcompat}', "");
Expect(1, 13312, '\p{^iscjkcompat}', "");
Expect(1, 13312, '\P{iscjkcompat}', "");
Expect(0, 13312, '\P{^iscjkcompat}', "");
Expect(1, 13311, '\p{	 Is_CJK_COMPAT}', "");
Expect(0, 13311, '\p{^	 Is_CJK_COMPAT}', "");
Expect(0, 13311, '\P{	 Is_CJK_COMPAT}', "");
Expect(1, 13311, '\P{^	 Is_CJK_COMPAT}', "");
Expect(0, 13312, '\p{	 Is_CJK_COMPAT}', "");
Expect(1, 13312, '\p{^	 Is_CJK_COMPAT}', "");
Expect(1, 13312, '\P{	 Is_CJK_COMPAT}', "");
Expect(0, 13312, '\P{^	 Is_CJK_COMPAT}', "");
Error('\p{/a/In_cjk_Compat}');
Error('\P{/a/In_cjk_Compat}');
Expect(1, 13311, '\p{incjkcompat}', "");
Expect(0, 13311, '\p{^incjkcompat}', "");
Expect(0, 13311, '\P{incjkcompat}', "");
Expect(1, 13311, '\P{^incjkcompat}', "");
Expect(0, 13312, '\p{incjkcompat}', "");
Expect(1, 13312, '\p{^incjkcompat}', "");
Expect(1, 13312, '\P{incjkcompat}', "");
Expect(0, 13312, '\P{^incjkcompat}', "");
Expect(1, 13311, '\p{_-In_cjk_Compat}', "");
Expect(0, 13311, '\p{^_-In_cjk_Compat}', "");
Expect(0, 13311, '\P{_-In_cjk_Compat}', "");
Expect(1, 13311, '\P{^_-In_cjk_Compat}', "");
Expect(0, 13312, '\p{_-In_cjk_Compat}', "");
Expect(1, 13312, '\p{^_-In_cjk_Compat}', "");
Expect(1, 13312, '\P{_-In_cjk_Compat}', "");
Expect(0, 13312, '\P{^_-In_cjk_Compat}', "");
Error('\p{--IN_malayalam:=}');
Error('\P{--IN_malayalam:=}');
Expect(1, 3455, '\p{inmalayalam}', "");
Expect(0, 3455, '\p{^inmalayalam}', "");
Expect(0, 3455, '\P{inmalayalam}', "");
Expect(1, 3455, '\P{^inmalayalam}', "");
Expect(0, 3456, '\p{inmalayalam}', "");
Expect(1, 3456, '\p{^inmalayalam}', "");
Expect(1, 3456, '\P{inmalayalam}', "");
Expect(0, 3456, '\P{^inmalayalam}', "");
Expect(1, 3455, '\p{	in_Malayalam}', "");
Expect(0, 3455, '\p{^	in_Malayalam}', "");
Expect(0, 3455, '\P{	in_Malayalam}', "");
Expect(1, 3455, '\P{^	in_Malayalam}', "");
Expect(0, 3456, '\p{	in_Malayalam}', "");
Expect(1, 3456, '\p{^	in_Malayalam}', "");
Expect(1, 3456, '\P{	in_Malayalam}', "");
Expect(0, 3456, '\P{^	in_Malayalam}', "");
Error('\p{-:=in_Malayalam}');
Error('\P{-:=in_Malayalam}');
Expect(1, 3455, '\p{	-in_MALAYALAM}', "");
Expect(0, 3455, '\p{^	-in_MALAYALAM}', "");
Expect(0, 3455, '\P{	-in_MALAYALAM}', "");
Expect(1, 3455, '\P{^	-in_MALAYALAM}', "");
Expect(0, 3456, '\p{	-in_MALAYALAM}', "");
Expect(1, 3456, '\p{^	-in_MALAYALAM}', "");
Expect(1, 3456, '\P{	-in_MALAYALAM}', "");
Expect(0, 3456, '\P{^	-in_MALAYALAM}', "");
Error('\p{/a/ 	Inscriptional_Pahlavi}');
Error('\P{/a/ 	Inscriptional_Pahlavi}');
Expect(1, 68479, '\p{inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{--INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68479, '\p{^--INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68479, '\P{--INSCRIPTIONAL_Pahlavi}', "");
Expect(1, 68479, '\P{^--INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68480, '\p{--INSCRIPTIONAL_Pahlavi}', "");
Expect(1, 68480, '\p{^--INSCRIPTIONAL_Pahlavi}', "");
Expect(1, 68480, '\P{--INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68480, '\P{^--INSCRIPTIONAL_Pahlavi}', "");
Error('\p{:=_is_Inscriptional_Pahlavi}');
Error('\P{:=_is_Inscriptional_Pahlavi}');
Expect(1, 68479, '\p{isinscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^isinscriptionalpahlavi}', "");
Expect(0, 68479, '\P{isinscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^isinscriptionalpahlavi}', "");
Expect(0, 68480, '\p{isinscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^isinscriptionalpahlavi}', "");
Expect(1, 68480, '\P{isinscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^isinscriptionalpahlavi}', "");
Expect(1, 68479, '\p{ is_INSCRIPTIONAL_pahlavi}', "");
Expect(0, 68479, '\p{^ is_INSCRIPTIONAL_pahlavi}', "");
Expect(0, 68479, '\P{ is_INSCRIPTIONAL_pahlavi}', "");
Expect(1, 68479, '\P{^ is_INSCRIPTIONAL_pahlavi}', "");
Expect(0, 68480, '\p{ is_INSCRIPTIONAL_pahlavi}', "");
Expect(1, 68480, '\p{^ is_INSCRIPTIONAL_pahlavi}', "");
Expect(1, 68480, '\P{ is_INSCRIPTIONAL_pahlavi}', "");
Expect(0, 68480, '\P{^ is_INSCRIPTIONAL_pahlavi}', "");
Error('\p{-/a/PHLI}');
Error('\P{-/a/PHLI}');
Expect(1, 68479, '\p{phli}', "");
Expect(0, 68479, '\p{^phli}', "");
Expect(0, 68479, '\P{phli}', "");
Expect(1, 68479, '\P{^phli}', "");
Expect(0, 68480, '\p{phli}', "");
Expect(1, 68480, '\p{^phli}', "");
Expect(1, 68480, '\P{phli}', "");
Expect(0, 68480, '\P{^phli}', "");
Expect(1, 68479, '\p{_	PHLI}', "");
Expect(0, 68479, '\p{^_	PHLI}', "");
Expect(0, 68479, '\P{_	PHLI}', "");
Expect(1, 68479, '\P{^_	PHLI}', "");
Expect(0, 68480, '\p{_	PHLI}', "");
Expect(1, 68480, '\p{^_	PHLI}', "");
Expect(1, 68480, '\P{_	PHLI}', "");
Expect(0, 68480, '\P{^_	PHLI}', "");
Error('\p{ is_phli/a/}');
Error('\P{ is_phli/a/}');
Expect(1, 68479, '\p{isphli}', "");
Expect(0, 68479, '\p{^isphli}', "");
Expect(0, 68479, '\P{isphli}', "");
Expect(1, 68479, '\P{^isphli}', "");
Expect(0, 68480, '\p{isphli}', "");
Expect(1, 68480, '\p{^isphli}', "");
Expect(1, 68480, '\P{isphli}', "");
Expect(0, 68480, '\P{^isphli}', "");
Expect(1, 68479, '\p{	Is_phli}', "");
Expect(0, 68479, '\p{^	Is_phli}', "");
Expect(0, 68479, '\P{	Is_phli}', "");
Expect(1, 68479, '\P{^	Is_phli}', "");
Expect(0, 68480, '\p{	Is_phli}', "");
Expect(1, 68480, '\p{^	Is_phli}', "");
Expect(1, 68480, '\P{	Is_phli}', "");
Expect(0, 68480, '\P{^	Is_phli}', "");
Error('\p{:= _Cham}');
Error('\P{:= _Cham}');
Expect(1, 43615, '\p{cham}', "");
Expect(0, 43615, '\p{^cham}', "");
Expect(0, 43615, '\P{cham}', "");
Expect(1, 43615, '\P{^cham}', "");
Expect(0, 43616, '\p{cham}', "");
Expect(1, 43616, '\p{^cham}', "");
Expect(1, 43616, '\P{cham}', "");
Expect(0, 43616, '\P{^cham}', "");
Expect(1, 43615, '\p{__cham}', "");
Expect(0, 43615, '\p{^__cham}', "");
Expect(0, 43615, '\P{__cham}', "");
Expect(1, 43615, '\P{^__cham}', "");
Expect(0, 43616, '\p{__cham}', "");
Expect(1, 43616, '\p{^__cham}', "");
Expect(1, 43616, '\P{__cham}', "");
Expect(0, 43616, '\P{^__cham}', "");
Error('\p{ :=IS_Cham}');
Error('\P{ :=IS_Cham}');
Expect(1, 43615, '\p{ischam}', "");
Expect(0, 43615, '\p{^ischam}', "");
Expect(0, 43615, '\P{ischam}', "");
Expect(1, 43615, '\P{^ischam}', "");
Expect(0, 43616, '\p{ischam}', "");
Expect(1, 43616, '\p{^ischam}', "");
Expect(1, 43616, '\P{ischam}', "");
Expect(0, 43616, '\P{^ischam}', "");
Expect(1, 43615, '\p{_Is_CHAM}', "");
Expect(0, 43615, '\p{^_Is_CHAM}', "");
Expect(0, 43615, '\P{_Is_CHAM}', "");
Expect(1, 43615, '\P{^_Is_CHAM}', "");
Expect(0, 43616, '\p{_Is_CHAM}', "");
Expect(1, 43616, '\p{^_Is_CHAM}', "");
Expect(1, 43616, '\P{_Is_CHAM}', "");
Expect(0, 43616, '\P{^_Is_CHAM}', "");
Error('\p{	Braille/a/}');
Error('\P{	Braille/a/}');
Expect(1, 10495, '\p{braille}', "");
Expect(0, 10495, '\p{^braille}', "");
Expect(0, 10495, '\P{braille}', "");
Expect(1, 10495, '\P{^braille}', "");
Expect(0, 10496, '\p{braille}', "");
Expect(1, 10496, '\p{^braille}', "");
Expect(1, 10496, '\P{braille}', "");
Expect(0, 10496, '\P{^braille}', "");
Expect(1, 10495, '\p{__BRAILLE}', "");
Expect(0, 10495, '\p{^__BRAILLE}', "");
Expect(0, 10495, '\P{__BRAILLE}', "");
Expect(1, 10495, '\P{^__BRAILLE}', "");
Expect(0, 10496, '\p{__BRAILLE}', "");
Expect(1, 10496, '\p{^__BRAILLE}', "");
Expect(1, 10496, '\P{__BRAILLE}', "");
Expect(0, 10496, '\P{^__BRAILLE}', "");
Error('\p{  Is_braille/a/}');
Error('\P{  Is_braille/a/}');
Expect(1, 10495, '\p{isbraille}', "");
Expect(0, 10495, '\p{^isbraille}', "");
Expect(0, 10495, '\P{isbraille}', "");
Expect(1, 10495, '\P{^isbraille}', "");
Expect(0, 10496, '\p{isbraille}', "");
Expect(1, 10496, '\p{^isbraille}', "");
Expect(1, 10496, '\P{isbraille}', "");
Expect(0, 10496, '\P{^isbraille}', "");
Expect(1, 10495, '\p{_-is_Braille}', "");
Expect(0, 10495, '\p{^_-is_Braille}', "");
Expect(0, 10495, '\P{_-is_Braille}', "");
Expect(1, 10495, '\P{^_-is_Braille}', "");
Expect(0, 10496, '\p{_-is_Braille}', "");
Expect(1, 10496, '\p{^_-is_Braille}', "");
Expect(1, 10496, '\P{_-is_Braille}', "");
Expect(0, 10496, '\P{^_-is_Braille}', "");
Error('\p{_ Brai/a/}');
Error('\P{_ Brai/a/}');
Expect(1, 10495, '\p{brai}', "");
Expect(0, 10495, '\p{^brai}', "");
Expect(0, 10495, '\P{brai}', "");
Expect(1, 10495, '\P{^brai}', "");
Expect(0, 10496, '\p{brai}', "");
Expect(1, 10496, '\p{^brai}', "");
Expect(1, 10496, '\P{brai}', "");
Expect(0, 10496, '\P{^brai}', "");
Expect(1, 10495, '\p{	 Brai}', "");
Expect(0, 10495, '\p{^	 Brai}', "");
Expect(0, 10495, '\P{	 Brai}', "");
Expect(1, 10495, '\P{^	 Brai}', "");
Expect(0, 10496, '\p{	 Brai}', "");
Expect(1, 10496, '\p{^	 Brai}', "");
Expect(1, 10496, '\P{	 Brai}', "");
Expect(0, 10496, '\P{^	 Brai}', "");
Error('\p{ -is_Brai:=}');
Error('\P{ -is_Brai:=}');
Expect(1, 10495, '\p{isbrai}', "");
Expect(0, 10495, '\p{^isbrai}', "");
Expect(0, 10495, '\P{isbrai}', "");
Expect(1, 10495, '\P{^isbrai}', "");
Expect(0, 10496, '\p{isbrai}', "");
Expect(1, 10496, '\p{^isbrai}', "");
Expect(1, 10496, '\P{isbrai}', "");
Expect(0, 10496, '\P{^isbrai}', "");
Expect(1, 10495, '\p{_-is_Brai}', "");
Expect(0, 10495, '\p{^_-is_Brai}', "");
Expect(0, 10495, '\P{_-is_Brai}', "");
Expect(1, 10495, '\P{^_-is_Brai}', "");
Expect(0, 10496, '\p{_-is_Brai}', "");
Expect(1, 10496, '\p{^_-is_Brai}', "");
Expect(1, 10496, '\P{_-is_Brai}', "");
Expect(0, 10496, '\P{^_-is_Brai}', "");
Error('\p{-	In_Old_Persian/a/}');
Error('\P{-	In_Old_Persian/a/}');
Expect(1, 66527, '\p{inoldpersian}', "");
Expect(0, 66527, '\p{^inoldpersian}', "");
Expect(0, 66527, '\P{inoldpersian}', "");
Expect(1, 66527, '\P{^inoldpersian}', "");
Expect(0, 66528, '\p{inoldpersian}', "");
Expect(1, 66528, '\p{^inoldpersian}', "");
Expect(1, 66528, '\P{inoldpersian}', "");
Expect(0, 66528, '\P{^inoldpersian}', "");
Expect(1, 66527, '\p{-_In_Old_persian}', "");
Expect(0, 66527, '\p{^-_In_Old_persian}', "");
Expect(0, 66527, '\P{-_In_Old_persian}', "");
Expect(1, 66527, '\P{^-_In_Old_persian}', "");
Expect(0, 66528, '\p{-_In_Old_persian}', "");
Expect(1, 66528, '\p{^-_In_Old_persian}', "");
Expect(1, 66528, '\P{-_In_Old_persian}', "");
Expect(0, 66528, '\P{^-_In_Old_persian}', "");
Error('\p{	 In_Old_PERSIAN:=}');
Error('\P{	 In_Old_PERSIAN:=}');
Expect(1, 66527, '\p{	In_OLD_Persian}', "");
Expect(0, 66527, '\p{^	In_OLD_Persian}', "");
Expect(0, 66527, '\P{	In_OLD_Persian}', "");
Expect(1, 66527, '\P{^	In_OLD_Persian}', "");
Expect(0, 66528, '\p{	In_OLD_Persian}', "");
Expect(1, 66528, '\p{^	In_OLD_Persian}', "");
Expect(1, 66528, '\P{	In_OLD_Persian}', "");
Expect(0, 66528, '\P{^	In_OLD_Persian}', "");
Error('\p{_BOX_DRAWING:=}');
Error('\P{_BOX_DRAWING:=}');
Expect(1, 9599, '\p{boxdrawing}', "");
Expect(0, 9599, '\p{^boxdrawing}', "");
Expect(0, 9599, '\P{boxdrawing}', "");
Expect(1, 9599, '\P{^boxdrawing}', "");
Expect(0, 9600, '\p{boxdrawing}', "");
Expect(1, 9600, '\p{^boxdrawing}', "");
Expect(1, 9600, '\P{boxdrawing}', "");
Expect(0, 9600, '\P{^boxdrawing}', "");
Expect(1, 9599, '\p{_-Box_drawing}', "");
Expect(0, 9599, '\p{^_-Box_drawing}', "");
Expect(0, 9599, '\P{_-Box_drawing}', "");
Expect(1, 9599, '\P{^_-Box_drawing}', "");
Expect(0, 9600, '\p{_-Box_drawing}', "");
Expect(1, 9600, '\p{^_-Box_drawing}', "");
Expect(1, 9600, '\P{_-Box_drawing}', "");
Expect(0, 9600, '\P{^_-Box_drawing}', "");
Error('\p{_:=Is_Box_DRAWING}');
Error('\P{_:=Is_Box_DRAWING}');
Expect(1, 9599, '\p{isboxdrawing}', "");
Expect(0, 9599, '\p{^isboxdrawing}', "");
Expect(0, 9599, '\P{isboxdrawing}', "");
Expect(1, 9599, '\P{^isboxdrawing}', "");
Expect(0, 9600, '\p{isboxdrawing}', "");
Expect(1, 9600, '\p{^isboxdrawing}', "");
Expect(1, 9600, '\P{isboxdrawing}', "");
Expect(0, 9600, '\P{^isboxdrawing}', "");
Expect(1, 9599, '\p{_-Is_BOX_DRAWING}', "");
Expect(0, 9599, '\p{^_-Is_BOX_DRAWING}', "");
Expect(0, 9599, '\P{_-Is_BOX_DRAWING}', "");
Expect(1, 9599, '\P{^_-Is_BOX_DRAWING}', "");
Expect(0, 9600, '\p{_-Is_BOX_DRAWING}', "");
Expect(1, 9600, '\p{^_-Is_BOX_DRAWING}', "");
Expect(1, 9600, '\P{_-Is_BOX_DRAWING}', "");
Expect(0, 9600, '\P{^_-Is_BOX_DRAWING}', "");
Error('\p{	:=in_Box_drawing}');
Error('\P{	:=in_Box_drawing}');
Expect(1, 9599, '\p{inboxdrawing}', "");
Expect(0, 9599, '\p{^inboxdrawing}', "");
Expect(0, 9599, '\P{inboxdrawing}', "");
Expect(1, 9599, '\P{^inboxdrawing}', "");
Expect(0, 9600, '\p{inboxdrawing}', "");
Expect(1, 9600, '\p{^inboxdrawing}', "");
Expect(1, 9600, '\P{inboxdrawing}', "");
Expect(0, 9600, '\P{^inboxdrawing}', "");
Expect(1, 9599, '\p{-	IN_box_Drawing}', "");
Expect(0, 9599, '\p{^-	IN_box_Drawing}', "");
Expect(0, 9599, '\P{-	IN_box_Drawing}', "");
Expect(1, 9599, '\P{^-	IN_box_Drawing}', "");
Expect(0, 9600, '\p{-	IN_box_Drawing}', "");
Expect(1, 9600, '\p{^-	IN_box_Drawing}', "");
Expect(1, 9600, '\P{-	IN_box_Drawing}', "");
Expect(0, 9600, '\P{^-	IN_box_Drawing}', "");
Error('\p{/a/_LINEAR_B}');
Error('\P{/a/_LINEAR_B}');
Expect(1, 65786, '\p{linearb}', "");
Expect(0, 65786, '\p{^linearb}', "");
Expect(0, 65786, '\P{linearb}', "");
Expect(1, 65786, '\P{^linearb}', "");
Expect(0, 65787, '\p{linearb}', "");
Expect(1, 65787, '\p{^linearb}', "");
Expect(1, 65787, '\P{linearb}', "");
Expect(0, 65787, '\P{^linearb}', "");
Expect(1, 65786, '\p{	 Linear_B}', "");
Expect(0, 65786, '\p{^	 Linear_B}', "");
Expect(0, 65786, '\P{	 Linear_B}', "");
Expect(1, 65786, '\P{^	 Linear_B}', "");
Expect(0, 65787, '\p{	 Linear_B}', "");
Expect(1, 65787, '\p{^	 Linear_B}', "");
Expect(1, 65787, '\P{	 Linear_B}', "");
Expect(0, 65787, '\P{^	 Linear_B}', "");
Error('\p{ IS_linear_B:=}');
Error('\P{ IS_linear_B:=}');
Expect(1, 65786, '\p{islinearb}', "");
Expect(0, 65786, '\p{^islinearb}', "");
Expect(0, 65786, '\P{islinearb}', "");
Expect(1, 65786, '\P{^islinearb}', "");
Expect(0, 65787, '\p{islinearb}', "");
Expect(1, 65787, '\p{^islinearb}', "");
Expect(1, 65787, '\P{islinearb}', "");
Expect(0, 65787, '\P{^islinearb}', "");
Expect(1, 65786, '\p{-_Is_LINEAR_B}', "");
Expect(0, 65786, '\p{^-_Is_LINEAR_B}', "");
Expect(0, 65786, '\P{-_Is_LINEAR_B}', "");
Expect(1, 65786, '\P{^-_Is_LINEAR_B}', "");
Expect(0, 65787, '\p{-_Is_LINEAR_B}', "");
Expect(1, 65787, '\p{^-_Is_LINEAR_B}', "");
Expect(1, 65787, '\P{-_Is_LINEAR_B}', "");
Expect(0, 65787, '\P{^-_Is_LINEAR_B}', "");
Error('\p{/a/_ Linb}');
Error('\P{/a/_ Linb}');
Expect(1, 65786, '\p{linb}', "");
Expect(0, 65786, '\p{^linb}', "");
Expect(0, 65786, '\P{linb}', "");
Expect(1, 65786, '\P{^linb}', "");
Expect(0, 65787, '\p{linb}', "");
Expect(1, 65787, '\p{^linb}', "");
Expect(1, 65787, '\P{linb}', "");
Expect(0, 65787, '\P{^linb}', "");
Expect(1, 65786, '\p{	 LINB}', "");
Expect(0, 65786, '\p{^	 LINB}', "");
Expect(0, 65786, '\P{	 LINB}', "");
Expect(1, 65786, '\P{^	 LINB}', "");
Expect(0, 65787, '\p{	 LINB}', "");
Expect(1, 65787, '\p{^	 LINB}', "");
Expect(1, 65787, '\P{	 LINB}', "");
Expect(0, 65787, '\P{^	 LINB}', "");
Error('\p{/a/IS_Linb}');
Error('\P{/a/IS_Linb}');
Expect(1, 65786, '\p{islinb}', "");
Expect(0, 65786, '\p{^islinb}', "");
Expect(0, 65786, '\P{islinb}', "");
Expect(1, 65786, '\P{^islinb}', "");
Expect(0, 65787, '\p{islinb}', "");
Expect(1, 65787, '\p{^islinb}', "");
Expect(1, 65787, '\P{islinb}', "");
Expect(0, 65787, '\P{^islinb}', "");
Expect(1, 65786, '\p{ _IS_Linb}', "");
Expect(0, 65786, '\p{^ _IS_Linb}', "");
Expect(0, 65786, '\P{ _IS_Linb}', "");
Expect(1, 65786, '\P{^ _IS_Linb}', "");
Expect(0, 65787, '\p{ _IS_Linb}', "");
Expect(1, 65787, '\p{^ _IS_Linb}', "");
Expect(1, 65787, '\P{ _IS_Linb}', "");
Expect(0, 65787, '\P{^ _IS_Linb}', "");
Error('\p{/a/--TIFINAGH}');
Error('\P{/a/--TIFINAGH}');
Expect(1, 11647, '\p{tifinagh}', "");
Expect(0, 11647, '\p{^tifinagh}', "");
Expect(0, 11647, '\P{tifinagh}', "");
Expect(1, 11647, '\P{^tifinagh}', "");
Expect(0, 11648, '\p{tifinagh}', "");
Expect(1, 11648, '\p{^tifinagh}', "");
Expect(1, 11648, '\P{tifinagh}', "");
Expect(0, 11648, '\P{^tifinagh}', "");
Expect(1, 11647, '\p{ Tifinagh}', "");
Expect(0, 11647, '\p{^ Tifinagh}', "");
Expect(0, 11647, '\P{ Tifinagh}', "");
Expect(1, 11647, '\P{^ Tifinagh}', "");
Expect(0, 11648, '\p{ Tifinagh}', "");
Expect(1, 11648, '\p{^ Tifinagh}', "");
Expect(1, 11648, '\P{ Tifinagh}', "");
Expect(0, 11648, '\P{^ Tifinagh}', "");
Error('\p{/a/	-is_Tifinagh}');
Error('\P{/a/	-is_Tifinagh}');
Expect(1, 11647, '\p{istifinagh}', "");
Expect(0, 11647, '\p{^istifinagh}', "");
Expect(0, 11647, '\P{istifinagh}', "");
Expect(1, 11647, '\P{^istifinagh}', "");
Expect(0, 11648, '\p{istifinagh}', "");
Expect(1, 11648, '\p{^istifinagh}', "");
Expect(1, 11648, '\P{istifinagh}', "");
Expect(0, 11648, '\P{^istifinagh}', "");
Expect(1, 11647, '\p{-IS_Tifinagh}', "");
Expect(0, 11647, '\p{^-IS_Tifinagh}', "");
Expect(0, 11647, '\P{-IS_Tifinagh}', "");
Expect(1, 11647, '\P{^-IS_Tifinagh}', "");
Expect(0, 11648, '\p{-IS_Tifinagh}', "");
Expect(1, 11648, '\p{^-IS_Tifinagh}', "");
Expect(1, 11648, '\P{-IS_Tifinagh}', "");
Expect(0, 11648, '\P{^-IS_Tifinagh}', "");
Error('\p{-Tfng/a/}');
Error('\P{-Tfng/a/}');
Expect(1, 11647, '\p{tfng}', "");
Expect(0, 11647, '\p{^tfng}', "");
Expect(0, 11647, '\P{tfng}', "");
Expect(1, 11647, '\P{^tfng}', "");
Expect(0, 11648, '\p{tfng}', "");
Expect(1, 11648, '\p{^tfng}', "");
Expect(1, 11648, '\P{tfng}', "");
Expect(0, 11648, '\P{^tfng}', "");
Expect(1, 11647, '\p{_ TFNG}', "");
Expect(0, 11647, '\p{^_ TFNG}', "");
Expect(0, 11647, '\P{_ TFNG}', "");
Expect(1, 11647, '\P{^_ TFNG}', "");
Expect(0, 11648, '\p{_ TFNG}', "");
Expect(1, 11648, '\p{^_ TFNG}', "");
Expect(1, 11648, '\P{_ TFNG}', "");
Expect(0, 11648, '\P{^_ TFNG}', "");
Error('\p{ -Is_Tfng/a/}');
Error('\P{ -Is_Tfng/a/}');
Expect(1, 11647, '\p{istfng}', "");
Expect(0, 11647, '\p{^istfng}', "");
Expect(0, 11647, '\P{istfng}', "");
Expect(1, 11647, '\P{^istfng}', "");
Expect(0, 11648, '\p{istfng}', "");
Expect(1, 11648, '\p{^istfng}', "");
Expect(1, 11648, '\P{istfng}', "");
Expect(0, 11648, '\P{^istfng}', "");
Expect(1, 11647, '\p{ _Is_Tfng}', "");
Expect(0, 11647, '\p{^ _Is_Tfng}', "");
Expect(0, 11647, '\P{ _Is_Tfng}', "");
Expect(1, 11647, '\P{^ _Is_Tfng}', "");
Expect(0, 11648, '\p{ _Is_Tfng}', "");
Expect(1, 11648, '\p{^ _Is_Tfng}', "");
Expect(1, 11648, '\P{ _Is_Tfng}', "");
Expect(0, 11648, '\P{^ _Is_Tfng}', "");
Error('\p{-/a/syloti_Nagri}');
Error('\P{-/a/syloti_Nagri}');
Expect(1, 43051, '\p{sylotinagri}', "");
Expect(0, 43051, '\p{^sylotinagri}', "");
Expect(0, 43051, '\P{sylotinagri}', "");
Expect(1, 43051, '\P{^sylotinagri}', "");
Expect(0, 43052, '\p{sylotinagri}', "");
Expect(1, 43052, '\p{^sylotinagri}', "");
Expect(1, 43052, '\P{sylotinagri}', "");
Expect(0, 43052, '\P{^sylotinagri}', "");
Expect(1, 43051, '\p{_	Syloti_Nagri}', "");
Expect(0, 43051, '\p{^_	Syloti_Nagri}', "");
Expect(0, 43051, '\P{_	Syloti_Nagri}', "");
Expect(1, 43051, '\P{^_	Syloti_Nagri}', "");
Expect(0, 43052, '\p{_	Syloti_Nagri}', "");
Expect(1, 43052, '\p{^_	Syloti_Nagri}', "");
Expect(1, 43052, '\P{_	Syloti_Nagri}', "");
Expect(0, 43052, '\P{^_	Syloti_Nagri}', "");
Error('\p{:=Is_SYLOTI_Nagri}');
Error('\P{:=Is_SYLOTI_Nagri}');
Expect(1, 43051, '\p{issylotinagri}', "");
Expect(0, 43051, '\p{^issylotinagri}', "");
Expect(0, 43051, '\P{issylotinagri}', "");
Expect(1, 43051, '\P{^issylotinagri}', "");
Expect(0, 43052, '\p{issylotinagri}', "");
Expect(1, 43052, '\p{^issylotinagri}', "");
Expect(1, 43052, '\P{issylotinagri}', "");
Expect(0, 43052, '\P{^issylotinagri}', "");
Expect(1, 43051, '\p{ Is_Syloti_nagri}', "");
Expect(0, 43051, '\p{^ Is_Syloti_nagri}', "");
Expect(0, 43051, '\P{ Is_Syloti_nagri}', "");
Expect(1, 43051, '\P{^ Is_Syloti_nagri}', "");
Expect(0, 43052, '\p{ Is_Syloti_nagri}', "");
Expect(1, 43052, '\p{^ Is_Syloti_nagri}', "");
Expect(1, 43052, '\P{ Is_Syloti_nagri}', "");
Expect(0, 43052, '\P{^ Is_Syloti_nagri}', "");
Error('\p{ Sylo:=}');
Error('\P{ Sylo:=}');
Expect(1, 43051, '\p{sylo}', "");
Expect(0, 43051, '\p{^sylo}', "");
Expect(0, 43051, '\P{sylo}', "");
Expect(1, 43051, '\P{^sylo}', "");
Expect(0, 43052, '\p{sylo}', "");
Expect(1, 43052, '\p{^sylo}', "");
Expect(1, 43052, '\P{sylo}', "");
Expect(0, 43052, '\P{^sylo}', "");
Expect(1, 43051, '\p{--SYLO}', "");
Expect(0, 43051, '\p{^--SYLO}', "");
Expect(0, 43051, '\P{--SYLO}', "");
Expect(1, 43051, '\P{^--SYLO}', "");
Expect(0, 43052, '\p{--SYLO}', "");
Expect(1, 43052, '\p{^--SYLO}', "");
Expect(1, 43052, '\P{--SYLO}', "");
Expect(0, 43052, '\P{^--SYLO}', "");
Error('\p{/a/	 Is_sylo}');
Error('\P{/a/	 Is_sylo}');
Expect(1, 43051, '\p{issylo}', "");
Expect(0, 43051, '\p{^issylo}', "");
Expect(0, 43051, '\P{issylo}', "");
Expect(1, 43051, '\P{^issylo}', "");
Expect(0, 43052, '\p{issylo}', "");
Expect(1, 43052, '\p{^issylo}', "");
Expect(1, 43052, '\P{issylo}', "");
Expect(0, 43052, '\P{^issylo}', "");
Expect(1, 43051, '\p{  is_Sylo}', "");
Expect(0, 43051, '\p{^  is_Sylo}', "");
Expect(0, 43051, '\P{  is_Sylo}', "");
Expect(1, 43051, '\P{^  is_Sylo}', "");
Expect(0, 43052, '\p{  is_Sylo}', "");
Expect(1, 43052, '\p{^  is_Sylo}', "");
Expect(1, 43052, '\P{  is_Sylo}', "");
Expect(0, 43052, '\P{^  is_Sylo}', "");
Error('\p{/a/ARABIC_presentation_FORMS_A}');
Error('\P{/a/ARABIC_presentation_FORMS_A}');
Expect(1, 65023, '\p{arabicpresentationformsa}', "");
Expect(0, 65023, '\p{^arabicpresentationformsa}', "");
Expect(0, 65023, '\P{arabicpresentationformsa}', "");
Expect(1, 65023, '\P{^arabicpresentationformsa}', "");
Expect(0, 65024, '\p{arabicpresentationformsa}', "");
Expect(1, 65024, '\p{^arabicpresentationformsa}', "");
Expect(1, 65024, '\P{arabicpresentationformsa}', "");
Expect(0, 65024, '\P{^arabicpresentationformsa}', "");
Expect(1, 65023, '\p{	 Arabic_PRESENTATION_FORMS_A}', "");
Expect(0, 65023, '\p{^	 Arabic_PRESENTATION_FORMS_A}', "");
Expect(0, 65023, '\P{	 Arabic_PRESENTATION_FORMS_A}', "");
Expect(1, 65023, '\P{^	 Arabic_PRESENTATION_FORMS_A}', "");
Expect(0, 65024, '\p{	 Arabic_PRESENTATION_FORMS_A}', "");
Expect(1, 65024, '\p{^	 Arabic_PRESENTATION_FORMS_A}', "");
Expect(1, 65024, '\P{	 Arabic_PRESENTATION_FORMS_A}', "");
Expect(0, 65024, '\P{^	 Arabic_PRESENTATION_FORMS_A}', "");
Error('\p{-	Is_Arabic_presentation_forms_a/a/}');
Error('\P{-	Is_Arabic_presentation_forms_a/a/}');
Expect(1, 65023, '\p{isarabicpresentationformsa}', "");
Expect(0, 65023, '\p{^isarabicpresentationformsa}', "");
Expect(0, 65023, '\P{isarabicpresentationformsa}', "");
Expect(1, 65023, '\P{^isarabicpresentationformsa}', "");
Expect(0, 65024, '\p{isarabicpresentationformsa}', "");
Expect(1, 65024, '\p{^isarabicpresentationformsa}', "");
Expect(1, 65024, '\P{isarabicpresentationformsa}', "");
Expect(0, 65024, '\P{^isarabicpresentationformsa}', "");
Expect(1, 65023, '\p{_-is_arabic_Presentation_Forms_A}', "");
Expect(0, 65023, '\p{^_-is_arabic_Presentation_Forms_A}', "");
Expect(0, 65023, '\P{_-is_arabic_Presentation_Forms_A}', "");
Expect(1, 65023, '\P{^_-is_arabic_Presentation_Forms_A}', "");
Expect(0, 65024, '\p{_-is_arabic_Presentation_Forms_A}', "");
Expect(1, 65024, '\p{^_-is_arabic_Presentation_Forms_A}', "");
Expect(1, 65024, '\P{_-is_arabic_Presentation_Forms_A}', "");
Expect(0, 65024, '\P{^_-is_arabic_Presentation_Forms_A}', "");
Error('\p{-:=in_arabic_presentation_Forms_A}');
Error('\P{-:=in_arabic_presentation_Forms_A}');
Expect(1, 65023, '\p{inarabicpresentationformsa}', "");
Expect(0, 65023, '\p{^inarabicpresentationformsa}', "");
Expect(0, 65023, '\P{inarabicpresentationformsa}', "");
Expect(1, 65023, '\P{^inarabicpresentationformsa}', "");
Expect(0, 65024, '\p{inarabicpresentationformsa}', "");
Expect(1, 65024, '\p{^inarabicpresentationformsa}', "");
Expect(1, 65024, '\P{inarabicpresentationformsa}', "");
Expect(0, 65024, '\P{^inarabicpresentationformsa}', "");
Expect(1, 65023, '\p{_	In_Arabic_Presentation_Forms_A}', "");
Expect(0, 65023, '\p{^_	In_Arabic_Presentation_Forms_A}', "");
Expect(0, 65023, '\P{_	In_Arabic_Presentation_Forms_A}', "");
Expect(1, 65023, '\P{^_	In_Arabic_Presentation_Forms_A}', "");
Expect(0, 65024, '\p{_	In_Arabic_Presentation_Forms_A}', "");
Expect(1, 65024, '\p{^_	In_Arabic_Presentation_Forms_A}', "");
Expect(1, 65024, '\P{_	In_Arabic_Presentation_Forms_A}', "");
Expect(0, 65024, '\P{^_	In_Arabic_Presentation_Forms_A}', "");
Error('\p{	:=ARABIC_pf_A}');
Error('\P{	:=ARABIC_pf_A}');
Expect(1, 65023, '\p{arabicpfa}', "");
Expect(0, 65023, '\p{^arabicpfa}', "");
Expect(0, 65023, '\P{arabicpfa}', "");
Expect(1, 65023, '\P{^arabicpfa}', "");
Expect(0, 65024, '\p{arabicpfa}', "");
Expect(1, 65024, '\p{^arabicpfa}', "");
Expect(1, 65024, '\P{arabicpfa}', "");
Expect(0, 65024, '\P{^arabicpfa}', "");
Expect(1, 65023, '\p{- Arabic_pf_A}', "");
Expect(0, 65023, '\p{^- Arabic_pf_A}', "");
Expect(0, 65023, '\P{- Arabic_pf_A}', "");
Expect(1, 65023, '\P{^- Arabic_pf_A}', "");
Expect(0, 65024, '\p{- Arabic_pf_A}', "");
Expect(1, 65024, '\p{^- Arabic_pf_A}', "");
Expect(1, 65024, '\P{- Arabic_pf_A}', "");
Expect(0, 65024, '\P{^- Arabic_pf_A}', "");
Error('\p{/a/-	Is_ARABIC_PF_A}');
Error('\P{/a/-	Is_ARABIC_PF_A}');
Expect(1, 65023, '\p{isarabicpfa}', "");
Expect(0, 65023, '\p{^isarabicpfa}', "");
Expect(0, 65023, '\P{isarabicpfa}', "");
Expect(1, 65023, '\P{^isarabicpfa}', "");
Expect(0, 65024, '\p{isarabicpfa}', "");
Expect(1, 65024, '\p{^isarabicpfa}', "");
Expect(1, 65024, '\P{isarabicpfa}', "");
Expect(0, 65024, '\P{^isarabicpfa}', "");
Expect(1, 65023, '\p{	_Is_Arabic_pf_A}', "");
Expect(0, 65023, '\p{^	_Is_Arabic_pf_A}', "");
Expect(0, 65023, '\P{	_Is_Arabic_pf_A}', "");
Expect(1, 65023, '\P{^	_Is_Arabic_pf_A}', "");
Expect(0, 65024, '\p{	_Is_Arabic_pf_A}', "");
Expect(1, 65024, '\p{^	_Is_Arabic_pf_A}', "");
Expect(1, 65024, '\P{	_Is_Arabic_pf_A}', "");
Expect(0, 65024, '\P{^	_Is_Arabic_pf_A}', "");
Error('\p{		In_Arabic_PF_a:=}');
Error('\P{		In_Arabic_PF_a:=}');
Expect(1, 65023, '\p{inarabicpfa}', "");
Expect(0, 65023, '\p{^inarabicpfa}', "");
Expect(0, 65023, '\P{inarabicpfa}', "");
Expect(1, 65023, '\P{^inarabicpfa}', "");
Expect(0, 65024, '\p{inarabicpfa}', "");
Expect(1, 65024, '\p{^inarabicpfa}', "");
Expect(1, 65024, '\P{inarabicpfa}', "");
Expect(0, 65024, '\P{^inarabicpfa}', "");
Expect(1, 65023, '\p{ 	in_ARABIC_pf_A}', "");
Expect(0, 65023, '\p{^ 	in_ARABIC_pf_A}', "");
Expect(0, 65023, '\P{ 	in_ARABIC_pf_A}', "");
Expect(1, 65023, '\P{^ 	in_ARABIC_pf_A}', "");
Expect(0, 65024, '\p{ 	in_ARABIC_pf_A}', "");
Expect(1, 65024, '\p{^ 	in_ARABIC_pf_A}', "");
Expect(1, 65024, '\P{ 	in_ARABIC_pf_A}', "");
Expect(0, 65024, '\P{^ 	in_ARABIC_pf_A}', "");
Error('\p{_	In_BATAK/a/}');
Error('\P{_	In_BATAK/a/}');
Expect(1, 7167, '\p{inbatak}', "");
Expect(0, 7167, '\p{^inbatak}', "");
Expect(0, 7167, '\P{inbatak}', "");
Expect(1, 7167, '\P{^inbatak}', "");
Expect(0, 7168, '\p{inbatak}', "");
Expect(1, 7168, '\p{^inbatak}', "");
Expect(1, 7168, '\P{inbatak}', "");
Expect(0, 7168, '\P{^inbatak}', "");
Expect(1, 7167, '\p{__in_BATAK}', "");
Expect(0, 7167, '\p{^__in_BATAK}', "");
Expect(0, 7167, '\P{__in_BATAK}', "");
Expect(1, 7167, '\P{^__in_BATAK}', "");
Expect(0, 7168, '\p{__in_BATAK}', "");
Expect(1, 7168, '\p{^__in_BATAK}', "");
Expect(1, 7168, '\P{__in_BATAK}', "");
Expect(0, 7168, '\P{^__in_BATAK}', "");
Error('\p{_	IN_batak/a/}');
Error('\P{_	IN_batak/a/}');
Expect(1, 7167, '\p{	_In_BATAK}', "");
Expect(0, 7167, '\p{^	_In_BATAK}', "");
Expect(0, 7167, '\P{	_In_BATAK}', "");
Expect(1, 7167, '\P{^	_In_BATAK}', "");
Expect(0, 7168, '\p{	_In_BATAK}', "");
Expect(1, 7168, '\p{^	_In_BATAK}', "");
Expect(1, 7168, '\P{	_In_BATAK}', "");
Expect(0, 7168, '\P{^	_In_BATAK}', "");
Error('\p{/a/balinese}');
Error('\P{/a/balinese}');
Expect(1, 7036, '\p{balinese}', "");
Expect(0, 7036, '\p{^balinese}', "");
Expect(0, 7036, '\P{balinese}', "");
Expect(1, 7036, '\P{^balinese}', "");
Expect(0, 7037, '\p{balinese}', "");
Expect(1, 7037, '\p{^balinese}', "");
Expect(1, 7037, '\P{balinese}', "");
Expect(0, 7037, '\P{^balinese}', "");
Expect(1, 7036, '\p{-balinese}', "");
Expect(0, 7036, '\p{^-balinese}', "");
Expect(0, 7036, '\P{-balinese}', "");
Expect(1, 7036, '\P{^-balinese}', "");
Expect(0, 7037, '\p{-balinese}', "");
Expect(1, 7037, '\p{^-balinese}', "");
Expect(1, 7037, '\P{-balinese}', "");
Expect(0, 7037, '\P{^-balinese}', "");
Error('\p{ /a/is_Balinese}');
Error('\P{ /a/is_Balinese}');
Expect(1, 7036, '\p{isbalinese}', "");
Expect(0, 7036, '\p{^isbalinese}', "");
Expect(0, 7036, '\P{isbalinese}', "");
Expect(1, 7036, '\P{^isbalinese}', "");
Expect(0, 7037, '\p{isbalinese}', "");
Expect(1, 7037, '\p{^isbalinese}', "");
Expect(1, 7037, '\P{isbalinese}', "");
Expect(0, 7037, '\P{^isbalinese}', "");
Expect(1, 7036, '\p{-is_BALINESE}', "");
Expect(0, 7036, '\p{^-is_BALINESE}', "");
Expect(0, 7036, '\P{-is_BALINESE}', "");
Expect(1, 7036, '\P{^-is_BALINESE}', "");
Expect(0, 7037, '\p{-is_BALINESE}', "");
Expect(1, 7037, '\p{^-is_BALINESE}', "");
Expect(1, 7037, '\P{-is_BALINESE}', "");
Expect(0, 7037, '\P{^-is_BALINESE}', "");
Error('\p{-Bali/a/}');
Error('\P{-Bali/a/}');
Expect(1, 7036, '\p{bali}', "");
Expect(0, 7036, '\p{^bali}', "");
Expect(0, 7036, '\P{bali}', "");
Expect(1, 7036, '\P{^bali}', "");
Expect(0, 7037, '\p{bali}', "");
Expect(1, 7037, '\p{^bali}', "");
Expect(1, 7037, '\P{bali}', "");
Expect(0, 7037, '\P{^bali}', "");
Expect(1, 7036, '\p{-	Bali}', "");
Expect(0, 7036, '\p{^-	Bali}', "");
Expect(0, 7036, '\P{-	Bali}', "");
Expect(1, 7036, '\P{^-	Bali}', "");
Expect(0, 7037, '\p{-	Bali}', "");
Expect(1, 7037, '\p{^-	Bali}', "");
Expect(1, 7037, '\P{-	Bali}', "");
Expect(0, 7037, '\P{^-	Bali}', "");
Error('\p{__Is_bali/a/}');
Error('\P{__Is_bali/a/}');
Expect(1, 7036, '\p{isbali}', "");
Expect(0, 7036, '\p{^isbali}', "");
Expect(0, 7036, '\P{isbali}', "");
Expect(1, 7036, '\P{^isbali}', "");
Expect(0, 7037, '\p{isbali}', "");
Expect(1, 7037, '\p{^isbali}', "");
Expect(1, 7037, '\P{isbali}', "");
Expect(0, 7037, '\P{^isbali}', "");
Expect(1, 7036, '\p{--Is_bali}', "");
Expect(0, 7036, '\p{^--Is_bali}', "");
Expect(0, 7036, '\P{--Is_bali}', "");
Expect(1, 7036, '\P{^--Is_bali}', "");
Expect(0, 7037, '\p{--Is_bali}', "");
Expect(1, 7037, '\p{^--Is_bali}', "");
Expect(1, 7037, '\P{--Is_bali}', "");
Expect(0, 7037, '\P{^--Is_bali}', "");
Error('\p{/a/Print}');
Error('\P{/a/Print}');
Expect(1, 1114109, '\p{print}', "");
Expect(0, 1114109, '\p{^print}', "");
Expect(0, 1114109, '\P{print}', "");
Expect(1, 1114109, '\P{^print}', "");
Expect(0, 918000, '\p{print}', "");
Expect(1, 918000, '\p{^print}', "");
Expect(1, 918000, '\P{print}', "");
Expect(0, 918000, '\P{^print}', "");
Expect(1, 1114109, '\p{- Print}', "");
Expect(0, 1114109, '\p{^- Print}', "");
Expect(0, 1114109, '\P{- Print}', "");
Expect(1, 1114109, '\P{^- Print}', "");
Expect(0, 918000, '\p{- Print}', "");
Expect(1, 918000, '\p{^- Print}', "");
Expect(1, 918000, '\P{- Print}', "");
Expect(0, 918000, '\P{^- Print}', "");
Error('\p{	/a/XPosixPrint}');
Error('\P{	/a/XPosixPrint}');
Expect(1, 1114109, '\p{xposixprint}', "");
Expect(0, 1114109, '\p{^xposixprint}', "");
Expect(0, 1114109, '\P{xposixprint}', "");
Expect(1, 1114109, '\P{^xposixprint}', "");
Expect(0, 918000, '\p{xposixprint}', "");
Expect(1, 918000, '\p{^xposixprint}', "");
Expect(1, 918000, '\P{xposixprint}', "");
Expect(0, 918000, '\P{^xposixprint}', "");
Expect(1, 1114109, '\p{	XPOSIXPRINT}', "");
Expect(0, 1114109, '\p{^	XPOSIXPRINT}', "");
Expect(0, 1114109, '\P{	XPOSIXPRINT}', "");
Expect(1, 1114109, '\P{^	XPOSIXPRINT}', "");
Expect(0, 918000, '\p{	XPOSIXPRINT}', "");
Expect(1, 918000, '\p{^	XPOSIXPRINT}', "");
Expect(1, 918000, '\P{	XPOSIXPRINT}', "");
Expect(0, 918000, '\P{^	XPOSIXPRINT}', "");
Error('\p{-Is_Print/a/}');
Error('\P{-Is_Print/a/}');
Expect(1, 1114109, '\p{isprint}', "");
Expect(0, 1114109, '\p{^isprint}', "");
Expect(0, 1114109, '\P{isprint}', "");
Expect(1, 1114109, '\P{^isprint}', "");
Expect(0, 918000, '\p{isprint}', "");
Expect(1, 918000, '\p{^isprint}', "");
Expect(1, 918000, '\P{isprint}', "");
Expect(0, 918000, '\P{^isprint}', "");
Expect(1, 1114109, '\p{_-Is_Print}', "");
Expect(0, 1114109, '\p{^_-Is_Print}', "");
Expect(0, 1114109, '\P{_-Is_Print}', "");
Expect(1, 1114109, '\P{^_-Is_Print}', "");
Expect(0, 918000, '\p{_-Is_Print}', "");
Expect(1, 918000, '\p{^_-Is_Print}', "");
Expect(1, 918000, '\P{_-Is_Print}', "");
Expect(0, 918000, '\P{^_-Is_Print}', "");
Error('\p{	/a/Is_XPOSIXPRINT}');
Error('\P{	/a/Is_XPOSIXPRINT}');
Expect(1, 1114109, '\p{isxposixprint}', "");
Expect(0, 1114109, '\p{^isxposixprint}', "");
Expect(0, 1114109, '\P{isxposixprint}', "");
Expect(1, 1114109, '\P{^isxposixprint}', "");
Expect(0, 918000, '\p{isxposixprint}', "");
Expect(1, 918000, '\p{^isxposixprint}', "");
Expect(1, 918000, '\P{isxposixprint}', "");
Expect(0, 918000, '\P{^isxposixprint}', "");
Expect(1, 1114109, '\p{ Is_XPOSIXPRINT}', "");
Expect(0, 1114109, '\p{^ Is_XPOSIXPRINT}', "");
Expect(0, 1114109, '\P{ Is_XPOSIXPRINT}', "");
Expect(1, 1114109, '\P{^ Is_XPOSIXPRINT}', "");
Expect(0, 918000, '\p{ Is_XPOSIXPRINT}', "");
Expect(1, 918000, '\p{^ Is_XPOSIXPRINT}', "");
Expect(1, 918000, '\P{ Is_XPOSIXPRINT}', "");
Expect(0, 918000, '\P{^ Is_XPOSIXPRINT}', "");
Error('\p{-:=IDS_Trinary_Operator}');
Error('\P{-:=IDS_Trinary_Operator}');
Expect(1, 12275, '\p{idstrinaryoperator}', "");
Expect(0, 12275, '\p{^idstrinaryoperator}', "");
Expect(0, 12275, '\P{idstrinaryoperator}', "");
Expect(1, 12275, '\P{^idstrinaryoperator}', "");
Expect(0, 12276, '\p{idstrinaryoperator}', "");
Expect(1, 12276, '\p{^idstrinaryoperator}', "");
Expect(1, 12276, '\P{idstrinaryoperator}', "");
Expect(0, 12276, '\P{^idstrinaryoperator}', "");
Expect(1, 12275, '\p{  IDS_Trinary_Operator}', "");
Expect(0, 12275, '\p{^  IDS_Trinary_Operator}', "");
Expect(0, 12275, '\P{  IDS_Trinary_Operator}', "");
Expect(1, 12275, '\P{^  IDS_Trinary_Operator}', "");
Expect(0, 12276, '\p{  IDS_Trinary_Operator}', "");
Expect(1, 12276, '\p{^  IDS_Trinary_Operator}', "");
Expect(1, 12276, '\P{  IDS_Trinary_Operator}', "");
Expect(0, 12276, '\P{^  IDS_Trinary_Operator}', "");
Error('\p{_:=Is_IDS_trinary_Operator}');
Error('\P{_:=Is_IDS_trinary_Operator}');
Expect(1, 12275, '\p{isidstrinaryoperator}', "");
Expect(0, 12275, '\p{^isidstrinaryoperator}', "");
Expect(0, 12275, '\P{isidstrinaryoperator}', "");
Expect(1, 12275, '\P{^isidstrinaryoperator}', "");
Expect(0, 12276, '\p{isidstrinaryoperator}', "");
Expect(1, 12276, '\p{^isidstrinaryoperator}', "");
Expect(1, 12276, '\P{isidstrinaryoperator}', "");
Expect(0, 12276, '\P{^isidstrinaryoperator}', "");
Expect(1, 12275, '\p{	 Is_IDS_Trinary_operator}', "");
Expect(0, 12275, '\p{^	 Is_IDS_Trinary_operator}', "");
Expect(0, 12275, '\P{	 Is_IDS_Trinary_operator}', "");
Expect(1, 12275, '\P{^	 Is_IDS_Trinary_operator}', "");
Expect(0, 12276, '\p{	 Is_IDS_Trinary_operator}', "");
Expect(1, 12276, '\p{^	 Is_IDS_Trinary_operator}', "");
Expect(1, 12276, '\P{	 Is_IDS_Trinary_operator}', "");
Expect(0, 12276, '\P{^	 Is_IDS_Trinary_operator}', "");
Error('\p{_:=IDST}');
Error('\P{_:=IDST}');
Expect(1, 12275, '\p{idst}', "");
Expect(0, 12275, '\p{^idst}', "");
Expect(0, 12275, '\P{idst}', "");
Expect(1, 12275, '\P{^idst}', "");
Expect(0, 12276, '\p{idst}', "");
Expect(1, 12276, '\p{^idst}', "");
Expect(1, 12276, '\P{idst}', "");
Expect(0, 12276, '\P{^idst}', "");
Expect(1, 12275, '\p{-_IDST}', "");
Expect(0, 12275, '\p{^-_IDST}', "");
Expect(0, 12275, '\P{-_IDST}', "");
Expect(1, 12275, '\P{^-_IDST}', "");
Expect(0, 12276, '\p{-_IDST}', "");
Expect(1, 12276, '\p{^-_IDST}', "");
Expect(1, 12276, '\P{-_IDST}', "");
Expect(0, 12276, '\P{^-_IDST}', "");
Error('\p{_	Is_IDST/a/}');
Error('\P{_	Is_IDST/a/}');
Expect(1, 12275, '\p{isidst}', "");
Expect(0, 12275, '\p{^isidst}', "");
Expect(0, 12275, '\P{isidst}', "");
Expect(1, 12275, '\P{^isidst}', "");
Expect(0, 12276, '\p{isidst}', "");
Expect(1, 12276, '\p{^isidst}', "");
Expect(1, 12276, '\P{isidst}', "");
Expect(0, 12276, '\P{^isidst}', "");
Expect(1, 12275, '\p{	is_IDST}', "");
Expect(0, 12275, '\p{^	is_IDST}', "");
Expect(0, 12275, '\P{	is_IDST}', "");
Expect(1, 12275, '\P{^	is_IDST}', "");
Expect(0, 12276, '\p{	is_IDST}', "");
Expect(1, 12276, '\p{^	is_IDST}', "");
Expect(1, 12276, '\P{	is_IDST}', "");
Expect(0, 12276, '\P{^	is_IDST}', "");
Error('\p{-:=in_VAI}');
Error('\P{-:=in_VAI}');
Expect(1, 42559, '\p{invai}', "");
Expect(0, 42559, '\p{^invai}', "");
Expect(0, 42559, '\P{invai}', "");
Expect(1, 42559, '\P{^invai}', "");
Expect(0, 42560, '\p{invai}', "");
Expect(1, 42560, '\p{^invai}', "");
Expect(1, 42560, '\P{invai}', "");
Expect(0, 42560, '\P{^invai}', "");
Expect(1, 42559, '\p{-_In_Vai}', "");
Expect(0, 42559, '\p{^-_In_Vai}', "");
Expect(0, 42559, '\P{-_In_Vai}', "");
Expect(1, 42559, '\P{^-_In_Vai}', "");
Expect(0, 42560, '\p{-_In_Vai}', "");
Expect(1, 42560, '\p{^-_In_Vai}', "");
Expect(1, 42560, '\P{-_In_Vai}', "");
Expect(0, 42560, '\P{^-_In_Vai}', "");
Error('\p{	In_VAI/a/}');
Error('\P{	In_VAI/a/}');
Expect(1, 42559, '\p{	in_VAI}', "");
Expect(0, 42559, '\p{^	in_VAI}', "");
Expect(0, 42559, '\P{	in_VAI}', "");
Expect(1, 42559, '\P{^	in_VAI}', "");
Expect(0, 42560, '\p{	in_VAI}', "");
Expect(1, 42560, '\p{^	in_VAI}', "");
Expect(1, 42560, '\P{	in_VAI}', "");
Expect(0, 42560, '\P{^	in_VAI}', "");
Error('\p{ thaana:=}');
Error('\P{ thaana:=}');
Expect(1, 1969, '\p{thaana}', "");
Expect(0, 1969, '\p{^thaana}', "");
Expect(0, 1969, '\P{thaana}', "");
Expect(1, 1969, '\P{^thaana}', "");
Expect(0, 1970, '\p{thaana}', "");
Expect(1, 1970, '\p{^thaana}', "");
Expect(1, 1970, '\P{thaana}', "");
Expect(0, 1970, '\P{^thaana}', "");
Expect(1, 1969, '\p{_-Thaana}', "");
Expect(0, 1969, '\p{^_-Thaana}', "");
Expect(0, 1969, '\P{_-Thaana}', "");
Expect(1, 1969, '\P{^_-Thaana}', "");
Expect(0, 1970, '\p{_-Thaana}', "");
Expect(1, 1970, '\p{^_-Thaana}', "");
Expect(1, 1970, '\P{_-Thaana}', "");
Expect(0, 1970, '\P{^_-Thaana}', "");
Error('\p{/a/	 IS_thaana}');
Error('\P{/a/	 IS_thaana}');
Expect(1, 1969, '\p{isthaana}', "");
Expect(0, 1969, '\p{^isthaana}', "");
Expect(0, 1969, '\P{isthaana}', "");
Expect(1, 1969, '\P{^isthaana}', "");
Expect(0, 1970, '\p{isthaana}', "");
Expect(1, 1970, '\p{^isthaana}', "");
Expect(1, 1970, '\P{isthaana}', "");
Expect(0, 1970, '\P{^isthaana}', "");
Expect(1, 1969, '\p{-Is_Thaana}', "");
Expect(0, 1969, '\p{^-Is_Thaana}', "");
Expect(0, 1969, '\P{-Is_Thaana}', "");
Expect(1, 1969, '\P{^-Is_Thaana}', "");
Expect(0, 1970, '\p{-Is_Thaana}', "");
Expect(1, 1970, '\p{^-Is_Thaana}', "");
Expect(1, 1970, '\P{-Is_Thaana}', "");
Expect(0, 1970, '\P{^-Is_Thaana}', "");
Error('\p{:=-thaa}');
Error('\P{:=-thaa}');
Expect(1, 1969, '\p{thaa}', "");
Expect(0, 1969, '\p{^thaa}', "");
Expect(0, 1969, '\P{thaa}', "");
Expect(1, 1969, '\P{^thaa}', "");
Expect(0, 1970, '\p{thaa}', "");
Expect(1, 1970, '\p{^thaa}', "");
Expect(1, 1970, '\P{thaa}', "");
Expect(0, 1970, '\P{^thaa}', "");
Expect(1, 1969, '\p{ _Thaa}', "");
Expect(0, 1969, '\p{^ _Thaa}', "");
Expect(0, 1969, '\P{ _Thaa}', "");
Expect(1, 1969, '\P{^ _Thaa}', "");
Expect(0, 1970, '\p{ _Thaa}', "");
Expect(1, 1970, '\p{^ _Thaa}', "");
Expect(1, 1970, '\P{ _Thaa}', "");
Expect(0, 1970, '\P{^ _Thaa}', "");
Error('\p{/a/	 Is_THAA}');
Error('\P{/a/	 Is_THAA}');
Expect(1, 1969, '\p{isthaa}', "");
Expect(0, 1969, '\p{^isthaa}', "");
Expect(0, 1969, '\P{isthaa}', "");
Expect(1, 1969, '\P{^isthaa}', "");
Expect(0, 1970, '\p{isthaa}', "");
Expect(1, 1970, '\p{^isthaa}', "");
Expect(1, 1970, '\P{isthaa}', "");
Expect(0, 1970, '\P{^isthaa}', "");
Expect(1, 1969, '\p{-_is_Thaa}', "");
Expect(0, 1969, '\p{^-_is_Thaa}', "");
Expect(0, 1969, '\P{-_is_Thaa}', "");
Expect(1, 1969, '\P{^-_is_Thaa}', "");
Expect(0, 1970, '\p{-_is_Thaa}', "");
Expect(1, 1970, '\p{^-_is_Thaa}', "");
Expect(1, 1970, '\P{-_is_Thaa}', "");
Expect(0, 1970, '\P{^-_is_Thaa}', "");
Error('\p{Other_ID_Start}');
Error('\P{Other_ID_Start}');
Error('\p{Is_Other_ID_Start}');
Error('\P{Is_Other_ID_Start}');
Error('\p{OIDS}');
Error('\P{OIDS}');
Error('\p{Is_OIDS}');
Error('\P{Is_OIDS}');
Error('\p{:=graph}');
Error('\P{:=graph}');
Expect(1, 1114109, '\p{graph}', "");
Expect(0, 1114109, '\p{^graph}', "");
Expect(0, 1114109, '\P{graph}', "");
Expect(1, 1114109, '\P{^graph}', "");
Expect(0, 918000, '\p{graph}', "");
Expect(1, 918000, '\p{^graph}', "");
Expect(1, 918000, '\P{graph}', "");
Expect(0, 918000, '\P{^graph}', "");
Expect(1, 1114109, '\p{	GRAPH}', "");
Expect(0, 1114109, '\p{^	GRAPH}', "");
Expect(0, 1114109, '\P{	GRAPH}', "");
Expect(1, 1114109, '\P{^	GRAPH}', "");
Expect(0, 918000, '\p{	GRAPH}', "");
Expect(1, 918000, '\p{^	GRAPH}', "");
Expect(1, 918000, '\P{	GRAPH}', "");
Expect(0, 918000, '\P{^	GRAPH}', "");
Error('\p{ :=XPosixGraph}');
Error('\P{ :=XPosixGraph}');
Expect(1, 1114109, '\p{xposixgraph}', "");
Expect(0, 1114109, '\p{^xposixgraph}', "");
Expect(0, 1114109, '\P{xposixgraph}', "");
Expect(1, 1114109, '\P{^xposixgraph}', "");
Expect(0, 918000, '\p{xposixgraph}', "");
Expect(1, 918000, '\p{^xposixgraph}', "");
Expect(1, 918000, '\P{xposixgraph}', "");
Expect(0, 918000, '\P{^xposixgraph}', "");
Expect(1, 1114109, '\p{	 xposixgraph}', "");
Expect(0, 1114109, '\p{^	 xposixgraph}', "");
Expect(0, 1114109, '\P{	 xposixgraph}', "");
Expect(1, 1114109, '\P{^	 xposixgraph}', "");
Expect(0, 918000, '\p{	 xposixgraph}', "");
Expect(1, 918000, '\p{^	 xposixgraph}', "");
Expect(1, 918000, '\P{	 xposixgraph}', "");
Expect(0, 918000, '\P{^	 xposixgraph}', "");
Error('\p{-/a/IS_Graph}');
Error('\P{-/a/IS_Graph}');
Expect(1, 1114109, '\p{isgraph}', "");
Expect(0, 1114109, '\p{^isgraph}', "");
Expect(0, 1114109, '\P{isgraph}', "");
Expect(1, 1114109, '\P{^isgraph}', "");
Expect(0, 918000, '\p{isgraph}', "");
Expect(1, 918000, '\p{^isgraph}', "");
Expect(1, 918000, '\P{isgraph}', "");
Expect(0, 918000, '\P{^isgraph}', "");
Expect(1, 1114109, '\p{-	Is_Graph}', "");
Expect(0, 1114109, '\p{^-	Is_Graph}', "");
Expect(0, 1114109, '\P{-	Is_Graph}', "");
Expect(1, 1114109, '\P{^-	Is_Graph}', "");
Expect(0, 918000, '\p{-	Is_Graph}', "");
Expect(1, 918000, '\p{^-	Is_Graph}', "");
Expect(1, 918000, '\P{-	Is_Graph}', "");
Expect(0, 918000, '\P{^-	Is_Graph}', "");
Error('\p{:=- Is_XPosixGraph}');
Error('\P{:=- Is_XPosixGraph}');
Expect(1, 1114109, '\p{isxposixgraph}', "");
Expect(0, 1114109, '\p{^isxposixgraph}', "");
Expect(0, 1114109, '\P{isxposixgraph}', "");
Expect(1, 1114109, '\P{^isxposixgraph}', "");
Expect(0, 918000, '\p{isxposixgraph}', "");
Expect(1, 918000, '\p{^isxposixgraph}', "");
Expect(1, 918000, '\P{isxposixgraph}', "");
Expect(0, 918000, '\P{^isxposixgraph}', "");
Expect(1, 1114109, '\p{ Is_xposixgraph}', "");
Expect(0, 1114109, '\p{^ Is_xposixgraph}', "");
Expect(0, 1114109, '\P{ Is_xposixgraph}', "");
Expect(1, 1114109, '\P{^ Is_xposixgraph}', "");
Expect(0, 918000, '\p{ Is_xposixgraph}', "");
Expect(1, 918000, '\p{^ Is_xposixgraph}', "");
Expect(1, 918000, '\P{ Is_xposixgraph}', "");
Expect(0, 918000, '\P{^ Is_xposixgraph}', "");
Error('\p{-:=Specials}');
Error('\P{-:=Specials}');
Expect(1, 65520, '\p{specials}', "");
Expect(0, 65520, '\p{^specials}', "");
Expect(0, 65520, '\P{specials}', "");
Expect(1, 65520, '\P{^specials}', "");
Expect(0, 65536, '\p{specials}', "");
Expect(1, 65536, '\p{^specials}', "");
Expect(1, 65536, '\P{specials}', "");
Expect(0, 65536, '\P{^specials}', "");
Expect(1, 65520, '\p{	 SPECIALS}', "");
Expect(0, 65520, '\p{^	 SPECIALS}', "");
Expect(0, 65520, '\P{	 SPECIALS}', "");
Expect(1, 65520, '\P{^	 SPECIALS}', "");
Expect(0, 65536, '\p{	 SPECIALS}', "");
Expect(1, 65536, '\p{^	 SPECIALS}', "");
Expect(1, 65536, '\P{	 SPECIALS}', "");
Expect(0, 65536, '\P{^	 SPECIALS}', "");
Error('\p{	/a/Is_Specials}');
Error('\P{	/a/Is_Specials}');
Expect(1, 65520, '\p{isspecials}', "");
Expect(0, 65520, '\p{^isspecials}', "");
Expect(0, 65520, '\P{isspecials}', "");
Expect(1, 65520, '\P{^isspecials}', "");
Expect(0, 65536, '\p{isspecials}', "");
Expect(1, 65536, '\p{^isspecials}', "");
Expect(1, 65536, '\P{isspecials}', "");
Expect(0, 65536, '\P{^isspecials}', "");
Expect(1, 65520, '\p{_-Is_specials}', "");
Expect(0, 65520, '\p{^_-Is_specials}', "");
Expect(0, 65520, '\P{_-Is_specials}', "");
Expect(1, 65520, '\P{^_-Is_specials}', "");
Expect(0, 65536, '\p{_-Is_specials}', "");
Expect(1, 65536, '\p{^_-Is_specials}', "");
Expect(1, 65536, '\P{_-Is_specials}', "");
Expect(0, 65536, '\P{^_-Is_specials}', "");
Error('\p{:=--In_specials}');
Error('\P{:=--In_specials}');
Expect(1, 65520, '\p{inspecials}', "");
Expect(0, 65520, '\p{^inspecials}', "");
Expect(0, 65520, '\P{inspecials}', "");
Expect(1, 65520, '\P{^inspecials}', "");
Expect(0, 65536, '\p{inspecials}', "");
Expect(1, 65536, '\p{^inspecials}', "");
Expect(1, 65536, '\P{inspecials}', "");
Expect(0, 65536, '\P{^inspecials}', "");
Expect(1, 65520, '\p{ IN_specials}', "");
Expect(0, 65520, '\p{^ IN_specials}', "");
Expect(0, 65520, '\P{ IN_specials}', "");
Expect(1, 65520, '\P{^ IN_specials}', "");
Expect(0, 65536, '\p{ IN_specials}', "");
Expect(1, 65536, '\p{^ IN_specials}', "");
Expect(1, 65536, '\P{ IN_specials}', "");
Expect(0, 65536, '\P{^ IN_specials}', "");
Error('\p{:=Bidi_mirrored}');
Error('\P{:=Bidi_mirrored}');
Expect(1, 120771, '\p{bidimirrored}', "");
Expect(0, 120771, '\p{^bidimirrored}', "");
Expect(0, 120771, '\P{bidimirrored}', "");
Expect(1, 120771, '\P{^bidimirrored}', "");
Expect(0, 120772, '\p{bidimirrored}', "");
Expect(1, 120772, '\p{^bidimirrored}', "");
Expect(1, 120772, '\P{bidimirrored}', "");
Expect(0, 120772, '\P{^bidimirrored}', "");
Expect(1, 120771, '\p{-_Bidi_Mirrored}', "");
Expect(0, 120771, '\p{^-_Bidi_Mirrored}', "");
Expect(0, 120771, '\P{-_Bidi_Mirrored}', "");
Expect(1, 120771, '\P{^-_Bidi_Mirrored}', "");
Expect(0, 120772, '\p{-_Bidi_Mirrored}', "");
Expect(1, 120772, '\p{^-_Bidi_Mirrored}', "");
Expect(1, 120772, '\P{-_Bidi_Mirrored}', "");
Expect(0, 120772, '\P{^-_Bidi_Mirrored}', "");
Error('\p{--Is_Bidi_mirrored/a/}');
Error('\P{--Is_Bidi_mirrored/a/}');
Expect(1, 120771, '\p{isbidimirrored}', "");
Expect(0, 120771, '\p{^isbidimirrored}', "");
Expect(0, 120771, '\P{isbidimirrored}', "");
Expect(1, 120771, '\P{^isbidimirrored}', "");
Expect(0, 120772, '\p{isbidimirrored}', "");
Expect(1, 120772, '\p{^isbidimirrored}', "");
Expect(1, 120772, '\P{isbidimirrored}', "");
Expect(0, 120772, '\P{^isbidimirrored}', "");
Expect(1, 120771, '\p{_-Is_Bidi_Mirrored}', "");
Expect(0, 120771, '\p{^_-Is_Bidi_Mirrored}', "");
Expect(0, 120771, '\P{_-Is_Bidi_Mirrored}', "");
Expect(1, 120771, '\P{^_-Is_Bidi_Mirrored}', "");
Expect(0, 120772, '\p{_-Is_Bidi_Mirrored}', "");
Expect(1, 120772, '\p{^_-Is_Bidi_Mirrored}', "");
Expect(1, 120772, '\P{_-Is_Bidi_Mirrored}', "");
Expect(0, 120772, '\P{^_-Is_Bidi_Mirrored}', "");
Error('\p{	/a/BIDI_M}');
Error('\P{	/a/BIDI_M}');
Expect(1, 120771, '\p{bidim}', "");
Expect(0, 120771, '\p{^bidim}', "");
Expect(0, 120771, '\P{bidim}', "");
Expect(1, 120771, '\P{^bidim}', "");
Expect(0, 120772, '\p{bidim}', "");
Expect(1, 120772, '\p{^bidim}', "");
Expect(1, 120772, '\P{bidim}', "");
Expect(0, 120772, '\P{^bidim}', "");
Expect(1, 120771, '\p{ 	Bidi_M}', "");
Expect(0, 120771, '\p{^ 	Bidi_M}', "");
Expect(0, 120771, '\P{ 	Bidi_M}', "");
Expect(1, 120771, '\P{^ 	Bidi_M}', "");
Expect(0, 120772, '\p{ 	Bidi_M}', "");
Expect(1, 120772, '\p{^ 	Bidi_M}', "");
Expect(1, 120772, '\P{ 	Bidi_M}', "");
Expect(0, 120772, '\P{^ 	Bidi_M}', "");
Error('\p{_/a/is_Bidi_m}');
Error('\P{_/a/is_Bidi_m}');
Expect(1, 120771, '\p{isbidim}', "");
Expect(0, 120771, '\p{^isbidim}', "");
Expect(0, 120771, '\P{isbidim}', "");
Expect(1, 120771, '\P{^isbidim}', "");
Expect(0, 120772, '\p{isbidim}', "");
Expect(1, 120772, '\p{^isbidim}', "");
Expect(1, 120772, '\P{isbidim}', "");
Expect(0, 120772, '\P{^isbidim}', "");
Expect(1, 120771, '\p{Is_bidi_M}', "");
Expect(0, 120771, '\p{^Is_bidi_M}', "");
Expect(0, 120771, '\P{Is_bidi_M}', "");
Expect(1, 120771, '\P{^Is_bidi_M}', "");
Expect(0, 120772, '\p{Is_bidi_M}', "");
Expect(1, 120772, '\p{^Is_bidi_M}', "");
Expect(1, 120772, '\P{Is_bidi_M}', "");
Expect(0, 120772, '\P{^Is_bidi_M}', "");
Error('\p{Katakana_Or_Hiragana}');
Error('\P{Katakana_Or_Hiragana}');
Error('\p{Is_Katakana_Or_Hiragana}');
Error('\P{Is_Katakana_Or_Hiragana}');
Error('\p{Hrkt}');
Error('\P{Hrkt}');
Error('\p{Is_Hrkt}');
Error('\P{Is_Hrkt}');
Error('\p{:=	SUPPLEMENTAL_mathematical_Operators}');
Error('\P{:=	SUPPLEMENTAL_mathematical_Operators}');
Expect(1, 11007, '\p{supplementalmathematicaloperators}', "");
Expect(0, 11007, '\p{^supplementalmathematicaloperators}', "");
Expect(0, 11007, '\P{supplementalmathematicaloperators}', "");
Expect(1, 11007, '\P{^supplementalmathematicaloperators}', "");
Expect(0, 11008, '\p{supplementalmathematicaloperators}', "");
Expect(1, 11008, '\p{^supplementalmathematicaloperators}', "");
Expect(1, 11008, '\P{supplementalmathematicaloperators}', "");
Expect(0, 11008, '\P{^supplementalmathematicaloperators}', "");
Expect(1, 11007, '\p{_SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(0, 11007, '\p{^_SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(0, 11007, '\P{_SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(1, 11007, '\P{^_SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(0, 11008, '\p{_SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(1, 11008, '\p{^_SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(1, 11008, '\P{_SUPPLEMENTAL_Mathematical_Operators}', "");
Expect(0, 11008, '\P{^_SUPPLEMENTAL_Mathematical_Operators}', "");
Error('\p{_-IS_Supplemental_MATHEMATICAL_operators:=}');
Error('\P{_-IS_Supplemental_MATHEMATICAL_operators:=}');
Expect(1, 11007, '\p{issupplementalmathematicaloperators}', "");
Expect(0, 11007, '\p{^issupplementalmathematicaloperators}', "");
Expect(0, 11007, '\P{issupplementalmathematicaloperators}', "");
Expect(1, 11007, '\P{^issupplementalmathematicaloperators}', "");
Expect(0, 11008, '\p{issupplementalmathematicaloperators}', "");
Expect(1, 11008, '\p{^issupplementalmathematicaloperators}', "");
Expect(1, 11008, '\P{issupplementalmathematicaloperators}', "");
Expect(0, 11008, '\P{^issupplementalmathematicaloperators}', "");
Expect(1, 11007, '\p{		Is_SUPPLEMENTAL_Mathematical_operators}', "");
Expect(0, 11007, '\p{^		Is_SUPPLEMENTAL_Mathematical_operators}', "");
Expect(0, 11007, '\P{		Is_SUPPLEMENTAL_Mathematical_operators}', "");
Expect(1, 11007, '\P{^		Is_SUPPLEMENTAL_Mathematical_operators}', "");
Expect(0, 11008, '\p{		Is_SUPPLEMENTAL_Mathematical_operators}', "");
Expect(1, 11008, '\p{^		Is_SUPPLEMENTAL_Mathematical_operators}', "");
Expect(1, 11008, '\P{		Is_SUPPLEMENTAL_Mathematical_operators}', "");
Expect(0, 11008, '\P{^		Is_SUPPLEMENTAL_Mathematical_operators}', "");
Error('\p{__IN_SUPPLEMENTAL_mathematical_Operators/a/}');
Error('\P{__IN_SUPPLEMENTAL_mathematical_Operators/a/}');
Expect(1, 11007, '\p{insupplementalmathematicaloperators}', "");
Expect(0, 11007, '\p{^insupplementalmathematicaloperators}', "");
Expect(0, 11007, '\P{insupplementalmathematicaloperators}', "");
Expect(1, 11007, '\P{^insupplementalmathematicaloperators}', "");
Expect(0, 11008, '\p{insupplementalmathematicaloperators}', "");
Expect(1, 11008, '\p{^insupplementalmathematicaloperators}', "");
Expect(1, 11008, '\P{insupplementalmathematicaloperators}', "");
Expect(0, 11008, '\P{^insupplementalmathematicaloperators}', "");
Expect(1, 11007, '\p{- IN_Supplemental_Mathematical_Operators}', "");
Expect(0, 11007, '\p{^- IN_Supplemental_Mathematical_Operators}', "");
Expect(0, 11007, '\P{- IN_Supplemental_Mathematical_Operators}', "");
Expect(1, 11007, '\P{^- IN_Supplemental_Mathematical_Operators}', "");
Expect(0, 11008, '\p{- IN_Supplemental_Mathematical_Operators}', "");
Expect(1, 11008, '\p{^- IN_Supplemental_Mathematical_Operators}', "");
Expect(1, 11008, '\P{- IN_Supplemental_Mathematical_Operators}', "");
Expect(0, 11008, '\P{^- IN_Supplemental_Mathematical_Operators}', "");
Error('\p{_:=sup_math_operators}');
Error('\P{_:=sup_math_operators}');
Expect(1, 11007, '\p{supmathoperators}', "");
Expect(0, 11007, '\p{^supmathoperators}', "");
Expect(0, 11007, '\P{supmathoperators}', "");
Expect(1, 11007, '\P{^supmathoperators}', "");
Expect(0, 11008, '\p{supmathoperators}', "");
Expect(1, 11008, '\p{^supmathoperators}', "");
Expect(1, 11008, '\P{supmathoperators}', "");
Expect(0, 11008, '\P{^supmathoperators}', "");
Expect(1, 11007, '\p{-_SUP_MATH_Operators}', "");
Expect(0, 11007, '\p{^-_SUP_MATH_Operators}', "");
Expect(0, 11007, '\P{-_SUP_MATH_Operators}', "");
Expect(1, 11007, '\P{^-_SUP_MATH_Operators}', "");
Expect(0, 11008, '\p{-_SUP_MATH_Operators}', "");
Expect(1, 11008, '\p{^-_SUP_MATH_Operators}', "");
Expect(1, 11008, '\P{-_SUP_MATH_Operators}', "");
Expect(0, 11008, '\P{^-_SUP_MATH_Operators}', "");
Error('\p{/a/IS_SUP_Math_operators}');
Error('\P{/a/IS_SUP_Math_operators}');
Expect(1, 11007, '\p{issupmathoperators}', "");
Expect(0, 11007, '\p{^issupmathoperators}', "");
Expect(0, 11007, '\P{issupmathoperators}', "");
Expect(1, 11007, '\P{^issupmathoperators}', "");
Expect(0, 11008, '\p{issupmathoperators}', "");
Expect(1, 11008, '\p{^issupmathoperators}', "");
Expect(1, 11008, '\P{issupmathoperators}', "");
Expect(0, 11008, '\P{^issupmathoperators}', "");
Expect(1, 11007, '\p{-_IS_Sup_MATH_Operators}', "");
Expect(0, 11007, '\p{^-_IS_Sup_MATH_Operators}', "");
Expect(0, 11007, '\P{-_IS_Sup_MATH_Operators}', "");
Expect(1, 11007, '\P{^-_IS_Sup_MATH_Operators}', "");
Expect(0, 11008, '\p{-_IS_Sup_MATH_Operators}', "");
Expect(1, 11008, '\p{^-_IS_Sup_MATH_Operators}', "");
Expect(1, 11008, '\P{-_IS_Sup_MATH_Operators}', "");
Expect(0, 11008, '\P{^-_IS_Sup_MATH_Operators}', "");
Error('\p{-:=IN_SUP_Math_operators}');
Error('\P{-:=IN_SUP_Math_operators}');
Expect(1, 11007, '\p{insupmathoperators}', "");
Expect(0, 11007, '\p{^insupmathoperators}', "");
Expect(0, 11007, '\P{insupmathoperators}', "");
Expect(1, 11007, '\P{^insupmathoperators}', "");
Expect(0, 11008, '\p{insupmathoperators}', "");
Expect(1, 11008, '\p{^insupmathoperators}', "");
Expect(1, 11008, '\P{insupmathoperators}', "");
Expect(0, 11008, '\P{^insupmathoperators}', "");
Expect(1, 11007, '\p{ -in_Sup_math_Operators}', "");
Expect(0, 11007, '\p{^ -in_Sup_math_Operators}', "");
Expect(0, 11007, '\P{ -in_Sup_math_Operators}', "");
Expect(1, 11007, '\P{^ -in_Sup_math_Operators}', "");
Expect(0, 11008, '\p{ -in_Sup_math_Operators}', "");
Expect(1, 11008, '\p{^ -in_Sup_math_Operators}', "");
Expect(1, 11008, '\P{ -in_Sup_math_Operators}', "");
Expect(0, 11008, '\P{^ -in_Sup_math_Operators}', "");
Error('\p{__OTHER_Symbol/a/}');
Error('\P{__OTHER_Symbol/a/}');
Expect(1, 128883, '\p{othersymbol}', "");
Expect(0, 128883, '\p{^othersymbol}', "");
Expect(0, 128883, '\P{othersymbol}', "");
Expect(1, 128883, '\P{^othersymbol}', "");
Expect(0, 128884, '\p{othersymbol}', "");
Expect(1, 128884, '\p{^othersymbol}', "");
Expect(1, 128884, '\P{othersymbol}', "");
Expect(0, 128884, '\P{^othersymbol}', "");
Expect(1, 128883, '\p{ 	Other_Symbol}', "");
Expect(0, 128883, '\p{^ 	Other_Symbol}', "");
Expect(0, 128883, '\P{ 	Other_Symbol}', "");
Expect(1, 128883, '\P{^ 	Other_Symbol}', "");
Expect(0, 128884, '\p{ 	Other_Symbol}', "");
Expect(1, 128884, '\p{^ 	Other_Symbol}', "");
Expect(1, 128884, '\P{ 	Other_Symbol}', "");
Expect(0, 128884, '\P{^ 	Other_Symbol}', "");
Error('\p{/a/Is_other_Symbol}');
Error('\P{/a/Is_other_Symbol}');
Expect(1, 128883, '\p{isothersymbol}', "");
Expect(0, 128883, '\p{^isothersymbol}', "");
Expect(0, 128883, '\P{isothersymbol}', "");
Expect(1, 128883, '\P{^isothersymbol}', "");
Expect(0, 128884, '\p{isothersymbol}', "");
Expect(1, 128884, '\p{^isothersymbol}', "");
Expect(1, 128884, '\P{isothersymbol}', "");
Expect(0, 128884, '\P{^isothersymbol}', "");
Expect(1, 128883, '\p{ Is_Other_SYMBOL}', "");
Expect(0, 128883, '\p{^ Is_Other_SYMBOL}', "");
Expect(0, 128883, '\P{ Is_Other_SYMBOL}', "");
Expect(1, 128883, '\P{^ Is_Other_SYMBOL}', "");
Expect(0, 128884, '\p{ Is_Other_SYMBOL}', "");
Expect(1, 128884, '\p{^ Is_Other_SYMBOL}', "");
Expect(1, 128884, '\P{ Is_Other_SYMBOL}', "");
Expect(0, 128884, '\P{^ Is_Other_SYMBOL}', "");
Error('\p{/a/	So}');
Error('\P{/a/	So}');
Expect(1, 128883, '\p{so}', "");
Expect(0, 128883, '\p{^so}', "");
Expect(0, 128883, '\P{so}', "");
Expect(1, 128883, '\P{^so}', "");
Expect(0, 128884, '\p{so}', "");
Expect(1, 128884, '\p{^so}', "");
Expect(1, 128884, '\P{so}', "");
Expect(0, 128884, '\P{^so}', "");
Expect(1, 128883, '\p{-so}', "");
Expect(0, 128883, '\p{^-so}', "");
Expect(0, 128883, '\P{-so}', "");
Expect(1, 128883, '\P{^-so}', "");
Expect(0, 128884, '\p{-so}', "");
Expect(1, 128884, '\p{^-so}', "");
Expect(1, 128884, '\P{-so}', "");
Expect(0, 128884, '\P{^-so}', "");
Error('\p{-:=Is_So}');
Error('\P{-:=Is_So}');
Expect(1, 128883, '\p{isso}', "");
Expect(0, 128883, '\p{^isso}', "");
Expect(0, 128883, '\P{isso}', "");
Expect(1, 128883, '\P{^isso}', "");
Expect(0, 128884, '\p{isso}', "");
Expect(1, 128884, '\p{^isso}', "");
Expect(1, 128884, '\P{isso}', "");
Expect(0, 128884, '\P{^isso}', "");
Expect(1, 128883, '\p{_IS_SO}', "");
Expect(0, 128883, '\p{^_IS_SO}', "");
Expect(0, 128883, '\P{_IS_SO}', "");
Expect(1, 128883, '\P{^_IS_SO}', "");
Expect(0, 128884, '\p{_IS_SO}', "");
Expect(1, 128884, '\p{^_IS_SO}', "");
Expect(1, 128884, '\P{_IS_SO}', "");
Expect(0, 128884, '\P{^_IS_SO}', "");
Error('\p{/a/ Any}');
Error('\P{/a/ Any}');
Expect(1, 1, '\p{any}', "");
Expect(0, 1, '\p{^any}', "");
Expect(0, 1, '\P{any}', "");
Expect(1, 1, '\P{^any}', "");
Expect(1, 1, '\p{-_Any}', "");
Expect(0, 1, '\p{^-_Any}', "");
Expect(0, 1, '\P{-_Any}', "");
Expect(1, 1, '\P{^-_Any}', "");
Error('\p{:=_ All}');
Error('\P{:=_ All}');
Expect(1, 1, '\p{all}', "");
Expect(0, 1, '\p{^all}', "");
Expect(0, 1, '\P{all}', "");
Expect(1, 1, '\P{^all}', "");
Expect(1, 1, '\p{		ALL}', "");
Expect(0, 1, '\p{^		ALL}', "");
Expect(0, 1, '\P{		ALL}', "");
Expect(1, 1, '\P{^		ALL}', "");
Error('\p{ IS_Any:=}');
Error('\P{ IS_Any:=}');
Expect(1, 1, '\p{isany}', "");
Expect(0, 1, '\p{^isany}', "");
Expect(0, 1, '\P{isany}', "");
Expect(1, 1, '\P{^isany}', "");
Expect(1, 1, '\p{_is_ANY}', "");
Expect(0, 1, '\p{^_is_ANY}', "");
Expect(0, 1, '\P{_is_ANY}', "");
Expect(1, 1, '\P{^_is_ANY}', "");
Error('\p{:=_IS_ALL}');
Error('\P{:=_IS_ALL}');
Expect(1, 1, '\p{isall}', "");
Expect(0, 1, '\p{^isall}', "");
Expect(0, 1, '\P{isall}', "");
Expect(1, 1, '\P{^isall}', "");
Expect(1, 1, '\p{ Is_all}', "");
Expect(0, 1, '\p{^ Is_all}', "");
Expect(0, 1, '\P{ Is_all}', "");
Expect(1, 1, '\P{^ Is_all}', "");
Error('\p{ 	Other_LETTER/a/}');
Error('\P{ 	Other_LETTER/a/}');
Expect(1, 195101, '\p{otherletter}', "");
Expect(0, 195101, '\p{^otherletter}', "");
Expect(0, 195101, '\P{otherletter}', "");
Expect(1, 195101, '\P{^otherletter}', "");
Expect(0, 195102, '\p{otherletter}', "");
Expect(1, 195102, '\p{^otherletter}', "");
Expect(1, 195102, '\P{otherletter}', "");
Expect(0, 195102, '\P{^otherletter}', "");
Expect(1, 195101, '\p{--other_Letter}', "");
Expect(0, 195101, '\p{^--other_Letter}', "");
Expect(0, 195101, '\P{--other_Letter}', "");
Expect(1, 195101, '\P{^--other_Letter}', "");
Expect(0, 195102, '\p{--other_Letter}', "");
Expect(1, 195102, '\p{^--other_Letter}', "");
Expect(1, 195102, '\P{--other_Letter}', "");
Expect(0, 195102, '\P{^--other_Letter}', "");
Error('\p{  IS_other_Letter/a/}');
Error('\P{  IS_other_Letter/a/}');
Expect(1, 195101, '\p{isotherletter}', "");
Expect(0, 195101, '\p{^isotherletter}', "");
Expect(0, 195101, '\P{isotherletter}', "");
Expect(1, 195101, '\P{^isotherletter}', "");
Expect(0, 195102, '\p{isotherletter}', "");
Expect(1, 195102, '\p{^isotherletter}', "");
Expect(1, 195102, '\P{isotherletter}', "");
Expect(0, 195102, '\P{^isotherletter}', "");
Expect(1, 195101, '\p{Is_other_letter}', "");
Expect(0, 195101, '\p{^Is_other_letter}', "");
Expect(0, 195101, '\P{Is_other_letter}', "");
Expect(1, 195101, '\P{^Is_other_letter}', "");
Expect(0, 195102, '\p{Is_other_letter}', "");
Expect(1, 195102, '\p{^Is_other_letter}', "");
Expect(1, 195102, '\P{Is_other_letter}', "");
Expect(0, 195102, '\P{^Is_other_letter}', "");
Error('\p{-	Lo:=}');
Error('\P{-	Lo:=}');
Expect(1, 195101, '\p{lo}', "");
Expect(0, 195101, '\p{^lo}', "");
Expect(0, 195101, '\P{lo}', "");
Expect(1, 195101, '\P{^lo}', "");
Expect(0, 195102, '\p{lo}', "");
Expect(1, 195102, '\p{^lo}', "");
Expect(1, 195102, '\P{lo}', "");
Expect(0, 195102, '\P{^lo}', "");
Expect(1, 195101, '\p{_	Lo}', "");
Expect(0, 195101, '\p{^_	Lo}', "");
Expect(0, 195101, '\P{_	Lo}', "");
Expect(1, 195101, '\P{^_	Lo}', "");
Expect(0, 195102, '\p{_	Lo}', "");
Expect(1, 195102, '\p{^_	Lo}', "");
Expect(1, 195102, '\P{_	Lo}', "");
Expect(0, 195102, '\P{^_	Lo}', "");
Error('\p{/a/Is_Lo}');
Error('\P{/a/Is_Lo}');
Expect(1, 195101, '\p{islo}', "");
Expect(0, 195101, '\p{^islo}', "");
Expect(0, 195101, '\P{islo}', "");
Expect(1, 195101, '\P{^islo}', "");
Expect(0, 195102, '\p{islo}', "");
Expect(1, 195102, '\p{^islo}', "");
Expect(1, 195102, '\P{islo}', "");
Expect(0, 195102, '\P{^islo}', "");
Expect(1, 195101, '\p{_ is_Lo}', "");
Expect(0, 195101, '\p{^_ is_Lo}', "");
Expect(0, 195101, '\P{_ is_Lo}', "");
Expect(1, 195101, '\P{^_ is_Lo}', "");
Expect(0, 195102, '\p{_ is_Lo}', "");
Expect(1, 195102, '\p{^_ is_Lo}', "");
Expect(1, 195102, '\P{_ is_Lo}', "");
Expect(0, 195102, '\P{^_ is_Lo}', "");
Error('\p{Other_Default_Ignorable_Code_Point}');
Error('\P{Other_Default_Ignorable_Code_Point}');
Error('\p{Is_Other_Default_Ignorable_Code_Point}');
Error('\P{Is_Other_Default_Ignorable_Code_Point}');
Error('\p{ODI}');
Error('\P{ODI}');
Error('\p{Is_ODI}');
Error('\P{Is_ODI}');
Error('\p{-:=Sundanese}');
Error('\P{-:=Sundanese}');
Expect(1, 7367, '\p{sundanese}', "");
Expect(0, 7367, '\p{^sundanese}', "");
Expect(0, 7367, '\P{sundanese}', "");
Expect(1, 7367, '\P{^sundanese}', "");
Expect(0, 7368, '\p{sundanese}', "");
Expect(1, 7368, '\p{^sundanese}', "");
Expect(1, 7368, '\P{sundanese}', "");
Expect(0, 7368, '\P{^sundanese}', "");
Expect(1, 7367, '\p{_ Sundanese}', "");
Expect(0, 7367, '\p{^_ Sundanese}', "");
Expect(0, 7367, '\P{_ Sundanese}', "");
Expect(1, 7367, '\P{^_ Sundanese}', "");
Expect(0, 7368, '\p{_ Sundanese}', "");
Expect(1, 7368, '\p{^_ Sundanese}', "");
Expect(1, 7368, '\P{_ Sundanese}', "");
Expect(0, 7368, '\P{^_ Sundanese}', "");
Error('\p{/a/is_Sundanese}');
Error('\P{/a/is_Sundanese}');
Expect(1, 7367, '\p{issundanese}', "");
Expect(0, 7367, '\p{^issundanese}', "");
Expect(0, 7367, '\P{issundanese}', "");
Expect(1, 7367, '\P{^issundanese}', "");
Expect(0, 7368, '\p{issundanese}', "");
Expect(1, 7368, '\p{^issundanese}', "");
Expect(1, 7368, '\P{issundanese}', "");
Expect(0, 7368, '\P{^issundanese}', "");
Expect(1, 7367, '\p{_Is_Sundanese}', "");
Expect(0, 7367, '\p{^_Is_Sundanese}', "");
Expect(0, 7367, '\P{_Is_Sundanese}', "");
Expect(1, 7367, '\P{^_Is_Sundanese}', "");
Expect(0, 7368, '\p{_Is_Sundanese}', "");
Expect(1, 7368, '\p{^_Is_Sundanese}', "");
Expect(1, 7368, '\P{_Is_Sundanese}', "");
Expect(0, 7368, '\P{^_Is_Sundanese}', "");
Error('\p{-_Sund:=}');
Error('\P{-_Sund:=}');
Expect(1, 7367, '\p{sund}', "");
Expect(0, 7367, '\p{^sund}', "");
Expect(0, 7367, '\P{sund}', "");
Expect(1, 7367, '\P{^sund}', "");
Expect(0, 7368, '\p{sund}', "");
Expect(1, 7368, '\p{^sund}', "");
Expect(1, 7368, '\P{sund}', "");
Expect(0, 7368, '\P{^sund}', "");
Expect(1, 7367, '\p{ Sund}', "");
Expect(0, 7367, '\p{^ Sund}', "");
Expect(0, 7367, '\P{ Sund}', "");
Expect(1, 7367, '\P{^ Sund}', "");
Expect(0, 7368, '\p{ Sund}', "");
Expect(1, 7368, '\p{^ Sund}', "");
Expect(1, 7368, '\P{ Sund}', "");
Expect(0, 7368, '\P{^ Sund}', "");
Error('\p{ :=IS_sund}');
Error('\P{ :=IS_sund}');
Expect(1, 7367, '\p{issund}', "");
Expect(0, 7367, '\p{^issund}', "");
Expect(0, 7367, '\P{issund}', "");
Expect(1, 7367, '\P{^issund}', "");
Expect(0, 7368, '\p{issund}', "");
Expect(1, 7368, '\p{^issund}', "");
Expect(1, 7368, '\P{issund}', "");
Expect(0, 7368, '\P{^issund}', "");
Expect(1, 7367, '\p{ Is_Sund}', "");
Expect(0, 7367, '\p{^ Is_Sund}', "");
Expect(0, 7367, '\P{ Is_Sund}', "");
Expect(1, 7367, '\P{^ Is_Sund}', "");
Expect(0, 7368, '\p{ Is_Sund}', "");
Expect(1, 7368, '\p{^ Is_Sund}', "");
Expect(1, 7368, '\P{ Is_Sund}', "");
Expect(0, 7368, '\P{^ Is_Sund}', "");
Error('\p{	 Carian/a/}');
Error('\P{	 Carian/a/}');
Expect(1, 66256, '\p{carian}', "");
Expect(0, 66256, '\p{^carian}', "");
Expect(0, 66256, '\P{carian}', "");
Expect(1, 66256, '\P{^carian}', "");
Expect(0, 66257, '\p{carian}', "");
Expect(1, 66257, '\p{^carian}', "");
Expect(1, 66257, '\P{carian}', "");
Expect(0, 66257, '\P{^carian}', "");
Expect(1, 66256, '\p{-Carian}', "");
Expect(0, 66256, '\p{^-Carian}', "");
Expect(0, 66256, '\P{-Carian}', "");
Expect(1, 66256, '\P{^-Carian}', "");
Expect(0, 66257, '\p{-Carian}', "");
Expect(1, 66257, '\p{^-Carian}', "");
Expect(1, 66257, '\P{-Carian}', "");
Expect(0, 66257, '\P{^-Carian}', "");
Error('\p{-/a/IS_Carian}');
Error('\P{-/a/IS_Carian}');
Expect(1, 66256, '\p{iscarian}', "");
Expect(0, 66256, '\p{^iscarian}', "");
Expect(0, 66256, '\P{iscarian}', "");
Expect(1, 66256, '\P{^iscarian}', "");
Expect(0, 66257, '\p{iscarian}', "");
Expect(1, 66257, '\p{^iscarian}', "");
Expect(1, 66257, '\P{iscarian}', "");
Expect(0, 66257, '\P{^iscarian}', "");
Expect(1, 66256, '\p{_	Is_CARIAN}', "");
Expect(0, 66256, '\p{^_	Is_CARIAN}', "");
Expect(0, 66256, '\P{_	Is_CARIAN}', "");
Expect(1, 66256, '\P{^_	Is_CARIAN}', "");
Expect(0, 66257, '\p{_	Is_CARIAN}', "");
Expect(1, 66257, '\p{^_	Is_CARIAN}', "");
Expect(1, 66257, '\P{_	Is_CARIAN}', "");
Expect(0, 66257, '\P{^_	Is_CARIAN}', "");
Error('\p{_cari:=}');
Error('\P{_cari:=}');
Expect(1, 66256, '\p{cari}', "");
Expect(0, 66256, '\p{^cari}', "");
Expect(0, 66256, '\P{cari}', "");
Expect(1, 66256, '\P{^cari}', "");
Expect(0, 66257, '\p{cari}', "");
Expect(1, 66257, '\p{^cari}', "");
Expect(1, 66257, '\P{cari}', "");
Expect(0, 66257, '\P{^cari}', "");
Expect(1, 66256, '\p{_	CARI}', "");
Expect(0, 66256, '\p{^_	CARI}', "");
Expect(0, 66256, '\P{_	CARI}', "");
Expect(1, 66256, '\P{^_	CARI}', "");
Expect(0, 66257, '\p{_	CARI}', "");
Expect(1, 66257, '\p{^_	CARI}', "");
Expect(1, 66257, '\P{_	CARI}', "");
Expect(0, 66257, '\P{^_	CARI}', "");
Error('\p{/a/  IS_CARI}');
Error('\P{/a/  IS_CARI}');
Expect(1, 66256, '\p{iscari}', "");
Expect(0, 66256, '\p{^iscari}', "");
Expect(0, 66256, '\P{iscari}', "");
Expect(1, 66256, '\P{^iscari}', "");
Expect(0, 66257, '\p{iscari}', "");
Expect(1, 66257, '\p{^iscari}', "");
Expect(1, 66257, '\P{iscari}', "");
Expect(0, 66257, '\P{^iscari}', "");
Expect(1, 66256, '\p{		Is_CARI}', "");
Expect(0, 66256, '\p{^		Is_CARI}', "");
Expect(0, 66256, '\P{		Is_CARI}', "");
Expect(1, 66256, '\P{^		Is_CARI}', "");
Expect(0, 66257, '\p{		Is_CARI}', "");
Expect(1, 66257, '\p{^		Is_CARI}', "");
Expect(1, 66257, '\P{		Is_CARI}', "");
Expect(0, 66257, '\P{^		Is_CARI}', "");
Error('\p{ old_ITALIC/a/}');
Error('\P{ old_ITALIC/a/}');
Expect(1, 66339, '\p{olditalic}', "");
Expect(0, 66339, '\p{^olditalic}', "");
Expect(0, 66339, '\P{olditalic}', "");
Expect(1, 66339, '\P{^olditalic}', "");
Expect(0, 66340, '\p{olditalic}', "");
Expect(1, 66340, '\p{^olditalic}', "");
Expect(1, 66340, '\P{olditalic}', "");
Expect(0, 66340, '\P{^olditalic}', "");
Expect(1, 66339, '\p{	_Old_ITALIC}', "");
Expect(0, 66339, '\p{^	_Old_ITALIC}', "");
Expect(0, 66339, '\P{	_Old_ITALIC}', "");
Expect(1, 66339, '\P{^	_Old_ITALIC}', "");
Expect(0, 66340, '\p{	_Old_ITALIC}', "");
Expect(1, 66340, '\p{^	_Old_ITALIC}', "");
Expect(1, 66340, '\P{	_Old_ITALIC}', "");
Expect(0, 66340, '\P{^	_Old_ITALIC}', "");
Error('\p{:=		Is_Old_italic}');
Error('\P{:=		Is_Old_italic}');
Expect(1, 66339, '\p{isolditalic}', "");
Expect(0, 66339, '\p{^isolditalic}', "");
Expect(0, 66339, '\P{isolditalic}', "");
Expect(1, 66339, '\P{^isolditalic}', "");
Expect(0, 66340, '\p{isolditalic}', "");
Expect(1, 66340, '\p{^isolditalic}', "");
Expect(1, 66340, '\P{isolditalic}', "");
Expect(0, 66340, '\P{^isolditalic}', "");
Expect(1, 66339, '\p{-Is_Old_italic}', "");
Expect(0, 66339, '\p{^-Is_Old_italic}', "");
Expect(0, 66339, '\P{-Is_Old_italic}', "");
Expect(1, 66339, '\P{^-Is_Old_italic}', "");
Expect(0, 66340, '\p{-Is_Old_italic}', "");
Expect(1, 66340, '\p{^-Is_Old_italic}', "");
Expect(1, 66340, '\P{-Is_Old_italic}', "");
Expect(0, 66340, '\P{^-Is_Old_italic}', "");
Error('\p{/a/Ital}');
Error('\P{/a/Ital}');
Expect(1, 66339, '\p{ital}', "");
Expect(0, 66339, '\p{^ital}', "");
Expect(0, 66339, '\P{ital}', "");
Expect(1, 66339, '\P{^ital}', "");
Expect(0, 66340, '\p{ital}', "");
Expect(1, 66340, '\p{^ital}', "");
Expect(1, 66340, '\P{ital}', "");
Expect(0, 66340, '\P{^ital}', "");
Expect(1, 66339, '\p{	ITAL}', "");
Expect(0, 66339, '\p{^	ITAL}', "");
Expect(0, 66339, '\P{	ITAL}', "");
Expect(1, 66339, '\P{^	ITAL}', "");
Expect(0, 66340, '\p{	ITAL}', "");
Expect(1, 66340, '\p{^	ITAL}', "");
Expect(1, 66340, '\P{	ITAL}', "");
Expect(0, 66340, '\P{^	ITAL}', "");
Error('\p{:=		IS_ital}');
Error('\P{:=		IS_ital}');
Expect(1, 66339, '\p{isital}', "");
Expect(0, 66339, '\p{^isital}', "");
Expect(0, 66339, '\P{isital}', "");
Expect(1, 66339, '\P{^isital}', "");
Expect(0, 66340, '\p{isital}', "");
Expect(1, 66340, '\p{^isital}', "");
Expect(1, 66340, '\P{isital}', "");
Expect(0, 66340, '\P{^isital}', "");
Expect(1, 66339, '\p{__Is_Ital}', "");
Expect(0, 66339, '\p{^__Is_Ital}', "");
Expect(0, 66339, '\P{__Is_Ital}', "");
Expect(1, 66339, '\P{^__Is_Ital}', "");
Expect(0, 66340, '\p{__Is_Ital}', "");
Expect(1, 66340, '\p{^__Is_Ital}', "");
Expect(1, 66340, '\P{__Is_Ital}', "");
Expect(0, 66340, '\P{^__Is_Ital}', "");
Error('\p{/a/	_XDIGIT}');
Error('\P{/a/	_XDIGIT}');
Expect(1, 65350, '\p{xdigit}', "");
Expect(0, 65350, '\p{^xdigit}', "");
Expect(0, 65350, '\P{xdigit}', "");
Expect(1, 65350, '\P{^xdigit}', "");
Expect(0, 65351, '\p{xdigit}', "");
Expect(1, 65351, '\p{^xdigit}', "");
Expect(1, 65351, '\P{xdigit}', "");
Expect(0, 65351, '\P{^xdigit}', "");
Expect(1, 65350, '\p{ 	xdigit}', "");
Expect(0, 65350, '\p{^ 	xdigit}', "");
Expect(0, 65350, '\P{ 	xdigit}', "");
Expect(1, 65350, '\P{^ 	xdigit}', "");
Expect(0, 65351, '\p{ 	xdigit}', "");
Expect(1, 65351, '\p{^ 	xdigit}', "");
Expect(1, 65351, '\P{ 	xdigit}', "");
Expect(0, 65351, '\P{^ 	xdigit}', "");
Error('\p{  XPosixXDigit:=}');
Error('\P{  XPosixXDigit:=}');
Expect(1, 65350, '\p{xposixxdigit}', "");
Expect(0, 65350, '\p{^xposixxdigit}', "");
Expect(0, 65350, '\P{xposixxdigit}', "");
Expect(1, 65350, '\P{^xposixxdigit}', "");
Expect(0, 65351, '\p{xposixxdigit}', "");
Expect(1, 65351, '\p{^xposixxdigit}', "");
Expect(1, 65351, '\P{xposixxdigit}', "");
Expect(0, 65351, '\P{^xposixxdigit}', "");
Expect(1, 65350, '\p{- XPosixXDigit}', "");
Expect(0, 65350, '\p{^- XPosixXDigit}', "");
Expect(0, 65350, '\P{- XPosixXDigit}', "");
Expect(1, 65350, '\P{^- XPosixXDigit}', "");
Expect(0, 65351, '\p{- XPosixXDigit}', "");
Expect(1, 65351, '\p{^- XPosixXDigit}', "");
Expect(1, 65351, '\P{- XPosixXDigit}', "");
Expect(0, 65351, '\P{^- XPosixXDigit}', "");
Error('\p{  Is_xdigit:=}');
Error('\P{  Is_xdigit:=}');
Expect(1, 65350, '\p{isxdigit}', "");
Expect(0, 65350, '\p{^isxdigit}', "");
Expect(0, 65350, '\P{isxdigit}', "");
Expect(1, 65350, '\P{^isxdigit}', "");
Expect(0, 65351, '\p{isxdigit}', "");
Expect(1, 65351, '\p{^isxdigit}', "");
Expect(1, 65351, '\P{isxdigit}', "");
Expect(0, 65351, '\P{^isxdigit}', "");
Expect(1, 65350, '\p{ Is_XDigit}', "");
Expect(0, 65350, '\p{^ Is_XDigit}', "");
Expect(0, 65350, '\P{ Is_XDigit}', "");
Expect(1, 65350, '\P{^ Is_XDigit}', "");
Expect(0, 65351, '\p{ Is_XDigit}', "");
Expect(1, 65351, '\p{^ Is_XDigit}', "");
Expect(1, 65351, '\P{ Is_XDigit}', "");
Expect(0, 65351, '\P{^ Is_XDigit}', "");
Error('\p{/a/ IS_xposixxdigit}');
Error('\P{/a/ IS_xposixxdigit}');
Expect(1, 65350, '\p{isxposixxdigit}', "");
Expect(0, 65350, '\p{^isxposixxdigit}', "");
Expect(0, 65350, '\P{isxposixxdigit}', "");
Expect(1, 65350, '\P{^isxposixxdigit}', "");
Expect(0, 65351, '\p{isxposixxdigit}', "");
Expect(1, 65351, '\p{^isxposixxdigit}', "");
Expect(1, 65351, '\P{isxposixxdigit}', "");
Expect(0, 65351, '\P{^isxposixxdigit}', "");
Expect(1, 65350, '\p{_	Is_XPosixXDigit}', "");
Expect(0, 65350, '\p{^_	Is_XPosixXDigit}', "");
Expect(0, 65350, '\P{_	Is_XPosixXDigit}', "");
Expect(1, 65350, '\P{^_	Is_XPosixXDigit}', "");
Expect(0, 65351, '\p{_	Is_XPosixXDigit}', "");
Expect(1, 65351, '\p{^_	Is_XPosixXDigit}', "");
Expect(1, 65351, '\P{_	Is_XPosixXDigit}', "");
Expect(0, 65351, '\P{^_	Is_XPosixXDigit}', "");
Error('\p{-	HEX_digit/a/}');
Error('\P{-	HEX_digit/a/}');
Expect(1, 65350, '\p{hexdigit}', "");
Expect(0, 65350, '\p{^hexdigit}', "");
Expect(0, 65350, '\P{hexdigit}', "");
Expect(1, 65350, '\P{^hexdigit}', "");
Expect(0, 65351, '\p{hexdigit}', "");
Expect(1, 65351, '\p{^hexdigit}', "");
Expect(1, 65351, '\P{hexdigit}', "");
Expect(0, 65351, '\P{^hexdigit}', "");
Expect(1, 65350, '\p{-HEX_DIGIT}', "");
Expect(0, 65350, '\p{^-HEX_DIGIT}', "");
Expect(0, 65350, '\P{-HEX_DIGIT}', "");
Expect(1, 65350, '\P{^-HEX_DIGIT}', "");
Expect(0, 65351, '\p{-HEX_DIGIT}', "");
Expect(1, 65351, '\p{^-HEX_DIGIT}', "");
Expect(1, 65351, '\P{-HEX_DIGIT}', "");
Expect(0, 65351, '\P{^-HEX_DIGIT}', "");
Error('\p{:= IS_Hex_DIGIT}');
Error('\P{:= IS_Hex_DIGIT}');
Expect(1, 65350, '\p{ishexdigit}', "");
Expect(0, 65350, '\p{^ishexdigit}', "");
Expect(0, 65350, '\P{ishexdigit}', "");
Expect(1, 65350, '\P{^ishexdigit}', "");
Expect(0, 65351, '\p{ishexdigit}', "");
Expect(1, 65351, '\p{^ishexdigit}', "");
Expect(1, 65351, '\P{ishexdigit}', "");
Expect(0, 65351, '\P{^ishexdigit}', "");
Expect(1, 65350, '\p{-_is_hex_digit}', "");
Expect(0, 65350, '\p{^-_is_hex_digit}', "");
Expect(0, 65350, '\P{-_is_hex_digit}', "");
Expect(1, 65350, '\P{^-_is_hex_digit}', "");
Expect(0, 65351, '\p{-_is_hex_digit}', "");
Expect(1, 65351, '\p{^-_is_hex_digit}', "");
Expect(1, 65351, '\P{-_is_hex_digit}', "");
Expect(0, 65351, '\P{^-_is_hex_digit}', "");
Error('\p{:= HEX}');
Error('\P{:= HEX}');
Expect(1, 65350, '\p{hex}', "");
Expect(0, 65350, '\p{^hex}', "");
Expect(0, 65350, '\P{hex}', "");
Expect(1, 65350, '\P{^hex}', "");
Expect(0, 65351, '\p{hex}', "");
Expect(1, 65351, '\p{^hex}', "");
Expect(1, 65351, '\P{hex}', "");
Expect(0, 65351, '\P{^hex}', "");
Expect(1, 65350, '\p{_	Hex}', "");
Expect(0, 65350, '\p{^_	Hex}', "");
Expect(0, 65350, '\P{_	Hex}', "");
Expect(1, 65350, '\P{^_	Hex}', "");
Expect(0, 65351, '\p{_	Hex}', "");
Expect(1, 65351, '\p{^_	Hex}', "");
Expect(1, 65351, '\P{_	Hex}', "");
Expect(0, 65351, '\P{^_	Hex}', "");
Error('\p{_:=Is_hex}');
Error('\P{_:=Is_hex}');
Expect(1, 65350, '\p{ishex}', "");
Expect(0, 65350, '\p{^ishex}', "");
Expect(0, 65350, '\P{ishex}', "");
Expect(1, 65350, '\P{^ishex}', "");
Expect(0, 65351, '\p{ishex}', "");
Expect(1, 65351, '\p{^ishex}', "");
Expect(1, 65351, '\P{ishex}', "");
Expect(0, 65351, '\P{^ishex}', "");
Expect(1, 65350, '\p{ -Is_HEX}', "");
Expect(0, 65350, '\p{^ -Is_HEX}', "");
Expect(0, 65350, '\P{ -Is_HEX}', "");
Expect(1, 65350, '\P{^ -Is_HEX}', "");
Expect(0, 65351, '\p{ -Is_HEX}', "");
Expect(1, 65351, '\p{^ -Is_HEX}', "");
Expect(1, 65351, '\P{ -Is_HEX}', "");
Expect(0, 65351, '\P{^ -Is_HEX}', "");
Error('\p{ _Titlecase_LETTER/a/}');
Error('\P{ _Titlecase_LETTER/a/}');
Expect(1, 8188, '\p{titlecaseletter}', "");
Expect(0, 8188, '\p{^titlecaseletter}', "");
Expect(0, 8188, '\P{titlecaseletter}', "");
Expect(1, 8188, '\P{^titlecaseletter}', "");
Expect(0, 8189, '\p{titlecaseletter}', "");
Expect(1, 8189, '\p{^titlecaseletter}', "");
Expect(1, 8189, '\P{titlecaseletter}', "");
Expect(0, 8189, '\P{^titlecaseletter}', "");
Expect(1, 8188, '\p{  TITLECASE_Letter}', "");
Expect(0, 8188, '\p{^  TITLECASE_Letter}', "");
Expect(0, 8188, '\P{  TITLECASE_Letter}', "");
Expect(1, 8188, '\P{^  TITLECASE_Letter}', "");
Expect(0, 8189, '\p{  TITLECASE_Letter}', "");
Expect(1, 8189, '\p{^  TITLECASE_Letter}', "");
Expect(1, 8189, '\P{  TITLECASE_Letter}', "");
Expect(0, 8189, '\P{^  TITLECASE_Letter}', "");
Error('\p{	/a/is_Titlecase_Letter}');
Error('\P{	/a/is_Titlecase_Letter}');
Expect(1, 8188, '\p{istitlecaseletter}', "");
Expect(0, 8188, '\p{^istitlecaseletter}', "");
Expect(0, 8188, '\P{istitlecaseletter}', "");
Expect(1, 8188, '\P{^istitlecaseletter}', "");
Expect(0, 8189, '\p{istitlecaseletter}', "");
Expect(1, 8189, '\p{^istitlecaseletter}', "");
Expect(1, 8189, '\P{istitlecaseletter}', "");
Expect(0, 8189, '\P{^istitlecaseletter}', "");
Expect(1, 8188, '\p{-Is_titlecase_letter}', "");
Expect(0, 8188, '\p{^-Is_titlecase_letter}', "");
Expect(0, 8188, '\P{-Is_titlecase_letter}', "");
Expect(1, 8188, '\P{^-Is_titlecase_letter}', "");
Expect(0, 8189, '\p{-Is_titlecase_letter}', "");
Expect(1, 8189, '\p{^-Is_titlecase_letter}', "");
Expect(1, 8189, '\P{-Is_titlecase_letter}', "");
Expect(0, 8189, '\P{^-Is_titlecase_letter}', "");
Error('\p{:=_-LT}');
Error('\P{:=_-LT}');
Expect(1, 8188, '\p{lt}', "");
Expect(0, 8188, '\p{^lt}', "");
Expect(0, 8188, '\P{lt}', "");
Expect(1, 8188, '\P{^lt}', "");
Expect(0, 8189, '\p{lt}', "");
Expect(1, 8189, '\p{^lt}', "");
Expect(1, 8189, '\P{lt}', "");
Expect(0, 8189, '\P{^lt}', "");
Expect(1, 8188, '\p{__Lt}', "");
Expect(0, 8188, '\p{^__Lt}', "");
Expect(0, 8188, '\P{__Lt}', "");
Expect(1, 8188, '\P{^__Lt}', "");
Expect(0, 8189, '\p{__Lt}', "");
Expect(1, 8189, '\p{^__Lt}', "");
Expect(1, 8189, '\P{__Lt}', "");
Expect(0, 8189, '\P{^__Lt}', "");
Error('\p{:= is_Lt}');
Error('\P{:= is_Lt}');
Expect(1, 8188, '\p{islt}', "");
Expect(0, 8188, '\p{^islt}', "");
Expect(0, 8188, '\P{islt}', "");
Expect(1, 8188, '\P{^islt}', "");
Expect(0, 8189, '\p{islt}', "");
Expect(1, 8189, '\p{^islt}', "");
Expect(1, 8189, '\P{islt}', "");
Expect(0, 8189, '\P{^islt}', "");
Expect(1, 8188, '\p{__Is_LT}', "");
Expect(0, 8188, '\p{^__Is_LT}', "");
Expect(0, 8188, '\P{__Is_LT}', "");
Expect(1, 8188, '\P{^__Is_LT}', "");
Expect(0, 8189, '\p{__Is_LT}', "");
Expect(1, 8189, '\p{^__Is_LT}', "");
Expect(1, 8189, '\P{__Is_LT}', "");
Expect(0, 8189, '\P{^__Is_LT}', "");
Error('\p{	_spacing_modifier_Letters/a/}');
Error('\P{	_spacing_modifier_Letters/a/}');
Expect(1, 767, '\p{spacingmodifierletters}', "");
Expect(0, 767, '\p{^spacingmodifierletters}', "");
Expect(0, 767, '\P{spacingmodifierletters}', "");
Expect(1, 767, '\P{^spacingmodifierletters}', "");
Expect(0, 768, '\p{spacingmodifierletters}', "");
Expect(1, 768, '\p{^spacingmodifierletters}', "");
Expect(1, 768, '\P{spacingmodifierletters}', "");
Expect(0, 768, '\P{^spacingmodifierletters}', "");
Expect(1, 767, '\p{ _SPACING_modifier_letters}', "");
Expect(0, 767, '\p{^ _SPACING_modifier_letters}', "");
Expect(0, 767, '\P{ _SPACING_modifier_letters}', "");
Expect(1, 767, '\P{^ _SPACING_modifier_letters}', "");
Expect(0, 768, '\p{ _SPACING_modifier_letters}', "");
Expect(1, 768, '\p{^ _SPACING_modifier_letters}', "");
Expect(1, 768, '\P{ _SPACING_modifier_letters}', "");
Expect(0, 768, '\P{^ _SPACING_modifier_letters}', "");
Error('\p{-IS_SPACING_modifier_letters/a/}');
Error('\P{-IS_SPACING_modifier_letters/a/}');
Expect(1, 767, '\p{isspacingmodifierletters}', "");
Expect(0, 767, '\p{^isspacingmodifierletters}', "");
Expect(0, 767, '\P{isspacingmodifierletters}', "");
Expect(1, 767, '\P{^isspacingmodifierletters}', "");
Expect(0, 768, '\p{isspacingmodifierletters}', "");
Expect(1, 768, '\p{^isspacingmodifierletters}', "");
Expect(1, 768, '\P{isspacingmodifierletters}', "");
Expect(0, 768, '\P{^isspacingmodifierletters}', "");
Expect(1, 767, '\p{ -Is_Spacing_Modifier_LETTERS}', "");
Expect(0, 767, '\p{^ -Is_Spacing_Modifier_LETTERS}', "");
Expect(0, 767, '\P{ -Is_Spacing_Modifier_LETTERS}', "");
Expect(1, 767, '\P{^ -Is_Spacing_Modifier_LETTERS}', "");
Expect(0, 768, '\p{ -Is_Spacing_Modifier_LETTERS}', "");
Expect(1, 768, '\p{^ -Is_Spacing_Modifier_LETTERS}', "");
Expect(1, 768, '\P{ -Is_Spacing_Modifier_LETTERS}', "");
Expect(0, 768, '\P{^ -Is_Spacing_Modifier_LETTERS}', "");
Error('\p{/a/ 	IN_Spacing_MODIFIER_LETTERS}');
Error('\P{/a/ 	IN_Spacing_MODIFIER_LETTERS}');
Expect(1, 767, '\p{inspacingmodifierletters}', "");
Expect(0, 767, '\p{^inspacingmodifierletters}', "");
Expect(0, 767, '\P{inspacingmodifierletters}', "");
Expect(1, 767, '\P{^inspacingmodifierletters}', "");
Expect(0, 768, '\p{inspacingmodifierletters}', "");
Expect(1, 768, '\p{^inspacingmodifierletters}', "");
Expect(1, 768, '\P{inspacingmodifierletters}', "");
Expect(0, 768, '\P{^inspacingmodifierletters}', "");
Expect(1, 767, '\p{-IN_SPACING_Modifier_LETTERS}', "");
Expect(0, 767, '\p{^-IN_SPACING_Modifier_LETTERS}', "");
Expect(0, 767, '\P{-IN_SPACING_Modifier_LETTERS}', "");
Expect(1, 767, '\P{^-IN_SPACING_Modifier_LETTERS}', "");
Expect(0, 768, '\p{-IN_SPACING_Modifier_LETTERS}', "");
Expect(1, 768, '\p{^-IN_SPACING_Modifier_LETTERS}', "");
Expect(1, 768, '\P{-IN_SPACING_Modifier_LETTERS}', "");
Expect(0, 768, '\P{^-IN_SPACING_Modifier_LETTERS}', "");
Error('\p{	Modifier_Letters/a/}');
Error('\P{	Modifier_Letters/a/}');
Expect(1, 767, '\p{modifierletters}', "");
Expect(0, 767, '\p{^modifierletters}', "");
Expect(0, 767, '\P{modifierletters}', "");
Expect(1, 767, '\P{^modifierletters}', "");
Expect(0, 768, '\p{modifierletters}', "");
Expect(1, 768, '\p{^modifierletters}', "");
Expect(1, 768, '\P{modifierletters}', "");
Expect(0, 768, '\P{^modifierletters}', "");
Expect(1, 767, '\p{ modifier_LETTERS}', "");
Expect(0, 767, '\p{^ modifier_LETTERS}', "");
Expect(0, 767, '\P{ modifier_LETTERS}', "");
Expect(1, 767, '\P{^ modifier_LETTERS}', "");
Expect(0, 768, '\p{ modifier_LETTERS}', "");
Expect(1, 768, '\p{^ modifier_LETTERS}', "");
Expect(1, 768, '\P{ modifier_LETTERS}', "");
Expect(0, 768, '\P{^ modifier_LETTERS}', "");
Error('\p{:=Is_MODIFIER_Letters}');
Error('\P{:=Is_MODIFIER_Letters}');
Expect(1, 767, '\p{ismodifierletters}', "");
Expect(0, 767, '\p{^ismodifierletters}', "");
Expect(0, 767, '\P{ismodifierletters}', "");
Expect(1, 767, '\P{^ismodifierletters}', "");
Expect(0, 768, '\p{ismodifierletters}', "");
Expect(1, 768, '\p{^ismodifierletters}', "");
Expect(1, 768, '\P{ismodifierletters}', "");
Expect(0, 768, '\P{^ismodifierletters}', "");
Expect(1, 767, '\p{__Is_modifier_Letters}', "");
Expect(0, 767, '\p{^__Is_modifier_Letters}', "");
Expect(0, 767, '\P{__Is_modifier_Letters}', "");
Expect(1, 767, '\P{^__Is_modifier_Letters}', "");
Expect(0, 768, '\p{__Is_modifier_Letters}', "");
Expect(1, 768, '\p{^__Is_modifier_Letters}', "");
Expect(1, 768, '\P{__Is_modifier_Letters}', "");
Expect(0, 768, '\P{^__Is_modifier_Letters}', "");
Error('\p{-	in_modifier_Letters/a/}');
Error('\P{-	in_modifier_Letters/a/}');
Expect(1, 767, '\p{inmodifierletters}', "");
Expect(0, 767, '\p{^inmodifierletters}', "");
Expect(0, 767, '\P{inmodifierletters}', "");
Expect(1, 767, '\P{^inmodifierletters}', "");
Expect(0, 768, '\p{inmodifierletters}', "");
Expect(1, 768, '\p{^inmodifierletters}', "");
Expect(1, 768, '\P{inmodifierletters}', "");
Expect(0, 768, '\P{^inmodifierletters}', "");
Expect(1, 767, '\p{-	in_Modifier_Letters}', "");
Expect(0, 767, '\p{^-	in_Modifier_Letters}', "");
Expect(0, 767, '\P{-	in_Modifier_Letters}', "");
Expect(1, 767, '\P{^-	in_Modifier_Letters}', "");
Expect(0, 768, '\p{-	in_Modifier_Letters}', "");
Expect(1, 768, '\p{^-	in_Modifier_Letters}', "");
Expect(1, 768, '\P{-	in_Modifier_Letters}', "");
Expect(0, 768, '\P{^-	in_Modifier_Letters}', "");
Error('\p{Grapheme_Link}');
Error('\P{Grapheme_Link}');
Error('\p{Is_Grapheme_Link}');
Error('\P{Is_Grapheme_Link}');
Error('\p{Gr_Link}');
Error('\P{Gr_Link}');
Error('\p{Is_Gr_Link}');
Error('\P{Is_Gr_Link}');
Error('\p{Other_Lowercase}');
Error('\P{Other_Lowercase}');
Error('\p{Is_Other_Lowercase}');
Error('\P{Is_Other_Lowercase}');
Error('\p{OLower}');
Error('\P{OLower}');
Error('\p{Is_OLower}');
Error('\P{Is_OLower}');
Error('\p{-/a/Arabic}');
Error('\P{-/a/Arabic}');
Expect(1, 126705, '\p{arabic}', "");
Expect(0, 126705, '\p{^arabic}', "");
Expect(0, 126705, '\P{arabic}', "");
Expect(1, 126705, '\P{^arabic}', "");
Expect(0, 126706, '\p{arabic}', "");
Expect(1, 126706, '\p{^arabic}', "");
Expect(1, 126706, '\P{arabic}', "");
Expect(0, 126706, '\P{^arabic}', "");
Expect(1, 126705, '\p{	 Arabic}', "");
Expect(0, 126705, '\p{^	 Arabic}', "");
Expect(0, 126705, '\P{	 Arabic}', "");
Expect(1, 126705, '\P{^	 Arabic}', "");
Expect(0, 126706, '\p{	 Arabic}', "");
Expect(1, 126706, '\p{^	 Arabic}', "");
Expect(1, 126706, '\P{	 Arabic}', "");
Expect(0, 126706, '\P{^	 Arabic}', "");
Error('\p{/a/-	Is_ARABIC}');
Error('\P{/a/-	Is_ARABIC}');
Expect(1, 126705, '\p{isarabic}', "");
Expect(0, 126705, '\p{^isarabic}', "");
Expect(0, 126705, '\P{isarabic}', "");
Expect(1, 126705, '\P{^isarabic}', "");
Expect(0, 126706, '\p{isarabic}', "");
Expect(1, 126706, '\p{^isarabic}', "");
Expect(1, 126706, '\P{isarabic}', "");
Expect(0, 126706, '\P{^isarabic}', "");
Expect(1, 126705, '\p{	 IS_ARABIC}', "");
Expect(0, 126705, '\p{^	 IS_ARABIC}', "");
Expect(0, 126705, '\P{	 IS_ARABIC}', "");
Expect(1, 126705, '\P{^	 IS_ARABIC}', "");
Expect(0, 126706, '\p{	 IS_ARABIC}', "");
Expect(1, 126706, '\p{^	 IS_ARABIC}', "");
Expect(1, 126706, '\P{	 IS_ARABIC}', "");
Expect(0, 126706, '\P{^	 IS_ARABIC}', "");
Error('\p{_/a/Arab}');
Error('\P{_/a/Arab}');
Expect(1, 126705, '\p{arab}', "");
Expect(0, 126705, '\p{^arab}', "");
Expect(0, 126705, '\P{arab}', "");
Expect(1, 126705, '\P{^arab}', "");
Expect(0, 126706, '\p{arab}', "");
Expect(1, 126706, '\p{^arab}', "");
Expect(1, 126706, '\P{arab}', "");
Expect(0, 126706, '\P{^arab}', "");
Expect(1, 126705, '\p{--arab}', "");
Expect(0, 126705, '\p{^--arab}', "");
Expect(0, 126705, '\P{--arab}', "");
Expect(1, 126705, '\P{^--arab}', "");
Expect(0, 126706, '\p{--arab}', "");
Expect(1, 126706, '\p{^--arab}', "");
Expect(1, 126706, '\P{--arab}', "");
Expect(0, 126706, '\P{^--arab}', "");
Error('\p{_/a/Is_Arab}');
Error('\P{_/a/Is_Arab}');
Expect(1, 126705, '\p{isarab}', "");
Expect(0, 126705, '\p{^isarab}', "");
Expect(0, 126705, '\P{isarab}', "");
Expect(1, 126705, '\P{^isarab}', "");
Expect(0, 126706, '\p{isarab}', "");
Expect(1, 126706, '\p{^isarab}', "");
Expect(1, 126706, '\P{isarab}', "");
Expect(0, 126706, '\P{^isarab}', "");
Expect(1, 126705, '\p{-	is_ARAB}', "");
Expect(0, 126705, '\p{^-	is_ARAB}', "");
Expect(0, 126705, '\P{-	is_ARAB}', "");
Expect(1, 126705, '\P{^-	is_ARAB}', "");
Expect(0, 126706, '\p{-	is_ARAB}', "");
Expect(1, 126706, '\p{^-	is_ARAB}', "");
Expect(1, 126706, '\P{-	is_ARAB}', "");
Expect(0, 126706, '\P{^-	is_ARAB}', "");
Error('\p{_ MEETEI_MAYEK_Extensions:=}');
Error('\P{_ MEETEI_MAYEK_Extensions:=}');
Expect(1, 43775, '\p{meeteimayekextensions}', "");
Expect(0, 43775, '\p{^meeteimayekextensions}', "");
Expect(0, 43775, '\P{meeteimayekextensions}', "");
Expect(1, 43775, '\P{^meeteimayekextensions}', "");
Expect(0, 43776, '\p{meeteimayekextensions}', "");
Expect(1, 43776, '\p{^meeteimayekextensions}', "");
Expect(1, 43776, '\P{meeteimayekextensions}', "");
Expect(0, 43776, '\P{^meeteimayekextensions}', "");
Expect(1, 43775, '\p{	 MEETEI_MAYEK_Extensions}', "");
Expect(0, 43775, '\p{^	 MEETEI_MAYEK_Extensions}', "");
Expect(0, 43775, '\P{	 MEETEI_MAYEK_Extensions}', "");
Expect(1, 43775, '\P{^	 MEETEI_MAYEK_Extensions}', "");
Expect(0, 43776, '\p{	 MEETEI_MAYEK_Extensions}', "");
Expect(1, 43776, '\p{^	 MEETEI_MAYEK_Extensions}', "");
Expect(1, 43776, '\P{	 MEETEI_MAYEK_Extensions}', "");
Expect(0, 43776, '\P{^	 MEETEI_MAYEK_Extensions}', "");
Error('\p{/a/ _is_Meetei_Mayek_Extensions}');
Error('\P{/a/ _is_Meetei_Mayek_Extensions}');
Expect(1, 43775, '\p{ismeeteimayekextensions}', "");
Expect(0, 43775, '\p{^ismeeteimayekextensions}', "");
Expect(0, 43775, '\P{ismeeteimayekextensions}', "");
Expect(1, 43775, '\P{^ismeeteimayekextensions}', "");
Expect(0, 43776, '\p{ismeeteimayekextensions}', "");
Expect(1, 43776, '\p{^ismeeteimayekextensions}', "");
Expect(1, 43776, '\P{ismeeteimayekextensions}', "");
Expect(0, 43776, '\P{^ismeeteimayekextensions}', "");
Expect(1, 43775, '\p{ IS_meetei_mayek_EXTENSIONS}', "");
Expect(0, 43775, '\p{^ IS_meetei_mayek_EXTENSIONS}', "");
Expect(0, 43775, '\P{ IS_meetei_mayek_EXTENSIONS}', "");
Expect(1, 43775, '\P{^ IS_meetei_mayek_EXTENSIONS}', "");
Expect(0, 43776, '\p{ IS_meetei_mayek_EXTENSIONS}', "");
Expect(1, 43776, '\p{^ IS_meetei_mayek_EXTENSIONS}', "");
Expect(1, 43776, '\P{ IS_meetei_mayek_EXTENSIONS}', "");
Expect(0, 43776, '\P{^ IS_meetei_mayek_EXTENSIONS}', "");
Error('\p{/a/	 In_MEETEI_Mayek_EXTENSIONS}');
Error('\P{/a/	 In_MEETEI_Mayek_EXTENSIONS}');
Expect(1, 43775, '\p{inmeeteimayekextensions}', "");
Expect(0, 43775, '\p{^inmeeteimayekextensions}', "");
Expect(0, 43775, '\P{inmeeteimayekextensions}', "");
Expect(1, 43775, '\P{^inmeeteimayekextensions}', "");
Expect(0, 43776, '\p{inmeeteimayekextensions}', "");
Expect(1, 43776, '\p{^inmeeteimayekextensions}', "");
Expect(1, 43776, '\P{inmeeteimayekextensions}', "");
Expect(0, 43776, '\P{^inmeeteimayekextensions}', "");
Expect(1, 43775, '\p{- in_Meetei_Mayek_extensions}', "");
Expect(0, 43775, '\p{^- in_Meetei_Mayek_extensions}', "");
Expect(0, 43775, '\P{- in_Meetei_Mayek_extensions}', "");
Expect(1, 43775, '\P{^- in_Meetei_Mayek_extensions}', "");
Expect(0, 43776, '\p{- in_Meetei_Mayek_extensions}', "");
Expect(1, 43776, '\p{^- in_Meetei_Mayek_extensions}', "");
Expect(1, 43776, '\P{- in_Meetei_Mayek_extensions}', "");
Expect(0, 43776, '\P{^- in_Meetei_Mayek_extensions}', "");
Error('\p{ :=Meetei_mayek_Ext}');
Error('\P{ :=Meetei_mayek_Ext}');
Expect(1, 43775, '\p{meeteimayekext}', "");
Expect(0, 43775, '\p{^meeteimayekext}', "");
Expect(0, 43775, '\P{meeteimayekext}', "");
Expect(1, 43775, '\P{^meeteimayekext}', "");
Expect(0, 43776, '\p{meeteimayekext}', "");
Expect(1, 43776, '\p{^meeteimayekext}', "");
Expect(1, 43776, '\P{meeteimayekext}', "");
Expect(0, 43776, '\P{^meeteimayekext}', "");
Expect(1, 43775, '\p{--MEETEI_mayek_EXT}', "");
Expect(0, 43775, '\p{^--MEETEI_mayek_EXT}', "");
Expect(0, 43775, '\P{--MEETEI_mayek_EXT}', "");
Expect(1, 43775, '\P{^--MEETEI_mayek_EXT}', "");
Expect(0, 43776, '\p{--MEETEI_mayek_EXT}', "");
Expect(1, 43776, '\p{^--MEETEI_mayek_EXT}', "");
Expect(1, 43776, '\P{--MEETEI_mayek_EXT}', "");
Expect(0, 43776, '\P{^--MEETEI_mayek_EXT}', "");
Error('\p{-_Is_Meetei_MAYEK_EXT/a/}');
Error('\P{-_Is_Meetei_MAYEK_EXT/a/}');
Expect(1, 43775, '\p{ismeeteimayekext}', "");
Expect(0, 43775, '\p{^ismeeteimayekext}', "");
Expect(0, 43775, '\P{ismeeteimayekext}', "");
Expect(1, 43775, '\P{^ismeeteimayekext}', "");
Expect(0, 43776, '\p{ismeeteimayekext}', "");
Expect(1, 43776, '\p{^ismeeteimayekext}', "");
Expect(1, 43776, '\P{ismeeteimayekext}', "");
Expect(0, 43776, '\P{^ismeeteimayekext}', "");
Expect(1, 43775, '\p{-is_Meetei_Mayek_Ext}', "");
Expect(0, 43775, '\p{^-is_Meetei_Mayek_Ext}', "");
Expect(0, 43775, '\P{-is_Meetei_Mayek_Ext}', "");
Expect(1, 43775, '\P{^-is_Meetei_Mayek_Ext}', "");
Expect(0, 43776, '\p{-is_Meetei_Mayek_Ext}', "");
Expect(1, 43776, '\p{^-is_Meetei_Mayek_Ext}', "");
Expect(1, 43776, '\P{-is_Meetei_Mayek_Ext}', "");
Expect(0, 43776, '\P{^-is_Meetei_Mayek_Ext}', "");
Error('\p{/a/_ in_Meetei_Mayek_Ext}');
Error('\P{/a/_ in_Meetei_Mayek_Ext}');
Expect(1, 43775, '\p{inmeeteimayekext}', "");
Expect(0, 43775, '\p{^inmeeteimayekext}', "");
Expect(0, 43775, '\P{inmeeteimayekext}', "");
Expect(1, 43775, '\P{^inmeeteimayekext}', "");
Expect(0, 43776, '\p{inmeeteimayekext}', "");
Expect(1, 43776, '\p{^inmeeteimayekext}', "");
Expect(1, 43776, '\P{inmeeteimayekext}', "");
Expect(0, 43776, '\P{^inmeeteimayekext}', "");
Expect(1, 43775, '\p{ -IN_Meetei_mayek_EXT}', "");
Expect(0, 43775, '\p{^ -IN_Meetei_mayek_EXT}', "");
Expect(0, 43775, '\P{ -IN_Meetei_mayek_EXT}', "");
Expect(1, 43775, '\P{^ -IN_Meetei_mayek_EXT}', "");
Expect(0, 43776, '\p{ -IN_Meetei_mayek_EXT}', "");
Expect(1, 43776, '\p{^ -IN_Meetei_mayek_EXT}', "");
Expect(1, 43776, '\P{ -IN_Meetei_mayek_EXT}', "");
Expect(0, 43776, '\P{^ -IN_Meetei_mayek_EXT}', "");
Error('\p{--deseret/a/}');
Error('\P{--deseret/a/}');
Expect(1, 66639, '\p{deseret}', "");
Expect(0, 66639, '\p{^deseret}', "");
Expect(0, 66639, '\P{deseret}', "");
Expect(1, 66639, '\P{^deseret}', "");
Expect(0, 66640, '\p{deseret}', "");
Expect(1, 66640, '\p{^deseret}', "");
Expect(1, 66640, '\P{deseret}', "");
Expect(0, 66640, '\P{^deseret}', "");
Expect(1, 66639, '\p{	DESERET}', "");
Expect(0, 66639, '\p{^	DESERET}', "");
Expect(0, 66639, '\P{	DESERET}', "");
Expect(1, 66639, '\P{^	DESERET}', "");
Expect(0, 66640, '\p{	DESERET}', "");
Expect(1, 66640, '\p{^	DESERET}', "");
Expect(1, 66640, '\P{	DESERET}', "");
Expect(0, 66640, '\P{^	DESERET}', "");
Error('\p{ /a/is_deseret}');
Error('\P{ /a/is_deseret}');
Expect(1, 66639, '\p{isdeseret}', "");
Expect(0, 66639, '\p{^isdeseret}', "");
Expect(0, 66639, '\P{isdeseret}', "");
Expect(1, 66639, '\P{^isdeseret}', "");
Expect(0, 66640, '\p{isdeseret}', "");
Expect(1, 66640, '\p{^isdeseret}', "");
Expect(1, 66640, '\P{isdeseret}', "");
Expect(0, 66640, '\P{^isdeseret}', "");
Expect(1, 66639, '\p{ -Is_Deseret}', "");
Expect(0, 66639, '\p{^ -Is_Deseret}', "");
Expect(0, 66639, '\P{ -Is_Deseret}', "");
Expect(1, 66639, '\P{^ -Is_Deseret}', "");
Expect(0, 66640, '\p{ -Is_Deseret}', "");
Expect(1, 66640, '\p{^ -Is_Deseret}', "");
Expect(1, 66640, '\P{ -Is_Deseret}', "");
Expect(0, 66640, '\P{^ -Is_Deseret}', "");
Error('\p{	/a/Dsrt}');
Error('\P{	/a/Dsrt}');
Expect(1, 66639, '\p{dsrt}', "");
Expect(0, 66639, '\p{^dsrt}', "");
Expect(0, 66639, '\P{dsrt}', "");
Expect(1, 66639, '\P{^dsrt}', "");
Expect(0, 66640, '\p{dsrt}', "");
Expect(1, 66640, '\p{^dsrt}', "");
Expect(1, 66640, '\P{dsrt}', "");
Expect(0, 66640, '\P{^dsrt}', "");
Expect(1, 66639, '\p{  dsrt}', "");
Expect(0, 66639, '\p{^  dsrt}', "");
Expect(0, 66639, '\P{  dsrt}', "");
Expect(1, 66639, '\P{^  dsrt}', "");
Expect(0, 66640, '\p{  dsrt}', "");
Expect(1, 66640, '\p{^  dsrt}', "");
Expect(1, 66640, '\P{  dsrt}', "");
Expect(0, 66640, '\P{^  dsrt}', "");
Error('\p{_/a/Is_Dsrt}');
Error('\P{_/a/Is_Dsrt}');
Expect(1, 66639, '\p{isdsrt}', "");
Expect(0, 66639, '\p{^isdsrt}', "");
Expect(0, 66639, '\P{isdsrt}', "");
Expect(1, 66639, '\P{^isdsrt}', "");
Expect(0, 66640, '\p{isdsrt}', "");
Expect(1, 66640, '\p{^isdsrt}', "");
Expect(1, 66640, '\P{isdsrt}', "");
Expect(0, 66640, '\P{^isdsrt}', "");
Expect(1, 66639, '\p{_Is_Dsrt}', "");
Expect(0, 66639, '\p{^_Is_Dsrt}', "");
Expect(0, 66639, '\P{_Is_Dsrt}', "");
Expect(1, 66639, '\P{^_Is_Dsrt}', "");
Expect(0, 66640, '\p{_Is_Dsrt}', "");
Expect(1, 66640, '\p{^_Is_Dsrt}', "");
Expect(1, 66640, '\P{_Is_Dsrt}', "");
Expect(0, 66640, '\P{^_Is_Dsrt}', "");
Error('\p{:=-	Cased_letter}');
Error('\P{:=-	Cased_letter}');
Expect(1, 120779, '\p{casedletter}', "");
Expect(0, 120779, '\p{^casedletter}', "");
Expect(0, 120779, '\P{casedletter}', "");
Expect(1, 120779, '\P{^casedletter}', "");
Expect(0, 120780, '\p{casedletter}', "");
Expect(1, 120780, '\p{^casedletter}', "");
Expect(1, 120780, '\P{casedletter}', "");
Expect(0, 120780, '\P{^casedletter}', "");
Expect(1, 120779, '\p{	_cased_LETTER}', "");
Expect(0, 120779, '\p{^	_cased_LETTER}', "");
Expect(0, 120779, '\P{	_cased_LETTER}', "");
Expect(1, 120779, '\P{^	_cased_LETTER}', "");
Expect(0, 120780, '\p{	_cased_LETTER}', "");
Expect(1, 120780, '\p{^	_cased_LETTER}', "");
Expect(1, 120780, '\P{	_cased_LETTER}', "");
Expect(0, 120780, '\P{^	_cased_LETTER}', "");
Error('\p{_:=Is_Cased_letter}');
Error('\P{_:=Is_Cased_letter}');
Expect(1, 120779, '\p{iscasedletter}', "");
Expect(0, 120779, '\p{^iscasedletter}', "");
Expect(0, 120779, '\P{iscasedletter}', "");
Expect(1, 120779, '\P{^iscasedletter}', "");
Expect(0, 120780, '\p{iscasedletter}', "");
Expect(1, 120780, '\p{^iscasedletter}', "");
Expect(1, 120780, '\P{iscasedletter}', "");
Expect(0, 120780, '\P{^iscasedletter}', "");
Expect(1, 120779, '\p{_	Is_CASED_letter}', "");
Expect(0, 120779, '\p{^_	Is_CASED_letter}', "");
Expect(0, 120779, '\P{_	Is_CASED_letter}', "");
Expect(1, 120779, '\P{^_	Is_CASED_letter}', "");
Expect(0, 120780, '\p{_	Is_CASED_letter}', "");
Expect(1, 120780, '\p{^_	Is_CASED_letter}', "");
Expect(1, 120780, '\P{_	Is_CASED_letter}', "");
Expect(0, 120780, '\P{^_	Is_CASED_letter}', "");
Error('\p{/a/-	LC}');
Error('\P{/a/-	LC}');
Expect(1, 120779, '\p{lc}', "");
Expect(0, 120779, '\p{^lc}', "");
Expect(0, 120779, '\P{lc}', "");
Expect(1, 120779, '\P{^lc}', "");
Expect(0, 120780, '\p{lc}', "");
Expect(1, 120780, '\p{^lc}', "");
Expect(1, 120780, '\P{lc}', "");
Expect(0, 120780, '\P{^lc}', "");
Expect(1, 120779, '\p{	 LC}', "");
Expect(0, 120779, '\p{^	 LC}', "");
Expect(0, 120779, '\P{	 LC}', "");
Expect(1, 120779, '\P{^	 LC}', "");
Expect(0, 120780, '\p{	 LC}', "");
Expect(1, 120780, '\p{^	 LC}', "");
Expect(1, 120780, '\P{	 LC}', "");
Expect(0, 120780, '\P{^	 LC}', "");
Error('\p{_:=is_lc}');
Error('\P{_:=is_lc}');
Expect(1, 120779, '\p{islc}', "");
Expect(0, 120779, '\p{^islc}', "");
Expect(0, 120779, '\P{islc}', "");
Expect(1, 120779, '\P{^islc}', "");
Expect(0, 120780, '\p{islc}', "");
Expect(1, 120780, '\p{^islc}', "");
Expect(1, 120780, '\P{islc}', "");
Expect(0, 120780, '\P{^islc}', "");
Expect(1, 120779, '\p{ 	Is_lc}', "");
Expect(0, 120779, '\p{^ 	Is_lc}', "");
Expect(0, 120779, '\P{ 	Is_lc}', "");
Expect(1, 120779, '\P{^ 	Is_lc}', "");
Expect(0, 120780, '\p{ 	Is_lc}', "");
Expect(1, 120780, '\p{^ 	Is_lc}', "");
Expect(1, 120780, '\P{ 	Is_lc}', "");
Expect(0, 120780, '\P{^ 	Is_lc}', "");
Error('\p{-L_/a/}');
Error('\P{-L_/a/}');
Expect(1, 120779, '\p{l_}', "");
Expect(0, 120779, '\p{^l_}', "");
Expect(0, 120779, '\P{l_}', "");
Expect(1, 120779, '\P{^l_}', "");
Expect(0, 120780, '\p{l_}', "");
Expect(1, 120780, '\p{^l_}', "");
Expect(1, 120780, '\P{l_}', "");
Expect(0, 120780, '\P{^l_}', "");
Expect(1, 120779, '\p{_L_}', "");
Expect(0, 120779, '\p{^_L_}', "");
Expect(0, 120779, '\P{_L_}', "");
Expect(1, 120779, '\P{^_L_}', "");
Expect(0, 120780, '\p{_L_}', "");
Expect(1, 120780, '\p{^_L_}', "");
Expect(1, 120780, '\P{_L_}', "");
Expect(0, 120780, '\P{^_L_}', "");
Error('\p{- Is_L_:=}');
Error('\P{- Is_L_:=}');
Expect(1, 120779, '\p{isl_}', "");
Expect(0, 120779, '\p{^isl_}', "");
Expect(0, 120779, '\P{isl_}', "");
Expect(1, 120779, '\P{^isl_}', "");
Expect(0, 120780, '\p{isl_}', "");
Expect(1, 120780, '\p{^isl_}', "");
Expect(1, 120780, '\P{isl_}', "");
Expect(0, 120780, '\P{^isl_}', "");
Expect(1, 120779, '\p{-_is_l_}', "");
Expect(0, 120779, '\p{^-_is_l_}', "");
Expect(0, 120779, '\P{-_is_l_}', "");
Expect(1, 120779, '\P{^-_is_l_}', "");
Expect(0, 120780, '\p{-_is_l_}', "");
Expect(1, 120780, '\p{^-_is_l_}', "");
Expect(1, 120780, '\P{-_is_l_}', "");
Expect(0, 120780, '\P{^-_is_l_}', "");
Error('\p{:= L&}');
Error('\P{:= L&}');
Expect(1, 120779, '\p{l&}', "");
Expect(0, 120779, '\p{^l&}', "");
Expect(0, 120779, '\P{l&}', "");
Expect(1, 120779, '\P{^l&}', "");
Expect(0, 120780, '\p{l&}', "");
Expect(1, 120780, '\p{^l&}', "");
Expect(1, 120780, '\P{l&}', "");
Expect(0, 120780, '\P{^l&}', "");
Expect(1, 120779, '\p{ L&}', "");
Expect(0, 120779, '\p{^ L&}', "");
Expect(0, 120779, '\P{ L&}', "");
Expect(1, 120779, '\P{^ L&}', "");
Expect(0, 120780, '\p{ L&}', "");
Expect(1, 120780, '\p{^ L&}', "");
Expect(1, 120780, '\P{ L&}', "");
Expect(0, 120780, '\P{^ L&}', "");
Error('\p{:= IS_L&}');
Error('\P{:= IS_L&}');
Expect(1, 120779, '\p{isl&}', "");
Expect(0, 120779, '\p{^isl&}', "");
Expect(0, 120779, '\P{isl&}', "");
Expect(1, 120779, '\P{^isl&}', "");
Expect(0, 120780, '\p{isl&}', "");
Expect(1, 120780, '\p{^isl&}', "");
Expect(1, 120780, '\P{isl&}', "");
Expect(0, 120780, '\P{^isl&}', "");
Expect(1, 120779, '\p{	is_L&}', "");
Expect(0, 120779, '\p{^	is_L&}', "");
Expect(0, 120779, '\P{	is_L&}', "");
Expect(1, 120779, '\P{^	is_L&}', "");
Expect(0, 120780, '\p{	is_L&}', "");
Expect(1, 120780, '\p{^	is_L&}', "");
Expect(1, 120780, '\P{	is_L&}', "");
Expect(0, 120780, '\P{^	is_L&}', "");
Error('\p{_ Symbol:=}');
Error('\P{_ Symbol:=}');
Expect(1, 128883, '\p{symbol}', "");
Expect(0, 128883, '\p{^symbol}', "");
Expect(0, 128883, '\P{symbol}', "");
Expect(1, 128883, '\P{^symbol}', "");
Expect(0, 128884, '\p{symbol}', "");
Expect(1, 128884, '\p{^symbol}', "");
Expect(1, 128884, '\P{symbol}', "");
Expect(0, 128884, '\P{^symbol}', "");
Expect(1, 128883, '\p{ _Symbol}', "");
Expect(0, 128883, '\p{^ _Symbol}', "");
Expect(0, 128883, '\P{ _Symbol}', "");
Expect(1, 128883, '\P{^ _Symbol}', "");
Expect(0, 128884, '\p{ _Symbol}', "");
Expect(1, 128884, '\p{^ _Symbol}', "");
Expect(1, 128884, '\P{ _Symbol}', "");
Expect(0, 128884, '\P{^ _Symbol}', "");
Error('\p{:=	 Is_Symbol}');
Error('\P{:=	 Is_Symbol}');
Expect(1, 128883, '\p{issymbol}', "");
Expect(0, 128883, '\p{^issymbol}', "");
Expect(0, 128883, '\P{issymbol}', "");
Expect(1, 128883, '\P{^issymbol}', "");
Expect(0, 128884, '\p{issymbol}', "");
Expect(1, 128884, '\p{^issymbol}', "");
Expect(1, 128884, '\P{issymbol}', "");
Expect(0, 128884, '\P{^issymbol}', "");
Expect(1, 128883, '\p{  is_Symbol}', "");
Expect(0, 128883, '\p{^  is_Symbol}', "");
Expect(0, 128883, '\P{  is_Symbol}', "");
Expect(1, 128883, '\P{^  is_Symbol}', "");
Expect(0, 128884, '\p{  is_Symbol}', "");
Expect(1, 128884, '\p{^  is_Symbol}', "");
Expect(1, 128884, '\P{  is_Symbol}', "");
Expect(0, 128884, '\P{^  is_Symbol}', "");
Error('\p{ :=S}');
Error('\P{ :=S}');
Expect(1, 128883, '\p{s}', "");
Expect(0, 128883, '\p{^s}', "");
Expect(0, 128883, '\P{s}', "");
Expect(1, 128883, '\P{^s}', "");
Expect(0, 128884, '\p{s}', "");
Expect(1, 128884, '\p{^s}', "");
Expect(1, 128884, '\P{s}', "");
Expect(0, 128884, '\P{^s}', "");
Expect(1, 128883, '\p{-	s}', "");
Expect(0, 128883, '\p{^-	s}', "");
Expect(0, 128883, '\P{-	s}', "");
Expect(1, 128883, '\P{^-	s}', "");
Expect(0, 128884, '\p{-	s}', "");
Expect(1, 128884, '\p{^-	s}', "");
Expect(1, 128884, '\P{-	s}', "");
Expect(0, 128884, '\P{^-	s}', "");
Error('\p{/a/ Is_S}');
Error('\P{/a/ Is_S}');
Expect(1, 128883, '\p{iss}', "");
Expect(0, 128883, '\p{^iss}', "");
Expect(0, 128883, '\P{iss}', "");
Expect(1, 128883, '\P{^iss}', "");
Expect(0, 128884, '\p{iss}', "");
Expect(1, 128884, '\p{^iss}', "");
Expect(1, 128884, '\P{iss}', "");
Expect(0, 128884, '\P{^iss}', "");
Expect(1, 128883, '\p{ 	IS_s}', "");
Expect(0, 128883, '\p{^ 	IS_s}', "");
Expect(0, 128883, '\P{ 	IS_s}', "");
Expect(1, 128883, '\P{^ 	IS_s}', "");
Expect(0, 128884, '\p{ 	IS_s}', "");
Expect(1, 128884, '\p{^ 	IS_s}', "");
Expect(1, 128884, '\P{ 	IS_s}', "");
Expect(0, 128884, '\P{^ 	IS_s}', "");
Error('\p{:=PosixSpace}');
Error('\P{:=PosixSpace}');
Expect(1, 32, '\p{posixspace}', "");
Expect(0, 32, '\p{^posixspace}', "");
Expect(0, 32, '\P{posixspace}', "");
Expect(1, 32, '\P{^posixspace}', "");
Expect(0, 33, '\p{posixspace}', "");
Expect(1, 33, '\p{^posixspace}', "");
Expect(1, 33, '\P{posixspace}', "");
Expect(0, 33, '\P{^posixspace}', "");
Expect(1, 32, '\p{		PosixSpace}', "");
Expect(0, 32, '\p{^		PosixSpace}', "");
Expect(0, 32, '\P{		PosixSpace}', "");
Expect(1, 32, '\P{^		PosixSpace}', "");
Expect(0, 33, '\p{		PosixSpace}', "");
Expect(1, 33, '\p{^		PosixSpace}', "");
Expect(1, 33, '\P{		PosixSpace}', "");
Expect(0, 33, '\P{^		PosixSpace}', "");
Error('\p{/a/-Is_PosixSpace}');
Error('\P{/a/-Is_PosixSpace}');
Expect(1, 32, '\p{isposixspace}', "");
Expect(0, 32, '\p{^isposixspace}', "");
Expect(0, 32, '\P{isposixspace}', "");
Expect(1, 32, '\P{^isposixspace}', "");
Expect(0, 33, '\p{isposixspace}', "");
Expect(1, 33, '\p{^isposixspace}', "");
Expect(1, 33, '\P{isposixspace}', "");
Expect(0, 33, '\P{^isposixspace}', "");
Expect(1, 32, '\p{-_IS_PosixSpace}', "");
Expect(0, 32, '\p{^-_IS_PosixSpace}', "");
Expect(0, 32, '\P{-_IS_PosixSpace}', "");
Expect(1, 32, '\P{^-_IS_PosixSpace}', "");
Expect(0, 33, '\p{-_IS_PosixSpace}', "");
Expect(1, 33, '\p{^-_IS_PosixSpace}', "");
Expect(1, 33, '\P{-_IS_PosixSpace}', "");
Expect(0, 33, '\P{^-_IS_PosixSpace}', "");
Error('\p{_/a/CJK_Compatibility_Forms}');
Error('\P{_/a/CJK_Compatibility_Forms}');
Expect(1, 65103, '\p{cjkcompatibilityforms}', "");
Expect(0, 65103, '\p{^cjkcompatibilityforms}', "");
Expect(0, 65103, '\P{cjkcompatibilityforms}', "");
Expect(1, 65103, '\P{^cjkcompatibilityforms}', "");
Expect(0, 65104, '\p{cjkcompatibilityforms}', "");
Expect(1, 65104, '\p{^cjkcompatibilityforms}', "");
Expect(1, 65104, '\P{cjkcompatibilityforms}', "");
Expect(0, 65104, '\P{^cjkcompatibilityforms}', "");
Expect(1, 65103, '\p{ _cjk_Compatibility_Forms}', "");
Expect(0, 65103, '\p{^ _cjk_Compatibility_Forms}', "");
Expect(0, 65103, '\P{ _cjk_Compatibility_Forms}', "");
Expect(1, 65103, '\P{^ _cjk_Compatibility_Forms}', "");
Expect(0, 65104, '\p{ _cjk_Compatibility_Forms}', "");
Expect(1, 65104, '\p{^ _cjk_Compatibility_Forms}', "");
Expect(1, 65104, '\P{ _cjk_Compatibility_Forms}', "");
Expect(0, 65104, '\P{^ _cjk_Compatibility_Forms}', "");
Error('\p{_/a/Is_CJK_Compatibility_forms}');
Error('\P{_/a/Is_CJK_Compatibility_forms}');
Expect(1, 65103, '\p{iscjkcompatibilityforms}', "");
Expect(0, 65103, '\p{^iscjkcompatibilityforms}', "");
Expect(0, 65103, '\P{iscjkcompatibilityforms}', "");
Expect(1, 65103, '\P{^iscjkcompatibilityforms}', "");
Expect(0, 65104, '\p{iscjkcompatibilityforms}', "");
Expect(1, 65104, '\p{^iscjkcompatibilityforms}', "");
Expect(1, 65104, '\P{iscjkcompatibilityforms}', "");
Expect(0, 65104, '\P{^iscjkcompatibilityforms}', "");
Expect(1, 65103, '\p{  IS_CJK_compatibility_FORMS}', "");
Expect(0, 65103, '\p{^  IS_CJK_compatibility_FORMS}', "");
Expect(0, 65103, '\P{  IS_CJK_compatibility_FORMS}', "");
Expect(1, 65103, '\P{^  IS_CJK_compatibility_FORMS}', "");
Expect(0, 65104, '\p{  IS_CJK_compatibility_FORMS}', "");
Expect(1, 65104, '\p{^  IS_CJK_compatibility_FORMS}', "");
Expect(1, 65104, '\P{  IS_CJK_compatibility_FORMS}', "");
Expect(0, 65104, '\P{^  IS_CJK_compatibility_FORMS}', "");
Error('\p{ /a/IN_CJK_COMPATIBILITY_Forms}');
Error('\P{ /a/IN_CJK_COMPATIBILITY_Forms}');
Expect(1, 65103, '\p{incjkcompatibilityforms}', "");
Expect(0, 65103, '\p{^incjkcompatibilityforms}', "");
Expect(0, 65103, '\P{incjkcompatibilityforms}', "");
Expect(1, 65103, '\P{^incjkcompatibilityforms}', "");
Expect(0, 65104, '\p{incjkcompatibilityforms}', "");
Expect(1, 65104, '\p{^incjkcompatibilityforms}', "");
Expect(1, 65104, '\P{incjkcompatibilityforms}', "");
Expect(0, 65104, '\P{^incjkcompatibilityforms}', "");
Expect(1, 65103, '\p{  IN_CJK_COMPATIBILITY_forms}', "");
Expect(0, 65103, '\p{^  IN_CJK_COMPATIBILITY_forms}', "");
Expect(0, 65103, '\P{  IN_CJK_COMPATIBILITY_forms}', "");
Expect(1, 65103, '\P{^  IN_CJK_COMPATIBILITY_forms}', "");
Expect(0, 65104, '\p{  IN_CJK_COMPATIBILITY_forms}', "");
Expect(1, 65104, '\p{^  IN_CJK_COMPATIBILITY_forms}', "");
Expect(1, 65104, '\P{  IN_CJK_COMPATIBILITY_forms}', "");
Expect(0, 65104, '\P{^  IN_CJK_COMPATIBILITY_forms}', "");
Error('\p{/a/-_CJK_Compat_Forms}');
Error('\P{/a/-_CJK_Compat_Forms}');
Expect(1, 65103, '\p{cjkcompatforms}', "");
Expect(0, 65103, '\p{^cjkcompatforms}', "");
Expect(0, 65103, '\P{cjkcompatforms}', "");
Expect(1, 65103, '\P{^cjkcompatforms}', "");
Expect(0, 65104, '\p{cjkcompatforms}', "");
Expect(1, 65104, '\p{^cjkcompatforms}', "");
Expect(1, 65104, '\P{cjkcompatforms}', "");
Expect(0, 65104, '\P{^cjkcompatforms}', "");
Expect(1, 65103, '\p{_cjk_Compat_FORMS}', "");
Expect(0, 65103, '\p{^_cjk_Compat_FORMS}', "");
Expect(0, 65103, '\P{_cjk_Compat_FORMS}', "");
Expect(1, 65103, '\P{^_cjk_Compat_FORMS}', "");
Expect(0, 65104, '\p{_cjk_Compat_FORMS}', "");
Expect(1, 65104, '\p{^_cjk_Compat_FORMS}', "");
Expect(1, 65104, '\P{_cjk_Compat_FORMS}', "");
Expect(0, 65104, '\P{^_cjk_Compat_FORMS}', "");
Error('\p{		Is_CJK_COMPAT_Forms/a/}');
Error('\P{		Is_CJK_COMPAT_Forms/a/}');
Expect(1, 65103, '\p{iscjkcompatforms}', "");
Expect(0, 65103, '\p{^iscjkcompatforms}', "");
Expect(0, 65103, '\P{iscjkcompatforms}', "");
Expect(1, 65103, '\P{^iscjkcompatforms}', "");
Expect(0, 65104, '\p{iscjkcompatforms}', "");
Expect(1, 65104, '\p{^iscjkcompatforms}', "");
Expect(1, 65104, '\P{iscjkcompatforms}', "");
Expect(0, 65104, '\P{^iscjkcompatforms}', "");
Expect(1, 65103, '\p{		is_cjk_compat_Forms}', "");
Expect(0, 65103, '\p{^		is_cjk_compat_Forms}', "");
Expect(0, 65103, '\P{		is_cjk_compat_Forms}', "");
Expect(1, 65103, '\P{^		is_cjk_compat_Forms}', "");
Expect(0, 65104, '\p{		is_cjk_compat_Forms}', "");
Expect(1, 65104, '\p{^		is_cjk_compat_Forms}', "");
Expect(1, 65104, '\P{		is_cjk_compat_Forms}', "");
Expect(0, 65104, '\P{^		is_cjk_compat_Forms}', "");
Error('\p{/a/	in_CJK_Compat_Forms}');
Error('\P{/a/	in_CJK_Compat_Forms}');
Expect(1, 65103, '\p{incjkcompatforms}', "");
Expect(0, 65103, '\p{^incjkcompatforms}', "");
Expect(0, 65103, '\P{incjkcompatforms}', "");
Expect(1, 65103, '\P{^incjkcompatforms}', "");
Expect(0, 65104, '\p{incjkcompatforms}', "");
Expect(1, 65104, '\p{^incjkcompatforms}', "");
Expect(1, 65104, '\P{incjkcompatforms}', "");
Expect(0, 65104, '\P{^incjkcompatforms}', "");
Expect(1, 65103, '\p{_-IN_cjk_compat_Forms}', "");
Expect(0, 65103, '\p{^_-IN_cjk_compat_Forms}', "");
Expect(0, 65103, '\P{_-IN_cjk_compat_Forms}', "");
Expect(1, 65103, '\P{^_-IN_cjk_compat_Forms}', "");
Expect(0, 65104, '\p{_-IN_cjk_compat_Forms}', "");
Expect(1, 65104, '\p{^_-IN_cjk_compat_Forms}', "");
Expect(1, 65104, '\P{_-IN_cjk_compat_Forms}', "");
Expect(0, 65104, '\P{^_-IN_cjk_compat_Forms}', "");
Error('\p{	currency_SYMBOLS/a/}');
Error('\P{	currency_SYMBOLS/a/}');
Expect(1, 8399, '\p{currencysymbols}', "");
Expect(0, 8399, '\p{^currencysymbols}', "");
Expect(0, 8399, '\P{currencysymbols}', "");
Expect(1, 8399, '\P{^currencysymbols}', "");
Expect(0, 8400, '\p{currencysymbols}', "");
Expect(1, 8400, '\p{^currencysymbols}', "");
Expect(1, 8400, '\P{currencysymbols}', "");
Expect(0, 8400, '\P{^currencysymbols}', "");
Expect(1, 8399, '\p{-CURRENCY_SYMBOLS}', "");
Expect(0, 8399, '\p{^-CURRENCY_SYMBOLS}', "");
Expect(0, 8399, '\P{-CURRENCY_SYMBOLS}', "");
Expect(1, 8399, '\P{^-CURRENCY_SYMBOLS}', "");
Expect(0, 8400, '\p{-CURRENCY_SYMBOLS}', "");
Expect(1, 8400, '\p{^-CURRENCY_SYMBOLS}', "");
Expect(1, 8400, '\P{-CURRENCY_SYMBOLS}', "");
Expect(0, 8400, '\P{^-CURRENCY_SYMBOLS}', "");
Error('\p{ :=IS_Currency_Symbols}');
Error('\P{ :=IS_Currency_Symbols}');
Expect(1, 8399, '\p{iscurrencysymbols}', "");
Expect(0, 8399, '\p{^iscurrencysymbols}', "");
Expect(0, 8399, '\P{iscurrencysymbols}', "");
Expect(1, 8399, '\P{^iscurrencysymbols}', "");
Expect(0, 8400, '\p{iscurrencysymbols}', "");
Expect(1, 8400, '\p{^iscurrencysymbols}', "");
Expect(1, 8400, '\P{iscurrencysymbols}', "");
Expect(0, 8400, '\P{^iscurrencysymbols}', "");
Expect(1, 8399, '\p{_IS_CURRENCY_Symbols}', "");
Expect(0, 8399, '\p{^_IS_CURRENCY_Symbols}', "");
Expect(0, 8399, '\P{_IS_CURRENCY_Symbols}', "");
Expect(1, 8399, '\P{^_IS_CURRENCY_Symbols}', "");
Expect(0, 8400, '\p{_IS_CURRENCY_Symbols}', "");
Expect(1, 8400, '\p{^_IS_CURRENCY_Symbols}', "");
Expect(1, 8400, '\P{_IS_CURRENCY_Symbols}', "");
Expect(0, 8400, '\P{^_IS_CURRENCY_Symbols}', "");
Error('\p{ :=In_Currency_Symbols}');
Error('\P{ :=In_Currency_Symbols}');
Expect(1, 8399, '\p{incurrencysymbols}', "");
Expect(0, 8399, '\p{^incurrencysymbols}', "");
Expect(0, 8399, '\P{incurrencysymbols}', "");
Expect(1, 8399, '\P{^incurrencysymbols}', "");
Expect(0, 8400, '\p{incurrencysymbols}', "");
Expect(1, 8400, '\p{^incurrencysymbols}', "");
Expect(1, 8400, '\P{incurrencysymbols}', "");
Expect(0, 8400, '\P{^incurrencysymbols}', "");
Expect(1, 8399, '\p{ in_CURRENCY_symbols}', "");
Expect(0, 8399, '\p{^ in_CURRENCY_symbols}', "");
Expect(0, 8399, '\P{ in_CURRENCY_symbols}', "");
Expect(1, 8399, '\P{^ in_CURRENCY_symbols}', "");
Expect(0, 8400, '\p{ in_CURRENCY_symbols}', "");
Expect(1, 8400, '\p{^ in_CURRENCY_symbols}', "");
Expect(1, 8400, '\P{ in_CURRENCY_symbols}', "");
Expect(0, 8400, '\P{^ in_CURRENCY_symbols}', "");
Error('\p{/a/Currency_Symbol}');
Error('\P{/a/Currency_Symbol}');
Expect(1, 65510, '\p{currencysymbol}', "");
Expect(0, 65510, '\p{^currencysymbol}', "");
Expect(0, 65510, '\P{currencysymbol}', "");
Expect(1, 65510, '\P{^currencysymbol}', "");
Expect(0, 65511, '\p{currencysymbol}', "");
Expect(1, 65511, '\p{^currencysymbol}', "");
Expect(1, 65511, '\P{currencysymbol}', "");
Expect(0, 65511, '\P{^currencysymbol}', "");
Expect(1, 65510, '\p{ -CURRENCY_symbol}', "");
Expect(0, 65510, '\p{^ -CURRENCY_symbol}', "");
Expect(0, 65510, '\P{ -CURRENCY_symbol}', "");
Expect(1, 65510, '\P{^ -CURRENCY_symbol}', "");
Expect(0, 65511, '\p{ -CURRENCY_symbol}', "");
Expect(1, 65511, '\p{^ -CURRENCY_symbol}', "");
Expect(1, 65511, '\P{ -CURRENCY_symbol}', "");
Expect(0, 65511, '\P{^ -CURRENCY_symbol}', "");
Error('\p{ 	is_CURRENCY_SYMBOL:=}');
Error('\P{ 	is_CURRENCY_SYMBOL:=}');
Expect(1, 65510, '\p{iscurrencysymbol}', "");
Expect(0, 65510, '\p{^iscurrencysymbol}', "");
Expect(0, 65510, '\P{iscurrencysymbol}', "");
Expect(1, 65510, '\P{^iscurrencysymbol}', "");
Expect(0, 65511, '\p{iscurrencysymbol}', "");
Expect(1, 65511, '\p{^iscurrencysymbol}', "");
Expect(1, 65511, '\P{iscurrencysymbol}', "");
Expect(0, 65511, '\P{^iscurrencysymbol}', "");
Expect(1, 65510, '\p{-Is_Currency_Symbol}', "");
Expect(0, 65510, '\p{^-Is_Currency_Symbol}', "");
Expect(0, 65510, '\P{-Is_Currency_Symbol}', "");
Expect(1, 65510, '\P{^-Is_Currency_Symbol}', "");
Expect(0, 65511, '\p{-Is_Currency_Symbol}', "");
Expect(1, 65511, '\p{^-Is_Currency_Symbol}', "");
Expect(1, 65511, '\P{-Is_Currency_Symbol}', "");
Expect(0, 65511, '\P{^-Is_Currency_Symbol}', "");
Error('\p{	/a/SC}');
Error('\P{	/a/SC}');
Expect(1, 65510, '\p{sc}', "");
Expect(0, 65510, '\p{^sc}', "");
Expect(0, 65510, '\P{sc}', "");
Expect(1, 65510, '\P{^sc}', "");
Expect(0, 65511, '\p{sc}', "");
Expect(1, 65511, '\p{^sc}', "");
Expect(1, 65511, '\P{sc}', "");
Expect(0, 65511, '\P{^sc}', "");
Expect(1, 65510, '\p{-SC}', "");
Expect(0, 65510, '\p{^-SC}', "");
Expect(0, 65510, '\P{-SC}', "");
Expect(1, 65510, '\P{^-SC}', "");
Expect(0, 65511, '\p{-SC}', "");
Expect(1, 65511, '\p{^-SC}', "");
Expect(1, 65511, '\P{-SC}', "");
Expect(0, 65511, '\P{^-SC}', "");
Error('\p{_/a/is_sc}');
Error('\P{_/a/is_sc}');
Expect(1, 65510, '\p{issc}', "");
Expect(0, 65510, '\p{^issc}', "");
Expect(0, 65510, '\P{issc}', "");
Expect(1, 65510, '\P{^issc}', "");
Expect(0, 65511, '\p{issc}', "");
Expect(1, 65511, '\p{^issc}', "");
Expect(1, 65511, '\P{issc}', "");
Expect(0, 65511, '\P{^issc}', "");
Expect(1, 65510, '\p{--is_Sc}', "");
Expect(0, 65510, '\p{^--is_Sc}', "");
Expect(0, 65510, '\P{--is_Sc}', "");
Expect(1, 65510, '\P{^--is_Sc}', "");
Expect(0, 65511, '\p{--is_Sc}', "");
Expect(1, 65511, '\p{^--is_Sc}', "");
Expect(1, 65511, '\P{--is_Sc}', "");
Expect(0, 65511, '\P{^--is_Sc}', "");
Error('\p{_VARIATION_selectors_Supplement/a/}');
Error('\P{_VARIATION_selectors_Supplement/a/}');
Expect(1, 917999, '\p{variationselectorssupplement}', "");
Expect(0, 917999, '\p{^variationselectorssupplement}', "");
Expect(0, 917999, '\P{variationselectorssupplement}', "");
Expect(1, 917999, '\P{^variationselectorssupplement}', "");
Expect(0, 918000, '\p{variationselectorssupplement}', "");
Expect(1, 918000, '\p{^variationselectorssupplement}', "");
Expect(1, 918000, '\P{variationselectorssupplement}', "");
Expect(0, 918000, '\P{^variationselectorssupplement}', "");
Expect(1, 917999, '\p{_-VARIATION_selectors_supplement}', "");
Expect(0, 917999, '\p{^_-VARIATION_selectors_supplement}', "");
Expect(0, 917999, '\P{_-VARIATION_selectors_supplement}', "");
Expect(1, 917999, '\P{^_-VARIATION_selectors_supplement}', "");
Expect(0, 918000, '\p{_-VARIATION_selectors_supplement}', "");
Expect(1, 918000, '\p{^_-VARIATION_selectors_supplement}', "");
Expect(1, 918000, '\P{_-VARIATION_selectors_supplement}', "");
Expect(0, 918000, '\P{^_-VARIATION_selectors_supplement}', "");
Error('\p{ -is_VARIATION_Selectors_Supplement/a/}');
Error('\P{ -is_VARIATION_Selectors_Supplement/a/}');
Expect(1, 917999, '\p{isvariationselectorssupplement}', "");
Expect(0, 917999, '\p{^isvariationselectorssupplement}', "");
Expect(0, 917999, '\P{isvariationselectorssupplement}', "");
Expect(1, 917999, '\P{^isvariationselectorssupplement}', "");
Expect(0, 918000, '\p{isvariationselectorssupplement}', "");
Expect(1, 918000, '\p{^isvariationselectorssupplement}', "");
Expect(1, 918000, '\P{isvariationselectorssupplement}', "");
Expect(0, 918000, '\P{^isvariationselectorssupplement}', "");
Expect(1, 917999, '\p{-IS_VARIATION_selectors_Supplement}', "");
Expect(0, 917999, '\p{^-IS_VARIATION_selectors_Supplement}', "");
Expect(0, 917999, '\P{-IS_VARIATION_selectors_Supplement}', "");
Expect(1, 917999, '\P{^-IS_VARIATION_selectors_Supplement}', "");
Expect(0, 918000, '\p{-IS_VARIATION_selectors_Supplement}', "");
Expect(1, 918000, '\p{^-IS_VARIATION_selectors_Supplement}', "");
Expect(1, 918000, '\P{-IS_VARIATION_selectors_Supplement}', "");
Expect(0, 918000, '\P{^-IS_VARIATION_selectors_Supplement}', "");
Error('\p{/a/	-In_Variation_selectors_supplement}');
Error('\P{/a/	-In_Variation_selectors_supplement}');
Expect(1, 917999, '\p{invariationselectorssupplement}', "");
Expect(0, 917999, '\p{^invariationselectorssupplement}', "");
Expect(0, 917999, '\P{invariationselectorssupplement}', "");
Expect(1, 917999, '\P{^invariationselectorssupplement}', "");
Expect(0, 918000, '\p{invariationselectorssupplement}', "");
Expect(1, 918000, '\p{^invariationselectorssupplement}', "");
Expect(1, 918000, '\P{invariationselectorssupplement}', "");
Expect(0, 918000, '\P{^invariationselectorssupplement}', "");
Expect(1, 917999, '\p{	 In_Variation_Selectors_supplement}', "");
Expect(0, 917999, '\p{^	 In_Variation_Selectors_supplement}', "");
Expect(0, 917999, '\P{	 In_Variation_Selectors_supplement}', "");
Expect(1, 917999, '\P{^	 In_Variation_Selectors_supplement}', "");
Expect(0, 918000, '\p{	 In_Variation_Selectors_supplement}', "");
Expect(1, 918000, '\p{^	 In_Variation_Selectors_supplement}', "");
Expect(1, 918000, '\P{	 In_Variation_Selectors_supplement}', "");
Expect(0, 918000, '\P{^	 In_Variation_Selectors_supplement}', "");
Error('\p{	 VS_Sup:=}');
Error('\P{	 VS_Sup:=}');
Expect(1, 917999, '\p{vssup}', "");
Expect(0, 917999, '\p{^vssup}', "");
Expect(0, 917999, '\P{vssup}', "");
Expect(1, 917999, '\P{^vssup}', "");
Expect(0, 918000, '\p{vssup}', "");
Expect(1, 918000, '\p{^vssup}', "");
Expect(1, 918000, '\P{vssup}', "");
Expect(0, 918000, '\P{^vssup}', "");
Expect(1, 917999, '\p{	vs_sup}', "");
Expect(0, 917999, '\p{^	vs_sup}', "");
Expect(0, 917999, '\P{	vs_sup}', "");
Expect(1, 917999, '\P{^	vs_sup}', "");
Expect(0, 918000, '\p{	vs_sup}', "");
Expect(1, 918000, '\p{^	vs_sup}', "");
Expect(1, 918000, '\P{	vs_sup}', "");
Expect(0, 918000, '\P{^	vs_sup}', "");
Error('\p{/a/	-is_VS_sup}');
Error('\P{/a/	-is_VS_sup}');
Expect(1, 917999, '\p{isvssup}', "");
Expect(0, 917999, '\p{^isvssup}', "");
Expect(0, 917999, '\P{isvssup}', "");
Expect(1, 917999, '\P{^isvssup}', "");
Expect(0, 918000, '\p{isvssup}', "");
Expect(1, 918000, '\p{^isvssup}', "");
Expect(1, 918000, '\P{isvssup}', "");
Expect(0, 918000, '\P{^isvssup}', "");
Expect(1, 917999, '\p{_ IS_VS_Sup}', "");
Expect(0, 917999, '\p{^_ IS_VS_Sup}', "");
Expect(0, 917999, '\P{_ IS_VS_Sup}', "");
Expect(1, 917999, '\P{^_ IS_VS_Sup}', "");
Expect(0, 918000, '\p{_ IS_VS_Sup}', "");
Expect(1, 918000, '\p{^_ IS_VS_Sup}', "");
Expect(1, 918000, '\P{_ IS_VS_Sup}', "");
Expect(0, 918000, '\P{^_ IS_VS_Sup}', "");
Error('\p{:=__In_VS_SUP}');
Error('\P{:=__In_VS_SUP}');
Expect(1, 917999, '\p{invssup}', "");
Expect(0, 917999, '\p{^invssup}', "");
Expect(0, 917999, '\P{invssup}', "");
Expect(1, 917999, '\P{^invssup}', "");
Expect(0, 918000, '\p{invssup}', "");
Expect(1, 918000, '\p{^invssup}', "");
Expect(1, 918000, '\P{invssup}', "");
Expect(0, 918000, '\P{^invssup}', "");
Expect(1, 917999, '\p{-	in_VS_Sup}', "");
Expect(0, 917999, '\p{^-	in_VS_Sup}', "");
Expect(0, 917999, '\P{-	in_VS_Sup}', "");
Expect(1, 917999, '\P{^-	in_VS_Sup}', "");
Expect(0, 918000, '\p{-	in_VS_Sup}', "");
Expect(1, 918000, '\p{^-	in_VS_Sup}', "");
Expect(1, 918000, '\P{-	in_VS_Sup}', "");
Expect(0, 918000, '\P{^-	in_VS_Sup}', "");
Error('\p{ _HIGH_SURROGATES/a/}');
Error('\P{ _HIGH_SURROGATES/a/}');
Expect(1, 56191, '\p{highsurrogates}', "");
Expect(0, 56191, '\p{^highsurrogates}', "");
Expect(0, 56191, '\P{highsurrogates}', "");
Expect(1, 56191, '\P{^highsurrogates}', "");
Expect(0, 57344, '\p{highsurrogates}', "");
Expect(1, 57344, '\p{^highsurrogates}', "");
Expect(1, 57344, '\P{highsurrogates}', "");
Expect(0, 57344, '\P{^highsurrogates}', "");
Expect(1, 56191, '\p{ High_Surrogates}', "");
Expect(0, 56191, '\p{^ High_Surrogates}', "");
Expect(0, 56191, '\P{ High_Surrogates}', "");
Expect(1, 56191, '\P{^ High_Surrogates}', "");
Expect(0, 57344, '\p{ High_Surrogates}', "");
Expect(1, 57344, '\p{^ High_Surrogates}', "");
Expect(1, 57344, '\P{ High_Surrogates}', "");
Expect(0, 57344, '\P{^ High_Surrogates}', "");
Error('\p{:=_-Is_High_SURROGATES}');
Error('\P{:=_-Is_High_SURROGATES}');
Expect(1, 56191, '\p{ishighsurrogates}', "");
Expect(0, 56191, '\p{^ishighsurrogates}', "");
Expect(0, 56191, '\P{ishighsurrogates}', "");
Expect(1, 56191, '\P{^ishighsurrogates}', "");
Expect(0, 57344, '\p{ishighsurrogates}', "");
Expect(1, 57344, '\p{^ishighsurrogates}', "");
Expect(1, 57344, '\P{ishighsurrogates}', "");
Expect(0, 57344, '\P{^ishighsurrogates}', "");
Expect(1, 56191, '\p{_	IS_High_surrogates}', "");
Expect(0, 56191, '\p{^_	IS_High_surrogates}', "");
Expect(0, 56191, '\P{_	IS_High_surrogates}', "");
Expect(1, 56191, '\P{^_	IS_High_surrogates}', "");
Expect(0, 57344, '\p{_	IS_High_surrogates}', "");
Expect(1, 57344, '\p{^_	IS_High_surrogates}', "");
Expect(1, 57344, '\P{_	IS_High_surrogates}', "");
Expect(0, 57344, '\P{^_	IS_High_surrogates}', "");
Error('\p{__In_high_SURROGATES:=}');
Error('\P{__In_high_SURROGATES:=}');
Expect(1, 56191, '\p{inhighsurrogates}', "");
Expect(0, 56191, '\p{^inhighsurrogates}', "");
Expect(0, 56191, '\P{inhighsurrogates}', "");
Expect(1, 56191, '\P{^inhighsurrogates}', "");
Expect(0, 57344, '\p{inhighsurrogates}', "");
Expect(1, 57344, '\p{^inhighsurrogates}', "");
Expect(1, 57344, '\P{inhighsurrogates}', "");
Expect(0, 57344, '\P{^inhighsurrogates}', "");
Expect(1, 56191, '\p{-_In_High_Surrogates}', "");
Expect(0, 56191, '\p{^-_In_High_Surrogates}', "");
Expect(0, 56191, '\P{-_In_High_Surrogates}', "");
Expect(1, 56191, '\P{^-_In_High_Surrogates}', "");
Expect(0, 57344, '\p{-_In_High_Surrogates}', "");
Expect(1, 57344, '\p{^-_In_High_Surrogates}', "");
Expect(1, 57344, '\P{-_In_High_Surrogates}', "");
Expect(0, 57344, '\P{^-_In_High_Surrogates}', "");
Error('\p{_/a/In_DESERET}');
Error('\P{_/a/In_DESERET}');
Expect(1, 66639, '\p{indeseret}', "");
Expect(0, 66639, '\p{^indeseret}', "");
Expect(0, 66639, '\P{indeseret}', "");
Expect(1, 66639, '\P{^indeseret}', "");
Expect(0, 66640, '\p{indeseret}', "");
Expect(1, 66640, '\p{^indeseret}', "");
Expect(1, 66640, '\P{indeseret}', "");
Expect(0, 66640, '\P{^indeseret}', "");
Expect(1, 66639, '\p{ in_Deseret}', "");
Expect(0, 66639, '\p{^ in_Deseret}', "");
Expect(0, 66639, '\P{ in_Deseret}', "");
Expect(1, 66639, '\P{^ in_Deseret}', "");
Expect(0, 66640, '\p{ in_Deseret}', "");
Expect(1, 66640, '\p{^ in_Deseret}', "");
Expect(1, 66640, '\P{ in_Deseret}', "");
Expect(0, 66640, '\P{^ in_Deseret}', "");
Error('\p{  IN_Deseret:=}');
Error('\P{  IN_Deseret:=}');
Expect(1, 66639, '\p{_IN_DESERET}', "");
Expect(0, 66639, '\p{^_IN_DESERET}', "");
Expect(0, 66639, '\P{_IN_DESERET}', "");
Expect(1, 66639, '\P{^_IN_DESERET}', "");
Expect(0, 66640, '\p{_IN_DESERET}', "");
Expect(1, 66640, '\p{^_IN_DESERET}', "");
Expect(1, 66640, '\P{_IN_DESERET}', "");
Expect(0, 66640, '\P{^_IN_DESERET}', "");
Error('\p{Other_Alphabetic}');
Error('\P{Other_Alphabetic}');
Error('\p{Is_Other_Alphabetic}');
Error('\P{Is_Other_Alphabetic}');
Error('\p{OAlpha}');
Error('\P{OAlpha}');
Error('\p{Is_OAlpha}');
Error('\P{Is_OAlpha}');
Error('\p{		saurashtra/a/}');
Error('\P{		saurashtra/a/}');
Expect(1, 43225, '\p{saurashtra}', "");
Expect(0, 43225, '\p{^saurashtra}', "");
Expect(0, 43225, '\P{saurashtra}', "");
Expect(1, 43225, '\P{^saurashtra}', "");
Expect(0, 43226, '\p{saurashtra}', "");
Expect(1, 43226, '\p{^saurashtra}', "");
Expect(1, 43226, '\P{saurashtra}', "");
Expect(0, 43226, '\P{^saurashtra}', "");
Expect(1, 43225, '\p{-	Saurashtra}', "");
Expect(0, 43225, '\p{^-	Saurashtra}', "");
Expect(0, 43225, '\P{-	Saurashtra}', "");
Expect(1, 43225, '\P{^-	Saurashtra}', "");
Expect(0, 43226, '\p{-	Saurashtra}', "");
Expect(1, 43226, '\p{^-	Saurashtra}', "");
Expect(1, 43226, '\P{-	Saurashtra}', "");
Expect(0, 43226, '\P{^-	Saurashtra}', "");
Error('\p{	Is_Saurashtra:=}');
Error('\P{	Is_Saurashtra:=}');
Expect(1, 43225, '\p{issaurashtra}', "");
Expect(0, 43225, '\p{^issaurashtra}', "");
Expect(0, 43225, '\P{issaurashtra}', "");
Expect(1, 43225, '\P{^issaurashtra}', "");
Expect(0, 43226, '\p{issaurashtra}', "");
Expect(1, 43226, '\p{^issaurashtra}', "");
Expect(1, 43226, '\P{issaurashtra}', "");
Expect(0, 43226, '\P{^issaurashtra}', "");
Expect(1, 43225, '\p{_	IS_Saurashtra}', "");
Expect(0, 43225, '\p{^_	IS_Saurashtra}', "");
Expect(0, 43225, '\P{_	IS_Saurashtra}', "");
Expect(1, 43225, '\P{^_	IS_Saurashtra}', "");
Expect(0, 43226, '\p{_	IS_Saurashtra}', "");
Expect(1, 43226, '\p{^_	IS_Saurashtra}', "");
Expect(1, 43226, '\P{_	IS_Saurashtra}', "");
Expect(0, 43226, '\P{^_	IS_Saurashtra}', "");
Error('\p{:=-	SAUR}');
Error('\P{:=-	SAUR}');
Expect(1, 43225, '\p{saur}', "");
Expect(0, 43225, '\p{^saur}', "");
Expect(0, 43225, '\P{saur}', "");
Expect(1, 43225, '\P{^saur}', "");
Expect(0, 43226, '\p{saur}', "");
Expect(1, 43226, '\p{^saur}', "");
Expect(1, 43226, '\P{saur}', "");
Expect(0, 43226, '\P{^saur}', "");
Expect(1, 43225, '\p{--Saur}', "");
Expect(0, 43225, '\p{^--Saur}', "");
Expect(0, 43225, '\P{--Saur}', "");
Expect(1, 43225, '\P{^--Saur}', "");
Expect(0, 43226, '\p{--Saur}', "");
Expect(1, 43226, '\p{^--Saur}', "");
Expect(1, 43226, '\P{--Saur}', "");
Expect(0, 43226, '\P{^--Saur}', "");
Error('\p{	Is_SAUR:=}');
Error('\P{	Is_SAUR:=}');
Expect(1, 43225, '\p{issaur}', "");
Expect(0, 43225, '\p{^issaur}', "");
Expect(0, 43225, '\P{issaur}', "");
Expect(1, 43225, '\P{^issaur}', "");
Expect(0, 43226, '\p{issaur}', "");
Expect(1, 43226, '\p{^issaur}', "");
Expect(1, 43226, '\P{issaur}', "");
Expect(0, 43226, '\P{^issaur}', "");
Expect(1, 43225, '\p{- is_Saur}', "");
Expect(0, 43225, '\p{^- is_Saur}', "");
Expect(0, 43225, '\P{- is_Saur}', "");
Expect(1, 43225, '\P{^- is_Saur}', "");
Expect(0, 43226, '\p{- is_Saur}', "");
Expect(1, 43226, '\p{^- is_Saur}', "");
Expect(1, 43226, '\P{- is_Saur}', "");
Expect(0, 43226, '\P{^- is_Saur}', "");
Error('\p{	_X_Extend:=}');
Error('\P{	_X_Extend:=}');
Expect(1, 917999, '\p{_X_Extend}', "");
Expect(0, 917999, '\p{^_X_Extend}', "");
Expect(0, 917999, '\P{_X_Extend}', "");
Expect(1, 917999, '\P{^_X_Extend}', "");
Expect(0, 918000, '\p{_X_Extend}', "");
Expect(1, 918000, '\p{^_X_Extend}', "");
Expect(1, 918000, '\P{_X_Extend}', "");
Expect(0, 918000, '\P{^_X_Extend}', "");
Error('\p{	_X_EXTEND/a/}');
Error('\P{	_X_EXTEND/a/}');
Error('\p{/a/_	Tagbanwa}');
Error('\P{/a/_	Tagbanwa}');
Expect(1, 6003, '\p{tagbanwa}', "");
Expect(0, 6003, '\p{^tagbanwa}', "");
Expect(0, 6003, '\P{tagbanwa}', "");
Expect(1, 6003, '\P{^tagbanwa}', "");
Expect(0, 6004, '\p{tagbanwa}', "");
Expect(1, 6004, '\p{^tagbanwa}', "");
Expect(1, 6004, '\P{tagbanwa}', "");
Expect(0, 6004, '\P{^tagbanwa}', "");
Expect(1, 6003, '\p{ -TAGBANWA}', "");
Expect(0, 6003, '\p{^ -TAGBANWA}', "");
Expect(0, 6003, '\P{ -TAGBANWA}', "");
Expect(1, 6003, '\P{^ -TAGBANWA}', "");
Expect(0, 6004, '\p{ -TAGBANWA}', "");
Expect(1, 6004, '\p{^ -TAGBANWA}', "");
Expect(1, 6004, '\P{ -TAGBANWA}', "");
Expect(0, 6004, '\P{^ -TAGBANWA}', "");
Error('\p{ is_Tagbanwa/a/}');
Error('\P{ is_Tagbanwa/a/}');
Expect(1, 6003, '\p{istagbanwa}', "");
Expect(0, 6003, '\p{^istagbanwa}', "");
Expect(0, 6003, '\P{istagbanwa}', "");
Expect(1, 6003, '\P{^istagbanwa}', "");
Expect(0, 6004, '\p{istagbanwa}', "");
Expect(1, 6004, '\p{^istagbanwa}', "");
Expect(1, 6004, '\P{istagbanwa}', "");
Expect(0, 6004, '\P{^istagbanwa}', "");
Expect(1, 6003, '\p{- Is_TAGBANWA}', "");
Expect(0, 6003, '\p{^- Is_TAGBANWA}', "");
Expect(0, 6003, '\P{- Is_TAGBANWA}', "");
Expect(1, 6003, '\P{^- Is_TAGBANWA}', "");
Expect(0, 6004, '\p{- Is_TAGBANWA}', "");
Expect(1, 6004, '\p{^- Is_TAGBANWA}', "");
Expect(1, 6004, '\P{- Is_TAGBANWA}', "");
Expect(0, 6004, '\P{^- Is_TAGBANWA}', "");
Error('\p{:= TAGB}');
Error('\P{:= TAGB}');
Expect(1, 6003, '\p{tagb}', "");
Expect(0, 6003, '\p{^tagb}', "");
Expect(0, 6003, '\P{tagb}', "");
Expect(1, 6003, '\P{^tagb}', "");
Expect(0, 6004, '\p{tagb}', "");
Expect(1, 6004, '\p{^tagb}', "");
Expect(1, 6004, '\P{tagb}', "");
Expect(0, 6004, '\P{^tagb}', "");
Expect(1, 6003, '\p{ _tagb}', "");
Expect(0, 6003, '\p{^ _tagb}', "");
Expect(0, 6003, '\P{ _tagb}', "");
Expect(1, 6003, '\P{^ _tagb}', "");
Expect(0, 6004, '\p{ _tagb}', "");
Expect(1, 6004, '\p{^ _tagb}', "");
Expect(1, 6004, '\P{ _tagb}', "");
Expect(0, 6004, '\P{^ _tagb}', "");
Error('\p{  Is_Tagb/a/}');
Error('\P{  Is_Tagb/a/}');
Expect(1, 6003, '\p{istagb}', "");
Expect(0, 6003, '\p{^istagb}', "");
Expect(0, 6003, '\P{istagb}', "");
Expect(1, 6003, '\P{^istagb}', "");
Expect(0, 6004, '\p{istagb}', "");
Expect(1, 6004, '\p{^istagb}', "");
Expect(1, 6004, '\P{istagb}', "");
Expect(0, 6004, '\P{^istagb}', "");
Expect(1, 6003, '\p{	 Is_Tagb}', "");
Expect(0, 6003, '\p{^	 Is_Tagb}', "");
Expect(0, 6003, '\P{	 Is_Tagb}', "");
Expect(1, 6003, '\P{^	 Is_Tagb}', "");
Expect(0, 6004, '\p{	 Is_Tagb}', "");
Expect(1, 6004, '\p{^	 Is_Tagb}', "");
Expect(1, 6004, '\P{	 Is_Tagb}', "");
Expect(0, 6004, '\P{^	 Is_Tagb}', "");
Error('\p{-miscellaneous_mathematical_Symbols_A:=}');
Error('\P{-miscellaneous_mathematical_Symbols_A:=}');
Expect(1, 10223, '\p{miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\p{^miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\P{miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\P{^miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\p{miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\p{^miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\P{miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\P{^miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\p{ -Miscellaneous_Mathematical_Symbols_a}', "");
Expect(0, 10223, '\p{^ -Miscellaneous_Mathematical_Symbols_a}', "");
Expect(0, 10223, '\P{ -Miscellaneous_Mathematical_Symbols_a}', "");
Expect(1, 10223, '\P{^ -Miscellaneous_Mathematical_Symbols_a}', "");
Expect(0, 10224, '\p{ -Miscellaneous_Mathematical_Symbols_a}', "");
Expect(1, 10224, '\p{^ -Miscellaneous_Mathematical_Symbols_a}', "");
Expect(1, 10224, '\P{ -Miscellaneous_Mathematical_Symbols_a}', "");
Expect(0, 10224, '\P{^ -Miscellaneous_Mathematical_Symbols_a}', "");
Error('\p{ _Is_Miscellaneous_MATHEMATICAL_Symbols_A/a/}');
Error('\P{ _Is_Miscellaneous_MATHEMATICAL_Symbols_A/a/}');
Expect(1, 10223, '\p{ismiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\p{^ismiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\P{ismiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\P{^ismiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\p{ismiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\p{^ismiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\P{ismiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\P{^ismiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\p{--IS_Miscellaneous_mathematical_SYMBOLS_a}', "");
Expect(0, 10223, '\p{^--IS_Miscellaneous_mathematical_SYMBOLS_a}', "");
Expect(0, 10223, '\P{--IS_Miscellaneous_mathematical_SYMBOLS_a}', "");
Expect(1, 10223, '\P{^--IS_Miscellaneous_mathematical_SYMBOLS_a}', "");
Expect(0, 10224, '\p{--IS_Miscellaneous_mathematical_SYMBOLS_a}', "");
Expect(1, 10224, '\p{^--IS_Miscellaneous_mathematical_SYMBOLS_a}', "");
Expect(1, 10224, '\P{--IS_Miscellaneous_mathematical_SYMBOLS_a}', "");
Expect(0, 10224, '\P{^--IS_Miscellaneous_mathematical_SYMBOLS_a}', "");
Error('\p{_/a/IN_miscellaneous_Mathematical_Symbols_A}');
Error('\P{_/a/IN_miscellaneous_Mathematical_Symbols_A}');
Expect(1, 10223, '\p{inmiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\p{^inmiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\P{inmiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\P{^inmiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\p{inmiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\p{^inmiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\P{inmiscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\P{^inmiscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\p{-	IN_Miscellaneous_Mathematical_Symbols_A}', "");
Expect(0, 10223, '\p{^-	IN_Miscellaneous_Mathematical_Symbols_A}', "");
Expect(0, 10223, '\P{-	IN_Miscellaneous_Mathematical_Symbols_A}', "");
Expect(1, 10223, '\P{^-	IN_Miscellaneous_Mathematical_Symbols_A}', "");
Expect(0, 10224, '\p{-	IN_Miscellaneous_Mathematical_Symbols_A}', "");
Expect(1, 10224, '\p{^-	IN_Miscellaneous_Mathematical_Symbols_A}', "");
Expect(1, 10224, '\P{-	IN_Miscellaneous_Mathematical_Symbols_A}', "");
Expect(0, 10224, '\P{^-	IN_Miscellaneous_Mathematical_Symbols_A}', "");
Error('\p{/a/-_Misc_MATH_symbols_A}');
Error('\P{/a/-_Misc_MATH_symbols_A}');
Expect(1, 10223, '\p{miscmathsymbolsa}', "");
Expect(0, 10223, '\p{^miscmathsymbolsa}', "");
Expect(0, 10223, '\P{miscmathsymbolsa}', "");
Expect(1, 10223, '\P{^miscmathsymbolsa}', "");
Expect(0, 10224, '\p{miscmathsymbolsa}', "");
Expect(1, 10224, '\p{^miscmathsymbolsa}', "");
Expect(1, 10224, '\P{miscmathsymbolsa}', "");
Expect(0, 10224, '\P{^miscmathsymbolsa}', "");
Expect(1, 10223, '\p{	Misc_math_Symbols_A}', "");
Expect(0, 10223, '\p{^	Misc_math_Symbols_A}', "");
Expect(0, 10223, '\P{	Misc_math_Symbols_A}', "");
Expect(1, 10223, '\P{^	Misc_math_Symbols_A}', "");
Expect(0, 10224, '\p{	Misc_math_Symbols_A}', "");
Expect(1, 10224, '\p{^	Misc_math_Symbols_A}', "");
Expect(1, 10224, '\P{	Misc_math_Symbols_A}', "");
Expect(0, 10224, '\P{^	Misc_math_Symbols_A}', "");
Error('\p{/a/ 	is_Misc_Math_Symbols_A}');
Error('\P{/a/ 	is_Misc_Math_Symbols_A}');
Expect(1, 10223, '\p{ismiscmathsymbolsa}', "");
Expect(0, 10223, '\p{^ismiscmathsymbolsa}', "");
Expect(0, 10223, '\P{ismiscmathsymbolsa}', "");
Expect(1, 10223, '\P{^ismiscmathsymbolsa}', "");
Expect(0, 10224, '\p{ismiscmathsymbolsa}', "");
Expect(1, 10224, '\p{^ismiscmathsymbolsa}', "");
Expect(1, 10224, '\P{ismiscmathsymbolsa}', "");
Expect(0, 10224, '\P{^ismiscmathsymbolsa}', "");
Expect(1, 10223, '\p{	is_MISC_Math_SYMBOLS_A}', "");
Expect(0, 10223, '\p{^	is_MISC_Math_SYMBOLS_A}', "");
Expect(0, 10223, '\P{	is_MISC_Math_SYMBOLS_A}', "");
Expect(1, 10223, '\P{^	is_MISC_Math_SYMBOLS_A}', "");
Expect(0, 10224, '\p{	is_MISC_Math_SYMBOLS_A}', "");
Expect(1, 10224, '\p{^	is_MISC_Math_SYMBOLS_A}', "");
Expect(1, 10224, '\P{	is_MISC_Math_SYMBOLS_A}', "");
Expect(0, 10224, '\P{^	is_MISC_Math_SYMBOLS_A}', "");
Error('\p{/a/	_In_Misc_Math_Symbols_A}');
Error('\P{/a/	_In_Misc_Math_Symbols_A}');
Expect(1, 10223, '\p{inmiscmathsymbolsa}', "");
Expect(0, 10223, '\p{^inmiscmathsymbolsa}', "");
Expect(0, 10223, '\P{inmiscmathsymbolsa}', "");
Expect(1, 10223, '\P{^inmiscmathsymbolsa}', "");
Expect(0, 10224, '\p{inmiscmathsymbolsa}', "");
Expect(1, 10224, '\p{^inmiscmathsymbolsa}', "");
Expect(1, 10224, '\P{inmiscmathsymbolsa}', "");
Expect(0, 10224, '\P{^inmiscmathsymbolsa}', "");
Expect(1, 10223, '\p{	_in_misc_math_Symbols_A}', "");
Expect(0, 10223, '\p{^	_in_misc_math_Symbols_A}', "");
Expect(0, 10223, '\P{	_in_misc_math_Symbols_A}', "");
Expect(1, 10223, '\P{^	_in_misc_math_Symbols_A}', "");
Expect(0, 10224, '\p{	_in_misc_math_Symbols_A}', "");
Expect(1, 10224, '\p{^	_in_misc_math_Symbols_A}', "");
Expect(1, 10224, '\P{	_in_misc_math_Symbols_A}', "");
Expect(0, 10224, '\P{^	_in_misc_math_Symbols_A}', "");
Error('\p{	:=Khmer_Symbols}');
Error('\P{	:=Khmer_Symbols}');
Expect(1, 6655, '\p{khmersymbols}', "");
Expect(0, 6655, '\p{^khmersymbols}', "");
Expect(0, 6655, '\P{khmersymbols}', "");
Expect(1, 6655, '\P{^khmersymbols}', "");
Expect(0, 6656, '\p{khmersymbols}', "");
Expect(1, 6656, '\p{^khmersymbols}', "");
Expect(1, 6656, '\P{khmersymbols}', "");
Expect(0, 6656, '\P{^khmersymbols}', "");
Expect(1, 6655, '\p{ 	Khmer_Symbols}', "");
Expect(0, 6655, '\p{^ 	Khmer_Symbols}', "");
Expect(0, 6655, '\P{ 	Khmer_Symbols}', "");
Expect(1, 6655, '\P{^ 	Khmer_Symbols}', "");
Expect(0, 6656, '\p{ 	Khmer_Symbols}', "");
Expect(1, 6656, '\p{^ 	Khmer_Symbols}', "");
Expect(1, 6656, '\P{ 	Khmer_Symbols}', "");
Expect(0, 6656, '\P{^ 	Khmer_Symbols}', "");
Error('\p{/a/IS_KHMER_SYMBOLS}');
Error('\P{/a/IS_KHMER_SYMBOLS}');
Expect(1, 6655, '\p{iskhmersymbols}', "");
Expect(0, 6655, '\p{^iskhmersymbols}', "");
Expect(0, 6655, '\P{iskhmersymbols}', "");
Expect(1, 6655, '\P{^iskhmersymbols}', "");
Expect(0, 6656, '\p{iskhmersymbols}', "");
Expect(1, 6656, '\p{^iskhmersymbols}', "");
Expect(1, 6656, '\P{iskhmersymbols}', "");
Expect(0, 6656, '\P{^iskhmersymbols}', "");
Expect(1, 6655, '\p{_	IS_Khmer_Symbols}', "");
Expect(0, 6655, '\p{^_	IS_Khmer_Symbols}', "");
Expect(0, 6655, '\P{_	IS_Khmer_Symbols}', "");
Expect(1, 6655, '\P{^_	IS_Khmer_Symbols}', "");
Expect(0, 6656, '\p{_	IS_Khmer_Symbols}', "");
Expect(1, 6656, '\p{^_	IS_Khmer_Symbols}', "");
Expect(1, 6656, '\P{_	IS_Khmer_Symbols}', "");
Expect(0, 6656, '\P{^_	IS_Khmer_Symbols}', "");
Error('\p{/a/	IN_khmer_symbols}');
Error('\P{/a/	IN_khmer_symbols}');
Expect(1, 6655, '\p{inkhmersymbols}', "");
Expect(0, 6655, '\p{^inkhmersymbols}', "");
Expect(0, 6655, '\P{inkhmersymbols}', "");
Expect(1, 6655, '\P{^inkhmersymbols}', "");
Expect(0, 6656, '\p{inkhmersymbols}', "");
Expect(1, 6656, '\p{^inkhmersymbols}', "");
Expect(1, 6656, '\P{inkhmersymbols}', "");
Expect(0, 6656, '\P{^inkhmersymbols}', "");
Expect(1, 6655, '\p{ In_Khmer_SYMBOLS}', "");
Expect(0, 6655, '\p{^ In_Khmer_SYMBOLS}', "");
Expect(0, 6655, '\P{ In_Khmer_SYMBOLS}', "");
Expect(1, 6655, '\P{^ In_Khmer_SYMBOLS}', "");
Expect(0, 6656, '\p{ In_Khmer_SYMBOLS}', "");
Expect(1, 6656, '\p{^ In_Khmer_SYMBOLS}', "");
Expect(1, 6656, '\P{ In_Khmer_SYMBOLS}', "");
Expect(0, 6656, '\P{^ In_Khmer_SYMBOLS}', "");
Error('\p{- HANGUL_JAMO_Extended_A:=}');
Error('\P{- HANGUL_JAMO_Extended_A:=}');
Expect(1, 43391, '\p{hanguljamoextendeda}', "");
Expect(0, 43391, '\p{^hanguljamoextendeda}', "");
Expect(0, 43391, '\P{hanguljamoextendeda}', "");
Expect(1, 43391, '\P{^hanguljamoextendeda}', "");
Expect(0, 43392, '\p{hanguljamoextendeda}', "");
Expect(1, 43392, '\p{^hanguljamoextendeda}', "");
Expect(1, 43392, '\P{hanguljamoextendeda}', "");
Expect(0, 43392, '\P{^hanguljamoextendeda}', "");
Expect(1, 43391, '\p{_HANGUL_jamo_Extended_A}', "");
Expect(0, 43391, '\p{^_HANGUL_jamo_Extended_A}', "");
Expect(0, 43391, '\P{_HANGUL_jamo_Extended_A}', "");
Expect(1, 43391, '\P{^_HANGUL_jamo_Extended_A}', "");
Expect(0, 43392, '\p{_HANGUL_jamo_Extended_A}', "");
Expect(1, 43392, '\p{^_HANGUL_jamo_Extended_A}', "");
Expect(1, 43392, '\P{_HANGUL_jamo_Extended_A}', "");
Expect(0, 43392, '\P{^_HANGUL_jamo_Extended_A}', "");
Error('\p{/a/		Is_Hangul_JAMO_Extended_A}');
Error('\P{/a/		Is_Hangul_JAMO_Extended_A}');
Expect(1, 43391, '\p{ishanguljamoextendeda}', "");
Expect(0, 43391, '\p{^ishanguljamoextendeda}', "");
Expect(0, 43391, '\P{ishanguljamoextendeda}', "");
Expect(1, 43391, '\P{^ishanguljamoextendeda}', "");
Expect(0, 43392, '\p{ishanguljamoextendeda}', "");
Expect(1, 43392, '\p{^ishanguljamoextendeda}', "");
Expect(1, 43392, '\P{ishanguljamoextendeda}', "");
Expect(0, 43392, '\P{^ishanguljamoextendeda}', "");
Expect(1, 43391, '\p{-	is_HANGUL_jamo_Extended_a}', "");
Expect(0, 43391, '\p{^-	is_HANGUL_jamo_Extended_a}', "");
Expect(0, 43391, '\P{-	is_HANGUL_jamo_Extended_a}', "");
Expect(1, 43391, '\P{^-	is_HANGUL_jamo_Extended_a}', "");
Expect(0, 43392, '\p{-	is_HANGUL_jamo_Extended_a}', "");
Expect(1, 43392, '\p{^-	is_HANGUL_jamo_Extended_a}', "");
Expect(1, 43392, '\P{-	is_HANGUL_jamo_Extended_a}', "");
Expect(0, 43392, '\P{^-	is_HANGUL_jamo_Extended_a}', "");
Error('\p{:=In_HANGUL_Jamo_EXTENDED_A}');
Error('\P{:=In_HANGUL_Jamo_EXTENDED_A}');
Expect(1, 43391, '\p{inhanguljamoextendeda}', "");
Expect(0, 43391, '\p{^inhanguljamoextendeda}', "");
Expect(0, 43391, '\P{inhanguljamoextendeda}', "");
Expect(1, 43391, '\P{^inhanguljamoextendeda}', "");
Expect(0, 43392, '\p{inhanguljamoextendeda}', "");
Expect(1, 43392, '\p{^inhanguljamoextendeda}', "");
Expect(1, 43392, '\P{inhanguljamoextendeda}', "");
Expect(0, 43392, '\P{^inhanguljamoextendeda}', "");
Expect(1, 43391, '\p{In_Hangul_Jamo_EXTENDED_A}', "");
Expect(0, 43391, '\p{^In_Hangul_Jamo_EXTENDED_A}', "");
Expect(0, 43391, '\P{In_Hangul_Jamo_EXTENDED_A}', "");
Expect(1, 43391, '\P{^In_Hangul_Jamo_EXTENDED_A}', "");
Expect(0, 43392, '\p{In_Hangul_Jamo_EXTENDED_A}', "");
Expect(1, 43392, '\p{^In_Hangul_Jamo_EXTENDED_A}', "");
Expect(1, 43392, '\P{In_Hangul_Jamo_EXTENDED_A}', "");
Expect(0, 43392, '\P{^In_Hangul_Jamo_EXTENDED_A}', "");
Error('\p{	/a/Jamo_EXT_A}');
Error('\P{	/a/Jamo_EXT_A}');
Expect(1, 43391, '\p{jamoexta}', "");
Expect(0, 43391, '\p{^jamoexta}', "");
Expect(0, 43391, '\P{jamoexta}', "");
Expect(1, 43391, '\P{^jamoexta}', "");
Expect(0, 43392, '\p{jamoexta}', "");
Expect(1, 43392, '\p{^jamoexta}', "");
Expect(1, 43392, '\P{jamoexta}', "");
Expect(0, 43392, '\P{^jamoexta}', "");
Expect(1, 43391, '\p{_Jamo_Ext_A}', "");
Expect(0, 43391, '\p{^_Jamo_Ext_A}', "");
Expect(0, 43391, '\P{_Jamo_Ext_A}', "");
Expect(1, 43391, '\P{^_Jamo_Ext_A}', "");
Expect(0, 43392, '\p{_Jamo_Ext_A}', "");
Expect(1, 43392, '\p{^_Jamo_Ext_A}', "");
Expect(1, 43392, '\P{_Jamo_Ext_A}', "");
Expect(0, 43392, '\P{^_Jamo_Ext_A}', "");
Error('\p{  Is_Jamo_Ext_A/a/}');
Error('\P{  Is_Jamo_Ext_A/a/}');
Expect(1, 43391, '\p{isjamoexta}', "");
Expect(0, 43391, '\p{^isjamoexta}', "");
Expect(0, 43391, '\P{isjamoexta}', "");
Expect(1, 43391, '\P{^isjamoexta}', "");
Expect(0, 43392, '\p{isjamoexta}', "");
Expect(1, 43392, '\p{^isjamoexta}', "");
Expect(1, 43392, '\P{isjamoexta}', "");
Expect(0, 43392, '\P{^isjamoexta}', "");
Expect(1, 43391, '\p{		is_jamo_ext_A}', "");
Expect(0, 43391, '\p{^		is_jamo_ext_A}', "");
Expect(0, 43391, '\P{		is_jamo_ext_A}', "");
Expect(1, 43391, '\P{^		is_jamo_ext_A}', "");
Expect(0, 43392, '\p{		is_jamo_ext_A}', "");
Expect(1, 43392, '\p{^		is_jamo_ext_A}', "");
Expect(1, 43392, '\P{		is_jamo_ext_A}', "");
Expect(0, 43392, '\P{^		is_jamo_ext_A}', "");
Error('\p{/a/In_Jamo_Ext_A}');
Error('\P{/a/In_Jamo_Ext_A}');
Expect(1, 43391, '\p{injamoexta}', "");
Expect(0, 43391, '\p{^injamoexta}', "");
Expect(0, 43391, '\P{injamoexta}', "");
Expect(1, 43391, '\P{^injamoexta}', "");
Expect(0, 43392, '\p{injamoexta}', "");
Expect(1, 43392, '\p{^injamoexta}', "");
Expect(1, 43392, '\P{injamoexta}', "");
Expect(0, 43392, '\P{^injamoexta}', "");
Expect(1, 43391, '\p{_In_Jamo_ext_A}', "");
Expect(0, 43391, '\p{^_In_Jamo_ext_A}', "");
Expect(0, 43391, '\P{_In_Jamo_ext_A}', "");
Expect(1, 43391, '\P{^_In_Jamo_ext_A}', "");
Expect(0, 43392, '\p{_In_Jamo_ext_A}', "");
Expect(1, 43392, '\p{^_In_Jamo_ext_A}', "");
Expect(1, 43392, '\P{_In_Jamo_ext_A}', "");
Expect(0, 43392, '\P{^_In_Jamo_ext_A}', "");
Error('\p{		Math/a/}');
Error('\P{		Math/a/}');
Expect(1, 126705, '\p{math}', "");
Expect(0, 126705, '\p{^math}', "");
Expect(0, 126705, '\P{math}', "");
Expect(1, 126705, '\P{^math}', "");
Expect(0, 126706, '\p{math}', "");
Expect(1, 126706, '\p{^math}', "");
Expect(1, 126706, '\P{math}', "");
Expect(0, 126706, '\P{^math}', "");
Expect(1, 126705, '\p{_math}', "");
Expect(0, 126705, '\p{^_math}', "");
Expect(0, 126705, '\P{_math}', "");
Expect(1, 126705, '\P{^_math}', "");
Expect(0, 126706, '\p{_math}', "");
Expect(1, 126706, '\p{^_math}', "");
Expect(1, 126706, '\P{_math}', "");
Expect(0, 126706, '\P{^_math}', "");
Error('\p{_	IS_Math:=}');
Error('\P{_	IS_Math:=}');
Expect(1, 126705, '\p{ismath}', "");
Expect(0, 126705, '\p{^ismath}', "");
Expect(0, 126705, '\P{ismath}', "");
Expect(1, 126705, '\P{^ismath}', "");
Expect(0, 126706, '\p{ismath}', "");
Expect(1, 126706, '\p{^ismath}', "");
Expect(1, 126706, '\P{ismath}', "");
Expect(0, 126706, '\P{^ismath}', "");
Expect(1, 126705, '\p{-Is_MATH}', "");
Expect(0, 126705, '\p{^-Is_MATH}', "");
Expect(0, 126705, '\P{-Is_MATH}', "");
Expect(1, 126705, '\P{^-Is_MATH}', "");
Expect(0, 126706, '\p{-Is_MATH}', "");
Expect(1, 126706, '\p{^-Is_MATH}', "");
Expect(1, 126706, '\P{-Is_MATH}', "");
Expect(0, 126706, '\P{^-Is_MATH}', "");
Error('\p{-BUHID/a/}');
Error('\P{-BUHID/a/}');
Expect(1, 5971, '\p{buhid}', "");
Expect(0, 5971, '\p{^buhid}', "");
Expect(0, 5971, '\P{buhid}', "");
Expect(1, 5971, '\P{^buhid}', "");
Expect(0, 5972, '\p{buhid}', "");
Expect(1, 5972, '\p{^buhid}', "");
Expect(1, 5972, '\P{buhid}', "");
Expect(0, 5972, '\P{^buhid}', "");
Expect(1, 5971, '\p{-Buhid}', "");
Expect(0, 5971, '\p{^-Buhid}', "");
Expect(0, 5971, '\P{-Buhid}', "");
Expect(1, 5971, '\P{^-Buhid}', "");
Expect(0, 5972, '\p{-Buhid}', "");
Expect(1, 5972, '\p{^-Buhid}', "");
Expect(1, 5972, '\P{-Buhid}', "");
Expect(0, 5972, '\P{^-Buhid}', "");
Error('\p{:=_Is_Buhid}');
Error('\P{:=_Is_Buhid}');
Expect(1, 5971, '\p{isbuhid}', "");
Expect(0, 5971, '\p{^isbuhid}', "");
Expect(0, 5971, '\P{isbuhid}', "");
Expect(1, 5971, '\P{^isbuhid}', "");
Expect(0, 5972, '\p{isbuhid}', "");
Expect(1, 5972, '\p{^isbuhid}', "");
Expect(1, 5972, '\P{isbuhid}', "");
Expect(0, 5972, '\P{^isbuhid}', "");
Expect(1, 5971, '\p{_	is_buhid}', "");
Expect(0, 5971, '\p{^_	is_buhid}', "");
Expect(0, 5971, '\P{_	is_buhid}', "");
Expect(1, 5971, '\P{^_	is_buhid}', "");
Expect(0, 5972, '\p{_	is_buhid}', "");
Expect(1, 5972, '\p{^_	is_buhid}', "");
Expect(1, 5972, '\P{_	is_buhid}', "");
Expect(0, 5972, '\P{^_	is_buhid}', "");
Error('\p{_	buhd/a/}');
Error('\P{_	buhd/a/}');
Expect(1, 5971, '\p{buhd}', "");
Expect(0, 5971, '\p{^buhd}', "");
Expect(0, 5971, '\P{buhd}', "");
Expect(1, 5971, '\P{^buhd}', "");
Expect(0, 5972, '\p{buhd}', "");
Expect(1, 5972, '\p{^buhd}', "");
Expect(1, 5972, '\P{buhd}', "");
Expect(0, 5972, '\P{^buhd}', "");
Expect(1, 5971, '\p{_ buhd}', "");
Expect(0, 5971, '\p{^_ buhd}', "");
Expect(0, 5971, '\P{_ buhd}', "");
Expect(1, 5971, '\P{^_ buhd}', "");
Expect(0, 5972, '\p{_ buhd}', "");
Expect(1, 5972, '\p{^_ buhd}', "");
Expect(1, 5972, '\P{_ buhd}', "");
Expect(0, 5972, '\P{^_ buhd}', "");
Error('\p{:=-Is_Buhd}');
Error('\P{:=-Is_Buhd}');
Expect(1, 5971, '\p{isbuhd}', "");
Expect(0, 5971, '\p{^isbuhd}', "");
Expect(0, 5971, '\P{isbuhd}', "");
Expect(1, 5971, '\P{^isbuhd}', "");
Expect(0, 5972, '\p{isbuhd}', "");
Expect(1, 5972, '\p{^isbuhd}', "");
Expect(1, 5972, '\P{isbuhd}', "");
Expect(0, 5972, '\P{^isbuhd}', "");
Expect(1, 5971, '\p{	_Is_Buhd}', "");
Expect(0, 5971, '\p{^	_Is_Buhd}', "");
Expect(0, 5971, '\P{	_Is_Buhd}', "");
Expect(1, 5971, '\P{^	_Is_Buhd}', "");
Expect(0, 5972, '\p{	_Is_Buhd}', "");
Expect(1, 5972, '\p{^	_Is_Buhd}', "");
Expect(1, 5972, '\P{	_Is_Buhd}', "");
Expect(0, 5972, '\P{^	_Is_Buhd}', "");
Error('\p{:=Greek_AND_coptic}');
Error('\P{:=Greek_AND_coptic}');
Expect(1, 1023, '\p{greekandcoptic}', "");
Expect(0, 1023, '\p{^greekandcoptic}', "");
Expect(0, 1023, '\P{greekandcoptic}', "");
Expect(1, 1023, '\P{^greekandcoptic}', "");
Expect(0, 1024, '\p{greekandcoptic}', "");
Expect(1, 1024, '\p{^greekandcoptic}', "");
Expect(1, 1024, '\P{greekandcoptic}', "");
Expect(0, 1024, '\P{^greekandcoptic}', "");
Expect(1, 1023, '\p{	-Greek_And_Coptic}', "");
Expect(0, 1023, '\p{^	-Greek_And_Coptic}', "");
Expect(0, 1023, '\P{	-Greek_And_Coptic}', "");
Expect(1, 1023, '\P{^	-Greek_And_Coptic}', "");
Expect(0, 1024, '\p{	-Greek_And_Coptic}', "");
Expect(1, 1024, '\p{^	-Greek_And_Coptic}', "");
Expect(1, 1024, '\P{	-Greek_And_Coptic}', "");
Expect(0, 1024, '\P{^	-Greek_And_Coptic}', "");
Error('\p{:=-is_Greek_And_coptic}');
Error('\P{:=-is_Greek_And_coptic}');
Expect(1, 1023, '\p{isgreekandcoptic}', "");
Expect(0, 1023, '\p{^isgreekandcoptic}', "");
Expect(0, 1023, '\P{isgreekandcoptic}', "");
Expect(1, 1023, '\P{^isgreekandcoptic}', "");
Expect(0, 1024, '\p{isgreekandcoptic}', "");
Expect(1, 1024, '\p{^isgreekandcoptic}', "");
Expect(1, 1024, '\P{isgreekandcoptic}', "");
Expect(0, 1024, '\P{^isgreekandcoptic}', "");
Expect(1, 1023, '\p{Is_Greek_and_Coptic}', "");
Expect(0, 1023, '\p{^Is_Greek_and_Coptic}', "");
Expect(0, 1023, '\P{Is_Greek_and_Coptic}', "");
Expect(1, 1023, '\P{^Is_Greek_and_Coptic}', "");
Expect(0, 1024, '\p{Is_Greek_and_Coptic}', "");
Expect(1, 1024, '\p{^Is_Greek_and_Coptic}', "");
Expect(1, 1024, '\P{Is_Greek_and_Coptic}', "");
Expect(0, 1024, '\P{^Is_Greek_and_Coptic}', "");
Error('\p{_	in_GREEK_And_Coptic:=}');
Error('\P{_	in_GREEK_And_Coptic:=}');
Expect(1, 1023, '\p{ingreekandcoptic}', "");
Expect(0, 1023, '\p{^ingreekandcoptic}', "");
Expect(0, 1023, '\P{ingreekandcoptic}', "");
Expect(1, 1023, '\P{^ingreekandcoptic}', "");
Expect(0, 1024, '\p{ingreekandcoptic}', "");
Expect(1, 1024, '\p{^ingreekandcoptic}', "");
Expect(1, 1024, '\P{ingreekandcoptic}', "");
Expect(0, 1024, '\P{^ingreekandcoptic}', "");
Expect(1, 1023, '\p{_-in_Greek_And_coptic}', "");
Expect(0, 1023, '\p{^_-in_Greek_And_coptic}', "");
Expect(0, 1023, '\P{_-in_Greek_And_coptic}', "");
Expect(1, 1023, '\P{^_-in_Greek_And_coptic}', "");
Expect(0, 1024, '\p{_-in_Greek_And_coptic}', "");
Expect(1, 1024, '\p{^_-in_Greek_And_coptic}', "");
Expect(1, 1024, '\P{_-in_Greek_And_coptic}', "");
Expect(0, 1024, '\P{^_-in_Greek_And_coptic}', "");
Error('\p{ in_GREEK/a/}');
Error('\P{ in_GREEK/a/}');
Expect(1, 1023, '\p{ingreek}', "");
Expect(0, 1023, '\p{^ingreek}', "");
Expect(0, 1023, '\P{ingreek}', "");
Expect(1, 1023, '\P{^ingreek}', "");
Expect(0, 1024, '\p{ingreek}', "");
Expect(1, 1024, '\p{^ingreek}', "");
Expect(1, 1024, '\P{ingreek}', "");
Expect(0, 1024, '\P{^ingreek}', "");
Expect(1, 1023, '\p{-In_Greek}', "");
Expect(0, 1023, '\p{^-In_Greek}', "");
Expect(0, 1023, '\P{-In_Greek}', "");
Expect(1, 1023, '\P{^-In_Greek}', "");
Expect(0, 1024, '\p{-In_Greek}', "");
Expect(1, 1024, '\p{^-In_Greek}', "");
Expect(1, 1024, '\P{-In_Greek}', "");
Expect(0, 1024, '\P{^-In_Greek}', "");
Error('\p{-/a/white_space}');
Error('\P{-/a/white_space}');
Expect(1, 12288, '\p{whitespace}', "");
Expect(0, 12288, '\p{^whitespace}', "");
Expect(0, 12288, '\P{whitespace}', "");
Expect(1, 12288, '\P{^whitespace}', "");
Expect(0, 12289, '\p{whitespace}', "");
Expect(1, 12289, '\p{^whitespace}', "");
Expect(1, 12289, '\P{whitespace}', "");
Expect(0, 12289, '\P{^whitespace}', "");
Expect(1, 12288, '\p{-_WHITE_SPACE}', "");
Expect(0, 12288, '\p{^-_WHITE_SPACE}', "");
Expect(0, 12288, '\P{-_WHITE_SPACE}', "");
Expect(1, 12288, '\P{^-_WHITE_SPACE}', "");
Expect(0, 12289, '\p{-_WHITE_SPACE}', "");
Expect(1, 12289, '\p{^-_WHITE_SPACE}', "");
Expect(1, 12289, '\P{-_WHITE_SPACE}', "");
Expect(0, 12289, '\P{^-_WHITE_SPACE}', "");
Error('\p{_:=Is_White_space}');
Error('\P{_:=Is_White_space}');
Expect(1, 12288, '\p{iswhitespace}', "");
Expect(0, 12288, '\p{^iswhitespace}', "");
Expect(0, 12288, '\P{iswhitespace}', "");
Expect(1, 12288, '\P{^iswhitespace}', "");
Expect(0, 12289, '\p{iswhitespace}', "");
Expect(1, 12289, '\p{^iswhitespace}', "");
Expect(1, 12289, '\P{iswhitespace}', "");
Expect(0, 12289, '\P{^iswhitespace}', "");
Expect(1, 12288, '\p{	IS_WHITE_Space}', "");
Expect(0, 12288, '\p{^	IS_WHITE_Space}', "");
Expect(0, 12288, '\P{	IS_WHITE_Space}', "");
Expect(1, 12288, '\P{^	IS_WHITE_Space}', "");
Expect(0, 12289, '\p{	IS_WHITE_Space}', "");
Expect(1, 12289, '\p{^	IS_WHITE_Space}', "");
Expect(1, 12289, '\P{	IS_WHITE_Space}', "");
Expect(0, 12289, '\P{^	IS_WHITE_Space}', "");
Error('\p{:=	 WSpace}');
Error('\P{:=	 WSpace}');
Expect(1, 12288, '\p{wspace}', "");
Expect(0, 12288, '\p{^wspace}', "");
Expect(0, 12288, '\P{wspace}', "");
Expect(1, 12288, '\P{^wspace}', "");
Expect(0, 12289, '\p{wspace}', "");
Expect(1, 12289, '\p{^wspace}', "");
Expect(1, 12289, '\P{wspace}', "");
Expect(0, 12289, '\P{^wspace}', "");
Expect(1, 12288, '\p{-_WSPACE}', "");
Expect(0, 12288, '\p{^-_WSPACE}', "");
Expect(0, 12288, '\P{-_WSPACE}', "");
Expect(1, 12288, '\P{^-_WSPACE}', "");
Expect(0, 12289, '\p{-_WSPACE}', "");
Expect(1, 12289, '\p{^-_WSPACE}', "");
Expect(1, 12289, '\P{-_WSPACE}', "");
Expect(0, 12289, '\P{^-_WSPACE}', "");
Error('\p{:=Is_wspace}');
Error('\P{:=Is_wspace}');
Expect(1, 12288, '\p{iswspace}', "");
Expect(0, 12288, '\p{^iswspace}', "");
Expect(0, 12288, '\P{iswspace}', "");
Expect(1, 12288, '\P{^iswspace}', "");
Expect(0, 12289, '\p{iswspace}', "");
Expect(1, 12289, '\p{^iswspace}', "");
Expect(1, 12289, '\P{iswspace}', "");
Expect(0, 12289, '\P{^iswspace}', "");
Expect(1, 12288, '\p{__is_WSpace}', "");
Expect(0, 12288, '\p{^__is_WSpace}', "");
Expect(0, 12288, '\P{__is_WSpace}', "");
Expect(1, 12288, '\P{^__is_WSpace}', "");
Expect(0, 12289, '\p{__is_WSpace}', "");
Expect(1, 12289, '\p{^__is_WSpace}', "");
Expect(1, 12289, '\P{__is_WSpace}', "");
Expect(0, 12289, '\P{^__is_WSpace}', "");
Error('\p{/a/ In_CUNEIFORM}');
Error('\P{/a/ In_CUNEIFORM}');
Expect(1, 74751, '\p{incuneiform}', "");
Expect(0, 74751, '\p{^incuneiform}', "");
Expect(0, 74751, '\P{incuneiform}', "");
Expect(1, 74751, '\P{^incuneiform}', "");
Expect(0, 74752, '\p{incuneiform}', "");
Expect(1, 74752, '\p{^incuneiform}', "");
Expect(1, 74752, '\P{incuneiform}', "");
Expect(0, 74752, '\P{^incuneiform}', "");
Expect(1, 74751, '\p{-	in_Cuneiform}', "");
Expect(0, 74751, '\p{^-	in_Cuneiform}', "");
Expect(0, 74751, '\P{-	in_Cuneiform}', "");
Expect(1, 74751, '\P{^-	in_Cuneiform}', "");
Expect(0, 74752, '\p{-	in_Cuneiform}', "");
Expect(1, 74752, '\p{^-	in_Cuneiform}', "");
Expect(1, 74752, '\P{-	in_Cuneiform}', "");
Expect(0, 74752, '\P{^-	in_Cuneiform}', "");
Error('\p{:=		In_cuneiform}');
Error('\P{:=		In_cuneiform}');
Expect(1, 74751, '\p{	-In_Cuneiform}', "");
Expect(0, 74751, '\p{^	-In_Cuneiform}', "");
Expect(0, 74751, '\P{	-In_Cuneiform}', "");
Expect(1, 74751, '\P{^	-In_Cuneiform}', "");
Expect(0, 74752, '\p{	-In_Cuneiform}', "");
Expect(1, 74752, '\p{^	-In_Cuneiform}', "");
Expect(1, 74752, '\P{	-In_Cuneiform}', "");
Expect(0, 74752, '\P{^	-In_Cuneiform}', "");
Error('\p{	_CJK_UNIFIED_Ideographs_Extension_B:=}');
Error('\P{	_CJK_UNIFIED_Ideographs_Extension_B:=}');
Expect(1, 173791, '\p{cjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\p{^cjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\P{cjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\P{^cjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\p{cjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\p{^cjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\P{cjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\P{^cjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\p{ 	CJK_Unified_Ideographs_EXTENSION_B}', "");
Expect(0, 173791, '\p{^ 	CJK_Unified_Ideographs_EXTENSION_B}', "");
Expect(0, 173791, '\P{ 	CJK_Unified_Ideographs_EXTENSION_B}', "");
Expect(1, 173791, '\P{^ 	CJK_Unified_Ideographs_EXTENSION_B}', "");
Expect(0, 173792, '\p{ 	CJK_Unified_Ideographs_EXTENSION_B}', "");
Expect(1, 173792, '\p{^ 	CJK_Unified_Ideographs_EXTENSION_B}', "");
Expect(1, 173792, '\P{ 	CJK_Unified_Ideographs_EXTENSION_B}', "");
Expect(0, 173792, '\P{^ 	CJK_Unified_Ideographs_EXTENSION_B}', "");
Error('\p{-_IS_CJK_Unified_Ideographs_extension_B:=}');
Error('\P{-_IS_CJK_Unified_Ideographs_extension_B:=}');
Expect(1, 173791, '\p{iscjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\p{^iscjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\P{iscjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\P{^iscjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\p{iscjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\p{^iscjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\P{iscjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\P{^iscjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\p{ 	IS_CJK_Unified_Ideographs_Extension_B}', "");
Expect(0, 173791, '\p{^ 	IS_CJK_Unified_Ideographs_Extension_B}', "");
Expect(0, 173791, '\P{ 	IS_CJK_Unified_Ideographs_Extension_B}', "");
Expect(1, 173791, '\P{^ 	IS_CJK_Unified_Ideographs_Extension_B}', "");
Expect(0, 173792, '\p{ 	IS_CJK_Unified_Ideographs_Extension_B}', "");
Expect(1, 173792, '\p{^ 	IS_CJK_Unified_Ideographs_Extension_B}', "");
Expect(1, 173792, '\P{ 	IS_CJK_Unified_Ideographs_Extension_B}', "");
Expect(0, 173792, '\P{^ 	IS_CJK_Unified_Ideographs_Extension_B}', "");
Error('\p{	-In_CJK_unified_Ideographs_Extension_b:=}');
Error('\P{	-In_CJK_unified_Ideographs_Extension_b:=}');
Expect(1, 173791, '\p{incjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\p{^incjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\P{incjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\P{^incjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\p{incjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\p{^incjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\P{incjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\P{^incjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\p{		IN_cjk_Unified_Ideographs_Extension_B}', "");
Expect(0, 173791, '\p{^		IN_cjk_Unified_Ideographs_Extension_B}', "");
Expect(0, 173791, '\P{		IN_cjk_Unified_Ideographs_Extension_B}', "");
Expect(1, 173791, '\P{^		IN_cjk_Unified_Ideographs_Extension_B}', "");
Expect(0, 173792, '\p{		IN_cjk_Unified_Ideographs_Extension_B}', "");
Expect(1, 173792, '\p{^		IN_cjk_Unified_Ideographs_Extension_B}', "");
Expect(1, 173792, '\P{		IN_cjk_Unified_Ideographs_Extension_B}', "");
Expect(0, 173792, '\P{^		IN_cjk_Unified_Ideographs_Extension_B}', "");
Error('\p{ _cjk_Ext_B/a/}');
Error('\P{ _cjk_Ext_B/a/}');
Expect(1, 173791, '\p{cjkextb}', "");
Expect(0, 173791, '\p{^cjkextb}', "");
Expect(0, 173791, '\P{cjkextb}', "");
Expect(1, 173791, '\P{^cjkextb}', "");
Expect(0, 173792, '\p{cjkextb}', "");
Expect(1, 173792, '\p{^cjkextb}', "");
Expect(1, 173792, '\P{cjkextb}', "");
Expect(0, 173792, '\P{^cjkextb}', "");
Expect(1, 173791, '\p{CJK_ext_B}', "");
Expect(0, 173791, '\p{^CJK_ext_B}', "");
Expect(0, 173791, '\P{CJK_ext_B}', "");
Expect(1, 173791, '\P{^CJK_ext_B}', "");
Expect(0, 173792, '\p{CJK_ext_B}', "");
Expect(1, 173792, '\p{^CJK_ext_B}', "");
Expect(1, 173792, '\P{CJK_ext_B}', "");
Expect(0, 173792, '\P{^CJK_ext_B}', "");
Error('\p{Is_CJK_ext_B:=}');
Error('\P{Is_CJK_ext_B:=}');
Expect(1, 173791, '\p{iscjkextb}', "");
Expect(0, 173791, '\p{^iscjkextb}', "");
Expect(0, 173791, '\P{iscjkextb}', "");
Expect(1, 173791, '\P{^iscjkextb}', "");
Expect(0, 173792, '\p{iscjkextb}', "");
Expect(1, 173792, '\p{^iscjkextb}', "");
Expect(1, 173792, '\P{iscjkextb}', "");
Expect(0, 173792, '\P{^iscjkextb}', "");
Expect(1, 173791, '\p{-IS_CJK_Ext_b}', "");
Expect(0, 173791, '\p{^-IS_CJK_Ext_b}', "");
Expect(0, 173791, '\P{-IS_CJK_Ext_b}', "");
Expect(1, 173791, '\P{^-IS_CJK_Ext_b}', "");
Expect(0, 173792, '\p{-IS_CJK_Ext_b}', "");
Expect(1, 173792, '\p{^-IS_CJK_Ext_b}', "");
Expect(1, 173792, '\P{-IS_CJK_Ext_b}', "");
Expect(0, 173792, '\P{^-IS_CJK_Ext_b}', "");
Error('\p{in_CJK_EXT_B:=}');
Error('\P{in_CJK_EXT_B:=}');
Expect(1, 173791, '\p{incjkextb}', "");
Expect(0, 173791, '\p{^incjkextb}', "");
Expect(0, 173791, '\P{incjkextb}', "");
Expect(1, 173791, '\P{^incjkextb}', "");
Expect(0, 173792, '\p{incjkextb}', "");
Expect(1, 173792, '\p{^incjkextb}', "");
Expect(1, 173792, '\P{incjkextb}', "");
Expect(0, 173792, '\P{^incjkextb}', "");
Expect(1, 173791, '\p{-in_CJK_EXT_B}', "");
Expect(0, 173791, '\p{^-in_CJK_EXT_B}', "");
Expect(0, 173791, '\P{-in_CJK_EXT_B}', "");
Expect(1, 173791, '\P{^-in_CJK_EXT_B}', "");
Expect(0, 173792, '\p{-in_CJK_EXT_B}', "");
Expect(1, 173792, '\p{^-in_CJK_EXT_B}', "");
Expect(1, 173792, '\P{-in_CJK_EXT_B}', "");
Expect(0, 173792, '\P{^-in_CJK_EXT_B}', "");
Error('\p{_/a/Surrogate}');
Error('\P{_/a/Surrogate}');
Expect(1, 57343, '\p{surrogate}', "");
Expect(0, 57343, '\p{^surrogate}', "");
Expect(0, 57343, '\P{surrogate}', "");
Expect(1, 57343, '\P{^surrogate}', "");
Expect(0, 57344, '\p{surrogate}', "");
Expect(1, 57344, '\p{^surrogate}', "");
Expect(1, 57344, '\P{surrogate}', "");
Expect(0, 57344, '\P{^surrogate}', "");
Expect(1, 57343, '\p{	_Surrogate}', "");
Expect(0, 57343, '\p{^	_Surrogate}', "");
Expect(0, 57343, '\P{	_Surrogate}', "");
Expect(1, 57343, '\P{^	_Surrogate}', "");
Expect(0, 57344, '\p{	_Surrogate}', "");
Expect(1, 57344, '\p{^	_Surrogate}', "");
Expect(1, 57344, '\P{	_Surrogate}', "");
Expect(0, 57344, '\P{^	_Surrogate}', "");
Error('\p{/a/-Is_Surrogate}');
Error('\P{/a/-Is_Surrogate}');
Expect(1, 57343, '\p{issurrogate}', "");
Expect(0, 57343, '\p{^issurrogate}', "");
Expect(0, 57343, '\P{issurrogate}', "");
Expect(1, 57343, '\P{^issurrogate}', "");
Expect(0, 57344, '\p{issurrogate}', "");
Expect(1, 57344, '\p{^issurrogate}', "");
Expect(1, 57344, '\P{issurrogate}', "");
Expect(0, 57344, '\P{^issurrogate}', "");
Expect(1, 57343, '\p{-is_surrogate}', "");
Expect(0, 57343, '\p{^-is_surrogate}', "");
Expect(0, 57343, '\P{-is_surrogate}', "");
Expect(1, 57343, '\P{^-is_surrogate}', "");
Expect(0, 57344, '\p{-is_surrogate}', "");
Expect(1, 57344, '\p{^-is_surrogate}', "");
Expect(1, 57344, '\P{-is_surrogate}', "");
Expect(0, 57344, '\P{^-is_surrogate}', "");
Error('\p{ 	Cs/a/}');
Error('\P{ 	Cs/a/}');
Expect(1, 57343, '\p{cs}', "");
Expect(0, 57343, '\p{^cs}', "");
Expect(0, 57343, '\P{cs}', "");
Expect(1, 57343, '\P{^cs}', "");
Expect(0, 57344, '\p{cs}', "");
Expect(1, 57344, '\p{^cs}', "");
Expect(1, 57344, '\P{cs}', "");
Expect(0, 57344, '\P{^cs}', "");
Expect(1, 57343, '\p{--cs}', "");
Expect(0, 57343, '\p{^--cs}', "");
Expect(0, 57343, '\P{--cs}', "");
Expect(1, 57343, '\P{^--cs}', "");
Expect(0, 57344, '\p{--cs}', "");
Expect(1, 57344, '\p{^--cs}', "");
Expect(1, 57344, '\P{--cs}', "");
Expect(0, 57344, '\P{^--cs}', "");
Error('\p{/a/_	Is_Cs}');
Error('\P{/a/_	Is_Cs}');
Expect(1, 57343, '\p{iscs}', "");
Expect(0, 57343, '\p{^iscs}', "");
Expect(0, 57343, '\P{iscs}', "");
Expect(1, 57343, '\P{^iscs}', "");
Expect(0, 57344, '\p{iscs}', "");
Expect(1, 57344, '\p{^iscs}', "");
Expect(1, 57344, '\P{iscs}', "");
Expect(0, 57344, '\P{^iscs}', "");
Expect(1, 57343, '\p{_IS_CS}', "");
Expect(0, 57343, '\p{^_IS_CS}', "");
Expect(0, 57343, '\P{_IS_CS}', "");
Expect(1, 57343, '\P{^_IS_CS}', "");
Expect(0, 57344, '\p{_IS_CS}', "");
Expect(1, 57344, '\p{^_IS_CS}', "");
Expect(1, 57344, '\P{_IS_CS}', "");
Expect(0, 57344, '\P{^_IS_CS}', "");
Error('\p{ :=in_coptic}');
Error('\P{ :=in_coptic}');
Expect(1, 11519, '\p{incoptic}', "");
Expect(0, 11519, '\p{^incoptic}', "");
Expect(0, 11519, '\P{incoptic}', "");
Expect(1, 11519, '\P{^incoptic}', "");
Expect(0, 11520, '\p{incoptic}', "");
Expect(1, 11520, '\p{^incoptic}', "");
Expect(1, 11520, '\P{incoptic}', "");
Expect(0, 11520, '\P{^incoptic}', "");
Expect(1, 11519, '\p{_-In_Coptic}', "");
Expect(0, 11519, '\p{^_-In_Coptic}', "");
Expect(0, 11519, '\P{_-In_Coptic}', "");
Expect(1, 11519, '\P{^_-In_Coptic}', "");
Expect(0, 11520, '\p{_-In_Coptic}', "");
Expect(1, 11520, '\p{^_-In_Coptic}', "");
Expect(1, 11520, '\P{_-In_Coptic}', "");
Expect(0, 11520, '\P{^_-In_Coptic}', "");
Error('\p{ :=in_coptic}');
Error('\P{ :=in_coptic}');
Expect(1, 11519, '\p{-_In_Coptic}', "");
Expect(0, 11519, '\p{^-_In_Coptic}', "");
Expect(0, 11519, '\P{-_In_Coptic}', "");
Expect(1, 11519, '\P{^-_In_Coptic}', "");
Expect(0, 11520, '\p{-_In_Coptic}', "");
Expect(1, 11520, '\p{^-_In_Coptic}', "");
Expect(1, 11520, '\P{-_In_Coptic}', "");
Expect(0, 11520, '\P{^-_In_Coptic}', "");
Error('\p{/a/Khmer}');
Error('\P{/a/Khmer}');
Expect(1, 6655, '\p{khmer}', "");
Expect(0, 6655, '\p{^khmer}', "");
Expect(0, 6655, '\P{khmer}', "");
Expect(1, 6655, '\P{^khmer}', "");
Expect(0, 6656, '\p{khmer}', "");
Expect(1, 6656, '\p{^khmer}', "");
Expect(1, 6656, '\P{khmer}', "");
Expect(0, 6656, '\P{^khmer}', "");
Expect(1, 6655, '\p{-KHMER}', "");
Expect(0, 6655, '\p{^-KHMER}', "");
Expect(0, 6655, '\P{-KHMER}', "");
Expect(1, 6655, '\P{^-KHMER}', "");
Expect(0, 6656, '\p{-KHMER}', "");
Expect(1, 6656, '\p{^-KHMER}', "");
Expect(1, 6656, '\P{-KHMER}', "");
Expect(0, 6656, '\P{^-KHMER}', "");
Error('\p{	/a/Is_Khmer}');
Error('\P{	/a/Is_Khmer}');
Expect(1, 6655, '\p{iskhmer}', "");
Expect(0, 6655, '\p{^iskhmer}', "");
Expect(0, 6655, '\P{iskhmer}', "");
Expect(1, 6655, '\P{^iskhmer}', "");
Expect(0, 6656, '\p{iskhmer}', "");
Expect(1, 6656, '\p{^iskhmer}', "");
Expect(1, 6656, '\P{iskhmer}', "");
Expect(0, 6656, '\P{^iskhmer}', "");
Expect(1, 6655, '\p{	-IS_Khmer}', "");
Expect(0, 6655, '\p{^	-IS_Khmer}', "");
Expect(0, 6655, '\P{	-IS_Khmer}', "");
Expect(1, 6655, '\P{^	-IS_Khmer}', "");
Expect(0, 6656, '\p{	-IS_Khmer}', "");
Expect(1, 6656, '\p{^	-IS_Khmer}', "");
Expect(1, 6656, '\P{	-IS_Khmer}', "");
Expect(0, 6656, '\P{^	-IS_Khmer}', "");
Error('\p{-:=Khmr}');
Error('\P{-:=Khmr}');
Expect(1, 6655, '\p{khmr}', "");
Expect(0, 6655, '\p{^khmr}', "");
Expect(0, 6655, '\P{khmr}', "");
Expect(1, 6655, '\P{^khmr}', "");
Expect(0, 6656, '\p{khmr}', "");
Expect(1, 6656, '\p{^khmr}', "");
Expect(1, 6656, '\P{khmr}', "");
Expect(0, 6656, '\P{^khmr}', "");
Expect(1, 6655, '\p{  Khmr}', "");
Expect(0, 6655, '\p{^  Khmr}', "");
Expect(0, 6655, '\P{  Khmr}', "");
Expect(1, 6655, '\P{^  Khmr}', "");
Expect(0, 6656, '\p{  Khmr}', "");
Expect(1, 6656, '\p{^  Khmr}', "");
Expect(1, 6656, '\P{  Khmr}', "");
Expect(0, 6656, '\P{^  Khmr}', "");
Error('\p{ /a/is_Khmr}');
Error('\P{ /a/is_Khmr}');
Expect(1, 6655, '\p{iskhmr}', "");
Expect(0, 6655, '\p{^iskhmr}', "");
Expect(0, 6655, '\P{iskhmr}', "");
Expect(1, 6655, '\P{^iskhmr}', "");
Expect(0, 6656, '\p{iskhmr}', "");
Expect(1, 6656, '\p{^iskhmr}', "");
Expect(1, 6656, '\P{iskhmr}', "");
Expect(0, 6656, '\P{^iskhmr}', "");
Expect(1, 6655, '\p{	-IS_khmr}', "");
Expect(0, 6655, '\p{^	-IS_khmr}', "");
Expect(0, 6655, '\P{	-IS_khmr}', "");
Expect(1, 6655, '\P{^	-IS_khmr}', "");
Expect(0, 6656, '\p{	-IS_khmr}', "");
Expect(1, 6656, '\p{^	-IS_khmr}', "");
Expect(1, 6656, '\P{	-IS_khmr}', "");
Expect(0, 6656, '\P{^	-IS_khmr}', "");
Error('\p{- In_NKo/a/}');
Error('\P{- In_NKo/a/}');
Expect(1, 2047, '\p{innko}', "");
Expect(0, 2047, '\p{^innko}', "");
Expect(0, 2047, '\P{innko}', "");
Expect(1, 2047, '\P{^innko}', "");
Expect(0, 2048, '\p{innko}', "");
Expect(1, 2048, '\p{^innko}', "");
Expect(1, 2048, '\P{innko}', "");
Expect(0, 2048, '\P{^innko}', "");
Expect(1, 2047, '\p{- In_NKo}', "");
Expect(0, 2047, '\p{^- In_NKo}', "");
Expect(0, 2047, '\P{- In_NKo}', "");
Expect(1, 2047, '\P{^- In_NKo}', "");
Expect(0, 2048, '\p{- In_NKo}', "");
Expect(1, 2048, '\p{^- In_NKo}', "");
Expect(1, 2048, '\P{- In_NKo}', "");
Expect(0, 2048, '\P{^- In_NKo}', "");
Error('\p{/a/_-in_nko}');
Error('\P{/a/_-in_nko}');
Expect(1, 2047, '\p{__in_nko}', "");
Expect(0, 2047, '\p{^__in_nko}', "");
Expect(0, 2047, '\P{__in_nko}', "");
Expect(1, 2047, '\P{^__in_nko}', "");
Expect(0, 2048, '\p{__in_nko}', "");
Expect(1, 2048, '\p{^__in_nko}', "");
Expect(1, 2048, '\P{__in_nko}', "");
Expect(0, 2048, '\P{^__in_nko}', "");
Error('\p{-Linear_b_syllabary:=}');
Error('\P{-Linear_b_syllabary:=}');
Expect(1, 65663, '\p{linearbsyllabary}', "");
Expect(0, 65663, '\p{^linearbsyllabary}', "");
Expect(0, 65663, '\P{linearbsyllabary}', "");
Expect(1, 65663, '\P{^linearbsyllabary}', "");
Expect(0, 65664, '\p{linearbsyllabary}', "");
Expect(1, 65664, '\p{^linearbsyllabary}', "");
Expect(1, 65664, '\P{linearbsyllabary}', "");
Expect(0, 65664, '\P{^linearbsyllabary}', "");
Expect(1, 65663, '\p{ _Linear_b_SYLLABARY}', "");
Expect(0, 65663, '\p{^ _Linear_b_SYLLABARY}', "");
Expect(0, 65663, '\P{ _Linear_b_SYLLABARY}', "");
Expect(1, 65663, '\P{^ _Linear_b_SYLLABARY}', "");
Expect(0, 65664, '\p{ _Linear_b_SYLLABARY}', "");
Expect(1, 65664, '\p{^ _Linear_b_SYLLABARY}', "");
Expect(1, 65664, '\P{ _Linear_b_SYLLABARY}', "");
Expect(0, 65664, '\P{^ _Linear_b_SYLLABARY}', "");
Error('\p{ IS_Linear_B_Syllabary:=}');
Error('\P{ IS_Linear_B_Syllabary:=}');
Expect(1, 65663, '\p{islinearbsyllabary}', "");
Expect(0, 65663, '\p{^islinearbsyllabary}', "");
Expect(0, 65663, '\P{islinearbsyllabary}', "");
Expect(1, 65663, '\P{^islinearbsyllabary}', "");
Expect(0, 65664, '\p{islinearbsyllabary}', "");
Expect(1, 65664, '\p{^islinearbsyllabary}', "");
Expect(1, 65664, '\P{islinearbsyllabary}', "");
Expect(0, 65664, '\P{^islinearbsyllabary}', "");
Expect(1, 65663, '\p{Is_linear_B_SYLLABARY}', "");
Expect(0, 65663, '\p{^Is_linear_B_SYLLABARY}', "");
Expect(0, 65663, '\P{Is_linear_B_SYLLABARY}', "");
Expect(1, 65663, '\P{^Is_linear_B_SYLLABARY}', "");
Expect(0, 65664, '\p{Is_linear_B_SYLLABARY}', "");
Expect(1, 65664, '\p{^Is_linear_B_SYLLABARY}', "");
Expect(1, 65664, '\P{Is_linear_B_SYLLABARY}', "");
Expect(0, 65664, '\P{^Is_linear_B_SYLLABARY}', "");
Error('\p{/a/	in_LINEAR_b_syllabary}');
Error('\P{/a/	in_LINEAR_b_syllabary}');
Expect(1, 65663, '\p{inlinearbsyllabary}', "");
Expect(0, 65663, '\p{^inlinearbsyllabary}', "");
Expect(0, 65663, '\P{inlinearbsyllabary}', "");
Expect(1, 65663, '\P{^inlinearbsyllabary}', "");
Expect(0, 65664, '\p{inlinearbsyllabary}', "");
Expect(1, 65664, '\p{^inlinearbsyllabary}', "");
Expect(1, 65664, '\P{inlinearbsyllabary}', "");
Expect(0, 65664, '\P{^inlinearbsyllabary}', "");
Expect(1, 65663, '\p{-In_Linear_B_Syllabary}', "");
Expect(0, 65663, '\p{^-In_Linear_B_Syllabary}', "");
Expect(0, 65663, '\P{-In_Linear_B_Syllabary}', "");
Expect(1, 65663, '\P{^-In_Linear_B_Syllabary}', "");
Expect(0, 65664, '\p{-In_Linear_B_Syllabary}', "");
Expect(1, 65664, '\p{^-In_Linear_B_Syllabary}', "");
Expect(1, 65664, '\P{-In_Linear_B_Syllabary}', "");
Expect(0, 65664, '\P{^-In_Linear_B_Syllabary}', "");
Error('\p{ /a/HANGUL_Jamo_Extended_B}');
Error('\P{ /a/HANGUL_Jamo_Extended_B}');
Expect(1, 55295, '\p{hanguljamoextendedb}', "");
Expect(0, 55295, '\p{^hanguljamoextendedb}', "");
Expect(0, 55295, '\P{hanguljamoextendedb}', "");
Expect(1, 55295, '\P{^hanguljamoextendedb}', "");
Expect(0, 57344, '\p{hanguljamoextendedb}', "");
Expect(1, 57344, '\p{^hanguljamoextendedb}', "");
Expect(1, 57344, '\P{hanguljamoextendedb}', "");
Expect(0, 57344, '\P{^hanguljamoextendedb}', "");
Expect(1, 55295, '\p{Hangul_Jamo_Extended_b}', "");
Expect(0, 55295, '\p{^Hangul_Jamo_Extended_b}', "");
Expect(0, 55295, '\P{Hangul_Jamo_Extended_b}', "");
Expect(1, 55295, '\P{^Hangul_Jamo_Extended_b}', "");
Expect(0, 57344, '\p{Hangul_Jamo_Extended_b}', "");
Expect(1, 57344, '\p{^Hangul_Jamo_Extended_b}', "");
Expect(1, 57344, '\P{Hangul_Jamo_Extended_b}', "");
Expect(0, 57344, '\P{^Hangul_Jamo_Extended_b}', "");
Error('\p{/a/_-IS_Hangul_JAMO_Extended_B}');
Error('\P{/a/_-IS_Hangul_JAMO_Extended_B}');
Expect(1, 55295, '\p{ishanguljamoextendedb}', "");
Expect(0, 55295, '\p{^ishanguljamoextendedb}', "");
Expect(0, 55295, '\P{ishanguljamoextendedb}', "");
Expect(1, 55295, '\P{^ishanguljamoextendedb}', "");
Expect(0, 57344, '\p{ishanguljamoextendedb}', "");
Expect(1, 57344, '\p{^ishanguljamoextendedb}', "");
Expect(1, 57344, '\P{ishanguljamoextendedb}', "");
Expect(0, 57344, '\P{^ishanguljamoextendedb}', "");
Expect(1, 55295, '\p{- IS_hangul_jamo_extended_B}', "");
Expect(0, 55295, '\p{^- IS_hangul_jamo_extended_B}', "");
Expect(0, 55295, '\P{- IS_hangul_jamo_extended_B}', "");
Expect(1, 55295, '\P{^- IS_hangul_jamo_extended_B}', "");
Expect(0, 57344, '\p{- IS_hangul_jamo_extended_B}', "");
Expect(1, 57344, '\p{^- IS_hangul_jamo_extended_B}', "");
Expect(1, 57344, '\P{- IS_hangul_jamo_extended_B}', "");
Expect(0, 57344, '\P{^- IS_hangul_jamo_extended_B}', "");
Error('\p{  In_Hangul_Jamo_EXTENDED_b/a/}');
Error('\P{  In_Hangul_Jamo_EXTENDED_b/a/}');
Expect(1, 55295, '\p{inhanguljamoextendedb}', "");
Expect(0, 55295, '\p{^inhanguljamoextendedb}', "");
Expect(0, 55295, '\P{inhanguljamoextendedb}', "");
Expect(1, 55295, '\P{^inhanguljamoextendedb}', "");
Expect(0, 57344, '\p{inhanguljamoextendedb}', "");
Expect(1, 57344, '\p{^inhanguljamoextendedb}', "");
Expect(1, 57344, '\P{inhanguljamoextendedb}', "");
Expect(0, 57344, '\P{^inhanguljamoextendedb}', "");
Expect(1, 55295, '\p{ -In_Hangul_Jamo_Extended_B}', "");
Expect(0, 55295, '\p{^ -In_Hangul_Jamo_Extended_B}', "");
Expect(0, 55295, '\P{ -In_Hangul_Jamo_Extended_B}', "");
Expect(1, 55295, '\P{^ -In_Hangul_Jamo_Extended_B}', "");
Expect(0, 57344, '\p{ -In_Hangul_Jamo_Extended_B}', "");
Expect(1, 57344, '\p{^ -In_Hangul_Jamo_Extended_B}', "");
Expect(1, 57344, '\P{ -In_Hangul_Jamo_Extended_B}', "");
Expect(0, 57344, '\P{^ -In_Hangul_Jamo_Extended_B}', "");
Error('\p{_:=Jamo_Ext_B}');
Error('\P{_:=Jamo_Ext_B}');
Expect(1, 55295, '\p{jamoextb}', "");
Expect(0, 55295, '\p{^jamoextb}', "");
Expect(0, 55295, '\P{jamoextb}', "");
Expect(1, 55295, '\P{^jamoextb}', "");
Expect(0, 57344, '\p{jamoextb}', "");
Expect(1, 57344, '\p{^jamoextb}', "");
Expect(1, 57344, '\P{jamoextb}', "");
Expect(0, 57344, '\P{^jamoextb}', "");
Expect(1, 55295, '\p{- jamo_ext_B}', "");
Expect(0, 55295, '\p{^- jamo_ext_B}', "");
Expect(0, 55295, '\P{- jamo_ext_B}', "");
Expect(1, 55295, '\P{^- jamo_ext_B}', "");
Expect(0, 57344, '\p{- jamo_ext_B}', "");
Expect(1, 57344, '\p{^- jamo_ext_B}', "");
Expect(1, 57344, '\P{- jamo_ext_B}', "");
Expect(0, 57344, '\P{^- jamo_ext_B}', "");
Error('\p{/a/	 is_JAMO_Ext_B}');
Error('\P{/a/	 is_JAMO_Ext_B}');
Expect(1, 55295, '\p{isjamoextb}', "");
Expect(0, 55295, '\p{^isjamoextb}', "");
Expect(0, 55295, '\P{isjamoextb}', "");
Expect(1, 55295, '\P{^isjamoextb}', "");
Expect(0, 57344, '\p{isjamoextb}', "");
Expect(1, 57344, '\p{^isjamoextb}', "");
Expect(1, 57344, '\P{isjamoextb}', "");
Expect(0, 57344, '\P{^isjamoextb}', "");
Expect(1, 55295, '\p{-_Is_jamo_Ext_b}', "");
Expect(0, 55295, '\p{^-_Is_jamo_Ext_b}', "");
Expect(0, 55295, '\P{-_Is_jamo_Ext_b}', "");
Expect(1, 55295, '\P{^-_Is_jamo_Ext_b}', "");
Expect(0, 57344, '\p{-_Is_jamo_Ext_b}', "");
Expect(1, 57344, '\p{^-_Is_jamo_Ext_b}', "");
Expect(1, 57344, '\P{-_Is_jamo_Ext_b}', "");
Expect(0, 57344, '\P{^-_Is_jamo_Ext_b}', "");
Error('\p{-:=In_JAMO_Ext_B}');
Error('\P{-:=In_JAMO_Ext_B}');
Expect(1, 55295, '\p{injamoextb}', "");
Expect(0, 55295, '\p{^injamoextb}', "");
Expect(0, 55295, '\P{injamoextb}', "");
Expect(1, 55295, '\P{^injamoextb}', "");
Expect(0, 57344, '\p{injamoextb}', "");
Expect(1, 57344, '\p{^injamoextb}', "");
Expect(1, 57344, '\P{injamoextb}', "");
Expect(0, 57344, '\P{^injamoextb}', "");
Expect(1, 55295, '\p{	_In_jamo_Ext_B}', "");
Expect(0, 55295, '\p{^	_In_jamo_Ext_B}', "");
Expect(0, 55295, '\P{	_In_jamo_Ext_B}', "");
Expect(1, 55295, '\P{^	_In_jamo_Ext_B}', "");
Expect(0, 57344, '\p{	_In_jamo_Ext_B}', "");
Expect(1, 57344, '\p{^	_In_jamo_Ext_B}', "");
Expect(1, 57344, '\P{	_In_jamo_Ext_B}', "");
Expect(0, 57344, '\P{^	_In_jamo_Ext_B}', "");
Error('\p{/a/ 	in_BUGINESE}');
Error('\P{/a/ 	in_BUGINESE}');
Expect(1, 6687, '\p{inbuginese}', "");
Expect(0, 6687, '\p{^inbuginese}', "");
Expect(0, 6687, '\P{inbuginese}', "");
Expect(1, 6687, '\P{^inbuginese}', "");
Expect(0, 6688, '\p{inbuginese}', "");
Expect(1, 6688, '\p{^inbuginese}', "");
Expect(1, 6688, '\P{inbuginese}', "");
Expect(0, 6688, '\P{^inbuginese}', "");
Expect(1, 6687, '\p{ In_buginese}', "");
Expect(0, 6687, '\p{^ In_buginese}', "");
Expect(0, 6687, '\P{ In_buginese}', "");
Expect(1, 6687, '\P{^ In_buginese}', "");
Expect(0, 6688, '\p{ In_buginese}', "");
Expect(1, 6688, '\p{^ In_buginese}', "");
Expect(1, 6688, '\P{ In_buginese}', "");
Expect(0, 6688, '\P{^ In_buginese}', "");
Error('\p{:=_-IN_buginese}');
Error('\P{:=_-IN_buginese}');
Expect(1, 6687, '\p{-	In_Buginese}', "");
Expect(0, 6687, '\p{^-	In_Buginese}', "");
Expect(0, 6687, '\P{-	In_Buginese}', "");
Expect(1, 6687, '\P{^-	In_Buginese}', "");
Expect(0, 6688, '\p{-	In_Buginese}', "");
Expect(1, 6688, '\p{^-	In_Buginese}', "");
Expect(1, 6688, '\P{-	In_Buginese}', "");
Expect(0, 6688, '\P{^-	In_Buginese}', "");
Error('\p{_	IN_Tagbanwa:=}');
Error('\P{_	IN_Tagbanwa:=}');
Expect(1, 6015, '\p{intagbanwa}', "");
Expect(0, 6015, '\p{^intagbanwa}', "");
Expect(0, 6015, '\P{intagbanwa}', "");
Expect(1, 6015, '\P{^intagbanwa}', "");
Expect(0, 6016, '\p{intagbanwa}', "");
Expect(1, 6016, '\p{^intagbanwa}', "");
Expect(1, 6016, '\P{intagbanwa}', "");
Expect(0, 6016, '\P{^intagbanwa}', "");
Expect(1, 6015, '\p{- In_tagbanwa}', "");
Expect(0, 6015, '\p{^- In_tagbanwa}', "");
Expect(0, 6015, '\P{- In_tagbanwa}', "");
Expect(1, 6015, '\P{^- In_tagbanwa}', "");
Expect(0, 6016, '\p{- In_tagbanwa}', "");
Expect(1, 6016, '\p{^- In_tagbanwa}', "");
Expect(1, 6016, '\P{- In_tagbanwa}', "");
Expect(0, 6016, '\P{^- In_tagbanwa}', "");
Error('\p{:=_-In_Tagbanwa}');
Error('\P{:=_-In_Tagbanwa}');
Expect(1, 6015, '\p{-In_TAGBANWA}', "");
Expect(0, 6015, '\p{^-In_TAGBANWA}', "");
Expect(0, 6015, '\P{-In_TAGBANWA}', "");
Expect(1, 6015, '\P{^-In_TAGBANWA}', "");
Expect(0, 6016, '\p{-In_TAGBANWA}', "");
Expect(1, 6016, '\p{^-In_TAGBANWA}', "");
Expect(1, 6016, '\P{-In_TAGBANWA}', "");
Expect(0, 6016, '\P{^-In_TAGBANWA}', "");
Error('\p{		XID_Start/a/}');
Error('\P{		XID_Start/a/}');
Expect(1, 195101, '\p{xidstart}', "");
Expect(0, 195101, '\p{^xidstart}', "");
Expect(0, 195101, '\P{xidstart}', "");
Expect(1, 195101, '\P{^xidstart}', "");
Expect(0, 195102, '\p{xidstart}', "");
Expect(1, 195102, '\p{^xidstart}', "");
Expect(1, 195102, '\P{xidstart}', "");
Expect(0, 195102, '\P{^xidstart}', "");
Expect(1, 195101, '\p{  XID_start}', "");
Expect(0, 195101, '\p{^  XID_start}', "");
Expect(0, 195101, '\P{  XID_start}', "");
Expect(1, 195101, '\P{^  XID_start}', "");
Expect(0, 195102, '\p{  XID_start}', "");
Expect(1, 195102, '\p{^  XID_start}', "");
Expect(1, 195102, '\P{  XID_start}', "");
Expect(0, 195102, '\P{^  XID_start}', "");
Error('\p{:=	-IS_XID_start}');
Error('\P{:=	-IS_XID_start}');
Expect(1, 195101, '\p{isxidstart}', "");
Expect(0, 195101, '\p{^isxidstart}', "");
Expect(0, 195101, '\P{isxidstart}', "");
Expect(1, 195101, '\P{^isxidstart}', "");
Expect(0, 195102, '\p{isxidstart}', "");
Expect(1, 195102, '\p{^isxidstart}', "");
Expect(1, 195102, '\P{isxidstart}', "");
Expect(0, 195102, '\P{^isxidstart}', "");
Expect(1, 195101, '\p{- Is_XID_START}', "");
Expect(0, 195101, '\p{^- Is_XID_START}', "");
Expect(0, 195101, '\P{- Is_XID_START}', "");
Expect(1, 195101, '\P{^- Is_XID_START}', "");
Expect(0, 195102, '\p{- Is_XID_START}', "");
Expect(1, 195102, '\p{^- Is_XID_START}', "");
Expect(1, 195102, '\P{- Is_XID_START}', "");
Expect(0, 195102, '\P{^- Is_XID_START}', "");
Error('\p{		XIDS/a/}');
Error('\P{		XIDS/a/}');
Expect(1, 195101, '\p{xids}', "");
Expect(0, 195101, '\p{^xids}', "");
Expect(0, 195101, '\P{xids}', "");
Expect(1, 195101, '\P{^xids}', "");
Expect(0, 195102, '\p{xids}', "");
Expect(1, 195102, '\p{^xids}', "");
Expect(1, 195102, '\P{xids}', "");
Expect(0, 195102, '\P{^xids}', "");
Expect(1, 195101, '\p{	XIDS}', "");
Expect(0, 195101, '\p{^	XIDS}', "");
Expect(0, 195101, '\P{	XIDS}', "");
Expect(1, 195101, '\P{^	XIDS}', "");
Expect(0, 195102, '\p{	XIDS}', "");
Expect(1, 195102, '\p{^	XIDS}', "");
Expect(1, 195102, '\P{	XIDS}', "");
Expect(0, 195102, '\P{^	XIDS}', "");
Error('\p{	/a/Is_XIDS}');
Error('\P{	/a/Is_XIDS}');
Expect(1, 195101, '\p{isxids}', "");
Expect(0, 195101, '\p{^isxids}', "");
Expect(0, 195101, '\P{isxids}', "");
Expect(1, 195101, '\P{^isxids}', "");
Expect(0, 195102, '\p{isxids}', "");
Expect(1, 195102, '\p{^isxids}', "");
Expect(1, 195102, '\P{isxids}', "");
Expect(0, 195102, '\P{^isxids}', "");
Expect(1, 195101, '\p{	_Is_xids}', "");
Expect(0, 195101, '\p{^	_Is_xids}', "");
Expect(0, 195101, '\P{	_Is_xids}', "");
Expect(1, 195101, '\P{^	_Is_xids}', "");
Expect(0, 195102, '\p{	_Is_xids}', "");
Expect(1, 195102, '\p{^	_Is_xids}', "");
Expect(1, 195102, '\P{	_Is_xids}', "");
Expect(0, 195102, '\P{^	_Is_xids}', "");
Error('\p{/a/- modifier_Tone_Letters}');
Error('\P{/a/- modifier_Tone_Letters}');
Expect(1, 42783, '\p{modifiertoneletters}', "");
Expect(0, 42783, '\p{^modifiertoneletters}', "");
Expect(0, 42783, '\P{modifiertoneletters}', "");
Expect(1, 42783, '\P{^modifiertoneletters}', "");
Expect(0, 42784, '\p{modifiertoneletters}', "");
Expect(1, 42784, '\p{^modifiertoneletters}', "");
Expect(1, 42784, '\P{modifiertoneletters}', "");
Expect(0, 42784, '\P{^modifiertoneletters}', "");
Expect(1, 42783, '\p{	 MODIFIER_tone_Letters}', "");
Expect(0, 42783, '\p{^	 MODIFIER_tone_Letters}', "");
Expect(0, 42783, '\P{	 MODIFIER_tone_Letters}', "");
Expect(1, 42783, '\P{^	 MODIFIER_tone_Letters}', "");
Expect(0, 42784, '\p{	 MODIFIER_tone_Letters}', "");
Expect(1, 42784, '\p{^	 MODIFIER_tone_Letters}', "");
Expect(1, 42784, '\P{	 MODIFIER_tone_Letters}', "");
Expect(0, 42784, '\P{^	 MODIFIER_tone_Letters}', "");
Error('\p{-	is_MODIFIER_TONE_letters:=}');
Error('\P{-	is_MODIFIER_TONE_letters:=}');
Expect(1, 42783, '\p{ismodifiertoneletters}', "");
Expect(0, 42783, '\p{^ismodifiertoneletters}', "");
Expect(0, 42783, '\P{ismodifiertoneletters}', "");
Expect(1, 42783, '\P{^ismodifiertoneletters}', "");
Expect(0, 42784, '\p{ismodifiertoneletters}', "");
Expect(1, 42784, '\p{^ismodifiertoneletters}', "");
Expect(1, 42784, '\P{ismodifiertoneletters}', "");
Expect(0, 42784, '\P{^ismodifiertoneletters}', "");
Expect(1, 42783, '\p{ is_Modifier_Tone_Letters}', "");
Expect(0, 42783, '\p{^ is_Modifier_Tone_Letters}', "");
Expect(0, 42783, '\P{ is_Modifier_Tone_Letters}', "");
Expect(1, 42783, '\P{^ is_Modifier_Tone_Letters}', "");
Expect(0, 42784, '\p{ is_Modifier_Tone_Letters}', "");
Expect(1, 42784, '\p{^ is_Modifier_Tone_Letters}', "");
Expect(1, 42784, '\P{ is_Modifier_Tone_Letters}', "");
Expect(0, 42784, '\P{^ is_Modifier_Tone_Letters}', "");
Error('\p{:=--In_modifier_tone_Letters}');
Error('\P{:=--In_modifier_tone_Letters}');
Expect(1, 42783, '\p{inmodifiertoneletters}', "");
Expect(0, 42783, '\p{^inmodifiertoneletters}', "");
Expect(0, 42783, '\P{inmodifiertoneletters}', "");
Expect(1, 42783, '\P{^inmodifiertoneletters}', "");
Expect(0, 42784, '\p{inmodifiertoneletters}', "");
Expect(1, 42784, '\p{^inmodifiertoneletters}', "");
Expect(1, 42784, '\P{inmodifiertoneletters}', "");
Expect(0, 42784, '\P{^inmodifiertoneletters}', "");
Expect(1, 42783, '\p{ In_Modifier_Tone_letters}', "");
Expect(0, 42783, '\p{^ In_Modifier_Tone_letters}', "");
Expect(0, 42783, '\P{ In_Modifier_Tone_letters}', "");
Expect(1, 42783, '\P{^ In_Modifier_Tone_letters}', "");
Expect(0, 42784, '\p{ In_Modifier_Tone_letters}', "");
Expect(1, 42784, '\p{^ In_Modifier_Tone_letters}', "");
Expect(1, 42784, '\P{ In_Modifier_Tone_letters}', "");
Expect(0, 42784, '\P{^ In_Modifier_Tone_letters}', "");
Error('\p{/a/-Changes_when_Casefolded}');
Error('\P{/a/-Changes_when_Casefolded}');
Expect(1, 66599, '\p{changeswhencasefolded}', "");
Expect(0, 66599, '\p{^changeswhencasefolded}', "");
Expect(0, 66599, '\P{changeswhencasefolded}', "");
Expect(1, 66599, '\P{^changeswhencasefolded}', "");
Expect(0, 66600, '\p{changeswhencasefolded}', "");
Expect(1, 66600, '\p{^changeswhencasefolded}', "");
Expect(1, 66600, '\P{changeswhencasefolded}', "");
Expect(0, 66600, '\P{^changeswhencasefolded}', "");
Expect(1, 66599, '\p{ CHANGES_WHEN_Casefolded}', "");
Expect(0, 66599, '\p{^ CHANGES_WHEN_Casefolded}', "");
Expect(0, 66599, '\P{ CHANGES_WHEN_Casefolded}', "");
Expect(1, 66599, '\P{^ CHANGES_WHEN_Casefolded}', "");
Expect(0, 66600, '\p{ CHANGES_WHEN_Casefolded}', "");
Expect(1, 66600, '\p{^ CHANGES_WHEN_Casefolded}', "");
Expect(1, 66600, '\P{ CHANGES_WHEN_Casefolded}', "");
Expect(0, 66600, '\P{^ CHANGES_WHEN_Casefolded}', "");
Error('\p{	Is_Changes_When_casefolded:=}');
Error('\P{	Is_Changes_When_casefolded:=}');
Expect(1, 66599, '\p{ischangeswhencasefolded}', "");
Expect(0, 66599, '\p{^ischangeswhencasefolded}', "");
Expect(0, 66599, '\P{ischangeswhencasefolded}', "");
Expect(1, 66599, '\P{^ischangeswhencasefolded}', "");
Expect(0, 66600, '\p{ischangeswhencasefolded}', "");
Expect(1, 66600, '\p{^ischangeswhencasefolded}', "");
Expect(1, 66600, '\P{ischangeswhencasefolded}', "");
Expect(0, 66600, '\P{^ischangeswhencasefolded}', "");
Expect(1, 66599, '\p{ _Is_Changes_When_Casefolded}', "");
Expect(0, 66599, '\p{^ _Is_Changes_When_Casefolded}', "");
Expect(0, 66599, '\P{ _Is_Changes_When_Casefolded}', "");
Expect(1, 66599, '\P{^ _Is_Changes_When_Casefolded}', "");
Expect(0, 66600, '\p{ _Is_Changes_When_Casefolded}', "");
Expect(1, 66600, '\p{^ _Is_Changes_When_Casefolded}', "");
Expect(1, 66600, '\P{ _Is_Changes_When_Casefolded}', "");
Expect(0, 66600, '\P{^ _Is_Changes_When_Casefolded}', "");
Error('\p{ /a/CWCF}');
Error('\P{ /a/CWCF}');
Expect(1, 66599, '\p{cwcf}', "");
Expect(0, 66599, '\p{^cwcf}', "");
Expect(0, 66599, '\P{cwcf}', "");
Expect(1, 66599, '\P{^cwcf}', "");
Expect(0, 66600, '\p{cwcf}', "");
Expect(1, 66600, '\p{^cwcf}', "");
Expect(1, 66600, '\P{cwcf}', "");
Expect(0, 66600, '\P{^cwcf}', "");
Expect(1, 66599, '\p{-CWCF}', "");
Expect(0, 66599, '\p{^-CWCF}', "");
Expect(0, 66599, '\P{-CWCF}', "");
Expect(1, 66599, '\P{^-CWCF}', "");
Expect(0, 66600, '\p{-CWCF}', "");
Expect(1, 66600, '\p{^-CWCF}', "");
Expect(1, 66600, '\P{-CWCF}', "");
Expect(0, 66600, '\P{^-CWCF}', "");
Error('\p{	Is_CWCF/a/}');
Error('\P{	Is_CWCF/a/}');
Expect(1, 66599, '\p{iscwcf}', "");
Expect(0, 66599, '\p{^iscwcf}', "");
Expect(0, 66599, '\P{iscwcf}', "");
Expect(1, 66599, '\P{^iscwcf}', "");
Expect(0, 66600, '\p{iscwcf}', "");
Expect(1, 66600, '\p{^iscwcf}', "");
Expect(1, 66600, '\P{iscwcf}', "");
Expect(0, 66600, '\P{^iscwcf}', "");
Expect(1, 66599, '\p{  IS_cwcf}', "");
Expect(0, 66599, '\p{^  IS_cwcf}', "");
Expect(0, 66599, '\P{  IS_cwcf}', "");
Expect(1, 66599, '\P{^  IS_cwcf}', "");
Expect(0, 66600, '\p{  IS_cwcf}', "");
Expect(1, 66600, '\p{^  IS_cwcf}', "");
Expect(1, 66600, '\P{  IS_cwcf}', "");
Expect(0, 66600, '\P{^  IS_cwcf}', "");
Error('\p{	:=Lycian}');
Error('\P{	:=Lycian}');
Expect(1, 66204, '\p{lycian}', "");
Expect(0, 66204, '\p{^lycian}', "");
Expect(0, 66204, '\P{lycian}', "");
Expect(1, 66204, '\P{^lycian}', "");
Expect(0, 66205, '\p{lycian}', "");
Expect(1, 66205, '\p{^lycian}', "");
Expect(1, 66205, '\P{lycian}', "");
Expect(0, 66205, '\P{^lycian}', "");
Expect(1, 66204, '\p{_ Lycian}', "");
Expect(0, 66204, '\p{^_ Lycian}', "");
Expect(0, 66204, '\P{_ Lycian}', "");
Expect(1, 66204, '\P{^_ Lycian}', "");
Expect(0, 66205, '\p{_ Lycian}', "");
Expect(1, 66205, '\p{^_ Lycian}', "");
Expect(1, 66205, '\P{_ Lycian}', "");
Expect(0, 66205, '\P{^_ Lycian}', "");
Error('\p{:=is_LYCIAN}');
Error('\P{:=is_LYCIAN}');
Expect(1, 66204, '\p{islycian}', "");
Expect(0, 66204, '\p{^islycian}', "");
Expect(0, 66204, '\P{islycian}', "");
Expect(1, 66204, '\P{^islycian}', "");
Expect(0, 66205, '\p{islycian}', "");
Expect(1, 66205, '\p{^islycian}', "");
Expect(1, 66205, '\P{islycian}', "");
Expect(0, 66205, '\P{^islycian}', "");
Expect(1, 66204, '\p{	-IS_Lycian}', "");
Expect(0, 66204, '\p{^	-IS_Lycian}', "");
Expect(0, 66204, '\P{	-IS_Lycian}', "");
Expect(1, 66204, '\P{^	-IS_Lycian}', "");
Expect(0, 66205, '\p{	-IS_Lycian}', "");
Expect(1, 66205, '\p{^	-IS_Lycian}', "");
Expect(1, 66205, '\P{	-IS_Lycian}', "");
Expect(0, 66205, '\P{^	-IS_Lycian}', "");
Error('\p{ /a/LYCI}');
Error('\P{ /a/LYCI}');
Expect(1, 66204, '\p{lyci}', "");
Expect(0, 66204, '\p{^lyci}', "");
Expect(0, 66204, '\P{lyci}', "");
Expect(1, 66204, '\P{^lyci}', "");
Expect(0, 66205, '\p{lyci}', "");
Expect(1, 66205, '\p{^lyci}', "");
Expect(1, 66205, '\P{lyci}', "");
Expect(0, 66205, '\P{^lyci}', "");
Expect(1, 66204, '\p{ _lyci}', "");
Expect(0, 66204, '\p{^ _lyci}', "");
Expect(0, 66204, '\P{ _lyci}', "");
Expect(1, 66204, '\P{^ _lyci}', "");
Expect(0, 66205, '\p{ _lyci}', "");
Expect(1, 66205, '\p{^ _lyci}', "");
Expect(1, 66205, '\P{ _lyci}', "");
Expect(0, 66205, '\P{^ _lyci}', "");
Error('\p{:=	 Is_Lyci}');
Error('\P{:=	 Is_Lyci}');
Expect(1, 66204, '\p{islyci}', "");
Expect(0, 66204, '\p{^islyci}', "");
Expect(0, 66204, '\P{islyci}', "");
Expect(1, 66204, '\P{^islyci}', "");
Expect(0, 66205, '\p{islyci}', "");
Expect(1, 66205, '\p{^islyci}', "");
Expect(1, 66205, '\P{islyci}', "");
Expect(0, 66205, '\P{^islyci}', "");
Expect(1, 66204, '\p{	-is_Lyci}', "");
Expect(0, 66204, '\p{^	-is_Lyci}', "");
Expect(0, 66204, '\P{	-is_Lyci}', "");
Expect(1, 66204, '\P{^	-is_Lyci}', "");
Expect(0, 66205, '\p{	-is_Lyci}', "");
Expect(1, 66205, '\p{^	-is_Lyci}', "");
Expect(1, 66205, '\P{	-is_Lyci}', "");
Expect(0, 66205, '\P{^	-is_Lyci}', "");
Error('\p{:=- number}');
Error('\P{:=- number}');
Expect(1, 127242, '\p{number}', "");
Expect(0, 127242, '\p{^number}', "");
Expect(0, 127242, '\P{number}', "");
Expect(1, 127242, '\P{^number}', "");
Expect(0, 127243, '\p{number}', "");
Expect(1, 127243, '\p{^number}', "");
Expect(1, 127243, '\P{number}', "");
Expect(0, 127243, '\P{^number}', "");
Expect(1, 127242, '\p{-	Number}', "");
Expect(0, 127242, '\p{^-	Number}', "");
Expect(0, 127242, '\P{-	Number}', "");
Expect(1, 127242, '\P{^-	Number}', "");
Expect(0, 127243, '\p{-	Number}', "");
Expect(1, 127243, '\p{^-	Number}', "");
Expect(1, 127243, '\P{-	Number}', "");
Expect(0, 127243, '\P{^-	Number}', "");
Error('\p{ /a/IS_Number}');
Error('\P{ /a/IS_Number}');
Expect(1, 127242, '\p{isnumber}', "");
Expect(0, 127242, '\p{^isnumber}', "");
Expect(0, 127242, '\P{isnumber}', "");
Expect(1, 127242, '\P{^isnumber}', "");
Expect(0, 127243, '\p{isnumber}', "");
Expect(1, 127243, '\p{^isnumber}', "");
Expect(1, 127243, '\P{isnumber}', "");
Expect(0, 127243, '\P{^isnumber}', "");
Expect(1, 127242, '\p{	is_Number}', "");
Expect(0, 127242, '\p{^	is_Number}', "");
Expect(0, 127242, '\P{	is_Number}', "");
Expect(1, 127242, '\P{^	is_Number}', "");
Expect(0, 127243, '\p{	is_Number}', "");
Expect(1, 127243, '\p{^	is_Number}', "");
Expect(1, 127243, '\P{	is_Number}', "");
Expect(0, 127243, '\P{^	is_Number}', "");
Error('\p{-/a/N}');
Error('\P{-/a/N}');
Expect(1, 127242, '\p{n}', "");
Expect(0, 127242, '\p{^n}', "");
Expect(0, 127242, '\P{n}', "");
Expect(1, 127242, '\P{^n}', "");
Expect(0, 127243, '\p{n}', "");
Expect(1, 127243, '\p{^n}', "");
Expect(1, 127243, '\P{n}', "");
Expect(0, 127243, '\P{^n}', "");
Expect(1, 127242, '\p{	n}', "");
Expect(0, 127242, '\p{^	n}', "");
Expect(0, 127242, '\P{	n}', "");
Expect(1, 127242, '\P{^	n}', "");
Expect(0, 127243, '\p{	n}', "");
Expect(1, 127243, '\p{^	n}', "");
Expect(1, 127243, '\P{	n}', "");
Expect(0, 127243, '\P{^	n}', "");
Error('\p{/a/	is_n}');
Error('\P{/a/	is_n}');
Expect(1, 127242, '\p{isn}', "");
Expect(0, 127242, '\p{^isn}', "");
Expect(0, 127242, '\P{isn}', "");
Expect(1, 127242, '\P{^isn}', "");
Expect(0, 127243, '\p{isn}', "");
Expect(1, 127243, '\p{^isn}', "");
Expect(1, 127243, '\P{isn}', "");
Expect(0, 127243, '\P{^isn}', "");
Expect(1, 127242, '\p{ IS_N}', "");
Expect(0, 127242, '\p{^ IS_N}', "");
Expect(0, 127242, '\P{ IS_N}', "");
Expect(1, 127242, '\P{^ IS_N}', "");
Expect(0, 127243, '\p{ IS_N}', "");
Expect(1, 127243, '\p{^ IS_N}', "");
Expect(1, 127243, '\P{ IS_N}', "");
Expect(0, 127243, '\P{^ IS_N}', "");
Error('\p{/a/	POSIXDIGIT}');
Error('\P{/a/	POSIXDIGIT}');
Expect(1, 57, '\p{posixdigit}', "");
Expect(0, 57, '\p{^posixdigit}', "");
Expect(0, 57, '\P{posixdigit}', "");
Expect(1, 57, '\P{^posixdigit}', "");
Expect(0, 58, '\p{posixdigit}', "");
Expect(1, 58, '\p{^posixdigit}', "");
Expect(1, 58, '\P{posixdigit}', "");
Expect(0, 58, '\P{^posixdigit}', "");
Expect(1, 57, '\p{ 	PosixDigit}', "");
Expect(0, 57, '\p{^ 	PosixDigit}', "");
Expect(0, 57, '\P{ 	PosixDigit}', "");
Expect(1, 57, '\P{^ 	PosixDigit}', "");
Expect(0, 58, '\p{ 	PosixDigit}', "");
Expect(1, 58, '\p{^ 	PosixDigit}', "");
Expect(1, 58, '\P{ 	PosixDigit}', "");
Expect(0, 58, '\P{^ 	PosixDigit}', "");
Error('\p{:=--is_posixdigit}');
Error('\P{:=--is_posixdigit}');
Expect(1, 57, '\p{isposixdigit}', "");
Expect(0, 57, '\p{^isposixdigit}', "");
Expect(0, 57, '\P{isposixdigit}', "");
Expect(1, 57, '\P{^isposixdigit}', "");
Expect(0, 58, '\p{isposixdigit}', "");
Expect(1, 58, '\p{^isposixdigit}', "");
Expect(1, 58, '\P{isposixdigit}', "");
Expect(0, 58, '\P{^isposixdigit}', "");
Expect(1, 57, '\p{	IS_PosixDigit}', "");
Expect(0, 57, '\p{^	IS_PosixDigit}', "");
Expect(0, 57, '\P{	IS_PosixDigit}', "");
Expect(1, 57, '\P{^	IS_PosixDigit}', "");
Expect(0, 58, '\p{	IS_PosixDigit}', "");
Expect(1, 58, '\p{^	IS_PosixDigit}', "");
Expect(1, 58, '\P{	IS_PosixDigit}', "");
Expect(0, 58, '\P{^	IS_PosixDigit}', "");
Error('\p{	-Georgian:=}');
Error('\P{	-Georgian:=}');
Expect(1, 11565, '\p{georgian}', "");
Expect(0, 11565, '\p{^georgian}', "");
Expect(0, 11565, '\P{georgian}', "");
Expect(1, 11565, '\P{^georgian}', "");
Expect(0, 11566, '\p{georgian}', "");
Expect(1, 11566, '\p{^georgian}', "");
Expect(1, 11566, '\P{georgian}', "");
Expect(0, 11566, '\P{^georgian}', "");
Expect(1, 11565, '\p{-	Georgian}', "");
Expect(0, 11565, '\p{^-	Georgian}', "");
Expect(0, 11565, '\P{-	Georgian}', "");
Expect(1, 11565, '\P{^-	Georgian}', "");
Expect(0, 11566, '\p{-	Georgian}', "");
Expect(1, 11566, '\p{^-	Georgian}', "");
Expect(1, 11566, '\P{-	Georgian}', "");
Expect(0, 11566, '\P{^-	Georgian}', "");
Error('\p{IS_Georgian:=}');
Error('\P{IS_Georgian:=}');
Expect(1, 11565, '\p{isgeorgian}', "");
Expect(0, 11565, '\p{^isgeorgian}', "");
Expect(0, 11565, '\P{isgeorgian}', "");
Expect(1, 11565, '\P{^isgeorgian}', "");
Expect(0, 11566, '\p{isgeorgian}', "");
Expect(1, 11566, '\p{^isgeorgian}', "");
Expect(1, 11566, '\P{isgeorgian}', "");
Expect(0, 11566, '\P{^isgeorgian}', "");
Expect(1, 11565, '\p{-_IS_georgian}', "");
Expect(0, 11565, '\p{^-_IS_georgian}', "");
Expect(0, 11565, '\P{-_IS_georgian}', "");
Expect(1, 11565, '\P{^-_IS_georgian}', "");
Expect(0, 11566, '\p{-_IS_georgian}', "");
Expect(1, 11566, '\p{^-_IS_georgian}', "");
Expect(1, 11566, '\P{-_IS_georgian}', "");
Expect(0, 11566, '\P{^-_IS_georgian}', "");
Error('\p{:=-_GEOR}');
Error('\P{:=-_GEOR}');
Expect(1, 11565, '\p{geor}', "");
Expect(0, 11565, '\p{^geor}', "");
Expect(0, 11565, '\P{geor}', "");
Expect(1, 11565, '\P{^geor}', "");
Expect(0, 11566, '\p{geor}', "");
Expect(1, 11566, '\p{^geor}', "");
Expect(1, 11566, '\P{geor}', "");
Expect(0, 11566, '\P{^geor}', "");
Expect(1, 11565, '\p{-	GEOR}', "");
Expect(0, 11565, '\p{^-	GEOR}', "");
Expect(0, 11565, '\P{-	GEOR}', "");
Expect(1, 11565, '\P{^-	GEOR}', "");
Expect(0, 11566, '\p{-	GEOR}', "");
Expect(1, 11566, '\p{^-	GEOR}', "");
Expect(1, 11566, '\P{-	GEOR}', "");
Expect(0, 11566, '\P{^-	GEOR}', "");
Error('\p{ -Is_geor:=}');
Error('\P{ -Is_geor:=}');
Expect(1, 11565, '\p{isgeor}', "");
Expect(0, 11565, '\p{^isgeor}', "");
Expect(0, 11565, '\P{isgeor}', "");
Expect(1, 11565, '\P{^isgeor}', "");
Expect(0, 11566, '\p{isgeor}', "");
Expect(1, 11566, '\p{^isgeor}', "");
Expect(1, 11566, '\P{isgeor}', "");
Expect(0, 11566, '\P{^isgeor}', "");
Expect(1, 11565, '\p{ IS_geor}', "");
Expect(0, 11565, '\p{^ IS_geor}', "");
Expect(0, 11565, '\P{ IS_geor}', "");
Expect(1, 11565, '\P{^ IS_geor}', "");
Expect(0, 11566, '\p{ IS_geor}', "");
Expect(1, 11566, '\p{^ IS_geor}', "");
Expect(1, 11566, '\P{ IS_geor}', "");
Expect(0, 11566, '\P{^ IS_geor}', "");
Error('\p{/a/--Assigned}');
Error('\P{/a/--Assigned}');
Expect(1, 1114109, '\p{assigned}', "");
Expect(0, 1114109, '\p{^assigned}', "");
Expect(0, 1114109, '\P{assigned}', "");
Expect(1, 1114109, '\P{^assigned}', "");
Expect(0, 918000, '\p{assigned}', "");
Expect(1, 918000, '\p{^assigned}', "");
Expect(1, 918000, '\P{assigned}', "");
Expect(0, 918000, '\P{^assigned}', "");
Expect(1, 1114109, '\p{Assigned}', "");
Expect(0, 1114109, '\p{^Assigned}', "");
Expect(0, 1114109, '\P{Assigned}', "");
Expect(1, 1114109, '\P{^Assigned}', "");
Expect(0, 918000, '\p{Assigned}', "");
Expect(1, 918000, '\p{^Assigned}', "");
Expect(1, 918000, '\P{Assigned}', "");
Expect(0, 918000, '\P{^Assigned}', "");
Error('\p{/a/ IS_ASSIGNED}');
Error('\P{/a/ IS_ASSIGNED}');
Expect(1, 1114109, '\p{isassigned}', "");
Expect(0, 1114109, '\p{^isassigned}', "");
Expect(0, 1114109, '\P{isassigned}', "");
Expect(1, 1114109, '\P{^isassigned}', "");
Expect(0, 918000, '\p{isassigned}', "");
Expect(1, 918000, '\p{^isassigned}', "");
Expect(1, 918000, '\P{isassigned}', "");
Expect(0, 918000, '\P{^isassigned}', "");
Expect(1, 1114109, '\p{	Is_Assigned}', "");
Expect(0, 1114109, '\p{^	Is_Assigned}', "");
Expect(0, 1114109, '\P{	Is_Assigned}', "");
Expect(1, 1114109, '\P{^	Is_Assigned}', "");
Expect(0, 918000, '\p{	Is_Assigned}', "");
Expect(1, 918000, '\p{^	Is_Assigned}', "");
Expect(1, 918000, '\P{	Is_Assigned}', "");
Expect(0, 918000, '\P{^	Is_Assigned}', "");
Error('\p{	_Vertical_forms/a/}');
Error('\P{	_Vertical_forms/a/}');
Expect(1, 65055, '\p{verticalforms}', "");
Expect(0, 65055, '\p{^verticalforms}', "");
Expect(0, 65055, '\P{verticalforms}', "");
Expect(1, 65055, '\P{^verticalforms}', "");
Expect(0, 65056, '\p{verticalforms}', "");
Expect(1, 65056, '\p{^verticalforms}', "");
Expect(1, 65056, '\P{verticalforms}', "");
Expect(0, 65056, '\P{^verticalforms}', "");
Expect(1, 65055, '\p{ _vertical_FORMS}', "");
Expect(0, 65055, '\p{^ _vertical_FORMS}', "");
Expect(0, 65055, '\P{ _vertical_FORMS}', "");
Expect(1, 65055, '\P{^ _vertical_FORMS}', "");
Expect(0, 65056, '\p{ _vertical_FORMS}', "");
Expect(1, 65056, '\p{^ _vertical_FORMS}', "");
Expect(1, 65056, '\P{ _vertical_FORMS}', "");
Expect(0, 65056, '\P{^ _vertical_FORMS}', "");
Error('\p{	/a/IS_VERTICAL_Forms}');
Error('\P{	/a/IS_VERTICAL_Forms}');
Expect(1, 65055, '\p{isverticalforms}', "");
Expect(0, 65055, '\p{^isverticalforms}', "");
Expect(0, 65055, '\P{isverticalforms}', "");
Expect(1, 65055, '\P{^isverticalforms}', "");
Expect(0, 65056, '\p{isverticalforms}', "");
Expect(1, 65056, '\p{^isverticalforms}', "");
Expect(1, 65056, '\P{isverticalforms}', "");
Expect(0, 65056, '\P{^isverticalforms}', "");
Expect(1, 65055, '\p{_Is_VERTICAL_Forms}', "");
Expect(0, 65055, '\p{^_Is_VERTICAL_Forms}', "");
Expect(0, 65055, '\P{_Is_VERTICAL_Forms}', "");
Expect(1, 65055, '\P{^_Is_VERTICAL_Forms}', "");
Expect(0, 65056, '\p{_Is_VERTICAL_Forms}', "");
Expect(1, 65056, '\p{^_Is_VERTICAL_Forms}', "");
Expect(1, 65056, '\P{_Is_VERTICAL_Forms}', "");
Expect(0, 65056, '\P{^_Is_VERTICAL_Forms}', "");
Error('\p{	-IN_VERTICAL_forms:=}');
Error('\P{	-IN_VERTICAL_forms:=}');
Expect(1, 65055, '\p{inverticalforms}', "");
Expect(0, 65055, '\p{^inverticalforms}', "");
Expect(0, 65055, '\P{inverticalforms}', "");
Expect(1, 65055, '\P{^inverticalforms}', "");
Expect(0, 65056, '\p{inverticalforms}', "");
Expect(1, 65056, '\p{^inverticalforms}', "");
Expect(1, 65056, '\P{inverticalforms}', "");
Expect(0, 65056, '\P{^inverticalforms}', "");
Expect(1, 65055, '\p{-In_Vertical_FORMS}', "");
Expect(0, 65055, '\p{^-In_Vertical_FORMS}', "");
Expect(0, 65055, '\P{-In_Vertical_FORMS}', "");
Expect(1, 65055, '\P{^-In_Vertical_FORMS}', "");
Expect(0, 65056, '\p{-In_Vertical_FORMS}', "");
Expect(1, 65056, '\p{^-In_Vertical_FORMS}', "");
Expect(1, 65056, '\P{-In_Vertical_FORMS}', "");
Expect(0, 65056, '\P{^-In_Vertical_FORMS}', "");
Error('\p{-:=Kharoshthi}');
Error('\P{-:=Kharoshthi}');
Expect(1, 68184, '\p{kharoshthi}', "");
Expect(0, 68184, '\p{^kharoshthi}', "");
Expect(0, 68184, '\P{kharoshthi}', "");
Expect(1, 68184, '\P{^kharoshthi}', "");
Expect(0, 68185, '\p{kharoshthi}', "");
Expect(1, 68185, '\p{^kharoshthi}', "");
Expect(1, 68185, '\P{kharoshthi}', "");
Expect(0, 68185, '\P{^kharoshthi}', "");
Expect(1, 68184, '\p{_ kharoshthi}', "");
Expect(0, 68184, '\p{^_ kharoshthi}', "");
Expect(0, 68184, '\P{_ kharoshthi}', "");
Expect(1, 68184, '\P{^_ kharoshthi}', "");
Expect(0, 68185, '\p{_ kharoshthi}', "");
Expect(1, 68185, '\p{^_ kharoshthi}', "");
Expect(1, 68185, '\P{_ kharoshthi}', "");
Expect(0, 68185, '\P{^_ kharoshthi}', "");
Error('\p{/a/ -Is_kharoshthi}');
Error('\P{/a/ -Is_kharoshthi}');
Expect(1, 68184, '\p{iskharoshthi}', "");
Expect(0, 68184, '\p{^iskharoshthi}', "");
Expect(0, 68184, '\P{iskharoshthi}', "");
Expect(1, 68184, '\P{^iskharoshthi}', "");
Expect(0, 68185, '\p{iskharoshthi}', "");
Expect(1, 68185, '\p{^iskharoshthi}', "");
Expect(1, 68185, '\P{iskharoshthi}', "");
Expect(0, 68185, '\P{^iskharoshthi}', "");
Expect(1, 68184, '\p{  Is_KHAROSHTHI}', "");
Expect(0, 68184, '\p{^  Is_KHAROSHTHI}', "");
Expect(0, 68184, '\P{  Is_KHAROSHTHI}', "");
Expect(1, 68184, '\P{^  Is_KHAROSHTHI}', "");
Expect(0, 68185, '\p{  Is_KHAROSHTHI}', "");
Expect(1, 68185, '\p{^  Is_KHAROSHTHI}', "");
Expect(1, 68185, '\P{  Is_KHAROSHTHI}', "");
Expect(0, 68185, '\P{^  Is_KHAROSHTHI}', "");
Error('\p{ KHAR/a/}');
Error('\P{ KHAR/a/}');
Expect(1, 68184, '\p{khar}', "");
Expect(0, 68184, '\p{^khar}', "");
Expect(0, 68184, '\P{khar}', "");
Expect(1, 68184, '\P{^khar}', "");
Expect(0, 68185, '\p{khar}', "");
Expect(1, 68185, '\p{^khar}', "");
Expect(1, 68185, '\P{khar}', "");
Expect(0, 68185, '\P{^khar}', "");
Expect(1, 68184, '\p{  Khar}', "");
Expect(0, 68184, '\p{^  Khar}', "");
Expect(0, 68184, '\P{  Khar}', "");
Expect(1, 68184, '\P{^  Khar}', "");
Expect(0, 68185, '\p{  Khar}', "");
Expect(1, 68185, '\p{^  Khar}', "");
Expect(1, 68185, '\P{  Khar}', "");
Expect(0, 68185, '\P{^  Khar}', "");
Error('\p{ Is_Khar:=}');
Error('\P{ Is_Khar:=}');
Expect(1, 68184, '\p{iskhar}', "");
Expect(0, 68184, '\p{^iskhar}', "");
Expect(0, 68184, '\P{iskhar}', "");
Expect(1, 68184, '\P{^iskhar}', "");
Expect(0, 68185, '\p{iskhar}', "");
Expect(1, 68185, '\p{^iskhar}', "");
Expect(1, 68185, '\P{iskhar}', "");
Expect(0, 68185, '\P{^iskhar}', "");
Expect(1, 68184, '\p{ Is_Khar}', "");
Expect(0, 68184, '\p{^ Is_Khar}', "");
Expect(0, 68184, '\P{ Is_Khar}', "");
Expect(1, 68184, '\P{^ Is_Khar}', "");
Expect(0, 68185, '\p{ Is_Khar}', "");
Expect(1, 68185, '\p{^ Is_Khar}', "");
Expect(1, 68185, '\P{ Is_Khar}', "");
Expect(0, 68185, '\P{^ Is_Khar}', "");
Error('\p{:=		MATH_Symbol}');
Error('\P{:=		MATH_Symbol}');
Expect(1, 126705, '\p{mathsymbol}', "");
Expect(0, 126705, '\p{^mathsymbol}', "");
Expect(0, 126705, '\P{mathsymbol}', "");
Expect(1, 126705, '\P{^mathsymbol}', "");
Expect(0, 126706, '\p{mathsymbol}', "");
Expect(1, 126706, '\p{^mathsymbol}', "");
Expect(1, 126706, '\P{mathsymbol}', "");
Expect(0, 126706, '\P{^mathsymbol}', "");
Expect(1, 126705, '\p{	MATH_symbol}', "");
Expect(0, 126705, '\p{^	MATH_symbol}', "");
Expect(0, 126705, '\P{	MATH_symbol}', "");
Expect(1, 126705, '\P{^	MATH_symbol}', "");
Expect(0, 126706, '\p{	MATH_symbol}', "");
Expect(1, 126706, '\p{^	MATH_symbol}', "");
Expect(1, 126706, '\P{	MATH_symbol}', "");
Expect(0, 126706, '\P{^	MATH_symbol}', "");
Error('\p{/a/ 	IS_Math_SYMBOL}');
Error('\P{/a/ 	IS_Math_SYMBOL}');
Expect(1, 126705, '\p{ismathsymbol}', "");
Expect(0, 126705, '\p{^ismathsymbol}', "");
Expect(0, 126705, '\P{ismathsymbol}', "");
Expect(1, 126705, '\P{^ismathsymbol}', "");
Expect(0, 126706, '\p{ismathsymbol}', "");
Expect(1, 126706, '\p{^ismathsymbol}', "");
Expect(1, 126706, '\P{ismathsymbol}', "");
Expect(0, 126706, '\P{^ismathsymbol}', "");
Expect(1, 126705, '\p{	is_math_Symbol}', "");
Expect(0, 126705, '\p{^	is_math_Symbol}', "");
Expect(0, 126705, '\P{	is_math_Symbol}', "");
Expect(1, 126705, '\P{^	is_math_Symbol}', "");
Expect(0, 126706, '\p{	is_math_Symbol}', "");
Expect(1, 126706, '\p{^	is_math_Symbol}', "");
Expect(1, 126706, '\P{	is_math_Symbol}', "");
Expect(0, 126706, '\P{^	is_math_Symbol}', "");
Error('\p{-:=Sm}');
Error('\P{-:=Sm}');
Expect(1, 126705, '\p{sm}', "");
Expect(0, 126705, '\p{^sm}', "");
Expect(0, 126705, '\P{sm}', "");
Expect(1, 126705, '\P{^sm}', "");
Expect(0, 126706, '\p{sm}', "");
Expect(1, 126706, '\p{^sm}', "");
Expect(1, 126706, '\P{sm}', "");
Expect(0, 126706, '\P{^sm}', "");
Expect(1, 126705, '\p{-	sm}', "");
Expect(0, 126705, '\p{^-	sm}', "");
Expect(0, 126705, '\P{-	sm}', "");
Expect(1, 126705, '\P{^-	sm}', "");
Expect(0, 126706, '\p{-	sm}', "");
Expect(1, 126706, '\p{^-	sm}', "");
Expect(1, 126706, '\P{-	sm}', "");
Expect(0, 126706, '\P{^-	sm}', "");
Error('\p{/a/_ is_Sm}');
Error('\P{/a/_ is_Sm}');
Expect(1, 126705, '\p{issm}', "");
Expect(0, 126705, '\p{^issm}', "");
Expect(0, 126705, '\P{issm}', "");
Expect(1, 126705, '\P{^issm}', "");
Expect(0, 126706, '\p{issm}', "");
Expect(1, 126706, '\p{^issm}', "");
Expect(1, 126706, '\P{issm}', "");
Expect(0, 126706, '\P{^issm}', "");
Expect(1, 126705, '\p{_ Is_Sm}', "");
Expect(0, 126705, '\p{^_ Is_Sm}', "");
Expect(0, 126705, '\P{_ Is_Sm}', "");
Expect(1, 126705, '\P{^_ Is_Sm}', "");
Expect(0, 126706, '\p{_ Is_Sm}', "");
Expect(1, 126706, '\p{^_ Is_Sm}', "");
Expect(1, 126706, '\P{_ Is_Sm}', "");
Expect(0, 126706, '\P{^_ Is_Sm}', "");
Error('\p{_ Tagalog/a/}');
Error('\P{_ Tagalog/a/}');
Expect(1, 5908, '\p{tagalog}', "");
Expect(0, 5908, '\p{^tagalog}', "");
Expect(0, 5908, '\P{tagalog}', "");
Expect(1, 5908, '\P{^tagalog}', "");
Expect(0, 5909, '\p{tagalog}', "");
Expect(1, 5909, '\p{^tagalog}', "");
Expect(1, 5909, '\P{tagalog}', "");
Expect(0, 5909, '\P{^tagalog}', "");
Expect(1, 5908, '\p{	-tagalog}', "");
Expect(0, 5908, '\p{^	-tagalog}', "");
Expect(0, 5908, '\P{	-tagalog}', "");
Expect(1, 5908, '\P{^	-tagalog}', "");
Expect(0, 5909, '\p{	-tagalog}', "");
Expect(1, 5909, '\p{^	-tagalog}', "");
Expect(1, 5909, '\P{	-tagalog}', "");
Expect(0, 5909, '\P{^	-tagalog}', "");
Error('\p{	:=IS_Tagalog}');
Error('\P{	:=IS_Tagalog}');
Expect(1, 5908, '\p{istagalog}', "");
Expect(0, 5908, '\p{^istagalog}', "");
Expect(0, 5908, '\P{istagalog}', "");
Expect(1, 5908, '\P{^istagalog}', "");
Expect(0, 5909, '\p{istagalog}', "");
Expect(1, 5909, '\p{^istagalog}', "");
Expect(1, 5909, '\P{istagalog}', "");
Expect(0, 5909, '\P{^istagalog}', "");
Expect(1, 5908, '\p{ 	is_TAGALOG}', "");
Expect(0, 5908, '\p{^ 	is_TAGALOG}', "");
Expect(0, 5908, '\P{ 	is_TAGALOG}', "");
Expect(1, 5908, '\P{^ 	is_TAGALOG}', "");
Expect(0, 5909, '\p{ 	is_TAGALOG}', "");
Expect(1, 5909, '\p{^ 	is_TAGALOG}', "");
Expect(1, 5909, '\P{ 	is_TAGALOG}', "");
Expect(0, 5909, '\P{^ 	is_TAGALOG}', "");
Error('\p{/a/TGLG}');
Error('\P{/a/TGLG}');
Expect(1, 5908, '\p{tglg}', "");
Expect(0, 5908, '\p{^tglg}', "");
Expect(0, 5908, '\P{tglg}', "");
Expect(1, 5908, '\P{^tglg}', "");
Expect(0, 5909, '\p{tglg}', "");
Expect(1, 5909, '\p{^tglg}', "");
Expect(1, 5909, '\P{tglg}', "");
Expect(0, 5909, '\P{^tglg}', "");
Expect(1, 5908, '\p{-Tglg}', "");
Expect(0, 5908, '\p{^-Tglg}', "");
Expect(0, 5908, '\P{-Tglg}', "");
Expect(1, 5908, '\P{^-Tglg}', "");
Expect(0, 5909, '\p{-Tglg}', "");
Expect(1, 5909, '\p{^-Tglg}', "");
Expect(1, 5909, '\P{-Tglg}', "");
Expect(0, 5909, '\P{^-Tglg}', "");
Error('\p{:=  Is_Tglg}');
Error('\P{:=  Is_Tglg}');
Expect(1, 5908, '\p{istglg}', "");
Expect(0, 5908, '\p{^istglg}', "");
Expect(0, 5908, '\P{istglg}', "");
Expect(1, 5908, '\P{^istglg}', "");
Expect(0, 5909, '\p{istglg}', "");
Expect(1, 5909, '\p{^istglg}', "");
Expect(1, 5909, '\P{istglg}', "");
Expect(0, 5909, '\P{^istglg}', "");
Expect(1, 5908, '\p{	-Is_TGLG}', "");
Expect(0, 5908, '\p{^	-Is_TGLG}', "");
Expect(0, 5908, '\P{	-Is_TGLG}', "");
Expect(1, 5908, '\P{^	-Is_TGLG}', "");
Expect(0, 5909, '\p{	-Is_TGLG}', "");
Expect(1, 5909, '\p{^	-Is_TGLG}', "");
Expect(1, 5909, '\P{	-Is_TGLG}', "");
Expect(0, 5909, '\P{^	-Is_TGLG}', "");
Error('\p{	Syriac:=}');
Error('\P{	Syriac:=}');
Expect(1, 1871, '\p{syriac}', "");
Expect(0, 1871, '\p{^syriac}', "");
Expect(0, 1871, '\P{syriac}', "");
Expect(1, 1871, '\P{^syriac}', "");
Expect(0, 1872, '\p{syriac}', "");
Expect(1, 1872, '\p{^syriac}', "");
Expect(1, 1872, '\P{syriac}', "");
Expect(0, 1872, '\P{^syriac}', "");
Expect(1, 1871, '\p{--Syriac}', "");
Expect(0, 1871, '\p{^--Syriac}', "");
Expect(0, 1871, '\P{--Syriac}', "");
Expect(1, 1871, '\P{^--Syriac}', "");
Expect(0, 1872, '\p{--Syriac}', "");
Expect(1, 1872, '\p{^--Syriac}', "");
Expect(1, 1872, '\P{--Syriac}', "");
Expect(0, 1872, '\P{^--Syriac}', "");
Error('\p{__Is_syriac/a/}');
Error('\P{__Is_syriac/a/}');
Expect(1, 1871, '\p{issyriac}', "");
Expect(0, 1871, '\p{^issyriac}', "");
Expect(0, 1871, '\P{issyriac}', "");
Expect(1, 1871, '\P{^issyriac}', "");
Expect(0, 1872, '\p{issyriac}', "");
Expect(1, 1872, '\p{^issyriac}', "");
Expect(1, 1872, '\P{issyriac}', "");
Expect(0, 1872, '\P{^issyriac}', "");
Expect(1, 1871, '\p{ Is_Syriac}', "");
Expect(0, 1871, '\p{^ Is_Syriac}', "");
Expect(0, 1871, '\P{ Is_Syriac}', "");
Expect(1, 1871, '\P{^ Is_Syriac}', "");
Expect(0, 1872, '\p{ Is_Syriac}', "");
Expect(1, 1872, '\p{^ Is_Syriac}', "");
Expect(1, 1872, '\P{ Is_Syriac}', "");
Expect(0, 1872, '\P{^ Is_Syriac}', "");
Error('\p{:=_-SYRC}');
Error('\P{:=_-SYRC}');
Expect(1, 1871, '\p{syrc}', "");
Expect(0, 1871, '\p{^syrc}', "");
Expect(0, 1871, '\P{syrc}', "");
Expect(1, 1871, '\P{^syrc}', "");
Expect(0, 1872, '\p{syrc}', "");
Expect(1, 1872, '\p{^syrc}', "");
Expect(1, 1872, '\P{syrc}', "");
Expect(0, 1872, '\P{^syrc}', "");
Expect(1, 1871, '\p{		SYRC}', "");
Expect(0, 1871, '\p{^		SYRC}', "");
Expect(0, 1871, '\P{		SYRC}', "");
Expect(1, 1871, '\P{^		SYRC}', "");
Expect(0, 1872, '\p{		SYRC}', "");
Expect(1, 1872, '\p{^		SYRC}', "");
Expect(1, 1872, '\P{		SYRC}', "");
Expect(0, 1872, '\P{^		SYRC}', "");
Error('\p{/a/	 IS_Syrc}');
Error('\P{/a/	 IS_Syrc}');
Expect(1, 1871, '\p{issyrc}', "");
Expect(0, 1871, '\p{^issyrc}', "");
Expect(0, 1871, '\P{issyrc}', "");
Expect(1, 1871, '\P{^issyrc}', "");
Expect(0, 1872, '\p{issyrc}', "");
Expect(1, 1872, '\p{^issyrc}', "");
Expect(1, 1872, '\P{issyrc}', "");
Expect(0, 1872, '\P{^issyrc}', "");
Expect(1, 1871, '\p{	_Is_Syrc}', "");
Expect(0, 1871, '\p{^	_Is_Syrc}', "");
Expect(0, 1871, '\P{	_Is_Syrc}', "");
Expect(1, 1871, '\P{^	_Is_Syrc}', "");
Expect(0, 1872, '\p{	_Is_Syrc}', "");
Expect(1, 1872, '\p{^	_Is_Syrc}', "");
Expect(1, 1872, '\P{	_Is_Syrc}', "");
Expect(0, 1872, '\P{^	_Is_Syrc}', "");
Error('\p{:=In_phoenician}');
Error('\P{:=In_phoenician}');
Expect(1, 67871, '\p{inphoenician}', "");
Expect(0, 67871, '\p{^inphoenician}', "");
Expect(0, 67871, '\P{inphoenician}', "");
Expect(1, 67871, '\P{^inphoenician}', "");
Expect(0, 67872, '\p{inphoenician}', "");
Expect(1, 67872, '\p{^inphoenician}', "");
Expect(1, 67872, '\P{inphoenician}', "");
Expect(0, 67872, '\P{^inphoenician}', "");
Expect(1, 67871, '\p{_	In_Phoenician}', "");
Expect(0, 67871, '\p{^_	In_Phoenician}', "");
Expect(0, 67871, '\P{_	In_Phoenician}', "");
Expect(1, 67871, '\P{^_	In_Phoenician}', "");
Expect(0, 67872, '\p{_	In_Phoenician}', "");
Expect(1, 67872, '\p{^_	In_Phoenician}', "");
Expect(1, 67872, '\P{_	In_Phoenician}', "");
Expect(0, 67872, '\P{^_	In_Phoenician}', "");
Error('\p{_/a/In_Phoenician}');
Error('\P{_/a/In_Phoenician}');
Expect(1, 67871, '\p{	in_Phoenician}', "");
Expect(0, 67871, '\p{^	in_Phoenician}', "");
Expect(0, 67871, '\P{	in_Phoenician}', "");
Expect(1, 67871, '\P{^	in_Phoenician}', "");
Expect(0, 67872, '\p{	in_Phoenician}', "");
Expect(1, 67872, '\p{^	in_Phoenician}', "");
Expect(1, 67872, '\P{	in_Phoenician}', "");
Expect(0, 67872, '\P{^	in_Phoenician}', "");
Error('\p{:=-ancient_Greek_NUMBERS}');
Error('\P{:=-ancient_Greek_NUMBERS}');
Expect(1, 65935, '\p{ancientgreeknumbers}', "");
Expect(0, 65935, '\p{^ancientgreeknumbers}', "");
Expect(0, 65935, '\P{ancientgreeknumbers}', "");
Expect(1, 65935, '\P{^ancientgreeknumbers}', "");
Expect(0, 65936, '\p{ancientgreeknumbers}', "");
Expect(1, 65936, '\p{^ancientgreeknumbers}', "");
Expect(1, 65936, '\P{ancientgreeknumbers}', "");
Expect(0, 65936, '\P{^ancientgreeknumbers}', "");
Expect(1, 65935, '\p{	Ancient_GREEK_NUMBERS}', "");
Expect(0, 65935, '\p{^	Ancient_GREEK_NUMBERS}', "");
Expect(0, 65935, '\P{	Ancient_GREEK_NUMBERS}', "");
Expect(1, 65935, '\P{^	Ancient_GREEK_NUMBERS}', "");
Expect(0, 65936, '\p{	Ancient_GREEK_NUMBERS}', "");
Expect(1, 65936, '\p{^	Ancient_GREEK_NUMBERS}', "");
Expect(1, 65936, '\P{	Ancient_GREEK_NUMBERS}', "");
Expect(0, 65936, '\P{^	Ancient_GREEK_NUMBERS}', "");
Error('\p{/a/	 is_ancient_greek_NUMBERS}');
Error('\P{/a/	 is_ancient_greek_NUMBERS}');
Expect(1, 65935, '\p{isancientgreeknumbers}', "");
Expect(0, 65935, '\p{^isancientgreeknumbers}', "");
Expect(0, 65935, '\P{isancientgreeknumbers}', "");
Expect(1, 65935, '\P{^isancientgreeknumbers}', "");
Expect(0, 65936, '\p{isancientgreeknumbers}', "");
Expect(1, 65936, '\p{^isancientgreeknumbers}', "");
Expect(1, 65936, '\P{isancientgreeknumbers}', "");
Expect(0, 65936, '\P{^isancientgreeknumbers}', "");
Expect(1, 65935, '\p{_-Is_Ancient_greek_Numbers}', "");
Expect(0, 65935, '\p{^_-Is_Ancient_greek_Numbers}', "");
Expect(0, 65935, '\P{_-Is_Ancient_greek_Numbers}', "");
Expect(1, 65935, '\P{^_-Is_Ancient_greek_Numbers}', "");
Expect(0, 65936, '\p{_-Is_Ancient_greek_Numbers}', "");
Expect(1, 65936, '\p{^_-Is_Ancient_greek_Numbers}', "");
Expect(1, 65936, '\P{_-Is_Ancient_greek_Numbers}', "");
Expect(0, 65936, '\P{^_-Is_Ancient_greek_Numbers}', "");
Error('\p{/a/- IN_ancient_Greek_numbers}');
Error('\P{/a/- IN_ancient_Greek_numbers}');
Expect(1, 65935, '\p{inancientgreeknumbers}', "");
Expect(0, 65935, '\p{^inancientgreeknumbers}', "");
Expect(0, 65935, '\P{inancientgreeknumbers}', "");
Expect(1, 65935, '\P{^inancientgreeknumbers}', "");
Expect(0, 65936, '\p{inancientgreeknumbers}', "");
Expect(1, 65936, '\p{^inancientgreeknumbers}', "");
Expect(1, 65936, '\P{inancientgreeknumbers}', "");
Expect(0, 65936, '\P{^inancientgreeknumbers}', "");
Expect(1, 65935, '\p{-	in_Ancient_GREEK_numbers}', "");
Expect(0, 65935, '\p{^-	in_Ancient_GREEK_numbers}', "");
Expect(0, 65935, '\P{-	in_Ancient_GREEK_numbers}', "");
Expect(1, 65935, '\P{^-	in_Ancient_GREEK_numbers}', "");
Expect(0, 65936, '\p{-	in_Ancient_GREEK_numbers}', "");
Expect(1, 65936, '\p{^-	in_Ancient_GREEK_numbers}', "");
Expect(1, 65936, '\P{-	in_Ancient_GREEK_numbers}', "");
Expect(0, 65936, '\P{^-	in_Ancient_GREEK_numbers}', "");
Error('\p{/a/_Optical_character_recognition}');
Error('\P{/a/_Optical_character_recognition}');
Expect(1, 9311, '\p{opticalcharacterrecognition}', "");
Expect(0, 9311, '\p{^opticalcharacterrecognition}', "");
Expect(0, 9311, '\P{opticalcharacterrecognition}', "");
Expect(1, 9311, '\P{^opticalcharacterrecognition}', "");
Expect(0, 9312, '\p{opticalcharacterrecognition}', "");
Expect(1, 9312, '\p{^opticalcharacterrecognition}', "");
Expect(1, 9312, '\P{opticalcharacterrecognition}', "");
Expect(0, 9312, '\P{^opticalcharacterrecognition}', "");
Expect(1, 9311, '\p{- OPTICAL_CHARACTER_Recognition}', "");
Expect(0, 9311, '\p{^- OPTICAL_CHARACTER_Recognition}', "");
Expect(0, 9311, '\P{- OPTICAL_CHARACTER_Recognition}', "");
Expect(1, 9311, '\P{^- OPTICAL_CHARACTER_Recognition}', "");
Expect(0, 9312, '\p{- OPTICAL_CHARACTER_Recognition}', "");
Expect(1, 9312, '\p{^- OPTICAL_CHARACTER_Recognition}', "");
Expect(1, 9312, '\P{- OPTICAL_CHARACTER_Recognition}', "");
Expect(0, 9312, '\P{^- OPTICAL_CHARACTER_Recognition}', "");
Error('\p{-:=IS_Optical_CHARACTER_recognition}');
Error('\P{-:=IS_Optical_CHARACTER_recognition}');
Expect(1, 9311, '\p{isopticalcharacterrecognition}', "");
Expect(0, 9311, '\p{^isopticalcharacterrecognition}', "");
Expect(0, 9311, '\P{isopticalcharacterrecognition}', "");
Expect(1, 9311, '\P{^isopticalcharacterrecognition}', "");
Expect(0, 9312, '\p{isopticalcharacterrecognition}', "");
Expect(1, 9312, '\p{^isopticalcharacterrecognition}', "");
Expect(1, 9312, '\P{isopticalcharacterrecognition}', "");
Expect(0, 9312, '\P{^isopticalcharacterrecognition}', "");
Expect(1, 9311, '\p{- is_Optical_character_Recognition}', "");
Expect(0, 9311, '\p{^- is_Optical_character_Recognition}', "");
Expect(0, 9311, '\P{- is_Optical_character_Recognition}', "");
Expect(1, 9311, '\P{^- is_Optical_character_Recognition}', "");
Expect(0, 9312, '\p{- is_Optical_character_Recognition}', "");
Expect(1, 9312, '\p{^- is_Optical_character_Recognition}', "");
Expect(1, 9312, '\P{- is_Optical_character_Recognition}', "");
Expect(0, 9312, '\P{^- is_Optical_character_Recognition}', "");
Error('\p{-_IN_OPTICAL_Character_Recognition:=}');
Error('\P{-_IN_OPTICAL_Character_Recognition:=}');
Expect(1, 9311, '\p{inopticalcharacterrecognition}', "");
Expect(0, 9311, '\p{^inopticalcharacterrecognition}', "");
Expect(0, 9311, '\P{inopticalcharacterrecognition}', "");
Expect(1, 9311, '\P{^inopticalcharacterrecognition}', "");
Expect(0, 9312, '\p{inopticalcharacterrecognition}', "");
Expect(1, 9312, '\p{^inopticalcharacterrecognition}', "");
Expect(1, 9312, '\P{inopticalcharacterrecognition}', "");
Expect(0, 9312, '\P{^inopticalcharacterrecognition}', "");
Expect(1, 9311, '\p{		IN_optical_CHARACTER_RECOGNITION}', "");
Expect(0, 9311, '\p{^		IN_optical_CHARACTER_RECOGNITION}', "");
Expect(0, 9311, '\P{		IN_optical_CHARACTER_RECOGNITION}', "");
Expect(1, 9311, '\P{^		IN_optical_CHARACTER_RECOGNITION}', "");
Expect(0, 9312, '\p{		IN_optical_CHARACTER_RECOGNITION}', "");
Expect(1, 9312, '\p{^		IN_optical_CHARACTER_RECOGNITION}', "");
Expect(1, 9312, '\P{		IN_optical_CHARACTER_RECOGNITION}', "");
Expect(0, 9312, '\P{^		IN_optical_CHARACTER_RECOGNITION}', "");
Error('\p{ :=OCR}');
Error('\P{ :=OCR}');
Expect(1, 9311, '\p{ocr}', "");
Expect(0, 9311, '\p{^ocr}', "");
Expect(0, 9311, '\P{ocr}', "");
Expect(1, 9311, '\P{^ocr}', "");
Expect(0, 9312, '\p{ocr}', "");
Expect(1, 9312, '\p{^ocr}', "");
Expect(1, 9312, '\P{ocr}', "");
Expect(0, 9312, '\P{^ocr}', "");
Expect(1, 9311, '\p{	OCR}', "");
Expect(0, 9311, '\p{^	OCR}', "");
Expect(0, 9311, '\P{	OCR}', "");
Expect(1, 9311, '\P{^	OCR}', "");
Expect(0, 9312, '\p{	OCR}', "");
Expect(1, 9312, '\p{^	OCR}', "");
Expect(1, 9312, '\P{	OCR}', "");
Expect(0, 9312, '\P{^	OCR}', "");
Error('\p{:= Is_ocr}');
Error('\P{:= Is_ocr}');
Expect(1, 9311, '\p{isocr}', "");
Expect(0, 9311, '\p{^isocr}', "");
Expect(0, 9311, '\P{isocr}', "");
Expect(1, 9311, '\P{^isocr}', "");
Expect(0, 9312, '\p{isocr}', "");
Expect(1, 9312, '\p{^isocr}', "");
Expect(1, 9312, '\P{isocr}', "");
Expect(0, 9312, '\P{^isocr}', "");
Expect(1, 9311, '\p{-	Is_OCR}', "");
Expect(0, 9311, '\p{^-	Is_OCR}', "");
Expect(0, 9311, '\P{-	Is_OCR}', "");
Expect(1, 9311, '\P{^-	Is_OCR}', "");
Expect(0, 9312, '\p{-	Is_OCR}', "");
Expect(1, 9312, '\p{^-	Is_OCR}', "");
Expect(1, 9312, '\P{-	Is_OCR}', "");
Expect(0, 9312, '\P{^-	Is_OCR}', "");
Error('\p{ :=IN_OCR}');
Error('\P{ :=IN_OCR}');
Expect(1, 9311, '\p{inocr}', "");
Expect(0, 9311, '\p{^inocr}', "");
Expect(0, 9311, '\P{inocr}', "");
Expect(1, 9311, '\P{^inocr}', "");
Expect(0, 9312, '\p{inocr}', "");
Expect(1, 9312, '\p{^inocr}', "");
Expect(1, 9312, '\P{inocr}', "");
Expect(0, 9312, '\P{^inocr}', "");
Expect(1, 9311, '\p{_-in_ocr}', "");
Expect(0, 9311, '\p{^_-in_ocr}', "");
Expect(0, 9311, '\P{_-in_ocr}', "");
Expect(1, 9311, '\P{^_-in_ocr}', "");
Expect(0, 9312, '\p{_-in_ocr}', "");
Expect(1, 9312, '\p{^_-in_ocr}', "");
Expect(1, 9312, '\P{_-in_ocr}', "");
Expect(0, 9312, '\P{^_-in_ocr}', "");
Error('\p{:=  LIMBU}');
Error('\P{:=  LIMBU}');
Expect(1, 6479, '\p{limbu}', "");
Expect(0, 6479, '\p{^limbu}', "");
Expect(0, 6479, '\P{limbu}', "");
Expect(1, 6479, '\P{^limbu}', "");
Expect(0, 6480, '\p{limbu}', "");
Expect(1, 6480, '\p{^limbu}', "");
Expect(1, 6480, '\P{limbu}', "");
Expect(0, 6480, '\P{^limbu}', "");
Expect(1, 6479, '\p{_Limbu}', "");
Expect(0, 6479, '\p{^_Limbu}', "");
Expect(0, 6479, '\P{_Limbu}', "");
Expect(1, 6479, '\P{^_Limbu}', "");
Expect(0, 6480, '\p{_Limbu}', "");
Expect(1, 6480, '\p{^_Limbu}', "");
Expect(1, 6480, '\P{_Limbu}', "");
Expect(0, 6480, '\P{^_Limbu}', "");
Error('\p{_Is_limbu:=}');
Error('\P{_Is_limbu:=}');
Expect(1, 6479, '\p{islimbu}', "");
Expect(0, 6479, '\p{^islimbu}', "");
Expect(0, 6479, '\P{islimbu}', "");
Expect(1, 6479, '\P{^islimbu}', "");
Expect(0, 6480, '\p{islimbu}', "");
Expect(1, 6480, '\p{^islimbu}', "");
Expect(1, 6480, '\P{islimbu}', "");
Expect(0, 6480, '\P{^islimbu}', "");
Expect(1, 6479, '\p{	Is_LIMBU}', "");
Expect(0, 6479, '\p{^	Is_LIMBU}', "");
Expect(0, 6479, '\P{	Is_LIMBU}', "");
Expect(1, 6479, '\P{^	Is_LIMBU}', "");
Expect(0, 6480, '\p{	Is_LIMBU}', "");
Expect(1, 6480, '\p{^	Is_LIMBU}', "");
Expect(1, 6480, '\P{	Is_LIMBU}', "");
Expect(0, 6480, '\P{^	Is_LIMBU}', "");
Error('\p{/a/	-limb}');
Error('\P{/a/	-limb}');
Expect(1, 6479, '\p{limb}', "");
Expect(0, 6479, '\p{^limb}', "");
Expect(0, 6479, '\P{limb}', "");
Expect(1, 6479, '\P{^limb}', "");
Expect(0, 6480, '\p{limb}', "");
Expect(1, 6480, '\p{^limb}', "");
Expect(1, 6480, '\P{limb}', "");
Expect(0, 6480, '\P{^limb}', "");
Expect(1, 6479, '\p{- Limb}', "");
Expect(0, 6479, '\p{^- Limb}', "");
Expect(0, 6479, '\P{- Limb}', "");
Expect(1, 6479, '\P{^- Limb}', "");
Expect(0, 6480, '\p{- Limb}', "");
Expect(1, 6480, '\p{^- Limb}', "");
Expect(1, 6480, '\P{- Limb}', "");
Expect(0, 6480, '\P{^- Limb}', "");
Error('\p{ :=is_limb}');
Error('\P{ :=is_limb}');
Expect(1, 6479, '\p{islimb}', "");
Expect(0, 6479, '\p{^islimb}', "");
Expect(0, 6479, '\P{islimb}', "");
Expect(1, 6479, '\P{^islimb}', "");
Expect(0, 6480, '\p{islimb}', "");
Expect(1, 6480, '\p{^islimb}', "");
Expect(1, 6480, '\P{islimb}', "");
Expect(0, 6480, '\P{^islimb}', "");
Expect(1, 6479, '\p{-_IS_limb}', "");
Expect(0, 6479, '\p{^-_IS_limb}', "");
Expect(0, 6479, '\P{-_IS_limb}', "");
Expect(1, 6479, '\P{^-_IS_limb}', "");
Expect(0, 6480, '\p{-_IS_limb}', "");
Expect(1, 6480, '\p{^-_IS_limb}', "");
Expect(1, 6480, '\P{-_IS_limb}', "");
Expect(0, 6480, '\P{^-_IS_limb}', "");
Error('\p{ 	SHARADA/a/}');
Error('\P{ 	SHARADA/a/}');
Expect(1, 70105, '\p{sharada}', "");
Expect(0, 70105, '\p{^sharada}', "");
Expect(0, 70105, '\P{sharada}', "");
Expect(1, 70105, '\P{^sharada}', "");
Expect(0, 70106, '\p{sharada}', "");
Expect(1, 70106, '\p{^sharada}', "");
Expect(1, 70106, '\P{sharada}', "");
Expect(0, 70106, '\P{^sharada}', "");
Expect(1, 70105, '\p{	_sharada}', "");
Expect(0, 70105, '\p{^	_sharada}', "");
Expect(0, 70105, '\P{	_sharada}', "");
Expect(1, 70105, '\P{^	_sharada}', "");
Expect(0, 70106, '\p{	_sharada}', "");
Expect(1, 70106, '\p{^	_sharada}', "");
Expect(1, 70106, '\P{	_sharada}', "");
Expect(0, 70106, '\P{^	_sharada}', "");
Error('\p{/a/	Is_Sharada}');
Error('\P{/a/	Is_Sharada}');
Expect(1, 70105, '\p{issharada}', "");
Expect(0, 70105, '\p{^issharada}', "");
Expect(0, 70105, '\P{issharada}', "");
Expect(1, 70105, '\P{^issharada}', "");
Expect(0, 70106, '\p{issharada}', "");
Expect(1, 70106, '\p{^issharada}', "");
Expect(1, 70106, '\P{issharada}', "");
Expect(0, 70106, '\P{^issharada}', "");
Expect(1, 70105, '\p{-is_sharada}', "");
Expect(0, 70105, '\p{^-is_sharada}', "");
Expect(0, 70105, '\P{-is_sharada}', "");
Expect(1, 70105, '\P{^-is_sharada}', "");
Expect(0, 70106, '\p{-is_sharada}', "");
Expect(1, 70106, '\p{^-is_sharada}', "");
Expect(1, 70106, '\P{-is_sharada}', "");
Expect(0, 70106, '\P{^-is_sharada}', "");
Error('\p{/a/ _Shrd}');
Error('\P{/a/ _Shrd}');
Expect(1, 70105, '\p{shrd}', "");
Expect(0, 70105, '\p{^shrd}', "");
Expect(0, 70105, '\P{shrd}', "");
Expect(1, 70105, '\P{^shrd}', "");
Expect(0, 70106, '\p{shrd}', "");
Expect(1, 70106, '\p{^shrd}', "");
Expect(1, 70106, '\P{shrd}', "");
Expect(0, 70106, '\P{^shrd}', "");
Expect(1, 70105, '\p{--SHRD}', "");
Expect(0, 70105, '\p{^--SHRD}', "");
Expect(0, 70105, '\P{--SHRD}', "");
Expect(1, 70105, '\P{^--SHRD}', "");
Expect(0, 70106, '\p{--SHRD}', "");
Expect(1, 70106, '\p{^--SHRD}', "");
Expect(1, 70106, '\P{--SHRD}', "");
Expect(0, 70106, '\P{^--SHRD}', "");
Error('\p{:=_	IS_Shrd}');
Error('\P{:=_	IS_Shrd}');
Expect(1, 70105, '\p{isshrd}', "");
Expect(0, 70105, '\p{^isshrd}', "");
Expect(0, 70105, '\P{isshrd}', "");
Expect(1, 70105, '\P{^isshrd}', "");
Expect(0, 70106, '\p{isshrd}', "");
Expect(1, 70106, '\p{^isshrd}', "");
Expect(1, 70106, '\P{isshrd}', "");
Expect(0, 70106, '\P{^isshrd}', "");
Expect(1, 70105, '\p{	-Is_SHRD}', "");
Expect(0, 70105, '\p{^	-Is_SHRD}', "");
Expect(0, 70105, '\P{	-Is_SHRD}', "");
Expect(1, 70105, '\P{^	-Is_SHRD}', "");
Expect(0, 70106, '\p{	-Is_SHRD}', "");
Expect(1, 70106, '\p{^	-Is_SHRD}', "");
Expect(1, 70106, '\P{	-Is_SHRD}', "");
Expect(0, 70106, '\P{^	-Is_SHRD}', "");
Error('\p{ /a/Terminal_Punctuation}');
Error('\P{ /a/Terminal_Punctuation}');
Expect(1, 74867, '\p{terminalpunctuation}', "");
Expect(0, 74867, '\p{^terminalpunctuation}', "");
Expect(0, 74867, '\P{terminalpunctuation}', "");
Expect(1, 74867, '\P{^terminalpunctuation}', "");
Expect(0, 74868, '\p{terminalpunctuation}', "");
Expect(1, 74868, '\p{^terminalpunctuation}', "");
Expect(1, 74868, '\P{terminalpunctuation}', "");
Expect(0, 74868, '\P{^terminalpunctuation}', "");
Expect(1, 74867, '\p{_terminal_Punctuation}', "");
Expect(0, 74867, '\p{^_terminal_Punctuation}', "");
Expect(0, 74867, '\P{_terminal_Punctuation}', "");
Expect(1, 74867, '\P{^_terminal_Punctuation}', "");
Expect(0, 74868, '\p{_terminal_Punctuation}', "");
Expect(1, 74868, '\p{^_terminal_Punctuation}', "");
Expect(1, 74868, '\P{_terminal_Punctuation}', "");
Expect(0, 74868, '\P{^_terminal_Punctuation}', "");
Error('\p{-	is_TERMINAL_PUNCTUATION:=}');
Error('\P{-	is_TERMINAL_PUNCTUATION:=}');
Expect(1, 74867, '\p{isterminalpunctuation}', "");
Expect(0, 74867, '\p{^isterminalpunctuation}', "");
Expect(0, 74867, '\P{isterminalpunctuation}', "");
Expect(1, 74867, '\P{^isterminalpunctuation}', "");
Expect(0, 74868, '\p{isterminalpunctuation}', "");
Expect(1, 74868, '\p{^isterminalpunctuation}', "");
Expect(1, 74868, '\P{isterminalpunctuation}', "");
Expect(0, 74868, '\P{^isterminalpunctuation}', "");
Expect(1, 74867, '\p{-IS_Terminal_punctuation}', "");
Expect(0, 74867, '\p{^-IS_Terminal_punctuation}', "");
Expect(0, 74867, '\P{-IS_Terminal_punctuation}', "");
Expect(1, 74867, '\P{^-IS_Terminal_punctuation}', "");
Expect(0, 74868, '\p{-IS_Terminal_punctuation}', "");
Expect(1, 74868, '\p{^-IS_Terminal_punctuation}', "");
Expect(1, 74868, '\P{-IS_Terminal_punctuation}', "");
Expect(0, 74868, '\P{^-IS_Terminal_punctuation}', "");
Error('\p{/a/_ TERM}');
Error('\P{/a/_ TERM}');
Expect(1, 74867, '\p{term}', "");
Expect(0, 74867, '\p{^term}', "");
Expect(0, 74867, '\P{term}', "");
Expect(1, 74867, '\P{^term}', "");
Expect(0, 74868, '\p{term}', "");
Expect(1, 74868, '\p{^term}', "");
Expect(1, 74868, '\P{term}', "");
Expect(0, 74868, '\P{^term}', "");
Expect(1, 74867, '\p{-_Term}', "");
Expect(0, 74867, '\p{^-_Term}', "");
Expect(0, 74867, '\P{-_Term}', "");
Expect(1, 74867, '\P{^-_Term}', "");
Expect(0, 74868, '\p{-_Term}', "");
Expect(1, 74868, '\p{^-_Term}', "");
Expect(1, 74868, '\P{-_Term}', "");
Expect(0, 74868, '\P{^-_Term}', "");
Error('\p{_:=Is_term}');
Error('\P{_:=Is_term}');
Expect(1, 74867, '\p{isterm}', "");
Expect(0, 74867, '\p{^isterm}', "");
Expect(0, 74867, '\P{isterm}', "");
Expect(1, 74867, '\P{^isterm}', "");
Expect(0, 74868, '\p{isterm}', "");
Expect(1, 74868, '\p{^isterm}', "");
Expect(1, 74868, '\P{isterm}', "");
Expect(0, 74868, '\P{^isterm}', "");
Expect(1, 74867, '\p{ 	Is_Term}', "");
Expect(0, 74867, '\p{^ 	Is_Term}', "");
Expect(0, 74867, '\P{ 	Is_Term}', "");
Expect(1, 74867, '\P{^ 	Is_Term}', "");
Expect(0, 74868, '\p{ 	Is_Term}', "");
Expect(1, 74868, '\p{^ 	Is_Term}', "");
Expect(1, 74868, '\P{ 	Is_Term}', "");
Expect(0, 74868, '\P{^ 	Is_Term}', "");
Error('\p{Other_Uppercase}');
Error('\P{Other_Uppercase}');
Error('\p{Is_Other_Uppercase}');
Error('\P{Is_Other_Uppercase}');
Error('\p{OUpper}');
Error('\P{OUpper}');
Error('\p{Is_OUpper}');
Error('\P{Is_OUpper}');
Error('\p{_POSIXPUNCT:=}');
Error('\P{_POSIXPUNCT:=}');
Expect(1, 126, '\p{posixpunct}', "");
Expect(0, 126, '\p{^posixpunct}', "");
Expect(0, 126, '\P{posixpunct}', "");
Expect(1, 126, '\P{^posixpunct}', "");
Expect(0, 127, '\p{posixpunct}', "");
Expect(1, 127, '\p{^posixpunct}', "");
Expect(1, 127, '\P{posixpunct}', "");
Expect(0, 127, '\P{^posixpunct}', "");
Expect(1, 126, '\p{  PosixPunct}', "");
Expect(0, 126, '\p{^  PosixPunct}', "");
Expect(0, 126, '\P{  PosixPunct}', "");
Expect(1, 126, '\P{^  PosixPunct}', "");
Expect(0, 127, '\p{  PosixPunct}', "");
Expect(1, 127, '\p{^  PosixPunct}', "");
Expect(1, 127, '\P{  PosixPunct}', "");
Expect(0, 127, '\P{^  PosixPunct}', "");
Error('\p{:=is_POSIXPUNCT}');
Error('\P{:=is_POSIXPUNCT}');
Expect(1, 126, '\p{isposixpunct}', "");
Expect(0, 126, '\p{^isposixpunct}', "");
Expect(0, 126, '\P{isposixpunct}', "");
Expect(1, 126, '\P{^isposixpunct}', "");
Expect(0, 127, '\p{isposixpunct}', "");
Expect(1, 127, '\p{^isposixpunct}', "");
Expect(1, 127, '\P{isposixpunct}', "");
Expect(0, 127, '\P{^isposixpunct}', "");
Expect(1, 126, '\p{-_Is_PosixPunct}', "");
Expect(0, 126, '\p{^-_Is_PosixPunct}', "");
Expect(0, 126, '\P{-_Is_PosixPunct}', "");
Expect(1, 126, '\P{^-_Is_PosixPunct}', "");
Expect(0, 127, '\p{-_Is_PosixPunct}', "");
Expect(1, 127, '\p{^-_Is_PosixPunct}', "");
Expect(1, 127, '\P{-_Is_PosixPunct}', "");
Expect(0, 127, '\P{^-_Is_PosixPunct}', "");
Error('\p{/a/-BLOCK_ELEMENTS}');
Error('\P{/a/-BLOCK_ELEMENTS}');
Expect(1, 9631, '\p{blockelements}', "");
Expect(0, 9631, '\p{^blockelements}', "");
Expect(0, 9631, '\P{blockelements}', "");
Expect(1, 9631, '\P{^blockelements}', "");
Expect(0, 9632, '\p{blockelements}', "");
Expect(1, 9632, '\p{^blockelements}', "");
Expect(1, 9632, '\P{blockelements}', "");
Expect(0, 9632, '\P{^blockelements}', "");
Expect(1, 9631, '\p{-block_Elements}', "");
Expect(0, 9631, '\p{^-block_Elements}', "");
Expect(0, 9631, '\P{-block_Elements}', "");
Expect(1, 9631, '\P{^-block_Elements}', "");
Expect(0, 9632, '\p{-block_Elements}', "");
Expect(1, 9632, '\p{^-block_Elements}', "");
Expect(1, 9632, '\P{-block_Elements}', "");
Expect(0, 9632, '\P{^-block_Elements}', "");
Error('\p{/a/ _Is_Block_elements}');
Error('\P{/a/ _Is_Block_elements}');
Expect(1, 9631, '\p{isblockelements}', "");
Expect(0, 9631, '\p{^isblockelements}', "");
Expect(0, 9631, '\P{isblockelements}', "");
Expect(1, 9631, '\P{^isblockelements}', "");
Expect(0, 9632, '\p{isblockelements}', "");
Expect(1, 9632, '\p{^isblockelements}', "");
Expect(1, 9632, '\P{isblockelements}', "");
Expect(0, 9632, '\P{^isblockelements}', "");
Expect(1, 9631, '\p{ Is_Block_Elements}', "");
Expect(0, 9631, '\p{^ Is_Block_Elements}', "");
Expect(0, 9631, '\P{ Is_Block_Elements}', "");
Expect(1, 9631, '\P{^ Is_Block_Elements}', "");
Expect(0, 9632, '\p{ Is_Block_Elements}', "");
Expect(1, 9632, '\p{^ Is_Block_Elements}', "");
Expect(1, 9632, '\P{ Is_Block_Elements}', "");
Expect(0, 9632, '\P{^ Is_Block_Elements}', "");
Error('\p{ :=in_Block_Elements}');
Error('\P{ :=in_Block_Elements}');
Expect(1, 9631, '\p{inblockelements}', "");
Expect(0, 9631, '\p{^inblockelements}', "");
Expect(0, 9631, '\P{inblockelements}', "");
Expect(1, 9631, '\P{^inblockelements}', "");
Expect(0, 9632, '\p{inblockelements}', "");
Expect(1, 9632, '\p{^inblockelements}', "");
Expect(1, 9632, '\P{inblockelements}', "");
Expect(0, 9632, '\P{^inblockelements}', "");
Expect(1, 9631, '\p{	In_BLOCK_elements}', "");
Expect(0, 9631, '\p{^	In_BLOCK_elements}', "");
Expect(0, 9631, '\P{	In_BLOCK_elements}', "");
Expect(1, 9631, '\P{^	In_BLOCK_elements}', "");
Expect(0, 9632, '\p{	In_BLOCK_elements}', "");
Expect(1, 9632, '\p{^	In_BLOCK_elements}', "");
Expect(1, 9632, '\P{	In_BLOCK_elements}', "");
Expect(0, 9632, '\P{^	In_BLOCK_elements}', "");
Error('\p{:=In_KAITHI}');
Error('\P{:=In_KAITHI}');
Expect(1, 69839, '\p{inkaithi}', "");
Expect(0, 69839, '\p{^inkaithi}', "");
Expect(0, 69839, '\P{inkaithi}', "");
Expect(1, 69839, '\P{^inkaithi}', "");
Expect(0, 69840, '\p{inkaithi}', "");
Expect(1, 69840, '\p{^inkaithi}', "");
Expect(1, 69840, '\P{inkaithi}', "");
Expect(0, 69840, '\P{^inkaithi}', "");
Expect(1, 69839, '\p{--In_Kaithi}', "");
Expect(0, 69839, '\p{^--In_Kaithi}', "");
Expect(0, 69839, '\P{--In_Kaithi}', "");
Expect(1, 69839, '\P{^--In_Kaithi}', "");
Expect(0, 69840, '\p{--In_Kaithi}', "");
Expect(1, 69840, '\p{^--In_Kaithi}', "");
Expect(1, 69840, '\P{--In_Kaithi}', "");
Expect(0, 69840, '\P{^--In_Kaithi}', "");
Error('\p{:=	 In_kaithi}');
Error('\P{:=	 In_kaithi}');
Expect(1, 69839, '\p{--in_kaithi}', "");
Expect(0, 69839, '\p{^--in_kaithi}', "");
Expect(0, 69839, '\P{--in_kaithi}', "");
Expect(1, 69839, '\P{^--in_kaithi}', "");
Expect(0, 69840, '\p{--in_kaithi}', "");
Expect(1, 69840, '\p{^--in_kaithi}', "");
Expect(1, 69840, '\P{--in_kaithi}', "");
Expect(0, 69840, '\P{^--in_kaithi}', "");
Error('\p{ /a/Han}');
Error('\P{ /a/Han}');
Expect(1, 195101, '\p{han}', "");
Expect(0, 195101, '\p{^han}', "");
Expect(0, 195101, '\P{han}', "");
Expect(1, 195101, '\P{^han}', "");
Expect(0, 195102, '\p{han}', "");
Expect(1, 195102, '\p{^han}', "");
Expect(1, 195102, '\P{han}', "");
Expect(0, 195102, '\P{^han}', "");
Expect(1, 195101, '\p{-Han}', "");
Expect(0, 195101, '\p{^-Han}', "");
Expect(0, 195101, '\P{-Han}', "");
Expect(1, 195101, '\P{^-Han}', "");
Expect(0, 195102, '\p{-Han}', "");
Expect(1, 195102, '\p{^-Han}', "");
Expect(1, 195102, '\P{-Han}', "");
Expect(0, 195102, '\P{^-Han}', "");
Error('\p{_-Is_Han/a/}');
Error('\P{_-Is_Han/a/}');
Expect(1, 195101, '\p{ishan}', "");
Expect(0, 195101, '\p{^ishan}', "");
Expect(0, 195101, '\P{ishan}', "");
Expect(1, 195101, '\P{^ishan}', "");
Expect(0, 195102, '\p{ishan}', "");
Expect(1, 195102, '\p{^ishan}', "");
Expect(1, 195102, '\P{ishan}', "");
Expect(0, 195102, '\P{^ishan}', "");
Expect(1, 195101, '\p{_-is_Han}', "");
Expect(0, 195101, '\p{^_-is_Han}', "");
Expect(0, 195101, '\P{_-is_Han}', "");
Expect(1, 195101, '\P{^_-is_Han}', "");
Expect(0, 195102, '\p{_-is_Han}', "");
Expect(1, 195102, '\p{^_-is_Han}', "");
Expect(1, 195102, '\P{_-is_Han}', "");
Expect(0, 195102, '\P{^_-is_Han}', "");
Error('\p{ :=HANI}');
Error('\P{ :=HANI}');
Expect(1, 195101, '\p{hani}', "");
Expect(0, 195101, '\p{^hani}', "");
Expect(0, 195101, '\P{hani}', "");
Expect(1, 195101, '\P{^hani}', "");
Expect(0, 195102, '\p{hani}', "");
Expect(1, 195102, '\p{^hani}', "");
Expect(1, 195102, '\P{hani}', "");
Expect(0, 195102, '\P{^hani}', "");
Expect(1, 195101, '\p{-_Hani}', "");
Expect(0, 195101, '\p{^-_Hani}', "");
Expect(0, 195101, '\P{-_Hani}', "");
Expect(1, 195101, '\P{^-_Hani}', "");
Expect(0, 195102, '\p{-_Hani}', "");
Expect(1, 195102, '\p{^-_Hani}', "");
Expect(1, 195102, '\P{-_Hani}', "");
Expect(0, 195102, '\P{^-_Hani}', "");
Error('\p{-:=Is_hani}');
Error('\P{-:=Is_hani}');
Expect(1, 195101, '\p{ishani}', "");
Expect(0, 195101, '\p{^ishani}', "");
Expect(0, 195101, '\P{ishani}', "");
Expect(1, 195101, '\P{^ishani}', "");
Expect(0, 195102, '\p{ishani}', "");
Expect(1, 195102, '\p{^ishani}', "");
Expect(1, 195102, '\P{ishani}', "");
Expect(0, 195102, '\P{^ishani}', "");
Expect(1, 195101, '\p{-Is_hani}', "");
Expect(0, 195101, '\p{^-Is_hani}', "");
Expect(0, 195101, '\P{-Is_hani}', "");
Expect(1, 195101, '\P{^-Is_hani}', "");
Expect(0, 195102, '\p{-Is_hani}', "");
Expect(1, 195102, '\p{^-Is_hani}', "");
Expect(1, 195102, '\P{-Is_hani}', "");
Expect(0, 195102, '\P{^-Is_hani}', "");
Error('\p{:=- HANUNOO}');
Error('\P{:=- HANUNOO}');
Expect(1, 5940, '\p{hanunoo}', "");
Expect(0, 5940, '\p{^hanunoo}', "");
Expect(0, 5940, '\P{hanunoo}', "");
Expect(1, 5940, '\P{^hanunoo}', "");
Expect(0, 5941, '\p{hanunoo}', "");
Expect(1, 5941, '\p{^hanunoo}', "");
Expect(1, 5941, '\P{hanunoo}', "");
Expect(0, 5941, '\P{^hanunoo}', "");
Expect(1, 5940, '\p{-	hanunoo}', "");
Expect(0, 5940, '\p{^-	hanunoo}', "");
Expect(0, 5940, '\P{-	hanunoo}', "");
Expect(1, 5940, '\P{^-	hanunoo}', "");
Expect(0, 5941, '\p{-	hanunoo}', "");
Expect(1, 5941, '\p{^-	hanunoo}', "");
Expect(1, 5941, '\P{-	hanunoo}', "");
Expect(0, 5941, '\P{^-	hanunoo}', "");
Error('\p{/a/_	Is_HANUNOO}');
Error('\P{/a/_	Is_HANUNOO}');
Expect(1, 5940, '\p{ishanunoo}', "");
Expect(0, 5940, '\p{^ishanunoo}', "");
Expect(0, 5940, '\P{ishanunoo}', "");
Expect(1, 5940, '\P{^ishanunoo}', "");
Expect(0, 5941, '\p{ishanunoo}', "");
Expect(1, 5941, '\p{^ishanunoo}', "");
Expect(1, 5941, '\P{ishanunoo}', "");
Expect(0, 5941, '\P{^ishanunoo}', "");
Expect(1, 5940, '\p{- IS_hanunoo}', "");
Expect(0, 5940, '\p{^- IS_hanunoo}', "");
Expect(0, 5940, '\P{- IS_hanunoo}', "");
Expect(1, 5940, '\P{^- IS_hanunoo}', "");
Expect(0, 5941, '\p{- IS_hanunoo}', "");
Expect(1, 5941, '\p{^- IS_hanunoo}', "");
Expect(1, 5941, '\P{- IS_hanunoo}', "");
Expect(0, 5941, '\P{^- IS_hanunoo}', "");
Error('\p{-/a/HANO}');
Error('\P{-/a/HANO}');
Expect(1, 5940, '\p{hano}', "");
Expect(0, 5940, '\p{^hano}', "");
Expect(0, 5940, '\P{hano}', "");
Expect(1, 5940, '\P{^hano}', "");
Expect(0, 5941, '\p{hano}', "");
Expect(1, 5941, '\p{^hano}', "");
Expect(1, 5941, '\P{hano}', "");
Expect(0, 5941, '\P{^hano}', "");
Expect(1, 5940, '\p{_HANO}', "");
Expect(0, 5940, '\p{^_HANO}', "");
Expect(0, 5940, '\P{_HANO}', "");
Expect(1, 5940, '\P{^_HANO}', "");
Expect(0, 5941, '\p{_HANO}', "");
Expect(1, 5941, '\p{^_HANO}', "");
Expect(1, 5941, '\P{_HANO}', "");
Expect(0, 5941, '\P{^_HANO}', "");
Error('\p{:=Is_Hano}');
Error('\P{:=Is_Hano}');
Expect(1, 5940, '\p{ishano}', "");
Expect(0, 5940, '\p{^ishano}', "");
Expect(0, 5940, '\P{ishano}', "");
Expect(1, 5940, '\P{^ishano}', "");
Expect(0, 5941, '\p{ishano}', "");
Expect(1, 5941, '\p{^ishano}', "");
Expect(1, 5941, '\P{ishano}', "");
Expect(0, 5941, '\P{^ishano}', "");
Expect(1, 5940, '\p{--is_Hano}', "");
Expect(0, 5940, '\p{^--is_Hano}', "");
Expect(0, 5940, '\P{--is_Hano}', "");
Expect(1, 5940, '\P{^--is_Hano}', "");
Expect(0, 5941, '\p{--is_Hano}', "");
Expect(1, 5941, '\p{^--is_Hano}', "");
Expect(1, 5941, '\P{--is_Hano}', "");
Expect(0, 5941, '\P{^--is_Hano}', "");
Error('\p{_-In_Meetei_Mayek/a/}');
Error('\P{_-In_Meetei_Mayek/a/}');
Expect(1, 44031, '\p{inmeeteimayek}', "");
Expect(0, 44031, '\p{^inmeeteimayek}', "");
Expect(0, 44031, '\P{inmeeteimayek}', "");
Expect(1, 44031, '\P{^inmeeteimayek}', "");
Expect(0, 44032, '\p{inmeeteimayek}', "");
Expect(1, 44032, '\p{^inmeeteimayek}', "");
Expect(1, 44032, '\P{inmeeteimayek}', "");
Expect(0, 44032, '\P{^inmeeteimayek}', "");
Expect(1, 44031, '\p{_in_meetei_Mayek}', "");
Expect(0, 44031, '\p{^_in_meetei_Mayek}', "");
Expect(0, 44031, '\P{_in_meetei_Mayek}', "");
Expect(1, 44031, '\P{^_in_meetei_Mayek}', "");
Expect(0, 44032, '\p{_in_meetei_Mayek}', "");
Expect(1, 44032, '\p{^_in_meetei_Mayek}', "");
Expect(1, 44032, '\P{_in_meetei_Mayek}', "");
Expect(0, 44032, '\P{^_in_meetei_Mayek}', "");
Error('\p{:=	IN_Meetei_Mayek}');
Error('\P{:=	IN_Meetei_Mayek}');
Expect(1, 44031, '\p{-In_MEETEI_MAYEK}', "");
Expect(0, 44031, '\p{^-In_MEETEI_MAYEK}', "");
Expect(0, 44031, '\P{-In_MEETEI_MAYEK}', "");
Expect(1, 44031, '\P{^-In_MEETEI_MAYEK}', "");
Expect(0, 44032, '\p{-In_MEETEI_MAYEK}', "");
Expect(1, 44032, '\p{^-In_MEETEI_MAYEK}', "");
Expect(1, 44032, '\P{-In_MEETEI_MAYEK}', "");
Expect(0, 44032, '\P{^-In_MEETEI_MAYEK}', "");
Error('\p{/a/_	VertSpace}');
Error('\P{/a/_	VertSpace}');
Expect(1, 8233, '\p{vertspace}', "");
Expect(0, 8233, '\p{^vertspace}', "");
Expect(0, 8233, '\P{vertspace}', "");
Expect(1, 8233, '\P{^vertspace}', "");
Expect(0, 8234, '\p{vertspace}', "");
Expect(1, 8234, '\p{^vertspace}', "");
Expect(1, 8234, '\P{vertspace}', "");
Expect(0, 8234, '\P{^vertspace}', "");
Expect(1, 8233, '\p{  VertSpace}', "");
Expect(0, 8233, '\p{^  VertSpace}', "");
Expect(0, 8233, '\P{  VertSpace}', "");
Expect(1, 8233, '\P{^  VertSpace}', "");
Expect(0, 8234, '\p{  VertSpace}', "");
Expect(1, 8234, '\p{^  VertSpace}', "");
Expect(1, 8234, '\P{  VertSpace}', "");
Expect(0, 8234, '\P{^  VertSpace}', "");
Error('\p{_-IS_VertSpace:=}');
Error('\P{_-IS_VertSpace:=}');
Expect(1, 8233, '\p{isvertspace}', "");
Expect(0, 8233, '\p{^isvertspace}', "");
Expect(0, 8233, '\P{isvertspace}', "");
Expect(1, 8233, '\P{^isvertspace}', "");
Expect(0, 8234, '\p{isvertspace}', "");
Expect(1, 8234, '\p{^isvertspace}', "");
Expect(1, 8234, '\P{isvertspace}', "");
Expect(0, 8234, '\P{^isvertspace}', "");
Expect(1, 8233, '\p{ IS_vertspace}', "");
Expect(0, 8233, '\p{^ IS_vertspace}', "");
Expect(0, 8233, '\P{ IS_vertspace}', "");
Expect(1, 8233, '\P{^ IS_vertspace}', "");
Expect(0, 8234, '\p{ IS_vertspace}', "");
Expect(1, 8234, '\p{^ IS_vertspace}', "");
Expect(1, 8234, '\P{ IS_vertspace}', "");
Expect(0, 8234, '\P{^ IS_vertspace}', "");
Error('\p{:=  Letterlike_symbols}');
Error('\P{:=  Letterlike_symbols}');
Expect(1, 8527, '\p{letterlikesymbols}', "");
Expect(0, 8527, '\p{^letterlikesymbols}', "");
Expect(0, 8527, '\P{letterlikesymbols}', "");
Expect(1, 8527, '\P{^letterlikesymbols}', "");
Expect(0, 8528, '\p{letterlikesymbols}', "");
Expect(1, 8528, '\p{^letterlikesymbols}', "");
Expect(1, 8528, '\P{letterlikesymbols}', "");
Expect(0, 8528, '\P{^letterlikesymbols}', "");
Expect(1, 8527, '\p{-Letterlike_SYMBOLS}', "");
Expect(0, 8527, '\p{^-Letterlike_SYMBOLS}', "");
Expect(0, 8527, '\P{-Letterlike_SYMBOLS}', "");
Expect(1, 8527, '\P{^-Letterlike_SYMBOLS}', "");
Expect(0, 8528, '\p{-Letterlike_SYMBOLS}', "");
Expect(1, 8528, '\p{^-Letterlike_SYMBOLS}', "");
Expect(1, 8528, '\P{-Letterlike_SYMBOLS}', "");
Expect(0, 8528, '\P{^-Letterlike_SYMBOLS}', "");
Error('\p{:=-_Is_Letterlike_Symbols}');
Error('\P{:=-_Is_Letterlike_Symbols}');
Expect(1, 8527, '\p{isletterlikesymbols}', "");
Expect(0, 8527, '\p{^isletterlikesymbols}', "");
Expect(0, 8527, '\P{isletterlikesymbols}', "");
Expect(1, 8527, '\P{^isletterlikesymbols}', "");
Expect(0, 8528, '\p{isletterlikesymbols}', "");
Expect(1, 8528, '\p{^isletterlikesymbols}', "");
Expect(1, 8528, '\P{isletterlikesymbols}', "");
Expect(0, 8528, '\P{^isletterlikesymbols}', "");
Expect(1, 8527, '\p{-IS_letterlike_Symbols}', "");
Expect(0, 8527, '\p{^-IS_letterlike_Symbols}', "");
Expect(0, 8527, '\P{-IS_letterlike_Symbols}', "");
Expect(1, 8527, '\P{^-IS_letterlike_Symbols}', "");
Expect(0, 8528, '\p{-IS_letterlike_Symbols}', "");
Expect(1, 8528, '\p{^-IS_letterlike_Symbols}', "");
Expect(1, 8528, '\P{-IS_letterlike_Symbols}', "");
Expect(0, 8528, '\P{^-IS_letterlike_Symbols}', "");
Error('\p{	In_Letterlike_SYMBOLS/a/}');
Error('\P{	In_Letterlike_SYMBOLS/a/}');
Expect(1, 8527, '\p{inletterlikesymbols}', "");
Expect(0, 8527, '\p{^inletterlikesymbols}', "");
Expect(0, 8527, '\P{inletterlikesymbols}', "");
Expect(1, 8527, '\P{^inletterlikesymbols}', "");
Expect(0, 8528, '\p{inletterlikesymbols}', "");
Expect(1, 8528, '\p{^inletterlikesymbols}', "");
Expect(1, 8528, '\P{inletterlikesymbols}', "");
Expect(0, 8528, '\P{^inletterlikesymbols}', "");
Expect(1, 8527, '\p{-in_LETTERLIKE_symbols}', "");
Expect(0, 8527, '\p{^-in_LETTERLIKE_symbols}', "");
Expect(0, 8527, '\P{-in_LETTERLIKE_symbols}', "");
Expect(1, 8527, '\P{^-in_LETTERLIKE_symbols}', "");
Expect(0, 8528, '\p{-in_LETTERLIKE_symbols}', "");
Expect(1, 8528, '\p{^-in_LETTERLIKE_symbols}', "");
Expect(1, 8528, '\P{-in_LETTERLIKE_symbols}', "");
Expect(0, 8528, '\P{^-in_LETTERLIKE_symbols}', "");
Error('\p{_In_mongolian:=}');
Error('\P{_In_mongolian:=}');
Expect(1, 6319, '\p{inmongolian}', "");
Expect(0, 6319, '\p{^inmongolian}', "");
Expect(0, 6319, '\P{inmongolian}', "");
Expect(1, 6319, '\P{^inmongolian}', "");
Expect(0, 6320, '\p{inmongolian}', "");
Expect(1, 6320, '\p{^inmongolian}', "");
Expect(1, 6320, '\P{inmongolian}', "");
Expect(0, 6320, '\P{^inmongolian}', "");
Expect(1, 6319, '\p{	IN_Mongolian}', "");
Expect(0, 6319, '\p{^	IN_Mongolian}', "");
Expect(0, 6319, '\P{	IN_Mongolian}', "");
Expect(1, 6319, '\P{^	IN_Mongolian}', "");
Expect(0, 6320, '\p{	IN_Mongolian}', "");
Expect(1, 6320, '\p{^	IN_Mongolian}', "");
Expect(1, 6320, '\P{	IN_Mongolian}', "");
Expect(0, 6320, '\P{^	IN_Mongolian}', "");
Error('\p{/a/-	In_Mongolian}');
Error('\P{/a/-	In_Mongolian}');
Expect(1, 6319, '\p{--IN_Mongolian}', "");
Expect(0, 6319, '\p{^--IN_Mongolian}', "");
Expect(0, 6319, '\P{--IN_Mongolian}', "");
Expect(1, 6319, '\P{^--IN_Mongolian}', "");
Expect(0, 6320, '\p{--IN_Mongolian}', "");
Expect(1, 6320, '\p{^--IN_Mongolian}', "");
Expect(1, 6320, '\P{--IN_Mongolian}', "");
Expect(0, 6320, '\P{^--IN_Mongolian}', "");
Error('\p{/a/ _kangxi_Radicals}');
Error('\P{/a/ _kangxi_Radicals}');
Expect(1, 12255, '\p{kangxiradicals}', "");
Expect(0, 12255, '\p{^kangxiradicals}', "");
Expect(0, 12255, '\P{kangxiradicals}', "");
Expect(1, 12255, '\P{^kangxiradicals}', "");
Expect(0, 12256, '\p{kangxiradicals}', "");
Expect(1, 12256, '\p{^kangxiradicals}', "");
Expect(1, 12256, '\P{kangxiradicals}', "");
Expect(0, 12256, '\P{^kangxiradicals}', "");
Expect(1, 12255, '\p{ KANGXI_radicals}', "");
Expect(0, 12255, '\p{^ KANGXI_radicals}', "");
Expect(0, 12255, '\P{ KANGXI_radicals}', "");
Expect(1, 12255, '\P{^ KANGXI_radicals}', "");
Expect(0, 12256, '\p{ KANGXI_radicals}', "");
Expect(1, 12256, '\p{^ KANGXI_radicals}', "");
Expect(1, 12256, '\P{ KANGXI_radicals}', "");
Expect(0, 12256, '\P{^ KANGXI_radicals}', "");
Error('\p{-IS_KANGXI_radicals/a/}');
Error('\P{-IS_KANGXI_radicals/a/}');
Expect(1, 12255, '\p{iskangxiradicals}', "");
Expect(0, 12255, '\p{^iskangxiradicals}', "");
Expect(0, 12255, '\P{iskangxiradicals}', "");
Expect(1, 12255, '\P{^iskangxiradicals}', "");
Expect(0, 12256, '\p{iskangxiradicals}', "");
Expect(1, 12256, '\p{^iskangxiradicals}', "");
Expect(1, 12256, '\P{iskangxiradicals}', "");
Expect(0, 12256, '\P{^iskangxiradicals}', "");
Expect(1, 12255, '\p{-Is_kangxi_RADICALS}', "");
Expect(0, 12255, '\p{^-Is_kangxi_RADICALS}', "");
Expect(0, 12255, '\P{-Is_kangxi_RADICALS}', "");
Expect(1, 12255, '\P{^-Is_kangxi_RADICALS}', "");
Expect(0, 12256, '\p{-Is_kangxi_RADICALS}', "");
Expect(1, 12256, '\p{^-Is_kangxi_RADICALS}', "");
Expect(1, 12256, '\P{-Is_kangxi_RADICALS}', "");
Expect(0, 12256, '\P{^-Is_kangxi_RADICALS}', "");
Error('\p{/a/  IN_kangxi_Radicals}');
Error('\P{/a/  IN_kangxi_Radicals}');
Expect(1, 12255, '\p{inkangxiradicals}', "");
Expect(0, 12255, '\p{^inkangxiradicals}', "");
Expect(0, 12255, '\P{inkangxiradicals}', "");
Expect(1, 12255, '\P{^inkangxiradicals}', "");
Expect(0, 12256, '\p{inkangxiradicals}', "");
Expect(1, 12256, '\p{^inkangxiradicals}', "");
Expect(1, 12256, '\P{inkangxiradicals}', "");
Expect(0, 12256, '\P{^inkangxiradicals}', "");
Expect(1, 12255, '\p{_IN_Kangxi_Radicals}', "");
Expect(0, 12255, '\p{^_IN_Kangxi_Radicals}', "");
Expect(0, 12255, '\P{_IN_Kangxi_Radicals}', "");
Expect(1, 12255, '\P{^_IN_Kangxi_Radicals}', "");
Expect(0, 12256, '\p{_IN_Kangxi_Radicals}', "");
Expect(1, 12256, '\p{^_IN_Kangxi_Radicals}', "");
Expect(1, 12256, '\P{_IN_Kangxi_Radicals}', "");
Expect(0, 12256, '\P{^_IN_Kangxi_Radicals}', "");
Error('\p{ Kangxi:=}');
Error('\P{ Kangxi:=}');
Expect(1, 12255, '\p{kangxi}', "");
Expect(0, 12255, '\p{^kangxi}', "");
Expect(0, 12255, '\P{kangxi}', "");
Expect(1, 12255, '\P{^kangxi}', "");
Expect(0, 12256, '\p{kangxi}', "");
Expect(1, 12256, '\p{^kangxi}', "");
Expect(1, 12256, '\P{kangxi}', "");
Expect(0, 12256, '\P{^kangxi}', "");
Expect(1, 12255, '\p{	Kangxi}', "");
Expect(0, 12255, '\p{^	Kangxi}', "");
Expect(0, 12255, '\P{	Kangxi}', "");
Expect(1, 12255, '\P{^	Kangxi}', "");
Expect(0, 12256, '\p{	Kangxi}', "");
Expect(1, 12256, '\p{^	Kangxi}', "");
Expect(1, 12256, '\P{	Kangxi}', "");
Expect(0, 12256, '\P{^	Kangxi}', "");
Error('\p{ Is_Kangxi:=}');
Error('\P{ Is_Kangxi:=}');
Expect(1, 12255, '\p{iskangxi}', "");
Expect(0, 12255, '\p{^iskangxi}', "");
Expect(0, 12255, '\P{iskangxi}', "");
Expect(1, 12255, '\P{^iskangxi}', "");
Expect(0, 12256, '\p{iskangxi}', "");
Expect(1, 12256, '\p{^iskangxi}', "");
Expect(1, 12256, '\P{iskangxi}', "");
Expect(0, 12256, '\P{^iskangxi}', "");
Expect(1, 12255, '\p{-	Is_Kangxi}', "");
Expect(0, 12255, '\p{^-	Is_Kangxi}', "");
Expect(0, 12255, '\P{-	Is_Kangxi}', "");
Expect(1, 12255, '\P{^-	Is_Kangxi}', "");
Expect(0, 12256, '\p{-	Is_Kangxi}', "");
Expect(1, 12256, '\p{^-	Is_Kangxi}', "");
Expect(1, 12256, '\P{-	Is_Kangxi}', "");
Expect(0, 12256, '\P{^-	Is_Kangxi}', "");
Error('\p{  in_Kangxi/a/}');
Error('\P{  in_Kangxi/a/}');
Expect(1, 12255, '\p{inkangxi}', "");
Expect(0, 12255, '\p{^inkangxi}', "");
Expect(0, 12255, '\P{inkangxi}', "");
Expect(1, 12255, '\P{^inkangxi}', "");
Expect(0, 12256, '\p{inkangxi}', "");
Expect(1, 12256, '\p{^inkangxi}', "");
Expect(1, 12256, '\P{inkangxi}', "");
Expect(0, 12256, '\P{^inkangxi}', "");
Expect(1, 12255, '\p{ 	IN_kangxi}', "");
Expect(0, 12255, '\p{^ 	IN_kangxi}', "");
Expect(0, 12255, '\P{ 	IN_kangxi}', "");
Expect(1, 12255, '\P{^ 	IN_kangxi}', "");
Expect(0, 12256, '\p{ 	IN_kangxi}', "");
Expect(1, 12256, '\p{^ 	IN_kangxi}', "");
Expect(1, 12256, '\P{ 	IN_kangxi}', "");
Expect(0, 12256, '\P{^ 	IN_kangxi}', "");
Error('\p{:=_IDEOGRAPHIC}');
Error('\P{:=_IDEOGRAPHIC}');
Expect(1, 195101, '\p{ideographic}', "");
Expect(0, 195101, '\p{^ideographic}', "");
Expect(0, 195101, '\P{ideographic}', "");
Expect(1, 195101, '\P{^ideographic}', "");
Expect(0, 195102, '\p{ideographic}', "");
Expect(1, 195102, '\p{^ideographic}', "");
Expect(1, 195102, '\P{ideographic}', "");
Expect(0, 195102, '\P{^ideographic}', "");
Expect(1, 195101, '\p{ _ideographic}', "");
Expect(0, 195101, '\p{^ _ideographic}', "");
Expect(0, 195101, '\P{ _ideographic}', "");
Expect(1, 195101, '\P{^ _ideographic}', "");
Expect(0, 195102, '\p{ _ideographic}', "");
Expect(1, 195102, '\p{^ _ideographic}', "");
Expect(1, 195102, '\P{ _ideographic}', "");
Expect(0, 195102, '\P{^ _ideographic}', "");
Error('\p{:=_-is_Ideographic}');
Error('\P{:=_-is_Ideographic}');
Expect(1, 195101, '\p{isideographic}', "");
Expect(0, 195101, '\p{^isideographic}', "");
Expect(0, 195101, '\P{isideographic}', "");
Expect(1, 195101, '\P{^isideographic}', "");
Expect(0, 195102, '\p{isideographic}', "");
Expect(1, 195102, '\p{^isideographic}', "");
Expect(1, 195102, '\P{isideographic}', "");
Expect(0, 195102, '\P{^isideographic}', "");
Expect(1, 195101, '\p{		IS_Ideographic}', "");
Expect(0, 195101, '\p{^		IS_Ideographic}', "");
Expect(0, 195101, '\P{		IS_Ideographic}', "");
Expect(1, 195101, '\P{^		IS_Ideographic}', "");
Expect(0, 195102, '\p{		IS_Ideographic}', "");
Expect(1, 195102, '\p{^		IS_Ideographic}', "");
Expect(1, 195102, '\P{		IS_Ideographic}', "");
Expect(0, 195102, '\P{^		IS_Ideographic}', "");
Error('\p{Ideo:=}');
Error('\P{Ideo:=}');
Expect(1, 195101, '\p{ideo}', "");
Expect(0, 195101, '\p{^ideo}', "");
Expect(0, 195101, '\P{ideo}', "");
Expect(1, 195101, '\P{^ideo}', "");
Expect(0, 195102, '\p{ideo}', "");
Expect(1, 195102, '\p{^ideo}', "");
Expect(1, 195102, '\P{ideo}', "");
Expect(0, 195102, '\P{^ideo}', "");
Expect(1, 195101, '\p{	-Ideo}', "");
Expect(0, 195101, '\p{^	-Ideo}', "");
Expect(0, 195101, '\P{	-Ideo}', "");
Expect(1, 195101, '\P{^	-Ideo}', "");
Expect(0, 195102, '\p{	-Ideo}', "");
Expect(1, 195102, '\p{^	-Ideo}', "");
Expect(1, 195102, '\P{	-Ideo}', "");
Expect(0, 195102, '\P{^	-Ideo}', "");
Error('\p{ _IS_Ideo:=}');
Error('\P{ _IS_Ideo:=}');
Expect(1, 195101, '\p{isideo}', "");
Expect(0, 195101, '\p{^isideo}', "");
Expect(0, 195101, '\P{isideo}', "");
Expect(1, 195101, '\P{^isideo}', "");
Expect(0, 195102, '\p{isideo}', "");
Expect(1, 195102, '\p{^isideo}', "");
Expect(1, 195102, '\P{isideo}', "");
Expect(0, 195102, '\P{^isideo}', "");
Expect(1, 195101, '\p{	Is_ideo}', "");
Expect(0, 195101, '\p{^	Is_ideo}', "");
Expect(0, 195101, '\P{	Is_ideo}', "");
Expect(1, 195101, '\P{^	Is_ideo}', "");
Expect(0, 195102, '\p{	Is_ideo}', "");
Expect(1, 195102, '\p{^	Is_ideo}', "");
Expect(1, 195102, '\P{	Is_ideo}', "");
Expect(0, 195102, '\P{^	Is_ideo}', "");
Error('\p{_chakma/a/}');
Error('\P{_chakma/a/}');
Expect(1, 69955, '\p{chakma}', "");
Expect(0, 69955, '\p{^chakma}', "");
Expect(0, 69955, '\P{chakma}', "");
Expect(1, 69955, '\P{^chakma}', "");
Expect(0, 69956, '\p{chakma}', "");
Expect(1, 69956, '\p{^chakma}', "");
Expect(1, 69956, '\P{chakma}', "");
Expect(0, 69956, '\P{^chakma}', "");
Expect(1, 69955, '\p{__CHAKMA}', "");
Expect(0, 69955, '\p{^__CHAKMA}', "");
Expect(0, 69955, '\P{__CHAKMA}', "");
Expect(1, 69955, '\P{^__CHAKMA}', "");
Expect(0, 69956, '\p{__CHAKMA}', "");
Expect(1, 69956, '\p{^__CHAKMA}', "");
Expect(1, 69956, '\P{__CHAKMA}', "");
Expect(0, 69956, '\P{^__CHAKMA}', "");
Error('\p{ :=Is_Chakma}');
Error('\P{ :=Is_Chakma}');
Expect(1, 69955, '\p{ischakma}', "");
Expect(0, 69955, '\p{^ischakma}', "");
Expect(0, 69955, '\P{ischakma}', "");
Expect(1, 69955, '\P{^ischakma}', "");
Expect(0, 69956, '\p{ischakma}', "");
Expect(1, 69956, '\p{^ischakma}', "");
Expect(1, 69956, '\P{ischakma}', "");
Expect(0, 69956, '\P{^ischakma}', "");
Expect(1, 69955, '\p{  IS_Chakma}', "");
Expect(0, 69955, '\p{^  IS_Chakma}', "");
Expect(0, 69955, '\P{  IS_Chakma}', "");
Expect(1, 69955, '\P{^  IS_Chakma}', "");
Expect(0, 69956, '\p{  IS_Chakma}', "");
Expect(1, 69956, '\p{^  IS_Chakma}', "");
Expect(1, 69956, '\P{  IS_Chakma}', "");
Expect(0, 69956, '\P{^  IS_Chakma}', "");
Error('\p{:=_cakm}');
Error('\P{:=_cakm}');
Expect(1, 69955, '\p{cakm}', "");
Expect(0, 69955, '\p{^cakm}', "");
Expect(0, 69955, '\P{cakm}', "");
Expect(1, 69955, '\P{^cakm}', "");
Expect(0, 69956, '\p{cakm}', "");
Expect(1, 69956, '\p{^cakm}', "");
Expect(1, 69956, '\P{cakm}', "");
Expect(0, 69956, '\P{^cakm}', "");
Expect(1, 69955, '\p{_cakm}', "");
Expect(0, 69955, '\p{^_cakm}', "");
Expect(0, 69955, '\P{_cakm}', "");
Expect(1, 69955, '\P{^_cakm}', "");
Expect(0, 69956, '\p{_cakm}', "");
Expect(1, 69956, '\p{^_cakm}', "");
Expect(1, 69956, '\P{_cakm}', "");
Expect(0, 69956, '\P{^_cakm}', "");
Error('\p{:=_IS_Cakm}');
Error('\P{:=_IS_Cakm}');
Expect(1, 69955, '\p{iscakm}', "");
Expect(0, 69955, '\p{^iscakm}', "");
Expect(0, 69955, '\P{iscakm}', "");
Expect(1, 69955, '\P{^iscakm}', "");
Expect(0, 69956, '\p{iscakm}', "");
Expect(1, 69956, '\p{^iscakm}', "");
Expect(1, 69956, '\P{iscakm}', "");
Expect(0, 69956, '\P{^iscakm}', "");
Expect(1, 69955, '\p{_ Is_CAKM}', "");
Expect(0, 69955, '\p{^_ Is_CAKM}', "");
Expect(0, 69955, '\P{_ Is_CAKM}', "");
Expect(1, 69955, '\P{^_ Is_CAKM}', "");
Expect(0, 69956, '\p{_ Is_CAKM}', "");
Expect(1, 69956, '\p{^_ Is_CAKM}', "");
Expect(1, 69956, '\P{_ Is_CAKM}', "");
Expect(0, 69956, '\P{^_ Is_CAKM}', "");
Error('\p{_:=Variation_Selectors}');
Error('\P{_:=Variation_Selectors}');
Expect(1, 65039, '\p{variationselectors}', "");
Expect(0, 65039, '\p{^variationselectors}', "");
Expect(0, 65039, '\P{variationselectors}', "");
Expect(1, 65039, '\P{^variationselectors}', "");
Expect(0, 65040, '\p{variationselectors}', "");
Expect(1, 65040, '\p{^variationselectors}', "");
Expect(1, 65040, '\P{variationselectors}', "");
Expect(0, 65040, '\P{^variationselectors}', "");
Expect(1, 65039, '\p{-	Variation_Selectors}', "");
Expect(0, 65039, '\p{^-	Variation_Selectors}', "");
Expect(0, 65039, '\P{-	Variation_Selectors}', "");
Expect(1, 65039, '\P{^-	Variation_Selectors}', "");
Expect(0, 65040, '\p{-	Variation_Selectors}', "");
Expect(1, 65040, '\p{^-	Variation_Selectors}', "");
Expect(1, 65040, '\P{-	Variation_Selectors}', "");
Expect(0, 65040, '\P{^-	Variation_Selectors}', "");
Error('\p{/a/-IS_variation_Selectors}');
Error('\P{/a/-IS_variation_Selectors}');
Expect(1, 65039, '\p{isvariationselectors}', "");
Expect(0, 65039, '\p{^isvariationselectors}', "");
Expect(0, 65039, '\P{isvariationselectors}', "");
Expect(1, 65039, '\P{^isvariationselectors}', "");
Expect(0, 65040, '\p{isvariationselectors}', "");
Expect(1, 65040, '\p{^isvariationselectors}', "");
Expect(1, 65040, '\P{isvariationselectors}', "");
Expect(0, 65040, '\P{^isvariationselectors}', "");
Expect(1, 65039, '\p{	is_variation_selectors}', "");
Expect(0, 65039, '\p{^	is_variation_selectors}', "");
Expect(0, 65039, '\P{	is_variation_selectors}', "");
Expect(1, 65039, '\P{^	is_variation_selectors}', "");
Expect(0, 65040, '\p{	is_variation_selectors}', "");
Expect(1, 65040, '\p{^	is_variation_selectors}', "");
Expect(1, 65040, '\P{	is_variation_selectors}', "");
Expect(0, 65040, '\P{^	is_variation_selectors}', "");
Error('\p{	_In_Variation_SELECTORS/a/}');
Error('\P{	_In_Variation_SELECTORS/a/}');
Expect(1, 65039, '\p{invariationselectors}', "");
Expect(0, 65039, '\p{^invariationselectors}', "");
Expect(0, 65039, '\P{invariationselectors}', "");
Expect(1, 65039, '\P{^invariationselectors}', "");
Expect(0, 65040, '\p{invariationselectors}', "");
Expect(1, 65040, '\p{^invariationselectors}', "");
Expect(1, 65040, '\P{invariationselectors}', "");
Expect(0, 65040, '\P{^invariationselectors}', "");
Expect(1, 65039, '\p{	_in_Variation_Selectors}', "");
Expect(0, 65039, '\p{^	_in_Variation_Selectors}', "");
Expect(0, 65039, '\P{	_in_Variation_Selectors}', "");
Expect(1, 65039, '\P{^	_in_Variation_Selectors}', "");
Expect(0, 65040, '\p{	_in_Variation_Selectors}', "");
Expect(1, 65040, '\p{^	_in_Variation_Selectors}', "");
Expect(1, 65040, '\P{	_in_Variation_Selectors}', "");
Expect(0, 65040, '\P{^	_in_Variation_Selectors}', "");
Error('\p{_:=In_VS}');
Error('\P{_:=In_VS}');
Expect(1, 65039, '\p{invs}', "");
Expect(0, 65039, '\p{^invs}', "");
Expect(0, 65039, '\P{invs}', "");
Expect(1, 65039, '\P{^invs}', "");
Expect(0, 65040, '\p{invs}', "");
Expect(1, 65040, '\p{^invs}', "");
Expect(1, 65040, '\P{invs}', "");
Expect(0, 65040, '\P{^invs}', "");
Expect(1, 65039, '\p{In_VS}', "");
Expect(0, 65039, '\p{^In_VS}', "");
Expect(0, 65039, '\P{In_VS}', "");
Expect(1, 65039, '\P{^In_VS}', "");
Expect(0, 65040, '\p{In_VS}', "");
Expect(1, 65040, '\p{^In_VS}', "");
Expect(1, 65040, '\P{In_VS}', "");
Expect(0, 65040, '\P{^In_VS}', "");
Error('\p{/a/_	In_thaana}');
Error('\P{/a/_	In_thaana}');
Expect(1, 1983, '\p{inthaana}', "");
Expect(0, 1983, '\p{^inthaana}', "");
Expect(0, 1983, '\P{inthaana}', "");
Expect(1, 1983, '\P{^inthaana}', "");
Expect(0, 1984, '\p{inthaana}', "");
Expect(1, 1984, '\p{^inthaana}', "");
Expect(1, 1984, '\P{inthaana}', "");
Expect(0, 1984, '\P{^inthaana}', "");
Expect(1, 1983, '\p{_in_THAANA}', "");
Expect(0, 1983, '\p{^_in_THAANA}', "");
Expect(0, 1983, '\P{_in_THAANA}', "");
Expect(1, 1983, '\P{^_in_THAANA}', "");
Expect(0, 1984, '\p{_in_THAANA}', "");
Expect(1, 1984, '\p{^_in_THAANA}', "");
Expect(1, 1984, '\P{_in_THAANA}', "");
Expect(0, 1984, '\P{^_in_THAANA}', "");
Error('\p{--In_Thaana/a/}');
Error('\P{--In_Thaana/a/}');
Expect(1, 1983, '\p{_IN_Thaana}', "");
Expect(0, 1983, '\p{^_IN_Thaana}', "");
Expect(0, 1983, '\P{_IN_Thaana}', "");
Expect(1, 1983, '\P{^_IN_Thaana}', "");
Expect(0, 1984, '\p{_IN_Thaana}', "");
Expect(1, 1984, '\p{^_IN_Thaana}', "");
Expect(1, 1984, '\P{_IN_Thaana}', "");
Expect(0, 1984, '\P{^_IN_Thaana}', "");
Error('\p{:=changes_When_Uppercased}');
Error('\P{:=changes_When_Uppercased}');
Expect(1, 66639, '\p{changeswhenuppercased}', "");
Expect(0, 66639, '\p{^changeswhenuppercased}', "");
Expect(0, 66639, '\P{changeswhenuppercased}', "");
Expect(1, 66639, '\P{^changeswhenuppercased}', "");
Expect(0, 66640, '\p{changeswhenuppercased}', "");
Expect(1, 66640, '\p{^changeswhenuppercased}', "");
Expect(1, 66640, '\P{changeswhenuppercased}', "");
Expect(0, 66640, '\P{^changeswhenuppercased}', "");
Expect(1, 66639, '\p{- Changes_when_UPPERCASED}', "");
Expect(0, 66639, '\p{^- Changes_when_UPPERCASED}', "");
Expect(0, 66639, '\P{- Changes_when_UPPERCASED}', "");
Expect(1, 66639, '\P{^- Changes_when_UPPERCASED}', "");
Expect(0, 66640, '\p{- Changes_when_UPPERCASED}', "");
Expect(1, 66640, '\p{^- Changes_when_UPPERCASED}', "");
Expect(1, 66640, '\P{- Changes_when_UPPERCASED}', "");
Expect(0, 66640, '\P{^- Changes_when_UPPERCASED}', "");
Error('\p{/a/_	is_CHANGES_WHEN_Uppercased}');
Error('\P{/a/_	is_CHANGES_WHEN_Uppercased}');
Expect(1, 66639, '\p{ischangeswhenuppercased}', "");
Expect(0, 66639, '\p{^ischangeswhenuppercased}', "");
Expect(0, 66639, '\P{ischangeswhenuppercased}', "");
Expect(1, 66639, '\P{^ischangeswhenuppercased}', "");
Expect(0, 66640, '\p{ischangeswhenuppercased}', "");
Expect(1, 66640, '\p{^ischangeswhenuppercased}', "");
Expect(1, 66640, '\P{ischangeswhenuppercased}', "");
Expect(0, 66640, '\P{^ischangeswhenuppercased}', "");
Expect(1, 66639, '\p{_IS_Changes_WHEN_Uppercased}', "");
Expect(0, 66639, '\p{^_IS_Changes_WHEN_Uppercased}', "");
Expect(0, 66639, '\P{_IS_Changes_WHEN_Uppercased}', "");
Expect(1, 66639, '\P{^_IS_Changes_WHEN_Uppercased}', "");
Expect(0, 66640, '\p{_IS_Changes_WHEN_Uppercased}', "");
Expect(1, 66640, '\p{^_IS_Changes_WHEN_Uppercased}', "");
Expect(1, 66640, '\P{_IS_Changes_WHEN_Uppercased}', "");
Expect(0, 66640, '\P{^_IS_Changes_WHEN_Uppercased}', "");
Error('\p{/a/  CWU}');
Error('\P{/a/  CWU}');
Expect(1, 66639, '\p{cwu}', "");
Expect(0, 66639, '\p{^cwu}', "");
Expect(0, 66639, '\P{cwu}', "");
Expect(1, 66639, '\P{^cwu}', "");
Expect(0, 66640, '\p{cwu}', "");
Expect(1, 66640, '\p{^cwu}', "");
Expect(1, 66640, '\P{cwu}', "");
Expect(0, 66640, '\P{^cwu}', "");
Expect(1, 66639, '\p{_	CWU}', "");
Expect(0, 66639, '\p{^_	CWU}', "");
Expect(0, 66639, '\P{_	CWU}', "");
Expect(1, 66639, '\P{^_	CWU}', "");
Expect(0, 66640, '\p{_	CWU}', "");
Expect(1, 66640, '\p{^_	CWU}', "");
Expect(1, 66640, '\P{_	CWU}', "");
Expect(0, 66640, '\P{^_	CWU}', "");
Error('\p{ 	is_cwu:=}');
Error('\P{ 	is_cwu:=}');
Expect(1, 66639, '\p{iscwu}', "");
Expect(0, 66639, '\p{^iscwu}', "");
Expect(0, 66639, '\P{iscwu}', "");
Expect(1, 66639, '\P{^iscwu}', "");
Expect(0, 66640, '\p{iscwu}', "");
Expect(1, 66640, '\p{^iscwu}', "");
Expect(1, 66640, '\P{iscwu}', "");
Expect(0, 66640, '\P{^iscwu}', "");
Expect(1, 66639, '\p{--Is_CWU}', "");
Expect(0, 66639, '\p{^--Is_CWU}', "");
Expect(0, 66639, '\P{--Is_CWU}', "");
Expect(1, 66639, '\P{^--Is_CWU}', "");
Expect(0, 66640, '\p{--Is_CWU}', "");
Expect(1, 66640, '\p{^--Is_CWU}', "");
Expect(1, 66640, '\P{--Is_CWU}', "");
Expect(0, 66640, '\P{^--Is_CWU}', "");
Error('\p{:=_In_Syriac}');
Error('\P{:=_In_Syriac}');
Expect(1, 1871, '\p{insyriac}', "");
Expect(0, 1871, '\p{^insyriac}', "");
Expect(0, 1871, '\P{insyriac}', "");
Expect(1, 1871, '\P{^insyriac}', "");
Expect(0, 1872, '\p{insyriac}', "");
Expect(1, 1872, '\p{^insyriac}', "");
Expect(1, 1872, '\P{insyriac}', "");
Expect(0, 1872, '\P{^insyriac}', "");
Expect(1, 1871, '\p{ _IN_Syriac}', "");
Expect(0, 1871, '\p{^ _IN_Syriac}', "");
Expect(0, 1871, '\P{ _IN_Syriac}', "");
Expect(1, 1871, '\P{^ _IN_Syriac}', "");
Expect(0, 1872, '\p{ _IN_Syriac}', "");
Expect(1, 1872, '\p{^ _IN_Syriac}', "");
Expect(1, 1872, '\P{ _IN_Syriac}', "");
Expect(0, 1872, '\P{^ _IN_Syriac}', "");
Error('\p{:=_ In_SYRIAC}');
Error('\P{:=_ In_SYRIAC}');
Expect(1, 1871, '\p{_In_syriac}', "");
Expect(0, 1871, '\p{^_In_syriac}', "");
Expect(0, 1871, '\P{_In_syriac}', "");
Expect(1, 1871, '\P{^_In_syriac}', "");
Expect(0, 1872, '\p{_In_syriac}', "");
Expect(1, 1872, '\p{^_In_syriac}', "");
Expect(1, 1872, '\P{_In_syriac}', "");
Expect(0, 1872, '\P{^_In_syriac}', "");
Error('\p{- Miscellaneous_symbols:=}');
Error('\P{- Miscellaneous_symbols:=}');
Expect(1, 9983, '\p{miscellaneoussymbols}', "");
Expect(0, 9983, '\p{^miscellaneoussymbols}', "");
Expect(0, 9983, '\P{miscellaneoussymbols}', "");
Expect(1, 9983, '\P{^miscellaneoussymbols}', "");
Expect(0, 9984, '\p{miscellaneoussymbols}', "");
Expect(1, 9984, '\p{^miscellaneoussymbols}', "");
Expect(1, 9984, '\P{miscellaneoussymbols}', "");
Expect(0, 9984, '\P{^miscellaneoussymbols}', "");
Expect(1, 9983, '\p{ _miscellaneous_symbols}', "");
Expect(0, 9983, '\p{^ _miscellaneous_symbols}', "");
Expect(0, 9983, '\P{ _miscellaneous_symbols}', "");
Expect(1, 9983, '\P{^ _miscellaneous_symbols}', "");
Expect(0, 9984, '\p{ _miscellaneous_symbols}', "");
Expect(1, 9984, '\p{^ _miscellaneous_symbols}', "");
Expect(1, 9984, '\P{ _miscellaneous_symbols}', "");
Expect(0, 9984, '\P{^ _miscellaneous_symbols}', "");
Error('\p{	:=is_Miscellaneous_SYMBOLS}');
Error('\P{	:=is_Miscellaneous_SYMBOLS}');
Expect(1, 9983, '\p{ismiscellaneoussymbols}', "");
Expect(0, 9983, '\p{^ismiscellaneoussymbols}', "");
Expect(0, 9983, '\P{ismiscellaneoussymbols}', "");
Expect(1, 9983, '\P{^ismiscellaneoussymbols}', "");
Expect(0, 9984, '\p{ismiscellaneoussymbols}', "");
Expect(1, 9984, '\p{^ismiscellaneoussymbols}', "");
Expect(1, 9984, '\P{ismiscellaneoussymbols}', "");
Expect(0, 9984, '\P{^ismiscellaneoussymbols}', "");
Expect(1, 9983, '\p{-	is_Miscellaneous_symbols}', "");
Expect(0, 9983, '\p{^-	is_Miscellaneous_symbols}', "");
Expect(0, 9983, '\P{-	is_Miscellaneous_symbols}', "");
Expect(1, 9983, '\P{^-	is_Miscellaneous_symbols}', "");
Expect(0, 9984, '\p{-	is_Miscellaneous_symbols}', "");
Expect(1, 9984, '\p{^-	is_Miscellaneous_symbols}', "");
Expect(1, 9984, '\P{-	is_Miscellaneous_symbols}', "");
Expect(0, 9984, '\P{^-	is_Miscellaneous_symbols}', "");
Error('\p{ :=In_Miscellaneous_SYMBOLS}');
Error('\P{ :=In_Miscellaneous_SYMBOLS}');
Expect(1, 9983, '\p{inmiscellaneoussymbols}', "");
Expect(0, 9983, '\p{^inmiscellaneoussymbols}', "");
Expect(0, 9983, '\P{inmiscellaneoussymbols}', "");
Expect(1, 9983, '\P{^inmiscellaneoussymbols}', "");
Expect(0, 9984, '\p{inmiscellaneoussymbols}', "");
Expect(1, 9984, '\p{^inmiscellaneoussymbols}', "");
Expect(1, 9984, '\P{inmiscellaneoussymbols}', "");
Expect(0, 9984, '\P{^inmiscellaneoussymbols}', "");
Expect(1, 9983, '\p{  In_MISCELLANEOUS_Symbols}', "");
Expect(0, 9983, '\p{^  In_MISCELLANEOUS_Symbols}', "");
Expect(0, 9983, '\P{  In_MISCELLANEOUS_Symbols}', "");
Expect(1, 9983, '\P{^  In_MISCELLANEOUS_Symbols}', "");
Expect(0, 9984, '\p{  In_MISCELLANEOUS_Symbols}', "");
Expect(1, 9984, '\p{^  In_MISCELLANEOUS_Symbols}', "");
Expect(1, 9984, '\P{  In_MISCELLANEOUS_Symbols}', "");
Expect(0, 9984, '\P{^  In_MISCELLANEOUS_Symbols}', "");
Error('\p{	Misc_symbols:=}');
Error('\P{	Misc_symbols:=}');
Expect(1, 9983, '\p{miscsymbols}', "");
Expect(0, 9983, '\p{^miscsymbols}', "");
Expect(0, 9983, '\P{miscsymbols}', "");
Expect(1, 9983, '\P{^miscsymbols}', "");
Expect(0, 9984, '\p{miscsymbols}', "");
Expect(1, 9984, '\p{^miscsymbols}', "");
Expect(1, 9984, '\P{miscsymbols}', "");
Expect(0, 9984, '\P{^miscsymbols}', "");
Expect(1, 9983, '\p{	 misc_Symbols}', "");
Expect(0, 9983, '\p{^	 misc_Symbols}', "");
Expect(0, 9983, '\P{	 misc_Symbols}', "");
Expect(1, 9983, '\P{^	 misc_Symbols}', "");
Expect(0, 9984, '\p{	 misc_Symbols}', "");
Expect(1, 9984, '\p{^	 misc_Symbols}', "");
Expect(1, 9984, '\P{	 misc_Symbols}', "");
Expect(0, 9984, '\P{^	 misc_Symbols}', "");
Error('\p{ /a/Is_MISC_Symbols}');
Error('\P{ /a/Is_MISC_Symbols}');
Expect(1, 9983, '\p{ismiscsymbols}', "");
Expect(0, 9983, '\p{^ismiscsymbols}', "");
Expect(0, 9983, '\P{ismiscsymbols}', "");
Expect(1, 9983, '\P{^ismiscsymbols}', "");
Expect(0, 9984, '\p{ismiscsymbols}', "");
Expect(1, 9984, '\p{^ismiscsymbols}', "");
Expect(1, 9984, '\P{ismiscsymbols}', "");
Expect(0, 9984, '\P{^ismiscsymbols}', "");
Expect(1, 9983, '\p{Is_Misc_Symbols}', "");
Expect(0, 9983, '\p{^Is_Misc_Symbols}', "");
Expect(0, 9983, '\P{Is_Misc_Symbols}', "");
Expect(1, 9983, '\P{^Is_Misc_Symbols}', "");
Expect(0, 9984, '\p{Is_Misc_Symbols}', "");
Expect(1, 9984, '\p{^Is_Misc_Symbols}', "");
Expect(1, 9984, '\P{Is_Misc_Symbols}', "");
Expect(0, 9984, '\P{^Is_Misc_Symbols}', "");
Error('\p{		In_Misc_Symbols/a/}');
Error('\P{		In_Misc_Symbols/a/}');
Expect(1, 9983, '\p{inmiscsymbols}', "");
Expect(0, 9983, '\p{^inmiscsymbols}', "");
Expect(0, 9983, '\P{inmiscsymbols}', "");
Expect(1, 9983, '\P{^inmiscsymbols}', "");
Expect(0, 9984, '\p{inmiscsymbols}', "");
Expect(1, 9984, '\p{^inmiscsymbols}', "");
Expect(1, 9984, '\P{inmiscsymbols}', "");
Expect(0, 9984, '\P{^inmiscsymbols}', "");
Expect(1, 9983, '\p{-	In_Misc_Symbols}', "");
Expect(0, 9983, '\p{^-	In_Misc_Symbols}', "");
Expect(0, 9983, '\P{-	In_Misc_Symbols}', "");
Expect(1, 9983, '\P{^-	In_Misc_Symbols}', "");
Expect(0, 9984, '\p{-	In_Misc_Symbols}', "");
Expect(1, 9984, '\p{^-	In_Misc_Symbols}', "");
Expect(1, 9984, '\P{-	In_Misc_Symbols}', "");
Expect(0, 9984, '\P{^-	In_Misc_Symbols}', "");
Error('\p{ 	PHOENICIAN:=}');
Error('\P{ 	PHOENICIAN:=}');
Expect(1, 67871, '\p{phoenician}', "");
Expect(0, 67871, '\p{^phoenician}', "");
Expect(0, 67871, '\P{phoenician}', "");
Expect(1, 67871, '\P{^phoenician}', "");
Expect(0, 67872, '\p{phoenician}', "");
Expect(1, 67872, '\p{^phoenician}', "");
Expect(1, 67872, '\P{phoenician}', "");
Expect(0, 67872, '\P{^phoenician}', "");
Expect(1, 67871, '\p{	-PHOENICIAN}', "");
Expect(0, 67871, '\p{^	-PHOENICIAN}', "");
Expect(0, 67871, '\P{	-PHOENICIAN}', "");
Expect(1, 67871, '\P{^	-PHOENICIAN}', "");
Expect(0, 67872, '\p{	-PHOENICIAN}', "");
Expect(1, 67872, '\p{^	-PHOENICIAN}', "");
Expect(1, 67872, '\P{	-PHOENICIAN}', "");
Expect(0, 67872, '\P{^	-PHOENICIAN}', "");
Error('\p{/a/_	Is_Phoenician}');
Error('\P{/a/_	Is_Phoenician}');
Expect(1, 67871, '\p{isphoenician}', "");
Expect(0, 67871, '\p{^isphoenician}', "");
Expect(0, 67871, '\P{isphoenician}', "");
Expect(1, 67871, '\P{^isphoenician}', "");
Expect(0, 67872, '\p{isphoenician}', "");
Expect(1, 67872, '\p{^isphoenician}', "");
Expect(1, 67872, '\P{isphoenician}', "");
Expect(0, 67872, '\P{^isphoenician}', "");
Expect(1, 67871, '\p{	-Is_phoenician}', "");
Expect(0, 67871, '\p{^	-Is_phoenician}', "");
Expect(0, 67871, '\P{	-Is_phoenician}', "");
Expect(1, 67871, '\P{^	-Is_phoenician}', "");
Expect(0, 67872, '\p{	-Is_phoenician}', "");
Expect(1, 67872, '\p{^	-Is_phoenician}', "");
Expect(1, 67872, '\P{	-Is_phoenician}', "");
Expect(0, 67872, '\P{^	-Is_phoenician}', "");
Error('\p{-/a/Phnx}');
Error('\P{-/a/Phnx}');
Expect(1, 67871, '\p{phnx}', "");
Expect(0, 67871, '\p{^phnx}', "");
Expect(0, 67871, '\P{phnx}', "");
Expect(1, 67871, '\P{^phnx}', "");
Expect(0, 67872, '\p{phnx}', "");
Expect(1, 67872, '\p{^phnx}', "");
Expect(1, 67872, '\P{phnx}', "");
Expect(0, 67872, '\P{^phnx}', "");
Expect(1, 67871, '\p{		PHNX}', "");
Expect(0, 67871, '\p{^		PHNX}', "");
Expect(0, 67871, '\P{		PHNX}', "");
Expect(1, 67871, '\P{^		PHNX}', "");
Expect(0, 67872, '\p{		PHNX}', "");
Expect(1, 67872, '\p{^		PHNX}', "");
Expect(1, 67872, '\P{		PHNX}', "");
Expect(0, 67872, '\P{^		PHNX}', "");
Error('\p{/a/IS_PHNX}');
Error('\P{/a/IS_PHNX}');
Expect(1, 67871, '\p{isphnx}', "");
Expect(0, 67871, '\p{^isphnx}', "");
Expect(0, 67871, '\P{isphnx}', "");
Expect(1, 67871, '\P{^isphnx}', "");
Expect(0, 67872, '\p{isphnx}', "");
Expect(1, 67872, '\p{^isphnx}', "");
Expect(1, 67872, '\P{isphnx}', "");
Expect(0, 67872, '\P{^isphnx}', "");
Expect(1, 67871, '\p{-Is_Phnx}', "");
Expect(0, 67871, '\p{^-Is_Phnx}', "");
Expect(0, 67871, '\P{-Is_Phnx}', "");
Expect(1, 67871, '\P{^-Is_Phnx}', "");
Expect(0, 67872, '\p{-Is_Phnx}', "");
Expect(1, 67872, '\p{^-Is_Phnx}', "");
Expect(1, 67872, '\P{-Is_Phnx}', "");
Expect(0, 67872, '\P{^-Is_Phnx}', "");
Error('\p{	:=MYANMAR_EXTENDED_a}');
Error('\P{	:=MYANMAR_EXTENDED_a}');
Expect(1, 43647, '\p{myanmarextendeda}', "");
Expect(0, 43647, '\p{^myanmarextendeda}', "");
Expect(0, 43647, '\P{myanmarextendeda}', "");
Expect(1, 43647, '\P{^myanmarextendeda}', "");
Expect(0, 43648, '\p{myanmarextendeda}', "");
Expect(1, 43648, '\p{^myanmarextendeda}', "");
Expect(1, 43648, '\P{myanmarextendeda}', "");
Expect(0, 43648, '\P{^myanmarextendeda}', "");
Expect(1, 43647, '\p{Myanmar_EXTENDED_a}', "");
Expect(0, 43647, '\p{^Myanmar_EXTENDED_a}', "");
Expect(0, 43647, '\P{Myanmar_EXTENDED_a}', "");
Expect(1, 43647, '\P{^Myanmar_EXTENDED_a}', "");
Expect(0, 43648, '\p{Myanmar_EXTENDED_a}', "");
Expect(1, 43648, '\p{^Myanmar_EXTENDED_a}', "");
Expect(1, 43648, '\P{Myanmar_EXTENDED_a}', "");
Expect(0, 43648, '\P{^Myanmar_EXTENDED_a}', "");
Error('\p{ _is_Myanmar_EXTENDED_A:=}');
Error('\P{ _is_Myanmar_EXTENDED_A:=}');
Expect(1, 43647, '\p{ismyanmarextendeda}', "");
Expect(0, 43647, '\p{^ismyanmarextendeda}', "");
Expect(0, 43647, '\P{ismyanmarextendeda}', "");
Expect(1, 43647, '\P{^ismyanmarextendeda}', "");
Expect(0, 43648, '\p{ismyanmarextendeda}', "");
Expect(1, 43648, '\p{^ismyanmarextendeda}', "");
Expect(1, 43648, '\P{ismyanmarextendeda}', "");
Expect(0, 43648, '\P{^ismyanmarextendeda}', "");
Expect(1, 43647, '\p{	 Is_myanmar_Extended_A}', "");
Expect(0, 43647, '\p{^	 Is_myanmar_Extended_A}', "");
Expect(0, 43647, '\P{	 Is_myanmar_Extended_A}', "");
Expect(1, 43647, '\P{^	 Is_myanmar_Extended_A}', "");
Expect(0, 43648, '\p{	 Is_myanmar_Extended_A}', "");
Expect(1, 43648, '\p{^	 Is_myanmar_Extended_A}', "");
Expect(1, 43648, '\P{	 Is_myanmar_Extended_A}', "");
Expect(0, 43648, '\P{^	 Is_myanmar_Extended_A}', "");
Error('\p{ IN_MYANMAR_Extended_A/a/}');
Error('\P{ IN_MYANMAR_Extended_A/a/}');
Expect(1, 43647, '\p{inmyanmarextendeda}', "");
Expect(0, 43647, '\p{^inmyanmarextendeda}', "");
Expect(0, 43647, '\P{inmyanmarextendeda}', "");
Expect(1, 43647, '\P{^inmyanmarextendeda}', "");
Expect(0, 43648, '\p{inmyanmarextendeda}', "");
Expect(1, 43648, '\p{^inmyanmarextendeda}', "");
Expect(1, 43648, '\P{inmyanmarextendeda}', "");
Expect(0, 43648, '\P{^inmyanmarextendeda}', "");
Expect(1, 43647, '\p{IN_myanmar_Extended_A}', "");
Expect(0, 43647, '\p{^IN_myanmar_Extended_A}', "");
Expect(0, 43647, '\P{IN_myanmar_Extended_A}', "");
Expect(1, 43647, '\P{^IN_myanmar_Extended_A}', "");
Expect(0, 43648, '\p{IN_myanmar_Extended_A}', "");
Expect(1, 43648, '\p{^IN_myanmar_Extended_A}', "");
Expect(1, 43648, '\P{IN_myanmar_Extended_A}', "");
Expect(0, 43648, '\P{^IN_myanmar_Extended_A}', "");
Error('\p{_myanmar_ext_A:=}');
Error('\P{_myanmar_ext_A:=}');
Expect(1, 43647, '\p{myanmarexta}', "");
Expect(0, 43647, '\p{^myanmarexta}', "");
Expect(0, 43647, '\P{myanmarexta}', "");
Expect(1, 43647, '\P{^myanmarexta}', "");
Expect(0, 43648, '\p{myanmarexta}', "");
Expect(1, 43648, '\p{^myanmarexta}', "");
Expect(1, 43648, '\P{myanmarexta}', "");
Expect(0, 43648, '\P{^myanmarexta}', "");
Expect(1, 43647, '\p{	-myanmar_Ext_A}', "");
Expect(0, 43647, '\p{^	-myanmar_Ext_A}', "");
Expect(0, 43647, '\P{	-myanmar_Ext_A}', "");
Expect(1, 43647, '\P{^	-myanmar_Ext_A}', "");
Expect(0, 43648, '\p{	-myanmar_Ext_A}', "");
Expect(1, 43648, '\p{^	-myanmar_Ext_A}', "");
Expect(1, 43648, '\P{	-myanmar_Ext_A}', "");
Expect(0, 43648, '\P{^	-myanmar_Ext_A}', "");
Error('\p{ :=IS_myanmar_Ext_A}');
Error('\P{ :=IS_myanmar_Ext_A}');
Expect(1, 43647, '\p{ismyanmarexta}', "");
Expect(0, 43647, '\p{^ismyanmarexta}', "");
Expect(0, 43647, '\P{ismyanmarexta}', "");
Expect(1, 43647, '\P{^ismyanmarexta}', "");
Expect(0, 43648, '\p{ismyanmarexta}', "");
Expect(1, 43648, '\p{^ismyanmarexta}', "");
Expect(1, 43648, '\P{ismyanmarexta}', "");
Expect(0, 43648, '\P{^ismyanmarexta}', "");
Expect(1, 43647, '\p{  IS_Myanmar_ext_A}', "");
Expect(0, 43647, '\p{^  IS_Myanmar_ext_A}', "");
Expect(0, 43647, '\P{  IS_Myanmar_ext_A}', "");
Expect(1, 43647, '\P{^  IS_Myanmar_ext_A}', "");
Expect(0, 43648, '\p{  IS_Myanmar_ext_A}', "");
Expect(1, 43648, '\p{^  IS_Myanmar_ext_A}', "");
Expect(1, 43648, '\P{  IS_Myanmar_ext_A}', "");
Expect(0, 43648, '\P{^  IS_Myanmar_ext_A}', "");
Error('\p{:= _in_Myanmar_EXT_A}');
Error('\P{:= _in_Myanmar_EXT_A}');
Expect(1, 43647, '\p{inmyanmarexta}', "");
Expect(0, 43647, '\p{^inmyanmarexta}', "");
Expect(0, 43647, '\P{inmyanmarexta}', "");
Expect(1, 43647, '\P{^inmyanmarexta}', "");
Expect(0, 43648, '\p{inmyanmarexta}', "");
Expect(1, 43648, '\p{^inmyanmarexta}', "");
Expect(1, 43648, '\P{inmyanmarexta}', "");
Expect(0, 43648, '\P{^inmyanmarexta}', "");
Expect(1, 43647, '\p{- IN_Myanmar_Ext_A}', "");
Expect(0, 43647, '\p{^- IN_Myanmar_Ext_A}', "");
Expect(0, 43647, '\P{- IN_Myanmar_Ext_A}', "");
Expect(1, 43647, '\P{^- IN_Myanmar_Ext_A}', "");
Expect(0, 43648, '\p{- IN_Myanmar_Ext_A}', "");
Expect(1, 43648, '\p{^- IN_Myanmar_Ext_A}', "");
Expect(1, 43648, '\P{- IN_Myanmar_Ext_A}', "");
Expect(0, 43648, '\P{^- IN_Myanmar_Ext_A}', "");
Error('\p{  runic/a/}');
Error('\P{  runic/a/}');
Expect(1, 5872, '\p{runic}', "");
Expect(0, 5872, '\p{^runic}', "");
Expect(0, 5872, '\P{runic}', "");
Expect(1, 5872, '\P{^runic}', "");
Expect(0, 5873, '\p{runic}', "");
Expect(1, 5873, '\p{^runic}', "");
Expect(1, 5873, '\P{runic}', "");
Expect(0, 5873, '\P{^runic}', "");
Expect(1, 5872, '\p{ 	RUNIC}', "");
Expect(0, 5872, '\p{^ 	RUNIC}', "");
Expect(0, 5872, '\P{ 	RUNIC}', "");
Expect(1, 5872, '\P{^ 	RUNIC}', "");
Expect(0, 5873, '\p{ 	RUNIC}', "");
Expect(1, 5873, '\p{^ 	RUNIC}', "");
Expect(1, 5873, '\P{ 	RUNIC}', "");
Expect(0, 5873, '\P{^ 	RUNIC}', "");
Error('\p{ :=Is_RUNIC}');
Error('\P{ :=Is_RUNIC}');
Expect(1, 5872, '\p{isrunic}', "");
Expect(0, 5872, '\p{^isrunic}', "");
Expect(0, 5872, '\P{isrunic}', "");
Expect(1, 5872, '\P{^isrunic}', "");
Expect(0, 5873, '\p{isrunic}', "");
Expect(1, 5873, '\p{^isrunic}', "");
Expect(1, 5873, '\P{isrunic}', "");
Expect(0, 5873, '\P{^isrunic}', "");
Expect(1, 5872, '\p{_-Is_Runic}', "");
Expect(0, 5872, '\p{^_-Is_Runic}', "");
Expect(0, 5872, '\P{_-Is_Runic}', "");
Expect(1, 5872, '\P{^_-Is_Runic}', "");
Expect(0, 5873, '\p{_-Is_Runic}', "");
Expect(1, 5873, '\p{^_-Is_Runic}', "");
Expect(1, 5873, '\P{_-Is_Runic}', "");
Expect(0, 5873, '\P{^_-Is_Runic}', "");
Error('\p{	/a/runr}');
Error('\P{	/a/runr}');
Expect(1, 5872, '\p{runr}', "");
Expect(0, 5872, '\p{^runr}', "");
Expect(0, 5872, '\P{runr}', "");
Expect(1, 5872, '\P{^runr}', "");
Expect(0, 5873, '\p{runr}', "");
Expect(1, 5873, '\p{^runr}', "");
Expect(1, 5873, '\P{runr}', "");
Expect(0, 5873, '\P{^runr}', "");
Expect(1, 5872, '\p{ -runr}', "");
Expect(0, 5872, '\p{^ -runr}', "");
Expect(0, 5872, '\P{ -runr}', "");
Expect(1, 5872, '\P{^ -runr}', "");
Expect(0, 5873, '\p{ -runr}', "");
Expect(1, 5873, '\p{^ -runr}', "");
Expect(1, 5873, '\P{ -runr}', "");
Expect(0, 5873, '\P{^ -runr}', "");
Error('\p{/a/ 	is_Runr}');
Error('\P{/a/ 	is_Runr}');
Expect(1, 5872, '\p{isrunr}', "");
Expect(0, 5872, '\p{^isrunr}', "");
Expect(0, 5872, '\P{isrunr}', "");
Expect(1, 5872, '\P{^isrunr}', "");
Expect(0, 5873, '\p{isrunr}', "");
Expect(1, 5873, '\p{^isrunr}', "");
Expect(1, 5873, '\P{isrunr}', "");
Expect(0, 5873, '\P{^isrunr}', "");
Expect(1, 5872, '\p{_	IS_Runr}', "");
Expect(0, 5872, '\p{^_	IS_Runr}', "");
Expect(0, 5872, '\P{_	IS_Runr}', "");
Expect(1, 5872, '\P{^_	IS_Runr}', "");
Expect(0, 5873, '\p{_	IS_Runr}', "");
Expect(1, 5873, '\p{^_	IS_Runr}', "");
Expect(1, 5873, '\P{_	IS_Runr}', "");
Expect(0, 5873, '\P{^_	IS_Runr}', "");
Error('\p{Hangul_SYLLABLES:=}');
Error('\P{Hangul_SYLLABLES:=}');
Expect(1, 55215, '\p{hangulsyllables}', "");
Expect(0, 55215, '\p{^hangulsyllables}', "");
Expect(0, 55215, '\P{hangulsyllables}', "");
Expect(1, 55215, '\P{^hangulsyllables}', "");
Expect(0, 55216, '\p{hangulsyllables}', "");
Expect(1, 55216, '\p{^hangulsyllables}', "");
Expect(1, 55216, '\P{hangulsyllables}', "");
Expect(0, 55216, '\P{^hangulsyllables}', "");
Expect(1, 55215, '\p{		Hangul_Syllables}', "");
Expect(0, 55215, '\p{^		Hangul_Syllables}', "");
Expect(0, 55215, '\P{		Hangul_Syllables}', "");
Expect(1, 55215, '\P{^		Hangul_Syllables}', "");
Expect(0, 55216, '\p{		Hangul_Syllables}', "");
Expect(1, 55216, '\p{^		Hangul_Syllables}', "");
Expect(1, 55216, '\P{		Hangul_Syllables}', "");
Expect(0, 55216, '\P{^		Hangul_Syllables}', "");
Error('\p{:= IS_Hangul_Syllables}');
Error('\P{:= IS_Hangul_Syllables}');
Expect(1, 55215, '\p{ishangulsyllables}', "");
Expect(0, 55215, '\p{^ishangulsyllables}', "");
Expect(0, 55215, '\P{ishangulsyllables}', "");
Expect(1, 55215, '\P{^ishangulsyllables}', "");
Expect(0, 55216, '\p{ishangulsyllables}', "");
Expect(1, 55216, '\p{^ishangulsyllables}', "");
Expect(1, 55216, '\P{ishangulsyllables}', "");
Expect(0, 55216, '\P{^ishangulsyllables}', "");
Expect(1, 55215, '\p{--IS_HANGUL_syllables}', "");
Expect(0, 55215, '\p{^--IS_HANGUL_syllables}', "");
Expect(0, 55215, '\P{--IS_HANGUL_syllables}', "");
Expect(1, 55215, '\P{^--IS_HANGUL_syllables}', "");
Expect(0, 55216, '\p{--IS_HANGUL_syllables}', "");
Expect(1, 55216, '\p{^--IS_HANGUL_syllables}', "");
Expect(1, 55216, '\P{--IS_HANGUL_syllables}', "");
Expect(0, 55216, '\P{^--IS_HANGUL_syllables}', "");
Error('\p{/a/-in_Hangul_Syllables}');
Error('\P{/a/-in_Hangul_Syllables}');
Expect(1, 55215, '\p{inhangulsyllables}', "");
Expect(0, 55215, '\p{^inhangulsyllables}', "");
Expect(0, 55215, '\P{inhangulsyllables}', "");
Expect(1, 55215, '\P{^inhangulsyllables}', "");
Expect(0, 55216, '\p{inhangulsyllables}', "");
Expect(1, 55216, '\p{^inhangulsyllables}', "");
Expect(1, 55216, '\P{inhangulsyllables}', "");
Expect(0, 55216, '\P{^inhangulsyllables}', "");
Expect(1, 55215, '\p{_in_Hangul_Syllables}', "");
Expect(0, 55215, '\p{^_in_Hangul_Syllables}', "");
Expect(0, 55215, '\P{_in_Hangul_Syllables}', "");
Expect(1, 55215, '\P{^_in_Hangul_Syllables}', "");
Expect(0, 55216, '\p{_in_Hangul_Syllables}', "");
Expect(1, 55216, '\p{^_in_Hangul_Syllables}', "");
Expect(1, 55216, '\P{_in_Hangul_Syllables}', "");
Expect(0, 55216, '\P{^_in_Hangul_Syllables}', "");
Error('\p{_	In_HANGUL:=}');
Error('\P{_	In_HANGUL:=}');
Expect(1, 55215, '\p{inhangul}', "");
Expect(0, 55215, '\p{^inhangul}', "");
Expect(0, 55215, '\P{inhangul}', "");
Expect(1, 55215, '\P{^inhangul}', "");
Expect(0, 55216, '\p{inhangul}', "");
Expect(1, 55216, '\p{^inhangul}', "");
Expect(1, 55216, '\P{inhangul}', "");
Expect(0, 55216, '\P{^inhangul}', "");
Expect(1, 55215, '\p{_In_Hangul}', "");
Expect(0, 55215, '\p{^_In_Hangul}', "");
Expect(0, 55215, '\P{_In_Hangul}', "");
Expect(1, 55215, '\P{^_In_Hangul}', "");
Expect(0, 55216, '\p{_In_Hangul}', "");
Expect(1, 55216, '\p{^_In_Hangul}', "");
Expect(1, 55216, '\P{_In_Hangul}', "");
Expect(0, 55216, '\P{^_In_Hangul}', "");
Error('\p{--Kayah_li:=}');
Error('\P{--Kayah_li:=}');
Expect(1, 43311, '\p{kayahli}', "");
Expect(0, 43311, '\p{^kayahli}', "");
Expect(0, 43311, '\P{kayahli}', "");
Expect(1, 43311, '\P{^kayahli}', "");
Expect(0, 43312, '\p{kayahli}', "");
Expect(1, 43312, '\p{^kayahli}', "");
Expect(1, 43312, '\P{kayahli}', "");
Expect(0, 43312, '\P{^kayahli}', "");
Expect(1, 43311, '\p{	_KAYAH_li}', "");
Expect(0, 43311, '\p{^	_KAYAH_li}', "");
Expect(0, 43311, '\P{	_KAYAH_li}', "");
Expect(1, 43311, '\P{^	_KAYAH_li}', "");
Expect(0, 43312, '\p{	_KAYAH_li}', "");
Expect(1, 43312, '\p{^	_KAYAH_li}', "");
Expect(1, 43312, '\P{	_KAYAH_li}', "");
Expect(0, 43312, '\P{^	_KAYAH_li}', "");
Error('\p{-Is_kayah_LI/a/}');
Error('\P{-Is_kayah_LI/a/}');
Expect(1, 43311, '\p{iskayahli}', "");
Expect(0, 43311, '\p{^iskayahli}', "");
Expect(0, 43311, '\P{iskayahli}', "");
Expect(1, 43311, '\P{^iskayahli}', "");
Expect(0, 43312, '\p{iskayahli}', "");
Expect(1, 43312, '\p{^iskayahli}', "");
Expect(1, 43312, '\P{iskayahli}', "");
Expect(0, 43312, '\P{^iskayahli}', "");
Expect(1, 43311, '\p{-is_KAYAH_Li}', "");
Expect(0, 43311, '\p{^-is_KAYAH_Li}', "");
Expect(0, 43311, '\P{-is_KAYAH_Li}', "");
Expect(1, 43311, '\P{^-is_KAYAH_Li}', "");
Expect(0, 43312, '\p{-is_KAYAH_Li}', "");
Expect(1, 43312, '\p{^-is_KAYAH_Li}', "");
Expect(1, 43312, '\P{-is_KAYAH_Li}', "");
Expect(0, 43312, '\P{^-is_KAYAH_Li}', "");
Error('\p{_/a/KALI}');
Error('\P{_/a/KALI}');
Expect(1, 43311, '\p{kali}', "");
Expect(0, 43311, '\p{^kali}', "");
Expect(0, 43311, '\P{kali}', "");
Expect(1, 43311, '\P{^kali}', "");
Expect(0, 43312, '\p{kali}', "");
Expect(1, 43312, '\p{^kali}', "");
Expect(1, 43312, '\P{kali}', "");
Expect(0, 43312, '\P{^kali}', "");
Expect(1, 43311, '\p{__Kali}', "");
Expect(0, 43311, '\p{^__Kali}', "");
Expect(0, 43311, '\P{__Kali}', "");
Expect(1, 43311, '\P{^__Kali}', "");
Expect(0, 43312, '\p{__Kali}', "");
Expect(1, 43312, '\p{^__Kali}', "");
Expect(1, 43312, '\P{__Kali}', "");
Expect(0, 43312, '\P{^__Kali}', "");
Error('\p{_-is_kali:=}');
Error('\P{_-is_kali:=}');
Expect(1, 43311, '\p{iskali}', "");
Expect(0, 43311, '\p{^iskali}', "");
Expect(0, 43311, '\P{iskali}', "");
Expect(1, 43311, '\P{^iskali}', "");
Expect(0, 43312, '\p{iskali}', "");
Expect(1, 43312, '\p{^iskali}', "");
Expect(1, 43312, '\P{iskali}', "");
Expect(0, 43312, '\P{^iskali}', "");
Expect(1, 43311, '\p{_Is_Kali}', "");
Expect(0, 43311, '\p{^_Is_Kali}', "");
Expect(0, 43311, '\P{_Is_Kali}', "");
Expect(1, 43311, '\P{^_Is_Kali}', "");
Expect(0, 43312, '\p{_Is_Kali}', "");
Expect(1, 43312, '\p{^_Is_Kali}', "");
Expect(1, 43312, '\P{_Is_Kali}', "");
Expect(0, 43312, '\P{^_Is_Kali}', "");
Error('\p{_posixgraph/a/}');
Error('\P{_posixgraph/a/}');
Expect(1, 126, '\p{posixgraph}', "");
Expect(0, 126, '\p{^posixgraph}', "");
Expect(0, 126, '\P{posixgraph}', "");
Expect(1, 126, '\P{^posixgraph}', "");
Expect(0, 127, '\p{posixgraph}', "");
Expect(1, 127, '\p{^posixgraph}', "");
Expect(1, 127, '\P{posixgraph}', "");
Expect(0, 127, '\P{^posixgraph}', "");
Expect(1, 126, '\p{ -PosixGraph}', "");
Expect(0, 126, '\p{^ -PosixGraph}', "");
Expect(0, 126, '\P{ -PosixGraph}', "");
Expect(1, 126, '\P{^ -PosixGraph}', "");
Expect(0, 127, '\p{ -PosixGraph}', "");
Expect(1, 127, '\p{^ -PosixGraph}', "");
Expect(1, 127, '\P{ -PosixGraph}', "");
Expect(0, 127, '\P{^ -PosixGraph}', "");
Error('\p{	/a/Is_PosixGraph}');
Error('\P{	/a/Is_PosixGraph}');
Expect(1, 126, '\p{isposixgraph}', "");
Expect(0, 126, '\p{^isposixgraph}', "");
Expect(0, 126, '\P{isposixgraph}', "");
Expect(1, 126, '\P{^isposixgraph}', "");
Expect(0, 127, '\p{isposixgraph}', "");
Expect(1, 127, '\p{^isposixgraph}', "");
Expect(1, 127, '\P{isposixgraph}', "");
Expect(0, 127, '\P{^isposixgraph}', "");
Expect(1, 126, '\p{- Is_POSIXGRAPH}', "");
Expect(0, 126, '\p{^- Is_POSIXGRAPH}', "");
Expect(0, 126, '\P{- Is_POSIXGRAPH}', "");
Expect(1, 126, '\P{^- Is_POSIXGRAPH}', "");
Expect(0, 127, '\p{- Is_POSIXGRAPH}', "");
Expect(1, 127, '\p{^- Is_POSIXGRAPH}', "");
Expect(1, 127, '\P{- Is_POSIXGRAPH}', "");
Expect(0, 127, '\P{^- Is_POSIXGRAPH}', "");
Error('\p{/a/ _IN_Old_TURKIC}');
Error('\P{/a/ _IN_Old_TURKIC}');
Expect(1, 68687, '\p{inoldturkic}', "");
Expect(0, 68687, '\p{^inoldturkic}', "");
Expect(0, 68687, '\P{inoldturkic}', "");
Expect(1, 68687, '\P{^inoldturkic}', "");
Expect(0, 68688, '\p{inoldturkic}', "");
Expect(1, 68688, '\p{^inoldturkic}', "");
Expect(1, 68688, '\P{inoldturkic}', "");
Expect(0, 68688, '\P{^inoldturkic}', "");
Expect(1, 68687, '\p{  In_old_Turkic}', "");
Expect(0, 68687, '\p{^  In_old_Turkic}', "");
Expect(0, 68687, '\P{  In_old_Turkic}', "");
Expect(1, 68687, '\P{^  In_old_Turkic}', "");
Expect(0, 68688, '\p{  In_old_Turkic}', "");
Expect(1, 68688, '\p{^  In_old_Turkic}', "");
Expect(1, 68688, '\P{  In_old_Turkic}', "");
Expect(0, 68688, '\P{^  In_old_Turkic}', "");
Error('\p{:=-IN_Old_Turkic}');
Error('\P{:=-IN_Old_Turkic}');
Expect(1, 68687, '\p{- In_Old_Turkic}', "");
Expect(0, 68687, '\p{^- In_Old_Turkic}', "");
Expect(0, 68687, '\P{- In_Old_Turkic}', "");
Expect(1, 68687, '\P{^- In_Old_Turkic}', "");
Expect(0, 68688, '\p{- In_Old_Turkic}', "");
Expect(1, 68688, '\p{^- In_Old_Turkic}', "");
Expect(1, 68688, '\P{- In_Old_Turkic}', "");
Expect(0, 68688, '\P{^- In_Old_Turkic}', "");
Error('\p{-:=GRAPHEME_Base}');
Error('\P{-:=GRAPHEME_Base}');
Expect(1, 195101, '\p{graphemebase}', "");
Expect(0, 195101, '\p{^graphemebase}', "");
Expect(0, 195101, '\P{graphemebase}', "");
Expect(1, 195101, '\P{^graphemebase}', "");
Expect(0, 195102, '\p{graphemebase}', "");
Expect(1, 195102, '\p{^graphemebase}', "");
Expect(1, 195102, '\P{graphemebase}', "");
Expect(0, 195102, '\P{^graphemebase}', "");
Expect(1, 195101, '\p{	 grapheme_Base}', "");
Expect(0, 195101, '\p{^	 grapheme_Base}', "");
Expect(0, 195101, '\P{	 grapheme_Base}', "");
Expect(1, 195101, '\P{^	 grapheme_Base}', "");
Expect(0, 195102, '\p{	 grapheme_Base}', "");
Expect(1, 195102, '\p{^	 grapheme_Base}', "");
Expect(1, 195102, '\P{	 grapheme_Base}', "");
Expect(0, 195102, '\P{^	 grapheme_Base}', "");
Error('\p{ :=IS_GRAPHEME_Base}');
Error('\P{ :=IS_GRAPHEME_Base}');
Expect(1, 195101, '\p{isgraphemebase}', "");
Expect(0, 195101, '\p{^isgraphemebase}', "");
Expect(0, 195101, '\P{isgraphemebase}', "");
Expect(1, 195101, '\P{^isgraphemebase}', "");
Expect(0, 195102, '\p{isgraphemebase}', "");
Expect(1, 195102, '\p{^isgraphemebase}', "");
Expect(1, 195102, '\P{isgraphemebase}', "");
Expect(0, 195102, '\P{^isgraphemebase}', "");
Expect(1, 195101, '\p{ IS_grapheme_BASE}', "");
Expect(0, 195101, '\p{^ IS_grapheme_BASE}', "");
Expect(0, 195101, '\P{ IS_grapheme_BASE}', "");
Expect(1, 195101, '\P{^ IS_grapheme_BASE}', "");
Expect(0, 195102, '\p{ IS_grapheme_BASE}', "");
Expect(1, 195102, '\p{^ IS_grapheme_BASE}', "");
Expect(1, 195102, '\P{ IS_grapheme_BASE}', "");
Expect(0, 195102, '\P{^ IS_grapheme_BASE}', "");
Error('\p{-GR_BASE:=}');
Error('\P{-GR_BASE:=}');
Expect(1, 195101, '\p{grbase}', "");
Expect(0, 195101, '\p{^grbase}', "");
Expect(0, 195101, '\P{grbase}', "");
Expect(1, 195101, '\P{^grbase}', "");
Expect(0, 195102, '\p{grbase}', "");
Expect(1, 195102, '\p{^grbase}', "");
Expect(1, 195102, '\P{grbase}', "");
Expect(0, 195102, '\P{^grbase}', "");
Expect(1, 195101, '\p{__GR_Base}', "");
Expect(0, 195101, '\p{^__GR_Base}', "");
Expect(0, 195101, '\P{__GR_Base}', "");
Expect(1, 195101, '\P{^__GR_Base}', "");
Expect(0, 195102, '\p{__GR_Base}', "");
Expect(1, 195102, '\p{^__GR_Base}', "");
Expect(1, 195102, '\P{__GR_Base}', "");
Expect(0, 195102, '\P{^__GR_Base}', "");
Error('\p{	/a/is_GR_base}');
Error('\P{	/a/is_GR_base}');
Expect(1, 195101, '\p{isgrbase}', "");
Expect(0, 195101, '\p{^isgrbase}', "");
Expect(0, 195101, '\P{isgrbase}', "");
Expect(1, 195101, '\P{^isgrbase}', "");
Expect(0, 195102, '\p{isgrbase}', "");
Expect(1, 195102, '\p{^isgrbase}', "");
Expect(1, 195102, '\P{isgrbase}', "");
Expect(0, 195102, '\P{^isgrbase}', "");
Expect(1, 195101, '\p{	is_Gr_Base}', "");
Expect(0, 195101, '\p{^	is_Gr_Base}', "");
Expect(0, 195101, '\P{	is_Gr_Base}', "");
Expect(1, 195101, '\P{^	is_Gr_Base}', "");
Expect(0, 195102, '\p{	is_Gr_Base}', "");
Expect(1, 195102, '\p{^	is_Gr_Base}', "");
Expect(1, 195102, '\P{	is_Gr_Base}', "");
Expect(0, 195102, '\P{^	is_Gr_Base}', "");
Error('\p{/a/		in_Ugaritic}');
Error('\P{/a/		in_Ugaritic}');
Expect(1, 66463, '\p{inugaritic}', "");
Expect(0, 66463, '\p{^inugaritic}', "");
Expect(0, 66463, '\P{inugaritic}', "");
Expect(1, 66463, '\P{^inugaritic}', "");
Expect(0, 66464, '\p{inugaritic}', "");
Expect(1, 66464, '\p{^inugaritic}', "");
Expect(1, 66464, '\P{inugaritic}', "");
Expect(0, 66464, '\P{^inugaritic}', "");
Expect(1, 66463, '\p{		in_UGARITIC}', "");
Expect(0, 66463, '\p{^		in_UGARITIC}', "");
Expect(0, 66463, '\P{		in_UGARITIC}', "");
Expect(1, 66463, '\P{^		in_UGARITIC}', "");
Expect(0, 66464, '\p{		in_UGARITIC}', "");
Expect(1, 66464, '\p{^		in_UGARITIC}', "");
Expect(1, 66464, '\P{		in_UGARITIC}', "");
Expect(0, 66464, '\P{^		in_UGARITIC}', "");
Error('\p{/a/in_Ugaritic}');
Error('\P{/a/in_Ugaritic}');
Expect(1, 66463, '\p{-_in_UGARITIC}', "");
Expect(0, 66463, '\p{^-_in_UGARITIC}', "");
Expect(0, 66463, '\P{-_in_UGARITIC}', "");
Expect(1, 66463, '\P{^-_in_UGARITIC}', "");
Expect(0, 66464, '\p{-_in_UGARITIC}', "");
Expect(1, 66464, '\p{^-_in_UGARITIC}', "");
Expect(1, 66464, '\P{-_in_UGARITIC}', "");
Expect(0, 66464, '\P{^-_in_UGARITIC}', "");
Error('\p{	/a/Devanagari}');
Error('\P{	/a/Devanagari}');
Expect(1, 43259, '\p{devanagari}', "");
Expect(0, 43259, '\p{^devanagari}', "");
Expect(0, 43259, '\P{devanagari}', "");
Expect(1, 43259, '\P{^devanagari}', "");
Expect(0, 43260, '\p{devanagari}', "");
Expect(1, 43260, '\p{^devanagari}', "");
Expect(1, 43260, '\P{devanagari}', "");
Expect(0, 43260, '\P{^devanagari}', "");
Expect(1, 43259, '\p{_Devanagari}', "");
Expect(0, 43259, '\p{^_Devanagari}', "");
Expect(0, 43259, '\P{_Devanagari}', "");
Expect(1, 43259, '\P{^_Devanagari}', "");
Expect(0, 43260, '\p{_Devanagari}', "");
Expect(1, 43260, '\p{^_Devanagari}', "");
Expect(1, 43260, '\P{_Devanagari}', "");
Expect(0, 43260, '\P{^_Devanagari}', "");
Error('\p{:=_ is_Devanagari}');
Error('\P{:=_ is_Devanagari}');
Expect(1, 43259, '\p{isdevanagari}', "");
Expect(0, 43259, '\p{^isdevanagari}', "");
Expect(0, 43259, '\P{isdevanagari}', "");
Expect(1, 43259, '\P{^isdevanagari}', "");
Expect(0, 43260, '\p{isdevanagari}', "");
Expect(1, 43260, '\p{^isdevanagari}', "");
Expect(1, 43260, '\P{isdevanagari}', "");
Expect(0, 43260, '\P{^isdevanagari}', "");
Expect(1, 43259, '\p{_-Is_Devanagari}', "");
Expect(0, 43259, '\p{^_-Is_Devanagari}', "");
Expect(0, 43259, '\P{_-Is_Devanagari}', "");
Expect(1, 43259, '\P{^_-Is_Devanagari}', "");
Expect(0, 43260, '\p{_-Is_Devanagari}', "");
Expect(1, 43260, '\p{^_-Is_Devanagari}', "");
Expect(1, 43260, '\P{_-Is_Devanagari}', "");
Expect(0, 43260, '\P{^_-Is_Devanagari}', "");
Error('\p{/a/-_deva}');
Error('\P{/a/-_deva}');
Expect(1, 43259, '\p{deva}', "");
Expect(0, 43259, '\p{^deva}', "");
Expect(0, 43259, '\P{deva}', "");
Expect(1, 43259, '\P{^deva}', "");
Expect(0, 43260, '\p{deva}', "");
Expect(1, 43260, '\p{^deva}', "");
Expect(1, 43260, '\P{deva}', "");
Expect(0, 43260, '\P{^deva}', "");
Expect(1, 43259, '\p{Deva}', "");
Expect(0, 43259, '\p{^Deva}', "");
Expect(0, 43259, '\P{Deva}', "");
Expect(1, 43259, '\P{^Deva}', "");
Expect(0, 43260, '\p{Deva}', "");
Expect(1, 43260, '\p{^Deva}', "");
Expect(1, 43260, '\P{Deva}', "");
Expect(0, 43260, '\P{^Deva}', "");
Error('\p{/a/Is_DEVA}');
Error('\P{/a/Is_DEVA}');
Expect(1, 43259, '\p{isdeva}', "");
Expect(0, 43259, '\p{^isdeva}', "");
Expect(0, 43259, '\P{isdeva}', "");
Expect(1, 43259, '\P{^isdeva}', "");
Expect(0, 43260, '\p{isdeva}', "");
Expect(1, 43260, '\p{^isdeva}', "");
Expect(1, 43260, '\P{isdeva}', "");
Expect(0, 43260, '\P{^isdeva}', "");
Expect(1, 43259, '\p{-is_DEVA}', "");
Expect(0, 43259, '\p{^-is_DEVA}', "");
Expect(0, 43259, '\P{-is_DEVA}', "");
Expect(1, 43259, '\P{^-is_DEVA}', "");
Expect(0, 43260, '\p{-is_DEVA}', "");
Expect(1, 43260, '\p{^-is_DEVA}', "");
Expect(1, 43260, '\P{-is_DEVA}', "");
Expect(0, 43260, '\P{^-is_DEVA}', "");
Error('\p{ :=BRAHMI}');
Error('\P{ :=BRAHMI}');
Expect(1, 69743, '\p{brahmi}', "");
Expect(0, 69743, '\p{^brahmi}', "");
Expect(0, 69743, '\P{brahmi}', "");
Expect(1, 69743, '\P{^brahmi}', "");
Expect(0, 69744, '\p{brahmi}', "");
Expect(1, 69744, '\p{^brahmi}', "");
Expect(1, 69744, '\P{brahmi}', "");
Expect(0, 69744, '\P{^brahmi}', "");
Expect(1, 69743, '\p{	 Brahmi}', "");
Expect(0, 69743, '\p{^	 Brahmi}', "");
Expect(0, 69743, '\P{	 Brahmi}', "");
Expect(1, 69743, '\P{^	 Brahmi}', "");
Expect(0, 69744, '\p{	 Brahmi}', "");
Expect(1, 69744, '\p{^	 Brahmi}', "");
Expect(1, 69744, '\P{	 Brahmi}', "");
Expect(0, 69744, '\P{^	 Brahmi}', "");
Error('\p{:= Is_Brahmi}');
Error('\P{:= Is_Brahmi}');
Expect(1, 69743, '\p{isbrahmi}', "");
Expect(0, 69743, '\p{^isbrahmi}', "");
Expect(0, 69743, '\P{isbrahmi}', "");
Expect(1, 69743, '\P{^isbrahmi}', "");
Expect(0, 69744, '\p{isbrahmi}', "");
Expect(1, 69744, '\p{^isbrahmi}', "");
Expect(1, 69744, '\P{isbrahmi}', "");
Expect(0, 69744, '\P{^isbrahmi}', "");
Expect(1, 69743, '\p{	IS_brahmi}', "");
Expect(0, 69743, '\p{^	IS_brahmi}', "");
Expect(0, 69743, '\P{	IS_brahmi}', "");
Expect(1, 69743, '\P{^	IS_brahmi}', "");
Expect(0, 69744, '\p{	IS_brahmi}', "");
Expect(1, 69744, '\p{^	IS_brahmi}', "");
Expect(1, 69744, '\P{	IS_brahmi}', "");
Expect(0, 69744, '\P{^	IS_brahmi}', "");
Error('\p{:=_ BRAH}');
Error('\P{:=_ BRAH}');
Expect(1, 69743, '\p{brah}', "");
Expect(0, 69743, '\p{^brah}', "");
Expect(0, 69743, '\P{brah}', "");
Expect(1, 69743, '\P{^brah}', "");
Expect(0, 69744, '\p{brah}', "");
Expect(1, 69744, '\p{^brah}', "");
Expect(1, 69744, '\P{brah}', "");
Expect(0, 69744, '\P{^brah}', "");
Expect(1, 69743, '\p{-Brah}', "");
Expect(0, 69743, '\p{^-Brah}', "");
Expect(0, 69743, '\P{-Brah}', "");
Expect(1, 69743, '\P{^-Brah}', "");
Expect(0, 69744, '\p{-Brah}', "");
Expect(1, 69744, '\p{^-Brah}', "");
Expect(1, 69744, '\P{-Brah}', "");
Expect(0, 69744, '\P{^-Brah}', "");
Error('\p{:=-_IS_BRAH}');
Error('\P{:=-_IS_BRAH}');
Expect(1, 69743, '\p{isbrah}', "");
Expect(0, 69743, '\p{^isbrah}', "");
Expect(0, 69743, '\P{isbrah}', "");
Expect(1, 69743, '\P{^isbrah}', "");
Expect(0, 69744, '\p{isbrah}', "");
Expect(1, 69744, '\p{^isbrah}', "");
Expect(1, 69744, '\P{isbrah}', "");
Expect(0, 69744, '\P{^isbrah}', "");
Expect(1, 69743, '\p{	_Is_Brah}', "");
Expect(0, 69743, '\p{^	_Is_Brah}', "");
Expect(0, 69743, '\P{	_Is_Brah}', "");
Expect(1, 69743, '\P{^	_Is_Brah}', "");
Expect(0, 69744, '\p{	_Is_Brah}', "");
Expect(1, 69744, '\p{^	_Is_Brah}', "");
Expect(1, 69744, '\P{	_Is_Brah}', "");
Expect(0, 69744, '\P{^	_Is_Brah}', "");
Error('\p{:=	FINAL_Punctuation}');
Error('\P{:=	FINAL_Punctuation}');
Expect(1, 11809, '\p{finalpunctuation}', "");
Expect(0, 11809, '\p{^finalpunctuation}', "");
Expect(0, 11809, '\P{finalpunctuation}', "");
Expect(1, 11809, '\P{^finalpunctuation}', "");
Expect(0, 11810, '\p{finalpunctuation}', "");
Expect(1, 11810, '\p{^finalpunctuation}', "");
Expect(1, 11810, '\P{finalpunctuation}', "");
Expect(0, 11810, '\P{^finalpunctuation}', "");
Expect(1, 11809, '\p{-FINAL_Punctuation}', "");
Expect(0, 11809, '\p{^-FINAL_Punctuation}', "");
Expect(0, 11809, '\P{-FINAL_Punctuation}', "");
Expect(1, 11809, '\P{^-FINAL_Punctuation}', "");
Expect(0, 11810, '\p{-FINAL_Punctuation}', "");
Expect(1, 11810, '\p{^-FINAL_Punctuation}', "");
Expect(1, 11810, '\P{-FINAL_Punctuation}', "");
Expect(0, 11810, '\P{^-FINAL_Punctuation}', "");
Error('\p{- Is_FINAL_Punctuation/a/}');
Error('\P{- Is_FINAL_Punctuation/a/}');
Expect(1, 11809, '\p{isfinalpunctuation}', "");
Expect(0, 11809, '\p{^isfinalpunctuation}', "");
Expect(0, 11809, '\P{isfinalpunctuation}', "");
Expect(1, 11809, '\P{^isfinalpunctuation}', "");
Expect(0, 11810, '\p{isfinalpunctuation}', "");
Expect(1, 11810, '\p{^isfinalpunctuation}', "");
Expect(1, 11810, '\P{isfinalpunctuation}', "");
Expect(0, 11810, '\P{^isfinalpunctuation}', "");
Expect(1, 11809, '\p{	-Is_Final_Punctuation}', "");
Expect(0, 11809, '\p{^	-Is_Final_Punctuation}', "");
Expect(0, 11809, '\P{	-Is_Final_Punctuation}', "");
Expect(1, 11809, '\P{^	-Is_Final_Punctuation}', "");
Expect(0, 11810, '\p{	-Is_Final_Punctuation}', "");
Expect(1, 11810, '\p{^	-Is_Final_Punctuation}', "");
Expect(1, 11810, '\P{	-Is_Final_Punctuation}', "");
Expect(0, 11810, '\P{^	-Is_Final_Punctuation}', "");
Error('\p{/a/	Pf}');
Error('\P{/a/	Pf}');
Expect(1, 11809, '\p{pf}', "");
Expect(0, 11809, '\p{^pf}', "");
Expect(0, 11809, '\P{pf}', "");
Expect(1, 11809, '\P{^pf}', "");
Expect(0, 11810, '\p{pf}', "");
Expect(1, 11810, '\p{^pf}', "");
Expect(1, 11810, '\P{pf}', "");
Expect(0, 11810, '\P{^pf}', "");
Expect(1, 11809, '\p{-_pf}', "");
Expect(0, 11809, '\p{^-_pf}', "");
Expect(0, 11809, '\P{-_pf}', "");
Expect(1, 11809, '\P{^-_pf}', "");
Expect(0, 11810, '\p{-_pf}', "");
Expect(1, 11810, '\p{^-_pf}', "");
Expect(1, 11810, '\P{-_pf}', "");
Expect(0, 11810, '\P{^-_pf}', "");
Error('\p{/a/__is_PF}');
Error('\P{/a/__is_PF}');
Expect(1, 11809, '\p{ispf}', "");
Expect(0, 11809, '\p{^ispf}', "");
Expect(0, 11809, '\P{ispf}', "");
Expect(1, 11809, '\P{^ispf}', "");
Expect(0, 11810, '\p{ispf}', "");
Expect(1, 11810, '\p{^ispf}', "");
Expect(1, 11810, '\P{ispf}', "");
Expect(0, 11810, '\P{^ispf}', "");
Expect(1, 11809, '\p{	Is_PF}', "");
Expect(0, 11809, '\p{^	Is_PF}', "");
Expect(0, 11809, '\P{	Is_PF}', "");
Expect(1, 11809, '\P{^	Is_PF}', "");
Expect(0, 11810, '\p{	Is_PF}', "");
Expect(1, 11810, '\p{^	Is_PF}', "");
Expect(1, 11810, '\P{	Is_PF}', "");
Expect(0, 11810, '\P{^	Is_PF}', "");
Error('\p{/a/		modifier_Letter}');
Error('\P{/a/		modifier_Letter}');
Expect(1, 94111, '\p{modifierletter}', "");
Expect(0, 94111, '\p{^modifierletter}', "");
Expect(0, 94111, '\P{modifierletter}', "");
Expect(1, 94111, '\P{^modifierletter}', "");
Expect(0, 94112, '\p{modifierletter}', "");
Expect(1, 94112, '\p{^modifierletter}', "");
Expect(1, 94112, '\P{modifierletter}', "");
Expect(0, 94112, '\P{^modifierletter}', "");
Expect(1, 94111, '\p{_-modifier_letter}', "");
Expect(0, 94111, '\p{^_-modifier_letter}', "");
Expect(0, 94111, '\P{_-modifier_letter}', "");
Expect(1, 94111, '\P{^_-modifier_letter}', "");
Expect(0, 94112, '\p{_-modifier_letter}', "");
Expect(1, 94112, '\p{^_-modifier_letter}', "");
Expect(1, 94112, '\P{_-modifier_letter}', "");
Expect(0, 94112, '\P{^_-modifier_letter}', "");
Error('\p{-	IS_modifier_LETTER/a/}');
Error('\P{-	IS_modifier_LETTER/a/}');
Expect(1, 94111, '\p{ismodifierletter}', "");
Expect(0, 94111, '\p{^ismodifierletter}', "");
Expect(0, 94111, '\P{ismodifierletter}', "");
Expect(1, 94111, '\P{^ismodifierletter}', "");
Expect(0, 94112, '\p{ismodifierletter}', "");
Expect(1, 94112, '\p{^ismodifierletter}', "");
Expect(1, 94112, '\P{ismodifierletter}', "");
Expect(0, 94112, '\P{^ismodifierletter}', "");
Expect(1, 94111, '\p{_ Is_Modifier_Letter}', "");
Expect(0, 94111, '\p{^_ Is_Modifier_Letter}', "");
Expect(0, 94111, '\P{_ Is_Modifier_Letter}', "");
Expect(1, 94111, '\P{^_ Is_Modifier_Letter}', "");
Expect(0, 94112, '\p{_ Is_Modifier_Letter}', "");
Expect(1, 94112, '\p{^_ Is_Modifier_Letter}', "");
Expect(1, 94112, '\P{_ Is_Modifier_Letter}', "");
Expect(0, 94112, '\P{^_ Is_Modifier_Letter}', "");
Error('\p{lm/a/}');
Error('\P{lm/a/}');
Expect(1, 94111, '\p{lm}', "");
Expect(0, 94111, '\p{^lm}', "");
Expect(0, 94111, '\P{lm}', "");
Expect(1, 94111, '\P{^lm}', "");
Expect(0, 94112, '\p{lm}', "");
Expect(1, 94112, '\p{^lm}', "");
Expect(1, 94112, '\P{lm}', "");
Expect(0, 94112, '\P{^lm}', "");
Expect(1, 94111, '\p{_-LM}', "");
Expect(0, 94111, '\p{^_-LM}', "");
Expect(0, 94111, '\P{_-LM}', "");
Expect(1, 94111, '\P{^_-LM}', "");
Expect(0, 94112, '\p{_-LM}', "");
Expect(1, 94112, '\p{^_-LM}', "");
Expect(1, 94112, '\P{_-LM}', "");
Expect(0, 94112, '\P{^_-LM}', "");
Error('\p{ is_lm/a/}');
Error('\P{ is_lm/a/}');
Expect(1, 94111, '\p{islm}', "");
Expect(0, 94111, '\p{^islm}', "");
Expect(0, 94111, '\P{islm}', "");
Expect(1, 94111, '\P{^islm}', "");
Expect(0, 94112, '\p{islm}', "");
Expect(1, 94112, '\p{^islm}', "");
Expect(1, 94112, '\P{islm}', "");
Expect(0, 94112, '\P{^islm}', "");
Expect(1, 94111, '\p{  IS_LM}', "");
Expect(0, 94111, '\p{^  IS_LM}', "");
Expect(0, 94111, '\P{  IS_LM}', "");
Expect(1, 94111, '\P{^  IS_LM}', "");
Expect(0, 94112, '\p{  IS_LM}', "");
Expect(1, 94112, '\p{^  IS_LM}', "");
Expect(1, 94112, '\P{  IS_LM}', "");
Expect(0, 94112, '\P{^  IS_LM}', "");
Error('\p{_:=Kanbun}');
Error('\P{_:=Kanbun}');
Expect(1, 12703, '\p{kanbun}', "");
Expect(0, 12703, '\p{^kanbun}', "");
Expect(0, 12703, '\P{kanbun}', "");
Expect(1, 12703, '\P{^kanbun}', "");
Expect(0, 12704, '\p{kanbun}', "");
Expect(1, 12704, '\p{^kanbun}', "");
Expect(1, 12704, '\P{kanbun}', "");
Expect(0, 12704, '\P{^kanbun}', "");
Expect(1, 12703, '\p{ KANBUN}', "");
Expect(0, 12703, '\p{^ KANBUN}', "");
Expect(0, 12703, '\P{ KANBUN}', "");
Expect(1, 12703, '\P{^ KANBUN}', "");
Expect(0, 12704, '\p{ KANBUN}', "");
Expect(1, 12704, '\p{^ KANBUN}', "");
Expect(1, 12704, '\P{ KANBUN}', "");
Expect(0, 12704, '\P{^ KANBUN}', "");
Error('\p{/a/__is_Kanbun}');
Error('\P{/a/__is_Kanbun}');
Expect(1, 12703, '\p{iskanbun}', "");
Expect(0, 12703, '\p{^iskanbun}', "");
Expect(0, 12703, '\P{iskanbun}', "");
Expect(1, 12703, '\P{^iskanbun}', "");
Expect(0, 12704, '\p{iskanbun}', "");
Expect(1, 12704, '\p{^iskanbun}', "");
Expect(1, 12704, '\P{iskanbun}', "");
Expect(0, 12704, '\P{^iskanbun}', "");
Expect(1, 12703, '\p{ Is_Kanbun}', "");
Expect(0, 12703, '\p{^ Is_Kanbun}', "");
Expect(0, 12703, '\P{ Is_Kanbun}', "");
Expect(1, 12703, '\P{^ Is_Kanbun}', "");
Expect(0, 12704, '\p{ Is_Kanbun}', "");
Expect(1, 12704, '\p{^ Is_Kanbun}', "");
Expect(1, 12704, '\P{ Is_Kanbun}', "");
Expect(0, 12704, '\P{^ Is_Kanbun}', "");
Error('\p{--in_KANBUN:=}');
Error('\P{--in_KANBUN:=}');
Expect(1, 12703, '\p{inkanbun}', "");
Expect(0, 12703, '\p{^inkanbun}', "");
Expect(0, 12703, '\P{inkanbun}', "");
Expect(1, 12703, '\P{^inkanbun}', "");
Expect(0, 12704, '\p{inkanbun}', "");
Expect(1, 12704, '\p{^inkanbun}', "");
Expect(1, 12704, '\P{inkanbun}', "");
Expect(0, 12704, '\P{^inkanbun}', "");
Expect(1, 12703, '\p{  IN_Kanbun}', "");
Expect(0, 12703, '\p{^  IN_Kanbun}', "");
Expect(0, 12703, '\P{  IN_Kanbun}', "");
Expect(1, 12703, '\P{^  IN_Kanbun}', "");
Expect(0, 12704, '\p{  IN_Kanbun}', "");
Expect(1, 12704, '\p{^  IN_Kanbun}', "");
Expect(1, 12704, '\P{  IN_Kanbun}', "");
Expect(0, 12704, '\P{^  IN_Kanbun}', "");
Error('\p{ :=In_Devanagari}');
Error('\P{ :=In_Devanagari}');
Expect(1, 2431, '\p{indevanagari}', "");
Expect(0, 2431, '\p{^indevanagari}', "");
Expect(0, 2431, '\P{indevanagari}', "");
Expect(1, 2431, '\P{^indevanagari}', "");
Expect(0, 2432, '\p{indevanagari}', "");
Expect(1, 2432, '\p{^indevanagari}', "");
Expect(1, 2432, '\P{indevanagari}', "");
Expect(0, 2432, '\P{^indevanagari}', "");
Expect(1, 2431, '\p{_ IN_DEVANAGARI}', "");
Expect(0, 2431, '\p{^_ IN_DEVANAGARI}', "");
Expect(0, 2431, '\P{_ IN_DEVANAGARI}', "");
Expect(1, 2431, '\P{^_ IN_DEVANAGARI}', "");
Expect(0, 2432, '\p{_ IN_DEVANAGARI}', "");
Expect(1, 2432, '\p{^_ IN_DEVANAGARI}', "");
Expect(1, 2432, '\P{_ IN_DEVANAGARI}', "");
Expect(0, 2432, '\P{^_ IN_DEVANAGARI}', "");
Error('\p{/a/In_DEVANAGARI}');
Error('\P{/a/In_DEVANAGARI}');
Expect(1, 2431, '\p{--in_devanagari}', "");
Expect(0, 2431, '\p{^--in_devanagari}', "");
Expect(0, 2431, '\P{--in_devanagari}', "");
Expect(1, 2431, '\P{^--in_devanagari}', "");
Expect(0, 2432, '\p{--in_devanagari}', "");
Expect(1, 2432, '\p{^--in_devanagari}', "");
Expect(1, 2432, '\P{--in_devanagari}', "");
Expect(0, 2432, '\P{^--in_devanagari}', "");
Error('\p{ /a/TIBETAN}');
Error('\P{ /a/TIBETAN}');
Expect(1, 4058, '\p{tibetan}', "");
Expect(0, 4058, '\p{^tibetan}', "");
Expect(0, 4058, '\P{tibetan}', "");
Expect(1, 4058, '\P{^tibetan}', "");
Expect(0, 4059, '\p{tibetan}', "");
Expect(1, 4059, '\p{^tibetan}', "");
Expect(1, 4059, '\P{tibetan}', "");
Expect(0, 4059, '\P{^tibetan}', "");
Expect(1, 4058, '\p{ _tibetan}', "");
Expect(0, 4058, '\p{^ _tibetan}', "");
Expect(0, 4058, '\P{ _tibetan}', "");
Expect(1, 4058, '\P{^ _tibetan}', "");
Expect(0, 4059, '\p{ _tibetan}', "");
Expect(1, 4059, '\p{^ _tibetan}', "");
Expect(1, 4059, '\P{ _tibetan}', "");
Expect(0, 4059, '\P{^ _tibetan}', "");
Error('\p{/a/_IS_Tibetan}');
Error('\P{/a/_IS_Tibetan}');
Expect(1, 4058, '\p{istibetan}', "");
Expect(0, 4058, '\p{^istibetan}', "");
Expect(0, 4058, '\P{istibetan}', "");
Expect(1, 4058, '\P{^istibetan}', "");
Expect(0, 4059, '\p{istibetan}', "");
Expect(1, 4059, '\p{^istibetan}', "");
Expect(1, 4059, '\P{istibetan}', "");
Expect(0, 4059, '\P{^istibetan}', "");
Expect(1, 4058, '\p{_Is_Tibetan}', "");
Expect(0, 4058, '\p{^_Is_Tibetan}', "");
Expect(0, 4058, '\P{_Is_Tibetan}', "");
Expect(1, 4058, '\P{^_Is_Tibetan}', "");
Expect(0, 4059, '\p{_Is_Tibetan}', "");
Expect(1, 4059, '\p{^_Is_Tibetan}', "");
Expect(1, 4059, '\P{_Is_Tibetan}', "");
Expect(0, 4059, '\P{^_Is_Tibetan}', "");
Error('\p{/a/	Tibt}');
Error('\P{/a/	Tibt}');
Expect(1, 4058, '\p{tibt}', "");
Expect(0, 4058, '\p{^tibt}', "");
Expect(0, 4058, '\P{tibt}', "");
Expect(1, 4058, '\P{^tibt}', "");
Expect(0, 4059, '\p{tibt}', "");
Expect(1, 4059, '\p{^tibt}', "");
Expect(1, 4059, '\P{tibt}', "");
Expect(0, 4059, '\P{^tibt}', "");
Expect(1, 4058, '\p{__Tibt}', "");
Expect(0, 4058, '\p{^__Tibt}', "");
Expect(0, 4058, '\P{__Tibt}', "");
Expect(1, 4058, '\P{^__Tibt}', "");
Expect(0, 4059, '\p{__Tibt}', "");
Expect(1, 4059, '\p{^__Tibt}', "");
Expect(1, 4059, '\P{__Tibt}', "");
Expect(0, 4059, '\P{^__Tibt}', "");
Error('\p{ is_Tibt/a/}');
Error('\P{ is_Tibt/a/}');
Expect(1, 4058, '\p{istibt}', "");
Expect(0, 4058, '\p{^istibt}', "");
Expect(0, 4058, '\P{istibt}', "");
Expect(1, 4058, '\P{^istibt}', "");
Expect(0, 4059, '\p{istibt}', "");
Expect(1, 4059, '\p{^istibt}', "");
Expect(1, 4059, '\P{istibt}', "");
Expect(0, 4059, '\P{^istibt}', "");
Expect(1, 4058, '\p{__is_Tibt}', "");
Expect(0, 4058, '\p{^__is_Tibt}', "");
Expect(0, 4058, '\P{__is_Tibt}', "");
Expect(1, 4058, '\P{^__is_Tibt}', "");
Expect(0, 4059, '\p{__is_Tibt}', "");
Expect(1, 4059, '\p{^__is_Tibt}', "");
Expect(1, 4059, '\P{__is_Tibt}', "");
Expect(0, 4059, '\P{^__is_Tibt}', "");
Error('\p{ :=JAVANESE}');
Error('\P{ :=JAVANESE}');
Expect(1, 43487, '\p{javanese}', "");
Expect(0, 43487, '\p{^javanese}', "");
Expect(0, 43487, '\P{javanese}', "");
Expect(1, 43487, '\P{^javanese}', "");
Expect(0, 43488, '\p{javanese}', "");
Expect(1, 43488, '\p{^javanese}', "");
Expect(1, 43488, '\P{javanese}', "");
Expect(0, 43488, '\P{^javanese}', "");
Expect(1, 43487, '\p{-_Javanese}', "");
Expect(0, 43487, '\p{^-_Javanese}', "");
Expect(0, 43487, '\P{-_Javanese}', "");
Expect(1, 43487, '\P{^-_Javanese}', "");
Expect(0, 43488, '\p{-_Javanese}', "");
Expect(1, 43488, '\p{^-_Javanese}', "");
Expect(1, 43488, '\P{-_Javanese}', "");
Expect(0, 43488, '\P{^-_Javanese}', "");
Error('\p{:=  Is_Javanese}');
Error('\P{:=  Is_Javanese}');
Expect(1, 43487, '\p{isjavanese}', "");
Expect(0, 43487, '\p{^isjavanese}', "");
Expect(0, 43487, '\P{isjavanese}', "");
Expect(1, 43487, '\P{^isjavanese}', "");
Expect(0, 43488, '\p{isjavanese}', "");
Expect(1, 43488, '\p{^isjavanese}', "");
Expect(1, 43488, '\P{isjavanese}', "");
Expect(0, 43488, '\P{^isjavanese}', "");
Expect(1, 43487, '\p{-	Is_javanese}', "");
Expect(0, 43487, '\p{^-	Is_javanese}', "");
Expect(0, 43487, '\P{-	Is_javanese}', "");
Expect(1, 43487, '\P{^-	Is_javanese}', "");
Expect(0, 43488, '\p{-	Is_javanese}', "");
Expect(1, 43488, '\p{^-	Is_javanese}', "");
Expect(1, 43488, '\P{-	Is_javanese}', "");
Expect(0, 43488, '\P{^-	Is_javanese}', "");
Error('\p{	:=java}');
Error('\P{	:=java}');
Expect(1, 43487, '\p{java}', "");
Expect(0, 43487, '\p{^java}', "");
Expect(0, 43487, '\P{java}', "");
Expect(1, 43487, '\P{^java}', "");
Expect(0, 43488, '\p{java}', "");
Expect(1, 43488, '\p{^java}', "");
Expect(1, 43488, '\P{java}', "");
Expect(0, 43488, '\P{^java}', "");
Expect(1, 43487, '\p{ -JAVA}', "");
Expect(0, 43487, '\p{^ -JAVA}', "");
Expect(0, 43487, '\P{ -JAVA}', "");
Expect(1, 43487, '\P{^ -JAVA}', "");
Expect(0, 43488, '\p{ -JAVA}', "");
Expect(1, 43488, '\p{^ -JAVA}', "");
Expect(1, 43488, '\P{ -JAVA}', "");
Expect(0, 43488, '\P{^ -JAVA}', "");
Error('\p{:=	IS_JAVA}');
Error('\P{:=	IS_JAVA}');
Expect(1, 43487, '\p{isjava}', "");
Expect(0, 43487, '\p{^isjava}', "");
Expect(0, 43487, '\P{isjava}', "");
Expect(1, 43487, '\P{^isjava}', "");
Expect(0, 43488, '\p{isjava}', "");
Expect(1, 43488, '\p{^isjava}', "");
Expect(1, 43488, '\P{isjava}', "");
Expect(0, 43488, '\P{^isjava}', "");
Expect(1, 43487, '\p{_	IS_Java}', "");
Expect(0, 43487, '\p{^_	IS_Java}', "");
Expect(0, 43487, '\P{_	IS_Java}', "");
Expect(1, 43487, '\P{^_	IS_Java}', "");
Expect(0, 43488, '\p{_	IS_Java}', "");
Expect(1, 43488, '\p{^_	IS_Java}', "");
Expect(1, 43488, '\P{_	IS_Java}', "");
Expect(0, 43488, '\P{^_	IS_Java}', "");
Error('\p{-/a/telugu}');
Error('\P{-/a/telugu}');
Expect(1, 3199, '\p{telugu}', "");
Expect(0, 3199, '\p{^telugu}', "");
Expect(0, 3199, '\P{telugu}', "");
Expect(1, 3199, '\P{^telugu}', "");
Expect(0, 3200, '\p{telugu}', "");
Expect(1, 3200, '\p{^telugu}', "");
Expect(1, 3200, '\P{telugu}', "");
Expect(0, 3200, '\P{^telugu}', "");
Expect(1, 3199, '\p{_ TELUGU}', "");
Expect(0, 3199, '\p{^_ TELUGU}', "");
Expect(0, 3199, '\P{_ TELUGU}', "");
Expect(1, 3199, '\P{^_ TELUGU}', "");
Expect(0, 3200, '\p{_ TELUGU}', "");
Expect(1, 3200, '\p{^_ TELUGU}', "");
Expect(1, 3200, '\P{_ TELUGU}', "");
Expect(0, 3200, '\P{^_ TELUGU}', "");
Error('\p{_	is_Telugu/a/}');
Error('\P{_	is_Telugu/a/}');
Expect(1, 3199, '\p{istelugu}', "");
Expect(0, 3199, '\p{^istelugu}', "");
Expect(0, 3199, '\P{istelugu}', "");
Expect(1, 3199, '\P{^istelugu}', "");
Expect(0, 3200, '\p{istelugu}', "");
Expect(1, 3200, '\p{^istelugu}', "");
Expect(1, 3200, '\P{istelugu}', "");
Expect(0, 3200, '\P{^istelugu}', "");
Expect(1, 3199, '\p{ Is_Telugu}', "");
Expect(0, 3199, '\p{^ Is_Telugu}', "");
Expect(0, 3199, '\P{ Is_Telugu}', "");
Expect(1, 3199, '\P{^ Is_Telugu}', "");
Expect(0, 3200, '\p{ Is_Telugu}', "");
Expect(1, 3200, '\p{^ Is_Telugu}', "");
Expect(1, 3200, '\P{ Is_Telugu}', "");
Expect(0, 3200, '\P{^ Is_Telugu}', "");
Error('\p{__Telu/a/}');
Error('\P{__Telu/a/}');
Expect(1, 3199, '\p{telu}', "");
Expect(0, 3199, '\p{^telu}', "");
Expect(0, 3199, '\P{telu}', "");
Expect(1, 3199, '\P{^telu}', "");
Expect(0, 3200, '\p{telu}', "");
Expect(1, 3200, '\p{^telu}', "");
Expect(1, 3200, '\P{telu}', "");
Expect(0, 3200, '\P{^telu}', "");
Expect(1, 3199, '\p{__TELU}', "");
Expect(0, 3199, '\p{^__TELU}', "");
Expect(0, 3199, '\P{__TELU}', "");
Expect(1, 3199, '\P{^__TELU}', "");
Expect(0, 3200, '\p{__TELU}', "");
Expect(1, 3200, '\p{^__TELU}', "");
Expect(1, 3200, '\P{__TELU}', "");
Expect(0, 3200, '\P{^__TELU}', "");
Error('\p{--IS_TELU:=}');
Error('\P{--IS_TELU:=}');
Expect(1, 3199, '\p{istelu}', "");
Expect(0, 3199, '\p{^istelu}', "");
Expect(0, 3199, '\P{istelu}', "");
Expect(1, 3199, '\P{^istelu}', "");
Expect(0, 3200, '\p{istelu}', "");
Expect(1, 3200, '\p{^istelu}', "");
Expect(1, 3200, '\P{istelu}', "");
Expect(0, 3200, '\P{^istelu}', "");
Expect(1, 3199, '\p{ 	Is_Telu}', "");
Expect(0, 3199, '\p{^ 	Is_Telu}', "");
Expect(0, 3199, '\P{ 	Is_Telu}', "");
Expect(1, 3199, '\P{^ 	Is_Telu}', "");
Expect(0, 3200, '\p{ 	Is_Telu}', "");
Expect(1, 3200, '\p{^ 	Is_Telu}', "");
Expect(1, 3200, '\P{ 	Is_Telu}', "");
Expect(0, 3200, '\P{^ 	Is_Telu}', "");
Error('\p{/a/_	Yi_Radicals}');
Error('\P{/a/_	Yi_Radicals}');
Expect(1, 42191, '\p{yiradicals}', "");
Expect(0, 42191, '\p{^yiradicals}', "");
Expect(0, 42191, '\P{yiradicals}', "");
Expect(1, 42191, '\P{^yiradicals}', "");
Expect(0, 42192, '\p{yiradicals}', "");
Expect(1, 42192, '\p{^yiradicals}', "");
Expect(1, 42192, '\P{yiradicals}', "");
Expect(0, 42192, '\P{^yiradicals}', "");
Expect(1, 42191, '\p{  yi_Radicals}', "");
Expect(0, 42191, '\p{^  yi_Radicals}', "");
Expect(0, 42191, '\P{  yi_Radicals}', "");
Expect(1, 42191, '\P{^  yi_Radicals}', "");
Expect(0, 42192, '\p{  yi_Radicals}', "");
Expect(1, 42192, '\p{^  yi_Radicals}', "");
Expect(1, 42192, '\P{  yi_Radicals}', "");
Expect(0, 42192, '\P{^  yi_Radicals}', "");
Error('\p{/a/--is_YI_Radicals}');
Error('\P{/a/--is_YI_Radicals}');
Expect(1, 42191, '\p{isyiradicals}', "");
Expect(0, 42191, '\p{^isyiradicals}', "");
Expect(0, 42191, '\P{isyiradicals}', "");
Expect(1, 42191, '\P{^isyiradicals}', "");
Expect(0, 42192, '\p{isyiradicals}', "");
Expect(1, 42192, '\p{^isyiradicals}', "");
Expect(1, 42192, '\P{isyiradicals}', "");
Expect(0, 42192, '\P{^isyiradicals}', "");
Expect(1, 42191, '\p{- Is_YI_radicals}', "");
Expect(0, 42191, '\p{^- Is_YI_radicals}', "");
Expect(0, 42191, '\P{- Is_YI_radicals}', "");
Expect(1, 42191, '\P{^- Is_YI_radicals}', "");
Expect(0, 42192, '\p{- Is_YI_radicals}', "");
Expect(1, 42192, '\p{^- Is_YI_radicals}', "");
Expect(1, 42192, '\P{- Is_YI_radicals}', "");
Expect(0, 42192, '\P{^- Is_YI_radicals}', "");
Error('\p{	_IN_Yi_radicals:=}');
Error('\P{	_IN_Yi_radicals:=}');
Expect(1, 42191, '\p{inyiradicals}', "");
Expect(0, 42191, '\p{^inyiradicals}', "");
Expect(0, 42191, '\P{inyiradicals}', "");
Expect(1, 42191, '\P{^inyiradicals}', "");
Expect(0, 42192, '\p{inyiradicals}', "");
Expect(1, 42192, '\p{^inyiradicals}', "");
Expect(1, 42192, '\P{inyiradicals}', "");
Expect(0, 42192, '\P{^inyiradicals}', "");
Expect(1, 42191, '\p{ In_yi_Radicals}', "");
Expect(0, 42191, '\p{^ In_yi_Radicals}', "");
Expect(0, 42191, '\P{ In_yi_Radicals}', "");
Expect(1, 42191, '\P{^ In_yi_Radicals}', "");
Expect(0, 42192, '\p{ In_yi_Radicals}', "");
Expect(1, 42192, '\p{^ In_yi_Radicals}', "");
Expect(1, 42192, '\P{ In_yi_Radicals}', "");
Expect(0, 42192, '\P{^ In_yi_Radicals}', "");
Error('\p{	_In_Shavian/a/}');
Error('\P{	_In_Shavian/a/}');
Expect(1, 66687, '\p{inshavian}', "");
Expect(0, 66687, '\p{^inshavian}', "");
Expect(0, 66687, '\P{inshavian}', "");
Expect(1, 66687, '\P{^inshavian}', "");
Expect(0, 66688, '\p{inshavian}', "");
Expect(1, 66688, '\p{^inshavian}', "");
Expect(1, 66688, '\P{inshavian}', "");
Expect(0, 66688, '\P{^inshavian}', "");
Expect(1, 66687, '\p{	 in_Shavian}', "");
Expect(0, 66687, '\p{^	 in_Shavian}', "");
Expect(0, 66687, '\P{	 in_Shavian}', "");
Expect(1, 66687, '\P{^	 in_Shavian}', "");
Expect(0, 66688, '\p{	 in_Shavian}', "");
Expect(1, 66688, '\p{^	 in_Shavian}', "");
Expect(1, 66688, '\P{	 in_Shavian}', "");
Expect(0, 66688, '\P{^	 in_Shavian}', "");
Error('\p{-:=in_SHAVIAN}');
Error('\P{-:=in_SHAVIAN}');
Expect(1, 66687, '\p{	 in_shavian}', "");
Expect(0, 66687, '\p{^	 in_shavian}', "");
Expect(0, 66687, '\P{	 in_shavian}', "");
Expect(1, 66687, '\P{^	 in_shavian}', "");
Expect(0, 66688, '\p{	 in_shavian}', "");
Expect(1, 66688, '\p{^	 in_shavian}', "");
Expect(1, 66688, '\P{	 in_shavian}', "");
Expect(0, 66688, '\P{^	 in_shavian}', "");
Error('\p{	_Ancient_SYMBOLS:=}');
Error('\P{	_Ancient_SYMBOLS:=}');
Expect(1, 65999, '\p{ancientsymbols}', "");
Expect(0, 65999, '\p{^ancientsymbols}', "");
Expect(0, 65999, '\P{ancientsymbols}', "");
Expect(1, 65999, '\P{^ancientsymbols}', "");
Expect(0, 66000, '\p{ancientsymbols}', "");
Expect(1, 66000, '\p{^ancientsymbols}', "");
Expect(1, 66000, '\P{ancientsymbols}', "");
Expect(0, 66000, '\P{^ancientsymbols}', "");
Expect(1, 65999, '\p{	 ancient_Symbols}', "");
Expect(0, 65999, '\p{^	 ancient_Symbols}', "");
Expect(0, 65999, '\P{	 ancient_Symbols}', "");
Expect(1, 65999, '\P{^	 ancient_Symbols}', "");
Expect(0, 66000, '\p{	 ancient_Symbols}', "");
Expect(1, 66000, '\p{^	 ancient_Symbols}', "");
Expect(1, 66000, '\P{	 ancient_Symbols}', "");
Expect(0, 66000, '\P{^	 ancient_Symbols}', "");
Error('\p{_:=Is_Ancient_Symbols}');
Error('\P{_:=Is_Ancient_Symbols}');
Expect(1, 65999, '\p{isancientsymbols}', "");
Expect(0, 65999, '\p{^isancientsymbols}', "");
Expect(0, 65999, '\P{isancientsymbols}', "");
Expect(1, 65999, '\P{^isancientsymbols}', "");
Expect(0, 66000, '\p{isancientsymbols}', "");
Expect(1, 66000, '\p{^isancientsymbols}', "");
Expect(1, 66000, '\P{isancientsymbols}', "");
Expect(0, 66000, '\P{^isancientsymbols}', "");
Expect(1, 65999, '\p{	Is_ANCIENT_symbols}', "");
Expect(0, 65999, '\p{^	Is_ANCIENT_symbols}', "");
Expect(0, 65999, '\P{	Is_ANCIENT_symbols}', "");
Expect(1, 65999, '\P{^	Is_ANCIENT_symbols}', "");
Expect(0, 66000, '\p{	Is_ANCIENT_symbols}', "");
Expect(1, 66000, '\p{^	Is_ANCIENT_symbols}', "");
Expect(1, 66000, '\P{	Is_ANCIENT_symbols}', "");
Expect(0, 66000, '\P{^	Is_ANCIENT_symbols}', "");
Error('\p{-/a/In_Ancient_SYMBOLS}');
Error('\P{-/a/In_Ancient_SYMBOLS}');
Expect(1, 65999, '\p{inancientsymbols}', "");
Expect(0, 65999, '\p{^inancientsymbols}', "");
Expect(0, 65999, '\P{inancientsymbols}', "");
Expect(1, 65999, '\P{^inancientsymbols}', "");
Expect(0, 66000, '\p{inancientsymbols}', "");
Expect(1, 66000, '\p{^inancientsymbols}', "");
Expect(1, 66000, '\P{inancientsymbols}', "");
Expect(0, 66000, '\P{^inancientsymbols}', "");
Expect(1, 65999, '\p{	_in_Ancient_symbols}', "");
Expect(0, 65999, '\p{^	_in_Ancient_symbols}', "");
Expect(0, 65999, '\P{	_in_Ancient_symbols}', "");
Expect(1, 65999, '\P{^	_in_Ancient_symbols}', "");
Expect(0, 66000, '\p{	_in_Ancient_symbols}', "");
Expect(1, 66000, '\p{^	_in_Ancient_symbols}', "");
Expect(1, 66000, '\P{	_in_Ancient_symbols}', "");
Expect(0, 66000, '\P{^	_in_Ancient_symbols}', "");
Error('\p{	Quotation_mark:=}');
Error('\P{	Quotation_mark:=}');
Expect(1, 65379, '\p{quotationmark}', "");
Expect(0, 65379, '\p{^quotationmark}', "");
Expect(0, 65379, '\P{quotationmark}', "");
Expect(1, 65379, '\P{^quotationmark}', "");
Expect(0, 65380, '\p{quotationmark}', "");
Expect(1, 65380, '\p{^quotationmark}', "");
Expect(1, 65380, '\P{quotationmark}', "");
Expect(0, 65380, '\P{^quotationmark}', "");
Expect(1, 65379, '\p{- Quotation_Mark}', "");
Expect(0, 65379, '\p{^- Quotation_Mark}', "");
Expect(0, 65379, '\P{- Quotation_Mark}', "");
Expect(1, 65379, '\P{^- Quotation_Mark}', "");
Expect(0, 65380, '\p{- Quotation_Mark}', "");
Expect(1, 65380, '\p{^- Quotation_Mark}', "");
Expect(1, 65380, '\P{- Quotation_Mark}', "");
Expect(0, 65380, '\P{^- Quotation_Mark}', "");
Error('\p{-Is_QUOTATION_mark:=}');
Error('\P{-Is_QUOTATION_mark:=}');
Expect(1, 65379, '\p{isquotationmark}', "");
Expect(0, 65379, '\p{^isquotationmark}', "");
Expect(0, 65379, '\P{isquotationmark}', "");
Expect(1, 65379, '\P{^isquotationmark}', "");
Expect(0, 65380, '\p{isquotationmark}', "");
Expect(1, 65380, '\p{^isquotationmark}', "");
Expect(1, 65380, '\P{isquotationmark}', "");
Expect(0, 65380, '\P{^isquotationmark}', "");
Expect(1, 65379, '\p{-	IS_Quotation_Mark}', "");
Expect(0, 65379, '\p{^-	IS_Quotation_Mark}', "");
Expect(0, 65379, '\P{-	IS_Quotation_Mark}', "");
Expect(1, 65379, '\P{^-	IS_Quotation_Mark}', "");
Expect(0, 65380, '\p{-	IS_Quotation_Mark}', "");
Expect(1, 65380, '\p{^-	IS_Quotation_Mark}', "");
Expect(1, 65380, '\P{-	IS_Quotation_Mark}', "");
Expect(0, 65380, '\P{^-	IS_Quotation_Mark}', "");
Error('\p{	/a/qmark}');
Error('\P{	/a/qmark}');
Expect(1, 65379, '\p{qmark}', "");
Expect(0, 65379, '\p{^qmark}', "");
Expect(0, 65379, '\P{qmark}', "");
Expect(1, 65379, '\P{^qmark}', "");
Expect(0, 65380, '\p{qmark}', "");
Expect(1, 65380, '\p{^qmark}', "");
Expect(1, 65380, '\P{qmark}', "");
Expect(0, 65380, '\P{^qmark}', "");
Expect(1, 65379, '\p{-QMARK}', "");
Expect(0, 65379, '\p{^-QMARK}', "");
Expect(0, 65379, '\P{-QMARK}', "");
Expect(1, 65379, '\P{^-QMARK}', "");
Expect(0, 65380, '\p{-QMARK}', "");
Expect(1, 65380, '\p{^-QMARK}', "");
Expect(1, 65380, '\P{-QMARK}', "");
Expect(0, 65380, '\P{^-QMARK}', "");
Error('\p{	_Is_QMark/a/}');
Error('\P{	_Is_QMark/a/}');
Expect(1, 65379, '\p{isqmark}', "");
Expect(0, 65379, '\p{^isqmark}', "");
Expect(0, 65379, '\P{isqmark}', "");
Expect(1, 65379, '\P{^isqmark}', "");
Expect(0, 65380, '\p{isqmark}', "");
Expect(1, 65380, '\p{^isqmark}', "");
Expect(1, 65380, '\P{isqmark}', "");
Expect(0, 65380, '\P{^isqmark}', "");
Expect(1, 65379, '\p{ Is_qmark}', "");
Expect(0, 65379, '\p{^ Is_qmark}', "");
Expect(0, 65379, '\P{ Is_qmark}', "");
Expect(1, 65379, '\P{^ Is_qmark}', "");
Expect(0, 65380, '\p{ Is_qmark}', "");
Expect(1, 65380, '\p{^ Is_qmark}', "");
Expect(1, 65380, '\P{ Is_qmark}', "");
Expect(0, 65380, '\P{^ Is_qmark}', "");
Error('\p{	/a/in_tibetan}');
Error('\P{	/a/in_tibetan}');
Expect(1, 4095, '\p{intibetan}', "");
Expect(0, 4095, '\p{^intibetan}', "");
Expect(0, 4095, '\P{intibetan}', "");
Expect(1, 4095, '\P{^intibetan}', "");
Expect(0, 4096, '\p{intibetan}', "");
Expect(1, 4096, '\p{^intibetan}', "");
Expect(1, 4096, '\P{intibetan}', "");
Expect(0, 4096, '\P{^intibetan}', "");
Expect(1, 4095, '\p{ 	in_Tibetan}', "");
Expect(0, 4095, '\p{^ 	in_Tibetan}', "");
Expect(0, 4095, '\P{ 	in_Tibetan}', "");
Expect(1, 4095, '\P{^ 	in_Tibetan}', "");
Expect(0, 4096, '\p{ 	in_Tibetan}', "");
Expect(1, 4096, '\p{^ 	in_Tibetan}', "");
Expect(1, 4096, '\P{ 	in_Tibetan}', "");
Expect(0, 4096, '\P{^ 	in_Tibetan}', "");
Error('\p{/a/	IN_Tibetan}');
Error('\P{/a/	IN_Tibetan}');
Expect(1, 4095, '\p{_IN_Tibetan}', "");
Expect(0, 4095, '\p{^_IN_Tibetan}', "");
Expect(0, 4095, '\P{_IN_Tibetan}', "");
Expect(1, 4095, '\P{^_IN_Tibetan}', "");
Expect(0, 4096, '\p{_IN_Tibetan}', "");
Expect(1, 4096, '\p{^_IN_Tibetan}', "");
Expect(1, 4096, '\P{_IN_Tibetan}', "");
Expect(0, 4096, '\P{^_IN_Tibetan}', "");
Error('\p{_Tai_Xuan_JING_Symbols:=}');
Error('\P{_Tai_Xuan_JING_Symbols:=}');
Expect(1, 119647, '\p{taixuanjingsymbols}', "");
Expect(0, 119647, '\p{^taixuanjingsymbols}', "");
Expect(0, 119647, '\P{taixuanjingsymbols}', "");
Expect(1, 119647, '\P{^taixuanjingsymbols}', "");
Expect(0, 119648, '\p{taixuanjingsymbols}', "");
Expect(1, 119648, '\p{^taixuanjingsymbols}', "");
Expect(1, 119648, '\P{taixuanjingsymbols}', "");
Expect(0, 119648, '\P{^taixuanjingsymbols}', "");
Expect(1, 119647, '\p{ -TAI_XUAN_Jing_Symbols}', "");
Expect(0, 119647, '\p{^ -TAI_XUAN_Jing_Symbols}', "");
Expect(0, 119647, '\P{ -TAI_XUAN_Jing_Symbols}', "");
Expect(1, 119647, '\P{^ -TAI_XUAN_Jing_Symbols}', "");
Expect(0, 119648, '\p{ -TAI_XUAN_Jing_Symbols}', "");
Expect(1, 119648, '\p{^ -TAI_XUAN_Jing_Symbols}', "");
Expect(1, 119648, '\P{ -TAI_XUAN_Jing_Symbols}', "");
Expect(0, 119648, '\P{^ -TAI_XUAN_Jing_Symbols}', "");
Error('\p{:=is_tai_XUAN_JING_SYMBOLS}');
Error('\P{:=is_tai_XUAN_JING_SYMBOLS}');
Expect(1, 119647, '\p{istaixuanjingsymbols}', "");
Expect(0, 119647, '\p{^istaixuanjingsymbols}', "");
Expect(0, 119647, '\P{istaixuanjingsymbols}', "");
Expect(1, 119647, '\P{^istaixuanjingsymbols}', "");
Expect(0, 119648, '\p{istaixuanjingsymbols}', "");
Expect(1, 119648, '\p{^istaixuanjingsymbols}', "");
Expect(1, 119648, '\P{istaixuanjingsymbols}', "");
Expect(0, 119648, '\P{^istaixuanjingsymbols}', "");
Expect(1, 119647, '\p{_ Is_tai_XUAN_JING_symbols}', "");
Expect(0, 119647, '\p{^_ Is_tai_XUAN_JING_symbols}', "");
Expect(0, 119647, '\P{_ Is_tai_XUAN_JING_symbols}', "");
Expect(1, 119647, '\P{^_ Is_tai_XUAN_JING_symbols}', "");
Expect(0, 119648, '\p{_ Is_tai_XUAN_JING_symbols}', "");
Expect(1, 119648, '\p{^_ Is_tai_XUAN_JING_symbols}', "");
Expect(1, 119648, '\P{_ Is_tai_XUAN_JING_symbols}', "");
Expect(0, 119648, '\P{^_ Is_tai_XUAN_JING_symbols}', "");
Error('\p{-	in_Tai_Xuan_jing_Symbols:=}');
Error('\P{-	in_Tai_Xuan_jing_Symbols:=}');
Expect(1, 119647, '\p{intaixuanjingsymbols}', "");
Expect(0, 119647, '\p{^intaixuanjingsymbols}', "");
Expect(0, 119647, '\P{intaixuanjingsymbols}', "");
Expect(1, 119647, '\P{^intaixuanjingsymbols}', "");
Expect(0, 119648, '\p{intaixuanjingsymbols}', "");
Expect(1, 119648, '\p{^intaixuanjingsymbols}', "");
Expect(1, 119648, '\P{intaixuanjingsymbols}', "");
Expect(0, 119648, '\P{^intaixuanjingsymbols}', "");
Expect(1, 119647, '\p{-In_tai_XUAN_Jing_Symbols}', "");
Expect(0, 119647, '\p{^-In_tai_XUAN_Jing_Symbols}', "");
Expect(0, 119647, '\P{-In_tai_XUAN_Jing_Symbols}', "");
Expect(1, 119647, '\P{^-In_tai_XUAN_Jing_Symbols}', "");
Expect(0, 119648, '\p{-In_tai_XUAN_Jing_Symbols}', "");
Expect(1, 119648, '\p{^-In_tai_XUAN_Jing_Symbols}', "");
Expect(1, 119648, '\P{-In_tai_XUAN_Jing_Symbols}', "");
Expect(0, 119648, '\P{^-In_tai_XUAN_Jing_Symbols}', "");
Error('\p{	:=TAI_XUAN_Jing}');
Error('\P{	:=TAI_XUAN_Jing}');
Expect(1, 119647, '\p{taixuanjing}', "");
Expect(0, 119647, '\p{^taixuanjing}', "");
Expect(0, 119647, '\P{taixuanjing}', "");
Expect(1, 119647, '\P{^taixuanjing}', "");
Expect(0, 119648, '\p{taixuanjing}', "");
Expect(1, 119648, '\p{^taixuanjing}', "");
Expect(1, 119648, '\P{taixuanjing}', "");
Expect(0, 119648, '\P{^taixuanjing}', "");
Expect(1, 119647, '\p{_Tai_Xuan_JING}', "");
Expect(0, 119647, '\p{^_Tai_Xuan_JING}', "");
Expect(0, 119647, '\P{_Tai_Xuan_JING}', "");
Expect(1, 119647, '\P{^_Tai_Xuan_JING}', "");
Expect(0, 119648, '\p{_Tai_Xuan_JING}', "");
Expect(1, 119648, '\p{^_Tai_Xuan_JING}', "");
Expect(1, 119648, '\P{_Tai_Xuan_JING}', "");
Expect(0, 119648, '\P{^_Tai_Xuan_JING}', "");
Error('\p{/a/Is_TAI_Xuan_Jing}');
Error('\P{/a/Is_TAI_Xuan_Jing}');
Expect(1, 119647, '\p{istaixuanjing}', "");
Expect(0, 119647, '\p{^istaixuanjing}', "");
Expect(0, 119647, '\P{istaixuanjing}', "");
Expect(1, 119647, '\P{^istaixuanjing}', "");
Expect(0, 119648, '\p{istaixuanjing}', "");
Expect(1, 119648, '\p{^istaixuanjing}', "");
Expect(1, 119648, '\P{istaixuanjing}', "");
Expect(0, 119648, '\P{^istaixuanjing}', "");
Expect(1, 119647, '\p{_is_Tai_XUAN_jing}', "");
Expect(0, 119647, '\p{^_is_Tai_XUAN_jing}', "");
Expect(0, 119647, '\P{_is_Tai_XUAN_jing}', "");
Expect(1, 119647, '\P{^_is_Tai_XUAN_jing}', "");
Expect(0, 119648, '\p{_is_Tai_XUAN_jing}', "");
Expect(1, 119648, '\p{^_is_Tai_XUAN_jing}', "");
Expect(1, 119648, '\P{_is_Tai_XUAN_jing}', "");
Expect(0, 119648, '\P{^_is_Tai_XUAN_jing}', "");
Error('\p{_/a/in_Tai_Xuan_JING}');
Error('\P{_/a/in_Tai_Xuan_JING}');
Expect(1, 119647, '\p{intaixuanjing}', "");
Expect(0, 119647, '\p{^intaixuanjing}', "");
Expect(0, 119647, '\P{intaixuanjing}', "");
Expect(1, 119647, '\P{^intaixuanjing}', "");
Expect(0, 119648, '\p{intaixuanjing}', "");
Expect(1, 119648, '\p{^intaixuanjing}', "");
Expect(1, 119648, '\P{intaixuanjing}', "");
Expect(0, 119648, '\P{^intaixuanjing}', "");
Expect(1, 119647, '\p{-	In_tai_XUAN_JING}', "");
Expect(0, 119647, '\p{^-	In_tai_XUAN_JING}', "");
Expect(0, 119647, '\P{-	In_tai_XUAN_JING}', "");
Expect(1, 119647, '\P{^-	In_tai_XUAN_JING}', "");
Expect(0, 119648, '\p{-	In_tai_XUAN_JING}', "");
Expect(1, 119648, '\p{^-	In_tai_XUAN_JING}', "");
Expect(1, 119648, '\P{-	In_tai_XUAN_JING}', "");
Expect(0, 119648, '\P{^-	In_tai_XUAN_JING}', "");
Error('\p{_In_Carian/a/}');
Error('\P{_In_Carian/a/}');
Expect(1, 66271, '\p{incarian}', "");
Expect(0, 66271, '\p{^incarian}', "");
Expect(0, 66271, '\P{incarian}', "");
Expect(1, 66271, '\P{^incarian}', "");
Expect(0, 66272, '\p{incarian}', "");
Expect(1, 66272, '\p{^incarian}', "");
Expect(1, 66272, '\P{incarian}', "");
Expect(0, 66272, '\P{^incarian}', "");
Expect(1, 66271, '\p{	In_Carian}', "");
Expect(0, 66271, '\p{^	In_Carian}', "");
Expect(0, 66271, '\P{	In_Carian}', "");
Expect(1, 66271, '\P{^	In_Carian}', "");
Expect(0, 66272, '\p{	In_Carian}', "");
Expect(1, 66272, '\p{^	In_Carian}', "");
Expect(1, 66272, '\P{	In_Carian}', "");
Expect(0, 66272, '\P{^	In_Carian}', "");
Error('\p{ /a/In_CARIAN}');
Error('\P{ /a/In_CARIAN}');
Expect(1, 66271, '\p{-_IN_Carian}', "");
Expect(0, 66271, '\p{^-_IN_Carian}', "");
Expect(0, 66271, '\P{-_IN_Carian}', "");
Expect(1, 66271, '\P{^-_IN_Carian}', "");
Expect(0, 66272, '\p{-_IN_Carian}', "");
Expect(1, 66272, '\p{^-_IN_Carian}', "");
Expect(1, 66272, '\P{-_IN_Carian}', "");
Expect(0, 66272, '\P{^-_IN_Carian}', "");
Error('\p{:= _In_ARABIC}');
Error('\P{:= _In_ARABIC}');
Expect(1, 1791, '\p{inarabic}', "");
Expect(0, 1791, '\p{^inarabic}', "");
Expect(0, 1791, '\P{inarabic}', "");
Expect(1, 1791, '\P{^inarabic}', "");
Expect(0, 1792, '\p{inarabic}', "");
Expect(1, 1792, '\p{^inarabic}', "");
Expect(1, 1792, '\P{inarabic}', "");
Expect(0, 1792, '\P{^inarabic}', "");
Expect(1, 1791, '\p{-In_Arabic}', "");
Expect(0, 1791, '\p{^-In_Arabic}', "");
Expect(0, 1791, '\P{-In_Arabic}', "");
Expect(1, 1791, '\P{^-In_Arabic}', "");
Expect(0, 1792, '\p{-In_Arabic}', "");
Expect(1, 1792, '\p{^-In_Arabic}', "");
Expect(1, 1792, '\P{-In_Arabic}', "");
Expect(0, 1792, '\P{^-In_Arabic}', "");
Error('\p{/a/ In_Arabic}');
Error('\P{/a/ In_Arabic}');
Expect(1, 1791, '\p{	In_Arabic}', "");
Expect(0, 1791, '\p{^	In_Arabic}', "");
Expect(0, 1791, '\P{	In_Arabic}', "");
Expect(1, 1791, '\P{^	In_Arabic}', "");
Expect(0, 1792, '\p{	In_Arabic}', "");
Expect(1, 1792, '\p{^	In_Arabic}', "");
Expect(1, 1792, '\P{	In_Arabic}', "");
Expect(0, 1792, '\P{^	In_Arabic}', "");
Error('\p{/a/_-PosixCntrl}');
Error('\P{/a/_-PosixCntrl}');
Expect(1, 127, '\p{posixcntrl}', "");
Expect(0, 127, '\p{^posixcntrl}', "");
Expect(0, 127, '\P{posixcntrl}', "");
Expect(1, 127, '\P{^posixcntrl}', "");
Expect(0, 128, '\p{posixcntrl}', "");
Expect(1, 128, '\p{^posixcntrl}', "");
Expect(1, 128, '\P{posixcntrl}', "");
Expect(0, 128, '\P{^posixcntrl}', "");
Expect(1, 127, '\p{ -PosixCntrl}', "");
Expect(0, 127, '\p{^ -PosixCntrl}', "");
Expect(0, 127, '\P{ -PosixCntrl}', "");
Expect(1, 127, '\P{^ -PosixCntrl}', "");
Expect(0, 128, '\p{ -PosixCntrl}', "");
Expect(1, 128, '\p{^ -PosixCntrl}', "");
Expect(1, 128, '\P{ -PosixCntrl}', "");
Expect(0, 128, '\P{^ -PosixCntrl}', "");
Error('\p{ IS_POSIXCNTRL:=}');
Error('\P{ IS_POSIXCNTRL:=}');
Expect(1, 127, '\p{isposixcntrl}', "");
Expect(0, 127, '\p{^isposixcntrl}', "");
Expect(0, 127, '\P{isposixcntrl}', "");
Expect(1, 127, '\P{^isposixcntrl}', "");
Expect(0, 128, '\p{isposixcntrl}', "");
Expect(1, 128, '\p{^isposixcntrl}', "");
Expect(1, 128, '\P{isposixcntrl}', "");
Expect(0, 128, '\P{^isposixcntrl}', "");
Expect(1, 127, '\p{	-IS_PosixCntrl}', "");
Expect(0, 127, '\p{^	-IS_PosixCntrl}', "");
Expect(0, 127, '\P{	-IS_PosixCntrl}', "");
Expect(1, 127, '\P{^	-IS_PosixCntrl}', "");
Expect(0, 128, '\p{	-IS_PosixCntrl}', "");
Expect(1, 128, '\p{^	-IS_PosixCntrl}', "");
Expect(1, 128, '\P{	-IS_PosixCntrl}', "");
Expect(0, 128, '\P{^	-IS_PosixCntrl}', "");
Error('\p{:=	 In_GOTHIC}');
Error('\P{:=	 In_GOTHIC}');
Expect(1, 66383, '\p{ingothic}', "");
Expect(0, 66383, '\p{^ingothic}', "");
Expect(0, 66383, '\P{ingothic}', "");
Expect(1, 66383, '\P{^ingothic}', "");
Expect(0, 66384, '\p{ingothic}', "");
Expect(1, 66384, '\p{^ingothic}', "");
Expect(1, 66384, '\P{ingothic}', "");
Expect(0, 66384, '\P{^ingothic}', "");
Expect(1, 66383, '\p{- In_gothic}', "");
Expect(0, 66383, '\p{^- In_gothic}', "");
Expect(0, 66383, '\P{- In_gothic}', "");
Expect(1, 66383, '\P{^- In_gothic}', "");
Expect(0, 66384, '\p{- In_gothic}', "");
Expect(1, 66384, '\p{^- In_gothic}', "");
Expect(1, 66384, '\P{- In_gothic}', "");
Expect(0, 66384, '\P{^- In_gothic}', "");
Error('\p{--In_Gothic/a/}');
Error('\P{--In_Gothic/a/}');
Expect(1, 66383, '\p{ _in_gothic}', "");
Expect(0, 66383, '\p{^ _in_gothic}', "");
Expect(0, 66383, '\P{ _in_gothic}', "");
Expect(1, 66383, '\P{^ _in_gothic}', "");
Expect(0, 66384, '\p{ _in_gothic}', "");
Expect(1, 66384, '\p{^ _in_gothic}', "");
Expect(1, 66384, '\P{ _in_gothic}', "");
Expect(0, 66384, '\P{^ _in_gothic}', "");
Error('\p{_Sora_Sompeng:=}');
Error('\P{_Sora_Sompeng:=}');
Expect(1, 69881, '\p{sorasompeng}', "");
Expect(0, 69881, '\p{^sorasompeng}', "");
Expect(0, 69881, '\P{sorasompeng}', "");
Expect(1, 69881, '\P{^sorasompeng}', "");
Expect(0, 69882, '\p{sorasompeng}', "");
Expect(1, 69882, '\p{^sorasompeng}', "");
Expect(1, 69882, '\P{sorasompeng}', "");
Expect(0, 69882, '\P{^sorasompeng}', "");
Expect(1, 69881, '\p{	 sora_sompeng}', "");
Expect(0, 69881, '\p{^	 sora_sompeng}', "");
Expect(0, 69881, '\P{	 sora_sompeng}', "");
Expect(1, 69881, '\P{^	 sora_sompeng}', "");
Expect(0, 69882, '\p{	 sora_sompeng}', "");
Expect(1, 69882, '\p{^	 sora_sompeng}', "");
Expect(1, 69882, '\P{	 sora_sompeng}', "");
Expect(0, 69882, '\P{^	 sora_sompeng}', "");
Error('\p{_	Is_sora_Sompeng/a/}');
Error('\P{_	Is_sora_Sompeng/a/}');
Expect(1, 69881, '\p{issorasompeng}', "");
Expect(0, 69881, '\p{^issorasompeng}', "");
Expect(0, 69881, '\P{issorasompeng}', "");
Expect(1, 69881, '\P{^issorasompeng}', "");
Expect(0, 69882, '\p{issorasompeng}', "");
Expect(1, 69882, '\p{^issorasompeng}', "");
Expect(1, 69882, '\P{issorasompeng}', "");
Expect(0, 69882, '\P{^issorasompeng}', "");
Expect(1, 69881, '\p{__Is_Sora_Sompeng}', "");
Expect(0, 69881, '\p{^__Is_Sora_Sompeng}', "");
Expect(0, 69881, '\P{__Is_Sora_Sompeng}', "");
Expect(1, 69881, '\P{^__Is_Sora_Sompeng}', "");
Expect(0, 69882, '\p{__Is_Sora_Sompeng}', "");
Expect(1, 69882, '\p{^__Is_Sora_Sompeng}', "");
Expect(1, 69882, '\P{__Is_Sora_Sompeng}', "");
Expect(0, 69882, '\P{^__Is_Sora_Sompeng}', "");
Error('\p{_ Sora:=}');
Error('\P{_ Sora:=}');
Expect(1, 69881, '\p{sora}', "");
Expect(0, 69881, '\p{^sora}', "");
Expect(0, 69881, '\P{sora}', "");
Expect(1, 69881, '\P{^sora}', "");
Expect(0, 69882, '\p{sora}', "");
Expect(1, 69882, '\p{^sora}', "");
Expect(1, 69882, '\P{sora}', "");
Expect(0, 69882, '\P{^sora}', "");
Expect(1, 69881, '\p{		sora}', "");
Expect(0, 69881, '\p{^		sora}', "");
Expect(0, 69881, '\P{		sora}', "");
Expect(1, 69881, '\P{^		sora}', "");
Expect(0, 69882, '\p{		sora}', "");
Expect(1, 69882, '\p{^		sora}', "");
Expect(1, 69882, '\P{		sora}', "");
Expect(0, 69882, '\P{^		sora}', "");
Error('\p{:=-_Is_sora}');
Error('\P{:=-_Is_sora}');
Expect(1, 69881, '\p{issora}', "");
Expect(0, 69881, '\p{^issora}', "");
Expect(0, 69881, '\P{issora}', "");
Expect(1, 69881, '\P{^issora}', "");
Expect(0, 69882, '\p{issora}', "");
Expect(1, 69882, '\p{^issora}', "");
Expect(1, 69882, '\P{issora}', "");
Expect(0, 69882, '\P{^issora}', "");
Expect(1, 69881, '\p{IS_sora}', "");
Expect(0, 69881, '\p{^IS_sora}', "");
Expect(0, 69881, '\P{IS_sora}', "");
Expect(1, 69881, '\P{^IS_sora}', "");
Expect(0, 69882, '\p{IS_sora}', "");
Expect(1, 69882, '\p{^IS_sora}', "");
Expect(1, 69882, '\P{IS_sora}', "");
Expect(0, 69882, '\P{^IS_sora}', "");
Error('\p{:=  In_Telugu}');
Error('\P{:=  In_Telugu}');
Expect(1, 3199, '\p{intelugu}', "");
Expect(0, 3199, '\p{^intelugu}', "");
Expect(0, 3199, '\P{intelugu}', "");
Expect(1, 3199, '\P{^intelugu}', "");
Expect(0, 3200, '\p{intelugu}', "");
Expect(1, 3200, '\p{^intelugu}', "");
Expect(1, 3200, '\P{intelugu}', "");
Expect(0, 3200, '\P{^intelugu}', "");
Expect(1, 3199, '\p{_-IN_TELUGU}', "");
Expect(0, 3199, '\p{^_-IN_TELUGU}', "");
Expect(0, 3199, '\P{_-IN_TELUGU}', "");
Expect(1, 3199, '\P{^_-IN_TELUGU}', "");
Expect(0, 3200, '\p{_-IN_TELUGU}', "");
Expect(1, 3200, '\p{^_-IN_TELUGU}', "");
Expect(1, 3200, '\P{_-IN_TELUGU}', "");
Expect(0, 3200, '\P{^_-IN_TELUGU}', "");
Error('\p{	-In_TELUGU/a/}');
Error('\P{	-In_TELUGU/a/}');
Expect(1, 3199, '\p{_-In_Telugu}', "");
Expect(0, 3199, '\p{^_-In_Telugu}', "");
Expect(0, 3199, '\P{_-In_Telugu}', "");
Expect(1, 3199, '\P{^_-In_Telugu}', "");
Expect(0, 3200, '\p{_-In_Telugu}', "");
Expect(1, 3200, '\p{^_-In_Telugu}', "");
Expect(1, 3200, '\P{_-In_Telugu}', "");
Expect(0, 3200, '\P{^_-In_Telugu}', "");
Error('\p{_	In_TAI_Viet:=}');
Error('\P{_	In_TAI_Viet:=}');
Expect(1, 43743, '\p{intaiviet}', "");
Expect(0, 43743, '\p{^intaiviet}', "");
Expect(0, 43743, '\P{intaiviet}', "");
Expect(1, 43743, '\P{^intaiviet}', "");
Expect(0, 43744, '\p{intaiviet}', "");
Expect(1, 43744, '\p{^intaiviet}', "");
Expect(1, 43744, '\P{intaiviet}', "");
Expect(0, 43744, '\P{^intaiviet}', "");
Expect(1, 43743, '\p{--in_Tai_Viet}', "");
Expect(0, 43743, '\p{^--in_Tai_Viet}', "");
Expect(0, 43743, '\P{--in_Tai_Viet}', "");
Expect(1, 43743, '\P{^--in_Tai_Viet}', "");
Expect(0, 43744, '\p{--in_Tai_Viet}', "");
Expect(1, 43744, '\p{^--in_Tai_Viet}', "");
Expect(1, 43744, '\P{--in_Tai_Viet}', "");
Expect(0, 43744, '\P{^--in_Tai_Viet}', "");
Error('\p{:= -In_Tai_Viet}');
Error('\P{:= -In_Tai_Viet}');
Expect(1, 43743, '\p{-_In_TAI_Viet}', "");
Expect(0, 43743, '\p{^-_In_TAI_Viet}', "");
Expect(0, 43743, '\P{-_In_TAI_Viet}', "");
Expect(1, 43743, '\P{^-_In_TAI_Viet}', "");
Expect(0, 43744, '\p{-_In_TAI_Viet}', "");
Expect(1, 43744, '\p{^-_In_TAI_Viet}', "");
Expect(1, 43744, '\P{-_In_TAI_Viet}', "");
Expect(0, 43744, '\P{^-_In_TAI_Viet}', "");
Error('\p{/a/ IN_Syloti_NAGRI}');
Error('\P{/a/ IN_Syloti_NAGRI}');
Expect(1, 43055, '\p{insylotinagri}', "");
Expect(0, 43055, '\p{^insylotinagri}', "");
Expect(0, 43055, '\P{insylotinagri}', "");
Expect(1, 43055, '\P{^insylotinagri}', "");
Expect(0, 43056, '\p{insylotinagri}', "");
Expect(1, 43056, '\p{^insylotinagri}', "");
Expect(1, 43056, '\P{insylotinagri}', "");
Expect(0, 43056, '\P{^insylotinagri}', "");
Expect(1, 43055, '\p{-	In_Syloti_nagri}', "");
Expect(0, 43055, '\p{^-	In_Syloti_nagri}', "");
Expect(0, 43055, '\P{-	In_Syloti_nagri}', "");
Expect(1, 43055, '\P{^-	In_Syloti_nagri}', "");
Expect(0, 43056, '\p{-	In_Syloti_nagri}', "");
Expect(1, 43056, '\p{^-	In_Syloti_nagri}', "");
Expect(1, 43056, '\P{-	In_Syloti_nagri}', "");
Expect(0, 43056, '\P{^-	In_Syloti_nagri}', "");
Error('\p{	:=IN_Syloti_Nagri}');
Error('\P{	:=IN_Syloti_Nagri}');
Expect(1, 43055, '\p{__IN_Syloti_NAGRI}', "");
Expect(0, 43055, '\p{^__IN_Syloti_NAGRI}', "");
Expect(0, 43055, '\P{__IN_Syloti_NAGRI}', "");
Expect(1, 43055, '\P{^__IN_Syloti_NAGRI}', "");
Expect(0, 43056, '\p{__IN_Syloti_NAGRI}', "");
Expect(1, 43056, '\p{^__IN_Syloti_NAGRI}', "");
Expect(1, 43056, '\P{__IN_Syloti_NAGRI}', "");
Expect(0, 43056, '\P{^__IN_Syloti_NAGRI}', "");
Error('\p{:=_kaithi}');
Error('\P{:=_kaithi}');
Expect(1, 69825, '\p{kaithi}', "");
Expect(0, 69825, '\p{^kaithi}', "");
Expect(0, 69825, '\P{kaithi}', "");
Expect(1, 69825, '\P{^kaithi}', "");
Expect(0, 69826, '\p{kaithi}', "");
Expect(1, 69826, '\p{^kaithi}', "");
Expect(1, 69826, '\P{kaithi}', "");
Expect(0, 69826, '\P{^kaithi}', "");
Expect(1, 69825, '\p{ Kaithi}', "");
Expect(0, 69825, '\p{^ Kaithi}', "");
Expect(0, 69825, '\P{ Kaithi}', "");
Expect(1, 69825, '\P{^ Kaithi}', "");
Expect(0, 69826, '\p{ Kaithi}', "");
Expect(1, 69826, '\p{^ Kaithi}', "");
Expect(1, 69826, '\P{ Kaithi}', "");
Expect(0, 69826, '\P{^ Kaithi}', "");
Error('\p{:=		is_KAITHI}');
Error('\P{:=		is_KAITHI}');
Expect(1, 69825, '\p{iskaithi}', "");
Expect(0, 69825, '\p{^iskaithi}', "");
Expect(0, 69825, '\P{iskaithi}', "");
Expect(1, 69825, '\P{^iskaithi}', "");
Expect(0, 69826, '\p{iskaithi}', "");
Expect(1, 69826, '\p{^iskaithi}', "");
Expect(1, 69826, '\P{iskaithi}', "");
Expect(0, 69826, '\P{^iskaithi}', "");
Expect(1, 69825, '\p{__Is_KAITHI}', "");
Expect(0, 69825, '\p{^__Is_KAITHI}', "");
Expect(0, 69825, '\P{__Is_KAITHI}', "");
Expect(1, 69825, '\P{^__Is_KAITHI}', "");
Expect(0, 69826, '\p{__Is_KAITHI}', "");
Expect(1, 69826, '\p{^__Is_KAITHI}', "");
Expect(1, 69826, '\P{__Is_KAITHI}', "");
Expect(0, 69826, '\P{^__Is_KAITHI}', "");
Error('\p{/a/-	KTHI}');
Error('\P{/a/-	KTHI}');
Expect(1, 69825, '\p{kthi}', "");
Expect(0, 69825, '\p{^kthi}', "");
Expect(0, 69825, '\P{kthi}', "");
Expect(1, 69825, '\P{^kthi}', "");
Expect(0, 69826, '\p{kthi}', "");
Expect(1, 69826, '\p{^kthi}', "");
Expect(1, 69826, '\P{kthi}', "");
Expect(0, 69826, '\P{^kthi}', "");
Expect(1, 69825, '\p{_KTHI}', "");
Expect(0, 69825, '\p{^_KTHI}', "");
Expect(0, 69825, '\P{_KTHI}', "");
Expect(1, 69825, '\P{^_KTHI}', "");
Expect(0, 69826, '\p{_KTHI}', "");
Expect(1, 69826, '\p{^_KTHI}', "");
Expect(1, 69826, '\P{_KTHI}', "");
Expect(0, 69826, '\P{^_KTHI}', "");
Error('\p{/a/- IS_Kthi}');
Error('\P{/a/- IS_Kthi}');
Expect(1, 69825, '\p{iskthi}', "");
Expect(0, 69825, '\p{^iskthi}', "");
Expect(0, 69825, '\P{iskthi}', "");
Expect(1, 69825, '\P{^iskthi}', "");
Expect(0, 69826, '\p{iskthi}', "");
Expect(1, 69826, '\p{^iskthi}', "");
Expect(1, 69826, '\P{iskthi}', "");
Expect(0, 69826, '\P{^iskthi}', "");
Expect(1, 69825, '\p{	-IS_Kthi}', "");
Expect(0, 69825, '\p{^	-IS_Kthi}', "");
Expect(0, 69825, '\P{	-IS_Kthi}', "");
Expect(1, 69825, '\P{^	-IS_Kthi}', "");
Expect(0, 69826, '\p{	-IS_Kthi}', "");
Expect(1, 69826, '\p{^	-IS_Kthi}', "");
Expect(1, 69826, '\P{	-IS_Kthi}', "");
Expect(0, 69826, '\P{^	-IS_Kthi}', "");
Error('\p{	:=In_Old_ITALIC}');
Error('\P{	:=In_Old_ITALIC}');
Expect(1, 66351, '\p{inolditalic}', "");
Expect(0, 66351, '\p{^inolditalic}', "");
Expect(0, 66351, '\P{inolditalic}', "");
Expect(1, 66351, '\P{^inolditalic}', "");
Expect(0, 66352, '\p{inolditalic}', "");
Expect(1, 66352, '\p{^inolditalic}', "");
Expect(1, 66352, '\P{inolditalic}', "");
Expect(0, 66352, '\P{^inolditalic}', "");
Expect(1, 66351, '\p{-	IN_Old_ITALIC}', "");
Expect(0, 66351, '\p{^-	IN_Old_ITALIC}', "");
Expect(0, 66351, '\P{-	IN_Old_ITALIC}', "");
Expect(1, 66351, '\P{^-	IN_Old_ITALIC}', "");
Expect(0, 66352, '\p{-	IN_Old_ITALIC}', "");
Expect(1, 66352, '\p{^-	IN_Old_ITALIC}', "");
Expect(1, 66352, '\P{-	IN_Old_ITALIC}', "");
Expect(0, 66352, '\P{^-	IN_Old_ITALIC}', "");
Error('\p{/a/In_Old_italic}');
Error('\P{/a/In_Old_italic}');
Expect(1, 66351, '\p{-_in_Old_Italic}', "");
Expect(0, 66351, '\p{^-_in_Old_Italic}', "");
Expect(0, 66351, '\P{-_in_Old_Italic}', "");
Expect(1, 66351, '\P{^-_in_Old_Italic}', "");
Expect(0, 66352, '\p{-_in_Old_Italic}', "");
Expect(1, 66352, '\p{^-_in_Old_Italic}', "");
Expect(1, 66352, '\P{-_in_Old_Italic}', "");
Expect(0, 66352, '\P{^-_in_Old_Italic}', "");
Error('\p{_In_avestan/a/}');
Error('\P{_In_avestan/a/}');
Expect(1, 68415, '\p{inavestan}', "");
Expect(0, 68415, '\p{^inavestan}', "");
Expect(0, 68415, '\P{inavestan}', "");
Expect(1, 68415, '\P{^inavestan}', "");
Expect(0, 68416, '\p{inavestan}', "");
Expect(1, 68416, '\p{^inavestan}', "");
Expect(1, 68416, '\P{inavestan}', "");
Expect(0, 68416, '\P{^inavestan}', "");
Expect(1, 68415, '\p{_In_avestan}', "");
Expect(0, 68415, '\p{^_In_avestan}', "");
Expect(0, 68415, '\P{_In_avestan}', "");
Expect(1, 68415, '\P{^_In_avestan}', "");
Expect(0, 68416, '\p{_In_avestan}', "");
Expect(1, 68416, '\p{^_In_avestan}', "");
Expect(1, 68416, '\P{_In_avestan}', "");
Expect(0, 68416, '\P{^_In_avestan}', "");
Error('\p{-:=IN_avestan}');
Error('\P{-:=IN_avestan}');
Expect(1, 68415, '\p{-IN_Avestan}', "");
Expect(0, 68415, '\p{^-IN_Avestan}', "");
Expect(0, 68415, '\P{-IN_Avestan}', "");
Expect(1, 68415, '\P{^-IN_Avestan}', "");
Expect(0, 68416, '\p{-IN_Avestan}', "");
Expect(1, 68416, '\p{^-IN_Avestan}', "");
Expect(1, 68416, '\P{-IN_Avestan}', "");
Expect(0, 68416, '\P{^-IN_Avestan}', "");
Error('\p{:=-	IN_Javanese}');
Error('\P{:=-	IN_Javanese}');
Expect(1, 43487, '\p{injavanese}', "");
Expect(0, 43487, '\p{^injavanese}', "");
Expect(0, 43487, '\P{injavanese}', "");
Expect(1, 43487, '\P{^injavanese}', "");
Expect(0, 43488, '\p{injavanese}', "");
Expect(1, 43488, '\p{^injavanese}', "");
Expect(1, 43488, '\P{injavanese}', "");
Expect(0, 43488, '\P{^injavanese}', "");
Expect(1, 43487, '\p{ -In_Javanese}', "");
Expect(0, 43487, '\p{^ -In_Javanese}', "");
Expect(0, 43487, '\P{ -In_Javanese}', "");
Expect(1, 43487, '\P{^ -In_Javanese}', "");
Expect(0, 43488, '\p{ -In_Javanese}', "");
Expect(1, 43488, '\p{^ -In_Javanese}', "");
Expect(1, 43488, '\P{ -In_Javanese}', "");
Expect(0, 43488, '\P{^ -In_Javanese}', "");
Error('\p{:=	In_Javanese}');
Error('\P{:=	In_Javanese}');
Expect(1, 43487, '\p{In_javanese}', "");
Expect(0, 43487, '\p{^In_javanese}', "");
Expect(0, 43487, '\P{In_javanese}', "");
Expect(1, 43487, '\P{^In_javanese}', "");
Expect(0, 43488, '\p{In_javanese}', "");
Expect(1, 43488, '\p{^In_javanese}', "");
Expect(1, 43488, '\P{In_javanese}', "");
Expect(0, 43488, '\P{^In_javanese}', "");
Error('\p{_:=PosixAlnum}');
Error('\P{_:=PosixAlnum}');
Expect(1, 122, '\p{posixalnum}', "");
Expect(0, 122, '\p{^posixalnum}', "");
Expect(0, 122, '\P{posixalnum}', "");
Expect(1, 122, '\P{^posixalnum}', "");
Expect(0, 123, '\p{posixalnum}', "");
Expect(1, 123, '\p{^posixalnum}', "");
Expect(1, 123, '\P{posixalnum}', "");
Expect(0, 123, '\P{^posixalnum}', "");
Expect(1, 122, '\p{		POSIXALNUM}', "");
Expect(0, 122, '\p{^		POSIXALNUM}', "");
Expect(0, 122, '\P{		POSIXALNUM}', "");
Expect(1, 122, '\P{^		POSIXALNUM}', "");
Expect(0, 123, '\p{		POSIXALNUM}', "");
Expect(1, 123, '\p{^		POSIXALNUM}', "");
Expect(1, 123, '\P{		POSIXALNUM}', "");
Expect(0, 123, '\P{^		POSIXALNUM}', "");
Error('\p{-/a/IS_posixalnum}');
Error('\P{-/a/IS_posixalnum}');
Expect(1, 122, '\p{isposixalnum}', "");
Expect(0, 122, '\p{^isposixalnum}', "");
Expect(0, 122, '\P{isposixalnum}', "");
Expect(1, 122, '\P{^isposixalnum}', "");
Expect(0, 123, '\p{isposixalnum}', "");
Expect(1, 123, '\p{^isposixalnum}', "");
Expect(1, 123, '\P{isposixalnum}', "");
Expect(0, 123, '\P{^isposixalnum}', "");
Expect(1, 122, '\p{-_is_posixalnum}', "");
Expect(0, 122, '\p{^-_is_posixalnum}', "");
Expect(0, 122, '\P{-_is_posixalnum}', "");
Expect(1, 122, '\P{^-_is_posixalnum}', "");
Expect(0, 123, '\p{-_is_posixalnum}', "");
Expect(1, 123, '\p{^-_is_posixalnum}', "");
Expect(1, 123, '\P{-_is_posixalnum}', "");
Expect(0, 123, '\P{^-_is_posixalnum}', "");
Error('\p{/a/Arabic_Supplement}');
Error('\P{/a/Arabic_Supplement}');
Expect(1, 1919, '\p{arabicsupplement}', "");
Expect(0, 1919, '\p{^arabicsupplement}', "");
Expect(0, 1919, '\P{arabicsupplement}', "");
Expect(1, 1919, '\P{^arabicsupplement}', "");
Expect(0, 1920, '\p{arabicsupplement}', "");
Expect(1, 1920, '\p{^arabicsupplement}', "");
Expect(1, 1920, '\P{arabicsupplement}', "");
Expect(0, 1920, '\P{^arabicsupplement}', "");
Expect(1, 1919, '\p{ arabic_Supplement}', "");
Expect(0, 1919, '\p{^ arabic_Supplement}', "");
Expect(0, 1919, '\P{ arabic_Supplement}', "");
Expect(1, 1919, '\P{^ arabic_Supplement}', "");
Expect(0, 1920, '\p{ arabic_Supplement}', "");
Expect(1, 1920, '\p{^ arabic_Supplement}', "");
Expect(1, 1920, '\P{ arabic_Supplement}', "");
Expect(0, 1920, '\P{^ arabic_Supplement}', "");
Error('\p{ :=is_Arabic_SUPPLEMENT}');
Error('\P{ :=is_Arabic_SUPPLEMENT}');
Expect(1, 1919, '\p{isarabicsupplement}', "");
Expect(0, 1919, '\p{^isarabicsupplement}', "");
Expect(0, 1919, '\P{isarabicsupplement}', "");
Expect(1, 1919, '\P{^isarabicsupplement}', "");
Expect(0, 1920, '\p{isarabicsupplement}', "");
Expect(1, 1920, '\p{^isarabicsupplement}', "");
Expect(1, 1920, '\P{isarabicsupplement}', "");
Expect(0, 1920, '\P{^isarabicsupplement}', "");
Expect(1, 1919, '\p{_	is_Arabic_Supplement}', "");
Expect(0, 1919, '\p{^_	is_Arabic_Supplement}', "");
Expect(0, 1919, '\P{_	is_Arabic_Supplement}', "");
Expect(1, 1919, '\P{^_	is_Arabic_Supplement}', "");
Expect(0, 1920, '\p{_	is_Arabic_Supplement}', "");
Expect(1, 1920, '\p{^_	is_Arabic_Supplement}', "");
Expect(1, 1920, '\P{_	is_Arabic_Supplement}', "");
Expect(0, 1920, '\P{^_	is_Arabic_Supplement}', "");
Error('\p{-/a/IN_Arabic_Supplement}');
Error('\P{-/a/IN_Arabic_Supplement}');
Expect(1, 1919, '\p{inarabicsupplement}', "");
Expect(0, 1919, '\p{^inarabicsupplement}', "");
Expect(0, 1919, '\P{inarabicsupplement}', "");
Expect(1, 1919, '\P{^inarabicsupplement}', "");
Expect(0, 1920, '\p{inarabicsupplement}', "");
Expect(1, 1920, '\p{^inarabicsupplement}', "");
Expect(1, 1920, '\P{inarabicsupplement}', "");
Expect(0, 1920, '\P{^inarabicsupplement}', "");
Expect(1, 1919, '\p{ -In_Arabic_Supplement}', "");
Expect(0, 1919, '\p{^ -In_Arabic_Supplement}', "");
Expect(0, 1919, '\P{ -In_Arabic_Supplement}', "");
Expect(1, 1919, '\P{^ -In_Arabic_Supplement}', "");
Expect(0, 1920, '\p{ -In_Arabic_Supplement}', "");
Expect(1, 1920, '\p{^ -In_Arabic_Supplement}', "");
Expect(1, 1920, '\P{ -In_Arabic_Supplement}', "");
Expect(0, 1920, '\P{^ -In_Arabic_Supplement}', "");
Error('\p{ -Arabic_sup:=}');
Error('\P{ -Arabic_sup:=}');
Expect(1, 1919, '\p{arabicsup}', "");
Expect(0, 1919, '\p{^arabicsup}', "");
Expect(0, 1919, '\P{arabicsup}', "");
Expect(1, 1919, '\P{^arabicsup}', "");
Expect(0, 1920, '\p{arabicsup}', "");
Expect(1, 1920, '\p{^arabicsup}', "");
Expect(1, 1920, '\P{arabicsup}', "");
Expect(0, 1920, '\P{^arabicsup}', "");
Expect(1, 1919, '\p{__arabic_Sup}', "");
Expect(0, 1919, '\p{^__arabic_Sup}', "");
Expect(0, 1919, '\P{__arabic_Sup}', "");
Expect(1, 1919, '\P{^__arabic_Sup}', "");
Expect(0, 1920, '\p{__arabic_Sup}', "");
Expect(1, 1920, '\p{^__arabic_Sup}', "");
Expect(1, 1920, '\P{__arabic_Sup}', "");
Expect(0, 1920, '\P{^__arabic_Sup}', "");
Error('\p{/a/	IS_Arabic_Sup}');
Error('\P{/a/	IS_Arabic_Sup}');
Expect(1, 1919, '\p{isarabicsup}', "");
Expect(0, 1919, '\p{^isarabicsup}', "");
Expect(0, 1919, '\P{isarabicsup}', "");
Expect(1, 1919, '\P{^isarabicsup}', "");
Expect(0, 1920, '\p{isarabicsup}', "");
Expect(1, 1920, '\p{^isarabicsup}', "");
Expect(1, 1920, '\P{isarabicsup}', "");
Expect(0, 1920, '\P{^isarabicsup}', "");
Expect(1, 1919, '\p{_	is_ARABIC_Sup}', "");
Expect(0, 1919, '\p{^_	is_ARABIC_Sup}', "");
Expect(0, 1919, '\P{_	is_ARABIC_Sup}', "");
Expect(1, 1919, '\P{^_	is_ARABIC_Sup}', "");
Expect(0, 1920, '\p{_	is_ARABIC_Sup}', "");
Expect(1, 1920, '\p{^_	is_ARABIC_Sup}', "");
Expect(1, 1920, '\P{_	is_ARABIC_Sup}', "");
Expect(0, 1920, '\P{^_	is_ARABIC_Sup}', "");
Error('\p{_/a/In_arabic_Sup}');
Error('\P{_/a/In_arabic_Sup}');
Expect(1, 1919, '\p{inarabicsup}', "");
Expect(0, 1919, '\p{^inarabicsup}', "");
Expect(0, 1919, '\P{inarabicsup}', "");
Expect(1, 1919, '\P{^inarabicsup}', "");
Expect(0, 1920, '\p{inarabicsup}', "");
Expect(1, 1920, '\p{^inarabicsup}', "");
Expect(1, 1920, '\P{inarabicsup}', "");
Expect(0, 1920, '\P{^inarabicsup}', "");
Expect(1, 1919, '\p{  In_ARABIC_Sup}', "");
Expect(0, 1919, '\p{^  In_ARABIC_Sup}', "");
Expect(0, 1919, '\P{  In_ARABIC_Sup}', "");
Expect(1, 1919, '\P{^  In_ARABIC_Sup}', "");
Expect(0, 1920, '\p{  In_ARABIC_Sup}', "");
Expect(1, 1920, '\p{^  In_ARABIC_Sup}', "");
Expect(1, 1920, '\P{  In_ARABIC_Sup}', "");
Expect(0, 1920, '\P{^  In_ARABIC_Sup}', "");
Error('\p{:=_	CJK_Compatibility_ideographs_supplement}');
Error('\P{:=_	CJK_Compatibility_ideographs_supplement}');
Expect(1, 195103, '\p{cjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\p{^cjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\P{cjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\P{^cjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\p{cjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\p{^cjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\P{cjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\P{^cjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\p{	cjk_COMPATIBILITY_Ideographs_Supplement}', "");
Expect(0, 195103, '\p{^	cjk_COMPATIBILITY_Ideographs_Supplement}', "");
Expect(0, 195103, '\P{	cjk_COMPATIBILITY_Ideographs_Supplement}', "");
Expect(1, 195103, '\P{^	cjk_COMPATIBILITY_Ideographs_Supplement}', "");
Expect(0, 195104, '\p{	cjk_COMPATIBILITY_Ideographs_Supplement}', "");
Expect(1, 195104, '\p{^	cjk_COMPATIBILITY_Ideographs_Supplement}', "");
Expect(1, 195104, '\P{	cjk_COMPATIBILITY_Ideographs_Supplement}', "");
Expect(0, 195104, '\P{^	cjk_COMPATIBILITY_Ideographs_Supplement}', "");
Error('\p{ /a/IS_CJK_compatibility_Ideographs_supplement}');
Error('\P{ /a/IS_CJK_compatibility_Ideographs_supplement}');
Expect(1, 195103, '\p{iscjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\p{^iscjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\P{iscjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\P{^iscjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\p{iscjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\p{^iscjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\P{iscjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\P{^iscjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\p{ 	Is_cjk_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT}', "");
Expect(0, 195103, '\p{^ 	Is_cjk_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT}', "");
Expect(0, 195103, '\P{ 	Is_cjk_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT}', "");
Expect(1, 195103, '\P{^ 	Is_cjk_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT}', "");
Expect(0, 195104, '\p{ 	Is_cjk_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT}', "");
Expect(1, 195104, '\p{^ 	Is_cjk_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT}', "");
Expect(1, 195104, '\P{ 	Is_cjk_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT}', "");
Expect(0, 195104, '\P{^ 	Is_cjk_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT}', "");
Error('\p{/a/ -In_cjk_Compatibility_ideographs_SUPPLEMENT}');
Error('\P{/a/ -In_cjk_Compatibility_ideographs_SUPPLEMENT}');
Expect(1, 195103, '\p{incjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\p{^incjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\P{incjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\P{^incjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\p{incjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\p{^incjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\P{incjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\P{^incjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\p{ In_CJK_compatibility_Ideographs_SUPPLEMENT}', "");
Expect(0, 195103, '\p{^ In_CJK_compatibility_Ideographs_SUPPLEMENT}', "");
Expect(0, 195103, '\P{ In_CJK_compatibility_Ideographs_SUPPLEMENT}', "");
Expect(1, 195103, '\P{^ In_CJK_compatibility_Ideographs_SUPPLEMENT}', "");
Expect(0, 195104, '\p{ In_CJK_compatibility_Ideographs_SUPPLEMENT}', "");
Expect(1, 195104, '\p{^ In_CJK_compatibility_Ideographs_SUPPLEMENT}', "");
Expect(1, 195104, '\P{ In_CJK_compatibility_Ideographs_SUPPLEMENT}', "");
Expect(0, 195104, '\P{^ In_CJK_compatibility_Ideographs_SUPPLEMENT}', "");
Error('\p{_CJK_Compat_Ideographs_Sup/a/}');
Error('\P{_CJK_Compat_Ideographs_Sup/a/}');
Expect(1, 195103, '\p{cjkcompatideographssup}', "");
Expect(0, 195103, '\p{^cjkcompatideographssup}', "");
Expect(0, 195103, '\P{cjkcompatideographssup}', "");
Expect(1, 195103, '\P{^cjkcompatideographssup}', "");
Expect(0, 195104, '\p{cjkcompatideographssup}', "");
Expect(1, 195104, '\p{^cjkcompatideographssup}', "");
Expect(1, 195104, '\P{cjkcompatideographssup}', "");
Expect(0, 195104, '\P{^cjkcompatideographssup}', "");
Expect(1, 195103, '\p{	_CJK_compat_ideographs_Sup}', "");
Expect(0, 195103, '\p{^	_CJK_compat_ideographs_Sup}', "");
Expect(0, 195103, '\P{	_CJK_compat_ideographs_Sup}', "");
Expect(1, 195103, '\P{^	_CJK_compat_ideographs_Sup}', "");
Expect(0, 195104, '\p{	_CJK_compat_ideographs_Sup}', "");
Expect(1, 195104, '\p{^	_CJK_compat_ideographs_Sup}', "");
Expect(1, 195104, '\P{	_CJK_compat_ideographs_Sup}', "");
Expect(0, 195104, '\P{^	_CJK_compat_ideographs_Sup}', "");
Error('\p{:=Is_CJK_Compat_IDEOGRAPHS_sup}');
Error('\P{:=Is_CJK_Compat_IDEOGRAPHS_sup}');
Expect(1, 195103, '\p{iscjkcompatideographssup}', "");
Expect(0, 195103, '\p{^iscjkcompatideographssup}', "");
Expect(0, 195103, '\P{iscjkcompatideographssup}', "");
Expect(1, 195103, '\P{^iscjkcompatideographssup}', "");
Expect(0, 195104, '\p{iscjkcompatideographssup}', "");
Expect(1, 195104, '\p{^iscjkcompatideographssup}', "");
Expect(1, 195104, '\P{iscjkcompatideographssup}', "");
Expect(0, 195104, '\P{^iscjkcompatideographssup}', "");
Expect(1, 195103, '\p{Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(0, 195103, '\p{^Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(0, 195103, '\P{Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(1, 195103, '\P{^Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(0, 195104, '\p{Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(1, 195104, '\p{^Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(1, 195104, '\P{Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Expect(0, 195104, '\P{^Is_CJK_Compat_IDEOGRAPHS_Sup}', "");
Error('\p{		IN_cjk_compat_ideographs_sup:=}');
Error('\P{		IN_cjk_compat_ideographs_sup:=}');
Expect(1, 195103, '\p{incjkcompatideographssup}', "");
Expect(0, 195103, '\p{^incjkcompatideographssup}', "");
Expect(0, 195103, '\P{incjkcompatideographssup}', "");
Expect(1, 195103, '\P{^incjkcompatideographssup}', "");
Expect(0, 195104, '\p{incjkcompatideographssup}', "");
Expect(1, 195104, '\p{^incjkcompatideographssup}', "");
Expect(1, 195104, '\P{incjkcompatideographssup}', "");
Expect(0, 195104, '\P{^incjkcompatideographssup}', "");
Expect(1, 195103, '\p{	-IN_CJK_compat_Ideographs_sup}', "");
Expect(0, 195103, '\p{^	-IN_CJK_compat_Ideographs_sup}', "");
Expect(0, 195103, '\P{	-IN_CJK_compat_Ideographs_sup}', "");
Expect(1, 195103, '\P{^	-IN_CJK_compat_Ideographs_sup}', "");
Expect(0, 195104, '\p{	-IN_CJK_compat_Ideographs_sup}', "");
Expect(1, 195104, '\p{^	-IN_CJK_compat_Ideographs_sup}', "");
Expect(1, 195104, '\P{	-IN_CJK_compat_Ideographs_sup}', "");
Expect(0, 195104, '\P{^	-IN_CJK_compat_Ideographs_sup}', "");
Error('\p{	-in_gurmukhi:=}');
Error('\P{	-in_gurmukhi:=}');
Expect(1, 2687, '\p{ingurmukhi}', "");
Expect(0, 2687, '\p{^ingurmukhi}', "");
Expect(0, 2687, '\P{ingurmukhi}', "");
Expect(1, 2687, '\P{^ingurmukhi}', "");
Expect(0, 2688, '\p{ingurmukhi}', "");
Expect(1, 2688, '\p{^ingurmukhi}', "");
Expect(1, 2688, '\P{ingurmukhi}', "");
Expect(0, 2688, '\P{^ingurmukhi}', "");
Expect(1, 2687, '\p{ 	in_Gurmukhi}', "");
Expect(0, 2687, '\p{^ 	in_Gurmukhi}', "");
Expect(0, 2687, '\P{ 	in_Gurmukhi}', "");
Expect(1, 2687, '\P{^ 	in_Gurmukhi}', "");
Expect(0, 2688, '\p{ 	in_Gurmukhi}', "");
Expect(1, 2688, '\p{^ 	in_Gurmukhi}', "");
Expect(1, 2688, '\P{ 	in_Gurmukhi}', "");
Expect(0, 2688, '\P{^ 	in_Gurmukhi}', "");
Error('\p{:=- in_Gurmukhi}');
Error('\P{:=- in_Gurmukhi}');
Expect(1, 2687, '\p{	_In_Gurmukhi}', "");
Expect(0, 2687, '\p{^	_In_Gurmukhi}', "");
Expect(0, 2687, '\P{	_In_Gurmukhi}', "");
Expect(1, 2687, '\P{^	_In_Gurmukhi}', "");
Expect(0, 2688, '\p{	_In_Gurmukhi}', "");
Expect(1, 2688, '\p{^	_In_Gurmukhi}', "");
Expect(1, 2688, '\P{	_In_Gurmukhi}', "");
Expect(0, 2688, '\P{^	_In_Gurmukhi}', "");
Error('\p{/a/	_COMBABOVE}');
Error('\P{/a/	_COMBABOVE}');
Expect(1, 119364, '\p{_CombAbove}', "");
Expect(0, 119364, '\p{^_CombAbove}', "");
Expect(0, 119364, '\P{_CombAbove}', "");
Expect(1, 119364, '\P{^_CombAbove}', "");
Expect(0, 119365, '\p{_CombAbove}', "");
Expect(1, 119365, '\p{^_CombAbove}', "");
Expect(1, 119365, '\P{_CombAbove}', "");
Expect(0, 119365, '\P{^_CombAbove}', "");
Error('\p{	__CombAbove:=}');
Error('\P{	__CombAbove:=}');
Expect(1, 119364, '\p{_COMBABOVE}', "");
Expect(0, 119364, '\p{^_COMBABOVE}', "");
Expect(0, 119364, '\P{_COMBABOVE}', "");
Expect(1, 119364, '\P{^_COMBABOVE}', "");
Expect(0, 119365, '\p{_COMBABOVE}', "");
Expect(1, 119365, '\p{^_COMBABOVE}', "");
Expect(1, 119365, '\P{_COMBABOVE}', "");
Expect(0, 119365, '\P{^_COMBABOVE}', "");
Error('\p{Other_ID_Continue}');
Error('\P{Other_ID_Continue}');
Error('\p{Is_Other_ID_Continue}');
Error('\P{Is_Other_ID_Continue}');
Error('\p{OIDC}');
Error('\P{OIDC}');
Error('\p{Is_OIDC}');
Error('\P{Is_OIDC}');
Error('\p{Bamum:=}');
Error('\P{Bamum:=}');
Expect(1, 92728, '\p{bamum}', "");
Expect(0, 92728, '\p{^bamum}', "");
Expect(0, 92728, '\P{bamum}', "");
Expect(1, 92728, '\P{^bamum}', "");
Expect(0, 92729, '\p{bamum}', "");
Expect(1, 92729, '\p{^bamum}', "");
Expect(1, 92729, '\P{bamum}', "");
Expect(0, 92729, '\P{^bamum}', "");
Expect(1, 92728, '\p{-Bamum}', "");
Expect(0, 92728, '\p{^-Bamum}', "");
Expect(0, 92728, '\P{-Bamum}', "");
Expect(1, 92728, '\P{^-Bamum}', "");
Expect(0, 92729, '\p{-Bamum}', "");
Expect(1, 92729, '\p{^-Bamum}', "");
Expect(1, 92729, '\P{-Bamum}', "");
Expect(0, 92729, '\P{^-Bamum}', "");
Error('\p{/a/Is_Bamum}');
Error('\P{/a/Is_Bamum}');
Expect(1, 92728, '\p{isbamum}', "");
Expect(0, 92728, '\p{^isbamum}', "");
Expect(0, 92728, '\P{isbamum}', "");
Expect(1, 92728, '\P{^isbamum}', "");
Expect(0, 92729, '\p{isbamum}', "");
Expect(1, 92729, '\p{^isbamum}', "");
Expect(1, 92729, '\P{isbamum}', "");
Expect(0, 92729, '\P{^isbamum}', "");
Expect(1, 92728, '\p{_	Is_Bamum}', "");
Expect(0, 92728, '\p{^_	Is_Bamum}', "");
Expect(0, 92728, '\P{_	Is_Bamum}', "");
Expect(1, 92728, '\P{^_	Is_Bamum}', "");
Expect(0, 92729, '\p{_	Is_Bamum}', "");
Expect(1, 92729, '\p{^_	Is_Bamum}', "");
Expect(1, 92729, '\P{_	Is_Bamum}', "");
Expect(0, 92729, '\P{^_	Is_Bamum}', "");
Error('\p{/a/__bamu}');
Error('\P{/a/__bamu}');
Expect(1, 92728, '\p{bamu}', "");
Expect(0, 92728, '\p{^bamu}', "");
Expect(0, 92728, '\P{bamu}', "");
Expect(1, 92728, '\P{^bamu}', "");
Expect(0, 92729, '\p{bamu}', "");
Expect(1, 92729, '\p{^bamu}', "");
Expect(1, 92729, '\P{bamu}', "");
Expect(0, 92729, '\P{^bamu}', "");
Expect(1, 92728, '\p{	Bamu}', "");
Expect(0, 92728, '\p{^	Bamu}', "");
Expect(0, 92728, '\P{	Bamu}', "");
Expect(1, 92728, '\P{^	Bamu}', "");
Expect(0, 92729, '\p{	Bamu}', "");
Expect(1, 92729, '\p{^	Bamu}', "");
Expect(1, 92729, '\P{	Bamu}', "");
Expect(0, 92729, '\P{^	Bamu}', "");
Error('\p{:= -Is_Bamu}');
Error('\P{:= -Is_Bamu}');
Expect(1, 92728, '\p{isbamu}', "");
Expect(0, 92728, '\p{^isbamu}', "");
Expect(0, 92728, '\P{isbamu}', "");
Expect(1, 92728, '\P{^isbamu}', "");
Expect(0, 92729, '\p{isbamu}', "");
Expect(1, 92729, '\p{^isbamu}', "");
Expect(1, 92729, '\P{isbamu}', "");
Expect(0, 92729, '\P{^isbamu}', "");
Expect(1, 92728, '\p{ 	IS_bamu}', "");
Expect(0, 92728, '\p{^ 	IS_bamu}', "");
Expect(0, 92728, '\P{ 	IS_bamu}', "");
Expect(1, 92728, '\P{^ 	IS_bamu}', "");
Expect(0, 92729, '\p{ 	IS_bamu}', "");
Expect(1, 92729, '\p{^ 	IS_bamu}', "");
Expect(1, 92729, '\P{ 	IS_bamu}', "");
Expect(0, 92729, '\P{^ 	IS_bamu}', "");
Error('\p{	In_bengali:=}');
Error('\P{	In_bengali:=}');
Expect(1, 2559, '\p{inbengali}', "");
Expect(0, 2559, '\p{^inbengali}', "");
Expect(0, 2559, '\P{inbengali}', "");
Expect(1, 2559, '\P{^inbengali}', "");
Expect(0, 2560, '\p{inbengali}', "");
Expect(1, 2560, '\p{^inbengali}', "");
Expect(1, 2560, '\P{inbengali}', "");
Expect(0, 2560, '\P{^inbengali}', "");
Expect(1, 2559, '\p{_ In_Bengali}', "");
Expect(0, 2559, '\p{^_ In_Bengali}', "");
Expect(0, 2559, '\P{_ In_Bengali}', "");
Expect(1, 2559, '\P{^_ In_Bengali}', "");
Expect(0, 2560, '\p{_ In_Bengali}', "");
Expect(1, 2560, '\p{^_ In_Bengali}', "");
Expect(1, 2560, '\P{_ In_Bengali}', "");
Expect(0, 2560, '\P{^_ In_Bengali}', "");
Error('\p{:=-	IN_BENGALI}');
Error('\P{:=-	IN_BENGALI}');
Expect(1, 2559, '\p{_	IN_Bengali}', "");
Expect(0, 2559, '\p{^_	IN_Bengali}', "");
Expect(0, 2559, '\P{_	IN_Bengali}', "");
Expect(1, 2559, '\P{^_	IN_Bengali}', "");
Expect(0, 2560, '\p{_	IN_Bengali}', "");
Expect(1, 2560, '\p{^_	IN_Bengali}', "");
Expect(1, 2560, '\P{_	IN_Bengali}', "");
Expect(0, 2560, '\P{^_	IN_Bengali}', "");
Error('\p{/a/_Cuneiform}');
Error('\P{/a/_Cuneiform}');
Expect(1, 74867, '\p{cuneiform}', "");
Expect(0, 74867, '\p{^cuneiform}', "");
Expect(0, 74867, '\P{cuneiform}', "");
Expect(1, 74867, '\P{^cuneiform}', "");
Expect(0, 74868, '\p{cuneiform}', "");
Expect(1, 74868, '\p{^cuneiform}', "");
Expect(1, 74868, '\P{cuneiform}', "");
Expect(0, 74868, '\P{^cuneiform}', "");
Expect(1, 74867, '\p{ -Cuneiform}', "");
Expect(0, 74867, '\p{^ -Cuneiform}', "");
Expect(0, 74867, '\P{ -Cuneiform}', "");
Expect(1, 74867, '\P{^ -Cuneiform}', "");
Expect(0, 74868, '\p{ -Cuneiform}', "");
Expect(1, 74868, '\p{^ -Cuneiform}', "");
Expect(1, 74868, '\P{ -Cuneiform}', "");
Expect(0, 74868, '\P{^ -Cuneiform}', "");
Error('\p{ _is_CUNEIFORM:=}');
Error('\P{ _is_CUNEIFORM:=}');
Expect(1, 74867, '\p{iscuneiform}', "");
Expect(0, 74867, '\p{^iscuneiform}', "");
Expect(0, 74867, '\P{iscuneiform}', "");
Expect(1, 74867, '\P{^iscuneiform}', "");
Expect(0, 74868, '\p{iscuneiform}', "");
Expect(1, 74868, '\p{^iscuneiform}', "");
Expect(1, 74868, '\P{iscuneiform}', "");
Expect(0, 74868, '\P{^iscuneiform}', "");
Expect(1, 74867, '\p{-_IS_cuneiform}', "");
Expect(0, 74867, '\p{^-_IS_cuneiform}', "");
Expect(0, 74867, '\P{-_IS_cuneiform}', "");
Expect(1, 74867, '\P{^-_IS_cuneiform}', "");
Expect(0, 74868, '\p{-_IS_cuneiform}', "");
Expect(1, 74868, '\p{^-_IS_cuneiform}', "");
Expect(1, 74868, '\P{-_IS_cuneiform}', "");
Expect(0, 74868, '\P{^-_IS_cuneiform}', "");
Error('\p{	 Xsux/a/}');
Error('\P{	 Xsux/a/}');
Expect(1, 74867, '\p{xsux}', "");
Expect(0, 74867, '\p{^xsux}', "");
Expect(0, 74867, '\P{xsux}', "");
Expect(1, 74867, '\P{^xsux}', "");
Expect(0, 74868, '\p{xsux}', "");
Expect(1, 74868, '\p{^xsux}', "");
Expect(1, 74868, '\P{xsux}', "");
Expect(0, 74868, '\P{^xsux}', "");
Expect(1, 74867, '\p{-Xsux}', "");
Expect(0, 74867, '\p{^-Xsux}', "");
Expect(0, 74867, '\P{-Xsux}', "");
Expect(1, 74867, '\P{^-Xsux}', "");
Expect(0, 74868, '\p{-Xsux}', "");
Expect(1, 74868, '\p{^-Xsux}', "");
Expect(1, 74868, '\P{-Xsux}', "");
Expect(0, 74868, '\P{^-Xsux}', "");
Error('\p{/a/_ Is_Xsux}');
Error('\P{/a/_ Is_Xsux}');
Expect(1, 74867, '\p{isxsux}', "");
Expect(0, 74867, '\p{^isxsux}', "");
Expect(0, 74867, '\P{isxsux}', "");
Expect(1, 74867, '\P{^isxsux}', "");
Expect(0, 74868, '\p{isxsux}', "");
Expect(1, 74868, '\p{^isxsux}', "");
Expect(1, 74868, '\P{isxsux}', "");
Expect(0, 74868, '\P{^isxsux}', "");
Expect(1, 74867, '\p{-is_xsux}', "");
Expect(0, 74867, '\p{^-is_xsux}', "");
Expect(0, 74867, '\P{-is_xsux}', "");
Expect(1, 74867, '\P{^-is_xsux}', "");
Expect(0, 74868, '\p{-is_xsux}', "");
Expect(1, 74868, '\p{^-is_xsux}', "");
Expect(1, 74868, '\P{-is_xsux}', "");
Expect(0, 74868, '\P{^-is_xsux}', "");
Error('\p{/a/	IN_Chakma}');
Error('\P{/a/	IN_Chakma}');
Expect(1, 69967, '\p{inchakma}', "");
Expect(0, 69967, '\p{^inchakma}', "");
Expect(0, 69967, '\P{inchakma}', "");
Expect(1, 69967, '\P{^inchakma}', "");
Expect(0, 69968, '\p{inchakma}', "");
Expect(1, 69968, '\p{^inchakma}', "");
Expect(1, 69968, '\P{inchakma}', "");
Expect(0, 69968, '\P{^inchakma}', "");
Expect(1, 69967, '\p{	in_chakma}', "");
Expect(0, 69967, '\p{^	in_chakma}', "");
Expect(0, 69967, '\P{	in_chakma}', "");
Expect(1, 69967, '\P{^	in_chakma}', "");
Expect(0, 69968, '\p{	in_chakma}', "");
Expect(1, 69968, '\p{^	in_chakma}', "");
Expect(1, 69968, '\P{	in_chakma}', "");
Expect(0, 69968, '\P{^	in_chakma}', "");
Error('\p{/a/In_Chakma}');
Error('\P{/a/In_Chakma}');
Expect(1, 69967, '\p{ In_chakma}', "");
Expect(0, 69967, '\p{^ In_chakma}', "");
Expect(0, 69967, '\P{ In_chakma}', "");
Expect(1, 69967, '\P{^ In_chakma}', "");
Expect(0, 69968, '\p{ In_chakma}', "");
Expect(1, 69968, '\p{^ In_chakma}', "");
Expect(1, 69968, '\P{ In_chakma}', "");
Expect(0, 69968, '\P{^ In_chakma}', "");
Error('\p{	/a/CASED}');
Error('\P{	/a/CASED}');
Expect(1, 120779, '\p{cased}', "");
Expect(0, 120779, '\p{^cased}', "");
Expect(0, 120779, '\P{cased}', "");
Expect(1, 120779, '\P{^cased}', "");
Expect(0, 120780, '\p{cased}', "");
Expect(1, 120780, '\p{^cased}', "");
Expect(1, 120780, '\P{cased}', "");
Expect(0, 120780, '\P{^cased}', "");
Expect(1, 120779, '\p{-CASED}', "");
Expect(0, 120779, '\p{^-CASED}', "");
Expect(0, 120779, '\P{-CASED}', "");
Expect(1, 120779, '\P{^-CASED}', "");
Expect(0, 120780, '\p{-CASED}', "");
Expect(1, 120780, '\p{^-CASED}', "");
Expect(1, 120780, '\P{-CASED}', "");
Expect(0, 120780, '\P{^-CASED}', "");
Error('\p{/a/_Is_CASED}');
Error('\P{/a/_Is_CASED}');
Expect(1, 120779, '\p{iscased}', "");
Expect(0, 120779, '\p{^iscased}', "");
Expect(0, 120779, '\P{iscased}', "");
Expect(1, 120779, '\P{^iscased}', "");
Expect(0, 120780, '\p{iscased}', "");
Expect(1, 120780, '\p{^iscased}', "");
Expect(1, 120780, '\P{iscased}', "");
Expect(0, 120780, '\P{^iscased}', "");
Expect(1, 120779, '\p{	-Is_Cased}', "");
Expect(0, 120779, '\p{^	-Is_Cased}', "");
Expect(0, 120779, '\P{	-Is_Cased}', "");
Expect(1, 120779, '\P{^	-Is_Cased}', "");
Expect(0, 120780, '\p{	-Is_Cased}', "");
Expect(1, 120780, '\p{^	-Is_Cased}', "");
Expect(1, 120780, '\P{	-Is_Cased}', "");
Expect(0, 120780, '\P{^	-Is_Cased}', "");
Error('\p{:=In_Katakana}');
Error('\P{:=In_Katakana}');
Expect(1, 12543, '\p{inkatakana}', "");
Expect(0, 12543, '\p{^inkatakana}', "");
Expect(0, 12543, '\P{inkatakana}', "");
Expect(1, 12543, '\P{^inkatakana}', "");
Expect(0, 12544, '\p{inkatakana}', "");
Expect(1, 12544, '\p{^inkatakana}', "");
Expect(1, 12544, '\P{inkatakana}', "");
Expect(0, 12544, '\P{^inkatakana}', "");
Expect(1, 12543, '\p{- In_Katakana}', "");
Expect(0, 12543, '\p{^- In_Katakana}', "");
Expect(0, 12543, '\P{- In_Katakana}', "");
Expect(1, 12543, '\P{^- In_Katakana}', "");
Expect(0, 12544, '\p{- In_Katakana}', "");
Expect(1, 12544, '\p{^- In_Katakana}', "");
Expect(1, 12544, '\P{- In_Katakana}', "");
Expect(0, 12544, '\P{^- In_Katakana}', "");
Error('\p{	:=In_Katakana}');
Error('\P{	:=In_Katakana}');
Expect(1, 12543, '\p{	 In_katakana}', "");
Expect(0, 12543, '\p{^	 In_katakana}', "");
Expect(0, 12543, '\P{	 In_katakana}', "");
Expect(1, 12543, '\P{^	 In_katakana}', "");
Expect(0, 12544, '\p{	 In_katakana}', "");
Expect(1, 12544, '\p{^	 In_katakana}', "");
Expect(1, 12544, '\P{	 In_katakana}', "");
Expect(0, 12544, '\P{^	 In_katakana}', "");
Error('\p{:=	General_PUNCTUATION}');
Error('\P{:=	General_PUNCTUATION}');
Expect(1, 8303, '\p{generalpunctuation}', "");
Expect(0, 8303, '\p{^generalpunctuation}', "");
Expect(0, 8303, '\P{generalpunctuation}', "");
Expect(1, 8303, '\P{^generalpunctuation}', "");
Expect(0, 8304, '\p{generalpunctuation}', "");
Expect(1, 8304, '\p{^generalpunctuation}', "");
Expect(1, 8304, '\P{generalpunctuation}', "");
Expect(0, 8304, '\P{^generalpunctuation}', "");
Expect(1, 8303, '\p{ 	general_punctuation}', "");
Expect(0, 8303, '\p{^ 	general_punctuation}', "");
Expect(0, 8303, '\P{ 	general_punctuation}', "");
Expect(1, 8303, '\P{^ 	general_punctuation}', "");
Expect(0, 8304, '\p{ 	general_punctuation}', "");
Expect(1, 8304, '\p{^ 	general_punctuation}', "");
Expect(1, 8304, '\P{ 	general_punctuation}', "");
Expect(0, 8304, '\P{^ 	general_punctuation}', "");
Error('\p{/a/ -Is_General_Punctuation}');
Error('\P{/a/ -Is_General_Punctuation}');
Expect(1, 8303, '\p{isgeneralpunctuation}', "");
Expect(0, 8303, '\p{^isgeneralpunctuation}', "");
Expect(0, 8303, '\P{isgeneralpunctuation}', "");
Expect(1, 8303, '\P{^isgeneralpunctuation}', "");
Expect(0, 8304, '\p{isgeneralpunctuation}', "");
Expect(1, 8304, '\p{^isgeneralpunctuation}', "");
Expect(1, 8304, '\P{isgeneralpunctuation}', "");
Expect(0, 8304, '\P{^isgeneralpunctuation}', "");
Expect(1, 8303, '\p{-Is_General_PUNCTUATION}', "");
Expect(0, 8303, '\p{^-Is_General_PUNCTUATION}', "");
Expect(0, 8303, '\P{-Is_General_PUNCTUATION}', "");
Expect(1, 8303, '\P{^-Is_General_PUNCTUATION}', "");
Expect(0, 8304, '\p{-Is_General_PUNCTUATION}', "");
Expect(1, 8304, '\p{^-Is_General_PUNCTUATION}', "");
Expect(1, 8304, '\P{-Is_General_PUNCTUATION}', "");
Expect(0, 8304, '\P{^-Is_General_PUNCTUATION}', "");
Error('\p{/a/_In_General_Punctuation}');
Error('\P{/a/_In_General_Punctuation}');
Expect(1, 8303, '\p{ingeneralpunctuation}', "");
Expect(0, 8303, '\p{^ingeneralpunctuation}', "");
Expect(0, 8303, '\P{ingeneralpunctuation}', "");
Expect(1, 8303, '\P{^ingeneralpunctuation}', "");
Expect(0, 8304, '\p{ingeneralpunctuation}', "");
Expect(1, 8304, '\p{^ingeneralpunctuation}', "");
Expect(1, 8304, '\P{ingeneralpunctuation}', "");
Expect(0, 8304, '\P{^ingeneralpunctuation}', "");
Expect(1, 8303, '\p{-IN_GENERAL_punctuation}', "");
Expect(0, 8303, '\p{^-IN_GENERAL_punctuation}', "");
Expect(0, 8303, '\P{-IN_GENERAL_punctuation}', "");
Expect(1, 8303, '\P{^-IN_GENERAL_punctuation}', "");
Expect(0, 8304, '\p{-IN_GENERAL_punctuation}', "");
Expect(1, 8304, '\p{^-IN_GENERAL_punctuation}', "");
Expect(1, 8304, '\P{-IN_GENERAL_punctuation}', "");
Expect(0, 8304, '\P{^-IN_GENERAL_punctuation}', "");
Error('\p{:=IN_Punctuation}');
Error('\P{:=IN_Punctuation}');
Expect(1, 8303, '\p{inpunctuation}', "");
Expect(0, 8303, '\p{^inpunctuation}', "");
Expect(0, 8303, '\P{inpunctuation}', "");
Expect(1, 8303, '\P{^inpunctuation}', "");
Expect(0, 8304, '\p{inpunctuation}', "");
Expect(1, 8304, '\p{^inpunctuation}', "");
Expect(1, 8304, '\P{inpunctuation}', "");
Expect(0, 8304, '\P{^inpunctuation}', "");
Expect(1, 8303, '\p{__In_punctuation}', "");
Expect(0, 8303, '\p{^__In_punctuation}', "");
Expect(0, 8303, '\P{__In_punctuation}', "");
Expect(1, 8303, '\P{^__In_punctuation}', "");
Expect(0, 8304, '\p{__In_punctuation}', "");
Expect(1, 8304, '\p{^__In_punctuation}', "");
Expect(1, 8304, '\P{__In_punctuation}', "");
Expect(0, 8304, '\P{^__In_punctuation}', "");
Error('\p{  In_TAI_Tham/a/}');
Error('\P{  In_TAI_Tham/a/}');
Expect(1, 6831, '\p{intaitham}', "");
Expect(0, 6831, '\p{^intaitham}', "");
Expect(0, 6831, '\P{intaitham}', "");
Expect(1, 6831, '\P{^intaitham}', "");
Expect(0, 6832, '\p{intaitham}', "");
Expect(1, 6832, '\p{^intaitham}', "");
Expect(1, 6832, '\P{intaitham}', "");
Expect(0, 6832, '\P{^intaitham}', "");
Expect(1, 6831, '\p{	 In_Tai_THAM}', "");
Expect(0, 6831, '\p{^	 In_Tai_THAM}', "");
Expect(0, 6831, '\P{	 In_Tai_THAM}', "");
Expect(1, 6831, '\P{^	 In_Tai_THAM}', "");
Expect(0, 6832, '\p{	 In_Tai_THAM}', "");
Expect(1, 6832, '\p{^	 In_Tai_THAM}', "");
Expect(1, 6832, '\P{	 In_Tai_THAM}', "");
Expect(0, 6832, '\P{^	 In_Tai_THAM}', "");
Error('\p{	 In_Tai_Tham:=}');
Error('\P{	 In_Tai_Tham:=}');
Expect(1, 6831, '\p{	 In_TAI_Tham}', "");
Expect(0, 6831, '\p{^	 In_TAI_Tham}', "");
Expect(0, 6831, '\P{	 In_TAI_Tham}', "");
Expect(1, 6831, '\P{^	 In_TAI_Tham}', "");
Expect(0, 6832, '\p{	 In_TAI_Tham}', "");
Expect(1, 6832, '\p{^	 In_TAI_Tham}', "");
Expect(1, 6832, '\P{	 In_TAI_Tham}', "");
Expect(0, 6832, '\P{^	 In_TAI_Tham}', "");
Error('\p{:=	-in_Sharada}');
Error('\P{:=	-in_Sharada}');
Expect(1, 70111, '\p{insharada}', "");
Expect(0, 70111, '\p{^insharada}', "");
Expect(0, 70111, '\P{insharada}', "");
Expect(1, 70111, '\P{^insharada}', "");
Expect(0, 70112, '\p{insharada}', "");
Expect(1, 70112, '\p{^insharada}', "");
Expect(1, 70112, '\P{insharada}', "");
Expect(0, 70112, '\P{^insharada}', "");
Expect(1, 70111, '\p{_in_sharada}', "");
Expect(0, 70111, '\p{^_in_sharada}', "");
Expect(0, 70111, '\P{_in_sharada}', "");
Expect(1, 70111, '\P{^_in_sharada}', "");
Expect(0, 70112, '\p{_in_sharada}', "");
Expect(1, 70112, '\p{^_in_sharada}', "");
Expect(1, 70112, '\P{_in_sharada}', "");
Expect(0, 70112, '\P{^_in_sharada}', "");
Error('\p{/a/- In_sharada}');
Error('\P{/a/- In_sharada}');
Expect(1, 70111, '\p{ In_SHARADA}', "");
Expect(0, 70111, '\p{^ In_SHARADA}', "");
Expect(0, 70111, '\P{ In_SHARADA}', "");
Expect(1, 70111, '\P{^ In_SHARADA}', "");
Expect(0, 70112, '\p{ In_SHARADA}', "");
Expect(1, 70112, '\p{^ In_SHARADA}', "");
Expect(1, 70112, '\P{ In_SHARADA}', "");
Expect(0, 70112, '\P{^ In_SHARADA}', "");
Error('\p{-IN_tagalog/a/}');
Error('\P{-IN_tagalog/a/}');
Expect(1, 5919, '\p{intagalog}', "");
Expect(0, 5919, '\p{^intagalog}', "");
Expect(0, 5919, '\P{intagalog}', "");
Expect(1, 5919, '\P{^intagalog}', "");
Expect(0, 5920, '\p{intagalog}', "");
Expect(1, 5920, '\p{^intagalog}', "");
Expect(1, 5920, '\P{intagalog}', "");
Expect(0, 5920, '\P{^intagalog}', "");
Expect(1, 5919, '\p{-_IN_Tagalog}', "");
Expect(0, 5919, '\p{^-_IN_Tagalog}', "");
Expect(0, 5919, '\P{-_IN_Tagalog}', "");
Expect(1, 5919, '\P{^-_IN_Tagalog}', "");
Expect(0, 5920, '\p{-_IN_Tagalog}', "");
Expect(1, 5920, '\p{^-_IN_Tagalog}', "");
Expect(1, 5920, '\P{-_IN_Tagalog}', "");
Expect(0, 5920, '\P{^-_IN_Tagalog}', "");
Error('\p{:=- in_Tagalog}');
Error('\P{:=- in_Tagalog}');
Expect(1, 5919, '\p{-in_Tagalog}', "");
Expect(0, 5919, '\p{^-in_Tagalog}', "");
Expect(0, 5919, '\P{-in_Tagalog}', "");
Expect(1, 5919, '\P{^-in_Tagalog}', "");
Expect(0, 5920, '\p{-in_Tagalog}', "");
Expect(1, 5920, '\p{^-in_Tagalog}', "");
Expect(1, 5920, '\P{-in_Tagalog}', "");
Expect(0, 5920, '\P{^-in_Tagalog}', "");
Error('\p{ Counting_Rod_numerals/a/}');
Error('\P{ Counting_Rod_numerals/a/}');
Expect(1, 119679, '\p{countingrodnumerals}', "");
Expect(0, 119679, '\p{^countingrodnumerals}', "");
Expect(0, 119679, '\P{countingrodnumerals}', "");
Expect(1, 119679, '\P{^countingrodnumerals}', "");
Expect(0, 119680, '\p{countingrodnumerals}', "");
Expect(1, 119680, '\p{^countingrodnumerals}', "");
Expect(1, 119680, '\P{countingrodnumerals}', "");
Expect(0, 119680, '\P{^countingrodnumerals}', "");
Expect(1, 119679, '\p{_ Counting_Rod_numerals}', "");
Expect(0, 119679, '\p{^_ Counting_Rod_numerals}', "");
Expect(0, 119679, '\P{_ Counting_Rod_numerals}', "");
Expect(1, 119679, '\P{^_ Counting_Rod_numerals}', "");
Expect(0, 119680, '\p{_ Counting_Rod_numerals}', "");
Expect(1, 119680, '\p{^_ Counting_Rod_numerals}', "");
Expect(1, 119680, '\P{_ Counting_Rod_numerals}', "");
Expect(0, 119680, '\P{^_ Counting_Rod_numerals}', "");
Error('\p{_IS_COUNTING_Rod_numerals/a/}');
Error('\P{_IS_COUNTING_Rod_numerals/a/}');
Expect(1, 119679, '\p{iscountingrodnumerals}', "");
Expect(0, 119679, '\p{^iscountingrodnumerals}', "");
Expect(0, 119679, '\P{iscountingrodnumerals}', "");
Expect(1, 119679, '\P{^iscountingrodnumerals}', "");
Expect(0, 119680, '\p{iscountingrodnumerals}', "");
Expect(1, 119680, '\p{^iscountingrodnumerals}', "");
Expect(1, 119680, '\P{iscountingrodnumerals}', "");
Expect(0, 119680, '\P{^iscountingrodnumerals}', "");
Expect(1, 119679, '\p{_-IS_COUNTING_Rod_NUMERALS}', "");
Expect(0, 119679, '\p{^_-IS_COUNTING_Rod_NUMERALS}', "");
Expect(0, 119679, '\P{_-IS_COUNTING_Rod_NUMERALS}', "");
Expect(1, 119679, '\P{^_-IS_COUNTING_Rod_NUMERALS}', "");
Expect(0, 119680, '\p{_-IS_COUNTING_Rod_NUMERALS}', "");
Expect(1, 119680, '\p{^_-IS_COUNTING_Rod_NUMERALS}', "");
Expect(1, 119680, '\P{_-IS_COUNTING_Rod_NUMERALS}', "");
Expect(0, 119680, '\P{^_-IS_COUNTING_Rod_NUMERALS}', "");
Error('\p{:=	 In_counting_ROD_Numerals}');
Error('\P{:=	 In_counting_ROD_Numerals}');
Expect(1, 119679, '\p{incountingrodnumerals}', "");
Expect(0, 119679, '\p{^incountingrodnumerals}', "");
Expect(0, 119679, '\P{incountingrodnumerals}', "");
Expect(1, 119679, '\P{^incountingrodnumerals}', "");
Expect(0, 119680, '\p{incountingrodnumerals}', "");
Expect(1, 119680, '\p{^incountingrodnumerals}', "");
Expect(1, 119680, '\P{incountingrodnumerals}', "");
Expect(0, 119680, '\P{^incountingrodnumerals}', "");
Expect(1, 119679, '\p{	-In_COUNTING_ROD_NUMERALS}', "");
Expect(0, 119679, '\p{^	-In_COUNTING_ROD_NUMERALS}', "");
Expect(0, 119679, '\P{	-In_COUNTING_ROD_NUMERALS}', "");
Expect(1, 119679, '\P{^	-In_COUNTING_ROD_NUMERALS}', "");
Expect(0, 119680, '\p{	-In_COUNTING_ROD_NUMERALS}', "");
Expect(1, 119680, '\p{^	-In_COUNTING_ROD_NUMERALS}', "");
Expect(1, 119680, '\P{	-In_COUNTING_ROD_NUMERALS}', "");
Expect(0, 119680, '\P{^	-In_COUNTING_ROD_NUMERALS}', "");
Error('\p{- Counting_Rod/a/}');
Error('\P{- Counting_Rod/a/}');
Expect(1, 119679, '\p{countingrod}', "");
Expect(0, 119679, '\p{^countingrod}', "");
Expect(0, 119679, '\P{countingrod}', "");
Expect(1, 119679, '\P{^countingrod}', "");
Expect(0, 119680, '\p{countingrod}', "");
Expect(1, 119680, '\p{^countingrod}', "");
Expect(1, 119680, '\P{countingrod}', "");
Expect(0, 119680, '\P{^countingrod}', "");
Expect(1, 119679, '\p{_Counting_rod}', "");
Expect(0, 119679, '\p{^_Counting_rod}', "");
Expect(0, 119679, '\P{_Counting_rod}', "");
Expect(1, 119679, '\P{^_Counting_rod}', "");
Expect(0, 119680, '\p{_Counting_rod}', "");
Expect(1, 119680, '\p{^_Counting_rod}', "");
Expect(1, 119680, '\P{_Counting_rod}', "");
Expect(0, 119680, '\P{^_Counting_rod}', "");
Error('\p{:= IS_Counting_Rod}');
Error('\P{:= IS_Counting_Rod}');
Expect(1, 119679, '\p{iscountingrod}', "");
Expect(0, 119679, '\p{^iscountingrod}', "");
Expect(0, 119679, '\P{iscountingrod}', "");
Expect(1, 119679, '\P{^iscountingrod}', "");
Expect(0, 119680, '\p{iscountingrod}', "");
Expect(1, 119680, '\p{^iscountingrod}', "");
Expect(1, 119680, '\P{iscountingrod}', "");
Expect(0, 119680, '\P{^iscountingrod}', "");
Expect(1, 119679, '\p{--IS_COUNTING_rod}', "");
Expect(0, 119679, '\p{^--IS_COUNTING_rod}', "");
Expect(0, 119679, '\P{--IS_COUNTING_rod}', "");
Expect(1, 119679, '\P{^--IS_COUNTING_rod}', "");
Expect(0, 119680, '\p{--IS_COUNTING_rod}', "");
Expect(1, 119680, '\p{^--IS_COUNTING_rod}', "");
Expect(1, 119680, '\P{--IS_COUNTING_rod}', "");
Expect(0, 119680, '\P{^--IS_COUNTING_rod}', "");
Error('\p{	in_counting_ROD:=}');
Error('\P{	in_counting_ROD:=}');
Expect(1, 119679, '\p{incountingrod}', "");
Expect(0, 119679, '\p{^incountingrod}', "");
Expect(0, 119679, '\P{incountingrod}', "");
Expect(1, 119679, '\P{^incountingrod}', "");
Expect(0, 119680, '\p{incountingrod}', "");
Expect(1, 119680, '\p{^incountingrod}', "");
Expect(1, 119680, '\P{incountingrod}', "");
Expect(0, 119680, '\P{^incountingrod}', "");
Expect(1, 119679, '\p{-_In_counting_rod}', "");
Expect(0, 119679, '\p{^-_In_counting_rod}', "");
Expect(0, 119679, '\P{-_In_counting_rod}', "");
Expect(1, 119679, '\P{^-_In_counting_rod}', "");
Expect(0, 119680, '\p{-_In_counting_rod}', "");
Expect(1, 119680, '\p{^-_In_counting_rod}', "");
Expect(1, 119680, '\P{-_In_counting_rod}', "");
Expect(0, 119680, '\P{^-_In_counting_rod}', "");
Error('\p{/a/-_in_LEPCHA}');
Error('\P{/a/-_in_LEPCHA}');
Expect(1, 7247, '\p{inlepcha}', "");
Expect(0, 7247, '\p{^inlepcha}', "");
Expect(0, 7247, '\P{inlepcha}', "");
Expect(1, 7247, '\P{^inlepcha}', "");
Expect(0, 7248, '\p{inlepcha}', "");
Expect(1, 7248, '\p{^inlepcha}', "");
Expect(1, 7248, '\P{inlepcha}', "");
Expect(0, 7248, '\P{^inlepcha}', "");
Expect(1, 7247, '\p{--in_LEPCHA}', "");
Expect(0, 7247, '\p{^--in_LEPCHA}', "");
Expect(0, 7247, '\P{--in_LEPCHA}', "");
Expect(1, 7247, '\P{^--in_LEPCHA}', "");
Expect(0, 7248, '\p{--in_LEPCHA}', "");
Expect(1, 7248, '\p{^--in_LEPCHA}', "");
Expect(1, 7248, '\P{--in_LEPCHA}', "");
Expect(0, 7248, '\P{^--in_LEPCHA}', "");
Error('\p{_in_LEPCHA/a/}');
Error('\P{_in_LEPCHA/a/}');
Expect(1, 7247, '\p{-	In_Lepcha}', "");
Expect(0, 7247, '\p{^-	In_Lepcha}', "");
Expect(0, 7247, '\P{-	In_Lepcha}', "");
Expect(1, 7247, '\P{^-	In_Lepcha}', "");
Expect(0, 7248, '\p{-	In_Lepcha}', "");
Expect(1, 7248, '\p{^-	In_Lepcha}', "");
Expect(1, 7248, '\P{-	In_Lepcha}', "");
Expect(0, 7248, '\P{^-	In_Lepcha}', "");
Error('\p{- in_SORA_Sompeng/a/}');
Error('\P{- in_SORA_Sompeng/a/}');
Expect(1, 69887, '\p{insorasompeng}', "");
Expect(0, 69887, '\p{^insorasompeng}', "");
Expect(0, 69887, '\P{insorasompeng}', "");
Expect(1, 69887, '\P{^insorasompeng}', "");
Expect(0, 69888, '\p{insorasompeng}', "");
Expect(1, 69888, '\p{^insorasompeng}', "");
Expect(1, 69888, '\P{insorasompeng}', "");
Expect(0, 69888, '\P{^insorasompeng}', "");
Expect(1, 69887, '\p{--In_Sora_Sompeng}', "");
Expect(0, 69887, '\p{^--In_Sora_Sompeng}', "");
Expect(0, 69887, '\P{--In_Sora_Sompeng}', "");
Expect(1, 69887, '\P{^--In_Sora_Sompeng}', "");
Expect(0, 69888, '\p{--In_Sora_Sompeng}', "");
Expect(1, 69888, '\p{^--In_Sora_Sompeng}', "");
Expect(1, 69888, '\P{--In_Sora_Sompeng}', "");
Expect(0, 69888, '\P{^--In_Sora_Sompeng}', "");
Error('\p{ :=In_Sora_Sompeng}');
Error('\P{ :=In_Sora_Sompeng}');
Expect(1, 69887, '\p{- IN_Sora_SOMPENG}', "");
Expect(0, 69887, '\p{^- IN_Sora_SOMPENG}', "");
Expect(0, 69887, '\P{- IN_Sora_SOMPENG}', "");
Expect(1, 69887, '\P{^- IN_Sora_SOMPENG}', "");
Expect(0, 69888, '\p{- IN_Sora_SOMPENG}', "");
Expect(1, 69888, '\p{^- IN_Sora_SOMPENG}', "");
Expect(1, 69888, '\P{- IN_Sora_SOMPENG}', "");
Expect(0, 69888, '\P{^- IN_Sora_SOMPENG}', "");
Error('\p{-:=In_CHAM}');
Error('\P{-:=In_CHAM}');
Expect(1, 43615, '\p{incham}', "");
Expect(0, 43615, '\p{^incham}', "");
Expect(0, 43615, '\P{incham}', "");
Expect(1, 43615, '\P{^incham}', "");
Expect(0, 43616, '\p{incham}', "");
Expect(1, 43616, '\p{^incham}', "");
Expect(1, 43616, '\P{incham}', "");
Expect(0, 43616, '\P{^incham}', "");
Expect(1, 43615, '\p{ in_Cham}', "");
Expect(0, 43615, '\p{^ in_Cham}', "");
Expect(0, 43615, '\P{ in_Cham}', "");
Expect(1, 43615, '\P{^ in_Cham}', "");
Expect(0, 43616, '\p{ in_Cham}', "");
Expect(1, 43616, '\p{^ in_Cham}', "");
Expect(1, 43616, '\P{ in_Cham}', "");
Expect(0, 43616, '\P{^ in_Cham}', "");
Error('\p{:=	-IN_Cham}');
Error('\P{:=	-IN_Cham}');
Expect(1, 43615, '\p{	-In_cham}', "");
Expect(0, 43615, '\p{^	-In_cham}', "");
Expect(0, 43615, '\P{	-In_cham}', "");
Expect(1, 43615, '\P{^	-In_cham}', "");
Expect(0, 43616, '\p{	-In_cham}', "");
Expect(1, 43616, '\p{^	-In_cham}', "");
Expect(1, 43616, '\P{	-In_cham}', "");
Expect(0, 43616, '\P{^	-In_cham}', "");
Error('\p{_:=In_Bopomofo}');
Error('\P{_:=In_Bopomofo}');
Expect(1, 12591, '\p{inbopomofo}', "");
Expect(0, 12591, '\p{^inbopomofo}', "");
Expect(0, 12591, '\P{inbopomofo}', "");
Expect(1, 12591, '\P{^inbopomofo}', "");
Expect(0, 12592, '\p{inbopomofo}', "");
Expect(1, 12592, '\p{^inbopomofo}', "");
Expect(1, 12592, '\P{inbopomofo}', "");
Expect(0, 12592, '\P{^inbopomofo}', "");
Expect(1, 12591, '\p{ in_bopomofo}', "");
Expect(0, 12591, '\p{^ in_bopomofo}', "");
Expect(0, 12591, '\P{ in_bopomofo}', "");
Expect(1, 12591, '\P{^ in_bopomofo}', "");
Expect(0, 12592, '\p{ in_bopomofo}', "");
Expect(1, 12592, '\p{^ in_bopomofo}', "");
Expect(1, 12592, '\P{ in_bopomofo}', "");
Expect(0, 12592, '\P{^ in_bopomofo}', "");
Error('\p{/a/in_BOPOMOFO}');
Error('\P{/a/in_BOPOMOFO}');
Expect(1, 12591, '\p{ IN_bopomofo}', "");
Expect(0, 12591, '\p{^ IN_bopomofo}', "");
Expect(0, 12591, '\P{ IN_bopomofo}', "");
Expect(1, 12591, '\P{^ IN_bopomofo}', "");
Expect(0, 12592, '\p{ IN_bopomofo}', "");
Expect(1, 12592, '\p{^ IN_bopomofo}', "");
Expect(1, 12592, '\P{ IN_bopomofo}', "");
Expect(0, 12592, '\P{^ IN_bopomofo}', "");
Error('\p{:= -In_miao}');
Error('\P{:= -In_miao}');
Expect(1, 94111, '\p{inmiao}', "");
Expect(0, 94111, '\p{^inmiao}', "");
Expect(0, 94111, '\P{inmiao}', "");
Expect(1, 94111, '\P{^inmiao}', "");
Expect(0, 94112, '\p{inmiao}', "");
Expect(1, 94112, '\p{^inmiao}', "");
Expect(1, 94112, '\P{inmiao}', "");
Expect(0, 94112, '\P{^inmiao}', "");
Expect(1, 94111, '\p{	In_Miao}', "");
Expect(0, 94111, '\p{^	In_Miao}', "");
Expect(0, 94111, '\P{	In_Miao}', "");
Expect(1, 94111, '\P{^	In_Miao}', "");
Expect(0, 94112, '\p{	In_Miao}', "");
Expect(1, 94112, '\p{^	In_Miao}', "");
Expect(1, 94112, '\P{	In_Miao}', "");
Expect(0, 94112, '\P{^	In_Miao}', "");
Error('\p{/a/--In_Miao}');
Error('\P{/a/--In_Miao}');
Expect(1, 94111, '\p{-_In_MIAO}', "");
Expect(0, 94111, '\p{^-_In_MIAO}', "");
Expect(0, 94111, '\P{-_In_MIAO}', "");
Expect(1, 94111, '\P{^-_In_MIAO}', "");
Expect(0, 94112, '\p{-_In_MIAO}', "");
Expect(1, 94112, '\p{^-_In_MIAO}', "");
Expect(1, 94112, '\P{-_In_MIAO}', "");
Expect(0, 94112, '\P{^-_In_MIAO}', "");
Error('\p{/a/_arrows}');
Error('\P{/a/_arrows}');
Expect(1, 8703, '\p{arrows}', "");
Expect(0, 8703, '\p{^arrows}', "");
Expect(0, 8703, '\P{arrows}', "");
Expect(1, 8703, '\P{^arrows}', "");
Expect(0, 8704, '\p{arrows}', "");
Expect(1, 8704, '\p{^arrows}', "");
Expect(1, 8704, '\P{arrows}', "");
Expect(0, 8704, '\P{^arrows}', "");
Expect(1, 8703, '\p{--arrows}', "");
Expect(0, 8703, '\p{^--arrows}', "");
Expect(0, 8703, '\P{--arrows}', "");
Expect(1, 8703, '\P{^--arrows}', "");
Expect(0, 8704, '\p{--arrows}', "");
Expect(1, 8704, '\p{^--arrows}', "");
Expect(1, 8704, '\P{--arrows}', "");
Expect(0, 8704, '\P{^--arrows}', "");
Error('\p{__is_ARROWS:=}');
Error('\P{__is_ARROWS:=}');
Expect(1, 8703, '\p{isarrows}', "");
Expect(0, 8703, '\p{^isarrows}', "");
Expect(0, 8703, '\P{isarrows}', "");
Expect(1, 8703, '\P{^isarrows}', "");
Expect(0, 8704, '\p{isarrows}', "");
Expect(1, 8704, '\p{^isarrows}', "");
Expect(1, 8704, '\P{isarrows}', "");
Expect(0, 8704, '\P{^isarrows}', "");
Expect(1, 8703, '\p{ IS_ARROWS}', "");
Expect(0, 8703, '\p{^ IS_ARROWS}', "");
Expect(0, 8703, '\P{ IS_ARROWS}', "");
Expect(1, 8703, '\P{^ IS_ARROWS}', "");
Expect(0, 8704, '\p{ IS_ARROWS}', "");
Expect(1, 8704, '\p{^ IS_ARROWS}', "");
Expect(1, 8704, '\P{ IS_ARROWS}', "");
Expect(0, 8704, '\P{^ IS_ARROWS}', "");
Error('\p{	/a/In_ARROWS}');
Error('\P{	/a/In_ARROWS}');
Expect(1, 8703, '\p{inarrows}', "");
Expect(0, 8703, '\p{^inarrows}', "");
Expect(0, 8703, '\P{inarrows}', "");
Expect(1, 8703, '\P{^inarrows}', "");
Expect(0, 8704, '\p{inarrows}', "");
Expect(1, 8704, '\p{^inarrows}', "");
Expect(1, 8704, '\P{inarrows}', "");
Expect(0, 8704, '\P{^inarrows}', "");
Expect(1, 8703, '\p{ 	In_ARROWS}', "");
Expect(0, 8703, '\p{^ 	In_ARROWS}', "");
Expect(0, 8703, '\P{ 	In_ARROWS}', "");
Expect(1, 8703, '\P{^ 	In_ARROWS}', "");
Expect(0, 8704, '\p{ 	In_ARROWS}', "");
Expect(1, 8704, '\p{^ 	In_ARROWS}', "");
Expect(1, 8704, '\P{ 	In_ARROWS}', "");
Expect(0, 8704, '\P{^ 	In_ARROWS}', "");
Error('\p{:=	-In_Lao}');
Error('\P{:=	-In_Lao}');
Expect(1, 3839, '\p{inlao}', "");
Expect(0, 3839, '\p{^inlao}', "");
Expect(0, 3839, '\P{inlao}', "");
Expect(1, 3839, '\P{^inlao}', "");
Expect(0, 3840, '\p{inlao}', "");
Expect(1, 3840, '\p{^inlao}', "");
Expect(1, 3840, '\P{inlao}', "");
Expect(0, 3840, '\P{^inlao}', "");
Expect(1, 3839, '\p{- In_Lao}', "");
Expect(0, 3839, '\p{^- In_Lao}', "");
Expect(0, 3839, '\P{- In_Lao}', "");
Expect(1, 3839, '\P{^- In_Lao}', "");
Expect(0, 3840, '\p{- In_Lao}', "");
Expect(1, 3840, '\p{^- In_Lao}', "");
Expect(1, 3840, '\P{- In_Lao}', "");
Expect(0, 3840, '\P{^- In_Lao}', "");
Error('\p{- In_Lao/a/}');
Error('\P{- In_Lao/a/}');
Expect(1, 3839, '\p{ _IN_Lao}', "");
Expect(0, 3839, '\p{^ _IN_Lao}', "");
Expect(0, 3839, '\P{ _IN_Lao}', "");
Expect(1, 3839, '\P{^ _IN_Lao}', "");
Expect(0, 3840, '\p{ _IN_Lao}', "");
Expect(1, 3840, '\p{^ _IN_Lao}', "");
Expect(1, 3840, '\P{ _IN_Lao}', "");
Expect(0, 3840, '\P{^ _IN_Lao}', "");
Error('\p{:=  CONNECTOR_Punctuation}');
Error('\P{:=  CONNECTOR_Punctuation}');
Expect(1, 65343, '\p{connectorpunctuation}', "");
Expect(0, 65343, '\p{^connectorpunctuation}', "");
Expect(0, 65343, '\P{connectorpunctuation}', "");
Expect(1, 65343, '\P{^connectorpunctuation}', "");
Expect(0, 65344, '\p{connectorpunctuation}', "");
Expect(1, 65344, '\p{^connectorpunctuation}', "");
Expect(1, 65344, '\P{connectorpunctuation}', "");
Expect(0, 65344, '\P{^connectorpunctuation}', "");
Expect(1, 65343, '\p{  Connector_punctuation}', "");
Expect(0, 65343, '\p{^  Connector_punctuation}', "");
Expect(0, 65343, '\P{  Connector_punctuation}', "");
Expect(1, 65343, '\P{^  Connector_punctuation}', "");
Expect(0, 65344, '\p{  Connector_punctuation}', "");
Expect(1, 65344, '\p{^  Connector_punctuation}', "");
Expect(1, 65344, '\P{  Connector_punctuation}', "");
Expect(0, 65344, '\P{^  Connector_punctuation}', "");
Error('\p{	IS_CONNECTOR_Punctuation:=}');
Error('\P{	IS_CONNECTOR_Punctuation:=}');
Expect(1, 65343, '\p{isconnectorpunctuation}', "");
Expect(0, 65343, '\p{^isconnectorpunctuation}', "");
Expect(0, 65343, '\P{isconnectorpunctuation}', "");
Expect(1, 65343, '\P{^isconnectorpunctuation}', "");
Expect(0, 65344, '\p{isconnectorpunctuation}', "");
Expect(1, 65344, '\p{^isconnectorpunctuation}', "");
Expect(1, 65344, '\P{isconnectorpunctuation}', "");
Expect(0, 65344, '\P{^isconnectorpunctuation}', "");
Expect(1, 65343, '\p{- IS_CONNECTOR_PUNCTUATION}', "");
Expect(0, 65343, '\p{^- IS_CONNECTOR_PUNCTUATION}', "");
Expect(0, 65343, '\P{- IS_CONNECTOR_PUNCTUATION}', "");
Expect(1, 65343, '\P{^- IS_CONNECTOR_PUNCTUATION}', "");
Expect(0, 65344, '\p{- IS_CONNECTOR_PUNCTUATION}', "");
Expect(1, 65344, '\p{^- IS_CONNECTOR_PUNCTUATION}', "");
Expect(1, 65344, '\P{- IS_CONNECTOR_PUNCTUATION}', "");
Expect(0, 65344, '\P{^- IS_CONNECTOR_PUNCTUATION}', "");
Error('\p{_Pc/a/}');
Error('\P{_Pc/a/}');
Expect(1, 65343, '\p{pc}', "");
Expect(0, 65343, '\p{^pc}', "");
Expect(0, 65343, '\P{pc}', "");
Expect(1, 65343, '\P{^pc}', "");
Expect(0, 65344, '\p{pc}', "");
Expect(1, 65344, '\p{^pc}', "");
Expect(1, 65344, '\P{pc}', "");
Expect(0, 65344, '\P{^pc}', "");
Expect(1, 65343, '\p{	 Pc}', "");
Expect(0, 65343, '\p{^	 Pc}', "");
Expect(0, 65343, '\P{	 Pc}', "");
Expect(1, 65343, '\P{^	 Pc}', "");
Expect(0, 65344, '\p{	 Pc}', "");
Expect(1, 65344, '\p{^	 Pc}', "");
Expect(1, 65344, '\P{	 Pc}', "");
Expect(0, 65344, '\P{^	 Pc}', "");
Error('\p{ :=IS_Pc}');
Error('\P{ :=IS_Pc}');
Expect(1, 65343, '\p{ispc}', "");
Expect(0, 65343, '\p{^ispc}', "");
Expect(0, 65343, '\P{ispc}', "");
Expect(1, 65343, '\P{^ispc}', "");
Expect(0, 65344, '\p{ispc}', "");
Expect(1, 65344, '\p{^ispc}', "");
Expect(1, 65344, '\P{ispc}', "");
Expect(0, 65344, '\P{^ispc}', "");
Expect(1, 65343, '\p{ 	is_Pc}', "");
Expect(0, 65343, '\p{^ 	is_Pc}', "");
Expect(0, 65343, '\P{ 	is_Pc}', "");
Expect(1, 65343, '\P{^ 	is_Pc}', "");
Expect(0, 65344, '\p{ 	is_Pc}', "");
Expect(1, 65344, '\p{^ 	is_Pc}', "");
Expect(1, 65344, '\P{ 	is_Pc}', "");
Expect(0, 65344, '\P{^ 	is_Pc}', "");
Error('\p{-	IN_Inscriptional_PAHLAVI:=}');
Error('\P{-	IN_Inscriptional_PAHLAVI:=}');
Expect(1, 68479, '\p{ininscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^ininscriptionalpahlavi}', "");
Expect(0, 68479, '\P{ininscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^ininscriptionalpahlavi}', "");
Expect(0, 68480, '\p{ininscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^ininscriptionalpahlavi}', "");
Expect(1, 68480, '\P{ininscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^ininscriptionalpahlavi}', "");
Expect(1, 68479, '\p{_ In_INSCRIPTIONAL_PAHLAVI}', "");
Expect(0, 68479, '\p{^_ In_INSCRIPTIONAL_PAHLAVI}', "");
Expect(0, 68479, '\P{_ In_INSCRIPTIONAL_PAHLAVI}', "");
Expect(1, 68479, '\P{^_ In_INSCRIPTIONAL_PAHLAVI}', "");
Expect(0, 68480, '\p{_ In_INSCRIPTIONAL_PAHLAVI}', "");
Expect(1, 68480, '\p{^_ In_INSCRIPTIONAL_PAHLAVI}', "");
Expect(1, 68480, '\P{_ In_INSCRIPTIONAL_PAHLAVI}', "");
Expect(0, 68480, '\P{^_ In_INSCRIPTIONAL_PAHLAVI}', "");
Error('\p{/a/	In_Inscriptional_pahlavi}');
Error('\P{/a/	In_Inscriptional_pahlavi}');
Expect(1, 68479, '\p{_-In_inscriptional_Pahlavi}', "");
Expect(0, 68479, '\p{^_-In_inscriptional_Pahlavi}', "");
Expect(0, 68479, '\P{_-In_inscriptional_Pahlavi}', "");
Expect(1, 68479, '\P{^_-In_inscriptional_Pahlavi}', "");
Expect(0, 68480, '\p{_-In_inscriptional_Pahlavi}', "");
Expect(1, 68480, '\p{^_-In_inscriptional_Pahlavi}', "");
Expect(1, 68480, '\P{_-In_inscriptional_Pahlavi}', "");
Expect(0, 68480, '\P{^_-In_inscriptional_Pahlavi}', "");
Error('\p{	/a/_X_LV_LVT_V}');
Error('\P{	/a/_X_LV_LVT_V}');
Expect(1, 55238, '\p{_X_LV_LVT_V}', "");
Expect(0, 55238, '\p{^_X_LV_LVT_V}', "");
Expect(0, 55238, '\P{_X_LV_LVT_V}', "");
Expect(1, 55238, '\P{^_X_LV_LVT_V}', "");
Expect(0, 55239, '\p{_X_LV_LVT_V}', "");
Expect(1, 55239, '\p{^_X_LV_LVT_V}', "");
Expect(1, 55239, '\P{_X_LV_LVT_V}', "");
Expect(0, 55239, '\P{^_X_LV_LVT_V}', "");
Error('\p{ /a/_X_LV_LVT_V}');
Error('\P{ /a/_X_LV_LVT_V}');
Error('\p{	 Private_Use/a/}');
Error('\P{	 Private_Use/a/}');
Expect(1, 1114109, '\p{privateuse}', "");
Expect(0, 1114109, '\p{^privateuse}', "");
Expect(0, 1114109, '\P{privateuse}', "");
Expect(1, 1114109, '\P{^privateuse}', "");
Expect(0, 63744, '\p{privateuse}', "");
Expect(1, 63744, '\p{^privateuse}', "");
Expect(1, 63744, '\P{privateuse}', "");
Expect(0, 63744, '\P{^privateuse}', "");
Expect(1, 1114109, '\p{	PRIVATE_USE}', "");
Expect(0, 1114109, '\p{^	PRIVATE_USE}', "");
Expect(0, 1114109, '\P{	PRIVATE_USE}', "");
Expect(1, 1114109, '\P{^	PRIVATE_USE}', "");
Expect(0, 63744, '\p{	PRIVATE_USE}', "");
Expect(1, 63744, '\p{^	PRIVATE_USE}', "");
Expect(1, 63744, '\P{	PRIVATE_USE}', "");
Expect(0, 63744, '\P{^	PRIVATE_USE}', "");
Error('\p{ 	is_PRIVATE_use:=}');
Error('\P{ 	is_PRIVATE_use:=}');
Expect(1, 1114109, '\p{isprivateuse}', "");
Expect(0, 1114109, '\p{^isprivateuse}', "");
Expect(0, 1114109, '\P{isprivateuse}', "");
Expect(1, 1114109, '\P{^isprivateuse}', "");
Expect(0, 63744, '\p{isprivateuse}', "");
Expect(1, 63744, '\p{^isprivateuse}', "");
Expect(1, 63744, '\P{isprivateuse}', "");
Expect(0, 63744, '\P{^isprivateuse}', "");
Expect(1, 1114109, '\p{_	is_Private_use}', "");
Expect(0, 1114109, '\p{^_	is_Private_use}', "");
Expect(0, 1114109, '\P{_	is_Private_use}', "");
Expect(1, 1114109, '\P{^_	is_Private_use}', "");
Expect(0, 63744, '\p{_	is_Private_use}', "");
Expect(1, 63744, '\p{^_	is_Private_use}', "");
Expect(1, 63744, '\P{_	is_Private_use}', "");
Expect(0, 63744, '\P{^_	is_Private_use}', "");
Error('\p{- Co/a/}');
Error('\P{- Co/a/}');
Expect(1, 1114109, '\p{co}', "");
Expect(0, 1114109, '\p{^co}', "");
Expect(0, 1114109, '\P{co}', "");
Expect(1, 1114109, '\P{^co}', "");
Expect(0, 63744, '\p{co}', "");
Expect(1, 63744, '\p{^co}', "");
Expect(1, 63744, '\P{co}', "");
Expect(0, 63744, '\P{^co}', "");
Expect(1, 1114109, '\p{		Co}', "");
Expect(0, 1114109, '\p{^		Co}', "");
Expect(0, 1114109, '\P{		Co}', "");
Expect(1, 1114109, '\P{^		Co}', "");
Expect(0, 63744, '\p{		Co}', "");
Expect(1, 63744, '\p{^		Co}', "");
Expect(1, 63744, '\P{		Co}', "");
Expect(0, 63744, '\P{^		Co}', "");
Error('\p{	:=IS_Co}');
Error('\P{	:=IS_Co}');
Expect(1, 1114109, '\p{isco}', "");
Expect(0, 1114109, '\p{^isco}', "");
Expect(0, 1114109, '\P{isco}', "");
Expect(1, 1114109, '\P{^isco}', "");
Expect(0, 63744, '\p{isco}', "");
Expect(1, 63744, '\p{^isco}', "");
Expect(1, 63744, '\P{isco}', "");
Expect(0, 63744, '\P{^isco}', "");
Expect(1, 1114109, '\p{-is_Co}', "");
Expect(0, 1114109, '\p{^-is_Co}', "");
Expect(0, 1114109, '\P{-is_Co}', "");
Expect(1, 1114109, '\P{^-is_Co}', "");
Expect(0, 63744, '\p{-is_Co}', "");
Expect(1, 63744, '\p{^-is_Co}', "");
Expect(1, 63744, '\P{-is_Co}', "");
Expect(0, 63744, '\P{^-is_Co}', "");
Error('\p{/a/_	In_LISU}');
Error('\P{/a/_	In_LISU}');
Expect(1, 42239, '\p{inlisu}', "");
Expect(0, 42239, '\p{^inlisu}', "");
Expect(0, 42239, '\P{inlisu}', "");
Expect(1, 42239, '\P{^inlisu}', "");
Expect(0, 42240, '\p{inlisu}', "");
Expect(1, 42240, '\p{^inlisu}', "");
Expect(1, 42240, '\P{inlisu}', "");
Expect(0, 42240, '\P{^inlisu}', "");
Expect(1, 42239, '\p{	IN_lisu}', "");
Expect(0, 42239, '\p{^	IN_lisu}', "");
Expect(0, 42239, '\P{	IN_lisu}', "");
Expect(1, 42239, '\P{^	IN_lisu}', "");
Expect(0, 42240, '\p{	IN_lisu}', "");
Expect(1, 42240, '\p{^	IN_lisu}', "");
Expect(1, 42240, '\P{	IN_lisu}', "");
Expect(0, 42240, '\P{^	IN_lisu}', "");
Error('\p{/a/_IN_LISU}');
Error('\P{/a/_IN_LISU}');
Expect(1, 42239, '\p{-	In_Lisu}', "");
Expect(0, 42239, '\p{^-	In_Lisu}', "");
Expect(0, 42239, '\P{-	In_Lisu}', "");
Expect(1, 42239, '\P{^-	In_Lisu}', "");
Expect(0, 42240, '\p{-	In_Lisu}', "");
Expect(1, 42240, '\p{^-	In_Lisu}', "");
Expect(1, 42240, '\P{-	In_Lisu}', "");
Expect(0, 42240, '\P{^-	In_Lisu}', "");
Error('\p{ :=Dash_Punctuation}');
Error('\P{ :=Dash_Punctuation}');
Expect(1, 65293, '\p{dashpunctuation}', "");
Expect(0, 65293, '\p{^dashpunctuation}', "");
Expect(0, 65293, '\P{dashpunctuation}', "");
Expect(1, 65293, '\P{^dashpunctuation}', "");
Expect(0, 65294, '\p{dashpunctuation}', "");
Expect(1, 65294, '\p{^dashpunctuation}', "");
Expect(1, 65294, '\P{dashpunctuation}', "");
Expect(0, 65294, '\P{^dashpunctuation}', "");
Expect(1, 65293, '\p{--DASH_punctuation}', "");
Expect(0, 65293, '\p{^--DASH_punctuation}', "");
Expect(0, 65293, '\P{--DASH_punctuation}', "");
Expect(1, 65293, '\P{^--DASH_punctuation}', "");
Expect(0, 65294, '\p{--DASH_punctuation}', "");
Expect(1, 65294, '\p{^--DASH_punctuation}', "");
Expect(1, 65294, '\P{--DASH_punctuation}', "");
Expect(0, 65294, '\P{^--DASH_punctuation}', "");
Error('\p{ :=is_Dash_Punctuation}');
Error('\P{ :=is_Dash_Punctuation}');
Expect(1, 65293, '\p{isdashpunctuation}', "");
Expect(0, 65293, '\p{^isdashpunctuation}', "");
Expect(0, 65293, '\P{isdashpunctuation}', "");
Expect(1, 65293, '\P{^isdashpunctuation}', "");
Expect(0, 65294, '\p{isdashpunctuation}', "");
Expect(1, 65294, '\p{^isdashpunctuation}', "");
Expect(1, 65294, '\P{isdashpunctuation}', "");
Expect(0, 65294, '\P{^isdashpunctuation}', "");
Expect(1, 65293, '\p{  Is_Dash_Punctuation}', "");
Expect(0, 65293, '\p{^  Is_Dash_Punctuation}', "");
Expect(0, 65293, '\P{  Is_Dash_Punctuation}', "");
Expect(1, 65293, '\P{^  Is_Dash_Punctuation}', "");
Expect(0, 65294, '\p{  Is_Dash_Punctuation}', "");
Expect(1, 65294, '\p{^  Is_Dash_Punctuation}', "");
Expect(1, 65294, '\P{  Is_Dash_Punctuation}', "");
Expect(0, 65294, '\P{^  Is_Dash_Punctuation}', "");
Error('\p{/a/-	pd}');
Error('\P{/a/-	pd}');
Expect(1, 65293, '\p{pd}', "");
Expect(0, 65293, '\p{^pd}', "");
Expect(0, 65293, '\P{pd}', "");
Expect(1, 65293, '\P{^pd}', "");
Expect(0, 65294, '\p{pd}', "");
Expect(1, 65294, '\p{^pd}', "");
Expect(1, 65294, '\P{pd}', "");
Expect(0, 65294, '\P{^pd}', "");
Expect(1, 65293, '\p{-Pd}', "");
Expect(0, 65293, '\p{^-Pd}', "");
Expect(0, 65293, '\P{-Pd}', "");
Expect(1, 65293, '\P{^-Pd}', "");
Expect(0, 65294, '\p{-Pd}', "");
Expect(1, 65294, '\p{^-Pd}', "");
Expect(1, 65294, '\P{-Pd}', "");
Expect(0, 65294, '\P{^-Pd}', "");
Error('\p{_:=Is_Pd}');
Error('\P{_:=Is_Pd}');
Expect(1, 65293, '\p{ispd}', "");
Expect(0, 65293, '\p{^ispd}', "");
Expect(0, 65293, '\P{ispd}', "");
Expect(1, 65293, '\P{^ispd}', "");
Expect(0, 65294, '\p{ispd}', "");
Expect(1, 65294, '\p{^ispd}', "");
Expect(1, 65294, '\P{ispd}', "");
Expect(0, 65294, '\P{^ispd}', "");
Expect(1, 65293, '\p{--Is_pd}', "");
Expect(0, 65293, '\p{^--Is_pd}', "");
Expect(0, 65293, '\P{--Is_pd}', "");
Expect(1, 65293, '\P{^--Is_pd}', "");
Expect(0, 65294, '\p{--Is_pd}', "");
Expect(1, 65294, '\p{^--Is_pd}', "");
Expect(1, 65294, '\P{--Is_pd}', "");
Expect(0, 65294, '\P{^--Is_pd}', "");
Error('\p{/a/ _In_Khmer}');
Error('\P{/a/ _In_Khmer}');
Expect(1, 6143, '\p{inkhmer}', "");
Expect(0, 6143, '\p{^inkhmer}', "");
Expect(0, 6143, '\P{inkhmer}', "");
Expect(1, 6143, '\P{^inkhmer}', "");
Expect(0, 6144, '\p{inkhmer}', "");
Expect(1, 6144, '\p{^inkhmer}', "");
Expect(1, 6144, '\P{inkhmer}', "");
Expect(0, 6144, '\P{^inkhmer}', "");
Expect(1, 6143, '\p{-In_Khmer}', "");
Expect(0, 6143, '\p{^-In_Khmer}', "");
Expect(0, 6143, '\P{-In_Khmer}', "");
Expect(1, 6143, '\P{^-In_Khmer}', "");
Expect(0, 6144, '\p{-In_Khmer}', "");
Expect(1, 6144, '\p{^-In_Khmer}', "");
Expect(1, 6144, '\P{-In_Khmer}', "");
Expect(0, 6144, '\P{^-In_Khmer}', "");
Error('\p{	 in_Khmer/a/}');
Error('\P{	 in_Khmer/a/}');
Expect(1, 6143, '\p{  in_Khmer}', "");
Expect(0, 6143, '\p{^  in_Khmer}', "");
Expect(0, 6143, '\P{  in_Khmer}', "");
Expect(1, 6143, '\P{^  in_Khmer}', "");
Expect(0, 6144, '\p{  in_Khmer}', "");
Expect(1, 6144, '\p{^  in_Khmer}', "");
Expect(1, 6144, '\P{  in_Khmer}', "");
Expect(0, 6144, '\P{^  in_Khmer}', "");
Error('\p{ /a/In_SAMARITAN}');
Error('\P{ /a/In_SAMARITAN}');
Expect(1, 2111, '\p{insamaritan}', "");
Expect(0, 2111, '\p{^insamaritan}', "");
Expect(0, 2111, '\P{insamaritan}', "");
Expect(1, 2111, '\P{^insamaritan}', "");
Expect(0, 2112, '\p{insamaritan}', "");
Expect(1, 2112, '\p{^insamaritan}', "");
Expect(1, 2112, '\P{insamaritan}', "");
Expect(0, 2112, '\P{^insamaritan}', "");
Expect(1, 2111, '\p{-In_Samaritan}', "");
Expect(0, 2111, '\p{^-In_Samaritan}', "");
Expect(0, 2111, '\P{-In_Samaritan}', "");
Expect(1, 2111, '\P{^-In_Samaritan}', "");
Expect(0, 2112, '\p{-In_Samaritan}', "");
Expect(1, 2112, '\p{^-In_Samaritan}', "");
Expect(1, 2112, '\P{-In_Samaritan}', "");
Expect(0, 2112, '\P{^-In_Samaritan}', "");
Error('\p{:=	_IN_Samaritan}');
Error('\P{:=	_IN_Samaritan}');
Expect(1, 2111, '\p{  In_SAMARITAN}', "");
Expect(0, 2111, '\p{^  In_SAMARITAN}', "");
Expect(0, 2111, '\P{  In_SAMARITAN}', "");
Expect(1, 2111, '\P{^  In_SAMARITAN}', "");
Expect(0, 2112, '\p{  In_SAMARITAN}', "");
Expect(1, 2112, '\p{^  In_SAMARITAN}', "");
Expect(1, 2112, '\P{  In_SAMARITAN}', "");
Expect(0, 2112, '\P{^  In_SAMARITAN}', "");
Error('\p{:=		in_Georgian}');
Error('\P{:=		in_Georgian}');
Expect(1, 4351, '\p{ingeorgian}', "");
Expect(0, 4351, '\p{^ingeorgian}', "");
Expect(0, 4351, '\P{ingeorgian}', "");
Expect(1, 4351, '\P{^ingeorgian}', "");
Expect(0, 4352, '\p{ingeorgian}', "");
Expect(1, 4352, '\p{^ingeorgian}', "");
Expect(1, 4352, '\P{ingeorgian}', "");
Expect(0, 4352, '\P{^ingeorgian}', "");
Expect(1, 4351, '\p{_In_GEORGIAN}', "");
Expect(0, 4351, '\p{^_In_GEORGIAN}', "");
Expect(0, 4351, '\P{_In_GEORGIAN}', "");
Expect(1, 4351, '\P{^_In_GEORGIAN}', "");
Expect(0, 4352, '\p{_In_GEORGIAN}', "");
Expect(1, 4352, '\p{^_In_GEORGIAN}', "");
Expect(1, 4352, '\P{_In_GEORGIAN}', "");
Expect(0, 4352, '\P{^_In_GEORGIAN}', "");
Error('\p{_/a/In_Georgian}');
Error('\P{_/a/In_Georgian}');
Expect(1, 4351, '\p{ In_Georgian}', "");
Expect(0, 4351, '\p{^ In_Georgian}', "");
Expect(0, 4351, '\P{ In_Georgian}', "");
Expect(1, 4351, '\P{^ In_Georgian}', "");
Expect(0, 4352, '\p{ In_Georgian}', "");
Expect(1, 4352, '\p{^ In_Georgian}', "");
Expect(1, 4352, '\P{ In_Georgian}', "");
Expect(0, 4352, '\P{^ In_Georgian}', "");
Error('\p{/a/		IN_Limbu}');
Error('\P{/a/		IN_Limbu}');
Expect(1, 6479, '\p{inlimbu}', "");
Expect(0, 6479, '\p{^inlimbu}', "");
Expect(0, 6479, '\P{inlimbu}', "");
Expect(1, 6479, '\P{^inlimbu}', "");
Expect(0, 6480, '\p{inlimbu}', "");
Expect(1, 6480, '\p{^inlimbu}', "");
Expect(1, 6480, '\P{inlimbu}', "");
Expect(0, 6480, '\P{^inlimbu}', "");
Expect(1, 6479, '\p{in_limbu}', "");
Expect(0, 6479, '\p{^in_limbu}', "");
Expect(0, 6479, '\P{in_limbu}', "");
Expect(1, 6479, '\P{^in_limbu}', "");
Expect(0, 6480, '\p{in_limbu}', "");
Expect(1, 6480, '\p{^in_limbu}', "");
Expect(1, 6480, '\P{in_limbu}', "");
Expect(0, 6480, '\P{^in_limbu}', "");
Error('\p{/a/-IN_Limbu}');
Error('\P{/a/-IN_Limbu}');
Expect(1, 6479, '\p{_	In_Limbu}', "");
Expect(0, 6479, '\p{^_	In_Limbu}', "");
Expect(0, 6479, '\P{_	In_Limbu}', "");
Expect(1, 6479, '\P{^_	In_Limbu}', "");
Expect(0, 6480, '\p{_	In_Limbu}', "");
Expect(1, 6480, '\p{^_	In_Limbu}', "");
Expect(1, 6480, '\P{_	In_Limbu}', "");
Expect(0, 6480, '\P{^_	In_Limbu}', "");
Error('\p{ :=In_LYDIAN}');
Error('\P{ :=In_LYDIAN}');
Expect(1, 67903, '\p{inlydian}', "");
Expect(0, 67903, '\p{^inlydian}', "");
Expect(0, 67903, '\P{inlydian}', "");
Expect(1, 67903, '\P{^inlydian}', "");
Expect(0, 67904, '\p{inlydian}', "");
Expect(1, 67904, '\p{^inlydian}', "");
Expect(1, 67904, '\P{inlydian}', "");
Expect(0, 67904, '\P{^inlydian}', "");
Expect(1, 67903, '\p{	-IN_lydian}', "");
Expect(0, 67903, '\p{^	-IN_lydian}', "");
Expect(0, 67903, '\P{	-IN_lydian}', "");
Expect(1, 67903, '\P{^	-IN_lydian}', "");
Expect(0, 67904, '\p{	-IN_lydian}', "");
Expect(1, 67904, '\p{^	-IN_lydian}', "");
Expect(1, 67904, '\P{	-IN_lydian}', "");
Expect(0, 67904, '\P{^	-IN_lydian}', "");
Error('\p{__IN_LYDIAN:=}');
Error('\P{__IN_LYDIAN:=}');
Expect(1, 67903, '\p{	in_Lydian}', "");
Expect(0, 67903, '\p{^	in_Lydian}', "");
Expect(0, 67903, '\P{	in_Lydian}', "");
Expect(1, 67903, '\P{^	in_Lydian}', "");
Expect(0, 67904, '\p{	in_Lydian}', "");
Expect(1, 67904, '\p{^	in_Lydian}', "");
Expect(1, 67904, '\P{	in_Lydian}', "");
Expect(0, 67904, '\P{^	in_Lydian}', "");
Error('\p{_In_Old_SOUTH_arabian/a/}');
Error('\P{_In_Old_SOUTH_arabian/a/}');
Expect(1, 68223, '\p{inoldsoutharabian}', "");
Expect(0, 68223, '\p{^inoldsoutharabian}', "");
Expect(0, 68223, '\P{inoldsoutharabian}', "");
Expect(1, 68223, '\P{^inoldsoutharabian}', "");
Expect(0, 68224, '\p{inoldsoutharabian}', "");
Expect(1, 68224, '\p{^inoldsoutharabian}', "");
Expect(1, 68224, '\P{inoldsoutharabian}', "");
Expect(0, 68224, '\P{^inoldsoutharabian}', "");
Expect(1, 68223, '\p{-in_Old_south_Arabian}', "");
Expect(0, 68223, '\p{^-in_Old_south_Arabian}', "");
Expect(0, 68223, '\P{-in_Old_south_Arabian}', "");
Expect(1, 68223, '\P{^-in_Old_south_Arabian}', "");
Expect(0, 68224, '\p{-in_Old_south_Arabian}', "");
Expect(1, 68224, '\p{^-in_Old_south_Arabian}', "");
Expect(1, 68224, '\P{-in_Old_south_Arabian}', "");
Expect(0, 68224, '\P{^-in_Old_south_Arabian}', "");
Error('\p{	in_Old_South_arabian:=}');
Error('\P{	in_Old_South_arabian:=}');
Expect(1, 68223, '\p{	in_Old_south_arabian}', "");
Expect(0, 68223, '\p{^	in_Old_south_arabian}', "");
Expect(0, 68223, '\P{	in_Old_south_arabian}', "");
Expect(1, 68223, '\P{^	in_Old_south_arabian}', "");
Expect(0, 68224, '\p{	in_Old_south_arabian}', "");
Expect(1, 68224, '\p{^	in_Old_south_arabian}', "");
Expect(1, 68224, '\P{	in_Old_south_arabian}', "");
Expect(0, 68224, '\P{^	in_Old_south_arabian}', "");
Error('\p{:=	-In_Kharoshthi}');
Error('\P{:=	-In_Kharoshthi}');
Expect(1, 68191, '\p{inkharoshthi}', "");
Expect(0, 68191, '\p{^inkharoshthi}', "");
Expect(0, 68191, '\P{inkharoshthi}', "");
Expect(1, 68191, '\P{^inkharoshthi}', "");
Expect(0, 68192, '\p{inkharoshthi}', "");
Expect(1, 68192, '\p{^inkharoshthi}', "");
Expect(1, 68192, '\P{inkharoshthi}', "");
Expect(0, 68192, '\P{^inkharoshthi}', "");
Expect(1, 68191, '\p{-	In_kharoshthi}', "");
Expect(0, 68191, '\p{^-	In_kharoshthi}', "");
Expect(0, 68191, '\P{-	In_kharoshthi}', "");
Expect(1, 68191, '\P{^-	In_kharoshthi}', "");
Expect(0, 68192, '\p{-	In_kharoshthi}', "");
Expect(1, 68192, '\p{^-	In_kharoshthi}', "");
Expect(1, 68192, '\P{-	In_kharoshthi}', "");
Expect(0, 68192, '\P{^-	In_kharoshthi}', "");
Error('\p{	 in_KHAROSHTHI:=}');
Error('\P{	 in_KHAROSHTHI:=}');
Expect(1, 68191, '\p{ In_KHAROSHTHI}', "");
Expect(0, 68191, '\p{^ In_KHAROSHTHI}', "");
Expect(0, 68191, '\P{ In_KHAROSHTHI}', "");
Expect(1, 68191, '\P{^ In_KHAROSHTHI}', "");
Expect(0, 68192, '\p{ In_KHAROSHTHI}', "");
Expect(1, 68192, '\p{^ In_KHAROSHTHI}', "");
Expect(1, 68192, '\P{ In_KHAROSHTHI}', "");
Expect(0, 68192, '\P{^ In_KHAROSHTHI}', "");
Error('\p{-	IN_BAMUM/a/}');
Error('\P{-	IN_BAMUM/a/}');
Expect(1, 42751, '\p{inbamum}', "");
Expect(0, 42751, '\p{^inbamum}', "");
Expect(0, 42751, '\P{inbamum}', "");
Expect(1, 42751, '\P{^inbamum}', "");
Expect(0, 42752, '\p{inbamum}', "");
Expect(1, 42752, '\p{^inbamum}', "");
Expect(1, 42752, '\P{inbamum}', "");
Expect(0, 42752, '\P{^inbamum}', "");
Expect(1, 42751, '\p{  in_bamum}', "");
Expect(0, 42751, '\p{^  in_bamum}', "");
Expect(0, 42751, '\P{  in_bamum}', "");
Expect(1, 42751, '\P{^  in_bamum}', "");
Expect(0, 42752, '\p{  in_bamum}', "");
Expect(1, 42752, '\p{^  in_bamum}', "");
Expect(1, 42752, '\P{  in_bamum}', "");
Expect(0, 42752, '\P{^  in_bamum}', "");
Error('\p{:=IN_bamum}');
Error('\P{:=IN_bamum}');
Expect(1, 42751, '\p{	_IN_Bamum}', "");
Expect(0, 42751, '\p{^	_IN_Bamum}', "");
Expect(0, 42751, '\P{	_IN_Bamum}', "");
Expect(1, 42751, '\P{^	_IN_Bamum}', "");
Expect(0, 42752, '\p{	_IN_Bamum}', "");
Expect(1, 42752, '\p{^	_IN_Bamum}', "");
Expect(1, 42752, '\P{	_IN_Bamum}', "");
Expect(0, 42752, '\P{^	_IN_Bamum}', "");
Error('\p{/a/ _In_MEROITIC_Cursive}');
Error('\P{/a/ _In_MEROITIC_Cursive}');
Expect(1, 68095, '\p{inmeroiticcursive}', "");
Expect(0, 68095, '\p{^inmeroiticcursive}', "");
Expect(0, 68095, '\P{inmeroiticcursive}', "");
Expect(1, 68095, '\P{^inmeroiticcursive}', "");
Expect(0, 68096, '\p{inmeroiticcursive}', "");
Expect(1, 68096, '\p{^inmeroiticcursive}', "");
Expect(1, 68096, '\P{inmeroiticcursive}', "");
Expect(0, 68096, '\P{^inmeroiticcursive}', "");
Expect(1, 68095, '\p{_	In_Meroitic_Cursive}', "");
Expect(0, 68095, '\p{^_	In_Meroitic_Cursive}', "");
Expect(0, 68095, '\P{_	In_Meroitic_Cursive}', "");
Expect(1, 68095, '\P{^_	In_Meroitic_Cursive}', "");
Expect(0, 68096, '\p{_	In_Meroitic_Cursive}', "");
Expect(1, 68096, '\p{^_	In_Meroitic_Cursive}', "");
Expect(1, 68096, '\P{_	In_Meroitic_Cursive}', "");
Expect(0, 68096, '\P{^_	In_Meroitic_Cursive}', "");
Error('\p{_In_meroitic_Cursive/a/}');
Error('\P{_In_meroitic_Cursive/a/}');
Expect(1, 68095, '\p{-In_Meroitic_cursive}', "");
Expect(0, 68095, '\p{^-In_Meroitic_cursive}', "");
Expect(0, 68095, '\P{-In_Meroitic_cursive}', "");
Expect(1, 68095, '\P{^-In_Meroitic_cursive}', "");
Expect(0, 68096, '\p{-In_Meroitic_cursive}', "");
Expect(1, 68096, '\p{^-In_Meroitic_cursive}', "");
Expect(1, 68096, '\P{-In_Meroitic_cursive}', "");
Expect(0, 68096, '\P{^-In_Meroitic_cursive}', "");
Error('\p{	 Changes_When_casemapped/a/}');
Error('\P{	 Changes_When_casemapped/a/}');
Expect(1, 66639, '\p{changeswhencasemapped}', "");
Expect(0, 66639, '\p{^changeswhencasemapped}', "");
Expect(0, 66639, '\P{changeswhencasemapped}', "");
Expect(1, 66639, '\P{^changeswhencasemapped}', "");
Expect(0, 66640, '\p{changeswhencasemapped}', "");
Expect(1, 66640, '\p{^changeswhencasemapped}', "");
Expect(1, 66640, '\P{changeswhencasemapped}', "");
Expect(0, 66640, '\P{^changeswhencasemapped}', "");
Expect(1, 66639, '\p{ 	Changes_When_CASEMAPPED}', "");
Expect(0, 66639, '\p{^ 	Changes_When_CASEMAPPED}', "");
Expect(0, 66639, '\P{ 	Changes_When_CASEMAPPED}', "");
Expect(1, 66639, '\P{^ 	Changes_When_CASEMAPPED}', "");
Expect(0, 66640, '\p{ 	Changes_When_CASEMAPPED}', "");
Expect(1, 66640, '\p{^ 	Changes_When_CASEMAPPED}', "");
Expect(1, 66640, '\P{ 	Changes_When_CASEMAPPED}', "");
Expect(0, 66640, '\P{^ 	Changes_When_CASEMAPPED}', "");
Error('\p{_/a/IS_Changes_WHEN_CASEMAPPED}');
Error('\P{_/a/IS_Changes_WHEN_CASEMAPPED}');
Expect(1, 66639, '\p{ischangeswhencasemapped}', "");
Expect(0, 66639, '\p{^ischangeswhencasemapped}', "");
Expect(0, 66639, '\P{ischangeswhencasemapped}', "");
Expect(1, 66639, '\P{^ischangeswhencasemapped}', "");
Expect(0, 66640, '\p{ischangeswhencasemapped}', "");
Expect(1, 66640, '\p{^ischangeswhencasemapped}', "");
Expect(1, 66640, '\P{ischangeswhencasemapped}', "");
Expect(0, 66640, '\P{^ischangeswhencasemapped}', "");
Expect(1, 66639, '\p{- IS_CHANGES_when_Casemapped}', "");
Expect(0, 66639, '\p{^- IS_CHANGES_when_Casemapped}', "");
Expect(0, 66639, '\P{- IS_CHANGES_when_Casemapped}', "");
Expect(1, 66639, '\P{^- IS_CHANGES_when_Casemapped}', "");
Expect(0, 66640, '\p{- IS_CHANGES_when_Casemapped}', "");
Expect(1, 66640, '\p{^- IS_CHANGES_when_Casemapped}', "");
Expect(1, 66640, '\P{- IS_CHANGES_when_Casemapped}', "");
Expect(0, 66640, '\P{^- IS_CHANGES_when_Casemapped}', "");
Error('\p{	:=cwcm}');
Error('\P{	:=cwcm}');
Expect(1, 66639, '\p{cwcm}', "");
Expect(0, 66639, '\p{^cwcm}', "");
Expect(0, 66639, '\P{cwcm}', "");
Expect(1, 66639, '\P{^cwcm}', "");
Expect(0, 66640, '\p{cwcm}', "");
Expect(1, 66640, '\p{^cwcm}', "");
Expect(1, 66640, '\P{cwcm}', "");
Expect(0, 66640, '\P{^cwcm}', "");
Expect(1, 66639, '\p{_CWCM}', "");
Expect(0, 66639, '\p{^_CWCM}', "");
Expect(0, 66639, '\P{_CWCM}', "");
Expect(1, 66639, '\P{^_CWCM}', "");
Expect(0, 66640, '\p{_CWCM}', "");
Expect(1, 66640, '\p{^_CWCM}', "");
Expect(1, 66640, '\P{_CWCM}', "");
Expect(0, 66640, '\P{^_CWCM}', "");
Error('\p{/a/-Is_CWCM}');
Error('\P{/a/-Is_CWCM}');
Expect(1, 66639, '\p{iscwcm}', "");
Expect(0, 66639, '\p{^iscwcm}', "");
Expect(0, 66639, '\P{iscwcm}', "");
Expect(1, 66639, '\P{^iscwcm}', "");
Expect(0, 66640, '\p{iscwcm}', "");
Expect(1, 66640, '\p{^iscwcm}', "");
Expect(1, 66640, '\P{iscwcm}', "");
Expect(0, 66640, '\P{^iscwcm}', "");
Expect(1, 66639, '\p{ IS_cwcm}', "");
Expect(0, 66639, '\p{^ IS_cwcm}', "");
Expect(0, 66639, '\P{ IS_cwcm}', "");
Expect(1, 66639, '\P{^ IS_cwcm}', "");
Expect(0, 66640, '\p{ IS_cwcm}', "");
Expect(1, 66640, '\p{^ IS_cwcm}', "");
Expect(1, 66640, '\P{ IS_cwcm}', "");
Expect(0, 66640, '\P{^ IS_cwcm}', "");
Error('\p{-_Dash/a/}');
Error('\P{-_Dash/a/}');
Expect(1, 65293, '\p{dash}', "");
Expect(0, 65293, '\p{^dash}', "");
Expect(0, 65293, '\P{dash}', "");
Expect(1, 65293, '\P{^dash}', "");
Expect(0, 65294, '\p{dash}', "");
Expect(1, 65294, '\p{^dash}', "");
Expect(1, 65294, '\P{dash}', "");
Expect(0, 65294, '\P{^dash}', "");
Expect(1, 65293, '\p{ 	Dash}', "");
Expect(0, 65293, '\p{^ 	Dash}', "");
Expect(0, 65293, '\P{ 	Dash}', "");
Expect(1, 65293, '\P{^ 	Dash}', "");
Expect(0, 65294, '\p{ 	Dash}', "");
Expect(1, 65294, '\p{^ 	Dash}', "");
Expect(1, 65294, '\P{ 	Dash}', "");
Expect(0, 65294, '\P{^ 	Dash}', "");
Error('\p{__is_DASH/a/}');
Error('\P{__is_DASH/a/}');
Expect(1, 65293, '\p{isdash}', "");
Expect(0, 65293, '\p{^isdash}', "");
Expect(0, 65293, '\P{isdash}', "");
Expect(1, 65293, '\P{^isdash}', "");
Expect(0, 65294, '\p{isdash}', "");
Expect(1, 65294, '\p{^isdash}', "");
Expect(1, 65294, '\P{isdash}', "");
Expect(0, 65294, '\P{^isdash}', "");
Expect(1, 65293, '\p{	 IS_DASH}', "");
Expect(0, 65293, '\p{^	 IS_DASH}', "");
Expect(0, 65293, '\P{	 IS_DASH}', "");
Expect(1, 65293, '\P{^	 IS_DASH}', "");
Expect(0, 65294, '\p{	 IS_DASH}', "");
Expect(1, 65294, '\p{^	 IS_DASH}', "");
Expect(1, 65294, '\P{	 IS_DASH}', "");
Expect(0, 65294, '\P{^	 IS_DASH}', "");
Error('\p{/a/-In_tamil}');
Error('\P{/a/-In_tamil}');
Expect(1, 3071, '\p{intamil}', "");
Expect(0, 3071, '\p{^intamil}', "");
Expect(0, 3071, '\P{intamil}', "");
Expect(1, 3071, '\P{^intamil}', "");
Expect(0, 3072, '\p{intamil}', "");
Expect(1, 3072, '\p{^intamil}', "");
Expect(1, 3072, '\P{intamil}', "");
Expect(0, 3072, '\P{^intamil}', "");
Expect(1, 3071, '\p{	_In_tamil}', "");
Expect(0, 3071, '\p{^	_In_tamil}', "");
Expect(0, 3071, '\P{	_In_tamil}', "");
Expect(1, 3071, '\P{^	_In_tamil}', "");
Expect(0, 3072, '\p{	_In_tamil}', "");
Expect(1, 3072, '\p{^	_In_tamil}', "");
Expect(1, 3072, '\P{	_In_tamil}', "");
Expect(0, 3072, '\P{^	_In_tamil}', "");
Error('\p{ In_TAMIL:=}');
Error('\P{ In_TAMIL:=}');
Expect(1, 3071, '\p{-In_Tamil}', "");
Expect(0, 3071, '\p{^-In_Tamil}', "");
Expect(0, 3071, '\P{-In_Tamil}', "");
Expect(1, 3071, '\P{^-In_Tamil}', "");
Expect(0, 3072, '\p{-In_Tamil}', "");
Expect(1, 3072, '\p{^-In_Tamil}', "");
Expect(1, 3072, '\P{-In_Tamil}', "");
Expect(0, 3072, '\P{^-In_Tamil}', "");
Error('\p{ IN_ethiopic/a/}');
Error('\P{ IN_ethiopic/a/}');
Expect(1, 4991, '\p{inethiopic}', "");
Expect(0, 4991, '\p{^inethiopic}', "");
Expect(0, 4991, '\P{inethiopic}', "");
Expect(1, 4991, '\P{^inethiopic}', "");
Expect(0, 4992, '\p{inethiopic}', "");
Expect(1, 4992, '\p{^inethiopic}', "");
Expect(1, 4992, '\P{inethiopic}', "");
Expect(0, 4992, '\P{^inethiopic}', "");
Expect(1, 4991, '\p{ -IN_Ethiopic}', "");
Expect(0, 4991, '\p{^ -IN_Ethiopic}', "");
Expect(0, 4991, '\P{ -IN_Ethiopic}', "");
Expect(1, 4991, '\P{^ -IN_Ethiopic}', "");
Expect(0, 4992, '\p{ -IN_Ethiopic}', "");
Expect(1, 4992, '\p{^ -IN_Ethiopic}', "");
Expect(1, 4992, '\P{ -IN_Ethiopic}', "");
Expect(0, 4992, '\P{^ -IN_Ethiopic}', "");
Error('\p{	:=In_Ethiopic}');
Error('\P{	:=In_Ethiopic}');
Expect(1, 4991, '\p{-_in_Ethiopic}', "");
Expect(0, 4991, '\p{^-_in_Ethiopic}', "");
Expect(0, 4991, '\P{-_in_Ethiopic}', "");
Expect(1, 4991, '\P{^-_in_Ethiopic}', "");
Expect(0, 4992, '\p{-_in_Ethiopic}', "");
Expect(1, 4992, '\p{^-_in_Ethiopic}', "");
Expect(1, 4992, '\P{-_in_Ethiopic}', "");
Expect(0, 4992, '\P{^-_in_Ethiopic}', "");
Error('\p{_:=In_Tai_Le}');
Error('\P{_:=In_Tai_Le}');
Expect(1, 6527, '\p{intaile}', "");
Expect(0, 6527, '\p{^intaile}', "");
Expect(0, 6527, '\P{intaile}', "");
Expect(1, 6527, '\P{^intaile}', "");
Expect(0, 6528, '\p{intaile}', "");
Expect(1, 6528, '\p{^intaile}', "");
Expect(1, 6528, '\P{intaile}', "");
Expect(0, 6528, '\P{^intaile}', "");
Expect(1, 6527, '\p{ in_Tai_Le}', "");
Expect(0, 6527, '\p{^ in_Tai_Le}', "");
Expect(0, 6527, '\P{ in_Tai_Le}', "");
Expect(1, 6527, '\P{^ in_Tai_Le}', "");
Expect(0, 6528, '\p{ in_Tai_Le}', "");
Expect(1, 6528, '\p{^ in_Tai_Le}', "");
Expect(1, 6528, '\P{ in_Tai_Le}', "");
Expect(0, 6528, '\P{^ in_Tai_Le}', "");
Error('\p{/a/	IN_Tai_Le}');
Error('\P{/a/	IN_Tai_Le}');
Expect(1, 6527, '\p{		In_TAI_Le}', "");
Expect(0, 6527, '\p{^		In_TAI_Le}', "");
Expect(0, 6527, '\P{		In_TAI_Le}', "");
Expect(1, 6527, '\P{^		In_TAI_Le}', "");
Expect(0, 6528, '\p{		In_TAI_Le}', "");
Expect(1, 6528, '\p{^		In_TAI_Le}', "");
Expect(1, 6528, '\P{		In_TAI_Le}', "");
Expect(0, 6528, '\P{^		In_TAI_Le}', "");
Error('\p{	/a/IN_Hebrew}');
Error('\P{	/a/IN_Hebrew}');
Expect(1, 1535, '\p{inhebrew}', "");
Expect(0, 1535, '\p{^inhebrew}', "");
Expect(0, 1535, '\P{inhebrew}', "");
Expect(1, 1535, '\P{^inhebrew}', "");
Expect(0, 1536, '\p{inhebrew}', "");
Expect(1, 1536, '\p{^inhebrew}', "");
Expect(1, 1536, '\P{inhebrew}', "");
Expect(0, 1536, '\P{^inhebrew}', "");
Expect(1, 1535, '\p{ In_hebrew}', "");
Expect(0, 1535, '\p{^ In_hebrew}', "");
Expect(0, 1535, '\P{ In_hebrew}', "");
Expect(1, 1535, '\P{^ In_hebrew}', "");
Expect(0, 1536, '\p{ In_hebrew}', "");
Expect(1, 1536, '\p{^ In_hebrew}', "");
Expect(1, 1536, '\P{ In_hebrew}', "");
Expect(0, 1536, '\P{^ In_hebrew}', "");
Error('\p{	/a/IN_Hebrew}');
Error('\P{	/a/IN_Hebrew}');
Expect(1, 1535, '\p{_-In_hebrew}', "");
Expect(0, 1535, '\p{^_-In_hebrew}', "");
Expect(0, 1535, '\P{_-In_hebrew}', "");
Expect(1, 1535, '\P{^_-In_hebrew}', "");
Expect(0, 1536, '\p{_-In_hebrew}', "");
Expect(1, 1536, '\p{^_-In_hebrew}', "");
Expect(1, 1536, '\P{_-In_hebrew}', "");
Expect(0, 1536, '\P{^_-In_hebrew}', "");
Error('\p{_ _Perl_non_Final_Folds:=}');
Error('\P{_ _Perl_non_Final_Folds:=}');
Expect(1, 8060, '\p{_PERL_non_Final_FOLDS}', "");
Expect(0, 8060, '\p{^_PERL_non_Final_FOLDS}', "");
Expect(0, 8060, '\P{_PERL_non_Final_FOLDS}', "");
Expect(1, 8060, '\P{^_PERL_non_Final_FOLDS}', "");
Expect(0, 8061, '\p{_PERL_non_Final_FOLDS}', "");
Expect(1, 8061, '\p{^_PERL_non_Final_FOLDS}', "");
Expect(1, 8061, '\P{_PERL_non_Final_FOLDS}', "");
Expect(0, 8061, '\P{^_PERL_non_Final_FOLDS}', "");
Error('\p{:=-_perl_NON_final_Folds}');
Error('\P{:=-_perl_NON_final_Folds}');
Expect(1, 8060, '\p{_Perl_non_Final_Folds}', "");
Expect(0, 8060, '\p{^_Perl_non_Final_Folds}', "");
Expect(0, 8060, '\P{_Perl_non_Final_Folds}', "");
Expect(1, 8060, '\P{^_Perl_non_Final_Folds}', "");
Expect(0, 8061, '\p{_Perl_non_Final_Folds}', "");
Expect(1, 8061, '\p{^_Perl_non_Final_Folds}', "");
Expect(1, 8061, '\P{_Perl_non_Final_Folds}', "");
Expect(0, 8061, '\P{^_Perl_non_Final_Folds}', "");
Error('\p{/a/_-In_INSCRIPTIONAL_Parthian}');
Error('\P{/a/_-In_INSCRIPTIONAL_Parthian}');
Expect(1, 68447, '\p{ininscriptionalparthian}', "");
Expect(0, 68447, '\p{^ininscriptionalparthian}', "");
Expect(0, 68447, '\P{ininscriptionalparthian}', "");
Expect(1, 68447, '\P{^ininscriptionalparthian}', "");
Expect(0, 68448, '\p{ininscriptionalparthian}', "");
Expect(1, 68448, '\p{^ininscriptionalparthian}', "");
Expect(1, 68448, '\P{ininscriptionalparthian}', "");
Expect(0, 68448, '\P{^ininscriptionalparthian}', "");
Expect(1, 68447, '\p{ IN_Inscriptional_parthian}', "");
Expect(0, 68447, '\p{^ IN_Inscriptional_parthian}', "");
Expect(0, 68447, '\P{ IN_Inscriptional_parthian}', "");
Expect(1, 68447, '\P{^ IN_Inscriptional_parthian}', "");
Expect(0, 68448, '\p{ IN_Inscriptional_parthian}', "");
Expect(1, 68448, '\p{^ IN_Inscriptional_parthian}', "");
Expect(1, 68448, '\P{ IN_Inscriptional_parthian}', "");
Expect(0, 68448, '\P{^ IN_Inscriptional_parthian}', "");
Error('\p{:=	In_Inscriptional_parthian}');
Error('\P{:=	In_Inscriptional_parthian}');
Expect(1, 68447, '\p{	-in_INSCRIPTIONAL_Parthian}', "");
Expect(0, 68447, '\p{^	-in_INSCRIPTIONAL_Parthian}', "");
Expect(0, 68447, '\P{	-in_INSCRIPTIONAL_Parthian}', "");
Expect(1, 68447, '\P{^	-in_INSCRIPTIONAL_Parthian}', "");
Expect(0, 68448, '\p{	-in_INSCRIPTIONAL_Parthian}', "");
Expect(1, 68448, '\p{^	-in_INSCRIPTIONAL_Parthian}', "");
Expect(1, 68448, '\P{	-in_INSCRIPTIONAL_Parthian}', "");
Expect(0, 68448, '\P{^	-in_INSCRIPTIONAL_Parthian}', "");
Error('\p{_in_New_TAI_Lue/a/}');
Error('\P{_in_New_TAI_Lue/a/}');
Expect(1, 6623, '\p{innewtailue}', "");
Expect(0, 6623, '\p{^innewtailue}', "");
Expect(0, 6623, '\P{innewtailue}', "");
Expect(1, 6623, '\P{^innewtailue}', "");
Expect(0, 6624, '\p{innewtailue}', "");
Expect(1, 6624, '\p{^innewtailue}', "");
Expect(1, 6624, '\P{innewtailue}', "");
Expect(0, 6624, '\P{^innewtailue}', "");
Expect(1, 6623, '\p{-_IN_New_Tai_lue}', "");
Expect(0, 6623, '\p{^-_IN_New_Tai_lue}', "");
Expect(0, 6623, '\P{-_IN_New_Tai_lue}', "");
Expect(1, 6623, '\P{^-_IN_New_Tai_lue}', "");
Expect(0, 6624, '\p{-_IN_New_Tai_lue}', "");
Expect(1, 6624, '\p{^-_IN_New_Tai_lue}', "");
Expect(1, 6624, '\P{-_IN_New_Tai_lue}', "");
Expect(0, 6624, '\P{^-_IN_New_Tai_lue}', "");
Error('\p{_:=In_New_Tai_Lue}');
Error('\P{_:=In_New_Tai_Lue}');
Expect(1, 6623, '\p{-	IN_NEW_Tai_Lue}', "");
Expect(0, 6623, '\p{^-	IN_NEW_Tai_Lue}', "");
Expect(0, 6623, '\P{-	IN_NEW_Tai_Lue}', "");
Expect(1, 6623, '\P{^-	IN_NEW_Tai_Lue}', "");
Expect(0, 6624, '\p{-	IN_NEW_Tai_Lue}', "");
Expect(1, 6624, '\p{^-	IN_NEW_Tai_Lue}', "");
Expect(1, 6624, '\P{-	IN_NEW_Tai_Lue}', "");
Expect(0, 6624, '\P{^-	IN_NEW_Tai_Lue}', "");
Error('\p{:=	-In_Takri}');
Error('\P{:=	-In_Takri}');
Expect(1, 71375, '\p{intakri}', "");
Expect(0, 71375, '\p{^intakri}', "");
Expect(0, 71375, '\P{intakri}', "");
Expect(1, 71375, '\P{^intakri}', "");
Expect(0, 71376, '\p{intakri}', "");
Expect(1, 71376, '\p{^intakri}', "");
Expect(1, 71376, '\P{intakri}', "");
Expect(0, 71376, '\P{^intakri}', "");
Expect(1, 71375, '\p{-_In_TAKRI}', "");
Expect(0, 71375, '\p{^-_In_TAKRI}', "");
Expect(0, 71375, '\P{-_In_TAKRI}', "");
Expect(1, 71375, '\P{^-_In_TAKRI}', "");
Expect(0, 71376, '\p{-_In_TAKRI}', "");
Expect(1, 71376, '\p{^-_In_TAKRI}', "");
Expect(1, 71376, '\P{-_In_TAKRI}', "");
Expect(0, 71376, '\P{^-_In_TAKRI}', "");
Error('\p{/a/-	In_Takri}');
Error('\P{/a/-	In_Takri}');
Expect(1, 71375, '\p{--in_Takri}', "");
Expect(0, 71375, '\p{^--in_Takri}', "");
Expect(0, 71375, '\P{--in_Takri}', "");
Expect(1, 71375, '\P{^--in_Takri}', "");
Expect(0, 71376, '\p{--in_Takri}', "");
Expect(1, 71376, '\p{^--in_Takri}', "");
Expect(1, 71376, '\P{--in_Takri}', "");
Expect(0, 71376, '\P{^--in_Takri}', "");
Error('\p{:=_in_IMPERIAL_Aramaic}');
Error('\P{:=_in_IMPERIAL_Aramaic}');
Expect(1, 67679, '\p{inimperialaramaic}', "");
Expect(0, 67679, '\p{^inimperialaramaic}', "");
Expect(0, 67679, '\P{inimperialaramaic}', "");
Expect(1, 67679, '\P{^inimperialaramaic}', "");
Expect(0, 67680, '\p{inimperialaramaic}', "");
Expect(1, 67680, '\p{^inimperialaramaic}', "");
Expect(1, 67680, '\P{inimperialaramaic}', "");
Expect(0, 67680, '\P{^inimperialaramaic}', "");
Expect(1, 67679, '\p{	-in_IMPERIAL_Aramaic}', "");
Expect(0, 67679, '\p{^	-in_IMPERIAL_Aramaic}', "");
Expect(0, 67679, '\P{	-in_IMPERIAL_Aramaic}', "");
Expect(1, 67679, '\P{^	-in_IMPERIAL_Aramaic}', "");
Expect(0, 67680, '\p{	-in_IMPERIAL_Aramaic}', "");
Expect(1, 67680, '\p{^	-in_IMPERIAL_Aramaic}', "");
Expect(1, 67680, '\P{	-in_IMPERIAL_Aramaic}', "");
Expect(0, 67680, '\P{^	-in_IMPERIAL_Aramaic}', "");
Error('\p{-	in_imperial_aramaic:=}');
Error('\P{-	in_imperial_aramaic:=}');
Expect(1, 67679, '\p{in_Imperial_Aramaic}', "");
Expect(0, 67679, '\p{^in_Imperial_Aramaic}', "");
Expect(0, 67679, '\P{in_Imperial_Aramaic}', "");
Expect(1, 67679, '\P{^in_Imperial_Aramaic}', "");
Expect(0, 67680, '\p{in_Imperial_Aramaic}', "");
Expect(1, 67680, '\p{^in_Imperial_Aramaic}', "");
Expect(1, 67680, '\P{in_Imperial_Aramaic}', "");
Expect(0, 67680, '\P{^in_Imperial_Aramaic}', "");
Error('\p{/a/_	in_Saurashtra}');
Error('\P{/a/_	in_Saurashtra}');
Expect(1, 43231, '\p{insaurashtra}', "");
Expect(0, 43231, '\p{^insaurashtra}', "");
Expect(0, 43231, '\P{insaurashtra}', "");
Expect(1, 43231, '\P{^insaurashtra}', "");
Expect(0, 43232, '\p{insaurashtra}', "");
Expect(1, 43232, '\p{^insaurashtra}', "");
Expect(1, 43232, '\P{insaurashtra}', "");
Expect(0, 43232, '\P{^insaurashtra}', "");
Expect(1, 43231, '\p{-in_saurashtra}', "");
Expect(0, 43231, '\p{^-in_saurashtra}', "");
Expect(0, 43231, '\P{-in_saurashtra}', "");
Expect(1, 43231, '\P{^-in_saurashtra}', "");
Expect(0, 43232, '\p{-in_saurashtra}', "");
Expect(1, 43232, '\p{^-in_saurashtra}', "");
Expect(1, 43232, '\P{-in_saurashtra}', "");
Expect(0, 43232, '\P{^-in_saurashtra}', "");
Error('\p{:=-In_SAURASHTRA}');
Error('\P{:=-In_SAURASHTRA}');
Expect(1, 43231, '\p{	in_Saurashtra}', "");
Expect(0, 43231, '\p{^	in_Saurashtra}', "");
Expect(0, 43231, '\P{	in_Saurashtra}', "");
Expect(1, 43231, '\P{^	in_Saurashtra}', "");
Expect(0, 43232, '\p{	in_Saurashtra}', "");
Expect(1, 43232, '\p{^	in_Saurashtra}', "");
Expect(1, 43232, '\P{	in_Saurashtra}', "");
Expect(0, 43232, '\P{^	in_Saurashtra}', "");
Error('\p{		IN_gujarati:=}');
Error('\P{		IN_gujarati:=}');
Expect(1, 2815, '\p{ingujarati}', "");
Expect(0, 2815, '\p{^ingujarati}', "");
Expect(0, 2815, '\P{ingujarati}', "");
Expect(1, 2815, '\P{^ingujarati}', "");
Expect(0, 2816, '\p{ingujarati}', "");
Expect(1, 2816, '\p{^ingujarati}', "");
Expect(1, 2816, '\P{ingujarati}', "");
Expect(0, 2816, '\P{^ingujarati}', "");
Expect(1, 2815, '\p{-IN_Gujarati}', "");
Expect(0, 2815, '\p{^-IN_Gujarati}', "");
Expect(0, 2815, '\P{-IN_Gujarati}', "");
Expect(1, 2815, '\P{^-IN_Gujarati}', "");
Expect(0, 2816, '\p{-IN_Gujarati}', "");
Expect(1, 2816, '\p{^-IN_Gujarati}', "");
Expect(1, 2816, '\P{-IN_Gujarati}', "");
Expect(0, 2816, '\P{^-IN_Gujarati}', "");
Error('\p{-/a/in_GUJARATI}');
Error('\P{-/a/in_GUJARATI}');
Expect(1, 2815, '\p{--In_Gujarati}', "");
Expect(0, 2815, '\p{^--In_Gujarati}', "");
Expect(0, 2815, '\P{--In_Gujarati}', "");
Expect(1, 2815, '\P{^--In_Gujarati}', "");
Expect(0, 2816, '\p{--In_Gujarati}', "");
Expect(1, 2816, '\p{^--In_Gujarati}', "");
Expect(1, 2816, '\P{--In_Gujarati}', "");
Expect(0, 2816, '\P{^--In_Gujarati}', "");
Error('\p{/a/_ In_Osmanya}');
Error('\P{/a/_ In_Osmanya}');
Expect(1, 66735, '\p{inosmanya}', "");
Expect(0, 66735, '\p{^inosmanya}', "");
Expect(0, 66735, '\P{inosmanya}', "");
Expect(1, 66735, '\P{^inosmanya}', "");
Expect(0, 66736, '\p{inosmanya}', "");
Expect(1, 66736, '\p{^inosmanya}', "");
Expect(1, 66736, '\P{inosmanya}', "");
Expect(0, 66736, '\P{^inosmanya}', "");
Expect(1, 66735, '\p{	 IN_OSMANYA}', "");
Expect(0, 66735, '\p{^	 IN_OSMANYA}', "");
Expect(0, 66735, '\P{	 IN_OSMANYA}', "");
Expect(1, 66735, '\P{^	 IN_OSMANYA}', "");
Expect(0, 66736, '\p{	 IN_OSMANYA}', "");
Expect(1, 66736, '\p{^	 IN_OSMANYA}', "");
Expect(1, 66736, '\P{	 IN_OSMANYA}', "");
Expect(0, 66736, '\P{^	 IN_OSMANYA}', "");
Error('\p{ :=in_Osmanya}');
Error('\P{ :=in_Osmanya}');
Expect(1, 66735, '\p{_	In_Osmanya}', "");
Expect(0, 66735, '\p{^_	In_Osmanya}', "");
Expect(0, 66735, '\P{_	In_Osmanya}', "");
Expect(1, 66735, '\P{^_	In_Osmanya}', "");
Expect(0, 66736, '\p{_	In_Osmanya}', "");
Expect(1, 66736, '\p{^_	In_Osmanya}', "");
Expect(1, 66736, '\P{_	In_Osmanya}', "");
Expect(0, 66736, '\P{^_	In_Osmanya}', "");
Error('\p{	/a/In_Tifinagh}');
Error('\P{	/a/In_Tifinagh}');
Expect(1, 11647, '\p{intifinagh}', "");
Expect(0, 11647, '\p{^intifinagh}', "");
Expect(0, 11647, '\P{intifinagh}', "");
Expect(1, 11647, '\P{^intifinagh}', "");
Expect(0, 11648, '\p{intifinagh}', "");
Expect(1, 11648, '\p{^intifinagh}', "");
Expect(1, 11648, '\P{intifinagh}', "");
Expect(0, 11648, '\P{^intifinagh}', "");
Expect(1, 11647, '\p{--In_tifinagh}', "");
Expect(0, 11647, '\p{^--In_tifinagh}', "");
Expect(0, 11647, '\P{--In_tifinagh}', "");
Expect(1, 11647, '\P{^--In_tifinagh}', "");
Expect(0, 11648, '\p{--In_tifinagh}', "");
Expect(1, 11648, '\p{^--In_tifinagh}', "");
Expect(1, 11648, '\P{--In_tifinagh}', "");
Expect(0, 11648, '\P{^--In_tifinagh}', "");
Error('\p{	In_Tifinagh/a/}');
Error('\P{	In_Tifinagh/a/}');
Expect(1, 11647, '\p{_	In_TIFINAGH}', "");
Expect(0, 11647, '\p{^_	In_TIFINAGH}', "");
Expect(0, 11647, '\P{_	In_TIFINAGH}', "");
Expect(1, 11647, '\P{^_	In_TIFINAGH}', "");
Expect(0, 11648, '\p{_	In_TIFINAGH}', "");
Expect(1, 11648, '\p{^_	In_TIFINAGH}', "");
Expect(1, 11648, '\P{_	In_TIFINAGH}', "");
Expect(0, 11648, '\P{^_	In_TIFINAGH}', "");
Error('\p{- posixprint:=}');
Error('\P{- posixprint:=}');
Expect(1, 126, '\p{posixprint}', "");
Expect(0, 126, '\p{^posixprint}', "");
Expect(0, 126, '\P{posixprint}', "");
Expect(1, 126, '\P{^posixprint}', "");
Expect(0, 127, '\p{posixprint}', "");
Expect(1, 127, '\p{^posixprint}', "");
Expect(1, 127, '\P{posixprint}', "");
Expect(0, 127, '\P{^posixprint}', "");
Expect(1, 126, '\p{ _POSIXPRINT}', "");
Expect(0, 126, '\p{^ _POSIXPRINT}', "");
Expect(0, 126, '\P{ _POSIXPRINT}', "");
Expect(1, 126, '\P{^ _POSIXPRINT}', "");
Expect(0, 127, '\p{ _POSIXPRINT}', "");
Expect(1, 127, '\p{^ _POSIXPRINT}', "");
Expect(1, 127, '\P{ _POSIXPRINT}', "");
Expect(0, 127, '\P{^ _POSIXPRINT}', "");
Error('\p{	:=is_PosixPrint}');
Error('\P{	:=is_PosixPrint}');
Expect(1, 126, '\p{isposixprint}', "");
Expect(0, 126, '\p{^isposixprint}', "");
Expect(0, 126, '\P{isposixprint}', "");
Expect(1, 126, '\P{^isposixprint}', "");
Expect(0, 127, '\p{isposixprint}', "");
Expect(1, 127, '\p{^isposixprint}', "");
Expect(1, 127, '\P{isposixprint}', "");
Expect(0, 127, '\P{^isposixprint}', "");
Expect(1, 126, '\p{_	is_POSIXPRINT}', "");
Expect(0, 126, '\p{^_	is_POSIXPRINT}', "");
Expect(0, 126, '\P{_	is_POSIXPRINT}', "");
Expect(1, 126, '\P{^_	is_POSIXPRINT}', "");
Expect(0, 127, '\p{_	is_POSIXPRINT}', "");
Expect(1, 127, '\p{^_	is_POSIXPRINT}', "");
Expect(1, 127, '\P{_	is_POSIXPRINT}', "");
Expect(0, 127, '\P{^_	is_POSIXPRINT}', "");
Error('\p{/a/In_Hanunoo}');
Error('\P{/a/In_Hanunoo}');
Expect(1, 5951, '\p{inhanunoo}', "");
Expect(0, 5951, '\p{^inhanunoo}', "");
Expect(0, 5951, '\P{inhanunoo}', "");
Expect(1, 5951, '\P{^inhanunoo}', "");
Expect(0, 5952, '\p{inhanunoo}', "");
Expect(1, 5952, '\p{^inhanunoo}', "");
Expect(1, 5952, '\P{inhanunoo}', "");
Expect(0, 5952, '\P{^inhanunoo}', "");
Expect(1, 5951, '\p{-in_Hanunoo}', "");
Expect(0, 5951, '\p{^-in_Hanunoo}', "");
Expect(0, 5951, '\P{-in_Hanunoo}', "");
Expect(1, 5951, '\P{^-in_Hanunoo}', "");
Expect(0, 5952, '\p{-in_Hanunoo}', "");
Expect(1, 5952, '\p{^-in_Hanunoo}', "");
Expect(1, 5952, '\P{-in_Hanunoo}', "");
Expect(0, 5952, '\P{^-in_Hanunoo}', "");
Error('\p{	In_Hanunoo/a/}');
Error('\P{	In_Hanunoo/a/}');
Expect(1, 5951, '\p{		in_Hanunoo}', "");
Expect(0, 5951, '\p{^		in_Hanunoo}', "");
Expect(0, 5951, '\P{		in_Hanunoo}', "");
Expect(1, 5951, '\P{^		in_Hanunoo}', "");
Expect(0, 5952, '\p{		in_Hanunoo}', "");
Expect(1, 5952, '\p{^		in_Hanunoo}', "");
Expect(1, 5952, '\P{		in_Hanunoo}', "");
Expect(0, 5952, '\P{^		in_Hanunoo}', "");
Error('\p{/a/_IN_Cyrillic}');
Error('\P{/a/_IN_Cyrillic}');
Expect(1, 1279, '\p{incyrillic}', "");
Expect(0, 1279, '\p{^incyrillic}', "");
Expect(0, 1279, '\P{incyrillic}', "");
Expect(1, 1279, '\P{^incyrillic}', "");
Expect(0, 1280, '\p{incyrillic}', "");
Expect(1, 1280, '\p{^incyrillic}', "");
Expect(1, 1280, '\P{incyrillic}', "");
Expect(0, 1280, '\P{^incyrillic}', "");
Expect(1, 1279, '\p{		in_Cyrillic}', "");
Expect(0, 1279, '\p{^		in_Cyrillic}', "");
Expect(0, 1279, '\P{		in_Cyrillic}', "");
Expect(1, 1279, '\P{^		in_Cyrillic}', "");
Expect(0, 1280, '\p{		in_Cyrillic}', "");
Expect(1, 1280, '\p{^		in_Cyrillic}', "");
Expect(1, 1280, '\P{		in_Cyrillic}', "");
Expect(0, 1280, '\P{^		in_Cyrillic}', "");
Error('\p{	_in_CYRILLIC/a/}');
Error('\P{	_in_CYRILLIC/a/}');
Expect(1, 1279, '\p{	in_CYRILLIC}', "");
Expect(0, 1279, '\p{^	in_CYRILLIC}', "");
Expect(0, 1279, '\P{	in_CYRILLIC}', "");
Expect(1, 1279, '\P{^	in_CYRILLIC}', "");
Expect(0, 1280, '\p{	in_CYRILLIC}', "");
Expect(1, 1280, '\p{^	in_CYRILLIC}', "");
Expect(1, 1280, '\P{	in_CYRILLIC}', "");
Expect(0, 1280, '\P{^	in_CYRILLIC}', "");
Error('\p{	In_buhid:=}');
Error('\P{	In_buhid:=}');
Expect(1, 5983, '\p{inbuhid}', "");
Expect(0, 5983, '\p{^inbuhid}', "");
Expect(0, 5983, '\P{inbuhid}', "");
Expect(1, 5983, '\P{^inbuhid}', "");
Expect(0, 5984, '\p{inbuhid}', "");
Expect(1, 5984, '\p{^inbuhid}', "");
Expect(1, 5984, '\P{inbuhid}', "");
Expect(0, 5984, '\P{^inbuhid}', "");
Expect(1, 5983, '\p{	-In_Buhid}', "");
Expect(0, 5983, '\p{^	-In_Buhid}', "");
Expect(0, 5983, '\P{	-In_Buhid}', "");
Expect(1, 5983, '\P{^	-In_Buhid}', "");
Expect(0, 5984, '\p{	-In_Buhid}', "");
Expect(1, 5984, '\p{^	-In_Buhid}', "");
Expect(1, 5984, '\P{	-In_Buhid}', "");
Expect(0, 5984, '\P{^	-In_Buhid}', "");
Error('\p{		In_Buhid:=}');
Error('\P{		In_Buhid:=}');
Expect(1, 5983, '\p{-in_buhid}', "");
Expect(0, 5983, '\p{^-in_buhid}', "");
Expect(0, 5983, '\P{-in_buhid}', "");
Expect(1, 5983, '\P{^-in_buhid}', "");
Expect(0, 5984, '\p{-in_buhid}', "");
Expect(1, 5984, '\p{^-in_buhid}', "");
Expect(1, 5984, '\P{-in_buhid}', "");
Expect(0, 5984, '\P{^-in_buhid}', "");
Error('\p{  in_Runic/a/}');
Error('\P{  in_Runic/a/}');
Expect(1, 5887, '\p{inrunic}', "");
Expect(0, 5887, '\p{^inrunic}', "");
Expect(0, 5887, '\P{inrunic}', "");
Expect(1, 5887, '\P{^inrunic}', "");
Expect(0, 5888, '\p{inrunic}', "");
Expect(1, 5888, '\p{^inrunic}', "");
Expect(1, 5888, '\P{inrunic}', "");
Expect(0, 5888, '\P{^inrunic}', "");
Expect(1, 5887, '\p{_In_runic}', "");
Expect(0, 5887, '\p{^_In_runic}', "");
Expect(0, 5887, '\P{_In_runic}', "");
Expect(1, 5887, '\P{^_In_runic}', "");
Expect(0, 5888, '\p{_In_runic}', "");
Expect(1, 5888, '\p{^_In_runic}', "");
Expect(1, 5888, '\P{_In_runic}', "");
Expect(0, 5888, '\P{^_In_runic}', "");
Error('\p{-	In_Runic/a/}');
Error('\P{-	In_Runic/a/}');
Expect(1, 5887, '\p{		IN_runic}', "");
Expect(0, 5887, '\p{^		IN_runic}', "");
Expect(0, 5887, '\P{		IN_runic}', "");
Expect(1, 5887, '\P{^		IN_runic}', "");
Expect(0, 5888, '\p{		IN_runic}', "");
Expect(1, 5888, '\p{^		IN_runic}', "");
Expect(1, 5888, '\P{		IN_runic}', "");
Expect(0, 5888, '\P{^		IN_runic}', "");
Error('\p{Block=:=CJK_Unified_Ideographs_EXTENSION_a}');
Error('\P{Block=:=CJK_Unified_Ideographs_EXTENSION_a}');
Expect(1, 19903, '\p{Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\p{^Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\P{Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\P{^Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\p{Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\p{^Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\P{Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\P{^Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\p{Block=_ CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19903, '\p{^Block=_ CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19903, '\P{Block=_ CJK_unified_Ideographs_extension_A}', "");
Expect(1, 19903, '\P{^Block=_ CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19904, '\p{Block=_ CJK_unified_Ideographs_extension_A}', "");
Expect(1, 19904, '\p{^Block=_ CJK_unified_Ideographs_extension_A}', "");
Expect(1, 19904, '\P{Block=_ CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19904, '\P{^Block=_ CJK_unified_Ideographs_extension_A}', "");
Error('\p{Blk=/a/-	cjk_EXT_A}');
Error('\P{Blk=/a/-	cjk_EXT_A}');
Expect(1, 19903, '\p{Blk:	cjkexta}', "");
Expect(0, 19903, '\p{^Blk:	cjkexta}', "");
Expect(0, 19903, '\P{Blk:	cjkexta}', "");
Expect(1, 19903, '\P{^Blk:	cjkexta}', "");
Expect(0, 19904, '\p{Blk:	cjkexta}', "");
Expect(1, 19904, '\p{^Blk:	cjkexta}', "");
Expect(1, 19904, '\P{Blk:	cjkexta}', "");
Expect(0, 19904, '\P{^Blk:	cjkexta}', "");
Expect(1, 19903, '\p{Blk=-CJK_ext_A}', "");
Expect(0, 19903, '\p{^Blk=-CJK_ext_A}', "");
Expect(0, 19903, '\P{Blk=-CJK_ext_A}', "");
Expect(1, 19903, '\P{^Blk=-CJK_ext_A}', "");
Expect(0, 19904, '\p{Blk=-CJK_ext_A}', "");
Expect(1, 19904, '\p{^Blk=-CJK_ext_A}', "");
Expect(1, 19904, '\P{Blk=-CJK_ext_A}', "");
Expect(0, 19904, '\P{^Blk=-CJK_ext_A}', "");
Error('\p{Is_Block=:= 	CJK_unified_Ideographs_extension_A}');
Error('\P{Is_Block=:= 	CJK_unified_Ideographs_extension_A}');
Expect(1, 19903, '\p{Is_Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\p{^Is_Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19903, '\P{Is_Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\P{^Is_Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\p{Is_Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\p{^Is_Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19904, '\P{Is_Block=cjkunifiedideographsextensiona}', "");
Expect(0, 19904, '\P{^Is_Block=cjkunifiedideographsextensiona}', "");
Expect(1, 19903, '\p{Is_Block: --CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19903, '\p{^Is_Block: --CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19903, '\P{Is_Block: --CJK_unified_Ideographs_extension_A}', "");
Expect(1, 19903, '\P{^Is_Block: --CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19904, '\p{Is_Block: --CJK_unified_Ideographs_extension_A}', "");
Expect(1, 19904, '\p{^Is_Block: --CJK_unified_Ideographs_extension_A}', "");
Expect(1, 19904, '\P{Is_Block: --CJK_unified_Ideographs_extension_A}', "");
Expect(0, 19904, '\P{^Is_Block: --CJK_unified_Ideographs_extension_A}', "");
Error('\p{Is_Blk=/a/		cjk_ext_A}');
Error('\P{Is_Blk=/a/		cjk_ext_A}');
Expect(1, 19903, '\p{Is_Blk=cjkexta}', "");
Expect(0, 19903, '\p{^Is_Blk=cjkexta}', "");
Expect(0, 19903, '\P{Is_Blk=cjkexta}', "");
Expect(1, 19903, '\P{^Is_Blk=cjkexta}', "");
Expect(0, 19904, '\p{Is_Blk=cjkexta}', "");
Expect(1, 19904, '\p{^Is_Blk=cjkexta}', "");
Expect(1, 19904, '\P{Is_Blk=cjkexta}', "");
Expect(0, 19904, '\P{^Is_Blk=cjkexta}', "");
Expect(1, 19903, '\p{Is_Blk:    CJK_ext_a}', "");
Expect(0, 19903, '\p{^Is_Blk:    CJK_ext_a}', "");
Expect(0, 19903, '\P{Is_Blk:    CJK_ext_a}', "");
Expect(1, 19903, '\P{^Is_Blk:    CJK_ext_a}', "");
Expect(0, 19904, '\p{Is_Blk:    CJK_ext_a}', "");
Expect(1, 19904, '\p{^Is_Blk:    CJK_ext_a}', "");
Expect(1, 19904, '\P{Is_Blk:    CJK_ext_a}', "");
Expect(0, 19904, '\P{^Is_Blk:    CJK_ext_a}', "");
Error('\p{Block=:=	Private_USE_Area}');
Error('\P{Block=:=	Private_USE_Area}');
Expect(1, 63743, '\p{Block=privateusearea}', "");
Expect(0, 63743, '\p{^Block=privateusearea}', "");
Expect(0, 63743, '\P{Block=privateusearea}', "");
Expect(1, 63743, '\P{^Block=privateusearea}', "");
Expect(0, 63744, '\p{Block=privateusearea}', "");
Expect(1, 63744, '\p{^Block=privateusearea}', "");
Expect(1, 63744, '\P{Block=privateusearea}', "");
Expect(0, 63744, '\P{^Block=privateusearea}', "");
Expect(1, 63743, '\p{Block= Private_Use_area}', "");
Expect(0, 63743, '\p{^Block= Private_Use_area}', "");
Expect(0, 63743, '\P{Block= Private_Use_area}', "");
Expect(1, 63743, '\P{^Block= Private_Use_area}', "");
Expect(0, 63744, '\p{Block= Private_Use_area}', "");
Expect(1, 63744, '\p{^Block= Private_Use_area}', "");
Expect(1, 63744, '\P{Block= Private_Use_area}', "");
Expect(0, 63744, '\P{^Block= Private_Use_area}', "");
Error('\p{Blk=/a/- PUA}');
Error('\P{Blk=/a/- PUA}');
Expect(1, 63743, '\p{Blk=pua}', "");
Expect(0, 63743, '\p{^Blk=pua}', "");
Expect(0, 63743, '\P{Blk=pua}', "");
Expect(1, 63743, '\P{^Blk=pua}', "");
Expect(0, 63744, '\p{Blk=pua}', "");
Expect(1, 63744, '\p{^Blk=pua}', "");
Expect(1, 63744, '\P{Blk=pua}', "");
Expect(0, 63744, '\P{^Blk=pua}', "");
Expect(1, 63743, '\p{Blk= _PUA}', "");
Expect(0, 63743, '\p{^Blk= _PUA}', "");
Expect(0, 63743, '\P{Blk= _PUA}', "");
Expect(1, 63743, '\P{^Blk= _PUA}', "");
Expect(0, 63744, '\p{Blk= _PUA}', "");
Expect(1, 63744, '\p{^Blk= _PUA}', "");
Expect(1, 63744, '\P{Blk= _PUA}', "");
Expect(0, 63744, '\P{^Blk= _PUA}', "");
Error('\p{Is_Block= PRIVATE_USE:=}');
Error('\P{Is_Block= PRIVATE_USE:=}');
Expect(1, 63743, '\p{Is_Block=privateuse}', "");
Expect(0, 63743, '\p{^Is_Block=privateuse}', "");
Expect(0, 63743, '\P{Is_Block=privateuse}', "");
Expect(1, 63743, '\P{^Is_Block=privateuse}', "");
Expect(0, 63744, '\p{Is_Block=privateuse}', "");
Expect(1, 63744, '\p{^Is_Block=privateuse}', "");
Expect(1, 63744, '\P{Is_Block=privateuse}', "");
Expect(0, 63744, '\P{^Is_Block=privateuse}', "");
Expect(1, 63743, '\p{Is_Block=-_PRIVATE_USE}', "");
Expect(0, 63743, '\p{^Is_Block=-_PRIVATE_USE}', "");
Expect(0, 63743, '\P{Is_Block=-_PRIVATE_USE}', "");
Expect(1, 63743, '\P{^Is_Block=-_PRIVATE_USE}', "");
Expect(0, 63744, '\p{Is_Block=-_PRIVATE_USE}', "");
Expect(1, 63744, '\p{^Is_Block=-_PRIVATE_USE}', "");
Expect(1, 63744, '\P{Is_Block=-_PRIVATE_USE}', "");
Expect(0, 63744, '\P{^Is_Block=-_PRIVATE_USE}', "");
Error('\p{Is_Blk=-/a/PRIVATE_Use_AREA}');
Error('\P{Is_Blk=-/a/PRIVATE_Use_AREA}');
Expect(1, 63743, '\p{Is_Blk:privateusearea}', "");
Expect(0, 63743, '\p{^Is_Blk:privateusearea}', "");
Expect(0, 63743, '\P{Is_Blk:privateusearea}', "");
Expect(1, 63743, '\P{^Is_Blk:privateusearea}', "");
Expect(0, 63744, '\p{Is_Blk:privateusearea}', "");
Expect(1, 63744, '\p{^Is_Blk:privateusearea}', "");
Expect(1, 63744, '\P{Is_Blk:privateusearea}', "");
Expect(0, 63744, '\P{^Is_Blk:privateusearea}', "");
Expect(1, 63743, '\p{Is_Blk=Private_Use_area}', "");
Expect(0, 63743, '\p{^Is_Blk=Private_Use_area}', "");
Expect(0, 63743, '\P{Is_Blk=Private_Use_area}', "");
Expect(1, 63743, '\P{^Is_Blk=Private_Use_area}', "");
Expect(0, 63744, '\p{Is_Blk=Private_Use_area}', "");
Expect(1, 63744, '\p{^Is_Blk=Private_Use_area}', "");
Expect(1, 63744, '\P{Is_Blk=Private_Use_area}', "");
Expect(0, 63744, '\P{^Is_Blk=Private_Use_area}', "");
Error('\p{Block=_/a/Dingbats}');
Error('\P{Block=_/a/Dingbats}');
Expect(1, 10175, '\p{Block=dingbats}', "");
Expect(0, 10175, '\p{^Block=dingbats}', "");
Expect(0, 10175, '\P{Block=dingbats}', "");
Expect(1, 10175, '\P{^Block=dingbats}', "");
Expect(0, 10176, '\p{Block=dingbats}', "");
Expect(1, 10176, '\p{^Block=dingbats}', "");
Expect(1, 10176, '\P{Block=dingbats}', "");
Expect(0, 10176, '\P{^Block=dingbats}', "");
Expect(1, 10175, '\p{Block=_Dingbats}', "");
Expect(0, 10175, '\p{^Block=_Dingbats}', "");
Expect(0, 10175, '\P{Block=_Dingbats}', "");
Expect(1, 10175, '\P{^Block=_Dingbats}', "");
Expect(0, 10176, '\p{Block=_Dingbats}', "");
Expect(1, 10176, '\p{^Block=_Dingbats}', "");
Expect(1, 10176, '\P{Block=_Dingbats}', "");
Expect(0, 10176, '\P{^Block=_Dingbats}', "");
Error('\p{Blk=/a/		Dingbats}');
Error('\P{Blk=/a/		Dingbats}');
Expect(1, 10175, '\p{Blk=dingbats}', "");
Expect(0, 10175, '\p{^Blk=dingbats}', "");
Expect(0, 10175, '\P{Blk=dingbats}', "");
Expect(1, 10175, '\P{^Blk=dingbats}', "");
Expect(0, 10176, '\p{Blk=dingbats}', "");
Expect(1, 10176, '\p{^Blk=dingbats}', "");
Expect(1, 10176, '\P{Blk=dingbats}', "");
Expect(0, 10176, '\P{^Blk=dingbats}', "");
Expect(1, 10175, '\p{Blk=		DINGBATS}', "");
Expect(0, 10175, '\p{^Blk=		DINGBATS}', "");
Expect(0, 10175, '\P{Blk=		DINGBATS}', "");
Expect(1, 10175, '\P{^Blk=		DINGBATS}', "");
Expect(0, 10176, '\p{Blk=		DINGBATS}', "");
Expect(1, 10176, '\p{^Blk=		DINGBATS}', "");
Expect(1, 10176, '\P{Blk=		DINGBATS}', "");
Expect(0, 10176, '\P{^Blk=		DINGBATS}', "");
Error('\p{Is_Block= Dingbats/a/}');
Error('\P{Is_Block= Dingbats/a/}');
Expect(1, 10175, '\p{Is_Block:dingbats}', "");
Expect(0, 10175, '\p{^Is_Block:dingbats}', "");
Expect(0, 10175, '\P{Is_Block:dingbats}', "");
Expect(1, 10175, '\P{^Is_Block:dingbats}', "");
Expect(0, 10176, '\p{Is_Block:dingbats}', "");
Expect(1, 10176, '\p{^Is_Block:dingbats}', "");
Expect(1, 10176, '\P{Is_Block:dingbats}', "");
Expect(0, 10176, '\P{^Is_Block:dingbats}', "");
Expect(1, 10175, '\p{Is_Block=	-dingbats}', "");
Expect(0, 10175, '\p{^Is_Block=	-dingbats}', "");
Expect(0, 10175, '\P{Is_Block=	-dingbats}', "");
Expect(1, 10175, '\P{^Is_Block=	-dingbats}', "");
Expect(0, 10176, '\p{Is_Block=	-dingbats}', "");
Expect(1, 10176, '\p{^Is_Block=	-dingbats}', "");
Expect(1, 10176, '\P{Is_Block=	-dingbats}', "");
Expect(0, 10176, '\P{^Is_Block=	-dingbats}', "");
Error('\p{Is_Blk=/a/Dingbats}');
Error('\P{Is_Blk=/a/Dingbats}');
Expect(1, 10175, '\p{Is_Blk=dingbats}', "");
Expect(0, 10175, '\p{^Is_Blk=dingbats}', "");
Expect(0, 10175, '\P{Is_Blk=dingbats}', "");
Expect(1, 10175, '\P{^Is_Blk=dingbats}', "");
Expect(0, 10176, '\p{Is_Blk=dingbats}', "");
Expect(1, 10176, '\p{^Is_Blk=dingbats}', "");
Expect(1, 10176, '\P{Is_Blk=dingbats}', "");
Expect(0, 10176, '\P{^Is_Blk=dingbats}', "");
Expect(1, 10175, '\p{Is_Blk:  dingbats}', "");
Expect(0, 10175, '\p{^Is_Blk:  dingbats}', "");
Expect(0, 10175, '\P{Is_Blk:  dingbats}', "");
Expect(1, 10175, '\P{^Is_Blk:  dingbats}', "");
Expect(0, 10176, '\p{Is_Blk:  dingbats}', "");
Expect(1, 10176, '\p{^Is_Blk:  dingbats}', "");
Expect(1, 10176, '\P{Is_Blk:  dingbats}', "");
Expect(0, 10176, '\P{^Is_Blk:  dingbats}', "");
Error('\p{Block= -Bamum/a/}');
Error('\P{Block= -Bamum/a/}');
Expect(1, 42751, '\p{Block=bamum}', "");
Expect(0, 42751, '\p{^Block=bamum}', "");
Expect(0, 42751, '\P{Block=bamum}', "");
Expect(1, 42751, '\P{^Block=bamum}', "");
Expect(0, 42752, '\p{Block=bamum}', "");
Expect(1, 42752, '\p{^Block=bamum}', "");
Expect(1, 42752, '\P{Block=bamum}', "");
Expect(0, 42752, '\P{^Block=bamum}', "");
Expect(1, 42751, '\p{Block=-_Bamum}', "");
Expect(0, 42751, '\p{^Block=-_Bamum}', "");
Expect(0, 42751, '\P{Block=-_Bamum}', "");
Expect(1, 42751, '\P{^Block=-_Bamum}', "");
Expect(0, 42752, '\p{Block=-_Bamum}', "");
Expect(1, 42752, '\p{^Block=-_Bamum}', "");
Expect(1, 42752, '\P{Block=-_Bamum}', "");
Expect(0, 42752, '\P{^Block=-_Bamum}', "");
Error('\p{Blk=--Bamum:=}');
Error('\P{Blk=--Bamum:=}');
Expect(1, 42751, '\p{Blk:   bamum}', "");
Expect(0, 42751, '\p{^Blk:   bamum}', "");
Expect(0, 42751, '\P{Blk:   bamum}', "");
Expect(1, 42751, '\P{^Blk:   bamum}', "");
Expect(0, 42752, '\p{Blk:   bamum}', "");
Expect(1, 42752, '\p{^Blk:   bamum}', "");
Expect(1, 42752, '\P{Blk:   bamum}', "");
Expect(0, 42752, '\P{^Blk:   bamum}', "");
Expect(1, 42751, '\p{Blk=	_Bamum}', "");
Expect(0, 42751, '\p{^Blk=	_Bamum}', "");
Expect(0, 42751, '\P{Blk=	_Bamum}', "");
Expect(1, 42751, '\P{^Blk=	_Bamum}', "");
Expect(0, 42752, '\p{Blk=	_Bamum}', "");
Expect(1, 42752, '\p{^Blk=	_Bamum}', "");
Expect(1, 42752, '\P{Blk=	_Bamum}', "");
Expect(0, 42752, '\P{^Blk=	_Bamum}', "");
Error('\p{Is_Block=	 Bamum:=}');
Error('\P{Is_Block=	 Bamum:=}');
Expect(1, 42751, '\p{Is_Block:bamum}', "");
Expect(0, 42751, '\p{^Is_Block:bamum}', "");
Expect(0, 42751, '\P{Is_Block:bamum}', "");
Expect(1, 42751, '\P{^Is_Block:bamum}', "");
Expect(0, 42752, '\p{Is_Block:bamum}', "");
Expect(1, 42752, '\p{^Is_Block:bamum}', "");
Expect(1, 42752, '\P{Is_Block:bamum}', "");
Expect(0, 42752, '\P{^Is_Block:bamum}', "");
Expect(1, 42751, '\p{Is_Block=_	Bamum}', "");
Expect(0, 42751, '\p{^Is_Block=_	Bamum}', "");
Expect(0, 42751, '\P{Is_Block=_	Bamum}', "");
Expect(1, 42751, '\P{^Is_Block=_	Bamum}', "");
Expect(0, 42752, '\p{Is_Block=_	Bamum}', "");
Expect(1, 42752, '\p{^Is_Block=_	Bamum}', "");
Expect(1, 42752, '\P{Is_Block=_	Bamum}', "");
Expect(0, 42752, '\P{^Is_Block=_	Bamum}', "");
Error('\p{Is_Blk=_/a/Bamum}');
Error('\P{Is_Blk=_/a/Bamum}');
Expect(1, 42751, '\p{Is_Blk=bamum}', "");
Expect(0, 42751, '\p{^Is_Blk=bamum}', "");
Expect(0, 42751, '\P{Is_Blk=bamum}', "");
Expect(1, 42751, '\P{^Is_Blk=bamum}', "");
Expect(0, 42752, '\p{Is_Blk=bamum}', "");
Expect(1, 42752, '\p{^Is_Blk=bamum}', "");
Expect(1, 42752, '\P{Is_Blk=bamum}', "");
Expect(0, 42752, '\P{^Is_Blk=bamum}', "");
Expect(1, 42751, '\p{Is_Blk=__bamum}', "");
Expect(0, 42751, '\p{^Is_Blk=__bamum}', "");
Expect(0, 42751, '\P{Is_Blk=__bamum}', "");
Expect(1, 42751, '\P{^Is_Blk=__bamum}', "");
Expect(0, 42752, '\p{Is_Blk=__bamum}', "");
Expect(1, 42752, '\p{^Is_Blk=__bamum}', "");
Expect(1, 42752, '\P{Is_Blk=__bamum}', "");
Expect(0, 42752, '\P{^Is_Blk=__bamum}', "");
Error('\p{Block: :=INSCRIPTIONAL_Parthian}');
Error('\P{Block: :=INSCRIPTIONAL_Parthian}');
Expect(1, 68447, '\p{Block=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Block=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Block=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Block=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Block=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Block=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Block=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Block=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Block:	 -Inscriptional_Parthian}', "");
Expect(0, 68447, '\p{^Block:	 -Inscriptional_Parthian}', "");
Expect(0, 68447, '\P{Block:	 -Inscriptional_Parthian}', "");
Expect(1, 68447, '\P{^Block:	 -Inscriptional_Parthian}', "");
Expect(0, 68448, '\p{Block:	 -Inscriptional_Parthian}', "");
Expect(1, 68448, '\p{^Block:	 -Inscriptional_Parthian}', "");
Expect(1, 68448, '\P{Block:	 -Inscriptional_Parthian}', "");
Expect(0, 68448, '\P{^Block:	 -Inscriptional_Parthian}', "");
Error('\p{Blk=:=-	Inscriptional_Parthian}');
Error('\P{Blk=:=-	Inscriptional_Parthian}');
Expect(1, 68447, '\p{Blk=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Blk=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Blk=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Blk=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Blk=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Blk=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Blk=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Blk=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Blk=--INSCRIPTIONAL_PARTHIAN}', "");
Expect(0, 68447, '\p{^Blk=--INSCRIPTIONAL_PARTHIAN}', "");
Expect(0, 68447, '\P{Blk=--INSCRIPTIONAL_PARTHIAN}', "");
Expect(1, 68447, '\P{^Blk=--INSCRIPTIONAL_PARTHIAN}', "");
Expect(0, 68448, '\p{Blk=--INSCRIPTIONAL_PARTHIAN}', "");
Expect(1, 68448, '\p{^Blk=--INSCRIPTIONAL_PARTHIAN}', "");
Expect(1, 68448, '\P{Blk=--INSCRIPTIONAL_PARTHIAN}', "");
Expect(0, 68448, '\P{^Blk=--INSCRIPTIONAL_PARTHIAN}', "");
Error('\p{Is_Block=/a/ Inscriptional_parthian}');
Error('\P{Is_Block=/a/ Inscriptional_parthian}');
Expect(1, 68447, '\p{Is_Block=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Is_Block=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Is_Block=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Is_Block=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Is_Block=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Is_Block=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Is_Block=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Is_Block=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Is_Block= -Inscriptional_Parthian}', "");
Expect(0, 68447, '\p{^Is_Block= -Inscriptional_Parthian}', "");
Expect(0, 68447, '\P{Is_Block= -Inscriptional_Parthian}', "");
Expect(1, 68447, '\P{^Is_Block= -Inscriptional_Parthian}', "");
Expect(0, 68448, '\p{Is_Block= -Inscriptional_Parthian}', "");
Expect(1, 68448, '\p{^Is_Block= -Inscriptional_Parthian}', "");
Expect(1, 68448, '\P{Is_Block= -Inscriptional_Parthian}', "");
Expect(0, 68448, '\P{^Is_Block= -Inscriptional_Parthian}', "");
Error('\p{Is_Blk= 	inscriptional_Parthian/a/}');
Error('\P{Is_Blk= 	inscriptional_Parthian/a/}');
Expect(1, 68447, '\p{Is_Blk=inscriptionalparthian}', "");
Expect(0, 68447, '\p{^Is_Blk=inscriptionalparthian}', "");
Expect(0, 68447, '\P{Is_Blk=inscriptionalparthian}', "");
Expect(1, 68447, '\P{^Is_Blk=inscriptionalparthian}', "");
Expect(0, 68448, '\p{Is_Blk=inscriptionalparthian}', "");
Expect(1, 68448, '\p{^Is_Blk=inscriptionalparthian}', "");
Expect(1, 68448, '\P{Is_Blk=inscriptionalparthian}', "");
Expect(0, 68448, '\P{^Is_Blk=inscriptionalparthian}', "");
Expect(1, 68447, '\p{Is_Blk=	_INSCRIPTIONAL_parthian}', "");
Expect(0, 68447, '\p{^Is_Blk=	_INSCRIPTIONAL_parthian}', "");
Expect(0, 68447, '\P{Is_Blk=	_INSCRIPTIONAL_parthian}', "");
Expect(1, 68447, '\P{^Is_Blk=	_INSCRIPTIONAL_parthian}', "");
Expect(0, 68448, '\p{Is_Blk=	_INSCRIPTIONAL_parthian}', "");
Expect(1, 68448, '\p{^Is_Blk=	_INSCRIPTIONAL_parthian}', "");
Expect(1, 68448, '\P{Is_Blk=	_INSCRIPTIONAL_parthian}', "");
Expect(0, 68448, '\P{^Is_Blk=	_INSCRIPTIONAL_parthian}', "");
Error('\p{Block=:= -Unified_CANADIAN_Aboriginal_Syllabics}');
Error('\P{Block=:= -Unified_CANADIAN_Aboriginal_Syllabics}');
Expect(1, 5759, '\p{Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\p{^Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\P{Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\P{^Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\p{Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\p{^Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\P{Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\P{^Block=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\p{Block: _UNIFIED_Canadian_ABORIGINAL_Syllabics}', "");
Expect(0, 5759, '\p{^Block: _UNIFIED_Canadian_ABORIGINAL_Syllabics}', "");
Expect(0, 5759, '\P{Block: _UNIFIED_Canadian_ABORIGINAL_Syllabics}', "");
Expect(1, 5759, '\P{^Block: _UNIFIED_Canadian_ABORIGINAL_Syllabics}', "");
Expect(0, 5760, '\p{Block: _UNIFIED_Canadian_ABORIGINAL_Syllabics}', "");
Expect(1, 5760, '\p{^Block: _UNIFIED_Canadian_ABORIGINAL_Syllabics}', "");
Expect(1, 5760, '\P{Block: _UNIFIED_Canadian_ABORIGINAL_Syllabics}', "");
Expect(0, 5760, '\P{^Block: _UNIFIED_Canadian_ABORIGINAL_Syllabics}', "");
Error('\p{Blk=/a/ucas}');
Error('\P{Blk=/a/ucas}');
Expect(1, 5759, '\p{Blk=ucas}', "");
Expect(0, 5759, '\p{^Blk=ucas}', "");
Expect(0, 5759, '\P{Blk=ucas}', "");
Expect(1, 5759, '\P{^Blk=ucas}', "");
Expect(0, 5760, '\p{Blk=ucas}', "");
Expect(1, 5760, '\p{^Blk=ucas}', "");
Expect(1, 5760, '\P{Blk=ucas}', "");
Expect(0, 5760, '\P{^Blk=ucas}', "");
Expect(1, 5759, '\p{Blk:- UCAS}', "");
Expect(0, 5759, '\p{^Blk:- UCAS}', "");
Expect(0, 5759, '\P{Blk:- UCAS}', "");
Expect(1, 5759, '\P{^Blk:- UCAS}', "");
Expect(0, 5760, '\p{Blk:- UCAS}', "");
Expect(1, 5760, '\p{^Blk:- UCAS}', "");
Expect(1, 5760, '\P{Blk:- UCAS}', "");
Expect(0, 5760, '\P{^Blk:- UCAS}', "");
Error('\p{Is_Block=- canadian_syllabics:=}');
Error('\P{Is_Block=- canadian_syllabics:=}');
Expect(1, 5759, '\p{Is_Block=canadiansyllabics}', "");
Expect(0, 5759, '\p{^Is_Block=canadiansyllabics}', "");
Expect(0, 5759, '\P{Is_Block=canadiansyllabics}', "");
Expect(1, 5759, '\P{^Is_Block=canadiansyllabics}', "");
Expect(0, 5760, '\p{Is_Block=canadiansyllabics}', "");
Expect(1, 5760, '\p{^Is_Block=canadiansyllabics}', "");
Expect(1, 5760, '\P{Is_Block=canadiansyllabics}', "");
Expect(0, 5760, '\P{^Is_Block=canadiansyllabics}', "");
Expect(1, 5759, '\p{Is_Block=-Canadian_Syllabics}', "");
Expect(0, 5759, '\p{^Is_Block=-Canadian_Syllabics}', "");
Expect(0, 5759, '\P{Is_Block=-Canadian_Syllabics}', "");
Expect(1, 5759, '\P{^Is_Block=-Canadian_Syllabics}', "");
Expect(0, 5760, '\p{Is_Block=-Canadian_Syllabics}', "");
Expect(1, 5760, '\p{^Is_Block=-Canadian_Syllabics}', "");
Expect(1, 5760, '\P{Is_Block=-Canadian_Syllabics}', "");
Expect(0, 5760, '\P{^Is_Block=-Canadian_Syllabics}', "");
Error('\p{Is_Blk=	:=unified_Canadian_Aboriginal_Syllabics}');
Error('\P{Is_Blk=	:=unified_Canadian_Aboriginal_Syllabics}');
Expect(1, 5759, '\p{Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\p{^Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5759, '\P{Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\P{^Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\p{Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\p{^Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5760, '\P{Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(0, 5760, '\P{^Is_Blk=unifiedcanadianaboriginalsyllabics}', "");
Expect(1, 5759, '\p{Is_Blk=unified_Canadian_ABORIGINAL_syllabics}', "");
Expect(0, 5759, '\p{^Is_Blk=unified_Canadian_ABORIGINAL_syllabics}', "");
Expect(0, 5759, '\P{Is_Blk=unified_Canadian_ABORIGINAL_syllabics}', "");
Expect(1, 5759, '\P{^Is_Blk=unified_Canadian_ABORIGINAL_syllabics}', "");
Expect(0, 5760, '\p{Is_Blk=unified_Canadian_ABORIGINAL_syllabics}', "");
Expect(1, 5760, '\p{^Is_Blk=unified_Canadian_ABORIGINAL_syllabics}', "");
Expect(1, 5760, '\P{Is_Blk=unified_Canadian_ABORIGINAL_syllabics}', "");
Expect(0, 5760, '\P{^Is_Blk=unified_Canadian_ABORIGINAL_syllabics}', "");
Error('\p{Block=__CONTROL_Pictures/a/}');
Error('\P{Block=__CONTROL_Pictures/a/}');
Expect(1, 9279, '\p{Block=controlpictures}', "");
Expect(0, 9279, '\p{^Block=controlpictures}', "");
Expect(0, 9279, '\P{Block=controlpictures}', "");
Expect(1, 9279, '\P{^Block=controlpictures}', "");
Expect(0, 9280, '\p{Block=controlpictures}', "");
Expect(1, 9280, '\p{^Block=controlpictures}', "");
Expect(1, 9280, '\P{Block=controlpictures}', "");
Expect(0, 9280, '\P{^Block=controlpictures}', "");
Expect(1, 9279, '\p{Block=  Control_pictures}', "");
Expect(0, 9279, '\p{^Block=  Control_pictures}', "");
Expect(0, 9279, '\P{Block=  Control_pictures}', "");
Expect(1, 9279, '\P{^Block=  Control_pictures}', "");
Expect(0, 9280, '\p{Block=  Control_pictures}', "");
Expect(1, 9280, '\p{^Block=  Control_pictures}', "");
Expect(1, 9280, '\P{Block=  Control_pictures}', "");
Expect(0, 9280, '\P{^Block=  Control_pictures}', "");
Error('\p{Blk=- CONTROL_PICTURES/a/}');
Error('\P{Blk=- CONTROL_PICTURES/a/}');
Expect(1, 9279, '\p{Blk:	controlpictures}', "");
Expect(0, 9279, '\p{^Blk:	controlpictures}', "");
Expect(0, 9279, '\P{Blk:	controlpictures}', "");
Expect(1, 9279, '\P{^Blk:	controlpictures}', "");
Expect(0, 9280, '\p{Blk:	controlpictures}', "");
Expect(1, 9280, '\p{^Blk:	controlpictures}', "");
Expect(1, 9280, '\P{Blk:	controlpictures}', "");
Expect(0, 9280, '\P{^Blk:	controlpictures}', "");
Expect(1, 9279, '\p{Blk=	Control_Pictures}', "");
Expect(0, 9279, '\p{^Blk=	Control_Pictures}', "");
Expect(0, 9279, '\P{Blk=	Control_Pictures}', "");
Expect(1, 9279, '\P{^Blk=	Control_Pictures}', "");
Expect(0, 9280, '\p{Blk=	Control_Pictures}', "");
Expect(1, 9280, '\p{^Blk=	Control_Pictures}', "");
Expect(1, 9280, '\P{Blk=	Control_Pictures}', "");
Expect(0, 9280, '\P{^Blk=	Control_Pictures}', "");
Error('\p{Is_Block=_	control_pictures/a/}');
Error('\P{Is_Block=_	control_pictures/a/}');
Expect(1, 9279, '\p{Is_Block=controlpictures}', "");
Expect(0, 9279, '\p{^Is_Block=controlpictures}', "");
Expect(0, 9279, '\P{Is_Block=controlpictures}', "");
Expect(1, 9279, '\P{^Is_Block=controlpictures}', "");
Expect(0, 9280, '\p{Is_Block=controlpictures}', "");
Expect(1, 9280, '\p{^Is_Block=controlpictures}', "");
Expect(1, 9280, '\P{Is_Block=controlpictures}', "");
Expect(0, 9280, '\P{^Is_Block=controlpictures}', "");
Expect(1, 9279, '\p{Is_Block=__Control_pictures}', "");
Expect(0, 9279, '\p{^Is_Block=__Control_pictures}', "");
Expect(0, 9279, '\P{Is_Block=__Control_pictures}', "");
Expect(1, 9279, '\P{^Is_Block=__Control_pictures}', "");
Expect(0, 9280, '\p{Is_Block=__Control_pictures}', "");
Expect(1, 9280, '\p{^Is_Block=__Control_pictures}', "");
Expect(1, 9280, '\P{Is_Block=__Control_pictures}', "");
Expect(0, 9280, '\P{^Is_Block=__Control_pictures}', "");
Error('\p{Is_Blk=	/a/CONTROL_pictures}');
Error('\P{Is_Blk=	/a/CONTROL_pictures}');
Expect(1, 9279, '\p{Is_Blk=controlpictures}', "");
Expect(0, 9279, '\p{^Is_Blk=controlpictures}', "");
Expect(0, 9279, '\P{Is_Blk=controlpictures}', "");
Expect(1, 9279, '\P{^Is_Blk=controlpictures}', "");
Expect(0, 9280, '\p{Is_Blk=controlpictures}', "");
Expect(1, 9280, '\p{^Is_Blk=controlpictures}', "");
Expect(1, 9280, '\P{Is_Blk=controlpictures}', "");
Expect(0, 9280, '\P{^Is_Blk=controlpictures}', "");
Expect(1, 9279, '\p{Is_Blk= 	CONTROL_PICTURES}', "");
Expect(0, 9279, '\p{^Is_Blk= 	CONTROL_PICTURES}', "");
Expect(0, 9279, '\P{Is_Blk= 	CONTROL_PICTURES}', "");
Expect(1, 9279, '\P{^Is_Blk= 	CONTROL_PICTURES}', "");
Expect(0, 9280, '\p{Is_Blk= 	CONTROL_PICTURES}', "");
Expect(1, 9280, '\p{^Is_Blk= 	CONTROL_PICTURES}', "");
Expect(1, 9280, '\P{Is_Blk= 	CONTROL_PICTURES}', "");
Expect(0, 9280, '\P{^Is_Blk= 	CONTROL_PICTURES}', "");
Error('\p{Block=-:=CJK_Symbols_and_Punctuation}');
Error('\P{Block=-:=CJK_Symbols_and_Punctuation}');
Expect(1, 12351, '\p{Block=cjksymbolsandpunctuation}', "");
Expect(0, 12351, '\p{^Block=cjksymbolsandpunctuation}', "");
Expect(0, 12351, '\P{Block=cjksymbolsandpunctuation}', "");
Expect(1, 12351, '\P{^Block=cjksymbolsandpunctuation}', "");
Expect(0, 12352, '\p{Block=cjksymbolsandpunctuation}', "");
Expect(1, 12352, '\p{^Block=cjksymbolsandpunctuation}', "");
Expect(1, 12352, '\P{Block=cjksymbolsandpunctuation}', "");
Expect(0, 12352, '\P{^Block=cjksymbolsandpunctuation}', "");
Expect(1, 12351, '\p{Block=__cjk_Symbols_And_PUNCTUATION}', "");
Expect(0, 12351, '\p{^Block=__cjk_Symbols_And_PUNCTUATION}', "");
Expect(0, 12351, '\P{Block=__cjk_Symbols_And_PUNCTUATION}', "");
Expect(1, 12351, '\P{^Block=__cjk_Symbols_And_PUNCTUATION}', "");
Expect(0, 12352, '\p{Block=__cjk_Symbols_And_PUNCTUATION}', "");
Expect(1, 12352, '\p{^Block=__cjk_Symbols_And_PUNCTUATION}', "");
Expect(1, 12352, '\P{Block=__cjk_Symbols_And_PUNCTUATION}', "");
Expect(0, 12352, '\P{^Block=__cjk_Symbols_And_PUNCTUATION}', "");
Error('\p{Blk=/a/_ CJK_Symbols}');
Error('\P{Blk=/a/_ CJK_Symbols}');
Expect(1, 12351, '\p{Blk=cjksymbols}', "");
Expect(0, 12351, '\p{^Blk=cjksymbols}', "");
Expect(0, 12351, '\P{Blk=cjksymbols}', "");
Expect(1, 12351, '\P{^Blk=cjksymbols}', "");
Expect(0, 12352, '\p{Blk=cjksymbols}', "");
Expect(1, 12352, '\p{^Blk=cjksymbols}', "");
Expect(1, 12352, '\P{Blk=cjksymbols}', "");
Expect(0, 12352, '\P{^Blk=cjksymbols}', "");
Expect(1, 12351, '\p{Blk=-_CJK_symbols}', "");
Expect(0, 12351, '\p{^Blk=-_CJK_symbols}', "");
Expect(0, 12351, '\P{Blk=-_CJK_symbols}', "");
Expect(1, 12351, '\P{^Blk=-_CJK_symbols}', "");
Expect(0, 12352, '\p{Blk=-_CJK_symbols}', "");
Expect(1, 12352, '\p{^Blk=-_CJK_symbols}', "");
Expect(1, 12352, '\P{Blk=-_CJK_symbols}', "");
Expect(0, 12352, '\P{^Blk=-_CJK_symbols}', "");
Error('\p{Is_Block=	/a/CJK_SYMBOLS_And_Punctuation}');
Error('\P{Is_Block=	/a/CJK_SYMBOLS_And_Punctuation}');
Expect(1, 12351, '\p{Is_Block=cjksymbolsandpunctuation}', "");
Expect(0, 12351, '\p{^Is_Block=cjksymbolsandpunctuation}', "");
Expect(0, 12351, '\P{Is_Block=cjksymbolsandpunctuation}', "");
Expect(1, 12351, '\P{^Is_Block=cjksymbolsandpunctuation}', "");
Expect(0, 12352, '\p{Is_Block=cjksymbolsandpunctuation}', "");
Expect(1, 12352, '\p{^Is_Block=cjksymbolsandpunctuation}', "");
Expect(1, 12352, '\P{Is_Block=cjksymbolsandpunctuation}', "");
Expect(0, 12352, '\P{^Is_Block=cjksymbolsandpunctuation}', "");
Expect(1, 12351, '\p{Is_Block=	CJK_symbols_And_punctuation}', "");
Expect(0, 12351, '\p{^Is_Block=	CJK_symbols_And_punctuation}', "");
Expect(0, 12351, '\P{Is_Block=	CJK_symbols_And_punctuation}', "");
Expect(1, 12351, '\P{^Is_Block=	CJK_symbols_And_punctuation}', "");
Expect(0, 12352, '\p{Is_Block=	CJK_symbols_And_punctuation}', "");
Expect(1, 12352, '\p{^Is_Block=	CJK_symbols_And_punctuation}', "");
Expect(1, 12352, '\P{Is_Block=	CJK_symbols_And_punctuation}', "");
Expect(0, 12352, '\P{^Is_Block=	CJK_symbols_And_punctuation}', "");
Error('\p{Is_Blk=/a/	CJK_SYMBOLS}');
Error('\P{Is_Blk=/a/	CJK_SYMBOLS}');
Expect(1, 12351, '\p{Is_Blk=cjksymbols}', "");
Expect(0, 12351, '\p{^Is_Blk=cjksymbols}', "");
Expect(0, 12351, '\P{Is_Blk=cjksymbols}', "");
Expect(1, 12351, '\P{^Is_Blk=cjksymbols}', "");
Expect(0, 12352, '\p{Is_Blk=cjksymbols}', "");
Expect(1, 12352, '\p{^Is_Blk=cjksymbols}', "");
Expect(1, 12352, '\P{Is_Blk=cjksymbols}', "");
Expect(0, 12352, '\P{^Is_Blk=cjksymbols}', "");
Expect(1, 12351, '\p{Is_Blk= 	CJK_Symbols}', "");
Expect(0, 12351, '\p{^Is_Blk= 	CJK_Symbols}', "");
Expect(0, 12351, '\P{Is_Blk= 	CJK_Symbols}', "");
Expect(1, 12351, '\P{^Is_Blk= 	CJK_Symbols}', "");
Expect(0, 12352, '\p{Is_Blk= 	CJK_Symbols}', "");
Expect(1, 12352, '\p{^Is_Blk= 	CJK_Symbols}', "");
Expect(1, 12352, '\P{Is_Blk= 	CJK_Symbols}', "");
Expect(0, 12352, '\P{^Is_Blk= 	CJK_Symbols}', "");
Error('\p{Block= -cjk_Unified_Ideographs_Extension_B/a/}');
Error('\P{Block= -cjk_Unified_Ideographs_Extension_B/a/}');
Expect(1, 173791, '\p{Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\p{^Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\P{Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\P{^Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\p{Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\p{^Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\P{Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\P{^Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\p{Block= _cjk_Unified_Ideographs_Extension_B}', "");
Expect(0, 173791, '\p{^Block= _cjk_Unified_Ideographs_Extension_B}', "");
Expect(0, 173791, '\P{Block= _cjk_Unified_Ideographs_Extension_B}', "");
Expect(1, 173791, '\P{^Block= _cjk_Unified_Ideographs_Extension_B}', "");
Expect(0, 173792, '\p{Block= _cjk_Unified_Ideographs_Extension_B}', "");
Expect(1, 173792, '\p{^Block= _cjk_Unified_Ideographs_Extension_B}', "");
Expect(1, 173792, '\P{Block= _cjk_Unified_Ideographs_Extension_B}', "");
Expect(0, 173792, '\P{^Block= _cjk_Unified_Ideographs_Extension_B}', "");
Error('\p{Blk=/a/cjk_Ext_b}');
Error('\P{Blk=/a/cjk_Ext_b}');
Expect(1, 173791, '\p{Blk=cjkextb}', "");
Expect(0, 173791, '\p{^Blk=cjkextb}', "");
Expect(0, 173791, '\P{Blk=cjkextb}', "");
Expect(1, 173791, '\P{^Blk=cjkextb}', "");
Expect(0, 173792, '\p{Blk=cjkextb}', "");
Expect(1, 173792, '\p{^Blk=cjkextb}', "");
Expect(1, 173792, '\P{Blk=cjkextb}', "");
Expect(0, 173792, '\P{^Blk=cjkextb}', "");
Expect(1, 173791, '\p{Blk=		CJK_Ext_B}', "");
Expect(0, 173791, '\p{^Blk=		CJK_Ext_B}', "");
Expect(0, 173791, '\P{Blk=		CJK_Ext_B}', "");
Expect(1, 173791, '\P{^Blk=		CJK_Ext_B}', "");
Expect(0, 173792, '\p{Blk=		CJK_Ext_B}', "");
Expect(1, 173792, '\p{^Blk=		CJK_Ext_B}', "");
Expect(1, 173792, '\P{Blk=		CJK_Ext_B}', "");
Expect(0, 173792, '\P{^Blk=		CJK_Ext_B}', "");
Error('\p{Is_Block=	/a/CJK_unified_IDEOGRAPHS_EXTENSION_B}');
Error('\P{Is_Block=	/a/CJK_unified_IDEOGRAPHS_EXTENSION_B}');
Expect(1, 173791, '\p{Is_Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\p{^Is_Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173791, '\P{Is_Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\P{^Is_Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\p{Is_Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\p{^Is_Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173792, '\P{Is_Block=cjkunifiedideographsextensionb}', "");
Expect(0, 173792, '\P{^Is_Block=cjkunifiedideographsextensionb}', "");
Expect(1, 173791, '\p{Is_Block=  CJK_UNIFIED_Ideographs_extension_B}', "");
Expect(0, 173791, '\p{^Is_Block=  CJK_UNIFIED_Ideographs_extension_B}', "");
Expect(0, 173791, '\P{Is_Block=  CJK_UNIFIED_Ideographs_extension_B}', "");
Expect(1, 173791, '\P{^Is_Block=  CJK_UNIFIED_Ideographs_extension_B}', "");
Expect(0, 173792, '\p{Is_Block=  CJK_UNIFIED_Ideographs_extension_B}', "");
Expect(1, 173792, '\p{^Is_Block=  CJK_UNIFIED_Ideographs_extension_B}', "");
Expect(1, 173792, '\P{Is_Block=  CJK_UNIFIED_Ideographs_extension_B}', "");
Expect(0, 173792, '\P{^Is_Block=  CJK_UNIFIED_Ideographs_extension_B}', "");
Error('\p{Is_Blk=_CJK_Ext_B:=}');
Error('\P{Is_Blk=_CJK_Ext_B:=}');
Expect(1, 173791, '\p{Is_Blk=cjkextb}', "");
Expect(0, 173791, '\p{^Is_Blk=cjkextb}', "");
Expect(0, 173791, '\P{Is_Blk=cjkextb}', "");
Expect(1, 173791, '\P{^Is_Blk=cjkextb}', "");
Expect(0, 173792, '\p{Is_Blk=cjkextb}', "");
Expect(1, 173792, '\p{^Is_Blk=cjkextb}', "");
Expect(1, 173792, '\P{Is_Blk=cjkextb}', "");
Expect(0, 173792, '\P{^Is_Blk=cjkextb}', "");
Expect(1, 173791, '\p{Is_Blk:  CJK_Ext_b}', "");
Expect(0, 173791, '\p{^Is_Blk:  CJK_Ext_b}', "");
Expect(0, 173791, '\P{Is_Blk:  CJK_Ext_b}', "");
Expect(1, 173791, '\P{^Is_Blk:  CJK_Ext_b}', "");
Expect(0, 173792, '\p{Is_Blk:  CJK_Ext_b}', "");
Expect(1, 173792, '\p{^Is_Blk:  CJK_Ext_b}', "");
Expect(1, 173792, '\P{Is_Blk:  CJK_Ext_b}', "");
Expect(0, 173792, '\P{^Is_Blk:  CJK_Ext_b}', "");
Error('\p{Block=	:=ARMENIAN}');
Error('\P{Block=	:=ARMENIAN}');
Expect(1, 1423, '\p{Block:armenian}', "");
Expect(0, 1423, '\p{^Block:armenian}', "");
Expect(0, 1423, '\P{Block:armenian}', "");
Expect(1, 1423, '\P{^Block:armenian}', "");
Expect(0, 1424, '\p{Block:armenian}', "");
Expect(1, 1424, '\p{^Block:armenian}', "");
Expect(1, 1424, '\P{Block:armenian}', "");
Expect(0, 1424, '\P{^Block:armenian}', "");
Expect(1, 1423, '\p{Block=- Armenian}', "");
Expect(0, 1423, '\p{^Block=- Armenian}', "");
Expect(0, 1423, '\P{Block=- Armenian}', "");
Expect(1, 1423, '\P{^Block=- Armenian}', "");
Expect(0, 1424, '\p{Block=- Armenian}', "");
Expect(1, 1424, '\p{^Block=- Armenian}', "");
Expect(1, 1424, '\P{Block=- Armenian}', "");
Expect(0, 1424, '\P{^Block=- Armenian}', "");
Error('\p{Blk=_	ARMENIAN/a/}');
Error('\P{Blk=_	ARMENIAN/a/}');
Expect(1, 1423, '\p{Blk=armenian}', "");
Expect(0, 1423, '\p{^Blk=armenian}', "");
Expect(0, 1423, '\P{Blk=armenian}', "");
Expect(1, 1423, '\P{^Blk=armenian}', "");
Expect(0, 1424, '\p{Blk=armenian}', "");
Expect(1, 1424, '\p{^Blk=armenian}', "");
Expect(1, 1424, '\P{Blk=armenian}', "");
Expect(0, 1424, '\P{^Blk=armenian}', "");
Expect(1, 1423, '\p{Blk=--armenian}', "");
Expect(0, 1423, '\p{^Blk=--armenian}', "");
Expect(0, 1423, '\P{Blk=--armenian}', "");
Expect(1, 1423, '\P{^Blk=--armenian}', "");
Expect(0, 1424, '\p{Blk=--armenian}', "");
Expect(1, 1424, '\p{^Blk=--armenian}', "");
Expect(1, 1424, '\P{Blk=--armenian}', "");
Expect(0, 1424, '\P{^Blk=--armenian}', "");
Error('\p{Is_Block=	:=armenian}');
Error('\P{Is_Block=	:=armenian}');
Expect(1, 1423, '\p{Is_Block=armenian}', "");
Expect(0, 1423, '\p{^Is_Block=armenian}', "");
Expect(0, 1423, '\P{Is_Block=armenian}', "");
Expect(1, 1423, '\P{^Is_Block=armenian}', "");
Expect(0, 1424, '\p{Is_Block=armenian}', "");
Expect(1, 1424, '\p{^Is_Block=armenian}', "");
Expect(1, 1424, '\P{Is_Block=armenian}', "");
Expect(0, 1424, '\P{^Is_Block=armenian}', "");
Expect(1, 1423, '\p{Is_Block:--Armenian}', "");
Expect(0, 1423, '\p{^Is_Block:--Armenian}', "");
Expect(0, 1423, '\P{Is_Block:--Armenian}', "");
Expect(1, 1423, '\P{^Is_Block:--Armenian}', "");
Expect(0, 1424, '\p{Is_Block:--Armenian}', "");
Expect(1, 1424, '\p{^Is_Block:--Armenian}', "");
Expect(1, 1424, '\P{Is_Block:--Armenian}', "");
Expect(0, 1424, '\P{^Is_Block:--Armenian}', "");
Error('\p{Is_Blk:   _	Armenian:=}');
Error('\P{Is_Blk:   _	Armenian:=}');
Expect(1, 1423, '\p{Is_Blk:	armenian}', "");
Expect(0, 1423, '\p{^Is_Blk:	armenian}', "");
Expect(0, 1423, '\P{Is_Blk:	armenian}', "");
Expect(1, 1423, '\P{^Is_Blk:	armenian}', "");
Expect(0, 1424, '\p{Is_Blk:	armenian}', "");
Expect(1, 1424, '\p{^Is_Blk:	armenian}', "");
Expect(1, 1424, '\P{Is_Blk:	armenian}', "");
Expect(0, 1424, '\P{^Is_Blk:	armenian}', "");
Expect(1, 1423, '\p{Is_Blk=_ARMENIAN}', "");
Expect(0, 1423, '\p{^Is_Blk=_ARMENIAN}', "");
Expect(0, 1423, '\P{Is_Blk=_ARMENIAN}', "");
Expect(1, 1423, '\P{^Is_Blk=_ARMENIAN}', "");
Expect(0, 1424, '\p{Is_Blk=_ARMENIAN}', "");
Expect(1, 1424, '\p{^Is_Blk=_ARMENIAN}', "");
Expect(1, 1424, '\P{Is_Blk=_ARMENIAN}', "");
Expect(0, 1424, '\P{^Is_Blk=_ARMENIAN}', "");
Error('\p{Block=	Ancient_greek_Musical_notation:=}');
Error('\P{Block=	Ancient_greek_Musical_notation:=}');
Expect(1, 119375, '\p{Block=ancientgreekmusicalnotation}', "");
Expect(0, 119375, '\p{^Block=ancientgreekmusicalnotation}', "");
Expect(0, 119375, '\P{Block=ancientgreekmusicalnotation}', "");
Expect(1, 119375, '\P{^Block=ancientgreekmusicalnotation}', "");
Expect(0, 119376, '\p{Block=ancientgreekmusicalnotation}', "");
Expect(1, 119376, '\p{^Block=ancientgreekmusicalnotation}', "");
Expect(1, 119376, '\P{Block=ancientgreekmusicalnotation}', "");
Expect(0, 119376, '\P{^Block=ancientgreekmusicalnotation}', "");
Expect(1, 119375, '\p{Block=	-ANCIENT_Greek_Musical_Notation}', "");
Expect(0, 119375, '\p{^Block=	-ANCIENT_Greek_Musical_Notation}', "");
Expect(0, 119375, '\P{Block=	-ANCIENT_Greek_Musical_Notation}', "");
Expect(1, 119375, '\P{^Block=	-ANCIENT_Greek_Musical_Notation}', "");
Expect(0, 119376, '\p{Block=	-ANCIENT_Greek_Musical_Notation}', "");
Expect(1, 119376, '\p{^Block=	-ANCIENT_Greek_Musical_Notation}', "");
Expect(1, 119376, '\P{Block=	-ANCIENT_Greek_Musical_Notation}', "");
Expect(0, 119376, '\P{^Block=	-ANCIENT_Greek_Musical_Notation}', "");
Error('\p{Blk=:=Ancient_GREEK_MUSIC}');
Error('\P{Blk=:=Ancient_GREEK_MUSIC}');
Expect(1, 119375, '\p{Blk=ancientgreekmusic}', "");
Expect(0, 119375, '\p{^Blk=ancientgreekmusic}', "");
Expect(0, 119375, '\P{Blk=ancientgreekmusic}', "");
Expect(1, 119375, '\P{^Blk=ancientgreekmusic}', "");
Expect(0, 119376, '\p{Blk=ancientgreekmusic}', "");
Expect(1, 119376, '\p{^Blk=ancientgreekmusic}', "");
Expect(1, 119376, '\P{Blk=ancientgreekmusic}', "");
Expect(0, 119376, '\P{^Blk=ancientgreekmusic}', "");
Expect(1, 119375, '\p{Blk=_Ancient_Greek_music}', "");
Expect(0, 119375, '\p{^Blk=_Ancient_Greek_music}', "");
Expect(0, 119375, '\P{Blk=_Ancient_Greek_music}', "");
Expect(1, 119375, '\P{^Blk=_Ancient_Greek_music}', "");
Expect(0, 119376, '\p{Blk=_Ancient_Greek_music}', "");
Expect(1, 119376, '\p{^Blk=_Ancient_Greek_music}', "");
Expect(1, 119376, '\P{Blk=_Ancient_Greek_music}', "");
Expect(0, 119376, '\P{^Blk=_Ancient_Greek_music}', "");
Error('\p{Is_Block=_Ancient_Greek_MUSICAL_notation/a/}');
Error('\P{Is_Block=_Ancient_Greek_MUSICAL_notation/a/}');
Expect(1, 119375, '\p{Is_Block=ancientgreekmusicalnotation}', "");
Expect(0, 119375, '\p{^Is_Block=ancientgreekmusicalnotation}', "");
Expect(0, 119375, '\P{Is_Block=ancientgreekmusicalnotation}', "");
Expect(1, 119375, '\P{^Is_Block=ancientgreekmusicalnotation}', "");
Expect(0, 119376, '\p{Is_Block=ancientgreekmusicalnotation}', "");
Expect(1, 119376, '\p{^Is_Block=ancientgreekmusicalnotation}', "");
Expect(1, 119376, '\P{Is_Block=ancientgreekmusicalnotation}', "");
Expect(0, 119376, '\P{^Is_Block=ancientgreekmusicalnotation}', "");
Expect(1, 119375, '\p{Is_Block= -Ancient_greek_MUSICAL_Notation}', "");
Expect(0, 119375, '\p{^Is_Block= -Ancient_greek_MUSICAL_Notation}', "");
Expect(0, 119375, '\P{Is_Block= -Ancient_greek_MUSICAL_Notation}', "");
Expect(1, 119375, '\P{^Is_Block= -Ancient_greek_MUSICAL_Notation}', "");
Expect(0, 119376, '\p{Is_Block= -Ancient_greek_MUSICAL_Notation}', "");
Expect(1, 119376, '\p{^Is_Block= -Ancient_greek_MUSICAL_Notation}', "");
Expect(1, 119376, '\P{Is_Block= -Ancient_greek_MUSICAL_Notation}', "");
Expect(0, 119376, '\P{^Is_Block= -Ancient_greek_MUSICAL_Notation}', "");
Error('\p{Is_Blk=:= ancient_Greek_Music}');
Error('\P{Is_Blk=:= ancient_Greek_Music}');
Expect(1, 119375, '\p{Is_Blk=ancientgreekmusic}', "");
Expect(0, 119375, '\p{^Is_Blk=ancientgreekmusic}', "");
Expect(0, 119375, '\P{Is_Blk=ancientgreekmusic}', "");
Expect(1, 119375, '\P{^Is_Blk=ancientgreekmusic}', "");
Expect(0, 119376, '\p{Is_Blk=ancientgreekmusic}', "");
Expect(1, 119376, '\p{^Is_Blk=ancientgreekmusic}', "");
Expect(1, 119376, '\P{Is_Blk=ancientgreekmusic}', "");
Expect(0, 119376, '\P{^Is_Blk=ancientgreekmusic}', "");
Expect(1, 119375, '\p{Is_Blk=-ANCIENT_greek_Music}', "");
Expect(0, 119375, '\p{^Is_Blk=-ANCIENT_greek_Music}', "");
Expect(0, 119375, '\P{Is_Blk=-ANCIENT_greek_Music}', "");
Expect(1, 119375, '\P{^Is_Blk=-ANCIENT_greek_Music}', "");
Expect(0, 119376, '\p{Is_Blk=-ANCIENT_greek_Music}', "");
Expect(1, 119376, '\p{^Is_Blk=-ANCIENT_greek_Music}', "");
Expect(1, 119376, '\P{Is_Blk=-ANCIENT_greek_Music}', "");
Expect(0, 119376, '\P{^Is_Blk=-ANCIENT_greek_Music}', "");
Error('\p{Block:   -PHAGS_Pa/a/}');
Error('\P{Block:   -PHAGS_Pa/a/}');
Expect(1, 43135, '\p{Block=phagspa}', "");
Expect(0, 43135, '\p{^Block=phagspa}', "");
Expect(0, 43135, '\P{Block=phagspa}', "");
Expect(1, 43135, '\P{^Block=phagspa}', "");
Expect(0, 43136, '\p{Block=phagspa}', "");
Expect(1, 43136, '\p{^Block=phagspa}', "");
Expect(1, 43136, '\P{Block=phagspa}', "");
Expect(0, 43136, '\P{^Block=phagspa}', "");
Expect(1, 43135, '\p{Block=	_Phags_Pa}', "");
Expect(0, 43135, '\p{^Block=	_Phags_Pa}', "");
Expect(0, 43135, '\P{Block=	_Phags_Pa}', "");
Expect(1, 43135, '\P{^Block=	_Phags_Pa}', "");
Expect(0, 43136, '\p{Block=	_Phags_Pa}', "");
Expect(1, 43136, '\p{^Block=	_Phags_Pa}', "");
Expect(1, 43136, '\P{Block=	_Phags_Pa}', "");
Expect(0, 43136, '\P{^Block=	_Phags_Pa}', "");
Error('\p{Blk=/a/Phags_Pa}');
Error('\P{Blk=/a/Phags_Pa}');
Expect(1, 43135, '\p{Blk=phagspa}', "");
Expect(0, 43135, '\p{^Blk=phagspa}', "");
Expect(0, 43135, '\P{Blk=phagspa}', "");
Expect(1, 43135, '\P{^Blk=phagspa}', "");
Expect(0, 43136, '\p{Blk=phagspa}', "");
Expect(1, 43136, '\p{^Blk=phagspa}', "");
Expect(1, 43136, '\P{Blk=phagspa}', "");
Expect(0, 43136, '\P{^Blk=phagspa}', "");
Expect(1, 43135, '\p{Blk=-	Phags_pa}', "");
Expect(0, 43135, '\p{^Blk=-	Phags_pa}', "");
Expect(0, 43135, '\P{Blk=-	Phags_pa}', "");
Expect(1, 43135, '\P{^Blk=-	Phags_pa}', "");
Expect(0, 43136, '\p{Blk=-	Phags_pa}', "");
Expect(1, 43136, '\p{^Blk=-	Phags_pa}', "");
Expect(1, 43136, '\P{Blk=-	Phags_pa}', "");
Expect(0, 43136, '\P{^Blk=-	Phags_pa}', "");
Error('\p{Is_Block= :=Phags_Pa}');
Error('\P{Is_Block= :=Phags_Pa}');
Expect(1, 43135, '\p{Is_Block:phagspa}', "");
Expect(0, 43135, '\p{^Is_Block:phagspa}', "");
Expect(0, 43135, '\P{Is_Block:phagspa}', "");
Expect(1, 43135, '\P{^Is_Block:phagspa}', "");
Expect(0, 43136, '\p{Is_Block:phagspa}', "");
Expect(1, 43136, '\p{^Is_Block:phagspa}', "");
Expect(1, 43136, '\P{Is_Block:phagspa}', "");
Expect(0, 43136, '\P{^Is_Block:phagspa}', "");
Expect(1, 43135, '\p{Is_Block:		-phags_PA}', "");
Expect(0, 43135, '\p{^Is_Block:		-phags_PA}', "");
Expect(0, 43135, '\P{Is_Block:		-phags_PA}', "");
Expect(1, 43135, '\P{^Is_Block:		-phags_PA}', "");
Expect(0, 43136, '\p{Is_Block:		-phags_PA}', "");
Expect(1, 43136, '\p{^Is_Block:		-phags_PA}', "");
Expect(1, 43136, '\P{Is_Block:		-phags_PA}', "");
Expect(0, 43136, '\P{^Is_Block:		-phags_PA}', "");
Error('\p{Is_Blk=	PHAGS_PA/a/}');
Error('\P{Is_Blk=	PHAGS_PA/a/}');
Expect(1, 43135, '\p{Is_Blk=phagspa}', "");
Expect(0, 43135, '\p{^Is_Blk=phagspa}', "");
Expect(0, 43135, '\P{Is_Blk=phagspa}', "");
Expect(1, 43135, '\P{^Is_Blk=phagspa}', "");
Expect(0, 43136, '\p{Is_Blk=phagspa}', "");
Expect(1, 43136, '\p{^Is_Blk=phagspa}', "");
Expect(1, 43136, '\P{Is_Blk=phagspa}', "");
Expect(0, 43136, '\P{^Is_Blk=phagspa}', "");
Expect(1, 43135, '\p{Is_Blk:   - Phags_PA}', "");
Expect(0, 43135, '\p{^Is_Blk:   - Phags_PA}', "");
Expect(0, 43135, '\P{Is_Blk:   - Phags_PA}', "");
Expect(1, 43135, '\P{^Is_Blk:   - Phags_PA}', "");
Expect(0, 43136, '\p{Is_Blk:   - Phags_PA}', "");
Expect(1, 43136, '\p{^Is_Blk:   - Phags_PA}', "");
Expect(1, 43136, '\P{Is_Blk:   - Phags_PA}', "");
Expect(0, 43136, '\P{^Is_Blk:   - Phags_PA}', "");
Error('\p{Block=	-Thaana/a/}');
Error('\P{Block=	-Thaana/a/}');
Expect(1, 1983, '\p{Block=thaana}', "");
Expect(0, 1983, '\p{^Block=thaana}', "");
Expect(0, 1983, '\P{Block=thaana}', "");
Expect(1, 1983, '\P{^Block=thaana}', "");
Expect(0, 1984, '\p{Block=thaana}', "");
Expect(1, 1984, '\p{^Block=thaana}', "");
Expect(1, 1984, '\P{Block=thaana}', "");
Expect(0, 1984, '\P{^Block=thaana}', "");
Expect(1, 1983, '\p{Block=	Thaana}', "");
Expect(0, 1983, '\p{^Block=	Thaana}', "");
Expect(0, 1983, '\P{Block=	Thaana}', "");
Expect(1, 1983, '\P{^Block=	Thaana}', "");
Expect(0, 1984, '\p{Block=	Thaana}', "");
Expect(1, 1984, '\p{^Block=	Thaana}', "");
Expect(1, 1984, '\P{Block=	Thaana}', "");
Expect(0, 1984, '\P{^Block=	Thaana}', "");
Error('\p{Blk=:= Thaana}');
Error('\P{Blk=:= Thaana}');
Expect(1, 1983, '\p{Blk=thaana}', "");
Expect(0, 1983, '\p{^Blk=thaana}', "");
Expect(0, 1983, '\P{Blk=thaana}', "");
Expect(1, 1983, '\P{^Blk=thaana}', "");
Expect(0, 1984, '\p{Blk=thaana}', "");
Expect(1, 1984, '\p{^Blk=thaana}', "");
Expect(1, 1984, '\P{Blk=thaana}', "");
Expect(0, 1984, '\P{^Blk=thaana}', "");
Expect(1, 1983, '\p{Blk=_ thaana}', "");
Expect(0, 1983, '\p{^Blk=_ thaana}', "");
Expect(0, 1983, '\P{Blk=_ thaana}', "");
Expect(1, 1983, '\P{^Blk=_ thaana}', "");
Expect(0, 1984, '\p{Blk=_ thaana}', "");
Expect(1, 1984, '\p{^Blk=_ thaana}', "");
Expect(1, 1984, '\P{Blk=_ thaana}', "");
Expect(0, 1984, '\P{^Blk=_ thaana}', "");
Error('\p{Is_Block=	:=Thaana}');
Error('\P{Is_Block=	:=Thaana}');
Expect(1, 1983, '\p{Is_Block=thaana}', "");
Expect(0, 1983, '\p{^Is_Block=thaana}', "");
Expect(0, 1983, '\P{Is_Block=thaana}', "");
Expect(1, 1983, '\P{^Is_Block=thaana}', "");
Expect(0, 1984, '\p{Is_Block=thaana}', "");
Expect(1, 1984, '\p{^Is_Block=thaana}', "");
Expect(1, 1984, '\P{Is_Block=thaana}', "");
Expect(0, 1984, '\P{^Is_Block=thaana}', "");
Expect(1, 1983, '\p{Is_Block=_THAANA}', "");
Expect(0, 1983, '\p{^Is_Block=_THAANA}', "");
Expect(0, 1983, '\P{Is_Block=_THAANA}', "");
Expect(1, 1983, '\P{^Is_Block=_THAANA}', "");
Expect(0, 1984, '\p{Is_Block=_THAANA}', "");
Expect(1, 1984, '\p{^Is_Block=_THAANA}', "");
Expect(1, 1984, '\P{Is_Block=_THAANA}', "");
Expect(0, 1984, '\P{^Is_Block=_THAANA}', "");
Error('\p{Is_Blk=-:=THAANA}');
Error('\P{Is_Blk=-:=THAANA}');
Expect(1, 1983, '\p{Is_Blk=thaana}', "");
Expect(0, 1983, '\p{^Is_Blk=thaana}', "");
Expect(0, 1983, '\P{Is_Blk=thaana}', "");
Expect(1, 1983, '\P{^Is_Blk=thaana}', "");
Expect(0, 1984, '\p{Is_Blk=thaana}', "");
Expect(1, 1984, '\p{^Is_Blk=thaana}', "");
Expect(1, 1984, '\P{Is_Blk=thaana}', "");
Expect(0, 1984, '\P{^Is_Blk=thaana}', "");
Expect(1, 1983, '\p{Is_Blk=_thaana}', "");
Expect(0, 1983, '\p{^Is_Blk=_thaana}', "");
Expect(0, 1983, '\P{Is_Blk=_thaana}', "");
Expect(1, 1983, '\P{^Is_Blk=_thaana}', "");
Expect(0, 1984, '\p{Is_Blk=_thaana}', "");
Expect(1, 1984, '\p{^Is_Blk=_thaana}', "");
Expect(1, 1984, '\P{Is_Blk=_thaana}', "");
Expect(0, 1984, '\P{^Is_Blk=_thaana}', "");
Error('\p{Block=-:=enclosed_Alphanumeric_SUPPLEMENT}');
Error('\P{Block=-:=enclosed_Alphanumeric_SUPPLEMENT}');
Expect(1, 127487, '\p{Block=enclosedalphanumericsupplement}', "");
Expect(0, 127487, '\p{^Block=enclosedalphanumericsupplement}', "");
Expect(0, 127487, '\P{Block=enclosedalphanumericsupplement}', "");
Expect(1, 127487, '\P{^Block=enclosedalphanumericsupplement}', "");
Expect(0, 127488, '\p{Block=enclosedalphanumericsupplement}', "");
Expect(1, 127488, '\p{^Block=enclosedalphanumericsupplement}', "");
Expect(1, 127488, '\P{Block=enclosedalphanumericsupplement}', "");
Expect(0, 127488, '\P{^Block=enclosedalphanumericsupplement}', "");
Expect(1, 127487, '\p{Block=  ENCLOSED_Alphanumeric_Supplement}', "");
Expect(0, 127487, '\p{^Block=  ENCLOSED_Alphanumeric_Supplement}', "");
Expect(0, 127487, '\P{Block=  ENCLOSED_Alphanumeric_Supplement}', "");
Expect(1, 127487, '\P{^Block=  ENCLOSED_Alphanumeric_Supplement}', "");
Expect(0, 127488, '\p{Block=  ENCLOSED_Alphanumeric_Supplement}', "");
Expect(1, 127488, '\p{^Block=  ENCLOSED_Alphanumeric_Supplement}', "");
Expect(1, 127488, '\P{Block=  ENCLOSED_Alphanumeric_Supplement}', "");
Expect(0, 127488, '\P{^Block=  ENCLOSED_Alphanumeric_Supplement}', "");
Error('\p{Blk=:=Enclosed_Alphanum_Sup}');
Error('\P{Blk=:=Enclosed_Alphanum_Sup}');
Expect(1, 127487, '\p{Blk=enclosedalphanumsup}', "");
Expect(0, 127487, '\p{^Blk=enclosedalphanumsup}', "");
Expect(0, 127487, '\P{Blk=enclosedalphanumsup}', "");
Expect(1, 127487, '\P{^Blk=enclosedalphanumsup}', "");
Expect(0, 127488, '\p{Blk=enclosedalphanumsup}', "");
Expect(1, 127488, '\p{^Blk=enclosedalphanumsup}', "");
Expect(1, 127488, '\P{Blk=enclosedalphanumsup}', "");
Expect(0, 127488, '\P{^Blk=enclosedalphanumsup}', "");
Expect(1, 127487, '\p{Blk:		enclosed_alphanum_SUP}', "");
Expect(0, 127487, '\p{^Blk:		enclosed_alphanum_SUP}', "");
Expect(0, 127487, '\P{Blk:		enclosed_alphanum_SUP}', "");
Expect(1, 127487, '\P{^Blk:		enclosed_alphanum_SUP}', "");
Expect(0, 127488, '\p{Blk:		enclosed_alphanum_SUP}', "");
Expect(1, 127488, '\p{^Blk:		enclosed_alphanum_SUP}', "");
Expect(1, 127488, '\P{Blk:		enclosed_alphanum_SUP}', "");
Expect(0, 127488, '\P{^Blk:		enclosed_alphanum_SUP}', "");
Error('\p{Is_Block= /a/enclosed_Alphanumeric_SUPPLEMENT}');
Error('\P{Is_Block= /a/enclosed_Alphanumeric_SUPPLEMENT}');
Expect(1, 127487, '\p{Is_Block=enclosedalphanumericsupplement}', "");
Expect(0, 127487, '\p{^Is_Block=enclosedalphanumericsupplement}', "");
Expect(0, 127487, '\P{Is_Block=enclosedalphanumericsupplement}', "");
Expect(1, 127487, '\P{^Is_Block=enclosedalphanumericsupplement}', "");
Expect(0, 127488, '\p{Is_Block=enclosedalphanumericsupplement}', "");
Expect(1, 127488, '\p{^Is_Block=enclosedalphanumericsupplement}', "");
Expect(1, 127488, '\P{Is_Block=enclosedalphanumericsupplement}', "");
Expect(0, 127488, '\P{^Is_Block=enclosedalphanumericsupplement}', "");
Expect(1, 127487, '\p{Is_Block= ENCLOSED_Alphanumeric_Supplement}', "");
Expect(0, 127487, '\p{^Is_Block= ENCLOSED_Alphanumeric_Supplement}', "");
Expect(0, 127487, '\P{Is_Block= ENCLOSED_Alphanumeric_Supplement}', "");
Expect(1, 127487, '\P{^Is_Block= ENCLOSED_Alphanumeric_Supplement}', "");
Expect(0, 127488, '\p{Is_Block= ENCLOSED_Alphanumeric_Supplement}', "");
Expect(1, 127488, '\p{^Is_Block= ENCLOSED_Alphanumeric_Supplement}', "");
Expect(1, 127488, '\P{Is_Block= ENCLOSED_Alphanumeric_Supplement}', "");
Expect(0, 127488, '\P{^Is_Block= ENCLOSED_Alphanumeric_Supplement}', "");
Error('\p{Is_Blk=:=-_enclosed_alphanum_sup}');
Error('\P{Is_Blk=:=-_enclosed_alphanum_sup}');
Expect(1, 127487, '\p{Is_Blk=enclosedalphanumsup}', "");
Expect(0, 127487, '\p{^Is_Blk=enclosedalphanumsup}', "");
Expect(0, 127487, '\P{Is_Blk=enclosedalphanumsup}', "");
Expect(1, 127487, '\P{^Is_Blk=enclosedalphanumsup}', "");
Expect(0, 127488, '\p{Is_Blk=enclosedalphanumsup}', "");
Expect(1, 127488, '\p{^Is_Blk=enclosedalphanumsup}', "");
Expect(1, 127488, '\P{Is_Blk=enclosedalphanumsup}', "");
Expect(0, 127488, '\P{^Is_Blk=enclosedalphanumsup}', "");
Expect(1, 127487, '\p{Is_Blk: 	 ENCLOSED_Alphanum_SUP}', "");
Expect(0, 127487, '\p{^Is_Blk: 	 ENCLOSED_Alphanum_SUP}', "");
Expect(0, 127487, '\P{Is_Blk: 	 ENCLOSED_Alphanum_SUP}', "");
Expect(1, 127487, '\P{^Is_Blk: 	 ENCLOSED_Alphanum_SUP}', "");
Expect(0, 127488, '\p{Is_Blk: 	 ENCLOSED_Alphanum_SUP}', "");
Expect(1, 127488, '\p{^Is_Blk: 	 ENCLOSED_Alphanum_SUP}', "");
Expect(1, 127488, '\P{Is_Blk: 	 ENCLOSED_Alphanum_SUP}', "");
Expect(0, 127488, '\P{^Is_Blk: 	 ENCLOSED_Alphanum_SUP}', "");
Error('\p{Block:	 CJK_Radicals_supplement/a/}');
Error('\P{Block:	 CJK_Radicals_supplement/a/}');
Expect(1, 12031, '\p{Block=cjkradicalssupplement}', "");
Expect(0, 12031, '\p{^Block=cjkradicalssupplement}', "");
Expect(0, 12031, '\P{Block=cjkradicalssupplement}', "");
Expect(1, 12031, '\P{^Block=cjkradicalssupplement}', "");
Expect(0, 12032, '\p{Block=cjkradicalssupplement}', "");
Expect(1, 12032, '\p{^Block=cjkradicalssupplement}', "");
Expect(1, 12032, '\P{Block=cjkradicalssupplement}', "");
Expect(0, 12032, '\P{^Block=cjkradicalssupplement}', "");
Expect(1, 12031, '\p{Block=__CJK_radicals_Supplement}', "");
Expect(0, 12031, '\p{^Block=__CJK_radicals_Supplement}', "");
Expect(0, 12031, '\P{Block=__CJK_radicals_Supplement}', "");
Expect(1, 12031, '\P{^Block=__CJK_radicals_Supplement}', "");
Expect(0, 12032, '\p{Block=__CJK_radicals_Supplement}', "");
Expect(1, 12032, '\p{^Block=__CJK_radicals_Supplement}', "");
Expect(1, 12032, '\P{Block=__CJK_radicals_Supplement}', "");
Expect(0, 12032, '\P{^Block=__CJK_radicals_Supplement}', "");
Error('\p{Blk=/a/ 	CJK_radicals_sup}');
Error('\P{Blk=/a/ 	CJK_radicals_sup}');
Expect(1, 12031, '\p{Blk=cjkradicalssup}', "");
Expect(0, 12031, '\p{^Blk=cjkradicalssup}', "");
Expect(0, 12031, '\P{Blk=cjkradicalssup}', "");
Expect(1, 12031, '\P{^Blk=cjkradicalssup}', "");
Expect(0, 12032, '\p{Blk=cjkradicalssup}', "");
Expect(1, 12032, '\p{^Blk=cjkradicalssup}', "");
Expect(1, 12032, '\P{Blk=cjkradicalssup}', "");
Expect(0, 12032, '\P{^Blk=cjkradicalssup}', "");
Expect(1, 12031, '\p{Blk=__cjk_Radicals_sup}', "");
Expect(0, 12031, '\p{^Blk=__cjk_Radicals_sup}', "");
Expect(0, 12031, '\P{Blk=__cjk_Radicals_sup}', "");
Expect(1, 12031, '\P{^Blk=__cjk_Radicals_sup}', "");
Expect(0, 12032, '\p{Blk=__cjk_Radicals_sup}', "");
Expect(1, 12032, '\p{^Blk=__cjk_Radicals_sup}', "");
Expect(1, 12032, '\P{Blk=__cjk_Radicals_sup}', "");
Expect(0, 12032, '\P{^Blk=__cjk_Radicals_sup}', "");
Error('\p{Is_Block= :=cjk_radicals_SUPPLEMENT}');
Error('\P{Is_Block= :=cjk_radicals_SUPPLEMENT}');
Expect(1, 12031, '\p{Is_Block=cjkradicalssupplement}', "");
Expect(0, 12031, '\p{^Is_Block=cjkradicalssupplement}', "");
Expect(0, 12031, '\P{Is_Block=cjkradicalssupplement}', "");
Expect(1, 12031, '\P{^Is_Block=cjkradicalssupplement}', "");
Expect(0, 12032, '\p{Is_Block=cjkradicalssupplement}', "");
Expect(1, 12032, '\p{^Is_Block=cjkradicalssupplement}', "");
Expect(1, 12032, '\P{Is_Block=cjkradicalssupplement}', "");
Expect(0, 12032, '\P{^Is_Block=cjkradicalssupplement}', "");
Expect(1, 12031, '\p{Is_Block=_-CJK_radicals_Supplement}', "");
Expect(0, 12031, '\p{^Is_Block=_-CJK_radicals_Supplement}', "");
Expect(0, 12031, '\P{Is_Block=_-CJK_radicals_Supplement}', "");
Expect(1, 12031, '\P{^Is_Block=_-CJK_radicals_Supplement}', "");
Expect(0, 12032, '\p{Is_Block=_-CJK_radicals_Supplement}', "");
Expect(1, 12032, '\p{^Is_Block=_-CJK_radicals_Supplement}', "");
Expect(1, 12032, '\P{Is_Block=_-CJK_radicals_Supplement}', "");
Expect(0, 12032, '\P{^Is_Block=_-CJK_radicals_Supplement}', "");
Error('\p{Is_Blk:	 	CJK_radicals_Sup/a/}');
Error('\P{Is_Blk:	 	CJK_radicals_Sup/a/}');
Expect(1, 12031, '\p{Is_Blk=cjkradicalssup}', "");
Expect(0, 12031, '\p{^Is_Blk=cjkradicalssup}', "");
Expect(0, 12031, '\P{Is_Blk=cjkradicalssup}', "");
Expect(1, 12031, '\P{^Is_Blk=cjkradicalssup}', "");
Expect(0, 12032, '\p{Is_Blk=cjkradicalssup}', "");
Expect(1, 12032, '\p{^Is_Blk=cjkradicalssup}', "");
Expect(1, 12032, '\P{Is_Blk=cjkradicalssup}', "");
Expect(0, 12032, '\P{^Is_Blk=cjkradicalssup}', "");
Expect(1, 12031, '\p{Is_Blk=	_CJK_Radicals_Sup}', "");
Expect(0, 12031, '\p{^Is_Blk=	_CJK_Radicals_Sup}', "");
Expect(0, 12031, '\P{Is_Blk=	_CJK_Radicals_Sup}', "");
Expect(1, 12031, '\P{^Is_Blk=	_CJK_Radicals_Sup}', "");
Expect(0, 12032, '\p{Is_Blk=	_CJK_Radicals_Sup}', "");
Expect(1, 12032, '\p{^Is_Blk=	_CJK_Radicals_Sup}', "");
Expect(1, 12032, '\P{Is_Blk=	_CJK_Radicals_Sup}', "");
Expect(0, 12032, '\P{^Is_Blk=	_CJK_Radicals_Sup}', "");
Error('\p{Block= /a/CJK_Compatibility_Forms}');
Error('\P{Block= /a/CJK_Compatibility_Forms}');
Expect(1, 65103, '\p{Block=cjkcompatibilityforms}', "");
Expect(0, 65103, '\p{^Block=cjkcompatibilityforms}', "");
Expect(0, 65103, '\P{Block=cjkcompatibilityforms}', "");
Expect(1, 65103, '\P{^Block=cjkcompatibilityforms}', "");
Expect(0, 65104, '\p{Block=cjkcompatibilityforms}', "");
Expect(1, 65104, '\p{^Block=cjkcompatibilityforms}', "");
Expect(1, 65104, '\P{Block=cjkcompatibilityforms}', "");
Expect(0, 65104, '\P{^Block=cjkcompatibilityforms}', "");
Expect(1, 65103, '\p{Block=	_CJK_COMPATIBILITY_Forms}', "");
Expect(0, 65103, '\p{^Block=	_CJK_COMPATIBILITY_Forms}', "");
Expect(0, 65103, '\P{Block=	_CJK_COMPATIBILITY_Forms}', "");
Expect(1, 65103, '\P{^Block=	_CJK_COMPATIBILITY_Forms}', "");
Expect(0, 65104, '\p{Block=	_CJK_COMPATIBILITY_Forms}', "");
Expect(1, 65104, '\p{^Block=	_CJK_COMPATIBILITY_Forms}', "");
Expect(1, 65104, '\P{Block=	_CJK_COMPATIBILITY_Forms}', "");
Expect(0, 65104, '\P{^Block=	_CJK_COMPATIBILITY_Forms}', "");
Error('\p{Blk=_CJK_COMPAT_Forms:=}');
Error('\P{Blk=_CJK_COMPAT_Forms:=}');
Expect(1, 65103, '\p{Blk=cjkcompatforms}', "");
Expect(0, 65103, '\p{^Blk=cjkcompatforms}', "");
Expect(0, 65103, '\P{Blk=cjkcompatforms}', "");
Expect(1, 65103, '\P{^Blk=cjkcompatforms}', "");
Expect(0, 65104, '\p{Blk=cjkcompatforms}', "");
Expect(1, 65104, '\p{^Blk=cjkcompatforms}', "");
Expect(1, 65104, '\P{Blk=cjkcompatforms}', "");
Expect(0, 65104, '\P{^Blk=cjkcompatforms}', "");
Expect(1, 65103, '\p{Blk=- CJK_COMPAT_forms}', "");
Expect(0, 65103, '\p{^Blk=- CJK_COMPAT_forms}', "");
Expect(0, 65103, '\P{Blk=- CJK_COMPAT_forms}', "");
Expect(1, 65103, '\P{^Blk=- CJK_COMPAT_forms}', "");
Expect(0, 65104, '\p{Blk=- CJK_COMPAT_forms}', "");
Expect(1, 65104, '\p{^Blk=- CJK_COMPAT_forms}', "");
Expect(1, 65104, '\P{Blk=- CJK_COMPAT_forms}', "");
Expect(0, 65104, '\P{^Blk=- CJK_COMPAT_forms}', "");
Error('\p{Is_Block=--CJK_Compatibility_Forms:=}');
Error('\P{Is_Block=--CJK_Compatibility_Forms:=}');
Expect(1, 65103, '\p{Is_Block=cjkcompatibilityforms}', "");
Expect(0, 65103, '\p{^Is_Block=cjkcompatibilityforms}', "");
Expect(0, 65103, '\P{Is_Block=cjkcompatibilityforms}', "");
Expect(1, 65103, '\P{^Is_Block=cjkcompatibilityforms}', "");
Expect(0, 65104, '\p{Is_Block=cjkcompatibilityforms}', "");
Expect(1, 65104, '\p{^Is_Block=cjkcompatibilityforms}', "");
Expect(1, 65104, '\P{Is_Block=cjkcompatibilityforms}', "");
Expect(0, 65104, '\P{^Is_Block=cjkcompatibilityforms}', "");
Expect(1, 65103, '\p{Is_Block=	-CJK_Compatibility_Forms}', "");
Expect(0, 65103, '\p{^Is_Block=	-CJK_Compatibility_Forms}', "");
Expect(0, 65103, '\P{Is_Block=	-CJK_Compatibility_Forms}', "");
Expect(1, 65103, '\P{^Is_Block=	-CJK_Compatibility_Forms}', "");
Expect(0, 65104, '\p{Is_Block=	-CJK_Compatibility_Forms}', "");
Expect(1, 65104, '\p{^Is_Block=	-CJK_Compatibility_Forms}', "");
Expect(1, 65104, '\P{Is_Block=	-CJK_Compatibility_Forms}', "");
Expect(0, 65104, '\P{^Is_Block=	-CJK_Compatibility_Forms}', "");
Error('\p{Is_Blk=:=_	CJK_COMPAT_forms}');
Error('\P{Is_Blk=:=_	CJK_COMPAT_forms}');
Expect(1, 65103, '\p{Is_Blk=cjkcompatforms}', "");
Expect(0, 65103, '\p{^Is_Blk=cjkcompatforms}', "");
Expect(0, 65103, '\P{Is_Blk=cjkcompatforms}', "");
Expect(1, 65103, '\P{^Is_Blk=cjkcompatforms}', "");
Expect(0, 65104, '\p{Is_Blk=cjkcompatforms}', "");
Expect(1, 65104, '\p{^Is_Blk=cjkcompatforms}', "");
Expect(1, 65104, '\P{Is_Blk=cjkcompatforms}', "");
Expect(0, 65104, '\P{^Is_Blk=cjkcompatforms}', "");
Expect(1, 65103, '\p{Is_Blk=__CJK_COMPAT_forms}', "");
Expect(0, 65103, '\p{^Is_Blk=__CJK_COMPAT_forms}', "");
Expect(0, 65103, '\P{Is_Blk=__CJK_COMPAT_forms}', "");
Expect(1, 65103, '\P{^Is_Blk=__CJK_COMPAT_forms}', "");
Expect(0, 65104, '\p{Is_Blk=__CJK_COMPAT_forms}', "");
Expect(1, 65104, '\p{^Is_Blk=__CJK_COMPAT_forms}', "");
Expect(1, 65104, '\P{Is_Blk=__CJK_COMPAT_forms}', "");
Expect(0, 65104, '\P{^Is_Blk=__CJK_COMPAT_forms}', "");
Error('\p{Block=:= Latin_1_supplement}');
Error('\P{Block=:= Latin_1_supplement}');
Expect(1, 255, '\p{Block=latin1supplement}', "");
Expect(0, 255, '\p{^Block=latin1supplement}', "");
Expect(0, 255, '\P{Block=latin1supplement}', "");
Expect(1, 255, '\P{^Block=latin1supplement}', "");
Expect(0, 256, '\p{Block=latin1supplement}', "");
Expect(1, 256, '\p{^Block=latin1supplement}', "");
Expect(1, 256, '\P{Block=latin1supplement}', "");
Expect(0, 256, '\P{^Block=latin1supplement}', "");
Expect(1, 255, '\p{Block=--Latin_1_SUPPLEMENT}', "");
Expect(0, 255, '\p{^Block=--Latin_1_SUPPLEMENT}', "");
Expect(0, 255, '\P{Block=--Latin_1_SUPPLEMENT}', "");
Expect(1, 255, '\P{^Block=--Latin_1_SUPPLEMENT}', "");
Expect(0, 256, '\p{Block=--Latin_1_SUPPLEMENT}', "");
Expect(1, 256, '\p{^Block=--Latin_1_SUPPLEMENT}', "");
Expect(1, 256, '\P{Block=--Latin_1_SUPPLEMENT}', "");
Expect(0, 256, '\P{^Block=--Latin_1_SUPPLEMENT}', "");
Error('\p{Blk=_latin_1_Sup/a/}');
Error('\P{Blk=_latin_1_Sup/a/}');
Expect(1, 255, '\p{Blk=latin1sup}', "");
Expect(0, 255, '\p{^Blk=latin1sup}', "");
Expect(0, 255, '\P{Blk=latin1sup}', "");
Expect(1, 255, '\P{^Blk=latin1sup}', "");
Expect(0, 256, '\p{Blk=latin1sup}', "");
Expect(1, 256, '\p{^Blk=latin1sup}', "");
Expect(1, 256, '\P{Blk=latin1sup}', "");
Expect(0, 256, '\P{^Blk=latin1sup}', "");
Expect(1, 255, '\p{Blk:   -	Latin_1_SUP}', "");
Expect(0, 255, '\p{^Blk:   -	Latin_1_SUP}', "");
Expect(0, 255, '\P{Blk:   -	Latin_1_SUP}', "");
Expect(1, 255, '\P{^Blk:   -	Latin_1_SUP}', "");
Expect(0, 256, '\p{Blk:   -	Latin_1_SUP}', "");
Expect(1, 256, '\p{^Blk:   -	Latin_1_SUP}', "");
Expect(1, 256, '\P{Blk:   -	Latin_1_SUP}', "");
Expect(0, 256, '\P{^Blk:   -	Latin_1_SUP}', "");
Error('\p{Is_Block:	:= 	latin_1}');
Error('\P{Is_Block:	:= 	latin_1}');
Expect(1, 255, '\p{Is_Block=latin1}', "");
Expect(0, 255, '\p{^Is_Block=latin1}', "");
Expect(0, 255, '\P{Is_Block=latin1}', "");
Expect(1, 255, '\P{^Is_Block=latin1}', "");
Expect(0, 256, '\p{Is_Block=latin1}', "");
Expect(1, 256, '\p{^Is_Block=latin1}', "");
Expect(1, 256, '\P{Is_Block=latin1}', "");
Expect(0, 256, '\P{^Is_Block=latin1}', "");
Expect(1, 255, '\p{Is_Block:__latin_1}', "");
Expect(0, 255, '\p{^Is_Block:__latin_1}', "");
Expect(0, 255, '\P{Is_Block:__latin_1}', "");
Expect(1, 255, '\P{^Is_Block:__latin_1}', "");
Expect(0, 256, '\p{Is_Block:__latin_1}', "");
Expect(1, 256, '\p{^Is_Block:__latin_1}', "");
Expect(1, 256, '\P{Is_Block:__latin_1}', "");
Expect(0, 256, '\P{^Is_Block:__latin_1}', "");
Error('\p{Is_Blk=_/a/Latin_1_SUPPLEMENT}');
Error('\P{Is_Blk=_/a/Latin_1_SUPPLEMENT}');
Expect(1, 255, '\p{Is_Blk=latin1supplement}', "");
Expect(0, 255, '\p{^Is_Blk=latin1supplement}', "");
Expect(0, 255, '\P{Is_Blk=latin1supplement}', "");
Expect(1, 255, '\P{^Is_Blk=latin1supplement}', "");
Expect(0, 256, '\p{Is_Blk=latin1supplement}', "");
Expect(1, 256, '\p{^Is_Blk=latin1supplement}', "");
Expect(1, 256, '\P{Is_Blk=latin1supplement}', "");
Expect(0, 256, '\P{^Is_Blk=latin1supplement}', "");
Expect(1, 255, '\p{Is_Blk=	_Latin_1_Supplement}', "");
Expect(0, 255, '\p{^Is_Blk=	_Latin_1_Supplement}', "");
Expect(0, 255, '\P{Is_Blk=	_Latin_1_Supplement}', "");
Expect(1, 255, '\P{^Is_Blk=	_Latin_1_Supplement}', "");
Expect(0, 256, '\p{Is_Blk=	_Latin_1_Supplement}', "");
Expect(1, 256, '\p{^Is_Blk=	_Latin_1_Supplement}', "");
Expect(1, 256, '\P{Is_Blk=	_Latin_1_Supplement}', "");
Expect(0, 256, '\P{^Is_Blk=	_Latin_1_Supplement}', "");
Error('\p{Block:	-latin_extended_B/a/}');
Error('\P{Block:	-latin_extended_B/a/}');
Expect(1, 591, '\p{Block=latinextendedb}', "");
Expect(0, 591, '\p{^Block=latinextendedb}', "");
Expect(0, 591, '\P{Block=latinextendedb}', "");
Expect(1, 591, '\P{^Block=latinextendedb}', "");
Expect(0, 592, '\p{Block=latinextendedb}', "");
Expect(1, 592, '\p{^Block=latinextendedb}', "");
Expect(1, 592, '\P{Block=latinextendedb}', "");
Expect(0, 592, '\P{^Block=latinextendedb}', "");
Expect(1, 591, '\p{Block:	latin_extended_B}', "");
Expect(0, 591, '\p{^Block:	latin_extended_B}', "");
Expect(0, 591, '\P{Block:	latin_extended_B}', "");
Expect(1, 591, '\P{^Block:	latin_extended_B}', "");
Expect(0, 592, '\p{Block:	latin_extended_B}', "");
Expect(1, 592, '\p{^Block:	latin_extended_B}', "");
Expect(1, 592, '\P{Block:	latin_extended_B}', "");
Expect(0, 592, '\P{^Block:	latin_extended_B}', "");
Error('\p{Blk=	/a/Latin_Ext_B}');
Error('\P{Blk=	/a/Latin_Ext_B}');
Expect(1, 591, '\p{Blk=latinextb}', "");
Expect(0, 591, '\p{^Blk=latinextb}', "");
Expect(0, 591, '\P{Blk=latinextb}', "");
Expect(1, 591, '\P{^Blk=latinextb}', "");
Expect(0, 592, '\p{Blk=latinextb}', "");
Expect(1, 592, '\p{^Blk=latinextb}', "");
Expect(1, 592, '\P{Blk=latinextb}', "");
Expect(0, 592, '\P{^Blk=latinextb}', "");
Expect(1, 591, '\p{Blk=	 Latin_Ext_B}', "");
Expect(0, 591, '\p{^Blk=	 Latin_Ext_B}', "");
Expect(0, 591, '\P{Blk=	 Latin_Ext_B}', "");
Expect(1, 591, '\P{^Blk=	 Latin_Ext_B}', "");
Expect(0, 592, '\p{Blk=	 Latin_Ext_B}', "");
Expect(1, 592, '\p{^Blk=	 Latin_Ext_B}', "");
Expect(1, 592, '\P{Blk=	 Latin_Ext_B}', "");
Expect(0, 592, '\P{^Blk=	 Latin_Ext_B}', "");
Error('\p{Is_Block= /a/latin_Extended_B}');
Error('\P{Is_Block= /a/latin_Extended_B}');
Expect(1, 591, '\p{Is_Block=latinextendedb}', "");
Expect(0, 591, '\p{^Is_Block=latinextendedb}', "");
Expect(0, 591, '\P{Is_Block=latinextendedb}', "");
Expect(1, 591, '\P{^Is_Block=latinextendedb}', "");
Expect(0, 592, '\p{Is_Block=latinextendedb}', "");
Expect(1, 592, '\p{^Is_Block=latinextendedb}', "");
Expect(1, 592, '\P{Is_Block=latinextendedb}', "");
Expect(0, 592, '\P{^Is_Block=latinextendedb}', "");
Expect(1, 591, '\p{Is_Block=-latin_Extended_B}', "");
Expect(0, 591, '\p{^Is_Block=-latin_Extended_B}', "");
Expect(0, 591, '\P{Is_Block=-latin_Extended_B}', "");
Expect(1, 591, '\P{^Is_Block=-latin_Extended_B}', "");
Expect(0, 592, '\p{Is_Block=-latin_Extended_B}', "");
Expect(1, 592, '\p{^Is_Block=-latin_Extended_B}', "");
Expect(1, 592, '\P{Is_Block=-latin_Extended_B}', "");
Expect(0, 592, '\P{^Is_Block=-latin_Extended_B}', "");
Error('\p{Is_Blk=	:=latin_Ext_b}');
Error('\P{Is_Blk=	:=latin_Ext_b}');
Expect(1, 591, '\p{Is_Blk=latinextb}', "");
Expect(0, 591, '\p{^Is_Blk=latinextb}', "");
Expect(0, 591, '\P{Is_Blk=latinextb}', "");
Expect(1, 591, '\P{^Is_Blk=latinextb}', "");
Expect(0, 592, '\p{Is_Blk=latinextb}', "");
Expect(1, 592, '\p{^Is_Blk=latinextb}', "");
Expect(1, 592, '\P{Is_Blk=latinextb}', "");
Expect(0, 592, '\P{^Is_Blk=latinextb}', "");
Expect(1, 591, '\p{Is_Blk=	LATIN_Ext_B}', "");
Expect(0, 591, '\p{^Is_Blk=	LATIN_Ext_B}', "");
Expect(0, 591, '\P{Is_Blk=	LATIN_Ext_B}', "");
Expect(1, 591, '\P{^Is_Blk=	LATIN_Ext_B}', "");
Expect(0, 592, '\p{Is_Blk=	LATIN_Ext_B}', "");
Expect(1, 592, '\p{^Is_Blk=	LATIN_Ext_B}', "");
Expect(1, 592, '\P{Is_Blk=	LATIN_Ext_B}', "");
Expect(0, 592, '\P{^Is_Blk=	LATIN_Ext_B}', "");
Error('\p{Block=/a/-Buhid}');
Error('\P{Block=/a/-Buhid}');
Expect(1, 5983, '\p{Block=buhid}', "");
Expect(0, 5983, '\p{^Block=buhid}', "");
Expect(0, 5983, '\P{Block=buhid}', "");
Expect(1, 5983, '\P{^Block=buhid}', "");
Expect(0, 5984, '\p{Block=buhid}', "");
Expect(1, 5984, '\p{^Block=buhid}', "");
Expect(1, 5984, '\P{Block=buhid}', "");
Expect(0, 5984, '\P{^Block=buhid}', "");
Expect(1, 5983, '\p{Block=__Buhid}', "");
Expect(0, 5983, '\p{^Block=__Buhid}', "");
Expect(0, 5983, '\P{Block=__Buhid}', "");
Expect(1, 5983, '\P{^Block=__Buhid}', "");
Expect(0, 5984, '\p{Block=__Buhid}', "");
Expect(1, 5984, '\p{^Block=__Buhid}', "");
Expect(1, 5984, '\P{Block=__Buhid}', "");
Expect(0, 5984, '\P{^Block=__Buhid}', "");
Error('\p{Blk= :=Buhid}');
Error('\P{Blk= :=Buhid}');
Expect(1, 5983, '\p{Blk=buhid}', "");
Expect(0, 5983, '\p{^Blk=buhid}', "");
Expect(0, 5983, '\P{Blk=buhid}', "");
Expect(1, 5983, '\P{^Blk=buhid}', "");
Expect(0, 5984, '\p{Blk=buhid}', "");
Expect(1, 5984, '\p{^Blk=buhid}', "");
Expect(1, 5984, '\P{Blk=buhid}', "");
Expect(0, 5984, '\P{^Blk=buhid}', "");
Expect(1, 5983, '\p{Blk=	-Buhid}', "");
Expect(0, 5983, '\p{^Blk=	-Buhid}', "");
Expect(0, 5983, '\P{Blk=	-Buhid}', "");
Expect(1, 5983, '\P{^Blk=	-Buhid}', "");
Expect(0, 5984, '\p{Blk=	-Buhid}', "");
Expect(1, 5984, '\p{^Blk=	-Buhid}', "");
Expect(1, 5984, '\P{Blk=	-Buhid}', "");
Expect(0, 5984, '\P{^Blk=	-Buhid}', "");
Error('\p{Is_Block=Buhid/a/}');
Error('\P{Is_Block=Buhid/a/}');
Expect(1, 5983, '\p{Is_Block=buhid}', "");
Expect(0, 5983, '\p{^Is_Block=buhid}', "");
Expect(0, 5983, '\P{Is_Block=buhid}', "");
Expect(1, 5983, '\P{^Is_Block=buhid}', "");
Expect(0, 5984, '\p{Is_Block=buhid}', "");
Expect(1, 5984, '\p{^Is_Block=buhid}', "");
Expect(1, 5984, '\P{Is_Block=buhid}', "");
Expect(0, 5984, '\P{^Is_Block=buhid}', "");
Expect(1, 5983, '\p{Is_Block=_ Buhid}', "");
Expect(0, 5983, '\p{^Is_Block=_ Buhid}', "");
Expect(0, 5983, '\P{Is_Block=_ Buhid}', "");
Expect(1, 5983, '\P{^Is_Block=_ Buhid}', "");
Expect(0, 5984, '\p{Is_Block=_ Buhid}', "");
Expect(1, 5984, '\p{^Is_Block=_ Buhid}', "");
Expect(1, 5984, '\P{Is_Block=_ Buhid}', "");
Expect(0, 5984, '\P{^Is_Block=_ Buhid}', "");
Error('\p{Is_Blk=-/a/Buhid}');
Error('\P{Is_Blk=-/a/Buhid}');
Expect(1, 5983, '\p{Is_Blk:	buhid}', "");
Expect(0, 5983, '\p{^Is_Blk:	buhid}', "");
Expect(0, 5983, '\P{Is_Blk:	buhid}', "");
Expect(1, 5983, '\P{^Is_Blk:	buhid}', "");
Expect(0, 5984, '\p{Is_Blk:	buhid}', "");
Expect(1, 5984, '\p{^Is_Blk:	buhid}', "");
Expect(1, 5984, '\P{Is_Blk:	buhid}', "");
Expect(0, 5984, '\P{^Is_Blk:	buhid}', "");
Expect(1, 5983, '\p{Is_Blk= BUHID}', "");
Expect(0, 5983, '\p{^Is_Blk= BUHID}', "");
Expect(0, 5983, '\P{Is_Blk= BUHID}', "");
Expect(1, 5983, '\P{^Is_Blk= BUHID}', "");
Expect(0, 5984, '\p{Is_Blk= BUHID}', "");
Expect(1, 5984, '\p{^Is_Blk= BUHID}', "");
Expect(1, 5984, '\P{Is_Blk= BUHID}', "");
Expect(0, 5984, '\P{^Is_Blk= BUHID}', "");
Error('\p{Block=_/a/Combining_Diacritical_Marks_Supplement}');
Error('\P{Block=_/a/Combining_Diacritical_Marks_Supplement}');
Expect(1, 7679, '\p{Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\p{^Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\P{Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\P{^Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\p{Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\p{^Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\P{Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\P{^Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\p{Block= combining_Diacritical_Marks_SUPPLEMENT}', "");
Expect(0, 7679, '\p{^Block= combining_Diacritical_Marks_SUPPLEMENT}', "");
Expect(0, 7679, '\P{Block= combining_Diacritical_Marks_SUPPLEMENT}', "");
Expect(1, 7679, '\P{^Block= combining_Diacritical_Marks_SUPPLEMENT}', "");
Expect(0, 7680, '\p{Block= combining_Diacritical_Marks_SUPPLEMENT}', "");
Expect(1, 7680, '\p{^Block= combining_Diacritical_Marks_SUPPLEMENT}', "");
Expect(1, 7680, '\P{Block= combining_Diacritical_Marks_SUPPLEMENT}', "");
Expect(0, 7680, '\P{^Block= combining_Diacritical_Marks_SUPPLEMENT}', "");
Error('\p{Blk=/a/__diacriticals_SUP}');
Error('\P{Blk=/a/__diacriticals_SUP}');
Expect(1, 7679, '\p{Blk=diacriticalssup}', "");
Expect(0, 7679, '\p{^Blk=diacriticalssup}', "");
Expect(0, 7679, '\P{Blk=diacriticalssup}', "");
Expect(1, 7679, '\P{^Blk=diacriticalssup}', "");
Expect(0, 7680, '\p{Blk=diacriticalssup}', "");
Expect(1, 7680, '\p{^Blk=diacriticalssup}', "");
Expect(1, 7680, '\P{Blk=diacriticalssup}', "");
Expect(0, 7680, '\P{^Blk=diacriticalssup}', "");
Expect(1, 7679, '\p{Blk:		Diacriticals_Sup}', "");
Expect(0, 7679, '\p{^Blk:		Diacriticals_Sup}', "");
Expect(0, 7679, '\P{Blk:		Diacriticals_Sup}', "");
Expect(1, 7679, '\P{^Blk:		Diacriticals_Sup}', "");
Expect(0, 7680, '\p{Blk:		Diacriticals_Sup}', "");
Expect(1, 7680, '\p{^Blk:		Diacriticals_Sup}', "");
Expect(1, 7680, '\P{Blk:		Diacriticals_Sup}', "");
Expect(0, 7680, '\P{^Blk:		Diacriticals_Sup}', "");
Error('\p{Is_Block=:= -combining_Diacritical_marks_supplement}');
Error('\P{Is_Block=:= -combining_Diacritical_marks_supplement}');
Expect(1, 7679, '\p{Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\p{^Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7679, '\P{Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\P{^Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\p{Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\p{^Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7680, '\P{Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(0, 7680, '\P{^Is_Block=combiningdiacriticalmarkssupplement}', "");
Expect(1, 7679, '\p{Is_Block:-combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(0, 7679, '\p{^Is_Block:-combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(0, 7679, '\P{Is_Block:-combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(1, 7679, '\P{^Is_Block:-combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(0, 7680, '\p{Is_Block:-combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(1, 7680, '\p{^Is_Block:-combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(1, 7680, '\P{Is_Block:-combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Expect(0, 7680, '\P{^Is_Block:-combining_DIACRITICAL_Marks_SUPPLEMENT}', "");
Error('\p{Is_Blk=_Diacriticals_sup/a/}');
Error('\P{Is_Blk=_Diacriticals_sup/a/}');
Expect(1, 7679, '\p{Is_Blk=diacriticalssup}', "");
Expect(0, 7679, '\p{^Is_Blk=diacriticalssup}', "");
Expect(0, 7679, '\P{Is_Blk=diacriticalssup}', "");
Expect(1, 7679, '\P{^Is_Blk=diacriticalssup}', "");
Expect(0, 7680, '\p{Is_Blk=diacriticalssup}', "");
Expect(1, 7680, '\p{^Is_Blk=diacriticalssup}', "");
Expect(1, 7680, '\P{Is_Blk=diacriticalssup}', "");
Expect(0, 7680, '\P{^Is_Blk=diacriticalssup}', "");
Expect(1, 7679, '\p{Is_Blk= _Diacriticals_SUP}', "");
Expect(0, 7679, '\p{^Is_Blk= _Diacriticals_SUP}', "");
Expect(0, 7679, '\P{Is_Blk= _Diacriticals_SUP}', "");
Expect(1, 7679, '\P{^Is_Blk= _Diacriticals_SUP}', "");
Expect(0, 7680, '\p{Is_Blk= _Diacriticals_SUP}', "");
Expect(1, 7680, '\p{^Is_Blk= _Diacriticals_SUP}', "");
Expect(1, 7680, '\P{Is_Blk= _Diacriticals_SUP}', "");
Expect(0, 7680, '\P{^Is_Blk= _Diacriticals_SUP}', "");
Error('\p{Block=:=Mathematical_alphanumeric_SYMBOLS}');
Error('\P{Block=:=Mathematical_alphanumeric_SYMBOLS}');
Expect(1, 120831, '\p{Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\p{^Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\P{Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\P{^Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\p{Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\p{^Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\P{Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\P{^Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\p{Block:    _mathematical_ALPHANUMERIC_Symbols}', "");
Expect(0, 120831, '\p{^Block:    _mathematical_ALPHANUMERIC_Symbols}', "");
Expect(0, 120831, '\P{Block:    _mathematical_ALPHANUMERIC_Symbols}', "");
Expect(1, 120831, '\P{^Block:    _mathematical_ALPHANUMERIC_Symbols}', "");
Expect(0, 120832, '\p{Block:    _mathematical_ALPHANUMERIC_Symbols}', "");
Expect(1, 120832, '\p{^Block:    _mathematical_ALPHANUMERIC_Symbols}', "");
Expect(1, 120832, '\P{Block:    _mathematical_ALPHANUMERIC_Symbols}', "");
Expect(0, 120832, '\P{^Block:    _mathematical_ALPHANUMERIC_Symbols}', "");
Error('\p{Blk= 	MATH_alphanum/a/}');
Error('\P{Blk= 	MATH_alphanum/a/}');
Expect(1, 120831, '\p{Blk=mathalphanum}', "");
Expect(0, 120831, '\p{^Blk=mathalphanum}', "");
Expect(0, 120831, '\P{Blk=mathalphanum}', "");
Expect(1, 120831, '\P{^Blk=mathalphanum}', "");
Expect(0, 120832, '\p{Blk=mathalphanum}', "");
Expect(1, 120832, '\p{^Blk=mathalphanum}', "");
Expect(1, 120832, '\P{Blk=mathalphanum}', "");
Expect(0, 120832, '\P{^Blk=mathalphanum}', "");
Expect(1, 120831, '\p{Blk= 	Math_Alphanum}', "");
Expect(0, 120831, '\p{^Blk= 	Math_Alphanum}', "");
Expect(0, 120831, '\P{Blk= 	Math_Alphanum}', "");
Expect(1, 120831, '\P{^Blk= 	Math_Alphanum}', "");
Expect(0, 120832, '\p{Blk= 	Math_Alphanum}', "");
Expect(1, 120832, '\p{^Blk= 	Math_Alphanum}', "");
Expect(1, 120832, '\P{Blk= 	Math_Alphanum}', "");
Expect(0, 120832, '\P{^Blk= 	Math_Alphanum}', "");
Error('\p{Is_Block=:=  MATHEMATICAL_ALPHANUMERIC_Symbols}');
Error('\P{Is_Block=:=  MATHEMATICAL_ALPHANUMERIC_Symbols}');
Expect(1, 120831, '\p{Is_Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\p{^Is_Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120831, '\P{Is_Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\P{^Is_Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\p{Is_Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\p{^Is_Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120832, '\P{Is_Block=mathematicalalphanumericsymbols}', "");
Expect(0, 120832, '\P{^Is_Block=mathematicalalphanumericsymbols}', "");
Expect(1, 120831, '\p{Is_Block:   MATHEMATICAL_alphanumeric_Symbols}', "");
Expect(0, 120831, '\p{^Is_Block:   MATHEMATICAL_alphanumeric_Symbols}', "");
Expect(0, 120831, '\P{Is_Block:   MATHEMATICAL_alphanumeric_Symbols}', "");
Expect(1, 120831, '\P{^Is_Block:   MATHEMATICAL_alphanumeric_Symbols}', "");
Expect(0, 120832, '\p{Is_Block:   MATHEMATICAL_alphanumeric_Symbols}', "");
Expect(1, 120832, '\p{^Is_Block:   MATHEMATICAL_alphanumeric_Symbols}', "");
Expect(1, 120832, '\P{Is_Block:   MATHEMATICAL_alphanumeric_Symbols}', "");
Expect(0, 120832, '\P{^Is_Block:   MATHEMATICAL_alphanumeric_Symbols}', "");
Error('\p{Is_Blk=_/a/MATH_Alphanum}');
Error('\P{Is_Blk=_/a/MATH_Alphanum}');
Expect(1, 120831, '\p{Is_Blk=mathalphanum}', "");
Expect(0, 120831, '\p{^Is_Blk=mathalphanum}', "");
Expect(0, 120831, '\P{Is_Blk=mathalphanum}', "");
Expect(1, 120831, '\P{^Is_Blk=mathalphanum}', "");
Expect(0, 120832, '\p{Is_Blk=mathalphanum}', "");
Expect(1, 120832, '\p{^Is_Blk=mathalphanum}', "");
Expect(1, 120832, '\P{Is_Blk=mathalphanum}', "");
Expect(0, 120832, '\P{^Is_Blk=mathalphanum}', "");
Expect(1, 120831, '\p{Is_Blk=- MATH_ALPHANUM}', "");
Expect(0, 120831, '\p{^Is_Blk=- MATH_ALPHANUM}', "");
Expect(0, 120831, '\P{Is_Blk=- MATH_ALPHANUM}', "");
Expect(1, 120831, '\P{^Is_Blk=- MATH_ALPHANUM}', "");
Expect(0, 120832, '\p{Is_Blk=- MATH_ALPHANUM}', "");
Expect(1, 120832, '\p{^Is_Blk=- MATH_ALPHANUM}', "");
Expect(1, 120832, '\P{Is_Blk=- MATH_ALPHANUM}', "");
Expect(0, 120832, '\P{^Is_Blk=- MATH_ALPHANUM}', "");
Error('\p{Block=/a/_Meetei_MAYEK_Extensions}');
Error('\P{Block=/a/_Meetei_MAYEK_Extensions}');
Expect(1, 43775, '\p{Block=meeteimayekextensions}', "");
Expect(0, 43775, '\p{^Block=meeteimayekextensions}', "");
Expect(0, 43775, '\P{Block=meeteimayekextensions}', "");
Expect(1, 43775, '\P{^Block=meeteimayekextensions}', "");
Expect(0, 43776, '\p{Block=meeteimayekextensions}', "");
Expect(1, 43776, '\p{^Block=meeteimayekextensions}', "");
Expect(1, 43776, '\P{Block=meeteimayekextensions}', "");
Expect(0, 43776, '\P{^Block=meeteimayekextensions}', "");
Expect(1, 43775, '\p{Block= _Meetei_MAYEK_Extensions}', "");
Expect(0, 43775, '\p{^Block= _Meetei_MAYEK_Extensions}', "");
Expect(0, 43775, '\P{Block= _Meetei_MAYEK_Extensions}', "");
Expect(1, 43775, '\P{^Block= _Meetei_MAYEK_Extensions}', "");
Expect(0, 43776, '\p{Block= _Meetei_MAYEK_Extensions}', "");
Expect(1, 43776, '\p{^Block= _Meetei_MAYEK_Extensions}', "");
Expect(1, 43776, '\P{Block= _Meetei_MAYEK_Extensions}', "");
Expect(0, 43776, '\P{^Block= _Meetei_MAYEK_Extensions}', "");
Error('\p{Blk=/a/ _Meetei_Mayek_ext}');
Error('\P{Blk=/a/ _Meetei_Mayek_ext}');
Expect(1, 43775, '\p{Blk=meeteimayekext}', "");
Expect(0, 43775, '\p{^Blk=meeteimayekext}', "");
Expect(0, 43775, '\P{Blk=meeteimayekext}', "");
Expect(1, 43775, '\P{^Blk=meeteimayekext}', "");
Expect(0, 43776, '\p{Blk=meeteimayekext}', "");
Expect(1, 43776, '\p{^Blk=meeteimayekext}', "");
Expect(1, 43776, '\P{Blk=meeteimayekext}', "");
Expect(0, 43776, '\P{^Blk=meeteimayekext}', "");
Expect(1, 43775, '\p{Blk=	meetei_Mayek_EXT}', "");
Expect(0, 43775, '\p{^Blk=	meetei_Mayek_EXT}', "");
Expect(0, 43775, '\P{Blk=	meetei_Mayek_EXT}', "");
Expect(1, 43775, '\P{^Blk=	meetei_Mayek_EXT}', "");
Expect(0, 43776, '\p{Blk=	meetei_Mayek_EXT}', "");
Expect(1, 43776, '\p{^Blk=	meetei_Mayek_EXT}', "");
Expect(1, 43776, '\P{Blk=	meetei_Mayek_EXT}', "");
Expect(0, 43776, '\P{^Blk=	meetei_Mayek_EXT}', "");
Error('\p{Is_Block=:=		meetei_Mayek_extensions}');
Error('\P{Is_Block=:=		meetei_Mayek_extensions}');
Expect(1, 43775, '\p{Is_Block=meeteimayekextensions}', "");
Expect(0, 43775, '\p{^Is_Block=meeteimayekextensions}', "");
Expect(0, 43775, '\P{Is_Block=meeteimayekextensions}', "");
Expect(1, 43775, '\P{^Is_Block=meeteimayekextensions}', "");
Expect(0, 43776, '\p{Is_Block=meeteimayekextensions}', "");
Expect(1, 43776, '\p{^Is_Block=meeteimayekextensions}', "");
Expect(1, 43776, '\P{Is_Block=meeteimayekextensions}', "");
Expect(0, 43776, '\P{^Is_Block=meeteimayekextensions}', "");
Expect(1, 43775, '\p{Is_Block=_-meetei_Mayek_EXTENSIONS}', "");
Expect(0, 43775, '\p{^Is_Block=_-meetei_Mayek_EXTENSIONS}', "");
Expect(0, 43775, '\P{Is_Block=_-meetei_Mayek_EXTENSIONS}', "");
Expect(1, 43775, '\P{^Is_Block=_-meetei_Mayek_EXTENSIONS}', "");
Expect(0, 43776, '\p{Is_Block=_-meetei_Mayek_EXTENSIONS}', "");
Expect(1, 43776, '\p{^Is_Block=_-meetei_Mayek_EXTENSIONS}', "");
Expect(1, 43776, '\P{Is_Block=_-meetei_Mayek_EXTENSIONS}', "");
Expect(0, 43776, '\P{^Is_Block=_-meetei_Mayek_EXTENSIONS}', "");
Error('\p{Is_Blk=	-Meetei_mayek_Ext:=}');
Error('\P{Is_Blk=	-Meetei_mayek_Ext:=}');
Expect(1, 43775, '\p{Is_Blk=meeteimayekext}', "");
Expect(0, 43775, '\p{^Is_Blk=meeteimayekext}', "");
Expect(0, 43775, '\P{Is_Blk=meeteimayekext}', "");
Expect(1, 43775, '\P{^Is_Blk=meeteimayekext}', "");
Expect(0, 43776, '\p{Is_Blk=meeteimayekext}', "");
Expect(1, 43776, '\p{^Is_Blk=meeteimayekext}', "");
Expect(1, 43776, '\P{Is_Blk=meeteimayekext}', "");
Expect(0, 43776, '\P{^Is_Blk=meeteimayekext}', "");
Expect(1, 43775, '\p{Is_Blk=  MEETEI_Mayek_Ext}', "");
Expect(0, 43775, '\p{^Is_Blk=  MEETEI_Mayek_Ext}', "");
Expect(0, 43775, '\P{Is_Blk=  MEETEI_Mayek_Ext}', "");
Expect(1, 43775, '\P{^Is_Blk=  MEETEI_Mayek_Ext}', "");
Expect(0, 43776, '\p{Is_Blk=  MEETEI_Mayek_Ext}', "");
Expect(1, 43776, '\p{^Is_Blk=  MEETEI_Mayek_Ext}', "");
Expect(1, 43776, '\P{Is_Blk=  MEETEI_Mayek_Ext}', "");
Expect(0, 43776, '\P{^Is_Blk=  MEETEI_Mayek_Ext}', "");
Error('\p{Block=__Hangul_SYLLABLES:=}');
Error('\P{Block=__Hangul_SYLLABLES:=}');
Expect(1, 55215, '\p{Block=hangulsyllables}', "");
Expect(0, 55215, '\p{^Block=hangulsyllables}', "");
Expect(0, 55215, '\P{Block=hangulsyllables}', "");
Expect(1, 55215, '\P{^Block=hangulsyllables}', "");
Expect(0, 55216, '\p{Block=hangulsyllables}', "");
Expect(1, 55216, '\p{^Block=hangulsyllables}', "");
Expect(1, 55216, '\P{Block=hangulsyllables}', "");
Expect(0, 55216, '\P{^Block=hangulsyllables}', "");
Expect(1, 55215, '\p{Block=-_HANGUL_Syllables}', "");
Expect(0, 55215, '\p{^Block=-_HANGUL_Syllables}', "");
Expect(0, 55215, '\P{Block=-_HANGUL_Syllables}', "");
Expect(1, 55215, '\P{^Block=-_HANGUL_Syllables}', "");
Expect(0, 55216, '\p{Block=-_HANGUL_Syllables}', "");
Expect(1, 55216, '\p{^Block=-_HANGUL_Syllables}', "");
Expect(1, 55216, '\P{Block=-_HANGUL_Syllables}', "");
Expect(0, 55216, '\P{^Block=-_HANGUL_Syllables}', "");
Error('\p{Blk=:=--hangul}');
Error('\P{Blk=:=--hangul}');
Expect(1, 55215, '\p{Blk=hangul}', "");
Expect(0, 55215, '\p{^Blk=hangul}', "");
Expect(0, 55215, '\P{Blk=hangul}', "");
Expect(1, 55215, '\P{^Blk=hangul}', "");
Expect(0, 55216, '\p{Blk=hangul}', "");
Expect(1, 55216, '\p{^Blk=hangul}', "");
Expect(1, 55216, '\P{Blk=hangul}', "");
Expect(0, 55216, '\P{^Blk=hangul}', "");
Expect(1, 55215, '\p{Blk=-Hangul}', "");
Expect(0, 55215, '\p{^Blk=-Hangul}', "");
Expect(0, 55215, '\P{Blk=-Hangul}', "");
Expect(1, 55215, '\P{^Blk=-Hangul}', "");
Expect(0, 55216, '\p{Blk=-Hangul}', "");
Expect(1, 55216, '\p{^Blk=-Hangul}', "");
Expect(1, 55216, '\P{Blk=-Hangul}', "");
Expect(0, 55216, '\P{^Blk=-Hangul}', "");
Error('\p{Is_Block=	/a/HANGUL_syllables}');
Error('\P{Is_Block=	/a/HANGUL_syllables}');
Expect(1, 55215, '\p{Is_Block:   hangulsyllables}', "");
Expect(0, 55215, '\p{^Is_Block:   hangulsyllables}', "");
Expect(0, 55215, '\P{Is_Block:   hangulsyllables}', "");
Expect(1, 55215, '\P{^Is_Block:   hangulsyllables}', "");
Expect(0, 55216, '\p{Is_Block:   hangulsyllables}', "");
Expect(1, 55216, '\p{^Is_Block:   hangulsyllables}', "");
Expect(1, 55216, '\P{Is_Block:   hangulsyllables}', "");
Expect(0, 55216, '\P{^Is_Block:   hangulsyllables}', "");
Expect(1, 55215, '\p{Is_Block: 	hangul_syllables}', "");
Expect(0, 55215, '\p{^Is_Block: 	hangul_syllables}', "");
Expect(0, 55215, '\P{Is_Block: 	hangul_syllables}', "");
Expect(1, 55215, '\P{^Is_Block: 	hangul_syllables}', "");
Expect(0, 55216, '\p{Is_Block: 	hangul_syllables}', "");
Expect(1, 55216, '\p{^Is_Block: 	hangul_syllables}', "");
Expect(1, 55216, '\P{Is_Block: 	hangul_syllables}', "");
Expect(0, 55216, '\P{^Is_Block: 	hangul_syllables}', "");
Error('\p{Is_Blk=	_Hangul/a/}');
Error('\P{Is_Blk=	_Hangul/a/}');
Expect(1, 55215, '\p{Is_Blk=hangul}', "");
Expect(0, 55215, '\p{^Is_Blk=hangul}', "");
Expect(0, 55215, '\P{Is_Blk=hangul}', "");
Expect(1, 55215, '\P{^Is_Blk=hangul}', "");
Expect(0, 55216, '\p{Is_Blk=hangul}', "");
Expect(1, 55216, '\p{^Is_Blk=hangul}', "");
Expect(1, 55216, '\P{Is_Blk=hangul}', "");
Expect(0, 55216, '\P{^Is_Blk=hangul}', "");
Expect(1, 55215, '\p{Is_Blk=--HANGUL}', "");
Expect(0, 55215, '\p{^Is_Blk=--HANGUL}', "");
Expect(0, 55215, '\P{Is_Blk=--HANGUL}', "");
Expect(1, 55215, '\P{^Is_Blk=--HANGUL}', "");
Expect(0, 55216, '\p{Is_Blk=--HANGUL}', "");
Expect(1, 55216, '\p{^Is_Blk=--HANGUL}', "");
Expect(1, 55216, '\P{Is_Blk=--HANGUL}', "");
Expect(0, 55216, '\P{^Is_Blk=--HANGUL}', "");
Error('\p{Block: /a/- Supplementary_private_USE_Area_b}');
Error('\P{Block: /a/- Supplementary_private_USE_Area_b}');
Expect(1, 1048576, '\p{Block=supplementaryprivateuseareab}', "");
Expect(0, 1048576, '\p{^Block=supplementaryprivateuseareab}', "");
Expect(0, 1048576, '\P{Block=supplementaryprivateuseareab}', "");
Expect(1, 1048576, '\P{^Block=supplementaryprivateuseareab}', "");
Expect(0, 1, '\p{Block=supplementaryprivateuseareab}', "");
Expect(1, 1, '\p{^Block=supplementaryprivateuseareab}', "");
Expect(1, 1, '\P{Block=supplementaryprivateuseareab}', "");
Expect(0, 1, '\P{^Block=supplementaryprivateuseareab}', "");
Expect(1, 1048576, '\p{Block= Supplementary_private_USE_Area_b}', "");
Expect(0, 1048576, '\p{^Block= Supplementary_private_USE_Area_b}', "");
Expect(0, 1048576, '\P{Block= Supplementary_private_USE_Area_b}', "");
Expect(1, 1048576, '\P{^Block= Supplementary_private_USE_Area_b}', "");
Expect(0, 1, '\p{Block= Supplementary_private_USE_Area_b}', "");
Expect(1, 1, '\p{^Block= Supplementary_private_USE_Area_b}', "");
Expect(1, 1, '\P{Block= Supplementary_private_USE_Area_b}', "");
Expect(0, 1, '\P{^Block= Supplementary_private_USE_Area_b}', "");
Error('\p{Blk: :=sup_pua_B}');
Error('\P{Blk: :=sup_pua_B}');
Expect(1, 1048576, '\p{Blk=suppuab}', "");
Expect(0, 1048576, '\p{^Blk=suppuab}', "");
Expect(0, 1048576, '\P{Blk=suppuab}', "");
Expect(1, 1048576, '\P{^Blk=suppuab}', "");
Expect(0, 1, '\p{Blk=suppuab}', "");
Expect(1, 1, '\p{^Blk=suppuab}', "");
Expect(1, 1, '\P{Blk=suppuab}', "");
Expect(0, 1, '\P{^Blk=suppuab}', "");
Expect(1, 1048576, '\p{Blk=_ Sup_PUA_B}', "");
Expect(0, 1048576, '\p{^Blk=_ Sup_PUA_B}', "");
Expect(0, 1048576, '\P{Blk=_ Sup_PUA_B}', "");
Expect(1, 1048576, '\P{^Blk=_ Sup_PUA_B}', "");
Expect(0, 1, '\p{Blk=_ Sup_PUA_B}', "");
Expect(1, 1, '\p{^Blk=_ Sup_PUA_B}', "");
Expect(1, 1, '\P{Blk=_ Sup_PUA_B}', "");
Expect(0, 1, '\P{^Blk=_ Sup_PUA_B}', "");
Error('\p{Is_Block= supplementary_PRIVATE_Use_Area_b:=}');
Error('\P{Is_Block= supplementary_PRIVATE_Use_Area_b:=}');
Expect(1, 1048576, '\p{Is_Block=supplementaryprivateuseareab}', "");
Expect(0, 1048576, '\p{^Is_Block=supplementaryprivateuseareab}', "");
Expect(0, 1048576, '\P{Is_Block=supplementaryprivateuseareab}', "");
Expect(1, 1048576, '\P{^Is_Block=supplementaryprivateuseareab}', "");
Expect(0, 1, '\p{Is_Block=supplementaryprivateuseareab}', "");
Expect(1, 1, '\p{^Is_Block=supplementaryprivateuseareab}', "");
Expect(1, 1, '\P{Is_Block=supplementaryprivateuseareab}', "");
Expect(0, 1, '\P{^Is_Block=supplementaryprivateuseareab}', "");
Expect(1, 1048576, '\p{Is_Block=_	SUPPLEMENTARY_Private_USE_area_B}', "");
Expect(0, 1048576, '\p{^Is_Block=_	SUPPLEMENTARY_Private_USE_area_B}', "");
Expect(0, 1048576, '\P{Is_Block=_	SUPPLEMENTARY_Private_USE_area_B}', "");
Expect(1, 1048576, '\P{^Is_Block=_	SUPPLEMENTARY_Private_USE_area_B}', "");
Expect(0, 1, '\p{Is_Block=_	SUPPLEMENTARY_Private_USE_area_B}', "");
Expect(1, 1, '\p{^Is_Block=_	SUPPLEMENTARY_Private_USE_area_B}', "");
Expect(1, 1, '\P{Is_Block=_	SUPPLEMENTARY_Private_USE_area_B}', "");
Expect(0, 1, '\P{^Is_Block=_	SUPPLEMENTARY_Private_USE_area_B}', "");
Error('\p{Is_Blk=	:=SUP_PUA_B}');
Error('\P{Is_Blk=	:=SUP_PUA_B}');
Expect(1, 1048576, '\p{Is_Blk=suppuab}', "");
Expect(0, 1048576, '\p{^Is_Blk=suppuab}', "");
Expect(0, 1048576, '\P{Is_Blk=suppuab}', "");
Expect(1, 1048576, '\P{^Is_Blk=suppuab}', "");
Expect(0, 1, '\p{Is_Blk=suppuab}', "");
Expect(1, 1, '\p{^Is_Blk=suppuab}', "");
Expect(1, 1, '\P{Is_Blk=suppuab}', "");
Expect(0, 1, '\P{^Is_Blk=suppuab}', "");
Expect(1, 1048576, '\p{Is_Blk=	_SUP_PUA_B}', "");
Expect(0, 1048576, '\p{^Is_Blk=	_SUP_PUA_B}', "");
Expect(0, 1048576, '\P{Is_Blk=	_SUP_PUA_B}', "");
Expect(1, 1048576, '\P{^Is_Blk=	_SUP_PUA_B}', "");
Expect(0, 1, '\p{Is_Blk=	_SUP_PUA_B}', "");
Expect(1, 1, '\p{^Is_Blk=	_SUP_PUA_B}', "");
Expect(1, 1, '\P{Is_Blk=	_SUP_PUA_B}', "");
Expect(0, 1, '\P{^Is_Blk=	_SUP_PUA_B}', "");
Error('\p{Block:   -Kana_supplement:=}');
Error('\P{Block:   -Kana_supplement:=}');
Expect(1, 110847, '\p{Block=kanasupplement}', "");
Expect(0, 110847, '\p{^Block=kanasupplement}', "");
Expect(0, 110847, '\P{Block=kanasupplement}', "");
Expect(1, 110847, '\P{^Block=kanasupplement}', "");
Expect(0, 110848, '\p{Block=kanasupplement}', "");
Expect(1, 110848, '\p{^Block=kanasupplement}', "");
Expect(1, 110848, '\P{Block=kanasupplement}', "");
Expect(0, 110848, '\P{^Block=kanasupplement}', "");
Expect(1, 110847, '\p{Block=	Kana_Supplement}', "");
Expect(0, 110847, '\p{^Block=	Kana_Supplement}', "");
Expect(0, 110847, '\P{Block=	Kana_Supplement}', "");
Expect(1, 110847, '\P{^Block=	Kana_Supplement}', "");
Expect(0, 110848, '\p{Block=	Kana_Supplement}', "");
Expect(1, 110848, '\p{^Block=	Kana_Supplement}', "");
Expect(1, 110848, '\P{Block=	Kana_Supplement}', "");
Expect(0, 110848, '\P{^Block=	Kana_Supplement}', "");
Error('\p{Blk=/a/Kana_sup}');
Error('\P{Blk=/a/Kana_sup}');
Expect(1, 110847, '\p{Blk=kanasup}', "");
Expect(0, 110847, '\p{^Blk=kanasup}', "");
Expect(0, 110847, '\P{Blk=kanasup}', "");
Expect(1, 110847, '\P{^Blk=kanasup}', "");
Expect(0, 110848, '\p{Blk=kanasup}', "");
Expect(1, 110848, '\p{^Blk=kanasup}', "");
Expect(1, 110848, '\P{Blk=kanasup}', "");
Expect(0, 110848, '\P{^Blk=kanasup}', "");
Expect(1, 110847, '\p{Blk=-KANA_Sup}', "");
Expect(0, 110847, '\p{^Blk=-KANA_Sup}', "");
Expect(0, 110847, '\P{Blk=-KANA_Sup}', "");
Expect(1, 110847, '\P{^Blk=-KANA_Sup}', "");
Expect(0, 110848, '\p{Blk=-KANA_Sup}', "");
Expect(1, 110848, '\p{^Blk=-KANA_Sup}', "");
Expect(1, 110848, '\P{Blk=-KANA_Sup}', "");
Expect(0, 110848, '\P{^Blk=-KANA_Sup}', "");
Error('\p{Is_Block=-:=KANA_Supplement}');
Error('\P{Is_Block=-:=KANA_Supplement}');
Expect(1, 110847, '\p{Is_Block=kanasupplement}', "");
Expect(0, 110847, '\p{^Is_Block=kanasupplement}', "");
Expect(0, 110847, '\P{Is_Block=kanasupplement}', "");
Expect(1, 110847, '\P{^Is_Block=kanasupplement}', "");
Expect(0, 110848, '\p{Is_Block=kanasupplement}', "");
Expect(1, 110848, '\p{^Is_Block=kanasupplement}', "");
Expect(1, 110848, '\P{Is_Block=kanasupplement}', "");
Expect(0, 110848, '\P{^Is_Block=kanasupplement}', "");
Expect(1, 110847, '\p{Is_Block=	Kana_Supplement}', "");
Expect(0, 110847, '\p{^Is_Block=	Kana_Supplement}', "");
Expect(0, 110847, '\P{Is_Block=	Kana_Supplement}', "");
Expect(1, 110847, '\P{^Is_Block=	Kana_Supplement}', "");
Expect(0, 110848, '\p{Is_Block=	Kana_Supplement}', "");
Expect(1, 110848, '\p{^Is_Block=	Kana_Supplement}', "");
Expect(1, 110848, '\P{Is_Block=	Kana_Supplement}', "");
Expect(0, 110848, '\P{^Is_Block=	Kana_Supplement}', "");
Error('\p{Is_Blk:	:= -KANA_Sup}');
Error('\P{Is_Blk:	:= -KANA_Sup}');
Expect(1, 110847, '\p{Is_Blk=kanasup}', "");
Expect(0, 110847, '\p{^Is_Blk=kanasup}', "");
Expect(0, 110847, '\P{Is_Blk=kanasup}', "");
Expect(1, 110847, '\P{^Is_Blk=kanasup}', "");
Expect(0, 110848, '\p{Is_Blk=kanasup}', "");
Expect(1, 110848, '\p{^Is_Blk=kanasup}', "");
Expect(1, 110848, '\P{Is_Blk=kanasup}', "");
Expect(0, 110848, '\P{^Is_Blk=kanasup}', "");
Expect(1, 110847, '\p{Is_Blk=	 KANA_SUP}', "");
Expect(0, 110847, '\p{^Is_Blk=	 KANA_SUP}', "");
Expect(0, 110847, '\P{Is_Blk=	 KANA_SUP}', "");
Expect(1, 110847, '\P{^Is_Blk=	 KANA_SUP}', "");
Expect(0, 110848, '\p{Is_Blk=	 KANA_SUP}', "");
Expect(1, 110848, '\p{^Is_Blk=	 KANA_SUP}', "");
Expect(1, 110848, '\P{Is_Blk=	 KANA_SUP}', "");
Expect(0, 110848, '\P{^Is_Blk=	 KANA_SUP}', "");
Error('\p{Block= _Miscellaneous_Symbols_AND_arrows:=}');
Error('\P{Block= _Miscellaneous_Symbols_AND_arrows:=}');
Expect(1, 11263, '\p{Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\p{^Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\P{Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\P{^Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\p{Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\p{^Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\P{Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\P{^Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\p{Block=-	Miscellaneous_SYMBOLS_AND_ARROWS}', "");
Expect(0, 11263, '\p{^Block=-	Miscellaneous_SYMBOLS_AND_ARROWS}', "");
Expect(0, 11263, '\P{Block=-	Miscellaneous_SYMBOLS_AND_ARROWS}', "");
Expect(1, 11263, '\P{^Block=-	Miscellaneous_SYMBOLS_AND_ARROWS}', "");
Expect(0, 11264, '\p{Block=-	Miscellaneous_SYMBOLS_AND_ARROWS}', "");
Expect(1, 11264, '\p{^Block=-	Miscellaneous_SYMBOLS_AND_ARROWS}', "");
Expect(1, 11264, '\P{Block=-	Miscellaneous_SYMBOLS_AND_ARROWS}', "");
Expect(0, 11264, '\P{^Block=-	Miscellaneous_SYMBOLS_AND_ARROWS}', "");
Error('\p{Blk: :=Misc_Arrows}');
Error('\P{Blk: :=Misc_Arrows}');
Expect(1, 11263, '\p{Blk=miscarrows}', "");
Expect(0, 11263, '\p{^Blk=miscarrows}', "");
Expect(0, 11263, '\P{Blk=miscarrows}', "");
Expect(1, 11263, '\P{^Blk=miscarrows}', "");
Expect(0, 11264, '\p{Blk=miscarrows}', "");
Expect(1, 11264, '\p{^Blk=miscarrows}', "");
Expect(1, 11264, '\P{Blk=miscarrows}', "");
Expect(0, 11264, '\P{^Blk=miscarrows}', "");
Expect(1, 11263, '\p{Blk=MISC_ARROWS}', "");
Expect(0, 11263, '\p{^Blk=MISC_ARROWS}', "");
Expect(0, 11263, '\P{Blk=MISC_ARROWS}', "");
Expect(1, 11263, '\P{^Blk=MISC_ARROWS}', "");
Expect(0, 11264, '\p{Blk=MISC_ARROWS}', "");
Expect(1, 11264, '\p{^Blk=MISC_ARROWS}', "");
Expect(1, 11264, '\P{Blk=MISC_ARROWS}', "");
Expect(0, 11264, '\P{^Blk=MISC_ARROWS}', "");
Error('\p{Is_Block=:=-_Miscellaneous_Symbols_and_Arrows}');
Error('\P{Is_Block=:=-_Miscellaneous_Symbols_and_Arrows}');
Expect(1, 11263, '\p{Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\p{^Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11263, '\P{Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\P{^Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\p{Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\p{^Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11264, '\P{Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(0, 11264, '\P{^Is_Block=miscellaneoussymbolsandarrows}', "");
Expect(1, 11263, '\p{Is_Block= 	MISCELLANEOUS_symbols_And_Arrows}', "");
Expect(0, 11263, '\p{^Is_Block= 	MISCELLANEOUS_symbols_And_Arrows}', "");
Expect(0, 11263, '\P{Is_Block= 	MISCELLANEOUS_symbols_And_Arrows}', "");
Expect(1, 11263, '\P{^Is_Block= 	MISCELLANEOUS_symbols_And_Arrows}', "");
Expect(0, 11264, '\p{Is_Block= 	MISCELLANEOUS_symbols_And_Arrows}', "");
Expect(1, 11264, '\p{^Is_Block= 	MISCELLANEOUS_symbols_And_Arrows}', "");
Expect(1, 11264, '\P{Is_Block= 	MISCELLANEOUS_symbols_And_Arrows}', "");
Expect(0, 11264, '\P{^Is_Block= 	MISCELLANEOUS_symbols_And_Arrows}', "");
Error('\p{Is_Blk=:=Misc_Arrows}');
Error('\P{Is_Blk=:=Misc_Arrows}');
Expect(1, 11263, '\p{Is_Blk=miscarrows}', "");
Expect(0, 11263, '\p{^Is_Blk=miscarrows}', "");
Expect(0, 11263, '\P{Is_Blk=miscarrows}', "");
Expect(1, 11263, '\P{^Is_Blk=miscarrows}', "");
Expect(0, 11264, '\p{Is_Blk=miscarrows}', "");
Expect(1, 11264, '\p{^Is_Blk=miscarrows}', "");
Expect(1, 11264, '\P{Is_Blk=miscarrows}', "");
Expect(0, 11264, '\P{^Is_Blk=miscarrows}', "");
Expect(1, 11263, '\p{Is_Blk=	-Misc_ARROWS}', "");
Expect(0, 11263, '\p{^Is_Blk=	-Misc_ARROWS}', "");
Expect(0, 11263, '\P{Is_Blk=	-Misc_ARROWS}', "");
Expect(1, 11263, '\P{^Is_Blk=	-Misc_ARROWS}', "");
Expect(0, 11264, '\p{Is_Blk=	-Misc_ARROWS}', "");
Expect(1, 11264, '\p{^Is_Blk=	-Misc_ARROWS}', "");
Expect(1, 11264, '\P{Is_Blk=	-Misc_ARROWS}', "");
Expect(0, 11264, '\P{^Is_Blk=	-Misc_ARROWS}', "");
Error('\p{Block=:=Mathematical_Operators}');
Error('\P{Block=:=Mathematical_Operators}');
Expect(1, 8959, '\p{Block=mathematicaloperators}', "");
Expect(0, 8959, '\p{^Block=mathematicaloperators}', "");
Expect(0, 8959, '\P{Block=mathematicaloperators}', "");
Expect(1, 8959, '\P{^Block=mathematicaloperators}', "");
Expect(0, 8960, '\p{Block=mathematicaloperators}', "");
Expect(1, 8960, '\p{^Block=mathematicaloperators}', "");
Expect(1, 8960, '\P{Block=mathematicaloperators}', "");
Expect(0, 8960, '\P{^Block=mathematicaloperators}', "");
Expect(1, 8959, '\p{Block= -Mathematical_Operators}', "");
Expect(0, 8959, '\p{^Block= -Mathematical_Operators}', "");
Expect(0, 8959, '\P{Block= -Mathematical_Operators}', "");
Expect(1, 8959, '\P{^Block= -Mathematical_Operators}', "");
Expect(0, 8960, '\p{Block= -Mathematical_Operators}', "");
Expect(1, 8960, '\p{^Block= -Mathematical_Operators}', "");
Expect(1, 8960, '\P{Block= -Mathematical_Operators}', "");
Expect(0, 8960, '\P{^Block= -Mathematical_Operators}', "");
Error('\p{Blk=/a/Math_Operators}');
Error('\P{Blk=/a/Math_Operators}');
Expect(1, 8959, '\p{Blk:	mathoperators}', "");
Expect(0, 8959, '\p{^Blk:	mathoperators}', "");
Expect(0, 8959, '\P{Blk:	mathoperators}', "");
Expect(1, 8959, '\P{^Blk:	mathoperators}', "");
Expect(0, 8960, '\p{Blk:	mathoperators}', "");
Expect(1, 8960, '\p{^Blk:	mathoperators}', "");
Expect(1, 8960, '\P{Blk:	mathoperators}', "");
Expect(0, 8960, '\P{^Blk:	mathoperators}', "");
Expect(1, 8959, '\p{Blk= _math_OPERATORS}', "");
Expect(0, 8959, '\p{^Blk= _math_OPERATORS}', "");
Expect(0, 8959, '\P{Blk= _math_OPERATORS}', "");
Expect(1, 8959, '\P{^Blk= _math_OPERATORS}', "");
Expect(0, 8960, '\p{Blk= _math_OPERATORS}', "");
Expect(1, 8960, '\p{^Blk= _math_OPERATORS}', "");
Expect(1, 8960, '\P{Blk= _math_OPERATORS}', "");
Expect(0, 8960, '\P{^Blk= _math_OPERATORS}', "");
Error('\p{Is_Block=	mathematical_Operators/a/}');
Error('\P{Is_Block=	mathematical_Operators/a/}');
Expect(1, 8959, '\p{Is_Block=mathematicaloperators}', "");
Expect(0, 8959, '\p{^Is_Block=mathematicaloperators}', "");
Expect(0, 8959, '\P{Is_Block=mathematicaloperators}', "");
Expect(1, 8959, '\P{^Is_Block=mathematicaloperators}', "");
Expect(0, 8960, '\p{Is_Block=mathematicaloperators}', "");
Expect(1, 8960, '\p{^Is_Block=mathematicaloperators}', "");
Expect(1, 8960, '\P{Is_Block=mathematicaloperators}', "");
Expect(0, 8960, '\P{^Is_Block=mathematicaloperators}', "");
Expect(1, 8959, '\p{Is_Block=		Mathematical_OPERATORS}', "");
Expect(0, 8959, '\p{^Is_Block=		Mathematical_OPERATORS}', "");
Expect(0, 8959, '\P{Is_Block=		Mathematical_OPERATORS}', "");
Expect(1, 8959, '\P{^Is_Block=		Mathematical_OPERATORS}', "");
Expect(0, 8960, '\p{Is_Block=		Mathematical_OPERATORS}', "");
Expect(1, 8960, '\p{^Is_Block=		Mathematical_OPERATORS}', "");
Expect(1, 8960, '\P{Is_Block=		Mathematical_OPERATORS}', "");
Expect(0, 8960, '\P{^Is_Block=		Mathematical_OPERATORS}', "");
Error('\p{Is_Blk= :=Math_Operators}');
Error('\P{Is_Blk= :=Math_Operators}');
Expect(1, 8959, '\p{Is_Blk=mathoperators}', "");
Expect(0, 8959, '\p{^Is_Blk=mathoperators}', "");
Expect(0, 8959, '\P{Is_Blk=mathoperators}', "");
Expect(1, 8959, '\P{^Is_Blk=mathoperators}', "");
Expect(0, 8960, '\p{Is_Blk=mathoperators}', "");
Expect(1, 8960, '\p{^Is_Blk=mathoperators}', "");
Expect(1, 8960, '\P{Is_Blk=mathoperators}', "");
Expect(0, 8960, '\P{^Is_Blk=mathoperators}', "");
Expect(1, 8959, '\p{Is_Blk= -math_operators}', "");
Expect(0, 8959, '\p{^Is_Blk= -math_operators}', "");
Expect(0, 8959, '\P{Is_Blk= -math_operators}', "");
Expect(1, 8959, '\P{^Is_Blk= -math_operators}', "");
Expect(0, 8960, '\p{Is_Blk= -math_operators}', "");
Expect(1, 8960, '\p{^Is_Blk= -math_operators}', "");
Expect(1, 8960, '\P{Is_Blk= -math_operators}', "");
Expect(0, 8960, '\P{^Is_Blk= -math_operators}', "");
Error('\p{Block=-/a/Yijing_hexagram_Symbols}');
Error('\P{Block=-/a/Yijing_hexagram_Symbols}');
Expect(1, 19967, '\p{Block=yijinghexagramsymbols}', "");
Expect(0, 19967, '\p{^Block=yijinghexagramsymbols}', "");
Expect(0, 19967, '\P{Block=yijinghexagramsymbols}', "");
Expect(1, 19967, '\P{^Block=yijinghexagramsymbols}', "");
Expect(0, 19968, '\p{Block=yijinghexagramsymbols}', "");
Expect(1, 19968, '\p{^Block=yijinghexagramsymbols}', "");
Expect(1, 19968, '\P{Block=yijinghexagramsymbols}', "");
Expect(0, 19968, '\P{^Block=yijinghexagramsymbols}', "");
Expect(1, 19967, '\p{Block=	Yijing_hexagram_Symbols}', "");
Expect(0, 19967, '\p{^Block=	Yijing_hexagram_Symbols}', "");
Expect(0, 19967, '\P{Block=	Yijing_hexagram_Symbols}', "");
Expect(1, 19967, '\P{^Block=	Yijing_hexagram_Symbols}', "");
Expect(0, 19968, '\p{Block=	Yijing_hexagram_Symbols}', "");
Expect(1, 19968, '\p{^Block=	Yijing_hexagram_Symbols}', "");
Expect(1, 19968, '\P{Block=	Yijing_hexagram_Symbols}', "");
Expect(0, 19968, '\P{^Block=	Yijing_hexagram_Symbols}', "");
Error('\p{Blk=:=--yijing}');
Error('\P{Blk=:=--yijing}');
Expect(1, 19967, '\p{Blk=yijing}', "");
Expect(0, 19967, '\p{^Blk=yijing}', "");
Expect(0, 19967, '\P{Blk=yijing}', "");
Expect(1, 19967, '\P{^Blk=yijing}', "");
Expect(0, 19968, '\p{Blk=yijing}', "");
Expect(1, 19968, '\p{^Blk=yijing}', "");
Expect(1, 19968, '\P{Blk=yijing}', "");
Expect(0, 19968, '\P{^Blk=yijing}', "");
Expect(1, 19967, '\p{Blk:	 YIJING}', "");
Expect(0, 19967, '\p{^Blk:	 YIJING}', "");
Expect(0, 19967, '\P{Blk:	 YIJING}', "");
Expect(1, 19967, '\P{^Blk:	 YIJING}', "");
Expect(0, 19968, '\p{Blk:	 YIJING}', "");
Expect(1, 19968, '\p{^Blk:	 YIJING}', "");
Expect(1, 19968, '\P{Blk:	 YIJING}', "");
Expect(0, 19968, '\P{^Blk:	 YIJING}', "");
Error('\p{Is_Block=_Yijing_HEXAGRAM_Symbols:=}');
Error('\P{Is_Block=_Yijing_HEXAGRAM_Symbols:=}');
Expect(1, 19967, '\p{Is_Block=yijinghexagramsymbols}', "");
Expect(0, 19967, '\p{^Is_Block=yijinghexagramsymbols}', "");
Expect(0, 19967, '\P{Is_Block=yijinghexagramsymbols}', "");
Expect(1, 19967, '\P{^Is_Block=yijinghexagramsymbols}', "");
Expect(0, 19968, '\p{Is_Block=yijinghexagramsymbols}', "");
Expect(1, 19968, '\p{^Is_Block=yijinghexagramsymbols}', "");
Expect(1, 19968, '\P{Is_Block=yijinghexagramsymbols}', "");
Expect(0, 19968, '\P{^Is_Block=yijinghexagramsymbols}', "");
Expect(1, 19967, '\p{Is_Block=  Yijing_HEXAGRAM_symbols}', "");
Expect(0, 19967, '\p{^Is_Block=  Yijing_HEXAGRAM_symbols}', "");
Expect(0, 19967, '\P{Is_Block=  Yijing_HEXAGRAM_symbols}', "");
Expect(1, 19967, '\P{^Is_Block=  Yijing_HEXAGRAM_symbols}', "");
Expect(0, 19968, '\p{Is_Block=  Yijing_HEXAGRAM_symbols}', "");
Expect(1, 19968, '\p{^Is_Block=  Yijing_HEXAGRAM_symbols}', "");
Expect(1, 19968, '\P{Is_Block=  Yijing_HEXAGRAM_symbols}', "");
Expect(0, 19968, '\P{^Is_Block=  Yijing_HEXAGRAM_symbols}', "");
Error('\p{Is_Blk=-_Yijing:=}');
Error('\P{Is_Blk=-_Yijing:=}');
Expect(1, 19967, '\p{Is_Blk=yijing}', "");
Expect(0, 19967, '\p{^Is_Blk=yijing}', "");
Expect(0, 19967, '\P{Is_Blk=yijing}', "");
Expect(1, 19967, '\P{^Is_Blk=yijing}', "");
Expect(0, 19968, '\p{Is_Blk=yijing}', "");
Expect(1, 19968, '\p{^Is_Blk=yijing}', "");
Expect(1, 19968, '\P{Is_Blk=yijing}', "");
Expect(0, 19968, '\P{^Is_Blk=yijing}', "");
Expect(1, 19967, '\p{Is_Blk:  Yijing}', "");
Expect(0, 19967, '\p{^Is_Blk:  Yijing}', "");
Expect(0, 19967, '\P{Is_Blk:  Yijing}', "");
Expect(1, 19967, '\P{^Is_Blk:  Yijing}', "");
Expect(0, 19968, '\p{Is_Blk:  Yijing}', "");
Expect(1, 19968, '\p{^Is_Blk:  Yijing}', "");
Expect(1, 19968, '\P{Is_Blk:  Yijing}', "");
Expect(0, 19968, '\P{^Is_Blk:  Yijing}', "");
Error('\p{Block=_-small_FORM_Variants/a/}');
Error('\P{Block=_-small_FORM_Variants/a/}');
Expect(1, 65135, '\p{Block=smallformvariants}', "");
Expect(0, 65135, '\p{^Block=smallformvariants}', "");
Expect(0, 65135, '\P{Block=smallformvariants}', "");
Expect(1, 65135, '\P{^Block=smallformvariants}', "");
Expect(0, 65136, '\p{Block=smallformvariants}', "");
Expect(1, 65136, '\p{^Block=smallformvariants}', "");
Expect(1, 65136, '\P{Block=smallformvariants}', "");
Expect(0, 65136, '\P{^Block=smallformvariants}', "");
Expect(1, 65135, '\p{Block=_	Small_Form_VARIANTS}', "");
Expect(0, 65135, '\p{^Block=_	Small_Form_VARIANTS}', "");
Expect(0, 65135, '\P{Block=_	Small_Form_VARIANTS}', "");
Expect(1, 65135, '\P{^Block=_	Small_Form_VARIANTS}', "");
Expect(0, 65136, '\p{Block=_	Small_Form_VARIANTS}', "");
Expect(1, 65136, '\p{^Block=_	Small_Form_VARIANTS}', "");
Expect(1, 65136, '\P{Block=_	Small_Form_VARIANTS}', "");
Expect(0, 65136, '\P{^Block=_	Small_Form_VARIANTS}', "");
Error('\p{Blk= :=Small_forms}');
Error('\P{Blk= :=Small_forms}');
Expect(1, 65135, '\p{Blk=smallforms}', "");
Expect(0, 65135, '\p{^Blk=smallforms}', "");
Expect(0, 65135, '\P{Blk=smallforms}', "");
Expect(1, 65135, '\P{^Blk=smallforms}', "");
Expect(0, 65136, '\p{Blk=smallforms}', "");
Expect(1, 65136, '\p{^Blk=smallforms}', "");
Expect(1, 65136, '\P{Blk=smallforms}', "");
Expect(0, 65136, '\P{^Blk=smallforms}', "");
Expect(1, 65135, '\p{Blk=_SMALL_Forms}', "");
Expect(0, 65135, '\p{^Blk=_SMALL_Forms}', "");
Expect(0, 65135, '\P{Blk=_SMALL_Forms}', "");
Expect(1, 65135, '\P{^Blk=_SMALL_Forms}', "");
Expect(0, 65136, '\p{Blk=_SMALL_Forms}', "");
Expect(1, 65136, '\p{^Blk=_SMALL_Forms}', "");
Expect(1, 65136, '\P{Blk=_SMALL_Forms}', "");
Expect(0, 65136, '\P{^Blk=_SMALL_Forms}', "");
Error('\p{Is_Block=/a/		small_Form_variants}');
Error('\P{Is_Block=/a/		small_Form_variants}');
Expect(1, 65135, '\p{Is_Block=smallformvariants}', "");
Expect(0, 65135, '\p{^Is_Block=smallformvariants}', "");
Expect(0, 65135, '\P{Is_Block=smallformvariants}', "");
Expect(1, 65135, '\P{^Is_Block=smallformvariants}', "");
Expect(0, 65136, '\p{Is_Block=smallformvariants}', "");
Expect(1, 65136, '\p{^Is_Block=smallformvariants}', "");
Expect(1, 65136, '\P{Is_Block=smallformvariants}', "");
Expect(0, 65136, '\P{^Is_Block=smallformvariants}', "");
Expect(1, 65135, '\p{Is_Block=_	small_Form_variants}', "");
Expect(0, 65135, '\p{^Is_Block=_	small_Form_variants}', "");
Expect(0, 65135, '\P{Is_Block=_	small_Form_variants}', "");
Expect(1, 65135, '\P{^Is_Block=_	small_Form_variants}', "");
Expect(0, 65136, '\p{Is_Block=_	small_Form_variants}', "");
Expect(1, 65136, '\p{^Is_Block=_	small_Form_variants}', "");
Expect(1, 65136, '\P{Is_Block=_	small_Form_variants}', "");
Expect(0, 65136, '\P{^Is_Block=_	small_Form_variants}', "");
Error('\p{Is_Blk=_-Small_forms/a/}');
Error('\P{Is_Blk=_-Small_forms/a/}');
Expect(1, 65135, '\p{Is_Blk=smallforms}', "");
Expect(0, 65135, '\p{^Is_Blk=smallforms}', "");
Expect(0, 65135, '\P{Is_Blk=smallforms}', "");
Expect(1, 65135, '\P{^Is_Blk=smallforms}', "");
Expect(0, 65136, '\p{Is_Blk=smallforms}', "");
Expect(1, 65136, '\p{^Is_Blk=smallforms}', "");
Expect(1, 65136, '\P{Is_Blk=smallforms}', "");
Expect(0, 65136, '\P{^Is_Blk=smallforms}', "");
Expect(1, 65135, '\p{Is_Blk=__Small_Forms}', "");
Expect(0, 65135, '\p{^Is_Blk=__Small_Forms}', "");
Expect(0, 65135, '\P{Is_Blk=__Small_Forms}', "");
Expect(1, 65135, '\P{^Is_Blk=__Small_Forms}', "");
Expect(0, 65136, '\p{Is_Blk=__Small_Forms}', "");
Expect(1, 65136, '\p{^Is_Blk=__Small_Forms}', "");
Expect(1, 65136, '\P{Is_Blk=__Small_Forms}', "");
Expect(0, 65136, '\P{^Is_Blk=__Small_Forms}', "");
Error('\p{Block=	/a/Egyptian_hieroglyphs}');
Error('\P{Block=	/a/Egyptian_hieroglyphs}');
Expect(1, 78895, '\p{Block=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{^Block=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{Block=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{^Block=egyptianhieroglyphs}', "");
Expect(0, 78896, '\p{Block=egyptianhieroglyphs}', "");
Expect(1, 78896, '\p{^Block=egyptianhieroglyphs}', "");
Expect(1, 78896, '\P{Block=egyptianhieroglyphs}', "");
Expect(0, 78896, '\P{^Block=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{Block=_ Egyptian_hieroglyphs}', "");
Expect(0, 78895, '\p{^Block=_ Egyptian_hieroglyphs}', "");
Expect(0, 78895, '\P{Block=_ Egyptian_hieroglyphs}', "");
Expect(1, 78895, '\P{^Block=_ Egyptian_hieroglyphs}', "");
Expect(0, 78896, '\p{Block=_ Egyptian_hieroglyphs}', "");
Expect(1, 78896, '\p{^Block=_ Egyptian_hieroglyphs}', "");
Expect(1, 78896, '\P{Block=_ Egyptian_hieroglyphs}', "");
Expect(0, 78896, '\P{^Block=_ Egyptian_hieroglyphs}', "");
Error('\p{Blk=-	Egyptian_hieroglyphs:=}');
Error('\P{Blk=-	Egyptian_hieroglyphs:=}');
Expect(1, 78895, '\p{Blk=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{^Blk=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{Blk=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{^Blk=egyptianhieroglyphs}', "");
Expect(0, 78896, '\p{Blk=egyptianhieroglyphs}', "");
Expect(1, 78896, '\p{^Blk=egyptianhieroglyphs}', "");
Expect(1, 78896, '\P{Blk=egyptianhieroglyphs}', "");
Expect(0, 78896, '\P{^Blk=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{Blk=	egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\p{^Blk=	egyptian_HIEROGLYPHS}', "");
Expect(0, 78895, '\P{Blk=	egyptian_HIEROGLYPHS}', "");
Expect(1, 78895, '\P{^Blk=	egyptian_HIEROGLYPHS}', "");
Expect(0, 78896, '\p{Blk=	egyptian_HIEROGLYPHS}', "");
Expect(1, 78896, '\p{^Blk=	egyptian_HIEROGLYPHS}', "");
Expect(1, 78896, '\P{Blk=	egyptian_HIEROGLYPHS}', "");
Expect(0, 78896, '\P{^Blk=	egyptian_HIEROGLYPHS}', "");
Error('\p{Is_Block=/a/- Egyptian_Hieroglyphs}');
Error('\P{Is_Block=/a/- Egyptian_Hieroglyphs}');
Expect(1, 78895, '\p{Is_Block:	egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{^Is_Block:	egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{Is_Block:	egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{^Is_Block:	egyptianhieroglyphs}', "");
Expect(0, 78896, '\p{Is_Block:	egyptianhieroglyphs}', "");
Expect(1, 78896, '\p{^Is_Block:	egyptianhieroglyphs}', "");
Expect(1, 78896, '\P{Is_Block:	egyptianhieroglyphs}', "");
Expect(0, 78896, '\P{^Is_Block:	egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{Is_Block: -EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78895, '\p{^Is_Block: -EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78895, '\P{Is_Block: -EGYPTIAN_Hieroglyphs}', "");
Expect(1, 78895, '\P{^Is_Block: -EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78896, '\p{Is_Block: -EGYPTIAN_Hieroglyphs}', "");
Expect(1, 78896, '\p{^Is_Block: -EGYPTIAN_Hieroglyphs}', "");
Expect(1, 78896, '\P{Is_Block: -EGYPTIAN_Hieroglyphs}', "");
Expect(0, 78896, '\P{^Is_Block: -EGYPTIAN_Hieroglyphs}', "");
Error('\p{Is_Blk=	/a/EGYPTIAN_HIEROGLYPHS}');
Error('\P{Is_Blk=	/a/EGYPTIAN_HIEROGLYPHS}');
Expect(1, 78895, '\p{Is_Blk=egyptianhieroglyphs}', "");
Expect(0, 78895, '\p{^Is_Blk=egyptianhieroglyphs}', "");
Expect(0, 78895, '\P{Is_Blk=egyptianhieroglyphs}', "");
Expect(1, 78895, '\P{^Is_Blk=egyptianhieroglyphs}', "");
Expect(0, 78896, '\p{Is_Blk=egyptianhieroglyphs}', "");
Expect(1, 78896, '\p{^Is_Blk=egyptianhieroglyphs}', "");
Expect(1, 78896, '\P{Is_Blk=egyptianhieroglyphs}', "");
Expect(0, 78896, '\P{^Is_Blk=egyptianhieroglyphs}', "");
Expect(1, 78895, '\p{Is_Blk= -egyptian_Hieroglyphs}', "");
Expect(0, 78895, '\p{^Is_Blk= -egyptian_Hieroglyphs}', "");
Expect(0, 78895, '\P{Is_Blk= -egyptian_Hieroglyphs}', "");
Expect(1, 78895, '\P{^Is_Blk= -egyptian_Hieroglyphs}', "");
Expect(0, 78896, '\p{Is_Blk= -egyptian_Hieroglyphs}', "");
Expect(1, 78896, '\p{^Is_Blk= -egyptian_Hieroglyphs}', "");
Expect(1, 78896, '\P{Is_Blk= -egyptian_Hieroglyphs}', "");
Expect(0, 78896, '\P{^Is_Blk= -egyptian_Hieroglyphs}', "");
Error('\p{Block=_ chakma:=}');
Error('\P{Block=_ chakma:=}');
Expect(1, 69967, '\p{Block=chakma}', "");
Expect(0, 69967, '\p{^Block=chakma}', "");
Expect(0, 69967, '\P{Block=chakma}', "");
Expect(1, 69967, '\P{^Block=chakma}', "");
Expect(0, 69968, '\p{Block=chakma}', "");
Expect(1, 69968, '\p{^Block=chakma}', "");
Expect(1, 69968, '\P{Block=chakma}', "");
Expect(0, 69968, '\P{^Block=chakma}', "");
Expect(1, 69967, '\p{Block=	Chakma}', "");
Expect(0, 69967, '\p{^Block=	Chakma}', "");
Expect(0, 69967, '\P{Block=	Chakma}', "");
Expect(1, 69967, '\P{^Block=	Chakma}', "");
Expect(0, 69968, '\p{Block=	Chakma}', "");
Expect(1, 69968, '\p{^Block=	Chakma}', "");
Expect(1, 69968, '\P{Block=	Chakma}', "");
Expect(0, 69968, '\P{^Block=	Chakma}', "");
Error('\p{Blk= /a/chakma}');
Error('\P{Blk= /a/chakma}');
Expect(1, 69967, '\p{Blk=chakma}', "");
Expect(0, 69967, '\p{^Blk=chakma}', "");
Expect(0, 69967, '\P{Blk=chakma}', "");
Expect(1, 69967, '\P{^Blk=chakma}', "");
Expect(0, 69968, '\p{Blk=chakma}', "");
Expect(1, 69968, '\p{^Blk=chakma}', "");
Expect(1, 69968, '\P{Blk=chakma}', "");
Expect(0, 69968, '\P{^Blk=chakma}', "");
Expect(1, 69967, '\p{Blk=- chakma}', "");
Expect(0, 69967, '\p{^Blk=- chakma}', "");
Expect(0, 69967, '\P{Blk=- chakma}', "");
Expect(1, 69967, '\P{^Blk=- chakma}', "");
Expect(0, 69968, '\p{Blk=- chakma}', "");
Expect(1, 69968, '\p{^Blk=- chakma}', "");
Expect(1, 69968, '\P{Blk=- chakma}', "");
Expect(0, 69968, '\P{^Blk=- chakma}', "");
Error('\p{Is_Block:    :=chakma}');
Error('\P{Is_Block:    :=chakma}');
Expect(1, 69967, '\p{Is_Block=chakma}', "");
Expect(0, 69967, '\p{^Is_Block=chakma}', "");
Expect(0, 69967, '\P{Is_Block=chakma}', "");
Expect(1, 69967, '\P{^Is_Block=chakma}', "");
Expect(0, 69968, '\p{Is_Block=chakma}', "");
Expect(1, 69968, '\p{^Is_Block=chakma}', "");
Expect(1, 69968, '\P{Is_Block=chakma}', "");
Expect(0, 69968, '\P{^Is_Block=chakma}', "");
Expect(1, 69967, '\p{Is_Block=		Chakma}', "");
Expect(0, 69967, '\p{^Is_Block=		Chakma}', "");
Expect(0, 69967, '\P{Is_Block=		Chakma}', "");
Expect(1, 69967, '\P{^Is_Block=		Chakma}', "");
Expect(0, 69968, '\p{Is_Block=		Chakma}', "");
Expect(1, 69968, '\p{^Is_Block=		Chakma}', "");
Expect(1, 69968, '\P{Is_Block=		Chakma}', "");
Expect(0, 69968, '\P{^Is_Block=		Chakma}', "");
Error('\p{Is_Blk=:= -Chakma}');
Error('\P{Is_Blk=:= -Chakma}');
Expect(1, 69967, '\p{Is_Blk=chakma}', "");
Expect(0, 69967, '\p{^Is_Blk=chakma}', "");
Expect(0, 69967, '\P{Is_Blk=chakma}', "");
Expect(1, 69967, '\P{^Is_Blk=chakma}', "");
Expect(0, 69968, '\p{Is_Blk=chakma}', "");
Expect(1, 69968, '\p{^Is_Blk=chakma}', "");
Expect(1, 69968, '\P{Is_Blk=chakma}', "");
Expect(0, 69968, '\P{^Is_Blk=chakma}', "");
Expect(1, 69967, '\p{Is_Blk=--CHAKMA}', "");
Expect(0, 69967, '\p{^Is_Blk=--CHAKMA}', "");
Expect(0, 69967, '\P{Is_Blk=--CHAKMA}', "");
Expect(1, 69967, '\P{^Is_Blk=--CHAKMA}', "");
Expect(0, 69968, '\p{Is_Blk=--CHAKMA}', "");
Expect(1, 69968, '\p{^Is_Blk=--CHAKMA}', "");
Expect(1, 69968, '\P{Is_Blk=--CHAKMA}', "");
Expect(0, 69968, '\P{^Is_Blk=--CHAKMA}', "");
Error('\p{Block=	/a/Optical_CHARACTER_recognition}');
Error('\P{Block=	/a/Optical_CHARACTER_recognition}');
Expect(1, 9311, '\p{Block:opticalcharacterrecognition}', "");
Expect(0, 9311, '\p{^Block:opticalcharacterrecognition}', "");
Expect(0, 9311, '\P{Block:opticalcharacterrecognition}', "");
Expect(1, 9311, '\P{^Block:opticalcharacterrecognition}', "");
Expect(0, 9312, '\p{Block:opticalcharacterrecognition}', "");
Expect(1, 9312, '\p{^Block:opticalcharacterrecognition}', "");
Expect(1, 9312, '\P{Block:opticalcharacterrecognition}', "");
Expect(0, 9312, '\P{^Block:opticalcharacterrecognition}', "");
Expect(1, 9311, '\p{Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(0, 9311, '\p{^Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(0, 9311, '\P{Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(1, 9311, '\P{^Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(0, 9312, '\p{Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(1, 9312, '\p{^Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(1, 9312, '\P{Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(0, 9312, '\P{^Block=-OPTICAL_Character_RECOGNITION}', "");
Error('\p{Blk=-_ocr:=}');
Error('\P{Blk=-_ocr:=}');
Expect(1, 9311, '\p{Blk=ocr}', "");
Expect(0, 9311, '\p{^Blk=ocr}', "");
Expect(0, 9311, '\P{Blk=ocr}', "");
Expect(1, 9311, '\P{^Blk=ocr}', "");
Expect(0, 9312, '\p{Blk=ocr}', "");
Expect(1, 9312, '\p{^Blk=ocr}', "");
Expect(1, 9312, '\P{Blk=ocr}', "");
Expect(0, 9312, '\P{^Blk=ocr}', "");
Expect(1, 9311, '\p{Blk=-OCR}', "");
Expect(0, 9311, '\p{^Blk=-OCR}', "");
Expect(0, 9311, '\P{Blk=-OCR}', "");
Expect(1, 9311, '\P{^Blk=-OCR}', "");
Expect(0, 9312, '\p{Blk=-OCR}', "");
Expect(1, 9312, '\p{^Blk=-OCR}', "");
Expect(1, 9312, '\P{Blk=-OCR}', "");
Expect(0, 9312, '\P{^Blk=-OCR}', "");
Error('\p{Is_Block=	_OPTICAL_Character_recognition:=}');
Error('\P{Is_Block=	_OPTICAL_Character_recognition:=}');
Expect(1, 9311, '\p{Is_Block=opticalcharacterrecognition}', "");
Expect(0, 9311, '\p{^Is_Block=opticalcharacterrecognition}', "");
Expect(0, 9311, '\P{Is_Block=opticalcharacterrecognition}', "");
Expect(1, 9311, '\P{^Is_Block=opticalcharacterrecognition}', "");
Expect(0, 9312, '\p{Is_Block=opticalcharacterrecognition}', "");
Expect(1, 9312, '\p{^Is_Block=opticalcharacterrecognition}', "");
Expect(1, 9312, '\P{Is_Block=opticalcharacterrecognition}', "");
Expect(0, 9312, '\P{^Is_Block=opticalcharacterrecognition}', "");
Expect(1, 9311, '\p{Is_Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(0, 9311, '\p{^Is_Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(0, 9311, '\P{Is_Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(1, 9311, '\P{^Is_Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(0, 9312, '\p{Is_Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(1, 9312, '\p{^Is_Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(1, 9312, '\P{Is_Block=-OPTICAL_Character_RECOGNITION}', "");
Expect(0, 9312, '\P{^Is_Block=-OPTICAL_Character_RECOGNITION}', "");
Error('\p{Is_Blk=:= ocr}');
Error('\P{Is_Blk=:= ocr}');
Expect(1, 9311, '\p{Is_Blk=ocr}', "");
Expect(0, 9311, '\p{^Is_Blk=ocr}', "");
Expect(0, 9311, '\P{Is_Blk=ocr}', "");
Expect(1, 9311, '\P{^Is_Blk=ocr}', "");
Expect(0, 9312, '\p{Is_Blk=ocr}', "");
Expect(1, 9312, '\p{^Is_Blk=ocr}', "");
Expect(1, 9312, '\P{Is_Blk=ocr}', "");
Expect(0, 9312, '\P{^Is_Blk=ocr}', "");
Expect(1, 9311, '\p{Is_Blk=-OCR}', "");
Expect(0, 9311, '\p{^Is_Blk=-OCR}', "");
Expect(0, 9311, '\P{Is_Blk=-OCR}', "");
Expect(1, 9311, '\P{^Is_Blk=-OCR}', "");
Expect(0, 9312, '\p{Is_Blk=-OCR}', "");
Expect(1, 9312, '\p{^Is_Blk=-OCR}', "");
Expect(1, 9312, '\P{Is_Blk=-OCR}', "");
Expect(0, 9312, '\P{^Is_Blk=-OCR}', "");
Error('\p{Block= /a/Ancient_Symbols}');
Error('\P{Block= /a/Ancient_Symbols}');
Expect(1, 65999, '\p{Block=ancientsymbols}', "");
Expect(0, 65999, '\p{^Block=ancientsymbols}', "");
Expect(0, 65999, '\P{Block=ancientsymbols}', "");
Expect(1, 65999, '\P{^Block=ancientsymbols}', "");
Expect(0, 66000, '\p{Block=ancientsymbols}', "");
Expect(1, 66000, '\p{^Block=ancientsymbols}', "");
Expect(1, 66000, '\P{Block=ancientsymbols}', "");
Expect(0, 66000, '\P{^Block=ancientsymbols}', "");
Expect(1, 65999, '\p{Block=_-ANCIENT_Symbols}', "");
Expect(0, 65999, '\p{^Block=_-ANCIENT_Symbols}', "");
Expect(0, 65999, '\P{Block=_-ANCIENT_Symbols}', "");
Expect(1, 65999, '\P{^Block=_-ANCIENT_Symbols}', "");
Expect(0, 66000, '\p{Block=_-ANCIENT_Symbols}', "");
Expect(1, 66000, '\p{^Block=_-ANCIENT_Symbols}', "");
Expect(1, 66000, '\P{Block=_-ANCIENT_Symbols}', "");
Expect(0, 66000, '\P{^Block=_-ANCIENT_Symbols}', "");
Error('\p{Blk: 	/a/ANCIENT_SYMBOLS}');
Error('\P{Blk: 	/a/ANCIENT_SYMBOLS}');
Expect(1, 65999, '\p{Blk:	ancientsymbols}', "");
Expect(0, 65999, '\p{^Blk:	ancientsymbols}', "");
Expect(0, 65999, '\P{Blk:	ancientsymbols}', "");
Expect(1, 65999, '\P{^Blk:	ancientsymbols}', "");
Expect(0, 66000, '\p{Blk:	ancientsymbols}', "");
Expect(1, 66000, '\p{^Blk:	ancientsymbols}', "");
Expect(1, 66000, '\P{Blk:	ancientsymbols}', "");
Expect(0, 66000, '\P{^Blk:	ancientsymbols}', "");
Expect(1, 65999, '\p{Blk=Ancient_Symbols}', "");
Expect(0, 65999, '\p{^Blk=Ancient_Symbols}', "");
Expect(0, 65999, '\P{Blk=Ancient_Symbols}', "");
Expect(1, 65999, '\P{^Blk=Ancient_Symbols}', "");
Expect(0, 66000, '\p{Blk=Ancient_Symbols}', "");
Expect(1, 66000, '\p{^Blk=Ancient_Symbols}', "");
Expect(1, 66000, '\P{Blk=Ancient_Symbols}', "");
Expect(0, 66000, '\P{^Blk=Ancient_Symbols}', "");
Error('\p{Is_Block=:=		Ancient_SYMBOLS}');
Error('\P{Is_Block=:=		Ancient_SYMBOLS}');
Expect(1, 65999, '\p{Is_Block=ancientsymbols}', "");
Expect(0, 65999, '\p{^Is_Block=ancientsymbols}', "");
Expect(0, 65999, '\P{Is_Block=ancientsymbols}', "");
Expect(1, 65999, '\P{^Is_Block=ancientsymbols}', "");
Expect(0, 66000, '\p{Is_Block=ancientsymbols}', "");
Expect(1, 66000, '\p{^Is_Block=ancientsymbols}', "");
Expect(1, 66000, '\P{Is_Block=ancientsymbols}', "");
Expect(0, 66000, '\P{^Is_Block=ancientsymbols}', "");
Expect(1, 65999, '\p{Is_Block=_ Ancient_Symbols}', "");
Expect(0, 65999, '\p{^Is_Block=_ Ancient_Symbols}', "");
Expect(0, 65999, '\P{Is_Block=_ Ancient_Symbols}', "");
Expect(1, 65999, '\P{^Is_Block=_ Ancient_Symbols}', "");
Expect(0, 66000, '\p{Is_Block=_ Ancient_Symbols}', "");
Expect(1, 66000, '\p{^Is_Block=_ Ancient_Symbols}', "");
Expect(1, 66000, '\P{Is_Block=_ Ancient_Symbols}', "");
Expect(0, 66000, '\P{^Is_Block=_ Ancient_Symbols}', "");
Error('\p{Is_Blk=ANCIENT_SYMBOLS:=}');
Error('\P{Is_Blk=ANCIENT_SYMBOLS:=}');
Expect(1, 65999, '\p{Is_Blk=ancientsymbols}', "");
Expect(0, 65999, '\p{^Is_Blk=ancientsymbols}', "");
Expect(0, 65999, '\P{Is_Blk=ancientsymbols}', "");
Expect(1, 65999, '\P{^Is_Blk=ancientsymbols}', "");
Expect(0, 66000, '\p{Is_Blk=ancientsymbols}', "");
Expect(1, 66000, '\p{^Is_Blk=ancientsymbols}', "");
Expect(1, 66000, '\P{Is_Blk=ancientsymbols}', "");
Expect(0, 66000, '\P{^Is_Blk=ancientsymbols}', "");
Expect(1, 65999, '\p{Is_Blk=	ancient_SYMBOLS}', "");
Expect(0, 65999, '\p{^Is_Blk=	ancient_SYMBOLS}', "");
Expect(0, 65999, '\P{Is_Blk=	ancient_SYMBOLS}', "");
Expect(1, 65999, '\P{^Is_Blk=	ancient_SYMBOLS}', "");
Expect(0, 66000, '\p{Is_Blk=	ancient_SYMBOLS}', "");
Expect(1, 66000, '\p{^Is_Blk=	ancient_SYMBOLS}', "");
Expect(1, 66000, '\P{Is_Blk=	ancient_SYMBOLS}', "");
Expect(0, 66000, '\P{^Is_Blk=	ancient_SYMBOLS}', "");
Error('\p{Block=-:=ALPHABETIC_Presentation_FORMS}');
Error('\P{Block=-:=ALPHABETIC_Presentation_FORMS}');
Expect(1, 64335, '\p{Block=alphabeticpresentationforms}', "");
Expect(0, 64335, '\p{^Block=alphabeticpresentationforms}', "");
Expect(0, 64335, '\P{Block=alphabeticpresentationforms}', "");
Expect(1, 64335, '\P{^Block=alphabeticpresentationforms}', "");
Expect(0, 64336, '\p{Block=alphabeticpresentationforms}', "");
Expect(1, 64336, '\p{^Block=alphabeticpresentationforms}', "");
Expect(1, 64336, '\P{Block=alphabeticpresentationforms}', "");
Expect(0, 64336, '\P{^Block=alphabeticpresentationforms}', "");
Expect(1, 64335, '\p{Block= ALPHABETIC_Presentation_Forms}', "");
Expect(0, 64335, '\p{^Block= ALPHABETIC_Presentation_Forms}', "");
Expect(0, 64335, '\P{Block= ALPHABETIC_Presentation_Forms}', "");
Expect(1, 64335, '\P{^Block= ALPHABETIC_Presentation_Forms}', "");
Expect(0, 64336, '\p{Block= ALPHABETIC_Presentation_Forms}', "");
Expect(1, 64336, '\p{^Block= ALPHABETIC_Presentation_Forms}', "");
Expect(1, 64336, '\P{Block= ALPHABETIC_Presentation_Forms}', "");
Expect(0, 64336, '\P{^Block= ALPHABETIC_Presentation_Forms}', "");
Error('\p{Blk=-/a/alphabetic_PF}');
Error('\P{Blk=-/a/alphabetic_PF}');
Expect(1, 64335, '\p{Blk=alphabeticpf}', "");
Expect(0, 64335, '\p{^Blk=alphabeticpf}', "");
Expect(0, 64335, '\P{Blk=alphabeticpf}', "");
Expect(1, 64335, '\P{^Blk=alphabeticpf}', "");
Expect(0, 64336, '\p{Blk=alphabeticpf}', "");
Expect(1, 64336, '\p{^Blk=alphabeticpf}', "");
Expect(1, 64336, '\P{Blk=alphabeticpf}', "");
Expect(0, 64336, '\P{^Blk=alphabeticpf}', "");
Expect(1, 64335, '\p{Blk=_ ALPHABETIC_PF}', "");
Expect(0, 64335, '\p{^Blk=_ ALPHABETIC_PF}', "");
Expect(0, 64335, '\P{Blk=_ ALPHABETIC_PF}', "");
Expect(1, 64335, '\P{^Blk=_ ALPHABETIC_PF}', "");
Expect(0, 64336, '\p{Blk=_ ALPHABETIC_PF}', "");
Expect(1, 64336, '\p{^Blk=_ ALPHABETIC_PF}', "");
Expect(1, 64336, '\P{Blk=_ ALPHABETIC_PF}', "");
Expect(0, 64336, '\P{^Blk=_ ALPHABETIC_PF}', "");
Error('\p{Is_Block=:=	_Alphabetic_PRESENTATION_Forms}');
Error('\P{Is_Block=:=	_Alphabetic_PRESENTATION_Forms}');
Expect(1, 64335, '\p{Is_Block=alphabeticpresentationforms}', "");
Expect(0, 64335, '\p{^Is_Block=alphabeticpresentationforms}', "");
Expect(0, 64335, '\P{Is_Block=alphabeticpresentationforms}', "");
Expect(1, 64335, '\P{^Is_Block=alphabeticpresentationforms}', "");
Expect(0, 64336, '\p{Is_Block=alphabeticpresentationforms}', "");
Expect(1, 64336, '\p{^Is_Block=alphabeticpresentationforms}', "");
Expect(1, 64336, '\P{Is_Block=alphabeticpresentationforms}', "");
Expect(0, 64336, '\P{^Is_Block=alphabeticpresentationforms}', "");
Expect(1, 64335, '\p{Is_Block: -_ALPHABETIC_PRESENTATION_Forms}', "");
Expect(0, 64335, '\p{^Is_Block: -_ALPHABETIC_PRESENTATION_Forms}', "");
Expect(0, 64335, '\P{Is_Block: -_ALPHABETIC_PRESENTATION_Forms}', "");
Expect(1, 64335, '\P{^Is_Block: -_ALPHABETIC_PRESENTATION_Forms}', "");
Expect(0, 64336, '\p{Is_Block: -_ALPHABETIC_PRESENTATION_Forms}', "");
Expect(1, 64336, '\p{^Is_Block: -_ALPHABETIC_PRESENTATION_Forms}', "");
Expect(1, 64336, '\P{Is_Block: -_ALPHABETIC_PRESENTATION_Forms}', "");
Expect(0, 64336, '\P{^Is_Block: -_ALPHABETIC_PRESENTATION_Forms}', "");
Error('\p{Is_Blk=:=_	Alphabetic_PF}');
Error('\P{Is_Blk=:=_	Alphabetic_PF}');
Expect(1, 64335, '\p{Is_Blk=alphabeticpf}', "");
Expect(0, 64335, '\p{^Is_Blk=alphabeticpf}', "");
Expect(0, 64335, '\P{Is_Blk=alphabeticpf}', "");
Expect(1, 64335, '\P{^Is_Blk=alphabeticpf}', "");
Expect(0, 64336, '\p{Is_Blk=alphabeticpf}', "");
Expect(1, 64336, '\p{^Is_Blk=alphabeticpf}', "");
Expect(1, 64336, '\P{Is_Blk=alphabeticpf}', "");
Expect(0, 64336, '\P{^Is_Blk=alphabeticpf}', "");
Expect(1, 64335, '\p{Is_Blk=--Alphabetic_PF}', "");
Expect(0, 64335, '\p{^Is_Blk=--Alphabetic_PF}', "");
Expect(0, 64335, '\P{Is_Blk=--Alphabetic_PF}', "");
Expect(1, 64335, '\P{^Is_Blk=--Alphabetic_PF}', "");
Expect(0, 64336, '\p{Is_Blk=--Alphabetic_PF}', "");
Expect(1, 64336, '\p{^Is_Blk=--Alphabetic_PF}', "");
Expect(1, 64336, '\P{Is_Blk=--Alphabetic_PF}', "");
Expect(0, 64336, '\P{^Is_Blk=--Alphabetic_PF}', "");
Error('\p{Block=Sora_Sompeng/a/}');
Error('\P{Block=Sora_Sompeng/a/}');
Expect(1, 69887, '\p{Block=sorasompeng}', "");
Expect(0, 69887, '\p{^Block=sorasompeng}', "");
Expect(0, 69887, '\P{Block=sorasompeng}', "");
Expect(1, 69887, '\P{^Block=sorasompeng}', "");
Expect(0, 69888, '\p{Block=sorasompeng}', "");
Expect(1, 69888, '\p{^Block=sorasompeng}', "");
Expect(1, 69888, '\P{Block=sorasompeng}', "");
Expect(0, 69888, '\P{^Block=sorasompeng}', "");
Expect(1, 69887, '\p{Block= sora_sompeng}', "");
Expect(0, 69887, '\p{^Block= sora_sompeng}', "");
Expect(0, 69887, '\P{Block= sora_sompeng}', "");
Expect(1, 69887, '\P{^Block= sora_sompeng}', "");
Expect(0, 69888, '\p{Block= sora_sompeng}', "");
Expect(1, 69888, '\p{^Block= sora_sompeng}', "");
Expect(1, 69888, '\P{Block= sora_sompeng}', "");
Expect(0, 69888, '\P{^Block= sora_sompeng}', "");
Error('\p{Blk=_-Sora_sompeng/a/}');
Error('\P{Blk=_-Sora_sompeng/a/}');
Expect(1, 69887, '\p{Blk=sorasompeng}', "");
Expect(0, 69887, '\p{^Blk=sorasompeng}', "");
Expect(0, 69887, '\P{Blk=sorasompeng}', "");
Expect(1, 69887, '\P{^Blk=sorasompeng}', "");
Expect(0, 69888, '\p{Blk=sorasompeng}', "");
Expect(1, 69888, '\p{^Blk=sorasompeng}', "");
Expect(1, 69888, '\P{Blk=sorasompeng}', "");
Expect(0, 69888, '\P{^Blk=sorasompeng}', "");
Expect(1, 69887, '\p{Blk=__sora_sompeng}', "");
Expect(0, 69887, '\p{^Blk=__sora_sompeng}', "");
Expect(0, 69887, '\P{Blk=__sora_sompeng}', "");
Expect(1, 69887, '\P{^Blk=__sora_sompeng}', "");
Expect(0, 69888, '\p{Blk=__sora_sompeng}', "");
Expect(1, 69888, '\p{^Blk=__sora_sompeng}', "");
Expect(1, 69888, '\P{Blk=__sora_sompeng}', "");
Expect(0, 69888, '\P{^Blk=__sora_sompeng}', "");
Error('\p{Is_Block= _SORA_Sompeng:=}');
Error('\P{Is_Block= _SORA_Sompeng:=}');
Expect(1, 69887, '\p{Is_Block=sorasompeng}', "");
Expect(0, 69887, '\p{^Is_Block=sorasompeng}', "");
Expect(0, 69887, '\P{Is_Block=sorasompeng}', "");
Expect(1, 69887, '\P{^Is_Block=sorasompeng}', "");
Expect(0, 69888, '\p{Is_Block=sorasompeng}', "");
Expect(1, 69888, '\p{^Is_Block=sorasompeng}', "");
Expect(1, 69888, '\P{Is_Block=sorasompeng}', "");
Expect(0, 69888, '\P{^Is_Block=sorasompeng}', "");
Expect(1, 69887, '\p{Is_Block=__sora_Sompeng}', "");
Expect(0, 69887, '\p{^Is_Block=__sora_Sompeng}', "");
Expect(0, 69887, '\P{Is_Block=__sora_Sompeng}', "");
Expect(1, 69887, '\P{^Is_Block=__sora_Sompeng}', "");
Expect(0, 69888, '\p{Is_Block=__sora_Sompeng}', "");
Expect(1, 69888, '\p{^Is_Block=__sora_Sompeng}', "");
Expect(1, 69888, '\P{Is_Block=__sora_Sompeng}', "");
Expect(0, 69888, '\P{^Is_Block=__sora_Sompeng}', "");
Error('\p{Is_Blk=/a/ _SORA_SOMPENG}');
Error('\P{Is_Blk=/a/ _SORA_SOMPENG}');
Expect(1, 69887, '\p{Is_Blk=sorasompeng}', "");
Expect(0, 69887, '\p{^Is_Blk=sorasompeng}', "");
Expect(0, 69887, '\P{Is_Blk=sorasompeng}', "");
Expect(1, 69887, '\P{^Is_Blk=sorasompeng}', "");
Expect(0, 69888, '\p{Is_Blk=sorasompeng}', "");
Expect(1, 69888, '\p{^Is_Blk=sorasompeng}', "");
Expect(1, 69888, '\P{Is_Blk=sorasompeng}', "");
Expect(0, 69888, '\P{^Is_Blk=sorasompeng}', "");
Expect(1, 69887, '\p{Is_Blk=_sora_sompeng}', "");
Expect(0, 69887, '\p{^Is_Blk=_sora_sompeng}', "");
Expect(0, 69887, '\P{Is_Blk=_sora_sompeng}', "");
Expect(1, 69887, '\P{^Is_Blk=_sora_sompeng}', "");
Expect(0, 69888, '\p{Is_Blk=_sora_sompeng}', "");
Expect(1, 69888, '\p{^Is_Blk=_sora_sompeng}', "");
Expect(1, 69888, '\P{Is_Blk=_sora_sompeng}', "");
Expect(0, 69888, '\P{^Is_Blk=_sora_sompeng}', "");
Error('\p{Block=/a/	 DEVANAGARI}');
Error('\P{Block=/a/	 DEVANAGARI}');
Expect(1, 2431, '\p{Block=devanagari}', "");
Expect(0, 2431, '\p{^Block=devanagari}', "");
Expect(0, 2431, '\P{Block=devanagari}', "");
Expect(1, 2431, '\P{^Block=devanagari}', "");
Expect(0, 2432, '\p{Block=devanagari}', "");
Expect(1, 2432, '\p{^Block=devanagari}', "");
Expect(1, 2432, '\P{Block=devanagari}', "");
Expect(0, 2432, '\P{^Block=devanagari}', "");
Expect(1, 2431, '\p{Block=	_devanagari}', "");
Expect(0, 2431, '\p{^Block=	_devanagari}', "");
Expect(0, 2431, '\P{Block=	_devanagari}', "");
Expect(1, 2431, '\P{^Block=	_devanagari}', "");
Expect(0, 2432, '\p{Block=	_devanagari}', "");
Expect(1, 2432, '\p{^Block=	_devanagari}', "");
Expect(1, 2432, '\P{Block=	_devanagari}', "");
Expect(0, 2432, '\P{^Block=	_devanagari}', "");
Error('\p{Blk=	Devanagari:=}');
Error('\P{Blk=	Devanagari:=}');
Expect(1, 2431, '\p{Blk=devanagari}', "");
Expect(0, 2431, '\p{^Blk=devanagari}', "");
Expect(0, 2431, '\P{Blk=devanagari}', "");
Expect(1, 2431, '\P{^Blk=devanagari}', "");
Expect(0, 2432, '\p{Blk=devanagari}', "");
Expect(1, 2432, '\p{^Blk=devanagari}', "");
Expect(1, 2432, '\P{Blk=devanagari}', "");
Expect(0, 2432, '\P{^Blk=devanagari}', "");
Expect(1, 2431, '\p{Blk:   	-DEVANAGARI}', "");
Expect(0, 2431, '\p{^Blk:   	-DEVANAGARI}', "");
Expect(0, 2431, '\P{Blk:   	-DEVANAGARI}', "");
Expect(1, 2431, '\P{^Blk:   	-DEVANAGARI}', "");
Expect(0, 2432, '\p{Blk:   	-DEVANAGARI}', "");
Expect(1, 2432, '\p{^Blk:   	-DEVANAGARI}', "");
Expect(1, 2432, '\P{Blk:   	-DEVANAGARI}', "");
Expect(0, 2432, '\P{^Blk:   	-DEVANAGARI}', "");
Error('\p{Is_Block=:=_Devanagari}');
Error('\P{Is_Block=:=_Devanagari}');
Expect(1, 2431, '\p{Is_Block=devanagari}', "");
Expect(0, 2431, '\p{^Is_Block=devanagari}', "");
Expect(0, 2431, '\P{Is_Block=devanagari}', "");
Expect(1, 2431, '\P{^Is_Block=devanagari}', "");
Expect(0, 2432, '\p{Is_Block=devanagari}', "");
Expect(1, 2432, '\p{^Is_Block=devanagari}', "");
Expect(1, 2432, '\P{Is_Block=devanagari}', "");
Expect(0, 2432, '\P{^Is_Block=devanagari}', "");
Expect(1, 2431, '\p{Is_Block=		Devanagari}', "");
Expect(0, 2431, '\p{^Is_Block=		Devanagari}', "");
Expect(0, 2431, '\P{Is_Block=		Devanagari}', "");
Expect(1, 2431, '\P{^Is_Block=		Devanagari}', "");
Expect(0, 2432, '\p{Is_Block=		Devanagari}', "");
Expect(1, 2432, '\p{^Is_Block=		Devanagari}', "");
Expect(1, 2432, '\P{Is_Block=		Devanagari}', "");
Expect(0, 2432, '\P{^Is_Block=		Devanagari}', "");
Error('\p{Is_Blk=/a/ DEVANAGARI}');
Error('\P{Is_Blk=/a/ DEVANAGARI}');
Expect(1, 2431, '\p{Is_Blk=devanagari}', "");
Expect(0, 2431, '\p{^Is_Blk=devanagari}', "");
Expect(0, 2431, '\P{Is_Blk=devanagari}', "");
Expect(1, 2431, '\P{^Is_Blk=devanagari}', "");
Expect(0, 2432, '\p{Is_Blk=devanagari}', "");
Expect(1, 2432, '\p{^Is_Blk=devanagari}', "");
Expect(1, 2432, '\P{Is_Blk=devanagari}', "");
Expect(0, 2432, '\P{^Is_Blk=devanagari}', "");
Expect(1, 2431, '\p{Is_Blk= devanagari}', "");
Expect(0, 2431, '\p{^Is_Blk= devanagari}', "");
Expect(0, 2431, '\P{Is_Blk= devanagari}', "");
Expect(1, 2431, '\P{^Is_Blk= devanagari}', "");
Expect(0, 2432, '\p{Is_Blk= devanagari}', "");
Expect(1, 2432, '\p{^Is_Blk= devanagari}', "");
Expect(1, 2432, '\P{Is_Blk= devanagari}', "");
Expect(0, 2432, '\P{^Is_Blk= devanagari}', "");
Error('\p{Block= 	buginese/a/}');
Error('\P{Block= 	buginese/a/}');
Expect(1, 6687, '\p{Block=buginese}', "");
Expect(0, 6687, '\p{^Block=buginese}', "");
Expect(0, 6687, '\P{Block=buginese}', "");
Expect(1, 6687, '\P{^Block=buginese}', "");
Expect(0, 6688, '\p{Block=buginese}', "");
Expect(1, 6688, '\p{^Block=buginese}', "");
Expect(1, 6688, '\P{Block=buginese}', "");
Expect(0, 6688, '\P{^Block=buginese}', "");
Expect(1, 6687, '\p{Block=	_BUGINESE}', "");
Expect(0, 6687, '\p{^Block=	_BUGINESE}', "");
Expect(0, 6687, '\P{Block=	_BUGINESE}', "");
Expect(1, 6687, '\P{^Block=	_BUGINESE}', "");
Expect(0, 6688, '\p{Block=	_BUGINESE}', "");
Expect(1, 6688, '\p{^Block=	_BUGINESE}', "");
Expect(1, 6688, '\P{Block=	_BUGINESE}', "");
Expect(0, 6688, '\P{^Block=	_BUGINESE}', "");
Error('\p{Blk=/a/-Buginese}');
Error('\P{Blk=/a/-Buginese}');
Expect(1, 6687, '\p{Blk=buginese}', "");
Expect(0, 6687, '\p{^Blk=buginese}', "");
Expect(0, 6687, '\P{Blk=buginese}', "");
Expect(1, 6687, '\P{^Blk=buginese}', "");
Expect(0, 6688, '\p{Blk=buginese}', "");
Expect(1, 6688, '\p{^Blk=buginese}', "");
Expect(1, 6688, '\P{Blk=buginese}', "");
Expect(0, 6688, '\P{^Blk=buginese}', "");
Expect(1, 6687, '\p{Blk=		Buginese}', "");
Expect(0, 6687, '\p{^Blk=		Buginese}', "");
Expect(0, 6687, '\P{Blk=		Buginese}', "");
Expect(1, 6687, '\P{^Blk=		Buginese}', "");
Expect(0, 6688, '\p{Blk=		Buginese}', "");
Expect(1, 6688, '\p{^Blk=		Buginese}', "");
Expect(1, 6688, '\P{Blk=		Buginese}', "");
Expect(0, 6688, '\P{^Blk=		Buginese}', "");
Error('\p{Is_Block=/a/  buginese}');
Error('\P{Is_Block=/a/  buginese}');
Expect(1, 6687, '\p{Is_Block=buginese}', "");
Expect(0, 6687, '\p{^Is_Block=buginese}', "");
Expect(0, 6687, '\P{Is_Block=buginese}', "");
Expect(1, 6687, '\P{^Is_Block=buginese}', "");
Expect(0, 6688, '\p{Is_Block=buginese}', "");
Expect(1, 6688, '\p{^Is_Block=buginese}', "");
Expect(1, 6688, '\P{Is_Block=buginese}', "");
Expect(0, 6688, '\P{^Is_Block=buginese}', "");
Expect(1, 6687, '\p{Is_Block=		BUGINESE}', "");
Expect(0, 6687, '\p{^Is_Block=		BUGINESE}', "");
Expect(0, 6687, '\P{Is_Block=		BUGINESE}', "");
Expect(1, 6687, '\P{^Is_Block=		BUGINESE}', "");
Expect(0, 6688, '\p{Is_Block=		BUGINESE}', "");
Expect(1, 6688, '\p{^Is_Block=		BUGINESE}', "");
Expect(1, 6688, '\P{Is_Block=		BUGINESE}', "");
Expect(0, 6688, '\P{^Is_Block=		BUGINESE}', "");
Error('\p{Is_Blk=-Buginese/a/}');
Error('\P{Is_Blk=-Buginese/a/}');
Expect(1, 6687, '\p{Is_Blk=buginese}', "");
Expect(0, 6687, '\p{^Is_Blk=buginese}', "");
Expect(0, 6687, '\P{Is_Blk=buginese}', "");
Expect(1, 6687, '\P{^Is_Blk=buginese}', "");
Expect(0, 6688, '\p{Is_Blk=buginese}', "");
Expect(1, 6688, '\p{^Is_Blk=buginese}', "");
Expect(1, 6688, '\P{Is_Blk=buginese}', "");
Expect(0, 6688, '\P{^Is_Blk=buginese}', "");
Expect(1, 6687, '\p{Is_Blk=	buginese}', "");
Expect(0, 6687, '\p{^Is_Blk=	buginese}', "");
Expect(0, 6687, '\P{Is_Blk=	buginese}', "");
Expect(1, 6687, '\P{^Is_Blk=	buginese}', "");
Expect(0, 6688, '\p{Is_Blk=	buginese}', "");
Expect(1, 6688, '\p{^Is_Blk=	buginese}', "");
Expect(1, 6688, '\P{Is_Blk=	buginese}', "");
Expect(0, 6688, '\P{^Is_Blk=	buginese}', "");
Error('\p{Block=_/a/Unified_canadian_Aboriginal_Syllabics_Extended}');
Error('\P{Block=_/a/Unified_canadian_Aboriginal_Syllabics_Extended}');
Expect(1, 6399, '\p{Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\p{^Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\P{Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\P{^Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\p{Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\p{^Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\P{Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\P{^Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\p{Block= _Unified_canadian_Aboriginal_Syllabics_Extended}', "");
Expect(0, 6399, '\p{^Block= _Unified_canadian_Aboriginal_Syllabics_Extended}', "");
Expect(0, 6399, '\P{Block= _Unified_canadian_Aboriginal_Syllabics_Extended}', "");
Expect(1, 6399, '\P{^Block= _Unified_canadian_Aboriginal_Syllabics_Extended}', "");
Expect(0, 6400, '\p{Block= _Unified_canadian_Aboriginal_Syllabics_Extended}', "");
Expect(1, 6400, '\p{^Block= _Unified_canadian_Aboriginal_Syllabics_Extended}', "");
Expect(1, 6400, '\P{Block= _Unified_canadian_Aboriginal_Syllabics_Extended}', "");
Expect(0, 6400, '\P{^Block= _Unified_canadian_Aboriginal_Syllabics_Extended}', "");
Error('\p{Blk=/a/ -UCAS_Ext}');
Error('\P{Blk=/a/ -UCAS_Ext}');
Expect(1, 6399, '\p{Blk=ucasext}', "");
Expect(0, 6399, '\p{^Blk=ucasext}', "");
Expect(0, 6399, '\P{Blk=ucasext}', "");
Expect(1, 6399, '\P{^Blk=ucasext}', "");
Expect(0, 6400, '\p{Blk=ucasext}', "");
Expect(1, 6400, '\p{^Blk=ucasext}', "");
Expect(1, 6400, '\P{Blk=ucasext}', "");
Expect(0, 6400, '\P{^Blk=ucasext}', "");
Expect(1, 6399, '\p{Blk=	_UCAS_Ext}', "");
Expect(0, 6399, '\p{^Blk=	_UCAS_Ext}', "");
Expect(0, 6399, '\P{Blk=	_UCAS_Ext}', "");
Expect(1, 6399, '\P{^Blk=	_UCAS_Ext}', "");
Expect(0, 6400, '\p{Blk=	_UCAS_Ext}', "");
Expect(1, 6400, '\p{^Blk=	_UCAS_Ext}', "");
Expect(1, 6400, '\P{Blk=	_UCAS_Ext}', "");
Expect(0, 6400, '\P{^Blk=	_UCAS_Ext}', "");
Error('\p{Is_Block=	:=unified_CANADIAN_ABORIGINAL_SYLLABICS_EXTENDED}');
Error('\P{Is_Block=	:=unified_CANADIAN_ABORIGINAL_SYLLABICS_EXTENDED}');
Expect(1, 6399, '\p{Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\p{^Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6399, '\P{Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\P{^Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\p{Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\p{^Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6400, '\P{Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(0, 6400, '\P{^Is_Block=unifiedcanadianaboriginalsyllabicsextended}', "");
Expect(1, 6399, '\p{Is_Block=	unified_CANADIAN_Aboriginal_Syllabics_Extended}', "");
Expect(0, 6399, '\p{^Is_Block=	unified_CANADIAN_Aboriginal_Syllabics_Extended}', "");
Expect(0, 6399, '\P{Is_Block=	unified_CANADIAN_Aboriginal_Syllabics_Extended}', "");
Expect(1, 6399, '\P{^Is_Block=	unified_CANADIAN_Aboriginal_Syllabics_Extended}', "");
Expect(0, 6400, '\p{Is_Block=	unified_CANADIAN_Aboriginal_Syllabics_Extended}', "");
Expect(1, 6400, '\p{^Is_Block=	unified_CANADIAN_Aboriginal_Syllabics_Extended}', "");
Expect(1, 6400, '\P{Is_Block=	unified_CANADIAN_Aboriginal_Syllabics_Extended}', "");
Expect(0, 6400, '\P{^Is_Block=	unified_CANADIAN_Aboriginal_Syllabics_Extended}', "");
Error('\p{Is_Blk=:= _UCAS_ext}');
Error('\P{Is_Blk=:= _UCAS_ext}');
Expect(1, 6399, '\p{Is_Blk: ucasext}', "");
Expect(0, 6399, '\p{^Is_Blk: ucasext}', "");
Expect(0, 6399, '\P{Is_Blk: ucasext}', "");
Expect(1, 6399, '\P{^Is_Blk: ucasext}', "");
Expect(0, 6400, '\p{Is_Blk: ucasext}', "");
Expect(1, 6400, '\p{^Is_Blk: ucasext}', "");
Expect(1, 6400, '\P{Is_Blk: ucasext}', "");
Expect(0, 6400, '\P{^Is_Blk: ucasext}', "");
Expect(1, 6399, '\p{Is_Blk=	_UCAS_EXT}', "");
Expect(0, 6399, '\p{^Is_Blk=	_UCAS_EXT}', "");
Expect(0, 6399, '\P{Is_Blk=	_UCAS_EXT}', "");
Expect(1, 6399, '\P{^Is_Blk=	_UCAS_EXT}', "");
Expect(0, 6400, '\p{Is_Blk=	_UCAS_EXT}', "");
Expect(1, 6400, '\p{^Is_Blk=	_UCAS_EXT}', "");
Expect(1, 6400, '\P{Is_Blk=	_UCAS_EXT}', "");
Expect(0, 6400, '\P{^Is_Blk=	_UCAS_EXT}', "");
Error('\p{Block=-Arabic_PRESENTATION_forms_B:=}');
Error('\P{Block=-Arabic_PRESENTATION_forms_B:=}');
Expect(1, 65279, '\p{Block=arabicpresentationformsb}', "");
Expect(0, 65279, '\p{^Block=arabicpresentationformsb}', "");
Expect(0, 65279, '\P{Block=arabicpresentationformsb}', "");
Expect(1, 65279, '\P{^Block=arabicpresentationformsb}', "");
Expect(0, 65280, '\p{Block=arabicpresentationformsb}', "");
Expect(1, 65280, '\p{^Block=arabicpresentationformsb}', "");
Expect(1, 65280, '\P{Block=arabicpresentationformsb}', "");
Expect(0, 65280, '\P{^Block=arabicpresentationformsb}', "");
Expect(1, 65279, '\p{Block: _	Arabic_PRESENTATION_FORMS_B}', "");
Expect(0, 65279, '\p{^Block: _	Arabic_PRESENTATION_FORMS_B}', "");
Expect(0, 65279, '\P{Block: _	Arabic_PRESENTATION_FORMS_B}', "");
Expect(1, 65279, '\P{^Block: _	Arabic_PRESENTATION_FORMS_B}', "");
Expect(0, 65280, '\p{Block: _	Arabic_PRESENTATION_FORMS_B}', "");
Expect(1, 65280, '\p{^Block: _	Arabic_PRESENTATION_FORMS_B}', "");
Expect(1, 65280, '\P{Block: _	Arabic_PRESENTATION_FORMS_B}', "");
Expect(0, 65280, '\P{^Block: _	Arabic_PRESENTATION_FORMS_B}', "");
Error('\p{Blk=	 Arabic_PF_b:=}');
Error('\P{Blk=	 Arabic_PF_b:=}');
Expect(1, 65279, '\p{Blk=arabicpfb}', "");
Expect(0, 65279, '\p{^Blk=arabicpfb}', "");
Expect(0, 65279, '\P{Blk=arabicpfb}', "");
Expect(1, 65279, '\P{^Blk=arabicpfb}', "");
Expect(0, 65280, '\p{Blk=arabicpfb}', "");
Expect(1, 65280, '\p{^Blk=arabicpfb}', "");
Expect(1, 65280, '\P{Blk=arabicpfb}', "");
Expect(0, 65280, '\P{^Blk=arabicpfb}', "");
Expect(1, 65279, '\p{Blk=_	Arabic_PF_b}', "");
Expect(0, 65279, '\p{^Blk=_	Arabic_PF_b}', "");
Expect(0, 65279, '\P{Blk=_	Arabic_PF_b}', "");
Expect(1, 65279, '\P{^Blk=_	Arabic_PF_b}', "");
Expect(0, 65280, '\p{Blk=_	Arabic_PF_b}', "");
Expect(1, 65280, '\p{^Blk=_	Arabic_PF_b}', "");
Expect(1, 65280, '\P{Blk=_	Arabic_PF_b}', "");
Expect(0, 65280, '\P{^Blk=_	Arabic_PF_b}', "");
Error('\p{Is_Block=_arabic_Presentation_FORMS_B:=}');
Error('\P{Is_Block=_arabic_Presentation_FORMS_B:=}');
Expect(1, 65279, '\p{Is_Block=arabicpresentationformsb}', "");
Expect(0, 65279, '\p{^Is_Block=arabicpresentationformsb}', "");
Expect(0, 65279, '\P{Is_Block=arabicpresentationformsb}', "");
Expect(1, 65279, '\P{^Is_Block=arabicpresentationformsb}', "");
Expect(0, 65280, '\p{Is_Block=arabicpresentationformsb}', "");
Expect(1, 65280, '\p{^Is_Block=arabicpresentationformsb}', "");
Expect(1, 65280, '\P{Is_Block=arabicpresentationformsb}', "");
Expect(0, 65280, '\P{^Is_Block=arabicpresentationformsb}', "");
Expect(1, 65279, '\p{Is_Block=		Arabic_PRESENTATION_FORMS_B}', "");
Expect(0, 65279, '\p{^Is_Block=		Arabic_PRESENTATION_FORMS_B}', "");
Expect(0, 65279, '\P{Is_Block=		Arabic_PRESENTATION_FORMS_B}', "");
Expect(1, 65279, '\P{^Is_Block=		Arabic_PRESENTATION_FORMS_B}', "");
Expect(0, 65280, '\p{Is_Block=		Arabic_PRESENTATION_FORMS_B}', "");
Expect(1, 65280, '\p{^Is_Block=		Arabic_PRESENTATION_FORMS_B}', "");
Expect(1, 65280, '\P{Is_Block=		Arabic_PRESENTATION_FORMS_B}', "");
Expect(0, 65280, '\P{^Is_Block=		Arabic_PRESENTATION_FORMS_B}', "");
Error('\p{Is_Blk=:=-arabic_pf_B}');
Error('\P{Is_Blk=:=-arabic_pf_B}');
Expect(1, 65279, '\p{Is_Blk=arabicpfb}', "");
Expect(0, 65279, '\p{^Is_Blk=arabicpfb}', "");
Expect(0, 65279, '\P{Is_Blk=arabicpfb}', "");
Expect(1, 65279, '\P{^Is_Blk=arabicpfb}', "");
Expect(0, 65280, '\p{Is_Blk=arabicpfb}', "");
Expect(1, 65280, '\p{^Is_Blk=arabicpfb}', "");
Expect(1, 65280, '\P{Is_Blk=arabicpfb}', "");
Expect(0, 65280, '\P{^Is_Blk=arabicpfb}', "");
Expect(1, 65279, '\p{Is_Blk=  arabic_pf_B}', "");
Expect(0, 65279, '\p{^Is_Blk=  arabic_pf_B}', "");
Expect(0, 65279, '\P{Is_Blk=  arabic_pf_B}', "");
Expect(1, 65279, '\P{^Is_Blk=  arabic_pf_B}', "");
Expect(0, 65280, '\p{Is_Blk=  arabic_pf_B}', "");
Expect(1, 65280, '\p{^Is_Blk=  arabic_pf_B}', "");
Expect(1, 65280, '\P{Is_Blk=  arabic_pf_B}', "");
Expect(0, 65280, '\P{^Is_Blk=  arabic_pf_B}', "");
Error('\p{Block=	 Gurmukhi/a/}');
Error('\P{Block=	 Gurmukhi/a/}');
Expect(1, 2687, '\p{Block=gurmukhi}', "");
Expect(0, 2687, '\p{^Block=gurmukhi}', "");
Expect(0, 2687, '\P{Block=gurmukhi}', "");
Expect(1, 2687, '\P{^Block=gurmukhi}', "");
Expect(0, 2688, '\p{Block=gurmukhi}', "");
Expect(1, 2688, '\p{^Block=gurmukhi}', "");
Expect(1, 2688, '\P{Block=gurmukhi}', "");
Expect(0, 2688, '\P{^Block=gurmukhi}', "");
Expect(1, 2687, '\p{Block=-_gurmukhi}', "");
Expect(0, 2687, '\p{^Block=-_gurmukhi}', "");
Expect(0, 2687, '\P{Block=-_gurmukhi}', "");
Expect(1, 2687, '\P{^Block=-_gurmukhi}', "");
Expect(0, 2688, '\p{Block=-_gurmukhi}', "");
Expect(1, 2688, '\p{^Block=-_gurmukhi}', "");
Expect(1, 2688, '\P{Block=-_gurmukhi}', "");
Expect(0, 2688, '\P{^Block=-_gurmukhi}', "");
Error('\p{Blk= :=Gurmukhi}');
Error('\P{Blk= :=Gurmukhi}');
Expect(1, 2687, '\p{Blk=gurmukhi}', "");
Expect(0, 2687, '\p{^Blk=gurmukhi}', "");
Expect(0, 2687, '\P{Blk=gurmukhi}', "");
Expect(1, 2687, '\P{^Blk=gurmukhi}', "");
Expect(0, 2688, '\p{Blk=gurmukhi}', "");
Expect(1, 2688, '\p{^Blk=gurmukhi}', "");
Expect(1, 2688, '\P{Blk=gurmukhi}', "");
Expect(0, 2688, '\P{^Blk=gurmukhi}', "");
Expect(1, 2687, '\p{Blk=_gurmukhi}', "");
Expect(0, 2687, '\p{^Blk=_gurmukhi}', "");
Expect(0, 2687, '\P{Blk=_gurmukhi}', "");
Expect(1, 2687, '\P{^Blk=_gurmukhi}', "");
Expect(0, 2688, '\p{Blk=_gurmukhi}', "");
Expect(1, 2688, '\p{^Blk=_gurmukhi}', "");
Expect(1, 2688, '\P{Blk=_gurmukhi}', "");
Expect(0, 2688, '\P{^Blk=_gurmukhi}', "");
Error('\p{Is_Block=_ Gurmukhi:=}');
Error('\P{Is_Block=_ Gurmukhi:=}');
Expect(1, 2687, '\p{Is_Block=gurmukhi}', "");
Expect(0, 2687, '\p{^Is_Block=gurmukhi}', "");
Expect(0, 2687, '\P{Is_Block=gurmukhi}', "");
Expect(1, 2687, '\P{^Is_Block=gurmukhi}', "");
Expect(0, 2688, '\p{Is_Block=gurmukhi}', "");
Expect(1, 2688, '\p{^Is_Block=gurmukhi}', "");
Expect(1, 2688, '\P{Is_Block=gurmukhi}', "");
Expect(0, 2688, '\P{^Is_Block=gurmukhi}', "");
Expect(1, 2687, '\p{Is_Block=-	gurmukhi}', "");
Expect(0, 2687, '\p{^Is_Block=-	gurmukhi}', "");
Expect(0, 2687, '\P{Is_Block=-	gurmukhi}', "");
Expect(1, 2687, '\P{^Is_Block=-	gurmukhi}', "");
Expect(0, 2688, '\p{Is_Block=-	gurmukhi}', "");
Expect(1, 2688, '\p{^Is_Block=-	gurmukhi}', "");
Expect(1, 2688, '\P{Is_Block=-	gurmukhi}', "");
Expect(0, 2688, '\P{^Is_Block=-	gurmukhi}', "");
Error('\p{Is_Blk=:= 	GURMUKHI}');
Error('\P{Is_Blk=:= 	GURMUKHI}');
Expect(1, 2687, '\p{Is_Blk=gurmukhi}', "");
Expect(0, 2687, '\p{^Is_Blk=gurmukhi}', "");
Expect(0, 2687, '\P{Is_Blk=gurmukhi}', "");
Expect(1, 2687, '\P{^Is_Blk=gurmukhi}', "");
Expect(0, 2688, '\p{Is_Blk=gurmukhi}', "");
Expect(1, 2688, '\p{^Is_Blk=gurmukhi}', "");
Expect(1, 2688, '\P{Is_Blk=gurmukhi}', "");
Expect(0, 2688, '\P{^Is_Blk=gurmukhi}', "");
Expect(1, 2687, '\p{Is_Blk=_-gurmukhi}', "");
Expect(0, 2687, '\p{^Is_Blk=_-gurmukhi}', "");
Expect(0, 2687, '\P{Is_Blk=_-gurmukhi}', "");
Expect(1, 2687, '\P{^Is_Blk=_-gurmukhi}', "");
Expect(0, 2688, '\p{Is_Blk=_-gurmukhi}', "");
Expect(1, 2688, '\p{^Is_Blk=_-gurmukhi}', "");
Expect(1, 2688, '\P{Is_Blk=_-gurmukhi}', "");
Expect(0, 2688, '\P{^Is_Blk=_-gurmukhi}', "");
Error('\p{Block=-Greek_extended:=}');
Error('\P{Block=-Greek_extended:=}');
Expect(1, 8191, '\p{Block=greekextended}', "");
Expect(0, 8191, '\p{^Block=greekextended}', "");
Expect(0, 8191, '\P{Block=greekextended}', "");
Expect(1, 8191, '\P{^Block=greekextended}', "");
Expect(0, 8192, '\p{Block=greekextended}', "");
Expect(1, 8192, '\p{^Block=greekextended}', "");
Expect(1, 8192, '\P{Block=greekextended}', "");
Expect(0, 8192, '\P{^Block=greekextended}', "");
Expect(1, 8191, '\p{Block=_-greek_Extended}', "");
Expect(0, 8191, '\p{^Block=_-greek_Extended}', "");
Expect(0, 8191, '\P{Block=_-greek_Extended}', "");
Expect(1, 8191, '\P{^Block=_-greek_Extended}', "");
Expect(0, 8192, '\p{Block=_-greek_Extended}', "");
Expect(1, 8192, '\p{^Block=_-greek_Extended}', "");
Expect(1, 8192, '\P{Block=_-greek_Extended}', "");
Expect(0, 8192, '\P{^Block=_-greek_Extended}', "");
Error('\p{Blk=-:=Greek_ext}');
Error('\P{Blk=-:=Greek_ext}');
Expect(1, 8191, '\p{Blk=greekext}', "");
Expect(0, 8191, '\p{^Blk=greekext}', "");
Expect(0, 8191, '\P{Blk=greekext}', "");
Expect(1, 8191, '\P{^Blk=greekext}', "");
Expect(0, 8192, '\p{Blk=greekext}', "");
Expect(1, 8192, '\p{^Blk=greekext}', "");
Expect(1, 8192, '\P{Blk=greekext}', "");
Expect(0, 8192, '\P{^Blk=greekext}', "");
Expect(1, 8191, '\p{Blk=_-GREEK_EXT}', "");
Expect(0, 8191, '\p{^Blk=_-GREEK_EXT}', "");
Expect(0, 8191, '\P{Blk=_-GREEK_EXT}', "");
Expect(1, 8191, '\P{^Blk=_-GREEK_EXT}', "");
Expect(0, 8192, '\p{Blk=_-GREEK_EXT}', "");
Expect(1, 8192, '\p{^Blk=_-GREEK_EXT}', "");
Expect(1, 8192, '\P{Blk=_-GREEK_EXT}', "");
Expect(0, 8192, '\P{^Blk=_-GREEK_EXT}', "");
Error('\p{Is_Block=_/a/Greek_extended}');
Error('\P{Is_Block=_/a/Greek_extended}');
Expect(1, 8191, '\p{Is_Block=greekextended}', "");
Expect(0, 8191, '\p{^Is_Block=greekextended}', "");
Expect(0, 8191, '\P{Is_Block=greekextended}', "");
Expect(1, 8191, '\P{^Is_Block=greekextended}', "");
Expect(0, 8192, '\p{Is_Block=greekextended}', "");
Expect(1, 8192, '\p{^Is_Block=greekextended}', "");
Expect(1, 8192, '\P{Is_Block=greekextended}', "");
Expect(0, 8192, '\P{^Is_Block=greekextended}', "");
Expect(1, 8191, '\p{Is_Block=	-Greek_Extended}', "");
Expect(0, 8191, '\p{^Is_Block=	-Greek_Extended}', "");
Expect(0, 8191, '\P{Is_Block=	-Greek_Extended}', "");
Expect(1, 8191, '\P{^Is_Block=	-Greek_Extended}', "");
Expect(0, 8192, '\p{Is_Block=	-Greek_Extended}', "");
Expect(1, 8192, '\p{^Is_Block=	-Greek_Extended}', "");
Expect(1, 8192, '\P{Is_Block=	-Greek_Extended}', "");
Expect(0, 8192, '\P{^Is_Block=	-Greek_Extended}', "");
Error('\p{Is_Blk=/a/__GREEK_Ext}');
Error('\P{Is_Blk=/a/__GREEK_Ext}');
Expect(1, 8191, '\p{Is_Blk=greekext}', "");
Expect(0, 8191, '\p{^Is_Blk=greekext}', "");
Expect(0, 8191, '\P{Is_Blk=greekext}', "");
Expect(1, 8191, '\P{^Is_Blk=greekext}', "");
Expect(0, 8192, '\p{Is_Blk=greekext}', "");
Expect(1, 8192, '\p{^Is_Blk=greekext}', "");
Expect(1, 8192, '\P{Is_Blk=greekext}', "");
Expect(0, 8192, '\P{^Is_Blk=greekext}', "");
Expect(1, 8191, '\p{Is_Blk=_	Greek_ext}', "");
Expect(0, 8191, '\p{^Is_Blk=_	Greek_ext}', "");
Expect(0, 8191, '\P{Is_Blk=_	Greek_ext}', "");
Expect(1, 8191, '\P{^Is_Blk=_	Greek_ext}', "");
Expect(0, 8192, '\p{Is_Blk=_	Greek_ext}', "");
Expect(1, 8192, '\p{^Is_Blk=_	Greek_ext}', "");
Expect(1, 8192, '\P{Is_Blk=_	Greek_ext}', "");
Expect(0, 8192, '\P{^Is_Blk=_	Greek_ext}', "");
Error('\p{Block=	-enclosed_ideographic_supplement/a/}');
Error('\P{Block=	-enclosed_ideographic_supplement/a/}');
Expect(1, 127743, '\p{Block=enclosedideographicsupplement}', "");
Expect(0, 127743, '\p{^Block=enclosedideographicsupplement}', "");
Expect(0, 127743, '\P{Block=enclosedideographicsupplement}', "");
Expect(1, 127743, '\P{^Block=enclosedideographicsupplement}', "");
Expect(0, 127744, '\p{Block=enclosedideographicsupplement}', "");
Expect(1, 127744, '\p{^Block=enclosedideographicsupplement}', "");
Expect(1, 127744, '\P{Block=enclosedideographicsupplement}', "");
Expect(0, 127744, '\P{^Block=enclosedideographicsupplement}', "");
Expect(1, 127743, '\p{Block= _Enclosed_Ideographic_Supplement}', "");
Expect(0, 127743, '\p{^Block= _Enclosed_Ideographic_Supplement}', "");
Expect(0, 127743, '\P{Block= _Enclosed_Ideographic_Supplement}', "");
Expect(1, 127743, '\P{^Block= _Enclosed_Ideographic_Supplement}', "");
Expect(0, 127744, '\p{Block= _Enclosed_Ideographic_Supplement}', "");
Expect(1, 127744, '\p{^Block= _Enclosed_Ideographic_Supplement}', "");
Expect(1, 127744, '\P{Block= _Enclosed_Ideographic_Supplement}', "");
Expect(0, 127744, '\P{^Block= _Enclosed_Ideographic_Supplement}', "");
Error('\p{Blk=	 enclosed_IDEOGRAPHIC_Sup:=}');
Error('\P{Blk=	 enclosed_IDEOGRAPHIC_Sup:=}');
Expect(1, 127743, '\p{Blk=enclosedideographicsup}', "");
Expect(0, 127743, '\p{^Blk=enclosedideographicsup}', "");
Expect(0, 127743, '\P{Blk=enclosedideographicsup}', "");
Expect(1, 127743, '\P{^Blk=enclosedideographicsup}', "");
Expect(0, 127744, '\p{Blk=enclosedideographicsup}', "");
Expect(1, 127744, '\p{^Blk=enclosedideographicsup}', "");
Expect(1, 127744, '\P{Blk=enclosedideographicsup}', "");
Expect(0, 127744, '\P{^Blk=enclosedideographicsup}', "");
Expect(1, 127743, '\p{Blk=	_Enclosed_ideographic_SUP}', "");
Expect(0, 127743, '\p{^Blk=	_Enclosed_ideographic_SUP}', "");
Expect(0, 127743, '\P{Blk=	_Enclosed_ideographic_SUP}', "");
Expect(1, 127743, '\P{^Blk=	_Enclosed_ideographic_SUP}', "");
Expect(0, 127744, '\p{Blk=	_Enclosed_ideographic_SUP}', "");
Expect(1, 127744, '\p{^Blk=	_Enclosed_ideographic_SUP}', "");
Expect(1, 127744, '\P{Blk=	_Enclosed_ideographic_SUP}', "");
Expect(0, 127744, '\P{^Blk=	_Enclosed_ideographic_SUP}', "");
Error('\p{Is_Block=:= _ENCLOSED_ideographic_supplement}');
Error('\P{Is_Block=:= _ENCLOSED_ideographic_supplement}');
Expect(1, 127743, '\p{Is_Block=enclosedideographicsupplement}', "");
Expect(0, 127743, '\p{^Is_Block=enclosedideographicsupplement}', "");
Expect(0, 127743, '\P{Is_Block=enclosedideographicsupplement}', "");
Expect(1, 127743, '\P{^Is_Block=enclosedideographicsupplement}', "");
Expect(0, 127744, '\p{Is_Block=enclosedideographicsupplement}', "");
Expect(1, 127744, '\p{^Is_Block=enclosedideographicsupplement}', "");
Expect(1, 127744, '\P{Is_Block=enclosedideographicsupplement}', "");
Expect(0, 127744, '\P{^Is_Block=enclosedideographicsupplement}', "");
Expect(1, 127743, '\p{Is_Block= 	Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127743, '\p{^Is_Block= 	Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127743, '\P{Is_Block= 	Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(1, 127743, '\P{^Is_Block= 	Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127744, '\p{Is_Block= 	Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(1, 127744, '\p{^Is_Block= 	Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(1, 127744, '\P{Is_Block= 	Enclosed_IDEOGRAPHIC_Supplement}', "");
Expect(0, 127744, '\P{^Is_Block= 	Enclosed_IDEOGRAPHIC_Supplement}', "");
Error('\p{Is_Blk: /a/_Enclosed_Ideographic_sup}');
Error('\P{Is_Blk: /a/_Enclosed_Ideographic_sup}');
Expect(1, 127743, '\p{Is_Blk=enclosedideographicsup}', "");
Expect(0, 127743, '\p{^Is_Blk=enclosedideographicsup}', "");
Expect(0, 127743, '\P{Is_Blk=enclosedideographicsup}', "");
Expect(1, 127743, '\P{^Is_Blk=enclosedideographicsup}', "");
Expect(0, 127744, '\p{Is_Blk=enclosedideographicsup}', "");
Expect(1, 127744, '\p{^Is_Blk=enclosedideographicsup}', "");
Expect(1, 127744, '\P{Is_Blk=enclosedideographicsup}', "");
Expect(0, 127744, '\P{^Is_Blk=enclosedideographicsup}', "");
Expect(1, 127743, '\p{Is_Blk=- Enclosed_Ideographic_sup}', "");
Expect(0, 127743, '\p{^Is_Blk=- Enclosed_Ideographic_sup}', "");
Expect(0, 127743, '\P{Is_Blk=- Enclosed_Ideographic_sup}', "");
Expect(1, 127743, '\P{^Is_Blk=- Enclosed_Ideographic_sup}', "");
Expect(0, 127744, '\p{Is_Blk=- Enclosed_Ideographic_sup}', "");
Expect(1, 127744, '\p{^Is_Blk=- Enclosed_Ideographic_sup}', "");
Expect(1, 127744, '\P{Is_Blk=- Enclosed_Ideographic_sup}', "");
Expect(0, 127744, '\P{^Is_Blk=- Enclosed_Ideographic_sup}', "");
Error('\p{Block=/a/-	phonetic_EXTENSIONS_Supplement}');
Error('\P{Block=/a/-	phonetic_EXTENSIONS_Supplement}');
Expect(1, 7615, '\p{Block=phoneticextensionssupplement}', "");
Expect(0, 7615, '\p{^Block=phoneticextensionssupplement}', "");
Expect(0, 7615, '\P{Block=phoneticextensionssupplement}', "");
Expect(1, 7615, '\P{^Block=phoneticextensionssupplement}', "");
Expect(0, 7616, '\p{Block=phoneticextensionssupplement}', "");
Expect(1, 7616, '\p{^Block=phoneticextensionssupplement}', "");
Expect(1, 7616, '\P{Block=phoneticextensionssupplement}', "");
Expect(0, 7616, '\P{^Block=phoneticextensionssupplement}', "");
Expect(1, 7615, '\p{Block= _PHONETIC_Extensions_SUPPLEMENT}', "");
Expect(0, 7615, '\p{^Block= _PHONETIC_Extensions_SUPPLEMENT}', "");
Expect(0, 7615, '\P{Block= _PHONETIC_Extensions_SUPPLEMENT}', "");
Expect(1, 7615, '\P{^Block= _PHONETIC_Extensions_SUPPLEMENT}', "");
Expect(0, 7616, '\p{Block= _PHONETIC_Extensions_SUPPLEMENT}', "");
Expect(1, 7616, '\p{^Block= _PHONETIC_Extensions_SUPPLEMENT}', "");
Expect(1, 7616, '\P{Block= _PHONETIC_Extensions_SUPPLEMENT}', "");
Expect(0, 7616, '\P{^Block= _PHONETIC_Extensions_SUPPLEMENT}', "");
Error('\p{Blk=	:=Phonetic_ext_sup}');
Error('\P{Blk=	:=Phonetic_ext_sup}');
Expect(1, 7615, '\p{Blk=phoneticextsup}', "");
Expect(0, 7615, '\p{^Blk=phoneticextsup}', "");
Expect(0, 7615, '\P{Blk=phoneticextsup}', "");
Expect(1, 7615, '\P{^Blk=phoneticextsup}', "");
Expect(0, 7616, '\p{Blk=phoneticextsup}', "");
Expect(1, 7616, '\p{^Blk=phoneticextsup}', "");
Expect(1, 7616, '\P{Blk=phoneticextsup}', "");
Expect(0, 7616, '\P{^Blk=phoneticextsup}', "");
Expect(1, 7615, '\p{Blk=_Phonetic_ext_sup}', "");
Expect(0, 7615, '\p{^Blk=_Phonetic_ext_sup}', "");
Expect(0, 7615, '\P{Blk=_Phonetic_ext_sup}', "");
Expect(1, 7615, '\P{^Blk=_Phonetic_ext_sup}', "");
Expect(0, 7616, '\p{Blk=_Phonetic_ext_sup}', "");
Expect(1, 7616, '\p{^Blk=_Phonetic_ext_sup}', "");
Expect(1, 7616, '\P{Blk=_Phonetic_ext_sup}', "");
Expect(0, 7616, '\P{^Blk=_Phonetic_ext_sup}', "");
Error('\p{Is_Block=-Phonetic_EXTENSIONS_supplement/a/}');
Error('\P{Is_Block=-Phonetic_EXTENSIONS_supplement/a/}');
Expect(1, 7615, '\p{Is_Block=phoneticextensionssupplement}', "");
Expect(0, 7615, '\p{^Is_Block=phoneticextensionssupplement}', "");
Expect(0, 7615, '\P{Is_Block=phoneticextensionssupplement}', "");
Expect(1, 7615, '\P{^Is_Block=phoneticextensionssupplement}', "");
Expect(0, 7616, '\p{Is_Block=phoneticextensionssupplement}', "");
Expect(1, 7616, '\p{^Is_Block=phoneticextensionssupplement}', "");
Expect(1, 7616, '\P{Is_Block=phoneticextensionssupplement}', "");
Expect(0, 7616, '\P{^Is_Block=phoneticextensionssupplement}', "");
Expect(1, 7615, '\p{Is_Block= phonetic_extensions_Supplement}', "");
Expect(0, 7615, '\p{^Is_Block= phonetic_extensions_Supplement}', "");
Expect(0, 7615, '\P{Is_Block= phonetic_extensions_Supplement}', "");
Expect(1, 7615, '\P{^Is_Block= phonetic_extensions_Supplement}', "");
Expect(0, 7616, '\p{Is_Block= phonetic_extensions_Supplement}', "");
Expect(1, 7616, '\p{^Is_Block= phonetic_extensions_Supplement}', "");
Expect(1, 7616, '\P{Is_Block= phonetic_extensions_Supplement}', "");
Expect(0, 7616, '\P{^Is_Block= phonetic_extensions_Supplement}', "");
Error('\p{Is_Blk=_/a/Phonetic_Ext_Sup}');
Error('\P{Is_Blk=_/a/Phonetic_Ext_Sup}');
Expect(1, 7615, '\p{Is_Blk=phoneticextsup}', "");
Expect(0, 7615, '\p{^Is_Blk=phoneticextsup}', "");
Expect(0, 7615, '\P{Is_Blk=phoneticextsup}', "");
Expect(1, 7615, '\P{^Is_Blk=phoneticextsup}', "");
Expect(0, 7616, '\p{Is_Blk=phoneticextsup}', "");
Expect(1, 7616, '\p{^Is_Blk=phoneticextsup}', "");
Expect(1, 7616, '\P{Is_Blk=phoneticextsup}', "");
Expect(0, 7616, '\P{^Is_Blk=phoneticextsup}', "");
Expect(1, 7615, '\p{Is_Blk=- phonetic_ext_Sup}', "");
Expect(0, 7615, '\p{^Is_Blk=- phonetic_ext_Sup}', "");
Expect(0, 7615, '\P{Is_Blk=- phonetic_ext_Sup}', "");
Expect(1, 7615, '\P{^Is_Blk=- phonetic_ext_Sup}', "");
Expect(0, 7616, '\p{Is_Blk=- phonetic_ext_Sup}', "");
Expect(1, 7616, '\p{^Is_Blk=- phonetic_ext_Sup}', "");
Expect(1, 7616, '\P{Is_Blk=- phonetic_ext_Sup}', "");
Expect(0, 7616, '\P{^Is_Blk=- phonetic_ext_Sup}', "");
Error('\p{Block= /a/Batak}');
Error('\P{Block= /a/Batak}');
Expect(1, 7167, '\p{Block=batak}', "");
Expect(0, 7167, '\p{^Block=batak}', "");
Expect(0, 7167, '\P{Block=batak}', "");
Expect(1, 7167, '\P{^Block=batak}', "");
Expect(0, 7168, '\p{Block=batak}', "");
Expect(1, 7168, '\p{^Block=batak}', "");
Expect(1, 7168, '\P{Block=batak}', "");
Expect(0, 7168, '\P{^Block=batak}', "");
Expect(1, 7167, '\p{Block=  BATAK}', "");
Expect(0, 7167, '\p{^Block=  BATAK}', "");
Expect(0, 7167, '\P{Block=  BATAK}', "");
Expect(1, 7167, '\P{^Block=  BATAK}', "");
Expect(0, 7168, '\p{Block=  BATAK}', "");
Expect(1, 7168, '\p{^Block=  BATAK}', "");
Expect(1, 7168, '\P{Block=  BATAK}', "");
Expect(0, 7168, '\P{^Block=  BATAK}', "");
Error('\p{Blk=	/a/Batak}');
Error('\P{Blk=	/a/Batak}');
Expect(1, 7167, '\p{Blk=batak}', "");
Expect(0, 7167, '\p{^Blk=batak}', "");
Expect(0, 7167, '\P{Blk=batak}', "");
Expect(1, 7167, '\P{^Blk=batak}', "");
Expect(0, 7168, '\p{Blk=batak}', "");
Expect(1, 7168, '\p{^Blk=batak}', "");
Expect(1, 7168, '\P{Blk=batak}', "");
Expect(0, 7168, '\P{^Blk=batak}', "");
Expect(1, 7167, '\p{Blk: -	BATAK}', "");
Expect(0, 7167, '\p{^Blk: -	BATAK}', "");
Expect(0, 7167, '\P{Blk: -	BATAK}', "");
Expect(1, 7167, '\P{^Blk: -	BATAK}', "");
Expect(0, 7168, '\p{Blk: -	BATAK}', "");
Expect(1, 7168, '\p{^Blk: -	BATAK}', "");
Expect(1, 7168, '\P{Blk: -	BATAK}', "");
Expect(0, 7168, '\P{^Blk: -	BATAK}', "");
Error('\p{Is_Block=	/a/BATAK}');
Error('\P{Is_Block=	/a/BATAK}');
Expect(1, 7167, '\p{Is_Block=batak}', "");
Expect(0, 7167, '\p{^Is_Block=batak}', "");
Expect(0, 7167, '\P{Is_Block=batak}', "");
Expect(1, 7167, '\P{^Is_Block=batak}', "");
Expect(0, 7168, '\p{Is_Block=batak}', "");
Expect(1, 7168, '\p{^Is_Block=batak}', "");
Expect(1, 7168, '\P{Is_Block=batak}', "");
Expect(0, 7168, '\P{^Is_Block=batak}', "");
Expect(1, 7167, '\p{Is_Block=_Batak}', "");
Expect(0, 7167, '\p{^Is_Block=_Batak}', "");
Expect(0, 7167, '\P{Is_Block=_Batak}', "");
Expect(1, 7167, '\P{^Is_Block=_Batak}', "");
Expect(0, 7168, '\p{Is_Block=_Batak}', "");
Expect(1, 7168, '\p{^Is_Block=_Batak}', "");
Expect(1, 7168, '\P{Is_Block=_Batak}', "");
Expect(0, 7168, '\P{^Is_Block=_Batak}', "");
Error('\p{Is_Blk=:= 	Batak}');
Error('\P{Is_Blk=:= 	Batak}');
Expect(1, 7167, '\p{Is_Blk=batak}', "");
Expect(0, 7167, '\p{^Is_Blk=batak}', "");
Expect(0, 7167, '\P{Is_Blk=batak}', "");
Expect(1, 7167, '\P{^Is_Blk=batak}', "");
Expect(0, 7168, '\p{Is_Blk=batak}', "");
Expect(1, 7168, '\p{^Is_Blk=batak}', "");
Expect(1, 7168, '\P{Is_Blk=batak}', "");
Expect(0, 7168, '\P{^Is_Blk=batak}', "");
Expect(1, 7167, '\p{Is_Blk=	 BATAK}', "");
Expect(0, 7167, '\p{^Is_Blk=	 BATAK}', "");
Expect(0, 7167, '\P{Is_Blk=	 BATAK}', "");
Expect(1, 7167, '\P{^Is_Blk=	 BATAK}', "");
Expect(0, 7168, '\p{Is_Blk=	 BATAK}', "");
Expect(1, 7168, '\p{^Is_Blk=	 BATAK}', "");
Expect(1, 7168, '\P{Is_Blk=	 BATAK}', "");
Expect(0, 7168, '\P{^Is_Blk=	 BATAK}', "");
Error('\p{Block=:=- MEROITIC_Hieroglyphs}');
Error('\P{Block=:=- MEROITIC_Hieroglyphs}');
Expect(1, 67999, '\p{Block:   meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Block:   meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Block:   meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Block:   meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Block:   meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Block:   meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Block:   meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Block:   meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Block=--Meroitic_hieroglyphs}', "");
Expect(0, 67999, '\p{^Block=--Meroitic_hieroglyphs}', "");
Expect(0, 67999, '\P{Block=--Meroitic_hieroglyphs}', "");
Expect(1, 67999, '\P{^Block=--Meroitic_hieroglyphs}', "");
Expect(0, 68000, '\p{Block=--Meroitic_hieroglyphs}', "");
Expect(1, 68000, '\p{^Block=--Meroitic_hieroglyphs}', "");
Expect(1, 68000, '\P{Block=--Meroitic_hieroglyphs}', "");
Expect(0, 68000, '\P{^Block=--Meroitic_hieroglyphs}', "");
Error('\p{Blk=/a/ -meroitic_hieroglyphs}');
Error('\P{Blk=/a/ -meroitic_hieroglyphs}');
Expect(1, 67999, '\p{Blk: meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Blk: meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Blk: meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Blk: meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Blk: meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Blk: meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Blk: meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Blk: meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Blk= _meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\p{^Blk= _meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\P{Blk= _meroitic_Hieroglyphs}', "");
Expect(1, 67999, '\P{^Blk= _meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\p{Blk= _meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\p{^Blk= _meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\P{Blk= _meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\P{^Blk= _meroitic_Hieroglyphs}', "");
Error('\p{Is_Block=-_MEROITIC_Hieroglyphs/a/}');
Error('\P{Is_Block=-_MEROITIC_Hieroglyphs/a/}');
Expect(1, 67999, '\p{Is_Block=meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Block=meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Is_Block=meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Block=meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Is_Block=meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Block=meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Is_Block=meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Block=meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Is_Block=_	Meroitic_HIEROGLYPHS}', "");
Expect(0, 67999, '\p{^Is_Block=_	Meroitic_HIEROGLYPHS}', "");
Expect(0, 67999, '\P{Is_Block=_	Meroitic_HIEROGLYPHS}', "");
Expect(1, 67999, '\P{^Is_Block=_	Meroitic_HIEROGLYPHS}', "");
Expect(0, 68000, '\p{Is_Block=_	Meroitic_HIEROGLYPHS}', "");
Expect(1, 68000, '\p{^Is_Block=_	Meroitic_HIEROGLYPHS}', "");
Expect(1, 68000, '\P{Is_Block=_	Meroitic_HIEROGLYPHS}', "");
Expect(0, 68000, '\P{^Is_Block=_	Meroitic_HIEROGLYPHS}', "");
Error('\p{Is_Blk=:=	 Meroitic_Hieroglyphs}');
Error('\P{Is_Blk=:=	 Meroitic_Hieroglyphs}');
Expect(1, 67999, '\p{Is_Blk=meroitichieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Blk=meroitichieroglyphs}', "");
Expect(0, 67999, '\P{Is_Blk=meroitichieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Blk=meroitichieroglyphs}', "");
Expect(0, 68000, '\p{Is_Blk=meroitichieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Blk=meroitichieroglyphs}', "");
Expect(1, 68000, '\P{Is_Blk=meroitichieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Blk=meroitichieroglyphs}', "");
Expect(1, 67999, '\p{Is_Blk=--Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\p{^Is_Blk=--Meroitic_Hieroglyphs}', "");
Expect(0, 67999, '\P{Is_Blk=--Meroitic_Hieroglyphs}', "");
Expect(1, 67999, '\P{^Is_Blk=--Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\p{Is_Blk=--Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\p{^Is_Blk=--Meroitic_Hieroglyphs}', "");
Expect(1, 68000, '\P{Is_Blk=--Meroitic_Hieroglyphs}', "");
Expect(0, 68000, '\P{^Is_Blk=--Meroitic_Hieroglyphs}', "");
Error('\p{Block=	:=katakana_PHONETIC_extensions}');
Error('\P{Block=	:=katakana_PHONETIC_extensions}');
Expect(1, 12799, '\p{Block=katakanaphoneticextensions}', "");
Expect(0, 12799, '\p{^Block=katakanaphoneticextensions}', "");
Expect(0, 12799, '\P{Block=katakanaphoneticextensions}', "");
Expect(1, 12799, '\P{^Block=katakanaphoneticextensions}', "");
Expect(0, 12800, '\p{Block=katakanaphoneticextensions}', "");
Expect(1, 12800, '\p{^Block=katakanaphoneticextensions}', "");
Expect(1, 12800, '\P{Block=katakanaphoneticextensions}', "");
Expect(0, 12800, '\P{^Block=katakanaphoneticextensions}', "");
Expect(1, 12799, '\p{Block=__katakana_phonetic_EXTENSIONS}', "");
Expect(0, 12799, '\p{^Block=__katakana_phonetic_EXTENSIONS}', "");
Expect(0, 12799, '\P{Block=__katakana_phonetic_EXTENSIONS}', "");
Expect(1, 12799, '\P{^Block=__katakana_phonetic_EXTENSIONS}', "");
Expect(0, 12800, '\p{Block=__katakana_phonetic_EXTENSIONS}', "");
Expect(1, 12800, '\p{^Block=__katakana_phonetic_EXTENSIONS}', "");
Expect(1, 12800, '\P{Block=__katakana_phonetic_EXTENSIONS}', "");
Expect(0, 12800, '\P{^Block=__katakana_phonetic_EXTENSIONS}', "");
Error('\p{Blk=_:=katakana_Ext}');
Error('\P{Blk=_:=katakana_Ext}');
Expect(1, 12799, '\p{Blk=katakanaext}', "");
Expect(0, 12799, '\p{^Blk=katakanaext}', "");
Expect(0, 12799, '\P{Blk=katakanaext}', "");
Expect(1, 12799, '\P{^Blk=katakanaext}', "");
Expect(0, 12800, '\p{Blk=katakanaext}', "");
Expect(1, 12800, '\p{^Blk=katakanaext}', "");
Expect(1, 12800, '\P{Blk=katakanaext}', "");
Expect(0, 12800, '\P{^Blk=katakanaext}', "");
Expect(1, 12799, '\p{Blk=--katakana_Ext}', "");
Expect(0, 12799, '\p{^Blk=--katakana_Ext}', "");
Expect(0, 12799, '\P{Blk=--katakana_Ext}', "");
Expect(1, 12799, '\P{^Blk=--katakana_Ext}', "");
Expect(0, 12800, '\p{Blk=--katakana_Ext}', "");
Expect(1, 12800, '\p{^Blk=--katakana_Ext}', "");
Expect(1, 12800, '\P{Blk=--katakana_Ext}', "");
Expect(0, 12800, '\P{^Blk=--katakana_Ext}', "");
Error('\p{Is_Block=	Katakana_Phonetic_Extensions/a/}');
Error('\P{Is_Block=	Katakana_Phonetic_Extensions/a/}');
Expect(1, 12799, '\p{Is_Block=katakanaphoneticextensions}', "");
Expect(0, 12799, '\p{^Is_Block=katakanaphoneticextensions}', "");
Expect(0, 12799, '\P{Is_Block=katakanaphoneticextensions}', "");
Expect(1, 12799, '\P{^Is_Block=katakanaphoneticextensions}', "");
Expect(0, 12800, '\p{Is_Block=katakanaphoneticextensions}', "");
Expect(1, 12800, '\p{^Is_Block=katakanaphoneticextensions}', "");
Expect(1, 12800, '\P{Is_Block=katakanaphoneticextensions}', "");
Expect(0, 12800, '\P{^Is_Block=katakanaphoneticextensions}', "");
Expect(1, 12799, '\p{Is_Block=  Katakana_phonetic_Extensions}', "");
Expect(0, 12799, '\p{^Is_Block=  Katakana_phonetic_Extensions}', "");
Expect(0, 12799, '\P{Is_Block=  Katakana_phonetic_Extensions}', "");
Expect(1, 12799, '\P{^Is_Block=  Katakana_phonetic_Extensions}', "");
Expect(0, 12800, '\p{Is_Block=  Katakana_phonetic_Extensions}', "");
Expect(1, 12800, '\p{^Is_Block=  Katakana_phonetic_Extensions}', "");
Expect(1, 12800, '\P{Is_Block=  Katakana_phonetic_Extensions}', "");
Expect(0, 12800, '\P{^Is_Block=  Katakana_phonetic_Extensions}', "");
Error('\p{Is_Blk= :=Katakana_EXT}');
Error('\P{Is_Blk= :=Katakana_EXT}');
Expect(1, 12799, '\p{Is_Blk=katakanaext}', "");
Expect(0, 12799, '\p{^Is_Blk=katakanaext}', "");
Expect(0, 12799, '\P{Is_Blk=katakanaext}', "");
Expect(1, 12799, '\P{^Is_Blk=katakanaext}', "");
Expect(0, 12800, '\p{Is_Blk=katakanaext}', "");
Expect(1, 12800, '\p{^Is_Blk=katakanaext}', "");
Expect(1, 12800, '\P{Is_Blk=katakanaext}', "");
Expect(0, 12800, '\P{^Is_Blk=katakanaext}', "");
Expect(1, 12799, '\p{Is_Blk=-	KATAKANA_Ext}', "");
Expect(0, 12799, '\p{^Is_Blk=-	KATAKANA_Ext}', "");
Expect(0, 12799, '\P{Is_Blk=-	KATAKANA_Ext}', "");
Expect(1, 12799, '\P{^Is_Blk=-	KATAKANA_Ext}', "");
Expect(0, 12800, '\p{Is_Blk=-	KATAKANA_Ext}', "");
Expect(1, 12800, '\p{^Is_Blk=-	KATAKANA_Ext}', "");
Expect(1, 12800, '\P{Is_Blk=-	KATAKANA_Ext}', "");
Expect(0, 12800, '\P{^Is_Blk=-	KATAKANA_Ext}', "");
Error('\p{Block=-:=COMMON_INDIC_number_Forms}');
Error('\P{Block=-:=COMMON_INDIC_number_Forms}');
Expect(1, 43071, '\p{Block=commonindicnumberforms}', "");
Expect(0, 43071, '\p{^Block=commonindicnumberforms}', "");
Expect(0, 43071, '\P{Block=commonindicnumberforms}', "");
Expect(1, 43071, '\P{^Block=commonindicnumberforms}', "");
Expect(0, 43072, '\p{Block=commonindicnumberforms}', "");
Expect(1, 43072, '\p{^Block=commonindicnumberforms}', "");
Expect(1, 43072, '\P{Block=commonindicnumberforms}', "");
Expect(0, 43072, '\P{^Block=commonindicnumberforms}', "");
Expect(1, 43071, '\p{Block=-COMMON_indic_NUMBER_Forms}', "");
Expect(0, 43071, '\p{^Block=-COMMON_indic_NUMBER_Forms}', "");
Expect(0, 43071, '\P{Block=-COMMON_indic_NUMBER_Forms}', "");
Expect(1, 43071, '\P{^Block=-COMMON_indic_NUMBER_Forms}', "");
Expect(0, 43072, '\p{Block=-COMMON_indic_NUMBER_Forms}', "");
Expect(1, 43072, '\p{^Block=-COMMON_indic_NUMBER_Forms}', "");
Expect(1, 43072, '\P{Block=-COMMON_indic_NUMBER_Forms}', "");
Expect(0, 43072, '\P{^Block=-COMMON_indic_NUMBER_Forms}', "");
Error('\p{Blk=:= -Indic_Number_FORMS}');
Error('\P{Blk=:= -Indic_Number_FORMS}');
Expect(1, 43071, '\p{Blk=indicnumberforms}', "");
Expect(0, 43071, '\p{^Blk=indicnumberforms}', "");
Expect(0, 43071, '\P{Blk=indicnumberforms}', "");
Expect(1, 43071, '\P{^Blk=indicnumberforms}', "");
Expect(0, 43072, '\p{Blk=indicnumberforms}', "");
Expect(1, 43072, '\p{^Blk=indicnumberforms}', "");
Expect(1, 43072, '\P{Blk=indicnumberforms}', "");
Expect(0, 43072, '\P{^Blk=indicnumberforms}', "");
Expect(1, 43071, '\p{Blk=-	indic_number_FORMS}', "");
Expect(0, 43071, '\p{^Blk=-	indic_number_FORMS}', "");
Expect(0, 43071, '\P{Blk=-	indic_number_FORMS}', "");
Expect(1, 43071, '\P{^Blk=-	indic_number_FORMS}', "");
Expect(0, 43072, '\p{Blk=-	indic_number_FORMS}', "");
Expect(1, 43072, '\p{^Blk=-	indic_number_FORMS}', "");
Expect(1, 43072, '\P{Blk=-	indic_number_FORMS}', "");
Expect(0, 43072, '\P{^Blk=-	indic_number_FORMS}', "");
Error('\p{Is_Block=_ COMMON_Indic_number_FORMS/a/}');
Error('\P{Is_Block=_ COMMON_Indic_number_FORMS/a/}');
Expect(1, 43071, '\p{Is_Block=commonindicnumberforms}', "");
Expect(0, 43071, '\p{^Is_Block=commonindicnumberforms}', "");
Expect(0, 43071, '\P{Is_Block=commonindicnumberforms}', "");
Expect(1, 43071, '\P{^Is_Block=commonindicnumberforms}', "");
Expect(0, 43072, '\p{Is_Block=commonindicnumberforms}', "");
Expect(1, 43072, '\p{^Is_Block=commonindicnumberforms}', "");
Expect(1, 43072, '\P{Is_Block=commonindicnumberforms}', "");
Expect(0, 43072, '\P{^Is_Block=commonindicnumberforms}', "");
Expect(1, 43071, '\p{Is_Block=_Common_Indic_NUMBER_Forms}', "");
Expect(0, 43071, '\p{^Is_Block=_Common_Indic_NUMBER_Forms}', "");
Expect(0, 43071, '\P{Is_Block=_Common_Indic_NUMBER_Forms}', "");
Expect(1, 43071, '\P{^Is_Block=_Common_Indic_NUMBER_Forms}', "");
Expect(0, 43072, '\p{Is_Block=_Common_Indic_NUMBER_Forms}', "");
Expect(1, 43072, '\p{^Is_Block=_Common_Indic_NUMBER_Forms}', "");
Expect(1, 43072, '\P{Is_Block=_Common_Indic_NUMBER_Forms}', "");
Expect(0, 43072, '\P{^Is_Block=_Common_Indic_NUMBER_Forms}', "");
Error('\p{Is_Blk=/a/indic_Number_FORMS}');
Error('\P{Is_Blk=/a/indic_Number_FORMS}');
Expect(1, 43071, '\p{Is_Blk=indicnumberforms}', "");
Expect(0, 43071, '\p{^Is_Blk=indicnumberforms}', "");
Expect(0, 43071, '\P{Is_Blk=indicnumberforms}', "");
Expect(1, 43071, '\P{^Is_Blk=indicnumberforms}', "");
Expect(0, 43072, '\p{Is_Blk=indicnumberforms}', "");
Expect(1, 43072, '\p{^Is_Blk=indicnumberforms}', "");
Expect(1, 43072, '\P{Is_Blk=indicnumberforms}', "");
Expect(0, 43072, '\P{^Is_Blk=indicnumberforms}', "");
Expect(1, 43071, '\p{Is_Blk=-Indic_Number_Forms}', "");
Expect(0, 43071, '\p{^Is_Blk=-Indic_Number_Forms}', "");
Expect(0, 43071, '\P{Is_Blk=-Indic_Number_Forms}', "");
Expect(1, 43071, '\P{^Is_Blk=-Indic_Number_Forms}', "");
Expect(0, 43072, '\p{Is_Blk=-Indic_Number_Forms}', "");
Expect(1, 43072, '\p{^Is_Blk=-Indic_Number_Forms}', "");
Expect(1, 43072, '\P{Is_Blk=-Indic_Number_Forms}', "");
Expect(0, 43072, '\P{^Is_Blk=-Indic_Number_Forms}', "");
Error('\p{Block=:=MEETEI_MAYEK}');
Error('\P{Block=:=MEETEI_MAYEK}');
Expect(1, 44031, '\p{Block:meeteimayek}', "");
Expect(0, 44031, '\p{^Block:meeteimayek}', "");
Expect(0, 44031, '\P{Block:meeteimayek}', "");
Expect(1, 44031, '\P{^Block:meeteimayek}', "");
Expect(0, 44032, '\p{Block:meeteimayek}', "");
Expect(1, 44032, '\p{^Block:meeteimayek}', "");
Expect(1, 44032, '\P{Block:meeteimayek}', "");
Expect(0, 44032, '\P{^Block:meeteimayek}', "");
Expect(1, 44031, '\p{Block=-_Meetei_Mayek}', "");
Expect(0, 44031, '\p{^Block=-_Meetei_Mayek}', "");
Expect(0, 44031, '\P{Block=-_Meetei_Mayek}', "");
Expect(1, 44031, '\P{^Block=-_Meetei_Mayek}', "");
Expect(0, 44032, '\p{Block=-_Meetei_Mayek}', "");
Expect(1, 44032, '\p{^Block=-_Meetei_Mayek}', "");
Expect(1, 44032, '\P{Block=-_Meetei_Mayek}', "");
Expect(0, 44032, '\P{^Block=-_Meetei_Mayek}', "");
Error('\p{Blk:   /a/MEETEI_MAYEK}');
Error('\P{Blk:   /a/MEETEI_MAYEK}');
Expect(1, 44031, '\p{Blk:meeteimayek}', "");
Expect(0, 44031, '\p{^Blk:meeteimayek}', "");
Expect(0, 44031, '\P{Blk:meeteimayek}', "");
Expect(1, 44031, '\P{^Blk:meeteimayek}', "");
Expect(0, 44032, '\p{Blk:meeteimayek}', "");
Expect(1, 44032, '\p{^Blk:meeteimayek}', "");
Expect(1, 44032, '\P{Blk:meeteimayek}', "");
Expect(0, 44032, '\P{^Blk:meeteimayek}', "");
Expect(1, 44031, '\p{Blk= Meetei_MAYEK}', "");
Expect(0, 44031, '\p{^Blk= Meetei_MAYEK}', "");
Expect(0, 44031, '\P{Blk= Meetei_MAYEK}', "");
Expect(1, 44031, '\P{^Blk= Meetei_MAYEK}', "");
Expect(0, 44032, '\p{Blk= Meetei_MAYEK}', "");
Expect(1, 44032, '\p{^Blk= Meetei_MAYEK}', "");
Expect(1, 44032, '\P{Blk= Meetei_MAYEK}', "");
Expect(0, 44032, '\P{^Blk= Meetei_MAYEK}', "");
Error('\p{Is_Block=-	Meetei_mayek/a/}');
Error('\P{Is_Block=-	Meetei_mayek/a/}');
Expect(1, 44031, '\p{Is_Block=meeteimayek}', "");
Expect(0, 44031, '\p{^Is_Block=meeteimayek}', "");
Expect(0, 44031, '\P{Is_Block=meeteimayek}', "");
Expect(1, 44031, '\P{^Is_Block=meeteimayek}', "");
Expect(0, 44032, '\p{Is_Block=meeteimayek}', "");
Expect(1, 44032, '\p{^Is_Block=meeteimayek}', "");
Expect(1, 44032, '\P{Is_Block=meeteimayek}', "");
Expect(0, 44032, '\P{^Is_Block=meeteimayek}', "");
Expect(1, 44031, '\p{Is_Block=-	Meetei_MAYEK}', "");
Expect(0, 44031, '\p{^Is_Block=-	Meetei_MAYEK}', "");
Expect(0, 44031, '\P{Is_Block=-	Meetei_MAYEK}', "");
Expect(1, 44031, '\P{^Is_Block=-	Meetei_MAYEK}', "");
Expect(0, 44032, '\p{Is_Block=-	Meetei_MAYEK}', "");
Expect(1, 44032, '\p{^Is_Block=-	Meetei_MAYEK}', "");
Expect(1, 44032, '\P{Is_Block=-	Meetei_MAYEK}', "");
Expect(0, 44032, '\P{^Is_Block=-	Meetei_MAYEK}', "");
Error('\p{Is_Blk=-	Meetei_MAYEK:=}');
Error('\P{Is_Blk=-	Meetei_MAYEK:=}');
Expect(1, 44031, '\p{Is_Blk=meeteimayek}', "");
Expect(0, 44031, '\p{^Is_Blk=meeteimayek}', "");
Expect(0, 44031, '\P{Is_Blk=meeteimayek}', "");
Expect(1, 44031, '\P{^Is_Blk=meeteimayek}', "");
Expect(0, 44032, '\p{Is_Blk=meeteimayek}', "");
Expect(1, 44032, '\p{^Is_Blk=meeteimayek}', "");
Expect(1, 44032, '\P{Is_Blk=meeteimayek}', "");
Expect(0, 44032, '\P{^Is_Blk=meeteimayek}', "");
Expect(1, 44031, '\p{Is_Blk=_-meetei_Mayek}', "");
Expect(0, 44031, '\p{^Is_Blk=_-meetei_Mayek}', "");
Expect(0, 44031, '\P{Is_Blk=_-meetei_Mayek}', "");
Expect(1, 44031, '\P{^Is_Blk=_-meetei_Mayek}', "");
Expect(0, 44032, '\p{Is_Blk=_-meetei_Mayek}', "");
Expect(1, 44032, '\p{^Is_Blk=_-meetei_Mayek}', "");
Expect(1, 44032, '\P{Is_Blk=_-meetei_Mayek}', "");
Expect(0, 44032, '\P{^Is_Blk=_-meetei_Mayek}', "");
Error('\p{Block=:= carian}');
Error('\P{Block=:= carian}');
Expect(1, 66271, '\p{Block=carian}', "");
Expect(0, 66271, '\p{^Block=carian}', "");
Expect(0, 66271, '\P{Block=carian}', "");
Expect(1, 66271, '\P{^Block=carian}', "");
Expect(0, 66272, '\p{Block=carian}', "");
Expect(1, 66272, '\p{^Block=carian}', "");
Expect(1, 66272, '\P{Block=carian}', "");
Expect(0, 66272, '\P{^Block=carian}', "");
Expect(1, 66271, '\p{Block=__Carian}', "");
Expect(0, 66271, '\p{^Block=__Carian}', "");
Expect(0, 66271, '\P{Block=__Carian}', "");
Expect(1, 66271, '\P{^Block=__Carian}', "");
Expect(0, 66272, '\p{Block=__Carian}', "");
Expect(1, 66272, '\p{^Block=__Carian}', "");
Expect(1, 66272, '\P{Block=__Carian}', "");
Expect(0, 66272, '\P{^Block=__Carian}', "");
Error('\p{Blk=_:=CARIAN}');
Error('\P{Blk=_:=CARIAN}');
Expect(1, 66271, '\p{Blk=carian}', "");
Expect(0, 66271, '\p{^Blk=carian}', "");
Expect(0, 66271, '\P{Blk=carian}', "");
Expect(1, 66271, '\P{^Blk=carian}', "");
Expect(0, 66272, '\p{Blk=carian}', "");
Expect(1, 66272, '\p{^Blk=carian}', "");
Expect(1, 66272, '\P{Blk=carian}', "");
Expect(0, 66272, '\P{^Blk=carian}', "");
Expect(1, 66271, '\p{Blk=	_Carian}', "");
Expect(0, 66271, '\p{^Blk=	_Carian}', "");
Expect(0, 66271, '\P{Blk=	_Carian}', "");
Expect(1, 66271, '\P{^Blk=	_Carian}', "");
Expect(0, 66272, '\p{Blk=	_Carian}', "");
Expect(1, 66272, '\p{^Blk=	_Carian}', "");
Expect(1, 66272, '\P{Blk=	_Carian}', "");
Expect(0, 66272, '\P{^Blk=	_Carian}', "");
Error('\p{Is_Block: _CARIAN:=}');
Error('\P{Is_Block: _CARIAN:=}');
Expect(1, 66271, '\p{Is_Block=carian}', "");
Expect(0, 66271, '\p{^Is_Block=carian}', "");
Expect(0, 66271, '\P{Is_Block=carian}', "");
Expect(1, 66271, '\P{^Is_Block=carian}', "");
Expect(0, 66272, '\p{Is_Block=carian}', "");
Expect(1, 66272, '\p{^Is_Block=carian}', "");
Expect(1, 66272, '\P{Is_Block=carian}', "");
Expect(0, 66272, '\P{^Is_Block=carian}', "");
Expect(1, 66271, '\p{Is_Block=	-CARIAN}', "");
Expect(0, 66271, '\p{^Is_Block=	-CARIAN}', "");
Expect(0, 66271, '\P{Is_Block=	-CARIAN}', "");
Expect(1, 66271, '\P{^Is_Block=	-CARIAN}', "");
Expect(0, 66272, '\p{Is_Block=	-CARIAN}', "");
Expect(1, 66272, '\p{^Is_Block=	-CARIAN}', "");
Expect(1, 66272, '\P{Is_Block=	-CARIAN}', "");
Expect(0, 66272, '\P{^Is_Block=	-CARIAN}', "");
Error('\p{Is_Blk=-/a/Carian}');
Error('\P{Is_Blk=-/a/Carian}');
Expect(1, 66271, '\p{Is_Blk=carian}', "");
Expect(0, 66271, '\p{^Is_Blk=carian}', "");
Expect(0, 66271, '\P{Is_Blk=carian}', "");
Expect(1, 66271, '\P{^Is_Blk=carian}', "");
Expect(0, 66272, '\p{Is_Blk=carian}', "");
Expect(1, 66272, '\p{^Is_Blk=carian}', "");
Expect(1, 66272, '\P{Is_Blk=carian}', "");
Expect(0, 66272, '\P{^Is_Blk=carian}', "");
Expect(1, 66271, '\p{Is_Blk=_	CARIAN}', "");
Expect(0, 66271, '\p{^Is_Blk=_	CARIAN}', "");
Expect(0, 66271, '\P{Is_Blk=_	CARIAN}', "");
Expect(1, 66271, '\P{^Is_Blk=_	CARIAN}', "");
Expect(0, 66272, '\p{Is_Blk=_	CARIAN}', "");
Expect(1, 66272, '\p{^Is_Blk=_	CARIAN}', "");
Expect(1, 66272, '\P{Is_Blk=_	CARIAN}', "");
Expect(0, 66272, '\P{^Is_Blk=_	CARIAN}', "");
Error('\p{Block=:=	Arabic_SUPPLEMENT}');
Error('\P{Block=:=	Arabic_SUPPLEMENT}');
Expect(1, 1919, '\p{Block=arabicsupplement}', "");
Expect(0, 1919, '\p{^Block=arabicsupplement}', "");
Expect(0, 1919, '\P{Block=arabicsupplement}', "");
Expect(1, 1919, '\P{^Block=arabicsupplement}', "");
Expect(0, 1920, '\p{Block=arabicsupplement}', "");
Expect(1, 1920, '\p{^Block=arabicsupplement}', "");
Expect(1, 1920, '\P{Block=arabicsupplement}', "");
Expect(0, 1920, '\P{^Block=arabicsupplement}', "");
Expect(1, 1919, '\p{Block=-_Arabic_Supplement}', "");
Expect(0, 1919, '\p{^Block=-_Arabic_Supplement}', "");
Expect(0, 1919, '\P{Block=-_Arabic_Supplement}', "");
Expect(1, 1919, '\P{^Block=-_Arabic_Supplement}', "");
Expect(0, 1920, '\p{Block=-_Arabic_Supplement}', "");
Expect(1, 1920, '\p{^Block=-_Arabic_Supplement}', "");
Expect(1, 1920, '\P{Block=-_Arabic_Supplement}', "");
Expect(0, 1920, '\P{^Block=-_Arabic_Supplement}', "");
Error('\p{Blk=-:=ARABIC_Sup}');
Error('\P{Blk=-:=ARABIC_Sup}');
Expect(1, 1919, '\p{Blk=arabicsup}', "");
Expect(0, 1919, '\p{^Blk=arabicsup}', "");
Expect(0, 1919, '\P{Blk=arabicsup}', "");
Expect(1, 1919, '\P{^Blk=arabicsup}', "");
Expect(0, 1920, '\p{Blk=arabicsup}', "");
Expect(1, 1920, '\p{^Blk=arabicsup}', "");
Expect(1, 1920, '\P{Blk=arabicsup}', "");
Expect(0, 1920, '\P{^Blk=arabicsup}', "");
Expect(1, 1919, '\p{Blk=_ Arabic_SUP}', "");
Expect(0, 1919, '\p{^Blk=_ Arabic_SUP}', "");
Expect(0, 1919, '\P{Blk=_ Arabic_SUP}', "");
Expect(1, 1919, '\P{^Blk=_ Arabic_SUP}', "");
Expect(0, 1920, '\p{Blk=_ Arabic_SUP}', "");
Expect(1, 1920, '\p{^Blk=_ Arabic_SUP}', "");
Expect(1, 1920, '\P{Blk=_ Arabic_SUP}', "");
Expect(0, 1920, '\P{^Blk=_ Arabic_SUP}', "");
Error('\p{Is_Block:	:=ARABIC_SUPPLEMENT}');
Error('\P{Is_Block:	:=ARABIC_SUPPLEMENT}');
Expect(1, 1919, '\p{Is_Block=arabicsupplement}', "");
Expect(0, 1919, '\p{^Is_Block=arabicsupplement}', "");
Expect(0, 1919, '\P{Is_Block=arabicsupplement}', "");
Expect(1, 1919, '\P{^Is_Block=arabicsupplement}', "");
Expect(0, 1920, '\p{Is_Block=arabicsupplement}', "");
Expect(1, 1920, '\p{^Is_Block=arabicsupplement}', "");
Expect(1, 1920, '\P{Is_Block=arabicsupplement}', "");
Expect(0, 1920, '\P{^Is_Block=arabicsupplement}', "");
Expect(1, 1919, '\p{Is_Block=_-Arabic_SUPPLEMENT}', "");
Expect(0, 1919, '\p{^Is_Block=_-Arabic_SUPPLEMENT}', "");
Expect(0, 1919, '\P{Is_Block=_-Arabic_SUPPLEMENT}', "");
Expect(1, 1919, '\P{^Is_Block=_-Arabic_SUPPLEMENT}', "");
Expect(0, 1920, '\p{Is_Block=_-Arabic_SUPPLEMENT}', "");
Expect(1, 1920, '\p{^Is_Block=_-Arabic_SUPPLEMENT}', "");
Expect(1, 1920, '\P{Is_Block=_-Arabic_SUPPLEMENT}', "");
Expect(0, 1920, '\P{^Is_Block=_-Arabic_SUPPLEMENT}', "");
Error('\p{Is_Blk=/a/-_Arabic_Sup}');
Error('\P{Is_Blk=/a/-_Arabic_Sup}');
Expect(1, 1919, '\p{Is_Blk=arabicsup}', "");
Expect(0, 1919, '\p{^Is_Blk=arabicsup}', "");
Expect(0, 1919, '\P{Is_Blk=arabicsup}', "");
Expect(1, 1919, '\P{^Is_Blk=arabicsup}', "");
Expect(0, 1920, '\p{Is_Blk=arabicsup}', "");
Expect(1, 1920, '\p{^Is_Blk=arabicsup}', "");
Expect(1, 1920, '\P{Is_Blk=arabicsup}', "");
Expect(0, 1920, '\P{^Is_Blk=arabicsup}', "");
Expect(1, 1919, '\p{Is_Blk=_-Arabic_sup}', "");
Expect(0, 1919, '\p{^Is_Blk=_-Arabic_sup}', "");
Expect(0, 1919, '\P{Is_Blk=_-Arabic_sup}', "");
Expect(1, 1919, '\P{^Is_Blk=_-Arabic_sup}', "");
Expect(0, 1920, '\p{Is_Blk=_-Arabic_sup}', "");
Expect(1, 1920, '\p{^Is_Blk=_-Arabic_sup}', "");
Expect(1, 1920, '\P{Is_Blk=_-Arabic_sup}', "");
Expect(0, 1920, '\P{^Is_Blk=_-Arabic_sup}', "");
Error('\p{Block=:=_	ALCHEMICAL_SYMBOLS}');
Error('\P{Block=:=_	ALCHEMICAL_SYMBOLS}');
Expect(1, 128895, '\p{Block=alchemicalsymbols}', "");
Expect(0, 128895, '\p{^Block=alchemicalsymbols}', "");
Expect(0, 128895, '\P{Block=alchemicalsymbols}', "");
Expect(1, 128895, '\P{^Block=alchemicalsymbols}', "");
Expect(0, 128896, '\p{Block=alchemicalsymbols}', "");
Expect(1, 128896, '\p{^Block=alchemicalsymbols}', "");
Expect(1, 128896, '\P{Block=alchemicalsymbols}', "");
Expect(0, 128896, '\P{^Block=alchemicalsymbols}', "");
Expect(1, 128895, '\p{Block=alchemical_SYMBOLS}', "");
Expect(0, 128895, '\p{^Block=alchemical_SYMBOLS}', "");
Expect(0, 128895, '\P{Block=alchemical_SYMBOLS}', "");
Expect(1, 128895, '\P{^Block=alchemical_SYMBOLS}', "");
Expect(0, 128896, '\p{Block=alchemical_SYMBOLS}', "");
Expect(1, 128896, '\p{^Block=alchemical_SYMBOLS}', "");
Expect(1, 128896, '\P{Block=alchemical_SYMBOLS}', "");
Expect(0, 128896, '\P{^Block=alchemical_SYMBOLS}', "");
Error('\p{Blk=	/a/Alchemical}');
Error('\P{Blk=	/a/Alchemical}');
Expect(1, 128895, '\p{Blk=alchemical}', "");
Expect(0, 128895, '\p{^Blk=alchemical}', "");
Expect(0, 128895, '\P{Blk=alchemical}', "");
Expect(1, 128895, '\P{^Blk=alchemical}', "");
Expect(0, 128896, '\p{Blk=alchemical}', "");
Expect(1, 128896, '\p{^Blk=alchemical}', "");
Expect(1, 128896, '\P{Blk=alchemical}', "");
Expect(0, 128896, '\P{^Blk=alchemical}', "");
Expect(1, 128895, '\p{Blk=-alchemical}', "");
Expect(0, 128895, '\p{^Blk=-alchemical}', "");
Expect(0, 128895, '\P{Blk=-alchemical}', "");
Expect(1, 128895, '\P{^Blk=-alchemical}', "");
Expect(0, 128896, '\p{Blk=-alchemical}', "");
Expect(1, 128896, '\p{^Blk=-alchemical}', "");
Expect(1, 128896, '\P{Blk=-alchemical}', "");
Expect(0, 128896, '\P{^Blk=-alchemical}', "");
Error('\p{Is_Block:/a/-	Alchemical_symbols}');
Error('\P{Is_Block:/a/-	Alchemical_symbols}');
Expect(1, 128895, '\p{Is_Block:   alchemicalsymbols}', "");
Expect(0, 128895, '\p{^Is_Block:   alchemicalsymbols}', "");
Expect(0, 128895, '\P{Is_Block:   alchemicalsymbols}', "");
Expect(1, 128895, '\P{^Is_Block:   alchemicalsymbols}', "");
Expect(0, 128896, '\p{Is_Block:   alchemicalsymbols}', "");
Expect(1, 128896, '\p{^Is_Block:   alchemicalsymbols}', "");
Expect(1, 128896, '\P{Is_Block:   alchemicalsymbols}', "");
Expect(0, 128896, '\P{^Is_Block:   alchemicalsymbols}', "");
Expect(1, 128895, '\p{Is_Block= ALCHEMICAL_Symbols}', "");
Expect(0, 128895, '\p{^Is_Block= ALCHEMICAL_Symbols}', "");
Expect(0, 128895, '\P{Is_Block= ALCHEMICAL_Symbols}', "");
Expect(1, 128895, '\P{^Is_Block= ALCHEMICAL_Symbols}', "");
Expect(0, 128896, '\p{Is_Block= ALCHEMICAL_Symbols}', "");
Expect(1, 128896, '\p{^Is_Block= ALCHEMICAL_Symbols}', "");
Expect(1, 128896, '\P{Is_Block= ALCHEMICAL_Symbols}', "");
Expect(0, 128896, '\P{^Is_Block= ALCHEMICAL_Symbols}', "");
Error('\p{Is_Blk=-Alchemical:=}');
Error('\P{Is_Blk=-Alchemical:=}');
Expect(1, 128895, '\p{Is_Blk=alchemical}', "");
Expect(0, 128895, '\p{^Is_Blk=alchemical}', "");
Expect(0, 128895, '\P{Is_Blk=alchemical}', "");
Expect(1, 128895, '\P{^Is_Blk=alchemical}', "");
Expect(0, 128896, '\p{Is_Blk=alchemical}', "");
Expect(1, 128896, '\p{^Is_Blk=alchemical}', "");
Expect(1, 128896, '\P{Is_Blk=alchemical}', "");
Expect(0, 128896, '\P{^Is_Blk=alchemical}', "");
Expect(1, 128895, '\p{Is_Blk=__alchemical}', "");
Expect(0, 128895, '\p{^Is_Blk=__alchemical}', "");
Expect(0, 128895, '\P{Is_Blk=__alchemical}', "");
Expect(1, 128895, '\P{^Is_Blk=__alchemical}', "");
Expect(0, 128896, '\p{Is_Blk=__alchemical}', "");
Expect(1, 128896, '\p{^Is_Blk=__alchemical}', "");
Expect(1, 128896, '\P{Is_Blk=__alchemical}', "");
Expect(0, 128896, '\P{^Is_Blk=__alchemical}', "");
Error('\p{Block=:=Supplemental_Mathematical_operators}');
Error('\P{Block=:=Supplemental_Mathematical_operators}');
Expect(1, 11007, '\p{Block=supplementalmathematicaloperators}', "");
Expect(0, 11007, '\p{^Block=supplementalmathematicaloperators}', "");
Expect(0, 11007, '\P{Block=supplementalmathematicaloperators}', "");
Expect(1, 11007, '\P{^Block=supplementalmathematicaloperators}', "");
Expect(0, 11008, '\p{Block=supplementalmathematicaloperators}', "");
Expect(1, 11008, '\p{^Block=supplementalmathematicaloperators}', "");
Expect(1, 11008, '\P{Block=supplementalmathematicaloperators}', "");
Expect(0, 11008, '\P{^Block=supplementalmathematicaloperators}', "");
Expect(1, 11007, '\p{Block=-	supplemental_Mathematical_OPERATORS}', "");
Expect(0, 11007, '\p{^Block=-	supplemental_Mathematical_OPERATORS}', "");
Expect(0, 11007, '\P{Block=-	supplemental_Mathematical_OPERATORS}', "");
Expect(1, 11007, '\P{^Block=-	supplemental_Mathematical_OPERATORS}', "");
Expect(0, 11008, '\p{Block=-	supplemental_Mathematical_OPERATORS}', "");
Expect(1, 11008, '\p{^Block=-	supplemental_Mathematical_OPERATORS}', "");
Expect(1, 11008, '\P{Block=-	supplemental_Mathematical_OPERATORS}', "");
Expect(0, 11008, '\P{^Block=-	supplemental_Mathematical_OPERATORS}', "");
Error('\p{Blk=:=-Sup_Math_Operators}');
Error('\P{Blk=:=-Sup_Math_Operators}');
Expect(1, 11007, '\p{Blk=supmathoperators}', "");
Expect(0, 11007, '\p{^Blk=supmathoperators}', "");
Expect(0, 11007, '\P{Blk=supmathoperators}', "");
Expect(1, 11007, '\P{^Blk=supmathoperators}', "");
Expect(0, 11008, '\p{Blk=supmathoperators}', "");
Expect(1, 11008, '\p{^Blk=supmathoperators}', "");
Expect(1, 11008, '\P{Blk=supmathoperators}', "");
Expect(0, 11008, '\P{^Blk=supmathoperators}', "");
Expect(1, 11007, '\p{Blk:_	Sup_MATH_operators}', "");
Expect(0, 11007, '\p{^Blk:_	Sup_MATH_operators}', "");
Expect(0, 11007, '\P{Blk:_	Sup_MATH_operators}', "");
Expect(1, 11007, '\P{^Blk:_	Sup_MATH_operators}', "");
Expect(0, 11008, '\p{Blk:_	Sup_MATH_operators}', "");
Expect(1, 11008, '\p{^Blk:_	Sup_MATH_operators}', "");
Expect(1, 11008, '\P{Blk:_	Sup_MATH_operators}', "");
Expect(0, 11008, '\P{^Blk:_	Sup_MATH_operators}', "");
Error('\p{Is_Block=:=_SUPPLEMENTAL_mathematical_Operators}');
Error('\P{Is_Block=:=_SUPPLEMENTAL_mathematical_Operators}');
Expect(1, 11007, '\p{Is_Block=supplementalmathematicaloperators}', "");
Expect(0, 11007, '\p{^Is_Block=supplementalmathematicaloperators}', "");
Expect(0, 11007, '\P{Is_Block=supplementalmathematicaloperators}', "");
Expect(1, 11007, '\P{^Is_Block=supplementalmathematicaloperators}', "");
Expect(0, 11008, '\p{Is_Block=supplementalmathematicaloperators}', "");
Expect(1, 11008, '\p{^Is_Block=supplementalmathematicaloperators}', "");
Expect(1, 11008, '\P{Is_Block=supplementalmathematicaloperators}', "");
Expect(0, 11008, '\P{^Is_Block=supplementalmathematicaloperators}', "");
Expect(1, 11007, '\p{Is_Block=_-Supplemental_mathematical_Operators}', "");
Expect(0, 11007, '\p{^Is_Block=_-Supplemental_mathematical_Operators}', "");
Expect(0, 11007, '\P{Is_Block=_-Supplemental_mathematical_Operators}', "");
Expect(1, 11007, '\P{^Is_Block=_-Supplemental_mathematical_Operators}', "");
Expect(0, 11008, '\p{Is_Block=_-Supplemental_mathematical_Operators}', "");
Expect(1, 11008, '\p{^Is_Block=_-Supplemental_mathematical_Operators}', "");
Expect(1, 11008, '\P{Is_Block=_-Supplemental_mathematical_Operators}', "");
Expect(0, 11008, '\P{^Is_Block=_-Supplemental_mathematical_Operators}', "");
Error('\p{Is_Blk=	_Sup_Math_OPERATORS/a/}');
Error('\P{Is_Blk=	_Sup_Math_OPERATORS/a/}');
Expect(1, 11007, '\p{Is_Blk=supmathoperators}', "");
Expect(0, 11007, '\p{^Is_Blk=supmathoperators}', "");
Expect(0, 11007, '\P{Is_Blk=supmathoperators}', "");
Expect(1, 11007, '\P{^Is_Blk=supmathoperators}', "");
Expect(0, 11008, '\p{Is_Blk=supmathoperators}', "");
Expect(1, 11008, '\p{^Is_Blk=supmathoperators}', "");
Expect(1, 11008, '\P{Is_Blk=supmathoperators}', "");
Expect(0, 11008, '\P{^Is_Blk=supmathoperators}', "");
Expect(1, 11007, '\p{Is_Blk= _sup_Math_operators}', "");
Expect(0, 11007, '\p{^Is_Blk= _sup_Math_operators}', "");
Expect(0, 11007, '\P{Is_Blk= _sup_Math_operators}', "");
Expect(1, 11007, '\P{^Is_Blk= _sup_Math_operators}', "");
Expect(0, 11008, '\p{Is_Blk= _sup_Math_operators}', "");
Expect(1, 11008, '\p{^Is_Blk= _sup_Math_operators}', "");
Expect(1, 11008, '\P{Is_Blk= _sup_Math_operators}', "");
Expect(0, 11008, '\P{^Is_Blk= _sup_Math_operators}', "");
Error('\p{Block= :=Shavian}');
Error('\P{Block= :=Shavian}');
Expect(1, 66687, '\p{Block=shavian}', "");
Expect(0, 66687, '\p{^Block=shavian}', "");
Expect(0, 66687, '\P{Block=shavian}', "");
Expect(1, 66687, '\P{^Block=shavian}', "");
Expect(0, 66688, '\p{Block=shavian}', "");
Expect(1, 66688, '\p{^Block=shavian}', "");
Expect(1, 66688, '\P{Block=shavian}', "");
Expect(0, 66688, '\P{^Block=shavian}', "");
Expect(1, 66687, '\p{Block=-SHAVIAN}', "");
Expect(0, 66687, '\p{^Block=-SHAVIAN}', "");
Expect(0, 66687, '\P{Block=-SHAVIAN}', "");
Expect(1, 66687, '\P{^Block=-SHAVIAN}', "");
Expect(0, 66688, '\p{Block=-SHAVIAN}', "");
Expect(1, 66688, '\p{^Block=-SHAVIAN}', "");
Expect(1, 66688, '\P{Block=-SHAVIAN}', "");
Expect(0, 66688, '\P{^Block=-SHAVIAN}', "");
Error('\p{Blk=/a/Shavian}');
Error('\P{Blk=/a/Shavian}');
Expect(1, 66687, '\p{Blk:   shavian}', "");
Expect(0, 66687, '\p{^Blk:   shavian}', "");
Expect(0, 66687, '\P{Blk:   shavian}', "");
Expect(1, 66687, '\P{^Blk:   shavian}', "");
Expect(0, 66688, '\p{Blk:   shavian}', "");
Expect(1, 66688, '\p{^Blk:   shavian}', "");
Expect(1, 66688, '\P{Blk:   shavian}', "");
Expect(0, 66688, '\P{^Blk:   shavian}', "");
Expect(1, 66687, '\p{Blk:  Shavian}', "");
Expect(0, 66687, '\p{^Blk:  Shavian}', "");
Expect(0, 66687, '\P{Blk:  Shavian}', "");
Expect(1, 66687, '\P{^Blk:  Shavian}', "");
Expect(0, 66688, '\p{Blk:  Shavian}', "");
Expect(1, 66688, '\p{^Blk:  Shavian}', "");
Expect(1, 66688, '\P{Blk:  Shavian}', "");
Expect(0, 66688, '\P{^Blk:  Shavian}', "");
Error('\p{Is_Block=:=_shavian}');
Error('\P{Is_Block=:=_shavian}');
Expect(1, 66687, '\p{Is_Block=shavian}', "");
Expect(0, 66687, '\p{^Is_Block=shavian}', "");
Expect(0, 66687, '\P{Is_Block=shavian}', "");
Expect(1, 66687, '\P{^Is_Block=shavian}', "");
Expect(0, 66688, '\p{Is_Block=shavian}', "");
Expect(1, 66688, '\p{^Is_Block=shavian}', "");
Expect(1, 66688, '\P{Is_Block=shavian}', "");
Expect(0, 66688, '\P{^Is_Block=shavian}', "");
Expect(1, 66687, '\p{Is_Block= shavian}', "");
Expect(0, 66687, '\p{^Is_Block= shavian}', "");
Expect(0, 66687, '\P{Is_Block= shavian}', "");
Expect(1, 66687, '\P{^Is_Block= shavian}', "");
Expect(0, 66688, '\p{Is_Block= shavian}', "");
Expect(1, 66688, '\p{^Is_Block= shavian}', "");
Expect(1, 66688, '\P{Is_Block= shavian}', "");
Expect(0, 66688, '\P{^Is_Block= shavian}', "");
Error('\p{Is_Blk:    :=SHAVIAN}');
Error('\P{Is_Blk:    :=SHAVIAN}');
Expect(1, 66687, '\p{Is_Blk=shavian}', "");
Expect(0, 66687, '\p{^Is_Blk=shavian}', "");
Expect(0, 66687, '\P{Is_Blk=shavian}', "");
Expect(1, 66687, '\P{^Is_Blk=shavian}', "");
Expect(0, 66688, '\p{Is_Blk=shavian}', "");
Expect(1, 66688, '\p{^Is_Blk=shavian}', "");
Expect(1, 66688, '\P{Is_Blk=shavian}', "");
Expect(0, 66688, '\P{^Is_Blk=shavian}', "");
Expect(1, 66687, '\p{Is_Blk=__shavian}', "");
Expect(0, 66687, '\p{^Is_Blk=__shavian}', "");
Expect(0, 66687, '\P{Is_Blk=__shavian}', "");
Expect(1, 66687, '\P{^Is_Blk=__shavian}', "");
Expect(0, 66688, '\p{Is_Blk=__shavian}', "");
Expect(1, 66688, '\p{^Is_Blk=__shavian}', "");
Expect(1, 66688, '\P{Is_Blk=__shavian}', "");
Expect(0, 66688, '\P{^Is_Blk=__shavian}', "");
Error('\p{Block=- Lydian/a/}');
Error('\P{Block=- Lydian/a/}');
Expect(1, 67903, '\p{Block=lydian}', "");
Expect(0, 67903, '\p{^Block=lydian}', "");
Expect(0, 67903, '\P{Block=lydian}', "");
Expect(1, 67903, '\P{^Block=lydian}', "");
Expect(0, 67904, '\p{Block=lydian}', "");
Expect(1, 67904, '\p{^Block=lydian}', "");
Expect(1, 67904, '\P{Block=lydian}', "");
Expect(0, 67904, '\P{^Block=lydian}', "");
Expect(1, 67903, '\p{Block=-	LYDIAN}', "");
Expect(0, 67903, '\p{^Block=-	LYDIAN}', "");
Expect(0, 67903, '\P{Block=-	LYDIAN}', "");
Expect(1, 67903, '\P{^Block=-	LYDIAN}', "");
Expect(0, 67904, '\p{Block=-	LYDIAN}', "");
Expect(1, 67904, '\p{^Block=-	LYDIAN}', "");
Expect(1, 67904, '\P{Block=-	LYDIAN}', "");
Expect(0, 67904, '\P{^Block=-	LYDIAN}', "");
Error('\p{Blk=/a/ _lydian}');
Error('\P{Blk=/a/ _lydian}');
Expect(1, 67903, '\p{Blk=lydian}', "");
Expect(0, 67903, '\p{^Blk=lydian}', "");
Expect(0, 67903, '\P{Blk=lydian}', "");
Expect(1, 67903, '\P{^Blk=lydian}', "");
Expect(0, 67904, '\p{Blk=lydian}', "");
Expect(1, 67904, '\p{^Blk=lydian}', "");
Expect(1, 67904, '\P{Blk=lydian}', "");
Expect(0, 67904, '\P{^Blk=lydian}', "");
Expect(1, 67903, '\p{Blk=		Lydian}', "");
Expect(0, 67903, '\p{^Blk=		Lydian}', "");
Expect(0, 67903, '\P{Blk=		Lydian}', "");
Expect(1, 67903, '\P{^Blk=		Lydian}', "");
Expect(0, 67904, '\p{Blk=		Lydian}', "");
Expect(1, 67904, '\p{^Blk=		Lydian}', "");
Expect(1, 67904, '\P{Blk=		Lydian}', "");
Expect(0, 67904, '\P{^Blk=		Lydian}', "");
Error('\p{Is_Block=:=	lydian}');
Error('\P{Is_Block=:=	lydian}');
Expect(1, 67903, '\p{Is_Block=lydian}', "");
Expect(0, 67903, '\p{^Is_Block=lydian}', "");
Expect(0, 67903, '\P{Is_Block=lydian}', "");
Expect(1, 67903, '\P{^Is_Block=lydian}', "");
Expect(0, 67904, '\p{Is_Block=lydian}', "");
Expect(1, 67904, '\p{^Is_Block=lydian}', "");
Expect(1, 67904, '\P{Is_Block=lydian}', "");
Expect(0, 67904, '\P{^Is_Block=lydian}', "");
Expect(1, 67903, '\p{Is_Block=-lydian}', "");
Expect(0, 67903, '\p{^Is_Block=-lydian}', "");
Expect(0, 67903, '\P{Is_Block=-lydian}', "");
Expect(1, 67903, '\P{^Is_Block=-lydian}', "");
Expect(0, 67904, '\p{Is_Block=-lydian}', "");
Expect(1, 67904, '\p{^Is_Block=-lydian}', "");
Expect(1, 67904, '\P{Is_Block=-lydian}', "");
Expect(0, 67904, '\P{^Is_Block=-lydian}', "");
Error('\p{Is_Blk=	Lydian:=}');
Error('\P{Is_Blk=	Lydian:=}');
Expect(1, 67903, '\p{Is_Blk=lydian}', "");
Expect(0, 67903, '\p{^Is_Blk=lydian}', "");
Expect(0, 67903, '\P{Is_Blk=lydian}', "");
Expect(1, 67903, '\P{^Is_Blk=lydian}', "");
Expect(0, 67904, '\p{Is_Blk=lydian}', "");
Expect(1, 67904, '\p{^Is_Blk=lydian}', "");
Expect(1, 67904, '\P{Is_Blk=lydian}', "");
Expect(0, 67904, '\P{^Is_Blk=lydian}', "");
Expect(1, 67903, '\p{Is_Blk=	-lydian}', "");
Expect(0, 67903, '\p{^Is_Blk=	-lydian}', "");
Expect(0, 67903, '\P{Is_Blk=	-lydian}', "");
Expect(1, 67903, '\P{^Is_Blk=	-lydian}', "");
Expect(0, 67904, '\p{Is_Blk=	-lydian}', "");
Expect(1, 67904, '\p{^Is_Blk=	-lydian}', "");
Expect(1, 67904, '\P{Is_Blk=	-lydian}', "");
Expect(0, 67904, '\P{^Is_Blk=	-lydian}', "");
Error('\p{Block=-:=MONGOLIAN}');
Error('\P{Block=-:=MONGOLIAN}');
Expect(1, 6319, '\p{Block=mongolian}', "");
Expect(0, 6319, '\p{^Block=mongolian}', "");
Expect(0, 6319, '\P{Block=mongolian}', "");
Expect(1, 6319, '\P{^Block=mongolian}', "");
Expect(0, 6320, '\p{Block=mongolian}', "");
Expect(1, 6320, '\p{^Block=mongolian}', "");
Expect(1, 6320, '\P{Block=mongolian}', "");
Expect(0, 6320, '\P{^Block=mongolian}', "");
Expect(1, 6319, '\p{Block=_Mongolian}', "");
Expect(0, 6319, '\p{^Block=_Mongolian}', "");
Expect(0, 6319, '\P{Block=_Mongolian}', "");
Expect(1, 6319, '\P{^Block=_Mongolian}', "");
Expect(0, 6320, '\p{Block=_Mongolian}', "");
Expect(1, 6320, '\p{^Block=_Mongolian}', "");
Expect(1, 6320, '\P{Block=_Mongolian}', "");
Expect(0, 6320, '\P{^Block=_Mongolian}', "");
Error('\p{Blk:    /a/mongolian}');
Error('\P{Blk:    /a/mongolian}');
Expect(1, 6319, '\p{Blk=mongolian}', "");
Expect(0, 6319, '\p{^Blk=mongolian}', "");
Expect(0, 6319, '\P{Blk=mongolian}', "");
Expect(1, 6319, '\P{^Blk=mongolian}', "");
Expect(0, 6320, '\p{Blk=mongolian}', "");
Expect(1, 6320, '\p{^Blk=mongolian}', "");
Expect(1, 6320, '\P{Blk=mongolian}', "");
Expect(0, 6320, '\P{^Blk=mongolian}', "");
Expect(1, 6319, '\p{Blk=-_mongolian}', "");
Expect(0, 6319, '\p{^Blk=-_mongolian}', "");
Expect(0, 6319, '\P{Blk=-_mongolian}', "");
Expect(1, 6319, '\P{^Blk=-_mongolian}', "");
Expect(0, 6320, '\p{Blk=-_mongolian}', "");
Expect(1, 6320, '\p{^Blk=-_mongolian}', "");
Expect(1, 6320, '\P{Blk=-_mongolian}', "");
Expect(0, 6320, '\P{^Blk=-_mongolian}', "");
Error('\p{Is_Block=/a/ Mongolian}');
Error('\P{Is_Block=/a/ Mongolian}');
Expect(1, 6319, '\p{Is_Block=mongolian}', "");
Expect(0, 6319, '\p{^Is_Block=mongolian}', "");
Expect(0, 6319, '\P{Is_Block=mongolian}', "");
Expect(1, 6319, '\P{^Is_Block=mongolian}', "");
Expect(0, 6320, '\p{Is_Block=mongolian}', "");
Expect(1, 6320, '\p{^Is_Block=mongolian}', "");
Expect(1, 6320, '\P{Is_Block=mongolian}', "");
Expect(0, 6320, '\P{^Is_Block=mongolian}', "");
Expect(1, 6319, '\p{Is_Block=_Mongolian}', "");
Expect(0, 6319, '\p{^Is_Block=_Mongolian}', "");
Expect(0, 6319, '\P{Is_Block=_Mongolian}', "");
Expect(1, 6319, '\P{^Is_Block=_Mongolian}', "");
Expect(0, 6320, '\p{Is_Block=_Mongolian}', "");
Expect(1, 6320, '\p{^Is_Block=_Mongolian}', "");
Expect(1, 6320, '\P{Is_Block=_Mongolian}', "");
Expect(0, 6320, '\P{^Is_Block=_Mongolian}', "");
Error('\p{Is_Blk=/a/--MONGOLIAN}');
Error('\P{Is_Blk=/a/--MONGOLIAN}');
Expect(1, 6319, '\p{Is_Blk=mongolian}', "");
Expect(0, 6319, '\p{^Is_Blk=mongolian}', "");
Expect(0, 6319, '\P{Is_Blk=mongolian}', "");
Expect(1, 6319, '\P{^Is_Blk=mongolian}', "");
Expect(0, 6320, '\p{Is_Blk=mongolian}', "");
Expect(1, 6320, '\p{^Is_Blk=mongolian}', "");
Expect(1, 6320, '\P{Is_Blk=mongolian}', "");
Expect(0, 6320, '\P{^Is_Blk=mongolian}', "");
Expect(1, 6319, '\p{Is_Blk=-Mongolian}', "");
Expect(0, 6319, '\p{^Is_Blk=-Mongolian}', "");
Expect(0, 6319, '\P{Is_Blk=-Mongolian}', "");
Expect(1, 6319, '\P{^Is_Blk=-Mongolian}', "");
Expect(0, 6320, '\p{Is_Blk=-Mongolian}', "");
Expect(1, 6320, '\p{^Is_Blk=-Mongolian}', "");
Expect(1, 6320, '\P{Is_Blk=-Mongolian}', "");
Expect(0, 6320, '\P{^Is_Blk=-Mongolian}', "");
Error('\p{Block=/a/	Hiragana}');
Error('\P{Block=/a/	Hiragana}');
Expect(1, 12447, '\p{Block=hiragana}', "");
Expect(0, 12447, '\p{^Block=hiragana}', "");
Expect(0, 12447, '\P{Block=hiragana}', "");
Expect(1, 12447, '\P{^Block=hiragana}', "");
Expect(0, 12448, '\p{Block=hiragana}', "");
Expect(1, 12448, '\p{^Block=hiragana}', "");
Expect(1, 12448, '\P{Block=hiragana}', "");
Expect(0, 12448, '\P{^Block=hiragana}', "");
Expect(1, 12447, '\p{Block:	 hiragana}', "");
Expect(0, 12447, '\p{^Block:	 hiragana}', "");
Expect(0, 12447, '\P{Block:	 hiragana}', "");
Expect(1, 12447, '\P{^Block:	 hiragana}', "");
Expect(0, 12448, '\p{Block:	 hiragana}', "");
Expect(1, 12448, '\p{^Block:	 hiragana}', "");
Expect(1, 12448, '\P{Block:	 hiragana}', "");
Expect(0, 12448, '\P{^Block:	 hiragana}', "");
Error('\p{Blk=-hiragana:=}');
Error('\P{Blk=-hiragana:=}');
Expect(1, 12447, '\p{Blk=hiragana}', "");
Expect(0, 12447, '\p{^Blk=hiragana}', "");
Expect(0, 12447, '\P{Blk=hiragana}', "");
Expect(1, 12447, '\P{^Blk=hiragana}', "");
Expect(0, 12448, '\p{Blk=hiragana}', "");
Expect(1, 12448, '\p{^Blk=hiragana}', "");
Expect(1, 12448, '\P{Blk=hiragana}', "");
Expect(0, 12448, '\P{^Blk=hiragana}', "");
Expect(1, 12447, '\p{Blk:-	HIRAGANA}', "");
Expect(0, 12447, '\p{^Blk:-	HIRAGANA}', "");
Expect(0, 12447, '\P{Blk:-	HIRAGANA}', "");
Expect(1, 12447, '\P{^Blk:-	HIRAGANA}', "");
Expect(0, 12448, '\p{Blk:-	HIRAGANA}', "");
Expect(1, 12448, '\p{^Blk:-	HIRAGANA}', "");
Expect(1, 12448, '\P{Blk:-	HIRAGANA}', "");
Expect(0, 12448, '\P{^Blk:-	HIRAGANA}', "");
Error('\p{Is_Block=/a/	_Hiragana}');
Error('\P{Is_Block=/a/	_Hiragana}');
Expect(1, 12447, '\p{Is_Block=hiragana}', "");
Expect(0, 12447, '\p{^Is_Block=hiragana}', "");
Expect(0, 12447, '\P{Is_Block=hiragana}', "");
Expect(1, 12447, '\P{^Is_Block=hiragana}', "");
Expect(0, 12448, '\p{Is_Block=hiragana}', "");
Expect(1, 12448, '\p{^Is_Block=hiragana}', "");
Expect(1, 12448, '\P{Is_Block=hiragana}', "");
Expect(0, 12448, '\P{^Is_Block=hiragana}', "");
Expect(1, 12447, '\p{Is_Block=_	hiragana}', "");
Expect(0, 12447, '\p{^Is_Block=_	hiragana}', "");
Expect(0, 12447, '\P{Is_Block=_	hiragana}', "");
Expect(1, 12447, '\P{^Is_Block=_	hiragana}', "");
Expect(0, 12448, '\p{Is_Block=_	hiragana}', "");
Expect(1, 12448, '\p{^Is_Block=_	hiragana}', "");
Expect(1, 12448, '\P{Is_Block=_	hiragana}', "");
Expect(0, 12448, '\P{^Is_Block=_	hiragana}', "");
Error('\p{Is_Blk=--Hiragana/a/}');
Error('\P{Is_Blk=--Hiragana/a/}');
Expect(1, 12447, '\p{Is_Blk: hiragana}', "");
Expect(0, 12447, '\p{^Is_Blk: hiragana}', "");
Expect(0, 12447, '\P{Is_Blk: hiragana}', "");
Expect(1, 12447, '\P{^Is_Blk: hiragana}', "");
Expect(0, 12448, '\p{Is_Blk: hiragana}', "");
Expect(1, 12448, '\p{^Is_Blk: hiragana}', "");
Expect(1, 12448, '\P{Is_Blk: hiragana}', "");
Expect(0, 12448, '\P{^Is_Blk: hiragana}', "");
Expect(1, 12447, '\p{Is_Blk=  Hiragana}', "");
Expect(0, 12447, '\p{^Is_Blk=  Hiragana}', "");
Expect(0, 12447, '\P{Is_Blk=  Hiragana}', "");
Expect(1, 12447, '\P{^Is_Blk=  Hiragana}', "");
Expect(0, 12448, '\p{Is_Blk=  Hiragana}', "");
Expect(1, 12448, '\p{^Is_Blk=  Hiragana}', "");
Expect(1, 12448, '\P{Is_Blk=  Hiragana}', "");
Expect(0, 12448, '\P{^Is_Blk=  Hiragana}', "");
Error('\p{Block:	_:=lepcha}');
Error('\P{Block:	_:=lepcha}');
Expect(1, 7247, '\p{Block=lepcha}', "");
Expect(0, 7247, '\p{^Block=lepcha}', "");
Expect(0, 7247, '\P{Block=lepcha}', "");
Expect(1, 7247, '\P{^Block=lepcha}', "");
Expect(0, 7248, '\p{Block=lepcha}', "");
Expect(1, 7248, '\p{^Block=lepcha}', "");
Expect(1, 7248, '\P{Block=lepcha}', "");
Expect(0, 7248, '\P{^Block=lepcha}', "");
Expect(1, 7247, '\p{Block=		Lepcha}', "");
Expect(0, 7247, '\p{^Block=		Lepcha}', "");
Expect(0, 7247, '\P{Block=		Lepcha}', "");
Expect(1, 7247, '\P{^Block=		Lepcha}', "");
Expect(0, 7248, '\p{Block=		Lepcha}', "");
Expect(1, 7248, '\p{^Block=		Lepcha}', "");
Expect(1, 7248, '\P{Block=		Lepcha}', "");
Expect(0, 7248, '\P{^Block=		Lepcha}', "");
Error('\p{Blk=-/a/Lepcha}');
Error('\P{Blk=-/a/Lepcha}');
Expect(1, 7247, '\p{Blk=lepcha}', "");
Expect(0, 7247, '\p{^Blk=lepcha}', "");
Expect(0, 7247, '\P{Blk=lepcha}', "");
Expect(1, 7247, '\P{^Blk=lepcha}', "");
Expect(0, 7248, '\p{Blk=lepcha}', "");
Expect(1, 7248, '\p{^Blk=lepcha}', "");
Expect(1, 7248, '\P{Blk=lepcha}', "");
Expect(0, 7248, '\P{^Blk=lepcha}', "");
Expect(1, 7247, '\p{Blk=_-LEPCHA}', "");
Expect(0, 7247, '\p{^Blk=_-LEPCHA}', "");
Expect(0, 7247, '\P{Blk=_-LEPCHA}', "");
Expect(1, 7247, '\P{^Blk=_-LEPCHA}', "");
Expect(0, 7248, '\p{Blk=_-LEPCHA}', "");
Expect(1, 7248, '\p{^Blk=_-LEPCHA}', "");
Expect(1, 7248, '\P{Blk=_-LEPCHA}', "");
Expect(0, 7248, '\P{^Blk=_-LEPCHA}', "");
Error('\p{Is_Block=_:=lepcha}');
Error('\P{Is_Block=_:=lepcha}');
Expect(1, 7247, '\p{Is_Block:lepcha}', "");
Expect(0, 7247, '\p{^Is_Block:lepcha}', "");
Expect(0, 7247, '\P{Is_Block:lepcha}', "");
Expect(1, 7247, '\P{^Is_Block:lepcha}', "");
Expect(0, 7248, '\p{Is_Block:lepcha}', "");
Expect(1, 7248, '\p{^Is_Block:lepcha}', "");
Expect(1, 7248, '\P{Is_Block:lepcha}', "");
Expect(0, 7248, '\P{^Is_Block:lepcha}', "");
Expect(1, 7247, '\p{Is_Block= -Lepcha}', "");
Expect(0, 7247, '\p{^Is_Block= -Lepcha}', "");
Expect(0, 7247, '\P{Is_Block= -Lepcha}', "");
Expect(1, 7247, '\P{^Is_Block= -Lepcha}', "");
Expect(0, 7248, '\p{Is_Block= -Lepcha}', "");
Expect(1, 7248, '\p{^Is_Block= -Lepcha}', "");
Expect(1, 7248, '\P{Is_Block= -Lepcha}', "");
Expect(0, 7248, '\P{^Is_Block= -Lepcha}', "");
Error('\p{Is_Blk=-_Lepcha:=}');
Error('\P{Is_Blk=-_Lepcha:=}');
Expect(1, 7247, '\p{Is_Blk=lepcha}', "");
Expect(0, 7247, '\p{^Is_Blk=lepcha}', "");
Expect(0, 7247, '\P{Is_Blk=lepcha}', "");
Expect(1, 7247, '\P{^Is_Blk=lepcha}', "");
Expect(0, 7248, '\p{Is_Blk=lepcha}', "");
Expect(1, 7248, '\p{^Is_Blk=lepcha}', "");
Expect(1, 7248, '\P{Is_Blk=lepcha}', "");
Expect(0, 7248, '\P{^Is_Blk=lepcha}', "");
Expect(1, 7247, '\p{Is_Blk=	LEPCHA}', "");
Expect(0, 7247, '\p{^Is_Blk=	LEPCHA}', "");
Expect(0, 7247, '\P{Is_Blk=	LEPCHA}', "");
Expect(1, 7247, '\P{^Is_Blk=	LEPCHA}', "");
Expect(0, 7248, '\p{Is_Blk=	LEPCHA}', "");
Expect(1, 7248, '\p{^Is_Blk=	LEPCHA}', "");
Expect(1, 7248, '\P{Is_Blk=	LEPCHA}', "");
Expect(0, 7248, '\P{^Is_Blk=	LEPCHA}', "");
Error('\p{Block=	CJK_UNIFIED_Ideographs/a/}');
Error('\P{Block=	CJK_UNIFIED_Ideographs/a/}');
Expect(1, 40959, '\p{Block=cjkunifiedideographs}', "");
Expect(0, 40959, '\p{^Block=cjkunifiedideographs}', "");
Expect(0, 40959, '\P{Block=cjkunifiedideographs}', "");
Expect(1, 40959, '\P{^Block=cjkunifiedideographs}', "");
Expect(0, 40960, '\p{Block=cjkunifiedideographs}', "");
Expect(1, 40960, '\p{^Block=cjkunifiedideographs}', "");
Expect(1, 40960, '\P{Block=cjkunifiedideographs}', "");
Expect(0, 40960, '\P{^Block=cjkunifiedideographs}', "");
Expect(1, 40959, '\p{Block= CJK_UNIFIED_Ideographs}', "");
Expect(0, 40959, '\p{^Block= CJK_UNIFIED_Ideographs}', "");
Expect(0, 40959, '\P{Block= CJK_UNIFIED_Ideographs}', "");
Expect(1, 40959, '\P{^Block= CJK_UNIFIED_Ideographs}', "");
Expect(0, 40960, '\p{Block= CJK_UNIFIED_Ideographs}', "");
Expect(1, 40960, '\p{^Block= CJK_UNIFIED_Ideographs}', "");
Expect(1, 40960, '\P{Block= CJK_UNIFIED_Ideographs}', "");
Expect(0, 40960, '\P{^Block= CJK_UNIFIED_Ideographs}', "");
Error('\p{Blk=:=_-CJK}');
Error('\P{Blk=:=_-CJK}');
Expect(1, 40959, '\p{Blk: cjk}', "");
Expect(0, 40959, '\p{^Blk: cjk}', "");
Expect(0, 40959, '\P{Blk: cjk}', "");
Expect(1, 40959, '\P{^Blk: cjk}', "");
Expect(0, 40960, '\p{Blk: cjk}', "");
Expect(1, 40960, '\p{^Blk: cjk}', "");
Expect(1, 40960, '\P{Blk: cjk}', "");
Expect(0, 40960, '\P{^Blk: cjk}', "");
Expect(1, 40959, '\p{Blk=-	CJK}', "");
Expect(0, 40959, '\p{^Blk=-	CJK}', "");
Expect(0, 40959, '\P{Blk=-	CJK}', "");
Expect(1, 40959, '\P{^Blk=-	CJK}', "");
Expect(0, 40960, '\p{Blk=-	CJK}', "");
Expect(1, 40960, '\p{^Blk=-	CJK}', "");
Expect(1, 40960, '\P{Blk=-	CJK}', "");
Expect(0, 40960, '\P{^Blk=-	CJK}', "");
Error('\p{Is_Block= /a/cjk_UNIFIED_IDEOGRAPHS}');
Error('\P{Is_Block= /a/cjk_UNIFIED_IDEOGRAPHS}');
Expect(1, 40959, '\p{Is_Block=cjkunifiedideographs}', "");
Expect(0, 40959, '\p{^Is_Block=cjkunifiedideographs}', "");
Expect(0, 40959, '\P{Is_Block=cjkunifiedideographs}', "");
Expect(1, 40959, '\P{^Is_Block=cjkunifiedideographs}', "");
Expect(0, 40960, '\p{Is_Block=cjkunifiedideographs}', "");
Expect(1, 40960, '\p{^Is_Block=cjkunifiedideographs}', "");
Expect(1, 40960, '\P{Is_Block=cjkunifiedideographs}', "");
Expect(0, 40960, '\P{^Is_Block=cjkunifiedideographs}', "");
Expect(1, 40959, '\p{Is_Block=_-CJK_Unified_ideographs}', "");
Expect(0, 40959, '\p{^Is_Block=_-CJK_Unified_ideographs}', "");
Expect(0, 40959, '\P{Is_Block=_-CJK_Unified_ideographs}', "");
Expect(1, 40959, '\P{^Is_Block=_-CJK_Unified_ideographs}', "");
Expect(0, 40960, '\p{Is_Block=_-CJK_Unified_ideographs}', "");
Expect(1, 40960, '\p{^Is_Block=_-CJK_Unified_ideographs}', "");
Expect(1, 40960, '\P{Is_Block=_-CJK_Unified_ideographs}', "");
Expect(0, 40960, '\P{^Is_Block=_-CJK_Unified_ideographs}', "");
Error('\p{Is_Blk=- cjk:=}');
Error('\P{Is_Blk=- cjk:=}');
Expect(1, 40959, '\p{Is_Blk:cjk}', "");
Expect(0, 40959, '\p{^Is_Blk:cjk}', "");
Expect(0, 40959, '\P{Is_Blk:cjk}', "");
Expect(1, 40959, '\P{^Is_Blk:cjk}', "");
Expect(0, 40960, '\p{Is_Blk:cjk}', "");
Expect(1, 40960, '\p{^Is_Blk:cjk}', "");
Expect(1, 40960, '\P{Is_Blk:cjk}', "");
Expect(0, 40960, '\P{^Is_Blk:cjk}', "");
Expect(1, 40959, '\p{Is_Blk=-_cjk}', "");
Expect(0, 40959, '\p{^Is_Blk=-_cjk}', "");
Expect(0, 40959, '\P{Is_Blk=-_cjk}', "");
Expect(1, 40959, '\P{^Is_Blk=-_cjk}', "");
Expect(0, 40960, '\p{Is_Blk=-_cjk}', "");
Expect(1, 40960, '\p{^Is_Blk=-_cjk}', "");
Expect(1, 40960, '\P{Is_Blk=-_cjk}', "");
Expect(0, 40960, '\P{^Is_Blk=-_cjk}', "");
Error('\p{Block= /a/Braille_patterns}');
Error('\P{Block= /a/Braille_patterns}');
Expect(1, 10495, '\p{Block:braillepatterns}', "");
Expect(0, 10495, '\p{^Block:braillepatterns}', "");
Expect(0, 10495, '\P{Block:braillepatterns}', "");
Expect(1, 10495, '\P{^Block:braillepatterns}', "");
Expect(0, 10496, '\p{Block:braillepatterns}', "");
Expect(1, 10496, '\p{^Block:braillepatterns}', "");
Expect(1, 10496, '\P{Block:braillepatterns}', "");
Expect(0, 10496, '\P{^Block:braillepatterns}', "");
Expect(1, 10495, '\p{Block=	_braille_Patterns}', "");
Expect(0, 10495, '\p{^Block=	_braille_Patterns}', "");
Expect(0, 10495, '\P{Block=	_braille_Patterns}', "");
Expect(1, 10495, '\P{^Block=	_braille_Patterns}', "");
Expect(0, 10496, '\p{Block=	_braille_Patterns}', "");
Expect(1, 10496, '\p{^Block=	_braille_Patterns}', "");
Expect(1, 10496, '\P{Block=	_braille_Patterns}', "");
Expect(0, 10496, '\P{^Block=	_braille_Patterns}', "");
Error('\p{Blk=  Braille/a/}');
Error('\P{Blk=  Braille/a/}');
Expect(1, 10495, '\p{Blk=braille}', "");
Expect(0, 10495, '\p{^Blk=braille}', "");
Expect(0, 10495, '\P{Blk=braille}', "");
Expect(1, 10495, '\P{^Blk=braille}', "");
Expect(0, 10496, '\p{Blk=braille}', "");
Expect(1, 10496, '\p{^Blk=braille}', "");
Expect(1, 10496, '\P{Blk=braille}', "");
Expect(0, 10496, '\P{^Blk=braille}', "");
Expect(1, 10495, '\p{Blk=-	braille}', "");
Expect(0, 10495, '\p{^Blk=-	braille}', "");
Expect(0, 10495, '\P{Blk=-	braille}', "");
Expect(1, 10495, '\P{^Blk=-	braille}', "");
Expect(0, 10496, '\p{Blk=-	braille}', "");
Expect(1, 10496, '\p{^Blk=-	braille}', "");
Expect(1, 10496, '\P{Blk=-	braille}', "");
Expect(0, 10496, '\P{^Blk=-	braille}', "");
Error('\p{Is_Block=-_Braille_Patterns/a/}');
Error('\P{Is_Block=-_Braille_Patterns/a/}');
Expect(1, 10495, '\p{Is_Block=braillepatterns}', "");
Expect(0, 10495, '\p{^Is_Block=braillepatterns}', "");
Expect(0, 10495, '\P{Is_Block=braillepatterns}', "");
Expect(1, 10495, '\P{^Is_Block=braillepatterns}', "");
Expect(0, 10496, '\p{Is_Block=braillepatterns}', "");
Expect(1, 10496, '\p{^Is_Block=braillepatterns}', "");
Expect(1, 10496, '\P{Is_Block=braillepatterns}', "");
Expect(0, 10496, '\P{^Is_Block=braillepatterns}', "");
Expect(1, 10495, '\p{Is_Block:	- BRAILLE_Patterns}', "");
Expect(0, 10495, '\p{^Is_Block:	- BRAILLE_Patterns}', "");
Expect(0, 10495, '\P{Is_Block:	- BRAILLE_Patterns}', "");
Expect(1, 10495, '\P{^Is_Block:	- BRAILLE_Patterns}', "");
Expect(0, 10496, '\p{Is_Block:	- BRAILLE_Patterns}', "");
Expect(1, 10496, '\p{^Is_Block:	- BRAILLE_Patterns}', "");
Expect(1, 10496, '\P{Is_Block:	- BRAILLE_Patterns}', "");
Expect(0, 10496, '\P{^Is_Block:	- BRAILLE_Patterns}', "");
Error('\p{Is_Blk= /a/Braille}');
Error('\P{Is_Blk= /a/Braille}');
Expect(1, 10495, '\p{Is_Blk=braille}', "");
Expect(0, 10495, '\p{^Is_Blk=braille}', "");
Expect(0, 10495, '\P{Is_Blk=braille}', "");
Expect(1, 10495, '\P{^Is_Blk=braille}', "");
Expect(0, 10496, '\p{Is_Blk=braille}', "");
Expect(1, 10496, '\p{^Is_Blk=braille}', "");
Expect(1, 10496, '\P{Is_Blk=braille}', "");
Expect(0, 10496, '\P{^Is_Blk=braille}', "");
Expect(1, 10495, '\p{Is_Blk: 		Braille}', "");
Expect(0, 10495, '\p{^Is_Blk: 		Braille}', "");
Expect(0, 10495, '\P{Is_Blk: 		Braille}', "");
Expect(1, 10495, '\P{^Is_Blk: 		Braille}', "");
Expect(0, 10496, '\p{Is_Blk: 		Braille}', "");
Expect(1, 10496, '\p{^Is_Blk: 		Braille}', "");
Expect(1, 10496, '\P{Is_Blk: 		Braille}', "");
Expect(0, 10496, '\P{^Is_Blk: 		Braille}', "");
Error('\p{Block=-	Ancient_Greek_numbers/a/}');
Error('\P{Block=-	Ancient_Greek_numbers/a/}');
Expect(1, 65935, '\p{Block=ancientgreeknumbers}', "");
Expect(0, 65935, '\p{^Block=ancientgreeknumbers}', "");
Expect(0, 65935, '\P{Block=ancientgreeknumbers}', "");
Expect(1, 65935, '\P{^Block=ancientgreeknumbers}', "");
Expect(0, 65936, '\p{Block=ancientgreeknumbers}', "");
Expect(1, 65936, '\p{^Block=ancientgreeknumbers}', "");
Expect(1, 65936, '\P{Block=ancientgreeknumbers}', "");
Expect(0, 65936, '\P{^Block=ancientgreeknumbers}', "");
Expect(1, 65935, '\p{Block=	-Ancient_GREEK_Numbers}', "");
Expect(0, 65935, '\p{^Block=	-Ancient_GREEK_Numbers}', "");
Expect(0, 65935, '\P{Block=	-Ancient_GREEK_Numbers}', "");
Expect(1, 65935, '\P{^Block=	-Ancient_GREEK_Numbers}', "");
Expect(0, 65936, '\p{Block=	-Ancient_GREEK_Numbers}', "");
Expect(1, 65936, '\p{^Block=	-Ancient_GREEK_Numbers}', "");
Expect(1, 65936, '\P{Block=	-Ancient_GREEK_Numbers}', "");
Expect(0, 65936, '\P{^Block=	-Ancient_GREEK_Numbers}', "");
Error('\p{Blk=/a/  Ancient_GREEK_Numbers}');
Error('\P{Blk=/a/  Ancient_GREEK_Numbers}');
Expect(1, 65935, '\p{Blk=ancientgreeknumbers}', "");
Expect(0, 65935, '\p{^Blk=ancientgreeknumbers}', "");
Expect(0, 65935, '\P{Blk=ancientgreeknumbers}', "");
Expect(1, 65935, '\P{^Blk=ancientgreeknumbers}', "");
Expect(0, 65936, '\p{Blk=ancientgreeknumbers}', "");
Expect(1, 65936, '\p{^Blk=ancientgreeknumbers}', "");
Expect(1, 65936, '\P{Blk=ancientgreeknumbers}', "");
Expect(0, 65936, '\P{^Blk=ancientgreeknumbers}', "");
Expect(1, 65935, '\p{Blk= Ancient_GREEK_NUMBERS}', "");
Expect(0, 65935, '\p{^Blk= Ancient_GREEK_NUMBERS}', "");
Expect(0, 65935, '\P{Blk= Ancient_GREEK_NUMBERS}', "");
Expect(1, 65935, '\P{^Blk= Ancient_GREEK_NUMBERS}', "");
Expect(0, 65936, '\p{Blk= Ancient_GREEK_NUMBERS}', "");
Expect(1, 65936, '\p{^Blk= Ancient_GREEK_NUMBERS}', "");
Expect(1, 65936, '\P{Blk= Ancient_GREEK_NUMBERS}', "");
Expect(0, 65936, '\P{^Blk= Ancient_GREEK_NUMBERS}', "");
Error('\p{Is_Block=	/a/Ancient_Greek_Numbers}');
Error('\P{Is_Block=	/a/Ancient_Greek_Numbers}');
Expect(1, 65935, '\p{Is_Block:	ancientgreeknumbers}', "");
Expect(0, 65935, '\p{^Is_Block:	ancientgreeknumbers}', "");
Expect(0, 65935, '\P{Is_Block:	ancientgreeknumbers}', "");
Expect(1, 65935, '\P{^Is_Block:	ancientgreeknumbers}', "");
Expect(0, 65936, '\p{Is_Block:	ancientgreeknumbers}', "");
Expect(1, 65936, '\p{^Is_Block:	ancientgreeknumbers}', "");
Expect(1, 65936, '\P{Is_Block:	ancientgreeknumbers}', "");
Expect(0, 65936, '\P{^Is_Block:	ancientgreeknumbers}', "");
Expect(1, 65935, '\p{Is_Block= -Ancient_Greek_NUMBERS}', "");
Expect(0, 65935, '\p{^Is_Block= -Ancient_Greek_NUMBERS}', "");
Expect(0, 65935, '\P{Is_Block= -Ancient_Greek_NUMBERS}', "");
Expect(1, 65935, '\P{^Is_Block= -Ancient_Greek_NUMBERS}', "");
Expect(0, 65936, '\p{Is_Block= -Ancient_Greek_NUMBERS}', "");
Expect(1, 65936, '\p{^Is_Block= -Ancient_Greek_NUMBERS}', "");
Expect(1, 65936, '\P{Is_Block= -Ancient_Greek_NUMBERS}', "");
Expect(0, 65936, '\P{^Is_Block= -Ancient_Greek_NUMBERS}', "");
Error('\p{Is_Blk=- Ancient_greek_numbers:=}');
Error('\P{Is_Blk=- Ancient_greek_numbers:=}');
Expect(1, 65935, '\p{Is_Blk=ancientgreeknumbers}', "");
Expect(0, 65935, '\p{^Is_Blk=ancientgreeknumbers}', "");
Expect(0, 65935, '\P{Is_Blk=ancientgreeknumbers}', "");
Expect(1, 65935, '\P{^Is_Blk=ancientgreeknumbers}', "");
Expect(0, 65936, '\p{Is_Blk=ancientgreeknumbers}', "");
Expect(1, 65936, '\p{^Is_Blk=ancientgreeknumbers}', "");
Expect(1, 65936, '\P{Is_Blk=ancientgreeknumbers}', "");
Expect(0, 65936, '\P{^Is_Blk=ancientgreeknumbers}', "");
Expect(1, 65935, '\p{Is_Blk: -Ancient_GREEK_NUMBERS}', "");
Expect(0, 65935, '\p{^Is_Blk: -Ancient_GREEK_NUMBERS}', "");
Expect(0, 65935, '\P{Is_Blk: -Ancient_GREEK_NUMBERS}', "");
Expect(1, 65935, '\P{^Is_Blk: -Ancient_GREEK_NUMBERS}', "");
Expect(0, 65936, '\p{Is_Blk: -Ancient_GREEK_NUMBERS}', "");
Expect(1, 65936, '\p{^Is_Blk: -Ancient_GREEK_NUMBERS}', "");
Expect(1, 65936, '\P{Is_Blk: -Ancient_GREEK_NUMBERS}', "");
Expect(0, 65936, '\P{^Is_Blk: -Ancient_GREEK_NUMBERS}', "");
Error('\p{Block=	NKo:=}');
Error('\P{Block=	NKo:=}');
Expect(1, 2047, '\p{Block=nko}', "");
Expect(0, 2047, '\p{^Block=nko}', "");
Expect(0, 2047, '\P{Block=nko}', "");
Expect(1, 2047, '\P{^Block=nko}', "");
Expect(0, 2048, '\p{Block=nko}', "");
Expect(1, 2048, '\p{^Block=nko}', "");
Expect(1, 2048, '\P{Block=nko}', "");
Expect(0, 2048, '\P{^Block=nko}', "");
Expect(1, 2047, '\p{Block=--NKo}', "");
Expect(0, 2047, '\p{^Block=--NKo}', "");
Expect(0, 2047, '\P{Block=--NKo}', "");
Expect(1, 2047, '\P{^Block=--NKo}', "");
Expect(0, 2048, '\p{Block=--NKo}', "");
Expect(1, 2048, '\p{^Block=--NKo}', "");
Expect(1, 2048, '\P{Block=--NKo}', "");
Expect(0, 2048, '\P{^Block=--NKo}', "");
Error('\p{Blk=-/a/nko}');
Error('\P{Blk=-/a/nko}');
Expect(1, 2047, '\p{Blk=nko}', "");
Expect(0, 2047, '\p{^Blk=nko}', "");
Expect(0, 2047, '\P{Blk=nko}', "");
Expect(1, 2047, '\P{^Blk=nko}', "");
Expect(0, 2048, '\p{Blk=nko}', "");
Expect(1, 2048, '\p{^Blk=nko}', "");
Expect(1, 2048, '\P{Blk=nko}', "");
Expect(0, 2048, '\P{^Blk=nko}', "");
Expect(1, 2047, '\p{Blk=-_NKo}', "");
Expect(0, 2047, '\p{^Blk=-_NKo}', "");
Expect(0, 2047, '\P{Blk=-_NKo}', "");
Expect(1, 2047, '\P{^Blk=-_NKo}', "");
Expect(0, 2048, '\p{Blk=-_NKo}', "");
Expect(1, 2048, '\p{^Blk=-_NKo}', "");
Expect(1, 2048, '\P{Blk=-_NKo}', "");
Expect(0, 2048, '\P{^Blk=-_NKo}', "");
Error('\p{Is_Block=	 NKO/a/}');
Error('\P{Is_Block=	 NKO/a/}');
Expect(1, 2047, '\p{Is_Block=nko}', "");
Expect(0, 2047, '\p{^Is_Block=nko}', "");
Expect(0, 2047, '\P{Is_Block=nko}', "");
Expect(1, 2047, '\P{^Is_Block=nko}', "");
Expect(0, 2048, '\p{Is_Block=nko}', "");
Expect(1, 2048, '\p{^Is_Block=nko}', "");
Expect(1, 2048, '\P{Is_Block=nko}', "");
Expect(0, 2048, '\P{^Is_Block=nko}', "");
Expect(1, 2047, '\p{Is_Block=_NKo}', "");
Expect(0, 2047, '\p{^Is_Block=_NKo}', "");
Expect(0, 2047, '\P{Is_Block=_NKo}', "");
Expect(1, 2047, '\P{^Is_Block=_NKo}', "");
Expect(0, 2048, '\p{Is_Block=_NKo}', "");
Expect(1, 2048, '\p{^Is_Block=_NKo}', "");
Expect(1, 2048, '\P{Is_Block=_NKo}', "");
Expect(0, 2048, '\P{^Is_Block=_NKo}', "");
Error('\p{Is_Blk=	 NKO/a/}');
Error('\P{Is_Blk=	 NKO/a/}');
Expect(1, 2047, '\p{Is_Blk:	nko}', "");
Expect(0, 2047, '\p{^Is_Blk:	nko}', "");
Expect(0, 2047, '\P{Is_Blk:	nko}', "");
Expect(1, 2047, '\P{^Is_Blk:	nko}', "");
Expect(0, 2048, '\p{Is_Blk:	nko}', "");
Expect(1, 2048, '\p{^Is_Blk:	nko}', "");
Expect(1, 2048, '\P{Is_Blk:	nko}', "");
Expect(0, 2048, '\P{^Is_Blk:	nko}', "");
Expect(1, 2047, '\p{Is_Blk= _NKo}', "");
Expect(0, 2047, '\p{^Is_Blk= _NKo}', "");
Expect(0, 2047, '\P{Is_Blk= _NKo}', "");
Expect(1, 2047, '\P{^Is_Blk= _NKo}', "");
Expect(0, 2048, '\p{Is_Blk= _NKo}', "");
Expect(1, 2048, '\p{^Is_Blk= _NKo}', "");
Expect(1, 2048, '\P{Is_Blk= _NKo}', "");
Expect(0, 2048, '\P{^Is_Blk= _NKo}', "");
Error('\p{Block=/a/-_General_Punctuation}');
Error('\P{Block=/a/-_General_Punctuation}');
Expect(1, 8303, '\p{Block:   generalpunctuation}', "");
Expect(0, 8303, '\p{^Block:   generalpunctuation}', "");
Expect(0, 8303, '\P{Block:   generalpunctuation}', "");
Expect(1, 8303, '\P{^Block:   generalpunctuation}', "");
Expect(0, 8304, '\p{Block:   generalpunctuation}', "");
Expect(1, 8304, '\p{^Block:   generalpunctuation}', "");
Expect(1, 8304, '\P{Block:   generalpunctuation}', "");
Expect(0, 8304, '\P{^Block:   generalpunctuation}', "");
Expect(1, 8303, '\p{Block=_	General_PUNCTUATION}', "");
Expect(0, 8303, '\p{^Block=_	General_PUNCTUATION}', "");
Expect(0, 8303, '\P{Block=_	General_PUNCTUATION}', "");
Expect(1, 8303, '\P{^Block=_	General_PUNCTUATION}', "");
Expect(0, 8304, '\p{Block=_	General_PUNCTUATION}', "");
Expect(1, 8304, '\p{^Block=_	General_PUNCTUATION}', "");
Expect(1, 8304, '\P{Block=_	General_PUNCTUATION}', "");
Expect(0, 8304, '\P{^Block=_	General_PUNCTUATION}', "");
Error('\p{Blk=:=punctuation}');
Error('\P{Blk=:=punctuation}');
Expect(1, 8303, '\p{Blk=punctuation}', "");
Expect(0, 8303, '\p{^Blk=punctuation}', "");
Expect(0, 8303, '\P{Blk=punctuation}', "");
Expect(1, 8303, '\P{^Blk=punctuation}', "");
Expect(0, 8304, '\p{Blk=punctuation}', "");
Expect(1, 8304, '\p{^Blk=punctuation}', "");
Expect(1, 8304, '\P{Blk=punctuation}', "");
Expect(0, 8304, '\P{^Blk=punctuation}', "");
Expect(1, 8303, '\p{Blk=_	PUNCTUATION}', "");
Expect(0, 8303, '\p{^Blk=_	PUNCTUATION}', "");
Expect(0, 8303, '\P{Blk=_	PUNCTUATION}', "");
Expect(1, 8303, '\P{^Blk=_	PUNCTUATION}', "");
Expect(0, 8304, '\p{Blk=_	PUNCTUATION}', "");
Expect(1, 8304, '\p{^Blk=_	PUNCTUATION}', "");
Expect(1, 8304, '\P{Blk=_	PUNCTUATION}', "");
Expect(0, 8304, '\P{^Blk=_	PUNCTUATION}', "");
Error('\p{Is_Block:   -:=GENERAL_Punctuation}');
Error('\P{Is_Block:   -:=GENERAL_Punctuation}');
Expect(1, 8303, '\p{Is_Block=generalpunctuation}', "");
Expect(0, 8303, '\p{^Is_Block=generalpunctuation}', "");
Expect(0, 8303, '\P{Is_Block=generalpunctuation}', "");
Expect(1, 8303, '\P{^Is_Block=generalpunctuation}', "");
Expect(0, 8304, '\p{Is_Block=generalpunctuation}', "");
Expect(1, 8304, '\p{^Is_Block=generalpunctuation}', "");
Expect(1, 8304, '\P{Is_Block=generalpunctuation}', "");
Expect(0, 8304, '\P{^Is_Block=generalpunctuation}', "");
Expect(1, 8303, '\p{Is_Block=- General_Punctuation}', "");
Expect(0, 8303, '\p{^Is_Block=- General_Punctuation}', "");
Expect(0, 8303, '\P{Is_Block=- General_Punctuation}', "");
Expect(1, 8303, '\P{^Is_Block=- General_Punctuation}', "");
Expect(0, 8304, '\p{Is_Block=- General_Punctuation}', "");
Expect(1, 8304, '\p{^Is_Block=- General_Punctuation}', "");
Expect(1, 8304, '\P{Is_Block=- General_Punctuation}', "");
Expect(0, 8304, '\P{^Is_Block=- General_Punctuation}', "");
Error('\p{Is_Blk=/a/ -PUNCTUATION}');
Error('\P{Is_Blk=/a/ -PUNCTUATION}');
Expect(1, 8303, '\p{Is_Blk=punctuation}', "");
Expect(0, 8303, '\p{^Is_Blk=punctuation}', "");
Expect(0, 8303, '\P{Is_Blk=punctuation}', "");
Expect(1, 8303, '\P{^Is_Blk=punctuation}', "");
Expect(0, 8304, '\p{Is_Blk=punctuation}', "");
Expect(1, 8304, '\p{^Is_Blk=punctuation}', "");
Expect(1, 8304, '\P{Is_Blk=punctuation}', "");
Expect(0, 8304, '\P{^Is_Blk=punctuation}', "");
Expect(1, 8303, '\p{Is_Blk=	_PUNCTUATION}', "");
Expect(0, 8303, '\p{^Is_Blk=	_PUNCTUATION}', "");
Expect(0, 8303, '\P{Is_Blk=	_PUNCTUATION}', "");
Expect(1, 8303, '\P{^Is_Blk=	_PUNCTUATION}', "");
Expect(0, 8304, '\p{Is_Blk=	_PUNCTUATION}', "");
Expect(1, 8304, '\p{^Is_Blk=	_PUNCTUATION}', "");
Expect(1, 8304, '\P{Is_Blk=	_PUNCTUATION}', "");
Expect(0, 8304, '\P{^Is_Blk=	_PUNCTUATION}', "");
Error('\p{Block=:=	-BAMUM_SUPPLEMENT}');
Error('\P{Block=:=	-BAMUM_SUPPLEMENT}');
Expect(1, 92735, '\p{Block=bamumsupplement}', "");
Expect(0, 92735, '\p{^Block=bamumsupplement}', "");
Expect(0, 92735, '\P{Block=bamumsupplement}', "");
Expect(1, 92735, '\P{^Block=bamumsupplement}', "");
Expect(0, 92736, '\p{Block=bamumsupplement}', "");
Expect(1, 92736, '\p{^Block=bamumsupplement}', "");
Expect(1, 92736, '\P{Block=bamumsupplement}', "");
Expect(0, 92736, '\P{^Block=bamumsupplement}', "");
Expect(1, 92735, '\p{Block:    _Bamum_Supplement}', "");
Expect(0, 92735, '\p{^Block:    _Bamum_Supplement}', "");
Expect(0, 92735, '\P{Block:    _Bamum_Supplement}', "");
Expect(1, 92735, '\P{^Block:    _Bamum_Supplement}', "");
Expect(0, 92736, '\p{Block:    _Bamum_Supplement}', "");
Expect(1, 92736, '\p{^Block:    _Bamum_Supplement}', "");
Expect(1, 92736, '\P{Block:    _Bamum_Supplement}', "");
Expect(0, 92736, '\P{^Block:    _Bamum_Supplement}', "");
Error('\p{Blk=:=__BAMUM_Sup}');
Error('\P{Blk=:=__BAMUM_Sup}');
Expect(1, 92735, '\p{Blk=bamumsup}', "");
Expect(0, 92735, '\p{^Blk=bamumsup}', "");
Expect(0, 92735, '\P{Blk=bamumsup}', "");
Expect(1, 92735, '\P{^Blk=bamumsup}', "");
Expect(0, 92736, '\p{Blk=bamumsup}', "");
Expect(1, 92736, '\p{^Blk=bamumsup}', "");
Expect(1, 92736, '\P{Blk=bamumsup}', "");
Expect(0, 92736, '\P{^Blk=bamumsup}', "");
Expect(1, 92735, '\p{Blk=-Bamum_SUP}', "");
Expect(0, 92735, '\p{^Blk=-Bamum_SUP}', "");
Expect(0, 92735, '\P{Blk=-Bamum_SUP}', "");
Expect(1, 92735, '\P{^Blk=-Bamum_SUP}', "");
Expect(0, 92736, '\p{Blk=-Bamum_SUP}', "");
Expect(1, 92736, '\p{^Blk=-Bamum_SUP}', "");
Expect(1, 92736, '\P{Blk=-Bamum_SUP}', "");
Expect(0, 92736, '\P{^Blk=-Bamum_SUP}', "");
Error('\p{Is_Block=		Bamum_supplement:=}');
Error('\P{Is_Block=		Bamum_supplement:=}');
Expect(1, 92735, '\p{Is_Block=bamumsupplement}', "");
Expect(0, 92735, '\p{^Is_Block=bamumsupplement}', "");
Expect(0, 92735, '\P{Is_Block=bamumsupplement}', "");
Expect(1, 92735, '\P{^Is_Block=bamumsupplement}', "");
Expect(0, 92736, '\p{Is_Block=bamumsupplement}', "");
Expect(1, 92736, '\p{^Is_Block=bamumsupplement}', "");
Expect(1, 92736, '\P{Is_Block=bamumsupplement}', "");
Expect(0, 92736, '\P{^Is_Block=bamumsupplement}', "");
Expect(1, 92735, '\p{Is_Block= -Bamum_Supplement}', "");
Expect(0, 92735, '\p{^Is_Block= -Bamum_Supplement}', "");
Expect(0, 92735, '\P{Is_Block= -Bamum_Supplement}', "");
Expect(1, 92735, '\P{^Is_Block= -Bamum_Supplement}', "");
Expect(0, 92736, '\p{Is_Block= -Bamum_Supplement}', "");
Expect(1, 92736, '\p{^Is_Block= -Bamum_Supplement}', "");
Expect(1, 92736, '\P{Is_Block= -Bamum_Supplement}', "");
Expect(0, 92736, '\P{^Is_Block= -Bamum_Supplement}', "");
Error('\p{Is_Blk:   -_Bamum_Sup:=}');
Error('\P{Is_Blk:   -_Bamum_Sup:=}');
Expect(1, 92735, '\p{Is_Blk=bamumsup}', "");
Expect(0, 92735, '\p{^Is_Blk=bamumsup}', "");
Expect(0, 92735, '\P{Is_Blk=bamumsup}', "");
Expect(1, 92735, '\P{^Is_Blk=bamumsup}', "");
Expect(0, 92736, '\p{Is_Blk=bamumsup}', "");
Expect(1, 92736, '\p{^Is_Blk=bamumsup}', "");
Expect(1, 92736, '\P{Is_Blk=bamumsup}', "");
Expect(0, 92736, '\P{^Is_Blk=bamumsup}', "");
Expect(1, 92735, '\p{Is_Blk=_Bamum_Sup}', "");
Expect(0, 92735, '\p{^Is_Blk=_Bamum_Sup}', "");
Expect(0, 92735, '\P{Is_Blk=_Bamum_Sup}', "");
Expect(1, 92735, '\P{^Is_Blk=_Bamum_Sup}', "");
Expect(0, 92736, '\p{Is_Blk=_Bamum_Sup}', "");
Expect(1, 92736, '\p{^Is_Blk=_Bamum_Sup}', "");
Expect(1, 92736, '\P{Is_Blk=_Bamum_Sup}', "");
Expect(0, 92736, '\P{^Is_Blk=_Bamum_Sup}', "");
Error('\p{Block=/a/  MUSICAL_SYMBOLS}');
Error('\P{Block=/a/  MUSICAL_SYMBOLS}');
Expect(1, 119295, '\p{Block:	musicalsymbols}', "");
Expect(0, 119295, '\p{^Block:	musicalsymbols}', "");
Expect(0, 119295, '\P{Block:	musicalsymbols}', "");
Expect(1, 119295, '\P{^Block:	musicalsymbols}', "");
Expect(0, 119296, '\p{Block:	musicalsymbols}', "");
Expect(1, 119296, '\p{^Block:	musicalsymbols}', "");
Expect(1, 119296, '\P{Block:	musicalsymbols}', "");
Expect(0, 119296, '\P{^Block:	musicalsymbols}', "");
Expect(1, 119295, '\p{Block=	-MUSICAL_Symbols}', "");
Expect(0, 119295, '\p{^Block=	-MUSICAL_Symbols}', "");
Expect(0, 119295, '\P{Block=	-MUSICAL_Symbols}', "");
Expect(1, 119295, '\P{^Block=	-MUSICAL_Symbols}', "");
Expect(0, 119296, '\p{Block=	-MUSICAL_Symbols}', "");
Expect(1, 119296, '\p{^Block=	-MUSICAL_Symbols}', "");
Expect(1, 119296, '\P{Block=	-MUSICAL_Symbols}', "");
Expect(0, 119296, '\P{^Block=	-MUSICAL_Symbols}', "");
Error('\p{Blk=-:=Music}');
Error('\P{Blk=-:=Music}');
Expect(1, 119295, '\p{Blk=music}', "");
Expect(0, 119295, '\p{^Blk=music}', "");
Expect(0, 119295, '\P{Blk=music}', "");
Expect(1, 119295, '\P{^Blk=music}', "");
Expect(0, 119296, '\p{Blk=music}', "");
Expect(1, 119296, '\p{^Blk=music}', "");
Expect(1, 119296, '\P{Blk=music}', "");
Expect(0, 119296, '\P{^Blk=music}', "");
Expect(1, 119295, '\p{Blk: _ Music}', "");
Expect(0, 119295, '\p{^Blk: _ Music}', "");
Expect(0, 119295, '\P{Blk: _ Music}', "");
Expect(1, 119295, '\P{^Blk: _ Music}', "");
Expect(0, 119296, '\p{Blk: _ Music}', "");
Expect(1, 119296, '\p{^Blk: _ Music}', "");
Expect(1, 119296, '\P{Blk: _ Music}', "");
Expect(0, 119296, '\P{^Blk: _ Music}', "");
Error('\p{Is_Block=:=__Musical_Symbols}');
Error('\P{Is_Block=:=__Musical_Symbols}');
Expect(1, 119295, '\p{Is_Block=musicalsymbols}', "");
Expect(0, 119295, '\p{^Is_Block=musicalsymbols}', "");
Expect(0, 119295, '\P{Is_Block=musicalsymbols}', "");
Expect(1, 119295, '\P{^Is_Block=musicalsymbols}', "");
Expect(0, 119296, '\p{Is_Block=musicalsymbols}', "");
Expect(1, 119296, '\p{^Is_Block=musicalsymbols}', "");
Expect(1, 119296, '\P{Is_Block=musicalsymbols}', "");
Expect(0, 119296, '\P{^Is_Block=musicalsymbols}', "");
Expect(1, 119295, '\p{Is_Block= -Musical_symbols}', "");
Expect(0, 119295, '\p{^Is_Block= -Musical_symbols}', "");
Expect(0, 119295, '\P{Is_Block= -Musical_symbols}', "");
Expect(1, 119295, '\P{^Is_Block= -Musical_symbols}', "");
Expect(0, 119296, '\p{Is_Block= -Musical_symbols}', "");
Expect(1, 119296, '\p{^Is_Block= -Musical_symbols}', "");
Expect(1, 119296, '\P{Is_Block= -Musical_symbols}', "");
Expect(0, 119296, '\P{^Is_Block= -Musical_symbols}', "");
Error('\p{Is_Blk=_ Music:=}');
Error('\P{Is_Blk=_ Music:=}');
Expect(1, 119295, '\p{Is_Blk=music}', "");
Expect(0, 119295, '\p{^Is_Blk=music}', "");
Expect(0, 119295, '\P{Is_Blk=music}', "");
Expect(1, 119295, '\P{^Is_Blk=music}', "");
Expect(0, 119296, '\p{Is_Blk=music}', "");
Expect(1, 119296, '\p{^Is_Blk=music}', "");
Expect(1, 119296, '\P{Is_Blk=music}', "");
Expect(0, 119296, '\P{^Is_Blk=music}', "");
Expect(1, 119295, '\p{Is_Blk: _music}', "");
Expect(0, 119295, '\p{^Is_Blk: _music}', "");
Expect(0, 119295, '\P{Is_Blk: _music}', "");
Expect(1, 119295, '\P{^Is_Blk: _music}', "");
Expect(0, 119296, '\p{Is_Blk: _music}', "");
Expect(1, 119296, '\p{^Is_Blk: _music}', "");
Expect(1, 119296, '\P{Is_Blk: _music}', "");
Expect(0, 119296, '\P{^Is_Blk: _music}', "");
Error('\p{Block=_ kangxi_RADICALS:=}');
Error('\P{Block=_ kangxi_RADICALS:=}');
Expect(1, 12255, '\p{Block=kangxiradicals}', "");
Expect(0, 12255, '\p{^Block=kangxiradicals}', "");
Expect(0, 12255, '\P{Block=kangxiradicals}', "");
Expect(1, 12255, '\P{^Block=kangxiradicals}', "");
Expect(0, 12256, '\p{Block=kangxiradicals}', "");
Expect(1, 12256, '\p{^Block=kangxiradicals}', "");
Expect(1, 12256, '\P{Block=kangxiradicals}', "");
Expect(0, 12256, '\P{^Block=kangxiradicals}', "");
Expect(1, 12255, '\p{Block=	Kangxi_Radicals}', "");
Expect(0, 12255, '\p{^Block=	Kangxi_Radicals}', "");
Expect(0, 12255, '\P{Block=	Kangxi_Radicals}', "");
Expect(1, 12255, '\P{^Block=	Kangxi_Radicals}', "");
Expect(0, 12256, '\p{Block=	Kangxi_Radicals}', "");
Expect(1, 12256, '\p{^Block=	Kangxi_Radicals}', "");
Expect(1, 12256, '\P{Block=	Kangxi_Radicals}', "");
Expect(0, 12256, '\P{^Block=	Kangxi_Radicals}', "");
Error('\p{Blk=:=kangxi}');
Error('\P{Blk=:=kangxi}');
Expect(1, 12255, '\p{Blk=kangxi}', "");
Expect(0, 12255, '\p{^Blk=kangxi}', "");
Expect(0, 12255, '\P{Blk=kangxi}', "");
Expect(1, 12255, '\P{^Blk=kangxi}', "");
Expect(0, 12256, '\p{Blk=kangxi}', "");
Expect(1, 12256, '\p{^Blk=kangxi}', "");
Expect(1, 12256, '\P{Blk=kangxi}', "");
Expect(0, 12256, '\P{^Blk=kangxi}', "");
Expect(1, 12255, '\p{Blk=-_kangxi}', "");
Expect(0, 12255, '\p{^Blk=-_kangxi}', "");
Expect(0, 12255, '\P{Blk=-_kangxi}', "");
Expect(1, 12255, '\P{^Blk=-_kangxi}', "");
Expect(0, 12256, '\p{Blk=-_kangxi}', "");
Expect(1, 12256, '\p{^Blk=-_kangxi}', "");
Expect(1, 12256, '\P{Blk=-_kangxi}', "");
Expect(0, 12256, '\P{^Blk=-_kangxi}', "");
Error('\p{Is_Block=-kangxi_Radicals:=}');
Error('\P{Is_Block=-kangxi_Radicals:=}');
Expect(1, 12255, '\p{Is_Block=kangxiradicals}', "");
Expect(0, 12255, '\p{^Is_Block=kangxiradicals}', "");
Expect(0, 12255, '\P{Is_Block=kangxiradicals}', "");
Expect(1, 12255, '\P{^Is_Block=kangxiradicals}', "");
Expect(0, 12256, '\p{Is_Block=kangxiradicals}', "");
Expect(1, 12256, '\p{^Is_Block=kangxiradicals}', "");
Expect(1, 12256, '\P{Is_Block=kangxiradicals}', "");
Expect(0, 12256, '\P{^Is_Block=kangxiradicals}', "");
Expect(1, 12255, '\p{Is_Block=- KANGXI_Radicals}', "");
Expect(0, 12255, '\p{^Is_Block=- KANGXI_Radicals}', "");
Expect(0, 12255, '\P{Is_Block=- KANGXI_Radicals}', "");
Expect(1, 12255, '\P{^Is_Block=- KANGXI_Radicals}', "");
Expect(0, 12256, '\p{Is_Block=- KANGXI_Radicals}', "");
Expect(1, 12256, '\p{^Is_Block=- KANGXI_Radicals}', "");
Expect(1, 12256, '\P{Is_Block=- KANGXI_Radicals}', "");
Expect(0, 12256, '\P{^Is_Block=- KANGXI_Radicals}', "");
Error('\p{Is_Blk=/a/-Kangxi}');
Error('\P{Is_Blk=/a/-Kangxi}');
Expect(1, 12255, '\p{Is_Blk:   kangxi}', "");
Expect(0, 12255, '\p{^Is_Blk:   kangxi}', "");
Expect(0, 12255, '\P{Is_Blk:   kangxi}', "");
Expect(1, 12255, '\P{^Is_Blk:   kangxi}', "");
Expect(0, 12256, '\p{Is_Blk:   kangxi}', "");
Expect(1, 12256, '\p{^Is_Blk:   kangxi}', "");
Expect(1, 12256, '\P{Is_Blk:   kangxi}', "");
Expect(0, 12256, '\P{^Is_Blk:   kangxi}', "");
Expect(1, 12255, '\p{Is_Blk= kangxi}', "");
Expect(0, 12255, '\p{^Is_Blk= kangxi}', "");
Expect(0, 12255, '\P{Is_Blk= kangxi}', "");
Expect(1, 12255, '\P{^Is_Blk= kangxi}', "");
Expect(0, 12256, '\p{Is_Blk= kangxi}', "");
Expect(1, 12256, '\p{^Is_Blk= kangxi}', "");
Expect(1, 12256, '\P{Is_Blk= kangxi}', "");
Expect(0, 12256, '\P{^Is_Blk= kangxi}', "");
Error('\p{Block=:=_lycian}');
Error('\P{Block=:=_lycian}');
Expect(1, 66207, '\p{Block=lycian}', "");
Expect(0, 66207, '\p{^Block=lycian}', "");
Expect(0, 66207, '\P{Block=lycian}', "");
Expect(1, 66207, '\P{^Block=lycian}', "");
Expect(0, 66208, '\p{Block=lycian}', "");
Expect(1, 66208, '\p{^Block=lycian}', "");
Expect(1, 66208, '\P{Block=lycian}', "");
Expect(0, 66208, '\P{^Block=lycian}', "");
Expect(1, 66207, '\p{Block: _ LYCIAN}', "");
Expect(0, 66207, '\p{^Block: _ LYCIAN}', "");
Expect(0, 66207, '\P{Block: _ LYCIAN}', "");
Expect(1, 66207, '\P{^Block: _ LYCIAN}', "");
Expect(0, 66208, '\p{Block: _ LYCIAN}', "");
Expect(1, 66208, '\p{^Block: _ LYCIAN}', "");
Expect(1, 66208, '\P{Block: _ LYCIAN}', "");
Expect(0, 66208, '\P{^Block: _ LYCIAN}', "");
Error('\p{Blk=- lycian:=}');
Error('\P{Blk=- lycian:=}');
Expect(1, 66207, '\p{Blk=lycian}', "");
Expect(0, 66207, '\p{^Blk=lycian}', "");
Expect(0, 66207, '\P{Blk=lycian}', "");
Expect(1, 66207, '\P{^Blk=lycian}', "");
Expect(0, 66208, '\p{Blk=lycian}', "");
Expect(1, 66208, '\p{^Blk=lycian}', "");
Expect(1, 66208, '\P{Blk=lycian}', "");
Expect(0, 66208, '\P{^Blk=lycian}', "");
Expect(1, 66207, '\p{Blk=Lycian}', "");
Expect(0, 66207, '\p{^Blk=Lycian}', "");
Expect(0, 66207, '\P{Blk=Lycian}', "");
Expect(1, 66207, '\P{^Blk=Lycian}', "");
Expect(0, 66208, '\p{Blk=Lycian}', "");
Expect(1, 66208, '\p{^Blk=Lycian}', "");
Expect(1, 66208, '\P{Blk=Lycian}', "");
Expect(0, 66208, '\P{^Blk=Lycian}', "");
Error('\p{Is_Block=/a/	-Lycian}');
Error('\P{Is_Block=/a/	-Lycian}');
Expect(1, 66207, '\p{Is_Block=lycian}', "");
Expect(0, 66207, '\p{^Is_Block=lycian}', "");
Expect(0, 66207, '\P{Is_Block=lycian}', "");
Expect(1, 66207, '\P{^Is_Block=lycian}', "");
Expect(0, 66208, '\p{Is_Block=lycian}', "");
Expect(1, 66208, '\p{^Is_Block=lycian}', "");
Expect(1, 66208, '\P{Is_Block=lycian}', "");
Expect(0, 66208, '\P{^Is_Block=lycian}', "");
Expect(1, 66207, '\p{Is_Block=	Lycian}', "");
Expect(0, 66207, '\p{^Is_Block=	Lycian}', "");
Expect(0, 66207, '\P{Is_Block=	Lycian}', "");
Expect(1, 66207, '\P{^Is_Block=	Lycian}', "");
Expect(0, 66208, '\p{Is_Block=	Lycian}', "");
Expect(1, 66208, '\p{^Is_Block=	Lycian}', "");
Expect(1, 66208, '\P{Is_Block=	Lycian}', "");
Expect(0, 66208, '\P{^Is_Block=	Lycian}', "");
Error('\p{Is_Blk=	_Lycian/a/}');
Error('\P{Is_Blk=	_Lycian/a/}');
Expect(1, 66207, '\p{Is_Blk:lycian}', "");
Expect(0, 66207, '\p{^Is_Blk:lycian}', "");
Expect(0, 66207, '\P{Is_Blk:lycian}', "");
Expect(1, 66207, '\P{^Is_Blk:lycian}', "");
Expect(0, 66208, '\p{Is_Blk:lycian}', "");
Expect(1, 66208, '\p{^Is_Blk:lycian}', "");
Expect(1, 66208, '\P{Is_Blk:lycian}', "");
Expect(0, 66208, '\P{^Is_Blk:lycian}', "");
Expect(1, 66207, '\p{Is_Blk=	 lycian}', "");
Expect(0, 66207, '\p{^Is_Blk=	 lycian}', "");
Expect(0, 66207, '\P{Is_Blk=	 lycian}', "");
Expect(1, 66207, '\P{^Is_Blk=	 lycian}', "");
Expect(0, 66208, '\p{Is_Blk=	 lycian}', "");
Expect(1, 66208, '\p{^Is_Blk=	 lycian}', "");
Expect(1, 66208, '\P{Is_Blk=	 lycian}', "");
Expect(0, 66208, '\P{^Is_Blk=	 lycian}', "");
Error('\p{Block=/a/-	tai_XUAN_Jing_SYMBOLS}');
Error('\P{Block=/a/-	tai_XUAN_Jing_SYMBOLS}');
Expect(1, 119647, '\p{Block=taixuanjingsymbols}', "");
Expect(0, 119647, '\p{^Block=taixuanjingsymbols}', "");
Expect(0, 119647, '\P{Block=taixuanjingsymbols}', "");
Expect(1, 119647, '\P{^Block=taixuanjingsymbols}', "");
Expect(0, 119648, '\p{Block=taixuanjingsymbols}', "");
Expect(1, 119648, '\p{^Block=taixuanjingsymbols}', "");
Expect(1, 119648, '\P{Block=taixuanjingsymbols}', "");
Expect(0, 119648, '\P{^Block=taixuanjingsymbols}', "");
Expect(1, 119647, '\p{Block= 	Tai_xuan_jing_symbols}', "");
Expect(0, 119647, '\p{^Block= 	Tai_xuan_jing_symbols}', "");
Expect(0, 119647, '\P{Block= 	Tai_xuan_jing_symbols}', "");
Expect(1, 119647, '\P{^Block= 	Tai_xuan_jing_symbols}', "");
Expect(0, 119648, '\p{Block= 	Tai_xuan_jing_symbols}', "");
Expect(1, 119648, '\p{^Block= 	Tai_xuan_jing_symbols}', "");
Expect(1, 119648, '\P{Block= 	Tai_xuan_jing_symbols}', "");
Expect(0, 119648, '\P{^Block= 	Tai_xuan_jing_symbols}', "");
Error('\p{Blk:-/a/tai_xuan_Jing}');
Error('\P{Blk:-/a/tai_xuan_Jing}');
Expect(1, 119647, '\p{Blk=taixuanjing}', "");
Expect(0, 119647, '\p{^Blk=taixuanjing}', "");
Expect(0, 119647, '\P{Blk=taixuanjing}', "");
Expect(1, 119647, '\P{^Blk=taixuanjing}', "");
Expect(0, 119648, '\p{Blk=taixuanjing}', "");
Expect(1, 119648, '\p{^Blk=taixuanjing}', "");
Expect(1, 119648, '\P{Blk=taixuanjing}', "");
Expect(0, 119648, '\P{^Blk=taixuanjing}', "");
Expect(1, 119647, '\p{Blk=	-tai_xuan_Jing}', "");
Expect(0, 119647, '\p{^Blk=	-tai_xuan_Jing}', "");
Expect(0, 119647, '\P{Blk=	-tai_xuan_Jing}', "");
Expect(1, 119647, '\P{^Blk=	-tai_xuan_Jing}', "");
Expect(0, 119648, '\p{Blk=	-tai_xuan_Jing}', "");
Expect(1, 119648, '\p{^Blk=	-tai_xuan_Jing}', "");
Expect(1, 119648, '\P{Blk=	-tai_xuan_Jing}', "");
Expect(0, 119648, '\P{^Blk=	-tai_xuan_Jing}', "");
Error('\p{Is_Block=:=_TAI_xuan_Jing_symbols}');
Error('\P{Is_Block=:=_TAI_xuan_Jing_symbols}');
Expect(1, 119647, '\p{Is_Block:   taixuanjingsymbols}', "");
Expect(0, 119647, '\p{^Is_Block:   taixuanjingsymbols}', "");
Expect(0, 119647, '\P{Is_Block:   taixuanjingsymbols}', "");
Expect(1, 119647, '\P{^Is_Block:   taixuanjingsymbols}', "");
Expect(0, 119648, '\p{Is_Block:   taixuanjingsymbols}', "");
Expect(1, 119648, '\p{^Is_Block:   taixuanjingsymbols}', "");
Expect(1, 119648, '\P{Is_Block:   taixuanjingsymbols}', "");
Expect(0, 119648, '\P{^Is_Block:   taixuanjingsymbols}', "");
Expect(1, 119647, '\p{Is_Block:   -_tai_xuan_Jing_Symbols}', "");
Expect(0, 119647, '\p{^Is_Block:   -_tai_xuan_Jing_Symbols}', "");
Expect(0, 119647, '\P{Is_Block:   -_tai_xuan_Jing_Symbols}', "");
Expect(1, 119647, '\P{^Is_Block:   -_tai_xuan_Jing_Symbols}', "");
Expect(0, 119648, '\p{Is_Block:   -_tai_xuan_Jing_Symbols}', "");
Expect(1, 119648, '\p{^Is_Block:   -_tai_xuan_Jing_Symbols}', "");
Expect(1, 119648, '\P{Is_Block:   -_tai_xuan_Jing_Symbols}', "");
Expect(0, 119648, '\P{^Is_Block:   -_tai_xuan_Jing_Symbols}', "");
Error('\p{Is_Blk=-:=TAI_XUAN_jing}');
Error('\P{Is_Blk=-:=TAI_XUAN_jing}');
Expect(1, 119647, '\p{Is_Blk=taixuanjing}', "");
Expect(0, 119647, '\p{^Is_Blk=taixuanjing}', "");
Expect(0, 119647, '\P{Is_Blk=taixuanjing}', "");
Expect(1, 119647, '\P{^Is_Blk=taixuanjing}', "");
Expect(0, 119648, '\p{Is_Blk=taixuanjing}', "");
Expect(1, 119648, '\p{^Is_Blk=taixuanjing}', "");
Expect(1, 119648, '\P{Is_Blk=taixuanjing}', "");
Expect(0, 119648, '\P{^Is_Blk=taixuanjing}', "");
Expect(1, 119647, '\p{Is_Blk=_	Tai_xuan_jing}', "");
Expect(0, 119647, '\p{^Is_Blk=_	Tai_xuan_jing}', "");
Expect(0, 119647, '\P{Is_Blk=_	Tai_xuan_jing}', "");
Expect(1, 119647, '\P{^Is_Blk=_	Tai_xuan_jing}', "");
Expect(0, 119648, '\p{Is_Blk=_	Tai_xuan_jing}', "");
Expect(1, 119648, '\p{^Is_Blk=_	Tai_xuan_jing}', "");
Expect(1, 119648, '\P{Is_Blk=_	Tai_xuan_jing}', "");
Expect(0, 119648, '\P{^Is_Blk=_	Tai_xuan_jing}', "");
Error('\p{Block=-:=Geometric_Shapes}');
Error('\P{Block=-:=Geometric_Shapes}');
Expect(1, 9727, '\p{Block=geometricshapes}', "");
Expect(0, 9727, '\p{^Block=geometricshapes}', "");
Expect(0, 9727, '\P{Block=geometricshapes}', "");
Expect(1, 9727, '\P{^Block=geometricshapes}', "");
Expect(0, 9728, '\p{Block=geometricshapes}', "");
Expect(1, 9728, '\p{^Block=geometricshapes}', "");
Expect(1, 9728, '\P{Block=geometricshapes}', "");
Expect(0, 9728, '\P{^Block=geometricshapes}', "");
Expect(1, 9727, '\p{Block=-_GEOMETRIC_shapes}', "");
Expect(0, 9727, '\p{^Block=-_GEOMETRIC_shapes}', "");
Expect(0, 9727, '\P{Block=-_GEOMETRIC_shapes}', "");
Expect(1, 9727, '\P{^Block=-_GEOMETRIC_shapes}', "");
Expect(0, 9728, '\p{Block=-_GEOMETRIC_shapes}', "");
Expect(1, 9728, '\p{^Block=-_GEOMETRIC_shapes}', "");
Expect(1, 9728, '\P{Block=-_GEOMETRIC_shapes}', "");
Expect(0, 9728, '\P{^Block=-_GEOMETRIC_shapes}', "");
Error('\p{Blk=-_Geometric_Shapes:=}');
Error('\P{Blk=-_Geometric_Shapes:=}');
Expect(1, 9727, '\p{Blk:   geometricshapes}', "");
Expect(0, 9727, '\p{^Blk:   geometricshapes}', "");
Expect(0, 9727, '\P{Blk:   geometricshapes}', "");
Expect(1, 9727, '\P{^Blk:   geometricshapes}', "");
Expect(0, 9728, '\p{Blk:   geometricshapes}', "");
Expect(1, 9728, '\p{^Blk:   geometricshapes}', "");
Expect(1, 9728, '\P{Blk:   geometricshapes}', "");
Expect(0, 9728, '\P{^Blk:   geometricshapes}', "");
Expect(1, 9727, '\p{Blk=_	Geometric_SHAPES}', "");
Expect(0, 9727, '\p{^Blk=_	Geometric_SHAPES}', "");
Expect(0, 9727, '\P{Blk=_	Geometric_SHAPES}', "");
Expect(1, 9727, '\P{^Blk=_	Geometric_SHAPES}', "");
Expect(0, 9728, '\p{Blk=_	Geometric_SHAPES}', "");
Expect(1, 9728, '\p{^Blk=_	Geometric_SHAPES}', "");
Expect(1, 9728, '\P{Blk=_	Geometric_SHAPES}', "");
Expect(0, 9728, '\P{^Blk=_	Geometric_SHAPES}', "");
Error('\p{Is_Block:	-geometric_SHAPES:=}');
Error('\P{Is_Block:	-geometric_SHAPES:=}');
Expect(1, 9727, '\p{Is_Block=geometricshapes}', "");
Expect(0, 9727, '\p{^Is_Block=geometricshapes}', "");
Expect(0, 9727, '\P{Is_Block=geometricshapes}', "");
Expect(1, 9727, '\P{^Is_Block=geometricshapes}', "");
Expect(0, 9728, '\p{Is_Block=geometricshapes}', "");
Expect(1, 9728, '\p{^Is_Block=geometricshapes}', "");
Expect(1, 9728, '\P{Is_Block=geometricshapes}', "");
Expect(0, 9728, '\P{^Is_Block=geometricshapes}', "");
Expect(1, 9727, '\p{Is_Block=_	GEOMETRIC_Shapes}', "");
Expect(0, 9727, '\p{^Is_Block=_	GEOMETRIC_Shapes}', "");
Expect(0, 9727, '\P{Is_Block=_	GEOMETRIC_Shapes}', "");
Expect(1, 9727, '\P{^Is_Block=_	GEOMETRIC_Shapes}', "");
Expect(0, 9728, '\p{Is_Block=_	GEOMETRIC_Shapes}', "");
Expect(1, 9728, '\p{^Is_Block=_	GEOMETRIC_Shapes}', "");
Expect(1, 9728, '\P{Is_Block=_	GEOMETRIC_Shapes}', "");
Expect(0, 9728, '\P{^Is_Block=_	GEOMETRIC_Shapes}', "");
Error('\p{Is_Blk=	:=Geometric_SHAPES}');
Error('\P{Is_Blk=	:=Geometric_SHAPES}');
Expect(1, 9727, '\p{Is_Blk=geometricshapes}', "");
Expect(0, 9727, '\p{^Is_Blk=geometricshapes}', "");
Expect(0, 9727, '\P{Is_Blk=geometricshapes}', "");
Expect(1, 9727, '\P{^Is_Blk=geometricshapes}', "");
Expect(0, 9728, '\p{Is_Blk=geometricshapes}', "");
Expect(1, 9728, '\p{^Is_Blk=geometricshapes}', "");
Expect(1, 9728, '\P{Is_Blk=geometricshapes}', "");
Expect(0, 9728, '\P{^Is_Blk=geometricshapes}', "");
Expect(1, 9727, '\p{Is_Blk=_Geometric_shapes}', "");
Expect(0, 9727, '\p{^Is_Blk=_Geometric_shapes}', "");
Expect(0, 9727, '\P{Is_Blk=_Geometric_shapes}', "");
Expect(1, 9727, '\P{^Is_Blk=_Geometric_shapes}', "");
Expect(0, 9728, '\p{Is_Blk=_Geometric_shapes}', "");
Expect(1, 9728, '\p{^Is_Blk=_Geometric_shapes}', "");
Expect(1, 9728, '\P{Is_Blk=_Geometric_shapes}', "");
Expect(0, 9728, '\P{^Is_Blk=_Geometric_shapes}', "");
Error('\p{Block=_:=Supplementary_PRIVATE_Use_area_a}');
Error('\P{Block=_:=Supplementary_PRIVATE_Use_area_a}');
Expect(1, 983040, '\p{Block=supplementaryprivateuseareaa}', "");
Expect(0, 983040, '\p{^Block=supplementaryprivateuseareaa}', "");
Expect(0, 983040, '\P{Block=supplementaryprivateuseareaa}', "");
Expect(1, 983040, '\P{^Block=supplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\p{Block=supplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\p{^Block=supplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\P{Block=supplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\P{^Block=supplementaryprivateuseareaa}', "");
Expect(1, 983040, '\p{Block=-SUPPLEMENTARY_Private_Use_AREA_A}', "");
Expect(0, 983040, '\p{^Block=-SUPPLEMENTARY_Private_Use_AREA_A}', "");
Expect(0, 983040, '\P{Block=-SUPPLEMENTARY_Private_Use_AREA_A}', "");
Expect(1, 983040, '\P{^Block=-SUPPLEMENTARY_Private_Use_AREA_A}', "");
Expect(0, 1048576, '\p{Block=-SUPPLEMENTARY_Private_Use_AREA_A}', "");
Expect(1, 1048576, '\p{^Block=-SUPPLEMENTARY_Private_Use_AREA_A}', "");
Expect(1, 1048576, '\P{Block=-SUPPLEMENTARY_Private_Use_AREA_A}', "");
Expect(0, 1048576, '\P{^Block=-SUPPLEMENTARY_Private_Use_AREA_A}', "");
Error('\p{Blk=	 Sup_pua_a/a/}');
Error('\P{Blk=	 Sup_pua_a/a/}');
Expect(1, 983040, '\p{Blk=suppuaa}', "");
Expect(0, 983040, '\p{^Blk=suppuaa}', "");
Expect(0, 983040, '\P{Blk=suppuaa}', "");
Expect(1, 983040, '\P{^Blk=suppuaa}', "");
Expect(0, 1048576, '\p{Blk=suppuaa}', "");
Expect(1, 1048576, '\p{^Blk=suppuaa}', "");
Expect(1, 1048576, '\P{Blk=suppuaa}', "");
Expect(0, 1048576, '\P{^Blk=suppuaa}', "");
Expect(1, 983040, '\p{Blk:   	Sup_PUA_A}', "");
Expect(0, 983040, '\p{^Blk:   	Sup_PUA_A}', "");
Expect(0, 983040, '\P{Blk:   	Sup_PUA_A}', "");
Expect(1, 983040, '\P{^Blk:   	Sup_PUA_A}', "");
Expect(0, 1048576, '\p{Blk:   	Sup_PUA_A}', "");
Expect(1, 1048576, '\p{^Blk:   	Sup_PUA_A}', "");
Expect(1, 1048576, '\P{Blk:   	Sup_PUA_A}', "");
Expect(0, 1048576, '\P{^Blk:   	Sup_PUA_A}', "");
Error('\p{Is_Block=/a/  SUPPLEMENTARY_private_use_area_A}');
Error('\P{Is_Block=/a/  SUPPLEMENTARY_private_use_area_A}');
Expect(1, 983040, '\p{Is_Block=supplementaryprivateuseareaa}', "");
Expect(0, 983040, '\p{^Is_Block=supplementaryprivateuseareaa}', "");
Expect(0, 983040, '\P{Is_Block=supplementaryprivateuseareaa}', "");
Expect(1, 983040, '\P{^Is_Block=supplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\p{Is_Block=supplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\p{^Is_Block=supplementaryprivateuseareaa}', "");
Expect(1, 1048576, '\P{Is_Block=supplementaryprivateuseareaa}', "");
Expect(0, 1048576, '\P{^Is_Block=supplementaryprivateuseareaa}', "");
Expect(1, 983040, '\p{Is_Block=	_Supplementary_Private_Use_AREA_A}', "");
Expect(0, 983040, '\p{^Is_Block=	_Supplementary_Private_Use_AREA_A}', "");
Expect(0, 983040, '\P{Is_Block=	_Supplementary_Private_Use_AREA_A}', "");
Expect(1, 983040, '\P{^Is_Block=	_Supplementary_Private_Use_AREA_A}', "");
Expect(0, 1048576, '\p{Is_Block=	_Supplementary_Private_Use_AREA_A}', "");
Expect(1, 1048576, '\p{^Is_Block=	_Supplementary_Private_Use_AREA_A}', "");
Expect(1, 1048576, '\P{Is_Block=	_Supplementary_Private_Use_AREA_A}', "");
Expect(0, 1048576, '\P{^Is_Block=	_Supplementary_Private_Use_AREA_A}', "");
Error('\p{Is_Blk=/a/_-Sup_pua_A}');
Error('\P{Is_Blk=/a/_-Sup_pua_A}');
Expect(1, 983040, '\p{Is_Blk=suppuaa}', "");
Expect(0, 983040, '\p{^Is_Blk=suppuaa}', "");
Expect(0, 983040, '\P{Is_Blk=suppuaa}', "");
Expect(1, 983040, '\P{^Is_Blk=suppuaa}', "");
Expect(0, 1048576, '\p{Is_Blk=suppuaa}', "");
Expect(1, 1048576, '\p{^Is_Blk=suppuaa}', "");
Expect(1, 1048576, '\P{Is_Blk=suppuaa}', "");
Expect(0, 1048576, '\P{^Is_Blk=suppuaa}', "");
Expect(1, 983040, '\p{Is_Blk=_	SUP_PUA_A}', "");
Expect(0, 983040, '\p{^Is_Blk=_	SUP_PUA_A}', "");
Expect(0, 983040, '\P{Is_Blk=_	SUP_PUA_A}', "");
Expect(1, 983040, '\P{^Is_Blk=_	SUP_PUA_A}', "");
Expect(0, 1048576, '\p{Is_Blk=_	SUP_PUA_A}', "");
Expect(1, 1048576, '\p{^Is_Blk=_	SUP_PUA_A}', "");
Expect(1, 1048576, '\P{Is_Blk=_	SUP_PUA_A}', "");
Expect(0, 1048576, '\P{^Is_Blk=_	SUP_PUA_A}', "");
Error('\p{Block=:=Khmer_Symbols}');
Error('\P{Block=:=Khmer_Symbols}');
Expect(1, 6655, '\p{Block=khmersymbols}', "");
Expect(0, 6655, '\p{^Block=khmersymbols}', "");
Expect(0, 6655, '\P{Block=khmersymbols}', "");
Expect(1, 6655, '\P{^Block=khmersymbols}', "");
Expect(0, 6656, '\p{Block=khmersymbols}', "");
Expect(1, 6656, '\p{^Block=khmersymbols}', "");
Expect(1, 6656, '\P{Block=khmersymbols}', "");
Expect(0, 6656, '\P{^Block=khmersymbols}', "");
Expect(1, 6655, '\p{Block= 	Khmer_Symbols}', "");
Expect(0, 6655, '\p{^Block= 	Khmer_Symbols}', "");
Expect(0, 6655, '\P{Block= 	Khmer_Symbols}', "");
Expect(1, 6655, '\P{^Block= 	Khmer_Symbols}', "");
Expect(0, 6656, '\p{Block= 	Khmer_Symbols}', "");
Expect(1, 6656, '\p{^Block= 	Khmer_Symbols}', "");
Expect(1, 6656, '\P{Block= 	Khmer_Symbols}', "");
Expect(0, 6656, '\P{^Block= 	Khmer_Symbols}', "");
Error('\p{Blk=_/a/Khmer_Symbols}');
Error('\P{Blk=_/a/Khmer_Symbols}');
Expect(1, 6655, '\p{Blk=khmersymbols}', "");
Expect(0, 6655, '\p{^Blk=khmersymbols}', "");
Expect(0, 6655, '\P{Blk=khmersymbols}', "");
Expect(1, 6655, '\P{^Blk=khmersymbols}', "");
Expect(0, 6656, '\p{Blk=khmersymbols}', "");
Expect(1, 6656, '\p{^Blk=khmersymbols}', "");
Expect(1, 6656, '\P{Blk=khmersymbols}', "");
Expect(0, 6656, '\P{^Blk=khmersymbols}', "");
Expect(1, 6655, '\p{Blk= khmer_SYMBOLS}', "");
Expect(0, 6655, '\p{^Blk= khmer_SYMBOLS}', "");
Expect(0, 6655, '\P{Blk= khmer_SYMBOLS}', "");
Expect(1, 6655, '\P{^Blk= khmer_SYMBOLS}', "");
Expect(0, 6656, '\p{Blk= khmer_SYMBOLS}', "");
Expect(1, 6656, '\p{^Blk= khmer_SYMBOLS}', "");
Expect(1, 6656, '\P{Blk= khmer_SYMBOLS}', "");
Expect(0, 6656, '\P{^Blk= khmer_SYMBOLS}', "");
Error('\p{Is_Block=--Khmer_Symbols/a/}');
Error('\P{Is_Block=--Khmer_Symbols/a/}');
Expect(1, 6655, '\p{Is_Block=khmersymbols}', "");
Expect(0, 6655, '\p{^Is_Block=khmersymbols}', "");
Expect(0, 6655, '\P{Is_Block=khmersymbols}', "");
Expect(1, 6655, '\P{^Is_Block=khmersymbols}', "");
Expect(0, 6656, '\p{Is_Block=khmersymbols}', "");
Expect(1, 6656, '\p{^Is_Block=khmersymbols}', "");
Expect(1, 6656, '\P{Is_Block=khmersymbols}', "");
Expect(0, 6656, '\P{^Is_Block=khmersymbols}', "");
Expect(1, 6655, '\p{Is_Block= 	Khmer_Symbols}', "");
Expect(0, 6655, '\p{^Is_Block= 	Khmer_Symbols}', "");
Expect(0, 6655, '\P{Is_Block= 	Khmer_Symbols}', "");
Expect(1, 6655, '\P{^Is_Block= 	Khmer_Symbols}', "");
Expect(0, 6656, '\p{Is_Block= 	Khmer_Symbols}', "");
Expect(1, 6656, '\p{^Is_Block= 	Khmer_Symbols}', "");
Expect(1, 6656, '\P{Is_Block= 	Khmer_Symbols}', "");
Expect(0, 6656, '\P{^Is_Block= 	Khmer_Symbols}', "");
Error('\p{Is_Blk=-/a/khmer_Symbols}');
Error('\P{Is_Blk=-/a/khmer_Symbols}');
Expect(1, 6655, '\p{Is_Blk=khmersymbols}', "");
Expect(0, 6655, '\p{^Is_Blk=khmersymbols}', "");
Expect(0, 6655, '\P{Is_Blk=khmersymbols}', "");
Expect(1, 6655, '\P{^Is_Blk=khmersymbols}', "");
Expect(0, 6656, '\p{Is_Blk=khmersymbols}', "");
Expect(1, 6656, '\p{^Is_Blk=khmersymbols}', "");
Expect(1, 6656, '\P{Is_Blk=khmersymbols}', "");
Expect(0, 6656, '\P{^Is_Blk=khmersymbols}', "");
Expect(1, 6655, '\p{Is_Blk= _khmer_SYMBOLS}', "");
Expect(0, 6655, '\p{^Is_Blk= _khmer_SYMBOLS}', "");
Expect(0, 6655, '\P{Is_Blk= _khmer_SYMBOLS}', "");
Expect(1, 6655, '\P{^Is_Blk= _khmer_SYMBOLS}', "");
Expect(0, 6656, '\p{Is_Blk= _khmer_SYMBOLS}', "");
Expect(1, 6656, '\p{^Is_Blk= _khmer_SYMBOLS}', "");
Expect(1, 6656, '\P{Is_Blk= _khmer_SYMBOLS}', "");
Expect(0, 6656, '\P{^Is_Blk= _khmer_SYMBOLS}', "");
Error('\p{Block=:=hangul_JAMO}');
Error('\P{Block=:=hangul_JAMO}');
Expect(1, 4607, '\p{Block=hanguljamo}', "");
Expect(0, 4607, '\p{^Block=hanguljamo}', "");
Expect(0, 4607, '\P{Block=hanguljamo}', "");
Expect(1, 4607, '\P{^Block=hanguljamo}', "");
Expect(0, 4608, '\p{Block=hanguljamo}', "");
Expect(1, 4608, '\p{^Block=hanguljamo}', "");
Expect(1, 4608, '\P{Block=hanguljamo}', "");
Expect(0, 4608, '\P{^Block=hanguljamo}', "");
Expect(1, 4607, '\p{Block=_Hangul_jamo}', "");
Expect(0, 4607, '\p{^Block=_Hangul_jamo}', "");
Expect(0, 4607, '\P{Block=_Hangul_jamo}', "");
Expect(1, 4607, '\P{^Block=_Hangul_jamo}', "");
Expect(0, 4608, '\p{Block=_Hangul_jamo}', "");
Expect(1, 4608, '\p{^Block=_Hangul_jamo}', "");
Expect(1, 4608, '\P{Block=_Hangul_jamo}', "");
Expect(0, 4608, '\P{^Block=_Hangul_jamo}', "");
Error('\p{Blk=_Jamo:=}');
Error('\P{Blk=_Jamo:=}');
Expect(1, 4607, '\p{Blk=jamo}', "");
Expect(0, 4607, '\p{^Blk=jamo}', "");
Expect(0, 4607, '\P{Blk=jamo}', "");
Expect(1, 4607, '\P{^Blk=jamo}', "");
Expect(0, 4608, '\p{Blk=jamo}', "");
Expect(1, 4608, '\p{^Blk=jamo}', "");
Expect(1, 4608, '\P{Blk=jamo}', "");
Expect(0, 4608, '\P{^Blk=jamo}', "");
Expect(1, 4607, '\p{Blk=- Jamo}', "");
Expect(0, 4607, '\p{^Blk=- Jamo}', "");
Expect(0, 4607, '\P{Blk=- Jamo}', "");
Expect(1, 4607, '\P{^Blk=- Jamo}', "");
Expect(0, 4608, '\p{Blk=- Jamo}', "");
Expect(1, 4608, '\p{^Blk=- Jamo}', "");
Expect(1, 4608, '\P{Blk=- Jamo}', "");
Expect(0, 4608, '\P{^Blk=- Jamo}', "");
Error('\p{Is_Block= _HANGUL_jamo/a/}');
Error('\P{Is_Block= _HANGUL_jamo/a/}');
Expect(1, 4607, '\p{Is_Block=hanguljamo}', "");
Expect(0, 4607, '\p{^Is_Block=hanguljamo}', "");
Expect(0, 4607, '\P{Is_Block=hanguljamo}', "");
Expect(1, 4607, '\P{^Is_Block=hanguljamo}', "");
Expect(0, 4608, '\p{Is_Block=hanguljamo}', "");
Expect(1, 4608, '\p{^Is_Block=hanguljamo}', "");
Expect(1, 4608, '\P{Is_Block=hanguljamo}', "");
Expect(0, 4608, '\P{^Is_Block=hanguljamo}', "");
Expect(1, 4607, '\p{Is_Block=-	HANGUL_Jamo}', "");
Expect(0, 4607, '\p{^Is_Block=-	HANGUL_Jamo}', "");
Expect(0, 4607, '\P{Is_Block=-	HANGUL_Jamo}', "");
Expect(1, 4607, '\P{^Is_Block=-	HANGUL_Jamo}', "");
Expect(0, 4608, '\p{Is_Block=-	HANGUL_Jamo}', "");
Expect(1, 4608, '\p{^Is_Block=-	HANGUL_Jamo}', "");
Expect(1, 4608, '\P{Is_Block=-	HANGUL_Jamo}', "");
Expect(0, 4608, '\P{^Is_Block=-	HANGUL_Jamo}', "");
Error('\p{Is_Blk=_ jamo/a/}');
Error('\P{Is_Blk=_ jamo/a/}');
Expect(1, 4607, '\p{Is_Blk=jamo}', "");
Expect(0, 4607, '\p{^Is_Blk=jamo}', "");
Expect(0, 4607, '\P{Is_Blk=jamo}', "");
Expect(1, 4607, '\P{^Is_Blk=jamo}', "");
Expect(0, 4608, '\p{Is_Blk=jamo}', "");
Expect(1, 4608, '\p{^Is_Blk=jamo}', "");
Expect(1, 4608, '\P{Is_Blk=jamo}', "");
Expect(0, 4608, '\P{^Is_Blk=jamo}', "");
Expect(1, 4607, '\p{Is_Blk= Jamo}', "");
Expect(0, 4607, '\p{^Is_Blk= Jamo}', "");
Expect(0, 4607, '\P{Is_Blk= Jamo}', "");
Expect(1, 4607, '\P{^Is_Blk= Jamo}', "");
Expect(0, 4608, '\p{Is_Blk= Jamo}', "");
Expect(1, 4608, '\p{^Is_Blk= Jamo}', "");
Expect(1, 4608, '\P{Is_Blk= Jamo}', "");
Expect(0, 4608, '\P{^Is_Blk= Jamo}', "");
Error('\p{Block=- Latin_Extended_A/a/}');
Error('\P{Block=- Latin_Extended_A/a/}');
Expect(1, 383, '\p{Block=latinextendeda}', "");
Expect(0, 383, '\p{^Block=latinextendeda}', "");
Expect(0, 383, '\P{Block=latinextendeda}', "");
Expect(1, 383, '\P{^Block=latinextendeda}', "");
Expect(0, 384, '\p{Block=latinextendeda}', "");
Expect(1, 384, '\p{^Block=latinextendeda}', "");
Expect(1, 384, '\P{Block=latinextendeda}', "");
Expect(0, 384, '\P{^Block=latinextendeda}', "");
Expect(1, 383, '\p{Block=	 Latin_Extended_A}', "");
Expect(0, 383, '\p{^Block=	 Latin_Extended_A}', "");
Expect(0, 383, '\P{Block=	 Latin_Extended_A}', "");
Expect(1, 383, '\P{^Block=	 Latin_Extended_A}', "");
Expect(0, 384, '\p{Block=	 Latin_Extended_A}', "");
Expect(1, 384, '\p{^Block=	 Latin_Extended_A}', "");
Expect(1, 384, '\P{Block=	 Latin_Extended_A}', "");
Expect(0, 384, '\P{^Block=	 Latin_Extended_A}', "");
Error('\p{Blk=:= 	LATIN_EXT_A}');
Error('\P{Blk=:= 	LATIN_EXT_A}');
Expect(1, 383, '\p{Blk=latinexta}', "");
Expect(0, 383, '\p{^Blk=latinexta}', "");
Expect(0, 383, '\P{Blk=latinexta}', "");
Expect(1, 383, '\P{^Blk=latinexta}', "");
Expect(0, 384, '\p{Blk=latinexta}', "");
Expect(1, 384, '\p{^Blk=latinexta}', "");
Expect(1, 384, '\P{Blk=latinexta}', "");
Expect(0, 384, '\P{^Blk=latinexta}', "");
Expect(1, 383, '\p{Blk=- latin_Ext_A}', "");
Expect(0, 383, '\p{^Blk=- latin_Ext_A}', "");
Expect(0, 383, '\P{Blk=- latin_Ext_A}', "");
Expect(1, 383, '\P{^Blk=- latin_Ext_A}', "");
Expect(0, 384, '\p{Blk=- latin_Ext_A}', "");
Expect(1, 384, '\p{^Blk=- latin_Ext_A}', "");
Expect(1, 384, '\P{Blk=- latin_Ext_A}', "");
Expect(0, 384, '\P{^Blk=- latin_Ext_A}', "");
Error('\p{Is_Block= latin_extended_A:=}');
Error('\P{Is_Block= latin_extended_A:=}');
Expect(1, 383, '\p{Is_Block:   latinextendeda}', "");
Expect(0, 383, '\p{^Is_Block:   latinextendeda}', "");
Expect(0, 383, '\P{Is_Block:   latinextendeda}', "");
Expect(1, 383, '\P{^Is_Block:   latinextendeda}', "");
Expect(0, 384, '\p{Is_Block:   latinextendeda}', "");
Expect(1, 384, '\p{^Is_Block:   latinextendeda}', "");
Expect(1, 384, '\P{Is_Block:   latinextendeda}', "");
Expect(0, 384, '\P{^Is_Block:   latinextendeda}', "");
Expect(1, 383, '\p{Is_Block:   --Latin_Extended_a}', "");
Expect(0, 383, '\p{^Is_Block:   --Latin_Extended_a}', "");
Expect(0, 383, '\P{Is_Block:   --Latin_Extended_a}', "");
Expect(1, 383, '\P{^Is_Block:   --Latin_Extended_a}', "");
Expect(0, 384, '\p{Is_Block:   --Latin_Extended_a}', "");
Expect(1, 384, '\p{^Is_Block:   --Latin_Extended_a}', "");
Expect(1, 384, '\P{Is_Block:   --Latin_Extended_a}', "");
Expect(0, 384, '\P{^Is_Block:   --Latin_Extended_a}', "");
Error('\p{Is_Blk=/a/ _Latin_Ext_A}');
Error('\P{Is_Blk=/a/ _Latin_Ext_A}');
Expect(1, 383, '\p{Is_Blk=latinexta}', "");
Expect(0, 383, '\p{^Is_Blk=latinexta}', "");
Expect(0, 383, '\P{Is_Blk=latinexta}', "");
Expect(1, 383, '\P{^Is_Blk=latinexta}', "");
Expect(0, 384, '\p{Is_Blk=latinexta}', "");
Expect(1, 384, '\p{^Is_Blk=latinexta}', "");
Expect(1, 384, '\P{Is_Blk=latinexta}', "");
Expect(0, 384, '\P{^Is_Blk=latinexta}', "");
Expect(1, 383, '\p{Is_Blk=-	Latin_Ext_A}', "");
Expect(0, 383, '\p{^Is_Blk=-	Latin_Ext_A}', "");
Expect(0, 383, '\P{Is_Blk=-	Latin_Ext_A}', "");
Expect(1, 383, '\P{^Is_Blk=-	Latin_Ext_A}', "");
Expect(0, 384, '\p{Is_Blk=-	Latin_Ext_A}', "");
Expect(1, 384, '\p{^Is_Blk=-	Latin_Ext_A}', "");
Expect(1, 384, '\P{Is_Blk=-	Latin_Ext_A}', "");
Expect(0, 384, '\P{^Is_Blk=-	Latin_Ext_A}', "");
Error('\p{Block=_-CJK_compatibility_ideographs_Supplement:=}');
Error('\P{Block=_-CJK_compatibility_ideographs_Supplement:=}');
Expect(1, 195103, '\p{Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\p{^Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\P{Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\P{^Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\p{Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\p{^Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\P{Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\P{^Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\p{Block=CJK_Compatibility_IDEOGRAPHS_Supplement}', "");
Expect(0, 195103, '\p{^Block=CJK_Compatibility_IDEOGRAPHS_Supplement}', "");
Expect(0, 195103, '\P{Block=CJK_Compatibility_IDEOGRAPHS_Supplement}', "");
Expect(1, 195103, '\P{^Block=CJK_Compatibility_IDEOGRAPHS_Supplement}', "");
Expect(0, 195104, '\p{Block=CJK_Compatibility_IDEOGRAPHS_Supplement}', "");
Expect(1, 195104, '\p{^Block=CJK_Compatibility_IDEOGRAPHS_Supplement}', "");
Expect(1, 195104, '\P{Block=CJK_Compatibility_IDEOGRAPHS_Supplement}', "");
Expect(0, 195104, '\P{^Block=CJK_Compatibility_IDEOGRAPHS_Supplement}', "");
Error('\p{Blk= CJK_COMPAT_Ideographs_SUP:=}');
Error('\P{Blk= CJK_COMPAT_Ideographs_SUP:=}');
Expect(1, 195103, '\p{Blk=cjkcompatideographssup}', "");
Expect(0, 195103, '\p{^Blk=cjkcompatideographssup}', "");
Expect(0, 195103, '\P{Blk=cjkcompatideographssup}', "");
Expect(1, 195103, '\P{^Blk=cjkcompatideographssup}', "");
Expect(0, 195104, '\p{Blk=cjkcompatideographssup}', "");
Expect(1, 195104, '\p{^Blk=cjkcompatideographssup}', "");
Expect(1, 195104, '\P{Blk=cjkcompatideographssup}', "");
Expect(0, 195104, '\P{^Blk=cjkcompatideographssup}', "");
Expect(1, 195103, '\p{Blk= 	cjk_compat_ideographs_SUP}', "");
Expect(0, 195103, '\p{^Blk= 	cjk_compat_ideographs_SUP}', "");
Expect(0, 195103, '\P{Blk= 	cjk_compat_ideographs_SUP}', "");
Expect(1, 195103, '\P{^Blk= 	cjk_compat_ideographs_SUP}', "");
Expect(0, 195104, '\p{Blk= 	cjk_compat_ideographs_SUP}', "");
Expect(1, 195104, '\p{^Blk= 	cjk_compat_ideographs_SUP}', "");
Expect(1, 195104, '\P{Blk= 	cjk_compat_ideographs_SUP}', "");
Expect(0, 195104, '\P{^Blk= 	cjk_compat_ideographs_SUP}', "");
Error('\p{Is_Block= cjk_Compatibility_Ideographs_SUPPLEMENT/a/}');
Error('\P{Is_Block= cjk_Compatibility_Ideographs_SUPPLEMENT/a/}');
Expect(1, 195103, '\p{Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\p{^Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195103, '\P{Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\P{^Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\p{Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\p{^Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195104, '\P{Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(0, 195104, '\P{^Is_Block=cjkcompatibilityideographssupplement}', "");
Expect(1, 195103, '\p{Is_Block=_CJK_Compatibility_ideographs_SUPPLEMENT}', "");
Expect(0, 195103, '\p{^Is_Block=_CJK_Compatibility_ideographs_SUPPLEMENT}', "");
Expect(0, 195103, '\P{Is_Block=_CJK_Compatibility_ideographs_SUPPLEMENT}', "");
Expect(1, 195103, '\P{^Is_Block=_CJK_Compatibility_ideographs_SUPPLEMENT}', "");
Expect(0, 195104, '\p{Is_Block=_CJK_Compatibility_ideographs_SUPPLEMENT}', "");
Expect(1, 195104, '\p{^Is_Block=_CJK_Compatibility_ideographs_SUPPLEMENT}', "");
Expect(1, 195104, '\P{Is_Block=_CJK_Compatibility_ideographs_SUPPLEMENT}', "");
Expect(0, 195104, '\P{^Is_Block=_CJK_Compatibility_ideographs_SUPPLEMENT}', "");
Error('\p{Is_Blk=:=	 cjk_Compat_Ideographs_Sup}');
Error('\P{Is_Blk=:=	 cjk_Compat_Ideographs_Sup}');
Expect(1, 195103, '\p{Is_Blk=cjkcompatideographssup}', "");
Expect(0, 195103, '\p{^Is_Blk=cjkcompatideographssup}', "");
Expect(0, 195103, '\P{Is_Blk=cjkcompatideographssup}', "");
Expect(1, 195103, '\P{^Is_Blk=cjkcompatideographssup}', "");
Expect(0, 195104, '\p{Is_Blk=cjkcompatideographssup}', "");
Expect(1, 195104, '\p{^Is_Blk=cjkcompatideographssup}', "");
Expect(1, 195104, '\P{Is_Blk=cjkcompatideographssup}', "");
Expect(0, 195104, '\P{^Is_Blk=cjkcompatideographssup}', "");
Expect(1, 195103, '\p{Is_Blk= _cjk_compat_ideographs_Sup}', "");
Expect(0, 195103, '\p{^Is_Blk= _cjk_compat_ideographs_Sup}', "");
Expect(0, 195103, '\P{Is_Blk= _cjk_compat_ideographs_Sup}', "");
Expect(1, 195103, '\P{^Is_Blk= _cjk_compat_ideographs_Sup}', "");
Expect(0, 195104, '\p{Is_Blk= _cjk_compat_ideographs_Sup}', "");
Expect(1, 195104, '\p{^Is_Blk= _cjk_compat_ideographs_Sup}', "");
Expect(1, 195104, '\P{Is_Blk= _cjk_compat_ideographs_Sup}', "");
Expect(0, 195104, '\P{^Is_Blk= _cjk_compat_ideographs_Sup}', "");
Error('\p{Block= -ipa_Extensions:=}');
Error('\P{Block= -ipa_Extensions:=}');
Expect(1, 687, '\p{Block=ipaextensions}', "");
Expect(0, 687, '\p{^Block=ipaextensions}', "");
Expect(0, 687, '\P{Block=ipaextensions}', "");
Expect(1, 687, '\P{^Block=ipaextensions}', "");
Expect(0, 688, '\p{Block=ipaextensions}', "");
Expect(1, 688, '\p{^Block=ipaextensions}', "");
Expect(1, 688, '\P{Block=ipaextensions}', "");
Expect(0, 688, '\P{^Block=ipaextensions}', "");
Expect(1, 687, '\p{Block=_ipa_Extensions}', "");
Expect(0, 687, '\p{^Block=_ipa_Extensions}', "");
Expect(0, 687, '\P{Block=_ipa_Extensions}', "");
Expect(1, 687, '\P{^Block=_ipa_Extensions}', "");
Expect(0, 688, '\p{Block=_ipa_Extensions}', "");
Expect(1, 688, '\p{^Block=_ipa_Extensions}', "");
Expect(1, 688, '\P{Block=_ipa_Extensions}', "");
Expect(0, 688, '\P{^Block=_ipa_Extensions}', "");
Error('\p{Blk= :=IPA_ext}');
Error('\P{Blk= :=IPA_ext}');
Expect(1, 687, '\p{Blk=ipaext}', "");
Expect(0, 687, '\p{^Blk=ipaext}', "");
Expect(0, 687, '\P{Blk=ipaext}', "");
Expect(1, 687, '\P{^Blk=ipaext}', "");
Expect(0, 688, '\p{Blk=ipaext}', "");
Expect(1, 688, '\p{^Blk=ipaext}', "");
Expect(1, 688, '\P{Blk=ipaext}', "");
Expect(0, 688, '\P{^Blk=ipaext}', "");
Expect(1, 687, '\p{Blk=_ IPA_Ext}', "");
Expect(0, 687, '\p{^Blk=_ IPA_Ext}', "");
Expect(0, 687, '\P{Blk=_ IPA_Ext}', "");
Expect(1, 687, '\P{^Blk=_ IPA_Ext}', "");
Expect(0, 688, '\p{Blk=_ IPA_Ext}', "");
Expect(1, 688, '\p{^Blk=_ IPA_Ext}', "");
Expect(1, 688, '\P{Blk=_ IPA_Ext}', "");
Expect(0, 688, '\P{^Blk=_ IPA_Ext}', "");
Error('\p{Is_Block=/a/IPA_Extensions}');
Error('\P{Is_Block=/a/IPA_Extensions}');
Expect(1, 687, '\p{Is_Block=ipaextensions}', "");
Expect(0, 687, '\p{^Is_Block=ipaextensions}', "");
Expect(0, 687, '\P{Is_Block=ipaextensions}', "");
Expect(1, 687, '\P{^Is_Block=ipaextensions}', "");
Expect(0, 688, '\p{Is_Block=ipaextensions}', "");
Expect(1, 688, '\p{^Is_Block=ipaextensions}', "");
Expect(1, 688, '\P{Is_Block=ipaextensions}', "");
Expect(0, 688, '\P{^Is_Block=ipaextensions}', "");
Expect(1, 687, '\p{Is_Block= IPA_Extensions}', "");
Expect(0, 687, '\p{^Is_Block= IPA_Extensions}', "");
Expect(0, 687, '\P{Is_Block= IPA_Extensions}', "");
Expect(1, 687, '\P{^Is_Block= IPA_Extensions}', "");
Expect(0, 688, '\p{Is_Block= IPA_Extensions}', "");
Expect(1, 688, '\p{^Is_Block= IPA_Extensions}', "");
Expect(1, 688, '\P{Is_Block= IPA_Extensions}', "");
Expect(0, 688, '\P{^Is_Block= IPA_Extensions}', "");
Error('\p{Is_Blk=/a/  IPA_EXT}');
Error('\P{Is_Blk=/a/  IPA_EXT}');
Expect(1, 687, '\p{Is_Blk=ipaext}', "");
Expect(0, 687, '\p{^Is_Blk=ipaext}', "");
Expect(0, 687, '\P{Is_Blk=ipaext}', "");
Expect(1, 687, '\P{^Is_Blk=ipaext}', "");
Expect(0, 688, '\p{Is_Blk=ipaext}', "");
Expect(1, 688, '\p{^Is_Blk=ipaext}', "");
Expect(1, 688, '\P{Is_Blk=ipaext}', "");
Expect(0, 688, '\P{^Is_Blk=ipaext}', "");
Expect(1, 687, '\p{Is_Blk=_ipa_Ext}', "");
Expect(0, 687, '\p{^Is_Blk=_ipa_Ext}', "");
Expect(0, 687, '\P{Is_Blk=_ipa_Ext}', "");
Expect(1, 687, '\P{^Is_Blk=_ipa_Ext}', "");
Expect(0, 688, '\p{Is_Blk=_ipa_Ext}', "");
Expect(1, 688, '\p{^Is_Blk=_ipa_Ext}', "");
Expect(1, 688, '\P{Is_Blk=_ipa_Ext}', "");
Expect(0, 688, '\P{^Is_Blk=_ipa_Ext}', "");
Error('\p{Block= :=Kannada}');
Error('\P{Block= :=Kannada}');
Expect(1, 3327, '\p{Block=kannada}', "");
Expect(0, 3327, '\p{^Block=kannada}', "");
Expect(0, 3327, '\P{Block=kannada}', "");
Expect(1, 3327, '\P{^Block=kannada}', "");
Expect(0, 3328, '\p{Block=kannada}', "");
Expect(1, 3328, '\p{^Block=kannada}', "");
Expect(1, 3328, '\P{Block=kannada}', "");
Expect(0, 3328, '\P{^Block=kannada}', "");
Expect(1, 3327, '\p{Block=  Kannada}', "");
Expect(0, 3327, '\p{^Block=  Kannada}', "");
Expect(0, 3327, '\P{Block=  Kannada}', "");
Expect(1, 3327, '\P{^Block=  Kannada}', "");
Expect(0, 3328, '\p{Block=  Kannada}', "");
Expect(1, 3328, '\p{^Block=  Kannada}', "");
Expect(1, 3328, '\P{Block=  Kannada}', "");
Expect(0, 3328, '\P{^Block=  Kannada}', "");
Error('\p{Blk=	:=Kannada}');
Error('\P{Blk=	:=Kannada}');
Expect(1, 3327, '\p{Blk=kannada}', "");
Expect(0, 3327, '\p{^Blk=kannada}', "");
Expect(0, 3327, '\P{Blk=kannada}', "");
Expect(1, 3327, '\P{^Blk=kannada}', "");
Expect(0, 3328, '\p{Blk=kannada}', "");
Expect(1, 3328, '\p{^Blk=kannada}', "");
Expect(1, 3328, '\P{Blk=kannada}', "");
Expect(0, 3328, '\P{^Blk=kannada}', "");
Expect(1, 3327, '\p{Blk: 	-kannada}', "");
Expect(0, 3327, '\p{^Blk: 	-kannada}', "");
Expect(0, 3327, '\P{Blk: 	-kannada}', "");
Expect(1, 3327, '\P{^Blk: 	-kannada}', "");
Expect(0, 3328, '\p{Blk: 	-kannada}', "");
Expect(1, 3328, '\p{^Blk: 	-kannada}', "");
Expect(1, 3328, '\P{Blk: 	-kannada}', "");
Expect(0, 3328, '\P{^Blk: 	-kannada}', "");
Error('\p{Is_Block=:=  Kannada}');
Error('\P{Is_Block=:=  Kannada}');
Expect(1, 3327, '\p{Is_Block=kannada}', "");
Expect(0, 3327, '\p{^Is_Block=kannada}', "");
Expect(0, 3327, '\P{Is_Block=kannada}', "");
Expect(1, 3327, '\P{^Is_Block=kannada}', "");
Expect(0, 3328, '\p{Is_Block=kannada}', "");
Expect(1, 3328, '\p{^Is_Block=kannada}', "");
Expect(1, 3328, '\P{Is_Block=kannada}', "");
Expect(0, 3328, '\P{^Is_Block=kannada}', "");
Expect(1, 3327, '\p{Is_Block=_-kannada}', "");
Expect(0, 3327, '\p{^Is_Block=_-kannada}', "");
Expect(0, 3327, '\P{Is_Block=_-kannada}', "");
Expect(1, 3327, '\P{^Is_Block=_-kannada}', "");
Expect(0, 3328, '\p{Is_Block=_-kannada}', "");
Expect(1, 3328, '\p{^Is_Block=_-kannada}', "");
Expect(1, 3328, '\P{Is_Block=_-kannada}', "");
Expect(0, 3328, '\P{^Is_Block=_-kannada}', "");
Error('\p{Is_Blk=/a/	_Kannada}');
Error('\P{Is_Blk=/a/	_Kannada}');
Expect(1, 3327, '\p{Is_Blk=kannada}', "");
Expect(0, 3327, '\p{^Is_Blk=kannada}', "");
Expect(0, 3327, '\P{Is_Blk=kannada}', "");
Expect(1, 3327, '\P{^Is_Blk=kannada}', "");
Expect(0, 3328, '\p{Is_Blk=kannada}', "");
Expect(1, 3328, '\p{^Is_Blk=kannada}', "");
Expect(1, 3328, '\P{Is_Blk=kannada}', "");
Expect(0, 3328, '\P{^Is_Blk=kannada}', "");
Expect(1, 3327, '\p{Is_Blk=	Kannada}', "");
Expect(0, 3327, '\p{^Is_Blk=	Kannada}', "");
Expect(0, 3327, '\P{Is_Blk=	Kannada}', "");
Expect(1, 3327, '\P{^Is_Blk=	Kannada}', "");
Expect(0, 3328, '\p{Is_Blk=	Kannada}', "");
Expect(1, 3328, '\p{^Is_Blk=	Kannada}', "");
Expect(1, 3328, '\P{Is_Blk=	Kannada}', "");
Expect(0, 3328, '\P{^Is_Blk=	Kannada}', "");
Error('\p{Block=-	GREEK_AND_coptic:=}');
Error('\P{Block=-	GREEK_AND_coptic:=}');
Expect(1, 1023, '\p{Block=greekandcoptic}', "");
Expect(0, 1023, '\p{^Block=greekandcoptic}', "");
Expect(0, 1023, '\P{Block=greekandcoptic}', "");
Expect(1, 1023, '\P{^Block=greekandcoptic}', "");
Expect(0, 1024, '\p{Block=greekandcoptic}', "");
Expect(1, 1024, '\p{^Block=greekandcoptic}', "");
Expect(1, 1024, '\P{Block=greekandcoptic}', "");
Expect(0, 1024, '\P{^Block=greekandcoptic}', "");
Expect(1, 1023, '\p{Block=_ Greek_AND_Coptic}', "");
Expect(0, 1023, '\p{^Block=_ Greek_AND_Coptic}', "");
Expect(0, 1023, '\P{Block=_ Greek_AND_Coptic}', "");
Expect(1, 1023, '\P{^Block=_ Greek_AND_Coptic}', "");
Expect(0, 1024, '\p{Block=_ Greek_AND_Coptic}', "");
Expect(1, 1024, '\p{^Block=_ Greek_AND_Coptic}', "");
Expect(1, 1024, '\P{Block=_ Greek_AND_Coptic}', "");
Expect(0, 1024, '\P{^Block=_ Greek_AND_Coptic}', "");
Error('\p{Blk:	 greek:=}');
Error('\P{Blk:	 greek:=}');
Expect(1, 1023, '\p{Blk=greek}', "");
Expect(0, 1023, '\p{^Blk=greek}', "");
Expect(0, 1023, '\P{Blk=greek}', "");
Expect(1, 1023, '\P{^Blk=greek}', "");
Expect(0, 1024, '\p{Blk=greek}', "");
Expect(1, 1024, '\p{^Blk=greek}', "");
Expect(1, 1024, '\P{Blk=greek}', "");
Expect(0, 1024, '\P{^Blk=greek}', "");
Expect(1, 1023, '\p{Blk=-_Greek}', "");
Expect(0, 1023, '\p{^Blk=-_Greek}', "");
Expect(0, 1023, '\P{Blk=-_Greek}', "");
Expect(1, 1023, '\P{^Blk=-_Greek}', "");
Expect(0, 1024, '\p{Blk=-_Greek}', "");
Expect(1, 1024, '\p{^Blk=-_Greek}', "");
Expect(1, 1024, '\P{Blk=-_Greek}', "");
Expect(0, 1024, '\P{^Blk=-_Greek}', "");
Error('\p{Is_Block=/a/ -Greek_And_coptic}');
Error('\P{Is_Block=/a/ -Greek_And_coptic}');
Expect(1, 1023, '\p{Is_Block=greekandcoptic}', "");
Expect(0, 1023, '\p{^Is_Block=greekandcoptic}', "");
Expect(0, 1023, '\P{Is_Block=greekandcoptic}', "");
Expect(1, 1023, '\P{^Is_Block=greekandcoptic}', "");
Expect(0, 1024, '\p{Is_Block=greekandcoptic}', "");
Expect(1, 1024, '\p{^Is_Block=greekandcoptic}', "");
Expect(1, 1024, '\P{Is_Block=greekandcoptic}', "");
Expect(0, 1024, '\P{^Is_Block=greekandcoptic}', "");
Expect(1, 1023, '\p{Is_Block=_-Greek_AND_Coptic}', "");
Expect(0, 1023, '\p{^Is_Block=_-Greek_AND_Coptic}', "");
Expect(0, 1023, '\P{Is_Block=_-Greek_AND_Coptic}', "");
Expect(1, 1023, '\P{^Is_Block=_-Greek_AND_Coptic}', "");
Expect(0, 1024, '\p{Is_Block=_-Greek_AND_Coptic}', "");
Expect(1, 1024, '\p{^Is_Block=_-Greek_AND_Coptic}', "");
Expect(1, 1024, '\P{Is_Block=_-Greek_AND_Coptic}', "");
Expect(0, 1024, '\P{^Is_Block=_-Greek_AND_Coptic}', "");
Error('\p{Is_Blk=/a/ -Greek}');
Error('\P{Is_Blk=/a/ -Greek}');
Expect(1, 1023, '\p{Is_Blk=greek}', "");
Expect(0, 1023, '\p{^Is_Blk=greek}', "");
Expect(0, 1023, '\P{Is_Blk=greek}', "");
Expect(1, 1023, '\P{^Is_Blk=greek}', "");
Expect(0, 1024, '\p{Is_Blk=greek}', "");
Expect(1, 1024, '\p{^Is_Blk=greek}', "");
Expect(1, 1024, '\P{Is_Blk=greek}', "");
Expect(0, 1024, '\P{^Is_Blk=greek}', "");
Expect(1, 1023, '\p{Is_Blk=-Greek}', "");
Expect(0, 1023, '\p{^Is_Blk=-Greek}', "");
Expect(0, 1023, '\P{Is_Blk=-Greek}', "");
Expect(1, 1023, '\P{^Is_Blk=-Greek}', "");
Expect(0, 1024, '\p{Is_Blk=-Greek}', "");
Expect(1, 1024, '\p{^Is_Blk=-Greek}', "");
Expect(1, 1024, '\P{Is_Blk=-Greek}', "");
Expect(0, 1024, '\P{^Is_Blk=-Greek}', "");
Error('\p{Block=_ Rumi_Numeral_symbols/a/}');
Error('\P{Block=_ Rumi_Numeral_symbols/a/}');
Expect(1, 69247, '\p{Block=ruminumeralsymbols}', "");
Expect(0, 69247, '\p{^Block=ruminumeralsymbols}', "");
Expect(0, 69247, '\P{Block=ruminumeralsymbols}', "");
Expect(1, 69247, '\P{^Block=ruminumeralsymbols}', "");
Expect(0, 69248, '\p{Block=ruminumeralsymbols}', "");
Expect(1, 69248, '\p{^Block=ruminumeralsymbols}', "");
Expect(1, 69248, '\P{Block=ruminumeralsymbols}', "");
Expect(0, 69248, '\P{^Block=ruminumeralsymbols}', "");
Expect(1, 69247, '\p{Block:   -rumi_numeral_Symbols}', "");
Expect(0, 69247, '\p{^Block:   -rumi_numeral_Symbols}', "");
Expect(0, 69247, '\P{Block:   -rumi_numeral_Symbols}', "");
Expect(1, 69247, '\P{^Block:   -rumi_numeral_Symbols}', "");
Expect(0, 69248, '\p{Block:   -rumi_numeral_Symbols}', "");
Expect(1, 69248, '\p{^Block:   -rumi_numeral_Symbols}', "");
Expect(1, 69248, '\P{Block:   -rumi_numeral_Symbols}', "");
Expect(0, 69248, '\P{^Block:   -rumi_numeral_Symbols}', "");
Error('\p{Blk:     rumi:=}');
Error('\P{Blk:     rumi:=}');
Expect(1, 69247, '\p{Blk=rumi}', "");
Expect(0, 69247, '\p{^Blk=rumi}', "");
Expect(0, 69247, '\P{Blk=rumi}', "");
Expect(1, 69247, '\P{^Blk=rumi}', "");
Expect(0, 69248, '\p{Blk=rumi}', "");
Expect(1, 69248, '\p{^Blk=rumi}', "");
Expect(1, 69248, '\P{Blk=rumi}', "");
Expect(0, 69248, '\P{^Blk=rumi}', "");
Expect(1, 69247, '\p{Blk=  Rumi}', "");
Expect(0, 69247, '\p{^Blk=  Rumi}', "");
Expect(0, 69247, '\P{Blk=  Rumi}', "");
Expect(1, 69247, '\P{^Blk=  Rumi}', "");
Expect(0, 69248, '\p{Blk=  Rumi}', "");
Expect(1, 69248, '\p{^Blk=  Rumi}', "");
Expect(1, 69248, '\P{Blk=  Rumi}', "");
Expect(0, 69248, '\P{^Blk=  Rumi}', "");
Error('\p{Is_Block=:=__Rumi_numeral_Symbols}');
Error('\P{Is_Block=:=__Rumi_numeral_Symbols}');
Expect(1, 69247, '\p{Is_Block=ruminumeralsymbols}', "");
Expect(0, 69247, '\p{^Is_Block=ruminumeralsymbols}', "");
Expect(0, 69247, '\P{Is_Block=ruminumeralsymbols}', "");
Expect(1, 69247, '\P{^Is_Block=ruminumeralsymbols}', "");
Expect(0, 69248, '\p{Is_Block=ruminumeralsymbols}', "");
Expect(1, 69248, '\p{^Is_Block=ruminumeralsymbols}', "");
Expect(1, 69248, '\P{Is_Block=ruminumeralsymbols}', "");
Expect(0, 69248, '\P{^Is_Block=ruminumeralsymbols}', "");
Expect(1, 69247, '\p{Is_Block=-Rumi_Numeral_Symbols}', "");
Expect(0, 69247, '\p{^Is_Block=-Rumi_Numeral_Symbols}', "");
Expect(0, 69247, '\P{Is_Block=-Rumi_Numeral_Symbols}', "");
Expect(1, 69247, '\P{^Is_Block=-Rumi_Numeral_Symbols}', "");
Expect(0, 69248, '\p{Is_Block=-Rumi_Numeral_Symbols}', "");
Expect(1, 69248, '\p{^Is_Block=-Rumi_Numeral_Symbols}', "");
Expect(1, 69248, '\P{Is_Block=-Rumi_Numeral_Symbols}', "");
Expect(0, 69248, '\P{^Is_Block=-Rumi_Numeral_Symbols}', "");
Error('\p{Is_Blk=-Rumi:=}');
Error('\P{Is_Blk=-Rumi:=}');
Expect(1, 69247, '\p{Is_Blk=rumi}', "");
Expect(0, 69247, '\p{^Is_Blk=rumi}', "");
Expect(0, 69247, '\P{Is_Blk=rumi}', "");
Expect(1, 69247, '\P{^Is_Blk=rumi}', "");
Expect(0, 69248, '\p{Is_Blk=rumi}', "");
Expect(1, 69248, '\p{^Is_Blk=rumi}', "");
Expect(1, 69248, '\P{Is_Blk=rumi}', "");
Expect(0, 69248, '\P{^Is_Blk=rumi}', "");
Expect(1, 69247, '\p{Is_Blk= rumi}', "");
Expect(0, 69247, '\p{^Is_Blk= rumi}', "");
Expect(0, 69247, '\P{Is_Blk= rumi}', "");
Expect(1, 69247, '\P{^Is_Blk= rumi}', "");
Expect(0, 69248, '\p{Is_Blk= rumi}', "");
Expect(1, 69248, '\p{^Is_Blk= rumi}', "");
Expect(1, 69248, '\P{Is_Blk= rumi}', "");
Expect(0, 69248, '\P{^Is_Blk= rumi}', "");
Error('\p{Block=/a/	Ethiopic_extended}');
Error('\P{Block=/a/	Ethiopic_extended}');
Expect(1, 11743, '\p{Block=ethiopicextended}', "");
Expect(0, 11743, '\p{^Block=ethiopicextended}', "");
Expect(0, 11743, '\P{Block=ethiopicextended}', "");
Expect(1, 11743, '\P{^Block=ethiopicextended}', "");
Expect(0, 11744, '\p{Block=ethiopicextended}', "");
Expect(1, 11744, '\p{^Block=ethiopicextended}', "");
Expect(1, 11744, '\P{Block=ethiopicextended}', "");
Expect(0, 11744, '\P{^Block=ethiopicextended}', "");
Expect(1, 11743, '\p{Block: _	Ethiopic_EXTENDED}', "");
Expect(0, 11743, '\p{^Block: _	Ethiopic_EXTENDED}', "");
Expect(0, 11743, '\P{Block: _	Ethiopic_EXTENDED}', "");
Expect(1, 11743, '\P{^Block: _	Ethiopic_EXTENDED}', "");
Expect(0, 11744, '\p{Block: _	Ethiopic_EXTENDED}', "");
Expect(1, 11744, '\p{^Block: _	Ethiopic_EXTENDED}', "");
Expect(1, 11744, '\P{Block: _	Ethiopic_EXTENDED}', "");
Expect(0, 11744, '\P{^Block: _	Ethiopic_EXTENDED}', "");
Error('\p{Blk=	:=Ethiopic_ext}');
Error('\P{Blk=	:=Ethiopic_ext}');
Expect(1, 11743, '\p{Blk=ethiopicext}', "");
Expect(0, 11743, '\p{^Blk=ethiopicext}', "");
Expect(0, 11743, '\P{Blk=ethiopicext}', "");
Expect(1, 11743, '\P{^Blk=ethiopicext}', "");
Expect(0, 11744, '\p{Blk=ethiopicext}', "");
Expect(1, 11744, '\p{^Blk=ethiopicext}', "");
Expect(1, 11744, '\P{Blk=ethiopicext}', "");
Expect(0, 11744, '\P{^Blk=ethiopicext}', "");
Expect(1, 11743, '\p{Blk=	-ETHIOPIC_EXT}', "");
Expect(0, 11743, '\p{^Blk=	-ETHIOPIC_EXT}', "");
Expect(0, 11743, '\P{Blk=	-ETHIOPIC_EXT}', "");
Expect(1, 11743, '\P{^Blk=	-ETHIOPIC_EXT}', "");
Expect(0, 11744, '\p{Blk=	-ETHIOPIC_EXT}', "");
Expect(1, 11744, '\p{^Blk=	-ETHIOPIC_EXT}', "");
Expect(1, 11744, '\P{Blk=	-ETHIOPIC_EXT}', "");
Expect(0, 11744, '\P{^Blk=	-ETHIOPIC_EXT}', "");
Error('\p{Is_Block=-Ethiopic_extended:=}');
Error('\P{Is_Block=-Ethiopic_extended:=}');
Expect(1, 11743, '\p{Is_Block:   ethiopicextended}', "");
Expect(0, 11743, '\p{^Is_Block:   ethiopicextended}', "");
Expect(0, 11743, '\P{Is_Block:   ethiopicextended}', "");
Expect(1, 11743, '\P{^Is_Block:   ethiopicextended}', "");
Expect(0, 11744, '\p{Is_Block:   ethiopicextended}', "");
Expect(1, 11744, '\p{^Is_Block:   ethiopicextended}', "");
Expect(1, 11744, '\P{Is_Block:   ethiopicextended}', "");
Expect(0, 11744, '\P{^Is_Block:   ethiopicextended}', "");
Expect(1, 11743, '\p{Is_Block=__ethiopic_Extended}', "");
Expect(0, 11743, '\p{^Is_Block=__ethiopic_Extended}', "");
Expect(0, 11743, '\P{Is_Block=__ethiopic_Extended}', "");
Expect(1, 11743, '\P{^Is_Block=__ethiopic_Extended}', "");
Expect(0, 11744, '\p{Is_Block=__ethiopic_Extended}', "");
Expect(1, 11744, '\p{^Is_Block=__ethiopic_Extended}', "");
Expect(1, 11744, '\P{Is_Block=__ethiopic_Extended}', "");
Expect(0, 11744, '\P{^Is_Block=__ethiopic_Extended}', "");
Error('\p{Is_Blk=:= _ETHIOPIC_Ext}');
Error('\P{Is_Blk=:= _ETHIOPIC_Ext}');
Expect(1, 11743, '\p{Is_Blk: ethiopicext}', "");
Expect(0, 11743, '\p{^Is_Blk: ethiopicext}', "");
Expect(0, 11743, '\P{Is_Blk: ethiopicext}', "");
Expect(1, 11743, '\P{^Is_Blk: ethiopicext}', "");
Expect(0, 11744, '\p{Is_Blk: ethiopicext}', "");
Expect(1, 11744, '\p{^Is_Blk: ethiopicext}', "");
Expect(1, 11744, '\P{Is_Blk: ethiopicext}', "");
Expect(0, 11744, '\P{^Is_Blk: ethiopicext}', "");
Expect(1, 11743, '\p{Is_Blk= -Ethiopic_Ext}', "");
Expect(0, 11743, '\p{^Is_Blk= -Ethiopic_Ext}', "");
Expect(0, 11743, '\P{Is_Blk= -Ethiopic_Ext}', "");
Expect(1, 11743, '\P{^Is_Blk= -Ethiopic_Ext}', "");
Expect(0, 11744, '\p{Is_Blk= -Ethiopic_Ext}', "");
Expect(1, 11744, '\p{^Is_Blk= -Ethiopic_Ext}', "");
Expect(1, 11744, '\P{Is_Blk= -Ethiopic_Ext}', "");
Expect(0, 11744, '\P{^Is_Blk= -Ethiopic_Ext}', "");
Error('\p{Block::=-	MANDAIC}');
Error('\P{Block::=-	MANDAIC}');
Expect(1, 2143, '\p{Block=mandaic}', "");
Expect(0, 2143, '\p{^Block=mandaic}', "");
Expect(0, 2143, '\P{Block=mandaic}', "");
Expect(1, 2143, '\P{^Block=mandaic}', "");
Expect(0, 2144, '\p{Block=mandaic}', "");
Expect(1, 2144, '\p{^Block=mandaic}', "");
Expect(1, 2144, '\P{Block=mandaic}', "");
Expect(0, 2144, '\P{^Block=mandaic}', "");
Expect(1, 2143, '\p{Block=_-Mandaic}', "");
Expect(0, 2143, '\p{^Block=_-Mandaic}', "");
Expect(0, 2143, '\P{Block=_-Mandaic}', "");
Expect(1, 2143, '\P{^Block=_-Mandaic}', "");
Expect(0, 2144, '\p{Block=_-Mandaic}', "");
Expect(1, 2144, '\p{^Block=_-Mandaic}', "");
Expect(1, 2144, '\P{Block=_-Mandaic}', "");
Expect(0, 2144, '\P{^Block=_-Mandaic}', "");
Error('\p{Blk=_mandaic/a/}');
Error('\P{Blk=_mandaic/a/}');
Expect(1, 2143, '\p{Blk=mandaic}', "");
Expect(0, 2143, '\p{^Blk=mandaic}', "");
Expect(0, 2143, '\P{Blk=mandaic}', "");
Expect(1, 2143, '\P{^Blk=mandaic}', "");
Expect(0, 2144, '\p{Blk=mandaic}', "");
Expect(1, 2144, '\p{^Blk=mandaic}', "");
Expect(1, 2144, '\P{Blk=mandaic}', "");
Expect(0, 2144, '\P{^Blk=mandaic}', "");
Expect(1, 2143, '\p{Blk=-MANDAIC}', "");
Expect(0, 2143, '\p{^Blk=-MANDAIC}', "");
Expect(0, 2143, '\P{Blk=-MANDAIC}', "");
Expect(1, 2143, '\P{^Blk=-MANDAIC}', "");
Expect(0, 2144, '\p{Blk=-MANDAIC}', "");
Expect(1, 2144, '\p{^Blk=-MANDAIC}', "");
Expect(1, 2144, '\P{Blk=-MANDAIC}', "");
Expect(0, 2144, '\P{^Blk=-MANDAIC}', "");
Error('\p{Is_Block=/a/		Mandaic}');
Error('\P{Is_Block=/a/		Mandaic}');
Expect(1, 2143, '\p{Is_Block=mandaic}', "");
Expect(0, 2143, '\p{^Is_Block=mandaic}', "");
Expect(0, 2143, '\P{Is_Block=mandaic}', "");
Expect(1, 2143, '\P{^Is_Block=mandaic}', "");
Expect(0, 2144, '\p{Is_Block=mandaic}', "");
Expect(1, 2144, '\p{^Is_Block=mandaic}', "");
Expect(1, 2144, '\P{Is_Block=mandaic}', "");
Expect(0, 2144, '\P{^Is_Block=mandaic}', "");
Expect(1, 2143, '\p{Is_Block:   MANDAIC}', "");
Expect(0, 2143, '\p{^Is_Block:   MANDAIC}', "");
Expect(0, 2143, '\P{Is_Block:   MANDAIC}', "");
Expect(1, 2143, '\P{^Is_Block:   MANDAIC}', "");
Expect(0, 2144, '\p{Is_Block:   MANDAIC}', "");
Expect(1, 2144, '\p{^Is_Block:   MANDAIC}', "");
Expect(1, 2144, '\P{Is_Block:   MANDAIC}', "");
Expect(0, 2144, '\P{^Is_Block:   MANDAIC}', "");
Error('\p{Is_Blk=:=	Mandaic}');
Error('\P{Is_Blk=:=	Mandaic}');
Expect(1, 2143, '\p{Is_Blk:mandaic}', "");
Expect(0, 2143, '\p{^Is_Blk:mandaic}', "");
Expect(0, 2143, '\P{Is_Blk:mandaic}', "");
Expect(1, 2143, '\P{^Is_Blk:mandaic}', "");
Expect(0, 2144, '\p{Is_Blk:mandaic}', "");
Expect(1, 2144, '\p{^Is_Blk:mandaic}', "");
Expect(1, 2144, '\P{Is_Blk:mandaic}', "");
Expect(0, 2144, '\P{^Is_Blk:mandaic}', "");
Expect(1, 2143, '\p{Is_Blk=- MANDAIC}', "");
Expect(0, 2143, '\p{^Is_Blk=- MANDAIC}', "");
Expect(0, 2143, '\P{Is_Blk=- MANDAIC}', "");
Expect(1, 2143, '\P{^Is_Blk=- MANDAIC}', "");
Expect(0, 2144, '\p{Is_Blk=- MANDAIC}', "");
Expect(1, 2144, '\p{^Is_Blk=- MANDAIC}', "");
Expect(1, 2144, '\P{Is_Blk=- MANDAIC}', "");
Expect(0, 2144, '\P{^Is_Blk=- MANDAIC}', "");
Error('\p{Block=_:=cypriot_Syllabary}');
Error('\P{Block=_:=cypriot_Syllabary}');
Expect(1, 67647, '\p{Block=cypriotsyllabary}', "");
Expect(0, 67647, '\p{^Block=cypriotsyllabary}', "");
Expect(0, 67647, '\P{Block=cypriotsyllabary}', "");
Expect(1, 67647, '\P{^Block=cypriotsyllabary}', "");
Expect(0, 67648, '\p{Block=cypriotsyllabary}', "");
Expect(1, 67648, '\p{^Block=cypriotsyllabary}', "");
Expect(1, 67648, '\P{Block=cypriotsyllabary}', "");
Expect(0, 67648, '\P{^Block=cypriotsyllabary}', "");
Expect(1, 67647, '\p{Block=-CYPRIOT_syllabary}', "");
Expect(0, 67647, '\p{^Block=-CYPRIOT_syllabary}', "");
Expect(0, 67647, '\P{Block=-CYPRIOT_syllabary}', "");
Expect(1, 67647, '\P{^Block=-CYPRIOT_syllabary}', "");
Expect(0, 67648, '\p{Block=-CYPRIOT_syllabary}', "");
Expect(1, 67648, '\p{^Block=-CYPRIOT_syllabary}', "");
Expect(1, 67648, '\P{Block=-CYPRIOT_syllabary}', "");
Expect(0, 67648, '\P{^Block=-CYPRIOT_syllabary}', "");
Error('\p{Blk=_cypriot_Syllabary:=}');
Error('\P{Blk=_cypriot_Syllabary:=}');
Expect(1, 67647, '\p{Blk: cypriotsyllabary}', "");
Expect(0, 67647, '\p{^Blk: cypriotsyllabary}', "");
Expect(0, 67647, '\P{Blk: cypriotsyllabary}', "");
Expect(1, 67647, '\P{^Blk: cypriotsyllabary}', "");
Expect(0, 67648, '\p{Blk: cypriotsyllabary}', "");
Expect(1, 67648, '\p{^Blk: cypriotsyllabary}', "");
Expect(1, 67648, '\P{Blk: cypriotsyllabary}', "");
Expect(0, 67648, '\P{^Blk: cypriotsyllabary}', "");
Expect(1, 67647, '\p{Blk=-_Cypriot_syllabary}', "");
Expect(0, 67647, '\p{^Blk=-_Cypriot_syllabary}', "");
Expect(0, 67647, '\P{Blk=-_Cypriot_syllabary}', "");
Expect(1, 67647, '\P{^Blk=-_Cypriot_syllabary}', "");
Expect(0, 67648, '\p{Blk=-_Cypriot_syllabary}', "");
Expect(1, 67648, '\p{^Blk=-_Cypriot_syllabary}', "");
Expect(1, 67648, '\P{Blk=-_Cypriot_syllabary}', "");
Expect(0, 67648, '\P{^Blk=-_Cypriot_syllabary}', "");
Error('\p{Is_Block=-/a/CYPRIOT_syllabary}');
Error('\P{Is_Block=-/a/CYPRIOT_syllabary}');
Expect(1, 67647, '\p{Is_Block=cypriotsyllabary}', "");
Expect(0, 67647, '\p{^Is_Block=cypriotsyllabary}', "");
Expect(0, 67647, '\P{Is_Block=cypriotsyllabary}', "");
Expect(1, 67647, '\P{^Is_Block=cypriotsyllabary}', "");
Expect(0, 67648, '\p{Is_Block=cypriotsyllabary}', "");
Expect(1, 67648, '\p{^Is_Block=cypriotsyllabary}', "");
Expect(1, 67648, '\P{Is_Block=cypriotsyllabary}', "");
Expect(0, 67648, '\P{^Is_Block=cypriotsyllabary}', "");
Expect(1, 67647, '\p{Is_Block=__CYPRIOT_syllabary}', "");
Expect(0, 67647, '\p{^Is_Block=__CYPRIOT_syllabary}', "");
Expect(0, 67647, '\P{Is_Block=__CYPRIOT_syllabary}', "");
Expect(1, 67647, '\P{^Is_Block=__CYPRIOT_syllabary}', "");
Expect(0, 67648, '\p{Is_Block=__CYPRIOT_syllabary}', "");
Expect(1, 67648, '\p{^Is_Block=__CYPRIOT_syllabary}', "");
Expect(1, 67648, '\P{Is_Block=__CYPRIOT_syllabary}', "");
Expect(0, 67648, '\P{^Is_Block=__CYPRIOT_syllabary}', "");
Error('\p{Is_Blk:		CYPRIOT_Syllabary/a/}');
Error('\P{Is_Blk:		CYPRIOT_Syllabary/a/}');
Expect(1, 67647, '\p{Is_Blk:cypriotsyllabary}', "");
Expect(0, 67647, '\p{^Is_Blk:cypriotsyllabary}', "");
Expect(0, 67647, '\P{Is_Blk:cypriotsyllabary}', "");
Expect(1, 67647, '\P{^Is_Blk:cypriotsyllabary}', "");
Expect(0, 67648, '\p{Is_Blk:cypriotsyllabary}', "");
Expect(1, 67648, '\p{^Is_Blk:cypriotsyllabary}', "");
Expect(1, 67648, '\P{Is_Blk:cypriotsyllabary}', "");
Expect(0, 67648, '\P{^Is_Blk:cypriotsyllabary}', "");
Expect(1, 67647, '\p{Is_Blk=Cypriot_Syllabary}', "");
Expect(0, 67647, '\p{^Is_Blk=Cypriot_Syllabary}', "");
Expect(0, 67647, '\P{Is_Blk=Cypriot_Syllabary}', "");
Expect(1, 67647, '\P{^Is_Blk=Cypriot_Syllabary}', "");
Expect(0, 67648, '\p{Is_Blk=Cypriot_Syllabary}', "");
Expect(1, 67648, '\p{^Is_Blk=Cypriot_Syllabary}', "");
Expect(1, 67648, '\P{Is_Blk=Cypriot_Syllabary}', "");
Expect(0, 67648, '\P{^Is_Blk=Cypriot_Syllabary}', "");
Error('\p{Block:   :=_coptic}');
Error('\P{Block:   :=_coptic}');
Expect(1, 11519, '\p{Block=coptic}', "");
Expect(0, 11519, '\p{^Block=coptic}', "");
Expect(0, 11519, '\P{Block=coptic}', "");
Expect(1, 11519, '\P{^Block=coptic}', "");
Expect(0, 11520, '\p{Block=coptic}', "");
Expect(1, 11520, '\p{^Block=coptic}', "");
Expect(1, 11520, '\P{Block=coptic}', "");
Expect(0, 11520, '\P{^Block=coptic}', "");
Expect(1, 11519, '\p{Block=- coptic}', "");
Expect(0, 11519, '\p{^Block=- coptic}', "");
Expect(0, 11519, '\P{Block=- coptic}', "");
Expect(1, 11519, '\P{^Block=- coptic}', "");
Expect(0, 11520, '\p{Block=- coptic}', "");
Expect(1, 11520, '\p{^Block=- coptic}', "");
Expect(1, 11520, '\P{Block=- coptic}', "");
Expect(0, 11520, '\P{^Block=- coptic}', "");
Error('\p{Blk=	coptic:=}');
Error('\P{Blk=	coptic:=}');
Expect(1, 11519, '\p{Blk:coptic}', "");
Expect(0, 11519, '\p{^Blk:coptic}', "");
Expect(0, 11519, '\P{Blk:coptic}', "");
Expect(1, 11519, '\P{^Blk:coptic}', "");
Expect(0, 11520, '\p{Blk:coptic}', "");
Expect(1, 11520, '\p{^Blk:coptic}', "");
Expect(1, 11520, '\P{Blk:coptic}', "");
Expect(0, 11520, '\P{^Blk:coptic}', "");
Expect(1, 11519, '\p{Blk=  Coptic}', "");
Expect(0, 11519, '\p{^Blk=  Coptic}', "");
Expect(0, 11519, '\P{Blk=  Coptic}', "");
Expect(1, 11519, '\P{^Blk=  Coptic}', "");
Expect(0, 11520, '\p{Blk=  Coptic}', "");
Expect(1, 11520, '\p{^Blk=  Coptic}', "");
Expect(1, 11520, '\P{Blk=  Coptic}', "");
Expect(0, 11520, '\P{^Blk=  Coptic}', "");
Error('\p{Is_Block=		Coptic/a/}');
Error('\P{Is_Block=		Coptic/a/}');
Expect(1, 11519, '\p{Is_Block:   coptic}', "");
Expect(0, 11519, '\p{^Is_Block:   coptic}', "");
Expect(0, 11519, '\P{Is_Block:   coptic}', "");
Expect(1, 11519, '\P{^Is_Block:   coptic}', "");
Expect(0, 11520, '\p{Is_Block:   coptic}', "");
Expect(1, 11520, '\p{^Is_Block:   coptic}', "");
Expect(1, 11520, '\P{Is_Block:   coptic}', "");
Expect(0, 11520, '\P{^Is_Block:   coptic}', "");
Expect(1, 11519, '\p{Is_Block=_	coptic}', "");
Expect(0, 11519, '\p{^Is_Block=_	coptic}', "");
Expect(0, 11519, '\P{Is_Block=_	coptic}', "");
Expect(1, 11519, '\P{^Is_Block=_	coptic}', "");
Expect(0, 11520, '\p{Is_Block=_	coptic}', "");
Expect(1, 11520, '\p{^Is_Block=_	coptic}', "");
Expect(1, 11520, '\P{Is_Block=_	coptic}', "");
Expect(0, 11520, '\P{^Is_Block=_	coptic}', "");
Error('\p{Is_Blk=/a/	Coptic}');
Error('\P{Is_Blk=/a/	Coptic}');
Expect(1, 11519, '\p{Is_Blk=coptic}', "");
Expect(0, 11519, '\p{^Is_Blk=coptic}', "");
Expect(0, 11519, '\P{Is_Blk=coptic}', "");
Expect(1, 11519, '\P{^Is_Blk=coptic}', "");
Expect(0, 11520, '\p{Is_Blk=coptic}', "");
Expect(1, 11520, '\p{^Is_Blk=coptic}', "");
Expect(1, 11520, '\P{Is_Blk=coptic}', "");
Expect(0, 11520, '\P{^Is_Blk=coptic}', "");
Expect(1, 11519, '\p{Is_Blk=		Coptic}', "");
Expect(0, 11519, '\p{^Is_Blk=		Coptic}', "");
Expect(0, 11519, '\P{Is_Blk=		Coptic}', "");
Expect(1, 11519, '\P{^Is_Blk=		Coptic}', "");
Expect(0, 11520, '\p{Is_Blk=		Coptic}', "");
Expect(1, 11520, '\p{^Is_Blk=		Coptic}', "");
Expect(1, 11520, '\P{Is_Blk=		Coptic}', "");
Expect(0, 11520, '\P{^Is_Blk=		Coptic}', "");
Error('\p{Block=__SUPPLEMENTAL_ARROWS_B/a/}');
Error('\P{Block=__SUPPLEMENTAL_ARROWS_B/a/}');
Expect(1, 10623, '\p{Block=supplementalarrowsb}', "");
Expect(0, 10623, '\p{^Block=supplementalarrowsb}', "");
Expect(0, 10623, '\P{Block=supplementalarrowsb}', "");
Expect(1, 10623, '\P{^Block=supplementalarrowsb}', "");
Expect(0, 10624, '\p{Block=supplementalarrowsb}', "");
Expect(1, 10624, '\p{^Block=supplementalarrowsb}', "");
Expect(1, 10624, '\P{Block=supplementalarrowsb}', "");
Expect(0, 10624, '\P{^Block=supplementalarrowsb}', "");
Expect(1, 10623, '\p{Block=_-SUPPLEMENTAL_ARROWS_b}', "");
Expect(0, 10623, '\p{^Block=_-SUPPLEMENTAL_ARROWS_b}', "");
Expect(0, 10623, '\P{Block=_-SUPPLEMENTAL_ARROWS_b}', "");
Expect(1, 10623, '\P{^Block=_-SUPPLEMENTAL_ARROWS_b}', "");
Expect(0, 10624, '\p{Block=_-SUPPLEMENTAL_ARROWS_b}', "");
Expect(1, 10624, '\p{^Block=_-SUPPLEMENTAL_ARROWS_b}', "");
Expect(1, 10624, '\P{Block=_-SUPPLEMENTAL_ARROWS_b}', "");
Expect(0, 10624, '\P{^Block=_-SUPPLEMENTAL_ARROWS_b}', "");
Error('\p{Blk=:=-Sup_Arrows_b}');
Error('\P{Blk=:=-Sup_Arrows_b}');
Expect(1, 10623, '\p{Blk=suparrowsb}', "");
Expect(0, 10623, '\p{^Blk=suparrowsb}', "");
Expect(0, 10623, '\P{Blk=suparrowsb}', "");
Expect(1, 10623, '\P{^Blk=suparrowsb}', "");
Expect(0, 10624, '\p{Blk=suparrowsb}', "");
Expect(1, 10624, '\p{^Blk=suparrowsb}', "");
Expect(1, 10624, '\P{Blk=suparrowsb}', "");
Expect(0, 10624, '\P{^Blk=suparrowsb}', "");
Expect(1, 10623, '\p{Blk=	 Sup_arrows_B}', "");
Expect(0, 10623, '\p{^Blk=	 Sup_arrows_B}', "");
Expect(0, 10623, '\P{Blk=	 Sup_arrows_B}', "");
Expect(1, 10623, '\P{^Blk=	 Sup_arrows_B}', "");
Expect(0, 10624, '\p{Blk=	 Sup_arrows_B}', "");
Expect(1, 10624, '\p{^Blk=	 Sup_arrows_B}', "");
Expect(1, 10624, '\P{Blk=	 Sup_arrows_B}', "");
Expect(0, 10624, '\P{^Blk=	 Sup_arrows_B}', "");
Error('\p{Is_Block=/a/-	SUPPLEMENTAL_Arrows_B}');
Error('\P{Is_Block=/a/-	SUPPLEMENTAL_Arrows_B}');
Expect(1, 10623, '\p{Is_Block=supplementalarrowsb}', "");
Expect(0, 10623, '\p{^Is_Block=supplementalarrowsb}', "");
Expect(0, 10623, '\P{Is_Block=supplementalarrowsb}', "");
Expect(1, 10623, '\P{^Is_Block=supplementalarrowsb}', "");
Expect(0, 10624, '\p{Is_Block=supplementalarrowsb}', "");
Expect(1, 10624, '\p{^Is_Block=supplementalarrowsb}', "");
Expect(1, 10624, '\P{Is_Block=supplementalarrowsb}', "");
Expect(0, 10624, '\P{^Is_Block=supplementalarrowsb}', "");
Expect(1, 10623, '\p{Is_Block:		Supplemental_ARROWS_B}', "");
Expect(0, 10623, '\p{^Is_Block:		Supplemental_ARROWS_B}', "");
Expect(0, 10623, '\P{Is_Block:		Supplemental_ARROWS_B}', "");
Expect(1, 10623, '\P{^Is_Block:		Supplemental_ARROWS_B}', "");
Expect(0, 10624, '\p{Is_Block:		Supplemental_ARROWS_B}', "");
Expect(1, 10624, '\p{^Is_Block:		Supplemental_ARROWS_B}', "");
Expect(1, 10624, '\P{Is_Block:		Supplemental_ARROWS_B}', "");
Expect(0, 10624, '\P{^Is_Block:		Supplemental_ARROWS_B}', "");
Error('\p{Is_Blk:	:=SUP_Arrows_b}');
Error('\P{Is_Blk:	:=SUP_Arrows_b}');
Expect(1, 10623, '\p{Is_Blk=suparrowsb}', "");
Expect(0, 10623, '\p{^Is_Blk=suparrowsb}', "");
Expect(0, 10623, '\P{Is_Blk=suparrowsb}', "");
Expect(1, 10623, '\P{^Is_Blk=suparrowsb}', "");
Expect(0, 10624, '\p{Is_Blk=suparrowsb}', "");
Expect(1, 10624, '\p{^Is_Blk=suparrowsb}', "");
Expect(1, 10624, '\P{Is_Blk=suparrowsb}', "");
Expect(0, 10624, '\P{^Is_Blk=suparrowsb}', "");
Expect(1, 10623, '\p{Is_Blk=_Sup_arrows_B}', "");
Expect(0, 10623, '\p{^Is_Blk=_Sup_arrows_B}', "");
Expect(0, 10623, '\P{Is_Blk=_Sup_arrows_B}', "");
Expect(1, 10623, '\P{^Is_Blk=_Sup_arrows_B}', "");
Expect(0, 10624, '\p{Is_Blk=_Sup_arrows_B}', "");
Expect(1, 10624, '\p{^Is_Blk=_Sup_arrows_B}', "");
Expect(1, 10624, '\P{Is_Blk=_Sup_arrows_B}', "");
Expect(0, 10624, '\P{^Is_Blk=_Sup_arrows_B}', "");
Error('\p{Block: -/a/Superscripts_AND_subscripts}');
Error('\P{Block: -/a/Superscripts_AND_subscripts}');
Expect(1, 8351, '\p{Block=superscriptsandsubscripts}', "");
Expect(0, 8351, '\p{^Block=superscriptsandsubscripts}', "");
Expect(0, 8351, '\P{Block=superscriptsandsubscripts}', "");
Expect(1, 8351, '\P{^Block=superscriptsandsubscripts}', "");
Expect(0, 8352, '\p{Block=superscriptsandsubscripts}', "");
Expect(1, 8352, '\p{^Block=superscriptsandsubscripts}', "");
Expect(1, 8352, '\P{Block=superscriptsandsubscripts}', "");
Expect(0, 8352, '\P{^Block=superscriptsandsubscripts}', "");
Expect(1, 8351, '\p{Block= _SUPERSCRIPTS_and_SUBSCRIPTS}', "");
Expect(0, 8351, '\p{^Block= _SUPERSCRIPTS_and_SUBSCRIPTS}', "");
Expect(0, 8351, '\P{Block= _SUPERSCRIPTS_and_SUBSCRIPTS}', "");
Expect(1, 8351, '\P{^Block= _SUPERSCRIPTS_and_SUBSCRIPTS}', "");
Expect(0, 8352, '\p{Block= _SUPERSCRIPTS_and_SUBSCRIPTS}', "");
Expect(1, 8352, '\p{^Block= _SUPERSCRIPTS_and_SUBSCRIPTS}', "");
Expect(1, 8352, '\P{Block= _SUPERSCRIPTS_and_SUBSCRIPTS}', "");
Expect(0, 8352, '\P{^Block= _SUPERSCRIPTS_and_SUBSCRIPTS}', "");
Error('\p{Blk=_SUPER_and_sub:=}');
Error('\P{Blk=_SUPER_and_sub:=}');
Expect(1, 8351, '\p{Blk=superandsub}', "");
Expect(0, 8351, '\p{^Blk=superandsub}', "");
Expect(0, 8351, '\P{Blk=superandsub}', "");
Expect(1, 8351, '\P{^Blk=superandsub}', "");
Expect(0, 8352, '\p{Blk=superandsub}', "");
Expect(1, 8352, '\p{^Blk=superandsub}', "");
Expect(1, 8352, '\P{Blk=superandsub}', "");
Expect(0, 8352, '\P{^Blk=superandsub}', "");
Expect(1, 8351, '\p{Blk=_ super_And_SUB}', "");
Expect(0, 8351, '\p{^Blk=_ super_And_SUB}', "");
Expect(0, 8351, '\P{Blk=_ super_And_SUB}', "");
Expect(1, 8351, '\P{^Blk=_ super_And_SUB}', "");
Expect(0, 8352, '\p{Blk=_ super_And_SUB}', "");
Expect(1, 8352, '\p{^Blk=_ super_And_SUB}', "");
Expect(1, 8352, '\P{Blk=_ super_And_SUB}', "");
Expect(0, 8352, '\P{^Blk=_ super_And_SUB}', "");
Error('\p{Is_Block=:=SUPERSCRIPTS_And_Subscripts}');
Error('\P{Is_Block=:=SUPERSCRIPTS_And_Subscripts}');
Expect(1, 8351, '\p{Is_Block=superscriptsandsubscripts}', "");
Expect(0, 8351, '\p{^Is_Block=superscriptsandsubscripts}', "");
Expect(0, 8351, '\P{Is_Block=superscriptsandsubscripts}', "");
Expect(1, 8351, '\P{^Is_Block=superscriptsandsubscripts}', "");
Expect(0, 8352, '\p{Is_Block=superscriptsandsubscripts}', "");
Expect(1, 8352, '\p{^Is_Block=superscriptsandsubscripts}', "");
Expect(1, 8352, '\P{Is_Block=superscriptsandsubscripts}', "");
Expect(0, 8352, '\P{^Is_Block=superscriptsandsubscripts}', "");
Expect(1, 8351, '\p{Is_Block=-_Superscripts_And_Subscripts}', "");
Expect(0, 8351, '\p{^Is_Block=-_Superscripts_And_Subscripts}', "");
Expect(0, 8351, '\P{Is_Block=-_Superscripts_And_Subscripts}', "");
Expect(1, 8351, '\P{^Is_Block=-_Superscripts_And_Subscripts}', "");
Expect(0, 8352, '\p{Is_Block=-_Superscripts_And_Subscripts}', "");
Expect(1, 8352, '\p{^Is_Block=-_Superscripts_And_Subscripts}', "");
Expect(1, 8352, '\P{Is_Block=-_Superscripts_And_Subscripts}', "");
Expect(0, 8352, '\P{^Is_Block=-_Superscripts_And_Subscripts}', "");
Error('\p{Is_Blk=-/a/super_And_Sub}');
Error('\P{Is_Blk=-/a/super_And_Sub}');
Expect(1, 8351, '\p{Is_Blk=superandsub}', "");
Expect(0, 8351, '\p{^Is_Blk=superandsub}', "");
Expect(0, 8351, '\P{Is_Blk=superandsub}', "");
Expect(1, 8351, '\P{^Is_Blk=superandsub}', "");
Expect(0, 8352, '\p{Is_Blk=superandsub}', "");
Expect(1, 8352, '\p{^Is_Blk=superandsub}', "");
Expect(1, 8352, '\P{Is_Blk=superandsub}', "");
Expect(0, 8352, '\P{^Is_Blk=superandsub}', "");
Expect(1, 8351, '\p{Is_Blk=_	SUPER_AND_Sub}', "");
Expect(0, 8351, '\p{^Is_Blk=_	SUPER_AND_Sub}', "");
Expect(0, 8351, '\P{Is_Blk=_	SUPER_AND_Sub}', "");
Expect(1, 8351, '\P{^Is_Blk=_	SUPER_AND_Sub}', "");
Expect(0, 8352, '\p{Is_Blk=_	SUPER_AND_Sub}', "");
Expect(1, 8352, '\p{^Is_Blk=_	SUPER_AND_Sub}', "");
Expect(1, 8352, '\P{Is_Blk=_	SUPER_AND_Sub}', "");
Expect(0, 8352, '\P{^Is_Blk=_	SUPER_AND_Sub}', "");
Error('\p{Block: /a/PHAISTOS_Disc}');
Error('\P{Block: /a/PHAISTOS_Disc}');
Expect(1, 66047, '\p{Block=phaistosdisc}', "");
Expect(0, 66047, '\p{^Block=phaistosdisc}', "");
Expect(0, 66047, '\P{Block=phaistosdisc}', "");
Expect(1, 66047, '\P{^Block=phaistosdisc}', "");
Expect(0, 66048, '\p{Block=phaistosdisc}', "");
Expect(1, 66048, '\p{^Block=phaistosdisc}', "");
Expect(1, 66048, '\P{Block=phaistosdisc}', "");
Expect(0, 66048, '\P{^Block=phaistosdisc}', "");
Expect(1, 66047, '\p{Block: -Phaistos_Disc}', "");
Expect(0, 66047, '\p{^Block: -Phaistos_Disc}', "");
Expect(0, 66047, '\P{Block: -Phaistos_Disc}', "");
Expect(1, 66047, '\P{^Block: -Phaistos_Disc}', "");
Expect(0, 66048, '\p{Block: -Phaistos_Disc}', "");
Expect(1, 66048, '\p{^Block: -Phaistos_Disc}', "");
Expect(1, 66048, '\P{Block: -Phaistos_Disc}', "");
Expect(0, 66048, '\P{^Block: -Phaistos_Disc}', "");
Error('\p{Blk=	/a/Phaistos}');
Error('\P{Blk=	/a/Phaistos}');
Expect(1, 66047, '\p{Blk=phaistos}', "");
Expect(0, 66047, '\p{^Blk=phaistos}', "");
Expect(0, 66047, '\P{Blk=phaistos}', "");
Expect(1, 66047, '\P{^Blk=phaistos}', "");
Expect(0, 66048, '\p{Blk=phaistos}', "");
Expect(1, 66048, '\p{^Blk=phaistos}', "");
Expect(1, 66048, '\P{Blk=phaistos}', "");
Expect(0, 66048, '\P{^Blk=phaistos}', "");
Error('\p{Is_Block=/a/ 	PHAISTOS_DISC}');
Error('\P{Is_Block=/a/ 	PHAISTOS_DISC}');
Expect(1, 66047, '\p{Is_Block=phaistosdisc}', "");
Expect(0, 66047, '\p{^Is_Block=phaistosdisc}', "");
Expect(0, 66047, '\P{Is_Block=phaistosdisc}', "");
Expect(1, 66047, '\P{^Is_Block=phaistosdisc}', "");
Expect(0, 66048, '\p{Is_Block=phaistosdisc}', "");
Expect(1, 66048, '\p{^Is_Block=phaistosdisc}', "");
Expect(1, 66048, '\P{Is_Block=phaistosdisc}', "");
Expect(0, 66048, '\P{^Is_Block=phaistosdisc}', "");
Expect(1, 66047, '\p{Is_Block=_-PHAISTOS_DISC}', "");
Expect(0, 66047, '\p{^Is_Block=_-PHAISTOS_DISC}', "");
Expect(0, 66047, '\P{Is_Block=_-PHAISTOS_DISC}', "");
Expect(1, 66047, '\P{^Is_Block=_-PHAISTOS_DISC}', "");
Expect(0, 66048, '\p{Is_Block=_-PHAISTOS_DISC}', "");
Expect(1, 66048, '\p{^Is_Block=_-PHAISTOS_DISC}', "");
Expect(1, 66048, '\P{Is_Block=_-PHAISTOS_DISC}', "");
Expect(0, 66048, '\P{^Is_Block=_-PHAISTOS_DISC}', "");
Error('\p{Is_Blk=_/a/PHAISTOS}');
Error('\P{Is_Blk=_/a/PHAISTOS}');
Expect(1, 66047, '\p{Is_Blk=phaistos}', "");
Expect(0, 66047, '\p{^Is_Blk=phaistos}', "");
Expect(0, 66047, '\P{Is_Blk=phaistos}', "");
Expect(1, 66047, '\P{^Is_Blk=phaistos}', "");
Expect(0, 66048, '\p{Is_Blk=phaistos}', "");
Expect(1, 66048, '\p{^Is_Blk=phaistos}', "");
Expect(1, 66048, '\P{Is_Blk=phaistos}', "");
Expect(0, 66048, '\P{^Is_Blk=phaistos}', "");
Expect(1, 66047, '\p{Is_Blk=--Phaistos}', "");
Expect(0, 66047, '\p{^Is_Blk=--Phaistos}', "");
Expect(0, 66047, '\P{Is_Blk=--Phaistos}', "");
Expect(1, 66047, '\P{^Is_Blk=--Phaistos}', "");
Expect(0, 66048, '\p{Is_Blk=--Phaistos}', "");
Expect(1, 66048, '\p{^Is_Blk=--Phaistos}', "");
Expect(1, 66048, '\P{Is_Blk=--Phaistos}', "");
Expect(0, 66048, '\P{^Is_Blk=--Phaistos}', "");
Error('\p{Block=/a/	 phonetic_Extensions}');
Error('\P{Block=/a/	 phonetic_Extensions}');
Expect(1, 7551, '\p{Block=phoneticextensions}', "");
Expect(0, 7551, '\p{^Block=phoneticextensions}', "");
Expect(0, 7551, '\P{Block=phoneticextensions}', "");
Expect(1, 7551, '\P{^Block=phoneticextensions}', "");
Expect(0, 7552, '\p{Block=phoneticextensions}', "");
Expect(1, 7552, '\p{^Block=phoneticextensions}', "");
Expect(1, 7552, '\P{Block=phoneticextensions}', "");
Expect(0, 7552, '\P{^Block=phoneticextensions}', "");
Expect(1, 7551, '\p{Block=- Phonetic_Extensions}', "");
Expect(0, 7551, '\p{^Block=- Phonetic_Extensions}', "");
Expect(0, 7551, '\P{Block=- Phonetic_Extensions}', "");
Expect(1, 7551, '\P{^Block=- Phonetic_Extensions}', "");
Expect(0, 7552, '\p{Block=- Phonetic_Extensions}', "");
Expect(1, 7552, '\p{^Block=- Phonetic_Extensions}', "");
Expect(1, 7552, '\P{Block=- Phonetic_Extensions}', "");
Expect(0, 7552, '\P{^Block=- Phonetic_Extensions}', "");
Error('\p{Blk=:= 	Phonetic_Ext}');
Error('\P{Blk=:= 	Phonetic_Ext}');
Expect(1, 7551, '\p{Blk=phoneticext}', "");
Expect(0, 7551, '\p{^Blk=phoneticext}', "");
Expect(0, 7551, '\P{Blk=phoneticext}', "");
Expect(1, 7551, '\P{^Blk=phoneticext}', "");
Expect(0, 7552, '\p{Blk=phoneticext}', "");
Expect(1, 7552, '\p{^Blk=phoneticext}', "");
Expect(1, 7552, '\P{Blk=phoneticext}', "");
Expect(0, 7552, '\P{^Blk=phoneticext}', "");
Expect(1, 7551, '\p{Blk:_Phonetic_Ext}', "");
Expect(0, 7551, '\p{^Blk:_Phonetic_Ext}', "");
Expect(0, 7551, '\P{Blk:_Phonetic_Ext}', "");
Expect(1, 7551, '\P{^Blk:_Phonetic_Ext}', "");
Expect(0, 7552, '\p{Blk:_Phonetic_Ext}', "");
Expect(1, 7552, '\p{^Blk:_Phonetic_Ext}', "");
Expect(1, 7552, '\P{Blk:_Phonetic_Ext}', "");
Expect(0, 7552, '\P{^Blk:_Phonetic_Ext}', "");
Error('\p{Is_Block= 	Phonetic_EXTENSIONS/a/}');
Error('\P{Is_Block= 	Phonetic_EXTENSIONS/a/}');
Expect(1, 7551, '\p{Is_Block=phoneticextensions}', "");
Expect(0, 7551, '\p{^Is_Block=phoneticextensions}', "");
Expect(0, 7551, '\P{Is_Block=phoneticextensions}', "");
Expect(1, 7551, '\P{^Is_Block=phoneticextensions}', "");
Expect(0, 7552, '\p{Is_Block=phoneticextensions}', "");
Expect(1, 7552, '\p{^Is_Block=phoneticextensions}', "");
Expect(1, 7552, '\P{Is_Block=phoneticextensions}', "");
Expect(0, 7552, '\P{^Is_Block=phoneticextensions}', "");
Expect(1, 7551, '\p{Is_Block= Phonetic_Extensions}', "");
Expect(0, 7551, '\p{^Is_Block= Phonetic_Extensions}', "");
Expect(0, 7551, '\P{Is_Block= Phonetic_Extensions}', "");
Expect(1, 7551, '\P{^Is_Block= Phonetic_Extensions}', "");
Expect(0, 7552, '\p{Is_Block= Phonetic_Extensions}', "");
Expect(1, 7552, '\p{^Is_Block= Phonetic_Extensions}', "");
Expect(1, 7552, '\P{Is_Block= Phonetic_Extensions}', "");
Expect(0, 7552, '\P{^Is_Block= Phonetic_Extensions}', "");
Error('\p{Is_Blk=  Phonetic_Ext:=}');
Error('\P{Is_Blk=  Phonetic_Ext:=}');
Expect(1, 7551, '\p{Is_Blk=phoneticext}', "");
Expect(0, 7551, '\p{^Is_Blk=phoneticext}', "");
Expect(0, 7551, '\P{Is_Blk=phoneticext}', "");
Expect(1, 7551, '\P{^Is_Blk=phoneticext}', "");
Expect(0, 7552, '\p{Is_Blk=phoneticext}', "");
Expect(1, 7552, '\p{^Is_Blk=phoneticext}', "");
Expect(1, 7552, '\P{Is_Blk=phoneticext}', "");
Expect(0, 7552, '\P{^Is_Blk=phoneticext}', "");
Expect(1, 7551, '\p{Is_Blk=	Phonetic_Ext}', "");
Expect(0, 7551, '\p{^Is_Blk=	Phonetic_Ext}', "");
Expect(0, 7551, '\P{Is_Blk=	Phonetic_Ext}', "");
Expect(1, 7551, '\P{^Is_Blk=	Phonetic_Ext}', "");
Expect(0, 7552, '\p{Is_Blk=	Phonetic_Ext}', "");
Expect(1, 7552, '\p{^Is_Blk=	Phonetic_Ext}', "");
Expect(1, 7552, '\P{Is_Blk=	Phonetic_Ext}', "");
Expect(0, 7552, '\P{^Is_Blk=	Phonetic_Ext}', "");
Error('\p{Block=_-CJK_unified_IDEOGRAPHS_extension_C/a/}');
Error('\P{Block=_-CJK_unified_IDEOGRAPHS_extension_C/a/}');
Expect(1, 177983, '\p{Block:	cjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\p{^Block:	cjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\P{Block:	cjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\P{^Block:	cjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\p{Block:	cjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\p{^Block:	cjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\P{Block:	cjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\P{^Block:	cjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\p{Block=	CJK_Unified_IDEOGRAPHS_extension_C}', "");
Expect(0, 177983, '\p{^Block=	CJK_Unified_IDEOGRAPHS_extension_C}', "");
Expect(0, 177983, '\P{Block=	CJK_Unified_IDEOGRAPHS_extension_C}', "");
Expect(1, 177983, '\P{^Block=	CJK_Unified_IDEOGRAPHS_extension_C}', "");
Expect(0, 177984, '\p{Block=	CJK_Unified_IDEOGRAPHS_extension_C}', "");
Expect(1, 177984, '\p{^Block=	CJK_Unified_IDEOGRAPHS_extension_C}', "");
Expect(1, 177984, '\P{Block=	CJK_Unified_IDEOGRAPHS_extension_C}', "");
Expect(0, 177984, '\P{^Block=	CJK_Unified_IDEOGRAPHS_extension_C}', "");
Error('\p{Blk=:=	CJK_Ext_C}');
Error('\P{Blk=:=	CJK_Ext_C}');
Expect(1, 177983, '\p{Blk=cjkextc}', "");
Expect(0, 177983, '\p{^Blk=cjkextc}', "");
Expect(0, 177983, '\P{Blk=cjkextc}', "");
Expect(1, 177983, '\P{^Blk=cjkextc}', "");
Expect(0, 177984, '\p{Blk=cjkextc}', "");
Expect(1, 177984, '\p{^Blk=cjkextc}', "");
Expect(1, 177984, '\P{Blk=cjkextc}', "");
Expect(0, 177984, '\P{^Blk=cjkextc}', "");
Expect(1, 177983, '\p{Blk=-	CJK_Ext_c}', "");
Expect(0, 177983, '\p{^Blk=-	CJK_Ext_c}', "");
Expect(0, 177983, '\P{Blk=-	CJK_Ext_c}', "");
Expect(1, 177983, '\P{^Blk=-	CJK_Ext_c}', "");
Expect(0, 177984, '\p{Blk=-	CJK_Ext_c}', "");
Expect(1, 177984, '\p{^Blk=-	CJK_Ext_c}', "");
Expect(1, 177984, '\P{Blk=-	CJK_Ext_c}', "");
Expect(0, 177984, '\P{^Blk=-	CJK_Ext_c}', "");
Error('\p{Is_Block=-:=cjk_UNIFIED_IDEOGRAPHS_EXTENSION_c}');
Error('\P{Is_Block=-:=cjk_UNIFIED_IDEOGRAPHS_EXTENSION_c}');
Expect(1, 177983, '\p{Is_Block=cjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\p{^Is_Block=cjkunifiedideographsextensionc}', "");
Expect(0, 177983, '\P{Is_Block=cjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\P{^Is_Block=cjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\p{Is_Block=cjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\p{^Is_Block=cjkunifiedideographsextensionc}', "");
Expect(1, 177984, '\P{Is_Block=cjkunifiedideographsextensionc}', "");
Expect(0, 177984, '\P{^Is_Block=cjkunifiedideographsextensionc}', "");
Expect(1, 177983, '\p{Is_Block=__cjk_Unified_IDEOGRAPHS_Extension_c}', "");
Expect(0, 177983, '\p{^Is_Block=__cjk_Unified_IDEOGRAPHS_Extension_c}', "");
Expect(0, 177983, '\P{Is_Block=__cjk_Unified_IDEOGRAPHS_Extension_c}', "");
Expect(1, 177983, '\P{^Is_Block=__cjk_Unified_IDEOGRAPHS_Extension_c}', "");
Expect(0, 177984, '\p{Is_Block=__cjk_Unified_IDEOGRAPHS_Extension_c}', "");
Expect(1, 177984, '\p{^Is_Block=__cjk_Unified_IDEOGRAPHS_Extension_c}', "");
Expect(1, 177984, '\P{Is_Block=__cjk_Unified_IDEOGRAPHS_Extension_c}', "");
Expect(0, 177984, '\P{^Is_Block=__cjk_Unified_IDEOGRAPHS_Extension_c}', "");
Error('\p{Is_Blk=_CJK_ext_C:=}');
Error('\P{Is_Blk=_CJK_ext_C:=}');
Expect(1, 177983, '\p{Is_Blk:   cjkextc}', "");
Expect(0, 177983, '\p{^Is_Blk:   cjkextc}', "");
Expect(0, 177983, '\P{Is_Blk:   cjkextc}', "");
Expect(1, 177983, '\P{^Is_Blk:   cjkextc}', "");
Expect(0, 177984, '\p{Is_Blk:   cjkextc}', "");
Expect(1, 177984, '\p{^Is_Blk:   cjkextc}', "");
Expect(1, 177984, '\P{Is_Blk:   cjkextc}', "");
Expect(0, 177984, '\P{^Is_Blk:   cjkextc}', "");
Expect(1, 177983, '\p{Is_Blk:    	cjk_Ext_c}', "");
Expect(0, 177983, '\p{^Is_Blk:    	cjk_Ext_c}', "");
Expect(0, 177983, '\P{Is_Blk:    	cjk_Ext_c}', "");
Expect(1, 177983, '\P{^Is_Blk:    	cjk_Ext_c}', "");
Expect(0, 177984, '\p{Is_Blk:    	cjk_Ext_c}', "");
Expect(1, 177984, '\p{^Is_Blk:    	cjk_Ext_c}', "");
Expect(1, 177984, '\P{Is_Blk:    	cjk_Ext_c}', "");
Expect(0, 177984, '\P{^Is_Blk:    	cjk_Ext_c}', "");
Error('\p{Block=:=_TIBETAN}');
Error('\P{Block=:=_TIBETAN}');
Expect(1, 4095, '\p{Block=tibetan}', "");
Expect(0, 4095, '\p{^Block=tibetan}', "");
Expect(0, 4095, '\P{Block=tibetan}', "");
Expect(1, 4095, '\P{^Block=tibetan}', "");
Expect(0, 4096, '\p{Block=tibetan}', "");
Expect(1, 4096, '\p{^Block=tibetan}', "");
Expect(1, 4096, '\P{Block=tibetan}', "");
Expect(0, 4096, '\P{^Block=tibetan}', "");
Expect(1, 4095, '\p{Block=		tibetan}', "");
Expect(0, 4095, '\p{^Block=		tibetan}', "");
Expect(0, 4095, '\P{Block=		tibetan}', "");
Expect(1, 4095, '\P{^Block=		tibetan}', "");
Expect(0, 4096, '\p{Block=		tibetan}', "");
Expect(1, 4096, '\p{^Block=		tibetan}', "");
Expect(1, 4096, '\P{Block=		tibetan}', "");
Expect(0, 4096, '\P{^Block=		tibetan}', "");
Error('\p{Blk=-:=TIBETAN}');
Error('\P{Blk=-:=TIBETAN}');
Expect(1, 4095, '\p{Blk: tibetan}', "");
Expect(0, 4095, '\p{^Blk: tibetan}', "");
Expect(0, 4095, '\P{Blk: tibetan}', "");
Expect(1, 4095, '\P{^Blk: tibetan}', "");
Expect(0, 4096, '\p{Blk: tibetan}', "");
Expect(1, 4096, '\p{^Blk: tibetan}', "");
Expect(1, 4096, '\P{Blk: tibetan}', "");
Expect(0, 4096, '\P{^Blk: tibetan}', "");
Expect(1, 4095, '\p{Blk=	 tibetan}', "");
Expect(0, 4095, '\p{^Blk=	 tibetan}', "");
Expect(0, 4095, '\P{Blk=	 tibetan}', "");
Expect(1, 4095, '\P{^Blk=	 tibetan}', "");
Expect(0, 4096, '\p{Blk=	 tibetan}', "");
Expect(1, 4096, '\p{^Blk=	 tibetan}', "");
Expect(1, 4096, '\P{Blk=	 tibetan}', "");
Expect(0, 4096, '\P{^Blk=	 tibetan}', "");
Error('\p{Is_Block=-	TIBETAN:=}');
Error('\P{Is_Block=-	TIBETAN:=}');
Expect(1, 4095, '\p{Is_Block=tibetan}', "");
Expect(0, 4095, '\p{^Is_Block=tibetan}', "");
Expect(0, 4095, '\P{Is_Block=tibetan}', "");
Expect(1, 4095, '\P{^Is_Block=tibetan}', "");
Expect(0, 4096, '\p{Is_Block=tibetan}', "");
Expect(1, 4096, '\p{^Is_Block=tibetan}', "");
Expect(1, 4096, '\P{Is_Block=tibetan}', "");
Expect(0, 4096, '\P{^Is_Block=tibetan}', "");
Expect(1, 4095, '\p{Is_Block=- Tibetan}', "");
Expect(0, 4095, '\p{^Is_Block=- Tibetan}', "");
Expect(0, 4095, '\P{Is_Block=- Tibetan}', "");
Expect(1, 4095, '\P{^Is_Block=- Tibetan}', "");
Expect(0, 4096, '\p{Is_Block=- Tibetan}', "");
Expect(1, 4096, '\p{^Is_Block=- Tibetan}', "");
Expect(1, 4096, '\P{Is_Block=- Tibetan}', "");
Expect(0, 4096, '\P{^Is_Block=- Tibetan}', "");
Error('\p{Is_Blk=/a/TIBETAN}');
Error('\P{Is_Blk=/a/TIBETAN}');
Expect(1, 4095, '\p{Is_Blk=tibetan}', "");
Expect(0, 4095, '\p{^Is_Blk=tibetan}', "");
Expect(0, 4095, '\P{Is_Blk=tibetan}', "");
Expect(1, 4095, '\P{^Is_Blk=tibetan}', "");
Expect(0, 4096, '\p{Is_Blk=tibetan}', "");
Expect(1, 4096, '\p{^Is_Blk=tibetan}', "");
Expect(1, 4096, '\P{Is_Blk=tibetan}', "");
Expect(0, 4096, '\P{^Is_Blk=tibetan}', "");
Expect(1, 4095, '\p{Is_Blk=		TIBETAN}', "");
Expect(0, 4095, '\p{^Is_Blk=		TIBETAN}', "");
Expect(0, 4095, '\P{Is_Blk=		TIBETAN}', "");
Expect(1, 4095, '\P{^Is_Blk=		TIBETAN}', "");
Expect(0, 4096, '\p{Is_Blk=		TIBETAN}', "");
Expect(1, 4096, '\p{^Is_Blk=		TIBETAN}', "");
Expect(1, 4096, '\P{Is_Blk=		TIBETAN}', "");
Expect(0, 4096, '\P{^Is_Blk=		TIBETAN}', "");
Error('\p{Block:   /a/		Domino_TILES}');
Error('\P{Block:   /a/		Domino_TILES}');
Expect(1, 127135, '\p{Block=dominotiles}', "");
Expect(0, 127135, '\p{^Block=dominotiles}', "");
Expect(0, 127135, '\P{Block=dominotiles}', "");
Expect(1, 127135, '\P{^Block=dominotiles}', "");
Expect(0, 127136, '\p{Block=dominotiles}', "");
Expect(1, 127136, '\p{^Block=dominotiles}', "");
Expect(1, 127136, '\P{Block=dominotiles}', "");
Expect(0, 127136, '\P{^Block=dominotiles}', "");
Expect(1, 127135, '\p{Block=-Domino_Tiles}', "");
Expect(0, 127135, '\p{^Block=-Domino_Tiles}', "");
Expect(0, 127135, '\P{Block=-Domino_Tiles}', "");
Expect(1, 127135, '\P{^Block=-Domino_Tiles}', "");
Expect(0, 127136, '\p{Block=-Domino_Tiles}', "");
Expect(1, 127136, '\p{^Block=-Domino_Tiles}', "");
Expect(1, 127136, '\P{Block=-Domino_Tiles}', "");
Expect(0, 127136, '\P{^Block=-Domino_Tiles}', "");
Error('\p{Blk=	/a/Domino}');
Error('\P{Blk=	/a/Domino}');
Expect(1, 127135, '\p{Blk=domino}', "");
Expect(0, 127135, '\p{^Blk=domino}', "");
Expect(0, 127135, '\P{Blk=domino}', "");
Expect(1, 127135, '\P{^Blk=domino}', "");
Expect(0, 127136, '\p{Blk=domino}', "");
Expect(1, 127136, '\p{^Blk=domino}', "");
Expect(1, 127136, '\P{Blk=domino}', "");
Expect(0, 127136, '\P{^Blk=domino}', "");
Expect(1, 127135, '\p{Blk:	- Domino}', "");
Expect(0, 127135, '\p{^Blk:	- Domino}', "");
Expect(0, 127135, '\P{Blk:	- Domino}', "");
Expect(1, 127135, '\P{^Blk:	- Domino}', "");
Expect(0, 127136, '\p{Blk:	- Domino}', "");
Expect(1, 127136, '\p{^Blk:	- Domino}', "");
Expect(1, 127136, '\P{Blk:	- Domino}', "");
Expect(0, 127136, '\P{^Blk:	- Domino}', "");
Error('\p{Is_Block=:=Domino_tiles}');
Error('\P{Is_Block=:=Domino_tiles}');
Expect(1, 127135, '\p{Is_Block=dominotiles}', "");
Expect(0, 127135, '\p{^Is_Block=dominotiles}', "");
Expect(0, 127135, '\P{Is_Block=dominotiles}', "");
Expect(1, 127135, '\P{^Is_Block=dominotiles}', "");
Expect(0, 127136, '\p{Is_Block=dominotiles}', "");
Expect(1, 127136, '\p{^Is_Block=dominotiles}', "");
Expect(1, 127136, '\P{Is_Block=dominotiles}', "");
Expect(0, 127136, '\P{^Is_Block=dominotiles}', "");
Expect(1, 127135, '\p{Is_Block=_ DOMINO_Tiles}', "");
Expect(0, 127135, '\p{^Is_Block=_ DOMINO_Tiles}', "");
Expect(0, 127135, '\P{Is_Block=_ DOMINO_Tiles}', "");
Expect(1, 127135, '\P{^Is_Block=_ DOMINO_Tiles}', "");
Expect(0, 127136, '\p{Is_Block=_ DOMINO_Tiles}', "");
Expect(1, 127136, '\p{^Is_Block=_ DOMINO_Tiles}', "");
Expect(1, 127136, '\P{Is_Block=_ DOMINO_Tiles}', "");
Expect(0, 127136, '\P{^Is_Block=_ DOMINO_Tiles}', "");
Error('\p{Is_Blk=:=	_Domino}');
Error('\P{Is_Blk=:=	_Domino}');
Expect(1, 127135, '\p{Is_Blk=domino}', "");
Expect(0, 127135, '\p{^Is_Blk=domino}', "");
Expect(0, 127135, '\P{Is_Blk=domino}', "");
Expect(1, 127135, '\P{^Is_Blk=domino}', "");
Expect(0, 127136, '\p{Is_Blk=domino}', "");
Expect(1, 127136, '\p{^Is_Blk=domino}', "");
Expect(1, 127136, '\P{Is_Blk=domino}', "");
Expect(0, 127136, '\P{^Is_Blk=domino}', "");
Expect(1, 127135, '\p{Is_Blk= 	domino}', "");
Expect(0, 127135, '\p{^Is_Blk= 	domino}', "");
Expect(0, 127135, '\P{Is_Blk= 	domino}', "");
Expect(1, 127135, '\P{^Is_Blk= 	domino}', "");
Expect(0, 127136, '\p{Is_Blk= 	domino}', "");
Expect(1, 127136, '\p{^Is_Blk= 	domino}', "");
Expect(1, 127136, '\P{Is_Blk= 	domino}', "");
Expect(0, 127136, '\P{^Is_Blk= 	domino}', "");
Error('\p{Block=/a/Bopomofo_EXTENDED}');
Error('\P{Block=/a/Bopomofo_EXTENDED}');
Expect(1, 12735, '\p{Block=bopomofoextended}', "");
Expect(0, 12735, '\p{^Block=bopomofoextended}', "");
Expect(0, 12735, '\P{Block=bopomofoextended}', "");
Expect(1, 12735, '\P{^Block=bopomofoextended}', "");
Expect(0, 12736, '\p{Block=bopomofoextended}', "");
Expect(1, 12736, '\p{^Block=bopomofoextended}', "");
Expect(1, 12736, '\P{Block=bopomofoextended}', "");
Expect(0, 12736, '\P{^Block=bopomofoextended}', "");
Expect(1, 12735, '\p{Block=-BOPOMOFO_Extended}', "");
Expect(0, 12735, '\p{^Block=-BOPOMOFO_Extended}', "");
Expect(0, 12735, '\P{Block=-BOPOMOFO_Extended}', "");
Expect(1, 12735, '\P{^Block=-BOPOMOFO_Extended}', "");
Expect(0, 12736, '\p{Block=-BOPOMOFO_Extended}', "");
Expect(1, 12736, '\p{^Block=-BOPOMOFO_Extended}', "");
Expect(1, 12736, '\P{Block=-BOPOMOFO_Extended}', "");
Expect(0, 12736, '\P{^Block=-BOPOMOFO_Extended}', "");
Error('\p{Blk=--Bopomofo_Ext/a/}');
Error('\P{Blk=--Bopomofo_Ext/a/}');
Expect(1, 12735, '\p{Blk=bopomofoext}', "");
Expect(0, 12735, '\p{^Blk=bopomofoext}', "");
Expect(0, 12735, '\P{Blk=bopomofoext}', "");
Expect(1, 12735, '\P{^Blk=bopomofoext}', "");
Expect(0, 12736, '\p{Blk=bopomofoext}', "");
Expect(1, 12736, '\p{^Blk=bopomofoext}', "");
Expect(1, 12736, '\P{Blk=bopomofoext}', "");
Expect(0, 12736, '\P{^Blk=bopomofoext}', "");
Expect(1, 12735, '\p{Blk=_	BOPOMOFO_ext}', "");
Expect(0, 12735, '\p{^Blk=_	BOPOMOFO_ext}', "");
Expect(0, 12735, '\P{Blk=_	BOPOMOFO_ext}', "");
Expect(1, 12735, '\P{^Blk=_	BOPOMOFO_ext}', "");
Expect(0, 12736, '\p{Blk=_	BOPOMOFO_ext}', "");
Expect(1, 12736, '\p{^Blk=_	BOPOMOFO_ext}', "");
Expect(1, 12736, '\P{Blk=_	BOPOMOFO_ext}', "");
Expect(0, 12736, '\P{^Blk=_	BOPOMOFO_ext}', "");
Error('\p{Is_Block=:=_bopomofo_Extended}');
Error('\P{Is_Block=:=_bopomofo_Extended}');
Expect(1, 12735, '\p{Is_Block=bopomofoextended}', "");
Expect(0, 12735, '\p{^Is_Block=bopomofoextended}', "");
Expect(0, 12735, '\P{Is_Block=bopomofoextended}', "");
Expect(1, 12735, '\P{^Is_Block=bopomofoextended}', "");
Expect(0, 12736, '\p{Is_Block=bopomofoextended}', "");
Expect(1, 12736, '\p{^Is_Block=bopomofoextended}', "");
Expect(1, 12736, '\P{Is_Block=bopomofoextended}', "");
Expect(0, 12736, '\P{^Is_Block=bopomofoextended}', "");
Expect(1, 12735, '\p{Is_Block:_	Bopomofo_EXTENDED}', "");
Expect(0, 12735, '\p{^Is_Block:_	Bopomofo_EXTENDED}', "");
Expect(0, 12735, '\P{Is_Block:_	Bopomofo_EXTENDED}', "");
Expect(1, 12735, '\P{^Is_Block:_	Bopomofo_EXTENDED}', "");
Expect(0, 12736, '\p{Is_Block:_	Bopomofo_EXTENDED}', "");
Expect(1, 12736, '\p{^Is_Block:_	Bopomofo_EXTENDED}', "");
Expect(1, 12736, '\P{Is_Block:_	Bopomofo_EXTENDED}', "");
Expect(0, 12736, '\P{^Is_Block:_	Bopomofo_EXTENDED}', "");
Error('\p{Is_Blk=_/a/Bopomofo_ext}');
Error('\P{Is_Blk=_/a/Bopomofo_ext}');
Expect(1, 12735, '\p{Is_Blk=bopomofoext}', "");
Expect(0, 12735, '\p{^Is_Blk=bopomofoext}', "");
Expect(0, 12735, '\P{Is_Blk=bopomofoext}', "");
Expect(1, 12735, '\P{^Is_Blk=bopomofoext}', "");
Expect(0, 12736, '\p{Is_Blk=bopomofoext}', "");
Expect(1, 12736, '\p{^Is_Blk=bopomofoext}', "");
Expect(1, 12736, '\P{Is_Blk=bopomofoext}', "");
Expect(0, 12736, '\P{^Is_Blk=bopomofoext}', "");
Expect(1, 12735, '\p{Is_Blk=_	BOPOMOFO_EXT}', "");
Expect(0, 12735, '\p{^Is_Blk=_	BOPOMOFO_EXT}', "");
Expect(0, 12735, '\P{Is_Blk=_	BOPOMOFO_EXT}', "");
Expect(1, 12735, '\P{^Is_Blk=_	BOPOMOFO_EXT}', "");
Expect(0, 12736, '\p{Is_Blk=_	BOPOMOFO_EXT}', "");
Expect(1, 12736, '\p{^Is_Blk=_	BOPOMOFO_EXT}', "");
Expect(1, 12736, '\P{Is_Blk=_	BOPOMOFO_EXT}', "");
Expect(0, 12736, '\P{^Is_Blk=_	BOPOMOFO_EXT}', "");
Error('\p{Block=/a/	 supplemental_ARROWS_A}');
Error('\P{Block=/a/	 supplemental_ARROWS_A}');
Expect(1, 10239, '\p{Block=supplementalarrowsa}', "");
Expect(0, 10239, '\p{^Block=supplementalarrowsa}', "");
Expect(0, 10239, '\P{Block=supplementalarrowsa}', "");
Expect(1, 10239, '\P{^Block=supplementalarrowsa}', "");
Expect(0, 10240, '\p{Block=supplementalarrowsa}', "");
Expect(1, 10240, '\p{^Block=supplementalarrowsa}', "");
Expect(1, 10240, '\P{Block=supplementalarrowsa}', "");
Expect(0, 10240, '\P{^Block=supplementalarrowsa}', "");
Expect(1, 10239, '\p{Block= 	Supplemental_ARROWS_A}', "");
Expect(0, 10239, '\p{^Block= 	Supplemental_ARROWS_A}', "");
Expect(0, 10239, '\P{Block= 	Supplemental_ARROWS_A}', "");
Expect(1, 10239, '\P{^Block= 	Supplemental_ARROWS_A}', "");
Expect(0, 10240, '\p{Block= 	Supplemental_ARROWS_A}', "");
Expect(1, 10240, '\p{^Block= 	Supplemental_ARROWS_A}', "");
Expect(1, 10240, '\P{Block= 	Supplemental_ARROWS_A}', "");
Expect(0, 10240, '\P{^Block= 	Supplemental_ARROWS_A}', "");
Error('\p{Blk=-:=sup_arrows_A}');
Error('\P{Blk=-:=sup_arrows_A}');
Expect(1, 10239, '\p{Blk=suparrowsa}', "");
Expect(0, 10239, '\p{^Blk=suparrowsa}', "");
Expect(0, 10239, '\P{Blk=suparrowsa}', "");
Expect(1, 10239, '\P{^Blk=suparrowsa}', "");
Expect(0, 10240, '\p{Blk=suparrowsa}', "");
Expect(1, 10240, '\p{^Blk=suparrowsa}', "");
Expect(1, 10240, '\P{Blk=suparrowsa}', "");
Expect(0, 10240, '\P{^Blk=suparrowsa}', "");
Expect(1, 10239, '\p{Blk=		SUP_Arrows_A}', "");
Expect(0, 10239, '\p{^Blk=		SUP_Arrows_A}', "");
Expect(0, 10239, '\P{Blk=		SUP_Arrows_A}', "");
Expect(1, 10239, '\P{^Blk=		SUP_Arrows_A}', "");
Expect(0, 10240, '\p{Blk=		SUP_Arrows_A}', "");
Expect(1, 10240, '\p{^Blk=		SUP_Arrows_A}', "");
Expect(1, 10240, '\P{Blk=		SUP_Arrows_A}', "");
Expect(0, 10240, '\P{^Blk=		SUP_Arrows_A}', "");
Error('\p{Is_Block:  :=supplemental_Arrows_a}');
Error('\P{Is_Block:  :=supplemental_Arrows_a}');
Expect(1, 10239, '\p{Is_Block=supplementalarrowsa}', "");
Expect(0, 10239, '\p{^Is_Block=supplementalarrowsa}', "");
Expect(0, 10239, '\P{Is_Block=supplementalarrowsa}', "");
Expect(1, 10239, '\P{^Is_Block=supplementalarrowsa}', "");
Expect(0, 10240, '\p{Is_Block=supplementalarrowsa}', "");
Expect(1, 10240, '\p{^Is_Block=supplementalarrowsa}', "");
Expect(1, 10240, '\P{Is_Block=supplementalarrowsa}', "");
Expect(0, 10240, '\P{^Is_Block=supplementalarrowsa}', "");
Expect(1, 10239, '\p{Is_Block:   __Supplemental_Arrows_A}', "");
Expect(0, 10239, '\p{^Is_Block:   __Supplemental_Arrows_A}', "");
Expect(0, 10239, '\P{Is_Block:   __Supplemental_Arrows_A}', "");
Expect(1, 10239, '\P{^Is_Block:   __Supplemental_Arrows_A}', "");
Expect(0, 10240, '\p{Is_Block:   __Supplemental_Arrows_A}', "");
Expect(1, 10240, '\p{^Is_Block:   __Supplemental_Arrows_A}', "");
Expect(1, 10240, '\P{Is_Block:   __Supplemental_Arrows_A}', "");
Expect(0, 10240, '\P{^Is_Block:   __Supplemental_Arrows_A}', "");
Error('\p{Is_Blk=:=		Sup_Arrows_A}');
Error('\P{Is_Blk=:=		Sup_Arrows_A}');
Expect(1, 10239, '\p{Is_Blk=suparrowsa}', "");
Expect(0, 10239, '\p{^Is_Blk=suparrowsa}', "");
Expect(0, 10239, '\P{Is_Blk=suparrowsa}', "");
Expect(1, 10239, '\P{^Is_Blk=suparrowsa}', "");
Expect(0, 10240, '\p{Is_Blk=suparrowsa}', "");
Expect(1, 10240, '\p{^Is_Blk=suparrowsa}', "");
Expect(1, 10240, '\P{Is_Blk=suparrowsa}', "");
Expect(0, 10240, '\P{^Is_Blk=suparrowsa}', "");
Expect(1, 10239, '\p{Is_Blk:	__Sup_Arrows_A}', "");
Expect(0, 10239, '\p{^Is_Blk:	__Sup_Arrows_A}', "");
Expect(0, 10239, '\P{Is_Blk:	__Sup_Arrows_A}', "");
Expect(1, 10239, '\P{^Is_Blk:	__Sup_Arrows_A}', "");
Expect(0, 10240, '\p{Is_Blk:	__Sup_Arrows_A}', "");
Expect(1, 10240, '\p{^Is_Blk:	__Sup_Arrows_A}', "");
Expect(1, 10240, '\P{Is_Blk:	__Sup_Arrows_A}', "");
Expect(0, 10240, '\P{^Is_Blk:	__Sup_Arrows_A}', "");
Error('\p{Block=:=Combining_Diacritical_MARKS_for_SYMBOLS}');
Error('\P{Block=:=Combining_Diacritical_MARKS_for_SYMBOLS}');
Expect(1, 8447, '\p{Block=combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\p{^Block=combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\P{Block=combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\P{^Block=combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\p{Block=combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\p{^Block=combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\P{Block=combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\P{^Block=combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\p{Block=-Combining_Diacritical_Marks_For_Symbols}', "");
Expect(0, 8447, '\p{^Block=-Combining_Diacritical_Marks_For_Symbols}', "");
Expect(0, 8447, '\P{Block=-Combining_Diacritical_Marks_For_Symbols}', "");
Expect(1, 8447, '\P{^Block=-Combining_Diacritical_Marks_For_Symbols}', "");
Expect(0, 8448, '\p{Block=-Combining_Diacritical_Marks_For_Symbols}', "");
Expect(1, 8448, '\p{^Block=-Combining_Diacritical_Marks_For_Symbols}', "");
Expect(1, 8448, '\P{Block=-Combining_Diacritical_Marks_For_Symbols}', "");
Expect(0, 8448, '\P{^Block=-Combining_Diacritical_Marks_For_Symbols}', "");
Error('\p{Blk=	/a/diacriticals_FOR_SYMBOLS}');
Error('\P{Blk=	/a/diacriticals_FOR_SYMBOLS}');
Expect(1, 8447, '\p{Blk=diacriticalsforsymbols}', "");
Expect(0, 8447, '\p{^Blk=diacriticalsforsymbols}', "");
Expect(0, 8447, '\P{Blk=diacriticalsforsymbols}', "");
Expect(1, 8447, '\P{^Blk=diacriticalsforsymbols}', "");
Expect(0, 8448, '\p{Blk=diacriticalsforsymbols}', "");
Expect(1, 8448, '\p{^Blk=diacriticalsforsymbols}', "");
Expect(1, 8448, '\P{Blk=diacriticalsforsymbols}', "");
Expect(0, 8448, '\P{^Blk=diacriticalsforsymbols}', "");
Expect(1, 8447, '\p{Blk=--Diacriticals_For_Symbols}', "");
Expect(0, 8447, '\p{^Blk=--Diacriticals_For_Symbols}', "");
Expect(0, 8447, '\P{Blk=--Diacriticals_For_Symbols}', "");
Expect(1, 8447, '\P{^Blk=--Diacriticals_For_Symbols}', "");
Expect(0, 8448, '\p{Blk=--Diacriticals_For_Symbols}', "");
Expect(1, 8448, '\p{^Blk=--Diacriticals_For_Symbols}', "");
Expect(1, 8448, '\P{Blk=--Diacriticals_For_Symbols}', "");
Expect(0, 8448, '\P{^Blk=--Diacriticals_For_Symbols}', "");
Error('\p{Is_Block= /a/COMBINING_Marks_For_Symbols}');
Error('\P{Is_Block= /a/COMBINING_Marks_For_Symbols}');
Expect(1, 8447, '\p{Is_Block=combiningmarksforsymbols}', "");
Expect(0, 8447, '\p{^Is_Block=combiningmarksforsymbols}', "");
Expect(0, 8447, '\P{Is_Block=combiningmarksforsymbols}', "");
Expect(1, 8447, '\P{^Is_Block=combiningmarksforsymbols}', "");
Expect(0, 8448, '\p{Is_Block=combiningmarksforsymbols}', "");
Expect(1, 8448, '\p{^Is_Block=combiningmarksforsymbols}', "");
Expect(1, 8448, '\P{Is_Block=combiningmarksforsymbols}', "");
Expect(0, 8448, '\P{^Is_Block=combiningmarksforsymbols}', "");
Expect(1, 8447, '\p{Is_Block=-	COMBINING_Marks_for_Symbols}', "");
Expect(0, 8447, '\p{^Is_Block=-	COMBINING_Marks_for_Symbols}', "");
Expect(0, 8447, '\P{Is_Block=-	COMBINING_Marks_for_Symbols}', "");
Expect(1, 8447, '\P{^Is_Block=-	COMBINING_Marks_for_Symbols}', "");
Expect(0, 8448, '\p{Is_Block=-	COMBINING_Marks_for_Symbols}', "");
Expect(1, 8448, '\p{^Is_Block=-	COMBINING_Marks_for_Symbols}', "");
Expect(1, 8448, '\P{Is_Block=-	COMBINING_Marks_for_Symbols}', "");
Expect(0, 8448, '\P{^Is_Block=-	COMBINING_Marks_for_Symbols}', "");
Error('\p{Is_Blk=--COMBINING_Diacritical_MARKS_for_SYMBOLS/a/}');
Error('\P{Is_Blk=--COMBINING_Diacritical_MARKS_for_SYMBOLS/a/}');
Expect(1, 8447, '\p{Is_Blk:	combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\p{^Is_Blk:	combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8447, '\P{Is_Blk:	combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\P{^Is_Blk:	combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\p{Is_Blk:	combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\p{^Is_Blk:	combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8448, '\P{Is_Blk:	combiningdiacriticalmarksforsymbols}', "");
Expect(0, 8448, '\P{^Is_Blk:	combiningdiacriticalmarksforsymbols}', "");
Expect(1, 8447, '\p{Is_Blk=	_Combining_diacritical_Marks_For_symbols}', "");
Expect(0, 8447, '\p{^Is_Blk=	_Combining_diacritical_Marks_For_symbols}', "");
Expect(0, 8447, '\P{Is_Blk=	_Combining_diacritical_Marks_For_symbols}', "");
Expect(1, 8447, '\P{^Is_Blk=	_Combining_diacritical_Marks_For_symbols}', "");
Expect(0, 8448, '\p{Is_Blk=	_Combining_diacritical_Marks_For_symbols}', "");
Expect(1, 8448, '\p{^Is_Blk=	_Combining_diacritical_Marks_For_symbols}', "");
Expect(1, 8448, '\P{Is_Blk=	_Combining_diacritical_Marks_For_symbols}', "");
Expect(0, 8448, '\P{^Is_Blk=	_Combining_diacritical_Marks_For_symbols}', "");
Error('\p{Block=		HANGUL_compatibility_Jamo/a/}');
Error('\P{Block=		HANGUL_compatibility_Jamo/a/}');
Expect(1, 12687, '\p{Block=hangulcompatibilityjamo}', "");
Expect(0, 12687, '\p{^Block=hangulcompatibilityjamo}', "");
Expect(0, 12687, '\P{Block=hangulcompatibilityjamo}', "");
Expect(1, 12687, '\P{^Block=hangulcompatibilityjamo}', "");
Expect(0, 12688, '\p{Block=hangulcompatibilityjamo}', "");
Expect(1, 12688, '\p{^Block=hangulcompatibilityjamo}', "");
Expect(1, 12688, '\P{Block=hangulcompatibilityjamo}', "");
Expect(0, 12688, '\P{^Block=hangulcompatibilityjamo}', "");
Expect(1, 12687, '\p{Block=	-Hangul_COMPATIBILITY_jamo}', "");
Expect(0, 12687, '\p{^Block=	-Hangul_COMPATIBILITY_jamo}', "");
Expect(0, 12687, '\P{Block=	-Hangul_COMPATIBILITY_jamo}', "");
Expect(1, 12687, '\P{^Block=	-Hangul_COMPATIBILITY_jamo}', "");
Expect(0, 12688, '\p{Block=	-Hangul_COMPATIBILITY_jamo}', "");
Expect(1, 12688, '\p{^Block=	-Hangul_COMPATIBILITY_jamo}', "");
Expect(1, 12688, '\P{Block=	-Hangul_COMPATIBILITY_jamo}', "");
Expect(0, 12688, '\P{^Block=	-Hangul_COMPATIBILITY_jamo}', "");
Error('\p{Blk= :=COMPAT_JAMO}');
Error('\P{Blk= :=COMPAT_JAMO}');
Expect(1, 12687, '\p{Blk=compatjamo}', "");
Expect(0, 12687, '\p{^Blk=compatjamo}', "");
Expect(0, 12687, '\P{Blk=compatjamo}', "");
Expect(1, 12687, '\P{^Blk=compatjamo}', "");
Expect(0, 12688, '\p{Blk=compatjamo}', "");
Expect(1, 12688, '\p{^Blk=compatjamo}', "");
Expect(1, 12688, '\P{Blk=compatjamo}', "");
Expect(0, 12688, '\P{^Blk=compatjamo}', "");
Expect(1, 12687, '\p{Blk=-Compat_Jamo}', "");
Expect(0, 12687, '\p{^Blk=-Compat_Jamo}', "");
Expect(0, 12687, '\P{Blk=-Compat_Jamo}', "");
Expect(1, 12687, '\P{^Blk=-Compat_Jamo}', "");
Expect(0, 12688, '\p{Blk=-Compat_Jamo}', "");
Expect(1, 12688, '\p{^Blk=-Compat_Jamo}', "");
Expect(1, 12688, '\P{Blk=-Compat_Jamo}', "");
Expect(0, 12688, '\P{^Blk=-Compat_Jamo}', "");
Error('\p{Is_Block=_ Hangul_COMPATIBILITY_JAMO:=}');
Error('\P{Is_Block=_ Hangul_COMPATIBILITY_JAMO:=}');
Expect(1, 12687, '\p{Is_Block=hangulcompatibilityjamo}', "");
Expect(0, 12687, '\p{^Is_Block=hangulcompatibilityjamo}', "");
Expect(0, 12687, '\P{Is_Block=hangulcompatibilityjamo}', "");
Expect(1, 12687, '\P{^Is_Block=hangulcompatibilityjamo}', "");
Expect(0, 12688, '\p{Is_Block=hangulcompatibilityjamo}', "");
Expect(1, 12688, '\p{^Is_Block=hangulcompatibilityjamo}', "");
Expect(1, 12688, '\P{Is_Block=hangulcompatibilityjamo}', "");
Expect(0, 12688, '\P{^Is_Block=hangulcompatibilityjamo}', "");
Expect(1, 12687, '\p{Is_Block= 	Hangul_Compatibility_jamo}', "");
Expect(0, 12687, '\p{^Is_Block= 	Hangul_Compatibility_jamo}', "");
Expect(0, 12687, '\P{Is_Block= 	Hangul_Compatibility_jamo}', "");
Expect(1, 12687, '\P{^Is_Block= 	Hangul_Compatibility_jamo}', "");
Expect(0, 12688, '\p{Is_Block= 	Hangul_Compatibility_jamo}', "");
Expect(1, 12688, '\p{^Is_Block= 	Hangul_Compatibility_jamo}', "");
Expect(1, 12688, '\P{Is_Block= 	Hangul_Compatibility_jamo}', "");
Expect(0, 12688, '\P{^Is_Block= 	Hangul_Compatibility_jamo}', "");
Error('\p{Is_Blk=_Compat_jamo:=}');
Error('\P{Is_Blk=_Compat_jamo:=}');
Expect(1, 12687, '\p{Is_Blk=compatjamo}', "");
Expect(0, 12687, '\p{^Is_Blk=compatjamo}', "");
Expect(0, 12687, '\P{Is_Blk=compatjamo}', "");
Expect(1, 12687, '\P{^Is_Blk=compatjamo}', "");
Expect(0, 12688, '\p{Is_Blk=compatjamo}', "");
Expect(1, 12688, '\p{^Is_Blk=compatjamo}', "");
Expect(1, 12688, '\P{Is_Blk=compatjamo}', "");
Expect(0, 12688, '\P{^Is_Blk=compatjamo}', "");
Expect(1, 12687, '\p{Is_Blk: COMPAT_Jamo}', "");
Expect(0, 12687, '\p{^Is_Blk: COMPAT_Jamo}', "");
Expect(0, 12687, '\P{Is_Blk: COMPAT_Jamo}', "");
Expect(1, 12687, '\P{^Is_Blk: COMPAT_Jamo}', "");
Expect(0, 12688, '\p{Is_Blk: COMPAT_Jamo}', "");
Expect(1, 12688, '\p{^Is_Blk: COMPAT_Jamo}', "");
Expect(1, 12688, '\P{Is_Blk: COMPAT_Jamo}', "");
Expect(0, 12688, '\P{^Is_Blk: COMPAT_Jamo}', "");
Error('\p{Block=:=Sundanese_supplement}');
Error('\P{Block=:=Sundanese_supplement}');
Expect(1, 7375, '\p{Block=sundanesesupplement}', "");
Expect(0, 7375, '\p{^Block=sundanesesupplement}', "");
Expect(0, 7375, '\P{Block=sundanesesupplement}', "");
Expect(1, 7375, '\P{^Block=sundanesesupplement}', "");
Expect(0, 7376, '\p{Block=sundanesesupplement}', "");
Expect(1, 7376, '\p{^Block=sundanesesupplement}', "");
Expect(1, 7376, '\P{Block=sundanesesupplement}', "");
Expect(0, 7376, '\P{^Block=sundanesesupplement}', "");
Expect(1, 7375, '\p{Block:   	 Sundanese_Supplement}', "");
Expect(0, 7375, '\p{^Block:   	 Sundanese_Supplement}', "");
Expect(0, 7375, '\P{Block:   	 Sundanese_Supplement}', "");
Expect(1, 7375, '\P{^Block:   	 Sundanese_Supplement}', "");
Expect(0, 7376, '\p{Block:   	 Sundanese_Supplement}', "");
Expect(1, 7376, '\p{^Block:   	 Sundanese_Supplement}', "");
Expect(1, 7376, '\P{Block:   	 Sundanese_Supplement}', "");
Expect(0, 7376, '\P{^Block:   	 Sundanese_Supplement}', "");
Error('\p{Blk=/a/-sundanese_Sup}');
Error('\P{Blk=/a/-sundanese_Sup}');
Expect(1, 7375, '\p{Blk=sundanesesup}', "");
Expect(0, 7375, '\p{^Blk=sundanesesup}', "");
Expect(0, 7375, '\P{Blk=sundanesesup}', "");
Expect(1, 7375, '\P{^Blk=sundanesesup}', "");
Expect(0, 7376, '\p{Blk=sundanesesup}', "");
Expect(1, 7376, '\p{^Blk=sundanesesup}', "");
Expect(1, 7376, '\P{Blk=sundanesesup}', "");
Expect(0, 7376, '\P{^Blk=sundanesesup}', "");
Expect(1, 7375, '\p{Blk= Sundanese_SUP}', "");
Expect(0, 7375, '\p{^Blk= Sundanese_SUP}', "");
Expect(0, 7375, '\P{Blk= Sundanese_SUP}', "");
Expect(1, 7375, '\P{^Blk= Sundanese_SUP}', "");
Expect(0, 7376, '\p{Blk= Sundanese_SUP}', "");
Expect(1, 7376, '\p{^Blk= Sundanese_SUP}', "");
Expect(1, 7376, '\P{Blk= Sundanese_SUP}', "");
Expect(0, 7376, '\P{^Blk= Sundanese_SUP}', "");
Error('\p{Is_Block=/a/- SUNDANESE_Supplement}');
Error('\P{Is_Block=/a/- SUNDANESE_Supplement}');
Expect(1, 7375, '\p{Is_Block=sundanesesupplement}', "");
Expect(0, 7375, '\p{^Is_Block=sundanesesupplement}', "");
Expect(0, 7375, '\P{Is_Block=sundanesesupplement}', "");
Expect(1, 7375, '\P{^Is_Block=sundanesesupplement}', "");
Expect(0, 7376, '\p{Is_Block=sundanesesupplement}', "");
Expect(1, 7376, '\p{^Is_Block=sundanesesupplement}', "");
Expect(1, 7376, '\P{Is_Block=sundanesesupplement}', "");
Expect(0, 7376, '\P{^Is_Block=sundanesesupplement}', "");
Expect(1, 7375, '\p{Is_Block=	 SUNDANESE_Supplement}', "");
Expect(0, 7375, '\p{^Is_Block=	 SUNDANESE_Supplement}', "");
Expect(0, 7375, '\P{Is_Block=	 SUNDANESE_Supplement}', "");
Expect(1, 7375, '\P{^Is_Block=	 SUNDANESE_Supplement}', "");
Expect(0, 7376, '\p{Is_Block=	 SUNDANESE_Supplement}', "");
Expect(1, 7376, '\p{^Is_Block=	 SUNDANESE_Supplement}', "");
Expect(1, 7376, '\P{Is_Block=	 SUNDANESE_Supplement}', "");
Expect(0, 7376, '\P{^Is_Block=	 SUNDANESE_Supplement}', "");
Error('\p{Is_Blk=		SUNDANESE_SUP/a/}');
Error('\P{Is_Blk=		SUNDANESE_SUP/a/}');
Expect(1, 7375, '\p{Is_Blk=sundanesesup}', "");
Expect(0, 7375, '\p{^Is_Blk=sundanesesup}', "");
Expect(0, 7375, '\P{Is_Blk=sundanesesup}', "");
Expect(1, 7375, '\P{^Is_Blk=sundanesesup}', "");
Expect(0, 7376, '\p{Is_Blk=sundanesesup}', "");
Expect(1, 7376, '\p{^Is_Blk=sundanesesup}', "");
Expect(1, 7376, '\P{Is_Blk=sundanesesup}', "");
Expect(0, 7376, '\P{^Is_Blk=sundanesesup}', "");
Expect(1, 7375, '\p{Is_Blk:  _Sundanese_Sup}', "");
Expect(0, 7375, '\p{^Is_Blk:  _Sundanese_Sup}', "");
Expect(0, 7375, '\P{Is_Blk:  _Sundanese_Sup}', "");
Expect(1, 7375, '\P{^Is_Blk:  _Sundanese_Sup}', "");
Expect(0, 7376, '\p{Is_Blk:  _Sundanese_Sup}', "");
Expect(1, 7376, '\p{^Is_Blk:  _Sundanese_Sup}', "");
Expect(1, 7376, '\P{Is_Blk:  _Sundanese_Sup}', "");
Expect(0, 7376, '\P{^Is_Blk:  _Sundanese_Sup}', "");
Error('\p{Block=--arabic_MATHEMATICAL_Alphabetic_Symbols:=}');
Error('\P{Block=--arabic_MATHEMATICAL_Alphabetic_Symbols:=}');
Expect(1, 126719, '\p{Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\p{^Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\P{Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\P{^Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\p{Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\p{^Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\P{Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\P{^Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\p{Block=- arabic_mathematical_Alphabetic_Symbols}', "");
Expect(0, 126719, '\p{^Block=- arabic_mathematical_Alphabetic_Symbols}', "");
Expect(0, 126719, '\P{Block=- arabic_mathematical_Alphabetic_Symbols}', "");
Expect(1, 126719, '\P{^Block=- arabic_mathematical_Alphabetic_Symbols}', "");
Expect(0, 126720, '\p{Block=- arabic_mathematical_Alphabetic_Symbols}', "");
Expect(1, 126720, '\p{^Block=- arabic_mathematical_Alphabetic_Symbols}', "");
Expect(1, 126720, '\P{Block=- arabic_mathematical_Alphabetic_Symbols}', "");
Expect(0, 126720, '\P{^Block=- arabic_mathematical_Alphabetic_Symbols}', "");
Error('\p{Blk=/a/ -arabic_Math}');
Error('\P{Blk=/a/ -arabic_Math}');
Expect(1, 126719, '\p{Blk=arabicmath}', "");
Expect(0, 126719, '\p{^Blk=arabicmath}', "");
Expect(0, 126719, '\P{Blk=arabicmath}', "");
Expect(1, 126719, '\P{^Blk=arabicmath}', "");
Expect(0, 126720, '\p{Blk=arabicmath}', "");
Expect(1, 126720, '\p{^Blk=arabicmath}', "");
Expect(1, 126720, '\P{Blk=arabicmath}', "");
Expect(0, 126720, '\P{^Blk=arabicmath}', "");
Expect(1, 126719, '\p{Blk=-_arabic_math}', "");
Expect(0, 126719, '\p{^Blk=-_arabic_math}', "");
Expect(0, 126719, '\P{Blk=-_arabic_math}', "");
Expect(1, 126719, '\P{^Blk=-_arabic_math}', "");
Expect(0, 126720, '\p{Blk=-_arabic_math}', "");
Expect(1, 126720, '\p{^Blk=-_arabic_math}', "");
Expect(1, 126720, '\P{Blk=-_arabic_math}', "");
Expect(0, 126720, '\P{^Blk=-_arabic_math}', "");
Error('\p{Is_Block=/a/	Arabic_MATHEMATICAL_Alphabetic_SYMBOLS}');
Error('\P{Is_Block=/a/	Arabic_MATHEMATICAL_Alphabetic_SYMBOLS}');
Expect(1, 126719, '\p{Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\p{^Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126719, '\P{Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\P{^Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\p{Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\p{^Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126720, '\P{Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(0, 126720, '\P{^Is_Block=arabicmathematicalalphabeticsymbols}', "");
Expect(1, 126719, '\p{Is_Block=	 Arabic_MATHEMATICAL_Alphabetic_Symbols}', "");
Expect(0, 126719, '\p{^Is_Block=	 Arabic_MATHEMATICAL_Alphabetic_Symbols}', "");
Expect(0, 126719, '\P{Is_Block=	 Arabic_MATHEMATICAL_Alphabetic_Symbols}', "");
Expect(1, 126719, '\P{^Is_Block=	 Arabic_MATHEMATICAL_Alphabetic_Symbols}', "");
Expect(0, 126720, '\p{Is_Block=	 Arabic_MATHEMATICAL_Alphabetic_Symbols}', "");
Expect(1, 126720, '\p{^Is_Block=	 Arabic_MATHEMATICAL_Alphabetic_Symbols}', "");
Expect(1, 126720, '\P{Is_Block=	 Arabic_MATHEMATICAL_Alphabetic_Symbols}', "");
Expect(0, 126720, '\P{^Is_Block=	 Arabic_MATHEMATICAL_Alphabetic_Symbols}', "");
Error('\p{Is_Blk=	:=Arabic_Math}');
Error('\P{Is_Blk=	:=Arabic_Math}');
Expect(1, 126719, '\p{Is_Blk=arabicmath}', "");
Expect(0, 126719, '\p{^Is_Blk=arabicmath}', "");
Expect(0, 126719, '\P{Is_Blk=arabicmath}', "");
Expect(1, 126719, '\P{^Is_Blk=arabicmath}', "");
Expect(0, 126720, '\p{Is_Blk=arabicmath}', "");
Expect(1, 126720, '\p{^Is_Blk=arabicmath}', "");
Expect(1, 126720, '\P{Is_Blk=arabicmath}', "");
Expect(0, 126720, '\P{^Is_Blk=arabicmath}', "");
Expect(1, 126719, '\p{Is_Blk=_-ARABIC_math}', "");
Expect(0, 126719, '\p{^Is_Blk=_-ARABIC_math}', "");
Expect(0, 126719, '\P{Is_Blk=_-ARABIC_math}', "");
Expect(1, 126719, '\P{^Is_Blk=_-ARABIC_math}', "");
Expect(0, 126720, '\p{Is_Blk=_-ARABIC_math}', "");
Expect(1, 126720, '\p{^Is_Blk=_-ARABIC_math}', "");
Expect(1, 126720, '\P{Is_Blk=_-ARABIC_math}', "");
Expect(0, 126720, '\P{^Is_Blk=_-ARABIC_math}', "");
Error('\p{Block: := BOPOMOFO}');
Error('\P{Block: := BOPOMOFO}');
Expect(1, 12591, '\p{Block=bopomofo}', "");
Expect(0, 12591, '\p{^Block=bopomofo}', "");
Expect(0, 12591, '\P{Block=bopomofo}', "");
Expect(1, 12591, '\P{^Block=bopomofo}', "");
Expect(0, 12592, '\p{Block=bopomofo}', "");
Expect(1, 12592, '\p{^Block=bopomofo}', "");
Expect(1, 12592, '\P{Block=bopomofo}', "");
Expect(0, 12592, '\P{^Block=bopomofo}', "");
Expect(1, 12591, '\p{Block=_	Bopomofo}', "");
Expect(0, 12591, '\p{^Block=_	Bopomofo}', "");
Expect(0, 12591, '\P{Block=_	Bopomofo}', "");
Expect(1, 12591, '\P{^Block=_	Bopomofo}', "");
Expect(0, 12592, '\p{Block=_	Bopomofo}', "");
Expect(1, 12592, '\p{^Block=_	Bopomofo}', "");
Expect(1, 12592, '\P{Block=_	Bopomofo}', "");
Expect(0, 12592, '\P{^Block=_	Bopomofo}', "");
Error('\p{Blk=-/a/Bopomofo}');
Error('\P{Blk=-/a/Bopomofo}');
Expect(1, 12591, '\p{Blk=bopomofo}', "");
Expect(0, 12591, '\p{^Blk=bopomofo}', "");
Expect(0, 12591, '\P{Blk=bopomofo}', "");
Expect(1, 12591, '\P{^Blk=bopomofo}', "");
Expect(0, 12592, '\p{Blk=bopomofo}', "");
Expect(1, 12592, '\p{^Blk=bopomofo}', "");
Expect(1, 12592, '\P{Blk=bopomofo}', "");
Expect(0, 12592, '\P{^Blk=bopomofo}', "");
Expect(1, 12591, '\p{Blk=_-Bopomofo}', "");
Expect(0, 12591, '\p{^Blk=_-Bopomofo}', "");
Expect(0, 12591, '\P{Blk=_-Bopomofo}', "");
Expect(1, 12591, '\P{^Blk=_-Bopomofo}', "");
Expect(0, 12592, '\p{Blk=_-Bopomofo}', "");
Expect(1, 12592, '\p{^Blk=_-Bopomofo}', "");
Expect(1, 12592, '\P{Blk=_-Bopomofo}', "");
Expect(0, 12592, '\P{^Blk=_-Bopomofo}', "");
Error('\p{Is_Block:		BOPOMOFO/a/}');
Error('\P{Is_Block:		BOPOMOFO/a/}');
Expect(1, 12591, '\p{Is_Block=bopomofo}', "");
Expect(0, 12591, '\p{^Is_Block=bopomofo}', "");
Expect(0, 12591, '\P{Is_Block=bopomofo}', "");
Expect(1, 12591, '\P{^Is_Block=bopomofo}', "");
Expect(0, 12592, '\p{Is_Block=bopomofo}', "");
Expect(1, 12592, '\p{^Is_Block=bopomofo}', "");
Expect(1, 12592, '\P{Is_Block=bopomofo}', "");
Expect(0, 12592, '\P{^Is_Block=bopomofo}', "");
Expect(1, 12591, '\p{Is_Block=_ bopomofo}', "");
Expect(0, 12591, '\p{^Is_Block=_ bopomofo}', "");
Expect(0, 12591, '\P{Is_Block=_ bopomofo}', "");
Expect(1, 12591, '\P{^Is_Block=_ bopomofo}', "");
Expect(0, 12592, '\p{Is_Block=_ bopomofo}', "");
Expect(1, 12592, '\p{^Is_Block=_ bopomofo}', "");
Expect(1, 12592, '\P{Is_Block=_ bopomofo}', "");
Expect(0, 12592, '\P{^Is_Block=_ bopomofo}', "");
Error('\p{Is_Blk=	bopomofo:=}');
Error('\P{Is_Blk=	bopomofo:=}');
Expect(1, 12591, '\p{Is_Blk=bopomofo}', "");
Expect(0, 12591, '\p{^Is_Blk=bopomofo}', "");
Expect(0, 12591, '\P{Is_Blk=bopomofo}', "");
Expect(1, 12591, '\P{^Is_Blk=bopomofo}', "");
Expect(0, 12592, '\p{Is_Blk=bopomofo}', "");
Expect(1, 12592, '\p{^Is_Blk=bopomofo}', "");
Expect(1, 12592, '\P{Is_Blk=bopomofo}', "");
Expect(0, 12592, '\P{^Is_Blk=bopomofo}', "");
Expect(1, 12591, '\p{Is_Blk= -bopomofo}', "");
Expect(0, 12591, '\p{^Is_Blk= -bopomofo}', "");
Expect(0, 12591, '\P{Is_Blk= -bopomofo}', "");
Expect(1, 12591, '\P{^Is_Blk= -bopomofo}', "");
Expect(0, 12592, '\p{Is_Blk= -bopomofo}', "");
Expect(1, 12592, '\p{^Is_Blk= -bopomofo}', "");
Expect(1, 12592, '\P{Is_Blk= -bopomofo}', "");
Expect(0, 12592, '\P{^Is_Blk= -bopomofo}', "");
Error('\p{Block::=MISCELLANEOUS_Technical}');
Error('\P{Block::=MISCELLANEOUS_Technical}');
Expect(1, 9215, '\p{Block:miscellaneoustechnical}', "");
Expect(0, 9215, '\p{^Block:miscellaneoustechnical}', "");
Expect(0, 9215, '\P{Block:miscellaneoustechnical}', "");
Expect(1, 9215, '\P{^Block:miscellaneoustechnical}', "");
Expect(0, 9216, '\p{Block:miscellaneoustechnical}', "");
Expect(1, 9216, '\p{^Block:miscellaneoustechnical}', "");
Expect(1, 9216, '\P{Block:miscellaneoustechnical}', "");
Expect(0, 9216, '\P{^Block:miscellaneoustechnical}', "");
Expect(1, 9215, '\p{Block=-Miscellaneous_TECHNICAL}', "");
Expect(0, 9215, '\p{^Block=-Miscellaneous_TECHNICAL}', "");
Expect(0, 9215, '\P{Block=-Miscellaneous_TECHNICAL}', "");
Expect(1, 9215, '\P{^Block=-Miscellaneous_TECHNICAL}', "");
Expect(0, 9216, '\p{Block=-Miscellaneous_TECHNICAL}', "");
Expect(1, 9216, '\p{^Block=-Miscellaneous_TECHNICAL}', "");
Expect(1, 9216, '\P{Block=-Miscellaneous_TECHNICAL}', "");
Expect(0, 9216, '\P{^Block=-Miscellaneous_TECHNICAL}', "");
Error('\p{Blk:	:=Misc_Technical}');
Error('\P{Blk:	:=Misc_Technical}');
Expect(1, 9215, '\p{Blk=misctechnical}', "");
Expect(0, 9215, '\p{^Blk=misctechnical}', "");
Expect(0, 9215, '\P{Blk=misctechnical}', "");
Expect(1, 9215, '\P{^Blk=misctechnical}', "");
Expect(0, 9216, '\p{Blk=misctechnical}', "");
Expect(1, 9216, '\p{^Blk=misctechnical}', "");
Expect(1, 9216, '\P{Blk=misctechnical}', "");
Expect(0, 9216, '\P{^Blk=misctechnical}', "");
Expect(1, 9215, '\p{Blk=-misc_TECHNICAL}', "");
Expect(0, 9215, '\p{^Blk=-misc_TECHNICAL}', "");
Expect(0, 9215, '\P{Blk=-misc_TECHNICAL}', "");
Expect(1, 9215, '\P{^Blk=-misc_TECHNICAL}', "");
Expect(0, 9216, '\p{Blk=-misc_TECHNICAL}', "");
Expect(1, 9216, '\p{^Blk=-misc_TECHNICAL}', "");
Expect(1, 9216, '\P{Blk=-misc_TECHNICAL}', "");
Expect(0, 9216, '\P{^Blk=-misc_TECHNICAL}', "");
Error('\p{Is_Block:   /a/Miscellaneous_technical}');
Error('\P{Is_Block:   /a/Miscellaneous_technical}');
Expect(1, 9215, '\p{Is_Block=miscellaneoustechnical}', "");
Expect(0, 9215, '\p{^Is_Block=miscellaneoustechnical}', "");
Expect(0, 9215, '\P{Is_Block=miscellaneoustechnical}', "");
Expect(1, 9215, '\P{^Is_Block=miscellaneoustechnical}', "");
Expect(0, 9216, '\p{Is_Block=miscellaneoustechnical}', "");
Expect(1, 9216, '\p{^Is_Block=miscellaneoustechnical}', "");
Expect(1, 9216, '\P{Is_Block=miscellaneoustechnical}', "");
Expect(0, 9216, '\P{^Is_Block=miscellaneoustechnical}', "");
Expect(1, 9215, '\p{Is_Block= _Miscellaneous_Technical}', "");
Expect(0, 9215, '\p{^Is_Block= _Miscellaneous_Technical}', "");
Expect(0, 9215, '\P{Is_Block= _Miscellaneous_Technical}', "");
Expect(1, 9215, '\P{^Is_Block= _Miscellaneous_Technical}', "");
Expect(0, 9216, '\p{Is_Block= _Miscellaneous_Technical}', "");
Expect(1, 9216, '\p{^Is_Block= _Miscellaneous_Technical}', "");
Expect(1, 9216, '\P{Is_Block= _Miscellaneous_Technical}', "");
Expect(0, 9216, '\P{^Is_Block= _Miscellaneous_Technical}', "");
Error('\p{Is_Blk:   _ Misc_technical:=}');
Error('\P{Is_Blk:   _ Misc_technical:=}');
Expect(1, 9215, '\p{Is_Blk:   misctechnical}', "");
Expect(0, 9215, '\p{^Is_Blk:   misctechnical}', "");
Expect(0, 9215, '\P{Is_Blk:   misctechnical}', "");
Expect(1, 9215, '\P{^Is_Blk:   misctechnical}', "");
Expect(0, 9216, '\p{Is_Blk:   misctechnical}', "");
Expect(1, 9216, '\p{^Is_Blk:   misctechnical}', "");
Expect(1, 9216, '\P{Is_Blk:   misctechnical}', "");
Expect(0, 9216, '\P{^Is_Blk:   misctechnical}', "");
Expect(1, 9215, '\p{Is_Blk:   _misc_technical}', "");
Expect(0, 9215, '\p{^Is_Blk:   _misc_technical}', "");
Expect(0, 9215, '\P{Is_Blk:   _misc_technical}', "");
Expect(1, 9215, '\P{^Is_Blk:   _misc_technical}', "");
Expect(0, 9216, '\p{Is_Blk:   _misc_technical}', "");
Expect(1, 9216, '\p{^Is_Blk:   _misc_technical}', "");
Expect(1, 9216, '\P{Is_Blk:   _misc_technical}', "");
Expect(0, 9216, '\P{^Is_Blk:   _misc_technical}', "");
Error('\p{Block=/a/-_Mahjong_tiles}');
Error('\P{Block=/a/-_Mahjong_tiles}');
Expect(1, 127023, '\p{Block=mahjongtiles}', "");
Expect(0, 127023, '\p{^Block=mahjongtiles}', "");
Expect(0, 127023, '\P{Block=mahjongtiles}', "");
Expect(1, 127023, '\P{^Block=mahjongtiles}', "");
Expect(0, 127024, '\p{Block=mahjongtiles}', "");
Expect(1, 127024, '\p{^Block=mahjongtiles}', "");
Expect(1, 127024, '\P{Block=mahjongtiles}', "");
Expect(0, 127024, '\P{^Block=mahjongtiles}', "");
Expect(1, 127023, '\p{Block:    _Mahjong_TILES}', "");
Expect(0, 127023, '\p{^Block:    _Mahjong_TILES}', "");
Expect(0, 127023, '\P{Block:    _Mahjong_TILES}', "");
Expect(1, 127023, '\P{^Block:    _Mahjong_TILES}', "");
Expect(0, 127024, '\p{Block:    _Mahjong_TILES}', "");
Expect(1, 127024, '\p{^Block:    _Mahjong_TILES}', "");
Expect(1, 127024, '\P{Block:    _Mahjong_TILES}', "");
Expect(0, 127024, '\P{^Block:    _Mahjong_TILES}', "");
Error('\p{Blk=	 mahjong:=}');
Error('\P{Blk=	 mahjong:=}');
Expect(1, 127023, '\p{Blk=mahjong}', "");
Expect(0, 127023, '\p{^Blk=mahjong}', "");
Expect(0, 127023, '\P{Blk=mahjong}', "");
Expect(1, 127023, '\P{^Blk=mahjong}', "");
Expect(0, 127024, '\p{Blk=mahjong}', "");
Expect(1, 127024, '\p{^Blk=mahjong}', "");
Expect(1, 127024, '\P{Blk=mahjong}', "");
Expect(0, 127024, '\P{^Blk=mahjong}', "");
Expect(1, 127023, '\p{Blk= 	MAHJONG}', "");
Expect(0, 127023, '\p{^Blk= 	MAHJONG}', "");
Expect(0, 127023, '\P{Blk= 	MAHJONG}', "");
Expect(1, 127023, '\P{^Blk= 	MAHJONG}', "");
Expect(0, 127024, '\p{Blk= 	MAHJONG}', "");
Expect(1, 127024, '\p{^Blk= 	MAHJONG}', "");
Expect(1, 127024, '\P{Blk= 	MAHJONG}', "");
Expect(0, 127024, '\P{^Blk= 	MAHJONG}', "");
Error('\p{Is_Block:	-	MAHJONG_TILES/a/}');
Error('\P{Is_Block:	-	MAHJONG_TILES/a/}');
Expect(1, 127023, '\p{Is_Block=mahjongtiles}', "");
Expect(0, 127023, '\p{^Is_Block=mahjongtiles}', "");
Expect(0, 127023, '\P{Is_Block=mahjongtiles}', "");
Expect(1, 127023, '\P{^Is_Block=mahjongtiles}', "");
Expect(0, 127024, '\p{Is_Block=mahjongtiles}', "");
Expect(1, 127024, '\p{^Is_Block=mahjongtiles}', "");
Expect(1, 127024, '\P{Is_Block=mahjongtiles}', "");
Expect(0, 127024, '\P{^Is_Block=mahjongtiles}', "");
Expect(1, 127023, '\p{Is_Block=	mahjong_TILES}', "");
Expect(0, 127023, '\p{^Is_Block=	mahjong_TILES}', "");
Expect(0, 127023, '\P{Is_Block=	mahjong_TILES}', "");
Expect(1, 127023, '\P{^Is_Block=	mahjong_TILES}', "");
Expect(0, 127024, '\p{Is_Block=	mahjong_TILES}', "");
Expect(1, 127024, '\p{^Is_Block=	mahjong_TILES}', "");
Expect(1, 127024, '\P{Is_Block=	mahjong_TILES}', "");
Expect(0, 127024, '\P{^Is_Block=	mahjong_TILES}', "");
Error('\p{Is_Blk=/a/mahjong}');
Error('\P{Is_Blk=/a/mahjong}');
Expect(1, 127023, '\p{Is_Blk=mahjong}', "");
Expect(0, 127023, '\p{^Is_Blk=mahjong}', "");
Expect(0, 127023, '\P{Is_Blk=mahjong}', "");
Expect(1, 127023, '\P{^Is_Blk=mahjong}', "");
Expect(0, 127024, '\p{Is_Blk=mahjong}', "");
Expect(1, 127024, '\p{^Is_Blk=mahjong}', "");
Expect(1, 127024, '\P{Is_Blk=mahjong}', "");
Expect(0, 127024, '\P{^Is_Blk=mahjong}', "");
Expect(1, 127023, '\p{Is_Blk=__Mahjong}', "");
Expect(0, 127023, '\p{^Is_Blk=__Mahjong}', "");
Expect(0, 127023, '\P{Is_Blk=__Mahjong}', "");
Expect(1, 127023, '\P{^Is_Blk=__Mahjong}', "");
Expect(0, 127024, '\p{Is_Blk=__Mahjong}', "");
Expect(1, 127024, '\p{^Is_Blk=__Mahjong}', "");
Expect(1, 127024, '\P{Is_Blk=__Mahjong}', "");
Expect(0, 127024, '\P{^Is_Blk=__Mahjong}', "");
Error('\p{Block=	:=block_elements}');
Error('\P{Block=	:=block_elements}');
Expect(1, 9631, '\p{Block=blockelements}', "");
Expect(0, 9631, '\p{^Block=blockelements}', "");
Expect(0, 9631, '\P{Block=blockelements}', "");
Expect(1, 9631, '\P{^Block=blockelements}', "");
Expect(0, 9632, '\p{Block=blockelements}', "");
Expect(1, 9632, '\p{^Block=blockelements}', "");
Expect(1, 9632, '\P{Block=blockelements}', "");
Expect(0, 9632, '\P{^Block=blockelements}', "");
Expect(1, 9631, '\p{Block= -Block_ELEMENTS}', "");
Expect(0, 9631, '\p{^Block= -Block_ELEMENTS}', "");
Expect(0, 9631, '\P{Block= -Block_ELEMENTS}', "");
Expect(1, 9631, '\P{^Block= -Block_ELEMENTS}', "");
Expect(0, 9632, '\p{Block= -Block_ELEMENTS}', "");
Expect(1, 9632, '\p{^Block= -Block_ELEMENTS}', "");
Expect(1, 9632, '\P{Block= -Block_ELEMENTS}', "");
Expect(0, 9632, '\P{^Block= -Block_ELEMENTS}', "");
Error('\p{Blk=_ block_elements:=}');
Error('\P{Blk=_ block_elements:=}');
Expect(1, 9631, '\p{Blk=blockelements}', "");
Expect(0, 9631, '\p{^Blk=blockelements}', "");
Expect(0, 9631, '\P{Blk=blockelements}', "");
Expect(1, 9631, '\P{^Blk=blockelements}', "");
Expect(0, 9632, '\p{Blk=blockelements}', "");
Expect(1, 9632, '\p{^Blk=blockelements}', "");
Expect(1, 9632, '\P{Blk=blockelements}', "");
Expect(0, 9632, '\P{^Blk=blockelements}', "");
Expect(1, 9631, '\p{Blk=-_Block_Elements}', "");
Expect(0, 9631, '\p{^Blk=-_Block_Elements}', "");
Expect(0, 9631, '\P{Blk=-_Block_Elements}', "");
Expect(1, 9631, '\P{^Blk=-_Block_Elements}', "");
Expect(0, 9632, '\p{Blk=-_Block_Elements}', "");
Expect(1, 9632, '\p{^Blk=-_Block_Elements}', "");
Expect(1, 9632, '\P{Blk=-_Block_Elements}', "");
Expect(0, 9632, '\P{^Blk=-_Block_Elements}', "");
Error('\p{Is_Block=-	BLOCK_elements/a/}');
Error('\P{Is_Block=-	BLOCK_elements/a/}');
Expect(1, 9631, '\p{Is_Block=blockelements}', "");
Expect(0, 9631, '\p{^Is_Block=blockelements}', "");
Expect(0, 9631, '\P{Is_Block=blockelements}', "");
Expect(1, 9631, '\P{^Is_Block=blockelements}', "");
Expect(0, 9632, '\p{Is_Block=blockelements}', "");
Expect(1, 9632, '\p{^Is_Block=blockelements}', "");
Expect(1, 9632, '\P{Is_Block=blockelements}', "");
Expect(0, 9632, '\P{^Is_Block=blockelements}', "");
Expect(1, 9631, '\p{Is_Block=_ Block_Elements}', "");
Expect(0, 9631, '\p{^Is_Block=_ Block_Elements}', "");
Expect(0, 9631, '\P{Is_Block=_ Block_Elements}', "");
Expect(1, 9631, '\P{^Is_Block=_ Block_Elements}', "");
Expect(0, 9632, '\p{Is_Block=_ Block_Elements}', "");
Expect(1, 9632, '\p{^Is_Block=_ Block_Elements}', "");
Expect(1, 9632, '\P{Is_Block=_ Block_Elements}', "");
Expect(0, 9632, '\P{^Is_Block=_ Block_Elements}', "");
Error('\p{Is_Blk=-/a/Block_elements}');
Error('\P{Is_Blk=-/a/Block_elements}');
Expect(1, 9631, '\p{Is_Blk=blockelements}', "");
Expect(0, 9631, '\p{^Is_Blk=blockelements}', "");
Expect(0, 9631, '\P{Is_Blk=blockelements}', "");
Expect(1, 9631, '\P{^Is_Blk=blockelements}', "");
Expect(0, 9632, '\p{Is_Blk=blockelements}', "");
Expect(1, 9632, '\p{^Is_Blk=blockelements}', "");
Expect(1, 9632, '\P{Is_Blk=blockelements}', "");
Expect(0, 9632, '\P{^Is_Blk=blockelements}', "");
Expect(1, 9631, '\p{Is_Blk=_ block_Elements}', "");
Expect(0, 9631, '\p{^Is_Blk=_ block_Elements}', "");
Expect(0, 9631, '\P{Is_Blk=_ block_Elements}', "");
Expect(1, 9631, '\P{^Is_Blk=_ block_Elements}', "");
Expect(0, 9632, '\p{Is_Blk=_ block_Elements}', "");
Expect(1, 9632, '\p{^Is_Blk=_ block_Elements}', "");
Expect(1, 9632, '\P{Is_Blk=_ block_Elements}', "");
Expect(0, 9632, '\P{^Is_Blk=_ block_Elements}', "");
Error('\p{Block=-/a/cjk_STROKES}');
Error('\P{Block=-/a/cjk_STROKES}');
Expect(1, 12783, '\p{Block=cjkstrokes}', "");
Expect(0, 12783, '\p{^Block=cjkstrokes}', "");
Expect(0, 12783, '\P{Block=cjkstrokes}', "");
Expect(1, 12783, '\P{^Block=cjkstrokes}', "");
Expect(0, 12784, '\p{Block=cjkstrokes}', "");
Expect(1, 12784, '\p{^Block=cjkstrokes}', "");
Expect(1, 12784, '\P{Block=cjkstrokes}', "");
Expect(0, 12784, '\P{^Block=cjkstrokes}', "");
Expect(1, 12783, '\p{Block= CJK_strokes}', "");
Expect(0, 12783, '\p{^Block= CJK_strokes}', "");
Expect(0, 12783, '\P{Block= CJK_strokes}', "");
Expect(1, 12783, '\P{^Block= CJK_strokes}', "");
Expect(0, 12784, '\p{Block= CJK_strokes}', "");
Expect(1, 12784, '\p{^Block= CJK_strokes}', "");
Expect(1, 12784, '\P{Block= CJK_strokes}', "");
Expect(0, 12784, '\P{^Block= CJK_strokes}', "");
Error('\p{Blk:   /a/_-CJK_Strokes}');
Error('\P{Blk:   /a/_-CJK_Strokes}');
Expect(1, 12783, '\p{Blk=cjkstrokes}', "");
Expect(0, 12783, '\p{^Blk=cjkstrokes}', "");
Expect(0, 12783, '\P{Blk=cjkstrokes}', "");
Expect(1, 12783, '\P{^Blk=cjkstrokes}', "");
Expect(0, 12784, '\p{Blk=cjkstrokes}', "");
Expect(1, 12784, '\p{^Blk=cjkstrokes}', "");
Expect(1, 12784, '\P{Blk=cjkstrokes}', "");
Expect(0, 12784, '\P{^Blk=cjkstrokes}', "");
Expect(1, 12783, '\p{Blk:		cjk_Strokes}', "");
Expect(0, 12783, '\p{^Blk:		cjk_Strokes}', "");
Expect(0, 12783, '\P{Blk:		cjk_Strokes}', "");
Expect(1, 12783, '\P{^Blk:		cjk_Strokes}', "");
Expect(0, 12784, '\p{Blk:		cjk_Strokes}', "");
Expect(1, 12784, '\p{^Blk:		cjk_Strokes}', "");
Expect(1, 12784, '\P{Blk:		cjk_Strokes}', "");
Expect(0, 12784, '\P{^Blk:		cjk_Strokes}', "");
Error('\p{Is_Block= cjk_STROKES:=}');
Error('\P{Is_Block= cjk_STROKES:=}');
Expect(1, 12783, '\p{Is_Block=cjkstrokes}', "");
Expect(0, 12783, '\p{^Is_Block=cjkstrokes}', "");
Expect(0, 12783, '\P{Is_Block=cjkstrokes}', "");
Expect(1, 12783, '\P{^Is_Block=cjkstrokes}', "");
Expect(0, 12784, '\p{Is_Block=cjkstrokes}', "");
Expect(1, 12784, '\p{^Is_Block=cjkstrokes}', "");
Expect(1, 12784, '\P{Is_Block=cjkstrokes}', "");
Expect(0, 12784, '\P{^Is_Block=cjkstrokes}', "");
Expect(1, 12783, '\p{Is_Block:   -CJK_strokes}', "");
Expect(0, 12783, '\p{^Is_Block:   -CJK_strokes}', "");
Expect(0, 12783, '\P{Is_Block:   -CJK_strokes}', "");
Expect(1, 12783, '\P{^Is_Block:   -CJK_strokes}', "");
Expect(0, 12784, '\p{Is_Block:   -CJK_strokes}', "");
Expect(1, 12784, '\p{^Is_Block:   -CJK_strokes}', "");
Expect(1, 12784, '\P{Is_Block:   -CJK_strokes}', "");
Expect(0, 12784, '\P{^Is_Block:   -CJK_strokes}', "");
Error('\p{Is_Blk:	/a/ _CJK_strokes}');
Error('\P{Is_Blk:	/a/ _CJK_strokes}');
Expect(1, 12783, '\p{Is_Blk=cjkstrokes}', "");
Expect(0, 12783, '\p{^Is_Blk=cjkstrokes}', "");
Expect(0, 12783, '\P{Is_Blk=cjkstrokes}', "");
Expect(1, 12783, '\P{^Is_Blk=cjkstrokes}', "");
Expect(0, 12784, '\p{Is_Blk=cjkstrokes}', "");
Expect(1, 12784, '\p{^Is_Blk=cjkstrokes}', "");
Expect(1, 12784, '\P{Is_Blk=cjkstrokes}', "");
Expect(0, 12784, '\P{^Is_Blk=cjkstrokes}', "");
Expect(1, 12783, '\p{Is_Blk=	CJK_Strokes}', "");
Expect(0, 12783, '\p{^Is_Blk=	CJK_Strokes}', "");
Expect(0, 12783, '\P{Is_Blk=	CJK_Strokes}', "");
Expect(1, 12783, '\P{^Is_Blk=	CJK_Strokes}', "");
Expect(0, 12784, '\p{Is_Blk=	CJK_Strokes}', "");
Expect(1, 12784, '\p{^Is_Blk=	CJK_Strokes}', "");
Expect(1, 12784, '\P{Is_Blk=	CJK_Strokes}', "");
Expect(0, 12784, '\P{^Is_Blk=	CJK_Strokes}', "");
Error('\p{Block=:=_No_BLOCK}');
Error('\P{Block=:=_No_BLOCK}');
Expect(1, 918000, '\p{Block=noblock}', "");
Expect(0, 918000, '\p{^Block=noblock}', "");
Expect(0, 918000, '\P{Block=noblock}', "");
Expect(1, 918000, '\P{^Block=noblock}', "");
Expect(0, 983040, '\p{Block=noblock}', "");
Expect(1, 983040, '\p{^Block=noblock}', "");
Expect(1, 983040, '\P{Block=noblock}', "");
Expect(0, 983040, '\P{^Block=noblock}', "");
Expect(1, 918000, '\p{Block=	No_block}', "");
Expect(0, 918000, '\p{^Block=	No_block}', "");
Expect(0, 918000, '\P{Block=	No_block}', "");
Expect(1, 918000, '\P{^Block=	No_block}', "");
Expect(0, 983040, '\p{Block=	No_block}', "");
Expect(1, 983040, '\p{^Block=	No_block}', "");
Expect(1, 983040, '\P{Block=	No_block}', "");
Expect(0, 983040, '\P{^Block=	No_block}', "");
Error('\p{Blk=	/a/NB}');
Error('\P{Blk=	/a/NB}');
Expect(1, 918000, '\p{Blk: nb}', "");
Expect(0, 918000, '\p{^Blk: nb}', "");
Expect(0, 918000, '\P{Blk: nb}', "");
Expect(1, 918000, '\P{^Blk: nb}', "");
Expect(0, 983040, '\p{Blk: nb}', "");
Expect(1, 983040, '\p{^Blk: nb}', "");
Expect(1, 983040, '\P{Blk: nb}', "");
Expect(0, 983040, '\P{^Blk: nb}', "");
Expect(1, 918000, '\p{Blk=	NB}', "");
Expect(0, 918000, '\p{^Blk=	NB}', "");
Expect(0, 918000, '\P{Blk=	NB}', "");
Expect(1, 918000, '\P{^Blk=	NB}', "");
Expect(0, 983040, '\p{Blk=	NB}', "");
Expect(1, 983040, '\p{^Blk=	NB}', "");
Expect(1, 983040, '\P{Blk=	NB}', "");
Expect(0, 983040, '\P{^Blk=	NB}', "");
Error('\p{Is_Block=_NO_BLOCK:=}');
Error('\P{Is_Block=_NO_BLOCK:=}');
Expect(1, 918000, '\p{Is_Block=noblock}', "");
Expect(0, 918000, '\p{^Is_Block=noblock}', "");
Expect(0, 918000, '\P{Is_Block=noblock}', "");
Expect(1, 918000, '\P{^Is_Block=noblock}', "");
Expect(0, 983040, '\p{Is_Block=noblock}', "");
Expect(1, 983040, '\p{^Is_Block=noblock}', "");
Expect(1, 983040, '\P{Is_Block=noblock}', "");
Expect(0, 983040, '\P{^Is_Block=noblock}', "");
Expect(1, 918000, '\p{Is_Block= No_Block}', "");
Expect(0, 918000, '\p{^Is_Block= No_Block}', "");
Expect(0, 918000, '\P{Is_Block= No_Block}', "");
Expect(1, 918000, '\P{^Is_Block= No_Block}', "");
Expect(0, 983040, '\p{Is_Block= No_Block}', "");
Expect(1, 983040, '\p{^Is_Block= No_Block}', "");
Expect(1, 983040, '\P{Is_Block= No_Block}', "");
Expect(0, 983040, '\P{^Is_Block= No_Block}', "");
Error('\p{Is_Blk= 	NB/a/}');
Error('\P{Is_Blk= 	NB/a/}');
Expect(1, 918000, '\p{Is_Blk=nb}', "");
Expect(0, 918000, '\p{^Is_Blk=nb}', "");
Expect(0, 918000, '\P{Is_Blk=nb}', "");
Expect(1, 918000, '\P{^Is_Blk=nb}', "");
Expect(0, 983040, '\p{Is_Blk=nb}', "");
Expect(1, 983040, '\p{^Is_Blk=nb}', "");
Expect(1, 983040, '\P{Is_Blk=nb}', "");
Expect(0, 983040, '\P{^Is_Blk=nb}', "");
Expect(1, 918000, '\p{Is_Blk: -NB}', "");
Expect(0, 918000, '\p{^Is_Blk: -NB}', "");
Expect(0, 918000, '\P{Is_Blk: -NB}', "");
Expect(1, 918000, '\P{^Is_Blk: -NB}', "");
Expect(0, 983040, '\p{Is_Blk: -NB}', "");
Expect(1, 983040, '\p{^Is_Blk: -NB}', "");
Expect(1, 983040, '\P{Is_Blk: -NB}', "");
Expect(0, 983040, '\P{^Is_Blk: -NB}', "");
Error('\p{Block= -GEORGIAN_Supplement:=}');
Error('\P{Block= -GEORGIAN_Supplement:=}');
Expect(1, 11567, '\p{Block:	georgiansupplement}', "");
Expect(0, 11567, '\p{^Block:	georgiansupplement}', "");
Expect(0, 11567, '\P{Block:	georgiansupplement}', "");
Expect(1, 11567, '\P{^Block:	georgiansupplement}', "");
Expect(0, 11568, '\p{Block:	georgiansupplement}', "");
Expect(1, 11568, '\p{^Block:	georgiansupplement}', "");
Expect(1, 11568, '\P{Block:	georgiansupplement}', "");
Expect(0, 11568, '\P{^Block:	georgiansupplement}', "");
Expect(1, 11567, '\p{Block=- Georgian_Supplement}', "");
Expect(0, 11567, '\p{^Block=- Georgian_Supplement}', "");
Expect(0, 11567, '\P{Block=- Georgian_Supplement}', "");
Expect(1, 11567, '\P{^Block=- Georgian_Supplement}', "");
Expect(0, 11568, '\p{Block=- Georgian_Supplement}', "");
Expect(1, 11568, '\p{^Block=- Georgian_Supplement}', "");
Expect(1, 11568, '\P{Block=- Georgian_Supplement}', "");
Expect(0, 11568, '\P{^Block=- Georgian_Supplement}', "");
Error('\p{Blk=:=GEORGIAN_Sup}');
Error('\P{Blk=:=GEORGIAN_Sup}');
Expect(1, 11567, '\p{Blk=georgiansup}', "");
Expect(0, 11567, '\p{^Blk=georgiansup}', "");
Expect(0, 11567, '\P{Blk=georgiansup}', "");
Expect(1, 11567, '\P{^Blk=georgiansup}', "");
Expect(0, 11568, '\p{Blk=georgiansup}', "");
Expect(1, 11568, '\p{^Blk=georgiansup}', "");
Expect(1, 11568, '\P{Blk=georgiansup}', "");
Expect(0, 11568, '\P{^Blk=georgiansup}', "");
Expect(1, 11567, '\p{Blk=_GEORGIAN_SUP}', "");
Expect(0, 11567, '\p{^Blk=_GEORGIAN_SUP}', "");
Expect(0, 11567, '\P{Blk=_GEORGIAN_SUP}', "");
Expect(1, 11567, '\P{^Blk=_GEORGIAN_SUP}', "");
Expect(0, 11568, '\p{Blk=_GEORGIAN_SUP}', "");
Expect(1, 11568, '\p{^Blk=_GEORGIAN_SUP}', "");
Expect(1, 11568, '\P{Blk=_GEORGIAN_SUP}', "");
Expect(0, 11568, '\P{^Blk=_GEORGIAN_SUP}', "");
Error('\p{Is_Block=:=Georgian_Supplement}');
Error('\P{Is_Block=:=Georgian_Supplement}');
Expect(1, 11567, '\p{Is_Block=georgiansupplement}', "");
Expect(0, 11567, '\p{^Is_Block=georgiansupplement}', "");
Expect(0, 11567, '\P{Is_Block=georgiansupplement}', "");
Expect(1, 11567, '\P{^Is_Block=georgiansupplement}', "");
Expect(0, 11568, '\p{Is_Block=georgiansupplement}', "");
Expect(1, 11568, '\p{^Is_Block=georgiansupplement}', "");
Expect(1, 11568, '\P{Is_Block=georgiansupplement}', "");
Expect(0, 11568, '\P{^Is_Block=georgiansupplement}', "");
Expect(1, 11567, '\p{Is_Block=	georgian_supplement}', "");
Expect(0, 11567, '\p{^Is_Block=	georgian_supplement}', "");
Expect(0, 11567, '\P{Is_Block=	georgian_supplement}', "");
Expect(1, 11567, '\P{^Is_Block=	georgian_supplement}', "");
Expect(0, 11568, '\p{Is_Block=	georgian_supplement}', "");
Expect(1, 11568, '\p{^Is_Block=	georgian_supplement}', "");
Expect(1, 11568, '\P{Is_Block=	georgian_supplement}', "");
Expect(0, 11568, '\P{^Is_Block=	georgian_supplement}', "");
Error('\p{Is_Blk= /a/georgian_sup}');
Error('\P{Is_Blk= /a/georgian_sup}');
Expect(1, 11567, '\p{Is_Blk=georgiansup}', "");
Expect(0, 11567, '\p{^Is_Blk=georgiansup}', "");
Expect(0, 11567, '\P{Is_Blk=georgiansup}', "");
Expect(1, 11567, '\P{^Is_Blk=georgiansup}', "");
Expect(0, 11568, '\p{Is_Blk=georgiansup}', "");
Expect(1, 11568, '\p{^Is_Blk=georgiansup}', "");
Expect(1, 11568, '\P{Is_Blk=georgiansup}', "");
Expect(0, 11568, '\P{^Is_Blk=georgiansup}', "");
Expect(1, 11567, '\p{Is_Blk:   	-georgian_Sup}', "");
Expect(0, 11567, '\p{^Is_Blk:   	-georgian_Sup}', "");
Expect(0, 11567, '\P{Is_Blk:   	-georgian_Sup}', "");
Expect(1, 11567, '\P{^Is_Blk:   	-georgian_Sup}', "");
Expect(0, 11568, '\p{Is_Blk:   	-georgian_Sup}', "");
Expect(1, 11568, '\p{^Is_Blk:   	-georgian_Sup}', "");
Expect(1, 11568, '\P{Is_Blk:   	-georgian_Sup}', "");
Expect(0, 11568, '\P{^Is_Blk:   	-georgian_Sup}', "");
Error('\p{Block=Vedic_Extensions:=}');
Error('\P{Block=Vedic_Extensions:=}');
Expect(1, 7423, '\p{Block=vedicextensions}', "");
Expect(0, 7423, '\p{^Block=vedicextensions}', "");
Expect(0, 7423, '\P{Block=vedicextensions}', "");
Expect(1, 7423, '\P{^Block=vedicextensions}', "");
Expect(0, 7424, '\p{Block=vedicextensions}', "");
Expect(1, 7424, '\p{^Block=vedicextensions}', "");
Expect(1, 7424, '\P{Block=vedicextensions}', "");
Expect(0, 7424, '\P{^Block=vedicextensions}', "");
Expect(1, 7423, '\p{Block=__Vedic_extensions}', "");
Expect(0, 7423, '\p{^Block=__Vedic_extensions}', "");
Expect(0, 7423, '\P{Block=__Vedic_extensions}', "");
Expect(1, 7423, '\P{^Block=__Vedic_extensions}', "");
Expect(0, 7424, '\p{Block=__Vedic_extensions}', "");
Expect(1, 7424, '\p{^Block=__Vedic_extensions}', "");
Expect(1, 7424, '\P{Block=__Vedic_extensions}', "");
Expect(0, 7424, '\P{^Block=__Vedic_extensions}', "");
Error('\p{Blk=:=_Vedic_Ext}');
Error('\P{Blk=:=_Vedic_Ext}');
Expect(1, 7423, '\p{Blk=vedicext}', "");
Expect(0, 7423, '\p{^Blk=vedicext}', "");
Expect(0, 7423, '\P{Blk=vedicext}', "");
Expect(1, 7423, '\P{^Blk=vedicext}', "");
Expect(0, 7424, '\p{Blk=vedicext}', "");
Expect(1, 7424, '\p{^Blk=vedicext}', "");
Expect(1, 7424, '\P{Blk=vedicext}', "");
Expect(0, 7424, '\P{^Blk=vedicext}', "");
Expect(1, 7423, '\p{Blk= _Vedic_Ext}', "");
Expect(0, 7423, '\p{^Blk= _Vedic_Ext}', "");
Expect(0, 7423, '\P{Blk= _Vedic_Ext}', "");
Expect(1, 7423, '\P{^Blk= _Vedic_Ext}', "");
Expect(0, 7424, '\p{Blk= _Vedic_Ext}', "");
Expect(1, 7424, '\p{^Blk= _Vedic_Ext}', "");
Expect(1, 7424, '\P{Blk= _Vedic_Ext}', "");
Expect(0, 7424, '\P{^Blk= _Vedic_Ext}', "");
Error('\p{Is_Block:   _:=Vedic_Extensions}');
Error('\P{Is_Block:   _:=Vedic_Extensions}');
Expect(1, 7423, '\p{Is_Block=vedicextensions}', "");
Expect(0, 7423, '\p{^Is_Block=vedicextensions}', "");
Expect(0, 7423, '\P{Is_Block=vedicextensions}', "");
Expect(1, 7423, '\P{^Is_Block=vedicextensions}', "");
Expect(0, 7424, '\p{Is_Block=vedicextensions}', "");
Expect(1, 7424, '\p{^Is_Block=vedicextensions}', "");
Expect(1, 7424, '\P{Is_Block=vedicextensions}', "");
Expect(0, 7424, '\P{^Is_Block=vedicextensions}', "");
Expect(1, 7423, '\p{Is_Block=	Vedic_Extensions}', "");
Expect(0, 7423, '\p{^Is_Block=	Vedic_Extensions}', "");
Expect(0, 7423, '\P{Is_Block=	Vedic_Extensions}', "");
Expect(1, 7423, '\P{^Is_Block=	Vedic_Extensions}', "");
Expect(0, 7424, '\p{Is_Block=	Vedic_Extensions}', "");
Expect(1, 7424, '\p{^Is_Block=	Vedic_Extensions}', "");
Expect(1, 7424, '\P{Is_Block=	Vedic_Extensions}', "");
Expect(0, 7424, '\P{^Is_Block=	Vedic_Extensions}', "");
Error('\p{Is_Blk=-/a/Vedic_Ext}');
Error('\P{Is_Blk=-/a/Vedic_Ext}');
Expect(1, 7423, '\p{Is_Blk:   vedicext}', "");
Expect(0, 7423, '\p{^Is_Blk:   vedicext}', "");
Expect(0, 7423, '\P{Is_Blk:   vedicext}', "");
Expect(1, 7423, '\P{^Is_Blk:   vedicext}', "");
Expect(0, 7424, '\p{Is_Blk:   vedicext}', "");
Expect(1, 7424, '\p{^Is_Blk:   vedicext}', "");
Expect(1, 7424, '\P{Is_Blk:   vedicext}', "");
Expect(0, 7424, '\P{^Is_Blk:   vedicext}', "");
Expect(1, 7423, '\p{Is_Blk=_Vedic_Ext}', "");
Expect(0, 7423, '\p{^Is_Blk=_Vedic_Ext}', "");
Expect(0, 7423, '\P{Is_Blk=_Vedic_Ext}', "");
Expect(1, 7423, '\P{^Is_Blk=_Vedic_Ext}', "");
Expect(0, 7424, '\p{Is_Blk=_Vedic_Ext}', "");
Expect(1, 7424, '\p{^Is_Blk=_Vedic_Ext}', "");
Expect(1, 7424, '\P{Is_Blk=_Vedic_Ext}', "");
Expect(0, 7424, '\P{^Is_Blk=_Vedic_Ext}', "");
Error('\p{Block=_:=Tai_VIET}');
Error('\P{Block=_:=Tai_VIET}');
Expect(1, 43743, '\p{Block=taiviet}', "");
Expect(0, 43743, '\p{^Block=taiviet}', "");
Expect(0, 43743, '\P{Block=taiviet}', "");
Expect(1, 43743, '\P{^Block=taiviet}', "");
Expect(0, 43744, '\p{Block=taiviet}', "");
Expect(1, 43744, '\p{^Block=taiviet}', "");
Expect(1, 43744, '\P{Block=taiviet}', "");
Expect(0, 43744, '\P{^Block=taiviet}', "");
Expect(1, 43743, '\p{Block=--TAI_Viet}', "");
Expect(0, 43743, '\p{^Block=--TAI_Viet}', "");
Expect(0, 43743, '\P{Block=--TAI_Viet}', "");
Expect(1, 43743, '\P{^Block=--TAI_Viet}', "");
Expect(0, 43744, '\p{Block=--TAI_Viet}', "");
Expect(1, 43744, '\p{^Block=--TAI_Viet}', "");
Expect(1, 43744, '\P{Block=--TAI_Viet}', "");
Expect(0, 43744, '\P{^Block=--TAI_Viet}', "");
Error('\p{Blk=:=-	TAI_viet}');
Error('\P{Blk=:=-	TAI_viet}');
Expect(1, 43743, '\p{Blk=taiviet}', "");
Expect(0, 43743, '\p{^Blk=taiviet}', "");
Expect(0, 43743, '\P{Blk=taiviet}', "");
Expect(1, 43743, '\P{^Blk=taiviet}', "");
Expect(0, 43744, '\p{Blk=taiviet}', "");
Expect(1, 43744, '\p{^Blk=taiviet}', "");
Expect(1, 43744, '\P{Blk=taiviet}', "");
Expect(0, 43744, '\P{^Blk=taiviet}', "");
Expect(1, 43743, '\p{Blk=-	tai_Viet}', "");
Expect(0, 43743, '\p{^Blk=-	tai_Viet}', "");
Expect(0, 43743, '\P{Blk=-	tai_Viet}', "");
Expect(1, 43743, '\P{^Blk=-	tai_Viet}', "");
Expect(0, 43744, '\p{Blk=-	tai_Viet}', "");
Expect(1, 43744, '\p{^Blk=-	tai_Viet}', "");
Expect(1, 43744, '\P{Blk=-	tai_Viet}', "");
Expect(0, 43744, '\P{^Blk=-	tai_Viet}', "");
Error('\p{Is_Block=_	Tai_Viet/a/}');
Error('\P{Is_Block=_	Tai_Viet/a/}');
Expect(1, 43743, '\p{Is_Block=taiviet}', "");
Expect(0, 43743, '\p{^Is_Block=taiviet}', "");
Expect(0, 43743, '\P{Is_Block=taiviet}', "");
Expect(1, 43743, '\P{^Is_Block=taiviet}', "");
Expect(0, 43744, '\p{Is_Block=taiviet}', "");
Expect(1, 43744, '\p{^Is_Block=taiviet}', "");
Expect(1, 43744, '\P{Is_Block=taiviet}', "");
Expect(0, 43744, '\P{^Is_Block=taiviet}', "");
Expect(1, 43743, '\p{Is_Block=_ Tai_VIET}', "");
Expect(0, 43743, '\p{^Is_Block=_ Tai_VIET}', "");
Expect(0, 43743, '\P{Is_Block=_ Tai_VIET}', "");
Expect(1, 43743, '\P{^Is_Block=_ Tai_VIET}', "");
Expect(0, 43744, '\p{Is_Block=_ Tai_VIET}', "");
Expect(1, 43744, '\p{^Is_Block=_ Tai_VIET}', "");
Expect(1, 43744, '\P{Is_Block=_ Tai_VIET}', "");
Expect(0, 43744, '\P{^Is_Block=_ Tai_VIET}', "");
Error('\p{Is_Blk=:=_Tai_Viet}');
Error('\P{Is_Blk=:=_Tai_Viet}');
Expect(1, 43743, '\p{Is_Blk=taiviet}', "");
Expect(0, 43743, '\p{^Is_Blk=taiviet}', "");
Expect(0, 43743, '\P{Is_Blk=taiviet}', "");
Expect(1, 43743, '\P{^Is_Blk=taiviet}', "");
Expect(0, 43744, '\p{Is_Blk=taiviet}', "");
Expect(1, 43744, '\p{^Is_Blk=taiviet}', "");
Expect(1, 43744, '\P{Is_Blk=taiviet}', "");
Expect(0, 43744, '\P{^Is_Blk=taiviet}', "");
Expect(1, 43743, '\p{Is_Blk=_Tai_Viet}', "");
Expect(0, 43743, '\p{^Is_Blk=_Tai_Viet}', "");
Expect(0, 43743, '\P{Is_Blk=_Tai_Viet}', "");
Expect(1, 43743, '\P{^Is_Blk=_Tai_Viet}', "");
Expect(0, 43744, '\p{Is_Blk=_Tai_Viet}', "");
Expect(1, 43744, '\p{^Is_Blk=_Tai_Viet}', "");
Expect(1, 43744, '\P{Is_Blk=_Tai_Viet}', "");
Expect(0, 43744, '\P{^Is_Blk=_Tai_Viet}', "");
Error('\p{Block=/a/_Lisu}');
Error('\P{Block=/a/_Lisu}');
Expect(1, 42239, '\p{Block=lisu}', "");
Expect(0, 42239, '\p{^Block=lisu}', "");
Expect(0, 42239, '\P{Block=lisu}', "");
Expect(1, 42239, '\P{^Block=lisu}', "");
Expect(0, 42240, '\p{Block=lisu}', "");
Expect(1, 42240, '\p{^Block=lisu}', "");
Expect(1, 42240, '\P{Block=lisu}', "");
Expect(0, 42240, '\P{^Block=lisu}', "");
Expect(1, 42239, '\p{Block=_-Lisu}', "");
Expect(0, 42239, '\p{^Block=_-Lisu}', "");
Expect(0, 42239, '\P{Block=_-Lisu}', "");
Expect(1, 42239, '\P{^Block=_-Lisu}', "");
Expect(0, 42240, '\p{Block=_-Lisu}', "");
Expect(1, 42240, '\p{^Block=_-Lisu}', "");
Expect(1, 42240, '\P{Block=_-Lisu}', "");
Expect(0, 42240, '\P{^Block=_-Lisu}', "");
Error('\p{Blk= :=Lisu}');
Error('\P{Blk= :=Lisu}');
Expect(1, 42239, '\p{Blk=lisu}', "");
Expect(0, 42239, '\p{^Blk=lisu}', "");
Expect(0, 42239, '\P{Blk=lisu}', "");
Expect(1, 42239, '\P{^Blk=lisu}', "");
Expect(0, 42240, '\p{Blk=lisu}', "");
Expect(1, 42240, '\p{^Blk=lisu}', "");
Expect(1, 42240, '\P{Blk=lisu}', "");
Expect(0, 42240, '\P{^Blk=lisu}', "");
Expect(1, 42239, '\p{Blk=-_LISU}', "");
Expect(0, 42239, '\p{^Blk=-_LISU}', "");
Expect(0, 42239, '\P{Blk=-_LISU}', "");
Expect(1, 42239, '\P{^Blk=-_LISU}', "");
Expect(0, 42240, '\p{Blk=-_LISU}', "");
Expect(1, 42240, '\p{^Blk=-_LISU}', "");
Expect(1, 42240, '\P{Blk=-_LISU}', "");
Expect(0, 42240, '\P{^Blk=-_LISU}', "");
Error('\p{Is_Block=:=-	Lisu}');
Error('\P{Is_Block=:=-	Lisu}');
Expect(1, 42239, '\p{Is_Block=lisu}', "");
Expect(0, 42239, '\p{^Is_Block=lisu}', "");
Expect(0, 42239, '\P{Is_Block=lisu}', "");
Expect(1, 42239, '\P{^Is_Block=lisu}', "");
Expect(0, 42240, '\p{Is_Block=lisu}', "");
Expect(1, 42240, '\p{^Is_Block=lisu}', "");
Expect(1, 42240, '\P{Is_Block=lisu}', "");
Expect(0, 42240, '\P{^Is_Block=lisu}', "");
Expect(1, 42239, '\p{Is_Block=--Lisu}', "");
Expect(0, 42239, '\p{^Is_Block=--Lisu}', "");
Expect(0, 42239, '\P{Is_Block=--Lisu}', "");
Expect(1, 42239, '\P{^Is_Block=--Lisu}', "");
Expect(0, 42240, '\p{Is_Block=--Lisu}', "");
Expect(1, 42240, '\p{^Is_Block=--Lisu}', "");
Expect(1, 42240, '\P{Is_Block=--Lisu}', "");
Expect(0, 42240, '\P{^Is_Block=--Lisu}', "");
Error('\p{Is_Blk=:=__lisu}');
Error('\P{Is_Blk=:=__lisu}');
Expect(1, 42239, '\p{Is_Blk=lisu}', "");
Expect(0, 42239, '\p{^Is_Blk=lisu}', "");
Expect(0, 42239, '\P{Is_Blk=lisu}', "");
Expect(1, 42239, '\P{^Is_Blk=lisu}', "");
Expect(0, 42240, '\p{Is_Blk=lisu}', "");
Expect(1, 42240, '\p{^Is_Blk=lisu}', "");
Expect(1, 42240, '\P{Is_Blk=lisu}', "");
Expect(0, 42240, '\P{^Is_Blk=lisu}', "");
Expect(1, 42239, '\p{Is_Blk=_-LISU}', "");
Expect(0, 42239, '\p{^Is_Blk=_-LISU}', "");
Expect(0, 42239, '\P{Is_Blk=_-LISU}', "");
Expect(1, 42239, '\P{^Is_Blk=_-LISU}', "");
Expect(0, 42240, '\p{Is_Blk=_-LISU}', "");
Expect(1, 42240, '\p{^Is_Blk=_-LISU}', "");
Expect(1, 42240, '\P{Is_Blk=_-LISU}', "");
Expect(0, 42240, '\P{^Is_Blk=_-LISU}', "");
Error('\p{Block= Arabic_Presentation_Forms_a/a/}');
Error('\P{Block= Arabic_Presentation_Forms_a/a/}');
Expect(1, 65023, '\p{Block=arabicpresentationformsa}', "");
Expect(0, 65023, '\p{^Block=arabicpresentationformsa}', "");
Expect(0, 65023, '\P{Block=arabicpresentationformsa}', "");
Expect(1, 65023, '\P{^Block=arabicpresentationformsa}', "");
Expect(0, 65024, '\p{Block=arabicpresentationformsa}', "");
Expect(1, 65024, '\p{^Block=arabicpresentationformsa}', "");
Expect(1, 65024, '\P{Block=arabicpresentationformsa}', "");
Expect(0, 65024, '\P{^Block=arabicpresentationformsa}', "");
Expect(1, 65023, '\p{Block=	ARABIC_presentation_forms_A}', "");
Expect(0, 65023, '\p{^Block=	ARABIC_presentation_forms_A}', "");
Expect(0, 65023, '\P{Block=	ARABIC_presentation_forms_A}', "");
Expect(1, 65023, '\P{^Block=	ARABIC_presentation_forms_A}', "");
Expect(0, 65024, '\p{Block=	ARABIC_presentation_forms_A}', "");
Expect(1, 65024, '\p{^Block=	ARABIC_presentation_forms_A}', "");
Expect(1, 65024, '\P{Block=	ARABIC_presentation_forms_A}', "");
Expect(0, 65024, '\P{^Block=	ARABIC_presentation_forms_A}', "");
Error('\p{Blk=/a/	 ARABIC_PF_a}');
Error('\P{Blk=/a/	 ARABIC_PF_a}');
Expect(1, 65023, '\p{Blk=arabicpfa}', "");
Expect(0, 65023, '\p{^Blk=arabicpfa}', "");
Expect(0, 65023, '\P{Blk=arabicpfa}', "");
Expect(1, 65023, '\P{^Blk=arabicpfa}', "");
Expect(0, 65024, '\p{Blk=arabicpfa}', "");
Expect(1, 65024, '\p{^Blk=arabicpfa}', "");
Expect(1, 65024, '\P{Blk=arabicpfa}', "");
Expect(0, 65024, '\P{^Blk=arabicpfa}', "");
Expect(1, 65023, '\p{Blk=	-arabic_pf_a}', "");
Expect(0, 65023, '\p{^Blk=	-arabic_pf_a}', "");
Expect(0, 65023, '\P{Blk=	-arabic_pf_a}', "");
Expect(1, 65023, '\P{^Blk=	-arabic_pf_a}', "");
Expect(0, 65024, '\p{Blk=	-arabic_pf_a}', "");
Expect(1, 65024, '\p{^Blk=	-arabic_pf_a}', "");
Expect(1, 65024, '\P{Blk=	-arabic_pf_a}', "");
Expect(0, 65024, '\P{^Blk=	-arabic_pf_a}', "");
Error('\p{Is_Block= /a/Arabic_Presentation_forms_a}');
Error('\P{Is_Block= /a/Arabic_Presentation_forms_a}');
Expect(1, 65023, '\p{Is_Block=arabicpresentationformsa}', "");
Expect(0, 65023, '\p{^Is_Block=arabicpresentationformsa}', "");
Expect(0, 65023, '\P{Is_Block=arabicpresentationformsa}', "");
Expect(1, 65023, '\P{^Is_Block=arabicpresentationformsa}', "");
Expect(0, 65024, '\p{Is_Block=arabicpresentationformsa}', "");
Expect(1, 65024, '\p{^Is_Block=arabicpresentationformsa}', "");
Expect(1, 65024, '\P{Is_Block=arabicpresentationformsa}', "");
Expect(0, 65024, '\P{^Is_Block=arabicpresentationformsa}', "");
Expect(1, 65023, '\p{Is_Block= 	Arabic_Presentation_FORMS_a}', "");
Expect(0, 65023, '\p{^Is_Block= 	Arabic_Presentation_FORMS_a}', "");
Expect(0, 65023, '\P{Is_Block= 	Arabic_Presentation_FORMS_a}', "");
Expect(1, 65023, '\P{^Is_Block= 	Arabic_Presentation_FORMS_a}', "");
Expect(0, 65024, '\p{Is_Block= 	Arabic_Presentation_FORMS_a}', "");
Expect(1, 65024, '\p{^Is_Block= 	Arabic_Presentation_FORMS_a}', "");
Expect(1, 65024, '\P{Is_Block= 	Arabic_Presentation_FORMS_a}', "");
Expect(0, 65024, '\P{^Is_Block= 	Arabic_Presentation_FORMS_a}', "");
Error('\p{Is_Blk=_ARABIC_PF_A/a/}');
Error('\P{Is_Blk=_ARABIC_PF_A/a/}');
Expect(1, 65023, '\p{Is_Blk=arabicpfa}', "");
Expect(0, 65023, '\p{^Is_Blk=arabicpfa}', "");
Expect(0, 65023, '\P{Is_Blk=arabicpfa}', "");
Expect(1, 65023, '\P{^Is_Blk=arabicpfa}', "");
Expect(0, 65024, '\p{Is_Blk=arabicpfa}', "");
Expect(1, 65024, '\p{^Is_Blk=arabicpfa}', "");
Expect(1, 65024, '\P{Is_Blk=arabicpfa}', "");
Expect(0, 65024, '\P{^Is_Blk=arabicpfa}', "");
Expect(1, 65023, '\p{Is_Blk= _Arabic_PF_A}', "");
Expect(0, 65023, '\p{^Is_Blk= _Arabic_PF_A}', "");
Expect(0, 65023, '\P{Is_Blk= _Arabic_PF_A}', "");
Expect(1, 65023, '\P{^Is_Blk= _Arabic_PF_A}', "");
Expect(0, 65024, '\p{Is_Blk= _Arabic_PF_A}', "");
Expect(1, 65024, '\p{^Is_Blk= _Arabic_PF_A}', "");
Expect(1, 65024, '\P{Is_Blk= _Arabic_PF_A}', "");
Expect(0, 65024, '\P{^Is_Blk= _Arabic_PF_A}', "");
Error('\p{Block=:=__basic_latin}');
Error('\P{Block=:=__basic_latin}');
Expect(1, 127, '\p{Block=basiclatin}', "");
Expect(0, 127, '\p{^Block=basiclatin}', "");
Expect(0, 127, '\P{Block=basiclatin}', "");
Expect(1, 127, '\P{^Block=basiclatin}', "");
Expect(0, 128, '\p{Block=basiclatin}', "");
Expect(1, 128, '\p{^Block=basiclatin}', "");
Expect(1, 128, '\P{Block=basiclatin}', "");
Expect(0, 128, '\P{^Block=basiclatin}', "");
Expect(1, 127, '\p{Block= Basic_latin}', "");
Expect(0, 127, '\p{^Block= Basic_latin}', "");
Expect(0, 127, '\P{Block= Basic_latin}', "");
Expect(1, 127, '\P{^Block= Basic_latin}', "");
Expect(0, 128, '\p{Block= Basic_latin}', "");
Expect(1, 128, '\p{^Block= Basic_latin}', "");
Expect(1, 128, '\P{Block= Basic_latin}', "");
Expect(0, 128, '\P{^Block= Basic_latin}', "");
Error('\p{Blk=:=- ASCII}');
Error('\P{Blk=:=- ASCII}');
Expect(1, 127, '\p{Blk=ascii}', "");
Expect(0, 127, '\p{^Blk=ascii}', "");
Expect(0, 127, '\P{Blk=ascii}', "");
Expect(1, 127, '\P{^Blk=ascii}', "");
Expect(0, 128, '\p{Blk=ascii}', "");
Expect(1, 128, '\p{^Blk=ascii}', "");
Expect(1, 128, '\P{Blk=ascii}', "");
Expect(0, 128, '\P{^Blk=ascii}', "");
Expect(1, 127, '\p{Blk= _ascii}', "");
Expect(0, 127, '\p{^Blk= _ascii}', "");
Expect(0, 127, '\P{Blk= _ascii}', "");
Expect(1, 127, '\P{^Blk= _ascii}', "");
Expect(0, 128, '\p{Blk= _ascii}', "");
Expect(1, 128, '\p{^Blk= _ascii}', "");
Expect(1, 128, '\P{Blk= _ascii}', "");
Expect(0, 128, '\P{^Blk= _ascii}', "");
Error('\p{Is_Block=-_Basic_Latin:=}');
Error('\P{Is_Block=-_Basic_Latin:=}');
Expect(1, 127, '\p{Is_Block=basiclatin}', "");
Expect(0, 127, '\p{^Is_Block=basiclatin}', "");
Expect(0, 127, '\P{Is_Block=basiclatin}', "");
Expect(1, 127, '\P{^Is_Block=basiclatin}', "");
Expect(0, 128, '\p{Is_Block=basiclatin}', "");
Expect(1, 128, '\p{^Is_Block=basiclatin}', "");
Expect(1, 128, '\P{Is_Block=basiclatin}', "");
Expect(0, 128, '\P{^Is_Block=basiclatin}', "");
Expect(1, 127, '\p{Is_Block=	Basic_Latin}', "");
Expect(0, 127, '\p{^Is_Block=	Basic_Latin}', "");
Expect(0, 127, '\P{Is_Block=	Basic_Latin}', "");
Expect(1, 127, '\P{^Is_Block=	Basic_Latin}', "");
Expect(0, 128, '\p{Is_Block=	Basic_Latin}', "");
Expect(1, 128, '\p{^Is_Block=	Basic_Latin}', "");
Expect(1, 128, '\P{Is_Block=	Basic_Latin}', "");
Expect(0, 128, '\P{^Is_Block=	Basic_Latin}', "");
Error('\p{Is_Blk=:= ASCII}');
Error('\P{Is_Blk=:= ASCII}');
Expect(1, 127, '\p{Is_Blk=ascii}', "");
Expect(0, 127, '\p{^Is_Blk=ascii}', "");
Expect(0, 127, '\P{Is_Blk=ascii}', "");
Expect(1, 127, '\P{^Is_Blk=ascii}', "");
Expect(0, 128, '\p{Is_Blk=ascii}', "");
Expect(1, 128, '\p{^Is_Blk=ascii}', "");
Expect(1, 128, '\P{Is_Blk=ascii}', "");
Expect(0, 128, '\P{^Is_Blk=ascii}', "");
Expect(1, 127, '\p{Is_Blk= 	ASCII}', "");
Expect(0, 127, '\p{^Is_Blk= 	ASCII}', "");
Expect(0, 127, '\P{Is_Blk= 	ASCII}', "");
Expect(1, 127, '\P{^Is_Blk= 	ASCII}', "");
Expect(0, 128, '\p{Is_Blk= 	ASCII}', "");
Expect(1, 128, '\p{^Is_Blk= 	ASCII}', "");
Expect(1, 128, '\P{Is_Blk= 	ASCII}', "");
Expect(0, 128, '\P{^Is_Blk= 	ASCII}', "");
Error('\p{Block=/a/_Miscellaneous_Symbols_and_Pictographs}');
Error('\P{Block=/a/_Miscellaneous_Symbols_and_Pictographs}');
Expect(1, 128511, '\p{Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\p{^Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\P{Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\P{^Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\p{Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\p{^Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\P{Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\P{^Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\p{Block=--Miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(0, 128511, '\p{^Block=--Miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(0, 128511, '\P{Block=--Miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(1, 128511, '\P{^Block=--Miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(0, 128512, '\p{Block=--Miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(1, 128512, '\p{^Block=--Miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(1, 128512, '\P{Block=--Miscellaneous_SYMBOLS_And_pictographs}', "");
Expect(0, 128512, '\P{^Block=--Miscellaneous_SYMBOLS_And_pictographs}', "");
Error('\p{Blk=	/a/misc_Pictographs}');
Error('\P{Blk=	/a/misc_Pictographs}');
Expect(1, 128511, '\p{Blk: miscpictographs}', "");
Expect(0, 128511, '\p{^Blk: miscpictographs}', "");
Expect(0, 128511, '\P{Blk: miscpictographs}', "");
Expect(1, 128511, '\P{^Blk: miscpictographs}', "");
Expect(0, 128512, '\p{Blk: miscpictographs}', "");
Expect(1, 128512, '\p{^Blk: miscpictographs}', "");
Expect(1, 128512, '\P{Blk: miscpictographs}', "");
Expect(0, 128512, '\P{^Blk: miscpictographs}', "");
Expect(1, 128511, '\p{Blk=  misc_Pictographs}', "");
Expect(0, 128511, '\p{^Blk=  misc_Pictographs}', "");
Expect(0, 128511, '\P{Blk=  misc_Pictographs}', "");
Expect(1, 128511, '\P{^Blk=  misc_Pictographs}', "");
Expect(0, 128512, '\p{Blk=  misc_Pictographs}', "");
Expect(1, 128512, '\p{^Blk=  misc_Pictographs}', "");
Expect(1, 128512, '\P{Blk=  misc_Pictographs}', "");
Expect(0, 128512, '\P{^Blk=  misc_Pictographs}', "");
Error('\p{Is_Block=	/a/MISCELLANEOUS_SYMBOLS_And_PICTOGRAPHS}');
Error('\P{Is_Block=	/a/MISCELLANEOUS_SYMBOLS_And_PICTOGRAPHS}');
Expect(1, 128511, '\p{Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\p{^Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128511, '\P{Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\P{^Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\p{Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\p{^Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128512, '\P{Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(0, 128512, '\P{^Is_Block=miscellaneoussymbolsandpictographs}', "");
Expect(1, 128511, '\p{Is_Block=		miscellaneous_Symbols_and_pictographs}', "");
Expect(0, 128511, '\p{^Is_Block=		miscellaneous_Symbols_and_pictographs}', "");
Expect(0, 128511, '\P{Is_Block=		miscellaneous_Symbols_and_pictographs}', "");
Expect(1, 128511, '\P{^Is_Block=		miscellaneous_Symbols_and_pictographs}', "");
Expect(0, 128512, '\p{Is_Block=		miscellaneous_Symbols_and_pictographs}', "");
Expect(1, 128512, '\p{^Is_Block=		miscellaneous_Symbols_and_pictographs}', "");
Expect(1, 128512, '\P{Is_Block=		miscellaneous_Symbols_and_pictographs}', "");
Expect(0, 128512, '\P{^Is_Block=		miscellaneous_Symbols_and_pictographs}', "");
Error('\p{Is_Blk=-Misc_PICTOGRAPHS/a/}');
Error('\P{Is_Blk=-Misc_PICTOGRAPHS/a/}');
Expect(1, 128511, '\p{Is_Blk=miscpictographs}', "");
Expect(0, 128511, '\p{^Is_Blk=miscpictographs}', "");
Expect(0, 128511, '\P{Is_Blk=miscpictographs}', "");
Expect(1, 128511, '\P{^Is_Blk=miscpictographs}', "");
Expect(0, 128512, '\p{Is_Blk=miscpictographs}', "");
Expect(1, 128512, '\p{^Is_Blk=miscpictographs}', "");
Expect(1, 128512, '\P{Is_Blk=miscpictographs}', "");
Expect(0, 128512, '\P{^Is_Blk=miscpictographs}', "");
Expect(1, 128511, '\p{Is_Blk:		-misc_pictographs}', "");
Expect(0, 128511, '\p{^Is_Blk:		-misc_pictographs}', "");
Expect(0, 128511, '\P{Is_Blk:		-misc_pictographs}', "");
Expect(1, 128511, '\P{^Is_Blk:		-misc_pictographs}', "");
Expect(0, 128512, '\p{Is_Blk:		-misc_pictographs}', "");
Expect(1, 128512, '\p{^Is_Blk:		-misc_pictographs}', "");
Expect(1, 128512, '\P{Is_Blk:		-misc_pictographs}', "");
Expect(0, 128512, '\P{^Is_Blk:		-misc_pictographs}', "");
Error('\p{Block=:= -Playing_Cards}');
Error('\P{Block=:= -Playing_Cards}');
Expect(1, 127231, '\p{Block=playingcards}', "");
Expect(0, 127231, '\p{^Block=playingcards}', "");
Expect(0, 127231, '\P{Block=playingcards}', "");
Expect(1, 127231, '\P{^Block=playingcards}', "");
Expect(0, 127232, '\p{Block=playingcards}', "");
Expect(1, 127232, '\p{^Block=playingcards}', "");
Expect(1, 127232, '\P{Block=playingcards}', "");
Expect(0, 127232, '\P{^Block=playingcards}', "");
Expect(1, 127231, '\p{Block: _ Playing_cards}', "");
Expect(0, 127231, '\p{^Block: _ Playing_cards}', "");
Expect(0, 127231, '\P{Block: _ Playing_cards}', "");
Expect(1, 127231, '\P{^Block: _ Playing_cards}', "");
Expect(0, 127232, '\p{Block: _ Playing_cards}', "");
Expect(1, 127232, '\p{^Block: _ Playing_cards}', "");
Expect(1, 127232, '\P{Block: _ Playing_cards}', "");
Expect(0, 127232, '\P{^Block: _ Playing_cards}', "");
Error('\p{Blk: /a/Playing_CARDS}');
Error('\P{Blk: /a/Playing_CARDS}');
Expect(1, 127231, '\p{Blk=playingcards}', "");
Expect(0, 127231, '\p{^Blk=playingcards}', "");
Expect(0, 127231, '\P{Blk=playingcards}', "");
Expect(1, 127231, '\P{^Blk=playingcards}', "");
Expect(0, 127232, '\p{Blk=playingcards}', "");
Expect(1, 127232, '\p{^Blk=playingcards}', "");
Expect(1, 127232, '\P{Blk=playingcards}', "");
Expect(0, 127232, '\P{^Blk=playingcards}', "");
Expect(1, 127231, '\p{Blk= 	Playing_cards}', "");
Expect(0, 127231, '\p{^Blk= 	Playing_cards}', "");
Expect(0, 127231, '\P{Blk= 	Playing_cards}', "");
Expect(1, 127231, '\P{^Blk= 	Playing_cards}', "");
Expect(0, 127232, '\p{Blk= 	Playing_cards}', "");
Expect(1, 127232, '\p{^Blk= 	Playing_cards}', "");
Expect(1, 127232, '\P{Blk= 	Playing_cards}', "");
Expect(0, 127232, '\P{^Blk= 	Playing_cards}', "");
Error('\p{Is_Block= 	Playing_Cards:=}');
Error('\P{Is_Block= 	Playing_Cards:=}');
Expect(1, 127231, '\p{Is_Block=playingcards}', "");
Expect(0, 127231, '\p{^Is_Block=playingcards}', "");
Expect(0, 127231, '\P{Is_Block=playingcards}', "");
Expect(1, 127231, '\P{^Is_Block=playingcards}', "");
Expect(0, 127232, '\p{Is_Block=playingcards}', "");
Expect(1, 127232, '\p{^Is_Block=playingcards}', "");
Expect(1, 127232, '\P{Is_Block=playingcards}', "");
Expect(0, 127232, '\P{^Is_Block=playingcards}', "");
Expect(1, 127231, '\p{Is_Block=-Playing_cards}', "");
Expect(0, 127231, '\p{^Is_Block=-Playing_cards}', "");
Expect(0, 127231, '\P{Is_Block=-Playing_cards}', "");
Expect(1, 127231, '\P{^Is_Block=-Playing_cards}', "");
Expect(0, 127232, '\p{Is_Block=-Playing_cards}', "");
Expect(1, 127232, '\p{^Is_Block=-Playing_cards}', "");
Expect(1, 127232, '\P{Is_Block=-Playing_cards}', "");
Expect(0, 127232, '\P{^Is_Block=-Playing_cards}', "");
Error('\p{Is_Blk= :=PLAYING_Cards}');
Error('\P{Is_Blk= :=PLAYING_Cards}');
Expect(1, 127231, '\p{Is_Blk=playingcards}', "");
Expect(0, 127231, '\p{^Is_Blk=playingcards}', "");
Expect(0, 127231, '\P{Is_Blk=playingcards}', "");
Expect(1, 127231, '\P{^Is_Blk=playingcards}', "");
Expect(0, 127232, '\p{Is_Blk=playingcards}', "");
Expect(1, 127232, '\p{^Is_Blk=playingcards}', "");
Expect(1, 127232, '\P{Is_Blk=playingcards}', "");
Expect(0, 127232, '\P{^Is_Blk=playingcards}', "");
Expect(1, 127231, '\p{Is_Blk=-_Playing_Cards}', "");
Expect(0, 127231, '\p{^Is_Blk=-_Playing_Cards}', "");
Expect(0, 127231, '\P{Is_Blk=-_Playing_Cards}', "");
Expect(1, 127231, '\P{^Is_Blk=-_Playing_Cards}', "");
Expect(0, 127232, '\p{Is_Blk=-_Playing_Cards}', "");
Expect(1, 127232, '\p{^Is_Blk=-_Playing_Cards}', "");
Expect(1, 127232, '\P{Is_Blk=-_Playing_Cards}', "");
Expect(0, 127232, '\P{^Is_Blk=-_Playing_Cards}', "");
Error('\p{Block= Javanese/a/}');
Error('\P{Block= Javanese/a/}');
Expect(1, 43487, '\p{Block=javanese}', "");
Expect(0, 43487, '\p{^Block=javanese}', "");
Expect(0, 43487, '\P{Block=javanese}', "");
Expect(1, 43487, '\P{^Block=javanese}', "");
Expect(0, 43488, '\p{Block=javanese}', "");
Expect(1, 43488, '\p{^Block=javanese}', "");
Expect(1, 43488, '\P{Block=javanese}', "");
Expect(0, 43488, '\P{^Block=javanese}', "");
Expect(1, 43487, '\p{Block=-	JAVANESE}', "");
Expect(0, 43487, '\p{^Block=-	JAVANESE}', "");
Expect(0, 43487, '\P{Block=-	JAVANESE}', "");
Expect(1, 43487, '\P{^Block=-	JAVANESE}', "");
Expect(0, 43488, '\p{Block=-	JAVANESE}', "");
Expect(1, 43488, '\p{^Block=-	JAVANESE}', "");
Expect(1, 43488, '\P{Block=-	JAVANESE}', "");
Expect(0, 43488, '\P{^Block=-	JAVANESE}', "");
Error('\p{Blk= /a/Javanese}');
Error('\P{Blk= /a/Javanese}');
Expect(1, 43487, '\p{Blk: javanese}', "");
Expect(0, 43487, '\p{^Blk: javanese}', "");
Expect(0, 43487, '\P{Blk: javanese}', "");
Expect(1, 43487, '\P{^Blk: javanese}', "");
Expect(0, 43488, '\p{Blk: javanese}', "");
Expect(1, 43488, '\p{^Blk: javanese}', "");
Expect(1, 43488, '\P{Blk: javanese}', "");
Expect(0, 43488, '\P{^Blk: javanese}', "");
Expect(1, 43487, '\p{Blk=	Javanese}', "");
Expect(0, 43487, '\p{^Blk=	Javanese}', "");
Expect(0, 43487, '\P{Blk=	Javanese}', "");
Expect(1, 43487, '\P{^Blk=	Javanese}', "");
Expect(0, 43488, '\p{Blk=	Javanese}', "");
Expect(1, 43488, '\p{^Blk=	Javanese}', "");
Expect(1, 43488, '\P{Blk=	Javanese}', "");
Expect(0, 43488, '\P{^Blk=	Javanese}', "");
Error('\p{Is_Block=:=Javanese}');
Error('\P{Is_Block=:=Javanese}');
Expect(1, 43487, '\p{Is_Block:	javanese}', "");
Expect(0, 43487, '\p{^Is_Block:	javanese}', "");
Expect(0, 43487, '\P{Is_Block:	javanese}', "");
Expect(1, 43487, '\P{^Is_Block:	javanese}', "");
Expect(0, 43488, '\p{Is_Block:	javanese}', "");
Expect(1, 43488, '\p{^Is_Block:	javanese}', "");
Expect(1, 43488, '\P{Is_Block:	javanese}', "");
Expect(0, 43488, '\P{^Is_Block:	javanese}', "");
Expect(1, 43487, '\p{Is_Block=-Javanese}', "");
Expect(0, 43487, '\p{^Is_Block=-Javanese}', "");
Expect(0, 43487, '\P{Is_Block=-Javanese}', "");
Expect(1, 43487, '\P{^Is_Block=-Javanese}', "");
Expect(0, 43488, '\p{Is_Block=-Javanese}', "");
Expect(1, 43488, '\p{^Is_Block=-Javanese}', "");
Expect(1, 43488, '\P{Is_Block=-Javanese}', "");
Expect(0, 43488, '\P{^Is_Block=-Javanese}', "");
Error('\p{Is_Blk=:=Javanese}');
Error('\P{Is_Blk=:=Javanese}');
Expect(1, 43487, '\p{Is_Blk=javanese}', "");
Expect(0, 43487, '\p{^Is_Blk=javanese}', "");
Expect(0, 43487, '\P{Is_Blk=javanese}', "");
Expect(1, 43487, '\P{^Is_Blk=javanese}', "");
Expect(0, 43488, '\p{Is_Blk=javanese}', "");
Expect(1, 43488, '\p{^Is_Blk=javanese}', "");
Expect(1, 43488, '\P{Is_Blk=javanese}', "");
Expect(0, 43488, '\P{^Is_Blk=javanese}', "");
Expect(1, 43487, '\p{Is_Blk=		javanese}', "");
Expect(0, 43487, '\p{^Is_Blk=		javanese}', "");
Expect(0, 43487, '\P{Is_Blk=		javanese}', "");
Expect(1, 43487, '\P{^Is_Blk=		javanese}', "");
Expect(0, 43488, '\p{Is_Blk=		javanese}', "");
Expect(1, 43488, '\p{^Is_Blk=		javanese}', "");
Expect(1, 43488, '\P{Is_Blk=		javanese}', "");
Expect(0, 43488, '\P{^Is_Blk=		javanese}', "");
Error('\p{Block: /a/LOW_Surrogates}');
Error('\P{Block: /a/LOW_Surrogates}');
Expect(1, 57343, '\p{Block=lowsurrogates}', "");
Expect(0, 57343, '\p{^Block=lowsurrogates}', "");
Expect(0, 57343, '\P{Block=lowsurrogates}', "");
Expect(1, 57343, '\P{^Block=lowsurrogates}', "");
Expect(0, 57344, '\p{Block=lowsurrogates}', "");
Expect(1, 57344, '\p{^Block=lowsurrogates}', "");
Expect(1, 57344, '\P{Block=lowsurrogates}', "");
Expect(0, 57344, '\P{^Block=lowsurrogates}', "");
Expect(1, 57343, '\p{Block=_Low_Surrogates}', "");
Expect(0, 57343, '\p{^Block=_Low_Surrogates}', "");
Expect(0, 57343, '\P{Block=_Low_Surrogates}', "");
Expect(1, 57343, '\P{^Block=_Low_Surrogates}', "");
Expect(0, 57344, '\p{Block=_Low_Surrogates}', "");
Expect(1, 57344, '\p{^Block=_Low_Surrogates}', "");
Expect(1, 57344, '\P{Block=_Low_Surrogates}', "");
Expect(0, 57344, '\P{^Block=_Low_Surrogates}', "");
Error('\p{Blk=:=	_LOW_SURROGATES}');
Error('\P{Blk=:=	_LOW_SURROGATES}');
Expect(1, 57343, '\p{Blk=lowsurrogates}', "");
Expect(0, 57343, '\p{^Blk=lowsurrogates}', "");
Expect(0, 57343, '\P{Blk=lowsurrogates}', "");
Expect(1, 57343, '\P{^Blk=lowsurrogates}', "");
Expect(0, 57344, '\p{Blk=lowsurrogates}', "");
Expect(1, 57344, '\p{^Blk=lowsurrogates}', "");
Expect(1, 57344, '\P{Blk=lowsurrogates}', "");
Expect(0, 57344, '\P{^Blk=lowsurrogates}', "");
Expect(1, 57343, '\p{Blk=	Low_SURROGATES}', "");
Expect(0, 57343, '\p{^Blk=	Low_SURROGATES}', "");
Expect(0, 57343, '\P{Blk=	Low_SURROGATES}', "");
Expect(1, 57343, '\P{^Blk=	Low_SURROGATES}', "");
Expect(0, 57344, '\p{Blk=	Low_SURROGATES}', "");
Expect(1, 57344, '\p{^Blk=	Low_SURROGATES}', "");
Expect(1, 57344, '\P{Blk=	Low_SURROGATES}', "");
Expect(0, 57344, '\P{^Blk=	Low_SURROGATES}', "");
Error('\p{Is_Block=/a/	_Low_Surrogates}');
Error('\P{Is_Block=/a/	_Low_Surrogates}');
Expect(1, 57343, '\p{Is_Block=lowsurrogates}', "");
Expect(0, 57343, '\p{^Is_Block=lowsurrogates}', "");
Expect(0, 57343, '\P{Is_Block=lowsurrogates}', "");
Expect(1, 57343, '\P{^Is_Block=lowsurrogates}', "");
Expect(0, 57344, '\p{Is_Block=lowsurrogates}', "");
Expect(1, 57344, '\p{^Is_Block=lowsurrogates}', "");
Expect(1, 57344, '\P{Is_Block=lowsurrogates}', "");
Expect(0, 57344, '\P{^Is_Block=lowsurrogates}', "");
Expect(1, 57343, '\p{Is_Block= _Low_surrogates}', "");
Expect(0, 57343, '\p{^Is_Block= _Low_surrogates}', "");
Expect(0, 57343, '\P{Is_Block= _Low_surrogates}', "");
Expect(1, 57343, '\P{^Is_Block= _Low_surrogates}', "");
Expect(0, 57344, '\p{Is_Block= _Low_surrogates}', "");
Expect(1, 57344, '\p{^Is_Block= _Low_surrogates}', "");
Expect(1, 57344, '\P{Is_Block= _Low_surrogates}', "");
Expect(0, 57344, '\P{^Is_Block= _Low_surrogates}', "");
Error('\p{Is_Blk= :=Low_SURROGATES}');
Error('\P{Is_Blk= :=Low_SURROGATES}');
Expect(1, 57343, '\p{Is_Blk=lowsurrogates}', "");
Expect(0, 57343, '\p{^Is_Blk=lowsurrogates}', "");
Expect(0, 57343, '\P{Is_Blk=lowsurrogates}', "");
Expect(1, 57343, '\P{^Is_Blk=lowsurrogates}', "");
Expect(0, 57344, '\p{Is_Blk=lowsurrogates}', "");
Expect(1, 57344, '\p{^Is_Blk=lowsurrogates}', "");
Expect(1, 57344, '\P{Is_Blk=lowsurrogates}', "");
Expect(0, 57344, '\P{^Is_Blk=lowsurrogates}', "");
Expect(1, 57343, '\p{Is_Blk=_-Low_surrogates}', "");
Expect(0, 57343, '\p{^Is_Blk=_-Low_surrogates}', "");
Expect(0, 57343, '\P{Is_Blk=_-Low_surrogates}', "");
Expect(1, 57343, '\P{^Is_Blk=_-Low_surrogates}', "");
Expect(0, 57344, '\p{Is_Blk=_-Low_surrogates}', "");
Expect(1, 57344, '\p{^Is_Blk=_-Low_surrogates}', "");
Expect(1, 57344, '\P{Is_Blk=_-Low_surrogates}', "");
Expect(0, 57344, '\P{^Is_Blk=_-Low_surrogates}', "");
Error('\p{Block=:=	MEROITIC_Cursive}');
Error('\P{Block=:=	MEROITIC_Cursive}');
Expect(1, 68095, '\p{Block:meroiticcursive}', "");
Expect(0, 68095, '\p{^Block:meroiticcursive}', "");
Expect(0, 68095, '\P{Block:meroiticcursive}', "");
Expect(1, 68095, '\P{^Block:meroiticcursive}', "");
Expect(0, 68096, '\p{Block:meroiticcursive}', "");
Expect(1, 68096, '\p{^Block:meroiticcursive}', "");
Expect(1, 68096, '\P{Block:meroiticcursive}', "");
Expect(0, 68096, '\P{^Block:meroiticcursive}', "");
Expect(1, 68095, '\p{Block=meroitic_CURSIVE}', "");
Expect(0, 68095, '\p{^Block=meroitic_CURSIVE}', "");
Expect(0, 68095, '\P{Block=meroitic_CURSIVE}', "");
Expect(1, 68095, '\P{^Block=meroitic_CURSIVE}', "");
Expect(0, 68096, '\p{Block=meroitic_CURSIVE}', "");
Expect(1, 68096, '\p{^Block=meroitic_CURSIVE}', "");
Expect(1, 68096, '\P{Block=meroitic_CURSIVE}', "");
Expect(0, 68096, '\P{^Block=meroitic_CURSIVE}', "");
Error('\p{Blk=/a/	-MEROITIC_CURSIVE}');
Error('\P{Blk=/a/	-MEROITIC_CURSIVE}');
Expect(1, 68095, '\p{Blk=meroiticcursive}', "");
Expect(0, 68095, '\p{^Blk=meroiticcursive}', "");
Expect(0, 68095, '\P{Blk=meroiticcursive}', "");
Expect(1, 68095, '\P{^Blk=meroiticcursive}', "");
Expect(0, 68096, '\p{Blk=meroiticcursive}', "");
Expect(1, 68096, '\p{^Blk=meroiticcursive}', "");
Expect(1, 68096, '\P{Blk=meroiticcursive}', "");
Expect(0, 68096, '\P{^Blk=meroiticcursive}', "");
Expect(1, 68095, '\p{Blk:	 MEROITIC_Cursive}', "");
Expect(0, 68095, '\p{^Blk:	 MEROITIC_Cursive}', "");
Expect(0, 68095, '\P{Blk:	 MEROITIC_Cursive}', "");
Expect(1, 68095, '\P{^Blk:	 MEROITIC_Cursive}', "");
Expect(0, 68096, '\p{Blk:	 MEROITIC_Cursive}', "");
Expect(1, 68096, '\p{^Blk:	 MEROITIC_Cursive}', "");
Expect(1, 68096, '\P{Blk:	 MEROITIC_Cursive}', "");
Expect(0, 68096, '\P{^Blk:	 MEROITIC_Cursive}', "");
Error('\p{Is_Block=meroitic_CURSIVE:=}');
Error('\P{Is_Block=meroitic_CURSIVE:=}');
Expect(1, 68095, '\p{Is_Block=meroiticcursive}', "");
Expect(0, 68095, '\p{^Is_Block=meroiticcursive}', "");
Expect(0, 68095, '\P{Is_Block=meroiticcursive}', "");
Expect(1, 68095, '\P{^Is_Block=meroiticcursive}', "");
Expect(0, 68096, '\p{Is_Block=meroiticcursive}', "");
Expect(1, 68096, '\p{^Is_Block=meroiticcursive}', "");
Expect(1, 68096, '\P{Is_Block=meroiticcursive}', "");
Expect(0, 68096, '\P{^Is_Block=meroiticcursive}', "");
Expect(1, 68095, '\p{Is_Block=	_Meroitic_Cursive}', "");
Expect(0, 68095, '\p{^Is_Block=	_Meroitic_Cursive}', "");
Expect(0, 68095, '\P{Is_Block=	_Meroitic_Cursive}', "");
Expect(1, 68095, '\P{^Is_Block=	_Meroitic_Cursive}', "");
Expect(0, 68096, '\p{Is_Block=	_Meroitic_Cursive}', "");
Expect(1, 68096, '\p{^Is_Block=	_Meroitic_Cursive}', "");
Expect(1, 68096, '\P{Is_Block=	_Meroitic_Cursive}', "");
Expect(0, 68096, '\P{^Is_Block=	_Meroitic_Cursive}', "");
Error('\p{Is_Blk=/a/meroitic_Cursive}');
Error('\P{Is_Blk=/a/meroitic_Cursive}');
Expect(1, 68095, '\p{Is_Blk=meroiticcursive}', "");
Expect(0, 68095, '\p{^Is_Blk=meroiticcursive}', "");
Expect(0, 68095, '\P{Is_Blk=meroiticcursive}', "");
Expect(1, 68095, '\P{^Is_Blk=meroiticcursive}', "");
Expect(0, 68096, '\p{Is_Blk=meroiticcursive}', "");
Expect(1, 68096, '\p{^Is_Blk=meroiticcursive}', "");
Expect(1, 68096, '\P{Is_Blk=meroiticcursive}', "");
Expect(0, 68096, '\P{^Is_Blk=meroiticcursive}', "");
Expect(1, 68095, '\p{Is_Blk=	 MEROITIC_Cursive}', "");
Expect(0, 68095, '\p{^Is_Blk=	 MEROITIC_Cursive}', "");
Expect(0, 68095, '\P{Is_Blk=	 MEROITIC_Cursive}', "");
Expect(1, 68095, '\P{^Is_Blk=	 MEROITIC_Cursive}', "");
Expect(0, 68096, '\p{Is_Blk=	 MEROITIC_Cursive}', "");
Expect(1, 68096, '\p{^Is_Blk=	 MEROITIC_Cursive}', "");
Expect(1, 68096, '\P{Is_Blk=	 MEROITIC_Cursive}', "");
Expect(0, 68096, '\P{^Is_Blk=	 MEROITIC_Cursive}', "");
Error('\p{Block=/a/ 	cuneiform}');
Error('\P{Block=/a/ 	cuneiform}');
Expect(1, 74751, '\p{Block=cuneiform}', "");
Expect(0, 74751, '\p{^Block=cuneiform}', "");
Expect(0, 74751, '\P{Block=cuneiform}', "");
Expect(1, 74751, '\P{^Block=cuneiform}', "");
Expect(0, 74752, '\p{Block=cuneiform}', "");
Expect(1, 74752, '\p{^Block=cuneiform}', "");
Expect(1, 74752, '\P{Block=cuneiform}', "");
Expect(0, 74752, '\P{^Block=cuneiform}', "");
Expect(1, 74751, '\p{Block=		cuneiform}', "");
Expect(0, 74751, '\p{^Block=		cuneiform}', "");
Expect(0, 74751, '\P{Block=		cuneiform}', "");
Expect(1, 74751, '\P{^Block=		cuneiform}', "");
Expect(0, 74752, '\p{Block=		cuneiform}', "");
Expect(1, 74752, '\p{^Block=		cuneiform}', "");
Expect(1, 74752, '\P{Block=		cuneiform}', "");
Expect(0, 74752, '\P{^Block=		cuneiform}', "");
Error('\p{Blk:	  Cuneiform/a/}');
Error('\P{Blk:	  Cuneiform/a/}');
Expect(1, 74751, '\p{Blk=cuneiform}', "");
Expect(0, 74751, '\p{^Blk=cuneiform}', "");
Expect(0, 74751, '\P{Blk=cuneiform}', "");
Expect(1, 74751, '\P{^Blk=cuneiform}', "");
Expect(0, 74752, '\p{Blk=cuneiform}', "");
Expect(1, 74752, '\p{^Blk=cuneiform}', "");
Expect(1, 74752, '\P{Blk=cuneiform}', "");
Expect(0, 74752, '\P{^Blk=cuneiform}', "");
Expect(1, 74751, '\p{Blk=-	cuneiform}', "");
Expect(0, 74751, '\p{^Blk=-	cuneiform}', "");
Expect(0, 74751, '\P{Blk=-	cuneiform}', "");
Expect(1, 74751, '\P{^Blk=-	cuneiform}', "");
Expect(0, 74752, '\p{Blk=-	cuneiform}', "");
Expect(1, 74752, '\p{^Blk=-	cuneiform}', "");
Expect(1, 74752, '\P{Blk=-	cuneiform}', "");
Expect(0, 74752, '\P{^Blk=-	cuneiform}', "");
Error('\p{Is_Block=	Cuneiform/a/}');
Error('\P{Is_Block=	Cuneiform/a/}');
Expect(1, 74751, '\p{Is_Block:cuneiform}', "");
Expect(0, 74751, '\p{^Is_Block:cuneiform}', "");
Expect(0, 74751, '\P{Is_Block:cuneiform}', "");
Expect(1, 74751, '\P{^Is_Block:cuneiform}', "");
Expect(0, 74752, '\p{Is_Block:cuneiform}', "");
Expect(1, 74752, '\p{^Is_Block:cuneiform}', "");
Expect(1, 74752, '\P{Is_Block:cuneiform}', "");
Expect(0, 74752, '\P{^Is_Block:cuneiform}', "");
Expect(1, 74751, '\p{Is_Block=		Cuneiform}', "");
Expect(0, 74751, '\p{^Is_Block=		Cuneiform}', "");
Expect(0, 74751, '\P{Is_Block=		Cuneiform}', "");
Expect(1, 74751, '\P{^Is_Block=		Cuneiform}', "");
Expect(0, 74752, '\p{Is_Block=		Cuneiform}', "");
Expect(1, 74752, '\p{^Is_Block=		Cuneiform}', "");
Expect(1, 74752, '\P{Is_Block=		Cuneiform}', "");
Expect(0, 74752, '\P{^Is_Block=		Cuneiform}', "");
Error('\p{Is_Blk=-/a/cuneiform}');
Error('\P{Is_Blk=-/a/cuneiform}');
Expect(1, 74751, '\p{Is_Blk=cuneiform}', "");
Expect(0, 74751, '\p{^Is_Blk=cuneiform}', "");
Expect(0, 74751, '\P{Is_Blk=cuneiform}', "");
Expect(1, 74751, '\P{^Is_Blk=cuneiform}', "");
Expect(0, 74752, '\p{Is_Blk=cuneiform}', "");
Expect(1, 74752, '\p{^Is_Blk=cuneiform}', "");
Expect(1, 74752, '\P{Is_Blk=cuneiform}', "");
Expect(0, 74752, '\P{^Is_Blk=cuneiform}', "");
Expect(1, 74751, '\p{Is_Blk=_	CUNEIFORM}', "");
Expect(0, 74751, '\p{^Is_Blk=_	CUNEIFORM}', "");
Expect(0, 74751, '\P{Is_Blk=_	CUNEIFORM}', "");
Expect(1, 74751, '\P{^Is_Blk=_	CUNEIFORM}', "");
Expect(0, 74752, '\p{Is_Blk=_	CUNEIFORM}', "");
Expect(1, 74752, '\p{^Is_Blk=_	CUNEIFORM}', "");
Expect(1, 74752, '\P{Is_Blk=_	CUNEIFORM}', "");
Expect(0, 74752, '\P{^Is_Blk=_	CUNEIFORM}', "");
Error('\p{Block=/a/VARIATION_Selectors_SUPPLEMENT}');
Error('\P{Block=/a/VARIATION_Selectors_SUPPLEMENT}');
Expect(1, 917999, '\p{Block:	variationselectorssupplement}', "");
Expect(0, 917999, '\p{^Block:	variationselectorssupplement}', "");
Expect(0, 917999, '\P{Block:	variationselectorssupplement}', "");
Expect(1, 917999, '\P{^Block:	variationselectorssupplement}', "");
Expect(0, 918000, '\p{Block:	variationselectorssupplement}', "");
Expect(1, 918000, '\p{^Block:	variationselectorssupplement}', "");
Expect(1, 918000, '\P{Block:	variationselectorssupplement}', "");
Expect(0, 918000, '\P{^Block:	variationselectorssupplement}', "");
Expect(1, 917999, '\p{Block=- variation_selectors_SUPPLEMENT}', "");
Expect(0, 917999, '\p{^Block=- variation_selectors_SUPPLEMENT}', "");
Expect(0, 917999, '\P{Block=- variation_selectors_SUPPLEMENT}', "");
Expect(1, 917999, '\P{^Block=- variation_selectors_SUPPLEMENT}', "");
Expect(0, 918000, '\p{Block=- variation_selectors_SUPPLEMENT}', "");
Expect(1, 918000, '\p{^Block=- variation_selectors_SUPPLEMENT}', "");
Expect(1, 918000, '\P{Block=- variation_selectors_SUPPLEMENT}', "");
Expect(0, 918000, '\P{^Block=- variation_selectors_SUPPLEMENT}', "");
Error('\p{Blk=-/a/vs_Sup}');
Error('\P{Blk=-/a/vs_Sup}');
Expect(1, 917999, '\p{Blk=vssup}', "");
Expect(0, 917999, '\p{^Blk=vssup}', "");
Expect(0, 917999, '\P{Blk=vssup}', "");
Expect(1, 917999, '\P{^Blk=vssup}', "");
Expect(0, 918000, '\p{Blk=vssup}', "");
Expect(1, 918000, '\p{^Blk=vssup}', "");
Expect(1, 918000, '\P{Blk=vssup}', "");
Expect(0, 918000, '\P{^Blk=vssup}', "");
Expect(1, 917999, '\p{Blk=	vs_Sup}', "");
Expect(0, 917999, '\p{^Blk=	vs_Sup}', "");
Expect(0, 917999, '\P{Blk=	vs_Sup}', "");
Expect(1, 917999, '\P{^Blk=	vs_Sup}', "");
Expect(0, 918000, '\p{Blk=	vs_Sup}', "");
Expect(1, 918000, '\p{^Blk=	vs_Sup}', "");
Expect(1, 918000, '\P{Blk=	vs_Sup}', "");
Expect(0, 918000, '\P{^Blk=	vs_Sup}', "");
Error('\p{Is_Block=:=		Variation_selectors_SUPPLEMENT}');
Error('\P{Is_Block=:=		Variation_selectors_SUPPLEMENT}');
Expect(1, 917999, '\p{Is_Block=variationselectorssupplement}', "");
Expect(0, 917999, '\p{^Is_Block=variationselectorssupplement}', "");
Expect(0, 917999, '\P{Is_Block=variationselectorssupplement}', "");
Expect(1, 917999, '\P{^Is_Block=variationselectorssupplement}', "");
Expect(0, 918000, '\p{Is_Block=variationselectorssupplement}', "");
Expect(1, 918000, '\p{^Is_Block=variationselectorssupplement}', "");
Expect(1, 918000, '\P{Is_Block=variationselectorssupplement}', "");
Expect(0, 918000, '\P{^Is_Block=variationselectorssupplement}', "");
Expect(1, 917999, '\p{Is_Block=	Variation_Selectors_supplement}', "");
Expect(0, 917999, '\p{^Is_Block=	Variation_Selectors_supplement}', "");
Expect(0, 917999, '\P{Is_Block=	Variation_Selectors_supplement}', "");
Expect(1, 917999, '\P{^Is_Block=	Variation_Selectors_supplement}', "");
Expect(0, 918000, '\p{Is_Block=	Variation_Selectors_supplement}', "");
Expect(1, 918000, '\p{^Is_Block=	Variation_Selectors_supplement}', "");
Expect(1, 918000, '\P{Is_Block=	Variation_Selectors_supplement}', "");
Expect(0, 918000, '\P{^Is_Block=	Variation_Selectors_supplement}', "");
Error('\p{Is_Blk=	:=VS_SUP}');
Error('\P{Is_Blk=	:=VS_SUP}');
Expect(1, 917999, '\p{Is_Blk=vssup}', "");
Expect(0, 917999, '\p{^Is_Blk=vssup}', "");
Expect(0, 917999, '\P{Is_Blk=vssup}', "");
Expect(1, 917999, '\P{^Is_Blk=vssup}', "");
Expect(0, 918000, '\p{Is_Blk=vssup}', "");
Expect(1, 918000, '\p{^Is_Blk=vssup}', "");
Expect(1, 918000, '\P{Is_Blk=vssup}', "");
Expect(0, 918000, '\P{^Is_Blk=vssup}', "");
Expect(1, 917999, '\p{Is_Blk=		vs_Sup}', "");
Expect(0, 917999, '\p{^Is_Blk=		vs_Sup}', "");
Expect(0, 917999, '\P{Is_Blk=		vs_Sup}', "");
Expect(1, 917999, '\P{^Is_Blk=		vs_Sup}', "");
Expect(0, 918000, '\p{Is_Blk=		vs_Sup}', "");
Expect(1, 918000, '\p{^Is_Blk=		vs_Sup}', "");
Expect(1, 918000, '\P{Is_Blk=		vs_Sup}', "");
Expect(0, 918000, '\P{^Is_Blk=		vs_Sup}', "");
Error('\p{Block=_:=LATIN_Extended_D}');
Error('\P{Block=_:=LATIN_Extended_D}');
Expect(1, 43007, '\p{Block=latinextendedd}', "");
Expect(0, 43007, '\p{^Block=latinextendedd}', "");
Expect(0, 43007, '\P{Block=latinextendedd}', "");
Expect(1, 43007, '\P{^Block=latinextendedd}', "");
Expect(0, 43008, '\p{Block=latinextendedd}', "");
Expect(1, 43008, '\p{^Block=latinextendedd}', "");
Expect(1, 43008, '\P{Block=latinextendedd}', "");
Expect(0, 43008, '\P{^Block=latinextendedd}', "");
Expect(1, 43007, '\p{Block=_LATIN_Extended_D}', "");
Expect(0, 43007, '\p{^Block=_LATIN_Extended_D}', "");
Expect(0, 43007, '\P{Block=_LATIN_Extended_D}', "");
Expect(1, 43007, '\P{^Block=_LATIN_Extended_D}', "");
Expect(0, 43008, '\p{Block=_LATIN_Extended_D}', "");
Expect(1, 43008, '\p{^Block=_LATIN_Extended_D}', "");
Expect(1, 43008, '\P{Block=_LATIN_Extended_D}', "");
Expect(0, 43008, '\P{^Block=_LATIN_Extended_D}', "");
Error('\p{Blk=/a/	LATIN_Ext_D}');
Error('\P{Blk=/a/	LATIN_Ext_D}');
Expect(1, 43007, '\p{Blk=latinextd}', "");
Expect(0, 43007, '\p{^Blk=latinextd}', "");
Expect(0, 43007, '\P{Blk=latinextd}', "");
Expect(1, 43007, '\P{^Blk=latinextd}', "");
Expect(0, 43008, '\p{Blk=latinextd}', "");
Expect(1, 43008, '\p{^Blk=latinextd}', "");
Expect(1, 43008, '\P{Blk=latinextd}', "");
Expect(0, 43008, '\P{^Blk=latinextd}', "");
Expect(1, 43007, '\p{Blk=--latin_EXT_D}', "");
Expect(0, 43007, '\p{^Blk=--latin_EXT_D}', "");
Expect(0, 43007, '\P{Blk=--latin_EXT_D}', "");
Expect(1, 43007, '\P{^Blk=--latin_EXT_D}', "");
Expect(0, 43008, '\p{Blk=--latin_EXT_D}', "");
Expect(1, 43008, '\p{^Blk=--latin_EXT_D}', "");
Expect(1, 43008, '\P{Blk=--latin_EXT_D}', "");
Expect(0, 43008, '\P{^Blk=--latin_EXT_D}', "");
Error('\p{Is_Block=:=--latin_EXTENDED_d}');
Error('\P{Is_Block=:=--latin_EXTENDED_d}');
Expect(1, 43007, '\p{Is_Block=latinextendedd}', "");
Expect(0, 43007, '\p{^Is_Block=latinextendedd}', "");
Expect(0, 43007, '\P{Is_Block=latinextendedd}', "");
Expect(1, 43007, '\P{^Is_Block=latinextendedd}', "");
Expect(0, 43008, '\p{Is_Block=latinextendedd}', "");
Expect(1, 43008, '\p{^Is_Block=latinextendedd}', "");
Expect(1, 43008, '\P{Is_Block=latinextendedd}', "");
Expect(0, 43008, '\P{^Is_Block=latinextendedd}', "");
Expect(1, 43007, '\p{Is_Block=-LATIN_EXTENDED_D}', "");
Expect(0, 43007, '\p{^Is_Block=-LATIN_EXTENDED_D}', "");
Expect(0, 43007, '\P{Is_Block=-LATIN_EXTENDED_D}', "");
Expect(1, 43007, '\P{^Is_Block=-LATIN_EXTENDED_D}', "");
Expect(0, 43008, '\p{Is_Block=-LATIN_EXTENDED_D}', "");
Expect(1, 43008, '\p{^Is_Block=-LATIN_EXTENDED_D}', "");
Expect(1, 43008, '\P{Is_Block=-LATIN_EXTENDED_D}', "");
Expect(0, 43008, '\P{^Is_Block=-LATIN_EXTENDED_D}', "");
Error('\p{Is_Blk=	/a/Latin_EXT_D}');
Error('\P{Is_Blk=	/a/Latin_EXT_D}');
Expect(1, 43007, '\p{Is_Blk=latinextd}', "");
Expect(0, 43007, '\p{^Is_Blk=latinextd}', "");
Expect(0, 43007, '\P{Is_Blk=latinextd}', "");
Expect(1, 43007, '\P{^Is_Blk=latinextd}', "");
Expect(0, 43008, '\p{Is_Blk=latinextd}', "");
Expect(1, 43008, '\p{^Is_Blk=latinextd}', "");
Expect(1, 43008, '\P{Is_Blk=latinextd}', "");
Expect(0, 43008, '\P{^Is_Blk=latinextd}', "");
Expect(1, 43007, '\p{Is_Blk:	_Latin_Ext_D}', "");
Expect(0, 43007, '\p{^Is_Blk:	_Latin_Ext_D}', "");
Expect(0, 43007, '\P{Is_Blk:	_Latin_Ext_D}', "");
Expect(1, 43007, '\P{^Is_Blk:	_Latin_Ext_D}', "");
Expect(0, 43008, '\p{Is_Blk:	_Latin_Ext_D}', "");
Expect(1, 43008, '\p{^Is_Blk:	_Latin_Ext_D}', "");
Expect(1, 43008, '\P{Is_Blk:	_Latin_Ext_D}', "");
Expect(0, 43008, '\P{^Is_Blk:	_Latin_Ext_D}', "");
Error('\p{Block=:=_Combining_Half_Marks}');
Error('\P{Block=:=_Combining_Half_Marks}');
Expect(1, 65071, '\p{Block=combininghalfmarks}', "");
Expect(0, 65071, '\p{^Block=combininghalfmarks}', "");
Expect(0, 65071, '\P{Block=combininghalfmarks}', "");
Expect(1, 65071, '\P{^Block=combininghalfmarks}', "");
Expect(0, 65072, '\p{Block=combininghalfmarks}', "");
Expect(1, 65072, '\p{^Block=combininghalfmarks}', "");
Expect(1, 65072, '\P{Block=combininghalfmarks}', "");
Expect(0, 65072, '\P{^Block=combininghalfmarks}', "");
Expect(1, 65071, '\p{Block=_Combining_HALF_MARKS}', "");
Expect(0, 65071, '\p{^Block=_Combining_HALF_MARKS}', "");
Expect(0, 65071, '\P{Block=_Combining_HALF_MARKS}', "");
Expect(1, 65071, '\P{^Block=_Combining_HALF_MARKS}', "");
Expect(0, 65072, '\p{Block=_Combining_HALF_MARKS}', "");
Expect(1, 65072, '\p{^Block=_Combining_HALF_MARKS}', "");
Expect(1, 65072, '\P{Block=_Combining_HALF_MARKS}', "");
Expect(0, 65072, '\P{^Block=_Combining_HALF_MARKS}', "");
Error('\p{Blk=-/a/Half_Marks}');
Error('\P{Blk=-/a/Half_Marks}');
Expect(1, 65071, '\p{Blk=halfmarks}', "");
Expect(0, 65071, '\p{^Blk=halfmarks}', "");
Expect(0, 65071, '\P{Blk=halfmarks}', "");
Expect(1, 65071, '\P{^Blk=halfmarks}', "");
Expect(0, 65072, '\p{Blk=halfmarks}', "");
Expect(1, 65072, '\p{^Blk=halfmarks}', "");
Expect(1, 65072, '\P{Blk=halfmarks}', "");
Expect(0, 65072, '\P{^Blk=halfmarks}', "");
Expect(1, 65071, '\p{Blk=-HALF_Marks}', "");
Expect(0, 65071, '\p{^Blk=-HALF_Marks}', "");
Expect(0, 65071, '\P{Blk=-HALF_Marks}', "");
Expect(1, 65071, '\P{^Blk=-HALF_Marks}', "");
Expect(0, 65072, '\p{Blk=-HALF_Marks}', "");
Expect(1, 65072, '\p{^Blk=-HALF_Marks}', "");
Expect(1, 65072, '\P{Blk=-HALF_Marks}', "");
Expect(0, 65072, '\P{^Blk=-HALF_Marks}', "");
Error('\p{Is_Block=_ COMBINING_Half_Marks:=}');
Error('\P{Is_Block=_ COMBINING_Half_Marks:=}');
Expect(1, 65071, '\p{Is_Block=combininghalfmarks}', "");
Expect(0, 65071, '\p{^Is_Block=combininghalfmarks}', "");
Expect(0, 65071, '\P{Is_Block=combininghalfmarks}', "");
Expect(1, 65071, '\P{^Is_Block=combininghalfmarks}', "");
Expect(0, 65072, '\p{Is_Block=combininghalfmarks}', "");
Expect(1, 65072, '\p{^Is_Block=combininghalfmarks}', "");
Expect(1, 65072, '\P{Is_Block=combininghalfmarks}', "");
Expect(0, 65072, '\P{^Is_Block=combininghalfmarks}', "");
Expect(1, 65071, '\p{Is_Block=_Combining_Half_Marks}', "");
Expect(0, 65071, '\p{^Is_Block=_Combining_Half_Marks}', "");
Expect(0, 65071, '\P{Is_Block=_Combining_Half_Marks}', "");
Expect(1, 65071, '\P{^Is_Block=_Combining_Half_Marks}', "");
Expect(0, 65072, '\p{Is_Block=_Combining_Half_Marks}', "");
Expect(1, 65072, '\p{^Is_Block=_Combining_Half_Marks}', "");
Expect(1, 65072, '\P{Is_Block=_Combining_Half_Marks}', "");
Expect(0, 65072, '\P{^Is_Block=_Combining_Half_Marks}', "");
Error('\p{Is_Blk=-	Half_MARKS:=}');
Error('\P{Is_Blk=-	Half_MARKS:=}');
Expect(1, 65071, '\p{Is_Blk=halfmarks}', "");
Expect(0, 65071, '\p{^Is_Blk=halfmarks}', "");
Expect(0, 65071, '\P{Is_Blk=halfmarks}', "");
Expect(1, 65071, '\P{^Is_Blk=halfmarks}', "");
Expect(0, 65072, '\p{Is_Blk=halfmarks}', "");
Expect(1, 65072, '\p{^Is_Blk=halfmarks}', "");
Expect(1, 65072, '\P{Is_Blk=halfmarks}', "");
Expect(0, 65072, '\P{^Is_Blk=halfmarks}', "");
Expect(1, 65071, '\p{Is_Blk=	-half_Marks}', "");
Expect(0, 65071, '\p{^Is_Blk=	-half_Marks}', "");
Expect(0, 65071, '\P{Is_Blk=	-half_Marks}', "");
Expect(1, 65071, '\P{^Is_Blk=	-half_Marks}', "");
Expect(0, 65072, '\p{Is_Blk=	-half_Marks}', "");
Expect(1, 65072, '\p{^Is_Blk=	-half_Marks}', "");
Expect(1, 65072, '\P{Is_Blk=	-half_Marks}', "");
Expect(0, 65072, '\P{^Is_Blk=	-half_Marks}', "");
Error('\p{Block= /a/Currency_symbols}');
Error('\P{Block= /a/Currency_symbols}');
Expect(1, 8399, '\p{Block=currencysymbols}', "");
Expect(0, 8399, '\p{^Block=currencysymbols}', "");
Expect(0, 8399, '\P{Block=currencysymbols}', "");
Expect(1, 8399, '\P{^Block=currencysymbols}', "");
Expect(0, 8400, '\p{Block=currencysymbols}', "");
Expect(1, 8400, '\p{^Block=currencysymbols}', "");
Expect(1, 8400, '\P{Block=currencysymbols}', "");
Expect(0, 8400, '\P{^Block=currencysymbols}', "");
Expect(1, 8399, '\p{Block=		CURRENCY_SYMBOLS}', "");
Expect(0, 8399, '\p{^Block=		CURRENCY_SYMBOLS}', "");
Expect(0, 8399, '\P{Block=		CURRENCY_SYMBOLS}', "");
Expect(1, 8399, '\P{^Block=		CURRENCY_SYMBOLS}', "");
Expect(0, 8400, '\p{Block=		CURRENCY_SYMBOLS}', "");
Expect(1, 8400, '\p{^Block=		CURRENCY_SYMBOLS}', "");
Expect(1, 8400, '\P{Block=		CURRENCY_SYMBOLS}', "");
Expect(0, 8400, '\P{^Block=		CURRENCY_SYMBOLS}', "");
Error('\p{Blk: /a/  currency_Symbols}');
Error('\P{Blk: /a/  currency_Symbols}');
Expect(1, 8399, '\p{Blk=currencysymbols}', "");
Expect(0, 8399, '\p{^Blk=currencysymbols}', "");
Expect(0, 8399, '\P{Blk=currencysymbols}', "");
Expect(1, 8399, '\P{^Blk=currencysymbols}', "");
Expect(0, 8400, '\p{Blk=currencysymbols}', "");
Expect(1, 8400, '\p{^Blk=currencysymbols}', "");
Expect(1, 8400, '\P{Blk=currencysymbols}', "");
Expect(0, 8400, '\P{^Blk=currencysymbols}', "");
Expect(1, 8399, '\p{Blk:  Currency_SYMBOLS}', "");
Expect(0, 8399, '\p{^Blk:  Currency_SYMBOLS}', "");
Expect(0, 8399, '\P{Blk:  Currency_SYMBOLS}', "");
Expect(1, 8399, '\P{^Blk:  Currency_SYMBOLS}', "");
Expect(0, 8400, '\p{Blk:  Currency_SYMBOLS}', "");
Expect(1, 8400, '\p{^Blk:  Currency_SYMBOLS}', "");
Expect(1, 8400, '\P{Blk:  Currency_SYMBOLS}', "");
Expect(0, 8400, '\P{^Blk:  Currency_SYMBOLS}', "");
Error('\p{Is_Block=/a/-	CURRENCY_Symbols}');
Error('\P{Is_Block=/a/-	CURRENCY_Symbols}');
Expect(1, 8399, '\p{Is_Block=currencysymbols}', "");
Expect(0, 8399, '\p{^Is_Block=currencysymbols}', "");
Expect(0, 8399, '\P{Is_Block=currencysymbols}', "");
Expect(1, 8399, '\P{^Is_Block=currencysymbols}', "");
Expect(0, 8400, '\p{Is_Block=currencysymbols}', "");
Expect(1, 8400, '\p{^Is_Block=currencysymbols}', "");
Expect(1, 8400, '\P{Is_Block=currencysymbols}', "");
Expect(0, 8400, '\P{^Is_Block=currencysymbols}', "");
Expect(1, 8399, '\p{Is_Block=	 CURRENCY_Symbols}', "");
Expect(0, 8399, '\p{^Is_Block=	 CURRENCY_Symbols}', "");
Expect(0, 8399, '\P{Is_Block=	 CURRENCY_Symbols}', "");
Expect(1, 8399, '\P{^Is_Block=	 CURRENCY_Symbols}', "");
Expect(0, 8400, '\p{Is_Block=	 CURRENCY_Symbols}', "");
Expect(1, 8400, '\p{^Is_Block=	 CURRENCY_Symbols}', "");
Expect(1, 8400, '\P{Is_Block=	 CURRENCY_Symbols}', "");
Expect(0, 8400, '\P{^Is_Block=	 CURRENCY_Symbols}', "");
Error('\p{Is_Blk=/a/_currency_SYMBOLS}');
Error('\P{Is_Blk=/a/_currency_SYMBOLS}');
Expect(1, 8399, '\p{Is_Blk=currencysymbols}', "");
Expect(0, 8399, '\p{^Is_Blk=currencysymbols}', "");
Expect(0, 8399, '\P{Is_Blk=currencysymbols}', "");
Expect(1, 8399, '\P{^Is_Blk=currencysymbols}', "");
Expect(0, 8400, '\p{Is_Blk=currencysymbols}', "");
Expect(1, 8400, '\p{^Is_Blk=currencysymbols}', "");
Expect(1, 8400, '\P{Is_Blk=currencysymbols}', "");
Expect(0, 8400, '\P{^Is_Blk=currencysymbols}', "");
Expect(1, 8399, '\p{Is_Blk=- currency_Symbols}', "");
Expect(0, 8399, '\p{^Is_Blk=- currency_Symbols}', "");
Expect(0, 8399, '\P{Is_Blk=- currency_Symbols}', "");
Expect(1, 8399, '\P{^Is_Blk=- currency_Symbols}', "");
Expect(0, 8400, '\p{Is_Blk=- currency_Symbols}', "");
Expect(1, 8400, '\p{^Is_Blk=- currency_Symbols}', "");
Expect(1, 8400, '\P{Is_Blk=- currency_Symbols}', "");
Expect(0, 8400, '\P{^Is_Blk=- currency_Symbols}', "");
Error('\p{Block= /a/Sinhala}');
Error('\P{Block= /a/Sinhala}');
Expect(1, 3583, '\p{Block=sinhala}', "");
Expect(0, 3583, '\p{^Block=sinhala}', "");
Expect(0, 3583, '\P{Block=sinhala}', "");
Expect(1, 3583, '\P{^Block=sinhala}', "");
Expect(0, 3584, '\p{Block=sinhala}', "");
Expect(1, 3584, '\p{^Block=sinhala}', "");
Expect(1, 3584, '\P{Block=sinhala}', "");
Expect(0, 3584, '\P{^Block=sinhala}', "");
Expect(1, 3583, '\p{Block=	 SINHALA}', "");
Expect(0, 3583, '\p{^Block=	 SINHALA}', "");
Expect(0, 3583, '\P{Block=	 SINHALA}', "");
Expect(1, 3583, '\P{^Block=	 SINHALA}', "");
Expect(0, 3584, '\p{Block=	 SINHALA}', "");
Expect(1, 3584, '\p{^Block=	 SINHALA}', "");
Expect(1, 3584, '\P{Block=	 SINHALA}', "");
Expect(0, 3584, '\P{^Block=	 SINHALA}', "");
Error('\p{Blk=_:=SINHALA}');
Error('\P{Blk=_:=SINHALA}');
Expect(1, 3583, '\p{Blk=sinhala}', "");
Expect(0, 3583, '\p{^Blk=sinhala}', "");
Expect(0, 3583, '\P{Blk=sinhala}', "");
Expect(1, 3583, '\P{^Blk=sinhala}', "");
Expect(0, 3584, '\p{Blk=sinhala}', "");
Expect(1, 3584, '\p{^Blk=sinhala}', "");
Expect(1, 3584, '\P{Blk=sinhala}', "");
Expect(0, 3584, '\P{^Blk=sinhala}', "");
Expect(1, 3583, '\p{Blk=	Sinhala}', "");
Expect(0, 3583, '\p{^Blk=	Sinhala}', "");
Expect(0, 3583, '\P{Blk=	Sinhala}', "");
Expect(1, 3583, '\P{^Blk=	Sinhala}', "");
Expect(0, 3584, '\p{Blk=	Sinhala}', "");
Expect(1, 3584, '\p{^Blk=	Sinhala}', "");
Expect(1, 3584, '\P{Blk=	Sinhala}', "");
Expect(0, 3584, '\P{^Blk=	Sinhala}', "");
Error('\p{Is_Block= 	Sinhala/a/}');
Error('\P{Is_Block= 	Sinhala/a/}');
Expect(1, 3583, '\p{Is_Block=sinhala}', "");
Expect(0, 3583, '\p{^Is_Block=sinhala}', "");
Expect(0, 3583, '\P{Is_Block=sinhala}', "");
Expect(1, 3583, '\P{^Is_Block=sinhala}', "");
Expect(0, 3584, '\p{Is_Block=sinhala}', "");
Expect(1, 3584, '\p{^Is_Block=sinhala}', "");
Expect(1, 3584, '\P{Is_Block=sinhala}', "");
Expect(0, 3584, '\P{^Is_Block=sinhala}', "");
Expect(1, 3583, '\p{Is_Block= SINHALA}', "");
Expect(0, 3583, '\p{^Is_Block= SINHALA}', "");
Expect(0, 3583, '\P{Is_Block= SINHALA}', "");
Expect(1, 3583, '\P{^Is_Block= SINHALA}', "");
Expect(0, 3584, '\p{Is_Block= SINHALA}', "");
Expect(1, 3584, '\p{^Is_Block= SINHALA}', "");
Expect(1, 3584, '\P{Is_Block= SINHALA}', "");
Expect(0, 3584, '\P{^Is_Block= SINHALA}', "");
Error('\p{Is_Blk=/a/-Sinhala}');
Error('\P{Is_Blk=/a/-Sinhala}');
Expect(1, 3583, '\p{Is_Blk=sinhala}', "");
Expect(0, 3583, '\p{^Is_Blk=sinhala}', "");
Expect(0, 3583, '\P{Is_Blk=sinhala}', "");
Expect(1, 3583, '\P{^Is_Blk=sinhala}', "");
Expect(0, 3584, '\p{Is_Blk=sinhala}', "");
Expect(1, 3584, '\p{^Is_Blk=sinhala}', "");
Expect(1, 3584, '\P{Is_Blk=sinhala}', "");
Expect(0, 3584, '\P{^Is_Blk=sinhala}', "");
Expect(1, 3583, '\p{Is_Blk=		sinhala}', "");
Expect(0, 3583, '\p{^Is_Blk=		sinhala}', "");
Expect(0, 3583, '\P{Is_Blk=		sinhala}', "");
Expect(1, 3583, '\P{^Is_Blk=		sinhala}', "");
Expect(0, 3584, '\p{Is_Blk=		sinhala}', "");
Expect(1, 3584, '\p{^Is_Blk=		sinhala}', "");
Expect(1, 3584, '\P{Is_Blk=		sinhala}', "");
Expect(0, 3584, '\P{^Is_Blk=		sinhala}', "");
Error('\p{Block=:=	imperial_Aramaic}');
Error('\P{Block=:=	imperial_Aramaic}');
Expect(1, 67679, '\p{Block=imperialaramaic}', "");
Expect(0, 67679, '\p{^Block=imperialaramaic}', "");
Expect(0, 67679, '\P{Block=imperialaramaic}', "");
Expect(1, 67679, '\P{^Block=imperialaramaic}', "");
Expect(0, 67680, '\p{Block=imperialaramaic}', "");
Expect(1, 67680, '\p{^Block=imperialaramaic}', "");
Expect(1, 67680, '\P{Block=imperialaramaic}', "");
Expect(0, 67680, '\P{^Block=imperialaramaic}', "");
Expect(1, 67679, '\p{Block= IMPERIAL_Aramaic}', "");
Expect(0, 67679, '\p{^Block= IMPERIAL_Aramaic}', "");
Expect(0, 67679, '\P{Block= IMPERIAL_Aramaic}', "");
Expect(1, 67679, '\P{^Block= IMPERIAL_Aramaic}', "");
Expect(0, 67680, '\p{Block= IMPERIAL_Aramaic}', "");
Expect(1, 67680, '\p{^Block= IMPERIAL_Aramaic}', "");
Expect(1, 67680, '\P{Block= IMPERIAL_Aramaic}', "");
Expect(0, 67680, '\P{^Block= IMPERIAL_Aramaic}', "");
Error('\p{Blk=	/a/IMPERIAL_Aramaic}');
Error('\P{Blk=	/a/IMPERIAL_Aramaic}');
Expect(1, 67679, '\p{Blk:   imperialaramaic}', "");
Expect(0, 67679, '\p{^Blk:   imperialaramaic}', "");
Expect(0, 67679, '\P{Blk:   imperialaramaic}', "");
Expect(1, 67679, '\P{^Blk:   imperialaramaic}', "");
Expect(0, 67680, '\p{Blk:   imperialaramaic}', "");
Expect(1, 67680, '\p{^Blk:   imperialaramaic}', "");
Expect(1, 67680, '\P{Blk:   imperialaramaic}', "");
Expect(0, 67680, '\P{^Blk:   imperialaramaic}', "");
Expect(1, 67679, '\p{Blk= Imperial_aramaic}', "");
Expect(0, 67679, '\p{^Blk= Imperial_aramaic}', "");
Expect(0, 67679, '\P{Blk= Imperial_aramaic}', "");
Expect(1, 67679, '\P{^Blk= Imperial_aramaic}', "");
Expect(0, 67680, '\p{Blk= Imperial_aramaic}', "");
Expect(1, 67680, '\p{^Blk= Imperial_aramaic}', "");
Expect(1, 67680, '\P{Blk= Imperial_aramaic}', "");
Expect(0, 67680, '\P{^Blk= Imperial_aramaic}', "");
Error('\p{Is_Block=:=-	IMPERIAL_ARAMAIC}');
Error('\P{Is_Block=:=-	IMPERIAL_ARAMAIC}');
Expect(1, 67679, '\p{Is_Block=imperialaramaic}', "");
Expect(0, 67679, '\p{^Is_Block=imperialaramaic}', "");
Expect(0, 67679, '\P{Is_Block=imperialaramaic}', "");
Expect(1, 67679, '\P{^Is_Block=imperialaramaic}', "");
Expect(0, 67680, '\p{Is_Block=imperialaramaic}', "");
Expect(1, 67680, '\p{^Is_Block=imperialaramaic}', "");
Expect(1, 67680, '\P{Is_Block=imperialaramaic}', "");
Expect(0, 67680, '\P{^Is_Block=imperialaramaic}', "");
Expect(1, 67679, '\p{Is_Block: imperial_aramaic}', "");
Expect(0, 67679, '\p{^Is_Block: imperial_aramaic}', "");
Expect(0, 67679, '\P{Is_Block: imperial_aramaic}', "");
Expect(1, 67679, '\P{^Is_Block: imperial_aramaic}', "");
Expect(0, 67680, '\p{Is_Block: imperial_aramaic}', "");
Expect(1, 67680, '\p{^Is_Block: imperial_aramaic}', "");
Expect(1, 67680, '\P{Is_Block: imperial_aramaic}', "");
Expect(0, 67680, '\P{^Is_Block: imperial_aramaic}', "");
Error('\p{Is_Blk=/a/	Imperial_Aramaic}');
Error('\P{Is_Blk=/a/	Imperial_Aramaic}');
Expect(1, 67679, '\p{Is_Blk=imperialaramaic}', "");
Expect(0, 67679, '\p{^Is_Blk=imperialaramaic}', "");
Expect(0, 67679, '\P{Is_Blk=imperialaramaic}', "");
Expect(1, 67679, '\P{^Is_Blk=imperialaramaic}', "");
Expect(0, 67680, '\p{Is_Blk=imperialaramaic}', "");
Expect(1, 67680, '\p{^Is_Blk=imperialaramaic}', "");
Expect(1, 67680, '\P{Is_Blk=imperialaramaic}', "");
Expect(0, 67680, '\P{^Is_Blk=imperialaramaic}', "");
Expect(1, 67679, '\p{Is_Blk:-_Imperial_aramaic}', "");
Expect(0, 67679, '\p{^Is_Blk:-_Imperial_aramaic}', "");
Expect(0, 67679, '\P{Is_Blk:-_Imperial_aramaic}', "");
Expect(1, 67679, '\P{^Is_Blk:-_Imperial_aramaic}', "");
Expect(0, 67680, '\p{Is_Blk:-_Imperial_aramaic}', "");
Expect(1, 67680, '\p{^Is_Blk:-_Imperial_aramaic}', "");
Expect(1, 67680, '\P{Is_Blk:-_Imperial_aramaic}', "");
Expect(0, 67680, '\P{^Is_Blk:-_Imperial_aramaic}', "");
Error('\p{Block=/a/ GLAGOLITIC}');
Error('\P{Block=/a/ GLAGOLITIC}');
Expect(1, 11359, '\p{Block=glagolitic}', "");
Expect(0, 11359, '\p{^Block=glagolitic}', "");
Expect(0, 11359, '\P{Block=glagolitic}', "");
Expect(1, 11359, '\P{^Block=glagolitic}', "");
Expect(0, 11360, '\p{Block=glagolitic}', "");
Expect(1, 11360, '\p{^Block=glagolitic}', "");
Expect(1, 11360, '\P{Block=glagolitic}', "");
Expect(0, 11360, '\P{^Block=glagolitic}', "");
Expect(1, 11359, '\p{Block:_	glagolitic}', "");
Expect(0, 11359, '\p{^Block:_	glagolitic}', "");
Expect(0, 11359, '\P{Block:_	glagolitic}', "");
Expect(1, 11359, '\P{^Block:_	glagolitic}', "");
Expect(0, 11360, '\p{Block:_	glagolitic}', "");
Expect(1, 11360, '\p{^Block:_	glagolitic}', "");
Expect(1, 11360, '\P{Block:_	glagolitic}', "");
Expect(0, 11360, '\P{^Block:_	glagolitic}', "");
Error('\p{Blk=:=_	GLAGOLITIC}');
Error('\P{Blk=:=_	GLAGOLITIC}');
Expect(1, 11359, '\p{Blk=glagolitic}', "");
Expect(0, 11359, '\p{^Blk=glagolitic}', "");
Expect(0, 11359, '\P{Blk=glagolitic}', "");
Expect(1, 11359, '\P{^Blk=glagolitic}', "");
Expect(0, 11360, '\p{Blk=glagolitic}', "");
Expect(1, 11360, '\p{^Blk=glagolitic}', "");
Expect(1, 11360, '\P{Blk=glagolitic}', "");
Expect(0, 11360, '\P{^Blk=glagolitic}', "");
Expect(1, 11359, '\p{Blk:   _Glagolitic}', "");
Expect(0, 11359, '\p{^Blk:   _Glagolitic}', "");
Expect(0, 11359, '\P{Blk:   _Glagolitic}', "");
Expect(1, 11359, '\P{^Blk:   _Glagolitic}', "");
Expect(0, 11360, '\p{Blk:   _Glagolitic}', "");
Expect(1, 11360, '\p{^Blk:   _Glagolitic}', "");
Expect(1, 11360, '\P{Blk:   _Glagolitic}', "");
Expect(0, 11360, '\P{^Blk:   _Glagolitic}', "");
Error('\p{Is_Block=_	Glagolitic/a/}');
Error('\P{Is_Block=_	Glagolitic/a/}');
Expect(1, 11359, '\p{Is_Block=glagolitic}', "");
Expect(0, 11359, '\p{^Is_Block=glagolitic}', "");
Expect(0, 11359, '\P{Is_Block=glagolitic}', "");
Expect(1, 11359, '\P{^Is_Block=glagolitic}', "");
Expect(0, 11360, '\p{Is_Block=glagolitic}', "");
Expect(1, 11360, '\p{^Is_Block=glagolitic}', "");
Expect(1, 11360, '\P{Is_Block=glagolitic}', "");
Expect(0, 11360, '\P{^Is_Block=glagolitic}', "");
Expect(1, 11359, '\p{Is_Block:   	Glagolitic}', "");
Expect(0, 11359, '\p{^Is_Block:   	Glagolitic}', "");
Expect(0, 11359, '\P{Is_Block:   	Glagolitic}', "");
Expect(1, 11359, '\P{^Is_Block:   	Glagolitic}', "");
Expect(0, 11360, '\p{Is_Block:   	Glagolitic}', "");
Expect(1, 11360, '\p{^Is_Block:   	Glagolitic}', "");
Expect(1, 11360, '\P{Is_Block:   	Glagolitic}', "");
Expect(0, 11360, '\P{^Is_Block:   	Glagolitic}', "");
Error('\p{Is_Blk=/a/_ Glagolitic}');
Error('\P{Is_Blk=/a/_ Glagolitic}');
Expect(1, 11359, '\p{Is_Blk:	glagolitic}', "");
Expect(0, 11359, '\p{^Is_Blk:	glagolitic}', "");
Expect(0, 11359, '\P{Is_Blk:	glagolitic}', "");
Expect(1, 11359, '\P{^Is_Blk:	glagolitic}', "");
Expect(0, 11360, '\p{Is_Blk:	glagolitic}', "");
Expect(1, 11360, '\p{^Is_Blk:	glagolitic}', "");
Expect(1, 11360, '\P{Is_Blk:	glagolitic}', "");
Expect(0, 11360, '\P{^Is_Blk:	glagolitic}', "");
Expect(1, 11359, '\p{Is_Blk= 	glagolitic}', "");
Expect(0, 11359, '\p{^Is_Blk= 	glagolitic}', "");
Expect(0, 11359, '\P{Is_Blk= 	glagolitic}', "");
Expect(1, 11359, '\P{^Is_Blk= 	glagolitic}', "");
Expect(0, 11360, '\p{Is_Blk= 	glagolitic}', "");
Expect(1, 11360, '\p{^Is_Blk= 	glagolitic}', "");
Expect(1, 11360, '\P{Is_Blk= 	glagolitic}', "");
Expect(0, 11360, '\P{^Is_Blk= 	glagolitic}', "");
Error('\p{Block=-/a/kharoshthi}');
Error('\P{Block=-/a/kharoshthi}');
Expect(1, 68191, '\p{Block=kharoshthi}', "");
Expect(0, 68191, '\p{^Block=kharoshthi}', "");
Expect(0, 68191, '\P{Block=kharoshthi}', "");
Expect(1, 68191, '\P{^Block=kharoshthi}', "");
Expect(0, 68192, '\p{Block=kharoshthi}', "");
Expect(1, 68192, '\p{^Block=kharoshthi}', "");
Expect(1, 68192, '\P{Block=kharoshthi}', "");
Expect(0, 68192, '\P{^Block=kharoshthi}', "");
Expect(1, 68191, '\p{Block=	 Kharoshthi}', "");
Expect(0, 68191, '\p{^Block=	 Kharoshthi}', "");
Expect(0, 68191, '\P{Block=	 Kharoshthi}', "");
Expect(1, 68191, '\P{^Block=	 Kharoshthi}', "");
Expect(0, 68192, '\p{Block=	 Kharoshthi}', "");
Expect(1, 68192, '\p{^Block=	 Kharoshthi}', "");
Expect(1, 68192, '\P{Block=	 Kharoshthi}', "");
Expect(0, 68192, '\P{^Block=	 Kharoshthi}', "");
Error('\p{Blk=_-KHAROSHTHI/a/}');
Error('\P{Blk=_-KHAROSHTHI/a/}');
Expect(1, 68191, '\p{Blk=kharoshthi}', "");
Expect(0, 68191, '\p{^Blk=kharoshthi}', "");
Expect(0, 68191, '\P{Blk=kharoshthi}', "");
Expect(1, 68191, '\P{^Blk=kharoshthi}', "");
Expect(0, 68192, '\p{Blk=kharoshthi}', "");
Expect(1, 68192, '\p{^Blk=kharoshthi}', "");
Expect(1, 68192, '\P{Blk=kharoshthi}', "");
Expect(0, 68192, '\P{^Blk=kharoshthi}', "");
Expect(1, 68191, '\p{Blk=	Kharoshthi}', "");
Expect(0, 68191, '\p{^Blk=	Kharoshthi}', "");
Expect(0, 68191, '\P{Blk=	Kharoshthi}', "");
Expect(1, 68191, '\P{^Blk=	Kharoshthi}', "");
Expect(0, 68192, '\p{Blk=	Kharoshthi}', "");
Expect(1, 68192, '\p{^Blk=	Kharoshthi}', "");
Expect(1, 68192, '\P{Blk=	Kharoshthi}', "");
Expect(0, 68192, '\P{^Blk=	Kharoshthi}', "");
Error('\p{Is_Block= 	Kharoshthi/a/}');
Error('\P{Is_Block= 	Kharoshthi/a/}');
Expect(1, 68191, '\p{Is_Block=kharoshthi}', "");
Expect(0, 68191, '\p{^Is_Block=kharoshthi}', "");
Expect(0, 68191, '\P{Is_Block=kharoshthi}', "");
Expect(1, 68191, '\P{^Is_Block=kharoshthi}', "");
Expect(0, 68192, '\p{Is_Block=kharoshthi}', "");
Expect(1, 68192, '\p{^Is_Block=kharoshthi}', "");
Expect(1, 68192, '\P{Is_Block=kharoshthi}', "");
Expect(0, 68192, '\P{^Is_Block=kharoshthi}', "");
Expect(1, 68191, '\p{Is_Block=- Kharoshthi}', "");
Expect(0, 68191, '\p{^Is_Block=- Kharoshthi}', "");
Expect(0, 68191, '\P{Is_Block=- Kharoshthi}', "");
Expect(1, 68191, '\P{^Is_Block=- Kharoshthi}', "");
Expect(0, 68192, '\p{Is_Block=- Kharoshthi}', "");
Expect(1, 68192, '\p{^Is_Block=- Kharoshthi}', "");
Expect(1, 68192, '\P{Is_Block=- Kharoshthi}', "");
Expect(0, 68192, '\P{^Is_Block=- Kharoshthi}', "");
Error('\p{Is_Blk=:=Kharoshthi}');
Error('\P{Is_Blk=:=Kharoshthi}');
Expect(1, 68191, '\p{Is_Blk=kharoshthi}', "");
Expect(0, 68191, '\p{^Is_Blk=kharoshthi}', "");
Expect(0, 68191, '\P{Is_Blk=kharoshthi}', "");
Expect(1, 68191, '\P{^Is_Blk=kharoshthi}', "");
Expect(0, 68192, '\p{Is_Blk=kharoshthi}', "");
Expect(1, 68192, '\p{^Is_Blk=kharoshthi}', "");
Expect(1, 68192, '\P{Is_Blk=kharoshthi}', "");
Expect(0, 68192, '\P{^Is_Blk=kharoshthi}', "");
Expect(1, 68191, '\p{Is_Blk=-KHAROSHTHI}', "");
Expect(0, 68191, '\p{^Is_Blk=-KHAROSHTHI}', "");
Expect(0, 68191, '\P{Is_Blk=-KHAROSHTHI}', "");
Expect(1, 68191, '\P{^Is_Blk=-KHAROSHTHI}', "");
Expect(0, 68192, '\p{Is_Blk=-KHAROSHTHI}', "");
Expect(1, 68192, '\p{^Is_Blk=-KHAROSHTHI}', "");
Expect(1, 68192, '\P{Is_Blk=-KHAROSHTHI}', "");
Expect(0, 68192, '\P{^Is_Blk=-KHAROSHTHI}', "");
Error('\p{Block=/a/ _YI_Syllables}');
Error('\P{Block=/a/ _YI_Syllables}');
Expect(1, 42127, '\p{Block:   yisyllables}', "");
Expect(0, 42127, '\p{^Block:   yisyllables}', "");
Expect(0, 42127, '\P{Block:   yisyllables}', "");
Expect(1, 42127, '\P{^Block:   yisyllables}', "");
Expect(0, 42128, '\p{Block:   yisyllables}', "");
Expect(1, 42128, '\p{^Block:   yisyllables}', "");
Expect(1, 42128, '\P{Block:   yisyllables}', "");
Expect(0, 42128, '\P{^Block:   yisyllables}', "");
Expect(1, 42127, '\p{Block=  Yi_Syllables}', "");
Expect(0, 42127, '\p{^Block=  Yi_Syllables}', "");
Expect(0, 42127, '\P{Block=  Yi_Syllables}', "");
Expect(1, 42127, '\P{^Block=  Yi_Syllables}', "");
Expect(0, 42128, '\p{Block=  Yi_Syllables}', "");
Expect(1, 42128, '\p{^Block=  Yi_Syllables}', "");
Expect(1, 42128, '\P{Block=  Yi_Syllables}', "");
Expect(0, 42128, '\P{^Block=  Yi_Syllables}', "");
Error('\p{Blk=/a/_	yi_SYLLABLES}');
Error('\P{Blk=/a/_	yi_SYLLABLES}');
Expect(1, 42127, '\p{Blk=yisyllables}', "");
Expect(0, 42127, '\p{^Blk=yisyllables}', "");
Expect(0, 42127, '\P{Blk=yisyllables}', "");
Expect(1, 42127, '\P{^Blk=yisyllables}', "");
Expect(0, 42128, '\p{Blk=yisyllables}', "");
Expect(1, 42128, '\p{^Blk=yisyllables}', "");
Expect(1, 42128, '\P{Blk=yisyllables}', "");
Expect(0, 42128, '\P{^Blk=yisyllables}', "");
Expect(1, 42127, '\p{Blk=Yi_Syllables}', "");
Expect(0, 42127, '\p{^Blk=Yi_Syllables}', "");
Expect(0, 42127, '\P{Blk=Yi_Syllables}', "");
Expect(1, 42127, '\P{^Blk=Yi_Syllables}', "");
Expect(0, 42128, '\p{Blk=Yi_Syllables}', "");
Expect(1, 42128, '\p{^Blk=Yi_Syllables}', "");
Expect(1, 42128, '\P{Blk=Yi_Syllables}', "");
Expect(0, 42128, '\P{^Blk=Yi_Syllables}', "");
Error('\p{Is_Block=_-YI_Syllables/a/}');
Error('\P{Is_Block=_-YI_Syllables/a/}');
Expect(1, 42127, '\p{Is_Block:	yisyllables}', "");
Expect(0, 42127, '\p{^Is_Block:	yisyllables}', "");
Expect(0, 42127, '\P{Is_Block:	yisyllables}', "");
Expect(1, 42127, '\P{^Is_Block:	yisyllables}', "");
Expect(0, 42128, '\p{Is_Block:	yisyllables}', "");
Expect(1, 42128, '\p{^Is_Block:	yisyllables}', "");
Expect(1, 42128, '\P{Is_Block:	yisyllables}', "");
Expect(0, 42128, '\P{^Is_Block:	yisyllables}', "");
Expect(1, 42127, '\p{Is_Block=_yi_SYLLABLES}', "");
Expect(0, 42127, '\p{^Is_Block=_yi_SYLLABLES}', "");
Expect(0, 42127, '\P{Is_Block=_yi_SYLLABLES}', "");
Expect(1, 42127, '\P{^Is_Block=_yi_SYLLABLES}', "");
Expect(0, 42128, '\p{Is_Block=_yi_SYLLABLES}', "");
Expect(1, 42128, '\p{^Is_Block=_yi_SYLLABLES}', "");
Expect(1, 42128, '\P{Is_Block=_yi_SYLLABLES}', "");
Expect(0, 42128, '\P{^Is_Block=_yi_SYLLABLES}', "");
Error('\p{Is_Blk=/a/	YI_syllables}');
Error('\P{Is_Blk=/a/	YI_syllables}');
Expect(1, 42127, '\p{Is_Blk=yisyllables}', "");
Expect(0, 42127, '\p{^Is_Blk=yisyllables}', "");
Expect(0, 42127, '\P{Is_Blk=yisyllables}', "");
Expect(1, 42127, '\P{^Is_Blk=yisyllables}', "");
Expect(0, 42128, '\p{Is_Blk=yisyllables}', "");
Expect(1, 42128, '\p{^Is_Blk=yisyllables}', "");
Expect(1, 42128, '\P{Is_Blk=yisyllables}', "");
Expect(0, 42128, '\P{^Is_Blk=yisyllables}', "");
Expect(1, 42127, '\p{Is_Blk=	 YI_Syllables}', "");
Expect(0, 42127, '\p{^Is_Blk=	 YI_Syllables}', "");
Expect(0, 42127, '\P{Is_Blk=	 YI_Syllables}', "");
Expect(1, 42127, '\P{^Is_Blk=	 YI_Syllables}', "");
Expect(0, 42128, '\p{Is_Blk=	 YI_Syllables}', "");
Expect(1, 42128, '\p{^Is_Blk=	 YI_Syllables}', "");
Expect(1, 42128, '\P{Is_Blk=	 YI_Syllables}', "");
Expect(0, 42128, '\P{^Is_Blk=	 YI_Syllables}', "");
Error('\p{Block:_:=Combining_Diacritical_marks}');
Error('\P{Block:_:=Combining_Diacritical_marks}');
Expect(1, 879, '\p{Block=combiningdiacriticalmarks}', "");
Expect(0, 879, '\p{^Block=combiningdiacriticalmarks}', "");
Expect(0, 879, '\P{Block=combiningdiacriticalmarks}', "");
Expect(1, 879, '\P{^Block=combiningdiacriticalmarks}', "");
Expect(0, 880, '\p{Block=combiningdiacriticalmarks}', "");
Expect(1, 880, '\p{^Block=combiningdiacriticalmarks}', "");
Expect(1, 880, '\P{Block=combiningdiacriticalmarks}', "");
Expect(0, 880, '\P{^Block=combiningdiacriticalmarks}', "");
Expect(1, 879, '\p{Block=	 Combining_Diacritical_Marks}', "");
Expect(0, 879, '\p{^Block=	 Combining_Diacritical_Marks}', "");
Expect(0, 879, '\P{Block=	 Combining_Diacritical_Marks}', "");
Expect(1, 879, '\P{^Block=	 Combining_Diacritical_Marks}', "");
Expect(0, 880, '\p{Block=	 Combining_Diacritical_Marks}', "");
Expect(1, 880, '\p{^Block=	 Combining_Diacritical_Marks}', "");
Expect(1, 880, '\P{Block=	 Combining_Diacritical_Marks}', "");
Expect(0, 880, '\P{^Block=	 Combining_Diacritical_Marks}', "");
Error('\p{Blk=-/a/DIACRITICALS}');
Error('\P{Blk=-/a/DIACRITICALS}');
Expect(1, 879, '\p{Blk=diacriticals}', "");
Expect(0, 879, '\p{^Blk=diacriticals}', "");
Expect(0, 879, '\P{Blk=diacriticals}', "");
Expect(1, 879, '\P{^Blk=diacriticals}', "");
Expect(0, 880, '\p{Blk=diacriticals}', "");
Expect(1, 880, '\p{^Blk=diacriticals}', "");
Expect(1, 880, '\P{Blk=diacriticals}', "");
Expect(0, 880, '\P{^Blk=diacriticals}', "");
Expect(1, 879, '\p{Blk=	_Diacriticals}', "");
Expect(0, 879, '\p{^Blk=	_Diacriticals}', "");
Expect(0, 879, '\P{Blk=	_Diacriticals}', "");
Expect(1, 879, '\P{^Blk=	_Diacriticals}', "");
Expect(0, 880, '\p{Blk=	_Diacriticals}', "");
Expect(1, 880, '\p{^Blk=	_Diacriticals}', "");
Expect(1, 880, '\P{Blk=	_Diacriticals}', "");
Expect(0, 880, '\P{^Blk=	_Diacriticals}', "");
Error('\p{Is_Block=/a/	Combining_Diacritical_MARKS}');
Error('\P{Is_Block=/a/	Combining_Diacritical_MARKS}');
Expect(1, 879, '\p{Is_Block=combiningdiacriticalmarks}', "");
Expect(0, 879, '\p{^Is_Block=combiningdiacriticalmarks}', "");
Expect(0, 879, '\P{Is_Block=combiningdiacriticalmarks}', "");
Expect(1, 879, '\P{^Is_Block=combiningdiacriticalmarks}', "");
Expect(0, 880, '\p{Is_Block=combiningdiacriticalmarks}', "");
Expect(1, 880, '\p{^Is_Block=combiningdiacriticalmarks}', "");
Expect(1, 880, '\P{Is_Block=combiningdiacriticalmarks}', "");
Expect(0, 880, '\P{^Is_Block=combiningdiacriticalmarks}', "");
Expect(1, 879, '\p{Is_Block=	-COMBINING_Diacritical_Marks}', "");
Expect(0, 879, '\p{^Is_Block=	-COMBINING_Diacritical_Marks}', "");
Expect(0, 879, '\P{Is_Block=	-COMBINING_Diacritical_Marks}', "");
Expect(1, 879, '\P{^Is_Block=	-COMBINING_Diacritical_Marks}', "");
Expect(0, 880, '\p{Is_Block=	-COMBINING_Diacritical_Marks}', "");
Expect(1, 880, '\p{^Is_Block=	-COMBINING_Diacritical_Marks}', "");
Expect(1, 880, '\P{Is_Block=	-COMBINING_Diacritical_Marks}', "");
Expect(0, 880, '\P{^Is_Block=	-COMBINING_Diacritical_Marks}', "");
Error('\p{Is_Blk=/a/  diacriticals}');
Error('\P{Is_Blk=/a/  diacriticals}');
Expect(1, 879, '\p{Is_Blk=diacriticals}', "");
Expect(0, 879, '\p{^Is_Blk=diacriticals}', "");
Expect(0, 879, '\P{Is_Blk=diacriticals}', "");
Expect(1, 879, '\P{^Is_Blk=diacriticals}', "");
Expect(0, 880, '\p{Is_Blk=diacriticals}', "");
Expect(1, 880, '\p{^Is_Blk=diacriticals}', "");
Expect(1, 880, '\P{Is_Blk=diacriticals}', "");
Expect(0, 880, '\P{^Is_Blk=diacriticals}', "");
Expect(1, 879, '\p{Is_Blk=-_Diacriticals}', "");
Expect(0, 879, '\p{^Is_Blk=-_Diacriticals}', "");
Expect(0, 879, '\P{Is_Blk=-_Diacriticals}', "");
Expect(1, 879, '\P{^Is_Blk=-_Diacriticals}', "");
Expect(0, 880, '\p{Is_Blk=-_Diacriticals}', "");
Expect(1, 880, '\p{^Is_Blk=-_Diacriticals}', "");
Expect(1, 880, '\P{Is_Blk=-_Diacriticals}', "");
Expect(0, 880, '\P{^Is_Blk=-_Diacriticals}', "");
Error('\p{Block=:=	 Enclosed_cjk_LETTERS_AND_Months}');
Error('\P{Block=:=	 Enclosed_cjk_LETTERS_AND_Months}');
Expect(1, 13055, '\p{Block=enclosedcjklettersandmonths}', "");
Expect(0, 13055, '\p{^Block=enclosedcjklettersandmonths}', "");
Expect(0, 13055, '\P{Block=enclosedcjklettersandmonths}', "");
Expect(1, 13055, '\P{^Block=enclosedcjklettersandmonths}', "");
Expect(0, 13056, '\p{Block=enclosedcjklettersandmonths}', "");
Expect(1, 13056, '\p{^Block=enclosedcjklettersandmonths}', "");
Expect(1, 13056, '\P{Block=enclosedcjklettersandmonths}', "");
Expect(0, 13056, '\P{^Block=enclosedcjklettersandmonths}', "");
Expect(1, 13055, '\p{Block=_ ENCLOSED_cjk_LETTERS_AND_MONTHS}', "");
Expect(0, 13055, '\p{^Block=_ ENCLOSED_cjk_LETTERS_AND_MONTHS}', "");
Expect(0, 13055, '\P{Block=_ ENCLOSED_cjk_LETTERS_AND_MONTHS}', "");
Expect(1, 13055, '\P{^Block=_ ENCLOSED_cjk_LETTERS_AND_MONTHS}', "");
Expect(0, 13056, '\p{Block=_ ENCLOSED_cjk_LETTERS_AND_MONTHS}', "");
Expect(1, 13056, '\p{^Block=_ ENCLOSED_cjk_LETTERS_AND_MONTHS}', "");
Expect(1, 13056, '\P{Block=_ ENCLOSED_cjk_LETTERS_AND_MONTHS}', "");
Expect(0, 13056, '\P{^Block=_ ENCLOSED_cjk_LETTERS_AND_MONTHS}', "");
Error('\p{Blk=	/a/ENCLOSED_CJK}');
Error('\P{Blk=	/a/ENCLOSED_CJK}');
Expect(1, 13055, '\p{Blk=enclosedcjk}', "");
Expect(0, 13055, '\p{^Blk=enclosedcjk}', "");
Expect(0, 13055, '\P{Blk=enclosedcjk}', "");
Expect(1, 13055, '\P{^Blk=enclosedcjk}', "");
Expect(0, 13056, '\p{Blk=enclosedcjk}', "");
Expect(1, 13056, '\p{^Blk=enclosedcjk}', "");
Expect(1, 13056, '\P{Blk=enclosedcjk}', "");
Expect(0, 13056, '\P{^Blk=enclosedcjk}', "");
Expect(1, 13055, '\p{Blk= Enclosed_cjk}', "");
Expect(0, 13055, '\p{^Blk= Enclosed_cjk}', "");
Expect(0, 13055, '\P{Blk= Enclosed_cjk}', "");
Expect(1, 13055, '\P{^Blk= Enclosed_cjk}', "");
Expect(0, 13056, '\p{Blk= Enclosed_cjk}', "");
Expect(1, 13056, '\p{^Blk= Enclosed_cjk}', "");
Expect(1, 13056, '\P{Blk= Enclosed_cjk}', "");
Expect(0, 13056, '\P{^Blk= Enclosed_cjk}', "");
Error('\p{Is_Block=-/a/Enclosed_CJK_LETTERS_And_Months}');
Error('\P{Is_Block=-/a/Enclosed_CJK_LETTERS_And_Months}');
Expect(1, 13055, '\p{Is_Block=enclosedcjklettersandmonths}', "");
Expect(0, 13055, '\p{^Is_Block=enclosedcjklettersandmonths}', "");
Expect(0, 13055, '\P{Is_Block=enclosedcjklettersandmonths}', "");
Expect(1, 13055, '\P{^Is_Block=enclosedcjklettersandmonths}', "");
Expect(0, 13056, '\p{Is_Block=enclosedcjklettersandmonths}', "");
Expect(1, 13056, '\p{^Is_Block=enclosedcjklettersandmonths}', "");
Expect(1, 13056, '\P{Is_Block=enclosedcjklettersandmonths}', "");
Expect(0, 13056, '\P{^Is_Block=enclosedcjklettersandmonths}', "");
Expect(1, 13055, '\p{Is_Block=__ENCLOSED_cjk_letters_And_Months}', "");
Expect(0, 13055, '\p{^Is_Block=__ENCLOSED_cjk_letters_And_Months}', "");
Expect(0, 13055, '\P{Is_Block=__ENCLOSED_cjk_letters_And_Months}', "");
Expect(1, 13055, '\P{^Is_Block=__ENCLOSED_cjk_letters_And_Months}', "");
Expect(0, 13056, '\p{Is_Block=__ENCLOSED_cjk_letters_And_Months}', "");
Expect(1, 13056, '\p{^Is_Block=__ENCLOSED_cjk_letters_And_Months}', "");
Expect(1, 13056, '\P{Is_Block=__ENCLOSED_cjk_letters_And_Months}', "");
Expect(0, 13056, '\P{^Is_Block=__ENCLOSED_cjk_letters_And_Months}', "");
Error('\p{Is_Blk=:=  ENCLOSED_cjk}');
Error('\P{Is_Blk=:=  ENCLOSED_cjk}');
Expect(1, 13055, '\p{Is_Blk=enclosedcjk}', "");
Expect(0, 13055, '\p{^Is_Blk=enclosedcjk}', "");
Expect(0, 13055, '\P{Is_Blk=enclosedcjk}', "");
Expect(1, 13055, '\P{^Is_Blk=enclosedcjk}', "");
Expect(0, 13056, '\p{Is_Blk=enclosedcjk}', "");
Expect(1, 13056, '\p{^Is_Blk=enclosedcjk}', "");
Expect(1, 13056, '\P{Is_Blk=enclosedcjk}', "");
Expect(0, 13056, '\P{^Is_Blk=enclosedcjk}', "");
Expect(1, 13055, '\p{Is_Blk=Enclosed_cjk}', "");
Expect(0, 13055, '\p{^Is_Blk=Enclosed_cjk}', "");
Expect(0, 13055, '\P{Is_Blk=Enclosed_cjk}', "");
Expect(1, 13055, '\P{^Is_Blk=Enclosed_cjk}', "");
Expect(0, 13056, '\p{Is_Blk=Enclosed_cjk}', "");
Expect(1, 13056, '\p{^Is_Blk=Enclosed_cjk}', "");
Expect(1, 13056, '\P{Is_Blk=Enclosed_cjk}', "");
Expect(0, 13056, '\P{^Is_Blk=Enclosed_cjk}', "");
Error('\p{Block=/a/Miscellaneous_Mathematical_Symbols_b}');
Error('\P{Block=/a/Miscellaneous_Mathematical_Symbols_b}');
Expect(1, 10751, '\p{Block=miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\p{^Block=miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\P{Block=miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\P{^Block=miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\p{Block=miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\p{^Block=miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\P{Block=miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\P{^Block=miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\p{Block=Miscellaneous_mathematical_Symbols_B}', "");
Expect(0, 10751, '\p{^Block=Miscellaneous_mathematical_Symbols_B}', "");
Expect(0, 10751, '\P{Block=Miscellaneous_mathematical_Symbols_B}', "");
Expect(1, 10751, '\P{^Block=Miscellaneous_mathematical_Symbols_B}', "");
Expect(0, 10752, '\p{Block=Miscellaneous_mathematical_Symbols_B}', "");
Expect(1, 10752, '\p{^Block=Miscellaneous_mathematical_Symbols_B}', "");
Expect(1, 10752, '\P{Block=Miscellaneous_mathematical_Symbols_B}', "");
Expect(0, 10752, '\P{^Block=Miscellaneous_mathematical_Symbols_B}', "");
Error('\p{Blk=/a/ 	Misc_Math_symbols_B}');
Error('\P{Blk=/a/ 	Misc_Math_symbols_B}');
Expect(1, 10751, '\p{Blk=miscmathsymbolsb}', "");
Expect(0, 10751, '\p{^Blk=miscmathsymbolsb}', "");
Expect(0, 10751, '\P{Blk=miscmathsymbolsb}', "");
Expect(1, 10751, '\P{^Blk=miscmathsymbolsb}', "");
Expect(0, 10752, '\p{Blk=miscmathsymbolsb}', "");
Expect(1, 10752, '\p{^Blk=miscmathsymbolsb}', "");
Expect(1, 10752, '\P{Blk=miscmathsymbolsb}', "");
Expect(0, 10752, '\P{^Blk=miscmathsymbolsb}', "");
Expect(1, 10751, '\p{Blk=Misc_Math_Symbols_B}', "");
Expect(0, 10751, '\p{^Blk=Misc_Math_Symbols_B}', "");
Expect(0, 10751, '\P{Blk=Misc_Math_Symbols_B}', "");
Expect(1, 10751, '\P{^Blk=Misc_Math_Symbols_B}', "");
Expect(0, 10752, '\p{Blk=Misc_Math_Symbols_B}', "");
Expect(1, 10752, '\p{^Blk=Misc_Math_Symbols_B}', "");
Expect(1, 10752, '\P{Blk=Misc_Math_Symbols_B}', "");
Expect(0, 10752, '\P{^Blk=Misc_Math_Symbols_B}', "");
Error('\p{Is_Block=_miscellaneous_Mathematical_Symbols_b/a/}');
Error('\P{Is_Block=_miscellaneous_Mathematical_Symbols_b/a/}');
Expect(1, 10751, '\p{Is_Block=miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\p{^Is_Block=miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10751, '\P{Is_Block=miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\P{^Is_Block=miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\p{Is_Block=miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\p{^Is_Block=miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10752, '\P{Is_Block=miscellaneousmathematicalsymbolsb}', "");
Expect(0, 10752, '\P{^Is_Block=miscellaneousmathematicalsymbolsb}', "");
Expect(1, 10751, '\p{Is_Block: 	Miscellaneous_MATHEMATICAL_Symbols_b}', "");
Expect(0, 10751, '\p{^Is_Block: 	Miscellaneous_MATHEMATICAL_Symbols_b}', "");
Expect(0, 10751, '\P{Is_Block: 	Miscellaneous_MATHEMATICAL_Symbols_b}', "");
Expect(1, 10751, '\P{^Is_Block: 	Miscellaneous_MATHEMATICAL_Symbols_b}', "");
Expect(0, 10752, '\p{Is_Block: 	Miscellaneous_MATHEMATICAL_Symbols_b}', "");
Expect(1, 10752, '\p{^Is_Block: 	Miscellaneous_MATHEMATICAL_Symbols_b}', "");
Expect(1, 10752, '\P{Is_Block: 	Miscellaneous_MATHEMATICAL_Symbols_b}', "");
Expect(0, 10752, '\P{^Is_Block: 	Miscellaneous_MATHEMATICAL_Symbols_b}', "");
Error('\p{Is_Blk::=-Misc_math_SYMBOLS_b}');
Error('\P{Is_Blk::=-Misc_math_SYMBOLS_b}');
Expect(1, 10751, '\p{Is_Blk=miscmathsymbolsb}', "");
Expect(0, 10751, '\p{^Is_Blk=miscmathsymbolsb}', "");
Expect(0, 10751, '\P{Is_Blk=miscmathsymbolsb}', "");
Expect(1, 10751, '\P{^Is_Blk=miscmathsymbolsb}', "");
Expect(0, 10752, '\p{Is_Blk=miscmathsymbolsb}', "");
Expect(1, 10752, '\p{^Is_Blk=miscmathsymbolsb}', "");
Expect(1, 10752, '\P{Is_Blk=miscmathsymbolsb}', "");
Expect(0, 10752, '\P{^Is_Blk=miscmathsymbolsb}', "");
Expect(1, 10751, '\p{Is_Blk=		Misc_Math_Symbols_B}', "");
Expect(0, 10751, '\p{^Is_Blk=		Misc_Math_Symbols_B}', "");
Expect(0, 10751, '\P{Is_Blk=		Misc_Math_Symbols_B}', "");
Expect(1, 10751, '\P{^Is_Blk=		Misc_Math_Symbols_B}', "");
Expect(0, 10752, '\p{Is_Blk=		Misc_Math_Symbols_B}', "");
Expect(1, 10752, '\p{^Is_Blk=		Misc_Math_Symbols_B}', "");
Expect(1, 10752, '\P{Is_Blk=		Misc_Math_Symbols_B}', "");
Expect(0, 10752, '\P{^Is_Blk=		Misc_Math_Symbols_B}', "");
Error('\p{Block= Arabic_EXTENDED_A:=}');
Error('\P{Block= Arabic_EXTENDED_A:=}');
Expect(1, 2303, '\p{Block=arabicextendeda}', "");
Expect(0, 2303, '\p{^Block=arabicextendeda}', "");
Expect(0, 2303, '\P{Block=arabicextendeda}', "");
Expect(1, 2303, '\P{^Block=arabicextendeda}', "");
Expect(0, 2304, '\p{Block=arabicextendeda}', "");
Expect(1, 2304, '\p{^Block=arabicextendeda}', "");
Expect(1, 2304, '\P{Block=arabicextendeda}', "");
Expect(0, 2304, '\P{^Block=arabicextendeda}', "");
Expect(1, 2303, '\p{Block:   -arabic_Extended_A}', "");
Expect(0, 2303, '\p{^Block:   -arabic_Extended_A}', "");
Expect(0, 2303, '\P{Block:   -arabic_Extended_A}', "");
Expect(1, 2303, '\P{^Block:   -arabic_Extended_A}', "");
Expect(0, 2304, '\p{Block:   -arabic_Extended_A}', "");
Expect(1, 2304, '\p{^Block:   -arabic_Extended_A}', "");
Expect(1, 2304, '\P{Block:   -arabic_Extended_A}', "");
Expect(0, 2304, '\P{^Block:   -arabic_Extended_A}', "");
Error('\p{Blk=_/a/Arabic_EXT_A}');
Error('\P{Blk=_/a/Arabic_EXT_A}');
Expect(1, 2303, '\p{Blk=arabicexta}', "");
Expect(0, 2303, '\p{^Blk=arabicexta}', "");
Expect(0, 2303, '\P{Blk=arabicexta}', "");
Expect(1, 2303, '\P{^Blk=arabicexta}', "");
Expect(0, 2304, '\p{Blk=arabicexta}', "");
Expect(1, 2304, '\p{^Blk=arabicexta}', "");
Expect(1, 2304, '\P{Blk=arabicexta}', "");
Expect(0, 2304, '\P{^Blk=arabicexta}', "");
Expect(1, 2303, '\p{Blk= -Arabic_EXT_a}', "");
Expect(0, 2303, '\p{^Blk= -Arabic_EXT_a}', "");
Expect(0, 2303, '\P{Blk= -Arabic_EXT_a}', "");
Expect(1, 2303, '\P{^Blk= -Arabic_EXT_a}', "");
Expect(0, 2304, '\p{Blk= -Arabic_EXT_a}', "");
Expect(1, 2304, '\p{^Blk= -Arabic_EXT_a}', "");
Expect(1, 2304, '\P{Blk= -Arabic_EXT_a}', "");
Expect(0, 2304, '\P{^Blk= -Arabic_EXT_a}', "");
Error('\p{Is_Block= :=Arabic_Extended_A}');
Error('\P{Is_Block= :=Arabic_Extended_A}');
Expect(1, 2303, '\p{Is_Block: arabicextendeda}', "");
Expect(0, 2303, '\p{^Is_Block: arabicextendeda}', "");
Expect(0, 2303, '\P{Is_Block: arabicextendeda}', "");
Expect(1, 2303, '\P{^Is_Block: arabicextendeda}', "");
Expect(0, 2304, '\p{Is_Block: arabicextendeda}', "");
Expect(1, 2304, '\p{^Is_Block: arabicextendeda}', "");
Expect(1, 2304, '\P{Is_Block: arabicextendeda}', "");
Expect(0, 2304, '\P{^Is_Block: arabicextendeda}', "");
Expect(1, 2303, '\p{Is_Block= -Arabic_extended_A}', "");
Expect(0, 2303, '\p{^Is_Block= -Arabic_extended_A}', "");
Expect(0, 2303, '\P{Is_Block= -Arabic_extended_A}', "");
Expect(1, 2303, '\P{^Is_Block= -Arabic_extended_A}', "");
Expect(0, 2304, '\p{Is_Block= -Arabic_extended_A}', "");
Expect(1, 2304, '\p{^Is_Block= -Arabic_extended_A}', "");
Expect(1, 2304, '\P{Is_Block= -Arabic_extended_A}', "");
Expect(0, 2304, '\P{^Is_Block= -Arabic_extended_A}', "");
Error('\p{Is_Blk=-/a/arabic_Ext_A}');
Error('\P{Is_Blk=-/a/arabic_Ext_A}');
Expect(1, 2303, '\p{Is_Blk=arabicexta}', "");
Expect(0, 2303, '\p{^Is_Blk=arabicexta}', "");
Expect(0, 2303, '\P{Is_Blk=arabicexta}', "");
Expect(1, 2303, '\P{^Is_Blk=arabicexta}', "");
Expect(0, 2304, '\p{Is_Blk=arabicexta}', "");
Expect(1, 2304, '\p{^Is_Blk=arabicexta}', "");
Expect(1, 2304, '\P{Is_Blk=arabicexta}', "");
Expect(0, 2304, '\P{^Is_Blk=arabicexta}', "");
Expect(1, 2303, '\p{Is_Blk=	-arabic_ext_A}', "");
Expect(0, 2303, '\p{^Is_Blk=	-arabic_ext_A}', "");
Expect(0, 2303, '\P{Is_Blk=	-arabic_ext_A}', "");
Expect(1, 2303, '\P{^Is_Blk=	-arabic_ext_A}', "");
Expect(0, 2304, '\p{Is_Blk=	-arabic_ext_A}', "");
Expect(1, 2304, '\p{^Is_Blk=	-arabic_ext_A}', "");
Expect(1, 2304, '\P{Is_Blk=	-arabic_ext_A}', "");
Expect(0, 2304, '\P{^Is_Blk=	-arabic_ext_A}', "");
Error('\p{Block=/a/-Cuneiform_NUMBERS_and_Punctuation}');
Error('\P{Block=/a/-Cuneiform_NUMBERS_and_Punctuation}');
Expect(1, 74879, '\p{Block=cuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\p{^Block=cuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\P{Block=cuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\P{^Block=cuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\p{Block=cuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\p{^Block=cuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\P{Block=cuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\P{^Block=cuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\p{Block=	_Cuneiform_NUMBERS_And_Punctuation}', "");
Expect(0, 74879, '\p{^Block=	_Cuneiform_NUMBERS_And_Punctuation}', "");
Expect(0, 74879, '\P{Block=	_Cuneiform_NUMBERS_And_Punctuation}', "");
Expect(1, 74879, '\P{^Block=	_Cuneiform_NUMBERS_And_Punctuation}', "");
Expect(0, 74880, '\p{Block=	_Cuneiform_NUMBERS_And_Punctuation}', "");
Expect(1, 74880, '\p{^Block=	_Cuneiform_NUMBERS_And_Punctuation}', "");
Expect(1, 74880, '\P{Block=	_Cuneiform_NUMBERS_And_Punctuation}', "");
Expect(0, 74880, '\P{^Block=	_Cuneiform_NUMBERS_And_Punctuation}', "");
Error('\p{Blk=_ cuneiform_Numbers/a/}');
Error('\P{Blk=_ cuneiform_Numbers/a/}');
Expect(1, 74879, '\p{Blk=cuneiformnumbers}', "");
Expect(0, 74879, '\p{^Blk=cuneiformnumbers}', "");
Expect(0, 74879, '\P{Blk=cuneiformnumbers}', "");
Expect(1, 74879, '\P{^Blk=cuneiformnumbers}', "");
Expect(0, 74880, '\p{Blk=cuneiformnumbers}', "");
Expect(1, 74880, '\p{^Blk=cuneiformnumbers}', "");
Expect(1, 74880, '\P{Blk=cuneiformnumbers}', "");
Expect(0, 74880, '\P{^Blk=cuneiformnumbers}', "");
Expect(1, 74879, '\p{Blk=  CUNEIFORM_numbers}', "");
Expect(0, 74879, '\p{^Blk=  CUNEIFORM_numbers}', "");
Expect(0, 74879, '\P{Blk=  CUNEIFORM_numbers}', "");
Expect(1, 74879, '\P{^Blk=  CUNEIFORM_numbers}', "");
Expect(0, 74880, '\p{Blk=  CUNEIFORM_numbers}', "");
Expect(1, 74880, '\p{^Blk=  CUNEIFORM_numbers}', "");
Expect(1, 74880, '\P{Blk=  CUNEIFORM_numbers}', "");
Expect(0, 74880, '\P{^Blk=  CUNEIFORM_numbers}', "");
Error('\p{Is_Block=	Cuneiform_Numbers_And_Punctuation/a/}');
Error('\P{Is_Block=	Cuneiform_Numbers_And_Punctuation/a/}');
Expect(1, 74879, '\p{Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\p{^Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(0, 74879, '\P{Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\P{^Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\p{Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\p{^Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(1, 74880, '\P{Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(0, 74880, '\P{^Is_Block=cuneiformnumbersandpunctuation}', "");
Expect(1, 74879, '\p{Is_Block=  cuneiform_Numbers_And_Punctuation}', "");
Expect(0, 74879, '\p{^Is_Block=  cuneiform_Numbers_And_Punctuation}', "");
Expect(0, 74879, '\P{Is_Block=  cuneiform_Numbers_And_Punctuation}', "");
Expect(1, 74879, '\P{^Is_Block=  cuneiform_Numbers_And_Punctuation}', "");
Expect(0, 74880, '\p{Is_Block=  cuneiform_Numbers_And_Punctuation}', "");
Expect(1, 74880, '\p{^Is_Block=  cuneiform_Numbers_And_Punctuation}', "");
Expect(1, 74880, '\P{Is_Block=  cuneiform_Numbers_And_Punctuation}', "");
Expect(0, 74880, '\P{^Is_Block=  cuneiform_Numbers_And_Punctuation}', "");
Error('\p{Is_Blk=-:=CUNEIFORM_Numbers}');
Error('\P{Is_Blk=-:=CUNEIFORM_Numbers}');
Expect(1, 74879, '\p{Is_Blk=cuneiformnumbers}', "");
Expect(0, 74879, '\p{^Is_Blk=cuneiformnumbers}', "");
Expect(0, 74879, '\P{Is_Blk=cuneiformnumbers}', "");
Expect(1, 74879, '\P{^Is_Blk=cuneiformnumbers}', "");
Expect(0, 74880, '\p{Is_Blk=cuneiformnumbers}', "");
Expect(1, 74880, '\p{^Is_Blk=cuneiformnumbers}', "");
Expect(1, 74880, '\P{Is_Blk=cuneiformnumbers}', "");
Expect(0, 74880, '\P{^Is_Blk=cuneiformnumbers}', "");
Expect(1, 74879, '\p{Is_Blk=--Cuneiform_NUMBERS}', "");
Expect(0, 74879, '\p{^Is_Blk=--Cuneiform_NUMBERS}', "");
Expect(0, 74879, '\P{Is_Blk=--Cuneiform_NUMBERS}', "");
Expect(1, 74879, '\P{^Is_Blk=--Cuneiform_NUMBERS}', "");
Expect(0, 74880, '\p{Is_Blk=--Cuneiform_NUMBERS}', "");
Expect(1, 74880, '\p{^Is_Blk=--Cuneiform_NUMBERS}', "");
Expect(1, 74880, '\P{Is_Blk=--Cuneiform_NUMBERS}', "");
Expect(0, 74880, '\P{^Is_Blk=--Cuneiform_NUMBERS}', "");
Error('\p{Block=:=-	katakana}');
Error('\P{Block=:=-	katakana}');
Expect(1, 12543, '\p{Block=katakana}', "");
Expect(0, 12543, '\p{^Block=katakana}', "");
Expect(0, 12543, '\P{Block=katakana}', "");
Expect(1, 12543, '\P{^Block=katakana}', "");
Expect(0, 12544, '\p{Block=katakana}', "");
Expect(1, 12544, '\p{^Block=katakana}', "");
Expect(1, 12544, '\P{Block=katakana}', "");
Expect(0, 12544, '\P{^Block=katakana}', "");
Expect(1, 12543, '\p{Block=	Katakana}', "");
Expect(0, 12543, '\p{^Block=	Katakana}', "");
Expect(0, 12543, '\P{Block=	Katakana}', "");
Expect(1, 12543, '\P{^Block=	Katakana}', "");
Expect(0, 12544, '\p{Block=	Katakana}', "");
Expect(1, 12544, '\p{^Block=	Katakana}', "");
Expect(1, 12544, '\P{Block=	Katakana}', "");
Expect(0, 12544, '\P{^Block=	Katakana}', "");
Error('\p{Blk=:=		Katakana}');
Error('\P{Blk=:=		Katakana}');
Expect(1, 12543, '\p{Blk:   katakana}', "");
Expect(0, 12543, '\p{^Blk:   katakana}', "");
Expect(0, 12543, '\P{Blk:   katakana}', "");
Expect(1, 12543, '\P{^Blk:   katakana}', "");
Expect(0, 12544, '\p{Blk:   katakana}', "");
Expect(1, 12544, '\p{^Blk:   katakana}', "");
Expect(1, 12544, '\P{Blk:   katakana}', "");
Expect(0, 12544, '\P{^Blk:   katakana}', "");
Expect(1, 12543, '\p{Blk=_KATAKANA}', "");
Expect(0, 12543, '\p{^Blk=_KATAKANA}', "");
Expect(0, 12543, '\P{Blk=_KATAKANA}', "");
Expect(1, 12543, '\P{^Blk=_KATAKANA}', "");
Expect(0, 12544, '\p{Blk=_KATAKANA}', "");
Expect(1, 12544, '\p{^Blk=_KATAKANA}', "");
Expect(1, 12544, '\P{Blk=_KATAKANA}', "");
Expect(0, 12544, '\P{^Blk=_KATAKANA}', "");
Error('\p{Is_Block:	  katakana:=}');
Error('\P{Is_Block:	  katakana:=}');
Expect(1, 12543, '\p{Is_Block=katakana}', "");
Expect(0, 12543, '\p{^Is_Block=katakana}', "");
Expect(0, 12543, '\P{Is_Block=katakana}', "");
Expect(1, 12543, '\P{^Is_Block=katakana}', "");
Expect(0, 12544, '\p{Is_Block=katakana}', "");
Expect(1, 12544, '\p{^Is_Block=katakana}', "");
Expect(1, 12544, '\P{Is_Block=katakana}', "");
Expect(0, 12544, '\P{^Is_Block=katakana}', "");
Expect(1, 12543, '\p{Is_Block=_	KATAKANA}', "");
Expect(0, 12543, '\p{^Is_Block=_	KATAKANA}', "");
Expect(0, 12543, '\P{Is_Block=_	KATAKANA}', "");
Expect(1, 12543, '\P{^Is_Block=_	KATAKANA}', "");
Expect(0, 12544, '\p{Is_Block=_	KATAKANA}', "");
Expect(1, 12544, '\p{^Is_Block=_	KATAKANA}', "");
Expect(1, 12544, '\P{Is_Block=_	KATAKANA}', "");
Expect(0, 12544, '\P{^Is_Block=_	KATAKANA}', "");
Error('\p{Is_Blk=:=--KATAKANA}');
Error('\P{Is_Blk=:=--KATAKANA}');
Expect(1, 12543, '\p{Is_Blk=katakana}', "");
Expect(0, 12543, '\p{^Is_Blk=katakana}', "");
Expect(0, 12543, '\P{Is_Blk=katakana}', "");
Expect(1, 12543, '\P{^Is_Blk=katakana}', "");
Expect(0, 12544, '\p{Is_Blk=katakana}', "");
Expect(1, 12544, '\p{^Is_Blk=katakana}', "");
Expect(1, 12544, '\P{Is_Blk=katakana}', "");
Expect(0, 12544, '\P{^Is_Blk=katakana}', "");
Expect(1, 12543, '\p{Is_Blk=  Katakana}', "");
Expect(0, 12543, '\p{^Is_Blk=  Katakana}', "");
Expect(0, 12543, '\P{Is_Blk=  Katakana}', "");
Expect(1, 12543, '\P{^Is_Blk=  Katakana}', "");
Expect(0, 12544, '\p{Is_Blk=  Katakana}', "");
Expect(1, 12544, '\p{^Is_Blk=  Katakana}', "");
Expect(1, 12544, '\P{Is_Blk=  Katakana}', "");
Expect(0, 12544, '\P{^Is_Blk=  Katakana}', "");
Error('\p{Block=_:=Tags}');
Error('\P{Block=_:=Tags}');
Expect(1, 917631, '\p{Block=tags}', "");
Expect(0, 917631, '\p{^Block=tags}', "");
Expect(0, 917631, '\P{Block=tags}', "");
Expect(1, 917631, '\P{^Block=tags}', "");
Expect(0, 917632, '\p{Block=tags}', "");
Expect(1, 917632, '\p{^Block=tags}', "");
Expect(1, 917632, '\P{Block=tags}', "");
Expect(0, 917632, '\P{^Block=tags}', "");
Expect(1, 917631, '\p{Block=Tags}', "");
Expect(0, 917631, '\p{^Block=Tags}', "");
Expect(0, 917631, '\P{Block=Tags}', "");
Expect(1, 917631, '\P{^Block=Tags}', "");
Expect(0, 917632, '\p{Block=Tags}', "");
Expect(1, 917632, '\p{^Block=Tags}', "");
Expect(1, 917632, '\P{Block=Tags}', "");
Expect(0, 917632, '\P{^Block=Tags}', "");
Error('\p{Blk=:=- TAGS}');
Error('\P{Blk=:=- TAGS}');
Expect(1, 917631, '\p{Blk=tags}', "");
Expect(0, 917631, '\p{^Blk=tags}', "");
Expect(0, 917631, '\P{Blk=tags}', "");
Expect(1, 917631, '\P{^Blk=tags}', "");
Expect(0, 917632, '\p{Blk=tags}', "");
Expect(1, 917632, '\p{^Blk=tags}', "");
Expect(1, 917632, '\P{Blk=tags}', "");
Expect(0, 917632, '\P{^Blk=tags}', "");
Expect(1, 917631, '\p{Blk= tags}', "");
Expect(0, 917631, '\p{^Blk= tags}', "");
Expect(0, 917631, '\P{Blk= tags}', "");
Expect(1, 917631, '\P{^Blk= tags}', "");
Expect(0, 917632, '\p{Blk= tags}', "");
Expect(1, 917632, '\p{^Blk= tags}', "");
Expect(1, 917632, '\P{Blk= tags}', "");
Expect(0, 917632, '\P{^Blk= tags}', "");
Error('\p{Is_Block=	/a/Tags}');
Error('\P{Is_Block=	/a/Tags}');
Expect(1, 917631, '\p{Is_Block=tags}', "");
Expect(0, 917631, '\p{^Is_Block=tags}', "");
Expect(0, 917631, '\P{Is_Block=tags}', "");
Expect(1, 917631, '\P{^Is_Block=tags}', "");
Expect(0, 917632, '\p{Is_Block=tags}', "");
Expect(1, 917632, '\p{^Is_Block=tags}', "");
Expect(1, 917632, '\P{Is_Block=tags}', "");
Expect(0, 917632, '\P{^Is_Block=tags}', "");
Expect(1, 917631, '\p{Is_Block:	_Tags}', "");
Expect(0, 917631, '\p{^Is_Block:	_Tags}', "");
Expect(0, 917631, '\P{Is_Block:	_Tags}', "");
Expect(1, 917631, '\P{^Is_Block:	_Tags}', "");
Expect(0, 917632, '\p{Is_Block:	_Tags}', "");
Expect(1, 917632, '\p{^Is_Block:	_Tags}', "");
Expect(1, 917632, '\P{Is_Block:	_Tags}', "");
Expect(0, 917632, '\P{^Is_Block:	_Tags}', "");
Error('\p{Is_Blk=	TAGS:=}');
Error('\P{Is_Blk=	TAGS:=}');
Expect(1, 917631, '\p{Is_Blk=tags}', "");
Expect(0, 917631, '\p{^Is_Blk=tags}', "");
Expect(0, 917631, '\P{Is_Blk=tags}', "");
Expect(1, 917631, '\P{^Is_Blk=tags}', "");
Expect(0, 917632, '\p{Is_Blk=tags}', "");
Expect(1, 917632, '\p{^Is_Blk=tags}', "");
Expect(1, 917632, '\P{Is_Blk=tags}', "");
Expect(0, 917632, '\P{^Is_Blk=tags}', "");
Expect(1, 917631, '\p{Is_Blk=	tags}', "");
Expect(0, 917631, '\p{^Is_Blk=	tags}', "");
Expect(0, 917631, '\P{Is_Blk=	tags}', "");
Expect(1, 917631, '\P{^Is_Blk=	tags}', "");
Expect(0, 917632, '\p{Is_Blk=	tags}', "");
Expect(1, 917632, '\p{^Is_Blk=	tags}', "");
Expect(1, 917632, '\P{Is_Blk=	tags}', "");
Expect(0, 917632, '\P{^Is_Blk=	tags}', "");
Error('\p{Block=_	MISCELLANEOUS_Mathematical_Symbols_A/a/}');
Error('\P{Block=_	MISCELLANEOUS_Mathematical_Symbols_A/a/}');
Expect(1, 10223, '\p{Block=miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\p{^Block=miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\P{Block=miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\P{^Block=miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\p{Block=miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\p{^Block=miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\P{Block=miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\P{^Block=miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\p{Block=-	Miscellaneous_MATHEMATICAL_Symbols_A}', "");
Expect(0, 10223, '\p{^Block=-	Miscellaneous_MATHEMATICAL_Symbols_A}', "");
Expect(0, 10223, '\P{Block=-	Miscellaneous_MATHEMATICAL_Symbols_A}', "");
Expect(1, 10223, '\P{^Block=-	Miscellaneous_MATHEMATICAL_Symbols_A}', "");
Expect(0, 10224, '\p{Block=-	Miscellaneous_MATHEMATICAL_Symbols_A}', "");
Expect(1, 10224, '\p{^Block=-	Miscellaneous_MATHEMATICAL_Symbols_A}', "");
Expect(1, 10224, '\P{Block=-	Miscellaneous_MATHEMATICAL_Symbols_A}', "");
Expect(0, 10224, '\P{^Block=-	Miscellaneous_MATHEMATICAL_Symbols_A}', "");
Error('\p{Blk:			misc_Math_Symbols_A/a/}');
Error('\P{Blk:			misc_Math_Symbols_A/a/}');
Expect(1, 10223, '\p{Blk:	miscmathsymbolsa}', "");
Expect(0, 10223, '\p{^Blk:	miscmathsymbolsa}', "");
Expect(0, 10223, '\P{Blk:	miscmathsymbolsa}', "");
Expect(1, 10223, '\P{^Blk:	miscmathsymbolsa}', "");
Expect(0, 10224, '\p{Blk:	miscmathsymbolsa}', "");
Expect(1, 10224, '\p{^Blk:	miscmathsymbolsa}', "");
Expect(1, 10224, '\P{Blk:	miscmathsymbolsa}', "");
Expect(0, 10224, '\P{^Blk:	miscmathsymbolsa}', "");
Expect(1, 10223, '\p{Blk=_	Misc_math_Symbols_A}', "");
Expect(0, 10223, '\p{^Blk=_	Misc_math_Symbols_A}', "");
Expect(0, 10223, '\P{Blk=_	Misc_math_Symbols_A}', "");
Expect(1, 10223, '\P{^Blk=_	Misc_math_Symbols_A}', "");
Expect(0, 10224, '\p{Blk=_	Misc_math_Symbols_A}', "");
Expect(1, 10224, '\p{^Blk=_	Misc_math_Symbols_A}', "");
Expect(1, 10224, '\P{Blk=_	Misc_math_Symbols_A}', "");
Expect(0, 10224, '\P{^Blk=_	Misc_math_Symbols_A}', "");
Error('\p{Is_Block=/a/  Miscellaneous_Mathematical_symbols_A}');
Error('\P{Is_Block=/a/  Miscellaneous_Mathematical_symbols_A}');
Expect(1, 10223, '\p{Is_Block: miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\p{^Is_Block: miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10223, '\P{Is_Block: miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\P{^Is_Block: miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\p{Is_Block: miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\p{^Is_Block: miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10224, '\P{Is_Block: miscellaneousmathematicalsymbolsa}', "");
Expect(0, 10224, '\P{^Is_Block: miscellaneousmathematicalsymbolsa}', "");
Expect(1, 10223, '\p{Is_Block=	 MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(0, 10223, '\p{^Is_Block=	 MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(0, 10223, '\P{Is_Block=	 MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(1, 10223, '\P{^Is_Block=	 MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(0, 10224, '\p{Is_Block=	 MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(1, 10224, '\p{^Is_Block=	 MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(1, 10224, '\P{Is_Block=	 MISCELLANEOUS_Mathematical_Symbols_A}', "");
Expect(0, 10224, '\P{^Is_Block=	 MISCELLANEOUS_Mathematical_Symbols_A}', "");
Error('\p{Is_Blk=-/a/Misc_Math_SYMBOLS_A}');
Error('\P{Is_Blk=-/a/Misc_Math_SYMBOLS_A}');
Expect(1, 10223, '\p{Is_Blk=miscmathsymbolsa}', "");
Expect(0, 10223, '\p{^Is_Blk=miscmathsymbolsa}', "");
Expect(0, 10223, '\P{Is_Blk=miscmathsymbolsa}', "");
Expect(1, 10223, '\P{^Is_Blk=miscmathsymbolsa}', "");
Expect(0, 10224, '\p{Is_Blk=miscmathsymbolsa}', "");
Expect(1, 10224, '\p{^Is_Blk=miscmathsymbolsa}', "");
Expect(1, 10224, '\P{Is_Blk=miscmathsymbolsa}', "");
Expect(0, 10224, '\P{^Is_Blk=miscmathsymbolsa}', "");
Expect(1, 10223, '\p{Is_Blk=-_misc_Math_Symbols_A}', "");
Expect(0, 10223, '\p{^Is_Blk=-_misc_Math_Symbols_A}', "");
Expect(0, 10223, '\P{Is_Blk=-_misc_Math_Symbols_A}', "");
Expect(1, 10223, '\P{^Is_Blk=-_misc_Math_Symbols_A}', "");
Expect(0, 10224, '\p{Is_Blk=-_misc_Math_Symbols_A}', "");
Expect(1, 10224, '\p{^Is_Blk=-_misc_Math_Symbols_A}', "");
Expect(1, 10224, '\P{Is_Blk=-_misc_Math_Symbols_A}', "");
Expect(0, 10224, '\P{^Is_Blk=-_misc_Math_Symbols_A}', "");
Error('\p{Block=_cjk_unified_IDEOGRAPHS_Extension_D/a/}');
Error('\P{Block=_cjk_unified_IDEOGRAPHS_Extension_D/a/}');
Expect(1, 178207, '\p{Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\p{^Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\P{Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\P{^Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\p{Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\p{^Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\P{Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\P{^Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\p{Block= CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(0, 178207, '\p{^Block= CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(0, 178207, '\P{Block= CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(1, 178207, '\P{^Block= CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(0, 178208, '\p{Block= CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(1, 178208, '\p{^Block= CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(1, 178208, '\P{Block= CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(0, 178208, '\P{^Block= CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Error('\p{Blk=	CJK_Ext_D/a/}');
Error('\P{Blk=	CJK_Ext_D/a/}');
Expect(1, 178207, '\p{Blk=cjkextd}', "");
Expect(0, 178207, '\p{^Blk=cjkextd}', "");
Expect(0, 178207, '\P{Blk=cjkextd}', "");
Expect(1, 178207, '\P{^Blk=cjkextd}', "");
Expect(0, 178208, '\p{Blk=cjkextd}', "");
Expect(1, 178208, '\p{^Blk=cjkextd}', "");
Expect(1, 178208, '\P{Blk=cjkextd}', "");
Expect(0, 178208, '\P{^Blk=cjkextd}', "");
Expect(1, 178207, '\p{Blk=	cjk_Ext_D}', "");
Expect(0, 178207, '\p{^Blk=	cjk_Ext_D}', "");
Expect(0, 178207, '\P{Blk=	cjk_Ext_D}', "");
Expect(1, 178207, '\P{^Blk=	cjk_Ext_D}', "");
Expect(0, 178208, '\p{Blk=	cjk_Ext_D}', "");
Expect(1, 178208, '\p{^Blk=	cjk_Ext_D}', "");
Expect(1, 178208, '\P{Blk=	cjk_Ext_D}', "");
Expect(0, 178208, '\P{^Blk=	cjk_Ext_D}', "");
Error('\p{Is_Block=	:=cjk_Unified_Ideographs_extension_D}');
Error('\P{Is_Block=	:=cjk_Unified_Ideographs_extension_D}');
Expect(1, 178207, '\p{Is_Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\p{^Is_Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178207, '\P{Is_Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\P{^Is_Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\p{Is_Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\p{^Is_Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178208, '\P{Is_Block=cjkunifiedideographsextensiond}', "");
Expect(0, 178208, '\P{^Is_Block=cjkunifiedideographsextensiond}', "");
Expect(1, 178207, '\p{Is_Block= 	CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(0, 178207, '\p{^Is_Block= 	CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(0, 178207, '\P{Is_Block= 	CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(1, 178207, '\P{^Is_Block= 	CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(0, 178208, '\p{Is_Block= 	CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(1, 178208, '\p{^Is_Block= 	CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(1, 178208, '\P{Is_Block= 	CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Expect(0, 178208, '\P{^Is_Block= 	CJK_Unified_IDEOGRAPHS_Extension_D}', "");
Error('\p{Is_Blk: /a/	CJK_EXT_D}');
Error('\P{Is_Blk: /a/	CJK_EXT_D}');
Expect(1, 178207, '\p{Is_Blk=cjkextd}', "");
Expect(0, 178207, '\p{^Is_Blk=cjkextd}', "");
Expect(0, 178207, '\P{Is_Blk=cjkextd}', "");
Expect(1, 178207, '\P{^Is_Blk=cjkextd}', "");
Expect(0, 178208, '\p{Is_Blk=cjkextd}', "");
Expect(1, 178208, '\p{^Is_Blk=cjkextd}', "");
Expect(1, 178208, '\P{Is_Blk=cjkextd}', "");
Expect(0, 178208, '\P{^Is_Blk=cjkextd}', "");
Expect(1, 178207, '\p{Is_Blk=	CJK_ext_D}', "");
Expect(0, 178207, '\p{^Is_Blk=	CJK_ext_D}', "");
Expect(0, 178207, '\P{Is_Blk=	CJK_ext_D}', "");
Expect(1, 178207, '\P{^Is_Blk=	CJK_ext_D}', "");
Expect(0, 178208, '\p{Is_Blk=	CJK_ext_D}', "");
Expect(1, 178208, '\p{^Is_Blk=	CJK_ext_D}', "");
Expect(1, 178208, '\P{Is_Blk=	CJK_ext_D}', "");
Expect(0, 178208, '\P{^Is_Blk=	CJK_ext_D}', "");
Error('\p{Block=	/a/Letterlike_Symbols}');
Error('\P{Block=	/a/Letterlike_Symbols}');
Expect(1, 8527, '\p{Block=letterlikesymbols}', "");
Expect(0, 8527, '\p{^Block=letterlikesymbols}', "");
Expect(0, 8527, '\P{Block=letterlikesymbols}', "");
Expect(1, 8527, '\P{^Block=letterlikesymbols}', "");
Expect(0, 8528, '\p{Block=letterlikesymbols}', "");
Expect(1, 8528, '\p{^Block=letterlikesymbols}', "");
Expect(1, 8528, '\P{Block=letterlikesymbols}', "");
Expect(0, 8528, '\P{^Block=letterlikesymbols}', "");
Expect(1, 8527, '\p{Block= LETTERLIKE_Symbols}', "");
Expect(0, 8527, '\p{^Block= LETTERLIKE_Symbols}', "");
Expect(0, 8527, '\P{Block= LETTERLIKE_Symbols}', "");
Expect(1, 8527, '\P{^Block= LETTERLIKE_Symbols}', "");
Expect(0, 8528, '\p{Block= LETTERLIKE_Symbols}', "");
Expect(1, 8528, '\p{^Block= LETTERLIKE_Symbols}', "");
Expect(1, 8528, '\P{Block= LETTERLIKE_Symbols}', "");
Expect(0, 8528, '\P{^Block= LETTERLIKE_Symbols}', "");
Error('\p{Blk=:=_	letterlike_Symbols}');
Error('\P{Blk=:=_	letterlike_Symbols}');
Expect(1, 8527, '\p{Blk=letterlikesymbols}', "");
Expect(0, 8527, '\p{^Blk=letterlikesymbols}', "");
Expect(0, 8527, '\P{Blk=letterlikesymbols}', "");
Expect(1, 8527, '\P{^Blk=letterlikesymbols}', "");
Expect(0, 8528, '\p{Blk=letterlikesymbols}', "");
Expect(1, 8528, '\p{^Blk=letterlikesymbols}', "");
Expect(1, 8528, '\P{Blk=letterlikesymbols}', "");
Expect(0, 8528, '\P{^Blk=letterlikesymbols}', "");
Expect(1, 8527, '\p{Blk=		Letterlike_SYMBOLS}', "");
Expect(0, 8527, '\p{^Blk=		Letterlike_SYMBOLS}', "");
Expect(0, 8527, '\P{Blk=		Letterlike_SYMBOLS}', "");
Expect(1, 8527, '\P{^Blk=		Letterlike_SYMBOLS}', "");
Expect(0, 8528, '\p{Blk=		Letterlike_SYMBOLS}', "");
Expect(1, 8528, '\p{^Blk=		Letterlike_SYMBOLS}', "");
Expect(1, 8528, '\P{Blk=		Letterlike_SYMBOLS}', "");
Expect(0, 8528, '\P{^Blk=		Letterlike_SYMBOLS}', "");
Error('\p{Is_Block= _letterlike_Symbols/a/}');
Error('\P{Is_Block= _letterlike_Symbols/a/}');
Expect(1, 8527, '\p{Is_Block=letterlikesymbols}', "");
Expect(0, 8527, '\p{^Is_Block=letterlikesymbols}', "");
Expect(0, 8527, '\P{Is_Block=letterlikesymbols}', "");
Expect(1, 8527, '\P{^Is_Block=letterlikesymbols}', "");
Expect(0, 8528, '\p{Is_Block=letterlikesymbols}', "");
Expect(1, 8528, '\p{^Is_Block=letterlikesymbols}', "");
Expect(1, 8528, '\P{Is_Block=letterlikesymbols}', "");
Expect(0, 8528, '\P{^Is_Block=letterlikesymbols}', "");
Expect(1, 8527, '\p{Is_Block=		Letterlike_symbols}', "");
Expect(0, 8527, '\p{^Is_Block=		Letterlike_symbols}', "");
Expect(0, 8527, '\P{Is_Block=		Letterlike_symbols}', "");
Expect(1, 8527, '\P{^Is_Block=		Letterlike_symbols}', "");
Expect(0, 8528, '\p{Is_Block=		Letterlike_symbols}', "");
Expect(1, 8528, '\p{^Is_Block=		Letterlike_symbols}', "");
Expect(1, 8528, '\P{Is_Block=		Letterlike_symbols}', "");
Expect(0, 8528, '\P{^Is_Block=		Letterlike_symbols}', "");
Error('\p{Is_Blk=	-LETTERLIKE_SYMBOLS:=}');
Error('\P{Is_Blk=	-LETTERLIKE_SYMBOLS:=}');
Expect(1, 8527, '\p{Is_Blk=letterlikesymbols}', "");
Expect(0, 8527, '\p{^Is_Blk=letterlikesymbols}', "");
Expect(0, 8527, '\P{Is_Blk=letterlikesymbols}', "");
Expect(1, 8527, '\P{^Is_Blk=letterlikesymbols}', "");
Expect(0, 8528, '\p{Is_Blk=letterlikesymbols}', "");
Expect(1, 8528, '\p{^Is_Blk=letterlikesymbols}', "");
Expect(1, 8528, '\P{Is_Blk=letterlikesymbols}', "");
Expect(0, 8528, '\P{^Is_Blk=letterlikesymbols}', "");
Expect(1, 8527, '\p{Is_Blk=-LETTERLIKE_Symbols}', "");
Expect(0, 8527, '\p{^Is_Blk=-LETTERLIKE_Symbols}', "");
Expect(0, 8527, '\P{Is_Blk=-LETTERLIKE_Symbols}', "");
Expect(1, 8527, '\P{^Is_Blk=-LETTERLIKE_Symbols}', "");
Expect(0, 8528, '\p{Is_Blk=-LETTERLIKE_Symbols}', "");
Expect(1, 8528, '\p{^Is_Blk=-LETTERLIKE_Symbols}', "");
Expect(1, 8528, '\P{Is_Blk=-LETTERLIKE_Symbols}', "");
Expect(0, 8528, '\P{^Is_Blk=-LETTERLIKE_Symbols}', "");
Error('\p{Block=:=	Linear_B_IDEOGRAMS}');
Error('\P{Block=:=	Linear_B_IDEOGRAMS}');
Expect(1, 65791, '\p{Block=linearbideograms}', "");
Expect(0, 65791, '\p{^Block=linearbideograms}', "");
Expect(0, 65791, '\P{Block=linearbideograms}', "");
Expect(1, 65791, '\P{^Block=linearbideograms}', "");
Expect(0, 65792, '\p{Block=linearbideograms}', "");
Expect(1, 65792, '\p{^Block=linearbideograms}', "");
Expect(1, 65792, '\P{Block=linearbideograms}', "");
Expect(0, 65792, '\P{^Block=linearbideograms}', "");
Expect(1, 65791, '\p{Block= -LINEAR_B_Ideograms}', "");
Expect(0, 65791, '\p{^Block= -LINEAR_B_Ideograms}', "");
Expect(0, 65791, '\P{Block= -LINEAR_B_Ideograms}', "");
Expect(1, 65791, '\P{^Block= -LINEAR_B_Ideograms}', "");
Expect(0, 65792, '\p{Block= -LINEAR_B_Ideograms}', "");
Expect(1, 65792, '\p{^Block= -LINEAR_B_Ideograms}', "");
Expect(1, 65792, '\P{Block= -LINEAR_B_Ideograms}', "");
Expect(0, 65792, '\P{^Block= -LINEAR_B_Ideograms}', "");
Error('\p{Blk=_	LINEAR_B_ideograms/a/}');
Error('\P{Blk=_	LINEAR_B_ideograms/a/}');
Expect(1, 65791, '\p{Blk=linearbideograms}', "");
Expect(0, 65791, '\p{^Blk=linearbideograms}', "");
Expect(0, 65791, '\P{Blk=linearbideograms}', "");
Expect(1, 65791, '\P{^Blk=linearbideograms}', "");
Expect(0, 65792, '\p{Blk=linearbideograms}', "");
Expect(1, 65792, '\p{^Blk=linearbideograms}', "");
Expect(1, 65792, '\P{Blk=linearbideograms}', "");
Expect(0, 65792, '\P{^Blk=linearbideograms}', "");
Expect(1, 65791, '\p{Blk=-	Linear_b_IDEOGRAMS}', "");
Expect(0, 65791, '\p{^Blk=-	Linear_b_IDEOGRAMS}', "");
Expect(0, 65791, '\P{Blk=-	Linear_b_IDEOGRAMS}', "");
Expect(1, 65791, '\P{^Blk=-	Linear_b_IDEOGRAMS}', "");
Expect(0, 65792, '\p{Blk=-	Linear_b_IDEOGRAMS}', "");
Expect(1, 65792, '\p{^Blk=-	Linear_b_IDEOGRAMS}', "");
Expect(1, 65792, '\P{Blk=-	Linear_b_IDEOGRAMS}', "");
Expect(0, 65792, '\P{^Blk=-	Linear_b_IDEOGRAMS}', "");
Error('\p{Is_Block= 	LINEAR_B_Ideograms/a/}');
Error('\P{Is_Block= 	LINEAR_B_Ideograms/a/}');
Expect(1, 65791, '\p{Is_Block=linearbideograms}', "");
Expect(0, 65791, '\p{^Is_Block=linearbideograms}', "");
Expect(0, 65791, '\P{Is_Block=linearbideograms}', "");
Expect(1, 65791, '\P{^Is_Block=linearbideograms}', "");
Expect(0, 65792, '\p{Is_Block=linearbideograms}', "");
Expect(1, 65792, '\p{^Is_Block=linearbideograms}', "");
Expect(1, 65792, '\P{Is_Block=linearbideograms}', "");
Expect(0, 65792, '\P{^Is_Block=linearbideograms}', "");
Expect(1, 65791, '\p{Is_Block=- linear_B_ideograms}', "");
Expect(0, 65791, '\p{^Is_Block=- linear_B_ideograms}', "");
Expect(0, 65791, '\P{Is_Block=- linear_B_ideograms}', "");
Expect(1, 65791, '\P{^Is_Block=- linear_B_ideograms}', "");
Expect(0, 65792, '\p{Is_Block=- linear_B_ideograms}', "");
Expect(1, 65792, '\p{^Is_Block=- linear_B_ideograms}', "");
Expect(1, 65792, '\P{Is_Block=- linear_B_ideograms}', "");
Expect(0, 65792, '\P{^Is_Block=- linear_B_ideograms}', "");
Error('\p{Is_Blk=	Linear_B_ideograms:=}');
Error('\P{Is_Blk=	Linear_B_ideograms:=}');
Expect(1, 65791, '\p{Is_Blk=linearbideograms}', "");
Expect(0, 65791, '\p{^Is_Blk=linearbideograms}', "");
Expect(0, 65791, '\P{Is_Blk=linearbideograms}', "");
Expect(1, 65791, '\P{^Is_Blk=linearbideograms}', "");
Expect(0, 65792, '\p{Is_Blk=linearbideograms}', "");
Expect(1, 65792, '\p{^Is_Blk=linearbideograms}', "");
Expect(1, 65792, '\P{Is_Blk=linearbideograms}', "");
Expect(0, 65792, '\P{^Is_Blk=linearbideograms}', "");
Expect(1, 65791, '\p{Is_Blk=_Linear_B_Ideograms}', "");
Expect(0, 65791, '\p{^Is_Blk=_Linear_B_Ideograms}', "");
Expect(0, 65791, '\P{Is_Blk=_Linear_B_Ideograms}', "");
Expect(1, 65791, '\P{^Is_Blk=_Linear_B_Ideograms}', "");
Expect(0, 65792, '\p{Is_Blk=_Linear_B_Ideograms}', "");
Expect(1, 65792, '\p{^Is_Blk=_Linear_B_Ideograms}', "");
Expect(1, 65792, '\P{Is_Blk=_Linear_B_Ideograms}', "");
Expect(0, 65792, '\P{^Is_Blk=_Linear_B_Ideograms}', "");
Error('\p{Block=-:=halfwidth_and_Fullwidth_Forms}');
Error('\P{Block=-:=halfwidth_and_Fullwidth_Forms}');
Expect(1, 65519, '\p{Block:   halfwidthandfullwidthforms}', "");
Expect(0, 65519, '\p{^Block:   halfwidthandfullwidthforms}', "");
Expect(0, 65519, '\P{Block:   halfwidthandfullwidthforms}', "");
Expect(1, 65519, '\P{^Block:   halfwidthandfullwidthforms}', "");
Expect(0, 65520, '\p{Block:   halfwidthandfullwidthforms}', "");
Expect(1, 65520, '\p{^Block:   halfwidthandfullwidthforms}', "");
Expect(1, 65520, '\P{Block:   halfwidthandfullwidthforms}', "");
Expect(0, 65520, '\P{^Block:   halfwidthandfullwidthforms}', "");
Expect(1, 65519, '\p{Block=	halfwidth_AND_Fullwidth_Forms}', "");
Expect(0, 65519, '\p{^Block=	halfwidth_AND_Fullwidth_Forms}', "");
Expect(0, 65519, '\P{Block=	halfwidth_AND_Fullwidth_Forms}', "");
Expect(1, 65519, '\P{^Block=	halfwidth_AND_Fullwidth_Forms}', "");
Expect(0, 65520, '\p{Block=	halfwidth_AND_Fullwidth_Forms}', "");
Expect(1, 65520, '\p{^Block=	halfwidth_AND_Fullwidth_Forms}', "");
Expect(1, 65520, '\P{Block=	halfwidth_AND_Fullwidth_Forms}', "");
Expect(0, 65520, '\P{^Block=	halfwidth_AND_Fullwidth_Forms}', "");
Error('\p{Blk=/a/_-half_and_FULL_FORMS}');
Error('\P{Blk=/a/_-half_and_FULL_FORMS}');
Expect(1, 65519, '\p{Blk:   halfandfullforms}', "");
Expect(0, 65519, '\p{^Blk:   halfandfullforms}', "");
Expect(0, 65519, '\P{Blk:   halfandfullforms}', "");
Expect(1, 65519, '\P{^Blk:   halfandfullforms}', "");
Expect(0, 65520, '\p{Blk:   halfandfullforms}', "");
Expect(1, 65520, '\p{^Blk:   halfandfullforms}', "");
Expect(1, 65520, '\P{Blk:   halfandfullforms}', "");
Expect(0, 65520, '\P{^Blk:   halfandfullforms}', "");
Expect(1, 65519, '\p{Blk= HALF_and_Full_Forms}', "");
Expect(0, 65519, '\p{^Blk= HALF_and_Full_Forms}', "");
Expect(0, 65519, '\P{Blk= HALF_and_Full_Forms}', "");
Expect(1, 65519, '\P{^Blk= HALF_and_Full_Forms}', "");
Expect(0, 65520, '\p{Blk= HALF_and_Full_Forms}', "");
Expect(1, 65520, '\p{^Blk= HALF_and_Full_Forms}', "");
Expect(1, 65520, '\P{Blk= HALF_and_Full_Forms}', "");
Expect(0, 65520, '\P{^Blk= HALF_and_Full_Forms}', "");
Error('\p{Is_Block=_Halfwidth_And_FULLWIDTH_Forms:=}');
Error('\P{Is_Block=_Halfwidth_And_FULLWIDTH_Forms:=}');
Expect(1, 65519, '\p{Is_Block=halfwidthandfullwidthforms}', "");
Expect(0, 65519, '\p{^Is_Block=halfwidthandfullwidthforms}', "");
Expect(0, 65519, '\P{Is_Block=halfwidthandfullwidthforms}', "");
Expect(1, 65519, '\P{^Is_Block=halfwidthandfullwidthforms}', "");
Expect(0, 65520, '\p{Is_Block=halfwidthandfullwidthforms}', "");
Expect(1, 65520, '\p{^Is_Block=halfwidthandfullwidthforms}', "");
Expect(1, 65520, '\P{Is_Block=halfwidthandfullwidthforms}', "");
Expect(0, 65520, '\P{^Is_Block=halfwidthandfullwidthforms}', "");
Expect(1, 65519, '\p{Is_Block=	 HALFWIDTH_And_Fullwidth_Forms}', "");
Expect(0, 65519, '\p{^Is_Block=	 HALFWIDTH_And_Fullwidth_Forms}', "");
Expect(0, 65519, '\P{Is_Block=	 HALFWIDTH_And_Fullwidth_Forms}', "");
Expect(1, 65519, '\P{^Is_Block=	 HALFWIDTH_And_Fullwidth_Forms}', "");
Expect(0, 65520, '\p{Is_Block=	 HALFWIDTH_And_Fullwidth_Forms}', "");
Expect(1, 65520, '\p{^Is_Block=	 HALFWIDTH_And_Fullwidth_Forms}', "");
Expect(1, 65520, '\P{Is_Block=	 HALFWIDTH_And_Fullwidth_Forms}', "");
Expect(0, 65520, '\P{^Is_Block=	 HALFWIDTH_And_Fullwidth_Forms}', "");
Error('\p{Is_Blk=_Half_AND_FULL_Forms/a/}');
Error('\P{Is_Blk=_Half_AND_FULL_Forms/a/}');
Expect(1, 65519, '\p{Is_Blk=halfandfullforms}', "");
Expect(0, 65519, '\p{^Is_Blk=halfandfullforms}', "");
Expect(0, 65519, '\P{Is_Blk=halfandfullforms}', "");
Expect(1, 65519, '\P{^Is_Blk=halfandfullforms}', "");
Expect(0, 65520, '\p{Is_Blk=halfandfullforms}', "");
Expect(1, 65520, '\p{^Is_Blk=halfandfullforms}', "");
Expect(1, 65520, '\P{Is_Blk=halfandfullforms}', "");
Expect(0, 65520, '\P{^Is_Blk=halfandfullforms}', "");
Expect(1, 65519, '\p{Is_Blk=	-Half_AND_Full_FORMS}', "");
Expect(0, 65519, '\p{^Is_Blk=	-Half_AND_Full_FORMS}', "");
Expect(0, 65519, '\P{Is_Blk=	-Half_AND_Full_FORMS}', "");
Expect(1, 65519, '\P{^Is_Blk=	-Half_AND_Full_FORMS}', "");
Expect(0, 65520, '\p{Is_Blk=	-Half_AND_Full_FORMS}', "");
Expect(1, 65520, '\p{^Is_Blk=	-Half_AND_Full_FORMS}', "");
Expect(1, 65520, '\P{Is_Blk=	-Half_AND_Full_FORMS}', "");
Expect(0, 65520, '\P{^Is_Blk=	-Half_AND_Full_FORMS}', "");
Error('\p{Block=/a/- cherokee}');
Error('\P{Block=/a/- cherokee}');
Expect(1, 5119, '\p{Block:   cherokee}', "");
Expect(0, 5119, '\p{^Block:   cherokee}', "");
Expect(0, 5119, '\P{Block:   cherokee}', "");
Expect(1, 5119, '\P{^Block:   cherokee}', "");
Expect(0, 5120, '\p{Block:   cherokee}', "");
Expect(1, 5120, '\p{^Block:   cherokee}', "");
Expect(1, 5120, '\P{Block:   cherokee}', "");
Expect(0, 5120, '\P{^Block:   cherokee}', "");
Expect(1, 5119, '\p{Block=- Cherokee}', "");
Expect(0, 5119, '\p{^Block=- Cherokee}', "");
Expect(0, 5119, '\P{Block=- Cherokee}', "");
Expect(1, 5119, '\P{^Block=- Cherokee}', "");
Expect(0, 5120, '\p{Block=- Cherokee}', "");
Expect(1, 5120, '\p{^Block=- Cherokee}', "");
Expect(1, 5120, '\P{Block=- Cherokee}', "");
Expect(0, 5120, '\P{^Block=- Cherokee}', "");
Error('\p{Blk=:=_Cherokee}');
Error('\P{Blk=:=_Cherokee}');
Expect(1, 5119, '\p{Blk=cherokee}', "");
Expect(0, 5119, '\p{^Blk=cherokee}', "");
Expect(0, 5119, '\P{Blk=cherokee}', "");
Expect(1, 5119, '\P{^Blk=cherokee}', "");
Expect(0, 5120, '\p{Blk=cherokee}', "");
Expect(1, 5120, '\p{^Blk=cherokee}', "");
Expect(1, 5120, '\P{Blk=cherokee}', "");
Expect(0, 5120, '\P{^Blk=cherokee}', "");
Expect(1, 5119, '\p{Blk=_-CHEROKEE}', "");
Expect(0, 5119, '\p{^Blk=_-CHEROKEE}', "");
Expect(0, 5119, '\P{Blk=_-CHEROKEE}', "");
Expect(1, 5119, '\P{^Blk=_-CHEROKEE}', "");
Expect(0, 5120, '\p{Blk=_-CHEROKEE}', "");
Expect(1, 5120, '\p{^Blk=_-CHEROKEE}', "");
Expect(1, 5120, '\P{Blk=_-CHEROKEE}', "");
Expect(0, 5120, '\P{^Blk=_-CHEROKEE}', "");
Error('\p{Is_Block=	 Cherokee:=}');
Error('\P{Is_Block=	 Cherokee:=}');
Expect(1, 5119, '\p{Is_Block=cherokee}', "");
Expect(0, 5119, '\p{^Is_Block=cherokee}', "");
Expect(0, 5119, '\P{Is_Block=cherokee}', "");
Expect(1, 5119, '\P{^Is_Block=cherokee}', "");
Expect(0, 5120, '\p{Is_Block=cherokee}', "");
Expect(1, 5120, '\p{^Is_Block=cherokee}', "");
Expect(1, 5120, '\P{Is_Block=cherokee}', "");
Expect(0, 5120, '\P{^Is_Block=cherokee}', "");
Expect(1, 5119, '\p{Is_Block=_-Cherokee}', "");
Expect(0, 5119, '\p{^Is_Block=_-Cherokee}', "");
Expect(0, 5119, '\P{Is_Block=_-Cherokee}', "");
Expect(1, 5119, '\P{^Is_Block=_-Cherokee}', "");
Expect(0, 5120, '\p{Is_Block=_-Cherokee}', "");
Expect(1, 5120, '\p{^Is_Block=_-Cherokee}', "");
Expect(1, 5120, '\P{Is_Block=_-Cherokee}', "");
Expect(0, 5120, '\P{^Is_Block=_-Cherokee}', "");
Error('\p{Is_Blk=_:=CHEROKEE}');
Error('\P{Is_Blk=_:=CHEROKEE}');
Expect(1, 5119, '\p{Is_Blk=cherokee}', "");
Expect(0, 5119, '\p{^Is_Blk=cherokee}', "");
Expect(0, 5119, '\P{Is_Blk=cherokee}', "");
Expect(1, 5119, '\P{^Is_Blk=cherokee}', "");
Expect(0, 5120, '\p{Is_Blk=cherokee}', "");
Expect(1, 5120, '\p{^Is_Blk=cherokee}', "");
Expect(1, 5120, '\P{Is_Blk=cherokee}', "");
Expect(0, 5120, '\P{^Is_Blk=cherokee}', "");
Expect(1, 5119, '\p{Is_Blk:	-cherokee}', "");
Expect(0, 5119, '\p{^Is_Blk:	-cherokee}', "");
Expect(0, 5119, '\P{Is_Blk:	-cherokee}', "");
Expect(1, 5119, '\P{^Is_Blk:	-cherokee}', "");
Expect(0, 5120, '\p{Is_Blk:	-cherokee}', "");
Expect(1, 5120, '\p{^Is_Blk:	-cherokee}', "");
Expect(1, 5120, '\P{Is_Blk:	-cherokee}', "");
Expect(0, 5120, '\P{^Is_Blk:	-cherokee}', "");
Error('\p{Block=	:=ogham}');
Error('\P{Block=	:=ogham}');
Expect(1, 5791, '\p{Block=ogham}', "");
Expect(0, 5791, '\p{^Block=ogham}', "");
Expect(0, 5791, '\P{Block=ogham}', "");
Expect(1, 5791, '\P{^Block=ogham}', "");
Expect(0, 5792, '\p{Block=ogham}', "");
Expect(1, 5792, '\p{^Block=ogham}', "");
Expect(1, 5792, '\P{Block=ogham}', "");
Expect(0, 5792, '\P{^Block=ogham}', "");
Expect(1, 5791, '\p{Block=-_ogham}', "");
Expect(0, 5791, '\p{^Block=-_ogham}', "");
Expect(0, 5791, '\P{Block=-_ogham}', "");
Expect(1, 5791, '\P{^Block=-_ogham}', "");
Expect(0, 5792, '\p{Block=-_ogham}', "");
Expect(1, 5792, '\p{^Block=-_ogham}', "");
Expect(1, 5792, '\P{Block=-_ogham}', "");
Expect(0, 5792, '\P{^Block=-_ogham}', "");
Error('\p{Blk=-_Ogham:=}');
Error('\P{Blk=-_Ogham:=}');
Expect(1, 5791, '\p{Blk=ogham}', "");
Expect(0, 5791, '\p{^Blk=ogham}', "");
Expect(0, 5791, '\P{Blk=ogham}', "");
Expect(1, 5791, '\P{^Blk=ogham}', "");
Expect(0, 5792, '\p{Blk=ogham}', "");
Expect(1, 5792, '\p{^Blk=ogham}', "");
Expect(1, 5792, '\P{Blk=ogham}', "");
Expect(0, 5792, '\P{^Blk=ogham}', "");
Expect(1, 5791, '\p{Blk=--Ogham}', "");
Expect(0, 5791, '\p{^Blk=--Ogham}', "");
Expect(0, 5791, '\P{Blk=--Ogham}', "");
Expect(1, 5791, '\P{^Blk=--Ogham}', "");
Expect(0, 5792, '\p{Blk=--Ogham}', "");
Expect(1, 5792, '\p{^Blk=--Ogham}', "");
Expect(1, 5792, '\P{Blk=--Ogham}', "");
Expect(0, 5792, '\P{^Blk=--Ogham}', "");
Error('\p{Is_Block= /a/Ogham}');
Error('\P{Is_Block= /a/Ogham}');
Expect(1, 5791, '\p{Is_Block=ogham}', "");
Expect(0, 5791, '\p{^Is_Block=ogham}', "");
Expect(0, 5791, '\P{Is_Block=ogham}', "");
Expect(1, 5791, '\P{^Is_Block=ogham}', "");
Expect(0, 5792, '\p{Is_Block=ogham}', "");
Expect(1, 5792, '\p{^Is_Block=ogham}', "");
Expect(1, 5792, '\P{Is_Block=ogham}', "");
Expect(0, 5792, '\P{^Is_Block=ogham}', "");
Expect(1, 5791, '\p{Is_Block=	_Ogham}', "");
Expect(0, 5791, '\p{^Is_Block=	_Ogham}', "");
Expect(0, 5791, '\P{Is_Block=	_Ogham}', "");
Expect(1, 5791, '\P{^Is_Block=	_Ogham}', "");
Expect(0, 5792, '\p{Is_Block=	_Ogham}', "");
Expect(1, 5792, '\p{^Is_Block=	_Ogham}', "");
Expect(1, 5792, '\P{Is_Block=	_Ogham}', "");
Expect(0, 5792, '\P{^Is_Block=	_Ogham}', "");
Error('\p{Is_Blk= 	Ogham/a/}');
Error('\P{Is_Blk= 	Ogham/a/}');
Expect(1, 5791, '\p{Is_Blk:	ogham}', "");
Expect(0, 5791, '\p{^Is_Blk:	ogham}', "");
Expect(0, 5791, '\P{Is_Blk:	ogham}', "");
Expect(1, 5791, '\P{^Is_Blk:	ogham}', "");
Expect(0, 5792, '\p{Is_Blk:	ogham}', "");
Expect(1, 5792, '\p{^Is_Blk:	ogham}', "");
Expect(1, 5792, '\P{Is_Blk:	ogham}', "");
Expect(0, 5792, '\P{^Is_Blk:	ogham}', "");
Expect(1, 5791, '\p{Is_Blk=__ogham}', "");
Expect(0, 5791, '\p{^Is_Blk=__ogham}', "");
Expect(0, 5791, '\P{Is_Blk=__ogham}', "");
Expect(1, 5791, '\P{^Is_Blk=__ogham}', "");
Expect(0, 5792, '\p{Is_Blk=__ogham}', "");
Expect(1, 5792, '\p{^Is_Blk=__ogham}', "");
Expect(1, 5792, '\P{Is_Blk=__ogham}', "");
Expect(0, 5792, '\P{^Is_Blk=__ogham}', "");
Error('\p{Block=_:=Samaritan}');
Error('\P{Block=_:=Samaritan}');
Expect(1, 2111, '\p{Block=samaritan}', "");
Expect(0, 2111, '\p{^Block=samaritan}', "");
Expect(0, 2111, '\P{Block=samaritan}', "");
Expect(1, 2111, '\P{^Block=samaritan}', "");
Expect(0, 2112, '\p{Block=samaritan}', "");
Expect(1, 2112, '\p{^Block=samaritan}', "");
Expect(1, 2112, '\P{Block=samaritan}', "");
Expect(0, 2112, '\P{^Block=samaritan}', "");
Expect(1, 2111, '\p{Block= -Samaritan}', "");
Expect(0, 2111, '\p{^Block= -Samaritan}', "");
Expect(0, 2111, '\P{Block= -Samaritan}', "");
Expect(1, 2111, '\P{^Block= -Samaritan}', "");
Expect(0, 2112, '\p{Block= -Samaritan}', "");
Expect(1, 2112, '\p{^Block= -Samaritan}', "");
Expect(1, 2112, '\P{Block= -Samaritan}', "");
Expect(0, 2112, '\P{^Block= -Samaritan}', "");
Error('\p{Blk=_-SAMARITAN:=}');
Error('\P{Blk=_-SAMARITAN:=}');
Expect(1, 2111, '\p{Blk=samaritan}', "");
Expect(0, 2111, '\p{^Blk=samaritan}', "");
Expect(0, 2111, '\P{Blk=samaritan}', "");
Expect(1, 2111, '\P{^Blk=samaritan}', "");
Expect(0, 2112, '\p{Blk=samaritan}', "");
Expect(1, 2112, '\p{^Blk=samaritan}', "");
Expect(1, 2112, '\P{Blk=samaritan}', "");
Expect(0, 2112, '\P{^Blk=samaritan}', "");
Expect(1, 2111, '\p{Blk=	 Samaritan}', "");
Expect(0, 2111, '\p{^Blk=	 Samaritan}', "");
Expect(0, 2111, '\P{Blk=	 Samaritan}', "");
Expect(1, 2111, '\P{^Blk=	 Samaritan}', "");
Expect(0, 2112, '\p{Blk=	 Samaritan}', "");
Expect(1, 2112, '\p{^Blk=	 Samaritan}', "");
Expect(1, 2112, '\P{Blk=	 Samaritan}', "");
Expect(0, 2112, '\P{^Blk=	 Samaritan}', "");
Error('\p{Is_Block=:=	SAMARITAN}');
Error('\P{Is_Block=:=	SAMARITAN}');
Expect(1, 2111, '\p{Is_Block=samaritan}', "");
Expect(0, 2111, '\p{^Is_Block=samaritan}', "");
Expect(0, 2111, '\P{Is_Block=samaritan}', "");
Expect(1, 2111, '\P{^Is_Block=samaritan}', "");
Expect(0, 2112, '\p{Is_Block=samaritan}', "");
Expect(1, 2112, '\p{^Is_Block=samaritan}', "");
Expect(1, 2112, '\P{Is_Block=samaritan}', "");
Expect(0, 2112, '\P{^Is_Block=samaritan}', "");
Expect(1, 2111, '\p{Is_Block=	 Samaritan}', "");
Expect(0, 2111, '\p{^Is_Block=	 Samaritan}', "");
Expect(0, 2111, '\P{Is_Block=	 Samaritan}', "");
Expect(1, 2111, '\P{^Is_Block=	 Samaritan}', "");
Expect(0, 2112, '\p{Is_Block=	 Samaritan}', "");
Expect(1, 2112, '\p{^Is_Block=	 Samaritan}', "");
Expect(1, 2112, '\P{Is_Block=	 Samaritan}', "");
Expect(0, 2112, '\P{^Is_Block=	 Samaritan}', "");
Error('\p{Is_Blk=_/a/SAMARITAN}');
Error('\P{Is_Blk=_/a/SAMARITAN}');
Expect(1, 2111, '\p{Is_Blk=samaritan}', "");
Expect(0, 2111, '\p{^Is_Blk=samaritan}', "");
Expect(0, 2111, '\P{Is_Blk=samaritan}', "");
Expect(1, 2111, '\P{^Is_Blk=samaritan}', "");
Expect(0, 2112, '\p{Is_Blk=samaritan}', "");
Expect(1, 2112, '\p{^Is_Blk=samaritan}', "");
Expect(1, 2112, '\P{Is_Blk=samaritan}', "");
Expect(0, 2112, '\P{^Is_Blk=samaritan}', "");
Expect(1, 2111, '\p{Is_Blk=_	Samaritan}', "");
Expect(0, 2111, '\p{^Is_Blk=_	Samaritan}', "");
Expect(0, 2111, '\P{Is_Blk=_	Samaritan}', "");
Expect(1, 2111, '\P{^Is_Blk=_	Samaritan}', "");
Expect(0, 2112, '\p{Is_Blk=_	Samaritan}', "");
Expect(1, 2112, '\p{^Is_Blk=_	Samaritan}', "");
Expect(1, 2112, '\P{Is_Blk=_	Samaritan}', "");
Expect(0, 2112, '\P{^Is_Blk=_	Samaritan}', "");
Error('\p{Block=-/a/LIMBU}');
Error('\P{Block=-/a/LIMBU}');
Expect(1, 6479, '\p{Block=limbu}', "");
Expect(0, 6479, '\p{^Block=limbu}', "");
Expect(0, 6479, '\P{Block=limbu}', "");
Expect(1, 6479, '\P{^Block=limbu}', "");
Expect(0, 6480, '\p{Block=limbu}', "");
Expect(1, 6480, '\p{^Block=limbu}', "");
Expect(1, 6480, '\P{Block=limbu}', "");
Expect(0, 6480, '\P{^Block=limbu}', "");
Expect(1, 6479, '\p{Block=  Limbu}', "");
Expect(0, 6479, '\p{^Block=  Limbu}', "");
Expect(0, 6479, '\P{Block=  Limbu}', "");
Expect(1, 6479, '\P{^Block=  Limbu}', "");
Expect(0, 6480, '\p{Block=  Limbu}', "");
Expect(1, 6480, '\p{^Block=  Limbu}', "");
Expect(1, 6480, '\P{Block=  Limbu}', "");
Expect(0, 6480, '\P{^Block=  Limbu}', "");
Error('\p{Blk=/a/_ LIMBU}');
Error('\P{Blk=/a/_ LIMBU}');
Expect(1, 6479, '\p{Blk=limbu}', "");
Expect(0, 6479, '\p{^Blk=limbu}', "");
Expect(0, 6479, '\P{Blk=limbu}', "");
Expect(1, 6479, '\P{^Blk=limbu}', "");
Expect(0, 6480, '\p{Blk=limbu}', "");
Expect(1, 6480, '\p{^Blk=limbu}', "");
Expect(1, 6480, '\P{Blk=limbu}', "");
Expect(0, 6480, '\P{^Blk=limbu}', "");
Expect(1, 6479, '\p{Blk=-_limbu}', "");
Expect(0, 6479, '\p{^Blk=-_limbu}', "");
Expect(0, 6479, '\P{Blk=-_limbu}', "");
Expect(1, 6479, '\P{^Blk=-_limbu}', "");
Expect(0, 6480, '\p{Blk=-_limbu}', "");
Expect(1, 6480, '\p{^Blk=-_limbu}', "");
Expect(1, 6480, '\P{Blk=-_limbu}', "");
Expect(0, 6480, '\P{^Blk=-_limbu}', "");
Error('\p{Is_Block= LIMBU:=}');
Error('\P{Is_Block= LIMBU:=}');
Expect(1, 6479, '\p{Is_Block=limbu}', "");
Expect(0, 6479, '\p{^Is_Block=limbu}', "");
Expect(0, 6479, '\P{Is_Block=limbu}', "");
Expect(1, 6479, '\P{^Is_Block=limbu}', "");
Expect(0, 6480, '\p{Is_Block=limbu}', "");
Expect(1, 6480, '\p{^Is_Block=limbu}', "");
Expect(1, 6480, '\P{Is_Block=limbu}', "");
Expect(0, 6480, '\P{^Is_Block=limbu}', "");
Expect(1, 6479, '\p{Is_Block=  Limbu}', "");
Expect(0, 6479, '\p{^Is_Block=  Limbu}', "");
Expect(0, 6479, '\P{Is_Block=  Limbu}', "");
Expect(1, 6479, '\P{^Is_Block=  Limbu}', "");
Expect(0, 6480, '\p{Is_Block=  Limbu}', "");
Expect(1, 6480, '\p{^Is_Block=  Limbu}', "");
Expect(1, 6480, '\P{Is_Block=  Limbu}', "");
Expect(0, 6480, '\P{^Is_Block=  Limbu}', "");
Error('\p{Is_Blk=:=-_limbu}');
Error('\P{Is_Blk=:=-_limbu}');
Expect(1, 6479, '\p{Is_Blk=limbu}', "");
Expect(0, 6479, '\p{^Is_Blk=limbu}', "");
Expect(0, 6479, '\P{Is_Blk=limbu}', "");
Expect(1, 6479, '\P{^Is_Blk=limbu}', "");
Expect(0, 6480, '\p{Is_Blk=limbu}', "");
Expect(1, 6480, '\p{^Is_Blk=limbu}', "");
Expect(1, 6480, '\P{Is_Blk=limbu}', "");
Expect(0, 6480, '\P{^Is_Blk=limbu}', "");
Expect(1, 6479, '\p{Is_Blk=  Limbu}', "");
Expect(0, 6479, '\p{^Is_Blk=  Limbu}', "");
Expect(0, 6479, '\P{Is_Blk=  Limbu}', "");
Expect(1, 6479, '\P{^Is_Blk=  Limbu}', "");
Expect(0, 6480, '\p{Is_Blk=  Limbu}', "");
Expect(1, 6480, '\p{^Is_Blk=  Limbu}', "");
Expect(1, 6480, '\P{Is_Blk=  Limbu}', "");
Expect(0, 6480, '\P{^Is_Blk=  Limbu}', "");
Error('\p{Block=_:=Ethiopic}');
Error('\P{Block=_:=Ethiopic}');
Expect(1, 4991, '\p{Block:	ethiopic}', "");
Expect(0, 4991, '\p{^Block:	ethiopic}', "");
Expect(0, 4991, '\P{Block:	ethiopic}', "");
Expect(1, 4991, '\P{^Block:	ethiopic}', "");
Expect(0, 4992, '\p{Block:	ethiopic}', "");
Expect(1, 4992, '\p{^Block:	ethiopic}', "");
Expect(1, 4992, '\P{Block:	ethiopic}', "");
Expect(0, 4992, '\P{^Block:	ethiopic}', "");
Expect(1, 4991, '\p{Block:	 ethiopic}', "");
Expect(0, 4991, '\p{^Block:	 ethiopic}', "");
Expect(0, 4991, '\P{Block:	 ethiopic}', "");
Expect(1, 4991, '\P{^Block:	 ethiopic}', "");
Expect(0, 4992, '\p{Block:	 ethiopic}', "");
Expect(1, 4992, '\p{^Block:	 ethiopic}', "");
Expect(1, 4992, '\P{Block:	 ethiopic}', "");
Expect(0, 4992, '\P{^Block:	 ethiopic}', "");
Error('\p{Blk=/a/Ethiopic}');
Error('\P{Blk=/a/Ethiopic}');
Expect(1, 4991, '\p{Blk=ethiopic}', "");
Expect(0, 4991, '\p{^Blk=ethiopic}', "");
Expect(0, 4991, '\P{Blk=ethiopic}', "");
Expect(1, 4991, '\P{^Blk=ethiopic}', "");
Expect(0, 4992, '\p{Blk=ethiopic}', "");
Expect(1, 4992, '\p{^Blk=ethiopic}', "");
Expect(1, 4992, '\P{Blk=ethiopic}', "");
Expect(0, 4992, '\P{^Blk=ethiopic}', "");
Expect(1, 4991, '\p{Blk= ETHIOPIC}', "");
Expect(0, 4991, '\p{^Blk= ETHIOPIC}', "");
Expect(0, 4991, '\P{Blk= ETHIOPIC}', "");
Expect(1, 4991, '\P{^Blk= ETHIOPIC}', "");
Expect(0, 4992, '\p{Blk= ETHIOPIC}', "");
Expect(1, 4992, '\p{^Blk= ETHIOPIC}', "");
Expect(1, 4992, '\P{Blk= ETHIOPIC}', "");
Expect(0, 4992, '\P{^Blk= ETHIOPIC}', "");
Error('\p{Is_Block=:= 	ETHIOPIC}');
Error('\P{Is_Block=:= 	ETHIOPIC}');
Expect(1, 4991, '\p{Is_Block=ethiopic}', "");
Expect(0, 4991, '\p{^Is_Block=ethiopic}', "");
Expect(0, 4991, '\P{Is_Block=ethiopic}', "");
Expect(1, 4991, '\P{^Is_Block=ethiopic}', "");
Expect(0, 4992, '\p{Is_Block=ethiopic}', "");
Expect(1, 4992, '\p{^Is_Block=ethiopic}', "");
Expect(1, 4992, '\P{Is_Block=ethiopic}', "");
Expect(0, 4992, '\P{^Is_Block=ethiopic}', "");
Expect(1, 4991, '\p{Is_Block=--ethiopic}', "");
Expect(0, 4991, '\p{^Is_Block=--ethiopic}', "");
Expect(0, 4991, '\P{Is_Block=--ethiopic}', "");
Expect(1, 4991, '\P{^Is_Block=--ethiopic}', "");
Expect(0, 4992, '\p{Is_Block=--ethiopic}', "");
Expect(1, 4992, '\p{^Is_Block=--ethiopic}', "");
Expect(1, 4992, '\P{Is_Block=--ethiopic}', "");
Expect(0, 4992, '\P{^Is_Block=--ethiopic}', "");
Error('\p{Is_Blk=_:=ETHIOPIC}');
Error('\P{Is_Blk=_:=ETHIOPIC}');
Expect(1, 4991, '\p{Is_Blk=ethiopic}', "");
Expect(0, 4991, '\p{^Is_Blk=ethiopic}', "");
Expect(0, 4991, '\P{Is_Blk=ethiopic}', "");
Expect(1, 4991, '\P{^Is_Blk=ethiopic}', "");
Expect(0, 4992, '\p{Is_Blk=ethiopic}', "");
Expect(1, 4992, '\p{^Is_Blk=ethiopic}', "");
Expect(1, 4992, '\P{Is_Blk=ethiopic}', "");
Expect(0, 4992, '\P{^Is_Blk=ethiopic}', "");
Expect(1, 4991, '\p{Is_Blk=	_Ethiopic}', "");
Expect(0, 4991, '\p{^Is_Blk=	_Ethiopic}', "");
Expect(0, 4991, '\P{Is_Blk=	_Ethiopic}', "");
Expect(1, 4991, '\P{^Is_Blk=	_Ethiopic}', "");
Expect(0, 4992, '\p{Is_Blk=	_Ethiopic}', "");
Expect(1, 4992, '\p{^Is_Blk=	_Ethiopic}', "");
Expect(1, 4992, '\P{Is_Blk=	_Ethiopic}', "");
Expect(0, 4992, '\P{^Is_Blk=	_Ethiopic}', "");
Error('\p{Block= 	BOX_drawing:=}');
Error('\P{Block= 	BOX_drawing:=}');
Expect(1, 9599, '\p{Block=boxdrawing}', "");
Expect(0, 9599, '\p{^Block=boxdrawing}', "");
Expect(0, 9599, '\P{Block=boxdrawing}', "");
Expect(1, 9599, '\P{^Block=boxdrawing}', "");
Expect(0, 9600, '\p{Block=boxdrawing}', "");
Expect(1, 9600, '\p{^Block=boxdrawing}', "");
Expect(1, 9600, '\P{Block=boxdrawing}', "");
Expect(0, 9600, '\P{^Block=boxdrawing}', "");
Expect(1, 9599, '\p{Block=_BOX_Drawing}', "");
Expect(0, 9599, '\p{^Block=_BOX_Drawing}', "");
Expect(0, 9599, '\P{Block=_BOX_Drawing}', "");
Expect(1, 9599, '\P{^Block=_BOX_Drawing}', "");
Expect(0, 9600, '\p{Block=_BOX_Drawing}', "");
Expect(1, 9600, '\p{^Block=_BOX_Drawing}', "");
Expect(1, 9600, '\P{Block=_BOX_Drawing}', "");
Expect(0, 9600, '\P{^Block=_BOX_Drawing}', "");
Error('\p{Blk= :=box_drawing}');
Error('\P{Blk= :=box_drawing}');
Expect(1, 9599, '\p{Blk=boxdrawing}', "");
Expect(0, 9599, '\p{^Blk=boxdrawing}', "");
Expect(0, 9599, '\P{Blk=boxdrawing}', "");
Expect(1, 9599, '\P{^Blk=boxdrawing}', "");
Expect(0, 9600, '\p{Blk=boxdrawing}', "");
Expect(1, 9600, '\p{^Blk=boxdrawing}', "");
Expect(1, 9600, '\P{Blk=boxdrawing}', "");
Expect(0, 9600, '\P{^Blk=boxdrawing}', "");
Expect(1, 9599, '\p{Blk=-	Box_drawing}', "");
Expect(0, 9599, '\p{^Blk=-	Box_drawing}', "");
Expect(0, 9599, '\P{Blk=-	Box_drawing}', "");
Expect(1, 9599, '\P{^Blk=-	Box_drawing}', "");
Expect(0, 9600, '\p{Blk=-	Box_drawing}', "");
Expect(1, 9600, '\p{^Blk=-	Box_drawing}', "");
Expect(1, 9600, '\P{Blk=-	Box_drawing}', "");
Expect(0, 9600, '\P{^Blk=-	Box_drawing}', "");
Error('\p{Is_Block=	:=Box_DRAWING}');
Error('\P{Is_Block=	:=Box_DRAWING}');
Expect(1, 9599, '\p{Is_Block:	boxdrawing}', "");
Expect(0, 9599, '\p{^Is_Block:	boxdrawing}', "");
Expect(0, 9599, '\P{Is_Block:	boxdrawing}', "");
Expect(1, 9599, '\P{^Is_Block:	boxdrawing}', "");
Expect(0, 9600, '\p{Is_Block:	boxdrawing}', "");
Expect(1, 9600, '\p{^Is_Block:	boxdrawing}', "");
Expect(1, 9600, '\P{Is_Block:	boxdrawing}', "");
Expect(0, 9600, '\P{^Is_Block:	boxdrawing}', "");
Expect(1, 9599, '\p{Is_Block:    	box_Drawing}', "");
Expect(0, 9599, '\p{^Is_Block:    	box_Drawing}', "");
Expect(0, 9599, '\P{Is_Block:    	box_Drawing}', "");
Expect(1, 9599, '\P{^Is_Block:    	box_Drawing}', "");
Expect(0, 9600, '\p{Is_Block:    	box_Drawing}', "");
Expect(1, 9600, '\p{^Is_Block:    	box_Drawing}', "");
Expect(1, 9600, '\P{Is_Block:    	box_Drawing}', "");
Expect(0, 9600, '\P{^Is_Block:    	box_Drawing}', "");
Error('\p{Is_Blk=:=	-BOX_DRAWING}');
Error('\P{Is_Blk=:=	-BOX_DRAWING}');
Expect(1, 9599, '\p{Is_Blk:boxdrawing}', "");
Expect(0, 9599, '\p{^Is_Blk:boxdrawing}', "");
Expect(0, 9599, '\P{Is_Blk:boxdrawing}', "");
Expect(1, 9599, '\P{^Is_Blk:boxdrawing}', "");
Expect(0, 9600, '\p{Is_Blk:boxdrawing}', "");
Expect(1, 9600, '\p{^Is_Blk:boxdrawing}', "");
Expect(1, 9600, '\P{Is_Blk:boxdrawing}', "");
Expect(0, 9600, '\P{^Is_Blk:boxdrawing}', "");
Expect(1, 9599, '\p{Is_Blk= 	Box_DRAWING}', "");
Expect(0, 9599, '\p{^Is_Blk= 	Box_DRAWING}', "");
Expect(0, 9599, '\P{Is_Blk= 	Box_DRAWING}', "");
Expect(1, 9599, '\P{^Is_Blk= 	Box_DRAWING}', "");
Expect(0, 9600, '\p{Is_Blk= 	Box_DRAWING}', "");
Expect(1, 9600, '\p{^Is_Blk= 	Box_DRAWING}', "");
Expect(1, 9600, '\P{Is_Blk= 	Box_DRAWING}', "");
Expect(0, 9600, '\P{^Is_Blk= 	Box_DRAWING}', "");
Error('\p{Block=:=old_Turkic}');
Error('\P{Block=:=old_Turkic}');
Expect(1, 68687, '\p{Block:	oldturkic}', "");
Expect(0, 68687, '\p{^Block:	oldturkic}', "");
Expect(0, 68687, '\P{Block:	oldturkic}', "");
Expect(1, 68687, '\P{^Block:	oldturkic}', "");
Expect(0, 68688, '\p{Block:	oldturkic}', "");
Expect(1, 68688, '\p{^Block:	oldturkic}', "");
Expect(1, 68688, '\P{Block:	oldturkic}', "");
Expect(0, 68688, '\P{^Block:	oldturkic}', "");
Expect(1, 68687, '\p{Block=_old_TURKIC}', "");
Expect(0, 68687, '\p{^Block=_old_TURKIC}', "");
Expect(0, 68687, '\P{Block=_old_TURKIC}', "");
Expect(1, 68687, '\P{^Block=_old_TURKIC}', "");
Expect(0, 68688, '\p{Block=_old_TURKIC}', "");
Expect(1, 68688, '\p{^Block=_old_TURKIC}', "");
Expect(1, 68688, '\P{Block=_old_TURKIC}', "");
Expect(0, 68688, '\P{^Block=_old_TURKIC}', "");
Error('\p{Blk=/a/-	OLD_Turkic}');
Error('\P{Blk=/a/-	OLD_Turkic}');
Expect(1, 68687, '\p{Blk=oldturkic}', "");
Expect(0, 68687, '\p{^Blk=oldturkic}', "");
Expect(0, 68687, '\P{Blk=oldturkic}', "");
Expect(1, 68687, '\P{^Blk=oldturkic}', "");
Expect(0, 68688, '\p{Blk=oldturkic}', "");
Expect(1, 68688, '\p{^Blk=oldturkic}', "");
Expect(1, 68688, '\P{Blk=oldturkic}', "");
Expect(0, 68688, '\P{^Blk=oldturkic}', "");
Expect(1, 68687, '\p{Blk=_old_Turkic}', "");
Expect(0, 68687, '\p{^Blk=_old_Turkic}', "");
Expect(0, 68687, '\P{Blk=_old_Turkic}', "");
Expect(1, 68687, '\P{^Blk=_old_Turkic}', "");
Expect(0, 68688, '\p{Blk=_old_Turkic}', "");
Expect(1, 68688, '\p{^Blk=_old_Turkic}', "");
Expect(1, 68688, '\P{Blk=_old_Turkic}', "");
Expect(0, 68688, '\P{^Blk=_old_Turkic}', "");
Error('\p{Is_Block=/a/__Old_TURKIC}');
Error('\P{Is_Block=/a/__Old_TURKIC}');
Expect(1, 68687, '\p{Is_Block=oldturkic}', "");
Expect(0, 68687, '\p{^Is_Block=oldturkic}', "");
Expect(0, 68687, '\P{Is_Block=oldturkic}', "");
Expect(1, 68687, '\P{^Is_Block=oldturkic}', "");
Expect(0, 68688, '\p{Is_Block=oldturkic}', "");
Expect(1, 68688, '\p{^Is_Block=oldturkic}', "");
Expect(1, 68688, '\P{Is_Block=oldturkic}', "");
Expect(0, 68688, '\P{^Is_Block=oldturkic}', "");
Expect(1, 68687, '\p{Is_Block= 	Old_turkic}', "");
Expect(0, 68687, '\p{^Is_Block= 	Old_turkic}', "");
Expect(0, 68687, '\P{Is_Block= 	Old_turkic}', "");
Expect(1, 68687, '\P{^Is_Block= 	Old_turkic}', "");
Expect(0, 68688, '\p{Is_Block= 	Old_turkic}', "");
Expect(1, 68688, '\p{^Is_Block= 	Old_turkic}', "");
Expect(1, 68688, '\P{Is_Block= 	Old_turkic}', "");
Expect(0, 68688, '\P{^Is_Block= 	Old_turkic}', "");
Error('\p{Is_Blk=:=- OLD_Turkic}');
Error('\P{Is_Blk=:=- OLD_Turkic}');
Expect(1, 68687, '\p{Is_Blk=oldturkic}', "");
Expect(0, 68687, '\p{^Is_Blk=oldturkic}', "");
Expect(0, 68687, '\P{Is_Blk=oldturkic}', "");
Expect(1, 68687, '\P{^Is_Blk=oldturkic}', "");
Expect(0, 68688, '\p{Is_Blk=oldturkic}', "");
Expect(1, 68688, '\p{^Is_Blk=oldturkic}', "");
Expect(1, 68688, '\P{Is_Blk=oldturkic}', "");
Expect(0, 68688, '\P{^Is_Blk=oldturkic}', "");
Expect(1, 68687, '\p{Is_Blk=Old_TURKIC}', "");
Expect(0, 68687, '\p{^Is_Blk=Old_TURKIC}', "");
Expect(0, 68687, '\P{Is_Blk=Old_TURKIC}', "");
Expect(1, 68687, '\P{^Is_Blk=Old_TURKIC}', "");
Expect(0, 68688, '\p{Is_Blk=Old_TURKIC}', "");
Expect(1, 68688, '\p{^Is_Blk=Old_TURKIC}', "");
Expect(1, 68688, '\P{Is_Blk=Old_TURKIC}', "");
Expect(0, 68688, '\P{^Is_Blk=Old_TURKIC}', "");
Error('\p{Block=-/a/Latin_EXTENDED_c}');
Error('\P{Block=-/a/Latin_EXTENDED_c}');
Expect(1, 11391, '\p{Block=latinextendedc}', "");
Expect(0, 11391, '\p{^Block=latinextendedc}', "");
Expect(0, 11391, '\P{Block=latinextendedc}', "");
Expect(1, 11391, '\P{^Block=latinextendedc}', "");
Expect(0, 11392, '\p{Block=latinextendedc}', "");
Expect(1, 11392, '\p{^Block=latinextendedc}', "");
Expect(1, 11392, '\P{Block=latinextendedc}', "");
Expect(0, 11392, '\P{^Block=latinextendedc}', "");
Expect(1, 11391, '\p{Block=		LATIN_extended_C}', "");
Expect(0, 11391, '\p{^Block=		LATIN_extended_C}', "");
Expect(0, 11391, '\P{Block=		LATIN_extended_C}', "");
Expect(1, 11391, '\P{^Block=		LATIN_extended_C}', "");
Expect(0, 11392, '\p{Block=		LATIN_extended_C}', "");
Expect(1, 11392, '\p{^Block=		LATIN_extended_C}', "");
Expect(1, 11392, '\P{Block=		LATIN_extended_C}', "");
Expect(0, 11392, '\P{^Block=		LATIN_extended_C}', "");
Error('\p{Blk=- Latin_Ext_C/a/}');
Error('\P{Blk=- Latin_Ext_C/a/}');
Expect(1, 11391, '\p{Blk=latinextc}', "");
Expect(0, 11391, '\p{^Blk=latinextc}', "");
Expect(0, 11391, '\P{Blk=latinextc}', "");
Expect(1, 11391, '\P{^Blk=latinextc}', "");
Expect(0, 11392, '\p{Blk=latinextc}', "");
Expect(1, 11392, '\p{^Blk=latinextc}', "");
Expect(1, 11392, '\P{Blk=latinextc}', "");
Expect(0, 11392, '\P{^Blk=latinextc}', "");
Expect(1, 11391, '\p{Blk= Latin_ext_C}', "");
Expect(0, 11391, '\p{^Blk= Latin_ext_C}', "");
Expect(0, 11391, '\P{Blk= Latin_ext_C}', "");
Expect(1, 11391, '\P{^Blk= Latin_ext_C}', "");
Expect(0, 11392, '\p{Blk= Latin_ext_C}', "");
Expect(1, 11392, '\p{^Blk= Latin_ext_C}', "");
Expect(1, 11392, '\P{Blk= Latin_ext_C}', "");
Expect(0, 11392, '\P{^Blk= Latin_ext_C}', "");
Error('\p{Is_Block:	_/a/latin_Extended_C}');
Error('\P{Is_Block:	_/a/latin_Extended_C}');
Expect(1, 11391, '\p{Is_Block=latinextendedc}', "");
Expect(0, 11391, '\p{^Is_Block=latinextendedc}', "");
Expect(0, 11391, '\P{Is_Block=latinextendedc}', "");
Expect(1, 11391, '\P{^Is_Block=latinextendedc}', "");
Expect(0, 11392, '\p{Is_Block=latinextendedc}', "");
Expect(1, 11392, '\p{^Is_Block=latinextendedc}', "");
Expect(1, 11392, '\P{Is_Block=latinextendedc}', "");
Expect(0, 11392, '\P{^Is_Block=latinextendedc}', "");
Expect(1, 11391, '\p{Is_Block=_ Latin_EXTENDED_C}', "");
Expect(0, 11391, '\p{^Is_Block=_ Latin_EXTENDED_C}', "");
Expect(0, 11391, '\P{Is_Block=_ Latin_EXTENDED_C}', "");
Expect(1, 11391, '\P{^Is_Block=_ Latin_EXTENDED_C}', "");
Expect(0, 11392, '\p{Is_Block=_ Latin_EXTENDED_C}', "");
Expect(1, 11392, '\p{^Is_Block=_ Latin_EXTENDED_C}', "");
Expect(1, 11392, '\P{Is_Block=_ Latin_EXTENDED_C}', "");
Expect(0, 11392, '\P{^Is_Block=_ Latin_EXTENDED_C}', "");
Error('\p{Is_Blk:	 -LATIN_ext_C/a/}');
Error('\P{Is_Blk:	 -LATIN_ext_C/a/}');
Expect(1, 11391, '\p{Is_Blk=latinextc}', "");
Expect(0, 11391, '\p{^Is_Blk=latinextc}', "");
Expect(0, 11391, '\P{Is_Blk=latinextc}', "");
Expect(1, 11391, '\P{^Is_Blk=latinextc}', "");
Expect(0, 11392, '\p{Is_Blk=latinextc}', "");
Expect(1, 11392, '\p{^Is_Blk=latinextc}', "");
Expect(1, 11392, '\P{Is_Blk=latinextc}', "");
Expect(0, 11392, '\P{^Is_Blk=latinextc}', "");
Expect(1, 11391, '\p{Is_Blk=		latin_Ext_C}', "");
Expect(0, 11391, '\p{^Is_Blk=		latin_Ext_C}', "");
Expect(0, 11391, '\P{Is_Blk=		latin_Ext_C}', "");
Expect(1, 11391, '\P{^Is_Blk=		latin_Ext_C}', "");
Expect(0, 11392, '\p{Is_Blk=		latin_Ext_C}', "");
Expect(1, 11392, '\p{^Is_Blk=		latin_Ext_C}', "");
Expect(1, 11392, '\P{Is_Blk=		latin_Ext_C}', "");
Expect(0, 11392, '\P{^Is_Blk=		latin_Ext_C}', "");
Error('\p{Block=/a/	_Ol_Chiki}');
Error('\P{Block=/a/	_Ol_Chiki}');
Expect(1, 7295, '\p{Block=olchiki}', "");
Expect(0, 7295, '\p{^Block=olchiki}', "");
Expect(0, 7295, '\P{Block=olchiki}', "");
Expect(1, 7295, '\P{^Block=olchiki}', "");
Expect(0, 7296, '\p{Block=olchiki}', "");
Expect(1, 7296, '\p{^Block=olchiki}', "");
Expect(1, 7296, '\P{Block=olchiki}', "");
Expect(0, 7296, '\P{^Block=olchiki}', "");
Expect(1, 7295, '\p{Block= _ol_chiki}', "");
Expect(0, 7295, '\p{^Block= _ol_chiki}', "");
Expect(0, 7295, '\P{Block= _ol_chiki}', "");
Expect(1, 7295, '\P{^Block= _ol_chiki}', "");
Expect(0, 7296, '\p{Block= _ol_chiki}', "");
Expect(1, 7296, '\p{^Block= _ol_chiki}', "");
Expect(1, 7296, '\P{Block= _ol_chiki}', "");
Expect(0, 7296, '\P{^Block= _ol_chiki}', "");
Error('\p{Blk= OL_CHIKI:=}');
Error('\P{Blk= OL_CHIKI:=}');
Expect(1, 7295, '\p{Blk=olchiki}', "");
Expect(0, 7295, '\p{^Blk=olchiki}', "");
Expect(0, 7295, '\P{Blk=olchiki}', "");
Expect(1, 7295, '\P{^Blk=olchiki}', "");
Expect(0, 7296, '\p{Blk=olchiki}', "");
Expect(1, 7296, '\p{^Blk=olchiki}', "");
Expect(1, 7296, '\P{Blk=olchiki}', "");
Expect(0, 7296, '\P{^Blk=olchiki}', "");
Expect(1, 7295, '\p{Blk:	ol_Chiki}', "");
Expect(0, 7295, '\p{^Blk:	ol_Chiki}', "");
Expect(0, 7295, '\P{Blk:	ol_Chiki}', "");
Expect(1, 7295, '\P{^Blk:	ol_Chiki}', "");
Expect(0, 7296, '\p{Blk:	ol_Chiki}', "");
Expect(1, 7296, '\p{^Blk:	ol_Chiki}', "");
Expect(1, 7296, '\P{Blk:	ol_Chiki}', "");
Expect(0, 7296, '\P{^Blk:	ol_Chiki}', "");
Error('\p{Is_Block=:=-Ol_Chiki}');
Error('\P{Is_Block=:=-Ol_Chiki}');
Expect(1, 7295, '\p{Is_Block=olchiki}', "");
Expect(0, 7295, '\p{^Is_Block=olchiki}', "");
Expect(0, 7295, '\P{Is_Block=olchiki}', "");
Expect(1, 7295, '\P{^Is_Block=olchiki}', "");
Expect(0, 7296, '\p{Is_Block=olchiki}', "");
Expect(1, 7296, '\p{^Is_Block=olchiki}', "");
Expect(1, 7296, '\P{Is_Block=olchiki}', "");
Expect(0, 7296, '\P{^Is_Block=olchiki}', "");
Expect(1, 7295, '\p{Is_Block= 	OL_Chiki}', "");
Expect(0, 7295, '\p{^Is_Block= 	OL_Chiki}', "");
Expect(0, 7295, '\P{Is_Block= 	OL_Chiki}', "");
Expect(1, 7295, '\P{^Is_Block= 	OL_Chiki}', "");
Expect(0, 7296, '\p{Is_Block= 	OL_Chiki}', "");
Expect(1, 7296, '\p{^Is_Block= 	OL_Chiki}', "");
Expect(1, 7296, '\P{Is_Block= 	OL_Chiki}', "");
Expect(0, 7296, '\P{^Is_Block= 	OL_Chiki}', "");
Error('\p{Is_Blk= :=Ol_Chiki}');
Error('\P{Is_Blk= :=Ol_Chiki}');
Expect(1, 7295, '\p{Is_Blk=olchiki}', "");
Expect(0, 7295, '\p{^Is_Blk=olchiki}', "");
Expect(0, 7295, '\P{Is_Blk=olchiki}', "");
Expect(1, 7295, '\P{^Is_Blk=olchiki}', "");
Expect(0, 7296, '\p{Is_Blk=olchiki}', "");
Expect(1, 7296, '\p{^Is_Blk=olchiki}', "");
Expect(1, 7296, '\P{Is_Blk=olchiki}', "");
Expect(0, 7296, '\P{^Is_Blk=olchiki}', "");
Expect(1, 7295, '\p{Is_Blk=_ ol_chiki}', "");
Expect(0, 7295, '\p{^Is_Blk=_ ol_chiki}', "");
Expect(0, 7295, '\P{Is_Blk=_ ol_chiki}', "");
Expect(1, 7295, '\P{^Is_Blk=_ ol_chiki}', "");
Expect(0, 7296, '\p{Is_Blk=_ ol_chiki}', "");
Expect(1, 7296, '\p{^Is_Blk=_ ol_chiki}', "");
Expect(1, 7296, '\P{Is_Blk=_ ol_chiki}', "");
Expect(0, 7296, '\P{^Is_Blk=_ ol_chiki}', "");
Error('\p{Block=/a/ supplemental_PUNCTUATION}');
Error('\P{Block=/a/ supplemental_PUNCTUATION}');
Expect(1, 11903, '\p{Block=supplementalpunctuation}', "");
Expect(0, 11903, '\p{^Block=supplementalpunctuation}', "");
Expect(0, 11903, '\P{Block=supplementalpunctuation}', "");
Expect(1, 11903, '\P{^Block=supplementalpunctuation}', "");
Expect(0, 11904, '\p{Block=supplementalpunctuation}', "");
Expect(1, 11904, '\p{^Block=supplementalpunctuation}', "");
Expect(1, 11904, '\P{Block=supplementalpunctuation}', "");
Expect(0, 11904, '\P{^Block=supplementalpunctuation}', "");
Expect(1, 11903, '\p{Block:    SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11903, '\p{^Block:    SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11903, '\P{Block:    SUPPLEMENTAL_Punctuation}', "");
Expect(1, 11903, '\P{^Block:    SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11904, '\p{Block:    SUPPLEMENTAL_Punctuation}', "");
Expect(1, 11904, '\p{^Block:    SUPPLEMENTAL_Punctuation}', "");
Expect(1, 11904, '\P{Block:    SUPPLEMENTAL_Punctuation}', "");
Expect(0, 11904, '\P{^Block:    SUPPLEMENTAL_Punctuation}', "");
Error('\p{Blk=/a/		SUP_punctuation}');
Error('\P{Blk=/a/		SUP_punctuation}');
Expect(1, 11903, '\p{Blk=suppunctuation}', "");
Expect(0, 11903, '\p{^Blk=suppunctuation}', "");
Expect(0, 11903, '\P{Blk=suppunctuation}', "");
Expect(1, 11903, '\P{^Blk=suppunctuation}', "");
Expect(0, 11904, '\p{Blk=suppunctuation}', "");
Expect(1, 11904, '\p{^Blk=suppunctuation}', "");
Expect(1, 11904, '\P{Blk=suppunctuation}', "");
Expect(0, 11904, '\P{^Blk=suppunctuation}', "");
Expect(1, 11903, '\p{Blk=_ SUP_Punctuation}', "");
Expect(0, 11903, '\p{^Blk=_ SUP_Punctuation}', "");
Expect(0, 11903, '\P{Blk=_ SUP_Punctuation}', "");
Expect(1, 11903, '\P{^Blk=_ SUP_Punctuation}', "");
Expect(0, 11904, '\p{Blk=_ SUP_Punctuation}', "");
Expect(1, 11904, '\p{^Blk=_ SUP_Punctuation}', "");
Expect(1, 11904, '\P{Blk=_ SUP_Punctuation}', "");
Expect(0, 11904, '\P{^Blk=_ SUP_Punctuation}', "");
Error('\p{Is_Block=SUPPLEMENTAL_Punctuation:=}');
Error('\P{Is_Block=SUPPLEMENTAL_Punctuation:=}');
Expect(1, 11903, '\p{Is_Block=supplementalpunctuation}', "");
Expect(0, 11903, '\p{^Is_Block=supplementalpunctuation}', "");
Expect(0, 11903, '\P{Is_Block=supplementalpunctuation}', "");
Expect(1, 11903, '\P{^Is_Block=supplementalpunctuation}', "");
Expect(0, 11904, '\p{Is_Block=supplementalpunctuation}', "");
Expect(1, 11904, '\p{^Is_Block=supplementalpunctuation}', "");
Expect(1, 11904, '\P{Is_Block=supplementalpunctuation}', "");
Expect(0, 11904, '\P{^Is_Block=supplementalpunctuation}', "");
Expect(1, 11903, '\p{Is_Block= -Supplemental_punctuation}', "");
Expect(0, 11903, '\p{^Is_Block= -Supplemental_punctuation}', "");
Expect(0, 11903, '\P{Is_Block= -Supplemental_punctuation}', "");
Expect(1, 11903, '\P{^Is_Block= -Supplemental_punctuation}', "");
Expect(0, 11904, '\p{Is_Block= -Supplemental_punctuation}', "");
Expect(1, 11904, '\p{^Is_Block= -Supplemental_punctuation}', "");
Expect(1, 11904, '\P{Is_Block= -Supplemental_punctuation}', "");
Expect(0, 11904, '\P{^Is_Block= -Supplemental_punctuation}', "");
Error('\p{Is_Blk=-:=Sup_PUNCTUATION}');
Error('\P{Is_Blk=-:=Sup_PUNCTUATION}');
Expect(1, 11903, '\p{Is_Blk=suppunctuation}', "");
Expect(0, 11903, '\p{^Is_Blk=suppunctuation}', "");
Expect(0, 11903, '\P{Is_Blk=suppunctuation}', "");
Expect(1, 11903, '\P{^Is_Blk=suppunctuation}', "");
Expect(0, 11904, '\p{Is_Blk=suppunctuation}', "");
Expect(1, 11904, '\p{^Is_Blk=suppunctuation}', "");
Expect(1, 11904, '\P{Is_Blk=suppunctuation}', "");
Expect(0, 11904, '\P{^Is_Blk=suppunctuation}', "");
Expect(1, 11903, '\p{Is_Blk:   -	Sup_Punctuation}', "");
Expect(0, 11903, '\p{^Is_Blk:   -	Sup_Punctuation}', "");
Expect(0, 11903, '\P{Is_Blk:   -	Sup_Punctuation}', "");
Expect(1, 11903, '\P{^Is_Blk:   -	Sup_Punctuation}', "");
Expect(0, 11904, '\p{Is_Blk:   -	Sup_Punctuation}', "");
Expect(1, 11904, '\p{^Is_Blk:   -	Sup_Punctuation}', "");
Expect(1, 11904, '\P{Is_Blk:   -	Sup_Punctuation}', "");
Expect(0, 11904, '\P{^Is_Blk:   -	Sup_Punctuation}', "");
Error('\p{Block=-/a/ETHIOPIC_Extended_A}');
Error('\P{Block=-/a/ETHIOPIC_Extended_A}');
Expect(1, 43823, '\p{Block=ethiopicextendeda}', "");
Expect(0, 43823, '\p{^Block=ethiopicextendeda}', "");
Expect(0, 43823, '\P{Block=ethiopicextendeda}', "");
Expect(1, 43823, '\P{^Block=ethiopicextendeda}', "");
Expect(0, 43824, '\p{Block=ethiopicextendeda}', "");
Expect(1, 43824, '\p{^Block=ethiopicextendeda}', "");
Expect(1, 43824, '\P{Block=ethiopicextendeda}', "");
Expect(0, 43824, '\P{^Block=ethiopicextendeda}', "");
Expect(1, 43823, '\p{Block=-	ETHIOPIC_Extended_A}', "");
Expect(0, 43823, '\p{^Block=-	ETHIOPIC_Extended_A}', "");
Expect(0, 43823, '\P{Block=-	ETHIOPIC_Extended_A}', "");
Expect(1, 43823, '\P{^Block=-	ETHIOPIC_Extended_A}', "");
Expect(0, 43824, '\p{Block=-	ETHIOPIC_Extended_A}', "");
Expect(1, 43824, '\p{^Block=-	ETHIOPIC_Extended_A}', "");
Expect(1, 43824, '\P{Block=-	ETHIOPIC_Extended_A}', "");
Expect(0, 43824, '\P{^Block=-	ETHIOPIC_Extended_A}', "");
Error('\p{Blk:	/a/--Ethiopic_Ext_A}');
Error('\P{Blk:	/a/--Ethiopic_Ext_A}');
Expect(1, 43823, '\p{Blk: ethiopicexta}', "");
Expect(0, 43823, '\p{^Blk: ethiopicexta}', "");
Expect(0, 43823, '\P{Blk: ethiopicexta}', "");
Expect(1, 43823, '\P{^Blk: ethiopicexta}', "");
Expect(0, 43824, '\p{Blk: ethiopicexta}', "");
Expect(1, 43824, '\p{^Blk: ethiopicexta}', "");
Expect(1, 43824, '\P{Blk: ethiopicexta}', "");
Expect(0, 43824, '\P{^Blk: ethiopicexta}', "");
Expect(1, 43823, '\p{Blk=	-ETHIOPIC_Ext_A}', "");
Expect(0, 43823, '\p{^Blk=	-ETHIOPIC_Ext_A}', "");
Expect(0, 43823, '\P{Blk=	-ETHIOPIC_Ext_A}', "");
Expect(1, 43823, '\P{^Blk=	-ETHIOPIC_Ext_A}', "");
Expect(0, 43824, '\p{Blk=	-ETHIOPIC_Ext_A}', "");
Expect(1, 43824, '\p{^Blk=	-ETHIOPIC_Ext_A}', "");
Expect(1, 43824, '\P{Blk=	-ETHIOPIC_Ext_A}', "");
Expect(0, 43824, '\P{^Blk=	-ETHIOPIC_Ext_A}', "");
Error('\p{Is_Block= ethiopic_Extended_A:=}');
Error('\P{Is_Block= ethiopic_Extended_A:=}');
Expect(1, 43823, '\p{Is_Block=ethiopicextendeda}', "");
Expect(0, 43823, '\p{^Is_Block=ethiopicextendeda}', "");
Expect(0, 43823, '\P{Is_Block=ethiopicextendeda}', "");
Expect(1, 43823, '\P{^Is_Block=ethiopicextendeda}', "");
Expect(0, 43824, '\p{Is_Block=ethiopicextendeda}', "");
Expect(1, 43824, '\p{^Is_Block=ethiopicextendeda}', "");
Expect(1, 43824, '\P{Is_Block=ethiopicextendeda}', "");
Expect(0, 43824, '\P{^Is_Block=ethiopicextendeda}', "");
Expect(1, 43823, '\p{Is_Block:- Ethiopic_extended_A}', "");
Expect(0, 43823, '\p{^Is_Block:- Ethiopic_extended_A}', "");
Expect(0, 43823, '\P{Is_Block:- Ethiopic_extended_A}', "");
Expect(1, 43823, '\P{^Is_Block:- Ethiopic_extended_A}', "");
Expect(0, 43824, '\p{Is_Block:- Ethiopic_extended_A}', "");
Expect(1, 43824, '\p{^Is_Block:- Ethiopic_extended_A}', "");
Expect(1, 43824, '\P{Is_Block:- Ethiopic_extended_A}', "");
Expect(0, 43824, '\P{^Is_Block:- Ethiopic_extended_A}', "");
Error('\p{Is_Blk=/a/	_Ethiopic_ext_A}');
Error('\P{Is_Blk=/a/	_Ethiopic_ext_A}');
Expect(1, 43823, '\p{Is_Blk=ethiopicexta}', "");
Expect(0, 43823, '\p{^Is_Blk=ethiopicexta}', "");
Expect(0, 43823, '\P{Is_Blk=ethiopicexta}', "");
Expect(1, 43823, '\P{^Is_Blk=ethiopicexta}', "");
Expect(0, 43824, '\p{Is_Blk=ethiopicexta}', "");
Expect(1, 43824, '\p{^Is_Blk=ethiopicexta}', "");
Expect(1, 43824, '\P{Is_Blk=ethiopicexta}', "");
Expect(0, 43824, '\P{^Is_Blk=ethiopicexta}', "");
Expect(1, 43823, '\p{Is_Blk=_	Ethiopic_Ext_A}', "");
Expect(0, 43823, '\p{^Is_Blk=_	Ethiopic_Ext_A}', "");
Expect(0, 43823, '\P{Is_Blk=_	Ethiopic_Ext_A}', "");
Expect(1, 43823, '\P{^Is_Blk=_	Ethiopic_Ext_A}', "");
Expect(0, 43824, '\p{Is_Blk=_	Ethiopic_Ext_A}', "");
Expect(1, 43824, '\p{^Is_Blk=_	Ethiopic_Ext_A}', "");
Expect(1, 43824, '\P{Is_Blk=_	Ethiopic_Ext_A}', "");
Expect(0, 43824, '\P{^Is_Blk=_	Ethiopic_Ext_A}', "");
Error('\p{Block=_-Hangul_Jamo_Extended_a:=}');
Error('\P{Block=_-Hangul_Jamo_Extended_a:=}');
Expect(1, 43391, '\p{Block=hanguljamoextendeda}', "");
Expect(0, 43391, '\p{^Block=hanguljamoextendeda}', "");
Expect(0, 43391, '\P{Block=hanguljamoextendeda}', "");
Expect(1, 43391, '\P{^Block=hanguljamoextendeda}', "");
Expect(0, 43392, '\p{Block=hanguljamoextendeda}', "");
Expect(1, 43392, '\p{^Block=hanguljamoextendeda}', "");
Expect(1, 43392, '\P{Block=hanguljamoextendeda}', "");
Expect(0, 43392, '\P{^Block=hanguljamoextendeda}', "");
Expect(1, 43391, '\p{Block=_	hangul_JAMO_EXTENDED_A}', "");
Expect(0, 43391, '\p{^Block=_	hangul_JAMO_EXTENDED_A}', "");
Expect(0, 43391, '\P{Block=_	hangul_JAMO_EXTENDED_A}', "");
Expect(1, 43391, '\P{^Block=_	hangul_JAMO_EXTENDED_A}', "");
Expect(0, 43392, '\p{Block=_	hangul_JAMO_EXTENDED_A}', "");
Expect(1, 43392, '\p{^Block=_	hangul_JAMO_EXTENDED_A}', "");
Expect(1, 43392, '\P{Block=_	hangul_JAMO_EXTENDED_A}', "");
Expect(0, 43392, '\P{^Block=_	hangul_JAMO_EXTENDED_A}', "");
Error('\p{Blk=	/a/jamo_Ext_a}');
Error('\P{Blk=	/a/jamo_Ext_a}');
Expect(1, 43391, '\p{Blk=jamoexta}', "");
Expect(0, 43391, '\p{^Blk=jamoexta}', "");
Expect(0, 43391, '\P{Blk=jamoexta}', "");
Expect(1, 43391, '\P{^Blk=jamoexta}', "");
Expect(0, 43392, '\p{Blk=jamoexta}', "");
Expect(1, 43392, '\p{^Blk=jamoexta}', "");
Expect(1, 43392, '\P{Blk=jamoexta}', "");
Expect(0, 43392, '\P{^Blk=jamoexta}', "");
Expect(1, 43391, '\p{Blk=_ JAMO_Ext_A}', "");
Expect(0, 43391, '\p{^Blk=_ JAMO_Ext_A}', "");
Expect(0, 43391, '\P{Blk=_ JAMO_Ext_A}', "");
Expect(1, 43391, '\P{^Blk=_ JAMO_Ext_A}', "");
Expect(0, 43392, '\p{Blk=_ JAMO_Ext_A}', "");
Expect(1, 43392, '\p{^Blk=_ JAMO_Ext_A}', "");
Expect(1, 43392, '\P{Blk=_ JAMO_Ext_A}', "");
Expect(0, 43392, '\P{^Blk=_ JAMO_Ext_A}', "");
Error('\p{Is_Block=/a/ -HANGUL_jamo_EXTENDED_a}');
Error('\P{Is_Block=/a/ -HANGUL_jamo_EXTENDED_a}');
Expect(1, 43391, '\p{Is_Block=hanguljamoextendeda}', "");
Expect(0, 43391, '\p{^Is_Block=hanguljamoextendeda}', "");
Expect(0, 43391, '\P{Is_Block=hanguljamoextendeda}', "");
Expect(1, 43391, '\P{^Is_Block=hanguljamoextendeda}', "");
Expect(0, 43392, '\p{Is_Block=hanguljamoextendeda}', "");
Expect(1, 43392, '\p{^Is_Block=hanguljamoextendeda}', "");
Expect(1, 43392, '\P{Is_Block=hanguljamoextendeda}', "");
Expect(0, 43392, '\P{^Is_Block=hanguljamoextendeda}', "");
Expect(1, 43391, '\p{Is_Block=_-Hangul_Jamo_Extended_A}', "");
Expect(0, 43391, '\p{^Is_Block=_-Hangul_Jamo_Extended_A}', "");
Expect(0, 43391, '\P{Is_Block=_-Hangul_Jamo_Extended_A}', "");
Expect(1, 43391, '\P{^Is_Block=_-Hangul_Jamo_Extended_A}', "");
Expect(0, 43392, '\p{Is_Block=_-Hangul_Jamo_Extended_A}', "");
Expect(1, 43392, '\p{^Is_Block=_-Hangul_Jamo_Extended_A}', "");
Expect(1, 43392, '\P{Is_Block=_-Hangul_Jamo_Extended_A}', "");
Expect(0, 43392, '\P{^Is_Block=_-Hangul_Jamo_Extended_A}', "");
Error('\p{Is_Blk=_/a/Jamo_EXT_A}');
Error('\P{Is_Blk=_/a/Jamo_EXT_A}');
Expect(1, 43391, '\p{Is_Blk=jamoexta}', "");
Expect(0, 43391, '\p{^Is_Blk=jamoexta}', "");
Expect(0, 43391, '\P{Is_Blk=jamoexta}', "");
Expect(1, 43391, '\P{^Is_Blk=jamoexta}', "");
Expect(0, 43392, '\p{Is_Blk=jamoexta}', "");
Expect(1, 43392, '\p{^Is_Blk=jamoexta}', "");
Expect(1, 43392, '\P{Is_Blk=jamoexta}', "");
Expect(0, 43392, '\P{^Is_Blk=jamoexta}', "");
Expect(1, 43391, '\p{Is_Blk=__jamo_Ext_A}', "");
Expect(0, 43391, '\p{^Is_Blk=__jamo_Ext_A}', "");
Expect(0, 43391, '\P{Is_Blk=__jamo_Ext_A}', "");
Expect(1, 43391, '\P{^Is_Blk=__jamo_Ext_A}', "");
Expect(0, 43392, '\p{Is_Blk=__jamo_Ext_A}', "");
Expect(1, 43392, '\p{^Is_Blk=__jamo_Ext_A}', "");
Expect(1, 43392, '\P{Is_Blk=__jamo_Ext_A}', "");
Expect(0, 43392, '\P{^Is_Blk=__jamo_Ext_A}', "");
Error('\p{Block:	 :=kanbun}');
Error('\P{Block:	 :=kanbun}');
Expect(1, 12703, '\p{Block=kanbun}', "");
Expect(0, 12703, '\p{^Block=kanbun}', "");
Expect(0, 12703, '\P{Block=kanbun}', "");
Expect(1, 12703, '\P{^Block=kanbun}', "");
Expect(0, 12704, '\p{Block=kanbun}', "");
Expect(1, 12704, '\p{^Block=kanbun}', "");
Expect(1, 12704, '\P{Block=kanbun}', "");
Expect(0, 12704, '\P{^Block=kanbun}', "");
Expect(1, 12703, '\p{Block= -kanbun}', "");
Expect(0, 12703, '\p{^Block= -kanbun}', "");
Expect(0, 12703, '\P{Block= -kanbun}', "");
Expect(1, 12703, '\P{^Block= -kanbun}', "");
Expect(0, 12704, '\p{Block= -kanbun}', "");
Expect(1, 12704, '\p{^Block= -kanbun}', "");
Expect(1, 12704, '\P{Block= -kanbun}', "");
Expect(0, 12704, '\P{^Block= -kanbun}', "");
Error('\p{Blk=_	KANBUN:=}');
Error('\P{Blk=_	KANBUN:=}');
Expect(1, 12703, '\p{Blk:kanbun}', "");
Expect(0, 12703, '\p{^Blk:kanbun}', "");
Expect(0, 12703, '\P{Blk:kanbun}', "");
Expect(1, 12703, '\P{^Blk:kanbun}', "");
Expect(0, 12704, '\p{Blk:kanbun}', "");
Expect(1, 12704, '\p{^Blk:kanbun}', "");
Expect(1, 12704, '\P{Blk:kanbun}', "");
Expect(0, 12704, '\P{^Blk:kanbun}', "");
Expect(1, 12703, '\p{Blk=Kanbun}', "");
Expect(0, 12703, '\p{^Blk=Kanbun}', "");
Expect(0, 12703, '\P{Blk=Kanbun}', "");
Expect(1, 12703, '\P{^Blk=Kanbun}', "");
Expect(0, 12704, '\p{Blk=Kanbun}', "");
Expect(1, 12704, '\p{^Blk=Kanbun}', "");
Expect(1, 12704, '\P{Blk=Kanbun}', "");
Expect(0, 12704, '\P{^Blk=Kanbun}', "");
Error('\p{Is_Block=	 Kanbun/a/}');
Error('\P{Is_Block=	 Kanbun/a/}');
Expect(1, 12703, '\p{Is_Block=kanbun}', "");
Expect(0, 12703, '\p{^Is_Block=kanbun}', "");
Expect(0, 12703, '\P{Is_Block=kanbun}', "");
Expect(1, 12703, '\P{^Is_Block=kanbun}', "");
Expect(0, 12704, '\p{Is_Block=kanbun}', "");
Expect(1, 12704, '\p{^Is_Block=kanbun}', "");
Expect(1, 12704, '\P{Is_Block=kanbun}', "");
Expect(0, 12704, '\P{^Is_Block=kanbun}', "");
Expect(1, 12703, '\p{Is_Block=_ KANBUN}', "");
Expect(0, 12703, '\p{^Is_Block=_ KANBUN}', "");
Expect(0, 12703, '\P{Is_Block=_ KANBUN}', "");
Expect(1, 12703, '\P{^Is_Block=_ KANBUN}', "");
Expect(0, 12704, '\p{Is_Block=_ KANBUN}', "");
Expect(1, 12704, '\p{^Is_Block=_ KANBUN}', "");
Expect(1, 12704, '\P{Is_Block=_ KANBUN}', "");
Expect(0, 12704, '\P{^Is_Block=_ KANBUN}', "");
Error('\p{Is_Blk=:=- kanbun}');
Error('\P{Is_Blk=:=- kanbun}');
Expect(1, 12703, '\p{Is_Blk=kanbun}', "");
Expect(0, 12703, '\p{^Is_Blk=kanbun}', "");
Expect(0, 12703, '\P{Is_Blk=kanbun}', "");
Expect(1, 12703, '\P{^Is_Blk=kanbun}', "");
Expect(0, 12704, '\p{Is_Blk=kanbun}', "");
Expect(1, 12704, '\p{^Is_Blk=kanbun}', "");
Expect(1, 12704, '\P{Is_Blk=kanbun}', "");
Expect(0, 12704, '\P{^Is_Blk=kanbun}', "");
Expect(1, 12703, '\p{Is_Blk= _Kanbun}', "");
Expect(0, 12703, '\p{^Is_Blk= _Kanbun}', "");
Expect(0, 12703, '\P{Is_Blk= _Kanbun}', "");
Expect(1, 12703, '\P{^Is_Blk= _Kanbun}', "");
Expect(0, 12704, '\p{Is_Blk= _Kanbun}', "");
Expect(1, 12704, '\p{^Is_Blk= _Kanbun}', "");
Expect(1, 12704, '\P{Is_Blk= _Kanbun}', "");
Expect(0, 12704, '\P{^Is_Blk= _Kanbun}', "");
Error('\p{Block=:= _HEBREW}');
Error('\P{Block=:= _HEBREW}');
Expect(1, 1535, '\p{Block=hebrew}', "");
Expect(0, 1535, '\p{^Block=hebrew}', "");
Expect(0, 1535, '\P{Block=hebrew}', "");
Expect(1, 1535, '\P{^Block=hebrew}', "");
Expect(0, 1536, '\p{Block=hebrew}', "");
Expect(1, 1536, '\p{^Block=hebrew}', "");
Expect(1, 1536, '\P{Block=hebrew}', "");
Expect(0, 1536, '\P{^Block=hebrew}', "");
Expect(1, 1535, '\p{Block=-_Hebrew}', "");
Expect(0, 1535, '\p{^Block=-_Hebrew}', "");
Expect(0, 1535, '\P{Block=-_Hebrew}', "");
Expect(1, 1535, '\P{^Block=-_Hebrew}', "");
Expect(0, 1536, '\p{Block=-_Hebrew}', "");
Expect(1, 1536, '\p{^Block=-_Hebrew}', "");
Expect(1, 1536, '\P{Block=-_Hebrew}', "");
Expect(0, 1536, '\P{^Block=-_Hebrew}', "");
Error('\p{Blk:	-	HEBREW/a/}');
Error('\P{Blk:	-	HEBREW/a/}');
Expect(1, 1535, '\p{Blk:hebrew}', "");
Expect(0, 1535, '\p{^Blk:hebrew}', "");
Expect(0, 1535, '\P{Blk:hebrew}', "");
Expect(1, 1535, '\P{^Blk:hebrew}', "");
Expect(0, 1536, '\p{Blk:hebrew}', "");
Expect(1, 1536, '\p{^Blk:hebrew}', "");
Expect(1, 1536, '\P{Blk:hebrew}', "");
Expect(0, 1536, '\P{^Blk:hebrew}', "");
Expect(1, 1535, '\p{Blk=	-HEBREW}', "");
Expect(0, 1535, '\p{^Blk=	-HEBREW}', "");
Expect(0, 1535, '\P{Blk=	-HEBREW}', "");
Expect(1, 1535, '\P{^Blk=	-HEBREW}', "");
Expect(0, 1536, '\p{Blk=	-HEBREW}', "");
Expect(1, 1536, '\p{^Blk=	-HEBREW}', "");
Expect(1, 1536, '\P{Blk=	-HEBREW}', "");
Expect(0, 1536, '\P{^Blk=	-HEBREW}', "");
Error('\p{Is_Block:	_:=Hebrew}');
Error('\P{Is_Block:	_:=Hebrew}');
Expect(1, 1535, '\p{Is_Block=hebrew}', "");
Expect(0, 1535, '\p{^Is_Block=hebrew}', "");
Expect(0, 1535, '\P{Is_Block=hebrew}', "");
Expect(1, 1535, '\P{^Is_Block=hebrew}', "");
Expect(0, 1536, '\p{Is_Block=hebrew}', "");
Expect(1, 1536, '\p{^Is_Block=hebrew}', "");
Expect(1, 1536, '\P{Is_Block=hebrew}', "");
Expect(0, 1536, '\P{^Is_Block=hebrew}', "");
Expect(1, 1535, '\p{Is_Block=	 Hebrew}', "");
Expect(0, 1535, '\p{^Is_Block=	 Hebrew}', "");
Expect(0, 1535, '\P{Is_Block=	 Hebrew}', "");
Expect(1, 1535, '\P{^Is_Block=	 Hebrew}', "");
Expect(0, 1536, '\p{Is_Block=	 Hebrew}', "");
Expect(1, 1536, '\p{^Is_Block=	 Hebrew}', "");
Expect(1, 1536, '\P{Is_Block=	 Hebrew}', "");
Expect(0, 1536, '\P{^Is_Block=	 Hebrew}', "");
Error('\p{Is_Blk= :=Hebrew}');
Error('\P{Is_Blk= :=Hebrew}');
Expect(1, 1535, '\p{Is_Blk=hebrew}', "");
Expect(0, 1535, '\p{^Is_Blk=hebrew}', "");
Expect(0, 1535, '\P{Is_Blk=hebrew}', "");
Expect(1, 1535, '\P{^Is_Blk=hebrew}', "");
Expect(0, 1536, '\p{Is_Blk=hebrew}', "");
Expect(1, 1536, '\p{^Is_Blk=hebrew}', "");
Expect(1, 1536, '\P{Is_Blk=hebrew}', "");
Expect(0, 1536, '\P{^Is_Blk=hebrew}', "");
Expect(1, 1535, '\p{Is_Blk= _HEBREW}', "");
Expect(0, 1535, '\p{^Is_Blk= _HEBREW}', "");
Expect(0, 1535, '\P{Is_Blk= _HEBREW}', "");
Expect(1, 1535, '\P{^Is_Blk= _HEBREW}', "");
Expect(0, 1536, '\p{Is_Blk= _HEBREW}', "");
Expect(1, 1536, '\p{^Is_Blk= _HEBREW}', "");
Expect(1, 1536, '\P{Is_Blk= _HEBREW}', "");
Expect(0, 1536, '\P{^Is_Blk= _HEBREW}', "");
Error('\p{Block= Cyrillic_Extended_A/a/}');
Error('\P{Block= Cyrillic_Extended_A/a/}');
Expect(1, 11775, '\p{Block=cyrillicextendeda}', "");
Expect(0, 11775, '\p{^Block=cyrillicextendeda}', "");
Expect(0, 11775, '\P{Block=cyrillicextendeda}', "");
Expect(1, 11775, '\P{^Block=cyrillicextendeda}', "");
Expect(0, 11776, '\p{Block=cyrillicextendeda}', "");
Expect(1, 11776, '\p{^Block=cyrillicextendeda}', "");
Expect(1, 11776, '\P{Block=cyrillicextendeda}', "");
Expect(0, 11776, '\P{^Block=cyrillicextendeda}', "");
Expect(1, 11775, '\p{Block=-cyrillic_EXTENDED_A}', "");
Expect(0, 11775, '\p{^Block=-cyrillic_EXTENDED_A}', "");
Expect(0, 11775, '\P{Block=-cyrillic_EXTENDED_A}', "");
Expect(1, 11775, '\P{^Block=-cyrillic_EXTENDED_A}', "");
Expect(0, 11776, '\p{Block=-cyrillic_EXTENDED_A}', "");
Expect(1, 11776, '\p{^Block=-cyrillic_EXTENDED_A}', "");
Expect(1, 11776, '\P{Block=-cyrillic_EXTENDED_A}', "");
Expect(0, 11776, '\P{^Block=-cyrillic_EXTENDED_A}', "");
Error('\p{Blk=	/a/cyrillic_ext_A}');
Error('\P{Blk=	/a/cyrillic_ext_A}');
Expect(1, 11775, '\p{Blk=cyrillicexta}', "");
Expect(0, 11775, '\p{^Blk=cyrillicexta}', "");
Expect(0, 11775, '\P{Blk=cyrillicexta}', "");
Expect(1, 11775, '\P{^Blk=cyrillicexta}', "");
Expect(0, 11776, '\p{Blk=cyrillicexta}', "");
Expect(1, 11776, '\p{^Blk=cyrillicexta}', "");
Expect(1, 11776, '\P{Blk=cyrillicexta}', "");
Expect(0, 11776, '\P{^Blk=cyrillicexta}', "");
Expect(1, 11775, '\p{Blk=  CYRILLIC_Ext_A}', "");
Expect(0, 11775, '\p{^Blk=  CYRILLIC_Ext_A}', "");
Expect(0, 11775, '\P{Blk=  CYRILLIC_Ext_A}', "");
Expect(1, 11775, '\P{^Blk=  CYRILLIC_Ext_A}', "");
Expect(0, 11776, '\p{Blk=  CYRILLIC_Ext_A}', "");
Expect(1, 11776, '\p{^Blk=  CYRILLIC_Ext_A}', "");
Expect(1, 11776, '\P{Blk=  CYRILLIC_Ext_A}', "");
Expect(0, 11776, '\P{^Blk=  CYRILLIC_Ext_A}', "");
Error('\p{Is_Block=	_cyrillic_extended_a:=}');
Error('\P{Is_Block=	_cyrillic_extended_a:=}');
Expect(1, 11775, '\p{Is_Block=cyrillicextendeda}', "");
Expect(0, 11775, '\p{^Is_Block=cyrillicextendeda}', "");
Expect(0, 11775, '\P{Is_Block=cyrillicextendeda}', "");
Expect(1, 11775, '\P{^Is_Block=cyrillicextendeda}', "");
Expect(0, 11776, '\p{Is_Block=cyrillicextendeda}', "");
Expect(1, 11776, '\p{^Is_Block=cyrillicextendeda}', "");
Expect(1, 11776, '\P{Is_Block=cyrillicextendeda}', "");
Expect(0, 11776, '\P{^Is_Block=cyrillicextendeda}', "");
Expect(1, 11775, '\p{Is_Block=	Cyrillic_EXTENDED_A}', "");
Expect(0, 11775, '\p{^Is_Block=	Cyrillic_EXTENDED_A}', "");
Expect(0, 11775, '\P{Is_Block=	Cyrillic_EXTENDED_A}', "");
Expect(1, 11775, '\P{^Is_Block=	Cyrillic_EXTENDED_A}', "");
Expect(0, 11776, '\p{Is_Block=	Cyrillic_EXTENDED_A}', "");
Expect(1, 11776, '\p{^Is_Block=	Cyrillic_EXTENDED_A}', "");
Expect(1, 11776, '\P{Is_Block=	Cyrillic_EXTENDED_A}', "");
Expect(0, 11776, '\P{^Is_Block=	Cyrillic_EXTENDED_A}', "");
Error('\p{Is_Blk=	:=Cyrillic_ext_A}');
Error('\P{Is_Blk=	:=Cyrillic_ext_A}');
Expect(1, 11775, '\p{Is_Blk=cyrillicexta}', "");
Expect(0, 11775, '\p{^Is_Blk=cyrillicexta}', "");
Expect(0, 11775, '\P{Is_Blk=cyrillicexta}', "");
Expect(1, 11775, '\P{^Is_Blk=cyrillicexta}', "");
Expect(0, 11776, '\p{Is_Blk=cyrillicexta}', "");
Expect(1, 11776, '\p{^Is_Blk=cyrillicexta}', "");
Expect(1, 11776, '\P{Is_Blk=cyrillicexta}', "");
Expect(0, 11776, '\P{^Is_Blk=cyrillicexta}', "");
Expect(1, 11775, '\p{Is_Blk=-_Cyrillic_Ext_A}', "");
Expect(0, 11775, '\p{^Is_Blk=-_Cyrillic_Ext_A}', "");
Expect(0, 11775, '\P{Is_Blk=-_Cyrillic_Ext_A}', "");
Expect(1, 11775, '\P{^Is_Blk=-_Cyrillic_Ext_A}', "");
Expect(0, 11776, '\p{Is_Blk=-_Cyrillic_Ext_A}', "");
Expect(1, 11776, '\p{^Is_Blk=-_Cyrillic_Ext_A}', "");
Expect(1, 11776, '\P{Is_Blk=-_Cyrillic_Ext_A}', "");
Expect(0, 11776, '\P{^Is_Blk=-_Cyrillic_Ext_A}', "");
Error('\p{Block: /a/	AVESTAN}');
Error('\P{Block: /a/	AVESTAN}');
Expect(1, 68415, '\p{Block=avestan}', "");
Expect(0, 68415, '\p{^Block=avestan}', "");
Expect(0, 68415, '\P{Block=avestan}', "");
Expect(1, 68415, '\P{^Block=avestan}', "");
Expect(0, 68416, '\p{Block=avestan}', "");
Expect(1, 68416, '\p{^Block=avestan}', "");
Expect(1, 68416, '\P{Block=avestan}', "");
Expect(0, 68416, '\P{^Block=avestan}', "");
Expect(1, 68415, '\p{Block=	 Avestan}', "");
Expect(0, 68415, '\p{^Block=	 Avestan}', "");
Expect(0, 68415, '\P{Block=	 Avestan}', "");
Expect(1, 68415, '\P{^Block=	 Avestan}', "");
Expect(0, 68416, '\p{Block=	 Avestan}', "");
Expect(1, 68416, '\p{^Block=	 Avestan}', "");
Expect(1, 68416, '\P{Block=	 Avestan}', "");
Expect(0, 68416, '\P{^Block=	 Avestan}', "");
Error('\p{Blk=-_AVESTAN/a/}');
Error('\P{Blk=-_AVESTAN/a/}');
Expect(1, 68415, '\p{Blk:	avestan}', "");
Expect(0, 68415, '\p{^Blk:	avestan}', "");
Expect(0, 68415, '\P{Blk:	avestan}', "");
Expect(1, 68415, '\P{^Blk:	avestan}', "");
Expect(0, 68416, '\p{Blk:	avestan}', "");
Expect(1, 68416, '\p{^Blk:	avestan}', "");
Expect(1, 68416, '\P{Blk:	avestan}', "");
Expect(0, 68416, '\P{^Blk:	avestan}', "");
Expect(1, 68415, '\p{Blk=-_Avestan}', "");
Expect(0, 68415, '\p{^Blk=-_Avestan}', "");
Expect(0, 68415, '\P{Blk=-_Avestan}', "");
Expect(1, 68415, '\P{^Blk=-_Avestan}', "");
Expect(0, 68416, '\p{Blk=-_Avestan}', "");
Expect(1, 68416, '\p{^Blk=-_Avestan}', "");
Expect(1, 68416, '\P{Blk=-_Avestan}', "");
Expect(0, 68416, '\P{^Blk=-_Avestan}', "");
Error('\p{Is_Block:/a/-	avestan}');
Error('\P{Is_Block:/a/-	avestan}');
Expect(1, 68415, '\p{Is_Block=avestan}', "");
Expect(0, 68415, '\p{^Is_Block=avestan}', "");
Expect(0, 68415, '\P{Is_Block=avestan}', "");
Expect(1, 68415, '\P{^Is_Block=avestan}', "");
Expect(0, 68416, '\p{Is_Block=avestan}', "");
Expect(1, 68416, '\p{^Is_Block=avestan}', "");
Expect(1, 68416, '\P{Is_Block=avestan}', "");
Expect(0, 68416, '\P{^Is_Block=avestan}', "");
Expect(1, 68415, '\p{Is_Block=_ Avestan}', "");
Expect(0, 68415, '\p{^Is_Block=_ Avestan}', "");
Expect(0, 68415, '\P{Is_Block=_ Avestan}', "");
Expect(1, 68415, '\P{^Is_Block=_ Avestan}', "");
Expect(0, 68416, '\p{Is_Block=_ Avestan}', "");
Expect(1, 68416, '\p{^Is_Block=_ Avestan}', "");
Expect(1, 68416, '\P{Is_Block=_ Avestan}', "");
Expect(0, 68416, '\P{^Is_Block=_ Avestan}', "");
Error('\p{Is_Blk=-/a/avestan}');
Error('\P{Is_Blk=-/a/avestan}');
Expect(1, 68415, '\p{Is_Blk=avestan}', "");
Expect(0, 68415, '\p{^Is_Blk=avestan}', "");
Expect(0, 68415, '\P{Is_Blk=avestan}', "");
Expect(1, 68415, '\P{^Is_Blk=avestan}', "");
Expect(0, 68416, '\p{Is_Blk=avestan}', "");
Expect(1, 68416, '\p{^Is_Blk=avestan}', "");
Expect(1, 68416, '\P{Is_Blk=avestan}', "");
Expect(0, 68416, '\P{^Is_Blk=avestan}', "");
Expect(1, 68415, '\p{Is_Blk=-_AVESTAN}', "");
Expect(0, 68415, '\p{^Is_Blk=-_AVESTAN}', "");
Expect(0, 68415, '\P{Is_Blk=-_AVESTAN}', "");
Expect(1, 68415, '\P{^Is_Blk=-_AVESTAN}', "");
Expect(0, 68416, '\p{Is_Blk=-_AVESTAN}', "");
Expect(1, 68416, '\p{^Is_Blk=-_AVESTAN}', "");
Expect(1, 68416, '\P{Is_Blk=-_AVESTAN}', "");
Expect(0, 68416, '\P{^Is_Blk=-_AVESTAN}', "");
Error('\p{Block=:=	COUNTING_rod_Numerals}');
Error('\P{Block=:=	COUNTING_rod_Numerals}');
Expect(1, 119679, '\p{Block=countingrodnumerals}', "");
Expect(0, 119679, '\p{^Block=countingrodnumerals}', "");
Expect(0, 119679, '\P{Block=countingrodnumerals}', "");
Expect(1, 119679, '\P{^Block=countingrodnumerals}', "");
Expect(0, 119680, '\p{Block=countingrodnumerals}', "");
Expect(1, 119680, '\p{^Block=countingrodnumerals}', "");
Expect(1, 119680, '\P{Block=countingrodnumerals}', "");
Expect(0, 119680, '\P{^Block=countingrodnumerals}', "");
Expect(1, 119679, '\p{Block=_-COUNTING_Rod_Numerals}', "");
Expect(0, 119679, '\p{^Block=_-COUNTING_Rod_Numerals}', "");
Expect(0, 119679, '\P{Block=_-COUNTING_Rod_Numerals}', "");
Expect(1, 119679, '\P{^Block=_-COUNTING_Rod_Numerals}', "");
Expect(0, 119680, '\p{Block=_-COUNTING_Rod_Numerals}', "");
Expect(1, 119680, '\p{^Block=_-COUNTING_Rod_Numerals}', "");
Expect(1, 119680, '\P{Block=_-COUNTING_Rod_Numerals}', "");
Expect(0, 119680, '\P{^Block=_-COUNTING_Rod_Numerals}', "");
Error('\p{Blk=	:=Counting_ROD}');
Error('\P{Blk=	:=Counting_ROD}');
Expect(1, 119679, '\p{Blk=countingrod}', "");
Expect(0, 119679, '\p{^Blk=countingrod}', "");
Expect(0, 119679, '\P{Blk=countingrod}', "");
Expect(1, 119679, '\P{^Blk=countingrod}', "");
Expect(0, 119680, '\p{Blk=countingrod}', "");
Expect(1, 119680, '\p{^Blk=countingrod}', "");
Expect(1, 119680, '\P{Blk=countingrod}', "");
Expect(0, 119680, '\P{^Blk=countingrod}', "");
Expect(1, 119679, '\p{Blk=-COUNTING_rod}', "");
Expect(0, 119679, '\p{^Blk=-COUNTING_rod}', "");
Expect(0, 119679, '\P{Blk=-COUNTING_rod}', "");
Expect(1, 119679, '\P{^Blk=-COUNTING_rod}', "");
Expect(0, 119680, '\p{Blk=-COUNTING_rod}', "");
Expect(1, 119680, '\p{^Blk=-COUNTING_rod}', "");
Expect(1, 119680, '\P{Blk=-COUNTING_rod}', "");
Expect(0, 119680, '\P{^Blk=-COUNTING_rod}', "");
Error('\p{Is_Block:  :=counting_ROD_numerals}');
Error('\P{Is_Block:  :=counting_ROD_numerals}');
Expect(1, 119679, '\p{Is_Block=countingrodnumerals}', "");
Expect(0, 119679, '\p{^Is_Block=countingrodnumerals}', "");
Expect(0, 119679, '\P{Is_Block=countingrodnumerals}', "");
Expect(1, 119679, '\P{^Is_Block=countingrodnumerals}', "");
Expect(0, 119680, '\p{Is_Block=countingrodnumerals}', "");
Expect(1, 119680, '\p{^Is_Block=countingrodnumerals}', "");
Expect(1, 119680, '\P{Is_Block=countingrodnumerals}', "");
Expect(0, 119680, '\P{^Is_Block=countingrodnumerals}', "");
Expect(1, 119679, '\p{Is_Block=	_COUNTING_Rod_numerals}', "");
Expect(0, 119679, '\p{^Is_Block=	_COUNTING_Rod_numerals}', "");
Expect(0, 119679, '\P{Is_Block=	_COUNTING_Rod_numerals}', "");
Expect(1, 119679, '\P{^Is_Block=	_COUNTING_Rod_numerals}', "");
Expect(0, 119680, '\p{Is_Block=	_COUNTING_Rod_numerals}', "");
Expect(1, 119680, '\p{^Is_Block=	_COUNTING_Rod_numerals}', "");
Expect(1, 119680, '\P{Is_Block=	_COUNTING_Rod_numerals}', "");
Expect(0, 119680, '\P{^Is_Block=	_COUNTING_Rod_numerals}', "");
Error('\p{Is_Blk=-:=Counting_Rod}');
Error('\P{Is_Blk=-:=Counting_Rod}');
Expect(1, 119679, '\p{Is_Blk: countingrod}', "");
Expect(0, 119679, '\p{^Is_Blk: countingrod}', "");
Expect(0, 119679, '\P{Is_Blk: countingrod}', "");
Expect(1, 119679, '\P{^Is_Blk: countingrod}', "");
Expect(0, 119680, '\p{Is_Blk: countingrod}', "");
Expect(1, 119680, '\p{^Is_Blk: countingrod}', "");
Expect(1, 119680, '\P{Is_Blk: countingrod}', "");
Expect(0, 119680, '\P{^Is_Blk: countingrod}', "");
Expect(1, 119679, '\p{Is_Blk=	Counting_ROD}', "");
Expect(0, 119679, '\p{^Is_Blk=	Counting_ROD}', "");
Expect(0, 119679, '\P{Is_Blk=	Counting_ROD}', "");
Expect(1, 119679, '\P{^Is_Blk=	Counting_ROD}', "");
Expect(0, 119680, '\p{Is_Blk=	Counting_ROD}', "");
Expect(1, 119680, '\p{^Is_Blk=	Counting_ROD}', "");
Expect(1, 119680, '\P{Is_Blk=	Counting_ROD}', "");
Expect(0, 119680, '\P{^Is_Blk=	Counting_ROD}', "");
Error('\p{Block= -Byzantine_musical_symbols:=}');
Error('\P{Block= -Byzantine_musical_symbols:=}');
Expect(1, 119039, '\p{Block=byzantinemusicalsymbols}', "");
Expect(0, 119039, '\p{^Block=byzantinemusicalsymbols}', "");
Expect(0, 119039, '\P{Block=byzantinemusicalsymbols}', "");
Expect(1, 119039, '\P{^Block=byzantinemusicalsymbols}', "");
Expect(0, 119040, '\p{Block=byzantinemusicalsymbols}', "");
Expect(1, 119040, '\p{^Block=byzantinemusicalsymbols}', "");
Expect(1, 119040, '\P{Block=byzantinemusicalsymbols}', "");
Expect(0, 119040, '\P{^Block=byzantinemusicalsymbols}', "");
Expect(1, 119039, '\p{Block=-_Byzantine_Musical_Symbols}', "");
Expect(0, 119039, '\p{^Block=-_Byzantine_Musical_Symbols}', "");
Expect(0, 119039, '\P{Block=-_Byzantine_Musical_Symbols}', "");
Expect(1, 119039, '\P{^Block=-_Byzantine_Musical_Symbols}', "");
Expect(0, 119040, '\p{Block=-_Byzantine_Musical_Symbols}', "");
Expect(1, 119040, '\p{^Block=-_Byzantine_Musical_Symbols}', "");
Expect(1, 119040, '\P{Block=-_Byzantine_Musical_Symbols}', "");
Expect(0, 119040, '\P{^Block=-_Byzantine_Musical_Symbols}', "");
Error('\p{Blk=	byzantine_Music/a/}');
Error('\P{Blk=	byzantine_Music/a/}');
Expect(1, 119039, '\p{Blk=byzantinemusic}', "");
Expect(0, 119039, '\p{^Blk=byzantinemusic}', "");
Expect(0, 119039, '\P{Blk=byzantinemusic}', "");
Expect(1, 119039, '\P{^Blk=byzantinemusic}', "");
Expect(0, 119040, '\p{Blk=byzantinemusic}', "");
Expect(1, 119040, '\p{^Blk=byzantinemusic}', "");
Expect(1, 119040, '\P{Blk=byzantinemusic}', "");
Expect(0, 119040, '\P{^Blk=byzantinemusic}', "");
Expect(1, 119039, '\p{Blk=-_Byzantine_Music}', "");
Expect(0, 119039, '\p{^Blk=-_Byzantine_Music}', "");
Expect(0, 119039, '\P{Blk=-_Byzantine_Music}', "");
Expect(1, 119039, '\P{^Blk=-_Byzantine_Music}', "");
Expect(0, 119040, '\p{Blk=-_Byzantine_Music}', "");
Expect(1, 119040, '\p{^Blk=-_Byzantine_Music}', "");
Expect(1, 119040, '\P{Blk=-_Byzantine_Music}', "");
Expect(0, 119040, '\P{^Blk=-_Byzantine_Music}', "");
Error('\p{Is_Block=/a/_byzantine_MUSICAL_Symbols}');
Error('\P{Is_Block=/a/_byzantine_MUSICAL_Symbols}');
Expect(1, 119039, '\p{Is_Block=byzantinemusicalsymbols}', "");
Expect(0, 119039, '\p{^Is_Block=byzantinemusicalsymbols}', "");
Expect(0, 119039, '\P{Is_Block=byzantinemusicalsymbols}', "");
Expect(1, 119039, '\P{^Is_Block=byzantinemusicalsymbols}', "");
Expect(0, 119040, '\p{Is_Block=byzantinemusicalsymbols}', "");
Expect(1, 119040, '\p{^Is_Block=byzantinemusicalsymbols}', "");
Expect(1, 119040, '\P{Is_Block=byzantinemusicalsymbols}', "");
Expect(0, 119040, '\P{^Is_Block=byzantinemusicalsymbols}', "");
Expect(1, 119039, '\p{Is_Block= _BYZANTINE_musical_Symbols}', "");
Expect(0, 119039, '\p{^Is_Block= _BYZANTINE_musical_Symbols}', "");
Expect(0, 119039, '\P{Is_Block= _BYZANTINE_musical_Symbols}', "");
Expect(1, 119039, '\P{^Is_Block= _BYZANTINE_musical_Symbols}', "");
Expect(0, 119040, '\p{Is_Block= _BYZANTINE_musical_Symbols}', "");
Expect(1, 119040, '\p{^Is_Block= _BYZANTINE_musical_Symbols}', "");
Expect(1, 119040, '\P{Is_Block= _BYZANTINE_musical_Symbols}', "");
Expect(0, 119040, '\P{^Is_Block= _BYZANTINE_musical_Symbols}', "");
Error('\p{Is_Blk=:= 	Byzantine_music}');
Error('\P{Is_Blk=:= 	Byzantine_music}');
Expect(1, 119039, '\p{Is_Blk=byzantinemusic}', "");
Expect(0, 119039, '\p{^Is_Blk=byzantinemusic}', "");
Expect(0, 119039, '\P{Is_Blk=byzantinemusic}', "");
Expect(1, 119039, '\P{^Is_Blk=byzantinemusic}', "");
Expect(0, 119040, '\p{Is_Blk=byzantinemusic}', "");
Expect(1, 119040, '\p{^Is_Blk=byzantinemusic}', "");
Expect(1, 119040, '\P{Is_Blk=byzantinemusic}', "");
Expect(0, 119040, '\P{^Is_Blk=byzantinemusic}', "");
Expect(1, 119039, '\p{Is_Blk= _Byzantine_Music}', "");
Expect(0, 119039, '\p{^Is_Blk= _Byzantine_Music}', "");
Expect(0, 119039, '\P{Is_Blk= _Byzantine_Music}', "");
Expect(1, 119039, '\P{^Is_Blk= _Byzantine_Music}', "");
Expect(0, 119040, '\p{Is_Blk= _Byzantine_Music}', "");
Expect(1, 119040, '\p{^Is_Blk= _Byzantine_Music}', "");
Expect(1, 119040, '\P{Is_Blk= _Byzantine_Music}', "");
Expect(0, 119040, '\P{^Is_Blk= _Byzantine_Music}', "");
Error('\p{Block=_Tamil:=}');
Error('\P{Block=_Tamil:=}');
Expect(1, 3071, '\p{Block=tamil}', "");
Expect(0, 3071, '\p{^Block=tamil}', "");
Expect(0, 3071, '\P{Block=tamil}', "");
Expect(1, 3071, '\P{^Block=tamil}', "");
Expect(0, 3072, '\p{Block=tamil}', "");
Expect(1, 3072, '\p{^Block=tamil}', "");
Expect(1, 3072, '\P{Block=tamil}', "");
Expect(0, 3072, '\P{^Block=tamil}', "");
Expect(1, 3071, '\p{Block=	 TAMIL}', "");
Expect(0, 3071, '\p{^Block=	 TAMIL}', "");
Expect(0, 3071, '\P{Block=	 TAMIL}', "");
Expect(1, 3071, '\P{^Block=	 TAMIL}', "");
Expect(0, 3072, '\p{Block=	 TAMIL}', "");
Expect(1, 3072, '\p{^Block=	 TAMIL}', "");
Expect(1, 3072, '\P{Block=	 TAMIL}', "");
Expect(0, 3072, '\P{^Block=	 TAMIL}', "");
Error('\p{Blk= 	Tamil/a/}');
Error('\P{Blk= 	Tamil/a/}');
Expect(1, 3071, '\p{Blk=tamil}', "");
Expect(0, 3071, '\p{^Blk=tamil}', "");
Expect(0, 3071, '\P{Blk=tamil}', "");
Expect(1, 3071, '\P{^Blk=tamil}', "");
Expect(0, 3072, '\p{Blk=tamil}', "");
Expect(1, 3072, '\p{^Blk=tamil}', "");
Expect(1, 3072, '\P{Blk=tamil}', "");
Expect(0, 3072, '\P{^Blk=tamil}', "");
Expect(1, 3071, '\p{Blk=	_tamil}', "");
Expect(0, 3071, '\p{^Blk=	_tamil}', "");
Expect(0, 3071, '\P{Blk=	_tamil}', "");
Expect(1, 3071, '\P{^Blk=	_tamil}', "");
Expect(0, 3072, '\p{Blk=	_tamil}', "");
Expect(1, 3072, '\p{^Blk=	_tamil}', "");
Expect(1, 3072, '\P{Blk=	_tamil}', "");
Expect(0, 3072, '\P{^Blk=	_tamil}', "");
Error('\p{Is_Block= -tamil:=}');
Error('\P{Is_Block= -tamil:=}');
Expect(1, 3071, '\p{Is_Block=tamil}', "");
Expect(0, 3071, '\p{^Is_Block=tamil}', "");
Expect(0, 3071, '\P{Is_Block=tamil}', "");
Expect(1, 3071, '\P{^Is_Block=tamil}', "");
Expect(0, 3072, '\p{Is_Block=tamil}', "");
Expect(1, 3072, '\p{^Is_Block=tamil}', "");
Expect(1, 3072, '\P{Is_Block=tamil}', "");
Expect(0, 3072, '\P{^Is_Block=tamil}', "");
Expect(1, 3071, '\p{Is_Block=_-Tamil}', "");
Expect(0, 3071, '\p{^Is_Block=_-Tamil}', "");
Expect(0, 3071, '\P{Is_Block=_-Tamil}', "");
Expect(1, 3071, '\P{^Is_Block=_-Tamil}', "");
Expect(0, 3072, '\p{Is_Block=_-Tamil}', "");
Expect(1, 3072, '\p{^Is_Block=_-Tamil}', "");
Expect(1, 3072, '\P{Is_Block=_-Tamil}', "");
Expect(0, 3072, '\P{^Is_Block=_-Tamil}', "");
Error('\p{Is_Blk= Tamil/a/}');
Error('\P{Is_Blk= Tamil/a/}');
Expect(1, 3071, '\p{Is_Blk=tamil}', "");
Expect(0, 3071, '\p{^Is_Blk=tamil}', "");
Expect(0, 3071, '\P{Is_Blk=tamil}', "");
Expect(1, 3071, '\P{^Is_Blk=tamil}', "");
Expect(0, 3072, '\p{Is_Blk=tamil}', "");
Expect(1, 3072, '\p{^Is_Blk=tamil}', "");
Expect(1, 3072, '\P{Is_Blk=tamil}', "");
Expect(0, 3072, '\P{^Is_Blk=tamil}', "");
Expect(1, 3071, '\p{Is_Blk=--TAMIL}', "");
Expect(0, 3071, '\p{^Is_Blk=--TAMIL}', "");
Expect(0, 3071, '\P{Is_Blk=--TAMIL}', "");
Expect(1, 3071, '\P{^Is_Blk=--TAMIL}', "");
Expect(0, 3072, '\p{Is_Blk=--TAMIL}', "");
Expect(1, 3072, '\p{^Is_Blk=--TAMIL}', "");
Expect(1, 3072, '\P{Is_Blk=--TAMIL}', "");
Expect(0, 3072, '\P{^Is_Blk=--TAMIL}', "");
Error('\p{Block= -Specials/a/}');
Error('\P{Block= -Specials/a/}');
Expect(1, 65520, '\p{Block=specials}', "");
Expect(0, 65520, '\p{^Block=specials}', "");
Expect(0, 65520, '\P{Block=specials}', "");
Expect(1, 65520, '\P{^Block=specials}', "");
Expect(0, 65536, '\p{Block=specials}', "");
Expect(1, 65536, '\p{^Block=specials}', "");
Expect(1, 65536, '\P{Block=specials}', "");
Expect(0, 65536, '\P{^Block=specials}', "");
Expect(1, 65520, '\p{Block=_Specials}', "");
Expect(0, 65520, '\p{^Block=_Specials}', "");
Expect(0, 65520, '\P{Block=_Specials}', "");
Expect(1, 65520, '\P{^Block=_Specials}', "");
Expect(0, 65536, '\p{Block=_Specials}', "");
Expect(1, 65536, '\p{^Block=_Specials}', "");
Expect(1, 65536, '\P{Block=_Specials}', "");
Expect(0, 65536, '\P{^Block=_Specials}', "");
Error('\p{Blk:   /a/	SPECIALS}');
Error('\P{Blk:   /a/	SPECIALS}');
Expect(1, 65520, '\p{Blk:specials}', "");
Expect(0, 65520, '\p{^Blk:specials}', "");
Expect(0, 65520, '\P{Blk:specials}', "");
Expect(1, 65520, '\P{^Blk:specials}', "");
Expect(0, 65536, '\p{Blk:specials}', "");
Expect(1, 65536, '\p{^Blk:specials}', "");
Expect(1, 65536, '\P{Blk:specials}', "");
Expect(0, 65536, '\P{^Blk:specials}', "");
Expect(1, 65520, '\p{Blk=_Specials}', "");
Expect(0, 65520, '\p{^Blk=_Specials}', "");
Expect(0, 65520, '\P{Blk=_Specials}', "");
Expect(1, 65520, '\P{^Blk=_Specials}', "");
Expect(0, 65536, '\p{Blk=_Specials}', "");
Expect(1, 65536, '\p{^Blk=_Specials}', "");
Expect(1, 65536, '\P{Blk=_Specials}', "");
Expect(0, 65536, '\P{^Blk=_Specials}', "");
Error('\p{Is_Block=_specials:=}');
Error('\P{Is_Block=_specials:=}');
Expect(1, 65520, '\p{Is_Block:   specials}', "");
Expect(0, 65520, '\p{^Is_Block:   specials}', "");
Expect(0, 65520, '\P{Is_Block:   specials}', "");
Expect(1, 65520, '\P{^Is_Block:   specials}', "");
Expect(0, 65536, '\p{Is_Block:   specials}', "");
Expect(1, 65536, '\p{^Is_Block:   specials}', "");
Expect(1, 65536, '\P{Is_Block:   specials}', "");
Expect(0, 65536, '\P{^Is_Block:   specials}', "");
Expect(1, 65520, '\p{Is_Block=--Specials}', "");
Expect(0, 65520, '\p{^Is_Block=--Specials}', "");
Expect(0, 65520, '\P{Is_Block=--Specials}', "");
Expect(1, 65520, '\P{^Is_Block=--Specials}', "");
Expect(0, 65536, '\p{Is_Block=--Specials}', "");
Expect(1, 65536, '\p{^Is_Block=--Specials}', "");
Expect(1, 65536, '\P{Is_Block=--Specials}', "");
Expect(0, 65536, '\P{^Is_Block=--Specials}', "");
Error('\p{Is_Blk=/a/-specials}');
Error('\P{Is_Blk=/a/-specials}');
Expect(1, 65520, '\p{Is_Blk:specials}', "");
Expect(0, 65520, '\p{^Is_Blk:specials}', "");
Expect(0, 65520, '\P{Is_Blk:specials}', "");
Expect(1, 65520, '\P{^Is_Blk:specials}', "");
Expect(0, 65536, '\p{Is_Blk:specials}', "");
Expect(1, 65536, '\p{^Is_Blk:specials}', "");
Expect(1, 65536, '\P{Is_Blk:specials}', "");
Expect(0, 65536, '\P{^Is_Blk:specials}', "");
Expect(1, 65520, '\p{Is_Blk= _SPECIALS}', "");
Expect(0, 65520, '\p{^Is_Blk= _SPECIALS}', "");
Expect(0, 65520, '\P{Is_Blk= _SPECIALS}', "");
Expect(1, 65520, '\P{^Is_Blk= _SPECIALS}', "");
Expect(0, 65536, '\p{Is_Blk= _SPECIALS}', "");
Expect(1, 65536, '\p{^Is_Blk= _SPECIALS}', "");
Expect(1, 65536, '\P{Is_Blk= _SPECIALS}', "");
Expect(0, 65536, '\P{^Is_Blk= _SPECIALS}', "");
Error('\p{Block: :=Ethiopic_supplement}');
Error('\P{Block: :=Ethiopic_supplement}');
Expect(1, 5023, '\p{Block=ethiopicsupplement}', "");
Expect(0, 5023, '\p{^Block=ethiopicsupplement}', "");
Expect(0, 5023, '\P{Block=ethiopicsupplement}', "");
Expect(1, 5023, '\P{^Block=ethiopicsupplement}', "");
Expect(0, 5024, '\p{Block=ethiopicsupplement}', "");
Expect(1, 5024, '\p{^Block=ethiopicsupplement}', "");
Expect(1, 5024, '\P{Block=ethiopicsupplement}', "");
Expect(0, 5024, '\P{^Block=ethiopicsupplement}', "");
Expect(1, 5023, '\p{Block= _Ethiopic_supplement}', "");
Expect(0, 5023, '\p{^Block= _Ethiopic_supplement}', "");
Expect(0, 5023, '\P{Block= _Ethiopic_supplement}', "");
Expect(1, 5023, '\P{^Block= _Ethiopic_supplement}', "");
Expect(0, 5024, '\p{Block= _Ethiopic_supplement}', "");
Expect(1, 5024, '\p{^Block= _Ethiopic_supplement}', "");
Expect(1, 5024, '\P{Block= _Ethiopic_supplement}', "");
Expect(0, 5024, '\P{^Block= _Ethiopic_supplement}', "");
Error('\p{Blk=	-Ethiopic_SUP:=}');
Error('\P{Blk=	-Ethiopic_SUP:=}');
Expect(1, 5023, '\p{Blk=ethiopicsup}', "");
Expect(0, 5023, '\p{^Blk=ethiopicsup}', "");
Expect(0, 5023, '\P{Blk=ethiopicsup}', "");
Expect(1, 5023, '\P{^Blk=ethiopicsup}', "");
Expect(0, 5024, '\p{Blk=ethiopicsup}', "");
Expect(1, 5024, '\p{^Blk=ethiopicsup}', "");
Expect(1, 5024, '\P{Blk=ethiopicsup}', "");
Expect(0, 5024, '\P{^Blk=ethiopicsup}', "");
Expect(1, 5023, '\p{Blk=_-Ethiopic_Sup}', "");
Expect(0, 5023, '\p{^Blk=_-Ethiopic_Sup}', "");
Expect(0, 5023, '\P{Blk=_-Ethiopic_Sup}', "");
Expect(1, 5023, '\P{^Blk=_-Ethiopic_Sup}', "");
Expect(0, 5024, '\p{Blk=_-Ethiopic_Sup}', "");
Expect(1, 5024, '\p{^Blk=_-Ethiopic_Sup}', "");
Expect(1, 5024, '\P{Blk=_-Ethiopic_Sup}', "");
Expect(0, 5024, '\P{^Blk=_-Ethiopic_Sup}', "");
Error('\p{Is_Block=_-Ethiopic_supplement/a/}');
Error('\P{Is_Block=_-Ethiopic_supplement/a/}');
Expect(1, 5023, '\p{Is_Block=ethiopicsupplement}', "");
Expect(0, 5023, '\p{^Is_Block=ethiopicsupplement}', "");
Expect(0, 5023, '\P{Is_Block=ethiopicsupplement}', "");
Expect(1, 5023, '\P{^Is_Block=ethiopicsupplement}', "");
Expect(0, 5024, '\p{Is_Block=ethiopicsupplement}', "");
Expect(1, 5024, '\p{^Is_Block=ethiopicsupplement}', "");
Expect(1, 5024, '\P{Is_Block=ethiopicsupplement}', "");
Expect(0, 5024, '\P{^Is_Block=ethiopicsupplement}', "");
Expect(1, 5023, '\p{Is_Block=	-ethiopic_Supplement}', "");
Expect(0, 5023, '\p{^Is_Block=	-ethiopic_Supplement}', "");
Expect(0, 5023, '\P{Is_Block=	-ethiopic_Supplement}', "");
Expect(1, 5023, '\P{^Is_Block=	-ethiopic_Supplement}', "");
Expect(0, 5024, '\p{Is_Block=	-ethiopic_Supplement}', "");
Expect(1, 5024, '\p{^Is_Block=	-ethiopic_Supplement}', "");
Expect(1, 5024, '\P{Is_Block=	-ethiopic_Supplement}', "");
Expect(0, 5024, '\P{^Is_Block=	-ethiopic_Supplement}', "");
Error('\p{Is_Blk= _Ethiopic_sup:=}');
Error('\P{Is_Blk= _Ethiopic_sup:=}');
Expect(1, 5023, '\p{Is_Blk=ethiopicsup}', "");
Expect(0, 5023, '\p{^Is_Blk=ethiopicsup}', "");
Expect(0, 5023, '\P{Is_Blk=ethiopicsup}', "");
Expect(1, 5023, '\P{^Is_Blk=ethiopicsup}', "");
Expect(0, 5024, '\p{Is_Blk=ethiopicsup}', "");
Expect(1, 5024, '\p{^Is_Blk=ethiopicsup}', "");
Expect(1, 5024, '\P{Is_Blk=ethiopicsup}', "");
Expect(0, 5024, '\P{^Is_Blk=ethiopicsup}', "");
Expect(1, 5023, '\p{Is_Blk=Ethiopic_Sup}', "");
Expect(0, 5023, '\p{^Is_Blk=Ethiopic_Sup}', "");
Expect(0, 5023, '\P{Is_Blk=Ethiopic_Sup}', "");
Expect(1, 5023, '\P{^Is_Blk=Ethiopic_Sup}', "");
Expect(0, 5024, '\p{Is_Blk=Ethiopic_Sup}', "");
Expect(1, 5024, '\p{^Is_Blk=Ethiopic_Sup}', "");
Expect(1, 5024, '\P{Is_Blk=Ethiopic_Sup}', "");
Expect(0, 5024, '\P{^Is_Blk=Ethiopic_Sup}', "");
Error('\p{Block= RUNIC/a/}');
Error('\P{Block= RUNIC/a/}');
Expect(1, 5887, '\p{Block=runic}', "");
Expect(0, 5887, '\p{^Block=runic}', "");
Expect(0, 5887, '\P{Block=runic}', "");
Expect(1, 5887, '\P{^Block=runic}', "");
Expect(0, 5888, '\p{Block=runic}', "");
Expect(1, 5888, '\p{^Block=runic}', "");
Expect(1, 5888, '\P{Block=runic}', "");
Expect(0, 5888, '\P{^Block=runic}', "");
Expect(1, 5887, '\p{Block:	_RUNIC}', "");
Expect(0, 5887, '\p{^Block:	_RUNIC}', "");
Expect(0, 5887, '\P{Block:	_RUNIC}', "");
Expect(1, 5887, '\P{^Block:	_RUNIC}', "");
Expect(0, 5888, '\p{Block:	_RUNIC}', "");
Expect(1, 5888, '\p{^Block:	_RUNIC}', "");
Expect(1, 5888, '\P{Block:	_RUNIC}', "");
Expect(0, 5888, '\P{^Block:	_RUNIC}', "");
Error('\p{Blk=/a/	Runic}');
Error('\P{Blk=/a/	Runic}');
Expect(1, 5887, '\p{Blk=runic}', "");
Expect(0, 5887, '\p{^Blk=runic}', "");
Expect(0, 5887, '\P{Blk=runic}', "");
Expect(1, 5887, '\P{^Blk=runic}', "");
Expect(0, 5888, '\p{Blk=runic}', "");
Expect(1, 5888, '\p{^Blk=runic}', "");
Expect(1, 5888, '\P{Blk=runic}', "");
Expect(0, 5888, '\P{^Blk=runic}', "");
Expect(1, 5887, '\p{Blk=_-Runic}', "");
Expect(0, 5887, '\p{^Blk=_-Runic}', "");
Expect(0, 5887, '\P{Blk=_-Runic}', "");
Expect(1, 5887, '\P{^Blk=_-Runic}', "");
Expect(0, 5888, '\p{Blk=_-Runic}', "");
Expect(1, 5888, '\p{^Blk=_-Runic}', "");
Expect(1, 5888, '\P{Blk=_-Runic}', "");
Expect(0, 5888, '\P{^Blk=_-Runic}', "");
Error('\p{Is_Block= :=RUNIC}');
Error('\P{Is_Block= :=RUNIC}');
Expect(1, 5887, '\p{Is_Block=runic}', "");
Expect(0, 5887, '\p{^Is_Block=runic}', "");
Expect(0, 5887, '\P{Is_Block=runic}', "");
Expect(1, 5887, '\P{^Is_Block=runic}', "");
Expect(0, 5888, '\p{Is_Block=runic}', "");
Expect(1, 5888, '\p{^Is_Block=runic}', "");
Expect(1, 5888, '\P{Is_Block=runic}', "");
Expect(0, 5888, '\P{^Is_Block=runic}', "");
Expect(1, 5887, '\p{Is_Block= runic}', "");
Expect(0, 5887, '\p{^Is_Block= runic}', "");
Expect(0, 5887, '\P{Is_Block= runic}', "");
Expect(1, 5887, '\P{^Is_Block= runic}', "");
Expect(0, 5888, '\p{Is_Block= runic}', "");
Expect(1, 5888, '\p{^Is_Block= runic}', "");
Expect(1, 5888, '\P{Is_Block= runic}', "");
Expect(0, 5888, '\P{^Is_Block= runic}', "");
Error('\p{Is_Blk=/a/ 	Runic}');
Error('\P{Is_Blk=/a/ 	Runic}');
Expect(1, 5887, '\p{Is_Blk=runic}', "");
Expect(0, 5887, '\p{^Is_Blk=runic}', "");
Expect(0, 5887, '\P{Is_Blk=runic}', "");
Expect(1, 5887, '\P{^Is_Blk=runic}', "");
Expect(0, 5888, '\p{Is_Blk=runic}', "");
Expect(1, 5888, '\p{^Is_Blk=runic}', "");
Expect(1, 5888, '\P{Is_Blk=runic}', "");
Expect(0, 5888, '\P{^Is_Blk=runic}', "");
Expect(1, 5887, '\p{Is_Blk: _ RUNIC}', "");
Expect(0, 5887, '\p{^Is_Blk: _ RUNIC}', "");
Expect(0, 5887, '\P{Is_Blk: _ RUNIC}', "");
Expect(1, 5887, '\P{^Is_Blk: _ RUNIC}', "");
Expect(0, 5888, '\p{Is_Blk: _ RUNIC}', "");
Expect(1, 5888, '\p{^Is_Blk: _ RUNIC}', "");
Expect(1, 5888, '\P{Is_Blk: _ RUNIC}', "");
Expect(0, 5888, '\P{^Is_Blk: _ RUNIC}', "");
Error('\p{Block=:=		Emoticons}');
Error('\P{Block=:=		Emoticons}');
Expect(1, 128591, '\p{Block=emoticons}', "");
Expect(0, 128591, '\p{^Block=emoticons}', "");
Expect(0, 128591, '\P{Block=emoticons}', "");
Expect(1, 128591, '\P{^Block=emoticons}', "");
Expect(0, 128592, '\p{Block=emoticons}', "");
Expect(1, 128592, '\p{^Block=emoticons}', "");
Expect(1, 128592, '\P{Block=emoticons}', "");
Expect(0, 128592, '\P{^Block=emoticons}', "");
Expect(1, 128591, '\p{Block=  EMOTICONS}', "");
Expect(0, 128591, '\p{^Block=  EMOTICONS}', "");
Expect(0, 128591, '\P{Block=  EMOTICONS}', "");
Expect(1, 128591, '\P{^Block=  EMOTICONS}', "");
Expect(0, 128592, '\p{Block=  EMOTICONS}', "");
Expect(1, 128592, '\p{^Block=  EMOTICONS}', "");
Expect(1, 128592, '\P{Block=  EMOTICONS}', "");
Expect(0, 128592, '\P{^Block=  EMOTICONS}', "");
Error('\p{Blk=_:=Emoticons}');
Error('\P{Blk=_:=Emoticons}');
Expect(1, 128591, '\p{Blk: emoticons}', "");
Expect(0, 128591, '\p{^Blk: emoticons}', "");
Expect(0, 128591, '\P{Blk: emoticons}', "");
Expect(1, 128591, '\P{^Blk: emoticons}', "");
Expect(0, 128592, '\p{Blk: emoticons}', "");
Expect(1, 128592, '\p{^Blk: emoticons}', "");
Expect(1, 128592, '\P{Blk: emoticons}', "");
Expect(0, 128592, '\P{^Blk: emoticons}', "");
Expect(1, 128591, '\p{Blk=_emoticons}', "");
Expect(0, 128591, '\p{^Blk=_emoticons}', "");
Expect(0, 128591, '\P{Blk=_emoticons}', "");
Expect(1, 128591, '\P{^Blk=_emoticons}', "");
Expect(0, 128592, '\p{Blk=_emoticons}', "");
Expect(1, 128592, '\p{^Blk=_emoticons}', "");
Expect(1, 128592, '\P{Blk=_emoticons}', "");
Expect(0, 128592, '\P{^Blk=_emoticons}', "");
Error('\p{Is_Block= :=Emoticons}');
Error('\P{Is_Block= :=Emoticons}');
Expect(1, 128591, '\p{Is_Block=emoticons}', "");
Expect(0, 128591, '\p{^Is_Block=emoticons}', "");
Expect(0, 128591, '\P{Is_Block=emoticons}', "");
Expect(1, 128591, '\P{^Is_Block=emoticons}', "");
Expect(0, 128592, '\p{Is_Block=emoticons}', "");
Expect(1, 128592, '\p{^Is_Block=emoticons}', "");
Expect(1, 128592, '\P{Is_Block=emoticons}', "");
Expect(0, 128592, '\P{^Is_Block=emoticons}', "");
Expect(1, 128591, '\p{Is_Block=- Emoticons}', "");
Expect(0, 128591, '\p{^Is_Block=- Emoticons}', "");
Expect(0, 128591, '\P{Is_Block=- Emoticons}', "");
Expect(1, 128591, '\P{^Is_Block=- Emoticons}', "");
Expect(0, 128592, '\p{Is_Block=- Emoticons}', "");
Expect(1, 128592, '\p{^Is_Block=- Emoticons}', "");
Expect(1, 128592, '\P{Is_Block=- Emoticons}', "");
Expect(0, 128592, '\P{^Is_Block=- Emoticons}', "");
Error('\p{Is_Blk=--EMOTICONS:=}');
Error('\P{Is_Blk=--EMOTICONS:=}');
Expect(1, 128591, '\p{Is_Blk:	emoticons}', "");
Expect(0, 128591, '\p{^Is_Blk:	emoticons}', "");
Expect(0, 128591, '\P{Is_Blk:	emoticons}', "");
Expect(1, 128591, '\P{^Is_Blk:	emoticons}', "");
Expect(0, 128592, '\p{Is_Blk:	emoticons}', "");
Expect(1, 128592, '\p{^Is_Blk:	emoticons}', "");
Expect(1, 128592, '\P{Is_Blk:	emoticons}', "");
Expect(0, 128592, '\P{^Is_Blk:	emoticons}', "");
Expect(1, 128591, '\p{Is_Blk=- EMOTICONS}', "");
Expect(0, 128591, '\p{^Is_Blk=- EMOTICONS}', "");
Expect(0, 128591, '\P{Is_Blk=- EMOTICONS}', "");
Expect(1, 128591, '\P{^Is_Blk=- EMOTICONS}', "");
Expect(0, 128592, '\p{Is_Blk=- EMOTICONS}', "");
Expect(1, 128592, '\p{^Is_Blk=- EMOTICONS}', "");
Expect(1, 128592, '\P{Is_Blk=- EMOTICONS}', "");
Expect(0, 128592, '\P{^Is_Blk=- EMOTICONS}', "");
Error('\p{Block=	:=Arabic}');
Error('\P{Block=	:=Arabic}');
Expect(1, 1791, '\p{Block=arabic}', "");
Expect(0, 1791, '\p{^Block=arabic}', "");
Expect(0, 1791, '\P{Block=arabic}', "");
Expect(1, 1791, '\P{^Block=arabic}', "");
Expect(0, 1792, '\p{Block=arabic}', "");
Expect(1, 1792, '\p{^Block=arabic}', "");
Expect(1, 1792, '\P{Block=arabic}', "");
Expect(0, 1792, '\P{^Block=arabic}', "");
Expect(1, 1791, '\p{Block=	_arabic}', "");
Expect(0, 1791, '\p{^Block=	_arabic}', "");
Expect(0, 1791, '\P{Block=	_arabic}', "");
Expect(1, 1791, '\P{^Block=	_arabic}', "");
Expect(0, 1792, '\p{Block=	_arabic}', "");
Expect(1, 1792, '\p{^Block=	_arabic}', "");
Expect(1, 1792, '\P{Block=	_arabic}', "");
Expect(0, 1792, '\P{^Block=	_arabic}', "");
Error('\p{Blk=_:=arabic}');
Error('\P{Blk=_:=arabic}');
Expect(1, 1791, '\p{Blk=arabic}', "");
Expect(0, 1791, '\p{^Blk=arabic}', "");
Expect(0, 1791, '\P{Blk=arabic}', "");
Expect(1, 1791, '\P{^Blk=arabic}', "");
Expect(0, 1792, '\p{Blk=arabic}', "");
Expect(1, 1792, '\p{^Blk=arabic}', "");
Expect(1, 1792, '\P{Blk=arabic}', "");
Expect(0, 1792, '\P{^Blk=arabic}', "");
Expect(1, 1791, '\p{Blk=	Arabic}', "");
Expect(0, 1791, '\p{^Blk=	Arabic}', "");
Expect(0, 1791, '\P{Blk=	Arabic}', "");
Expect(1, 1791, '\P{^Blk=	Arabic}', "");
Expect(0, 1792, '\p{Blk=	Arabic}', "");
Expect(1, 1792, '\p{^Blk=	Arabic}', "");
Expect(1, 1792, '\P{Blk=	Arabic}', "");
Expect(0, 1792, '\P{^Blk=	Arabic}', "");
Error('\p{Is_Block= :=Arabic}');
Error('\P{Is_Block= :=Arabic}');
Expect(1, 1791, '\p{Is_Block=arabic}', "");
Expect(0, 1791, '\p{^Is_Block=arabic}', "");
Expect(0, 1791, '\P{Is_Block=arabic}', "");
Expect(1, 1791, '\P{^Is_Block=arabic}', "");
Expect(0, 1792, '\p{Is_Block=arabic}', "");
Expect(1, 1792, '\p{^Is_Block=arabic}', "");
Expect(1, 1792, '\P{Is_Block=arabic}', "");
Expect(0, 1792, '\P{^Is_Block=arabic}', "");
Expect(1, 1791, '\p{Is_Block=--arabic}', "");
Expect(0, 1791, '\p{^Is_Block=--arabic}', "");
Expect(0, 1791, '\P{Is_Block=--arabic}', "");
Expect(1, 1791, '\P{^Is_Block=--arabic}', "");
Expect(0, 1792, '\p{Is_Block=--arabic}', "");
Expect(1, 1792, '\p{^Is_Block=--arabic}', "");
Expect(1, 1792, '\P{Is_Block=--arabic}', "");
Expect(0, 1792, '\P{^Is_Block=--arabic}', "");
Error('\p{Is_Blk=-	arabic/a/}');
Error('\P{Is_Blk=-	arabic/a/}');
Expect(1, 1791, '\p{Is_Blk=arabic}', "");
Expect(0, 1791, '\p{^Is_Blk=arabic}', "");
Expect(0, 1791, '\P{Is_Blk=arabic}', "");
Expect(1, 1791, '\P{^Is_Blk=arabic}', "");
Expect(0, 1792, '\p{Is_Blk=arabic}', "");
Expect(1, 1792, '\p{^Is_Blk=arabic}', "");
Expect(1, 1792, '\P{Is_Blk=arabic}', "");
Expect(0, 1792, '\P{^Is_Blk=arabic}', "");
Expect(1, 1791, '\p{Is_Blk=--arabic}', "");
Expect(0, 1791, '\p{^Is_Blk=--arabic}', "");
Expect(0, 1791, '\P{Is_Blk=--arabic}', "");
Expect(1, 1791, '\P{^Is_Blk=--arabic}', "");
Expect(0, 1792, '\p{Is_Blk=--arabic}', "");
Expect(1, 1792, '\p{^Is_Blk=--arabic}', "");
Expect(1, 1792, '\P{Is_Blk=--arabic}', "");
Expect(0, 1792, '\P{^Is_Blk=--arabic}', "");
Error('\p{Block= cyrillic_Supplement/a/}');
Error('\P{Block= cyrillic_Supplement/a/}');
Expect(1, 1327, '\p{Block=cyrillicsupplement}', "");
Expect(0, 1327, '\p{^Block=cyrillicsupplement}', "");
Expect(0, 1327, '\P{Block=cyrillicsupplement}', "");
Expect(1, 1327, '\P{^Block=cyrillicsupplement}', "");
Expect(0, 1328, '\p{Block=cyrillicsupplement}', "");
Expect(1, 1328, '\p{^Block=cyrillicsupplement}', "");
Expect(1, 1328, '\P{Block=cyrillicsupplement}', "");
Expect(0, 1328, '\P{^Block=cyrillicsupplement}', "");
Expect(1, 1327, '\p{Block=  cyrillic_Supplement}', "");
Expect(0, 1327, '\p{^Block=  cyrillic_Supplement}', "");
Expect(0, 1327, '\P{Block=  cyrillic_Supplement}', "");
Expect(1, 1327, '\P{^Block=  cyrillic_Supplement}', "");
Expect(0, 1328, '\p{Block=  cyrillic_Supplement}', "");
Expect(1, 1328, '\p{^Block=  cyrillic_Supplement}', "");
Expect(1, 1328, '\P{Block=  cyrillic_Supplement}', "");
Expect(0, 1328, '\P{^Block=  cyrillic_Supplement}', "");
Error('\p{Blk=	Cyrillic_Sup/a/}');
Error('\P{Blk=	Cyrillic_Sup/a/}');
Expect(1, 1327, '\p{Blk=cyrillicsup}', "");
Expect(0, 1327, '\p{^Blk=cyrillicsup}', "");
Expect(0, 1327, '\P{Blk=cyrillicsup}', "");
Expect(1, 1327, '\P{^Blk=cyrillicsup}', "");
Expect(0, 1328, '\p{Blk=cyrillicsup}', "");
Expect(1, 1328, '\p{^Blk=cyrillicsup}', "");
Expect(1, 1328, '\P{Blk=cyrillicsup}', "");
Expect(0, 1328, '\P{^Blk=cyrillicsup}', "");
Expect(1, 1327, '\p{Blk=	 cyrillic_Sup}', "");
Expect(0, 1327, '\p{^Blk=	 cyrillic_Sup}', "");
Expect(0, 1327, '\P{Blk=	 cyrillic_Sup}', "");
Expect(1, 1327, '\P{^Blk=	 cyrillic_Sup}', "");
Expect(0, 1328, '\p{Blk=	 cyrillic_Sup}', "");
Expect(1, 1328, '\p{^Blk=	 cyrillic_Sup}', "");
Expect(1, 1328, '\P{Blk=	 cyrillic_Sup}', "");
Expect(0, 1328, '\P{^Blk=	 cyrillic_Sup}', "");
Error('\p{Is_Block=  cyrillic_Supplementary/a/}');
Error('\P{Is_Block=  cyrillic_Supplementary/a/}');
Expect(1, 1327, '\p{Is_Block=cyrillicsupplementary}', "");
Expect(0, 1327, '\p{^Is_Block=cyrillicsupplementary}', "");
Expect(0, 1327, '\P{Is_Block=cyrillicsupplementary}', "");
Expect(1, 1327, '\P{^Is_Block=cyrillicsupplementary}', "");
Expect(0, 1328, '\p{Is_Block=cyrillicsupplementary}', "");
Expect(1, 1328, '\p{^Is_Block=cyrillicsupplementary}', "");
Expect(1, 1328, '\P{Is_Block=cyrillicsupplementary}', "");
Expect(0, 1328, '\P{^Is_Block=cyrillicsupplementary}', "");
Expect(1, 1327, '\p{Is_Block=-cyrillic_SUPPLEMENTARY}', "");
Expect(0, 1327, '\p{^Is_Block=-cyrillic_SUPPLEMENTARY}', "");
Expect(0, 1327, '\P{Is_Block=-cyrillic_SUPPLEMENTARY}', "");
Expect(1, 1327, '\P{^Is_Block=-cyrillic_SUPPLEMENTARY}', "");
Expect(0, 1328, '\p{Is_Block=-cyrillic_SUPPLEMENTARY}', "");
Expect(1, 1328, '\p{^Is_Block=-cyrillic_SUPPLEMENTARY}', "");
Expect(1, 1328, '\P{Is_Block=-cyrillic_SUPPLEMENTARY}', "");
Expect(0, 1328, '\P{^Is_Block=-cyrillic_SUPPLEMENTARY}', "");
Error('\p{Is_Blk=:= _Cyrillic_SUPPLEMENT}');
Error('\P{Is_Blk=:= _Cyrillic_SUPPLEMENT}');
Expect(1, 1327, '\p{Is_Blk=cyrillicsupplement}', "");
Expect(0, 1327, '\p{^Is_Blk=cyrillicsupplement}', "");
Expect(0, 1327, '\P{Is_Blk=cyrillicsupplement}', "");
Expect(1, 1327, '\P{^Is_Blk=cyrillicsupplement}', "");
Expect(0, 1328, '\p{Is_Blk=cyrillicsupplement}', "");
Expect(1, 1328, '\p{^Is_Blk=cyrillicsupplement}', "");
Expect(1, 1328, '\P{Is_Blk=cyrillicsupplement}', "");
Expect(0, 1328, '\P{^Is_Blk=cyrillicsupplement}', "");
Expect(1, 1327, '\p{Is_Blk= _Cyrillic_SUPPLEMENT}', "");
Expect(0, 1327, '\p{^Is_Blk= _Cyrillic_SUPPLEMENT}', "");
Expect(0, 1327, '\P{Is_Blk= _Cyrillic_SUPPLEMENT}', "");
Expect(1, 1327, '\P{^Is_Blk= _Cyrillic_SUPPLEMENT}', "");
Expect(0, 1328, '\p{Is_Blk= _Cyrillic_SUPPLEMENT}', "");
Expect(1, 1328, '\p{^Is_Blk= _Cyrillic_SUPPLEMENT}', "");
Expect(1, 1328, '\P{Is_Blk= _Cyrillic_SUPPLEMENT}', "");
Expect(0, 1328, '\P{^Is_Blk= _Cyrillic_SUPPLEMENT}', "");
Error('\p{Block=-_transport_and_Map_Symbols:=}');
Error('\P{Block=-_transport_and_Map_Symbols:=}');
Expect(1, 128767, '\p{Block:   transportandmapsymbols}', "");
Expect(0, 128767, '\p{^Block:   transportandmapsymbols}', "");
Expect(0, 128767, '\P{Block:   transportandmapsymbols}', "");
Expect(1, 128767, '\P{^Block:   transportandmapsymbols}', "");
Expect(0, 128768, '\p{Block:   transportandmapsymbols}', "");
Expect(1, 128768, '\p{^Block:   transportandmapsymbols}', "");
Expect(1, 128768, '\P{Block:   transportandmapsymbols}', "");
Expect(0, 128768, '\P{^Block:   transportandmapsymbols}', "");
Expect(1, 128767, '\p{Block=  Transport_and_MAP_Symbols}', "");
Expect(0, 128767, '\p{^Block=  Transport_and_MAP_Symbols}', "");
Expect(0, 128767, '\P{Block=  Transport_and_MAP_Symbols}', "");
Expect(1, 128767, '\P{^Block=  Transport_and_MAP_Symbols}', "");
Expect(0, 128768, '\p{Block=  Transport_and_MAP_Symbols}', "");
Expect(1, 128768, '\p{^Block=  Transport_and_MAP_Symbols}', "");
Expect(1, 128768, '\P{Block=  Transport_and_MAP_Symbols}', "");
Expect(0, 128768, '\P{^Block=  Transport_and_MAP_Symbols}', "");
Error('\p{Blk:  :=Transport_and_map}');
Error('\P{Blk:  :=Transport_and_map}');
Expect(1, 128767, '\p{Blk=transportandmap}', "");
Expect(0, 128767, '\p{^Blk=transportandmap}', "");
Expect(0, 128767, '\P{Blk=transportandmap}', "");
Expect(1, 128767, '\P{^Blk=transportandmap}', "");
Expect(0, 128768, '\p{Blk=transportandmap}', "");
Expect(1, 128768, '\p{^Blk=transportandmap}', "");
Expect(1, 128768, '\P{Blk=transportandmap}', "");
Expect(0, 128768, '\P{^Blk=transportandmap}', "");
Expect(1, 128767, '\p{Blk= _TRANSPORT_And_Map}', "");
Expect(0, 128767, '\p{^Blk= _TRANSPORT_And_Map}', "");
Expect(0, 128767, '\P{Blk= _TRANSPORT_And_Map}', "");
Expect(1, 128767, '\P{^Blk= _TRANSPORT_And_Map}', "");
Expect(0, 128768, '\p{Blk= _TRANSPORT_And_Map}', "");
Expect(1, 128768, '\p{^Blk= _TRANSPORT_And_Map}', "");
Expect(1, 128768, '\P{Blk= _TRANSPORT_And_Map}', "");
Expect(0, 128768, '\P{^Blk= _TRANSPORT_And_Map}', "");
Error('\p{Is_Block=_:=Transport_And_Map_SYMBOLS}');
Error('\P{Is_Block=_:=Transport_And_Map_SYMBOLS}');
Expect(1, 128767, '\p{Is_Block:	transportandmapsymbols}', "");
Expect(0, 128767, '\p{^Is_Block:	transportandmapsymbols}', "");
Expect(0, 128767, '\P{Is_Block:	transportandmapsymbols}', "");
Expect(1, 128767, '\P{^Is_Block:	transportandmapsymbols}', "");
Expect(0, 128768, '\p{Is_Block:	transportandmapsymbols}', "");
Expect(1, 128768, '\p{^Is_Block:	transportandmapsymbols}', "");
Expect(1, 128768, '\P{Is_Block:	transportandmapsymbols}', "");
Expect(0, 128768, '\P{^Is_Block:	transportandmapsymbols}', "");
Expect(1, 128767, '\p{Is_Block=_Transport_AND_MAP_SYMBOLS}', "");
Expect(0, 128767, '\p{^Is_Block=_Transport_AND_MAP_SYMBOLS}', "");
Expect(0, 128767, '\P{Is_Block=_Transport_AND_MAP_SYMBOLS}', "");
Expect(1, 128767, '\P{^Is_Block=_Transport_AND_MAP_SYMBOLS}', "");
Expect(0, 128768, '\p{Is_Block=_Transport_AND_MAP_SYMBOLS}', "");
Expect(1, 128768, '\p{^Is_Block=_Transport_AND_MAP_SYMBOLS}', "");
Expect(1, 128768, '\P{Is_Block=_Transport_AND_MAP_SYMBOLS}', "");
Expect(0, 128768, '\P{^Is_Block=_Transport_AND_MAP_SYMBOLS}', "");
Error('\p{Is_Blk=:=	_Transport_AND_map}');
Error('\P{Is_Blk=:=	_Transport_AND_map}');
Expect(1, 128767, '\p{Is_Blk: transportandmap}', "");
Expect(0, 128767, '\p{^Is_Blk: transportandmap}', "");
Expect(0, 128767, '\P{Is_Blk: transportandmap}', "");
Expect(1, 128767, '\P{^Is_Blk: transportandmap}', "");
Expect(0, 128768, '\p{Is_Blk: transportandmap}', "");
Expect(1, 128768, '\p{^Is_Blk: transportandmap}', "");
Expect(1, 128768, '\P{Is_Blk: transportandmap}', "");
Expect(0, 128768, '\P{^Is_Blk: transportandmap}', "");
Expect(1, 128767, '\p{Is_Blk= 	TRANSPORT_And_MAP}', "");
Expect(0, 128767, '\p{^Is_Blk= 	TRANSPORT_And_MAP}', "");
Expect(0, 128767, '\P{Is_Blk= 	TRANSPORT_And_MAP}', "");
Expect(1, 128767, '\P{^Is_Blk= 	TRANSPORT_And_MAP}', "");
Expect(0, 128768, '\p{Is_Blk= 	TRANSPORT_And_MAP}', "");
Expect(1, 128768, '\p{^Is_Blk= 	TRANSPORT_And_MAP}', "");
Expect(1, 128768, '\P{Is_Blk= 	TRANSPORT_And_MAP}', "");
Expect(0, 128768, '\P{^Is_Blk= 	TRANSPORT_And_MAP}', "");
Error('\p{Block:   old_South_arabian:=}');
Error('\P{Block:   old_South_arabian:=}');
Expect(1, 68223, '\p{Block=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Block=oldsoutharabian}', "");
Expect(0, 68223, '\P{Block=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Block=oldsoutharabian}', "");
Expect(0, 68224, '\p{Block=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Block=oldsoutharabian}', "");
Expect(1, 68224, '\P{Block=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Block=oldsoutharabian}', "");
Expect(1, 68223, '\p{Block:   _ OLD_SOUTH_arabian}', "");
Expect(0, 68223, '\p{^Block:   _ OLD_SOUTH_arabian}', "");
Expect(0, 68223, '\P{Block:   _ OLD_SOUTH_arabian}', "");
Expect(1, 68223, '\P{^Block:   _ OLD_SOUTH_arabian}', "");
Expect(0, 68224, '\p{Block:   _ OLD_SOUTH_arabian}', "");
Expect(1, 68224, '\p{^Block:   _ OLD_SOUTH_arabian}', "");
Expect(1, 68224, '\P{Block:   _ OLD_SOUTH_arabian}', "");
Expect(0, 68224, '\P{^Block:   _ OLD_SOUTH_arabian}', "");
Error('\p{Blk=	:=Old_south_Arabian}');
Error('\P{Blk=	:=Old_south_Arabian}');
Expect(1, 68223, '\p{Blk=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Blk=oldsoutharabian}', "");
Expect(0, 68223, '\P{Blk=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Blk=oldsoutharabian}', "");
Expect(0, 68224, '\p{Blk=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Blk=oldsoutharabian}', "");
Expect(1, 68224, '\P{Blk=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Blk=oldsoutharabian}', "");
Expect(1, 68223, '\p{Blk= Old_South_Arabian}', "");
Expect(0, 68223, '\p{^Blk= Old_South_Arabian}', "");
Expect(0, 68223, '\P{Blk= Old_South_Arabian}', "");
Expect(1, 68223, '\P{^Blk= Old_South_Arabian}', "");
Expect(0, 68224, '\p{Blk= Old_South_Arabian}', "");
Expect(1, 68224, '\p{^Blk= Old_South_Arabian}', "");
Expect(1, 68224, '\P{Blk= Old_South_Arabian}', "");
Expect(0, 68224, '\P{^Blk= Old_South_Arabian}', "");
Error('\p{Is_Block=/a/ _Old_South_arabian}');
Error('\P{Is_Block=/a/ _Old_South_arabian}');
Expect(1, 68223, '\p{Is_Block=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Is_Block=oldsoutharabian}', "");
Expect(0, 68223, '\P{Is_Block=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Is_Block=oldsoutharabian}', "");
Expect(0, 68224, '\p{Is_Block=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Is_Block=oldsoutharabian}', "");
Expect(1, 68224, '\P{Is_Block=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Is_Block=oldsoutharabian}', "");
Expect(1, 68223, '\p{Is_Block=  Old_south_Arabian}', "");
Expect(0, 68223, '\p{^Is_Block=  Old_south_Arabian}', "");
Expect(0, 68223, '\P{Is_Block=  Old_south_Arabian}', "");
Expect(1, 68223, '\P{^Is_Block=  Old_south_Arabian}', "");
Expect(0, 68224, '\p{Is_Block=  Old_south_Arabian}', "");
Expect(1, 68224, '\p{^Is_Block=  Old_south_Arabian}', "");
Expect(1, 68224, '\P{Is_Block=  Old_south_Arabian}', "");
Expect(0, 68224, '\P{^Is_Block=  Old_south_Arabian}', "");
Error('\p{Is_Blk=	:=Old_south_ARABIAN}');
Error('\P{Is_Blk=	:=Old_south_ARABIAN}');
Expect(1, 68223, '\p{Is_Blk=oldsoutharabian}', "");
Expect(0, 68223, '\p{^Is_Blk=oldsoutharabian}', "");
Expect(0, 68223, '\P{Is_Blk=oldsoutharabian}', "");
Expect(1, 68223, '\P{^Is_Blk=oldsoutharabian}', "");
Expect(0, 68224, '\p{Is_Blk=oldsoutharabian}', "");
Expect(1, 68224, '\p{^Is_Blk=oldsoutharabian}', "");
Expect(1, 68224, '\P{Is_Blk=oldsoutharabian}', "");
Expect(0, 68224, '\P{^Is_Blk=oldsoutharabian}', "");
Expect(1, 68223, '\p{Is_Blk=-	OLD_SOUTH_Arabian}', "");
Expect(0, 68223, '\p{^Is_Blk=-	OLD_SOUTH_Arabian}', "");
Expect(0, 68223, '\P{Is_Blk=-	OLD_SOUTH_Arabian}', "");
Expect(1, 68223, '\P{^Is_Blk=-	OLD_SOUTH_Arabian}', "");
Expect(0, 68224, '\p{Is_Blk=-	OLD_SOUTH_Arabian}', "");
Expect(1, 68224, '\p{^Is_Blk=-	OLD_SOUTH_Arabian}', "");
Expect(1, 68224, '\P{Is_Blk=-	OLD_SOUTH_Arabian}', "");
Expect(0, 68224, '\P{^Is_Blk=-	OLD_SOUTH_Arabian}', "");
Error('\p{Block=-:=NUMBER_Forms}');
Error('\P{Block=-:=NUMBER_Forms}');
Expect(1, 8591, '\p{Block=numberforms}', "");
Expect(0, 8591, '\p{^Block=numberforms}', "");
Expect(0, 8591, '\P{Block=numberforms}', "");
Expect(1, 8591, '\P{^Block=numberforms}', "");
Expect(0, 8592, '\p{Block=numberforms}', "");
Expect(1, 8592, '\p{^Block=numberforms}', "");
Expect(1, 8592, '\P{Block=numberforms}', "");
Expect(0, 8592, '\P{^Block=numberforms}', "");
Expect(1, 8591, '\p{Block=- NUMBER_forms}', "");
Expect(0, 8591, '\p{^Block=- NUMBER_forms}', "");
Expect(0, 8591, '\P{Block=- NUMBER_forms}', "");
Expect(1, 8591, '\P{^Block=- NUMBER_forms}', "");
Expect(0, 8592, '\p{Block=- NUMBER_forms}', "");
Expect(1, 8592, '\p{^Block=- NUMBER_forms}', "");
Expect(1, 8592, '\P{Block=- NUMBER_forms}', "");
Expect(0, 8592, '\P{^Block=- NUMBER_forms}', "");
Error('\p{Blk=_-Number_Forms/a/}');
Error('\P{Blk=_-Number_Forms/a/}');
Expect(1, 8591, '\p{Blk=numberforms}', "");
Expect(0, 8591, '\p{^Blk=numberforms}', "");
Expect(0, 8591, '\P{Blk=numberforms}', "");
Expect(1, 8591, '\P{^Blk=numberforms}', "");
Expect(0, 8592, '\p{Blk=numberforms}', "");
Expect(1, 8592, '\p{^Blk=numberforms}', "");
Expect(1, 8592, '\P{Blk=numberforms}', "");
Expect(0, 8592, '\P{^Blk=numberforms}', "");
Expect(1, 8591, '\p{Blk=	number_Forms}', "");
Expect(0, 8591, '\p{^Blk=	number_Forms}', "");
Expect(0, 8591, '\P{Blk=	number_Forms}', "");
Expect(1, 8591, '\P{^Blk=	number_Forms}', "");
Expect(0, 8592, '\p{Blk=	number_Forms}', "");
Expect(1, 8592, '\p{^Blk=	number_Forms}', "");
Expect(1, 8592, '\P{Blk=	number_Forms}', "");
Expect(0, 8592, '\P{^Blk=	number_Forms}', "");
Error('\p{Is_Block=:=_ Number_Forms}');
Error('\P{Is_Block=:=_ Number_Forms}');
Expect(1, 8591, '\p{Is_Block=numberforms}', "");
Expect(0, 8591, '\p{^Is_Block=numberforms}', "");
Expect(0, 8591, '\P{Is_Block=numberforms}', "");
Expect(1, 8591, '\P{^Is_Block=numberforms}', "");
Expect(0, 8592, '\p{Is_Block=numberforms}', "");
Expect(1, 8592, '\p{^Is_Block=numberforms}', "");
Expect(1, 8592, '\P{Is_Block=numberforms}', "");
Expect(0, 8592, '\P{^Is_Block=numberforms}', "");
Expect(1, 8591, '\p{Is_Block=	Number_Forms}', "");
Expect(0, 8591, '\p{^Is_Block=	Number_Forms}', "");
Expect(0, 8591, '\P{Is_Block=	Number_Forms}', "");
Expect(1, 8591, '\P{^Is_Block=	Number_Forms}', "");
Expect(0, 8592, '\p{Is_Block=	Number_Forms}', "");
Expect(1, 8592, '\p{^Is_Block=	Number_Forms}', "");
Expect(1, 8592, '\P{Is_Block=	Number_Forms}', "");
Expect(0, 8592, '\P{^Is_Block=	Number_Forms}', "");
Error('\p{Is_Blk=/a/-number_FORMS}');
Error('\P{Is_Blk=/a/-number_FORMS}');
Expect(1, 8591, '\p{Is_Blk:	numberforms}', "");
Expect(0, 8591, '\p{^Is_Blk:	numberforms}', "");
Expect(0, 8591, '\P{Is_Blk:	numberforms}', "");
Expect(1, 8591, '\P{^Is_Blk:	numberforms}', "");
Expect(0, 8592, '\p{Is_Blk:	numberforms}', "");
Expect(1, 8592, '\p{^Is_Blk:	numberforms}', "");
Expect(1, 8592, '\P{Is_Blk:	numberforms}', "");
Expect(0, 8592, '\P{^Is_Blk:	numberforms}', "");
Expect(1, 8591, '\p{Is_Blk=--Number_forms}', "");
Expect(0, 8591, '\p{^Is_Blk=--Number_forms}', "");
Expect(0, 8591, '\P{Is_Blk=--Number_forms}', "");
Expect(1, 8591, '\P{^Is_Blk=--Number_forms}', "");
Expect(0, 8592, '\p{Is_Blk=--Number_forms}', "");
Expect(1, 8592, '\p{^Is_Blk=--Number_forms}', "");
Expect(1, 8592, '\P{Is_Blk=--Number_forms}', "");
Expect(0, 8592, '\P{^Is_Blk=--Number_forms}', "");
Error('\p{Block= 	Deseret:=}');
Error('\P{Block= 	Deseret:=}');
Expect(1, 66639, '\p{Block=deseret}', "");
Expect(0, 66639, '\p{^Block=deseret}', "");
Expect(0, 66639, '\P{Block=deseret}', "");
Expect(1, 66639, '\P{^Block=deseret}', "");
Expect(0, 66640, '\p{Block=deseret}', "");
Expect(1, 66640, '\p{^Block=deseret}', "");
Expect(1, 66640, '\P{Block=deseret}', "");
Expect(0, 66640, '\P{^Block=deseret}', "");
Expect(1, 66639, '\p{Block=_	deseret}', "");
Expect(0, 66639, '\p{^Block=_	deseret}', "");
Expect(0, 66639, '\P{Block=_	deseret}', "");
Expect(1, 66639, '\P{^Block=_	deseret}', "");
Expect(0, 66640, '\p{Block=_	deseret}', "");
Expect(1, 66640, '\p{^Block=_	deseret}', "");
Expect(1, 66640, '\P{Block=_	deseret}', "");
Expect(0, 66640, '\P{^Block=_	deseret}', "");
Error('\p{Blk=/a/-deseret}');
Error('\P{Blk=/a/-deseret}');
Expect(1, 66639, '\p{Blk=deseret}', "");
Expect(0, 66639, '\p{^Blk=deseret}', "");
Expect(0, 66639, '\P{Blk=deseret}', "");
Expect(1, 66639, '\P{^Blk=deseret}', "");
Expect(0, 66640, '\p{Blk=deseret}', "");
Expect(1, 66640, '\p{^Blk=deseret}', "");
Expect(1, 66640, '\P{Blk=deseret}', "");
Expect(0, 66640, '\P{^Blk=deseret}', "");
Expect(1, 66639, '\p{Blk= _Deseret}', "");
Expect(0, 66639, '\p{^Blk= _Deseret}', "");
Expect(0, 66639, '\P{Blk= _Deseret}', "");
Expect(1, 66639, '\P{^Blk= _Deseret}', "");
Expect(0, 66640, '\p{Blk= _Deseret}', "");
Expect(1, 66640, '\p{^Blk= _Deseret}', "");
Expect(1, 66640, '\P{Blk= _Deseret}', "");
Expect(0, 66640, '\P{^Blk= _Deseret}', "");
Error('\p{Is_Block=:=Deseret}');
Error('\P{Is_Block=:=Deseret}');
Expect(1, 66639, '\p{Is_Block=deseret}', "");
Expect(0, 66639, '\p{^Is_Block=deseret}', "");
Expect(0, 66639, '\P{Is_Block=deseret}', "");
Expect(1, 66639, '\P{^Is_Block=deseret}', "");
Expect(0, 66640, '\p{Is_Block=deseret}', "");
Expect(1, 66640, '\p{^Is_Block=deseret}', "");
Expect(1, 66640, '\P{Is_Block=deseret}', "");
Expect(0, 66640, '\P{^Is_Block=deseret}', "");
Expect(1, 66639, '\p{Is_Block=Deseret}', "");
Expect(0, 66639, '\p{^Is_Block=Deseret}', "");
Expect(0, 66639, '\P{Is_Block=Deseret}', "");
Expect(1, 66639, '\P{^Is_Block=Deseret}', "");
Expect(0, 66640, '\p{Is_Block=Deseret}', "");
Expect(1, 66640, '\p{^Is_Block=Deseret}', "");
Expect(1, 66640, '\P{Is_Block=Deseret}', "");
Expect(0, 66640, '\P{^Is_Block=Deseret}', "");
Error('\p{Is_Blk= 	deseret:=}');
Error('\P{Is_Blk= 	deseret:=}');
Expect(1, 66639, '\p{Is_Blk=deseret}', "");
Expect(0, 66639, '\p{^Is_Blk=deseret}', "");
Expect(0, 66639, '\P{Is_Blk=deseret}', "");
Expect(1, 66639, '\P{^Is_Blk=deseret}', "");
Expect(0, 66640, '\p{Is_Blk=deseret}', "");
Expect(1, 66640, '\p{^Is_Blk=deseret}', "");
Expect(1, 66640, '\P{Is_Blk=deseret}', "");
Expect(0, 66640, '\P{^Is_Blk=deseret}', "");
Expect(1, 66639, '\p{Is_Blk=	 deseret}', "");
Expect(0, 66639, '\p{^Is_Blk=	 deseret}', "");
Expect(0, 66639, '\P{Is_Blk=	 deseret}', "");
Expect(1, 66639, '\P{^Is_Blk=	 deseret}', "");
Expect(0, 66640, '\p{Is_Blk=	 deseret}', "");
Expect(1, 66640, '\p{^Is_Blk=	 deseret}', "");
Expect(1, 66640, '\P{Is_Blk=	 deseret}', "");
Expect(0, 66640, '\P{^Is_Blk=	 deseret}', "");
Error('\p{Block:-:=Tifinagh}');
Error('\P{Block:-:=Tifinagh}');
Expect(1, 11647, '\p{Block=tifinagh}', "");
Expect(0, 11647, '\p{^Block=tifinagh}', "");
Expect(0, 11647, '\P{Block=tifinagh}', "");
Expect(1, 11647, '\P{^Block=tifinagh}', "");
Expect(0, 11648, '\p{Block=tifinagh}', "");
Expect(1, 11648, '\p{^Block=tifinagh}', "");
Expect(1, 11648, '\P{Block=tifinagh}', "");
Expect(0, 11648, '\P{^Block=tifinagh}', "");
Expect(1, 11647, '\p{Block=	 tifinagh}', "");
Expect(0, 11647, '\p{^Block=	 tifinagh}', "");
Expect(0, 11647, '\P{Block=	 tifinagh}', "");
Expect(1, 11647, '\P{^Block=	 tifinagh}', "");
Expect(0, 11648, '\p{Block=	 tifinagh}', "");
Expect(1, 11648, '\p{^Block=	 tifinagh}', "");
Expect(1, 11648, '\P{Block=	 tifinagh}', "");
Expect(0, 11648, '\P{^Block=	 tifinagh}', "");
Error('\p{Blk=__Tifinagh/a/}');
Error('\P{Blk=__Tifinagh/a/}');
Expect(1, 11647, '\p{Blk:tifinagh}', "");
Expect(0, 11647, '\p{^Blk:tifinagh}', "");
Expect(0, 11647, '\P{Blk:tifinagh}', "");
Expect(1, 11647, '\P{^Blk:tifinagh}', "");
Expect(0, 11648, '\p{Blk:tifinagh}', "");
Expect(1, 11648, '\p{^Blk:tifinagh}', "");
Expect(1, 11648, '\P{Blk:tifinagh}', "");
Expect(0, 11648, '\P{^Blk:tifinagh}', "");
Expect(1, 11647, '\p{Blk=  Tifinagh}', "");
Expect(0, 11647, '\p{^Blk=  Tifinagh}', "");
Expect(0, 11647, '\P{Blk=  Tifinagh}', "");
Expect(1, 11647, '\P{^Blk=  Tifinagh}', "");
Expect(0, 11648, '\p{Blk=  Tifinagh}', "");
Expect(1, 11648, '\p{^Blk=  Tifinagh}', "");
Expect(1, 11648, '\P{Blk=  Tifinagh}', "");
Expect(0, 11648, '\P{^Blk=  Tifinagh}', "");
Error('\p{Is_Block=- Tifinagh:=}');
Error('\P{Is_Block=- Tifinagh:=}');
Expect(1, 11647, '\p{Is_Block=tifinagh}', "");
Expect(0, 11647, '\p{^Is_Block=tifinagh}', "");
Expect(0, 11647, '\P{Is_Block=tifinagh}', "");
Expect(1, 11647, '\P{^Is_Block=tifinagh}', "");
Expect(0, 11648, '\p{Is_Block=tifinagh}', "");
Expect(1, 11648, '\p{^Is_Block=tifinagh}', "");
Expect(1, 11648, '\P{Is_Block=tifinagh}', "");
Expect(0, 11648, '\P{^Is_Block=tifinagh}', "");
Expect(1, 11647, '\p{Is_Block=	TIFINAGH}', "");
Expect(0, 11647, '\p{^Is_Block=	TIFINAGH}', "");
Expect(0, 11647, '\P{Is_Block=	TIFINAGH}', "");
Expect(1, 11647, '\P{^Is_Block=	TIFINAGH}', "");
Expect(0, 11648, '\p{Is_Block=	TIFINAGH}', "");
Expect(1, 11648, '\p{^Is_Block=	TIFINAGH}', "");
Expect(1, 11648, '\P{Is_Block=	TIFINAGH}', "");
Expect(0, 11648, '\P{^Is_Block=	TIFINAGH}', "");
Error('\p{Is_Blk=/a/ -Tifinagh}');
Error('\P{Is_Blk=/a/ -Tifinagh}');
Expect(1, 11647, '\p{Is_Blk=tifinagh}', "");
Expect(0, 11647, '\p{^Is_Blk=tifinagh}', "");
Expect(0, 11647, '\P{Is_Blk=tifinagh}', "");
Expect(1, 11647, '\P{^Is_Blk=tifinagh}', "");
Expect(0, 11648, '\p{Is_Blk=tifinagh}', "");
Expect(1, 11648, '\p{^Is_Blk=tifinagh}', "");
Expect(1, 11648, '\P{Is_Blk=tifinagh}', "");
Expect(0, 11648, '\P{^Is_Blk=tifinagh}', "");
Expect(1, 11647, '\p{Is_Blk= 	Tifinagh}', "");
Expect(0, 11647, '\p{^Is_Blk= 	Tifinagh}', "");
Expect(0, 11647, '\P{Is_Blk= 	Tifinagh}', "");
Expect(1, 11647, '\P{^Is_Blk= 	Tifinagh}', "");
Expect(0, 11648, '\p{Is_Blk= 	Tifinagh}', "");
Expect(1, 11648, '\p{^Is_Blk= 	Tifinagh}', "");
Expect(1, 11648, '\P{Is_Blk= 	Tifinagh}', "");
Expect(0, 11648, '\P{^Is_Blk= 	Tifinagh}', "");
Error('\p{Block=	_takri/a/}');
Error('\P{Block=	_takri/a/}');
Expect(1, 71375, '\p{Block=takri}', "");
Expect(0, 71375, '\p{^Block=takri}', "");
Expect(0, 71375, '\P{Block=takri}', "");
Expect(1, 71375, '\P{^Block=takri}', "");
Expect(0, 71376, '\p{Block=takri}', "");
Expect(1, 71376, '\p{^Block=takri}', "");
Expect(1, 71376, '\P{Block=takri}', "");
Expect(0, 71376, '\P{^Block=takri}', "");
Expect(1, 71375, '\p{Block=  Takri}', "");
Expect(0, 71375, '\p{^Block=  Takri}', "");
Expect(0, 71375, '\P{Block=  Takri}', "");
Expect(1, 71375, '\P{^Block=  Takri}', "");
Expect(0, 71376, '\p{Block=  Takri}', "");
Expect(1, 71376, '\p{^Block=  Takri}', "");
Expect(1, 71376, '\P{Block=  Takri}', "");
Expect(0, 71376, '\P{^Block=  Takri}', "");
Error('\p{Blk=:=	_takri}');
Error('\P{Blk=:=	_takri}');
Expect(1, 71375, '\p{Blk=takri}', "");
Expect(0, 71375, '\p{^Blk=takri}', "");
Expect(0, 71375, '\P{Blk=takri}', "");
Expect(1, 71375, '\P{^Blk=takri}', "");
Expect(0, 71376, '\p{Blk=takri}', "");
Expect(1, 71376, '\p{^Blk=takri}', "");
Expect(1, 71376, '\P{Blk=takri}', "");
Expect(0, 71376, '\P{^Blk=takri}', "");
Expect(1, 71375, '\p{Blk=	Takri}', "");
Expect(0, 71375, '\p{^Blk=	Takri}', "");
Expect(0, 71375, '\P{Blk=	Takri}', "");
Expect(1, 71375, '\P{^Blk=	Takri}', "");
Expect(0, 71376, '\p{Blk=	Takri}', "");
Expect(1, 71376, '\p{^Blk=	Takri}', "");
Expect(1, 71376, '\P{Blk=	Takri}', "");
Expect(0, 71376, '\P{^Blk=	Takri}', "");
Error('\p{Is_Block=:=takri}');
Error('\P{Is_Block=:=takri}');
Expect(1, 71375, '\p{Is_Block=takri}', "");
Expect(0, 71375, '\p{^Is_Block=takri}', "");
Expect(0, 71375, '\P{Is_Block=takri}', "");
Expect(1, 71375, '\P{^Is_Block=takri}', "");
Expect(0, 71376, '\p{Is_Block=takri}', "");
Expect(1, 71376, '\p{^Is_Block=takri}', "");
Expect(1, 71376, '\P{Is_Block=takri}', "");
Expect(0, 71376, '\P{^Is_Block=takri}', "");
Expect(1, 71375, '\p{Is_Block= _takri}', "");
Expect(0, 71375, '\p{^Is_Block= _takri}', "");
Expect(0, 71375, '\P{Is_Block= _takri}', "");
Expect(1, 71375, '\P{^Is_Block= _takri}', "");
Expect(0, 71376, '\p{Is_Block= _takri}', "");
Expect(1, 71376, '\p{^Is_Block= _takri}', "");
Expect(1, 71376, '\P{Is_Block= _takri}', "");
Expect(0, 71376, '\P{^Is_Block= _takri}', "");
Error('\p{Is_Blk=_takri:=}');
Error('\P{Is_Blk=_takri:=}');
Expect(1, 71375, '\p{Is_Blk=takri}', "");
Expect(0, 71375, '\p{^Is_Blk=takri}', "");
Expect(0, 71375, '\P{Is_Blk=takri}', "");
Expect(1, 71375, '\P{^Is_Blk=takri}', "");
Expect(0, 71376, '\p{Is_Blk=takri}', "");
Expect(1, 71376, '\p{^Is_Blk=takri}', "");
Expect(1, 71376, '\P{Is_Blk=takri}', "");
Expect(0, 71376, '\P{^Is_Blk=takri}', "");
Expect(1, 71375, '\p{Is_Blk=-	Takri}', "");
Expect(0, 71375, '\p{^Is_Blk=-	Takri}', "");
Expect(0, 71375, '\P{Is_Blk=-	Takri}', "");
Expect(1, 71375, '\P{^Is_Blk=-	Takri}', "");
Expect(0, 71376, '\p{Is_Blk=-	Takri}', "");
Expect(1, 71376, '\p{^Is_Blk=-	Takri}', "");
Expect(1, 71376, '\P{Is_Blk=-	Takri}', "");
Expect(0, 71376, '\P{^Is_Blk=-	Takri}', "");
Error('\p{Block=:= -CJK_compatibility}');
Error('\P{Block=:= -CJK_compatibility}');
Expect(1, 13311, '\p{Block=cjkcompatibility}', "");
Expect(0, 13311, '\p{^Block=cjkcompatibility}', "");
Expect(0, 13311, '\P{Block=cjkcompatibility}', "");
Expect(1, 13311, '\P{^Block=cjkcompatibility}', "");
Expect(0, 13312, '\p{Block=cjkcompatibility}', "");
Expect(1, 13312, '\p{^Block=cjkcompatibility}', "");
Expect(1, 13312, '\P{Block=cjkcompatibility}', "");
Expect(0, 13312, '\P{^Block=cjkcompatibility}', "");
Expect(1, 13311, '\p{Block=_ CJK_Compatibility}', "");
Expect(0, 13311, '\p{^Block=_ CJK_Compatibility}', "");
Expect(0, 13311, '\P{Block=_ CJK_Compatibility}', "");
Expect(1, 13311, '\P{^Block=_ CJK_Compatibility}', "");
Expect(0, 13312, '\p{Block=_ CJK_Compatibility}', "");
Expect(1, 13312, '\p{^Block=_ CJK_Compatibility}', "");
Expect(1, 13312, '\P{Block=_ CJK_Compatibility}', "");
Expect(0, 13312, '\P{^Block=_ CJK_Compatibility}', "");
Error('\p{Blk=/a/-CJK_Compat}');
Error('\P{Blk=/a/-CJK_Compat}');
Expect(1, 13311, '\p{Blk=cjkcompat}', "");
Expect(0, 13311, '\p{^Blk=cjkcompat}', "");
Expect(0, 13311, '\P{Blk=cjkcompat}', "");
Expect(1, 13311, '\P{^Blk=cjkcompat}', "");
Expect(0, 13312, '\p{Blk=cjkcompat}', "");
Expect(1, 13312, '\p{^Blk=cjkcompat}', "");
Expect(1, 13312, '\P{Blk=cjkcompat}', "");
Expect(0, 13312, '\P{^Blk=cjkcompat}', "");
Expect(1, 13311, '\p{Blk=_ cjk_Compat}', "");
Expect(0, 13311, '\p{^Blk=_ cjk_Compat}', "");
Expect(0, 13311, '\P{Blk=_ cjk_Compat}', "");
Expect(1, 13311, '\P{^Blk=_ cjk_Compat}', "");
Expect(0, 13312, '\p{Blk=_ cjk_Compat}', "");
Expect(1, 13312, '\p{^Blk=_ cjk_Compat}', "");
Expect(1, 13312, '\P{Blk=_ cjk_Compat}', "");
Expect(0, 13312, '\P{^Blk=_ cjk_Compat}', "");
Error('\p{Is_Block:	-CJK_COMPATIBILITY/a/}');
Error('\P{Is_Block:	-CJK_COMPATIBILITY/a/}');
Expect(1, 13311, '\p{Is_Block=cjkcompatibility}', "");
Expect(0, 13311, '\p{^Is_Block=cjkcompatibility}', "");
Expect(0, 13311, '\P{Is_Block=cjkcompatibility}', "");
Expect(1, 13311, '\P{^Is_Block=cjkcompatibility}', "");
Expect(0, 13312, '\p{Is_Block=cjkcompatibility}', "");
Expect(1, 13312, '\p{^Is_Block=cjkcompatibility}', "");
Expect(1, 13312, '\P{Is_Block=cjkcompatibility}', "");
Expect(0, 13312, '\P{^Is_Block=cjkcompatibility}', "");
Expect(1, 13311, '\p{Is_Block:		cjk_compatibility}', "");
Expect(0, 13311, '\p{^Is_Block:		cjk_compatibility}', "");
Expect(0, 13311, '\P{Is_Block:		cjk_compatibility}', "");
Expect(1, 13311, '\P{^Is_Block:		cjk_compatibility}', "");
Expect(0, 13312, '\p{Is_Block:		cjk_compatibility}', "");
Expect(1, 13312, '\p{^Is_Block:		cjk_compatibility}', "");
Expect(1, 13312, '\P{Is_Block:		cjk_compatibility}', "");
Expect(0, 13312, '\P{^Is_Block:		cjk_compatibility}', "");
Error('\p{Is_Blk=_-CJK_COMPAT/a/}');
Error('\P{Is_Blk=_-CJK_COMPAT/a/}');
Expect(1, 13311, '\p{Is_Blk=cjkcompat}', "");
Expect(0, 13311, '\p{^Is_Blk=cjkcompat}', "");
Expect(0, 13311, '\P{Is_Blk=cjkcompat}', "");
Expect(1, 13311, '\P{^Is_Blk=cjkcompat}', "");
Expect(0, 13312, '\p{Is_Blk=cjkcompat}', "");
Expect(1, 13312, '\p{^Is_Blk=cjkcompat}', "");
Expect(1, 13312, '\P{Is_Blk=cjkcompat}', "");
Expect(0, 13312, '\P{^Is_Blk=cjkcompat}', "");
Expect(1, 13311, '\p{Is_Blk= cjk_COMPAT}', "");
Expect(0, 13311, '\p{^Is_Blk= cjk_COMPAT}', "");
Expect(0, 13311, '\P{Is_Blk= cjk_COMPAT}', "");
Expect(1, 13311, '\P{^Is_Blk= cjk_COMPAT}', "");
Expect(0, 13312, '\p{Is_Blk= cjk_COMPAT}', "");
Expect(1, 13312, '\p{^Is_Blk= cjk_COMPAT}', "");
Expect(1, 13312, '\P{Is_Blk= cjk_COMPAT}', "");
Expect(0, 13312, '\P{^Is_Blk= cjk_COMPAT}', "");
Error('\p{Block=	/a/Ugaritic}');
Error('\P{Block=	/a/Ugaritic}');
Expect(1, 66463, '\p{Block:ugaritic}', "");
Expect(0, 66463, '\p{^Block:ugaritic}', "");
Expect(0, 66463, '\P{Block:ugaritic}', "");
Expect(1, 66463, '\P{^Block:ugaritic}', "");
Expect(0, 66464, '\p{Block:ugaritic}', "");
Expect(1, 66464, '\p{^Block:ugaritic}', "");
Expect(1, 66464, '\P{Block:ugaritic}', "");
Expect(0, 66464, '\P{^Block:ugaritic}', "");
Expect(1, 66463, '\p{Block:   -ugaritic}', "");
Expect(0, 66463, '\p{^Block:   -ugaritic}', "");
Expect(0, 66463, '\P{Block:   -ugaritic}', "");
Expect(1, 66463, '\P{^Block:   -ugaritic}', "");
Expect(0, 66464, '\p{Block:   -ugaritic}', "");
Expect(1, 66464, '\p{^Block:   -ugaritic}', "");
Expect(1, 66464, '\P{Block:   -ugaritic}', "");
Expect(0, 66464, '\P{^Block:   -ugaritic}', "");
Error('\p{Blk=:= UGARITIC}');
Error('\P{Blk=:= UGARITIC}');
Expect(1, 66463, '\p{Blk=ugaritic}', "");
Expect(0, 66463, '\p{^Blk=ugaritic}', "");
Expect(0, 66463, '\P{Blk=ugaritic}', "");
Expect(1, 66463, '\P{^Blk=ugaritic}', "");
Expect(0, 66464, '\p{Blk=ugaritic}', "");
Expect(1, 66464, '\p{^Blk=ugaritic}', "");
Expect(1, 66464, '\P{Blk=ugaritic}', "");
Expect(0, 66464, '\P{^Blk=ugaritic}', "");
Expect(1, 66463, '\p{Blk=_Ugaritic}', "");
Expect(0, 66463, '\p{^Blk=_Ugaritic}', "");
Expect(0, 66463, '\P{Blk=_Ugaritic}', "");
Expect(1, 66463, '\P{^Blk=_Ugaritic}', "");
Expect(0, 66464, '\p{Blk=_Ugaritic}', "");
Expect(1, 66464, '\p{^Blk=_Ugaritic}', "");
Expect(1, 66464, '\P{Blk=_Ugaritic}', "");
Expect(0, 66464, '\P{^Blk=_Ugaritic}', "");
Error('\p{Is_Block=/a/--Ugaritic}');
Error('\P{Is_Block=/a/--Ugaritic}');
Expect(1, 66463, '\p{Is_Block=ugaritic}', "");
Expect(0, 66463, '\p{^Is_Block=ugaritic}', "");
Expect(0, 66463, '\P{Is_Block=ugaritic}', "");
Expect(1, 66463, '\P{^Is_Block=ugaritic}', "");
Expect(0, 66464, '\p{Is_Block=ugaritic}', "");
Expect(1, 66464, '\p{^Is_Block=ugaritic}', "");
Expect(1, 66464, '\P{Is_Block=ugaritic}', "");
Expect(0, 66464, '\P{^Is_Block=ugaritic}', "");
Expect(1, 66463, '\p{Is_Block=	Ugaritic}', "");
Expect(0, 66463, '\p{^Is_Block=	Ugaritic}', "");
Expect(0, 66463, '\P{Is_Block=	Ugaritic}', "");
Expect(1, 66463, '\P{^Is_Block=	Ugaritic}', "");
Expect(0, 66464, '\p{Is_Block=	Ugaritic}', "");
Expect(1, 66464, '\p{^Is_Block=	Ugaritic}', "");
Expect(1, 66464, '\P{Is_Block=	Ugaritic}', "");
Expect(0, 66464, '\P{^Is_Block=	Ugaritic}', "");
Error('\p{Is_Blk=:=Ugaritic}');
Error('\P{Is_Blk=:=Ugaritic}');
Expect(1, 66463, '\p{Is_Blk=ugaritic}', "");
Expect(0, 66463, '\p{^Is_Blk=ugaritic}', "");
Expect(0, 66463, '\P{Is_Blk=ugaritic}', "");
Expect(1, 66463, '\P{^Is_Blk=ugaritic}', "");
Expect(0, 66464, '\p{Is_Blk=ugaritic}', "");
Expect(1, 66464, '\p{^Is_Blk=ugaritic}', "");
Expect(1, 66464, '\P{Is_Blk=ugaritic}', "");
Expect(0, 66464, '\P{^Is_Blk=ugaritic}', "");
Expect(1, 66463, '\p{Is_Blk= UGARITIC}', "");
Expect(0, 66463, '\p{^Is_Blk= UGARITIC}', "");
Expect(0, 66463, '\P{Is_Blk= UGARITIC}', "");
Expect(1, 66463, '\P{^Is_Blk= UGARITIC}', "");
Expect(0, 66464, '\p{Is_Blk= UGARITIC}', "");
Expect(1, 66464, '\p{^Is_Blk= UGARITIC}', "");
Expect(1, 66464, '\P{Is_Blk= UGARITIC}', "");
Expect(0, 66464, '\P{^Is_Blk= UGARITIC}', "");
Error('\p{Block=-kayah_Li:=}');
Error('\P{Block=-kayah_Li:=}');
Expect(1, 43311, '\p{Block=kayahli}', "");
Expect(0, 43311, '\p{^Block=kayahli}', "");
Expect(0, 43311, '\P{Block=kayahli}', "");
Expect(1, 43311, '\P{^Block=kayahli}', "");
Expect(0, 43312, '\p{Block=kayahli}', "");
Expect(1, 43312, '\p{^Block=kayahli}', "");
Expect(1, 43312, '\P{Block=kayahli}', "");
Expect(0, 43312, '\P{^Block=kayahli}', "");
Expect(1, 43311, '\p{Block=_-Kayah_LI}', "");
Expect(0, 43311, '\p{^Block=_-Kayah_LI}', "");
Expect(0, 43311, '\P{Block=_-Kayah_LI}', "");
Expect(1, 43311, '\P{^Block=_-Kayah_LI}', "");
Expect(0, 43312, '\p{Block=_-Kayah_LI}', "");
Expect(1, 43312, '\p{^Block=_-Kayah_LI}', "");
Expect(1, 43312, '\P{Block=_-Kayah_LI}', "");
Expect(0, 43312, '\P{^Block=_-Kayah_LI}', "");
Error('\p{Blk=	kayah_li/a/}');
Error('\P{Blk=	kayah_li/a/}');
Expect(1, 43311, '\p{Blk=kayahli}', "");
Expect(0, 43311, '\p{^Blk=kayahli}', "");
Expect(0, 43311, '\P{Blk=kayahli}', "");
Expect(1, 43311, '\P{^Blk=kayahli}', "");
Expect(0, 43312, '\p{Blk=kayahli}', "");
Expect(1, 43312, '\p{^Blk=kayahli}', "");
Expect(1, 43312, '\P{Blk=kayahli}', "");
Expect(0, 43312, '\P{^Blk=kayahli}', "");
Expect(1, 43311, '\p{Blk=-KAYAH_LI}', "");
Expect(0, 43311, '\p{^Blk=-KAYAH_LI}', "");
Expect(0, 43311, '\P{Blk=-KAYAH_LI}', "");
Expect(1, 43311, '\P{^Blk=-KAYAH_LI}', "");
Expect(0, 43312, '\p{Blk=-KAYAH_LI}', "");
Expect(1, 43312, '\p{^Blk=-KAYAH_LI}', "");
Expect(1, 43312, '\P{Blk=-KAYAH_LI}', "");
Expect(0, 43312, '\P{^Blk=-KAYAH_LI}', "");
Error('\p{Is_Block=:=	Kayah_li}');
Error('\P{Is_Block=:=	Kayah_li}');
Expect(1, 43311, '\p{Is_Block=kayahli}', "");
Expect(0, 43311, '\p{^Is_Block=kayahli}', "");
Expect(0, 43311, '\P{Is_Block=kayahli}', "");
Expect(1, 43311, '\P{^Is_Block=kayahli}', "");
Expect(0, 43312, '\p{Is_Block=kayahli}', "");
Expect(1, 43312, '\p{^Is_Block=kayahli}', "");
Expect(1, 43312, '\P{Is_Block=kayahli}', "");
Expect(0, 43312, '\P{^Is_Block=kayahli}', "");
Expect(1, 43311, '\p{Is_Block:--Kayah_Li}', "");
Expect(0, 43311, '\p{^Is_Block:--Kayah_Li}', "");
Expect(0, 43311, '\P{Is_Block:--Kayah_Li}', "");
Expect(1, 43311, '\P{^Is_Block:--Kayah_Li}', "");
Expect(0, 43312, '\p{Is_Block:--Kayah_Li}', "");
Expect(1, 43312, '\p{^Is_Block:--Kayah_Li}', "");
Expect(1, 43312, '\P{Is_Block:--Kayah_Li}', "");
Expect(0, 43312, '\P{^Is_Block:--Kayah_Li}', "");
Error('\p{Is_Blk=:= Kayah_li}');
Error('\P{Is_Blk=:= Kayah_li}');
Expect(1, 43311, '\p{Is_Blk=kayahli}', "");
Expect(0, 43311, '\p{^Is_Blk=kayahli}', "");
Expect(0, 43311, '\P{Is_Blk=kayahli}', "");
Expect(1, 43311, '\P{^Is_Blk=kayahli}', "");
Expect(0, 43312, '\p{Is_Blk=kayahli}', "");
Expect(1, 43312, '\p{^Is_Blk=kayahli}', "");
Expect(1, 43312, '\P{Is_Blk=kayahli}', "");
Expect(0, 43312, '\P{^Is_Blk=kayahli}', "");
Expect(1, 43311, '\p{Is_Blk=- KAYAH_li}', "");
Expect(0, 43311, '\p{^Is_Blk=- KAYAH_li}', "");
Expect(0, 43311, '\P{Is_Blk=- KAYAH_li}', "");
Expect(1, 43311, '\P{^Is_Blk=- KAYAH_li}', "");
Expect(0, 43312, '\p{Is_Blk=- KAYAH_li}', "");
Expect(1, 43312, '\p{^Is_Blk=- KAYAH_li}', "");
Expect(1, 43312, '\P{Is_Blk=- KAYAH_li}', "");
Expect(0, 43312, '\P{^Is_Blk=- KAYAH_li}', "");
Error('\p{Block=/a/cjk_Compatibility_Ideographs}');
Error('\P{Block=/a/cjk_Compatibility_Ideographs}');
Expect(1, 64255, '\p{Block=cjkcompatibilityideographs}', "");
Expect(0, 64255, '\p{^Block=cjkcompatibilityideographs}', "");
Expect(0, 64255, '\P{Block=cjkcompatibilityideographs}', "");
Expect(1, 64255, '\P{^Block=cjkcompatibilityideographs}', "");
Expect(0, 64256, '\p{Block=cjkcompatibilityideographs}', "");
Expect(1, 64256, '\p{^Block=cjkcompatibilityideographs}', "");
Expect(1, 64256, '\P{Block=cjkcompatibilityideographs}', "");
Expect(0, 64256, '\P{^Block=cjkcompatibilityideographs}', "");
Expect(1, 64255, '\p{Block= CJK_COMPATIBILITY_IDEOGRAPHS}', "");
Expect(0, 64255, '\p{^Block= CJK_COMPATIBILITY_IDEOGRAPHS}', "");
Expect(0, 64255, '\P{Block= CJK_COMPATIBILITY_IDEOGRAPHS}', "");
Expect(1, 64255, '\P{^Block= CJK_COMPATIBILITY_IDEOGRAPHS}', "");
Expect(0, 64256, '\p{Block= CJK_COMPATIBILITY_IDEOGRAPHS}', "");
Expect(1, 64256, '\p{^Block= CJK_COMPATIBILITY_IDEOGRAPHS}', "");
Expect(1, 64256, '\P{Block= CJK_COMPATIBILITY_IDEOGRAPHS}', "");
Expect(0, 64256, '\P{^Block= CJK_COMPATIBILITY_IDEOGRAPHS}', "");
Error('\p{Blk=	:=CJK_Compat_ideographs}');
Error('\P{Blk=	:=CJK_Compat_ideographs}');
Expect(1, 64255, '\p{Blk=cjkcompatideographs}', "");
Expect(0, 64255, '\p{^Blk=cjkcompatideographs}', "");
Expect(0, 64255, '\P{Blk=cjkcompatideographs}', "");
Expect(1, 64255, '\P{^Blk=cjkcompatideographs}', "");
Expect(0, 64256, '\p{Blk=cjkcompatideographs}', "");
Expect(1, 64256, '\p{^Blk=cjkcompatideographs}', "");
Expect(1, 64256, '\P{Blk=cjkcompatideographs}', "");
Expect(0, 64256, '\P{^Blk=cjkcompatideographs}', "");
Expect(1, 64255, '\p{Blk= -cjk_COMPAT_ideographs}', "");
Expect(0, 64255, '\p{^Blk= -cjk_COMPAT_ideographs}', "");
Expect(0, 64255, '\P{Blk= -cjk_COMPAT_ideographs}', "");
Expect(1, 64255, '\P{^Blk= -cjk_COMPAT_ideographs}', "");
Expect(0, 64256, '\p{Blk= -cjk_COMPAT_ideographs}', "");
Expect(1, 64256, '\p{^Blk= -cjk_COMPAT_ideographs}', "");
Expect(1, 64256, '\P{Blk= -cjk_COMPAT_ideographs}', "");
Expect(0, 64256, '\P{^Blk= -cjk_COMPAT_ideographs}', "");
Error('\p{Is_Block=:=--CJK_Compatibility_ideographs}');
Error('\P{Is_Block=:=--CJK_Compatibility_ideographs}');
Expect(1, 64255, '\p{Is_Block=cjkcompatibilityideographs}', "");
Expect(0, 64255, '\p{^Is_Block=cjkcompatibilityideographs}', "");
Expect(0, 64255, '\P{Is_Block=cjkcompatibilityideographs}', "");
Expect(1, 64255, '\P{^Is_Block=cjkcompatibilityideographs}', "");
Expect(0, 64256, '\p{Is_Block=cjkcompatibilityideographs}', "");
Expect(1, 64256, '\p{^Is_Block=cjkcompatibilityideographs}', "");
Expect(1, 64256, '\P{Is_Block=cjkcompatibilityideographs}', "");
Expect(0, 64256, '\P{^Is_Block=cjkcompatibilityideographs}', "");
Expect(1, 64255, '\p{Is_Block=_-CJK_Compatibility_Ideographs}', "");
Expect(0, 64255, '\p{^Is_Block=_-CJK_Compatibility_Ideographs}', "");
Expect(0, 64255, '\P{Is_Block=_-CJK_Compatibility_Ideographs}', "");
Expect(1, 64255, '\P{^Is_Block=_-CJK_Compatibility_Ideographs}', "");
Expect(0, 64256, '\p{Is_Block=_-CJK_Compatibility_Ideographs}', "");
Expect(1, 64256, '\p{^Is_Block=_-CJK_Compatibility_Ideographs}', "");
Expect(1, 64256, '\P{Is_Block=_-CJK_Compatibility_Ideographs}', "");
Expect(0, 64256, '\P{^Is_Block=_-CJK_Compatibility_Ideographs}', "");
Error('\p{Is_Blk=_:=CJK_Compat_IDEOGRAPHS}');
Error('\P{Is_Blk=_:=CJK_Compat_IDEOGRAPHS}');
Expect(1, 64255, '\p{Is_Blk: cjkcompatideographs}', "");
Expect(0, 64255, '\p{^Is_Blk: cjkcompatideographs}', "");
Expect(0, 64255, '\P{Is_Blk: cjkcompatideographs}', "");
Expect(1, 64255, '\P{^Is_Blk: cjkcompatideographs}', "");
Expect(0, 64256, '\p{Is_Blk: cjkcompatideographs}', "");
Expect(1, 64256, '\p{^Is_Blk: cjkcompatideographs}', "");
Expect(1, 64256, '\P{Is_Blk: cjkcompatideographs}', "");
Expect(0, 64256, '\P{^Is_Blk: cjkcompatideographs}', "");
Expect(1, 64255, '\p{Is_Blk= _CJK_compat_IDEOGRAPHS}', "");
Expect(0, 64255, '\p{^Is_Blk= _CJK_compat_IDEOGRAPHS}', "");
Expect(0, 64255, '\P{Is_Blk= _CJK_compat_IDEOGRAPHS}', "");
Expect(1, 64255, '\P{^Is_Blk= _CJK_compat_IDEOGRAPHS}', "");
Expect(0, 64256, '\p{Is_Blk= _CJK_compat_IDEOGRAPHS}', "");
Expect(1, 64256, '\p{^Is_Blk= _CJK_compat_IDEOGRAPHS}', "");
Expect(1, 64256, '\P{Is_Blk= _CJK_compat_IDEOGRAPHS}', "");
Expect(0, 64256, '\P{^Is_Blk= _CJK_compat_IDEOGRAPHS}', "");
Error('\p{Block=:=THAI}');
Error('\P{Block=:=THAI}');
Expect(1, 3711, '\p{Block=thai}', "");
Expect(0, 3711, '\p{^Block=thai}', "");
Expect(0, 3711, '\P{Block=thai}', "");
Expect(1, 3711, '\P{^Block=thai}', "");
Expect(0, 3712, '\p{Block=thai}', "");
Expect(1, 3712, '\p{^Block=thai}', "");
Expect(1, 3712, '\P{Block=thai}', "");
Expect(0, 3712, '\P{^Block=thai}', "");
Expect(1, 3711, '\p{Block=	Thai}', "");
Expect(0, 3711, '\p{^Block=	Thai}', "");
Expect(0, 3711, '\P{Block=	Thai}', "");
Expect(1, 3711, '\P{^Block=	Thai}', "");
Expect(0, 3712, '\p{Block=	Thai}', "");
Expect(1, 3712, '\p{^Block=	Thai}', "");
Expect(1, 3712, '\P{Block=	Thai}', "");
Expect(0, 3712, '\P{^Block=	Thai}', "");
Error('\p{Blk=/a/_ THAI}');
Error('\P{Blk=/a/_ THAI}');
Expect(1, 3711, '\p{Blk=thai}', "");
Expect(0, 3711, '\p{^Blk=thai}', "");
Expect(0, 3711, '\P{Blk=thai}', "");
Expect(1, 3711, '\P{^Blk=thai}', "");
Expect(0, 3712, '\p{Blk=thai}', "");
Expect(1, 3712, '\p{^Blk=thai}', "");
Expect(1, 3712, '\P{Blk=thai}', "");
Expect(0, 3712, '\P{^Blk=thai}', "");
Expect(1, 3711, '\p{Blk=_Thai}', "");
Expect(0, 3711, '\p{^Blk=_Thai}', "");
Expect(0, 3711, '\P{Blk=_Thai}', "");
Expect(1, 3711, '\P{^Blk=_Thai}', "");
Expect(0, 3712, '\p{Blk=_Thai}', "");
Expect(1, 3712, '\p{^Blk=_Thai}', "");
Expect(1, 3712, '\P{Blk=_Thai}', "");
Expect(0, 3712, '\P{^Blk=_Thai}', "");
Error('\p{Is_Block=	/a/Thai}');
Error('\P{Is_Block=	/a/Thai}');
Expect(1, 3711, '\p{Is_Block=thai}', "");
Expect(0, 3711, '\p{^Is_Block=thai}', "");
Expect(0, 3711, '\P{Is_Block=thai}', "");
Expect(1, 3711, '\P{^Is_Block=thai}', "");
Expect(0, 3712, '\p{Is_Block=thai}', "");
Expect(1, 3712, '\p{^Is_Block=thai}', "");
Expect(1, 3712, '\P{Is_Block=thai}', "");
Expect(0, 3712, '\P{^Is_Block=thai}', "");
Expect(1, 3711, '\p{Is_Block=_	Thai}', "");
Expect(0, 3711, '\p{^Is_Block=_	Thai}', "");
Expect(0, 3711, '\P{Is_Block=_	Thai}', "");
Expect(1, 3711, '\P{^Is_Block=_	Thai}', "");
Expect(0, 3712, '\p{Is_Block=_	Thai}', "");
Expect(1, 3712, '\p{^Is_Block=_	Thai}', "");
Expect(1, 3712, '\P{Is_Block=_	Thai}', "");
Expect(0, 3712, '\P{^Is_Block=_	Thai}', "");
Error('\p{Is_Blk=:=- Thai}');
Error('\P{Is_Blk=:=- Thai}');
Expect(1, 3711, '\p{Is_Blk:	thai}', "");
Expect(0, 3711, '\p{^Is_Blk:	thai}', "");
Expect(0, 3711, '\P{Is_Blk:	thai}', "");
Expect(1, 3711, '\P{^Is_Blk:	thai}', "");
Expect(0, 3712, '\p{Is_Blk:	thai}', "");
Expect(1, 3712, '\p{^Is_Blk:	thai}', "");
Expect(1, 3712, '\P{Is_Blk:	thai}', "");
Expect(0, 3712, '\P{^Is_Blk:	thai}', "");
Expect(1, 3711, '\p{Is_Blk=__Thai}', "");
Expect(0, 3711, '\p{^Is_Blk=__Thai}', "");
Expect(0, 3711, '\P{Is_Blk=__Thai}', "");
Expect(1, 3711, '\P{^Is_Blk=__Thai}', "");
Expect(0, 3712, '\p{Is_Blk=__Thai}', "");
Expect(1, 3712, '\p{^Is_Blk=__Thai}', "");
Expect(1, 3712, '\P{Is_Blk=__Thai}', "");
Expect(0, 3712, '\P{^Is_Blk=__Thai}', "");
Error('\p{Block=/a/	vai}');
Error('\P{Block=/a/	vai}');
Expect(1, 42559, '\p{Block=vai}', "");
Expect(0, 42559, '\p{^Block=vai}', "");
Expect(0, 42559, '\P{Block=vai}', "");
Expect(1, 42559, '\P{^Block=vai}', "");
Expect(0, 42560, '\p{Block=vai}', "");
Expect(1, 42560, '\p{^Block=vai}', "");
Expect(1, 42560, '\P{Block=vai}', "");
Expect(0, 42560, '\P{^Block=vai}', "");
Expect(1, 42559, '\p{Block=_Vai}', "");
Expect(0, 42559, '\p{^Block=_Vai}', "");
Expect(0, 42559, '\P{Block=_Vai}', "");
Expect(1, 42559, '\P{^Block=_Vai}', "");
Expect(0, 42560, '\p{Block=_Vai}', "");
Expect(1, 42560, '\p{^Block=_Vai}', "");
Expect(1, 42560, '\P{Block=_Vai}', "");
Expect(0, 42560, '\P{^Block=_Vai}', "");
Error('\p{Blk=	/a/VAI}');
Error('\P{Blk=	/a/VAI}');
Expect(1, 42559, '\p{Blk=vai}', "");
Expect(0, 42559, '\p{^Blk=vai}', "");
Expect(0, 42559, '\P{Blk=vai}', "");
Expect(1, 42559, '\P{^Blk=vai}', "");
Expect(0, 42560, '\p{Blk=vai}', "");
Expect(1, 42560, '\p{^Blk=vai}', "");
Expect(1, 42560, '\P{Blk=vai}', "");
Expect(0, 42560, '\P{^Blk=vai}', "");
Expect(1, 42559, '\p{Blk:	_Vai}', "");
Expect(0, 42559, '\p{^Blk:	_Vai}', "");
Expect(0, 42559, '\P{Blk:	_Vai}', "");
Expect(1, 42559, '\P{^Blk:	_Vai}', "");
Expect(0, 42560, '\p{Blk:	_Vai}', "");
Expect(1, 42560, '\p{^Blk:	_Vai}', "");
Expect(1, 42560, '\P{Blk:	_Vai}', "");
Expect(0, 42560, '\P{^Blk:	_Vai}', "");
Error('\p{Is_Block: 		vai:=}');
Error('\P{Is_Block: 		vai:=}');
Expect(1, 42559, '\p{Is_Block=vai}', "");
Expect(0, 42559, '\p{^Is_Block=vai}', "");
Expect(0, 42559, '\P{Is_Block=vai}', "");
Expect(1, 42559, '\P{^Is_Block=vai}', "");
Expect(0, 42560, '\p{Is_Block=vai}', "");
Expect(1, 42560, '\p{^Is_Block=vai}', "");
Expect(1, 42560, '\P{Is_Block=vai}', "");
Expect(0, 42560, '\P{^Is_Block=vai}', "");
Expect(1, 42559, '\p{Is_Block=-vai}', "");
Expect(0, 42559, '\p{^Is_Block=-vai}', "");
Expect(0, 42559, '\P{Is_Block=-vai}', "");
Expect(1, 42559, '\P{^Is_Block=-vai}', "");
Expect(0, 42560, '\p{Is_Block=-vai}', "");
Expect(1, 42560, '\p{^Is_Block=-vai}', "");
Expect(1, 42560, '\P{Is_Block=-vai}', "");
Expect(0, 42560, '\P{^Is_Block=-vai}', "");
Error('\p{Is_Blk=-:=Vai}');
Error('\P{Is_Blk=-:=Vai}');
Expect(1, 42559, '\p{Is_Blk=vai}', "");
Expect(0, 42559, '\p{^Is_Blk=vai}', "");
Expect(0, 42559, '\P{Is_Blk=vai}', "");
Expect(1, 42559, '\P{^Is_Blk=vai}', "");
Expect(0, 42560, '\p{Is_Blk=vai}', "");
Expect(1, 42560, '\p{^Is_Blk=vai}', "");
Expect(1, 42560, '\P{Is_Blk=vai}', "");
Expect(0, 42560, '\P{^Is_Blk=vai}', "");
Expect(1, 42559, '\p{Is_Blk=	Vai}', "");
Expect(0, 42559, '\p{^Is_Blk=	Vai}', "");
Expect(0, 42559, '\P{Is_Blk=	Vai}', "");
Expect(1, 42559, '\P{^Is_Blk=	Vai}', "");
Expect(0, 42560, '\p{Is_Blk=	Vai}', "");
Expect(1, 42560, '\p{^Is_Blk=	Vai}', "");
Expect(1, 42560, '\P{Is_Blk=	Vai}', "");
Expect(0, 42560, '\P{^Is_Blk=	Vai}', "");
Error('\p{Block= /a/Rejang}');
Error('\P{Block= /a/Rejang}');
Expect(1, 43359, '\p{Block=rejang}', "");
Expect(0, 43359, '\p{^Block=rejang}', "");
Expect(0, 43359, '\P{Block=rejang}', "");
Expect(1, 43359, '\P{^Block=rejang}', "");
Expect(0, 43360, '\p{Block=rejang}', "");
Expect(1, 43360, '\p{^Block=rejang}', "");
Expect(1, 43360, '\P{Block=rejang}', "");
Expect(0, 43360, '\P{^Block=rejang}', "");
Expect(1, 43359, '\p{Block= _REJANG}', "");
Expect(0, 43359, '\p{^Block= _REJANG}', "");
Expect(0, 43359, '\P{Block= _REJANG}', "");
Expect(1, 43359, '\P{^Block= _REJANG}', "");
Expect(0, 43360, '\p{Block= _REJANG}', "");
Expect(1, 43360, '\p{^Block= _REJANG}', "");
Expect(1, 43360, '\P{Block= _REJANG}', "");
Expect(0, 43360, '\P{^Block= _REJANG}', "");
Error('\p{Blk::= Rejang}');
Error('\P{Blk::= Rejang}');
Expect(1, 43359, '\p{Blk=rejang}', "");
Expect(0, 43359, '\p{^Blk=rejang}', "");
Expect(0, 43359, '\P{Blk=rejang}', "");
Expect(1, 43359, '\P{^Blk=rejang}', "");
Expect(0, 43360, '\p{Blk=rejang}', "");
Expect(1, 43360, '\p{^Blk=rejang}', "");
Expect(1, 43360, '\P{Blk=rejang}', "");
Expect(0, 43360, '\P{^Blk=rejang}', "");
Expect(1, 43359, '\p{Blk=-Rejang}', "");
Expect(0, 43359, '\p{^Blk=-Rejang}', "");
Expect(0, 43359, '\P{Blk=-Rejang}', "");
Expect(1, 43359, '\P{^Blk=-Rejang}', "");
Expect(0, 43360, '\p{Blk=-Rejang}', "");
Expect(1, 43360, '\p{^Blk=-Rejang}', "");
Expect(1, 43360, '\P{Blk=-Rejang}', "");
Expect(0, 43360, '\P{^Blk=-Rejang}', "");
Error('\p{Is_Block=:=_-REJANG}');
Error('\P{Is_Block=:=_-REJANG}');
Expect(1, 43359, '\p{Is_Block=rejang}', "");
Expect(0, 43359, '\p{^Is_Block=rejang}', "");
Expect(0, 43359, '\P{Is_Block=rejang}', "");
Expect(1, 43359, '\P{^Is_Block=rejang}', "");
Expect(0, 43360, '\p{Is_Block=rejang}', "");
Expect(1, 43360, '\p{^Is_Block=rejang}', "");
Expect(1, 43360, '\P{Is_Block=rejang}', "");
Expect(0, 43360, '\P{^Is_Block=rejang}', "");
Expect(1, 43359, '\p{Is_Block=_	rejang}', "");
Expect(0, 43359, '\p{^Is_Block=_	rejang}', "");
Expect(0, 43359, '\P{Is_Block=_	rejang}', "");
Expect(1, 43359, '\P{^Is_Block=_	rejang}', "");
Expect(0, 43360, '\p{Is_Block=_	rejang}', "");
Expect(1, 43360, '\p{^Is_Block=_	rejang}', "");
Expect(1, 43360, '\P{Is_Block=_	rejang}', "");
Expect(0, 43360, '\P{^Is_Block=_	rejang}', "");
Error('\p{Is_Blk=/a/	 REJANG}');
Error('\P{Is_Blk=/a/	 REJANG}');
Expect(1, 43359, '\p{Is_Blk=rejang}', "");
Expect(0, 43359, '\p{^Is_Blk=rejang}', "");
Expect(0, 43359, '\P{Is_Blk=rejang}', "");
Expect(1, 43359, '\P{^Is_Blk=rejang}', "");
Expect(0, 43360, '\p{Is_Blk=rejang}', "");
Expect(1, 43360, '\p{^Is_Blk=rejang}', "");
Expect(1, 43360, '\P{Is_Blk=rejang}', "");
Expect(0, 43360, '\P{^Is_Blk=rejang}', "");
Expect(1, 43359, '\p{Is_Blk=-Rejang}', "");
Expect(0, 43359, '\p{^Is_Blk=-Rejang}', "");
Expect(0, 43359, '\P{Is_Blk=-Rejang}', "");
Expect(1, 43359, '\P{^Is_Blk=-Rejang}', "");
Expect(0, 43360, '\p{Is_Blk=-Rejang}', "");
Expect(1, 43360, '\p{^Is_Blk=-Rejang}', "");
Expect(1, 43360, '\P{Is_Blk=-Rejang}', "");
Expect(0, 43360, '\P{^Is_Blk=-Rejang}', "");
Error('\p{Block=/a/-bengali}');
Error('\P{Block=/a/-bengali}');
Expect(1, 2559, '\p{Block=bengali}', "");
Expect(0, 2559, '\p{^Block=bengali}', "");
Expect(0, 2559, '\P{Block=bengali}', "");
Expect(1, 2559, '\P{^Block=bengali}', "");
Expect(0, 2560, '\p{Block=bengali}', "");
Expect(1, 2560, '\p{^Block=bengali}', "");
Expect(1, 2560, '\P{Block=bengali}', "");
Expect(0, 2560, '\P{^Block=bengali}', "");
Expect(1, 2559, '\p{Block=	_BENGALI}', "");
Expect(0, 2559, '\p{^Block=	_BENGALI}', "");
Expect(0, 2559, '\P{Block=	_BENGALI}', "");
Expect(1, 2559, '\P{^Block=	_BENGALI}', "");
Expect(0, 2560, '\p{Block=	_BENGALI}', "");
Expect(1, 2560, '\p{^Block=	_BENGALI}', "");
Expect(1, 2560, '\P{Block=	_BENGALI}', "");
Expect(0, 2560, '\P{^Block=	_BENGALI}', "");
Error('\p{Blk=	_BENGALI/a/}');
Error('\P{Blk=	_BENGALI/a/}');
Expect(1, 2559, '\p{Blk=bengali}', "");
Expect(0, 2559, '\p{^Blk=bengali}', "");
Expect(0, 2559, '\P{Blk=bengali}', "");
Expect(1, 2559, '\P{^Blk=bengali}', "");
Expect(0, 2560, '\p{Blk=bengali}', "");
Expect(1, 2560, '\p{^Blk=bengali}', "");
Expect(1, 2560, '\P{Blk=bengali}', "");
Expect(0, 2560, '\P{^Blk=bengali}', "");
Expect(1, 2559, '\p{Blk: _ BENGALI}', "");
Expect(0, 2559, '\p{^Blk: _ BENGALI}', "");
Expect(0, 2559, '\P{Blk: _ BENGALI}', "");
Expect(1, 2559, '\P{^Blk: _ BENGALI}', "");
Expect(0, 2560, '\p{Blk: _ BENGALI}', "");
Expect(1, 2560, '\p{^Blk: _ BENGALI}', "");
Expect(1, 2560, '\P{Blk: _ BENGALI}', "");
Expect(0, 2560, '\P{^Blk: _ BENGALI}', "");
Error('\p{Is_Block=-:=Bengali}');
Error('\P{Is_Block=-:=Bengali}');
Expect(1, 2559, '\p{Is_Block=bengali}', "");
Expect(0, 2559, '\p{^Is_Block=bengali}', "");
Expect(0, 2559, '\P{Is_Block=bengali}', "");
Expect(1, 2559, '\P{^Is_Block=bengali}', "");
Expect(0, 2560, '\p{Is_Block=bengali}', "");
Expect(1, 2560, '\p{^Is_Block=bengali}', "");
Expect(1, 2560, '\P{Is_Block=bengali}', "");
Expect(0, 2560, '\P{^Is_Block=bengali}', "");
Expect(1, 2559, '\p{Is_Block=__BENGALI}', "");
Expect(0, 2559, '\p{^Is_Block=__BENGALI}', "");
Expect(0, 2559, '\P{Is_Block=__BENGALI}', "");
Expect(1, 2559, '\P{^Is_Block=__BENGALI}', "");
Expect(0, 2560, '\p{Is_Block=__BENGALI}', "");
Expect(1, 2560, '\p{^Is_Block=__BENGALI}', "");
Expect(1, 2560, '\P{Is_Block=__BENGALI}', "");
Expect(0, 2560, '\P{^Is_Block=__BENGALI}', "");
Error('\p{Is_Blk= BENGALI:=}');
Error('\P{Is_Blk= BENGALI:=}');
Expect(1, 2559, '\p{Is_Blk=bengali}', "");
Expect(0, 2559, '\p{^Is_Blk=bengali}', "");
Expect(0, 2559, '\P{Is_Blk=bengali}', "");
Expect(1, 2559, '\P{^Is_Blk=bengali}', "");
Expect(0, 2560, '\p{Is_Blk=bengali}', "");
Expect(1, 2560, '\p{^Is_Blk=bengali}', "");
Expect(1, 2560, '\P{Is_Blk=bengali}', "");
Expect(0, 2560, '\P{^Is_Blk=bengali}', "");
Expect(1, 2559, '\p{Is_Blk=	 BENGALI}', "");
Expect(0, 2559, '\p{^Is_Blk=	 BENGALI}', "");
Expect(0, 2559, '\P{Is_Blk=	 BENGALI}', "");
Expect(1, 2559, '\P{^Is_Blk=	 BENGALI}', "");
Expect(0, 2560, '\p{Is_Blk=	 BENGALI}', "");
Expect(1, 2560, '\p{^Is_Blk=	 BENGALI}', "");
Expect(1, 2560, '\P{Is_Blk=	 BENGALI}', "");
Expect(0, 2560, '\P{^Is_Blk=	 BENGALI}', "");
Error('\p{Block=:=_ Saurashtra}');
Error('\P{Block=:=_ Saurashtra}');
Expect(1, 43231, '\p{Block=saurashtra}', "");
Expect(0, 43231, '\p{^Block=saurashtra}', "");
Expect(0, 43231, '\P{Block=saurashtra}', "");
Expect(1, 43231, '\P{^Block=saurashtra}', "");
Expect(0, 43232, '\p{Block=saurashtra}', "");
Expect(1, 43232, '\p{^Block=saurashtra}', "");
Expect(1, 43232, '\P{Block=saurashtra}', "");
Expect(0, 43232, '\P{^Block=saurashtra}', "");
Expect(1, 43231, '\p{Block=_Saurashtra}', "");
Expect(0, 43231, '\p{^Block=_Saurashtra}', "");
Expect(0, 43231, '\P{Block=_Saurashtra}', "");
Expect(1, 43231, '\P{^Block=_Saurashtra}', "");
Expect(0, 43232, '\p{Block=_Saurashtra}', "");
Expect(1, 43232, '\p{^Block=_Saurashtra}', "");
Expect(1, 43232, '\P{Block=_Saurashtra}', "");
Expect(0, 43232, '\P{^Block=_Saurashtra}', "");
Error('\p{Blk=:=Saurashtra}');
Error('\P{Blk=:=Saurashtra}');
Expect(1, 43231, '\p{Blk:   saurashtra}', "");
Expect(0, 43231, '\p{^Blk:   saurashtra}', "");
Expect(0, 43231, '\P{Blk:   saurashtra}', "");
Expect(1, 43231, '\P{^Blk:   saurashtra}', "");
Expect(0, 43232, '\p{Blk:   saurashtra}', "");
Expect(1, 43232, '\p{^Blk:   saurashtra}', "");
Expect(1, 43232, '\P{Blk:   saurashtra}', "");
Expect(0, 43232, '\P{^Blk:   saurashtra}', "");
Expect(1, 43231, '\p{Blk:  SAURASHTRA}', "");
Expect(0, 43231, '\p{^Blk:  SAURASHTRA}', "");
Expect(0, 43231, '\P{Blk:  SAURASHTRA}', "");
Expect(1, 43231, '\P{^Blk:  SAURASHTRA}', "");
Expect(0, 43232, '\p{Blk:  SAURASHTRA}', "");
Expect(1, 43232, '\p{^Blk:  SAURASHTRA}', "");
Expect(1, 43232, '\P{Blk:  SAURASHTRA}', "");
Expect(0, 43232, '\P{^Blk:  SAURASHTRA}', "");
Error('\p{Is_Block= Saurashtra:=}');
Error('\P{Is_Block= Saurashtra:=}');
Expect(1, 43231, '\p{Is_Block:   saurashtra}', "");
Expect(0, 43231, '\p{^Is_Block:   saurashtra}', "");
Expect(0, 43231, '\P{Is_Block:   saurashtra}', "");
Expect(1, 43231, '\P{^Is_Block:   saurashtra}', "");
Expect(0, 43232, '\p{Is_Block:   saurashtra}', "");
Expect(1, 43232, '\p{^Is_Block:   saurashtra}', "");
Expect(1, 43232, '\P{Is_Block:   saurashtra}', "");
Expect(0, 43232, '\P{^Is_Block:   saurashtra}', "");
Expect(1, 43231, '\p{Is_Block:  Saurashtra}', "");
Expect(0, 43231, '\p{^Is_Block:  Saurashtra}', "");
Expect(0, 43231, '\P{Is_Block:  Saurashtra}', "");
Expect(1, 43231, '\P{^Is_Block:  Saurashtra}', "");
Expect(0, 43232, '\p{Is_Block:  Saurashtra}', "");
Expect(1, 43232, '\p{^Is_Block:  Saurashtra}', "");
Expect(1, 43232, '\P{Is_Block:  Saurashtra}', "");
Expect(0, 43232, '\P{^Is_Block:  Saurashtra}', "");
Error('\p{Is_Blk=/a/-	Saurashtra}');
Error('\P{Is_Blk=/a/-	Saurashtra}');
Expect(1, 43231, '\p{Is_Blk=saurashtra}', "");
Expect(0, 43231, '\p{^Is_Blk=saurashtra}', "");
Expect(0, 43231, '\P{Is_Blk=saurashtra}', "");
Expect(1, 43231, '\P{^Is_Blk=saurashtra}', "");
Expect(0, 43232, '\p{Is_Blk=saurashtra}', "");
Expect(1, 43232, '\p{^Is_Blk=saurashtra}', "");
Expect(1, 43232, '\P{Is_Blk=saurashtra}', "");
Expect(0, 43232, '\P{^Is_Blk=saurashtra}', "");
Expect(1, 43231, '\p{Is_Blk:	 Saurashtra}', "");
Expect(0, 43231, '\p{^Is_Blk:	 Saurashtra}', "");
Expect(0, 43231, '\P{Is_Blk:	 Saurashtra}', "");
Expect(1, 43231, '\P{^Is_Blk:	 Saurashtra}', "");
Expect(0, 43232, '\p{Is_Blk:	 Saurashtra}', "");
Expect(1, 43232, '\p{^Is_Blk:	 Saurashtra}', "");
Expect(1, 43232, '\P{Is_Blk:	 Saurashtra}', "");
Expect(0, 43232, '\P{^Is_Blk:	 Saurashtra}', "");
Error('\p{Block=/a/_-osmanya}');
Error('\P{Block=/a/_-osmanya}');
Expect(1, 66735, '\p{Block:   osmanya}', "");
Expect(0, 66735, '\p{^Block:   osmanya}', "");
Expect(0, 66735, '\P{Block:   osmanya}', "");
Expect(1, 66735, '\P{^Block:   osmanya}', "");
Expect(0, 66736, '\p{Block:   osmanya}', "");
Expect(1, 66736, '\p{^Block:   osmanya}', "");
Expect(1, 66736, '\P{Block:   osmanya}', "");
Expect(0, 66736, '\P{^Block:   osmanya}', "");
Expect(1, 66735, '\p{Block:   _ Osmanya}', "");
Expect(0, 66735, '\p{^Block:   _ Osmanya}', "");
Expect(0, 66735, '\P{Block:   _ Osmanya}', "");
Expect(1, 66735, '\P{^Block:   _ Osmanya}', "");
Expect(0, 66736, '\p{Block:   _ Osmanya}', "");
Expect(1, 66736, '\p{^Block:   _ Osmanya}', "");
Expect(1, 66736, '\P{Block:   _ Osmanya}', "");
Expect(0, 66736, '\P{^Block:   _ Osmanya}', "");
Error('\p{Blk=:=-_osmanya}');
Error('\P{Blk=:=-_osmanya}');
Expect(1, 66735, '\p{Blk=osmanya}', "");
Expect(0, 66735, '\p{^Blk=osmanya}', "");
Expect(0, 66735, '\P{Blk=osmanya}', "");
Expect(1, 66735, '\P{^Blk=osmanya}', "");
Expect(0, 66736, '\p{Blk=osmanya}', "");
Expect(1, 66736, '\p{^Blk=osmanya}', "");
Expect(1, 66736, '\P{Blk=osmanya}', "");
Expect(0, 66736, '\P{^Blk=osmanya}', "");
Expect(1, 66735, '\p{Blk=		osmanya}', "");
Expect(0, 66735, '\p{^Blk=		osmanya}', "");
Expect(0, 66735, '\P{Blk=		osmanya}', "");
Expect(1, 66735, '\P{^Blk=		osmanya}', "");
Expect(0, 66736, '\p{Blk=		osmanya}', "");
Expect(1, 66736, '\p{^Blk=		osmanya}', "");
Expect(1, 66736, '\P{Blk=		osmanya}', "");
Expect(0, 66736, '\P{^Blk=		osmanya}', "");
Error('\p{Is_Block=	Osmanya:=}');
Error('\P{Is_Block=	Osmanya:=}');
Expect(1, 66735, '\p{Is_Block=osmanya}', "");
Expect(0, 66735, '\p{^Is_Block=osmanya}', "");
Expect(0, 66735, '\P{Is_Block=osmanya}', "");
Expect(1, 66735, '\P{^Is_Block=osmanya}', "");
Expect(0, 66736, '\p{Is_Block=osmanya}', "");
Expect(1, 66736, '\p{^Is_Block=osmanya}', "");
Expect(1, 66736, '\P{Is_Block=osmanya}', "");
Expect(0, 66736, '\P{^Is_Block=osmanya}', "");
Expect(1, 66735, '\p{Is_Block=-OSMANYA}', "");
Expect(0, 66735, '\p{^Is_Block=-OSMANYA}', "");
Expect(0, 66735, '\P{Is_Block=-OSMANYA}', "");
Expect(1, 66735, '\P{^Is_Block=-OSMANYA}', "");
Expect(0, 66736, '\p{Is_Block=-OSMANYA}', "");
Expect(1, 66736, '\p{^Is_Block=-OSMANYA}', "");
Expect(1, 66736, '\P{Is_Block=-OSMANYA}', "");
Expect(0, 66736, '\P{^Is_Block=-OSMANYA}', "");
Error('\p{Is_Blk:_:=Osmanya}');
Error('\P{Is_Blk:_:=Osmanya}');
Expect(1, 66735, '\p{Is_Blk=osmanya}', "");
Expect(0, 66735, '\p{^Is_Blk=osmanya}', "");
Expect(0, 66735, '\P{Is_Blk=osmanya}', "");
Expect(1, 66735, '\P{^Is_Blk=osmanya}', "");
Expect(0, 66736, '\p{Is_Blk=osmanya}', "");
Expect(1, 66736, '\p{^Is_Blk=osmanya}', "");
Expect(1, 66736, '\P{Is_Blk=osmanya}', "");
Expect(0, 66736, '\P{^Is_Blk=osmanya}', "");
Expect(1, 66735, '\p{Is_Blk=_-osmanya}', "");
Expect(0, 66735, '\p{^Is_Blk=_-osmanya}', "");
Expect(0, 66735, '\P{Is_Blk=_-osmanya}', "");
Expect(1, 66735, '\P{^Is_Blk=_-osmanya}', "");
Expect(0, 66736, '\p{Is_Blk=_-osmanya}', "");
Expect(1, 66736, '\p{^Is_Blk=_-osmanya}', "");
Expect(1, 66736, '\P{Is_Blk=_-osmanya}', "");
Expect(0, 66736, '\P{^Is_Blk=_-osmanya}', "");
Error('\p{Block= 	gujarati/a/}');
Error('\P{Block= 	gujarati/a/}');
Expect(1, 2815, '\p{Block=gujarati}', "");
Expect(0, 2815, '\p{^Block=gujarati}', "");
Expect(0, 2815, '\P{Block=gujarati}', "");
Expect(1, 2815, '\P{^Block=gujarati}', "");
Expect(0, 2816, '\p{Block=gujarati}', "");
Expect(1, 2816, '\p{^Block=gujarati}', "");
Expect(1, 2816, '\P{Block=gujarati}', "");
Expect(0, 2816, '\P{^Block=gujarati}', "");
Expect(1, 2815, '\p{Block=	 gujarati}', "");
Expect(0, 2815, '\p{^Block=	 gujarati}', "");
Expect(0, 2815, '\P{Block=	 gujarati}', "");
Expect(1, 2815, '\P{^Block=	 gujarati}', "");
Expect(0, 2816, '\p{Block=	 gujarati}', "");
Expect(1, 2816, '\p{^Block=	 gujarati}', "");
Expect(1, 2816, '\P{Block=	 gujarati}', "");
Expect(0, 2816, '\P{^Block=	 gujarati}', "");
Error('\p{Blk=	/a/gujarati}');
Error('\P{Blk=	/a/gujarati}');
Expect(1, 2815, '\p{Blk=gujarati}', "");
Expect(0, 2815, '\p{^Blk=gujarati}', "");
Expect(0, 2815, '\P{Blk=gujarati}', "");
Expect(1, 2815, '\P{^Blk=gujarati}', "");
Expect(0, 2816, '\p{Blk=gujarati}', "");
Expect(1, 2816, '\p{^Blk=gujarati}', "");
Expect(1, 2816, '\P{Blk=gujarati}', "");
Expect(0, 2816, '\P{^Blk=gujarati}', "");
Expect(1, 2815, '\p{Blk:   _ Gujarati}', "");
Expect(0, 2815, '\p{^Blk:   _ Gujarati}', "");
Expect(0, 2815, '\P{Blk:   _ Gujarati}', "");
Expect(1, 2815, '\P{^Blk:   _ Gujarati}', "");
Expect(0, 2816, '\p{Blk:   _ Gujarati}', "");
Expect(1, 2816, '\p{^Blk:   _ Gujarati}', "");
Expect(1, 2816, '\P{Blk:   _ Gujarati}', "");
Expect(0, 2816, '\P{^Blk:   _ Gujarati}', "");
Error('\p{Is_Block= _Gujarati/a/}');
Error('\P{Is_Block= _Gujarati/a/}');
Expect(1, 2815, '\p{Is_Block=gujarati}', "");
Expect(0, 2815, '\p{^Is_Block=gujarati}', "");
Expect(0, 2815, '\P{Is_Block=gujarati}', "");
Expect(1, 2815, '\P{^Is_Block=gujarati}', "");
Expect(0, 2816, '\p{Is_Block=gujarati}', "");
Expect(1, 2816, '\p{^Is_Block=gujarati}', "");
Expect(1, 2816, '\P{Is_Block=gujarati}', "");
Expect(0, 2816, '\P{^Is_Block=gujarati}', "");
Expect(1, 2815, '\p{Is_Block=	GUJARATI}', "");
Expect(0, 2815, '\p{^Is_Block=	GUJARATI}', "");
Expect(0, 2815, '\P{Is_Block=	GUJARATI}', "");
Expect(1, 2815, '\P{^Is_Block=	GUJARATI}', "");
Expect(0, 2816, '\p{Is_Block=	GUJARATI}', "");
Expect(1, 2816, '\p{^Is_Block=	GUJARATI}', "");
Expect(1, 2816, '\P{Is_Block=	GUJARATI}', "");
Expect(0, 2816, '\P{^Is_Block=	GUJARATI}', "");
Error('\p{Is_Blk=_Gujarati:=}');
Error('\P{Is_Blk=_Gujarati:=}');
Expect(1, 2815, '\p{Is_Blk=gujarati}', "");
Expect(0, 2815, '\p{^Is_Blk=gujarati}', "");
Expect(0, 2815, '\P{Is_Blk=gujarati}', "");
Expect(1, 2815, '\P{^Is_Blk=gujarati}', "");
Expect(0, 2816, '\p{Is_Blk=gujarati}', "");
Expect(1, 2816, '\p{^Is_Blk=gujarati}', "");
Expect(1, 2816, '\P{Is_Blk=gujarati}', "");
Expect(0, 2816, '\P{^Is_Blk=gujarati}', "");
Expect(1, 2815, '\p{Is_Blk=_-Gujarati}', "");
Expect(0, 2815, '\p{^Is_Blk=_-Gujarati}', "");
Expect(0, 2815, '\P{Is_Blk=_-Gujarati}', "");
Expect(1, 2815, '\P{^Is_Blk=_-Gujarati}', "");
Expect(0, 2816, '\p{Is_Blk=_-Gujarati}', "");
Expect(1, 2816, '\p{^Is_Blk=_-Gujarati}', "");
Expect(1, 2816, '\P{Is_Blk=_-Gujarati}', "");
Expect(0, 2816, '\P{^Is_Blk=_-Gujarati}', "");
Error('\p{Block=/a/-_MISCELLANEOUS_SYMBOLS}');
Error('\P{Block=/a/-_MISCELLANEOUS_SYMBOLS}');
Expect(1, 9983, '\p{Block=miscellaneoussymbols}', "");
Expect(0, 9983, '\p{^Block=miscellaneoussymbols}', "");
Expect(0, 9983, '\P{Block=miscellaneoussymbols}', "");
Expect(1, 9983, '\P{^Block=miscellaneoussymbols}', "");
Expect(0, 9984, '\p{Block=miscellaneoussymbols}', "");
Expect(1, 9984, '\p{^Block=miscellaneoussymbols}', "");
Expect(1, 9984, '\P{Block=miscellaneoussymbols}', "");
Expect(0, 9984, '\P{^Block=miscellaneoussymbols}', "");
Expect(1, 9983, '\p{Block=		miscellaneous_SYMBOLS}', "");
Expect(0, 9983, '\p{^Block=		miscellaneous_SYMBOLS}', "");
Expect(0, 9983, '\P{Block=		miscellaneous_SYMBOLS}', "");
Expect(1, 9983, '\P{^Block=		miscellaneous_SYMBOLS}', "");
Expect(0, 9984, '\p{Block=		miscellaneous_SYMBOLS}', "");
Expect(1, 9984, '\p{^Block=		miscellaneous_SYMBOLS}', "");
Expect(1, 9984, '\P{Block=		miscellaneous_SYMBOLS}', "");
Expect(0, 9984, '\P{^Block=		miscellaneous_SYMBOLS}', "");
Error('\p{Blk=:=	-Misc_Symbols}');
Error('\P{Blk=:=	-Misc_Symbols}');
Expect(1, 9983, '\p{Blk=miscsymbols}', "");
Expect(0, 9983, '\p{^Blk=miscsymbols}', "");
Expect(0, 9983, '\P{Blk=miscsymbols}', "");
Expect(1, 9983, '\P{^Blk=miscsymbols}', "");
Expect(0, 9984, '\p{Blk=miscsymbols}', "");
Expect(1, 9984, '\p{^Blk=miscsymbols}', "");
Expect(1, 9984, '\P{Blk=miscsymbols}', "");
Expect(0, 9984, '\P{^Blk=miscsymbols}', "");
Expect(1, 9983, '\p{Blk=misc_Symbols}', "");
Expect(0, 9983, '\p{^Blk=misc_Symbols}', "");
Expect(0, 9983, '\P{Blk=misc_Symbols}', "");
Expect(1, 9983, '\P{^Blk=misc_Symbols}', "");
Expect(0, 9984, '\p{Blk=misc_Symbols}', "");
Expect(1, 9984, '\p{^Blk=misc_Symbols}', "");
Expect(1, 9984, '\P{Blk=misc_Symbols}', "");
Expect(0, 9984, '\P{^Blk=misc_Symbols}', "");
Error('\p{Is_Block=-MISCELLANEOUS_Symbols/a/}');
Error('\P{Is_Block=-MISCELLANEOUS_Symbols/a/}');
Expect(1, 9983, '\p{Is_Block=miscellaneoussymbols}', "");
Expect(0, 9983, '\p{^Is_Block=miscellaneoussymbols}', "");
Expect(0, 9983, '\P{Is_Block=miscellaneoussymbols}', "");
Expect(1, 9983, '\P{^Is_Block=miscellaneoussymbols}', "");
Expect(0, 9984, '\p{Is_Block=miscellaneoussymbols}', "");
Expect(1, 9984, '\p{^Is_Block=miscellaneoussymbols}', "");
Expect(1, 9984, '\P{Is_Block=miscellaneoussymbols}', "");
Expect(0, 9984, '\P{^Is_Block=miscellaneoussymbols}', "");
Expect(1, 9983, '\p{Is_Block: Miscellaneous_SYMBOLS}', "");
Expect(0, 9983, '\p{^Is_Block: Miscellaneous_SYMBOLS}', "");
Expect(0, 9983, '\P{Is_Block: Miscellaneous_SYMBOLS}', "");
Expect(1, 9983, '\P{^Is_Block: Miscellaneous_SYMBOLS}', "");
Expect(0, 9984, '\p{Is_Block: Miscellaneous_SYMBOLS}', "");
Expect(1, 9984, '\p{^Is_Block: Miscellaneous_SYMBOLS}', "");
Expect(1, 9984, '\P{Is_Block: Miscellaneous_SYMBOLS}', "");
Expect(0, 9984, '\P{^Is_Block: Miscellaneous_SYMBOLS}', "");
Error('\p{Is_Blk=:= _MISC_Symbols}');
Error('\P{Is_Blk=:= _MISC_Symbols}');
Expect(1, 9983, '\p{Is_Blk=miscsymbols}', "");
Expect(0, 9983, '\p{^Is_Blk=miscsymbols}', "");
Expect(0, 9983, '\P{Is_Blk=miscsymbols}', "");
Expect(1, 9983, '\P{^Is_Blk=miscsymbols}', "");
Expect(0, 9984, '\p{Is_Blk=miscsymbols}', "");
Expect(1, 9984, '\p{^Is_Blk=miscsymbols}', "");
Expect(1, 9984, '\P{Is_Blk=miscsymbols}', "");
Expect(0, 9984, '\P{^Is_Blk=miscsymbols}', "");
Expect(1, 9983, '\p{Is_Blk= _MISC_Symbols}', "");
Expect(0, 9983, '\p{^Is_Blk= _MISC_Symbols}', "");
Expect(0, 9983, '\P{Is_Blk= _MISC_Symbols}', "");
Expect(1, 9983, '\P{^Is_Blk= _MISC_Symbols}', "");
Expect(0, 9984, '\p{Is_Blk= _MISC_Symbols}', "");
Expect(1, 9984, '\p{^Is_Blk= _MISC_Symbols}', "");
Expect(1, 9984, '\P{Is_Blk= _MISC_Symbols}', "");
Expect(0, 9984, '\P{^Is_Blk= _MISC_Symbols}', "");
Error('\p{Block=Spacing_Modifier_letters/a/}');
Error('\P{Block=Spacing_Modifier_letters/a/}');
Expect(1, 767, '\p{Block=spacingmodifierletters}', "");
Expect(0, 767, '\p{^Block=spacingmodifierletters}', "");
Expect(0, 767, '\P{Block=spacingmodifierletters}', "");
Expect(1, 767, '\P{^Block=spacingmodifierletters}', "");
Expect(0, 768, '\p{Block=spacingmodifierletters}', "");
Expect(1, 768, '\p{^Block=spacingmodifierletters}', "");
Expect(1, 768, '\P{Block=spacingmodifierletters}', "");
Expect(0, 768, '\P{^Block=spacingmodifierletters}', "");
Expect(1, 767, '\p{Block= Spacing_Modifier_Letters}', "");
Expect(0, 767, '\p{^Block= Spacing_Modifier_Letters}', "");
Expect(0, 767, '\P{Block= Spacing_Modifier_Letters}', "");
Expect(1, 767, '\P{^Block= Spacing_Modifier_Letters}', "");
Expect(0, 768, '\p{Block= Spacing_Modifier_Letters}', "");
Expect(1, 768, '\p{^Block= Spacing_Modifier_Letters}', "");
Expect(1, 768, '\P{Block= Spacing_Modifier_Letters}', "");
Expect(0, 768, '\P{^Block= Spacing_Modifier_Letters}', "");
Error('\p{Blk=-:=Modifier_letters}');
Error('\P{Blk=-:=Modifier_letters}');
Expect(1, 767, '\p{Blk=modifierletters}', "");
Expect(0, 767, '\p{^Blk=modifierletters}', "");
Expect(0, 767, '\P{Blk=modifierletters}', "");
Expect(1, 767, '\P{^Blk=modifierletters}', "");
Expect(0, 768, '\p{Blk=modifierletters}', "");
Expect(1, 768, '\p{^Blk=modifierletters}', "");
Expect(1, 768, '\P{Blk=modifierletters}', "");
Expect(0, 768, '\P{^Blk=modifierletters}', "");
Expect(1, 767, '\p{Blk= modifier_LETTERS}', "");
Expect(0, 767, '\p{^Blk= modifier_LETTERS}', "");
Expect(0, 767, '\P{Blk= modifier_LETTERS}', "");
Expect(1, 767, '\P{^Blk= modifier_LETTERS}', "");
Expect(0, 768, '\p{Blk= modifier_LETTERS}', "");
Expect(1, 768, '\p{^Blk= modifier_LETTERS}', "");
Expect(1, 768, '\P{Blk= modifier_LETTERS}', "");
Expect(0, 768, '\P{^Blk= modifier_LETTERS}', "");
Error('\p{Is_Block=__SPACING_Modifier_Letters/a/}');
Error('\P{Is_Block=__SPACING_Modifier_Letters/a/}');
Expect(1, 767, '\p{Is_Block=spacingmodifierletters}', "");
Expect(0, 767, '\p{^Is_Block=spacingmodifierletters}', "");
Expect(0, 767, '\P{Is_Block=spacingmodifierletters}', "");
Expect(1, 767, '\P{^Is_Block=spacingmodifierletters}', "");
Expect(0, 768, '\p{Is_Block=spacingmodifierletters}', "");
Expect(1, 768, '\p{^Is_Block=spacingmodifierletters}', "");
Expect(1, 768, '\P{Is_Block=spacingmodifierletters}', "");
Expect(0, 768, '\P{^Is_Block=spacingmodifierletters}', "");
Expect(1, 767, '\p{Is_Block=_ Spacing_modifier_Letters}', "");
Expect(0, 767, '\p{^Is_Block=_ Spacing_modifier_Letters}', "");
Expect(0, 767, '\P{Is_Block=_ Spacing_modifier_Letters}', "");
Expect(1, 767, '\P{^Is_Block=_ Spacing_modifier_Letters}', "");
Expect(0, 768, '\p{Is_Block=_ Spacing_modifier_Letters}', "");
Expect(1, 768, '\p{^Is_Block=_ Spacing_modifier_Letters}', "");
Expect(1, 768, '\P{Is_Block=_ Spacing_modifier_Letters}', "");
Expect(0, 768, '\P{^Is_Block=_ Spacing_modifier_Letters}', "");
Error('\p{Is_Blk=/a/  Modifier_Letters}');
Error('\P{Is_Blk=/a/  Modifier_Letters}');
Expect(1, 767, '\p{Is_Blk=modifierletters}', "");
Expect(0, 767, '\p{^Is_Blk=modifierletters}', "");
Expect(0, 767, '\P{Is_Blk=modifierletters}', "");
Expect(1, 767, '\P{^Is_Blk=modifierletters}', "");
Expect(0, 768, '\p{Is_Blk=modifierletters}', "");
Expect(1, 768, '\p{^Is_Blk=modifierletters}', "");
Expect(1, 768, '\P{Is_Blk=modifierletters}', "");
Expect(0, 768, '\P{^Is_Blk=modifierletters}', "");
Expect(1, 767, '\p{Is_Blk=	_Modifier_letters}', "");
Expect(0, 767, '\p{^Is_Blk=	_Modifier_letters}', "");
Expect(0, 767, '\P{Is_Blk=	_Modifier_letters}', "");
Expect(1, 767, '\P{^Is_Blk=	_Modifier_letters}', "");
Expect(0, 768, '\p{Is_Blk=	_Modifier_letters}', "");
Expect(1, 768, '\p{^Is_Blk=	_Modifier_letters}', "");
Expect(1, 768, '\P{Is_Blk=	_Modifier_letters}', "");
Expect(0, 768, '\P{^Is_Blk=	_Modifier_letters}', "");
Error('\p{Block:   	Telugu:=}');
Error('\P{Block:   	Telugu:=}');
Expect(1, 3199, '\p{Block=telugu}', "");
Expect(0, 3199, '\p{^Block=telugu}', "");
Expect(0, 3199, '\P{Block=telugu}', "");
Expect(1, 3199, '\P{^Block=telugu}', "");
Expect(0, 3200, '\p{Block=telugu}', "");
Expect(1, 3200, '\p{^Block=telugu}', "");
Expect(1, 3200, '\P{Block=telugu}', "");
Expect(0, 3200, '\P{^Block=telugu}', "");
Expect(1, 3199, '\p{Block= Telugu}', "");
Expect(0, 3199, '\p{^Block= Telugu}', "");
Expect(0, 3199, '\P{Block= Telugu}', "");
Expect(1, 3199, '\P{^Block= Telugu}', "");
Expect(0, 3200, '\p{Block= Telugu}', "");
Expect(1, 3200, '\p{^Block= Telugu}', "");
Expect(1, 3200, '\P{Block= Telugu}', "");
Expect(0, 3200, '\P{^Block= Telugu}', "");
Error('\p{Blk=_	TELUGU:=}');
Error('\P{Blk=_	TELUGU:=}');
Expect(1, 3199, '\p{Blk=telugu}', "");
Expect(0, 3199, '\p{^Blk=telugu}', "");
Expect(0, 3199, '\P{Blk=telugu}', "");
Expect(1, 3199, '\P{^Blk=telugu}', "");
Expect(0, 3200, '\p{Blk=telugu}', "");
Expect(1, 3200, '\p{^Blk=telugu}', "");
Expect(1, 3200, '\P{Blk=telugu}', "");
Expect(0, 3200, '\P{^Blk=telugu}', "");
Expect(1, 3199, '\p{Blk=	-TELUGU}', "");
Expect(0, 3199, '\p{^Blk=	-TELUGU}', "");
Expect(0, 3199, '\P{Blk=	-TELUGU}', "");
Expect(1, 3199, '\P{^Blk=	-TELUGU}', "");
Expect(0, 3200, '\p{Blk=	-TELUGU}', "");
Expect(1, 3200, '\p{^Blk=	-TELUGU}', "");
Expect(1, 3200, '\P{Blk=	-TELUGU}', "");
Expect(0, 3200, '\P{^Blk=	-TELUGU}', "");
Error('\p{Is_Block=/a/-	TELUGU}');
Error('\P{Is_Block=/a/-	TELUGU}');
Expect(1, 3199, '\p{Is_Block=telugu}', "");
Expect(0, 3199, '\p{^Is_Block=telugu}', "");
Expect(0, 3199, '\P{Is_Block=telugu}', "");
Expect(1, 3199, '\P{^Is_Block=telugu}', "");
Expect(0, 3200, '\p{Is_Block=telugu}', "");
Expect(1, 3200, '\p{^Is_Block=telugu}', "");
Expect(1, 3200, '\P{Is_Block=telugu}', "");
Expect(0, 3200, '\P{^Is_Block=telugu}', "");
Expect(1, 3199, '\p{Is_Block=	TELUGU}', "");
Expect(0, 3199, '\p{^Is_Block=	TELUGU}', "");
Expect(0, 3199, '\P{Is_Block=	TELUGU}', "");
Expect(1, 3199, '\P{^Is_Block=	TELUGU}', "");
Expect(0, 3200, '\p{Is_Block=	TELUGU}', "");
Expect(1, 3200, '\p{^Is_Block=	TELUGU}', "");
Expect(1, 3200, '\P{Is_Block=	TELUGU}', "");
Expect(0, 3200, '\P{^Is_Block=	TELUGU}', "");
Error('\p{Is_Blk=:= _telugu}');
Error('\P{Is_Blk=:= _telugu}');
Expect(1, 3199, '\p{Is_Blk=telugu}', "");
Expect(0, 3199, '\p{^Is_Blk=telugu}', "");
Expect(0, 3199, '\P{Is_Blk=telugu}', "");
Expect(1, 3199, '\P{^Is_Blk=telugu}', "");
Expect(0, 3200, '\p{Is_Blk=telugu}', "");
Expect(1, 3200, '\p{^Is_Blk=telugu}', "");
Expect(1, 3200, '\P{Is_Blk=telugu}', "");
Expect(0, 3200, '\P{^Is_Blk=telugu}', "");
Expect(1, 3199, '\p{Is_Blk= 	telugu}', "");
Expect(0, 3199, '\p{^Is_Blk= 	telugu}', "");
Expect(0, 3199, '\P{Is_Blk= 	telugu}', "");
Expect(1, 3199, '\P{^Is_Blk= 	telugu}', "");
Expect(0, 3200, '\p{Is_Blk= 	telugu}', "");
Expect(1, 3200, '\p{^Is_Blk= 	telugu}', "");
Expect(1, 3200, '\P{Is_Blk= 	telugu}', "");
Expect(0, 3200, '\P{^Is_Blk= 	telugu}', "");
Error('\p{Block:   :=MYANMAR_Extended_A}');
Error('\P{Block:   :=MYANMAR_Extended_A}');
Expect(1, 43647, '\p{Block=myanmarextendeda}', "");
Expect(0, 43647, '\p{^Block=myanmarextendeda}', "");
Expect(0, 43647, '\P{Block=myanmarextendeda}', "");
Expect(1, 43647, '\P{^Block=myanmarextendeda}', "");
Expect(0, 43648, '\p{Block=myanmarextendeda}', "");
Expect(1, 43648, '\p{^Block=myanmarextendeda}', "");
Expect(1, 43648, '\P{Block=myanmarextendeda}', "");
Expect(0, 43648, '\P{^Block=myanmarextendeda}', "");
Expect(1, 43647, '\p{Block=	Myanmar_Extended_A}', "");
Expect(0, 43647, '\p{^Block=	Myanmar_Extended_A}', "");
Expect(0, 43647, '\P{Block=	Myanmar_Extended_A}', "");
Expect(1, 43647, '\P{^Block=	Myanmar_Extended_A}', "");
Expect(0, 43648, '\p{Block=	Myanmar_Extended_A}', "");
Expect(1, 43648, '\p{^Block=	Myanmar_Extended_A}', "");
Expect(1, 43648, '\P{Block=	Myanmar_Extended_A}', "");
Expect(0, 43648, '\P{^Block=	Myanmar_Extended_A}', "");
Error('\p{Blk=_:=Myanmar_Ext_A}');
Error('\P{Blk=_:=Myanmar_Ext_A}');
Expect(1, 43647, '\p{Blk: myanmarexta}', "");
Expect(0, 43647, '\p{^Blk: myanmarexta}', "");
Expect(0, 43647, '\P{Blk: myanmarexta}', "");
Expect(1, 43647, '\P{^Blk: myanmarexta}', "");
Expect(0, 43648, '\p{Blk: myanmarexta}', "");
Expect(1, 43648, '\p{^Blk: myanmarexta}', "");
Expect(1, 43648, '\P{Blk: myanmarexta}', "");
Expect(0, 43648, '\P{^Blk: myanmarexta}', "");
Expect(1, 43647, '\p{Blk=	-Myanmar_EXT_a}', "");
Expect(0, 43647, '\p{^Blk=	-Myanmar_EXT_a}', "");
Expect(0, 43647, '\P{Blk=	-Myanmar_EXT_a}', "");
Expect(1, 43647, '\P{^Blk=	-Myanmar_EXT_a}', "");
Expect(0, 43648, '\p{Blk=	-Myanmar_EXT_a}', "");
Expect(1, 43648, '\p{^Blk=	-Myanmar_EXT_a}', "");
Expect(1, 43648, '\P{Blk=	-Myanmar_EXT_a}', "");
Expect(0, 43648, '\P{^Blk=	-Myanmar_EXT_a}', "");
Error('\p{Is_Block=/a/	MYANMAR_EXTENDED_A}');
Error('\P{Is_Block=/a/	MYANMAR_EXTENDED_A}');
Expect(1, 43647, '\p{Is_Block=myanmarextendeda}', "");
Expect(0, 43647, '\p{^Is_Block=myanmarextendeda}', "");
Expect(0, 43647, '\P{Is_Block=myanmarextendeda}', "");
Expect(1, 43647, '\P{^Is_Block=myanmarextendeda}', "");
Expect(0, 43648, '\p{Is_Block=myanmarextendeda}', "");
Expect(1, 43648, '\p{^Is_Block=myanmarextendeda}', "");
Expect(1, 43648, '\P{Is_Block=myanmarextendeda}', "");
Expect(0, 43648, '\P{^Is_Block=myanmarextendeda}', "");
Expect(1, 43647, '\p{Is_Block=		myanmar_Extended_A}', "");
Expect(0, 43647, '\p{^Is_Block=		myanmar_Extended_A}', "");
Expect(0, 43647, '\P{Is_Block=		myanmar_Extended_A}', "");
Expect(1, 43647, '\P{^Is_Block=		myanmar_Extended_A}', "");
Expect(0, 43648, '\p{Is_Block=		myanmar_Extended_A}', "");
Expect(1, 43648, '\p{^Is_Block=		myanmar_Extended_A}', "");
Expect(1, 43648, '\P{Is_Block=		myanmar_Extended_A}', "");
Expect(0, 43648, '\P{^Is_Block=		myanmar_Extended_A}', "");
Error('\p{Is_Blk=_myanmar_Ext_A:=}');
Error('\P{Is_Blk=_myanmar_Ext_A:=}');
Expect(1, 43647, '\p{Is_Blk=myanmarexta}', "");
Expect(0, 43647, '\p{^Is_Blk=myanmarexta}', "");
Expect(0, 43647, '\P{Is_Blk=myanmarexta}', "");
Expect(1, 43647, '\P{^Is_Blk=myanmarexta}', "");
Expect(0, 43648, '\p{Is_Blk=myanmarexta}', "");
Expect(1, 43648, '\p{^Is_Blk=myanmarexta}', "");
Expect(1, 43648, '\P{Is_Blk=myanmarexta}', "");
Expect(0, 43648, '\P{^Is_Blk=myanmarexta}', "");
Expect(1, 43647, '\p{Is_Blk= 	Myanmar_EXT_A}', "");
Expect(0, 43647, '\p{^Is_Blk= 	Myanmar_EXT_A}', "");
Expect(0, 43647, '\P{Is_Blk= 	Myanmar_EXT_A}', "");
Expect(1, 43647, '\P{^Is_Blk= 	Myanmar_EXT_A}', "");
Expect(0, 43648, '\p{Is_Blk= 	Myanmar_EXT_A}', "");
Expect(1, 43648, '\p{^Is_Blk= 	Myanmar_EXT_A}', "");
Expect(1, 43648, '\P{Is_Blk= 	Myanmar_EXT_A}', "");
Expect(0, 43648, '\P{^Is_Blk= 	Myanmar_EXT_A}', "");
Error('\p{Block=-	KHMER:=}');
Error('\P{Block=-	KHMER:=}');
Expect(1, 6143, '\p{Block=khmer}', "");
Expect(0, 6143, '\p{^Block=khmer}', "");
Expect(0, 6143, '\P{Block=khmer}', "");
Expect(1, 6143, '\P{^Block=khmer}', "");
Expect(0, 6144, '\p{Block=khmer}', "");
Expect(1, 6144, '\p{^Block=khmer}', "");
Expect(1, 6144, '\P{Block=khmer}', "");
Expect(0, 6144, '\P{^Block=khmer}', "");
Expect(1, 6143, '\p{Block= 	Khmer}', "");
Expect(0, 6143, '\p{^Block= 	Khmer}', "");
Expect(0, 6143, '\P{Block= 	Khmer}', "");
Expect(1, 6143, '\P{^Block= 	Khmer}', "");
Expect(0, 6144, '\p{Block= 	Khmer}', "");
Expect(1, 6144, '\p{^Block= 	Khmer}', "");
Expect(1, 6144, '\P{Block= 	Khmer}', "");
Expect(0, 6144, '\P{^Block= 	Khmer}', "");
Error('\p{Blk= /a/KHMER}');
Error('\P{Blk= /a/KHMER}');
Expect(1, 6143, '\p{Blk: khmer}', "");
Expect(0, 6143, '\p{^Blk: khmer}', "");
Expect(0, 6143, '\P{Blk: khmer}', "");
Expect(1, 6143, '\P{^Blk: khmer}', "");
Expect(0, 6144, '\p{Blk: khmer}', "");
Expect(1, 6144, '\p{^Blk: khmer}', "");
Expect(1, 6144, '\P{Blk: khmer}', "");
Expect(0, 6144, '\P{^Blk: khmer}', "");
Expect(1, 6143, '\p{Blk= 	Khmer}', "");
Expect(0, 6143, '\p{^Blk= 	Khmer}', "");
Expect(0, 6143, '\P{Blk= 	Khmer}', "");
Expect(1, 6143, '\P{^Blk= 	Khmer}', "");
Expect(0, 6144, '\p{Blk= 	Khmer}', "");
Expect(1, 6144, '\p{^Blk= 	Khmer}', "");
Expect(1, 6144, '\P{Blk= 	Khmer}', "");
Expect(0, 6144, '\P{^Blk= 	Khmer}', "");
Error('\p{Is_Block=/a/	khmer}');
Error('\P{Is_Block=/a/	khmer}');
Expect(1, 6143, '\p{Is_Block=khmer}', "");
Expect(0, 6143, '\p{^Is_Block=khmer}', "");
Expect(0, 6143, '\P{Is_Block=khmer}', "");
Expect(1, 6143, '\P{^Is_Block=khmer}', "");
Expect(0, 6144, '\p{Is_Block=khmer}', "");
Expect(1, 6144, '\p{^Is_Block=khmer}', "");
Expect(1, 6144, '\P{Is_Block=khmer}', "");
Expect(0, 6144, '\P{^Is_Block=khmer}', "");
Expect(1, 6143, '\p{Is_Block=		khmer}', "");
Expect(0, 6143, '\p{^Is_Block=		khmer}', "");
Expect(0, 6143, '\P{Is_Block=		khmer}', "");
Expect(1, 6143, '\P{^Is_Block=		khmer}', "");
Expect(0, 6144, '\p{Is_Block=		khmer}', "");
Expect(1, 6144, '\p{^Is_Block=		khmer}', "");
Expect(1, 6144, '\P{Is_Block=		khmer}', "");
Expect(0, 6144, '\P{^Is_Block=		khmer}', "");
Error('\p{Is_Blk=	_KHMER:=}');
Error('\P{Is_Blk=	_KHMER:=}');
Expect(1, 6143, '\p{Is_Blk=khmer}', "");
Expect(0, 6143, '\p{^Is_Blk=khmer}', "");
Expect(0, 6143, '\P{Is_Blk=khmer}', "");
Expect(1, 6143, '\P{^Is_Blk=khmer}', "");
Expect(0, 6144, '\p{Is_Blk=khmer}', "");
Expect(1, 6144, '\p{^Is_Blk=khmer}', "");
Expect(1, 6144, '\P{Is_Blk=khmer}', "");
Expect(0, 6144, '\P{^Is_Blk=khmer}', "");
Expect(1, 6143, '\p{Is_Blk= Khmer}', "");
Expect(0, 6143, '\p{^Is_Blk= Khmer}', "");
Expect(0, 6143, '\P{Is_Blk= Khmer}', "");
Expect(1, 6143, '\P{^Is_Blk= Khmer}', "");
Expect(0, 6144, '\p{Is_Blk= Khmer}', "");
Expect(1, 6144, '\p{^Is_Blk= Khmer}', "");
Expect(1, 6144, '\P{Is_Blk= Khmer}', "");
Expect(0, 6144, '\P{^Is_Blk= Khmer}', "");
Error('\p{Block=:=		cyrillic_Extended_B}');
Error('\P{Block=:=		cyrillic_Extended_B}');
Expect(1, 42655, '\p{Block=cyrillicextendedb}', "");
Expect(0, 42655, '\p{^Block=cyrillicextendedb}', "");
Expect(0, 42655, '\P{Block=cyrillicextendedb}', "");
Expect(1, 42655, '\P{^Block=cyrillicextendedb}', "");
Expect(0, 42656, '\p{Block=cyrillicextendedb}', "");
Expect(1, 42656, '\p{^Block=cyrillicextendedb}', "");
Expect(1, 42656, '\P{Block=cyrillicextendedb}', "");
Expect(0, 42656, '\P{^Block=cyrillicextendedb}', "");
Expect(1, 42655, '\p{Block= _Cyrillic_Extended_B}', "");
Expect(0, 42655, '\p{^Block= _Cyrillic_Extended_B}', "");
Expect(0, 42655, '\P{Block= _Cyrillic_Extended_B}', "");
Expect(1, 42655, '\P{^Block= _Cyrillic_Extended_B}', "");
Expect(0, 42656, '\p{Block= _Cyrillic_Extended_B}', "");
Expect(1, 42656, '\p{^Block= _Cyrillic_Extended_B}', "");
Expect(1, 42656, '\P{Block= _Cyrillic_Extended_B}', "");
Expect(0, 42656, '\P{^Block= _Cyrillic_Extended_B}', "");
Error('\p{Blk=/a/	Cyrillic_EXT_B}');
Error('\P{Blk=/a/	Cyrillic_EXT_B}');
Expect(1, 42655, '\p{Blk=cyrillicextb}', "");
Expect(0, 42655, '\p{^Blk=cyrillicextb}', "");
Expect(0, 42655, '\P{Blk=cyrillicextb}', "");
Expect(1, 42655, '\P{^Blk=cyrillicextb}', "");
Expect(0, 42656, '\p{Blk=cyrillicextb}', "");
Expect(1, 42656, '\p{^Blk=cyrillicextb}', "");
Expect(1, 42656, '\P{Blk=cyrillicextb}', "");
Expect(0, 42656, '\P{^Blk=cyrillicextb}', "");
Expect(1, 42655, '\p{Blk=	 Cyrillic_Ext_B}', "");
Expect(0, 42655, '\p{^Blk=	 Cyrillic_Ext_B}', "");
Expect(0, 42655, '\P{Blk=	 Cyrillic_Ext_B}', "");
Expect(1, 42655, '\P{^Blk=	 Cyrillic_Ext_B}', "");
Expect(0, 42656, '\p{Blk=	 Cyrillic_Ext_B}', "");
Expect(1, 42656, '\p{^Blk=	 Cyrillic_Ext_B}', "");
Expect(1, 42656, '\P{Blk=	 Cyrillic_Ext_B}', "");
Expect(0, 42656, '\P{^Blk=	 Cyrillic_Ext_B}', "");
Error('\p{Is_Block=/a/_ Cyrillic_Extended_B}');
Error('\P{Is_Block=/a/_ Cyrillic_Extended_B}');
Expect(1, 42655, '\p{Is_Block=cyrillicextendedb}', "");
Expect(0, 42655, '\p{^Is_Block=cyrillicextendedb}', "");
Expect(0, 42655, '\P{Is_Block=cyrillicextendedb}', "");
Expect(1, 42655, '\P{^Is_Block=cyrillicextendedb}', "");
Expect(0, 42656, '\p{Is_Block=cyrillicextendedb}', "");
Expect(1, 42656, '\p{^Is_Block=cyrillicextendedb}', "");
Expect(1, 42656, '\P{Is_Block=cyrillicextendedb}', "");
Expect(0, 42656, '\P{^Is_Block=cyrillicextendedb}', "");
Expect(1, 42655, '\p{Is_Block=- CYRILLIC_Extended_B}', "");
Expect(0, 42655, '\p{^Is_Block=- CYRILLIC_Extended_B}', "");
Expect(0, 42655, '\P{Is_Block=- CYRILLIC_Extended_B}', "");
Expect(1, 42655, '\P{^Is_Block=- CYRILLIC_Extended_B}', "");
Expect(0, 42656, '\p{Is_Block=- CYRILLIC_Extended_B}', "");
Expect(1, 42656, '\p{^Is_Block=- CYRILLIC_Extended_B}', "");
Expect(1, 42656, '\P{Is_Block=- CYRILLIC_Extended_B}', "");
Expect(0, 42656, '\P{^Is_Block=- CYRILLIC_Extended_B}', "");
Error('\p{Is_Blk=/a/- CYRILLIC_Ext_B}');
Error('\P{Is_Blk=/a/- CYRILLIC_Ext_B}');
Expect(1, 42655, '\p{Is_Blk=cyrillicextb}', "");
Expect(0, 42655, '\p{^Is_Blk=cyrillicextb}', "");
Expect(0, 42655, '\P{Is_Blk=cyrillicextb}', "");
Expect(1, 42655, '\P{^Is_Blk=cyrillicextb}', "");
Expect(0, 42656, '\p{Is_Blk=cyrillicextb}', "");
Expect(1, 42656, '\p{^Is_Blk=cyrillicextb}', "");
Expect(1, 42656, '\P{Is_Blk=cyrillicextb}', "");
Expect(0, 42656, '\P{^Is_Blk=cyrillicextb}', "");
Expect(1, 42655, '\p{Is_Blk:  _Cyrillic_ext_B}', "");
Expect(0, 42655, '\p{^Is_Blk:  _Cyrillic_ext_B}', "");
Expect(0, 42655, '\P{Is_Blk:  _Cyrillic_ext_B}', "");
Expect(1, 42655, '\P{^Is_Blk:  _Cyrillic_ext_B}', "");
Expect(0, 42656, '\p{Is_Blk:  _Cyrillic_ext_B}', "");
Expect(1, 42656, '\p{^Is_Blk:  _Cyrillic_ext_B}', "");
Expect(1, 42656, '\P{Is_Blk:  _Cyrillic_ext_B}', "");
Expect(0, 42656, '\P{^Is_Blk:  _Cyrillic_ext_B}', "");
Error('\p{Block=:= Cyrillic}');
Error('\P{Block=:= Cyrillic}');
Expect(1, 1279, '\p{Block:   cyrillic}', "");
Expect(0, 1279, '\p{^Block:   cyrillic}', "");
Expect(0, 1279, '\P{Block:   cyrillic}', "");
Expect(1, 1279, '\P{^Block:   cyrillic}', "");
Expect(0, 1280, '\p{Block:   cyrillic}', "");
Expect(1, 1280, '\p{^Block:   cyrillic}', "");
Expect(1, 1280, '\P{Block:   cyrillic}', "");
Expect(0, 1280, '\P{^Block:   cyrillic}', "");
Expect(1, 1279, '\p{Block=__Cyrillic}', "");
Expect(0, 1279, '\p{^Block=__Cyrillic}', "");
Expect(0, 1279, '\P{Block=__Cyrillic}', "");
Expect(1, 1279, '\P{^Block=__Cyrillic}', "");
Expect(0, 1280, '\p{Block=__Cyrillic}', "");
Expect(1, 1280, '\p{^Block=__Cyrillic}', "");
Expect(1, 1280, '\P{Block=__Cyrillic}', "");
Expect(0, 1280, '\P{^Block=__Cyrillic}', "");
Error('\p{Blk=/a/_CYRILLIC}');
Error('\P{Blk=/a/_CYRILLIC}');
Expect(1, 1279, '\p{Blk: cyrillic}', "");
Expect(0, 1279, '\p{^Blk: cyrillic}', "");
Expect(0, 1279, '\P{Blk: cyrillic}', "");
Expect(1, 1279, '\P{^Blk: cyrillic}', "");
Expect(0, 1280, '\p{Blk: cyrillic}', "");
Expect(1, 1280, '\p{^Blk: cyrillic}', "");
Expect(1, 1280, '\P{Blk: cyrillic}', "");
Expect(0, 1280, '\P{^Blk: cyrillic}', "");
Expect(1, 1279, '\p{Blk=	Cyrillic}', "");
Expect(0, 1279, '\p{^Blk=	Cyrillic}', "");
Expect(0, 1279, '\P{Blk=	Cyrillic}', "");
Expect(1, 1279, '\P{^Blk=	Cyrillic}', "");
Expect(0, 1280, '\p{Blk=	Cyrillic}', "");
Expect(1, 1280, '\p{^Blk=	Cyrillic}', "");
Expect(1, 1280, '\P{Blk=	Cyrillic}', "");
Expect(0, 1280, '\P{^Blk=	Cyrillic}', "");
Error('\p{Is_Block=:=-Cyrillic}');
Error('\P{Is_Block=:=-Cyrillic}');
Expect(1, 1279, '\p{Is_Block=cyrillic}', "");
Expect(0, 1279, '\p{^Is_Block=cyrillic}', "");
Expect(0, 1279, '\P{Is_Block=cyrillic}', "");
Expect(1, 1279, '\P{^Is_Block=cyrillic}', "");
Expect(0, 1280, '\p{Is_Block=cyrillic}', "");
Expect(1, 1280, '\p{^Is_Block=cyrillic}', "");
Expect(1, 1280, '\P{Is_Block=cyrillic}', "");
Expect(0, 1280, '\P{^Is_Block=cyrillic}', "");
Expect(1, 1279, '\p{Is_Block=	_CYRILLIC}', "");
Expect(0, 1279, '\p{^Is_Block=	_CYRILLIC}', "");
Expect(0, 1279, '\P{Is_Block=	_CYRILLIC}', "");
Expect(1, 1279, '\P{^Is_Block=	_CYRILLIC}', "");
Expect(0, 1280, '\p{Is_Block=	_CYRILLIC}', "");
Expect(1, 1280, '\p{^Is_Block=	_CYRILLIC}', "");
Expect(1, 1280, '\P{Is_Block=	_CYRILLIC}', "");
Expect(0, 1280, '\P{^Is_Block=	_CYRILLIC}', "");
Error('\p{Is_Blk=:=_ CYRILLIC}');
Error('\P{Is_Blk=:=_ CYRILLIC}');
Expect(1, 1279, '\p{Is_Blk=cyrillic}', "");
Expect(0, 1279, '\p{^Is_Blk=cyrillic}', "");
Expect(0, 1279, '\P{Is_Blk=cyrillic}', "");
Expect(1, 1279, '\P{^Is_Blk=cyrillic}', "");
Expect(0, 1280, '\p{Is_Blk=cyrillic}', "");
Expect(1, 1280, '\p{^Is_Blk=cyrillic}', "");
Expect(1, 1280, '\P{Is_Blk=cyrillic}', "");
Expect(0, 1280, '\P{^Is_Blk=cyrillic}', "");
Expect(1, 1279, '\p{Is_Blk=	_cyrillic}', "");
Expect(0, 1279, '\p{^Is_Blk=	_cyrillic}', "");
Expect(0, 1279, '\P{Is_Blk=	_cyrillic}', "");
Expect(1, 1279, '\P{^Is_Blk=	_cyrillic}', "");
Expect(0, 1280, '\p{Is_Blk=	_cyrillic}', "");
Expect(1, 1280, '\p{^Is_Blk=	_cyrillic}', "");
Expect(1, 1280, '\P{Is_Blk=	_cyrillic}', "");
Expect(0, 1280, '\P{^Is_Blk=	_cyrillic}', "");
Error('\p{Block=/a/ _Syriac}');
Error('\P{Block=/a/ _Syriac}');
Expect(1, 1871, '\p{Block=syriac}', "");
Expect(0, 1871, '\p{^Block=syriac}', "");
Expect(0, 1871, '\P{Block=syriac}', "");
Expect(1, 1871, '\P{^Block=syriac}', "");
Expect(0, 1872, '\p{Block=syriac}', "");
Expect(1, 1872, '\p{^Block=syriac}', "");
Expect(1, 1872, '\P{Block=syriac}', "");
Expect(0, 1872, '\P{^Block=syriac}', "");
Expect(1, 1871, '\p{Block=	 SYRIAC}', "");
Expect(0, 1871, '\p{^Block=	 SYRIAC}', "");
Expect(0, 1871, '\P{Block=	 SYRIAC}', "");
Expect(1, 1871, '\P{^Block=	 SYRIAC}', "");
Expect(0, 1872, '\p{Block=	 SYRIAC}', "");
Expect(1, 1872, '\p{^Block=	 SYRIAC}', "");
Expect(1, 1872, '\P{Block=	 SYRIAC}', "");
Expect(0, 1872, '\P{^Block=	 SYRIAC}', "");
Error('\p{Blk=	-SYRIAC/a/}');
Error('\P{Blk=	-SYRIAC/a/}');
Expect(1, 1871, '\p{Blk=syriac}', "");
Expect(0, 1871, '\p{^Blk=syriac}', "");
Expect(0, 1871, '\P{Blk=syriac}', "");
Expect(1, 1871, '\P{^Blk=syriac}', "");
Expect(0, 1872, '\p{Blk=syriac}', "");
Expect(1, 1872, '\p{^Blk=syriac}', "");
Expect(1, 1872, '\P{Blk=syriac}', "");
Expect(0, 1872, '\P{^Blk=syriac}', "");
Expect(1, 1871, '\p{Blk=	Syriac}', "");
Expect(0, 1871, '\p{^Blk=	Syriac}', "");
Expect(0, 1871, '\P{Blk=	Syriac}', "");
Expect(1, 1871, '\P{^Blk=	Syriac}', "");
Expect(0, 1872, '\p{Blk=	Syriac}', "");
Expect(1, 1872, '\p{^Blk=	Syriac}', "");
Expect(1, 1872, '\P{Blk=	Syriac}', "");
Expect(0, 1872, '\P{^Blk=	Syriac}', "");
Error('\p{Is_Block:-:=Syriac}');
Error('\P{Is_Block:-:=Syriac}');
Expect(1, 1871, '\p{Is_Block=syriac}', "");
Expect(0, 1871, '\p{^Is_Block=syriac}', "");
Expect(0, 1871, '\P{Is_Block=syriac}', "");
Expect(1, 1871, '\P{^Is_Block=syriac}', "");
Expect(0, 1872, '\p{Is_Block=syriac}', "");
Expect(1, 1872, '\p{^Is_Block=syriac}', "");
Expect(1, 1872, '\P{Is_Block=syriac}', "");
Expect(0, 1872, '\P{^Is_Block=syriac}', "");
Expect(1, 1871, '\p{Is_Block=		syriac}', "");
Expect(0, 1871, '\p{^Is_Block=		syriac}', "");
Expect(0, 1871, '\P{Is_Block=		syriac}', "");
Expect(1, 1871, '\P{^Is_Block=		syriac}', "");
Expect(0, 1872, '\p{Is_Block=		syriac}', "");
Expect(1, 1872, '\p{^Is_Block=		syriac}', "");
Expect(1, 1872, '\P{Is_Block=		syriac}', "");
Expect(0, 1872, '\P{^Is_Block=		syriac}', "");
Error('\p{Is_Blk=	syriac:=}');
Error('\P{Is_Blk=	syriac:=}');
Expect(1, 1871, '\p{Is_Blk=syriac}', "");
Expect(0, 1871, '\p{^Is_Blk=syriac}', "");
Expect(0, 1871, '\P{Is_Blk=syriac}', "");
Expect(1, 1871, '\P{^Is_Blk=syriac}', "");
Expect(0, 1872, '\p{Is_Blk=syriac}', "");
Expect(1, 1872, '\p{^Is_Blk=syriac}', "");
Expect(1, 1872, '\P{Is_Blk=syriac}', "");
Expect(0, 1872, '\P{^Is_Blk=syriac}', "");
Expect(1, 1871, '\p{Is_Blk= SYRIAC}', "");
Expect(0, 1871, '\p{^Is_Blk= SYRIAC}', "");
Expect(0, 1871, '\P{Is_Blk= SYRIAC}', "");
Expect(1, 1871, '\P{^Is_Blk= SYRIAC}', "");
Expect(0, 1872, '\p{Is_Blk= SYRIAC}', "");
Expect(1, 1872, '\p{^Is_Blk= SYRIAC}', "");
Expect(1, 1872, '\P{Is_Blk= SYRIAC}', "");
Expect(0, 1872, '\P{^Is_Blk= SYRIAC}', "");
Error('\p{Block=	:=Cham}');
Error('\P{Block=	:=Cham}');
Expect(1, 43615, '\p{Block=cham}', "");
Expect(0, 43615, '\p{^Block=cham}', "");
Expect(0, 43615, '\P{Block=cham}', "");
Expect(1, 43615, '\P{^Block=cham}', "");
Expect(0, 43616, '\p{Block=cham}', "");
Expect(1, 43616, '\p{^Block=cham}', "");
Expect(1, 43616, '\P{Block=cham}', "");
Expect(0, 43616, '\P{^Block=cham}', "");
Expect(1, 43615, '\p{Block= _cham}', "");
Expect(0, 43615, '\p{^Block= _cham}', "");
Expect(0, 43615, '\P{Block= _cham}', "");
Expect(1, 43615, '\P{^Block= _cham}', "");
Expect(0, 43616, '\p{Block= _cham}', "");
Expect(1, 43616, '\p{^Block= _cham}', "");
Expect(1, 43616, '\P{Block= _cham}', "");
Expect(0, 43616, '\P{^Block= _cham}', "");
Error('\p{Blk=:=		Cham}');
Error('\P{Blk=:=		Cham}');
Expect(1, 43615, '\p{Blk=cham}', "");
Expect(0, 43615, '\p{^Blk=cham}', "");
Expect(0, 43615, '\P{Blk=cham}', "");
Expect(1, 43615, '\P{^Blk=cham}', "");
Expect(0, 43616, '\p{Blk=cham}', "");
Expect(1, 43616, '\p{^Blk=cham}', "");
Expect(1, 43616, '\P{Blk=cham}', "");
Expect(0, 43616, '\P{^Blk=cham}', "");
Expect(1, 43615, '\p{Blk=-cham}', "");
Expect(0, 43615, '\p{^Blk=-cham}', "");
Expect(0, 43615, '\P{Blk=-cham}', "");
Expect(1, 43615, '\P{^Blk=-cham}', "");
Expect(0, 43616, '\p{Blk=-cham}', "");
Expect(1, 43616, '\p{^Blk=-cham}', "");
Expect(1, 43616, '\P{Blk=-cham}', "");
Expect(0, 43616, '\P{^Blk=-cham}', "");
Error('\p{Is_Block: -/a/CHAM}');
Error('\P{Is_Block: -/a/CHAM}');
Expect(1, 43615, '\p{Is_Block=cham}', "");
Expect(0, 43615, '\p{^Is_Block=cham}', "");
Expect(0, 43615, '\P{Is_Block=cham}', "");
Expect(1, 43615, '\P{^Is_Block=cham}', "");
Expect(0, 43616, '\p{Is_Block=cham}', "");
Expect(1, 43616, '\p{^Is_Block=cham}', "");
Expect(1, 43616, '\P{Is_Block=cham}', "");
Expect(0, 43616, '\P{^Is_Block=cham}', "");
Expect(1, 43615, '\p{Is_Block=	Cham}', "");
Expect(0, 43615, '\p{^Is_Block=	Cham}', "");
Expect(0, 43615, '\P{Is_Block=	Cham}', "");
Expect(1, 43615, '\P{^Is_Block=	Cham}', "");
Expect(0, 43616, '\p{Is_Block=	Cham}', "");
Expect(1, 43616, '\p{^Is_Block=	Cham}', "");
Expect(1, 43616, '\P{Is_Block=	Cham}', "");
Expect(0, 43616, '\P{^Is_Block=	Cham}', "");
Error('\p{Is_Blk=:=Cham}');
Error('\P{Is_Blk=:=Cham}');
Expect(1, 43615, '\p{Is_Blk=cham}', "");
Expect(0, 43615, '\p{^Is_Blk=cham}', "");
Expect(0, 43615, '\P{Is_Blk=cham}', "");
Expect(1, 43615, '\P{^Is_Blk=cham}', "");
Expect(0, 43616, '\p{Is_Blk=cham}', "");
Expect(1, 43616, '\p{^Is_Blk=cham}', "");
Expect(1, 43616, '\P{Is_Blk=cham}', "");
Expect(0, 43616, '\P{^Is_Blk=cham}', "");
Expect(1, 43615, '\p{Is_Blk= 	Cham}', "");
Expect(0, 43615, '\p{^Is_Blk= 	Cham}', "");
Expect(0, 43615, '\P{Is_Blk= 	Cham}', "");
Expect(1, 43615, '\P{^Is_Blk= 	Cham}', "");
Expect(0, 43616, '\p{Is_Blk= 	Cham}', "");
Expect(1, 43616, '\p{^Is_Blk= 	Cham}', "");
Expect(1, 43616, '\P{Is_Blk= 	Cham}', "");
Expect(0, 43616, '\P{^Is_Blk= 	Cham}', "");
Error('\p{Block: :=Ideographic_DESCRIPTION_Characters}');
Error('\P{Block: :=Ideographic_DESCRIPTION_Characters}');
Expect(1, 12287, '\p{Block=ideographicdescriptioncharacters}', "");
Expect(0, 12287, '\p{^Block=ideographicdescriptioncharacters}', "");
Expect(0, 12287, '\P{Block=ideographicdescriptioncharacters}', "");
Expect(1, 12287, '\P{^Block=ideographicdescriptioncharacters}', "");
Expect(0, 12288, '\p{Block=ideographicdescriptioncharacters}', "");
Expect(1, 12288, '\p{^Block=ideographicdescriptioncharacters}', "");
Expect(1, 12288, '\P{Block=ideographicdescriptioncharacters}', "");
Expect(0, 12288, '\P{^Block=ideographicdescriptioncharacters}', "");
Expect(1, 12287, '\p{Block=_	Ideographic_description_CHARACTERS}', "");
Expect(0, 12287, '\p{^Block=_	Ideographic_description_CHARACTERS}', "");
Expect(0, 12287, '\P{Block=_	Ideographic_description_CHARACTERS}', "");
Expect(1, 12287, '\P{^Block=_	Ideographic_description_CHARACTERS}', "");
Expect(0, 12288, '\p{Block=_	Ideographic_description_CHARACTERS}', "");
Expect(1, 12288, '\p{^Block=_	Ideographic_description_CHARACTERS}', "");
Expect(1, 12288, '\P{Block=_	Ideographic_description_CHARACTERS}', "");
Expect(0, 12288, '\P{^Block=_	Ideographic_description_CHARACTERS}', "");
Error('\p{Blk=-/a/IDC}');
Error('\P{Blk=-/a/IDC}');
Expect(1, 12287, '\p{Blk=idc}', "");
Expect(0, 12287, '\p{^Blk=idc}', "");
Expect(0, 12287, '\P{Blk=idc}', "");
Expect(1, 12287, '\P{^Blk=idc}', "");
Expect(0, 12288, '\p{Blk=idc}', "");
Expect(1, 12288, '\p{^Blk=idc}', "");
Expect(1, 12288, '\P{Blk=idc}', "");
Expect(0, 12288, '\P{^Blk=idc}', "");
Expect(1, 12287, '\p{Blk=--IDC}', "");
Expect(0, 12287, '\p{^Blk=--IDC}', "");
Expect(0, 12287, '\P{Blk=--IDC}', "");
Expect(1, 12287, '\P{^Blk=--IDC}', "");
Expect(0, 12288, '\p{Blk=--IDC}', "");
Expect(1, 12288, '\p{^Blk=--IDC}', "");
Expect(1, 12288, '\P{Blk=--IDC}', "");
Expect(0, 12288, '\P{^Blk=--IDC}', "");
Error('\p{Is_Block=	ideographic_description_CHARACTERS:=}');
Error('\P{Is_Block=	ideographic_description_CHARACTERS:=}');
Expect(1, 12287, '\p{Is_Block=ideographicdescriptioncharacters}', "");
Expect(0, 12287, '\p{^Is_Block=ideographicdescriptioncharacters}', "");
Expect(0, 12287, '\P{Is_Block=ideographicdescriptioncharacters}', "");
Expect(1, 12287, '\P{^Is_Block=ideographicdescriptioncharacters}', "");
Expect(0, 12288, '\p{Is_Block=ideographicdescriptioncharacters}', "");
Expect(1, 12288, '\p{^Is_Block=ideographicdescriptioncharacters}', "");
Expect(1, 12288, '\P{Is_Block=ideographicdescriptioncharacters}', "");
Expect(0, 12288, '\P{^Is_Block=ideographicdescriptioncharacters}', "");
Expect(1, 12287, '\p{Is_Block=	-Ideographic_DESCRIPTION_characters}', "");
Expect(0, 12287, '\p{^Is_Block=	-Ideographic_DESCRIPTION_characters}', "");
Expect(0, 12287, '\P{Is_Block=	-Ideographic_DESCRIPTION_characters}', "");
Expect(1, 12287, '\P{^Is_Block=	-Ideographic_DESCRIPTION_characters}', "");
Expect(0, 12288, '\p{Is_Block=	-Ideographic_DESCRIPTION_characters}', "");
Expect(1, 12288, '\p{^Is_Block=	-Ideographic_DESCRIPTION_characters}', "");
Expect(1, 12288, '\P{Is_Block=	-Ideographic_DESCRIPTION_characters}', "");
Expect(0, 12288, '\P{^Is_Block=	-Ideographic_DESCRIPTION_characters}', "");
Error('\p{Is_Blk=:=idc}');
Error('\P{Is_Blk=:=idc}');
Expect(1, 12287, '\p{Is_Blk=idc}', "");
Expect(0, 12287, '\p{^Is_Blk=idc}', "");
Expect(0, 12287, '\P{Is_Blk=idc}', "");
Expect(1, 12287, '\P{^Is_Blk=idc}', "");
Expect(0, 12288, '\p{Is_Blk=idc}', "");
Expect(1, 12288, '\p{^Is_Blk=idc}', "");
Expect(1, 12288, '\P{Is_Blk=idc}', "");
Expect(0, 12288, '\P{^Is_Blk=idc}', "");
Expect(1, 12287, '\p{Is_Blk=-IDC}', "");
Expect(0, 12287, '\p{^Is_Blk=-IDC}', "");
Expect(0, 12287, '\P{Is_Blk=-IDC}', "");
Expect(1, 12287, '\P{^Is_Blk=-IDC}', "");
Expect(0, 12288, '\p{Is_Blk=-IDC}', "");
Expect(1, 12288, '\p{^Is_Blk=-IDC}', "");
Expect(1, 12288, '\P{Is_Blk=-IDC}', "");
Expect(0, 12288, '\P{^Is_Blk=-IDC}', "");
Error('\p{Block= Variation_Selectors:=}');
Error('\P{Block= Variation_Selectors:=}');
Expect(1, 65039, '\p{Block=variationselectors}', "");
Expect(0, 65039, '\p{^Block=variationselectors}', "");
Expect(0, 65039, '\P{Block=variationselectors}', "");
Expect(1, 65039, '\P{^Block=variationselectors}', "");
Expect(0, 65040, '\p{Block=variationselectors}', "");
Expect(1, 65040, '\p{^Block=variationselectors}', "");
Expect(1, 65040, '\P{Block=variationselectors}', "");
Expect(0, 65040, '\P{^Block=variationselectors}', "");
Expect(1, 65039, '\p{Block=-VARIATION_SELECTORS}', "");
Expect(0, 65039, '\p{^Block=-VARIATION_SELECTORS}', "");
Expect(0, 65039, '\P{Block=-VARIATION_SELECTORS}', "");
Expect(1, 65039, '\P{^Block=-VARIATION_SELECTORS}', "");
Expect(0, 65040, '\p{Block=-VARIATION_SELECTORS}', "");
Expect(1, 65040, '\p{^Block=-VARIATION_SELECTORS}', "");
Expect(1, 65040, '\P{Block=-VARIATION_SELECTORS}', "");
Expect(0, 65040, '\P{^Block=-VARIATION_SELECTORS}', "");
Error('\p{Blk= /a/VS}');
Error('\P{Blk= /a/VS}');
Expect(1, 65039, '\p{Blk=vs}', "");
Expect(0, 65039, '\p{^Blk=vs}', "");
Expect(0, 65039, '\P{Blk=vs}', "");
Expect(1, 65039, '\P{^Blk=vs}', "");
Expect(0, 65040, '\p{Blk=vs}', "");
Expect(1, 65040, '\p{^Blk=vs}', "");
Expect(1, 65040, '\P{Blk=vs}', "");
Expect(0, 65040, '\P{^Blk=vs}', "");
Expect(1, 65039, '\p{Blk: 	VS}', "");
Expect(0, 65039, '\p{^Blk: 	VS}', "");
Expect(0, 65039, '\P{Blk: 	VS}', "");
Expect(1, 65039, '\P{^Blk: 	VS}', "");
Expect(0, 65040, '\p{Blk: 	VS}', "");
Expect(1, 65040, '\p{^Blk: 	VS}', "");
Expect(1, 65040, '\P{Blk: 	VS}', "");
Expect(0, 65040, '\P{^Blk: 	VS}', "");
Error('\p{Is_Block=:=- Variation_Selectors}');
Error('\P{Is_Block=:=- Variation_Selectors}');
Expect(1, 65039, '\p{Is_Block=variationselectors}', "");
Expect(0, 65039, '\p{^Is_Block=variationselectors}', "");
Expect(0, 65039, '\P{Is_Block=variationselectors}', "");
Expect(1, 65039, '\P{^Is_Block=variationselectors}', "");
Expect(0, 65040, '\p{Is_Block=variationselectors}', "");
Expect(1, 65040, '\p{^Is_Block=variationselectors}', "");
Expect(1, 65040, '\P{Is_Block=variationselectors}', "");
Expect(0, 65040, '\P{^Is_Block=variationselectors}', "");
Expect(1, 65039, '\p{Is_Block=	 Variation_SELECTORS}', "");
Expect(0, 65039, '\p{^Is_Block=	 Variation_SELECTORS}', "");
Expect(0, 65039, '\P{Is_Block=	 Variation_SELECTORS}', "");
Expect(1, 65039, '\P{^Is_Block=	 Variation_SELECTORS}', "");
Expect(0, 65040, '\p{Is_Block=	 Variation_SELECTORS}', "");
Expect(1, 65040, '\p{^Is_Block=	 Variation_SELECTORS}', "");
Expect(1, 65040, '\P{Is_Block=	 Variation_SELECTORS}', "");
Expect(0, 65040, '\P{^Is_Block=	 Variation_SELECTORS}', "");
Error('\p{Is_Blk=:=-VS}');
Error('\P{Is_Blk=:=-VS}');
Expect(1, 65039, '\p{Is_Blk=vs}', "");
Expect(0, 65039, '\p{^Is_Blk=vs}', "");
Expect(0, 65039, '\P{Is_Blk=vs}', "");
Expect(1, 65039, '\P{^Is_Blk=vs}', "");
Expect(0, 65040, '\p{Is_Blk=vs}', "");
Expect(1, 65040, '\p{^Is_Blk=vs}', "");
Expect(1, 65040, '\P{Is_Blk=vs}', "");
Expect(0, 65040, '\P{^Is_Blk=vs}', "");
Expect(1, 65039, '\p{Is_Blk:_VS}', "");
Expect(0, 65039, '\p{^Is_Blk:_VS}', "");
Expect(0, 65039, '\P{Is_Blk:_VS}', "");
Expect(1, 65039, '\P{^Is_Blk:_VS}', "");
Expect(0, 65040, '\p{Is_Blk:_VS}', "");
Expect(1, 65040, '\p{^Is_Blk:_VS}', "");
Expect(1, 65040, '\P{Is_Blk:_VS}', "");
Expect(0, 65040, '\P{^Is_Blk:_VS}', "");
Error('\p{Block: NEW_Tai_LUE/a/}');
Error('\P{Block: NEW_Tai_LUE/a/}');
Expect(1, 6623, '\p{Block:	newtailue}', "");
Expect(0, 6623, '\p{^Block:	newtailue}', "");
Expect(0, 6623, '\P{Block:	newtailue}', "");
Expect(1, 6623, '\P{^Block:	newtailue}', "");
Expect(0, 6624, '\p{Block:	newtailue}', "");
Expect(1, 6624, '\p{^Block:	newtailue}', "");
Expect(1, 6624, '\P{Block:	newtailue}', "");
Expect(0, 6624, '\P{^Block:	newtailue}', "");
Expect(1, 6623, '\p{Block=-New_Tai_Lue}', "");
Expect(0, 6623, '\p{^Block=-New_Tai_Lue}', "");
Expect(0, 6623, '\P{Block=-New_Tai_Lue}', "");
Expect(1, 6623, '\P{^Block=-New_Tai_Lue}', "");
Expect(0, 6624, '\p{Block=-New_Tai_Lue}', "");
Expect(1, 6624, '\p{^Block=-New_Tai_Lue}', "");
Expect(1, 6624, '\P{Block=-New_Tai_Lue}', "");
Expect(0, 6624, '\P{^Block=-New_Tai_Lue}', "");
Error('\p{Blk= 	New_tai_lue:=}');
Error('\P{Blk= 	New_tai_lue:=}');
Expect(1, 6623, '\p{Blk=newtailue}', "");
Expect(0, 6623, '\p{^Blk=newtailue}', "");
Expect(0, 6623, '\P{Blk=newtailue}', "");
Expect(1, 6623, '\P{^Blk=newtailue}', "");
Expect(0, 6624, '\p{Blk=newtailue}', "");
Expect(1, 6624, '\p{^Blk=newtailue}', "");
Expect(1, 6624, '\P{Blk=newtailue}', "");
Expect(0, 6624, '\P{^Blk=newtailue}', "");
Expect(1, 6623, '\p{Blk=_New_TAI_Lue}', "");
Expect(0, 6623, '\p{^Blk=_New_TAI_Lue}', "");
Expect(0, 6623, '\P{Blk=_New_TAI_Lue}', "");
Expect(1, 6623, '\P{^Blk=_New_TAI_Lue}', "");
Expect(0, 6624, '\p{Blk=_New_TAI_Lue}', "");
Expect(1, 6624, '\p{^Blk=_New_TAI_Lue}', "");
Expect(1, 6624, '\P{Blk=_New_TAI_Lue}', "");
Expect(0, 6624, '\P{^Blk=_New_TAI_Lue}', "");
Error('\p{Is_Block=:= New_Tai_Lue}');
Error('\P{Is_Block=:= New_Tai_Lue}');
Expect(1, 6623, '\p{Is_Block=newtailue}', "");
Expect(0, 6623, '\p{^Is_Block=newtailue}', "");
Expect(0, 6623, '\P{Is_Block=newtailue}', "");
Expect(1, 6623, '\P{^Is_Block=newtailue}', "");
Expect(0, 6624, '\p{Is_Block=newtailue}', "");
Expect(1, 6624, '\p{^Is_Block=newtailue}', "");
Expect(1, 6624, '\P{Is_Block=newtailue}', "");
Expect(0, 6624, '\P{^Is_Block=newtailue}', "");
Expect(1, 6623, '\p{Is_Block=	-new_TAI_lue}', "");
Expect(0, 6623, '\p{^Is_Block=	-new_TAI_lue}', "");
Expect(0, 6623, '\P{Is_Block=	-new_TAI_lue}', "");
Expect(1, 6623, '\P{^Is_Block=	-new_TAI_lue}', "");
Expect(0, 6624, '\p{Is_Block=	-new_TAI_lue}', "");
Expect(1, 6624, '\p{^Is_Block=	-new_TAI_lue}', "");
Expect(1, 6624, '\P{Is_Block=	-new_TAI_lue}', "");
Expect(0, 6624, '\P{^Is_Block=	-new_TAI_lue}', "");
Error('\p{Is_Blk=	New_Tai_lue/a/}');
Error('\P{Is_Blk=	New_Tai_lue/a/}');
Expect(1, 6623, '\p{Is_Blk=newtailue}', "");
Expect(0, 6623, '\p{^Is_Blk=newtailue}', "");
Expect(0, 6623, '\P{Is_Blk=newtailue}', "");
Expect(1, 6623, '\P{^Is_Blk=newtailue}', "");
Expect(0, 6624, '\p{Is_Blk=newtailue}', "");
Expect(1, 6624, '\p{^Is_Blk=newtailue}', "");
Expect(1, 6624, '\P{Is_Blk=newtailue}', "");
Expect(0, 6624, '\P{^Is_Blk=newtailue}', "");
Expect(1, 6623, '\p{Is_Blk=-NEW_Tai_lue}', "");
Expect(0, 6623, '\p{^Is_Blk=-NEW_Tai_lue}', "");
Expect(0, 6623, '\P{Is_Blk=-NEW_Tai_lue}', "");
Expect(1, 6623, '\P{^Is_Blk=-NEW_Tai_lue}', "");
Expect(0, 6624, '\p{Is_Blk=-NEW_Tai_lue}', "");
Expect(1, 6624, '\p{^Is_Blk=-NEW_Tai_lue}', "");
Expect(1, 6624, '\P{Is_Blk=-NEW_Tai_lue}', "");
Expect(0, 6624, '\P{^Is_Blk=-NEW_Tai_lue}', "");
Error('\p{Block=_ HIGH_private_Use_SURROGATES:=}');
Error('\P{Block=_ HIGH_private_Use_SURROGATES:=}');
Expect(1, 56319, '\p{Block=highprivateusesurrogates}', "");
Expect(0, 56319, '\p{^Block=highprivateusesurrogates}', "");
Expect(0, 56319, '\P{Block=highprivateusesurrogates}', "");
Expect(1, 56319, '\P{^Block=highprivateusesurrogates}', "");
Expect(0, 57344, '\p{Block=highprivateusesurrogates}', "");
Expect(1, 57344, '\p{^Block=highprivateusesurrogates}', "");
Expect(1, 57344, '\P{Block=highprivateusesurrogates}', "");
Expect(0, 57344, '\P{^Block=highprivateusesurrogates}', "");
Expect(1, 56319, '\p{Block= _High_Private_USE_Surrogates}', "");
Expect(0, 56319, '\p{^Block= _High_Private_USE_Surrogates}', "");
Expect(0, 56319, '\P{Block= _High_Private_USE_Surrogates}', "");
Expect(1, 56319, '\P{^Block= _High_Private_USE_Surrogates}', "");
Expect(0, 57344, '\p{Block= _High_Private_USE_Surrogates}', "");
Expect(1, 57344, '\p{^Block= _High_Private_USE_Surrogates}', "");
Expect(1, 57344, '\P{Block= _High_Private_USE_Surrogates}', "");
Expect(0, 57344, '\P{^Block= _High_Private_USE_Surrogates}', "");
Error('\p{Blk:   :=	-High_PU_Surrogates}');
Error('\P{Blk:   :=	-High_PU_Surrogates}');
Expect(1, 56319, '\p{Blk=highpusurrogates}', "");
Expect(0, 56319, '\p{^Blk=highpusurrogates}', "");
Expect(0, 56319, '\P{Blk=highpusurrogates}', "");
Expect(1, 56319, '\P{^Blk=highpusurrogates}', "");
Expect(0, 57344, '\p{Blk=highpusurrogates}', "");
Expect(1, 57344, '\p{^Blk=highpusurrogates}', "");
Expect(1, 57344, '\P{Blk=highpusurrogates}', "");
Expect(0, 57344, '\P{^Blk=highpusurrogates}', "");
Expect(1, 56319, '\p{Blk= high_PU_surrogates}', "");
Expect(0, 56319, '\p{^Blk= high_PU_surrogates}', "");
Expect(0, 56319, '\P{Blk= high_PU_surrogates}', "");
Expect(1, 56319, '\P{^Blk= high_PU_surrogates}', "");
Expect(0, 57344, '\p{Blk= high_PU_surrogates}', "");
Expect(1, 57344, '\p{^Blk= high_PU_surrogates}', "");
Expect(1, 57344, '\P{Blk= high_PU_surrogates}', "");
Expect(0, 57344, '\P{^Blk= high_PU_surrogates}', "");
Error('\p{Is_Block= High_private_USE_Surrogates/a/}');
Error('\P{Is_Block= High_private_USE_Surrogates/a/}');
Expect(1, 56319, '\p{Is_Block=highprivateusesurrogates}', "");
Expect(0, 56319, '\p{^Is_Block=highprivateusesurrogates}', "");
Expect(0, 56319, '\P{Is_Block=highprivateusesurrogates}', "");
Expect(1, 56319, '\P{^Is_Block=highprivateusesurrogates}', "");
Expect(0, 57344, '\p{Is_Block=highprivateusesurrogates}', "");
Expect(1, 57344, '\p{^Is_Block=highprivateusesurrogates}', "");
Expect(1, 57344, '\P{Is_Block=highprivateusesurrogates}', "");
Expect(0, 57344, '\P{^Is_Block=highprivateusesurrogates}', "");
Expect(1, 56319, '\p{Is_Block=	-High_PRIVATE_Use_SURROGATES}', "");
Expect(0, 56319, '\p{^Is_Block=	-High_PRIVATE_Use_SURROGATES}', "");
Expect(0, 56319, '\P{Is_Block=	-High_PRIVATE_Use_SURROGATES}', "");
Expect(1, 56319, '\P{^Is_Block=	-High_PRIVATE_Use_SURROGATES}', "");
Expect(0, 57344, '\p{Is_Block=	-High_PRIVATE_Use_SURROGATES}', "");
Expect(1, 57344, '\p{^Is_Block=	-High_PRIVATE_Use_SURROGATES}', "");
Expect(1, 57344, '\P{Is_Block=	-High_PRIVATE_Use_SURROGATES}', "");
Expect(0, 57344, '\P{^Is_Block=	-High_PRIVATE_Use_SURROGATES}', "");
Error('\p{Is_Blk=:=	HIGH_pu_Surrogates}');
Error('\P{Is_Blk=:=	HIGH_pu_Surrogates}');
Expect(1, 56319, '\p{Is_Blk=highpusurrogates}', "");
Expect(0, 56319, '\p{^Is_Blk=highpusurrogates}', "");
Expect(0, 56319, '\P{Is_Blk=highpusurrogates}', "");
Expect(1, 56319, '\P{^Is_Blk=highpusurrogates}', "");
Expect(0, 57344, '\p{Is_Blk=highpusurrogates}', "");
Expect(1, 57344, '\p{^Is_Blk=highpusurrogates}', "");
Expect(1, 57344, '\P{Is_Blk=highpusurrogates}', "");
Expect(0, 57344, '\P{^Is_Blk=highpusurrogates}', "");
Expect(1, 56319, '\p{Is_Blk= High_PU_Surrogates}', "");
Expect(0, 56319, '\p{^Is_Blk= High_PU_Surrogates}', "");
Expect(0, 56319, '\P{Is_Blk= High_PU_Surrogates}', "");
Expect(1, 56319, '\P{^Is_Blk= High_PU_Surrogates}', "");
Expect(0, 57344, '\p{Is_Blk= High_PU_Surrogates}', "");
Expect(1, 57344, '\p{^Is_Blk= High_PU_Surrogates}', "");
Expect(1, 57344, '\P{Is_Blk= High_PU_Surrogates}', "");
Expect(0, 57344, '\P{^Is_Blk= High_PU_Surrogates}', "");
Error('\p{Block=_-PHOENICIAN:=}');
Error('\P{Block=_-PHOENICIAN:=}');
Expect(1, 67871, '\p{Block=phoenician}', "");
Expect(0, 67871, '\p{^Block=phoenician}', "");
Expect(0, 67871, '\P{Block=phoenician}', "");
Expect(1, 67871, '\P{^Block=phoenician}', "");
Expect(0, 67872, '\p{Block=phoenician}', "");
Expect(1, 67872, '\p{^Block=phoenician}', "");
Expect(1, 67872, '\P{Block=phoenician}', "");
Expect(0, 67872, '\P{^Block=phoenician}', "");
Expect(1, 67871, '\p{Block=-PHOENICIAN}', "");
Expect(0, 67871, '\p{^Block=-PHOENICIAN}', "");
Expect(0, 67871, '\P{Block=-PHOENICIAN}', "");
Expect(1, 67871, '\P{^Block=-PHOENICIAN}', "");
Expect(0, 67872, '\p{Block=-PHOENICIAN}', "");
Expect(1, 67872, '\p{^Block=-PHOENICIAN}', "");
Expect(1, 67872, '\P{Block=-PHOENICIAN}', "");
Expect(0, 67872, '\P{^Block=-PHOENICIAN}', "");
Error('\p{Blk=/a/_-Phoenician}');
Error('\P{Blk=/a/_-Phoenician}');
Expect(1, 67871, '\p{Blk=phoenician}', "");
Expect(0, 67871, '\p{^Blk=phoenician}', "");
Expect(0, 67871, '\P{Blk=phoenician}', "");
Expect(1, 67871, '\P{^Blk=phoenician}', "");
Expect(0, 67872, '\p{Blk=phoenician}', "");
Expect(1, 67872, '\p{^Blk=phoenician}', "");
Expect(1, 67872, '\P{Blk=phoenician}', "");
Expect(0, 67872, '\P{^Blk=phoenician}', "");
Expect(1, 67871, '\p{Blk=	_Phoenician}', "");
Expect(0, 67871, '\p{^Blk=	_Phoenician}', "");
Expect(0, 67871, '\P{Blk=	_Phoenician}', "");
Expect(1, 67871, '\P{^Blk=	_Phoenician}', "");
Expect(0, 67872, '\p{Blk=	_Phoenician}', "");
Expect(1, 67872, '\p{^Blk=	_Phoenician}', "");
Expect(1, 67872, '\P{Blk=	_Phoenician}', "");
Expect(0, 67872, '\P{^Blk=	_Phoenician}', "");
Error('\p{Is_Block= -PHOENICIAN:=}');
Error('\P{Is_Block= -PHOENICIAN:=}');
Expect(1, 67871, '\p{Is_Block=phoenician}', "");
Expect(0, 67871, '\p{^Is_Block=phoenician}', "");
Expect(0, 67871, '\P{Is_Block=phoenician}', "");
Expect(1, 67871, '\P{^Is_Block=phoenician}', "");
Expect(0, 67872, '\p{Is_Block=phoenician}', "");
Expect(1, 67872, '\p{^Is_Block=phoenician}', "");
Expect(1, 67872, '\P{Is_Block=phoenician}', "");
Expect(0, 67872, '\P{^Is_Block=phoenician}', "");
Expect(1, 67871, '\p{Is_Block=Phoenician}', "");
Expect(0, 67871, '\p{^Is_Block=Phoenician}', "");
Expect(0, 67871, '\P{Is_Block=Phoenician}', "");
Expect(1, 67871, '\P{^Is_Block=Phoenician}', "");
Expect(0, 67872, '\p{Is_Block=Phoenician}', "");
Expect(1, 67872, '\p{^Is_Block=Phoenician}', "");
Expect(1, 67872, '\P{Is_Block=Phoenician}', "");
Expect(0, 67872, '\P{^Is_Block=Phoenician}', "");
Error('\p{Is_Blk=-:=phoenician}');
Error('\P{Is_Blk=-:=phoenician}');
Expect(1, 67871, '\p{Is_Blk=phoenician}', "");
Expect(0, 67871, '\p{^Is_Blk=phoenician}', "");
Expect(0, 67871, '\P{Is_Blk=phoenician}', "");
Expect(1, 67871, '\P{^Is_Blk=phoenician}', "");
Expect(0, 67872, '\p{Is_Blk=phoenician}', "");
Expect(1, 67872, '\p{^Is_Blk=phoenician}', "");
Expect(1, 67872, '\P{Is_Blk=phoenician}', "");
Expect(0, 67872, '\P{^Is_Blk=phoenician}', "");
Expect(1, 67871, '\p{Is_Blk=	_PHOENICIAN}', "");
Expect(0, 67871, '\p{^Is_Blk=	_PHOENICIAN}', "");
Expect(0, 67871, '\P{Is_Blk=	_PHOENICIAN}', "");
Expect(1, 67871, '\P{^Is_Blk=	_PHOENICIAN}', "");
Expect(0, 67872, '\p{Is_Blk=	_PHOENICIAN}', "");
Expect(1, 67872, '\p{^Is_Blk=	_PHOENICIAN}', "");
Expect(1, 67872, '\P{Is_Blk=	_PHOENICIAN}', "");
Expect(0, 67872, '\P{^Is_Blk=	_PHOENICIAN}', "");
Error('\p{Block=/a/-Sharada}');
Error('\P{Block=/a/-Sharada}');
Expect(1, 70111, '\p{Block=sharada}', "");
Expect(0, 70111, '\p{^Block=sharada}', "");
Expect(0, 70111, '\P{Block=sharada}', "");
Expect(1, 70111, '\P{^Block=sharada}', "");
Expect(0, 70112, '\p{Block=sharada}', "");
Expect(1, 70112, '\p{^Block=sharada}', "");
Expect(1, 70112, '\P{Block=sharada}', "");
Expect(0, 70112, '\P{^Block=sharada}', "");
Expect(1, 70111, '\p{Block=_Sharada}', "");
Expect(0, 70111, '\p{^Block=_Sharada}', "");
Expect(0, 70111, '\P{Block=_Sharada}', "");
Expect(1, 70111, '\P{^Block=_Sharada}', "");
Expect(0, 70112, '\p{Block=_Sharada}', "");
Expect(1, 70112, '\p{^Block=_Sharada}', "");
Expect(1, 70112, '\P{Block=_Sharada}', "");
Expect(0, 70112, '\P{^Block=_Sharada}', "");
Error('\p{Blk=/a/-Sharada}');
Error('\P{Blk=/a/-Sharada}');
Expect(1, 70111, '\p{Blk=sharada}', "");
Expect(0, 70111, '\p{^Blk=sharada}', "");
Expect(0, 70111, '\P{Blk=sharada}', "");
Expect(1, 70111, '\P{^Blk=sharada}', "");
Expect(0, 70112, '\p{Blk=sharada}', "");
Expect(1, 70112, '\p{^Blk=sharada}', "");
Expect(1, 70112, '\P{Blk=sharada}', "");
Expect(0, 70112, '\P{^Blk=sharada}', "");
Expect(1, 70111, '\p{Blk=--SHARADA}', "");
Expect(0, 70111, '\p{^Blk=--SHARADA}', "");
Expect(0, 70111, '\P{Blk=--SHARADA}', "");
Expect(1, 70111, '\P{^Blk=--SHARADA}', "");
Expect(0, 70112, '\p{Blk=--SHARADA}', "");
Expect(1, 70112, '\p{^Blk=--SHARADA}', "");
Expect(1, 70112, '\P{Blk=--SHARADA}', "");
Expect(0, 70112, '\P{^Blk=--SHARADA}', "");
Error('\p{Is_Block=	Sharada/a/}');
Error('\P{Is_Block=	Sharada/a/}');
Expect(1, 70111, '\p{Is_Block=sharada}', "");
Expect(0, 70111, '\p{^Is_Block=sharada}', "");
Expect(0, 70111, '\P{Is_Block=sharada}', "");
Expect(1, 70111, '\P{^Is_Block=sharada}', "");
Expect(0, 70112, '\p{Is_Block=sharada}', "");
Expect(1, 70112, '\p{^Is_Block=sharada}', "");
Expect(1, 70112, '\P{Is_Block=sharada}', "");
Expect(0, 70112, '\P{^Is_Block=sharada}', "");
Expect(1, 70111, '\p{Is_Block=	SHARADA}', "");
Expect(0, 70111, '\p{^Is_Block=	SHARADA}', "");
Expect(0, 70111, '\P{Is_Block=	SHARADA}', "");
Expect(1, 70111, '\P{^Is_Block=	SHARADA}', "");
Expect(0, 70112, '\p{Is_Block=	SHARADA}', "");
Expect(1, 70112, '\p{^Is_Block=	SHARADA}', "");
Expect(1, 70112, '\P{Is_Block=	SHARADA}', "");
Expect(0, 70112, '\P{^Is_Block=	SHARADA}', "");
Error('\p{Is_Blk=  Sharada:=}');
Error('\P{Is_Blk=  Sharada:=}');
Expect(1, 70111, '\p{Is_Blk=sharada}', "");
Expect(0, 70111, '\p{^Is_Blk=sharada}', "");
Expect(0, 70111, '\P{Is_Blk=sharada}', "");
Expect(1, 70111, '\P{^Is_Blk=sharada}', "");
Expect(0, 70112, '\p{Is_Blk=sharada}', "");
Expect(1, 70112, '\p{^Is_Blk=sharada}', "");
Expect(1, 70112, '\P{Is_Blk=sharada}', "");
Expect(0, 70112, '\P{^Is_Blk=sharada}', "");
Expect(1, 70111, '\p{Is_Blk:   --sharada}', "");
Expect(0, 70111, '\p{^Is_Blk:   --sharada}', "");
Expect(0, 70111, '\P{Is_Blk:   --sharada}', "");
Expect(1, 70111, '\P{^Is_Blk:   --sharada}', "");
Expect(0, 70112, '\p{Is_Blk:   --sharada}', "");
Expect(1, 70112, '\p{^Is_Blk:   --sharada}', "");
Expect(1, 70112, '\P{Is_Blk:   --sharada}', "");
Expect(0, 70112, '\P{^Is_Blk:   --sharada}', "");
Error('\p{Block=-:=Old_Italic}');
Error('\P{Block=-:=Old_Italic}');
Expect(1, 66351, '\p{Block=olditalic}', "");
Expect(0, 66351, '\p{^Block=olditalic}', "");
Expect(0, 66351, '\P{Block=olditalic}', "");
Expect(1, 66351, '\P{^Block=olditalic}', "");
Expect(0, 66352, '\p{Block=olditalic}', "");
Expect(1, 66352, '\p{^Block=olditalic}', "");
Expect(1, 66352, '\P{Block=olditalic}', "");
Expect(0, 66352, '\P{^Block=olditalic}', "");
Expect(1, 66351, '\p{Block=		Old_Italic}', "");
Expect(0, 66351, '\p{^Block=		Old_Italic}', "");
Expect(0, 66351, '\P{Block=		Old_Italic}', "");
Expect(1, 66351, '\P{^Block=		Old_Italic}', "");
Expect(0, 66352, '\p{Block=		Old_Italic}', "");
Expect(1, 66352, '\p{^Block=		Old_Italic}', "");
Expect(1, 66352, '\P{Block=		Old_Italic}', "");
Expect(0, 66352, '\P{^Block=		Old_Italic}', "");
Error('\p{Blk= /a/Old_ITALIC}');
Error('\P{Blk= /a/Old_ITALIC}');
Expect(1, 66351, '\p{Blk=olditalic}', "");
Expect(0, 66351, '\p{^Blk=olditalic}', "");
Expect(0, 66351, '\P{Blk=olditalic}', "");
Expect(1, 66351, '\P{^Blk=olditalic}', "");
Expect(0, 66352, '\p{Blk=olditalic}', "");
Expect(1, 66352, '\p{^Blk=olditalic}', "");
Expect(1, 66352, '\P{Blk=olditalic}', "");
Expect(0, 66352, '\P{^Blk=olditalic}', "");
Expect(1, 66351, '\p{Blk=-	OLD_Italic}', "");
Expect(0, 66351, '\p{^Blk=-	OLD_Italic}', "");
Expect(0, 66351, '\P{Blk=-	OLD_Italic}', "");
Expect(1, 66351, '\P{^Blk=-	OLD_Italic}', "");
Expect(0, 66352, '\p{Blk=-	OLD_Italic}', "");
Expect(1, 66352, '\p{^Blk=-	OLD_Italic}', "");
Expect(1, 66352, '\P{Blk=-	OLD_Italic}', "");
Expect(0, 66352, '\P{^Blk=-	OLD_Italic}', "");
Error('\p{Is_Block=_Old_Italic/a/}');
Error('\P{Is_Block=_Old_Italic/a/}');
Expect(1, 66351, '\p{Is_Block:	olditalic}', "");
Expect(0, 66351, '\p{^Is_Block:	olditalic}', "");
Expect(0, 66351, '\P{Is_Block:	olditalic}', "");
Expect(1, 66351, '\P{^Is_Block:	olditalic}', "");
Expect(0, 66352, '\p{Is_Block:	olditalic}', "");
Expect(1, 66352, '\p{^Is_Block:	olditalic}', "");
Expect(1, 66352, '\P{Is_Block:	olditalic}', "");
Expect(0, 66352, '\P{^Is_Block:	olditalic}', "");
Expect(1, 66351, '\p{Is_Block= -Old_ITALIC}', "");
Expect(0, 66351, '\p{^Is_Block= -Old_ITALIC}', "");
Expect(0, 66351, '\P{Is_Block= -Old_ITALIC}', "");
Expect(1, 66351, '\P{^Is_Block= -Old_ITALIC}', "");
Expect(0, 66352, '\p{Is_Block= -Old_ITALIC}', "");
Expect(1, 66352, '\p{^Is_Block= -Old_ITALIC}', "");
Expect(1, 66352, '\P{Is_Block= -Old_ITALIC}', "");
Expect(0, 66352, '\P{^Is_Block= -Old_ITALIC}', "");
Error('\p{Is_Blk=:=_OLD_ITALIC}');
Error('\P{Is_Blk=:=_OLD_ITALIC}');
Expect(1, 66351, '\p{Is_Blk=olditalic}', "");
Expect(0, 66351, '\p{^Is_Blk=olditalic}', "");
Expect(0, 66351, '\P{Is_Blk=olditalic}', "");
Expect(1, 66351, '\P{^Is_Blk=olditalic}', "");
Expect(0, 66352, '\p{Is_Blk=olditalic}', "");
Expect(1, 66352, '\p{^Is_Blk=olditalic}', "");
Expect(1, 66352, '\P{Is_Blk=olditalic}', "");
Expect(0, 66352, '\P{^Is_Blk=olditalic}', "");
Expect(1, 66351, '\p{Is_Blk=	-OLD_Italic}', "");
Expect(0, 66351, '\p{^Is_Blk=	-OLD_Italic}', "");
Expect(0, 66351, '\P{Is_Blk=	-OLD_Italic}', "");
Expect(1, 66351, '\P{^Is_Blk=	-OLD_Italic}', "");
Expect(0, 66352, '\p{Is_Blk=	-OLD_Italic}', "");
Expect(1, 66352, '\p{^Is_Blk=	-OLD_Italic}', "");
Expect(1, 66352, '\P{Is_Blk=	-OLD_Italic}', "");
Expect(0, 66352, '\P{^Is_Blk=	-OLD_Italic}', "");
Error('\p{Block= /a/Gothic}');
Error('\P{Block= /a/Gothic}');
Expect(1, 66383, '\p{Block=gothic}', "");
Expect(0, 66383, '\p{^Block=gothic}', "");
Expect(0, 66383, '\P{Block=gothic}', "");
Expect(1, 66383, '\P{^Block=gothic}', "");
Expect(0, 66384, '\p{Block=gothic}', "");
Expect(1, 66384, '\p{^Block=gothic}', "");
Expect(1, 66384, '\P{Block=gothic}', "");
Expect(0, 66384, '\P{^Block=gothic}', "");
Expect(1, 66383, '\p{Block=_Gothic}', "");
Expect(0, 66383, '\p{^Block=_Gothic}', "");
Expect(0, 66383, '\P{Block=_Gothic}', "");
Expect(1, 66383, '\P{^Block=_Gothic}', "");
Expect(0, 66384, '\p{Block=_Gothic}', "");
Expect(1, 66384, '\p{^Block=_Gothic}', "");
Expect(1, 66384, '\P{Block=_Gothic}', "");
Expect(0, 66384, '\P{^Block=_Gothic}', "");
Error('\p{Blk=	/a/GOTHIC}');
Error('\P{Blk=	/a/GOTHIC}');
Expect(1, 66383, '\p{Blk=gothic}', "");
Expect(0, 66383, '\p{^Blk=gothic}', "");
Expect(0, 66383, '\P{Blk=gothic}', "");
Expect(1, 66383, '\P{^Blk=gothic}', "");
Expect(0, 66384, '\p{Blk=gothic}', "");
Expect(1, 66384, '\p{^Blk=gothic}', "");
Expect(1, 66384, '\P{Blk=gothic}', "");
Expect(0, 66384, '\P{^Blk=gothic}', "");
Expect(1, 66383, '\p{Blk= -GOTHIC}', "");
Expect(0, 66383, '\p{^Blk= -GOTHIC}', "");
Expect(0, 66383, '\P{Blk= -GOTHIC}', "");
Expect(1, 66383, '\P{^Blk= -GOTHIC}', "");
Expect(0, 66384, '\p{Blk= -GOTHIC}', "");
Expect(1, 66384, '\p{^Blk= -GOTHIC}', "");
Expect(1, 66384, '\P{Blk= -GOTHIC}', "");
Expect(0, 66384, '\P{^Blk= -GOTHIC}', "");
Error('\p{Is_Block=__Gothic/a/}');
Error('\P{Is_Block=__Gothic/a/}');
Expect(1, 66383, '\p{Is_Block:	gothic}', "");
Expect(0, 66383, '\p{^Is_Block:	gothic}', "");
Expect(0, 66383, '\P{Is_Block:	gothic}', "");
Expect(1, 66383, '\P{^Is_Block:	gothic}', "");
Expect(0, 66384, '\p{Is_Block:	gothic}', "");
Expect(1, 66384, '\p{^Is_Block:	gothic}', "");
Expect(1, 66384, '\P{Is_Block:	gothic}', "");
Expect(0, 66384, '\P{^Is_Block:	gothic}', "");
Expect(1, 66383, '\p{Is_Block=		GOTHIC}', "");
Expect(0, 66383, '\p{^Is_Block=		GOTHIC}', "");
Expect(0, 66383, '\P{Is_Block=		GOTHIC}', "");
Expect(1, 66383, '\P{^Is_Block=		GOTHIC}', "");
Expect(0, 66384, '\p{Is_Block=		GOTHIC}', "");
Expect(1, 66384, '\p{^Is_Block=		GOTHIC}', "");
Expect(1, 66384, '\P{Is_Block=		GOTHIC}', "");
Expect(0, 66384, '\P{^Is_Block=		GOTHIC}', "");
Error('\p{Is_Blk=/a/-	gothic}');
Error('\P{Is_Blk=/a/-	gothic}');
Expect(1, 66383, '\p{Is_Blk:gothic}', "");
Expect(0, 66383, '\p{^Is_Blk:gothic}', "");
Expect(0, 66383, '\P{Is_Blk:gothic}', "");
Expect(1, 66383, '\P{^Is_Blk:gothic}', "");
Expect(0, 66384, '\p{Is_Blk:gothic}', "");
Expect(1, 66384, '\p{^Is_Blk:gothic}', "");
Expect(1, 66384, '\P{Is_Blk:gothic}', "");
Expect(0, 66384, '\P{^Is_Blk:gothic}', "");
Expect(1, 66383, '\p{Is_Blk=	-Gothic}', "");
Expect(0, 66383, '\p{^Is_Blk=	-Gothic}', "");
Expect(0, 66383, '\P{Is_Blk=	-Gothic}', "");
Expect(1, 66383, '\P{^Is_Blk=	-Gothic}', "");
Expect(0, 66384, '\p{Is_Blk=	-Gothic}', "");
Expect(1, 66384, '\p{^Is_Blk=	-Gothic}', "");
Expect(1, 66384, '\P{Is_Blk=	-Gothic}', "");
Expect(0, 66384, '\P{^Is_Blk=	-Gothic}', "");
Error('\p{Block=	inscriptional_Pahlavi/a/}');
Error('\P{Block=	inscriptional_Pahlavi/a/}');
Expect(1, 68479, '\p{Block=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Block=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Block=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Block=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Block=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Block=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Block=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Block=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Block=__INSCRIPTIONAL_pahlavi}', "");
Expect(0, 68479, '\p{^Block=__INSCRIPTIONAL_pahlavi}', "");
Expect(0, 68479, '\P{Block=__INSCRIPTIONAL_pahlavi}', "");
Expect(1, 68479, '\P{^Block=__INSCRIPTIONAL_pahlavi}', "");
Expect(0, 68480, '\p{Block=__INSCRIPTIONAL_pahlavi}', "");
Expect(1, 68480, '\p{^Block=__INSCRIPTIONAL_pahlavi}', "");
Expect(1, 68480, '\P{Block=__INSCRIPTIONAL_pahlavi}', "");
Expect(0, 68480, '\P{^Block=__INSCRIPTIONAL_pahlavi}', "");
Error('\p{Blk=:=--Inscriptional_Pahlavi}');
Error('\P{Blk=:=--Inscriptional_Pahlavi}');
Expect(1, 68479, '\p{Blk=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Blk=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Blk=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Blk=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Blk=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Blk=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Blk=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Blk=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Blk=_-Inscriptional_PAHLAVI}', "");
Expect(0, 68479, '\p{^Blk=_-Inscriptional_PAHLAVI}', "");
Expect(0, 68479, '\P{Blk=_-Inscriptional_PAHLAVI}', "");
Expect(1, 68479, '\P{^Blk=_-Inscriptional_PAHLAVI}', "");
Expect(0, 68480, '\p{Blk=_-Inscriptional_PAHLAVI}', "");
Expect(1, 68480, '\p{^Blk=_-Inscriptional_PAHLAVI}', "");
Expect(1, 68480, '\P{Blk=_-Inscriptional_PAHLAVI}', "");
Expect(0, 68480, '\P{^Blk=_-Inscriptional_PAHLAVI}', "");
Error('\p{Is_Block=/a/		Inscriptional_Pahlavi}');
Error('\P{Is_Block=/a/		Inscriptional_Pahlavi}');
Expect(1, 68479, '\p{Is_Block=inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Is_Block=inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Is_Block=inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Is_Block=inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Is_Block=inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Is_Block=inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Is_Block=inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Is_Block=inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Is_Block= -INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68479, '\p{^Is_Block= -INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68479, '\P{Is_Block= -INSCRIPTIONAL_Pahlavi}', "");
Expect(1, 68479, '\P{^Is_Block= -INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68480, '\p{Is_Block= -INSCRIPTIONAL_Pahlavi}', "");
Expect(1, 68480, '\p{^Is_Block= -INSCRIPTIONAL_Pahlavi}', "");
Expect(1, 68480, '\P{Is_Block= -INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68480, '\P{^Is_Block= -INSCRIPTIONAL_Pahlavi}', "");
Error('\p{Is_Blk=	:=INSCRIPTIONAL_Pahlavi}');
Error('\P{Is_Blk=	:=INSCRIPTIONAL_Pahlavi}');
Expect(1, 68479, '\p{Is_Blk:   inscriptionalpahlavi}', "");
Expect(0, 68479, '\p{^Is_Blk:   inscriptionalpahlavi}', "");
Expect(0, 68479, '\P{Is_Blk:   inscriptionalpahlavi}', "");
Expect(1, 68479, '\P{^Is_Blk:   inscriptionalpahlavi}', "");
Expect(0, 68480, '\p{Is_Blk:   inscriptionalpahlavi}', "");
Expect(1, 68480, '\p{^Is_Blk:   inscriptionalpahlavi}', "");
Expect(1, 68480, '\P{Is_Blk:   inscriptionalpahlavi}', "");
Expect(0, 68480, '\P{^Is_Blk:   inscriptionalpahlavi}', "");
Expect(1, 68479, '\p{Is_Blk=	 INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68479, '\p{^Is_Blk=	 INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68479, '\P{Is_Blk=	 INSCRIPTIONAL_Pahlavi}', "");
Expect(1, 68479, '\P{^Is_Blk=	 INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68480, '\p{Is_Blk=	 INSCRIPTIONAL_Pahlavi}', "");
Expect(1, 68480, '\p{^Is_Blk=	 INSCRIPTIONAL_Pahlavi}', "");
Expect(1, 68480, '\P{Is_Blk=	 INSCRIPTIONAL_Pahlavi}', "");
Expect(0, 68480, '\P{^Is_Blk=	 INSCRIPTIONAL_Pahlavi}', "");
Error('\p{Block=:= tai_THAM}');
Error('\P{Block=:= tai_THAM}');
Expect(1, 6831, '\p{Block=taitham}', "");
Expect(0, 6831, '\p{^Block=taitham}', "");
Expect(0, 6831, '\P{Block=taitham}', "");
Expect(1, 6831, '\P{^Block=taitham}', "");
Expect(0, 6832, '\p{Block=taitham}', "");
Expect(1, 6832, '\p{^Block=taitham}', "");
Expect(1, 6832, '\P{Block=taitham}', "");
Expect(0, 6832, '\P{^Block=taitham}', "");
Expect(1, 6831, '\p{Block: _	Tai_Tham}', "");
Expect(0, 6831, '\p{^Block: _	Tai_Tham}', "");
Expect(0, 6831, '\P{Block: _	Tai_Tham}', "");
Expect(1, 6831, '\P{^Block: _	Tai_Tham}', "");
Expect(0, 6832, '\p{Block: _	Tai_Tham}', "");
Expect(1, 6832, '\p{^Block: _	Tai_Tham}', "");
Expect(1, 6832, '\P{Block: _	Tai_Tham}', "");
Expect(0, 6832, '\P{^Block: _	Tai_Tham}', "");
Error('\p{Blk=-:=Tai_Tham}');
Error('\P{Blk=-:=Tai_Tham}');
Expect(1, 6831, '\p{Blk=taitham}', "");
Expect(0, 6831, '\p{^Blk=taitham}', "");
Expect(0, 6831, '\P{Blk=taitham}', "");
Expect(1, 6831, '\P{^Blk=taitham}', "");
Expect(0, 6832, '\p{Blk=taitham}', "");
Expect(1, 6832, '\p{^Blk=taitham}', "");
Expect(1, 6832, '\P{Blk=taitham}', "");
Expect(0, 6832, '\P{^Blk=taitham}', "");
Expect(1, 6831, '\p{Blk=__Tai_Tham}', "");
Expect(0, 6831, '\p{^Blk=__Tai_Tham}', "");
Expect(0, 6831, '\P{Blk=__Tai_Tham}', "");
Expect(1, 6831, '\P{^Blk=__Tai_Tham}', "");
Expect(0, 6832, '\p{Blk=__Tai_Tham}', "");
Expect(1, 6832, '\p{^Blk=__Tai_Tham}', "");
Expect(1, 6832, '\P{Blk=__Tai_Tham}', "");
Expect(0, 6832, '\P{^Blk=__Tai_Tham}', "");
Error('\p{Is_Block=-_Tai_Tham/a/}');
Error('\P{Is_Block=-_Tai_Tham/a/}');
Expect(1, 6831, '\p{Is_Block=taitham}', "");
Expect(0, 6831, '\p{^Is_Block=taitham}', "");
Expect(0, 6831, '\P{Is_Block=taitham}', "");
Expect(1, 6831, '\P{^Is_Block=taitham}', "");
Expect(0, 6832, '\p{Is_Block=taitham}', "");
Expect(1, 6832, '\p{^Is_Block=taitham}', "");
Expect(1, 6832, '\P{Is_Block=taitham}', "");
Expect(0, 6832, '\P{^Is_Block=taitham}', "");
Expect(1, 6831, '\p{Is_Block=Tai_THAM}', "");
Expect(0, 6831, '\p{^Is_Block=Tai_THAM}', "");
Expect(0, 6831, '\P{Is_Block=Tai_THAM}', "");
Expect(1, 6831, '\P{^Is_Block=Tai_THAM}', "");
Expect(0, 6832, '\p{Is_Block=Tai_THAM}', "");
Expect(1, 6832, '\p{^Is_Block=Tai_THAM}', "");
Expect(1, 6832, '\P{Is_Block=Tai_THAM}', "");
Expect(0, 6832, '\P{^Is_Block=Tai_THAM}', "");
Error('\p{Is_Blk=:=	Tai_Tham}');
Error('\P{Is_Blk=:=	Tai_Tham}');
Expect(1, 6831, '\p{Is_Blk=taitham}', "");
Expect(0, 6831, '\p{^Is_Blk=taitham}', "");
Expect(0, 6831, '\P{Is_Blk=taitham}', "");
Expect(1, 6831, '\P{^Is_Blk=taitham}', "");
Expect(0, 6832, '\p{Is_Blk=taitham}', "");
Expect(1, 6832, '\p{^Is_Blk=taitham}', "");
Expect(1, 6832, '\P{Is_Blk=taitham}', "");
Expect(0, 6832, '\P{^Is_Blk=taitham}', "");
Expect(1, 6831, '\p{Is_Blk=	_TAI_tham}', "");
Expect(0, 6831, '\p{^Is_Blk=	_TAI_tham}', "");
Expect(0, 6831, '\P{Is_Blk=	_TAI_tham}', "");
Expect(1, 6831, '\P{^Is_Blk=	_TAI_tham}', "");
Expect(0, 6832, '\p{Is_Blk=	_TAI_tham}', "");
Expect(1, 6832, '\p{^Is_Blk=	_TAI_tham}', "");
Expect(1, 6832, '\P{Is_Blk=	_TAI_tham}', "");
Expect(0, 6832, '\P{^Is_Blk=	_TAI_tham}', "");
Error('\p{Block=:= 	tagbanwa}');
Error('\P{Block=:= 	tagbanwa}');
Expect(1, 6015, '\p{Block=tagbanwa}', "");
Expect(0, 6015, '\p{^Block=tagbanwa}', "");
Expect(0, 6015, '\P{Block=tagbanwa}', "");
Expect(1, 6015, '\P{^Block=tagbanwa}', "");
Expect(0, 6016, '\p{Block=tagbanwa}', "");
Expect(1, 6016, '\p{^Block=tagbanwa}', "");
Expect(1, 6016, '\P{Block=tagbanwa}', "");
Expect(0, 6016, '\P{^Block=tagbanwa}', "");
Expect(1, 6015, '\p{Block=_ TAGBANWA}', "");
Expect(0, 6015, '\p{^Block=_ TAGBANWA}', "");
Expect(0, 6015, '\P{Block=_ TAGBANWA}', "");
Expect(1, 6015, '\P{^Block=_ TAGBANWA}', "");
Expect(0, 6016, '\p{Block=_ TAGBANWA}', "");
Expect(1, 6016, '\p{^Block=_ TAGBANWA}', "");
Expect(1, 6016, '\P{Block=_ TAGBANWA}', "");
Expect(0, 6016, '\P{^Block=_ TAGBANWA}', "");
Error('\p{Blk=-	TAGBANWA/a/}');
Error('\P{Blk=-	TAGBANWA/a/}');
Expect(1, 6015, '\p{Blk=tagbanwa}', "");
Expect(0, 6015, '\p{^Blk=tagbanwa}', "");
Expect(0, 6015, '\P{Blk=tagbanwa}', "");
Expect(1, 6015, '\P{^Blk=tagbanwa}', "");
Expect(0, 6016, '\p{Blk=tagbanwa}', "");
Expect(1, 6016, '\p{^Blk=tagbanwa}', "");
Expect(1, 6016, '\P{Blk=tagbanwa}', "");
Expect(0, 6016, '\P{^Blk=tagbanwa}', "");
Expect(1, 6015, '\p{Blk=_	Tagbanwa}', "");
Expect(0, 6015, '\p{^Blk=_	Tagbanwa}', "");
Expect(0, 6015, '\P{Blk=_	Tagbanwa}', "");
Expect(1, 6015, '\P{^Blk=_	Tagbanwa}', "");
Expect(0, 6016, '\p{Blk=_	Tagbanwa}', "");
Expect(1, 6016, '\p{^Blk=_	Tagbanwa}', "");
Expect(1, 6016, '\P{Blk=_	Tagbanwa}', "");
Expect(0, 6016, '\P{^Blk=_	Tagbanwa}', "");
Error('\p{Is_Block= _Tagbanwa/a/}');
Error('\P{Is_Block= _Tagbanwa/a/}');
Expect(1, 6015, '\p{Is_Block=tagbanwa}', "");
Expect(0, 6015, '\p{^Is_Block=tagbanwa}', "");
Expect(0, 6015, '\P{Is_Block=tagbanwa}', "");
Expect(1, 6015, '\P{^Is_Block=tagbanwa}', "");
Expect(0, 6016, '\p{Is_Block=tagbanwa}', "");
Expect(1, 6016, '\p{^Is_Block=tagbanwa}', "");
Expect(1, 6016, '\P{Is_Block=tagbanwa}', "");
Expect(0, 6016, '\P{^Is_Block=tagbanwa}', "");
Expect(1, 6015, '\p{Is_Block=--tagbanwa}', "");
Expect(0, 6015, '\p{^Is_Block=--tagbanwa}', "");
Expect(0, 6015, '\P{Is_Block=--tagbanwa}', "");
Expect(1, 6015, '\P{^Is_Block=--tagbanwa}', "");
Expect(0, 6016, '\p{Is_Block=--tagbanwa}', "");
Expect(1, 6016, '\p{^Is_Block=--tagbanwa}', "");
Expect(1, 6016, '\P{Is_Block=--tagbanwa}', "");
Expect(0, 6016, '\P{^Is_Block=--tagbanwa}', "");
Error('\p{Is_Blk:	/a/ Tagbanwa}');
Error('\P{Is_Blk:	/a/ Tagbanwa}');
Expect(1, 6015, '\p{Is_Blk=tagbanwa}', "");
Expect(0, 6015, '\p{^Is_Blk=tagbanwa}', "");
Expect(0, 6015, '\P{Is_Blk=tagbanwa}', "");
Expect(1, 6015, '\P{^Is_Blk=tagbanwa}', "");
Expect(0, 6016, '\p{Is_Blk=tagbanwa}', "");
Expect(1, 6016, '\p{^Is_Blk=tagbanwa}', "");
Expect(1, 6016, '\P{Is_Blk=tagbanwa}', "");
Expect(0, 6016, '\P{^Is_Blk=tagbanwa}', "");
Expect(1, 6015, '\p{Is_Blk=	_tagbanwa}', "");
Expect(0, 6015, '\p{^Is_Blk=	_tagbanwa}', "");
Expect(0, 6015, '\P{Is_Blk=	_tagbanwa}', "");
Expect(1, 6015, '\P{^Is_Blk=	_tagbanwa}', "");
Expect(0, 6016, '\p{Is_Blk=	_tagbanwa}', "");
Expect(1, 6016, '\p{^Is_Blk=	_tagbanwa}', "");
Expect(1, 6016, '\P{Is_Blk=	_tagbanwa}', "");
Expect(0, 6016, '\P{^Is_Blk=	_tagbanwa}', "");
Error('\p{Block=:=-LAO}');
Error('\P{Block=:=-LAO}');
Expect(1, 3839, '\p{Block=lao}', "");
Expect(0, 3839, '\p{^Block=lao}', "");
Expect(0, 3839, '\P{Block=lao}', "");
Expect(1, 3839, '\P{^Block=lao}', "");
Expect(0, 3840, '\p{Block=lao}', "");
Expect(1, 3840, '\p{^Block=lao}', "");
Expect(1, 3840, '\P{Block=lao}', "");
Expect(0, 3840, '\P{^Block=lao}', "");
Expect(1, 3839, '\p{Block= 	Lao}', "");
Expect(0, 3839, '\p{^Block= 	Lao}', "");
Expect(0, 3839, '\P{Block= 	Lao}', "");
Expect(1, 3839, '\P{^Block= 	Lao}', "");
Expect(0, 3840, '\p{Block= 	Lao}', "");
Expect(1, 3840, '\p{^Block= 	Lao}', "");
Expect(1, 3840, '\P{Block= 	Lao}', "");
Expect(0, 3840, '\P{^Block= 	Lao}', "");
Error('\p{Blk=-/a/LAO}');
Error('\P{Blk=-/a/LAO}');
Expect(1, 3839, '\p{Blk=lao}', "");
Expect(0, 3839, '\p{^Blk=lao}', "");
Expect(0, 3839, '\P{Blk=lao}', "");
Expect(1, 3839, '\P{^Blk=lao}', "");
Expect(0, 3840, '\p{Blk=lao}', "");
Expect(1, 3840, '\p{^Blk=lao}', "");
Expect(1, 3840, '\P{Blk=lao}', "");
Expect(0, 3840, '\P{^Blk=lao}', "");
Expect(1, 3839, '\p{Blk=	Lao}', "");
Expect(0, 3839, '\p{^Blk=	Lao}', "");
Expect(0, 3839, '\P{Blk=	Lao}', "");
Expect(1, 3839, '\P{^Blk=	Lao}', "");
Expect(0, 3840, '\p{Blk=	Lao}', "");
Expect(1, 3840, '\p{^Blk=	Lao}', "");
Expect(1, 3840, '\P{Blk=	Lao}', "");
Expect(0, 3840, '\P{^Blk=	Lao}', "");
Error('\p{Is_Block= _Lao:=}');
Error('\P{Is_Block= _Lao:=}');
Expect(1, 3839, '\p{Is_Block=lao}', "");
Expect(0, 3839, '\p{^Is_Block=lao}', "");
Expect(0, 3839, '\P{Is_Block=lao}', "");
Expect(1, 3839, '\P{^Is_Block=lao}', "");
Expect(0, 3840, '\p{Is_Block=lao}', "");
Expect(1, 3840, '\p{^Is_Block=lao}', "");
Expect(1, 3840, '\P{Is_Block=lao}', "");
Expect(0, 3840, '\P{^Is_Block=lao}', "");
Expect(1, 3839, '\p{Is_Block= lao}', "");
Expect(0, 3839, '\p{^Is_Block= lao}', "");
Expect(0, 3839, '\P{Is_Block= lao}', "");
Expect(1, 3839, '\P{^Is_Block= lao}', "");
Expect(0, 3840, '\p{Is_Block= lao}', "");
Expect(1, 3840, '\p{^Is_Block= lao}', "");
Expect(1, 3840, '\P{Is_Block= lao}', "");
Expect(0, 3840, '\P{^Is_Block= lao}', "");
Error('\p{Is_Blk=_-lao:=}');
Error('\P{Is_Blk=_-lao:=}');
Expect(1, 3839, '\p{Is_Blk=lao}', "");
Expect(0, 3839, '\p{^Is_Blk=lao}', "");
Expect(0, 3839, '\P{Is_Blk=lao}', "");
Expect(1, 3839, '\P{^Is_Blk=lao}', "");
Expect(0, 3840, '\p{Is_Blk=lao}', "");
Expect(1, 3840, '\p{^Is_Blk=lao}', "");
Expect(1, 3840, '\P{Is_Blk=lao}', "");
Expect(0, 3840, '\P{^Is_Blk=lao}', "");
Expect(1, 3839, '\p{Is_Blk=_ Lao}', "");
Expect(0, 3839, '\p{^Is_Blk=_ Lao}', "");
Expect(0, 3839, '\P{Is_Blk=_ Lao}', "");
Expect(1, 3839, '\P{^Is_Blk=_ Lao}', "");
Expect(0, 3840, '\p{Is_Blk=_ Lao}', "");
Expect(1, 3840, '\p{^Is_Blk=_ Lao}', "");
Expect(1, 3840, '\P{Is_Blk=_ Lao}', "");
Expect(0, 3840, '\P{^Is_Blk=_ Lao}', "");
Error('\p{Block=	/a/Syloti_nagri}');
Error('\P{Block=	/a/Syloti_nagri}');
Expect(1, 43055, '\p{Block=sylotinagri}', "");
Expect(0, 43055, '\p{^Block=sylotinagri}', "");
Expect(0, 43055, '\P{Block=sylotinagri}', "");
Expect(1, 43055, '\P{^Block=sylotinagri}', "");
Expect(0, 43056, '\p{Block=sylotinagri}', "");
Expect(1, 43056, '\p{^Block=sylotinagri}', "");
Expect(1, 43056, '\P{Block=sylotinagri}', "");
Expect(0, 43056, '\P{^Block=sylotinagri}', "");
Expect(1, 43055, '\p{Block=  syloti_NAGRI}', "");
Expect(0, 43055, '\p{^Block=  syloti_NAGRI}', "");
Expect(0, 43055, '\P{Block=  syloti_NAGRI}', "");
Expect(1, 43055, '\P{^Block=  syloti_NAGRI}', "");
Expect(0, 43056, '\p{Block=  syloti_NAGRI}', "");
Expect(1, 43056, '\p{^Block=  syloti_NAGRI}', "");
Expect(1, 43056, '\P{Block=  syloti_NAGRI}', "");
Expect(0, 43056, '\P{^Block=  syloti_NAGRI}', "");
Error('\p{Blk=/a/ syloti_nagri}');
Error('\P{Blk=/a/ syloti_nagri}');
Expect(1, 43055, '\p{Blk=sylotinagri}', "");
Expect(0, 43055, '\p{^Blk=sylotinagri}', "");
Expect(0, 43055, '\P{Blk=sylotinagri}', "");
Expect(1, 43055, '\P{^Blk=sylotinagri}', "");
Expect(0, 43056, '\p{Blk=sylotinagri}', "");
Expect(1, 43056, '\p{^Blk=sylotinagri}', "");
Expect(1, 43056, '\P{Blk=sylotinagri}', "");
Expect(0, 43056, '\P{^Blk=sylotinagri}', "");
Expect(1, 43055, '\p{Blk= Syloti_NAGRI}', "");
Expect(0, 43055, '\p{^Blk= Syloti_NAGRI}', "");
Expect(0, 43055, '\P{Blk= Syloti_NAGRI}', "");
Expect(1, 43055, '\P{^Blk= Syloti_NAGRI}', "");
Expect(0, 43056, '\p{Blk= Syloti_NAGRI}', "");
Expect(1, 43056, '\p{^Blk= Syloti_NAGRI}', "");
Expect(1, 43056, '\P{Blk= Syloti_NAGRI}', "");
Expect(0, 43056, '\P{^Blk= Syloti_NAGRI}', "");
Error('\p{Is_Block=:=-Syloti_Nagri}');
Error('\P{Is_Block=:=-Syloti_Nagri}');
Expect(1, 43055, '\p{Is_Block=sylotinagri}', "");
Expect(0, 43055, '\p{^Is_Block=sylotinagri}', "");
Expect(0, 43055, '\P{Is_Block=sylotinagri}', "");
Expect(1, 43055, '\P{^Is_Block=sylotinagri}', "");
Expect(0, 43056, '\p{Is_Block=sylotinagri}', "");
Expect(1, 43056, '\p{^Is_Block=sylotinagri}', "");
Expect(1, 43056, '\P{Is_Block=sylotinagri}', "");
Expect(0, 43056, '\P{^Is_Block=sylotinagri}', "");
Expect(1, 43055, '\p{Is_Block=_ SYLOTI_Nagri}', "");
Expect(0, 43055, '\p{^Is_Block=_ SYLOTI_Nagri}', "");
Expect(0, 43055, '\P{Is_Block=_ SYLOTI_Nagri}', "");
Expect(1, 43055, '\P{^Is_Block=_ SYLOTI_Nagri}', "");
Expect(0, 43056, '\p{Is_Block=_ SYLOTI_Nagri}', "");
Expect(1, 43056, '\p{^Is_Block=_ SYLOTI_Nagri}', "");
Expect(1, 43056, '\P{Is_Block=_ SYLOTI_Nagri}', "");
Expect(0, 43056, '\P{^Is_Block=_ SYLOTI_Nagri}', "");
Error('\p{Is_Blk=/a/syloti_NAGRI}');
Error('\P{Is_Blk=/a/syloti_NAGRI}');
Expect(1, 43055, '\p{Is_Blk=sylotinagri}', "");
Expect(0, 43055, '\p{^Is_Blk=sylotinagri}', "");
Expect(0, 43055, '\P{Is_Blk=sylotinagri}', "");
Expect(1, 43055, '\P{^Is_Blk=sylotinagri}', "");
Expect(0, 43056, '\p{Is_Blk=sylotinagri}', "");
Expect(1, 43056, '\p{^Is_Blk=sylotinagri}', "");
Expect(1, 43056, '\P{Is_Blk=sylotinagri}', "");
Expect(0, 43056, '\P{^Is_Blk=sylotinagri}', "");
Expect(1, 43055, '\p{Is_Blk=_ syloti_nagri}', "");
Expect(0, 43055, '\p{^Is_Blk=_ syloti_nagri}', "");
Expect(0, 43055, '\P{Is_Blk=_ syloti_nagri}', "");
Expect(1, 43055, '\P{^Is_Blk=_ syloti_nagri}', "");
Expect(0, 43056, '\p{Is_Blk=_ syloti_nagri}', "");
Expect(1, 43056, '\p{^Is_Blk=_ syloti_nagri}', "");
Expect(1, 43056, '\P{Is_Blk=_ syloti_nagri}', "");
Expect(0, 43056, '\P{^Is_Blk=_ syloti_nagri}', "");
Error('\p{Block= latin_Extended_ADDITIONAL/a/}');
Error('\P{Block= latin_Extended_ADDITIONAL/a/}');
Expect(1, 7935, '\p{Block=latinextendedadditional}', "");
Expect(0, 7935, '\p{^Block=latinextendedadditional}', "");
Expect(0, 7935, '\P{Block=latinextendedadditional}', "");
Expect(1, 7935, '\P{^Block=latinextendedadditional}', "");
Expect(0, 7936, '\p{Block=latinextendedadditional}', "");
Expect(1, 7936, '\p{^Block=latinextendedadditional}', "");
Expect(1, 7936, '\P{Block=latinextendedadditional}', "");
Expect(0, 7936, '\P{^Block=latinextendedadditional}', "");
Expect(1, 7935, '\p{Block=-Latin_extended_ADDITIONAL}', "");
Expect(0, 7935, '\p{^Block=-Latin_extended_ADDITIONAL}', "");
Expect(0, 7935, '\P{Block=-Latin_extended_ADDITIONAL}', "");
Expect(1, 7935, '\P{^Block=-Latin_extended_ADDITIONAL}', "");
Expect(0, 7936, '\p{Block=-Latin_extended_ADDITIONAL}', "");
Expect(1, 7936, '\p{^Block=-Latin_extended_ADDITIONAL}', "");
Expect(1, 7936, '\P{Block=-Latin_extended_ADDITIONAL}', "");
Expect(0, 7936, '\P{^Block=-Latin_extended_ADDITIONAL}', "");
Error('\p{Blk=:=Latin_EXT_Additional}');
Error('\P{Blk=:=Latin_EXT_Additional}');
Expect(1, 7935, '\p{Blk=latinextadditional}', "");
Expect(0, 7935, '\p{^Blk=latinextadditional}', "");
Expect(0, 7935, '\P{Blk=latinextadditional}', "");
Expect(1, 7935, '\P{^Blk=latinextadditional}', "");
Expect(0, 7936, '\p{Blk=latinextadditional}', "");
Expect(1, 7936, '\p{^Blk=latinextadditional}', "");
Expect(1, 7936, '\P{Blk=latinextadditional}', "");
Expect(0, 7936, '\P{^Blk=latinextadditional}', "");
Expect(1, 7935, '\p{Blk=__Latin_ext_Additional}', "");
Expect(0, 7935, '\p{^Blk=__Latin_ext_Additional}', "");
Expect(0, 7935, '\P{Blk=__Latin_ext_Additional}', "");
Expect(1, 7935, '\P{^Blk=__Latin_ext_Additional}', "");
Expect(0, 7936, '\p{Blk=__Latin_ext_Additional}', "");
Expect(1, 7936, '\p{^Blk=__Latin_ext_Additional}', "");
Expect(1, 7936, '\P{Blk=__Latin_ext_Additional}', "");
Expect(0, 7936, '\P{^Blk=__Latin_ext_Additional}', "");
Error('\p{Is_Block=/a/ Latin_extended_ADDITIONAL}');
Error('\P{Is_Block=/a/ Latin_extended_ADDITIONAL}');
Expect(1, 7935, '\p{Is_Block=latinextendedadditional}', "");
Expect(0, 7935, '\p{^Is_Block=latinextendedadditional}', "");
Expect(0, 7935, '\P{Is_Block=latinextendedadditional}', "");
Expect(1, 7935, '\P{^Is_Block=latinextendedadditional}', "");
Expect(0, 7936, '\p{Is_Block=latinextendedadditional}', "");
Expect(1, 7936, '\p{^Is_Block=latinextendedadditional}', "");
Expect(1, 7936, '\P{Is_Block=latinextendedadditional}', "");
Expect(0, 7936, '\P{^Is_Block=latinextendedadditional}', "");
Expect(1, 7935, '\p{Is_Block=	 LATIN_extended_additional}', "");
Expect(0, 7935, '\p{^Is_Block=	 LATIN_extended_additional}', "");
Expect(0, 7935, '\P{Is_Block=	 LATIN_extended_additional}', "");
Expect(1, 7935, '\P{^Is_Block=	 LATIN_extended_additional}', "");
Expect(0, 7936, '\p{Is_Block=	 LATIN_extended_additional}', "");
Expect(1, 7936, '\p{^Is_Block=	 LATIN_extended_additional}', "");
Expect(1, 7936, '\P{Is_Block=	 LATIN_extended_additional}', "");
Expect(0, 7936, '\P{^Is_Block=	 LATIN_extended_additional}', "");
Error('\p{Is_Blk= /a/Latin_ext_Additional}');
Error('\P{Is_Blk= /a/Latin_ext_Additional}');
Expect(1, 7935, '\p{Is_Blk=latinextadditional}', "");
Expect(0, 7935, '\p{^Is_Blk=latinextadditional}', "");
Expect(0, 7935, '\P{Is_Blk=latinextadditional}', "");
Expect(1, 7935, '\P{^Is_Blk=latinextadditional}', "");
Expect(0, 7936, '\p{Is_Blk=latinextadditional}', "");
Expect(1, 7936, '\p{^Is_Blk=latinextadditional}', "");
Expect(1, 7936, '\P{Is_Blk=latinextadditional}', "");
Expect(0, 7936, '\P{^Is_Blk=latinextadditional}', "");
Expect(1, 7935, '\p{Is_Blk=	LATIN_ext_Additional}', "");
Expect(0, 7935, '\p{^Is_Blk=	LATIN_ext_Additional}', "");
Expect(0, 7935, '\P{Is_Blk=	LATIN_ext_Additional}', "");
Expect(1, 7935, '\P{^Is_Blk=	LATIN_ext_Additional}', "");
Expect(0, 7936, '\p{Is_Blk=	LATIN_ext_Additional}', "");
Expect(1, 7936, '\p{^Is_Blk=	LATIN_ext_Additional}', "");
Expect(1, 7936, '\P{Is_Blk=	LATIN_ext_Additional}', "");
Expect(0, 7936, '\P{^Is_Blk=	LATIN_ext_Additional}', "");
Error('\p{Block=  aegean_NUMBERS:=}');
Error('\P{Block=  aegean_NUMBERS:=}');
Expect(1, 65855, '\p{Block=aegeannumbers}', "");
Expect(0, 65855, '\p{^Block=aegeannumbers}', "");
Expect(0, 65855, '\P{Block=aegeannumbers}', "");
Expect(1, 65855, '\P{^Block=aegeannumbers}', "");
Expect(0, 65856, '\p{Block=aegeannumbers}', "");
Expect(1, 65856, '\p{^Block=aegeannumbers}', "");
Expect(1, 65856, '\P{Block=aegeannumbers}', "");
Expect(0, 65856, '\P{^Block=aegeannumbers}', "");
Expect(1, 65855, '\p{Block= -Aegean_Numbers}', "");
Expect(0, 65855, '\p{^Block= -Aegean_Numbers}', "");
Expect(0, 65855, '\P{Block= -Aegean_Numbers}', "");
Expect(1, 65855, '\P{^Block= -Aegean_Numbers}', "");
Expect(0, 65856, '\p{Block= -Aegean_Numbers}', "");
Expect(1, 65856, '\p{^Block= -Aegean_Numbers}', "");
Expect(1, 65856, '\P{Block= -Aegean_Numbers}', "");
Expect(0, 65856, '\P{^Block= -Aegean_Numbers}', "");
Error('\p{Blk=/a/aegean_Numbers}');
Error('\P{Blk=/a/aegean_Numbers}');
Expect(1, 65855, '\p{Blk:	aegeannumbers}', "");
Expect(0, 65855, '\p{^Blk:	aegeannumbers}', "");
Expect(0, 65855, '\P{Blk:	aegeannumbers}', "");
Expect(1, 65855, '\P{^Blk:	aegeannumbers}', "");
Expect(0, 65856, '\p{Blk:	aegeannumbers}', "");
Expect(1, 65856, '\p{^Blk:	aegeannumbers}', "");
Expect(1, 65856, '\P{Blk:	aegeannumbers}', "");
Expect(0, 65856, '\P{^Blk:	aegeannumbers}', "");
Expect(1, 65855, '\p{Blk=	-Aegean_Numbers}', "");
Expect(0, 65855, '\p{^Blk=	-Aegean_Numbers}', "");
Expect(0, 65855, '\P{Blk=	-Aegean_Numbers}', "");
Expect(1, 65855, '\P{^Blk=	-Aegean_Numbers}', "");
Expect(0, 65856, '\p{Blk=	-Aegean_Numbers}', "");
Expect(1, 65856, '\p{^Blk=	-Aegean_Numbers}', "");
Expect(1, 65856, '\P{Blk=	-Aegean_Numbers}', "");
Expect(0, 65856, '\P{^Blk=	-Aegean_Numbers}', "");
Error('\p{Is_Block=/a/	 aegean_numbers}');
Error('\P{Is_Block=/a/	 aegean_numbers}');
Expect(1, 65855, '\p{Is_Block=aegeannumbers}', "");
Expect(0, 65855, '\p{^Is_Block=aegeannumbers}', "");
Expect(0, 65855, '\P{Is_Block=aegeannumbers}', "");
Expect(1, 65855, '\P{^Is_Block=aegeannumbers}', "");
Expect(0, 65856, '\p{Is_Block=aegeannumbers}', "");
Expect(1, 65856, '\p{^Is_Block=aegeannumbers}', "");
Expect(1, 65856, '\P{Is_Block=aegeannumbers}', "");
Expect(0, 65856, '\P{^Is_Block=aegeannumbers}', "");
Expect(1, 65855, '\p{Is_Block: 	-Aegean_NUMBERS}', "");
Expect(0, 65855, '\p{^Is_Block: 	-Aegean_NUMBERS}', "");
Expect(0, 65855, '\P{Is_Block: 	-Aegean_NUMBERS}', "");
Expect(1, 65855, '\P{^Is_Block: 	-Aegean_NUMBERS}', "");
Expect(0, 65856, '\p{Is_Block: 	-Aegean_NUMBERS}', "");
Expect(1, 65856, '\p{^Is_Block: 	-Aegean_NUMBERS}', "");
Expect(1, 65856, '\P{Is_Block: 	-Aegean_NUMBERS}', "");
Expect(0, 65856, '\P{^Is_Block: 	-Aegean_NUMBERS}', "");
Error('\p{Is_Blk=/a/_aegean_Numbers}');
Error('\P{Is_Blk=/a/_aegean_Numbers}');
Expect(1, 65855, '\p{Is_Blk=aegeannumbers}', "");
Expect(0, 65855, '\p{^Is_Blk=aegeannumbers}', "");
Expect(0, 65855, '\P{Is_Blk=aegeannumbers}', "");
Expect(1, 65855, '\P{^Is_Blk=aegeannumbers}', "");
Expect(0, 65856, '\p{Is_Blk=aegeannumbers}', "");
Expect(1, 65856, '\p{^Is_Blk=aegeannumbers}', "");
Expect(1, 65856, '\P{Is_Blk=aegeannumbers}', "");
Expect(0, 65856, '\P{^Is_Blk=aegeannumbers}', "");
Expect(1, 65855, '\p{Is_Blk: _-Aegean_Numbers}', "");
Expect(0, 65855, '\p{^Is_Blk: _-Aegean_Numbers}', "");
Expect(0, 65855, '\P{Is_Blk: _-Aegean_Numbers}', "");
Expect(1, 65855, '\P{^Is_Blk: _-Aegean_Numbers}', "");
Expect(0, 65856, '\p{Is_Blk: _-Aegean_Numbers}', "");
Expect(1, 65856, '\p{^Is_Blk: _-Aegean_Numbers}', "");
Expect(1, 65856, '\P{Is_Blk: _-Aegean_Numbers}', "");
Expect(0, 65856, '\P{^Is_Blk: _-Aegean_Numbers}', "");
Error('\p{Block=_Old_PERSIAN:=}');
Error('\P{Block=_Old_PERSIAN:=}');
Expect(1, 66527, '\p{Block=oldpersian}', "");
Expect(0, 66527, '\p{^Block=oldpersian}', "");
Expect(0, 66527, '\P{Block=oldpersian}', "");
Expect(1, 66527, '\P{^Block=oldpersian}', "");
Expect(0, 66528, '\p{Block=oldpersian}', "");
Expect(1, 66528, '\p{^Block=oldpersian}', "");
Expect(1, 66528, '\P{Block=oldpersian}', "");
Expect(0, 66528, '\P{^Block=oldpersian}', "");
Expect(1, 66527, '\p{Block=__OLD_Persian}', "");
Expect(0, 66527, '\p{^Block=__OLD_Persian}', "");
Expect(0, 66527, '\P{Block=__OLD_Persian}', "");
Expect(1, 66527, '\P{^Block=__OLD_Persian}', "");
Expect(0, 66528, '\p{Block=__OLD_Persian}', "");
Expect(1, 66528, '\p{^Block=__OLD_Persian}', "");
Expect(1, 66528, '\P{Block=__OLD_Persian}', "");
Expect(0, 66528, '\P{^Block=__OLD_Persian}', "");
Error('\p{Blk=:=old_persian}');
Error('\P{Blk=:=old_persian}');
Expect(1, 66527, '\p{Blk=oldpersian}', "");
Expect(0, 66527, '\p{^Blk=oldpersian}', "");
Expect(0, 66527, '\P{Blk=oldpersian}', "");
Expect(1, 66527, '\P{^Blk=oldpersian}', "");
Expect(0, 66528, '\p{Blk=oldpersian}', "");
Expect(1, 66528, '\p{^Blk=oldpersian}', "");
Expect(1, 66528, '\P{Blk=oldpersian}', "");
Expect(0, 66528, '\P{^Blk=oldpersian}', "");
Expect(1, 66527, '\p{Blk=-Old_Persian}', "");
Expect(0, 66527, '\p{^Blk=-Old_Persian}', "");
Expect(0, 66527, '\P{Blk=-Old_Persian}', "");
Expect(1, 66527, '\P{^Blk=-Old_Persian}', "");
Expect(0, 66528, '\p{Blk=-Old_Persian}', "");
Expect(1, 66528, '\p{^Blk=-Old_Persian}', "");
Expect(1, 66528, '\P{Blk=-Old_Persian}', "");
Expect(0, 66528, '\P{^Blk=-Old_Persian}', "");
Error('\p{Is_Block=	OLD_persian:=}');
Error('\P{Is_Block=	OLD_persian:=}');
Expect(1, 66527, '\p{Is_Block=oldpersian}', "");
Expect(0, 66527, '\p{^Is_Block=oldpersian}', "");
Expect(0, 66527, '\P{Is_Block=oldpersian}', "");
Expect(1, 66527, '\P{^Is_Block=oldpersian}', "");
Expect(0, 66528, '\p{Is_Block=oldpersian}', "");
Expect(1, 66528, '\p{^Is_Block=oldpersian}', "");
Expect(1, 66528, '\P{Is_Block=oldpersian}', "");
Expect(0, 66528, '\P{^Is_Block=oldpersian}', "");
Expect(1, 66527, '\p{Is_Block=-Old_persian}', "");
Expect(0, 66527, '\p{^Is_Block=-Old_persian}', "");
Expect(0, 66527, '\P{Is_Block=-Old_persian}', "");
Expect(1, 66527, '\P{^Is_Block=-Old_persian}', "");
Expect(0, 66528, '\p{Is_Block=-Old_persian}', "");
Expect(1, 66528, '\p{^Is_Block=-Old_persian}', "");
Expect(1, 66528, '\P{Is_Block=-Old_persian}', "");
Expect(0, 66528, '\P{^Is_Block=-Old_persian}', "");
Error('\p{Is_Blk=/a/	Old_Persian}');
Error('\P{Is_Blk=/a/	Old_Persian}');
Expect(1, 66527, '\p{Is_Blk=oldpersian}', "");
Expect(0, 66527, '\p{^Is_Blk=oldpersian}', "");
Expect(0, 66527, '\P{Is_Blk=oldpersian}', "");
Expect(1, 66527, '\P{^Is_Blk=oldpersian}', "");
Expect(0, 66528, '\p{Is_Blk=oldpersian}', "");
Expect(1, 66528, '\p{^Is_Blk=oldpersian}', "");
Expect(1, 66528, '\P{Is_Blk=oldpersian}', "");
Expect(0, 66528, '\P{^Is_Blk=oldpersian}', "");
Expect(1, 66527, '\p{Is_Blk= -Old_Persian}', "");
Expect(0, 66527, '\p{^Is_Blk= -Old_Persian}', "");
Expect(0, 66527, '\P{Is_Blk= -Old_Persian}', "");
Expect(1, 66527, '\P{^Is_Blk= -Old_Persian}', "");
Expect(0, 66528, '\p{Is_Blk= -Old_Persian}', "");
Expect(1, 66528, '\p{^Is_Blk= -Old_Persian}', "");
Expect(1, 66528, '\P{Is_Blk= -Old_Persian}', "");
Expect(0, 66528, '\P{^Is_Blk= -Old_Persian}', "");
Error('\p{Block=	:=Modifier_Tone_Letters}');
Error('\P{Block=	:=Modifier_Tone_Letters}');
Expect(1, 42783, '\p{Block=modifiertoneletters}', "");
Expect(0, 42783, '\p{^Block=modifiertoneletters}', "");
Expect(0, 42783, '\P{Block=modifiertoneletters}', "");
Expect(1, 42783, '\P{^Block=modifiertoneletters}', "");
Expect(0, 42784, '\p{Block=modifiertoneletters}', "");
Expect(1, 42784, '\p{^Block=modifiertoneletters}', "");
Expect(1, 42784, '\P{Block=modifiertoneletters}', "");
Expect(0, 42784, '\P{^Block=modifiertoneletters}', "");
Expect(1, 42783, '\p{Block=	_modifier_Tone_Letters}', "");
Expect(0, 42783, '\p{^Block=	_modifier_Tone_Letters}', "");
Expect(0, 42783, '\P{Block=	_modifier_Tone_Letters}', "");
Expect(1, 42783, '\P{^Block=	_modifier_Tone_Letters}', "");
Expect(0, 42784, '\p{Block=	_modifier_Tone_Letters}', "");
Expect(1, 42784, '\p{^Block=	_modifier_Tone_Letters}', "");
Expect(1, 42784, '\P{Block=	_modifier_Tone_Letters}', "");
Expect(0, 42784, '\P{^Block=	_modifier_Tone_Letters}', "");
Error('\p{Blk= /a/Modifier_Tone_Letters}');
Error('\P{Blk= /a/Modifier_Tone_Letters}');
Expect(1, 42783, '\p{Blk=modifiertoneletters}', "");
Expect(0, 42783, '\p{^Blk=modifiertoneletters}', "");
Expect(0, 42783, '\P{Blk=modifiertoneletters}', "");
Expect(1, 42783, '\P{^Blk=modifiertoneletters}', "");
Expect(0, 42784, '\p{Blk=modifiertoneletters}', "");
Expect(1, 42784, '\p{^Blk=modifiertoneletters}', "");
Expect(1, 42784, '\P{Blk=modifiertoneletters}', "");
Expect(0, 42784, '\P{^Blk=modifiertoneletters}', "");
Expect(1, 42783, '\p{Blk=- Modifier_Tone_letters}', "");
Expect(0, 42783, '\p{^Blk=- Modifier_Tone_letters}', "");
Expect(0, 42783, '\P{Blk=- Modifier_Tone_letters}', "");
Expect(1, 42783, '\P{^Blk=- Modifier_Tone_letters}', "");
Expect(0, 42784, '\p{Blk=- Modifier_Tone_letters}', "");
Expect(1, 42784, '\p{^Blk=- Modifier_Tone_letters}', "");
Expect(1, 42784, '\P{Blk=- Modifier_Tone_letters}', "");
Expect(0, 42784, '\P{^Blk=- Modifier_Tone_letters}', "");
Error('\p{Is_Block= modifier_Tone_letters:=}');
Error('\P{Is_Block= modifier_Tone_letters:=}');
Expect(1, 42783, '\p{Is_Block=modifiertoneletters}', "");
Expect(0, 42783, '\p{^Is_Block=modifiertoneletters}', "");
Expect(0, 42783, '\P{Is_Block=modifiertoneletters}', "");
Expect(1, 42783, '\P{^Is_Block=modifiertoneletters}', "");
Expect(0, 42784, '\p{Is_Block=modifiertoneletters}', "");
Expect(1, 42784, '\p{^Is_Block=modifiertoneletters}', "");
Expect(1, 42784, '\P{Is_Block=modifiertoneletters}', "");
Expect(0, 42784, '\P{^Is_Block=modifiertoneletters}', "");
Expect(1, 42783, '\p{Is_Block=-	MODIFIER_Tone_Letters}', "");
Expect(0, 42783, '\p{^Is_Block=-	MODIFIER_Tone_Letters}', "");
Expect(0, 42783, '\P{Is_Block=-	MODIFIER_Tone_Letters}', "");
Expect(1, 42783, '\P{^Is_Block=-	MODIFIER_Tone_Letters}', "");
Expect(0, 42784, '\p{Is_Block=-	MODIFIER_Tone_Letters}', "");
Expect(1, 42784, '\p{^Is_Block=-	MODIFIER_Tone_Letters}', "");
Expect(1, 42784, '\P{Is_Block=-	MODIFIER_Tone_Letters}', "");
Expect(0, 42784, '\P{^Is_Block=-	MODIFIER_Tone_Letters}', "");
Error('\p{Is_Blk: 		modifier_TONE_LETTERS:=}');
Error('\P{Is_Blk: 		modifier_TONE_LETTERS:=}');
Expect(1, 42783, '\p{Is_Blk: modifiertoneletters}', "");
Expect(0, 42783, '\p{^Is_Blk: modifiertoneletters}', "");
Expect(0, 42783, '\P{Is_Blk: modifiertoneletters}', "");
Expect(1, 42783, '\P{^Is_Blk: modifiertoneletters}', "");
Expect(0, 42784, '\p{Is_Blk: modifiertoneletters}', "");
Expect(1, 42784, '\p{^Is_Blk: modifiertoneletters}', "");
Expect(1, 42784, '\P{Is_Blk: modifiertoneletters}', "");
Expect(0, 42784, '\P{^Is_Blk: modifiertoneletters}', "");
Expect(1, 42783, '\p{Is_Blk:	_modifier_tone_Letters}', "");
Expect(0, 42783, '\p{^Is_Blk:	_modifier_tone_Letters}', "");
Expect(0, 42783, '\P{Is_Blk:	_modifier_tone_Letters}', "");
Expect(1, 42783, '\P{^Is_Blk:	_modifier_tone_Letters}', "");
Expect(0, 42784, '\p{Is_Blk:	_modifier_tone_Letters}', "");
Expect(1, 42784, '\p{^Is_Blk:	_modifier_tone_Letters}', "");
Expect(1, 42784, '\P{Is_Blk:	_modifier_tone_Letters}', "");
Expect(0, 42784, '\P{^Is_Blk:	_modifier_tone_Letters}', "");
Error('\p{Block=:=-_HANGUL_JAMO_Extended_B}');
Error('\P{Block=:=-_HANGUL_JAMO_Extended_B}');
Expect(1, 55295, '\p{Block=hanguljamoextendedb}', "");
Expect(0, 55295, '\p{^Block=hanguljamoextendedb}', "");
Expect(0, 55295, '\P{Block=hanguljamoextendedb}', "");
Expect(1, 55295, '\P{^Block=hanguljamoextendedb}', "");
Expect(0, 57344, '\p{Block=hanguljamoextendedb}', "");
Expect(1, 57344, '\p{^Block=hanguljamoextendedb}', "");
Expect(1, 57344, '\P{Block=hanguljamoextendedb}', "");
Expect(0, 57344, '\P{^Block=hanguljamoextendedb}', "");
Expect(1, 55295, '\p{Block=	-hangul_Jamo_EXTENDED_B}', "");
Expect(0, 55295, '\p{^Block=	-hangul_Jamo_EXTENDED_B}', "");
Expect(0, 55295, '\P{Block=	-hangul_Jamo_EXTENDED_B}', "");
Expect(1, 55295, '\P{^Block=	-hangul_Jamo_EXTENDED_B}', "");
Expect(0, 57344, '\p{Block=	-hangul_Jamo_EXTENDED_B}', "");
Expect(1, 57344, '\p{^Block=	-hangul_Jamo_EXTENDED_B}', "");
Expect(1, 57344, '\P{Block=	-hangul_Jamo_EXTENDED_B}', "");
Expect(0, 57344, '\P{^Block=	-hangul_Jamo_EXTENDED_B}', "");
Error('\p{Blk=-JAMO_ext_B:=}');
Error('\P{Blk=-JAMO_ext_B:=}');
Expect(1, 55295, '\p{Blk=jamoextb}', "");
Expect(0, 55295, '\p{^Blk=jamoextb}', "");
Expect(0, 55295, '\P{Blk=jamoextb}', "");
Expect(1, 55295, '\P{^Blk=jamoextb}', "");
Expect(0, 57344, '\p{Blk=jamoextb}', "");
Expect(1, 57344, '\p{^Blk=jamoextb}', "");
Expect(1, 57344, '\P{Blk=jamoextb}', "");
Expect(0, 57344, '\P{^Blk=jamoextb}', "");
Expect(1, 55295, '\p{Blk=-_Jamo_EXT_b}', "");
Expect(0, 55295, '\p{^Blk=-_Jamo_EXT_b}', "");
Expect(0, 55295, '\P{Blk=-_Jamo_EXT_b}', "");
Expect(1, 55295, '\P{^Blk=-_Jamo_EXT_b}', "");
Expect(0, 57344, '\p{Blk=-_Jamo_EXT_b}', "");
Expect(1, 57344, '\p{^Blk=-_Jamo_EXT_b}', "");
Expect(1, 57344, '\P{Blk=-_Jamo_EXT_b}', "");
Expect(0, 57344, '\P{^Blk=-_Jamo_EXT_b}', "");
Error('\p{Is_Block=_/a/hangul_JAMO_Extended_B}');
Error('\P{Is_Block=_/a/hangul_JAMO_Extended_B}');
Expect(1, 55295, '\p{Is_Block=hanguljamoextendedb}', "");
Expect(0, 55295, '\p{^Is_Block=hanguljamoextendedb}', "");
Expect(0, 55295, '\P{Is_Block=hanguljamoextendedb}', "");
Expect(1, 55295, '\P{^Is_Block=hanguljamoextendedb}', "");
Expect(0, 57344, '\p{Is_Block=hanguljamoextendedb}', "");
Expect(1, 57344, '\p{^Is_Block=hanguljamoextendedb}', "");
Expect(1, 57344, '\P{Is_Block=hanguljamoextendedb}', "");
Expect(0, 57344, '\P{^Is_Block=hanguljamoextendedb}', "");
Expect(1, 55295, '\p{Is_Block=__HANGUL_Jamo_extended_B}', "");
Expect(0, 55295, '\p{^Is_Block=__HANGUL_Jamo_extended_B}', "");
Expect(0, 55295, '\P{Is_Block=__HANGUL_Jamo_extended_B}', "");
Expect(1, 55295, '\P{^Is_Block=__HANGUL_Jamo_extended_B}', "");
Expect(0, 57344, '\p{Is_Block=__HANGUL_Jamo_extended_B}', "");
Expect(1, 57344, '\p{^Is_Block=__HANGUL_Jamo_extended_B}', "");
Expect(1, 57344, '\P{Is_Block=__HANGUL_Jamo_extended_B}', "");
Expect(0, 57344, '\P{^Is_Block=__HANGUL_Jamo_extended_B}', "");
Error('\p{Is_Blk=	JAMO_EXT_B/a/}');
Error('\P{Is_Blk=	JAMO_EXT_B/a/}');
Expect(1, 55295, '\p{Is_Blk: jamoextb}', "");
Expect(0, 55295, '\p{^Is_Blk: jamoextb}', "");
Expect(0, 55295, '\P{Is_Blk: jamoextb}', "");
Expect(1, 55295, '\P{^Is_Blk: jamoextb}', "");
Expect(0, 57344, '\p{Is_Blk: jamoextb}', "");
Expect(1, 57344, '\p{^Is_Blk: jamoextb}', "");
Expect(1, 57344, '\P{Is_Blk: jamoextb}', "");
Expect(0, 57344, '\P{^Is_Blk: jamoextb}', "");
Expect(1, 55295, '\p{Is_Blk=--Jamo_Ext_B}', "");
Expect(0, 55295, '\p{^Is_Blk=--Jamo_Ext_B}', "");
Expect(0, 55295, '\P{Is_Blk=--Jamo_Ext_B}', "");
Expect(1, 55295, '\P{^Is_Blk=--Jamo_Ext_B}', "");
Expect(0, 57344, '\p{Is_Blk=--Jamo_Ext_B}', "");
Expect(1, 57344, '\p{^Is_Blk=--Jamo_Ext_B}', "");
Expect(1, 57344, '\P{Is_Blk=--Jamo_Ext_B}', "");
Expect(0, 57344, '\P{^Is_Blk=--Jamo_Ext_B}', "");
Error('\p{Block=-Linear_b_syllabary:=}');
Error('\P{Block=-Linear_b_syllabary:=}');
Expect(1, 65663, '\p{Block=linearbsyllabary}', "");
Expect(0, 65663, '\p{^Block=linearbsyllabary}', "");
Expect(0, 65663, '\P{Block=linearbsyllabary}', "");
Expect(1, 65663, '\P{^Block=linearbsyllabary}', "");
Expect(0, 65664, '\p{Block=linearbsyllabary}', "");
Expect(1, 65664, '\p{^Block=linearbsyllabary}', "");
Expect(1, 65664, '\P{Block=linearbsyllabary}', "");
Expect(0, 65664, '\P{^Block=linearbsyllabary}', "");
Expect(1, 65663, '\p{Block=-_LINEAR_B_SYLLABARY}', "");
Expect(0, 65663, '\p{^Block=-_LINEAR_B_SYLLABARY}', "");
Expect(0, 65663, '\P{Block=-_LINEAR_B_SYLLABARY}', "");
Expect(1, 65663, '\P{^Block=-_LINEAR_B_SYLLABARY}', "");
Expect(0, 65664, '\p{Block=-_LINEAR_B_SYLLABARY}', "");
Expect(1, 65664, '\p{^Block=-_LINEAR_B_SYLLABARY}', "");
Expect(1, 65664, '\P{Block=-_LINEAR_B_SYLLABARY}', "");
Expect(0, 65664, '\P{^Block=-_LINEAR_B_SYLLABARY}', "");
Error('\p{Blk=:=Linear_B_Syllabary}');
Error('\P{Blk=:=Linear_B_Syllabary}');
Expect(1, 65663, '\p{Blk=linearbsyllabary}', "");
Expect(0, 65663, '\p{^Blk=linearbsyllabary}', "");
Expect(0, 65663, '\P{Blk=linearbsyllabary}', "");
Expect(1, 65663, '\P{^Blk=linearbsyllabary}', "");
Expect(0, 65664, '\p{Blk=linearbsyllabary}', "");
Expect(1, 65664, '\p{^Blk=linearbsyllabary}', "");
Expect(1, 65664, '\P{Blk=linearbsyllabary}', "");
Expect(0, 65664, '\P{^Blk=linearbsyllabary}', "");
Expect(1, 65663, '\p{Blk= linear_B_Syllabary}', "");
Expect(0, 65663, '\p{^Blk= linear_B_Syllabary}', "");
Expect(0, 65663, '\P{Blk= linear_B_Syllabary}', "");
Expect(1, 65663, '\P{^Blk= linear_B_Syllabary}', "");
Expect(0, 65664, '\p{Blk= linear_B_Syllabary}', "");
Expect(1, 65664, '\p{^Blk= linear_B_Syllabary}', "");
Expect(1, 65664, '\P{Blk= linear_B_Syllabary}', "");
Expect(0, 65664, '\P{^Blk= linear_B_Syllabary}', "");
Error('\p{Is_Block: 	/a/linear_b_Syllabary}');
Error('\P{Is_Block: 	/a/linear_b_Syllabary}');
Expect(1, 65663, '\p{Is_Block:   linearbsyllabary}', "");
Expect(0, 65663, '\p{^Is_Block:   linearbsyllabary}', "");
Expect(0, 65663, '\P{Is_Block:   linearbsyllabary}', "");
Expect(1, 65663, '\P{^Is_Block:   linearbsyllabary}', "");
Expect(0, 65664, '\p{Is_Block:   linearbsyllabary}', "");
Expect(1, 65664, '\p{^Is_Block:   linearbsyllabary}', "");
Expect(1, 65664, '\P{Is_Block:   linearbsyllabary}', "");
Expect(0, 65664, '\P{^Is_Block:   linearbsyllabary}', "");
Expect(1, 65663, '\p{Is_Block=--LINEAR_B_SYLLABARY}', "");
Expect(0, 65663, '\p{^Is_Block=--LINEAR_B_SYLLABARY}', "");
Expect(0, 65663, '\P{Is_Block=--LINEAR_B_SYLLABARY}', "");
Expect(1, 65663, '\P{^Is_Block=--LINEAR_B_SYLLABARY}', "");
Expect(0, 65664, '\p{Is_Block=--LINEAR_B_SYLLABARY}', "");
Expect(1, 65664, '\p{^Is_Block=--LINEAR_B_SYLLABARY}', "");
Expect(1, 65664, '\P{Is_Block=--LINEAR_B_SYLLABARY}', "");
Expect(0, 65664, '\P{^Is_Block=--LINEAR_B_SYLLABARY}', "");
Error('\p{Is_Blk=-:=linear_B_syllabary}');
Error('\P{Is_Blk=-:=linear_B_syllabary}');
Expect(1, 65663, '\p{Is_Blk=linearbsyllabary}', "");
Expect(0, 65663, '\p{^Is_Blk=linearbsyllabary}', "");
Expect(0, 65663, '\P{Is_Blk=linearbsyllabary}', "");
Expect(1, 65663, '\P{^Is_Blk=linearbsyllabary}', "");
Expect(0, 65664, '\p{Is_Blk=linearbsyllabary}', "");
Expect(1, 65664, '\p{^Is_Blk=linearbsyllabary}', "");
Expect(1, 65664, '\P{Is_Blk=linearbsyllabary}', "");
Expect(0, 65664, '\P{^Is_Blk=linearbsyllabary}', "");
Expect(1, 65663, '\p{Is_Blk= -Linear_B_Syllabary}', "");
Expect(0, 65663, '\p{^Is_Blk= -Linear_B_Syllabary}', "");
Expect(0, 65663, '\P{Is_Blk= -Linear_B_Syllabary}', "");
Expect(1, 65663, '\P{^Is_Blk= -Linear_B_Syllabary}', "");
Expect(0, 65664, '\p{Is_Blk= -Linear_B_Syllabary}', "");
Expect(1, 65664, '\p{^Is_Blk= -Linear_B_Syllabary}', "");
Expect(1, 65664, '\P{Is_Blk= -Linear_B_Syllabary}', "");
Expect(0, 65664, '\P{^Is_Blk= -Linear_B_Syllabary}', "");
Error('\p{Block= /a/arrows}');
Error('\P{Block= /a/arrows}');
Expect(1, 8703, '\p{Block=arrows}', "");
Expect(0, 8703, '\p{^Block=arrows}', "");
Expect(0, 8703, '\P{Block=arrows}', "");
Expect(1, 8703, '\P{^Block=arrows}', "");
Expect(0, 8704, '\p{Block=arrows}', "");
Expect(1, 8704, '\p{^Block=arrows}', "");
Expect(1, 8704, '\P{Block=arrows}', "");
Expect(0, 8704, '\P{^Block=arrows}', "");
Expect(1, 8703, '\p{Block=--Arrows}', "");
Expect(0, 8703, '\p{^Block=--Arrows}', "");
Expect(0, 8703, '\P{Block=--Arrows}', "");
Expect(1, 8703, '\P{^Block=--Arrows}', "");
Expect(0, 8704, '\p{Block=--Arrows}', "");
Expect(1, 8704, '\p{^Block=--Arrows}', "");
Expect(1, 8704, '\P{Block=--Arrows}', "");
Expect(0, 8704, '\P{^Block=--Arrows}', "");
Error('\p{Blk=:= _Arrows}');
Error('\P{Blk=:= _Arrows}');
Expect(1, 8703, '\p{Blk=arrows}', "");
Expect(0, 8703, '\p{^Blk=arrows}', "");
Expect(0, 8703, '\P{Blk=arrows}', "");
Expect(1, 8703, '\P{^Blk=arrows}', "");
Expect(0, 8704, '\p{Blk=arrows}', "");
Expect(1, 8704, '\p{^Blk=arrows}', "");
Expect(1, 8704, '\P{Blk=arrows}', "");
Expect(0, 8704, '\P{^Blk=arrows}', "");
Expect(1, 8703, '\p{Blk=	 Arrows}', "");
Expect(0, 8703, '\p{^Blk=	 Arrows}', "");
Expect(0, 8703, '\P{Blk=	 Arrows}', "");
Expect(1, 8703, '\P{^Blk=	 Arrows}', "");
Expect(0, 8704, '\p{Blk=	 Arrows}', "");
Expect(1, 8704, '\p{^Blk=	 Arrows}', "");
Expect(1, 8704, '\P{Blk=	 Arrows}', "");
Expect(0, 8704, '\P{^Blk=	 Arrows}', "");
Error('\p{Is_Block= Arrows:=}');
Error('\P{Is_Block= Arrows:=}');
Expect(1, 8703, '\p{Is_Block=arrows}', "");
Expect(0, 8703, '\p{^Is_Block=arrows}', "");
Expect(0, 8703, '\P{Is_Block=arrows}', "");
Expect(1, 8703, '\P{^Is_Block=arrows}', "");
Expect(0, 8704, '\p{Is_Block=arrows}', "");
Expect(1, 8704, '\p{^Is_Block=arrows}', "");
Expect(1, 8704, '\P{Is_Block=arrows}', "");
Expect(0, 8704, '\P{^Is_Block=arrows}', "");
Expect(1, 8703, '\p{Is_Block=__ARROWS}', "");
Expect(0, 8703, '\p{^Is_Block=__ARROWS}', "");
Expect(0, 8703, '\P{Is_Block=__ARROWS}', "");
Expect(1, 8703, '\P{^Is_Block=__ARROWS}', "");
Expect(0, 8704, '\p{Is_Block=__ARROWS}', "");
Expect(1, 8704, '\p{^Is_Block=__ARROWS}', "");
Expect(1, 8704, '\P{Is_Block=__ARROWS}', "");
Expect(0, 8704, '\P{^Is_Block=__ARROWS}', "");
Error('\p{Is_Blk=	:=Arrows}');
Error('\P{Is_Blk=	:=Arrows}');
Expect(1, 8703, '\p{Is_Blk=arrows}', "");
Expect(0, 8703, '\p{^Is_Blk=arrows}', "");
Expect(0, 8703, '\P{Is_Blk=arrows}', "");
Expect(1, 8703, '\P{^Is_Blk=arrows}', "");
Expect(0, 8704, '\p{Is_Blk=arrows}', "");
Expect(1, 8704, '\p{^Is_Blk=arrows}', "");
Expect(1, 8704, '\P{Is_Blk=arrows}', "");
Expect(0, 8704, '\P{^Is_Blk=arrows}', "");
Expect(1, 8703, '\p{Is_Blk: -Arrows}', "");
Expect(0, 8703, '\p{^Is_Blk: -Arrows}', "");
Expect(0, 8703, '\P{Is_Blk: -Arrows}', "");
Expect(1, 8703, '\P{^Is_Blk: -Arrows}', "");
Expect(0, 8704, '\p{Is_Blk: -Arrows}', "");
Expect(1, 8704, '\p{^Is_Blk: -Arrows}', "");
Expect(1, 8704, '\P{Is_Blk: -Arrows}', "");
Expect(0, 8704, '\P{^Is_Blk: -Arrows}', "");
Error('\p{Block= -Malayalam:=}');
Error('\P{Block= -Malayalam:=}');
Expect(1, 3455, '\p{Block=malayalam}', "");
Expect(0, 3455, '\p{^Block=malayalam}', "");
Expect(0, 3455, '\P{Block=malayalam}', "");
Expect(1, 3455, '\P{^Block=malayalam}', "");
Expect(0, 3456, '\p{Block=malayalam}', "");
Expect(1, 3456, '\p{^Block=malayalam}', "");
Expect(1, 3456, '\P{Block=malayalam}', "");
Expect(0, 3456, '\P{^Block=malayalam}', "");
Expect(1, 3455, '\p{Block=Malayalam}', "");
Expect(0, 3455, '\p{^Block=Malayalam}', "");
Expect(0, 3455, '\P{Block=Malayalam}', "");
Expect(1, 3455, '\P{^Block=Malayalam}', "");
Expect(0, 3456, '\p{Block=Malayalam}', "");
Expect(1, 3456, '\p{^Block=Malayalam}', "");
Expect(1, 3456, '\P{Block=Malayalam}', "");
Expect(0, 3456, '\P{^Block=Malayalam}', "");
Error('\p{Blk: --Malayalam:=}');
Error('\P{Blk: --Malayalam:=}');
Expect(1, 3455, '\p{Blk=malayalam}', "");
Expect(0, 3455, '\p{^Blk=malayalam}', "");
Expect(0, 3455, '\P{Blk=malayalam}', "");
Expect(1, 3455, '\P{^Blk=malayalam}', "");
Expect(0, 3456, '\p{Blk=malayalam}', "");
Expect(1, 3456, '\p{^Blk=malayalam}', "");
Expect(1, 3456, '\P{Blk=malayalam}', "");
Expect(0, 3456, '\P{^Blk=malayalam}', "");
Expect(1, 3455, '\p{Blk=_-Malayalam}', "");
Expect(0, 3455, '\p{^Blk=_-Malayalam}', "");
Expect(0, 3455, '\P{Blk=_-Malayalam}', "");
Expect(1, 3455, '\P{^Blk=_-Malayalam}', "");
Expect(0, 3456, '\p{Blk=_-Malayalam}', "");
Expect(1, 3456, '\p{^Blk=_-Malayalam}', "");
Expect(1, 3456, '\P{Blk=_-Malayalam}', "");
Expect(0, 3456, '\P{^Blk=_-Malayalam}', "");
Error('\p{Is_Block=/a/ malayalam}');
Error('\P{Is_Block=/a/ malayalam}');
Expect(1, 3455, '\p{Is_Block=malayalam}', "");
Expect(0, 3455, '\p{^Is_Block=malayalam}', "");
Expect(0, 3455, '\P{Is_Block=malayalam}', "");
Expect(1, 3455, '\P{^Is_Block=malayalam}', "");
Expect(0, 3456, '\p{Is_Block=malayalam}', "");
Expect(1, 3456, '\p{^Is_Block=malayalam}', "");
Expect(1, 3456, '\P{Is_Block=malayalam}', "");
Expect(0, 3456, '\P{^Is_Block=malayalam}', "");
Expect(1, 3455, '\p{Is_Block=--malayalam}', "");
Expect(0, 3455, '\p{^Is_Block=--malayalam}', "");
Expect(0, 3455, '\P{Is_Block=--malayalam}', "");
Expect(1, 3455, '\P{^Is_Block=--malayalam}', "");
Expect(0, 3456, '\p{Is_Block=--malayalam}', "");
Expect(1, 3456, '\p{^Is_Block=--malayalam}', "");
Expect(1, 3456, '\P{Is_Block=--malayalam}', "");
Expect(0, 3456, '\P{^Is_Block=--malayalam}', "");
Error('\p{Is_Blk=-/a/Malayalam}');
Error('\P{Is_Blk=-/a/Malayalam}');
Expect(1, 3455, '\p{Is_Blk=malayalam}', "");
Expect(0, 3455, '\p{^Is_Blk=malayalam}', "");
Expect(0, 3455, '\P{Is_Blk=malayalam}', "");
Expect(1, 3455, '\P{^Is_Blk=malayalam}', "");
Expect(0, 3456, '\p{Is_Blk=malayalam}', "");
Expect(1, 3456, '\p{^Is_Blk=malayalam}', "");
Expect(1, 3456, '\P{Is_Blk=malayalam}', "");
Expect(0, 3456, '\P{^Is_Blk=malayalam}', "");
Expect(1, 3455, '\p{Is_Blk=	Malayalam}', "");
Expect(0, 3455, '\p{^Is_Blk=	Malayalam}', "");
Expect(0, 3455, '\P{Is_Blk=	Malayalam}', "");
Expect(1, 3455, '\P{^Is_Blk=	Malayalam}', "");
Expect(0, 3456, '\p{Is_Blk=	Malayalam}', "");
Expect(1, 3456, '\p{^Is_Blk=	Malayalam}', "");
Expect(1, 3456, '\P{Is_Blk=	Malayalam}', "");
Expect(0, 3456, '\P{^Is_Blk=	Malayalam}', "");
Error('\p{Block= _tai_Le:=}');
Error('\P{Block= _tai_Le:=}');
Expect(1, 6527, '\p{Block=taile}', "");
Expect(0, 6527, '\p{^Block=taile}', "");
Expect(0, 6527, '\P{Block=taile}', "");
Expect(1, 6527, '\P{^Block=taile}', "");
Expect(0, 6528, '\p{Block=taile}', "");
Expect(1, 6528, '\p{^Block=taile}', "");
Expect(1, 6528, '\P{Block=taile}', "");
Expect(0, 6528, '\P{^Block=taile}', "");
Expect(1, 6527, '\p{Block=_	Tai_Le}', "");
Expect(0, 6527, '\p{^Block=_	Tai_Le}', "");
Expect(0, 6527, '\P{Block=_	Tai_Le}', "");
Expect(1, 6527, '\P{^Block=_	Tai_Le}', "");
Expect(0, 6528, '\p{Block=_	Tai_Le}', "");
Expect(1, 6528, '\p{^Block=_	Tai_Le}', "");
Expect(1, 6528, '\P{Block=_	Tai_Le}', "");
Expect(0, 6528, '\P{^Block=_	Tai_Le}', "");
Error('\p{Blk:   :=-TAI_le}');
Error('\P{Blk:   :=-TAI_le}');
Expect(1, 6527, '\p{Blk=taile}', "");
Expect(0, 6527, '\p{^Blk=taile}', "");
Expect(0, 6527, '\P{Blk=taile}', "");
Expect(1, 6527, '\P{^Blk=taile}', "");
Expect(0, 6528, '\p{Blk=taile}', "");
Expect(1, 6528, '\p{^Blk=taile}', "");
Expect(1, 6528, '\P{Blk=taile}', "");
Expect(0, 6528, '\P{^Blk=taile}', "");
Expect(1, 6527, '\p{Blk= -Tai_Le}', "");
Expect(0, 6527, '\p{^Blk= -Tai_Le}', "");
Expect(0, 6527, '\P{Blk= -Tai_Le}', "");
Expect(1, 6527, '\P{^Blk= -Tai_Le}', "");
Expect(0, 6528, '\p{Blk= -Tai_Le}', "");
Expect(1, 6528, '\p{^Blk= -Tai_Le}', "");
Expect(1, 6528, '\P{Blk= -Tai_Le}', "");
Expect(0, 6528, '\P{^Blk= -Tai_Le}', "");
Error('\p{Is_Block:	 :=TAI_le}');
Error('\P{Is_Block:	 :=TAI_le}');
Expect(1, 6527, '\p{Is_Block=taile}', "");
Expect(0, 6527, '\p{^Is_Block=taile}', "");
Expect(0, 6527, '\P{Is_Block=taile}', "");
Expect(1, 6527, '\P{^Is_Block=taile}', "");
Expect(0, 6528, '\p{Is_Block=taile}', "");
Expect(1, 6528, '\p{^Is_Block=taile}', "");
Expect(1, 6528, '\P{Is_Block=taile}', "");
Expect(0, 6528, '\P{^Is_Block=taile}', "");
Expect(1, 6527, '\p{Is_Block=_-Tai_Le}', "");
Expect(0, 6527, '\p{^Is_Block=_-Tai_Le}', "");
Expect(0, 6527, '\P{Is_Block=_-Tai_Le}', "");
Expect(1, 6527, '\P{^Is_Block=_-Tai_Le}', "");
Expect(0, 6528, '\p{Is_Block=_-Tai_Le}', "");
Expect(1, 6528, '\p{^Is_Block=_-Tai_Le}', "");
Expect(1, 6528, '\P{Is_Block=_-Tai_Le}', "");
Expect(0, 6528, '\P{^Is_Block=_-Tai_Le}', "");
Error('\p{Is_Blk= _Tai_LE:=}');
Error('\P{Is_Blk= _Tai_LE:=}');
Expect(1, 6527, '\p{Is_Blk=taile}', "");
Expect(0, 6527, '\p{^Is_Blk=taile}', "");
Expect(0, 6527, '\P{Is_Blk=taile}', "");
Expect(1, 6527, '\P{^Is_Blk=taile}', "");
Expect(0, 6528, '\p{Is_Blk=taile}', "");
Expect(1, 6528, '\p{^Is_Blk=taile}', "");
Expect(1, 6528, '\P{Is_Blk=taile}', "");
Expect(0, 6528, '\P{^Is_Blk=taile}', "");
Expect(1, 6527, '\p{Is_Blk=	 Tai_Le}', "");
Expect(0, 6527, '\p{^Is_Blk=	 Tai_Le}', "");
Expect(0, 6527, '\P{Is_Blk=	 Tai_Le}', "");
Expect(1, 6527, '\P{^Is_Blk=	 Tai_Le}', "");
Expect(0, 6528, '\p{Is_Blk=	 Tai_Le}', "");
Expect(1, 6528, '\p{^Is_Blk=	 Tai_Le}', "");
Expect(1, 6528, '\P{Is_Blk=	 Tai_Le}', "");
Expect(0, 6528, '\P{^Is_Blk=	 Tai_Le}', "");
Error('\p{Block=_:=BRAHMI}');
Error('\P{Block=_:=BRAHMI}');
Expect(1, 69759, '\p{Block=brahmi}', "");
Expect(0, 69759, '\p{^Block=brahmi}', "");
Expect(0, 69759, '\P{Block=brahmi}', "");
Expect(1, 69759, '\P{^Block=brahmi}', "");
Expect(0, 69760, '\p{Block=brahmi}', "");
Expect(1, 69760, '\p{^Block=brahmi}', "");
Expect(1, 69760, '\P{Block=brahmi}', "");
Expect(0, 69760, '\P{^Block=brahmi}', "");
Expect(1, 69759, '\p{Block=-Brahmi}', "");
Expect(0, 69759, '\p{^Block=-Brahmi}', "");
Expect(0, 69759, '\P{Block=-Brahmi}', "");
Expect(1, 69759, '\P{^Block=-Brahmi}', "");
Expect(0, 69760, '\p{Block=-Brahmi}', "");
Expect(1, 69760, '\p{^Block=-Brahmi}', "");
Expect(1, 69760, '\P{Block=-Brahmi}', "");
Expect(0, 69760, '\P{^Block=-Brahmi}', "");
Error('\p{Blk=/a/ Brahmi}');
Error('\P{Blk=/a/ Brahmi}');
Expect(1, 69759, '\p{Blk=brahmi}', "");
Expect(0, 69759, '\p{^Blk=brahmi}', "");
Expect(0, 69759, '\P{Blk=brahmi}', "");
Expect(1, 69759, '\P{^Blk=brahmi}', "");
Expect(0, 69760, '\p{Blk=brahmi}', "");
Expect(1, 69760, '\p{^Blk=brahmi}', "");
Expect(1, 69760, '\P{Blk=brahmi}', "");
Expect(0, 69760, '\P{^Blk=brahmi}', "");
Expect(1, 69759, '\p{Blk=_BRAHMI}', "");
Expect(0, 69759, '\p{^Blk=_BRAHMI}', "");
Expect(0, 69759, '\P{Blk=_BRAHMI}', "");
Expect(1, 69759, '\P{^Blk=_BRAHMI}', "");
Expect(0, 69760, '\p{Blk=_BRAHMI}', "");
Expect(1, 69760, '\p{^Blk=_BRAHMI}', "");
Expect(1, 69760, '\P{Blk=_BRAHMI}', "");
Expect(0, 69760, '\P{^Blk=_BRAHMI}', "");
Error('\p{Is_Block=_:=Brahmi}');
Error('\P{Is_Block=_:=Brahmi}');
Expect(1, 69759, '\p{Is_Block: brahmi}', "");
Expect(0, 69759, '\p{^Is_Block: brahmi}', "");
Expect(0, 69759, '\P{Is_Block: brahmi}', "");
Expect(1, 69759, '\P{^Is_Block: brahmi}', "");
Expect(0, 69760, '\p{Is_Block: brahmi}', "");
Expect(1, 69760, '\p{^Is_Block: brahmi}', "");
Expect(1, 69760, '\P{Is_Block: brahmi}', "");
Expect(0, 69760, '\P{^Is_Block: brahmi}', "");
Expect(1, 69759, '\p{Is_Block=-	BRAHMI}', "");
Expect(0, 69759, '\p{^Is_Block=-	BRAHMI}', "");
Expect(0, 69759, '\P{Is_Block=-	BRAHMI}', "");
Expect(1, 69759, '\P{^Is_Block=-	BRAHMI}', "");
Expect(0, 69760, '\p{Is_Block=-	BRAHMI}', "");
Expect(1, 69760, '\p{^Is_Block=-	BRAHMI}', "");
Expect(1, 69760, '\P{Is_Block=-	BRAHMI}', "");
Expect(0, 69760, '\P{^Is_Block=-	BRAHMI}', "");
Error('\p{Is_Blk=/a/_BRAHMI}');
Error('\P{Is_Blk=/a/_BRAHMI}');
Expect(1, 69759, '\p{Is_Blk=brahmi}', "");
Expect(0, 69759, '\p{^Is_Blk=brahmi}', "");
Expect(0, 69759, '\P{Is_Blk=brahmi}', "");
Expect(1, 69759, '\P{^Is_Blk=brahmi}', "");
Expect(0, 69760, '\p{Is_Blk=brahmi}', "");
Expect(1, 69760, '\p{^Is_Blk=brahmi}', "");
Expect(1, 69760, '\P{Is_Blk=brahmi}', "");
Expect(0, 69760, '\P{^Is_Blk=brahmi}', "");
Expect(1, 69759, '\p{Is_Blk: 	Brahmi}', "");
Expect(0, 69759, '\p{^Is_Blk: 	Brahmi}', "");
Expect(0, 69759, '\P{Is_Blk: 	Brahmi}', "");
Expect(1, 69759, '\P{^Is_Blk: 	Brahmi}', "");
Expect(0, 69760, '\p{Is_Blk: 	Brahmi}', "");
Expect(1, 69760, '\p{^Is_Blk: 	Brahmi}', "");
Expect(1, 69760, '\P{Is_Blk: 	Brahmi}', "");
Expect(0, 69760, '\P{^Is_Blk: 	Brahmi}', "");
Error('\p{Block=-	Hanunoo/a/}');
Error('\P{Block=-	Hanunoo/a/}');
Expect(1, 5951, '\p{Block=hanunoo}', "");
Expect(0, 5951, '\p{^Block=hanunoo}', "");
Expect(0, 5951, '\P{Block=hanunoo}', "");
Expect(1, 5951, '\P{^Block=hanunoo}', "");
Expect(0, 5952, '\p{Block=hanunoo}', "");
Expect(1, 5952, '\p{^Block=hanunoo}', "");
Expect(1, 5952, '\P{Block=hanunoo}', "");
Expect(0, 5952, '\P{^Block=hanunoo}', "");
Expect(1, 5951, '\p{Block=_hanunoo}', "");
Expect(0, 5951, '\p{^Block=_hanunoo}', "");
Expect(0, 5951, '\P{Block=_hanunoo}', "");
Expect(1, 5951, '\P{^Block=_hanunoo}', "");
Expect(0, 5952, '\p{Block=_hanunoo}', "");
Expect(1, 5952, '\p{^Block=_hanunoo}', "");
Expect(1, 5952, '\P{Block=_hanunoo}', "");
Expect(0, 5952, '\P{^Block=_hanunoo}', "");
Error('\p{Blk=:=Hanunoo}');
Error('\P{Blk=:=Hanunoo}');
Expect(1, 5951, '\p{Blk=hanunoo}', "");
Expect(0, 5951, '\p{^Blk=hanunoo}', "");
Expect(0, 5951, '\P{Blk=hanunoo}', "");
Expect(1, 5951, '\P{^Blk=hanunoo}', "");
Expect(0, 5952, '\p{Blk=hanunoo}', "");
Expect(1, 5952, '\p{^Blk=hanunoo}', "");
Expect(1, 5952, '\P{Blk=hanunoo}', "");
Expect(0, 5952, '\P{^Blk=hanunoo}', "");
Expect(1, 5951, '\p{Blk=- HANUNOO}', "");
Expect(0, 5951, '\p{^Blk=- HANUNOO}', "");
Expect(0, 5951, '\P{Blk=- HANUNOO}', "");
Expect(1, 5951, '\P{^Blk=- HANUNOO}', "");
Expect(0, 5952, '\p{Blk=- HANUNOO}', "");
Expect(1, 5952, '\p{^Blk=- HANUNOO}', "");
Expect(1, 5952, '\P{Blk=- HANUNOO}', "");
Expect(0, 5952, '\P{^Blk=- HANUNOO}', "");
Error('\p{Is_Block=	:=HANUNOO}');
Error('\P{Is_Block=	:=HANUNOO}');
Expect(1, 5951, '\p{Is_Block=hanunoo}', "");
Expect(0, 5951, '\p{^Is_Block=hanunoo}', "");
Expect(0, 5951, '\P{Is_Block=hanunoo}', "");
Expect(1, 5951, '\P{^Is_Block=hanunoo}', "");
Expect(0, 5952, '\p{Is_Block=hanunoo}', "");
Expect(1, 5952, '\p{^Is_Block=hanunoo}', "");
Expect(1, 5952, '\P{Is_Block=hanunoo}', "");
Expect(0, 5952, '\P{^Is_Block=hanunoo}', "");
Expect(1, 5951, '\p{Is_Block=	HANUNOO}', "");
Expect(0, 5951, '\p{^Is_Block=	HANUNOO}', "");
Expect(0, 5951, '\P{Is_Block=	HANUNOO}', "");
Expect(1, 5951, '\P{^Is_Block=	HANUNOO}', "");
Expect(0, 5952, '\p{Is_Block=	HANUNOO}', "");
Expect(1, 5952, '\p{^Is_Block=	HANUNOO}', "");
Expect(1, 5952, '\P{Is_Block=	HANUNOO}', "");
Expect(0, 5952, '\P{^Is_Block=	HANUNOO}', "");
Error('\p{Is_Blk=_-Hanunoo/a/}');
Error('\P{Is_Blk=_-Hanunoo/a/}');
Expect(1, 5951, '\p{Is_Blk=hanunoo}', "");
Expect(0, 5951, '\p{^Is_Blk=hanunoo}', "");
Expect(0, 5951, '\P{Is_Blk=hanunoo}', "");
Expect(1, 5951, '\P{^Is_Blk=hanunoo}', "");
Expect(0, 5952, '\p{Is_Blk=hanunoo}', "");
Expect(1, 5952, '\p{^Is_Blk=hanunoo}', "");
Expect(1, 5952, '\P{Is_Blk=hanunoo}', "");
Expect(0, 5952, '\P{^Is_Blk=hanunoo}', "");
Expect(1, 5951, '\p{Is_Blk=Hanunoo}', "");
Expect(0, 5951, '\p{^Is_Blk=Hanunoo}', "");
Expect(0, 5951, '\P{Is_Blk=Hanunoo}', "");
Expect(1, 5951, '\P{^Is_Blk=Hanunoo}', "");
Expect(0, 5952, '\p{Is_Blk=Hanunoo}', "");
Expect(1, 5952, '\p{^Is_Blk=Hanunoo}', "");
Expect(1, 5952, '\P{Is_Blk=Hanunoo}', "");
Expect(0, 5952, '\P{^Is_Blk=Hanunoo}', "");
Error('\p{Block=-:=High_Surrogates}');
Error('\P{Block=-:=High_Surrogates}');
Expect(1, 56191, '\p{Block:   highsurrogates}', "");
Expect(0, 56191, '\p{^Block:   highsurrogates}', "");
Expect(0, 56191, '\P{Block:   highsurrogates}', "");
Expect(1, 56191, '\P{^Block:   highsurrogates}', "");
Expect(0, 57344, '\p{Block:   highsurrogates}', "");
Expect(1, 57344, '\p{^Block:   highsurrogates}', "");
Expect(1, 57344, '\P{Block:   highsurrogates}', "");
Expect(0, 57344, '\P{^Block:   highsurrogates}', "");
Expect(1, 56191, '\p{Block=High_Surrogates}', "");
Expect(0, 56191, '\p{^Block=High_Surrogates}', "");
Expect(0, 56191, '\P{Block=High_Surrogates}', "");
Expect(1, 56191, '\P{^Block=High_Surrogates}', "");
Expect(0, 57344, '\p{Block=High_Surrogates}', "");
Expect(1, 57344, '\p{^Block=High_Surrogates}', "");
Expect(1, 57344, '\P{Block=High_Surrogates}', "");
Expect(0, 57344, '\P{^Block=High_Surrogates}', "");
Error('\p{Blk= /a/High_Surrogates}');
Error('\P{Blk= /a/High_Surrogates}');
Expect(1, 56191, '\p{Blk=highsurrogates}', "");
Expect(0, 56191, '\p{^Blk=highsurrogates}', "");
Expect(0, 56191, '\P{Blk=highsurrogates}', "");
Expect(1, 56191, '\P{^Blk=highsurrogates}', "");
Expect(0, 57344, '\p{Blk=highsurrogates}', "");
Expect(1, 57344, '\p{^Blk=highsurrogates}', "");
Expect(1, 57344, '\P{Blk=highsurrogates}', "");
Expect(0, 57344, '\P{^Blk=highsurrogates}', "");
Expect(1, 56191, '\p{Blk:	high_SURROGATES}', "");
Expect(0, 56191, '\p{^Blk:	high_SURROGATES}', "");
Expect(0, 56191, '\P{Blk:	high_SURROGATES}', "");
Expect(1, 56191, '\P{^Blk:	high_SURROGATES}', "");
Expect(0, 57344, '\p{Blk:	high_SURROGATES}', "");
Expect(1, 57344, '\p{^Blk:	high_SURROGATES}', "");
Expect(1, 57344, '\P{Blk:	high_SURROGATES}', "");
Expect(0, 57344, '\P{^Blk:	high_SURROGATES}', "");
Error('\p{Is_Block= /a/high_surrogates}');
Error('\P{Is_Block= /a/high_surrogates}');
Expect(1, 56191, '\p{Is_Block=highsurrogates}', "");
Expect(0, 56191, '\p{^Is_Block=highsurrogates}', "");
Expect(0, 56191, '\P{Is_Block=highsurrogates}', "");
Expect(1, 56191, '\P{^Is_Block=highsurrogates}', "");
Expect(0, 57344, '\p{Is_Block=highsurrogates}', "");
Expect(1, 57344, '\p{^Is_Block=highsurrogates}', "");
Expect(1, 57344, '\P{Is_Block=highsurrogates}', "");
Expect(0, 57344, '\P{^Is_Block=highsurrogates}', "");
Expect(1, 56191, '\p{Is_Block:-High_Surrogates}', "");
Expect(0, 56191, '\p{^Is_Block:-High_Surrogates}', "");
Expect(0, 56191, '\P{Is_Block:-High_Surrogates}', "");
Expect(1, 56191, '\P{^Is_Block:-High_Surrogates}', "");
Expect(0, 57344, '\p{Is_Block:-High_Surrogates}', "");
Expect(1, 57344, '\p{^Is_Block:-High_Surrogates}', "");
Expect(1, 57344, '\P{Is_Block:-High_Surrogates}', "");
Expect(0, 57344, '\P{^Is_Block:-High_Surrogates}', "");
Error('\p{Is_Blk=-/a/HIGH_SURROGATES}');
Error('\P{Is_Blk=-/a/HIGH_SURROGATES}');
Expect(1, 56191, '\p{Is_Blk=highsurrogates}', "");
Expect(0, 56191, '\p{^Is_Blk=highsurrogates}', "");
Expect(0, 56191, '\P{Is_Blk=highsurrogates}', "");
Expect(1, 56191, '\P{^Is_Blk=highsurrogates}', "");
Expect(0, 57344, '\p{Is_Blk=highsurrogates}', "");
Expect(1, 57344, '\p{^Is_Blk=highsurrogates}', "");
Expect(1, 57344, '\P{Is_Blk=highsurrogates}', "");
Expect(0, 57344, '\P{^Is_Blk=highsurrogates}', "");
Expect(1, 56191, '\p{Is_Blk=- HIGH_Surrogates}', "");
Expect(0, 56191, '\p{^Is_Blk=- HIGH_Surrogates}', "");
Expect(0, 56191, '\P{Is_Blk=- HIGH_Surrogates}', "");
Expect(1, 56191, '\P{^Is_Blk=- HIGH_Surrogates}', "");
Expect(0, 57344, '\p{Is_Blk=- HIGH_Surrogates}', "");
Expect(1, 57344, '\p{^Is_Blk=- HIGH_Surrogates}', "");
Expect(1, 57344, '\P{Is_Blk=- HIGH_Surrogates}', "");
Expect(0, 57344, '\P{^Is_Blk=- HIGH_Surrogates}', "");
Error('\p{Block=:=-MYANMAR}');
Error('\P{Block=:=-MYANMAR}');
Expect(1, 4255, '\p{Block=myanmar}', "");
Expect(0, 4255, '\p{^Block=myanmar}', "");
Expect(0, 4255, '\P{Block=myanmar}', "");
Expect(1, 4255, '\P{^Block=myanmar}', "");
Expect(0, 4256, '\p{Block=myanmar}', "");
Expect(1, 4256, '\p{^Block=myanmar}', "");
Expect(1, 4256, '\P{Block=myanmar}', "");
Expect(0, 4256, '\P{^Block=myanmar}', "");
Expect(1, 4255, '\p{Block=-MYANMAR}', "");
Expect(0, 4255, '\p{^Block=-MYANMAR}', "");
Expect(0, 4255, '\P{Block=-MYANMAR}', "");
Expect(1, 4255, '\P{^Block=-MYANMAR}', "");
Expect(0, 4256, '\p{Block=-MYANMAR}', "");
Expect(1, 4256, '\p{^Block=-MYANMAR}', "");
Expect(1, 4256, '\P{Block=-MYANMAR}', "");
Expect(0, 4256, '\P{^Block=-MYANMAR}', "");
Error('\p{Blk=:=_	MYANMAR}');
Error('\P{Blk=:=_	MYANMAR}');
Expect(1, 4255, '\p{Blk=myanmar}', "");
Expect(0, 4255, '\p{^Blk=myanmar}', "");
Expect(0, 4255, '\P{Blk=myanmar}', "");
Expect(1, 4255, '\P{^Blk=myanmar}', "");
Expect(0, 4256, '\p{Blk=myanmar}', "");
Expect(1, 4256, '\p{^Blk=myanmar}', "");
Expect(1, 4256, '\P{Blk=myanmar}', "");
Expect(0, 4256, '\P{^Blk=myanmar}', "");
Expect(1, 4255, '\p{Blk=	Myanmar}', "");
Expect(0, 4255, '\p{^Blk=	Myanmar}', "");
Expect(0, 4255, '\P{Blk=	Myanmar}', "");
Expect(1, 4255, '\P{^Blk=	Myanmar}', "");
Expect(0, 4256, '\p{Blk=	Myanmar}', "");
Expect(1, 4256, '\p{^Blk=	Myanmar}', "");
Expect(1, 4256, '\P{Blk=	Myanmar}', "");
Expect(0, 4256, '\P{^Blk=	Myanmar}', "");
Error('\p{Is_Block:	_-MYANMAR:=}');
Error('\P{Is_Block:	_-MYANMAR:=}');
Expect(1, 4255, '\p{Is_Block=myanmar}', "");
Expect(0, 4255, '\p{^Is_Block=myanmar}', "");
Expect(0, 4255, '\P{Is_Block=myanmar}', "");
Expect(1, 4255, '\P{^Is_Block=myanmar}', "");
Expect(0, 4256, '\p{Is_Block=myanmar}', "");
Expect(1, 4256, '\p{^Is_Block=myanmar}', "");
Expect(1, 4256, '\P{Is_Block=myanmar}', "");
Expect(0, 4256, '\P{^Is_Block=myanmar}', "");
Expect(1, 4255, '\p{Is_Block=-	Myanmar}', "");
Expect(0, 4255, '\p{^Is_Block=-	Myanmar}', "");
Expect(0, 4255, '\P{Is_Block=-	Myanmar}', "");
Expect(1, 4255, '\P{^Is_Block=-	Myanmar}', "");
Expect(0, 4256, '\p{Is_Block=-	Myanmar}', "");
Expect(1, 4256, '\p{^Is_Block=-	Myanmar}', "");
Expect(1, 4256, '\P{Is_Block=-	Myanmar}', "");
Expect(0, 4256, '\P{^Is_Block=-	Myanmar}', "");
Error('\p{Is_Blk=/a/		MYANMAR}');
Error('\P{Is_Blk=/a/		MYANMAR}');
Expect(1, 4255, '\p{Is_Blk=myanmar}', "");
Expect(0, 4255, '\p{^Is_Blk=myanmar}', "");
Expect(0, 4255, '\P{Is_Blk=myanmar}', "");
Expect(1, 4255, '\P{^Is_Blk=myanmar}', "");
Expect(0, 4256, '\p{Is_Blk=myanmar}', "");
Expect(1, 4256, '\p{^Is_Blk=myanmar}', "");
Expect(1, 4256, '\P{Is_Blk=myanmar}', "");
Expect(0, 4256, '\P{^Is_Blk=myanmar}', "");
Expect(1, 4255, '\p{Is_Blk=	_MYANMAR}', "");
Expect(0, 4255, '\p{^Is_Blk=	_MYANMAR}', "");
Expect(0, 4255, '\P{Is_Blk=	_MYANMAR}', "");
Expect(1, 4255, '\P{^Is_Blk=	_MYANMAR}', "");
Expect(0, 4256, '\p{Is_Blk=	_MYANMAR}', "");
Expect(1, 4256, '\p{^Is_Blk=	_MYANMAR}', "");
Expect(1, 4256, '\P{Is_Blk=	_MYANMAR}', "");
Expect(0, 4256, '\P{^Is_Blk=	_MYANMAR}', "");
Error('\p{Block=/a/vertical_forms}');
Error('\P{Block=/a/vertical_forms}');
Expect(1, 65055, '\p{Block=verticalforms}', "");
Expect(0, 65055, '\p{^Block=verticalforms}', "");
Expect(0, 65055, '\P{Block=verticalforms}', "");
Expect(1, 65055, '\P{^Block=verticalforms}', "");
Expect(0, 65056, '\p{Block=verticalforms}', "");
Expect(1, 65056, '\p{^Block=verticalforms}', "");
Expect(1, 65056, '\P{Block=verticalforms}', "");
Expect(0, 65056, '\P{^Block=verticalforms}', "");
Expect(1, 65055, '\p{Block=- VERTICAL_Forms}', "");
Expect(0, 65055, '\p{^Block=- VERTICAL_Forms}', "");
Expect(0, 65055, '\P{Block=- VERTICAL_Forms}', "");
Expect(1, 65055, '\P{^Block=- VERTICAL_Forms}', "");
Expect(0, 65056, '\p{Block=- VERTICAL_Forms}', "");
Expect(1, 65056, '\p{^Block=- VERTICAL_Forms}', "");
Expect(1, 65056, '\P{Block=- VERTICAL_Forms}', "");
Expect(0, 65056, '\P{^Block=- VERTICAL_Forms}', "");
Error('\p{Blk=	:=Vertical_Forms}');
Error('\P{Blk=	:=Vertical_Forms}');
Expect(1, 65055, '\p{Blk=verticalforms}', "");
Expect(0, 65055, '\p{^Blk=verticalforms}', "");
Expect(0, 65055, '\P{Blk=verticalforms}', "");
Expect(1, 65055, '\P{^Blk=verticalforms}', "");
Expect(0, 65056, '\p{Blk=verticalforms}', "");
Expect(1, 65056, '\p{^Blk=verticalforms}', "");
Expect(1, 65056, '\P{Blk=verticalforms}', "");
Expect(0, 65056, '\P{^Blk=verticalforms}', "");
Expect(1, 65055, '\p{Blk:	  vertical_Forms}', "");
Expect(0, 65055, '\p{^Blk:	  vertical_Forms}', "");
Expect(0, 65055, '\P{Blk:	  vertical_Forms}', "");
Expect(1, 65055, '\P{^Blk:	  vertical_Forms}', "");
Expect(0, 65056, '\p{Blk:	  vertical_Forms}', "");
Expect(1, 65056, '\p{^Blk:	  vertical_Forms}', "");
Expect(1, 65056, '\P{Blk:	  vertical_Forms}', "");
Expect(0, 65056, '\P{^Blk:	  vertical_Forms}', "");
Error('\p{Is_Block=_ Vertical_Forms:=}');
Error('\P{Is_Block=_ Vertical_Forms:=}');
Expect(1, 65055, '\p{Is_Block=verticalforms}', "");
Expect(0, 65055, '\p{^Is_Block=verticalforms}', "");
Expect(0, 65055, '\P{Is_Block=verticalforms}', "");
Expect(1, 65055, '\P{^Is_Block=verticalforms}', "");
Expect(0, 65056, '\p{Is_Block=verticalforms}', "");
Expect(1, 65056, '\p{^Is_Block=verticalforms}', "");
Expect(1, 65056, '\P{Is_Block=verticalforms}', "");
Expect(0, 65056, '\P{^Is_Block=verticalforms}', "");
Expect(1, 65055, '\p{Is_Block= Vertical_forms}', "");
Expect(0, 65055, '\p{^Is_Block= Vertical_forms}', "");
Expect(0, 65055, '\P{Is_Block= Vertical_forms}', "");
Expect(1, 65055, '\P{^Is_Block= Vertical_forms}', "");
Expect(0, 65056, '\p{Is_Block= Vertical_forms}', "");
Expect(1, 65056, '\p{^Is_Block= Vertical_forms}', "");
Expect(1, 65056, '\P{Is_Block= Vertical_forms}', "");
Expect(0, 65056, '\P{^Is_Block= Vertical_forms}', "");
Error('\p{Is_Blk=-:=Vertical_Forms}');
Error('\P{Is_Blk=-:=Vertical_Forms}');
Expect(1, 65055, '\p{Is_Blk=verticalforms}', "");
Expect(0, 65055, '\p{^Is_Blk=verticalforms}', "");
Expect(0, 65055, '\P{Is_Blk=verticalforms}', "");
Expect(1, 65055, '\P{^Is_Blk=verticalforms}', "");
Expect(0, 65056, '\p{Is_Blk=verticalforms}', "");
Expect(1, 65056, '\p{^Is_Blk=verticalforms}', "");
Expect(1, 65056, '\P{Is_Blk=verticalforms}', "");
Expect(0, 65056, '\P{^Is_Blk=verticalforms}', "");
Expect(1, 65055, '\p{Is_Blk=_vertical_Forms}', "");
Expect(0, 65055, '\p{^Is_Blk=_vertical_Forms}', "");
Expect(0, 65055, '\P{Is_Blk=_vertical_Forms}', "");
Expect(1, 65055, '\P{^Is_Blk=_vertical_Forms}', "");
Expect(0, 65056, '\p{Is_Blk=_vertical_Forms}', "");
Expect(1, 65056, '\p{^Is_Blk=_vertical_Forms}', "");
Expect(1, 65056, '\P{Is_Blk=_vertical_Forms}', "");
Expect(0, 65056, '\P{^Is_Blk=_vertical_Forms}', "");
Error('\p{Block:   /a/	Kaithi}');
Error('\P{Block:   /a/	Kaithi}');
Expect(1, 69839, '\p{Block=kaithi}', "");
Expect(0, 69839, '\p{^Block=kaithi}', "");
Expect(0, 69839, '\P{Block=kaithi}', "");
Expect(1, 69839, '\P{^Block=kaithi}', "");
Expect(0, 69840, '\p{Block=kaithi}', "");
Expect(1, 69840, '\p{^Block=kaithi}', "");
Expect(1, 69840, '\P{Block=kaithi}', "");
Expect(0, 69840, '\P{^Block=kaithi}', "");
Expect(1, 69839, '\p{Block=	-KAITHI}', "");
Expect(0, 69839, '\p{^Block=	-KAITHI}', "");
Expect(0, 69839, '\P{Block=	-KAITHI}', "");
Expect(1, 69839, '\P{^Block=	-KAITHI}', "");
Expect(0, 69840, '\p{Block=	-KAITHI}', "");
Expect(1, 69840, '\p{^Block=	-KAITHI}', "");
Expect(1, 69840, '\P{Block=	-KAITHI}', "");
Expect(0, 69840, '\P{^Block=	-KAITHI}', "");
Error('\p{Blk= _KAITHI:=}');
Error('\P{Blk= _KAITHI:=}');
Expect(1, 69839, '\p{Blk=kaithi}', "");
Expect(0, 69839, '\p{^Blk=kaithi}', "");
Expect(0, 69839, '\P{Blk=kaithi}', "");
Expect(1, 69839, '\P{^Blk=kaithi}', "");
Expect(0, 69840, '\p{Blk=kaithi}', "");
Expect(1, 69840, '\p{^Blk=kaithi}', "");
Expect(1, 69840, '\P{Blk=kaithi}', "");
Expect(0, 69840, '\P{^Blk=kaithi}', "");
Expect(1, 69839, '\p{Blk= _kaithi}', "");
Expect(0, 69839, '\p{^Blk= _kaithi}', "");
Expect(0, 69839, '\P{Blk= _kaithi}', "");
Expect(1, 69839, '\P{^Blk= _kaithi}', "");
Expect(0, 69840, '\p{Blk= _kaithi}', "");
Expect(1, 69840, '\p{^Blk= _kaithi}', "");
Expect(1, 69840, '\P{Blk= _kaithi}', "");
Expect(0, 69840, '\P{^Blk= _kaithi}', "");
Error('\p{Is_Block= /a/kaithi}');
Error('\P{Is_Block= /a/kaithi}');
Expect(1, 69839, '\p{Is_Block=kaithi}', "");
Expect(0, 69839, '\p{^Is_Block=kaithi}', "");
Expect(0, 69839, '\P{Is_Block=kaithi}', "");
Expect(1, 69839, '\P{^Is_Block=kaithi}', "");
Expect(0, 69840, '\p{Is_Block=kaithi}', "");
Expect(1, 69840, '\p{^Is_Block=kaithi}', "");
Expect(1, 69840, '\P{Is_Block=kaithi}', "");
Expect(0, 69840, '\P{^Is_Block=kaithi}', "");
Expect(1, 69839, '\p{Is_Block=Kaithi}', "");
Expect(0, 69839, '\p{^Is_Block=Kaithi}', "");
Expect(0, 69839, '\P{Is_Block=Kaithi}', "");
Expect(1, 69839, '\P{^Is_Block=Kaithi}', "");
Expect(0, 69840, '\p{Is_Block=Kaithi}', "");
Expect(1, 69840, '\p{^Is_Block=Kaithi}', "");
Expect(1, 69840, '\P{Is_Block=Kaithi}', "");
Expect(0, 69840, '\P{^Is_Block=Kaithi}', "");
Error('\p{Is_Blk=		Kaithi:=}');
Error('\P{Is_Blk=		Kaithi:=}');
Expect(1, 69839, '\p{Is_Blk=kaithi}', "");
Expect(0, 69839, '\p{^Is_Blk=kaithi}', "");
Expect(0, 69839, '\P{Is_Blk=kaithi}', "");
Expect(1, 69839, '\P{^Is_Blk=kaithi}', "");
Expect(0, 69840, '\p{Is_Blk=kaithi}', "");
Expect(1, 69840, '\p{^Is_Blk=kaithi}', "");
Expect(1, 69840, '\P{Is_Blk=kaithi}', "");
Expect(0, 69840, '\P{^Is_Blk=kaithi}', "");
Expect(1, 69839, '\p{Is_Blk=--Kaithi}', "");
Expect(0, 69839, '\p{^Is_Blk=--Kaithi}', "");
Expect(0, 69839, '\P{Is_Blk=--Kaithi}', "");
Expect(1, 69839, '\P{^Is_Blk=--Kaithi}', "");
Expect(0, 69840, '\p{Is_Blk=--Kaithi}', "");
Expect(1, 69840, '\p{^Is_Blk=--Kaithi}', "");
Expect(1, 69840, '\P{Is_Blk=--Kaithi}', "");
Expect(0, 69840, '\P{^Is_Blk=--Kaithi}', "");
Error('\p{Block=-/a/Tagalog}');
Error('\P{Block=-/a/Tagalog}');
Expect(1, 5919, '\p{Block=tagalog}', "");
Expect(0, 5919, '\p{^Block=tagalog}', "");
Expect(0, 5919, '\P{Block=tagalog}', "");
Expect(1, 5919, '\P{^Block=tagalog}', "");
Expect(0, 5920, '\p{Block=tagalog}', "");
Expect(1, 5920, '\p{^Block=tagalog}', "");
Expect(1, 5920, '\P{Block=tagalog}', "");
Expect(0, 5920, '\P{^Block=tagalog}', "");
Expect(1, 5919, '\p{Block:		Tagalog}', "");
Expect(0, 5919, '\p{^Block:		Tagalog}', "");
Expect(0, 5919, '\P{Block:		Tagalog}', "");
Expect(1, 5919, '\P{^Block:		Tagalog}', "");
Expect(0, 5920, '\p{Block:		Tagalog}', "");
Expect(1, 5920, '\p{^Block:		Tagalog}', "");
Expect(1, 5920, '\P{Block:		Tagalog}', "");
Expect(0, 5920, '\P{^Block:		Tagalog}', "");
Error('\p{Blk= /a/tagalog}');
Error('\P{Blk= /a/tagalog}');
Expect(1, 5919, '\p{Blk=tagalog}', "");
Expect(0, 5919, '\p{^Blk=tagalog}', "");
Expect(0, 5919, '\P{Blk=tagalog}', "");
Expect(1, 5919, '\P{^Blk=tagalog}', "");
Expect(0, 5920, '\p{Blk=tagalog}', "");
Expect(1, 5920, '\p{^Blk=tagalog}', "");
Expect(1, 5920, '\P{Blk=tagalog}', "");
Expect(0, 5920, '\P{^Blk=tagalog}', "");
Expect(1, 5919, '\p{Blk=__TAGALOG}', "");
Expect(0, 5919, '\p{^Blk=__TAGALOG}', "");
Expect(0, 5919, '\P{Blk=__TAGALOG}', "");
Expect(1, 5919, '\P{^Blk=__TAGALOG}', "");
Expect(0, 5920, '\p{Blk=__TAGALOG}', "");
Expect(1, 5920, '\p{^Blk=__TAGALOG}', "");
Expect(1, 5920, '\P{Blk=__TAGALOG}', "");
Expect(0, 5920, '\P{^Blk=__TAGALOG}', "");
Error('\p{Is_Block:   __Tagalog/a/}');
Error('\P{Is_Block:   __Tagalog/a/}');
Expect(1, 5919, '\p{Is_Block=tagalog}', "");
Expect(0, 5919, '\p{^Is_Block=tagalog}', "");
Expect(0, 5919, '\P{Is_Block=tagalog}', "");
Expect(1, 5919, '\P{^Is_Block=tagalog}', "");
Expect(0, 5920, '\p{Is_Block=tagalog}', "");
Expect(1, 5920, '\p{^Is_Block=tagalog}', "");
Expect(1, 5920, '\P{Is_Block=tagalog}', "");
Expect(0, 5920, '\P{^Is_Block=tagalog}', "");
Expect(1, 5919, '\p{Is_Block=_tagalog}', "");
Expect(0, 5919, '\p{^Is_Block=_tagalog}', "");
Expect(0, 5919, '\P{Is_Block=_tagalog}', "");
Expect(1, 5919, '\P{^Is_Block=_tagalog}', "");
Expect(0, 5920, '\p{Is_Block=_tagalog}', "");
Expect(1, 5920, '\p{^Is_Block=_tagalog}', "");
Expect(1, 5920, '\P{Is_Block=_tagalog}', "");
Expect(0, 5920, '\P{^Is_Block=_tagalog}', "");
Error('\p{Is_Blk= TAGALOG/a/}');
Error('\P{Is_Blk= TAGALOG/a/}');
Expect(1, 5919, '\p{Is_Blk=tagalog}', "");
Expect(0, 5919, '\p{^Is_Blk=tagalog}', "");
Expect(0, 5919, '\P{Is_Blk=tagalog}', "");
Expect(1, 5919, '\P{^Is_Blk=tagalog}', "");
Expect(0, 5920, '\p{Is_Blk=tagalog}', "");
Expect(1, 5920, '\p{^Is_Blk=tagalog}', "");
Expect(1, 5920, '\P{Is_Blk=tagalog}', "");
Expect(0, 5920, '\P{^Is_Blk=tagalog}', "");
Expect(1, 5919, '\p{Is_Blk=		tagalog}', "");
Expect(0, 5919, '\p{^Is_Blk=		tagalog}', "");
Expect(0, 5919, '\P{Is_Blk=		tagalog}', "");
Expect(1, 5919, '\P{^Is_Blk=		tagalog}', "");
Expect(0, 5920, '\p{Is_Blk=		tagalog}', "");
Expect(1, 5920, '\p{^Is_Blk=		tagalog}', "");
Expect(1, 5920, '\P{Is_Blk=		tagalog}', "");
Expect(0, 5920, '\P{^Is_Blk=		tagalog}', "");
Error('\p{Block=:= 	Miao}');
Error('\P{Block=:= 	Miao}');
Expect(1, 94111, '\p{Block=miao}', "");
Expect(0, 94111, '\p{^Block=miao}', "");
Expect(0, 94111, '\P{Block=miao}', "");
Expect(1, 94111, '\P{^Block=miao}', "");
Expect(0, 94112, '\p{Block=miao}', "");
Expect(1, 94112, '\p{^Block=miao}', "");
Expect(1, 94112, '\P{Block=miao}', "");
Expect(0, 94112, '\P{^Block=miao}', "");
Expect(1, 94111, '\p{Block=_miao}', "");
Expect(0, 94111, '\p{^Block=_miao}', "");
Expect(0, 94111, '\P{Block=_miao}', "");
Expect(1, 94111, '\P{^Block=_miao}', "");
Expect(0, 94112, '\p{Block=_miao}', "");
Expect(1, 94112, '\p{^Block=_miao}', "");
Expect(1, 94112, '\P{Block=_miao}', "");
Expect(0, 94112, '\P{^Block=_miao}', "");
Error('\p{Blk=	/a/Miao}');
Error('\P{Blk=	/a/Miao}');
Expect(1, 94111, '\p{Blk=miao}', "");
Expect(0, 94111, '\p{^Blk=miao}', "");
Expect(0, 94111, '\P{Blk=miao}', "");
Expect(1, 94111, '\P{^Blk=miao}', "");
Expect(0, 94112, '\p{Blk=miao}', "");
Expect(1, 94112, '\p{^Blk=miao}', "");
Expect(1, 94112, '\P{Blk=miao}', "");
Expect(0, 94112, '\P{^Blk=miao}', "");
Expect(1, 94111, '\p{Blk=--miao}', "");
Expect(0, 94111, '\p{^Blk=--miao}', "");
Expect(0, 94111, '\P{Blk=--miao}', "");
Expect(1, 94111, '\P{^Blk=--miao}', "");
Expect(0, 94112, '\p{Blk=--miao}', "");
Expect(1, 94112, '\p{^Blk=--miao}', "");
Expect(1, 94112, '\P{Blk=--miao}', "");
Expect(0, 94112, '\P{^Blk=--miao}', "");
Error('\p{Is_Block=:=miao}');
Error('\P{Is_Block=:=miao}');
Expect(1, 94111, '\p{Is_Block=miao}', "");
Expect(0, 94111, '\p{^Is_Block=miao}', "");
Expect(0, 94111, '\P{Is_Block=miao}', "");
Expect(1, 94111, '\P{^Is_Block=miao}', "");
Expect(0, 94112, '\p{Is_Block=miao}', "");
Expect(1, 94112, '\p{^Is_Block=miao}', "");
Expect(1, 94112, '\P{Is_Block=miao}', "");
Expect(0, 94112, '\P{^Is_Block=miao}', "");
Expect(1, 94111, '\p{Is_Block= 	Miao}', "");
Expect(0, 94111, '\p{^Is_Block= 	Miao}', "");
Expect(0, 94111, '\P{Is_Block= 	Miao}', "");
Expect(1, 94111, '\P{^Is_Block= 	Miao}', "");
Expect(0, 94112, '\p{Is_Block= 	Miao}', "");
Expect(1, 94112, '\p{^Is_Block= 	Miao}', "");
Expect(1, 94112, '\P{Is_Block= 	Miao}', "");
Expect(0, 94112, '\P{^Is_Block= 	Miao}', "");
Error('\p{Is_Blk:	-:=MIAO}');
Error('\P{Is_Blk:	-:=MIAO}');
Expect(1, 94111, '\p{Is_Blk: miao}', "");
Expect(0, 94111, '\p{^Is_Blk: miao}', "");
Expect(0, 94111, '\P{Is_Blk: miao}', "");
Expect(1, 94111, '\P{^Is_Blk: miao}', "");
Expect(0, 94112, '\p{Is_Blk: miao}', "");
Expect(1, 94112, '\p{^Is_Blk: miao}', "");
Expect(1, 94112, '\P{Is_Blk: miao}', "");
Expect(0, 94112, '\P{^Is_Blk: miao}', "");
Expect(1, 94111, '\p{Is_Blk=-MIAO}', "");
Expect(0, 94111, '\p{^Is_Blk=-MIAO}', "");
Expect(0, 94111, '\P{Is_Blk=-MIAO}', "");
Expect(1, 94111, '\P{^Is_Blk=-MIAO}', "");
Expect(0, 94112, '\p{Is_Blk=-MIAO}', "");
Expect(1, 94112, '\p{^Is_Blk=-MIAO}', "");
Expect(1, 94112, '\P{Is_Blk=-MIAO}', "");
Expect(0, 94112, '\P{^Is_Blk=-MIAO}', "");
Error('\p{Block=_/a/Devanagari_extended}');
Error('\P{Block=_/a/Devanagari_extended}');
Expect(1, 43263, '\p{Block=devanagariextended}', "");
Expect(0, 43263, '\p{^Block=devanagariextended}', "");
Expect(0, 43263, '\P{Block=devanagariextended}', "");
Expect(1, 43263, '\P{^Block=devanagariextended}', "");
Expect(0, 43264, '\p{Block=devanagariextended}', "");
Expect(1, 43264, '\p{^Block=devanagariextended}', "");
Expect(1, 43264, '\P{Block=devanagariextended}', "");
Expect(0, 43264, '\P{^Block=devanagariextended}', "");
Expect(1, 43263, '\p{Block=-devanagari_Extended}', "");
Expect(0, 43263, '\p{^Block=-devanagari_Extended}', "");
Expect(0, 43263, '\P{Block=-devanagari_Extended}', "");
Expect(1, 43263, '\P{^Block=-devanagari_Extended}', "");
Expect(0, 43264, '\p{Block=-devanagari_Extended}', "");
Expect(1, 43264, '\p{^Block=-devanagari_Extended}', "");
Expect(1, 43264, '\P{Block=-devanagari_Extended}', "");
Expect(0, 43264, '\P{^Block=-devanagari_Extended}', "");
Error('\p{Blk=:= DEVANAGARI_EXT}');
Error('\P{Blk=:= DEVANAGARI_EXT}');
Expect(1, 43263, '\p{Blk=devanagariext}', "");
Expect(0, 43263, '\p{^Blk=devanagariext}', "");
Expect(0, 43263, '\P{Blk=devanagariext}', "");
Expect(1, 43263, '\P{^Blk=devanagariext}', "");
Expect(0, 43264, '\p{Blk=devanagariext}', "");
Expect(1, 43264, '\p{^Blk=devanagariext}', "");
Expect(1, 43264, '\P{Blk=devanagariext}', "");
Expect(0, 43264, '\P{^Blk=devanagariext}', "");
Expect(1, 43263, '\p{Blk=_devanagari_EXT}', "");
Expect(0, 43263, '\p{^Blk=_devanagari_EXT}', "");
Expect(0, 43263, '\P{Blk=_devanagari_EXT}', "");
Expect(1, 43263, '\P{^Blk=_devanagari_EXT}', "");
Expect(0, 43264, '\p{Blk=_devanagari_EXT}', "");
Expect(1, 43264, '\p{^Blk=_devanagari_EXT}', "");
Expect(1, 43264, '\P{Blk=_devanagari_EXT}', "");
Expect(0, 43264, '\P{^Blk=_devanagari_EXT}', "");
Error('\p{Is_Block=:=Devanagari_EXTENDED}');
Error('\P{Is_Block=:=Devanagari_EXTENDED}');
Expect(1, 43263, '\p{Is_Block=devanagariextended}', "");
Expect(0, 43263, '\p{^Is_Block=devanagariextended}', "");
Expect(0, 43263, '\P{Is_Block=devanagariextended}', "");
Expect(1, 43263, '\P{^Is_Block=devanagariextended}', "");
Expect(0, 43264, '\p{Is_Block=devanagariextended}', "");
Expect(1, 43264, '\p{^Is_Block=devanagariextended}', "");
Expect(1, 43264, '\P{Is_Block=devanagariextended}', "");
Expect(0, 43264, '\P{^Is_Block=devanagariextended}', "");
Expect(1, 43263, '\p{Is_Block=	-Devanagari_EXTENDED}', "");
Expect(0, 43263, '\p{^Is_Block=	-Devanagari_EXTENDED}', "");
Expect(0, 43263, '\P{Is_Block=	-Devanagari_EXTENDED}', "");
Expect(1, 43263, '\P{^Is_Block=	-Devanagari_EXTENDED}', "");
Expect(0, 43264, '\p{Is_Block=	-Devanagari_EXTENDED}', "");
Expect(1, 43264, '\p{^Is_Block=	-Devanagari_EXTENDED}', "");
Expect(1, 43264, '\P{Is_Block=	-Devanagari_EXTENDED}', "");
Expect(0, 43264, '\P{^Is_Block=	-Devanagari_EXTENDED}', "");
Error('\p{Is_Blk=	_Devanagari_ext:=}');
Error('\P{Is_Blk=	_Devanagari_ext:=}');
Expect(1, 43263, '\p{Is_Blk=devanagariext}', "");
Expect(0, 43263, '\p{^Is_Blk=devanagariext}', "");
Expect(0, 43263, '\P{Is_Blk=devanagariext}', "");
Expect(1, 43263, '\P{^Is_Blk=devanagariext}', "");
Expect(0, 43264, '\p{Is_Blk=devanagariext}', "");
Expect(1, 43264, '\p{^Is_Blk=devanagariext}', "");
Expect(1, 43264, '\P{Is_Blk=devanagariext}', "");
Expect(0, 43264, '\P{^Is_Blk=devanagariext}', "");
Expect(1, 43263, '\p{Is_Blk=__Devanagari_Ext}', "");
Expect(0, 43263, '\p{^Is_Blk=__Devanagari_Ext}', "");
Expect(0, 43263, '\P{Is_Blk=__Devanagari_Ext}', "");
Expect(1, 43263, '\P{^Is_Blk=__Devanagari_Ext}', "");
Expect(0, 43264, '\p{Is_Blk=__Devanagari_Ext}', "");
Expect(1, 43264, '\p{^Is_Blk=__Devanagari_Ext}', "");
Expect(1, 43264, '\P{Is_Blk=__Devanagari_Ext}', "");
Expect(0, 43264, '\P{^Is_Blk=__Devanagari_Ext}', "");
Error('\p{Block=/a/ SUNDANESE}');
Error('\P{Block=/a/ SUNDANESE}');
Expect(1, 7103, '\p{Block=sundanese}', "");
Expect(0, 7103, '\p{^Block=sundanese}', "");
Expect(0, 7103, '\P{Block=sundanese}', "");
Expect(1, 7103, '\P{^Block=sundanese}', "");
Expect(0, 7104, '\p{Block=sundanese}', "");
Expect(1, 7104, '\p{^Block=sundanese}', "");
Expect(1, 7104, '\P{Block=sundanese}', "");
Expect(0, 7104, '\P{^Block=sundanese}', "");
Expect(1, 7103, '\p{Block=__Sundanese}', "");
Expect(0, 7103, '\p{^Block=__Sundanese}', "");
Expect(0, 7103, '\P{Block=__Sundanese}', "");
Expect(1, 7103, '\P{^Block=__Sundanese}', "");
Expect(0, 7104, '\p{Block=__Sundanese}', "");
Expect(1, 7104, '\p{^Block=__Sundanese}', "");
Expect(1, 7104, '\P{Block=__Sundanese}', "");
Expect(0, 7104, '\P{^Block=__Sundanese}', "");
Error('\p{Blk=	/a/Sundanese}');
Error('\P{Blk=	/a/Sundanese}');
Expect(1, 7103, '\p{Blk=sundanese}', "");
Expect(0, 7103, '\p{^Blk=sundanese}', "");
Expect(0, 7103, '\P{Blk=sundanese}', "");
Expect(1, 7103, '\P{^Blk=sundanese}', "");
Expect(0, 7104, '\p{Blk=sundanese}', "");
Expect(1, 7104, '\p{^Blk=sundanese}', "");
Expect(1, 7104, '\P{Blk=sundanese}', "");
Expect(0, 7104, '\P{^Blk=sundanese}', "");
Expect(1, 7103, '\p{Blk=_ Sundanese}', "");
Expect(0, 7103, '\p{^Blk=_ Sundanese}', "");
Expect(0, 7103, '\P{Blk=_ Sundanese}', "");
Expect(1, 7103, '\P{^Blk=_ Sundanese}', "");
Expect(0, 7104, '\p{Blk=_ Sundanese}', "");
Expect(1, 7104, '\p{^Blk=_ Sundanese}', "");
Expect(1, 7104, '\P{Blk=_ Sundanese}', "");
Expect(0, 7104, '\P{^Blk=_ Sundanese}', "");
Error('\p{Is_Block=		SUNDANESE:=}');
Error('\P{Is_Block=		SUNDANESE:=}');
Expect(1, 7103, '\p{Is_Block=sundanese}', "");
Expect(0, 7103, '\p{^Is_Block=sundanese}', "");
Expect(0, 7103, '\P{Is_Block=sundanese}', "");
Expect(1, 7103, '\P{^Is_Block=sundanese}', "");
Expect(0, 7104, '\p{Is_Block=sundanese}', "");
Expect(1, 7104, '\p{^Is_Block=sundanese}', "");
Expect(1, 7104, '\P{Is_Block=sundanese}', "");
Expect(0, 7104, '\P{^Is_Block=sundanese}', "");
Expect(1, 7103, '\p{Is_Block=	_Sundanese}', "");
Expect(0, 7103, '\p{^Is_Block=	_Sundanese}', "");
Expect(0, 7103, '\P{Is_Block=	_Sundanese}', "");
Expect(1, 7103, '\P{^Is_Block=	_Sundanese}', "");
Expect(0, 7104, '\p{Is_Block=	_Sundanese}', "");
Expect(1, 7104, '\p{^Is_Block=	_Sundanese}', "");
Expect(1, 7104, '\P{Is_Block=	_Sundanese}', "");
Expect(0, 7104, '\P{^Is_Block=	_Sundanese}', "");
Error('\p{Is_Blk=/a/_Sundanese}');
Error('\P{Is_Blk=/a/_Sundanese}');
Expect(1, 7103, '\p{Is_Blk=sundanese}', "");
Expect(0, 7103, '\p{^Is_Blk=sundanese}', "");
Expect(0, 7103, '\P{Is_Blk=sundanese}', "");
Expect(1, 7103, '\P{^Is_Blk=sundanese}', "");
Expect(0, 7104, '\p{Is_Blk=sundanese}', "");
Expect(1, 7104, '\p{^Is_Blk=sundanese}', "");
Expect(1, 7104, '\P{Is_Blk=sundanese}', "");
Expect(0, 7104, '\P{^Is_Blk=sundanese}', "");
Expect(1, 7103, '\p{Is_Blk=	-SUNDANESE}', "");
Expect(0, 7103, '\p{^Is_Blk=	-SUNDANESE}', "");
Expect(0, 7103, '\P{Is_Blk=	-SUNDANESE}', "");
Expect(1, 7103, '\P{^Is_Blk=	-SUNDANESE}', "");
Expect(0, 7104, '\p{Is_Blk=	-SUNDANESE}', "");
Expect(1, 7104, '\p{^Is_Blk=	-SUNDANESE}', "");
Expect(1, 7104, '\P{Is_Blk=	-SUNDANESE}', "");
Expect(0, 7104, '\P{^Is_Blk=	-SUNDANESE}', "");
Error('\p{Block=_:=Enclosed_Alphanumerics}');
Error('\P{Block=_:=Enclosed_Alphanumerics}');
Expect(1, 9471, '\p{Block=enclosedalphanumerics}', "");
Expect(0, 9471, '\p{^Block=enclosedalphanumerics}', "");
Expect(0, 9471, '\P{Block=enclosedalphanumerics}', "");
Expect(1, 9471, '\P{^Block=enclosedalphanumerics}', "");
Expect(0, 9472, '\p{Block=enclosedalphanumerics}', "");
Expect(1, 9472, '\p{^Block=enclosedalphanumerics}', "");
Expect(1, 9472, '\P{Block=enclosedalphanumerics}', "");
Expect(0, 9472, '\P{^Block=enclosedalphanumerics}', "");
Expect(1, 9471, '\p{Block=  Enclosed_Alphanumerics}', "");
Expect(0, 9471, '\p{^Block=  Enclosed_Alphanumerics}', "");
Expect(0, 9471, '\P{Block=  Enclosed_Alphanumerics}', "");
Expect(1, 9471, '\P{^Block=  Enclosed_Alphanumerics}', "");
Expect(0, 9472, '\p{Block=  Enclosed_Alphanumerics}', "");
Expect(1, 9472, '\p{^Block=  Enclosed_Alphanumerics}', "");
Expect(1, 9472, '\P{Block=  Enclosed_Alphanumerics}', "");
Expect(0, 9472, '\P{^Block=  Enclosed_Alphanumerics}', "");
Error('\p{Blk=/a/ _Enclosed_ALPHANUM}');
Error('\P{Blk=/a/ _Enclosed_ALPHANUM}');
Expect(1, 9471, '\p{Blk:   enclosedalphanum}', "");
Expect(0, 9471, '\p{^Blk:   enclosedalphanum}', "");
Expect(0, 9471, '\P{Blk:   enclosedalphanum}', "");
Expect(1, 9471, '\P{^Blk:   enclosedalphanum}', "");
Expect(0, 9472, '\p{Blk:   enclosedalphanum}', "");
Expect(1, 9472, '\p{^Blk:   enclosedalphanum}', "");
Expect(1, 9472, '\P{Blk:   enclosedalphanum}', "");
Expect(0, 9472, '\P{^Blk:   enclosedalphanum}', "");
Expect(1, 9471, '\p{Blk=	 Enclosed_ALPHANUM}', "");
Expect(0, 9471, '\p{^Blk=	 Enclosed_ALPHANUM}', "");
Expect(0, 9471, '\P{Blk=	 Enclosed_ALPHANUM}', "");
Expect(1, 9471, '\P{^Blk=	 Enclosed_ALPHANUM}', "");
Expect(0, 9472, '\p{Blk=	 Enclosed_ALPHANUM}', "");
Expect(1, 9472, '\p{^Blk=	 Enclosed_ALPHANUM}', "");
Expect(1, 9472, '\P{Blk=	 Enclosed_ALPHANUM}', "");
Expect(0, 9472, '\P{^Blk=	 Enclosed_ALPHANUM}', "");
Error('\p{Is_Block=_/a/Enclosed_alphanumerics}');
Error('\P{Is_Block=_/a/Enclosed_alphanumerics}');
Expect(1, 9471, '\p{Is_Block=enclosedalphanumerics}', "");
Expect(0, 9471, '\p{^Is_Block=enclosedalphanumerics}', "");
Expect(0, 9471, '\P{Is_Block=enclosedalphanumerics}', "");
Expect(1, 9471, '\P{^Is_Block=enclosedalphanumerics}', "");
Expect(0, 9472, '\p{Is_Block=enclosedalphanumerics}', "");
Expect(1, 9472, '\p{^Is_Block=enclosedalphanumerics}', "");
Expect(1, 9472, '\P{Is_Block=enclosedalphanumerics}', "");
Expect(0, 9472, '\P{^Is_Block=enclosedalphanumerics}', "");
Expect(1, 9471, '\p{Is_Block=_	ENCLOSED_Alphanumerics}', "");
Expect(0, 9471, '\p{^Is_Block=_	ENCLOSED_Alphanumerics}', "");
Expect(0, 9471, '\P{Is_Block=_	ENCLOSED_Alphanumerics}', "");
Expect(1, 9471, '\P{^Is_Block=_	ENCLOSED_Alphanumerics}', "");
Expect(0, 9472, '\p{Is_Block=_	ENCLOSED_Alphanumerics}', "");
Expect(1, 9472, '\p{^Is_Block=_	ENCLOSED_Alphanumerics}', "");
Expect(1, 9472, '\P{Is_Block=_	ENCLOSED_Alphanumerics}', "");
Expect(0, 9472, '\P{^Is_Block=_	ENCLOSED_Alphanumerics}', "");
Error('\p{Is_Blk=	_enclosed_Alphanum/a/}');
Error('\P{Is_Blk=	_enclosed_Alphanum/a/}');
Expect(1, 9471, '\p{Is_Blk:enclosedalphanum}', "");
Expect(0, 9471, '\p{^Is_Blk:enclosedalphanum}', "");
Expect(0, 9471, '\P{Is_Blk:enclosedalphanum}', "");
Expect(1, 9471, '\P{^Is_Blk:enclosedalphanum}', "");
Expect(0, 9472, '\p{Is_Blk:enclosedalphanum}', "");
Expect(1, 9472, '\p{^Is_Blk:enclosedalphanum}', "");
Expect(1, 9472, '\P{Is_Blk:enclosedalphanum}', "");
Expect(0, 9472, '\P{^Is_Blk:enclosedalphanum}', "");
Expect(1, 9471, '\p{Is_Blk= -Enclosed_alphanum}', "");
Expect(0, 9471, '\p{^Is_Blk= -Enclosed_alphanum}', "");
Expect(0, 9471, '\P{Is_Blk= -Enclosed_alphanum}', "");
Expect(1, 9471, '\P{^Is_Blk= -Enclosed_alphanum}', "");
Expect(0, 9472, '\p{Is_Blk= -Enclosed_alphanum}', "");
Expect(1, 9472, '\p{^Is_Blk= -Enclosed_alphanum}', "");
Expect(1, 9472, '\P{Is_Blk= -Enclosed_alphanum}', "");
Expect(0, 9472, '\P{^Is_Blk= -Enclosed_alphanum}', "");
Error('\p{Block=_:=Oriya}');
Error('\P{Block=_:=Oriya}');
Expect(1, 2943, '\p{Block=oriya}', "");
Expect(0, 2943, '\p{^Block=oriya}', "");
Expect(0, 2943, '\P{Block=oriya}', "");
Expect(1, 2943, '\P{^Block=oriya}', "");
Expect(0, 2944, '\p{Block=oriya}', "");
Expect(1, 2944, '\p{^Block=oriya}', "");
Expect(1, 2944, '\P{Block=oriya}', "");
Expect(0, 2944, '\P{^Block=oriya}', "");
Expect(1, 2943, '\p{Block=_ORIYA}', "");
Expect(0, 2943, '\p{^Block=_ORIYA}', "");
Expect(0, 2943, '\P{Block=_ORIYA}', "");
Expect(1, 2943, '\P{^Block=_ORIYA}', "");
Expect(0, 2944, '\p{Block=_ORIYA}', "");
Expect(1, 2944, '\p{^Block=_ORIYA}', "");
Expect(1, 2944, '\P{Block=_ORIYA}', "");
Expect(0, 2944, '\P{^Block=_ORIYA}', "");
Error('\p{Blk=:=_oriya}');
Error('\P{Blk=:=_oriya}');
Expect(1, 2943, '\p{Blk=oriya}', "");
Expect(0, 2943, '\p{^Blk=oriya}', "");
Expect(0, 2943, '\P{Blk=oriya}', "");
Expect(1, 2943, '\P{^Blk=oriya}', "");
Expect(0, 2944, '\p{Blk=oriya}', "");
Expect(1, 2944, '\p{^Blk=oriya}', "");
Expect(1, 2944, '\P{Blk=oriya}', "");
Expect(0, 2944, '\P{^Blk=oriya}', "");
Expect(1, 2943, '\p{Blk= -Oriya}', "");
Expect(0, 2943, '\p{^Blk= -Oriya}', "");
Expect(0, 2943, '\P{Blk= -Oriya}', "");
Expect(1, 2943, '\P{^Blk= -Oriya}', "");
Expect(0, 2944, '\p{Blk= -Oriya}', "");
Expect(1, 2944, '\p{^Blk= -Oriya}', "");
Expect(1, 2944, '\P{Blk= -Oriya}', "");
Expect(0, 2944, '\P{^Blk= -Oriya}', "");
Error('\p{Is_Block=/a/_oriya}');
Error('\P{Is_Block=/a/_oriya}');
Expect(1, 2943, '\p{Is_Block=oriya}', "");
Expect(0, 2943, '\p{^Is_Block=oriya}', "");
Expect(0, 2943, '\P{Is_Block=oriya}', "");
Expect(1, 2943, '\P{^Is_Block=oriya}', "");
Expect(0, 2944, '\p{Is_Block=oriya}', "");
Expect(1, 2944, '\p{^Is_Block=oriya}', "");
Expect(1, 2944, '\P{Is_Block=oriya}', "");
Expect(0, 2944, '\P{^Is_Block=oriya}', "");
Expect(1, 2943, '\p{Is_Block: 	Oriya}', "");
Expect(0, 2943, '\p{^Is_Block: 	Oriya}', "");
Expect(0, 2943, '\P{Is_Block: 	Oriya}', "");
Expect(1, 2943, '\P{^Is_Block: 	Oriya}', "");
Expect(0, 2944, '\p{Is_Block: 	Oriya}', "");
Expect(1, 2944, '\p{^Is_Block: 	Oriya}', "");
Expect(1, 2944, '\P{Is_Block: 	Oriya}', "");
Expect(0, 2944, '\P{^Is_Block: 	Oriya}', "");
Error('\p{Is_Blk=	_Oriya/a/}');
Error('\P{Is_Blk=	_Oriya/a/}');
Expect(1, 2943, '\p{Is_Blk=oriya}', "");
Expect(0, 2943, '\p{^Is_Blk=oriya}', "");
Expect(0, 2943, '\P{Is_Blk=oriya}', "");
Expect(1, 2943, '\P{^Is_Blk=oriya}', "");
Expect(0, 2944, '\p{Is_Blk=oriya}', "");
Expect(1, 2944, '\p{^Is_Blk=oriya}', "");
Expect(1, 2944, '\P{Is_Blk=oriya}', "");
Expect(0, 2944, '\P{^Is_Blk=oriya}', "");
Expect(1, 2943, '\p{Is_Blk=	oriya}', "");
Expect(0, 2943, '\p{^Is_Blk=	oriya}', "");
Expect(0, 2943, '\P{Is_Blk=	oriya}', "");
Expect(1, 2943, '\P{^Is_Blk=	oriya}', "");
Expect(0, 2944, '\p{Is_Blk=	oriya}', "");
Expect(1, 2944, '\p{^Is_Blk=	oriya}', "");
Expect(1, 2944, '\P{Is_Blk=	oriya}', "");
Expect(0, 2944, '\P{^Is_Blk=	oriya}', "");
Error('\p{Block= balinese:=}');
Error('\P{Block= balinese:=}');
Expect(1, 7039, '\p{Block=balinese}', "");
Expect(0, 7039, '\p{^Block=balinese}', "");
Expect(0, 7039, '\P{Block=balinese}', "");
Expect(1, 7039, '\P{^Block=balinese}', "");
Expect(0, 7040, '\p{Block=balinese}', "");
Expect(1, 7040, '\p{^Block=balinese}', "");
Expect(1, 7040, '\P{Block=balinese}', "");
Expect(0, 7040, '\P{^Block=balinese}', "");
Expect(1, 7039, '\p{Block=_balinese}', "");
Expect(0, 7039, '\p{^Block=_balinese}', "");
Expect(0, 7039, '\P{Block=_balinese}', "");
Expect(1, 7039, '\P{^Block=_balinese}', "");
Expect(0, 7040, '\p{Block=_balinese}', "");
Expect(1, 7040, '\p{^Block=_balinese}', "");
Expect(1, 7040, '\P{Block=_balinese}', "");
Expect(0, 7040, '\P{^Block=_balinese}', "");
Error('\p{Blk=:= _BALINESE}');
Error('\P{Blk=:= _BALINESE}');
Expect(1, 7039, '\p{Blk=balinese}', "");
Expect(0, 7039, '\p{^Blk=balinese}', "");
Expect(0, 7039, '\P{Blk=balinese}', "");
Expect(1, 7039, '\P{^Blk=balinese}', "");
Expect(0, 7040, '\p{Blk=balinese}', "");
Expect(1, 7040, '\p{^Blk=balinese}', "");
Expect(1, 7040, '\P{Blk=balinese}', "");
Expect(0, 7040, '\P{^Blk=balinese}', "");
Expect(1, 7039, '\p{Blk=-Balinese}', "");
Expect(0, 7039, '\p{^Blk=-Balinese}', "");
Expect(0, 7039, '\P{Blk=-Balinese}', "");
Expect(1, 7039, '\P{^Blk=-Balinese}', "");
Expect(0, 7040, '\p{Blk=-Balinese}', "");
Expect(1, 7040, '\p{^Blk=-Balinese}', "");
Expect(1, 7040, '\P{Blk=-Balinese}', "");
Expect(0, 7040, '\P{^Blk=-Balinese}', "");
Error('\p{Is_Block=/a/ Balinese}');
Error('\P{Is_Block=/a/ Balinese}');
Expect(1, 7039, '\p{Is_Block: balinese}', "");
Expect(0, 7039, '\p{^Is_Block: balinese}', "");
Expect(0, 7039, '\P{Is_Block: balinese}', "");
Expect(1, 7039, '\P{^Is_Block: balinese}', "");
Expect(0, 7040, '\p{Is_Block: balinese}', "");
Expect(1, 7040, '\p{^Is_Block: balinese}', "");
Expect(1, 7040, '\P{Is_Block: balinese}', "");
Expect(0, 7040, '\P{^Is_Block: balinese}', "");
Expect(1, 7039, '\p{Is_Block=	Balinese}', "");
Expect(0, 7039, '\p{^Is_Block=	Balinese}', "");
Expect(0, 7039, '\P{Is_Block=	Balinese}', "");
Expect(1, 7039, '\P{^Is_Block=	Balinese}', "");
Expect(0, 7040, '\p{Is_Block=	Balinese}', "");
Expect(1, 7040, '\p{^Is_Block=	Balinese}', "");
Expect(1, 7040, '\P{Is_Block=	Balinese}', "");
Expect(0, 7040, '\P{^Is_Block=	Balinese}', "");
Error('\p{Is_Blk:	_-balinese:=}');
Error('\P{Is_Blk:	_-balinese:=}');
Expect(1, 7039, '\p{Is_Blk:	balinese}', "");
Expect(0, 7039, '\p{^Is_Blk:	balinese}', "");
Expect(0, 7039, '\P{Is_Blk:	balinese}', "");
Expect(1, 7039, '\P{^Is_Blk:	balinese}', "");
Expect(0, 7040, '\p{Is_Blk:	balinese}', "");
Expect(1, 7040, '\p{^Is_Blk:	balinese}', "");
Expect(1, 7040, '\P{Is_Blk:	balinese}', "");
Expect(0, 7040, '\P{^Is_Blk:	balinese}', "");
Expect(1, 7039, '\p{Is_Blk=__Balinese}', "");
Expect(0, 7039, '\p{^Is_Blk=__Balinese}', "");
Expect(0, 7039, '\P{Is_Blk=__Balinese}', "");
Expect(1, 7039, '\P{^Is_Blk=__Balinese}', "");
Expect(0, 7040, '\p{Is_Blk=__Balinese}', "");
Expect(1, 7040, '\p{^Is_Blk=__Balinese}', "");
Expect(1, 7040, '\P{Is_Blk=__Balinese}', "");
Expect(0, 7040, '\P{^Is_Blk=__Balinese}', "");
Error('\p{Block: Georgian:=}');
Error('\P{Block: Georgian:=}');
Expect(1, 4351, '\p{Block:   georgian}', "");
Expect(0, 4351, '\p{^Block:   georgian}', "");
Expect(0, 4351, '\P{Block:   georgian}', "");
Expect(1, 4351, '\P{^Block:   georgian}', "");
Expect(0, 4352, '\p{Block:   georgian}', "");
Expect(1, 4352, '\p{^Block:   georgian}', "");
Expect(1, 4352, '\P{Block:   georgian}', "");
Expect(0, 4352, '\P{^Block:   georgian}', "");
Expect(1, 4351, '\p{Block: -georgian}', "");
Expect(0, 4351, '\p{^Block: -georgian}', "");
Expect(0, 4351, '\P{Block: -georgian}', "");
Expect(1, 4351, '\P{^Block: -georgian}', "");
Expect(0, 4352, '\p{Block: -georgian}', "");
Expect(1, 4352, '\p{^Block: -georgian}', "");
Expect(1, 4352, '\P{Block: -georgian}', "");
Expect(0, 4352, '\P{^Block: -georgian}', "");
Error('\p{Blk:   _ Georgian:=}');
Error('\P{Blk:   _ Georgian:=}');
Expect(1, 4351, '\p{Blk=georgian}', "");
Expect(0, 4351, '\p{^Blk=georgian}', "");
Expect(0, 4351, '\P{Blk=georgian}', "");
Expect(1, 4351, '\P{^Blk=georgian}', "");
Expect(0, 4352, '\p{Blk=georgian}', "");
Expect(1, 4352, '\p{^Blk=georgian}', "");
Expect(1, 4352, '\P{Blk=georgian}', "");
Expect(0, 4352, '\P{^Blk=georgian}', "");
Expect(1, 4351, '\p{Blk=_Georgian}', "");
Expect(0, 4351, '\p{^Blk=_Georgian}', "");
Expect(0, 4351, '\P{Blk=_Georgian}', "");
Expect(1, 4351, '\P{^Blk=_Georgian}', "");
Expect(0, 4352, '\p{Blk=_Georgian}', "");
Expect(1, 4352, '\p{^Blk=_Georgian}', "");
Expect(1, 4352, '\P{Blk=_Georgian}', "");
Expect(0, 4352, '\P{^Blk=_Georgian}', "");
Error('\p{Is_Block:	Georgian/a/}');
Error('\P{Is_Block:	Georgian/a/}');
Expect(1, 4351, '\p{Is_Block:	georgian}', "");
Expect(0, 4351, '\p{^Is_Block:	georgian}', "");
Expect(0, 4351, '\P{Is_Block:	georgian}', "");
Expect(1, 4351, '\P{^Is_Block:	georgian}', "");
Expect(0, 4352, '\p{Is_Block:	georgian}', "");
Expect(1, 4352, '\p{^Is_Block:	georgian}', "");
Expect(1, 4352, '\P{Is_Block:	georgian}', "");
Expect(0, 4352, '\P{^Is_Block:	georgian}', "");
Expect(1, 4351, '\p{Is_Block=-Georgian}', "");
Expect(0, 4351, '\p{^Is_Block=-Georgian}', "");
Expect(0, 4351, '\P{Is_Block=-Georgian}', "");
Expect(1, 4351, '\P{^Is_Block=-Georgian}', "");
Expect(0, 4352, '\p{Is_Block=-Georgian}', "");
Expect(1, 4352, '\p{^Is_Block=-Georgian}', "");
Expect(1, 4352, '\P{Is_Block=-Georgian}', "");
Expect(0, 4352, '\P{^Is_Block=-Georgian}', "");
Error('\p{Is_Blk=-	Georgian:=}');
Error('\P{Is_Blk=-	Georgian:=}');
Expect(1, 4351, '\p{Is_Blk=georgian}', "");
Expect(0, 4351, '\p{^Is_Blk=georgian}', "");
Expect(0, 4351, '\P{Is_Blk=georgian}', "");
Expect(1, 4351, '\P{^Is_Blk=georgian}', "");
Expect(0, 4352, '\p{Is_Blk=georgian}', "");
Expect(1, 4352, '\p{^Is_Blk=georgian}', "");
Expect(1, 4352, '\P{Is_Blk=georgian}', "");
Expect(0, 4352, '\P{^Is_Blk=georgian}', "");
Expect(1, 4351, '\p{Is_Blk=	Georgian}', "");
Expect(0, 4351, '\p{^Is_Blk=	Georgian}', "");
Expect(0, 4351, '\P{Is_Blk=	Georgian}', "");
Expect(1, 4351, '\P{^Is_Blk=	Georgian}', "");
Expect(0, 4352, '\p{Is_Blk=	Georgian}', "");
Expect(1, 4352, '\p{^Is_Blk=	Georgian}', "");
Expect(1, 4352, '\P{Is_Blk=	Georgian}', "");
Expect(0, 4352, '\P{^Is_Blk=	Georgian}', "");
Error('\p{Block=:=-Yi_RADICALS}');
Error('\P{Block=:=-Yi_RADICALS}');
Expect(1, 42191, '\p{Block=yiradicals}', "");
Expect(0, 42191, '\p{^Block=yiradicals}', "");
Expect(0, 42191, '\P{Block=yiradicals}', "");
Expect(1, 42191, '\P{^Block=yiradicals}', "");
Expect(0, 42192, '\p{Block=yiradicals}', "");
Expect(1, 42192, '\p{^Block=yiradicals}', "");
Expect(1, 42192, '\P{Block=yiradicals}', "");
Expect(0, 42192, '\P{^Block=yiradicals}', "");
Expect(1, 42191, '\p{Block: Yi_Radicals}', "");
Expect(0, 42191, '\p{^Block: Yi_Radicals}', "");
Expect(0, 42191, '\P{Block: Yi_Radicals}', "");
Expect(1, 42191, '\P{^Block: Yi_Radicals}', "");
Expect(0, 42192, '\p{Block: Yi_Radicals}', "");
Expect(1, 42192, '\p{^Block: Yi_Radicals}', "");
Expect(1, 42192, '\P{Block: Yi_Radicals}', "");
Expect(0, 42192, '\P{^Block: Yi_Radicals}', "");
Error('\p{Blk=  yi_Radicals/a/}');
Error('\P{Blk=  yi_Radicals/a/}');
Expect(1, 42191, '\p{Blk=yiradicals}', "");
Expect(0, 42191, '\p{^Blk=yiradicals}', "");
Expect(0, 42191, '\P{Blk=yiradicals}', "");
Expect(1, 42191, '\P{^Blk=yiradicals}', "");
Expect(0, 42192, '\p{Blk=yiradicals}', "");
Expect(1, 42192, '\p{^Blk=yiradicals}', "");
Expect(1, 42192, '\P{Blk=yiradicals}', "");
Expect(0, 42192, '\P{^Blk=yiradicals}', "");
Expect(1, 42191, '\p{Blk= Yi_radicals}', "");
Expect(0, 42191, '\p{^Blk= Yi_radicals}', "");
Expect(0, 42191, '\P{Blk= Yi_radicals}', "");
Expect(1, 42191, '\P{^Blk= Yi_radicals}', "");
Expect(0, 42192, '\p{Blk= Yi_radicals}', "");
Expect(1, 42192, '\p{^Blk= Yi_radicals}', "");
Expect(1, 42192, '\P{Blk= Yi_radicals}', "");
Expect(0, 42192, '\P{^Blk= Yi_radicals}', "");
Error('\p{Is_Block:	:=_Yi_Radicals}');
Error('\P{Is_Block:	:=_Yi_Radicals}');
Expect(1, 42191, '\p{Is_Block=yiradicals}', "");
Expect(0, 42191, '\p{^Is_Block=yiradicals}', "");
Expect(0, 42191, '\P{Is_Block=yiradicals}', "");
Expect(1, 42191, '\P{^Is_Block=yiradicals}', "");
Expect(0, 42192, '\p{Is_Block=yiradicals}', "");
Expect(1, 42192, '\p{^Is_Block=yiradicals}', "");
Expect(1, 42192, '\P{Is_Block=yiradicals}', "");
Expect(0, 42192, '\P{^Is_Block=yiradicals}', "");
Expect(1, 42191, '\p{Is_Block= _Yi_Radicals}', "");
Expect(0, 42191, '\p{^Is_Block= _Yi_Radicals}', "");
Expect(0, 42191, '\P{Is_Block= _Yi_Radicals}', "");
Expect(1, 42191, '\P{^Is_Block= _Yi_Radicals}', "");
Expect(0, 42192, '\p{Is_Block= _Yi_Radicals}', "");
Expect(1, 42192, '\p{^Is_Block= _Yi_Radicals}', "");
Expect(1, 42192, '\P{Is_Block= _Yi_Radicals}', "");
Expect(0, 42192, '\P{^Is_Block= _Yi_Radicals}', "");
Error('\p{Is_Blk= -yi_RADICALS:=}');
Error('\P{Is_Blk= -yi_RADICALS:=}');
Expect(1, 42191, '\p{Is_Blk=yiradicals}', "");
Expect(0, 42191, '\p{^Is_Blk=yiradicals}', "");
Expect(0, 42191, '\P{Is_Blk=yiradicals}', "");
Expect(1, 42191, '\P{^Is_Blk=yiradicals}', "");
Expect(0, 42192, '\p{Is_Blk=yiradicals}', "");
Expect(1, 42192, '\p{^Is_Blk=yiradicals}', "");
Expect(1, 42192, '\P{Is_Blk=yiradicals}', "");
Expect(0, 42192, '\P{^Is_Blk=yiradicals}', "");
Expect(1, 42191, '\p{Is_Blk: --yi_Radicals}', "");
Expect(0, 42191, '\p{^Is_Blk: --yi_Radicals}', "");
Expect(0, 42191, '\P{Is_Blk: --yi_Radicals}', "");
Expect(1, 42191, '\P{^Is_Blk: --yi_Radicals}', "");
Expect(0, 42192, '\p{Is_Blk: --yi_Radicals}', "");
Expect(1, 42192, '\p{^Is_Blk: --yi_Radicals}', "");
Expect(1, 42192, '\P{Is_Blk: --yi_Radicals}', "");
Expect(0, 42192, '\P{^Is_Blk: --yi_Radicals}', "");
Error('\p{Grapheme_Extend= :=Yes}');
Error('\P{Grapheme_Extend= :=Yes}');
Expect(1, 917999, '\p{Grapheme_Extend=yes}', "");
Expect(0, 917999, '\p{^Grapheme_Extend=yes}', "");
Expect(0, 917999, '\P{Grapheme_Extend=yes}', "");
Expect(1, 917999, '\P{^Grapheme_Extend=yes}', "");
Expect(0, 918000, '\p{Grapheme_Extend=yes}', "");
Expect(1, 918000, '\p{^Grapheme_Extend=yes}', "");
Expect(1, 918000, '\P{Grapheme_Extend=yes}', "");
Expect(0, 918000, '\P{^Grapheme_Extend=yes}', "");
Expect(1, 917999, '\p{Grapheme_Extend=--Yes}', "");
Expect(0, 917999, '\p{^Grapheme_Extend=--Yes}', "");
Expect(0, 917999, '\P{Grapheme_Extend=--Yes}', "");
Expect(1, 917999, '\P{^Grapheme_Extend=--Yes}', "");
Expect(0, 918000, '\p{Grapheme_Extend=--Yes}', "");
Expect(1, 918000, '\p{^Grapheme_Extend=--Yes}', "");
Expect(1, 918000, '\P{Grapheme_Extend=--Yes}', "");
Expect(0, 918000, '\P{^Grapheme_Extend=--Yes}', "");
Error('\p{Gr_Ext= _Y:=}');
Error('\P{Gr_Ext= _Y:=}');
Expect(1, 917999, '\p{Gr_Ext=y}', "");
Expect(0, 917999, '\p{^Gr_Ext=y}', "");
Expect(0, 917999, '\P{Gr_Ext=y}', "");
Expect(1, 917999, '\P{^Gr_Ext=y}', "");
Expect(0, 918000, '\p{Gr_Ext=y}', "");
Expect(1, 918000, '\p{^Gr_Ext=y}', "");
Expect(1, 918000, '\P{Gr_Ext=y}', "");
Expect(0, 918000, '\P{^Gr_Ext=y}', "");
Expect(1, 917999, '\p{Gr_Ext=_-Y}', "");
Expect(0, 917999, '\p{^Gr_Ext=_-Y}', "");
Expect(0, 917999, '\P{Gr_Ext=_-Y}', "");
Expect(1, 917999, '\P{^Gr_Ext=_-Y}', "");
Expect(0, 918000, '\p{Gr_Ext=_-Y}', "");
Expect(1, 918000, '\p{^Gr_Ext=_-Y}', "");
Expect(1, 918000, '\P{Gr_Ext=_-Y}', "");
Expect(0, 918000, '\P{^Gr_Ext=_-Y}', "");
Error('\p{Is_Grapheme_Extend=-/a/T}');
Error('\P{Is_Grapheme_Extend=-/a/T}');
Expect(1, 917999, '\p{Is_Grapheme_Extend=t}', "");
Expect(0, 917999, '\p{^Is_Grapheme_Extend=t}', "");
Expect(0, 917999, '\P{Is_Grapheme_Extend=t}', "");
Expect(1, 917999, '\P{^Is_Grapheme_Extend=t}', "");
Expect(0, 918000, '\p{Is_Grapheme_Extend=t}', "");
Expect(1, 918000, '\p{^Is_Grapheme_Extend=t}', "");
Expect(1, 918000, '\P{Is_Grapheme_Extend=t}', "");
Expect(0, 918000, '\P{^Is_Grapheme_Extend=t}', "");
Expect(1, 917999, '\p{Is_Grapheme_Extend=-	t}', "");
Expect(0, 917999, '\p{^Is_Grapheme_Extend=-	t}', "");
Expect(0, 917999, '\P{Is_Grapheme_Extend=-	t}', "");
Expect(1, 917999, '\P{^Is_Grapheme_Extend=-	t}', "");
Expect(0, 918000, '\p{Is_Grapheme_Extend=-	t}', "");
Expect(1, 918000, '\p{^Is_Grapheme_Extend=-	t}', "");
Expect(1, 918000, '\P{Is_Grapheme_Extend=-	t}', "");
Expect(0, 918000, '\P{^Is_Grapheme_Extend=-	t}', "");
Error('\p{Is_Gr_Ext= 	True:=}');
Error('\P{Is_Gr_Ext= 	True:=}');
Expect(1, 917999, '\p{Is_Gr_Ext=true}', "");
Expect(0, 917999, '\p{^Is_Gr_Ext=true}', "");
Expect(0, 917999, '\P{Is_Gr_Ext=true}', "");
Expect(1, 917999, '\P{^Is_Gr_Ext=true}', "");
Expect(0, 918000, '\p{Is_Gr_Ext=true}', "");
Expect(1, 918000, '\p{^Is_Gr_Ext=true}', "");
Expect(1, 918000, '\P{Is_Gr_Ext=true}', "");
Expect(0, 918000, '\P{^Is_Gr_Ext=true}', "");
Expect(1, 917999, '\p{Is_Gr_Ext= 	true}', "");
Expect(0, 917999, '\p{^Is_Gr_Ext= 	true}', "");
Expect(0, 917999, '\P{Is_Gr_Ext= 	true}', "");
Expect(1, 917999, '\P{^Is_Gr_Ext= 	true}', "");
Expect(0, 918000, '\p{Is_Gr_Ext= 	true}', "");
Expect(1, 918000, '\p{^Is_Gr_Ext= 	true}', "");
Expect(1, 918000, '\P{Is_Gr_Ext= 	true}', "");
Expect(0, 918000, '\P{^Is_Gr_Ext= 	true}', "");
Error('\p{Grapheme_Extend=  no:=}');
Error('\P{Grapheme_Extend=  no:=}');
Expect(1, 918000, '\p{Grapheme_Extend:   no}', "");
Expect(0, 918000, '\p{^Grapheme_Extend:   no}', "");
Expect(0, 918000, '\P{Grapheme_Extend:   no}', "");
Expect(1, 918000, '\P{^Grapheme_Extend:   no}', "");
Expect(0, 917999, '\p{Grapheme_Extend:   no}', "");
Expect(1, 917999, '\p{^Grapheme_Extend:   no}', "");
Expect(1, 917999, '\P{Grapheme_Extend:   no}', "");
Expect(0, 917999, '\P{^Grapheme_Extend:   no}', "");
Expect(1, 918000, '\p{Grapheme_Extend=	 NO}', "");
Expect(0, 918000, '\p{^Grapheme_Extend=	 NO}', "");
Expect(0, 918000, '\P{Grapheme_Extend=	 NO}', "");
Expect(1, 918000, '\P{^Grapheme_Extend=	 NO}', "");
Expect(0, 917999, '\p{Grapheme_Extend=	 NO}', "");
Expect(1, 917999, '\p{^Grapheme_Extend=	 NO}', "");
Expect(1, 917999, '\P{Grapheme_Extend=	 NO}', "");
Expect(0, 917999, '\P{^Grapheme_Extend=	 NO}', "");
Error('\p{Gr_Ext= N/a/}');
Error('\P{Gr_Ext= N/a/}');
Expect(1, 918000, '\p{Gr_Ext=n}', "");
Expect(0, 918000, '\p{^Gr_Ext=n}', "");
Expect(0, 918000, '\P{Gr_Ext=n}', "");
Expect(1, 918000, '\P{^Gr_Ext=n}', "");
Expect(0, 917999, '\p{Gr_Ext=n}', "");
Expect(1, 917999, '\p{^Gr_Ext=n}', "");
Expect(1, 917999, '\P{Gr_Ext=n}', "");
Expect(0, 917999, '\P{^Gr_Ext=n}', "");
Expect(1, 918000, '\p{Gr_Ext= N}', "");
Expect(0, 918000, '\p{^Gr_Ext= N}', "");
Expect(0, 918000, '\P{Gr_Ext= N}', "");
Expect(1, 918000, '\P{^Gr_Ext= N}', "");
Expect(0, 917999, '\p{Gr_Ext= N}', "");
Expect(1, 917999, '\p{^Gr_Ext= N}', "");
Expect(1, 917999, '\P{Gr_Ext= N}', "");
Expect(0, 917999, '\P{^Gr_Ext= N}', "");
Error('\p{Is_Grapheme_Extend= F/a/}');
Error('\P{Is_Grapheme_Extend= F/a/}');
Expect(1, 918000, '\p{Is_Grapheme_Extend=f}', "");
Expect(0, 918000, '\p{^Is_Grapheme_Extend=f}', "");
Expect(0, 918000, '\P{Is_Grapheme_Extend=f}', "");
Expect(1, 918000, '\P{^Is_Grapheme_Extend=f}', "");
Expect(0, 917999, '\p{Is_Grapheme_Extend=f}', "");
Expect(1, 917999, '\p{^Is_Grapheme_Extend=f}', "");
Expect(1, 917999, '\P{Is_Grapheme_Extend=f}', "");
Expect(0, 917999, '\P{^Is_Grapheme_Extend=f}', "");
Expect(1, 918000, '\p{Is_Grapheme_Extend=_F}', "");
Expect(0, 918000, '\p{^Is_Grapheme_Extend=_F}', "");
Expect(0, 918000, '\P{Is_Grapheme_Extend=_F}', "");
Expect(1, 918000, '\P{^Is_Grapheme_Extend=_F}', "");
Expect(0, 917999, '\p{Is_Grapheme_Extend=_F}', "");
Expect(1, 917999, '\p{^Is_Grapheme_Extend=_F}', "");
Expect(1, 917999, '\P{Is_Grapheme_Extend=_F}', "");
Expect(0, 917999, '\P{^Is_Grapheme_Extend=_F}', "");
Error('\p{Is_Gr_Ext= /a/False}');
Error('\P{Is_Gr_Ext= /a/False}');
Expect(1, 918000, '\p{Is_Gr_Ext=false}', "");
Expect(0, 918000, '\p{^Is_Gr_Ext=false}', "");
Expect(0, 918000, '\P{Is_Gr_Ext=false}', "");
Expect(1, 918000, '\P{^Is_Gr_Ext=false}', "");
Expect(0, 917999, '\p{Is_Gr_Ext=false}', "");
Expect(1, 917999, '\p{^Is_Gr_Ext=false}', "");
Expect(1, 917999, '\P{Is_Gr_Ext=false}', "");
Expect(0, 917999, '\P{^Is_Gr_Ext=false}', "");
Expect(1, 918000, '\p{Is_Gr_Ext=_-False}', "");
Expect(0, 918000, '\p{^Is_Gr_Ext=_-False}', "");
Expect(0, 918000, '\P{Is_Gr_Ext=_-False}', "");
Expect(1, 918000, '\P{^Is_Gr_Ext=_-False}', "");
Expect(0, 917999, '\p{Is_Gr_Ext=_-False}', "");
Expect(1, 917999, '\p{^Is_Gr_Ext=_-False}', "");
Expect(1, 917999, '\P{Is_Gr_Ext=_-False}', "");
Expect(0, 917999, '\P{^Is_Gr_Ext=_-False}', "");
Test_X('÷ 0020 ÷ 0020 ÷');
Test_X('÷ 0020 × 0308 ÷ 0020 ÷');
Test_X('÷ 0020 ÷ 000D ÷');
Test_X('÷ 0020 × 0308 ÷ 000D ÷');
Test_X('÷ 0020 ÷ 000A ÷');
Test_X('÷ 0020 × 0308 ÷ 000A ÷');
Test_X('÷ 0020 ÷ 0001 ÷');
Test_X('÷ 0020 × 0308 ÷ 0001 ÷');
Test_X('÷ 0020 × 0300 ÷');
Test_X('÷ 0020 × 0308 × 0300 ÷');
Test_X('÷ 0020 × 0903 ÷');
Test_X('÷ 0020 × 0308 × 0903 ÷');
Test_X('÷ 0020 ÷ 1100 ÷');
Test_X('÷ 0020 × 0308 ÷ 1100 ÷');
Test_X('÷ 0020 ÷ 1160 ÷');
Test_X('÷ 0020 × 0308 ÷ 1160 ÷');
Test_X('÷ 0020 ÷ 11A8 ÷');
Test_X('÷ 0020 × 0308 ÷ 11A8 ÷');
Test_X('÷ 0020 ÷ AC00 ÷');
Test_X('÷ 0020 × 0308 ÷ AC00 ÷');
Test_X('÷ 0020 ÷ AC01 ÷');
Test_X('÷ 0020 × 0308 ÷ AC01 ÷');
Test_X('÷ 0020 ÷ 0378 ÷');
Test_X('÷ 0020 × 0308 ÷ 0378 ÷');
Test_X('÷ 0020 ÷ D800 ÷');
Test_X('÷ 0020 × 0308 ÷ D800 ÷');
Test_X('÷ 000D ÷ 0020 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 0020 ÷');
Test_X('÷ 000D ÷ 000D ÷');
Test_X('÷ 000D ÷ 0308 ÷ 000D ÷');
Test_X('÷ 000D × 000A ÷');
Test_X('÷ 000D ÷ 0308 ÷ 000A ÷');
Test_X('÷ 000D ÷ 0001 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 0001 ÷');
Test_X('÷ 000D ÷ 0300 ÷');
Test_X('÷ 000D ÷ 0308 × 0300 ÷');
Test_X('÷ 000D ÷ 0903 ÷');
Test_X('÷ 000D ÷ 0308 × 0903 ÷');
Test_X('÷ 000D ÷ 1100 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 1100 ÷');
Test_X('÷ 000D ÷ 1160 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 1160 ÷');
Test_X('÷ 000D ÷ 11A8 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 11A8 ÷');
Test_X('÷ 000D ÷ AC00 ÷');
Test_X('÷ 000D ÷ 0308 ÷ AC00 ÷');
Test_X('÷ 000D ÷ AC01 ÷');
Test_X('÷ 000D ÷ 0308 ÷ AC01 ÷');
Test_X('÷ 000D ÷ 0378 ÷');
Test_X('÷ 000D ÷ 0308 ÷ 0378 ÷');
Test_X('÷ 000D ÷ D800 ÷');
Test_X('÷ 000D ÷ 0308 ÷ D800 ÷');
Test_X('÷ 000A ÷ 0020 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 0020 ÷');
Test_X('÷ 000A ÷ 000D ÷');
Test_X('÷ 000A ÷ 0308 ÷ 000D ÷');
Test_X('÷ 000A ÷ 000A ÷');
Test_X('÷ 000A ÷ 0308 ÷ 000A ÷');
Test_X('÷ 000A ÷ 0001 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 0001 ÷');
Test_X('÷ 000A ÷ 0300 ÷');
Test_X('÷ 000A ÷ 0308 × 0300 ÷');
Test_X('÷ 000A ÷ 0903 ÷');
Test_X('÷ 000A ÷ 0308 × 0903 ÷');
Test_X('÷ 000A ÷ 1100 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 1100 ÷');
Test_X('÷ 000A ÷ 1160 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 1160 ÷');
Test_X('÷ 000A ÷ 11A8 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 11A8 ÷');
Test_X('÷ 000A ÷ AC00 ÷');
Test_X('÷ 000A ÷ 0308 ÷ AC00 ÷');
Test_X('÷ 000A ÷ AC01 ÷');
Test_X('÷ 000A ÷ 0308 ÷ AC01 ÷');
Test_X('÷ 000A ÷ 0378 ÷');
Test_X('÷ 000A ÷ 0308 ÷ 0378 ÷');
Test_X('÷ 000A ÷ D800 ÷');
Test_X('÷ 000A ÷ 0308 ÷ D800 ÷');
Test_X('÷ 0001 ÷ 0020 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 0020 ÷');
Test_X('÷ 0001 ÷ 000D ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 000D ÷');
Test_X('÷ 0001 ÷ 000A ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 000A ÷');
Test_X('÷ 0001 ÷ 0001 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 0001 ÷');
Test_X('÷ 0001 ÷ 0300 ÷');
Test_X('÷ 0001 ÷ 0308 × 0300 ÷');
Test_X('÷ 0001 ÷ 0903 ÷');
Test_X('÷ 0001 ÷ 0308 × 0903 ÷');
Test_X('÷ 0001 ÷ 1100 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 1100 ÷');
Test_X('÷ 0001 ÷ 1160 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 1160 ÷');
Test_X('÷ 0001 ÷ 11A8 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 11A8 ÷');
Test_X('÷ 0001 ÷ AC00 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ AC00 ÷');
Test_X('÷ 0001 ÷ AC01 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ AC01 ÷');
Test_X('÷ 0001 ÷ 0378 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ 0378 ÷');
Test_X('÷ 0001 ÷ D800 ÷');
Test_X('÷ 0001 ÷ 0308 ÷ D800 ÷');
Test_X('÷ 0300 ÷ 0020 ÷');
Test_X('÷ 0300 × 0308 ÷ 0020 ÷');
Test_X('÷ 0300 ÷ 000D ÷');
Test_X('÷ 0300 × 0308 ÷ 000D ÷');
Test_X('÷ 0300 ÷ 000A ÷');
Test_X('÷ 0300 × 0308 ÷ 000A ÷');
Test_X('÷ 0300 ÷ 0001 ÷');
Test_X('÷ 0300 × 0308 ÷ 0001 ÷');
Test_X('÷ 0300 × 0300 ÷');
Test_X('÷ 0300 × 0308 × 0300 ÷');
Test_X('÷ 0300 × 0903 ÷');
Test_X('÷ 0300 × 0308 × 0903 ÷');
Test_X('÷ 0300 ÷ 1100 ÷');
Test_X('÷ 0300 × 0308 ÷ 1100 ÷');
Test_X('÷ 0300 ÷ 1160 ÷');
Test_X('÷ 0300 × 0308 ÷ 1160 ÷');
Test_X('÷ 0300 ÷ 11A8 ÷');
Test_X('÷ 0300 × 0308 ÷ 11A8 ÷');
Test_X('÷ 0300 ÷ AC00 ÷');
Test_X('÷ 0300 × 0308 ÷ AC00 ÷');
Test_X('÷ 0300 ÷ AC01 ÷');
Test_X('÷ 0300 × 0308 ÷ AC01 ÷');
Test_X('÷ 0300 ÷ 0378 ÷');
Test_X('÷ 0300 × 0308 ÷ 0378 ÷');
Test_X('÷ 0300 ÷ D800 ÷');
Test_X('÷ 0300 × 0308 ÷ D800 ÷');
Test_X('÷ 0903 ÷ 0020 ÷');
Test_X('÷ 0903 × 0308 ÷ 0020 ÷');
Test_X('÷ 0903 ÷ 000D ÷');
Test_X('÷ 0903 × 0308 ÷ 000D ÷');
Test_X('÷ 0903 ÷ 000A ÷');
Test_X('÷ 0903 × 0308 ÷ 000A ÷');
Test_X('÷ 0903 ÷ 0001 ÷');
Test_X('÷ 0903 × 0308 ÷ 0001 ÷');
Test_X('÷ 0903 × 0300 ÷');
Test_X('÷ 0903 × 0308 × 0300 ÷');
Test_X('÷ 0903 × 0903 ÷');
Test_X('÷ 0903 × 0308 × 0903 ÷');
Test_X('÷ 0903 ÷ 1100 ÷');
Test_X('÷ 0903 × 0308 ÷ 1100 ÷');
Test_X('÷ 0903 ÷ 1160 ÷');
Test_X('÷ 0903 × 0308 ÷ 1160 ÷');
Test_X('÷ 0903 ÷ 11A8 ÷');
Test_X('÷ 0903 × 0308 ÷ 11A8 ÷');
Test_X('÷ 0903 ÷ AC00 ÷');
Test_X('÷ 0903 × 0308 ÷ AC00 ÷');
Test_X('÷ 0903 ÷ AC01 ÷');
Test_X('÷ 0903 × 0308 ÷ AC01 ÷');
Test_X('÷ 0903 ÷ 0378 ÷');
Test_X('÷ 0903 × 0308 ÷ 0378 ÷');
Test_X('÷ 0903 ÷ D800 ÷');
Test_X('÷ 0903 × 0308 ÷ D800 ÷');
Test_X('÷ 1100 ÷ 0020 ÷');
Test_X('÷ 1100 × 0308 ÷ 0020 ÷');
Test_X('÷ 1100 ÷ 000D ÷');
Test_X('÷ 1100 × 0308 ÷ 000D ÷');
Test_X('÷ 1100 ÷ 000A ÷');
Test_X('÷ 1100 × 0308 ÷ 000A ÷');
Test_X('÷ 1100 ÷ 0001 ÷');
Test_X('÷ 1100 × 0308 ÷ 0001 ÷');
Test_X('÷ 1100 × 0300 ÷');
Test_X('÷ 1100 × 0308 × 0300 ÷');
Test_X('÷ 1100 × 0903 ÷');
Test_X('÷ 1100 × 0308 × 0903 ÷');
Test_X('÷ 1100 × 1100 ÷');
Test_X('÷ 1100 × 0308 ÷ 1100 ÷');
Test_X('÷ 1100 × 1160 ÷');
Test_X('÷ 1100 × 0308 ÷ 1160 ÷');
Test_X('÷ 1100 ÷ 11A8 ÷');
Test_X('÷ 1100 × 0308 ÷ 11A8 ÷');
Test_X('÷ 1100 × AC00 ÷');
Test_X('÷ 1100 × 0308 ÷ AC00 ÷');
Test_X('÷ 1100 × AC01 ÷');
Test_X('÷ 1100 × 0308 ÷ AC01 ÷');
Test_X('÷ 1100 ÷ 0378 ÷');
Test_X('÷ 1100 × 0308 ÷ 0378 ÷');
Test_X('÷ 1100 ÷ D800 ÷');
Test_X('÷ 1100 × 0308 ÷ D800 ÷');
Test_X('÷ 1160 ÷ 0020 ÷');
Test_X('÷ 1160 × 0308 ÷ 0020 ÷');
Test_X('÷ 1160 ÷ 000D ÷');
Test_X('÷ 1160 × 0308 ÷ 000D ÷');
Test_X('÷ 1160 ÷ 000A ÷');
Test_X('÷ 1160 × 0308 ÷ 000A ÷');
Test_X('÷ 1160 ÷ 0001 ÷');
Test_X('÷ 1160 × 0308 ÷ 0001 ÷');
Test_X('÷ 1160 × 0300 ÷');
Test_X('÷ 1160 × 0308 × 0300 ÷');
Test_X('÷ 1160 × 0903 ÷');
Test_X('÷ 1160 × 0308 × 0903 ÷');
Test_X('÷ 1160 ÷ 1100 ÷');
Test_X('÷ 1160 × 0308 ÷ 1100 ÷');
Test_X('÷ 1160 × 1160 ÷');
Test_X('÷ 1160 × 0308 ÷ 1160 ÷');
Test_X('÷ 1160 × 11A8 ÷');
Test_X('÷ 1160 × 0308 ÷ 11A8 ÷');
Test_X('÷ 1160 ÷ AC00 ÷');
Test_X('÷ 1160 × 0308 ÷ AC00 ÷');
Test_X('÷ 1160 ÷ AC01 ÷');
Test_X('÷ 1160 × 0308 ÷ AC01 ÷');
Test_X('÷ 1160 ÷ 0378 ÷');
Test_X('÷ 1160 × 0308 ÷ 0378 ÷');
Test_X('÷ 1160 ÷ D800 ÷');
Test_X('÷ 1160 × 0308 ÷ D800 ÷');
Test_X('÷ 11A8 ÷ 0020 ÷');
Test_X('÷ 11A8 × 0308 ÷ 0020 ÷');
Test_X('÷ 11A8 ÷ 000D ÷');
Test_X('÷ 11A8 × 0308 ÷ 000D ÷');
Test_X('÷ 11A8 ÷ 000A ÷');
Test_X('÷ 11A8 × 0308 ÷ 000A ÷');
Test_X('÷ 11A8 ÷ 0001 ÷');
Test_X('÷ 11A8 × 0308 ÷ 0001 ÷');
Test_X('÷ 11A8 × 0300 ÷');
Test_X('÷ 11A8 × 0308 × 0300 ÷');
Test_X('÷ 11A8 × 0903 ÷');
Test_X('÷ 11A8 × 0308 × 0903 ÷');
Test_X('÷ 11A8 ÷ 1100 ÷');
Test_X('÷ 11A8 × 0308 ÷ 1100 ÷');
Test_X('÷ 11A8 ÷ 1160 ÷');
Test_X('÷ 11A8 × 0308 ÷ 1160 ÷');
Test_X('÷ 11A8 × 11A8 ÷');
Test_X('÷ 11A8 × 0308 ÷ 11A8 ÷');
Test_X('÷ 11A8 ÷ AC00 ÷');
Test_X('÷ 11A8 × 0308 ÷ AC00 ÷');
Test_X('÷ 11A8 ÷ AC01 ÷');
Test_X('÷ 11A8 × 0308 ÷ AC01 ÷');
Test_X('÷ 11A8 ÷ 0378 ÷');
Test_X('÷ 11A8 × 0308 ÷ 0378 ÷');
Test_X('÷ 11A8 ÷ D800 ÷');
Test_X('÷ 11A8 × 0308 ÷ D800 ÷');
Test_X('÷ AC00 ÷ 0020 ÷');
Test_X('÷ AC00 × 0308 ÷ 0020 ÷');
Test_X('÷ AC00 ÷ 000D ÷');
Test_X('÷ AC00 × 0308 ÷ 000D ÷');
Test_X('÷ AC00 ÷ 000A ÷');
Test_X('÷ AC00 × 0308 ÷ 000A ÷');
Test_X('÷ AC00 ÷ 0001 ÷');
Test_X('÷ AC00 × 0308 ÷ 0001 ÷');
Test_X('÷ AC00 × 0300 ÷');
Test_X('÷ AC00 × 0308 × 0300 ÷');
Test_X('÷ AC00 × 0903 ÷');
Test_X('÷ AC00 × 0308 × 0903 ÷');
Test_X('÷ AC00 ÷ 1100 ÷');
Test_X('÷ AC00 × 0308 ÷ 1100 ÷');
Test_X('÷ AC00 × 1160 ÷');
Test_X('÷ AC00 × 0308 ÷ 1160 ÷');
Test_X('÷ AC00 × 11A8 ÷');
Test_X('÷ AC00 × 0308 ÷ 11A8 ÷');
Test_X('÷ AC00 ÷ AC00 ÷');
Test_X('÷ AC00 × 0308 ÷ AC00 ÷');
Test_X('÷ AC00 ÷ AC01 ÷');
Test_X('÷ AC00 × 0308 ÷ AC01 ÷');
Test_X('÷ AC00 ÷ 0378 ÷');
Test_X('÷ AC00 × 0308 ÷ 0378 ÷');
Test_X('÷ AC00 ÷ D800 ÷');
Test_X('÷ AC00 × 0308 ÷ D800 ÷');
Test_X('÷ AC01 ÷ 0020 ÷');
Test_X('÷ AC01 × 0308 ÷ 0020 ÷');
Test_X('÷ AC01 ÷ 000D ÷');
Test_X('÷ AC01 × 0308 ÷ 000D ÷');
Test_X('÷ AC01 ÷ 000A ÷');
Test_X('÷ AC01 × 0308 ÷ 000A ÷');
Test_X('÷ AC01 ÷ 0001 ÷');
Test_X('÷ AC01 × 0308 ÷ 0001 ÷');
Test_X('÷ AC01 × 0300 ÷');
Test_X('÷ AC01 × 0308 × 0300 ÷');
Test_X('÷ AC01 × 0903 ÷');
Test_X('÷ AC01 × 0308 × 0903 ÷');
Test_X('÷ AC01 ÷ 1100 ÷');
Test_X('÷ AC01 × 0308 ÷ 1100 ÷');
Test_X('÷ AC01 ÷ 1160 ÷');
Test_X('÷ AC01 × 0308 ÷ 1160 ÷');
Test_X('÷ AC01 × 11A8 ÷');
Test_X('÷ AC01 × 0308 ÷ 11A8 ÷');
Test_X('÷ AC01 ÷ AC00 ÷');
Test_X('÷ AC01 × 0308 ÷ AC00 ÷');
Test_X('÷ AC01 ÷ AC01 ÷');
Test_X('÷ AC01 × 0308 ÷ AC01 ÷');
Test_X('÷ AC01 ÷ 0378 ÷');
Test_X('÷ AC01 × 0308 ÷ 0378 ÷');
Test_X('÷ AC01 ÷ D800 ÷');
Test_X('÷ AC01 × 0308 ÷ D800 ÷');
Test_X('÷ 0378 ÷ 0020 ÷');
Test_X('÷ 0378 × 0308 ÷ 0020 ÷');
Test_X('÷ 0378 ÷ 000D ÷');
Test_X('÷ 0378 × 0308 ÷ 000D ÷');
Test_X('÷ 0378 ÷ 000A ÷');
Test_X('÷ 0378 × 0308 ÷ 000A ÷');
Test_X('÷ 0378 ÷ 0001 ÷');
Test_X('÷ 0378 × 0308 ÷ 0001 ÷');
Test_X('÷ 0378 × 0300 ÷');
Test_X('÷ 0378 × 0308 × 0300 ÷');
Test_X('÷ 0378 × 0903 ÷');
Test_X('÷ 0378 × 0308 × 0903 ÷');
Test_X('÷ 0378 ÷ 1100 ÷');
Test_X('÷ 0378 × 0308 ÷ 1100 ÷');
Test_X('÷ 0378 ÷ 1160 ÷');
Test_X('÷ 0378 × 0308 ÷ 1160 ÷');
Test_X('÷ 0378 ÷ 11A8 ÷');
Test_X('÷ 0378 × 0308 ÷ 11A8 ÷');
Test_X('÷ 0378 ÷ AC00 ÷');
Test_X('÷ 0378 × 0308 ÷ AC00 ÷');
Test_X('÷ 0378 ÷ AC01 ÷');
Test_X('÷ 0378 × 0308 ÷ AC01 ÷');
Test_X('÷ 0378 ÷ 0378 ÷');
Test_X('÷ 0378 × 0308 ÷ 0378 ÷');
Test_X('÷ 0378 ÷ D800 ÷');
Test_X('÷ 0378 × 0308 ÷ D800 ÷');
Test_X('÷ D800 ÷ 0020 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 0020 ÷');
Test_X('÷ D800 ÷ 000D ÷');
Test_X('÷ D800 ÷ 0308 ÷ 000D ÷');
Test_X('÷ D800 ÷ 000A ÷');
Test_X('÷ D800 ÷ 0308 ÷ 000A ÷');
Test_X('÷ D800 ÷ 0001 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 0001 ÷');
Test_X('÷ D800 ÷ 0300 ÷');
Test_X('÷ D800 ÷ 0308 × 0300 ÷');
Test_X('÷ D800 ÷ 0903 ÷');
Test_X('÷ D800 ÷ 0308 × 0903 ÷');
Test_X('÷ D800 ÷ 1100 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 1100 ÷');
Test_X('÷ D800 ÷ 1160 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 1160 ÷');
Test_X('÷ D800 ÷ 11A8 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 11A8 ÷');
Test_X('÷ D800 ÷ AC00 ÷');
Test_X('÷ D800 ÷ 0308 ÷ AC00 ÷');
Test_X('÷ D800 ÷ AC01 ÷');
Test_X('÷ D800 ÷ 0308 ÷ AC01 ÷');
Test_X('÷ D800 ÷ 0378 ÷');
Test_X('÷ D800 ÷ 0308 ÷ 0378 ÷');
Test_X('÷ D800 ÷ D800 ÷');
Test_X('÷ D800 ÷ 0308 ÷ D800 ÷');
Finished();
